#include "Problem.h"
#include "Functions.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <stack>
#include "hdf5.h"
#include <gsl/gsl_rng.h>
#include <random>
#include "SAMRAI/pdat/CellVariable.h"
#include "SAMRAI/pdat/NodeData.h"
#include "SAMRAI/pdat/NodeVariable.h"
#include "LagrangianPolynomicRefine.h"


#include "SAMRAI/tbox/Array.h"
#include "SAMRAI/hier/BoundaryBox.h"
#include "SAMRAI/hier/BoxContainer.h"
#include "SAMRAI/geom/CartesianPatchGeometry.h"
#include "SAMRAI/hier/VariableDatabase.h"
#include "SAMRAI/hier/PatchDataRestartManager.h"
#include "SAMRAI/hier/Index.h"
#include "SAMRAI/tbox/PIO.h"
#include "SAMRAI/tbox/Utilities.h"
#include "SAMRAI/tbox/Timer.h"
#include "SAMRAI/tbox/TimerManager.h"

#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))
#define SIGN(X) (((X) > 0) - ((X) < 0))
#define isEven(a) ((a) % 2 == 0 ? true : false)
#define greaterEq(a,b) ((fabs((a) - (b))/1.0E-15 > 10 ? false : (floor(fabs((a) - (b))/1.0E-15) < 1)) || (b)<(a))
#define lessEq(a,b) ((fabs((a) - (b))/1.0E-15 > 10 ? false: (floor(fabs((a) - (b))/1.0E-15) < 1)) || (a)<(b))
#define equalsEq(a,b) ((fabs((a) - (b))/1.0E-15 > 10 ? false: (floor(fabs((a) - (b))/1.0E-15) < 1)))
#define reducePrecision(x, p) (floor(((x) * pow(10, (p)) + 0.5)) / pow(10, (p)))

using namespace external;

const gsl_rng_type * T;
gsl_rng *r_var;

//Timers
std::shared_ptr<tbox::Timer> t_step;
std::shared_ptr<tbox::Timer> t_moveParticles;
std::shared_ptr<tbox::Timer> t_output;

inline int Problem::GetExpoBase2(double d)
{
	int i = 0;
	((short *)(&i))[0] = (((short *)(&d))[3] & (short)32752); // _123456789ab____ & 0111111111110000
	return (i >> 4) - 1023;
}

bool	Problem::Equals(double d1, double d2)
{
	if (d1 == d2)
		return true;
	int e1 = GetExpoBase2(d1);
	int e2 = GetExpoBase2(d2);
	int e3 = GetExpoBase2(d1 - d2);
	if ((e3 - e2 < -48) && (e3 - e1 < -48))
		return true;
	return false;
}

/*
 * Constructor of the problem.
 */
Problem::Problem(const string& object_name, const tbox::Dimension& dim, std::shared_ptr<tbox::Database>& database, std::shared_ptr<geom::CartesianGridGeometry >& grid_geom, std::shared_ptr<hier::PatchHierarchy >& patch_hierarchy, MainRestartData& mrd, const double dt, const bool init_from_restart, const int console_output, const int timer_output, const int mesh_output_period, const vector<string> full_mesh_writer_variables, std::shared_ptr<appu::VisItDataWriter>& mesh_data_writer, const vector<int> slicer_output_period, const vector<set<string> > sliceVariables, vector<std::shared_ptr<SlicerDataWriter > >sliceWriters, const vector<int> sphere_output_period, const vector<set<string> > sphereVariables, vector<std::shared_ptr<SphereDataWriter > > sphereWriters, const vector<bool> slicer_analysis_output, const vector<bool> sphere_analysis_output, const vector<int> integration_output_period, const vector<set<string> > integralVariables, vector<std::shared_ptr<IntegrateDataWriter > > integrateDataWriters, const vector<bool> integration_analysis_output, const vector<int> point_output_period, const vector<set<string> > pointVariables, vector<std::shared_ptr<PointDataWriter > > pointDataWriters, const vector<bool> point_analysis_output): 
d_dim(dim), xfer::RefinePatchStrategy(), xfer::CoarsenPatchStrategy(), viz_mesh_dump_interval(mesh_output_period), d_full_mesh_writer_variables(full_mesh_writer_variables.begin(), full_mesh_writer_variables.end()), d_visit_data_writer(mesh_data_writer), d_output_interval(console_output), d_timer_output_interval(timer_output), d_sliceWriters(sliceWriters.begin(), sliceWriters.end()), d_sphereWriters(sphereWriters.begin(), sphereWriters.end()), d_slicer_output_period(slicer_output_period.begin(), slicer_output_period.end()), d_sphere_output_period(sphere_output_period.begin(), sphere_output_period.end()), analysis_slice_dump(slicer_analysis_output.begin(), slicer_analysis_output.end()), analysis_sphere_dump(sphere_analysis_output.begin(), sphere_analysis_output.end()), d_integrateDataWriters(integrateDataWriters.begin(), integrateDataWriters.end()), d_integration_output_period(integration_output_period.begin(), integration_output_period.end()), analysis_integration_dump(integration_analysis_output.begin(), integration_analysis_output.end()), d_pointDataWriters(pointDataWriters.begin(), pointDataWriters.end()), d_point_output_period(point_output_period.begin(), point_output_period.end()), analysis_point_dump(point_analysis_output.begin(), point_analysis_output.end())
{
	//Setup the timers
	t_step = tbox::TimerManager::getManager()->getTimer("Step");
	t_moveParticles = tbox::TimerManager::getManager()->getTimer("Move particles");
	t_output = tbox::TimerManager::getManager()->getTimer("OutputGeneration");

	//Output configuration
	next_console_output = d_output_interval;
	next_timer_output = d_timer_output_interval;

	//Get the object name, the grid geometry and the patch hierarchy
  	d_grid_geometry = grid_geom;
	d_patch_hierarchy = patch_hierarchy;
	d_object_name = object_name;
	d_init_from_restart = init_from_restart;
	initial_dt = dt;

	for (vector<set<string> >::const_iterator it = sliceVariables.begin(); it != sliceVariables.end(); ++it) {
		set<string> vars = *it;
		for (set<string>::const_iterator it2 = vars.begin() ; it2 != vars.end(); ++it2) {
			d_sliceVariables.push_back(vars);
		}
	}
	for (vector<set<string> >::const_iterator it = sphereVariables.begin(); it != sphereVariables.end(); ++it) {
		set<string> vars = *it;
		for (set<string>::const_iterator it2 = vars.begin() ; it2 != vars.end(); ++it2) {
			d_sphereVariables.push_back(vars);
		}
	}
	for (vector<set<string> >::const_iterator it = integralVariables.begin(); it != integralVariables.end(); ++it) {
		set<string> vars = *it;
		for (set<string>::const_iterator it2 = vars.begin() ; it2 != vars.end(); ++it2) {
			d_integralVariables.push_back(vars);
		}
	}
	for (vector<set<string> >::const_iterator it = pointVariables.begin(); it != pointVariables.end(); ++it) {
		set<string> vars = *it;
		for (set<string>::const_iterator it2 = vars.begin() ; it2 != vars.end(); ++it2) {
			d_pointVariables.push_back(vars);
		}
	}


	//Get parameters
    cout<<"Reading parameters"<<endl;
	minTableTemperature = database->getDouble("minTableTemperature");
	dissipation_factor_Atd_xx = database->getDouble("dissipation_factor_Atd_xx");
	dissipation_factor_Atd_xy = database->getDouble("dissipation_factor_Atd_xy");
	dissipation_factor_gtd_zz = database->getDouble("dissipation_factor_gtd_zz");
	dissipation_factor_Atd_xz = database->getDouble("dissipation_factor_Atd_xz");
	dissipation_factor_Gamh_y = database->getDouble("dissipation_factor_Gamh_y");
	dissipation_factor_Gamh_x = database->getDouble("dissipation_factor_Gamh_x");
	Atd_xx_falloff = database->getDouble("Atd_xx_falloff");
	dissipation_factor_Gamh_z = database->getDouble("dissipation_factor_Gamh_z");
	p_kappa_z2 = database->getDouble("p_kappa_z2");
	p_kappa_z1 = database->getDouble("p_kappa_z1");
	calculate_les_terms = database->getInteger("calculate_les_terms");
	Gamh_z_asymptotic = database->getDouble("Gamh_z_asymptotic");
	vacuum_tau_reset = database->getDouble("vacuum_tau_reset");
	dissipation_factor_Sfd_z = database->getDouble("dissipation_factor_Sfd_z");
	gtd_yz_asymptotic = database->getDouble("gtd_yz_asymptotic");
	dissipation_factor_Sfd_x = database->getDouble("dissipation_factor_Sfd_x");
	rho_2 = database->getDouble("rho_2");
	dissipation_factor_Sfd_y = database->getDouble("dissipation_factor_Sfd_y");
	rho_1 = database->getDouble("rho_1");
	rho_0 = database->getDouble("rho_0");
	gtd_xz_asymptotic = database->getDouble("gtd_xz_asymptotic");
	Betau_y_falloff = database->getDouble("Betau_y_falloff");
	gtd_xz_falloff = database->getDouble("gtd_xz_falloff");
	lambda_f1 = database->getDouble("lambda_f1");
	Betau_z_0 = database->getDouble("Betau_z_0");
	lambda_f0 = database->getDouble("lambda_f0");
	lambda_f3 = database->getDouble("lambda_f3");
	lambda_f2 = database->getDouble("lambda_f2");
	minTableEnergy = database->getDouble("minTableEnergy");
	Atd_zz_falloff = database->getDouble("Atd_zz_falloff");
	Gamh_z_falloff = database->getDouble("Gamh_z_falloff");
	max_error = database->getDouble("max_error");
	trK_falloff = database->getDouble("trK_falloff");
	energyShift = database->getDouble("energyShift");
	chi_falloff = database->getDouble("chi_falloff");
	dissipation_factor_tauf = database->getDouble("dissipation_factor_tauf");
	dissipation_factor_theta = database->getDouble("dissipation_factor_theta");
	gtd_yz_falloff = database->getDouble("gtd_yz_falloff");
	K_1 = database->getDouble("K_1");
	p_kappa_f = database->getDouble("p_kappa_f");
	K_0 = database->getDouble("K_0");
	K_3 = database->getDouble("K_3");
	K_2 = database->getDouble("K_2");
	gtd_xx_asymptotic = database->getDouble("gtd_xx_asymptotic");
	Betau_y_asymptotic = database->getDouble("Betau_y_asymptotic");
	dissipation_factor_Atd_zz = database->getDouble("dissipation_factor_Atd_zz");
	threshold_leakage_vacuum = database->getDouble("threshold_leakage_vacuum");
	epsilon_main = database->getDouble("epsilon_main");
	alpha_main = database->getDouble("alpha_main");
	dissipation_factor_Bfu_z = database->getDouble("dissipation_factor_Bfu_z");
	vacuum_ye_beta = database->getDouble("vacuum_ye_beta");
	dissipation_factor_Betau_z = database->getDouble("dissipation_factor_Betau_z");
	dissipation_factor_Bfu_x = database->getDouble("dissipation_factor_Bfu_x");
	dissipation_factor_Betau_y = database->getDouble("dissipation_factor_Betau_y");
	dissipation_factor_Bfu_y = database->getDouble("dissipation_factor_Bfu_y");
	dissipation_factor_Betau_x = database->getDouble("dissipation_factor_Betau_x");
	dissipation_factor_Alpha = database->getDouble("dissipation_factor_Alpha");
	gtd_zz_falloff = database->getDouble("gtd_zz_falloff");
	vacuum_rho_reset = database->getDouble("vacuum_rho_reset");
	theta_asymptotic = database->getDouble("theta_asymptotic");
	C_M = database->getDouble("C_M");
	p_feta = database->getDouble("p_feta");
	C_N = database->getDouble("C_N");
	initialYe = database->getDouble("initialYe");
	C_T = database->getDouble("C_T");
	Gamh_x_falloff = database->getDouble("Gamh_x_falloff");
	eta_damping_exp = database->getDouble("eta_damping_exp");
	Gamh_y_asymptotic = database->getDouble("Gamh_y_asymptotic");
	Betau_x_falloff = database->getDouble("Betau_x_falloff");
	dissipation_factor_Atd_yy = database->getDouble("dissipation_factor_Atd_yy");
	ch = database->getDouble("ch");
	dissipation_factor_Atd_yz = database->getDouble("dissipation_factor_Atd_yz");
	gtd_xy_falloff = database->getDouble("gtd_xy_falloff");
	gtd_zz_asymptotic = database->getDouble("gtd_zz_asymptotic");
	Alpha_asymptotic = database->getDouble("Alpha_asymptotic");
	threshold_vacuum = database->getDouble("threshold_vacuum");
	tov_facp = database->getDouble("tov_facp");
	vacuum_temp_reset = database->getDouble("vacuum_temp_reset");
	ye_maximum = database->getDouble("ye_maximum");
	Betau_x_asymptotic = database->getDouble("Betau_x_asymptotic");
	tov_Asize = database->getDouble("tov_Asize");
	Atd_yy_falloff = database->getDouble("Atd_yy_falloff");
	lambda_2 = database->getDouble("lambda_2");
	trK0 = database->getDouble("trK0");
	lambda_1 = database->getDouble("lambda_1");
	lambda_0 = database->getDouble("lambda_0");
	Gamh_x_asymptotic = database->getDouble("Gamh_x_asymptotic");
	gamma = database->getDouble("gamma");
	R_0 = database->getDouble("R_0");
	Atd_xy_asymptotic = database->getDouble("Atd_xy_asymptotic");
	tov_angle = database->getDouble("tov_angle");
	trK_asymptotic = database->getDouble("trK_asymptotic");
	vacuum_rho = database->getDouble("vacuum_rho");
	Atd_yy_asymptotic = database->getDouble("Atd_yy_asymptotic");
	threshold_sqBmax = database->getDouble("threshold_sqBmax");
	gtd_yy_asymptotic = database->getDouble("gtd_yy_asymptotic");
	p_kappa_cc = database->getDouble("p_kappa_cc");
	Betau_z_asymptotic = database->getDouble("Betau_z_asymptotic");
	Atd_xz_falloff = database->getDouble("Atd_xz_falloff");
	theta_falloff = database->getDouble("theta_falloff");
	vacuum_D = database->getDouble("vacuum_D");
	initialTemperature = database->getDouble("initialTemperature");
	vacuum_ye_reset = database->getDouble("vacuum_ye_reset");
	gtd_xx_falloff = database->getDouble("gtd_xx_falloff");
	externalCon2Prim = database->getInteger("externalCon2Prim");
	gtd_xy_asymptotic = database->getDouble("gtd_xy_asymptotic");
	threshold_Dles_max = database->getDouble("threshold_Dles_max");
	dissipation_factor_chi = database->getDouble("dissipation_factor_chi");
	Betau_x_0 = database->getDouble("Betau_x_0");
	Atd_yz_falloff = database->getDouble("Atd_yz_falloff");
	gamma_2 = database->getDouble("gamma_2");
	dissipation_factor_gtd_xz = database->getDouble("dissipation_factor_gtd_xz");
	gamma_3 = database->getDouble("gamma_3");
	gamma_0 = database->getDouble("gamma_0");
	dissipation_factor_phif = database->getDouble("dissipation_factor_phif");
	gamma_1 = database->getDouble("gamma_1");
	threshold_sqSmax = database->getDouble("threshold_sqSmax");
	dissipation_factor_Df = database->getDouble("dissipation_factor_Df");
	dissipation_factor_gtd_xx = database->getDouble("dissipation_factor_gtd_xx");
	dissipation_factor_gtd_xy = database->getDouble("dissipation_factor_gtd_xy");
	Atd_xy_falloff = database->getDouble("Atd_xy_falloff");
	dissipation_factor_trK = database->getDouble("dissipation_factor_trK");
	tend = database->getDouble("tend");
	a_1 = database->getDouble("a_1");
	Atd_xx_asymptotic = database->getDouble("Atd_xx_asymptotic");
	a_0 = database->getDouble("a_0");
	a_3 = database->getDouble("a_3");
	a_2 = database->getDouble("a_2");
	threshold_Dles_min = database->getDouble("threshold_Dles_min");
	Betau_y_0 = database->getDouble("Betau_y_0");
	vacuum_ye = database->getDouble("vacuum_ye");
	dissipation_factor_gtd_yy = database->getDouble("dissipation_factor_gtd_yy");
	do_leakage = database->getInteger("do_leakage");
	Gamh_y_falloff = database->getDouble("Gamh_y_falloff");
	dissipation_factor_gtd_yz = database->getDouble("dissipation_factor_gtd_yz");
	eos_type = database->getInteger("eos_type");
	Atd_zz_asymptotic = database->getDouble("Atd_zz_asymptotic");
	dissipation_factor_DYf = database->getDouble("dissipation_factor_DYf");
	Alpha_falloff = database->getDouble("Alpha_falloff");
	Betau_z_falloff = database->getDouble("Betau_z_falloff");
	chi_asymptotic = database->getDouble("chi_asymptotic");
	Atd_xz_asymptotic = database->getDouble("Atd_xz_asymptotic");
	Atd_yz_asymptotic = database->getDouble("Atd_yz_asymptotic");
	vacuum_P_reset = database->getDouble("vacuum_P_reset");
	vacuum_tau = database->getDouble("vacuum_tau");
	chi_floor = database->getDouble("chi_floor");
	gtd_yy_falloff = database->getDouble("gtd_yy_falloff");
	//Random initialization
	gsl_rng_env_setup();
	//Random for simulation
	const tbox::SAMRAI_MPI& mpi(tbox::SAMRAI_MPI::getSAMRAIWorld());
	int random_seed = database->getDouble("random_seed")*(mpi.getRank() + 1);
	r_var = gsl_rng_alloc(gsl_rng_ranlxs0);
	gsl_rng_set(r_var, random_seed);
	for (int il = 0; il < d_patch_hierarchy->getMaxNumberOfLevels(); il++) {
		bo_substep_iteration.push_back(0);
	}
	//Initialization from input file
	if (!d_init_from_restart) {
		next_mesh_dump_iteration = viz_mesh_dump_interval;
		for (std::vector<int>::iterator it = d_slicer_output_period.begin(); it != d_slicer_output_period.end(); ++it) {
			next_slice_dump_iteration.push_back((*it));
		}
		for (std::vector<int>::iterator it = d_sphere_output_period.begin(); it != d_sphere_output_period.end(); ++it) {
			next_sphere_dump_iteration.push_back((*it));
		}
		for (std::vector<int>::iterator it = d_integration_output_period.begin(); it != d_integration_output_period.end(); ++it) {
			next_integration_dump_iteration.push_back((*it));
		}
		for (std::vector<int>::iterator it = d_point_output_period.begin(); it != d_point_output_period.end(); ++it) {
			next_point_dump_iteration.push_back((*it));
		}

		//Iteration counter
		for (int il = 0; il < d_patch_hierarchy->getMaxNumberOfLevels(); il++) {
			current_iteration.push_back(0);
		}
	}
	//Initialization from restart file
	else {
		getFromRestart(mrd);
		if (d_slicer_output_period.size() < next_slice_dump_iteration.size()) {
			TBOX_ERROR("Number of slices cannot be reduced after a checkpoint.");
		}
		for (int il = 0; il < d_slicer_output_period.size(); il++) {
			if (il >= next_slice_dump_iteration.size()) {
				next_slice_dump_iteration.push_back(0);
			}
			if (next_slice_dump_iteration[il] == 0 && d_slicer_output_period[il] > 0) {
				next_slice_dump_iteration[il] = current_iteration[d_patch_hierarchy->getNumberOfLevels() - 1] + d_slicer_output_period[il];
			}
		}
		if (d_sphere_output_period.size() < next_sphere_dump_iteration.size()) {
			TBOX_ERROR("Number of spheres cannot be reduced after a checkpoint.");
		}
		for (int il = 0; il < d_sphere_output_period.size(); il++) {
			if (il >= next_sphere_dump_iteration.size()) {
				next_sphere_dump_iteration.push_back(0);
			}
			if (next_sphere_dump_iteration[il] == 0 && d_sphere_output_period[il] > 0) {
				next_sphere_dump_iteration[il] = current_iteration[d_patch_hierarchy->getNumberOfLevels() - 1] + d_sphere_output_period[il];
			}
		}
		if (d_integration_output_period.size() < next_integration_dump_iteration.size()) {
			TBOX_ERROR("Number of integrations cannot be reduced after a checkpoint.");
		}
		for (int il = 0; il < d_integration_output_period.size(); il++) {
			if (il >= next_integration_dump_iteration.size()) {
				next_integration_dump_iteration.push_back(0);
			}
			if (next_integration_dump_iteration[il] == 0 && d_integration_output_period[il] > 0) {
				next_integration_dump_iteration[il] = current_iteration[d_patch_hierarchy->getNumberOfLevels() - 1] + d_integration_output_period[il];
			}
		}
		if (d_point_output_period.size() < next_point_dump_iteration.size()) {
			TBOX_ERROR("Number of point output cannot be reduced after a checkpoint.");
		}
		for (int il = 0; il < d_point_output_period.size(); il++) {
			if (il >= next_point_dump_iteration.size()) {
				next_point_dump_iteration.push_back(0);
			}
			if (next_point_dump_iteration[il] == 0 && d_point_output_period[il] > 0) {
				next_point_dump_iteration[il] = current_iteration[d_patch_hierarchy->getNumberOfLevels() - 1] + d_point_output_period[il];
			}
		}
	}


	//External eos parameters
#ifdef EXTERNAL_EOS

	std::shared_ptr<tbox::Database> external_eos_db = database->getDatabase("external_EOS");
	Commons::ExternalEos::reprimand_eos_type = external_eos_db->getInteger("eos_type");
	Commons::ExternalEos::reprimand_atmo_Ye = external_eos_db->getDouble("atmo_Ye");
	Commons::ExternalEos::reprimand_max_z = external_eos_db->getDouble("max_z");
	Commons::ExternalEos::reprimand_max_b = external_eos_db->getDouble("max_b");
	Commons::ExternalEos::reprimand_c2p_acc = external_eos_db->getDouble("c2p_acc");
	Commons::ExternalEos::reprimand_atmo_rho = external_eos_db->getDouble("atmo_rho");
	Commons::ExternalEos::reprimand_rho_strict = external_eos_db->getDouble("rho_strict");
	Commons::ExternalEos::reprimand_max_rho = external_eos_db->getDouble("max_rho");
	Commons::ExternalEos::reprimand_max_eps = external_eos_db->getDouble("max_eps");
	Commons::ExternalEos::reprimand_gamma_th = external_eos_db->getDouble("gamma_th");
#endif

    	//Subcycling
	d_refinedTimeStepping = false;
	if (database->isString("subcycling")) {
		if (database->getString("subcycling") == "BERGER-OLIGER") {
			d_refinedTimeStepping = true;
		}
	}


	//Regridding options
	d_regridding = false;
	if (database->isDatabase("regridding")) {
		regridding_db = database->getDatabase("regridding");
		d_regridding_buffer = regridding_db->getDouble("regridding_buffer");
		int smallest_patch_size = d_patch_hierarchy->getSmallestPatchSize(0).min();
		for (int il = 1; il < d_patch_hierarchy->getMaxNumberOfLevels(); il++) {
			smallest_patch_size = MIN(smallest_patch_size, d_patch_hierarchy->getSmallestPatchSize(il).min());
		}
		if (d_regridding_buffer > smallest_patch_size) {
			TBOX_ERROR("Error: Regridding_buffer parameter ("<<d_regridding_buffer<<") cannot be greater than smallest_patch_size minimum value("<<smallest_patch_size<<")");
		}
		if (regridding_db->isString("regridding_type")) {
			d_regridding_type = regridding_db->getString("regridding_type");
			d_regridding_min_level = regridding_db->getInteger("regridding_min_level");
			d_regridding_max_level = regridding_db->getInteger("regridding_max_level");
			if (d_regridding_type == "GRADIENT") {
				d_regridding_field = regridding_db->getString("regridding_field");
				d_regridding_compressionFactor = regridding_db->getDouble("regridding_compressionFactor");
				d_regridding_mOffset = regridding_db->getDouble("regridding_mOffset");
				d_regridding = true;
			} else {
				if (d_regridding_type == "FUNCTION") {
					d_regridding_field = regridding_db->getString("regridding_function_field");
					d_regridding_threshold = regridding_db->getDouble("regridding_threshold");
					d_regridding = true;
				} else {
					if (d_regridding_type == "SHADOW") {
						std::string* fields = new std::string[2];
						regridding_db->getStringArray("regridding_fields", fields, 2);
						d_regridding_field = fields[0];
						d_regridding_field_shadow = fields[1];
						d_regridding_error = regridding_db->getDouble("regridding_error");
						d_regridding = true;
						delete[] fields;
					}
				}
			}
		}
	}

	//Stencil of the discretization method
	int maxratio = 1;
	for (int il = 1; il < d_patch_hierarchy->getMaxNumberOfLevels(); il++) {
		const hier::IntVector ratio = d_patch_hierarchy->getRatioToCoarserLevel(il);
		maxratio = MAX(maxratio, ratio.max());
	}
	//Minimum region thickness
	d_regionMinThickness = 3;
	d_ghost_width = 3;

	
	//Register Fields and temporal fields into the variable database
	hier::VariableDatabase *vdb = hier::VariableDatabase::getDatabase();
  	std::shared_ptr<hier::VariableContext> d_cont_curr(vdb->getContext("Current"));
	std::shared_ptr< pdat::CellVariable<int> > mask(std::shared_ptr< pdat::CellVariable<int> >(new pdat::CellVariable<int>(d_dim, "samrai_mask",1)));
	d_mask_id = vdb->registerVariableAndContext(mask ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	IntegrateDataWriter::setMaskVariable(d_mask_id);
	std::shared_ptr< pdat::NodeVariable<double> > interior(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "regridding_value",1)));
	d_interior_regridding_value_id = vdb->registerVariableAndContext(interior ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<int> > nonSync(std::shared_ptr< pdat::NodeVariable<int> >(new pdat::NodeVariable<int>(d_dim, "regridding_tag",1)));
	d_nonSync_regridding_tag_id = vdb->registerVariableAndContext(nonSync ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > interior_i(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "interior_i",1)));
	d_interior_i_id = vdb->registerVariableAndContext(interior_i ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > interior_j(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "interior_j",1)));
	d_interior_j_id = vdb->registerVariableAndContext(interior_j ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > interior_k(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "interior_k",1)));
	d_interior_k_id = vdb->registerVariableAndContext(interior_k ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > FOV_1(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "FOV_1",1)));
	d_FOV_1_id = vdb->registerVariableAndContext(FOV_1 ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_FOV_1_id);
	std::shared_ptr< pdat::NodeVariable<double> > FOV_xLower(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "FOV_xLower",1)));
	d_FOV_xLower_id = vdb->registerVariableAndContext(FOV_xLower ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_FOV_xLower_id);
	std::shared_ptr< pdat::NodeVariable<double> > FOV_xUpper(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "FOV_xUpper",1)));
	d_FOV_xUpper_id = vdb->registerVariableAndContext(FOV_xUpper ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_FOV_xUpper_id);
	std::shared_ptr< pdat::NodeVariable<double> > FOV_yLower(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "FOV_yLower",1)));
	d_FOV_yLower_id = vdb->registerVariableAndContext(FOV_yLower ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_FOV_yLower_id);
	std::shared_ptr< pdat::NodeVariable<double> > FOV_yUpper(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "FOV_yUpper",1)));
	d_FOV_yUpper_id = vdb->registerVariableAndContext(FOV_yUpper ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_FOV_yUpper_id);
	std::shared_ptr< pdat::NodeVariable<double> > FOV_zLower(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "FOV_zLower",1)));
	d_FOV_zLower_id = vdb->registerVariableAndContext(FOV_zLower ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_FOV_zLower_id);
	std::shared_ptr< pdat::NodeVariable<double> > FOV_zUpper(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "FOV_zUpper",1)));
	d_FOV_zUpper_id = vdb->registerVariableAndContext(FOV_zUpper ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_FOV_zUpper_id);
	std::shared_ptr< pdat::NodeVariable<double> > Sfd_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Sfd_x",1)));
	d_Sfd_x_id = vdb->registerVariableAndContext(Sfd_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Sfd_x_id);
	std::shared_ptr< pdat::NodeVariable<double> > Sfd_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Sfd_y",1)));
	d_Sfd_y_id = vdb->registerVariableAndContext(Sfd_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Sfd_y_id);
	std::shared_ptr< pdat::NodeVariable<double> > Sfd_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Sfd_z",1)));
	d_Sfd_z_id = vdb->registerVariableAndContext(Sfd_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Sfd_z_id);
	std::shared_ptr< pdat::NodeVariable<double> > Bfu_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Bfu_x",1)));
	d_Bfu_x_id = vdb->registerVariableAndContext(Bfu_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Bfu_x_id);
	std::shared_ptr< pdat::NodeVariable<double> > Bfu_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Bfu_y",1)));
	d_Bfu_y_id = vdb->registerVariableAndContext(Bfu_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Bfu_y_id);
	std::shared_ptr< pdat::NodeVariable<double> > Bfu_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Bfu_z",1)));
	d_Bfu_z_id = vdb->registerVariableAndContext(Bfu_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Bfu_z_id);
	std::shared_ptr< pdat::NodeVariable<double> > Df(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Df",1)));
	d_Df_id = vdb->registerVariableAndContext(Df ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Df_id);
	std::shared_ptr< pdat::NodeVariable<double> > DYf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "DYf",1)));
	d_DYf_id = vdb->registerVariableAndContext(DYf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_DYf_id);
	std::shared_ptr< pdat::NodeVariable<double> > tauf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "tauf",1)));
	d_tauf_id = vdb->registerVariableAndContext(tauf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_tauf_id);
	std::shared_ptr< pdat::NodeVariable<double> > phif(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "phif",1)));
	d_phif_id = vdb->registerVariableAndContext(phif ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_phif_id);
	std::shared_ptr< pdat::NodeVariable<double> > gtd_xx(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "gtd_xx",1)));
	d_gtd_xx_id = vdb->registerVariableAndContext(gtd_xx ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_gtd_xx_id);
	std::shared_ptr< pdat::NodeVariable<double> > gtd_xy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "gtd_xy",1)));
	d_gtd_xy_id = vdb->registerVariableAndContext(gtd_xy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_gtd_xy_id);
	std::shared_ptr< pdat::NodeVariable<double> > gtd_xz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "gtd_xz",1)));
	d_gtd_xz_id = vdb->registerVariableAndContext(gtd_xz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_gtd_xz_id);
	std::shared_ptr< pdat::NodeVariable<double> > gtd_yy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "gtd_yy",1)));
	d_gtd_yy_id = vdb->registerVariableAndContext(gtd_yy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_gtd_yy_id);
	std::shared_ptr< pdat::NodeVariable<double> > gtd_yz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "gtd_yz",1)));
	d_gtd_yz_id = vdb->registerVariableAndContext(gtd_yz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_gtd_yz_id);
	std::shared_ptr< pdat::NodeVariable<double> > gtd_zz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "gtd_zz",1)));
	d_gtd_zz_id = vdb->registerVariableAndContext(gtd_zz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_gtd_zz_id);
	std::shared_ptr< pdat::NodeVariable<double> > Atd_xx(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Atd_xx",1)));
	d_Atd_xx_id = vdb->registerVariableAndContext(Atd_xx ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Atd_xx_id);
	std::shared_ptr< pdat::NodeVariable<double> > Atd_xy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Atd_xy",1)));
	d_Atd_xy_id = vdb->registerVariableAndContext(Atd_xy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Atd_xy_id);
	std::shared_ptr< pdat::NodeVariable<double> > Atd_xz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Atd_xz",1)));
	d_Atd_xz_id = vdb->registerVariableAndContext(Atd_xz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Atd_xz_id);
	std::shared_ptr< pdat::NodeVariable<double> > Atd_yy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Atd_yy",1)));
	d_Atd_yy_id = vdb->registerVariableAndContext(Atd_yy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Atd_yy_id);
	std::shared_ptr< pdat::NodeVariable<double> > Atd_yz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Atd_yz",1)));
	d_Atd_yz_id = vdb->registerVariableAndContext(Atd_yz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Atd_yz_id);
	std::shared_ptr< pdat::NodeVariable<double> > Atd_zz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Atd_zz",1)));
	d_Atd_zz_id = vdb->registerVariableAndContext(Atd_zz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Atd_zz_id);
	std::shared_ptr< pdat::NodeVariable<double> > Gamh_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Gamh_x",1)));
	d_Gamh_x_id = vdb->registerVariableAndContext(Gamh_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Gamh_x_id);
	std::shared_ptr< pdat::NodeVariable<double> > Gamh_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Gamh_y",1)));
	d_Gamh_y_id = vdb->registerVariableAndContext(Gamh_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Gamh_y_id);
	std::shared_ptr< pdat::NodeVariable<double> > Gamh_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Gamh_z",1)));
	d_Gamh_z_id = vdb->registerVariableAndContext(Gamh_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Gamh_z_id);
	std::shared_ptr< pdat::NodeVariable<double> > Betau_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Betau_x",1)));
	d_Betau_x_id = vdb->registerVariableAndContext(Betau_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Betau_x_id);
	std::shared_ptr< pdat::NodeVariable<double> > Betau_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Betau_y",1)));
	d_Betau_y_id = vdb->registerVariableAndContext(Betau_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Betau_y_id);
	std::shared_ptr< pdat::NodeVariable<double> > Betau_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Betau_z",1)));
	d_Betau_z_id = vdb->registerVariableAndContext(Betau_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Betau_z_id);
	std::shared_ptr< pdat::NodeVariable<double> > Alpha(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Alpha",1)));
	d_Alpha_id = vdb->registerVariableAndContext(Alpha ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Alpha_id);
	std::shared_ptr< pdat::NodeVariable<double> > chi(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "chi",1)));
	d_chi_id = vdb->registerVariableAndContext(chi ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_chi_id);
	std::shared_ptr< pdat::NodeVariable<double> > trK(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "trK",1)));
	d_trK_id = vdb->registerVariableAndContext(trK ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_trK_id);
	std::shared_ptr< pdat::NodeVariable<double> > theta(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "theta",1)));
	d_theta_id = vdb->registerVariableAndContext(theta ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_theta_id);
	std::shared_ptr< pdat::NodeVariable<double> > rhof(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rhof",1)));
	d_rhof_id = vdb->registerVariableAndContext(rhof ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_rhof_id);
	std::shared_ptr< pdat::NodeVariable<double> > Yef(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Yef",1)));
	d_Yef_id = vdb->registerVariableAndContext(Yef ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Yef_id);
	std::shared_ptr< pdat::NodeVariable<double> > vfd_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "vfd_x",1)));
	d_vfd_x_id = vdb->registerVariableAndContext(vfd_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_vfd_x_id);
	std::shared_ptr< pdat::NodeVariable<double> > vfd_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "vfd_y",1)));
	d_vfd_y_id = vdb->registerVariableAndContext(vfd_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_vfd_y_id);
	std::shared_ptr< pdat::NodeVariable<double> > vfd_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "vfd_z",1)));
	d_vfd_z_id = vdb->registerVariableAndContext(vfd_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_vfd_z_id);
	std::shared_ptr< pdat::NodeVariable<double> > pf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "pf",1)));
	d_pf_id = vdb->registerVariableAndContext(pf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_pf_id);
	std::shared_ptr< pdat::NodeVariable<double> > epsf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "epsf",1)));
	d_epsf_id = vdb->registerVariableAndContext(epsf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_epsf_id);
	std::shared_ptr< pdat::NodeVariable<double> > Tf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Tf",1)));
	d_Tf_id = vdb->registerVariableAndContext(Tf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Tf_id);
	std::shared_ptr< pdat::NodeVariable<double> > sqcs(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "sqcs",1)));
	d_sqcs_id = vdb->registerVariableAndContext(sqcs ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_sqcs_id);
	std::shared_ptr< pdat::NodeVariable<double> > qnu(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "qnu",1)));
	d_qnu_id = vdb->registerVariableAndContext(qnu ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_qnu_id);
	std::shared_ptr< pdat::NodeVariable<double> > rnu(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rnu",1)));
	d_rnu_id = vdb->registerVariableAndContext(rnu ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_rnu_id);
	std::shared_ptr< pdat::NodeVariable<double> > optdepthe(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "optdepthe",1)));
	d_optdepthe_id = vdb->registerVariableAndContext(optdepthe ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_optdepthe_id);
	std::shared_ptr< pdat::NodeVariable<double> > optdeptha(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "optdeptha",1)));
	d_optdeptha_id = vdb->registerVariableAndContext(optdeptha ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_optdeptha_id);
	std::shared_ptr< pdat::NodeVariable<double> > optdepthx(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "optdepthx",1)));
	d_optdepthx_id = vdb->registerVariableAndContext(optdepthx ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_optdepthx_id);
	std::shared_ptr< pdat::NodeVariable<double> > chie(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "chie",1)));
	d_chie_id = vdb->registerVariableAndContext(chie ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_chie_id);
	std::shared_ptr< pdat::NodeVariable<double> > chia(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "chia",1)));
	d_chia_id = vdb->registerVariableAndContext(chia ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_chia_id);
	std::shared_ptr< pdat::NodeVariable<double> > chix(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "chix",1)));
	d_chix_id = vdb->registerVariableAndContext(chix ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_chix_id);
	std::shared_ptr< pdat::NodeVariable<double> > t_optdepthe(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "t_optdepthe",1)));
	d_t_optdepthe_id = vdb->registerVariableAndContext(t_optdepthe ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_t_optdepthe_id);
	std::shared_ptr< pdat::NodeVariable<double> > t_optdeptha(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "t_optdeptha",1)));
	d_t_optdeptha_id = vdb->registerVariableAndContext(t_optdeptha ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_t_optdeptha_id);
	std::shared_ptr< pdat::NodeVariable<double> > t_optdepthx(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "t_optdepthx",1)));
	d_t_optdepthx_id = vdb->registerVariableAndContext(t_optdepthx ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_t_optdepthx_id);
	std::shared_ptr< pdat::NodeVariable<double> > t_chie(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "t_chie",1)));
	d_t_chie_id = vdb->registerVariableAndContext(t_chie ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_t_chie_id);
	std::shared_ptr< pdat::NodeVariable<double> > t_chia(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "t_chia",1)));
	d_t_chia_id = vdb->registerVariableAndContext(t_chia ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_t_chia_id);
	std::shared_ptr< pdat::NodeVariable<double> > t_chix(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "t_chix",1)));
	d_t_chix_id = vdb->registerVariableAndContext(t_chix ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_t_chix_id);
	std::shared_ptr< pdat::NodeVariable<double> > Efu_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Efu_x",1)));
	d_Efu_x_id = vdb->registerVariableAndContext(Efu_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Efu_x_id);
	std::shared_ptr< pdat::NodeVariable<double> > Efu_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Efu_y",1)));
	d_Efu_y_id = vdb->registerVariableAndContext(Efu_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Efu_y_id);
	std::shared_ptr< pdat::NodeVariable<double> > Efu_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Efu_z",1)));
	d_Efu_z_id = vdb->registerVariableAndContext(Efu_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_Efu_z_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauN_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauN_x",1)));
	d_TauN_x_id = vdb->registerVariableAndContext(TauN_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauN_x_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauN_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauN_y",1)));
	d_TauN_y_id = vdb->registerVariableAndContext(TauN_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauN_y_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauN_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauN_z",1)));
	d_TauN_z_id = vdb->registerVariableAndContext(TauN_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauN_z_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauNe_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauNe_x",1)));
	d_TauNe_x_id = vdb->registerVariableAndContext(TauNe_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauNe_x_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauNe_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauNe_y",1)));
	d_TauNe_y_id = vdb->registerVariableAndContext(TauNe_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauNe_y_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauNe_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauNe_z",1)));
	d_TauNe_z_id = vdb->registerVariableAndContext(TauNe_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauNe_z_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauM_xy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauM_xy",1)));
	d_TauM_xy_id = vdb->registerVariableAndContext(TauM_xy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauM_xy_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauM_xz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauM_xz",1)));
	d_TauM_xz_id = vdb->registerVariableAndContext(TauM_xz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauM_xz_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauM_yz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauM_yz",1)));
	d_TauM_yz_id = vdb->registerVariableAndContext(TauM_yz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauM_yz_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauT_xx(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauT_xx",1)));
	d_TauT_xx_id = vdb->registerVariableAndContext(TauT_xx ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauT_xx_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauT_xy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauT_xy",1)));
	d_TauT_xy_id = vdb->registerVariableAndContext(TauT_xy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauT_xy_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauT_xz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauT_xz",1)));
	d_TauT_xz_id = vdb->registerVariableAndContext(TauT_xz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauT_xz_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauT_yy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauT_yy",1)));
	d_TauT_yy_id = vdb->registerVariableAndContext(TauT_yy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauT_yy_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauT_yz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauT_yz",1)));
	d_TauT_yz_id = vdb->registerVariableAndContext(TauT_yz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauT_yz_id);
	std::shared_ptr< pdat::NodeVariable<double> > TauT_zz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "TauT_zz",1)));
	d_TauT_zz_id = vdb->registerVariableAndContext(TauT_zz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_TauT_zz_id);
	std::shared_ptr< pdat::NodeVariable<double> > dpfdeps(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "dpfdeps",1)));
	d_dpfdeps_id = vdb->registerVariableAndContext(dpfdeps ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_dpfdeps_id);
	std::shared_ptr< pdat::NodeVariable<double> > dpfdrho(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "dpfdrho",1)));
	d_dpfdrho_id = vdb->registerVariableAndContext(dpfdrho ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_dpfdrho_id);
	std::shared_ptr< pdat::NodeVariable<double> > dpfdye(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "dpfdye",1)));
	d_dpfdye_id = vdb->registerVariableAndContext(dpfdye ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_dpfdye_id);
	std::shared_ptr< pdat::NodeVariable<double> > qnu_a(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "qnu_a",1)));
	d_qnu_a_id = vdb->registerVariableAndContext(qnu_a ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_qnu_a_id);
	std::shared_ptr< pdat::NodeVariable<double> > qnu_e(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "qnu_e",1)));
	d_qnu_e_id = vdb->registerVariableAndContext(qnu_e ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_qnu_e_id);
	std::shared_ptr< pdat::NodeVariable<double> > qnu_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "qnu_x",1)));
	d_qnu_x_id = vdb->registerVariableAndContext(qnu_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_qnu_x_id);
	std::shared_ptr< pdat::NodeVariable<double> > Rscalar(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Rscalar",1)));
	d_Rscalar_id = vdb->registerVariableAndContext(Rscalar ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > HamCon(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "HamCon",1)));
	d_HamCon_id = vdb->registerVariableAndContext(HamCon ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > MomCon_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "MomCon_x",1)));
	d_MomCon_x_id = vdb->registerVariableAndContext(MomCon_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > MomCon_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "MomCon_y",1)));
	d_MomCon_y_id = vdb->registerVariableAndContext(MomCon_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > MomCon_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "MomCon_z",1)));
	d_MomCon_z_id = vdb->registerVariableAndContext(MomCon_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > trA(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "trA",1)));
	d_trA_id = vdb->registerVariableAndContext(trA ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > detgtm1(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "detgtm1",1)));
	d_detgtm1_id = vdb->registerVariableAndContext(detgtm1 ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > M_ADM_surf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "M_ADM_surf",1)));
	d_M_ADM_surf_id = vdb->registerVariableAndContext(M_ADM_surf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Jz_ADM_surf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Jz_ADM_surf",1)));
	d_Jz_ADM_surf_id = vdb->registerVariableAndContext(Jz_ADM_surf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > M_Komar(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "M_Komar",1)));
	d_M_Komar_id = vdb->registerVariableAndContext(M_Komar ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Jz_Komar(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Jz_Komar",1)));
	d_Jz_Komar_id = vdb->registerVariableAndContext(Jz_Komar ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > psi4R(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "psi4R",1)));
	d_psi4R_id = vdb->registerVariableAndContext(psi4R ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > psi4I(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "psi4I",1)));
	d_psi4I_id = vdb->registerVariableAndContext(psi4I ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Z_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Z_x",1)));
	d_Z_x_id = vdb->registerVariableAndContext(Z_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Z_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Z_y",1)));
	d_Z_y_id = vdb->registerVariableAndContext(Z_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Z_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Z_z",1)));
	d_Z_z_id = vdb->registerVariableAndContext(Z_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > flux_mass(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "flux_mass",1)));
	d_flux_mass_id = vdb->registerVariableAndContext(flux_mass ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Sfd_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Sfd_x",1)));
	d_rk1Sfd_x_id = vdb->registerVariableAndContext(rk1Sfd_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Sfd_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Sfd_y",1)));
	d_rk1Sfd_y_id = vdb->registerVariableAndContext(rk1Sfd_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Sfd_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Sfd_z",1)));
	d_rk1Sfd_z_id = vdb->registerVariableAndContext(rk1Sfd_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Bfu_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Bfu_x",1)));
	d_rk1Bfu_x_id = vdb->registerVariableAndContext(rk1Bfu_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Bfu_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Bfu_y",1)));
	d_rk1Bfu_y_id = vdb->registerVariableAndContext(rk1Bfu_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Bfu_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Bfu_z",1)));
	d_rk1Bfu_z_id = vdb->registerVariableAndContext(rk1Bfu_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Df(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Df",1)));
	d_rk1Df_id = vdb->registerVariableAndContext(rk1Df ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1DYf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1DYf",1)));
	d_rk1DYf_id = vdb->registerVariableAndContext(rk1DYf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1tauf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1tauf",1)));
	d_rk1tauf_id = vdb->registerVariableAndContext(rk1tauf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1phif(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1phif",1)));
	d_rk1phif_id = vdb->registerVariableAndContext(rk1phif ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1gtd_xx(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1gtd_xx",1)));
	d_rk1gtd_xx_id = vdb->registerVariableAndContext(rk1gtd_xx ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1gtd_xy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1gtd_xy",1)));
	d_rk1gtd_xy_id = vdb->registerVariableAndContext(rk1gtd_xy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1gtd_xz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1gtd_xz",1)));
	d_rk1gtd_xz_id = vdb->registerVariableAndContext(rk1gtd_xz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1gtd_yy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1gtd_yy",1)));
	d_rk1gtd_yy_id = vdb->registerVariableAndContext(rk1gtd_yy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1gtd_yz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1gtd_yz",1)));
	d_rk1gtd_yz_id = vdb->registerVariableAndContext(rk1gtd_yz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1gtd_zz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1gtd_zz",1)));
	d_rk1gtd_zz_id = vdb->registerVariableAndContext(rk1gtd_zz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Atd_xx(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Atd_xx",1)));
	d_rk1Atd_xx_id = vdb->registerVariableAndContext(rk1Atd_xx ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Atd_xy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Atd_xy",1)));
	d_rk1Atd_xy_id = vdb->registerVariableAndContext(rk1Atd_xy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Atd_xz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Atd_xz",1)));
	d_rk1Atd_xz_id = vdb->registerVariableAndContext(rk1Atd_xz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Atd_yy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Atd_yy",1)));
	d_rk1Atd_yy_id = vdb->registerVariableAndContext(rk1Atd_yy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Atd_yz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Atd_yz",1)));
	d_rk1Atd_yz_id = vdb->registerVariableAndContext(rk1Atd_yz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Atd_zz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Atd_zz",1)));
	d_rk1Atd_zz_id = vdb->registerVariableAndContext(rk1Atd_zz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Gamh_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Gamh_x",1)));
	d_rk1Gamh_x_id = vdb->registerVariableAndContext(rk1Gamh_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Gamh_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Gamh_y",1)));
	d_rk1Gamh_y_id = vdb->registerVariableAndContext(rk1Gamh_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Gamh_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Gamh_z",1)));
	d_rk1Gamh_z_id = vdb->registerVariableAndContext(rk1Gamh_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Betau_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Betau_x",1)));
	d_rk1Betau_x_id = vdb->registerVariableAndContext(rk1Betau_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Betau_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Betau_y",1)));
	d_rk1Betau_y_id = vdb->registerVariableAndContext(rk1Betau_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Betau_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Betau_z",1)));
	d_rk1Betau_z_id = vdb->registerVariableAndContext(rk1Betau_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1Alpha(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1Alpha",1)));
	d_rk1Alpha_id = vdb->registerVariableAndContext(rk1Alpha ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1chi(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1chi",1)));
	d_rk1chi_id = vdb->registerVariableAndContext(rk1chi ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1trK(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1trK",1)));
	d_rk1trK_id = vdb->registerVariableAndContext(rk1trK ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk1theta(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk1theta",1)));
	d_rk1theta_id = vdb->registerVariableAndContext(rk1theta ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Sfd_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Sfd_x",1)));
	d_rk2Sfd_x_id = vdb->registerVariableAndContext(rk2Sfd_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Sfd_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Sfd_y",1)));
	d_rk2Sfd_y_id = vdb->registerVariableAndContext(rk2Sfd_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Sfd_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Sfd_z",1)));
	d_rk2Sfd_z_id = vdb->registerVariableAndContext(rk2Sfd_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Bfu_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Bfu_x",1)));
	d_rk2Bfu_x_id = vdb->registerVariableAndContext(rk2Bfu_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Bfu_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Bfu_y",1)));
	d_rk2Bfu_y_id = vdb->registerVariableAndContext(rk2Bfu_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Bfu_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Bfu_z",1)));
	d_rk2Bfu_z_id = vdb->registerVariableAndContext(rk2Bfu_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Df(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Df",1)));
	d_rk2Df_id = vdb->registerVariableAndContext(rk2Df ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2DYf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2DYf",1)));
	d_rk2DYf_id = vdb->registerVariableAndContext(rk2DYf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2tauf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2tauf",1)));
	d_rk2tauf_id = vdb->registerVariableAndContext(rk2tauf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2phif(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2phif",1)));
	d_rk2phif_id = vdb->registerVariableAndContext(rk2phif ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2gtd_xx(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2gtd_xx",1)));
	d_rk2gtd_xx_id = vdb->registerVariableAndContext(rk2gtd_xx ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2gtd_xy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2gtd_xy",1)));
	d_rk2gtd_xy_id = vdb->registerVariableAndContext(rk2gtd_xy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2gtd_xz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2gtd_xz",1)));
	d_rk2gtd_xz_id = vdb->registerVariableAndContext(rk2gtd_xz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2gtd_yy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2gtd_yy",1)));
	d_rk2gtd_yy_id = vdb->registerVariableAndContext(rk2gtd_yy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2gtd_yz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2gtd_yz",1)));
	d_rk2gtd_yz_id = vdb->registerVariableAndContext(rk2gtd_yz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2gtd_zz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2gtd_zz",1)));
	d_rk2gtd_zz_id = vdb->registerVariableAndContext(rk2gtd_zz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Atd_xx(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Atd_xx",1)));
	d_rk2Atd_xx_id = vdb->registerVariableAndContext(rk2Atd_xx ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Atd_xy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Atd_xy",1)));
	d_rk2Atd_xy_id = vdb->registerVariableAndContext(rk2Atd_xy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Atd_xz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Atd_xz",1)));
	d_rk2Atd_xz_id = vdb->registerVariableAndContext(rk2Atd_xz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Atd_yy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Atd_yy",1)));
	d_rk2Atd_yy_id = vdb->registerVariableAndContext(rk2Atd_yy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Atd_yz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Atd_yz",1)));
	d_rk2Atd_yz_id = vdb->registerVariableAndContext(rk2Atd_yz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Atd_zz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Atd_zz",1)));
	d_rk2Atd_zz_id = vdb->registerVariableAndContext(rk2Atd_zz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Gamh_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Gamh_x",1)));
	d_rk2Gamh_x_id = vdb->registerVariableAndContext(rk2Gamh_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Gamh_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Gamh_y",1)));
	d_rk2Gamh_y_id = vdb->registerVariableAndContext(rk2Gamh_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Gamh_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Gamh_z",1)));
	d_rk2Gamh_z_id = vdb->registerVariableAndContext(rk2Gamh_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Betau_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Betau_x",1)));
	d_rk2Betau_x_id = vdb->registerVariableAndContext(rk2Betau_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Betau_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Betau_y",1)));
	d_rk2Betau_y_id = vdb->registerVariableAndContext(rk2Betau_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Betau_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Betau_z",1)));
	d_rk2Betau_z_id = vdb->registerVariableAndContext(rk2Betau_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2Alpha(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2Alpha",1)));
	d_rk2Alpha_id = vdb->registerVariableAndContext(rk2Alpha ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2chi(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2chi",1)));
	d_rk2chi_id = vdb->registerVariableAndContext(rk2chi ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2trK(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2trK",1)));
	d_rk2trK_id = vdb->registerVariableAndContext(rk2trK ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk2theta(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk2theta",1)));
	d_rk2theta_id = vdb->registerVariableAndContext(rk2theta ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Sfd_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Sfd_x",1)));
	d_rk3Sfd_x_id = vdb->registerVariableAndContext(rk3Sfd_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Sfd_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Sfd_y",1)));
	d_rk3Sfd_y_id = vdb->registerVariableAndContext(rk3Sfd_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Sfd_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Sfd_z",1)));
	d_rk3Sfd_z_id = vdb->registerVariableAndContext(rk3Sfd_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Bfu_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Bfu_x",1)));
	d_rk3Bfu_x_id = vdb->registerVariableAndContext(rk3Bfu_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Bfu_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Bfu_y",1)));
	d_rk3Bfu_y_id = vdb->registerVariableAndContext(rk3Bfu_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Bfu_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Bfu_z",1)));
	d_rk3Bfu_z_id = vdb->registerVariableAndContext(rk3Bfu_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Df(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Df",1)));
	d_rk3Df_id = vdb->registerVariableAndContext(rk3Df ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3DYf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3DYf",1)));
	d_rk3DYf_id = vdb->registerVariableAndContext(rk3DYf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3tauf(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3tauf",1)));
	d_rk3tauf_id = vdb->registerVariableAndContext(rk3tauf ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3phif(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3phif",1)));
	d_rk3phif_id = vdb->registerVariableAndContext(rk3phif ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3gtd_xx(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3gtd_xx",1)));
	d_rk3gtd_xx_id = vdb->registerVariableAndContext(rk3gtd_xx ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3gtd_xy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3gtd_xy",1)));
	d_rk3gtd_xy_id = vdb->registerVariableAndContext(rk3gtd_xy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3gtd_xz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3gtd_xz",1)));
	d_rk3gtd_xz_id = vdb->registerVariableAndContext(rk3gtd_xz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3gtd_yy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3gtd_yy",1)));
	d_rk3gtd_yy_id = vdb->registerVariableAndContext(rk3gtd_yy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3gtd_yz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3gtd_yz",1)));
	d_rk3gtd_yz_id = vdb->registerVariableAndContext(rk3gtd_yz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3gtd_zz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3gtd_zz",1)));
	d_rk3gtd_zz_id = vdb->registerVariableAndContext(rk3gtd_zz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Atd_xx(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Atd_xx",1)));
	d_rk3Atd_xx_id = vdb->registerVariableAndContext(rk3Atd_xx ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Atd_xy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Atd_xy",1)));
	d_rk3Atd_xy_id = vdb->registerVariableAndContext(rk3Atd_xy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Atd_xz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Atd_xz",1)));
	d_rk3Atd_xz_id = vdb->registerVariableAndContext(rk3Atd_xz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Atd_yy(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Atd_yy",1)));
	d_rk3Atd_yy_id = vdb->registerVariableAndContext(rk3Atd_yy ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Atd_yz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Atd_yz",1)));
	d_rk3Atd_yz_id = vdb->registerVariableAndContext(rk3Atd_yz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Atd_zz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Atd_zz",1)));
	d_rk3Atd_zz_id = vdb->registerVariableAndContext(rk3Atd_zz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Gamh_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Gamh_x",1)));
	d_rk3Gamh_x_id = vdb->registerVariableAndContext(rk3Gamh_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Gamh_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Gamh_y",1)));
	d_rk3Gamh_y_id = vdb->registerVariableAndContext(rk3Gamh_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Gamh_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Gamh_z",1)));
	d_rk3Gamh_z_id = vdb->registerVariableAndContext(rk3Gamh_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Betau_x(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Betau_x",1)));
	d_rk3Betau_x_id = vdb->registerVariableAndContext(rk3Betau_x ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Betau_y(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Betau_y",1)));
	d_rk3Betau_y_id = vdb->registerVariableAndContext(rk3Betau_y ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Betau_z(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Betau_z",1)));
	d_rk3Betau_z_id = vdb->registerVariableAndContext(rk3Betau_z ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3Alpha(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3Alpha",1)));
	d_rk3Alpha_id = vdb->registerVariableAndContext(rk3Alpha ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3chi(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3chi",1)));
	d_rk3chi_id = vdb->registerVariableAndContext(rk3chi ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3trK(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3trK",1)));
	d_rk3trK_id = vdb->registerVariableAndContext(rk3trK ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > rk3theta(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "rk3theta",1)));
	d_rk3theta_id = vdb->registerVariableAndContext(rk3theta ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz",1)));
	d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id = vdb->registerVariableAndContext(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id);
	std::shared_ptr< pdat::NodeVariable<double> > d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz",1)));
	d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id = vdb->registerVariableAndContext(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id);
	std::shared_ptr< pdat::NodeVariable<double> > d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz",1)));
	d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id = vdb->registerVariableAndContext(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id);
	std::shared_ptr< pdat::NodeVariable<double> > stalled_1(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "stalled_1",1)));
	d_stalled_1_id = vdb->registerVariableAndContext(stalled_1 ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	hier::PatchDataRestartManager::getManager()->registerPatchDataForRestart(d_stalled_1_id);
	std::shared_ptr< pdat::NodeVariable<double> > Sfd_x_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Sfd_x_p",1)));
	d_Sfd_x_p_id = vdb->registerVariableAndContext(Sfd_x_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Sfd_y_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Sfd_y_p",1)));
	d_Sfd_y_p_id = vdb->registerVariableAndContext(Sfd_y_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Sfd_z_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Sfd_z_p",1)));
	d_Sfd_z_p_id = vdb->registerVariableAndContext(Sfd_z_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Bfu_x_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Bfu_x_p",1)));
	d_Bfu_x_p_id = vdb->registerVariableAndContext(Bfu_x_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Bfu_y_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Bfu_y_p",1)));
	d_Bfu_y_p_id = vdb->registerVariableAndContext(Bfu_y_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Bfu_z_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Bfu_z_p",1)));
	d_Bfu_z_p_id = vdb->registerVariableAndContext(Bfu_z_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Df_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Df_p",1)));
	d_Df_p_id = vdb->registerVariableAndContext(Df_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > DYf_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "DYf_p",1)));
	d_DYf_p_id = vdb->registerVariableAndContext(DYf_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > tauf_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "tauf_p",1)));
	d_tauf_p_id = vdb->registerVariableAndContext(tauf_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > phif_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "phif_p",1)));
	d_phif_p_id = vdb->registerVariableAndContext(phif_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > gtd_xx_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "gtd_xx_p",1)));
	d_gtd_xx_p_id = vdb->registerVariableAndContext(gtd_xx_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > gtd_xy_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "gtd_xy_p",1)));
	d_gtd_xy_p_id = vdb->registerVariableAndContext(gtd_xy_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > gtd_xz_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "gtd_xz_p",1)));
	d_gtd_xz_p_id = vdb->registerVariableAndContext(gtd_xz_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > gtd_yy_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "gtd_yy_p",1)));
	d_gtd_yy_p_id = vdb->registerVariableAndContext(gtd_yy_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > gtd_yz_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "gtd_yz_p",1)));
	d_gtd_yz_p_id = vdb->registerVariableAndContext(gtd_yz_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > gtd_zz_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "gtd_zz_p",1)));
	d_gtd_zz_p_id = vdb->registerVariableAndContext(gtd_zz_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Atd_xx_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Atd_xx_p",1)));
	d_Atd_xx_p_id = vdb->registerVariableAndContext(Atd_xx_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Atd_xy_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Atd_xy_p",1)));
	d_Atd_xy_p_id = vdb->registerVariableAndContext(Atd_xy_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Atd_xz_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Atd_xz_p",1)));
	d_Atd_xz_p_id = vdb->registerVariableAndContext(Atd_xz_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Atd_yy_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Atd_yy_p",1)));
	d_Atd_yy_p_id = vdb->registerVariableAndContext(Atd_yy_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Atd_yz_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Atd_yz_p",1)));
	d_Atd_yz_p_id = vdb->registerVariableAndContext(Atd_yz_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Atd_zz_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Atd_zz_p",1)));
	d_Atd_zz_p_id = vdb->registerVariableAndContext(Atd_zz_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Gamh_x_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Gamh_x_p",1)));
	d_Gamh_x_p_id = vdb->registerVariableAndContext(Gamh_x_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Gamh_y_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Gamh_y_p",1)));
	d_Gamh_y_p_id = vdb->registerVariableAndContext(Gamh_y_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Gamh_z_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Gamh_z_p",1)));
	d_Gamh_z_p_id = vdb->registerVariableAndContext(Gamh_z_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Betau_x_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Betau_x_p",1)));
	d_Betau_x_p_id = vdb->registerVariableAndContext(Betau_x_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Betau_y_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Betau_y_p",1)));
	d_Betau_y_p_id = vdb->registerVariableAndContext(Betau_y_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Betau_z_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Betau_z_p",1)));
	d_Betau_z_p_id = vdb->registerVariableAndContext(Betau_z_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > Alpha_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "Alpha_p",1)));
	d_Alpha_p_id = vdb->registerVariableAndContext(Alpha_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > chi_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "chi_p",1)));
	d_chi_p_id = vdb->registerVariableAndContext(chi_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > trK_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "trK_p",1)));
	d_trK_p_id = vdb->registerVariableAndContext(trK_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));
	std::shared_ptr< pdat::NodeVariable<double> > theta_p(std::shared_ptr< pdat::NodeVariable<double> >(new pdat::NodeVariable<double>(d_dim, "theta_p",1)));
	d_theta_p_id = vdb->registerVariableAndContext(theta_p ,d_cont_curr ,hier::IntVector(d_dim, d_ghost_width));


	//Refine and coarse algorithms

	d_bdry_fill_init = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_bdry_post_coarsen = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_bdry_fill_advance2 = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_bdry_fill_advance5 = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_bdry_fill_advance9 = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_bdry_fill_advance12 = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_bdry_fill_advance16 = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_bdry_fill_advance19 = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_bdry_fill_advance23 = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_bdry_fill_advance26 = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_bdry_fill_analysis1 = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_coarsen_algorithm = std::shared_ptr< xfer::CoarsenAlgorithm >(new xfer::CoarsenAlgorithm(d_dim));

	d_mapping_fill = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
    d_tagging_fill = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_fill_new_level    = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_fill_new_level_aux    = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());
	d_fill_new_level_der_aux    = std::shared_ptr< xfer::RefineAlgorithm >(new xfer::RefineAlgorithm());


	//mapping communication

	std::shared_ptr< hier::RefineOperator > refine_operator_map = d_grid_geometry->lookupRefineOperator(interior, "LINEAR_REFINE");
	d_mapping_fill->registerRefine(d_interior_regridding_value_id,d_interior_regridding_value_id,d_interior_regridding_value_id, refine_operator_map);
	d_mapping_fill->registerRefine(d_interior_i_id,d_interior_i_id,d_interior_i_id, refine_operator_map);
	d_mapping_fill->registerRefine(d_interior_j_id,d_interior_j_id,d_interior_j_id, refine_operator_map);
	d_mapping_fill->registerRefine(d_interior_k_id,d_interior_k_id,d_interior_k_id, refine_operator_map);
	d_mapping_fill->registerRefine(d_FOV_1_id,d_FOV_1_id,d_FOV_1_id, refine_operator_map);
	d_mapping_fill->registerRefine(d_FOV_xLower_id,d_FOV_xLower_id,d_FOV_xLower_id, refine_operator_map);
	d_mapping_fill->registerRefine(d_FOV_xUpper_id,d_FOV_xUpper_id,d_FOV_xUpper_id, refine_operator_map);
	d_mapping_fill->registerRefine(d_FOV_yLower_id,d_FOV_yLower_id,d_FOV_yLower_id, refine_operator_map);
	d_mapping_fill->registerRefine(d_FOV_yUpper_id,d_FOV_yUpper_id,d_FOV_yUpper_id, refine_operator_map);
	d_mapping_fill->registerRefine(d_FOV_zLower_id,d_FOV_zLower_id,d_FOV_zLower_id, refine_operator_map);
	d_mapping_fill->registerRefine(d_FOV_zUpper_id,d_FOV_zUpper_id,d_FOV_zUpper_id, refine_operator_map);
	d_mapping_fill->registerRefine(d_stalled_1_id,d_stalled_1_id,d_stalled_1_id, refine_operator_map);


    d_tagging_fill->registerRefine(d_nonSync_regridding_tag_id,d_nonSync_regridding_tag_id,d_nonSync_regridding_tag_id, d_grid_geometry->lookupRefineOperator(nonSync, "NO_REFINE"));

	//refine and coarsen operators
	string refine_op_name = "LINEAR_REFINE";
	int order = 0;
	if (database->isDatabase("regridding")) {
		regridding_db = database->getDatabase("regridding");
		if (regridding_db->isString("interpolator")) {
			refine_op_name = regridding_db->getString("interpolator");
			if (refine_op_name == "LINEAR_REFINE") {
				order = 1;
			}
			if (refine_op_name == "CUBIC_REFINE") {
				order = 3;
			}
			if (refine_op_name == "QUINTIC_REFINE") {
				order = 5;
			}
		}
	}
	std::shared_ptr< hier::RefineOperator > refine_operator, refine_operator_bound;
	std::shared_ptr< hier::CoarsenOperator > coarsen_operator = d_grid_geometry->lookupCoarsenOperator(FOV_1, "CONSTANT_COARSEN");
	if (order > 0) {
		std::shared_ptr< hier::RefineOperator > tmp_refine_operator(new LagrangianPolynomicRefine(false, order, d_patch_hierarchy, d_dim));
		refine_operator = tmp_refine_operator;
		std::shared_ptr< hier::RefineOperator > tmp_refine_operator_bound(new LagrangianPolynomicRefine(true, order, d_patch_hierarchy, d_dim));
		refine_operator_bound = tmp_refine_operator_bound;
	} else {
		refine_operator = d_grid_geometry->lookupRefineOperator(FOV_1, refine_op_name);
		refine_operator_bound = d_grid_geometry->lookupRefineOperator(FOV_1, refine_op_name);
	}

	std::shared_ptr<SAMRAI::hier::TimeInterpolateOperator> tio_mesh1(new TimeInterpolator(d_grid_geometry, "mesh"));
	time_interpolate_operator_mesh1 = std::dynamic_pointer_cast<TimeInterpolator>(tio_mesh1);


	//Register variables to the refineAlgorithm for boundaries

	if (d_refinedTimeStepping) {
		d_bdry_fill_advance2->registerRefine(d_rk1theta_id,d_rk1theta_id,d_theta_p_id,d_rk1theta_id,d_rk2theta_id,d_rk3theta_id,d_theta_id,d_rk1theta_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1trK_id,d_rk1trK_id,d_trK_p_id,d_rk1trK_id,d_rk2trK_id,d_rk3trK_id,d_trK_id,d_rk1trK_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1chi_id,d_rk1chi_id,d_chi_p_id,d_rk1chi_id,d_rk2chi_id,d_rk3chi_id,d_chi_id,d_rk1chi_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Alpha_id,d_rk1Alpha_id,d_Alpha_p_id,d_rk1Alpha_id,d_rk2Alpha_id,d_rk3Alpha_id,d_Alpha_id,d_rk1Alpha_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Betau_z_id,d_rk1Betau_z_id,d_Betau_z_p_id,d_rk1Betau_z_id,d_rk2Betau_z_id,d_rk3Betau_z_id,d_Betau_z_id,d_rk1Betau_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Betau_y_id,d_rk1Betau_y_id,d_Betau_y_p_id,d_rk1Betau_y_id,d_rk2Betau_y_id,d_rk3Betau_y_id,d_Betau_y_id,d_rk1Betau_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Betau_x_id,d_rk1Betau_x_id,d_Betau_x_p_id,d_rk1Betau_x_id,d_rk2Betau_x_id,d_rk3Betau_x_id,d_Betau_x_id,d_rk1Betau_x_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Gamh_z_id,d_rk1Gamh_z_id,d_Gamh_z_p_id,d_rk1Gamh_z_id,d_rk2Gamh_z_id,d_rk3Gamh_z_id,d_Gamh_z_id,d_rk1Gamh_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Gamh_y_id,d_rk1Gamh_y_id,d_Gamh_y_p_id,d_rk1Gamh_y_id,d_rk2Gamh_y_id,d_rk3Gamh_y_id,d_Gamh_y_id,d_rk1Gamh_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Gamh_x_id,d_rk1Gamh_x_id,d_Gamh_x_p_id,d_rk1Gamh_x_id,d_rk2Gamh_x_id,d_rk3Gamh_x_id,d_Gamh_x_id,d_rk1Gamh_x_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Atd_zz_id,d_rk1Atd_zz_id,d_Atd_zz_p_id,d_rk1Atd_zz_id,d_rk2Atd_zz_id,d_rk3Atd_zz_id,d_Atd_zz_id,d_rk1Atd_zz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Atd_yz_id,d_rk1Atd_yz_id,d_Atd_yz_p_id,d_rk1Atd_yz_id,d_rk2Atd_yz_id,d_rk3Atd_yz_id,d_Atd_yz_id,d_rk1Atd_yz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Atd_yy_id,d_rk1Atd_yy_id,d_Atd_yy_p_id,d_rk1Atd_yy_id,d_rk2Atd_yy_id,d_rk3Atd_yy_id,d_Atd_yy_id,d_rk1Atd_yy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Atd_xz_id,d_rk1Atd_xz_id,d_Atd_xz_p_id,d_rk1Atd_xz_id,d_rk2Atd_xz_id,d_rk3Atd_xz_id,d_Atd_xz_id,d_rk1Atd_xz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Atd_xy_id,d_rk1Atd_xy_id,d_Atd_xy_p_id,d_rk1Atd_xy_id,d_rk2Atd_xy_id,d_rk3Atd_xy_id,d_Atd_xy_id,d_rk1Atd_xy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Atd_xx_id,d_rk1Atd_xx_id,d_Atd_xx_p_id,d_rk1Atd_xx_id,d_rk2Atd_xx_id,d_rk3Atd_xx_id,d_Atd_xx_id,d_rk1Atd_xx_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1gtd_zz_id,d_rk1gtd_zz_id,d_gtd_zz_p_id,d_rk1gtd_zz_id,d_rk2gtd_zz_id,d_rk3gtd_zz_id,d_gtd_zz_id,d_rk1gtd_zz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1gtd_yz_id,d_rk1gtd_yz_id,d_gtd_yz_p_id,d_rk1gtd_yz_id,d_rk2gtd_yz_id,d_rk3gtd_yz_id,d_gtd_yz_id,d_rk1gtd_yz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1gtd_yy_id,d_rk1gtd_yy_id,d_gtd_yy_p_id,d_rk1gtd_yy_id,d_rk2gtd_yy_id,d_rk3gtd_yy_id,d_gtd_yy_id,d_rk1gtd_yy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1gtd_xz_id,d_rk1gtd_xz_id,d_gtd_xz_p_id,d_rk1gtd_xz_id,d_rk2gtd_xz_id,d_rk3gtd_xz_id,d_gtd_xz_id,d_rk1gtd_xz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1gtd_xy_id,d_rk1gtd_xy_id,d_gtd_xy_p_id,d_rk1gtd_xy_id,d_rk2gtd_xy_id,d_rk3gtd_xy_id,d_gtd_xy_id,d_rk1gtd_xy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1gtd_xx_id,d_rk1gtd_xx_id,d_gtd_xx_p_id,d_rk1gtd_xx_id,d_rk2gtd_xx_id,d_rk3gtd_xx_id,d_gtd_xx_id,d_rk1gtd_xx_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1phif_id,d_rk1phif_id,d_phif_p_id,d_rk1phif_id,d_rk2phif_id,d_rk3phif_id,d_phif_id,d_rk1phif_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1tauf_id,d_rk1tauf_id,d_tauf_p_id,d_rk1tauf_id,d_rk2tauf_id,d_rk3tauf_id,d_tauf_id,d_rk1tauf_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1DYf_id,d_rk1DYf_id,d_DYf_p_id,d_rk1DYf_id,d_rk2DYf_id,d_rk3DYf_id,d_DYf_id,d_rk1DYf_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Df_id,d_rk1Df_id,d_Df_p_id,d_rk1Df_id,d_rk2Df_id,d_rk3Df_id,d_Df_id,d_rk1Df_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Bfu_z_id,d_rk1Bfu_z_id,d_Bfu_z_p_id,d_rk1Bfu_z_id,d_rk2Bfu_z_id,d_rk3Bfu_z_id,d_Bfu_z_id,d_rk1Bfu_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Bfu_y_id,d_rk1Bfu_y_id,d_Bfu_y_p_id,d_rk1Bfu_y_id,d_rk2Bfu_y_id,d_rk3Bfu_y_id,d_Bfu_y_id,d_rk1Bfu_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Bfu_x_id,d_rk1Bfu_x_id,d_Bfu_x_p_id,d_rk1Bfu_x_id,d_rk2Bfu_x_id,d_rk3Bfu_x_id,d_Bfu_x_id,d_rk1Bfu_x_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Sfd_z_id,d_rk1Sfd_z_id,d_Sfd_z_p_id,d_rk1Sfd_z_id,d_rk2Sfd_z_id,d_rk3Sfd_z_id,d_Sfd_z_id,d_rk1Sfd_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Sfd_y_id,d_rk1Sfd_y_id,d_Sfd_y_p_id,d_rk1Sfd_y_id,d_rk2Sfd_y_id,d_rk3Sfd_y_id,d_Sfd_y_id,d_rk1Sfd_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance2->registerRefine(d_rk1Sfd_x_id,d_rk1Sfd_x_id,d_Sfd_x_p_id,d_rk1Sfd_x_id,d_rk2Sfd_x_id,d_rk3Sfd_x_id,d_Sfd_x_id,d_rk1Sfd_x_id,refine_operator,tio_mesh1);
	} else {
		d_bdry_fill_advance2->registerRefine(d_rk1theta_id,d_rk1theta_id,d_rk1theta_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1trK_id,d_rk1trK_id,d_rk1trK_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1chi_id,d_rk1chi_id,d_rk1chi_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Alpha_id,d_rk1Alpha_id,d_rk1Alpha_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Betau_z_id,d_rk1Betau_z_id,d_rk1Betau_z_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Betau_y_id,d_rk1Betau_y_id,d_rk1Betau_y_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Betau_x_id,d_rk1Betau_x_id,d_rk1Betau_x_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Gamh_z_id,d_rk1Gamh_z_id,d_rk1Gamh_z_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Gamh_y_id,d_rk1Gamh_y_id,d_rk1Gamh_y_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Gamh_x_id,d_rk1Gamh_x_id,d_rk1Gamh_x_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Atd_zz_id,d_rk1Atd_zz_id,d_rk1Atd_zz_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Atd_yz_id,d_rk1Atd_yz_id,d_rk1Atd_yz_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Atd_yy_id,d_rk1Atd_yy_id,d_rk1Atd_yy_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Atd_xz_id,d_rk1Atd_xz_id,d_rk1Atd_xz_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Atd_xy_id,d_rk1Atd_xy_id,d_rk1Atd_xy_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Atd_xx_id,d_rk1Atd_xx_id,d_rk1Atd_xx_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1gtd_zz_id,d_rk1gtd_zz_id,d_rk1gtd_zz_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1gtd_yz_id,d_rk1gtd_yz_id,d_rk1gtd_yz_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1gtd_yy_id,d_rk1gtd_yy_id,d_rk1gtd_yy_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1gtd_xz_id,d_rk1gtd_xz_id,d_rk1gtd_xz_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1gtd_xy_id,d_rk1gtd_xy_id,d_rk1gtd_xy_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1gtd_xx_id,d_rk1gtd_xx_id,d_rk1gtd_xx_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1phif_id,d_rk1phif_id,d_rk1phif_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1tauf_id,d_rk1tauf_id,d_rk1tauf_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1DYf_id,d_rk1DYf_id,d_rk1DYf_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Df_id,d_rk1Df_id,d_rk1Df_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Bfu_z_id,d_rk1Bfu_z_id,d_rk1Bfu_z_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Bfu_y_id,d_rk1Bfu_y_id,d_rk1Bfu_y_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Bfu_x_id,d_rk1Bfu_x_id,d_rk1Bfu_x_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Sfd_z_id,d_rk1Sfd_z_id,d_rk1Sfd_z_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Sfd_y_id,d_rk1Sfd_y_id,d_rk1Sfd_y_id,refine_operator);
		d_bdry_fill_advance2->registerRefine(d_rk1Sfd_x_id,d_rk1Sfd_x_id,d_rk1Sfd_x_id,refine_operator);
	}
	d_bdry_fill_advance5->registerRefine(d_rnu_id,d_rnu_id,d_rnu_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_qnu_id,d_qnu_id,d_qnu_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_qnu_x_id,d_qnu_x_id,d_qnu_x_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_qnu_a_id,d_qnu_a_id,d_qnu_a_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_qnu_e_id,d_qnu_e_id,d_qnu_e_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_chix_id,d_chix_id,d_chix_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_chia_id,d_chia_id,d_chia_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_chie_id,d_chie_id,d_chie_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_optdepthx_id,d_optdepthx_id,d_optdepthx_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_optdeptha_id,d_optdeptha_id,d_optdeptha_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_optdepthe_id,d_optdepthe_id,d_optdepthe_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauN_y_id,d_TauN_y_id,d_TauN_y_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauN_z_id,d_TauN_z_id,d_TauN_z_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauM_xz_id,d_TauM_xz_id,d_TauM_xz_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauT_xx_id,d_TauT_xx_id,d_TauT_xx_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauT_yy_id,d_TauT_yy_id,d_TauT_yy_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauT_zz_id,d_TauT_zz_id,d_TauT_zz_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauN_x_id,d_TauN_x_id,d_TauN_x_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauT_xz_id,d_TauT_xz_id,d_TauT_xz_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauNe_z_id,d_TauNe_z_id,d_TauNe_z_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauM_xy_id,d_TauM_xy_id,d_TauM_xy_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauM_yz_id,d_TauM_yz_id,d_TauM_yz_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauT_xy_id,d_TauT_xy_id,d_TauT_xy_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauT_yz_id,d_TauT_yz_id,d_TauT_yz_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauNe_y_id,d_TauNe_y_id,d_TauNe_y_id,refine_operator);
	d_bdry_fill_advance5->registerRefine(d_TauNe_x_id,d_TauNe_x_id,d_TauNe_x_id,refine_operator);
	if (d_refinedTimeStepping) {
		d_bdry_fill_advance9->registerRefine(d_rk2theta_id,d_rk2theta_id,d_theta_p_id,d_rk1theta_id,d_rk2theta_id,d_rk3theta_id,d_theta_id,d_rk2theta_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2trK_id,d_rk2trK_id,d_trK_p_id,d_rk1trK_id,d_rk2trK_id,d_rk3trK_id,d_trK_id,d_rk2trK_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2chi_id,d_rk2chi_id,d_chi_p_id,d_rk1chi_id,d_rk2chi_id,d_rk3chi_id,d_chi_id,d_rk2chi_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Alpha_id,d_rk2Alpha_id,d_Alpha_p_id,d_rk1Alpha_id,d_rk2Alpha_id,d_rk3Alpha_id,d_Alpha_id,d_rk2Alpha_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Betau_z_id,d_rk2Betau_z_id,d_Betau_z_p_id,d_rk1Betau_z_id,d_rk2Betau_z_id,d_rk3Betau_z_id,d_Betau_z_id,d_rk2Betau_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Betau_y_id,d_rk2Betau_y_id,d_Betau_y_p_id,d_rk1Betau_y_id,d_rk2Betau_y_id,d_rk3Betau_y_id,d_Betau_y_id,d_rk2Betau_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Betau_x_id,d_rk2Betau_x_id,d_Betau_x_p_id,d_rk1Betau_x_id,d_rk2Betau_x_id,d_rk3Betau_x_id,d_Betau_x_id,d_rk2Betau_x_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Gamh_z_id,d_rk2Gamh_z_id,d_Gamh_z_p_id,d_rk1Gamh_z_id,d_rk2Gamh_z_id,d_rk3Gamh_z_id,d_Gamh_z_id,d_rk2Gamh_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Gamh_y_id,d_rk2Gamh_y_id,d_Gamh_y_p_id,d_rk1Gamh_y_id,d_rk2Gamh_y_id,d_rk3Gamh_y_id,d_Gamh_y_id,d_rk2Gamh_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Gamh_x_id,d_rk2Gamh_x_id,d_Gamh_x_p_id,d_rk1Gamh_x_id,d_rk2Gamh_x_id,d_rk3Gamh_x_id,d_Gamh_x_id,d_rk2Gamh_x_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Atd_zz_id,d_rk2Atd_zz_id,d_Atd_zz_p_id,d_rk1Atd_zz_id,d_rk2Atd_zz_id,d_rk3Atd_zz_id,d_Atd_zz_id,d_rk2Atd_zz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Atd_yz_id,d_rk2Atd_yz_id,d_Atd_yz_p_id,d_rk1Atd_yz_id,d_rk2Atd_yz_id,d_rk3Atd_yz_id,d_Atd_yz_id,d_rk2Atd_yz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Atd_yy_id,d_rk2Atd_yy_id,d_Atd_yy_p_id,d_rk1Atd_yy_id,d_rk2Atd_yy_id,d_rk3Atd_yy_id,d_Atd_yy_id,d_rk2Atd_yy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Atd_xz_id,d_rk2Atd_xz_id,d_Atd_xz_p_id,d_rk1Atd_xz_id,d_rk2Atd_xz_id,d_rk3Atd_xz_id,d_Atd_xz_id,d_rk2Atd_xz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Atd_xy_id,d_rk2Atd_xy_id,d_Atd_xy_p_id,d_rk1Atd_xy_id,d_rk2Atd_xy_id,d_rk3Atd_xy_id,d_Atd_xy_id,d_rk2Atd_xy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Atd_xx_id,d_rk2Atd_xx_id,d_Atd_xx_p_id,d_rk1Atd_xx_id,d_rk2Atd_xx_id,d_rk3Atd_xx_id,d_Atd_xx_id,d_rk2Atd_xx_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2gtd_zz_id,d_rk2gtd_zz_id,d_gtd_zz_p_id,d_rk1gtd_zz_id,d_rk2gtd_zz_id,d_rk3gtd_zz_id,d_gtd_zz_id,d_rk2gtd_zz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2gtd_yz_id,d_rk2gtd_yz_id,d_gtd_yz_p_id,d_rk1gtd_yz_id,d_rk2gtd_yz_id,d_rk3gtd_yz_id,d_gtd_yz_id,d_rk2gtd_yz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2gtd_yy_id,d_rk2gtd_yy_id,d_gtd_yy_p_id,d_rk1gtd_yy_id,d_rk2gtd_yy_id,d_rk3gtd_yy_id,d_gtd_yy_id,d_rk2gtd_yy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2gtd_xz_id,d_rk2gtd_xz_id,d_gtd_xz_p_id,d_rk1gtd_xz_id,d_rk2gtd_xz_id,d_rk3gtd_xz_id,d_gtd_xz_id,d_rk2gtd_xz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2gtd_xy_id,d_rk2gtd_xy_id,d_gtd_xy_p_id,d_rk1gtd_xy_id,d_rk2gtd_xy_id,d_rk3gtd_xy_id,d_gtd_xy_id,d_rk2gtd_xy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2gtd_xx_id,d_rk2gtd_xx_id,d_gtd_xx_p_id,d_rk1gtd_xx_id,d_rk2gtd_xx_id,d_rk3gtd_xx_id,d_gtd_xx_id,d_rk2gtd_xx_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2phif_id,d_rk2phif_id,d_phif_p_id,d_rk1phif_id,d_rk2phif_id,d_rk3phif_id,d_phif_id,d_rk2phif_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2tauf_id,d_rk2tauf_id,d_tauf_p_id,d_rk1tauf_id,d_rk2tauf_id,d_rk3tauf_id,d_tauf_id,d_rk2tauf_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2DYf_id,d_rk2DYf_id,d_DYf_p_id,d_rk1DYf_id,d_rk2DYf_id,d_rk3DYf_id,d_DYf_id,d_rk2DYf_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Df_id,d_rk2Df_id,d_Df_p_id,d_rk1Df_id,d_rk2Df_id,d_rk3Df_id,d_Df_id,d_rk2Df_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Bfu_z_id,d_rk2Bfu_z_id,d_Bfu_z_p_id,d_rk1Bfu_z_id,d_rk2Bfu_z_id,d_rk3Bfu_z_id,d_Bfu_z_id,d_rk2Bfu_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Bfu_y_id,d_rk2Bfu_y_id,d_Bfu_y_p_id,d_rk1Bfu_y_id,d_rk2Bfu_y_id,d_rk3Bfu_y_id,d_Bfu_y_id,d_rk2Bfu_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Bfu_x_id,d_rk2Bfu_x_id,d_Bfu_x_p_id,d_rk1Bfu_x_id,d_rk2Bfu_x_id,d_rk3Bfu_x_id,d_Bfu_x_id,d_rk2Bfu_x_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Sfd_z_id,d_rk2Sfd_z_id,d_Sfd_z_p_id,d_rk1Sfd_z_id,d_rk2Sfd_z_id,d_rk3Sfd_z_id,d_Sfd_z_id,d_rk2Sfd_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Sfd_y_id,d_rk2Sfd_y_id,d_Sfd_y_p_id,d_rk1Sfd_y_id,d_rk2Sfd_y_id,d_rk3Sfd_y_id,d_Sfd_y_id,d_rk2Sfd_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance9->registerRefine(d_rk2Sfd_x_id,d_rk2Sfd_x_id,d_Sfd_x_p_id,d_rk1Sfd_x_id,d_rk2Sfd_x_id,d_rk3Sfd_x_id,d_Sfd_x_id,d_rk2Sfd_x_id,refine_operator,tio_mesh1);
	} else {
		d_bdry_fill_advance9->registerRefine(d_rk2theta_id,d_rk2theta_id,d_rk2theta_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2trK_id,d_rk2trK_id,d_rk2trK_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2chi_id,d_rk2chi_id,d_rk2chi_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Alpha_id,d_rk2Alpha_id,d_rk2Alpha_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Betau_z_id,d_rk2Betau_z_id,d_rk2Betau_z_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Betau_y_id,d_rk2Betau_y_id,d_rk2Betau_y_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Betau_x_id,d_rk2Betau_x_id,d_rk2Betau_x_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Gamh_z_id,d_rk2Gamh_z_id,d_rk2Gamh_z_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Gamh_y_id,d_rk2Gamh_y_id,d_rk2Gamh_y_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Gamh_x_id,d_rk2Gamh_x_id,d_rk2Gamh_x_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Atd_zz_id,d_rk2Atd_zz_id,d_rk2Atd_zz_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Atd_yz_id,d_rk2Atd_yz_id,d_rk2Atd_yz_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Atd_yy_id,d_rk2Atd_yy_id,d_rk2Atd_yy_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Atd_xz_id,d_rk2Atd_xz_id,d_rk2Atd_xz_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Atd_xy_id,d_rk2Atd_xy_id,d_rk2Atd_xy_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Atd_xx_id,d_rk2Atd_xx_id,d_rk2Atd_xx_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2gtd_zz_id,d_rk2gtd_zz_id,d_rk2gtd_zz_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2gtd_yz_id,d_rk2gtd_yz_id,d_rk2gtd_yz_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2gtd_yy_id,d_rk2gtd_yy_id,d_rk2gtd_yy_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2gtd_xz_id,d_rk2gtd_xz_id,d_rk2gtd_xz_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2gtd_xy_id,d_rk2gtd_xy_id,d_rk2gtd_xy_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2gtd_xx_id,d_rk2gtd_xx_id,d_rk2gtd_xx_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2phif_id,d_rk2phif_id,d_rk2phif_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2tauf_id,d_rk2tauf_id,d_rk2tauf_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2DYf_id,d_rk2DYf_id,d_rk2DYf_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Df_id,d_rk2Df_id,d_rk2Df_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Bfu_z_id,d_rk2Bfu_z_id,d_rk2Bfu_z_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Bfu_y_id,d_rk2Bfu_y_id,d_rk2Bfu_y_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Bfu_x_id,d_rk2Bfu_x_id,d_rk2Bfu_x_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Sfd_z_id,d_rk2Sfd_z_id,d_rk2Sfd_z_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Sfd_y_id,d_rk2Sfd_y_id,d_rk2Sfd_y_id,refine_operator);
		d_bdry_fill_advance9->registerRefine(d_rk2Sfd_x_id,d_rk2Sfd_x_id,d_rk2Sfd_x_id,refine_operator);
	}
	d_bdry_fill_advance12->registerRefine(d_rnu_id,d_rnu_id,d_rnu_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_qnu_id,d_qnu_id,d_qnu_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_qnu_x_id,d_qnu_x_id,d_qnu_x_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_qnu_a_id,d_qnu_a_id,d_qnu_a_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_qnu_e_id,d_qnu_e_id,d_qnu_e_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_chix_id,d_chix_id,d_chix_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_chia_id,d_chia_id,d_chia_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_chie_id,d_chie_id,d_chie_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_optdepthx_id,d_optdepthx_id,d_optdepthx_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_optdeptha_id,d_optdeptha_id,d_optdeptha_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_optdepthe_id,d_optdepthe_id,d_optdepthe_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauN_y_id,d_TauN_y_id,d_TauN_y_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauN_z_id,d_TauN_z_id,d_TauN_z_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauM_xz_id,d_TauM_xz_id,d_TauM_xz_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauT_xx_id,d_TauT_xx_id,d_TauT_xx_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauT_yy_id,d_TauT_yy_id,d_TauT_yy_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauT_zz_id,d_TauT_zz_id,d_TauT_zz_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauN_x_id,d_TauN_x_id,d_TauN_x_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauT_xz_id,d_TauT_xz_id,d_TauT_xz_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauNe_z_id,d_TauNe_z_id,d_TauNe_z_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauM_xy_id,d_TauM_xy_id,d_TauM_xy_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauM_yz_id,d_TauM_yz_id,d_TauM_yz_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauT_xy_id,d_TauT_xy_id,d_TauT_xy_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauT_yz_id,d_TauT_yz_id,d_TauT_yz_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauNe_y_id,d_TauNe_y_id,d_TauNe_y_id,refine_operator);
	d_bdry_fill_advance12->registerRefine(d_TauNe_x_id,d_TauNe_x_id,d_TauNe_x_id,refine_operator);
	if (d_refinedTimeStepping) {
		d_bdry_fill_advance16->registerRefine(d_rk3theta_id,d_rk3theta_id,d_theta_p_id,d_rk1theta_id,d_rk2theta_id,d_rk3theta_id,d_theta_id,d_rk3theta_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3trK_id,d_rk3trK_id,d_trK_p_id,d_rk1trK_id,d_rk2trK_id,d_rk3trK_id,d_trK_id,d_rk3trK_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3chi_id,d_rk3chi_id,d_chi_p_id,d_rk1chi_id,d_rk2chi_id,d_rk3chi_id,d_chi_id,d_rk3chi_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Alpha_id,d_rk3Alpha_id,d_Alpha_p_id,d_rk1Alpha_id,d_rk2Alpha_id,d_rk3Alpha_id,d_Alpha_id,d_rk3Alpha_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Betau_z_id,d_rk3Betau_z_id,d_Betau_z_p_id,d_rk1Betau_z_id,d_rk2Betau_z_id,d_rk3Betau_z_id,d_Betau_z_id,d_rk3Betau_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Betau_y_id,d_rk3Betau_y_id,d_Betau_y_p_id,d_rk1Betau_y_id,d_rk2Betau_y_id,d_rk3Betau_y_id,d_Betau_y_id,d_rk3Betau_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Betau_x_id,d_rk3Betau_x_id,d_Betau_x_p_id,d_rk1Betau_x_id,d_rk2Betau_x_id,d_rk3Betau_x_id,d_Betau_x_id,d_rk3Betau_x_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Gamh_z_id,d_rk3Gamh_z_id,d_Gamh_z_p_id,d_rk1Gamh_z_id,d_rk2Gamh_z_id,d_rk3Gamh_z_id,d_Gamh_z_id,d_rk3Gamh_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Gamh_y_id,d_rk3Gamh_y_id,d_Gamh_y_p_id,d_rk1Gamh_y_id,d_rk2Gamh_y_id,d_rk3Gamh_y_id,d_Gamh_y_id,d_rk3Gamh_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Gamh_x_id,d_rk3Gamh_x_id,d_Gamh_x_p_id,d_rk1Gamh_x_id,d_rk2Gamh_x_id,d_rk3Gamh_x_id,d_Gamh_x_id,d_rk3Gamh_x_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Atd_zz_id,d_rk3Atd_zz_id,d_Atd_zz_p_id,d_rk1Atd_zz_id,d_rk2Atd_zz_id,d_rk3Atd_zz_id,d_Atd_zz_id,d_rk3Atd_zz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Atd_yz_id,d_rk3Atd_yz_id,d_Atd_yz_p_id,d_rk1Atd_yz_id,d_rk2Atd_yz_id,d_rk3Atd_yz_id,d_Atd_yz_id,d_rk3Atd_yz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Atd_yy_id,d_rk3Atd_yy_id,d_Atd_yy_p_id,d_rk1Atd_yy_id,d_rk2Atd_yy_id,d_rk3Atd_yy_id,d_Atd_yy_id,d_rk3Atd_yy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Atd_xz_id,d_rk3Atd_xz_id,d_Atd_xz_p_id,d_rk1Atd_xz_id,d_rk2Atd_xz_id,d_rk3Atd_xz_id,d_Atd_xz_id,d_rk3Atd_xz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Atd_xy_id,d_rk3Atd_xy_id,d_Atd_xy_p_id,d_rk1Atd_xy_id,d_rk2Atd_xy_id,d_rk3Atd_xy_id,d_Atd_xy_id,d_rk3Atd_xy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Atd_xx_id,d_rk3Atd_xx_id,d_Atd_xx_p_id,d_rk1Atd_xx_id,d_rk2Atd_xx_id,d_rk3Atd_xx_id,d_Atd_xx_id,d_rk3Atd_xx_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3gtd_zz_id,d_rk3gtd_zz_id,d_gtd_zz_p_id,d_rk1gtd_zz_id,d_rk2gtd_zz_id,d_rk3gtd_zz_id,d_gtd_zz_id,d_rk3gtd_zz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3gtd_yz_id,d_rk3gtd_yz_id,d_gtd_yz_p_id,d_rk1gtd_yz_id,d_rk2gtd_yz_id,d_rk3gtd_yz_id,d_gtd_yz_id,d_rk3gtd_yz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3gtd_yy_id,d_rk3gtd_yy_id,d_gtd_yy_p_id,d_rk1gtd_yy_id,d_rk2gtd_yy_id,d_rk3gtd_yy_id,d_gtd_yy_id,d_rk3gtd_yy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3gtd_xz_id,d_rk3gtd_xz_id,d_gtd_xz_p_id,d_rk1gtd_xz_id,d_rk2gtd_xz_id,d_rk3gtd_xz_id,d_gtd_xz_id,d_rk3gtd_xz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3gtd_xy_id,d_rk3gtd_xy_id,d_gtd_xy_p_id,d_rk1gtd_xy_id,d_rk2gtd_xy_id,d_rk3gtd_xy_id,d_gtd_xy_id,d_rk3gtd_xy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3gtd_xx_id,d_rk3gtd_xx_id,d_gtd_xx_p_id,d_rk1gtd_xx_id,d_rk2gtd_xx_id,d_rk3gtd_xx_id,d_gtd_xx_id,d_rk3gtd_xx_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3phif_id,d_rk3phif_id,d_phif_p_id,d_rk1phif_id,d_rk2phif_id,d_rk3phif_id,d_phif_id,d_rk3phif_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3tauf_id,d_rk3tauf_id,d_tauf_p_id,d_rk1tauf_id,d_rk2tauf_id,d_rk3tauf_id,d_tauf_id,d_rk3tauf_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3DYf_id,d_rk3DYf_id,d_DYf_p_id,d_rk1DYf_id,d_rk2DYf_id,d_rk3DYf_id,d_DYf_id,d_rk3DYf_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Df_id,d_rk3Df_id,d_Df_p_id,d_rk1Df_id,d_rk2Df_id,d_rk3Df_id,d_Df_id,d_rk3Df_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Bfu_z_id,d_rk3Bfu_z_id,d_Bfu_z_p_id,d_rk1Bfu_z_id,d_rk2Bfu_z_id,d_rk3Bfu_z_id,d_Bfu_z_id,d_rk3Bfu_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Bfu_y_id,d_rk3Bfu_y_id,d_Bfu_y_p_id,d_rk1Bfu_y_id,d_rk2Bfu_y_id,d_rk3Bfu_y_id,d_Bfu_y_id,d_rk3Bfu_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Bfu_x_id,d_rk3Bfu_x_id,d_Bfu_x_p_id,d_rk1Bfu_x_id,d_rk2Bfu_x_id,d_rk3Bfu_x_id,d_Bfu_x_id,d_rk3Bfu_x_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Sfd_z_id,d_rk3Sfd_z_id,d_Sfd_z_p_id,d_rk1Sfd_z_id,d_rk2Sfd_z_id,d_rk3Sfd_z_id,d_Sfd_z_id,d_rk3Sfd_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Sfd_y_id,d_rk3Sfd_y_id,d_Sfd_y_p_id,d_rk1Sfd_y_id,d_rk2Sfd_y_id,d_rk3Sfd_y_id,d_Sfd_y_id,d_rk3Sfd_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance16->registerRefine(d_rk3Sfd_x_id,d_rk3Sfd_x_id,d_Sfd_x_p_id,d_rk1Sfd_x_id,d_rk2Sfd_x_id,d_rk3Sfd_x_id,d_Sfd_x_id,d_rk3Sfd_x_id,refine_operator,tio_mesh1);
	} else {
		d_bdry_fill_advance16->registerRefine(d_rk3theta_id,d_rk3theta_id,d_rk3theta_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3trK_id,d_rk3trK_id,d_rk3trK_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3chi_id,d_rk3chi_id,d_rk3chi_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Alpha_id,d_rk3Alpha_id,d_rk3Alpha_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Betau_z_id,d_rk3Betau_z_id,d_rk3Betau_z_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Betau_y_id,d_rk3Betau_y_id,d_rk3Betau_y_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Betau_x_id,d_rk3Betau_x_id,d_rk3Betau_x_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Gamh_z_id,d_rk3Gamh_z_id,d_rk3Gamh_z_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Gamh_y_id,d_rk3Gamh_y_id,d_rk3Gamh_y_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Gamh_x_id,d_rk3Gamh_x_id,d_rk3Gamh_x_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Atd_zz_id,d_rk3Atd_zz_id,d_rk3Atd_zz_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Atd_yz_id,d_rk3Atd_yz_id,d_rk3Atd_yz_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Atd_yy_id,d_rk3Atd_yy_id,d_rk3Atd_yy_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Atd_xz_id,d_rk3Atd_xz_id,d_rk3Atd_xz_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Atd_xy_id,d_rk3Atd_xy_id,d_rk3Atd_xy_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Atd_xx_id,d_rk3Atd_xx_id,d_rk3Atd_xx_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3gtd_zz_id,d_rk3gtd_zz_id,d_rk3gtd_zz_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3gtd_yz_id,d_rk3gtd_yz_id,d_rk3gtd_yz_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3gtd_yy_id,d_rk3gtd_yy_id,d_rk3gtd_yy_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3gtd_xz_id,d_rk3gtd_xz_id,d_rk3gtd_xz_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3gtd_xy_id,d_rk3gtd_xy_id,d_rk3gtd_xy_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3gtd_xx_id,d_rk3gtd_xx_id,d_rk3gtd_xx_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3phif_id,d_rk3phif_id,d_rk3phif_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3tauf_id,d_rk3tauf_id,d_rk3tauf_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3DYf_id,d_rk3DYf_id,d_rk3DYf_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Df_id,d_rk3Df_id,d_rk3Df_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Bfu_z_id,d_rk3Bfu_z_id,d_rk3Bfu_z_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Bfu_y_id,d_rk3Bfu_y_id,d_rk3Bfu_y_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Bfu_x_id,d_rk3Bfu_x_id,d_rk3Bfu_x_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Sfd_z_id,d_rk3Sfd_z_id,d_rk3Sfd_z_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Sfd_y_id,d_rk3Sfd_y_id,d_rk3Sfd_y_id,refine_operator);
		d_bdry_fill_advance16->registerRefine(d_rk3Sfd_x_id,d_rk3Sfd_x_id,d_rk3Sfd_x_id,refine_operator);
	}
	d_bdry_fill_advance19->registerRefine(d_rnu_id,d_rnu_id,d_rnu_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_qnu_id,d_qnu_id,d_qnu_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_qnu_x_id,d_qnu_x_id,d_qnu_x_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_qnu_a_id,d_qnu_a_id,d_qnu_a_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_qnu_e_id,d_qnu_e_id,d_qnu_e_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_chix_id,d_chix_id,d_chix_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_chia_id,d_chia_id,d_chia_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_chie_id,d_chie_id,d_chie_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_optdepthx_id,d_optdepthx_id,d_optdepthx_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_optdeptha_id,d_optdeptha_id,d_optdeptha_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_optdepthe_id,d_optdepthe_id,d_optdepthe_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauN_y_id,d_TauN_y_id,d_TauN_y_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauN_z_id,d_TauN_z_id,d_TauN_z_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauM_xz_id,d_TauM_xz_id,d_TauM_xz_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauT_xx_id,d_TauT_xx_id,d_TauT_xx_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauT_yy_id,d_TauT_yy_id,d_TauT_yy_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauT_zz_id,d_TauT_zz_id,d_TauT_zz_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauN_x_id,d_TauN_x_id,d_TauN_x_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauT_xz_id,d_TauT_xz_id,d_TauT_xz_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauNe_z_id,d_TauNe_z_id,d_TauNe_z_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauM_xy_id,d_TauM_xy_id,d_TauM_xy_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauM_yz_id,d_TauM_yz_id,d_TauM_yz_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauT_xy_id,d_TauT_xy_id,d_TauT_xy_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauT_yz_id,d_TauT_yz_id,d_TauT_yz_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauNe_y_id,d_TauNe_y_id,d_TauNe_y_id,refine_operator);
	d_bdry_fill_advance19->registerRefine(d_TauNe_x_id,d_TauNe_x_id,d_TauNe_x_id,refine_operator);
	if (d_refinedTimeStepping) {
		d_bdry_fill_advance23->registerRefine(d_theta_id,d_theta_id,d_theta_p_id,d_rk1theta_id,d_rk2theta_id,d_rk3theta_id,d_theta_id,d_theta_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_trK_id,d_trK_id,d_trK_p_id,d_rk1trK_id,d_rk2trK_id,d_rk3trK_id,d_trK_id,d_trK_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_chi_id,d_chi_id,d_chi_p_id,d_rk1chi_id,d_rk2chi_id,d_rk3chi_id,d_chi_id,d_chi_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Alpha_id,d_Alpha_id,d_Alpha_p_id,d_rk1Alpha_id,d_rk2Alpha_id,d_rk3Alpha_id,d_Alpha_id,d_Alpha_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Betau_z_id,d_Betau_z_id,d_Betau_z_p_id,d_rk1Betau_z_id,d_rk2Betau_z_id,d_rk3Betau_z_id,d_Betau_z_id,d_Betau_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Betau_y_id,d_Betau_y_id,d_Betau_y_p_id,d_rk1Betau_y_id,d_rk2Betau_y_id,d_rk3Betau_y_id,d_Betau_y_id,d_Betau_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Betau_x_id,d_Betau_x_id,d_Betau_x_p_id,d_rk1Betau_x_id,d_rk2Betau_x_id,d_rk3Betau_x_id,d_Betau_x_id,d_Betau_x_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Gamh_z_id,d_Gamh_z_id,d_Gamh_z_p_id,d_rk1Gamh_z_id,d_rk2Gamh_z_id,d_rk3Gamh_z_id,d_Gamh_z_id,d_Gamh_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Gamh_y_id,d_Gamh_y_id,d_Gamh_y_p_id,d_rk1Gamh_y_id,d_rk2Gamh_y_id,d_rk3Gamh_y_id,d_Gamh_y_id,d_Gamh_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Gamh_x_id,d_Gamh_x_id,d_Gamh_x_p_id,d_rk1Gamh_x_id,d_rk2Gamh_x_id,d_rk3Gamh_x_id,d_Gamh_x_id,d_Gamh_x_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Atd_zz_id,d_Atd_zz_id,d_Atd_zz_p_id,d_rk1Atd_zz_id,d_rk2Atd_zz_id,d_rk3Atd_zz_id,d_Atd_zz_id,d_Atd_zz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Atd_yz_id,d_Atd_yz_id,d_Atd_yz_p_id,d_rk1Atd_yz_id,d_rk2Atd_yz_id,d_rk3Atd_yz_id,d_Atd_yz_id,d_Atd_yz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Atd_yy_id,d_Atd_yy_id,d_Atd_yy_p_id,d_rk1Atd_yy_id,d_rk2Atd_yy_id,d_rk3Atd_yy_id,d_Atd_yy_id,d_Atd_yy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Atd_xz_id,d_Atd_xz_id,d_Atd_xz_p_id,d_rk1Atd_xz_id,d_rk2Atd_xz_id,d_rk3Atd_xz_id,d_Atd_xz_id,d_Atd_xz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Atd_xy_id,d_Atd_xy_id,d_Atd_xy_p_id,d_rk1Atd_xy_id,d_rk2Atd_xy_id,d_rk3Atd_xy_id,d_Atd_xy_id,d_Atd_xy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Atd_xx_id,d_Atd_xx_id,d_Atd_xx_p_id,d_rk1Atd_xx_id,d_rk2Atd_xx_id,d_rk3Atd_xx_id,d_Atd_xx_id,d_Atd_xx_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_gtd_zz_id,d_gtd_zz_id,d_gtd_zz_p_id,d_rk1gtd_zz_id,d_rk2gtd_zz_id,d_rk3gtd_zz_id,d_gtd_zz_id,d_gtd_zz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_gtd_yz_id,d_gtd_yz_id,d_gtd_yz_p_id,d_rk1gtd_yz_id,d_rk2gtd_yz_id,d_rk3gtd_yz_id,d_gtd_yz_id,d_gtd_yz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_gtd_yy_id,d_gtd_yy_id,d_gtd_yy_p_id,d_rk1gtd_yy_id,d_rk2gtd_yy_id,d_rk3gtd_yy_id,d_gtd_yy_id,d_gtd_yy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_gtd_xz_id,d_gtd_xz_id,d_gtd_xz_p_id,d_rk1gtd_xz_id,d_rk2gtd_xz_id,d_rk3gtd_xz_id,d_gtd_xz_id,d_gtd_xz_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_gtd_xy_id,d_gtd_xy_id,d_gtd_xy_p_id,d_rk1gtd_xy_id,d_rk2gtd_xy_id,d_rk3gtd_xy_id,d_gtd_xy_id,d_gtd_xy_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_gtd_xx_id,d_gtd_xx_id,d_gtd_xx_p_id,d_rk1gtd_xx_id,d_rk2gtd_xx_id,d_rk3gtd_xx_id,d_gtd_xx_id,d_gtd_xx_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_phif_id,d_phif_id,d_phif_p_id,d_rk1phif_id,d_rk2phif_id,d_rk3phif_id,d_phif_id,d_phif_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_tauf_id,d_tauf_id,d_tauf_p_id,d_rk1tauf_id,d_rk2tauf_id,d_rk3tauf_id,d_tauf_id,d_tauf_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_DYf_id,d_DYf_id,d_DYf_p_id,d_rk1DYf_id,d_rk2DYf_id,d_rk3DYf_id,d_DYf_id,d_DYf_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Df_id,d_Df_id,d_Df_p_id,d_rk1Df_id,d_rk2Df_id,d_rk3Df_id,d_Df_id,d_Df_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Bfu_z_id,d_Bfu_z_id,d_Bfu_z_p_id,d_rk1Bfu_z_id,d_rk2Bfu_z_id,d_rk3Bfu_z_id,d_Bfu_z_id,d_Bfu_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Bfu_y_id,d_Bfu_y_id,d_Bfu_y_p_id,d_rk1Bfu_y_id,d_rk2Bfu_y_id,d_rk3Bfu_y_id,d_Bfu_y_id,d_Bfu_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Bfu_x_id,d_Bfu_x_id,d_Bfu_x_p_id,d_rk1Bfu_x_id,d_rk2Bfu_x_id,d_rk3Bfu_x_id,d_Bfu_x_id,d_Bfu_x_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Sfd_z_id,d_Sfd_z_id,d_Sfd_z_p_id,d_rk1Sfd_z_id,d_rk2Sfd_z_id,d_rk3Sfd_z_id,d_Sfd_z_id,d_Sfd_z_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Sfd_y_id,d_Sfd_y_id,d_Sfd_y_p_id,d_rk1Sfd_y_id,d_rk2Sfd_y_id,d_rk3Sfd_y_id,d_Sfd_y_id,d_Sfd_y_id,refine_operator,tio_mesh1);
		d_bdry_fill_advance23->registerRefine(d_Sfd_x_id,d_Sfd_x_id,d_Sfd_x_p_id,d_rk1Sfd_x_id,d_rk2Sfd_x_id,d_rk3Sfd_x_id,d_Sfd_x_id,d_Sfd_x_id,refine_operator,tio_mesh1);
	} else {
		d_bdry_fill_advance23->registerRefine(d_theta_id,d_theta_id,d_theta_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_trK_id,d_trK_id,d_trK_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_chi_id,d_chi_id,d_chi_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Alpha_id,d_Alpha_id,d_Alpha_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Betau_z_id,d_Betau_z_id,d_Betau_z_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Betau_y_id,d_Betau_y_id,d_Betau_y_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Betau_x_id,d_Betau_x_id,d_Betau_x_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Gamh_z_id,d_Gamh_z_id,d_Gamh_z_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Gamh_y_id,d_Gamh_y_id,d_Gamh_y_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Gamh_x_id,d_Gamh_x_id,d_Gamh_x_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Atd_zz_id,d_Atd_zz_id,d_Atd_zz_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Atd_yz_id,d_Atd_yz_id,d_Atd_yz_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Atd_yy_id,d_Atd_yy_id,d_Atd_yy_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Atd_xz_id,d_Atd_xz_id,d_Atd_xz_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Atd_xy_id,d_Atd_xy_id,d_Atd_xy_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Atd_xx_id,d_Atd_xx_id,d_Atd_xx_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_gtd_zz_id,d_gtd_zz_id,d_gtd_zz_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_gtd_yz_id,d_gtd_yz_id,d_gtd_yz_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_gtd_yy_id,d_gtd_yy_id,d_gtd_yy_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_gtd_xz_id,d_gtd_xz_id,d_gtd_xz_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_gtd_xy_id,d_gtd_xy_id,d_gtd_xy_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_gtd_xx_id,d_gtd_xx_id,d_gtd_xx_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_phif_id,d_phif_id,d_phif_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_tauf_id,d_tauf_id,d_tauf_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_DYf_id,d_DYf_id,d_DYf_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Df_id,d_Df_id,d_Df_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Bfu_z_id,d_Bfu_z_id,d_Bfu_z_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Bfu_y_id,d_Bfu_y_id,d_Bfu_y_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Bfu_x_id,d_Bfu_x_id,d_Bfu_x_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Sfd_z_id,d_Sfd_z_id,d_Sfd_z_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Sfd_y_id,d_Sfd_y_id,d_Sfd_y_id,refine_operator);
		d_bdry_fill_advance23->registerRefine(d_Sfd_x_id,d_Sfd_x_id,d_Sfd_x_id,refine_operator);
	}
	d_bdry_fill_advance26->registerRefine(d_rnu_id,d_rnu_id,d_rnu_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_qnu_id,d_qnu_id,d_qnu_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_qnu_x_id,d_qnu_x_id,d_qnu_x_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_qnu_a_id,d_qnu_a_id,d_qnu_a_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_qnu_e_id,d_qnu_e_id,d_qnu_e_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_chix_id,d_chix_id,d_chix_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_chia_id,d_chia_id,d_chia_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_chie_id,d_chie_id,d_chie_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_optdepthx_id,d_optdepthx_id,d_optdepthx_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_optdeptha_id,d_optdeptha_id,d_optdeptha_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_optdepthe_id,d_optdepthe_id,d_optdepthe_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauN_y_id,d_TauN_y_id,d_TauN_y_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauN_z_id,d_TauN_z_id,d_TauN_z_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauM_xz_id,d_TauM_xz_id,d_TauM_xz_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauT_xx_id,d_TauT_xx_id,d_TauT_xx_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauT_yy_id,d_TauT_yy_id,d_TauT_yy_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauT_zz_id,d_TauT_zz_id,d_TauT_zz_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauN_x_id,d_TauN_x_id,d_TauN_x_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauT_xz_id,d_TauT_xz_id,d_TauT_xz_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauNe_z_id,d_TauNe_z_id,d_TauNe_z_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauM_xy_id,d_TauM_xy_id,d_TauM_xy_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauM_yz_id,d_TauM_yz_id,d_TauM_yz_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauT_xy_id,d_TauT_xy_id,d_TauT_xy_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauT_yz_id,d_TauT_yz_id,d_TauT_yz_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauNe_y_id,d_TauNe_y_id,d_TauNe_y_id,refine_operator);
	d_bdry_fill_advance26->registerRefine(d_TauNe_x_id,d_TauNe_x_id,d_TauNe_x_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_flux_mass_id,d_flux_mass_id,d_flux_mass_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_Z_z_id,d_Z_z_id,d_Z_z_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_Z_y_id,d_Z_y_id,d_Z_y_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_Z_x_id,d_Z_x_id,d_Z_x_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_psi4I_id,d_psi4I_id,d_psi4I_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_psi4R_id,d_psi4R_id,d_psi4R_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_Jz_Komar_id,d_Jz_Komar_id,d_Jz_Komar_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_M_Komar_id,d_M_Komar_id,d_M_Komar_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_Jz_ADM_surf_id,d_Jz_ADM_surf_id,d_Jz_ADM_surf_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_M_ADM_surf_id,d_M_ADM_surf_id,d_M_ADM_surf_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_detgtm1_id,d_detgtm1_id,d_detgtm1_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_trA_id,d_trA_id,d_trA_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_MomCon_z_id,d_MomCon_z_id,d_MomCon_z_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_MomCon_y_id,d_MomCon_y_id,d_MomCon_y_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_MomCon_x_id,d_MomCon_x_id,d_MomCon_x_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_HamCon_id,d_HamCon_id,d_HamCon_id,refine_operator);
	d_bdry_fill_analysis1->registerRefine(d_Rscalar_id,d_Rscalar_id,d_Rscalar_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Sfd_x_id,d_Sfd_x_id,d_Sfd_x_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Sfd_x_id,d_Sfd_x_id,d_Sfd_x_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Sfd_y_id,d_Sfd_y_id,d_Sfd_y_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Sfd_y_id,d_Sfd_y_id,d_Sfd_y_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Sfd_z_id,d_Sfd_z_id,d_Sfd_z_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Sfd_z_id,d_Sfd_z_id,d_Sfd_z_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Bfu_x_id,d_Bfu_x_id,d_Bfu_x_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Bfu_x_id,d_Bfu_x_id,d_Bfu_x_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Bfu_y_id,d_Bfu_y_id,d_Bfu_y_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Bfu_y_id,d_Bfu_y_id,d_Bfu_y_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Bfu_z_id,d_Bfu_z_id,d_Bfu_z_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Bfu_z_id,d_Bfu_z_id,d_Bfu_z_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Df_id,d_Df_id,d_Df_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Df_id,d_Df_id,d_Df_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_DYf_id,d_DYf_id,d_DYf_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_DYf_id,d_DYf_id,d_DYf_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_tauf_id,d_tauf_id,d_tauf_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_tauf_id,d_tauf_id,d_tauf_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_phif_id,d_phif_id,d_phif_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_phif_id,d_phif_id,d_phif_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_gtd_xx_id,d_gtd_xx_id,d_gtd_xx_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_gtd_xx_id,d_gtd_xx_id,d_gtd_xx_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_gtd_xy_id,d_gtd_xy_id,d_gtd_xy_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_gtd_xy_id,d_gtd_xy_id,d_gtd_xy_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_gtd_xz_id,d_gtd_xz_id,d_gtd_xz_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_gtd_xz_id,d_gtd_xz_id,d_gtd_xz_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_gtd_yy_id,d_gtd_yy_id,d_gtd_yy_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_gtd_yy_id,d_gtd_yy_id,d_gtd_yy_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_gtd_yz_id,d_gtd_yz_id,d_gtd_yz_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_gtd_yz_id,d_gtd_yz_id,d_gtd_yz_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_gtd_zz_id,d_gtd_zz_id,d_gtd_zz_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_gtd_zz_id,d_gtd_zz_id,d_gtd_zz_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Atd_xx_id,d_Atd_xx_id,d_Atd_xx_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Atd_xx_id,d_Atd_xx_id,d_Atd_xx_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Atd_xy_id,d_Atd_xy_id,d_Atd_xy_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Atd_xy_id,d_Atd_xy_id,d_Atd_xy_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Atd_xz_id,d_Atd_xz_id,d_Atd_xz_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Atd_xz_id,d_Atd_xz_id,d_Atd_xz_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Atd_yy_id,d_Atd_yy_id,d_Atd_yy_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Atd_yy_id,d_Atd_yy_id,d_Atd_yy_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Atd_yz_id,d_Atd_yz_id,d_Atd_yz_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Atd_yz_id,d_Atd_yz_id,d_Atd_yz_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Atd_zz_id,d_Atd_zz_id,d_Atd_zz_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Atd_zz_id,d_Atd_zz_id,d_Atd_zz_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Gamh_x_id,d_Gamh_x_id,d_Gamh_x_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Gamh_x_id,d_Gamh_x_id,d_Gamh_x_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Gamh_y_id,d_Gamh_y_id,d_Gamh_y_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Gamh_y_id,d_Gamh_y_id,d_Gamh_y_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Gamh_z_id,d_Gamh_z_id,d_Gamh_z_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Gamh_z_id,d_Gamh_z_id,d_Gamh_z_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Betau_x_id,d_Betau_x_id,d_Betau_x_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Betau_x_id,d_Betau_x_id,d_Betau_x_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Betau_y_id,d_Betau_y_id,d_Betau_y_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Betau_y_id,d_Betau_y_id,d_Betau_y_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Betau_z_id,d_Betau_z_id,d_Betau_z_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Betau_z_id,d_Betau_z_id,d_Betau_z_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Alpha_id,d_Alpha_id,d_Alpha_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Alpha_id,d_Alpha_id,d_Alpha_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_chi_id,d_chi_id,d_chi_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_chi_id,d_chi_id,d_chi_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_trK_id,d_trK_id,d_trK_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_trK_id,d_trK_id,d_trK_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_theta_id,d_theta_id,d_theta_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_theta_id,d_theta_id,d_theta_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_rhof_id,d_rhof_id,d_rhof_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_rhof_id,d_rhof_id,d_rhof_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Yef_id,d_Yef_id,d_Yef_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Yef_id,d_Yef_id,d_Yef_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_vfd_x_id,d_vfd_x_id,d_vfd_x_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_vfd_x_id,d_vfd_x_id,d_vfd_x_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_vfd_y_id,d_vfd_y_id,d_vfd_y_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_vfd_y_id,d_vfd_y_id,d_vfd_y_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_vfd_z_id,d_vfd_z_id,d_vfd_z_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_vfd_z_id,d_vfd_z_id,d_vfd_z_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_pf_id,d_pf_id,d_pf_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_pf_id,d_pf_id,d_pf_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_epsf_id,d_epsf_id,d_epsf_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_epsf_id,d_epsf_id,d_epsf_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Tf_id,d_Tf_id,d_Tf_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Tf_id,d_Tf_id,d_Tf_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_sqcs_id,d_sqcs_id,d_sqcs_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_sqcs_id,d_sqcs_id,d_sqcs_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_qnu_id,d_qnu_id,d_qnu_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_qnu_id,d_qnu_id,d_qnu_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_rnu_id,d_rnu_id,d_rnu_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_rnu_id,d_rnu_id,d_rnu_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_optdepthe_id,d_optdepthe_id,d_optdepthe_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_optdepthe_id,d_optdepthe_id,d_optdepthe_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_optdeptha_id,d_optdeptha_id,d_optdeptha_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_optdeptha_id,d_optdeptha_id,d_optdeptha_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_optdepthx_id,d_optdepthx_id,d_optdepthx_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_optdepthx_id,d_optdepthx_id,d_optdepthx_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_chie_id,d_chie_id,d_chie_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_chie_id,d_chie_id,d_chie_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_chia_id,d_chia_id,d_chia_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_chia_id,d_chia_id,d_chia_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_chix_id,d_chix_id,d_chix_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_chix_id,d_chix_id,d_chix_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_t_optdepthe_id,d_t_optdepthe_id,d_t_optdepthe_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_t_optdepthe_id,d_t_optdepthe_id,d_t_optdepthe_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_t_optdeptha_id,d_t_optdeptha_id,d_t_optdeptha_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_t_optdeptha_id,d_t_optdeptha_id,d_t_optdeptha_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_t_optdepthx_id,d_t_optdepthx_id,d_t_optdepthx_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_t_optdepthx_id,d_t_optdepthx_id,d_t_optdepthx_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_t_chie_id,d_t_chie_id,d_t_chie_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_t_chie_id,d_t_chie_id,d_t_chie_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_t_chia_id,d_t_chia_id,d_t_chia_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_t_chia_id,d_t_chia_id,d_t_chia_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_t_chix_id,d_t_chix_id,d_t_chix_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_t_chix_id,d_t_chix_id,d_t_chix_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Efu_x_id,d_Efu_x_id,d_Efu_x_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Efu_x_id,d_Efu_x_id,d_Efu_x_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Efu_y_id,d_Efu_y_id,d_Efu_y_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Efu_y_id,d_Efu_y_id,d_Efu_y_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_Efu_z_id,d_Efu_z_id,d_Efu_z_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_Efu_z_id,d_Efu_z_id,d_Efu_z_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauN_x_id,d_TauN_x_id,d_TauN_x_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauN_x_id,d_TauN_x_id,d_TauN_x_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauN_y_id,d_TauN_y_id,d_TauN_y_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauN_y_id,d_TauN_y_id,d_TauN_y_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauN_z_id,d_TauN_z_id,d_TauN_z_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauN_z_id,d_TauN_z_id,d_TauN_z_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauNe_x_id,d_TauNe_x_id,d_TauNe_x_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauNe_x_id,d_TauNe_x_id,d_TauNe_x_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauNe_y_id,d_TauNe_y_id,d_TauNe_y_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauNe_y_id,d_TauNe_y_id,d_TauNe_y_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauNe_z_id,d_TauNe_z_id,d_TauNe_z_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauNe_z_id,d_TauNe_z_id,d_TauNe_z_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauM_xy_id,d_TauM_xy_id,d_TauM_xy_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauM_xy_id,d_TauM_xy_id,d_TauM_xy_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauM_xz_id,d_TauM_xz_id,d_TauM_xz_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauM_xz_id,d_TauM_xz_id,d_TauM_xz_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauM_yz_id,d_TauM_yz_id,d_TauM_yz_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauM_yz_id,d_TauM_yz_id,d_TauM_yz_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauT_xx_id,d_TauT_xx_id,d_TauT_xx_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauT_xx_id,d_TauT_xx_id,d_TauT_xx_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauT_xy_id,d_TauT_xy_id,d_TauT_xy_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauT_xy_id,d_TauT_xy_id,d_TauT_xy_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauT_xz_id,d_TauT_xz_id,d_TauT_xz_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauT_xz_id,d_TauT_xz_id,d_TauT_xz_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauT_yy_id,d_TauT_yy_id,d_TauT_yy_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauT_yy_id,d_TauT_yy_id,d_TauT_yy_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauT_yz_id,d_TauT_yz_id,d_TauT_yz_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauT_yz_id,d_TauT_yz_id,d_TauT_yz_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_TauT_zz_id,d_TauT_zz_id,d_TauT_zz_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_TauT_zz_id,d_TauT_zz_id,d_TauT_zz_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_dpfdeps_id,d_dpfdeps_id,d_dpfdeps_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_dpfdeps_id,d_dpfdeps_id,d_dpfdeps_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_dpfdrho_id,d_dpfdrho_id,d_dpfdrho_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_dpfdrho_id,d_dpfdrho_id,d_dpfdrho_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_dpfdye_id,d_dpfdye_id,d_dpfdye_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_dpfdye_id,d_dpfdye_id,d_dpfdye_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_qnu_a_id,d_qnu_a_id,d_qnu_a_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_qnu_a_id,d_qnu_a_id,d_qnu_a_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_qnu_e_id,d_qnu_e_id,d_qnu_e_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_qnu_e_id,d_qnu_e_id,d_qnu_e_id,refine_operator);
	d_bdry_fill_init->registerRefine(d_qnu_x_id,d_qnu_x_id,d_qnu_x_id,refine_operator);
	d_bdry_post_coarsen->registerRefine(d_qnu_x_id,d_qnu_x_id,d_qnu_x_id,refine_operator);


	//Register variables to the refineAlgorithm for filling new levels on regridding
	d_fill_new_level->registerRefine(d_Sfd_x_id,d_Sfd_x_id,d_Sfd_x_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Sfd_y_id,d_Sfd_y_id,d_Sfd_y_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Sfd_z_id,d_Sfd_z_id,d_Sfd_z_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Bfu_x_id,d_Bfu_x_id,d_Bfu_x_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Bfu_y_id,d_Bfu_y_id,d_Bfu_y_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Bfu_z_id,d_Bfu_z_id,d_Bfu_z_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Df_id,d_Df_id,d_Df_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_DYf_id,d_DYf_id,d_DYf_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_tauf_id,d_tauf_id,d_tauf_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_phif_id,d_phif_id,d_phif_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_gtd_xx_id,d_gtd_xx_id,d_gtd_xx_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_gtd_xy_id,d_gtd_xy_id,d_gtd_xy_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_gtd_xz_id,d_gtd_xz_id,d_gtd_xz_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_gtd_yy_id,d_gtd_yy_id,d_gtd_yy_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_gtd_yz_id,d_gtd_yz_id,d_gtd_yz_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_gtd_zz_id,d_gtd_zz_id,d_gtd_zz_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Atd_xx_id,d_Atd_xx_id,d_Atd_xx_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Atd_xy_id,d_Atd_xy_id,d_Atd_xy_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Atd_xz_id,d_Atd_xz_id,d_Atd_xz_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Atd_yy_id,d_Atd_yy_id,d_Atd_yy_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Atd_yz_id,d_Atd_yz_id,d_Atd_yz_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Atd_zz_id,d_Atd_zz_id,d_Atd_zz_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Gamh_x_id,d_Gamh_x_id,d_Gamh_x_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Gamh_y_id,d_Gamh_y_id,d_Gamh_y_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Gamh_z_id,d_Gamh_z_id,d_Gamh_z_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Betau_x_id,d_Betau_x_id,d_Betau_x_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Betau_y_id,d_Betau_y_id,d_Betau_y_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Betau_z_id,d_Betau_z_id,d_Betau_z_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Alpha_id,d_Alpha_id,d_Alpha_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_chi_id,d_chi_id,d_chi_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_trK_id,d_trK_id,d_trK_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_theta_id,d_theta_id,d_theta_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_rhof_id,d_rhof_id,d_rhof_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_rhof_id,d_rhof_id,d_rhof_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_Yef_id,d_Yef_id,d_Yef_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Yef_id,d_Yef_id,d_Yef_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_vfd_x_id,d_vfd_x_id,d_vfd_x_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_vfd_x_id,d_vfd_x_id,d_vfd_x_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_vfd_y_id,d_vfd_y_id,d_vfd_y_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_vfd_y_id,d_vfd_y_id,d_vfd_y_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_vfd_z_id,d_vfd_z_id,d_vfd_z_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_vfd_z_id,d_vfd_z_id,d_vfd_z_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_pf_id,d_pf_id,d_pf_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_pf_id,d_pf_id,d_pf_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_epsf_id,d_epsf_id,d_epsf_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_epsf_id,d_epsf_id,d_epsf_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_Tf_id,d_Tf_id,d_Tf_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Tf_id,d_Tf_id,d_Tf_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_sqcs_id,d_sqcs_id,d_sqcs_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_sqcs_id,d_sqcs_id,d_sqcs_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_qnu_id,d_qnu_id,d_qnu_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_qnu_id,d_qnu_id,d_qnu_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_rnu_id,d_rnu_id,d_rnu_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_rnu_id,d_rnu_id,d_rnu_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_optdepthe_id,d_optdepthe_id,d_optdepthe_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_optdepthe_id,d_optdepthe_id,d_optdepthe_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_optdeptha_id,d_optdeptha_id,d_optdeptha_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_optdeptha_id,d_optdeptha_id,d_optdeptha_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_optdepthx_id,d_optdepthx_id,d_optdepthx_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_optdepthx_id,d_optdepthx_id,d_optdepthx_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_chie_id,d_chie_id,d_chie_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_chie_id,d_chie_id,d_chie_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_chia_id,d_chia_id,d_chia_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_chia_id,d_chia_id,d_chia_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_chix_id,d_chix_id,d_chix_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_chix_id,d_chix_id,d_chix_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_t_optdepthe_id,d_t_optdepthe_id,d_t_optdepthe_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_t_optdepthe_id,d_t_optdepthe_id,d_t_optdepthe_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_t_optdeptha_id,d_t_optdeptha_id,d_t_optdeptha_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_t_optdeptha_id,d_t_optdeptha_id,d_t_optdeptha_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_t_optdepthx_id,d_t_optdepthx_id,d_t_optdepthx_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_t_optdepthx_id,d_t_optdepthx_id,d_t_optdepthx_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_t_chie_id,d_t_chie_id,d_t_chie_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_t_chie_id,d_t_chie_id,d_t_chie_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_t_chia_id,d_t_chia_id,d_t_chia_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_t_chia_id,d_t_chia_id,d_t_chia_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_t_chix_id,d_t_chix_id,d_t_chix_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_t_chix_id,d_t_chix_id,d_t_chix_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_Efu_x_id,d_Efu_x_id,d_Efu_x_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Efu_x_id,d_Efu_x_id,d_Efu_x_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_Efu_y_id,d_Efu_y_id,d_Efu_y_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Efu_y_id,d_Efu_y_id,d_Efu_y_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_Efu_z_id,d_Efu_z_id,d_Efu_z_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Efu_z_id,d_Efu_z_id,d_Efu_z_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauN_x_id,d_TauN_x_id,d_TauN_x_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauN_x_id,d_TauN_x_id,d_TauN_x_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauN_y_id,d_TauN_y_id,d_TauN_y_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauN_y_id,d_TauN_y_id,d_TauN_y_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauN_z_id,d_TauN_z_id,d_TauN_z_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauN_z_id,d_TauN_z_id,d_TauN_z_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauNe_x_id,d_TauNe_x_id,d_TauNe_x_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauNe_x_id,d_TauNe_x_id,d_TauNe_x_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauNe_y_id,d_TauNe_y_id,d_TauNe_y_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauNe_y_id,d_TauNe_y_id,d_TauNe_y_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauNe_z_id,d_TauNe_z_id,d_TauNe_z_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauNe_z_id,d_TauNe_z_id,d_TauNe_z_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauM_xy_id,d_TauM_xy_id,d_TauM_xy_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauM_xy_id,d_TauM_xy_id,d_TauM_xy_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauM_xz_id,d_TauM_xz_id,d_TauM_xz_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauM_xz_id,d_TauM_xz_id,d_TauM_xz_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauM_yz_id,d_TauM_yz_id,d_TauM_yz_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauM_yz_id,d_TauM_yz_id,d_TauM_yz_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauT_xx_id,d_TauT_xx_id,d_TauT_xx_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauT_xx_id,d_TauT_xx_id,d_TauT_xx_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauT_xy_id,d_TauT_xy_id,d_TauT_xy_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauT_xy_id,d_TauT_xy_id,d_TauT_xy_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauT_xz_id,d_TauT_xz_id,d_TauT_xz_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauT_xz_id,d_TauT_xz_id,d_TauT_xz_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauT_yy_id,d_TauT_yy_id,d_TauT_yy_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauT_yy_id,d_TauT_yy_id,d_TauT_yy_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauT_yz_id,d_TauT_yz_id,d_TauT_yz_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauT_yz_id,d_TauT_yz_id,d_TauT_yz_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_TauT_zz_id,d_TauT_zz_id,d_TauT_zz_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_TauT_zz_id,d_TauT_zz_id,d_TauT_zz_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_dpfdeps_id,d_dpfdeps_id,d_dpfdeps_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_dpfdeps_id,d_dpfdeps_id,d_dpfdeps_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_dpfdrho_id,d_dpfdrho_id,d_dpfdrho_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_dpfdrho_id,d_dpfdrho_id,d_dpfdrho_id,refine_operator_bound);
	d_fill_new_level_aux->registerRefine(d_dpfdye_id,d_dpfdye_id,d_dpfdye_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_dpfdye_id,d_dpfdye_id,d_dpfdye_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_qnu_a_id,d_qnu_a_id,d_qnu_a_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_qnu_a_id,d_qnu_a_id,d_qnu_a_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_qnu_e_id,d_qnu_e_id,d_qnu_e_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_qnu_e_id,d_qnu_e_id,d_qnu_e_id,refine_operator_bound);
	d_fill_new_level_der_aux->registerRefine(d_qnu_x_id,d_qnu_x_id,d_qnu_x_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_qnu_x_id,d_qnu_x_id,d_qnu_x_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Rscalar_id,d_Rscalar_id,d_Rscalar_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_HamCon_id,d_HamCon_id,d_HamCon_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_MomCon_x_id,d_MomCon_x_id,d_MomCon_x_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_MomCon_y_id,d_MomCon_y_id,d_MomCon_y_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_MomCon_z_id,d_MomCon_z_id,d_MomCon_z_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_trA_id,d_trA_id,d_trA_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_detgtm1_id,d_detgtm1_id,d_detgtm1_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_M_ADM_surf_id,d_M_ADM_surf_id,d_M_ADM_surf_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Jz_ADM_surf_id,d_Jz_ADM_surf_id,d_Jz_ADM_surf_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_M_Komar_id,d_M_Komar_id,d_M_Komar_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Jz_Komar_id,d_Jz_Komar_id,d_Jz_Komar_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_psi4R_id,d_psi4R_id,d_psi4R_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_psi4I_id,d_psi4I_id,d_psi4I_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Z_x_id,d_Z_x_id,d_Z_x_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Z_y_id,d_Z_y_id,d_Z_y_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_Z_z_id,d_Z_z_id,d_Z_z_id,refine_operator_bound);
	d_fill_new_level->registerRefine(d_flux_mass_id,d_flux_mass_id,d_flux_mass_id,refine_operator_bound);


	//Register variables to the coarsenAlgorithm
	d_coarsen_algorithm->registerCoarsen(d_Sfd_x_id,d_Sfd_x_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Sfd_y_id,d_Sfd_y_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Sfd_z_id,d_Sfd_z_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Bfu_x_id,d_Bfu_x_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Bfu_y_id,d_Bfu_y_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Bfu_z_id,d_Bfu_z_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Df_id,d_Df_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_DYf_id,d_DYf_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_tauf_id,d_tauf_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_phif_id,d_phif_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_gtd_xx_id,d_gtd_xx_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_gtd_xy_id,d_gtd_xy_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_gtd_xz_id,d_gtd_xz_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_gtd_yy_id,d_gtd_yy_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_gtd_yz_id,d_gtd_yz_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_gtd_zz_id,d_gtd_zz_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Atd_xx_id,d_Atd_xx_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Atd_xy_id,d_Atd_xy_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Atd_xz_id,d_Atd_xz_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Atd_yy_id,d_Atd_yy_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Atd_yz_id,d_Atd_yz_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Atd_zz_id,d_Atd_zz_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Gamh_x_id,d_Gamh_x_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Gamh_y_id,d_Gamh_y_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Gamh_z_id,d_Gamh_z_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Betau_x_id,d_Betau_x_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Betau_y_id,d_Betau_y_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Betau_z_id,d_Betau_z_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Alpha_id,d_Alpha_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_chi_id,d_chi_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_trK_id,d_trK_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_theta_id,d_theta_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_rhof_id,d_rhof_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Yef_id,d_Yef_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_vfd_x_id,d_vfd_x_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_vfd_y_id,d_vfd_y_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_vfd_z_id,d_vfd_z_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_pf_id,d_pf_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_epsf_id,d_epsf_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Tf_id,d_Tf_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_sqcs_id,d_sqcs_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_qnu_id,d_qnu_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_rnu_id,d_rnu_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_optdepthe_id,d_optdepthe_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_optdeptha_id,d_optdeptha_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_optdepthx_id,d_optdepthx_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_chie_id,d_chie_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_chia_id,d_chia_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_chix_id,d_chix_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_t_optdepthe_id,d_t_optdepthe_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_t_optdeptha_id,d_t_optdeptha_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_t_optdepthx_id,d_t_optdepthx_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_t_chie_id,d_t_chie_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_t_chia_id,d_t_chia_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_t_chix_id,d_t_chix_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Efu_x_id,d_Efu_x_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Efu_y_id,d_Efu_y_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_Efu_z_id,d_Efu_z_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauN_x_id,d_TauN_x_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauN_y_id,d_TauN_y_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauN_z_id,d_TauN_z_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauNe_x_id,d_TauNe_x_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauNe_y_id,d_TauNe_y_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauNe_z_id,d_TauNe_z_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauM_xy_id,d_TauM_xy_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauM_xz_id,d_TauM_xz_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauM_yz_id,d_TauM_yz_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauT_xx_id,d_TauT_xx_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauT_xy_id,d_TauT_xy_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauT_xz_id,d_TauT_xz_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauT_yy_id,d_TauT_yy_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauT_yz_id,d_TauT_yz_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_TauT_zz_id,d_TauT_zz_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_dpfdeps_id,d_dpfdeps_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_dpfdrho_id,d_dpfdrho_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_dpfdye_id,d_dpfdye_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_qnu_a_id,d_qnu_a_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_qnu_e_id,d_qnu_e_id,coarsen_operator);
	d_coarsen_algorithm->registerCoarsen(d_qnu_x_id,d_qnu_x_id,coarsen_operator);

	readTable("DataLeakage.h5", &coord0_data_DataLeakage, coord0_size_DataLeakage, coord0_dx_DataLeakage, &coord1_data_DataLeakage, coord1_size_DataLeakage, coord1_dx_DataLeakage, &coord2_data_DataLeakage, coord2_size_DataLeakage, coord2_dx_DataLeakage, &vars_data_DataLeakage, vars_size_DataLeakage);
	readTable("Data.h5", &coord0_data_Data, coord0_size_Data, coord0_dx_Data, &coord1_data_Data, coord1_size_Data, coord1_dx_Data, &coord2_data_Data, coord2_size_Data, coord2_dx_Data, &vars_data_Data, vars_size_Data);


    Commons::initialization();
}

/*
 * Destructor.
 */
Problem::~Problem() 
{
} 

/*
 * Initialize the data from a given level.
 */
void Problem::initializeLevelData (
   const std::shared_ptr<hier::PatchHierarchy >& hierarchy , 
   const int level_number ,
   const double init_data_time ,
   const bool can_be_refined ,
   const bool initial_time ,
   const std::shared_ptr<hier::PatchLevel >& old_level ,
   const bool allocate_data )
{
    cout<<"Initializing level "<<level_number<<endl;
    tbox::MemoryUtilities::printMemoryInfo(cout);   
	std::shared_ptr< hier::PatchLevel > level(hierarchy->getPatchLevel(level_number));

	// Allocate storage needed to initialize level and fill data from coarser levels in AMR hierarchy.  
	level->allocatePatchData(d_interior_regridding_value_id, init_data_time);
	level->allocatePatchData(d_nonSync_regridding_tag_id, init_data_time);
	level->allocatePatchData(d_interior_i_id, init_data_time);
	level->allocatePatchData(d_interior_j_id, init_data_time);
	level->allocatePatchData(d_interior_k_id, init_data_time);
	level->allocatePatchData(d_FOV_1_id, init_data_time);
	level->allocatePatchData(d_FOV_xLower_id, init_data_time);
	level->allocatePatchData(d_FOV_xUpper_id, init_data_time);
	level->allocatePatchData(d_FOV_yLower_id, init_data_time);
	level->allocatePatchData(d_FOV_yUpper_id, init_data_time);
	level->allocatePatchData(d_FOV_zLower_id, init_data_time);
	level->allocatePatchData(d_FOV_zUpper_id, init_data_time);
	level->allocatePatchData(d_Sfd_x_id, init_data_time);
	level->allocatePatchData(d_Sfd_y_id, init_data_time);
	level->allocatePatchData(d_Sfd_z_id, init_data_time);
	level->allocatePatchData(d_Bfu_x_id, init_data_time);
	level->allocatePatchData(d_Bfu_y_id, init_data_time);
	level->allocatePatchData(d_Bfu_z_id, init_data_time);
	level->allocatePatchData(d_Df_id, init_data_time);
	level->allocatePatchData(d_DYf_id, init_data_time);
	level->allocatePatchData(d_tauf_id, init_data_time);
	level->allocatePatchData(d_phif_id, init_data_time);
	level->allocatePatchData(d_gtd_xx_id, init_data_time);
	level->allocatePatchData(d_gtd_xy_id, init_data_time);
	level->allocatePatchData(d_gtd_xz_id, init_data_time);
	level->allocatePatchData(d_gtd_yy_id, init_data_time);
	level->allocatePatchData(d_gtd_yz_id, init_data_time);
	level->allocatePatchData(d_gtd_zz_id, init_data_time);
	level->allocatePatchData(d_Atd_xx_id, init_data_time);
	level->allocatePatchData(d_Atd_xy_id, init_data_time);
	level->allocatePatchData(d_Atd_xz_id, init_data_time);
	level->allocatePatchData(d_Atd_yy_id, init_data_time);
	level->allocatePatchData(d_Atd_yz_id, init_data_time);
	level->allocatePatchData(d_Atd_zz_id, init_data_time);
	level->allocatePatchData(d_Gamh_x_id, init_data_time);
	level->allocatePatchData(d_Gamh_y_id, init_data_time);
	level->allocatePatchData(d_Gamh_z_id, init_data_time);
	level->allocatePatchData(d_Betau_x_id, init_data_time);
	level->allocatePatchData(d_Betau_y_id, init_data_time);
	level->allocatePatchData(d_Betau_z_id, init_data_time);
	level->allocatePatchData(d_Alpha_id, init_data_time);
	level->allocatePatchData(d_chi_id, init_data_time);
	level->allocatePatchData(d_trK_id, init_data_time);
	level->allocatePatchData(d_theta_id, init_data_time);
	level->allocatePatchData(d_rhof_id, init_data_time);
	level->allocatePatchData(d_Yef_id, init_data_time);
	level->allocatePatchData(d_vfd_x_id, init_data_time);
	level->allocatePatchData(d_vfd_y_id, init_data_time);
	level->allocatePatchData(d_vfd_z_id, init_data_time);
	level->allocatePatchData(d_pf_id, init_data_time);
	level->allocatePatchData(d_epsf_id, init_data_time);
	level->allocatePatchData(d_Tf_id, init_data_time);
	level->allocatePatchData(d_sqcs_id, init_data_time);
	level->allocatePatchData(d_qnu_id, init_data_time);
	level->allocatePatchData(d_rnu_id, init_data_time);
	level->allocatePatchData(d_optdepthe_id, init_data_time);
	level->allocatePatchData(d_optdeptha_id, init_data_time);
	level->allocatePatchData(d_optdepthx_id, init_data_time);
	level->allocatePatchData(d_chie_id, init_data_time);
	level->allocatePatchData(d_chia_id, init_data_time);
	level->allocatePatchData(d_chix_id, init_data_time);
	level->allocatePatchData(d_t_optdepthe_id, init_data_time);
	level->allocatePatchData(d_t_optdeptha_id, init_data_time);
	level->allocatePatchData(d_t_optdepthx_id, init_data_time);
	level->allocatePatchData(d_t_chie_id, init_data_time);
	level->allocatePatchData(d_t_chia_id, init_data_time);
	level->allocatePatchData(d_t_chix_id, init_data_time);
	level->allocatePatchData(d_Efu_x_id, init_data_time);
	level->allocatePatchData(d_Efu_y_id, init_data_time);
	level->allocatePatchData(d_Efu_z_id, init_data_time);
	level->allocatePatchData(d_TauN_x_id, init_data_time);
	level->allocatePatchData(d_TauN_y_id, init_data_time);
	level->allocatePatchData(d_TauN_z_id, init_data_time);
	level->allocatePatchData(d_TauNe_x_id, init_data_time);
	level->allocatePatchData(d_TauNe_y_id, init_data_time);
	level->allocatePatchData(d_TauNe_z_id, init_data_time);
	level->allocatePatchData(d_TauM_xy_id, init_data_time);
	level->allocatePatchData(d_TauM_xz_id, init_data_time);
	level->allocatePatchData(d_TauM_yz_id, init_data_time);
	level->allocatePatchData(d_TauT_xx_id, init_data_time);
	level->allocatePatchData(d_TauT_xy_id, init_data_time);
	level->allocatePatchData(d_TauT_xz_id, init_data_time);
	level->allocatePatchData(d_TauT_yy_id, init_data_time);
	level->allocatePatchData(d_TauT_yz_id, init_data_time);
	level->allocatePatchData(d_TauT_zz_id, init_data_time);
	level->allocatePatchData(d_dpfdeps_id, init_data_time);
	level->allocatePatchData(d_dpfdrho_id, init_data_time);
	level->allocatePatchData(d_dpfdye_id, init_data_time);
	level->allocatePatchData(d_qnu_a_id, init_data_time);
	level->allocatePatchData(d_qnu_e_id, init_data_time);
	level->allocatePatchData(d_qnu_x_id, init_data_time);
	level->allocatePatchData(d_Rscalar_id, init_data_time);
	level->allocatePatchData(d_HamCon_id, init_data_time);
	level->allocatePatchData(d_MomCon_x_id, init_data_time);
	level->allocatePatchData(d_MomCon_y_id, init_data_time);
	level->allocatePatchData(d_MomCon_z_id, init_data_time);
	level->allocatePatchData(d_trA_id, init_data_time);
	level->allocatePatchData(d_detgtm1_id, init_data_time);
	level->allocatePatchData(d_M_ADM_surf_id, init_data_time);
	level->allocatePatchData(d_Jz_ADM_surf_id, init_data_time);
	level->allocatePatchData(d_M_Komar_id, init_data_time);
	level->allocatePatchData(d_Jz_Komar_id, init_data_time);
	level->allocatePatchData(d_psi4R_id, init_data_time);
	level->allocatePatchData(d_psi4I_id, init_data_time);
	level->allocatePatchData(d_Z_x_id, init_data_time);
	level->allocatePatchData(d_Z_y_id, init_data_time);
	level->allocatePatchData(d_Z_z_id, init_data_time);
	level->allocatePatchData(d_flux_mass_id, init_data_time);
	level->allocatePatchData(d_rk1Sfd_x_id, init_data_time);
	level->allocatePatchData(d_rk1Sfd_y_id, init_data_time);
	level->allocatePatchData(d_rk1Sfd_z_id, init_data_time);
	level->allocatePatchData(d_rk1Bfu_x_id, init_data_time);
	level->allocatePatchData(d_rk1Bfu_y_id, init_data_time);
	level->allocatePatchData(d_rk1Bfu_z_id, init_data_time);
	level->allocatePatchData(d_rk1Df_id, init_data_time);
	level->allocatePatchData(d_rk1DYf_id, init_data_time);
	level->allocatePatchData(d_rk1tauf_id, init_data_time);
	level->allocatePatchData(d_rk1phif_id, init_data_time);
	level->allocatePatchData(d_rk1gtd_xx_id, init_data_time);
	level->allocatePatchData(d_rk1gtd_xy_id, init_data_time);
	level->allocatePatchData(d_rk1gtd_xz_id, init_data_time);
	level->allocatePatchData(d_rk1gtd_yy_id, init_data_time);
	level->allocatePatchData(d_rk1gtd_yz_id, init_data_time);
	level->allocatePatchData(d_rk1gtd_zz_id, init_data_time);
	level->allocatePatchData(d_rk1Atd_xx_id, init_data_time);
	level->allocatePatchData(d_rk1Atd_xy_id, init_data_time);
	level->allocatePatchData(d_rk1Atd_xz_id, init_data_time);
	level->allocatePatchData(d_rk1Atd_yy_id, init_data_time);
	level->allocatePatchData(d_rk1Atd_yz_id, init_data_time);
	level->allocatePatchData(d_rk1Atd_zz_id, init_data_time);
	level->allocatePatchData(d_rk1Gamh_x_id, init_data_time);
	level->allocatePatchData(d_rk1Gamh_y_id, init_data_time);
	level->allocatePatchData(d_rk1Gamh_z_id, init_data_time);
	level->allocatePatchData(d_rk1Betau_x_id, init_data_time);
	level->allocatePatchData(d_rk1Betau_y_id, init_data_time);
	level->allocatePatchData(d_rk1Betau_z_id, init_data_time);
	level->allocatePatchData(d_rk1Alpha_id, init_data_time);
	level->allocatePatchData(d_rk1chi_id, init_data_time);
	level->allocatePatchData(d_rk1trK_id, init_data_time);
	level->allocatePatchData(d_rk1theta_id, init_data_time);
	level->allocatePatchData(d_rk2Sfd_x_id, init_data_time);
	level->allocatePatchData(d_rk2Sfd_y_id, init_data_time);
	level->allocatePatchData(d_rk2Sfd_z_id, init_data_time);
	level->allocatePatchData(d_rk2Bfu_x_id, init_data_time);
	level->allocatePatchData(d_rk2Bfu_y_id, init_data_time);
	level->allocatePatchData(d_rk2Bfu_z_id, init_data_time);
	level->allocatePatchData(d_rk2Df_id, init_data_time);
	level->allocatePatchData(d_rk2DYf_id, init_data_time);
	level->allocatePatchData(d_rk2tauf_id, init_data_time);
	level->allocatePatchData(d_rk2phif_id, init_data_time);
	level->allocatePatchData(d_rk2gtd_xx_id, init_data_time);
	level->allocatePatchData(d_rk2gtd_xy_id, init_data_time);
	level->allocatePatchData(d_rk2gtd_xz_id, init_data_time);
	level->allocatePatchData(d_rk2gtd_yy_id, init_data_time);
	level->allocatePatchData(d_rk2gtd_yz_id, init_data_time);
	level->allocatePatchData(d_rk2gtd_zz_id, init_data_time);
	level->allocatePatchData(d_rk2Atd_xx_id, init_data_time);
	level->allocatePatchData(d_rk2Atd_xy_id, init_data_time);
	level->allocatePatchData(d_rk2Atd_xz_id, init_data_time);
	level->allocatePatchData(d_rk2Atd_yy_id, init_data_time);
	level->allocatePatchData(d_rk2Atd_yz_id, init_data_time);
	level->allocatePatchData(d_rk2Atd_zz_id, init_data_time);
	level->allocatePatchData(d_rk2Gamh_x_id, init_data_time);
	level->allocatePatchData(d_rk2Gamh_y_id, init_data_time);
	level->allocatePatchData(d_rk2Gamh_z_id, init_data_time);
	level->allocatePatchData(d_rk2Betau_x_id, init_data_time);
	level->allocatePatchData(d_rk2Betau_y_id, init_data_time);
	level->allocatePatchData(d_rk2Betau_z_id, init_data_time);
	level->allocatePatchData(d_rk2Alpha_id, init_data_time);
	level->allocatePatchData(d_rk2chi_id, init_data_time);
	level->allocatePatchData(d_rk2trK_id, init_data_time);
	level->allocatePatchData(d_rk2theta_id, init_data_time);
	level->allocatePatchData(d_rk3Sfd_x_id, init_data_time);
	level->allocatePatchData(d_rk3Sfd_y_id, init_data_time);
	level->allocatePatchData(d_rk3Sfd_z_id, init_data_time);
	level->allocatePatchData(d_rk3Bfu_x_id, init_data_time);
	level->allocatePatchData(d_rk3Bfu_y_id, init_data_time);
	level->allocatePatchData(d_rk3Bfu_z_id, init_data_time);
	level->allocatePatchData(d_rk3Df_id, init_data_time);
	level->allocatePatchData(d_rk3DYf_id, init_data_time);
	level->allocatePatchData(d_rk3tauf_id, init_data_time);
	level->allocatePatchData(d_rk3phif_id, init_data_time);
	level->allocatePatchData(d_rk3gtd_xx_id, init_data_time);
	level->allocatePatchData(d_rk3gtd_xy_id, init_data_time);
	level->allocatePatchData(d_rk3gtd_xz_id, init_data_time);
	level->allocatePatchData(d_rk3gtd_yy_id, init_data_time);
	level->allocatePatchData(d_rk3gtd_yz_id, init_data_time);
	level->allocatePatchData(d_rk3gtd_zz_id, init_data_time);
	level->allocatePatchData(d_rk3Atd_xx_id, init_data_time);
	level->allocatePatchData(d_rk3Atd_xy_id, init_data_time);
	level->allocatePatchData(d_rk3Atd_xz_id, init_data_time);
	level->allocatePatchData(d_rk3Atd_yy_id, init_data_time);
	level->allocatePatchData(d_rk3Atd_yz_id, init_data_time);
	level->allocatePatchData(d_rk3Atd_zz_id, init_data_time);
	level->allocatePatchData(d_rk3Gamh_x_id, init_data_time);
	level->allocatePatchData(d_rk3Gamh_y_id, init_data_time);
	level->allocatePatchData(d_rk3Gamh_z_id, init_data_time);
	level->allocatePatchData(d_rk3Betau_x_id, init_data_time);
	level->allocatePatchData(d_rk3Betau_y_id, init_data_time);
	level->allocatePatchData(d_rk3Betau_z_id, init_data_time);
	level->allocatePatchData(d_rk3Alpha_id, init_data_time);
	level->allocatePatchData(d_rk3chi_id, init_data_time);
	level->allocatePatchData(d_rk3trK_id, init_data_time);
	level->allocatePatchData(d_rk3theta_id, init_data_time);
	level->allocatePatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id, init_data_time);
	level->allocatePatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id, init_data_time);
	level->allocatePatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id, init_data_time);
	level->allocatePatchData(d_stalled_1_id, init_data_time);
	level->allocatePatchData(d_Sfd_x_p_id, init_data_time);
	level->allocatePatchData(d_Sfd_y_p_id, init_data_time);
	level->allocatePatchData(d_Sfd_z_p_id, init_data_time);
	level->allocatePatchData(d_Bfu_x_p_id, init_data_time);
	level->allocatePatchData(d_Bfu_y_p_id, init_data_time);
	level->allocatePatchData(d_Bfu_z_p_id, init_data_time);
	level->allocatePatchData(d_Df_p_id, init_data_time);
	level->allocatePatchData(d_DYf_p_id, init_data_time);
	level->allocatePatchData(d_tauf_p_id, init_data_time);
	level->allocatePatchData(d_phif_p_id, init_data_time);
	level->allocatePatchData(d_gtd_xx_p_id, init_data_time);
	level->allocatePatchData(d_gtd_xy_p_id, init_data_time);
	level->allocatePatchData(d_gtd_xz_p_id, init_data_time);
	level->allocatePatchData(d_gtd_yy_p_id, init_data_time);
	level->allocatePatchData(d_gtd_yz_p_id, init_data_time);
	level->allocatePatchData(d_gtd_zz_p_id, init_data_time);
	level->allocatePatchData(d_Atd_xx_p_id, init_data_time);
	level->allocatePatchData(d_Atd_xy_p_id, init_data_time);
	level->allocatePatchData(d_Atd_xz_p_id, init_data_time);
	level->allocatePatchData(d_Atd_yy_p_id, init_data_time);
	level->allocatePatchData(d_Atd_yz_p_id, init_data_time);
	level->allocatePatchData(d_Atd_zz_p_id, init_data_time);
	level->allocatePatchData(d_Gamh_x_p_id, init_data_time);
	level->allocatePatchData(d_Gamh_y_p_id, init_data_time);
	level->allocatePatchData(d_Gamh_z_p_id, init_data_time);
	level->allocatePatchData(d_Betau_x_p_id, init_data_time);
	level->allocatePatchData(d_Betau_y_p_id, init_data_time);
	level->allocatePatchData(d_Betau_z_p_id, init_data_time);
	level->allocatePatchData(d_Alpha_p_id, init_data_time);
	level->allocatePatchData(d_chi_p_id, init_data_time);
	level->allocatePatchData(d_trK_p_id, init_data_time);
	level->allocatePatchData(d_theta_p_id, init_data_time);
	level->allocatePatchData(d_mask_id, init_data_time);


	//Mapping the current data for new level.
	if (initial_time || level_number == 0) {
		mapDataOnPatch(init_data_time, initial_time, level_number, level);
	}

	//Fill a finer level with the data of the next coarse level.
	if ((level_number > 0) || old_level) {
		d_mapping_fill->createSchedule(level,old_level,level_number-1,hierarchy,this)->fillData(init_data_time, false);
		correctFOVS(level);
	}

	//Fill a finer level with the data of the next coarse level.
	if (!initial_time && ((level_number > 0) || old_level)) {
		d_fill_new_level->createSchedule(level,old_level,level_number-1,hierarchy,this)->fillData(init_data_time, false);
		postNewLevel(level);
		d_fill_new_level_aux->createSchedule(level,this)->fillData(init_data_time, false);
		d_fill_new_level_der_aux->createSchedule(level,old_level,level_number-1,hierarchy,this)->fillData(init_data_time, false);
	}

	//Interphase mapping
	if (initial_time || level_number == 0) {
		interphaseMapping(init_data_time, initial_time, level_number, level, 1);
	}


	//Initialize current data for new level.
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr< hier::Patch > patch = *p_it;
		if (initial_time) {
  		    initializeDataOnPatch(*patch, init_data_time, initial_time);
		}

	}
	//Post-initialization Sync.
    	if (initial_time || level_number == 0) {

		//First synchronization from initialization
		d_bdry_fill_init->createSchedule(level,this)->fillData(init_data_time, false);
		double current_time = init_data_time;
		const double level_ratio = level->getRatioToCoarserLevel().max();
		double simPlat_dt = 0;
		for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
			const std::shared_ptr< hier::Patch >& patch = *p_it;
			double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
			double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
			double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
			double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
			double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
			double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
			double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
			double* Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_id).get())->getPointer();
			double* Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_id).get())->getPointer();
			double* Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_id).get())->getPointer();
			double* Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_id).get())->getPointer();
			double* Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_id).get())->getPointer();
			double* Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_id).get())->getPointer();
			double* Df = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_id).get())->getPointer();
			double* DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_DYf_id).get())->getPointer();
			double* tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_tauf_id).get())->getPointer();
			double* phif = ((pdat::NodeData<double> *) patch->getPatchData(d_phif_id).get())->getPointer();
			double* gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_id).get())->getPointer();
			double* gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_id).get())->getPointer();
			double* gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_id).get())->getPointer();
			double* gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_id).get())->getPointer();
			double* gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_id).get())->getPointer();
			double* gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_id).get())->getPointer();
			double* Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xx_id).get())->getPointer();
			double* Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xy_id).get())->getPointer();
			double* Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xz_id).get())->getPointer();
			double* Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yy_id).get())->getPointer();
			double* Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yz_id).get())->getPointer();
			double* Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_zz_id).get())->getPointer();
			double* Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_x_id).get())->getPointer();
			double* Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_y_id).get())->getPointer();
			double* Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_z_id).get())->getPointer();
			double* Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_id).get())->getPointer();
			double* Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_id).get())->getPointer();
			double* Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_id).get())->getPointer();
			double* Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_id).get())->getPointer();
			double* chi = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_id).get())->getPointer();
			double* trK = ((pdat::NodeData<double> *) patch->getPatchData(d_trK_id).get())->getPointer();
			double* theta = ((pdat::NodeData<double> *) patch->getPatchData(d_theta_id).get())->getPointer();

			//Hard region field distance variables
			double* d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
			double* d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
			double* d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();

			//Get the dimensions of the patch
			const hier::Index boxfirst = patch->getBox().lower();
			const hier::Index boxlast = patch->getBox().upper();

			//Get delta spaces into an array. dx, dy, dz.
			const std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
			const double* dx  = patch_geom->getDx();
			int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
			int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
			int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = klast/2; k < klast; k++) {
			for(int j = jlast/2; j < jlast; j++) {
				for(int i = ilast/2; i < ilast; i++) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
							vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
							vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
							vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
							vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
							vector(Df, i, j, k) = vector(Df, i - 1, j, k);
							vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
							vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
							vector(phif, i, j, k) = vector(phif, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
							vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
							vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
							vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
							vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
							vector(Df, i, j, k) = vector(Df, i, j - 1, k);
							vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
							vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
							vector(phif, i, j, k) = vector(phif, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
							vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
							vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
							vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
							vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
							vector(Df, i, j, k) = vector(Df, i, j, k - 1);
							vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
							vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
							vector(phif, i, j, k) = vector(phif, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2 - 1; k >= 0; k--) {
			for(int j = jlast/2; j < jlast; j++) {
				for(int i = ilast/2; i < ilast; i++) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
							vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
							vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
							vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
							vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
							vector(Df, i, j, k) = vector(Df, i - 1, j, k);
							vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
							vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
							vector(phif, i, j, k) = vector(phif, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
							vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
							vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
							vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
							vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
							vector(Df, i, j, k) = vector(Df, i, j - 1, k);
							vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
							vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
							vector(phif, i, j, k) = vector(phif, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
							vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
							vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
							vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
							vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
							vector(Df, i, j, k) = vector(Df, i, j, k - 1);
							vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
							vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
							vector(phif, i, j, k) = vector(phif, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2 - 1; k >= 0; k--) {
			for(int j = jlast/2 - 1; j >= 0; j--) {
				for(int i = ilast/2; i < ilast; i++) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
							vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
							vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
							vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
							vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
							vector(Df, i, j, k) = vector(Df, i - 1, j, k);
							vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
							vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
							vector(phif, i, j, k) = vector(phif, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
							vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
							vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
							vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
							vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
							vector(Df, i, j, k) = vector(Df, i, j - 1, k);
							vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
							vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
							vector(phif, i, j, k) = vector(phif, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
							vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
							vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
							vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
							vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
							vector(Df, i, j, k) = vector(Df, i, j, k - 1);
							vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
							vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
							vector(phif, i, j, k) = vector(phif, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2; k < klast; k++) {
			for(int j = jlast/2 - 1; j >= 0; j--) {
				for(int i = ilast/2; i < ilast; i++) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
							vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
							vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
							vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
							vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
							vector(Df, i, j, k) = vector(Df, i - 1, j, k);
							vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
							vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
							vector(phif, i, j, k) = vector(phif, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
							vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
							vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
							vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
							vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
							vector(Df, i, j, k) = vector(Df, i, j - 1, k);
							vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
							vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
							vector(phif, i, j, k) = vector(phif, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
							vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
							vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
							vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
							vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
							vector(Df, i, j, k) = vector(Df, i, j, k - 1);
							vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
							vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
							vector(phif, i, j, k) = vector(phif, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2; k < klast; k++) {
			for(int j = jlast/2; j < jlast; j++) {
				for(int i = ilast/2 - 1; i >= 0; i--) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
							vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
							vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
							vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
							vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
							vector(Df, i, j, k) = vector(Df, i - 1, j, k);
							vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
							vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
							vector(phif, i, j, k) = vector(phif, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
							vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
							vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
							vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
							vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
							vector(Df, i, j, k) = vector(Df, i, j - 1, k);
							vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
							vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
							vector(phif, i, j, k) = vector(phif, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
							vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
							vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
							vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
							vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
							vector(Df, i, j, k) = vector(Df, i, j, k - 1);
							vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
							vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
							vector(phif, i, j, k) = vector(phif, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2 - 1; k >= 0; k--) {
			for(int j = jlast/2; j < jlast; j++) {
				for(int i = ilast/2 - 1; i >= 0; i--) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
							vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
							vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
							vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
							vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
							vector(Df, i, j, k) = vector(Df, i - 1, j, k);
							vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
							vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
							vector(phif, i, j, k) = vector(phif, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
							vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
							vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
							vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
							vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
							vector(Df, i, j, k) = vector(Df, i, j - 1, k);
							vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
							vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
							vector(phif, i, j, k) = vector(phif, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
							vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
							vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
							vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
							vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
							vector(Df, i, j, k) = vector(Df, i, j, k - 1);
							vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
							vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
							vector(phif, i, j, k) = vector(phif, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2 - 1; k >= 0; k--) {
			for(int j = jlast/2 - 1; j >= 0; j--) {
				for(int i = ilast/2 - 1; i >= 0; i--) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
							vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
							vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
							vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
							vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
							vector(Df, i, j, k) = vector(Df, i - 1, j, k);
							vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
							vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
							vector(phif, i, j, k) = vector(phif, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
							vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
							vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
							vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
							vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
							vector(Df, i, j, k) = vector(Df, i, j - 1, k);
							vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
							vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
							vector(phif, i, j, k) = vector(phif, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
							vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
							vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
							vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
							vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
							vector(Df, i, j, k) = vector(Df, i, j, k - 1);
							vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
							vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
							vector(phif, i, j, k) = vector(phif, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2; k < klast; k++) {
			for(int j = jlast/2 - 1; j >= 0; j--) {
				for(int i = ilast/2 - 1; i >= 0; i--) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
							vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
							vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
							vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
							vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
							vector(Df, i, j, k) = vector(Df, i - 1, j, k);
							vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
							vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
							vector(phif, i, j, k) = vector(phif, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
							vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
							vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
							vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
							vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
							vector(Df, i, j, k) = vector(Df, i, j - 1, k);
							vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
							vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
							vector(phif, i, j, k) = vector(phif, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
							vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
							vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
							vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
							vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
							vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
							vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
							vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
							vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
							vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
							vector(Df, i, j, k) = vector(Df, i, j, k - 1);
							vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
							vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
							vector(phif, i, j, k) = vector(phif, i, j, k - 1);
						}
					}
				}
			}
		}
		for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
			const std::shared_ptr< hier::Patch >& patch = *p_it;
			double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
			double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
			double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
			double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
			double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
			double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
			double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
			double* chi = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_id).get())->getPointer();
			double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
			double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
			double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
			double* Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_id).get())->getPointer();
			double* Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_id).get())->getPointer();
			double* Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_id).get())->getPointer();
			double* tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_tauf_id).get())->getPointer();
			double* gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_id).get())->getPointer();
			double* gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_id).get())->getPointer();
			double* gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_id).get())->getPointer();
			double* gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_id).get())->getPointer();
			double* gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_id).get())->getPointer();
			double* gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_id).get())->getPointer();
			double* Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_id).get())->getPointer();
			double* Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_id).get())->getPointer();
			double* Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_id).get())->getPointer();
			double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
			double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
			double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
			double* Df = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_id).get())->getPointer();
			double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
			double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
			double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
			double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
			double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
			double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
			double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
			double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
			double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
			double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
			double* Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_id).get())->getPointer();
			double* DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_DYf_id).get())->getPointer();
			double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
			double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
			double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
			double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
			double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
			double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
			double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
			double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
			double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
			double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
			double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
			double chi_max, h, sdetg, Bu_x, Bu_y, Bu_z, tau, inv_chi, Bfd_y, Bfd_z, Bfd_x, decay_factor, detgtd, idetgtd, gtu_zz, gtu_xx, gtu_xz, gtu_xy, Sfu_x, Su_x, vfu_x, Sd_x, Sd_y, gtu_yy, Sd_z, gtu_yz, vfu_z, Sfu_y, Su_y, Sfu_z, Su_z, vfu_y, sqvf, sqW, W, Bd_x, D, Bd_y, Bd_z, rhof_tmp, Yef_tmp, Tf_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, dpfdye_tmp, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tauf_tmp, Df_tmp, DYf_tmp, Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, Efu_x_tmp, Efu_y_tmp, Efu_z_tmp, vacuum_D_tmp, vacuum_tau_tmp, vacuum_rho_reset_tmp;

			//Get the dimensions of the patch
			const hier::Index boxfirst = patch->getBox().lower();
			const hier::Index boxlast = patch->getBox().upper();

			//Get delta spaces into an array. dx, dy, dz.
			const std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
			const double* dx  = patch_geom->getDx();
			int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
			int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
			int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			for (int k = 0; k < klast; k++) {
				for (int j = 0; j < jlast; j++) {
					for (int i = 0; i < ilast; i++) {
						chi_max = MAX(chi_floor, vector(chi, i, j, k));
						h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						sdetg = pow(chi_max, (-1.500000000000000));
						Bu_x = vector(Bfu_x, i, j, k) / sdetg;
						Bu_y = vector(Bfu_y, i, j, k) / sdetg;
						Bu_z = vector(Bfu_z, i, j, k) / sdetg;
						tau = vector(tauf, i, j, k) / sdetg;
						inv_chi = 1.0 / fabs(chi_max);
						Bfd_y = inv_chi * (vector(gtd_xy, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_yz, i, j, k) * vector(Bfu_z, i, j, k));
						Bfd_z = inv_chi * (vector(gtd_xz, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_yz, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Bfu_z, i, j, k));
						Bfd_x = inv_chi * (vector(gtd_xx, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_xy, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_xz, i, j, k) * vector(Bfu_z, i, j, k));
						decay_factor = MIN(1.0, pow((R_0 / MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))), eta_damping_exp));
						detgtd = vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) * vector(gtd_xx, i, j, k) + (-vector(gtd_yy, i, j, k) * (vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k))) + (-vector(gtd_zz, i, j, k) * (vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k))) + (-(vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k)) * vector(gtd_xx, i, j, k)) + 2.0 * vector(gtd_yz, i, j, k) * vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k);
						idetgtd = 1.0 / detgtd;
						gtu_zz = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_yy, i, j, k) - vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k));
						gtu_xx = idetgtd * (vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k));
						gtu_xz = idetgtd * (vector(gtd_xy, i, j, k) * vector(gtd_yz, i, j, k) - vector(gtd_yy, i, j, k) * vector(gtd_xz, i, j, k));
						gtu_xy = idetgtd * ((-vector(gtd_xy, i, j, k) * vector(gtd_zz, i, j, k)) + vector(gtd_yz, i, j, k) * vector(gtd_xz, i, j, k));
						Sfu_x = chi_max * (gtu_xx * vector(Sfd_x, i, j, k) + gtu_xy * vector(Sfd_y, i, j, k) + gtu_xz * vector(Sfd_z, i, j, k));
						Su_x = Sfu_x / sdetg;
						vfu_x = chi_max * (gtu_xx * vector(vfd_x, i, j, k) + gtu_xy * vector(vfd_y, i, j, k) + gtu_xz * vector(vfd_z, i, j, k));
						Sd_x = vector(Sfd_x, i, j, k) / sdetg;
						Sd_y = vector(Sfd_y, i, j, k) / sdetg;
						gtu_yy = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k));
						Sd_z = vector(Sfd_z, i, j, k) / sdetg;
						gtu_yz = idetgtd * ((-vector(gtd_xx, i, j, k) * vector(gtd_yz, i, j, k)) + vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k));
						vfu_z = chi_max * (gtu_xz * vector(vfd_x, i, j, k) + gtu_yz * vector(vfd_y, i, j, k) + gtu_zz * vector(vfd_z, i, j, k));
						Sfu_y = chi_max * (gtu_xy * vector(Sfd_x, i, j, k) + gtu_yy * vector(Sfd_y, i, j, k) + gtu_yz * vector(Sfd_z, i, j, k));
						Su_y = Sfu_y / sdetg;
						Sfu_z = chi_max * (gtu_xz * vector(Sfd_x, i, j, k) + gtu_yz * vector(Sfd_y, i, j, k) + gtu_zz * vector(Sfd_z, i, j, k));
						Su_z = Sfu_z / sdetg;
						vfu_y = chi_max * (gtu_xy * vector(vfd_x, i, j, k) + gtu_yy * vector(vfd_y, i, j, k) + gtu_yz * vector(vfd_z, i, j, k));
						sqvf = MIN(vfu_x * vector(vfd_x, i, j, k) + vfu_y * vector(vfd_y, i, j, k) + vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						sqW = 1.0 / (1.0 - sqvf);
						W = sqrt(sqW);
						Bd_x = Bfd_x / sdetg;
						D = vector(Df, i, j, k) / sdetg;
						Bd_y = Bfd_y / sdetg;
						Bd_z = Bfd_z / sdetg;
						vector(t_chie, i, j, k) = vector(chie, i, j, k);
						vector(t_chia, i, j, k) = vector(chia, i, j, k);
						vector(t_optdepthx, i, j, k) = vector(optdepthx, i, j, k);
						vector(t_optdeptha, i, j, k) = vector(optdeptha, i, j, k);
						vector(t_optdepthe, i, j, k) = vector(optdepthe, i, j, k);
						vector(Alpha, i, j, k) = fabs(vector(Alpha, i, j, k));
						rhof_tmp = vector(rhof, i, j, k);
						Yef_tmp = vector(DYf, i, j, k) / vector(Df, i, j, k);
						if (Yef_tmp < vacuum_ye) {
							vector(DYf, i, j, k) = vacuum_ye_reset * vector(Df, i, j, k);
							Yef_tmp = vacuum_ye_reset;
						}
						if (Yef_tmp > ye_maximum) {
							vector(DYf, i, j, k) = 0.9 * ye_maximum * vector(Df, i, j, k);
							Yef_tmp = 0.9 * ye_maximum;
						}
						Tf_tmp = vector(Tf, i, j, k);
						pf_tmp = vector(pf, i, j, k);
						epsf_tmp = vector(epsf, i, j, k);
						vfd_x_tmp = vector(vfd_x, i, j, k);
						vfd_y_tmp = vector(vfd_y, i, j, k);
						vfd_z_tmp = vector(vfd_z, i, j, k);
						sqcs_tmp = vector(sqcs, i, j, k);
						dpfdrho_tmp = vector(dpfdrho, i, j, k);
						dpfdeps_tmp = vector(dpfdeps, i, j, k);
						dpfdye_tmp = vector(dpfdye, i, j, k);
						Sfd_x_tmp = vector(Sfd_x, i, j, k);
						Sfd_y_tmp = vector(Sfd_y, i, j, k);
						Sfd_z_tmp = vector(Sfd_z, i, j, k);
						tauf_tmp = vector(tauf, i, j, k);
						Df_tmp = vector(Df, i, j, k);
						DYf_tmp = vector(DYf, i, j, k);
						Bfu_x_tmp = vector(Bfu_x, i, j, k);
						Bfu_y_tmp = vector(Bfu_y, i, j, k);
						Bfu_z_tmp = vector(Bfu_z, i, j, k);
						if (equalsEq(externalCon2Prim, 1.0)) {
							Efu_x_tmp = vector(Efu_x, i, j, k);
							Efu_y_tmp = vector(Efu_y, i, j, k);
							Efu_z_tmp = vector(Efu_z, i, j, k);
							externalCon2prim(Efu_x_tmp, Efu_y_tmp, Efu_z_tmp, rhof_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, Df_tmp, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tauf_tmp, W, vector(chi, i, j, k), gtu_xx, gtu_xy, gtu_xz, gtu_yy, gtu_yz, gtu_zz, vector(gtd_xx, i, j, k), vector(gtd_xy, i, j, k), vector(gtd_xz, i, j, k), vector(gtd_yy, i, j, k), vector(gtd_yz, i, j, k), vector(gtd_zz, i, j, k), Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, dx, simPlat_dt, ilast, jlast);
							vector(Efu_x, i, j, k) = Efu_x_tmp;
							vector(Efu_y, i, j, k) = Efu_y_tmp;
							vector(Efu_z, i, j, k) = Efu_z_tmp;
						} else {
							vacuum_D_tmp = vacuum_D * decay_factor;
							vacuum_tau_tmp = vacuum_tau * decay_factor;
							vacuum_rho_reset_tmp = vacuum_rho_reset * decay_factor;
							con2prim(eos_type, rhof_tmp, Yef_tmp, Tf_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, dpfdye_tmp, rho_0, rho_1, rho_2, a_0, a_1, a_2, a_3, K_0, K_1, K_2, K_3, gamma_0, gamma_1, gamma_2, gamma_3, D, vacuum_D_tmp, Df_tmp, DYf_tmp, sdetg, Sd_x, Sd_y, Sd_z, Su_x, Su_y, Su_z, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tau, vacuum_tau_tmp, tauf_tmp, W, h, vector(chi, i, j, k), gtu_xx, gtu_xy, gtu_xz, gtu_yy, gtu_yz, gtu_zz, vector(gtd_xx, i, j, k), vector(gtd_xy, i, j, k), vector(gtd_xz, i, j, k), vector(gtd_yy, i, j, k), vector(gtd_yz, i, j, k), vector(gtd_zz, i, j, k), Bd_x, Bd_y, Bd_z, Bu_x, Bu_y, Bu_z, Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, max_error, gamma, inv_chi, threshold_vacuum, threshold_sqSmax, threshold_sqBmax, energyShift, vacuum_ye_beta, vacuum_rho, vacuum_rho_reset_tmp, vacuum_P_reset, vacuum_ye_reset, vacuum_temp_reset, vacuum_tau_reset, minTableTemperature, minTableEnergy, coord0_data_Data, coord0_dx_Data, coord0_size_Data, coord1_data_Data, coord1_dx_Data, coord1_size_Data, coord2_data_Data, coord2_dx_Data, coord2_size_Data, vars_data_Data, vars_size_Data, dx, simPlat_dt, ilast, jlast);
							vector(Efu_x, i, j, k) = -((-vector(vfd_z, i, j, k) * Bfd_y) + vector(vfd_y, i, j, k) * Bfd_z) / (sdetg * sdetg);
							vector(Efu_y, i, j, k) = -(vector(vfd_z, i, j, k) * Bfd_x - vector(vfd_x, i, j, k) * Bfd_z) / (sdetg * sdetg);
							vector(Efu_z, i, j, k) = -((-vector(vfd_y, i, j, k) * Bfd_x) + vector(vfd_x, i, j, k) * Bfd_y) / (sdetg * sdetg);
						}
						vector(rhof, i, j, k) = rhof_tmp;
						vector(Yef, i, j, k) = Yef_tmp;
						vector(Tf, i, j, k) = Tf_tmp;
						vector(pf, i, j, k) = pf_tmp;
						vector(epsf, i, j, k) = epsf_tmp;
						vector(vfd_x, i, j, k) = vfd_x_tmp;
						vector(vfd_y, i, j, k) = vfd_y_tmp;
						vector(vfd_z, i, j, k) = vfd_z_tmp;
						vector(sqcs, i, j, k) = sqcs_tmp;
						vector(dpfdye, i, j, k) = dpfdye_tmp;
						vector(dpfdeps, i, j, k) = dpfdeps_tmp;
						vector(dpfdrho, i, j, k) = dpfdrho_tmp;
						vector(Sfd_x, i, j, k) = Sfd_x_tmp;
						vector(Sfd_y, i, j, k) = Sfd_y_tmp;
						vector(Sfd_z, i, j, k) = Sfd_z_tmp;
						vector(tauf, i, j, k) = tauf_tmp;
						vector(Df, i, j, k) = Df_tmp;
						vector(DYf, i, j, k) = DYf_tmp;
						vector(Bfu_x, i, j, k) = Bfu_x_tmp;
						vector(Bfu_y, i, j, k) = Bfu_y_tmp;
						vector(Bfu_z, i, j, k) = Bfu_z_tmp;
						vector(t_chix, i, j, k) = vector(chix, i, j, k);
					}
				}
			}
		}
		for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
			const std::shared_ptr< hier::Patch >& patch = *p_it;
			double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
			double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
			double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
			double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
			double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
			double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
			double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
			double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
			double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
			double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
			double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
			double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
			double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
			double* DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_DYf_id).get())->getPointer();
			double* Df = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_id).get())->getPointer();
			double* Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_id).get())->getPointer();
			double* Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_id).get())->getPointer();
			double* Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_id).get())->getPointer();
			double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
			double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
			double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
			double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
			double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
			double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
			double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
			double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
			double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
			double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
			double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
			double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
			double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
			double* chi = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_id).get())->getPointer();
			double* gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_id).get())->getPointer();
			double* gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_id).get())->getPointer();
			double* gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_id).get())->getPointer();
			double* gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_id).get())->getPointer();
			double* gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_id).get())->getPointer();
			double* gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_id).get())->getPointer();
			double* Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_id).get())->getPointer();
			double* Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_id).get())->getPointer();
			double* Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_id).get())->getPointer();
			double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
			double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
			double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
			double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
			double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
			double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
			double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
			double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
			double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
			double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
			double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
			double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
			double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
			double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
			double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
			double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
			double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
			double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
			double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
			double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
			double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
			double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
			double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
			double* qnu_e = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_e_id).get())->getPointer();
			double* qnu_a = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_a_id).get())->getPointer();
			double* qnu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_x_id).get())->getPointer();
			double* Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_id).get())->getPointer();
			double* Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_id).get())->getPointer();
			double* Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_id).get())->getPointer();
			double* Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_id).get())->getPointer();
			double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
			double d_optdeptheh_o0_t0_m0_l0, d_optdepthev_o0_t0_m0_l0, d_optdepthed_o0_t0_m0_l0, d_optdepthah_o0_t0_m0_l0, d_optdepthav_o0_t0_m0_l0, d_optdepthad_o0_t0_m0_l0, d_optdepthxh_o0_t0_m0_l0, d_optdepthxv_o0_t0_m0_l0, d_optdepthxd_o0_t0_m0_l0, d_chieh_o0_t0_m0_l0, d_chiev_o0_t0_m0_l0, d_chied_o0_t0_m0_l0, d_chiah_o0_t0_m0_l0, d_chiav_o0_t0_m0_l0, d_chiad_o0_t0_m0_l0, d_chixh_o0_t0_m0_l0, d_chixv_o0_t0_m0_l0, d_chixd_o0_t0_m0_l0, d_CovdDYf_z_o0_t0_m0_l0, d_CovdDYf_y_o0_t0_m0_l0, d_CovdDYf_x_o0_t0_m0_l0, d_CovdDf_z_o0_t0_m0_l0, d_CovdDf_y_o0_t0_m0_l0, d_CovdDf_x_o0_t0_m0_l0, d_CovdBfu_xx_o0_t0_m0_l0, d_CovdBfu_xy_o0_t0_m0_l0, d_CovdBfu_xz_o0_t0_m0_l0, d_CovdBfu_yx_o0_t0_m0_l0, d_CovdBfu_yy_o0_t0_m0_l0, d_CovdBfu_yz_o0_t0_m0_l0, d_CovdBfu_zx_o0_t0_m0_l0, d_CovdBfu_zy_o0_t0_m0_l0, d_CovdBfu_zz_o0_t0_m0_l0, d_CovdEfu_xx_o0_t0_m0_l0, d_CovdEfu_xy_o0_t0_m0_l0, d_CovdEfu_xz_o0_t0_m0_l0, d_CovdEfu_yx_o0_t0_m0_l0, d_CovdEfu_yy_o0_t0_m0_l0, d_CovdEfu_yz_o0_t0_m0_l0, d_CovdEfu_zx_o0_t0_m0_l0, d_CovdEfu_zy_o0_t0_m0_l0, d_CovdEfu_zz_o0_t0_m0_l0, d_Covdepsf_x_o0_t0_m0_l0, d_Covdepsf_y_o0_t0_m0_l0, d_Covdepsf_z_o0_t0_m0_l0, d_Covddpfdeps_x_o0_t0_m0_l0, d_Covddpfdeps_y_o0_t0_m0_l0, d_Covddpfdeps_z_o0_t0_m0_l0, d_Covddpfdye_x_o0_t0_m0_l0, d_Covddpfdye_y_o0_t0_m0_l0, d_Covddpfdye_z_o0_t0_m0_l0, d_Covddpfdrho_x_o0_t0_m0_l0, d_Covddpfdrho_y_o0_t0_m0_l0, d_Covddpfdrho_z_o0_t0_m0_l0, d_Covdvfd_xx_o0_t0_m0_l0, d_Covdvfd_xy_o0_t0_m0_l0, d_Covdvfd_xz_o0_t0_m0_l0, d_Covdvfd_yx_o0_t0_m0_l0, d_Covdvfd_yy_o0_t0_m0_l0, d_Covdvfd_yz_o0_t0_m0_l0, d_Covdvfd_zx_o0_t0_m0_l0, d_Covdvfd_zy_o0_t0_m0_l0, d_Covdvfd_zz_o0_t0_m0_l0, d_Covdrhof_x_o0_t0_m0_l0, d_Covdrhof_y_o0_t0_m0_l0, d_Covdrhof_z_o0_t0_m0_l0, d_Covdpf_x_o0_t0_m0_l0, d_Covdpf_y_o0_t0_m0_l0, d_Covdpf_z_o0_t0_m0_l0, d_CovdYef_x_o0_t0_m0_l0, d_CovdYef_y_o0_t0_m0_l0, d_CovdYef_z_o0_t0_m0_l0, chiev, chieh, Covdvfd_xx, Covdvfd_xz, Covdvfd_xy, chied, Covddpfdeps_x, Covddpfdeps_y, Covddpfdeps_z, Covdvfd_zx, Covdvfd_zz, Covdvfd_zy, Covdvfd_yy, Covdvfd_yx, Covdvfd_yz, Covdpf_x, Covdpf_y, Covdpf_z, chi_max, inv_chi, Bfd_y, Bfd_z, Bfd_x, sqBf, chixh, chixd, Covddpfdrho_z, Covddpfdrho_y, Covddpfdrho_x, sdetg, m_CovdDf_x_o0_t0_l0, CovdDf_x, m_CovdDf_y_o0_t0_l0, CovdDf_y, m_CovdDf_z_o0_t0_l0, CovdDf_z, m_CovdBfu_zz_o0_t0_l0, CovdBfu_zz, m_CovdBfu_zy_o0_t0_l0, CovdBfu_zy, m_CovdBfu_zx_o0_t0_l0, CovdBfu_zx, m_CovdBfu_yz_o0_t0_l0, CovdBfu_yz, m_CovdBfu_yy_o0_t0_l0, CovdBfu_yy, m_CovdBfu_yx_o0_t0_l0, CovdBfu_yx, m_CovdBfu_xy_o0_t0_l0, CovdBfu_xy, m_CovdBfu_xx_o0_t0_l0, CovdBfu_xx, m_CovdBfu_xz_o0_t0_l0, CovdBfu_xz, CovdBfd_zx, Covdrhof_z, Covdrhof_y, Covdrhof_x, CovdBfd_zz, CovdBfd_zy, m_CovdDYf_x_o0_t0_l0, CovdDYf_x, m_CovdDYf_y_o0_t0_l0, CovdDYf_y, m_CovdDYf_z_o0_t0_l0, CovdDYf_z, Efd_z, Efd_y, chixv, Efd_x, sqEf, optdepthed, CovdBfd_yy, CovdBfd_yx, CovdBfd_yz, optdeptheh, CovdBvf_z, optdepthev, CovdBvf_y, CovdBvf_x, CovdBfd_xx, CovdBfd_xz, CovdBfd_xy, optdepthxd, optdepthxh, optdepthxv, chiav, CovdEfu_zz, CovdEfu_zy, CovdEfu_zx, CovdEfd_zz, CovdEfd_zx, CovdEfd_zy, Bvf, optdepthah, optdepthad, h, chiah, chiad, CovdYef_y, optdepthav, CovdYef_z, CovdYef_x, detgtd, idetgtd, gtu_zz, gtu_yy, gtu_yz, gtu_xx, gtu_xz, CovuBfu_zx, CovuBfu_zz, CovuBfu_zy, Covdvfu_xz, CovuYef_z, Covdvfu_zz, Covdvfu_yz, Covuvfu_zz, Sfu_z, Covurhof_z, CovuBvf_z, gtu_xy, CovuBfu_yz, CovuBfu_yy, CovuBfu_yx, Covdvfu_xy, Covdvfu_xx, CovdinvsqW_x, CovuYef_x, CovuYef_y, Covdvfu_zy, Covdvfu_zx, CovdinvsqW_z, Covdvfu_yx, Covuvfu_zx, Covdvfu_yy, CovdinvsqW_y, CovuinvsqW_z, Covuvfu_zy, CovuBfu_xz, CovuBfu_xy, CovuBfu_xx, Sfu_y, Sfu_x, Covurhof_x, Covurhof_y, Covuvfu_yz, Covuvfu_yy, Covuvfu_yx, Covuvfu_xz, Covuvfu_xy, Covuvfu_xx, vfu_y, vfu_x, CovuBvf_x, CovuBvf_y, CovuinvsqW_y, CovuinvsqW_x, Covddpfdye_z, Covddpfdye_y, Covddpfdye_x, CovdEfu_xz, CovdEfu_xy, CovdEfu_xx, CovdEfd_xx, CovdEfd_xy, CovdEfd_xz, vfu_z, sqvf, sqW, PhiAh, Epf, Thetaf, Covdepsf_z, Covdhf_z, CovdEpf_z, CovdThetaf_z, Covdepsf_y, Covdhf_y, CovdEpf_y, CovdThetaf_y, Covdepsf_x, Covuepsf_z, Covuepsf_y, Covdhf_x, CovdEpf_x, CovdThetaf_x, CovuThetaf_z, CovuThetaf_x, CovuThetaf_y, Phivh_z, Phivh_x, Phivh_y, PhiMh_yz, PhiMh_xz, PhiMh_xy, Covuepsf_x, faceta, CovdEfu_yz, CovuEfu_zz, CovuEfu_xz, CovdEfu_yy, CovuEfu_zy, CovuEfu_xy, CovdEfu_yx, CovdEfd_yy, CovdEfd_yz, CovdEfd_yx, CovuEfu_zx, CovuEfu_xx, CovuEfu_yx, CovuEfu_yz, CovuEfu_yy, PhiThetah, HPres, HTheta, Hv_z, HN_z, HNe_z, Hv_y, HN_y, HM_yz, HNe_y, Hv_x, HN_x, HM_xz, HM_xy, HE_z, HE_x, HE_y, HT_zz, HT_yy, HT_yz, HT_xx, HT_xz, HT_xy, HNe_x, d_TauN_x_o0_t0_m0_l0, d_TauN_y_o0_t0_m0_l0, d_TauN_z_o0_t0_m0_l0, d_TauNe_x_o0_t0_m0_l0, d_TauNe_y_o0_t0_m0_l0, d_TauNe_z_o0_t0_m0_l0, d_TauM_xy_o0_t0_m0_l0, d_TauM_xz_o0_t0_m0_l0, d_TauM_yz_o0_t0_m0_l0, d_TauT_xx_o0_t0_m0_l0, d_TauT_xy_o0_t0_m0_l0, d_TauT_xz_o0_t0_m0_l0, d_TauT_yy_o0_t0_m0_l0, d_TauT_yz_o0_t0_m0_l0, d_TauT_zz_o0_t0_m0_l0, optdepthe_tmp, optdeptha_tmp, optdepthx_tmp, chie_tmp, chia_tmp, chix_tmp, qnu_tmp, rnu_tmp, qnu_e_tmp, qnu_a_tmp, qnu_x_tmp, betabeta, betav, velr, f_redshift;

			//Get the dimensions of the patch
			const hier::Index boxfirst = patch->getBox().lower();
			const hier::Index boxlast = patch->getBox().upper();

			//Get delta spaces into an array. dx, dy, dz.
			const std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
			const double* dx  = patch_geom->getDx();
			int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
			int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
			int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			for (int k = 0; k < klast; k++) {
				for (int j = 0; j < jlast; j++) {
					for (int i = 0; i < ilast; i++) {
						if ((i + 2 < ilast && i - 2 >= 0 && j + 2 < jlast && j - 2 >= 0 && k + 2 < klast && k - 2 >= 0) && ((i + 3 < ilast || !patch->getPatchGeometry()->getTouchesRegularBoundary(0, 1)) && (i - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(0, 0)) && (j + 3 < jlast || !patch->getPatchGeometry()->getTouchesRegularBoundary(1, 1)) && (j - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(1, 0)) && (k + 3 < klast || !patch->getPatchGeometry()->getTouchesRegularBoundary(2, 1)) && (k - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(2, 0)))) {
							d_optdeptheh_o0_t0_m0_l0 = eikonal_i(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_optdepthev_o0_t0_m0_l0 = eikonal_j(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_optdepthed_o0_t0_m0_l0 = eikonal_k(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_optdepthah_o0_t0_m0_l0 = eikonal_i(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_optdepthav_o0_t0_m0_l0 = eikonal_j(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_optdepthad_o0_t0_m0_l0 = eikonal_k(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_optdepthxh_o0_t0_m0_l0 = eikonal_i(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_optdepthxv_o0_t0_m0_l0 = eikonal_j(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_optdepthxd_o0_t0_m0_l0 = eikonal_k(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_chieh_o0_t0_m0_l0 = eikonal_i(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_chiev_o0_t0_m0_l0 = eikonal_j(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_chied_o0_t0_m0_l0 = eikonal_k(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_chiah_o0_t0_m0_l0 = eikonal_i(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_chiav_o0_t0_m0_l0 = eikonal_j(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_chiad_o0_t0_m0_l0 = eikonal_k(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_chixh_o0_t0_m0_l0 = eikonal_i(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_chixv_o0_t0_m0_l0 = eikonal_j(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_chixd_o0_t0_m0_l0 = eikonal_k(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
							if (calculate_les_terms) {
								d_CovdDYf_z_o0_t0_m0_l0 = D1CDO4_k(DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdDYf_y_o0_t0_m0_l0 = D1CDO4_j(DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdDYf_x_o0_t0_m0_l0 = D1CDO4_i(DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdDf_z_o0_t0_m0_l0 = D1CDO4_k(Df, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdDf_y_o0_t0_m0_l0 = D1CDO4_j(Df, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdDf_x_o0_t0_m0_l0 = D1CDO4_i(Df, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdBfu_xx_o0_t0_m0_l0 = D1CDO4_i(Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdBfu_xy_o0_t0_m0_l0 = D1CDO4_i(Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdBfu_xz_o0_t0_m0_l0 = D1CDO4_i(Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdBfu_yx_o0_t0_m0_l0 = D1CDO4_j(Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdBfu_yy_o0_t0_m0_l0 = D1CDO4_j(Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdBfu_yz_o0_t0_m0_l0 = D1CDO4_j(Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdBfu_zx_o0_t0_m0_l0 = D1CDO4_k(Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdBfu_zy_o0_t0_m0_l0 = D1CDO4_k(Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdBfu_zz_o0_t0_m0_l0 = D1CDO4_k(Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdEfu_xx_o0_t0_m0_l0 = D1CDO4_i(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdEfu_xy_o0_t0_m0_l0 = D1CDO4_i(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdEfu_xz_o0_t0_m0_l0 = D1CDO4_i(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdEfu_yx_o0_t0_m0_l0 = D1CDO4_j(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdEfu_yy_o0_t0_m0_l0 = D1CDO4_j(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdEfu_yz_o0_t0_m0_l0 = D1CDO4_j(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdEfu_zx_o0_t0_m0_l0 = D1CDO4_k(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdEfu_zy_o0_t0_m0_l0 = D1CDO4_k(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdEfu_zz_o0_t0_m0_l0 = D1CDO4_k(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdepsf_x_o0_t0_m0_l0 = D1CDO4_i(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdepsf_y_o0_t0_m0_l0 = D1CDO4_j(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdepsf_z_o0_t0_m0_l0 = D1CDO4_k(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covddpfdeps_x_o0_t0_m0_l0 = D1CDO4_i(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covddpfdeps_y_o0_t0_m0_l0 = D1CDO4_j(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covddpfdeps_z_o0_t0_m0_l0 = D1CDO4_k(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covddpfdye_x_o0_t0_m0_l0 = D1CDO4_i(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covddpfdye_y_o0_t0_m0_l0 = D1CDO4_j(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covddpfdye_z_o0_t0_m0_l0 = D1CDO4_k(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covddpfdrho_x_o0_t0_m0_l0 = D1CDO4_i(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covddpfdrho_y_o0_t0_m0_l0 = D1CDO4_j(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covddpfdrho_z_o0_t0_m0_l0 = D1CDO4_k(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdvfd_xx_o0_t0_m0_l0 = D1CDO4_i(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdvfd_xy_o0_t0_m0_l0 = D1CDO4_i(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdvfd_xz_o0_t0_m0_l0 = D1CDO4_i(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdvfd_yx_o0_t0_m0_l0 = D1CDO4_j(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdvfd_yy_o0_t0_m0_l0 = D1CDO4_j(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdvfd_yz_o0_t0_m0_l0 = D1CDO4_j(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdvfd_zx_o0_t0_m0_l0 = D1CDO4_k(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdvfd_zy_o0_t0_m0_l0 = D1CDO4_k(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdvfd_zz_o0_t0_m0_l0 = D1CDO4_k(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdrhof_x_o0_t0_m0_l0 = D1CDO4_i(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdrhof_y_o0_t0_m0_l0 = D1CDO4_j(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdrhof_z_o0_t0_m0_l0 = D1CDO4_k(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdpf_x_o0_t0_m0_l0 = D1CDO4_i(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdpf_y_o0_t0_m0_l0 = D1CDO4_j(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Covdpf_z_o0_t0_m0_l0 = D1CDO4_k(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdYef_x_o0_t0_m0_l0 = D1CDO4_i(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdYef_y_o0_t0_m0_l0 = D1CDO4_j(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_CovdYef_z_o0_t0_m0_l0 = D1CDO4_k(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
							}
							chiev = d_chiev_o0_t0_m0_l0;
							chieh = d_chieh_o0_t0_m0_l0;
							if (calculate_les_terms) {
								Covdvfd_xx = d_Covdvfd_xx_o0_t0_m0_l0;
								Covdvfd_xz = d_Covdvfd_xz_o0_t0_m0_l0;
								Covdvfd_xy = d_Covdvfd_xy_o0_t0_m0_l0;
							} else {
								Covdvfd_xx = 0.0;
								Covdvfd_xz = 0.0;
								Covdvfd_xy = 0.0;
							}
							chied = d_chied_o0_t0_m0_l0;
							if (calculate_les_terms) {
								Covddpfdeps_x = d_Covddpfdeps_x_o0_t0_m0_l0;
								Covddpfdeps_y = d_Covddpfdeps_y_o0_t0_m0_l0;
								Covddpfdeps_z = d_Covddpfdeps_z_o0_t0_m0_l0;
								Covdvfd_zx = d_Covdvfd_zx_o0_t0_m0_l0;
								Covdvfd_zz = d_Covdvfd_zz_o0_t0_m0_l0;
								Covdvfd_zy = d_Covdvfd_zy_o0_t0_m0_l0;
								Covdvfd_yy = d_Covdvfd_yy_o0_t0_m0_l0;
								Covdvfd_yx = d_Covdvfd_yx_o0_t0_m0_l0;
								Covdvfd_yz = d_Covdvfd_yz_o0_t0_m0_l0;
								Covdpf_x = d_Covdpf_x_o0_t0_m0_l0;
								Covdpf_y = d_Covdpf_y_o0_t0_m0_l0;
								Covdpf_z = d_Covdpf_z_o0_t0_m0_l0;
							} else {
								Covddpfdeps_x = 0.0;
								Covddpfdeps_y = 0.0;
								Covddpfdeps_z = 0.0;
								Covdvfd_zx = 0.0;
								Covdvfd_zz = 0.0;
								Covdvfd_zy = 0.0;
								Covdvfd_yy = 0.0;
								Covdvfd_yx = 0.0;
								Covdvfd_yz = 0.0;
								Covdpf_x = 0.0;
								Covdpf_y = 0.0;
								Covdpf_z = 0.0;
							}
							chi_max = MAX(chi_floor, vector(chi, i, j, k));
							inv_chi = 1.0 / fabs(chi_max);
							Bfd_y = inv_chi * (vector(gtd_xy, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_yz, i, j, k) * vector(Bfu_z, i, j, k));
							Bfd_z = inv_chi * (vector(gtd_xz, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_yz, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Bfu_z, i, j, k));
							Bfd_x = inv_chi * (vector(gtd_xx, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_xy, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_xz, i, j, k) * vector(Bfu_z, i, j, k));
							sqBf = vector(Bfu_x, i, j, k) * Bfd_x + vector(Bfu_y, i, j, k) * Bfd_y + vector(Bfu_z, i, j, k) * Bfd_z;
							chixh = d_chixh_o0_t0_m0_l0;
							chixd = d_chixd_o0_t0_m0_l0;
							if (calculate_les_terms) {
								Covddpfdrho_z = d_Covddpfdrho_z_o0_t0_m0_l0;
								Covddpfdrho_y = d_Covddpfdrho_y_o0_t0_m0_l0;
								Covddpfdrho_x = d_Covddpfdrho_x_o0_t0_m0_l0;
							} else {
								Covddpfdrho_z = 0.0;
								Covddpfdrho_y = 0.0;
								Covddpfdrho_x = 0.0;
							}
							sdetg = pow(chi_max, (-1.500000000000000));
							if (calculate_les_terms) {
								m_CovdDf_x_o0_t0_l0 = 1.0 / sdetg * d_CovdDf_x_o0_t0_m0_l0;
								CovdDf_x = m_CovdDf_x_o0_t0_l0;
								m_CovdDf_y_o0_t0_l0 = 1.0 / sdetg * d_CovdDf_y_o0_t0_m0_l0;
								CovdDf_y = m_CovdDf_y_o0_t0_l0;
								m_CovdDf_z_o0_t0_l0 = 1.0 / sdetg * d_CovdDf_z_o0_t0_m0_l0;
								CovdDf_z = m_CovdDf_z_o0_t0_l0;
								m_CovdBfu_zz_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_zz_o0_t0_m0_l0;
								CovdBfu_zz = m_CovdBfu_zz_o0_t0_l0;
								m_CovdBfu_zy_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_zy_o0_t0_m0_l0;
								CovdBfu_zy = m_CovdBfu_zy_o0_t0_l0;
								m_CovdBfu_zx_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_zx_o0_t0_m0_l0;
								CovdBfu_zx = m_CovdBfu_zx_o0_t0_l0;
								m_CovdBfu_yz_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_yz_o0_t0_m0_l0;
								CovdBfu_yz = m_CovdBfu_yz_o0_t0_l0;
								m_CovdBfu_yy_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_yy_o0_t0_m0_l0;
								CovdBfu_yy = m_CovdBfu_yy_o0_t0_l0;
								m_CovdBfu_yx_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_yx_o0_t0_m0_l0;
								CovdBfu_yx = m_CovdBfu_yx_o0_t0_l0;
								m_CovdBfu_xy_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_xy_o0_t0_m0_l0;
								CovdBfu_xy = m_CovdBfu_xy_o0_t0_l0;
								m_CovdBfu_xx_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_xx_o0_t0_m0_l0;
								CovdBfu_xx = m_CovdBfu_xx_o0_t0_l0;
								m_CovdBfu_xz_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_xz_o0_t0_m0_l0;
								CovdBfu_xz = m_CovdBfu_xz_o0_t0_l0;
								CovdBfd_zx = inv_chi * (vector(gtd_xx, i, j, k) * CovdBfu_zx + vector(gtd_xy, i, j, k) * CovdBfu_zy + vector(gtd_xz, i, j, k) * CovdBfu_zz);
								Covdrhof_z = d_Covdrhof_z_o0_t0_m0_l0;
								Covdrhof_y = d_Covdrhof_y_o0_t0_m0_l0;
								Covdrhof_x = d_Covdrhof_x_o0_t0_m0_l0;
								CovdBfd_zz = inv_chi * (vector(gtd_xz, i, j, k) * CovdBfu_zx + vector(gtd_yz, i, j, k) * CovdBfu_zy + vector(gtd_zz, i, j, k) * CovdBfu_zz);
								CovdBfd_zy = inv_chi * (vector(gtd_xy, i, j, k) * CovdBfu_zx + vector(gtd_yy, i, j, k) * CovdBfu_zy + vector(gtd_yz, i, j, k) * CovdBfu_zz);
								m_CovdDYf_x_o0_t0_l0 = 1.0 / sdetg * d_CovdDYf_x_o0_t0_m0_l0;
								CovdDYf_x = m_CovdDYf_x_o0_t0_l0;
								m_CovdDYf_y_o0_t0_l0 = 1.0 / sdetg * d_CovdDYf_y_o0_t0_m0_l0;
								CovdDYf_y = m_CovdDYf_y_o0_t0_l0;
								m_CovdDYf_z_o0_t0_l0 = 1.0 / sdetg * d_CovdDYf_z_o0_t0_m0_l0;
								CovdDYf_z = m_CovdDYf_z_o0_t0_l0;
							} else {
								CovdDf_x = 0.0;
								CovdDf_y = 0.0;
								CovdDf_z = 0.0;
								CovdBfu_zz = 0.0;
								CovdBfu_zy = 0.0;
								CovdBfu_zx = 0.0;
								CovdBfu_yz = 0.0;
								CovdBfu_yy = 0.0;
								CovdBfu_yx = 0.0;
								CovdBfu_xy = 0.0;
								CovdBfu_xx = 0.0;
								CovdBfu_xz = 0.0;
								CovdBfd_zx = 0.0;
								Covdrhof_z = 0.0;
								Covdrhof_y = 0.0;
								Covdrhof_x = 0.0;
								CovdBfd_zz = 0.0;
								CovdBfd_zy = 0.0;
								CovdDYf_x = 0.0;
								CovdDYf_y = 0.0;
								CovdDYf_z = 0.0;
							}
							Efd_z = inv_chi * (vector(gtd_xz, i, j, k) * vector(Efu_x, i, j, k) + vector(gtd_yz, i, j, k) * vector(Efu_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Efu_z, i, j, k));
							Efd_y = inv_chi * (vector(gtd_xy, i, j, k) * vector(Efu_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Efu_y, i, j, k) + vector(gtd_yz, i, j, k) * vector(Efu_z, i, j, k));
							chixv = d_chixv_o0_t0_m0_l0;
							Efd_x = inv_chi * (vector(gtd_xx, i, j, k) * vector(Efu_x, i, j, k) + vector(gtd_xy, i, j, k) * vector(Efu_y, i, j, k) + vector(gtd_xz, i, j, k) * vector(Efu_z, i, j, k));
							sqEf = vector(Efu_x, i, j, k) * Efd_x + vector(Efu_y, i, j, k) * Efd_y + vector(Efu_z, i, j, k) * Efd_z;
							optdepthed = d_optdepthed_o0_t0_m0_l0;
							if (calculate_les_terms) {
								CovdBfd_yy = inv_chi * (vector(gtd_xy, i, j, k) * CovdBfu_yx + vector(gtd_yy, i, j, k) * CovdBfu_yy + vector(gtd_yz, i, j, k) * CovdBfu_yz);
								CovdBfd_yx = inv_chi * (vector(gtd_xx, i, j, k) * CovdBfu_yx + vector(gtd_xy, i, j, k) * CovdBfu_yy + vector(gtd_xz, i, j, k) * CovdBfu_yz);
								CovdBfd_yz = inv_chi * (vector(gtd_xz, i, j, k) * CovdBfu_yx + vector(gtd_yz, i, j, k) * CovdBfu_yy + vector(gtd_zz, i, j, k) * CovdBfu_yz);
							} else {
								CovdBfd_yy = 0.0;
								CovdBfd_yx = 0.0;
								CovdBfd_yz = 0.0;
							}
							optdeptheh = d_optdeptheh_o0_t0_m0_l0;
							if (calculate_les_terms) {
								CovdBvf_z = vector(Bfu_x, i, j, k) * Covdvfd_zx + vector(vfd_x, i, j, k) * CovdBfu_zx + vector(Bfu_y, i, j, k) * Covdvfd_zy + vector(vfd_y, i, j, k) * CovdBfu_zy + vector(Bfu_z, i, j, k) * Covdvfd_zz + vector(vfd_z, i, j, k) * CovdBfu_zz;
							} else {
								CovdBvf_z = 0.0;
							}
							optdepthev = d_optdepthev_o0_t0_m0_l0;
							if (calculate_les_terms) {
								CovdBvf_y = vector(Bfu_x, i, j, k) * Covdvfd_yx + vector(vfd_x, i, j, k) * CovdBfu_yx + vector(Bfu_y, i, j, k) * Covdvfd_yy + vector(vfd_y, i, j, k) * CovdBfu_yy + vector(Bfu_z, i, j, k) * Covdvfd_yz + vector(vfd_z, i, j, k) * CovdBfu_yz;
								CovdBvf_x = vector(Bfu_x, i, j, k) * Covdvfd_xx + vector(vfd_x, i, j, k) * CovdBfu_xx + vector(Bfu_y, i, j, k) * Covdvfd_xy + vector(vfd_y, i, j, k) * CovdBfu_xy + vector(Bfu_z, i, j, k) * Covdvfd_xz + vector(vfd_z, i, j, k) * CovdBfu_xz;
								CovdBfd_xx = inv_chi * (vector(gtd_xx, i, j, k) * CovdBfu_xx + vector(gtd_xy, i, j, k) * CovdBfu_xy + vector(gtd_xz, i, j, k) * CovdBfu_xz);
								CovdBfd_xz = inv_chi * (vector(gtd_xz, i, j, k) * CovdBfu_xx + vector(gtd_yz, i, j, k) * CovdBfu_xy + vector(gtd_zz, i, j, k) * CovdBfu_xz);
								CovdBfd_xy = inv_chi * (vector(gtd_xy, i, j, k) * CovdBfu_xx + vector(gtd_yy, i, j, k) * CovdBfu_xy + vector(gtd_yz, i, j, k) * CovdBfu_xz);
							} else {
								CovdBvf_y = 0.0;
								CovdBvf_x = 0.0;
								CovdBfd_xx = 0.0;
								CovdBfd_xz = 0.0;
								CovdBfd_xy = 0.0;
							}
							optdepthxd = d_optdepthxd_o0_t0_m0_l0;
							optdepthxh = d_optdepthxh_o0_t0_m0_l0;
							optdepthxv = d_optdepthxv_o0_t0_m0_l0;
							chiav = d_chiav_o0_t0_m0_l0;
							if (calculate_les_terms) {
								CovdEfu_zz = d_CovdEfu_zz_o0_t0_m0_l0;
								CovdEfu_zy = d_CovdEfu_zy_o0_t0_m0_l0;
								CovdEfu_zx = d_CovdEfu_zx_o0_t0_m0_l0;
								CovdEfd_zz = inv_chi * (vector(gtd_xz, i, j, k) * CovdEfu_zx + vector(gtd_yz, i, j, k) * CovdEfu_zy + vector(gtd_zz, i, j, k) * CovdEfu_zz);
								CovdEfd_zx = inv_chi * (vector(gtd_xx, i, j, k) * CovdEfu_zx + vector(gtd_xy, i, j, k) * CovdEfu_zy + vector(gtd_xz, i, j, k) * CovdEfu_zz);
								CovdEfd_zy = inv_chi * (vector(gtd_xy, i, j, k) * CovdEfu_zx + vector(gtd_yy, i, j, k) * CovdEfu_zy + vector(gtd_yz, i, j, k) * CovdEfu_zz);
							} else {
								CovdEfu_zz = 0.0;
								CovdEfu_zy = 0.0;
								CovdEfu_zx = 0.0;
								CovdEfd_zz = 0.0;
								CovdEfd_zx = 0.0;
								CovdEfd_zy = 0.0;
							}
							Bvf = (vector(Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(Bfu_z, i, j, k) * vector(vfd_z, i, j, k)) / sdetg;
							optdepthah = d_optdepthah_o0_t0_m0_l0;
							optdepthad = d_optdepthad_o0_t0_m0_l0;
							h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
							chiah = d_chiah_o0_t0_m0_l0;
							chiad = d_chiad_o0_t0_m0_l0;
							if (calculate_les_terms) {
								CovdYef_y = d_CovdYef_y_o0_t0_m0_l0;
							} else {
								CovdYef_y = 0.0;
							}
							optdepthav = d_optdepthav_o0_t0_m0_l0;
							if (calculate_les_terms) {
								CovdYef_z = d_CovdYef_z_o0_t0_m0_l0;
								CovdYef_x = d_CovdYef_x_o0_t0_m0_l0;
							} else {
								CovdYef_z = 0.0;
								CovdYef_x = 0.0;
							}
							detgtd = vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) * vector(gtd_xx, i, j, k) + (-vector(gtd_yy, i, j, k) * (vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k))) + (-vector(gtd_zz, i, j, k) * (vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k))) + (-(vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k)) * vector(gtd_xx, i, j, k)) + 2.0 * vector(gtd_yz, i, j, k) * vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k);
							idetgtd = 1.0 / detgtd;
							gtu_zz = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_yy, i, j, k) - vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k));
							gtu_yy = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k));
							gtu_yz = idetgtd * ((-vector(gtd_xx, i, j, k) * vector(gtd_yz, i, j, k)) + vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k));
							gtu_xx = idetgtd * (vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k));
							gtu_xz = idetgtd * (vector(gtd_xy, i, j, k) * vector(gtd_yz, i, j, k) - vector(gtd_yy, i, j, k) * vector(gtd_xz, i, j, k));
							if (calculate_les_terms) {
								CovuBfu_zx = chi_max * (gtu_xz * CovdBfu_xx + gtu_yz * CovdBfu_yx + gtu_zz * CovdBfu_zx);
								CovuBfu_zz = chi_max * (gtu_xz * CovdBfu_xz + gtu_yz * CovdBfu_yz + gtu_zz * CovdBfu_zz);
								CovuBfu_zy = chi_max * (gtu_xz * CovdBfu_xy + gtu_yz * CovdBfu_yy + gtu_zz * CovdBfu_zy);
								Covdvfu_xz = chi_max * (gtu_xz * Covdvfd_xx + gtu_yz * Covdvfd_xy + gtu_zz * Covdvfd_xz);
								CovuYef_z = chi_max * (gtu_xz * CovdYef_x + gtu_yz * CovdYef_y + gtu_zz * CovdYef_z);
								Covdvfu_zz = chi_max * (gtu_xz * Covdvfd_zx + gtu_yz * Covdvfd_zy + gtu_zz * Covdvfd_zz);
								Covdvfu_yz = chi_max * (gtu_xz * Covdvfd_yx + gtu_yz * Covdvfd_yy + gtu_zz * Covdvfd_yz);
								Covuvfu_zz = chi_max * (gtu_xz * Covdvfu_xz + gtu_yz * Covdvfu_yz + gtu_zz * Covdvfu_zz);
							} else {
								CovuBfu_zx = 0.0;
								CovuBfu_zz = 0.0;
								CovuBfu_zy = 0.0;
								Covdvfu_xz = 0.0;
								CovuYef_z = 0.0;
								Covdvfu_zz = 0.0;
								Covdvfu_yz = 0.0;
								Covuvfu_zz = 0.0;
							}
							Sfu_z = chi_max * (gtu_xz * vector(Sfd_x, i, j, k) + gtu_yz * vector(Sfd_y, i, j, k) + gtu_zz * vector(Sfd_z, i, j, k));
							if (calculate_les_terms) {
								Covurhof_z = chi_max * (gtu_xz * Covdrhof_x + gtu_yz * Covdrhof_y + gtu_zz * Covdrhof_z);
								CovuBvf_z = chi_max * (gtu_xz * CovdBvf_x + gtu_yz * CovdBvf_y + gtu_zz * CovdBvf_z);
							} else {
								Covurhof_z = 0.0;
								CovuBvf_z = 0.0;
							}
							gtu_xy = idetgtd * ((-vector(gtd_xy, i, j, k) * vector(gtd_zz, i, j, k)) + vector(gtd_yz, i, j, k) * vector(gtd_xz, i, j, k));
							if (calculate_les_terms) {
								CovuBfu_yz = chi_max * (gtu_xy * CovdBfu_xz + gtu_yy * CovdBfu_yz + gtu_yz * CovdBfu_zz);
								CovuBfu_yy = chi_max * (gtu_xy * CovdBfu_xy + gtu_yy * CovdBfu_yy + gtu_yz * CovdBfu_zy);
								CovuBfu_yx = chi_max * (gtu_xy * CovdBfu_xx + gtu_yy * CovdBfu_yx + gtu_yz * CovdBfu_zx);
								Covdvfu_xy = chi_max * (gtu_xy * Covdvfd_xx + gtu_yy * Covdvfd_xy + gtu_yz * Covdvfd_xz);
								Covdvfu_xx = chi_max * (gtu_xx * Covdvfd_xx + gtu_xy * Covdvfd_xy + gtu_xz * Covdvfd_xz);
								CovdinvsqW_x = (-2.0 * vector(vfd_x, i, j, k) * Covdvfu_xx) + (-2.0 * vector(vfd_y, i, j, k) * Covdvfu_xy) + (-2.0 * vector(vfd_z, i, j, k) * Covdvfu_xz);
								CovuYef_x = chi_max * (gtu_xx * CovdYef_x + gtu_xy * CovdYef_y + gtu_xz * CovdYef_z);
								CovuYef_y = chi_max * (gtu_xy * CovdYef_x + gtu_yy * CovdYef_y + gtu_yz * CovdYef_z);
								Covdvfu_zy = chi_max * (gtu_xy * Covdvfd_zx + gtu_yy * Covdvfd_zy + gtu_yz * Covdvfd_zz);
								Covdvfu_zx = chi_max * (gtu_xx * Covdvfd_zx + gtu_xy * Covdvfd_zy + gtu_xz * Covdvfd_zz);
								CovdinvsqW_z = (-2.0 * vector(vfd_x, i, j, k) * Covdvfu_zx) + (-2.0 * vector(vfd_y, i, j, k) * Covdvfu_zy) + (-2.0 * vector(vfd_z, i, j, k) * Covdvfu_zz);
								Covdvfu_yx = chi_max * (gtu_xx * Covdvfd_yx + gtu_xy * Covdvfd_yy + gtu_xz * Covdvfd_yz);
								Covuvfu_zx = chi_max * (gtu_xz * Covdvfu_xx + gtu_yz * Covdvfu_yx + gtu_zz * Covdvfu_zx);
								Covdvfu_yy = chi_max * (gtu_xy * Covdvfd_yx + gtu_yy * Covdvfd_yy + gtu_yz * Covdvfd_yz);
								CovdinvsqW_y = (-2.0 * vector(vfd_x, i, j, k) * Covdvfu_yx) + (-2.0 * vector(vfd_y, i, j, k) * Covdvfu_yy) + (-2.0 * vector(vfd_z, i, j, k) * Covdvfu_yz);
								CovuinvsqW_z = chi_max * (gtu_xz * CovdinvsqW_x + gtu_yz * CovdinvsqW_y + gtu_zz * CovdinvsqW_z);
								Covuvfu_zy = chi_max * (gtu_xz * Covdvfu_xy + gtu_yz * Covdvfu_yy + gtu_zz * Covdvfu_zy);
								CovuBfu_xz = chi_max * (gtu_xx * CovdBfu_xz + gtu_xy * CovdBfu_yz + gtu_xz * CovdBfu_zz);
								CovuBfu_xy = chi_max * (gtu_xx * CovdBfu_xy + gtu_xy * CovdBfu_yy + gtu_xz * CovdBfu_zy);
								CovuBfu_xx = chi_max * (gtu_xx * CovdBfu_xx + gtu_xy * CovdBfu_yx + gtu_xz * CovdBfu_zx);
							} else {
								CovuBfu_yz = 0.0;
								CovuBfu_yy = 0.0;
								CovuBfu_yx = 0.0;
								Covdvfu_xy = 0.0;
								Covdvfu_xx = 0.0;
								CovdinvsqW_x = 0.0;
								CovuYef_x = 0.0;
								CovuYef_y = 0.0;
								Covdvfu_zy = 0.0;
								Covdvfu_zx = 0.0;
								CovdinvsqW_z = 0.0;
								Covdvfu_yx = 0.0;
								Covuvfu_zx = 0.0;
								Covdvfu_yy = 0.0;
								CovdinvsqW_y = 0.0;
								CovuinvsqW_z = 0.0;
								Covuvfu_zy = 0.0;
								CovuBfu_xz = 0.0;
								CovuBfu_xy = 0.0;
								CovuBfu_xx = 0.0;
							}
							Sfu_y = chi_max * (gtu_xy * vector(Sfd_x, i, j, k) + gtu_yy * vector(Sfd_y, i, j, k) + gtu_yz * vector(Sfd_z, i, j, k));
							Sfu_x = chi_max * (gtu_xx * vector(Sfd_x, i, j, k) + gtu_xy * vector(Sfd_y, i, j, k) + gtu_xz * vector(Sfd_z, i, j, k));
							if (calculate_les_terms) {
								Covurhof_x = chi_max * (gtu_xx * Covdrhof_x + gtu_xy * Covdrhof_y + gtu_xz * Covdrhof_z);
								Covurhof_y = chi_max * (gtu_xy * Covdrhof_x + gtu_yy * Covdrhof_y + gtu_yz * Covdrhof_z);
								Covuvfu_yz = chi_max * (gtu_xy * Covdvfu_xz + gtu_yy * Covdvfu_yz + gtu_yz * Covdvfu_zz);
								Covuvfu_yy = chi_max * (gtu_xy * Covdvfu_xy + gtu_yy * Covdvfu_yy + gtu_yz * Covdvfu_zy);
								Covuvfu_yx = chi_max * (gtu_xy * Covdvfu_xx + gtu_yy * Covdvfu_yx + gtu_yz * Covdvfu_zx);
								Covuvfu_xz = chi_max * (gtu_xx * Covdvfu_xz + gtu_xy * Covdvfu_yz + gtu_xz * Covdvfu_zz);
								Covuvfu_xy = chi_max * (gtu_xx * Covdvfu_xy + gtu_xy * Covdvfu_yy + gtu_xz * Covdvfu_zy);
								Covuvfu_xx = chi_max * (gtu_xx * Covdvfu_xx + gtu_xy * Covdvfu_yx + gtu_xz * Covdvfu_zx);
							} else {
								Covurhof_x = 0.0;
								Covurhof_y = 0.0;
								Covuvfu_yz = 0.0;
								Covuvfu_yy = 0.0;
								Covuvfu_yx = 0.0;
								Covuvfu_xz = 0.0;
								Covuvfu_xy = 0.0;
								Covuvfu_xx = 0.0;
							}
							vfu_y = chi_max * (gtu_xy * vector(vfd_x, i, j, k) + gtu_yy * vector(vfd_y, i, j, k) + gtu_yz * vector(vfd_z, i, j, k));
							vfu_x = chi_max * (gtu_xx * vector(vfd_x, i, j, k) + gtu_xy * vector(vfd_y, i, j, k) + gtu_xz * vector(vfd_z, i, j, k));
							if (calculate_les_terms) {
								CovuBvf_x = chi_max * (gtu_xx * CovdBvf_x + gtu_xy * CovdBvf_y + gtu_xz * CovdBvf_z);
								CovuBvf_y = chi_max * (gtu_xy * CovdBvf_x + gtu_yy * CovdBvf_y + gtu_yz * CovdBvf_z);
								CovuinvsqW_y = chi_max * (gtu_xy * CovdinvsqW_x + gtu_yy * CovdinvsqW_y + gtu_yz * CovdinvsqW_z);
								CovuinvsqW_x = chi_max * (gtu_xx * CovdinvsqW_x + gtu_xy * CovdinvsqW_y + gtu_xz * CovdinvsqW_z);
								Covddpfdye_z = d_Covddpfdye_z_o0_t0_m0_l0;
								Covddpfdye_y = d_Covddpfdye_y_o0_t0_m0_l0;
								Covddpfdye_x = d_Covddpfdye_x_o0_t0_m0_l0;
								CovdEfu_xz = d_CovdEfu_xz_o0_t0_m0_l0;
								CovdEfu_xy = d_CovdEfu_xy_o0_t0_m0_l0;
								CovdEfu_xx = d_CovdEfu_xx_o0_t0_m0_l0;
								CovdEfd_xx = inv_chi * (vector(gtd_xx, i, j, k) * CovdEfu_xx + vector(gtd_xy, i, j, k) * CovdEfu_xy + vector(gtd_xz, i, j, k) * CovdEfu_xz);
								CovdEfd_xy = inv_chi * (vector(gtd_xy, i, j, k) * CovdEfu_xx + vector(gtd_yy, i, j, k) * CovdEfu_xy + vector(gtd_yz, i, j, k) * CovdEfu_xz);
								CovdEfd_xz = inv_chi * (vector(gtd_xz, i, j, k) * CovdEfu_xx + vector(gtd_yz, i, j, k) * CovdEfu_xy + vector(gtd_zz, i, j, k) * CovdEfu_xz);
							} else {
								CovuBvf_x = 0.0;
								CovuBvf_y = 0.0;
								CovuinvsqW_y = 0.0;
								CovuinvsqW_x = 0.0;
								Covddpfdye_z = 0.0;
								Covddpfdye_y = 0.0;
								Covddpfdye_x = 0.0;
								CovdEfu_xz = 0.0;
								CovdEfu_xy = 0.0;
								CovdEfu_xx = 0.0;
								CovdEfd_xx = 0.0;
								CovdEfd_xy = 0.0;
								CovdEfd_xz = 0.0;
							}
							vfu_z = chi_max * (gtu_xz * vector(vfd_x, i, j, k) + gtu_yz * vector(vfd_y, i, j, k) + gtu_zz * vector(vfd_z, i, j, k));
							sqvf = MIN(vfu_x * vector(vfd_x, i, j, k) + vfu_y * vector(vfd_y, i, j, k) + vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
							sqW = 1.0 / (1.0 - sqvf);
							if (calculate_les_terms) {
								PhiAh = sqW * (vector(pf, i, j, k) * vector(dpfdeps, i, j, k) + (vector(rhof, i, j, k) * vector(rhof, i, j, k)) * vector(dpfdrho, i, j, k));
							} else {
								PhiAh = 0.0;
							}
							Epf = h * sqW;
							Thetaf = Epf + sqBf / (sdetg * sdetg);
							if (calculate_les_terms) {
								Covdepsf_z = d_Covdepsf_z_o0_t0_m0_l0;
								Covdhf_z = (1.0 + vector(epsf, i, j, k)) * Covdrhof_z + vector(rhof, i, j, k) * Covdepsf_z + Covdpf_z;
								CovdEpf_z = sqW * Covdhf_z + (-h * (sqW * sqW) * CovdinvsqW_z);
								CovdThetaf_z = CovdEpf_z + 2.0 * (Bfd_x * CovdBfu_zx) / sdetg + 2.0 * (Bfd_y * CovdBfu_zy) / sdetg + 2.0 * (Bfd_z * CovdBfu_zz) / sdetg;
								Covdepsf_y = d_Covdepsf_y_o0_t0_m0_l0;
								Covdhf_y = (1.0 + vector(epsf, i, j, k)) * Covdrhof_y + vector(rhof, i, j, k) * Covdepsf_y + Covdpf_y;
								CovdEpf_y = sqW * Covdhf_y + (-h * (sqW * sqW) * CovdinvsqW_y);
								CovdThetaf_y = CovdEpf_y + 2.0 * (Bfd_x * CovdBfu_yx) / sdetg + 2.0 * (Bfd_y * CovdBfu_yy) / sdetg + 2.0 * (Bfd_z * CovdBfu_yz) / sdetg;
								Covdepsf_x = d_Covdepsf_x_o0_t0_m0_l0;
								Covuepsf_z = chi_max * (gtu_xz * Covdepsf_x + gtu_yz * Covdepsf_y + gtu_zz * Covdepsf_z);
								Covuepsf_y = chi_max * (gtu_xy * Covdepsf_x + gtu_yy * Covdepsf_y + gtu_yz * Covdepsf_z);
								Covdhf_x = (1.0 + vector(epsf, i, j, k)) * Covdrhof_x + vector(rhof, i, j, k) * Covdepsf_x + Covdpf_x;
								CovdEpf_x = sqW * Covdhf_x + (-h * (sqW * sqW) * CovdinvsqW_x);
								CovdThetaf_x = CovdEpf_x + 2.0 * (Bfd_x * CovdBfu_xx) / sdetg + 2.0 * (Bfd_y * CovdBfu_xy) / sdetg + 2.0 * (Bfd_z * CovdBfu_xz) / sdetg;
								CovuThetaf_z = chi_max * (gtu_xz * CovdThetaf_x + gtu_yz * CovdThetaf_y + gtu_zz * CovdThetaf_z);
								CovuThetaf_x = chi_max * (gtu_xx * CovdThetaf_x + gtu_xy * CovdThetaf_y + gtu_xz * CovdThetaf_z);
								CovuThetaf_y = chi_max * (gtu_xy * CovdThetaf_x + gtu_yy * CovdThetaf_y + gtu_yz * CovdThetaf_z);
								Phivh_z = 2.0 * (CovuBvf_x * CovdBfu_xz + CovuBvf_y * CovdBfu_yz + CovuBvf_z * CovdBfu_zz + (-CovuThetaf_x * Covdvfu_xz) + (-CovuThetaf_y * Covdvfu_yz) + (-CovuThetaf_z * Covdvfu_zz) + (vector(Bfu_z, i, j, k) * (Thetaf * (CovuBfu_xx * Covdvfd_xx + CovuBfu_yx * Covdvfd_yx + CovuBfu_zx * Covdvfd_zx + CovuBfu_xy * Covdvfd_xy + CovuBfu_yy * Covdvfd_yy + CovuBfu_zy * Covdvfd_zy + CovuBfu_xz * Covdvfd_xz + CovuBfu_yz * Covdvfd_yz + CovuBfu_zz * Covdvfd_zz) + (Bfd_x * CovdBfu_xx * CovuBvf_x) / sdetg + (Bfd_y * CovdBfu_xy * CovuBvf_x) / sdetg + (Bfd_z * CovdBfu_xz * CovuBvf_x) / sdetg + (Bfd_x * CovdBfu_yx * CovuBvf_y) / sdetg + (Bfd_y * CovdBfu_yy * CovuBvf_y) / sdetg + (Bfd_z * CovdBfu_yz * CovuBvf_y) / sdetg + (Bfd_x * CovdBfu_zx * CovuBvf_z) / sdetg + (Bfd_y * CovdBfu_zy * CovuBvf_z) / sdetg + (Bfd_z * CovdBfu_zz * CovuBvf_z) / sdetg + (-(vector(Bfu_x, i, j, k) * Covdvfd_xx * CovuThetaf_x) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_xy * CovuThetaf_x) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_xz * CovuThetaf_x) / sdetg) + (-(vector(Bfu_x, i, j, k) * Covdvfd_yx * CovuThetaf_y) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_yy * CovuThetaf_y) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_yz * CovuThetaf_y) / sdetg) + (-(vector(Bfu_x, i, j, k) * Covdvfd_zx * CovuThetaf_z) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_zy * CovuThetaf_z) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_zz * CovuThetaf_z) / sdetg))) / (sdetg * Epf)) / Thetaf;
								Phivh_x = 2.0 * (CovuBvf_x * CovdBfu_xx + CovuBvf_y * CovdBfu_yx + CovuBvf_z * CovdBfu_zx + (-CovuThetaf_x * Covdvfu_xx) + (-CovuThetaf_y * Covdvfu_yx) + (-CovuThetaf_z * Covdvfu_zx) + (vector(Bfu_x, i, j, k) * (Thetaf * (CovuBfu_xx * Covdvfd_xx + CovuBfu_yx * Covdvfd_yx + CovuBfu_zx * Covdvfd_zx + CovuBfu_xy * Covdvfd_xy + CovuBfu_yy * Covdvfd_yy + CovuBfu_zy * Covdvfd_zy + CovuBfu_xz * Covdvfd_xz + CovuBfu_yz * Covdvfd_yz + CovuBfu_zz * Covdvfd_zz) + (Bfd_x * CovdBfu_xx * CovuBvf_x) / sdetg + (Bfd_y * CovdBfu_xy * CovuBvf_x) / sdetg + (Bfd_z * CovdBfu_xz * CovuBvf_x) / sdetg + (Bfd_x * CovdBfu_yx * CovuBvf_y) / sdetg + (Bfd_y * CovdBfu_yy * CovuBvf_y) / sdetg + (Bfd_z * CovdBfu_yz * CovuBvf_y) / sdetg + (Bfd_x * CovdBfu_zx * CovuBvf_z) / sdetg + (Bfd_y * CovdBfu_zy * CovuBvf_z) / sdetg + (Bfd_z * CovdBfu_zz * CovuBvf_z) / sdetg + (-(vector(Bfu_x, i, j, k) * Covdvfd_xx * CovuThetaf_x) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_xy * CovuThetaf_x) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_xz * CovuThetaf_x) / sdetg) + (-(vector(Bfu_x, i, j, k) * Covdvfd_yx * CovuThetaf_y) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_yy * CovuThetaf_y) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_yz * CovuThetaf_y) / sdetg) + (-(vector(Bfu_x, i, j, k) * Covdvfd_zx * CovuThetaf_z) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_zy * CovuThetaf_z) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_zz * CovuThetaf_z) / sdetg))) / (sdetg * Epf)) / Thetaf;
								Phivh_y = 2.0 * (CovuBvf_x * CovdBfu_xy + CovuBvf_y * CovdBfu_yy + CovuBvf_z * CovdBfu_zy + (-CovuThetaf_x * Covdvfu_xy) + (-CovuThetaf_y * Covdvfu_yy) + (-CovuThetaf_z * Covdvfu_zy) + (vector(Bfu_y, i, j, k) * (Thetaf * (CovuBfu_xx * Covdvfd_xx + CovuBfu_yx * Covdvfd_yx + CovuBfu_zx * Covdvfd_zx + CovuBfu_xy * Covdvfd_xy + CovuBfu_yy * Covdvfd_yy + CovuBfu_zy * Covdvfd_zy + CovuBfu_xz * Covdvfd_xz + CovuBfu_yz * Covdvfd_yz + CovuBfu_zz * Covdvfd_zz) + (Bfd_x * CovdBfu_xx * CovuBvf_x) / sdetg + (Bfd_y * CovdBfu_xy * CovuBvf_x) / sdetg + (Bfd_z * CovdBfu_xz * CovuBvf_x) / sdetg + (Bfd_x * CovdBfu_yx * CovuBvf_y) / sdetg + (Bfd_y * CovdBfu_yy * CovuBvf_y) / sdetg + (Bfd_z * CovdBfu_yz * CovuBvf_y) / sdetg + (Bfd_x * CovdBfu_zx * CovuBvf_z) / sdetg + (Bfd_y * CovdBfu_zy * CovuBvf_z) / sdetg + (Bfd_z * CovdBfu_zz * CovuBvf_z) / sdetg + (-(vector(Bfu_x, i, j, k) * Covdvfd_xx * CovuThetaf_x) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_xy * CovuThetaf_x) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_xz * CovuThetaf_x) / sdetg) + (-(vector(Bfu_x, i, j, k) * Covdvfd_yx * CovuThetaf_y) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_yy * CovuThetaf_y) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_yz * CovuThetaf_y) / sdetg) + (-(vector(Bfu_x, i, j, k) * Covdvfd_zx * CovuThetaf_z) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_zy * CovuThetaf_z) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_zz * CovuThetaf_z) / sdetg))) / (sdetg * Epf)) / Thetaf;
								PhiMh_yz = 2.0 * (Thetaf * ((-Covdvfu_xz * CovuBfu_xy) + Covdvfu_xy * CovuBfu_xz + (-Covdvfu_yz * CovuBfu_yy) + Covdvfu_yy * CovuBfu_yz + (-Covdvfu_zz * CovuBfu_zy) + Covdvfu_zy * CovuBfu_zz) + (vector(Bfu_z, i, j, k) * CovuBfu_xy * CovdBvf_x) / sdetg + (vector(Bfu_z, i, j, k) * CovuBfu_yy * CovdBvf_y) / sdetg + (vector(Bfu_z, i, j, k) * CovuBfu_zy * CovdBvf_z) / sdetg + (-(vector(Bfu_y, i, j, k) * CovuBfu_xz * CovdBvf_x) / sdetg) + (-(vector(Bfu_y, i, j, k) * CovuBfu_yz * CovdBvf_y) / sdetg) + (-(vector(Bfu_y, i, j, k) * CovuBfu_zz * CovdBvf_z) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfu_xy * CovuThetaf_x) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfu_yy * CovuThetaf_y) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfu_zy * CovuThetaf_z) / sdetg) + (vector(Bfu_y, i, j, k) * Covdvfu_xz * CovuThetaf_x) / sdetg + (vector(Bfu_y, i, j, k) * Covdvfu_yz * CovuThetaf_y) / sdetg + (vector(Bfu_y, i, j, k) * Covdvfu_zz * CovuThetaf_z) / sdetg) / Thetaf;
								PhiMh_xz = 2.0 * (Thetaf * ((-Covdvfu_xz * CovuBfu_xx) + Covdvfu_xx * CovuBfu_xz + (-Covdvfu_yz * CovuBfu_yx) + Covdvfu_yx * CovuBfu_yz + (-Covdvfu_zz * CovuBfu_zx) + Covdvfu_zx * CovuBfu_zz) + (vector(Bfu_z, i, j, k) * CovuBfu_xx * CovdBvf_x) / sdetg + (vector(Bfu_z, i, j, k) * CovuBfu_yx * CovdBvf_y) / sdetg + (vector(Bfu_z, i, j, k) * CovuBfu_zx * CovdBvf_z) / sdetg + (-(vector(Bfu_x, i, j, k) * CovuBfu_xz * CovdBvf_x) / sdetg) + (-(vector(Bfu_x, i, j, k) * CovuBfu_yz * CovdBvf_y) / sdetg) + (-(vector(Bfu_x, i, j, k) * CovuBfu_zz * CovdBvf_z) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfu_xx * CovuThetaf_x) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfu_yx * CovuThetaf_y) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfu_zx * CovuThetaf_z) / sdetg) + (vector(Bfu_x, i, j, k) * Covdvfu_xz * CovuThetaf_x) / sdetg + (vector(Bfu_x, i, j, k) * Covdvfu_yz * CovuThetaf_y) / sdetg + (vector(Bfu_x, i, j, k) * Covdvfu_zz * CovuThetaf_z) / sdetg) / Thetaf;
								PhiMh_xy = 2.0 * (Thetaf * ((-Covdvfu_xy * CovuBfu_xx) + Covdvfu_xx * CovuBfu_xy + (-Covdvfu_yy * CovuBfu_yx) + Covdvfu_yx * CovuBfu_yy + (-Covdvfu_zy * CovuBfu_zx) + Covdvfu_zx * CovuBfu_zy) + (vector(Bfu_y, i, j, k) * CovuBfu_xx * CovdBvf_x) / sdetg + (vector(Bfu_y, i, j, k) * CovuBfu_yx * CovdBvf_y) / sdetg + (vector(Bfu_y, i, j, k) * CovuBfu_zx * CovdBvf_z) / sdetg + (-(vector(Bfu_x, i, j, k) * CovuBfu_xy * CovdBvf_x) / sdetg) + (-(vector(Bfu_x, i, j, k) * CovuBfu_yy * CovdBvf_y) / sdetg) + (-(vector(Bfu_x, i, j, k) * CovuBfu_zy * CovdBvf_z) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfu_xx * CovuThetaf_x) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfu_yx * CovuThetaf_y) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfu_zx * CovuThetaf_z) / sdetg) + (vector(Bfu_x, i, j, k) * Covdvfu_xy * CovuThetaf_x) / sdetg + (vector(Bfu_x, i, j, k) * Covdvfu_yy * CovuThetaf_y) / sdetg + (vector(Bfu_x, i, j, k) * Covdvfu_zy * CovuThetaf_z) / sdetg) / Thetaf;
								Covuepsf_x = chi_max * (gtu_xx * Covdepsf_x + gtu_xy * Covdepsf_y + gtu_xz * Covdepsf_z);
							} else {
								Covdepsf_z = 0.0;
								Covdhf_z = 0.0;
								CovdEpf_z = 0.0;
								CovdThetaf_z = 0.0;
								Covdepsf_y = 0.0;
								Covdhf_y = 0.0;
								CovdEpf_y = 0.0;
								CovdThetaf_y = 0.0;
								Covdepsf_x = 0.0;
								Covuepsf_z = 0.0;
								Covuepsf_y = 0.0;
								Covdhf_x = 0.0;
								CovdEpf_x = 0.0;
								CovdThetaf_x = 0.0;
								CovuThetaf_z = 0.0;
								CovuThetaf_x = 0.0;
								CovuThetaf_y = 0.0;
								Phivh_z = 0.0;
								Phivh_x = 0.0;
								Phivh_y = 0.0;
								PhiMh_yz = 0.0;
								PhiMh_xz = 0.0;
								PhiMh_xy = 0.0;
								Covuepsf_x = 0.0;
							}
							faceta = 0.04166666666666667 * inv_chi * (dx[0] * dx[0]) * Smooth(threshold_Dles_min, threshold_Dles_max, vector(Df, i, j, k), dx, simPlat_dt, ilast, jlast);
							if (calculate_les_terms) {
								CovdEfu_yz = d_CovdEfu_yz_o0_t0_m0_l0;
								CovuEfu_zz = chi_max * (gtu_xz * CovdEfu_xz + gtu_yz * CovdEfu_yz + gtu_zz * CovdEfu_zz);
								CovuEfu_xz = chi_max * (gtu_xx * CovdEfu_xz + gtu_xy * CovdEfu_yz + gtu_xz * CovdEfu_zz);
								CovdEfu_yy = d_CovdEfu_yy_o0_t0_m0_l0;
								CovuEfu_zy = chi_max * (gtu_xz * CovdEfu_xy + gtu_yz * CovdEfu_yy + gtu_zz * CovdEfu_zy);
								CovuEfu_xy = chi_max * (gtu_xx * CovdEfu_xy + gtu_xy * CovdEfu_yy + gtu_xz * CovdEfu_zy);
								CovdEfu_yx = d_CovdEfu_yx_o0_t0_m0_l0;
								CovdEfd_yy = inv_chi * (vector(gtd_xy, i, j, k) * CovdEfu_yx + vector(gtd_yy, i, j, k) * CovdEfu_yy + vector(gtd_yz, i, j, k) * CovdEfu_yz);
								CovdEfd_yz = inv_chi * (vector(gtd_xz, i, j, k) * CovdEfu_yx + vector(gtd_yz, i, j, k) * CovdEfu_yy + vector(gtd_zz, i, j, k) * CovdEfu_yz);
								CovdEfd_yx = inv_chi * (vector(gtd_xx, i, j, k) * CovdEfu_yx + vector(gtd_xy, i, j, k) * CovdEfu_yy + vector(gtd_xz, i, j, k) * CovdEfu_yz);
								CovuEfu_zx = chi_max * (gtu_xz * CovdEfu_xx + gtu_yz * CovdEfu_yx + gtu_zz * CovdEfu_zx);
								CovuEfu_xx = chi_max * (gtu_xx * CovdEfu_xx + gtu_xy * CovdEfu_yx + gtu_xz * CovdEfu_zx);
								CovuEfu_yx = chi_max * (gtu_xy * CovdEfu_xx + gtu_yy * CovdEfu_yx + gtu_yz * CovdEfu_zx);
								CovuEfu_yz = chi_max * (gtu_xy * CovdEfu_xz + gtu_yy * CovdEfu_yz + gtu_yz * CovdEfu_zz);
								CovuEfu_yy = chi_max * (gtu_xy * CovdEfu_xy + gtu_yy * CovdEfu_yy + gtu_yz * CovdEfu_zy);
								PhiThetah = (Thetaf * (CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + (-CovdEfd_xx * CovuEfu_xx) + (-CovdEfd_yx * CovuEfu_yx) + (-CovdEfd_zx * CovuEfu_zx) + (-CovdEfd_xy * CovuEfu_xy) + (-CovdEfd_yy * CovuEfu_yy) + (-CovdEfd_zy * CovuEfu_zy) + (-CovdEfd_xz * CovuEfu_xz) + (-CovdEfd_yz * CovuEfu_yz) + (-CovdEfd_zz * CovuEfu_zz) + (-0.5 * (((-vector(vfd_y, i, j, k) * Bfd_x) + vector(vfd_x, i, j, k) * Bfd_y) * PhiMh_xy + (vector(vfd_x, i, j, k) * Bfd_z - vector(vfd_z, i, j, k) * Bfd_x) * PhiMh_xz + (-(vector(vfd_y, i, j, k) * Bfd_x - vector(vfd_x, i, j, k) * Bfd_y) * PhiMh_xy) + ((-vector(vfd_z, i, j, k) * Bfd_y) + vector(vfd_y, i, j, k) * Bfd_z) * PhiMh_yz + (-(vector(vfd_z, i, j, k) * Bfd_x - vector(vfd_x, i, j, k) * Bfd_z) * PhiMh_xz) + (-(vector(vfd_z, i, j, k) * Bfd_y - vector(vfd_y, i, j, k) * Bfd_z) * PhiMh_yz)) / sdetg))) / (Thetaf - sqEf);
								HPres = (Epf * sqW * (Thetaf - sqEf) * (vector(rhof, i, j, k) * (Covddpfdrho_x * Covurhof_x + Covddpfdrho_y * Covurhof_y + Covddpfdrho_z * Covurhof_z + Covddpfdeps_x * Covuepsf_x + Covddpfdeps_y * Covuepsf_y + Covddpfdeps_z * Covuepsf_z) + (-2.0 * vector(dpfdeps, i, j, k) * (Covurhof_x * Covdepsf_x + Covurhof_y * Covdepsf_y + Covurhof_z * Covdepsf_z)) + (-(Epf * vector(dpfdeps, i, j, k) - PhiAh) * (0.25 * sqW * (CovdinvsqW_x * CovuinvsqW_x + CovdinvsqW_y * CovuinvsqW_y + CovdinvsqW_z * CovuinvsqW_z) + (CovdinvsqW_x * Covurhof_x + CovdinvsqW_y * Covurhof_y + CovdinvsqW_z * Covurhof_z) / vector(rhof, i, j, k))) + (-2.0 * (vector(dpfdeps, i, j, k) * (CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + (-(sqW * sqW) * (CovuinvsqW_x * Covdhf_x + CovuinvsqW_y * Covdhf_y + CovuinvsqW_z * Covdhf_z)))) / sqW) + (-(Epf * vector(dpfdeps, i, j, k) + PhiAh) * (vector(vfd_x, i, j, k) * Phivh_x + vector(vfd_y, i, j, k) * Phivh_y + vector(vfd_z, i, j, k) * Phivh_z + Covdvfd_xx * Covuvfu_xx + Covdvfd_xy * Covuvfu_xy + Covdvfd_xz * Covuvfu_xz + Covdvfd_yx * Covuvfu_yx + Covdvfd_yy * Covuvfu_yy + Covdvfd_yz * Covuvfu_yz + Covdvfd_zx * Covuvfu_zx + Covdvfd_zy * Covuvfu_zy + Covdvfd_zz * Covuvfu_zz + sqW * (CovdinvsqW_x * CovuinvsqW_x + CovdinvsqW_y * CovuinvsqW_y + CovdinvsqW_z * CovuinvsqW_z))) + (PhiThetah * ((Epf * vector(dpfdeps, i, j, k) + PhiAh) * (Thetaf - sqEf) - (PhiAh * Thetaf) / sqW)) / (Epf * Thetaf))) / ((Epf * vector(rhof, i, j, k) - PhiAh) * (Thetaf - sqEf) * sqW + PhiAh * Thetaf) + Covddpfdye_x * CovuYef_x + Covddpfdye_y * CovuYef_y + Covddpfdye_z * CovuYef_z - 2.0 * vector(dpfdye, i, j, k) / vector(Df, i, j, k) * (CovdDf_x * CovuYef_x + CovdDf_y * CovuYef_y + CovdDf_z * CovuYef_z);
								HTheta = PhiThetah + (Thetaf * HPres) / (Thetaf - sqEf);
								Hv_z = Phivh_z + (-((vfu_z + (Bvf * vector(Bfu_z, i, j, k)) / (sdetg * Epf)) * HTheta) / Thetaf);
								HN_z = 2.0 * CovdDf_x * Covuvfu_xz + 2.0 * CovdDf_y * Covuvfu_yz + 2.0 * CovdDf_z * Covuvfu_zz + (vector(Df, i, j, k) * Hv_z) / sdetg;
								HNe_z = 2.0 * CovdDYf_x * Covuvfu_xz + 2.0 * CovdDYf_y * Covuvfu_yz + 2.0 * CovdDYf_z * Covuvfu_zz + (vector(DYf, i, j, k) * Hv_z) / sdetg;
								Hv_y = Phivh_y + (-((vfu_y + (Bvf * vector(Bfu_y, i, j, k)) / (sdetg * Epf)) * HTheta) / Thetaf);
								HN_y = 2.0 * CovdDf_x * Covuvfu_xy + 2.0 * CovdDf_y * Covuvfu_yy + 2.0 * CovdDf_z * Covuvfu_zy + (vector(Df, i, j, k) * Hv_y) / sdetg;
								HM_yz = (-2.0 * Covdvfu_xz * CovuBfu_xy) + 2.0 * Covdvfu_xy * CovuBfu_xz + (-2.0 * Covdvfu_yz * CovuBfu_yy) + 2.0 * Covdvfu_yy * CovuBfu_yz + (-2.0 * Covdvfu_zz * CovuBfu_zy) + 2.0 * Covdvfu_zy * CovuBfu_zz + (vector(Bfu_z, i, j, k) * Hv_y - vector(Bfu_y, i, j, k) * Hv_z) / sdetg;
								HNe_y = 2.0 * CovdDYf_x * Covuvfu_xy + 2.0 * CovdDYf_y * Covuvfu_yy + 2.0 * CovdDYf_z * Covuvfu_zy + (vector(DYf, i, j, k) * Hv_y) / sdetg;
								Hv_x = Phivh_x + (-((vfu_x + (Bvf * vector(Bfu_x, i, j, k)) / (sdetg * Epf)) * HTheta) / Thetaf);
								HN_x = 2.0 * CovdDf_x * Covuvfu_xx + 2.0 * CovdDf_y * Covuvfu_yx + 2.0 * CovdDf_z * Covuvfu_zx + (vector(Df, i, j, k) * Hv_x) / sdetg;
								HM_xz = (-2.0 * Covdvfu_xz * CovuBfu_xx) + 2.0 * Covdvfu_xx * CovuBfu_xz + (-2.0 * Covdvfu_yz * CovuBfu_yx) + 2.0 * Covdvfu_yx * CovuBfu_yz + (-2.0 * Covdvfu_zz * CovuBfu_zx) + 2.0 * Covdvfu_zx * CovuBfu_zz + (vector(Bfu_z, i, j, k) * Hv_x - vector(Bfu_x, i, j, k) * Hv_z) / sdetg;
								HM_xy = (-2.0 * Covdvfu_xy * CovuBfu_xx) + 2.0 * Covdvfu_xx * CovuBfu_xy + (-2.0 * Covdvfu_yy * CovuBfu_yx) + 2.0 * Covdvfu_yx * CovuBfu_yy + (-2.0 * Covdvfu_zy * CovuBfu_zx) + 2.0 * Covdvfu_zx * CovuBfu_zy + (vector(Bfu_y, i, j, k) * Hv_x - vector(Bfu_x, i, j, k) * Hv_y) / sdetg;
								HE_z = 0.5 * ((-2.0 * gtu_yz * HM_xz) + 2.0 * gtu_zz * HM_xy + 2.0 * gtu_xz * HM_yz) / sqrt(chi_max);
								HE_x = 0.5 * ((-2.0 * gtu_xy * HM_xz) + 2.0 * gtu_xz * HM_xy + 2.0 * gtu_xx * HM_yz) / sqrt(chi_max);
								HE_y = 0.5 * ((-2.0 * gtu_yy * HM_xz) + 2.0 * gtu_yz * HM_xy + 2.0 * gtu_xy * HM_yz) / sqrt(chi_max);
								HT_zz = 4.0 * CovdEpf_x * Covuvfu_xz * vfu_z + 4.0 * CovdEpf_y * Covuvfu_yz * vfu_z + 4.0 * CovdEpf_z * Covuvfu_zz * vfu_z + Epf * (2.0 * vfu_z * Hv_z + 2.0 * Covdvfu_xz * Covuvfu_xz + 2.0 * Covdvfu_yz * Covuvfu_yz + 2.0 * Covdvfu_zz * Covuvfu_zz) + 2.0 * (vfu_z * vfu_z) * HPres + (-2.0 * CovdBfu_xz * CovuBfu_xz) + (-2.0 * CovdBfu_yz * CovuBfu_yz) + (-2.0 * CovdBfu_zz * CovuBfu_zz) + (-2.0 * CovdEfu_xz * CovuEfu_xz) + (-2.0 * CovdEfu_yz * CovuEfu_yz) + (-2.0 * CovdEfu_zz * CovuEfu_zz) + (-2.00 * vector(Efu_z, i, j, k) * HE_z) + (chi_max * gtu_zz - vfu_z * vfu_z) * (HPres + CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + CovdEfd_xx * CovuEfu_xx + CovdEfd_yx * CovuEfu_yx + CovdEfd_zx * CovuEfu_zx + CovdEfd_xy * CovuEfu_xy + CovdEfd_yy * CovuEfu_yy + CovdEfd_zy * CovuEfu_zy + CovdEfd_xz * CovuEfu_xz + CovdEfd_yz * CovuEfu_yz + CovdEfd_zz * CovuEfu_zz + Efd_x * HE_x + Efd_y * HE_y + Efd_z * HE_z);
								HT_yy = 4.0 * CovdEpf_x * Covuvfu_xy * vfu_y + 4.0 * CovdEpf_y * Covuvfu_yy * vfu_y + 4.0 * CovdEpf_z * Covuvfu_zy * vfu_y + Epf * (2.0 * vfu_y * Hv_y + 2.0 * Covdvfu_xy * Covuvfu_xy + 2.0 * Covdvfu_yy * Covuvfu_yy + 2.0 * Covdvfu_zy * Covuvfu_zy) + 2.0 * (vfu_y * vfu_y) * HPres + (-2.0 * CovdBfu_xy * CovuBfu_xy) + (-2.0 * CovdBfu_yy * CovuBfu_yy) + (-2.0 * CovdBfu_zy * CovuBfu_zy) + (-2.0 * CovdEfu_xy * CovuEfu_xy) + (-2.0 * CovdEfu_yy * CovuEfu_yy) + (-2.0 * CovdEfu_zy * CovuEfu_zy) + (-2.00 * vector(Efu_y, i, j, k) * HE_y) + (chi_max * gtu_yy - vfu_y * vfu_y) * (HPres + CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + CovdEfd_xx * CovuEfu_xx + CovdEfd_yx * CovuEfu_yx + CovdEfd_zx * CovuEfu_zx + CovdEfd_xy * CovuEfu_xy + CovdEfd_yy * CovuEfu_yy + CovdEfd_zy * CovuEfu_zy + CovdEfd_xz * CovuEfu_xz + CovdEfd_yz * CovuEfu_yz + CovdEfd_zz * CovuEfu_zz + Efd_x * HE_x + Efd_y * HE_y + Efd_z * HE_z);
								HT_yz = 2.0 * CovdEpf_x * (Covuvfu_xz * vfu_y + Covuvfu_xy * vfu_z) + 2.0 * CovdEpf_y * (Covuvfu_yz * vfu_y + Covuvfu_yy * vfu_z) + 2.0 * CovdEpf_z * (Covuvfu_zz * vfu_y + Covuvfu_zy * vfu_z) + Epf * (vfu_y * Hv_z + vfu_z * Hv_y + 2.0 * Covdvfu_xy * Covuvfu_xz + 2.0 * Covdvfu_yy * Covuvfu_yz + 2.0 * Covdvfu_zy * Covuvfu_zz) + 2.0 * vfu_y * vfu_z * HPres + (-2.0 * CovdBfu_xy * CovuBfu_xz) + (-2.0 * CovdBfu_yy * CovuBfu_yz) + (-2.0 * CovdBfu_zy * CovuBfu_zz) + (-2.0 * CovdEfu_xy * CovuEfu_xz) + (-2.0 * CovdEfu_yy * CovuEfu_yz) + (-2.0 * CovdEfu_zy * CovuEfu_zz) + (-1.00 * vector(Efu_z, i, j, k) * HE_y) + (-1.00 * vector(Efu_y, i, j, k) * HE_z) + (chi_max * gtu_yz - vfu_z * vfu_y) * (HPres + CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + CovdEfd_xx * CovuEfu_xx + CovdEfd_yx * CovuEfu_yx + CovdEfd_zx * CovuEfu_zx + CovdEfd_xy * CovuEfu_xy + CovdEfd_yy * CovuEfu_yy + CovdEfd_zy * CovuEfu_zy + CovdEfd_xz * CovuEfu_xz + CovdEfd_yz * CovuEfu_yz + CovdEfd_zz * CovuEfu_zz + Efd_x * HE_x + Efd_y * HE_y + Efd_z * HE_z);
								HT_xx = 4.0 * CovdEpf_x * Covuvfu_xx * vfu_x + 4.0 * CovdEpf_y * Covuvfu_yx * vfu_x + 4.0 * CovdEpf_z * Covuvfu_zx * vfu_x + Epf * (2.0 * vfu_x * Hv_x + 2.0 * Covdvfu_xx * Covuvfu_xx + 2.0 * Covdvfu_yx * Covuvfu_yx + 2.0 * Covdvfu_zx * Covuvfu_zx) + 2.0 * (vfu_x * vfu_x) * HPres + (-2.0 * CovdBfu_xx * CovuBfu_xx) + (-2.0 * CovdBfu_yx * CovuBfu_yx) + (-2.0 * CovdBfu_zx * CovuBfu_zx) + (-2.0 * CovdEfu_xx * CovuEfu_xx) + (-2.0 * CovdEfu_yx * CovuEfu_yx) + (-2.0 * CovdEfu_zx * CovuEfu_zx) + (-2.00 * vector(Efu_x, i, j, k) * HE_x) + (chi_max * gtu_xx - vfu_x * vfu_x) * (HPres + CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + CovdEfd_xx * CovuEfu_xx + CovdEfd_yx * CovuEfu_yx + CovdEfd_zx * CovuEfu_zx + CovdEfd_xy * CovuEfu_xy + CovdEfd_yy * CovuEfu_yy + CovdEfd_zy * CovuEfu_zy + CovdEfd_xz * CovuEfu_xz + CovdEfd_yz * CovuEfu_yz + CovdEfd_zz * CovuEfu_zz + Efd_x * HE_x + Efd_y * HE_y + Efd_z * HE_z);
								HT_xz = 2.0 * CovdEpf_x * (Covuvfu_xz * vfu_x + Covuvfu_xx * vfu_z) + 2.0 * CovdEpf_y * (Covuvfu_yz * vfu_x + Covuvfu_yx * vfu_z) + 2.0 * CovdEpf_z * (Covuvfu_zz * vfu_x + Covuvfu_zx * vfu_z) + Epf * (vfu_x * Hv_z + vfu_z * Hv_x + 2.0 * Covdvfu_xx * Covuvfu_xz + 2.0 * Covdvfu_yx * Covuvfu_yz + 2.0 * Covdvfu_zx * Covuvfu_zz) + 2.0 * vfu_x * vfu_z * HPres + (-2.0 * CovdBfu_xx * CovuBfu_xz) + (-2.0 * CovdBfu_yx * CovuBfu_yz) + (-2.0 * CovdBfu_zx * CovuBfu_zz) + (-2.0 * CovdEfu_xx * CovuEfu_xz) + (-2.0 * CovdEfu_yx * CovuEfu_yz) + (-2.0 * CovdEfu_zx * CovuEfu_zz) + (-1.00 * vector(Efu_z, i, j, k) * HE_x) + (-1.00 * vector(Efu_x, i, j, k) * HE_z) + (chi_max * gtu_xz - vfu_z * vfu_x) * (HPres + CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + CovdEfd_xx * CovuEfu_xx + CovdEfd_yx * CovuEfu_yx + CovdEfd_zx * CovuEfu_zx + CovdEfd_xy * CovuEfu_xy + CovdEfd_yy * CovuEfu_yy + CovdEfd_zy * CovuEfu_zy + CovdEfd_xz * CovuEfu_xz + CovdEfd_yz * CovuEfu_yz + CovdEfd_zz * CovuEfu_zz + Efd_x * HE_x + Efd_y * HE_y + Efd_z * HE_z);
								HT_xy = 2.0 * CovdEpf_x * (Covuvfu_xy * vfu_x + Covuvfu_xx * vfu_y) + 2.0 * CovdEpf_y * (Covuvfu_yy * vfu_x + Covuvfu_yx * vfu_y) + 2.0 * CovdEpf_z * (Covuvfu_zy * vfu_x + Covuvfu_zx * vfu_y) + Epf * (vfu_x * Hv_y + vfu_y * Hv_x + 2.0 * Covdvfu_xx * Covuvfu_xy + 2.0 * Covdvfu_yx * Covuvfu_yy + 2.0 * Covdvfu_zx * Covuvfu_zy) + 2.0 * vfu_x * vfu_y * HPres + (-2.0 * CovdBfu_xx * CovuBfu_xy) + (-2.0 * CovdBfu_yx * CovuBfu_yy) + (-2.0 * CovdBfu_zx * CovuBfu_zy) + (-2.0 * CovdEfu_xx * CovuEfu_xy) + (-2.0 * CovdEfu_yx * CovuEfu_yy) + (-2.0 * CovdEfu_zx * CovuEfu_zy) + (-1.00 * vector(Efu_y, i, j, k) * HE_x) + (-1.00 * vector(Efu_x, i, j, k) * HE_y) + (chi_max * gtu_xy - vfu_y * vfu_x) * (HPres + CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + CovdEfd_xx * CovuEfu_xx + CovdEfd_yx * CovuEfu_yx + CovdEfd_zx * CovuEfu_zx + CovdEfd_xy * CovuEfu_xy + CovdEfd_yy * CovuEfu_yy + CovdEfd_zy * CovuEfu_zy + CovdEfd_xz * CovuEfu_xz + CovdEfd_yz * CovuEfu_yz + CovdEfd_zz * CovuEfu_zz + Efd_x * HE_x + Efd_y * HE_y + Efd_z * HE_z);
								HNe_x = 2.0 * CovdDYf_x * Covuvfu_xx + 2.0 * CovdDYf_y * Covuvfu_yx + 2.0 * CovdDYf_z * Covuvfu_zx + (vector(DYf, i, j, k) * Hv_x) / sdetg;
								d_TauN_x_o0_t0_m0_l0 = -faceta * C_N * HN_x;
								d_TauN_y_o0_t0_m0_l0 = -faceta * C_N * HN_y;
								d_TauN_z_o0_t0_m0_l0 = -faceta * C_N * HN_z;
								d_TauNe_x_o0_t0_m0_l0 = -faceta * C_N * HNe_x;
								d_TauNe_y_o0_t0_m0_l0 = -faceta * C_N * HNe_y;
								d_TauNe_z_o0_t0_m0_l0 = -faceta * C_N * HNe_z;
								d_TauM_xy_o0_t0_m0_l0 = -faceta * C_M * HM_xy;
								d_TauM_xz_o0_t0_m0_l0 = -faceta * C_M * HM_xz;
								d_TauM_yz_o0_t0_m0_l0 = -faceta * C_M * HM_yz;
								d_TauT_xx_o0_t0_m0_l0 = -faceta * C_T * HT_xx;
								d_TauT_xy_o0_t0_m0_l0 = -faceta * C_T * HT_xy;
								d_TauT_xz_o0_t0_m0_l0 = -faceta * C_T * HT_xz;
								d_TauT_yy_o0_t0_m0_l0 = -faceta * C_T * HT_yy;
								d_TauT_yz_o0_t0_m0_l0 = -faceta * C_T * HT_yz;
								d_TauT_zz_o0_t0_m0_l0 = -faceta * C_T * HT_zz;
								vector(TauNe_x, i, j, k) = d_TauNe_x_o0_t0_m0_l0;
								vector(TauNe_y, i, j, k) = d_TauNe_y_o0_t0_m0_l0;
								vector(TauT_yz, i, j, k) = d_TauT_yz_o0_t0_m0_l0;
								vector(TauT_xy, i, j, k) = d_TauT_xy_o0_t0_m0_l0;
								vector(TauM_yz, i, j, k) = d_TauM_yz_o0_t0_m0_l0;
								vector(TauM_xy, i, j, k) = d_TauM_xy_o0_t0_m0_l0;
								vector(TauNe_z, i, j, k) = d_TauNe_z_o0_t0_m0_l0;
								vector(TauT_xz, i, j, k) = d_TauT_xz_o0_t0_m0_l0;
								vector(TauN_x, i, j, k) = d_TauN_x_o0_t0_m0_l0;
								vector(TauT_zz, i, j, k) = d_TauT_zz_o0_t0_m0_l0;
								vector(TauT_yy, i, j, k) = d_TauT_yy_o0_t0_m0_l0;
								vector(TauT_xx, i, j, k) = d_TauT_xx_o0_t0_m0_l0;
								vector(TauM_xz, i, j, k) = d_TauM_xz_o0_t0_m0_l0;
								vector(TauN_z, i, j, k) = d_TauN_z_o0_t0_m0_l0;
								vector(TauN_y, i, j, k) = d_TauN_y_o0_t0_m0_l0;
							} else {
								CovdEfu_yz = 0.0;
								CovuEfu_zz = 0.0;
								CovuEfu_xz = 0.0;
								CovdEfu_yy = 0.0;
								CovuEfu_zy = 0.0;
								CovuEfu_xy = 0.0;
								CovdEfu_yx = 0.0;
								CovdEfd_yy = 0.0;
								CovdEfd_yz = 0.0;
								CovdEfd_yx = 0.0;
								CovuEfu_zx = 0.0;
								CovuEfu_xx = 0.0;
								CovuEfu_yx = 0.0;
								CovuEfu_yz = 0.0;
								CovuEfu_yy = 0.0;
								PhiThetah = 0.0;
								HPres = 0.0;
								HTheta = 0.0;
								Hv_z = 0.0;
								HN_z = 0.0;
								HNe_z = 0.0;
								Hv_y = 0.0;
								HN_y = 0.0;
								HM_yz = 0.0;
								HNe_y = 0.0;
								Hv_x = 0.0;
								HN_x = 0.0;
								HM_xz = 0.0;
								HM_xy = 0.0;
								HE_z = 0.0;
								HE_x = 0.0;
								HE_y = 0.0;
								HT_zz = 0.0;
								HT_yy = 0.0;
								HT_yz = 0.0;
								HT_xx = 0.0;
								HT_xz = 0.0;
								HT_xy = 0.0;
								HNe_x = 0.0;
							}
							optdepthe_tmp = vector(optdepthe, i, j, k);
							optdeptha_tmp = vector(optdeptha, i, j, k);
							optdepthx_tmp = vector(optdepthx, i, j, k);
							chie_tmp = vector(chie, i, j, k);
							chia_tmp = vector(chia, i, j, k);
							chix_tmp = vector(chix, i, j, k);
							qnu_tmp = vector(qnu, i, j, k);
							rnu_tmp = vector(rnu, i, j, k);
							qnu_e_tmp = vector(qnu_e, i, j, k);
							qnu_a_tmp = vector(qnu_a, i, j, k);
							qnu_x_tmp = vector(qnu_x, i, j, k);
							betabeta = vector(gtd_xx, i, j, k) * vector(Betau_x, i, j, k) * vector(Betau_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Betau_y, i, j, k) * vector(Betau_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Betau_z, i, j, k) * vector(Betau_z, i, j, k) + 2.0 * vector(gtd_xy, i, j, k) * vector(Betau_x, i, j, k) * vector(Betau_y, i, j, k) + 2.0 * vector(gtd_xz, i, j, k) * vector(Betau_x, i, j, k) * vector(Betau_z, i, j, k) + 2.0 * vector(gtd_yz, i, j, k) * vector(Betau_y, i, j, k) * vector(Betau_z, i, j, k);
							betav = vector(gtd_xx, i, j, k) * vector(Betau_x, i, j, k) * vector(vfd_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Betau_y, i, j, k) * vector(vfd_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Betau_z, i, j, k) * vector(vfd_z, i, j, k) + vector(gtd_xy, i, j, k) * vector(Betau_x, i, j, k) * vector(vfd_y, i, j, k) + vector(gtd_xy, i, j, k) * vector(Betau_y, i, j, k) * vector(vfd_x, i, j, k) + vector(gtd_xz, i, j, k) * vector(Betau_x, i, j, k) * vector(vfd_z, i, j, k) + vector(gtd_xz, i, j, k) * vector(Betau_z, i, j, k) * vector(vfd_x, i, j, k) + vector(gtd_yz, i, j, k) * vector(Betau_y, i, j, k) * vector(vfd_z, i, j, k) + vector(gtd_yz, i, j, k) * vector(Betau_z, i, j, k) * vector(vfd_y, i, j, k);
							velr = vector(vfd_x, i, j, k) * xcoord(i) + vector(vfd_y, i, j, k) * ycoord(j) + vector(vfd_z, i, j, k) * zcoord(k);
							velr = velr / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k) + 0.01);
							f_redshift = sqW * (vector(Alpha, i, j, k) - betav) * (1.0 + velr);
							f_redshift = f_redshift * sqrt(vector(Alpha, i, j, k) * vector(Alpha, i, j, k) - betabeta);
							leakage(do_leakage, vacuum_rho, threshold_leakage_vacuum, energyShift, optdepthe_tmp, optdeptha_tmp, optdepthx_tmp, qnu_e_tmp, qnu_a_tmp, qnu_x_tmp, chie_tmp, chia_tmp, chix_tmp, qnu_tmp, rnu_tmp, vector(rhof, i, j, k), vector(epsf, i, j, k), vector(Yef, i, j, k), vector(Tf, i, j, k), sdetg, optdeptheh, optdepthah, optdepthxh, optdepthev, optdepthav, optdepthxv, optdepthed, optdepthad, optdepthxd, chieh, chiah, chixh, chiev, chiav, chixv, chied, chiad, chixd, f_redshift, coord0_data_DataLeakage, coord0_dx_DataLeakage, coord0_size_DataLeakage, coord1_data_DataLeakage, coord1_dx_DataLeakage, coord1_size_DataLeakage, coord2_data_DataLeakage, coord2_dx_DataLeakage, coord2_size_DataLeakage, vars_data_DataLeakage, vars_size_DataLeakage, dx, simPlat_dt, ilast, jlast);
							vector(optdepthe, i, j, k) = optdepthe_tmp;
							vector(optdeptha, i, j, k) = optdeptha_tmp;
							vector(optdepthx, i, j, k) = optdepthx_tmp;
							vector(chie, i, j, k) = chie_tmp;
							vector(chia, i, j, k) = chia_tmp;
							vector(chix, i, j, k) = chix_tmp;
							vector(qnu_e, i, j, k) = qnu_e_tmp;
							vector(qnu_a, i, j, k) = qnu_a_tmp;
							vector(qnu_x, i, j, k) = qnu_x_tmp;
							vector(qnu, i, j, k) = qnu_tmp;
							vector(rnu, i, j, k) = rnu_tmp;
						}
					}
				}
			}
		}
		for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
			const std::shared_ptr< hier::Patch >& patch = *p_it;
			double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
			double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
			double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
			double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
			double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
			double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
			double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
			double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
			double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
			double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
			double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
			double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
			double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
			double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
			double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
			double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
			double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
			double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
			double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
			double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
			double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
			double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
			double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
			double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
			double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
			double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
			double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
			double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
			double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
			double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
			double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
			double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
			double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
			double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
			double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
			double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
			double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
			double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
			double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
			double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
			double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
			double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
			double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
			double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
			double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
			double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
			double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
			double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
			double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
			double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
			double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
			double* gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_id).get())->getPointer();
			double* gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_id).get())->getPointer();
			double* gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_id).get())->getPointer();
			double* gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_id).get())->getPointer();
			double* gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_id).get())->getPointer();
			double* gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_id).get())->getPointer();
			double* Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xx_id).get())->getPointer();
			double* Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xy_id).get())->getPointer();
			double* Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xz_id).get())->getPointer();
			double* Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yy_id).get())->getPointer();
			double* Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yz_id).get())->getPointer();
			double* Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_zz_id).get())->getPointer();
			double* Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_x_id).get())->getPointer();
			double* Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_y_id).get())->getPointer();
			double* Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_z_id).get())->getPointer();
			double* Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_id).get())->getPointer();
			double* Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_id).get())->getPointer();
			double* Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_id).get())->getPointer();
			double* Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_id).get())->getPointer();
			double* chi = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_id).get())->getPointer();
			double* trK = ((pdat::NodeData<double> *) patch->getPatchData(d_trK_id).get())->getPointer();
			double* theta = ((pdat::NodeData<double> *) patch->getPatchData(d_theta_id).get())->getPointer();

			//Hard region field distance variables
			double* d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
			double* d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
			double* d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();

			//Get the dimensions of the patch
			const hier::Index boxfirst = patch->getBox().lower();
			const hier::Index boxlast = patch->getBox().upper();

			//Get delta spaces into an array. dx, dy, dz.
			const std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
			const double* dx  = patch_geom->getDx();
			int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
			int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
			int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = klast/2; k < klast; k++) {
			for(int j = jlast/2; j < jlast; j++) {
				for(int i = ilast/2; i < ilast; i++) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
							vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
							vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
							vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
							vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
							vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
							vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
							vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
							vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
							vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
							vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
							vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
							vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
							vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
							vector(pf, i, j, k) = vector(pf, i - 1, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
							vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
							vector(chie, i, j, k) = vector(chie, i - 1, j, k);
							vector(chia, i, j, k) = vector(chia, i - 1, j, k);
							vector(chix, i, j, k) = vector(chix, i - 1, j, k);
							vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
							vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
							vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
							vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
							vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
							vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
							vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
							vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
							vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
							vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
							vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
							vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
							vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
							vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
							vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
							vector(pf, i, j, k) = vector(pf, i, j - 1, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
							vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
							vector(chie, i, j, k) = vector(chie, i, j - 1, k);
							vector(chia, i, j, k) = vector(chia, i, j - 1, k);
							vector(chix, i, j, k) = vector(chix, i, j - 1, k);
							vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
							vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
							vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
							vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
							vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
							vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
							vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
							vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
							vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
							vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
							vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
							vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
							vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
							vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
							vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
							vector(pf, i, j, k) = vector(pf, i, j, k - 1);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
							vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
							vector(chie, i, j, k) = vector(chie, i, j, k - 1);
							vector(chia, i, j, k) = vector(chia, i, j, k - 1);
							vector(chix, i, j, k) = vector(chix, i, j, k - 1);
							vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
							vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2 - 1; k >= 0; k--) {
			for(int j = jlast/2; j < jlast; j++) {
				for(int i = ilast/2; i < ilast; i++) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
							vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
							vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
							vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
							vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
							vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
							vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
							vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
							vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
							vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
							vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
							vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
							vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
							vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
							vector(pf, i, j, k) = vector(pf, i - 1, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
							vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
							vector(chie, i, j, k) = vector(chie, i - 1, j, k);
							vector(chia, i, j, k) = vector(chia, i - 1, j, k);
							vector(chix, i, j, k) = vector(chix, i - 1, j, k);
							vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
							vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
							vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
							vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
							vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
							vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
							vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
							vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
							vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
							vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
							vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
							vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
							vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
							vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
							vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
							vector(pf, i, j, k) = vector(pf, i, j - 1, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
							vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
							vector(chie, i, j, k) = vector(chie, i, j - 1, k);
							vector(chia, i, j, k) = vector(chia, i, j - 1, k);
							vector(chix, i, j, k) = vector(chix, i, j - 1, k);
							vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
							vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
							vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
							vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
							vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
							vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
							vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
							vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
							vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
							vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
							vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
							vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
							vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
							vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
							vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
							vector(pf, i, j, k) = vector(pf, i, j, k - 1);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
							vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
							vector(chie, i, j, k) = vector(chie, i, j, k - 1);
							vector(chia, i, j, k) = vector(chia, i, j, k - 1);
							vector(chix, i, j, k) = vector(chix, i, j, k - 1);
							vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
							vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2 - 1; k >= 0; k--) {
			for(int j = jlast/2 - 1; j >= 0; j--) {
				for(int i = ilast/2; i < ilast; i++) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
							vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
							vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
							vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
							vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
							vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
							vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
							vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
							vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
							vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
							vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
							vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
							vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
							vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
							vector(pf, i, j, k) = vector(pf, i - 1, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
							vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
							vector(chie, i, j, k) = vector(chie, i - 1, j, k);
							vector(chia, i, j, k) = vector(chia, i - 1, j, k);
							vector(chix, i, j, k) = vector(chix, i - 1, j, k);
							vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
							vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
							vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
							vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
							vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
							vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
							vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
							vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
							vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
							vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
							vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
							vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
							vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
							vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
							vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
							vector(pf, i, j, k) = vector(pf, i, j - 1, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
							vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
							vector(chie, i, j, k) = vector(chie, i, j - 1, k);
							vector(chia, i, j, k) = vector(chia, i, j - 1, k);
							vector(chix, i, j, k) = vector(chix, i, j - 1, k);
							vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
							vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
							vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
							vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
							vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
							vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
							vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
							vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
							vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
							vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
							vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
							vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
							vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
							vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
							vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
							vector(pf, i, j, k) = vector(pf, i, j, k - 1);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
							vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
							vector(chie, i, j, k) = vector(chie, i, j, k - 1);
							vector(chia, i, j, k) = vector(chia, i, j, k - 1);
							vector(chix, i, j, k) = vector(chix, i, j, k - 1);
							vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
							vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2; k < klast; k++) {
			for(int j = jlast/2 - 1; j >= 0; j--) {
				for(int i = ilast/2; i < ilast; i++) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
							vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
							vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
							vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
							vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
							vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
							vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
							vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
							vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
							vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
							vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
							vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
							vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
							vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
							vector(pf, i, j, k) = vector(pf, i - 1, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
							vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
							vector(chie, i, j, k) = vector(chie, i - 1, j, k);
							vector(chia, i, j, k) = vector(chia, i - 1, j, k);
							vector(chix, i, j, k) = vector(chix, i - 1, j, k);
							vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
							vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
							vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
							vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
							vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
							vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
							vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
							vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
							vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
							vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
							vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
							vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
							vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
							vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
							vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
							vector(pf, i, j, k) = vector(pf, i, j - 1, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
							vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
							vector(chie, i, j, k) = vector(chie, i, j - 1, k);
							vector(chia, i, j, k) = vector(chia, i, j - 1, k);
							vector(chix, i, j, k) = vector(chix, i, j - 1, k);
							vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
							vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
							vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
							vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
							vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
							vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
							vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
							vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
							vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
							vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
							vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
							vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
							vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
							vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
							vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
							vector(pf, i, j, k) = vector(pf, i, j, k - 1);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
							vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
							vector(chie, i, j, k) = vector(chie, i, j, k - 1);
							vector(chia, i, j, k) = vector(chia, i, j, k - 1);
							vector(chix, i, j, k) = vector(chix, i, j, k - 1);
							vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
							vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2; k < klast; k++) {
			for(int j = jlast/2; j < jlast; j++) {
				for(int i = ilast/2 - 1; i >= 0; i--) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
							vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
							vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
							vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
							vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
							vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
							vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
							vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
							vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
							vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
							vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
							vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
							vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
							vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
							vector(pf, i, j, k) = vector(pf, i - 1, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
							vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
							vector(chie, i, j, k) = vector(chie, i - 1, j, k);
							vector(chia, i, j, k) = vector(chia, i - 1, j, k);
							vector(chix, i, j, k) = vector(chix, i - 1, j, k);
							vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
							vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
							vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
							vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
							vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
							vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
							vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
							vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
							vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
							vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
							vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
							vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
							vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
							vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
							vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
							vector(pf, i, j, k) = vector(pf, i, j - 1, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
							vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
							vector(chie, i, j, k) = vector(chie, i, j - 1, k);
							vector(chia, i, j, k) = vector(chia, i, j - 1, k);
							vector(chix, i, j, k) = vector(chix, i, j - 1, k);
							vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
							vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
							vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
							vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
							vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
							vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
							vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
							vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
							vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
							vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
							vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
							vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
							vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
							vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
							vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
							vector(pf, i, j, k) = vector(pf, i, j, k - 1);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
							vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
							vector(chie, i, j, k) = vector(chie, i, j, k - 1);
							vector(chia, i, j, k) = vector(chia, i, j, k - 1);
							vector(chix, i, j, k) = vector(chix, i, j, k - 1);
							vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
							vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2 - 1; k >= 0; k--) {
			for(int j = jlast/2; j < jlast; j++) {
				for(int i = ilast/2 - 1; i >= 0; i--) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
							vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
							vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
							vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
							vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
							vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
							vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
							vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
							vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
							vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
							vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
							vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
							vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
							vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
							vector(pf, i, j, k) = vector(pf, i - 1, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
							vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
							vector(chie, i, j, k) = vector(chie, i - 1, j, k);
							vector(chia, i, j, k) = vector(chia, i - 1, j, k);
							vector(chix, i, j, k) = vector(chix, i - 1, j, k);
							vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
							vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
							vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
							vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
							vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
							vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
							vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
							vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
							vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
							vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
							vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
							vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
							vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
							vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
							vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
							vector(pf, i, j, k) = vector(pf, i, j - 1, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
							vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
							vector(chie, i, j, k) = vector(chie, i, j - 1, k);
							vector(chia, i, j, k) = vector(chia, i, j - 1, k);
							vector(chix, i, j, k) = vector(chix, i, j - 1, k);
							vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
							vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
							vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
							vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
							vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
							vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
							vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
							vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
							vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
							vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
							vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
							vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
							vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
							vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
							vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
							vector(pf, i, j, k) = vector(pf, i, j, k - 1);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
							vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
							vector(chie, i, j, k) = vector(chie, i, j, k - 1);
							vector(chia, i, j, k) = vector(chia, i, j, k - 1);
							vector(chix, i, j, k) = vector(chix, i, j, k - 1);
							vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
							vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2 - 1; k >= 0; k--) {
			for(int j = jlast/2 - 1; j >= 0; j--) {
				for(int i = ilast/2 - 1; i >= 0; i--) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
							vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
							vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
							vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
							vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
							vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
							vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
							vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
							vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
							vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
							vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
							vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
							vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
							vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
							vector(pf, i, j, k) = vector(pf, i - 1, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
							vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
							vector(chie, i, j, k) = vector(chie, i - 1, j, k);
							vector(chia, i, j, k) = vector(chia, i - 1, j, k);
							vector(chix, i, j, k) = vector(chix, i - 1, j, k);
							vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
							vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
							vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
							vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
							vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
							vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
							vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
							vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
							vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
							vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
							vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
							vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
							vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
							vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
							vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
							vector(pf, i, j, k) = vector(pf, i, j - 1, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
							vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
							vector(chie, i, j, k) = vector(chie, i, j - 1, k);
							vector(chia, i, j, k) = vector(chia, i, j - 1, k);
							vector(chix, i, j, k) = vector(chix, i, j - 1, k);
							vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
							vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
							vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
							vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
							vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
							vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
							vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
							vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
							vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
							vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
							vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
							vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
							vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
							vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
							vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
							vector(pf, i, j, k) = vector(pf, i, j, k - 1);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
							vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
							vector(chie, i, j, k) = vector(chie, i, j, k - 1);
							vector(chia, i, j, k) = vector(chia, i, j, k - 1);
							vector(chix, i, j, k) = vector(chix, i, j, k - 1);
							vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
							vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
						}
					}
				}
			}
		for(int k = klast/2; k < klast; k++) {
			for(int j = jlast/2 - 1; j >= 0; j--) {
				for(int i = ilast/2 - 1; i >= 0; i--) {
						if ((vector(FOV_xLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
							vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
							vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
							vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
							vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
							vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
							vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
							vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
							vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
							vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
							vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
						}
						if ((vector(FOV_xUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
							vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
							vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
							vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
							vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
							vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
							vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
							vector(pf, i, j, k) = vector(pf, i - 1, j, k);
							vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
							vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
							vector(chie, i, j, k) = vector(chie, i - 1, j, k);
							vector(chia, i, j, k) = vector(chia, i - 1, j, k);
							vector(chix, i, j, k) = vector(chix, i - 1, j, k);
							vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
							vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
						}
						if ((vector(FOV_yLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
							vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
							vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
							vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
							vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
							vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
							vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
							vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
							vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
							vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
							vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
						}
						if ((vector(FOV_yUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
							vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
							vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
							vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
							vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
							vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
							vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
							vector(pf, i, j, k) = vector(pf, i, j - 1, k);
							vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
							vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
							vector(chie, i, j, k) = vector(chie, i, j - 1, k);
							vector(chia, i, j, k) = vector(chia, i, j - 1, k);
							vector(chix, i, j, k) = vector(chix, i, j - 1, k);
							vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
							vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
						}
						if ((vector(FOV_zLower, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
							vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
							vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
							vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
							vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
							vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
							vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
							vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
							vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
							vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
							vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
						}
						if ((vector(FOV_zUpper, i, j, k) > 0)) {
							//Region field extrapolations
							if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
								extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
							}
							vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
							vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
							vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
							vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
							vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
							vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
							vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
							vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
							vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
							vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
							vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
							vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
							vector(pf, i, j, k) = vector(pf, i, j, k - 1);
							vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
							vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
							vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
							vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
							vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
							vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
							vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
							vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
							vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
							vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
							vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
							vector(chie, i, j, k) = vector(chie, i, j, k - 1);
							vector(chia, i, j, k) = vector(chia, i, j, k - 1);
							vector(chix, i, j, k) = vector(chix, i, j, k - 1);
							vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
							vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
							vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
							vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
							vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
							vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
							vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
							vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
							vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
							vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
							vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
							vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
							vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
							vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
							vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
							vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
							vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
						}
					}
				}
			}
		}
		//Last synchronization from initialization
		d_bdry_fill_init->createSchedule(level,this)->fillData(init_data_time, false);

    	}

    cout<<"Level "<<level_number<<" initialized"<<endl;
    tbox::MemoryUtilities::printMemoryInfo(cout);
}

void Problem::postInit(){
	double current_time = 0;
	for (int ln=0; ln<=d_patch_hierarchy->getFinestLevelNumber(); ++ln ) {
		std::shared_ptr<hier::PatchLevel > level(d_patch_hierarchy->getPatchLevel(ln));
		for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
			const std::shared_ptr<hier::Patch >& patch = *p_it;
	
			double* chi = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_id).get())->getPointer();
			double* gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_id).get())->getPointer();
			double* Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_x_id).get())->getPointer();
			double* Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_y_id).get())->getPointer();
			double* Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_z_id).get())->getPointer();
			double* gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_id).get())->getPointer();
			double* gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_id).get())->getPointer();
			double* gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_id).get())->getPointer();
			double* gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_id).get())->getPointer();
			double* gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_id).get())->getPointer();
			double* Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xy_id).get())->getPointer();
			double* trK = ((pdat::NodeData<double> *) patch->getPatchData(d_trK_id).get())->getPointer();
			double* Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xz_id).get())->getPointer();
			double* Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xx_id).get())->getPointer();
			double* Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yy_id).get())->getPointer();
			double* Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yz_id).get())->getPointer();
			double* Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_zz_id).get())->getPointer();
			double* Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_id).get())->getPointer();
			double* Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_id).get())->getPointer();
			double* Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_id).get())->getPointer();
			double* Df = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_id).get())->getPointer();
			double* tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_tauf_id).get())->getPointer();
			double* Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_id).get())->getPointer();
			double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
			double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
			double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
			double* Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_id).get())->getPointer();
			double* Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_id).get())->getPointer();
			double* Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_id).get())->getPointer();
			double* Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_id).get())->getPointer();
			double* Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_id).get())->getPointer();
			double* Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_id).get())->getPointer();
			double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
			double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
			double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
			double* Rscalar = ((pdat::NodeData<double> *) patch->getPatchData(d_Rscalar_id).get())->getPointer();
			double* HamCon = ((pdat::NodeData<double> *) patch->getPatchData(d_HamCon_id).get())->getPointer();
			double* MomCon_x = ((pdat::NodeData<double> *) patch->getPatchData(d_MomCon_x_id).get())->getPointer();
			double* MomCon_y = ((pdat::NodeData<double> *) patch->getPatchData(d_MomCon_y_id).get())->getPointer();
			double* MomCon_z = ((pdat::NodeData<double> *) patch->getPatchData(d_MomCon_z_id).get())->getPointer();
			double* trA = ((pdat::NodeData<double> *) patch->getPatchData(d_trA_id).get())->getPointer();
			double* detgtm1 = ((pdat::NodeData<double> *) patch->getPatchData(d_detgtm1_id).get())->getPointer();
			double* M_ADM_surf = ((pdat::NodeData<double> *) patch->getPatchData(d_M_ADM_surf_id).get())->getPointer();
			double* Jz_ADM_surf = ((pdat::NodeData<double> *) patch->getPatchData(d_Jz_ADM_surf_id).get())->getPointer();
			double* M_Komar = ((pdat::NodeData<double> *) patch->getPatchData(d_M_Komar_id).get())->getPointer();
			double* Jz_Komar = ((pdat::NodeData<double> *) patch->getPatchData(d_Jz_Komar_id).get())->getPointer();
			double* psi4R = ((pdat::NodeData<double> *) patch->getPatchData(d_psi4R_id).get())->getPointer();
			double* psi4I = ((pdat::NodeData<double> *) patch->getPatchData(d_psi4I_id).get())->getPointer();
			double* Z_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Z_x_id).get())->getPointer();
			double* Z_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Z_y_id).get())->getPointer();
			double* Z_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Z_z_id).get())->getPointer();
			double* flux_mass = ((pdat::NodeData<double> *) patch->getPatchData(d_flux_mass_id).get())->getPointer();
			double d_Rpd_xx_o0_t0_m0_l0, d_Rpd_xx_o0_t1_m0_l0, d_Rpd_xx_o0_t2_m0_l0, d_Rpd_xx_o0_t3_m0_l0, d_Rpd_xx_o0_t4_m0_l0, d_Rpd_xx_o0_t5_m0_l0, d_Rpd_xx_o0_t6_m0_l0, d_Rpd_xx_o0_t7_m0_l0, d_Rpd_xx_o0_t8_m0_l0, d_Ctd_xxx_o0_t0_m0_l0, d_Ctd_xxy_o0_t0_m0_l0, d_Ctd_xxz_o0_t0_m0_l0, d_Rtd_xx_o0_t0_m0_l0, d_Rtd_xx_o0_t1_m0_l0, d_Rtd_xx_o0_t2_m0_l0, d_Rtd_xx_o0_t3_m0_l0, d_Rtd_xx_o0_t4_m0_l0, d_Rtd_xx_o0_t5_m0_l0, d_Rtd_xx_o0_t6_m0_l0, d_Rtd_xy_o0_t6_m0_l0, d_Rtd_xz_o0_t6_m0_l0, d_Rtd_xx_o0_t7_m0_l0, d_Rtd_xy_o0_t7_m0_l0, d_Rtd_xz_o0_t7_m0_l0, d_Rtd_xx_o0_t8_m0_l0, d_Rtd_xy_o0_t8_m0_l0, d_Rtd_xz_o0_t8_m0_l0, d_Ctd_yxx_o0_t0_m0_l0, d_Ctd_xyy_o0_t0_m0_l0, d_Ctd_xyz_o0_t1_m0_l0, d_Rtd_xy_o0_t0_m0_l0, d_Rtd_xy_o0_t1_m0_l0, d_Rtd_xy_o0_t2_m0_l0, d_Rtd_xy_o0_t3_m0_l0, d_Rtd_xy_o0_t4_m0_l0, d_Rtd_xy_o0_t5_m0_l0, d_Ctd_zxx_o0_t0_m0_l0, d_Ctd_xyz_o0_t0_m0_l0, d_Ctd_xzz_o0_t0_m0_l0, d_Rtd_xz_o0_t0_m0_l0, d_Rtd_xz_o0_t1_m0_l0, d_Rtd_xz_o0_t2_m0_l0, d_Rtd_xz_o0_t3_m0_l0, d_Rtd_xz_o0_t4_m0_l0, d_Rtd_xz_o0_t5_m0_l0, d_Ctd_xyy_o0_t1_m0_l0, d_Ctd_yyy_o0_t0_m0_l0, d_Ctd_yyz_o0_t0_m0_l0, d_Rtd_yy_o0_t0_m0_l0, d_Rtd_yy_o0_t1_m0_l0, d_Rtd_yy_o0_t2_m0_l0, d_Rtd_yy_o0_t3_m0_l0, d_Rtd_yy_o0_t4_m0_l0, d_Rtd_yy_o0_t5_m0_l0, d_Ctd_xyz_o0_t2_m0_l0, d_Ctd_zyy_o0_t0_m0_l0, d_Ctd_yzz_o0_t0_m0_l0, d_Rtd_yz_o0_t0_m0_l0, d_Rtd_yz_o0_t1_m0_l0, d_Rtd_yz_o0_t2_m0_l0, d_Rtd_yz_o0_t3_m0_l0, d_Rtd_yz_o0_t4_m0_l0, d_Rtd_yz_o0_t5_m0_l0, d_Ctd_xzz_o0_t1_m0_l0, d_Ctd_yzz_o0_t1_m0_l0, d_Ctd_zzz_o0_t0_m0_l0, d_Rtd_zz_o0_t0_m0_l0, d_Rtd_zz_o0_t1_m0_l0, d_Rtd_zz_o0_t2_m0_l0, d_Rtd_zz_o0_t3_m0_l0, d_Rtd_zz_o0_t4_m0_l0, d_Rtd_zz_o0_t5_m0_l0, d_Del_Kd_xxy_o0_t0_m0_l0, d_Del_Kd_yxy_o0_t0_m0_l0, d_Del_Kd_zxy_o0_t0_m0_l0, d_Del_Kd_xxy_o0_t2_m0_l0, d_Del_Kd_yxx_o0_t2_m0_l0, d_Del_Kd_zxx_o0_t2_m0_l0, d_Del_Kd_xxz_o0_t0_m0_l0, d_Del_Kd_yxz_o0_t0_m0_l0, d_Del_Kd_zxz_o0_t0_m0_l0, d_MomCon_x_o0_t0_m0_l0, d_Del_Kd_yxx_o0_t0_m0_l0, d_Del_Kd_zxx_o0_t0_m0_l0, d_Del_Kd_xyy_o0_t0_m0_l0, d_MomCon_y_o0_t4_m0_l0, d_Del_Kd_zyy_o0_t0_m0_l0, d_Del_Kd_xyz_o0_t0_m0_l0, d_Del_Kd_yyz_o0_t0_m0_l0, d_Del_Kd_zyz_o0_t0_m0_l0, d_Del_Kd_xzz_o0_t0_m0_l0, d_Del_Kd_yzz_o0_t0_m0_l0, d_MomCon_z_o0_t11_m0_l0, inv_chi, g4d_zz, g4d_yz, g4d_yy, Bfd_y, Bfd_z, Bfd_x, sqBf, r_pos, ur_z, ur_y, uph_z, uph_x, uph_y, ur_x, uthd_y, uthd_x, uthd_z, sdetg, Tf4u_tt, m_Ctd_zzz_o0_t0_l0, Ctd_zzz, hf, wrr, m_Ctd_yyz_o0_t0_l0, Ctd_yyz, m_Ctd_zyy_o0_t1_l0, m_Ctd_zyy_o0_t0_l0, Ctd_zyy, m_Ctd_yyy_o0_t0_l0, Ctd_yyy, m_Ctd_xxz_o0_t0_l0, Ctd_xxz, m_Ctd_zxx_o0_t1_l0, m_Ctd_zxx_o0_t0_l0, Ctd_zxx, m_Ctd_xxx_o0_t0_l0, Ctd_xxx, m_Ctd_yzz_o0_t1_l0, m_Ctd_yzz_o0_t0_l0, Ctd_yzz, m_Ctd_zyz_o0_t0_l0, Ctd_zyz, m_Ctd_xxy_o0_t0_l0, Ctd_xxy, m_Ctd_yxx_o0_t1_l0, m_Ctd_yxx_o0_t0_l0, Ctd_yxx, m_Ctd_xyy_o0_t1_l0, m_Ctd_xyy_o0_t0_l0, Ctd_xyy, m_Ctd_yxy_o0_t0_l0, Ctd_yxy, m_Ctd_xyz_o0_t2_l0, m_Ctd_xyz_o0_t1_l0, m_Ctd_xyz_o0_t0_l0, Ctd_xyz, m_Ctd_yxz_o0_t2_l0, m_Ctd_yxz_o0_t0_l0, Ctd_yxz, m_Ctd_zxy_o0_t2_l0, Ctd_zxy, g4d_tt, g4d_ty, g4d_tx, g4d_tz, m_Ctd_xzz_o0_t1_l0, m_Ctd_xzz_o0_t0_l0, Ctd_xzz, m_Ctd_zxz_o0_t0_l0, Ctd_zxz, dSigma_y, dSigma_x, dSigma_z, detgtd, idetgtd, gtu_zz, gtu_yy, gtu_yz, gtu_xx, gtu_xz, uth_z, Atud_zx, Atud_zz, Atud_zy, Atu_zz, Sfu_z, Tf4u_tz, Ct_zxz, Ct_zxy, Ct_zxx, Ct_zyz, Ct_zyy, Ct_zzz, gtu_xy, Ct_xzz, uth_x, uth_y, Gamt_z, Zu_z, Atud_yy, Atud_yx, Atud_yz, Atu_yz, Atud_xx, Atud_xz, Atud_xy, Atu_xz, Sfu_y, Tf4u_ty, Sfu_x, Tf4u_tx, vfu_y, Ju_y, vfu_x, Ju_x, Atu_yy, Ct_yxx, Ct_yxz, Ct_yxy, Atu_xy, Atu_xx, Ct_yyy, Ct_yyz, Tu_ty, Tu_tx, Tu_tz, Ct_yzz, Gamt_y, Zu_y, Ct_xxy, vfu_z, sqvf, sqW, Ju_z, Ct_xxx, Ct_xxz, d_Rtd_xx_o0_t9_m0_l0, m_Rtd_xx_o0_t8_l0, m_Rtd_xx_o0_t7_l0, m_Rtd_xx_o0_t6_l0, m_Rtd_xx_o0_t5_l0, m_Rtd_xx_o0_t4_l0, m_Rtd_xx_o0_t3_l0, m_Rtd_xx_o0_t2_l0, m_Rtd_xx_o0_t1_l0, m_Rtd_xx_o0_t0_l0, Rtd_xx, g4d_xy, wphph, vph_y, vph_z, vph_x, wthph, g4d_xx, wrph, vr_x, vr_y, vr_z, wthr, g4d_xz, wthth, vth_x, mmR_xx, mmI_xx, vth_y, mmR_yy, mmR_xy, mmI_xy, mmI_yy, vth_z, mmR_yz, mmR_zz, mmR_xz, mmI_zz, mmI_xz, mmI_yz, Ct_xyz, d_Rtd_xz_o0_t12_m0_l0, m_Rtd_xz_o0_t11_l0, m_Rtd_xz_o0_t10_l0, m_Rtd_xz_o0_t9_l0, m_Rtd_xz_o0_t8_l0, m_Rtd_xz_o0_t7_l0, m_Rtd_xz_o0_t6_l0, m_Rtd_xz_o0_t5_l0, m_Rtd_xz_o0_t4_l0, m_Rtd_xz_o0_t3_l0, m_Rtd_xz_o0_t2_l0, m_Rtd_xz_o0_t1_l0, m_Rtd_xz_o0_t0_l0, Rtd_xz, d_Rtd_zz_o0_t9_m0_l0, m_Rtd_zz_o0_t8_l0, m_Rtd_zz_o0_t7_l0, m_Rtd_zz_o0_t6_l0, m_Rtd_zz_o0_t5_l0, m_Rtd_zz_o0_t4_l0, m_Rtd_zz_o0_t3_l0, m_Rtd_zz_o0_t2_l0, m_Rtd_zz_o0_t1_l0, m_Rtd_zz_o0_t0_l0, Rtd_zz, Ct_xyy, Gamt_x, Zu_x, m_Rpd_xx_o0_t22_l0, m_Rpd_xx_o0_t21_l0, m_Rpd_xx_o0_t20_l0, m_Rpd_xx_o0_t19_l0, m_Rpd_xx_o0_t18_l0, m_Rpd_xx_o0_t17_l0, m_Rpd_xx_o0_t16_l0, m_Rpd_xx_o0_t15_l0, m_Rpd_xx_o0_t14_l0, m_Rpd_xx_o0_t13_l0, m_Rpd_xx_o0_t12_l0, m_Rpd_xx_o0_t11_l0, m_Rpd_xx_o0_t10_l0, m_Rpd_xx_o0_t9_l0, m_Rpd_xx_o0_t8_l0, m_Rpd_xx_o0_t7_l0, m_Rpd_xx_o0_t6_l0, m_Rpd_xx_o0_t5_l0, m_Rpd_xx_o0_t4_l0, m_Rpd_xx_o0_t3_l0, m_Rpd_xx_o0_t2_l0, m_Rpd_xx_o0_t1_l0, m_Rpd_xx_o0_t0_l0, Rpd_xx, d_Del_Kd_yxx_o0_t1_m0_l0, m_Del_Kd_yxx_o0_t6_l0, m_Del_Kd_yxx_o0_t5_l0, m_Del_Kd_yxx_o0_t4_l0, m_Del_Kd_yxx_o0_t3_l0, m_Del_Kd_yxx_o0_t2_l0, m_Del_Kd_yxx_o0_t0_l0, Del_Kd_yxx, d_Del_Kd_yxz_o0_t1_m0_l0, m_Del_Kd_yxz_o0_t10_l0, m_Del_Kd_yxz_o0_t9_l0, m_Del_Kd_yxz_o0_t8_l0, m_Del_Kd_yxz_o0_t7_l0, m_Del_Kd_yxz_o0_t6_l0, m_Del_Kd_yxz_o0_t5_l0, m_Del_Kd_yxz_o0_t4_l0, m_Del_Kd_yxz_o0_t3_l0, m_Del_Kd_yxz_o0_t2_l0, m_Del_Kd_yxz_o0_t0_l0, Del_Kd_yxz, EWeyl_xx, d_Del_Kd_zxx_o0_t1_m0_l0, m_Del_Kd_zxx_o0_t6_l0, m_Del_Kd_zxx_o0_t5_l0, m_Del_Kd_zxx_o0_t4_l0, m_Del_Kd_zxx_o0_t3_l0, m_Del_Kd_zxx_o0_t2_l0, m_Del_Kd_zxx_o0_t0_l0, Del_Kd_zxx, d_Del_Kd_zyz_o0_t1_m0_l0, m_Del_Kd_zyz_o0_t10_l0, m_Del_Kd_zyz_o0_t9_l0, m_Del_Kd_zyz_o0_t8_l0, m_Del_Kd_zyz_o0_t7_l0, m_Del_Kd_zyz_o0_t6_l0, m_Del_Kd_zyz_o0_t5_l0, m_Del_Kd_zyz_o0_t4_l0, m_Del_Kd_zyz_o0_t3_l0, m_Del_Kd_zyz_o0_t2_l0, m_Del_Kd_zyz_o0_t0_l0, Del_Kd_zyz, d_Del_Kd_zyy_o0_t1_m0_l0, m_Del_Kd_zyy_o0_t6_l0, m_Del_Kd_zyy_o0_t5_l0, m_Del_Kd_zyy_o0_t4_l0, m_Del_Kd_zyy_o0_t3_l0, m_Del_Kd_zyy_o0_t2_l0, m_Del_Kd_zyy_o0_t0_l0, Del_Kd_zyy, d_Del_Kd_zxy_o0_t1_m0_l0, m_Del_Kd_zxy_o0_t7_l0, m_Del_Kd_zxy_o0_t6_l0, m_Del_Kd_zxy_o0_t5_l0, m_Del_Kd_zxy_o0_t4_l0, m_Del_Kd_zxy_o0_t2_l0, m_Del_Kd_zxy_o0_t0_l0, Del_Kd_zxy, d_Del_Kd_zxz_o0_t1_m0_l0, m_Del_Kd_zxz_o0_t10_l0, m_Del_Kd_zxz_o0_t9_l0, m_Del_Kd_zxz_o0_t8_l0, m_Del_Kd_zxz_o0_t7_l0, m_Del_Kd_zxz_o0_t6_l0, m_Del_Kd_zxz_o0_t5_l0, m_Del_Kd_zxz_o0_t4_l0, m_Del_Kd_zxz_o0_t3_l0, m_Del_Kd_zxz_o0_t2_l0, m_Del_Kd_zxz_o0_t0_l0, Del_Kd_zxz, m_Rpd_yz_o0_t25_l0, m_Rpd_yz_o0_t24_l0, m_Rpd_yz_o0_t23_l0, m_Rpd_yz_o0_t22_l0, m_Rpd_yz_o0_t21_l0, m_Rpd_yz_o0_t20_l0, m_Rpd_yz_o0_t19_l0, m_Rpd_yz_o0_t18_l0, m_Rpd_yz_o0_t17_l0, m_Rpd_yz_o0_t16_l0, m_Rpd_yz_o0_t15_l0, m_Rpd_yz_o0_t14_l0, m_Rpd_yz_o0_t13_l0, m_Rpd_yz_o0_t12_l0, m_Rpd_yz_o0_t11_l0, m_Rpd_yz_o0_t10_l0, m_Rpd_yz_o0_t9_l0, m_Rpd_yz_o0_t8_l0, m_Rpd_yz_o0_t7_l0, m_Rpd_yz_o0_t6_l0, m_Rpd_yz_o0_t5_l0, m_Rpd_yz_o0_t4_l0, m_Rpd_yz_o0_t3_l0, m_Rpd_yz_o0_t2_l0, m_Rpd_yz_o0_t1_l0, m_Rpd_yz_o0_t0_l0, Rpd_yz, d_Del_Kd_yzz_o0_t1_m0_l0, m_Del_Kd_yzz_o0_t6_l0, m_Del_Kd_yzz_o0_t5_l0, m_Del_Kd_yzz_o0_t4_l0, m_Del_Kd_yzz_o0_t3_l0, m_Del_Kd_yzz_o0_t2_l0, m_Del_Kd_yzz_o0_t0_l0, Del_Kd_yzz, d_Del_Kd_xxy_o0_t1_m0_l0, m_Del_Kd_xxy_o0_t10_l0, m_Del_Kd_xxy_o0_t9_l0, m_Del_Kd_xxy_o0_t8_l0, m_Del_Kd_xxy_o0_t7_l0, m_Del_Kd_xxy_o0_t6_l0, m_Del_Kd_xxy_o0_t5_l0, m_Del_Kd_xxy_o0_t4_l0, m_Del_Kd_xxy_o0_t3_l0, m_Del_Kd_xxy_o0_t2_l0, m_Del_Kd_xxy_o0_t0_l0, Del_Kd_xxy, d_Del_Kd_xzz_o0_t1_m0_l0, m_Del_Kd_xzz_o0_t6_l0, m_Del_Kd_xzz_o0_t5_l0, m_Del_Kd_xzz_o0_t4_l0, m_Del_Kd_xzz_o0_t3_l0, m_Del_Kd_xzz_o0_t2_l0, m_Del_Kd_xzz_o0_t0_l0, Del_Kd_xzz, d_Del_Kd_xyz_o0_t1_m0_l0, m_Del_Kd_xyz_o0_t2_l0, m_Del_Kd_xyz_o0_t0_l0, Del_Kd_xyz, BWeyl_xz, BWeyl_zz, BWeyl_yz, d_Del_Kd_xyy_o0_t1_m0_l0, m_Del_Kd_xyy_o0_t6_l0, m_Del_Kd_xyy_o0_t5_l0, m_Del_Kd_xyy_o0_t4_l0, m_Del_Kd_xyy_o0_t3_l0, m_Del_Kd_xyy_o0_t2_l0, m_Del_Kd_xyy_o0_t0_l0, Del_Kd_xyy, d_Del_Kd_xxz_o0_t1_m0_l0, m_Del_Kd_xxz_o0_t10_l0, m_Del_Kd_xxz_o0_t9_l0, m_Del_Kd_xxz_o0_t8_l0, m_Del_Kd_xxz_o0_t7_l0, m_Del_Kd_xxz_o0_t6_l0, m_Del_Kd_xxz_o0_t5_l0, m_Del_Kd_xxz_o0_t4_l0, m_Del_Kd_xxz_o0_t3_l0, m_Del_Kd_xxz_o0_t2_l0, m_Del_Kd_xxz_o0_t0_l0, Del_Kd_xxz, BWeyl_xx, BWeyl_zx, BWeyl_yx, m_Rpd_xz_o0_t25_l0, m_Rpd_xz_o0_t24_l0, m_Rpd_xz_o0_t23_l0, m_Rpd_xz_o0_t22_l0, m_Rpd_xz_o0_t21_l0, m_Rpd_xz_o0_t20_l0, m_Rpd_xz_o0_t19_l0, m_Rpd_xz_o0_t18_l0, m_Rpd_xz_o0_t17_l0, m_Rpd_xz_o0_t16_l0, m_Rpd_xz_o0_t15_l0, m_Rpd_xz_o0_t14_l0, m_Rpd_xz_o0_t13_l0, m_Rpd_xz_o0_t12_l0, m_Rpd_xz_o0_t11_l0, m_Rpd_xz_o0_t10_l0, m_Rpd_xz_o0_t9_l0, m_Rpd_xz_o0_t8_l0, m_Rpd_xz_o0_t7_l0, m_Rpd_xz_o0_t6_l0, m_Rpd_xz_o0_t5_l0, m_Rpd_xz_o0_t4_l0, m_Rpd_xz_o0_t3_l0, m_Rpd_xz_o0_t2_l0, m_Rpd_xz_o0_t1_l0, m_Rpd_xz_o0_t0_l0, Rpd_xz, EWeyl_xz, m_Rpd_xy_o0_t25_l0, m_Rpd_xy_o0_t24_l0, m_Rpd_xy_o0_t23_l0, m_Rpd_xy_o0_t22_l0, m_Rpd_xy_o0_t21_l0, m_Rpd_xy_o0_t20_l0, m_Rpd_xy_o0_t19_l0, m_Rpd_xy_o0_t18_l0, m_Rpd_xy_o0_t17_l0, m_Rpd_xy_o0_t16_l0, m_Rpd_xy_o0_t15_l0, m_Rpd_xy_o0_t14_l0, m_Rpd_xy_o0_t13_l0, m_Rpd_xy_o0_t12_l0, m_Rpd_xy_o0_t11_l0, m_Rpd_xy_o0_t10_l0, m_Rpd_xy_o0_t9_l0, m_Rpd_xy_o0_t8_l0, m_Rpd_xy_o0_t7_l0, m_Rpd_xy_o0_t6_l0, m_Rpd_xy_o0_t5_l0, m_Rpd_xy_o0_t4_l0, m_Rpd_xy_o0_t3_l0, m_Rpd_xy_o0_t2_l0, m_Rpd_xy_o0_t1_l0, m_Rpd_xy_o0_t0_l0, Rpd_xy, m_Rpd_zz_o0_t22_l0, m_Rpd_zz_o0_t21_l0, m_Rpd_zz_o0_t20_l0, m_Rpd_zz_o0_t19_l0, m_Rpd_zz_o0_t18_l0, m_Rpd_zz_o0_t17_l0, m_Rpd_zz_o0_t16_l0, m_Rpd_zz_o0_t15_l0, m_Rpd_zz_o0_t14_l0, m_Rpd_zz_o0_t13_l0, m_Rpd_zz_o0_t12_l0, m_Rpd_zz_o0_t11_l0, m_Rpd_zz_o0_t10_l0, m_Rpd_zz_o0_t9_l0, m_Rpd_zz_o0_t8_l0, m_Rpd_zz_o0_t7_l0, m_Rpd_zz_o0_t6_l0, m_Rpd_zz_o0_t5_l0, m_Rpd_zz_o0_t4_l0, m_Rpd_zz_o0_t3_l0, m_Rpd_zz_o0_t2_l0, m_Rpd_zz_o0_t1_l0, m_Rpd_zz_o0_t0_l0, Rpd_zz, EWeyl_zz, d_Del_Kd_yyz_o0_t1_m0_l0, m_Del_Kd_yyz_o0_t10_l0, m_Del_Kd_yyz_o0_t9_l0, m_Del_Kd_yyz_o0_t8_l0, m_Del_Kd_yyz_o0_t7_l0, m_Del_Kd_yyz_o0_t6_l0, m_Del_Kd_yyz_o0_t5_l0, m_Del_Kd_yyz_o0_t4_l0, m_Del_Kd_yyz_o0_t3_l0, m_Del_Kd_yyz_o0_t2_l0, m_Del_Kd_yyz_o0_t0_l0, Del_Kd_yyz, d_Del_Kd_yxy_o0_t1_m0_l0, m_Del_Kd_yxy_o0_t10_l0, m_Del_Kd_yxy_o0_t9_l0, m_Del_Kd_yxy_o0_t8_l0, m_Del_Kd_yxy_o0_t7_l0, m_Del_Kd_yxy_o0_t6_l0, m_Del_Kd_yxy_o0_t5_l0, m_Del_Kd_yxy_o0_t4_l0, m_Del_Kd_yxy_o0_t3_l0, m_Del_Kd_yxy_o0_t2_l0, m_Del_Kd_yxy_o0_t0_l0, Del_Kd_yxy, BWeyl_xy, BWeyl_zy, BWeyl_yy, m_Rpd_yy_o0_t22_l0, m_Rpd_yy_o0_t21_l0, m_Rpd_yy_o0_t20_l0, m_Rpd_yy_o0_t19_l0, m_Rpd_yy_o0_t18_l0, m_Rpd_yy_o0_t17_l0, m_Rpd_yy_o0_t16_l0, m_Rpd_yy_o0_t15_l0, m_Rpd_yy_o0_t14_l0, m_Rpd_yy_o0_t13_l0, m_Rpd_yy_o0_t12_l0, m_Rpd_yy_o0_t11_l0, m_Rpd_yy_o0_t10_l0, m_Rpd_yy_o0_t9_l0, m_Rpd_yy_o0_t8_l0, m_Rpd_yy_o0_t7_l0, m_Rpd_yy_o0_t6_l0, m_Rpd_yy_o0_t5_l0, m_Rpd_yy_o0_t4_l0, m_Rpd_yy_o0_t3_l0, m_Rpd_yy_o0_t2_l0, m_Rpd_yy_o0_t1_l0, m_Rpd_yy_o0_t0_l0, Rpd_yy, d_Rtd_xy_o0_t12_m0_l0, m_Rtd_xy_o0_t11_l0, m_Rtd_xy_o0_t10_l0, m_Rtd_xy_o0_t9_l0, m_Rtd_xy_o0_t8_l0, m_Rtd_xy_o0_t7_l0, m_Rtd_xy_o0_t6_l0, m_Rtd_xy_o0_t5_l0, m_Rtd_xy_o0_t4_l0, m_Rtd_xy_o0_t3_l0, m_Rtd_xy_o0_t2_l0, m_Rtd_xy_o0_t1_l0, m_Rtd_xy_o0_t0_l0, Rtd_xy, EWeyl_xy, d_Rtd_yy_o0_t9_m0_l0, m_Rtd_yy_o0_t8_l0, m_Rtd_yy_o0_t7_l0, m_Rtd_yy_o0_t6_l0, m_Rtd_yy_o0_t5_l0, m_Rtd_yy_o0_t4_l0, m_Rtd_yy_o0_t3_l0, m_Rtd_yy_o0_t2_l0, m_Rtd_yy_o0_t1_l0, m_Rtd_yy_o0_t0_l0, Rtd_yy, EWeyl_yy, d_Rtd_yz_o0_t12_m0_l0, m_Rtd_yz_o0_t11_l0, m_Rtd_yz_o0_t10_l0, m_Rtd_yz_o0_t9_l0, m_Rtd_yz_o0_t8_l0, m_Rtd_yz_o0_t7_l0, m_Rtd_yz_o0_t6_l0, m_Rtd_yz_o0_t5_l0, m_Rtd_yz_o0_t4_l0, m_Rtd_yz_o0_t3_l0, m_Rtd_yz_o0_t2_l0, m_Rtd_yz_o0_t1_l0, m_Rtd_yz_o0_t0_l0, Rtd_yz, EWeyl_yz, Bfvf, Sfuu_zz, Tf4u_zz, Tu_zz, Sfuu_yz, Tf4u_yz, Tu_yz, Sfuu_yy, Tf4u_yy, Tu_yy, Sfuu_xy, Tf4u_xy, Tu_xy, Sfuu_xx, Tf4u_xx, Tu_xx, Sfuu_xz, Tf4u_xz, Tu_xz, Tu_tt, Td_yz, Td_yy, Td_xz, Td_xy, Td_xx, Td_zz, TT, Td_tx, Td_tz, Td_ty, Jtd_ADM_z, Td_tt, Jtd_ADM_y, Jtd_ADM_x, rho_ADM, d_Rscalar_o0_t0_m0_l0, d_HamCon_o0_t0_m0_l0, d_MomCon_x_o0_t13_m0_l0, d_MomCon_y_o0_t13_m0_l0, d_MomCon_z_o0_t13_m0_l0, d_trA_o0_t0_m0_l0, d_detgtm1_o0_t0_m0_l0, d_M_ADM_surf_o0_t9_m0_l0, d_Jz_ADM_surf_o0_t0_m0_l0, d_M_Komar_o0_t0_m0_l0, d_Jz_Komar_o0_t0_m0_l0, d_psi4R_o0_t0_m0_l0, d_psi4I_o0_t0_m0_l0, d_Z_x_o0_t0_m0_l0, d_Z_y_o0_t0_m0_l0, d_Z_z_o0_t0_m0_l0, d_flux_mass_o0_t0_m0_l0, m_M_ADM_surf_o0_t8_l0, m_M_ADM_surf_o0_t7_l0, m_M_ADM_surf_o0_t6_l0, m_M_ADM_surf_o0_t5_l0, m_M_ADM_surf_o0_t4_l0, m_M_ADM_surf_o0_t3_l0, m_M_ADM_surf_o0_t2_l0, m_M_ADM_surf_o0_t1_l0, m_M_ADM_surf_o0_t0_l0, m_MomCon_z_o0_t12_l0, m_MomCon_z_o0_t11_l0, m_MomCon_z_o0_t10_l0, m_MomCon_z_o0_t9_l0, m_MomCon_z_o0_t8_l0, m_MomCon_z_o0_t7_l0, m_MomCon_z_o0_t6_l0, m_MomCon_z_o0_t5_l0, m_MomCon_z_o0_t4_l0, m_MomCon_z_o0_t3_l0, m_MomCon_z_o0_t2_l0, m_MomCon_z_o0_t1_l0, m_MomCon_z_o0_t0_l0, m_MomCon_y_o0_t12_l0, m_MomCon_y_o0_t11_l0, m_MomCon_y_o0_t10_l0, m_MomCon_y_o0_t9_l0, m_MomCon_y_o0_t8_l0, m_MomCon_y_o0_t7_l0, m_MomCon_y_o0_t6_l0, m_MomCon_y_o0_t5_l0, m_MomCon_y_o0_t4_l0, m_MomCon_y_o0_t3_l0, m_MomCon_y_o0_t2_l0, m_MomCon_y_o0_t1_l0, m_MomCon_y_o0_t0_l0, m_MomCon_x_o0_t12_l0, m_MomCon_x_o0_t11_l0, m_MomCon_x_o0_t10_l0, m_MomCon_x_o0_t9_l0, m_MomCon_x_o0_t8_l0, m_MomCon_x_o0_t7_l0, m_MomCon_x_o0_t6_l0, m_MomCon_x_o0_t5_l0, m_MomCon_x_o0_t4_l0, m_MomCon_x_o0_t3_l0, m_MomCon_x_o0_t2_l0, m_MomCon_x_o0_t1_l0, m_MomCon_x_o0_t0_l0;
	
			//Get the dimensions of the patch
			hier::Box pbox = patch->getBox();
			const hier::Index boxfirst = patch->getBox().lower();
			const hier::Index boxlast = patch->getBox().upper();
	
			//Get delta spaces into an array. dx, dy, dz.
			std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
			const double* dx  = patch_geom->getDx();
	
			//Auxiliary definitions
			int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
			int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
			int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
						if ((i + 2 < ilast && i - 2 >= 0 && j + 2 < jlast && j - 2 >= 0 && k + 2 < klast && k - 2 >= 0)) {
							d_Rpd_xx_o0_t0_m0_l0 = D1CDO4_i(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rpd_xx_o0_t1_m0_l0 = D1CDO4_j(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rpd_xx_o0_t2_m0_l0 = D1CDO4_k(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rpd_xx_o0_t3_m0_l0 = D2CDO4_i(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rpd_xx_o0_t4_m0_l0 = D1CDO4crossed_ij(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rpd_xx_o0_t5_m0_l0 = D1CDO4crossed_ik(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rpd_xx_o0_t6_m0_l0 = D2CDO4_j(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rpd_xx_o0_t7_m0_l0 = D1CDO4crossed_jk(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rpd_xx_o0_t8_m0_l0 = D2CDO4_k(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_xxx_o0_t0_m0_l0 = D1CDO4_i(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_xxy_o0_t0_m0_l0 = D1CDO4_j(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_xxz_o0_t0_m0_l0 = D1CDO4_k(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xx_o0_t0_m0_l0 = D2CDO4_i(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xx_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xx_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xx_o0_t3_m0_l0 = D2CDO4_j(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xx_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xx_o0_t5_m0_l0 = D2CDO4_k(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xx_o0_t6_m0_l0 = D1CDO4_i(Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xy_o0_t6_m0_l0 = D1CDO4_j(Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xz_o0_t6_m0_l0 = D1CDO4_k(Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xx_o0_t7_m0_l0 = D1CDO4_i(Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xy_o0_t7_m0_l0 = D1CDO4_j(Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xz_o0_t7_m0_l0 = D1CDO4_k(Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xx_o0_t8_m0_l0 = D1CDO4_i(Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xy_o0_t8_m0_l0 = D1CDO4_j(Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xz_o0_t8_m0_l0 = D1CDO4_k(Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_yxx_o0_t0_m0_l0 = D1CDO4_i(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_xyy_o0_t0_m0_l0 = D1CDO4_j(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_xyz_o0_t1_m0_l0 = D1CDO4_k(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xy_o0_t0_m0_l0 = D2CDO4_i(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xy_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xy_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xy_o0_t3_m0_l0 = D2CDO4_j(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xy_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xy_o0_t5_m0_l0 = D2CDO4_k(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_zxx_o0_t0_m0_l0 = D1CDO4_i(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_xyz_o0_t0_m0_l0 = D1CDO4_j(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_xzz_o0_t0_m0_l0 = D1CDO4_k(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xz_o0_t0_m0_l0 = D2CDO4_i(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xz_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xz_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xz_o0_t3_m0_l0 = D2CDO4_j(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xz_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_xz_o0_t5_m0_l0 = D2CDO4_k(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_xyy_o0_t1_m0_l0 = D1CDO4_i(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_yyy_o0_t0_m0_l0 = D1CDO4_j(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_yyz_o0_t0_m0_l0 = D1CDO4_k(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_yy_o0_t0_m0_l0 = D2CDO4_i(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_yy_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_yy_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_yy_o0_t3_m0_l0 = D2CDO4_j(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_yy_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_yy_o0_t5_m0_l0 = D2CDO4_k(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_xyz_o0_t2_m0_l0 = D1CDO4_i(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_zyy_o0_t0_m0_l0 = D1CDO4_j(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_yzz_o0_t0_m0_l0 = D1CDO4_k(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_yz_o0_t0_m0_l0 = D2CDO4_i(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_yz_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_yz_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_yz_o0_t3_m0_l0 = D2CDO4_j(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_yz_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_yz_o0_t5_m0_l0 = D2CDO4_k(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_xzz_o0_t1_m0_l0 = D1CDO4_i(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_yzz_o0_t1_m0_l0 = D1CDO4_j(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Ctd_zzz_o0_t0_m0_l0 = D1CDO4_k(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_zz_o0_t0_m0_l0 = D2CDO4_i(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_zz_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_zz_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_zz_o0_t3_m0_l0 = D2CDO4_j(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_zz_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Rtd_zz_o0_t5_m0_l0 = D2CDO4_k(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_xxy_o0_t0_m0_l0 = D1CDO4_i(Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_yxy_o0_t0_m0_l0 = D1CDO4_j(Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_zxy_o0_t0_m0_l0 = D1CDO4_k(Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_xxy_o0_t2_m0_l0 = D1CDO4_i(trK, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_yxx_o0_t2_m0_l0 = D1CDO4_j(trK, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_zxx_o0_t2_m0_l0 = D1CDO4_k(trK, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_xxz_o0_t0_m0_l0 = D1CDO4_i(Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_yxz_o0_t0_m0_l0 = D1CDO4_j(Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_zxz_o0_t0_m0_l0 = D1CDO4_k(Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_MomCon_x_o0_t0_m0_l0 = D1CDO4_i(Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_yxx_o0_t0_m0_l0 = D1CDO4_j(Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_zxx_o0_t0_m0_l0 = D1CDO4_k(Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_xyy_o0_t0_m0_l0 = D1CDO4_i(Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_MomCon_y_o0_t4_m0_l0 = D1CDO4_j(Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_zyy_o0_t0_m0_l0 = D1CDO4_k(Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_xyz_o0_t0_m0_l0 = D1CDO4_i(Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_yyz_o0_t0_m0_l0 = D1CDO4_j(Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_zyz_o0_t0_m0_l0 = D1CDO4_k(Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_xzz_o0_t0_m0_l0 = D1CDO4_i(Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Del_Kd_yzz_o0_t0_m0_l0 = D1CDO4_j(Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_MomCon_z_o0_t11_m0_l0 = D1CDO4_k(Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							inv_chi = 1.0 / fabs(vector(chi, i, j, k));
							g4d_zz = inv_chi * vector(gtd_zz, i, j, k);
							g4d_yz = inv_chi * vector(gtd_yz, i, j, k);
							g4d_yy = inv_chi * vector(gtd_yy, i, j, k);
							Bfd_y = inv_chi * vector(gtd_xy, i, j, k) * vector(Bfu_x, i, j, k) + inv_chi * vector(gtd_yy, i, j, k) * vector(Bfu_y, i, j, k) + inv_chi * vector(gtd_yz, i, j, k) * vector(Bfu_z, i, j, k);
							Bfd_z = inv_chi * vector(gtd_xz, i, j, k) * vector(Bfu_x, i, j, k) + inv_chi * vector(gtd_yz, i, j, k) * vector(Bfu_y, i, j, k) + inv_chi * vector(gtd_zz, i, j, k) * vector(Bfu_z, i, j, k);
							Bfd_x = inv_chi * vector(gtd_xx, i, j, k) * vector(Bfu_x, i, j, k) + inv_chi * vector(gtd_xy, i, j, k) * vector(Bfu_y, i, j, k) + inv_chi * vector(gtd_xz, i, j, k) * vector(Bfu_z, i, j, k);
							sqBf = vector(Bfu_x, i, j, k) * Bfd_x + vector(Bfu_y, i, j, k) * Bfd_y + vector(Bfu_z, i, j, k) * Bfd_z;
							r_pos = MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)));
							ur_z = zcoord(k);
							ur_y = ycoord(j);
							uph_z = 0.0;
							uph_x = -ycoord(j);
							uph_y = xcoord(i);
							ur_x = xcoord(i);
							uthd_y = ycoord(j) * zcoord(k);
							uthd_x = xcoord(i) * zcoord(k);
							uthd_z = (-xcoord(i) * xcoord(i)) + (-ycoord(j) * ycoord(j));
							sdetg = pow(vector(chi, i, j, k), (-1.500000000000000));
							Tf4u_tt = (vector(Df, i, j, k) + vector(tauf, i, j, k)) / (sdetg * (vector(Alpha, i, j, k) * vector(Alpha, i, j, k)));
							m_Ctd_zzz_o0_t0_l0 = 0.5 * d_Ctd_zzz_o0_t0_m0_l0;
							Ctd_zzz = m_Ctd_zzz_o0_t0_l0;
							hf = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
							wrr = inv_chi * vector(gtd_xx, i, j, k) * (ur_x * ur_x) + 2.0 * inv_chi * vector(gtd_xy, i, j, k) * ur_x * ur_y + 2.0 * inv_chi * vector(gtd_xz, i, j, k) * ur_x * ur_z + inv_chi * vector(gtd_yy, i, j, k) * (ur_y * ur_y) + 2.0 * inv_chi * vector(gtd_yz, i, j, k) * ur_y * ur_z + inv_chi * vector(gtd_zz, i, j, k) * (ur_z * ur_z);
							m_Ctd_yyz_o0_t0_l0 = 0.5 * d_Ctd_yyz_o0_t0_m0_l0;
							Ctd_yyz = m_Ctd_yyz_o0_t0_l0;
							m_Ctd_zyy_o0_t1_l0 = (-0.5) * d_Ctd_yyz_o0_t0_m0_l0;
							m_Ctd_zyy_o0_t0_l0 = 1.0 * d_Ctd_zyy_o0_t0_m0_l0;
							Ctd_zyy = m_Ctd_zyy_o0_t0_l0 + m_Ctd_zyy_o0_t1_l0;
							m_Ctd_yyy_o0_t0_l0 = 0.5 * d_Ctd_yyy_o0_t0_m0_l0;
							Ctd_yyy = m_Ctd_yyy_o0_t0_l0;
							m_Ctd_xxz_o0_t0_l0 = 0.5 * d_Ctd_xxz_o0_t0_m0_l0;
							Ctd_xxz = m_Ctd_xxz_o0_t0_l0;
							m_Ctd_zxx_o0_t1_l0 = (-0.5) * d_Ctd_xxz_o0_t0_m0_l0;
							m_Ctd_zxx_o0_t0_l0 = 1.0 * d_Ctd_zxx_o0_t0_m0_l0;
							Ctd_zxx = m_Ctd_zxx_o0_t0_l0 + m_Ctd_zxx_o0_t1_l0;
							m_Ctd_xxx_o0_t0_l0 = 0.5 * d_Ctd_xxx_o0_t0_m0_l0;
							Ctd_xxx = m_Ctd_xxx_o0_t0_l0;
							m_Ctd_yzz_o0_t1_l0 = (-0.5) * d_Ctd_yzz_o0_t1_m0_l0;
							m_Ctd_yzz_o0_t0_l0 = 1.0 * d_Ctd_yzz_o0_t0_m0_l0;
							Ctd_yzz = m_Ctd_yzz_o0_t0_l0 + m_Ctd_yzz_o0_t1_l0;
							m_Ctd_zyz_o0_t0_l0 = 0.5 * d_Ctd_yzz_o0_t1_m0_l0;
							Ctd_zyz = m_Ctd_zyz_o0_t0_l0;
							m_Ctd_xxy_o0_t0_l0 = 0.5 * d_Ctd_xxy_o0_t0_m0_l0;
							Ctd_xxy = m_Ctd_xxy_o0_t0_l0;
							m_Ctd_yxx_o0_t1_l0 = (-0.5) * d_Ctd_xxy_o0_t0_m0_l0;
							m_Ctd_yxx_o0_t0_l0 = 1.0 * d_Ctd_yxx_o0_t0_m0_l0;
							Ctd_yxx = m_Ctd_yxx_o0_t0_l0 + m_Ctd_yxx_o0_t1_l0;
							m_Ctd_xyy_o0_t1_l0 = (-0.5) * d_Ctd_xyy_o0_t1_m0_l0;
							m_Ctd_xyy_o0_t0_l0 = 1.0 * d_Ctd_xyy_o0_t0_m0_l0;
							Ctd_xyy = m_Ctd_xyy_o0_t0_l0 + m_Ctd_xyy_o0_t1_l0;
							m_Ctd_yxy_o0_t0_l0 = 0.5 * d_Ctd_xyy_o0_t1_m0_l0;
							Ctd_yxy = m_Ctd_yxy_o0_t0_l0;
							m_Ctd_xyz_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t2_m0_l0;
							m_Ctd_xyz_o0_t1_l0 = 0.5 * d_Ctd_xyz_o0_t1_m0_l0;
							m_Ctd_xyz_o0_t0_l0 = 0.5 * d_Ctd_xyz_o0_t0_m0_l0;
							Ctd_xyz = (m_Ctd_xyz_o0_t0_l0 + m_Ctd_xyz_o0_t1_l0) + m_Ctd_xyz_o0_t2_l0;
							m_Ctd_yxz_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t0_m0_l0;
							m_Ctd_yxz_o0_t0_l0 = 0.5 * d_Ctd_xyz_o0_t2_m0_l0;
							Ctd_yxz = (m_Ctd_yxz_o0_t0_l0 + m_Ctd_xyz_o0_t1_l0) + m_Ctd_yxz_o0_t2_l0;
							m_Ctd_zxy_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t1_m0_l0;
							Ctd_zxy = (m_Ctd_yxz_o0_t0_l0 + m_Ctd_xyz_o0_t0_l0) + m_Ctd_zxy_o0_t2_l0;
							g4d_tt = (-vector(Alpha, i, j, k) * vector(Alpha, i, j, k)) + inv_chi * (vector(gtd_xx, i, j, k) * (vector(Betau_x, i, j, k) * vector(Betau_x, i, j, k)) + 2.0 * vector(gtd_xy, i, j, k) * vector(Betau_y, i, j, k) * vector(Betau_x, i, j, k) + 2.0 * vector(gtd_xz, i, j, k) * vector(Betau_z, i, j, k) * vector(Betau_x, i, j, k) + vector(gtd_yy, i, j, k) * (vector(Betau_y, i, j, k) * vector(Betau_y, i, j, k)) + 2.0 * vector(gtd_yz, i, j, k) * vector(Betau_z, i, j, k) * vector(Betau_y, i, j, k) + vector(gtd_zz, i, j, k) * (vector(Betau_z, i, j, k) * vector(Betau_z, i, j, k)));
							g4d_ty = inv_chi * (vector(gtd_xy, i, j, k) * vector(Betau_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Betau_y, i, j, k) + vector(gtd_yz, i, j, k) * vector(Betau_z, i, j, k));
							g4d_tx = inv_chi * (vector(gtd_xx, i, j, k) * vector(Betau_x, i, j, k) + vector(gtd_xy, i, j, k) * vector(Betau_y, i, j, k) + vector(gtd_xz, i, j, k) * vector(Betau_z, i, j, k));
							g4d_tz = inv_chi * (vector(gtd_xz, i, j, k) * vector(Betau_x, i, j, k) + vector(gtd_yz, i, j, k) * vector(Betau_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Betau_z, i, j, k));
							m_Ctd_xzz_o0_t1_l0 = (-0.5) * d_Ctd_xzz_o0_t1_m0_l0;
							m_Ctd_xzz_o0_t0_l0 = 1.0 * d_Ctd_xzz_o0_t0_m0_l0;
							Ctd_xzz = m_Ctd_xzz_o0_t0_l0 + m_Ctd_xzz_o0_t1_l0;
							m_Ctd_zxz_o0_t0_l0 = 0.5 * d_Ctd_xzz_o0_t1_m0_l0;
							Ctd_zxz = m_Ctd_zxz_o0_t0_l0;
							dSigma_y = (1.0 * ur_y * MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))) / (pow(vector(chi, i, j, k), 1.500000000000000));
							dSigma_x = (1.0 * ur_x * MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))) / (pow(vector(chi, i, j, k), 1.500000000000000));
							dSigma_z = (1.0 * ur_z * MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))) / (pow(vector(chi, i, j, k), 1.500000000000000));
							detgtd = vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) * vector(gtd_xx, i, j, k) + (-vector(gtd_yy, i, j, k) * (vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k))) + (-vector(gtd_zz, i, j, k) * (vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k))) + (-(vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k)) * vector(gtd_xx, i, j, k)) + 2.0 * vector(gtd_yz, i, j, k) * vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k);
							idetgtd = 1.0 / detgtd;
							gtu_zz = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_yy, i, j, k) - vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k));
							gtu_yy = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k));
							gtu_yz = idetgtd * ((-vector(gtd_xx, i, j, k) * vector(gtd_yz, i, j, k)) + vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k));
							gtu_xx = idetgtd * (vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k));
							gtu_xz = idetgtd * (vector(gtd_xy, i, j, k) * vector(gtd_yz, i, j, k) - vector(gtd_yy, i, j, k) * vector(gtd_xz, i, j, k));
							uth_z = vector(chi, i, j, k) * gtu_xz * uthd_x + vector(chi, i, j, k) * gtu_yz * uthd_y + vector(chi, i, j, k) * gtu_zz * uthd_z;
							Atud_zx = gtu_xz * vector(Atd_xx, i, j, k) + gtu_yz * vector(Atd_xy, i, j, k) + gtu_zz * vector(Atd_xz, i, j, k);
							Atud_zz = gtu_xz * vector(Atd_xz, i, j, k) + gtu_yz * vector(Atd_yz, i, j, k) + gtu_zz * vector(Atd_zz, i, j, k);
							Atud_zy = gtu_xz * vector(Atd_xy, i, j, k) + gtu_yz * vector(Atd_yy, i, j, k) + gtu_zz * vector(Atd_yz, i, j, k);
							Atu_zz = Atud_zx * gtu_xz + Atud_zy * gtu_yz + Atud_zz * gtu_zz;
							Sfu_z = vector(chi, i, j, k) * gtu_xz * vector(Sfd_x, i, j, k) + vector(chi, i, j, k) * gtu_yz * vector(Sfd_y, i, j, k) + vector(chi, i, j, k) * gtu_zz * vector(Sfd_z, i, j, k);
							Tf4u_tz = Sfu_z / (sdetg * vector(Alpha, i, j, k)) + (-vector(Betau_z, i, j, k) * Tf4u_tt);
							Ct_zxz = gtu_xz * Ctd_xxz + gtu_yz * Ctd_yxz + gtu_zz * Ctd_zxz;
							Ct_zxy = gtu_xz * Ctd_xxy + gtu_yz * Ctd_yxy + gtu_zz * Ctd_zxy;
							Ct_zxx = gtu_xz * Ctd_xxx + gtu_yz * Ctd_yxx + gtu_zz * Ctd_zxx;
							Ct_zyz = gtu_xz * Ctd_xyz + gtu_yz * Ctd_yyz + gtu_zz * Ctd_zyz;
							Ct_zyy = gtu_xz * Ctd_xyy + gtu_yz * Ctd_yyy + gtu_zz * Ctd_zyy;
							Ct_zzz = gtu_xz * Ctd_xzz + gtu_yz * Ctd_yzz + gtu_zz * Ctd_zzz;
							gtu_xy = idetgtd * ((-vector(gtd_xy, i, j, k) * vector(gtd_zz, i, j, k)) + vector(gtd_yz, i, j, k) * vector(gtd_xz, i, j, k));
							Ct_xzz = gtu_xx * Ctd_xzz + gtu_xy * Ctd_yzz + gtu_xz * Ctd_zzz;
							uth_x = vector(chi, i, j, k) * gtu_xx * uthd_x + vector(chi, i, j, k) * gtu_xy * uthd_y + vector(chi, i, j, k) * gtu_xz * uthd_z;
							uth_y = vector(chi, i, j, k) * gtu_xy * uthd_x + vector(chi, i, j, k) * gtu_yy * uthd_y + vector(chi, i, j, k) * gtu_yz * uthd_z;
							Gamt_z = gtu_xx * Ct_zxx + 2.0 * gtu_xy * Ct_zxy + 2.0 * gtu_xz * Ct_zxz + gtu_yy * Ct_zyy + 2.0 * gtu_yz * Ct_zyz + gtu_zz * Ct_zzz;
							Zu_z = 0.5 * vector(chi, i, j, k) * (vector(Gamh_z, i, j, k) - Gamt_z);
							Atud_yy = gtu_xy * vector(Atd_xy, i, j, k) + gtu_yy * vector(Atd_yy, i, j, k) + gtu_yz * vector(Atd_yz, i, j, k);
							Atud_yx = gtu_xy * vector(Atd_xx, i, j, k) + gtu_yy * vector(Atd_xy, i, j, k) + gtu_yz * vector(Atd_xz, i, j, k);
							Atud_yz = gtu_xy * vector(Atd_xz, i, j, k) + gtu_yy * vector(Atd_yz, i, j, k) + gtu_yz * vector(Atd_zz, i, j, k);
							Atu_yz = Atud_yx * gtu_xz + Atud_yy * gtu_yz + Atud_yz * gtu_zz;
							Atud_xx = gtu_xx * vector(Atd_xx, i, j, k) + gtu_xy * vector(Atd_xy, i, j, k) + gtu_xz * vector(Atd_xz, i, j, k);
							Atud_xz = gtu_xx * vector(Atd_xz, i, j, k) + gtu_xy * vector(Atd_yz, i, j, k) + gtu_xz * vector(Atd_zz, i, j, k);
							Atud_xy = gtu_xx * vector(Atd_xy, i, j, k) + gtu_xy * vector(Atd_yy, i, j, k) + gtu_xz * vector(Atd_yz, i, j, k);
							Atu_xz = Atud_xx * gtu_xz + Atud_xy * gtu_yz + Atud_xz * gtu_zz;
							Sfu_y = vector(chi, i, j, k) * gtu_xy * vector(Sfd_x, i, j, k) + vector(chi, i, j, k) * gtu_yy * vector(Sfd_y, i, j, k) + vector(chi, i, j, k) * gtu_yz * vector(Sfd_z, i, j, k);
							Tf4u_ty = Sfu_y / (sdetg * vector(Alpha, i, j, k)) + (-vector(Betau_y, i, j, k) * Tf4u_tt);
							Sfu_x = vector(chi, i, j, k) * gtu_xx * vector(Sfd_x, i, j, k) + vector(chi, i, j, k) * gtu_xy * vector(Sfd_y, i, j, k) + vector(chi, i, j, k) * gtu_xz * vector(Sfd_z, i, j, k);
							Tf4u_tx = Sfu_x / (sdetg * vector(Alpha, i, j, k)) + (-vector(Betau_x, i, j, k) * Tf4u_tt);
							vfu_y = vector(chi, i, j, k) * gtu_xy * vector(vfd_x, i, j, k) + vector(chi, i, j, k) * gtu_yy * vector(vfd_y, i, j, k) + vector(chi, i, j, k) * gtu_yz * vector(vfd_z, i, j, k);
							Ju_y = vector(Df, i, j, k) * (vector(Alpha, i, j, k) * vfu_y - vector(Betau_y, i, j, k));
							vfu_x = vector(chi, i, j, k) * gtu_xx * vector(vfd_x, i, j, k) + vector(chi, i, j, k) * gtu_xy * vector(vfd_y, i, j, k) + vector(chi, i, j, k) * gtu_xz * vector(vfd_z, i, j, k);
							Ju_x = vector(Df, i, j, k) * (vector(Alpha, i, j, k) * vfu_x - vector(Betau_x, i, j, k));
							Atu_yy = Atud_yx * gtu_xy + Atud_yy * gtu_yy + Atud_yz * gtu_yz;
							Ct_yxx = gtu_xy * Ctd_xxx + gtu_yy * Ctd_yxx + gtu_yz * Ctd_zxx;
							Ct_yxz = gtu_xy * Ctd_xxz + gtu_yy * Ctd_yxz + gtu_yz * Ctd_zxz;
							Ct_yxy = gtu_xy * Ctd_xxy + gtu_yy * Ctd_yxy + gtu_yz * Ctd_zxy;
							Atu_xy = Atud_xx * gtu_xy + Atud_xy * gtu_yy + Atud_xz * gtu_yz;
							Atu_xx = Atud_xx * gtu_xx + Atud_xy * gtu_xy + Atud_xz * gtu_xz;
							Ct_yyy = gtu_xy * Ctd_xyy + gtu_yy * Ctd_yyy + gtu_yz * Ctd_zyy;
							Ct_yyz = gtu_xy * Ctd_xyz + gtu_yy * Ctd_yyz + gtu_yz * Ctd_zyz;
							Tu_ty = Tf4u_ty;
							Tu_tx = Tf4u_tx;
							Tu_tz = Tf4u_tz;
							Ct_yzz = gtu_xy * Ctd_xzz + gtu_yy * Ctd_yzz + gtu_yz * Ctd_zzz;
							Gamt_y = gtu_xx * Ct_yxx + 2.0 * gtu_xy * Ct_yxy + 2.0 * gtu_xz * Ct_yxz + gtu_yy * Ct_yyy + 2.0 * gtu_yz * Ct_yyz + gtu_zz * Ct_yzz;
							Zu_y = 0.5 * vector(chi, i, j, k) * (vector(Gamh_y, i, j, k) - Gamt_y);
							Ct_xxy = gtu_xx * Ctd_xxy + gtu_xy * Ctd_yxy + gtu_xz * Ctd_zxy;
							vfu_z = vector(chi, i, j, k) * gtu_xz * vector(vfd_x, i, j, k) + vector(chi, i, j, k) * gtu_yz * vector(vfd_y, i, j, k) + vector(chi, i, j, k) * gtu_zz * vector(vfd_z, i, j, k);
							sqvf = vfu_x * vector(vfd_x, i, j, k) + vfu_y * vector(vfd_y, i, j, k) + vfu_z * vector(vfd_z, i, j, k);
							sqW = 1.0 / (1.0 - sqvf);
							Ju_z = vector(Df, i, j, k) * (vector(Alpha, i, j, k) * vfu_z - vector(Betau_z, i, j, k));
							Ct_xxx = gtu_xx * Ctd_xxx + gtu_xy * Ctd_yxx + gtu_xz * Ctd_zxx;
							Ct_xxz = gtu_xx * Ctd_xxz + gtu_xy * Ctd_yxz + gtu_xz * Ctd_zxz;
							d_Rtd_xx_o0_t9_m0_l0 = gtu_zz * (3.0 * Ct_xxz * Ctd_xxz + 2.0 * Ct_yxz * Ctd_xyz + Ct_yxz * Ctd_yxz + 2.0 * Ct_zxz * Ctd_xzz + Ct_zxz * Ctd_zxz) + gtu_yz * (3.0 * Ct_xxy * Ctd_xxz + 2.0 * Ct_yxy * Ctd_xyz + Ct_yxy * Ctd_yxz + 2.0 * Ct_zxy * Ctd_xzz + Ct_zxy * Ctd_zxz) + gtu_xz * (3.0 * Ct_xxx * Ctd_xxz + 2.0 * Ct_yxx * Ctd_xyz + Ct_yxx * Ctd_yxz + 2.0 * Ct_zxx * Ctd_xzz + Ct_zxx * Ctd_zxz) + gtu_yz * (3.0 * Ct_xxz * Ctd_xxy + 2.0 * Ct_yxz * Ctd_xyy + Ct_yxz * Ctd_yxy + 2.0 * Ct_zxz * Ctd_xyz + Ct_zxz * Ctd_zxy) + gtu_yy * (3.0 * Ct_xxy * Ctd_xxy + 2.0 * Ct_yxy * Ctd_xyy + Ct_yxy * Ctd_yxy + 2.0 * Ct_zxy * Ctd_xyz + Ct_zxy * Ctd_zxy) + gtu_xy * (3.0 * Ct_xxx * Ctd_xxy + 2.0 * Ct_yxx * Ctd_xyy + Ct_yxx * Ctd_yxy + 2.0 * Ct_zxx * Ctd_xyz + Ct_zxx * Ctd_zxy) + gtu_xz * (3.0 * Ct_xxz * Ctd_xxx + 2.0 * Ct_yxz * Ctd_xxy + Ct_yxz * Ctd_yxx + 2.0 * Ct_zxz * Ctd_xxz + Ct_zxz * Ctd_zxx) + gtu_xy * (3.0 * Ct_xxy * Ctd_xxx + 2.0 * Ct_yxy * Ctd_xxy + Ct_yxy * Ctd_yxx + 2.0 * Ct_zxy * Ctd_xxz + Ct_zxy * Ctd_zxx) + gtu_xx * (3.0 * Ct_xxx * Ctd_xxx + 2.0 * Ct_yxx * Ctd_xxy + Ct_yxx * Ctd_yxx + 2.0 * Ct_zxx * Ctd_xxz + Ct_zxx * Ctd_zxx) + 1.0 * vector(Gamh_z, i, j, k) * Ctd_xxz + 1.0 * vector(Gamh_y, i, j, k) * Ctd_xxy + 1.0 * vector(Gamh_x, i, j, k) * Ctd_xxx;
							m_Rtd_xx_o0_t8_l0 = 1.0 * vector(gtd_xz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
							m_Rtd_xx_o0_t7_l0 = 1.0 * vector(gtd_xy, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
							m_Rtd_xx_o0_t6_l0 = 1.0 * vector(gtd_xx, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
							m_Rtd_xx_o0_t5_l0 = (-0.5 * gtu_zz) * d_Rtd_xx_o0_t5_m0_l0;
							m_Rtd_xx_o0_t4_l0 = (-1.0 * gtu_yz) * d_Rtd_xx_o0_t4_m0_l0;
							m_Rtd_xx_o0_t3_l0 = (-0.5 * gtu_yy) * d_Rtd_xx_o0_t3_m0_l0;
							m_Rtd_xx_o0_t2_l0 = (-1.0 * gtu_xz) * d_Rtd_xx_o0_t2_m0_l0;
							m_Rtd_xx_o0_t1_l0 = (-1.0 * gtu_xy) * d_Rtd_xx_o0_t1_m0_l0;
							m_Rtd_xx_o0_t0_l0 = (-0.5 * gtu_xx) * d_Rtd_xx_o0_t0_m0_l0;
							Rtd_xx = ((((((((m_Rtd_xx_o0_t0_l0 + m_Rtd_xx_o0_t1_l0) + m_Rtd_xx_o0_t2_l0) + m_Rtd_xx_o0_t3_l0) + m_Rtd_xx_o0_t4_l0) + m_Rtd_xx_o0_t5_l0) + m_Rtd_xx_o0_t6_l0) + m_Rtd_xx_o0_t7_l0) + m_Rtd_xx_o0_t8_l0) + d_Rtd_xx_o0_t9_m0_l0;
							g4d_xy = inv_chi * vector(gtd_xy, i, j, k);
							wphph = inv_chi * vector(gtd_xx, i, j, k) * (uph_x * uph_x) + 2.0 * inv_chi * vector(gtd_xy, i, j, k) * uph_x * uph_y + 2.0 * inv_chi * vector(gtd_xz, i, j, k) * uph_x * uph_z + inv_chi * vector(gtd_yy, i, j, k) * (uph_y * uph_y) + 2.0 * inv_chi * vector(gtd_yz, i, j, k) * uph_y * uph_z + inv_chi * vector(gtd_zz, i, j, k) * (uph_z * uph_z);
							vph_y = uph_y / MAX(0.0001, sqrt(wphph));
							vph_z = uph_z / MAX(0.0001, sqrt(wphph));
							vph_x = uph_x / MAX(0.0001, sqrt(wphph));
							wthph = inv_chi * vector(gtd_xx, i, j, k) * uth_x * vph_x + inv_chi * vector(gtd_xy, i, j, k) * uth_x * vph_y + inv_chi * vector(gtd_xz, i, j, k) * uth_x * vph_z + inv_chi * vector(gtd_xy, i, j, k) * uth_y * vph_x + inv_chi * vector(gtd_yy, i, j, k) * uth_y * vph_y + inv_chi * vector(gtd_yz, i, j, k) * uth_y * vph_z + inv_chi * vector(gtd_xz, i, j, k) * uth_z * vph_x + inv_chi * vector(gtd_yz, i, j, k) * uth_z * vph_y + inv_chi * vector(gtd_zz, i, j, k) * uth_z * vph_z;
							g4d_xx = inv_chi * vector(gtd_xx, i, j, k);
							wrph = inv_chi * vector(gtd_xx, i, j, k) * ur_x * vph_x + inv_chi * vector(gtd_xy, i, j, k) * ur_x * vph_y + inv_chi * vector(gtd_xz, i, j, k) * ur_x * vph_z + inv_chi * vector(gtd_xy, i, j, k) * ur_y * vph_x + inv_chi * vector(gtd_yy, i, j, k) * ur_y * vph_y + inv_chi * vector(gtd_yz, i, j, k) * ur_y * vph_z + inv_chi * vector(gtd_xz, i, j, k) * ur_z * vph_x + inv_chi * vector(gtd_yz, i, j, k) * ur_z * vph_y + inv_chi * vector(gtd_zz, i, j, k) * ur_z * vph_z;
							vr_x = ((-wrph * vph_x) + ur_x) / MAX(0.0001, sqrt(wrr));
							vr_y = ((-wrph * vph_y) + ur_y) / MAX(0.0001, sqrt(wrr));
							vr_z = ((-wrph * vph_z) + ur_z) / MAX(0.0001, sqrt(wrr));
							wthr = inv_chi * vector(gtd_xx, i, j, k) * uth_x * vr_x + inv_chi * vector(gtd_xy, i, j, k) * uth_x * vr_y + inv_chi * vector(gtd_xz, i, j, k) * uth_x * vr_z + inv_chi * vector(gtd_xy, i, j, k) * uth_y * vr_x + inv_chi * vector(gtd_yy, i, j, k) * uth_y * vr_y + inv_chi * vector(gtd_yz, i, j, k) * uth_y * vr_z + inv_chi * vector(gtd_xz, i, j, k) * uth_z * vr_x + inv_chi * vector(gtd_yz, i, j, k) * uth_z * vr_y + inv_chi * vector(gtd_zz, i, j, k) * uth_z * vr_z;
							g4d_xz = inv_chi * vector(gtd_xz, i, j, k);
							wthth = inv_chi * vector(gtd_xx, i, j, k) * (uth_x * uth_x) + 2.0 * inv_chi * vector(gtd_xy, i, j, k) * uth_x * uth_y + 2.0 * inv_chi * vector(gtd_xz, i, j, k) * uth_x * uth_z + inv_chi * vector(gtd_yy, i, j, k) * (uth_y * uth_y) + 2.0 * inv_chi * vector(gtd_yz, i, j, k) * uth_y * uth_z + inv_chi * vector(gtd_zz, i, j, k) * (uth_z * uth_z);
							vth_x = ((-wthph * vph_x) + (-wthr * vr_x) + uth_x) / MAX(0.0001, sqrt(wthth));
							mmR_xx = (-0.5 * (vph_x * vph_x)) + 0.5 * (vth_x * vth_x);
							mmI_xx = (-1.0) * vph_x * vth_x;
							vth_y = ((-wthph * vph_y) + (-wthr * vr_y) + uth_y) / MAX(0.0001, sqrt(wthth));
							mmR_yy = (-0.5 * (vph_y * vph_y)) + 0.5 * (vth_y * vth_y);
							mmR_xy = (-0.5 * vph_x * vph_y) + 0.5 * vth_x * vth_y;
							mmI_xy = (-0.5 * vph_x * vth_y) + (-0.5 * vph_y * vth_x);
							mmI_yy = (-1.0) * vph_y * vth_y;
							vth_z = ((-wthph * vph_z) + (-wthr * vr_z) + uth_z) / MAX(0.0001, sqrt(wthth));
							mmR_yz = (-0.5 * vph_y * vph_z) + 0.5 * vth_y * vth_z;
							mmR_zz = (-0.5 * (vph_z * vph_z)) + 0.5 * (vth_z * vth_z);
							mmR_xz = (-0.5 * vph_x * vph_z) + 0.5 * vth_x * vth_z;
							mmI_zz = (-1.0) * vph_z * vth_z;
							mmI_xz = (-0.5 * vph_x * vth_z) + (-0.5 * vph_z * vth_x);
							mmI_yz = (-0.5 * vph_y * vth_z) + (-0.5 * vph_z * vth_y);
							Ct_xyz = gtu_xx * Ctd_xyz + gtu_xy * Ctd_yyz + gtu_xz * Ctd_zyz;
							d_Rtd_xz_o0_t12_m0_l0 = gtu_zz * (Ct_xxz * Ctd_zxz + Ct_xxz * Ctd_xzz + Ct_xzz * Ctd_xxz + Ct_yxz * Ctd_zyz + Ct_yxz * Ctd_yzz + Ct_yzz * Ctd_xyz + 2.0 * Ct_zxz * Ctd_zzz + Ct_zzz * Ctd_xzz) + gtu_yz * (Ct_xxy * Ctd_zxz + Ct_xxy * Ctd_xzz + Ct_xyz * Ctd_xxz + Ct_yxy * Ctd_zyz + Ct_yxy * Ctd_yzz + Ct_yyz * Ctd_xyz + 2.0 * Ct_zxy * Ctd_zzz + Ct_zyz * Ctd_xzz) + gtu_xz * (Ct_xxx * Ctd_zxz + Ct_xxx * Ctd_xzz + Ct_xxz * Ctd_xxz + Ct_yxx * Ctd_zyz + Ct_yxx * Ctd_yzz + Ct_yxz * Ctd_xyz + 2.0 * Ct_zxx * Ctd_zzz + Ct_zxz * Ctd_xzz) + gtu_yz * (Ct_xxz * Ctd_zxy + Ct_xxz * Ctd_xyz + Ct_xzz * Ctd_xxy + Ct_yxz * Ctd_zyy + Ct_yxz * Ctd_yyz + Ct_yzz * Ctd_xyy + 2.0 * Ct_zxz * Ctd_zyz + Ct_zzz * Ctd_xyz) + gtu_yy * (Ct_xxy * Ctd_zxy + Ct_xxy * Ctd_xyz + Ct_xyz * Ctd_xxy + Ct_yxy * Ctd_zyy + Ct_yxy * Ctd_yyz + Ct_yyz * Ctd_xyy + 2.0 * Ct_zxy * Ctd_zyz + Ct_zyz * Ctd_xyz) + gtu_xy * (Ct_xxx * Ctd_zxy + Ct_xxx * Ctd_xyz + Ct_xxz * Ctd_xxy + Ct_yxx * Ctd_zyy + Ct_yxx * Ctd_yyz + Ct_yxz * Ctd_xyy + 2.0 * Ct_zxx * Ctd_zyz + Ct_zxz * Ctd_xyz) + gtu_xz * (Ct_xxz * Ctd_zxx + Ct_xxz * Ctd_xxz + Ct_xzz * Ctd_xxx + Ct_yxz * Ctd_zxy + Ct_yxz * Ctd_yxz + Ct_yzz * Ctd_xxy + 2.0 * Ct_zxz * Ctd_zxz + Ct_zzz * Ctd_xxz) + gtu_xy * (Ct_xxy * Ctd_zxx + Ct_xxy * Ctd_xxz + Ct_xyz * Ctd_xxx + Ct_yxy * Ctd_zxy + Ct_yxy * Ctd_yxz + Ct_yyz * Ctd_xxy + 2.0 * Ct_zxy * Ctd_zxz + Ct_zyz * Ctd_xxz) + gtu_xx * (Ct_xxx * Ctd_zxx + Ct_xxx * Ctd_xxz + Ct_xxz * Ctd_xxx + Ct_yxx * Ctd_zxy + Ct_yxx * Ctd_yxz + Ct_yxz * Ctd_xxy + 2.0 * Ct_zxx * Ctd_zxz + Ct_zxz * Ctd_xxz) + 0.5 * vector(Gamh_z, i, j, k) * (Ctd_xzz + Ctd_zxz) + 0.5 * vector(Gamh_y, i, j, k) * (Ctd_xyz + Ctd_zxy) + 0.5 * vector(Gamh_x, i, j, k) * (Ctd_xxz + Ctd_zxx);
							m_Rtd_xz_o0_t11_l0 = 0.5 * vector(gtd_zz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
							m_Rtd_xz_o0_t10_l0 = 0.5 * vector(gtd_yz, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
							m_Rtd_xz_o0_t9_l0 = 0.5 * vector(gtd_xz, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
							m_Rtd_xz_o0_t8_l0 = 0.5 * vector(gtd_xz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
							m_Rtd_xz_o0_t7_l0 = 0.5 * vector(gtd_xy, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
							m_Rtd_xz_o0_t6_l0 = 0.5 * vector(gtd_xx, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
							m_Rtd_xz_o0_t5_l0 = (-0.5 * gtu_zz) * d_Rtd_xz_o0_t5_m0_l0;
							m_Rtd_xz_o0_t4_l0 = (-1.0 * gtu_yz) * d_Rtd_xz_o0_t4_m0_l0;
							m_Rtd_xz_o0_t3_l0 = (-0.5 * gtu_yy) * d_Rtd_xz_o0_t3_m0_l0;
							m_Rtd_xz_o0_t2_l0 = (-1.0 * gtu_xz) * d_Rtd_xz_o0_t2_m0_l0;
							m_Rtd_xz_o0_t1_l0 = (-1.0 * gtu_xy) * d_Rtd_xz_o0_t1_m0_l0;
							m_Rtd_xz_o0_t0_l0 = (-0.5 * gtu_xx) * d_Rtd_xz_o0_t0_m0_l0;
							Rtd_xz = (((((((((((m_Rtd_xz_o0_t0_l0 + m_Rtd_xz_o0_t1_l0) + m_Rtd_xz_o0_t2_l0) + m_Rtd_xz_o0_t3_l0) + m_Rtd_xz_o0_t4_l0) + m_Rtd_xz_o0_t5_l0) + m_Rtd_xz_o0_t6_l0) + m_Rtd_xz_o0_t7_l0) + m_Rtd_xz_o0_t8_l0) + m_Rtd_xz_o0_t9_l0) + m_Rtd_xz_o0_t10_l0) + m_Rtd_xz_o0_t11_l0) + d_Rtd_xz_o0_t12_m0_l0;
							d_Rtd_zz_o0_t9_m0_l0 = gtu_zz * (2.0 * Ct_xzz * Ctd_zxz + Ct_xzz * Ctd_xzz + 2.0 * Ct_yzz * Ctd_zyz + Ct_yzz * Ctd_yzz + 3.0 * Ct_zzz * Ctd_zzz) + gtu_yz * (2.0 * Ct_xyz * Ctd_zxz + Ct_xyz * Ctd_xzz + 2.0 * Ct_yyz * Ctd_zyz + Ct_yyz * Ctd_yzz + 3.0 * Ct_zyz * Ctd_zzz) + gtu_xz * (2.0 * Ct_xxz * Ctd_zxz + Ct_xxz * Ctd_xzz + 2.0 * Ct_yxz * Ctd_zyz + Ct_yxz * Ctd_yzz + 3.0 * Ct_zxz * Ctd_zzz) + gtu_yz * (2.0 * Ct_xzz * Ctd_zxy + Ct_xzz * Ctd_xyz + 2.0 * Ct_yzz * Ctd_zyy + Ct_yzz * Ctd_yyz + 3.0 * Ct_zzz * Ctd_zyz) + gtu_yy * (2.0 * Ct_xyz * Ctd_zxy + Ct_xyz * Ctd_xyz + 2.0 * Ct_yyz * Ctd_zyy + Ct_yyz * Ctd_yyz + 3.0 * Ct_zyz * Ctd_zyz) + gtu_xy * (2.0 * Ct_xxz * Ctd_zxy + Ct_xxz * Ctd_xyz + 2.0 * Ct_yxz * Ctd_zyy + Ct_yxz * Ctd_yyz + 3.0 * Ct_zxz * Ctd_zyz) + gtu_xz * (2.0 * Ct_xzz * Ctd_zxx + Ct_xzz * Ctd_xxz + 2.0 * Ct_yzz * Ctd_zxy + Ct_yzz * Ctd_yxz + 3.0 * Ct_zzz * Ctd_zxz) + gtu_xy * (2.0 * Ct_xyz * Ctd_zxx + Ct_xyz * Ctd_xxz + 2.0 * Ct_yyz * Ctd_zxy + Ct_yyz * Ctd_yxz + 3.0 * Ct_zyz * Ctd_zxz) + gtu_xx * (2.0 * Ct_xxz * Ctd_zxx + Ct_xxz * Ctd_xxz + 2.0 * Ct_yxz * Ctd_zxy + Ct_yxz * Ctd_yxz + 3.0 * Ct_zxz * Ctd_zxz) + 1.0 * vector(Gamh_z, i, j, k) * Ctd_zzz + 1.0 * vector(Gamh_y, i, j, k) * Ctd_zyz + 1.0 * vector(Gamh_x, i, j, k) * Ctd_zxz;
							m_Rtd_zz_o0_t8_l0 = 1.0 * vector(gtd_zz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
							m_Rtd_zz_o0_t7_l0 = 1.0 * vector(gtd_yz, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
							m_Rtd_zz_o0_t6_l0 = 1.0 * vector(gtd_xz, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
							m_Rtd_zz_o0_t5_l0 = (-0.5 * gtu_zz) * d_Rtd_zz_o0_t5_m0_l0;
							m_Rtd_zz_o0_t4_l0 = (-1.0 * gtu_yz) * d_Rtd_zz_o0_t4_m0_l0;
							m_Rtd_zz_o0_t3_l0 = (-0.5 * gtu_yy) * d_Rtd_zz_o0_t3_m0_l0;
							m_Rtd_zz_o0_t2_l0 = (-1.0 * gtu_xz) * d_Rtd_zz_o0_t2_m0_l0;
							m_Rtd_zz_o0_t1_l0 = (-1.0 * gtu_xy) * d_Rtd_zz_o0_t1_m0_l0;
							m_Rtd_zz_o0_t0_l0 = (-0.5 * gtu_xx) * d_Rtd_zz_o0_t0_m0_l0;
							Rtd_zz = ((((((((m_Rtd_zz_o0_t0_l0 + m_Rtd_zz_o0_t1_l0) + m_Rtd_zz_o0_t2_l0) + m_Rtd_zz_o0_t3_l0) + m_Rtd_zz_o0_t4_l0) + m_Rtd_zz_o0_t5_l0) + m_Rtd_zz_o0_t6_l0) + m_Rtd_zz_o0_t7_l0) + m_Rtd_zz_o0_t8_l0) + d_Rtd_zz_o0_t9_m0_l0;
							Ct_xyy = gtu_xx * Ctd_xyy + gtu_xy * Ctd_yyy + gtu_xz * Ctd_zyy;
							Gamt_x = gtu_xx * Ct_xxx + 2.0 * gtu_xy * Ct_xxy + 2.0 * gtu_xz * Ct_xxz + gtu_yy * Ct_xyy + 2.0 * gtu_yz * Ct_xyz + gtu_zz * Ct_xzz;
							Zu_x = 0.5 * vector(chi, i, j, k) * (vector(Gamh_x, i, j, k) - Gamt_x);
							m_Rpd_xx_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t3_m0_l0;
							m_Rpd_xx_o0_t21_l0 = (-0.75 * vector(gtd_xx, i, j, k) * gtu_zz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xx_o0_t20_l0 = (-0.75 * vector(gtd_xx, i, j, k) * gtu_yy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xx_o0_t19_l0 = (-0.75 * vector(gtd_xx, i, j, k) * gtu_xx * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xx_o0_t18_l0 = 2.0 * inv_chi * Zu_z * vector(gtd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xx_o0_t17_l0 = 2.0 * inv_chi * Zu_y * vector(gtd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xx_o0_t16_l0 = 2.0 * inv_chi * Zu_x * vector(gtd_xx, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xx_o0_t15_l0 = (-1.500000000000000 * vector(gtd_xx, i, j, k) * gtu_yz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xx_o0_t14_l0 = (-1.500000000000000 * vector(gtd_xx, i, j, k) * gtu_xz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xx_o0_t13_l0 = (-1.500000000000000 * vector(gtd_xx, i, j, k) * gtu_xy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xx_o0_t12_l0 = (-0.5 * Ct_xxx) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xx_o0_t11_l0 = (-0.25 * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xx_o0_t10_l0 = (-0.5 * Ct_zxx) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xx_o0_t9_l0 = (-0.5 * Ct_yxx) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xx_o0_t8_l0 = 0.5 * vector(gtd_xx, i, j, k) * gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
							m_Rpd_xx_o0_t7_l0 = 1.0 * vector(gtd_xx, i, j, k) * gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
							m_Rpd_xx_o0_t6_l0 = 0.5 * vector(gtd_xx, i, j, k) * gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
							m_Rpd_xx_o0_t5_l0 = 1.0 * vector(gtd_xx, i, j, k) * gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
							m_Rpd_xx_o0_t4_l0 = 1.0 * vector(gtd_xx, i, j, k) * gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
							m_Rpd_xx_o0_t3_l0 = 0.5 * vector(gtd_xx, i, j, k) * gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
							m_Rpd_xx_o0_t2_l0 = (-0.5 * vector(gtd_xx, i, j, k) * vector(Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xx_o0_t1_l0 = (-0.5 * vector(gtd_xx, i, j, k) * vector(Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xx_o0_t0_l0 = (-0.5 * vector(gtd_xx, i, j, k) * vector(Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
							Rpd_xx = (((((((((((((((((((((m_Rpd_xx_o0_t0_l0 + m_Rpd_xx_o0_t1_l0) + m_Rpd_xx_o0_t2_l0) + m_Rpd_xx_o0_t3_l0) + m_Rpd_xx_o0_t4_l0) + m_Rpd_xx_o0_t5_l0) + m_Rpd_xx_o0_t6_l0) + m_Rpd_xx_o0_t7_l0) + m_Rpd_xx_o0_t8_l0) + m_Rpd_xx_o0_t9_l0) + m_Rpd_xx_o0_t10_l0) + m_Rpd_xx_o0_t11_l0) + m_Rpd_xx_o0_t12_l0) + m_Rpd_xx_o0_t13_l0) + m_Rpd_xx_o0_t14_l0) + m_Rpd_xx_o0_t15_l0) + m_Rpd_xx_o0_t16_l0) + m_Rpd_xx_o0_t17_l0) + m_Rpd_xx_o0_t18_l0) + m_Rpd_xx_o0_t19_l0) + m_Rpd_xx_o0_t20_l0) + m_Rpd_xx_o0_t21_l0) + m_Rpd_xx_o0_t22_l0;
							d_Del_Kd_yxx_o0_t1_m0_l0 = -inv_chi * (2.0 * vector(Atd_xx, i, j, k) * Ct_xxy + 2.0 * vector(Atd_xy, i, j, k) * Ct_yxy + 2.0 * vector(Atd_xz, i, j, k) * Ct_zxy);
							m_Del_Kd_yxx_o0_t6_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_yxx_o0_t5_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_yxx_o0_t4_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_yxx_o0_t3_l0 = 1.0 * (inv_chi * inv_chi) * vector(Atd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_yxx_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xx, i, j, k) * d_Del_Kd_yxx_o0_t2_m0_l0;
							m_Del_Kd_yxx_o0_t0_l0 = inv_chi * d_Del_Kd_yxx_o0_t0_m0_l0;
							Del_Kd_yxx = (((((m_Del_Kd_yxx_o0_t0_l0 + d_Del_Kd_yxx_o0_t1_m0_l0) + m_Del_Kd_yxx_o0_t2_l0) + m_Del_Kd_yxx_o0_t3_l0) + m_Del_Kd_yxx_o0_t4_l0) + m_Del_Kd_yxx_o0_t5_l0) + m_Del_Kd_yxx_o0_t6_l0;
							d_Del_Kd_yxz_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xz, i, j, k) * Ct_xxy + vector(Atd_xx, i, j, k) * Ct_xyz + vector(Atd_yz, i, j, k) * Ct_yxy + vector(Atd_xy, i, j, k) * Ct_yyz + vector(Atd_zz, i, j, k) * Ct_zxy + vector(Atd_xz, i, j, k) * Ct_zyz);
							m_Del_Kd_yxz_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_yxz_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_yxz_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_yxz_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_yxz_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_yxz_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_yxz_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_yxz_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_yz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_yxz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xz, i, j, k) * d_Del_Kd_yxx_o0_t2_m0_l0;
							m_Del_Kd_yxz_o0_t0_l0 = inv_chi * d_Del_Kd_yxz_o0_t0_m0_l0;
							Del_Kd_yxz = (((((((((m_Del_Kd_yxz_o0_t0_l0 + d_Del_Kd_yxz_o0_t1_m0_l0) + m_Del_Kd_yxz_o0_t2_l0) + m_Del_Kd_yxz_o0_t3_l0) + m_Del_Kd_yxz_o0_t4_l0) + m_Del_Kd_yxz_o0_t5_l0) + m_Del_Kd_yxz_o0_t6_l0) + m_Del_Kd_yxz_o0_t7_l0) + m_Del_Kd_yxz_o0_t8_l0) + m_Del_Kd_yxz_o0_t9_l0) + m_Del_Kd_yxz_o0_t10_l0;
							EWeyl_xx = (-inv_chi * Rpd_xx) + (-Rtd_xx) + inv_chi * (vector(Atd_xx, i, j, k) * Atud_xx + vector(Atd_xy, i, j, k) * Atud_yx + vector(Atd_xz, i, j, k) * Atud_zx) + (-0.3333333333333333 * inv_chi * vector(trK, i, j, k) * vector(Atd_xx, i, j, k)) + (-0.2222222222222222 * inv_chi * vector(gtd_xx, i, j, k) * (vector(trK, i, j, k) * vector(trK, i, j, k)));
							d_Del_Kd_zxx_o0_t1_m0_l0 = -inv_chi * (2.0 * vector(Atd_xx, i, j, k) * Ct_xxz + 2.0 * vector(Atd_xy, i, j, k) * Ct_yxz + 2.0 * vector(Atd_xz, i, j, k) * Ct_zxz);
							m_Del_Kd_zxx_o0_t6_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_zxx_o0_t5_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_zxx_o0_t4_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_zxx_o0_t3_l0 = 1.0 * (inv_chi * inv_chi) * vector(Atd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_zxx_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xx, i, j, k) * d_Del_Kd_zxx_o0_t2_m0_l0;
							m_Del_Kd_zxx_o0_t0_l0 = inv_chi * d_Del_Kd_zxx_o0_t0_m0_l0;
							Del_Kd_zxx = (((((m_Del_Kd_zxx_o0_t0_l0 + d_Del_Kd_zxx_o0_t1_m0_l0) + m_Del_Kd_zxx_o0_t2_l0) + m_Del_Kd_zxx_o0_t3_l0) + m_Del_Kd_zxx_o0_t4_l0) + m_Del_Kd_zxx_o0_t5_l0) + m_Del_Kd_zxx_o0_t6_l0;
							d_Del_Kd_zyz_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xz, i, j, k) * Ct_xyz + vector(Atd_xy, i, j, k) * Ct_xzz + vector(Atd_yz, i, j, k) * Ct_yyz + vector(Atd_yy, i, j, k) * Ct_yzz + vector(Atd_zz, i, j, k) * Ct_zyz + vector(Atd_yz, i, j, k) * Ct_zzz);
							m_Del_Kd_zyz_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_zz, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_zyz_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_zz, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_zyz_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_zz, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_zyz_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_zyz_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_zyz_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_zyz_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_zyz_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_zz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_zyz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_yz, i, j, k) * d_Del_Kd_zxx_o0_t2_m0_l0;
							m_Del_Kd_zyz_o0_t0_l0 = inv_chi * d_Del_Kd_zyz_o0_t0_m0_l0;
							Del_Kd_zyz = (((((((((m_Del_Kd_zyz_o0_t0_l0 + d_Del_Kd_zyz_o0_t1_m0_l0) + m_Del_Kd_zyz_o0_t2_l0) + m_Del_Kd_zyz_o0_t3_l0) + m_Del_Kd_zyz_o0_t4_l0) + m_Del_Kd_zyz_o0_t5_l0) + m_Del_Kd_zyz_o0_t6_l0) + m_Del_Kd_zyz_o0_t7_l0) + m_Del_Kd_zyz_o0_t8_l0) + m_Del_Kd_zyz_o0_t9_l0) + m_Del_Kd_zyz_o0_t10_l0;
							d_Del_Kd_zyy_o0_t1_m0_l0 = -inv_chi * (2.0 * vector(Atd_xy, i, j, k) * Ct_xyz + 2.0 * vector(Atd_yy, i, j, k) * Ct_yyz + 2.0 * vector(Atd_yz, i, j, k) * Ct_zyz);
							m_Del_Kd_zyy_o0_t6_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_zyy_o0_t5_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_zyy_o0_t4_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_zyy_o0_t3_l0 = 1.0 * (inv_chi * inv_chi) * vector(Atd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_zyy_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_yy, i, j, k) * d_Del_Kd_zxx_o0_t2_m0_l0;
							m_Del_Kd_zyy_o0_t0_l0 = inv_chi * d_Del_Kd_zyy_o0_t0_m0_l0;
							Del_Kd_zyy = (((((m_Del_Kd_zyy_o0_t0_l0 + d_Del_Kd_zyy_o0_t1_m0_l0) + m_Del_Kd_zyy_o0_t2_l0) + m_Del_Kd_zyy_o0_t3_l0) + m_Del_Kd_zyy_o0_t4_l0) + m_Del_Kd_zyy_o0_t5_l0) + m_Del_Kd_zyy_o0_t6_l0;
							d_Del_Kd_zxy_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xy, i, j, k) * Ct_xxz + vector(Atd_xx, i, j, k) * Ct_xyz + vector(Atd_yy, i, j, k) * Ct_yxz + vector(Atd_xy, i, j, k) * Ct_yyz + vector(Atd_yz, i, j, k) * Ct_zxz + vector(Atd_xz, i, j, k) * Ct_zyz);
							m_Del_Kd_zxy_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_zxy_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_zxy_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_zxy_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_zxy_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xy, i, j, k) * d_Del_Kd_zxx_o0_t2_m0_l0;
							m_Del_Kd_zxy_o0_t0_l0 = inv_chi * d_Del_Kd_zxy_o0_t0_m0_l0;
							Del_Kd_zxy = (((((((((m_Del_Kd_zxy_o0_t0_l0 + d_Del_Kd_zxy_o0_t1_m0_l0) + m_Del_Kd_zxy_o0_t2_l0) + m_Del_Kd_yxz_o0_t3_l0) + m_Del_Kd_zxy_o0_t4_l0) + m_Del_Kd_zxy_o0_t5_l0) + m_Del_Kd_zxy_o0_t6_l0) + m_Del_Kd_zxy_o0_t7_l0) + m_Del_Kd_yxz_o0_t8_l0) + m_Del_Kd_yxz_o0_t9_l0) + m_Del_Kd_yxz_o0_t10_l0;
							d_Del_Kd_zxz_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xz, i, j, k) * Ct_xxz + vector(Atd_xx, i, j, k) * Ct_xzz + vector(Atd_yz, i, j, k) * Ct_yxz + vector(Atd_xy, i, j, k) * Ct_yzz + vector(Atd_zz, i, j, k) * Ct_zxz + vector(Atd_xz, i, j, k) * Ct_zzz);
							m_Del_Kd_zxz_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_zz, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_zxz_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_zz, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_zxz_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_zz, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_zxz_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_zxz_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_zxz_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_zxz_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_zxz_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_zz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_zxz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xz, i, j, k) * d_Del_Kd_zxx_o0_t2_m0_l0;
							m_Del_Kd_zxz_o0_t0_l0 = inv_chi * d_Del_Kd_zxz_o0_t0_m0_l0;
							Del_Kd_zxz = (((((((((m_Del_Kd_zxz_o0_t0_l0 + d_Del_Kd_zxz_o0_t1_m0_l0) + m_Del_Kd_zxz_o0_t2_l0) + m_Del_Kd_zxz_o0_t3_l0) + m_Del_Kd_zxz_o0_t4_l0) + m_Del_Kd_zxz_o0_t5_l0) + m_Del_Kd_zxz_o0_t6_l0) + m_Del_Kd_zxz_o0_t7_l0) + m_Del_Kd_zxz_o0_t8_l0) + m_Del_Kd_zxz_o0_t9_l0) + m_Del_Kd_zxz_o0_t10_l0;
							m_Rpd_yz_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t7_m0_l0;
							m_Rpd_yz_o0_t24_l0 = (-0.75 * vector(gtd_yz, i, j, k) * gtu_zz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_yz_o0_t23_l0 = (-0.75 * vector(gtd_yz, i, j, k) * gtu_yy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yz_o0_t22_l0 = (-0.75 * vector(gtd_yz, i, j, k) * gtu_xx * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_yz_o0_t21_l0 = inv_chi * Zu_z * vector(gtd_zz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yz_o0_t20_l0 = inv_chi * Zu_y * vector(gtd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yz_o0_t19_l0 = inv_chi * Zu_x * vector(gtd_xz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yz_o0_t18_l0 = inv_chi * Zu_z * vector(gtd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_yz_o0_t17_l0 = inv_chi * Zu_y * vector(gtd_yy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_yz_o0_t16_l0 = inv_chi * Zu_x * vector(gtd_xy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_yz_o0_t15_l0 = (-1.500000000000000 * vector(gtd_yz, i, j, k) * gtu_yz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yz_o0_t14_l0 = (-1.500000000000000 * vector(gtd_yz, i, j, k) * gtu_xz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_yz_o0_t13_l0 = (-1.500000000000000 * vector(gtd_yz, i, j, k) * gtu_xy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_yz_o0_t12_l0 = (-0.5 * Ct_zyz) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_yz_o0_t11_l0 = (-0.5 * Ct_yyz) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yz_o0_t10_l0 = (-0.5 * Ct_xyz) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_yz_o0_t9_l0 = (-0.25 * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_yz_o0_t8_l0 = 0.5 * vector(gtd_yz, i, j, k) * gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
							m_Rpd_yz_o0_t7_l0 = 1.0 * vector(gtd_yz, i, j, k) * gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
							m_Rpd_yz_o0_t6_l0 = 0.5 * vector(gtd_yz, i, j, k) * gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
							m_Rpd_yz_o0_t5_l0 = 1.0 * vector(gtd_yz, i, j, k) * gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
							m_Rpd_yz_o0_t4_l0 = 1.0 * vector(gtd_yz, i, j, k) * gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
							m_Rpd_yz_o0_t3_l0 = 0.5 * vector(gtd_yz, i, j, k) * gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
							m_Rpd_yz_o0_t2_l0 = (-0.5 * vector(gtd_yz, i, j, k) * vector(Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_yz_o0_t1_l0 = (-0.5 * vector(gtd_yz, i, j, k) * vector(Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yz_o0_t0_l0 = (-0.5 * vector(gtd_yz, i, j, k) * vector(Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
							Rpd_yz = ((((((((((((((((((((((((m_Rpd_yz_o0_t0_l0 + m_Rpd_yz_o0_t1_l0) + m_Rpd_yz_o0_t2_l0) + m_Rpd_yz_o0_t3_l0) + m_Rpd_yz_o0_t4_l0) + m_Rpd_yz_o0_t5_l0) + m_Rpd_yz_o0_t6_l0) + m_Rpd_yz_o0_t7_l0) + m_Rpd_yz_o0_t8_l0) + m_Rpd_yz_o0_t9_l0) + m_Rpd_yz_o0_t10_l0) + m_Rpd_yz_o0_t11_l0) + m_Rpd_yz_o0_t12_l0) + m_Rpd_yz_o0_t13_l0) + m_Rpd_yz_o0_t14_l0) + m_Rpd_yz_o0_t15_l0) + m_Rpd_yz_o0_t16_l0) + m_Rpd_yz_o0_t17_l0) + m_Rpd_yz_o0_t18_l0) + m_Rpd_yz_o0_t19_l0) + m_Rpd_yz_o0_t20_l0) + m_Rpd_yz_o0_t21_l0) + m_Rpd_yz_o0_t22_l0) + m_Rpd_yz_o0_t23_l0) + m_Rpd_yz_o0_t24_l0) + m_Rpd_yz_o0_t25_l0;
							d_Del_Kd_yzz_o0_t1_m0_l0 = -inv_chi * (2.0 * vector(Atd_xz, i, j, k) * Ct_xyz + 2.0 * vector(Atd_yz, i, j, k) * Ct_yyz + 2.0 * vector(Atd_zz, i, j, k) * Ct_zyz);
							m_Del_Kd_yzz_o0_t6_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_yzz_o0_t5_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_yzz_o0_t4_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_yzz_o0_t3_l0 = 1.0 * (inv_chi * inv_chi) * vector(Atd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_yzz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_zz, i, j, k) * d_Del_Kd_yxx_o0_t2_m0_l0;
							m_Del_Kd_yzz_o0_t0_l0 = inv_chi * d_Del_Kd_yzz_o0_t0_m0_l0;
							Del_Kd_yzz = (((((m_Del_Kd_yzz_o0_t0_l0 + d_Del_Kd_yzz_o0_t1_m0_l0) + m_Del_Kd_yzz_o0_t2_l0) + m_Del_Kd_yzz_o0_t3_l0) + m_Del_Kd_yzz_o0_t4_l0) + m_Del_Kd_yzz_o0_t5_l0) + m_Del_Kd_yzz_o0_t6_l0;
							d_Del_Kd_xxy_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xy, i, j, k) * Ct_xxx + vector(Atd_xx, i, j, k) * Ct_xxy + vector(Atd_yy, i, j, k) * Ct_yxx + vector(Atd_xy, i, j, k) * Ct_yxy + vector(Atd_yz, i, j, k) * Ct_zxx + vector(Atd_xz, i, j, k) * Ct_zxy);
							m_Del_Kd_xxy_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_xxy_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_xxy_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_xxy_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xx, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_xxy_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xx, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_xxy_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xx, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_xxy_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xx, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_xxy_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_xxy_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xy, i, j, k) * d_Del_Kd_xxy_o0_t2_m0_l0;
							m_Del_Kd_xxy_o0_t0_l0 = inv_chi * d_Del_Kd_xxy_o0_t0_m0_l0;
							Del_Kd_xxy = (((((((((m_Del_Kd_xxy_o0_t0_l0 + d_Del_Kd_xxy_o0_t1_m0_l0) + m_Del_Kd_xxy_o0_t2_l0) + m_Del_Kd_xxy_o0_t3_l0) + m_Del_Kd_xxy_o0_t4_l0) + m_Del_Kd_xxy_o0_t5_l0) + m_Del_Kd_xxy_o0_t6_l0) + m_Del_Kd_xxy_o0_t7_l0) + m_Del_Kd_xxy_o0_t8_l0) + m_Del_Kd_xxy_o0_t9_l0) + m_Del_Kd_xxy_o0_t10_l0;
							d_Del_Kd_xzz_o0_t1_m0_l0 = -inv_chi * (2.0 * vector(Atd_xz, i, j, k) * Ct_xxz + 2.0 * vector(Atd_yz, i, j, k) * Ct_yxz + 2.0 * vector(Atd_zz, i, j, k) * Ct_zxz);
							m_Del_Kd_xzz_o0_t6_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_xzz_o0_t5_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_xzz_o0_t4_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_xzz_o0_t3_l0 = 1.0 * (inv_chi * inv_chi) * vector(Atd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_xzz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_zz, i, j, k) * d_Del_Kd_xxy_o0_t2_m0_l0;
							m_Del_Kd_xzz_o0_t0_l0 = inv_chi * d_Del_Kd_xzz_o0_t0_m0_l0;
							Del_Kd_xzz = (((((m_Del_Kd_xzz_o0_t0_l0 + d_Del_Kd_xzz_o0_t1_m0_l0) + m_Del_Kd_xzz_o0_t2_l0) + m_Del_Kd_xzz_o0_t3_l0) + m_Del_Kd_xzz_o0_t4_l0) + m_Del_Kd_xzz_o0_t5_l0) + m_Del_Kd_xzz_o0_t6_l0;
							d_Del_Kd_xyz_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xz, i, j, k) * Ct_xxy + vector(Atd_xy, i, j, k) * Ct_xxz + vector(Atd_yz, i, j, k) * Ct_yxy + vector(Atd_yy, i, j, k) * Ct_yxz + vector(Atd_zz, i, j, k) * Ct_zxy + vector(Atd_yz, i, j, k) * Ct_zxz);
							m_Del_Kd_xyz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_yz, i, j, k) * d_Del_Kd_xxy_o0_t2_m0_l0;
							m_Del_Kd_xyz_o0_t0_l0 = inv_chi * d_Del_Kd_xyz_o0_t0_m0_l0;
							Del_Kd_xyz = (((((((((m_Del_Kd_xyz_o0_t0_l0 + d_Del_Kd_xyz_o0_t1_m0_l0) + m_Del_Kd_xyz_o0_t2_l0) + m_Del_Kd_zxy_o0_t4_l0) + m_Del_Kd_yxz_o0_t4_l0) + m_Del_Kd_yxz_o0_t5_l0) + m_Del_Kd_yxz_o0_t6_l0) + m_Del_Kd_yxz_o0_t7_l0) + m_Del_Kd_zxy_o0_t5_l0) + m_Del_Kd_zxy_o0_t6_l0) + m_Del_Kd_zxy_o0_t7_l0;
							BWeyl_xz = (-vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxz) + vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxz + vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyz + (-vector(gtd_xx, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zyz) + (-vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xzz) + vector(gtd_xx, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yzz;
							BWeyl_zz = (-vector(gtd_zz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxz) + vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxz + vector(gtd_zz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyz + (-vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zyz) + (-vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xzz) + vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yzz;
							BWeyl_yz = (-vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxz) + vector(gtd_yy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxz + vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyz + (-vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zyz) + (-vector(gtd_yy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xzz) + vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yzz;
							d_Del_Kd_xyy_o0_t1_m0_l0 = -inv_chi * (2.0 * vector(Atd_xy, i, j, k) * Ct_xxy + 2.0 * vector(Atd_yy, i, j, k) * Ct_yxy + 2.0 * vector(Atd_yz, i, j, k) * Ct_zxy);
							m_Del_Kd_xyy_o0_t6_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_xyy_o0_t5_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_xyy_o0_t4_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_xyy_o0_t3_l0 = 1.0 * (inv_chi * inv_chi) * vector(Atd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_xyy_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_yy, i, j, k) * d_Del_Kd_xxy_o0_t2_m0_l0;
							m_Del_Kd_xyy_o0_t0_l0 = inv_chi * d_Del_Kd_xyy_o0_t0_m0_l0;
							Del_Kd_xyy = (((((m_Del_Kd_xyy_o0_t0_l0 + d_Del_Kd_xyy_o0_t1_m0_l0) + m_Del_Kd_xyy_o0_t2_l0) + m_Del_Kd_xyy_o0_t3_l0) + m_Del_Kd_xyy_o0_t4_l0) + m_Del_Kd_xyy_o0_t5_l0) + m_Del_Kd_xyy_o0_t6_l0;
							d_Del_Kd_xxz_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xz, i, j, k) * Ct_xxx + vector(Atd_xx, i, j, k) * Ct_xxz + vector(Atd_yz, i, j, k) * Ct_yxx + vector(Atd_xy, i, j, k) * Ct_yxz + vector(Atd_zz, i, j, k) * Ct_zxx + vector(Atd_xz, i, j, k) * Ct_zxz);
							m_Del_Kd_xxz_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_xxz_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_xxz_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_xxz_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xx, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_xxz_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xx, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_xxz_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xx, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_xxz_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xx, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_xxz_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_xxz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xz, i, j, k) * d_Del_Kd_xxy_o0_t2_m0_l0;
							m_Del_Kd_xxz_o0_t0_l0 = inv_chi * d_Del_Kd_xxz_o0_t0_m0_l0;
							Del_Kd_xxz = (((((((((m_Del_Kd_xxz_o0_t0_l0 + d_Del_Kd_xxz_o0_t1_m0_l0) + m_Del_Kd_xxz_o0_t2_l0) + m_Del_Kd_xxz_o0_t3_l0) + m_Del_Kd_xxz_o0_t4_l0) + m_Del_Kd_xxz_o0_t5_l0) + m_Del_Kd_xxz_o0_t6_l0) + m_Del_Kd_xxz_o0_t7_l0) + m_Del_Kd_xxz_o0_t8_l0) + m_Del_Kd_xxz_o0_t9_l0) + m_Del_Kd_xxz_o0_t10_l0;
							BWeyl_xx = (-vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxx) + vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxx + vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xxy + (-vector(gtd_xx, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxy) + (-vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xxz) + vector(gtd_xx, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxz;
							BWeyl_zx = (-vector(gtd_zz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxx) + vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxx + vector(gtd_zz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xxy + (-vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxy) + (-vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xxz) + vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxz;
							BWeyl_yx = (-vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxx) + vector(gtd_yy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxx + vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xxy + (-vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxy) + (-vector(gtd_yy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xxz) + vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxz;
							m_Rpd_xz_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t5_m0_l0;
							m_Rpd_xz_o0_t24_l0 = (-0.75 * vector(gtd_xz, i, j, k) * gtu_zz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xz_o0_t23_l0 = (-0.75 * vector(gtd_xz, i, j, k) * gtu_yy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xz_o0_t22_l0 = (-0.75 * vector(gtd_xz, i, j, k) * gtu_xx * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xz_o0_t21_l0 = inv_chi * Zu_z * vector(gtd_zz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xz_o0_t20_l0 = inv_chi * Zu_y * vector(gtd_yz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xz_o0_t19_l0 = inv_chi * Zu_x * vector(gtd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xz_o0_t18_l0 = inv_chi * Zu_z * vector(gtd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xz_o0_t17_l0 = inv_chi * Zu_y * vector(gtd_xy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xz_o0_t16_l0 = inv_chi * Zu_x * vector(gtd_xx, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xz_o0_t15_l0 = (-1.500000000000000 * vector(gtd_xz, i, j, k) * gtu_yz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xz_o0_t14_l0 = (-1.500000000000000 * vector(gtd_xz, i, j, k) * gtu_xz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xz_o0_t13_l0 = (-1.500000000000000 * vector(gtd_xz, i, j, k) * gtu_xy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xz_o0_t12_l0 = (-0.25 * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xz_o0_t11_l0 = 0.5 * vector(gtd_xz, i, j, k) * gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
							m_Rpd_xz_o0_t10_l0 = 1.0 * vector(gtd_xz, i, j, k) * gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
							m_Rpd_xz_o0_t9_l0 = 0.5 * vector(gtd_xz, i, j, k) * gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
							m_Rpd_xz_o0_t8_l0 = 1.0 * vector(gtd_xz, i, j, k) * gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
							m_Rpd_xz_o0_t7_l0 = 1.0 * vector(gtd_xz, i, j, k) * gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
							m_Rpd_xz_o0_t6_l0 = 0.5 * vector(gtd_xz, i, j, k) * gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
							m_Rpd_xz_o0_t5_l0 = (-0.5 * vector(gtd_xz, i, j, k) * vector(Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xz_o0_t4_l0 = (-0.5 * vector(gtd_xz, i, j, k) * vector(Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xz_o0_t3_l0 = (-0.5 * vector(gtd_xz, i, j, k) * vector(Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xz_o0_t2_l0 = (-0.5 * Ct_zxz) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xz_o0_t1_l0 = (-0.5 * Ct_yxz) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xz_o0_t0_l0 = (-0.5 * Ct_xxz) * d_Rpd_xx_o0_t0_m0_l0;
							Rpd_xz = ((((((((((((((((((((((((m_Rpd_xz_o0_t0_l0 + m_Rpd_xz_o0_t1_l0) + m_Rpd_xz_o0_t2_l0) + m_Rpd_xz_o0_t3_l0) + m_Rpd_xz_o0_t4_l0) + m_Rpd_xz_o0_t5_l0) + m_Rpd_xz_o0_t6_l0) + m_Rpd_xz_o0_t7_l0) + m_Rpd_xz_o0_t8_l0) + m_Rpd_xz_o0_t9_l0) + m_Rpd_xz_o0_t10_l0) + m_Rpd_xz_o0_t11_l0) + m_Rpd_xz_o0_t12_l0) + m_Rpd_xz_o0_t13_l0) + m_Rpd_xz_o0_t14_l0) + m_Rpd_xz_o0_t15_l0) + m_Rpd_xz_o0_t16_l0) + m_Rpd_xz_o0_t17_l0) + m_Rpd_xz_o0_t18_l0) + m_Rpd_xz_o0_t19_l0) + m_Rpd_xz_o0_t20_l0) + m_Rpd_xz_o0_t21_l0) + m_Rpd_xz_o0_t22_l0) + m_Rpd_xz_o0_t23_l0) + m_Rpd_xz_o0_t24_l0) + m_Rpd_xz_o0_t25_l0;
							EWeyl_xz = (-inv_chi * Rpd_xz) + (-Rtd_xz) + inv_chi * (vector(Atd_xx, i, j, k) * Atud_xz + vector(Atd_xy, i, j, k) * Atud_yz + vector(Atd_xz, i, j, k) * Atud_zz) + (-0.3333333333333333 * inv_chi * vector(trK, i, j, k) * vector(Atd_xz, i, j, k)) + (-0.2222222222222222 * inv_chi * vector(gtd_xz, i, j, k) * (vector(trK, i, j, k) * vector(trK, i, j, k)));
							m_Rpd_xy_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t4_m0_l0;
							m_Rpd_xy_o0_t24_l0 = (-0.75 * vector(gtd_xy, i, j, k) * gtu_zz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xy_o0_t23_l0 = (-0.75 * vector(gtd_xy, i, j, k) * gtu_yy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xy_o0_t22_l0 = (-0.75 * vector(gtd_xy, i, j, k) * gtu_xx * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xy_o0_t21_l0 = inv_chi * Zu_z * vector(gtd_yz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xy_o0_t20_l0 = inv_chi * Zu_y * vector(gtd_yy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xy_o0_t19_l0 = inv_chi * Zu_x * vector(gtd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xy_o0_t18_l0 = inv_chi * Zu_z * vector(gtd_xz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xy_o0_t17_l0 = inv_chi * Zu_y * vector(gtd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xy_o0_t16_l0 = inv_chi * Zu_x * vector(gtd_xx, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xy_o0_t15_l0 = (-1.500000000000000 * vector(gtd_xy, i, j, k) * gtu_yz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xy_o0_t14_l0 = (-1.500000000000000 * vector(gtd_xy, i, j, k) * gtu_xz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xy_o0_t13_l0 = (-1.500000000000000 * vector(gtd_xy, i, j, k) * gtu_xy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xy_o0_t12_l0 = (-0.25 * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xy_o0_t11_l0 = 0.5 * vector(gtd_xy, i, j, k) * gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
							m_Rpd_xy_o0_t10_l0 = 1.0 * vector(gtd_xy, i, j, k) * gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
							m_Rpd_xy_o0_t9_l0 = 0.5 * vector(gtd_xy, i, j, k) * gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
							m_Rpd_xy_o0_t8_l0 = 1.0 * vector(gtd_xy, i, j, k) * gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
							m_Rpd_xy_o0_t7_l0 = 1.0 * vector(gtd_xy, i, j, k) * gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
							m_Rpd_xy_o0_t6_l0 = 0.5 * vector(gtd_xy, i, j, k) * gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
							m_Rpd_xy_o0_t5_l0 = (-0.5 * vector(gtd_xy, i, j, k) * vector(Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xy_o0_t4_l0 = (-0.5 * vector(gtd_xy, i, j, k) * vector(Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xy_o0_t3_l0 = (-0.5 * vector(gtd_xy, i, j, k) * vector(Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_xy_o0_t2_l0 = (-0.5 * Ct_zxy) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_xy_o0_t1_l0 = (-0.5 * Ct_yxy) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_xy_o0_t0_l0 = (-0.5 * Ct_xxy) * d_Rpd_xx_o0_t0_m0_l0;
							Rpd_xy = ((((((((((((((((((((((((m_Rpd_xy_o0_t0_l0 + m_Rpd_xy_o0_t1_l0) + m_Rpd_xy_o0_t2_l0) + m_Rpd_xy_o0_t3_l0) + m_Rpd_xy_o0_t4_l0) + m_Rpd_xy_o0_t5_l0) + m_Rpd_xy_o0_t6_l0) + m_Rpd_xy_o0_t7_l0) + m_Rpd_xy_o0_t8_l0) + m_Rpd_xy_o0_t9_l0) + m_Rpd_xy_o0_t10_l0) + m_Rpd_xy_o0_t11_l0) + m_Rpd_xy_o0_t12_l0) + m_Rpd_xy_o0_t13_l0) + m_Rpd_xy_o0_t14_l0) + m_Rpd_xy_o0_t15_l0) + m_Rpd_xy_o0_t16_l0) + m_Rpd_xy_o0_t17_l0) + m_Rpd_xy_o0_t18_l0) + m_Rpd_xy_o0_t19_l0) + m_Rpd_xy_o0_t20_l0) + m_Rpd_xy_o0_t21_l0) + m_Rpd_xy_o0_t22_l0) + m_Rpd_xy_o0_t23_l0) + m_Rpd_xy_o0_t24_l0) + m_Rpd_xy_o0_t25_l0;
							m_Rpd_zz_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t8_m0_l0;
							m_Rpd_zz_o0_t21_l0 = (-0.75 * vector(gtd_zz, i, j, k) * gtu_zz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_zz_o0_t20_l0 = (-0.75 * vector(gtd_zz, i, j, k) * gtu_yy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_zz_o0_t19_l0 = (-0.75 * vector(gtd_zz, i, j, k) * gtu_xx * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_zz_o0_t18_l0 = 2.0 * inv_chi * Zu_z * vector(gtd_zz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_zz_o0_t17_l0 = 2.0 * inv_chi * Zu_y * vector(gtd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_zz_o0_t16_l0 = 2.0 * inv_chi * Zu_x * vector(gtd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_zz_o0_t15_l0 = (-1.500000000000000 * vector(gtd_zz, i, j, k) * gtu_yz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_zz_o0_t14_l0 = (-1.500000000000000 * vector(gtd_zz, i, j, k) * gtu_xz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_zz_o0_t13_l0 = (-1.500000000000000 * vector(gtd_zz, i, j, k) * gtu_xy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_zz_o0_t12_l0 = (-0.25 * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_zz_o0_t11_l0 = (-0.5 * Ct_zzz) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_zz_o0_t10_l0 = (-0.5 * Ct_yzz) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_zz_o0_t9_l0 = (-0.5 * Ct_xzz) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_zz_o0_t8_l0 = 0.5 * vector(gtd_zz, i, j, k) * gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
							m_Rpd_zz_o0_t7_l0 = 1.0 * vector(gtd_zz, i, j, k) * gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
							m_Rpd_zz_o0_t6_l0 = 0.5 * vector(gtd_zz, i, j, k) * gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
							m_Rpd_zz_o0_t5_l0 = 1.0 * vector(gtd_zz, i, j, k) * gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
							m_Rpd_zz_o0_t4_l0 = 1.0 * vector(gtd_zz, i, j, k) * gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
							m_Rpd_zz_o0_t3_l0 = 0.5 * vector(gtd_zz, i, j, k) * gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
							m_Rpd_zz_o0_t2_l0 = (-0.5 * vector(gtd_zz, i, j, k) * vector(Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_zz_o0_t1_l0 = (-0.5 * vector(gtd_zz, i, j, k) * vector(Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_zz_o0_t0_l0 = (-0.5 * vector(gtd_zz, i, j, k) * vector(Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
							Rpd_zz = (((((((((((((((((((((m_Rpd_zz_o0_t0_l0 + m_Rpd_zz_o0_t1_l0) + m_Rpd_zz_o0_t2_l0) + m_Rpd_zz_o0_t3_l0) + m_Rpd_zz_o0_t4_l0) + m_Rpd_zz_o0_t5_l0) + m_Rpd_zz_o0_t6_l0) + m_Rpd_zz_o0_t7_l0) + m_Rpd_zz_o0_t8_l0) + m_Rpd_zz_o0_t9_l0) + m_Rpd_zz_o0_t10_l0) + m_Rpd_zz_o0_t11_l0) + m_Rpd_zz_o0_t12_l0) + m_Rpd_zz_o0_t13_l0) + m_Rpd_zz_o0_t14_l0) + m_Rpd_zz_o0_t15_l0) + m_Rpd_zz_o0_t16_l0) + m_Rpd_zz_o0_t17_l0) + m_Rpd_zz_o0_t18_l0) + m_Rpd_zz_o0_t19_l0) + m_Rpd_zz_o0_t20_l0) + m_Rpd_zz_o0_t21_l0) + m_Rpd_zz_o0_t22_l0;
							EWeyl_zz = (-inv_chi * Rpd_zz) + (-Rtd_zz) + inv_chi * (vector(Atd_xz, i, j, k) * Atud_xz + vector(Atd_yz, i, j, k) * Atud_yz + vector(Atd_zz, i, j, k) * Atud_zz) + (-0.3333333333333333 * inv_chi * vector(trK, i, j, k) * vector(Atd_zz, i, j, k)) + (-0.2222222222222222 * inv_chi * vector(gtd_zz, i, j, k) * (vector(trK, i, j, k) * vector(trK, i, j, k)));
							d_Del_Kd_yyz_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xz, i, j, k) * Ct_xyy + vector(Atd_xy, i, j, k) * Ct_xyz + vector(Atd_yz, i, j, k) * Ct_yyy + vector(Atd_yy, i, j, k) * Ct_yyz + vector(Atd_zz, i, j, k) * Ct_zyy + vector(Atd_yz, i, j, k) * Ct_zyz);
							m_Del_Kd_yyz_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_yyz_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_yyz_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_yyz_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yy, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_yyz_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yy, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_yyz_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yy, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_yyz_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_yy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_yyz_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_yyz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_yz, i, j, k) * d_Del_Kd_yxx_o0_t2_m0_l0;
							m_Del_Kd_yyz_o0_t0_l0 = inv_chi * d_Del_Kd_yyz_o0_t0_m0_l0;
							Del_Kd_yyz = (((((((((m_Del_Kd_yyz_o0_t0_l0 + d_Del_Kd_yyz_o0_t1_m0_l0) + m_Del_Kd_yyz_o0_t2_l0) + m_Del_Kd_yyz_o0_t3_l0) + m_Del_Kd_yyz_o0_t4_l0) + m_Del_Kd_yyz_o0_t5_l0) + m_Del_Kd_yyz_o0_t6_l0) + m_Del_Kd_yyz_o0_t7_l0) + m_Del_Kd_yyz_o0_t8_l0) + m_Del_Kd_yyz_o0_t9_l0) + m_Del_Kd_yyz_o0_t10_l0;
							d_Del_Kd_yxy_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xy, i, j, k) * Ct_xxy + vector(Atd_xx, i, j, k) * Ct_xyy + vector(Atd_yy, i, j, k) * Ct_yxy + vector(Atd_xy, i, j, k) * Ct_yyy + vector(Atd_yz, i, j, k) * Ct_zxy + vector(Atd_xz, i, j, k) * Ct_zyy);
							m_Del_Kd_yxy_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yy, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_yxy_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yy, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_yxy_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yy, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_yxy_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
							m_Del_Kd_yxy_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_yxy_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_yxy_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Del_Kd_yxy_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_yy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
							m_Del_Kd_yxy_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xy, i, j, k) * d_Del_Kd_yxx_o0_t2_m0_l0;
							m_Del_Kd_yxy_o0_t0_l0 = inv_chi * d_Del_Kd_yxy_o0_t0_m0_l0;
							Del_Kd_yxy = (((((((((m_Del_Kd_yxy_o0_t0_l0 + d_Del_Kd_yxy_o0_t1_m0_l0) + m_Del_Kd_yxy_o0_t2_l0) + m_Del_Kd_yxy_o0_t3_l0) + m_Del_Kd_yxy_o0_t4_l0) + m_Del_Kd_yxy_o0_t5_l0) + m_Del_Kd_yxy_o0_t6_l0) + m_Del_Kd_yxy_o0_t7_l0) + m_Del_Kd_yxy_o0_t8_l0) + m_Del_Kd_yxy_o0_t9_l0) + m_Del_Kd_yxy_o0_t10_l0;
							BWeyl_xy = (-vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxy) + vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxy + vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyy + (-vector(gtd_xx, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zyy) + (-vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyz) + vector(gtd_xx, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yyz;
							BWeyl_zy = (-vector(gtd_zz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxy) + vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxy + vector(gtd_zz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyy + (-vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zyy) + (-vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyz) + vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yyz;
							BWeyl_yy = (-vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxy) + vector(gtd_yy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxy + vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyy + (-vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zyy) + (-vector(gtd_yy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyz) + vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yyz;
							m_Rpd_yy_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t6_m0_l0;
							m_Rpd_yy_o0_t21_l0 = (-0.75 * vector(gtd_yy, i, j, k) * gtu_zz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_yy_o0_t20_l0 = (-0.75 * vector(gtd_yy, i, j, k) * gtu_yy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yy_o0_t19_l0 = (-0.75 * vector(gtd_yy, i, j, k) * gtu_xx * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_yy_o0_t18_l0 = 2.0 * inv_chi * Zu_z * vector(gtd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yy_o0_t17_l0 = 2.0 * inv_chi * Zu_y * vector(gtd_yy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yy_o0_t16_l0 = 2.0 * inv_chi * Zu_x * vector(gtd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yy_o0_t15_l0 = (-1.500000000000000 * vector(gtd_yy, i, j, k) * gtu_yz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yy_o0_t14_l0 = (-1.500000000000000 * vector(gtd_yy, i, j, k) * gtu_xz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_yy_o0_t13_l0 = (-1.500000000000000 * vector(gtd_yy, i, j, k) * gtu_xy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_yy_o0_t12_l0 = (-0.5 * vector(gtd_yy, i, j, k) * vector(Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_yy_o0_t11_l0 = (-0.5 * vector(gtd_yy, i, j, k) * vector(Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yy_o0_t10_l0 = (-0.5 * vector(gtd_yy, i, j, k) * vector(Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
							m_Rpd_yy_o0_t9_l0 = 0.5 * vector(gtd_yy, i, j, k) * gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
							m_Rpd_yy_o0_t8_l0 = 1.0 * vector(gtd_yy, i, j, k) * gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
							m_Rpd_yy_o0_t7_l0 = 0.5 * vector(gtd_yy, i, j, k) * gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
							m_Rpd_yy_o0_t6_l0 = 1.0 * vector(gtd_yy, i, j, k) * gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
							m_Rpd_yy_o0_t5_l0 = 1.0 * vector(gtd_yy, i, j, k) * gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
							m_Rpd_yy_o0_t4_l0 = 0.5 * vector(gtd_yy, i, j, k) * gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
							m_Rpd_yy_o0_t3_l0 = (-0.25 * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yy_o0_t2_l0 = (-0.5 * Ct_zyy) * d_Rpd_xx_o0_t2_m0_l0;
							m_Rpd_yy_o0_t1_l0 = (-0.5 * Ct_yyy) * d_Rpd_xx_o0_t1_m0_l0;
							m_Rpd_yy_o0_t0_l0 = (-0.5 * Ct_xyy) * d_Rpd_xx_o0_t0_m0_l0;
							Rpd_yy = (((((((((((((((((((((m_Rpd_yy_o0_t0_l0 + m_Rpd_yy_o0_t1_l0) + m_Rpd_yy_o0_t2_l0) + m_Rpd_yy_o0_t3_l0) + m_Rpd_yy_o0_t4_l0) + m_Rpd_yy_o0_t5_l0) + m_Rpd_yy_o0_t6_l0) + m_Rpd_yy_o0_t7_l0) + m_Rpd_yy_o0_t8_l0) + m_Rpd_yy_o0_t9_l0) + m_Rpd_yy_o0_t10_l0) + m_Rpd_yy_o0_t11_l0) + m_Rpd_yy_o0_t12_l0) + m_Rpd_yy_o0_t13_l0) + m_Rpd_yy_o0_t14_l0) + m_Rpd_yy_o0_t15_l0) + m_Rpd_yy_o0_t16_l0) + m_Rpd_yy_o0_t17_l0) + m_Rpd_yy_o0_t18_l0) + m_Rpd_yy_o0_t19_l0) + m_Rpd_yy_o0_t20_l0) + m_Rpd_yy_o0_t21_l0) + m_Rpd_yy_o0_t22_l0;
							d_Rtd_xy_o0_t12_m0_l0 = gtu_zz * (Ct_xxz * Ctd_yxz + Ct_xxz * Ctd_xyz + Ct_xyz * Ctd_xxz + 2.0 * Ct_yxz * Ctd_yyz + Ct_yyz * Ctd_xyz + Ct_zxz * Ctd_yzz + Ct_zxz * Ctd_zyz + Ct_zyz * Ctd_xzz) + gtu_yz * (Ct_xxy * Ctd_yxz + Ct_xxy * Ctd_xyz + Ct_xyy * Ctd_xxz + 2.0 * Ct_yxy * Ctd_yyz + Ct_yyy * Ctd_xyz + Ct_zxy * Ctd_yzz + Ct_zxy * Ctd_zyz + Ct_zyy * Ctd_xzz) + gtu_xz * (Ct_xxx * Ctd_yxz + Ct_xxx * Ctd_xyz + Ct_xxy * Ctd_xxz + 2.0 * Ct_yxx * Ctd_yyz + Ct_yxy * Ctd_xyz + Ct_zxx * Ctd_yzz + Ct_zxx * Ctd_zyz + Ct_zxy * Ctd_xzz) + gtu_yz * (Ct_xxz * Ctd_yxy + Ct_xxz * Ctd_xyy + Ct_xyz * Ctd_xxy + 2.0 * Ct_yxz * Ctd_yyy + Ct_yyz * Ctd_xyy + Ct_zxz * Ctd_yyz + Ct_zxz * Ctd_zyy + Ct_zyz * Ctd_xyz) + gtu_yy * (Ct_xxy * Ctd_yxy + Ct_xxy * Ctd_xyy + Ct_xyy * Ctd_xxy + 2.0 * Ct_yxy * Ctd_yyy + Ct_yyy * Ctd_xyy + Ct_zxy * Ctd_yyz + Ct_zxy * Ctd_zyy + Ct_zyy * Ctd_xyz) + gtu_xy * (Ct_xxx * Ctd_yxy + Ct_xxx * Ctd_xyy + Ct_xxy * Ctd_xxy + 2.0 * Ct_yxx * Ctd_yyy + Ct_yxy * Ctd_xyy + Ct_zxx * Ctd_yyz + Ct_zxx * Ctd_zyy + Ct_zxy * Ctd_xyz) + gtu_xz * (Ct_xxz * Ctd_yxx + Ct_xxz * Ctd_xxy + Ct_xyz * Ctd_xxx + 2.0 * Ct_yxz * Ctd_yxy + Ct_yyz * Ctd_xxy + Ct_zxz * Ctd_yxz + Ct_zxz * Ctd_zxy + Ct_zyz * Ctd_xxz) + gtu_xy * (Ct_xxy * Ctd_yxx + Ct_xxy * Ctd_xxy + Ct_xyy * Ctd_xxx + 2.0 * Ct_yxy * Ctd_yxy + Ct_yyy * Ctd_xxy + Ct_zxy * Ctd_yxz + Ct_zxy * Ctd_zxy + Ct_zyy * Ctd_xxz) + gtu_xx * (Ct_xxx * Ctd_yxx + Ct_xxx * Ctd_xxy + Ct_xxy * Ctd_xxx + 2.0 * Ct_yxx * Ctd_yxy + Ct_yxy * Ctd_xxy + Ct_zxx * Ctd_yxz + Ct_zxx * Ctd_zxy + Ct_zxy * Ctd_xxz) + 0.5 * vector(Gamh_z, i, j, k) * (Ctd_xyz + Ctd_yxz) + 0.5 * vector(Gamh_y, i, j, k) * (Ctd_xyy + Ctd_yxy) + 0.5 * vector(Gamh_x, i, j, k) * (Ctd_xxy + Ctd_yxx);
							m_Rtd_xy_o0_t11_l0 = 0.5 * vector(gtd_yz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
							m_Rtd_xy_o0_t10_l0 = 0.5 * vector(gtd_yy, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
							m_Rtd_xy_o0_t9_l0 = 0.5 * vector(gtd_xy, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
							m_Rtd_xy_o0_t8_l0 = 0.5 * vector(gtd_xz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
							m_Rtd_xy_o0_t7_l0 = 0.5 * vector(gtd_xy, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
							m_Rtd_xy_o0_t6_l0 = 0.5 * vector(gtd_xx, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
							m_Rtd_xy_o0_t5_l0 = (-0.5 * gtu_zz) * d_Rtd_xy_o0_t5_m0_l0;
							m_Rtd_xy_o0_t4_l0 = (-1.0 * gtu_yz) * d_Rtd_xy_o0_t4_m0_l0;
							m_Rtd_xy_o0_t3_l0 = (-0.5 * gtu_yy) * d_Rtd_xy_o0_t3_m0_l0;
							m_Rtd_xy_o0_t2_l0 = (-1.0 * gtu_xz) * d_Rtd_xy_o0_t2_m0_l0;
							m_Rtd_xy_o0_t1_l0 = (-1.0 * gtu_xy) * d_Rtd_xy_o0_t1_m0_l0;
							m_Rtd_xy_o0_t0_l0 = (-0.5 * gtu_xx) * d_Rtd_xy_o0_t0_m0_l0;
							Rtd_xy = (((((((((((m_Rtd_xy_o0_t0_l0 + m_Rtd_xy_o0_t1_l0) + m_Rtd_xy_o0_t2_l0) + m_Rtd_xy_o0_t3_l0) + m_Rtd_xy_o0_t4_l0) + m_Rtd_xy_o0_t5_l0) + m_Rtd_xy_o0_t6_l0) + m_Rtd_xy_o0_t7_l0) + m_Rtd_xy_o0_t8_l0) + m_Rtd_xy_o0_t9_l0) + m_Rtd_xy_o0_t10_l0) + m_Rtd_xy_o0_t11_l0) + d_Rtd_xy_o0_t12_m0_l0;
							EWeyl_xy = (-inv_chi * Rpd_xy) + (-Rtd_xy) + inv_chi * (vector(Atd_xx, i, j, k) * Atud_xy + vector(Atd_xy, i, j, k) * Atud_yy + vector(Atd_xz, i, j, k) * Atud_zy) + (-0.3333333333333333 * inv_chi * vector(trK, i, j, k) * vector(Atd_xy, i, j, k)) + (-0.2222222222222222 * inv_chi * vector(gtd_xy, i, j, k) * (vector(trK, i, j, k) * vector(trK, i, j, k)));
							d_Rtd_yy_o0_t9_m0_l0 = gtu_zz * (2.0 * Ct_xyz * Ctd_yxz + Ct_xyz * Ctd_xyz + 3.0 * Ct_yyz * Ctd_yyz + 2.0 * Ct_zyz * Ctd_yzz + Ct_zyz * Ctd_zyz) + gtu_yz * (2.0 * Ct_xyy * Ctd_yxz + Ct_xyy * Ctd_xyz + 3.0 * Ct_yyy * Ctd_yyz + 2.0 * Ct_zyy * Ctd_yzz + Ct_zyy * Ctd_zyz) + gtu_xz * (2.0 * Ct_xxy * Ctd_yxz + Ct_xxy * Ctd_xyz + 3.0 * Ct_yxy * Ctd_yyz + 2.0 * Ct_zxy * Ctd_yzz + Ct_zxy * Ctd_zyz) + gtu_yz * (2.0 * Ct_xyz * Ctd_yxy + Ct_xyz * Ctd_xyy + 3.0 * Ct_yyz * Ctd_yyy + 2.0 * Ct_zyz * Ctd_yyz + Ct_zyz * Ctd_zyy) + gtu_yy * (2.0 * Ct_xyy * Ctd_yxy + Ct_xyy * Ctd_xyy + 3.0 * Ct_yyy * Ctd_yyy + 2.0 * Ct_zyy * Ctd_yyz + Ct_zyy * Ctd_zyy) + gtu_xy * (2.0 * Ct_xxy * Ctd_yxy + Ct_xxy * Ctd_xyy + 3.0 * Ct_yxy * Ctd_yyy + 2.0 * Ct_zxy * Ctd_yyz + Ct_zxy * Ctd_zyy) + gtu_xz * (2.0 * Ct_xyz * Ctd_yxx + Ct_xyz * Ctd_xxy + 3.0 * Ct_yyz * Ctd_yxy + 2.0 * Ct_zyz * Ctd_yxz + Ct_zyz * Ctd_zxy) + gtu_xy * (2.0 * Ct_xyy * Ctd_yxx + Ct_xyy * Ctd_xxy + 3.0 * Ct_yyy * Ctd_yxy + 2.0 * Ct_zyy * Ctd_yxz + Ct_zyy * Ctd_zxy) + gtu_xx * (2.0 * Ct_xxy * Ctd_yxx + Ct_xxy * Ctd_xxy + 3.0 * Ct_yxy * Ctd_yxy + 2.0 * Ct_zxy * Ctd_yxz + Ct_zxy * Ctd_zxy) + 1.0 * vector(Gamh_z, i, j, k) * Ctd_yyz + 1.0 * vector(Gamh_y, i, j, k) * Ctd_yyy + 1.0 * vector(Gamh_x, i, j, k) * Ctd_yxy;
							m_Rtd_yy_o0_t8_l0 = 1.0 * vector(gtd_yz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
							m_Rtd_yy_o0_t7_l0 = 1.0 * vector(gtd_yy, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
							m_Rtd_yy_o0_t6_l0 = 1.0 * vector(gtd_xy, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
							m_Rtd_yy_o0_t5_l0 = (-0.5 * gtu_zz) * d_Rtd_yy_o0_t5_m0_l0;
							m_Rtd_yy_o0_t4_l0 = (-1.0 * gtu_yz) * d_Rtd_yy_o0_t4_m0_l0;
							m_Rtd_yy_o0_t3_l0 = (-0.5 * gtu_yy) * d_Rtd_yy_o0_t3_m0_l0;
							m_Rtd_yy_o0_t2_l0 = (-1.0 * gtu_xz) * d_Rtd_yy_o0_t2_m0_l0;
							m_Rtd_yy_o0_t1_l0 = (-1.0 * gtu_xy) * d_Rtd_yy_o0_t1_m0_l0;
							m_Rtd_yy_o0_t0_l0 = (-0.5 * gtu_xx) * d_Rtd_yy_o0_t0_m0_l0;
							Rtd_yy = ((((((((m_Rtd_yy_o0_t0_l0 + m_Rtd_yy_o0_t1_l0) + m_Rtd_yy_o0_t2_l0) + m_Rtd_yy_o0_t3_l0) + m_Rtd_yy_o0_t4_l0) + m_Rtd_yy_o0_t5_l0) + m_Rtd_yy_o0_t6_l0) + m_Rtd_yy_o0_t7_l0) + m_Rtd_yy_o0_t8_l0) + d_Rtd_yy_o0_t9_m0_l0;
							EWeyl_yy = (-inv_chi * Rpd_yy) + (-Rtd_yy) + inv_chi * (vector(Atd_xy, i, j, k) * Atud_xy + vector(Atd_yy, i, j, k) * Atud_yy + vector(Atd_yz, i, j, k) * Atud_zy) + (-0.3333333333333333 * inv_chi * vector(trK, i, j, k) * vector(Atd_yy, i, j, k)) + (-0.2222222222222222 * inv_chi * vector(gtd_yy, i, j, k) * (vector(trK, i, j, k) * vector(trK, i, j, k)));
							d_Rtd_yz_o0_t12_m0_l0 = gtu_zz * (Ct_xyz * Ctd_zxz + Ct_xyz * Ctd_xzz + Ct_xzz * Ctd_yxz + Ct_yyz * Ctd_zyz + Ct_yyz * Ctd_yzz + Ct_yzz * Ctd_yyz + 2.0 * Ct_zyz * Ctd_zzz + Ct_zzz * Ctd_yzz) + gtu_yz * (Ct_xyy * Ctd_zxz + Ct_xyy * Ctd_xzz + Ct_xyz * Ctd_yxz + Ct_yyy * Ctd_zyz + Ct_yyy * Ctd_yzz + Ct_yyz * Ctd_yyz + 2.0 * Ct_zyy * Ctd_zzz + Ct_zyz * Ctd_yzz) + gtu_xz * (Ct_xxy * Ctd_zxz + Ct_xxy * Ctd_xzz + Ct_xxz * Ctd_yxz + Ct_yxy * Ctd_zyz + Ct_yxy * Ctd_yzz + Ct_yxz * Ctd_yyz + 2.0 * Ct_zxy * Ctd_zzz + Ct_zxz * Ctd_yzz) + gtu_yz * (Ct_xyz * Ctd_zxy + Ct_xyz * Ctd_xyz + Ct_xzz * Ctd_yxy + Ct_yyz * Ctd_zyy + Ct_yyz * Ctd_yyz + Ct_yzz * Ctd_yyy + 2.0 * Ct_zyz * Ctd_zyz + Ct_zzz * Ctd_yyz) + gtu_yy * (Ct_xyy * Ctd_zxy + Ct_xyy * Ctd_xyz + Ct_xyz * Ctd_yxy + Ct_yyy * Ctd_zyy + Ct_yyy * Ctd_yyz + Ct_yyz * Ctd_yyy + 2.0 * Ct_zyy * Ctd_zyz + Ct_zyz * Ctd_yyz) + gtu_xy * (Ct_xxy * Ctd_zxy + Ct_xxy * Ctd_xyz + Ct_xxz * Ctd_yxy + Ct_yxy * Ctd_zyy + Ct_yxy * Ctd_yyz + Ct_yxz * Ctd_yyy + 2.0 * Ct_zxy * Ctd_zyz + Ct_zxz * Ctd_yyz) + gtu_xz * (Ct_xyz * Ctd_zxx + Ct_xyz * Ctd_xxz + Ct_xzz * Ctd_yxx + Ct_yyz * Ctd_zxy + Ct_yyz * Ctd_yxz + Ct_yzz * Ctd_yxy + 2.0 * Ct_zyz * Ctd_zxz + Ct_zzz * Ctd_yxz) + gtu_xy * (Ct_xyy * Ctd_zxx + Ct_xyy * Ctd_xxz + Ct_xyz * Ctd_yxx + Ct_yyy * Ctd_zxy + Ct_yyy * Ctd_yxz + Ct_yyz * Ctd_yxy + 2.0 * Ct_zyy * Ctd_zxz + Ct_zyz * Ctd_yxz) + gtu_xx * (Ct_xxy * Ctd_zxx + Ct_xxy * Ctd_xxz + Ct_xxz * Ctd_yxx + Ct_yxy * Ctd_zxy + Ct_yxy * Ctd_yxz + Ct_yxz * Ctd_yxy + 2.0 * Ct_zxy * Ctd_zxz + Ct_zxz * Ctd_yxz) + 0.5 * vector(Gamh_z, i, j, k) * (Ctd_yzz + Ctd_zyz) + 0.5 * vector(Gamh_y, i, j, k) * (Ctd_yyz + Ctd_zyy) + 0.5 * vector(Gamh_x, i, j, k) * (Ctd_yxz + Ctd_zxy);
							m_Rtd_yz_o0_t11_l0 = 0.5 * vector(gtd_zz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
							m_Rtd_yz_o0_t10_l0 = 0.5 * vector(gtd_yz, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
							m_Rtd_yz_o0_t9_l0 = 0.5 * vector(gtd_xz, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
							m_Rtd_yz_o0_t8_l0 = 0.5 * vector(gtd_yz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
							m_Rtd_yz_o0_t7_l0 = 0.5 * vector(gtd_yy, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
							m_Rtd_yz_o0_t6_l0 = 0.5 * vector(gtd_xy, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
							m_Rtd_yz_o0_t5_l0 = (-0.5 * gtu_zz) * d_Rtd_yz_o0_t5_m0_l0;
							m_Rtd_yz_o0_t4_l0 = (-1.0 * gtu_yz) * d_Rtd_yz_o0_t4_m0_l0;
							m_Rtd_yz_o0_t3_l0 = (-0.5 * gtu_yy) * d_Rtd_yz_o0_t3_m0_l0;
							m_Rtd_yz_o0_t2_l0 = (-1.0 * gtu_xz) * d_Rtd_yz_o0_t2_m0_l0;
							m_Rtd_yz_o0_t1_l0 = (-1.0 * gtu_xy) * d_Rtd_yz_o0_t1_m0_l0;
							m_Rtd_yz_o0_t0_l0 = (-0.5 * gtu_xx) * d_Rtd_yz_o0_t0_m0_l0;
							Rtd_yz = (((((((((((m_Rtd_yz_o0_t0_l0 + m_Rtd_yz_o0_t1_l0) + m_Rtd_yz_o0_t2_l0) + m_Rtd_yz_o0_t3_l0) + m_Rtd_yz_o0_t4_l0) + m_Rtd_yz_o0_t5_l0) + m_Rtd_yz_o0_t6_l0) + m_Rtd_yz_o0_t7_l0) + m_Rtd_yz_o0_t8_l0) + m_Rtd_yz_o0_t9_l0) + m_Rtd_yz_o0_t10_l0) + m_Rtd_yz_o0_t11_l0) + d_Rtd_yz_o0_t12_m0_l0;
							EWeyl_yz = (-inv_chi * Rpd_yz) + (-Rtd_yz) + inv_chi * (vector(Atd_xy, i, j, k) * Atud_xz + vector(Atd_yy, i, j, k) * Atud_yz + vector(Atd_yz, i, j, k) * Atud_zz) + (-0.3333333333333333 * inv_chi * vector(trK, i, j, k) * vector(Atd_yz, i, j, k)) + (-0.2222222222222222 * inv_chi * vector(gtd_yz, i, j, k) * (vector(trK, i, j, k) * vector(trK, i, j, k)));
							Bfvf = vector(Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
							Sfuu_zz = 1.0 * Sfu_z * vfu_z + vector(chi, i, j, k) * gtu_zz * sdetg * vector(pf, i, j, k) + (-((-0.5 * vector(chi, i, j, k) * sqBf * gtu_zz) + vector(Bfu_z, i, j, k) * vector(Bfu_z, i, j, k)) / (sdetg * sqW)) + (-0.5 * (Bfvf * ((-Bfvf * vector(chi, i, j, k) * gtu_zz) + 2.0 * vector(Bfu_z, i, j, k) * vfu_z)) / sdetg);
							Tf4u_zz = Sfuu_zz / sdetg + (-(Sfu_z * vector(Betau_z, i, j, k)) / (sdetg * vector(Alpha, i, j, k))) + (-vector(Betau_z, i, j, k) * Tf4u_tz);
							Tu_zz = Tf4u_zz;
							Sfuu_yz = 0.5 * Sfu_y * vfu_z + 0.5 * Sfu_z * vfu_y + vector(chi, i, j, k) * gtu_yz * sdetg * vector(pf, i, j, k) + (-((-0.5 * vector(chi, i, j, k) * sqBf * gtu_yz) + vector(Bfu_y, i, j, k) * vector(Bfu_z, i, j, k)) / (sdetg * sqW)) + (-0.5 * (Bfvf * ((-Bfvf * vector(chi, i, j, k) * gtu_yz) + vector(Bfu_y, i, j, k) * vfu_z + vector(Bfu_z, i, j, k) * vfu_y)) / sdetg);
							Tf4u_yz = Sfuu_yz / sdetg + (-(Sfu_y * vector(Betau_z, i, j, k)) / (sdetg * vector(Alpha, i, j, k))) + (-vector(Betau_y, i, j, k) * Tf4u_tz);
							Tu_yz = Tf4u_yz;
							Sfuu_yy = 1.0 * Sfu_y * vfu_y + vector(chi, i, j, k) * gtu_yy * sdetg * vector(pf, i, j, k) + (-((-0.5 * vector(chi, i, j, k) * sqBf * gtu_yy) + vector(Bfu_y, i, j, k) * vector(Bfu_y, i, j, k)) / (sdetg * sqW)) + (-0.5 * (Bfvf * ((-Bfvf * vector(chi, i, j, k) * gtu_yy) + 2.0 * vector(Bfu_y, i, j, k) * vfu_y)) / sdetg);
							Tf4u_yy = Sfuu_yy / sdetg + (-(Sfu_y * vector(Betau_y, i, j, k)) / (sdetg * vector(Alpha, i, j, k))) + (-vector(Betau_y, i, j, k) * Tf4u_ty);
							Tu_yy = Tf4u_yy;
							Sfuu_xy = 0.5 * Sfu_x * vfu_y + 0.5 * Sfu_y * vfu_x + vector(chi, i, j, k) * gtu_xy * sdetg * vector(pf, i, j, k) + (-((-0.5 * vector(chi, i, j, k) * sqBf * gtu_xy) + vector(Bfu_x, i, j, k) * vector(Bfu_y, i, j, k)) / (sdetg * sqW)) + (-0.5 * (Bfvf * ((-Bfvf * vector(chi, i, j, k) * gtu_xy) + vector(Bfu_x, i, j, k) * vfu_y + vector(Bfu_y, i, j, k) * vfu_x)) / sdetg);
							Tf4u_xy = Sfuu_xy / sdetg + (-(Sfu_x * vector(Betau_y, i, j, k)) / (sdetg * vector(Alpha, i, j, k))) + (-vector(Betau_x, i, j, k) * Tf4u_ty);
							Tu_xy = Tf4u_xy;
							Sfuu_xx = 1.0 * Sfu_x * vfu_x + vector(chi, i, j, k) * gtu_xx * sdetg * vector(pf, i, j, k) + (-((-0.5 * vector(chi, i, j, k) * sqBf * gtu_xx) + vector(Bfu_x, i, j, k) * vector(Bfu_x, i, j, k)) / (sdetg * sqW)) + (-0.5 * (Bfvf * ((-Bfvf * vector(chi, i, j, k) * gtu_xx) + 2.0 * vector(Bfu_x, i, j, k) * vfu_x)) / sdetg);
							Tf4u_xx = Sfuu_xx / sdetg + (-(Sfu_x * vector(Betau_x, i, j, k)) / (sdetg * vector(Alpha, i, j, k))) + (-vector(Betau_x, i, j, k) * Tf4u_tx);
							Tu_xx = Tf4u_xx;
							Sfuu_xz = 0.5 * Sfu_x * vfu_z + 0.5 * Sfu_z * vfu_x + vector(chi, i, j, k) * gtu_xz * sdetg * vector(pf, i, j, k) + (-((-0.5 * vector(chi, i, j, k) * sqBf * gtu_xz) + vector(Bfu_x, i, j, k) * vector(Bfu_z, i, j, k)) / (sdetg * sqW)) + (-0.5 * (Bfvf * ((-Bfvf * vector(chi, i, j, k) * gtu_xz) + vector(Bfu_x, i, j, k) * vfu_z + vector(Bfu_z, i, j, k) * vfu_x)) / sdetg);
							Tf4u_xz = Sfuu_xz / sdetg + (-(Sfu_x * vector(Betau_z, i, j, k)) / (sdetg * vector(Alpha, i, j, k))) + (-vector(Betau_x, i, j, k) * Tf4u_tz);
							Tu_xz = Tf4u_xz;
							Tu_tt = Tf4u_tt;
							Td_yz = g4d_ty * g4d_tz * Tu_tt + g4d_xy * g4d_tz * Tu_tx + g4d_yy * g4d_tz * Tu_ty + g4d_yz * g4d_tz * Tu_tz + g4d_ty * g4d_xz * Tu_tx + g4d_xy * g4d_xz * Tu_xx + g4d_yy * g4d_xz * Tu_xy + g4d_yz * g4d_xz * Tu_xz + g4d_ty * g4d_yz * Tu_ty + g4d_xy * g4d_yz * Tu_xy + g4d_yy * g4d_yz * Tu_yy + (g4d_yz * g4d_yz) * Tu_yz + g4d_ty * g4d_zz * Tu_tz + g4d_xy * g4d_zz * Tu_xz + g4d_yy * g4d_zz * Tu_yz + g4d_yz * g4d_zz * Tu_zz;
							Td_yy = (g4d_ty * g4d_ty) * Tu_tt + 2.0 * g4d_xy * g4d_ty * Tu_tx + 2.0 * g4d_yy * g4d_ty * Tu_ty + 2.0 * g4d_yz * g4d_ty * Tu_tz + (g4d_xy * g4d_xy) * Tu_xx + 2.0 * g4d_yy * g4d_xy * Tu_xy + 2.0 * g4d_yz * g4d_xy * Tu_xz + (g4d_yy * g4d_yy) * Tu_yy + 2.0 * g4d_yz * g4d_yy * Tu_yz + (g4d_yz * g4d_yz) * Tu_zz;
							Td_xz = g4d_tx * g4d_tz * Tu_tt + g4d_xx * g4d_tz * Tu_tx + g4d_xy * g4d_tz * Tu_ty + g4d_xz * g4d_tz * Tu_tz + g4d_tx * g4d_xz * Tu_tx + g4d_xx * g4d_xz * Tu_xx + g4d_xy * g4d_xz * Tu_xy + (g4d_xz * g4d_xz) * Tu_xz + g4d_tx * g4d_yz * Tu_ty + g4d_xx * g4d_yz * Tu_xy + g4d_xy * g4d_yz * Tu_yy + g4d_xz * g4d_yz * Tu_yz + g4d_tx * g4d_zz * Tu_tz + g4d_xx * g4d_zz * Tu_xz + g4d_xy * g4d_zz * Tu_yz + g4d_xz * g4d_zz * Tu_zz;
							Td_xy = g4d_tx * g4d_ty * Tu_tt + g4d_xx * g4d_ty * Tu_tx + g4d_xy * g4d_ty * Tu_ty + g4d_xz * g4d_ty * Tu_tz + g4d_tx * g4d_xy * Tu_tx + g4d_xx * g4d_xy * Tu_xx + (g4d_xy * g4d_xy) * Tu_xy + g4d_xz * g4d_xy * Tu_xz + g4d_tx * g4d_yy * Tu_ty + g4d_xx * g4d_yy * Tu_xy + g4d_xy * g4d_yy * Tu_yy + g4d_xz * g4d_yy * Tu_yz + g4d_tx * g4d_yz * Tu_tz + g4d_xx * g4d_yz * Tu_xz + g4d_xy * g4d_yz * Tu_yz + g4d_xz * g4d_yz * Tu_zz;
							Td_xx = (g4d_tx * g4d_tx) * Tu_tt + 2.0 * g4d_xx * g4d_tx * Tu_tx + 2.0 * g4d_xy * g4d_tx * Tu_ty + 2.0 * g4d_xz * g4d_tx * Tu_tz + (g4d_xx * g4d_xx) * Tu_xx + 2.0 * g4d_xy * g4d_xx * Tu_xy + 2.0 * g4d_xz * g4d_xx * Tu_xz + (g4d_xy * g4d_xy) * Tu_yy + 2.0 * g4d_xz * g4d_xy * Tu_yz + (g4d_xz * g4d_xz) * Tu_zz;
							Td_zz = (g4d_tz * g4d_tz) * Tu_tt + 2.0 * g4d_xz * g4d_tz * Tu_tx + 2.0 * g4d_yz * g4d_tz * Tu_ty + 2.0 * g4d_zz * g4d_tz * Tu_tz + (g4d_xz * g4d_xz) * Tu_xx + 2.0 * g4d_yz * g4d_xz * Tu_xy + 2.0 * g4d_zz * g4d_xz * Tu_xz + (g4d_yz * g4d_yz) * Tu_yy + 2.0 * g4d_zz * g4d_yz * Tu_yz + (g4d_zz * g4d_zz) * Tu_zz;
							TT = g4d_tt * Tu_tt + 2.0 * g4d_tx * Tu_tx + 2.0 * g4d_ty * Tu_ty + 2.0 * g4d_tz * Tu_tz + g4d_xx * Tu_xx + 2.0 * g4d_xy * Tu_xy + 2.0 * g4d_xz * Tu_xz + g4d_yy * Tu_yy + 2.0 * g4d_yz * Tu_yz + g4d_zz * Tu_zz;
							Td_tx = g4d_tt * g4d_tx * Tu_tt + (g4d_tx * g4d_tx) * Tu_tx + g4d_ty * g4d_tx * Tu_ty + g4d_tz * g4d_tx * Tu_tz + g4d_tt * g4d_xx * Tu_tx + g4d_tx * g4d_xx * Tu_xx + g4d_ty * g4d_xx * Tu_xy + g4d_tz * g4d_xx * Tu_xz + g4d_tt * g4d_xy * Tu_ty + g4d_tx * g4d_xy * Tu_xy + g4d_ty * g4d_xy * Tu_yy + g4d_tz * g4d_xy * Tu_yz + g4d_tt * g4d_xz * Tu_tz + g4d_tx * g4d_xz * Tu_xz + g4d_ty * g4d_xz * Tu_yz + g4d_tz * g4d_xz * Tu_zz;
							Td_tz = g4d_tt * g4d_tz * Tu_tt + g4d_tx * g4d_tz * Tu_tx + g4d_ty * g4d_tz * Tu_ty + (g4d_tz * g4d_tz) * Tu_tz + g4d_tt * g4d_xz * Tu_tx + g4d_tx * g4d_xz * Tu_xx + g4d_ty * g4d_xz * Tu_xy + g4d_tz * g4d_xz * Tu_xz + g4d_tt * g4d_yz * Tu_ty + g4d_tx * g4d_yz * Tu_xy + g4d_ty * g4d_yz * Tu_yy + g4d_tz * g4d_yz * Tu_yz + g4d_tt * g4d_zz * Tu_tz + g4d_tx * g4d_zz * Tu_xz + g4d_ty * g4d_zz * Tu_yz + g4d_tz * g4d_zz * Tu_zz;
							Td_ty = g4d_tt * g4d_ty * Tu_tt + g4d_tx * g4d_ty * Tu_tx + (g4d_ty * g4d_ty) * Tu_ty + g4d_tz * g4d_ty * Tu_tz + g4d_tt * g4d_xy * Tu_tx + g4d_tx * g4d_xy * Tu_xx + g4d_ty * g4d_xy * Tu_xy + g4d_tz * g4d_xy * Tu_xz + g4d_tt * g4d_yy * Tu_ty + g4d_tx * g4d_yy * Tu_xy + g4d_ty * g4d_yy * Tu_yy + g4d_tz * g4d_yy * Tu_yz + g4d_tt * g4d_yz * Tu_tz + g4d_tx * g4d_yz * Tu_xz + g4d_ty * g4d_yz * Tu_yz + g4d_tz * g4d_yz * Tu_zz;
							Jtd_ADM_z = vector(Alpha, i, j, k) * ((vector(Betau_x, i, j, k) * Tu_tt + Tu_tx) * vector(gtd_xz, i, j, k) + (vector(Betau_y, i, j, k) * Tu_tt + Tu_ty) * vector(gtd_yz, i, j, k) + (vector(Betau_z, i, j, k) * Tu_tt + Tu_tz) * vector(gtd_zz, i, j, k));
							Td_tt = (g4d_tt * g4d_tt) * Tu_tt + 2.0 * g4d_tx * g4d_tt * Tu_tx + 2.0 * g4d_ty * g4d_tt * Tu_ty + 2.0 * g4d_tz * g4d_tt * Tu_tz + (g4d_tx * g4d_tx) * Tu_xx + 2.0 * g4d_ty * g4d_tx * Tu_xy + 2.0 * g4d_tz * g4d_tx * Tu_xz + (g4d_ty * g4d_ty) * Tu_yy + 2.0 * g4d_tz * g4d_ty * Tu_yz + (g4d_tz * g4d_tz) * Tu_zz;
							Jtd_ADM_y = vector(Alpha, i, j, k) * ((vector(Betau_x, i, j, k) * Tu_tt + Tu_tx) * vector(gtd_xy, i, j, k) + (vector(Betau_y, i, j, k) * Tu_tt + Tu_ty) * vector(gtd_yy, i, j, k) + (vector(Betau_z, i, j, k) * Tu_tt + Tu_tz) * vector(gtd_yz, i, j, k));
							Jtd_ADM_x = vector(Alpha, i, j, k) * ((vector(Betau_x, i, j, k) * Tu_tt + Tu_tx) * vector(gtd_xx, i, j, k) + (vector(Betau_y, i, j, k) * Tu_tt + Tu_ty) * vector(gtd_xy, i, j, k) + (vector(Betau_z, i, j, k) * Tu_tt + Tu_tz) * vector(gtd_xz, i, j, k));
							rho_ADM = (vector(Alpha, i, j, k) * vector(Alpha, i, j, k)) * Tu_tt;
							d_Rscalar_o0_t0_m0_l0 = (vector(chi, i, j, k) * Rtd_xx + Rpd_xx) * gtu_xx + 2.0 * (vector(chi, i, j, k) * Rtd_xy + Rpd_xy) * gtu_xy + 2.0 * (vector(chi, i, j, k) * Rtd_xz + Rpd_xz) * gtu_xz + (vector(chi, i, j, k) * Rtd_yy + Rpd_yy) * gtu_yy + 2.0 * (vector(chi, i, j, k) * Rtd_yz + Rpd_yz) * gtu_yz + (vector(chi, i, j, k) * Rtd_zz + Rpd_zz) * gtu_zz;
							d_HamCon_o0_t0_m0_l0 = (-16.0 * M_PI * rho_ADM) + (vector(chi, i, j, k) * Rtd_xx + Rpd_xx) * gtu_xx + 2.0 * (vector(chi, i, j, k) * Rtd_xy + Rpd_xy) * gtu_xy + 2.0 * (vector(chi, i, j, k) * Rtd_xz + Rpd_xz) * gtu_xz + (vector(chi, i, j, k) * Rtd_yy + Rpd_yy) * gtu_yy + 2.0 * (vector(chi, i, j, k) * Rtd_yz + Rpd_yz) * gtu_yz + (vector(chi, i, j, k) * Rtd_zz + Rpd_zz) * gtu_zz + (-vector(Atd_xx, i, j, k) * Atu_xx) + (-2.0 * vector(Atd_xy, i, j, k) * Atu_xy) + (-2.0 * vector(Atd_xz, i, j, k) * Atu_xz) + (-vector(Atd_yy, i, j, k) * Atu_yy) + (-2.0 * vector(Atd_yz, i, j, k) * Atu_yz) + (-vector(Atd_zz, i, j, k) * Atu_zz) + 0.6666666666666667 * (vector(trK, i, j, k) * vector(trK, i, j, k));
							d_MomCon_x_o0_t13_m0_l0 = (-8.0) * M_PI * Jtd_ADM_x * inv_chi + (-Ct_xxz * Atud_zx) + (-Ct_xxy * Atud_yx) + (-Ct_xxx * Atud_xx) + (-Gamt_z * vector(Atd_xz, i, j, k)) + (-Gamt_y * vector(Atd_xy, i, j, k)) + (-Gamt_x * vector(Atd_xx, i, j, k)) + (-Ct_zxz * Atud_zz) + (-Ct_zxy * Atud_yz) + (-Ct_zxx * Atud_xz) + (-Ct_yxz * Atud_zy) + (-Ct_yxy * Atud_yy) + (-Ct_yxx * Atud_xy);
							d_MomCon_y_o0_t13_m0_l0 = (-8.0) * M_PI * Jtd_ADM_y * inv_chi + (-Gamt_z * vector(Atd_yz, i, j, k)) + (-Gamt_y * vector(Atd_yy, i, j, k)) + (-Gamt_x * vector(Atd_xy, i, j, k)) + (-Ct_zyz * Atud_zz) + (-Ct_zyy * Atud_yz) + (-Ct_zxy * Atud_xz) + (-Ct_yyz * Atud_zy) + (-Ct_yyy * Atud_yy) + (-Ct_yxy * Atud_xy) + (-Ct_xyz * Atud_zx) + (-Ct_xyy * Atud_yx) + (-Ct_xxy * Atud_xx);
							d_MomCon_z_o0_t13_m0_l0 = (-Gamt_z * vector(Atd_zz, i, j, k)) + (-Gamt_y * vector(Atd_yz, i, j, k)) + (-Gamt_x * vector(Atd_xz, i, j, k)) + (-Ct_zzz * Atud_zz) + (-Ct_zyz * Atud_yz) + (-Ct_zxz * Atud_xz) + (-Ct_yzz * Atud_zy) + (-Ct_yyz * Atud_yy) + (-Ct_yxz * Atud_xy) + (-Ct_xzz * Atud_zx) + (-Ct_xyz * Atud_yx) + (-Ct_xxz * Atud_xx) + (-8.0) * M_PI * Jtd_ADM_z * inv_chi;
							d_trA_o0_t0_m0_l0 = Atud_xx + Atud_yy + Atud_zz;
							d_detgtm1_o0_t0_m0_l0 = detgtd + (-1.0);
							d_M_ADM_surf_o0_t9_m0_l0 = 0.0198943678864869 * vector(chi, i, j, k) * Gamt_z * dSigma_z + 0.0198943678864869 * vector(chi, i, j, k) * Gamt_y * dSigma_y + 0.0198943678864869 * vector(chi, i, j, k) * Gamt_x * dSigma_x;
							d_Jz_ADM_surf_o0_t0_m0_l0 = 0.0397887357729738 * vector(chi, i, j, k) * (uph_x * (0.3333333333333333 * vector(trK, i, j, k) * gtu_xx + Atu_xx) + uph_y * (0.3333333333333333 * vector(trK, i, j, k) * gtu_xy + Atu_xy) + uph_z * (0.3333333333333333 * vector(trK, i, j, k) * gtu_xz + Atu_xz)) * dSigma_x + 0.0397887357729738 * vector(chi, i, j, k) * (uph_x * (0.3333333333333333 * vector(trK, i, j, k) * gtu_xy + Atu_xy) + uph_y * (0.3333333333333333 * vector(trK, i, j, k) * gtu_yy + Atu_yy) + uph_z * (0.3333333333333333 * vector(trK, i, j, k) * gtu_yz + Atu_yz)) * dSigma_y + 0.0397887357729738 * vector(chi, i, j, k) * (uph_x * (0.3333333333333333 * vector(trK, i, j, k) * gtu_xz + Atu_xz) + uph_y * (0.3333333333333333 * vector(trK, i, j, k) * gtu_yz + Atu_yz) + uph_z * (0.3333333333333333 * vector(trK, i, j, k) * gtu_zz + Atu_zz)) * dSigma_z;
							d_M_Komar_o0_t0_m0_l0 = 2.0 * (((-0.5 * TT * g4d_tt) + Td_tt) / vector(Alpha, i, j, k) + (-(((-0.5 * TT * g4d_tx) + Td_tx) * vector(Betau_x, i, j, k)) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_ty) + Td_ty) * vector(Betau_y, i, j, k)) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_tz) + Td_tz) * vector(Betau_z, i, j, k)) / vector(Alpha, i, j, k))) / (pow(vector(chi, i, j, k), 1.500000000000000));
							d_Jz_Komar_o0_t0_m0_l0 = -((((-0.5 * TT * g4d_tx) + Td_tx) * vph_x) / vector(Alpha, i, j, k) + (-(((-0.5 * TT * g4d_xx) + Td_xx) * vector(Betau_x, i, j, k) * vph_x) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_xy) + Td_xy) * vector(Betau_y, i, j, k) * vph_x) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_xz) + Td_xz) * vector(Betau_z, i, j, k) * vph_x) / vector(Alpha, i, j, k)) + (((-0.5 * TT * g4d_ty) + Td_ty) * vph_y) / vector(Alpha, i, j, k) + (-(((-0.5 * TT * g4d_xy) + Td_xy) * vector(Betau_x, i, j, k) * vph_y) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_yy) + Td_yy) * vector(Betau_y, i, j, k) * vph_y) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_yz) + Td_yz) * vector(Betau_z, i, j, k) * vph_y) / vector(Alpha, i, j, k)) + (((-0.5 * TT * g4d_tz) + Td_tz) * vph_z) / vector(Alpha, i, j, k) + (-(((-0.5 * TT * g4d_xz) + Td_xz) * vector(Betau_x, i, j, k) * vph_z) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_yz) + Td_yz) * vector(Betau_y, i, j, k) * vph_z) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_zz) + Td_zz) * vector(Betau_z, i, j, k) * vph_z) / vector(Alpha, i, j, k))) / (pow(vector(chi, i, j, k), 1.500000000000000));
							d_psi4R_o0_t0_m0_l0 = MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * ((-BWeyl_xx * mmI_xx) + EWeyl_xx * mmR_xx + (-BWeyl_yx * mmI_xy) + 2.0 * EWeyl_xy * mmR_xy + (-BWeyl_zx * mmI_xz) + 2.0 * EWeyl_xz * mmR_xz + (-BWeyl_xy * mmI_xy) + (-BWeyl_yy * mmI_yy) + EWeyl_yy * mmR_yy + (-BWeyl_zy * mmI_yz) + 2.0 * EWeyl_yz * mmR_yz + (-BWeyl_xz * mmI_xz) + (-BWeyl_yz * mmI_yz) + (-BWeyl_zz * mmI_zz) + EWeyl_zz * mmR_zz);
							d_psi4I_o0_t0_m0_l0 = MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * (BWeyl_xx * mmR_xx + EWeyl_xx * mmI_xx + BWeyl_yx * mmR_xy + 2.0 * EWeyl_xy * mmI_xy + BWeyl_zx * mmR_xz + 2.0 * EWeyl_xz * mmI_xz + BWeyl_xy * mmR_xy + BWeyl_yy * mmR_yy + EWeyl_yy * mmI_yy + BWeyl_zy * mmR_yz + 2.0 * EWeyl_yz * mmI_yz + BWeyl_xz * mmR_xz + BWeyl_yz * mmR_yz + BWeyl_zz * mmR_zz + EWeyl_zz * mmI_zz);
							d_Z_x_o0_t0_m0_l0 = Zu_x;
							d_Z_y_o0_t0_m0_l0 = Zu_y;
							d_Z_z_o0_t0_m0_l0 = Zu_z;
							d_flux_mass_o0_t0_m0_l0 = (Ju_x * xcoord(i) + Ju_y * ycoord(j) + Ju_z * zcoord(k)) * r_pos * SIGN(MAX(vector(Alpha, i, j, k) * hf * sqrt(sqW) - 1.0, 0.0));
							m_M_ADM_surf_o0_t8_l0 = 0.125 * (gtu_zz * dSigma_z) / M_PI * d_Rpd_xx_o0_t2_m0_l0;
							m_M_ADM_surf_o0_t7_l0 = 0.125 * (gtu_yz * dSigma_z) / M_PI * d_Rpd_xx_o0_t1_m0_l0;
							m_M_ADM_surf_o0_t6_l0 = 0.125 * (gtu_xz * dSigma_z) / M_PI * d_Rpd_xx_o0_t0_m0_l0;
							m_M_ADM_surf_o0_t5_l0 = 0.125 * (gtu_yz * dSigma_y) / M_PI * d_Rpd_xx_o0_t2_m0_l0;
							m_M_ADM_surf_o0_t4_l0 = 0.125 * (gtu_yy * dSigma_y) / M_PI * d_Rpd_xx_o0_t1_m0_l0;
							m_M_ADM_surf_o0_t3_l0 = 0.125 * (gtu_xy * dSigma_y) / M_PI * d_Rpd_xx_o0_t0_m0_l0;
							m_M_ADM_surf_o0_t2_l0 = 0.125 * (gtu_xz * dSigma_x) / M_PI * d_Rpd_xx_o0_t2_m0_l0;
							m_M_ADM_surf_o0_t1_l0 = 0.125 * (gtu_xy * dSigma_x) / M_PI * d_Rpd_xx_o0_t1_m0_l0;
							m_M_ADM_surf_o0_t0_l0 = 0.125 * (gtu_xx * dSigma_x) / M_PI * d_Rpd_xx_o0_t0_m0_l0;
							m_MomCon_z_o0_t12_l0 = (-0.6666666666666667) * d_Del_Kd_zxx_o0_t2_m0_l0;
							m_MomCon_z_o0_t11_l0 = gtu_zz * d_MomCon_z_o0_t11_m0_l0;
							m_MomCon_z_o0_t10_l0 = gtu_yz * d_Del_Kd_zyz_o0_t0_m0_l0;
							m_MomCon_z_o0_t9_l0 = gtu_xz * d_Del_Kd_zxz_o0_t0_m0_l0;
							m_MomCon_z_o0_t8_l0 = gtu_yz * d_Del_Kd_yzz_o0_t0_m0_l0;
							m_MomCon_z_o0_t7_l0 = gtu_yy * d_Del_Kd_yyz_o0_t0_m0_l0;
							m_MomCon_z_o0_t6_l0 = gtu_xy * d_Del_Kd_yxz_o0_t0_m0_l0;
							m_MomCon_z_o0_t5_l0 = gtu_xz * d_Del_Kd_xzz_o0_t0_m0_l0;
							m_MomCon_z_o0_t4_l0 = gtu_xy * d_Del_Kd_xyz_o0_t0_m0_l0;
							m_MomCon_z_o0_t3_l0 = gtu_xx * d_Del_Kd_xxz_o0_t0_m0_l0;
							m_MomCon_z_o0_t2_l0 = (-1.500000000000000 * inv_chi * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
							m_MomCon_z_o0_t1_l0 = (-1.500000000000000 * inv_chi * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
							m_MomCon_z_o0_t0_l0 = (-1.500000000000000 * inv_chi * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
							m_MomCon_y_o0_t12_l0 = (-0.6666666666666667) * d_Del_Kd_yxx_o0_t2_m0_l0;
							m_MomCon_y_o0_t11_l0 = (-1.500000000000000 * inv_chi * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
							m_MomCon_y_o0_t10_l0 = (-1.500000000000000 * inv_chi * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
							m_MomCon_y_o0_t9_l0 = (-1.500000000000000 * inv_chi * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
							m_MomCon_y_o0_t8_l0 = gtu_zz * d_Del_Kd_zyz_o0_t0_m0_l0;
							m_MomCon_y_o0_t7_l0 = gtu_yz * d_Del_Kd_zyy_o0_t0_m0_l0;
							m_MomCon_y_o0_t6_l0 = gtu_xz * d_Del_Kd_zxy_o0_t0_m0_l0;
							m_MomCon_y_o0_t5_l0 = gtu_yz * d_Del_Kd_yyz_o0_t0_m0_l0;
							m_MomCon_y_o0_t4_l0 = gtu_yy * d_MomCon_y_o0_t4_m0_l0;
							m_MomCon_y_o0_t3_l0 = gtu_xy * d_Del_Kd_yxy_o0_t0_m0_l0;
							m_MomCon_y_o0_t2_l0 = gtu_xz * d_Del_Kd_xyz_o0_t0_m0_l0;
							m_MomCon_y_o0_t1_l0 = gtu_xy * d_Del_Kd_xyy_o0_t0_m0_l0;
							m_MomCon_y_o0_t0_l0 = gtu_xx * d_Del_Kd_xxy_o0_t0_m0_l0;
							m_MomCon_x_o0_t12_l0 = (-0.6666666666666667) * d_Del_Kd_xxy_o0_t2_m0_l0;
							m_MomCon_x_o0_t11_l0 = (-1.500000000000000 * inv_chi * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
							m_MomCon_x_o0_t10_l0 = (-1.500000000000000 * inv_chi * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
							m_MomCon_x_o0_t9_l0 = (-1.500000000000000 * inv_chi * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
							m_MomCon_x_o0_t8_l0 = gtu_zz * d_Del_Kd_zxz_o0_t0_m0_l0;
							m_MomCon_x_o0_t7_l0 = gtu_yz * d_Del_Kd_zxy_o0_t0_m0_l0;
							m_MomCon_x_o0_t6_l0 = gtu_xz * d_Del_Kd_zxx_o0_t0_m0_l0;
							m_MomCon_x_o0_t5_l0 = gtu_yz * d_Del_Kd_yxz_o0_t0_m0_l0;
							m_MomCon_x_o0_t4_l0 = gtu_yy * d_Del_Kd_yxy_o0_t0_m0_l0;
							m_MomCon_x_o0_t3_l0 = gtu_xy * d_Del_Kd_yxx_o0_t0_m0_l0;
							m_MomCon_x_o0_t2_l0 = gtu_xz * d_Del_Kd_xxz_o0_t0_m0_l0;
							m_MomCon_x_o0_t1_l0 = gtu_xy * d_Del_Kd_xxy_o0_t0_m0_l0;
							m_MomCon_x_o0_t0_l0 = gtu_xx * d_MomCon_x_o0_t0_m0_l0;
							vector(Rscalar, i, j, k) = d_Rscalar_o0_t0_m0_l0;
							vector(HamCon, i, j, k) = d_HamCon_o0_t0_m0_l0;
							vector(MomCon_x, i, j, k) = ((((((((((((m_MomCon_x_o0_t0_l0 + m_MomCon_x_o0_t1_l0) + m_MomCon_x_o0_t2_l0) + m_MomCon_x_o0_t3_l0) + m_MomCon_x_o0_t4_l0) + m_MomCon_x_o0_t5_l0) + m_MomCon_x_o0_t6_l0) + m_MomCon_x_o0_t7_l0) + m_MomCon_x_o0_t8_l0) + m_MomCon_x_o0_t9_l0) + m_MomCon_x_o0_t10_l0) + m_MomCon_x_o0_t11_l0) + m_MomCon_x_o0_t12_l0) + d_MomCon_x_o0_t13_m0_l0;
							vector(MomCon_y, i, j, k) = ((((((((((((m_MomCon_y_o0_t0_l0 + m_MomCon_y_o0_t1_l0) + m_MomCon_y_o0_t2_l0) + m_MomCon_y_o0_t3_l0) + m_MomCon_y_o0_t4_l0) + m_MomCon_y_o0_t5_l0) + m_MomCon_y_o0_t6_l0) + m_MomCon_y_o0_t7_l0) + m_MomCon_y_o0_t8_l0) + m_MomCon_y_o0_t9_l0) + m_MomCon_y_o0_t10_l0) + m_MomCon_y_o0_t11_l0) + m_MomCon_y_o0_t12_l0) + d_MomCon_y_o0_t13_m0_l0;
							vector(MomCon_z, i, j, k) = ((((((((((((m_MomCon_z_o0_t0_l0 + m_MomCon_z_o0_t1_l0) + m_MomCon_z_o0_t2_l0) + m_MomCon_z_o0_t3_l0) + m_MomCon_z_o0_t4_l0) + m_MomCon_z_o0_t5_l0) + m_MomCon_z_o0_t6_l0) + m_MomCon_z_o0_t7_l0) + m_MomCon_z_o0_t8_l0) + m_MomCon_z_o0_t9_l0) + m_MomCon_z_o0_t10_l0) + m_MomCon_z_o0_t11_l0) + m_MomCon_z_o0_t12_l0) + d_MomCon_z_o0_t13_m0_l0;
							vector(trA, i, j, k) = d_trA_o0_t0_m0_l0;
							vector(detgtm1, i, j, k) = d_detgtm1_o0_t0_m0_l0;
							vector(M_ADM_surf, i, j, k) = ((((((((m_M_ADM_surf_o0_t0_l0 + m_M_ADM_surf_o0_t1_l0) + m_M_ADM_surf_o0_t2_l0) + m_M_ADM_surf_o0_t3_l0) + m_M_ADM_surf_o0_t4_l0) + m_M_ADM_surf_o0_t5_l0) + m_M_ADM_surf_o0_t6_l0) + m_M_ADM_surf_o0_t7_l0) + m_M_ADM_surf_o0_t8_l0) + d_M_ADM_surf_o0_t9_m0_l0;
							vector(Jz_ADM_surf, i, j, k) = d_Jz_ADM_surf_o0_t0_m0_l0;
							vector(M_Komar, i, j, k) = d_M_Komar_o0_t0_m0_l0;
							vector(Jz_Komar, i, j, k) = d_Jz_Komar_o0_t0_m0_l0;
							vector(psi4R, i, j, k) = d_psi4R_o0_t0_m0_l0;
							vector(psi4I, i, j, k) = d_psi4I_o0_t0_m0_l0;
							vector(Z_x, i, j, k) = d_Z_x_o0_t0_m0_l0;
							vector(Z_y, i, j, k) = d_Z_y_o0_t0_m0_l0;
							vector(Z_z, i, j, k) = d_Z_z_o0_t0_m0_l0;
							vector(flux_mass, i, j, k) = d_flux_mass_o0_t0_m0_l0;
						}
					}
				}
			}
		}
	}
	
}


void Problem::initializeLevelIntegrator(
   const std::shared_ptr<mesh::GriddingAlgorithmStrategy>& gridding_alg)
{
}

double Problem::getLevelDt(
   const std::shared_ptr<hier::PatchLevel>& level,
   const double dt_time,
   const bool initial_time)
{
  
   TBOX_ASSERT(level);

   if (level->getLevelNumber() == 0) return initial_dt;

    double dt = initial_dt;
    const hier::IntVector ratio = level->getRatioToLevelZero();
    double local_dt = dt;
    for (int i = 0; i < 2; i++) {
        if (local_dt > dt / ratio[i]) {
            local_dt = dt / ratio[i];
        }
    }
    return local_dt;
}

double Problem::getMaxFinerLevelDt(
   const int finer_level_number,
   const double coarse_dt,
   const hier::IntVector& ratio)
{
   NULL_USE(finer_level_number);

   TBOX_ASSERT(ratio.min() > 0);

   return coarse_dt / double(ratio.max());
}

void Problem::standardLevelSynchronization(
   const std::shared_ptr<hier::PatchHierarchy>& hierarchy,
   const int coarsest_level,
   const int finest_level,
   const double sync_time,
   const std::vector<double>& old_times)
{

}

void Problem::synchronizeNewLevels(
   const std::shared_ptr<hier::PatchHierarchy>& hierarchy,
   const int coarsest_level,
   const int finest_level,
   const double sync_time,
   const bool initial_time)
{

    //Not needed, but not absolutely sure
    /*for (int fine_ln = finest_level; fine_ln > coarsest_level; --fine_ln) {
        const int coarse_ln = fine_ln - 1;
        std::shared_ptr<hier::PatchLevel> fine_level(hierarchy->getPatchLevel(fine_ln));
        d_bdry_fill_init->createSchedule(fine_level, coarse_ln, hierarchy, this)->fillData(sync_time, true);
    }*/
}

void Problem::resetTimeDependentData(
   const std::shared_ptr<hier::PatchLevel>& level,
   const double new_time,
   const bool can_be_refined)
{
   TBOX_ASSERT(level);
   level->setTime(new_time);
}

void Problem::resetDataToPreadvanceState(
   const std::shared_ptr<hier::PatchLevel>& level)
{
    //cout<<"resetDataToPreadvanceState"<<endl;
}

/*
 * Map data on a patch. This mapping is done only at the begining of the simulation.
 */
void Problem::mapDataOnPatch(const double time, const bool initial_time, const int ln, const std::shared_ptr< hier::PatchLevel >& level)
{
	(void) time;
	const tbox::SAMRAI_MPI& mpi(tbox::SAMRAI_MPI::getSAMRAIWorld());
   	if (initial_time || ln == 0) {

		// Mapping		
		int i, iterm, previousMapi, iWallAcc;
		bool interiorMapi;
		double iMapStart, iMapEnd;
		int j, jterm, previousMapj, jWallAcc;
		bool interiorMapj;
		double jMapStart, jMapEnd;
		int k, kterm, previousMapk, kWallAcc;
		bool interiorMapk;
		double kMapStart, kMapEnd;
		int minBlock[3], maxBlock[3], unionsI, facePointI, ie1, ie2, ie3, proc, pcounter, working, finished, pred;
		double maxDistance, e1, e2, e3;
		bool done, modif, workingArray[mpi.getSize()], finishedArray[mpi.getSize()], workingGlobal, finishedGlobal, workingPatchArray[level->getLocalNumberOfPatches()], finishedPatchArray[level->getLocalNumberOfPatches()], workingPatchGlobal, finishedPatchGlobal;
		int nodes = mpi.getSize();
		int patches = level->getLocalNumberOfPatches();

		double SQRT3INV = 1.0/sqrt(3.0);

		if (ln == 0) {
			//FOV initialization
			for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
				const std::shared_ptr< hier::Patch >& patch = *p_it;

				//Get the dimensions of the patch
				hier::Box pbox = patch->getBox();
				double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
				double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
				double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
				double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
				double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
				double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
				double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
				const hier::Index boxfirst = patch->getBox().lower();
				const hier::Index boxlast  = patch->getBox().upper();

				//Get delta spaces into an array. dx, dy, dz.
				const std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
				const double* dx  = patch_geom->getDx();

				int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
				int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
				int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;

				for (i = 0; i < ilast; i++) {
					for (j = 0; j < jlast; j++) {
						for (k = 0; k < klast; k++) {
							vector(FOV_1, i, j, k) = 0;
							vector(FOV_xLower, i, j, k) = 0;
							vector(FOV_xUpper, i, j, k) = 0;
							vector(FOV_yLower, i, j, k) = 0;
							vector(FOV_yUpper, i, j, k) = 0;
							vector(FOV_zLower, i, j, k) = 0;
							vector(FOV_zUpper, i, j, k) = 0;
						}
					}
				}
			}
			d_mapping_fill->createSchedule(level, level)->fillData(initial_time, true);

			//Region: mainI
			for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
				const std::shared_ptr< hier::Patch >& patch = *p_it;

				//Get the dimensions of the patch
				hier::Box pbox = patch->getBox();
				double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
				double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
				double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
				double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
				double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
				double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
				double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
				double* interior_i = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_i_id).get())->getPointer();
				double* interior_j = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_j_id).get())->getPointer();
				double* interior_k = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_k_id).get())->getPointer();
				double* interior = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_regridding_value_id).get())->getPointer();
				int* nonSync = ((pdat::NodeData<int> *) patch->getPatchData(d_nonSync_regridding_tag_id).get())->getPointer();
				const hier::Index boxfirst = patch->getBox().lower();
				const hier::Index boxlast  = patch->getBox().upper();

				//Get delta spaces into an array. dx, dy, dz.
				const std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
				const double* dx  = patch_geom->getDx();

				int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
				int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
				int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;

				iMapStart = 0;
				iMapEnd = d_grid_geometry->getPhysicalDomain().front().numberCells()[0];
				jMapStart = 0;
				jMapEnd = d_grid_geometry->getPhysicalDomain().front().numberCells()[1];
				kMapStart = 0;
				kMapEnd = d_grid_geometry->getPhysicalDomain().front().numberCells()[2];
				for (i = iMapStart; i <= iMapEnd; i++) {
					for (j = jMapStart; j <= jMapEnd; j++) {
						for (k = kMapStart; k <= kMapEnd; k++) {
							if (i >= boxfirst(0) - d_ghost_width && i <= boxlast(0) + d_ghost_width && j >= boxfirst(1) - d_ghost_width && j <= boxlast(1) + d_ghost_width && k >= boxfirst(2) - d_ghost_width && k <= boxlast(2) + d_ghost_width) {
								vector(FOV_1, i - boxfirst(0) + d_ghost_width, j - boxfirst(1) + d_ghost_width, k - boxfirst(2) + d_ghost_width) = 100;
								vector(FOV_xLower, i - boxfirst(0) + d_ghost_width, j - boxfirst(1) + d_ghost_width, k - boxfirst(2) + d_ghost_width) = 0;
								vector(FOV_xUpper, i - boxfirst(0) + d_ghost_width, j - boxfirst(1) + d_ghost_width, k - boxfirst(2) + d_ghost_width) = 0;
								vector(FOV_yLower, i - boxfirst(0) + d_ghost_width, j - boxfirst(1) + d_ghost_width, k - boxfirst(2) + d_ghost_width) = 0;
								vector(FOV_yUpper, i - boxfirst(0) + d_ghost_width, j - boxfirst(1) + d_ghost_width, k - boxfirst(2) + d_ghost_width) = 0;
								vector(FOV_zLower, i - boxfirst(0) + d_ghost_width, j - boxfirst(1) + d_ghost_width, k - boxfirst(2) + d_ghost_width) = 0;
								vector(FOV_zUpper, i - boxfirst(0) + d_ghost_width, j - boxfirst(1) + d_ghost_width, k - boxfirst(2) + d_ghost_width) = 0;
							}
						}
					}
				}
				//Check stencil
				for (k = 0; k < klast; k++) {
					for (j = 0; j < jlast; j++) {
						for (i = 0; i < ilast; i++) {
							vector(interior_i, i, j, k) = 0;
							vector(interior_j, i, j, k) = 0;
							vector(interior_k, i, j, k) = 0;
						}
					}
				}
				for (k = 0; k < klast; k++) {
					for (j = 0; j < jlast; j++) {
						for (i = 0; i < ilast; i++) {
							if (vector(FOV_1, i, j, k) > 0) {
								setStencilLimits(patch, i, j, k, d_FOV_1_id);
							}
						}
					}
				}
			}
			d_mapping_fill->createSchedule(level, level)->fillData(initial_time, true);
			for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
				const std::shared_ptr< hier::Patch >& patch = *p_it;

				//Get the dimensions of the patch
				hier::Box pbox = patch->getBox();
				double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
				double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
				double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
				double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
				double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
				double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
				double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
				double* interior_i = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_i_id).get())->getPointer();
				double* interior_j = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_j_id).get())->getPointer();
				double* interior_k = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_k_id).get())->getPointer();
				double* interior = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_regridding_value_id).get())->getPointer();
				int* nonSync = ((pdat::NodeData<int> *) patch->getPatchData(d_nonSync_regridding_tag_id).get())->getPointer();
				const hier::Index boxfirst = patch->getBox().lower();
				const hier::Index boxlast  = patch->getBox().upper();

				//Get delta spaces into an array. dx, dy, dz.
				const std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
				const double* dx  = patch_geom->getDx();

				int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
				int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
				int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;

				for (k = 0; k < klast; k++) {
					for (j = 0; j < jlast; j++) {
						for (i = 0; i < ilast; i++) {
							if ((abs(vector(interior_i, i, j, k)) > 1 || abs(vector(interior_j, i, j, k)) > 1 || abs(vector(interior_k, i, j, k)) > 1) && (i < d_ghost_width || i >= ilast - d_ghost_width || j < d_ghost_width || j >= jlast - d_ghost_width || k < d_ghost_width || k >= klast - d_ghost_width)) {
								checkStencil(patch, i, j, k, d_FOV_1_id);
							}
						}
					}
				}
				for (k = 0; k < klast; k++) {
					for (j = 0; j < jlast; j++) {
						for (i = 0; i < ilast; i++) {
							if (vector(interior_i, i, j, k) != 0 || vector(interior_j, i, j, k) != 0 || vector(interior_k, i, j, k) != 0) {
								vector(FOV_1, i, j, k) = 100;
								vector(FOV_xLower, i, j, k) = 0;
								vector(FOV_xUpper, i, j, k) = 0;
								vector(FOV_yLower, i, j, k) = 0;
								vector(FOV_yUpper, i, j, k) = 0;
								vector(FOV_zLower, i, j, k) = 0;
								vector(FOV_zUpper, i, j, k) = 0;
							}
						}
					}
				}
			}
			d_mapping_fill->createSchedule(level, level)->fillData(initial_time, true);

		}
		//Boundaries Mapping
		for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
			const std::shared_ptr< hier::Patch >& patch = *p_it;

			//Get the dimensions of the patch
			hier::Box pbox = patch->getBox();
			double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
			double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
			double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
			double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
			double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
			double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
			double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
			const hier::Index boxfirst = patch->getBox().lower();
			const hier::Index boxlast  = patch->getBox().upper();

			//Get delta spaces into an array. dx, dy, dz.
			const std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
			const double* dx  = patch_geom->getDx();

			int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
			int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
			int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;

			//z-Upper
			if (patch->getPatchGeometry()->getTouchesRegularBoundary(2, 1)) {
				for (k = klast - d_ghost_width; k < klast; k++) {
					for (j = 0; j < jlast; j++) {
						for (i = 0; i < ilast; i++) {
							vector(FOV_zUpper, i, j, k) = 100;
							vector(FOV_1, i, j, k) = 0;
							vector(FOV_xLower, i, j, k) = 0;
							vector(FOV_xUpper, i, j, k) = 0;
							vector(FOV_yLower, i, j, k) = 0;
							vector(FOV_yUpper, i, j, k) = 0;
							vector(FOV_zLower, i, j, k) = 0;
						}
					}
				}
			}
			//z-Lower
			if (patch->getPatchGeometry()->getTouchesRegularBoundary(2, 0)) {
				for (k = 0; k < d_ghost_width; k++) {
					for (j = 0; j < jlast; j++) {
						for (i = 0; i < ilast; i++) {
							vector(FOV_zLower, i, j, k) = 100;
							vector(FOV_1, i, j, k) = 0;
							vector(FOV_xLower, i, j, k) = 0;
							vector(FOV_xUpper, i, j, k) = 0;
							vector(FOV_yLower, i, j, k) = 0;
							vector(FOV_yUpper, i, j, k) = 0;
							vector(FOV_zUpper, i, j, k) = 0;
						}
					}
				}
			}
			//y-Upper
			if (patch->getPatchGeometry()->getTouchesRegularBoundary(1, 1)) {
				for (k = 0; k < klast; k++) {
					for (j = jlast - d_ghost_width; j < jlast; j++) {
						for (i = 0; i < ilast; i++) {
							vector(FOV_yUpper, i, j, k) = 100;
							vector(FOV_1, i, j, k) = 0;
							vector(FOV_xLower, i, j, k) = 0;
							vector(FOV_xUpper, i, j, k) = 0;
							vector(FOV_yLower, i, j, k) = 0;
							vector(FOV_zLower, i, j, k) = 0;
							vector(FOV_zUpper, i, j, k) = 0;
						}
					}
				}
			}
			//y-Lower
			if (patch->getPatchGeometry()->getTouchesRegularBoundary(1, 0)) {
				for (k = 0; k < klast; k++) {
					for (j = 0; j < d_ghost_width; j++) {
						for (i = 0; i < ilast; i++) {
							vector(FOV_yLower, i, j, k) = 100;
							vector(FOV_1, i, j, k) = 0;
							vector(FOV_xLower, i, j, k) = 0;
							vector(FOV_xUpper, i, j, k) = 0;
							vector(FOV_yUpper, i, j, k) = 0;
							vector(FOV_zLower, i, j, k) = 0;
							vector(FOV_zUpper, i, j, k) = 0;
						}
					}
				}
			}
			//x-Upper
			if (patch->getPatchGeometry()->getTouchesRegularBoundary(0, 1)) {
				for (k = 0; k < klast; k++) {
					for (j = 0; j < jlast; j++) {
						for (i = ilast - d_ghost_width; i < ilast; i++) {
							vector(FOV_xUpper, i, j, k) = 100;
							vector(FOV_1, i, j, k) = 0;
							vector(FOV_xLower, i, j, k) = 0;
							vector(FOV_yLower, i, j, k) = 0;
							vector(FOV_yUpper, i, j, k) = 0;
							vector(FOV_zLower, i, j, k) = 0;
							vector(FOV_zUpper, i, j, k) = 0;
						}
					}
				}
			}
			//x-Lower
			if (patch->getPatchGeometry()->getTouchesRegularBoundary(0, 0)) {
				for (k = 0; k < klast; k++) {
					for (j = 0; j < jlast; j++) {
						for (i = 0; i < d_ghost_width; i++) {
							vector(FOV_xLower, i, j, k) = 100;
							vector(FOV_1, i, j, k) = 0;
							vector(FOV_xUpper, i, j, k) = 0;
							vector(FOV_yLower, i, j, k) = 0;
							vector(FOV_yUpper, i, j, k) = 0;
							vector(FOV_zLower, i, j, k) = 0;
							vector(FOV_zUpper, i, j, k) = 0;
						}
					}
				}
			}
		}
		d_mapping_fill->createSchedule(level, level)->fillData(initial_time, true);




   	}
}


/*
 * Sets the limit for the checkstencil routine
 */
void Problem::setStencilLimits(std::shared_ptr< hier::Patch > patch, int i, int j, int k, int v) const {
	double* FOV = ((pdat::NodeData<double> *) patch->getPatchData(v).get())->getPointer();
	double* interior_i = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_i_id).get())->getPointer();
	double* interior_j = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_j_id).get())->getPointer();
	double* interior_k = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_k_id).get())->getPointer();
	//Get the dimensions of the patch
	const hier::Index boxfirst = patch->getBox().lower();
	const hier::Index boxlast  = patch->getBox().upper();
	//Auxiliary definitions
	int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
	int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
	int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;

	int iStart, iEnd, currentGhosti, otherSideShifti, jStart, jEnd, currentGhostj, otherSideShiftj, kStart, kEnd, currentGhostk, otherSideShiftk, shift;

	currentGhosti = d_ghost_width - 1;
	otherSideShifti = 0;
	currentGhostj = d_ghost_width - 1;
	otherSideShiftj = 0;
	currentGhostk = d_ghost_width - 1;
	otherSideShiftk = 0;
	//Checking width
	if ((i + 1 < ilast && vector(FOV, i + 1, j, k) == 0) ||  (i - 1 >= 0 && vector(FOV, i - 1, j, k) == 0)) {
		if (i + 1 < ilast && vector(FOV, i + 1, j, k) > 0) {
			bool stop_counting = false;
			for(int iti = i + 1; iti <= i + d_ghost_width - 1 && currentGhosti > 0; iti++) {
				if (iti < ilast  && vector(FOV, iti, j, k) > 0 && stop_counting == false) {
					currentGhosti--;
				} else {
					//First not interior point found
					if (iti < ilast  && vector(FOV, iti, j, k) == 0) {
						stop_counting = true;
					}
					//Physical boundary reach
					if (iti >= ilast - 3 && patch->getPatchGeometry()->getTouchesRegularBoundary (0, 1)) {
						stop_counting = true;
						//Calculate the number of cells the limit cannot grow
						if (iti + currentGhosti/2 >= ilast - 3) {
							otherSideShifti = (iti  + currentGhosti/2) - (ilast - 3);
						}
					}
				}
			}
		}
		if (i - 1 >= 0 && vector(FOV, i - 1, j, k) > 0) {
			bool stop_counting = false;
			for(int iti = i - 1; iti >= i - d_ghost_width + 1 && currentGhosti > 0; iti--) {
				if (iti >= 0  && vector(FOV, iti, j, k) > 0) {
					currentGhosti--;
				} else {
					//First not interior point found
					if (iti >= 0 && vector(FOV, iti, j, k) == 0) {
						stop_counting = true;
					}
					//Physical boundary reach
					if (iti < 3 && patch->getPatchGeometry()->getTouchesRegularBoundary (0, 0)) {
						stop_counting = true;
						//calculate the number of cells the limit cannot grow
						if (iti  -  ((currentGhosti) - currentGhosti/2) < 3) {
							otherSideShifti = 3 - (iti  - ((currentGhosti) - currentGhosti/2));
						}
					}
				}
			}
		}
		if (currentGhosti > 0) {
			if (i + 1 < ilast && vector(FOV, i + 1, j, k) > 0) {
				shift = 0;
				if(patch->getPatchGeometry()->getTouchesRegularBoundary (0, 0)) {
					while(i - ((currentGhosti) - currentGhosti/2) + shift < 3) {
						shift++;
					}
				}
				iStart = (currentGhosti - currentGhosti/2) - shift - otherSideShifti;
				iEnd = 0;
			} else {
				if (i - 1 >= 0 && vector(FOV, i - 1, j, k) > 0) {
					shift = 0;
					if(patch->getPatchGeometry()->getTouchesRegularBoundary (0, 1)) {
						while(i + currentGhosti/2 + shift >= ilast - 3) {
							shift--;
						}
					}
					iStart = 0;
					iEnd = currentGhosti/2 + shift + otherSideShifti;
				} else {
					shift = 0;
					if(patch->getPatchGeometry()->getTouchesRegularBoundary (0, 0)) {
						while(i - ((currentGhosti) - currentGhosti/2) + shift < 3) {
							shift++;
						}
					}
					if(patch->getPatchGeometry()->getTouchesRegularBoundary (0, 1)) {
						while(i + currentGhosti/2 + shift >= ilast - 3) {
							shift--;
						}
					}
					iStart = (currentGhosti - currentGhosti/2) - shift;
					iEnd = currentGhosti/2 + shift;
				}
			}
		} else {
			iStart = 0;
			iEnd = 0;
		}
	} else {
		iStart = 0;
		iEnd = 0;
	}
	if ((j + 1 < jlast && vector(FOV, i, j + 1, k) == 0) ||  (j - 1 >= 0 && vector(FOV, i, j - 1, k) == 0)) {
		if (j + 1 < jlast && vector(FOV, i, j + 1, k) > 0) {
			bool stop_counting = false;
			for(int itj = j + 1; itj <= j + d_ghost_width - 1 && currentGhostj > 0; itj++) {
				if (itj < jlast  && vector(FOV, i, itj, k) > 0 && stop_counting == false) {
					currentGhostj--;
				} else {
					//First not interior point found
					if (itj < jlast  && vector(FOV, i, itj, k) == 0) {
						stop_counting = true;
					}
					//Physical boundary reach
					if (itj >= jlast - 3 && patch->getPatchGeometry()->getTouchesRegularBoundary (1, 1)) {
						stop_counting = true;
						//Calculate the number of cells the limit cannot grow
						if (itj + currentGhostj/2 >= jlast - 3) {
							otherSideShiftj = (itj  + currentGhostj/2) - (jlast - 3);
						}
					}
				}
			}
		}
		if (j - 1 >= 0 && vector(FOV, i, j - 1, k) > 0) {
			bool stop_counting = false;
			for(int itj = j - 1; itj >= j - d_ghost_width + 1 && currentGhostj > 0; itj--) {
				if (itj >= 0  && vector(FOV, i, itj, k) > 0) {
					currentGhostj--;
				} else {
					//First not interior point found
					if (itj >= 0 && vector(FOV, i, itj, k) == 0) {
						stop_counting = true;
					}
					//Physical boundary reach
					if (itj < 3 && patch->getPatchGeometry()->getTouchesRegularBoundary (1, 0)) {
						stop_counting = true;
						//calculate the number of cells the limit cannot grow
						if (itj  -  ((currentGhostj) - currentGhostj/2) < 3) {
							otherSideShiftj = 3 - (itj  - ((currentGhostj) - currentGhostj/2));
						}
					}
				}
			}
		}
		if (currentGhostj > 0) {
			if (j + 1 < jlast && vector(FOV, i, j + 1, k) > 0) {
				shift = 0;
				if(patch->getPatchGeometry()->getTouchesRegularBoundary (1, 0)) {
					while(j - ((currentGhostj) - currentGhostj/2) + shift < 3) {
						shift++;
					}
				}
				jStart = (currentGhostj - currentGhostj/2) - shift - otherSideShiftj;
				jEnd = 0;
			} else {
				if (j - 1 >= 0 && vector(FOV, i, j - 1, k) > 0) {
					shift = 0;
					if(patch->getPatchGeometry()->getTouchesRegularBoundary (1, 1)) {
						while(j + currentGhostj/2 + shift >= jlast - 3) {
							shift--;
						}
					}
					jStart = 0;
					jEnd = currentGhostj/2 + shift + otherSideShiftj;
				} else {
					shift = 0;
					if(patch->getPatchGeometry()->getTouchesRegularBoundary (1, 0)) {
						while(j - ((currentGhostj) - currentGhostj/2) + shift < 3) {
							shift++;
						}
					}
					if(patch->getPatchGeometry()->getTouchesRegularBoundary (1, 1)) {
						while(j + currentGhostj/2 + shift >= jlast - 3) {
							shift--;
						}
					}
					jStart = (currentGhostj - currentGhostj/2) - shift;
					jEnd = currentGhostj/2 + shift;
				}
			}
		} else {
			jStart = 0;
			jEnd = 0;
		}
	} else {
		jStart = 0;
		jEnd = 0;
	}
	if ((k + 1 < klast && vector(FOV, i, j, k + 1) == 0) ||  (k - 1 >= 0 && vector(FOV, i, j, k - 1) == 0)) {
		if (k + 1 < klast && vector(FOV, i, j, k + 1) > 0) {
			bool stop_counting = false;
			for(int itk = k + 1; itk <= k + d_ghost_width - 1 && currentGhostk > 0; itk++) {
				if (itk < klast  && vector(FOV, i, j, itk) > 0 && stop_counting == false) {
					currentGhostk--;
				} else {
					//First not interior point found
					if (itk < klast  && vector(FOV, i, j, itk) == 0) {
						stop_counting = true;
					}
					//Physical boundary reach
					if (itk >= klast - 3 && patch->getPatchGeometry()->getTouchesRegularBoundary (2, 1)) {
						stop_counting = true;
						//Calculate the number of cells the limit cannot grow
						if (itk + currentGhostk/2 >= klast - 3) {
							otherSideShiftk = (itk  + currentGhostk/2) - (klast - 3);
						}
					}
				}
			}
		}
		if (k - 1 >= 0 && vector(FOV, i, j, k - 1) > 0) {
			bool stop_counting = false;
			for(int itk = k - 1; itk >= k - d_ghost_width + 1 && currentGhostk > 0; itk--) {
				if (itk >= 0  && vector(FOV, i, j, itk) > 0) {
					currentGhostk--;
				} else {
					//First not interior point found
					if (itk >= 0 && vector(FOV, i, j, itk) == 0) {
						stop_counting = true;
					}
					//Physical boundary reach
					if (itk < 3 && patch->getPatchGeometry()->getTouchesRegularBoundary (2, 0)) {
						stop_counting = true;
						//calculate the number of cells the limit cannot grow
						if (itk  -  ((currentGhostk) - currentGhostk/2) < 3) {
							otherSideShiftk = 3 - (itk  - ((currentGhostk) - currentGhostk/2));
						}
					}
				}
			}
		}
		if (currentGhostk > 0) {
			if (k + 1 < klast && vector(FOV, i, j, k + 1) > 0) {
				shift = 0;
				if(patch->getPatchGeometry()->getTouchesRegularBoundary (2, 0)) {
					while(k - ((currentGhostk) - currentGhostk/2) + shift < 3) {
						shift++;
					}
				}
				kStart = (currentGhostk - currentGhostk/2) - shift - otherSideShiftk;
				kEnd = 0;
			} else {
				if (k - 1 >= 0 && vector(FOV, i, j, k - 1) > 0) {
					shift = 0;
					if(patch->getPatchGeometry()->getTouchesRegularBoundary (2, 1)) {
						while(k + currentGhostk/2 + shift >= klast - 3) {
							shift--;
						}
					}
					kStart = 0;
					kEnd = currentGhostk/2 + shift + otherSideShiftk;
				} else {
					shift = 0;
					if(patch->getPatchGeometry()->getTouchesRegularBoundary (2, 0)) {
						while(k - ((currentGhostk) - currentGhostk/2) + shift < 3) {
							shift++;
						}
					}
					if(patch->getPatchGeometry()->getTouchesRegularBoundary (2, 1)) {
						while(k + currentGhostk/2 + shift >= klast - 3) {
							shift--;
						}
					}
					kStart = (currentGhostk - currentGhostk/2) - shift;
					kEnd = currentGhostk/2 + shift;
				}
			}
		} else {
			kStart = 0;
			kEnd = 0;
		}
	} else {
		kStart = 0;
		kEnd = 0;
	}
	//Assigning stencil limits
	for(int iti = i - iStart; iti <= i + iEnd; iti++) {
		for(int itj = j - jStart; itj <= j + jEnd; itj++) {
			for(int itk = k - kStart; itk <= k + kEnd; itk++) {
				if(iti >= 0 && iti < ilast && itj >= 0 && itj < jlast && itk >= 0 && itk < klast && vector(FOV, iti, itj, itk) == 0) {
					if (i - iti < 0) {
						vector(interior_i, iti, itj, itk) = - (iStart + 1) - (i - iti);
					} else {
						vector(interior_i, iti, itj, itk) = (iStart + 1) - (i - iti);
					}
					if (j - itj < 0) {
						vector(interior_j, iti, itj, itk) = - (jStart + 1) - (j - itj);
					} else {
						vector(interior_j, iti, itj, itk) = (jStart + 1) - (j - itj);
					}
					if (k - itk < 0) {
						vector(interior_k, iti, itj, itk) = - (kStart + 1) - (k - itk);
					} else {
						vector(interior_k, iti, itj, itk) = (kStart + 1) - (k - itk);
					}
				}
			}
		}
	}
}

/*
 * Checks if the point has a stencil width
 */
void Problem::checkStencil(std::shared_ptr< hier::Patch > patch, int i, int j, int k, int v) const {
	double* FOV = ((pdat::NodeData<double> *) patch->getPatchData(v).get())->getPointer();
	double* interior_i = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_i_id).get())->getPointer();
	double* interior_j = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_j_id).get())->getPointer();
	double* interior_k = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_k_id).get())->getPointer();
	//Get the dimensions of the patch
	const hier::Index boxfirst = patch->getBox().lower();
	const hier::Index boxlast  = patch->getBox().upper();
	//Auxiliary definitions
	int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
	int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
	int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;

	int i_i = vector(interior_i, i, j, k);
	int iStart = MAX(0, i_i) - 1;
	int iEnd = MAX(0, -i_i) - 1;
	int i_j = vector(interior_j, i, j, k);
	int jStart = MAX(0, i_j) - 1;
	int jEnd = MAX(0, -i_j) - 1;
	int i_k = vector(interior_k, i, j, k);
	int kStart = MAX(0, i_k) - 1;
	int kEnd = MAX(0, -i_k) - 1;

	for(int iti = i - iStart; iti <= i + iEnd; iti++) {
		for(int itj = j - jStart; itj <= j + jEnd; itj++) {
			for(int itk = k - kStart; itk <= k + kEnd; itk++) {
				if(iti >= 0 && iti < ilast && itj >= 0 && itj < jlast && itk >= 0 && itk < klast && vector(FOV, iti, itj, itk) == 0) {
					vector(interior_i, iti, itj, itk) = i_i  - (i - iti);
					vector(interior_j, iti, itj, itk) = i_j  - (j - itj);
					vector(interior_k, iti, itj, itk) = i_k  - (k - itk);
				}
			}
		}
	}
}


// Point class for the floodfill algorithm
class Point {
private:
public:
	int i, j, k;
};

void Problem::floodfill(std::shared_ptr< hier::Patch > patch, int i, int j, int k, int pred, int seg) const {

	double* FOV;
	switch(seg) {
		case 1:
			FOV = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		break;
	}
	int* nonSync = ((pdat::NodeData<int> *) patch->getPatchData(d_nonSync_regridding_tag_id).get())->getPointer();
	double* interior = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_regridding_value_id).get())->getPointer();
	//Get the dimensions of the patch
	const hier::Index boxfirst = patch->getBox().lower();
	const hier::Index boxlast  = patch->getBox().upper();
	//Auxiliary definitions
	int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
	int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
	int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;

	stack<Point> mystack;

	Point p;
	p.i = i;
	p.j = j;
	p.k = k;

	mystack.push(p);
	while(mystack.size() > 0) {
		p = mystack.top();
		mystack.pop();
		if (vector(nonSync, p.i, p.j, p.k) == 0) {
			vector(nonSync, p.i, p.j, p.k) = pred;
			vector(interior, p.i, p.j, p.k) = pred;
			if (pred == 2) {
				vector(FOV, p.i, p.j, p.k) = 100;
			}
			if (p.i - 1 >= 0 && vector(nonSync, p.i - 1, p.j, p.k) == 0) {
				Point np;
				np.i = p.i-1;
				np.j = p.j;
				np.k = p.k;
				mystack.push(np);
			}
			if (p.i + 1 < ilast && vector(nonSync, p.i + 1, p.j, p.k) == 0) {
				Point np;
				np.i = p.i+1;
				np.j = p.j;
				np.k = p.k;
				mystack.push(np);
			}
			if (p.j - 1 >= 0 && vector(nonSync, p.i, p.j - 1, p.k) == 0) {
				Point np;
				np.i = p.i;
				np.j = p.j-1;
				np.k = p.k;
				mystack.push(np);
			}
			if (p.j + 1 < jlast && vector(nonSync, p.i, p.j + 1, p.k) == 0) {
				Point np;
				np.i = p.i;
				np.j = p.j+1;
				np.k = p.k;
				mystack.push(np);
			}
			if (p.k - 1 >= 0 && vector(nonSync, p.i, p.j, p.k - 1) == 0) {
				Point np;
				np.i = p.i;
				np.j = p.j;
				np.k = p.k-1;
				mystack.push(np);
			}
			if (p.k + 1 < klast && vector(nonSync, p.i, p.j, p.k + 1) == 0) {
				Point np;
				np.i = p.i;
				np.j = p.j;
				np.k = p.k+1;
				mystack.push(np);
			}
		}
	}
}


/*
 * FOV correction for AMR
 */
void Problem::correctFOVS(const std::shared_ptr< hier::PatchLevel >& level) {
	int i, j, k;
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr< hier::Patch >& patch = *p_it;

		//Get the dimensions of the patch
		hier::Box pbox = patch->getBox();
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast  = patch->getBox().upper();

		//Get delta spaces into an array. dx, dy, dz.
		const std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();

		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;

		for (i = 0; i < ilast; i++) {
			for (j = 0; j < jlast; j++) {
				for (k = 0; k < klast; k++) {
					if (vector(FOV_xLower, i, j, k) > 0) {
						vector(FOV_xLower, i, j, k) = 100;
						vector(FOV_1, i, j, k) = 0;
						vector(FOV_xUpper, i, j, k) = 0;
						vector(FOV_yLower, i, j, k) = 0;
						vector(FOV_yUpper, i, j, k) = 0;
						vector(FOV_zLower, i, j, k) = 0;
						vector(FOV_zUpper, i, j, k) = 0;
					}
					if (vector(FOV_xUpper, i, j, k) > 0) {
						vector(FOV_xUpper, i, j, k) = 100;
						vector(FOV_1, i, j, k) = 0;
						vector(FOV_xLower, i, j, k) = 0;
						vector(FOV_yLower, i, j, k) = 0;
						vector(FOV_yUpper, i, j, k) = 0;
						vector(FOV_zLower, i, j, k) = 0;
						vector(FOV_zUpper, i, j, k) = 0;
					}
					if (vector(FOV_yLower, i, j, k) > 0) {
						vector(FOV_yLower, i, j, k) = 100;
						vector(FOV_1, i, j, k) = 0;
						vector(FOV_xLower, i, j, k) = 0;
						vector(FOV_xUpper, i, j, k) = 0;
						vector(FOV_yUpper, i, j, k) = 0;
						vector(FOV_zLower, i, j, k) = 0;
						vector(FOV_zUpper, i, j, k) = 0;
					}
					if (vector(FOV_yUpper, i, j, k) > 0) {
						vector(FOV_yUpper, i, j, k) = 100;
						vector(FOV_1, i, j, k) = 0;
						vector(FOV_xLower, i, j, k) = 0;
						vector(FOV_xUpper, i, j, k) = 0;
						vector(FOV_yLower, i, j, k) = 0;
						vector(FOV_zLower, i, j, k) = 0;
						vector(FOV_zUpper, i, j, k) = 0;
					}
					if (vector(FOV_zLower, i, j, k) > 0) {
						vector(FOV_zLower, i, j, k) = 100;
						vector(FOV_1, i, j, k) = 0;
						vector(FOV_xLower, i, j, k) = 0;
						vector(FOV_xUpper, i, j, k) = 0;
						vector(FOV_yLower, i, j, k) = 0;
						vector(FOV_yUpper, i, j, k) = 0;
						vector(FOV_zUpper, i, j, k) = 0;
					}
					if (vector(FOV_zUpper, i, j, k) > 0) {
						vector(FOV_zUpper, i, j, k) = 100;
						vector(FOV_1, i, j, k) = 0;
						vector(FOV_xLower, i, j, k) = 0;
						vector(FOV_xUpper, i, j, k) = 0;
						vector(FOV_yLower, i, j, k) = 0;
						vector(FOV_yUpper, i, j, k) = 0;
						vector(FOV_zLower, i, j, k) = 0;
					}
					if (vector(FOV_1, i, j, k) > 0) {
						vector(FOV_1, i, j, k) = 100;
						vector(FOV_xLower, i, j, k) = 0;
						vector(FOV_xUpper, i, j, k) = 0;
						vector(FOV_yLower, i, j, k) = 0;
						vector(FOV_yUpper, i, j, k) = 0;
						vector(FOV_zLower, i, j, k) = 0;
						vector(FOV_zUpper, i, j, k) = 0;
					}
				}
			}
		}
	}
}




void Problem::interphaseMapping(const double time ,const bool initial_time,const int ln, const std::shared_ptr< hier::PatchLevel >& level, const int remesh) {
	const tbox::SAMRAI_MPI& mpi(tbox::SAMRAI_MPI::getSAMRAIWorld());
	if (remesh == 1) {
		//Update extrapolation variables
		//Calculation of hard region distance variables
		for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
			const std::shared_ptr< hier::Patch >& patch = *p_it;

			//Get the dimensions of the patch
			const hier::Index boxfirst = patch->getBox().lower();
			const hier::Index boxlast  = patch->getBox().upper();
			const hier::IntVector ratio = level->getRatioToCoarserLevel();
			double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
			double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
			double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
			double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
			double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
			double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
			double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
			double* stalled_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_stalled_1_id).get())->getPointer();
			//Hard region field distance variables
			double* d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
			double* d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
			double* d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();

			int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
			int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
			int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			for (int i = 0; i < ilast; i++) {
				for (int j = 0; j < jlast; j++) {
					for (int k = 0; k < klast; k++) {
						vector(stalled_1, i, j, k) = checkStalled(patch, i, j, k, d_FOV_1_id);
						vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = 0;
						vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = 0;
						vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = 0;
					}
				}
			}
		}
		d_mapping_fill->createSchedule(level, level)->fillData(time, true);
		for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
			const std::shared_ptr< hier::Patch >& patch = *p_it;

			//Get the dimensions of the patch
			const hier::Index boxfirst = patch->getBox().lower();
			const hier::Index boxlast  = patch->getBox().upper();
			const hier::IntVector ratio = level->getRatioToCoarserLevel();
			double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
			double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
			double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
			double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
			double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
			double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
			double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
			double* stalled_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_stalled_1_id).get())->getPointer();
			//Hard region field distance variables
			double* d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
			double* d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
			double* d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();

			int reGrid_i;
			if (ratio(0) == 0) reGrid_i = 1;
			else           reGrid_i = ratio(0);
			int reGrid_j;
			if (ratio(1) == 0) reGrid_j = 1;
			else           reGrid_j = ratio(1);
			int reGrid_k;
			if (ratio(2) == 0) reGrid_k = 1;
			else           reGrid_k = ratio(2);

			int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
			int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
			int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			int dist_i_tmp, dist_i_tmp_p, dist_i_tmp_m, dist_i_tmp_r;
			int dist_j_tmp, dist_j_tmp_p, dist_j_tmp_m, dist_j_tmp_r;
			int dist_k_tmp, dist_k_tmp_p, dist_k_tmp_m, dist_k_tmp_r;
			int dist_r, dist_r_tmp;

			for (int i = 0; i < ilast; i++) {
				for (int j = 0; j < jlast; j++) {
					for (int k = 0; k < klast; k++) {
						if (vector(FOV_zUpper, i, j, k) > 0) {
							dist_i_tmp = 999;
							dist_i_tmp_p = 999;
							dist_i_tmp_m = 999;
							dist_i_tmp_r = 999;
							dist_j_tmp = 999;
							dist_j_tmp_p = 999;
							dist_j_tmp_m = 999;
							dist_j_tmp_r = 999;
							dist_k_tmp = 999;
							dist_k_tmp_p = 999;
							dist_k_tmp_m = 999;
							dist_k_tmp_r = 999;
							dist_r = 999;
							for (int Dist_i = -3 * reGrid_i; Dist_i <= 3 * reGrid_i; Dist_i++) {
								for (int Dist_j = -3 * reGrid_j; Dist_j <= 3 * reGrid_j; Dist_j++) {
									for (int Dist_k = -3 * reGrid_k; Dist_k <= 3 * reGrid_k; Dist_k++) {
										if (i + Dist_i >= 0 && i + Dist_i < ilast && j + Dist_j >= 0 && j + Dist_j < jlast && k + Dist_k >= 0 && k + Dist_k < klast && (!vector(stalled_1, i + Dist_i, j + Dist_j, k + Dist_k))) {
											if (Dist_i < 0 && Dist_j == 0 && Dist_k == 0 && fabs(Dist_i) < fabs(dist_i_tmp_m)) {
												dist_i_tmp_m = -Dist_i;
											}
											if (Dist_i > 0 && Dist_j == 0 && Dist_k == 0 && fabs(Dist_i) < fabs(dist_i_tmp_p)) {
												dist_i_tmp_p = -Dist_i;
											}
											if (Dist_j < 0 && Dist_i == 0 && Dist_k == 0 && fabs(Dist_j) < fabs(dist_j_tmp_m)) {
												dist_j_tmp_m = -Dist_j;
											}
											if (Dist_j > 0 && Dist_i == 0 && Dist_k == 0 && fabs(Dist_j) < fabs(dist_j_tmp_p)) {
												dist_j_tmp_p = -Dist_j;
											}
											if (Dist_k < 0 && Dist_i == 0 && Dist_j == 0 && fabs(Dist_k) < fabs(dist_k_tmp_m)) {
												dist_k_tmp_m = -Dist_k;
											}
											if (Dist_k > 0 && Dist_i == 0 && Dist_j == 0 && fabs(Dist_k) < fabs(dist_k_tmp_p)) {
												dist_k_tmp_p = -Dist_k;
											}
											dist_r_tmp = Dist_i * Dist_i + Dist_j * Dist_j + Dist_k * Dist_k;
											if (dist_r_tmp < dist_r) {
												dist_r = dist_r_tmp;
												dist_i_tmp_r = -Dist_i;
												dist_j_tmp_r = -Dist_j;
												dist_k_tmp_r = -Dist_k;
											}
										}
									}
								}
							}
							if (fabs(dist_i_tmp_m) == fabs(dist_i_tmp_p) && dist_i_tmp_m < 999) {
								dist_i_tmp = 0;
								bool enough = true;
								for (int Dist_i = 0; Dist_i <= 2 && enough; Dist_i++) {
									if (!(i - dist_i_tmp_p + Dist_i < ilast && (!vector(stalled_1, i - dist_i_tmp_p + Dist_i, j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_i_tmp = dist_i_tmp_p;
								}
								enough = true;
								for (int Dist_i = -2; Dist_i <= 0 && enough; Dist_i++) {
									if (!(i - dist_i_tmp_m + Dist_i >= 0 && (!vector(stalled_1, i - dist_i_tmp_m + Dist_i, j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_i_tmp = dist_i_tmp_m;
								}
							} else if (fabs(dist_i_tmp_m) < fabs(dist_i_tmp_p)) {
								dist_i_tmp = dist_i_tmp_m;
							} else {
								dist_i_tmp = dist_i_tmp_p;
							}
							if (fabs(dist_j_tmp_m) == fabs(dist_j_tmp_p) && dist_j_tmp_m < 999) {
								dist_j_tmp = 0;
								bool enough = true;
								for (int Dist_j = 0; Dist_j <= 2 && enough; Dist_j++) {
									if (!(j - dist_j_tmp_p + Dist_j < jlast && (!vector(stalled_1, i, j - dist_j_tmp_p + Dist_j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_j_tmp = dist_j_tmp_p;
								}
								enough = true;
								for (int Dist_j = -2; Dist_j <= 0 && enough; Dist_j++) {
									if (!(j - dist_j_tmp_m + Dist_j >= 0 && (!vector(stalled_1, i, j - dist_j_tmp_m + Dist_j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_j_tmp = dist_j_tmp_m;
								}
							} else if (fabs(dist_j_tmp_m) < fabs(dist_j_tmp_p)) {
								dist_j_tmp = dist_j_tmp_m;
							} else {
								dist_j_tmp = dist_j_tmp_p;
							}
							if (fabs(dist_k_tmp_m) == fabs(dist_k_tmp_p) && dist_k_tmp_m < 999) {
								dist_k_tmp = 0;
								bool enough = true;
								for (int Dist_k = 0; Dist_k <= 2 && enough; Dist_k++) {
									if (!(k - dist_k_tmp_p + Dist_k < klast && (!vector(stalled_1, i, j, k - dist_k_tmp_p + Dist_k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_k_tmp = dist_k_tmp_p;
								}
								enough = true;
								for (int Dist_k = -2; Dist_k <= 0 && enough; Dist_k++) {
									if (!(k - dist_k_tmp_m + Dist_k >= 0 && (!vector(stalled_1, i, j, k - dist_k_tmp_m + Dist_k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_k_tmp = dist_k_tmp_m;
								}
							} else if (fabs(dist_k_tmp_m) < fabs(dist_k_tmp_p)) {
								dist_k_tmp = dist_k_tmp_m;
							} else {
								dist_k_tmp = dist_k_tmp_p;
							}
							if (dist_i_tmp == 999 && dist_j_tmp == 999 && dist_k_tmp == 999) {
								dist_i_tmp = dist_i_tmp_r;
								dist_j_tmp = dist_j_tmp_r;
								dist_k_tmp = dist_k_tmp_r;
							}
							if (dist_i_tmp != 999 && ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_i_tmp) < fabs(vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_i_tmp;
							}
							if (dist_j_tmp != 999 && ((vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_j_tmp) < fabs(vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_j_tmp;
							}
							if (dist_k_tmp != 999 && ((vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_k_tmp) < fabs(vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_k_tmp;
							}
						}
						if (vector(FOV_zLower, i, j, k) > 0) {
							dist_i_tmp = 999;
							dist_i_tmp_p = 999;
							dist_i_tmp_m = 999;
							dist_i_tmp_r = 999;
							dist_j_tmp = 999;
							dist_j_tmp_p = 999;
							dist_j_tmp_m = 999;
							dist_j_tmp_r = 999;
							dist_k_tmp = 999;
							dist_k_tmp_p = 999;
							dist_k_tmp_m = 999;
							dist_k_tmp_r = 999;
							dist_r = 999;
							for (int Dist_i = -3 * reGrid_i; Dist_i <= 3 * reGrid_i; Dist_i++) {
								for (int Dist_j = -3 * reGrid_j; Dist_j <= 3 * reGrid_j; Dist_j++) {
									for (int Dist_k = -3 * reGrid_k; Dist_k <= 3 * reGrid_k; Dist_k++) {
										if (i + Dist_i >= 0 && i + Dist_i < ilast && j + Dist_j >= 0 && j + Dist_j < jlast && k + Dist_k >= 0 && k + Dist_k < klast && (!vector(stalled_1, i + Dist_i, j + Dist_j, k + Dist_k))) {
											if (Dist_i < 0 && Dist_j == 0 && Dist_k == 0 && fabs(Dist_i) < fabs(dist_i_tmp_m)) {
												dist_i_tmp_m = -Dist_i;
											}
											if (Dist_i > 0 && Dist_j == 0 && Dist_k == 0 && fabs(Dist_i) < fabs(dist_i_tmp_p)) {
												dist_i_tmp_p = -Dist_i;
											}
											if (Dist_j < 0 && Dist_i == 0 && Dist_k == 0 && fabs(Dist_j) < fabs(dist_j_tmp_m)) {
												dist_j_tmp_m = -Dist_j;
											}
											if (Dist_j > 0 && Dist_i == 0 && Dist_k == 0 && fabs(Dist_j) < fabs(dist_j_tmp_p)) {
												dist_j_tmp_p = -Dist_j;
											}
											if (Dist_k < 0 && Dist_i == 0 && Dist_j == 0 && fabs(Dist_k) < fabs(dist_k_tmp_m)) {
												dist_k_tmp_m = -Dist_k;
											}
											if (Dist_k > 0 && Dist_i == 0 && Dist_j == 0 && fabs(Dist_k) < fabs(dist_k_tmp_p)) {
												dist_k_tmp_p = -Dist_k;
											}
											dist_r_tmp = Dist_i * Dist_i + Dist_j * Dist_j + Dist_k * Dist_k;
											if (dist_r_tmp < dist_r) {
												dist_r = dist_r_tmp;
												dist_i_tmp_r = -Dist_i;
												dist_j_tmp_r = -Dist_j;
												dist_k_tmp_r = -Dist_k;
											}
										}
									}
								}
							}
							if (fabs(dist_i_tmp_m) == fabs(dist_i_tmp_p) && dist_i_tmp_m < 999) {
								dist_i_tmp = 0;
								bool enough = true;
								for (int Dist_i = 0; Dist_i <= 2 && enough; Dist_i++) {
									if (!(i - dist_i_tmp_p + Dist_i < ilast && (!vector(stalled_1, i - dist_i_tmp_p + Dist_i, j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_i_tmp = dist_i_tmp_p;
								}
								enough = true;
								for (int Dist_i = -2; Dist_i <= 0 && enough; Dist_i++) {
									if (!(i - dist_i_tmp_m + Dist_i >= 0 && (!vector(stalled_1, i - dist_i_tmp_m + Dist_i, j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_i_tmp = dist_i_tmp_m;
								}
							} else if (fabs(dist_i_tmp_m) < fabs(dist_i_tmp_p)) {
								dist_i_tmp = dist_i_tmp_m;
							} else {
								dist_i_tmp = dist_i_tmp_p;
							}
							if (fabs(dist_j_tmp_m) == fabs(dist_j_tmp_p) && dist_j_tmp_m < 999) {
								dist_j_tmp = 0;
								bool enough = true;
								for (int Dist_j = 0; Dist_j <= 2 && enough; Dist_j++) {
									if (!(j - dist_j_tmp_p + Dist_j < jlast && (!vector(stalled_1, i, j - dist_j_tmp_p + Dist_j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_j_tmp = dist_j_tmp_p;
								}
								enough = true;
								for (int Dist_j = -2; Dist_j <= 0 && enough; Dist_j++) {
									if (!(j - dist_j_tmp_m + Dist_j >= 0 && (!vector(stalled_1, i, j - dist_j_tmp_m + Dist_j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_j_tmp = dist_j_tmp_m;
								}
							} else if (fabs(dist_j_tmp_m) < fabs(dist_j_tmp_p)) {
								dist_j_tmp = dist_j_tmp_m;
							} else {
								dist_j_tmp = dist_j_tmp_p;
							}
							if (fabs(dist_k_tmp_m) == fabs(dist_k_tmp_p) && dist_k_tmp_m < 999) {
								dist_k_tmp = 0;
								bool enough = true;
								for (int Dist_k = 0; Dist_k <= 2 && enough; Dist_k++) {
									if (!(k - dist_k_tmp_p + Dist_k < klast && (!vector(stalled_1, i, j, k - dist_k_tmp_p + Dist_k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_k_tmp = dist_k_tmp_p;
								}
								enough = true;
								for (int Dist_k = -2; Dist_k <= 0 && enough; Dist_k++) {
									if (!(k - dist_k_tmp_m + Dist_k >= 0 && (!vector(stalled_1, i, j, k - dist_k_tmp_m + Dist_k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_k_tmp = dist_k_tmp_m;
								}
							} else if (fabs(dist_k_tmp_m) < fabs(dist_k_tmp_p)) {
								dist_k_tmp = dist_k_tmp_m;
							} else {
								dist_k_tmp = dist_k_tmp_p;
							}
							if (dist_i_tmp == 999 && dist_j_tmp == 999 && dist_k_tmp == 999) {
								dist_i_tmp = dist_i_tmp_r;
								dist_j_tmp = dist_j_tmp_r;
								dist_k_tmp = dist_k_tmp_r;
							}
							if (dist_i_tmp != 999 && ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_i_tmp) < fabs(vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_i_tmp;
							}
							if (dist_j_tmp != 999 && ((vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_j_tmp) < fabs(vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_j_tmp;
							}
							if (dist_k_tmp != 999 && ((vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_k_tmp) < fabs(vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_k_tmp;
							}
						}
						if (vector(FOV_yUpper, i, j, k) > 0) {
							dist_i_tmp = 999;
							dist_i_tmp_p = 999;
							dist_i_tmp_m = 999;
							dist_i_tmp_r = 999;
							dist_j_tmp = 999;
							dist_j_tmp_p = 999;
							dist_j_tmp_m = 999;
							dist_j_tmp_r = 999;
							dist_k_tmp = 999;
							dist_k_tmp_p = 999;
							dist_k_tmp_m = 999;
							dist_k_tmp_r = 999;
							dist_r = 999;
							for (int Dist_i = -3 * reGrid_i; Dist_i <= 3 * reGrid_i; Dist_i++) {
								for (int Dist_j = -3 * reGrid_j; Dist_j <= 3 * reGrid_j; Dist_j++) {
									for (int Dist_k = -3 * reGrid_k; Dist_k <= 3 * reGrid_k; Dist_k++) {
										if (i + Dist_i >= 0 && i + Dist_i < ilast && j + Dist_j >= 0 && j + Dist_j < jlast && k + Dist_k >= 0 && k + Dist_k < klast && (!vector(stalled_1, i + Dist_i, j + Dist_j, k + Dist_k))) {
											if (Dist_i < 0 && Dist_j == 0 && Dist_k == 0 && fabs(Dist_i) < fabs(dist_i_tmp_m)) {
												dist_i_tmp_m = -Dist_i;
											}
											if (Dist_i > 0 && Dist_j == 0 && Dist_k == 0 && fabs(Dist_i) < fabs(dist_i_tmp_p)) {
												dist_i_tmp_p = -Dist_i;
											}
											if (Dist_j < 0 && Dist_i == 0 && Dist_k == 0 && fabs(Dist_j) < fabs(dist_j_tmp_m)) {
												dist_j_tmp_m = -Dist_j;
											}
											if (Dist_j > 0 && Dist_i == 0 && Dist_k == 0 && fabs(Dist_j) < fabs(dist_j_tmp_p)) {
												dist_j_tmp_p = -Dist_j;
											}
											if (Dist_k < 0 && Dist_i == 0 && Dist_j == 0 && fabs(Dist_k) < fabs(dist_k_tmp_m)) {
												dist_k_tmp_m = -Dist_k;
											}
											if (Dist_k > 0 && Dist_i == 0 && Dist_j == 0 && fabs(Dist_k) < fabs(dist_k_tmp_p)) {
												dist_k_tmp_p = -Dist_k;
											}
											dist_r_tmp = Dist_i * Dist_i + Dist_j * Dist_j + Dist_k * Dist_k;
											if (dist_r_tmp < dist_r) {
												dist_r = dist_r_tmp;
												dist_i_tmp_r = -Dist_i;
												dist_j_tmp_r = -Dist_j;
												dist_k_tmp_r = -Dist_k;
											}
										}
									}
								}
							}
							if (fabs(dist_i_tmp_m) == fabs(dist_i_tmp_p) && dist_i_tmp_m < 999) {
								dist_i_tmp = 0;
								bool enough = true;
								for (int Dist_i = 0; Dist_i <= 2 && enough; Dist_i++) {
									if (!(i - dist_i_tmp_p + Dist_i < ilast && (!vector(stalled_1, i - dist_i_tmp_p + Dist_i, j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_i_tmp = dist_i_tmp_p;
								}
								enough = true;
								for (int Dist_i = -2; Dist_i <= 0 && enough; Dist_i++) {
									if (!(i - dist_i_tmp_m + Dist_i >= 0 && (!vector(stalled_1, i - dist_i_tmp_m + Dist_i, j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_i_tmp = dist_i_tmp_m;
								}
							} else if (fabs(dist_i_tmp_m) < fabs(dist_i_tmp_p)) {
								dist_i_tmp = dist_i_tmp_m;
							} else {
								dist_i_tmp = dist_i_tmp_p;
							}
							if (fabs(dist_j_tmp_m) == fabs(dist_j_tmp_p) && dist_j_tmp_m < 999) {
								dist_j_tmp = 0;
								bool enough = true;
								for (int Dist_j = 0; Dist_j <= 2 && enough; Dist_j++) {
									if (!(j - dist_j_tmp_p + Dist_j < jlast && (!vector(stalled_1, i, j - dist_j_tmp_p + Dist_j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_j_tmp = dist_j_tmp_p;
								}
								enough = true;
								for (int Dist_j = -2; Dist_j <= 0 && enough; Dist_j++) {
									if (!(j - dist_j_tmp_m + Dist_j >= 0 && (!vector(stalled_1, i, j - dist_j_tmp_m + Dist_j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_j_tmp = dist_j_tmp_m;
								}
							} else if (fabs(dist_j_tmp_m) < fabs(dist_j_tmp_p)) {
								dist_j_tmp = dist_j_tmp_m;
							} else {
								dist_j_tmp = dist_j_tmp_p;
							}
							if (fabs(dist_k_tmp_m) == fabs(dist_k_tmp_p) && dist_k_tmp_m < 999) {
								dist_k_tmp = 0;
								bool enough = true;
								for (int Dist_k = 0; Dist_k <= 2 && enough; Dist_k++) {
									if (!(k - dist_k_tmp_p + Dist_k < klast && (!vector(stalled_1, i, j, k - dist_k_tmp_p + Dist_k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_k_tmp = dist_k_tmp_p;
								}
								enough = true;
								for (int Dist_k = -2; Dist_k <= 0 && enough; Dist_k++) {
									if (!(k - dist_k_tmp_m + Dist_k >= 0 && (!vector(stalled_1, i, j, k - dist_k_tmp_m + Dist_k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_k_tmp = dist_k_tmp_m;
								}
							} else if (fabs(dist_k_tmp_m) < fabs(dist_k_tmp_p)) {
								dist_k_tmp = dist_k_tmp_m;
							} else {
								dist_k_tmp = dist_k_tmp_p;
							}
							if (dist_i_tmp == 999 && dist_j_tmp == 999 && dist_k_tmp == 999) {
								dist_i_tmp = dist_i_tmp_r;
								dist_j_tmp = dist_j_tmp_r;
								dist_k_tmp = dist_k_tmp_r;
							}
							if (dist_i_tmp != 999 && ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_i_tmp) < fabs(vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_i_tmp;
							}
							if (dist_j_tmp != 999 && ((vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_j_tmp) < fabs(vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_j_tmp;
							}
							if (dist_k_tmp != 999 && ((vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_k_tmp) < fabs(vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_k_tmp;
							}
						}
						if (vector(FOV_yLower, i, j, k) > 0) {
							dist_i_tmp = 999;
							dist_i_tmp_p = 999;
							dist_i_tmp_m = 999;
							dist_i_tmp_r = 999;
							dist_j_tmp = 999;
							dist_j_tmp_p = 999;
							dist_j_tmp_m = 999;
							dist_j_tmp_r = 999;
							dist_k_tmp = 999;
							dist_k_tmp_p = 999;
							dist_k_tmp_m = 999;
							dist_k_tmp_r = 999;
							dist_r = 999;
							for (int Dist_i = -3 * reGrid_i; Dist_i <= 3 * reGrid_i; Dist_i++) {
								for (int Dist_j = -3 * reGrid_j; Dist_j <= 3 * reGrid_j; Dist_j++) {
									for (int Dist_k = -3 * reGrid_k; Dist_k <= 3 * reGrid_k; Dist_k++) {
										if (i + Dist_i >= 0 && i + Dist_i < ilast && j + Dist_j >= 0 && j + Dist_j < jlast && k + Dist_k >= 0 && k + Dist_k < klast && (!vector(stalled_1, i + Dist_i, j + Dist_j, k + Dist_k))) {
											if (Dist_i < 0 && Dist_j == 0 && Dist_k == 0 && fabs(Dist_i) < fabs(dist_i_tmp_m)) {
												dist_i_tmp_m = -Dist_i;
											}
											if (Dist_i > 0 && Dist_j == 0 && Dist_k == 0 && fabs(Dist_i) < fabs(dist_i_tmp_p)) {
												dist_i_tmp_p = -Dist_i;
											}
											if (Dist_j < 0 && Dist_i == 0 && Dist_k == 0 && fabs(Dist_j) < fabs(dist_j_tmp_m)) {
												dist_j_tmp_m = -Dist_j;
											}
											if (Dist_j > 0 && Dist_i == 0 && Dist_k == 0 && fabs(Dist_j) < fabs(dist_j_tmp_p)) {
												dist_j_tmp_p = -Dist_j;
											}
											if (Dist_k < 0 && Dist_i == 0 && Dist_j == 0 && fabs(Dist_k) < fabs(dist_k_tmp_m)) {
												dist_k_tmp_m = -Dist_k;
											}
											if (Dist_k > 0 && Dist_i == 0 && Dist_j == 0 && fabs(Dist_k) < fabs(dist_k_tmp_p)) {
												dist_k_tmp_p = -Dist_k;
											}
											dist_r_tmp = Dist_i * Dist_i + Dist_j * Dist_j + Dist_k * Dist_k;
											if (dist_r_tmp < dist_r) {
												dist_r = dist_r_tmp;
												dist_i_tmp_r = -Dist_i;
												dist_j_tmp_r = -Dist_j;
												dist_k_tmp_r = -Dist_k;
											}
										}
									}
								}
							}
							if (fabs(dist_i_tmp_m) == fabs(dist_i_tmp_p) && dist_i_tmp_m < 999) {
								dist_i_tmp = 0;
								bool enough = true;
								for (int Dist_i = 0; Dist_i <= 2 && enough; Dist_i++) {
									if (!(i - dist_i_tmp_p + Dist_i < ilast && (!vector(stalled_1, i - dist_i_tmp_p + Dist_i, j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_i_tmp = dist_i_tmp_p;
								}
								enough = true;
								for (int Dist_i = -2; Dist_i <= 0 && enough; Dist_i++) {
									if (!(i - dist_i_tmp_m + Dist_i >= 0 && (!vector(stalled_1, i - dist_i_tmp_m + Dist_i, j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_i_tmp = dist_i_tmp_m;
								}
							} else if (fabs(dist_i_tmp_m) < fabs(dist_i_tmp_p)) {
								dist_i_tmp = dist_i_tmp_m;
							} else {
								dist_i_tmp = dist_i_tmp_p;
							}
							if (fabs(dist_j_tmp_m) == fabs(dist_j_tmp_p) && dist_j_tmp_m < 999) {
								dist_j_tmp = 0;
								bool enough = true;
								for (int Dist_j = 0; Dist_j <= 2 && enough; Dist_j++) {
									if (!(j - dist_j_tmp_p + Dist_j < jlast && (!vector(stalled_1, i, j - dist_j_tmp_p + Dist_j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_j_tmp = dist_j_tmp_p;
								}
								enough = true;
								for (int Dist_j = -2; Dist_j <= 0 && enough; Dist_j++) {
									if (!(j - dist_j_tmp_m + Dist_j >= 0 && (!vector(stalled_1, i, j - dist_j_tmp_m + Dist_j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_j_tmp = dist_j_tmp_m;
								}
							} else if (fabs(dist_j_tmp_m) < fabs(dist_j_tmp_p)) {
								dist_j_tmp = dist_j_tmp_m;
							} else {
								dist_j_tmp = dist_j_tmp_p;
							}
							if (fabs(dist_k_tmp_m) == fabs(dist_k_tmp_p) && dist_k_tmp_m < 999) {
								dist_k_tmp = 0;
								bool enough = true;
								for (int Dist_k = 0; Dist_k <= 2 && enough; Dist_k++) {
									if (!(k - dist_k_tmp_p + Dist_k < klast && (!vector(stalled_1, i, j, k - dist_k_tmp_p + Dist_k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_k_tmp = dist_k_tmp_p;
								}
								enough = true;
								for (int Dist_k = -2; Dist_k <= 0 && enough; Dist_k++) {
									if (!(k - dist_k_tmp_m + Dist_k >= 0 && (!vector(stalled_1, i, j, k - dist_k_tmp_m + Dist_k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_k_tmp = dist_k_tmp_m;
								}
							} else if (fabs(dist_k_tmp_m) < fabs(dist_k_tmp_p)) {
								dist_k_tmp = dist_k_tmp_m;
							} else {
								dist_k_tmp = dist_k_tmp_p;
							}
							if (dist_i_tmp == 999 && dist_j_tmp == 999 && dist_k_tmp == 999) {
								dist_i_tmp = dist_i_tmp_r;
								dist_j_tmp = dist_j_tmp_r;
								dist_k_tmp = dist_k_tmp_r;
							}
							if (dist_i_tmp != 999 && ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_i_tmp) < fabs(vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_i_tmp;
							}
							if (dist_j_tmp != 999 && ((vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_j_tmp) < fabs(vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_j_tmp;
							}
							if (dist_k_tmp != 999 && ((vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_k_tmp) < fabs(vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_k_tmp;
							}
						}
						if (vector(FOV_xUpper, i, j, k) > 0) {
							dist_i_tmp = 999;
							dist_i_tmp_p = 999;
							dist_i_tmp_m = 999;
							dist_i_tmp_r = 999;
							dist_j_tmp = 999;
							dist_j_tmp_p = 999;
							dist_j_tmp_m = 999;
							dist_j_tmp_r = 999;
							dist_k_tmp = 999;
							dist_k_tmp_p = 999;
							dist_k_tmp_m = 999;
							dist_k_tmp_r = 999;
							dist_r = 999;
							for (int Dist_i = -3 * reGrid_i; Dist_i <= 3 * reGrid_i; Dist_i++) {
								for (int Dist_j = -3 * reGrid_j; Dist_j <= 3 * reGrid_j; Dist_j++) {
									for (int Dist_k = -3 * reGrid_k; Dist_k <= 3 * reGrid_k; Dist_k++) {
										if (i + Dist_i >= 0 && i + Dist_i < ilast && j + Dist_j >= 0 && j + Dist_j < jlast && k + Dist_k >= 0 && k + Dist_k < klast && (!vector(stalled_1, i + Dist_i, j + Dist_j, k + Dist_k))) {
											if (Dist_i < 0 && Dist_j == 0 && Dist_k == 0 && fabs(Dist_i) < fabs(dist_i_tmp_m)) {
												dist_i_tmp_m = -Dist_i;
											}
											if (Dist_i > 0 && Dist_j == 0 && Dist_k == 0 && fabs(Dist_i) < fabs(dist_i_tmp_p)) {
												dist_i_tmp_p = -Dist_i;
											}
											if (Dist_j < 0 && Dist_i == 0 && Dist_k == 0 && fabs(Dist_j) < fabs(dist_j_tmp_m)) {
												dist_j_tmp_m = -Dist_j;
											}
											if (Dist_j > 0 && Dist_i == 0 && Dist_k == 0 && fabs(Dist_j) < fabs(dist_j_tmp_p)) {
												dist_j_tmp_p = -Dist_j;
											}
											if (Dist_k < 0 && Dist_i == 0 && Dist_j == 0 && fabs(Dist_k) < fabs(dist_k_tmp_m)) {
												dist_k_tmp_m = -Dist_k;
											}
											if (Dist_k > 0 && Dist_i == 0 && Dist_j == 0 && fabs(Dist_k) < fabs(dist_k_tmp_p)) {
												dist_k_tmp_p = -Dist_k;
											}
											dist_r_tmp = Dist_i * Dist_i + Dist_j * Dist_j + Dist_k * Dist_k;
											if (dist_r_tmp < dist_r) {
												dist_r = dist_r_tmp;
												dist_i_tmp_r = -Dist_i;
												dist_j_tmp_r = -Dist_j;
												dist_k_tmp_r = -Dist_k;
											}
										}
									}
								}
							}
							if (fabs(dist_i_tmp_m) == fabs(dist_i_tmp_p) && dist_i_tmp_m < 999) {
								dist_i_tmp = 0;
								bool enough = true;
								for (int Dist_i = 0; Dist_i <= 2 && enough; Dist_i++) {
									if (!(i - dist_i_tmp_p + Dist_i < ilast && (!vector(stalled_1, i - dist_i_tmp_p + Dist_i, j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_i_tmp = dist_i_tmp_p;
								}
								enough = true;
								for (int Dist_i = -2; Dist_i <= 0 && enough; Dist_i++) {
									if (!(i - dist_i_tmp_m + Dist_i >= 0 && (!vector(stalled_1, i - dist_i_tmp_m + Dist_i, j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_i_tmp = dist_i_tmp_m;
								}
							} else if (fabs(dist_i_tmp_m) < fabs(dist_i_tmp_p)) {
								dist_i_tmp = dist_i_tmp_m;
							} else {
								dist_i_tmp = dist_i_tmp_p;
							}
							if (fabs(dist_j_tmp_m) == fabs(dist_j_tmp_p) && dist_j_tmp_m < 999) {
								dist_j_tmp = 0;
								bool enough = true;
								for (int Dist_j = 0; Dist_j <= 2 && enough; Dist_j++) {
									if (!(j - dist_j_tmp_p + Dist_j < jlast && (!vector(stalled_1, i, j - dist_j_tmp_p + Dist_j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_j_tmp = dist_j_tmp_p;
								}
								enough = true;
								for (int Dist_j = -2; Dist_j <= 0 && enough; Dist_j++) {
									if (!(j - dist_j_tmp_m + Dist_j >= 0 && (!vector(stalled_1, i, j - dist_j_tmp_m + Dist_j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_j_tmp = dist_j_tmp_m;
								}
							} else if (fabs(dist_j_tmp_m) < fabs(dist_j_tmp_p)) {
								dist_j_tmp = dist_j_tmp_m;
							} else {
								dist_j_tmp = dist_j_tmp_p;
							}
							if (fabs(dist_k_tmp_m) == fabs(dist_k_tmp_p) && dist_k_tmp_m < 999) {
								dist_k_tmp = 0;
								bool enough = true;
								for (int Dist_k = 0; Dist_k <= 2 && enough; Dist_k++) {
									if (!(k - dist_k_tmp_p + Dist_k < klast && (!vector(stalled_1, i, j, k - dist_k_tmp_p + Dist_k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_k_tmp = dist_k_tmp_p;
								}
								enough = true;
								for (int Dist_k = -2; Dist_k <= 0 && enough; Dist_k++) {
									if (!(k - dist_k_tmp_m + Dist_k >= 0 && (!vector(stalled_1, i, j, k - dist_k_tmp_m + Dist_k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_k_tmp = dist_k_tmp_m;
								}
							} else if (fabs(dist_k_tmp_m) < fabs(dist_k_tmp_p)) {
								dist_k_tmp = dist_k_tmp_m;
							} else {
								dist_k_tmp = dist_k_tmp_p;
							}
							if (dist_i_tmp == 999 && dist_j_tmp == 999 && dist_k_tmp == 999) {
								dist_i_tmp = dist_i_tmp_r;
								dist_j_tmp = dist_j_tmp_r;
								dist_k_tmp = dist_k_tmp_r;
							}
							if (dist_i_tmp != 999 && ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_i_tmp) < fabs(vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_i_tmp;
							}
							if (dist_j_tmp != 999 && ((vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_j_tmp) < fabs(vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_j_tmp;
							}
							if (dist_k_tmp != 999 && ((vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_k_tmp) < fabs(vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_k_tmp;
							}
						}
						if (vector(FOV_xLower, i, j, k) > 0) {
							dist_i_tmp = 999;
							dist_i_tmp_p = 999;
							dist_i_tmp_m = 999;
							dist_i_tmp_r = 999;
							dist_j_tmp = 999;
							dist_j_tmp_p = 999;
							dist_j_tmp_m = 999;
							dist_j_tmp_r = 999;
							dist_k_tmp = 999;
							dist_k_tmp_p = 999;
							dist_k_tmp_m = 999;
							dist_k_tmp_r = 999;
							dist_r = 999;
							for (int Dist_i = -3 * reGrid_i; Dist_i <= 3 * reGrid_i; Dist_i++) {
								for (int Dist_j = -3 * reGrid_j; Dist_j <= 3 * reGrid_j; Dist_j++) {
									for (int Dist_k = -3 * reGrid_k; Dist_k <= 3 * reGrid_k; Dist_k++) {
										if (i + Dist_i >= 0 && i + Dist_i < ilast && j + Dist_j >= 0 && j + Dist_j < jlast && k + Dist_k >= 0 && k + Dist_k < klast && (!vector(stalled_1, i + Dist_i, j + Dist_j, k + Dist_k))) {
											if (Dist_i < 0 && Dist_j == 0 && Dist_k == 0 && fabs(Dist_i) < fabs(dist_i_tmp_m)) {
												dist_i_tmp_m = -Dist_i;
											}
											if (Dist_i > 0 && Dist_j == 0 && Dist_k == 0 && fabs(Dist_i) < fabs(dist_i_tmp_p)) {
												dist_i_tmp_p = -Dist_i;
											}
											if (Dist_j < 0 && Dist_i == 0 && Dist_k == 0 && fabs(Dist_j) < fabs(dist_j_tmp_m)) {
												dist_j_tmp_m = -Dist_j;
											}
											if (Dist_j > 0 && Dist_i == 0 && Dist_k == 0 && fabs(Dist_j) < fabs(dist_j_tmp_p)) {
												dist_j_tmp_p = -Dist_j;
											}
											if (Dist_k < 0 && Dist_i == 0 && Dist_j == 0 && fabs(Dist_k) < fabs(dist_k_tmp_m)) {
												dist_k_tmp_m = -Dist_k;
											}
											if (Dist_k > 0 && Dist_i == 0 && Dist_j == 0 && fabs(Dist_k) < fabs(dist_k_tmp_p)) {
												dist_k_tmp_p = -Dist_k;
											}
											dist_r_tmp = Dist_i * Dist_i + Dist_j * Dist_j + Dist_k * Dist_k;
											if (dist_r_tmp < dist_r) {
												dist_r = dist_r_tmp;
												dist_i_tmp_r = -Dist_i;
												dist_j_tmp_r = -Dist_j;
												dist_k_tmp_r = -Dist_k;
											}
										}
									}
								}
							}
							if (fabs(dist_i_tmp_m) == fabs(dist_i_tmp_p) && dist_i_tmp_m < 999) {
								dist_i_tmp = 0;
								bool enough = true;
								for (int Dist_i = 0; Dist_i <= 2 && enough; Dist_i++) {
									if (!(i - dist_i_tmp_p + Dist_i < ilast && (!vector(stalled_1, i - dist_i_tmp_p + Dist_i, j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_i_tmp = dist_i_tmp_p;
								}
								enough = true;
								for (int Dist_i = -2; Dist_i <= 0 && enough; Dist_i++) {
									if (!(i - dist_i_tmp_m + Dist_i >= 0 && (!vector(stalled_1, i - dist_i_tmp_m + Dist_i, j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_i_tmp = dist_i_tmp_m;
								}
							} else if (fabs(dist_i_tmp_m) < fabs(dist_i_tmp_p)) {
								dist_i_tmp = dist_i_tmp_m;
							} else {
								dist_i_tmp = dist_i_tmp_p;
							}
							if (fabs(dist_j_tmp_m) == fabs(dist_j_tmp_p) && dist_j_tmp_m < 999) {
								dist_j_tmp = 0;
								bool enough = true;
								for (int Dist_j = 0; Dist_j <= 2 && enough; Dist_j++) {
									if (!(j - dist_j_tmp_p + Dist_j < jlast && (!vector(stalled_1, i, j - dist_j_tmp_p + Dist_j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_j_tmp = dist_j_tmp_p;
								}
								enough = true;
								for (int Dist_j = -2; Dist_j <= 0 && enough; Dist_j++) {
									if (!(j - dist_j_tmp_m + Dist_j >= 0 && (!vector(stalled_1, i, j - dist_j_tmp_m + Dist_j, k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_j_tmp = dist_j_tmp_m;
								}
							} else if (fabs(dist_j_tmp_m) < fabs(dist_j_tmp_p)) {
								dist_j_tmp = dist_j_tmp_m;
							} else {
								dist_j_tmp = dist_j_tmp_p;
							}
							if (fabs(dist_k_tmp_m) == fabs(dist_k_tmp_p) && dist_k_tmp_m < 999) {
								dist_k_tmp = 0;
								bool enough = true;
								for (int Dist_k = 0; Dist_k <= 2 && enough; Dist_k++) {
									if (!(k - dist_k_tmp_p + Dist_k < klast && (!vector(stalled_1, i, j, k - dist_k_tmp_p + Dist_k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_k_tmp = dist_k_tmp_p;
								}
								enough = true;
								for (int Dist_k = -2; Dist_k <= 0 && enough; Dist_k++) {
									if (!(k - dist_k_tmp_m + Dist_k >= 0 && (!vector(stalled_1, i, j, k - dist_k_tmp_m + Dist_k)))) {
										enough = false;
									}
								}
								if (enough) {
									dist_k_tmp = dist_k_tmp_m;
								}
							} else if (fabs(dist_k_tmp_m) < fabs(dist_k_tmp_p)) {
								dist_k_tmp = dist_k_tmp_m;
							} else {
								dist_k_tmp = dist_k_tmp_p;
							}
							if (dist_i_tmp == 999 && dist_j_tmp == 999 && dist_k_tmp == 999) {
								dist_i_tmp = dist_i_tmp_r;
								dist_j_tmp = dist_j_tmp_r;
								dist_k_tmp = dist_k_tmp_r;
							}
							if (dist_i_tmp != 999 && ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_i_tmp) < fabs(vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_i_tmp;
							}
							if (dist_j_tmp != 999 && ((vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_j_tmp) < fabs(vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_j_tmp;
							}
							if (dist_k_tmp != 999 && ((vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) == 0) || (fabs(dist_k_tmp) < fabs(vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k))))) {
								vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) = dist_k_tmp;
							}
						}
					}
				}
			}
		}
		d_mapping_fill->createSchedule(level, level)->fillData(initial_time, true);
	}
}


/*
 * Checks if the point has to be stalled
 */
bool Problem::checkStalled(std::shared_ptr< hier::Patch > patch, int i, int j, int k, int v) const {
	double* FOV = ((pdat::NodeData<double> *) patch->getPatchData(v).get())->getPointer();
	double* interior = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_regridding_value_id).get())->getPointer();
	//Get the dimensions of the patch
	const hier::Index boxfirst = patch->getBox().lower();
	const hier::Index boxlast  = patch->getBox().upper();
	//Auxiliary definitions
	int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
	int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
	int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;

	int stencilAcc, stencilAccMax_i, stencilAccMax_j, stencilAccMax_k;
	bool notEnoughStencil = false;
	int FOV_threshold = 0;
	if (vector(FOV, i, j, k) <= FOV_threshold) {
		notEnoughStencil = true;
	} else {
		stencilAcc = 0;
		stencilAccMax_i = 0;
		for (int it1 = MAX(i-d_regionMinThickness, 0); it1 <= MIN(i+d_regionMinThickness, ilast - 1); it1++) {
			if (vector(FOV, it1, j, k) > FOV_threshold) {
				stencilAcc++;
			} else {
				stencilAccMax_i = MAX(stencilAccMax_i, stencilAcc);
				stencilAcc = 0;
			}
		}
		stencilAccMax_i = MAX(stencilAccMax_i, stencilAcc);
		stencilAcc = 0;
		stencilAccMax_j = 0;
		for (int jt1 = MAX(j-d_regionMinThickness, 0); jt1 <= MIN(j+d_regionMinThickness, jlast - 1); jt1++) {
			if (vector(FOV, i, jt1, k) > FOV_threshold) {
				stencilAcc++;
			} else {
				stencilAccMax_j = MAX(stencilAccMax_j, stencilAcc);
				stencilAcc = 0;
			}
		}
		stencilAccMax_j = MAX(stencilAccMax_j, stencilAcc);
		stencilAcc = 0;
		stencilAccMax_k = 0;
		for (int kt1 = MAX(k-d_regionMinThickness, 0); kt1 <= MIN(k+d_regionMinThickness, klast - 1); kt1++) {
			if (vector(FOV, i, j, kt1) > FOV_threshold) {
				stencilAcc++;
			} else {
				stencilAccMax_k = MAX(stencilAccMax_k, stencilAcc);
				stencilAcc = 0;
			}
		}
		stencilAccMax_k = MAX(stencilAccMax_k, stencilAcc);
		if ((stencilAccMax_i < d_regionMinThickness) || (stencilAccMax_j < d_regionMinThickness) || (stencilAccMax_k < d_regionMinThickness)) {
			notEnoughStencil = true;
		}
	}
	return notEnoughStencil;
}






/*
 * Initialize data on a patch. This initialization is done only at the begining of the simulation.
 */
void Problem::initializeDataOnPatch(hier::Patch& patch, 
                                    const double time,
                                    const bool initial_time)
{
	(void) time;
   	if (initial_time) {
		// Initial conditions		
		//Get fields, auxiliary fields and local variables that are going to be used.
		double* Sfd_x = ((pdat::NodeData<double> *) patch.getPatchData(d_Sfd_x_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Sfd_x_id).get())->fillAll(0);
		double* Sfd_y = ((pdat::NodeData<double> *) patch.getPatchData(d_Sfd_y_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Sfd_y_id).get())->fillAll(0);
		double* Sfd_z = ((pdat::NodeData<double> *) patch.getPatchData(d_Sfd_z_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Sfd_z_id).get())->fillAll(0);
		double* Bfu_x = ((pdat::NodeData<double> *) patch.getPatchData(d_Bfu_x_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Bfu_x_id).get())->fillAll(0);
		double* Bfu_y = ((pdat::NodeData<double> *) patch.getPatchData(d_Bfu_y_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Bfu_y_id).get())->fillAll(0);
		double* Bfu_z = ((pdat::NodeData<double> *) patch.getPatchData(d_Bfu_z_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Bfu_z_id).get())->fillAll(0);
		double* Df = ((pdat::NodeData<double> *) patch.getPatchData(d_Df_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Df_id).get())->fillAll(0);
		double* DYf = ((pdat::NodeData<double> *) patch.getPatchData(d_DYf_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_DYf_id).get())->fillAll(0);
		double* tauf = ((pdat::NodeData<double> *) patch.getPatchData(d_tauf_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_tauf_id).get())->fillAll(0);
		double* phif = ((pdat::NodeData<double> *) patch.getPatchData(d_phif_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_phif_id).get())->fillAll(0);
		double* gtd_xx = ((pdat::NodeData<double> *) patch.getPatchData(d_gtd_xx_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_gtd_xx_id).get())->fillAll(0);
		double* gtd_xy = ((pdat::NodeData<double> *) patch.getPatchData(d_gtd_xy_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_gtd_xy_id).get())->fillAll(0);
		double* gtd_xz = ((pdat::NodeData<double> *) patch.getPatchData(d_gtd_xz_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_gtd_xz_id).get())->fillAll(0);
		double* gtd_yy = ((pdat::NodeData<double> *) patch.getPatchData(d_gtd_yy_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_gtd_yy_id).get())->fillAll(0);
		double* gtd_yz = ((pdat::NodeData<double> *) patch.getPatchData(d_gtd_yz_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_gtd_yz_id).get())->fillAll(0);
		double* gtd_zz = ((pdat::NodeData<double> *) patch.getPatchData(d_gtd_zz_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_gtd_zz_id).get())->fillAll(0);
		double* Atd_xx = ((pdat::NodeData<double> *) patch.getPatchData(d_Atd_xx_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Atd_xx_id).get())->fillAll(0);
		double* Atd_xy = ((pdat::NodeData<double> *) patch.getPatchData(d_Atd_xy_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Atd_xy_id).get())->fillAll(0);
		double* Atd_xz = ((pdat::NodeData<double> *) patch.getPatchData(d_Atd_xz_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Atd_xz_id).get())->fillAll(0);
		double* Atd_yy = ((pdat::NodeData<double> *) patch.getPatchData(d_Atd_yy_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Atd_yy_id).get())->fillAll(0);
		double* Atd_yz = ((pdat::NodeData<double> *) patch.getPatchData(d_Atd_yz_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Atd_yz_id).get())->fillAll(0);
		double* Atd_zz = ((pdat::NodeData<double> *) patch.getPatchData(d_Atd_zz_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Atd_zz_id).get())->fillAll(0);
		double* Gamh_x = ((pdat::NodeData<double> *) patch.getPatchData(d_Gamh_x_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Gamh_x_id).get())->fillAll(0);
		double* Gamh_y = ((pdat::NodeData<double> *) patch.getPatchData(d_Gamh_y_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Gamh_y_id).get())->fillAll(0);
		double* Gamh_z = ((pdat::NodeData<double> *) patch.getPatchData(d_Gamh_z_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Gamh_z_id).get())->fillAll(0);
		double* Betau_x = ((pdat::NodeData<double> *) patch.getPatchData(d_Betau_x_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Betau_x_id).get())->fillAll(0);
		double* Betau_y = ((pdat::NodeData<double> *) patch.getPatchData(d_Betau_y_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Betau_y_id).get())->fillAll(0);
		double* Betau_z = ((pdat::NodeData<double> *) patch.getPatchData(d_Betau_z_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Betau_z_id).get())->fillAll(0);
		double* Alpha = ((pdat::NodeData<double> *) patch.getPatchData(d_Alpha_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Alpha_id).get())->fillAll(0);
		double* chi = ((pdat::NodeData<double> *) patch.getPatchData(d_chi_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_chi_id).get())->fillAll(0);
		double* trK = ((pdat::NodeData<double> *) patch.getPatchData(d_trK_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_trK_id).get())->fillAll(0);
		double* theta = ((pdat::NodeData<double> *) patch.getPatchData(d_theta_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_theta_id).get())->fillAll(0);
		double* rhof = ((pdat::NodeData<double> *) patch.getPatchData(d_rhof_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_rhof_id).get())->fillAll(0);
		double* Yef = ((pdat::NodeData<double> *) patch.getPatchData(d_Yef_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Yef_id).get())->fillAll(0);
		double* vfd_x = ((pdat::NodeData<double> *) patch.getPatchData(d_vfd_x_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_vfd_x_id).get())->fillAll(0);
		double* vfd_y = ((pdat::NodeData<double> *) patch.getPatchData(d_vfd_y_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_vfd_y_id).get())->fillAll(0);
		double* vfd_z = ((pdat::NodeData<double> *) patch.getPatchData(d_vfd_z_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_vfd_z_id).get())->fillAll(0);
		double* pf = ((pdat::NodeData<double> *) patch.getPatchData(d_pf_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_pf_id).get())->fillAll(0);
		double* epsf = ((pdat::NodeData<double> *) patch.getPatchData(d_epsf_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_epsf_id).get())->fillAll(0);
		double* Tf = ((pdat::NodeData<double> *) patch.getPatchData(d_Tf_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Tf_id).get())->fillAll(0);
		double* sqcs = ((pdat::NodeData<double> *) patch.getPatchData(d_sqcs_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_sqcs_id).get())->fillAll(0);
		double* qnu = ((pdat::NodeData<double> *) patch.getPatchData(d_qnu_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_qnu_id).get())->fillAll(0);
		double* rnu = ((pdat::NodeData<double> *) patch.getPatchData(d_rnu_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_rnu_id).get())->fillAll(0);
		double* optdepthe = ((pdat::NodeData<double> *) patch.getPatchData(d_optdepthe_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_optdepthe_id).get())->fillAll(0);
		double* optdeptha = ((pdat::NodeData<double> *) patch.getPatchData(d_optdeptha_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_optdeptha_id).get())->fillAll(0);
		double* optdepthx = ((pdat::NodeData<double> *) patch.getPatchData(d_optdepthx_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_optdepthx_id).get())->fillAll(0);
		double* chie = ((pdat::NodeData<double> *) patch.getPatchData(d_chie_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_chie_id).get())->fillAll(0);
		double* chia = ((pdat::NodeData<double> *) patch.getPatchData(d_chia_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_chia_id).get())->fillAll(0);
		double* chix = ((pdat::NodeData<double> *) patch.getPatchData(d_chix_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_chix_id).get())->fillAll(0);
		double* t_optdepthe = ((pdat::NodeData<double> *) patch.getPatchData(d_t_optdepthe_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_t_optdepthe_id).get())->fillAll(0);
		double* t_optdeptha = ((pdat::NodeData<double> *) patch.getPatchData(d_t_optdeptha_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_t_optdeptha_id).get())->fillAll(0);
		double* t_optdepthx = ((pdat::NodeData<double> *) patch.getPatchData(d_t_optdepthx_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_t_optdepthx_id).get())->fillAll(0);
		double* t_chie = ((pdat::NodeData<double> *) patch.getPatchData(d_t_chie_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_t_chie_id).get())->fillAll(0);
		double* t_chia = ((pdat::NodeData<double> *) patch.getPatchData(d_t_chia_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_t_chia_id).get())->fillAll(0);
		double* t_chix = ((pdat::NodeData<double> *) patch.getPatchData(d_t_chix_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_t_chix_id).get())->fillAll(0);
		double* Efu_x = ((pdat::NodeData<double> *) patch.getPatchData(d_Efu_x_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Efu_x_id).get())->fillAll(0);
		double* Efu_y = ((pdat::NodeData<double> *) patch.getPatchData(d_Efu_y_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Efu_y_id).get())->fillAll(0);
		double* Efu_z = ((pdat::NodeData<double> *) patch.getPatchData(d_Efu_z_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_Efu_z_id).get())->fillAll(0);
		double* TauN_x = ((pdat::NodeData<double> *) patch.getPatchData(d_TauN_x_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauN_x_id).get())->fillAll(0);
		double* TauN_y = ((pdat::NodeData<double> *) patch.getPatchData(d_TauN_y_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauN_y_id).get())->fillAll(0);
		double* TauN_z = ((pdat::NodeData<double> *) patch.getPatchData(d_TauN_z_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauN_z_id).get())->fillAll(0);
		double* TauNe_x = ((pdat::NodeData<double> *) patch.getPatchData(d_TauNe_x_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauNe_x_id).get())->fillAll(0);
		double* TauNe_y = ((pdat::NodeData<double> *) patch.getPatchData(d_TauNe_y_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauNe_y_id).get())->fillAll(0);
		double* TauNe_z = ((pdat::NodeData<double> *) patch.getPatchData(d_TauNe_z_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauNe_z_id).get())->fillAll(0);
		double* TauM_xy = ((pdat::NodeData<double> *) patch.getPatchData(d_TauM_xy_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauM_xy_id).get())->fillAll(0);
		double* TauM_xz = ((pdat::NodeData<double> *) patch.getPatchData(d_TauM_xz_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauM_xz_id).get())->fillAll(0);
		double* TauM_yz = ((pdat::NodeData<double> *) patch.getPatchData(d_TauM_yz_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauM_yz_id).get())->fillAll(0);
		double* TauT_xx = ((pdat::NodeData<double> *) patch.getPatchData(d_TauT_xx_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauT_xx_id).get())->fillAll(0);
		double* TauT_xy = ((pdat::NodeData<double> *) patch.getPatchData(d_TauT_xy_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauT_xy_id).get())->fillAll(0);
		double* TauT_xz = ((pdat::NodeData<double> *) patch.getPatchData(d_TauT_xz_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauT_xz_id).get())->fillAll(0);
		double* TauT_yy = ((pdat::NodeData<double> *) patch.getPatchData(d_TauT_yy_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauT_yy_id).get())->fillAll(0);
		double* TauT_yz = ((pdat::NodeData<double> *) patch.getPatchData(d_TauT_yz_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauT_yz_id).get())->fillAll(0);
		double* TauT_zz = ((pdat::NodeData<double> *) patch.getPatchData(d_TauT_zz_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_TauT_zz_id).get())->fillAll(0);
		double* dpfdeps = ((pdat::NodeData<double> *) patch.getPatchData(d_dpfdeps_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_dpfdeps_id).get())->fillAll(0);
		double* dpfdrho = ((pdat::NodeData<double> *) patch.getPatchData(d_dpfdrho_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_dpfdrho_id).get())->fillAll(0);
		double* dpfdye = ((pdat::NodeData<double> *) patch.getPatchData(d_dpfdye_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_dpfdye_id).get())->fillAll(0);
		double* qnu_a = ((pdat::NodeData<double> *) patch.getPatchData(d_qnu_a_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_qnu_a_id).get())->fillAll(0);
		double* qnu_e = ((pdat::NodeData<double> *) patch.getPatchData(d_qnu_e_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_qnu_e_id).get())->fillAll(0);
		double* qnu_x = ((pdat::NodeData<double> *) patch.getPatchData(d_qnu_x_id).get())->getPointer();
		((pdat::NodeData<double> *) patch.getPatchData(d_qnu_x_id).get())->fillAll(0);
		
		//Get the dimensions of the patch
		hier::Box pbox = patch.getBox();
		double* FOV_1 = ((pdat::NodeData<double> *) patch.getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch.getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch.getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch.getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch.getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch.getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch.getPatchData(d_FOV_zUpper_id).get())->getPointer();
		const hier::Index boxfirst = patch.getBox().lower();
		const hier::Index boxlast  = patch.getBox().upper();
		
		//Get delta spaces into an array. dx, dy, dz.
		const std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch.getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
		
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		SAMRAI_External_Data::external_loadData(76, &eos_type, &initialYe, &initialTemperature, &minTableTemperature, &minTableEnergy, &energyShift, &vacuum_ye_beta, &vacuum_P_reset, &vacuum_ye_reset, &vacuum_temp_reset, &vacuum_tau_reset, &vacuum_rho_reset, &vacuum_ye, &vacuum_D, &vacuum_tau, &ye_maximum, Tf, Yef, DYf, epsf, tauf, rhof, Df, vfd_x, vfd_y, vfd_z, Sfd_x, Sfd_y, Sfd_z, pf, sqcs, Bfu_x, Bfu_y, Bfu_z, gtd_xx, gtd_xy, gtd_xz, gtd_yy, gtd_yz, gtd_zz, Atd_xx, Atd_xy, Atd_xz, Atd_yy, Atd_yz, Atd_zz, Alpha, chi, trK, theta, Gamh_x, Gamh_y, Gamh_z, Betau_x, Betau_y, Betau_z, &gamma, &vacuum_rho, &tov_angle, &tov_Asize, &tov_facp, &rho_0, &rho_1, &rho_2, &a_0, &a_1, &a_2, &a_3, &K_0, &K_1, &K_2, &K_3, &gamma_0, &gamma_1, &gamma_2, &gamma_3, dx, d_grid_geometry->getXLower()[0], d_grid_geometry->getXLower()[1], d_grid_geometry->getXLower()[2], boxfirst[0], boxfirst[1], boxfirst[2], d_ghost_width, ilast, jlast, klast);
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					if (vector(FOV_1, i, j, k) > 0) {
						vector(optdepthe, i, j, k) = 0.0;
						vector(optdeptha, i, j, k) = 0.0;
						vector(optdepthx, i, j, k) = 0.0;
						vector(chie, i, j, k) = 0.0;
						vector(chia, i, j, k) = 0.0;
						vector(chix, i, j, k) = 0.0;
					}
		
				}
			}
		}
		

   	}
}



/*
 * Gets the coarser patch that contains the box
 */
const std::shared_ptr<hier::Patch >& Problem::getCoarserPatch(
	const std::shared_ptr< hier::PatchLevel >& level,
	const hier::Box interior, 
	const hier::IntVector ratio)
{
	const hier::Box& coarsenBox = hier::Box::coarsen(interior, ratio);
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr< hier::Patch >& patch = *p_it;
		const hier::Box& interior_C = patch->getBox();
		if (interior_C.intersects(coarsenBox)) {
			return patch;		
		}
	}
    return NULL;
}

/*
 * Reset the hierarchy-dependent internal information.
 */
void Problem::resetHierarchyConfiguration (
   const std::shared_ptr<hier::PatchHierarchy >& new_hierarchy ,
   int coarsest_level ,
   int finest_level )
{
	int finest_hiera_level = new_hierarchy->getFinestLevelNumber();

   	//  If we have added or removed a level, resize the schedule arrays

	d_bdry_sched_advance2.resize(finest_hiera_level+1);
	d_bdry_sched_advance5.resize(finest_hiera_level+1);
	d_bdry_sched_advance9.resize(finest_hiera_level+1);
	d_bdry_sched_advance12.resize(finest_hiera_level+1);
	d_bdry_sched_advance16.resize(finest_hiera_level+1);
	d_bdry_sched_advance19.resize(finest_hiera_level+1);
	d_bdry_sched_advance23.resize(finest_hiera_level+1);
	d_bdry_sched_advance26.resize(finest_hiera_level+1);
	d_bdry_sched_analysis1.resize(finest_hiera_level+1);
	d_coarsen_schedule.resize(finest_hiera_level+1);
	d_bdry_sched_postCoarsen.resize(finest_hiera_level+1);
	//  Build coarsen and refine communication schedules.
	for (int ln = coarsest_level; ln <= finest_hiera_level; ln++) {
		std::shared_ptr< hier::PatchLevel > level(new_hierarchy->getPatchLevel(ln));
		d_bdry_sched_advance2[ln] = d_bdry_fill_advance2->createSchedule(level,ln-1,new_hierarchy,this);
		d_bdry_sched_advance5[ln] = d_bdry_fill_advance5->createSchedule(level,ln-1,new_hierarchy,this);
		d_bdry_sched_advance9[ln] = d_bdry_fill_advance9->createSchedule(level,ln-1,new_hierarchy,this);
		d_bdry_sched_advance12[ln] = d_bdry_fill_advance12->createSchedule(level,ln-1,new_hierarchy,this);
		d_bdry_sched_advance16[ln] = d_bdry_fill_advance16->createSchedule(level,ln-1,new_hierarchy,this);
		d_bdry_sched_advance19[ln] = d_bdry_fill_advance19->createSchedule(level,ln-1,new_hierarchy,this);
		d_bdry_sched_advance23[ln] = d_bdry_fill_advance23->createSchedule(level,ln-1,new_hierarchy,this);
		d_bdry_sched_advance26[ln] = d_bdry_fill_advance26->createSchedule(level,ln-1,new_hierarchy,this);
		d_bdry_sched_analysis1[ln] = d_bdry_fill_analysis1->createSchedule(level,this);
		d_bdry_sched_postCoarsen[ln] = d_bdry_post_coarsen->createSchedule(level);
		// coarsen schedule only for levels > 0
		if (ln > 0) {
			std::shared_ptr< hier::PatchLevel > coarser_level(new_hierarchy->getPatchLevel(ln-1));
			d_coarsen_schedule[ln] = d_coarsen_algorithm->createSchedule(coarser_level, level, NULL);
		}
	}

}

/* 
 * Calculation of auxiliary terms after a new level is regridded.
 */
void Problem::postNewLevel(const std::shared_ptr< hier::PatchLevel >& level) {
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr< hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
		double* chi = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_id).get())->getPointer();
		double* Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_id).get())->getPointer();
		double* Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_id).get())->getPointer();
		double* tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_tauf_id).get())->getPointer();
		double* gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_id).get())->getPointer();
		double* gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_id).get())->getPointer();
		double* gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_id).get())->getPointer();
		double* gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_id).get())->getPointer();
		double* gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_id).get())->getPointer();
		double* gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_id).get())->getPointer();
		double* Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_id).get())->getPointer();
		double* Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_id).get())->getPointer();
		double* Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* Df = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_id).get())->getPointer();
		double* DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_DYf_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double chi_max, h, sdetg, Bu_x, Bu_y, Bu_z, tau, inv_chi, Bfd_y, Bfd_z, Bfd_x, decay_factor, detgtd, idetgtd, gtu_zz, gtu_xx, gtu_xz, gtu_xy, Sfu_x, Su_x, vfu_x, Sd_x, Sd_y, gtu_yy, Sd_z, gtu_yz, vfu_z, Sfu_y, Su_y, Sfu_z, Su_z, vfu_y, sqvf, sqW, W, Bd_x, D, Bd_y, Bd_z, rhof_tmp, Yef_tmp, Tf_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, dpfdye_tmp, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tauf_tmp, Df_tmp, DYf_tmp, Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, Efu_x_tmp, Efu_y_tmp, Efu_z_tmp, vacuum_D_tmp, vacuum_tau_tmp, vacuum_rho_reset_tmp;

		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
		const double simPlat_dt = patch->getPatchData(d_FOV_1_id)->getTime();

		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					if ((vector(FOV_1, i, j, k) > 0 || (((i + 1 < ilast && vector(FOV_1, i + 1, j, k) > 0) || (i + 2 < ilast && vector(FOV_1, i + 2, j, k) > 0) || (j + 1 < jlast && vector(FOV_1, i, j + 1, k) > 0) || (j + 2 < jlast && vector(FOV_1, i, j + 2, k) > 0) || (k + 1 < klast && vector(FOV_1, i, j, k + 1) > 0) || (k + 2 < klast && vector(FOV_1, i, j, k + 2) > 0)) || ((i - 1 >= 0 && vector(FOV_1, i - 1, j, k) > 0) || (i - 2 >= 0 && vector(FOV_1, i - 2, j, k) > 0) || (j - 1 >= 0 && vector(FOV_1, i, j - 1, k) > 0) || (j - 2 >= 0 && vector(FOV_1, i, j - 2, k) > 0) || (k - 1 >= 0 && vector(FOV_1, i, j, k - 1) > 0) || (k - 2 >= 0 && vector(FOV_1, i, j, k - 2) > 0))) )) {
						chi_max = MAX(chi_floor, vector(chi, i, j, k));
						h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						sdetg = pow(chi_max, (-1.500000000000000));
						Bu_x = vector(Bfu_x, i, j, k) / sdetg;
						Bu_y = vector(Bfu_y, i, j, k) / sdetg;
						Bu_z = vector(Bfu_z, i, j, k) / sdetg;
						tau = vector(tauf, i, j, k) / sdetg;
						inv_chi = 1.0 / fabs(chi_max);
						Bfd_y = inv_chi * (vector(gtd_xy, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_yz, i, j, k) * vector(Bfu_z, i, j, k));
						Bfd_z = inv_chi * (vector(gtd_xz, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_yz, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Bfu_z, i, j, k));
						Bfd_x = inv_chi * (vector(gtd_xx, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_xy, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_xz, i, j, k) * vector(Bfu_z, i, j, k));
						decay_factor = MIN(1.0, pow((R_0 / MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))), eta_damping_exp));
						detgtd = vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) * vector(gtd_xx, i, j, k) + (-vector(gtd_yy, i, j, k) * (vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k))) + (-vector(gtd_zz, i, j, k) * (vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k))) + (-(vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k)) * vector(gtd_xx, i, j, k)) + 2.0 * vector(gtd_yz, i, j, k) * vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k);
						idetgtd = 1.0 / detgtd;
						gtu_zz = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_yy, i, j, k) - vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k));
						gtu_xx = idetgtd * (vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k));
						gtu_xz = idetgtd * (vector(gtd_xy, i, j, k) * vector(gtd_yz, i, j, k) - vector(gtd_yy, i, j, k) * vector(gtd_xz, i, j, k));
						gtu_xy = idetgtd * ((-vector(gtd_xy, i, j, k) * vector(gtd_zz, i, j, k)) + vector(gtd_yz, i, j, k) * vector(gtd_xz, i, j, k));
						Sfu_x = chi_max * (gtu_xx * vector(Sfd_x, i, j, k) + gtu_xy * vector(Sfd_y, i, j, k) + gtu_xz * vector(Sfd_z, i, j, k));
						Su_x = Sfu_x / sdetg;
						vfu_x = chi_max * (gtu_xx * vector(vfd_x, i, j, k) + gtu_xy * vector(vfd_y, i, j, k) + gtu_xz * vector(vfd_z, i, j, k));
						Sd_x = vector(Sfd_x, i, j, k) / sdetg;
						Sd_y = vector(Sfd_y, i, j, k) / sdetg;
						gtu_yy = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k));
						Sd_z = vector(Sfd_z, i, j, k) / sdetg;
						gtu_yz = idetgtd * ((-vector(gtd_xx, i, j, k) * vector(gtd_yz, i, j, k)) + vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k));
						vfu_z = chi_max * (gtu_xz * vector(vfd_x, i, j, k) + gtu_yz * vector(vfd_y, i, j, k) + gtu_zz * vector(vfd_z, i, j, k));
						Sfu_y = chi_max * (gtu_xy * vector(Sfd_x, i, j, k) + gtu_yy * vector(Sfd_y, i, j, k) + gtu_yz * vector(Sfd_z, i, j, k));
						Su_y = Sfu_y / sdetg;
						Sfu_z = chi_max * (gtu_xz * vector(Sfd_x, i, j, k) + gtu_yz * vector(Sfd_y, i, j, k) + gtu_zz * vector(Sfd_z, i, j, k));
						Su_z = Sfu_z / sdetg;
						vfu_y = chi_max * (gtu_xy * vector(vfd_x, i, j, k) + gtu_yy * vector(vfd_y, i, j, k) + gtu_yz * vector(vfd_z, i, j, k));
						sqvf = MIN(vfu_x * vector(vfd_x, i, j, k) + vfu_y * vector(vfd_y, i, j, k) + vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						sqW = 1.0 / (1.0 - sqvf);
						W = sqrt(sqW);
						Bd_x = Bfd_x / sdetg;
						D = vector(Df, i, j, k) / sdetg;
						Bd_y = Bfd_y / sdetg;
						Bd_z = Bfd_z / sdetg;
						vector(t_chie, i, j, k) = vector(chie, i, j, k);
						vector(t_chia, i, j, k) = vector(chia, i, j, k);
						vector(t_optdepthx, i, j, k) = vector(optdepthx, i, j, k);
						vector(t_optdeptha, i, j, k) = vector(optdeptha, i, j, k);
						vector(t_optdepthe, i, j, k) = vector(optdepthe, i, j, k);
						vector(Alpha, i, j, k) = fabs(vector(Alpha, i, j, k));
						rhof_tmp = vector(rhof, i, j, k);
						Yef_tmp = vector(DYf, i, j, k) / vector(Df, i, j, k);
						if (Yef_tmp < vacuum_ye) {
							vector(DYf, i, j, k) = vacuum_ye_reset * vector(Df, i, j, k);
							Yef_tmp = vacuum_ye_reset;
						}
						if (Yef_tmp > ye_maximum) {
							vector(DYf, i, j, k) = 0.9 * ye_maximum * vector(Df, i, j, k);
							Yef_tmp = 0.9 * ye_maximum;
						}
						Tf_tmp = vector(Tf, i, j, k);
						pf_tmp = vector(pf, i, j, k);
						epsf_tmp = vector(epsf, i, j, k);
						vfd_x_tmp = vector(vfd_x, i, j, k);
						vfd_y_tmp = vector(vfd_y, i, j, k);
						vfd_z_tmp = vector(vfd_z, i, j, k);
						sqcs_tmp = vector(sqcs, i, j, k);
						dpfdrho_tmp = vector(dpfdrho, i, j, k);
						dpfdeps_tmp = vector(dpfdeps, i, j, k);
						dpfdye_tmp = vector(dpfdye, i, j, k);
						Sfd_x_tmp = vector(Sfd_x, i, j, k);
						Sfd_y_tmp = vector(Sfd_y, i, j, k);
						Sfd_z_tmp = vector(Sfd_z, i, j, k);
						tauf_tmp = vector(tauf, i, j, k);
						Df_tmp = vector(Df, i, j, k);
						DYf_tmp = vector(DYf, i, j, k);
						Bfu_x_tmp = vector(Bfu_x, i, j, k);
						Bfu_y_tmp = vector(Bfu_y, i, j, k);
						Bfu_z_tmp = vector(Bfu_z, i, j, k);
						if (equalsEq(externalCon2Prim, 1.0)) {
							Efu_x_tmp = vector(Efu_x, i, j, k);
							Efu_y_tmp = vector(Efu_y, i, j, k);
							Efu_z_tmp = vector(Efu_z, i, j, k);
							externalCon2prim(Efu_x_tmp, Efu_y_tmp, Efu_z_tmp, rhof_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, Df_tmp, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tauf_tmp, W, vector(chi, i, j, k), gtu_xx, gtu_xy, gtu_xz, gtu_yy, gtu_yz, gtu_zz, vector(gtd_xx, i, j, k), vector(gtd_xy, i, j, k), vector(gtd_xz, i, j, k), vector(gtd_yy, i, j, k), vector(gtd_yz, i, j, k), vector(gtd_zz, i, j, k), Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, dx, simPlat_dt, ilast, jlast);
							vector(Efu_x, i, j, k) = Efu_x_tmp;
							vector(Efu_y, i, j, k) = Efu_y_tmp;
							vector(Efu_z, i, j, k) = Efu_z_tmp;
						} else {
							vacuum_D_tmp = vacuum_D * decay_factor;
							vacuum_tau_tmp = vacuum_tau * decay_factor;
							vacuum_rho_reset_tmp = vacuum_rho_reset * decay_factor;
							con2prim(eos_type, rhof_tmp, Yef_tmp, Tf_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, dpfdye_tmp, rho_0, rho_1, rho_2, a_0, a_1, a_2, a_3, K_0, K_1, K_2, K_3, gamma_0, gamma_1, gamma_2, gamma_3, D, vacuum_D_tmp, Df_tmp, DYf_tmp, sdetg, Sd_x, Sd_y, Sd_z, Su_x, Su_y, Su_z, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tau, vacuum_tau_tmp, tauf_tmp, W, h, vector(chi, i, j, k), gtu_xx, gtu_xy, gtu_xz, gtu_yy, gtu_yz, gtu_zz, vector(gtd_xx, i, j, k), vector(gtd_xy, i, j, k), vector(gtd_xz, i, j, k), vector(gtd_yy, i, j, k), vector(gtd_yz, i, j, k), vector(gtd_zz, i, j, k), Bd_x, Bd_y, Bd_z, Bu_x, Bu_y, Bu_z, Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, max_error, gamma, inv_chi, threshold_vacuum, threshold_sqSmax, threshold_sqBmax, energyShift, vacuum_ye_beta, vacuum_rho, vacuum_rho_reset_tmp, vacuum_P_reset, vacuum_ye_reset, vacuum_temp_reset, vacuum_tau_reset, minTableTemperature, minTableEnergy, coord0_data_Data, coord0_dx_Data, coord0_size_Data, coord1_data_Data, coord1_dx_Data, coord1_size_Data, coord2_data_Data, coord2_dx_Data, coord2_size_Data, vars_data_Data, vars_size_Data, dx, simPlat_dt, ilast, jlast);
							vector(Efu_x, i, j, k) = -((-vector(vfd_z, i, j, k) * Bfd_y) + vector(vfd_y, i, j, k) * Bfd_z) / (sdetg * sdetg);
							vector(Efu_y, i, j, k) = -(vector(vfd_z, i, j, k) * Bfd_x - vector(vfd_x, i, j, k) * Bfd_z) / (sdetg * sdetg);
							vector(Efu_z, i, j, k) = -((-vector(vfd_y, i, j, k) * Bfd_x) + vector(vfd_x, i, j, k) * Bfd_y) / (sdetg * sdetg);
						}
						vector(rhof, i, j, k) = rhof_tmp;
						vector(Yef, i, j, k) = Yef_tmp;
						vector(Tf, i, j, k) = Tf_tmp;
						vector(pf, i, j, k) = pf_tmp;
						vector(epsf, i, j, k) = epsf_tmp;
						vector(vfd_x, i, j, k) = vfd_x_tmp;
						vector(vfd_y, i, j, k) = vfd_y_tmp;
						vector(vfd_z, i, j, k) = vfd_z_tmp;
						vector(sqcs, i, j, k) = sqcs_tmp;
						vector(dpfdye, i, j, k) = dpfdye_tmp;
						vector(dpfdeps, i, j, k) = dpfdeps_tmp;
						vector(dpfdrho, i, j, k) = dpfdrho_tmp;
						vector(Sfd_x, i, j, k) = Sfd_x_tmp;
						vector(Sfd_y, i, j, k) = Sfd_y_tmp;
						vector(Sfd_z, i, j, k) = Sfd_z_tmp;
						vector(tauf, i, j, k) = tauf_tmp;
						vector(Df, i, j, k) = Df_tmp;
						vector(DYf, i, j, k) = DYf_tmp;
						vector(Bfu_x, i, j, k) = Bfu_x_tmp;
						vector(Bfu_y, i, j, k) = Bfu_y_tmp;
						vector(Bfu_z, i, j, k) = Bfu_z_tmp;
						vector(t_chix, i, j, k) = vector(chix, i, j, k);
					}
				}
			}
		}
	}
}

/*
 * This method sets the physical boundary conditions.
 */
void Problem::setPhysicalBoundaryConditions(
   hier::Patch& patch,
   const double fill_time,
   const hier::IntVector& ghost_width_to_fill)
{
	//Boundary must not be implemented in this method
}

/*
 * Set up external plotter to plot internal data from this class.
 * Register variables appropriate for plotting.
 */
int Problem::setupPlotterMesh(appu::VisItDataWriter &plotter) const {
	if (!d_patch_hierarchy) {
		TBOX_ERROR(d_object_name << ": No hierarchy inn"
			<< " Problem::setupPlottern"
			<< "The hierarchy must be set before callingn"
			<< "this function.n");
	}
	plotter.registerPlotQuantity("FOV_1","SCALAR",d_FOV_1_id);
	hier::VariableDatabase *vdb = hier::VariableDatabase::getDatabase();
	for (set<string>::const_iterator it = d_full_mesh_writer_variables.begin() ; it != d_full_mesh_writer_variables.end(); ++it) {
		string var_to_register = *it;
		if (!(vdb->checkVariableExists(var_to_register))) {
			TBOX_ERROR(d_object_name << ": Variable selected for 3D write not found:" <<  var_to_register);
		}
		int var_id = vdb->getVariable(var_to_register)->getInstanceIdentifier();
		plotter.registerPlotQuantity(var_to_register,"SCALAR",var_id);
	}
	return 0;
}
/*
 * Set up external plotter to plot sliced data from this class.
 * Register variables appropriate for plotting.
 */
int Problem::setupSlicePlotter(vector<std::shared_ptr<SlicerDataWriter> > &plotters) const {
	if (!d_patch_hierarchy) {
	TBOX_ERROR(d_object_name << ": No hierarchy in\n"
		<< " Problem::setupSlicePlotter\n"
		<< "The hierarchy must be set before calling\n"
		<< "this function.\n");
	}
	int i = 0;
	for (vector<std::shared_ptr<SlicerDataWriter> >::const_iterator it = plotters.begin(); it != plotters.end(); ++it) {
		std::shared_ptr<SlicerDataWriter> plotter = *it;
		plotter->registerPlotQuantity("FOV_1","SCALAR",d_FOV_1_id);
		hier::VariableDatabase *vdb = hier::VariableDatabase::getDatabase();
		set<string> variables = d_sliceVariables[i];
		for (set<string>::const_iterator it2 = variables.begin() ; it2 != variables.end(); ++it2) {
			string var_to_register = *it2;
			if (!(vdb->checkVariableExists(var_to_register))) {
				TBOX_ERROR(d_object_name << ": Variable selected for Slice not found:" <<  var_to_register);
			}
			int var_id = vdb->getVariable(var_to_register)->getInstanceIdentifier();
			plotter->registerPlotQuantity(var_to_register,"SCALAR",var_id);
		}
		i++;
	}

	return 0;
}

/*
 * Set up external plotter to plot spherical data from this class.
 * Register variables appropriate for plotting.
 */
int Problem::setupSpherePlotter(vector<std::shared_ptr<SphereDataWriter> > &plotters) const {
	if (!d_patch_hierarchy) {
	TBOX_ERROR(d_object_name << ": No hierarchy in\n"
		<< " Problem::setupSpherePlotter\n"
		<< "The hierarchy must be set before calling\n"
		<< "this function.\n");
	}
	int i = 0;
	for (vector<std::shared_ptr<SphereDataWriter> >::const_iterator it = plotters.begin(); it != plotters.end(); ++it) {
		std::shared_ptr<SphereDataWriter> plotter = *it;
		plotter->registerPlotQuantity("FOV_1","SCALAR",d_FOV_1_id);
		hier::VariableDatabase *vdb = hier::VariableDatabase::getDatabase();
		set<string> variables = d_sphereVariables[i];
		for (set<string>::const_iterator it2 = variables.begin() ; it2 != variables.end(); ++it2) {
			string var_to_register = *it2;
			if (!(vdb->checkVariableExists(var_to_register))) {
				TBOX_ERROR(d_object_name << ": Variable selected for Sphere not found:" <<  var_to_register);
			}
			int var_id = vdb->getVariable(var_to_register)->getInstanceIdentifier();
			plotter->registerPlotQuantity(var_to_register,"SCALAR",var_id);
		}
		i++;
	}

	return 0;
}
/*
 * Set up external plotter to plot integration data from this class.
 * Register variables appropriate for plotting.
 */
int Problem::setupIntegralPlotter(vector<std::shared_ptr<IntegrateDataWriter> > &plotters) const {
	if (!d_patch_hierarchy) {
		TBOX_ERROR(d_object_name << ": No hierarchy inn"
		<< " Problem::setupIntegralPlottern"
		<< "The hierarchy must be set before callingn"
		<< "this function.n");
	}
	int i = 0;
	for (vector<std::shared_ptr<IntegrateDataWriter> >::const_iterator it = plotters.begin(); it != plotters.end(); ++it) {
		std::shared_ptr<IntegrateDataWriter> plotter = *it;
		hier::VariableDatabase *vdb = hier::VariableDatabase::getDatabase();
		set<string> variables = d_integralVariables[i];
		for (set<string>::const_iterator it2 = variables.begin() ; it2 != variables.end(); ++it2) {
			string var_to_register = *it2;
			if (!(vdb->checkVariableExists(var_to_register))) {
				TBOX_ERROR(d_object_name << ": Variable selected for Integration not found:" <<  var_to_register);
			}
			int var_id = vdb->getVariable(var_to_register)->getInstanceIdentifier();
			plotter->registerPlotQuantity(var_to_register,"SCALAR",var_id);
		}
		i++;
	}
	return 0;
}
/*
 * Set up external plotter to plot point data from this class.
 * Register variables appropriate for plotting.
 */
int Problem::setupPointPlotter(vector<std::shared_ptr<PointDataWriter> > &plotters) const {
	int i = 0;
	for (vector<std::shared_ptr<PointDataWriter> >::const_iterator it = plotters.begin(); it != plotters.end(); ++it) {
		std::shared_ptr<PointDataWriter> plotter = *it;
		hier::VariableDatabase *vdb = hier::VariableDatabase::getDatabase();
		set<string> variables = d_pointVariables[i];
		for (set<string>::const_iterator it2 = variables.begin() ; it2 != variables.end(); ++it2) {
			string var_to_register = *it2;
			if (!(vdb->checkVariableExists(var_to_register))) {
				TBOX_ERROR(d_object_name << ": Variable selected for Point not found:" <<  var_to_register);
			}
			int var_id = vdb->getVariable(var_to_register)->getInstanceIdentifier();
			plotter->registerPlotQuantity(var_to_register,"SCALAR",var_id);
		}
		i++;
	}
	return 0;
}


/*
 * Perform a single step from the discretization schema algorithm   
 */
double Problem::advanceLevel(
   const std::shared_ptr<hier::PatchLevel>& level,
   const std::shared_ptr<hier::PatchHierarchy>& hierarchy,
   const double current_time,
   const double new_time,
   const bool first_step,
   const bool last_step,
   const bool regrid_advance)
{
	const tbox::SAMRAI_MPI& mpi(tbox::SAMRAI_MPI::getSAMRAIWorld());

	const int ln = level->getLevelNumber();
	const double simPlat_dt = new_time - current_time;
	const double level_ratio = level->getRatioToCoarserLevel().max();
	if (first_step) {
		bo_substep_iteration[ln] = 0;
	}
	else {
		bo_substep_iteration[ln] = bo_substep_iteration[ln] + 1;
	}
	time_interpolate_operator_mesh1->setRatio(level_ratio);
	time_interpolate_operator_mesh1->setStep(0);
	time_interpolate_operator_mesh1->setTimeSubstepNumber(bo_substep_iteration[ln]);

	if (d_refinedTimeStepping && first_step && ln > 0) {
		current_iteration[ln] = (current_iteration[ln - 1] - 1) * hierarchy->getRatioToCoarserLevel(ln).max() + 1;
	} else {
		current_iteration[ln] = current_iteration[ln] + 1;
	}
	int previous_iteration = current_iteration[ln] - 1;
	int outputCycle = current_iteration[ln];
	int maxLevels = hierarchy->getMaxNumberOfLevels();
	if (maxLevels > ln + 1) {
		int currentLevelNumber = ln;
		while (currentLevelNumber < maxLevels - 1) {
			int ratio = hierarchy->getRatioToCoarserLevel(currentLevelNumber + 1).max();
			outputCycle = outputCycle * ratio;
			previous_iteration = previous_iteration * ratio;
			currentLevelNumber++;
		}
	}

	t_step->start();
  	// Shifting time
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch > patch = *p_it;
		std::shared_ptr< pdat::NodeData<double> > tauf(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_tauf_id)));
		std::shared_ptr< pdat::NodeData<double> > tauf_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_tauf_p_id)));
		std::shared_ptr< pdat::NodeData<double> > trK(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_trK_id)));
		std::shared_ptr< pdat::NodeData<double> > trK_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_trK_p_id)));
		std::shared_ptr< pdat::NodeData<double> > theta(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_theta_id)));
		std::shared_ptr< pdat::NodeData<double> > theta_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_theta_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Bfu_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Bfu_x_id)));
		std::shared_ptr< pdat::NodeData<double> > Bfu_x_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Bfu_x_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Bfu_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Bfu_y_id)));
		std::shared_ptr< pdat::NodeData<double> > Bfu_y_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Bfu_y_p_id)));
		std::shared_ptr< pdat::NodeData<double> > gtd_yy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_gtd_yy_id)));
		std::shared_ptr< pdat::NodeData<double> > gtd_yy_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_gtd_yy_p_id)));
		std::shared_ptr< pdat::NodeData<double> > gtd_yz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_gtd_yz_id)));
		std::shared_ptr< pdat::NodeData<double> > gtd_yz_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_gtd_yz_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Bfu_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Bfu_z_id)));
		std::shared_ptr< pdat::NodeData<double> > Bfu_z_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Bfu_z_p_id)));
		std::shared_ptr< pdat::NodeData<double> > phif(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_phif_id)));
		std::shared_ptr< pdat::NodeData<double> > phif_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_phif_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Df(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Df_id)));
		std::shared_ptr< pdat::NodeData<double> > Df_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Df_p_id)));
		std::shared_ptr< pdat::NodeData<double> > chi(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_chi_id)));
		std::shared_ptr< pdat::NodeData<double> > chi_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_chi_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Atd_zz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Atd_zz_id)));
		std::shared_ptr< pdat::NodeData<double> > Atd_zz_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Atd_zz_p_id)));
		std::shared_ptr< pdat::NodeData<double> > gtd_zz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_gtd_zz_id)));
		std::shared_ptr< pdat::NodeData<double> > gtd_zz_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_gtd_zz_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Sfd_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Sfd_y_id)));
		std::shared_ptr< pdat::NodeData<double> > Sfd_y_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Sfd_y_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Sfd_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Sfd_x_id)));
		std::shared_ptr< pdat::NodeData<double> > Sfd_x_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Sfd_x_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Sfd_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Sfd_z_id)));
		std::shared_ptr< pdat::NodeData<double> > Sfd_z_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Sfd_z_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Atd_yy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Atd_yy_id)));
		std::shared_ptr< pdat::NodeData<double> > Atd_yy_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Atd_yy_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Atd_yz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Atd_yz_id)));
		std::shared_ptr< pdat::NodeData<double> > Atd_yz_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Atd_yz_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Betau_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Betau_y_id)));
		std::shared_ptr< pdat::NodeData<double> > Betau_y_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Betau_y_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Betau_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Betau_x_id)));
		std::shared_ptr< pdat::NodeData<double> > Betau_x_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Betau_x_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Gamh_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Gamh_z_id)));
		std::shared_ptr< pdat::NodeData<double> > Gamh_z_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Gamh_z_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Betau_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Betau_z_id)));
		std::shared_ptr< pdat::NodeData<double> > Betau_z_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Betau_z_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Atd_xz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Atd_xz_id)));
		std::shared_ptr< pdat::NodeData<double> > Atd_xz_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Atd_xz_p_id)));
		std::shared_ptr< pdat::NodeData<double> > DYf(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_DYf_id)));
		std::shared_ptr< pdat::NodeData<double> > DYf_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_DYf_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Atd_xx(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Atd_xx_id)));
		std::shared_ptr< pdat::NodeData<double> > Atd_xx_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Atd_xx_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Atd_xy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Atd_xy_id)));
		std::shared_ptr< pdat::NodeData<double> > Atd_xy_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Atd_xy_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Alpha(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Alpha_id)));
		std::shared_ptr< pdat::NodeData<double> > Alpha_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Alpha_p_id)));
		std::shared_ptr< pdat::NodeData<double> > gtd_xx(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_gtd_xx_id)));
		std::shared_ptr< pdat::NodeData<double> > gtd_xx_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_gtd_xx_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Gamh_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Gamh_x_id)));
		std::shared_ptr< pdat::NodeData<double> > Gamh_x_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Gamh_x_p_id)));
		std::shared_ptr< pdat::NodeData<double> > gtd_xy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_gtd_xy_id)));
		std::shared_ptr< pdat::NodeData<double> > gtd_xy_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_gtd_xy_p_id)));
		std::shared_ptr< pdat::NodeData<double> > Gamh_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Gamh_y_id)));
		std::shared_ptr< pdat::NodeData<double> > Gamh_y_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_Gamh_y_p_id)));
		std::shared_ptr< pdat::NodeData<double> > gtd_xz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_gtd_xz_id)));
		std::shared_ptr< pdat::NodeData<double> > gtd_xz_p(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_gtd_xz_p_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1tauf(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1tauf_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1trK(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1trK_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1theta(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1theta_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Bfu_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Bfu_x_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Bfu_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Bfu_y_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1gtd_yy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1gtd_yy_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1gtd_yz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1gtd_yz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Bfu_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Bfu_z_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1phif(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1phif_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Df(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Df_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1chi(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1chi_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Atd_zz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Atd_zz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1gtd_zz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1gtd_zz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Sfd_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Sfd_y_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Sfd_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Sfd_x_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Sfd_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Sfd_z_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Atd_yy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Atd_yy_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Atd_yz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Atd_yz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Betau_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Betau_y_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Betau_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Betau_x_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Gamh_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Gamh_z_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Betau_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Betau_z_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Atd_xz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Atd_xz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1DYf(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1DYf_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Atd_xx(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Atd_xx_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Atd_xy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Atd_xy_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Alpha(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Alpha_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1gtd_xx(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1gtd_xx_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Gamh_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Gamh_x_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1gtd_xy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1gtd_xy_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1Gamh_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1Gamh_y_id)));
		std::shared_ptr< pdat::NodeData<double> > rk1gtd_xz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk1gtd_xz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2tauf(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2tauf_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2trK(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2trK_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2theta(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2theta_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Bfu_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Bfu_x_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Bfu_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Bfu_y_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2gtd_yy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2gtd_yy_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2gtd_yz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2gtd_yz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Bfu_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Bfu_z_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2phif(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2phif_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Df(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Df_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2chi(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2chi_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Atd_zz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Atd_zz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2gtd_zz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2gtd_zz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Sfd_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Sfd_y_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Sfd_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Sfd_x_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Sfd_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Sfd_z_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Atd_yy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Atd_yy_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Atd_yz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Atd_yz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Betau_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Betau_y_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Betau_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Betau_x_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Gamh_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Gamh_z_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Betau_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Betau_z_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Atd_xz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Atd_xz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2DYf(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2DYf_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Atd_xx(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Atd_xx_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Atd_xy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Atd_xy_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Alpha(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Alpha_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2gtd_xx(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2gtd_xx_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Gamh_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Gamh_x_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2gtd_xy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2gtd_xy_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2Gamh_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2Gamh_y_id)));
		std::shared_ptr< pdat::NodeData<double> > rk2gtd_xz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk2gtd_xz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3tauf(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3tauf_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3trK(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3trK_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3theta(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3theta_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Bfu_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Bfu_x_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Bfu_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Bfu_y_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3gtd_yy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3gtd_yy_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3gtd_yz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3gtd_yz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Bfu_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Bfu_z_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3phif(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3phif_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Df(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Df_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3chi(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3chi_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Atd_zz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Atd_zz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3gtd_zz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3gtd_zz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Sfd_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Sfd_y_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Sfd_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Sfd_x_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Sfd_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Sfd_z_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Atd_yy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Atd_yy_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Atd_yz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Atd_yz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Betau_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Betau_y_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Betau_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Betau_x_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Gamh_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Gamh_z_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Betau_z(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Betau_z_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Atd_xz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Atd_xz_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3DYf(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3DYf_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Atd_xx(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Atd_xx_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Atd_xy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Atd_xy_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Alpha(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Alpha_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3gtd_xx(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3gtd_xx_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Gamh_x(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Gamh_x_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3gtd_xy(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3gtd_xy_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3Gamh_y(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3Gamh_y_id)));
		std::shared_ptr< pdat::NodeData<double> > rk3gtd_xz(SAMRAI_SHARED_PTR_CAST<pdat::NodeData<double>, hier::PatchData>(patch->getPatchData(d_rk3gtd_xz_id)));
		gtd_xz_p->copy(*gtd_xz);
		Gamh_y_p->copy(*Gamh_y);
		gtd_xy_p->copy(*gtd_xy);
		Gamh_x_p->copy(*Gamh_x);
		gtd_xx_p->copy(*gtd_xx);
		Alpha_p->copy(*Alpha);
		Atd_xy_p->copy(*Atd_xy);
		Atd_xx_p->copy(*Atd_xx);
		DYf_p->copy(*DYf);
		Atd_xz_p->copy(*Atd_xz);
		Betau_z_p->copy(*Betau_z);
		Gamh_z_p->copy(*Gamh_z);
		Betau_x_p->copy(*Betau_x);
		Betau_y_p->copy(*Betau_y);
		Atd_yz_p->copy(*Atd_yz);
		Atd_yy_p->copy(*Atd_yy);
		Sfd_z_p->copy(*Sfd_z);
		Sfd_x_p->copy(*Sfd_x);
		Sfd_y_p->copy(*Sfd_y);
		gtd_zz_p->copy(*gtd_zz);
		Atd_zz_p->copy(*Atd_zz);
		chi_p->copy(*chi);
		Df_p->copy(*Df);
		phif_p->copy(*phif);
		Bfu_z_p->copy(*Bfu_z);
		gtd_yz_p->copy(*gtd_yz);
		gtd_yy_p->copy(*gtd_yy);
		Bfu_y_p->copy(*Bfu_y);
		Bfu_x_p->copy(*Bfu_x);
		theta_p->copy(*theta);
		trK_p->copy(*trK);
		tauf_p->copy(*tauf);
		tauf_p->setTime(current_time);
		trK_p->setTime(current_time);
		theta_p->setTime(current_time);
		Bfu_x_p->setTime(current_time);
		Bfu_y_p->setTime(current_time);
		gtd_yy_p->setTime(current_time);
		gtd_yz_p->setTime(current_time);
		Bfu_z_p->setTime(current_time);
		phif_p->setTime(current_time);
		Df_p->setTime(current_time);
		chi_p->setTime(current_time);
		Atd_zz_p->setTime(current_time);
		gtd_zz_p->setTime(current_time);
		Sfd_y_p->setTime(current_time);
		Sfd_x_p->setTime(current_time);
		Sfd_z_p->setTime(current_time);
		Atd_yy_p->setTime(current_time);
		Atd_yz_p->setTime(current_time);
		Betau_y_p->setTime(current_time);
		Betau_x_p->setTime(current_time);
		Gamh_z_p->setTime(current_time);
		Betau_z_p->setTime(current_time);
		Atd_xz_p->setTime(current_time);
		DYf_p->setTime(current_time);
		Atd_xx_p->setTime(current_time);
		Atd_xy_p->setTime(current_time);
		Alpha_p->setTime(current_time);
		gtd_xx_p->setTime(current_time);
		Gamh_x_p->setTime(current_time);
		gtd_xy_p->setTime(current_time);
		Gamh_y_p->setTime(current_time);
		gtd_xz_p->setTime(current_time);
		rk1tauf->setTime(current_time + simPlat_dt * 0.5);
		rk1trK->setTime(current_time + simPlat_dt * 0.5);
		rk1theta->setTime(current_time + simPlat_dt * 0.5);
		rk1Bfu_x->setTime(current_time + simPlat_dt * 0.5);
		rk1Bfu_y->setTime(current_time + simPlat_dt * 0.5);
		rk1gtd_yy->setTime(current_time + simPlat_dt * 0.5);
		rk1gtd_yz->setTime(current_time + simPlat_dt * 0.5);
		rk1Bfu_z->setTime(current_time + simPlat_dt * 0.5);
		rk1phif->setTime(current_time + simPlat_dt * 0.5);
		rk1Df->setTime(current_time + simPlat_dt * 0.5);
		rk1chi->setTime(current_time + simPlat_dt * 0.5);
		rk1Atd_zz->setTime(current_time + simPlat_dt * 0.5);
		rk1gtd_zz->setTime(current_time + simPlat_dt * 0.5);
		rk1Sfd_y->setTime(current_time + simPlat_dt * 0.5);
		rk1Sfd_x->setTime(current_time + simPlat_dt * 0.5);
		rk1Sfd_z->setTime(current_time + simPlat_dt * 0.5);
		rk1Atd_yy->setTime(current_time + simPlat_dt * 0.5);
		rk1Atd_yz->setTime(current_time + simPlat_dt * 0.5);
		rk1Betau_y->setTime(current_time + simPlat_dt * 0.5);
		rk1Betau_x->setTime(current_time + simPlat_dt * 0.5);
		rk1Gamh_z->setTime(current_time + simPlat_dt * 0.5);
		rk1Betau_z->setTime(current_time + simPlat_dt * 0.5);
		rk1Atd_xz->setTime(current_time + simPlat_dt * 0.5);
		rk1DYf->setTime(current_time + simPlat_dt * 0.5);
		rk1Atd_xx->setTime(current_time + simPlat_dt * 0.5);
		rk1Atd_xy->setTime(current_time + simPlat_dt * 0.5);
		rk1Alpha->setTime(current_time + simPlat_dt * 0.5);
		rk1gtd_xx->setTime(current_time + simPlat_dt * 0.5);
		rk1Gamh_x->setTime(current_time + simPlat_dt * 0.5);
		rk1gtd_xy->setTime(current_time + simPlat_dt * 0.5);
		rk1Gamh_y->setTime(current_time + simPlat_dt * 0.5);
		rk1gtd_xz->setTime(current_time + simPlat_dt * 0.5);
		rk2tauf->setTime(current_time + simPlat_dt * 0.5);
		rk2trK->setTime(current_time + simPlat_dt * 0.5);
		rk2theta->setTime(current_time + simPlat_dt * 0.5);
		rk2Bfu_x->setTime(current_time + simPlat_dt * 0.5);
		rk2Bfu_y->setTime(current_time + simPlat_dt * 0.5);
		rk2gtd_yy->setTime(current_time + simPlat_dt * 0.5);
		rk2gtd_yz->setTime(current_time + simPlat_dt * 0.5);
		rk2Bfu_z->setTime(current_time + simPlat_dt * 0.5);
		rk2phif->setTime(current_time + simPlat_dt * 0.5);
		rk2Df->setTime(current_time + simPlat_dt * 0.5);
		rk2chi->setTime(current_time + simPlat_dt * 0.5);
		rk2Atd_zz->setTime(current_time + simPlat_dt * 0.5);
		rk2gtd_zz->setTime(current_time + simPlat_dt * 0.5);
		rk2Sfd_y->setTime(current_time + simPlat_dt * 0.5);
		rk2Sfd_x->setTime(current_time + simPlat_dt * 0.5);
		rk2Sfd_z->setTime(current_time + simPlat_dt * 0.5);
		rk2Atd_yy->setTime(current_time + simPlat_dt * 0.5);
		rk2Atd_yz->setTime(current_time + simPlat_dt * 0.5);
		rk2Betau_y->setTime(current_time + simPlat_dt * 0.5);
		rk2Betau_x->setTime(current_time + simPlat_dt * 0.5);
		rk2Gamh_z->setTime(current_time + simPlat_dt * 0.5);
		rk2Betau_z->setTime(current_time + simPlat_dt * 0.5);
		rk2Atd_xz->setTime(current_time + simPlat_dt * 0.5);
		rk2DYf->setTime(current_time + simPlat_dt * 0.5);
		rk2Atd_xx->setTime(current_time + simPlat_dt * 0.5);
		rk2Atd_xy->setTime(current_time + simPlat_dt * 0.5);
		rk2Alpha->setTime(current_time + simPlat_dt * 0.5);
		rk2gtd_xx->setTime(current_time + simPlat_dt * 0.5);
		rk2Gamh_x->setTime(current_time + simPlat_dt * 0.5);
		rk2gtd_xy->setTime(current_time + simPlat_dt * 0.5);
		rk2Gamh_y->setTime(current_time + simPlat_dt * 0.5);
		rk2gtd_xz->setTime(current_time + simPlat_dt * 0.5);
		rk3tauf->setTime(current_time + simPlat_dt);
		rk3trK->setTime(current_time + simPlat_dt);
		rk3theta->setTime(current_time + simPlat_dt);
		rk3Bfu_x->setTime(current_time + simPlat_dt);
		rk3Bfu_y->setTime(current_time + simPlat_dt);
		rk3gtd_yy->setTime(current_time + simPlat_dt);
		rk3gtd_yz->setTime(current_time + simPlat_dt);
		rk3Bfu_z->setTime(current_time + simPlat_dt);
		rk3phif->setTime(current_time + simPlat_dt);
		rk3Df->setTime(current_time + simPlat_dt);
		rk3chi->setTime(current_time + simPlat_dt);
		rk3Atd_zz->setTime(current_time + simPlat_dt);
		rk3gtd_zz->setTime(current_time + simPlat_dt);
		rk3Sfd_y->setTime(current_time + simPlat_dt);
		rk3Sfd_x->setTime(current_time + simPlat_dt);
		rk3Sfd_z->setTime(current_time + simPlat_dt);
		rk3Atd_yy->setTime(current_time + simPlat_dt);
		rk3Atd_yz->setTime(current_time + simPlat_dt);
		rk3Betau_y->setTime(current_time + simPlat_dt);
		rk3Betau_x->setTime(current_time + simPlat_dt);
		rk3Gamh_z->setTime(current_time + simPlat_dt);
		rk3Betau_z->setTime(current_time + simPlat_dt);
		rk3Atd_xz->setTime(current_time + simPlat_dt);
		rk3DYf->setTime(current_time + simPlat_dt);
		rk3Atd_xx->setTime(current_time + simPlat_dt);
		rk3Atd_xy->setTime(current_time + simPlat_dt);
		rk3Alpha->setTime(current_time + simPlat_dt);
		rk3gtd_xx->setTime(current_time + simPlat_dt);
		rk3Gamh_x->setTime(current_time + simPlat_dt);
		rk3gtd_xy->setTime(current_time + simPlat_dt);
		rk3Gamh_y->setTime(current_time + simPlat_dt);
		rk3gtd_xz->setTime(current_time + simPlat_dt);
		tauf->setTime(current_time + simPlat_dt);
		trK->setTime(current_time + simPlat_dt);
		theta->setTime(current_time + simPlat_dt);
		Bfu_x->setTime(current_time + simPlat_dt);
		Bfu_y->setTime(current_time + simPlat_dt);
		gtd_yy->setTime(current_time + simPlat_dt);
		gtd_yz->setTime(current_time + simPlat_dt);
		Bfu_z->setTime(current_time + simPlat_dt);
		phif->setTime(current_time + simPlat_dt);
		Df->setTime(current_time + simPlat_dt);
		chi->setTime(current_time + simPlat_dt);
		Atd_zz->setTime(current_time + simPlat_dt);
		gtd_zz->setTime(current_time + simPlat_dt);
		Sfd_y->setTime(current_time + simPlat_dt);
		Sfd_x->setTime(current_time + simPlat_dt);
		Sfd_z->setTime(current_time + simPlat_dt);
		Atd_yy->setTime(current_time + simPlat_dt);
		Atd_yz->setTime(current_time + simPlat_dt);
		Betau_y->setTime(current_time + simPlat_dt);
		Betau_x->setTime(current_time + simPlat_dt);
		Gamh_z->setTime(current_time + simPlat_dt);
		Betau_z->setTime(current_time + simPlat_dt);
		Atd_xz->setTime(current_time + simPlat_dt);
		DYf->setTime(current_time + simPlat_dt);
		Atd_xx->setTime(current_time + simPlat_dt);
		Atd_xy->setTime(current_time + simPlat_dt);
		Alpha->setTime(current_time + simPlat_dt);
		gtd_xx->setTime(current_time + simPlat_dt);
		Gamh_x->setTime(current_time + simPlat_dt);
		gtd_xy->setTime(current_time + simPlat_dt);
		Gamh_y->setTime(current_time + simPlat_dt);
		gtd_xz->setTime(current_time + simPlat_dt);
	}
  	// Evolution
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		double* chi_p = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_p_id).get())->getPointer();
		double* gtd_xy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_p_id).get())->getPointer();
		double* Bfu_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_p_id).get())->getPointer();
		double* gtd_yy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_p_id).get())->getPointer();
		double* Bfu_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_p_id).get())->getPointer();
		double* gtd_yz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_p_id).get())->getPointer();
		double* Bfu_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_p_id).get())->getPointer();
		double* gtd_xz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_p_id).get())->getPointer();
		double* gtd_zz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_p_id).get())->getPointer();
		double* gtd_xx_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_p_id).get())->getPointer();
		double* Sfd_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_p_id).get())->getPointer();
		double* Sfd_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_p_id).get())->getPointer();
		double* Sfd_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_p_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* Betau_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_p_id).get())->getPointer();
		double* Alpha_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_p_id).get())->getPointer();
		double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
		double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
		double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
		double* phif_p = ((pdat::NodeData<double> *) patch->getPatchData(d_phif_p_id).get())->getPointer();
		double* Betau_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_p_id).get())->getPointer();
		double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
		double* Betau_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_p_id).get())->getPointer();
		double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
		double* Df_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_p_id).get())->getPointer();
		double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
		double* DYf_p = ((pdat::NodeData<double> *) patch->getPatchData(d_DYf_p_id).get())->getPointer();
		double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
		double* tauf_p = ((pdat::NodeData<double> *) patch->getPatchData(d_tauf_p_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
		double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
		double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
		double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
		double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
		double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
		double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
		double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
		double* Gamh_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_x_p_id).get())->getPointer();
		double* Gamh_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_y_p_id).get())->getPointer();
		double* Gamh_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_z_p_id).get())->getPointer();
		double* Atd_xx_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xx_p_id).get())->getPointer();
		double* Atd_xy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xy_p_id).get())->getPointer();
		double* Atd_xz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xz_p_id).get())->getPointer();
		double* Atd_yy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yy_p_id).get())->getPointer();
		double* Atd_yz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yz_p_id).get())->getPointer();
		double* Atd_zz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_zz_p_id).get())->getPointer();
		double* trK_p = ((pdat::NodeData<double> *) patch->getPatchData(d_trK_p_id).get())->getPointer();
		double* theta_p = ((pdat::NodeData<double> *) patch->getPatchData(d_theta_p_id).get())->getPointer();
		double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
		double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
		double* rk1Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_x_id).get())->getPointer();
		double* rk1Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_y_id).get())->getPointer();
		double* rk1Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_z_id).get())->getPointer();
		double* rk1Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_x_id).get())->getPointer();
		double* rk1Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_y_id).get())->getPointer();
		double* rk1Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_z_id).get())->getPointer();
		double* rk1Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Df_id).get())->getPointer();
		double* rk1DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1DYf_id).get())->getPointer();
		double* rk1tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1tauf_id).get())->getPointer();
		double* rk1phif = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1phif_id).get())->getPointer();
		double* rk1gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xx_id).get())->getPointer();
		double* rk1gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xy_id).get())->getPointer();
		double* rk1gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xz_id).get())->getPointer();
		double* rk1gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yy_id).get())->getPointer();
		double* rk1gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yz_id).get())->getPointer();
		double* rk1gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_zz_id).get())->getPointer();
		double* rk1Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xx_id).get())->getPointer();
		double* rk1Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xy_id).get())->getPointer();
		double* rk1Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xz_id).get())->getPointer();
		double* rk1Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_yy_id).get())->getPointer();
		double* rk1Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_yz_id).get())->getPointer();
		double* rk1Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_zz_id).get())->getPointer();
		double* rk1Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_x_id).get())->getPointer();
		double* rk1Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_y_id).get())->getPointer();
		double* rk1Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_z_id).get())->getPointer();
		double* rk1Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_x_id).get())->getPointer();
		double* rk1Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_y_id).get())->getPointer();
		double* rk1Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_z_id).get())->getPointer();
		double* rk1Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Alpha_id).get())->getPointer();
		double* rk1chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1chi_id).get())->getPointer();
		double* rk1trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1trK_id).get())->getPointer();
		double* rk1theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1theta_id).get())->getPointer();
		double chi_max_p, sdetg_p, inv_chi_p, Bfd_y_p, Bfd_z_p, Bfd_x_p, sqBf_p, detgtd_p, idetgtd_p, gtu_zz_p, gtu_xx_p, gtu_xz_p, gtu_xy_p, Sfu_x_p, vfu_x_p, Bfvf_p, gtu_yy_p, gtu_yz_p, vfu_z_p, Sfu_y_p, Sfu_z_p, vfu_y_p, sqvf_p, sqW_p, Sfuu_zz_p, Sfuu_xx_p, Sfuu_xz_p, Sfuu_xy_p, Sfud_xx_p, Sfud_xy_p, Sfud_xz_p, Sfuu_yz_p, Sfud_zy_p, Sfud_zz_p, Sfud_zx_p, Sfuu_yy_p, Sfud_yx_p, Sfud_yy_p, Sfud_yz_p, FluxiSfd_x, FluxiSfd_y, FluxiSfd_z, FluxiBfu_x, FluxiBfu_y, FluxiBfu_z, FluxiDf, FluxiDYf, Fluxitauf, Fluxiphif, h_p, sqbc_p, sqca_p, sqa_p, SpeediSfd_x, SpeediSfd_y, SpeediSfd_z, SpeediBfu_x, SpeediBfu_y, SpeediBfu_z, SpeediDf, SpeediDYf, Speeditauf, Speediphif, Fluxm3iSfd_x, Fluxm3iSfd_y, Fluxm3iSfd_z, Fluxm3iBfu_x, Fluxm3iBfu_y, Fluxm3iBfu_z, Fluxm3iDf, Fluxm3iDYf, Fluxm3itauf, Fluxm3iphif, Fluxm2iSfd_x, Fluxm2iSfd_y, Fluxm2iSfd_z, Fluxm2iBfu_x, Fluxm2iBfu_y, Fluxm2iBfu_z, Fluxm2iDf, Fluxm2iDYf, Fluxm2itauf, Fluxm2iphif, Fluxm1iSfd_x, Fluxm1iSfd_y, Fluxm1iSfd_z, Fluxm1iBfu_x, Fluxm1iBfu_y, Fluxm1iBfu_z, Fluxm1iDf, Fluxm1iDYf, Fluxm1itauf, Fluxm1iphif, Fluxp1iSfd_x, Fluxp1iSfd_y, Fluxp1iSfd_z, Fluxp1iBfu_x, Fluxp1iBfu_y, Fluxp1iBfu_z, Fluxp1iDf, Fluxp1iDYf, Fluxp1itauf, Fluxp1iphif, Fluxp2iSfd_x, Fluxp2iSfd_y, Fluxp2iSfd_z, Fluxp2iBfu_x, Fluxp2iBfu_y, Fluxp2iBfu_z, Fluxp2iDf, Fluxp2iDYf, Fluxp2itauf, Fluxp2iphif, Fluxp3iSfd_x, Fluxp3iSfd_y, Fluxp3iSfd_z, Fluxp3iBfu_x, Fluxp3iBfu_y, Fluxp3iBfu_z, Fluxp3iDf, Fluxp3iDYf, Fluxp3itauf, Fluxp3iphif, LLFpiSfd_x, LLFpiSfd_y, LLFpiSfd_z, LLFpiBfu_x, LLFpiBfu_y, LLFpiBfu_z, LLFpiDf, LLFpiDYf, LLFpitauf, LLFpiphif, LLFniSfd_x, LLFniSfd_y, LLFniSfd_z, LLFniBfu_x, LLFniBfu_y, LLFniBfu_z, LLFniDf, LLFniDYf, LLFnitauf, LLFniphif, LLFp_m3iSfd_x, LLFp_m3iSfd_y, LLFp_m3iSfd_z, LLFp_m3iBfu_x, LLFp_m3iBfu_y, LLFp_m3iBfu_z, LLFp_m3iDf, LLFp_m3iDYf, LLFp_m3itauf, LLFp_m3iphif, LLFn_m3iSfd_x, LLFn_m3iSfd_y, LLFn_m3iSfd_z, LLFn_m3iBfu_x, LLFn_m3iBfu_y, LLFn_m3iBfu_z, LLFn_m3iDf, LLFn_m3iDYf, LLFn_m3itauf, LLFn_m3iphif, LLFp_m2iSfd_x, LLFp_m2iSfd_y, LLFp_m2iSfd_z, LLFp_m2iBfu_x, LLFp_m2iBfu_y, LLFp_m2iBfu_z, LLFp_m2iDf, LLFp_m2iDYf, LLFp_m2itauf, LLFp_m2iphif, LLFn_m2iSfd_x, LLFn_m2iSfd_y, LLFn_m2iSfd_z, LLFn_m2iBfu_x, LLFn_m2iBfu_y, LLFn_m2iBfu_z, LLFn_m2iDf, LLFn_m2iDYf, LLFn_m2itauf, LLFn_m2iphif, LLFp_m1iSfd_x, LLFp_m1iSfd_y, LLFp_m1iSfd_z, LLFp_m1iBfu_x, LLFp_m1iBfu_y, LLFp_m1iBfu_z, LLFp_m1iDf, LLFp_m1iDYf, LLFp_m1itauf, LLFp_m1iphif, LLFn_m1iSfd_x, LLFn_m1iSfd_y, LLFn_m1iSfd_z, LLFn_m1iBfu_x, LLFn_m1iBfu_y, LLFn_m1iBfu_z, LLFn_m1iDf, LLFn_m1iDYf, LLFn_m1itauf, LLFn_m1iphif, LLFp_p1iSfd_x, LLFp_p1iSfd_y, LLFp_p1iSfd_z, LLFp_p1iBfu_x, LLFp_p1iBfu_y, LLFp_p1iBfu_z, LLFp_p1iDf, LLFp_p1iDYf, LLFp_p1itauf, LLFp_p1iphif, LLFn_p1iSfd_x, LLFn_p1iSfd_y, LLFn_p1iSfd_z, LLFn_p1iBfu_x, LLFn_p1iBfu_y, LLFn_p1iBfu_z, LLFn_p1iDf, LLFn_p1iDYf, LLFn_p1itauf, LLFn_p1iphif, LLFp_p2iSfd_x, LLFp_p2iSfd_y, LLFp_p2iSfd_z, LLFp_p2iBfu_x, LLFp_p2iBfu_y, LLFp_p2iBfu_z, LLFp_p2iDf, LLFp_p2iDYf, LLFp_p2itauf, LLFp_p2iphif, LLFn_p2iSfd_x, LLFn_p2iSfd_y, LLFn_p2iSfd_z, LLFn_p2iBfu_x, LLFn_p2iBfu_y, LLFn_p2iBfu_z, LLFn_p2iDf, LLFn_p2iDYf, LLFn_p2itauf, LLFn_p2iphif, LLFp_p3iSfd_x, LLFp_p3iSfd_y, LLFp_p3iSfd_z, LLFp_p3iBfu_x, LLFp_p3iBfu_y, LLFp_p3iBfu_z, LLFp_p3iDf, LLFp_p3iDYf, LLFp_p3itauf, LLFp_p3iphif, LLFn_p3iSfd_x, LLFn_p3iSfd_y, LLFn_p3iSfd_z, LLFn_p3iBfu_x, LLFn_p3iBfu_y, LLFn_p3iBfu_z, LLFn_p3iDf, LLFn_p3iDYf, LLFn_p3itauf, LLFn_p3iphif, FluxjSfd_x, FluxjSfd_y, FluxjSfd_z, FluxjBfu_x, FluxjBfu_y, FluxjBfu_z, FluxjDf, FluxjDYf, Fluxjtauf, Fluxjphif, SpeedjSfd_x, SpeedjSfd_y, SpeedjSfd_z, SpeedjBfu_x, SpeedjBfu_y, SpeedjBfu_z, SpeedjDf, SpeedjDYf, Speedjtauf, Speedjphif, Fluxm3jSfd_x, Fluxm3jSfd_y, Fluxm3jSfd_z, Fluxm3jBfu_x, Fluxm3jBfu_y, Fluxm3jBfu_z, Fluxm3jDf, Fluxm3jDYf, Fluxm3jtauf, Fluxm3jphif, Fluxm2jSfd_x, Fluxm2jSfd_y, Fluxm2jSfd_z, Fluxm2jBfu_x, Fluxm2jBfu_y, Fluxm2jBfu_z, Fluxm2jDf, Fluxm2jDYf, Fluxm2jtauf, Fluxm2jphif, Fluxm1jSfd_x, Fluxm1jSfd_y, Fluxm1jSfd_z, Fluxm1jBfu_x, Fluxm1jBfu_y, Fluxm1jBfu_z, Fluxm1jDf, Fluxm1jDYf, Fluxm1jtauf, Fluxm1jphif, Fluxp1jSfd_x, Fluxp1jSfd_y, Fluxp1jSfd_z, Fluxp1jBfu_x, Fluxp1jBfu_y, Fluxp1jBfu_z, Fluxp1jDf, Fluxp1jDYf, Fluxp1jtauf, Fluxp1jphif, Fluxp2jSfd_x, Fluxp2jSfd_y, Fluxp2jSfd_z, Fluxp2jBfu_x, Fluxp2jBfu_y, Fluxp2jBfu_z, Fluxp2jDf, Fluxp2jDYf, Fluxp2jtauf, Fluxp2jphif, Fluxp3jSfd_x, Fluxp3jSfd_y, Fluxp3jSfd_z, Fluxp3jBfu_x, Fluxp3jBfu_y, Fluxp3jBfu_z, Fluxp3jDf, Fluxp3jDYf, Fluxp3jtauf, Fluxp3jphif, LLFpjSfd_x, LLFpjSfd_y, LLFpjSfd_z, LLFpjBfu_x, LLFpjBfu_y, LLFpjBfu_z, LLFpjDf, LLFpjDYf, LLFpjtauf, LLFpjphif, LLFnjSfd_x, LLFnjSfd_y, LLFnjSfd_z, LLFnjBfu_x, LLFnjBfu_y, LLFnjBfu_z, LLFnjDf, LLFnjDYf, LLFnjtauf, LLFnjphif, LLFp_m3jSfd_x, LLFp_m3jSfd_y, LLFp_m3jSfd_z, LLFp_m3jBfu_x, LLFp_m3jBfu_y, LLFp_m3jBfu_z, LLFp_m3jDf, LLFp_m3jDYf, LLFp_m3jtauf, LLFp_m3jphif, LLFn_m3jSfd_x, LLFn_m3jSfd_y, LLFn_m3jSfd_z, LLFn_m3jBfu_x, LLFn_m3jBfu_y, LLFn_m3jBfu_z, LLFn_m3jDf, LLFn_m3jDYf, LLFn_m3jtauf, LLFn_m3jphif, LLFp_m2jSfd_x, LLFp_m2jSfd_y, LLFp_m2jSfd_z, LLFp_m2jBfu_x, LLFp_m2jBfu_y, LLFp_m2jBfu_z, LLFp_m2jDf, LLFp_m2jDYf, LLFp_m2jtauf, LLFp_m2jphif, LLFn_m2jSfd_x, LLFn_m2jSfd_y, LLFn_m2jSfd_z, LLFn_m2jBfu_x, LLFn_m2jBfu_y, LLFn_m2jBfu_z, LLFn_m2jDf, LLFn_m2jDYf, LLFn_m2jtauf, LLFn_m2jphif, LLFp_m1jSfd_x, LLFp_m1jSfd_y, LLFp_m1jSfd_z, LLFp_m1jBfu_x, LLFp_m1jBfu_y, LLFp_m1jBfu_z, LLFp_m1jDf, LLFp_m1jDYf, LLFp_m1jtauf, LLFp_m1jphif, LLFn_m1jSfd_x, LLFn_m1jSfd_y, LLFn_m1jSfd_z, LLFn_m1jBfu_x, LLFn_m1jBfu_y, LLFn_m1jBfu_z, LLFn_m1jDf, LLFn_m1jDYf, LLFn_m1jtauf, LLFn_m1jphif, LLFp_p1jSfd_x, LLFp_p1jSfd_y, LLFp_p1jSfd_z, LLFp_p1jBfu_x, LLFp_p1jBfu_y, LLFp_p1jBfu_z, LLFp_p1jDf, LLFp_p1jDYf, LLFp_p1jtauf, LLFp_p1jphif, LLFn_p1jSfd_x, LLFn_p1jSfd_y, LLFn_p1jSfd_z, LLFn_p1jBfu_x, LLFn_p1jBfu_y, LLFn_p1jBfu_z, LLFn_p1jDf, LLFn_p1jDYf, LLFn_p1jtauf, LLFn_p1jphif, LLFp_p2jSfd_x, LLFp_p2jSfd_y, LLFp_p2jSfd_z, LLFp_p2jBfu_x, LLFp_p2jBfu_y, LLFp_p2jBfu_z, LLFp_p2jDf, LLFp_p2jDYf, LLFp_p2jtauf, LLFp_p2jphif, LLFn_p2jSfd_x, LLFn_p2jSfd_y, LLFn_p2jSfd_z, LLFn_p2jBfu_x, LLFn_p2jBfu_y, LLFn_p2jBfu_z, LLFn_p2jDf, LLFn_p2jDYf, LLFn_p2jtauf, LLFn_p2jphif, LLFp_p3jSfd_x, LLFp_p3jSfd_y, LLFp_p3jSfd_z, LLFp_p3jBfu_x, LLFp_p3jBfu_y, LLFp_p3jBfu_z, LLFp_p3jDf, LLFp_p3jDYf, LLFp_p3jtauf, LLFp_p3jphif, LLFn_p3jSfd_x, LLFn_p3jSfd_y, LLFn_p3jSfd_z, LLFn_p3jBfu_x, LLFn_p3jBfu_y, LLFn_p3jBfu_z, LLFn_p3jDf, LLFn_p3jDYf, LLFn_p3jtauf, LLFn_p3jphif, FluxkSfd_x, FluxkSfd_y, FluxkSfd_z, FluxkBfu_x, FluxkBfu_y, FluxkBfu_z, FluxkDf, FluxkDYf, Fluxktauf, Fluxkphif, SpeedkSfd_x, SpeedkSfd_y, SpeedkSfd_z, SpeedkBfu_x, SpeedkBfu_y, SpeedkBfu_z, SpeedkDf, SpeedkDYf, Speedktauf, Speedkphif, Fluxm3kSfd_x, Fluxm3kSfd_y, Fluxm3kSfd_z, Fluxm3kBfu_x, Fluxm3kBfu_y, Fluxm3kBfu_z, Fluxm3kDf, Fluxm3kDYf, Fluxm3ktauf, Fluxm3kphif, Fluxm2kSfd_x, Fluxm2kSfd_y, Fluxm2kSfd_z, Fluxm2kBfu_x, Fluxm2kBfu_y, Fluxm2kBfu_z, Fluxm2kDf, Fluxm2kDYf, Fluxm2ktauf, Fluxm2kphif, Fluxm1kSfd_x, Fluxm1kSfd_y, Fluxm1kSfd_z, Fluxm1kBfu_x, Fluxm1kBfu_y, Fluxm1kBfu_z, Fluxm1kDf, Fluxm1kDYf, Fluxm1ktauf, Fluxm1kphif, Fluxp1kSfd_x, Fluxp1kSfd_y, Fluxp1kSfd_z, Fluxp1kBfu_x, Fluxp1kBfu_y, Fluxp1kBfu_z, Fluxp1kDf, Fluxp1kDYf, Fluxp1ktauf, Fluxp1kphif, Fluxp2kSfd_x, Fluxp2kSfd_y, Fluxp2kSfd_z, Fluxp2kBfu_x, Fluxp2kBfu_y, Fluxp2kBfu_z, Fluxp2kDf, Fluxp2kDYf, Fluxp2ktauf, Fluxp2kphif, Fluxp3kSfd_x, Fluxp3kSfd_y, Fluxp3kSfd_z, Fluxp3kBfu_x, Fluxp3kBfu_y, Fluxp3kBfu_z, Fluxp3kDf, Fluxp3kDYf, Fluxp3ktauf, Fluxp3kphif, LLFpkSfd_x, LLFpkSfd_y, LLFpkSfd_z, LLFpkBfu_x, LLFpkBfu_y, LLFpkBfu_z, LLFpkDf, LLFpkDYf, LLFpktauf, LLFpkphif, LLFnkSfd_x, LLFnkSfd_y, LLFnkSfd_z, LLFnkBfu_x, LLFnkBfu_y, LLFnkBfu_z, LLFnkDf, LLFnkDYf, LLFnktauf, LLFnkphif, LLFp_m3kSfd_x, LLFp_m3kSfd_y, LLFp_m3kSfd_z, LLFp_m3kBfu_x, LLFp_m3kBfu_y, LLFp_m3kBfu_z, LLFp_m3kDf, LLFp_m3kDYf, LLFp_m3ktauf, LLFp_m3kphif, LLFn_m3kSfd_x, LLFn_m3kSfd_y, LLFn_m3kSfd_z, LLFn_m3kBfu_x, LLFn_m3kBfu_y, LLFn_m3kBfu_z, LLFn_m3kDf, LLFn_m3kDYf, LLFn_m3ktauf, LLFn_m3kphif, LLFp_m2kSfd_x, LLFp_m2kSfd_y, LLFp_m2kSfd_z, LLFp_m2kBfu_x, LLFp_m2kBfu_y, LLFp_m2kBfu_z, LLFp_m2kDf, LLFp_m2kDYf, LLFp_m2ktauf, LLFp_m2kphif, LLFn_m2kSfd_x, LLFn_m2kSfd_y, LLFn_m2kSfd_z, LLFn_m2kBfu_x, LLFn_m2kBfu_y, LLFn_m2kBfu_z, LLFn_m2kDf, LLFn_m2kDYf, LLFn_m2ktauf, LLFn_m2kphif, LLFp_m1kSfd_x, LLFp_m1kSfd_y, LLFp_m1kSfd_z, LLFp_m1kBfu_x, LLFp_m1kBfu_y, LLFp_m1kBfu_z, LLFp_m1kDf, LLFp_m1kDYf, LLFp_m1ktauf, LLFp_m1kphif, LLFn_m1kSfd_x, LLFn_m1kSfd_y, LLFn_m1kSfd_z, LLFn_m1kBfu_x, LLFn_m1kBfu_y, LLFn_m1kBfu_z, LLFn_m1kDf, LLFn_m1kDYf, LLFn_m1ktauf, LLFn_m1kphif, LLFp_p1kSfd_x, LLFp_p1kSfd_y, LLFp_p1kSfd_z, LLFp_p1kBfu_x, LLFp_p1kBfu_y, LLFp_p1kBfu_z, LLFp_p1kDf, LLFp_p1kDYf, LLFp_p1ktauf, LLFp_p1kphif, LLFn_p1kSfd_x, LLFn_p1kSfd_y, LLFn_p1kSfd_z, LLFn_p1kBfu_x, LLFn_p1kBfu_y, LLFn_p1kBfu_z, LLFn_p1kDf, LLFn_p1kDYf, LLFn_p1ktauf, LLFn_p1kphif, LLFp_p2kSfd_x, LLFp_p2kSfd_y, LLFp_p2kSfd_z, LLFp_p2kBfu_x, LLFp_p2kBfu_y, LLFp_p2kBfu_z, LLFp_p2kDf, LLFp_p2kDYf, LLFp_p2ktauf, LLFp_p2kphif, LLFn_p2kSfd_x, LLFn_p2kSfd_y, LLFn_p2kSfd_z, LLFn_p2kBfu_x, LLFn_p2kBfu_y, LLFn_p2kBfu_z, LLFn_p2kDf, LLFn_p2kDYf, LLFn_p2ktauf, LLFn_p2kphif, LLFp_p3kSfd_x, LLFp_p3kSfd_y, LLFp_p3kSfd_z, LLFp_p3kBfu_x, LLFp_p3kBfu_y, LLFp_p3kBfu_z, LLFp_p3kDf, LLFp_p3kDYf, LLFp_p3ktauf, LLFp_p3kphif, LLFn_p3kSfd_x, LLFn_p3kSfd_y, LLFn_p3kSfd_z, LLFn_p3kBfu_x, LLFn_p3kBfu_y, LLFn_p3kBfu_z, LLFn_p3kDf, LLFn_p3kDYf, LLFn_p3ktauf, LLFn_p3kphif, Sfd_x_lni, Sfd_y_lni, Sfd_z_lni, Bfu_x_lni, Bfu_y_lni, Bfu_z_lni, Df_lni, DYf_lni, tauf_lni, phif_lni, Sfd_x_lnj, Sfd_y_lnj, Sfd_z_lnj, Bfu_x_lnj, Bfu_y_lnj, Bfu_z_lnj, Df_lnj, DYf_lnj, tauf_lnj, phif_lnj, Sfd_x_lnk, Sfd_y_lnk, Sfd_z_lnk, Bfu_x_lnk, Bfu_y_lnk, Bfu_z_lnk, Df_lnk, DYf_lnk, tauf_lnk, phif_lnk, Sfd_x_rni, Sfd_y_rni, Sfd_z_rni, Bfu_x_rni, Bfu_y_rni, Bfu_z_rni, Df_rni, DYf_rni, tauf_rni, phif_rni, Sfd_x_rnj, Sfd_y_rnj, Sfd_z_rnj, Bfu_x_rnj, Bfu_y_rnj, Bfu_z_rnj, Df_rnj, DYf_rnj, tauf_rnj, phif_rnj, Sfd_x_rnk, Sfd_y_rnk, Sfd_z_rnk, Bfu_x_rnk, Bfu_y_rnk, Bfu_z_rnk, Df_rnk, DYf_rnk, tauf_rnk, phif_rnk, Sfd_x_lpi, Sfd_y_lpi, Sfd_z_lpi, Bfu_x_lpi, Bfu_y_lpi, Bfu_z_lpi, Df_lpi, DYf_lpi, tauf_lpi, phif_lpi, Sfd_x_lpj, Sfd_y_lpj, Sfd_z_lpj, Bfu_x_lpj, Bfu_y_lpj, Bfu_z_lpj, Df_lpj, DYf_lpj, tauf_lpj, phif_lpj, Sfd_x_lpk, Sfd_y_lpk, Sfd_z_lpk, Bfu_x_lpk, Bfu_y_lpk, Bfu_z_lpk, Df_lpk, DYf_lpk, tauf_lpk, phif_lpk, Sfd_x_rpi, Sfd_y_rpi, Sfd_z_rpi, Bfu_x_rpi, Bfu_y_rpi, Bfu_z_rpi, Df_rpi, DYf_rpi, tauf_rpi, phif_rpi, Sfd_x_rpj, Sfd_y_rpj, Sfd_z_rpj, Bfu_x_rpj, Bfu_y_rpj, Bfu_z_rpj, Df_rpj, DYf_rpj, tauf_rpj, phif_rpj, Sfd_x_rpk, Sfd_y_rpk, Sfd_z_rpk, Bfu_x_rpk, Bfu_y_rpk, Bfu_z_rpk, Df_rpk, DYf_rpk, tauf_rpk, phif_rpk, FluxniSfd_x, FluxniSfd_y, FluxniSfd_z, FluxniBfu_x, FluxniBfu_y, FluxniBfu_z, FluxniDf, FluxniDYf, Fluxnitauf, Fluxniphif, FluxnjSfd_x, FluxnjSfd_y, FluxnjSfd_z, FluxnjBfu_x, FluxnjBfu_y, FluxnjBfu_z, FluxnjDf, FluxnjDYf, Fluxnjtauf, Fluxnjphif, FluxnkSfd_x, FluxnkSfd_y, FluxnkSfd_z, FluxnkBfu_x, FluxnkBfu_y, FluxnkBfu_z, FluxnkDf, FluxnkDYf, Fluxnktauf, Fluxnkphif, FluxpiSfd_x, FluxpiSfd_y, FluxpiSfd_z, FluxpiBfu_x, FluxpiBfu_y, FluxpiBfu_z, FluxpiDf, FluxpiDYf, Fluxpitauf, Fluxpiphif, FluxpjSfd_x, FluxpjSfd_y, FluxpjSfd_z, FluxpjBfu_x, FluxpjBfu_y, FluxpjBfu_z, FluxpjDf, FluxpjDYf, Fluxpjtauf, Fluxpjphif, FluxpkSfd_x, FluxpkSfd_y, FluxpkSfd_z, FluxpkBfu_x, FluxpkBfu_y, FluxpkBfu_z, FluxpkDf, FluxpkDYf, Fluxpktauf, Fluxpkphif, d_div_Beta_o0_t0_m0_l0, d_Betau_x_o1_t0_m0_l0, d_Betau_x_o2_t0_m0_l0, d_Sfd_y_o0_t7_m0_l0, d_Betau_x_o1_t1_m0_l0, d_Betau_x_o2_t1_m0_l0, d_Sfd_z_o0_t7_m0_l0, d_Betau_x_o1_t2_m0_l0, d_Betau_x_o2_t2_m0_l0, d_d_div_Beta_x_o0_t0_m0_l0, d_d_div_Beta_y_o0_t0_m0_l0, d_d_div_Beta_z_o0_t0_m0_l0, d_Gamh_x_o0_t15_m0_l0, d_Gamh_x_o0_t16_m0_l0, d_Gamh_x_o0_t17_m0_l0, d_Sfd_x_o0_t8_m0_l0, d_Betau_y_o1_t0_m0_l0, d_Betau_y_o2_t0_m0_l0, d_div_Beta_o0_t1_m0_l0, d_Betau_y_o1_t1_m0_l0, d_Betau_y_o2_t1_m0_l0, d_Sfd_z_o0_t8_m0_l0, d_Betau_y_o1_t2_m0_l0, d_Betau_y_o2_t2_m0_l0, d_Gamh_y_o0_t6_m0_l0, d_d_div_Beta_x_o0_t1_m0_l0, d_Gamh_y_o0_t8_m0_l0, d_d_div_Beta_y_o0_t1_m0_l0, d_d_div_Beta_z_o0_t1_m0_l0, d_Gamh_y_o0_t11_m0_l0, d_Sfd_x_o0_t9_m0_l0, d_Betau_z_o1_t0_m0_l0, d_Betau_z_o2_t0_m0_l0, d_Sfd_y_o0_t9_m0_l0, d_Betau_z_o1_t1_m0_l0, d_Betau_z_o2_t1_m0_l0, d_div_Beta_o0_t2_m0_l0, d_Betau_z_o1_t2_m0_l0, d_Betau_z_o2_t2_m0_l0, d_Gamh_z_o0_t0_m0_l0, d_Gamh_z_o0_t1_m0_l0, d_d_div_Beta_x_o0_t2_m0_l0, d_Gamh_z_o0_t3_m0_l0, d_d_div_Beta_y_o0_t2_m0_l0, d_d_div_Beta_z_o0_t2_m0_l0, d_Ctd_xxx_o0_t0_m0_l0, d_gtd_xx_o1_t0_m0_l0, d_gtd_xx_o2_t0_m0_l0, d_Ctd_xxy_o0_t0_m0_l0, d_gtd_xx_o1_t1_m0_l0, d_gtd_xx_o2_t1_m0_l0, d_Ctd_xxz_o0_t0_m0_l0, d_gtd_xx_o1_t2_m0_l0, d_gtd_xx_o2_t2_m0_l0, d_Ctd_yxx_o0_t0_m0_l0, d_gtd_xy_o1_t0_m0_l0, d_gtd_xy_o2_t0_m0_l0, d_Ctd_xyy_o0_t0_m0_l0, d_gtd_xy_o1_t1_m0_l0, d_gtd_xy_o2_t1_m0_l0, d_Ctd_xyz_o0_t1_m0_l0, d_gtd_xy_o1_t2_m0_l0, d_gtd_xy_o2_t2_m0_l0, d_Ctd_xyy_o0_t1_m0_l0, d_gtd_yy_o1_t0_m0_l0, d_gtd_yy_o2_t0_m0_l0, d_Ctd_yyy_o0_t0_m0_l0, d_gtd_yy_o1_t1_m0_l0, d_gtd_yy_o2_t1_m0_l0, d_Ctd_yyz_o0_t0_m0_l0, d_gtd_yy_o1_t2_m0_l0, d_gtd_yy_o2_t2_m0_l0, d_Ctd_zxx_o0_t0_m0_l0, d_gtd_xz_o1_t0_m0_l0, d_gtd_xz_o2_t0_m0_l0, d_Ctd_xyz_o0_t0_m0_l0, d_gtd_xz_o1_t1_m0_l0, d_gtd_xz_o2_t1_m0_l0, d_Ctd_xzz_o0_t0_m0_l0, d_gtd_xz_o1_t2_m0_l0, d_gtd_xz_o2_t2_m0_l0, d_Ctd_xyz_o0_t2_m0_l0, d_gtd_yz_o1_t0_m0_l0, d_gtd_yz_o2_t0_m0_l0, d_Ctd_zyy_o0_t0_m0_l0, d_gtd_yz_o1_t1_m0_l0, d_gtd_yz_o2_t1_m0_l0, d_Ctd_yzz_o0_t0_m0_l0, d_gtd_yz_o1_t2_m0_l0, d_gtd_yz_o2_t2_m0_l0, d_Ctd_xzz_o0_t1_m0_l0, d_gtd_zz_o1_t0_m0_l0, d_gtd_zz_o2_t0_m0_l0, d_Ctd_yzz_o0_t1_m0_l0, d_gtd_zz_o1_t1_m0_l0, d_gtd_zz_o2_t1_m0_l0, d_Ctd_zzz_o0_t0_m0_l0, d_gtd_zz_o1_t2_m0_l0, d_gtd_zz_o2_t2_m0_l0, d_Rpd_xx_o0_t0_m0_l0, d_chi_o1_t0_m0_l0, d_chi_o2_t0_m0_l0, d_Rpd_xx_o0_t1_m0_l0, d_chi_o1_t1_m0_l0, d_chi_o2_t1_m0_l0, d_Rpd_xx_o0_t2_m0_l0, d_chi_o1_t2_m0_l0, d_chi_o2_t2_m0_l0, d_Rtd_xx_o0_t6_m0_l0, d_Gamh_x_o1_t0_m0_l0, d_Gamh_x_o2_t0_m0_l0, d_Rtd_xy_o0_t6_m0_l0, d_Gamh_x_o1_t1_m0_l0, d_Gamh_x_o2_t1_m0_l0, d_Rtd_xz_o0_t6_m0_l0, d_Gamh_x_o1_t2_m0_l0, d_Gamh_x_o2_t2_m0_l0, d_Rtd_xx_o0_t7_m0_l0, d_Gamh_y_o1_t0_m0_l0, d_Gamh_y_o2_t0_m0_l0, d_Rtd_xy_o0_t7_m0_l0, d_Gamh_y_o1_t1_m0_l0, d_Gamh_y_o2_t1_m0_l0, d_Rtd_xz_o0_t7_m0_l0, d_Gamh_y_o1_t2_m0_l0, d_Gamh_y_o2_t2_m0_l0, d_Rtd_xx_o0_t8_m0_l0, d_Gamh_z_o1_t0_m0_l0, d_Gamh_z_o2_t0_m0_l0, d_Rtd_xy_o0_t8_m0_l0, d_Gamh_z_o1_t1_m0_l0, d_Gamh_z_o2_t1_m0_l0, d_Rtd_xz_o0_t8_m0_l0, d_Gamh_z_o1_t2_m0_l0, d_Gamh_z_o2_t2_m0_l0, d_Psi1_xx_o0_t1_m0_l0, d_Alpha_o1_t0_m0_l0, d_Alpha_o2_t0_m0_l0, d_Psi1_xx_o0_t2_m0_l0, d_Alpha_o1_t1_m0_l0, d_Alpha_o2_t1_m0_l0, d_Psi1_xx_o0_t3_m0_l0, d_Alpha_o1_t2_m0_l0, d_Alpha_o2_t2_m0_l0, d_Psi1_xx_o0_t0_m0_l0, d_Psi1_xy_o0_t0_m0_l0, d_Psi1_xz_o0_t0_m0_l0, d_Psi1_yy_o0_t0_m0_l0, d_Psi1_yz_o0_t0_m0_l0, d_Psi1_zz_o0_t0_m0_l0, d_Atd_xx_o1_t0_m0_l0, d_Atd_xx_o2_t0_m0_l0, d_Atd_xx_o1_t1_m0_l0, d_Atd_xx_o2_t1_m0_l0, d_Atd_xx_o1_t2_m0_l0, d_Atd_xx_o2_t2_m0_l0, d_Atd_xy_o1_t0_m0_l0, d_Atd_xy_o2_t0_m0_l0, d_Atd_xy_o1_t1_m0_l0, d_Atd_xy_o2_t1_m0_l0, d_Atd_xy_o1_t2_m0_l0, d_Atd_xy_o2_t2_m0_l0, d_Atd_xz_o1_t0_m0_l0, d_Atd_xz_o2_t0_m0_l0, d_Atd_xz_o1_t1_m0_l0, d_Atd_xz_o2_t1_m0_l0, d_Atd_xz_o1_t2_m0_l0, d_Atd_xz_o2_t2_m0_l0, d_Atd_yy_o1_t0_m0_l0, d_Atd_yy_o2_t0_m0_l0, d_Atd_yy_o1_t1_m0_l0, d_Atd_yy_o2_t1_m0_l0, d_Atd_yy_o1_t2_m0_l0, d_Atd_yy_o2_t2_m0_l0, d_Atd_yz_o1_t0_m0_l0, d_Atd_yz_o2_t0_m0_l0, d_Atd_yz_o1_t1_m0_l0, d_Atd_yz_o2_t1_m0_l0, d_Atd_yz_o1_t2_m0_l0, d_Atd_yz_o2_t2_m0_l0, d_Atd_zz_o1_t0_m0_l0, d_Atd_zz_o2_t0_m0_l0, d_Atd_zz_o1_t1_m0_l0, d_Atd_zz_o2_t1_m0_l0, d_Atd_zz_o1_t2_m0_l0, d_Atd_zz_o2_t2_m0_l0, d_Gamh_x_o0_t0_m0_l0, d_trK_o1_t0_m0_l0, d_trK_o2_t0_m0_l0, d_Gamh_x_o0_t1_m0_l0, d_trK_o1_t1_m0_l0, d_trK_o2_t1_m0_l0, d_Gamh_x_o0_t2_m0_l0, d_trK_o1_t2_m0_l0, d_trK_o2_t2_m0_l0, d_Gamh_x_o0_t3_m0_l0, d_theta_o1_t0_m0_l0, d_theta_o2_t0_m0_l0, d_Gamh_x_o0_t4_m0_l0, d_theta_o1_t1_m0_l0, d_theta_o2_t1_m0_l0, d_Gamh_x_o0_t5_m0_l0, d_theta_o1_t2_m0_l0, d_theta_o2_t2_m0_l0, d_Rtd_xx_o0_t0_m0_l0, d_Rtd_xx_o0_t1_m0_l0, d_Rtd_xx_o0_t2_m0_l0, d_Rtd_xx_o0_t3_m0_l0, d_Rtd_xx_o0_t4_m0_l0, d_Rtd_xx_o0_t5_m0_l0, d_Rtd_xy_o0_t0_m0_l0, d_Rtd_xy_o0_t1_m0_l0, d_Rtd_xy_o0_t2_m0_l0, d_Rtd_xy_o0_t3_m0_l0, d_Rtd_xy_o0_t4_m0_l0, d_Rtd_xy_o0_t5_m0_l0, d_Rtd_yy_o0_t0_m0_l0, d_Rtd_yy_o0_t1_m0_l0, d_Rtd_yy_o0_t2_m0_l0, d_Rtd_yy_o0_t3_m0_l0, d_Rtd_yy_o0_t4_m0_l0, d_Rtd_yy_o0_t5_m0_l0, d_Rtd_xz_o0_t0_m0_l0, d_Rtd_xz_o0_t1_m0_l0, d_Rtd_xz_o0_t2_m0_l0, d_Rtd_xz_o0_t3_m0_l0, d_Rtd_xz_o0_t4_m0_l0, d_Rtd_xz_o0_t5_m0_l0, d_Rtd_yz_o0_t0_m0_l0, d_Rtd_yz_o0_t1_m0_l0, d_Rtd_yz_o0_t2_m0_l0, d_Rtd_yz_o0_t3_m0_l0, d_Rtd_yz_o0_t4_m0_l0, d_Rtd_yz_o0_t5_m0_l0, d_Rtd_zz_o0_t0_m0_l0, d_Rtd_zz_o0_t1_m0_l0, d_Rtd_zz_o0_t2_m0_l0, d_Rtd_zz_o0_t3_m0_l0, d_Rtd_zz_o0_t4_m0_l0, d_Rtd_zz_o0_t5_m0_l0, d_Rpd_xx_o0_t3_m0_l0, d_Rpd_xx_o0_t4_m0_l0, d_Rpd_xx_o0_t5_m0_l0, d_Rpd_xx_o0_t6_m0_l0, d_Rpd_xx_o0_t7_m0_l0, d_Rpd_xx_o0_t8_m0_l0, m_Ctd_yyz_o0_t0_l0_p, Ctd_yyz_p, m_Ctd_yyy_o0_t0_l0_p, Ctd_yyy_p, div_Beta_p, m_Ctd_xxz_o0_t0_l0_p, Ctd_xxz_p, m_Ctd_xxx_o0_t0_l0_p, Ctd_xxx_p, m_Ctd_yzz_o0_t1_l0_p, Ctd_yzz_p, m_Ctd_xxy_o0_t0_l0_p, Ctd_xxy_p, kappa_z2_p, m_Ctd_xyy_o0_t1_l0_p, Ctd_xyy_p, m_Ctd_xyz_o0_t2_l0_p, m_Ctd_xyz_o0_t1_l0_p, m_Ctd_xyz_o0_t0_l0_p, Ctd_xyz_p, d_div_Beta_y_p, d_div_Beta_x_p, d_div_Beta_z_p, m_Ctd_xzz_o0_t1_l0_p, Ctd_xzz_p, m_Ctd_zxz_o0_t0_l0_p, Ctd_zxz_p, m_Ctd_zxx_o0_t1_l0_p, Ctd_zxx_p, m_Ctd_zyy_o0_t1_l0_p, Ctd_zyy_p, m_Ctd_zyz_o0_t0_l0_p, Ctd_zyz_p, Tf4u_tt_p, Betatd_x_p, Betatd_z_p, Betatd_y_p, Ct_zyz_p, Ct_zyy_p, Atud_zx_p, Atud_zz_p, Atud_zy_p, Atu_zz_p, Tf4u_tz_p, Tu_tz_p, Atud_yy_p, Atud_yx_p, Atud_yz_p, Atu_yz_p, Atu_yy_p, Ct_yyy_p, Ct_yyz_p, Atud_xx_p, trAt_p, Atud_xz_p, Atud_xy_p, Atu_xz_p, Atu_xy_p, Atu_xx_p, Tf4u_ty_p, Tu_ty_p, Tf4u_tx_p, Tu_tx_p, decay_factor_p, kappa_z1_p, kappa_f_p, feta_p, m_Ctd_zzz_o0_t0_l0_p, Ctd_zzz_p, Ct_xzz_p, Ct_zzz_p, Ct_yzz_p, m_Ctd_yxy_o0_t0_l0_p, Ctd_yxy_p, Ct_xyz_p, m_Ctd_yxz_o0_t2_l0_p, m_Ctd_yxz_o0_t0_l0_p, Ctd_yxz_p, Ct_zxz_p, Ct_yxz_p, m_Ctd_zxy_o0_t2_l0_p, Ctd_zxy_p, Ct_zxy_p, Ct_yxy_p, Ct_xxy_p, Ct_xxz_p, Ct_xyy_p, m_Ctd_yxx_o0_t1_l0_p, Ctd_yxx_p, Ct_zxx_p, Gamt_z_p, Zu_z_p, Ct_yxx_p, Gamt_y_p, Zu_y_p, Ct_xxx_p, Gamt_x_p, Zu_x_p, d_Rtd_xx_o0_t9_m0_l0_p, m_Rtd_xx_o0_t8_l0_p, m_Rtd_xx_o0_t7_l0_p, m_Rtd_xx_o0_t6_l0_p, m_Rtd_xx_o0_t5_l0_p, m_Rtd_xx_o0_t4_l0_p, m_Rtd_xx_o0_t3_l0_p, m_Rtd_xx_o0_t2_l0_p, m_Rtd_xx_o0_t1_l0_p, m_Rtd_xx_o0_t0_l0_p, Rtd_xx_p, d_Rtd_xz_o0_t12_m0_l0_p, m_Rtd_xz_o0_t11_l0_p, m_Rtd_xz_o0_t10_l0_p, m_Rtd_xz_o0_t9_l0_p, m_Rtd_xz_o0_t8_l0_p, m_Rtd_xz_o0_t7_l0_p, m_Rtd_xz_o0_t6_l0_p, m_Rtd_xz_o0_t5_l0_p, m_Rtd_xz_o0_t4_l0_p, m_Rtd_xz_o0_t3_l0_p, m_Rtd_xz_o0_t2_l0_p, m_Rtd_xz_o0_t1_l0_p, m_Rtd_xz_o0_t0_l0_p, Rtd_xz_p, d_Rtd_zz_o0_t9_m0_l0_p, m_Rtd_zz_o0_t8_l0_p, m_Rtd_zz_o0_t7_l0_p, m_Rtd_zz_o0_t6_l0_p, m_Rtd_zz_o0_t5_l0_p, m_Rtd_zz_o0_t4_l0_p, m_Rtd_zz_o0_t3_l0_p, m_Rtd_zz_o0_t2_l0_p, m_Rtd_zz_o0_t1_l0_p, m_Rtd_zz_o0_t0_l0_p, Rtd_zz_p, d_Rtd_xy_o0_t12_m0_l0_p, m_Rtd_xy_o0_t11_l0_p, m_Rtd_xy_o0_t10_l0_p, m_Rtd_xy_o0_t9_l0_p, m_Rtd_xy_o0_t8_l0_p, m_Rtd_xy_o0_t7_l0_p, m_Rtd_xy_o0_t6_l0_p, m_Rtd_xy_o0_t5_l0_p, m_Rtd_xy_o0_t4_l0_p, m_Rtd_xy_o0_t3_l0_p, m_Rtd_xy_o0_t2_l0_p, m_Rtd_xy_o0_t1_l0_p, m_Rtd_xy_o0_t0_l0_p, Rtd_xy_p, d_Rtd_yy_o0_t9_m0_l0_p, m_Rtd_yy_o0_t8_l0_p, m_Rtd_yy_o0_t7_l0_p, m_Rtd_yy_o0_t6_l0_p, m_Rtd_yy_o0_t5_l0_p, m_Rtd_yy_o0_t4_l0_p, m_Rtd_yy_o0_t3_l0_p, m_Rtd_yy_o0_t2_l0_p, m_Rtd_yy_o0_t1_l0_p, m_Rtd_yy_o0_t0_l0_p, Rtd_yy_p, d_Rtd_yz_o0_t12_m0_l0_p, m_Rtd_yz_o0_t11_l0_p, m_Rtd_yz_o0_t10_l0_p, m_Rtd_yz_o0_t9_l0_p, m_Rtd_yz_o0_t8_l0_p, m_Rtd_yz_o0_t7_l0_p, m_Rtd_yz_o0_t6_l0_p, m_Rtd_yz_o0_t5_l0_p, m_Rtd_yz_o0_t4_l0_p, m_Rtd_yz_o0_t3_l0_p, m_Rtd_yz_o0_t2_l0_p, m_Rtd_yz_o0_t1_l0_p, m_Rtd_yz_o0_t0_l0_p, Rtd_yz_p, m_Rpd_xx_o0_t22_l0_p, m_Rpd_xx_o0_t21_l0_p, m_Rpd_xx_o0_t20_l0_p, m_Rpd_xx_o0_t19_l0_p, m_Rpd_xx_o0_t18_l0_p, m_Rpd_xx_o0_t17_l0_p, m_Rpd_xx_o0_t16_l0_p, m_Rpd_xx_o0_t15_l0_p, m_Rpd_xx_o0_t14_l0_p, m_Rpd_xx_o0_t13_l0_p, m_Rpd_xx_o0_t12_l0_p, m_Rpd_xx_o0_t11_l0_p, m_Rpd_xx_o0_t10_l0_p, m_Rpd_xx_o0_t9_l0_p, m_Rpd_xx_o0_t8_l0_p, m_Rpd_xx_o0_t7_l0_p, m_Rpd_xx_o0_t6_l0_p, m_Rpd_xx_o0_t5_l0_p, m_Rpd_xx_o0_t4_l0_p, m_Rpd_xx_o0_t3_l0_p, m_Rpd_xx_o0_t2_l0_p, m_Rpd_xx_o0_t1_l0_p, m_Rpd_xx_o0_t0_l0_p, Rpd_xx_p, m_Rpd_yy_o0_t22_l0_p, m_Rpd_yy_o0_t21_l0_p, m_Rpd_yy_o0_t20_l0_p, m_Rpd_yy_o0_t19_l0_p, m_Rpd_yy_o0_t18_l0_p, m_Rpd_yy_o0_t17_l0_p, m_Rpd_yy_o0_t16_l0_p, m_Rpd_yy_o0_t15_l0_p, m_Rpd_yy_o0_t14_l0_p, m_Rpd_yy_o0_t13_l0_p, m_Rpd_yy_o0_t12_l0_p, m_Rpd_yy_o0_t11_l0_p, m_Rpd_yy_o0_t10_l0_p, m_Rpd_yy_o0_t9_l0_p, m_Rpd_yy_o0_t8_l0_p, m_Rpd_yy_o0_t7_l0_p, m_Rpd_yy_o0_t6_l0_p, m_Rpd_yy_o0_t5_l0_p, m_Rpd_yy_o0_t4_l0_p, m_Rpd_yy_o0_t3_l0_p, m_Rpd_yy_o0_t2_l0_p, m_Rpd_yy_o0_t1_l0_p, m_Rpd_yy_o0_t0_l0_p, Rpd_yy_p, m_Rpd_yz_o0_t25_l0_p, m_Rpd_yz_o0_t24_l0_p, m_Rpd_yz_o0_t23_l0_p, m_Rpd_yz_o0_t22_l0_p, m_Rpd_yz_o0_t21_l0_p, m_Rpd_yz_o0_t20_l0_p, m_Rpd_yz_o0_t19_l0_p, m_Rpd_yz_o0_t18_l0_p, m_Rpd_yz_o0_t17_l0_p, m_Rpd_yz_o0_t16_l0_p, m_Rpd_yz_o0_t15_l0_p, m_Rpd_yz_o0_t14_l0_p, m_Rpd_yz_o0_t13_l0_p, m_Rpd_yz_o0_t12_l0_p, m_Rpd_yz_o0_t11_l0_p, m_Rpd_yz_o0_t10_l0_p, m_Rpd_yz_o0_t9_l0_p, m_Rpd_yz_o0_t8_l0_p, m_Rpd_yz_o0_t7_l0_p, m_Rpd_yz_o0_t6_l0_p, m_Rpd_yz_o0_t5_l0_p, m_Rpd_yz_o0_t4_l0_p, m_Rpd_yz_o0_t3_l0_p, m_Rpd_yz_o0_t2_l0_p, m_Rpd_yz_o0_t1_l0_p, m_Rpd_yz_o0_t0_l0_p, Rpd_yz_p, m_Rpd_xz_o0_t25_l0_p, m_Rpd_xz_o0_t24_l0_p, m_Rpd_xz_o0_t23_l0_p, m_Rpd_xz_o0_t22_l0_p, m_Rpd_xz_o0_t21_l0_p, m_Rpd_xz_o0_t20_l0_p, m_Rpd_xz_o0_t19_l0_p, m_Rpd_xz_o0_t18_l0_p, m_Rpd_xz_o0_t17_l0_p, m_Rpd_xz_o0_t16_l0_p, m_Rpd_xz_o0_t15_l0_p, m_Rpd_xz_o0_t14_l0_p, m_Rpd_xz_o0_t13_l0_p, m_Rpd_xz_o0_t12_l0_p, m_Rpd_xz_o0_t11_l0_p, m_Rpd_xz_o0_t10_l0_p, m_Rpd_xz_o0_t9_l0_p, m_Rpd_xz_o0_t8_l0_p, m_Rpd_xz_o0_t7_l0_p, m_Rpd_xz_o0_t6_l0_p, m_Rpd_xz_o0_t5_l0_p, m_Rpd_xz_o0_t4_l0_p, m_Rpd_xz_o0_t3_l0_p, m_Rpd_xz_o0_t2_l0_p, m_Rpd_xz_o0_t1_l0_p, m_Rpd_xz_o0_t0_l0_p, Rpd_xz_p, m_Rpd_xy_o0_t25_l0_p, m_Rpd_xy_o0_t24_l0_p, m_Rpd_xy_o0_t23_l0_p, m_Rpd_xy_o0_t22_l0_p, m_Rpd_xy_o0_t21_l0_p, m_Rpd_xy_o0_t20_l0_p, m_Rpd_xy_o0_t19_l0_p, m_Rpd_xy_o0_t18_l0_p, m_Rpd_xy_o0_t17_l0_p, m_Rpd_xy_o0_t16_l0_p, m_Rpd_xy_o0_t15_l0_p, m_Rpd_xy_o0_t14_l0_p, m_Rpd_xy_o0_t13_l0_p, m_Rpd_xy_o0_t12_l0_p, m_Rpd_xy_o0_t11_l0_p, m_Rpd_xy_o0_t10_l0_p, m_Rpd_xy_o0_t9_l0_p, m_Rpd_xy_o0_t8_l0_p, m_Rpd_xy_o0_t7_l0_p, m_Rpd_xy_o0_t6_l0_p, m_Rpd_xy_o0_t5_l0_p, m_Rpd_xy_o0_t4_l0_p, m_Rpd_xy_o0_t3_l0_p, m_Rpd_xy_o0_t2_l0_p, m_Rpd_xy_o0_t1_l0_p, m_Rpd_xy_o0_t0_l0_p, Rpd_xy_p, m_Rpd_zz_o0_t22_l0_p, m_Rpd_zz_o0_t21_l0_p, m_Rpd_zz_o0_t20_l0_p, m_Rpd_zz_o0_t19_l0_p, m_Rpd_zz_o0_t18_l0_p, m_Rpd_zz_o0_t17_l0_p, m_Rpd_zz_o0_t16_l0_p, m_Rpd_zz_o0_t15_l0_p, m_Rpd_zz_o0_t14_l0_p, m_Rpd_zz_o0_t13_l0_p, m_Rpd_zz_o0_t12_l0_p, m_Rpd_zz_o0_t11_l0_p, m_Rpd_zz_o0_t10_l0_p, m_Rpd_zz_o0_t9_l0_p, m_Rpd_zz_o0_t8_l0_p, m_Rpd_zz_o0_t7_l0_p, m_Rpd_zz_o0_t6_l0_p, m_Rpd_zz_o0_t5_l0_p, m_Rpd_zz_o0_t4_l0_p, m_Rpd_zz_o0_t3_l0_p, m_Rpd_zz_o0_t2_l0_p, m_Rpd_zz_o0_t1_l0_p, m_Rpd_zz_o0_t0_l0_p, Rpd_zz_p, Rscalar_p, Tf4u_zz_p, Tu_zz_p, Tf4u_yz_p, Tu_yz_p, Tf4u_yy_p, Tu_yy_p, Tf4u_xy_p, Tu_xy_p, Tf4u_xx_p, Tu_xx_p, Tf4u_xz_p, Tu_xz_p, trSf_p, Tu_tt_p, pTtd_ADM_xz_p, pTtd_ADM_xy_p, pTtd_ADM_xx_p, d_Psi1_xx_o0_t5_m0_l0_p, m_Psi1_xx_o0_t4_l0_p, m_Psi1_xx_o0_t3_l0_p, m_Psi1_xx_o0_t2_l0_p, m_Psi1_xx_o0_t1_l0_p, m_Psi1_xx_o0_t0_l0_p, Psi1_xx_p, d_Psi1_xy_o0_t6_m0_l0_p, m_Psi1_xy_o0_t5_l0_p, m_Psi1_xy_o0_t4_l0_p, m_Psi1_xy_o0_t3_l0_p, m_Psi1_xy_o0_t2_l0_p, m_Psi1_xy_o0_t1_l0_p, m_Psi1_xy_o0_t0_l0_p, Psi1_xy_p, d_Psi1_xz_o0_t6_m0_l0_p, m_Psi1_xz_o0_t5_l0_p, m_Psi1_xz_o0_t4_l0_p, m_Psi1_xz_o0_t3_l0_p, m_Psi1_xz_o0_t2_l0_p, m_Psi1_xz_o0_t1_l0_p, m_Psi1_xz_o0_t0_l0_p, Psi1_xz_p, Jtd_ADM_z_p, pTtd_ADM_zz_p, d_Psi1_zz_o0_t5_m0_l0_p, m_Psi1_zz_o0_t4_l0_p, m_Psi1_zz_o0_t3_l0_p, m_Psi1_zz_o0_t2_l0_p, m_Psi1_zz_o0_t1_l0_p, m_Psi1_zz_o0_t0_l0_p, Psi1_zz_p, Jtd_ADM_y_p, Jtd_ADM_x_p, pTtd_ADM_yz_p, d_Psi1_yz_o0_t6_m0_l0_p, m_Psi1_yz_o0_t5_l0_p, m_Psi1_yz_o0_t4_l0_p, m_Psi1_yz_o0_t3_l0_p, m_Psi1_yz_o0_t2_l0_p, m_Psi1_yz_o0_t1_l0_p, m_Psi1_yz_o0_t0_l0_p, Psi1_yz_p, pTtd_ADM_yy_p, tr_pT_p, d_Psi1_yy_o0_t5_m0_l0_p, m_Psi1_yy_o0_t4_l0_p, m_Psi1_yy_o0_t3_l0_p, m_Psi1_yy_o0_t2_l0_p, m_Psi1_yy_o0_t1_l0_p, m_Psi1_yy_o0_t0_l0_p, Psi1_yy_p, trPsi1_p, Psi1TF_yz_p, Psi1TF_yy_p, Psi1TF_xy_p, Psi1TF_xx_p, Psi1TF_xz_p, Psi1TF_zz_p, kappa_cc_p, rho_ADM_p, d_Sfd_x_o0_t11_m0_l0, d_Sfd_y_o0_t11_m0_l0, d_Sfd_z_o0_t11_m0_l0, d_DYf_o0_t0_m0_l0, d_tauf_o0_t3_m0_l0, d_gtd_xx_o0_t3_m0_l0, d_gtd_xy_o0_t6_m0_l0, d_gtd_xz_o0_t6_m0_l0, d_gtd_yy_o0_t3_m0_l0, d_gtd_yz_o0_t6_m0_l0, d_gtd_zz_o0_t3_m0_l0, d_Atd_xx_o0_t3_m0_l0, d_Atd_xy_o0_t6_m0_l0, d_Atd_xz_o0_t6_m0_l0, d_Atd_yy_o0_t3_m0_l0, d_Atd_yz_o0_t6_m0_l0, d_Atd_zz_o0_t3_m0_l0, d_Gamh_x_o0_t24_m0_l0, d_Gamh_y_o0_t24_m0_l0, d_Gamh_z_o0_t24_m0_l0, d_Betau_x_o0_t0_m0_l0, d_Betau_y_o0_t0_m0_l0, d_Betau_z_o0_t0_m0_l0, d_Alpha_o0_t0_m0_l0, d_chi_o0_t0_m0_l0, d_trK_o0_t21_m0_l0, d_theta_o0_t3_m0_l0, RHS_Bfu_x, RHS_Bfu_y, RHS_Bfu_z, RHS_tauf, RHS_phif, RHS_Sfd_x, RHS_Sfd_y, RHS_Sfd_z, RHS_Df, RHS_DYf, m_theta_o2_t2_l0, m_theta_o2_t1_l0, m_theta_o2_t0_l0, m_theta_o1_t2_l0, m_theta_o1_t1_l0, m_theta_o1_t0_l0, m_theta_o0_t2_l0, m_theta_o0_t1_l0, m_theta_o0_t0_l0, m_trK_o2_t2_l0, m_trK_o2_t1_l0, m_trK_o2_t0_l0, m_trK_o1_t2_l0, m_trK_o1_t1_l0, m_trK_o1_t0_l0, m_trK_o0_t20_l0, m_trK_o0_t19_l0, m_trK_o0_t18_l0, m_trK_o0_t17_l0, m_trK_o0_t16_l0, m_trK_o0_t15_l0, m_trK_o0_t14_l0, m_trK_o0_t13_l0, m_trK_o0_t12_l0, m_trK_o0_t11_l0, m_trK_o0_t10_l0, m_trK_o0_t9_l0, m_trK_o0_t8_l0, m_trK_o0_t7_l0, m_trK_o0_t6_l0, m_trK_o0_t5_l0, m_trK_o0_t4_l0, m_trK_o0_t3_l0, m_trK_o0_t2_l0, m_trK_o0_t1_l0, m_trK_o0_t0_l0, m_chi_o2_t2_l0, m_chi_o2_t1_l0, m_chi_o2_t0_l0, m_chi_o1_t2_l0, m_chi_o1_t1_l0, m_chi_o1_t0_l0, m_Alpha_o2_t2_l0, m_Alpha_o2_t1_l0, m_Alpha_o2_t0_l0, m_Alpha_o1_t2_l0, m_Alpha_o1_t1_l0, m_Alpha_o1_t0_l0, m_Betau_z_o2_t2_l0, m_Betau_z_o2_t1_l0, m_Betau_z_o2_t0_l0, m_Betau_z_o1_t2_l0, m_Betau_z_o1_t1_l0, m_Betau_z_o1_t0_l0, m_Betau_y_o2_t2_l0, m_Betau_y_o2_t1_l0, m_Betau_y_o2_t0_l0, m_Betau_y_o1_t2_l0, m_Betau_y_o1_t1_l0, m_Betau_y_o1_t0_l0, m_Betau_x_o2_t2_l0, m_Betau_x_o2_t1_l0, m_Betau_x_o2_t0_l0, m_Betau_x_o1_t2_l0, m_Betau_x_o1_t1_l0, m_Betau_x_o1_t0_l0, m_Gamh_z_o2_t2_l0, m_Gamh_z_o2_t1_l0, m_Gamh_z_o2_t0_l0, m_Gamh_z_o1_t2_l0, m_Gamh_z_o1_t1_l0, m_Gamh_z_o1_t0_l0, m_Gamh_z_o0_t23_l0, m_Gamh_z_o0_t22_l0, m_Gamh_z_o0_t21_l0, m_Gamh_z_o0_t20_l0, m_Gamh_z_o0_t19_l0, m_Gamh_z_o0_t18_l0, m_Gamh_z_o0_t17_l0, m_Gamh_z_o0_t16_l0, m_Gamh_z_o0_t15_l0, m_Gamh_z_o0_t14_l0, m_Gamh_z_o0_t13_l0, m_Gamh_z_o0_t12_l0, m_Gamh_z_o0_t11_l0, m_Gamh_z_o0_t10_l0, m_Gamh_z_o0_t9_l0, m_Gamh_z_o0_t8_l0, m_Gamh_z_o0_t7_l0, m_Gamh_z_o0_t6_l0, m_Gamh_z_o0_t5_l0, m_Gamh_z_o0_t4_l0, m_Gamh_z_o0_t3_l0, m_Gamh_z_o0_t2_l0, m_Gamh_z_o0_t1_l0, m_Gamh_z_o0_t0_l0, m_Gamh_y_o2_t2_l0, m_Gamh_y_o2_t1_l0, m_Gamh_y_o2_t0_l0, m_Gamh_y_o1_t2_l0, m_Gamh_y_o1_t1_l0, m_Gamh_y_o1_t0_l0, m_Gamh_y_o0_t23_l0, m_Gamh_y_o0_t22_l0, m_Gamh_y_o0_t21_l0, m_Gamh_y_o0_t20_l0, m_Gamh_y_o0_t19_l0, m_Gamh_y_o0_t18_l0, m_Gamh_y_o0_t17_l0, m_Gamh_y_o0_t16_l0, m_Gamh_y_o0_t15_l0, m_Gamh_y_o0_t14_l0, m_Gamh_y_o0_t13_l0, m_Gamh_y_o0_t12_l0, m_Gamh_y_o0_t11_l0, m_Gamh_y_o0_t10_l0, m_Gamh_y_o0_t9_l0, m_Gamh_y_o0_t8_l0, m_Gamh_y_o0_t7_l0, m_Gamh_y_o0_t6_l0, m_Gamh_y_o0_t5_l0, m_Gamh_y_o0_t4_l0, m_Gamh_y_o0_t3_l0, m_Gamh_y_o0_t2_l0, m_Gamh_y_o0_t1_l0, m_Gamh_y_o0_t0_l0, m_Gamh_x_o2_t2_l0, m_Gamh_x_o2_t1_l0, m_Gamh_x_o2_t0_l0, m_Gamh_x_o1_t2_l0, m_Gamh_x_o1_t1_l0, m_Gamh_x_o1_t0_l0, m_Gamh_x_o0_t23_l0, m_Gamh_x_o0_t22_l0, m_Gamh_x_o0_t21_l0, m_Gamh_x_o0_t20_l0, m_Gamh_x_o0_t19_l0, m_Gamh_x_o0_t18_l0, m_Gamh_x_o0_t17_l0, m_Gamh_x_o0_t16_l0, m_Gamh_x_o0_t15_l0, m_Gamh_x_o0_t14_l0, m_Gamh_x_o0_t13_l0, m_Gamh_x_o0_t12_l0, m_Gamh_x_o0_t11_l0, m_Gamh_x_o0_t10_l0, m_Gamh_x_o0_t9_l0, m_Gamh_x_o0_t8_l0, m_Gamh_x_o0_t7_l0, m_Gamh_x_o0_t6_l0, m_Gamh_x_o0_t5_l0, m_Gamh_x_o0_t4_l0, m_Gamh_x_o0_t3_l0, m_Gamh_x_o0_t2_l0, m_Gamh_x_o0_t1_l0, m_Gamh_x_o0_t0_l0, m_Atd_zz_o2_t2_l0, m_Atd_zz_o2_t1_l0, m_Atd_zz_o2_t0_l0, m_Atd_zz_o1_t2_l0, m_Atd_zz_o1_t1_l0, m_Atd_zz_o1_t0_l0, m_Atd_zz_o0_t2_l0, m_Atd_zz_o0_t1_l0, m_Atd_zz_o0_t0_l0, m_Atd_yz_o2_t2_l0, m_Atd_yz_o2_t1_l0, m_Atd_yz_o2_t0_l0, m_Atd_yz_o1_t2_l0, m_Atd_yz_o1_t1_l0, m_Atd_yz_o1_t0_l0, m_Atd_yz_o0_t5_l0, m_Atd_yz_o0_t4_l0, m_Atd_yz_o0_t3_l0, m_Atd_yz_o0_t2_l0, m_Atd_yz_o0_t1_l0, m_Atd_yz_o0_t0_l0, m_Atd_yy_o2_t2_l0, m_Atd_yy_o2_t1_l0, m_Atd_yy_o2_t0_l0, m_Atd_yy_o1_t2_l0, m_Atd_yy_o1_t1_l0, m_Atd_yy_o1_t0_l0, m_Atd_yy_o0_t2_l0, m_Atd_yy_o0_t1_l0, m_Atd_yy_o0_t0_l0, m_Atd_xz_o2_t2_l0, m_Atd_xz_o2_t1_l0, m_Atd_xz_o2_t0_l0, m_Atd_xz_o1_t2_l0, m_Atd_xz_o1_t1_l0, m_Atd_xz_o1_t0_l0, m_Atd_xz_o0_t5_l0, m_Atd_xz_o0_t4_l0, m_Atd_xz_o0_t3_l0, m_Atd_xz_o0_t2_l0, m_Atd_xz_o0_t1_l0, m_Atd_xz_o0_t0_l0, m_Atd_xy_o2_t2_l0, m_Atd_xy_o2_t1_l0, m_Atd_xy_o2_t0_l0, m_Atd_xy_o1_t2_l0, m_Atd_xy_o1_t1_l0, m_Atd_xy_o1_t0_l0, m_Atd_xy_o0_t5_l0, m_Atd_xy_o0_t4_l0, m_Atd_xy_o0_t3_l0, m_Atd_xy_o0_t2_l0, m_Atd_xy_o0_t1_l0, m_Atd_xy_o0_t0_l0, m_Atd_xx_o2_t2_l0, m_Atd_xx_o2_t1_l0, m_Atd_xx_o2_t0_l0, m_Atd_xx_o1_t2_l0, m_Atd_xx_o1_t1_l0, m_Atd_xx_o1_t0_l0, m_Atd_xx_o0_t2_l0, m_Atd_xx_o0_t1_l0, m_Atd_xx_o0_t0_l0, m_gtd_zz_o2_t2_l0, m_gtd_zz_o2_t1_l0, m_gtd_zz_o2_t0_l0, m_gtd_zz_o1_t2_l0, m_gtd_zz_o1_t1_l0, m_gtd_zz_o1_t0_l0, m_gtd_zz_o0_t2_l0, m_gtd_zz_o0_t1_l0, m_gtd_zz_o0_t0_l0, m_gtd_yz_o2_t2_l0, m_gtd_yz_o2_t1_l0, m_gtd_yz_o2_t0_l0, m_gtd_yz_o1_t2_l0, m_gtd_yz_o1_t1_l0, m_gtd_yz_o1_t0_l0, m_gtd_yz_o0_t5_l0, m_gtd_yz_o0_t4_l0, m_gtd_yz_o0_t3_l0, m_gtd_yz_o0_t2_l0, m_gtd_yz_o0_t1_l0, m_gtd_yz_o0_t0_l0, m_gtd_yy_o2_t2_l0, m_gtd_yy_o2_t1_l0, m_gtd_yy_o2_t0_l0, m_gtd_yy_o1_t2_l0, m_gtd_yy_o1_t1_l0, m_gtd_yy_o1_t0_l0, m_gtd_yy_o0_t2_l0, m_gtd_yy_o0_t1_l0, m_gtd_yy_o0_t0_l0, m_gtd_xz_o2_t2_l0, m_gtd_xz_o2_t1_l0, m_gtd_xz_o2_t0_l0, m_gtd_xz_o1_t2_l0, m_gtd_xz_o1_t1_l0, m_gtd_xz_o1_t0_l0, m_gtd_xz_o0_t5_l0, m_gtd_xz_o0_t4_l0, m_gtd_xz_o0_t3_l0, m_gtd_xz_o0_t2_l0, m_gtd_xz_o0_t1_l0, m_gtd_xz_o0_t0_l0, m_gtd_xy_o2_t2_l0, m_gtd_xy_o2_t1_l0, m_gtd_xy_o2_t0_l0, m_gtd_xy_o1_t2_l0, m_gtd_xy_o1_t1_l0, m_gtd_xy_o1_t0_l0, m_gtd_xy_o0_t5_l0, m_gtd_xy_o0_t4_l0, m_gtd_xy_o0_t3_l0, m_gtd_xy_o0_t2_l0, m_gtd_xy_o0_t1_l0, m_gtd_xy_o0_t0_l0, m_gtd_xx_o2_t2_l0, m_gtd_xx_o2_t1_l0, m_gtd_xx_o2_t0_l0, m_gtd_xx_o1_t2_l0, m_gtd_xx_o1_t1_l0, m_gtd_xx_o1_t0_l0, m_gtd_xx_o0_t2_l0, m_gtd_xx_o0_t1_l0, m_gtd_xx_o0_t0_l0, m_phif_o0_t2_l0, m_phif_o0_t1_l0, m_phif_o0_t0_l0, m_tauf_o0_t2_l0, m_tauf_o0_t1_l0, m_tauf_o0_t0_l0, m_Bfu_z_o0_t5_l0, m_Bfu_z_o0_t4_l0, m_Bfu_z_o0_t3_l0, m_Bfu_z_o0_t2_l0, m_Bfu_z_o0_t1_l0, m_Bfu_z_o0_t0_l0, m_Bfu_y_o0_t5_l0, m_Bfu_y_o0_t4_l0, m_Bfu_y_o0_t3_l0, m_Bfu_y_o0_t2_l0, m_Bfu_y_o0_t1_l0, m_Bfu_y_o0_t0_l0, m_Bfu_x_o0_t5_l0, m_Bfu_x_o0_t4_l0, m_Bfu_x_o0_t3_l0, m_Bfu_x_o0_t2_l0, m_Bfu_x_o0_t1_l0, m_Bfu_x_o0_t0_l0, m_Sfd_z_o0_t10_l0, m_Sfd_z_o0_t9_l0, m_Sfd_z_o0_t8_l0, m_Sfd_z_o0_t7_l0, m_Sfd_z_o0_t6_l0, m_Sfd_z_o0_t5_l0, m_Sfd_z_o0_t4_l0, m_Sfd_z_o0_t3_l0, m_Sfd_z_o0_t2_l0, m_Sfd_z_o0_t1_l0, m_Sfd_z_o0_t0_l0, m_Sfd_y_o0_t10_l0, m_Sfd_y_o0_t9_l0, m_Sfd_y_o0_t8_l0, m_Sfd_y_o0_t7_l0, m_Sfd_y_o0_t6_l0, m_Sfd_y_o0_t5_l0, m_Sfd_y_o0_t4_l0, m_Sfd_y_o0_t3_l0, m_Sfd_y_o0_t2_l0, m_Sfd_y_o0_t1_l0, m_Sfd_y_o0_t0_l0, m_Sfd_x_o0_t10_l0, m_Sfd_x_o0_t9_l0, m_Sfd_x_o0_t8_l0, m_Sfd_x_o0_t7_l0, m_Sfd_x_o0_t6_l0, m_Sfd_x_o0_t5_l0, m_Sfd_x_o0_t4_l0, m_Sfd_x_o0_t3_l0, m_Sfd_x_o0_t2_l0, m_Sfd_x_o0_t1_l0, m_Sfd_x_o0_t0_l0, RHS_gtd_xx, RHS_gtd_xy, RHS_gtd_xz, RHS_gtd_yy, RHS_gtd_yz, RHS_gtd_zz, RHS_Atd_xx, RHS_Atd_xy, RHS_Atd_xz, RHS_Atd_yy, RHS_Atd_yz, RHS_Atd_zz, RHS_Gamh_x, RHS_Gamh_y, RHS_Gamh_z, RHS_Betau_x, RHS_Betau_y, RHS_Betau_z, RHS_Alpha, RHS_chi, RHS_trK, RHS_theta, n_x, n_y, n_z, interaction_index, mod_normal, i_d_Atd_xx_o0_t0_m0_l0, i_d_Atd_xx_o0_t1_m0_l0, i_d_Atd_xx_o0_t2_m0_l0, i_d_Atd_xy_o0_t0_m0_l0, i_d_Atd_xy_o0_t1_m0_l0, i_d_Atd_xy_o0_t2_m0_l0, i_d_Atd_xz_o0_t0_m0_l0, i_d_Atd_xz_o0_t1_m0_l0, i_d_Atd_xz_o0_t2_m0_l0, i_d_Atd_yy_o0_t0_m0_l0, i_d_Atd_yy_o0_t1_m0_l0, i_d_Atd_yy_o0_t2_m0_l0, i_d_Atd_yz_o0_t0_m0_l0, i_d_Atd_yz_o0_t1_m0_l0, i_d_Atd_yz_o0_t2_m0_l0, i_d_Atd_zz_o0_t0_m0_l0, i_d_Atd_zz_o0_t1_m0_l0, i_d_Atd_zz_o0_t2_m0_l0, i_d_gtd_xx_o0_t3_m0_l0, i_d_gtd_xy_o0_t3_m0_l0, i_d_gtd_xz_o0_t3_m0_l0, i_d_gtd_yy_o0_t3_m0_l0, i_d_gtd_yz_o0_t3_m0_l0, i_d_gtd_zz_o0_t3_m0_l0, i_d_Atd_xx_o0_t3_m0_l0, i_d_Atd_xy_o0_t3_m0_l0, i_d_Atd_xz_o0_t3_m0_l0, i_d_Atd_yy_o0_t3_m0_l0, i_d_Atd_yz_o0_t3_m0_l0, i_d_Atd_zz_o0_t3_m0_l0, i_d_Gamh_x_o0_t3_m0_l0, i_d_Gamh_y_o0_t3_m0_l0, i_d_Gamh_z_o0_t3_m0_l0, i_d_Betau_x_o0_t3_m0_l0, i_d_Betau_y_o0_t3_m0_l0, i_d_Betau_z_o0_t3_m0_l0, i_d_Alpha_o0_t3_m0_l0, i_d_chi_o0_t3_m0_l0, i_d_trK_o0_t3_m0_l0, i_d_theta_o0_t3_m0_l0, i_m_gtd_xx_o0_t0_l0, i_m_gtd_xx_o0_t1_l0, i_m_gtd_xx_o0_t2_l0, i_m_gtd_xy_o0_t0_l0, i_m_gtd_xy_o0_t1_l0, i_m_gtd_xy_o0_t2_l0, i_m_gtd_xz_o0_t0_l0, i_m_gtd_xz_o0_t1_l0, i_m_gtd_xz_o0_t2_l0, i_m_gtd_yy_o0_t0_l0, i_m_gtd_yy_o0_t1_l0, i_m_gtd_yy_o0_t2_l0, i_m_gtd_yz_o0_t0_l0, i_m_gtd_yz_o0_t1_l0, i_m_gtd_yz_o0_t2_l0, i_m_gtd_zz_o0_t0_l0, i_m_gtd_zz_o0_t1_l0, i_m_gtd_zz_o0_t2_l0, i_m_Atd_xx_o0_t0_l0, i_m_Atd_xx_o0_t1_l0, i_m_Atd_xx_o0_t2_l0, i_m_Atd_xy_o0_t0_l0, i_m_Atd_xy_o0_t1_l0, i_m_Atd_xy_o0_t2_l0, i_m_Atd_xz_o0_t0_l0, i_m_Atd_xz_o0_t1_l0, i_m_Atd_xz_o0_t2_l0, i_m_Atd_yy_o0_t0_l0, i_m_Atd_yy_o0_t1_l0, i_m_Atd_yy_o0_t2_l0, i_m_Atd_yz_o0_t0_l0, i_m_Atd_yz_o0_t1_l0, i_m_Atd_yz_o0_t2_l0, i_m_Atd_zz_o0_t0_l0, i_m_Atd_zz_o0_t1_l0, i_m_Atd_zz_o0_t2_l0, i_m_Gamh_x_o0_t0_l0, i_m_Gamh_x_o0_t1_l0, i_m_Gamh_x_o0_t2_l0, i_m_Gamh_y_o0_t0_l0, i_m_Gamh_y_o0_t1_l0, i_m_Gamh_y_o0_t2_l0, i_m_Gamh_z_o0_t0_l0, i_m_Gamh_z_o0_t1_l0, i_m_Gamh_z_o0_t2_l0, i_m_Betau_x_o0_t0_l0, i_m_Betau_x_o0_t1_l0, i_m_Betau_x_o0_t2_l0, i_m_Betau_y_o0_t0_l0, i_m_Betau_y_o0_t1_l0, i_m_Betau_y_o0_t2_l0, i_m_Betau_z_o0_t0_l0, i_m_Betau_z_o0_t1_l0, i_m_Betau_z_o0_t2_l0, i_m_Alpha_o0_t0_l0, i_m_Alpha_o0_t1_l0, i_m_Alpha_o0_t2_l0, i_m_chi_o0_t0_l0, i_m_chi_o0_t1_l0, i_m_chi_o0_t2_l0, i_m_trK_o0_t0_l0, i_m_trK_o0_t1_l0, i_m_trK_o0_t2_l0, i_m_theta_o0_t0_l0, i_m_theta_o0_t1_l0, i_m_theta_o0_t2_l0;
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					if ((i + 3 < ilast && i - 3 >= 0 && j + 3 < jlast && j - 3 >= 0 && k + 3 < klast && k - 3 >= 0)) {
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j, k) * vector(Bfu_x_p, i, j, k) + vector(gtd_yy_p, i, j, k) * vector(Bfu_y_p, i, j, k) + vector(gtd_yz_p, i, j, k) * vector(Bfu_z_p, i, j, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j, k) * vector(Bfu_x_p, i, j, k) + vector(gtd_yz_p, i, j, k) * vector(Bfu_y_p, i, j, k) + vector(gtd_zz_p, i, j, k) * vector(Bfu_z_p, i, j, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j, k) * vector(Bfu_x_p, i, j, k) + vector(gtd_xy_p, i, j, k) * vector(Bfu_y_p, i, j, k) + vector(gtd_xz_p, i, j, k) * vector(Bfu_z_p, i, j, k));
						sqBf_p = vector(Bfu_x_p, i, j, k) * Bfd_x_p + vector(Bfu_y_p, i, j, k) * Bfd_y_p + vector(Bfu_z_p, i, j, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j, k) * vector(gtd_zz_p, i, j, k) * vector(gtd_xx_p, i, j, k) + (-vector(gtd_yy_p, i, j, k) * (vector(gtd_xz_p, i, j, k) * vector(gtd_xz_p, i, j, k))) + (-vector(gtd_zz_p, i, j, k) * (vector(gtd_xy_p, i, j, k) * vector(gtd_xy_p, i, j, k))) + (-(vector(gtd_yz_p, i, j, k) * vector(gtd_yz_p, i, j, k)) * vector(gtd_xx_p, i, j, k)) + 2.0 * vector(gtd_yz_p, i, j, k) * vector(gtd_xy_p, i, j, k) * vector(gtd_xz_p, i, j, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j, k) * vector(gtd_yy_p, i, j, k) - vector(gtd_xy_p, i, j, k) * vector(gtd_xy_p, i, j, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j, k) * vector(gtd_zz_p, i, j, k) - vector(gtd_yz_p, i, j, k) * vector(gtd_yz_p, i, j, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j, k) * vector(gtd_yz_p, i, j, k) - vector(gtd_yy_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j, k) * vector(gtd_zz_p, i, j, k)) + vector(gtd_yz_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j, k) + gtu_xy_p * vector(Sfd_y_p, i, j, k) + gtu_xz_p * vector(Sfd_z_p, i, j, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j, k) + gtu_xy_p * vector(vfd_y, i, j, k) + gtu_xz_p * vector(vfd_z, i, j, k));
						Bfvf_p = vector(Bfu_x_p, i, j, k) * vector(vfd_x, i, j, k) + vector(Bfu_y_p, i, j, k) * vector(vfd_y, i, j, k) + vector(Bfu_z_p, i, j, k) * vector(vfd_z, i, j, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j, k) * vector(gtd_zz_p, i, j, k) - vector(gtd_xz_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j, k) * vector(gtd_yz_p, i, j, k)) + vector(gtd_xy_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j, k) + gtu_yz_p * vector(vfd_y, i, j, k) + gtu_zz_p * vector(vfd_z, i, j, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j, k) + gtu_yy_p * vector(Sfd_y_p, i, j, k) + gtu_yz_p * vector(Sfd_z_p, i, j, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j, k) + gtu_yz_p * vector(Sfd_y_p, i, j, k) + gtu_zz_p * vector(Sfd_z_p, i, j, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j, k) + gtu_yy_p * vector(vfd_y, i, j, k) + gtu_yz_p * vector(vfd_z, i, j, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j, k) + vfu_y_p * vector(vfd_y, i, j, k) + vfu_z_p * vector(vfd_z, i, j, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j, k) * vector(Bfu_z_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j, k) * vector(Bfu_x_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k) * vector(Bfu_z_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k) * vfu_z_p + vector(Bfu_z_p, i, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k) * vector(Bfu_y_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k) * vfu_y_p + vector(Bfu_y_p, i, j, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j, k) + Sfuu_xy_p * vector(gtd_xy_p, i, j, k) + Sfuu_xz_p * vector(gtd_xz_p, i, j, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j, k) + Sfuu_xy_p * vector(gtd_yy_p, i, j, k) + Sfuu_xz_p * vector(gtd_yz_p, i, j, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j, k) + Sfuu_xy_p * vector(gtd_yz_p, i, j, k) + Sfuu_xz_p * vector(gtd_zz_p, i, j, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k) * vector(Bfu_z_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k) * vfu_z_p + vector(Bfu_z_p, i, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j, k) + Sfuu_yz_p * vector(gtd_yy_p, i, j, k) + Sfuu_zz_p * vector(gtd_yz_p, i, j, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k) + Sfuu_zz_p * vector(gtd_zz_p, i, j, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j, k) + Sfuu_yz_p * vector(gtd_xy_p, i, j, k) + Sfuu_zz_p * vector(gtd_xz_p, i, j, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j, k) * vector(Bfu_y_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j, k) + Sfuu_yy_p * vector(gtd_xy_p, i, j, k) + Sfuu_yz_p * vector(gtd_xz_p, i, j, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j, k) + Sfuu_yy_p * vector(gtd_yy_p, i, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j, k) + Sfuu_yy_p * vector(gtd_yz_p, i, j, k) + Sfuu_yz_p * vector(gtd_zz_p, i, j, k));
						FluxiSfd_x = FiSfd_x_mainI(vector(Sfd_x_p, i, j, k), vector(gtd_xx_p, i, j, k), vector(gtd_xy_p, i, j, k), vector(gtd_xz_p, i, j, k), vector(Betau_x_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauT_xx, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_xz, i, j, k), inv_chi_p, sdetg_p, Sfud_xx_p, dx, simPlat_dt, ilast, jlast);
						FluxiSfd_y = FiSfd_y_mainI(vector(Sfd_y_p, i, j, k), vector(gtd_xy_p, i, j, k), vector(gtd_yy_p, i, j, k), vector(gtd_yz_p, i, j, k), vector(Betau_x_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauT_xx, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_xz, i, j, k), inv_chi_p, sdetg_p, Sfud_xy_p, dx, simPlat_dt, ilast, jlast);
						FluxiSfd_z = FiSfd_z_mainI(vector(Sfd_z_p, i, j, k), vector(gtd_xz_p, i, j, k), vector(gtd_yz_p, i, j, k), vector(gtd_zz_p, i, j, k), vector(Betau_x_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauT_xx, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_xz, i, j, k), inv_chi_p, sdetg_p, Sfud_xz_p, dx, simPlat_dt, ilast, jlast);
						FluxiBfu_x = FiBfu_x_mainI(vector(phif_p, i, j, k), vector(Alpha_p, i, j, k), chi_max_p, gtu_xx_p, dx, simPlat_dt, ilast, jlast);
						FluxiBfu_y = FiBfu_y_mainI(vector(Bfu_x_p, i, j, k), vector(Bfu_y_p, i, j, k), vector(phif_p, i, j, k), vector(Betau_x_p, i, j, k), vector(Betau_y_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauM_xy, i, j, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						FluxiBfu_z = FiBfu_z_mainI(vector(Bfu_x_p, i, j, k), vector(Bfu_z_p, i, j, k), vector(phif_p, i, j, k), vector(Betau_x_p, i, j, k), vector(Betau_z_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauM_xz, i, j, k), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						FluxiDf = FiDf_mainI(vector(Df_p, i, j, k), vector(Betau_x_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauN_x, i, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						FluxiDYf = FiDYf_mainI(vector(DYf_p, i, j, k), vector(Betau_x_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauNe_x, i, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxitauf = Fitauf_mainI(vector(Df_p, i, j, k), vector(tauf_p, i, j, k), vector(Betau_x_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauN_x, i, j, k), sdetg_p, Sfu_x_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxiphif = Fiphif_mainI(vector(Bfu_x_p, i, j, k), vector(phif_p, i, j, k), vector(Betau_x_p, i, j, k), vector(Alpha_p, i, j, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j, k) + sqca_p - vector(sqcs, i, j, k) * sqca_p;
						SpeediSfd_x = fabs(fabs(vector(Betau_x_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p))))));
						SpeediSfd_y = fabs(fabs(vector(Betau_x_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p))))));
						SpeediSfd_z = fabs(fabs(vector(Betau_x_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p))))));
						SpeediBfu_x = fabs(fabs(vector(Betau_x_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p))))));
						SpeediBfu_y = fabs(fabs(vector(Betau_x_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p))))));
						SpeediBfu_z = fabs(fabs(vector(Betau_x_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p))))));
						SpeediDf = fabs(fabs(vector(Betau_x_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p))))));
						SpeediDYf = fabs(fabs(vector(Betau_x_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p))))));
						Speeditauf = fabs(fabs(vector(Betau_x_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p))))));
						Speediphif = fabs(fabs(vector(Betau_x_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i - 3, j, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i - 3, j, k) * vector(Bfu_x_p, i - 3, j, k) + vector(gtd_yy_p, i - 3, j, k) * vector(Bfu_y_p, i - 3, j, k) + vector(gtd_yz_p, i - 3, j, k) * vector(Bfu_z_p, i - 3, j, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i - 3, j, k) * vector(Bfu_x_p, i - 3, j, k) + vector(gtd_yz_p, i - 3, j, k) * vector(Bfu_y_p, i - 3, j, k) + vector(gtd_zz_p, i - 3, j, k) * vector(Bfu_z_p, i - 3, j, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i - 3, j, k) * vector(Bfu_x_p, i - 3, j, k) + vector(gtd_xy_p, i - 3, j, k) * vector(Bfu_y_p, i - 3, j, k) + vector(gtd_xz_p, i - 3, j, k) * vector(Bfu_z_p, i - 3, j, k));
						sqBf_p = vector(Bfu_x_p, i - 3, j, k) * Bfd_x_p + vector(Bfu_y_p, i - 3, j, k) * Bfd_y_p + vector(Bfu_z_p, i - 3, j, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i - 3, j, k) * vector(gtd_zz_p, i - 3, j, k) * vector(gtd_xx_p, i - 3, j, k) + (-vector(gtd_yy_p, i - 3, j, k) * (vector(gtd_xz_p, i - 3, j, k) * vector(gtd_xz_p, i - 3, j, k))) + (-vector(gtd_zz_p, i - 3, j, k) * (vector(gtd_xy_p, i - 3, j, k) * vector(gtd_xy_p, i - 3, j, k))) + (-(vector(gtd_yz_p, i - 3, j, k) * vector(gtd_yz_p, i - 3, j, k)) * vector(gtd_xx_p, i - 3, j, k)) + 2.0 * vector(gtd_yz_p, i - 3, j, k) * vector(gtd_xy_p, i - 3, j, k) * vector(gtd_xz_p, i - 3, j, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i - 3, j, k) * vector(gtd_yy_p, i - 3, j, k) - vector(gtd_xy_p, i - 3, j, k) * vector(gtd_xy_p, i - 3, j, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i - 3, j, k) * vector(gtd_zz_p, i - 3, j, k) - vector(gtd_yz_p, i - 3, j, k) * vector(gtd_yz_p, i - 3, j, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i - 3, j, k) * vector(gtd_yz_p, i - 3, j, k) - vector(gtd_yy_p, i - 3, j, k) * vector(gtd_xz_p, i - 3, j, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i - 3, j, k) * vector(gtd_zz_p, i - 3, j, k)) + vector(gtd_yz_p, i - 3, j, k) * vector(gtd_xz_p, i - 3, j, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i - 3, j, k) + gtu_xy_p * vector(Sfd_y_p, i - 3, j, k) + gtu_xz_p * vector(Sfd_z_p, i - 3, j, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i - 3, j, k) + gtu_xy_p * vector(vfd_y, i - 3, j, k) + gtu_xz_p * vector(vfd_z, i - 3, j, k));
						Bfvf_p = vector(Bfu_x_p, i - 3, j, k) * vector(vfd_x, i - 3, j, k) + vector(Bfu_y_p, i - 3, j, k) * vector(vfd_y, i - 3, j, k) + vector(Bfu_z_p, i - 3, j, k) * vector(vfd_z, i - 3, j, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i - 3, j, k) * vector(gtd_zz_p, i - 3, j, k) - vector(gtd_xz_p, i - 3, j, k) * vector(gtd_xz_p, i - 3, j, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i - 3, j, k) * vector(gtd_yz_p, i - 3, j, k)) + vector(gtd_xy_p, i - 3, j, k) * vector(gtd_xz_p, i - 3, j, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i - 3, j, k) + gtu_yz_p * vector(vfd_y, i - 3, j, k) + gtu_zz_p * vector(vfd_z, i - 3, j, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i - 3, j, k) + gtu_yy_p * vector(Sfd_y_p, i - 3, j, k) + gtu_yz_p * vector(Sfd_z_p, i - 3, j, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i - 3, j, k) + gtu_yz_p * vector(Sfd_y_p, i - 3, j, k) + gtu_zz_p * vector(Sfd_z_p, i - 3, j, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i - 3, j, k) + gtu_yy_p * vector(vfd_y, i - 3, j, k) + gtu_yz_p * vector(vfd_z, i - 3, j, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i - 3, j, k) + vfu_y_p * vector(vfd_y, i - 3, j, k) + vfu_z_p * vector(vfd_z, i - 3, j, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i - 3, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i - 3, j, k) * vector(Bfu_z_p, i - 3, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i - 3, j, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i - 3, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i - 3, j, k) * vector(Bfu_x_p, i - 3, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i - 3, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i - 3, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i - 3, j, k) * vector(Bfu_z_p, i - 3, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i - 3, j, k) * vfu_z_p + vector(Bfu_z_p, i - 3, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i - 3, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i - 3, j, k) * vector(Bfu_y_p, i - 3, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i - 3, j, k) * vfu_y_p + vector(Bfu_y_p, i - 3, j, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i - 3, j, k) + Sfuu_xy_p * vector(gtd_xy_p, i - 3, j, k) + Sfuu_xz_p * vector(gtd_xz_p, i - 3, j, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i - 3, j, k) + Sfuu_xy_p * vector(gtd_yy_p, i - 3, j, k) + Sfuu_xz_p * vector(gtd_yz_p, i - 3, j, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i - 3, j, k) + Sfuu_xy_p * vector(gtd_yz_p, i - 3, j, k) + Sfuu_xz_p * vector(gtd_zz_p, i - 3, j, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i - 3, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i - 3, j, k) * vector(Bfu_z_p, i - 3, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i - 3, j, k) * vfu_z_p + vector(Bfu_z_p, i - 3, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i - 3, j, k) + Sfuu_yz_p * vector(gtd_yy_p, i - 3, j, k) + Sfuu_zz_p * vector(gtd_yz_p, i - 3, j, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i - 3, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i - 3, j, k) + Sfuu_zz_p * vector(gtd_zz_p, i - 3, j, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i - 3, j, k) + Sfuu_yz_p * vector(gtd_xy_p, i - 3, j, k) + Sfuu_zz_p * vector(gtd_xz_p, i - 3, j, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i - 3, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i - 3, j, k) * vector(Bfu_y_p, i - 3, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i - 3, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i - 3, j, k) + Sfuu_yy_p * vector(gtd_xy_p, i - 3, j, k) + Sfuu_yz_p * vector(gtd_xz_p, i - 3, j, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i - 3, j, k) + Sfuu_yy_p * vector(gtd_yy_p, i - 3, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i - 3, j, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i - 3, j, k) + Sfuu_yy_p * vector(gtd_yz_p, i - 3, j, k) + Sfuu_yz_p * vector(gtd_zz_p, i - 3, j, k));
						Fluxm3iSfd_x = FiSfd_x_mainI(vector(Sfd_x_p, i - 3, j, k), vector(gtd_xx_p, i - 3, j, k), vector(gtd_xy_p, i - 3, j, k), vector(gtd_xz_p, i - 3, j, k), vector(Betau_x_p, i - 3, j, k), vector(Alpha_p, i - 3, j, k), vector(TauT_xx, i - 3, j, k), vector(TauT_xy, i - 3, j, k), vector(TauT_xz, i - 3, j, k), inv_chi_p, sdetg_p, Sfud_xx_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3iSfd_y = FiSfd_y_mainI(vector(Sfd_y_p, i - 3, j, k), vector(gtd_xy_p, i - 3, j, k), vector(gtd_yy_p, i - 3, j, k), vector(gtd_yz_p, i - 3, j, k), vector(Betau_x_p, i - 3, j, k), vector(Alpha_p, i - 3, j, k), vector(TauT_xx, i - 3, j, k), vector(TauT_xy, i - 3, j, k), vector(TauT_xz, i - 3, j, k), inv_chi_p, sdetg_p, Sfud_xy_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3iSfd_z = FiSfd_z_mainI(vector(Sfd_z_p, i - 3, j, k), vector(gtd_xz_p, i - 3, j, k), vector(gtd_yz_p, i - 3, j, k), vector(gtd_zz_p, i - 3, j, k), vector(Betau_x_p, i - 3, j, k), vector(Alpha_p, i - 3, j, k), vector(TauT_xx, i - 3, j, k), vector(TauT_xy, i - 3, j, k), vector(TauT_xz, i - 3, j, k), inv_chi_p, sdetg_p, Sfud_xz_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3iBfu_x = FiBfu_x_mainI(vector(phif_p, i - 3, j, k), vector(Alpha_p, i - 3, j, k), chi_max_p, gtu_xx_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3iBfu_y = FiBfu_y_mainI(vector(Bfu_x_p, i - 3, j, k), vector(Bfu_y_p, i - 3, j, k), vector(phif_p, i - 3, j, k), vector(Betau_x_p, i - 3, j, k), vector(Betau_y_p, i - 3, j, k), vector(Alpha_p, i - 3, j, k), vector(TauM_xy, i - 3, j, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3iBfu_z = FiBfu_z_mainI(vector(Bfu_x_p, i - 3, j, k), vector(Bfu_z_p, i - 3, j, k), vector(phif_p, i - 3, j, k), vector(Betau_x_p, i - 3, j, k), vector(Betau_z_p, i - 3, j, k), vector(Alpha_p, i - 3, j, k), vector(TauM_xz, i - 3, j, k), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3iDf = FiDf_mainI(vector(Df_p, i - 3, j, k), vector(Betau_x_p, i - 3, j, k), vector(Alpha_p, i - 3, j, k), vector(TauN_x, i - 3, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3iDYf = FiDYf_mainI(vector(DYf_p, i - 3, j, k), vector(Betau_x_p, i - 3, j, k), vector(Alpha_p, i - 3, j, k), vector(TauNe_x, i - 3, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3itauf = Fitauf_mainI(vector(Df_p, i - 3, j, k), vector(tauf_p, i - 3, j, k), vector(Betau_x_p, i - 3, j, k), vector(Alpha_p, i - 3, j, k), vector(TauN_x, i - 3, j, k), sdetg_p, Sfu_x_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3iphif = Fiphif_mainI(vector(Bfu_x_p, i - 3, j, k), vector(phif_p, i - 3, j, k), vector(Betau_x_p, i - 3, j, k), vector(Alpha_p, i - 3, j, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i - 3, j, k) * (1.0 + vector(epsf, i - 3, j, k)) + vector(pf, i - 3, j, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i - 3, j, k) + sqca_p - vector(sqcs, i - 3, j, k) * sqca_p;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(Betau_x_p, i - 3, j, k)) + vector(Alpha_p, i - 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(Betau_x_p, i - 3, j, k)) + vector(Alpha_p, i - 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(Betau_x_p, i - 3, j, k)) + vector(Alpha_p, i - 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(Betau_x_p, i - 3, j, k)) + vector(Alpha_p, i - 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(Betau_x_p, i - 3, j, k)) + vector(Alpha_p, i - 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(Betau_x_p, i - 3, j, k)) + vector(Alpha_p, i - 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(Betau_x_p, i - 3, j, k)) + vector(Alpha_p, i - 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(Betau_x_p, i - 3, j, k)) + vector(Alpha_p, i - 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(Betau_x_p, i - 3, j, k)) + vector(Alpha_p, i - 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(Betau_x_p, i - 3, j, k)) + vector(Alpha_p, i - 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i - 2, j, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i - 2, j, k) * vector(Bfu_x_p, i - 2, j, k) + vector(gtd_yy_p, i - 2, j, k) * vector(Bfu_y_p, i - 2, j, k) + vector(gtd_yz_p, i - 2, j, k) * vector(Bfu_z_p, i - 2, j, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i - 2, j, k) * vector(Bfu_x_p, i - 2, j, k) + vector(gtd_yz_p, i - 2, j, k) * vector(Bfu_y_p, i - 2, j, k) + vector(gtd_zz_p, i - 2, j, k) * vector(Bfu_z_p, i - 2, j, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i - 2, j, k) * vector(Bfu_x_p, i - 2, j, k) + vector(gtd_xy_p, i - 2, j, k) * vector(Bfu_y_p, i - 2, j, k) + vector(gtd_xz_p, i - 2, j, k) * vector(Bfu_z_p, i - 2, j, k));
						sqBf_p = vector(Bfu_x_p, i - 2, j, k) * Bfd_x_p + vector(Bfu_y_p, i - 2, j, k) * Bfd_y_p + vector(Bfu_z_p, i - 2, j, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i - 2, j, k) * vector(gtd_zz_p, i - 2, j, k) * vector(gtd_xx_p, i - 2, j, k) + (-vector(gtd_yy_p, i - 2, j, k) * (vector(gtd_xz_p, i - 2, j, k) * vector(gtd_xz_p, i - 2, j, k))) + (-vector(gtd_zz_p, i - 2, j, k) * (vector(gtd_xy_p, i - 2, j, k) * vector(gtd_xy_p, i - 2, j, k))) + (-(vector(gtd_yz_p, i - 2, j, k) * vector(gtd_yz_p, i - 2, j, k)) * vector(gtd_xx_p, i - 2, j, k)) + 2.0 * vector(gtd_yz_p, i - 2, j, k) * vector(gtd_xy_p, i - 2, j, k) * vector(gtd_xz_p, i - 2, j, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i - 2, j, k) * vector(gtd_yy_p, i - 2, j, k) - vector(gtd_xy_p, i - 2, j, k) * vector(gtd_xy_p, i - 2, j, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i - 2, j, k) * vector(gtd_zz_p, i - 2, j, k) - vector(gtd_yz_p, i - 2, j, k) * vector(gtd_yz_p, i - 2, j, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i - 2, j, k) * vector(gtd_yz_p, i - 2, j, k) - vector(gtd_yy_p, i - 2, j, k) * vector(gtd_xz_p, i - 2, j, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i - 2, j, k) * vector(gtd_zz_p, i - 2, j, k)) + vector(gtd_yz_p, i - 2, j, k) * vector(gtd_xz_p, i - 2, j, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i - 2, j, k) + gtu_xy_p * vector(Sfd_y_p, i - 2, j, k) + gtu_xz_p * vector(Sfd_z_p, i - 2, j, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i - 2, j, k) + gtu_xy_p * vector(vfd_y, i - 2, j, k) + gtu_xz_p * vector(vfd_z, i - 2, j, k));
						Bfvf_p = vector(Bfu_x_p, i - 2, j, k) * vector(vfd_x, i - 2, j, k) + vector(Bfu_y_p, i - 2, j, k) * vector(vfd_y, i - 2, j, k) + vector(Bfu_z_p, i - 2, j, k) * vector(vfd_z, i - 2, j, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i - 2, j, k) * vector(gtd_zz_p, i - 2, j, k) - vector(gtd_xz_p, i - 2, j, k) * vector(gtd_xz_p, i - 2, j, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i - 2, j, k) * vector(gtd_yz_p, i - 2, j, k)) + vector(gtd_xy_p, i - 2, j, k) * vector(gtd_xz_p, i - 2, j, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i - 2, j, k) + gtu_yz_p * vector(vfd_y, i - 2, j, k) + gtu_zz_p * vector(vfd_z, i - 2, j, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i - 2, j, k) + gtu_yy_p * vector(Sfd_y_p, i - 2, j, k) + gtu_yz_p * vector(Sfd_z_p, i - 2, j, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i - 2, j, k) + gtu_yz_p * vector(Sfd_y_p, i - 2, j, k) + gtu_zz_p * vector(Sfd_z_p, i - 2, j, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i - 2, j, k) + gtu_yy_p * vector(vfd_y, i - 2, j, k) + gtu_yz_p * vector(vfd_z, i - 2, j, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i - 2, j, k) + vfu_y_p * vector(vfd_y, i - 2, j, k) + vfu_z_p * vector(vfd_z, i - 2, j, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i - 2, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i - 2, j, k) * vector(Bfu_z_p, i - 2, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i - 2, j, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i - 2, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i - 2, j, k) * vector(Bfu_x_p, i - 2, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i - 2, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i - 2, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i - 2, j, k) * vector(Bfu_z_p, i - 2, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i - 2, j, k) * vfu_z_p + vector(Bfu_z_p, i - 2, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i - 2, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i - 2, j, k) * vector(Bfu_y_p, i - 2, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i - 2, j, k) * vfu_y_p + vector(Bfu_y_p, i - 2, j, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i - 2, j, k) + Sfuu_xy_p * vector(gtd_xy_p, i - 2, j, k) + Sfuu_xz_p * vector(gtd_xz_p, i - 2, j, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i - 2, j, k) + Sfuu_xy_p * vector(gtd_yy_p, i - 2, j, k) + Sfuu_xz_p * vector(gtd_yz_p, i - 2, j, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i - 2, j, k) + Sfuu_xy_p * vector(gtd_yz_p, i - 2, j, k) + Sfuu_xz_p * vector(gtd_zz_p, i - 2, j, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i - 2, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i - 2, j, k) * vector(Bfu_z_p, i - 2, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i - 2, j, k) * vfu_z_p + vector(Bfu_z_p, i - 2, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i - 2, j, k) + Sfuu_yz_p * vector(gtd_yy_p, i - 2, j, k) + Sfuu_zz_p * vector(gtd_yz_p, i - 2, j, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i - 2, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i - 2, j, k) + Sfuu_zz_p * vector(gtd_zz_p, i - 2, j, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i - 2, j, k) + Sfuu_yz_p * vector(gtd_xy_p, i - 2, j, k) + Sfuu_zz_p * vector(gtd_xz_p, i - 2, j, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i - 2, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i - 2, j, k) * vector(Bfu_y_p, i - 2, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i - 2, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i - 2, j, k) + Sfuu_yy_p * vector(gtd_xy_p, i - 2, j, k) + Sfuu_yz_p * vector(gtd_xz_p, i - 2, j, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i - 2, j, k) + Sfuu_yy_p * vector(gtd_yy_p, i - 2, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i - 2, j, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i - 2, j, k) + Sfuu_yy_p * vector(gtd_yz_p, i - 2, j, k) + Sfuu_yz_p * vector(gtd_zz_p, i - 2, j, k));
						Fluxm2iSfd_x = FiSfd_x_mainI(vector(Sfd_x_p, i - 2, j, k), vector(gtd_xx_p, i - 2, j, k), vector(gtd_xy_p, i - 2, j, k), vector(gtd_xz_p, i - 2, j, k), vector(Betau_x_p, i - 2, j, k), vector(Alpha_p, i - 2, j, k), vector(TauT_xx, i - 2, j, k), vector(TauT_xy, i - 2, j, k), vector(TauT_xz, i - 2, j, k), inv_chi_p, sdetg_p, Sfud_xx_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2iSfd_y = FiSfd_y_mainI(vector(Sfd_y_p, i - 2, j, k), vector(gtd_xy_p, i - 2, j, k), vector(gtd_yy_p, i - 2, j, k), vector(gtd_yz_p, i - 2, j, k), vector(Betau_x_p, i - 2, j, k), vector(Alpha_p, i - 2, j, k), vector(TauT_xx, i - 2, j, k), vector(TauT_xy, i - 2, j, k), vector(TauT_xz, i - 2, j, k), inv_chi_p, sdetg_p, Sfud_xy_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2iSfd_z = FiSfd_z_mainI(vector(Sfd_z_p, i - 2, j, k), vector(gtd_xz_p, i - 2, j, k), vector(gtd_yz_p, i - 2, j, k), vector(gtd_zz_p, i - 2, j, k), vector(Betau_x_p, i - 2, j, k), vector(Alpha_p, i - 2, j, k), vector(TauT_xx, i - 2, j, k), vector(TauT_xy, i - 2, j, k), vector(TauT_xz, i - 2, j, k), inv_chi_p, sdetg_p, Sfud_xz_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2iBfu_x = FiBfu_x_mainI(vector(phif_p, i - 2, j, k), vector(Alpha_p, i - 2, j, k), chi_max_p, gtu_xx_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2iBfu_y = FiBfu_y_mainI(vector(Bfu_x_p, i - 2, j, k), vector(Bfu_y_p, i - 2, j, k), vector(phif_p, i - 2, j, k), vector(Betau_x_p, i - 2, j, k), vector(Betau_y_p, i - 2, j, k), vector(Alpha_p, i - 2, j, k), vector(TauM_xy, i - 2, j, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2iBfu_z = FiBfu_z_mainI(vector(Bfu_x_p, i - 2, j, k), vector(Bfu_z_p, i - 2, j, k), vector(phif_p, i - 2, j, k), vector(Betau_x_p, i - 2, j, k), vector(Betau_z_p, i - 2, j, k), vector(Alpha_p, i - 2, j, k), vector(TauM_xz, i - 2, j, k), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2iDf = FiDf_mainI(vector(Df_p, i - 2, j, k), vector(Betau_x_p, i - 2, j, k), vector(Alpha_p, i - 2, j, k), vector(TauN_x, i - 2, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2iDYf = FiDYf_mainI(vector(DYf_p, i - 2, j, k), vector(Betau_x_p, i - 2, j, k), vector(Alpha_p, i - 2, j, k), vector(TauNe_x, i - 2, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2itauf = Fitauf_mainI(vector(Df_p, i - 2, j, k), vector(tauf_p, i - 2, j, k), vector(Betau_x_p, i - 2, j, k), vector(Alpha_p, i - 2, j, k), vector(TauN_x, i - 2, j, k), sdetg_p, Sfu_x_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2iphif = Fiphif_mainI(vector(Bfu_x_p, i - 2, j, k), vector(phif_p, i - 2, j, k), vector(Betau_x_p, i - 2, j, k), vector(Alpha_p, i - 2, j, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i - 2, j, k) * (1.0 + vector(epsf, i - 2, j, k)) + vector(pf, i - 2, j, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i - 2, j, k) + sqca_p - vector(sqcs, i - 2, j, k) * sqca_p;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(Betau_x_p, i - 2, j, k)) + vector(Alpha_p, i - 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(Betau_x_p, i - 2, j, k)) + vector(Alpha_p, i - 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(Betau_x_p, i - 2, j, k)) + vector(Alpha_p, i - 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(Betau_x_p, i - 2, j, k)) + vector(Alpha_p, i - 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(Betau_x_p, i - 2, j, k)) + vector(Alpha_p, i - 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(Betau_x_p, i - 2, j, k)) + vector(Alpha_p, i - 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(Betau_x_p, i - 2, j, k)) + vector(Alpha_p, i - 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(Betau_x_p, i - 2, j, k)) + vector(Alpha_p, i - 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(Betau_x_p, i - 2, j, k)) + vector(Alpha_p, i - 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(Betau_x_p, i - 2, j, k)) + vector(Alpha_p, i - 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i - 1, j, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i - 1, j, k) * vector(Bfu_x_p, i - 1, j, k) + vector(gtd_yy_p, i - 1, j, k) * vector(Bfu_y_p, i - 1, j, k) + vector(gtd_yz_p, i - 1, j, k) * vector(Bfu_z_p, i - 1, j, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i - 1, j, k) * vector(Bfu_x_p, i - 1, j, k) + vector(gtd_yz_p, i - 1, j, k) * vector(Bfu_y_p, i - 1, j, k) + vector(gtd_zz_p, i - 1, j, k) * vector(Bfu_z_p, i - 1, j, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i - 1, j, k) * vector(Bfu_x_p, i - 1, j, k) + vector(gtd_xy_p, i - 1, j, k) * vector(Bfu_y_p, i - 1, j, k) + vector(gtd_xz_p, i - 1, j, k) * vector(Bfu_z_p, i - 1, j, k));
						sqBf_p = vector(Bfu_x_p, i - 1, j, k) * Bfd_x_p + vector(Bfu_y_p, i - 1, j, k) * Bfd_y_p + vector(Bfu_z_p, i - 1, j, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i - 1, j, k) * vector(gtd_zz_p, i - 1, j, k) * vector(gtd_xx_p, i - 1, j, k) + (-vector(gtd_yy_p, i - 1, j, k) * (vector(gtd_xz_p, i - 1, j, k) * vector(gtd_xz_p, i - 1, j, k))) + (-vector(gtd_zz_p, i - 1, j, k) * (vector(gtd_xy_p, i - 1, j, k) * vector(gtd_xy_p, i - 1, j, k))) + (-(vector(gtd_yz_p, i - 1, j, k) * vector(gtd_yz_p, i - 1, j, k)) * vector(gtd_xx_p, i - 1, j, k)) + 2.0 * vector(gtd_yz_p, i - 1, j, k) * vector(gtd_xy_p, i - 1, j, k) * vector(gtd_xz_p, i - 1, j, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i - 1, j, k) * vector(gtd_yy_p, i - 1, j, k) - vector(gtd_xy_p, i - 1, j, k) * vector(gtd_xy_p, i - 1, j, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i - 1, j, k) * vector(gtd_zz_p, i - 1, j, k) - vector(gtd_yz_p, i - 1, j, k) * vector(gtd_yz_p, i - 1, j, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i - 1, j, k) * vector(gtd_yz_p, i - 1, j, k) - vector(gtd_yy_p, i - 1, j, k) * vector(gtd_xz_p, i - 1, j, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i - 1, j, k) * vector(gtd_zz_p, i - 1, j, k)) + vector(gtd_yz_p, i - 1, j, k) * vector(gtd_xz_p, i - 1, j, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i - 1, j, k) + gtu_xy_p * vector(Sfd_y_p, i - 1, j, k) + gtu_xz_p * vector(Sfd_z_p, i - 1, j, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i - 1, j, k) + gtu_xy_p * vector(vfd_y, i - 1, j, k) + gtu_xz_p * vector(vfd_z, i - 1, j, k));
						Bfvf_p = vector(Bfu_x_p, i - 1, j, k) * vector(vfd_x, i - 1, j, k) + vector(Bfu_y_p, i - 1, j, k) * vector(vfd_y, i - 1, j, k) + vector(Bfu_z_p, i - 1, j, k) * vector(vfd_z, i - 1, j, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i - 1, j, k) * vector(gtd_zz_p, i - 1, j, k) - vector(gtd_xz_p, i - 1, j, k) * vector(gtd_xz_p, i - 1, j, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i - 1, j, k) * vector(gtd_yz_p, i - 1, j, k)) + vector(gtd_xy_p, i - 1, j, k) * vector(gtd_xz_p, i - 1, j, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i - 1, j, k) + gtu_yz_p * vector(vfd_y, i - 1, j, k) + gtu_zz_p * vector(vfd_z, i - 1, j, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i - 1, j, k) + gtu_yy_p * vector(Sfd_y_p, i - 1, j, k) + gtu_yz_p * vector(Sfd_z_p, i - 1, j, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i - 1, j, k) + gtu_yz_p * vector(Sfd_y_p, i - 1, j, k) + gtu_zz_p * vector(Sfd_z_p, i - 1, j, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i - 1, j, k) + gtu_yy_p * vector(vfd_y, i - 1, j, k) + gtu_yz_p * vector(vfd_z, i - 1, j, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i - 1, j, k) + vfu_y_p * vector(vfd_y, i - 1, j, k) + vfu_z_p * vector(vfd_z, i - 1, j, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i - 1, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i - 1, j, k) * vector(Bfu_z_p, i - 1, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i - 1, j, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i - 1, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i - 1, j, k) * vector(Bfu_x_p, i - 1, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i - 1, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i - 1, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i - 1, j, k) * vector(Bfu_z_p, i - 1, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i - 1, j, k) * vfu_z_p + vector(Bfu_z_p, i - 1, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i - 1, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i - 1, j, k) * vector(Bfu_y_p, i - 1, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i - 1, j, k) * vfu_y_p + vector(Bfu_y_p, i - 1, j, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i - 1, j, k) + Sfuu_xy_p * vector(gtd_xy_p, i - 1, j, k) + Sfuu_xz_p * vector(gtd_xz_p, i - 1, j, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i - 1, j, k) + Sfuu_xy_p * vector(gtd_yy_p, i - 1, j, k) + Sfuu_xz_p * vector(gtd_yz_p, i - 1, j, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i - 1, j, k) + Sfuu_xy_p * vector(gtd_yz_p, i - 1, j, k) + Sfuu_xz_p * vector(gtd_zz_p, i - 1, j, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i - 1, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i - 1, j, k) * vector(Bfu_z_p, i - 1, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i - 1, j, k) * vfu_z_p + vector(Bfu_z_p, i - 1, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i - 1, j, k) + Sfuu_yz_p * vector(gtd_yy_p, i - 1, j, k) + Sfuu_zz_p * vector(gtd_yz_p, i - 1, j, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i - 1, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i - 1, j, k) + Sfuu_zz_p * vector(gtd_zz_p, i - 1, j, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i - 1, j, k) + Sfuu_yz_p * vector(gtd_xy_p, i - 1, j, k) + Sfuu_zz_p * vector(gtd_xz_p, i - 1, j, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i - 1, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i - 1, j, k) * vector(Bfu_y_p, i - 1, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i - 1, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i - 1, j, k) + Sfuu_yy_p * vector(gtd_xy_p, i - 1, j, k) + Sfuu_yz_p * vector(gtd_xz_p, i - 1, j, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i - 1, j, k) + Sfuu_yy_p * vector(gtd_yy_p, i - 1, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i - 1, j, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i - 1, j, k) + Sfuu_yy_p * vector(gtd_yz_p, i - 1, j, k) + Sfuu_yz_p * vector(gtd_zz_p, i - 1, j, k));
						Fluxm1iSfd_x = FiSfd_x_mainI(vector(Sfd_x_p, i - 1, j, k), vector(gtd_xx_p, i - 1, j, k), vector(gtd_xy_p, i - 1, j, k), vector(gtd_xz_p, i - 1, j, k), vector(Betau_x_p, i - 1, j, k), vector(Alpha_p, i - 1, j, k), vector(TauT_xx, i - 1, j, k), vector(TauT_xy, i - 1, j, k), vector(TauT_xz, i - 1, j, k), inv_chi_p, sdetg_p, Sfud_xx_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1iSfd_y = FiSfd_y_mainI(vector(Sfd_y_p, i - 1, j, k), vector(gtd_xy_p, i - 1, j, k), vector(gtd_yy_p, i - 1, j, k), vector(gtd_yz_p, i - 1, j, k), vector(Betau_x_p, i - 1, j, k), vector(Alpha_p, i - 1, j, k), vector(TauT_xx, i - 1, j, k), vector(TauT_xy, i - 1, j, k), vector(TauT_xz, i - 1, j, k), inv_chi_p, sdetg_p, Sfud_xy_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1iSfd_z = FiSfd_z_mainI(vector(Sfd_z_p, i - 1, j, k), vector(gtd_xz_p, i - 1, j, k), vector(gtd_yz_p, i - 1, j, k), vector(gtd_zz_p, i - 1, j, k), vector(Betau_x_p, i - 1, j, k), vector(Alpha_p, i - 1, j, k), vector(TauT_xx, i - 1, j, k), vector(TauT_xy, i - 1, j, k), vector(TauT_xz, i - 1, j, k), inv_chi_p, sdetg_p, Sfud_xz_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1iBfu_x = FiBfu_x_mainI(vector(phif_p, i - 1, j, k), vector(Alpha_p, i - 1, j, k), chi_max_p, gtu_xx_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1iBfu_y = FiBfu_y_mainI(vector(Bfu_x_p, i - 1, j, k), vector(Bfu_y_p, i - 1, j, k), vector(phif_p, i - 1, j, k), vector(Betau_x_p, i - 1, j, k), vector(Betau_y_p, i - 1, j, k), vector(Alpha_p, i - 1, j, k), vector(TauM_xy, i - 1, j, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1iBfu_z = FiBfu_z_mainI(vector(Bfu_x_p, i - 1, j, k), vector(Bfu_z_p, i - 1, j, k), vector(phif_p, i - 1, j, k), vector(Betau_x_p, i - 1, j, k), vector(Betau_z_p, i - 1, j, k), vector(Alpha_p, i - 1, j, k), vector(TauM_xz, i - 1, j, k), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1iDf = FiDf_mainI(vector(Df_p, i - 1, j, k), vector(Betau_x_p, i - 1, j, k), vector(Alpha_p, i - 1, j, k), vector(TauN_x, i - 1, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1iDYf = FiDYf_mainI(vector(DYf_p, i - 1, j, k), vector(Betau_x_p, i - 1, j, k), vector(Alpha_p, i - 1, j, k), vector(TauNe_x, i - 1, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1itauf = Fitauf_mainI(vector(Df_p, i - 1, j, k), vector(tauf_p, i - 1, j, k), vector(Betau_x_p, i - 1, j, k), vector(Alpha_p, i - 1, j, k), vector(TauN_x, i - 1, j, k), sdetg_p, Sfu_x_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1iphif = Fiphif_mainI(vector(Bfu_x_p, i - 1, j, k), vector(phif_p, i - 1, j, k), vector(Betau_x_p, i - 1, j, k), vector(Alpha_p, i - 1, j, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i - 1, j, k) * (1.0 + vector(epsf, i - 1, j, k)) + vector(pf, i - 1, j, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i - 1, j, k) + sqca_p - vector(sqcs, i - 1, j, k) * sqca_p;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(Betau_x_p, i - 1, j, k)) + vector(Alpha_p, i - 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(Betau_x_p, i - 1, j, k)) + vector(Alpha_p, i - 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(Betau_x_p, i - 1, j, k)) + vector(Alpha_p, i - 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(Betau_x_p, i - 1, j, k)) + vector(Alpha_p, i - 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(Betau_x_p, i - 1, j, k)) + vector(Alpha_p, i - 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(Betau_x_p, i - 1, j, k)) + vector(Alpha_p, i - 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(Betau_x_p, i - 1, j, k)) + vector(Alpha_p, i - 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(Betau_x_p, i - 1, j, k)) + vector(Alpha_p, i - 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(Betau_x_p, i - 1, j, k)) + vector(Alpha_p, i - 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(Betau_x_p, i - 1, j, k)) + vector(Alpha_p, i - 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i + 1, j, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i + 1, j, k) * vector(Bfu_x_p, i + 1, j, k) + vector(gtd_yy_p, i + 1, j, k) * vector(Bfu_y_p, i + 1, j, k) + vector(gtd_yz_p, i + 1, j, k) * vector(Bfu_z_p, i + 1, j, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i + 1, j, k) * vector(Bfu_x_p, i + 1, j, k) + vector(gtd_yz_p, i + 1, j, k) * vector(Bfu_y_p, i + 1, j, k) + vector(gtd_zz_p, i + 1, j, k) * vector(Bfu_z_p, i + 1, j, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i + 1, j, k) * vector(Bfu_x_p, i + 1, j, k) + vector(gtd_xy_p, i + 1, j, k) * vector(Bfu_y_p, i + 1, j, k) + vector(gtd_xz_p, i + 1, j, k) * vector(Bfu_z_p, i + 1, j, k));
						sqBf_p = vector(Bfu_x_p, i + 1, j, k) * Bfd_x_p + vector(Bfu_y_p, i + 1, j, k) * Bfd_y_p + vector(Bfu_z_p, i + 1, j, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i + 1, j, k) * vector(gtd_zz_p, i + 1, j, k) * vector(gtd_xx_p, i + 1, j, k) + (-vector(gtd_yy_p, i + 1, j, k) * (vector(gtd_xz_p, i + 1, j, k) * vector(gtd_xz_p, i + 1, j, k))) + (-vector(gtd_zz_p, i + 1, j, k) * (vector(gtd_xy_p, i + 1, j, k) * vector(gtd_xy_p, i + 1, j, k))) + (-(vector(gtd_yz_p, i + 1, j, k) * vector(gtd_yz_p, i + 1, j, k)) * vector(gtd_xx_p, i + 1, j, k)) + 2.0 * vector(gtd_yz_p, i + 1, j, k) * vector(gtd_xy_p, i + 1, j, k) * vector(gtd_xz_p, i + 1, j, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i + 1, j, k) * vector(gtd_yy_p, i + 1, j, k) - vector(gtd_xy_p, i + 1, j, k) * vector(gtd_xy_p, i + 1, j, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i + 1, j, k) * vector(gtd_zz_p, i + 1, j, k) - vector(gtd_yz_p, i + 1, j, k) * vector(gtd_yz_p, i + 1, j, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i + 1, j, k) * vector(gtd_yz_p, i + 1, j, k) - vector(gtd_yy_p, i + 1, j, k) * vector(gtd_xz_p, i + 1, j, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i + 1, j, k) * vector(gtd_zz_p, i + 1, j, k)) + vector(gtd_yz_p, i + 1, j, k) * vector(gtd_xz_p, i + 1, j, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i + 1, j, k) + gtu_xy_p * vector(Sfd_y_p, i + 1, j, k) + gtu_xz_p * vector(Sfd_z_p, i + 1, j, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i + 1, j, k) + gtu_xy_p * vector(vfd_y, i + 1, j, k) + gtu_xz_p * vector(vfd_z, i + 1, j, k));
						Bfvf_p = vector(Bfu_x_p, i + 1, j, k) * vector(vfd_x, i + 1, j, k) + vector(Bfu_y_p, i + 1, j, k) * vector(vfd_y, i + 1, j, k) + vector(Bfu_z_p, i + 1, j, k) * vector(vfd_z, i + 1, j, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i + 1, j, k) * vector(gtd_zz_p, i + 1, j, k) - vector(gtd_xz_p, i + 1, j, k) * vector(gtd_xz_p, i + 1, j, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i + 1, j, k) * vector(gtd_yz_p, i + 1, j, k)) + vector(gtd_xy_p, i + 1, j, k) * vector(gtd_xz_p, i + 1, j, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i + 1, j, k) + gtu_yz_p * vector(vfd_y, i + 1, j, k) + gtu_zz_p * vector(vfd_z, i + 1, j, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i + 1, j, k) + gtu_yy_p * vector(Sfd_y_p, i + 1, j, k) + gtu_yz_p * vector(Sfd_z_p, i + 1, j, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i + 1, j, k) + gtu_yz_p * vector(Sfd_y_p, i + 1, j, k) + gtu_zz_p * vector(Sfd_z_p, i + 1, j, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i + 1, j, k) + gtu_yy_p * vector(vfd_y, i + 1, j, k) + gtu_yz_p * vector(vfd_z, i + 1, j, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i + 1, j, k) + vfu_y_p * vector(vfd_y, i + 1, j, k) + vfu_z_p * vector(vfd_z, i + 1, j, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i + 1, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i + 1, j, k) * vector(Bfu_z_p, i + 1, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i + 1, j, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i + 1, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i + 1, j, k) * vector(Bfu_x_p, i + 1, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i + 1, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i + 1, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i + 1, j, k) * vector(Bfu_z_p, i + 1, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i + 1, j, k) * vfu_z_p + vector(Bfu_z_p, i + 1, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i + 1, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i + 1, j, k) * vector(Bfu_y_p, i + 1, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i + 1, j, k) * vfu_y_p + vector(Bfu_y_p, i + 1, j, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i + 1, j, k) + Sfuu_xy_p * vector(gtd_xy_p, i + 1, j, k) + Sfuu_xz_p * vector(gtd_xz_p, i + 1, j, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i + 1, j, k) + Sfuu_xy_p * vector(gtd_yy_p, i + 1, j, k) + Sfuu_xz_p * vector(gtd_yz_p, i + 1, j, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i + 1, j, k) + Sfuu_xy_p * vector(gtd_yz_p, i + 1, j, k) + Sfuu_xz_p * vector(gtd_zz_p, i + 1, j, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i + 1, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i + 1, j, k) * vector(Bfu_z_p, i + 1, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i + 1, j, k) * vfu_z_p + vector(Bfu_z_p, i + 1, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i + 1, j, k) + Sfuu_yz_p * vector(gtd_yy_p, i + 1, j, k) + Sfuu_zz_p * vector(gtd_yz_p, i + 1, j, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i + 1, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i + 1, j, k) + Sfuu_zz_p * vector(gtd_zz_p, i + 1, j, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i + 1, j, k) + Sfuu_yz_p * vector(gtd_xy_p, i + 1, j, k) + Sfuu_zz_p * vector(gtd_xz_p, i + 1, j, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i + 1, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i + 1, j, k) * vector(Bfu_y_p, i + 1, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i + 1, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i + 1, j, k) + Sfuu_yy_p * vector(gtd_xy_p, i + 1, j, k) + Sfuu_yz_p * vector(gtd_xz_p, i + 1, j, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i + 1, j, k) + Sfuu_yy_p * vector(gtd_yy_p, i + 1, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i + 1, j, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i + 1, j, k) + Sfuu_yy_p * vector(gtd_yz_p, i + 1, j, k) + Sfuu_yz_p * vector(gtd_zz_p, i + 1, j, k));
						Fluxp1iSfd_x = FiSfd_x_mainI(vector(Sfd_x_p, i + 1, j, k), vector(gtd_xx_p, i + 1, j, k), vector(gtd_xy_p, i + 1, j, k), vector(gtd_xz_p, i + 1, j, k), vector(Betau_x_p, i + 1, j, k), vector(Alpha_p, i + 1, j, k), vector(TauT_xx, i + 1, j, k), vector(TauT_xy, i + 1, j, k), vector(TauT_xz, i + 1, j, k), inv_chi_p, sdetg_p, Sfud_xx_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1iSfd_y = FiSfd_y_mainI(vector(Sfd_y_p, i + 1, j, k), vector(gtd_xy_p, i + 1, j, k), vector(gtd_yy_p, i + 1, j, k), vector(gtd_yz_p, i + 1, j, k), vector(Betau_x_p, i + 1, j, k), vector(Alpha_p, i + 1, j, k), vector(TauT_xx, i + 1, j, k), vector(TauT_xy, i + 1, j, k), vector(TauT_xz, i + 1, j, k), inv_chi_p, sdetg_p, Sfud_xy_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1iSfd_z = FiSfd_z_mainI(vector(Sfd_z_p, i + 1, j, k), vector(gtd_xz_p, i + 1, j, k), vector(gtd_yz_p, i + 1, j, k), vector(gtd_zz_p, i + 1, j, k), vector(Betau_x_p, i + 1, j, k), vector(Alpha_p, i + 1, j, k), vector(TauT_xx, i + 1, j, k), vector(TauT_xy, i + 1, j, k), vector(TauT_xz, i + 1, j, k), inv_chi_p, sdetg_p, Sfud_xz_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1iBfu_x = FiBfu_x_mainI(vector(phif_p, i + 1, j, k), vector(Alpha_p, i + 1, j, k), chi_max_p, gtu_xx_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1iBfu_y = FiBfu_y_mainI(vector(Bfu_x_p, i + 1, j, k), vector(Bfu_y_p, i + 1, j, k), vector(phif_p, i + 1, j, k), vector(Betau_x_p, i + 1, j, k), vector(Betau_y_p, i + 1, j, k), vector(Alpha_p, i + 1, j, k), vector(TauM_xy, i + 1, j, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1iBfu_z = FiBfu_z_mainI(vector(Bfu_x_p, i + 1, j, k), vector(Bfu_z_p, i + 1, j, k), vector(phif_p, i + 1, j, k), vector(Betau_x_p, i + 1, j, k), vector(Betau_z_p, i + 1, j, k), vector(Alpha_p, i + 1, j, k), vector(TauM_xz, i + 1, j, k), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1iDf = FiDf_mainI(vector(Df_p, i + 1, j, k), vector(Betau_x_p, i + 1, j, k), vector(Alpha_p, i + 1, j, k), vector(TauN_x, i + 1, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1iDYf = FiDYf_mainI(vector(DYf_p, i + 1, j, k), vector(Betau_x_p, i + 1, j, k), vector(Alpha_p, i + 1, j, k), vector(TauNe_x, i + 1, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1itauf = Fitauf_mainI(vector(Df_p, i + 1, j, k), vector(tauf_p, i + 1, j, k), vector(Betau_x_p, i + 1, j, k), vector(Alpha_p, i + 1, j, k), vector(TauN_x, i + 1, j, k), sdetg_p, Sfu_x_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1iphif = Fiphif_mainI(vector(Bfu_x_p, i + 1, j, k), vector(phif_p, i + 1, j, k), vector(Betau_x_p, i + 1, j, k), vector(Alpha_p, i + 1, j, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i + 1, j, k) * (1.0 + vector(epsf, i + 1, j, k)) + vector(pf, i + 1, j, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i + 1, j, k) + sqca_p - vector(sqcs, i + 1, j, k) * sqca_p;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(Betau_x_p, i + 1, j, k)) + vector(Alpha_p, i + 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(Betau_x_p, i + 1, j, k)) + vector(Alpha_p, i + 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(Betau_x_p, i + 1, j, k)) + vector(Alpha_p, i + 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(Betau_x_p, i + 1, j, k)) + vector(Alpha_p, i + 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(Betau_x_p, i + 1, j, k)) + vector(Alpha_p, i + 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(Betau_x_p, i + 1, j, k)) + vector(Alpha_p, i + 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(Betau_x_p, i + 1, j, k)) + vector(Alpha_p, i + 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(Betau_x_p, i + 1, j, k)) + vector(Alpha_p, i + 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(Betau_x_p, i + 1, j, k)) + vector(Alpha_p, i + 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(Betau_x_p, i + 1, j, k)) + vector(Alpha_p, i + 1, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i + 2, j, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i + 2, j, k) * vector(Bfu_x_p, i + 2, j, k) + vector(gtd_yy_p, i + 2, j, k) * vector(Bfu_y_p, i + 2, j, k) + vector(gtd_yz_p, i + 2, j, k) * vector(Bfu_z_p, i + 2, j, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i + 2, j, k) * vector(Bfu_x_p, i + 2, j, k) + vector(gtd_yz_p, i + 2, j, k) * vector(Bfu_y_p, i + 2, j, k) + vector(gtd_zz_p, i + 2, j, k) * vector(Bfu_z_p, i + 2, j, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i + 2, j, k) * vector(Bfu_x_p, i + 2, j, k) + vector(gtd_xy_p, i + 2, j, k) * vector(Bfu_y_p, i + 2, j, k) + vector(gtd_xz_p, i + 2, j, k) * vector(Bfu_z_p, i + 2, j, k));
						sqBf_p = vector(Bfu_x_p, i + 2, j, k) * Bfd_x_p + vector(Bfu_y_p, i + 2, j, k) * Bfd_y_p + vector(Bfu_z_p, i + 2, j, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i + 2, j, k) * vector(gtd_zz_p, i + 2, j, k) * vector(gtd_xx_p, i + 2, j, k) + (-vector(gtd_yy_p, i + 2, j, k) * (vector(gtd_xz_p, i + 2, j, k) * vector(gtd_xz_p, i + 2, j, k))) + (-vector(gtd_zz_p, i + 2, j, k) * (vector(gtd_xy_p, i + 2, j, k) * vector(gtd_xy_p, i + 2, j, k))) + (-(vector(gtd_yz_p, i + 2, j, k) * vector(gtd_yz_p, i + 2, j, k)) * vector(gtd_xx_p, i + 2, j, k)) + 2.0 * vector(gtd_yz_p, i + 2, j, k) * vector(gtd_xy_p, i + 2, j, k) * vector(gtd_xz_p, i + 2, j, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i + 2, j, k) * vector(gtd_yy_p, i + 2, j, k) - vector(gtd_xy_p, i + 2, j, k) * vector(gtd_xy_p, i + 2, j, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i + 2, j, k) * vector(gtd_zz_p, i + 2, j, k) - vector(gtd_yz_p, i + 2, j, k) * vector(gtd_yz_p, i + 2, j, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i + 2, j, k) * vector(gtd_yz_p, i + 2, j, k) - vector(gtd_yy_p, i + 2, j, k) * vector(gtd_xz_p, i + 2, j, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i + 2, j, k) * vector(gtd_zz_p, i + 2, j, k)) + vector(gtd_yz_p, i + 2, j, k) * vector(gtd_xz_p, i + 2, j, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i + 2, j, k) + gtu_xy_p * vector(Sfd_y_p, i + 2, j, k) + gtu_xz_p * vector(Sfd_z_p, i + 2, j, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i + 2, j, k) + gtu_xy_p * vector(vfd_y, i + 2, j, k) + gtu_xz_p * vector(vfd_z, i + 2, j, k));
						Bfvf_p = vector(Bfu_x_p, i + 2, j, k) * vector(vfd_x, i + 2, j, k) + vector(Bfu_y_p, i + 2, j, k) * vector(vfd_y, i + 2, j, k) + vector(Bfu_z_p, i + 2, j, k) * vector(vfd_z, i + 2, j, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i + 2, j, k) * vector(gtd_zz_p, i + 2, j, k) - vector(gtd_xz_p, i + 2, j, k) * vector(gtd_xz_p, i + 2, j, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i + 2, j, k) * vector(gtd_yz_p, i + 2, j, k)) + vector(gtd_xy_p, i + 2, j, k) * vector(gtd_xz_p, i + 2, j, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i + 2, j, k) + gtu_yz_p * vector(vfd_y, i + 2, j, k) + gtu_zz_p * vector(vfd_z, i + 2, j, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i + 2, j, k) + gtu_yy_p * vector(Sfd_y_p, i + 2, j, k) + gtu_yz_p * vector(Sfd_z_p, i + 2, j, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i + 2, j, k) + gtu_yz_p * vector(Sfd_y_p, i + 2, j, k) + gtu_zz_p * vector(Sfd_z_p, i + 2, j, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i + 2, j, k) + gtu_yy_p * vector(vfd_y, i + 2, j, k) + gtu_yz_p * vector(vfd_z, i + 2, j, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i + 2, j, k) + vfu_y_p * vector(vfd_y, i + 2, j, k) + vfu_z_p * vector(vfd_z, i + 2, j, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i + 2, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i + 2, j, k) * vector(Bfu_z_p, i + 2, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i + 2, j, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i + 2, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i + 2, j, k) * vector(Bfu_x_p, i + 2, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i + 2, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i + 2, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i + 2, j, k) * vector(Bfu_z_p, i + 2, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i + 2, j, k) * vfu_z_p + vector(Bfu_z_p, i + 2, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i + 2, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i + 2, j, k) * vector(Bfu_y_p, i + 2, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i + 2, j, k) * vfu_y_p + vector(Bfu_y_p, i + 2, j, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i + 2, j, k) + Sfuu_xy_p * vector(gtd_xy_p, i + 2, j, k) + Sfuu_xz_p * vector(gtd_xz_p, i + 2, j, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i + 2, j, k) + Sfuu_xy_p * vector(gtd_yy_p, i + 2, j, k) + Sfuu_xz_p * vector(gtd_yz_p, i + 2, j, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i + 2, j, k) + Sfuu_xy_p * vector(gtd_yz_p, i + 2, j, k) + Sfuu_xz_p * vector(gtd_zz_p, i + 2, j, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i + 2, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i + 2, j, k) * vector(Bfu_z_p, i + 2, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i + 2, j, k) * vfu_z_p + vector(Bfu_z_p, i + 2, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i + 2, j, k) + Sfuu_yz_p * vector(gtd_yy_p, i + 2, j, k) + Sfuu_zz_p * vector(gtd_yz_p, i + 2, j, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i + 2, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i + 2, j, k) + Sfuu_zz_p * vector(gtd_zz_p, i + 2, j, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i + 2, j, k) + Sfuu_yz_p * vector(gtd_xy_p, i + 2, j, k) + Sfuu_zz_p * vector(gtd_xz_p, i + 2, j, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i + 2, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i + 2, j, k) * vector(Bfu_y_p, i + 2, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i + 2, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i + 2, j, k) + Sfuu_yy_p * vector(gtd_xy_p, i + 2, j, k) + Sfuu_yz_p * vector(gtd_xz_p, i + 2, j, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i + 2, j, k) + Sfuu_yy_p * vector(gtd_yy_p, i + 2, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i + 2, j, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i + 2, j, k) + Sfuu_yy_p * vector(gtd_yz_p, i + 2, j, k) + Sfuu_yz_p * vector(gtd_zz_p, i + 2, j, k));
						Fluxp2iSfd_x = FiSfd_x_mainI(vector(Sfd_x_p, i + 2, j, k), vector(gtd_xx_p, i + 2, j, k), vector(gtd_xy_p, i + 2, j, k), vector(gtd_xz_p, i + 2, j, k), vector(Betau_x_p, i + 2, j, k), vector(Alpha_p, i + 2, j, k), vector(TauT_xx, i + 2, j, k), vector(TauT_xy, i + 2, j, k), vector(TauT_xz, i + 2, j, k), inv_chi_p, sdetg_p, Sfud_xx_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2iSfd_y = FiSfd_y_mainI(vector(Sfd_y_p, i + 2, j, k), vector(gtd_xy_p, i + 2, j, k), vector(gtd_yy_p, i + 2, j, k), vector(gtd_yz_p, i + 2, j, k), vector(Betau_x_p, i + 2, j, k), vector(Alpha_p, i + 2, j, k), vector(TauT_xx, i + 2, j, k), vector(TauT_xy, i + 2, j, k), vector(TauT_xz, i + 2, j, k), inv_chi_p, sdetg_p, Sfud_xy_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2iSfd_z = FiSfd_z_mainI(vector(Sfd_z_p, i + 2, j, k), vector(gtd_xz_p, i + 2, j, k), vector(gtd_yz_p, i + 2, j, k), vector(gtd_zz_p, i + 2, j, k), vector(Betau_x_p, i + 2, j, k), vector(Alpha_p, i + 2, j, k), vector(TauT_xx, i + 2, j, k), vector(TauT_xy, i + 2, j, k), vector(TauT_xz, i + 2, j, k), inv_chi_p, sdetg_p, Sfud_xz_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2iBfu_x = FiBfu_x_mainI(vector(phif_p, i + 2, j, k), vector(Alpha_p, i + 2, j, k), chi_max_p, gtu_xx_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2iBfu_y = FiBfu_y_mainI(vector(Bfu_x_p, i + 2, j, k), vector(Bfu_y_p, i + 2, j, k), vector(phif_p, i + 2, j, k), vector(Betau_x_p, i + 2, j, k), vector(Betau_y_p, i + 2, j, k), vector(Alpha_p, i + 2, j, k), vector(TauM_xy, i + 2, j, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2iBfu_z = FiBfu_z_mainI(vector(Bfu_x_p, i + 2, j, k), vector(Bfu_z_p, i + 2, j, k), vector(phif_p, i + 2, j, k), vector(Betau_x_p, i + 2, j, k), vector(Betau_z_p, i + 2, j, k), vector(Alpha_p, i + 2, j, k), vector(TauM_xz, i + 2, j, k), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2iDf = FiDf_mainI(vector(Df_p, i + 2, j, k), vector(Betau_x_p, i + 2, j, k), vector(Alpha_p, i + 2, j, k), vector(TauN_x, i + 2, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2iDYf = FiDYf_mainI(vector(DYf_p, i + 2, j, k), vector(Betau_x_p, i + 2, j, k), vector(Alpha_p, i + 2, j, k), vector(TauNe_x, i + 2, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2itauf = Fitauf_mainI(vector(Df_p, i + 2, j, k), vector(tauf_p, i + 2, j, k), vector(Betau_x_p, i + 2, j, k), vector(Alpha_p, i + 2, j, k), vector(TauN_x, i + 2, j, k), sdetg_p, Sfu_x_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2iphif = Fiphif_mainI(vector(Bfu_x_p, i + 2, j, k), vector(phif_p, i + 2, j, k), vector(Betau_x_p, i + 2, j, k), vector(Alpha_p, i + 2, j, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i + 2, j, k) * (1.0 + vector(epsf, i + 2, j, k)) + vector(pf, i + 2, j, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i + 2, j, k) + sqca_p - vector(sqcs, i + 2, j, k) * sqca_p;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(Betau_x_p, i + 2, j, k)) + vector(Alpha_p, i + 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(Betau_x_p, i + 2, j, k)) + vector(Alpha_p, i + 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(Betau_x_p, i + 2, j, k)) + vector(Alpha_p, i + 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(Betau_x_p, i + 2, j, k)) + vector(Alpha_p, i + 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(Betau_x_p, i + 2, j, k)) + vector(Alpha_p, i + 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(Betau_x_p, i + 2, j, k)) + vector(Alpha_p, i + 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(Betau_x_p, i + 2, j, k)) + vector(Alpha_p, i + 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(Betau_x_p, i + 2, j, k)) + vector(Alpha_p, i + 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(Betau_x_p, i + 2, j, k)) + vector(Alpha_p, i + 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(Betau_x_p, i + 2, j, k)) + vector(Alpha_p, i + 2, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i + 3, j, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i + 3, j, k) * vector(Bfu_x_p, i + 3, j, k) + vector(gtd_yy_p, i + 3, j, k) * vector(Bfu_y_p, i + 3, j, k) + vector(gtd_yz_p, i + 3, j, k) * vector(Bfu_z_p, i + 3, j, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i + 3, j, k) * vector(Bfu_x_p, i + 3, j, k) + vector(gtd_yz_p, i + 3, j, k) * vector(Bfu_y_p, i + 3, j, k) + vector(gtd_zz_p, i + 3, j, k) * vector(Bfu_z_p, i + 3, j, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i + 3, j, k) * vector(Bfu_x_p, i + 3, j, k) + vector(gtd_xy_p, i + 3, j, k) * vector(Bfu_y_p, i + 3, j, k) + vector(gtd_xz_p, i + 3, j, k) * vector(Bfu_z_p, i + 3, j, k));
						sqBf_p = vector(Bfu_x_p, i + 3, j, k) * Bfd_x_p + vector(Bfu_y_p, i + 3, j, k) * Bfd_y_p + vector(Bfu_z_p, i + 3, j, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i + 3, j, k) * vector(gtd_zz_p, i + 3, j, k) * vector(gtd_xx_p, i + 3, j, k) + (-vector(gtd_yy_p, i + 3, j, k) * (vector(gtd_xz_p, i + 3, j, k) * vector(gtd_xz_p, i + 3, j, k))) + (-vector(gtd_zz_p, i + 3, j, k) * (vector(gtd_xy_p, i + 3, j, k) * vector(gtd_xy_p, i + 3, j, k))) + (-(vector(gtd_yz_p, i + 3, j, k) * vector(gtd_yz_p, i + 3, j, k)) * vector(gtd_xx_p, i + 3, j, k)) + 2.0 * vector(gtd_yz_p, i + 3, j, k) * vector(gtd_xy_p, i + 3, j, k) * vector(gtd_xz_p, i + 3, j, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i + 3, j, k) * vector(gtd_yy_p, i + 3, j, k) - vector(gtd_xy_p, i + 3, j, k) * vector(gtd_xy_p, i + 3, j, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i + 3, j, k) * vector(gtd_zz_p, i + 3, j, k) - vector(gtd_yz_p, i + 3, j, k) * vector(gtd_yz_p, i + 3, j, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i + 3, j, k) * vector(gtd_yz_p, i + 3, j, k) - vector(gtd_yy_p, i + 3, j, k) * vector(gtd_xz_p, i + 3, j, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i + 3, j, k) * vector(gtd_zz_p, i + 3, j, k)) + vector(gtd_yz_p, i + 3, j, k) * vector(gtd_xz_p, i + 3, j, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i + 3, j, k) + gtu_xy_p * vector(Sfd_y_p, i + 3, j, k) + gtu_xz_p * vector(Sfd_z_p, i + 3, j, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i + 3, j, k) + gtu_xy_p * vector(vfd_y, i + 3, j, k) + gtu_xz_p * vector(vfd_z, i + 3, j, k));
						Bfvf_p = vector(Bfu_x_p, i + 3, j, k) * vector(vfd_x, i + 3, j, k) + vector(Bfu_y_p, i + 3, j, k) * vector(vfd_y, i + 3, j, k) + vector(Bfu_z_p, i + 3, j, k) * vector(vfd_z, i + 3, j, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i + 3, j, k) * vector(gtd_zz_p, i + 3, j, k) - vector(gtd_xz_p, i + 3, j, k) * vector(gtd_xz_p, i + 3, j, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i + 3, j, k) * vector(gtd_yz_p, i + 3, j, k)) + vector(gtd_xy_p, i + 3, j, k) * vector(gtd_xz_p, i + 3, j, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i + 3, j, k) + gtu_yz_p * vector(vfd_y, i + 3, j, k) + gtu_zz_p * vector(vfd_z, i + 3, j, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i + 3, j, k) + gtu_yy_p * vector(Sfd_y_p, i + 3, j, k) + gtu_yz_p * vector(Sfd_z_p, i + 3, j, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i + 3, j, k) + gtu_yz_p * vector(Sfd_y_p, i + 3, j, k) + gtu_zz_p * vector(Sfd_z_p, i + 3, j, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i + 3, j, k) + gtu_yy_p * vector(vfd_y, i + 3, j, k) + gtu_yz_p * vector(vfd_z, i + 3, j, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i + 3, j, k) + vfu_y_p * vector(vfd_y, i + 3, j, k) + vfu_z_p * vector(vfd_z, i + 3, j, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i + 3, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i + 3, j, k) * vector(Bfu_z_p, i + 3, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i + 3, j, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i + 3, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i + 3, j, k) * vector(Bfu_x_p, i + 3, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i + 3, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i + 3, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i + 3, j, k) * vector(Bfu_z_p, i + 3, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i + 3, j, k) * vfu_z_p + vector(Bfu_z_p, i + 3, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i + 3, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i + 3, j, k) * vector(Bfu_y_p, i + 3, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i + 3, j, k) * vfu_y_p + vector(Bfu_y_p, i + 3, j, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i + 3, j, k) + Sfuu_xy_p * vector(gtd_xy_p, i + 3, j, k) + Sfuu_xz_p * vector(gtd_xz_p, i + 3, j, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i + 3, j, k) + Sfuu_xy_p * vector(gtd_yy_p, i + 3, j, k) + Sfuu_xz_p * vector(gtd_yz_p, i + 3, j, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i + 3, j, k) + Sfuu_xy_p * vector(gtd_yz_p, i + 3, j, k) + Sfuu_xz_p * vector(gtd_zz_p, i + 3, j, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i + 3, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i + 3, j, k) * vector(Bfu_z_p, i + 3, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i + 3, j, k) * vfu_z_p + vector(Bfu_z_p, i + 3, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i + 3, j, k) + Sfuu_yz_p * vector(gtd_yy_p, i + 3, j, k) + Sfuu_zz_p * vector(gtd_yz_p, i + 3, j, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i + 3, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i + 3, j, k) + Sfuu_zz_p * vector(gtd_zz_p, i + 3, j, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i + 3, j, k) + Sfuu_yz_p * vector(gtd_xy_p, i + 3, j, k) + Sfuu_zz_p * vector(gtd_xz_p, i + 3, j, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i + 3, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i + 3, j, k) * vector(Bfu_y_p, i + 3, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i + 3, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i + 3, j, k) + Sfuu_yy_p * vector(gtd_xy_p, i + 3, j, k) + Sfuu_yz_p * vector(gtd_xz_p, i + 3, j, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i + 3, j, k) + Sfuu_yy_p * vector(gtd_yy_p, i + 3, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i + 3, j, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i + 3, j, k) + Sfuu_yy_p * vector(gtd_yz_p, i + 3, j, k) + Sfuu_yz_p * vector(gtd_zz_p, i + 3, j, k));
						Fluxp3iSfd_x = FiSfd_x_mainI(vector(Sfd_x_p, i + 3, j, k), vector(gtd_xx_p, i + 3, j, k), vector(gtd_xy_p, i + 3, j, k), vector(gtd_xz_p, i + 3, j, k), vector(Betau_x_p, i + 3, j, k), vector(Alpha_p, i + 3, j, k), vector(TauT_xx, i + 3, j, k), vector(TauT_xy, i + 3, j, k), vector(TauT_xz, i + 3, j, k), inv_chi_p, sdetg_p, Sfud_xx_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3iSfd_y = FiSfd_y_mainI(vector(Sfd_y_p, i + 3, j, k), vector(gtd_xy_p, i + 3, j, k), vector(gtd_yy_p, i + 3, j, k), vector(gtd_yz_p, i + 3, j, k), vector(Betau_x_p, i + 3, j, k), vector(Alpha_p, i + 3, j, k), vector(TauT_xx, i + 3, j, k), vector(TauT_xy, i + 3, j, k), vector(TauT_xz, i + 3, j, k), inv_chi_p, sdetg_p, Sfud_xy_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3iSfd_z = FiSfd_z_mainI(vector(Sfd_z_p, i + 3, j, k), vector(gtd_xz_p, i + 3, j, k), vector(gtd_yz_p, i + 3, j, k), vector(gtd_zz_p, i + 3, j, k), vector(Betau_x_p, i + 3, j, k), vector(Alpha_p, i + 3, j, k), vector(TauT_xx, i + 3, j, k), vector(TauT_xy, i + 3, j, k), vector(TauT_xz, i + 3, j, k), inv_chi_p, sdetg_p, Sfud_xz_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3iBfu_x = FiBfu_x_mainI(vector(phif_p, i + 3, j, k), vector(Alpha_p, i + 3, j, k), chi_max_p, gtu_xx_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3iBfu_y = FiBfu_y_mainI(vector(Bfu_x_p, i + 3, j, k), vector(Bfu_y_p, i + 3, j, k), vector(phif_p, i + 3, j, k), vector(Betau_x_p, i + 3, j, k), vector(Betau_y_p, i + 3, j, k), vector(Alpha_p, i + 3, j, k), vector(TauM_xy, i + 3, j, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3iBfu_z = FiBfu_z_mainI(vector(Bfu_x_p, i + 3, j, k), vector(Bfu_z_p, i + 3, j, k), vector(phif_p, i + 3, j, k), vector(Betau_x_p, i + 3, j, k), vector(Betau_z_p, i + 3, j, k), vector(Alpha_p, i + 3, j, k), vector(TauM_xz, i + 3, j, k), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3iDf = FiDf_mainI(vector(Df_p, i + 3, j, k), vector(Betau_x_p, i + 3, j, k), vector(Alpha_p, i + 3, j, k), vector(TauN_x, i + 3, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3iDYf = FiDYf_mainI(vector(DYf_p, i + 3, j, k), vector(Betau_x_p, i + 3, j, k), vector(Alpha_p, i + 3, j, k), vector(TauNe_x, i + 3, j, k), sdetg_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3itauf = Fitauf_mainI(vector(Df_p, i + 3, j, k), vector(tauf_p, i + 3, j, k), vector(Betau_x_p, i + 3, j, k), vector(Alpha_p, i + 3, j, k), vector(TauN_x, i + 3, j, k), sdetg_p, Sfu_x_p, vfu_x_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3iphif = Fiphif_mainI(vector(Bfu_x_p, i + 3, j, k), vector(phif_p, i + 3, j, k), vector(Betau_x_p, i + 3, j, k), vector(Alpha_p, i + 3, j, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i + 3, j, k) * (1.0 + vector(epsf, i + 3, j, k)) + vector(pf, i + 3, j, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i + 3, j, k) + sqca_p - vector(sqcs, i + 3, j, k) * sqca_p;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(Betau_x_p, i + 3, j, k)) + vector(Alpha_p, i + 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(Betau_x_p, i + 3, j, k)) + vector(Alpha_p, i + 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(Betau_x_p, i + 3, j, k)) + vector(Alpha_p, i + 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(Betau_x_p, i + 3, j, k)) + vector(Alpha_p, i + 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(Betau_x_p, i + 3, j, k)) + vector(Alpha_p, i + 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(Betau_x_p, i + 3, j, k)) + vector(Alpha_p, i + 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(Betau_x_p, i + 3, j, k)) + vector(Alpha_p, i + 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(Betau_x_p, i + 3, j, k)) + vector(Alpha_p, i + 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(Betau_x_p, i + 3, j, k)) + vector(Alpha_p, i + 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(Betau_x_p, i + 3, j, k)) + vector(Alpha_p, i + 3, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_x_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_xx_p - (1.0 - sqa_p) * (vfu_x_p * vfu_x_p)))))));
						LLFpiSfd_x = 0.5 * (FluxiSfd_x + fabs(SpeediSfd_x) * vector(Sfd_x_p, i, j, k));
						LLFpiSfd_y = 0.5 * (FluxiSfd_y + fabs(SpeediSfd_y) * vector(Sfd_y_p, i, j, k));
						LLFpiSfd_z = 0.5 * (FluxiSfd_z + fabs(SpeediSfd_z) * vector(Sfd_z_p, i, j, k));
						LLFpiBfu_x = 0.5 * (FluxiBfu_x + fabs(SpeediBfu_x) * vector(Bfu_x_p, i, j, k));
						LLFpiBfu_y = 0.5 * (FluxiBfu_y + fabs(SpeediBfu_y) * vector(Bfu_y_p, i, j, k));
						LLFpiBfu_z = 0.5 * (FluxiBfu_z + fabs(SpeediBfu_z) * vector(Bfu_z_p, i, j, k));
						LLFpiDf = 0.5 * (FluxiDf + fabs(SpeediDf) * vector(Df_p, i, j, k));
						LLFpiDYf = 0.5 * (FluxiDYf + fabs(SpeediDYf) * vector(DYf_p, i, j, k));
						LLFpitauf = 0.5 * (Fluxitauf + fabs(Speeditauf) * vector(tauf_p, i, j, k));
						LLFpiphif = 0.5 * (Fluxiphif + fabs(Speediphif) * vector(phif_p, i, j, k));
						LLFniSfd_x = 0.5 * (FluxiSfd_x - fabs(SpeediSfd_x) * vector(Sfd_x_p, i, j, k));
						LLFniSfd_y = 0.5 * (FluxiSfd_y - fabs(SpeediSfd_y) * vector(Sfd_y_p, i, j, k));
						LLFniSfd_z = 0.5 * (FluxiSfd_z - fabs(SpeediSfd_z) * vector(Sfd_z_p, i, j, k));
						LLFniBfu_x = 0.5 * (FluxiBfu_x - fabs(SpeediBfu_x) * vector(Bfu_x_p, i, j, k));
						LLFniBfu_y = 0.5 * (FluxiBfu_y - fabs(SpeediBfu_y) * vector(Bfu_y_p, i, j, k));
						LLFniBfu_z = 0.5 * (FluxiBfu_z - fabs(SpeediBfu_z) * vector(Bfu_z_p, i, j, k));
						LLFniDf = 0.5 * (FluxiDf - fabs(SpeediDf) * vector(Df_p, i, j, k));
						LLFniDYf = 0.5 * (FluxiDYf - fabs(SpeediDYf) * vector(DYf_p, i, j, k));
						LLFnitauf = 0.5 * (Fluxitauf - fabs(Speeditauf) * vector(tauf_p, i, j, k));
						LLFniphif = 0.5 * (Fluxiphif - fabs(Speediphif) * vector(phif_p, i, j, k));
						LLFp_m3iSfd_x = 0.5 * (Fluxm3iSfd_x + fabs(SpeediSfd_x) * vector(Sfd_x_p, i - 3, j, k));
						LLFp_m3iSfd_y = 0.5 * (Fluxm3iSfd_y + fabs(SpeediSfd_y) * vector(Sfd_y_p, i - 3, j, k));
						LLFp_m3iSfd_z = 0.5 * (Fluxm3iSfd_z + fabs(SpeediSfd_z) * vector(Sfd_z_p, i - 3, j, k));
						LLFp_m3iBfu_x = 0.5 * (Fluxm3iBfu_x + fabs(SpeediBfu_x) * vector(Bfu_x_p, i - 3, j, k));
						LLFp_m3iBfu_y = 0.5 * (Fluxm3iBfu_y + fabs(SpeediBfu_y) * vector(Bfu_y_p, i - 3, j, k));
						LLFp_m3iBfu_z = 0.5 * (Fluxm3iBfu_z + fabs(SpeediBfu_z) * vector(Bfu_z_p, i - 3, j, k));
						LLFp_m3iDf = 0.5 * (Fluxm3iDf + fabs(SpeediDf) * vector(Df_p, i - 3, j, k));
						LLFp_m3iDYf = 0.5 * (Fluxm3iDYf + fabs(SpeediDYf) * vector(DYf_p, i - 3, j, k));
						LLFp_m3itauf = 0.5 * (Fluxm3itauf + fabs(Speeditauf) * vector(tauf_p, i - 3, j, k));
						LLFp_m3iphif = 0.5 * (Fluxm3iphif + fabs(Speediphif) * vector(phif_p, i - 3, j, k));
						LLFn_m3iSfd_x = 0.5 * (Fluxm3iSfd_x - fabs(SpeediSfd_x) * vector(Sfd_x_p, i - 3, j, k));
						LLFn_m3iSfd_y = 0.5 * (Fluxm3iSfd_y - fabs(SpeediSfd_y) * vector(Sfd_y_p, i - 3, j, k));
						LLFn_m3iSfd_z = 0.5 * (Fluxm3iSfd_z - fabs(SpeediSfd_z) * vector(Sfd_z_p, i - 3, j, k));
						LLFn_m3iBfu_x = 0.5 * (Fluxm3iBfu_x - fabs(SpeediBfu_x) * vector(Bfu_x_p, i - 3, j, k));
						LLFn_m3iBfu_y = 0.5 * (Fluxm3iBfu_y - fabs(SpeediBfu_y) * vector(Bfu_y_p, i - 3, j, k));
						LLFn_m3iBfu_z = 0.5 * (Fluxm3iBfu_z - fabs(SpeediBfu_z) * vector(Bfu_z_p, i - 3, j, k));
						LLFn_m3iDf = 0.5 * (Fluxm3iDf - fabs(SpeediDf) * vector(Df_p, i - 3, j, k));
						LLFn_m3iDYf = 0.5 * (Fluxm3iDYf - fabs(SpeediDYf) * vector(DYf_p, i - 3, j, k));
						LLFn_m3itauf = 0.5 * (Fluxm3itauf - fabs(Speeditauf) * vector(tauf_p, i - 3, j, k));
						LLFn_m3iphif = 0.5 * (Fluxm3iphif - fabs(Speediphif) * vector(phif_p, i - 3, j, k));
						LLFp_m2iSfd_x = 0.5 * (Fluxm2iSfd_x + fabs(SpeediSfd_x) * vector(Sfd_x_p, i - 2, j, k));
						LLFp_m2iSfd_y = 0.5 * (Fluxm2iSfd_y + fabs(SpeediSfd_y) * vector(Sfd_y_p, i - 2, j, k));
						LLFp_m2iSfd_z = 0.5 * (Fluxm2iSfd_z + fabs(SpeediSfd_z) * vector(Sfd_z_p, i - 2, j, k));
						LLFp_m2iBfu_x = 0.5 * (Fluxm2iBfu_x + fabs(SpeediBfu_x) * vector(Bfu_x_p, i - 2, j, k));
						LLFp_m2iBfu_y = 0.5 * (Fluxm2iBfu_y + fabs(SpeediBfu_y) * vector(Bfu_y_p, i - 2, j, k));
						LLFp_m2iBfu_z = 0.5 * (Fluxm2iBfu_z + fabs(SpeediBfu_z) * vector(Bfu_z_p, i - 2, j, k));
						LLFp_m2iDf = 0.5 * (Fluxm2iDf + fabs(SpeediDf) * vector(Df_p, i - 2, j, k));
						LLFp_m2iDYf = 0.5 * (Fluxm2iDYf + fabs(SpeediDYf) * vector(DYf_p, i - 2, j, k));
						LLFp_m2itauf = 0.5 * (Fluxm2itauf + fabs(Speeditauf) * vector(tauf_p, i - 2, j, k));
						LLFp_m2iphif = 0.5 * (Fluxm2iphif + fabs(Speediphif) * vector(phif_p, i - 2, j, k));
						LLFn_m2iSfd_x = 0.5 * (Fluxm2iSfd_x - fabs(SpeediSfd_x) * vector(Sfd_x_p, i - 2, j, k));
						LLFn_m2iSfd_y = 0.5 * (Fluxm2iSfd_y - fabs(SpeediSfd_y) * vector(Sfd_y_p, i - 2, j, k));
						LLFn_m2iSfd_z = 0.5 * (Fluxm2iSfd_z - fabs(SpeediSfd_z) * vector(Sfd_z_p, i - 2, j, k));
						LLFn_m2iBfu_x = 0.5 * (Fluxm2iBfu_x - fabs(SpeediBfu_x) * vector(Bfu_x_p, i - 2, j, k));
						LLFn_m2iBfu_y = 0.5 * (Fluxm2iBfu_y - fabs(SpeediBfu_y) * vector(Bfu_y_p, i - 2, j, k));
						LLFn_m2iBfu_z = 0.5 * (Fluxm2iBfu_z - fabs(SpeediBfu_z) * vector(Bfu_z_p, i - 2, j, k));
						LLFn_m2iDf = 0.5 * (Fluxm2iDf - fabs(SpeediDf) * vector(Df_p, i - 2, j, k));
						LLFn_m2iDYf = 0.5 * (Fluxm2iDYf - fabs(SpeediDYf) * vector(DYf_p, i - 2, j, k));
						LLFn_m2itauf = 0.5 * (Fluxm2itauf - fabs(Speeditauf) * vector(tauf_p, i - 2, j, k));
						LLFn_m2iphif = 0.5 * (Fluxm2iphif - fabs(Speediphif) * vector(phif_p, i - 2, j, k));
						LLFp_m1iSfd_x = 0.5 * (Fluxm1iSfd_x + fabs(SpeediSfd_x) * vector(Sfd_x_p, i - 1, j, k));
						LLFp_m1iSfd_y = 0.5 * (Fluxm1iSfd_y + fabs(SpeediSfd_y) * vector(Sfd_y_p, i - 1, j, k));
						LLFp_m1iSfd_z = 0.5 * (Fluxm1iSfd_z + fabs(SpeediSfd_z) * vector(Sfd_z_p, i - 1, j, k));
						LLFp_m1iBfu_x = 0.5 * (Fluxm1iBfu_x + fabs(SpeediBfu_x) * vector(Bfu_x_p, i - 1, j, k));
						LLFp_m1iBfu_y = 0.5 * (Fluxm1iBfu_y + fabs(SpeediBfu_y) * vector(Bfu_y_p, i - 1, j, k));
						LLFp_m1iBfu_z = 0.5 * (Fluxm1iBfu_z + fabs(SpeediBfu_z) * vector(Bfu_z_p, i - 1, j, k));
						LLFp_m1iDf = 0.5 * (Fluxm1iDf + fabs(SpeediDf) * vector(Df_p, i - 1, j, k));
						LLFp_m1iDYf = 0.5 * (Fluxm1iDYf + fabs(SpeediDYf) * vector(DYf_p, i - 1, j, k));
						LLFp_m1itauf = 0.5 * (Fluxm1itauf + fabs(Speeditauf) * vector(tauf_p, i - 1, j, k));
						LLFp_m1iphif = 0.5 * (Fluxm1iphif + fabs(Speediphif) * vector(phif_p, i - 1, j, k));
						LLFn_m1iSfd_x = 0.5 * (Fluxm1iSfd_x - fabs(SpeediSfd_x) * vector(Sfd_x_p, i - 1, j, k));
						LLFn_m1iSfd_y = 0.5 * (Fluxm1iSfd_y - fabs(SpeediSfd_y) * vector(Sfd_y_p, i - 1, j, k));
						LLFn_m1iSfd_z = 0.5 * (Fluxm1iSfd_z - fabs(SpeediSfd_z) * vector(Sfd_z_p, i - 1, j, k));
						LLFn_m1iBfu_x = 0.5 * (Fluxm1iBfu_x - fabs(SpeediBfu_x) * vector(Bfu_x_p, i - 1, j, k));
						LLFn_m1iBfu_y = 0.5 * (Fluxm1iBfu_y - fabs(SpeediBfu_y) * vector(Bfu_y_p, i - 1, j, k));
						LLFn_m1iBfu_z = 0.5 * (Fluxm1iBfu_z - fabs(SpeediBfu_z) * vector(Bfu_z_p, i - 1, j, k));
						LLFn_m1iDf = 0.5 * (Fluxm1iDf - fabs(SpeediDf) * vector(Df_p, i - 1, j, k));
						LLFn_m1iDYf = 0.5 * (Fluxm1iDYf - fabs(SpeediDYf) * vector(DYf_p, i - 1, j, k));
						LLFn_m1itauf = 0.5 * (Fluxm1itauf - fabs(Speeditauf) * vector(tauf_p, i - 1, j, k));
						LLFn_m1iphif = 0.5 * (Fluxm1iphif - fabs(Speediphif) * vector(phif_p, i - 1, j, k));
						LLFp_p1iSfd_x = 0.5 * (Fluxp1iSfd_x + fabs(SpeediSfd_x) * vector(Sfd_x_p, i + 1, j, k));
						LLFp_p1iSfd_y = 0.5 * (Fluxp1iSfd_y + fabs(SpeediSfd_y) * vector(Sfd_y_p, i + 1, j, k));
						LLFp_p1iSfd_z = 0.5 * (Fluxp1iSfd_z + fabs(SpeediSfd_z) * vector(Sfd_z_p, i + 1, j, k));
						LLFp_p1iBfu_x = 0.5 * (Fluxp1iBfu_x + fabs(SpeediBfu_x) * vector(Bfu_x_p, i + 1, j, k));
						LLFp_p1iBfu_y = 0.5 * (Fluxp1iBfu_y + fabs(SpeediBfu_y) * vector(Bfu_y_p, i + 1, j, k));
						LLFp_p1iBfu_z = 0.5 * (Fluxp1iBfu_z + fabs(SpeediBfu_z) * vector(Bfu_z_p, i + 1, j, k));
						LLFp_p1iDf = 0.5 * (Fluxp1iDf + fabs(SpeediDf) * vector(Df_p, i + 1, j, k));
						LLFp_p1iDYf = 0.5 * (Fluxp1iDYf + fabs(SpeediDYf) * vector(DYf_p, i + 1, j, k));
						LLFp_p1itauf = 0.5 * (Fluxp1itauf + fabs(Speeditauf) * vector(tauf_p, i + 1, j, k));
						LLFp_p1iphif = 0.5 * (Fluxp1iphif + fabs(Speediphif) * vector(phif_p, i + 1, j, k));
						LLFn_p1iSfd_x = 0.5 * (Fluxp1iSfd_x - fabs(SpeediSfd_x) * vector(Sfd_x_p, i + 1, j, k));
						LLFn_p1iSfd_y = 0.5 * (Fluxp1iSfd_y - fabs(SpeediSfd_y) * vector(Sfd_y_p, i + 1, j, k));
						LLFn_p1iSfd_z = 0.5 * (Fluxp1iSfd_z - fabs(SpeediSfd_z) * vector(Sfd_z_p, i + 1, j, k));
						LLFn_p1iBfu_x = 0.5 * (Fluxp1iBfu_x - fabs(SpeediBfu_x) * vector(Bfu_x_p, i + 1, j, k));
						LLFn_p1iBfu_y = 0.5 * (Fluxp1iBfu_y - fabs(SpeediBfu_y) * vector(Bfu_y_p, i + 1, j, k));
						LLFn_p1iBfu_z = 0.5 * (Fluxp1iBfu_z - fabs(SpeediBfu_z) * vector(Bfu_z_p, i + 1, j, k));
						LLFn_p1iDf = 0.5 * (Fluxp1iDf - fabs(SpeediDf) * vector(Df_p, i + 1, j, k));
						LLFn_p1iDYf = 0.5 * (Fluxp1iDYf - fabs(SpeediDYf) * vector(DYf_p, i + 1, j, k));
						LLFn_p1itauf = 0.5 * (Fluxp1itauf - fabs(Speeditauf) * vector(tauf_p, i + 1, j, k));
						LLFn_p1iphif = 0.5 * (Fluxp1iphif - fabs(Speediphif) * vector(phif_p, i + 1, j, k));
						LLFp_p2iSfd_x = 0.5 * (Fluxp2iSfd_x + fabs(SpeediSfd_x) * vector(Sfd_x_p, i + 2, j, k));
						LLFp_p2iSfd_y = 0.5 * (Fluxp2iSfd_y + fabs(SpeediSfd_y) * vector(Sfd_y_p, i + 2, j, k));
						LLFp_p2iSfd_z = 0.5 * (Fluxp2iSfd_z + fabs(SpeediSfd_z) * vector(Sfd_z_p, i + 2, j, k));
						LLFp_p2iBfu_x = 0.5 * (Fluxp2iBfu_x + fabs(SpeediBfu_x) * vector(Bfu_x_p, i + 2, j, k));
						LLFp_p2iBfu_y = 0.5 * (Fluxp2iBfu_y + fabs(SpeediBfu_y) * vector(Bfu_y_p, i + 2, j, k));
						LLFp_p2iBfu_z = 0.5 * (Fluxp2iBfu_z + fabs(SpeediBfu_z) * vector(Bfu_z_p, i + 2, j, k));
						LLFp_p2iDf = 0.5 * (Fluxp2iDf + fabs(SpeediDf) * vector(Df_p, i + 2, j, k));
						LLFp_p2iDYf = 0.5 * (Fluxp2iDYf + fabs(SpeediDYf) * vector(DYf_p, i + 2, j, k));
						LLFp_p2itauf = 0.5 * (Fluxp2itauf + fabs(Speeditauf) * vector(tauf_p, i + 2, j, k));
						LLFp_p2iphif = 0.5 * (Fluxp2iphif + fabs(Speediphif) * vector(phif_p, i + 2, j, k));
						LLFn_p2iSfd_x = 0.5 * (Fluxp2iSfd_x - fabs(SpeediSfd_x) * vector(Sfd_x_p, i + 2, j, k));
						LLFn_p2iSfd_y = 0.5 * (Fluxp2iSfd_y - fabs(SpeediSfd_y) * vector(Sfd_y_p, i + 2, j, k));
						LLFn_p2iSfd_z = 0.5 * (Fluxp2iSfd_z - fabs(SpeediSfd_z) * vector(Sfd_z_p, i + 2, j, k));
						LLFn_p2iBfu_x = 0.5 * (Fluxp2iBfu_x - fabs(SpeediBfu_x) * vector(Bfu_x_p, i + 2, j, k));
						LLFn_p2iBfu_y = 0.5 * (Fluxp2iBfu_y - fabs(SpeediBfu_y) * vector(Bfu_y_p, i + 2, j, k));
						LLFn_p2iBfu_z = 0.5 * (Fluxp2iBfu_z - fabs(SpeediBfu_z) * vector(Bfu_z_p, i + 2, j, k));
						LLFn_p2iDf = 0.5 * (Fluxp2iDf - fabs(SpeediDf) * vector(Df_p, i + 2, j, k));
						LLFn_p2iDYf = 0.5 * (Fluxp2iDYf - fabs(SpeediDYf) * vector(DYf_p, i + 2, j, k));
						LLFn_p2itauf = 0.5 * (Fluxp2itauf - fabs(Speeditauf) * vector(tauf_p, i + 2, j, k));
						LLFn_p2iphif = 0.5 * (Fluxp2iphif - fabs(Speediphif) * vector(phif_p, i + 2, j, k));
						LLFp_p3iSfd_x = 0.5 * (Fluxp3iSfd_x + fabs(SpeediSfd_x) * vector(Sfd_x_p, i + 3, j, k));
						LLFp_p3iSfd_y = 0.5 * (Fluxp3iSfd_y + fabs(SpeediSfd_y) * vector(Sfd_y_p, i + 3, j, k));
						LLFp_p3iSfd_z = 0.5 * (Fluxp3iSfd_z + fabs(SpeediSfd_z) * vector(Sfd_z_p, i + 3, j, k));
						LLFp_p3iBfu_x = 0.5 * (Fluxp3iBfu_x + fabs(SpeediBfu_x) * vector(Bfu_x_p, i + 3, j, k));
						LLFp_p3iBfu_y = 0.5 * (Fluxp3iBfu_y + fabs(SpeediBfu_y) * vector(Bfu_y_p, i + 3, j, k));
						LLFp_p3iBfu_z = 0.5 * (Fluxp3iBfu_z + fabs(SpeediBfu_z) * vector(Bfu_z_p, i + 3, j, k));
						LLFp_p3iDf = 0.5 * (Fluxp3iDf + fabs(SpeediDf) * vector(Df_p, i + 3, j, k));
						LLFp_p3iDYf = 0.5 * (Fluxp3iDYf + fabs(SpeediDYf) * vector(DYf_p, i + 3, j, k));
						LLFp_p3itauf = 0.5 * (Fluxp3itauf + fabs(Speeditauf) * vector(tauf_p, i + 3, j, k));
						LLFp_p3iphif = 0.5 * (Fluxp3iphif + fabs(Speediphif) * vector(phif_p, i + 3, j, k));
						LLFn_p3iSfd_x = 0.5 * (Fluxp3iSfd_x - fabs(SpeediSfd_x) * vector(Sfd_x_p, i + 3, j, k));
						LLFn_p3iSfd_y = 0.5 * (Fluxp3iSfd_y - fabs(SpeediSfd_y) * vector(Sfd_y_p, i + 3, j, k));
						LLFn_p3iSfd_z = 0.5 * (Fluxp3iSfd_z - fabs(SpeediSfd_z) * vector(Sfd_z_p, i + 3, j, k));
						LLFn_p3iBfu_x = 0.5 * (Fluxp3iBfu_x - fabs(SpeediBfu_x) * vector(Bfu_x_p, i + 3, j, k));
						LLFn_p3iBfu_y = 0.5 * (Fluxp3iBfu_y - fabs(SpeediBfu_y) * vector(Bfu_y_p, i + 3, j, k));
						LLFn_p3iBfu_z = 0.5 * (Fluxp3iBfu_z - fabs(SpeediBfu_z) * vector(Bfu_z_p, i + 3, j, k));
						LLFn_p3iDf = 0.5 * (Fluxp3iDf - fabs(SpeediDf) * vector(Df_p, i + 3, j, k));
						LLFn_p3iDYf = 0.5 * (Fluxp3iDYf - fabs(SpeediDYf) * vector(DYf_p, i + 3, j, k));
						LLFn_p3itauf = 0.5 * (Fluxp3itauf - fabs(Speeditauf) * vector(tauf_p, i + 3, j, k));
						LLFn_p3iphif = 0.5 * (Fluxp3iphif - fabs(Speediphif) * vector(phif_p, i + 3, j, k));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j, k) * vector(Bfu_x_p, i, j, k) + vector(gtd_yy_p, i, j, k) * vector(Bfu_y_p, i, j, k) + vector(gtd_yz_p, i, j, k) * vector(Bfu_z_p, i, j, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j, k) * vector(Bfu_x_p, i, j, k) + vector(gtd_yz_p, i, j, k) * vector(Bfu_y_p, i, j, k) + vector(gtd_zz_p, i, j, k) * vector(Bfu_z_p, i, j, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j, k) * vector(Bfu_x_p, i, j, k) + vector(gtd_xy_p, i, j, k) * vector(Bfu_y_p, i, j, k) + vector(gtd_xz_p, i, j, k) * vector(Bfu_z_p, i, j, k));
						sqBf_p = vector(Bfu_x_p, i, j, k) * Bfd_x_p + vector(Bfu_y_p, i, j, k) * Bfd_y_p + vector(Bfu_z_p, i, j, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j, k) * vector(gtd_zz_p, i, j, k) * vector(gtd_xx_p, i, j, k) + (-vector(gtd_yy_p, i, j, k) * (vector(gtd_xz_p, i, j, k) * vector(gtd_xz_p, i, j, k))) + (-vector(gtd_zz_p, i, j, k) * (vector(gtd_xy_p, i, j, k) * vector(gtd_xy_p, i, j, k))) + (-(vector(gtd_yz_p, i, j, k) * vector(gtd_yz_p, i, j, k)) * vector(gtd_xx_p, i, j, k)) + 2.0 * vector(gtd_yz_p, i, j, k) * vector(gtd_xy_p, i, j, k) * vector(gtd_xz_p, i, j, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j, k) * vector(gtd_yy_p, i, j, k) - vector(gtd_xy_p, i, j, k) * vector(gtd_xy_p, i, j, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j, k) * vector(gtd_zz_p, i, j, k) - vector(gtd_yz_p, i, j, k) * vector(gtd_yz_p, i, j, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j, k) * vector(gtd_yz_p, i, j, k) - vector(gtd_yy_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j, k) * vector(gtd_zz_p, i, j, k)) + vector(gtd_yz_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j, k) + gtu_xy_p * vector(Sfd_y_p, i, j, k) + gtu_xz_p * vector(Sfd_z_p, i, j, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j, k) + gtu_xy_p * vector(vfd_y, i, j, k) + gtu_xz_p * vector(vfd_z, i, j, k));
						Bfvf_p = vector(Bfu_x_p, i, j, k) * vector(vfd_x, i, j, k) + vector(Bfu_y_p, i, j, k) * vector(vfd_y, i, j, k) + vector(Bfu_z_p, i, j, k) * vector(vfd_z, i, j, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j, k) * vector(gtd_zz_p, i, j, k) - vector(gtd_xz_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j, k) * vector(gtd_yz_p, i, j, k)) + vector(gtd_xy_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j, k) + gtu_yz_p * vector(vfd_y, i, j, k) + gtu_zz_p * vector(vfd_z, i, j, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j, k) + gtu_yy_p * vector(Sfd_y_p, i, j, k) + gtu_yz_p * vector(Sfd_z_p, i, j, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j, k) + gtu_yz_p * vector(Sfd_y_p, i, j, k) + gtu_zz_p * vector(Sfd_z_p, i, j, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j, k) + gtu_yy_p * vector(vfd_y, i, j, k) + gtu_yz_p * vector(vfd_z, i, j, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j, k) + vfu_y_p * vector(vfd_y, i, j, k) + vfu_z_p * vector(vfd_z, i, j, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j, k) * vector(Bfu_z_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j, k) * vector(Bfu_x_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k) * vector(Bfu_z_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k) * vfu_z_p + vector(Bfu_z_p, i, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k) * vector(Bfu_y_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k) * vfu_y_p + vector(Bfu_y_p, i, j, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j, k) + Sfuu_xy_p * vector(gtd_xy_p, i, j, k) + Sfuu_xz_p * vector(gtd_xz_p, i, j, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j, k) + Sfuu_xy_p * vector(gtd_yy_p, i, j, k) + Sfuu_xz_p * vector(gtd_yz_p, i, j, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j, k) + Sfuu_xy_p * vector(gtd_yz_p, i, j, k) + Sfuu_xz_p * vector(gtd_zz_p, i, j, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k) * vector(Bfu_z_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k) * vfu_z_p + vector(Bfu_z_p, i, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j, k) + Sfuu_yz_p * vector(gtd_yy_p, i, j, k) + Sfuu_zz_p * vector(gtd_yz_p, i, j, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k) + Sfuu_zz_p * vector(gtd_zz_p, i, j, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j, k) + Sfuu_yz_p * vector(gtd_xy_p, i, j, k) + Sfuu_zz_p * vector(gtd_xz_p, i, j, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j, k) * vector(Bfu_y_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j, k) + Sfuu_yy_p * vector(gtd_xy_p, i, j, k) + Sfuu_yz_p * vector(gtd_xz_p, i, j, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j, k) + Sfuu_yy_p * vector(gtd_yy_p, i, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j, k) + Sfuu_yy_p * vector(gtd_yz_p, i, j, k) + Sfuu_yz_p * vector(gtd_zz_p, i, j, k));
						FluxjSfd_x = FjSfd_x_mainI(vector(Sfd_x_p, i, j, k), vector(gtd_xx_p, i, j, k), vector(gtd_xy_p, i, j, k), vector(gtd_xz_p, i, j, k), vector(Betau_y_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_yy, i, j, k), vector(TauT_yz, i, j, k), inv_chi_p, sdetg_p, Sfud_yx_p, dx, simPlat_dt, ilast, jlast);
						FluxjSfd_y = FjSfd_y_mainI(vector(Sfd_y_p, i, j, k), vector(gtd_xy_p, i, j, k), vector(gtd_yy_p, i, j, k), vector(gtd_yz_p, i, j, k), vector(Betau_y_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_yy, i, j, k), vector(TauT_yz, i, j, k), inv_chi_p, sdetg_p, Sfud_yy_p, dx, simPlat_dt, ilast, jlast);
						FluxjSfd_z = FjSfd_z_mainI(vector(Sfd_z_p, i, j, k), vector(gtd_xz_p, i, j, k), vector(gtd_yz_p, i, j, k), vector(gtd_zz_p, i, j, k), vector(Betau_y_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_yy, i, j, k), vector(TauT_yz, i, j, k), inv_chi_p, sdetg_p, Sfud_yz_p, dx, simPlat_dt, ilast, jlast);
						FluxjBfu_x = FjBfu_x_mainI(vector(Bfu_x_p, i, j, k), vector(Bfu_y_p, i, j, k), vector(phif_p, i, j, k), vector(Betau_x_p, i, j, k), vector(Betau_y_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauM_xy, i, j, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						FluxjBfu_y = FjBfu_y_mainI(vector(phif_p, i, j, k), vector(Alpha_p, i, j, k), chi_max_p, gtu_yy_p, dx, simPlat_dt, ilast, jlast);
						FluxjBfu_z = FjBfu_z_mainI(vector(Bfu_y_p, i, j, k), vector(Bfu_z_p, i, j, k), vector(phif_p, i, j, k), vector(Betau_y_p, i, j, k), vector(Betau_z_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauM_yz, i, j, k), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						FluxjDf = FjDf_mainI(vector(Df_p, i, j, k), vector(Betau_y_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauN_y, i, j, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						FluxjDYf = FjDYf_mainI(vector(DYf_p, i, j, k), vector(Betau_y_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauNe_y, i, j, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxjtauf = Fjtauf_mainI(vector(Df_p, i, j, k), vector(tauf_p, i, j, k), vector(Betau_y_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauN_y, i, j, k), sdetg_p, Sfu_y_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxjphif = Fjphif_mainI(vector(Bfu_y_p, i, j, k), vector(phif_p, i, j, k), vector(Betau_y_p, i, j, k), vector(Alpha_p, i, j, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j, k) + sqca_p - vector(sqcs, i, j, k) * sqca_p;
						SpeedjSfd_x = fabs(fabs(vector(Betau_y_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p))))));
						SpeedjSfd_y = fabs(fabs(vector(Betau_y_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p))))));
						SpeedjSfd_z = fabs(fabs(vector(Betau_y_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p))))));
						SpeedjBfu_x = fabs(fabs(vector(Betau_y_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p))))));
						SpeedjBfu_y = fabs(fabs(vector(Betau_y_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p))))));
						SpeedjBfu_z = fabs(fabs(vector(Betau_y_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p))))));
						SpeedjDf = fabs(fabs(vector(Betau_y_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p))))));
						SpeedjDYf = fabs(fabs(vector(Betau_y_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p))))));
						Speedjtauf = fabs(fabs(vector(Betau_y_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p))))));
						Speedjphif = fabs(fabs(vector(Betau_y_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j - 3, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j - 3, k) * vector(Bfu_x_p, i, j - 3, k) + vector(gtd_yy_p, i, j - 3, k) * vector(Bfu_y_p, i, j - 3, k) + vector(gtd_yz_p, i, j - 3, k) * vector(Bfu_z_p, i, j - 3, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j - 3, k) * vector(Bfu_x_p, i, j - 3, k) + vector(gtd_yz_p, i, j - 3, k) * vector(Bfu_y_p, i, j - 3, k) + vector(gtd_zz_p, i, j - 3, k) * vector(Bfu_z_p, i, j - 3, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j - 3, k) * vector(Bfu_x_p, i, j - 3, k) + vector(gtd_xy_p, i, j - 3, k) * vector(Bfu_y_p, i, j - 3, k) + vector(gtd_xz_p, i, j - 3, k) * vector(Bfu_z_p, i, j - 3, k));
						sqBf_p = vector(Bfu_x_p, i, j - 3, k) * Bfd_x_p + vector(Bfu_y_p, i, j - 3, k) * Bfd_y_p + vector(Bfu_z_p, i, j - 3, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j - 3, k) * vector(gtd_zz_p, i, j - 3, k) * vector(gtd_xx_p, i, j - 3, k) + (-vector(gtd_yy_p, i, j - 3, k) * (vector(gtd_xz_p, i, j - 3, k) * vector(gtd_xz_p, i, j - 3, k))) + (-vector(gtd_zz_p, i, j - 3, k) * (vector(gtd_xy_p, i, j - 3, k) * vector(gtd_xy_p, i, j - 3, k))) + (-(vector(gtd_yz_p, i, j - 3, k) * vector(gtd_yz_p, i, j - 3, k)) * vector(gtd_xx_p, i, j - 3, k)) + 2.0 * vector(gtd_yz_p, i, j - 3, k) * vector(gtd_xy_p, i, j - 3, k) * vector(gtd_xz_p, i, j - 3, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j - 3, k) * vector(gtd_yy_p, i, j - 3, k) - vector(gtd_xy_p, i, j - 3, k) * vector(gtd_xy_p, i, j - 3, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j - 3, k) * vector(gtd_zz_p, i, j - 3, k) - vector(gtd_yz_p, i, j - 3, k) * vector(gtd_yz_p, i, j - 3, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j - 3, k) * vector(gtd_yz_p, i, j - 3, k) - vector(gtd_yy_p, i, j - 3, k) * vector(gtd_xz_p, i, j - 3, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j - 3, k) * vector(gtd_zz_p, i, j - 3, k)) + vector(gtd_yz_p, i, j - 3, k) * vector(gtd_xz_p, i, j - 3, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j - 3, k) + gtu_xy_p * vector(Sfd_y_p, i, j - 3, k) + gtu_xz_p * vector(Sfd_z_p, i, j - 3, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j - 3, k) + gtu_xy_p * vector(vfd_y, i, j - 3, k) + gtu_xz_p * vector(vfd_z, i, j - 3, k));
						Bfvf_p = vector(Bfu_x_p, i, j - 3, k) * vector(vfd_x, i, j - 3, k) + vector(Bfu_y_p, i, j - 3, k) * vector(vfd_y, i, j - 3, k) + vector(Bfu_z_p, i, j - 3, k) * vector(vfd_z, i, j - 3, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j - 3, k) * vector(gtd_zz_p, i, j - 3, k) - vector(gtd_xz_p, i, j - 3, k) * vector(gtd_xz_p, i, j - 3, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j - 3, k) * vector(gtd_yz_p, i, j - 3, k)) + vector(gtd_xy_p, i, j - 3, k) * vector(gtd_xz_p, i, j - 3, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j - 3, k) + gtu_yz_p * vector(vfd_y, i, j - 3, k) + gtu_zz_p * vector(vfd_z, i, j - 3, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j - 3, k) + gtu_yy_p * vector(Sfd_y_p, i, j - 3, k) + gtu_yz_p * vector(Sfd_z_p, i, j - 3, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j - 3, k) + gtu_yz_p * vector(Sfd_y_p, i, j - 3, k) + gtu_zz_p * vector(Sfd_z_p, i, j - 3, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j - 3, k) + gtu_yy_p * vector(vfd_y, i, j - 3, k) + gtu_yz_p * vector(vfd_z, i, j - 3, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j - 3, k) + vfu_y_p * vector(vfd_y, i, j - 3, k) + vfu_z_p * vector(vfd_z, i, j - 3, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j - 3, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j - 3, k) * vector(Bfu_z_p, i, j - 3, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j - 3, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j - 3, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j - 3, k) * vector(Bfu_x_p, i, j - 3, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j - 3, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j - 3, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j - 3, k) * vector(Bfu_z_p, i, j - 3, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j - 3, k) * vfu_z_p + vector(Bfu_z_p, i, j - 3, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j - 3, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j - 3, k) * vector(Bfu_y_p, i, j - 3, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j - 3, k) * vfu_y_p + vector(Bfu_y_p, i, j - 3, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j - 3, k) + Sfuu_xy_p * vector(gtd_xy_p, i, j - 3, k) + Sfuu_xz_p * vector(gtd_xz_p, i, j - 3, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j - 3, k) + Sfuu_xy_p * vector(gtd_yy_p, i, j - 3, k) + Sfuu_xz_p * vector(gtd_yz_p, i, j - 3, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j - 3, k) + Sfuu_xy_p * vector(gtd_yz_p, i, j - 3, k) + Sfuu_xz_p * vector(gtd_zz_p, i, j - 3, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j - 3, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j - 3, k) * vector(Bfu_z_p, i, j - 3, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j - 3, k) * vfu_z_p + vector(Bfu_z_p, i, j - 3, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j - 3, k) + Sfuu_yz_p * vector(gtd_yy_p, i, j - 3, k) + Sfuu_zz_p * vector(gtd_yz_p, i, j - 3, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j - 3, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j - 3, k) + Sfuu_zz_p * vector(gtd_zz_p, i, j - 3, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j - 3, k) + Sfuu_yz_p * vector(gtd_xy_p, i, j - 3, k) + Sfuu_zz_p * vector(gtd_xz_p, i, j - 3, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j - 3, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j - 3, k) * vector(Bfu_y_p, i, j - 3, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j - 3, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j - 3, k) + Sfuu_yy_p * vector(gtd_xy_p, i, j - 3, k) + Sfuu_yz_p * vector(gtd_xz_p, i, j - 3, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j - 3, k) + Sfuu_yy_p * vector(gtd_yy_p, i, j - 3, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j - 3, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j - 3, k) + Sfuu_yy_p * vector(gtd_yz_p, i, j - 3, k) + Sfuu_yz_p * vector(gtd_zz_p, i, j - 3, k));
						Fluxm3jSfd_x = FjSfd_x_mainI(vector(Sfd_x_p, i, j - 3, k), vector(gtd_xx_p, i, j - 3, k), vector(gtd_xy_p, i, j - 3, k), vector(gtd_xz_p, i, j - 3, k), vector(Betau_y_p, i, j - 3, k), vector(Alpha_p, i, j - 3, k), vector(TauT_xy, i, j - 3, k), vector(TauT_yy, i, j - 3, k), vector(TauT_yz, i, j - 3, k), inv_chi_p, sdetg_p, Sfud_yx_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3jSfd_y = FjSfd_y_mainI(vector(Sfd_y_p, i, j - 3, k), vector(gtd_xy_p, i, j - 3, k), vector(gtd_yy_p, i, j - 3, k), vector(gtd_yz_p, i, j - 3, k), vector(Betau_y_p, i, j - 3, k), vector(Alpha_p, i, j - 3, k), vector(TauT_xy, i, j - 3, k), vector(TauT_yy, i, j - 3, k), vector(TauT_yz, i, j - 3, k), inv_chi_p, sdetg_p, Sfud_yy_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3jSfd_z = FjSfd_z_mainI(vector(Sfd_z_p, i, j - 3, k), vector(gtd_xz_p, i, j - 3, k), vector(gtd_yz_p, i, j - 3, k), vector(gtd_zz_p, i, j - 3, k), vector(Betau_y_p, i, j - 3, k), vector(Alpha_p, i, j - 3, k), vector(TauT_xy, i, j - 3, k), vector(TauT_yy, i, j - 3, k), vector(TauT_yz, i, j - 3, k), inv_chi_p, sdetg_p, Sfud_yz_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3jBfu_x = FjBfu_x_mainI(vector(Bfu_x_p, i, j - 3, k), vector(Bfu_y_p, i, j - 3, k), vector(phif_p, i, j - 3, k), vector(Betau_x_p, i, j - 3, k), vector(Betau_y_p, i, j - 3, k), vector(Alpha_p, i, j - 3, k), vector(TauM_xy, i, j - 3, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3jBfu_y = FjBfu_y_mainI(vector(phif_p, i, j - 3, k), vector(Alpha_p, i, j - 3, k), chi_max_p, gtu_yy_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3jBfu_z = FjBfu_z_mainI(vector(Bfu_y_p, i, j - 3, k), vector(Bfu_z_p, i, j - 3, k), vector(phif_p, i, j - 3, k), vector(Betau_y_p, i, j - 3, k), vector(Betau_z_p, i, j - 3, k), vector(Alpha_p, i, j - 3, k), vector(TauM_yz, i, j - 3, k), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3jDf = FjDf_mainI(vector(Df_p, i, j - 3, k), vector(Betau_y_p, i, j - 3, k), vector(Alpha_p, i, j - 3, k), vector(TauN_y, i, j - 3, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3jDYf = FjDYf_mainI(vector(DYf_p, i, j - 3, k), vector(Betau_y_p, i, j - 3, k), vector(Alpha_p, i, j - 3, k), vector(TauNe_y, i, j - 3, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3jtauf = Fjtauf_mainI(vector(Df_p, i, j - 3, k), vector(tauf_p, i, j - 3, k), vector(Betau_y_p, i, j - 3, k), vector(Alpha_p, i, j - 3, k), vector(TauN_y, i, j - 3, k), sdetg_p, Sfu_y_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3jphif = Fjphif_mainI(vector(Bfu_y_p, i, j - 3, k), vector(phif_p, i, j - 3, k), vector(Betau_y_p, i, j - 3, k), vector(Alpha_p, i, j - 3, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j - 3, k) * (1.0 + vector(epsf, i, j - 3, k)) + vector(pf, i, j - 3, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j - 3, k) + sqca_p - vector(sqcs, i, j - 3, k) * sqca_p;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(Betau_y_p, i, j - 3, k)) + vector(Alpha_p, i, j - 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(Betau_y_p, i, j - 3, k)) + vector(Alpha_p, i, j - 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(Betau_y_p, i, j - 3, k)) + vector(Alpha_p, i, j - 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(Betau_y_p, i, j - 3, k)) + vector(Alpha_p, i, j - 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(Betau_y_p, i, j - 3, k)) + vector(Alpha_p, i, j - 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(Betau_y_p, i, j - 3, k)) + vector(Alpha_p, i, j - 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(Betau_y_p, i, j - 3, k)) + vector(Alpha_p, i, j - 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(Betau_y_p, i, j - 3, k)) + vector(Alpha_p, i, j - 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(Betau_y_p, i, j - 3, k)) + vector(Alpha_p, i, j - 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(Betau_y_p, i, j - 3, k)) + vector(Alpha_p, i, j - 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j - 2, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j - 2, k) * vector(Bfu_x_p, i, j - 2, k) + vector(gtd_yy_p, i, j - 2, k) * vector(Bfu_y_p, i, j - 2, k) + vector(gtd_yz_p, i, j - 2, k) * vector(Bfu_z_p, i, j - 2, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j - 2, k) * vector(Bfu_x_p, i, j - 2, k) + vector(gtd_yz_p, i, j - 2, k) * vector(Bfu_y_p, i, j - 2, k) + vector(gtd_zz_p, i, j - 2, k) * vector(Bfu_z_p, i, j - 2, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j - 2, k) * vector(Bfu_x_p, i, j - 2, k) + vector(gtd_xy_p, i, j - 2, k) * vector(Bfu_y_p, i, j - 2, k) + vector(gtd_xz_p, i, j - 2, k) * vector(Bfu_z_p, i, j - 2, k));
						sqBf_p = vector(Bfu_x_p, i, j - 2, k) * Bfd_x_p + vector(Bfu_y_p, i, j - 2, k) * Bfd_y_p + vector(Bfu_z_p, i, j - 2, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j - 2, k) * vector(gtd_zz_p, i, j - 2, k) * vector(gtd_xx_p, i, j - 2, k) + (-vector(gtd_yy_p, i, j - 2, k) * (vector(gtd_xz_p, i, j - 2, k) * vector(gtd_xz_p, i, j - 2, k))) + (-vector(gtd_zz_p, i, j - 2, k) * (vector(gtd_xy_p, i, j - 2, k) * vector(gtd_xy_p, i, j - 2, k))) + (-(vector(gtd_yz_p, i, j - 2, k) * vector(gtd_yz_p, i, j - 2, k)) * vector(gtd_xx_p, i, j - 2, k)) + 2.0 * vector(gtd_yz_p, i, j - 2, k) * vector(gtd_xy_p, i, j - 2, k) * vector(gtd_xz_p, i, j - 2, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j - 2, k) * vector(gtd_yy_p, i, j - 2, k) - vector(gtd_xy_p, i, j - 2, k) * vector(gtd_xy_p, i, j - 2, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j - 2, k) * vector(gtd_zz_p, i, j - 2, k) - vector(gtd_yz_p, i, j - 2, k) * vector(gtd_yz_p, i, j - 2, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j - 2, k) * vector(gtd_yz_p, i, j - 2, k) - vector(gtd_yy_p, i, j - 2, k) * vector(gtd_xz_p, i, j - 2, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j - 2, k) * vector(gtd_zz_p, i, j - 2, k)) + vector(gtd_yz_p, i, j - 2, k) * vector(gtd_xz_p, i, j - 2, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j - 2, k) + gtu_xy_p * vector(Sfd_y_p, i, j - 2, k) + gtu_xz_p * vector(Sfd_z_p, i, j - 2, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j - 2, k) + gtu_xy_p * vector(vfd_y, i, j - 2, k) + gtu_xz_p * vector(vfd_z, i, j - 2, k));
						Bfvf_p = vector(Bfu_x_p, i, j - 2, k) * vector(vfd_x, i, j - 2, k) + vector(Bfu_y_p, i, j - 2, k) * vector(vfd_y, i, j - 2, k) + vector(Bfu_z_p, i, j - 2, k) * vector(vfd_z, i, j - 2, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j - 2, k) * vector(gtd_zz_p, i, j - 2, k) - vector(gtd_xz_p, i, j - 2, k) * vector(gtd_xz_p, i, j - 2, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j - 2, k) * vector(gtd_yz_p, i, j - 2, k)) + vector(gtd_xy_p, i, j - 2, k) * vector(gtd_xz_p, i, j - 2, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j - 2, k) + gtu_yz_p * vector(vfd_y, i, j - 2, k) + gtu_zz_p * vector(vfd_z, i, j - 2, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j - 2, k) + gtu_yy_p * vector(Sfd_y_p, i, j - 2, k) + gtu_yz_p * vector(Sfd_z_p, i, j - 2, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j - 2, k) + gtu_yz_p * vector(Sfd_y_p, i, j - 2, k) + gtu_zz_p * vector(Sfd_z_p, i, j - 2, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j - 2, k) + gtu_yy_p * vector(vfd_y, i, j - 2, k) + gtu_yz_p * vector(vfd_z, i, j - 2, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j - 2, k) + vfu_y_p * vector(vfd_y, i, j - 2, k) + vfu_z_p * vector(vfd_z, i, j - 2, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j - 2, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j - 2, k) * vector(Bfu_z_p, i, j - 2, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j - 2, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j - 2, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j - 2, k) * vector(Bfu_x_p, i, j - 2, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j - 2, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j - 2, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j - 2, k) * vector(Bfu_z_p, i, j - 2, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j - 2, k) * vfu_z_p + vector(Bfu_z_p, i, j - 2, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j - 2, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j - 2, k) * vector(Bfu_y_p, i, j - 2, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j - 2, k) * vfu_y_p + vector(Bfu_y_p, i, j - 2, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j - 2, k) + Sfuu_xy_p * vector(gtd_xy_p, i, j - 2, k) + Sfuu_xz_p * vector(gtd_xz_p, i, j - 2, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j - 2, k) + Sfuu_xy_p * vector(gtd_yy_p, i, j - 2, k) + Sfuu_xz_p * vector(gtd_yz_p, i, j - 2, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j - 2, k) + Sfuu_xy_p * vector(gtd_yz_p, i, j - 2, k) + Sfuu_xz_p * vector(gtd_zz_p, i, j - 2, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j - 2, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j - 2, k) * vector(Bfu_z_p, i, j - 2, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j - 2, k) * vfu_z_p + vector(Bfu_z_p, i, j - 2, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j - 2, k) + Sfuu_yz_p * vector(gtd_yy_p, i, j - 2, k) + Sfuu_zz_p * vector(gtd_yz_p, i, j - 2, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j - 2, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j - 2, k) + Sfuu_zz_p * vector(gtd_zz_p, i, j - 2, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j - 2, k) + Sfuu_yz_p * vector(gtd_xy_p, i, j - 2, k) + Sfuu_zz_p * vector(gtd_xz_p, i, j - 2, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j - 2, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j - 2, k) * vector(Bfu_y_p, i, j - 2, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j - 2, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j - 2, k) + Sfuu_yy_p * vector(gtd_xy_p, i, j - 2, k) + Sfuu_yz_p * vector(gtd_xz_p, i, j - 2, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j - 2, k) + Sfuu_yy_p * vector(gtd_yy_p, i, j - 2, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j - 2, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j - 2, k) + Sfuu_yy_p * vector(gtd_yz_p, i, j - 2, k) + Sfuu_yz_p * vector(gtd_zz_p, i, j - 2, k));
						Fluxm2jSfd_x = FjSfd_x_mainI(vector(Sfd_x_p, i, j - 2, k), vector(gtd_xx_p, i, j - 2, k), vector(gtd_xy_p, i, j - 2, k), vector(gtd_xz_p, i, j - 2, k), vector(Betau_y_p, i, j - 2, k), vector(Alpha_p, i, j - 2, k), vector(TauT_xy, i, j - 2, k), vector(TauT_yy, i, j - 2, k), vector(TauT_yz, i, j - 2, k), inv_chi_p, sdetg_p, Sfud_yx_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2jSfd_y = FjSfd_y_mainI(vector(Sfd_y_p, i, j - 2, k), vector(gtd_xy_p, i, j - 2, k), vector(gtd_yy_p, i, j - 2, k), vector(gtd_yz_p, i, j - 2, k), vector(Betau_y_p, i, j - 2, k), vector(Alpha_p, i, j - 2, k), vector(TauT_xy, i, j - 2, k), vector(TauT_yy, i, j - 2, k), vector(TauT_yz, i, j - 2, k), inv_chi_p, sdetg_p, Sfud_yy_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2jSfd_z = FjSfd_z_mainI(vector(Sfd_z_p, i, j - 2, k), vector(gtd_xz_p, i, j - 2, k), vector(gtd_yz_p, i, j - 2, k), vector(gtd_zz_p, i, j - 2, k), vector(Betau_y_p, i, j - 2, k), vector(Alpha_p, i, j - 2, k), vector(TauT_xy, i, j - 2, k), vector(TauT_yy, i, j - 2, k), vector(TauT_yz, i, j - 2, k), inv_chi_p, sdetg_p, Sfud_yz_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2jBfu_x = FjBfu_x_mainI(vector(Bfu_x_p, i, j - 2, k), vector(Bfu_y_p, i, j - 2, k), vector(phif_p, i, j - 2, k), vector(Betau_x_p, i, j - 2, k), vector(Betau_y_p, i, j - 2, k), vector(Alpha_p, i, j - 2, k), vector(TauM_xy, i, j - 2, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2jBfu_y = FjBfu_y_mainI(vector(phif_p, i, j - 2, k), vector(Alpha_p, i, j - 2, k), chi_max_p, gtu_yy_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2jBfu_z = FjBfu_z_mainI(vector(Bfu_y_p, i, j - 2, k), vector(Bfu_z_p, i, j - 2, k), vector(phif_p, i, j - 2, k), vector(Betau_y_p, i, j - 2, k), vector(Betau_z_p, i, j - 2, k), vector(Alpha_p, i, j - 2, k), vector(TauM_yz, i, j - 2, k), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2jDf = FjDf_mainI(vector(Df_p, i, j - 2, k), vector(Betau_y_p, i, j - 2, k), vector(Alpha_p, i, j - 2, k), vector(TauN_y, i, j - 2, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2jDYf = FjDYf_mainI(vector(DYf_p, i, j - 2, k), vector(Betau_y_p, i, j - 2, k), vector(Alpha_p, i, j - 2, k), vector(TauNe_y, i, j - 2, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2jtauf = Fjtauf_mainI(vector(Df_p, i, j - 2, k), vector(tauf_p, i, j - 2, k), vector(Betau_y_p, i, j - 2, k), vector(Alpha_p, i, j - 2, k), vector(TauN_y, i, j - 2, k), sdetg_p, Sfu_y_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2jphif = Fjphif_mainI(vector(Bfu_y_p, i, j - 2, k), vector(phif_p, i, j - 2, k), vector(Betau_y_p, i, j - 2, k), vector(Alpha_p, i, j - 2, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j - 2, k) * (1.0 + vector(epsf, i, j - 2, k)) + vector(pf, i, j - 2, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j - 2, k) + sqca_p - vector(sqcs, i, j - 2, k) * sqca_p;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(Betau_y_p, i, j - 2, k)) + vector(Alpha_p, i, j - 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(Betau_y_p, i, j - 2, k)) + vector(Alpha_p, i, j - 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(Betau_y_p, i, j - 2, k)) + vector(Alpha_p, i, j - 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(Betau_y_p, i, j - 2, k)) + vector(Alpha_p, i, j - 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(Betau_y_p, i, j - 2, k)) + vector(Alpha_p, i, j - 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(Betau_y_p, i, j - 2, k)) + vector(Alpha_p, i, j - 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(Betau_y_p, i, j - 2, k)) + vector(Alpha_p, i, j - 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(Betau_y_p, i, j - 2, k)) + vector(Alpha_p, i, j - 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(Betau_y_p, i, j - 2, k)) + vector(Alpha_p, i, j - 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(Betau_y_p, i, j - 2, k)) + vector(Alpha_p, i, j - 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j - 1, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j - 1, k) * vector(Bfu_x_p, i, j - 1, k) + vector(gtd_yy_p, i, j - 1, k) * vector(Bfu_y_p, i, j - 1, k) + vector(gtd_yz_p, i, j - 1, k) * vector(Bfu_z_p, i, j - 1, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j - 1, k) * vector(Bfu_x_p, i, j - 1, k) + vector(gtd_yz_p, i, j - 1, k) * vector(Bfu_y_p, i, j - 1, k) + vector(gtd_zz_p, i, j - 1, k) * vector(Bfu_z_p, i, j - 1, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j - 1, k) * vector(Bfu_x_p, i, j - 1, k) + vector(gtd_xy_p, i, j - 1, k) * vector(Bfu_y_p, i, j - 1, k) + vector(gtd_xz_p, i, j - 1, k) * vector(Bfu_z_p, i, j - 1, k));
						sqBf_p = vector(Bfu_x_p, i, j - 1, k) * Bfd_x_p + vector(Bfu_y_p, i, j - 1, k) * Bfd_y_p + vector(Bfu_z_p, i, j - 1, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j - 1, k) * vector(gtd_zz_p, i, j - 1, k) * vector(gtd_xx_p, i, j - 1, k) + (-vector(gtd_yy_p, i, j - 1, k) * (vector(gtd_xz_p, i, j - 1, k) * vector(gtd_xz_p, i, j - 1, k))) + (-vector(gtd_zz_p, i, j - 1, k) * (vector(gtd_xy_p, i, j - 1, k) * vector(gtd_xy_p, i, j - 1, k))) + (-(vector(gtd_yz_p, i, j - 1, k) * vector(gtd_yz_p, i, j - 1, k)) * vector(gtd_xx_p, i, j - 1, k)) + 2.0 * vector(gtd_yz_p, i, j - 1, k) * vector(gtd_xy_p, i, j - 1, k) * vector(gtd_xz_p, i, j - 1, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j - 1, k) * vector(gtd_yy_p, i, j - 1, k) - vector(gtd_xy_p, i, j - 1, k) * vector(gtd_xy_p, i, j - 1, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j - 1, k) * vector(gtd_zz_p, i, j - 1, k) - vector(gtd_yz_p, i, j - 1, k) * vector(gtd_yz_p, i, j - 1, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j - 1, k) * vector(gtd_yz_p, i, j - 1, k) - vector(gtd_yy_p, i, j - 1, k) * vector(gtd_xz_p, i, j - 1, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j - 1, k) * vector(gtd_zz_p, i, j - 1, k)) + vector(gtd_yz_p, i, j - 1, k) * vector(gtd_xz_p, i, j - 1, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j - 1, k) + gtu_xy_p * vector(Sfd_y_p, i, j - 1, k) + gtu_xz_p * vector(Sfd_z_p, i, j - 1, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j - 1, k) + gtu_xy_p * vector(vfd_y, i, j - 1, k) + gtu_xz_p * vector(vfd_z, i, j - 1, k));
						Bfvf_p = vector(Bfu_x_p, i, j - 1, k) * vector(vfd_x, i, j - 1, k) + vector(Bfu_y_p, i, j - 1, k) * vector(vfd_y, i, j - 1, k) + vector(Bfu_z_p, i, j - 1, k) * vector(vfd_z, i, j - 1, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j - 1, k) * vector(gtd_zz_p, i, j - 1, k) - vector(gtd_xz_p, i, j - 1, k) * vector(gtd_xz_p, i, j - 1, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j - 1, k) * vector(gtd_yz_p, i, j - 1, k)) + vector(gtd_xy_p, i, j - 1, k) * vector(gtd_xz_p, i, j - 1, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j - 1, k) + gtu_yz_p * vector(vfd_y, i, j - 1, k) + gtu_zz_p * vector(vfd_z, i, j - 1, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j - 1, k) + gtu_yy_p * vector(Sfd_y_p, i, j - 1, k) + gtu_yz_p * vector(Sfd_z_p, i, j - 1, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j - 1, k) + gtu_yz_p * vector(Sfd_y_p, i, j - 1, k) + gtu_zz_p * vector(Sfd_z_p, i, j - 1, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j - 1, k) + gtu_yy_p * vector(vfd_y, i, j - 1, k) + gtu_yz_p * vector(vfd_z, i, j - 1, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j - 1, k) + vfu_y_p * vector(vfd_y, i, j - 1, k) + vfu_z_p * vector(vfd_z, i, j - 1, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j - 1, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j - 1, k) * vector(Bfu_z_p, i, j - 1, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j - 1, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j - 1, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j - 1, k) * vector(Bfu_x_p, i, j - 1, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j - 1, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j - 1, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j - 1, k) * vector(Bfu_z_p, i, j - 1, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j - 1, k) * vfu_z_p + vector(Bfu_z_p, i, j - 1, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j - 1, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j - 1, k) * vector(Bfu_y_p, i, j - 1, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j - 1, k) * vfu_y_p + vector(Bfu_y_p, i, j - 1, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j - 1, k) + Sfuu_xy_p * vector(gtd_xy_p, i, j - 1, k) + Sfuu_xz_p * vector(gtd_xz_p, i, j - 1, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j - 1, k) + Sfuu_xy_p * vector(gtd_yy_p, i, j - 1, k) + Sfuu_xz_p * vector(gtd_yz_p, i, j - 1, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j - 1, k) + Sfuu_xy_p * vector(gtd_yz_p, i, j - 1, k) + Sfuu_xz_p * vector(gtd_zz_p, i, j - 1, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j - 1, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j - 1, k) * vector(Bfu_z_p, i, j - 1, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j - 1, k) * vfu_z_p + vector(Bfu_z_p, i, j - 1, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j - 1, k) + Sfuu_yz_p * vector(gtd_yy_p, i, j - 1, k) + Sfuu_zz_p * vector(gtd_yz_p, i, j - 1, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j - 1, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j - 1, k) + Sfuu_zz_p * vector(gtd_zz_p, i, j - 1, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j - 1, k) + Sfuu_yz_p * vector(gtd_xy_p, i, j - 1, k) + Sfuu_zz_p * vector(gtd_xz_p, i, j - 1, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j - 1, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j - 1, k) * vector(Bfu_y_p, i, j - 1, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j - 1, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j - 1, k) + Sfuu_yy_p * vector(gtd_xy_p, i, j - 1, k) + Sfuu_yz_p * vector(gtd_xz_p, i, j - 1, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j - 1, k) + Sfuu_yy_p * vector(gtd_yy_p, i, j - 1, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j - 1, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j - 1, k) + Sfuu_yy_p * vector(gtd_yz_p, i, j - 1, k) + Sfuu_yz_p * vector(gtd_zz_p, i, j - 1, k));
						Fluxm1jSfd_x = FjSfd_x_mainI(vector(Sfd_x_p, i, j - 1, k), vector(gtd_xx_p, i, j - 1, k), vector(gtd_xy_p, i, j - 1, k), vector(gtd_xz_p, i, j - 1, k), vector(Betau_y_p, i, j - 1, k), vector(Alpha_p, i, j - 1, k), vector(TauT_xy, i, j - 1, k), vector(TauT_yy, i, j - 1, k), vector(TauT_yz, i, j - 1, k), inv_chi_p, sdetg_p, Sfud_yx_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1jSfd_y = FjSfd_y_mainI(vector(Sfd_y_p, i, j - 1, k), vector(gtd_xy_p, i, j - 1, k), vector(gtd_yy_p, i, j - 1, k), vector(gtd_yz_p, i, j - 1, k), vector(Betau_y_p, i, j - 1, k), vector(Alpha_p, i, j - 1, k), vector(TauT_xy, i, j - 1, k), vector(TauT_yy, i, j - 1, k), vector(TauT_yz, i, j - 1, k), inv_chi_p, sdetg_p, Sfud_yy_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1jSfd_z = FjSfd_z_mainI(vector(Sfd_z_p, i, j - 1, k), vector(gtd_xz_p, i, j - 1, k), vector(gtd_yz_p, i, j - 1, k), vector(gtd_zz_p, i, j - 1, k), vector(Betau_y_p, i, j - 1, k), vector(Alpha_p, i, j - 1, k), vector(TauT_xy, i, j - 1, k), vector(TauT_yy, i, j - 1, k), vector(TauT_yz, i, j - 1, k), inv_chi_p, sdetg_p, Sfud_yz_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1jBfu_x = FjBfu_x_mainI(vector(Bfu_x_p, i, j - 1, k), vector(Bfu_y_p, i, j - 1, k), vector(phif_p, i, j - 1, k), vector(Betau_x_p, i, j - 1, k), vector(Betau_y_p, i, j - 1, k), vector(Alpha_p, i, j - 1, k), vector(TauM_xy, i, j - 1, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1jBfu_y = FjBfu_y_mainI(vector(phif_p, i, j - 1, k), vector(Alpha_p, i, j - 1, k), chi_max_p, gtu_yy_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1jBfu_z = FjBfu_z_mainI(vector(Bfu_y_p, i, j - 1, k), vector(Bfu_z_p, i, j - 1, k), vector(phif_p, i, j - 1, k), vector(Betau_y_p, i, j - 1, k), vector(Betau_z_p, i, j - 1, k), vector(Alpha_p, i, j - 1, k), vector(TauM_yz, i, j - 1, k), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1jDf = FjDf_mainI(vector(Df_p, i, j - 1, k), vector(Betau_y_p, i, j - 1, k), vector(Alpha_p, i, j - 1, k), vector(TauN_y, i, j - 1, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1jDYf = FjDYf_mainI(vector(DYf_p, i, j - 1, k), vector(Betau_y_p, i, j - 1, k), vector(Alpha_p, i, j - 1, k), vector(TauNe_y, i, j - 1, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1jtauf = Fjtauf_mainI(vector(Df_p, i, j - 1, k), vector(tauf_p, i, j - 1, k), vector(Betau_y_p, i, j - 1, k), vector(Alpha_p, i, j - 1, k), vector(TauN_y, i, j - 1, k), sdetg_p, Sfu_y_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1jphif = Fjphif_mainI(vector(Bfu_y_p, i, j - 1, k), vector(phif_p, i, j - 1, k), vector(Betau_y_p, i, j - 1, k), vector(Alpha_p, i, j - 1, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j - 1, k) * (1.0 + vector(epsf, i, j - 1, k)) + vector(pf, i, j - 1, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j - 1, k) + sqca_p - vector(sqcs, i, j - 1, k) * sqca_p;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(Betau_y_p, i, j - 1, k)) + vector(Alpha_p, i, j - 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(Betau_y_p, i, j - 1, k)) + vector(Alpha_p, i, j - 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(Betau_y_p, i, j - 1, k)) + vector(Alpha_p, i, j - 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(Betau_y_p, i, j - 1, k)) + vector(Alpha_p, i, j - 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(Betau_y_p, i, j - 1, k)) + vector(Alpha_p, i, j - 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(Betau_y_p, i, j - 1, k)) + vector(Alpha_p, i, j - 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(Betau_y_p, i, j - 1, k)) + vector(Alpha_p, i, j - 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(Betau_y_p, i, j - 1, k)) + vector(Alpha_p, i, j - 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(Betau_y_p, i, j - 1, k)) + vector(Alpha_p, i, j - 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(Betau_y_p, i, j - 1, k)) + vector(Alpha_p, i, j - 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j + 1, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j + 1, k) * vector(Bfu_x_p, i, j + 1, k) + vector(gtd_yy_p, i, j + 1, k) * vector(Bfu_y_p, i, j + 1, k) + vector(gtd_yz_p, i, j + 1, k) * vector(Bfu_z_p, i, j + 1, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j + 1, k) * vector(Bfu_x_p, i, j + 1, k) + vector(gtd_yz_p, i, j + 1, k) * vector(Bfu_y_p, i, j + 1, k) + vector(gtd_zz_p, i, j + 1, k) * vector(Bfu_z_p, i, j + 1, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j + 1, k) * vector(Bfu_x_p, i, j + 1, k) + vector(gtd_xy_p, i, j + 1, k) * vector(Bfu_y_p, i, j + 1, k) + vector(gtd_xz_p, i, j + 1, k) * vector(Bfu_z_p, i, j + 1, k));
						sqBf_p = vector(Bfu_x_p, i, j + 1, k) * Bfd_x_p + vector(Bfu_y_p, i, j + 1, k) * Bfd_y_p + vector(Bfu_z_p, i, j + 1, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j + 1, k) * vector(gtd_zz_p, i, j + 1, k) * vector(gtd_xx_p, i, j + 1, k) + (-vector(gtd_yy_p, i, j + 1, k) * (vector(gtd_xz_p, i, j + 1, k) * vector(gtd_xz_p, i, j + 1, k))) + (-vector(gtd_zz_p, i, j + 1, k) * (vector(gtd_xy_p, i, j + 1, k) * vector(gtd_xy_p, i, j + 1, k))) + (-(vector(gtd_yz_p, i, j + 1, k) * vector(gtd_yz_p, i, j + 1, k)) * vector(gtd_xx_p, i, j + 1, k)) + 2.0 * vector(gtd_yz_p, i, j + 1, k) * vector(gtd_xy_p, i, j + 1, k) * vector(gtd_xz_p, i, j + 1, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j + 1, k) * vector(gtd_yy_p, i, j + 1, k) - vector(gtd_xy_p, i, j + 1, k) * vector(gtd_xy_p, i, j + 1, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j + 1, k) * vector(gtd_zz_p, i, j + 1, k) - vector(gtd_yz_p, i, j + 1, k) * vector(gtd_yz_p, i, j + 1, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j + 1, k) * vector(gtd_yz_p, i, j + 1, k) - vector(gtd_yy_p, i, j + 1, k) * vector(gtd_xz_p, i, j + 1, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j + 1, k) * vector(gtd_zz_p, i, j + 1, k)) + vector(gtd_yz_p, i, j + 1, k) * vector(gtd_xz_p, i, j + 1, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j + 1, k) + gtu_xy_p * vector(Sfd_y_p, i, j + 1, k) + gtu_xz_p * vector(Sfd_z_p, i, j + 1, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j + 1, k) + gtu_xy_p * vector(vfd_y, i, j + 1, k) + gtu_xz_p * vector(vfd_z, i, j + 1, k));
						Bfvf_p = vector(Bfu_x_p, i, j + 1, k) * vector(vfd_x, i, j + 1, k) + vector(Bfu_y_p, i, j + 1, k) * vector(vfd_y, i, j + 1, k) + vector(Bfu_z_p, i, j + 1, k) * vector(vfd_z, i, j + 1, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j + 1, k) * vector(gtd_zz_p, i, j + 1, k) - vector(gtd_xz_p, i, j + 1, k) * vector(gtd_xz_p, i, j + 1, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j + 1, k) * vector(gtd_yz_p, i, j + 1, k)) + vector(gtd_xy_p, i, j + 1, k) * vector(gtd_xz_p, i, j + 1, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j + 1, k) + gtu_yz_p * vector(vfd_y, i, j + 1, k) + gtu_zz_p * vector(vfd_z, i, j + 1, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j + 1, k) + gtu_yy_p * vector(Sfd_y_p, i, j + 1, k) + gtu_yz_p * vector(Sfd_z_p, i, j + 1, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j + 1, k) + gtu_yz_p * vector(Sfd_y_p, i, j + 1, k) + gtu_zz_p * vector(Sfd_z_p, i, j + 1, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j + 1, k) + gtu_yy_p * vector(vfd_y, i, j + 1, k) + gtu_yz_p * vector(vfd_z, i, j + 1, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j + 1, k) + vfu_y_p * vector(vfd_y, i, j + 1, k) + vfu_z_p * vector(vfd_z, i, j + 1, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j + 1, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j + 1, k) * vector(Bfu_z_p, i, j + 1, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j + 1, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j + 1, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j + 1, k) * vector(Bfu_x_p, i, j + 1, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j + 1, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j + 1, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j + 1, k) * vector(Bfu_z_p, i, j + 1, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j + 1, k) * vfu_z_p + vector(Bfu_z_p, i, j + 1, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j + 1, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j + 1, k) * vector(Bfu_y_p, i, j + 1, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j + 1, k) * vfu_y_p + vector(Bfu_y_p, i, j + 1, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j + 1, k) + Sfuu_xy_p * vector(gtd_xy_p, i, j + 1, k) + Sfuu_xz_p * vector(gtd_xz_p, i, j + 1, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j + 1, k) + Sfuu_xy_p * vector(gtd_yy_p, i, j + 1, k) + Sfuu_xz_p * vector(gtd_yz_p, i, j + 1, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j + 1, k) + Sfuu_xy_p * vector(gtd_yz_p, i, j + 1, k) + Sfuu_xz_p * vector(gtd_zz_p, i, j + 1, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j + 1, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j + 1, k) * vector(Bfu_z_p, i, j + 1, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j + 1, k) * vfu_z_p + vector(Bfu_z_p, i, j + 1, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j + 1, k) + Sfuu_yz_p * vector(gtd_yy_p, i, j + 1, k) + Sfuu_zz_p * vector(gtd_yz_p, i, j + 1, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j + 1, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j + 1, k) + Sfuu_zz_p * vector(gtd_zz_p, i, j + 1, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j + 1, k) + Sfuu_yz_p * vector(gtd_xy_p, i, j + 1, k) + Sfuu_zz_p * vector(gtd_xz_p, i, j + 1, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j + 1, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j + 1, k) * vector(Bfu_y_p, i, j + 1, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j + 1, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j + 1, k) + Sfuu_yy_p * vector(gtd_xy_p, i, j + 1, k) + Sfuu_yz_p * vector(gtd_xz_p, i, j + 1, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j + 1, k) + Sfuu_yy_p * vector(gtd_yy_p, i, j + 1, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j + 1, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j + 1, k) + Sfuu_yy_p * vector(gtd_yz_p, i, j + 1, k) + Sfuu_yz_p * vector(gtd_zz_p, i, j + 1, k));
						Fluxp1jSfd_x = FjSfd_x_mainI(vector(Sfd_x_p, i, j + 1, k), vector(gtd_xx_p, i, j + 1, k), vector(gtd_xy_p, i, j + 1, k), vector(gtd_xz_p, i, j + 1, k), vector(Betau_y_p, i, j + 1, k), vector(Alpha_p, i, j + 1, k), vector(TauT_xy, i, j + 1, k), vector(TauT_yy, i, j + 1, k), vector(TauT_yz, i, j + 1, k), inv_chi_p, sdetg_p, Sfud_yx_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1jSfd_y = FjSfd_y_mainI(vector(Sfd_y_p, i, j + 1, k), vector(gtd_xy_p, i, j + 1, k), vector(gtd_yy_p, i, j + 1, k), vector(gtd_yz_p, i, j + 1, k), vector(Betau_y_p, i, j + 1, k), vector(Alpha_p, i, j + 1, k), vector(TauT_xy, i, j + 1, k), vector(TauT_yy, i, j + 1, k), vector(TauT_yz, i, j + 1, k), inv_chi_p, sdetg_p, Sfud_yy_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1jSfd_z = FjSfd_z_mainI(vector(Sfd_z_p, i, j + 1, k), vector(gtd_xz_p, i, j + 1, k), vector(gtd_yz_p, i, j + 1, k), vector(gtd_zz_p, i, j + 1, k), vector(Betau_y_p, i, j + 1, k), vector(Alpha_p, i, j + 1, k), vector(TauT_xy, i, j + 1, k), vector(TauT_yy, i, j + 1, k), vector(TauT_yz, i, j + 1, k), inv_chi_p, sdetg_p, Sfud_yz_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1jBfu_x = FjBfu_x_mainI(vector(Bfu_x_p, i, j + 1, k), vector(Bfu_y_p, i, j + 1, k), vector(phif_p, i, j + 1, k), vector(Betau_x_p, i, j + 1, k), vector(Betau_y_p, i, j + 1, k), vector(Alpha_p, i, j + 1, k), vector(TauM_xy, i, j + 1, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1jBfu_y = FjBfu_y_mainI(vector(phif_p, i, j + 1, k), vector(Alpha_p, i, j + 1, k), chi_max_p, gtu_yy_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1jBfu_z = FjBfu_z_mainI(vector(Bfu_y_p, i, j + 1, k), vector(Bfu_z_p, i, j + 1, k), vector(phif_p, i, j + 1, k), vector(Betau_y_p, i, j + 1, k), vector(Betau_z_p, i, j + 1, k), vector(Alpha_p, i, j + 1, k), vector(TauM_yz, i, j + 1, k), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1jDf = FjDf_mainI(vector(Df_p, i, j + 1, k), vector(Betau_y_p, i, j + 1, k), vector(Alpha_p, i, j + 1, k), vector(TauN_y, i, j + 1, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1jDYf = FjDYf_mainI(vector(DYf_p, i, j + 1, k), vector(Betau_y_p, i, j + 1, k), vector(Alpha_p, i, j + 1, k), vector(TauNe_y, i, j + 1, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1jtauf = Fjtauf_mainI(vector(Df_p, i, j + 1, k), vector(tauf_p, i, j + 1, k), vector(Betau_y_p, i, j + 1, k), vector(Alpha_p, i, j + 1, k), vector(TauN_y, i, j + 1, k), sdetg_p, Sfu_y_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1jphif = Fjphif_mainI(vector(Bfu_y_p, i, j + 1, k), vector(phif_p, i, j + 1, k), vector(Betau_y_p, i, j + 1, k), vector(Alpha_p, i, j + 1, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j + 1, k) * (1.0 + vector(epsf, i, j + 1, k)) + vector(pf, i, j + 1, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j + 1, k) + sqca_p - vector(sqcs, i, j + 1, k) * sqca_p;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(Betau_y_p, i, j + 1, k)) + vector(Alpha_p, i, j + 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(Betau_y_p, i, j + 1, k)) + vector(Alpha_p, i, j + 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(Betau_y_p, i, j + 1, k)) + vector(Alpha_p, i, j + 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(Betau_y_p, i, j + 1, k)) + vector(Alpha_p, i, j + 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(Betau_y_p, i, j + 1, k)) + vector(Alpha_p, i, j + 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(Betau_y_p, i, j + 1, k)) + vector(Alpha_p, i, j + 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(Betau_y_p, i, j + 1, k)) + vector(Alpha_p, i, j + 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(Betau_y_p, i, j + 1, k)) + vector(Alpha_p, i, j + 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(Betau_y_p, i, j + 1, k)) + vector(Alpha_p, i, j + 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(Betau_y_p, i, j + 1, k)) + vector(Alpha_p, i, j + 1, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j + 2, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j + 2, k) * vector(Bfu_x_p, i, j + 2, k) + vector(gtd_yy_p, i, j + 2, k) * vector(Bfu_y_p, i, j + 2, k) + vector(gtd_yz_p, i, j + 2, k) * vector(Bfu_z_p, i, j + 2, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j + 2, k) * vector(Bfu_x_p, i, j + 2, k) + vector(gtd_yz_p, i, j + 2, k) * vector(Bfu_y_p, i, j + 2, k) + vector(gtd_zz_p, i, j + 2, k) * vector(Bfu_z_p, i, j + 2, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j + 2, k) * vector(Bfu_x_p, i, j + 2, k) + vector(gtd_xy_p, i, j + 2, k) * vector(Bfu_y_p, i, j + 2, k) + vector(gtd_xz_p, i, j + 2, k) * vector(Bfu_z_p, i, j + 2, k));
						sqBf_p = vector(Bfu_x_p, i, j + 2, k) * Bfd_x_p + vector(Bfu_y_p, i, j + 2, k) * Bfd_y_p + vector(Bfu_z_p, i, j + 2, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j + 2, k) * vector(gtd_zz_p, i, j + 2, k) * vector(gtd_xx_p, i, j + 2, k) + (-vector(gtd_yy_p, i, j + 2, k) * (vector(gtd_xz_p, i, j + 2, k) * vector(gtd_xz_p, i, j + 2, k))) + (-vector(gtd_zz_p, i, j + 2, k) * (vector(gtd_xy_p, i, j + 2, k) * vector(gtd_xy_p, i, j + 2, k))) + (-(vector(gtd_yz_p, i, j + 2, k) * vector(gtd_yz_p, i, j + 2, k)) * vector(gtd_xx_p, i, j + 2, k)) + 2.0 * vector(gtd_yz_p, i, j + 2, k) * vector(gtd_xy_p, i, j + 2, k) * vector(gtd_xz_p, i, j + 2, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j + 2, k) * vector(gtd_yy_p, i, j + 2, k) - vector(gtd_xy_p, i, j + 2, k) * vector(gtd_xy_p, i, j + 2, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j + 2, k) * vector(gtd_zz_p, i, j + 2, k) - vector(gtd_yz_p, i, j + 2, k) * vector(gtd_yz_p, i, j + 2, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j + 2, k) * vector(gtd_yz_p, i, j + 2, k) - vector(gtd_yy_p, i, j + 2, k) * vector(gtd_xz_p, i, j + 2, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j + 2, k) * vector(gtd_zz_p, i, j + 2, k)) + vector(gtd_yz_p, i, j + 2, k) * vector(gtd_xz_p, i, j + 2, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j + 2, k) + gtu_xy_p * vector(Sfd_y_p, i, j + 2, k) + gtu_xz_p * vector(Sfd_z_p, i, j + 2, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j + 2, k) + gtu_xy_p * vector(vfd_y, i, j + 2, k) + gtu_xz_p * vector(vfd_z, i, j + 2, k));
						Bfvf_p = vector(Bfu_x_p, i, j + 2, k) * vector(vfd_x, i, j + 2, k) + vector(Bfu_y_p, i, j + 2, k) * vector(vfd_y, i, j + 2, k) + vector(Bfu_z_p, i, j + 2, k) * vector(vfd_z, i, j + 2, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j + 2, k) * vector(gtd_zz_p, i, j + 2, k) - vector(gtd_xz_p, i, j + 2, k) * vector(gtd_xz_p, i, j + 2, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j + 2, k) * vector(gtd_yz_p, i, j + 2, k)) + vector(gtd_xy_p, i, j + 2, k) * vector(gtd_xz_p, i, j + 2, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j + 2, k) + gtu_yz_p * vector(vfd_y, i, j + 2, k) + gtu_zz_p * vector(vfd_z, i, j + 2, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j + 2, k) + gtu_yy_p * vector(Sfd_y_p, i, j + 2, k) + gtu_yz_p * vector(Sfd_z_p, i, j + 2, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j + 2, k) + gtu_yz_p * vector(Sfd_y_p, i, j + 2, k) + gtu_zz_p * vector(Sfd_z_p, i, j + 2, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j + 2, k) + gtu_yy_p * vector(vfd_y, i, j + 2, k) + gtu_yz_p * vector(vfd_z, i, j + 2, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j + 2, k) + vfu_y_p * vector(vfd_y, i, j + 2, k) + vfu_z_p * vector(vfd_z, i, j + 2, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j + 2, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j + 2, k) * vector(Bfu_z_p, i, j + 2, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j + 2, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j + 2, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j + 2, k) * vector(Bfu_x_p, i, j + 2, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j + 2, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j + 2, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j + 2, k) * vector(Bfu_z_p, i, j + 2, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j + 2, k) * vfu_z_p + vector(Bfu_z_p, i, j + 2, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j + 2, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j + 2, k) * vector(Bfu_y_p, i, j + 2, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j + 2, k) * vfu_y_p + vector(Bfu_y_p, i, j + 2, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j + 2, k) + Sfuu_xy_p * vector(gtd_xy_p, i, j + 2, k) + Sfuu_xz_p * vector(gtd_xz_p, i, j + 2, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j + 2, k) + Sfuu_xy_p * vector(gtd_yy_p, i, j + 2, k) + Sfuu_xz_p * vector(gtd_yz_p, i, j + 2, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j + 2, k) + Sfuu_xy_p * vector(gtd_yz_p, i, j + 2, k) + Sfuu_xz_p * vector(gtd_zz_p, i, j + 2, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j + 2, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j + 2, k) * vector(Bfu_z_p, i, j + 2, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j + 2, k) * vfu_z_p + vector(Bfu_z_p, i, j + 2, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j + 2, k) + Sfuu_yz_p * vector(gtd_yy_p, i, j + 2, k) + Sfuu_zz_p * vector(gtd_yz_p, i, j + 2, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j + 2, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j + 2, k) + Sfuu_zz_p * vector(gtd_zz_p, i, j + 2, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j + 2, k) + Sfuu_yz_p * vector(gtd_xy_p, i, j + 2, k) + Sfuu_zz_p * vector(gtd_xz_p, i, j + 2, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j + 2, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j + 2, k) * vector(Bfu_y_p, i, j + 2, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j + 2, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j + 2, k) + Sfuu_yy_p * vector(gtd_xy_p, i, j + 2, k) + Sfuu_yz_p * vector(gtd_xz_p, i, j + 2, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j + 2, k) + Sfuu_yy_p * vector(gtd_yy_p, i, j + 2, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j + 2, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j + 2, k) + Sfuu_yy_p * vector(gtd_yz_p, i, j + 2, k) + Sfuu_yz_p * vector(gtd_zz_p, i, j + 2, k));
						Fluxp2jSfd_x = FjSfd_x_mainI(vector(Sfd_x_p, i, j + 2, k), vector(gtd_xx_p, i, j + 2, k), vector(gtd_xy_p, i, j + 2, k), vector(gtd_xz_p, i, j + 2, k), vector(Betau_y_p, i, j + 2, k), vector(Alpha_p, i, j + 2, k), vector(TauT_xy, i, j + 2, k), vector(TauT_yy, i, j + 2, k), vector(TauT_yz, i, j + 2, k), inv_chi_p, sdetg_p, Sfud_yx_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2jSfd_y = FjSfd_y_mainI(vector(Sfd_y_p, i, j + 2, k), vector(gtd_xy_p, i, j + 2, k), vector(gtd_yy_p, i, j + 2, k), vector(gtd_yz_p, i, j + 2, k), vector(Betau_y_p, i, j + 2, k), vector(Alpha_p, i, j + 2, k), vector(TauT_xy, i, j + 2, k), vector(TauT_yy, i, j + 2, k), vector(TauT_yz, i, j + 2, k), inv_chi_p, sdetg_p, Sfud_yy_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2jSfd_z = FjSfd_z_mainI(vector(Sfd_z_p, i, j + 2, k), vector(gtd_xz_p, i, j + 2, k), vector(gtd_yz_p, i, j + 2, k), vector(gtd_zz_p, i, j + 2, k), vector(Betau_y_p, i, j + 2, k), vector(Alpha_p, i, j + 2, k), vector(TauT_xy, i, j + 2, k), vector(TauT_yy, i, j + 2, k), vector(TauT_yz, i, j + 2, k), inv_chi_p, sdetg_p, Sfud_yz_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2jBfu_x = FjBfu_x_mainI(vector(Bfu_x_p, i, j + 2, k), vector(Bfu_y_p, i, j + 2, k), vector(phif_p, i, j + 2, k), vector(Betau_x_p, i, j + 2, k), vector(Betau_y_p, i, j + 2, k), vector(Alpha_p, i, j + 2, k), vector(TauM_xy, i, j + 2, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2jBfu_y = FjBfu_y_mainI(vector(phif_p, i, j + 2, k), vector(Alpha_p, i, j + 2, k), chi_max_p, gtu_yy_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2jBfu_z = FjBfu_z_mainI(vector(Bfu_y_p, i, j + 2, k), vector(Bfu_z_p, i, j + 2, k), vector(phif_p, i, j + 2, k), vector(Betau_y_p, i, j + 2, k), vector(Betau_z_p, i, j + 2, k), vector(Alpha_p, i, j + 2, k), vector(TauM_yz, i, j + 2, k), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2jDf = FjDf_mainI(vector(Df_p, i, j + 2, k), vector(Betau_y_p, i, j + 2, k), vector(Alpha_p, i, j + 2, k), vector(TauN_y, i, j + 2, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2jDYf = FjDYf_mainI(vector(DYf_p, i, j + 2, k), vector(Betau_y_p, i, j + 2, k), vector(Alpha_p, i, j + 2, k), vector(TauNe_y, i, j + 2, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2jtauf = Fjtauf_mainI(vector(Df_p, i, j + 2, k), vector(tauf_p, i, j + 2, k), vector(Betau_y_p, i, j + 2, k), vector(Alpha_p, i, j + 2, k), vector(TauN_y, i, j + 2, k), sdetg_p, Sfu_y_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2jphif = Fjphif_mainI(vector(Bfu_y_p, i, j + 2, k), vector(phif_p, i, j + 2, k), vector(Betau_y_p, i, j + 2, k), vector(Alpha_p, i, j + 2, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j + 2, k) * (1.0 + vector(epsf, i, j + 2, k)) + vector(pf, i, j + 2, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j + 2, k) + sqca_p - vector(sqcs, i, j + 2, k) * sqca_p;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(Betau_y_p, i, j + 2, k)) + vector(Alpha_p, i, j + 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(Betau_y_p, i, j + 2, k)) + vector(Alpha_p, i, j + 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(Betau_y_p, i, j + 2, k)) + vector(Alpha_p, i, j + 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(Betau_y_p, i, j + 2, k)) + vector(Alpha_p, i, j + 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(Betau_y_p, i, j + 2, k)) + vector(Alpha_p, i, j + 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(Betau_y_p, i, j + 2, k)) + vector(Alpha_p, i, j + 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(Betau_y_p, i, j + 2, k)) + vector(Alpha_p, i, j + 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(Betau_y_p, i, j + 2, k)) + vector(Alpha_p, i, j + 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(Betau_y_p, i, j + 2, k)) + vector(Alpha_p, i, j + 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(Betau_y_p, i, j + 2, k)) + vector(Alpha_p, i, j + 2, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j + 3, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j + 3, k) * vector(Bfu_x_p, i, j + 3, k) + vector(gtd_yy_p, i, j + 3, k) * vector(Bfu_y_p, i, j + 3, k) + vector(gtd_yz_p, i, j + 3, k) * vector(Bfu_z_p, i, j + 3, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j + 3, k) * vector(Bfu_x_p, i, j + 3, k) + vector(gtd_yz_p, i, j + 3, k) * vector(Bfu_y_p, i, j + 3, k) + vector(gtd_zz_p, i, j + 3, k) * vector(Bfu_z_p, i, j + 3, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j + 3, k) * vector(Bfu_x_p, i, j + 3, k) + vector(gtd_xy_p, i, j + 3, k) * vector(Bfu_y_p, i, j + 3, k) + vector(gtd_xz_p, i, j + 3, k) * vector(Bfu_z_p, i, j + 3, k));
						sqBf_p = vector(Bfu_x_p, i, j + 3, k) * Bfd_x_p + vector(Bfu_y_p, i, j + 3, k) * Bfd_y_p + vector(Bfu_z_p, i, j + 3, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j + 3, k) * vector(gtd_zz_p, i, j + 3, k) * vector(gtd_xx_p, i, j + 3, k) + (-vector(gtd_yy_p, i, j + 3, k) * (vector(gtd_xz_p, i, j + 3, k) * vector(gtd_xz_p, i, j + 3, k))) + (-vector(gtd_zz_p, i, j + 3, k) * (vector(gtd_xy_p, i, j + 3, k) * vector(gtd_xy_p, i, j + 3, k))) + (-(vector(gtd_yz_p, i, j + 3, k) * vector(gtd_yz_p, i, j + 3, k)) * vector(gtd_xx_p, i, j + 3, k)) + 2.0 * vector(gtd_yz_p, i, j + 3, k) * vector(gtd_xy_p, i, j + 3, k) * vector(gtd_xz_p, i, j + 3, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j + 3, k) * vector(gtd_yy_p, i, j + 3, k) - vector(gtd_xy_p, i, j + 3, k) * vector(gtd_xy_p, i, j + 3, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j + 3, k) * vector(gtd_zz_p, i, j + 3, k) - vector(gtd_yz_p, i, j + 3, k) * vector(gtd_yz_p, i, j + 3, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j + 3, k) * vector(gtd_yz_p, i, j + 3, k) - vector(gtd_yy_p, i, j + 3, k) * vector(gtd_xz_p, i, j + 3, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j + 3, k) * vector(gtd_zz_p, i, j + 3, k)) + vector(gtd_yz_p, i, j + 3, k) * vector(gtd_xz_p, i, j + 3, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j + 3, k) + gtu_xy_p * vector(Sfd_y_p, i, j + 3, k) + gtu_xz_p * vector(Sfd_z_p, i, j + 3, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j + 3, k) + gtu_xy_p * vector(vfd_y, i, j + 3, k) + gtu_xz_p * vector(vfd_z, i, j + 3, k));
						Bfvf_p = vector(Bfu_x_p, i, j + 3, k) * vector(vfd_x, i, j + 3, k) + vector(Bfu_y_p, i, j + 3, k) * vector(vfd_y, i, j + 3, k) + vector(Bfu_z_p, i, j + 3, k) * vector(vfd_z, i, j + 3, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j + 3, k) * vector(gtd_zz_p, i, j + 3, k) - vector(gtd_xz_p, i, j + 3, k) * vector(gtd_xz_p, i, j + 3, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j + 3, k) * vector(gtd_yz_p, i, j + 3, k)) + vector(gtd_xy_p, i, j + 3, k) * vector(gtd_xz_p, i, j + 3, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j + 3, k) + gtu_yz_p * vector(vfd_y, i, j + 3, k) + gtu_zz_p * vector(vfd_z, i, j + 3, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j + 3, k) + gtu_yy_p * vector(Sfd_y_p, i, j + 3, k) + gtu_yz_p * vector(Sfd_z_p, i, j + 3, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j + 3, k) + gtu_yz_p * vector(Sfd_y_p, i, j + 3, k) + gtu_zz_p * vector(Sfd_z_p, i, j + 3, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j + 3, k) + gtu_yy_p * vector(vfd_y, i, j + 3, k) + gtu_yz_p * vector(vfd_z, i, j + 3, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j + 3, k) + vfu_y_p * vector(vfd_y, i, j + 3, k) + vfu_z_p * vector(vfd_z, i, j + 3, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j + 3, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j + 3, k) * vector(Bfu_z_p, i, j + 3, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j + 3, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j + 3, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j + 3, k) * vector(Bfu_x_p, i, j + 3, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j + 3, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j + 3, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j + 3, k) * vector(Bfu_z_p, i, j + 3, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j + 3, k) * vfu_z_p + vector(Bfu_z_p, i, j + 3, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j + 3, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j + 3, k) * vector(Bfu_y_p, i, j + 3, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j + 3, k) * vfu_y_p + vector(Bfu_y_p, i, j + 3, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j + 3, k) + Sfuu_xy_p * vector(gtd_xy_p, i, j + 3, k) + Sfuu_xz_p * vector(gtd_xz_p, i, j + 3, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j + 3, k) + Sfuu_xy_p * vector(gtd_yy_p, i, j + 3, k) + Sfuu_xz_p * vector(gtd_yz_p, i, j + 3, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j + 3, k) + Sfuu_xy_p * vector(gtd_yz_p, i, j + 3, k) + Sfuu_xz_p * vector(gtd_zz_p, i, j + 3, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j + 3, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j + 3, k) * vector(Bfu_z_p, i, j + 3, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j + 3, k) * vfu_z_p + vector(Bfu_z_p, i, j + 3, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j + 3, k) + Sfuu_yz_p * vector(gtd_yy_p, i, j + 3, k) + Sfuu_zz_p * vector(gtd_yz_p, i, j + 3, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j + 3, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j + 3, k) + Sfuu_zz_p * vector(gtd_zz_p, i, j + 3, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j + 3, k) + Sfuu_yz_p * vector(gtd_xy_p, i, j + 3, k) + Sfuu_zz_p * vector(gtd_xz_p, i, j + 3, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j + 3, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j + 3, k) * vector(Bfu_y_p, i, j + 3, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j + 3, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j + 3, k) + Sfuu_yy_p * vector(gtd_xy_p, i, j + 3, k) + Sfuu_yz_p * vector(gtd_xz_p, i, j + 3, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j + 3, k) + Sfuu_yy_p * vector(gtd_yy_p, i, j + 3, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j + 3, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j + 3, k) + Sfuu_yy_p * vector(gtd_yz_p, i, j + 3, k) + Sfuu_yz_p * vector(gtd_zz_p, i, j + 3, k));
						Fluxp3jSfd_x = FjSfd_x_mainI(vector(Sfd_x_p, i, j + 3, k), vector(gtd_xx_p, i, j + 3, k), vector(gtd_xy_p, i, j + 3, k), vector(gtd_xz_p, i, j + 3, k), vector(Betau_y_p, i, j + 3, k), vector(Alpha_p, i, j + 3, k), vector(TauT_xy, i, j + 3, k), vector(TauT_yy, i, j + 3, k), vector(TauT_yz, i, j + 3, k), inv_chi_p, sdetg_p, Sfud_yx_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3jSfd_y = FjSfd_y_mainI(vector(Sfd_y_p, i, j + 3, k), vector(gtd_xy_p, i, j + 3, k), vector(gtd_yy_p, i, j + 3, k), vector(gtd_yz_p, i, j + 3, k), vector(Betau_y_p, i, j + 3, k), vector(Alpha_p, i, j + 3, k), vector(TauT_xy, i, j + 3, k), vector(TauT_yy, i, j + 3, k), vector(TauT_yz, i, j + 3, k), inv_chi_p, sdetg_p, Sfud_yy_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3jSfd_z = FjSfd_z_mainI(vector(Sfd_z_p, i, j + 3, k), vector(gtd_xz_p, i, j + 3, k), vector(gtd_yz_p, i, j + 3, k), vector(gtd_zz_p, i, j + 3, k), vector(Betau_y_p, i, j + 3, k), vector(Alpha_p, i, j + 3, k), vector(TauT_xy, i, j + 3, k), vector(TauT_yy, i, j + 3, k), vector(TauT_yz, i, j + 3, k), inv_chi_p, sdetg_p, Sfud_yz_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3jBfu_x = FjBfu_x_mainI(vector(Bfu_x_p, i, j + 3, k), vector(Bfu_y_p, i, j + 3, k), vector(phif_p, i, j + 3, k), vector(Betau_x_p, i, j + 3, k), vector(Betau_y_p, i, j + 3, k), vector(Alpha_p, i, j + 3, k), vector(TauM_xy, i, j + 3, k), chi_max_p, gtu_xy_p, sdetg_p, vfu_x_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3jBfu_y = FjBfu_y_mainI(vector(phif_p, i, j + 3, k), vector(Alpha_p, i, j + 3, k), chi_max_p, gtu_yy_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3jBfu_z = FjBfu_z_mainI(vector(Bfu_y_p, i, j + 3, k), vector(Bfu_z_p, i, j + 3, k), vector(phif_p, i, j + 3, k), vector(Betau_y_p, i, j + 3, k), vector(Betau_z_p, i, j + 3, k), vector(Alpha_p, i, j + 3, k), vector(TauM_yz, i, j + 3, k), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3jDf = FjDf_mainI(vector(Df_p, i, j + 3, k), vector(Betau_y_p, i, j + 3, k), vector(Alpha_p, i, j + 3, k), vector(TauN_y, i, j + 3, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3jDYf = FjDYf_mainI(vector(DYf_p, i, j + 3, k), vector(Betau_y_p, i, j + 3, k), vector(Alpha_p, i, j + 3, k), vector(TauNe_y, i, j + 3, k), sdetg_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3jtauf = Fjtauf_mainI(vector(Df_p, i, j + 3, k), vector(tauf_p, i, j + 3, k), vector(Betau_y_p, i, j + 3, k), vector(Alpha_p, i, j + 3, k), vector(TauN_y, i, j + 3, k), sdetg_p, Sfu_y_p, vfu_y_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3jphif = Fjphif_mainI(vector(Bfu_y_p, i, j + 3, k), vector(phif_p, i, j + 3, k), vector(Betau_y_p, i, j + 3, k), vector(Alpha_p, i, j + 3, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j + 3, k) * (1.0 + vector(epsf, i, j + 3, k)) + vector(pf, i, j + 3, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j + 3, k) + sqca_p - vector(sqcs, i, j + 3, k) * sqca_p;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(Betau_y_p, i, j + 3, k)) + vector(Alpha_p, i, j + 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(Betau_y_p, i, j + 3, k)) + vector(Alpha_p, i, j + 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(Betau_y_p, i, j + 3, k)) + vector(Alpha_p, i, j + 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(Betau_y_p, i, j + 3, k)) + vector(Alpha_p, i, j + 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(Betau_y_p, i, j + 3, k)) + vector(Alpha_p, i, j + 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(Betau_y_p, i, j + 3, k)) + vector(Alpha_p, i, j + 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(Betau_y_p, i, j + 3, k)) + vector(Alpha_p, i, j + 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(Betau_y_p, i, j + 3, k)) + vector(Alpha_p, i, j + 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(Betau_y_p, i, j + 3, k)) + vector(Alpha_p, i, j + 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(Betau_y_p, i, j + 3, k)) + vector(Alpha_p, i, j + 3, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_y_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_yy_p - (1.0 - sqa_p) * (vfu_y_p * vfu_y_p)))))));
						LLFpjSfd_x = 0.5 * (FluxjSfd_x + fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j, k));
						LLFpjSfd_y = 0.5 * (FluxjSfd_y + fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j, k));
						LLFpjSfd_z = 0.5 * (FluxjSfd_z + fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j, k));
						LLFpjBfu_x = 0.5 * (FluxjBfu_x + fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j, k));
						LLFpjBfu_y = 0.5 * (FluxjBfu_y + fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j, k));
						LLFpjBfu_z = 0.5 * (FluxjBfu_z + fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j, k));
						LLFpjDf = 0.5 * (FluxjDf + fabs(SpeedjDf) * vector(Df_p, i, j, k));
						LLFpjDYf = 0.5 * (FluxjDYf + fabs(SpeedjDYf) * vector(DYf_p, i, j, k));
						LLFpjtauf = 0.5 * (Fluxjtauf + fabs(Speedjtauf) * vector(tauf_p, i, j, k));
						LLFpjphif = 0.5 * (Fluxjphif + fabs(Speedjphif) * vector(phif_p, i, j, k));
						LLFnjSfd_x = 0.5 * (FluxjSfd_x - fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j, k));
						LLFnjSfd_y = 0.5 * (FluxjSfd_y - fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j, k));
						LLFnjSfd_z = 0.5 * (FluxjSfd_z - fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j, k));
						LLFnjBfu_x = 0.5 * (FluxjBfu_x - fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j, k));
						LLFnjBfu_y = 0.5 * (FluxjBfu_y - fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j, k));
						LLFnjBfu_z = 0.5 * (FluxjBfu_z - fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j, k));
						LLFnjDf = 0.5 * (FluxjDf - fabs(SpeedjDf) * vector(Df_p, i, j, k));
						LLFnjDYf = 0.5 * (FluxjDYf - fabs(SpeedjDYf) * vector(DYf_p, i, j, k));
						LLFnjtauf = 0.5 * (Fluxjtauf - fabs(Speedjtauf) * vector(tauf_p, i, j, k));
						LLFnjphif = 0.5 * (Fluxjphif - fabs(Speedjphif) * vector(phif_p, i, j, k));
						LLFp_m3jSfd_x = 0.5 * (Fluxm3jSfd_x + fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j - 3, k));
						LLFp_m3jSfd_y = 0.5 * (Fluxm3jSfd_y + fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j - 3, k));
						LLFp_m3jSfd_z = 0.5 * (Fluxm3jSfd_z + fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j - 3, k));
						LLFp_m3jBfu_x = 0.5 * (Fluxm3jBfu_x + fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j - 3, k));
						LLFp_m3jBfu_y = 0.5 * (Fluxm3jBfu_y + fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j - 3, k));
						LLFp_m3jBfu_z = 0.5 * (Fluxm3jBfu_z + fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j - 3, k));
						LLFp_m3jDf = 0.5 * (Fluxm3jDf + fabs(SpeedjDf) * vector(Df_p, i, j - 3, k));
						LLFp_m3jDYf = 0.5 * (Fluxm3jDYf + fabs(SpeedjDYf) * vector(DYf_p, i, j - 3, k));
						LLFp_m3jtauf = 0.5 * (Fluxm3jtauf + fabs(Speedjtauf) * vector(tauf_p, i, j - 3, k));
						LLFp_m3jphif = 0.5 * (Fluxm3jphif + fabs(Speedjphif) * vector(phif_p, i, j - 3, k));
						LLFn_m3jSfd_x = 0.5 * (Fluxm3jSfd_x - fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j - 3, k));
						LLFn_m3jSfd_y = 0.5 * (Fluxm3jSfd_y - fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j - 3, k));
						LLFn_m3jSfd_z = 0.5 * (Fluxm3jSfd_z - fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j - 3, k));
						LLFn_m3jBfu_x = 0.5 * (Fluxm3jBfu_x - fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j - 3, k));
						LLFn_m3jBfu_y = 0.5 * (Fluxm3jBfu_y - fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j - 3, k));
						LLFn_m3jBfu_z = 0.5 * (Fluxm3jBfu_z - fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j - 3, k));
						LLFn_m3jDf = 0.5 * (Fluxm3jDf - fabs(SpeedjDf) * vector(Df_p, i, j - 3, k));
						LLFn_m3jDYf = 0.5 * (Fluxm3jDYf - fabs(SpeedjDYf) * vector(DYf_p, i, j - 3, k));
						LLFn_m3jtauf = 0.5 * (Fluxm3jtauf - fabs(Speedjtauf) * vector(tauf_p, i, j - 3, k));
						LLFn_m3jphif = 0.5 * (Fluxm3jphif - fabs(Speedjphif) * vector(phif_p, i, j - 3, k));
						LLFp_m2jSfd_x = 0.5 * (Fluxm2jSfd_x + fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j - 2, k));
						LLFp_m2jSfd_y = 0.5 * (Fluxm2jSfd_y + fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j - 2, k));
						LLFp_m2jSfd_z = 0.5 * (Fluxm2jSfd_z + fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j - 2, k));
						LLFp_m2jBfu_x = 0.5 * (Fluxm2jBfu_x + fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j - 2, k));
						LLFp_m2jBfu_y = 0.5 * (Fluxm2jBfu_y + fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j - 2, k));
						LLFp_m2jBfu_z = 0.5 * (Fluxm2jBfu_z + fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j - 2, k));
						LLFp_m2jDf = 0.5 * (Fluxm2jDf + fabs(SpeedjDf) * vector(Df_p, i, j - 2, k));
						LLFp_m2jDYf = 0.5 * (Fluxm2jDYf + fabs(SpeedjDYf) * vector(DYf_p, i, j - 2, k));
						LLFp_m2jtauf = 0.5 * (Fluxm2jtauf + fabs(Speedjtauf) * vector(tauf_p, i, j - 2, k));
						LLFp_m2jphif = 0.5 * (Fluxm2jphif + fabs(Speedjphif) * vector(phif_p, i, j - 2, k));
						LLFn_m2jSfd_x = 0.5 * (Fluxm2jSfd_x - fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j - 2, k));
						LLFn_m2jSfd_y = 0.5 * (Fluxm2jSfd_y - fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j - 2, k));
						LLFn_m2jSfd_z = 0.5 * (Fluxm2jSfd_z - fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j - 2, k));
						LLFn_m2jBfu_x = 0.5 * (Fluxm2jBfu_x - fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j - 2, k));
						LLFn_m2jBfu_y = 0.5 * (Fluxm2jBfu_y - fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j - 2, k));
						LLFn_m2jBfu_z = 0.5 * (Fluxm2jBfu_z - fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j - 2, k));
						LLFn_m2jDf = 0.5 * (Fluxm2jDf - fabs(SpeedjDf) * vector(Df_p, i, j - 2, k));
						LLFn_m2jDYf = 0.5 * (Fluxm2jDYf - fabs(SpeedjDYf) * vector(DYf_p, i, j - 2, k));
						LLFn_m2jtauf = 0.5 * (Fluxm2jtauf - fabs(Speedjtauf) * vector(tauf_p, i, j - 2, k));
						LLFn_m2jphif = 0.5 * (Fluxm2jphif - fabs(Speedjphif) * vector(phif_p, i, j - 2, k));
						LLFp_m1jSfd_x = 0.5 * (Fluxm1jSfd_x + fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j - 1, k));
						LLFp_m1jSfd_y = 0.5 * (Fluxm1jSfd_y + fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j - 1, k));
						LLFp_m1jSfd_z = 0.5 * (Fluxm1jSfd_z + fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j - 1, k));
						LLFp_m1jBfu_x = 0.5 * (Fluxm1jBfu_x + fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j - 1, k));
						LLFp_m1jBfu_y = 0.5 * (Fluxm1jBfu_y + fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j - 1, k));
						LLFp_m1jBfu_z = 0.5 * (Fluxm1jBfu_z + fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j - 1, k));
						LLFp_m1jDf = 0.5 * (Fluxm1jDf + fabs(SpeedjDf) * vector(Df_p, i, j - 1, k));
						LLFp_m1jDYf = 0.5 * (Fluxm1jDYf + fabs(SpeedjDYf) * vector(DYf_p, i, j - 1, k));
						LLFp_m1jtauf = 0.5 * (Fluxm1jtauf + fabs(Speedjtauf) * vector(tauf_p, i, j - 1, k));
						LLFp_m1jphif = 0.5 * (Fluxm1jphif + fabs(Speedjphif) * vector(phif_p, i, j - 1, k));
						LLFn_m1jSfd_x = 0.5 * (Fluxm1jSfd_x - fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j - 1, k));
						LLFn_m1jSfd_y = 0.5 * (Fluxm1jSfd_y - fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j - 1, k));
						LLFn_m1jSfd_z = 0.5 * (Fluxm1jSfd_z - fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j - 1, k));
						LLFn_m1jBfu_x = 0.5 * (Fluxm1jBfu_x - fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j - 1, k));
						LLFn_m1jBfu_y = 0.5 * (Fluxm1jBfu_y - fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j - 1, k));
						LLFn_m1jBfu_z = 0.5 * (Fluxm1jBfu_z - fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j - 1, k));
						LLFn_m1jDf = 0.5 * (Fluxm1jDf - fabs(SpeedjDf) * vector(Df_p, i, j - 1, k));
						LLFn_m1jDYf = 0.5 * (Fluxm1jDYf - fabs(SpeedjDYf) * vector(DYf_p, i, j - 1, k));
						LLFn_m1jtauf = 0.5 * (Fluxm1jtauf - fabs(Speedjtauf) * vector(tauf_p, i, j - 1, k));
						LLFn_m1jphif = 0.5 * (Fluxm1jphif - fabs(Speedjphif) * vector(phif_p, i, j - 1, k));
						LLFp_p1jSfd_x = 0.5 * (Fluxp1jSfd_x + fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j + 1, k));
						LLFp_p1jSfd_y = 0.5 * (Fluxp1jSfd_y + fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j + 1, k));
						LLFp_p1jSfd_z = 0.5 * (Fluxp1jSfd_z + fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j + 1, k));
						LLFp_p1jBfu_x = 0.5 * (Fluxp1jBfu_x + fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j + 1, k));
						LLFp_p1jBfu_y = 0.5 * (Fluxp1jBfu_y + fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j + 1, k));
						LLFp_p1jBfu_z = 0.5 * (Fluxp1jBfu_z + fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j + 1, k));
						LLFp_p1jDf = 0.5 * (Fluxp1jDf + fabs(SpeedjDf) * vector(Df_p, i, j + 1, k));
						LLFp_p1jDYf = 0.5 * (Fluxp1jDYf + fabs(SpeedjDYf) * vector(DYf_p, i, j + 1, k));
						LLFp_p1jtauf = 0.5 * (Fluxp1jtauf + fabs(Speedjtauf) * vector(tauf_p, i, j + 1, k));
						LLFp_p1jphif = 0.5 * (Fluxp1jphif + fabs(Speedjphif) * vector(phif_p, i, j + 1, k));
						LLFn_p1jSfd_x = 0.5 * (Fluxp1jSfd_x - fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j + 1, k));
						LLFn_p1jSfd_y = 0.5 * (Fluxp1jSfd_y - fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j + 1, k));
						LLFn_p1jSfd_z = 0.5 * (Fluxp1jSfd_z - fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j + 1, k));
						LLFn_p1jBfu_x = 0.5 * (Fluxp1jBfu_x - fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j + 1, k));
						LLFn_p1jBfu_y = 0.5 * (Fluxp1jBfu_y - fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j + 1, k));
						LLFn_p1jBfu_z = 0.5 * (Fluxp1jBfu_z - fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j + 1, k));
						LLFn_p1jDf = 0.5 * (Fluxp1jDf - fabs(SpeedjDf) * vector(Df_p, i, j + 1, k));
						LLFn_p1jDYf = 0.5 * (Fluxp1jDYf - fabs(SpeedjDYf) * vector(DYf_p, i, j + 1, k));
						LLFn_p1jtauf = 0.5 * (Fluxp1jtauf - fabs(Speedjtauf) * vector(tauf_p, i, j + 1, k));
						LLFn_p1jphif = 0.5 * (Fluxp1jphif - fabs(Speedjphif) * vector(phif_p, i, j + 1, k));
						LLFp_p2jSfd_x = 0.5 * (Fluxp2jSfd_x + fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j + 2, k));
						LLFp_p2jSfd_y = 0.5 * (Fluxp2jSfd_y + fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j + 2, k));
						LLFp_p2jSfd_z = 0.5 * (Fluxp2jSfd_z + fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j + 2, k));
						LLFp_p2jBfu_x = 0.5 * (Fluxp2jBfu_x + fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j + 2, k));
						LLFp_p2jBfu_y = 0.5 * (Fluxp2jBfu_y + fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j + 2, k));
						LLFp_p2jBfu_z = 0.5 * (Fluxp2jBfu_z + fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j + 2, k));
						LLFp_p2jDf = 0.5 * (Fluxp2jDf + fabs(SpeedjDf) * vector(Df_p, i, j + 2, k));
						LLFp_p2jDYf = 0.5 * (Fluxp2jDYf + fabs(SpeedjDYf) * vector(DYf_p, i, j + 2, k));
						LLFp_p2jtauf = 0.5 * (Fluxp2jtauf + fabs(Speedjtauf) * vector(tauf_p, i, j + 2, k));
						LLFp_p2jphif = 0.5 * (Fluxp2jphif + fabs(Speedjphif) * vector(phif_p, i, j + 2, k));
						LLFn_p2jSfd_x = 0.5 * (Fluxp2jSfd_x - fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j + 2, k));
						LLFn_p2jSfd_y = 0.5 * (Fluxp2jSfd_y - fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j + 2, k));
						LLFn_p2jSfd_z = 0.5 * (Fluxp2jSfd_z - fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j + 2, k));
						LLFn_p2jBfu_x = 0.5 * (Fluxp2jBfu_x - fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j + 2, k));
						LLFn_p2jBfu_y = 0.5 * (Fluxp2jBfu_y - fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j + 2, k));
						LLFn_p2jBfu_z = 0.5 * (Fluxp2jBfu_z - fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j + 2, k));
						LLFn_p2jDf = 0.5 * (Fluxp2jDf - fabs(SpeedjDf) * vector(Df_p, i, j + 2, k));
						LLFn_p2jDYf = 0.5 * (Fluxp2jDYf - fabs(SpeedjDYf) * vector(DYf_p, i, j + 2, k));
						LLFn_p2jtauf = 0.5 * (Fluxp2jtauf - fabs(Speedjtauf) * vector(tauf_p, i, j + 2, k));
						LLFn_p2jphif = 0.5 * (Fluxp2jphif - fabs(Speedjphif) * vector(phif_p, i, j + 2, k));
						LLFp_p3jSfd_x = 0.5 * (Fluxp3jSfd_x + fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j + 3, k));
						LLFp_p3jSfd_y = 0.5 * (Fluxp3jSfd_y + fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j + 3, k));
						LLFp_p3jSfd_z = 0.5 * (Fluxp3jSfd_z + fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j + 3, k));
						LLFp_p3jBfu_x = 0.5 * (Fluxp3jBfu_x + fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j + 3, k));
						LLFp_p3jBfu_y = 0.5 * (Fluxp3jBfu_y + fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j + 3, k));
						LLFp_p3jBfu_z = 0.5 * (Fluxp3jBfu_z + fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j + 3, k));
						LLFp_p3jDf = 0.5 * (Fluxp3jDf + fabs(SpeedjDf) * vector(Df_p, i, j + 3, k));
						LLFp_p3jDYf = 0.5 * (Fluxp3jDYf + fabs(SpeedjDYf) * vector(DYf_p, i, j + 3, k));
						LLFp_p3jtauf = 0.5 * (Fluxp3jtauf + fabs(Speedjtauf) * vector(tauf_p, i, j + 3, k));
						LLFp_p3jphif = 0.5 * (Fluxp3jphif + fabs(Speedjphif) * vector(phif_p, i, j + 3, k));
						LLFn_p3jSfd_x = 0.5 * (Fluxp3jSfd_x - fabs(SpeedjSfd_x) * vector(Sfd_x_p, i, j + 3, k));
						LLFn_p3jSfd_y = 0.5 * (Fluxp3jSfd_y - fabs(SpeedjSfd_y) * vector(Sfd_y_p, i, j + 3, k));
						LLFn_p3jSfd_z = 0.5 * (Fluxp3jSfd_z - fabs(SpeedjSfd_z) * vector(Sfd_z_p, i, j + 3, k));
						LLFn_p3jBfu_x = 0.5 * (Fluxp3jBfu_x - fabs(SpeedjBfu_x) * vector(Bfu_x_p, i, j + 3, k));
						LLFn_p3jBfu_y = 0.5 * (Fluxp3jBfu_y - fabs(SpeedjBfu_y) * vector(Bfu_y_p, i, j + 3, k));
						LLFn_p3jBfu_z = 0.5 * (Fluxp3jBfu_z - fabs(SpeedjBfu_z) * vector(Bfu_z_p, i, j + 3, k));
						LLFn_p3jDf = 0.5 * (Fluxp3jDf - fabs(SpeedjDf) * vector(Df_p, i, j + 3, k));
						LLFn_p3jDYf = 0.5 * (Fluxp3jDYf - fabs(SpeedjDYf) * vector(DYf_p, i, j + 3, k));
						LLFn_p3jtauf = 0.5 * (Fluxp3jtauf - fabs(Speedjtauf) * vector(tauf_p, i, j + 3, k));
						LLFn_p3jphif = 0.5 * (Fluxp3jphif - fabs(Speedjphif) * vector(phif_p, i, j + 3, k));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j, k));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j, k) * vector(Bfu_x_p, i, j, k) + vector(gtd_yy_p, i, j, k) * vector(Bfu_y_p, i, j, k) + vector(gtd_yz_p, i, j, k) * vector(Bfu_z_p, i, j, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j, k) * vector(Bfu_x_p, i, j, k) + vector(gtd_yz_p, i, j, k) * vector(Bfu_y_p, i, j, k) + vector(gtd_zz_p, i, j, k) * vector(Bfu_z_p, i, j, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j, k) * vector(Bfu_x_p, i, j, k) + vector(gtd_xy_p, i, j, k) * vector(Bfu_y_p, i, j, k) + vector(gtd_xz_p, i, j, k) * vector(Bfu_z_p, i, j, k));
						sqBf_p = vector(Bfu_x_p, i, j, k) * Bfd_x_p + vector(Bfu_y_p, i, j, k) * Bfd_y_p + vector(Bfu_z_p, i, j, k) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j, k) * vector(gtd_zz_p, i, j, k) * vector(gtd_xx_p, i, j, k) + (-vector(gtd_yy_p, i, j, k) * (vector(gtd_xz_p, i, j, k) * vector(gtd_xz_p, i, j, k))) + (-vector(gtd_zz_p, i, j, k) * (vector(gtd_xy_p, i, j, k) * vector(gtd_xy_p, i, j, k))) + (-(vector(gtd_yz_p, i, j, k) * vector(gtd_yz_p, i, j, k)) * vector(gtd_xx_p, i, j, k)) + 2.0 * vector(gtd_yz_p, i, j, k) * vector(gtd_xy_p, i, j, k) * vector(gtd_xz_p, i, j, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j, k) * vector(gtd_yy_p, i, j, k) - vector(gtd_xy_p, i, j, k) * vector(gtd_xy_p, i, j, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j, k) * vector(gtd_zz_p, i, j, k) - vector(gtd_yz_p, i, j, k) * vector(gtd_yz_p, i, j, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j, k) * vector(gtd_yz_p, i, j, k) - vector(gtd_yy_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j, k) * vector(gtd_zz_p, i, j, k)) + vector(gtd_yz_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j, k) + gtu_xy_p * vector(Sfd_y_p, i, j, k) + gtu_xz_p * vector(Sfd_z_p, i, j, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j, k) + gtu_xy_p * vector(vfd_y, i, j, k) + gtu_xz_p * vector(vfd_z, i, j, k));
						Bfvf_p = vector(Bfu_x_p, i, j, k) * vector(vfd_x, i, j, k) + vector(Bfu_y_p, i, j, k) * vector(vfd_y, i, j, k) + vector(Bfu_z_p, i, j, k) * vector(vfd_z, i, j, k);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j, k) * vector(gtd_zz_p, i, j, k) - vector(gtd_xz_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j, k) * vector(gtd_yz_p, i, j, k)) + vector(gtd_xy_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j, k) + gtu_yz_p * vector(vfd_y, i, j, k) + gtu_zz_p * vector(vfd_z, i, j, k));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j, k) + gtu_yy_p * vector(Sfd_y_p, i, j, k) + gtu_yz_p * vector(Sfd_z_p, i, j, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j, k) + gtu_yz_p * vector(Sfd_y_p, i, j, k) + gtu_zz_p * vector(Sfd_z_p, i, j, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j, k) + gtu_yy_p * vector(vfd_y, i, j, k) + gtu_yz_p * vector(vfd_z, i, j, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j, k) + vfu_y_p * vector(vfd_y, i, j, k) + vfu_z_p * vector(vfd_z, i, j, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j, k) * vector(Bfu_z_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j, k) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j, k) * vector(Bfu_x_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k) * vector(Bfu_z_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k) * vfu_z_p + vector(Bfu_z_p, i, j, k) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k) * vector(Bfu_y_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k) * vfu_y_p + vector(Bfu_y_p, i, j, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j, k) + Sfuu_xy_p * vector(gtd_xy_p, i, j, k) + Sfuu_xz_p * vector(gtd_xz_p, i, j, k));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j, k) + Sfuu_xy_p * vector(gtd_yy_p, i, j, k) + Sfuu_xz_p * vector(gtd_yz_p, i, j, k));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j, k) + Sfuu_xy_p * vector(gtd_yz_p, i, j, k) + Sfuu_xz_p * vector(gtd_zz_p, i, j, k));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k) * vector(Bfu_z_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k) * vfu_z_p + vector(Bfu_z_p, i, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j, k) + Sfuu_yz_p * vector(gtd_yy_p, i, j, k) + Sfuu_zz_p * vector(gtd_yz_p, i, j, k));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k) + Sfuu_zz_p * vector(gtd_zz_p, i, j, k));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j, k) + Sfuu_yz_p * vector(gtd_xy_p, i, j, k) + Sfuu_zz_p * vector(gtd_xz_p, i, j, k));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j, k) * vector(Bfu_y_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j, k) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j, k) + Sfuu_yy_p * vector(gtd_xy_p, i, j, k) + Sfuu_yz_p * vector(gtd_xz_p, i, j, k));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j, k) + Sfuu_yy_p * vector(gtd_yy_p, i, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j, k) + Sfuu_yy_p * vector(gtd_yz_p, i, j, k) + Sfuu_yz_p * vector(gtd_zz_p, i, j, k));
						FluxkSfd_x = FkSfd_x_mainI(vector(Sfd_x_p, i, j, k), vector(gtd_xx_p, i, j, k), vector(gtd_xy_p, i, j, k), vector(gtd_xz_p, i, j, k), vector(Betau_z_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauT_xz, i, j, k), vector(TauT_yz, i, j, k), vector(TauT_zz, i, j, k), inv_chi_p, sdetg_p, Sfud_zx_p, dx, simPlat_dt, ilast, jlast);
						FluxkSfd_y = FkSfd_y_mainI(vector(Sfd_y_p, i, j, k), vector(gtd_xy_p, i, j, k), vector(gtd_yy_p, i, j, k), vector(gtd_yz_p, i, j, k), vector(Betau_z_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauT_xz, i, j, k), vector(TauT_yz, i, j, k), vector(TauT_zz, i, j, k), inv_chi_p, sdetg_p, Sfud_zy_p, dx, simPlat_dt, ilast, jlast);
						FluxkSfd_z = FkSfd_z_mainI(vector(Sfd_z_p, i, j, k), vector(gtd_xz_p, i, j, k), vector(gtd_yz_p, i, j, k), vector(gtd_zz_p, i, j, k), vector(Betau_z_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauT_xz, i, j, k), vector(TauT_yz, i, j, k), vector(TauT_zz, i, j, k), inv_chi_p, sdetg_p, Sfud_zz_p, dx, simPlat_dt, ilast, jlast);
						FluxkBfu_x = FkBfu_x_mainI(vector(Bfu_x_p, i, j, k), vector(Bfu_z_p, i, j, k), vector(phif_p, i, j, k), vector(Betau_x_p, i, j, k), vector(Betau_z_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauM_xz, i, j, k), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						FluxkBfu_y = FkBfu_y_mainI(vector(Bfu_y_p, i, j, k), vector(Bfu_z_p, i, j, k), vector(phif_p, i, j, k), vector(Betau_y_p, i, j, k), vector(Betau_z_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauM_yz, i, j, k), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						FluxkBfu_z = FkBfu_z_mainI(vector(phif_p, i, j, k), vector(Alpha_p, i, j, k), chi_max_p, gtu_zz_p, dx, simPlat_dt, ilast, jlast);
						FluxkDf = FkDf_mainI(vector(Df_p, i, j, k), vector(Betau_z_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauN_z, i, j, k), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						FluxkDYf = FkDYf_mainI(vector(DYf_p, i, j, k), vector(Betau_z_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauNe_z, i, j, k), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxktauf = Fktauf_mainI(vector(Df_p, i, j, k), vector(tauf_p, i, j, k), vector(Betau_z_p, i, j, k), vector(Alpha_p, i, j, k), vector(TauN_z, i, j, k), sdetg_p, Sfu_z_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxkphif = Fkphif_mainI(vector(Bfu_z_p, i, j, k), vector(phif_p, i, j, k), vector(Betau_z_p, i, j, k), vector(Alpha_p, i, j, k), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j, k) + sqca_p - vector(sqcs, i, j, k) * sqca_p;
						SpeedkSfd_x = fabs(fabs(vector(Betau_z_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p))))));
						SpeedkSfd_y = fabs(fabs(vector(Betau_z_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p))))));
						SpeedkSfd_z = fabs(fabs(vector(Betau_z_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p))))));
						SpeedkBfu_x = fabs(fabs(vector(Betau_z_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p))))));
						SpeedkBfu_y = fabs(fabs(vector(Betau_z_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p))))));
						SpeedkBfu_z = fabs(fabs(vector(Betau_z_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p))))));
						SpeedkDf = fabs(fabs(vector(Betau_z_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p))))));
						SpeedkDYf = fabs(fabs(vector(Betau_z_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p))))));
						Speedktauf = fabs(fabs(vector(Betau_z_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p))))));
						Speedkphif = fabs(fabs(vector(Betau_z_p, i, j, k)) + vector(Alpha_p, i, j, k) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j, k - 3));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j, k - 3) * vector(Bfu_x_p, i, j, k - 3) + vector(gtd_yy_p, i, j, k - 3) * vector(Bfu_y_p, i, j, k - 3) + vector(gtd_yz_p, i, j, k - 3) * vector(Bfu_z_p, i, j, k - 3));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j, k - 3) * vector(Bfu_x_p, i, j, k - 3) + vector(gtd_yz_p, i, j, k - 3) * vector(Bfu_y_p, i, j, k - 3) + vector(gtd_zz_p, i, j, k - 3) * vector(Bfu_z_p, i, j, k - 3));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j, k - 3) * vector(Bfu_x_p, i, j, k - 3) + vector(gtd_xy_p, i, j, k - 3) * vector(Bfu_y_p, i, j, k - 3) + vector(gtd_xz_p, i, j, k - 3) * vector(Bfu_z_p, i, j, k - 3));
						sqBf_p = vector(Bfu_x_p, i, j, k - 3) * Bfd_x_p + vector(Bfu_y_p, i, j, k - 3) * Bfd_y_p + vector(Bfu_z_p, i, j, k - 3) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j, k - 3) * vector(gtd_zz_p, i, j, k - 3) * vector(gtd_xx_p, i, j, k - 3) + (-vector(gtd_yy_p, i, j, k - 3) * (vector(gtd_xz_p, i, j, k - 3) * vector(gtd_xz_p, i, j, k - 3))) + (-vector(gtd_zz_p, i, j, k - 3) * (vector(gtd_xy_p, i, j, k - 3) * vector(gtd_xy_p, i, j, k - 3))) + (-(vector(gtd_yz_p, i, j, k - 3) * vector(gtd_yz_p, i, j, k - 3)) * vector(gtd_xx_p, i, j, k - 3)) + 2.0 * vector(gtd_yz_p, i, j, k - 3) * vector(gtd_xy_p, i, j, k - 3) * vector(gtd_xz_p, i, j, k - 3);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j, k - 3) * vector(gtd_yy_p, i, j, k - 3) - vector(gtd_xy_p, i, j, k - 3) * vector(gtd_xy_p, i, j, k - 3));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j, k - 3) * vector(gtd_zz_p, i, j, k - 3) - vector(gtd_yz_p, i, j, k - 3) * vector(gtd_yz_p, i, j, k - 3));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j, k - 3) * vector(gtd_yz_p, i, j, k - 3) - vector(gtd_yy_p, i, j, k - 3) * vector(gtd_xz_p, i, j, k - 3));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j, k - 3) * vector(gtd_zz_p, i, j, k - 3)) + vector(gtd_yz_p, i, j, k - 3) * vector(gtd_xz_p, i, j, k - 3));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j, k - 3) + gtu_xy_p * vector(Sfd_y_p, i, j, k - 3) + gtu_xz_p * vector(Sfd_z_p, i, j, k - 3));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j, k - 3) + gtu_xy_p * vector(vfd_y, i, j, k - 3) + gtu_xz_p * vector(vfd_z, i, j, k - 3));
						Bfvf_p = vector(Bfu_x_p, i, j, k - 3) * vector(vfd_x, i, j, k - 3) + vector(Bfu_y_p, i, j, k - 3) * vector(vfd_y, i, j, k - 3) + vector(Bfu_z_p, i, j, k - 3) * vector(vfd_z, i, j, k - 3);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j, k - 3) * vector(gtd_zz_p, i, j, k - 3) - vector(gtd_xz_p, i, j, k - 3) * vector(gtd_xz_p, i, j, k - 3));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j, k - 3) * vector(gtd_yz_p, i, j, k - 3)) + vector(gtd_xy_p, i, j, k - 3) * vector(gtd_xz_p, i, j, k - 3));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j, k - 3) + gtu_yz_p * vector(vfd_y, i, j, k - 3) + gtu_zz_p * vector(vfd_z, i, j, k - 3));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j, k - 3) + gtu_yy_p * vector(Sfd_y_p, i, j, k - 3) + gtu_yz_p * vector(Sfd_z_p, i, j, k - 3));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j, k - 3) + gtu_yz_p * vector(Sfd_y_p, i, j, k - 3) + gtu_zz_p * vector(Sfd_z_p, i, j, k - 3));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j, k - 3) + gtu_yy_p * vector(vfd_y, i, j, k - 3) + gtu_yz_p * vector(vfd_z, i, j, k - 3));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j, k - 3) + vfu_y_p * vector(vfd_y, i, j, k - 3) + vfu_z_p * vector(vfd_z, i, j, k - 3), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j, k - 3) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j, k - 3) * vector(Bfu_z_p, i, j, k - 3)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j, k - 3) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j, k - 3) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j, k - 3) * vector(Bfu_x_p, i, j, k - 3)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j, k - 3) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j, k - 3) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k - 3) * vector(Bfu_z_p, i, j, k - 3)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k - 3) * vfu_z_p + vector(Bfu_z_p, i, j, k - 3) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j, k - 3) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k - 3) * vector(Bfu_y_p, i, j, k - 3)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k - 3) * vfu_y_p + vector(Bfu_y_p, i, j, k - 3) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j, k - 3) + Sfuu_xy_p * vector(gtd_xy_p, i, j, k - 3) + Sfuu_xz_p * vector(gtd_xz_p, i, j, k - 3));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j, k - 3) + Sfuu_xy_p * vector(gtd_yy_p, i, j, k - 3) + Sfuu_xz_p * vector(gtd_yz_p, i, j, k - 3));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j, k - 3) + Sfuu_xy_p * vector(gtd_yz_p, i, j, k - 3) + Sfuu_xz_p * vector(gtd_zz_p, i, j, k - 3));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j, k - 3) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k - 3) * vector(Bfu_z_p, i, j, k - 3)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k - 3) * vfu_z_p + vector(Bfu_z_p, i, j, k - 3) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j, k - 3) + Sfuu_yz_p * vector(gtd_yy_p, i, j, k - 3) + Sfuu_zz_p * vector(gtd_yz_p, i, j, k - 3));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j, k - 3) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k - 3) + Sfuu_zz_p * vector(gtd_zz_p, i, j, k - 3));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j, k - 3) + Sfuu_yz_p * vector(gtd_xy_p, i, j, k - 3) + Sfuu_zz_p * vector(gtd_xz_p, i, j, k - 3));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j, k - 3) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j, k - 3) * vector(Bfu_y_p, i, j, k - 3)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j, k - 3) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j, k - 3) + Sfuu_yy_p * vector(gtd_xy_p, i, j, k - 3) + Sfuu_yz_p * vector(gtd_xz_p, i, j, k - 3));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j, k - 3) + Sfuu_yy_p * vector(gtd_yy_p, i, j, k - 3) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k - 3));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j, k - 3) + Sfuu_yy_p * vector(gtd_yz_p, i, j, k - 3) + Sfuu_yz_p * vector(gtd_zz_p, i, j, k - 3));
						Fluxm3kSfd_x = FkSfd_x_mainI(vector(Sfd_x_p, i, j, k - 3), vector(gtd_xx_p, i, j, k - 3), vector(gtd_xy_p, i, j, k - 3), vector(gtd_xz_p, i, j, k - 3), vector(Betau_z_p, i, j, k - 3), vector(Alpha_p, i, j, k - 3), vector(TauT_xz, i, j, k - 3), vector(TauT_yz, i, j, k - 3), vector(TauT_zz, i, j, k - 3), inv_chi_p, sdetg_p, Sfud_zx_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3kSfd_y = FkSfd_y_mainI(vector(Sfd_y_p, i, j, k - 3), vector(gtd_xy_p, i, j, k - 3), vector(gtd_yy_p, i, j, k - 3), vector(gtd_yz_p, i, j, k - 3), vector(Betau_z_p, i, j, k - 3), vector(Alpha_p, i, j, k - 3), vector(TauT_xz, i, j, k - 3), vector(TauT_yz, i, j, k - 3), vector(TauT_zz, i, j, k - 3), inv_chi_p, sdetg_p, Sfud_zy_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3kSfd_z = FkSfd_z_mainI(vector(Sfd_z_p, i, j, k - 3), vector(gtd_xz_p, i, j, k - 3), vector(gtd_yz_p, i, j, k - 3), vector(gtd_zz_p, i, j, k - 3), vector(Betau_z_p, i, j, k - 3), vector(Alpha_p, i, j, k - 3), vector(TauT_xz, i, j, k - 3), vector(TauT_yz, i, j, k - 3), vector(TauT_zz, i, j, k - 3), inv_chi_p, sdetg_p, Sfud_zz_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3kBfu_x = FkBfu_x_mainI(vector(Bfu_x_p, i, j, k - 3), vector(Bfu_z_p, i, j, k - 3), vector(phif_p, i, j, k - 3), vector(Betau_x_p, i, j, k - 3), vector(Betau_z_p, i, j, k - 3), vector(Alpha_p, i, j, k - 3), vector(TauM_xz, i, j, k - 3), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3kBfu_y = FkBfu_y_mainI(vector(Bfu_y_p, i, j, k - 3), vector(Bfu_z_p, i, j, k - 3), vector(phif_p, i, j, k - 3), vector(Betau_y_p, i, j, k - 3), vector(Betau_z_p, i, j, k - 3), vector(Alpha_p, i, j, k - 3), vector(TauM_yz, i, j, k - 3), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3kBfu_z = FkBfu_z_mainI(vector(phif_p, i, j, k - 3), vector(Alpha_p, i, j, k - 3), chi_max_p, gtu_zz_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3kDf = FkDf_mainI(vector(Df_p, i, j, k - 3), vector(Betau_z_p, i, j, k - 3), vector(Alpha_p, i, j, k - 3), vector(TauN_z, i, j, k - 3), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3kDYf = FkDYf_mainI(vector(DYf_p, i, j, k - 3), vector(Betau_z_p, i, j, k - 3), vector(Alpha_p, i, j, k - 3), vector(TauNe_z, i, j, k - 3), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3ktauf = Fktauf_mainI(vector(Df_p, i, j, k - 3), vector(tauf_p, i, j, k - 3), vector(Betau_z_p, i, j, k - 3), vector(Alpha_p, i, j, k - 3), vector(TauN_z, i, j, k - 3), sdetg_p, Sfu_z_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm3kphif = Fkphif_mainI(vector(Bfu_z_p, i, j, k - 3), vector(phif_p, i, j, k - 3), vector(Betau_z_p, i, j, k - 3), vector(Alpha_p, i, j, k - 3), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j, k - 3) * (1.0 + vector(epsf, i, j, k - 3)) + vector(pf, i, j, k - 3);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j, k - 3) + sqca_p - vector(sqcs, i, j, k - 3) * sqca_p;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(Betau_z_p, i, j, k - 3)) + vector(Alpha_p, i, j, k - 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(Betau_z_p, i, j, k - 3)) + vector(Alpha_p, i, j, k - 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(Betau_z_p, i, j, k - 3)) + vector(Alpha_p, i, j, k - 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(Betau_z_p, i, j, k - 3)) + vector(Alpha_p, i, j, k - 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(Betau_z_p, i, j, k - 3)) + vector(Alpha_p, i, j, k - 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(Betau_z_p, i, j, k - 3)) + vector(Alpha_p, i, j, k - 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(Betau_z_p, i, j, k - 3)) + vector(Alpha_p, i, j, k - 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(Betau_z_p, i, j, k - 3)) + vector(Alpha_p, i, j, k - 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(Betau_z_p, i, j, k - 3)) + vector(Alpha_p, i, j, k - 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(Betau_z_p, i, j, k - 3)) + vector(Alpha_p, i, j, k - 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j, k - 2));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j, k - 2) * vector(Bfu_x_p, i, j, k - 2) + vector(gtd_yy_p, i, j, k - 2) * vector(Bfu_y_p, i, j, k - 2) + vector(gtd_yz_p, i, j, k - 2) * vector(Bfu_z_p, i, j, k - 2));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j, k - 2) * vector(Bfu_x_p, i, j, k - 2) + vector(gtd_yz_p, i, j, k - 2) * vector(Bfu_y_p, i, j, k - 2) + vector(gtd_zz_p, i, j, k - 2) * vector(Bfu_z_p, i, j, k - 2));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j, k - 2) * vector(Bfu_x_p, i, j, k - 2) + vector(gtd_xy_p, i, j, k - 2) * vector(Bfu_y_p, i, j, k - 2) + vector(gtd_xz_p, i, j, k - 2) * vector(Bfu_z_p, i, j, k - 2));
						sqBf_p = vector(Bfu_x_p, i, j, k - 2) * Bfd_x_p + vector(Bfu_y_p, i, j, k - 2) * Bfd_y_p + vector(Bfu_z_p, i, j, k - 2) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j, k - 2) * vector(gtd_zz_p, i, j, k - 2) * vector(gtd_xx_p, i, j, k - 2) + (-vector(gtd_yy_p, i, j, k - 2) * (vector(gtd_xz_p, i, j, k - 2) * vector(gtd_xz_p, i, j, k - 2))) + (-vector(gtd_zz_p, i, j, k - 2) * (vector(gtd_xy_p, i, j, k - 2) * vector(gtd_xy_p, i, j, k - 2))) + (-(vector(gtd_yz_p, i, j, k - 2) * vector(gtd_yz_p, i, j, k - 2)) * vector(gtd_xx_p, i, j, k - 2)) + 2.0 * vector(gtd_yz_p, i, j, k - 2) * vector(gtd_xy_p, i, j, k - 2) * vector(gtd_xz_p, i, j, k - 2);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j, k - 2) * vector(gtd_yy_p, i, j, k - 2) - vector(gtd_xy_p, i, j, k - 2) * vector(gtd_xy_p, i, j, k - 2));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j, k - 2) * vector(gtd_zz_p, i, j, k - 2) - vector(gtd_yz_p, i, j, k - 2) * vector(gtd_yz_p, i, j, k - 2));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j, k - 2) * vector(gtd_yz_p, i, j, k - 2) - vector(gtd_yy_p, i, j, k - 2) * vector(gtd_xz_p, i, j, k - 2));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j, k - 2) * vector(gtd_zz_p, i, j, k - 2)) + vector(gtd_yz_p, i, j, k - 2) * vector(gtd_xz_p, i, j, k - 2));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j, k - 2) + gtu_xy_p * vector(Sfd_y_p, i, j, k - 2) + gtu_xz_p * vector(Sfd_z_p, i, j, k - 2));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j, k - 2) + gtu_xy_p * vector(vfd_y, i, j, k - 2) + gtu_xz_p * vector(vfd_z, i, j, k - 2));
						Bfvf_p = vector(Bfu_x_p, i, j, k - 2) * vector(vfd_x, i, j, k - 2) + vector(Bfu_y_p, i, j, k - 2) * vector(vfd_y, i, j, k - 2) + vector(Bfu_z_p, i, j, k - 2) * vector(vfd_z, i, j, k - 2);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j, k - 2) * vector(gtd_zz_p, i, j, k - 2) - vector(gtd_xz_p, i, j, k - 2) * vector(gtd_xz_p, i, j, k - 2));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j, k - 2) * vector(gtd_yz_p, i, j, k - 2)) + vector(gtd_xy_p, i, j, k - 2) * vector(gtd_xz_p, i, j, k - 2));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j, k - 2) + gtu_yz_p * vector(vfd_y, i, j, k - 2) + gtu_zz_p * vector(vfd_z, i, j, k - 2));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j, k - 2) + gtu_yy_p * vector(Sfd_y_p, i, j, k - 2) + gtu_yz_p * vector(Sfd_z_p, i, j, k - 2));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j, k - 2) + gtu_yz_p * vector(Sfd_y_p, i, j, k - 2) + gtu_zz_p * vector(Sfd_z_p, i, j, k - 2));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j, k - 2) + gtu_yy_p * vector(vfd_y, i, j, k - 2) + gtu_yz_p * vector(vfd_z, i, j, k - 2));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j, k - 2) + vfu_y_p * vector(vfd_y, i, j, k - 2) + vfu_z_p * vector(vfd_z, i, j, k - 2), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j, k - 2) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j, k - 2) * vector(Bfu_z_p, i, j, k - 2)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j, k - 2) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j, k - 2) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j, k - 2) * vector(Bfu_x_p, i, j, k - 2)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j, k - 2) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j, k - 2) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k - 2) * vector(Bfu_z_p, i, j, k - 2)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k - 2) * vfu_z_p + vector(Bfu_z_p, i, j, k - 2) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j, k - 2) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k - 2) * vector(Bfu_y_p, i, j, k - 2)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k - 2) * vfu_y_p + vector(Bfu_y_p, i, j, k - 2) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j, k - 2) + Sfuu_xy_p * vector(gtd_xy_p, i, j, k - 2) + Sfuu_xz_p * vector(gtd_xz_p, i, j, k - 2));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j, k - 2) + Sfuu_xy_p * vector(gtd_yy_p, i, j, k - 2) + Sfuu_xz_p * vector(gtd_yz_p, i, j, k - 2));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j, k - 2) + Sfuu_xy_p * vector(gtd_yz_p, i, j, k - 2) + Sfuu_xz_p * vector(gtd_zz_p, i, j, k - 2));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j, k - 2) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k - 2) * vector(Bfu_z_p, i, j, k - 2)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k - 2) * vfu_z_p + vector(Bfu_z_p, i, j, k - 2) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j, k - 2) + Sfuu_yz_p * vector(gtd_yy_p, i, j, k - 2) + Sfuu_zz_p * vector(gtd_yz_p, i, j, k - 2));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j, k - 2) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k - 2) + Sfuu_zz_p * vector(gtd_zz_p, i, j, k - 2));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j, k - 2) + Sfuu_yz_p * vector(gtd_xy_p, i, j, k - 2) + Sfuu_zz_p * vector(gtd_xz_p, i, j, k - 2));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j, k - 2) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j, k - 2) * vector(Bfu_y_p, i, j, k - 2)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j, k - 2) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j, k - 2) + Sfuu_yy_p * vector(gtd_xy_p, i, j, k - 2) + Sfuu_yz_p * vector(gtd_xz_p, i, j, k - 2));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j, k - 2) + Sfuu_yy_p * vector(gtd_yy_p, i, j, k - 2) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k - 2));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j, k - 2) + Sfuu_yy_p * vector(gtd_yz_p, i, j, k - 2) + Sfuu_yz_p * vector(gtd_zz_p, i, j, k - 2));
						Fluxm2kSfd_x = FkSfd_x_mainI(vector(Sfd_x_p, i, j, k - 2), vector(gtd_xx_p, i, j, k - 2), vector(gtd_xy_p, i, j, k - 2), vector(gtd_xz_p, i, j, k - 2), vector(Betau_z_p, i, j, k - 2), vector(Alpha_p, i, j, k - 2), vector(TauT_xz, i, j, k - 2), vector(TauT_yz, i, j, k - 2), vector(TauT_zz, i, j, k - 2), inv_chi_p, sdetg_p, Sfud_zx_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2kSfd_y = FkSfd_y_mainI(vector(Sfd_y_p, i, j, k - 2), vector(gtd_xy_p, i, j, k - 2), vector(gtd_yy_p, i, j, k - 2), vector(gtd_yz_p, i, j, k - 2), vector(Betau_z_p, i, j, k - 2), vector(Alpha_p, i, j, k - 2), vector(TauT_xz, i, j, k - 2), vector(TauT_yz, i, j, k - 2), vector(TauT_zz, i, j, k - 2), inv_chi_p, sdetg_p, Sfud_zy_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2kSfd_z = FkSfd_z_mainI(vector(Sfd_z_p, i, j, k - 2), vector(gtd_xz_p, i, j, k - 2), vector(gtd_yz_p, i, j, k - 2), vector(gtd_zz_p, i, j, k - 2), vector(Betau_z_p, i, j, k - 2), vector(Alpha_p, i, j, k - 2), vector(TauT_xz, i, j, k - 2), vector(TauT_yz, i, j, k - 2), vector(TauT_zz, i, j, k - 2), inv_chi_p, sdetg_p, Sfud_zz_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2kBfu_x = FkBfu_x_mainI(vector(Bfu_x_p, i, j, k - 2), vector(Bfu_z_p, i, j, k - 2), vector(phif_p, i, j, k - 2), vector(Betau_x_p, i, j, k - 2), vector(Betau_z_p, i, j, k - 2), vector(Alpha_p, i, j, k - 2), vector(TauM_xz, i, j, k - 2), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2kBfu_y = FkBfu_y_mainI(vector(Bfu_y_p, i, j, k - 2), vector(Bfu_z_p, i, j, k - 2), vector(phif_p, i, j, k - 2), vector(Betau_y_p, i, j, k - 2), vector(Betau_z_p, i, j, k - 2), vector(Alpha_p, i, j, k - 2), vector(TauM_yz, i, j, k - 2), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2kBfu_z = FkBfu_z_mainI(vector(phif_p, i, j, k - 2), vector(Alpha_p, i, j, k - 2), chi_max_p, gtu_zz_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2kDf = FkDf_mainI(vector(Df_p, i, j, k - 2), vector(Betau_z_p, i, j, k - 2), vector(Alpha_p, i, j, k - 2), vector(TauN_z, i, j, k - 2), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2kDYf = FkDYf_mainI(vector(DYf_p, i, j, k - 2), vector(Betau_z_p, i, j, k - 2), vector(Alpha_p, i, j, k - 2), vector(TauNe_z, i, j, k - 2), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2ktauf = Fktauf_mainI(vector(Df_p, i, j, k - 2), vector(tauf_p, i, j, k - 2), vector(Betau_z_p, i, j, k - 2), vector(Alpha_p, i, j, k - 2), vector(TauN_z, i, j, k - 2), sdetg_p, Sfu_z_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm2kphif = Fkphif_mainI(vector(Bfu_z_p, i, j, k - 2), vector(phif_p, i, j, k - 2), vector(Betau_z_p, i, j, k - 2), vector(Alpha_p, i, j, k - 2), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j, k - 2) * (1.0 + vector(epsf, i, j, k - 2)) + vector(pf, i, j, k - 2);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j, k - 2) + sqca_p - vector(sqcs, i, j, k - 2) * sqca_p;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(Betau_z_p, i, j, k - 2)) + vector(Alpha_p, i, j, k - 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(Betau_z_p, i, j, k - 2)) + vector(Alpha_p, i, j, k - 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(Betau_z_p, i, j, k - 2)) + vector(Alpha_p, i, j, k - 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(Betau_z_p, i, j, k - 2)) + vector(Alpha_p, i, j, k - 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(Betau_z_p, i, j, k - 2)) + vector(Alpha_p, i, j, k - 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(Betau_z_p, i, j, k - 2)) + vector(Alpha_p, i, j, k - 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(Betau_z_p, i, j, k - 2)) + vector(Alpha_p, i, j, k - 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(Betau_z_p, i, j, k - 2)) + vector(Alpha_p, i, j, k - 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(Betau_z_p, i, j, k - 2)) + vector(Alpha_p, i, j, k - 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(Betau_z_p, i, j, k - 2)) + vector(Alpha_p, i, j, k - 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j, k - 1));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j, k - 1) * vector(Bfu_x_p, i, j, k - 1) + vector(gtd_yy_p, i, j, k - 1) * vector(Bfu_y_p, i, j, k - 1) + vector(gtd_yz_p, i, j, k - 1) * vector(Bfu_z_p, i, j, k - 1));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j, k - 1) * vector(Bfu_x_p, i, j, k - 1) + vector(gtd_yz_p, i, j, k - 1) * vector(Bfu_y_p, i, j, k - 1) + vector(gtd_zz_p, i, j, k - 1) * vector(Bfu_z_p, i, j, k - 1));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j, k - 1) * vector(Bfu_x_p, i, j, k - 1) + vector(gtd_xy_p, i, j, k - 1) * vector(Bfu_y_p, i, j, k - 1) + vector(gtd_xz_p, i, j, k - 1) * vector(Bfu_z_p, i, j, k - 1));
						sqBf_p = vector(Bfu_x_p, i, j, k - 1) * Bfd_x_p + vector(Bfu_y_p, i, j, k - 1) * Bfd_y_p + vector(Bfu_z_p, i, j, k - 1) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j, k - 1) * vector(gtd_zz_p, i, j, k - 1) * vector(gtd_xx_p, i, j, k - 1) + (-vector(gtd_yy_p, i, j, k - 1) * (vector(gtd_xz_p, i, j, k - 1) * vector(gtd_xz_p, i, j, k - 1))) + (-vector(gtd_zz_p, i, j, k - 1) * (vector(gtd_xy_p, i, j, k - 1) * vector(gtd_xy_p, i, j, k - 1))) + (-(vector(gtd_yz_p, i, j, k - 1) * vector(gtd_yz_p, i, j, k - 1)) * vector(gtd_xx_p, i, j, k - 1)) + 2.0 * vector(gtd_yz_p, i, j, k - 1) * vector(gtd_xy_p, i, j, k - 1) * vector(gtd_xz_p, i, j, k - 1);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j, k - 1) * vector(gtd_yy_p, i, j, k - 1) - vector(gtd_xy_p, i, j, k - 1) * vector(gtd_xy_p, i, j, k - 1));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j, k - 1) * vector(gtd_zz_p, i, j, k - 1) - vector(gtd_yz_p, i, j, k - 1) * vector(gtd_yz_p, i, j, k - 1));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j, k - 1) * vector(gtd_yz_p, i, j, k - 1) - vector(gtd_yy_p, i, j, k - 1) * vector(gtd_xz_p, i, j, k - 1));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j, k - 1) * vector(gtd_zz_p, i, j, k - 1)) + vector(gtd_yz_p, i, j, k - 1) * vector(gtd_xz_p, i, j, k - 1));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j, k - 1) + gtu_xy_p * vector(Sfd_y_p, i, j, k - 1) + gtu_xz_p * vector(Sfd_z_p, i, j, k - 1));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j, k - 1) + gtu_xy_p * vector(vfd_y, i, j, k - 1) + gtu_xz_p * vector(vfd_z, i, j, k - 1));
						Bfvf_p = vector(Bfu_x_p, i, j, k - 1) * vector(vfd_x, i, j, k - 1) + vector(Bfu_y_p, i, j, k - 1) * vector(vfd_y, i, j, k - 1) + vector(Bfu_z_p, i, j, k - 1) * vector(vfd_z, i, j, k - 1);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j, k - 1) * vector(gtd_zz_p, i, j, k - 1) - vector(gtd_xz_p, i, j, k - 1) * vector(gtd_xz_p, i, j, k - 1));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j, k - 1) * vector(gtd_yz_p, i, j, k - 1)) + vector(gtd_xy_p, i, j, k - 1) * vector(gtd_xz_p, i, j, k - 1));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j, k - 1) + gtu_yz_p * vector(vfd_y, i, j, k - 1) + gtu_zz_p * vector(vfd_z, i, j, k - 1));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j, k - 1) + gtu_yy_p * vector(Sfd_y_p, i, j, k - 1) + gtu_yz_p * vector(Sfd_z_p, i, j, k - 1));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j, k - 1) + gtu_yz_p * vector(Sfd_y_p, i, j, k - 1) + gtu_zz_p * vector(Sfd_z_p, i, j, k - 1));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j, k - 1) + gtu_yy_p * vector(vfd_y, i, j, k - 1) + gtu_yz_p * vector(vfd_z, i, j, k - 1));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j, k - 1) + vfu_y_p * vector(vfd_y, i, j, k - 1) + vfu_z_p * vector(vfd_z, i, j, k - 1), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j, k - 1) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j, k - 1) * vector(Bfu_z_p, i, j, k - 1)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j, k - 1) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j, k - 1) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j, k - 1) * vector(Bfu_x_p, i, j, k - 1)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j, k - 1) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j, k - 1) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k - 1) * vector(Bfu_z_p, i, j, k - 1)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k - 1) * vfu_z_p + vector(Bfu_z_p, i, j, k - 1) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j, k - 1) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k - 1) * vector(Bfu_y_p, i, j, k - 1)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k - 1) * vfu_y_p + vector(Bfu_y_p, i, j, k - 1) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j, k - 1) + Sfuu_xy_p * vector(gtd_xy_p, i, j, k - 1) + Sfuu_xz_p * vector(gtd_xz_p, i, j, k - 1));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j, k - 1) + Sfuu_xy_p * vector(gtd_yy_p, i, j, k - 1) + Sfuu_xz_p * vector(gtd_yz_p, i, j, k - 1));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j, k - 1) + Sfuu_xy_p * vector(gtd_yz_p, i, j, k - 1) + Sfuu_xz_p * vector(gtd_zz_p, i, j, k - 1));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j, k - 1) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k - 1) * vector(Bfu_z_p, i, j, k - 1)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k - 1) * vfu_z_p + vector(Bfu_z_p, i, j, k - 1) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j, k - 1) + Sfuu_yz_p * vector(gtd_yy_p, i, j, k - 1) + Sfuu_zz_p * vector(gtd_yz_p, i, j, k - 1));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j, k - 1) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k - 1) + Sfuu_zz_p * vector(gtd_zz_p, i, j, k - 1));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j, k - 1) + Sfuu_yz_p * vector(gtd_xy_p, i, j, k - 1) + Sfuu_zz_p * vector(gtd_xz_p, i, j, k - 1));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j, k - 1) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j, k - 1) * vector(Bfu_y_p, i, j, k - 1)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j, k - 1) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j, k - 1) + Sfuu_yy_p * vector(gtd_xy_p, i, j, k - 1) + Sfuu_yz_p * vector(gtd_xz_p, i, j, k - 1));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j, k - 1) + Sfuu_yy_p * vector(gtd_yy_p, i, j, k - 1) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k - 1));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j, k - 1) + Sfuu_yy_p * vector(gtd_yz_p, i, j, k - 1) + Sfuu_yz_p * vector(gtd_zz_p, i, j, k - 1));
						Fluxm1kSfd_x = FkSfd_x_mainI(vector(Sfd_x_p, i, j, k - 1), vector(gtd_xx_p, i, j, k - 1), vector(gtd_xy_p, i, j, k - 1), vector(gtd_xz_p, i, j, k - 1), vector(Betau_z_p, i, j, k - 1), vector(Alpha_p, i, j, k - 1), vector(TauT_xz, i, j, k - 1), vector(TauT_yz, i, j, k - 1), vector(TauT_zz, i, j, k - 1), inv_chi_p, sdetg_p, Sfud_zx_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1kSfd_y = FkSfd_y_mainI(vector(Sfd_y_p, i, j, k - 1), vector(gtd_xy_p, i, j, k - 1), vector(gtd_yy_p, i, j, k - 1), vector(gtd_yz_p, i, j, k - 1), vector(Betau_z_p, i, j, k - 1), vector(Alpha_p, i, j, k - 1), vector(TauT_xz, i, j, k - 1), vector(TauT_yz, i, j, k - 1), vector(TauT_zz, i, j, k - 1), inv_chi_p, sdetg_p, Sfud_zy_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1kSfd_z = FkSfd_z_mainI(vector(Sfd_z_p, i, j, k - 1), vector(gtd_xz_p, i, j, k - 1), vector(gtd_yz_p, i, j, k - 1), vector(gtd_zz_p, i, j, k - 1), vector(Betau_z_p, i, j, k - 1), vector(Alpha_p, i, j, k - 1), vector(TauT_xz, i, j, k - 1), vector(TauT_yz, i, j, k - 1), vector(TauT_zz, i, j, k - 1), inv_chi_p, sdetg_p, Sfud_zz_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1kBfu_x = FkBfu_x_mainI(vector(Bfu_x_p, i, j, k - 1), vector(Bfu_z_p, i, j, k - 1), vector(phif_p, i, j, k - 1), vector(Betau_x_p, i, j, k - 1), vector(Betau_z_p, i, j, k - 1), vector(Alpha_p, i, j, k - 1), vector(TauM_xz, i, j, k - 1), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1kBfu_y = FkBfu_y_mainI(vector(Bfu_y_p, i, j, k - 1), vector(Bfu_z_p, i, j, k - 1), vector(phif_p, i, j, k - 1), vector(Betau_y_p, i, j, k - 1), vector(Betau_z_p, i, j, k - 1), vector(Alpha_p, i, j, k - 1), vector(TauM_yz, i, j, k - 1), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1kBfu_z = FkBfu_z_mainI(vector(phif_p, i, j, k - 1), vector(Alpha_p, i, j, k - 1), chi_max_p, gtu_zz_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1kDf = FkDf_mainI(vector(Df_p, i, j, k - 1), vector(Betau_z_p, i, j, k - 1), vector(Alpha_p, i, j, k - 1), vector(TauN_z, i, j, k - 1), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1kDYf = FkDYf_mainI(vector(DYf_p, i, j, k - 1), vector(Betau_z_p, i, j, k - 1), vector(Alpha_p, i, j, k - 1), vector(TauNe_z, i, j, k - 1), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1ktauf = Fktauf_mainI(vector(Df_p, i, j, k - 1), vector(tauf_p, i, j, k - 1), vector(Betau_z_p, i, j, k - 1), vector(Alpha_p, i, j, k - 1), vector(TauN_z, i, j, k - 1), sdetg_p, Sfu_z_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxm1kphif = Fkphif_mainI(vector(Bfu_z_p, i, j, k - 1), vector(phif_p, i, j, k - 1), vector(Betau_z_p, i, j, k - 1), vector(Alpha_p, i, j, k - 1), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j, k - 1) * (1.0 + vector(epsf, i, j, k - 1)) + vector(pf, i, j, k - 1);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j, k - 1) + sqca_p - vector(sqcs, i, j, k - 1) * sqca_p;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(Betau_z_p, i, j, k - 1)) + vector(Alpha_p, i, j, k - 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(Betau_z_p, i, j, k - 1)) + vector(Alpha_p, i, j, k - 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(Betau_z_p, i, j, k - 1)) + vector(Alpha_p, i, j, k - 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(Betau_z_p, i, j, k - 1)) + vector(Alpha_p, i, j, k - 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(Betau_z_p, i, j, k - 1)) + vector(Alpha_p, i, j, k - 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(Betau_z_p, i, j, k - 1)) + vector(Alpha_p, i, j, k - 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(Betau_z_p, i, j, k - 1)) + vector(Alpha_p, i, j, k - 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(Betau_z_p, i, j, k - 1)) + vector(Alpha_p, i, j, k - 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(Betau_z_p, i, j, k - 1)) + vector(Alpha_p, i, j, k - 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(Betau_z_p, i, j, k - 1)) + vector(Alpha_p, i, j, k - 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j, k + 1));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j, k + 1) * vector(Bfu_x_p, i, j, k + 1) + vector(gtd_yy_p, i, j, k + 1) * vector(Bfu_y_p, i, j, k + 1) + vector(gtd_yz_p, i, j, k + 1) * vector(Bfu_z_p, i, j, k + 1));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j, k + 1) * vector(Bfu_x_p, i, j, k + 1) + vector(gtd_yz_p, i, j, k + 1) * vector(Bfu_y_p, i, j, k + 1) + vector(gtd_zz_p, i, j, k + 1) * vector(Bfu_z_p, i, j, k + 1));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j, k + 1) * vector(Bfu_x_p, i, j, k + 1) + vector(gtd_xy_p, i, j, k + 1) * vector(Bfu_y_p, i, j, k + 1) + vector(gtd_xz_p, i, j, k + 1) * vector(Bfu_z_p, i, j, k + 1));
						sqBf_p = vector(Bfu_x_p, i, j, k + 1) * Bfd_x_p + vector(Bfu_y_p, i, j, k + 1) * Bfd_y_p + vector(Bfu_z_p, i, j, k + 1) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j, k + 1) * vector(gtd_zz_p, i, j, k + 1) * vector(gtd_xx_p, i, j, k + 1) + (-vector(gtd_yy_p, i, j, k + 1) * (vector(gtd_xz_p, i, j, k + 1) * vector(gtd_xz_p, i, j, k + 1))) + (-vector(gtd_zz_p, i, j, k + 1) * (vector(gtd_xy_p, i, j, k + 1) * vector(gtd_xy_p, i, j, k + 1))) + (-(vector(gtd_yz_p, i, j, k + 1) * vector(gtd_yz_p, i, j, k + 1)) * vector(gtd_xx_p, i, j, k + 1)) + 2.0 * vector(gtd_yz_p, i, j, k + 1) * vector(gtd_xy_p, i, j, k + 1) * vector(gtd_xz_p, i, j, k + 1);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j, k + 1) * vector(gtd_yy_p, i, j, k + 1) - vector(gtd_xy_p, i, j, k + 1) * vector(gtd_xy_p, i, j, k + 1));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j, k + 1) * vector(gtd_zz_p, i, j, k + 1) - vector(gtd_yz_p, i, j, k + 1) * vector(gtd_yz_p, i, j, k + 1));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j, k + 1) * vector(gtd_yz_p, i, j, k + 1) - vector(gtd_yy_p, i, j, k + 1) * vector(gtd_xz_p, i, j, k + 1));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j, k + 1) * vector(gtd_zz_p, i, j, k + 1)) + vector(gtd_yz_p, i, j, k + 1) * vector(gtd_xz_p, i, j, k + 1));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j, k + 1) + gtu_xy_p * vector(Sfd_y_p, i, j, k + 1) + gtu_xz_p * vector(Sfd_z_p, i, j, k + 1));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j, k + 1) + gtu_xy_p * vector(vfd_y, i, j, k + 1) + gtu_xz_p * vector(vfd_z, i, j, k + 1));
						Bfvf_p = vector(Bfu_x_p, i, j, k + 1) * vector(vfd_x, i, j, k + 1) + vector(Bfu_y_p, i, j, k + 1) * vector(vfd_y, i, j, k + 1) + vector(Bfu_z_p, i, j, k + 1) * vector(vfd_z, i, j, k + 1);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j, k + 1) * vector(gtd_zz_p, i, j, k + 1) - vector(gtd_xz_p, i, j, k + 1) * vector(gtd_xz_p, i, j, k + 1));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j, k + 1) * vector(gtd_yz_p, i, j, k + 1)) + vector(gtd_xy_p, i, j, k + 1) * vector(gtd_xz_p, i, j, k + 1));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j, k + 1) + gtu_yz_p * vector(vfd_y, i, j, k + 1) + gtu_zz_p * vector(vfd_z, i, j, k + 1));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j, k + 1) + gtu_yy_p * vector(Sfd_y_p, i, j, k + 1) + gtu_yz_p * vector(Sfd_z_p, i, j, k + 1));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j, k + 1) + gtu_yz_p * vector(Sfd_y_p, i, j, k + 1) + gtu_zz_p * vector(Sfd_z_p, i, j, k + 1));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j, k + 1) + gtu_yy_p * vector(vfd_y, i, j, k + 1) + gtu_yz_p * vector(vfd_z, i, j, k + 1));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j, k + 1) + vfu_y_p * vector(vfd_y, i, j, k + 1) + vfu_z_p * vector(vfd_z, i, j, k + 1), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j, k + 1) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j, k + 1) * vector(Bfu_z_p, i, j, k + 1)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j, k + 1) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j, k + 1) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j, k + 1) * vector(Bfu_x_p, i, j, k + 1)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j, k + 1) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j, k + 1) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k + 1) * vector(Bfu_z_p, i, j, k + 1)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k + 1) * vfu_z_p + vector(Bfu_z_p, i, j, k + 1) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j, k + 1) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k + 1) * vector(Bfu_y_p, i, j, k + 1)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k + 1) * vfu_y_p + vector(Bfu_y_p, i, j, k + 1) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j, k + 1) + Sfuu_xy_p * vector(gtd_xy_p, i, j, k + 1) + Sfuu_xz_p * vector(gtd_xz_p, i, j, k + 1));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j, k + 1) + Sfuu_xy_p * vector(gtd_yy_p, i, j, k + 1) + Sfuu_xz_p * vector(gtd_yz_p, i, j, k + 1));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j, k + 1) + Sfuu_xy_p * vector(gtd_yz_p, i, j, k + 1) + Sfuu_xz_p * vector(gtd_zz_p, i, j, k + 1));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j, k + 1) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k + 1) * vector(Bfu_z_p, i, j, k + 1)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k + 1) * vfu_z_p + vector(Bfu_z_p, i, j, k + 1) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j, k + 1) + Sfuu_yz_p * vector(gtd_yy_p, i, j, k + 1) + Sfuu_zz_p * vector(gtd_yz_p, i, j, k + 1));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j, k + 1) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k + 1) + Sfuu_zz_p * vector(gtd_zz_p, i, j, k + 1));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j, k + 1) + Sfuu_yz_p * vector(gtd_xy_p, i, j, k + 1) + Sfuu_zz_p * vector(gtd_xz_p, i, j, k + 1));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j, k + 1) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j, k + 1) * vector(Bfu_y_p, i, j, k + 1)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j, k + 1) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j, k + 1) + Sfuu_yy_p * vector(gtd_xy_p, i, j, k + 1) + Sfuu_yz_p * vector(gtd_xz_p, i, j, k + 1));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j, k + 1) + Sfuu_yy_p * vector(gtd_yy_p, i, j, k + 1) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k + 1));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j, k + 1) + Sfuu_yy_p * vector(gtd_yz_p, i, j, k + 1) + Sfuu_yz_p * vector(gtd_zz_p, i, j, k + 1));
						Fluxp1kSfd_x = FkSfd_x_mainI(vector(Sfd_x_p, i, j, k + 1), vector(gtd_xx_p, i, j, k + 1), vector(gtd_xy_p, i, j, k + 1), vector(gtd_xz_p, i, j, k + 1), vector(Betau_z_p, i, j, k + 1), vector(Alpha_p, i, j, k + 1), vector(TauT_xz, i, j, k + 1), vector(TauT_yz, i, j, k + 1), vector(TauT_zz, i, j, k + 1), inv_chi_p, sdetg_p, Sfud_zx_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1kSfd_y = FkSfd_y_mainI(vector(Sfd_y_p, i, j, k + 1), vector(gtd_xy_p, i, j, k + 1), vector(gtd_yy_p, i, j, k + 1), vector(gtd_yz_p, i, j, k + 1), vector(Betau_z_p, i, j, k + 1), vector(Alpha_p, i, j, k + 1), vector(TauT_xz, i, j, k + 1), vector(TauT_yz, i, j, k + 1), vector(TauT_zz, i, j, k + 1), inv_chi_p, sdetg_p, Sfud_zy_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1kSfd_z = FkSfd_z_mainI(vector(Sfd_z_p, i, j, k + 1), vector(gtd_xz_p, i, j, k + 1), vector(gtd_yz_p, i, j, k + 1), vector(gtd_zz_p, i, j, k + 1), vector(Betau_z_p, i, j, k + 1), vector(Alpha_p, i, j, k + 1), vector(TauT_xz, i, j, k + 1), vector(TauT_yz, i, j, k + 1), vector(TauT_zz, i, j, k + 1), inv_chi_p, sdetg_p, Sfud_zz_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1kBfu_x = FkBfu_x_mainI(vector(Bfu_x_p, i, j, k + 1), vector(Bfu_z_p, i, j, k + 1), vector(phif_p, i, j, k + 1), vector(Betau_x_p, i, j, k + 1), vector(Betau_z_p, i, j, k + 1), vector(Alpha_p, i, j, k + 1), vector(TauM_xz, i, j, k + 1), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1kBfu_y = FkBfu_y_mainI(vector(Bfu_y_p, i, j, k + 1), vector(Bfu_z_p, i, j, k + 1), vector(phif_p, i, j, k + 1), vector(Betau_y_p, i, j, k + 1), vector(Betau_z_p, i, j, k + 1), vector(Alpha_p, i, j, k + 1), vector(TauM_yz, i, j, k + 1), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1kBfu_z = FkBfu_z_mainI(vector(phif_p, i, j, k + 1), vector(Alpha_p, i, j, k + 1), chi_max_p, gtu_zz_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1kDf = FkDf_mainI(vector(Df_p, i, j, k + 1), vector(Betau_z_p, i, j, k + 1), vector(Alpha_p, i, j, k + 1), vector(TauN_z, i, j, k + 1), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1kDYf = FkDYf_mainI(vector(DYf_p, i, j, k + 1), vector(Betau_z_p, i, j, k + 1), vector(Alpha_p, i, j, k + 1), vector(TauNe_z, i, j, k + 1), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1ktauf = Fktauf_mainI(vector(Df_p, i, j, k + 1), vector(tauf_p, i, j, k + 1), vector(Betau_z_p, i, j, k + 1), vector(Alpha_p, i, j, k + 1), vector(TauN_z, i, j, k + 1), sdetg_p, Sfu_z_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp1kphif = Fkphif_mainI(vector(Bfu_z_p, i, j, k + 1), vector(phif_p, i, j, k + 1), vector(Betau_z_p, i, j, k + 1), vector(Alpha_p, i, j, k + 1), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j, k + 1) * (1.0 + vector(epsf, i, j, k + 1)) + vector(pf, i, j, k + 1);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j, k + 1) + sqca_p - vector(sqcs, i, j, k + 1) * sqca_p;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(Betau_z_p, i, j, k + 1)) + vector(Alpha_p, i, j, k + 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(Betau_z_p, i, j, k + 1)) + vector(Alpha_p, i, j, k + 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(Betau_z_p, i, j, k + 1)) + vector(Alpha_p, i, j, k + 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(Betau_z_p, i, j, k + 1)) + vector(Alpha_p, i, j, k + 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(Betau_z_p, i, j, k + 1)) + vector(Alpha_p, i, j, k + 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(Betau_z_p, i, j, k + 1)) + vector(Alpha_p, i, j, k + 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(Betau_z_p, i, j, k + 1)) + vector(Alpha_p, i, j, k + 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(Betau_z_p, i, j, k + 1)) + vector(Alpha_p, i, j, k + 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(Betau_z_p, i, j, k + 1)) + vector(Alpha_p, i, j, k + 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(Betau_z_p, i, j, k + 1)) + vector(Alpha_p, i, j, k + 1) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j, k + 2));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j, k + 2) * vector(Bfu_x_p, i, j, k + 2) + vector(gtd_yy_p, i, j, k + 2) * vector(Bfu_y_p, i, j, k + 2) + vector(gtd_yz_p, i, j, k + 2) * vector(Bfu_z_p, i, j, k + 2));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j, k + 2) * vector(Bfu_x_p, i, j, k + 2) + vector(gtd_yz_p, i, j, k + 2) * vector(Bfu_y_p, i, j, k + 2) + vector(gtd_zz_p, i, j, k + 2) * vector(Bfu_z_p, i, j, k + 2));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j, k + 2) * vector(Bfu_x_p, i, j, k + 2) + vector(gtd_xy_p, i, j, k + 2) * vector(Bfu_y_p, i, j, k + 2) + vector(gtd_xz_p, i, j, k + 2) * vector(Bfu_z_p, i, j, k + 2));
						sqBf_p = vector(Bfu_x_p, i, j, k + 2) * Bfd_x_p + vector(Bfu_y_p, i, j, k + 2) * Bfd_y_p + vector(Bfu_z_p, i, j, k + 2) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j, k + 2) * vector(gtd_zz_p, i, j, k + 2) * vector(gtd_xx_p, i, j, k + 2) + (-vector(gtd_yy_p, i, j, k + 2) * (vector(gtd_xz_p, i, j, k + 2) * vector(gtd_xz_p, i, j, k + 2))) + (-vector(gtd_zz_p, i, j, k + 2) * (vector(gtd_xy_p, i, j, k + 2) * vector(gtd_xy_p, i, j, k + 2))) + (-(vector(gtd_yz_p, i, j, k + 2) * vector(gtd_yz_p, i, j, k + 2)) * vector(gtd_xx_p, i, j, k + 2)) + 2.0 * vector(gtd_yz_p, i, j, k + 2) * vector(gtd_xy_p, i, j, k + 2) * vector(gtd_xz_p, i, j, k + 2);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j, k + 2) * vector(gtd_yy_p, i, j, k + 2) - vector(gtd_xy_p, i, j, k + 2) * vector(gtd_xy_p, i, j, k + 2));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j, k + 2) * vector(gtd_zz_p, i, j, k + 2) - vector(gtd_yz_p, i, j, k + 2) * vector(gtd_yz_p, i, j, k + 2));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j, k + 2) * vector(gtd_yz_p, i, j, k + 2) - vector(gtd_yy_p, i, j, k + 2) * vector(gtd_xz_p, i, j, k + 2));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j, k + 2) * vector(gtd_zz_p, i, j, k + 2)) + vector(gtd_yz_p, i, j, k + 2) * vector(gtd_xz_p, i, j, k + 2));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j, k + 2) + gtu_xy_p * vector(Sfd_y_p, i, j, k + 2) + gtu_xz_p * vector(Sfd_z_p, i, j, k + 2));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j, k + 2) + gtu_xy_p * vector(vfd_y, i, j, k + 2) + gtu_xz_p * vector(vfd_z, i, j, k + 2));
						Bfvf_p = vector(Bfu_x_p, i, j, k + 2) * vector(vfd_x, i, j, k + 2) + vector(Bfu_y_p, i, j, k + 2) * vector(vfd_y, i, j, k + 2) + vector(Bfu_z_p, i, j, k + 2) * vector(vfd_z, i, j, k + 2);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j, k + 2) * vector(gtd_zz_p, i, j, k + 2) - vector(gtd_xz_p, i, j, k + 2) * vector(gtd_xz_p, i, j, k + 2));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j, k + 2) * vector(gtd_yz_p, i, j, k + 2)) + vector(gtd_xy_p, i, j, k + 2) * vector(gtd_xz_p, i, j, k + 2));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j, k + 2) + gtu_yz_p * vector(vfd_y, i, j, k + 2) + gtu_zz_p * vector(vfd_z, i, j, k + 2));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j, k + 2) + gtu_yy_p * vector(Sfd_y_p, i, j, k + 2) + gtu_yz_p * vector(Sfd_z_p, i, j, k + 2));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j, k + 2) + gtu_yz_p * vector(Sfd_y_p, i, j, k + 2) + gtu_zz_p * vector(Sfd_z_p, i, j, k + 2));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j, k + 2) + gtu_yy_p * vector(vfd_y, i, j, k + 2) + gtu_yz_p * vector(vfd_z, i, j, k + 2));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j, k + 2) + vfu_y_p * vector(vfd_y, i, j, k + 2) + vfu_z_p * vector(vfd_z, i, j, k + 2), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j, k + 2) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j, k + 2) * vector(Bfu_z_p, i, j, k + 2)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j, k + 2) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j, k + 2) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j, k + 2) * vector(Bfu_x_p, i, j, k + 2)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j, k + 2) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j, k + 2) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k + 2) * vector(Bfu_z_p, i, j, k + 2)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k + 2) * vfu_z_p + vector(Bfu_z_p, i, j, k + 2) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j, k + 2) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k + 2) * vector(Bfu_y_p, i, j, k + 2)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k + 2) * vfu_y_p + vector(Bfu_y_p, i, j, k + 2) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j, k + 2) + Sfuu_xy_p * vector(gtd_xy_p, i, j, k + 2) + Sfuu_xz_p * vector(gtd_xz_p, i, j, k + 2));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j, k + 2) + Sfuu_xy_p * vector(gtd_yy_p, i, j, k + 2) + Sfuu_xz_p * vector(gtd_yz_p, i, j, k + 2));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j, k + 2) + Sfuu_xy_p * vector(gtd_yz_p, i, j, k + 2) + Sfuu_xz_p * vector(gtd_zz_p, i, j, k + 2));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j, k + 2) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k + 2) * vector(Bfu_z_p, i, j, k + 2)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k + 2) * vfu_z_p + vector(Bfu_z_p, i, j, k + 2) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j, k + 2) + Sfuu_yz_p * vector(gtd_yy_p, i, j, k + 2) + Sfuu_zz_p * vector(gtd_yz_p, i, j, k + 2));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j, k + 2) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k + 2) + Sfuu_zz_p * vector(gtd_zz_p, i, j, k + 2));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j, k + 2) + Sfuu_yz_p * vector(gtd_xy_p, i, j, k + 2) + Sfuu_zz_p * vector(gtd_xz_p, i, j, k + 2));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j, k + 2) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j, k + 2) * vector(Bfu_y_p, i, j, k + 2)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j, k + 2) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j, k + 2) + Sfuu_yy_p * vector(gtd_xy_p, i, j, k + 2) + Sfuu_yz_p * vector(gtd_xz_p, i, j, k + 2));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j, k + 2) + Sfuu_yy_p * vector(gtd_yy_p, i, j, k + 2) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k + 2));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j, k + 2) + Sfuu_yy_p * vector(gtd_yz_p, i, j, k + 2) + Sfuu_yz_p * vector(gtd_zz_p, i, j, k + 2));
						Fluxp2kSfd_x = FkSfd_x_mainI(vector(Sfd_x_p, i, j, k + 2), vector(gtd_xx_p, i, j, k + 2), vector(gtd_xy_p, i, j, k + 2), vector(gtd_xz_p, i, j, k + 2), vector(Betau_z_p, i, j, k + 2), vector(Alpha_p, i, j, k + 2), vector(TauT_xz, i, j, k + 2), vector(TauT_yz, i, j, k + 2), vector(TauT_zz, i, j, k + 2), inv_chi_p, sdetg_p, Sfud_zx_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2kSfd_y = FkSfd_y_mainI(vector(Sfd_y_p, i, j, k + 2), vector(gtd_xy_p, i, j, k + 2), vector(gtd_yy_p, i, j, k + 2), vector(gtd_yz_p, i, j, k + 2), vector(Betau_z_p, i, j, k + 2), vector(Alpha_p, i, j, k + 2), vector(TauT_xz, i, j, k + 2), vector(TauT_yz, i, j, k + 2), vector(TauT_zz, i, j, k + 2), inv_chi_p, sdetg_p, Sfud_zy_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2kSfd_z = FkSfd_z_mainI(vector(Sfd_z_p, i, j, k + 2), vector(gtd_xz_p, i, j, k + 2), vector(gtd_yz_p, i, j, k + 2), vector(gtd_zz_p, i, j, k + 2), vector(Betau_z_p, i, j, k + 2), vector(Alpha_p, i, j, k + 2), vector(TauT_xz, i, j, k + 2), vector(TauT_yz, i, j, k + 2), vector(TauT_zz, i, j, k + 2), inv_chi_p, sdetg_p, Sfud_zz_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2kBfu_x = FkBfu_x_mainI(vector(Bfu_x_p, i, j, k + 2), vector(Bfu_z_p, i, j, k + 2), vector(phif_p, i, j, k + 2), vector(Betau_x_p, i, j, k + 2), vector(Betau_z_p, i, j, k + 2), vector(Alpha_p, i, j, k + 2), vector(TauM_xz, i, j, k + 2), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2kBfu_y = FkBfu_y_mainI(vector(Bfu_y_p, i, j, k + 2), vector(Bfu_z_p, i, j, k + 2), vector(phif_p, i, j, k + 2), vector(Betau_y_p, i, j, k + 2), vector(Betau_z_p, i, j, k + 2), vector(Alpha_p, i, j, k + 2), vector(TauM_yz, i, j, k + 2), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2kBfu_z = FkBfu_z_mainI(vector(phif_p, i, j, k + 2), vector(Alpha_p, i, j, k + 2), chi_max_p, gtu_zz_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2kDf = FkDf_mainI(vector(Df_p, i, j, k + 2), vector(Betau_z_p, i, j, k + 2), vector(Alpha_p, i, j, k + 2), vector(TauN_z, i, j, k + 2), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2kDYf = FkDYf_mainI(vector(DYf_p, i, j, k + 2), vector(Betau_z_p, i, j, k + 2), vector(Alpha_p, i, j, k + 2), vector(TauNe_z, i, j, k + 2), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2ktauf = Fktauf_mainI(vector(Df_p, i, j, k + 2), vector(tauf_p, i, j, k + 2), vector(Betau_z_p, i, j, k + 2), vector(Alpha_p, i, j, k + 2), vector(TauN_z, i, j, k + 2), sdetg_p, Sfu_z_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp2kphif = Fkphif_mainI(vector(Bfu_z_p, i, j, k + 2), vector(phif_p, i, j, k + 2), vector(Betau_z_p, i, j, k + 2), vector(Alpha_p, i, j, k + 2), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j, k + 2) * (1.0 + vector(epsf, i, j, k + 2)) + vector(pf, i, j, k + 2);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j, k + 2) + sqca_p - vector(sqcs, i, j, k + 2) * sqca_p;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(Betau_z_p, i, j, k + 2)) + vector(Alpha_p, i, j, k + 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(Betau_z_p, i, j, k + 2)) + vector(Alpha_p, i, j, k + 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(Betau_z_p, i, j, k + 2)) + vector(Alpha_p, i, j, k + 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(Betau_z_p, i, j, k + 2)) + vector(Alpha_p, i, j, k + 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(Betau_z_p, i, j, k + 2)) + vector(Alpha_p, i, j, k + 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(Betau_z_p, i, j, k + 2)) + vector(Alpha_p, i, j, k + 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(Betau_z_p, i, j, k + 2)) + vector(Alpha_p, i, j, k + 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(Betau_z_p, i, j, k + 2)) + vector(Alpha_p, i, j, k + 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(Betau_z_p, i, j, k + 2)) + vector(Alpha_p, i, j, k + 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(Betau_z_p, i, j, k + 2)) + vector(Alpha_p, i, j, k + 2) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j, k + 3));
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j, k + 3) * vector(Bfu_x_p, i, j, k + 3) + vector(gtd_yy_p, i, j, k + 3) * vector(Bfu_y_p, i, j, k + 3) + vector(gtd_yz_p, i, j, k + 3) * vector(Bfu_z_p, i, j, k + 3));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j, k + 3) * vector(Bfu_x_p, i, j, k + 3) + vector(gtd_yz_p, i, j, k + 3) * vector(Bfu_y_p, i, j, k + 3) + vector(gtd_zz_p, i, j, k + 3) * vector(Bfu_z_p, i, j, k + 3));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j, k + 3) * vector(Bfu_x_p, i, j, k + 3) + vector(gtd_xy_p, i, j, k + 3) * vector(Bfu_y_p, i, j, k + 3) + vector(gtd_xz_p, i, j, k + 3) * vector(Bfu_z_p, i, j, k + 3));
						sqBf_p = vector(Bfu_x_p, i, j, k + 3) * Bfd_x_p + vector(Bfu_y_p, i, j, k + 3) * Bfd_y_p + vector(Bfu_z_p, i, j, k + 3) * Bfd_z_p;
						detgtd_p = vector(gtd_yy_p, i, j, k + 3) * vector(gtd_zz_p, i, j, k + 3) * vector(gtd_xx_p, i, j, k + 3) + (-vector(gtd_yy_p, i, j, k + 3) * (vector(gtd_xz_p, i, j, k + 3) * vector(gtd_xz_p, i, j, k + 3))) + (-vector(gtd_zz_p, i, j, k + 3) * (vector(gtd_xy_p, i, j, k + 3) * vector(gtd_xy_p, i, j, k + 3))) + (-(vector(gtd_yz_p, i, j, k + 3) * vector(gtd_yz_p, i, j, k + 3)) * vector(gtd_xx_p, i, j, k + 3)) + 2.0 * vector(gtd_yz_p, i, j, k + 3) * vector(gtd_xy_p, i, j, k + 3) * vector(gtd_xz_p, i, j, k + 3);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j, k + 3) * vector(gtd_yy_p, i, j, k + 3) - vector(gtd_xy_p, i, j, k + 3) * vector(gtd_xy_p, i, j, k + 3));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j, k + 3) * vector(gtd_zz_p, i, j, k + 3) - vector(gtd_yz_p, i, j, k + 3) * vector(gtd_yz_p, i, j, k + 3));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j, k + 3) * vector(gtd_yz_p, i, j, k + 3) - vector(gtd_yy_p, i, j, k + 3) * vector(gtd_xz_p, i, j, k + 3));
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j, k + 3) * vector(gtd_zz_p, i, j, k + 3)) + vector(gtd_yz_p, i, j, k + 3) * vector(gtd_xz_p, i, j, k + 3));
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j, k + 3) + gtu_xy_p * vector(Sfd_y_p, i, j, k + 3) + gtu_xz_p * vector(Sfd_z_p, i, j, k + 3));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j, k + 3) + gtu_xy_p * vector(vfd_y, i, j, k + 3) + gtu_xz_p * vector(vfd_z, i, j, k + 3));
						Bfvf_p = vector(Bfu_x_p, i, j, k + 3) * vector(vfd_x, i, j, k + 3) + vector(Bfu_y_p, i, j, k + 3) * vector(vfd_y, i, j, k + 3) + vector(Bfu_z_p, i, j, k + 3) * vector(vfd_z, i, j, k + 3);
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j, k + 3) * vector(gtd_zz_p, i, j, k + 3) - vector(gtd_xz_p, i, j, k + 3) * vector(gtd_xz_p, i, j, k + 3));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j, k + 3) * vector(gtd_yz_p, i, j, k + 3)) + vector(gtd_xy_p, i, j, k + 3) * vector(gtd_xz_p, i, j, k + 3));
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j, k + 3) + gtu_yz_p * vector(vfd_y, i, j, k + 3) + gtu_zz_p * vector(vfd_z, i, j, k + 3));
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j, k + 3) + gtu_yy_p * vector(Sfd_y_p, i, j, k + 3) + gtu_yz_p * vector(Sfd_z_p, i, j, k + 3));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j, k + 3) + gtu_yz_p * vector(Sfd_y_p, i, j, k + 3) + gtu_zz_p * vector(Sfd_z_p, i, j, k + 3));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j, k + 3) + gtu_yy_p * vector(vfd_y, i, j, k + 3) + gtu_yz_p * vector(vfd_z, i, j, k + 3));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j, k + 3) + vfu_y_p * vector(vfd_y, i, j, k + 3) + vfu_z_p * vector(vfd_z, i, j, k + 3), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j, k + 3) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j, k + 3) * vector(Bfu_z_p, i, j, k + 3)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j, k + 3) * vfu_z_p)) / sdetg_p);
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j, k + 3) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j, k + 3) * vector(Bfu_x_p, i, j, k + 3)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j, k + 3) * vfu_x_p)) / sdetg_p);
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j, k + 3) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k + 3) * vector(Bfu_z_p, i, j, k + 3)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k + 3) * vfu_z_p + vector(Bfu_z_p, i, j, k + 3) * vfu_x_p)) / sdetg_p);
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j, k + 3) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k + 3) * vector(Bfu_y_p, i, j, k + 3)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k + 3) * vfu_y_p + vector(Bfu_y_p, i, j, k + 3) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j, k + 3) + Sfuu_xy_p * vector(gtd_xy_p, i, j, k + 3) + Sfuu_xz_p * vector(gtd_xz_p, i, j, k + 3));
						Sfud_xy_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xy_p, i, j, k + 3) + Sfuu_xy_p * vector(gtd_yy_p, i, j, k + 3) + Sfuu_xz_p * vector(gtd_yz_p, i, j, k + 3));
						Sfud_xz_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xz_p, i, j, k + 3) + Sfuu_xy_p * vector(gtd_yz_p, i, j, k + 3) + Sfuu_xz_p * vector(gtd_zz_p, i, j, k + 3));
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j, k + 3) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k + 3) * vector(Bfu_z_p, i, j, k + 3)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k + 3) * vfu_z_p + vector(Bfu_z_p, i, j, k + 3) * vfu_y_p)) / sdetg_p);
						Sfud_zy_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xy_p, i, j, k + 3) + Sfuu_yz_p * vector(gtd_yy_p, i, j, k + 3) + Sfuu_zz_p * vector(gtd_yz_p, i, j, k + 3));
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j, k + 3) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k + 3) + Sfuu_zz_p * vector(gtd_zz_p, i, j, k + 3));
						Sfud_zx_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xx_p, i, j, k + 3) + Sfuu_yz_p * vector(gtd_xy_p, i, j, k + 3) + Sfuu_zz_p * vector(gtd_xz_p, i, j, k + 3));
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j, k + 3) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j, k + 3) * vector(Bfu_y_p, i, j, k + 3)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j, k + 3) * vfu_y_p)) / sdetg_p);
						Sfud_yx_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xx_p, i, j, k + 3) + Sfuu_yy_p * vector(gtd_xy_p, i, j, k + 3) + Sfuu_yz_p * vector(gtd_xz_p, i, j, k + 3));
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j, k + 3) + Sfuu_yy_p * vector(gtd_yy_p, i, j, k + 3) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k + 3));
						Sfud_yz_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xz_p, i, j, k + 3) + Sfuu_yy_p * vector(gtd_yz_p, i, j, k + 3) + Sfuu_yz_p * vector(gtd_zz_p, i, j, k + 3));
						Fluxp3kSfd_x = FkSfd_x_mainI(vector(Sfd_x_p, i, j, k + 3), vector(gtd_xx_p, i, j, k + 3), vector(gtd_xy_p, i, j, k + 3), vector(gtd_xz_p, i, j, k + 3), vector(Betau_z_p, i, j, k + 3), vector(Alpha_p, i, j, k + 3), vector(TauT_xz, i, j, k + 3), vector(TauT_yz, i, j, k + 3), vector(TauT_zz, i, j, k + 3), inv_chi_p, sdetg_p, Sfud_zx_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3kSfd_y = FkSfd_y_mainI(vector(Sfd_y_p, i, j, k + 3), vector(gtd_xy_p, i, j, k + 3), vector(gtd_yy_p, i, j, k + 3), vector(gtd_yz_p, i, j, k + 3), vector(Betau_z_p, i, j, k + 3), vector(Alpha_p, i, j, k + 3), vector(TauT_xz, i, j, k + 3), vector(TauT_yz, i, j, k + 3), vector(TauT_zz, i, j, k + 3), inv_chi_p, sdetg_p, Sfud_zy_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3kSfd_z = FkSfd_z_mainI(vector(Sfd_z_p, i, j, k + 3), vector(gtd_xz_p, i, j, k + 3), vector(gtd_yz_p, i, j, k + 3), vector(gtd_zz_p, i, j, k + 3), vector(Betau_z_p, i, j, k + 3), vector(Alpha_p, i, j, k + 3), vector(TauT_xz, i, j, k + 3), vector(TauT_yz, i, j, k + 3), vector(TauT_zz, i, j, k + 3), inv_chi_p, sdetg_p, Sfud_zz_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3kBfu_x = FkBfu_x_mainI(vector(Bfu_x_p, i, j, k + 3), vector(Bfu_z_p, i, j, k + 3), vector(phif_p, i, j, k + 3), vector(Betau_x_p, i, j, k + 3), vector(Betau_z_p, i, j, k + 3), vector(Alpha_p, i, j, k + 3), vector(TauM_xz, i, j, k + 3), chi_max_p, gtu_xz_p, sdetg_p, vfu_x_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3kBfu_y = FkBfu_y_mainI(vector(Bfu_y_p, i, j, k + 3), vector(Bfu_z_p, i, j, k + 3), vector(phif_p, i, j, k + 3), vector(Betau_y_p, i, j, k + 3), vector(Betau_z_p, i, j, k + 3), vector(Alpha_p, i, j, k + 3), vector(TauM_yz, i, j, k + 3), chi_max_p, gtu_yz_p, sdetg_p, vfu_y_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3kBfu_z = FkBfu_z_mainI(vector(phif_p, i, j, k + 3), vector(Alpha_p, i, j, k + 3), chi_max_p, gtu_zz_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3kDf = FkDf_mainI(vector(Df_p, i, j, k + 3), vector(Betau_z_p, i, j, k + 3), vector(Alpha_p, i, j, k + 3), vector(TauN_z, i, j, k + 3), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3kDYf = FkDYf_mainI(vector(DYf_p, i, j, k + 3), vector(Betau_z_p, i, j, k + 3), vector(Alpha_p, i, j, k + 3), vector(TauNe_z, i, j, k + 3), sdetg_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3ktauf = Fktauf_mainI(vector(Df_p, i, j, k + 3), vector(tauf_p, i, j, k + 3), vector(Betau_z_p, i, j, k + 3), vector(Alpha_p, i, j, k + 3), vector(TauN_z, i, j, k + 3), sdetg_p, Sfu_z_p, vfu_z_p, dx, simPlat_dt, ilast, jlast);
						Fluxp3kphif = Fkphif_mainI(vector(Bfu_z_p, i, j, k + 3), vector(phif_p, i, j, k + 3), vector(Betau_z_p, i, j, k + 3), vector(Alpha_p, i, j, k + 3), ch, dx, simPlat_dt, ilast, jlast);
						h_p = vector(rhof, i, j, k + 3) * (1.0 + vector(epsf, i, j, k + 3)) + vector(pf, i, j, k + 3);
						sqbc_p = sqBf_p / sqW_p + Bfvf_p * Bfvf_p;
						sqca_p = fabs(sqbc_p / (h_p + sqbc_p));
						sqa_p = vector(sqcs, i, j, k + 3) + sqca_p - vector(sqcs, i, j, k + 3) * sqca_p;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(Betau_z_p, i, j, k + 3)) + vector(Alpha_p, i, j, k + 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(Betau_z_p, i, j, k + 3)) + vector(Alpha_p, i, j, k + 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(Betau_z_p, i, j, k + 3)) + vector(Alpha_p, i, j, k + 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(Betau_z_p, i, j, k + 3)) + vector(Alpha_p, i, j, k + 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(Betau_z_p, i, j, k + 3)) + vector(Alpha_p, i, j, k + 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(Betau_z_p, i, j, k + 3)) + vector(Alpha_p, i, j, k + 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(Betau_z_p, i, j, k + 3)) + vector(Alpha_p, i, j, k + 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(Betau_z_p, i, j, k + 3)) + vector(Alpha_p, i, j, k + 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(Betau_z_p, i, j, k + 3)) + vector(Alpha_p, i, j, k + 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(Betau_z_p, i, j, k + 3)) + vector(Alpha_p, i, j, k + 3) / MAX(1.0 - sqvf_p * sqa_p, 0.0000000001) * (fabs((1.0 - sqa_p) * vfu_z_p) + sqrt(fabs(sqa_p * (1.0 - sqvf_p) * ((1.0 - sqvf_p * sqa_p) * chi_max_p * gtu_zz_p - (1.0 - sqa_p) * (vfu_z_p * vfu_z_p)))))));
						LLFpkSfd_x = 0.5 * (FluxkSfd_x + fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k));
						LLFpkSfd_y = 0.5 * (FluxkSfd_y + fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k));
						LLFpkSfd_z = 0.5 * (FluxkSfd_z + fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k));
						LLFpkBfu_x = 0.5 * (FluxkBfu_x + fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k));
						LLFpkBfu_y = 0.5 * (FluxkBfu_y + fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k));
						LLFpkBfu_z = 0.5 * (FluxkBfu_z + fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k));
						LLFpkDf = 0.5 * (FluxkDf + fabs(SpeedkDf) * vector(Df_p, i, j, k));
						LLFpkDYf = 0.5 * (FluxkDYf + fabs(SpeedkDYf) * vector(DYf_p, i, j, k));
						LLFpktauf = 0.5 * (Fluxktauf + fabs(Speedktauf) * vector(tauf_p, i, j, k));
						LLFpkphif = 0.5 * (Fluxkphif + fabs(Speedkphif) * vector(phif_p, i, j, k));
						LLFnkSfd_x = 0.5 * (FluxkSfd_x - fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k));
						LLFnkSfd_y = 0.5 * (FluxkSfd_y - fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k));
						LLFnkSfd_z = 0.5 * (FluxkSfd_z - fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k));
						LLFnkBfu_x = 0.5 * (FluxkBfu_x - fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k));
						LLFnkBfu_y = 0.5 * (FluxkBfu_y - fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k));
						LLFnkBfu_z = 0.5 * (FluxkBfu_z - fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k));
						LLFnkDf = 0.5 * (FluxkDf - fabs(SpeedkDf) * vector(Df_p, i, j, k));
						LLFnkDYf = 0.5 * (FluxkDYf - fabs(SpeedkDYf) * vector(DYf_p, i, j, k));
						LLFnktauf = 0.5 * (Fluxktauf - fabs(Speedktauf) * vector(tauf_p, i, j, k));
						LLFnkphif = 0.5 * (Fluxkphif - fabs(Speedkphif) * vector(phif_p, i, j, k));
						LLFp_m3kSfd_x = 0.5 * (Fluxm3kSfd_x + fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k - 3));
						LLFp_m3kSfd_y = 0.5 * (Fluxm3kSfd_y + fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k - 3));
						LLFp_m3kSfd_z = 0.5 * (Fluxm3kSfd_z + fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k - 3));
						LLFp_m3kBfu_x = 0.5 * (Fluxm3kBfu_x + fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k - 3));
						LLFp_m3kBfu_y = 0.5 * (Fluxm3kBfu_y + fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k - 3));
						LLFp_m3kBfu_z = 0.5 * (Fluxm3kBfu_z + fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k - 3));
						LLFp_m3kDf = 0.5 * (Fluxm3kDf + fabs(SpeedkDf) * vector(Df_p, i, j, k - 3));
						LLFp_m3kDYf = 0.5 * (Fluxm3kDYf + fabs(SpeedkDYf) * vector(DYf_p, i, j, k - 3));
						LLFp_m3ktauf = 0.5 * (Fluxm3ktauf + fabs(Speedktauf) * vector(tauf_p, i, j, k - 3));
						LLFp_m3kphif = 0.5 * (Fluxm3kphif + fabs(Speedkphif) * vector(phif_p, i, j, k - 3));
						LLFn_m3kSfd_x = 0.5 * (Fluxm3kSfd_x - fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k - 3));
						LLFn_m3kSfd_y = 0.5 * (Fluxm3kSfd_y - fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k - 3));
						LLFn_m3kSfd_z = 0.5 * (Fluxm3kSfd_z - fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k - 3));
						LLFn_m3kBfu_x = 0.5 * (Fluxm3kBfu_x - fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k - 3));
						LLFn_m3kBfu_y = 0.5 * (Fluxm3kBfu_y - fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k - 3));
						LLFn_m3kBfu_z = 0.5 * (Fluxm3kBfu_z - fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k - 3));
						LLFn_m3kDf = 0.5 * (Fluxm3kDf - fabs(SpeedkDf) * vector(Df_p, i, j, k - 3));
						LLFn_m3kDYf = 0.5 * (Fluxm3kDYf - fabs(SpeedkDYf) * vector(DYf_p, i, j, k - 3));
						LLFn_m3ktauf = 0.5 * (Fluxm3ktauf - fabs(Speedktauf) * vector(tauf_p, i, j, k - 3));
						LLFn_m3kphif = 0.5 * (Fluxm3kphif - fabs(Speedkphif) * vector(phif_p, i, j, k - 3));
						LLFp_m2kSfd_x = 0.5 * (Fluxm2kSfd_x + fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k - 2));
						LLFp_m2kSfd_y = 0.5 * (Fluxm2kSfd_y + fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k - 2));
						LLFp_m2kSfd_z = 0.5 * (Fluxm2kSfd_z + fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k - 2));
						LLFp_m2kBfu_x = 0.5 * (Fluxm2kBfu_x + fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k - 2));
						LLFp_m2kBfu_y = 0.5 * (Fluxm2kBfu_y + fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k - 2));
						LLFp_m2kBfu_z = 0.5 * (Fluxm2kBfu_z + fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k - 2));
						LLFp_m2kDf = 0.5 * (Fluxm2kDf + fabs(SpeedkDf) * vector(Df_p, i, j, k - 2));
						LLFp_m2kDYf = 0.5 * (Fluxm2kDYf + fabs(SpeedkDYf) * vector(DYf_p, i, j, k - 2));
						LLFp_m2ktauf = 0.5 * (Fluxm2ktauf + fabs(Speedktauf) * vector(tauf_p, i, j, k - 2));
						LLFp_m2kphif = 0.5 * (Fluxm2kphif + fabs(Speedkphif) * vector(phif_p, i, j, k - 2));
						LLFn_m2kSfd_x = 0.5 * (Fluxm2kSfd_x - fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k - 2));
						LLFn_m2kSfd_y = 0.5 * (Fluxm2kSfd_y - fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k - 2));
						LLFn_m2kSfd_z = 0.5 * (Fluxm2kSfd_z - fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k - 2));
						LLFn_m2kBfu_x = 0.5 * (Fluxm2kBfu_x - fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k - 2));
						LLFn_m2kBfu_y = 0.5 * (Fluxm2kBfu_y - fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k - 2));
						LLFn_m2kBfu_z = 0.5 * (Fluxm2kBfu_z - fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k - 2));
						LLFn_m2kDf = 0.5 * (Fluxm2kDf - fabs(SpeedkDf) * vector(Df_p, i, j, k - 2));
						LLFn_m2kDYf = 0.5 * (Fluxm2kDYf - fabs(SpeedkDYf) * vector(DYf_p, i, j, k - 2));
						LLFn_m2ktauf = 0.5 * (Fluxm2ktauf - fabs(Speedktauf) * vector(tauf_p, i, j, k - 2));
						LLFn_m2kphif = 0.5 * (Fluxm2kphif - fabs(Speedkphif) * vector(phif_p, i, j, k - 2));
						LLFp_m1kSfd_x = 0.5 * (Fluxm1kSfd_x + fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k - 1));
						LLFp_m1kSfd_y = 0.5 * (Fluxm1kSfd_y + fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k - 1));
						LLFp_m1kSfd_z = 0.5 * (Fluxm1kSfd_z + fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k - 1));
						LLFp_m1kBfu_x = 0.5 * (Fluxm1kBfu_x + fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k - 1));
						LLFp_m1kBfu_y = 0.5 * (Fluxm1kBfu_y + fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k - 1));
						LLFp_m1kBfu_z = 0.5 * (Fluxm1kBfu_z + fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k - 1));
						LLFp_m1kDf = 0.5 * (Fluxm1kDf + fabs(SpeedkDf) * vector(Df_p, i, j, k - 1));
						LLFp_m1kDYf = 0.5 * (Fluxm1kDYf + fabs(SpeedkDYf) * vector(DYf_p, i, j, k - 1));
						LLFp_m1ktauf = 0.5 * (Fluxm1ktauf + fabs(Speedktauf) * vector(tauf_p, i, j, k - 1));
						LLFp_m1kphif = 0.5 * (Fluxm1kphif + fabs(Speedkphif) * vector(phif_p, i, j, k - 1));
						LLFn_m1kSfd_x = 0.5 * (Fluxm1kSfd_x - fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k - 1));
						LLFn_m1kSfd_y = 0.5 * (Fluxm1kSfd_y - fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k - 1));
						LLFn_m1kSfd_z = 0.5 * (Fluxm1kSfd_z - fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k - 1));
						LLFn_m1kBfu_x = 0.5 * (Fluxm1kBfu_x - fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k - 1));
						LLFn_m1kBfu_y = 0.5 * (Fluxm1kBfu_y - fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k - 1));
						LLFn_m1kBfu_z = 0.5 * (Fluxm1kBfu_z - fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k - 1));
						LLFn_m1kDf = 0.5 * (Fluxm1kDf - fabs(SpeedkDf) * vector(Df_p, i, j, k - 1));
						LLFn_m1kDYf = 0.5 * (Fluxm1kDYf - fabs(SpeedkDYf) * vector(DYf_p, i, j, k - 1));
						LLFn_m1ktauf = 0.5 * (Fluxm1ktauf - fabs(Speedktauf) * vector(tauf_p, i, j, k - 1));
						LLFn_m1kphif = 0.5 * (Fluxm1kphif - fabs(Speedkphif) * vector(phif_p, i, j, k - 1));
						LLFp_p1kSfd_x = 0.5 * (Fluxp1kSfd_x + fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k + 1));
						LLFp_p1kSfd_y = 0.5 * (Fluxp1kSfd_y + fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k + 1));
						LLFp_p1kSfd_z = 0.5 * (Fluxp1kSfd_z + fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k + 1));
						LLFp_p1kBfu_x = 0.5 * (Fluxp1kBfu_x + fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k + 1));
						LLFp_p1kBfu_y = 0.5 * (Fluxp1kBfu_y + fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k + 1));
						LLFp_p1kBfu_z = 0.5 * (Fluxp1kBfu_z + fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k + 1));
						LLFp_p1kDf = 0.5 * (Fluxp1kDf + fabs(SpeedkDf) * vector(Df_p, i, j, k + 1));
						LLFp_p1kDYf = 0.5 * (Fluxp1kDYf + fabs(SpeedkDYf) * vector(DYf_p, i, j, k + 1));
						LLFp_p1ktauf = 0.5 * (Fluxp1ktauf + fabs(Speedktauf) * vector(tauf_p, i, j, k + 1));
						LLFp_p1kphif = 0.5 * (Fluxp1kphif + fabs(Speedkphif) * vector(phif_p, i, j, k + 1));
						LLFn_p1kSfd_x = 0.5 * (Fluxp1kSfd_x - fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k + 1));
						LLFn_p1kSfd_y = 0.5 * (Fluxp1kSfd_y - fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k + 1));
						LLFn_p1kSfd_z = 0.5 * (Fluxp1kSfd_z - fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k + 1));
						LLFn_p1kBfu_x = 0.5 * (Fluxp1kBfu_x - fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k + 1));
						LLFn_p1kBfu_y = 0.5 * (Fluxp1kBfu_y - fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k + 1));
						LLFn_p1kBfu_z = 0.5 * (Fluxp1kBfu_z - fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k + 1));
						LLFn_p1kDf = 0.5 * (Fluxp1kDf - fabs(SpeedkDf) * vector(Df_p, i, j, k + 1));
						LLFn_p1kDYf = 0.5 * (Fluxp1kDYf - fabs(SpeedkDYf) * vector(DYf_p, i, j, k + 1));
						LLFn_p1ktauf = 0.5 * (Fluxp1ktauf - fabs(Speedktauf) * vector(tauf_p, i, j, k + 1));
						LLFn_p1kphif = 0.5 * (Fluxp1kphif - fabs(Speedkphif) * vector(phif_p, i, j, k + 1));
						LLFp_p2kSfd_x = 0.5 * (Fluxp2kSfd_x + fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k + 2));
						LLFp_p2kSfd_y = 0.5 * (Fluxp2kSfd_y + fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k + 2));
						LLFp_p2kSfd_z = 0.5 * (Fluxp2kSfd_z + fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k + 2));
						LLFp_p2kBfu_x = 0.5 * (Fluxp2kBfu_x + fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k + 2));
						LLFp_p2kBfu_y = 0.5 * (Fluxp2kBfu_y + fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k + 2));
						LLFp_p2kBfu_z = 0.5 * (Fluxp2kBfu_z + fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k + 2));
						LLFp_p2kDf = 0.5 * (Fluxp2kDf + fabs(SpeedkDf) * vector(Df_p, i, j, k + 2));
						LLFp_p2kDYf = 0.5 * (Fluxp2kDYf + fabs(SpeedkDYf) * vector(DYf_p, i, j, k + 2));
						LLFp_p2ktauf = 0.5 * (Fluxp2ktauf + fabs(Speedktauf) * vector(tauf_p, i, j, k + 2));
						LLFp_p2kphif = 0.5 * (Fluxp2kphif + fabs(Speedkphif) * vector(phif_p, i, j, k + 2));
						LLFn_p2kSfd_x = 0.5 * (Fluxp2kSfd_x - fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k + 2));
						LLFn_p2kSfd_y = 0.5 * (Fluxp2kSfd_y - fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k + 2));
						LLFn_p2kSfd_z = 0.5 * (Fluxp2kSfd_z - fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k + 2));
						LLFn_p2kBfu_x = 0.5 * (Fluxp2kBfu_x - fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k + 2));
						LLFn_p2kBfu_y = 0.5 * (Fluxp2kBfu_y - fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k + 2));
						LLFn_p2kBfu_z = 0.5 * (Fluxp2kBfu_z - fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k + 2));
						LLFn_p2kDf = 0.5 * (Fluxp2kDf - fabs(SpeedkDf) * vector(Df_p, i, j, k + 2));
						LLFn_p2kDYf = 0.5 * (Fluxp2kDYf - fabs(SpeedkDYf) * vector(DYf_p, i, j, k + 2));
						LLFn_p2ktauf = 0.5 * (Fluxp2ktauf - fabs(Speedktauf) * vector(tauf_p, i, j, k + 2));
						LLFn_p2kphif = 0.5 * (Fluxp2kphif - fabs(Speedkphif) * vector(phif_p, i, j, k + 2));
						LLFp_p3kSfd_x = 0.5 * (Fluxp3kSfd_x + fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k + 3));
						LLFp_p3kSfd_y = 0.5 * (Fluxp3kSfd_y + fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k + 3));
						LLFp_p3kSfd_z = 0.5 * (Fluxp3kSfd_z + fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k + 3));
						LLFp_p3kBfu_x = 0.5 * (Fluxp3kBfu_x + fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k + 3));
						LLFp_p3kBfu_y = 0.5 * (Fluxp3kBfu_y + fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k + 3));
						LLFp_p3kBfu_z = 0.5 * (Fluxp3kBfu_z + fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k + 3));
						LLFp_p3kDf = 0.5 * (Fluxp3kDf + fabs(SpeedkDf) * vector(Df_p, i, j, k + 3));
						LLFp_p3kDYf = 0.5 * (Fluxp3kDYf + fabs(SpeedkDYf) * vector(DYf_p, i, j, k + 3));
						LLFp_p3ktauf = 0.5 * (Fluxp3ktauf + fabs(Speedktauf) * vector(tauf_p, i, j, k + 3));
						LLFp_p3kphif = 0.5 * (Fluxp3kphif + fabs(Speedkphif) * vector(phif_p, i, j, k + 3));
						LLFn_p3kSfd_x = 0.5 * (Fluxp3kSfd_x - fabs(SpeedkSfd_x) * vector(Sfd_x_p, i, j, k + 3));
						LLFn_p3kSfd_y = 0.5 * (Fluxp3kSfd_y - fabs(SpeedkSfd_y) * vector(Sfd_y_p, i, j, k + 3));
						LLFn_p3kSfd_z = 0.5 * (Fluxp3kSfd_z - fabs(SpeedkSfd_z) * vector(Sfd_z_p, i, j, k + 3));
						LLFn_p3kBfu_x = 0.5 * (Fluxp3kBfu_x - fabs(SpeedkBfu_x) * vector(Bfu_x_p, i, j, k + 3));
						LLFn_p3kBfu_y = 0.5 * (Fluxp3kBfu_y - fabs(SpeedkBfu_y) * vector(Bfu_y_p, i, j, k + 3));
						LLFn_p3kBfu_z = 0.5 * (Fluxp3kBfu_z - fabs(SpeedkBfu_z) * vector(Bfu_z_p, i, j, k + 3));
						LLFn_p3kDf = 0.5 * (Fluxp3kDf - fabs(SpeedkDf) * vector(Df_p, i, j, k + 3));
						LLFn_p3kDYf = 0.5 * (Fluxp3kDYf - fabs(SpeedkDYf) * vector(DYf_p, i, j, k + 3));
						LLFn_p3ktauf = 0.5 * (Fluxp3ktauf - fabs(Speedktauf) * vector(tauf_p, i, j, k + 3));
						LLFn_p3kphif = 0.5 * (Fluxp3kphif - fabs(Speedkphif) * vector(phif_p, i, j, k + 3));
						Sfd_x_lni = MP5(LLFp_m3iSfd_x, LLFp_m2iSfd_x, LLFp_m1iSfd_x, LLFpiSfd_x, LLFp_p1iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lni = MP5(LLFp_m3iSfd_y, LLFp_m2iSfd_y, LLFp_m1iSfd_y, LLFpiSfd_y, LLFp_p1iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lni = MP5(LLFp_m3iSfd_z, LLFp_m2iSfd_z, LLFp_m1iSfd_z, LLFpiSfd_z, LLFp_p1iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lni = MP5(LLFp_m3iBfu_x, LLFp_m2iBfu_x, LLFp_m1iBfu_x, LLFpiBfu_x, LLFp_p1iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lni = MP5(LLFp_m3iBfu_y, LLFp_m2iBfu_y, LLFp_m1iBfu_y, LLFpiBfu_y, LLFp_p1iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lni = MP5(LLFp_m3iBfu_z, LLFp_m2iBfu_z, LLFp_m1iBfu_z, LLFpiBfu_z, LLFp_p1iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lni = MP5(LLFp_m3iDf, LLFp_m2iDf, LLFp_m1iDf, LLFpiDf, LLFp_p1iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lni = MP5(LLFp_m3iDYf, LLFp_m2iDYf, LLFp_m1iDYf, LLFpiDYf, LLFp_p1iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lni = MP5(LLFp_m3itauf, LLFp_m2itauf, LLFp_m1itauf, LLFpitauf, LLFp_p1itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lni = MP5(LLFp_m3iphif, LLFp_m2iphif, LLFp_m1iphif, LLFpiphif, LLFp_p1iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lnj = MP5(LLFp_m3jSfd_x, LLFp_m2jSfd_x, LLFp_m1jSfd_x, LLFpjSfd_x, LLFp_p1jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lnj = MP5(LLFp_m3jSfd_y, LLFp_m2jSfd_y, LLFp_m1jSfd_y, LLFpjSfd_y, LLFp_p1jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lnj = MP5(LLFp_m3jSfd_z, LLFp_m2jSfd_z, LLFp_m1jSfd_z, LLFpjSfd_z, LLFp_p1jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lnj = MP5(LLFp_m3jBfu_x, LLFp_m2jBfu_x, LLFp_m1jBfu_x, LLFpjBfu_x, LLFp_p1jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lnj = MP5(LLFp_m3jBfu_y, LLFp_m2jBfu_y, LLFp_m1jBfu_y, LLFpjBfu_y, LLFp_p1jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lnj = MP5(LLFp_m3jBfu_z, LLFp_m2jBfu_z, LLFp_m1jBfu_z, LLFpjBfu_z, LLFp_p1jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lnj = MP5(LLFp_m3jDf, LLFp_m2jDf, LLFp_m1jDf, LLFpjDf, LLFp_p1jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lnj = MP5(LLFp_m3jDYf, LLFp_m2jDYf, LLFp_m1jDYf, LLFpjDYf, LLFp_p1jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lnj = MP5(LLFp_m3jtauf, LLFp_m2jtauf, LLFp_m1jtauf, LLFpjtauf, LLFp_p1jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lnj = MP5(LLFp_m3jphif, LLFp_m2jphif, LLFp_m1jphif, LLFpjphif, LLFp_p1jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lnk = MP5(LLFp_m3kSfd_x, LLFp_m2kSfd_x, LLFp_m1kSfd_x, LLFpkSfd_x, LLFp_p1kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lnk = MP5(LLFp_m3kSfd_y, LLFp_m2kSfd_y, LLFp_m1kSfd_y, LLFpkSfd_y, LLFp_p1kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lnk = MP5(LLFp_m3kSfd_z, LLFp_m2kSfd_z, LLFp_m1kSfd_z, LLFpkSfd_z, LLFp_p1kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lnk = MP5(LLFp_m3kBfu_x, LLFp_m2kBfu_x, LLFp_m1kBfu_x, LLFpkBfu_x, LLFp_p1kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lnk = MP5(LLFp_m3kBfu_y, LLFp_m2kBfu_y, LLFp_m1kBfu_y, LLFpkBfu_y, LLFp_p1kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lnk = MP5(LLFp_m3kBfu_z, LLFp_m2kBfu_z, LLFp_m1kBfu_z, LLFpkBfu_z, LLFp_p1kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lnk = MP5(LLFp_m3kDf, LLFp_m2kDf, LLFp_m1kDf, LLFpkDf, LLFp_p1kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lnk = MP5(LLFp_m3kDYf, LLFp_m2kDYf, LLFp_m1kDYf, LLFpkDYf, LLFp_p1kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lnk = MP5(LLFp_m3ktauf, LLFp_m2ktauf, LLFp_m1ktauf, LLFpktauf, LLFp_p1ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lnk = MP5(LLFp_m3kphif, LLFp_m2kphif, LLFp_m1kphif, LLFpkphif, LLFp_p1kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rni = MP5(LLFn_p2iSfd_x, LLFn_p1iSfd_x, LLFniSfd_x, LLFn_m1iSfd_x, LLFn_m2iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rni = MP5(LLFn_p2iSfd_y, LLFn_p1iSfd_y, LLFniSfd_y, LLFn_m1iSfd_y, LLFn_m2iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rni = MP5(LLFn_p2iSfd_z, LLFn_p1iSfd_z, LLFniSfd_z, LLFn_m1iSfd_z, LLFn_m2iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rni = MP5(LLFn_p2iBfu_x, LLFn_p1iBfu_x, LLFniBfu_x, LLFn_m1iBfu_x, LLFn_m2iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rni = MP5(LLFn_p2iBfu_y, LLFn_p1iBfu_y, LLFniBfu_y, LLFn_m1iBfu_y, LLFn_m2iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rni = MP5(LLFn_p2iBfu_z, LLFn_p1iBfu_z, LLFniBfu_z, LLFn_m1iBfu_z, LLFn_m2iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rni = MP5(LLFn_p2iDf, LLFn_p1iDf, LLFniDf, LLFn_m1iDf, LLFn_m2iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rni = MP5(LLFn_p2iDYf, LLFn_p1iDYf, LLFniDYf, LLFn_m1iDYf, LLFn_m2iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rni = MP5(LLFn_p2itauf, LLFn_p1itauf, LLFnitauf, LLFn_m1itauf, LLFn_m2itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rni = MP5(LLFn_p2iphif, LLFn_p1iphif, LLFniphif, LLFn_m1iphif, LLFn_m2iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rnj = MP5(LLFn_p2jSfd_x, LLFn_p1jSfd_x, LLFnjSfd_x, LLFn_m1jSfd_x, LLFn_m2jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rnj = MP5(LLFn_p2jSfd_y, LLFn_p1jSfd_y, LLFnjSfd_y, LLFn_m1jSfd_y, LLFn_m2jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rnj = MP5(LLFn_p2jSfd_z, LLFn_p1jSfd_z, LLFnjSfd_z, LLFn_m1jSfd_z, LLFn_m2jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rnj = MP5(LLFn_p2jBfu_x, LLFn_p1jBfu_x, LLFnjBfu_x, LLFn_m1jBfu_x, LLFn_m2jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rnj = MP5(LLFn_p2jBfu_y, LLFn_p1jBfu_y, LLFnjBfu_y, LLFn_m1jBfu_y, LLFn_m2jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rnj = MP5(LLFn_p2jBfu_z, LLFn_p1jBfu_z, LLFnjBfu_z, LLFn_m1jBfu_z, LLFn_m2jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rnj = MP5(LLFn_p2jDf, LLFn_p1jDf, LLFnjDf, LLFn_m1jDf, LLFn_m2jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rnj = MP5(LLFn_p2jDYf, LLFn_p1jDYf, LLFnjDYf, LLFn_m1jDYf, LLFn_m2jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rnj = MP5(LLFn_p2jtauf, LLFn_p1jtauf, LLFnjtauf, LLFn_m1jtauf, LLFn_m2jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rnj = MP5(LLFn_p2jphif, LLFn_p1jphif, LLFnjphif, LLFn_m1jphif, LLFn_m2jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rnk = MP5(LLFn_p2kSfd_x, LLFn_p1kSfd_x, LLFnkSfd_x, LLFn_m1kSfd_x, LLFn_m2kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rnk = MP5(LLFn_p2kSfd_y, LLFn_p1kSfd_y, LLFnkSfd_y, LLFn_m1kSfd_y, LLFn_m2kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rnk = MP5(LLFn_p2kSfd_z, LLFn_p1kSfd_z, LLFnkSfd_z, LLFn_m1kSfd_z, LLFn_m2kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rnk = MP5(LLFn_p2kBfu_x, LLFn_p1kBfu_x, LLFnkBfu_x, LLFn_m1kBfu_x, LLFn_m2kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rnk = MP5(LLFn_p2kBfu_y, LLFn_p1kBfu_y, LLFnkBfu_y, LLFn_m1kBfu_y, LLFn_m2kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rnk = MP5(LLFn_p2kBfu_z, LLFn_p1kBfu_z, LLFnkBfu_z, LLFn_m1kBfu_z, LLFn_m2kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rnk = MP5(LLFn_p2kDf, LLFn_p1kDf, LLFnkDf, LLFn_m1kDf, LLFn_m2kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rnk = MP5(LLFn_p2kDYf, LLFn_p1kDYf, LLFnkDYf, LLFn_m1kDYf, LLFn_m2kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rnk = MP5(LLFn_p2ktauf, LLFn_p1ktauf, LLFnktauf, LLFn_m1ktauf, LLFn_m2ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rnk = MP5(LLFn_p2kphif, LLFn_p1kphif, LLFnkphif, LLFn_m1kphif, LLFn_m2kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lpi = MP5(LLFp_m2iSfd_x, LLFp_m1iSfd_x, LLFpiSfd_x, LLFp_p1iSfd_x, LLFp_p2iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lpi = MP5(LLFp_m2iSfd_y, LLFp_m1iSfd_y, LLFpiSfd_y, LLFp_p1iSfd_y, LLFp_p2iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lpi = MP5(LLFp_m2iSfd_z, LLFp_m1iSfd_z, LLFpiSfd_z, LLFp_p1iSfd_z, LLFp_p2iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lpi = MP5(LLFp_m2iBfu_x, LLFp_m1iBfu_x, LLFpiBfu_x, LLFp_p1iBfu_x, LLFp_p2iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lpi = MP5(LLFp_m2iBfu_y, LLFp_m1iBfu_y, LLFpiBfu_y, LLFp_p1iBfu_y, LLFp_p2iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lpi = MP5(LLFp_m2iBfu_z, LLFp_m1iBfu_z, LLFpiBfu_z, LLFp_p1iBfu_z, LLFp_p2iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lpi = MP5(LLFp_m2iDf, LLFp_m1iDf, LLFpiDf, LLFp_p1iDf, LLFp_p2iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lpi = MP5(LLFp_m2iDYf, LLFp_m1iDYf, LLFpiDYf, LLFp_p1iDYf, LLFp_p2iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lpi = MP5(LLFp_m2itauf, LLFp_m1itauf, LLFpitauf, LLFp_p1itauf, LLFp_p2itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lpi = MP5(LLFp_m2iphif, LLFp_m1iphif, LLFpiphif, LLFp_p1iphif, LLFp_p2iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lpj = MP5(LLFp_m2jSfd_x, LLFp_m1jSfd_x, LLFpjSfd_x, LLFp_p1jSfd_x, LLFp_p2jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lpj = MP5(LLFp_m2jSfd_y, LLFp_m1jSfd_y, LLFpjSfd_y, LLFp_p1jSfd_y, LLFp_p2jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lpj = MP5(LLFp_m2jSfd_z, LLFp_m1jSfd_z, LLFpjSfd_z, LLFp_p1jSfd_z, LLFp_p2jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lpj = MP5(LLFp_m2jBfu_x, LLFp_m1jBfu_x, LLFpjBfu_x, LLFp_p1jBfu_x, LLFp_p2jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lpj = MP5(LLFp_m2jBfu_y, LLFp_m1jBfu_y, LLFpjBfu_y, LLFp_p1jBfu_y, LLFp_p2jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lpj = MP5(LLFp_m2jBfu_z, LLFp_m1jBfu_z, LLFpjBfu_z, LLFp_p1jBfu_z, LLFp_p2jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lpj = MP5(LLFp_m2jDf, LLFp_m1jDf, LLFpjDf, LLFp_p1jDf, LLFp_p2jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lpj = MP5(LLFp_m2jDYf, LLFp_m1jDYf, LLFpjDYf, LLFp_p1jDYf, LLFp_p2jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lpj = MP5(LLFp_m2jtauf, LLFp_m1jtauf, LLFpjtauf, LLFp_p1jtauf, LLFp_p2jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lpj = MP5(LLFp_m2jphif, LLFp_m1jphif, LLFpjphif, LLFp_p1jphif, LLFp_p2jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lpk = MP5(LLFp_m2kSfd_x, LLFp_m1kSfd_x, LLFpkSfd_x, LLFp_p1kSfd_x, LLFp_p2kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lpk = MP5(LLFp_m2kSfd_y, LLFp_m1kSfd_y, LLFpkSfd_y, LLFp_p1kSfd_y, LLFp_p2kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lpk = MP5(LLFp_m2kSfd_z, LLFp_m1kSfd_z, LLFpkSfd_z, LLFp_p1kSfd_z, LLFp_p2kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lpk = MP5(LLFp_m2kBfu_x, LLFp_m1kBfu_x, LLFpkBfu_x, LLFp_p1kBfu_x, LLFp_p2kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lpk = MP5(LLFp_m2kBfu_y, LLFp_m1kBfu_y, LLFpkBfu_y, LLFp_p1kBfu_y, LLFp_p2kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lpk = MP5(LLFp_m2kBfu_z, LLFp_m1kBfu_z, LLFpkBfu_z, LLFp_p1kBfu_z, LLFp_p2kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lpk = MP5(LLFp_m2kDf, LLFp_m1kDf, LLFpkDf, LLFp_p1kDf, LLFp_p2kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lpk = MP5(LLFp_m2kDYf, LLFp_m1kDYf, LLFpkDYf, LLFp_p1kDYf, LLFp_p2kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lpk = MP5(LLFp_m2ktauf, LLFp_m1ktauf, LLFpktauf, LLFp_p1ktauf, LLFp_p2ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lpk = MP5(LLFp_m2kphif, LLFp_m1kphif, LLFpkphif, LLFp_p1kphif, LLFp_p2kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rpi = MP5(LLFn_p3iSfd_x, LLFn_p2iSfd_x, LLFn_p1iSfd_x, LLFniSfd_x, LLFn_m1iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rpi = MP5(LLFn_p3iSfd_y, LLFn_p2iSfd_y, LLFn_p1iSfd_y, LLFniSfd_y, LLFn_m1iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rpi = MP5(LLFn_p3iSfd_z, LLFn_p2iSfd_z, LLFn_p1iSfd_z, LLFniSfd_z, LLFn_m1iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rpi = MP5(LLFn_p3iBfu_x, LLFn_p2iBfu_x, LLFn_p1iBfu_x, LLFniBfu_x, LLFn_m1iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rpi = MP5(LLFn_p3iBfu_y, LLFn_p2iBfu_y, LLFn_p1iBfu_y, LLFniBfu_y, LLFn_m1iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rpi = MP5(LLFn_p3iBfu_z, LLFn_p2iBfu_z, LLFn_p1iBfu_z, LLFniBfu_z, LLFn_m1iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rpi = MP5(LLFn_p3iDf, LLFn_p2iDf, LLFn_p1iDf, LLFniDf, LLFn_m1iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rpi = MP5(LLFn_p3iDYf, LLFn_p2iDYf, LLFn_p1iDYf, LLFniDYf, LLFn_m1iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rpi = MP5(LLFn_p3itauf, LLFn_p2itauf, LLFn_p1itauf, LLFnitauf, LLFn_m1itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rpi = MP5(LLFn_p3iphif, LLFn_p2iphif, LLFn_p1iphif, LLFniphif, LLFn_m1iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rpj = MP5(LLFn_p3jSfd_x, LLFn_p2jSfd_x, LLFn_p1jSfd_x, LLFnjSfd_x, LLFn_m1jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rpj = MP5(LLFn_p3jSfd_y, LLFn_p2jSfd_y, LLFn_p1jSfd_y, LLFnjSfd_y, LLFn_m1jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rpj = MP5(LLFn_p3jSfd_z, LLFn_p2jSfd_z, LLFn_p1jSfd_z, LLFnjSfd_z, LLFn_m1jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rpj = MP5(LLFn_p3jBfu_x, LLFn_p2jBfu_x, LLFn_p1jBfu_x, LLFnjBfu_x, LLFn_m1jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rpj = MP5(LLFn_p3jBfu_y, LLFn_p2jBfu_y, LLFn_p1jBfu_y, LLFnjBfu_y, LLFn_m1jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rpj = MP5(LLFn_p3jBfu_z, LLFn_p2jBfu_z, LLFn_p1jBfu_z, LLFnjBfu_z, LLFn_m1jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rpj = MP5(LLFn_p3jDf, LLFn_p2jDf, LLFn_p1jDf, LLFnjDf, LLFn_m1jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rpj = MP5(LLFn_p3jDYf, LLFn_p2jDYf, LLFn_p1jDYf, LLFnjDYf, LLFn_m1jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rpj = MP5(LLFn_p3jtauf, LLFn_p2jtauf, LLFn_p1jtauf, LLFnjtauf, LLFn_m1jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rpj = MP5(LLFn_p3jphif, LLFn_p2jphif, LLFn_p1jphif, LLFnjphif, LLFn_m1jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rpk = MP5(LLFn_p3kSfd_x, LLFn_p2kSfd_x, LLFn_p1kSfd_x, LLFnkSfd_x, LLFn_m1kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rpk = MP5(LLFn_p3kSfd_y, LLFn_p2kSfd_y, LLFn_p1kSfd_y, LLFnkSfd_y, LLFn_m1kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rpk = MP5(LLFn_p3kSfd_z, LLFn_p2kSfd_z, LLFn_p1kSfd_z, LLFnkSfd_z, LLFn_m1kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rpk = MP5(LLFn_p3kBfu_x, LLFn_p2kBfu_x, LLFn_p1kBfu_x, LLFnkBfu_x, LLFn_m1kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rpk = MP5(LLFn_p3kBfu_y, LLFn_p2kBfu_y, LLFn_p1kBfu_y, LLFnkBfu_y, LLFn_m1kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rpk = MP5(LLFn_p3kBfu_z, LLFn_p2kBfu_z, LLFn_p1kBfu_z, LLFnkBfu_z, LLFn_m1kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rpk = MP5(LLFn_p3kDf, LLFn_p2kDf, LLFn_p1kDf, LLFnkDf, LLFn_m1kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rpk = MP5(LLFn_p3kDYf, LLFn_p2kDYf, LLFn_p1kDYf, LLFnkDYf, LLFn_m1kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rpk = MP5(LLFn_p3ktauf, LLFn_p2ktauf, LLFn_p1ktauf, LLFnktauf, LLFn_m1ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rpk = MP5(LLFn_p3kphif, LLFn_p2kphif, LLFn_p1kphif, LLFnkphif, LLFn_m1kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						FluxniSfd_x = Sfd_x_lni + Sfd_x_rni;
						FluxniSfd_y = Sfd_y_lni + Sfd_y_rni;
						FluxniSfd_z = Sfd_z_lni + Sfd_z_rni;
						FluxniBfu_x = Bfu_x_lni + Bfu_x_rni;
						FluxniBfu_y = Bfu_y_lni + Bfu_y_rni;
						FluxniBfu_z = Bfu_z_lni + Bfu_z_rni;
						FluxniDf = Df_lni + Df_rni;
						FluxniDYf = DYf_lni + DYf_rni;
						Fluxnitauf = tauf_lni + tauf_rni;
						Fluxniphif = phif_lni + phif_rni;
						FluxnjSfd_x = Sfd_x_lnj + Sfd_x_rnj;
						FluxnjSfd_y = Sfd_y_lnj + Sfd_y_rnj;
						FluxnjSfd_z = Sfd_z_lnj + Sfd_z_rnj;
						FluxnjBfu_x = Bfu_x_lnj + Bfu_x_rnj;
						FluxnjBfu_y = Bfu_y_lnj + Bfu_y_rnj;
						FluxnjBfu_z = Bfu_z_lnj + Bfu_z_rnj;
						FluxnjDf = Df_lnj + Df_rnj;
						FluxnjDYf = DYf_lnj + DYf_rnj;
						Fluxnjtauf = tauf_lnj + tauf_rnj;
						Fluxnjphif = phif_lnj + phif_rnj;
						FluxnkSfd_x = Sfd_x_lnk + Sfd_x_rnk;
						FluxnkSfd_y = Sfd_y_lnk + Sfd_y_rnk;
						FluxnkSfd_z = Sfd_z_lnk + Sfd_z_rnk;
						FluxnkBfu_x = Bfu_x_lnk + Bfu_x_rnk;
						FluxnkBfu_y = Bfu_y_lnk + Bfu_y_rnk;
						FluxnkBfu_z = Bfu_z_lnk + Bfu_z_rnk;
						FluxnkDf = Df_lnk + Df_rnk;
						FluxnkDYf = DYf_lnk + DYf_rnk;
						Fluxnktauf = tauf_lnk + tauf_rnk;
						Fluxnkphif = phif_lnk + phif_rnk;
						FluxpiSfd_x = Sfd_x_lpi + Sfd_x_rpi;
						FluxpiSfd_y = Sfd_y_lpi + Sfd_y_rpi;
						FluxpiSfd_z = Sfd_z_lpi + Sfd_z_rpi;
						FluxpiBfu_x = Bfu_x_lpi + Bfu_x_rpi;
						FluxpiBfu_y = Bfu_y_lpi + Bfu_y_rpi;
						FluxpiBfu_z = Bfu_z_lpi + Bfu_z_rpi;
						FluxpiDf = Df_lpi + Df_rpi;
						FluxpiDYf = DYf_lpi + DYf_rpi;
						Fluxpitauf = tauf_lpi + tauf_rpi;
						Fluxpiphif = phif_lpi + phif_rpi;
						FluxpjSfd_x = Sfd_x_lpj + Sfd_x_rpj;
						FluxpjSfd_y = Sfd_y_lpj + Sfd_y_rpj;
						FluxpjSfd_z = Sfd_z_lpj + Sfd_z_rpj;
						FluxpjBfu_x = Bfu_x_lpj + Bfu_x_rpj;
						FluxpjBfu_y = Bfu_y_lpj + Bfu_y_rpj;
						FluxpjBfu_z = Bfu_z_lpj + Bfu_z_rpj;
						FluxpjDf = Df_lpj + Df_rpj;
						FluxpjDYf = DYf_lpj + DYf_rpj;
						Fluxpjtauf = tauf_lpj + tauf_rpj;
						Fluxpjphif = phif_lpj + phif_rpj;
						FluxpkSfd_x = Sfd_x_lpk + Sfd_x_rpk;
						FluxpkSfd_y = Sfd_y_lpk + Sfd_y_rpk;
						FluxpkSfd_z = Sfd_z_lpk + Sfd_z_rpk;
						FluxpkBfu_x = Bfu_x_lpk + Bfu_x_rpk;
						FluxpkBfu_y = Bfu_y_lpk + Bfu_y_rpk;
						FluxpkBfu_z = Bfu_z_lpk + Bfu_z_rpk;
						FluxpkDf = Df_lpk + Df_rpk;
						FluxpkDYf = DYf_lpk + DYf_rpk;
						Fluxpktauf = tauf_lpk + tauf_rpk;
						Fluxpkphif = phif_lpk + phif_rpk;
						d_div_Beta_o0_t0_m0_l0 = D1CDO4_i(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o1_t0_m0_l0 = lieforward_i(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o2_t0_m0_l0 = liebackward_i(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_y_o0_t7_m0_l0 = D1CDO4_j(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o1_t1_m0_l0 = lieforward_j(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o2_t1_m0_l0 = liebackward_j(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_z_o0_t7_m0_l0 = D1CDO4_k(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o1_t2_m0_l0 = lieforward_k(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o2_t2_m0_l0 = liebackward_k(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_x_o0_t0_m0_l0 = D2CDO4_i(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_y_o0_t0_m0_l0 = D1CDO4crossed_ij(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_z_o0_t0_m0_l0 = D1CDO4crossed_ik(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t15_m0_l0 = D2CDO4_j(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t16_m0_l0 = D1CDO4crossed_jk(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t17_m0_l0 = D2CDO4_k(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_x_o0_t8_m0_l0 = D1CDO4_i(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o1_t0_m0_l0 = lieforward_i(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o2_t0_m0_l0 = liebackward_i(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_div_Beta_o0_t1_m0_l0 = D1CDO4_j(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o1_t1_m0_l0 = lieforward_j(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o2_t1_m0_l0 = liebackward_j(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_z_o0_t8_m0_l0 = D1CDO4_k(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o1_t2_m0_l0 = lieforward_k(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o2_t2_m0_l0 = liebackward_k(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o0_t6_m0_l0 = D2CDO4_i(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_x_o0_t1_m0_l0 = D1CDO4crossed_ij(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o0_t8_m0_l0 = D1CDO4crossed_ik(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_y_o0_t1_m0_l0 = D2CDO4_j(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_z_o0_t1_m0_l0 = D1CDO4crossed_jk(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o0_t11_m0_l0 = D2CDO4_k(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_x_o0_t9_m0_l0 = D1CDO4_i(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o1_t0_m0_l0 = lieforward_i(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o2_t0_m0_l0 = liebackward_i(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_y_o0_t9_m0_l0 = D1CDO4_j(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o1_t1_m0_l0 = lieforward_j(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o2_t1_m0_l0 = liebackward_j(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_div_Beta_o0_t2_m0_l0 = D1CDO4_k(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o1_t2_m0_l0 = lieforward_k(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o2_t2_m0_l0 = liebackward_k(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o0_t0_m0_l0 = D2CDO4_i(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o0_t1_m0_l0 = D1CDO4crossed_ij(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_x_o0_t2_m0_l0 = D1CDO4crossed_ik(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o0_t3_m0_l0 = D2CDO4_j(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_y_o0_t2_m0_l0 = D1CDO4crossed_jk(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_z_o0_t2_m0_l0 = D2CDO4_k(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xxx_o0_t0_m0_l0 = D1CDO4_i(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o1_t0_m0_l0 = lieforward_i(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o2_t0_m0_l0 = liebackward_i(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xxy_o0_t0_m0_l0 = D1CDO4_j(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o1_t1_m0_l0 = lieforward_j(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o2_t1_m0_l0 = liebackward_j(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xxz_o0_t0_m0_l0 = D1CDO4_k(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o1_t2_m0_l0 = lieforward_k(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o2_t2_m0_l0 = liebackward_k(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yxx_o0_t0_m0_l0 = D1CDO4_i(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o1_t0_m0_l0 = lieforward_i(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o2_t0_m0_l0 = liebackward_i(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyy_o0_t0_m0_l0 = D1CDO4_j(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o1_t1_m0_l0 = lieforward_j(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o2_t1_m0_l0 = liebackward_j(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyz_o0_t1_m0_l0 = D1CDO4_k(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o1_t2_m0_l0 = lieforward_k(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o2_t2_m0_l0 = liebackward_k(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyy_o0_t1_m0_l0 = D1CDO4_i(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o1_t0_m0_l0 = lieforward_i(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o2_t0_m0_l0 = liebackward_i(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yyy_o0_t0_m0_l0 = D1CDO4_j(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o1_t1_m0_l0 = lieforward_j(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o2_t1_m0_l0 = liebackward_j(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yyz_o0_t0_m0_l0 = D1CDO4_k(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o1_t2_m0_l0 = lieforward_k(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o2_t2_m0_l0 = liebackward_k(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_zxx_o0_t0_m0_l0 = D1CDO4_i(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o1_t0_m0_l0 = lieforward_i(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o2_t0_m0_l0 = liebackward_i(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyz_o0_t0_m0_l0 = D1CDO4_j(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o1_t1_m0_l0 = lieforward_j(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o2_t1_m0_l0 = liebackward_j(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xzz_o0_t0_m0_l0 = D1CDO4_k(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o1_t2_m0_l0 = lieforward_k(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o2_t2_m0_l0 = liebackward_k(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyz_o0_t2_m0_l0 = D1CDO4_i(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o1_t0_m0_l0 = lieforward_i(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o2_t0_m0_l0 = liebackward_i(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_zyy_o0_t0_m0_l0 = D1CDO4_j(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o1_t1_m0_l0 = lieforward_j(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o2_t1_m0_l0 = liebackward_j(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yzz_o0_t0_m0_l0 = D1CDO4_k(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o1_t2_m0_l0 = lieforward_k(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o2_t2_m0_l0 = liebackward_k(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xzz_o0_t1_m0_l0 = D1CDO4_i(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o1_t0_m0_l0 = lieforward_i(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o2_t0_m0_l0 = liebackward_i(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yzz_o0_t1_m0_l0 = D1CDO4_j(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o1_t1_m0_l0 = lieforward_j(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o2_t1_m0_l0 = liebackward_j(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_zzz_o0_t0_m0_l0 = D1CDO4_k(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o1_t2_m0_l0 = lieforward_k(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o2_t2_m0_l0 = liebackward_k(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t0_m0_l0 = D1CDO4_i(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o1_t0_m0_l0 = lieforward_i(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o2_t0_m0_l0 = liebackward_i(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t1_m0_l0 = D1CDO4_j(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o1_t1_m0_l0 = lieforward_j(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o2_t1_m0_l0 = liebackward_j(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t2_m0_l0 = D1CDO4_k(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o1_t2_m0_l0 = lieforward_k(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o2_t2_m0_l0 = liebackward_k(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t6_m0_l0 = D1CDO4_i(Gamh_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o1_t0_m0_l0 = liebackward_i(Gamh_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o2_t0_m0_l0 = lieforward_i(Gamh_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t6_m0_l0 = D1CDO4_j(Gamh_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o1_t1_m0_l0 = liebackward_j(Gamh_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o2_t1_m0_l0 = lieforward_j(Gamh_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t6_m0_l0 = D1CDO4_k(Gamh_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o1_t2_m0_l0 = liebackward_k(Gamh_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o2_t2_m0_l0 = lieforward_k(Gamh_x_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t7_m0_l0 = D1CDO4_i(Gamh_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o1_t0_m0_l0 = liebackward_i(Gamh_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o2_t0_m0_l0 = lieforward_i(Gamh_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t7_m0_l0 = D1CDO4_j(Gamh_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o1_t1_m0_l0 = liebackward_j(Gamh_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o2_t1_m0_l0 = lieforward_j(Gamh_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t7_m0_l0 = D1CDO4_k(Gamh_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o1_t2_m0_l0 = liebackward_k(Gamh_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o2_t2_m0_l0 = lieforward_k(Gamh_y_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t8_m0_l0 = D1CDO4_i(Gamh_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o1_t0_m0_l0 = liebackward_i(Gamh_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o2_t0_m0_l0 = lieforward_i(Gamh_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t8_m0_l0 = D1CDO4_j(Gamh_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o1_t1_m0_l0 = liebackward_j(Gamh_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o2_t1_m0_l0 = lieforward_j(Gamh_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t8_m0_l0 = D1CDO4_k(Gamh_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o1_t2_m0_l0 = liebackward_k(Gamh_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o2_t2_m0_l0 = lieforward_k(Gamh_z_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t1_m0_l0 = D1CDO4_i(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o1_t0_m0_l0 = lieforward_i(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o2_t0_m0_l0 = liebackward_i(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t2_m0_l0 = D1CDO4_j(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o1_t1_m0_l0 = lieforward_j(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o2_t1_m0_l0 = liebackward_j(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t3_m0_l0 = D1CDO4_k(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o1_t2_m0_l0 = lieforward_k(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o2_t2_m0_l0 = liebackward_k(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t0_m0_l0 = D2CDO4_i(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xy_o0_t0_m0_l0 = D1CDO4crossed_ij(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xz_o0_t0_m0_l0 = D1CDO4crossed_ik(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_yy_o0_t0_m0_l0 = D2CDO4_j(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_yz_o0_t0_m0_l0 = D1CDO4crossed_jk(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_zz_o0_t0_m0_l0 = D2CDO4_k(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o1_t0_m0_l0 = lieforward_i(Atd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o2_t0_m0_l0 = liebackward_i(Atd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o1_t1_m0_l0 = lieforward_j(Atd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o2_t1_m0_l0 = liebackward_j(Atd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o1_t2_m0_l0 = lieforward_k(Atd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o2_t2_m0_l0 = liebackward_k(Atd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o1_t0_m0_l0 = lieforward_i(Atd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o2_t0_m0_l0 = liebackward_i(Atd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o1_t1_m0_l0 = lieforward_j(Atd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o2_t1_m0_l0 = liebackward_j(Atd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o1_t2_m0_l0 = lieforward_k(Atd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o2_t2_m0_l0 = liebackward_k(Atd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o1_t0_m0_l0 = lieforward_i(Atd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o2_t0_m0_l0 = liebackward_i(Atd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o1_t1_m0_l0 = lieforward_j(Atd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o2_t1_m0_l0 = liebackward_j(Atd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o1_t2_m0_l0 = lieforward_k(Atd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o2_t2_m0_l0 = liebackward_k(Atd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o1_t0_m0_l0 = lieforward_i(Atd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o2_t0_m0_l0 = liebackward_i(Atd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o1_t1_m0_l0 = lieforward_j(Atd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o2_t1_m0_l0 = liebackward_j(Atd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o1_t2_m0_l0 = lieforward_k(Atd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o2_t2_m0_l0 = liebackward_k(Atd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o1_t0_m0_l0 = lieforward_i(Atd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o2_t0_m0_l0 = liebackward_i(Atd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o1_t1_m0_l0 = lieforward_j(Atd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o2_t1_m0_l0 = liebackward_j(Atd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o1_t2_m0_l0 = lieforward_k(Atd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o2_t2_m0_l0 = liebackward_k(Atd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o1_t0_m0_l0 = lieforward_i(Atd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o2_t0_m0_l0 = liebackward_i(Atd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o1_t1_m0_l0 = lieforward_j(Atd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o2_t1_m0_l0 = liebackward_j(Atd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o1_t2_m0_l0 = lieforward_k(Atd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o2_t2_m0_l0 = liebackward_k(Atd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t0_m0_l0 = D1CDO4_i(trK_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o1_t0_m0_l0 = liebackward_i(trK_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o2_t0_m0_l0 = lieforward_i(trK_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t1_m0_l0 = D1CDO4_j(trK_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o1_t1_m0_l0 = liebackward_j(trK_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o2_t1_m0_l0 = lieforward_j(trK_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t2_m0_l0 = D1CDO4_k(trK_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o1_t2_m0_l0 = liebackward_k(trK_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o2_t2_m0_l0 = lieforward_k(trK_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t3_m0_l0 = D1CDO4_i(theta_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o1_t0_m0_l0 = lieforward_i(theta_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o2_t0_m0_l0 = liebackward_i(theta_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t4_m0_l0 = D1CDO4_j(theta_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o1_t1_m0_l0 = lieforward_j(theta_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o2_t1_m0_l0 = liebackward_j(theta_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t5_m0_l0 = D1CDO4_k(theta_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o1_t2_m0_l0 = lieforward_k(theta_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o2_t2_m0_l0 = liebackward_k(theta_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t0_m0_l0 = D2CDO4_i(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t3_m0_l0 = D2CDO4_j(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t5_m0_l0 = D2CDO4_k(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t0_m0_l0 = D2CDO4_i(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t3_m0_l0 = D2CDO4_j(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t5_m0_l0 = D2CDO4_k(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t0_m0_l0 = D2CDO4_i(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t3_m0_l0 = D2CDO4_j(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t5_m0_l0 = D2CDO4_k(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t0_m0_l0 = D2CDO4_i(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t3_m0_l0 = D2CDO4_j(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t5_m0_l0 = D2CDO4_k(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t0_m0_l0 = D2CDO4_i(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t3_m0_l0 = D2CDO4_j(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t5_m0_l0 = D2CDO4_k(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t0_m0_l0 = D2CDO4_i(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t3_m0_l0 = D2CDO4_j(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t5_m0_l0 = D2CDO4_k(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t3_m0_l0 = D2CDO4_i(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t4_m0_l0 = D1CDO4crossed_ij(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t5_m0_l0 = D1CDO4crossed_ik(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t6_m0_l0 = D2CDO4_j(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t7_m0_l0 = D1CDO4crossed_jk(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t8_m0_l0 = D2CDO4_k(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast);
						m_Ctd_yyz_o0_t0_l0_p = 0.5 * d_Ctd_yyz_o0_t0_m0_l0;
						Ctd_yyz_p = m_Ctd_yyz_o0_t0_l0_p;
						m_Ctd_yyy_o0_t0_l0_p = 0.5 * d_Ctd_yyy_o0_t0_m0_l0;
						Ctd_yyy_p = m_Ctd_yyy_o0_t0_l0_p;
						div_Beta_p = (d_div_Beta_o0_t0_m0_l0 + d_div_Beta_o0_t1_m0_l0) + d_div_Beta_o0_t2_m0_l0;
						m_Ctd_xxz_o0_t0_l0_p = 0.5 * d_Ctd_xxz_o0_t0_m0_l0;
						Ctd_xxz_p = m_Ctd_xxz_o0_t0_l0_p;
						m_Ctd_xxx_o0_t0_l0_p = 0.5 * d_Ctd_xxx_o0_t0_m0_l0;
						Ctd_xxx_p = m_Ctd_xxx_o0_t0_l0_p;
						m_Ctd_yzz_o0_t1_l0_p = (-0.5) * d_Ctd_yzz_o0_t1_m0_l0;
						Ctd_yzz_p = d_Ctd_yzz_o0_t0_m0_l0 + m_Ctd_yzz_o0_t1_l0_p;
						m_Ctd_xxy_o0_t0_l0_p = 0.5 * d_Ctd_xxy_o0_t0_m0_l0;
						Ctd_xxy_p = m_Ctd_xxy_o0_t0_l0_p;
						kappa_z2_p = p_kappa_z2;
						m_Ctd_xyy_o0_t1_l0_p = (-0.5) * d_Ctd_xyy_o0_t1_m0_l0;
						Ctd_xyy_p = d_Ctd_xyy_o0_t0_m0_l0 + m_Ctd_xyy_o0_t1_l0_p;
						m_Ctd_xyz_o0_t2_l0_p = (-0.5) * d_Ctd_xyz_o0_t2_m0_l0;
						m_Ctd_xyz_o0_t1_l0_p = 0.5 * d_Ctd_xyz_o0_t1_m0_l0;
						m_Ctd_xyz_o0_t0_l0_p = 0.5 * d_Ctd_xyz_o0_t0_m0_l0;
						Ctd_xyz_p = (m_Ctd_xyz_o0_t0_l0_p + m_Ctd_xyz_o0_t1_l0_p) + m_Ctd_xyz_o0_t2_l0_p;
						d_div_Beta_y_p = (d_d_div_Beta_y_o0_t0_m0_l0 + d_d_div_Beta_y_o0_t1_m0_l0) + d_d_div_Beta_y_o0_t2_m0_l0;
						d_div_Beta_x_p = (d_d_div_Beta_x_o0_t0_m0_l0 + d_d_div_Beta_x_o0_t1_m0_l0) + d_d_div_Beta_x_o0_t2_m0_l0;
						d_div_Beta_z_p = (d_d_div_Beta_z_o0_t0_m0_l0 + d_d_div_Beta_z_o0_t1_m0_l0) + d_d_div_Beta_z_o0_t2_m0_l0;
						m_Ctd_xzz_o0_t1_l0_p = (-0.5) * d_Ctd_xzz_o0_t1_m0_l0;
						Ctd_xzz_p = d_Ctd_xzz_o0_t0_m0_l0 + m_Ctd_xzz_o0_t1_l0_p;
						m_Ctd_zxz_o0_t0_l0_p = 0.5 * d_Ctd_xzz_o0_t1_m0_l0;
						Ctd_zxz_p = m_Ctd_zxz_o0_t0_l0_p;
						m_Ctd_zxx_o0_t1_l0_p = (-0.5) * d_Ctd_xxz_o0_t0_m0_l0;
						Ctd_zxx_p = d_Ctd_zxx_o0_t0_m0_l0 + m_Ctd_zxx_o0_t1_l0_p;
						detgtd_p = vector(gtd_yy_p, i, j, k) * vector(gtd_zz_p, i, j, k) * vector(gtd_xx_p, i, j, k) + (-vector(gtd_yy_p, i, j, k) * (vector(gtd_xz_p, i, j, k) * vector(gtd_xz_p, i, j, k))) + (-vector(gtd_zz_p, i, j, k) * (vector(gtd_xy_p, i, j, k) * vector(gtd_xy_p, i, j, k))) + (-(vector(gtd_yz_p, i, j, k) * vector(gtd_yz_p, i, j, k)) * vector(gtd_xx_p, i, j, k)) + 2.0 * vector(gtd_yz_p, i, j, k) * vector(gtd_xy_p, i, j, k) * vector(gtd_xz_p, i, j, k);
						chi_max_p = MAX(chi_floor, vector(chi_p, i, j, k));
						inv_chi_p = 1.0 / fabs(chi_max_p);
						Bfd_y_p = inv_chi_p * (vector(gtd_xy_p, i, j, k) * vector(Bfu_x_p, i, j, k) + vector(gtd_yy_p, i, j, k) * vector(Bfu_y_p, i, j, k) + vector(gtd_yz_p, i, j, k) * vector(Bfu_z_p, i, j, k));
						Bfd_z_p = inv_chi_p * (vector(gtd_xz_p, i, j, k) * vector(Bfu_x_p, i, j, k) + vector(gtd_yz_p, i, j, k) * vector(Bfu_y_p, i, j, k) + vector(gtd_zz_p, i, j, k) * vector(Bfu_z_p, i, j, k));
						Bfd_x_p = inv_chi_p * (vector(gtd_xx_p, i, j, k) * vector(Bfu_x_p, i, j, k) + vector(gtd_xy_p, i, j, k) * vector(Bfu_y_p, i, j, k) + vector(gtd_xz_p, i, j, k) * vector(Bfu_z_p, i, j, k));
						sqBf_p = vector(Bfu_x_p, i, j, k) * Bfd_x_p + vector(Bfu_y_p, i, j, k) * Bfd_y_p + vector(Bfu_z_p, i, j, k) * Bfd_z_p;
						m_Ctd_zyy_o0_t1_l0_p = (-0.5) * d_Ctd_yyz_o0_t0_m0_l0;
						Ctd_zyy_p = d_Ctd_zyy_o0_t0_m0_l0 + m_Ctd_zyy_o0_t1_l0_p;
						m_Ctd_zyz_o0_t0_l0_p = 0.5 * d_Ctd_yzz_o0_t1_m0_l0;
						Ctd_zyz_p = m_Ctd_zyz_o0_t0_l0_p;
						sdetg_p = pow(chi_max_p, (-1.500000000000000));
						Tf4u_tt_p = (vector(Df_p, i, j, k) + vector(tauf_p, i, j, k)) / (sdetg_p * (vector(Alpha_p, i, j, k) * vector(Alpha_p, i, j, k)));
						Betatd_x_p = vector(Betau_x_p, i, j, k) * vector(gtd_xx_p, i, j, k) + vector(Betau_y_p, i, j, k) * vector(gtd_xy_p, i, j, k) + vector(Betau_z_p, i, j, k) * vector(gtd_xz_p, i, j, k);
						Betatd_z_p = vector(Betau_x_p, i, j, k) * vector(gtd_xz_p, i, j, k) + vector(Betau_y_p, i, j, k) * vector(gtd_yz_p, i, j, k) + vector(Betau_z_p, i, j, k) * vector(gtd_zz_p, i, j, k);
						Betatd_y_p = vector(Betau_x_p, i, j, k) * vector(gtd_xy_p, i, j, k) + vector(Betau_y_p, i, j, k) * vector(gtd_yy_p, i, j, k) + vector(Betau_z_p, i, j, k) * vector(gtd_yz_p, i, j, k);
						idetgtd_p = 1.0 / detgtd_p;
						gtu_zz_p = idetgtd_p * (vector(gtd_xx_p, i, j, k) * vector(gtd_yy_p, i, j, k) - vector(gtd_xy_p, i, j, k) * vector(gtd_xy_p, i, j, k));
						gtu_yy_p = idetgtd_p * (vector(gtd_xx_p, i, j, k) * vector(gtd_zz_p, i, j, k) - vector(gtd_xz_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						gtu_yz_p = idetgtd_p * ((-vector(gtd_xx_p, i, j, k) * vector(gtd_yz_p, i, j, k)) + vector(gtd_xy_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						gtu_xx_p = idetgtd_p * (vector(gtd_yy_p, i, j, k) * vector(gtd_zz_p, i, j, k) - vector(gtd_yz_p, i, j, k) * vector(gtd_yz_p, i, j, k));
						gtu_xz_p = idetgtd_p * (vector(gtd_xy_p, i, j, k) * vector(gtd_yz_p, i, j, k) - vector(gtd_yy_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						Ct_zyz_p = gtu_xz_p * Ctd_xyz_p + gtu_yz_p * Ctd_yyz_p + gtu_zz_p * Ctd_zyz_p;
						Ct_zyy_p = gtu_xz_p * Ctd_xyy_p + gtu_yz_p * Ctd_yyy_p + gtu_zz_p * Ctd_zyy_p;
						Atud_zx_p = gtu_xz_p * vector(Atd_xx_p, i, j, k) + gtu_yz_p * vector(Atd_xy_p, i, j, k) + gtu_zz_p * vector(Atd_xz_p, i, j, k);
						Atud_zz_p = gtu_xz_p * vector(Atd_xz_p, i, j, k) + gtu_yz_p * vector(Atd_yz_p, i, j, k) + gtu_zz_p * vector(Atd_zz_p, i, j, k);
						Atud_zy_p = gtu_xz_p * vector(Atd_xy_p, i, j, k) + gtu_yz_p * vector(Atd_yy_p, i, j, k) + gtu_zz_p * vector(Atd_yz_p, i, j, k);
						Atu_zz_p = Atud_zx_p * gtu_xz_p + Atud_zy_p * gtu_yz_p + Atud_zz_p * gtu_zz_p;
						vfu_z_p = chi_max_p * (gtu_xz_p * vector(vfd_x, i, j, k) + gtu_yz_p * vector(vfd_y, i, j, k) + gtu_zz_p * vector(vfd_z, i, j, k));
						Sfu_z_p = chi_max_p * (gtu_xz_p * vector(Sfd_x_p, i, j, k) + gtu_yz_p * vector(Sfd_y_p, i, j, k) + gtu_zz_p * vector(Sfd_z_p, i, j, k));
						Tf4u_tz_p = Sfu_z_p / (sdetg_p * vector(Alpha_p, i, j, k)) + (-vector(Betau_z_p, i, j, k) * Tf4u_tt_p);
						Tu_tz_p = Tf4u_tz_p;
						gtu_xy_p = idetgtd_p * ((-vector(gtd_xy_p, i, j, k) * vector(gtd_zz_p, i, j, k)) + vector(gtd_yz_p, i, j, k) * vector(gtd_xz_p, i, j, k));
						vfu_y_p = chi_max_p * (gtu_xy_p * vector(vfd_x, i, j, k) + gtu_yy_p * vector(vfd_y, i, j, k) + gtu_yz_p * vector(vfd_z, i, j, k));
						vfu_x_p = chi_max_p * (gtu_xx_p * vector(vfd_x, i, j, k) + gtu_xy_p * vector(vfd_y, i, j, k) + gtu_xz_p * vector(vfd_z, i, j, k));
						sqvf_p = MIN(vfu_x_p * vector(vfd_x, i, j, k) + vfu_y_p * vector(vfd_y, i, j, k) + vfu_z_p * vector(vfd_z, i, j, k), 0.9999999999);
						sqW_p = 1.0 / (1.0 - sqvf_p);
						Atud_yy_p = gtu_xy_p * vector(Atd_xy_p, i, j, k) + gtu_yy_p * vector(Atd_yy_p, i, j, k) + gtu_yz_p * vector(Atd_yz_p, i, j, k);
						Atud_yx_p = gtu_xy_p * vector(Atd_xx_p, i, j, k) + gtu_yy_p * vector(Atd_xy_p, i, j, k) + gtu_yz_p * vector(Atd_xz_p, i, j, k);
						Atud_yz_p = gtu_xy_p * vector(Atd_xz_p, i, j, k) + gtu_yy_p * vector(Atd_yz_p, i, j, k) + gtu_yz_p * vector(Atd_zz_p, i, j, k);
						Atu_yz_p = Atud_yx_p * gtu_xz_p + Atud_yy_p * gtu_yz_p + Atud_yz_p * gtu_zz_p;
						Atu_yy_p = Atud_yx_p * gtu_xy_p + Atud_yy_p * gtu_yy_p + Atud_yz_p * gtu_yz_p;
						Ct_yyy_p = gtu_xy_p * Ctd_xyy_p + gtu_yy_p * Ctd_yyy_p + gtu_yz_p * Ctd_zyy_p;
						Ct_yyz_p = gtu_xy_p * Ctd_xyz_p + gtu_yy_p * Ctd_yyz_p + gtu_yz_p * Ctd_zyz_p;
						Atud_xx_p = gtu_xx_p * vector(Atd_xx_p, i, j, k) + gtu_xy_p * vector(Atd_xy_p, i, j, k) + gtu_xz_p * vector(Atd_xz_p, i, j, k);
						trAt_p = Atud_xx_p + Atud_yy_p + Atud_zz_p;
						Atud_xz_p = gtu_xx_p * vector(Atd_xz_p, i, j, k) + gtu_xy_p * vector(Atd_yz_p, i, j, k) + gtu_xz_p * vector(Atd_zz_p, i, j, k);
						Atud_xy_p = gtu_xx_p * vector(Atd_xy_p, i, j, k) + gtu_xy_p * vector(Atd_yy_p, i, j, k) + gtu_xz_p * vector(Atd_yz_p, i, j, k);
						Atu_xz_p = Atud_xx_p * gtu_xz_p + Atud_xy_p * gtu_yz_p + Atud_xz_p * gtu_zz_p;
						Atu_xy_p = Atud_xx_p * gtu_xy_p + Atud_xy_p * gtu_yy_p + Atud_xz_p * gtu_yz_p;
						Atu_xx_p = Atud_xx_p * gtu_xx_p + Atud_xy_p * gtu_xy_p + Atud_xz_p * gtu_xz_p;
						Sfu_y_p = chi_max_p * (gtu_xy_p * vector(Sfd_x_p, i, j, k) + gtu_yy_p * vector(Sfd_y_p, i, j, k) + gtu_yz_p * vector(Sfd_z_p, i, j, k));
						Tf4u_ty_p = Sfu_y_p / (sdetg_p * vector(Alpha_p, i, j, k)) + (-vector(Betau_y_p, i, j, k) * Tf4u_tt_p);
						Tu_ty_p = Tf4u_ty_p;
						Sfu_x_p = chi_max_p * (gtu_xx_p * vector(Sfd_x_p, i, j, k) + gtu_xy_p * vector(Sfd_y_p, i, j, k) + gtu_xz_p * vector(Sfd_z_p, i, j, k));
						Tf4u_tx_p = Sfu_x_p / (sdetg_p * vector(Alpha_p, i, j, k)) + (-vector(Betau_x_p, i, j, k) * Tf4u_tt_p);
						Tu_tx_p = Tf4u_tx_p;
						decay_factor_p = MIN(1.0, pow((R_0 / MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))), eta_damping_exp));
						kappa_z1_p = MIN(p_kappa_z1, p_kappa_z1 * decay_factor_p) / vector(Alpha_p, i, j, k);
						kappa_f_p = MIN(p_kappa_f, p_kappa_f * decay_factor_p);
						feta_p = MIN(p_feta, p_feta * decay_factor_p);
						m_Ctd_zzz_o0_t0_l0_p = 0.5 * d_Ctd_zzz_o0_t0_m0_l0;
						Ctd_zzz_p = m_Ctd_zzz_o0_t0_l0_p;
						Ct_xzz_p = gtu_xx_p * Ctd_xzz_p + gtu_xy_p * Ctd_yzz_p + gtu_xz_p * Ctd_zzz_p;
						Ct_zzz_p = gtu_xz_p * Ctd_xzz_p + gtu_yz_p * Ctd_yzz_p + gtu_zz_p * Ctd_zzz_p;
						Ct_yzz_p = gtu_xy_p * Ctd_xzz_p + gtu_yy_p * Ctd_yzz_p + gtu_yz_p * Ctd_zzz_p;
						m_Ctd_yxy_o0_t0_l0_p = 0.5 * d_Ctd_xyy_o0_t1_m0_l0;
						Ctd_yxy_p = m_Ctd_yxy_o0_t0_l0_p;
						Ct_xyz_p = gtu_xx_p * Ctd_xyz_p + gtu_xy_p * Ctd_yyz_p + gtu_xz_p * Ctd_zyz_p;
						m_Ctd_yxz_o0_t2_l0_p = (-0.5) * d_Ctd_xyz_o0_t0_m0_l0;
						m_Ctd_yxz_o0_t0_l0_p = 0.5 * d_Ctd_xyz_o0_t2_m0_l0;
						Ctd_yxz_p = (m_Ctd_yxz_o0_t0_l0_p + m_Ctd_xyz_o0_t1_l0_p) + m_Ctd_yxz_o0_t2_l0_p;
						Ct_zxz_p = gtu_xz_p * Ctd_xxz_p + gtu_yz_p * Ctd_yxz_p + gtu_zz_p * Ctd_zxz_p;
						Ct_yxz_p = gtu_xy_p * Ctd_xxz_p + gtu_yy_p * Ctd_yxz_p + gtu_yz_p * Ctd_zxz_p;
						m_Ctd_zxy_o0_t2_l0_p = (-0.5) * d_Ctd_xyz_o0_t1_m0_l0;
						Ctd_zxy_p = (m_Ctd_yxz_o0_t0_l0_p + m_Ctd_xyz_o0_t0_l0_p) + m_Ctd_zxy_o0_t2_l0_p;
						Ct_zxy_p = gtu_xz_p * Ctd_xxy_p + gtu_yz_p * Ctd_yxy_p + gtu_zz_p * Ctd_zxy_p;
						Ct_yxy_p = gtu_xy_p * Ctd_xxy_p + gtu_yy_p * Ctd_yxy_p + gtu_yz_p * Ctd_zxy_p;
						Ct_xxy_p = gtu_xx_p * Ctd_xxy_p + gtu_xy_p * Ctd_yxy_p + gtu_xz_p * Ctd_zxy_p;
						Ct_xxz_p = gtu_xx_p * Ctd_xxz_p + gtu_xy_p * Ctd_yxz_p + gtu_xz_p * Ctd_zxz_p;
						Ct_xyy_p = gtu_xx_p * Ctd_xyy_p + gtu_xy_p * Ctd_yyy_p + gtu_xz_p * Ctd_zyy_p;
						m_Ctd_yxx_o0_t1_l0_p = (-0.5) * d_Ctd_xxy_o0_t0_m0_l0;
						Ctd_yxx_p = d_Ctd_yxx_o0_t0_m0_l0 + m_Ctd_yxx_o0_t1_l0_p;
						Ct_zxx_p = gtu_xz_p * Ctd_xxx_p + gtu_yz_p * Ctd_yxx_p + gtu_zz_p * Ctd_zxx_p;
						Gamt_z_p = gtu_xx_p * Ct_zxx_p + 2.0 * gtu_xy_p * Ct_zxy_p + 2.0 * gtu_xz_p * Ct_zxz_p + gtu_yy_p * Ct_zyy_p + 2.0 * gtu_yz_p * Ct_zyz_p + gtu_zz_p * Ct_zzz_p;
						Zu_z_p = 0.5 * chi_max_p * (vector(Gamh_z_p, i, j, k) - Gamt_z_p);
						Ct_yxx_p = gtu_xy_p * Ctd_xxx_p + gtu_yy_p * Ctd_yxx_p + gtu_yz_p * Ctd_zxx_p;
						Gamt_y_p = gtu_xx_p * Ct_yxx_p + 2.0 * gtu_xy_p * Ct_yxy_p + 2.0 * gtu_xz_p * Ct_yxz_p + gtu_yy_p * Ct_yyy_p + 2.0 * gtu_yz_p * Ct_yyz_p + gtu_zz_p * Ct_yzz_p;
						Zu_y_p = 0.5 * chi_max_p * (vector(Gamh_y_p, i, j, k) - Gamt_y_p);
						Ct_xxx_p = gtu_xx_p * Ctd_xxx_p + gtu_xy_p * Ctd_yxx_p + gtu_xz_p * Ctd_zxx_p;
						Gamt_x_p = gtu_xx_p * Ct_xxx_p + 2.0 * gtu_xy_p * Ct_xxy_p + 2.0 * gtu_xz_p * Ct_xxz_p + gtu_yy_p * Ct_xyy_p + 2.0 * gtu_yz_p * Ct_xyz_p + gtu_zz_p * Ct_xzz_p;
						Zu_x_p = 0.5 * chi_max_p * (vector(Gamh_x_p, i, j, k) - Gamt_x_p);
						d_Rtd_xx_o0_t9_m0_l0_p = gtu_zz_p * (3.0 * Ct_xxz_p * Ctd_xxz_p + 2.0 * Ct_yxz_p * Ctd_xyz_p + Ct_yxz_p * Ctd_yxz_p + 2.0 * Ct_zxz_p * Ctd_xzz_p + Ct_zxz_p * Ctd_zxz_p) + gtu_yz_p * (3.0 * Ct_xxy_p * Ctd_xxz_p + 2.0 * Ct_yxy_p * Ctd_xyz_p + Ct_yxy_p * Ctd_yxz_p + 2.0 * Ct_zxy_p * Ctd_xzz_p + Ct_zxy_p * Ctd_zxz_p) + gtu_xz_p * (3.0 * Ct_xxx_p * Ctd_xxz_p + 2.0 * Ct_yxx_p * Ctd_xyz_p + Ct_yxx_p * Ctd_yxz_p + 2.0 * Ct_zxx_p * Ctd_xzz_p + Ct_zxx_p * Ctd_zxz_p) + gtu_yz_p * (3.0 * Ct_xxz_p * Ctd_xxy_p + 2.0 * Ct_yxz_p * Ctd_xyy_p + Ct_yxz_p * Ctd_yxy_p + 2.0 * Ct_zxz_p * Ctd_xyz_p + Ct_zxz_p * Ctd_zxy_p) + gtu_yy_p * (3.0 * Ct_xxy_p * Ctd_xxy_p + 2.0 * Ct_yxy_p * Ctd_xyy_p + Ct_yxy_p * Ctd_yxy_p + 2.0 * Ct_zxy_p * Ctd_xyz_p + Ct_zxy_p * Ctd_zxy_p) + gtu_xy_p * (3.0 * Ct_xxx_p * Ctd_xxy_p + 2.0 * Ct_yxx_p * Ctd_xyy_p + Ct_yxx_p * Ctd_yxy_p + 2.0 * Ct_zxx_p * Ctd_xyz_p + Ct_zxx_p * Ctd_zxy_p) + gtu_xz_p * (3.0 * Ct_xxz_p * Ctd_xxx_p + 2.0 * Ct_yxz_p * Ctd_xxy_p + Ct_yxz_p * Ctd_yxx_p + 2.0 * Ct_zxz_p * Ctd_xxz_p + Ct_zxz_p * Ctd_zxx_p) + gtu_xy_p * (3.0 * Ct_xxy_p * Ctd_xxx_p + 2.0 * Ct_yxy_p * Ctd_xxy_p + Ct_yxy_p * Ctd_yxx_p + 2.0 * Ct_zxy_p * Ctd_xxz_p + Ct_zxy_p * Ctd_zxx_p) + gtu_xx_p * (3.0 * Ct_xxx_p * Ctd_xxx_p + 2.0 * Ct_yxx_p * Ctd_xxy_p + Ct_yxx_p * Ctd_yxx_p + 2.0 * Ct_zxx_p * Ctd_xxz_p + Ct_zxx_p * Ctd_zxx_p) + vector(Gamh_z_p, i, j, k) * Ctd_xxz_p + vector(Gamh_y_p, i, j, k) * Ctd_xxy_p + vector(Gamh_x_p, i, j, k) * Ctd_xxx_p;
						m_Rtd_xx_o0_t8_l0_p = vector(gtd_xz_p, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
						m_Rtd_xx_o0_t7_l0_p = vector(gtd_xy_p, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
						m_Rtd_xx_o0_t6_l0_p = vector(gtd_xx_p, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
						m_Rtd_xx_o0_t5_l0_p = (-0.5 * gtu_zz_p) * d_Rtd_xx_o0_t5_m0_l0;
						m_Rtd_xx_o0_t4_l0_p = (-gtu_yz_p) * d_Rtd_xx_o0_t4_m0_l0;
						m_Rtd_xx_o0_t3_l0_p = (-0.5 * gtu_yy_p) * d_Rtd_xx_o0_t3_m0_l0;
						m_Rtd_xx_o0_t2_l0_p = (-gtu_xz_p) * d_Rtd_xx_o0_t2_m0_l0;
						m_Rtd_xx_o0_t1_l0_p = (-gtu_xy_p) * d_Rtd_xx_o0_t1_m0_l0;
						m_Rtd_xx_o0_t0_l0_p = (-0.5 * gtu_xx_p) * d_Rtd_xx_o0_t0_m0_l0;
						Rtd_xx_p = ((((((((m_Rtd_xx_o0_t0_l0_p + m_Rtd_xx_o0_t1_l0_p) + m_Rtd_xx_o0_t2_l0_p) + m_Rtd_xx_o0_t3_l0_p) + m_Rtd_xx_o0_t4_l0_p) + m_Rtd_xx_o0_t5_l0_p) + m_Rtd_xx_o0_t6_l0_p) + m_Rtd_xx_o0_t7_l0_p) + m_Rtd_xx_o0_t8_l0_p) + d_Rtd_xx_o0_t9_m0_l0_p;
						d_Rtd_xz_o0_t12_m0_l0_p = gtu_zz_p * (Ct_xxz_p * Ctd_zxz_p + Ct_xxz_p * Ctd_xzz_p + Ct_xzz_p * Ctd_xxz_p + Ct_yxz_p * Ctd_zyz_p + Ct_yxz_p * Ctd_yzz_p + Ct_yzz_p * Ctd_xyz_p + 2.0 * Ct_zxz_p * Ctd_zzz_p + Ct_zzz_p * Ctd_xzz_p) + gtu_yz_p * (Ct_xxy_p * Ctd_zxz_p + Ct_xxy_p * Ctd_xzz_p + Ct_xyz_p * Ctd_xxz_p + Ct_yxy_p * Ctd_zyz_p + Ct_yxy_p * Ctd_yzz_p + Ct_yyz_p * Ctd_xyz_p + 2.0 * Ct_zxy_p * Ctd_zzz_p + Ct_zyz_p * Ctd_xzz_p) + gtu_xz_p * (Ct_xxx_p * Ctd_zxz_p + Ct_xxx_p * Ctd_xzz_p + Ct_xxz_p * Ctd_xxz_p + Ct_yxx_p * Ctd_zyz_p + Ct_yxx_p * Ctd_yzz_p + Ct_yxz_p * Ctd_xyz_p + 2.0 * Ct_zxx_p * Ctd_zzz_p + Ct_zxz_p * Ctd_xzz_p) + gtu_yz_p * (Ct_xxz_p * Ctd_zxy_p + Ct_xxz_p * Ctd_xyz_p + Ct_xzz_p * Ctd_xxy_p + Ct_yxz_p * Ctd_zyy_p + Ct_yxz_p * Ctd_yyz_p + Ct_yzz_p * Ctd_xyy_p + 2.0 * Ct_zxz_p * Ctd_zyz_p + Ct_zzz_p * Ctd_xyz_p) + gtu_yy_p * (Ct_xxy_p * Ctd_zxy_p + Ct_xxy_p * Ctd_xyz_p + Ct_xyz_p * Ctd_xxy_p + Ct_yxy_p * Ctd_zyy_p + Ct_yxy_p * Ctd_yyz_p + Ct_yyz_p * Ctd_xyy_p + 2.0 * Ct_zxy_p * Ctd_zyz_p + Ct_zyz_p * Ctd_xyz_p) + gtu_xy_p * (Ct_xxx_p * Ctd_zxy_p + Ct_xxx_p * Ctd_xyz_p + Ct_xxz_p * Ctd_xxy_p + Ct_yxx_p * Ctd_zyy_p + Ct_yxx_p * Ctd_yyz_p + Ct_yxz_p * Ctd_xyy_p + 2.0 * Ct_zxx_p * Ctd_zyz_p + Ct_zxz_p * Ctd_xyz_p) + gtu_xz_p * (Ct_xxz_p * Ctd_zxx_p + Ct_xxz_p * Ctd_xxz_p + Ct_xzz_p * Ctd_xxx_p + Ct_yxz_p * Ctd_zxy_p + Ct_yxz_p * Ctd_yxz_p + Ct_yzz_p * Ctd_xxy_p + 2.0 * Ct_zxz_p * Ctd_zxz_p + Ct_zzz_p * Ctd_xxz_p) + gtu_xy_p * (Ct_xxy_p * Ctd_zxx_p + Ct_xxy_p * Ctd_xxz_p + Ct_xyz_p * Ctd_xxx_p + Ct_yxy_p * Ctd_zxy_p + Ct_yxy_p * Ctd_yxz_p + Ct_yyz_p * Ctd_xxy_p + 2.0 * Ct_zxy_p * Ctd_zxz_p + Ct_zyz_p * Ctd_xxz_p) + gtu_xx_p * (Ct_xxx_p * Ctd_zxx_p + Ct_xxx_p * Ctd_xxz_p + Ct_xxz_p * Ctd_xxx_p + Ct_yxx_p * Ctd_zxy_p + Ct_yxx_p * Ctd_yxz_p + Ct_yxz_p * Ctd_xxy_p + 2.0 * Ct_zxx_p * Ctd_zxz_p + Ct_zxz_p * Ctd_xxz_p) + 0.5 * vector(Gamh_z_p, i, j, k) * (Ctd_xzz_p + Ctd_zxz_p) + 0.5 * vector(Gamh_y_p, i, j, k) * (Ctd_xyz_p + Ctd_zxy_p) + 0.5 * vector(Gamh_x_p, i, j, k) * (Ctd_xxz_p + Ctd_zxx_p);
						m_Rtd_xz_o0_t11_l0_p = 0.5 * vector(gtd_zz_p, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
						m_Rtd_xz_o0_t10_l0_p = 0.5 * vector(gtd_yz_p, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
						m_Rtd_xz_o0_t9_l0_p = 0.5 * vector(gtd_xz_p, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
						m_Rtd_xz_o0_t8_l0_p = 0.5 * vector(gtd_xz_p, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
						m_Rtd_xz_o0_t7_l0_p = 0.5 * vector(gtd_xy_p, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
						m_Rtd_xz_o0_t6_l0_p = 0.5 * vector(gtd_xx_p, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
						m_Rtd_xz_o0_t5_l0_p = (-0.5 * gtu_zz_p) * d_Rtd_xz_o0_t5_m0_l0;
						m_Rtd_xz_o0_t4_l0_p = (-gtu_yz_p) * d_Rtd_xz_o0_t4_m0_l0;
						m_Rtd_xz_o0_t3_l0_p = (-0.5 * gtu_yy_p) * d_Rtd_xz_o0_t3_m0_l0;
						m_Rtd_xz_o0_t2_l0_p = (-gtu_xz_p) * d_Rtd_xz_o0_t2_m0_l0;
						m_Rtd_xz_o0_t1_l0_p = (-gtu_xy_p) * d_Rtd_xz_o0_t1_m0_l0;
						m_Rtd_xz_o0_t0_l0_p = (-0.5 * gtu_xx_p) * d_Rtd_xz_o0_t0_m0_l0;
						Rtd_xz_p = (((((((((((m_Rtd_xz_o0_t0_l0_p + m_Rtd_xz_o0_t1_l0_p) + m_Rtd_xz_o0_t2_l0_p) + m_Rtd_xz_o0_t3_l0_p) + m_Rtd_xz_o0_t4_l0_p) + m_Rtd_xz_o0_t5_l0_p) + m_Rtd_xz_o0_t6_l0_p) + m_Rtd_xz_o0_t7_l0_p) + m_Rtd_xz_o0_t8_l0_p) + m_Rtd_xz_o0_t9_l0_p) + m_Rtd_xz_o0_t10_l0_p) + m_Rtd_xz_o0_t11_l0_p) + d_Rtd_xz_o0_t12_m0_l0_p;
						d_Rtd_zz_o0_t9_m0_l0_p = gtu_zz_p * (2.0 * Ct_xzz_p * Ctd_zxz_p + Ct_xzz_p * Ctd_xzz_p + 2.0 * Ct_yzz_p * Ctd_zyz_p + Ct_yzz_p * Ctd_yzz_p + 3.0 * Ct_zzz_p * Ctd_zzz_p) + gtu_yz_p * (2.0 * Ct_xyz_p * Ctd_zxz_p + Ct_xyz_p * Ctd_xzz_p + 2.0 * Ct_yyz_p * Ctd_zyz_p + Ct_yyz_p * Ctd_yzz_p + 3.0 * Ct_zyz_p * Ctd_zzz_p) + gtu_xz_p * (2.0 * Ct_xxz_p * Ctd_zxz_p + Ct_xxz_p * Ctd_xzz_p + 2.0 * Ct_yxz_p * Ctd_zyz_p + Ct_yxz_p * Ctd_yzz_p + 3.0 * Ct_zxz_p * Ctd_zzz_p) + gtu_yz_p * (2.0 * Ct_xzz_p * Ctd_zxy_p + Ct_xzz_p * Ctd_xyz_p + 2.0 * Ct_yzz_p * Ctd_zyy_p + Ct_yzz_p * Ctd_yyz_p + 3.0 * Ct_zzz_p * Ctd_zyz_p) + gtu_yy_p * (2.0 * Ct_xyz_p * Ctd_zxy_p + Ct_xyz_p * Ctd_xyz_p + 2.0 * Ct_yyz_p * Ctd_zyy_p + Ct_yyz_p * Ctd_yyz_p + 3.0 * Ct_zyz_p * Ctd_zyz_p) + gtu_xy_p * (2.0 * Ct_xxz_p * Ctd_zxy_p + Ct_xxz_p * Ctd_xyz_p + 2.0 * Ct_yxz_p * Ctd_zyy_p + Ct_yxz_p * Ctd_yyz_p + 3.0 * Ct_zxz_p * Ctd_zyz_p) + gtu_xz_p * (2.0 * Ct_xzz_p * Ctd_zxx_p + Ct_xzz_p * Ctd_xxz_p + 2.0 * Ct_yzz_p * Ctd_zxy_p + Ct_yzz_p * Ctd_yxz_p + 3.0 * Ct_zzz_p * Ctd_zxz_p) + gtu_xy_p * (2.0 * Ct_xyz_p * Ctd_zxx_p + Ct_xyz_p * Ctd_xxz_p + 2.0 * Ct_yyz_p * Ctd_zxy_p + Ct_yyz_p * Ctd_yxz_p + 3.0 * Ct_zyz_p * Ctd_zxz_p) + gtu_xx_p * (2.0 * Ct_xxz_p * Ctd_zxx_p + Ct_xxz_p * Ctd_xxz_p + 2.0 * Ct_yxz_p * Ctd_zxy_p + Ct_yxz_p * Ctd_yxz_p + 3.0 * Ct_zxz_p * Ctd_zxz_p) + vector(Gamh_z_p, i, j, k) * Ctd_zzz_p + vector(Gamh_y_p, i, j, k) * Ctd_zyz_p + vector(Gamh_x_p, i, j, k) * Ctd_zxz_p;
						m_Rtd_zz_o0_t8_l0_p = vector(gtd_zz_p, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
						m_Rtd_zz_o0_t7_l0_p = vector(gtd_yz_p, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
						m_Rtd_zz_o0_t6_l0_p = vector(gtd_xz_p, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
						m_Rtd_zz_o0_t5_l0_p = (-0.5 * gtu_zz_p) * d_Rtd_zz_o0_t5_m0_l0;
						m_Rtd_zz_o0_t4_l0_p = (-gtu_yz_p) * d_Rtd_zz_o0_t4_m0_l0;
						m_Rtd_zz_o0_t3_l0_p = (-0.5 * gtu_yy_p) * d_Rtd_zz_o0_t3_m0_l0;
						m_Rtd_zz_o0_t2_l0_p = (-gtu_xz_p) * d_Rtd_zz_o0_t2_m0_l0;
						m_Rtd_zz_o0_t1_l0_p = (-gtu_xy_p) * d_Rtd_zz_o0_t1_m0_l0;
						m_Rtd_zz_o0_t0_l0_p = (-0.5 * gtu_xx_p) * d_Rtd_zz_o0_t0_m0_l0;
						Rtd_zz_p = ((((((((m_Rtd_zz_o0_t0_l0_p + m_Rtd_zz_o0_t1_l0_p) + m_Rtd_zz_o0_t2_l0_p) + m_Rtd_zz_o0_t3_l0_p) + m_Rtd_zz_o0_t4_l0_p) + m_Rtd_zz_o0_t5_l0_p) + m_Rtd_zz_o0_t6_l0_p) + m_Rtd_zz_o0_t7_l0_p) + m_Rtd_zz_o0_t8_l0_p) + d_Rtd_zz_o0_t9_m0_l0_p;
						d_Rtd_xy_o0_t12_m0_l0_p = gtu_zz_p * (Ct_xxz_p * Ctd_yxz_p + Ct_xxz_p * Ctd_xyz_p + Ct_xyz_p * Ctd_xxz_p + 2.0 * Ct_yxz_p * Ctd_yyz_p + Ct_yyz_p * Ctd_xyz_p + Ct_zxz_p * Ctd_yzz_p + Ct_zxz_p * Ctd_zyz_p + Ct_zyz_p * Ctd_xzz_p) + gtu_yz_p * (Ct_xxy_p * Ctd_yxz_p + Ct_xxy_p * Ctd_xyz_p + Ct_xyy_p * Ctd_xxz_p + 2.0 * Ct_yxy_p * Ctd_yyz_p + Ct_yyy_p * Ctd_xyz_p + Ct_zxy_p * Ctd_yzz_p + Ct_zxy_p * Ctd_zyz_p + Ct_zyy_p * Ctd_xzz_p) + gtu_xz_p * (Ct_xxx_p * Ctd_yxz_p + Ct_xxx_p * Ctd_xyz_p + Ct_xxy_p * Ctd_xxz_p + 2.0 * Ct_yxx_p * Ctd_yyz_p + Ct_yxy_p * Ctd_xyz_p + Ct_zxx_p * Ctd_yzz_p + Ct_zxx_p * Ctd_zyz_p + Ct_zxy_p * Ctd_xzz_p) + gtu_yz_p * (Ct_xxz_p * Ctd_yxy_p + Ct_xxz_p * Ctd_xyy_p + Ct_xyz_p * Ctd_xxy_p + 2.0 * Ct_yxz_p * Ctd_yyy_p + Ct_yyz_p * Ctd_xyy_p + Ct_zxz_p * Ctd_yyz_p + Ct_zxz_p * Ctd_zyy_p + Ct_zyz_p * Ctd_xyz_p) + gtu_yy_p * (Ct_xxy_p * Ctd_yxy_p + Ct_xxy_p * Ctd_xyy_p + Ct_xyy_p * Ctd_xxy_p + 2.0 * Ct_yxy_p * Ctd_yyy_p + Ct_yyy_p * Ctd_xyy_p + Ct_zxy_p * Ctd_yyz_p + Ct_zxy_p * Ctd_zyy_p + Ct_zyy_p * Ctd_xyz_p) + gtu_xy_p * (Ct_xxx_p * Ctd_yxy_p + Ct_xxx_p * Ctd_xyy_p + Ct_xxy_p * Ctd_xxy_p + 2.0 * Ct_yxx_p * Ctd_yyy_p + Ct_yxy_p * Ctd_xyy_p + Ct_zxx_p * Ctd_yyz_p + Ct_zxx_p * Ctd_zyy_p + Ct_zxy_p * Ctd_xyz_p) + gtu_xz_p * (Ct_xxz_p * Ctd_yxx_p + Ct_xxz_p * Ctd_xxy_p + Ct_xyz_p * Ctd_xxx_p + 2.0 * Ct_yxz_p * Ctd_yxy_p + Ct_yyz_p * Ctd_xxy_p + Ct_zxz_p * Ctd_yxz_p + Ct_zxz_p * Ctd_zxy_p + Ct_zyz_p * Ctd_xxz_p) + gtu_xy_p * (Ct_xxy_p * Ctd_yxx_p + Ct_xxy_p * Ctd_xxy_p + Ct_xyy_p * Ctd_xxx_p + 2.0 * Ct_yxy_p * Ctd_yxy_p + Ct_yyy_p * Ctd_xxy_p + Ct_zxy_p * Ctd_yxz_p + Ct_zxy_p * Ctd_zxy_p + Ct_zyy_p * Ctd_xxz_p) + gtu_xx_p * (Ct_xxx_p * Ctd_yxx_p + Ct_xxx_p * Ctd_xxy_p + Ct_xxy_p * Ctd_xxx_p + 2.0 * Ct_yxx_p * Ctd_yxy_p + Ct_yxy_p * Ctd_xxy_p + Ct_zxx_p * Ctd_yxz_p + Ct_zxx_p * Ctd_zxy_p + Ct_zxy_p * Ctd_xxz_p) + 0.5 * vector(Gamh_z_p, i, j, k) * (Ctd_xyz_p + Ctd_yxz_p) + 0.5 * vector(Gamh_y_p, i, j, k) * (Ctd_xyy_p + Ctd_yxy_p) + 0.5 * vector(Gamh_x_p, i, j, k) * (Ctd_xxy_p + Ctd_yxx_p);
						m_Rtd_xy_o0_t11_l0_p = 0.5 * vector(gtd_yz_p, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
						m_Rtd_xy_o0_t10_l0_p = 0.5 * vector(gtd_yy_p, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
						m_Rtd_xy_o0_t9_l0_p = 0.5 * vector(gtd_xy_p, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
						m_Rtd_xy_o0_t8_l0_p = 0.5 * vector(gtd_xz_p, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
						m_Rtd_xy_o0_t7_l0_p = 0.5 * vector(gtd_xy_p, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
						m_Rtd_xy_o0_t6_l0_p = 0.5 * vector(gtd_xx_p, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
						m_Rtd_xy_o0_t5_l0_p = (-0.5 * gtu_zz_p) * d_Rtd_xy_o0_t5_m0_l0;
						m_Rtd_xy_o0_t4_l0_p = (-gtu_yz_p) * d_Rtd_xy_o0_t4_m0_l0;
						m_Rtd_xy_o0_t3_l0_p = (-0.5 * gtu_yy_p) * d_Rtd_xy_o0_t3_m0_l0;
						m_Rtd_xy_o0_t2_l0_p = (-gtu_xz_p) * d_Rtd_xy_o0_t2_m0_l0;
						m_Rtd_xy_o0_t1_l0_p = (-gtu_xy_p) * d_Rtd_xy_o0_t1_m0_l0;
						m_Rtd_xy_o0_t0_l0_p = (-0.5 * gtu_xx_p) * d_Rtd_xy_o0_t0_m0_l0;
						Rtd_xy_p = (((((((((((m_Rtd_xy_o0_t0_l0_p + m_Rtd_xy_o0_t1_l0_p) + m_Rtd_xy_o0_t2_l0_p) + m_Rtd_xy_o0_t3_l0_p) + m_Rtd_xy_o0_t4_l0_p) + m_Rtd_xy_o0_t5_l0_p) + m_Rtd_xy_o0_t6_l0_p) + m_Rtd_xy_o0_t7_l0_p) + m_Rtd_xy_o0_t8_l0_p) + m_Rtd_xy_o0_t9_l0_p) + m_Rtd_xy_o0_t10_l0_p) + m_Rtd_xy_o0_t11_l0_p) + d_Rtd_xy_o0_t12_m0_l0_p;
						d_Rtd_yy_o0_t9_m0_l0_p = gtu_zz_p * (2.0 * Ct_xyz_p * Ctd_yxz_p + Ct_xyz_p * Ctd_xyz_p + 3.0 * Ct_yyz_p * Ctd_yyz_p + 2.0 * Ct_zyz_p * Ctd_yzz_p + Ct_zyz_p * Ctd_zyz_p) + gtu_yz_p * (2.0 * Ct_xyy_p * Ctd_yxz_p + Ct_xyy_p * Ctd_xyz_p + 3.0 * Ct_yyy_p * Ctd_yyz_p + 2.0 * Ct_zyy_p * Ctd_yzz_p + Ct_zyy_p * Ctd_zyz_p) + gtu_xz_p * (2.0 * Ct_xxy_p * Ctd_yxz_p + Ct_xxy_p * Ctd_xyz_p + 3.0 * Ct_yxy_p * Ctd_yyz_p + 2.0 * Ct_zxy_p * Ctd_yzz_p + Ct_zxy_p * Ctd_zyz_p) + gtu_yz_p * (2.0 * Ct_xyz_p * Ctd_yxy_p + Ct_xyz_p * Ctd_xyy_p + 3.0 * Ct_yyz_p * Ctd_yyy_p + 2.0 * Ct_zyz_p * Ctd_yyz_p + Ct_zyz_p * Ctd_zyy_p) + gtu_yy_p * (2.0 * Ct_xyy_p * Ctd_yxy_p + Ct_xyy_p * Ctd_xyy_p + 3.0 * Ct_yyy_p * Ctd_yyy_p + 2.0 * Ct_zyy_p * Ctd_yyz_p + Ct_zyy_p * Ctd_zyy_p) + gtu_xy_p * (2.0 * Ct_xxy_p * Ctd_yxy_p + Ct_xxy_p * Ctd_xyy_p + 3.0 * Ct_yxy_p * Ctd_yyy_p + 2.0 * Ct_zxy_p * Ctd_yyz_p + Ct_zxy_p * Ctd_zyy_p) + gtu_xz_p * (2.0 * Ct_xyz_p * Ctd_yxx_p + Ct_xyz_p * Ctd_xxy_p + 3.0 * Ct_yyz_p * Ctd_yxy_p + 2.0 * Ct_zyz_p * Ctd_yxz_p + Ct_zyz_p * Ctd_zxy_p) + gtu_xy_p * (2.0 * Ct_xyy_p * Ctd_yxx_p + Ct_xyy_p * Ctd_xxy_p + 3.0 * Ct_yyy_p * Ctd_yxy_p + 2.0 * Ct_zyy_p * Ctd_yxz_p + Ct_zyy_p * Ctd_zxy_p) + gtu_xx_p * (2.0 * Ct_xxy_p * Ctd_yxx_p + Ct_xxy_p * Ctd_xxy_p + 3.0 * Ct_yxy_p * Ctd_yxy_p + 2.0 * Ct_zxy_p * Ctd_yxz_p + Ct_zxy_p * Ctd_zxy_p) + vector(Gamh_z_p, i, j, k) * Ctd_yyz_p + vector(Gamh_y_p, i, j, k) * Ctd_yyy_p + vector(Gamh_x_p, i, j, k) * Ctd_yxy_p;
						m_Rtd_yy_o0_t8_l0_p = vector(gtd_yz_p, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
						m_Rtd_yy_o0_t7_l0_p = vector(gtd_yy_p, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
						m_Rtd_yy_o0_t6_l0_p = vector(gtd_xy_p, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
						m_Rtd_yy_o0_t5_l0_p = (-0.5 * gtu_zz_p) * d_Rtd_yy_o0_t5_m0_l0;
						m_Rtd_yy_o0_t4_l0_p = (-gtu_yz_p) * d_Rtd_yy_o0_t4_m0_l0;
						m_Rtd_yy_o0_t3_l0_p = (-0.5 * gtu_yy_p) * d_Rtd_yy_o0_t3_m0_l0;
						m_Rtd_yy_o0_t2_l0_p = (-gtu_xz_p) * d_Rtd_yy_o0_t2_m0_l0;
						m_Rtd_yy_o0_t1_l0_p = (-gtu_xy_p) * d_Rtd_yy_o0_t1_m0_l0;
						m_Rtd_yy_o0_t0_l0_p = (-0.5 * gtu_xx_p) * d_Rtd_yy_o0_t0_m0_l0;
						Rtd_yy_p = ((((((((m_Rtd_yy_o0_t0_l0_p + m_Rtd_yy_o0_t1_l0_p) + m_Rtd_yy_o0_t2_l0_p) + m_Rtd_yy_o0_t3_l0_p) + m_Rtd_yy_o0_t4_l0_p) + m_Rtd_yy_o0_t5_l0_p) + m_Rtd_yy_o0_t6_l0_p) + m_Rtd_yy_o0_t7_l0_p) + m_Rtd_yy_o0_t8_l0_p) + d_Rtd_yy_o0_t9_m0_l0_p;
						d_Rtd_yz_o0_t12_m0_l0_p = gtu_zz_p * (Ct_xyz_p * Ctd_zxz_p + Ct_xyz_p * Ctd_xzz_p + Ct_xzz_p * Ctd_yxz_p + Ct_yyz_p * Ctd_zyz_p + Ct_yyz_p * Ctd_yzz_p + Ct_yzz_p * Ctd_yyz_p + 2.0 * Ct_zyz_p * Ctd_zzz_p + Ct_zzz_p * Ctd_yzz_p) + gtu_yz_p * (Ct_xyy_p * Ctd_zxz_p + Ct_xyy_p * Ctd_xzz_p + Ct_xyz_p * Ctd_yxz_p + Ct_yyy_p * Ctd_zyz_p + Ct_yyy_p * Ctd_yzz_p + Ct_yyz_p * Ctd_yyz_p + 2.0 * Ct_zyy_p * Ctd_zzz_p + Ct_zyz_p * Ctd_yzz_p) + gtu_xz_p * (Ct_xxy_p * Ctd_zxz_p + Ct_xxy_p * Ctd_xzz_p + Ct_xxz_p * Ctd_yxz_p + Ct_yxy_p * Ctd_zyz_p + Ct_yxy_p * Ctd_yzz_p + Ct_yxz_p * Ctd_yyz_p + 2.0 * Ct_zxy_p * Ctd_zzz_p + Ct_zxz_p * Ctd_yzz_p) + gtu_yz_p * (Ct_xyz_p * Ctd_zxy_p + Ct_xyz_p * Ctd_xyz_p + Ct_xzz_p * Ctd_yxy_p + Ct_yyz_p * Ctd_zyy_p + Ct_yyz_p * Ctd_yyz_p + Ct_yzz_p * Ctd_yyy_p + 2.0 * Ct_zyz_p * Ctd_zyz_p + Ct_zzz_p * Ctd_yyz_p) + gtu_yy_p * (Ct_xyy_p * Ctd_zxy_p + Ct_xyy_p * Ctd_xyz_p + Ct_xyz_p * Ctd_yxy_p + Ct_yyy_p * Ctd_zyy_p + Ct_yyy_p * Ctd_yyz_p + Ct_yyz_p * Ctd_yyy_p + 2.0 * Ct_zyy_p * Ctd_zyz_p + Ct_zyz_p * Ctd_yyz_p) + gtu_xy_p * (Ct_xxy_p * Ctd_zxy_p + Ct_xxy_p * Ctd_xyz_p + Ct_xxz_p * Ctd_yxy_p + Ct_yxy_p * Ctd_zyy_p + Ct_yxy_p * Ctd_yyz_p + Ct_yxz_p * Ctd_yyy_p + 2.0 * Ct_zxy_p * Ctd_zyz_p + Ct_zxz_p * Ctd_yyz_p) + gtu_xz_p * (Ct_xyz_p * Ctd_zxx_p + Ct_xyz_p * Ctd_xxz_p + Ct_xzz_p * Ctd_yxx_p + Ct_yyz_p * Ctd_zxy_p + Ct_yyz_p * Ctd_yxz_p + Ct_yzz_p * Ctd_yxy_p + 2.0 * Ct_zyz_p * Ctd_zxz_p + Ct_zzz_p * Ctd_yxz_p) + gtu_xy_p * (Ct_xyy_p * Ctd_zxx_p + Ct_xyy_p * Ctd_xxz_p + Ct_xyz_p * Ctd_yxx_p + Ct_yyy_p * Ctd_zxy_p + Ct_yyy_p * Ctd_yxz_p + Ct_yyz_p * Ctd_yxy_p + 2.0 * Ct_zyy_p * Ctd_zxz_p + Ct_zyz_p * Ctd_yxz_p) + gtu_xx_p * (Ct_xxy_p * Ctd_zxx_p + Ct_xxy_p * Ctd_xxz_p + Ct_xxz_p * Ctd_yxx_p + Ct_yxy_p * Ctd_zxy_p + Ct_yxy_p * Ctd_yxz_p + Ct_yxz_p * Ctd_yxy_p + 2.0 * Ct_zxy_p * Ctd_zxz_p + Ct_zxz_p * Ctd_yxz_p) + 0.5 * vector(Gamh_z_p, i, j, k) * (Ctd_yzz_p + Ctd_zyz_p) + 0.5 * vector(Gamh_y_p, i, j, k) * (Ctd_yyz_p + Ctd_zyy_p) + 0.5 * vector(Gamh_x_p, i, j, k) * (Ctd_yxz_p + Ctd_zxy_p);
						m_Rtd_yz_o0_t11_l0_p = 0.5 * vector(gtd_zz_p, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
						m_Rtd_yz_o0_t10_l0_p = 0.5 * vector(gtd_yz_p, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
						m_Rtd_yz_o0_t9_l0_p = 0.5 * vector(gtd_xz_p, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
						m_Rtd_yz_o0_t8_l0_p = 0.5 * vector(gtd_yz_p, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
						m_Rtd_yz_o0_t7_l0_p = 0.5 * vector(gtd_yy_p, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
						m_Rtd_yz_o0_t6_l0_p = 0.5 * vector(gtd_xy_p, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
						m_Rtd_yz_o0_t5_l0_p = (-0.5 * gtu_zz_p) * d_Rtd_yz_o0_t5_m0_l0;
						m_Rtd_yz_o0_t4_l0_p = (-gtu_yz_p) * d_Rtd_yz_o0_t4_m0_l0;
						m_Rtd_yz_o0_t3_l0_p = (-0.5 * gtu_yy_p) * d_Rtd_yz_o0_t3_m0_l0;
						m_Rtd_yz_o0_t2_l0_p = (-gtu_xz_p) * d_Rtd_yz_o0_t2_m0_l0;
						m_Rtd_yz_o0_t1_l0_p = (-gtu_xy_p) * d_Rtd_yz_o0_t1_m0_l0;
						m_Rtd_yz_o0_t0_l0_p = (-0.5 * gtu_xx_p) * d_Rtd_yz_o0_t0_m0_l0;
						Rtd_yz_p = (((((((((((m_Rtd_yz_o0_t0_l0_p + m_Rtd_yz_o0_t1_l0_p) + m_Rtd_yz_o0_t2_l0_p) + m_Rtd_yz_o0_t3_l0_p) + m_Rtd_yz_o0_t4_l0_p) + m_Rtd_yz_o0_t5_l0_p) + m_Rtd_yz_o0_t6_l0_p) + m_Rtd_yz_o0_t7_l0_p) + m_Rtd_yz_o0_t8_l0_p) + m_Rtd_yz_o0_t9_l0_p) + m_Rtd_yz_o0_t10_l0_p) + m_Rtd_yz_o0_t11_l0_p) + d_Rtd_yz_o0_t12_m0_l0_p;
						m_Rpd_xx_o0_t22_l0_p = 0.5 * d_Rpd_xx_o0_t3_m0_l0;
						m_Rpd_xx_o0_t21_l0_p = (-0.75 * vector(gtd_xx_p, i, j, k) * gtu_zz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xx_o0_t20_l0_p = (-0.75 * vector(gtd_xx_p, i, j, k) * gtu_yy_p * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xx_o0_t19_l0_p = (-0.75 * vector(gtd_xx_p, i, j, k) * gtu_xx_p * inv_chi_p) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xx_o0_t18_l0_p = 2.0 * inv_chi_p * Zu_z_p * vector(gtd_xz_p, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xx_o0_t17_l0_p = 2.0 * inv_chi_p * Zu_y_p * vector(gtd_xy_p, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xx_o0_t16_l0_p = 2.0 * inv_chi_p * Zu_x_p * vector(gtd_xx_p, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xx_o0_t15_l0_p = (-1.500000000000000 * vector(gtd_xx_p, i, j, k) * gtu_yz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xx_o0_t14_l0_p = (-1.500000000000000 * vector(gtd_xx_p, i, j, k) * gtu_xz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xx_o0_t13_l0_p = (-1.500000000000000 * vector(gtd_xx_p, i, j, k) * gtu_xy_p * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xx_o0_t12_l0_p = (-0.25 * inv_chi_p) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xx_o0_t11_l0_p = (-0.5 * Ct_zxx_p) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xx_o0_t10_l0_p = (-0.5 * Ct_yxx_p) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xx_o0_t9_l0_p = (-0.5 * Ct_xxx_p) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xx_o0_t8_l0_p = 0.5 * vector(gtd_xx_p, i, j, k) * gtu_zz_p * d_Rpd_xx_o0_t8_m0_l0;
						m_Rpd_xx_o0_t7_l0_p = vector(gtd_xx_p, i, j, k) * gtu_yz_p * d_Rpd_xx_o0_t7_m0_l0;
						m_Rpd_xx_o0_t6_l0_p = 0.5 * vector(gtd_xx_p, i, j, k) * gtu_yy_p * d_Rpd_xx_o0_t6_m0_l0;
						m_Rpd_xx_o0_t5_l0_p = vector(gtd_xx_p, i, j, k) * gtu_xz_p * d_Rpd_xx_o0_t5_m0_l0;
						m_Rpd_xx_o0_t4_l0_p = vector(gtd_xx_p, i, j, k) * gtu_xy_p * d_Rpd_xx_o0_t4_m0_l0;
						m_Rpd_xx_o0_t3_l0_p = 0.5 * vector(gtd_xx_p, i, j, k) * gtu_xx_p * d_Rpd_xx_o0_t3_m0_l0;
						m_Rpd_xx_o0_t2_l0_p = (-0.5 * vector(gtd_xx_p, i, j, k) * vector(Gamh_z_p, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xx_o0_t1_l0_p = (-0.5 * vector(gtd_xx_p, i, j, k) * vector(Gamh_y_p, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xx_o0_t0_l0_p = (-0.5 * vector(gtd_xx_p, i, j, k) * vector(Gamh_x_p, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						Rpd_xx_p = (((((((((((((((((((((m_Rpd_xx_o0_t0_l0_p + m_Rpd_xx_o0_t1_l0_p) + m_Rpd_xx_o0_t2_l0_p) + m_Rpd_xx_o0_t3_l0_p) + m_Rpd_xx_o0_t4_l0_p) + m_Rpd_xx_o0_t5_l0_p) + m_Rpd_xx_o0_t6_l0_p) + m_Rpd_xx_o0_t7_l0_p) + m_Rpd_xx_o0_t8_l0_p) + m_Rpd_xx_o0_t9_l0_p) + m_Rpd_xx_o0_t10_l0_p) + m_Rpd_xx_o0_t11_l0_p) + m_Rpd_xx_o0_t12_l0_p) + m_Rpd_xx_o0_t13_l0_p) + m_Rpd_xx_o0_t14_l0_p) + m_Rpd_xx_o0_t15_l0_p) + m_Rpd_xx_o0_t16_l0_p) + m_Rpd_xx_o0_t17_l0_p) + m_Rpd_xx_o0_t18_l0_p) + m_Rpd_xx_o0_t19_l0_p) + m_Rpd_xx_o0_t20_l0_p) + m_Rpd_xx_o0_t21_l0_p) + m_Rpd_xx_o0_t22_l0_p;
						m_Rpd_yy_o0_t22_l0_p = 0.5 * d_Rpd_xx_o0_t6_m0_l0;
						m_Rpd_yy_o0_t21_l0_p = (-0.75 * vector(gtd_yy_p, i, j, k) * gtu_zz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_yy_o0_t20_l0_p = (-0.75 * vector(gtd_yy_p, i, j, k) * gtu_yy_p * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yy_o0_t19_l0_p = (-0.75 * vector(gtd_yy_p, i, j, k) * gtu_xx_p * inv_chi_p) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_yy_o0_t18_l0_p = 2.0 * inv_chi_p * Zu_z_p * vector(gtd_yz_p, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yy_o0_t17_l0_p = 2.0 * inv_chi_p * Zu_y_p * vector(gtd_yy_p, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yy_o0_t16_l0_p = 2.0 * inv_chi_p * Zu_x_p * vector(gtd_xy_p, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yy_o0_t15_l0_p = (-1.500000000000000 * vector(gtd_yy_p, i, j, k) * gtu_xz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_yy_o0_t14_l0_p = (-1.500000000000000 * vector(gtd_yy_p, i, j, k) * gtu_xy_p * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_yy_o0_t13_l0_p = (-0.25 * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yy_o0_t12_l0_p = (-0.5 * Ct_zyy_p) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_yy_o0_t11_l0_p = (-0.5 * Ct_yyy_p) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yy_o0_t10_l0_p = (-0.5 * Ct_xyy_p) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_yy_o0_t9_l0_p = 0.5 * vector(gtd_yy_p, i, j, k) * gtu_zz_p * d_Rpd_xx_o0_t8_m0_l0;
						m_Rpd_yy_o0_t8_l0_p = vector(gtd_yy_p, i, j, k) * gtu_yz_p * d_Rpd_xx_o0_t7_m0_l0;
						m_Rpd_yy_o0_t7_l0_p = 0.5 * vector(gtd_yy_p, i, j, k) * gtu_yy_p * d_Rpd_xx_o0_t6_m0_l0;
						m_Rpd_yy_o0_t6_l0_p = vector(gtd_yy_p, i, j, k) * gtu_xz_p * d_Rpd_xx_o0_t5_m0_l0;
						m_Rpd_yy_o0_t5_l0_p = vector(gtd_yy_p, i, j, k) * gtu_xy_p * d_Rpd_xx_o0_t4_m0_l0;
						m_Rpd_yy_o0_t4_l0_p = 0.5 * vector(gtd_yy_p, i, j, k) * gtu_xx_p * d_Rpd_xx_o0_t3_m0_l0;
						m_Rpd_yy_o0_t3_l0_p = (-0.5 * vector(gtd_yy_p, i, j, k) * vector(Gamh_z_p, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_yy_o0_t2_l0_p = (-0.5 * vector(gtd_yy_p, i, j, k) * vector(Gamh_y_p, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yy_o0_t1_l0_p = (-0.5 * vector(gtd_yy_p, i, j, k) * vector(Gamh_x_p, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_yy_o0_t0_l0_p = (-1.500000000000000 * vector(gtd_yy_p, i, j, k) * gtu_yz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						Rpd_yy_p = (((((((((((((((((((((m_Rpd_yy_o0_t0_l0_p + m_Rpd_yy_o0_t1_l0_p) + m_Rpd_yy_o0_t2_l0_p) + m_Rpd_yy_o0_t3_l0_p) + m_Rpd_yy_o0_t4_l0_p) + m_Rpd_yy_o0_t5_l0_p) + m_Rpd_yy_o0_t6_l0_p) + m_Rpd_yy_o0_t7_l0_p) + m_Rpd_yy_o0_t8_l0_p) + m_Rpd_yy_o0_t9_l0_p) + m_Rpd_yy_o0_t10_l0_p) + m_Rpd_yy_o0_t11_l0_p) + m_Rpd_yy_o0_t12_l0_p) + m_Rpd_yy_o0_t13_l0_p) + m_Rpd_yy_o0_t14_l0_p) + m_Rpd_yy_o0_t15_l0_p) + m_Rpd_yy_o0_t16_l0_p) + m_Rpd_yy_o0_t17_l0_p) + m_Rpd_yy_o0_t18_l0_p) + m_Rpd_yy_o0_t19_l0_p) + m_Rpd_yy_o0_t20_l0_p) + m_Rpd_yy_o0_t21_l0_p) + m_Rpd_yy_o0_t22_l0_p;
						m_Rpd_yz_o0_t25_l0_p = 0.5 * d_Rpd_xx_o0_t7_m0_l0;
						m_Rpd_yz_o0_t24_l0_p = (-0.75 * vector(gtd_yz_p, i, j, k) * gtu_zz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_yz_o0_t23_l0_p = (-0.75 * vector(gtd_yz_p, i, j, k) * gtu_yy_p * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yz_o0_t22_l0_p = (-0.75 * vector(gtd_yz_p, i, j, k) * gtu_xx_p * inv_chi_p) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_yz_o0_t21_l0_p = inv_chi_p * Zu_z_p * vector(gtd_zz_p, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yz_o0_t20_l0_p = inv_chi_p * Zu_y_p * vector(gtd_yz_p, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yz_o0_t19_l0_p = inv_chi_p * Zu_x_p * vector(gtd_xz_p, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yz_o0_t18_l0_p = inv_chi_p * Zu_z_p * vector(gtd_yz_p, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_yz_o0_t17_l0_p = inv_chi_p * Zu_y_p * vector(gtd_yy_p, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_yz_o0_t16_l0_p = inv_chi_p * Zu_x_p * vector(gtd_xy_p, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_yz_o0_t15_l0_p = (-0.5 * Ct_zyz_p) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_yz_o0_t14_l0_p = (-0.5 * Ct_yyz_p) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yz_o0_t13_l0_p = (-0.5 * Ct_xyz_p) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_yz_o0_t12_l0_p = (-0.25 * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_yz_o0_t11_l0_p = 0.5 * vector(gtd_yz_p, i, j, k) * gtu_zz_p * d_Rpd_xx_o0_t8_m0_l0;
						m_Rpd_yz_o0_t10_l0_p = vector(gtd_yz_p, i, j, k) * gtu_yz_p * d_Rpd_xx_o0_t7_m0_l0;
						m_Rpd_yz_o0_t9_l0_p = 0.5 * vector(gtd_yz_p, i, j, k) * gtu_yy_p * d_Rpd_xx_o0_t6_m0_l0;
						m_Rpd_yz_o0_t8_l0_p = vector(gtd_yz_p, i, j, k) * gtu_xz_p * d_Rpd_xx_o0_t5_m0_l0;
						m_Rpd_yz_o0_t7_l0_p = vector(gtd_yz_p, i, j, k) * gtu_xy_p * d_Rpd_xx_o0_t4_m0_l0;
						m_Rpd_yz_o0_t6_l0_p = 0.5 * vector(gtd_yz_p, i, j, k) * gtu_xx_p * d_Rpd_xx_o0_t3_m0_l0;
						m_Rpd_yz_o0_t5_l0_p = (-0.5 * vector(gtd_yz_p, i, j, k) * vector(Gamh_z_p, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_yz_o0_t4_l0_p = (-0.5 * vector(gtd_yz_p, i, j, k) * vector(Gamh_y_p, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yz_o0_t3_l0_p = (-0.5 * vector(gtd_yz_p, i, j, k) * vector(Gamh_x_p, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_yz_o0_t2_l0_p = (-1.500000000000000 * vector(gtd_yz_p, i, j, k) * gtu_yz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_yz_o0_t1_l0_p = (-1.500000000000000 * vector(gtd_yz_p, i, j, k) * gtu_xz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_yz_o0_t0_l0_p = (-1.500000000000000 * vector(gtd_yz_p, i, j, k) * gtu_xy_p * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						Rpd_yz_p = ((((((((((((((((((((((((m_Rpd_yz_o0_t0_l0_p + m_Rpd_yz_o0_t1_l0_p) + m_Rpd_yz_o0_t2_l0_p) + m_Rpd_yz_o0_t3_l0_p) + m_Rpd_yz_o0_t4_l0_p) + m_Rpd_yz_o0_t5_l0_p) + m_Rpd_yz_o0_t6_l0_p) + m_Rpd_yz_o0_t7_l0_p) + m_Rpd_yz_o0_t8_l0_p) + m_Rpd_yz_o0_t9_l0_p) + m_Rpd_yz_o0_t10_l0_p) + m_Rpd_yz_o0_t11_l0_p) + m_Rpd_yz_o0_t12_l0_p) + m_Rpd_yz_o0_t13_l0_p) + m_Rpd_yz_o0_t14_l0_p) + m_Rpd_yz_o0_t15_l0_p) + m_Rpd_yz_o0_t16_l0_p) + m_Rpd_yz_o0_t17_l0_p) + m_Rpd_yz_o0_t18_l0_p) + m_Rpd_yz_o0_t19_l0_p) + m_Rpd_yz_o0_t20_l0_p) + m_Rpd_yz_o0_t21_l0_p) + m_Rpd_yz_o0_t22_l0_p) + m_Rpd_yz_o0_t23_l0_p) + m_Rpd_yz_o0_t24_l0_p) + m_Rpd_yz_o0_t25_l0_p;
						m_Rpd_xz_o0_t25_l0_p = 0.5 * d_Rpd_xx_o0_t5_m0_l0;
						m_Rpd_xz_o0_t24_l0_p = (-0.75 * vector(gtd_xz_p, i, j, k) * gtu_zz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xz_o0_t23_l0_p = (-0.75 * vector(gtd_xz_p, i, j, k) * gtu_yy_p * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xz_o0_t22_l0_p = (-0.75 * vector(gtd_xz_p, i, j, k) * gtu_xx_p * inv_chi_p) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xz_o0_t21_l0_p = inv_chi_p * Zu_z_p * vector(gtd_zz_p, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xz_o0_t20_l0_p = inv_chi_p * Zu_y_p * vector(gtd_yz_p, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xz_o0_t19_l0_p = inv_chi_p * Zu_x_p * vector(gtd_xz_p, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xz_o0_t18_l0_p = inv_chi_p * Zu_z_p * vector(gtd_xz_p, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xz_o0_t17_l0_p = inv_chi_p * Zu_y_p * vector(gtd_xy_p, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xz_o0_t16_l0_p = inv_chi_p * Zu_x_p * vector(gtd_xx_p, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xz_o0_t15_l0_p = (-1.500000000000000 * vector(gtd_xz_p, i, j, k) * gtu_yz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xz_o0_t14_l0_p = (-1.500000000000000 * vector(gtd_xz_p, i, j, k) * gtu_xz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xz_o0_t13_l0_p = (-1.500000000000000 * vector(gtd_xz_p, i, j, k) * gtu_xy_p * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xz_o0_t12_l0_p = (-0.5 * Ct_zxz_p) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xz_o0_t11_l0_p = (-0.5 * Ct_yxz_p) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xz_o0_t10_l0_p = (-0.5 * Ct_xxz_p) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xz_o0_t9_l0_p = (-0.25 * inv_chi_p) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xz_o0_t8_l0_p = 0.5 * vector(gtd_xz_p, i, j, k) * gtu_zz_p * d_Rpd_xx_o0_t8_m0_l0;
						m_Rpd_xz_o0_t7_l0_p = vector(gtd_xz_p, i, j, k) * gtu_yz_p * d_Rpd_xx_o0_t7_m0_l0;
						m_Rpd_xz_o0_t6_l0_p = 0.5 * vector(gtd_xz_p, i, j, k) * gtu_yy_p * d_Rpd_xx_o0_t6_m0_l0;
						m_Rpd_xz_o0_t5_l0_p = vector(gtd_xz_p, i, j, k) * gtu_xz_p * d_Rpd_xx_o0_t5_m0_l0;
						m_Rpd_xz_o0_t4_l0_p = vector(gtd_xz_p, i, j, k) * gtu_xy_p * d_Rpd_xx_o0_t4_m0_l0;
						m_Rpd_xz_o0_t3_l0_p = 0.5 * vector(gtd_xz_p, i, j, k) * gtu_xx_p * d_Rpd_xx_o0_t3_m0_l0;
						m_Rpd_xz_o0_t2_l0_p = (-0.5 * vector(gtd_xz_p, i, j, k) * vector(Gamh_z_p, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xz_o0_t1_l0_p = (-0.5 * vector(gtd_xz_p, i, j, k) * vector(Gamh_y_p, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xz_o0_t0_l0_p = (-0.5 * vector(gtd_xz_p, i, j, k) * vector(Gamh_x_p, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						Rpd_xz_p = ((((((((((((((((((((((((m_Rpd_xz_o0_t0_l0_p + m_Rpd_xz_o0_t1_l0_p) + m_Rpd_xz_o0_t2_l0_p) + m_Rpd_xz_o0_t3_l0_p) + m_Rpd_xz_o0_t4_l0_p) + m_Rpd_xz_o0_t5_l0_p) + m_Rpd_xz_o0_t6_l0_p) + m_Rpd_xz_o0_t7_l0_p) + m_Rpd_xz_o0_t8_l0_p) + m_Rpd_xz_o0_t9_l0_p) + m_Rpd_xz_o0_t10_l0_p) + m_Rpd_xz_o0_t11_l0_p) + m_Rpd_xz_o0_t12_l0_p) + m_Rpd_xz_o0_t13_l0_p) + m_Rpd_xz_o0_t14_l0_p) + m_Rpd_xz_o0_t15_l0_p) + m_Rpd_xz_o0_t16_l0_p) + m_Rpd_xz_o0_t17_l0_p) + m_Rpd_xz_o0_t18_l0_p) + m_Rpd_xz_o0_t19_l0_p) + m_Rpd_xz_o0_t20_l0_p) + m_Rpd_xz_o0_t21_l0_p) + m_Rpd_xz_o0_t22_l0_p) + m_Rpd_xz_o0_t23_l0_p) + m_Rpd_xz_o0_t24_l0_p) + m_Rpd_xz_o0_t25_l0_p;
						m_Rpd_xy_o0_t25_l0_p = 0.5 * d_Rpd_xx_o0_t4_m0_l0;
						m_Rpd_xy_o0_t24_l0_p = (-0.75 * vector(gtd_xy_p, i, j, k) * gtu_zz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xy_o0_t23_l0_p = (-0.75 * vector(gtd_xy_p, i, j, k) * gtu_yy_p * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xy_o0_t22_l0_p = (-0.75 * vector(gtd_xy_p, i, j, k) * gtu_xx_p * inv_chi_p) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xy_o0_t21_l0_p = inv_chi_p * Zu_z_p * vector(gtd_yz_p, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xy_o0_t20_l0_p = inv_chi_p * Zu_y_p * vector(gtd_yy_p, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xy_o0_t19_l0_p = inv_chi_p * Zu_x_p * vector(gtd_xy_p, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xy_o0_t18_l0_p = inv_chi_p * Zu_z_p * vector(gtd_xz_p, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xy_o0_t17_l0_p = inv_chi_p * Zu_y_p * vector(gtd_xy_p, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xy_o0_t16_l0_p = inv_chi_p * Zu_x_p * vector(gtd_xx_p, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xy_o0_t15_l0_p = (-1.500000000000000 * vector(gtd_xy_p, i, j, k) * gtu_yz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xy_o0_t14_l0_p = (-1.500000000000000 * vector(gtd_xy_p, i, j, k) * gtu_xz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xy_o0_t13_l0_p = (-1.500000000000000 * vector(gtd_xy_p, i, j, k) * gtu_xy_p * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xy_o0_t12_l0_p = (-0.5 * Ct_zxy_p) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xy_o0_t11_l0_p = (-0.5 * Ct_yxy_p) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xy_o0_t10_l0_p = (-0.5 * Ct_xxy_p) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xy_o0_t9_l0_p = 0.5 * vector(gtd_xy_p, i, j, k) * gtu_zz_p * d_Rpd_xx_o0_t8_m0_l0;
						m_Rpd_xy_o0_t8_l0_p = vector(gtd_xy_p, i, j, k) * gtu_yz_p * d_Rpd_xx_o0_t7_m0_l0;
						m_Rpd_xy_o0_t7_l0_p = 0.5 * vector(gtd_xy_p, i, j, k) * gtu_yy_p * d_Rpd_xx_o0_t6_m0_l0;
						m_Rpd_xy_o0_t6_l0_p = vector(gtd_xy_p, i, j, k) * gtu_xz_p * d_Rpd_xx_o0_t5_m0_l0;
						m_Rpd_xy_o0_t5_l0_p = vector(gtd_xy_p, i, j, k) * gtu_xy_p * d_Rpd_xx_o0_t4_m0_l0;
						m_Rpd_xy_o0_t4_l0_p = 0.5 * vector(gtd_xy_p, i, j, k) * gtu_xx_p * d_Rpd_xx_o0_t3_m0_l0;
						m_Rpd_xy_o0_t3_l0_p = (-0.5 * vector(gtd_xy_p, i, j, k) * vector(Gamh_z_p, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_xy_o0_t2_l0_p = (-0.5 * vector(gtd_xy_p, i, j, k) * vector(Gamh_y_p, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_xy_o0_t1_l0_p = (-0.5 * vector(gtd_xy_p, i, j, k) * vector(Gamh_x_p, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_xy_o0_t0_l0_p = (-0.25 * inv_chi_p) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						Rpd_xy_p = ((((((((((((((((((((((((m_Rpd_xy_o0_t0_l0_p + m_Rpd_xy_o0_t1_l0_p) + m_Rpd_xy_o0_t2_l0_p) + m_Rpd_xy_o0_t3_l0_p) + m_Rpd_xy_o0_t4_l0_p) + m_Rpd_xy_o0_t5_l0_p) + m_Rpd_xy_o0_t6_l0_p) + m_Rpd_xy_o0_t7_l0_p) + m_Rpd_xy_o0_t8_l0_p) + m_Rpd_xy_o0_t9_l0_p) + m_Rpd_xy_o0_t10_l0_p) + m_Rpd_xy_o0_t11_l0_p) + m_Rpd_xy_o0_t12_l0_p) + m_Rpd_xy_o0_t13_l0_p) + m_Rpd_xy_o0_t14_l0_p) + m_Rpd_xy_o0_t15_l0_p) + m_Rpd_xy_o0_t16_l0_p) + m_Rpd_xy_o0_t17_l0_p) + m_Rpd_xy_o0_t18_l0_p) + m_Rpd_xy_o0_t19_l0_p) + m_Rpd_xy_o0_t20_l0_p) + m_Rpd_xy_o0_t21_l0_p) + m_Rpd_xy_o0_t22_l0_p) + m_Rpd_xy_o0_t23_l0_p) + m_Rpd_xy_o0_t24_l0_p) + m_Rpd_xy_o0_t25_l0_p;
						m_Rpd_zz_o0_t22_l0_p = 0.5 * d_Rpd_xx_o0_t8_m0_l0;
						m_Rpd_zz_o0_t21_l0_p = (-0.75 * vector(gtd_zz_p, i, j, k) * gtu_zz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_zz_o0_t20_l0_p = (-0.75 * vector(gtd_zz_p, i, j, k) * gtu_yy_p * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_zz_o0_t19_l0_p = (-0.75 * vector(gtd_zz_p, i, j, k) * gtu_xx_p * inv_chi_p) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_zz_o0_t18_l0_p = 2.0 * inv_chi_p * Zu_z_p * vector(gtd_zz_p, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_zz_o0_t17_l0_p = 2.0 * inv_chi_p * Zu_y_p * vector(gtd_yz_p, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_zz_o0_t16_l0_p = 2.0 * inv_chi_p * Zu_x_p * vector(gtd_xz_p, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_zz_o0_t15_l0_p = (-0.25 * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_zz_o0_t14_l0_p = (-0.5 * Ct_zzz_p) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_zz_o0_t13_l0_p = (-0.5 * Ct_yzz_p) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_zz_o0_t12_l0_p = (-0.5 * Ct_xzz_p) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_zz_o0_t11_l0_p = vector(gtd_zz_p, i, j, k) * gtu_xz_p * d_Rpd_xx_o0_t5_m0_l0;
						m_Rpd_zz_o0_t10_l0_p = vector(gtd_zz_p, i, j, k) * gtu_xy_p * d_Rpd_xx_o0_t4_m0_l0;
						m_Rpd_zz_o0_t9_l0_p = 0.5 * vector(gtd_zz_p, i, j, k) * gtu_xx_p * d_Rpd_xx_o0_t3_m0_l0;
						m_Rpd_zz_o0_t8_l0_p = (-0.5 * vector(gtd_zz_p, i, j, k) * vector(Gamh_z_p, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Rpd_zz_o0_t7_l0_p = (-0.5 * vector(gtd_zz_p, i, j, k) * vector(Gamh_y_p, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_zz_o0_t6_l0_p = (-0.5 * vector(gtd_zz_p, i, j, k) * vector(Gamh_x_p, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_zz_o0_t5_l0_p = 0.5 * vector(gtd_zz_p, i, j, k) * gtu_zz_p * d_Rpd_xx_o0_t8_m0_l0;
						m_Rpd_zz_o0_t4_l0_p = vector(gtd_zz_p, i, j, k) * gtu_yz_p * d_Rpd_xx_o0_t7_m0_l0;
						m_Rpd_zz_o0_t3_l0_p = 0.5 * vector(gtd_zz_p, i, j, k) * gtu_yy_p * d_Rpd_xx_o0_t6_m0_l0;
						m_Rpd_zz_o0_t2_l0_p = (-1.500000000000000 * vector(gtd_zz_p, i, j, k) * gtu_yz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Rpd_zz_o0_t1_l0_p = (-1.500000000000000 * vector(gtd_zz_p, i, j, k) * gtu_xz_p * inv_chi_p) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Rpd_zz_o0_t0_l0_p = (-1.500000000000000 * vector(gtd_zz_p, i, j, k) * gtu_xy_p * inv_chi_p) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						Rpd_zz_p = (((((((((((((((((((((m_Rpd_zz_o0_t0_l0_p + m_Rpd_zz_o0_t1_l0_p) + m_Rpd_zz_o0_t2_l0_p) + m_Rpd_zz_o0_t3_l0_p) + m_Rpd_zz_o0_t4_l0_p) + m_Rpd_zz_o0_t5_l0_p) + m_Rpd_zz_o0_t6_l0_p) + m_Rpd_zz_o0_t7_l0_p) + m_Rpd_zz_o0_t8_l0_p) + m_Rpd_zz_o0_t9_l0_p) + m_Rpd_zz_o0_t10_l0_p) + m_Rpd_zz_o0_t11_l0_p) + m_Rpd_zz_o0_t12_l0_p) + m_Rpd_zz_o0_t13_l0_p) + m_Rpd_zz_o0_t14_l0_p) + m_Rpd_zz_o0_t15_l0_p) + m_Rpd_zz_o0_t16_l0_p) + m_Rpd_zz_o0_t17_l0_p) + m_Rpd_zz_o0_t18_l0_p) + m_Rpd_zz_o0_t19_l0_p) + m_Rpd_zz_o0_t20_l0_p) + m_Rpd_zz_o0_t21_l0_p) + m_Rpd_zz_o0_t22_l0_p;
						Rscalar_p = (chi_max_p * Rtd_xx_p + Rpd_xx_p) * gtu_xx_p + 2.0 * (chi_max_p * Rtd_xy_p + Rpd_xy_p) * gtu_xy_p + 2.0 * (chi_max_p * Rtd_xz_p + Rpd_xz_p) * gtu_xz_p + (chi_max_p * Rtd_yy_p + Rpd_yy_p) * gtu_yy_p + 2.0 * (chi_max_p * Rtd_yz_p + Rpd_yz_p) * gtu_yz_p + (chi_max_p * Rtd_zz_p + Rpd_zz_p) * gtu_zz_p;
						Bfvf_p = vector(Bfu_x_p, i, j, k) * vector(vfd_x, i, j, k) + vector(Bfu_y_p, i, j, k) * vector(vfd_y, i, j, k) + vector(Bfu_z_p, i, j, k) * vector(vfd_z, i, j, k);
						Sfuu_zz_p = Sfu_z_p * vfu_z_p + chi_max_p * gtu_zz_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_zz_p) + vector(Bfu_z_p, i, j, k) * vector(Bfu_z_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_zz_p) + 2.0 * vector(Bfu_z_p, i, j, k) * vfu_z_p)) / sdetg_p);
						Tf4u_zz_p = Sfuu_zz_p / sdetg_p + (-(Sfu_z_p * vector(Betau_z_p, i, j, k)) / (sdetg_p * vector(Alpha_p, i, j, k))) + (-vector(Betau_z_p, i, j, k) * Tf4u_tz_p);
						Tu_zz_p = Tf4u_zz_p;
						Sfuu_yz_p = 0.5 * Sfu_y_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_y_p + chi_max_p * gtu_yz_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k) * vector(Bfu_z_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yz_p) + vector(Bfu_y_p, i, j, k) * vfu_z_p + vector(Bfu_z_p, i, j, k) * vfu_y_p)) / sdetg_p);
						Tf4u_yz_p = Sfuu_yz_p / sdetg_p + (-(Sfu_y_p * vector(Betau_z_p, i, j, k)) / (sdetg_p * vector(Alpha_p, i, j, k))) + (-vector(Betau_y_p, i, j, k) * Tf4u_tz_p);
						Tu_yz_p = Tf4u_yz_p;
						Sfuu_yy_p = Sfu_y_p * vfu_y_p + chi_max_p * gtu_yy_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_yy_p) + vector(Bfu_y_p, i, j, k) * vector(Bfu_y_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_yy_p) + 2.0 * vector(Bfu_y_p, i, j, k) * vfu_y_p)) / sdetg_p);
						Tf4u_yy_p = Sfuu_yy_p / sdetg_p + (-(Sfu_y_p * vector(Betau_y_p, i, j, k)) / (sdetg_p * vector(Alpha_p, i, j, k))) + (-vector(Betau_y_p, i, j, k) * Tf4u_ty_p);
						Tu_yy_p = Tf4u_yy_p;
						Sfuu_xy_p = 0.5 * Sfu_x_p * vfu_y_p + 0.5 * Sfu_y_p * vfu_x_p + chi_max_p * gtu_xy_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k) * vector(Bfu_y_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xy_p) + vector(Bfu_x_p, i, j, k) * vfu_y_p + vector(Bfu_y_p, i, j, k) * vfu_x_p)) / sdetg_p);
						Sfud_yy_p = inv_chi_p * (Sfuu_xy_p * vector(gtd_xy_p, i, j, k) + Sfuu_yy_p * vector(gtd_yy_p, i, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k));
						Tf4u_xy_p = Sfuu_xy_p / sdetg_p + (-(Sfu_x_p * vector(Betau_y_p, i, j, k)) / (sdetg_p * vector(Alpha_p, i, j, k))) + (-vector(Betau_x_p, i, j, k) * Tf4u_ty_p);
						Tu_xy_p = Tf4u_xy_p;
						Sfuu_xx_p = Sfu_x_p * vfu_x_p + chi_max_p * gtu_xx_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xx_p) + vector(Bfu_x_p, i, j, k) * vector(Bfu_x_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xx_p) + 2.0 * vector(Bfu_x_p, i, j, k) * vfu_x_p)) / sdetg_p);
						Tf4u_xx_p = Sfuu_xx_p / sdetg_p + (-(Sfu_x_p * vector(Betau_x_p, i, j, k)) / (sdetg_p * vector(Alpha_p, i, j, k))) + (-vector(Betau_x_p, i, j, k) * Tf4u_tx_p);
						Tu_xx_p = Tf4u_xx_p;
						Sfuu_xz_p = 0.5 * Sfu_x_p * vfu_z_p + 0.5 * Sfu_z_p * vfu_x_p + chi_max_p * gtu_xz_p * sdetg_p * vector(pf, i, j, k) + (-((-0.5 * sqBf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k) * vector(Bfu_z_p, i, j, k)) / (sdetg_p * sqW_p)) + (-0.5 * (Bfvf_p * ((-Bfvf_p * chi_max_p * gtu_xz_p) + vector(Bfu_x_p, i, j, k) * vfu_z_p + vector(Bfu_z_p, i, j, k) * vfu_x_p)) / sdetg_p);
						Sfud_xx_p = inv_chi_p * (Sfuu_xx_p * vector(gtd_xx_p, i, j, k) + Sfuu_xy_p * vector(gtd_xy_p, i, j, k) + Sfuu_xz_p * vector(gtd_xz_p, i, j, k));
						Tf4u_xz_p = Sfuu_xz_p / sdetg_p + (-(Sfu_x_p * vector(Betau_z_p, i, j, k)) / (sdetg_p * vector(Alpha_p, i, j, k))) + (-vector(Betau_x_p, i, j, k) * Tf4u_tz_p);
						Tu_xz_p = Tf4u_xz_p;
						Sfud_zz_p = inv_chi_p * (Sfuu_xz_p * vector(gtd_xz_p, i, j, k) + Sfuu_yz_p * vector(gtd_yz_p, i, j, k) + Sfuu_zz_p * vector(gtd_zz_p, i, j, k));
						trSf_p = Sfud_xx_p + Sfud_yy_p + Sfud_zz_p;
						Tu_tt_p = Tf4u_tt_p;
						pTtd_ADM_xz_p = Betatd_x_p * Betatd_z_p * Tu_tt_p + Betatd_x_p * (vector(gtd_xz_p, i, j, k) * Tu_tx_p + vector(gtd_yz_p, i, j, k) * Tu_ty_p + vector(gtd_zz_p, i, j, k) * Tu_tz_p) + Betatd_z_p * (vector(gtd_xx_p, i, j, k) * Tu_tx_p + vector(gtd_xy_p, i, j, k) * Tu_ty_p + vector(gtd_xz_p, i, j, k) * Tu_tz_p) + vector(gtd_xx_p, i, j, k) * vector(gtd_xz_p, i, j, k) * Tu_xx_p + vector(gtd_xx_p, i, j, k) * vector(gtd_yz_p, i, j, k) * Tu_xy_p + vector(gtd_xx_p, i, j, k) * vector(gtd_zz_p, i, j, k) * Tu_xz_p + vector(gtd_xy_p, i, j, k) * vector(gtd_xz_p, i, j, k) * Tu_xy_p + vector(gtd_xy_p, i, j, k) * vector(gtd_yz_p, i, j, k) * Tu_yy_p + vector(gtd_xy_p, i, j, k) * vector(gtd_zz_p, i, j, k) * Tu_yz_p + (vector(gtd_xz_p, i, j, k) * vector(gtd_xz_p, i, j, k)) * Tu_xz_p + vector(gtd_xz_p, i, j, k) * vector(gtd_yz_p, i, j, k) * Tu_yz_p + vector(gtd_xz_p, i, j, k) * vector(gtd_zz_p, i, j, k) * Tu_zz_p;
						pTtd_ADM_xy_p = Betatd_x_p * Betatd_y_p * Tu_tt_p + Betatd_x_p * (vector(gtd_xy_p, i, j, k) * Tu_tx_p + vector(gtd_yy_p, i, j, k) * Tu_ty_p + vector(gtd_yz_p, i, j, k) * Tu_tz_p) + Betatd_y_p * (vector(gtd_xx_p, i, j, k) * Tu_tx_p + vector(gtd_xy_p, i, j, k) * Tu_ty_p + vector(gtd_xz_p, i, j, k) * Tu_tz_p) + vector(gtd_xx_p, i, j, k) * vector(gtd_xy_p, i, j, k) * Tu_xx_p + vector(gtd_xx_p, i, j, k) * vector(gtd_yy_p, i, j, k) * Tu_xy_p + vector(gtd_xx_p, i, j, k) * vector(gtd_yz_p, i, j, k) * Tu_xz_p + (vector(gtd_xy_p, i, j, k) * vector(gtd_xy_p, i, j, k)) * Tu_xy_p + vector(gtd_xy_p, i, j, k) * vector(gtd_yy_p, i, j, k) * Tu_yy_p + vector(gtd_xy_p, i, j, k) * vector(gtd_yz_p, i, j, k) * Tu_yz_p + vector(gtd_xz_p, i, j, k) * vector(gtd_xy_p, i, j, k) * Tu_xz_p + vector(gtd_xz_p, i, j, k) * vector(gtd_yy_p, i, j, k) * Tu_yz_p + vector(gtd_xz_p, i, j, k) * vector(gtd_yz_p, i, j, k) * Tu_zz_p;
						pTtd_ADM_xx_p = (Betatd_x_p * Betatd_x_p) * Tu_tt_p + 2.0 * Betatd_x_p * (vector(gtd_xx_p, i, j, k) * Tu_tx_p + vector(gtd_xy_p, i, j, k) * Tu_ty_p + vector(gtd_xz_p, i, j, k) * Tu_tz_p) + (vector(gtd_xx_p, i, j, k) * vector(gtd_xx_p, i, j, k)) * Tu_xx_p + 2.0 * vector(gtd_xx_p, i, j, k) * vector(gtd_xy_p, i, j, k) * Tu_xy_p + 2.0 * vector(gtd_xx_p, i, j, k) * vector(gtd_xz_p, i, j, k) * Tu_xz_p + (vector(gtd_xy_p, i, j, k) * vector(gtd_xy_p, i, j, k)) * Tu_yy_p + 2.0 * vector(gtd_xy_p, i, j, k) * vector(gtd_xz_p, i, j, k) * Tu_yz_p + (vector(gtd_xz_p, i, j, k) * vector(gtd_xz_p, i, j, k)) * Tu_zz_p;
						d_Psi1_xx_o0_t5_m0_l0_p = (-8.0) * M_PI * inv_chi_p * vector(Alpha_p, i, j, k) * pTtd_ADM_xx_p + vector(Alpha_p, i, j, k) * Rpd_xx_p + chi_max_p * vector(Alpha_p, i, j, k) * Rtd_xx_p;
						m_Psi1_xx_o0_t4_l0_p = (-chi_max_p * inv_chi_p) * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Psi1_xx_o0_t3_l0_p = chi_max_p * Ct_zxx_p * d_Psi1_xx_o0_t3_m0_l0;
						m_Psi1_xx_o0_t2_l0_p = chi_max_p * Ct_yxx_p * d_Psi1_xx_o0_t2_m0_l0;
						m_Psi1_xx_o0_t1_l0_p = chi_max_p * Ct_xxx_p * d_Psi1_xx_o0_t1_m0_l0;
						m_Psi1_xx_o0_t0_l0_p = (-chi_max_p) * d_Psi1_xx_o0_t0_m0_l0;
						Psi1_xx_p = ((((m_Psi1_xx_o0_t0_l0_p + m_Psi1_xx_o0_t1_l0_p) + m_Psi1_xx_o0_t2_l0_p) + m_Psi1_xx_o0_t3_l0_p) + m_Psi1_xx_o0_t4_l0_p) + d_Psi1_xx_o0_t5_m0_l0_p;
						d_Psi1_xy_o0_t6_m0_l0_p = (-8.0) * M_PI * inv_chi_p * vector(Alpha_p, i, j, k) * pTtd_ADM_xy_p + vector(Alpha_p, i, j, k) * Rpd_xy_p + chi_max_p * vector(Alpha_p, i, j, k) * Rtd_xy_p;
						m_Psi1_xy_o0_t5_l0_p = (-0.5 * chi_max_p * inv_chi_p) * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Psi1_xy_o0_t4_l0_p = (-0.5 * chi_max_p * inv_chi_p) * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Psi1_xy_o0_t3_l0_p = chi_max_p * Ct_zxy_p * d_Psi1_xx_o0_t3_m0_l0;
						m_Psi1_xy_o0_t2_l0_p = chi_max_p * Ct_yxy_p * d_Psi1_xx_o0_t2_m0_l0;
						m_Psi1_xy_o0_t1_l0_p = chi_max_p * Ct_xxy_p * d_Psi1_xx_o0_t1_m0_l0;
						m_Psi1_xy_o0_t0_l0_p = (-chi_max_p) * d_Psi1_xy_o0_t0_m0_l0;
						Psi1_xy_p = (((((m_Psi1_xy_o0_t0_l0_p + m_Psi1_xy_o0_t1_l0_p) + m_Psi1_xy_o0_t2_l0_p) + m_Psi1_xy_o0_t3_l0_p) + m_Psi1_xy_o0_t4_l0_p) + m_Psi1_xy_o0_t5_l0_p) + d_Psi1_xy_o0_t6_m0_l0_p;
						d_Psi1_xz_o0_t6_m0_l0_p = (-8.0) * M_PI * inv_chi_p * vector(Alpha_p, i, j, k) * pTtd_ADM_xz_p + vector(Alpha_p, i, j, k) * Rpd_xz_p + chi_max_p * vector(Alpha_p, i, j, k) * Rtd_xz_p;
						m_Psi1_xz_o0_t5_l0_p = (-0.5 * chi_max_p * inv_chi_p) * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_Psi1_xz_o0_t4_l0_p = (-0.5 * chi_max_p * inv_chi_p) * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_Psi1_xz_o0_t3_l0_p = chi_max_p * Ct_zxz_p * d_Psi1_xx_o0_t3_m0_l0;
						m_Psi1_xz_o0_t2_l0_p = chi_max_p * Ct_yxz_p * d_Psi1_xx_o0_t2_m0_l0;
						m_Psi1_xz_o0_t1_l0_p = chi_max_p * Ct_xxz_p * d_Psi1_xx_o0_t1_m0_l0;
						m_Psi1_xz_o0_t0_l0_p = (-chi_max_p) * d_Psi1_xz_o0_t0_m0_l0;
						Psi1_xz_p = (((((m_Psi1_xz_o0_t0_l0_p + m_Psi1_xz_o0_t1_l0_p) + m_Psi1_xz_o0_t2_l0_p) + m_Psi1_xz_o0_t3_l0_p) + m_Psi1_xz_o0_t4_l0_p) + m_Psi1_xz_o0_t5_l0_p) + d_Psi1_xz_o0_t6_m0_l0_p;
						Jtd_ADM_z_p = vector(Alpha_p, i, j, k) * ((vector(Betau_x_p, i, j, k) * Tu_tt_p + Tu_tx_p) * vector(gtd_xz_p, i, j, k) + (vector(Betau_y_p, i, j, k) * Tu_tt_p + Tu_ty_p) * vector(gtd_yz_p, i, j, k) + (vector(Betau_z_p, i, j, k) * Tu_tt_p + Tu_tz_p) * vector(gtd_zz_p, i, j, k));
						pTtd_ADM_zz_p = (Betatd_z_p * Betatd_z_p) * Tu_tt_p + 2.0 * Betatd_z_p * (vector(gtd_xz_p, i, j, k) * Tu_tx_p + vector(gtd_yz_p, i, j, k) * Tu_ty_p + vector(gtd_zz_p, i, j, k) * Tu_tz_p) + (vector(gtd_xz_p, i, j, k) * vector(gtd_xz_p, i, j, k)) * Tu_xx_p + 2.0 * vector(gtd_xz_p, i, j, k) * vector(gtd_yz_p, i, j, k) * Tu_xy_p + 2.0 * vector(gtd_xz_p, i, j, k) * vector(gtd_zz_p, i, j, k) * Tu_xz_p + (vector(gtd_yz_p, i, j, k) * vector(gtd_yz_p, i, j, k)) * Tu_yy_p + 2.0 * vector(gtd_yz_p, i, j, k) * vector(gtd_zz_p, i, j, k) * Tu_yz_p + (vector(gtd_zz_p, i, j, k) * vector(gtd_zz_p, i, j, k)) * Tu_zz_p;
						d_Psi1_zz_o0_t5_m0_l0_p = (-8.0) * M_PI * inv_chi_p * vector(Alpha_p, i, j, k) * pTtd_ADM_zz_p + vector(Alpha_p, i, j, k) * Rpd_zz_p + chi_max_p * vector(Alpha_p, i, j, k) * Rtd_zz_p;
						m_Psi1_zz_o0_t4_l0_p = (-chi_max_p * inv_chi_p) * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_Psi1_zz_o0_t3_l0_p = chi_max_p * Ct_zzz_p * d_Psi1_xx_o0_t3_m0_l0;
						m_Psi1_zz_o0_t2_l0_p = chi_max_p * Ct_yzz_p * d_Psi1_xx_o0_t2_m0_l0;
						m_Psi1_zz_o0_t1_l0_p = chi_max_p * Ct_xzz_p * d_Psi1_xx_o0_t1_m0_l0;
						m_Psi1_zz_o0_t0_l0_p = (-chi_max_p) * d_Psi1_zz_o0_t0_m0_l0;
						Psi1_zz_p = ((((m_Psi1_zz_o0_t0_l0_p + m_Psi1_zz_o0_t1_l0_p) + m_Psi1_zz_o0_t2_l0_p) + m_Psi1_zz_o0_t3_l0_p) + m_Psi1_zz_o0_t4_l0_p) + d_Psi1_zz_o0_t5_m0_l0_p;
						Jtd_ADM_y_p = vector(Alpha_p, i, j, k) * ((vector(Betau_x_p, i, j, k) * Tu_tt_p + Tu_tx_p) * vector(gtd_xy_p, i, j, k) + (vector(Betau_y_p, i, j, k) * Tu_tt_p + Tu_ty_p) * vector(gtd_yy_p, i, j, k) + (vector(Betau_z_p, i, j, k) * Tu_tt_p + Tu_tz_p) * vector(gtd_yz_p, i, j, k));
						Jtd_ADM_x_p = vector(Alpha_p, i, j, k) * ((vector(Betau_x_p, i, j, k) * Tu_tt_p + Tu_tx_p) * vector(gtd_xx_p, i, j, k) + (vector(Betau_y_p, i, j, k) * Tu_tt_p + Tu_ty_p) * vector(gtd_xy_p, i, j, k) + (vector(Betau_z_p, i, j, k) * Tu_tt_p + Tu_tz_p) * vector(gtd_xz_p, i, j, k));
						pTtd_ADM_yz_p = Betatd_y_p * Betatd_z_p * Tu_tt_p + Betatd_y_p * (vector(gtd_xz_p, i, j, k) * Tu_tx_p + vector(gtd_yz_p, i, j, k) * Tu_ty_p + vector(gtd_zz_p, i, j, k) * Tu_tz_p) + Betatd_z_p * (vector(gtd_xy_p, i, j, k) * Tu_tx_p + vector(gtd_yy_p, i, j, k) * Tu_ty_p + vector(gtd_yz_p, i, j, k) * Tu_tz_p) + vector(gtd_xy_p, i, j, k) * vector(gtd_xz_p, i, j, k) * Tu_xx_p + vector(gtd_xy_p, i, j, k) * vector(gtd_yz_p, i, j, k) * Tu_xy_p + vector(gtd_xy_p, i, j, k) * vector(gtd_zz_p, i, j, k) * Tu_xz_p + vector(gtd_yy_p, i, j, k) * vector(gtd_xz_p, i, j, k) * Tu_xy_p + vector(gtd_yy_p, i, j, k) * vector(gtd_yz_p, i, j, k) * Tu_yy_p + vector(gtd_yy_p, i, j, k) * vector(gtd_zz_p, i, j, k) * Tu_yz_p + vector(gtd_yz_p, i, j, k) * vector(gtd_xz_p, i, j, k) * Tu_xz_p + (vector(gtd_yz_p, i, j, k) * vector(gtd_yz_p, i, j, k)) * Tu_yz_p + vector(gtd_yz_p, i, j, k) * vector(gtd_zz_p, i, j, k) * Tu_zz_p;
						d_Psi1_yz_o0_t6_m0_l0_p = (-8.0) * M_PI * inv_chi_p * vector(Alpha_p, i, j, k) * pTtd_ADM_yz_p + vector(Alpha_p, i, j, k) * Rpd_yz_p + chi_max_p * vector(Alpha_p, i, j, k) * Rtd_yz_p;
						m_Psi1_yz_o0_t5_l0_p = (-0.5 * chi_max_p * inv_chi_p) * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Psi1_yz_o0_t4_l0_p = (-0.5 * chi_max_p * inv_chi_p) * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_Psi1_yz_o0_t3_l0_p = chi_max_p * Ct_zyz_p * d_Psi1_xx_o0_t3_m0_l0;
						m_Psi1_yz_o0_t2_l0_p = chi_max_p * Ct_yyz_p * d_Psi1_xx_o0_t2_m0_l0;
						m_Psi1_yz_o0_t1_l0_p = chi_max_p * Ct_xyz_p * d_Psi1_xx_o0_t1_m0_l0;
						m_Psi1_yz_o0_t0_l0_p = (-chi_max_p) * d_Psi1_yz_o0_t0_m0_l0;
						Psi1_yz_p = (((((m_Psi1_yz_o0_t0_l0_p + m_Psi1_yz_o0_t1_l0_p) + m_Psi1_yz_o0_t2_l0_p) + m_Psi1_yz_o0_t3_l0_p) + m_Psi1_yz_o0_t4_l0_p) + m_Psi1_yz_o0_t5_l0_p) + d_Psi1_yz_o0_t6_m0_l0_p;
						pTtd_ADM_yy_p = (Betatd_y_p * Betatd_y_p) * Tu_tt_p + 2.0 * Betatd_y_p * (vector(gtd_xy_p, i, j, k) * Tu_tx_p + vector(gtd_yy_p, i, j, k) * Tu_ty_p + vector(gtd_yz_p, i, j, k) * Tu_tz_p) + (vector(gtd_xy_p, i, j, k) * vector(gtd_xy_p, i, j, k)) * Tu_xx_p + 2.0 * vector(gtd_xy_p, i, j, k) * vector(gtd_yy_p, i, j, k) * Tu_xy_p + 2.0 * vector(gtd_xy_p, i, j, k) * vector(gtd_yz_p, i, j, k) * Tu_xz_p + (vector(gtd_yy_p, i, j, k) * vector(gtd_yy_p, i, j, k)) * Tu_yy_p + 2.0 * vector(gtd_yy_p, i, j, k) * vector(gtd_yz_p, i, j, k) * Tu_yz_p + (vector(gtd_yz_p, i, j, k) * vector(gtd_yz_p, i, j, k)) * Tu_zz_p;
						tr_pT_p = inv_chi_p * (gtu_xx_p * pTtd_ADM_xx_p + 2.0 * gtu_xy_p * pTtd_ADM_xy_p + 2.0 * gtu_xz_p * pTtd_ADM_xz_p + gtu_yy_p * pTtd_ADM_yy_p + 2.0 * gtu_yz_p * pTtd_ADM_yz_p + gtu_zz_p * pTtd_ADM_zz_p);
						d_Psi1_yy_o0_t5_m0_l0_p = (-8.0) * M_PI * inv_chi_p * vector(Alpha_p, i, j, k) * pTtd_ADM_yy_p + vector(Alpha_p, i, j, k) * Rpd_yy_p + chi_max_p * vector(Alpha_p, i, j, k) * Rtd_yy_p;
						m_Psi1_yy_o0_t4_l0_p = (-chi_max_p * inv_chi_p) * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_Psi1_yy_o0_t3_l0_p = chi_max_p * Ct_zyy_p * d_Psi1_xx_o0_t3_m0_l0;
						m_Psi1_yy_o0_t2_l0_p = chi_max_p * Ct_yyy_p * d_Psi1_xx_o0_t2_m0_l0;
						m_Psi1_yy_o0_t1_l0_p = chi_max_p * Ct_xyy_p * d_Psi1_xx_o0_t1_m0_l0;
						m_Psi1_yy_o0_t0_l0_p = (-chi_max_p) * d_Psi1_yy_o0_t0_m0_l0;
						Psi1_yy_p = ((((m_Psi1_yy_o0_t0_l0_p + m_Psi1_yy_o0_t1_l0_p) + m_Psi1_yy_o0_t2_l0_p) + m_Psi1_yy_o0_t3_l0_p) + m_Psi1_yy_o0_t4_l0_p) + d_Psi1_yy_o0_t5_m0_l0_p;
						trPsi1_p = Psi1_xx_p * gtu_xx_p + 2.0 * Psi1_xy_p * gtu_xy_p + 2.0 * Psi1_xz_p * gtu_xz_p + Psi1_yy_p * gtu_yy_p + 2.0 * Psi1_yz_p * gtu_yz_p + Psi1_zz_p * gtu_zz_p;
						Psi1TF_yz_p = (-0.3333333333333333 * trPsi1_p * vector(gtd_yz_p, i, j, k)) + Psi1_yz_p;
						Psi1TF_yy_p = (-0.3333333333333333 * trPsi1_p * vector(gtd_yy_p, i, j, k)) + Psi1_yy_p;
						Psi1TF_xy_p = (-0.3333333333333333 * trPsi1_p * vector(gtd_xy_p, i, j, k)) + Psi1_xy_p;
						Psi1TF_xx_p = (-0.3333333333333333 * trPsi1_p * vector(gtd_xx_p, i, j, k)) + Psi1_xx_p;
						Psi1TF_xz_p = (-0.3333333333333333 * trPsi1_p * vector(gtd_xz_p, i, j, k)) + Psi1_xz_p;
						Psi1TF_zz_p = (-0.3333333333333333 * trPsi1_p * vector(gtd_zz_p, i, j, k)) + Psi1_zz_p;
						kappa_cc_p = MIN(p_kappa_cc, p_kappa_cc * decay_factor_p) / vector(Alpha_p, i, j, k);
						rho_ADM_p = (vector(Alpha_p, i, j, k) * vector(Alpha_p, i, j, k)) * Tu_tt_p;
						d_Sfd_x_o0_t11_m0_l0 = vector(Alpha_p, i, j, k) * sdetg_p * sqW_p * vector(qnu, i, j, k) * vfu_x_p;
						d_Sfd_y_o0_t11_m0_l0 = vector(Alpha_p, i, j, k) * sdetg_p * sqW_p * vector(qnu, i, j, k) * vfu_y_p;
						d_Sfd_z_o0_t11_m0_l0 = vector(Alpha_p, i, j, k) * sdetg_p * sqW_p * vector(qnu, i, j, k) * vfu_z_p;
						d_DYf_o0_t0_m0_l0 = vector(Alpha_p, i, j, k) * sdetg_p * vector(rnu, i, j, k);
						d_tauf_o0_t3_m0_l0 = vector(Alpha_p, i, j, k) * sdetg_p * sqW_p * vector(qnu, i, j, k);
						d_gtd_xx_o0_t3_m0_l0 = (-0.3333333333333333) * kappa_cc_p * vector(Alpha_p, i, j, k) * vector(gtd_xx_p, i, j, k) * log(detgtd_p) + (-2.0) * vector(Alpha_p, i, j, k) * ((-0.3333333333333333 * trAt_p * lambda_0 * vector(gtd_xx_p, i, j, k)) + vector(Atd_xx_p, i, j, k)) + (-0.6666666666666667) * vector(gtd_xx_p, i, j, k) * div_Beta_p;
						d_gtd_xy_o0_t6_m0_l0 = (-0.3333333333333333) * kappa_cc_p * vector(Alpha_p, i, j, k) * vector(gtd_xy_p, i, j, k) * log(detgtd_p) + (-2.0) * vector(Alpha_p, i, j, k) * ((-0.3333333333333333 * trAt_p * lambda_0 * vector(gtd_xy_p, i, j, k)) + vector(Atd_xy_p, i, j, k)) + (-0.6666666666666667) * vector(gtd_xy_p, i, j, k) * div_Beta_p;
						d_gtd_xz_o0_t6_m0_l0 = (-0.3333333333333333) * kappa_cc_p * vector(Alpha_p, i, j, k) * vector(gtd_xz_p, i, j, k) * log(detgtd_p) + (-2.0) * vector(Alpha_p, i, j, k) * ((-0.3333333333333333 * trAt_p * lambda_0 * vector(gtd_xz_p, i, j, k)) + vector(Atd_xz_p, i, j, k)) + (-0.6666666666666667) * vector(gtd_xz_p, i, j, k) * div_Beta_p;
						d_gtd_yy_o0_t3_m0_l0 = (-0.3333333333333333) * kappa_cc_p * vector(Alpha_p, i, j, k) * vector(gtd_yy_p, i, j, k) * log(detgtd_p) + (-2.0) * vector(Alpha_p, i, j, k) * ((-0.3333333333333333 * trAt_p * lambda_0 * vector(gtd_yy_p, i, j, k)) + vector(Atd_yy_p, i, j, k)) + (-0.6666666666666667) * vector(gtd_yy_p, i, j, k) * div_Beta_p;
						d_gtd_yz_o0_t6_m0_l0 = (-0.3333333333333333) * kappa_cc_p * vector(Alpha_p, i, j, k) * vector(gtd_yz_p, i, j, k) * log(detgtd_p) + (-2.0) * vector(Alpha_p, i, j, k) * ((-0.3333333333333333 * trAt_p * lambda_0 * vector(gtd_yz_p, i, j, k)) + vector(Atd_yz_p, i, j, k)) + (-0.6666666666666667) * vector(gtd_yz_p, i, j, k) * div_Beta_p;
						d_gtd_zz_o0_t3_m0_l0 = (-0.3333333333333333) * kappa_cc_p * vector(Alpha_p, i, j, k) * vector(gtd_zz_p, i, j, k) * log(detgtd_p) + (-2.0) * vector(Alpha_p, i, j, k) * ((-0.3333333333333333 * trAt_p * lambda_0 * vector(gtd_zz_p, i, j, k)) + vector(Atd_zz_p, i, j, k)) + (-0.6666666666666667) * vector(gtd_zz_p, i, j, k) * div_Beta_p;
						d_Atd_xx_o0_t3_m0_l0 = (-0.3333333333333333) * kappa_cc_p * vector(Alpha_p, i, j, k) * vector(gtd_xx_p, i, j, k) * trAt_p + vector(Alpha_p, i, j, k) * (vector(trK_p, i, j, k) * vector(Atd_xx_p, i, j, k) + (-2.0 * vector(Atd_xx_p, i, j, k) * Atud_xx_p) + (-2.0 * vector(Atd_xy_p, i, j, k) * Atud_yx_p) + (-2.0 * vector(Atd_xz_p, i, j, k) * Atud_zx_p)) + Psi1TF_xx_p + (-0.6666666666666667) * vector(Atd_xx_p, i, j, k) * div_Beta_p;
						d_Atd_xy_o0_t6_m0_l0 = (-0.3333333333333333) * kappa_cc_p * vector(Alpha_p, i, j, k) * vector(gtd_xy_p, i, j, k) * trAt_p + vector(Alpha_p, i, j, k) * (vector(trK_p, i, j, k) * vector(Atd_xy_p, i, j, k) + (-2.0 * vector(Atd_xx_p, i, j, k) * Atud_xy_p) + (-2.0 * vector(Atd_xy_p, i, j, k) * Atud_yy_p) + (-2.0 * vector(Atd_xz_p, i, j, k) * Atud_zy_p)) + Psi1TF_xy_p + (-0.6666666666666667) * vector(Atd_xy_p, i, j, k) * div_Beta_p;
						d_Atd_xz_o0_t6_m0_l0 = (-0.3333333333333333) * kappa_cc_p * vector(Alpha_p, i, j, k) * vector(gtd_xz_p, i, j, k) * trAt_p + vector(Alpha_p, i, j, k) * (vector(trK_p, i, j, k) * vector(Atd_xz_p, i, j, k) + (-2.0 * vector(Atd_xx_p, i, j, k) * Atud_xz_p) + (-2.0 * vector(Atd_xy_p, i, j, k) * Atud_yz_p) + (-2.0 * vector(Atd_xz_p, i, j, k) * Atud_zz_p)) + Psi1TF_xz_p + (-0.6666666666666667) * vector(Atd_xz_p, i, j, k) * div_Beta_p;
						d_Atd_yy_o0_t3_m0_l0 = (-0.3333333333333333) * kappa_cc_p * vector(Alpha_p, i, j, k) * vector(gtd_yy_p, i, j, k) * trAt_p + vector(Alpha_p, i, j, k) * (vector(trK_p, i, j, k) * vector(Atd_yy_p, i, j, k) + (-2.0 * vector(Atd_xy_p, i, j, k) * Atud_xy_p) + (-2.0 * vector(Atd_yy_p, i, j, k) * Atud_yy_p) + (-2.0 * vector(Atd_yz_p, i, j, k) * Atud_zy_p)) + Psi1TF_yy_p + (-0.6666666666666667) * vector(Atd_yy_p, i, j, k) * div_Beta_p;
						d_Atd_yz_o0_t6_m0_l0 = (-0.3333333333333333) * kappa_cc_p * vector(Alpha_p, i, j, k) * vector(gtd_yz_p, i, j, k) * trAt_p + vector(Alpha_p, i, j, k) * (vector(trK_p, i, j, k) * vector(Atd_yz_p, i, j, k) + (-2.0 * vector(Atd_xy_p, i, j, k) * Atud_xz_p) + (-2.0 * vector(Atd_yy_p, i, j, k) * Atud_yz_p) + (-2.0 * vector(Atd_yz_p, i, j, k) * Atud_zz_p)) + Psi1TF_yz_p + (-0.6666666666666667) * vector(Atd_yz_p, i, j, k) * div_Beta_p;
						d_Atd_zz_o0_t3_m0_l0 = (-0.3333333333333333) * kappa_cc_p * vector(Alpha_p, i, j, k) * vector(gtd_zz_p, i, j, k) * trAt_p + vector(Alpha_p, i, j, k) * (vector(trK_p, i, j, k) * vector(Atd_zz_p, i, j, k) + (-2.0 * vector(Atd_xz_p, i, j, k) * Atud_xz_p) + (-2.0 * vector(Atd_yz_p, i, j, k) * Atud_yz_p) + (-2.0 * vector(Atd_zz_p, i, j, k) * Atud_zz_p)) + Psi1TF_zz_p + (-0.6666666666666667) * vector(Atd_zz_p, i, j, k) * div_Beta_p;
						d_Gamh_x_o0_t24_m0_l0 = (-2.0) * vector(Alpha_p, i, j, k) * inv_chi_p * Zu_x_p * (kappa_z1_p + 1.333333333333333 * vector(theta_p, i, j, k) + 0.6666666666666667 * vector(trK_p, i, j, k)) + 0.3333333333333333 * gtu_xz_p * d_div_Beta_z_p + 0.3333333333333333 * gtu_xy_p * d_div_Beta_y_p + 0.3333333333333333 * gtu_xx_p * d_div_Beta_x_p + (-50.26548245743669) * vector(Alpha_p, i, j, k) * inv_chi_p * (gtu_xx_p * Jtd_ADM_x_p + gtu_xy_p * Jtd_ADM_y_p + gtu_xz_p * Jtd_ADM_z_p) + 2.0 * vector(Alpha_p, i, j, k) * (Ct_xxx_p * Atu_xx_p + 2.0 * Ct_xxy_p * Atu_xy_p + 2.0 * Ct_xxz_p * Atu_xz_p + Ct_xyy_p * Atu_yy_p + 2.0 * Ct_xyz_p * Atu_yz_p + Ct_xzz_p * Atu_zz_p) + 0.6666666666666667 * vector(Gamh_x_p, i, j, k) * div_Beta_p;
						d_Gamh_y_o0_t24_m0_l0 = (-2.0) * vector(Alpha_p, i, j, k) * inv_chi_p * Zu_y_p * (kappa_z1_p + 1.333333333333333 * vector(theta_p, i, j, k) + 0.6666666666666667 * vector(trK_p, i, j, k)) + (-50.26548245743669) * vector(Alpha_p, i, j, k) * inv_chi_p * (gtu_xy_p * Jtd_ADM_x_p + gtu_yy_p * Jtd_ADM_y_p + gtu_yz_p * Jtd_ADM_z_p) + 0.3333333333333333 * gtu_yy_p * d_div_Beta_y_p + 0.3333333333333333 * gtu_xy_p * d_div_Beta_x_p + 2.0 * vector(Alpha_p, i, j, k) * (Ct_yxx_p * Atu_xx_p + 2.0 * Ct_yxy_p * Atu_xy_p + 2.0 * Ct_yxz_p * Atu_xz_p + Ct_yyy_p * Atu_yy_p + 2.0 * Ct_yyz_p * Atu_yz_p + Ct_yzz_p * Atu_zz_p) + 0.6666666666666667 * vector(Gamh_y_p, i, j, k) * div_Beta_p + 0.3333333333333333 * gtu_yz_p * d_div_Beta_z_p;
						d_Gamh_z_o0_t24_m0_l0 = (-2.0) * vector(Alpha_p, i, j, k) * inv_chi_p * Zu_z_p * (kappa_z1_p + 1.333333333333333 * vector(theta_p, i, j, k) + 0.6666666666666667 * vector(trK_p, i, j, k)) + (-50.26548245743669) * vector(Alpha_p, i, j, k) * inv_chi_p * (gtu_xz_p * Jtd_ADM_x_p + gtu_yz_p * Jtd_ADM_y_p + gtu_zz_p * Jtd_ADM_z_p) + 2.0 * vector(Alpha_p, i, j, k) * (Ct_zxx_p * Atu_xx_p + 2.0 * Ct_zxy_p * Atu_xy_p + 2.0 * Ct_zxz_p * Atu_xz_p + Ct_zyy_p * Atu_yy_p + 2.0 * Ct_zyz_p * Atu_yz_p + Ct_zzz_p * Atu_zz_p) + 0.6666666666666667 * vector(Gamh_z_p, i, j, k) * div_Beta_p + 0.3333333333333333 * gtu_zz_p * d_div_Beta_z_p + 0.3333333333333333 * gtu_yz_p * d_div_Beta_y_p + 0.3333333333333333 * gtu_xz_p * d_div_Beta_x_p;
						d_Betau_x_o0_t0_m0_l0 = (Betau_x_0 - vector(Betau_x_p, i, j, k)) * feta_p + 0.75 * (vector(Alpha_p, i, j, k) * lambda_f1 + lambda_f0) * vector(Gamh_x_p, i, j, k);
						d_Betau_y_o0_t0_m0_l0 = (Betau_y_0 - vector(Betau_y_p, i, j, k)) * feta_p + 0.75 * (vector(Alpha_p, i, j, k) * lambda_f1 + lambda_f0) * vector(Gamh_y_p, i, j, k);
						d_Betau_z_o0_t0_m0_l0 = (Betau_z_0 - vector(Betau_z_p, i, j, k)) * feta_p + 0.75 * (vector(Alpha_p, i, j, k) * lambda_f1 + lambda_f0) * vector(Gamh_z_p, i, j, k);
						d_Alpha_o0_t0_m0_l0 = (-2.0) * (vector(Alpha_p, i, j, k) * lambda_f3 + lambda_f2) * vector(Alpha_p, i, j, k) * (vector(trK_p, i, j, k) - trK0);
						d_chi_o0_t0_m0_l0 = 0.6666666666666667 * chi_max_p * (vector(Alpha_p, i, j, k) * (2.0 * vector(theta_p, i, j, k) + vector(trK_p, i, j, k)) - div_Beta_p);
						d_trK_o0_t21_m0_l0 = vector(Alpha_p, i, j, k) * kappa_z1_p * (1.0 - kappa_z2_p) * vector(theta_p, i, j, k) + 0.3333333333333333 * vector(Alpha_p, i, j, k) * ((2.0 * vector(theta_p, i, j, k) + vector(trK_p, i, j, k)) * (2.0 * vector(theta_p, i, j, k) + vector(trK_p, i, j, k))) + vector(Alpha_p, i, j, k) * (vector(Atd_xx_p, i, j, k) * Atu_xx_p + 2.0 * vector(Atd_xy_p, i, j, k) * Atu_xy_p + 2.0 * vector(Atd_xz_p, i, j, k) * Atu_xz_p + vector(Atd_yy_p, i, j, k) * Atu_yy_p + 2.0 * vector(Atd_yz_p, i, j, k) * Atu_yz_p + vector(Atd_zz_p, i, j, k) * Atu_zz_p) + 4.0 * vector(Alpha_p, i, j, k) * M_PI * (rho_ADM_p + tr_pT_p);
						d_theta_o0_t3_m0_l0 = (-vector(Alpha_p, i, j, k) * kappa_z1_p * (2.0 + kappa_z2_p) * vector(theta_p, i, j, k)) + (-8.0) * vector(Alpha_p, i, j, k) * M_PI * rho_ADM_p + 0.5 * vector(Alpha_p, i, j, k) * (Rscalar_p + 0.6666666666666667 * (vector(trK_p, i, j, k) * vector(trK_p, i, j, k)) + 0.6666666666666667 * vector(theta_p, i, j, k) * ((-2.0 * vector(theta_p, i, j, k)) + vector(trK_p, i, j, k)) + (-vector(Atd_xx_p, i, j, k) * Atu_xx_p) + (-2.0 * vector(Atd_xy_p, i, j, k) * Atu_xy_p) + (-2.0 * vector(Atd_xz_p, i, j, k) * Atu_xz_p) + (-vector(Atd_yy_p, i, j, k) * Atu_yy_p) + (-2.0 * vector(Atd_yz_p, i, j, k) * Atu_yz_p) + (-vector(Atd_zz_p, i, j, k) * Atu_zz_p));
						RHS_Bfu_x = -vector(phif_p, i, j, k) * vector(Alpha_p, i, j, k) * chi_max_p * Gamt_x_p;
						RHS_Bfu_y = -vector(phif_p, i, j, k) * vector(Alpha_p, i, j, k) * chi_max_p * Gamt_y_p;
						RHS_Bfu_z = -vector(phif_p, i, j, k) * vector(Alpha_p, i, j, k) * chi_max_p * Gamt_z_p;
						RHS_tauf = (vector(Alpha_p, i, j, k) * (Sfuu_xx_p * vector(Atd_xx_p, i, j, k) + 2.0 * Sfuu_xy_p * vector(Atd_xy_p, i, j, k) + 2.0 * Sfuu_xz_p * vector(Atd_xz_p, i, j, k) + Sfuu_yy_p * vector(Atd_yy_p, i, j, k) + 2.0 * Sfuu_yz_p * vector(Atd_yz_p, i, j, k) + Sfuu_zz_p * vector(Atd_zz_p, i, j, k))) / chi_max_p + 0.3333333333333333 * vector(Alpha_p, i, j, k) * trSf_p * vector(trK_p, i, j, k);
						RHS_phif = (-vector(Alpha_p, i, j, k) * kappa_f_p * vector(phif_p, i, j, k)) + (-vector(Alpha_p, i, j, k) * vector(phif_p, i, j, k) * vector(trK_p, i, j, k));
						RHS_Sfd_x = -FVM_i(FluxpiSfd_x, FluxniSfd_x, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_x = RHS_Sfd_x - FVM_j(FluxpjSfd_x, FluxnjSfd_x, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_x = RHS_Sfd_x - FVM_k(FluxpkSfd_x, FluxnkSfd_x, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_y = -FVM_i(FluxpiSfd_y, FluxniSfd_y, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_y = RHS_Sfd_y - FVM_j(FluxpjSfd_y, FluxnjSfd_y, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_y = RHS_Sfd_y - FVM_k(FluxpkSfd_y, FluxnkSfd_y, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_z = -FVM_i(FluxpiSfd_z, FluxniSfd_z, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_z = RHS_Sfd_z - FVM_j(FluxpjSfd_z, FluxnjSfd_z, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_z = RHS_Sfd_z - FVM_k(FluxpkSfd_z, FluxnkSfd_z, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_x = RHS_Bfu_x - FVM_i(FluxpiBfu_x, FluxniBfu_x, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_x = RHS_Bfu_x - FVM_j(FluxpjBfu_x, FluxnjBfu_x, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_x = RHS_Bfu_x - FVM_k(FluxpkBfu_x, FluxnkBfu_x, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_y = RHS_Bfu_y - FVM_i(FluxpiBfu_y, FluxniBfu_y, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_y = RHS_Bfu_y - FVM_j(FluxpjBfu_y, FluxnjBfu_y, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_y = RHS_Bfu_y - FVM_k(FluxpkBfu_y, FluxnkBfu_y, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_z = RHS_Bfu_z - FVM_i(FluxpiBfu_z, FluxniBfu_z, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_z = RHS_Bfu_z - FVM_j(FluxpjBfu_z, FluxnjBfu_z, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_z = RHS_Bfu_z - FVM_k(FluxpkBfu_z, FluxnkBfu_z, dx, simPlat_dt, ilast, jlast);
						RHS_Df = -FVM_i(FluxpiDf, FluxniDf, dx, simPlat_dt, ilast, jlast);
						RHS_Df = RHS_Df - FVM_j(FluxpjDf, FluxnjDf, dx, simPlat_dt, ilast, jlast);
						RHS_Df = RHS_Df - FVM_k(FluxpkDf, FluxnkDf, dx, simPlat_dt, ilast, jlast);
						RHS_DYf = -FVM_i(FluxpiDYf, FluxniDYf, dx, simPlat_dt, ilast, jlast);
						RHS_DYf = RHS_DYf - FVM_j(FluxpjDYf, FluxnjDYf, dx, simPlat_dt, ilast, jlast);
						RHS_DYf = RHS_DYf - FVM_k(FluxpkDYf, FluxnkDYf, dx, simPlat_dt, ilast, jlast);
						RHS_tauf = RHS_tauf - FVM_i(Fluxpitauf, Fluxnitauf, dx, simPlat_dt, ilast, jlast);
						RHS_tauf = RHS_tauf - FVM_j(Fluxpjtauf, Fluxnjtauf, dx, simPlat_dt, ilast, jlast);
						RHS_tauf = RHS_tauf - FVM_k(Fluxpktauf, Fluxnktauf, dx, simPlat_dt, ilast, jlast);
						RHS_phif = RHS_phif - FVM_i(Fluxpiphif, Fluxniphif, dx, simPlat_dt, ilast, jlast);
						RHS_phif = RHS_phif - FVM_j(Fluxpjphif, Fluxnjphif, dx, simPlat_dt, ilast, jlast);
						RHS_phif = RHS_phif - FVM_k(Fluxpkphif, Fluxnkphif, dx, simPlat_dt, ilast, jlast);
						m_theta_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_theta_o2_t2_m0_l0;
						m_theta_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_theta_o2_t1_m0_l0;
						m_theta_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_theta_o2_t0_m0_l0;
						m_theta_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_theta_o1_t2_m0_l0;
						m_theta_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_theta_o1_t1_m0_l0;
						m_theta_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_theta_o1_t0_m0_l0;
						m_theta_o0_t2_l0 = (-Zu_z_p) * d_Psi1_xx_o0_t3_m0_l0;
						m_theta_o0_t1_l0 = (-Zu_y_p) * d_Psi1_xx_o0_t2_m0_l0;
						m_theta_o0_t0_l0 = (-Zu_x_p) * d_Psi1_xx_o0_t1_m0_l0;
						m_trK_o2_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_trK_o2_t2_m0_l0;
						m_trK_o2_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_trK_o2_t1_m0_l0;
						m_trK_o2_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_trK_o2_t0_m0_l0;
						m_trK_o1_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_trK_o1_t2_m0_l0;
						m_trK_o1_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_trK_o1_t1_m0_l0;
						m_trK_o1_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_trK_o1_t0_m0_l0;
						m_trK_o0_t20_l0 = 2.0 * Zu_z_p * d_Psi1_xx_o0_t3_m0_l0;
						m_trK_o0_t19_l0 = 2.0 * Zu_y_p * d_Psi1_xx_o0_t2_m0_l0;
						m_trK_o0_t18_l0 = 2.0 * Zu_x_p * d_Psi1_xx_o0_t1_m0_l0;
						m_trK_o0_t17_l0 = (-2.0 * chi_max_p * gtu_xz_p) * d_Psi1_xz_o0_t0_m0_l0;
						m_trK_o0_t16_l0 = (-2.0 * chi_max_p * gtu_xy_p) * d_Psi1_xy_o0_t0_m0_l0;
						m_trK_o0_t15_l0 = 0.5 * gtu_zz_p * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_trK_o0_t14_l0 = 0.5 * gtu_yz_p * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_trK_o0_t13_l0 = 0.5 * gtu_xz_p * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_trK_o0_t12_l0 = 0.5 * gtu_yz_p * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_trK_o0_t11_l0 = 0.5 * gtu_yy_p * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_trK_o0_t10_l0 = 0.5 * gtu_xy_p * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_trK_o0_t9_l0 = 0.5 * gtu_xz_p * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_trK_o0_t8_l0 = 0.5 * gtu_xy_p * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_trK_o0_t7_l0 = 0.5 * gtu_xx_p * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_trK_o0_t6_l0 = chi_max_p * Gamt_z_p * d_Psi1_xx_o0_t3_m0_l0;
						m_trK_o0_t5_l0 = chi_max_p * Gamt_y_p * d_Psi1_xx_o0_t2_m0_l0;
						m_trK_o0_t4_l0 = chi_max_p * Gamt_x_p * d_Psi1_xx_o0_t1_m0_l0;
						m_trK_o0_t3_l0 = (-chi_max_p * gtu_zz_p) * d_Psi1_zz_o0_t0_m0_l0;
						m_trK_o0_t2_l0 = (-2.0 * chi_max_p * gtu_yz_p) * d_Psi1_yz_o0_t0_m0_l0;
						m_trK_o0_t1_l0 = (-chi_max_p * gtu_yy_p) * d_Psi1_yy_o0_t0_m0_l0;
						m_trK_o0_t0_l0 = (-chi_max_p * gtu_xx_p) * d_Psi1_xx_o0_t0_m0_l0;
						m_chi_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_chi_o2_t2_m0_l0;
						m_chi_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_chi_o2_t1_m0_l0;
						m_chi_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_chi_o2_t0_m0_l0;
						m_chi_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_chi_o1_t2_m0_l0;
						m_chi_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_chi_o1_t1_m0_l0;
						m_chi_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_chi_o1_t0_m0_l0;
						m_Alpha_o2_t2_l0 = lambda_1 * MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Alpha_o2_t2_m0_l0;
						m_Alpha_o2_t1_l0 = lambda_1 * MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Alpha_o2_t1_m0_l0;
						m_Alpha_o2_t0_l0 = lambda_1 * MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Alpha_o2_t0_m0_l0;
						m_Alpha_o1_t2_l0 = lambda_1 * MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Alpha_o1_t2_m0_l0;
						m_Alpha_o1_t1_l0 = lambda_1 * MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Alpha_o1_t1_m0_l0;
						m_Alpha_o1_t0_l0 = lambda_1 * MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Alpha_o1_t0_m0_l0;
						m_Betau_z_o2_t2_l0 = lambda_2 * MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Betau_z_o2_t2_m0_l0;
						m_Betau_z_o2_t1_l0 = lambda_2 * MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Betau_z_o2_t1_m0_l0;
						m_Betau_z_o2_t0_l0 = lambda_2 * MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Betau_z_o2_t0_m0_l0;
						m_Betau_z_o1_t2_l0 = lambda_2 * MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Betau_z_o1_t2_m0_l0;
						m_Betau_z_o1_t1_l0 = lambda_2 * MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Betau_z_o1_t1_m0_l0;
						m_Betau_z_o1_t0_l0 = lambda_2 * MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Betau_z_o1_t0_m0_l0;
						m_Betau_y_o2_t2_l0 = lambda_2 * MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Betau_y_o2_t2_m0_l0;
						m_Betau_y_o2_t1_l0 = lambda_2 * MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Betau_y_o2_t1_m0_l0;
						m_Betau_y_o2_t0_l0 = lambda_2 * MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Betau_y_o2_t0_m0_l0;
						m_Betau_y_o1_t2_l0 = lambda_2 * MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Betau_y_o1_t2_m0_l0;
						m_Betau_y_o1_t1_l0 = lambda_2 * MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Betau_y_o1_t1_m0_l0;
						m_Betau_y_o1_t0_l0 = lambda_2 * MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Betau_y_o1_t0_m0_l0;
						m_Betau_x_o2_t2_l0 = lambda_2 * MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Betau_x_o2_t2_m0_l0;
						m_Betau_x_o2_t1_l0 = lambda_2 * MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Betau_x_o2_t1_m0_l0;
						m_Betau_x_o2_t0_l0 = lambda_2 * MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Betau_x_o2_t0_m0_l0;
						m_Betau_x_o1_t2_l0 = lambda_2 * MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Betau_x_o1_t2_m0_l0;
						m_Betau_x_o1_t1_l0 = lambda_2 * MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Betau_x_o1_t1_m0_l0;
						m_Betau_x_o1_t0_l0 = lambda_2 * MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Betau_x_o1_t0_m0_l0;
						m_Gamh_z_o2_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Gamh_z_o2_t2_m0_l0;
						m_Gamh_z_o2_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Gamh_z_o2_t1_m0_l0;
						m_Gamh_z_o2_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Gamh_z_o2_t0_m0_l0;
						m_Gamh_z_o1_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Gamh_z_o1_t2_m0_l0;
						m_Gamh_z_o1_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Gamh_z_o1_t1_m0_l0;
						m_Gamh_z_o1_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Gamh_z_o1_t0_m0_l0;
						m_Gamh_z_o0_t23_l0 = (-3.0 * vector(Alpha_p, i, j, k) * inv_chi_p * Atu_zz_p) * d_Rpd_xx_o0_t2_m0_l0;
						m_Gamh_z_o0_t22_l0 = (-3.0 * vector(Alpha_p, i, j, k) * inv_chi_p * Atu_yz_p) * d_Rpd_xx_o0_t1_m0_l0;
						m_Gamh_z_o0_t21_l0 = (-3.0 * vector(Alpha_p, i, j, k) * inv_chi_p * Atu_xz_p) * d_Rpd_xx_o0_t0_m0_l0;
						m_Gamh_z_o0_t20_l0 = (-2.0 * vector(theta_p, i, j, k) * gtu_zz_p) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_z_o0_t19_l0 = (-2.0 * vector(theta_p, i, j, k) * gtu_yz_p) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_z_o0_t18_l0 = (-2.0 * vector(theta_p, i, j, k) * gtu_xz_p) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_z_o0_t17_l0 = (-0.6666666666666667 * vector(Alpha_p, i, j, k) * gtu_zz_p) * d_Gamh_x_o0_t5_m0_l0;
						m_Gamh_z_o0_t16_l0 = (-0.6666666666666667 * vector(Alpha_p, i, j, k) * gtu_yz_p) * d_Gamh_x_o0_t4_m0_l0;
						m_Gamh_z_o0_t15_l0 = (-0.6666666666666667 * vector(Alpha_p, i, j, k) * gtu_xz_p) * d_Gamh_x_o0_t3_m0_l0;
						m_Gamh_z_o0_t14_l0 = (-1.333333333333333 * vector(Alpha_p, i, j, k) * gtu_zz_p) * d_Gamh_x_o0_t2_m0_l0;
						m_Gamh_z_o0_t13_l0 = (-1.333333333333333 * vector(Alpha_p, i, j, k) * gtu_yz_p) * d_Gamh_x_o0_t1_m0_l0;
						m_Gamh_z_o0_t12_l0 = (-1.333333333333333 * vector(Alpha_p, i, j, k) * gtu_xz_p) * d_Gamh_x_o0_t0_m0_l0;
						m_Gamh_z_o0_t11_l0 = (-vector(Gamh_z_p, i, j, k)) * d_div_Beta_o0_t2_m0_l0;
						m_Gamh_z_o0_t10_l0 = (-vector(Gamh_y_p, i, j, k)) * d_Sfd_y_o0_t9_m0_l0;
						m_Gamh_z_o0_t9_l0 = (-vector(Gamh_x_p, i, j, k)) * d_Sfd_x_o0_t9_m0_l0;
						m_Gamh_z_o0_t8_l0 = (-2.0 * Atu_zz_p) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_z_o0_t7_l0 = (-2.0 * Atu_yz_p) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_z_o0_t6_l0 = (-2.0 * Atu_xz_p) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_z_o0_t5_l0 = gtu_zz_p * d_d_div_Beta_z_o0_t2_m0_l0;
						m_Gamh_z_o0_t4_l0 = 2.0 * gtu_yz_p * d_d_div_Beta_y_o0_t2_m0_l0;
						m_Gamh_z_o0_t3_l0 = gtu_yy_p * d_Gamh_z_o0_t3_m0_l0;
						m_Gamh_z_o0_t2_l0 = 2.0 * gtu_xz_p * d_d_div_Beta_x_o0_t2_m0_l0;
						m_Gamh_z_o0_t1_l0 = 2.0 * gtu_xy_p * d_Gamh_z_o0_t1_m0_l0;
						m_Gamh_z_o0_t0_l0 = gtu_xx_p * d_Gamh_z_o0_t0_m0_l0;
						m_Gamh_y_o2_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Gamh_y_o2_t2_m0_l0;
						m_Gamh_y_o2_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Gamh_y_o2_t1_m0_l0;
						m_Gamh_y_o2_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Gamh_y_o2_t0_m0_l0;
						m_Gamh_y_o1_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Gamh_y_o1_t2_m0_l0;
						m_Gamh_y_o1_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Gamh_y_o1_t1_m0_l0;
						m_Gamh_y_o1_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Gamh_y_o1_t0_m0_l0;
						m_Gamh_y_o0_t23_l0 = (-3.0 * vector(Alpha_p, i, j, k) * inv_chi_p * Atu_yz_p) * d_Rpd_xx_o0_t2_m0_l0;
						m_Gamh_y_o0_t22_l0 = (-3.0 * vector(Alpha_p, i, j, k) * inv_chi_p * Atu_yy_p) * d_Rpd_xx_o0_t1_m0_l0;
						m_Gamh_y_o0_t21_l0 = (-3.0 * vector(Alpha_p, i, j, k) * inv_chi_p * Atu_xy_p) * d_Rpd_xx_o0_t0_m0_l0;
						m_Gamh_y_o0_t20_l0 = (-2.0 * vector(theta_p, i, j, k) * gtu_yz_p) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_y_o0_t19_l0 = (-2.0 * vector(theta_p, i, j, k) * gtu_yy_p) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_y_o0_t18_l0 = (-2.0 * vector(theta_p, i, j, k) * gtu_xy_p) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_y_o0_t17_l0 = (-0.6666666666666667 * vector(Alpha_p, i, j, k) * gtu_yz_p) * d_Gamh_x_o0_t5_m0_l0;
						m_Gamh_y_o0_t16_l0 = (-0.6666666666666667 * vector(Alpha_p, i, j, k) * gtu_yy_p) * d_Gamh_x_o0_t4_m0_l0;
						m_Gamh_y_o0_t15_l0 = (-0.6666666666666667 * vector(Alpha_p, i, j, k) * gtu_xy_p) * d_Gamh_x_o0_t3_m0_l0;
						m_Gamh_y_o0_t14_l0 = (-1.333333333333333 * vector(Alpha_p, i, j, k) * gtu_yz_p) * d_Gamh_x_o0_t2_m0_l0;
						m_Gamh_y_o0_t13_l0 = (-1.333333333333333 * vector(Alpha_p, i, j, k) * gtu_yy_p) * d_Gamh_x_o0_t1_m0_l0;
						m_Gamh_y_o0_t12_l0 = (-1.333333333333333 * vector(Alpha_p, i, j, k) * gtu_xy_p) * d_Gamh_x_o0_t0_m0_l0;
						m_Gamh_y_o0_t11_l0 = gtu_zz_p * d_Gamh_y_o0_t11_m0_l0;
						m_Gamh_y_o0_t10_l0 = 2.0 * gtu_yz_p * d_d_div_Beta_z_o0_t1_m0_l0;
						m_Gamh_y_o0_t9_l0 = gtu_yy_p * d_d_div_Beta_y_o0_t1_m0_l0;
						m_Gamh_y_o0_t8_l0 = 2.0 * gtu_xz_p * d_Gamh_y_o0_t8_m0_l0;
						m_Gamh_y_o0_t7_l0 = 2.0 * gtu_xy_p * d_d_div_Beta_x_o0_t1_m0_l0;
						m_Gamh_y_o0_t6_l0 = gtu_xx_p * d_Gamh_y_o0_t6_m0_l0;
						m_Gamh_y_o0_t5_l0 = (-vector(Gamh_z_p, i, j, k)) * d_Sfd_z_o0_t8_m0_l0;
						m_Gamh_y_o0_t4_l0 = (-vector(Gamh_y_p, i, j, k)) * d_div_Beta_o0_t1_m0_l0;
						m_Gamh_y_o0_t3_l0 = (-vector(Gamh_x_p, i, j, k)) * d_Sfd_x_o0_t8_m0_l0;
						m_Gamh_y_o0_t2_l0 = (-2.0 * Atu_yz_p) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_y_o0_t1_l0 = (-2.0 * Atu_yy_p) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_y_o0_t0_l0 = (-2.0 * Atu_xy_p) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_x_o2_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Gamh_x_o2_t2_m0_l0;
						m_Gamh_x_o2_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Gamh_x_o2_t1_m0_l0;
						m_Gamh_x_o2_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Gamh_x_o2_t0_m0_l0;
						m_Gamh_x_o1_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Gamh_x_o1_t2_m0_l0;
						m_Gamh_x_o1_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Gamh_x_o1_t1_m0_l0;
						m_Gamh_x_o1_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Gamh_x_o1_t0_m0_l0;
						m_Gamh_x_o0_t23_l0 = (-3.0 * vector(Alpha_p, i, j, k) * inv_chi_p * Atu_xz_p) * d_Rpd_xx_o0_t2_m0_l0;
						m_Gamh_x_o0_t22_l0 = (-3.0 * vector(Alpha_p, i, j, k) * inv_chi_p * Atu_xy_p) * d_Rpd_xx_o0_t1_m0_l0;
						m_Gamh_x_o0_t21_l0 = (-3.0 * vector(Alpha_p, i, j, k) * inv_chi_p * Atu_xx_p) * d_Rpd_xx_o0_t0_m0_l0;
						m_Gamh_x_o0_t20_l0 = (-2.0 * Atu_xz_p) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_x_o0_t19_l0 = (-2.0 * Atu_xy_p) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_x_o0_t18_l0 = (-2.0 * Atu_xx_p) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_x_o0_t17_l0 = gtu_zz_p * d_Gamh_x_o0_t17_m0_l0;
						m_Gamh_x_o0_t16_l0 = 2.0 * gtu_yz_p * d_Gamh_x_o0_t16_m0_l0;
						m_Gamh_x_o0_t15_l0 = gtu_yy_p * d_Gamh_x_o0_t15_m0_l0;
						m_Gamh_x_o0_t14_l0 = 2.0 * gtu_xz_p * d_d_div_Beta_z_o0_t0_m0_l0;
						m_Gamh_x_o0_t13_l0 = 2.0 * gtu_xy_p * d_d_div_Beta_y_o0_t0_m0_l0;
						m_Gamh_x_o0_t12_l0 = gtu_xx_p * d_d_div_Beta_x_o0_t0_m0_l0;
						m_Gamh_x_o0_t11_l0 = (-vector(Gamh_z_p, i, j, k)) * d_Sfd_z_o0_t7_m0_l0;
						m_Gamh_x_o0_t10_l0 = (-vector(Gamh_y_p, i, j, k)) * d_Sfd_y_o0_t7_m0_l0;
						m_Gamh_x_o0_t9_l0 = (-vector(Gamh_x_p, i, j, k)) * d_div_Beta_o0_t0_m0_l0;
						m_Gamh_x_o0_t8_l0 = (-2.0 * vector(theta_p, i, j, k) * gtu_xz_p) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_x_o0_t7_l0 = (-2.0 * vector(theta_p, i, j, k) * gtu_xy_p) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_x_o0_t6_l0 = (-2.0 * vector(theta_p, i, j, k) * gtu_xx_p) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_x_o0_t5_l0 = (-0.6666666666666667 * vector(Alpha_p, i, j, k) * gtu_xz_p) * d_Gamh_x_o0_t5_m0_l0;
						m_Gamh_x_o0_t4_l0 = (-0.6666666666666667 * vector(Alpha_p, i, j, k) * gtu_xy_p) * d_Gamh_x_o0_t4_m0_l0;
						m_Gamh_x_o0_t3_l0 = (-0.6666666666666667 * vector(Alpha_p, i, j, k) * gtu_xx_p) * d_Gamh_x_o0_t3_m0_l0;
						m_Gamh_x_o0_t2_l0 = (-1.333333333333333 * vector(Alpha_p, i, j, k) * gtu_xz_p) * d_Gamh_x_o0_t2_m0_l0;
						m_Gamh_x_o0_t1_l0 = (-1.333333333333333 * vector(Alpha_p, i, j, k) * gtu_xy_p) * d_Gamh_x_o0_t1_m0_l0;
						m_Gamh_x_o0_t0_l0 = (-1.333333333333333 * vector(Alpha_p, i, j, k) * gtu_xx_p) * d_Gamh_x_o0_t0_m0_l0;
						m_Atd_zz_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Atd_zz_o2_t2_m0_l0;
						m_Atd_zz_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Atd_zz_o2_t1_m0_l0;
						m_Atd_zz_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Atd_zz_o2_t0_m0_l0;
						m_Atd_zz_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Atd_zz_o1_t2_m0_l0;
						m_Atd_zz_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Atd_zz_o1_t1_m0_l0;
						m_Atd_zz_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Atd_zz_o1_t0_m0_l0;
						m_Atd_zz_o0_t2_l0 = 2.0 * vector(Atd_zz_p, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Atd_zz_o0_t1_l0 = 2.0 * vector(Atd_yz_p, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Atd_zz_o0_t0_l0 = 2.0 * vector(Atd_xz_p, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Atd_yz_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Atd_yz_o2_t2_m0_l0;
						m_Atd_yz_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Atd_yz_o2_t1_m0_l0;
						m_Atd_yz_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Atd_yz_o2_t0_m0_l0;
						m_Atd_yz_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Atd_yz_o1_t2_m0_l0;
						m_Atd_yz_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Atd_yz_o1_t1_m0_l0;
						m_Atd_yz_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Atd_yz_o1_t0_m0_l0;
						m_Atd_yz_o0_t5_l0 = vector(Atd_zz_p, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Atd_yz_o0_t4_l0 = vector(Atd_yz_p, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Atd_yz_o0_t3_l0 = vector(Atd_xz_p, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Atd_yz_o0_t2_l0 = vector(Atd_yz_p, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Atd_yz_o0_t1_l0 = vector(Atd_yy_p, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Atd_yz_o0_t0_l0 = vector(Atd_xy_p, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Atd_yy_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Atd_yy_o2_t2_m0_l0;
						m_Atd_yy_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Atd_yy_o2_t1_m0_l0;
						m_Atd_yy_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Atd_yy_o2_t0_m0_l0;
						m_Atd_yy_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Atd_yy_o1_t2_m0_l0;
						m_Atd_yy_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Atd_yy_o1_t1_m0_l0;
						m_Atd_yy_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Atd_yy_o1_t0_m0_l0;
						m_Atd_yy_o0_t2_l0 = 2.0 * vector(Atd_yz_p, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Atd_yy_o0_t1_l0 = 2.0 * vector(Atd_yy_p, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Atd_yy_o0_t0_l0 = 2.0 * vector(Atd_xy_p, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Atd_xz_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Atd_xz_o2_t2_m0_l0;
						m_Atd_xz_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Atd_xz_o2_t1_m0_l0;
						m_Atd_xz_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Atd_xz_o2_t0_m0_l0;
						m_Atd_xz_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Atd_xz_o1_t2_m0_l0;
						m_Atd_xz_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Atd_xz_o1_t1_m0_l0;
						m_Atd_xz_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Atd_xz_o1_t0_m0_l0;
						m_Atd_xz_o0_t5_l0 = vector(Atd_zz_p, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Atd_xz_o0_t4_l0 = vector(Atd_yz_p, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Atd_xz_o0_t3_l0 = vector(Atd_xz_p, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_Atd_xz_o0_t2_l0 = vector(Atd_xz_p, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Atd_xz_o0_t1_l0 = vector(Atd_xy_p, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Atd_xz_o0_t0_l0 = vector(Atd_xx_p, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Atd_xy_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Atd_xy_o2_t2_m0_l0;
						m_Atd_xy_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Atd_xy_o2_t1_m0_l0;
						m_Atd_xy_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Atd_xy_o2_t0_m0_l0;
						m_Atd_xy_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Atd_xy_o1_t2_m0_l0;
						m_Atd_xy_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Atd_xy_o1_t1_m0_l0;
						m_Atd_xy_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Atd_xy_o1_t0_m0_l0;
						m_Atd_xy_o0_t5_l0 = vector(Atd_yz_p, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Atd_xy_o0_t4_l0 = vector(Atd_yy_p, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Atd_xy_o0_t3_l0 = vector(Atd_xy_p, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_Atd_xy_o0_t2_l0 = vector(Atd_xz_p, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Atd_xy_o0_t1_l0 = vector(Atd_xy_p, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Atd_xy_o0_t0_l0 = vector(Atd_xx_p, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Atd_xx_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_Atd_xx_o2_t2_m0_l0;
						m_Atd_xx_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_Atd_xx_o2_t1_m0_l0;
						m_Atd_xx_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_Atd_xx_o2_t0_m0_l0;
						m_Atd_xx_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_Atd_xx_o1_t2_m0_l0;
						m_Atd_xx_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_Atd_xx_o1_t1_m0_l0;
						m_Atd_xx_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_Atd_xx_o1_t0_m0_l0;
						m_Atd_xx_o0_t2_l0 = 2.0 * vector(Atd_xz_p, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Atd_xx_o0_t1_l0 = 2.0 * vector(Atd_xy_p, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Atd_xx_o0_t0_l0 = 2.0 * vector(Atd_xx_p, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_gtd_zz_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_gtd_zz_o2_t2_m0_l0;
						m_gtd_zz_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_gtd_zz_o2_t1_m0_l0;
						m_gtd_zz_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_gtd_zz_o2_t0_m0_l0;
						m_gtd_zz_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_gtd_zz_o1_t2_m0_l0;
						m_gtd_zz_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_gtd_zz_o1_t1_m0_l0;
						m_gtd_zz_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_gtd_zz_o1_t0_m0_l0;
						m_gtd_zz_o0_t2_l0 = 2.0 * vector(gtd_zz_p, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_gtd_zz_o0_t1_l0 = 2.0 * vector(gtd_yz_p, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_gtd_zz_o0_t0_l0 = 2.0 * vector(gtd_xz_p, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_gtd_yz_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_gtd_yz_o2_t2_m0_l0;
						m_gtd_yz_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_gtd_yz_o2_t1_m0_l0;
						m_gtd_yz_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_gtd_yz_o2_t0_m0_l0;
						m_gtd_yz_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_gtd_yz_o1_t2_m0_l0;
						m_gtd_yz_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_gtd_yz_o1_t1_m0_l0;
						m_gtd_yz_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_gtd_yz_o1_t0_m0_l0;
						m_gtd_yz_o0_t5_l0 = vector(gtd_zz_p, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_gtd_yz_o0_t4_l0 = vector(gtd_yz_p, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_gtd_yz_o0_t3_l0 = vector(gtd_xz_p, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_gtd_yz_o0_t2_l0 = vector(gtd_yz_p, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_gtd_yz_o0_t1_l0 = vector(gtd_yy_p, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_gtd_yz_o0_t0_l0 = vector(gtd_xy_p, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_gtd_yy_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_gtd_yy_o2_t2_m0_l0;
						m_gtd_yy_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_gtd_yy_o2_t1_m0_l0;
						m_gtd_yy_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_gtd_yy_o2_t0_m0_l0;
						m_gtd_yy_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_gtd_yy_o1_t2_m0_l0;
						m_gtd_yy_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_gtd_yy_o1_t1_m0_l0;
						m_gtd_yy_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_gtd_yy_o1_t0_m0_l0;
						m_gtd_yy_o0_t2_l0 = 2.0 * vector(gtd_yz_p, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_gtd_yy_o0_t1_l0 = 2.0 * vector(gtd_yy_p, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_gtd_yy_o0_t0_l0 = 2.0 * vector(gtd_xy_p, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_gtd_xz_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_gtd_xz_o2_t2_m0_l0;
						m_gtd_xz_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_gtd_xz_o2_t1_m0_l0;
						m_gtd_xz_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_gtd_xz_o2_t0_m0_l0;
						m_gtd_xz_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_gtd_xz_o1_t2_m0_l0;
						m_gtd_xz_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_gtd_xz_o1_t1_m0_l0;
						m_gtd_xz_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_gtd_xz_o1_t0_m0_l0;
						m_gtd_xz_o0_t5_l0 = vector(gtd_zz_p, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_gtd_xz_o0_t4_l0 = vector(gtd_yz_p, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_gtd_xz_o0_t3_l0 = vector(gtd_xz_p, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_gtd_xz_o0_t2_l0 = vector(gtd_xz_p, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_gtd_xz_o0_t1_l0 = vector(gtd_xy_p, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_gtd_xz_o0_t0_l0 = vector(gtd_xx_p, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_gtd_xy_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_gtd_xy_o2_t2_m0_l0;
						m_gtd_xy_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_gtd_xy_o2_t1_m0_l0;
						m_gtd_xy_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_gtd_xy_o2_t0_m0_l0;
						m_gtd_xy_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_gtd_xy_o1_t2_m0_l0;
						m_gtd_xy_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_gtd_xy_o1_t1_m0_l0;
						m_gtd_xy_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_gtd_xy_o1_t0_m0_l0;
						m_gtd_xy_o0_t5_l0 = vector(gtd_yz_p, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_gtd_xy_o0_t4_l0 = vector(gtd_yy_p, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_gtd_xy_o0_t3_l0 = vector(gtd_xy_p, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_gtd_xy_o0_t2_l0 = vector(gtd_xz_p, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_gtd_xy_o0_t1_l0 = vector(gtd_xy_p, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_gtd_xy_o0_t0_l0 = vector(gtd_xx_p, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_gtd_xx_o2_t2_l0 = MIN(0.0, vector(Betau_z_p, i, j, k)) * d_gtd_xx_o2_t2_m0_l0;
						m_gtd_xx_o2_t1_l0 = MIN(0.0, vector(Betau_y_p, i, j, k)) * d_gtd_xx_o2_t1_m0_l0;
						m_gtd_xx_o2_t0_l0 = MIN(0.0, vector(Betau_x_p, i, j, k)) * d_gtd_xx_o2_t0_m0_l0;
						m_gtd_xx_o1_t2_l0 = MAX(0.0, vector(Betau_z_p, i, j, k)) * d_gtd_xx_o1_t2_m0_l0;
						m_gtd_xx_o1_t1_l0 = MAX(0.0, vector(Betau_y_p, i, j, k)) * d_gtd_xx_o1_t1_m0_l0;
						m_gtd_xx_o1_t0_l0 = MAX(0.0, vector(Betau_x_p, i, j, k)) * d_gtd_xx_o1_t0_m0_l0;
						m_gtd_xx_o0_t2_l0 = 2.0 * vector(gtd_xz_p, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_gtd_xx_o0_t1_l0 = 2.0 * vector(gtd_xy_p, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_gtd_xx_o0_t0_l0 = 2.0 * vector(gtd_xx_p, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_phif_o0_t2_l0 = (ch * ch) * vector(Bfu_z_p, i, j, k) * d_Psi1_xx_o0_t3_m0_l0;
						m_phif_o0_t1_l0 = (ch * ch) * vector(Bfu_y_p, i, j, k) * d_Psi1_xx_o0_t2_m0_l0;
						m_phif_o0_t0_l0 = (ch * ch) * vector(Bfu_x_p, i, j, k) * d_Psi1_xx_o0_t1_m0_l0;
						m_tauf_o0_t2_l0 = (-Sfu_z_p) * d_Psi1_xx_o0_t3_m0_l0;
						m_tauf_o0_t1_l0 = (-Sfu_y_p) * d_Psi1_xx_o0_t2_m0_l0;
						m_tauf_o0_t0_l0 = (-Sfu_x_p) * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_z_o0_t5_l0 = vector(phif_p, i, j, k) * gtu_zz_p * chi_max_p * d_Psi1_xx_o0_t3_m0_l0;
						m_Bfu_z_o0_t4_l0 = vector(phif_p, i, j, k) * gtu_yz_p * chi_max_p * d_Psi1_xx_o0_t2_m0_l0;
						m_Bfu_z_o0_t3_l0 = vector(phif_p, i, j, k) * gtu_xz_p * chi_max_p * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_z_o0_t2_l0 = (-0.5 * vector(phif_p, i, j, k) * gtu_zz_p * vector(Alpha_p, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Bfu_z_o0_t1_l0 = (-0.5 * vector(phif_p, i, j, k) * gtu_yz_p * vector(Alpha_p, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Bfu_z_o0_t0_l0 = (-0.5 * vector(phif_p, i, j, k) * gtu_xz_p * vector(Alpha_p, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Bfu_y_o0_t5_l0 = vector(phif_p, i, j, k) * gtu_yz_p * chi_max_p * d_Psi1_xx_o0_t3_m0_l0;
						m_Bfu_y_o0_t4_l0 = vector(phif_p, i, j, k) * gtu_yy_p * chi_max_p * d_Psi1_xx_o0_t2_m0_l0;
						m_Bfu_y_o0_t3_l0 = vector(phif_p, i, j, k) * gtu_xy_p * chi_max_p * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_y_o0_t2_l0 = (-0.5 * vector(phif_p, i, j, k) * gtu_yz_p * vector(Alpha_p, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Bfu_y_o0_t1_l0 = (-0.5 * vector(phif_p, i, j, k) * gtu_yy_p * vector(Alpha_p, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Bfu_y_o0_t0_l0 = (-0.5 * vector(phif_p, i, j, k) * gtu_xy_p * vector(Alpha_p, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Bfu_x_o0_t5_l0 = vector(phif_p, i, j, k) * gtu_xz_p * chi_max_p * d_Psi1_xx_o0_t3_m0_l0;
						m_Bfu_x_o0_t4_l0 = vector(phif_p, i, j, k) * gtu_xy_p * chi_max_p * d_Psi1_xx_o0_t2_m0_l0;
						m_Bfu_x_o0_t3_l0 = vector(phif_p, i, j, k) * gtu_xx_p * chi_max_p * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_x_o0_t2_l0 = (-0.5 * vector(phif_p, i, j, k) * gtu_xz_p * vector(Alpha_p, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Bfu_x_o0_t1_l0 = (-0.5 * vector(phif_p, i, j, k) * gtu_xy_p * vector(Alpha_p, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Bfu_x_o0_t0_l0 = (-0.5 * vector(phif_p, i, j, k) * gtu_xx_p * vector(Alpha_p, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Sfd_z_o0_t10_l0 = (-(vector(Df_p, i, j, k) + vector(tauf_p, i, j, k))) * d_Psi1_xx_o0_t3_m0_l0;
						m_Sfd_z_o0_t9_l0 = vector(Sfd_z_p, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Sfd_z_o0_t8_l0 = vector(Sfd_y_p, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Sfd_z_o0_t7_l0 = vector(Sfd_x_p, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Sfd_z_o0_t6_l0 = (-0.5 * (vector(Alpha_p, i, j, k) * trSf_p) / chi_max_p) * d_Rpd_xx_o0_t2_m0_l0;
						m_Sfd_z_o0_t5_l0 = 0.5 * (vector(Alpha_p, i, j, k) * Sfuu_zz_p) / chi_max_p * d_Ctd_zzz_o0_t0_m0_l0;
						m_Sfd_z_o0_t4_l0 = (vector(Alpha_p, i, j, k) * Sfuu_yz_p) / chi_max_p * d_Ctd_yzz_o0_t0_m0_l0;
						m_Sfd_z_o0_t3_l0 = 0.5 * (vector(Alpha_p, i, j, k) * Sfuu_yy_p) / chi_max_p * d_Ctd_yyz_o0_t0_m0_l0;
						m_Sfd_z_o0_t2_l0 = (vector(Alpha_p, i, j, k) * Sfuu_xz_p) / chi_max_p * d_Ctd_xzz_o0_t0_m0_l0;
						m_Sfd_z_o0_t1_l0 = (vector(Alpha_p, i, j, k) * Sfuu_xy_p) / chi_max_p * d_Ctd_xyz_o0_t1_m0_l0;
						m_Sfd_z_o0_t0_l0 = 0.5 * (vector(Alpha_p, i, j, k) * Sfuu_xx_p) / chi_max_p * d_Ctd_xxz_o0_t0_m0_l0;
						m_Sfd_y_o0_t10_l0 = (-(vector(Df_p, i, j, k) + vector(tauf_p, i, j, k))) * d_Psi1_xx_o0_t2_m0_l0;
						m_Sfd_y_o0_t9_l0 = vector(Sfd_z_p, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Sfd_y_o0_t8_l0 = vector(Sfd_y_p, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Sfd_y_o0_t7_l0 = vector(Sfd_x_p, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Sfd_y_o0_t6_l0 = (-0.5 * (vector(Alpha_p, i, j, k) * trSf_p) / chi_max_p) * d_Rpd_xx_o0_t1_m0_l0;
						m_Sfd_y_o0_t5_l0 = 0.5 * (vector(Alpha_p, i, j, k) * Sfuu_zz_p) / chi_max_p * d_Ctd_yzz_o0_t1_m0_l0;
						m_Sfd_y_o0_t4_l0 = (vector(Alpha_p, i, j, k) * Sfuu_yz_p) / chi_max_p * d_Ctd_zyy_o0_t0_m0_l0;
						m_Sfd_y_o0_t3_l0 = 0.5 * (vector(Alpha_p, i, j, k) * Sfuu_yy_p) / chi_max_p * d_Ctd_yyy_o0_t0_m0_l0;
						m_Sfd_y_o0_t2_l0 = (vector(Alpha_p, i, j, k) * Sfuu_xz_p) / chi_max_p * d_Ctd_xyz_o0_t0_m0_l0;
						m_Sfd_y_o0_t1_l0 = (vector(Alpha_p, i, j, k) * Sfuu_xy_p) / chi_max_p * d_Ctd_xyy_o0_t0_m0_l0;
						m_Sfd_y_o0_t0_l0 = 0.5 * (vector(Alpha_p, i, j, k) * Sfuu_xx_p) / chi_max_p * d_Ctd_xxy_o0_t0_m0_l0;
						m_Sfd_x_o0_t10_l0 = (-(vector(Df_p, i, j, k) + vector(tauf_p, i, j, k))) * d_Psi1_xx_o0_t1_m0_l0;
						m_Sfd_x_o0_t9_l0 = vector(Sfd_z_p, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Sfd_x_o0_t8_l0 = vector(Sfd_y_p, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Sfd_x_o0_t7_l0 = vector(Sfd_x_p, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_Sfd_x_o0_t6_l0 = (-0.5 * (vector(Alpha_p, i, j, k) * trSf_p) / chi_max_p) * d_Rpd_xx_o0_t0_m0_l0;
						m_Sfd_x_o0_t5_l0 = 0.5 * (vector(Alpha_p, i, j, k) * Sfuu_zz_p) / chi_max_p * d_Ctd_xzz_o0_t1_m0_l0;
						m_Sfd_x_o0_t4_l0 = (vector(Alpha_p, i, j, k) * Sfuu_yz_p) / chi_max_p * d_Ctd_xyz_o0_t2_m0_l0;
						m_Sfd_x_o0_t3_l0 = 0.5 * (vector(Alpha_p, i, j, k) * Sfuu_yy_p) / chi_max_p * d_Ctd_xyy_o0_t1_m0_l0;
						m_Sfd_x_o0_t2_l0 = (vector(Alpha_p, i, j, k) * Sfuu_xz_p) / chi_max_p * d_Ctd_zxx_o0_t0_m0_l0;
						m_Sfd_x_o0_t1_l0 = (vector(Alpha_p, i, j, k) * Sfuu_xy_p) / chi_max_p * d_Ctd_yxx_o0_t0_m0_l0;
						m_Sfd_x_o0_t0_l0 = 0.5 * (vector(Alpha_p, i, j, k) * Sfuu_xx_p) / chi_max_p * d_Ctd_xxx_o0_t0_m0_l0;
						RHS_Sfd_x = RHS_Sfd_x + (((((((((((m_Sfd_x_o0_t0_l0 + m_Sfd_x_o0_t1_l0) + m_Sfd_x_o0_t2_l0) + m_Sfd_x_o0_t3_l0) + m_Sfd_x_o0_t4_l0) + m_Sfd_x_o0_t5_l0) + m_Sfd_x_o0_t6_l0) + m_Sfd_x_o0_t7_l0) + m_Sfd_x_o0_t8_l0) + m_Sfd_x_o0_t9_l0) + m_Sfd_x_o0_t10_l0) + d_Sfd_x_o0_t11_m0_l0);
						RHS_Sfd_y = RHS_Sfd_y + (((((((((((m_Sfd_y_o0_t0_l0 + m_Sfd_y_o0_t1_l0) + m_Sfd_y_o0_t2_l0) + m_Sfd_y_o0_t3_l0) + m_Sfd_y_o0_t4_l0) + m_Sfd_y_o0_t5_l0) + m_Sfd_y_o0_t6_l0) + m_Sfd_y_o0_t7_l0) + m_Sfd_y_o0_t8_l0) + m_Sfd_y_o0_t9_l0) + m_Sfd_y_o0_t10_l0) + d_Sfd_y_o0_t11_m0_l0);
						RHS_Sfd_z = RHS_Sfd_z + (((((((((((m_Sfd_z_o0_t0_l0 + m_Sfd_z_o0_t1_l0) + m_Sfd_z_o0_t2_l0) + m_Sfd_z_o0_t3_l0) + m_Sfd_z_o0_t4_l0) + m_Sfd_z_o0_t5_l0) + m_Sfd_z_o0_t6_l0) + m_Sfd_z_o0_t7_l0) + m_Sfd_z_o0_t8_l0) + m_Sfd_z_o0_t9_l0) + m_Sfd_z_o0_t10_l0) + d_Sfd_z_o0_t11_m0_l0);
						RHS_Bfu_x = RHS_Bfu_x + (((((m_Bfu_x_o0_t0_l0 + m_Bfu_x_o0_t1_l0) + m_Bfu_x_o0_t2_l0) + m_Bfu_x_o0_t3_l0) + m_Bfu_x_o0_t4_l0) + m_Bfu_x_o0_t5_l0);
						RHS_Bfu_y = RHS_Bfu_y + (((((m_Bfu_y_o0_t0_l0 + m_Bfu_y_o0_t1_l0) + m_Bfu_y_o0_t2_l0) + m_Bfu_y_o0_t3_l0) + m_Bfu_y_o0_t4_l0) + m_Bfu_y_o0_t5_l0);
						RHS_Bfu_z = RHS_Bfu_z + (((((m_Bfu_z_o0_t0_l0 + m_Bfu_z_o0_t1_l0) + m_Bfu_z_o0_t2_l0) + m_Bfu_z_o0_t3_l0) + m_Bfu_z_o0_t4_l0) + m_Bfu_z_o0_t5_l0);
						RHS_DYf = RHS_DYf + d_DYf_o0_t0_m0_l0;
						RHS_tauf = RHS_tauf + (((m_tauf_o0_t0_l0 + m_tauf_o0_t1_l0) + m_tauf_o0_t2_l0) + d_tauf_o0_t3_m0_l0);
						RHS_phif = RHS_phif + ((m_phif_o0_t0_l0 + m_phif_o0_t1_l0) + m_phif_o0_t2_l0);
						RHS_gtd_xx = ((((((((m_gtd_xx_o0_t0_l0 + m_gtd_xx_o0_t1_l0) + m_gtd_xx_o0_t2_l0) + d_gtd_xx_o0_t3_m0_l0) + m_gtd_xx_o1_t0_l0) + m_gtd_xx_o1_t1_l0) + m_gtd_xx_o1_t2_l0) + m_gtd_xx_o2_t0_l0) + m_gtd_xx_o2_t1_l0) + m_gtd_xx_o2_t2_l0;
						RHS_gtd_xy = (((((((((((m_gtd_xy_o0_t0_l0 + m_gtd_xy_o0_t1_l0) + m_gtd_xy_o0_t2_l0) + m_gtd_xy_o0_t3_l0) + m_gtd_xy_o0_t4_l0) + m_gtd_xy_o0_t5_l0) + d_gtd_xy_o0_t6_m0_l0) + m_gtd_xy_o1_t0_l0) + m_gtd_xy_o1_t1_l0) + m_gtd_xy_o1_t2_l0) + m_gtd_xy_o2_t0_l0) + m_gtd_xy_o2_t1_l0) + m_gtd_xy_o2_t2_l0;
						RHS_gtd_xz = (((((((((((m_gtd_xz_o0_t0_l0 + m_gtd_xz_o0_t1_l0) + m_gtd_xz_o0_t2_l0) + m_gtd_xz_o0_t3_l0) + m_gtd_xz_o0_t4_l0) + m_gtd_xz_o0_t5_l0) + d_gtd_xz_o0_t6_m0_l0) + m_gtd_xz_o1_t0_l0) + m_gtd_xz_o1_t1_l0) + m_gtd_xz_o1_t2_l0) + m_gtd_xz_o2_t0_l0) + m_gtd_xz_o2_t1_l0) + m_gtd_xz_o2_t2_l0;
						RHS_gtd_yy = ((((((((m_gtd_yy_o0_t0_l0 + m_gtd_yy_o0_t1_l0) + m_gtd_yy_o0_t2_l0) + d_gtd_yy_o0_t3_m0_l0) + m_gtd_yy_o1_t0_l0) + m_gtd_yy_o1_t1_l0) + m_gtd_yy_o1_t2_l0) + m_gtd_yy_o2_t0_l0) + m_gtd_yy_o2_t1_l0) + m_gtd_yy_o2_t2_l0;
						RHS_gtd_yz = (((((((((((m_gtd_yz_o0_t0_l0 + m_gtd_yz_o0_t1_l0) + m_gtd_yz_o0_t2_l0) + m_gtd_yz_o0_t3_l0) + m_gtd_yz_o0_t4_l0) + m_gtd_yz_o0_t5_l0) + d_gtd_yz_o0_t6_m0_l0) + m_gtd_yz_o1_t0_l0) + m_gtd_yz_o1_t1_l0) + m_gtd_yz_o1_t2_l0) + m_gtd_yz_o2_t0_l0) + m_gtd_yz_o2_t1_l0) + m_gtd_yz_o2_t2_l0;
						RHS_gtd_zz = ((((((((m_gtd_zz_o0_t0_l0 + m_gtd_zz_o0_t1_l0) + m_gtd_zz_o0_t2_l0) + d_gtd_zz_o0_t3_m0_l0) + m_gtd_zz_o1_t0_l0) + m_gtd_zz_o1_t1_l0) + m_gtd_zz_o1_t2_l0) + m_gtd_zz_o2_t0_l0) + m_gtd_zz_o2_t1_l0) + m_gtd_zz_o2_t2_l0;
						RHS_Atd_xx = ((((((((m_Atd_xx_o0_t0_l0 + m_Atd_xx_o0_t1_l0) + m_Atd_xx_o0_t2_l0) + d_Atd_xx_o0_t3_m0_l0) + m_Atd_xx_o1_t0_l0) + m_Atd_xx_o1_t1_l0) + m_Atd_xx_o1_t2_l0) + m_Atd_xx_o2_t0_l0) + m_Atd_xx_o2_t1_l0) + m_Atd_xx_o2_t2_l0;
						RHS_Atd_xy = (((((((((((m_Atd_xy_o0_t0_l0 + m_Atd_xy_o0_t1_l0) + m_Atd_xy_o0_t2_l0) + m_Atd_xy_o0_t3_l0) + m_Atd_xy_o0_t4_l0) + m_Atd_xy_o0_t5_l0) + d_Atd_xy_o0_t6_m0_l0) + m_Atd_xy_o1_t0_l0) + m_Atd_xy_o1_t1_l0) + m_Atd_xy_o1_t2_l0) + m_Atd_xy_o2_t0_l0) + m_Atd_xy_o2_t1_l0) + m_Atd_xy_o2_t2_l0;
						RHS_Atd_xz = (((((((((((m_Atd_xz_o0_t0_l0 + m_Atd_xz_o0_t1_l0) + m_Atd_xz_o0_t2_l0) + m_Atd_xz_o0_t3_l0) + m_Atd_xz_o0_t4_l0) + m_Atd_xz_o0_t5_l0) + d_Atd_xz_o0_t6_m0_l0) + m_Atd_xz_o1_t0_l0) + m_Atd_xz_o1_t1_l0) + m_Atd_xz_o1_t2_l0) + m_Atd_xz_o2_t0_l0) + m_Atd_xz_o2_t1_l0) + m_Atd_xz_o2_t2_l0;
						RHS_Atd_yy = ((((((((m_Atd_yy_o0_t0_l0 + m_Atd_yy_o0_t1_l0) + m_Atd_yy_o0_t2_l0) + d_Atd_yy_o0_t3_m0_l0) + m_Atd_yy_o1_t0_l0) + m_Atd_yy_o1_t1_l0) + m_Atd_yy_o1_t2_l0) + m_Atd_yy_o2_t0_l0) + m_Atd_yy_o2_t1_l0) + m_Atd_yy_o2_t2_l0;
						RHS_Atd_yz = (((((((((((m_Atd_yz_o0_t0_l0 + m_Atd_yz_o0_t1_l0) + m_Atd_yz_o0_t2_l0) + m_Atd_yz_o0_t3_l0) + m_Atd_yz_o0_t4_l0) + m_Atd_yz_o0_t5_l0) + d_Atd_yz_o0_t6_m0_l0) + m_Atd_yz_o1_t0_l0) + m_Atd_yz_o1_t1_l0) + m_Atd_yz_o1_t2_l0) + m_Atd_yz_o2_t0_l0) + m_Atd_yz_o2_t1_l0) + m_Atd_yz_o2_t2_l0;
						RHS_Atd_zz = ((((((((m_Atd_zz_o0_t0_l0 + m_Atd_zz_o0_t1_l0) + m_Atd_zz_o0_t2_l0) + d_Atd_zz_o0_t3_m0_l0) + m_Atd_zz_o1_t0_l0) + m_Atd_zz_o1_t1_l0) + m_Atd_zz_o1_t2_l0) + m_Atd_zz_o2_t0_l0) + m_Atd_zz_o2_t1_l0) + m_Atd_zz_o2_t2_l0;
						RHS_Gamh_x = (((((((((((((((((((((((((((((m_Gamh_x_o0_t0_l0 + m_Gamh_x_o0_t1_l0) + m_Gamh_x_o0_t2_l0) + m_Gamh_x_o0_t3_l0) + m_Gamh_x_o0_t4_l0) + m_Gamh_x_o0_t5_l0) + m_Gamh_x_o0_t6_l0) + m_Gamh_x_o0_t7_l0) + m_Gamh_x_o0_t8_l0) + m_Gamh_x_o0_t9_l0) + m_Gamh_x_o0_t10_l0) + m_Gamh_x_o0_t11_l0) + m_Gamh_x_o0_t12_l0) + m_Gamh_x_o0_t13_l0) + m_Gamh_x_o0_t14_l0) + m_Gamh_x_o0_t15_l0) + m_Gamh_x_o0_t16_l0) + m_Gamh_x_o0_t17_l0) + m_Gamh_x_o0_t18_l0) + m_Gamh_x_o0_t19_l0) + m_Gamh_x_o0_t20_l0) + m_Gamh_x_o0_t21_l0) + m_Gamh_x_o0_t22_l0) + m_Gamh_x_o0_t23_l0) + d_Gamh_x_o0_t24_m0_l0) + m_Gamh_x_o1_t0_l0) + m_Gamh_x_o1_t1_l0) + m_Gamh_x_o1_t2_l0) + m_Gamh_x_o2_t0_l0) + m_Gamh_x_o2_t1_l0) + m_Gamh_x_o2_t2_l0;
						RHS_Gamh_y = (((((((((((((((((((((((((((((m_Gamh_y_o0_t0_l0 + m_Gamh_y_o0_t1_l0) + m_Gamh_y_o0_t2_l0) + m_Gamh_y_o0_t3_l0) + m_Gamh_y_o0_t4_l0) + m_Gamh_y_o0_t5_l0) + m_Gamh_y_o0_t6_l0) + m_Gamh_y_o0_t7_l0) + m_Gamh_y_o0_t8_l0) + m_Gamh_y_o0_t9_l0) + m_Gamh_y_o0_t10_l0) + m_Gamh_y_o0_t11_l0) + m_Gamh_y_o0_t12_l0) + m_Gamh_y_o0_t13_l0) + m_Gamh_y_o0_t14_l0) + m_Gamh_y_o0_t15_l0) + m_Gamh_y_o0_t16_l0) + m_Gamh_y_o0_t17_l0) + m_Gamh_y_o0_t18_l0) + m_Gamh_y_o0_t19_l0) + m_Gamh_y_o0_t20_l0) + m_Gamh_y_o0_t21_l0) + m_Gamh_y_o0_t22_l0) + m_Gamh_y_o0_t23_l0) + d_Gamh_y_o0_t24_m0_l0) + m_Gamh_y_o1_t0_l0) + m_Gamh_y_o1_t1_l0) + m_Gamh_y_o1_t2_l0) + m_Gamh_y_o2_t0_l0) + m_Gamh_y_o2_t1_l0) + m_Gamh_y_o2_t2_l0;
						RHS_Gamh_z = (((((((((((((((((((((((((((((m_Gamh_z_o0_t0_l0 + m_Gamh_z_o0_t1_l0) + m_Gamh_z_o0_t2_l0) + m_Gamh_z_o0_t3_l0) + m_Gamh_z_o0_t4_l0) + m_Gamh_z_o0_t5_l0) + m_Gamh_z_o0_t6_l0) + m_Gamh_z_o0_t7_l0) + m_Gamh_z_o0_t8_l0) + m_Gamh_z_o0_t9_l0) + m_Gamh_z_o0_t10_l0) + m_Gamh_z_o0_t11_l0) + m_Gamh_z_o0_t12_l0) + m_Gamh_z_o0_t13_l0) + m_Gamh_z_o0_t14_l0) + m_Gamh_z_o0_t15_l0) + m_Gamh_z_o0_t16_l0) + m_Gamh_z_o0_t17_l0) + m_Gamh_z_o0_t18_l0) + m_Gamh_z_o0_t19_l0) + m_Gamh_z_o0_t20_l0) + m_Gamh_z_o0_t21_l0) + m_Gamh_z_o0_t22_l0) + m_Gamh_z_o0_t23_l0) + d_Gamh_z_o0_t24_m0_l0) + m_Gamh_z_o1_t0_l0) + m_Gamh_z_o1_t1_l0) + m_Gamh_z_o1_t2_l0) + m_Gamh_z_o2_t0_l0) + m_Gamh_z_o2_t1_l0) + m_Gamh_z_o2_t2_l0;
						RHS_Betau_x = (((((d_Betau_x_o0_t0_m0_l0 + m_Betau_x_o1_t0_l0) + m_Betau_x_o1_t1_l0) + m_Betau_x_o1_t2_l0) + m_Betau_x_o2_t0_l0) + m_Betau_x_o2_t1_l0) + m_Betau_x_o2_t2_l0;
						RHS_Betau_y = (((((d_Betau_y_o0_t0_m0_l0 + m_Betau_y_o1_t0_l0) + m_Betau_y_o1_t1_l0) + m_Betau_y_o1_t2_l0) + m_Betau_y_o2_t0_l0) + m_Betau_y_o2_t1_l0) + m_Betau_y_o2_t2_l0;
						RHS_Betau_z = (((((d_Betau_z_o0_t0_m0_l0 + m_Betau_z_o1_t0_l0) + m_Betau_z_o1_t1_l0) + m_Betau_z_o1_t2_l0) + m_Betau_z_o2_t0_l0) + m_Betau_z_o2_t1_l0) + m_Betau_z_o2_t2_l0;
						RHS_Alpha = (((((d_Alpha_o0_t0_m0_l0 + m_Alpha_o1_t0_l0) + m_Alpha_o1_t1_l0) + m_Alpha_o1_t2_l0) + m_Alpha_o2_t0_l0) + m_Alpha_o2_t1_l0) + m_Alpha_o2_t2_l0;
						RHS_chi = (((((d_chi_o0_t0_m0_l0 + m_chi_o1_t0_l0) + m_chi_o1_t1_l0) + m_chi_o1_t2_l0) + m_chi_o2_t0_l0) + m_chi_o2_t1_l0) + m_chi_o2_t2_l0;
						RHS_trK = ((((((((((((((((((((((((((m_trK_o0_t0_l0 + m_trK_o0_t1_l0) + m_trK_o0_t2_l0) + m_trK_o0_t3_l0) + m_trK_o0_t4_l0) + m_trK_o0_t5_l0) + m_trK_o0_t6_l0) + m_trK_o0_t7_l0) + m_trK_o0_t8_l0) + m_trK_o0_t9_l0) + m_trK_o0_t10_l0) + m_trK_o0_t11_l0) + m_trK_o0_t12_l0) + m_trK_o0_t13_l0) + m_trK_o0_t14_l0) + m_trK_o0_t15_l0) + m_trK_o0_t16_l0) + m_trK_o0_t17_l0) + m_trK_o0_t18_l0) + m_trK_o0_t19_l0) + m_trK_o0_t20_l0) + d_trK_o0_t21_m0_l0) + m_trK_o1_t0_l0) + m_trK_o1_t1_l0) + m_trK_o1_t2_l0) + m_trK_o2_t0_l0) + m_trK_o2_t1_l0) + m_trK_o2_t2_l0;
						RHS_theta = ((((((((m_theta_o0_t0_l0 + m_theta_o0_t1_l0) + m_theta_o0_t2_l0) + d_theta_o0_t3_m0_l0) + m_theta_o1_t0_l0) + m_theta_o1_t1_l0) + m_theta_o1_t2_l0) + m_theta_o2_t0_l0) + m_theta_o2_t1_l0) + m_theta_o2_t2_l0;
						n_x = 0.0;
						n_y = 0.0;
						n_z = 0.0;
						interaction_index = 0.0;
						if ((((((vector(FOV_xLower, i + 1, j, k) > 0.0) || (vector(FOV_xUpper, i + 1, j, k) > 0.0)) || (vector(FOV_yLower, i + 1, j, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k) > 0.0)) || (vector(FOV_zLower, i + 1, j, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x + 1.0 / dx[0];
						}
						if ((((((vector(FOV_xLower, i - 1, j, k) > 0.0) || (vector(FOV_xUpper, i - 1, j, k) > 0.0)) || (vector(FOV_yLower, i - 1, j, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k) > 0.0)) || (vector(FOV_zLower, i - 1, j, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x - 1.0 / dx[0];
						}
						if ((((((vector(FOV_xLower, i, j + 1, k) > 0.0) || (vector(FOV_xUpper, i, j + 1, k) > 0.0)) || (vector(FOV_yLower, i, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k) > 0.0)) || (vector(FOV_zLower, i, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y + 1.0 / dx[1];
						}
						if ((((((vector(FOV_xLower, i, j - 1, k) > 0.0) || (vector(FOV_xUpper, i, j - 1, k) > 0.0)) || (vector(FOV_yLower, i, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k) > 0.0)) || (vector(FOV_zLower, i, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y - 1.0 / dx[1];
						}
						if ((((((vector(FOV_xLower, i, j, k + 1) > 0.0) || (vector(FOV_xUpper, i, j, k + 1) > 0.0)) || (vector(FOV_yLower, i, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j, k + 1) > 0.0)) || (vector(FOV_zLower, i, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j, k + 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z + 1.0 / dx[2];
						}
						if ((((((vector(FOV_xLower, i, j, k - 1) > 0.0) || (vector(FOV_xUpper, i, j, k - 1) > 0.0)) || (vector(FOV_yLower, i, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j, k - 1) > 0.0)) || (vector(FOV_zLower, i, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j, k - 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z - 1.0 / dx[2];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k) > 0.0)) || (vector(FOV_yLower, i + 1, j, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k) > 0.0)) || (vector(FOV_zLower, i + 1, j, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x + 1.0 / dx[0];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i - 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k) > 0.0)) || (vector(FOV_yLower, i - 1, j, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k) > 0.0)) || (vector(FOV_zLower, i - 1, j, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x - 1.0 / dx[0];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k) > 0.0)) || (vector(FOV_yLower, i, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k) > 0.0)) || (vector(FOV_zLower, i, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y + 1.0 / dx[1];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j - 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k) > 0.0)) || (vector(FOV_yLower, i, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k) > 0.0)) || (vector(FOV_zLower, i, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y - 1.0 / dx[1];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j, k + 1) > 0.0)) || (vector(FOV_yLower, i, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j, k + 1) > 0.0)) || (vector(FOV_zLower, i, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j, k + 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z + 1.0 / dx[2];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k - 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j, k - 1) > 0.0)) || (vector(FOV_yLower, i, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j, k - 1) > 0.0)) || (vector(FOV_zLower, i, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j, k - 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z - 1.0 / dx[2];
						}
						if ((!equalsEq(n_x, 0.0) || !equalsEq(n_y, 0.0)) || !equalsEq(n_z, 0.0)) {
							mod_normal = sqrt((n_x * n_x + n_y * n_y) + n_z * n_z);
							n_x = n_x / mod_normal;
							n_y = n_y / mod_normal;
							n_z = n_z / mod_normal;
						}
						if (equalsEq(interaction_index, 1.0)) {
							i_d_Atd_xx_o0_t0_m0_l0 = D1CDO4_i(Atd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xx_o0_t1_m0_l0 = D1CDO4_j(Atd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xx_o0_t2_m0_l0 = D1CDO4_k(Atd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xy_o0_t0_m0_l0 = D1CDO4_i(Atd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xy_o0_t1_m0_l0 = D1CDO4_j(Atd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xy_o0_t2_m0_l0 = D1CDO4_k(Atd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xz_o0_t0_m0_l0 = D1CDO4_i(Atd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xz_o0_t1_m0_l0 = D1CDO4_j(Atd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xz_o0_t2_m0_l0 = D1CDO4_k(Atd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yy_o0_t0_m0_l0 = D1CDO4_i(Atd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yy_o0_t1_m0_l0 = D1CDO4_j(Atd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yy_o0_t2_m0_l0 = D1CDO4_k(Atd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yz_o0_t0_m0_l0 = D1CDO4_i(Atd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yz_o0_t1_m0_l0 = D1CDO4_j(Atd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yz_o0_t2_m0_l0 = D1CDO4_k(Atd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_zz_o0_t0_m0_l0 = D1CDO4_i(Atd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_zz_o0_t1_m0_l0 = D1CDO4_j(Atd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_zz_o0_t2_m0_l0 = D1CDO4_k(Atd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_gtd_xx_o0_t3_m0_l0 = -gtd_xx_falloff * (vector(gtd_xx_p, i, j, k) - gtd_xx_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_xy_o0_t3_m0_l0 = -gtd_xy_falloff * (vector(gtd_xy_p, i, j, k) - gtd_xy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_xz_o0_t3_m0_l0 = -gtd_xz_falloff * (vector(gtd_xz_p, i, j, k) - gtd_xz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_yy_o0_t3_m0_l0 = -gtd_yy_falloff * (vector(gtd_yy_p, i, j, k) - gtd_yy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_yz_o0_t3_m0_l0 = -gtd_yz_falloff * (vector(gtd_yz_p, i, j, k) - gtd_yz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_zz_o0_t3_m0_l0 = -gtd_zz_falloff * (vector(gtd_zz_p, i, j, k) - gtd_zz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_xx_o0_t3_m0_l0 = -Atd_xx_falloff * (vector(Atd_xx_p, i, j, k) - Atd_xx_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_xy_o0_t3_m0_l0 = -Atd_xy_falloff * (vector(Atd_xy_p, i, j, k) - Atd_xy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_xz_o0_t3_m0_l0 = -Atd_xz_falloff * (vector(Atd_xz_p, i, j, k) - Atd_xz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_yy_o0_t3_m0_l0 = -Atd_yy_falloff * (vector(Atd_yy_p, i, j, k) - Atd_yy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_yz_o0_t3_m0_l0 = -Atd_yz_falloff * (vector(Atd_yz_p, i, j, k) - Atd_yz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_zz_o0_t3_m0_l0 = -Atd_zz_falloff * (vector(Atd_zz_p, i, j, k) - Atd_zz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Gamh_x_o0_t3_m0_l0 = -Gamh_x_falloff * (vector(Gamh_x_p, i, j, k) - Gamh_x_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Gamh_y_o0_t3_m0_l0 = -Gamh_y_falloff * (vector(Gamh_y_p, i, j, k) - Gamh_y_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Gamh_z_o0_t3_m0_l0 = -Gamh_z_falloff * (vector(Gamh_z_p, i, j, k) - Gamh_z_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Betau_x_o0_t3_m0_l0 = -Betau_x_falloff * (vector(Betau_x_p, i, j, k) - Betau_x_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Betau_y_o0_t3_m0_l0 = -Betau_y_falloff * (vector(Betau_y_p, i, j, k) - Betau_y_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Betau_z_o0_t3_m0_l0 = -Betau_z_falloff * (vector(Betau_z_p, i, j, k) - Betau_z_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Alpha_o0_t3_m0_l0 = -Alpha_falloff * (vector(Alpha_p, i, j, k) - Alpha_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_chi_o0_t3_m0_l0 = -chi_falloff * (vector(chi_p, i, j, k) - chi_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_trK_o0_t3_m0_l0 = -trK_falloff * (vector(trK_p, i, j, k) - trK_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_theta_o0_t3_m0_l0 = -theta_falloff * (vector(theta_p, i, j, k) - theta_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_m_gtd_xx_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xxx_o0_t0_m0_l0;
							i_m_gtd_xx_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xxy_o0_t0_m0_l0;
							i_m_gtd_xx_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xxz_o0_t0_m0_l0;
							i_m_gtd_xy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yxx_o0_t0_m0_l0;
							i_m_gtd_xy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyy_o0_t0_m0_l0;
							i_m_gtd_xy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyz_o0_t1_m0_l0;
							i_m_gtd_xz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_zxx_o0_t0_m0_l0;
							i_m_gtd_xz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyz_o0_t0_m0_l0;
							i_m_gtd_xz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xzz_o0_t0_m0_l0;
							i_m_gtd_yy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyy_o0_t1_m0_l0;
							i_m_gtd_yy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yyy_o0_t0_m0_l0;
							i_m_gtd_yy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yyz_o0_t0_m0_l0;
							i_m_gtd_yz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyz_o0_t2_m0_l0;
							i_m_gtd_yz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_zyy_o0_t0_m0_l0;
							i_m_gtd_yz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yzz_o0_t0_m0_l0;
							i_m_gtd_zz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xzz_o0_t1_m0_l0;
							i_m_gtd_zz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yzz_o0_t1_m0_l0;
							i_m_gtd_zz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_zzz_o0_t0_m0_l0;
							i_m_Atd_xx_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xx_o0_t0_m0_l0;
							i_m_Atd_xx_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xx_o0_t1_m0_l0;
							i_m_Atd_xx_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xx_o0_t2_m0_l0;
							i_m_Atd_xy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xy_o0_t0_m0_l0;
							i_m_Atd_xy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xy_o0_t1_m0_l0;
							i_m_Atd_xy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xy_o0_t2_m0_l0;
							i_m_Atd_xz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xz_o0_t0_m0_l0;
							i_m_Atd_xz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xz_o0_t1_m0_l0;
							i_m_Atd_xz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xz_o0_t2_m0_l0;
							i_m_Atd_yy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yy_o0_t0_m0_l0;
							i_m_Atd_yy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yy_o0_t1_m0_l0;
							i_m_Atd_yy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yy_o0_t2_m0_l0;
							i_m_Atd_yz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yz_o0_t0_m0_l0;
							i_m_Atd_yz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yz_o0_t1_m0_l0;
							i_m_Atd_yz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yz_o0_t2_m0_l0;
							i_m_Atd_zz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_zz_o0_t0_m0_l0;
							i_m_Atd_zz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_zz_o0_t1_m0_l0;
							i_m_Atd_zz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_zz_o0_t2_m0_l0;
							i_m_Gamh_x_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xx_o0_t6_m0_l0;
							i_m_Gamh_x_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xy_o0_t6_m0_l0;
							i_m_Gamh_x_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xz_o0_t6_m0_l0;
							i_m_Gamh_y_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xx_o0_t7_m0_l0;
							i_m_Gamh_y_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xy_o0_t7_m0_l0;
							i_m_Gamh_y_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xz_o0_t7_m0_l0;
							i_m_Gamh_z_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xx_o0_t8_m0_l0;
							i_m_Gamh_z_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xy_o0_t8_m0_l0;
							i_m_Gamh_z_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xz_o0_t8_m0_l0;
							i_m_Betau_x_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_div_Beta_o0_t0_m0_l0;
							i_m_Betau_x_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_y_o0_t7_m0_l0;
							i_m_Betau_x_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_z_o0_t7_m0_l0;
							i_m_Betau_y_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_x_o0_t8_m0_l0;
							i_m_Betau_y_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_div_Beta_o0_t1_m0_l0;
							i_m_Betau_y_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_z_o0_t8_m0_l0;
							i_m_Betau_z_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_x_o0_t9_m0_l0;
							i_m_Betau_z_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_y_o0_t9_m0_l0;
							i_m_Betau_z_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_div_Beta_o0_t2_m0_l0;
							i_m_Alpha_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Psi1_xx_o0_t1_m0_l0;
							i_m_Alpha_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Psi1_xx_o0_t2_m0_l0;
							i_m_Alpha_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Psi1_xx_o0_t3_m0_l0;
							i_m_chi_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rpd_xx_o0_t0_m0_l0;
							i_m_chi_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rpd_xx_o0_t1_m0_l0;
							i_m_chi_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rpd_xx_o0_t2_m0_l0;
							i_m_trK_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t0_m0_l0;
							i_m_trK_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t1_m0_l0;
							i_m_trK_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t2_m0_l0;
							i_m_theta_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t3_m0_l0;
							i_m_theta_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t4_m0_l0;
							i_m_theta_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t5_m0_l0;
							RHS_gtd_xx = ((i_m_gtd_xx_o0_t0_l0 + i_m_gtd_xx_o0_t1_l0) + i_m_gtd_xx_o0_t2_l0) + i_d_gtd_xx_o0_t3_m0_l0;
							RHS_gtd_xy = ((i_m_gtd_xy_o0_t0_l0 + i_m_gtd_xy_o0_t1_l0) + i_m_gtd_xy_o0_t2_l0) + i_d_gtd_xy_o0_t3_m0_l0;
							RHS_gtd_xz = ((i_m_gtd_xz_o0_t0_l0 + i_m_gtd_xz_o0_t1_l0) + i_m_gtd_xz_o0_t2_l0) + i_d_gtd_xz_o0_t3_m0_l0;
							RHS_gtd_yy = ((i_m_gtd_yy_o0_t0_l0 + i_m_gtd_yy_o0_t1_l0) + i_m_gtd_yy_o0_t2_l0) + i_d_gtd_yy_o0_t3_m0_l0;
							RHS_gtd_yz = ((i_m_gtd_yz_o0_t0_l0 + i_m_gtd_yz_o0_t1_l0) + i_m_gtd_yz_o0_t2_l0) + i_d_gtd_yz_o0_t3_m0_l0;
							RHS_gtd_zz = ((i_m_gtd_zz_o0_t0_l0 + i_m_gtd_zz_o0_t1_l0) + i_m_gtd_zz_o0_t2_l0) + i_d_gtd_zz_o0_t3_m0_l0;
							RHS_Atd_xx = ((i_m_Atd_xx_o0_t0_l0 + i_m_Atd_xx_o0_t1_l0) + i_m_Atd_xx_o0_t2_l0) + i_d_Atd_xx_o0_t3_m0_l0;
							RHS_Atd_xy = ((i_m_Atd_xy_o0_t0_l0 + i_m_Atd_xy_o0_t1_l0) + i_m_Atd_xy_o0_t2_l0) + i_d_Atd_xy_o0_t3_m0_l0;
							RHS_Atd_xz = ((i_m_Atd_xz_o0_t0_l0 + i_m_Atd_xz_o0_t1_l0) + i_m_Atd_xz_o0_t2_l0) + i_d_Atd_xz_o0_t3_m0_l0;
							RHS_Atd_yy = ((i_m_Atd_yy_o0_t0_l0 + i_m_Atd_yy_o0_t1_l0) + i_m_Atd_yy_o0_t2_l0) + i_d_Atd_yy_o0_t3_m0_l0;
							RHS_Atd_yz = ((i_m_Atd_yz_o0_t0_l0 + i_m_Atd_yz_o0_t1_l0) + i_m_Atd_yz_o0_t2_l0) + i_d_Atd_yz_o0_t3_m0_l0;
							RHS_Atd_zz = ((i_m_Atd_zz_o0_t0_l0 + i_m_Atd_zz_o0_t1_l0) + i_m_Atd_zz_o0_t2_l0) + i_d_Atd_zz_o0_t3_m0_l0;
							RHS_Gamh_x = ((i_m_Gamh_x_o0_t0_l0 + i_m_Gamh_x_o0_t1_l0) + i_m_Gamh_x_o0_t2_l0) + i_d_Gamh_x_o0_t3_m0_l0;
							RHS_Gamh_y = ((i_m_Gamh_y_o0_t0_l0 + i_m_Gamh_y_o0_t1_l0) + i_m_Gamh_y_o0_t2_l0) + i_d_Gamh_y_o0_t3_m0_l0;
							RHS_Gamh_z = ((i_m_Gamh_z_o0_t0_l0 + i_m_Gamh_z_o0_t1_l0) + i_m_Gamh_z_o0_t2_l0) + i_d_Gamh_z_o0_t3_m0_l0;
							RHS_Betau_x = ((i_m_Betau_x_o0_t0_l0 + i_m_Betau_x_o0_t1_l0) + i_m_Betau_x_o0_t2_l0) + i_d_Betau_x_o0_t3_m0_l0;
							RHS_Betau_y = ((i_m_Betau_y_o0_t0_l0 + i_m_Betau_y_o0_t1_l0) + i_m_Betau_y_o0_t2_l0) + i_d_Betau_y_o0_t3_m0_l0;
							RHS_Betau_z = ((i_m_Betau_z_o0_t0_l0 + i_m_Betau_z_o0_t1_l0) + i_m_Betau_z_o0_t2_l0) + i_d_Betau_z_o0_t3_m0_l0;
							RHS_Alpha = ((i_m_Alpha_o0_t0_l0 + i_m_Alpha_o0_t1_l0) + i_m_Alpha_o0_t2_l0) + i_d_Alpha_o0_t3_m0_l0;
							RHS_chi = ((i_m_chi_o0_t0_l0 + i_m_chi_o0_t1_l0) + i_m_chi_o0_t2_l0) + i_d_chi_o0_t3_m0_l0;
							RHS_trK = ((i_m_trK_o0_t0_l0 + i_m_trK_o0_t1_l0) + i_m_trK_o0_t2_l0) + i_d_trK_o0_t3_m0_l0;
							RHS_theta = ((i_m_theta_o0_t0_l0 + i_m_theta_o0_t1_l0) + i_m_theta_o0_t2_l0) + i_d_theta_o0_t3_m0_l0;
						}
						if (dissipation_factor_Sfd_x > 0.0) {
							RHS_Sfd_x = RHS_Sfd_x + dissipation_factor_Sfd_x * (meshDissipation_i(Sfd_x_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Sfd_x_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Sfd_x_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Sfd_y > 0.0) {
							RHS_Sfd_y = RHS_Sfd_y + dissipation_factor_Sfd_y * (meshDissipation_i(Sfd_y_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Sfd_y_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Sfd_y_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Sfd_z > 0.0) {
							RHS_Sfd_z = RHS_Sfd_z + dissipation_factor_Sfd_z * (meshDissipation_i(Sfd_z_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Sfd_z_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Sfd_z_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Bfu_x > 0.0) {
							RHS_Bfu_x = RHS_Bfu_x + dissipation_factor_Bfu_x * (meshDissipation_i(Bfu_x_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Bfu_x_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Bfu_x_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Bfu_y > 0.0) {
							RHS_Bfu_y = RHS_Bfu_y + dissipation_factor_Bfu_y * (meshDissipation_i(Bfu_y_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Bfu_y_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Bfu_y_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Bfu_z > 0.0) {
							RHS_Bfu_z = RHS_Bfu_z + dissipation_factor_Bfu_z * (meshDissipation_i(Bfu_z_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Bfu_z_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Bfu_z_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Df > 0.0) {
							RHS_Df = RHS_Df + dissipation_factor_Df * (meshDissipation_i(Df_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Df_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Df_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_DYf > 0.0) {
							RHS_DYf = RHS_DYf + dissipation_factor_DYf * (meshDissipation_i(DYf_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(DYf_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(DYf_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_tauf > 0.0) {
							RHS_tauf = RHS_tauf + dissipation_factor_tauf * (meshDissipation_i(tauf_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(tauf_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(tauf_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_phif > 0.0) {
							RHS_phif = RHS_phif + dissipation_factor_phif * (meshDissipation_i(phif_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(phif_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(phif_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_xx > 0.0) {
							RHS_gtd_xx = RHS_gtd_xx + dissipation_factor_gtd_xx * (meshDissipation_i(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(gtd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_xy > 0.0) {
							RHS_gtd_xy = RHS_gtd_xy + dissipation_factor_gtd_xy * (meshDissipation_i(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(gtd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_xz > 0.0) {
							RHS_gtd_xz = RHS_gtd_xz + dissipation_factor_gtd_xz * (meshDissipation_i(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(gtd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_yy > 0.0) {
							RHS_gtd_yy = RHS_gtd_yy + dissipation_factor_gtd_yy * (meshDissipation_i(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(gtd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_yz > 0.0) {
							RHS_gtd_yz = RHS_gtd_yz + dissipation_factor_gtd_yz * (meshDissipation_i(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(gtd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_zz > 0.0) {
							RHS_gtd_zz = RHS_gtd_zz + dissipation_factor_gtd_zz * (meshDissipation_i(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(gtd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_xx > 0.0) {
							RHS_Atd_xx = RHS_Atd_xx + dissipation_factor_Atd_xx * (meshDissipation_i(Atd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Atd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Atd_xx_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_xy > 0.0) {
							RHS_Atd_xy = RHS_Atd_xy + dissipation_factor_Atd_xy * (meshDissipation_i(Atd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Atd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Atd_xy_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_xz > 0.0) {
							RHS_Atd_xz = RHS_Atd_xz + dissipation_factor_Atd_xz * (meshDissipation_i(Atd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Atd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Atd_xz_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_yy > 0.0) {
							RHS_Atd_yy = RHS_Atd_yy + dissipation_factor_Atd_yy * (meshDissipation_i(Atd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Atd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Atd_yy_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_yz > 0.0) {
							RHS_Atd_yz = RHS_Atd_yz + dissipation_factor_Atd_yz * (meshDissipation_i(Atd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Atd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Atd_yz_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_zz > 0.0) {
							RHS_Atd_zz = RHS_Atd_zz + dissipation_factor_Atd_zz * (meshDissipation_i(Atd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Atd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Atd_zz_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Gamh_x > 0.0) {
							RHS_Gamh_x = RHS_Gamh_x + dissipation_factor_Gamh_x * (meshDissipation_i(Gamh_x_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Gamh_x_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Gamh_x_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Gamh_y > 0.0) {
							RHS_Gamh_y = RHS_Gamh_y + dissipation_factor_Gamh_y * (meshDissipation_i(Gamh_y_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Gamh_y_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Gamh_y_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Gamh_z > 0.0) {
							RHS_Gamh_z = RHS_Gamh_z + dissipation_factor_Gamh_z * (meshDissipation_i(Gamh_z_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Gamh_z_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Gamh_z_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Betau_x > 0.0) {
							RHS_Betau_x = RHS_Betau_x + dissipation_factor_Betau_x * (meshDissipation_i(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Betau_x_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Betau_y > 0.0) {
							RHS_Betau_y = RHS_Betau_y + dissipation_factor_Betau_y * (meshDissipation_i(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Betau_y_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Betau_z > 0.0) {
							RHS_Betau_z = RHS_Betau_z + dissipation_factor_Betau_z * (meshDissipation_i(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Betau_z_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Alpha > 0.0) {
							RHS_Alpha = RHS_Alpha + dissipation_factor_Alpha * (meshDissipation_i(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(Alpha_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_chi > 0.0) {
							RHS_chi = RHS_chi + dissipation_factor_chi * (meshDissipation_i(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(chi_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_trK > 0.0) {
							RHS_trK = RHS_trK + dissipation_factor_trK * (meshDissipation_i(trK_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(trK_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(trK_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_theta > 0.0) {
							RHS_theta = RHS_theta + dissipation_factor_theta * (meshDissipation_i(theta_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(theta_p, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(theta_p, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						vector(rk1Sfd_x, i, j, k) = RK4P1_(RHS_Sfd_x, vector(Sfd_x_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Sfd_y, i, j, k) = RK4P1_(RHS_Sfd_y, vector(Sfd_y_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Sfd_z, i, j, k) = RK4P1_(RHS_Sfd_z, vector(Sfd_z_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Bfu_x, i, j, k) = RK4P1_(RHS_Bfu_x, vector(Bfu_x_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Bfu_y, i, j, k) = RK4P1_(RHS_Bfu_y, vector(Bfu_y_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Bfu_z, i, j, k) = RK4P1_(RHS_Bfu_z, vector(Bfu_z_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Df, i, j, k) = RK4P1_(RHS_Df, vector(Df_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1DYf, i, j, k) = RK4P1_(RHS_DYf, vector(DYf_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1tauf, i, j, k) = RK4P1_(RHS_tauf, vector(tauf_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1phif, i, j, k) = RK4P1_(RHS_phif, vector(phif_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1gtd_xx, i, j, k) = RK4P1_(RHS_gtd_xx, vector(gtd_xx_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1gtd_xy, i, j, k) = RK4P1_(RHS_gtd_xy, vector(gtd_xy_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1gtd_xz, i, j, k) = RK4P1_(RHS_gtd_xz, vector(gtd_xz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1gtd_yy, i, j, k) = RK4P1_(RHS_gtd_yy, vector(gtd_yy_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1gtd_yz, i, j, k) = RK4P1_(RHS_gtd_yz, vector(gtd_yz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1gtd_zz, i, j, k) = RK4P1_(RHS_gtd_zz, vector(gtd_zz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Atd_xx, i, j, k) = RK4P1_(RHS_Atd_xx, vector(Atd_xx_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Atd_xy, i, j, k) = RK4P1_(RHS_Atd_xy, vector(Atd_xy_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Atd_xz, i, j, k) = RK4P1_(RHS_Atd_xz, vector(Atd_xz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Atd_yy, i, j, k) = RK4P1_(RHS_Atd_yy, vector(Atd_yy_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Atd_yz, i, j, k) = RK4P1_(RHS_Atd_yz, vector(Atd_yz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Atd_zz, i, j, k) = RK4P1_(RHS_Atd_zz, vector(Atd_zz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Gamh_x, i, j, k) = RK4P1_(RHS_Gamh_x, vector(Gamh_x_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Gamh_y, i, j, k) = RK4P1_(RHS_Gamh_y, vector(Gamh_y_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Gamh_z, i, j, k) = RK4P1_(RHS_Gamh_z, vector(Gamh_z_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Betau_x, i, j, k) = RK4P1_(RHS_Betau_x, vector(Betau_x_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Betau_y, i, j, k) = RK4P1_(RHS_Betau_y, vector(Betau_y_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Betau_z, i, j, k) = RK4P1_(RHS_Betau_z, vector(Betau_z_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1Alpha, i, j, k) = RK4P1_(RHS_Alpha, vector(Alpha_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1chi, i, j, k) = RK4P1_(RHS_chi, vector(chi_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1trK, i, j, k) = RK4P1_(RHS_trK, vector(trK_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk1theta, i, j, k) = RK4P1_(RHS_theta, vector(theta_p, i, j, k), dx, simPlat_dt, ilast, jlast);
					}
				}
			}
		}
	}
	//Fill ghosts and periodical boundaries
	time_interpolate_operator_mesh1->setStep(1);
	d_bdry_sched_advance2[ln]->fillData(current_time + simPlat_dt * 0.5, false);
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		//Hard region field distance variables
		double* d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
	
		double* rk1Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_x_id).get())->getPointer();
		double* rk1Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_y_id).get())->getPointer();
		double* rk1Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_z_id).get())->getPointer();
		double* rk1Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_x_id).get())->getPointer();
		double* rk1Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_y_id).get())->getPointer();
		double* rk1Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_z_id).get())->getPointer();
		double* rk1Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Df_id).get())->getPointer();
		double* rk1DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1DYf_id).get())->getPointer();
		double* rk1tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1tauf_id).get())->getPointer();
		double* rk1phif = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1phif_id).get())->getPointer();
		double* rk1gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xx_id).get())->getPointer();
		double* rk1gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xy_id).get())->getPointer();
		double* rk1gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xz_id).get())->getPointer();
		double* rk1gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yy_id).get())->getPointer();
		double* rk1gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yz_id).get())->getPointer();
		double* rk1gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_zz_id).get())->getPointer();
		double* rk1Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xx_id).get())->getPointer();
		double* rk1Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xy_id).get())->getPointer();
		double* rk1Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xz_id).get())->getPointer();
		double* rk1Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_yy_id).get())->getPointer();
		double* rk1Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_yz_id).get())->getPointer();
		double* rk1Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_zz_id).get())->getPointer();
		double* rk1Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_x_id).get())->getPointer();
		double* rk1Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_y_id).get())->getPointer();
		double* rk1Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_z_id).get())->getPointer();
		double* rk1Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_x_id).get())->getPointer();
		double* rk1Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_y_id).get())->getPointer();
		double* rk1Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_z_id).get())->getPointer();
		double* rk1Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Alpha_id).get())->getPointer();
		double* rk1chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1chi_id).get())->getPointer();
		double* rk1trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1trK_id).get())->getPointer();
		double* rk1theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1theta_id).get())->getPointer();
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, d_ghost_width, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, d_ghost_width, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, d_ghost_width, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, d_ghost_width, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, d_ghost_width, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, d_ghost_width, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, d_ghost_width, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, d_ghost_width, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, d_ghost_width, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i - 1, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i - 1, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i - 1, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i - 1, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i - 1, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i - 1, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i - 1, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i - 1, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i - 1, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, d_ghost_width, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, d_ghost_width, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, d_ghost_width, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, d_ghost_width, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, d_ghost_width, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, d_ghost_width, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, d_ghost_width, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, d_ghost_width, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, d_ghost_width, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j - 1, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j - 1, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j - 1, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j - 1, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j - 1, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j - 1, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j - 1, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j - 1, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j - 1, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, d_ghost_width);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, d_ghost_width);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, d_ghost_width);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, d_ghost_width);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, d_ghost_width);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, d_ghost_width);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, d_ghost_width);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, d_ghost_width);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, d_ghost_width);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, k - 1);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, k - 1);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, k - 1);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, k - 1);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, k - 1);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, k - 1);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, k - 1);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, k - 1);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, k - 1);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, d_ghost_width, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, d_ghost_width, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, d_ghost_width, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, d_ghost_width, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, d_ghost_width, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, d_ghost_width, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, d_ghost_width, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, d_ghost_width, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, d_ghost_width, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i - 1, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i - 1, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i - 1, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i - 1, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i - 1, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i - 1, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i - 1, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i - 1, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i - 1, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, d_ghost_width, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, d_ghost_width, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, d_ghost_width, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, d_ghost_width, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, d_ghost_width, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, d_ghost_width, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, d_ghost_width, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, d_ghost_width, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, d_ghost_width, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j - 1, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j - 1, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j - 1, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j - 1, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j - 1, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j - 1, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j - 1, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j - 1, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j - 1, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, d_ghost_width);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, d_ghost_width);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, d_ghost_width);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, d_ghost_width);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, d_ghost_width);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, d_ghost_width);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, d_ghost_width);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, d_ghost_width);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, d_ghost_width);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, k - 1);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, k - 1);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, k - 1);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, k - 1);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, k - 1);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, k - 1);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, k - 1);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, k - 1);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, k - 1);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, d_ghost_width, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, d_ghost_width, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, d_ghost_width, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, d_ghost_width, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, d_ghost_width, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, d_ghost_width, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, d_ghost_width, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, d_ghost_width, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, d_ghost_width, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i - 1, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i - 1, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i - 1, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i - 1, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i - 1, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i - 1, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i - 1, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i - 1, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i - 1, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, d_ghost_width, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, d_ghost_width, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, d_ghost_width, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, d_ghost_width, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, d_ghost_width, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, d_ghost_width, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, d_ghost_width, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, d_ghost_width, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, d_ghost_width, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j - 1, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j - 1, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j - 1, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j - 1, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j - 1, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j - 1, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j - 1, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j - 1, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j - 1, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, d_ghost_width);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, d_ghost_width);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, d_ghost_width);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, d_ghost_width);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, d_ghost_width);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, d_ghost_width);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, d_ghost_width);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, d_ghost_width);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, d_ghost_width);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, k - 1);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, k - 1);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, k - 1);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, k - 1);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, k - 1);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, k - 1);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, k - 1);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, k - 1);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, k - 1);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, d_ghost_width, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, d_ghost_width, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, d_ghost_width, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, d_ghost_width, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, d_ghost_width, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, d_ghost_width, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, d_ghost_width, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, d_ghost_width, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, d_ghost_width, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i - 1, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i - 1, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i - 1, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i - 1, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i - 1, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i - 1, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i - 1, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i - 1, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i - 1, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, d_ghost_width, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, d_ghost_width, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, d_ghost_width, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, d_ghost_width, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, d_ghost_width, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, d_ghost_width, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, d_ghost_width, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, d_ghost_width, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, d_ghost_width, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j - 1, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j - 1, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j - 1, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j - 1, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j - 1, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j - 1, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j - 1, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j - 1, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j - 1, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, d_ghost_width);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, d_ghost_width);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, d_ghost_width);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, d_ghost_width);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, d_ghost_width);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, d_ghost_width);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, d_ghost_width);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, d_ghost_width);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, d_ghost_width);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, k - 1);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, k - 1);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, k - 1);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, k - 1);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, k - 1);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, k - 1);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, k - 1);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, k - 1);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, k - 1);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, d_ghost_width, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, d_ghost_width, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, d_ghost_width, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, d_ghost_width, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, d_ghost_width, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, d_ghost_width, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, d_ghost_width, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, d_ghost_width, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, d_ghost_width, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i - 1, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i - 1, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i - 1, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i - 1, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i - 1, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i - 1, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i - 1, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i - 1, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i - 1, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, d_ghost_width, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, d_ghost_width, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, d_ghost_width, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, d_ghost_width, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, d_ghost_width, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, d_ghost_width, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, d_ghost_width, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, d_ghost_width, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, d_ghost_width, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j - 1, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j - 1, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j - 1, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j - 1, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j - 1, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j - 1, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j - 1, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j - 1, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j - 1, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, d_ghost_width);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, d_ghost_width);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, d_ghost_width);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, d_ghost_width);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, d_ghost_width);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, d_ghost_width);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, d_ghost_width);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, d_ghost_width);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, d_ghost_width);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, k - 1);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, k - 1);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, k - 1);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, k - 1);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, k - 1);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, k - 1);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, k - 1);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, k - 1);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, k - 1);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, d_ghost_width, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, d_ghost_width, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, d_ghost_width, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, d_ghost_width, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, d_ghost_width, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, d_ghost_width, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, d_ghost_width, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, d_ghost_width, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, d_ghost_width, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i - 1, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i - 1, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i - 1, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i - 1, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i - 1, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i - 1, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i - 1, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i - 1, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i - 1, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, d_ghost_width, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, d_ghost_width, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, d_ghost_width, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, d_ghost_width, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, d_ghost_width, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, d_ghost_width, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, d_ghost_width, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, d_ghost_width, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, d_ghost_width, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j - 1, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j - 1, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j - 1, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j - 1, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j - 1, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j - 1, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j - 1, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j - 1, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j - 1, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, d_ghost_width);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, d_ghost_width);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, d_ghost_width);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, d_ghost_width);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, d_ghost_width);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, d_ghost_width);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, d_ghost_width);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, d_ghost_width);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, d_ghost_width);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, k - 1);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, k - 1);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, k - 1);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, k - 1);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, k - 1);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, k - 1);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, k - 1);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, k - 1);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, k - 1);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, d_ghost_width, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, d_ghost_width, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, d_ghost_width, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, d_ghost_width, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, d_ghost_width, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, d_ghost_width, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, d_ghost_width, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, d_ghost_width, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, d_ghost_width, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i - 1, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i - 1, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i - 1, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i - 1, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i - 1, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i - 1, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i - 1, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i - 1, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i - 1, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, d_ghost_width, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, d_ghost_width, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, d_ghost_width, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, d_ghost_width, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, d_ghost_width, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, d_ghost_width, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, d_ghost_width, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, d_ghost_width, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, d_ghost_width, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j - 1, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j - 1, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j - 1, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j - 1, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j - 1, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j - 1, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j - 1, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j - 1, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j - 1, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, d_ghost_width);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, d_ghost_width);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, d_ghost_width);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, d_ghost_width);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, d_ghost_width);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, d_ghost_width);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, d_ghost_width);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, d_ghost_width);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, d_ghost_width);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, k - 1);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, k - 1);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, k - 1);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, k - 1);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, k - 1);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, k - 1);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, k - 1);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, k - 1);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, k - 1);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, d_ghost_width, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, d_ghost_width, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, d_ghost_width, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, d_ghost_width, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, d_ghost_width, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, d_ghost_width, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, d_ghost_width, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, d_ghost_width, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, d_ghost_width, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i - 1, j, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i - 1, j, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i - 1, j, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i - 1, j, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i - 1, j, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i - 1, j, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i - 1, j, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i - 1, j, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i - 1, j, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, d_ghost_width, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, d_ghost_width, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, d_ghost_width, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, d_ghost_width, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, d_ghost_width, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, d_ghost_width, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, d_ghost_width, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, d_ghost_width, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, d_ghost_width, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j - 1, k);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j - 1, k);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j - 1, k);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j - 1, k);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j - 1, k);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j - 1, k);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j - 1, k);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j - 1, k);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j - 1, k);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, d_ghost_width);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, d_ghost_width);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, d_ghost_width);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, d_ghost_width);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, d_ghost_width);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, d_ghost_width);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, d_ghost_width);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, d_ghost_width);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, d_ghost_width);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk1Sfd_x, i, j, k) = vector(rk1Sfd_x, i, j, k - 1);
						vector(rk1Sfd_y, i, j, k) = vector(rk1Sfd_y, i, j, k - 1);
						vector(rk1Sfd_z, i, j, k) = vector(rk1Sfd_z, i, j, k - 1);
						vector(rk1Bfu_x, i, j, k) = vector(rk1Bfu_x, i, j, k - 1);
						vector(rk1Bfu_y, i, j, k) = vector(rk1Bfu_y, i, j, k - 1);
						vector(rk1Bfu_z, i, j, k) = vector(rk1Bfu_z, i, j, k - 1);
						vector(rk1Df, i, j, k) = vector(rk1Df, i, j, k - 1);
						vector(rk1DYf, i, j, k) = vector(rk1DYf, i, j, k - 1);
						vector(rk1tauf, i, j, k) = vector(rk1tauf, i, j, k - 1);
						vector(rk1phif, i, j, k) = vector(rk1phif, i, j, k - 1);
					}
				}
			}
		}
	}
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		double* rk1chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1chi_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* rk1Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_x_id).get())->getPointer();
		double* rk1Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_y_id).get())->getPointer();
		double* rk1Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_z_id).get())->getPointer();
		double* rk1tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1tauf_id).get())->getPointer();
		double* rk1gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xy_id).get())->getPointer();
		double* rk1gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yy_id).get())->getPointer();
		double* rk1gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yz_id).get())->getPointer();
		double* rk1gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xz_id).get())->getPointer();
		double* rk1gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_zz_id).get())->getPointer();
		double* rk1gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xx_id).get())->getPointer();
		double* rk1Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_x_id).get())->getPointer();
		double* rk1Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_y_id).get())->getPointer();
		double* rk1Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_z_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* rk1Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Df_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* rk1Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Alpha_id).get())->getPointer();
		double* rk1DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1DYf_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double rk1chi_max, rk1h, rk1sdetg, rk1Bu_x, rk1Bu_y, rk1Bu_z, rk1tau, rk1inv_chi, rk1Bfd_y, rk1Bfd_z, rk1Bfd_x, rk1decay_factor, rk1detgtd, rk1idetgtd, rk1gtu_zz, rk1gtu_xx, rk1gtu_xz, rk1gtu_xy, rk1Sfu_x, rk1Su_x, rk1vfu_x, rk1Sd_x, rk1Sd_y, rk1gtu_yy, rk1Sd_z, rk1gtu_yz, rk1vfu_z, rk1Sfu_y, rk1Su_y, rk1Sfu_z, rk1Su_z, rk1vfu_y, rk1sqvf, rk1sqW, rk1W, rk1Bd_x, rk1D, rk1Bd_y, rk1Bd_z, rhof_tmp, Yef_tmp, Tf_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, dpfdye_tmp, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tauf_tmp, Df_tmp, DYf_tmp, Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, Efu_x_tmp, Efu_y_tmp, Efu_z_tmp, vacuum_D_tmp, vacuum_tau_tmp, vacuum_rho_reset_tmp;
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j, k));
					rk1h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
					rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
					rk1Bu_x = vector(rk1Bfu_x, i, j, k) / rk1sdetg;
					rk1Bu_y = vector(rk1Bfu_y, i, j, k) / rk1sdetg;
					rk1Bu_z = vector(rk1Bfu_z, i, j, k) / rk1sdetg;
					rk1tau = vector(rk1tauf, i, j, k) / rk1sdetg;
					rk1inv_chi = 1.0 / fabs(rk1chi_max);
					rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_yy, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Bfu_z, i, j, k));
					rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_zz, i, j, k) * vector(rk1Bfu_z, i, j, k));
					rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_xy, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_xz, i, j, k) * vector(rk1Bfu_z, i, j, k));
					rk1decay_factor = MIN(1.0, pow((R_0 / MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))), eta_damping_exp));
					rk1detgtd = vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) * vector(rk1gtd_xx, i, j, k) + (-vector(rk1gtd_yy, i, j, k) * (vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k))) + (-vector(rk1gtd_zz, i, j, k) * (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k))) + (-(vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k)) * vector(rk1gtd_xx, i, j, k)) + 2.0 * vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k);
					rk1idetgtd = 1.0 / rk1detgtd;
					rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yy, i, j, k) - vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k));
					rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) - vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k));
					rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_yz, i, j, k) - vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_xz, i, j, k));
					rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_zz, i, j, k)) + vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xz, i, j, k));
					rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j, k) + rk1gtu_xy * vector(rk1Sfd_y, i, j, k) + rk1gtu_xz * vector(rk1Sfd_z, i, j, k));
					rk1Su_x = rk1Sfu_x / rk1sdetg;
					rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j, k) + rk1gtu_xy * vector(vfd_y, i, j, k) + rk1gtu_xz * vector(vfd_z, i, j, k));
					rk1Sd_x = vector(rk1Sfd_x, i, j, k) / rk1sdetg;
					rk1Sd_y = vector(rk1Sfd_y, i, j, k) / rk1sdetg;
					rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_zz, i, j, k) - vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k));
					rk1Sd_z = vector(rk1Sfd_z, i, j, k) / rk1sdetg;
					rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yz, i, j, k)) + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k));
					rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j, k) + rk1gtu_yz * vector(vfd_y, i, j, k) + rk1gtu_zz * vector(vfd_z, i, j, k));
					rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j, k) + rk1gtu_yy * vector(rk1Sfd_y, i, j, k) + rk1gtu_yz * vector(rk1Sfd_z, i, j, k));
					rk1Su_y = rk1Sfu_y / rk1sdetg;
					rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j, k) + rk1gtu_yz * vector(rk1Sfd_y, i, j, k) + rk1gtu_zz * vector(rk1Sfd_z, i, j, k));
					rk1Su_z = rk1Sfu_z / rk1sdetg;
					rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j, k) + rk1gtu_yy * vector(vfd_y, i, j, k) + rk1gtu_yz * vector(vfd_z, i, j, k));
					rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j, k) + rk1vfu_y * vector(vfd_y, i, j, k) + rk1vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
					rk1sqW = 1.0 / (1.0 - rk1sqvf);
					rk1W = sqrt(rk1sqW);
					rk1Bd_x = rk1Bfd_x / rk1sdetg;
					rk1D = vector(rk1Df, i, j, k) / rk1sdetg;
					rk1Bd_y = rk1Bfd_y / rk1sdetg;
					rk1Bd_z = rk1Bfd_z / rk1sdetg;
					vector(t_chie, i, j, k) = vector(chie, i, j, k);
					vector(t_chia, i, j, k) = vector(chia, i, j, k);
					vector(t_optdepthx, i, j, k) = vector(optdepthx, i, j, k);
					vector(t_optdeptha, i, j, k) = vector(optdeptha, i, j, k);
					vector(t_optdepthe, i, j, k) = vector(optdepthe, i, j, k);
					vector(rk1Alpha, i, j, k) = fabs(vector(rk1Alpha, i, j, k));
					rhof_tmp = vector(rhof, i, j, k);
					Yef_tmp = vector(rk1DYf, i, j, k) / vector(rk1Df, i, j, k);
					if (Yef_tmp < vacuum_ye) {
						vector(rk1DYf, i, j, k) = vacuum_ye_reset * vector(rk1Df, i, j, k);
						Yef_tmp = vacuum_ye_reset;
					}
					if (Yef_tmp > ye_maximum) {
						vector(rk1DYf, i, j, k) = 0.9 * ye_maximum * vector(rk1Df, i, j, k);
						Yef_tmp = 0.9 * ye_maximum;
					}
					Tf_tmp = vector(Tf, i, j, k);
					pf_tmp = vector(pf, i, j, k);
					epsf_tmp = vector(epsf, i, j, k);
					vfd_x_tmp = vector(vfd_x, i, j, k);
					vfd_y_tmp = vector(vfd_y, i, j, k);
					vfd_z_tmp = vector(vfd_z, i, j, k);
					sqcs_tmp = vector(sqcs, i, j, k);
					dpfdrho_tmp = vector(dpfdrho, i, j, k);
					dpfdeps_tmp = vector(dpfdeps, i, j, k);
					dpfdye_tmp = vector(dpfdye, i, j, k);
					Sfd_x_tmp = vector(rk1Sfd_x, i, j, k);
					Sfd_y_tmp = vector(rk1Sfd_y, i, j, k);
					Sfd_z_tmp = vector(rk1Sfd_z, i, j, k);
					tauf_tmp = vector(rk1tauf, i, j, k);
					Df_tmp = vector(rk1Df, i, j, k);
					DYf_tmp = vector(rk1DYf, i, j, k);
					Bfu_x_tmp = vector(rk1Bfu_x, i, j, k);
					Bfu_y_tmp = vector(rk1Bfu_y, i, j, k);
					Bfu_z_tmp = vector(rk1Bfu_z, i, j, k);
					if (equalsEq(externalCon2Prim, 1.0)) {
						Efu_x_tmp = vector(Efu_x, i, j, k);
						Efu_y_tmp = vector(Efu_y, i, j, k);
						Efu_z_tmp = vector(Efu_z, i, j, k);
						externalCon2prim(Efu_x_tmp, Efu_y_tmp, Efu_z_tmp, rhof_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, Df_tmp, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tauf_tmp, rk1W, vector(rk1chi, i, j, k), rk1gtu_xx, rk1gtu_xy, rk1gtu_xz, rk1gtu_yy, rk1gtu_yz, rk1gtu_zz, vector(rk1gtd_xx, i, j, k), vector(rk1gtd_xy, i, j, k), vector(rk1gtd_xz, i, j, k), vector(rk1gtd_yy, i, j, k), vector(rk1gtd_yz, i, j, k), vector(rk1gtd_zz, i, j, k), Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, dx, simPlat_dt, ilast, jlast);
						vector(Efu_x, i, j, k) = Efu_x_tmp;
						vector(Efu_y, i, j, k) = Efu_y_tmp;
						vector(Efu_z, i, j, k) = Efu_z_tmp;
					} else {
						vacuum_D_tmp = vacuum_D * rk1decay_factor;
						vacuum_tau_tmp = vacuum_tau * rk1decay_factor;
						vacuum_rho_reset_tmp = vacuum_rho_reset * rk1decay_factor;
						con2prim(eos_type, rhof_tmp, Yef_tmp, Tf_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, dpfdye_tmp, rho_0, rho_1, rho_2, a_0, a_1, a_2, a_3, K_0, K_1, K_2, K_3, gamma_0, gamma_1, gamma_2, gamma_3, rk1D, vacuum_D_tmp, Df_tmp, DYf_tmp, rk1sdetg, rk1Sd_x, rk1Sd_y, rk1Sd_z, rk1Su_x, rk1Su_y, rk1Su_z, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, rk1tau, vacuum_tau_tmp, tauf_tmp, rk1W, rk1h, vector(rk1chi, i, j, k), rk1gtu_xx, rk1gtu_xy, rk1gtu_xz, rk1gtu_yy, rk1gtu_yz, rk1gtu_zz, vector(rk1gtd_xx, i, j, k), vector(rk1gtd_xy, i, j, k), vector(rk1gtd_xz, i, j, k), vector(rk1gtd_yy, i, j, k), vector(rk1gtd_yz, i, j, k), vector(rk1gtd_zz, i, j, k), rk1Bd_x, rk1Bd_y, rk1Bd_z, rk1Bu_x, rk1Bu_y, rk1Bu_z, Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, max_error, gamma, rk1inv_chi, threshold_vacuum, threshold_sqSmax, threshold_sqBmax, energyShift, vacuum_ye_beta, vacuum_rho, vacuum_rho_reset_tmp, vacuum_P_reset, vacuum_ye_reset, vacuum_temp_reset, vacuum_tau_reset, minTableTemperature, minTableEnergy, coord0_data_Data, coord0_dx_Data, coord0_size_Data, coord1_data_Data, coord1_dx_Data, coord1_size_Data, coord2_data_Data, coord2_dx_Data, coord2_size_Data, vars_data_Data, vars_size_Data, dx, simPlat_dt, ilast, jlast);
						vector(Efu_x, i, j, k) = -((-vector(vfd_z, i, j, k) * rk1Bfd_y) + vector(vfd_y, i, j, k) * rk1Bfd_z) / (rk1sdetg * rk1sdetg);
						vector(Efu_y, i, j, k) = -(vector(vfd_z, i, j, k) * rk1Bfd_x - vector(vfd_x, i, j, k) * rk1Bfd_z) / (rk1sdetg * rk1sdetg);
						vector(Efu_z, i, j, k) = -((-vector(vfd_y, i, j, k) * rk1Bfd_x) + vector(vfd_x, i, j, k) * rk1Bfd_y) / (rk1sdetg * rk1sdetg);
					}
					vector(rhof, i, j, k) = rhof_tmp;
					vector(Yef, i, j, k) = Yef_tmp;
					vector(Tf, i, j, k) = Tf_tmp;
					vector(pf, i, j, k) = pf_tmp;
					vector(epsf, i, j, k) = epsf_tmp;
					vector(vfd_x, i, j, k) = vfd_x_tmp;
					vector(vfd_y, i, j, k) = vfd_y_tmp;
					vector(vfd_z, i, j, k) = vfd_z_tmp;
					vector(sqcs, i, j, k) = sqcs_tmp;
					vector(dpfdye, i, j, k) = dpfdye_tmp;
					vector(dpfdeps, i, j, k) = dpfdeps_tmp;
					vector(dpfdrho, i, j, k) = dpfdrho_tmp;
					vector(rk1Sfd_x, i, j, k) = Sfd_x_tmp;
					vector(rk1Sfd_y, i, j, k) = Sfd_y_tmp;
					vector(rk1Sfd_z, i, j, k) = Sfd_z_tmp;
					vector(rk1tauf, i, j, k) = tauf_tmp;
					vector(rk1Df, i, j, k) = Df_tmp;
					vector(rk1DYf, i, j, k) = DYf_tmp;
					vector(rk1Bfu_x, i, j, k) = Bfu_x_tmp;
					vector(rk1Bfu_y, i, j, k) = Bfu_y_tmp;
					vector(rk1Bfu_z, i, j, k) = Bfu_z_tmp;
					vector(t_chix, i, j, k) = vector(chix, i, j, k);
				}
			}
		}
	}
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* rk1DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1DYf_id).get())->getPointer();
		double* rk1Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Df_id).get())->getPointer();
		double* rk1Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_x_id).get())->getPointer();
		double* rk1Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_y_id).get())->getPointer();
		double* rk1Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_z_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* rk1chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1chi_id).get())->getPointer();
		double* rk1gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xy_id).get())->getPointer();
		double* rk1gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yy_id).get())->getPointer();
		double* rk1gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yz_id).get())->getPointer();
		double* rk1gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xz_id).get())->getPointer();
		double* rk1gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_zz_id).get())->getPointer();
		double* rk1gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xx_id).get())->getPointer();
		double* rk1Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_x_id).get())->getPointer();
		double* rk1Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_y_id).get())->getPointer();
		double* rk1Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_z_id).get())->getPointer();
		double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
		double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
		double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
		double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
		double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
		double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
		double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
		double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
		double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
		double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
		double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
		double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
		double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
		double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
		double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
		double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
		double* qnu_e = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_e_id).get())->getPointer();
		double* qnu_a = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_a_id).get())->getPointer();
		double* qnu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_x_id).get())->getPointer();
		double* rk1Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_x_id).get())->getPointer();
		double* rk1Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_y_id).get())->getPointer();
		double* rk1Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_z_id).get())->getPointer();
		double* rk1Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Alpha_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double d_optdeptheh_o0_t0_m0_l0, d_optdepthev_o0_t0_m0_l0, d_optdepthed_o0_t0_m0_l0, d_optdepthah_o0_t0_m0_l0, d_optdepthav_o0_t0_m0_l0, d_optdepthad_o0_t0_m0_l0, d_optdepthxh_o0_t0_m0_l0, d_optdepthxv_o0_t0_m0_l0, d_optdepthxd_o0_t0_m0_l0, d_chieh_o0_t0_m0_l0, d_chiev_o0_t0_m0_l0, d_chied_o0_t0_m0_l0, d_chiah_o0_t0_m0_l0, d_chiav_o0_t0_m0_l0, d_chiad_o0_t0_m0_l0, d_chixh_o0_t0_m0_l0, d_chixv_o0_t0_m0_l0, d_chixd_o0_t0_m0_l0, d_CovdDYf_z_o0_t0_m0_l0, d_CovdDYf_y_o0_t0_m0_l0, d_CovdDYf_x_o0_t0_m0_l0, d_CovdDf_z_o0_t0_m0_l0, d_CovdDf_y_o0_t0_m0_l0, d_CovdDf_x_o0_t0_m0_l0, d_CovdBfu_xx_o0_t0_m0_l0, d_CovdBfu_xy_o0_t0_m0_l0, d_CovdBfu_xz_o0_t0_m0_l0, d_CovdBfu_yx_o0_t0_m0_l0, d_CovdBfu_yy_o0_t0_m0_l0, d_CovdBfu_yz_o0_t0_m0_l0, d_CovdBfu_zx_o0_t0_m0_l0, d_CovdBfu_zy_o0_t0_m0_l0, d_CovdBfu_zz_o0_t0_m0_l0, d_CovdEfu_xx_o0_t0_m0_l0, d_CovdEfu_xy_o0_t0_m0_l0, d_CovdEfu_xz_o0_t0_m0_l0, d_CovdEfu_yx_o0_t0_m0_l0, d_CovdEfu_yy_o0_t0_m0_l0, d_CovdEfu_yz_o0_t0_m0_l0, d_CovdEfu_zx_o0_t0_m0_l0, d_CovdEfu_zy_o0_t0_m0_l0, d_CovdEfu_zz_o0_t0_m0_l0, d_Covdepsf_x_o0_t0_m0_l0, d_Covdepsf_y_o0_t0_m0_l0, d_Covdepsf_z_o0_t0_m0_l0, d_Covddpfdeps_x_o0_t0_m0_l0, d_Covddpfdeps_y_o0_t0_m0_l0, d_Covddpfdeps_z_o0_t0_m0_l0, d_Covddpfdye_x_o0_t0_m0_l0, d_Covddpfdye_y_o0_t0_m0_l0, d_Covddpfdye_z_o0_t0_m0_l0, d_Covddpfdrho_x_o0_t0_m0_l0, d_Covddpfdrho_y_o0_t0_m0_l0, d_Covddpfdrho_z_o0_t0_m0_l0, d_Covdvfd_xx_o0_t0_m0_l0, d_Covdvfd_xy_o0_t0_m0_l0, d_Covdvfd_xz_o0_t0_m0_l0, d_Covdvfd_yx_o0_t0_m0_l0, d_Covdvfd_yy_o0_t0_m0_l0, d_Covdvfd_yz_o0_t0_m0_l0, d_Covdvfd_zx_o0_t0_m0_l0, d_Covdvfd_zy_o0_t0_m0_l0, d_Covdvfd_zz_o0_t0_m0_l0, d_Covdrhof_x_o0_t0_m0_l0, d_Covdrhof_y_o0_t0_m0_l0, d_Covdrhof_z_o0_t0_m0_l0, d_Covdpf_x_o0_t0_m0_l0, d_Covdpf_y_o0_t0_m0_l0, d_Covdpf_z_o0_t0_m0_l0, d_CovdYef_x_o0_t0_m0_l0, d_CovdYef_y_o0_t0_m0_l0, d_CovdYef_z_o0_t0_m0_l0, rk1chiev, rk1chieh, rk1Covdvfd_xx, rk1Covdvfd_xz, rk1Covdvfd_xy, rk1chied, rk1Covddpfdeps_x, rk1Covddpfdeps_y, rk1Covddpfdeps_z, rk1Covdvfd_zx, rk1Covdvfd_zz, rk1Covdvfd_zy, rk1Covdvfd_yy, rk1Covdvfd_yx, rk1Covdvfd_yz, rk1Covdpf_x, rk1Covdpf_y, rk1Covdpf_z, rk1chi_max, rk1inv_chi, rk1Bfd_y, rk1Bfd_z, rk1Bfd_x, rk1sqBf, rk1chixh, rk1chixd, rk1Covddpfdrho_z, rk1Covddpfdrho_y, rk1Covddpfdrho_x, rk1sdetg, rk1m_CovdDf_x_o0_t0_l0, rk1CovdDf_x, rk1m_CovdDf_y_o0_t0_l0, rk1CovdDf_y, rk1m_CovdDf_z_o0_t0_l0, rk1CovdDf_z, rk1m_CovdBfu_zz_o0_t0_l0, rk1CovdBfu_zz, rk1m_CovdBfu_zy_o0_t0_l0, rk1CovdBfu_zy, rk1m_CovdBfu_zx_o0_t0_l0, rk1CovdBfu_zx, rk1m_CovdBfu_yz_o0_t0_l0, rk1CovdBfu_yz, rk1m_CovdBfu_yy_o0_t0_l0, rk1CovdBfu_yy, rk1m_CovdBfu_yx_o0_t0_l0, rk1CovdBfu_yx, rk1m_CovdBfu_xy_o0_t0_l0, rk1CovdBfu_xy, rk1m_CovdBfu_xx_o0_t0_l0, rk1CovdBfu_xx, rk1m_CovdBfu_xz_o0_t0_l0, rk1CovdBfu_xz, rk1CovdBfd_zx, rk1Covdrhof_z, rk1Covdrhof_y, rk1Covdrhof_x, rk1CovdBfd_zz, rk1CovdBfd_zy, rk1m_CovdDYf_x_o0_t0_l0, rk1CovdDYf_x, rk1m_CovdDYf_y_o0_t0_l0, rk1CovdDYf_y, rk1m_CovdDYf_z_o0_t0_l0, rk1CovdDYf_z, rk1Efd_z, rk1Efd_y, rk1chixv, rk1Efd_x, rk1sqEf, rk1optdepthed, rk1CovdBfd_yy, rk1CovdBfd_yx, rk1CovdBfd_yz, rk1optdeptheh, rk1CovdBvf_z, rk1optdepthev, rk1CovdBvf_y, rk1CovdBvf_x, rk1CovdBfd_xx, rk1CovdBfd_xz, rk1CovdBfd_xy, rk1optdepthxd, rk1optdepthxh, rk1optdepthxv, rk1chiav, rk1CovdEfu_zz, rk1CovdEfu_zy, rk1CovdEfu_zx, rk1CovdEfd_zz, rk1CovdEfd_zx, rk1CovdEfd_zy, rk1Bvf, rk1optdepthah, rk1optdepthad, rk1h, rk1chiah, rk1chiad, rk1CovdYef_y, rk1optdepthav, rk1CovdYef_z, rk1CovdYef_x, rk1detgtd, rk1idetgtd, rk1gtu_zz, rk1gtu_yy, rk1gtu_yz, rk1gtu_xx, rk1gtu_xz, rk1CovuBfu_zx, rk1CovuBfu_zz, rk1CovuBfu_zy, rk1Covdvfu_xz, rk1CovuYef_z, rk1Covdvfu_zz, rk1Covdvfu_yz, rk1Covuvfu_zz, rk1Sfu_z, rk1Covurhof_z, rk1CovuBvf_z, rk1gtu_xy, rk1CovuBfu_yz, rk1CovuBfu_yy, rk1CovuBfu_yx, rk1Covdvfu_xy, rk1Covdvfu_xx, rk1CovdinvsqW_x, rk1CovuYef_x, rk1CovuYef_y, rk1Covdvfu_zy, rk1Covdvfu_zx, rk1CovdinvsqW_z, rk1Covdvfu_yx, rk1Covuvfu_zx, rk1Covdvfu_yy, rk1CovdinvsqW_y, rk1CovuinvsqW_z, rk1Covuvfu_zy, rk1CovuBfu_xz, rk1CovuBfu_xy, rk1CovuBfu_xx, rk1Sfu_y, rk1Sfu_x, rk1Covurhof_x, rk1Covurhof_y, rk1Covuvfu_yz, rk1Covuvfu_yy, rk1Covuvfu_yx, rk1Covuvfu_xz, rk1Covuvfu_xy, rk1Covuvfu_xx, rk1vfu_y, rk1vfu_x, rk1CovuBvf_x, rk1CovuBvf_y, rk1CovuinvsqW_y, rk1CovuinvsqW_x, rk1Covddpfdye_z, rk1Covddpfdye_y, rk1Covddpfdye_x, rk1CovdEfu_xz, rk1CovdEfu_xy, rk1CovdEfu_xx, rk1CovdEfd_xx, rk1CovdEfd_xy, rk1CovdEfd_xz, rk1vfu_z, rk1sqvf, rk1sqW, rk1PhiAh, rk1Epf, rk1Thetaf, rk1Covdepsf_z, rk1Covdhf_z, rk1CovdEpf_z, rk1CovdThetaf_z, rk1Covdepsf_y, rk1Covdhf_y, rk1CovdEpf_y, rk1CovdThetaf_y, rk1Covdepsf_x, rk1Covuepsf_z, rk1Covuepsf_y, rk1Covdhf_x, rk1CovdEpf_x, rk1CovdThetaf_x, rk1CovuThetaf_z, rk1CovuThetaf_x, rk1CovuThetaf_y, rk1Phivh_z, rk1Phivh_x, rk1Phivh_y, rk1PhiMh_yz, rk1PhiMh_xz, rk1PhiMh_xy, rk1Covuepsf_x, rk1faceta, rk1CovdEfu_yz, rk1CovuEfu_zz, rk1CovuEfu_xz, rk1CovdEfu_yy, rk1CovuEfu_zy, rk1CovuEfu_xy, rk1CovdEfu_yx, rk1CovdEfd_yy, rk1CovdEfd_yz, rk1CovdEfd_yx, rk1CovuEfu_zx, rk1CovuEfu_xx, rk1CovuEfu_yx, rk1CovuEfu_yz, rk1CovuEfu_yy, rk1PhiThetah, rk1HPres, rk1HTheta, rk1Hv_z, rk1HN_z, rk1HNe_z, rk1Hv_y, rk1HN_y, rk1HM_yz, rk1HNe_y, rk1Hv_x, rk1HN_x, rk1HM_xz, rk1HM_xy, rk1HE_z, rk1HE_x, rk1HE_y, rk1HT_zz, rk1HT_yy, rk1HT_yz, rk1HT_xx, rk1HT_xz, rk1HT_xy, rk1HNe_x, d_TauN_x_o0_t0_m0_l0, d_TauN_y_o0_t0_m0_l0, d_TauN_z_o0_t0_m0_l0, d_TauNe_x_o0_t0_m0_l0, d_TauNe_y_o0_t0_m0_l0, d_TauNe_z_o0_t0_m0_l0, d_TauM_xy_o0_t0_m0_l0, d_TauM_xz_o0_t0_m0_l0, d_TauM_yz_o0_t0_m0_l0, d_TauT_xx_o0_t0_m0_l0, d_TauT_xy_o0_t0_m0_l0, d_TauT_xz_o0_t0_m0_l0, d_TauT_yy_o0_t0_m0_l0, d_TauT_yz_o0_t0_m0_l0, d_TauT_zz_o0_t0_m0_l0, optdepthe_tmp, optdeptha_tmp, optdepthx_tmp, chie_tmp, chia_tmp, chix_tmp, qnu_tmp, rnu_tmp, qnu_e_tmp, qnu_a_tmp, qnu_x_tmp, betabeta, betav, velr, f_redshift;
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					if ((i + 2 < ilast && i - 2 >= 0 && j + 2 < jlast && j - 2 >= 0 && k + 2 < klast && k - 2 >= 0) && ((i + 3 < ilast || !patch->getPatchGeometry()->getTouchesRegularBoundary(0, 1)) && (i - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(0, 0)) && (j + 3 < jlast || !patch->getPatchGeometry()->getTouchesRegularBoundary(1, 1)) && (j - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(1, 0)) && (k + 3 < klast || !patch->getPatchGeometry()->getTouchesRegularBoundary(2, 1)) && (k - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(2, 0)))) {
						d_optdeptheh_o0_t0_m0_l0 = eikonal_i(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthev_o0_t0_m0_l0 = eikonal_j(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthed_o0_t0_m0_l0 = eikonal_k(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthah_o0_t0_m0_l0 = eikonal_i(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthav_o0_t0_m0_l0 = eikonal_j(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthad_o0_t0_m0_l0 = eikonal_k(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthxh_o0_t0_m0_l0 = eikonal_i(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthxv_o0_t0_m0_l0 = eikonal_j(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthxd_o0_t0_m0_l0 = eikonal_k(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chieh_o0_t0_m0_l0 = eikonal_i(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiev_o0_t0_m0_l0 = eikonal_j(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chied_o0_t0_m0_l0 = eikonal_k(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiah_o0_t0_m0_l0 = eikonal_i(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiav_o0_t0_m0_l0 = eikonal_j(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiad_o0_t0_m0_l0 = eikonal_k(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chixh_o0_t0_m0_l0 = eikonal_i(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chixv_o0_t0_m0_l0 = eikonal_j(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chixd_o0_t0_m0_l0 = eikonal_k(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
						if (calculate_les_terms) {
							d_CovdDYf_z_o0_t0_m0_l0 = D1CDO4_k(rk1DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDYf_y_o0_t0_m0_l0 = D1CDO4_j(rk1DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDYf_x_o0_t0_m0_l0 = D1CDO4_i(rk1DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDf_z_o0_t0_m0_l0 = D1CDO4_k(rk1Df, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDf_y_o0_t0_m0_l0 = D1CDO4_j(rk1Df, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDf_x_o0_t0_m0_l0 = D1CDO4_i(rk1Df, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_xx_o0_t0_m0_l0 = D1CDO4_i(rk1Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_xy_o0_t0_m0_l0 = D1CDO4_i(rk1Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_xz_o0_t0_m0_l0 = D1CDO4_i(rk1Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_yx_o0_t0_m0_l0 = D1CDO4_j(rk1Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_yy_o0_t0_m0_l0 = D1CDO4_j(rk1Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_yz_o0_t0_m0_l0 = D1CDO4_j(rk1Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_zx_o0_t0_m0_l0 = D1CDO4_k(rk1Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_zy_o0_t0_m0_l0 = D1CDO4_k(rk1Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_zz_o0_t0_m0_l0 = D1CDO4_k(rk1Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_xx_o0_t0_m0_l0 = D1CDO4_i(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_xy_o0_t0_m0_l0 = D1CDO4_i(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_xz_o0_t0_m0_l0 = D1CDO4_i(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_yx_o0_t0_m0_l0 = D1CDO4_j(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_yy_o0_t0_m0_l0 = D1CDO4_j(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_yz_o0_t0_m0_l0 = D1CDO4_j(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_zx_o0_t0_m0_l0 = D1CDO4_k(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_zy_o0_t0_m0_l0 = D1CDO4_k(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_zz_o0_t0_m0_l0 = D1CDO4_k(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdepsf_x_o0_t0_m0_l0 = D1CDO4_i(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdepsf_y_o0_t0_m0_l0 = D1CDO4_j(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdepsf_z_o0_t0_m0_l0 = D1CDO4_k(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdeps_x_o0_t0_m0_l0 = D1CDO4_i(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdeps_y_o0_t0_m0_l0 = D1CDO4_j(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdeps_z_o0_t0_m0_l0 = D1CDO4_k(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdye_x_o0_t0_m0_l0 = D1CDO4_i(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdye_y_o0_t0_m0_l0 = D1CDO4_j(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdye_z_o0_t0_m0_l0 = D1CDO4_k(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdrho_x_o0_t0_m0_l0 = D1CDO4_i(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdrho_y_o0_t0_m0_l0 = D1CDO4_j(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdrho_z_o0_t0_m0_l0 = D1CDO4_k(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_xx_o0_t0_m0_l0 = D1CDO4_i(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_xy_o0_t0_m0_l0 = D1CDO4_i(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_xz_o0_t0_m0_l0 = D1CDO4_i(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_yx_o0_t0_m0_l0 = D1CDO4_j(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_yy_o0_t0_m0_l0 = D1CDO4_j(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_yz_o0_t0_m0_l0 = D1CDO4_j(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_zx_o0_t0_m0_l0 = D1CDO4_k(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_zy_o0_t0_m0_l0 = D1CDO4_k(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_zz_o0_t0_m0_l0 = D1CDO4_k(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdrhof_x_o0_t0_m0_l0 = D1CDO4_i(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdrhof_y_o0_t0_m0_l0 = D1CDO4_j(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdrhof_z_o0_t0_m0_l0 = D1CDO4_k(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdpf_x_o0_t0_m0_l0 = D1CDO4_i(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdpf_y_o0_t0_m0_l0 = D1CDO4_j(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdpf_z_o0_t0_m0_l0 = D1CDO4_k(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdYef_x_o0_t0_m0_l0 = D1CDO4_i(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdYef_y_o0_t0_m0_l0 = D1CDO4_j(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdYef_z_o0_t0_m0_l0 = D1CDO4_k(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
						}
						rk1chiev = d_chiev_o0_t0_m0_l0;
						rk1chieh = d_chieh_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk1Covdvfd_xx = d_Covdvfd_xx_o0_t0_m0_l0;
							rk1Covdvfd_xz = d_Covdvfd_xz_o0_t0_m0_l0;
							rk1Covdvfd_xy = d_Covdvfd_xy_o0_t0_m0_l0;
						} else {
							rk1Covdvfd_xx = 0.0;
							rk1Covdvfd_xz = 0.0;
							rk1Covdvfd_xy = 0.0;
						}
						rk1chied = d_chied_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk1Covddpfdeps_x = d_Covddpfdeps_x_o0_t0_m0_l0;
							rk1Covddpfdeps_y = d_Covddpfdeps_y_o0_t0_m0_l0;
							rk1Covddpfdeps_z = d_Covddpfdeps_z_o0_t0_m0_l0;
							rk1Covdvfd_zx = d_Covdvfd_zx_o0_t0_m0_l0;
							rk1Covdvfd_zz = d_Covdvfd_zz_o0_t0_m0_l0;
							rk1Covdvfd_zy = d_Covdvfd_zy_o0_t0_m0_l0;
							rk1Covdvfd_yy = d_Covdvfd_yy_o0_t0_m0_l0;
							rk1Covdvfd_yx = d_Covdvfd_yx_o0_t0_m0_l0;
							rk1Covdvfd_yz = d_Covdvfd_yz_o0_t0_m0_l0;
							rk1Covdpf_x = d_Covdpf_x_o0_t0_m0_l0;
							rk1Covdpf_y = d_Covdpf_y_o0_t0_m0_l0;
							rk1Covdpf_z = d_Covdpf_z_o0_t0_m0_l0;
						} else {
							rk1Covddpfdeps_x = 0.0;
							rk1Covddpfdeps_y = 0.0;
							rk1Covddpfdeps_z = 0.0;
							rk1Covdvfd_zx = 0.0;
							rk1Covdvfd_zz = 0.0;
							rk1Covdvfd_zy = 0.0;
							rk1Covdvfd_yy = 0.0;
							rk1Covdvfd_yx = 0.0;
							rk1Covdvfd_yz = 0.0;
							rk1Covdpf_x = 0.0;
							rk1Covdpf_y = 0.0;
							rk1Covdpf_z = 0.0;
						}
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j, k));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_yy, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_zz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_xy, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_xz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1sqBf = vector(rk1Bfu_x, i, j, k) * rk1Bfd_x + vector(rk1Bfu_y, i, j, k) * rk1Bfd_y + vector(rk1Bfu_z, i, j, k) * rk1Bfd_z;
						rk1chixh = d_chixh_o0_t0_m0_l0;
						rk1chixd = d_chixd_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk1Covddpfdrho_z = d_Covddpfdrho_z_o0_t0_m0_l0;
							rk1Covddpfdrho_y = d_Covddpfdrho_y_o0_t0_m0_l0;
							rk1Covddpfdrho_x = d_Covddpfdrho_x_o0_t0_m0_l0;
						} else {
							rk1Covddpfdrho_z = 0.0;
							rk1Covddpfdrho_y = 0.0;
							rk1Covddpfdrho_x = 0.0;
						}
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						if (calculate_les_terms) {
							rk1m_CovdDf_x_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdDf_x_o0_t0_m0_l0;
							rk1CovdDf_x = rk1m_CovdDf_x_o0_t0_l0;
							rk1m_CovdDf_y_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdDf_y_o0_t0_m0_l0;
							rk1CovdDf_y = rk1m_CovdDf_y_o0_t0_l0;
							rk1m_CovdDf_z_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdDf_z_o0_t0_m0_l0;
							rk1CovdDf_z = rk1m_CovdDf_z_o0_t0_l0;
							rk1m_CovdBfu_zz_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdBfu_zz_o0_t0_m0_l0;
							rk1CovdBfu_zz = rk1m_CovdBfu_zz_o0_t0_l0;
							rk1m_CovdBfu_zy_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdBfu_zy_o0_t0_m0_l0;
							rk1CovdBfu_zy = rk1m_CovdBfu_zy_o0_t0_l0;
							rk1m_CovdBfu_zx_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdBfu_zx_o0_t0_m0_l0;
							rk1CovdBfu_zx = rk1m_CovdBfu_zx_o0_t0_l0;
							rk1m_CovdBfu_yz_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdBfu_yz_o0_t0_m0_l0;
							rk1CovdBfu_yz = rk1m_CovdBfu_yz_o0_t0_l0;
							rk1m_CovdBfu_yy_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdBfu_yy_o0_t0_m0_l0;
							rk1CovdBfu_yy = rk1m_CovdBfu_yy_o0_t0_l0;
							rk1m_CovdBfu_yx_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdBfu_yx_o0_t0_m0_l0;
							rk1CovdBfu_yx = rk1m_CovdBfu_yx_o0_t0_l0;
							rk1m_CovdBfu_xy_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdBfu_xy_o0_t0_m0_l0;
							rk1CovdBfu_xy = rk1m_CovdBfu_xy_o0_t0_l0;
							rk1m_CovdBfu_xx_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdBfu_xx_o0_t0_m0_l0;
							rk1CovdBfu_xx = rk1m_CovdBfu_xx_o0_t0_l0;
							rk1m_CovdBfu_xz_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdBfu_xz_o0_t0_m0_l0;
							rk1CovdBfu_xz = rk1m_CovdBfu_xz_o0_t0_l0;
							rk1CovdBfd_zx = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * rk1CovdBfu_zx + vector(rk1gtd_xy, i, j, k) * rk1CovdBfu_zy + vector(rk1gtd_xz, i, j, k) * rk1CovdBfu_zz);
							rk1Covdrhof_z = d_Covdrhof_z_o0_t0_m0_l0;
							rk1Covdrhof_y = d_Covdrhof_y_o0_t0_m0_l0;
							rk1Covdrhof_x = d_Covdrhof_x_o0_t0_m0_l0;
							rk1CovdBfd_zz = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * rk1CovdBfu_zx + vector(rk1gtd_yz, i, j, k) * rk1CovdBfu_zy + vector(rk1gtd_zz, i, j, k) * rk1CovdBfu_zz);
							rk1CovdBfd_zy = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * rk1CovdBfu_zx + vector(rk1gtd_yy, i, j, k) * rk1CovdBfu_zy + vector(rk1gtd_yz, i, j, k) * rk1CovdBfu_zz);
							rk1m_CovdDYf_x_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdDYf_x_o0_t0_m0_l0;
							rk1CovdDYf_x = rk1m_CovdDYf_x_o0_t0_l0;
							rk1m_CovdDYf_y_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdDYf_y_o0_t0_m0_l0;
							rk1CovdDYf_y = rk1m_CovdDYf_y_o0_t0_l0;
							rk1m_CovdDYf_z_o0_t0_l0 = 1.0 / rk1sdetg * d_CovdDYf_z_o0_t0_m0_l0;
							rk1CovdDYf_z = rk1m_CovdDYf_z_o0_t0_l0;
						} else {
							rk1CovdDf_x = 0.0;
							rk1CovdDf_y = 0.0;
							rk1CovdDf_z = 0.0;
							rk1CovdBfu_zz = 0.0;
							rk1CovdBfu_zy = 0.0;
							rk1CovdBfu_zx = 0.0;
							rk1CovdBfu_yz = 0.0;
							rk1CovdBfu_yy = 0.0;
							rk1CovdBfu_yx = 0.0;
							rk1CovdBfu_xy = 0.0;
							rk1CovdBfu_xx = 0.0;
							rk1CovdBfu_xz = 0.0;
							rk1CovdBfd_zx = 0.0;
							rk1Covdrhof_z = 0.0;
							rk1Covdrhof_y = 0.0;
							rk1Covdrhof_x = 0.0;
							rk1CovdBfd_zz = 0.0;
							rk1CovdBfd_zy = 0.0;
							rk1CovdDYf_x = 0.0;
							rk1CovdDYf_y = 0.0;
							rk1CovdDYf_z = 0.0;
						}
						rk1Efd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * vector(Efu_x, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(Efu_y, i, j, k) + vector(rk1gtd_zz, i, j, k) * vector(Efu_z, i, j, k));
						rk1Efd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * vector(Efu_x, i, j, k) + vector(rk1gtd_yy, i, j, k) * vector(Efu_y, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(Efu_z, i, j, k));
						rk1chixv = d_chixv_o0_t0_m0_l0;
						rk1Efd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * vector(Efu_x, i, j, k) + vector(rk1gtd_xy, i, j, k) * vector(Efu_y, i, j, k) + vector(rk1gtd_xz, i, j, k) * vector(Efu_z, i, j, k));
						rk1sqEf = vector(Efu_x, i, j, k) * rk1Efd_x + vector(Efu_y, i, j, k) * rk1Efd_y + vector(Efu_z, i, j, k) * rk1Efd_z;
						rk1optdepthed = d_optdepthed_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk1CovdBfd_yy = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * rk1CovdBfu_yx + vector(rk1gtd_yy, i, j, k) * rk1CovdBfu_yy + vector(rk1gtd_yz, i, j, k) * rk1CovdBfu_yz);
							rk1CovdBfd_yx = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * rk1CovdBfu_yx + vector(rk1gtd_xy, i, j, k) * rk1CovdBfu_yy + vector(rk1gtd_xz, i, j, k) * rk1CovdBfu_yz);
							rk1CovdBfd_yz = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * rk1CovdBfu_yx + vector(rk1gtd_yz, i, j, k) * rk1CovdBfu_yy + vector(rk1gtd_zz, i, j, k) * rk1CovdBfu_yz);
						} else {
							rk1CovdBfd_yy = 0.0;
							rk1CovdBfd_yx = 0.0;
							rk1CovdBfd_yz = 0.0;
						}
						rk1optdeptheh = d_optdeptheh_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk1CovdBvf_z = vector(rk1Bfu_x, i, j, k) * rk1Covdvfd_zx + vector(vfd_x, i, j, k) * rk1CovdBfu_zx + vector(rk1Bfu_y, i, j, k) * rk1Covdvfd_zy + vector(vfd_y, i, j, k) * rk1CovdBfu_zy + vector(rk1Bfu_z, i, j, k) * rk1Covdvfd_zz + vector(vfd_z, i, j, k) * rk1CovdBfu_zz;
						} else {
							rk1CovdBvf_z = 0.0;
						}
						rk1optdepthev = d_optdepthev_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk1CovdBvf_y = vector(rk1Bfu_x, i, j, k) * rk1Covdvfd_yx + vector(vfd_x, i, j, k) * rk1CovdBfu_yx + vector(rk1Bfu_y, i, j, k) * rk1Covdvfd_yy + vector(vfd_y, i, j, k) * rk1CovdBfu_yy + vector(rk1Bfu_z, i, j, k) * rk1Covdvfd_yz + vector(vfd_z, i, j, k) * rk1CovdBfu_yz;
							rk1CovdBvf_x = vector(rk1Bfu_x, i, j, k) * rk1Covdvfd_xx + vector(vfd_x, i, j, k) * rk1CovdBfu_xx + vector(rk1Bfu_y, i, j, k) * rk1Covdvfd_xy + vector(vfd_y, i, j, k) * rk1CovdBfu_xy + vector(rk1Bfu_z, i, j, k) * rk1Covdvfd_xz + vector(vfd_z, i, j, k) * rk1CovdBfu_xz;
							rk1CovdBfd_xx = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * rk1CovdBfu_xx + vector(rk1gtd_xy, i, j, k) * rk1CovdBfu_xy + vector(rk1gtd_xz, i, j, k) * rk1CovdBfu_xz);
							rk1CovdBfd_xz = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * rk1CovdBfu_xx + vector(rk1gtd_yz, i, j, k) * rk1CovdBfu_xy + vector(rk1gtd_zz, i, j, k) * rk1CovdBfu_xz);
							rk1CovdBfd_xy = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * rk1CovdBfu_xx + vector(rk1gtd_yy, i, j, k) * rk1CovdBfu_xy + vector(rk1gtd_yz, i, j, k) * rk1CovdBfu_xz);
						} else {
							rk1CovdBvf_y = 0.0;
							rk1CovdBvf_x = 0.0;
							rk1CovdBfd_xx = 0.0;
							rk1CovdBfd_xz = 0.0;
							rk1CovdBfd_xy = 0.0;
						}
						rk1optdepthxd = d_optdepthxd_o0_t0_m0_l0;
						rk1optdepthxh = d_optdepthxh_o0_t0_m0_l0;
						rk1optdepthxv = d_optdepthxv_o0_t0_m0_l0;
						rk1chiav = d_chiav_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk1CovdEfu_zz = d_CovdEfu_zz_o0_t0_m0_l0;
							rk1CovdEfu_zy = d_CovdEfu_zy_o0_t0_m0_l0;
							rk1CovdEfu_zx = d_CovdEfu_zx_o0_t0_m0_l0;
							rk1CovdEfd_zz = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * rk1CovdEfu_zx + vector(rk1gtd_yz, i, j, k) * rk1CovdEfu_zy + vector(rk1gtd_zz, i, j, k) * rk1CovdEfu_zz);
							rk1CovdEfd_zx = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * rk1CovdEfu_zx + vector(rk1gtd_xy, i, j, k) * rk1CovdEfu_zy + vector(rk1gtd_xz, i, j, k) * rk1CovdEfu_zz);
							rk1CovdEfd_zy = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * rk1CovdEfu_zx + vector(rk1gtd_yy, i, j, k) * rk1CovdEfu_zy + vector(rk1gtd_yz, i, j, k) * rk1CovdEfu_zz);
						} else {
							rk1CovdEfu_zz = 0.0;
							rk1CovdEfu_zy = 0.0;
							rk1CovdEfu_zx = 0.0;
							rk1CovdEfd_zz = 0.0;
							rk1CovdEfd_zx = 0.0;
							rk1CovdEfd_zy = 0.0;
						}
						rk1Bvf = (vector(rk1Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk1Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk1Bfu_z, i, j, k) * vector(vfd_z, i, j, k)) / rk1sdetg;
						rk1optdepthah = d_optdepthah_o0_t0_m0_l0;
						rk1optdepthad = d_optdepthad_o0_t0_m0_l0;
						rk1h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						rk1chiah = d_chiah_o0_t0_m0_l0;
						rk1chiad = d_chiad_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk1CovdYef_y = d_CovdYef_y_o0_t0_m0_l0;
						} else {
							rk1CovdYef_y = 0.0;
						}
						rk1optdepthav = d_optdepthav_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk1CovdYef_z = d_CovdYef_z_o0_t0_m0_l0;
							rk1CovdYef_x = d_CovdYef_x_o0_t0_m0_l0;
						} else {
							rk1CovdYef_z = 0.0;
							rk1CovdYef_x = 0.0;
						}
						rk1detgtd = vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) * vector(rk1gtd_xx, i, j, k) + (-vector(rk1gtd_yy, i, j, k) * (vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k))) + (-vector(rk1gtd_zz, i, j, k) * (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k))) + (-(vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k)) * vector(rk1gtd_xx, i, j, k)) + 2.0 * vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yy, i, j, k) - vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k));
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_zz, i, j, k) - vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yz, i, j, k)) + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) - vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_yz, i, j, k) - vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_xz, i, j, k));
						if (calculate_les_terms) {
							rk1CovuBfu_zx = rk1chi_max * (rk1gtu_xz * rk1CovdBfu_xx + rk1gtu_yz * rk1CovdBfu_yx + rk1gtu_zz * rk1CovdBfu_zx);
							rk1CovuBfu_zz = rk1chi_max * (rk1gtu_xz * rk1CovdBfu_xz + rk1gtu_yz * rk1CovdBfu_yz + rk1gtu_zz * rk1CovdBfu_zz);
							rk1CovuBfu_zy = rk1chi_max * (rk1gtu_xz * rk1CovdBfu_xy + rk1gtu_yz * rk1CovdBfu_yy + rk1gtu_zz * rk1CovdBfu_zy);
							rk1Covdvfu_xz = rk1chi_max * (rk1gtu_xz * rk1Covdvfd_xx + rk1gtu_yz * rk1Covdvfd_xy + rk1gtu_zz * rk1Covdvfd_xz);
							rk1CovuYef_z = rk1chi_max * (rk1gtu_xz * rk1CovdYef_x + rk1gtu_yz * rk1CovdYef_y + rk1gtu_zz * rk1CovdYef_z);
							rk1Covdvfu_zz = rk1chi_max * (rk1gtu_xz * rk1Covdvfd_zx + rk1gtu_yz * rk1Covdvfd_zy + rk1gtu_zz * rk1Covdvfd_zz);
							rk1Covdvfu_yz = rk1chi_max * (rk1gtu_xz * rk1Covdvfd_yx + rk1gtu_yz * rk1Covdvfd_yy + rk1gtu_zz * rk1Covdvfd_yz);
							rk1Covuvfu_zz = rk1chi_max * (rk1gtu_xz * rk1Covdvfu_xz + rk1gtu_yz * rk1Covdvfu_yz + rk1gtu_zz * rk1Covdvfu_zz);
						} else {
							rk1CovuBfu_zx = 0.0;
							rk1CovuBfu_zz = 0.0;
							rk1CovuBfu_zy = 0.0;
							rk1Covdvfu_xz = 0.0;
							rk1CovuYef_z = 0.0;
							rk1Covdvfu_zz = 0.0;
							rk1Covdvfu_yz = 0.0;
							rk1Covuvfu_zz = 0.0;
						}
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j, k) + rk1gtu_yz * vector(rk1Sfd_y, i, j, k) + rk1gtu_zz * vector(rk1Sfd_z, i, j, k));
						if (calculate_les_terms) {
							rk1Covurhof_z = rk1chi_max * (rk1gtu_xz * rk1Covdrhof_x + rk1gtu_yz * rk1Covdrhof_y + rk1gtu_zz * rk1Covdrhof_z);
							rk1CovuBvf_z = rk1chi_max * (rk1gtu_xz * rk1CovdBvf_x + rk1gtu_yz * rk1CovdBvf_y + rk1gtu_zz * rk1CovdBvf_z);
						} else {
							rk1Covurhof_z = 0.0;
							rk1CovuBvf_z = 0.0;
						}
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_zz, i, j, k)) + vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xz, i, j, k));
						if (calculate_les_terms) {
							rk1CovuBfu_yz = rk1chi_max * (rk1gtu_xy * rk1CovdBfu_xz + rk1gtu_yy * rk1CovdBfu_yz + rk1gtu_yz * rk1CovdBfu_zz);
							rk1CovuBfu_yy = rk1chi_max * (rk1gtu_xy * rk1CovdBfu_xy + rk1gtu_yy * rk1CovdBfu_yy + rk1gtu_yz * rk1CovdBfu_zy);
							rk1CovuBfu_yx = rk1chi_max * (rk1gtu_xy * rk1CovdBfu_xx + rk1gtu_yy * rk1CovdBfu_yx + rk1gtu_yz * rk1CovdBfu_zx);
							rk1Covdvfu_xy = rk1chi_max * (rk1gtu_xy * rk1Covdvfd_xx + rk1gtu_yy * rk1Covdvfd_xy + rk1gtu_yz * rk1Covdvfd_xz);
							rk1Covdvfu_xx = rk1chi_max * (rk1gtu_xx * rk1Covdvfd_xx + rk1gtu_xy * rk1Covdvfd_xy + rk1gtu_xz * rk1Covdvfd_xz);
							rk1CovdinvsqW_x = (-2.0 * vector(vfd_x, i, j, k) * rk1Covdvfu_xx) + (-2.0 * vector(vfd_y, i, j, k) * rk1Covdvfu_xy) + (-2.0 * vector(vfd_z, i, j, k) * rk1Covdvfu_xz);
							rk1CovuYef_x = rk1chi_max * (rk1gtu_xx * rk1CovdYef_x + rk1gtu_xy * rk1CovdYef_y + rk1gtu_xz * rk1CovdYef_z);
							rk1CovuYef_y = rk1chi_max * (rk1gtu_xy * rk1CovdYef_x + rk1gtu_yy * rk1CovdYef_y + rk1gtu_yz * rk1CovdYef_z);
							rk1Covdvfu_zy = rk1chi_max * (rk1gtu_xy * rk1Covdvfd_zx + rk1gtu_yy * rk1Covdvfd_zy + rk1gtu_yz * rk1Covdvfd_zz);
							rk1Covdvfu_zx = rk1chi_max * (rk1gtu_xx * rk1Covdvfd_zx + rk1gtu_xy * rk1Covdvfd_zy + rk1gtu_xz * rk1Covdvfd_zz);
							rk1CovdinvsqW_z = (-2.0 * vector(vfd_x, i, j, k) * rk1Covdvfu_zx) + (-2.0 * vector(vfd_y, i, j, k) * rk1Covdvfu_zy) + (-2.0 * vector(vfd_z, i, j, k) * rk1Covdvfu_zz);
							rk1Covdvfu_yx = rk1chi_max * (rk1gtu_xx * rk1Covdvfd_yx + rk1gtu_xy * rk1Covdvfd_yy + rk1gtu_xz * rk1Covdvfd_yz);
							rk1Covuvfu_zx = rk1chi_max * (rk1gtu_xz * rk1Covdvfu_xx + rk1gtu_yz * rk1Covdvfu_yx + rk1gtu_zz * rk1Covdvfu_zx);
							rk1Covdvfu_yy = rk1chi_max * (rk1gtu_xy * rk1Covdvfd_yx + rk1gtu_yy * rk1Covdvfd_yy + rk1gtu_yz * rk1Covdvfd_yz);
							rk1CovdinvsqW_y = (-2.0 * vector(vfd_x, i, j, k) * rk1Covdvfu_yx) + (-2.0 * vector(vfd_y, i, j, k) * rk1Covdvfu_yy) + (-2.0 * vector(vfd_z, i, j, k) * rk1Covdvfu_yz);
							rk1CovuinvsqW_z = rk1chi_max * (rk1gtu_xz * rk1CovdinvsqW_x + rk1gtu_yz * rk1CovdinvsqW_y + rk1gtu_zz * rk1CovdinvsqW_z);
							rk1Covuvfu_zy = rk1chi_max * (rk1gtu_xz * rk1Covdvfu_xy + rk1gtu_yz * rk1Covdvfu_yy + rk1gtu_zz * rk1Covdvfu_zy);
							rk1CovuBfu_xz = rk1chi_max * (rk1gtu_xx * rk1CovdBfu_xz + rk1gtu_xy * rk1CovdBfu_yz + rk1gtu_xz * rk1CovdBfu_zz);
							rk1CovuBfu_xy = rk1chi_max * (rk1gtu_xx * rk1CovdBfu_xy + rk1gtu_xy * rk1CovdBfu_yy + rk1gtu_xz * rk1CovdBfu_zy);
							rk1CovuBfu_xx = rk1chi_max * (rk1gtu_xx * rk1CovdBfu_xx + rk1gtu_xy * rk1CovdBfu_yx + rk1gtu_xz * rk1CovdBfu_zx);
						} else {
							rk1CovuBfu_yz = 0.0;
							rk1CovuBfu_yy = 0.0;
							rk1CovuBfu_yx = 0.0;
							rk1Covdvfu_xy = 0.0;
							rk1Covdvfu_xx = 0.0;
							rk1CovdinvsqW_x = 0.0;
							rk1CovuYef_x = 0.0;
							rk1CovuYef_y = 0.0;
							rk1Covdvfu_zy = 0.0;
							rk1Covdvfu_zx = 0.0;
							rk1CovdinvsqW_z = 0.0;
							rk1Covdvfu_yx = 0.0;
							rk1Covuvfu_zx = 0.0;
							rk1Covdvfu_yy = 0.0;
							rk1CovdinvsqW_y = 0.0;
							rk1CovuinvsqW_z = 0.0;
							rk1Covuvfu_zy = 0.0;
							rk1CovuBfu_xz = 0.0;
							rk1CovuBfu_xy = 0.0;
							rk1CovuBfu_xx = 0.0;
						}
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j, k) + rk1gtu_yy * vector(rk1Sfd_y, i, j, k) + rk1gtu_yz * vector(rk1Sfd_z, i, j, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j, k) + rk1gtu_xy * vector(rk1Sfd_y, i, j, k) + rk1gtu_xz * vector(rk1Sfd_z, i, j, k));
						if (calculate_les_terms) {
							rk1Covurhof_x = rk1chi_max * (rk1gtu_xx * rk1Covdrhof_x + rk1gtu_xy * rk1Covdrhof_y + rk1gtu_xz * rk1Covdrhof_z);
							rk1Covurhof_y = rk1chi_max * (rk1gtu_xy * rk1Covdrhof_x + rk1gtu_yy * rk1Covdrhof_y + rk1gtu_yz * rk1Covdrhof_z);
							rk1Covuvfu_yz = rk1chi_max * (rk1gtu_xy * rk1Covdvfu_xz + rk1gtu_yy * rk1Covdvfu_yz + rk1gtu_yz * rk1Covdvfu_zz);
							rk1Covuvfu_yy = rk1chi_max * (rk1gtu_xy * rk1Covdvfu_xy + rk1gtu_yy * rk1Covdvfu_yy + rk1gtu_yz * rk1Covdvfu_zy);
							rk1Covuvfu_yx = rk1chi_max * (rk1gtu_xy * rk1Covdvfu_xx + rk1gtu_yy * rk1Covdvfu_yx + rk1gtu_yz * rk1Covdvfu_zx);
							rk1Covuvfu_xz = rk1chi_max * (rk1gtu_xx * rk1Covdvfu_xz + rk1gtu_xy * rk1Covdvfu_yz + rk1gtu_xz * rk1Covdvfu_zz);
							rk1Covuvfu_xy = rk1chi_max * (rk1gtu_xx * rk1Covdvfu_xy + rk1gtu_xy * rk1Covdvfu_yy + rk1gtu_xz * rk1Covdvfu_zy);
							rk1Covuvfu_xx = rk1chi_max * (rk1gtu_xx * rk1Covdvfu_xx + rk1gtu_xy * rk1Covdvfu_yx + rk1gtu_xz * rk1Covdvfu_zx);
						} else {
							rk1Covurhof_x = 0.0;
							rk1Covurhof_y = 0.0;
							rk1Covuvfu_yz = 0.0;
							rk1Covuvfu_yy = 0.0;
							rk1Covuvfu_yx = 0.0;
							rk1Covuvfu_xz = 0.0;
							rk1Covuvfu_xy = 0.0;
							rk1Covuvfu_xx = 0.0;
						}
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j, k) + rk1gtu_yy * vector(vfd_y, i, j, k) + rk1gtu_yz * vector(vfd_z, i, j, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j, k) + rk1gtu_xy * vector(vfd_y, i, j, k) + rk1gtu_xz * vector(vfd_z, i, j, k));
						if (calculate_les_terms) {
							rk1CovuBvf_x = rk1chi_max * (rk1gtu_xx * rk1CovdBvf_x + rk1gtu_xy * rk1CovdBvf_y + rk1gtu_xz * rk1CovdBvf_z);
							rk1CovuBvf_y = rk1chi_max * (rk1gtu_xy * rk1CovdBvf_x + rk1gtu_yy * rk1CovdBvf_y + rk1gtu_yz * rk1CovdBvf_z);
							rk1CovuinvsqW_y = rk1chi_max * (rk1gtu_xy * rk1CovdinvsqW_x + rk1gtu_yy * rk1CovdinvsqW_y + rk1gtu_yz * rk1CovdinvsqW_z);
							rk1CovuinvsqW_x = rk1chi_max * (rk1gtu_xx * rk1CovdinvsqW_x + rk1gtu_xy * rk1CovdinvsqW_y + rk1gtu_xz * rk1CovdinvsqW_z);
							rk1Covddpfdye_z = d_Covddpfdye_z_o0_t0_m0_l0;
							rk1Covddpfdye_y = d_Covddpfdye_y_o0_t0_m0_l0;
							rk1Covddpfdye_x = d_Covddpfdye_x_o0_t0_m0_l0;
							rk1CovdEfu_xz = d_CovdEfu_xz_o0_t0_m0_l0;
							rk1CovdEfu_xy = d_CovdEfu_xy_o0_t0_m0_l0;
							rk1CovdEfu_xx = d_CovdEfu_xx_o0_t0_m0_l0;
							rk1CovdEfd_xx = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * rk1CovdEfu_xx + vector(rk1gtd_xy, i, j, k) * rk1CovdEfu_xy + vector(rk1gtd_xz, i, j, k) * rk1CovdEfu_xz);
							rk1CovdEfd_xy = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * rk1CovdEfu_xx + vector(rk1gtd_yy, i, j, k) * rk1CovdEfu_xy + vector(rk1gtd_yz, i, j, k) * rk1CovdEfu_xz);
							rk1CovdEfd_xz = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * rk1CovdEfu_xx + vector(rk1gtd_yz, i, j, k) * rk1CovdEfu_xy + vector(rk1gtd_zz, i, j, k) * rk1CovdEfu_xz);
						} else {
							rk1CovuBvf_x = 0.0;
							rk1CovuBvf_y = 0.0;
							rk1CovuinvsqW_y = 0.0;
							rk1CovuinvsqW_x = 0.0;
							rk1Covddpfdye_z = 0.0;
							rk1Covddpfdye_y = 0.0;
							rk1Covddpfdye_x = 0.0;
							rk1CovdEfu_xz = 0.0;
							rk1CovdEfu_xy = 0.0;
							rk1CovdEfu_xx = 0.0;
							rk1CovdEfd_xx = 0.0;
							rk1CovdEfd_xy = 0.0;
							rk1CovdEfd_xz = 0.0;
						}
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j, k) + rk1gtu_yz * vector(vfd_y, i, j, k) + rk1gtu_zz * vector(vfd_z, i, j, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j, k) + rk1vfu_y * vector(vfd_y, i, j, k) + rk1vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						if (calculate_les_terms) {
							rk1PhiAh = rk1sqW * (vector(pf, i, j, k) * vector(dpfdeps, i, j, k) + (vector(rhof, i, j, k) * vector(rhof, i, j, k)) * vector(dpfdrho, i, j, k));
						} else {
							rk1PhiAh = 0.0;
						}
						rk1Epf = rk1h * rk1sqW;
						rk1Thetaf = rk1Epf + rk1sqBf / (rk1sdetg * rk1sdetg);
						if (calculate_les_terms) {
							rk1Covdepsf_z = d_Covdepsf_z_o0_t0_m0_l0;
							rk1Covdhf_z = (1.0 + vector(epsf, i, j, k)) * rk1Covdrhof_z + vector(rhof, i, j, k) * rk1Covdepsf_z + rk1Covdpf_z;
							rk1CovdEpf_z = rk1sqW * rk1Covdhf_z + (-rk1h * (rk1sqW * rk1sqW) * rk1CovdinvsqW_z);
							rk1CovdThetaf_z = rk1CovdEpf_z + 2.0 * (rk1Bfd_x * rk1CovdBfu_zx) / rk1sdetg + 2.0 * (rk1Bfd_y * rk1CovdBfu_zy) / rk1sdetg + 2.0 * (rk1Bfd_z * rk1CovdBfu_zz) / rk1sdetg;
							rk1Covdepsf_y = d_Covdepsf_y_o0_t0_m0_l0;
							rk1Covdhf_y = (1.0 + vector(epsf, i, j, k)) * rk1Covdrhof_y + vector(rhof, i, j, k) * rk1Covdepsf_y + rk1Covdpf_y;
							rk1CovdEpf_y = rk1sqW * rk1Covdhf_y + (-rk1h * (rk1sqW * rk1sqW) * rk1CovdinvsqW_y);
							rk1CovdThetaf_y = rk1CovdEpf_y + 2.0 * (rk1Bfd_x * rk1CovdBfu_yx) / rk1sdetg + 2.0 * (rk1Bfd_y * rk1CovdBfu_yy) / rk1sdetg + 2.0 * (rk1Bfd_z * rk1CovdBfu_yz) / rk1sdetg;
							rk1Covdepsf_x = d_Covdepsf_x_o0_t0_m0_l0;
							rk1Covuepsf_z = rk1chi_max * (rk1gtu_xz * rk1Covdepsf_x + rk1gtu_yz * rk1Covdepsf_y + rk1gtu_zz * rk1Covdepsf_z);
							rk1Covuepsf_y = rk1chi_max * (rk1gtu_xy * rk1Covdepsf_x + rk1gtu_yy * rk1Covdepsf_y + rk1gtu_yz * rk1Covdepsf_z);
							rk1Covdhf_x = (1.0 + vector(epsf, i, j, k)) * rk1Covdrhof_x + vector(rhof, i, j, k) * rk1Covdepsf_x + rk1Covdpf_x;
							rk1CovdEpf_x = rk1sqW * rk1Covdhf_x + (-rk1h * (rk1sqW * rk1sqW) * rk1CovdinvsqW_x);
							rk1CovdThetaf_x = rk1CovdEpf_x + 2.0 * (rk1Bfd_x * rk1CovdBfu_xx) / rk1sdetg + 2.0 * (rk1Bfd_y * rk1CovdBfu_xy) / rk1sdetg + 2.0 * (rk1Bfd_z * rk1CovdBfu_xz) / rk1sdetg;
							rk1CovuThetaf_z = rk1chi_max * (rk1gtu_xz * rk1CovdThetaf_x + rk1gtu_yz * rk1CovdThetaf_y + rk1gtu_zz * rk1CovdThetaf_z);
							rk1CovuThetaf_x = rk1chi_max * (rk1gtu_xx * rk1CovdThetaf_x + rk1gtu_xy * rk1CovdThetaf_y + rk1gtu_xz * rk1CovdThetaf_z);
							rk1CovuThetaf_y = rk1chi_max * (rk1gtu_xy * rk1CovdThetaf_x + rk1gtu_yy * rk1CovdThetaf_y + rk1gtu_yz * rk1CovdThetaf_z);
							rk1Phivh_z = 2.0 * (rk1CovuBvf_x * rk1CovdBfu_xz + rk1CovuBvf_y * rk1CovdBfu_yz + rk1CovuBvf_z * rk1CovdBfu_zz + (-rk1CovuThetaf_x * rk1Covdvfu_xz) + (-rk1CovuThetaf_y * rk1Covdvfu_yz) + (-rk1CovuThetaf_z * rk1Covdvfu_zz) + (vector(rk1Bfu_z, i, j, k) * (rk1Thetaf * (rk1CovuBfu_xx * rk1Covdvfd_xx + rk1CovuBfu_yx * rk1Covdvfd_yx + rk1CovuBfu_zx * rk1Covdvfd_zx + rk1CovuBfu_xy * rk1Covdvfd_xy + rk1CovuBfu_yy * rk1Covdvfd_yy + rk1CovuBfu_zy * rk1Covdvfd_zy + rk1CovuBfu_xz * rk1Covdvfd_xz + rk1CovuBfu_yz * rk1Covdvfd_yz + rk1CovuBfu_zz * rk1Covdvfd_zz) + (rk1Bfd_x * rk1CovdBfu_xx * rk1CovuBvf_x) / rk1sdetg + (rk1Bfd_y * rk1CovdBfu_xy * rk1CovuBvf_x) / rk1sdetg + (rk1Bfd_z * rk1CovdBfu_xz * rk1CovuBvf_x) / rk1sdetg + (rk1Bfd_x * rk1CovdBfu_yx * rk1CovuBvf_y) / rk1sdetg + (rk1Bfd_y * rk1CovdBfu_yy * rk1CovuBvf_y) / rk1sdetg + (rk1Bfd_z * rk1CovdBfu_yz * rk1CovuBvf_y) / rk1sdetg + (rk1Bfd_x * rk1CovdBfu_zx * rk1CovuBvf_z) / rk1sdetg + (rk1Bfd_y * rk1CovdBfu_zy * rk1CovuBvf_z) / rk1sdetg + (rk1Bfd_z * rk1CovdBfu_zz * rk1CovuBvf_z) / rk1sdetg + (-(vector(rk1Bfu_x, i, j, k) * rk1Covdvfd_xx * rk1CovuThetaf_x) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1Covdvfd_xy * rk1CovuThetaf_x) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfd_xz * rk1CovuThetaf_x) / rk1sdetg) + (-(vector(rk1Bfu_x, i, j, k) * rk1Covdvfd_yx * rk1CovuThetaf_y) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1Covdvfd_yy * rk1CovuThetaf_y) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfd_yz * rk1CovuThetaf_y) / rk1sdetg) + (-(vector(rk1Bfu_x, i, j, k) * rk1Covdvfd_zx * rk1CovuThetaf_z) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1Covdvfd_zy * rk1CovuThetaf_z) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfd_zz * rk1CovuThetaf_z) / rk1sdetg))) / (rk1sdetg * rk1Epf)) / rk1Thetaf;
							rk1Phivh_x = 2.0 * (rk1CovuBvf_x * rk1CovdBfu_xx + rk1CovuBvf_y * rk1CovdBfu_yx + rk1CovuBvf_z * rk1CovdBfu_zx + (-rk1CovuThetaf_x * rk1Covdvfu_xx) + (-rk1CovuThetaf_y * rk1Covdvfu_yx) + (-rk1CovuThetaf_z * rk1Covdvfu_zx) + (vector(rk1Bfu_x, i, j, k) * (rk1Thetaf * (rk1CovuBfu_xx * rk1Covdvfd_xx + rk1CovuBfu_yx * rk1Covdvfd_yx + rk1CovuBfu_zx * rk1Covdvfd_zx + rk1CovuBfu_xy * rk1Covdvfd_xy + rk1CovuBfu_yy * rk1Covdvfd_yy + rk1CovuBfu_zy * rk1Covdvfd_zy + rk1CovuBfu_xz * rk1Covdvfd_xz + rk1CovuBfu_yz * rk1Covdvfd_yz + rk1CovuBfu_zz * rk1Covdvfd_zz) + (rk1Bfd_x * rk1CovdBfu_xx * rk1CovuBvf_x) / rk1sdetg + (rk1Bfd_y * rk1CovdBfu_xy * rk1CovuBvf_x) / rk1sdetg + (rk1Bfd_z * rk1CovdBfu_xz * rk1CovuBvf_x) / rk1sdetg + (rk1Bfd_x * rk1CovdBfu_yx * rk1CovuBvf_y) / rk1sdetg + (rk1Bfd_y * rk1CovdBfu_yy * rk1CovuBvf_y) / rk1sdetg + (rk1Bfd_z * rk1CovdBfu_yz * rk1CovuBvf_y) / rk1sdetg + (rk1Bfd_x * rk1CovdBfu_zx * rk1CovuBvf_z) / rk1sdetg + (rk1Bfd_y * rk1CovdBfu_zy * rk1CovuBvf_z) / rk1sdetg + (rk1Bfd_z * rk1CovdBfu_zz * rk1CovuBvf_z) / rk1sdetg + (-(vector(rk1Bfu_x, i, j, k) * rk1Covdvfd_xx * rk1CovuThetaf_x) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1Covdvfd_xy * rk1CovuThetaf_x) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfd_xz * rk1CovuThetaf_x) / rk1sdetg) + (-(vector(rk1Bfu_x, i, j, k) * rk1Covdvfd_yx * rk1CovuThetaf_y) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1Covdvfd_yy * rk1CovuThetaf_y) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfd_yz * rk1CovuThetaf_y) / rk1sdetg) + (-(vector(rk1Bfu_x, i, j, k) * rk1Covdvfd_zx * rk1CovuThetaf_z) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1Covdvfd_zy * rk1CovuThetaf_z) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfd_zz * rk1CovuThetaf_z) / rk1sdetg))) / (rk1sdetg * rk1Epf)) / rk1Thetaf;
							rk1Phivh_y = 2.0 * (rk1CovuBvf_x * rk1CovdBfu_xy + rk1CovuBvf_y * rk1CovdBfu_yy + rk1CovuBvf_z * rk1CovdBfu_zy + (-rk1CovuThetaf_x * rk1Covdvfu_xy) + (-rk1CovuThetaf_y * rk1Covdvfu_yy) + (-rk1CovuThetaf_z * rk1Covdvfu_zy) + (vector(rk1Bfu_y, i, j, k) * (rk1Thetaf * (rk1CovuBfu_xx * rk1Covdvfd_xx + rk1CovuBfu_yx * rk1Covdvfd_yx + rk1CovuBfu_zx * rk1Covdvfd_zx + rk1CovuBfu_xy * rk1Covdvfd_xy + rk1CovuBfu_yy * rk1Covdvfd_yy + rk1CovuBfu_zy * rk1Covdvfd_zy + rk1CovuBfu_xz * rk1Covdvfd_xz + rk1CovuBfu_yz * rk1Covdvfd_yz + rk1CovuBfu_zz * rk1Covdvfd_zz) + (rk1Bfd_x * rk1CovdBfu_xx * rk1CovuBvf_x) / rk1sdetg + (rk1Bfd_y * rk1CovdBfu_xy * rk1CovuBvf_x) / rk1sdetg + (rk1Bfd_z * rk1CovdBfu_xz * rk1CovuBvf_x) / rk1sdetg + (rk1Bfd_x * rk1CovdBfu_yx * rk1CovuBvf_y) / rk1sdetg + (rk1Bfd_y * rk1CovdBfu_yy * rk1CovuBvf_y) / rk1sdetg + (rk1Bfd_z * rk1CovdBfu_yz * rk1CovuBvf_y) / rk1sdetg + (rk1Bfd_x * rk1CovdBfu_zx * rk1CovuBvf_z) / rk1sdetg + (rk1Bfd_y * rk1CovdBfu_zy * rk1CovuBvf_z) / rk1sdetg + (rk1Bfd_z * rk1CovdBfu_zz * rk1CovuBvf_z) / rk1sdetg + (-(vector(rk1Bfu_x, i, j, k) * rk1Covdvfd_xx * rk1CovuThetaf_x) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1Covdvfd_xy * rk1CovuThetaf_x) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfd_xz * rk1CovuThetaf_x) / rk1sdetg) + (-(vector(rk1Bfu_x, i, j, k) * rk1Covdvfd_yx * rk1CovuThetaf_y) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1Covdvfd_yy * rk1CovuThetaf_y) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfd_yz * rk1CovuThetaf_y) / rk1sdetg) + (-(vector(rk1Bfu_x, i, j, k) * rk1Covdvfd_zx * rk1CovuThetaf_z) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1Covdvfd_zy * rk1CovuThetaf_z) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfd_zz * rk1CovuThetaf_z) / rk1sdetg))) / (rk1sdetg * rk1Epf)) / rk1Thetaf;
							rk1PhiMh_yz = 2.0 * (rk1Thetaf * ((-rk1Covdvfu_xz * rk1CovuBfu_xy) + rk1Covdvfu_xy * rk1CovuBfu_xz + (-rk1Covdvfu_yz * rk1CovuBfu_yy) + rk1Covdvfu_yy * rk1CovuBfu_yz + (-rk1Covdvfu_zz * rk1CovuBfu_zy) + rk1Covdvfu_zy * rk1CovuBfu_zz) + (vector(rk1Bfu_z, i, j, k) * rk1CovuBfu_xy * rk1CovdBvf_x) / rk1sdetg + (vector(rk1Bfu_z, i, j, k) * rk1CovuBfu_yy * rk1CovdBvf_y) / rk1sdetg + (vector(rk1Bfu_z, i, j, k) * rk1CovuBfu_zy * rk1CovdBvf_z) / rk1sdetg + (-(vector(rk1Bfu_y, i, j, k) * rk1CovuBfu_xz * rk1CovdBvf_x) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1CovuBfu_yz * rk1CovdBvf_y) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1CovuBfu_zz * rk1CovdBvf_z) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfu_xy * rk1CovuThetaf_x) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfu_yy * rk1CovuThetaf_y) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfu_zy * rk1CovuThetaf_z) / rk1sdetg) + (vector(rk1Bfu_y, i, j, k) * rk1Covdvfu_xz * rk1CovuThetaf_x) / rk1sdetg + (vector(rk1Bfu_y, i, j, k) * rk1Covdvfu_yz * rk1CovuThetaf_y) / rk1sdetg + (vector(rk1Bfu_y, i, j, k) * rk1Covdvfu_zz * rk1CovuThetaf_z) / rk1sdetg) / rk1Thetaf;
							rk1PhiMh_xz = 2.0 * (rk1Thetaf * ((-rk1Covdvfu_xz * rk1CovuBfu_xx) + rk1Covdvfu_xx * rk1CovuBfu_xz + (-rk1Covdvfu_yz * rk1CovuBfu_yx) + rk1Covdvfu_yx * rk1CovuBfu_yz + (-rk1Covdvfu_zz * rk1CovuBfu_zx) + rk1Covdvfu_zx * rk1CovuBfu_zz) + (vector(rk1Bfu_z, i, j, k) * rk1CovuBfu_xx * rk1CovdBvf_x) / rk1sdetg + (vector(rk1Bfu_z, i, j, k) * rk1CovuBfu_yx * rk1CovdBvf_y) / rk1sdetg + (vector(rk1Bfu_z, i, j, k) * rk1CovuBfu_zx * rk1CovdBvf_z) / rk1sdetg + (-(vector(rk1Bfu_x, i, j, k) * rk1CovuBfu_xz * rk1CovdBvf_x) / rk1sdetg) + (-(vector(rk1Bfu_x, i, j, k) * rk1CovuBfu_yz * rk1CovdBvf_y) / rk1sdetg) + (-(vector(rk1Bfu_x, i, j, k) * rk1CovuBfu_zz * rk1CovdBvf_z) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfu_xx * rk1CovuThetaf_x) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfu_yx * rk1CovuThetaf_y) / rk1sdetg) + (-(vector(rk1Bfu_z, i, j, k) * rk1Covdvfu_zx * rk1CovuThetaf_z) / rk1sdetg) + (vector(rk1Bfu_x, i, j, k) * rk1Covdvfu_xz * rk1CovuThetaf_x) / rk1sdetg + (vector(rk1Bfu_x, i, j, k) * rk1Covdvfu_yz * rk1CovuThetaf_y) / rk1sdetg + (vector(rk1Bfu_x, i, j, k) * rk1Covdvfu_zz * rk1CovuThetaf_z) / rk1sdetg) / rk1Thetaf;
							rk1PhiMh_xy = 2.0 * (rk1Thetaf * ((-rk1Covdvfu_xy * rk1CovuBfu_xx) + rk1Covdvfu_xx * rk1CovuBfu_xy + (-rk1Covdvfu_yy * rk1CovuBfu_yx) + rk1Covdvfu_yx * rk1CovuBfu_yy + (-rk1Covdvfu_zy * rk1CovuBfu_zx) + rk1Covdvfu_zx * rk1CovuBfu_zy) + (vector(rk1Bfu_y, i, j, k) * rk1CovuBfu_xx * rk1CovdBvf_x) / rk1sdetg + (vector(rk1Bfu_y, i, j, k) * rk1CovuBfu_yx * rk1CovdBvf_y) / rk1sdetg + (vector(rk1Bfu_y, i, j, k) * rk1CovuBfu_zx * rk1CovdBvf_z) / rk1sdetg + (-(vector(rk1Bfu_x, i, j, k) * rk1CovuBfu_xy * rk1CovdBvf_x) / rk1sdetg) + (-(vector(rk1Bfu_x, i, j, k) * rk1CovuBfu_yy * rk1CovdBvf_y) / rk1sdetg) + (-(vector(rk1Bfu_x, i, j, k) * rk1CovuBfu_zy * rk1CovdBvf_z) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1Covdvfu_xx * rk1CovuThetaf_x) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1Covdvfu_yx * rk1CovuThetaf_y) / rk1sdetg) + (-(vector(rk1Bfu_y, i, j, k) * rk1Covdvfu_zx * rk1CovuThetaf_z) / rk1sdetg) + (vector(rk1Bfu_x, i, j, k) * rk1Covdvfu_xy * rk1CovuThetaf_x) / rk1sdetg + (vector(rk1Bfu_x, i, j, k) * rk1Covdvfu_yy * rk1CovuThetaf_y) / rk1sdetg + (vector(rk1Bfu_x, i, j, k) * rk1Covdvfu_zy * rk1CovuThetaf_z) / rk1sdetg) / rk1Thetaf;
							rk1Covuepsf_x = rk1chi_max * (rk1gtu_xx * rk1Covdepsf_x + rk1gtu_xy * rk1Covdepsf_y + rk1gtu_xz * rk1Covdepsf_z);
						} else {
							rk1Covdepsf_z = 0.0;
							rk1Covdhf_z = 0.0;
							rk1CovdEpf_z = 0.0;
							rk1CovdThetaf_z = 0.0;
							rk1Covdepsf_y = 0.0;
							rk1Covdhf_y = 0.0;
							rk1CovdEpf_y = 0.0;
							rk1CovdThetaf_y = 0.0;
							rk1Covdepsf_x = 0.0;
							rk1Covuepsf_z = 0.0;
							rk1Covuepsf_y = 0.0;
							rk1Covdhf_x = 0.0;
							rk1CovdEpf_x = 0.0;
							rk1CovdThetaf_x = 0.0;
							rk1CovuThetaf_z = 0.0;
							rk1CovuThetaf_x = 0.0;
							rk1CovuThetaf_y = 0.0;
							rk1Phivh_z = 0.0;
							rk1Phivh_x = 0.0;
							rk1Phivh_y = 0.0;
							rk1PhiMh_yz = 0.0;
							rk1PhiMh_xz = 0.0;
							rk1PhiMh_xy = 0.0;
							rk1Covuepsf_x = 0.0;
						}
						rk1faceta = 0.04166666666666667 * rk1inv_chi * (dx[0] * dx[0]) * Smooth(threshold_Dles_min, threshold_Dles_max, vector(rk1Df, i, j, k), dx, simPlat_dt, ilast, jlast);
						if (calculate_les_terms) {
							rk1CovdEfu_yz = d_CovdEfu_yz_o0_t0_m0_l0;
							rk1CovuEfu_zz = rk1chi_max * (rk1gtu_xz * rk1CovdEfu_xz + rk1gtu_yz * rk1CovdEfu_yz + rk1gtu_zz * rk1CovdEfu_zz);
							rk1CovuEfu_xz = rk1chi_max * (rk1gtu_xx * rk1CovdEfu_xz + rk1gtu_xy * rk1CovdEfu_yz + rk1gtu_xz * rk1CovdEfu_zz);
							rk1CovdEfu_yy = d_CovdEfu_yy_o0_t0_m0_l0;
							rk1CovuEfu_zy = rk1chi_max * (rk1gtu_xz * rk1CovdEfu_xy + rk1gtu_yz * rk1CovdEfu_yy + rk1gtu_zz * rk1CovdEfu_zy);
							rk1CovuEfu_xy = rk1chi_max * (rk1gtu_xx * rk1CovdEfu_xy + rk1gtu_xy * rk1CovdEfu_yy + rk1gtu_xz * rk1CovdEfu_zy);
							rk1CovdEfu_yx = d_CovdEfu_yx_o0_t0_m0_l0;
							rk1CovdEfd_yy = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * rk1CovdEfu_yx + vector(rk1gtd_yy, i, j, k) * rk1CovdEfu_yy + vector(rk1gtd_yz, i, j, k) * rk1CovdEfu_yz);
							rk1CovdEfd_yz = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * rk1CovdEfu_yx + vector(rk1gtd_yz, i, j, k) * rk1CovdEfu_yy + vector(rk1gtd_zz, i, j, k) * rk1CovdEfu_yz);
							rk1CovdEfd_yx = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * rk1CovdEfu_yx + vector(rk1gtd_xy, i, j, k) * rk1CovdEfu_yy + vector(rk1gtd_xz, i, j, k) * rk1CovdEfu_yz);
							rk1CovuEfu_zx = rk1chi_max * (rk1gtu_xz * rk1CovdEfu_xx + rk1gtu_yz * rk1CovdEfu_yx + rk1gtu_zz * rk1CovdEfu_zx);
							rk1CovuEfu_xx = rk1chi_max * (rk1gtu_xx * rk1CovdEfu_xx + rk1gtu_xy * rk1CovdEfu_yx + rk1gtu_xz * rk1CovdEfu_zx);
							rk1CovuEfu_yx = rk1chi_max * (rk1gtu_xy * rk1CovdEfu_xx + rk1gtu_yy * rk1CovdEfu_yx + rk1gtu_yz * rk1CovdEfu_zx);
							rk1CovuEfu_yz = rk1chi_max * (rk1gtu_xy * rk1CovdEfu_xz + rk1gtu_yy * rk1CovdEfu_yz + rk1gtu_yz * rk1CovdEfu_zz);
							rk1CovuEfu_yy = rk1chi_max * (rk1gtu_xy * rk1CovdEfu_xy + rk1gtu_yy * rk1CovdEfu_yy + rk1gtu_yz * rk1CovdEfu_zy);
							rk1PhiThetah = (rk1Thetaf * (rk1CovdBfd_xx * rk1CovuBfu_xx + rk1CovdBfd_yx * rk1CovuBfu_yx + rk1CovdBfd_zx * rk1CovuBfu_zx + rk1CovdBfd_xy * rk1CovuBfu_xy + rk1CovdBfd_yy * rk1CovuBfu_yy + rk1CovdBfd_zy * rk1CovuBfu_zy + rk1CovdBfd_xz * rk1CovuBfu_xz + rk1CovdBfd_yz * rk1CovuBfu_yz + rk1CovdBfd_zz * rk1CovuBfu_zz + (-rk1CovdEfd_xx * rk1CovuEfu_xx) + (-rk1CovdEfd_yx * rk1CovuEfu_yx) + (-rk1CovdEfd_zx * rk1CovuEfu_zx) + (-rk1CovdEfd_xy * rk1CovuEfu_xy) + (-rk1CovdEfd_yy * rk1CovuEfu_yy) + (-rk1CovdEfd_zy * rk1CovuEfu_zy) + (-rk1CovdEfd_xz * rk1CovuEfu_xz) + (-rk1CovdEfd_yz * rk1CovuEfu_yz) + (-rk1CovdEfd_zz * rk1CovuEfu_zz) + (-0.5 * (((-vector(vfd_y, i, j, k) * rk1Bfd_x) + vector(vfd_x, i, j, k) * rk1Bfd_y) * rk1PhiMh_xy + (vector(vfd_x, i, j, k) * rk1Bfd_z - vector(vfd_z, i, j, k) * rk1Bfd_x) * rk1PhiMh_xz + (-(vector(vfd_y, i, j, k) * rk1Bfd_x - vector(vfd_x, i, j, k) * rk1Bfd_y) * rk1PhiMh_xy) + ((-vector(vfd_z, i, j, k) * rk1Bfd_y) + vector(vfd_y, i, j, k) * rk1Bfd_z) * rk1PhiMh_yz + (-(vector(vfd_z, i, j, k) * rk1Bfd_x - vector(vfd_x, i, j, k) * rk1Bfd_z) * rk1PhiMh_xz) + (-(vector(vfd_z, i, j, k) * rk1Bfd_y - vector(vfd_y, i, j, k) * rk1Bfd_z) * rk1PhiMh_yz)) / rk1sdetg))) / (rk1Thetaf - rk1sqEf);
							rk1HPres = (rk1Epf * rk1sqW * (rk1Thetaf - rk1sqEf) * (vector(rhof, i, j, k) * (rk1Covddpfdrho_x * rk1Covurhof_x + rk1Covddpfdrho_y * rk1Covurhof_y + rk1Covddpfdrho_z * rk1Covurhof_z + rk1Covddpfdeps_x * rk1Covuepsf_x + rk1Covddpfdeps_y * rk1Covuepsf_y + rk1Covddpfdeps_z * rk1Covuepsf_z) + (-2.0 * vector(dpfdeps, i, j, k) * (rk1Covurhof_x * rk1Covdepsf_x + rk1Covurhof_y * rk1Covdepsf_y + rk1Covurhof_z * rk1Covdepsf_z)) + (-(rk1Epf * vector(dpfdeps, i, j, k) - rk1PhiAh) * (0.25 * rk1sqW * (rk1CovdinvsqW_x * rk1CovuinvsqW_x + rk1CovdinvsqW_y * rk1CovuinvsqW_y + rk1CovdinvsqW_z * rk1CovuinvsqW_z) + (rk1CovdinvsqW_x * rk1Covurhof_x + rk1CovdinvsqW_y * rk1Covurhof_y + rk1CovdinvsqW_z * rk1Covurhof_z) / vector(rhof, i, j, k))) + (-2.0 * (vector(dpfdeps, i, j, k) * (rk1CovdBfd_xx * rk1CovuBfu_xx + rk1CovdBfd_yx * rk1CovuBfu_yx + rk1CovdBfd_zx * rk1CovuBfu_zx + rk1CovdBfd_xy * rk1CovuBfu_xy + rk1CovdBfd_yy * rk1CovuBfu_yy + rk1CovdBfd_zy * rk1CovuBfu_zy + rk1CovdBfd_xz * rk1CovuBfu_xz + rk1CovdBfd_yz * rk1CovuBfu_yz + rk1CovdBfd_zz * rk1CovuBfu_zz + (-(rk1sqW * rk1sqW) * (rk1CovuinvsqW_x * rk1Covdhf_x + rk1CovuinvsqW_y * rk1Covdhf_y + rk1CovuinvsqW_z * rk1Covdhf_z)))) / rk1sqW) + (-(rk1Epf * vector(dpfdeps, i, j, k) + rk1PhiAh) * (vector(vfd_x, i, j, k) * rk1Phivh_x + vector(vfd_y, i, j, k) * rk1Phivh_y + vector(vfd_z, i, j, k) * rk1Phivh_z + rk1Covdvfd_xx * rk1Covuvfu_xx + rk1Covdvfd_xy * rk1Covuvfu_xy + rk1Covdvfd_xz * rk1Covuvfu_xz + rk1Covdvfd_yx * rk1Covuvfu_yx + rk1Covdvfd_yy * rk1Covuvfu_yy + rk1Covdvfd_yz * rk1Covuvfu_yz + rk1Covdvfd_zx * rk1Covuvfu_zx + rk1Covdvfd_zy * rk1Covuvfu_zy + rk1Covdvfd_zz * rk1Covuvfu_zz + rk1sqW * (rk1CovdinvsqW_x * rk1CovuinvsqW_x + rk1CovdinvsqW_y * rk1CovuinvsqW_y + rk1CovdinvsqW_z * rk1CovuinvsqW_z))) + (rk1PhiThetah * ((rk1Epf * vector(dpfdeps, i, j, k) + rk1PhiAh) * (rk1Thetaf - rk1sqEf) - (rk1PhiAh * rk1Thetaf) / rk1sqW)) / (rk1Epf * rk1Thetaf))) / ((rk1Epf * vector(rhof, i, j, k) - rk1PhiAh) * (rk1Thetaf - rk1sqEf) * rk1sqW + rk1PhiAh * rk1Thetaf) + rk1Covddpfdye_x * rk1CovuYef_x + rk1Covddpfdye_y * rk1CovuYef_y + rk1Covddpfdye_z * rk1CovuYef_z - 2.0 * vector(dpfdye, i, j, k) / vector(rk1Df, i, j, k) * (rk1CovdDf_x * rk1CovuYef_x + rk1CovdDf_y * rk1CovuYef_y + rk1CovdDf_z * rk1CovuYef_z);
							rk1HTheta = rk1PhiThetah + (rk1Thetaf * rk1HPres) / (rk1Thetaf - rk1sqEf);
							rk1Hv_z = rk1Phivh_z + (-((rk1vfu_z + (rk1Bvf * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1Epf)) * rk1HTheta) / rk1Thetaf);
							rk1HN_z = 2.0 * rk1CovdDf_x * rk1Covuvfu_xz + 2.0 * rk1CovdDf_y * rk1Covuvfu_yz + 2.0 * rk1CovdDf_z * rk1Covuvfu_zz + (vector(rk1Df, i, j, k) * rk1Hv_z) / rk1sdetg;
							rk1HNe_z = 2.0 * rk1CovdDYf_x * rk1Covuvfu_xz + 2.0 * rk1CovdDYf_y * rk1Covuvfu_yz + 2.0 * rk1CovdDYf_z * rk1Covuvfu_zz + (vector(rk1DYf, i, j, k) * rk1Hv_z) / rk1sdetg;
							rk1Hv_y = rk1Phivh_y + (-((rk1vfu_y + (rk1Bvf * vector(rk1Bfu_y, i, j, k)) / (rk1sdetg * rk1Epf)) * rk1HTheta) / rk1Thetaf);
							rk1HN_y = 2.0 * rk1CovdDf_x * rk1Covuvfu_xy + 2.0 * rk1CovdDf_y * rk1Covuvfu_yy + 2.0 * rk1CovdDf_z * rk1Covuvfu_zy + (vector(rk1Df, i, j, k) * rk1Hv_y) / rk1sdetg;
							rk1HM_yz = (-2.0 * rk1Covdvfu_xz * rk1CovuBfu_xy) + 2.0 * rk1Covdvfu_xy * rk1CovuBfu_xz + (-2.0 * rk1Covdvfu_yz * rk1CovuBfu_yy) + 2.0 * rk1Covdvfu_yy * rk1CovuBfu_yz + (-2.0 * rk1Covdvfu_zz * rk1CovuBfu_zy) + 2.0 * rk1Covdvfu_zy * rk1CovuBfu_zz + (vector(rk1Bfu_z, i, j, k) * rk1Hv_y - vector(rk1Bfu_y, i, j, k) * rk1Hv_z) / rk1sdetg;
							rk1HNe_y = 2.0 * rk1CovdDYf_x * rk1Covuvfu_xy + 2.0 * rk1CovdDYf_y * rk1Covuvfu_yy + 2.0 * rk1CovdDYf_z * rk1Covuvfu_zy + (vector(rk1DYf, i, j, k) * rk1Hv_y) / rk1sdetg;
							rk1Hv_x = rk1Phivh_x + (-((rk1vfu_x + (rk1Bvf * vector(rk1Bfu_x, i, j, k)) / (rk1sdetg * rk1Epf)) * rk1HTheta) / rk1Thetaf);
							rk1HN_x = 2.0 * rk1CovdDf_x * rk1Covuvfu_xx + 2.0 * rk1CovdDf_y * rk1Covuvfu_yx + 2.0 * rk1CovdDf_z * rk1Covuvfu_zx + (vector(rk1Df, i, j, k) * rk1Hv_x) / rk1sdetg;
							rk1HM_xz = (-2.0 * rk1Covdvfu_xz * rk1CovuBfu_xx) + 2.0 * rk1Covdvfu_xx * rk1CovuBfu_xz + (-2.0 * rk1Covdvfu_yz * rk1CovuBfu_yx) + 2.0 * rk1Covdvfu_yx * rk1CovuBfu_yz + (-2.0 * rk1Covdvfu_zz * rk1CovuBfu_zx) + 2.0 * rk1Covdvfu_zx * rk1CovuBfu_zz + (vector(rk1Bfu_z, i, j, k) * rk1Hv_x - vector(rk1Bfu_x, i, j, k) * rk1Hv_z) / rk1sdetg;
							rk1HM_xy = (-2.0 * rk1Covdvfu_xy * rk1CovuBfu_xx) + 2.0 * rk1Covdvfu_xx * rk1CovuBfu_xy + (-2.0 * rk1Covdvfu_yy * rk1CovuBfu_yx) + 2.0 * rk1Covdvfu_yx * rk1CovuBfu_yy + (-2.0 * rk1Covdvfu_zy * rk1CovuBfu_zx) + 2.0 * rk1Covdvfu_zx * rk1CovuBfu_zy + (vector(rk1Bfu_y, i, j, k) * rk1Hv_x - vector(rk1Bfu_x, i, j, k) * rk1Hv_y) / rk1sdetg;
							rk1HE_z = 0.5 * ((-2.0 * rk1gtu_yz * rk1HM_xz) + 2.0 * rk1gtu_zz * rk1HM_xy + 2.0 * rk1gtu_xz * rk1HM_yz) / sqrt(rk1chi_max);
							rk1HE_x = 0.5 * ((-2.0 * rk1gtu_xy * rk1HM_xz) + 2.0 * rk1gtu_xz * rk1HM_xy + 2.0 * rk1gtu_xx * rk1HM_yz) / sqrt(rk1chi_max);
							rk1HE_y = 0.5 * ((-2.0 * rk1gtu_yy * rk1HM_xz) + 2.0 * rk1gtu_yz * rk1HM_xy + 2.0 * rk1gtu_xy * rk1HM_yz) / sqrt(rk1chi_max);
							rk1HT_zz = 4.0 * rk1CovdEpf_x * rk1Covuvfu_xz * rk1vfu_z + 4.0 * rk1CovdEpf_y * rk1Covuvfu_yz * rk1vfu_z + 4.0 * rk1CovdEpf_z * rk1Covuvfu_zz * rk1vfu_z + rk1Epf * (2.0 * rk1vfu_z * rk1Hv_z + 2.0 * rk1Covdvfu_xz * rk1Covuvfu_xz + 2.0 * rk1Covdvfu_yz * rk1Covuvfu_yz + 2.0 * rk1Covdvfu_zz * rk1Covuvfu_zz) + 2.0 * (rk1vfu_z * rk1vfu_z) * rk1HPres + (-2.0 * rk1CovdBfu_xz * rk1CovuBfu_xz) + (-2.0 * rk1CovdBfu_yz * rk1CovuBfu_yz) + (-2.0 * rk1CovdBfu_zz * rk1CovuBfu_zz) + (-2.0 * rk1CovdEfu_xz * rk1CovuEfu_xz) + (-2.0 * rk1CovdEfu_yz * rk1CovuEfu_yz) + (-2.0 * rk1CovdEfu_zz * rk1CovuEfu_zz) + (-2.00 * vector(Efu_z, i, j, k) * rk1HE_z) + (rk1chi_max * rk1gtu_zz - rk1vfu_z * rk1vfu_z) * (rk1HPres + rk1CovdBfd_xx * rk1CovuBfu_xx + rk1CovdBfd_yx * rk1CovuBfu_yx + rk1CovdBfd_zx * rk1CovuBfu_zx + rk1CovdBfd_xy * rk1CovuBfu_xy + rk1CovdBfd_yy * rk1CovuBfu_yy + rk1CovdBfd_zy * rk1CovuBfu_zy + rk1CovdBfd_xz * rk1CovuBfu_xz + rk1CovdBfd_yz * rk1CovuBfu_yz + rk1CovdBfd_zz * rk1CovuBfu_zz + rk1CovdEfd_xx * rk1CovuEfu_xx + rk1CovdEfd_yx * rk1CovuEfu_yx + rk1CovdEfd_zx * rk1CovuEfu_zx + rk1CovdEfd_xy * rk1CovuEfu_xy + rk1CovdEfd_yy * rk1CovuEfu_yy + rk1CovdEfd_zy * rk1CovuEfu_zy + rk1CovdEfd_xz * rk1CovuEfu_xz + rk1CovdEfd_yz * rk1CovuEfu_yz + rk1CovdEfd_zz * rk1CovuEfu_zz + rk1Efd_x * rk1HE_x + rk1Efd_y * rk1HE_y + rk1Efd_z * rk1HE_z);
							rk1HT_yy = 4.0 * rk1CovdEpf_x * rk1Covuvfu_xy * rk1vfu_y + 4.0 * rk1CovdEpf_y * rk1Covuvfu_yy * rk1vfu_y + 4.0 * rk1CovdEpf_z * rk1Covuvfu_zy * rk1vfu_y + rk1Epf * (2.0 * rk1vfu_y * rk1Hv_y + 2.0 * rk1Covdvfu_xy * rk1Covuvfu_xy + 2.0 * rk1Covdvfu_yy * rk1Covuvfu_yy + 2.0 * rk1Covdvfu_zy * rk1Covuvfu_zy) + 2.0 * (rk1vfu_y * rk1vfu_y) * rk1HPres + (-2.0 * rk1CovdBfu_xy * rk1CovuBfu_xy) + (-2.0 * rk1CovdBfu_yy * rk1CovuBfu_yy) + (-2.0 * rk1CovdBfu_zy * rk1CovuBfu_zy) + (-2.0 * rk1CovdEfu_xy * rk1CovuEfu_xy) + (-2.0 * rk1CovdEfu_yy * rk1CovuEfu_yy) + (-2.0 * rk1CovdEfu_zy * rk1CovuEfu_zy) + (-2.00 * vector(Efu_y, i, j, k) * rk1HE_y) + (rk1chi_max * rk1gtu_yy - rk1vfu_y * rk1vfu_y) * (rk1HPres + rk1CovdBfd_xx * rk1CovuBfu_xx + rk1CovdBfd_yx * rk1CovuBfu_yx + rk1CovdBfd_zx * rk1CovuBfu_zx + rk1CovdBfd_xy * rk1CovuBfu_xy + rk1CovdBfd_yy * rk1CovuBfu_yy + rk1CovdBfd_zy * rk1CovuBfu_zy + rk1CovdBfd_xz * rk1CovuBfu_xz + rk1CovdBfd_yz * rk1CovuBfu_yz + rk1CovdBfd_zz * rk1CovuBfu_zz + rk1CovdEfd_xx * rk1CovuEfu_xx + rk1CovdEfd_yx * rk1CovuEfu_yx + rk1CovdEfd_zx * rk1CovuEfu_zx + rk1CovdEfd_xy * rk1CovuEfu_xy + rk1CovdEfd_yy * rk1CovuEfu_yy + rk1CovdEfd_zy * rk1CovuEfu_zy + rk1CovdEfd_xz * rk1CovuEfu_xz + rk1CovdEfd_yz * rk1CovuEfu_yz + rk1CovdEfd_zz * rk1CovuEfu_zz + rk1Efd_x * rk1HE_x + rk1Efd_y * rk1HE_y + rk1Efd_z * rk1HE_z);
							rk1HT_yz = 2.0 * rk1CovdEpf_x * (rk1Covuvfu_xz * rk1vfu_y + rk1Covuvfu_xy * rk1vfu_z) + 2.0 * rk1CovdEpf_y * (rk1Covuvfu_yz * rk1vfu_y + rk1Covuvfu_yy * rk1vfu_z) + 2.0 * rk1CovdEpf_z * (rk1Covuvfu_zz * rk1vfu_y + rk1Covuvfu_zy * rk1vfu_z) + rk1Epf * (rk1vfu_y * rk1Hv_z + rk1vfu_z * rk1Hv_y + 2.0 * rk1Covdvfu_xy * rk1Covuvfu_xz + 2.0 * rk1Covdvfu_yy * rk1Covuvfu_yz + 2.0 * rk1Covdvfu_zy * rk1Covuvfu_zz) + 2.0 * rk1vfu_y * rk1vfu_z * rk1HPres + (-2.0 * rk1CovdBfu_xy * rk1CovuBfu_xz) + (-2.0 * rk1CovdBfu_yy * rk1CovuBfu_yz) + (-2.0 * rk1CovdBfu_zy * rk1CovuBfu_zz) + (-2.0 * rk1CovdEfu_xy * rk1CovuEfu_xz) + (-2.0 * rk1CovdEfu_yy * rk1CovuEfu_yz) + (-2.0 * rk1CovdEfu_zy * rk1CovuEfu_zz) + (-1.00 * vector(Efu_z, i, j, k) * rk1HE_y) + (-1.00 * vector(Efu_y, i, j, k) * rk1HE_z) + (rk1chi_max * rk1gtu_yz - rk1vfu_z * rk1vfu_y) * (rk1HPres + rk1CovdBfd_xx * rk1CovuBfu_xx + rk1CovdBfd_yx * rk1CovuBfu_yx + rk1CovdBfd_zx * rk1CovuBfu_zx + rk1CovdBfd_xy * rk1CovuBfu_xy + rk1CovdBfd_yy * rk1CovuBfu_yy + rk1CovdBfd_zy * rk1CovuBfu_zy + rk1CovdBfd_xz * rk1CovuBfu_xz + rk1CovdBfd_yz * rk1CovuBfu_yz + rk1CovdBfd_zz * rk1CovuBfu_zz + rk1CovdEfd_xx * rk1CovuEfu_xx + rk1CovdEfd_yx * rk1CovuEfu_yx + rk1CovdEfd_zx * rk1CovuEfu_zx + rk1CovdEfd_xy * rk1CovuEfu_xy + rk1CovdEfd_yy * rk1CovuEfu_yy + rk1CovdEfd_zy * rk1CovuEfu_zy + rk1CovdEfd_xz * rk1CovuEfu_xz + rk1CovdEfd_yz * rk1CovuEfu_yz + rk1CovdEfd_zz * rk1CovuEfu_zz + rk1Efd_x * rk1HE_x + rk1Efd_y * rk1HE_y + rk1Efd_z * rk1HE_z);
							rk1HT_xx = 4.0 * rk1CovdEpf_x * rk1Covuvfu_xx * rk1vfu_x + 4.0 * rk1CovdEpf_y * rk1Covuvfu_yx * rk1vfu_x + 4.0 * rk1CovdEpf_z * rk1Covuvfu_zx * rk1vfu_x + rk1Epf * (2.0 * rk1vfu_x * rk1Hv_x + 2.0 * rk1Covdvfu_xx * rk1Covuvfu_xx + 2.0 * rk1Covdvfu_yx * rk1Covuvfu_yx + 2.0 * rk1Covdvfu_zx * rk1Covuvfu_zx) + 2.0 * (rk1vfu_x * rk1vfu_x) * rk1HPres + (-2.0 * rk1CovdBfu_xx * rk1CovuBfu_xx) + (-2.0 * rk1CovdBfu_yx * rk1CovuBfu_yx) + (-2.0 * rk1CovdBfu_zx * rk1CovuBfu_zx) + (-2.0 * rk1CovdEfu_xx * rk1CovuEfu_xx) + (-2.0 * rk1CovdEfu_yx * rk1CovuEfu_yx) + (-2.0 * rk1CovdEfu_zx * rk1CovuEfu_zx) + (-2.00 * vector(Efu_x, i, j, k) * rk1HE_x) + (rk1chi_max * rk1gtu_xx - rk1vfu_x * rk1vfu_x) * (rk1HPres + rk1CovdBfd_xx * rk1CovuBfu_xx + rk1CovdBfd_yx * rk1CovuBfu_yx + rk1CovdBfd_zx * rk1CovuBfu_zx + rk1CovdBfd_xy * rk1CovuBfu_xy + rk1CovdBfd_yy * rk1CovuBfu_yy + rk1CovdBfd_zy * rk1CovuBfu_zy + rk1CovdBfd_xz * rk1CovuBfu_xz + rk1CovdBfd_yz * rk1CovuBfu_yz + rk1CovdBfd_zz * rk1CovuBfu_zz + rk1CovdEfd_xx * rk1CovuEfu_xx + rk1CovdEfd_yx * rk1CovuEfu_yx + rk1CovdEfd_zx * rk1CovuEfu_zx + rk1CovdEfd_xy * rk1CovuEfu_xy + rk1CovdEfd_yy * rk1CovuEfu_yy + rk1CovdEfd_zy * rk1CovuEfu_zy + rk1CovdEfd_xz * rk1CovuEfu_xz + rk1CovdEfd_yz * rk1CovuEfu_yz + rk1CovdEfd_zz * rk1CovuEfu_zz + rk1Efd_x * rk1HE_x + rk1Efd_y * rk1HE_y + rk1Efd_z * rk1HE_z);
							rk1HT_xz = 2.0 * rk1CovdEpf_x * (rk1Covuvfu_xz * rk1vfu_x + rk1Covuvfu_xx * rk1vfu_z) + 2.0 * rk1CovdEpf_y * (rk1Covuvfu_yz * rk1vfu_x + rk1Covuvfu_yx * rk1vfu_z) + 2.0 * rk1CovdEpf_z * (rk1Covuvfu_zz * rk1vfu_x + rk1Covuvfu_zx * rk1vfu_z) + rk1Epf * (rk1vfu_x * rk1Hv_z + rk1vfu_z * rk1Hv_x + 2.0 * rk1Covdvfu_xx * rk1Covuvfu_xz + 2.0 * rk1Covdvfu_yx * rk1Covuvfu_yz + 2.0 * rk1Covdvfu_zx * rk1Covuvfu_zz) + 2.0 * rk1vfu_x * rk1vfu_z * rk1HPres + (-2.0 * rk1CovdBfu_xx * rk1CovuBfu_xz) + (-2.0 * rk1CovdBfu_yx * rk1CovuBfu_yz) + (-2.0 * rk1CovdBfu_zx * rk1CovuBfu_zz) + (-2.0 * rk1CovdEfu_xx * rk1CovuEfu_xz) + (-2.0 * rk1CovdEfu_yx * rk1CovuEfu_yz) + (-2.0 * rk1CovdEfu_zx * rk1CovuEfu_zz) + (-1.00 * vector(Efu_z, i, j, k) * rk1HE_x) + (-1.00 * vector(Efu_x, i, j, k) * rk1HE_z) + (rk1chi_max * rk1gtu_xz - rk1vfu_z * rk1vfu_x) * (rk1HPres + rk1CovdBfd_xx * rk1CovuBfu_xx + rk1CovdBfd_yx * rk1CovuBfu_yx + rk1CovdBfd_zx * rk1CovuBfu_zx + rk1CovdBfd_xy * rk1CovuBfu_xy + rk1CovdBfd_yy * rk1CovuBfu_yy + rk1CovdBfd_zy * rk1CovuBfu_zy + rk1CovdBfd_xz * rk1CovuBfu_xz + rk1CovdBfd_yz * rk1CovuBfu_yz + rk1CovdBfd_zz * rk1CovuBfu_zz + rk1CovdEfd_xx * rk1CovuEfu_xx + rk1CovdEfd_yx * rk1CovuEfu_yx + rk1CovdEfd_zx * rk1CovuEfu_zx + rk1CovdEfd_xy * rk1CovuEfu_xy + rk1CovdEfd_yy * rk1CovuEfu_yy + rk1CovdEfd_zy * rk1CovuEfu_zy + rk1CovdEfd_xz * rk1CovuEfu_xz + rk1CovdEfd_yz * rk1CovuEfu_yz + rk1CovdEfd_zz * rk1CovuEfu_zz + rk1Efd_x * rk1HE_x + rk1Efd_y * rk1HE_y + rk1Efd_z * rk1HE_z);
							rk1HT_xy = 2.0 * rk1CovdEpf_x * (rk1Covuvfu_xy * rk1vfu_x + rk1Covuvfu_xx * rk1vfu_y) + 2.0 * rk1CovdEpf_y * (rk1Covuvfu_yy * rk1vfu_x + rk1Covuvfu_yx * rk1vfu_y) + 2.0 * rk1CovdEpf_z * (rk1Covuvfu_zy * rk1vfu_x + rk1Covuvfu_zx * rk1vfu_y) + rk1Epf * (rk1vfu_x * rk1Hv_y + rk1vfu_y * rk1Hv_x + 2.0 * rk1Covdvfu_xx * rk1Covuvfu_xy + 2.0 * rk1Covdvfu_yx * rk1Covuvfu_yy + 2.0 * rk1Covdvfu_zx * rk1Covuvfu_zy) + 2.0 * rk1vfu_x * rk1vfu_y * rk1HPres + (-2.0 * rk1CovdBfu_xx * rk1CovuBfu_xy) + (-2.0 * rk1CovdBfu_yx * rk1CovuBfu_yy) + (-2.0 * rk1CovdBfu_zx * rk1CovuBfu_zy) + (-2.0 * rk1CovdEfu_xx * rk1CovuEfu_xy) + (-2.0 * rk1CovdEfu_yx * rk1CovuEfu_yy) + (-2.0 * rk1CovdEfu_zx * rk1CovuEfu_zy) + (-1.00 * vector(Efu_y, i, j, k) * rk1HE_x) + (-1.00 * vector(Efu_x, i, j, k) * rk1HE_y) + (rk1chi_max * rk1gtu_xy - rk1vfu_y * rk1vfu_x) * (rk1HPres + rk1CovdBfd_xx * rk1CovuBfu_xx + rk1CovdBfd_yx * rk1CovuBfu_yx + rk1CovdBfd_zx * rk1CovuBfu_zx + rk1CovdBfd_xy * rk1CovuBfu_xy + rk1CovdBfd_yy * rk1CovuBfu_yy + rk1CovdBfd_zy * rk1CovuBfu_zy + rk1CovdBfd_xz * rk1CovuBfu_xz + rk1CovdBfd_yz * rk1CovuBfu_yz + rk1CovdBfd_zz * rk1CovuBfu_zz + rk1CovdEfd_xx * rk1CovuEfu_xx + rk1CovdEfd_yx * rk1CovuEfu_yx + rk1CovdEfd_zx * rk1CovuEfu_zx + rk1CovdEfd_xy * rk1CovuEfu_xy + rk1CovdEfd_yy * rk1CovuEfu_yy + rk1CovdEfd_zy * rk1CovuEfu_zy + rk1CovdEfd_xz * rk1CovuEfu_xz + rk1CovdEfd_yz * rk1CovuEfu_yz + rk1CovdEfd_zz * rk1CovuEfu_zz + rk1Efd_x * rk1HE_x + rk1Efd_y * rk1HE_y + rk1Efd_z * rk1HE_z);
							rk1HNe_x = 2.0 * rk1CovdDYf_x * rk1Covuvfu_xx + 2.0 * rk1CovdDYf_y * rk1Covuvfu_yx + 2.0 * rk1CovdDYf_z * rk1Covuvfu_zx + (vector(rk1DYf, i, j, k) * rk1Hv_x) / rk1sdetg;
							d_TauN_x_o0_t0_m0_l0 = -rk1faceta * C_N * rk1HN_x;
							d_TauN_y_o0_t0_m0_l0 = -rk1faceta * C_N * rk1HN_y;
							d_TauN_z_o0_t0_m0_l0 = -rk1faceta * C_N * rk1HN_z;
							d_TauNe_x_o0_t0_m0_l0 = -rk1faceta * C_N * rk1HNe_x;
							d_TauNe_y_o0_t0_m0_l0 = -rk1faceta * C_N * rk1HNe_y;
							d_TauNe_z_o0_t0_m0_l0 = -rk1faceta * C_N * rk1HNe_z;
							d_TauM_xy_o0_t0_m0_l0 = -rk1faceta * C_M * rk1HM_xy;
							d_TauM_xz_o0_t0_m0_l0 = -rk1faceta * C_M * rk1HM_xz;
							d_TauM_yz_o0_t0_m0_l0 = -rk1faceta * C_M * rk1HM_yz;
							d_TauT_xx_o0_t0_m0_l0 = -rk1faceta * C_T * rk1HT_xx;
							d_TauT_xy_o0_t0_m0_l0 = -rk1faceta * C_T * rk1HT_xy;
							d_TauT_xz_o0_t0_m0_l0 = -rk1faceta * C_T * rk1HT_xz;
							d_TauT_yy_o0_t0_m0_l0 = -rk1faceta * C_T * rk1HT_yy;
							d_TauT_yz_o0_t0_m0_l0 = -rk1faceta * C_T * rk1HT_yz;
							d_TauT_zz_o0_t0_m0_l0 = -rk1faceta * C_T * rk1HT_zz;
							vector(TauNe_x, i, j, k) = d_TauNe_x_o0_t0_m0_l0;
							vector(TauNe_y, i, j, k) = d_TauNe_y_o0_t0_m0_l0;
							vector(TauT_yz, i, j, k) = d_TauT_yz_o0_t0_m0_l0;
							vector(TauT_xy, i, j, k) = d_TauT_xy_o0_t0_m0_l0;
							vector(TauM_yz, i, j, k) = d_TauM_yz_o0_t0_m0_l0;
							vector(TauM_xy, i, j, k) = d_TauM_xy_o0_t0_m0_l0;
							vector(TauNe_z, i, j, k) = d_TauNe_z_o0_t0_m0_l0;
							vector(TauT_xz, i, j, k) = d_TauT_xz_o0_t0_m0_l0;
							vector(TauN_x, i, j, k) = d_TauN_x_o0_t0_m0_l0;
							vector(TauT_zz, i, j, k) = d_TauT_zz_o0_t0_m0_l0;
							vector(TauT_yy, i, j, k) = d_TauT_yy_o0_t0_m0_l0;
							vector(TauT_xx, i, j, k) = d_TauT_xx_o0_t0_m0_l0;
							vector(TauM_xz, i, j, k) = d_TauM_xz_o0_t0_m0_l0;
							vector(TauN_z, i, j, k) = d_TauN_z_o0_t0_m0_l0;
							vector(TauN_y, i, j, k) = d_TauN_y_o0_t0_m0_l0;
						} else {
							rk1CovdEfu_yz = 0.0;
							rk1CovuEfu_zz = 0.0;
							rk1CovuEfu_xz = 0.0;
							rk1CovdEfu_yy = 0.0;
							rk1CovuEfu_zy = 0.0;
							rk1CovuEfu_xy = 0.0;
							rk1CovdEfu_yx = 0.0;
							rk1CovdEfd_yy = 0.0;
							rk1CovdEfd_yz = 0.0;
							rk1CovdEfd_yx = 0.0;
							rk1CovuEfu_zx = 0.0;
							rk1CovuEfu_xx = 0.0;
							rk1CovuEfu_yx = 0.0;
							rk1CovuEfu_yz = 0.0;
							rk1CovuEfu_yy = 0.0;
							rk1PhiThetah = 0.0;
							rk1HPres = 0.0;
							rk1HTheta = 0.0;
							rk1Hv_z = 0.0;
							rk1HN_z = 0.0;
							rk1HNe_z = 0.0;
							rk1Hv_y = 0.0;
							rk1HN_y = 0.0;
							rk1HM_yz = 0.0;
							rk1HNe_y = 0.0;
							rk1Hv_x = 0.0;
							rk1HN_x = 0.0;
							rk1HM_xz = 0.0;
							rk1HM_xy = 0.0;
							rk1HE_z = 0.0;
							rk1HE_x = 0.0;
							rk1HE_y = 0.0;
							rk1HT_zz = 0.0;
							rk1HT_yy = 0.0;
							rk1HT_yz = 0.0;
							rk1HT_xx = 0.0;
							rk1HT_xz = 0.0;
							rk1HT_xy = 0.0;
							rk1HNe_x = 0.0;
						}
						optdepthe_tmp = vector(optdepthe, i, j, k);
						optdeptha_tmp = vector(optdeptha, i, j, k);
						optdepthx_tmp = vector(optdepthx, i, j, k);
						chie_tmp = vector(chie, i, j, k);
						chia_tmp = vector(chia, i, j, k);
						chix_tmp = vector(chix, i, j, k);
						qnu_tmp = vector(qnu, i, j, k);
						rnu_tmp = vector(rnu, i, j, k);
						qnu_e_tmp = vector(qnu_e, i, j, k);
						qnu_a_tmp = vector(qnu_a, i, j, k);
						qnu_x_tmp = vector(qnu_x, i, j, k);
						betabeta = vector(rk1gtd_xx, i, j, k) * vector(rk1Betau_x, i, j, k) * vector(rk1Betau_x, i, j, k) + vector(rk1gtd_yy, i, j, k) * vector(rk1Betau_y, i, j, k) * vector(rk1Betau_y, i, j, k) + vector(rk1gtd_zz, i, j, k) * vector(rk1Betau_z, i, j, k) * vector(rk1Betau_z, i, j, k) + 2.0 * vector(rk1gtd_xy, i, j, k) * vector(rk1Betau_x, i, j, k) * vector(rk1Betau_y, i, j, k) + 2.0 * vector(rk1gtd_xz, i, j, k) * vector(rk1Betau_x, i, j, k) * vector(rk1Betau_z, i, j, k) + 2.0 * vector(rk1gtd_yz, i, j, k) * vector(rk1Betau_y, i, j, k) * vector(rk1Betau_z, i, j, k);
						betav = vector(rk1gtd_xx, i, j, k) * vector(rk1Betau_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk1gtd_yy, i, j, k) * vector(rk1Betau_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk1gtd_zz, i, j, k) * vector(rk1Betau_z, i, j, k) * vector(vfd_z, i, j, k) + vector(rk1gtd_xy, i, j, k) * vector(rk1Betau_x, i, j, k) * vector(vfd_y, i, j, k) + vector(rk1gtd_xy, i, j, k) * vector(rk1Betau_y, i, j, k) * vector(vfd_x, i, j, k) + vector(rk1gtd_xz, i, j, k) * vector(rk1Betau_x, i, j, k) * vector(vfd_z, i, j, k) + vector(rk1gtd_xz, i, j, k) * vector(rk1Betau_z, i, j, k) * vector(vfd_x, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Betau_y, i, j, k) * vector(vfd_z, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Betau_z, i, j, k) * vector(vfd_y, i, j, k);
						velr = vector(vfd_x, i, j, k) * xcoord(i) + vector(vfd_y, i, j, k) * ycoord(j) + vector(vfd_z, i, j, k) * zcoord(k);
						velr = velr / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k) + 0.01);
						f_redshift = rk1sqW * (vector(rk1Alpha, i, j, k) - betav) * (1.0 + velr);
						f_redshift = f_redshift * sqrt(vector(rk1Alpha, i, j, k) * vector(rk1Alpha, i, j, k) - betabeta);
						leakage(do_leakage, vacuum_rho, threshold_leakage_vacuum, energyShift, optdepthe_tmp, optdeptha_tmp, optdepthx_tmp, qnu_e_tmp, qnu_a_tmp, qnu_x_tmp, chie_tmp, chia_tmp, chix_tmp, qnu_tmp, rnu_tmp, vector(rhof, i, j, k), vector(epsf, i, j, k), vector(Yef, i, j, k), vector(Tf, i, j, k), rk1sdetg, rk1optdeptheh, rk1optdepthah, rk1optdepthxh, rk1optdepthev, rk1optdepthav, rk1optdepthxv, rk1optdepthed, rk1optdepthad, rk1optdepthxd, rk1chieh, rk1chiah, rk1chixh, rk1chiev, rk1chiav, rk1chixv, rk1chied, rk1chiad, rk1chixd, f_redshift, coord0_data_DataLeakage, coord0_dx_DataLeakage, coord0_size_DataLeakage, coord1_data_DataLeakage, coord1_dx_DataLeakage, coord1_size_DataLeakage, coord2_data_DataLeakage, coord2_dx_DataLeakage, coord2_size_DataLeakage, vars_data_DataLeakage, vars_size_DataLeakage, dx, simPlat_dt, ilast, jlast);
						vector(optdepthe, i, j, k) = optdepthe_tmp;
						vector(optdeptha, i, j, k) = optdeptha_tmp;
						vector(optdepthx, i, j, k) = optdepthx_tmp;
						vector(chie, i, j, k) = chie_tmp;
						vector(chia, i, j, k) = chia_tmp;
						vector(chix, i, j, k) = chix_tmp;
						vector(qnu_e, i, j, k) = qnu_e_tmp;
						vector(qnu_a, i, j, k) = qnu_a_tmp;
						vector(qnu_x, i, j, k) = qnu_x_tmp;
						vector(qnu, i, j, k) = qnu_tmp;
						vector(rnu, i, j, k) = rnu_tmp;
					}
				}
			}
		}
	}
	//Fill ghosts and periodical boundaries
	d_bdry_sched_advance5[ln]->fillData(current_time, false);
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		//Hard region field distance variables
		double* d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
	
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
		double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
		double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
		double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
		double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
		double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
		double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
		double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
		double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
		double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
		double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
		double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
		double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
		double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
		double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
		double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
		double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
		double* rk1gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xx_id).get())->getPointer();
		double* rk1gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xy_id).get())->getPointer();
		double* rk1gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xz_id).get())->getPointer();
		double* rk1gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yy_id).get())->getPointer();
		double* rk1gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yz_id).get())->getPointer();
		double* rk1gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_zz_id).get())->getPointer();
		double* rk1Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xx_id).get())->getPointer();
		double* rk1Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xy_id).get())->getPointer();
		double* rk1Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xz_id).get())->getPointer();
		double* rk1Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_yy_id).get())->getPointer();
		double* rk1Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_yz_id).get())->getPointer();
		double* rk1Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_zz_id).get())->getPointer();
		double* rk1Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_x_id).get())->getPointer();
		double* rk1Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_y_id).get())->getPointer();
		double* rk1Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_z_id).get())->getPointer();
		double* rk1Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_x_id).get())->getPointer();
		double* rk1Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_y_id).get())->getPointer();
		double* rk1Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_z_id).get())->getPointer();
		double* rk1Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Alpha_id).get())->getPointer();
		double* rk1chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1chi_id).get())->getPointer();
		double* rk1trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1trK_id).get())->getPointer();
		double* rk1theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1theta_id).get())->getPointer();
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk1theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
	}
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		double* rk1chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1chi_id).get())->getPointer();
		double* rk1gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xy_id).get())->getPointer();
		double* rk1Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_x_id).get())->getPointer();
		double* rk1gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yy_id).get())->getPointer();
		double* rk1Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_y_id).get())->getPointer();
		double* rk1gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yz_id).get())->getPointer();
		double* rk1Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_z_id).get())->getPointer();
		double* rk1gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xz_id).get())->getPointer();
		double* rk1gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_zz_id).get())->getPointer();
		double* rk1gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xx_id).get())->getPointer();
		double* rk1Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_x_id).get())->getPointer();
		double* rk1Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_y_id).get())->getPointer();
		double* rk1Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_z_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* rk1Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_x_id).get())->getPointer();
		double* rk1Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Alpha_id).get())->getPointer();
		double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
		double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
		double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
		double* rk1phif = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1phif_id).get())->getPointer();
		double* rk1Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_y_id).get())->getPointer();
		double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
		double* rk1Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_z_id).get())->getPointer();
		double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
		double* rk1Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Df_id).get())->getPointer();
		double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
		double* rk1DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1DYf_id).get())->getPointer();
		double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
		double* rk1tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1tauf_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
		double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
		double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
		double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
		double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
		double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
		double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
		double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
		double* rk1Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_x_id).get())->getPointer();
		double* rk1Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_y_id).get())->getPointer();
		double* rk1Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_z_id).get())->getPointer();
		double* rk1Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xx_id).get())->getPointer();
		double* rk1Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xy_id).get())->getPointer();
		double* rk1Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xz_id).get())->getPointer();
		double* rk1Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_yy_id).get())->getPointer();
		double* rk1Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_yz_id).get())->getPointer();
		double* rk1Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_zz_id).get())->getPointer();
		double* rk1trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1trK_id).get())->getPointer();
		double* rk1theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1theta_id).get())->getPointer();
		double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
		double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
		double* rk2Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_x_id).get())->getPointer();
		double* Sfd_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_p_id).get())->getPointer();
		double* rk2Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_y_id).get())->getPointer();
		double* Sfd_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_p_id).get())->getPointer();
		double* rk2Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_z_id).get())->getPointer();
		double* Sfd_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_p_id).get())->getPointer();
		double* rk2Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_x_id).get())->getPointer();
		double* Bfu_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_p_id).get())->getPointer();
		double* rk2Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_y_id).get())->getPointer();
		double* Bfu_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_p_id).get())->getPointer();
		double* rk2Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_z_id).get())->getPointer();
		double* Bfu_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_p_id).get())->getPointer();
		double* rk2Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Df_id).get())->getPointer();
		double* Df_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_p_id).get())->getPointer();
		double* rk2DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2DYf_id).get())->getPointer();
		double* DYf_p = ((pdat::NodeData<double> *) patch->getPatchData(d_DYf_p_id).get())->getPointer();
		double* rk2tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2tauf_id).get())->getPointer();
		double* tauf_p = ((pdat::NodeData<double> *) patch->getPatchData(d_tauf_p_id).get())->getPointer();
		double* rk2phif = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2phif_id).get())->getPointer();
		double* phif_p = ((pdat::NodeData<double> *) patch->getPatchData(d_phif_p_id).get())->getPointer();
		double* rk2gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xx_id).get())->getPointer();
		double* gtd_xx_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_p_id).get())->getPointer();
		double* rk2gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xy_id).get())->getPointer();
		double* gtd_xy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_p_id).get())->getPointer();
		double* rk2gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xz_id).get())->getPointer();
		double* gtd_xz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_p_id).get())->getPointer();
		double* rk2gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yy_id).get())->getPointer();
		double* gtd_yy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_p_id).get())->getPointer();
		double* rk2gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yz_id).get())->getPointer();
		double* gtd_yz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_p_id).get())->getPointer();
		double* rk2gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_zz_id).get())->getPointer();
		double* gtd_zz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_p_id).get())->getPointer();
		double* rk2Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xx_id).get())->getPointer();
		double* Atd_xx_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xx_p_id).get())->getPointer();
		double* rk2Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xy_id).get())->getPointer();
		double* Atd_xy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xy_p_id).get())->getPointer();
		double* rk2Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xz_id).get())->getPointer();
		double* Atd_xz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xz_p_id).get())->getPointer();
		double* rk2Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_yy_id).get())->getPointer();
		double* Atd_yy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yy_p_id).get())->getPointer();
		double* rk2Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_yz_id).get())->getPointer();
		double* Atd_yz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yz_p_id).get())->getPointer();
		double* rk2Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_zz_id).get())->getPointer();
		double* Atd_zz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_zz_p_id).get())->getPointer();
		double* rk2Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_x_id).get())->getPointer();
		double* Gamh_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_x_p_id).get())->getPointer();
		double* rk2Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_y_id).get())->getPointer();
		double* Gamh_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_y_p_id).get())->getPointer();
		double* rk2Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_z_id).get())->getPointer();
		double* Gamh_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_z_p_id).get())->getPointer();
		double* rk2Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_x_id).get())->getPointer();
		double* Betau_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_p_id).get())->getPointer();
		double* rk2Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_y_id).get())->getPointer();
		double* Betau_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_p_id).get())->getPointer();
		double* rk2Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_z_id).get())->getPointer();
		double* Betau_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_p_id).get())->getPointer();
		double* rk2Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Alpha_id).get())->getPointer();
		double* Alpha_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_p_id).get())->getPointer();
		double* rk2chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2chi_id).get())->getPointer();
		double* chi_p = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_p_id).get())->getPointer();
		double* rk2trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2trK_id).get())->getPointer();
		double* trK_p = ((pdat::NodeData<double> *) patch->getPatchData(d_trK_p_id).get())->getPointer();
		double* rk2theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2theta_id).get())->getPointer();
		double* theta_p = ((pdat::NodeData<double> *) patch->getPatchData(d_theta_p_id).get())->getPointer();
		double rk1chi_max, rk1sdetg, rk1inv_chi, rk1Bfd_y, rk1Bfd_z, rk1Bfd_x, rk1sqBf, rk1detgtd, rk1idetgtd, rk1gtu_zz, rk1gtu_xx, rk1gtu_xz, rk1gtu_xy, rk1Sfu_x, rk1vfu_x, rk1Bfvf, rk1gtu_yy, rk1gtu_yz, rk1vfu_z, rk1Sfu_y, rk1Sfu_z, rk1vfu_y, rk1sqvf, rk1sqW, rk1Sfuu_zz, rk1Sfuu_xx, rk1Sfuu_xz, rk1Sfuu_xy, rk1Sfud_xx, rk1Sfud_xy, rk1Sfud_xz, rk1Sfuu_yz, rk1Sfud_zy, rk1Sfud_zz, rk1Sfud_zx, rk1Sfuu_yy, rk1Sfud_yx, rk1Sfud_yy, rk1Sfud_yz, FluxiSfd_x, FluxiSfd_y, FluxiSfd_z, FluxiBfu_x, FluxiBfu_y, FluxiBfu_z, FluxiDf, FluxiDYf, Fluxitauf, Fluxiphif, rk1h, rk1sqbc, rk1sqca, rk1sqa, SpeediSfd_x, SpeediSfd_y, SpeediSfd_z, SpeediBfu_x, SpeediBfu_y, SpeediBfu_z, SpeediDf, SpeediDYf, Speeditauf, Speediphif, Fluxm3iSfd_x, Fluxm3iSfd_y, Fluxm3iSfd_z, Fluxm3iBfu_x, Fluxm3iBfu_y, Fluxm3iBfu_z, Fluxm3iDf, Fluxm3iDYf, Fluxm3itauf, Fluxm3iphif, Fluxm2iSfd_x, Fluxm2iSfd_y, Fluxm2iSfd_z, Fluxm2iBfu_x, Fluxm2iBfu_y, Fluxm2iBfu_z, Fluxm2iDf, Fluxm2iDYf, Fluxm2itauf, Fluxm2iphif, Fluxm1iSfd_x, Fluxm1iSfd_y, Fluxm1iSfd_z, Fluxm1iBfu_x, Fluxm1iBfu_y, Fluxm1iBfu_z, Fluxm1iDf, Fluxm1iDYf, Fluxm1itauf, Fluxm1iphif, Fluxp1iSfd_x, Fluxp1iSfd_y, Fluxp1iSfd_z, Fluxp1iBfu_x, Fluxp1iBfu_y, Fluxp1iBfu_z, Fluxp1iDf, Fluxp1iDYf, Fluxp1itauf, Fluxp1iphif, Fluxp2iSfd_x, Fluxp2iSfd_y, Fluxp2iSfd_z, Fluxp2iBfu_x, Fluxp2iBfu_y, Fluxp2iBfu_z, Fluxp2iDf, Fluxp2iDYf, Fluxp2itauf, Fluxp2iphif, Fluxp3iSfd_x, Fluxp3iSfd_y, Fluxp3iSfd_z, Fluxp3iBfu_x, Fluxp3iBfu_y, Fluxp3iBfu_z, Fluxp3iDf, Fluxp3iDYf, Fluxp3itauf, Fluxp3iphif, LLFpiSfd_x, LLFpiSfd_y, LLFpiSfd_z, LLFpiBfu_x, LLFpiBfu_y, LLFpiBfu_z, LLFpiDf, LLFpiDYf, LLFpitauf, LLFpiphif, LLFniSfd_x, LLFniSfd_y, LLFniSfd_z, LLFniBfu_x, LLFniBfu_y, LLFniBfu_z, LLFniDf, LLFniDYf, LLFnitauf, LLFniphif, LLFp_m3iSfd_x, LLFp_m3iSfd_y, LLFp_m3iSfd_z, LLFp_m3iBfu_x, LLFp_m3iBfu_y, LLFp_m3iBfu_z, LLFp_m3iDf, LLFp_m3iDYf, LLFp_m3itauf, LLFp_m3iphif, LLFn_m3iSfd_x, LLFn_m3iSfd_y, LLFn_m3iSfd_z, LLFn_m3iBfu_x, LLFn_m3iBfu_y, LLFn_m3iBfu_z, LLFn_m3iDf, LLFn_m3iDYf, LLFn_m3itauf, LLFn_m3iphif, LLFp_m2iSfd_x, LLFp_m2iSfd_y, LLFp_m2iSfd_z, LLFp_m2iBfu_x, LLFp_m2iBfu_y, LLFp_m2iBfu_z, LLFp_m2iDf, LLFp_m2iDYf, LLFp_m2itauf, LLFp_m2iphif, LLFn_m2iSfd_x, LLFn_m2iSfd_y, LLFn_m2iSfd_z, LLFn_m2iBfu_x, LLFn_m2iBfu_y, LLFn_m2iBfu_z, LLFn_m2iDf, LLFn_m2iDYf, LLFn_m2itauf, LLFn_m2iphif, LLFp_m1iSfd_x, LLFp_m1iSfd_y, LLFp_m1iSfd_z, LLFp_m1iBfu_x, LLFp_m1iBfu_y, LLFp_m1iBfu_z, LLFp_m1iDf, LLFp_m1iDYf, LLFp_m1itauf, LLFp_m1iphif, LLFn_m1iSfd_x, LLFn_m1iSfd_y, LLFn_m1iSfd_z, LLFn_m1iBfu_x, LLFn_m1iBfu_y, LLFn_m1iBfu_z, LLFn_m1iDf, LLFn_m1iDYf, LLFn_m1itauf, LLFn_m1iphif, LLFp_p1iSfd_x, LLFp_p1iSfd_y, LLFp_p1iSfd_z, LLFp_p1iBfu_x, LLFp_p1iBfu_y, LLFp_p1iBfu_z, LLFp_p1iDf, LLFp_p1iDYf, LLFp_p1itauf, LLFp_p1iphif, LLFn_p1iSfd_x, LLFn_p1iSfd_y, LLFn_p1iSfd_z, LLFn_p1iBfu_x, LLFn_p1iBfu_y, LLFn_p1iBfu_z, LLFn_p1iDf, LLFn_p1iDYf, LLFn_p1itauf, LLFn_p1iphif, LLFp_p2iSfd_x, LLFp_p2iSfd_y, LLFp_p2iSfd_z, LLFp_p2iBfu_x, LLFp_p2iBfu_y, LLFp_p2iBfu_z, LLFp_p2iDf, LLFp_p2iDYf, LLFp_p2itauf, LLFp_p2iphif, LLFn_p2iSfd_x, LLFn_p2iSfd_y, LLFn_p2iSfd_z, LLFn_p2iBfu_x, LLFn_p2iBfu_y, LLFn_p2iBfu_z, LLFn_p2iDf, LLFn_p2iDYf, LLFn_p2itauf, LLFn_p2iphif, LLFp_p3iSfd_x, LLFp_p3iSfd_y, LLFp_p3iSfd_z, LLFp_p3iBfu_x, LLFp_p3iBfu_y, LLFp_p3iBfu_z, LLFp_p3iDf, LLFp_p3iDYf, LLFp_p3itauf, LLFp_p3iphif, LLFn_p3iSfd_x, LLFn_p3iSfd_y, LLFn_p3iSfd_z, LLFn_p3iBfu_x, LLFn_p3iBfu_y, LLFn_p3iBfu_z, LLFn_p3iDf, LLFn_p3iDYf, LLFn_p3itauf, LLFn_p3iphif, FluxjSfd_x, FluxjSfd_y, FluxjSfd_z, FluxjBfu_x, FluxjBfu_y, FluxjBfu_z, FluxjDf, FluxjDYf, Fluxjtauf, Fluxjphif, SpeedjSfd_x, SpeedjSfd_y, SpeedjSfd_z, SpeedjBfu_x, SpeedjBfu_y, SpeedjBfu_z, SpeedjDf, SpeedjDYf, Speedjtauf, Speedjphif, Fluxm3jSfd_x, Fluxm3jSfd_y, Fluxm3jSfd_z, Fluxm3jBfu_x, Fluxm3jBfu_y, Fluxm3jBfu_z, Fluxm3jDf, Fluxm3jDYf, Fluxm3jtauf, Fluxm3jphif, Fluxm2jSfd_x, Fluxm2jSfd_y, Fluxm2jSfd_z, Fluxm2jBfu_x, Fluxm2jBfu_y, Fluxm2jBfu_z, Fluxm2jDf, Fluxm2jDYf, Fluxm2jtauf, Fluxm2jphif, Fluxm1jSfd_x, Fluxm1jSfd_y, Fluxm1jSfd_z, Fluxm1jBfu_x, Fluxm1jBfu_y, Fluxm1jBfu_z, Fluxm1jDf, Fluxm1jDYf, Fluxm1jtauf, Fluxm1jphif, Fluxp1jSfd_x, Fluxp1jSfd_y, Fluxp1jSfd_z, Fluxp1jBfu_x, Fluxp1jBfu_y, Fluxp1jBfu_z, Fluxp1jDf, Fluxp1jDYf, Fluxp1jtauf, Fluxp1jphif, Fluxp2jSfd_x, Fluxp2jSfd_y, Fluxp2jSfd_z, Fluxp2jBfu_x, Fluxp2jBfu_y, Fluxp2jBfu_z, Fluxp2jDf, Fluxp2jDYf, Fluxp2jtauf, Fluxp2jphif, Fluxp3jSfd_x, Fluxp3jSfd_y, Fluxp3jSfd_z, Fluxp3jBfu_x, Fluxp3jBfu_y, Fluxp3jBfu_z, Fluxp3jDf, Fluxp3jDYf, Fluxp3jtauf, Fluxp3jphif, LLFpjSfd_x, LLFpjSfd_y, LLFpjSfd_z, LLFpjBfu_x, LLFpjBfu_y, LLFpjBfu_z, LLFpjDf, LLFpjDYf, LLFpjtauf, LLFpjphif, LLFnjSfd_x, LLFnjSfd_y, LLFnjSfd_z, LLFnjBfu_x, LLFnjBfu_y, LLFnjBfu_z, LLFnjDf, LLFnjDYf, LLFnjtauf, LLFnjphif, LLFp_m3jSfd_x, LLFp_m3jSfd_y, LLFp_m3jSfd_z, LLFp_m3jBfu_x, LLFp_m3jBfu_y, LLFp_m3jBfu_z, LLFp_m3jDf, LLFp_m3jDYf, LLFp_m3jtauf, LLFp_m3jphif, LLFn_m3jSfd_x, LLFn_m3jSfd_y, LLFn_m3jSfd_z, LLFn_m3jBfu_x, LLFn_m3jBfu_y, LLFn_m3jBfu_z, LLFn_m3jDf, LLFn_m3jDYf, LLFn_m3jtauf, LLFn_m3jphif, LLFp_m2jSfd_x, LLFp_m2jSfd_y, LLFp_m2jSfd_z, LLFp_m2jBfu_x, LLFp_m2jBfu_y, LLFp_m2jBfu_z, LLFp_m2jDf, LLFp_m2jDYf, LLFp_m2jtauf, LLFp_m2jphif, LLFn_m2jSfd_x, LLFn_m2jSfd_y, LLFn_m2jSfd_z, LLFn_m2jBfu_x, LLFn_m2jBfu_y, LLFn_m2jBfu_z, LLFn_m2jDf, LLFn_m2jDYf, LLFn_m2jtauf, LLFn_m2jphif, LLFp_m1jSfd_x, LLFp_m1jSfd_y, LLFp_m1jSfd_z, LLFp_m1jBfu_x, LLFp_m1jBfu_y, LLFp_m1jBfu_z, LLFp_m1jDf, LLFp_m1jDYf, LLFp_m1jtauf, LLFp_m1jphif, LLFn_m1jSfd_x, LLFn_m1jSfd_y, LLFn_m1jSfd_z, LLFn_m1jBfu_x, LLFn_m1jBfu_y, LLFn_m1jBfu_z, LLFn_m1jDf, LLFn_m1jDYf, LLFn_m1jtauf, LLFn_m1jphif, LLFp_p1jSfd_x, LLFp_p1jSfd_y, LLFp_p1jSfd_z, LLFp_p1jBfu_x, LLFp_p1jBfu_y, LLFp_p1jBfu_z, LLFp_p1jDf, LLFp_p1jDYf, LLFp_p1jtauf, LLFp_p1jphif, LLFn_p1jSfd_x, LLFn_p1jSfd_y, LLFn_p1jSfd_z, LLFn_p1jBfu_x, LLFn_p1jBfu_y, LLFn_p1jBfu_z, LLFn_p1jDf, LLFn_p1jDYf, LLFn_p1jtauf, LLFn_p1jphif, LLFp_p2jSfd_x, LLFp_p2jSfd_y, LLFp_p2jSfd_z, LLFp_p2jBfu_x, LLFp_p2jBfu_y, LLFp_p2jBfu_z, LLFp_p2jDf, LLFp_p2jDYf, LLFp_p2jtauf, LLFp_p2jphif, LLFn_p2jSfd_x, LLFn_p2jSfd_y, LLFn_p2jSfd_z, LLFn_p2jBfu_x, LLFn_p2jBfu_y, LLFn_p2jBfu_z, LLFn_p2jDf, LLFn_p2jDYf, LLFn_p2jtauf, LLFn_p2jphif, LLFp_p3jSfd_x, LLFp_p3jSfd_y, LLFp_p3jSfd_z, LLFp_p3jBfu_x, LLFp_p3jBfu_y, LLFp_p3jBfu_z, LLFp_p3jDf, LLFp_p3jDYf, LLFp_p3jtauf, LLFp_p3jphif, LLFn_p3jSfd_x, LLFn_p3jSfd_y, LLFn_p3jSfd_z, LLFn_p3jBfu_x, LLFn_p3jBfu_y, LLFn_p3jBfu_z, LLFn_p3jDf, LLFn_p3jDYf, LLFn_p3jtauf, LLFn_p3jphif, FluxkSfd_x, FluxkSfd_y, FluxkSfd_z, FluxkBfu_x, FluxkBfu_y, FluxkBfu_z, FluxkDf, FluxkDYf, Fluxktauf, Fluxkphif, SpeedkSfd_x, SpeedkSfd_y, SpeedkSfd_z, SpeedkBfu_x, SpeedkBfu_y, SpeedkBfu_z, SpeedkDf, SpeedkDYf, Speedktauf, Speedkphif, Fluxm3kSfd_x, Fluxm3kSfd_y, Fluxm3kSfd_z, Fluxm3kBfu_x, Fluxm3kBfu_y, Fluxm3kBfu_z, Fluxm3kDf, Fluxm3kDYf, Fluxm3ktauf, Fluxm3kphif, Fluxm2kSfd_x, Fluxm2kSfd_y, Fluxm2kSfd_z, Fluxm2kBfu_x, Fluxm2kBfu_y, Fluxm2kBfu_z, Fluxm2kDf, Fluxm2kDYf, Fluxm2ktauf, Fluxm2kphif, Fluxm1kSfd_x, Fluxm1kSfd_y, Fluxm1kSfd_z, Fluxm1kBfu_x, Fluxm1kBfu_y, Fluxm1kBfu_z, Fluxm1kDf, Fluxm1kDYf, Fluxm1ktauf, Fluxm1kphif, Fluxp1kSfd_x, Fluxp1kSfd_y, Fluxp1kSfd_z, Fluxp1kBfu_x, Fluxp1kBfu_y, Fluxp1kBfu_z, Fluxp1kDf, Fluxp1kDYf, Fluxp1ktauf, Fluxp1kphif, Fluxp2kSfd_x, Fluxp2kSfd_y, Fluxp2kSfd_z, Fluxp2kBfu_x, Fluxp2kBfu_y, Fluxp2kBfu_z, Fluxp2kDf, Fluxp2kDYf, Fluxp2ktauf, Fluxp2kphif, Fluxp3kSfd_x, Fluxp3kSfd_y, Fluxp3kSfd_z, Fluxp3kBfu_x, Fluxp3kBfu_y, Fluxp3kBfu_z, Fluxp3kDf, Fluxp3kDYf, Fluxp3ktauf, Fluxp3kphif, LLFpkSfd_x, LLFpkSfd_y, LLFpkSfd_z, LLFpkBfu_x, LLFpkBfu_y, LLFpkBfu_z, LLFpkDf, LLFpkDYf, LLFpktauf, LLFpkphif, LLFnkSfd_x, LLFnkSfd_y, LLFnkSfd_z, LLFnkBfu_x, LLFnkBfu_y, LLFnkBfu_z, LLFnkDf, LLFnkDYf, LLFnktauf, LLFnkphif, LLFp_m3kSfd_x, LLFp_m3kSfd_y, LLFp_m3kSfd_z, LLFp_m3kBfu_x, LLFp_m3kBfu_y, LLFp_m3kBfu_z, LLFp_m3kDf, LLFp_m3kDYf, LLFp_m3ktauf, LLFp_m3kphif, LLFn_m3kSfd_x, LLFn_m3kSfd_y, LLFn_m3kSfd_z, LLFn_m3kBfu_x, LLFn_m3kBfu_y, LLFn_m3kBfu_z, LLFn_m3kDf, LLFn_m3kDYf, LLFn_m3ktauf, LLFn_m3kphif, LLFp_m2kSfd_x, LLFp_m2kSfd_y, LLFp_m2kSfd_z, LLFp_m2kBfu_x, LLFp_m2kBfu_y, LLFp_m2kBfu_z, LLFp_m2kDf, LLFp_m2kDYf, LLFp_m2ktauf, LLFp_m2kphif, LLFn_m2kSfd_x, LLFn_m2kSfd_y, LLFn_m2kSfd_z, LLFn_m2kBfu_x, LLFn_m2kBfu_y, LLFn_m2kBfu_z, LLFn_m2kDf, LLFn_m2kDYf, LLFn_m2ktauf, LLFn_m2kphif, LLFp_m1kSfd_x, LLFp_m1kSfd_y, LLFp_m1kSfd_z, LLFp_m1kBfu_x, LLFp_m1kBfu_y, LLFp_m1kBfu_z, LLFp_m1kDf, LLFp_m1kDYf, LLFp_m1ktauf, LLFp_m1kphif, LLFn_m1kSfd_x, LLFn_m1kSfd_y, LLFn_m1kSfd_z, LLFn_m1kBfu_x, LLFn_m1kBfu_y, LLFn_m1kBfu_z, LLFn_m1kDf, LLFn_m1kDYf, LLFn_m1ktauf, LLFn_m1kphif, LLFp_p1kSfd_x, LLFp_p1kSfd_y, LLFp_p1kSfd_z, LLFp_p1kBfu_x, LLFp_p1kBfu_y, LLFp_p1kBfu_z, LLFp_p1kDf, LLFp_p1kDYf, LLFp_p1ktauf, LLFp_p1kphif, LLFn_p1kSfd_x, LLFn_p1kSfd_y, LLFn_p1kSfd_z, LLFn_p1kBfu_x, LLFn_p1kBfu_y, LLFn_p1kBfu_z, LLFn_p1kDf, LLFn_p1kDYf, LLFn_p1ktauf, LLFn_p1kphif, LLFp_p2kSfd_x, LLFp_p2kSfd_y, LLFp_p2kSfd_z, LLFp_p2kBfu_x, LLFp_p2kBfu_y, LLFp_p2kBfu_z, LLFp_p2kDf, LLFp_p2kDYf, LLFp_p2ktauf, LLFp_p2kphif, LLFn_p2kSfd_x, LLFn_p2kSfd_y, LLFn_p2kSfd_z, LLFn_p2kBfu_x, LLFn_p2kBfu_y, LLFn_p2kBfu_z, LLFn_p2kDf, LLFn_p2kDYf, LLFn_p2ktauf, LLFn_p2kphif, LLFp_p3kSfd_x, LLFp_p3kSfd_y, LLFp_p3kSfd_z, LLFp_p3kBfu_x, LLFp_p3kBfu_y, LLFp_p3kBfu_z, LLFp_p3kDf, LLFp_p3kDYf, LLFp_p3ktauf, LLFp_p3kphif, LLFn_p3kSfd_x, LLFn_p3kSfd_y, LLFn_p3kSfd_z, LLFn_p3kBfu_x, LLFn_p3kBfu_y, LLFn_p3kBfu_z, LLFn_p3kDf, LLFn_p3kDYf, LLFn_p3ktauf, LLFn_p3kphif, Sfd_x_lni, Sfd_y_lni, Sfd_z_lni, Bfu_x_lni, Bfu_y_lni, Bfu_z_lni, Df_lni, DYf_lni, tauf_lni, phif_lni, Sfd_x_lnj, Sfd_y_lnj, Sfd_z_lnj, Bfu_x_lnj, Bfu_y_lnj, Bfu_z_lnj, Df_lnj, DYf_lnj, tauf_lnj, phif_lnj, Sfd_x_lnk, Sfd_y_lnk, Sfd_z_lnk, Bfu_x_lnk, Bfu_y_lnk, Bfu_z_lnk, Df_lnk, DYf_lnk, tauf_lnk, phif_lnk, Sfd_x_rni, Sfd_y_rni, Sfd_z_rni, Bfu_x_rni, Bfu_y_rni, Bfu_z_rni, Df_rni, DYf_rni, tauf_rni, phif_rni, Sfd_x_rnj, Sfd_y_rnj, Sfd_z_rnj, Bfu_x_rnj, Bfu_y_rnj, Bfu_z_rnj, Df_rnj, DYf_rnj, tauf_rnj, phif_rnj, Sfd_x_rnk, Sfd_y_rnk, Sfd_z_rnk, Bfu_x_rnk, Bfu_y_rnk, Bfu_z_rnk, Df_rnk, DYf_rnk, tauf_rnk, phif_rnk, Sfd_x_lpi, Sfd_y_lpi, Sfd_z_lpi, Bfu_x_lpi, Bfu_y_lpi, Bfu_z_lpi, Df_lpi, DYf_lpi, tauf_lpi, phif_lpi, Sfd_x_lpj, Sfd_y_lpj, Sfd_z_lpj, Bfu_x_lpj, Bfu_y_lpj, Bfu_z_lpj, Df_lpj, DYf_lpj, tauf_lpj, phif_lpj, Sfd_x_lpk, Sfd_y_lpk, Sfd_z_lpk, Bfu_x_lpk, Bfu_y_lpk, Bfu_z_lpk, Df_lpk, DYf_lpk, tauf_lpk, phif_lpk, Sfd_x_rpi, Sfd_y_rpi, Sfd_z_rpi, Bfu_x_rpi, Bfu_y_rpi, Bfu_z_rpi, Df_rpi, DYf_rpi, tauf_rpi, phif_rpi, Sfd_x_rpj, Sfd_y_rpj, Sfd_z_rpj, Bfu_x_rpj, Bfu_y_rpj, Bfu_z_rpj, Df_rpj, DYf_rpj, tauf_rpj, phif_rpj, Sfd_x_rpk, Sfd_y_rpk, Sfd_z_rpk, Bfu_x_rpk, Bfu_y_rpk, Bfu_z_rpk, Df_rpk, DYf_rpk, tauf_rpk, phif_rpk, FluxniSfd_x, FluxniSfd_y, FluxniSfd_z, FluxniBfu_x, FluxniBfu_y, FluxniBfu_z, FluxniDf, FluxniDYf, Fluxnitauf, Fluxniphif, FluxnjSfd_x, FluxnjSfd_y, FluxnjSfd_z, FluxnjBfu_x, FluxnjBfu_y, FluxnjBfu_z, FluxnjDf, FluxnjDYf, Fluxnjtauf, Fluxnjphif, FluxnkSfd_x, FluxnkSfd_y, FluxnkSfd_z, FluxnkBfu_x, FluxnkBfu_y, FluxnkBfu_z, FluxnkDf, FluxnkDYf, Fluxnktauf, Fluxnkphif, FluxpiSfd_x, FluxpiSfd_y, FluxpiSfd_z, FluxpiBfu_x, FluxpiBfu_y, FluxpiBfu_z, FluxpiDf, FluxpiDYf, Fluxpitauf, Fluxpiphif, FluxpjSfd_x, FluxpjSfd_y, FluxpjSfd_z, FluxpjBfu_x, FluxpjBfu_y, FluxpjBfu_z, FluxpjDf, FluxpjDYf, Fluxpjtauf, Fluxpjphif, FluxpkSfd_x, FluxpkSfd_y, FluxpkSfd_z, FluxpkBfu_x, FluxpkBfu_y, FluxpkBfu_z, FluxpkDf, FluxpkDYf, Fluxpktauf, Fluxpkphif, d_div_Beta_o0_t0_m0_l0, d_Betau_x_o1_t0_m0_l0, d_Betau_x_o2_t0_m0_l0, d_Sfd_y_o0_t7_m0_l0, d_Betau_x_o1_t1_m0_l0, d_Betau_x_o2_t1_m0_l0, d_Sfd_z_o0_t7_m0_l0, d_Betau_x_o1_t2_m0_l0, d_Betau_x_o2_t2_m0_l0, d_d_div_Beta_x_o0_t0_m0_l0, d_d_div_Beta_y_o0_t0_m0_l0, d_d_div_Beta_z_o0_t0_m0_l0, d_Gamh_x_o0_t15_m0_l0, d_Gamh_x_o0_t16_m0_l0, d_Gamh_x_o0_t17_m0_l0, d_Sfd_x_o0_t8_m0_l0, d_Betau_y_o1_t0_m0_l0, d_Betau_y_o2_t0_m0_l0, d_div_Beta_o0_t1_m0_l0, d_Betau_y_o1_t1_m0_l0, d_Betau_y_o2_t1_m0_l0, d_Sfd_z_o0_t8_m0_l0, d_Betau_y_o1_t2_m0_l0, d_Betau_y_o2_t2_m0_l0, d_Gamh_y_o0_t6_m0_l0, d_d_div_Beta_x_o0_t1_m0_l0, d_Gamh_y_o0_t8_m0_l0, d_d_div_Beta_y_o0_t1_m0_l0, d_d_div_Beta_z_o0_t1_m0_l0, d_Gamh_y_o0_t11_m0_l0, d_Sfd_x_o0_t9_m0_l0, d_Betau_z_o1_t0_m0_l0, d_Betau_z_o2_t0_m0_l0, d_Sfd_y_o0_t9_m0_l0, d_Betau_z_o1_t1_m0_l0, d_Betau_z_o2_t1_m0_l0, d_div_Beta_o0_t2_m0_l0, d_Betau_z_o1_t2_m0_l0, d_Betau_z_o2_t2_m0_l0, d_Gamh_z_o0_t0_m0_l0, d_Gamh_z_o0_t1_m0_l0, d_d_div_Beta_x_o0_t2_m0_l0, d_Gamh_z_o0_t3_m0_l0, d_d_div_Beta_y_o0_t2_m0_l0, d_d_div_Beta_z_o0_t2_m0_l0, d_Ctd_xxx_o0_t0_m0_l0, d_gtd_xx_o1_t0_m0_l0, d_gtd_xx_o2_t0_m0_l0, d_Ctd_xxy_o0_t0_m0_l0, d_gtd_xx_o1_t1_m0_l0, d_gtd_xx_o2_t1_m0_l0, d_Ctd_xxz_o0_t0_m0_l0, d_gtd_xx_o1_t2_m0_l0, d_gtd_xx_o2_t2_m0_l0, d_Ctd_yxx_o0_t0_m0_l0, d_gtd_xy_o1_t0_m0_l0, d_gtd_xy_o2_t0_m0_l0, d_Ctd_xyy_o0_t0_m0_l0, d_gtd_xy_o1_t1_m0_l0, d_gtd_xy_o2_t1_m0_l0, d_Ctd_xyz_o0_t1_m0_l0, d_gtd_xy_o1_t2_m0_l0, d_gtd_xy_o2_t2_m0_l0, d_Ctd_xyy_o0_t1_m0_l0, d_gtd_yy_o1_t0_m0_l0, d_gtd_yy_o2_t0_m0_l0, d_Ctd_yyy_o0_t0_m0_l0, d_gtd_yy_o1_t1_m0_l0, d_gtd_yy_o2_t1_m0_l0, d_Ctd_yyz_o0_t0_m0_l0, d_gtd_yy_o1_t2_m0_l0, d_gtd_yy_o2_t2_m0_l0, d_Ctd_zxx_o0_t0_m0_l0, d_gtd_xz_o1_t0_m0_l0, d_gtd_xz_o2_t0_m0_l0, d_Ctd_xyz_o0_t0_m0_l0, d_gtd_xz_o1_t1_m0_l0, d_gtd_xz_o2_t1_m0_l0, d_Ctd_xzz_o0_t0_m0_l0, d_gtd_xz_o1_t2_m0_l0, d_gtd_xz_o2_t2_m0_l0, d_Ctd_xyz_o0_t2_m0_l0, d_gtd_yz_o1_t0_m0_l0, d_gtd_yz_o2_t0_m0_l0, d_Ctd_zyy_o0_t0_m0_l0, d_gtd_yz_o1_t1_m0_l0, d_gtd_yz_o2_t1_m0_l0, d_Ctd_yzz_o0_t0_m0_l0, d_gtd_yz_o1_t2_m0_l0, d_gtd_yz_o2_t2_m0_l0, d_Ctd_xzz_o0_t1_m0_l0, d_gtd_zz_o1_t0_m0_l0, d_gtd_zz_o2_t0_m0_l0, d_Ctd_yzz_o0_t1_m0_l0, d_gtd_zz_o1_t1_m0_l0, d_gtd_zz_o2_t1_m0_l0, d_Ctd_zzz_o0_t0_m0_l0, d_gtd_zz_o1_t2_m0_l0, d_gtd_zz_o2_t2_m0_l0, d_Rpd_xx_o0_t0_m0_l0, d_chi_o1_t0_m0_l0, d_chi_o2_t0_m0_l0, d_Rpd_xx_o0_t1_m0_l0, d_chi_o1_t1_m0_l0, d_chi_o2_t1_m0_l0, d_Rpd_xx_o0_t2_m0_l0, d_chi_o1_t2_m0_l0, d_chi_o2_t2_m0_l0, d_Rtd_xx_o0_t6_m0_l0, d_Gamh_x_o1_t0_m0_l0, d_Gamh_x_o2_t0_m0_l0, d_Rtd_xy_o0_t6_m0_l0, d_Gamh_x_o1_t1_m0_l0, d_Gamh_x_o2_t1_m0_l0, d_Rtd_xz_o0_t6_m0_l0, d_Gamh_x_o1_t2_m0_l0, d_Gamh_x_o2_t2_m0_l0, d_Rtd_xx_o0_t7_m0_l0, d_Gamh_y_o1_t0_m0_l0, d_Gamh_y_o2_t0_m0_l0, d_Rtd_xy_o0_t7_m0_l0, d_Gamh_y_o1_t1_m0_l0, d_Gamh_y_o2_t1_m0_l0, d_Rtd_xz_o0_t7_m0_l0, d_Gamh_y_o1_t2_m0_l0, d_Gamh_y_o2_t2_m0_l0, d_Rtd_xx_o0_t8_m0_l0, d_Gamh_z_o1_t0_m0_l0, d_Gamh_z_o2_t0_m0_l0, d_Rtd_xy_o0_t8_m0_l0, d_Gamh_z_o1_t1_m0_l0, d_Gamh_z_o2_t1_m0_l0, d_Rtd_xz_o0_t8_m0_l0, d_Gamh_z_o1_t2_m0_l0, d_Gamh_z_o2_t2_m0_l0, d_Psi1_xx_o0_t1_m0_l0, d_Alpha_o1_t0_m0_l0, d_Alpha_o2_t0_m0_l0, d_Psi1_xx_o0_t2_m0_l0, d_Alpha_o1_t1_m0_l0, d_Alpha_o2_t1_m0_l0, d_Psi1_xx_o0_t3_m0_l0, d_Alpha_o1_t2_m0_l0, d_Alpha_o2_t2_m0_l0, d_Psi1_xx_o0_t0_m0_l0, d_Psi1_xy_o0_t0_m0_l0, d_Psi1_xz_o0_t0_m0_l0, d_Psi1_yy_o0_t0_m0_l0, d_Psi1_yz_o0_t0_m0_l0, d_Psi1_zz_o0_t0_m0_l0, d_Atd_xx_o1_t0_m0_l0, d_Atd_xx_o2_t0_m0_l0, d_Atd_xx_o1_t1_m0_l0, d_Atd_xx_o2_t1_m0_l0, d_Atd_xx_o1_t2_m0_l0, d_Atd_xx_o2_t2_m0_l0, d_Atd_xy_o1_t0_m0_l0, d_Atd_xy_o2_t0_m0_l0, d_Atd_xy_o1_t1_m0_l0, d_Atd_xy_o2_t1_m0_l0, d_Atd_xy_o1_t2_m0_l0, d_Atd_xy_o2_t2_m0_l0, d_Atd_xz_o1_t0_m0_l0, d_Atd_xz_o2_t0_m0_l0, d_Atd_xz_o1_t1_m0_l0, d_Atd_xz_o2_t1_m0_l0, d_Atd_xz_o1_t2_m0_l0, d_Atd_xz_o2_t2_m0_l0, d_Atd_yy_o1_t0_m0_l0, d_Atd_yy_o2_t0_m0_l0, d_Atd_yy_o1_t1_m0_l0, d_Atd_yy_o2_t1_m0_l0, d_Atd_yy_o1_t2_m0_l0, d_Atd_yy_o2_t2_m0_l0, d_Atd_yz_o1_t0_m0_l0, d_Atd_yz_o2_t0_m0_l0, d_Atd_yz_o1_t1_m0_l0, d_Atd_yz_o2_t1_m0_l0, d_Atd_yz_o1_t2_m0_l0, d_Atd_yz_o2_t2_m0_l0, d_Atd_zz_o1_t0_m0_l0, d_Atd_zz_o2_t0_m0_l0, d_Atd_zz_o1_t1_m0_l0, d_Atd_zz_o2_t1_m0_l0, d_Atd_zz_o1_t2_m0_l0, d_Atd_zz_o2_t2_m0_l0, d_Gamh_x_o0_t0_m0_l0, d_trK_o1_t0_m0_l0, d_trK_o2_t0_m0_l0, d_Gamh_x_o0_t1_m0_l0, d_trK_o1_t1_m0_l0, d_trK_o2_t1_m0_l0, d_Gamh_x_o0_t2_m0_l0, d_trK_o1_t2_m0_l0, d_trK_o2_t2_m0_l0, d_Gamh_x_o0_t3_m0_l0, d_theta_o1_t0_m0_l0, d_theta_o2_t0_m0_l0, d_Gamh_x_o0_t4_m0_l0, d_theta_o1_t1_m0_l0, d_theta_o2_t1_m0_l0, d_Gamh_x_o0_t5_m0_l0, d_theta_o1_t2_m0_l0, d_theta_o2_t2_m0_l0, d_Rtd_xx_o0_t0_m0_l0, d_Rtd_xx_o0_t1_m0_l0, d_Rtd_xx_o0_t2_m0_l0, d_Rtd_xx_o0_t3_m0_l0, d_Rtd_xx_o0_t4_m0_l0, d_Rtd_xx_o0_t5_m0_l0, d_Rtd_xy_o0_t0_m0_l0, d_Rtd_xy_o0_t1_m0_l0, d_Rtd_xy_o0_t2_m0_l0, d_Rtd_xy_o0_t3_m0_l0, d_Rtd_xy_o0_t4_m0_l0, d_Rtd_xy_o0_t5_m0_l0, d_Rtd_yy_o0_t0_m0_l0, d_Rtd_yy_o0_t1_m0_l0, d_Rtd_yy_o0_t2_m0_l0, d_Rtd_yy_o0_t3_m0_l0, d_Rtd_yy_o0_t4_m0_l0, d_Rtd_yy_o0_t5_m0_l0, d_Rtd_xz_o0_t0_m0_l0, d_Rtd_xz_o0_t1_m0_l0, d_Rtd_xz_o0_t2_m0_l0, d_Rtd_xz_o0_t3_m0_l0, d_Rtd_xz_o0_t4_m0_l0, d_Rtd_xz_o0_t5_m0_l0, d_Rtd_yz_o0_t0_m0_l0, d_Rtd_yz_o0_t1_m0_l0, d_Rtd_yz_o0_t2_m0_l0, d_Rtd_yz_o0_t3_m0_l0, d_Rtd_yz_o0_t4_m0_l0, d_Rtd_yz_o0_t5_m0_l0, d_Rtd_zz_o0_t0_m0_l0, d_Rtd_zz_o0_t1_m0_l0, d_Rtd_zz_o0_t2_m0_l0, d_Rtd_zz_o0_t3_m0_l0, d_Rtd_zz_o0_t4_m0_l0, d_Rtd_zz_o0_t5_m0_l0, d_Rpd_xx_o0_t3_m0_l0, d_Rpd_xx_o0_t4_m0_l0, d_Rpd_xx_o0_t5_m0_l0, d_Rpd_xx_o0_t6_m0_l0, d_Rpd_xx_o0_t7_m0_l0, d_Rpd_xx_o0_t8_m0_l0, rk1m_Ctd_yyz_o0_t0_l0, rk1Ctd_yyz, rk1m_Ctd_yyy_o0_t0_l0, rk1Ctd_yyy, rk1div_Beta, rk1m_Ctd_xxz_o0_t0_l0, rk1Ctd_xxz, rk1m_Ctd_xxx_o0_t0_l0, rk1Ctd_xxx, rk1m_Ctd_yzz_o0_t1_l0, rk1Ctd_yzz, rk1m_Ctd_xxy_o0_t0_l0, rk1Ctd_xxy, rk1kappa_z2, rk1m_Ctd_xyy_o0_t1_l0, rk1Ctd_xyy, rk1m_Ctd_xyz_o0_t2_l0, rk1m_Ctd_xyz_o0_t1_l0, rk1m_Ctd_xyz_o0_t0_l0, rk1Ctd_xyz, rk1d_div_Beta_y, rk1d_div_Beta_x, rk1d_div_Beta_z, rk1m_Ctd_xzz_o0_t1_l0, rk1Ctd_xzz, rk1m_Ctd_zxz_o0_t0_l0, rk1Ctd_zxz, rk1m_Ctd_zxx_o0_t1_l0, rk1Ctd_zxx, rk1m_Ctd_zyy_o0_t1_l0, rk1Ctd_zyy, rk1m_Ctd_zyz_o0_t0_l0, rk1Ctd_zyz, rk1Tf4u_tt, rk1Betatd_x, rk1Betatd_z, rk1Betatd_y, rk1Ct_zyz, rk1Ct_zyy, rk1Atud_zx, rk1Atud_zz, rk1Atud_zy, rk1Atu_zz, rk1Tf4u_tz, rk1Tu_tz, rk1Atud_yy, rk1Atud_yx, rk1Atud_yz, rk1Atu_yz, rk1Atu_yy, rk1Ct_yyy, rk1Ct_yyz, rk1Atud_xx, rk1trAt, rk1Atud_xz, rk1Atud_xy, rk1Atu_xz, rk1Atu_xy, rk1Atu_xx, rk1Tf4u_ty, rk1Tu_ty, rk1Tf4u_tx, rk1Tu_tx, rk1decay_factor, rk1kappa_z1, rk1kappa_f, rk1feta, rk1m_Ctd_zzz_o0_t0_l0, rk1Ctd_zzz, rk1Ct_xzz, rk1Ct_zzz, rk1Ct_yzz, rk1m_Ctd_yxy_o0_t0_l0, rk1Ctd_yxy, rk1Ct_xyz, rk1m_Ctd_yxz_o0_t2_l0, rk1m_Ctd_yxz_o0_t0_l0, rk1Ctd_yxz, rk1Ct_zxz, rk1Ct_yxz, rk1m_Ctd_zxy_o0_t2_l0, rk1Ctd_zxy, rk1Ct_zxy, rk1Ct_yxy, rk1Ct_xxy, rk1Ct_xxz, rk1Ct_xyy, rk1m_Ctd_yxx_o0_t1_l0, rk1Ctd_yxx, rk1Ct_zxx, rk1Gamt_z, rk1Zu_z, rk1Ct_yxx, rk1Gamt_y, rk1Zu_y, rk1Ct_xxx, rk1Gamt_x, rk1Zu_x, rk1d_Rtd_xx_o0_t9_m0_l0, rk1m_Rtd_xx_o0_t8_l0, rk1m_Rtd_xx_o0_t7_l0, rk1m_Rtd_xx_o0_t6_l0, rk1m_Rtd_xx_o0_t5_l0, rk1m_Rtd_xx_o0_t4_l0, rk1m_Rtd_xx_o0_t3_l0, rk1m_Rtd_xx_o0_t2_l0, rk1m_Rtd_xx_o0_t1_l0, rk1m_Rtd_xx_o0_t0_l0, rk1Rtd_xx, rk1d_Rtd_xz_o0_t12_m0_l0, rk1m_Rtd_xz_o0_t11_l0, rk1m_Rtd_xz_o0_t10_l0, rk1m_Rtd_xz_o0_t9_l0, rk1m_Rtd_xz_o0_t8_l0, rk1m_Rtd_xz_o0_t7_l0, rk1m_Rtd_xz_o0_t6_l0, rk1m_Rtd_xz_o0_t5_l0, rk1m_Rtd_xz_o0_t4_l0, rk1m_Rtd_xz_o0_t3_l0, rk1m_Rtd_xz_o0_t2_l0, rk1m_Rtd_xz_o0_t1_l0, rk1m_Rtd_xz_o0_t0_l0, rk1Rtd_xz, rk1d_Rtd_zz_o0_t9_m0_l0, rk1m_Rtd_zz_o0_t8_l0, rk1m_Rtd_zz_o0_t7_l0, rk1m_Rtd_zz_o0_t6_l0, rk1m_Rtd_zz_o0_t5_l0, rk1m_Rtd_zz_o0_t4_l0, rk1m_Rtd_zz_o0_t3_l0, rk1m_Rtd_zz_o0_t2_l0, rk1m_Rtd_zz_o0_t1_l0, rk1m_Rtd_zz_o0_t0_l0, rk1Rtd_zz, rk1d_Rtd_xy_o0_t12_m0_l0, rk1m_Rtd_xy_o0_t11_l0, rk1m_Rtd_xy_o0_t10_l0, rk1m_Rtd_xy_o0_t9_l0, rk1m_Rtd_xy_o0_t8_l0, rk1m_Rtd_xy_o0_t7_l0, rk1m_Rtd_xy_o0_t6_l0, rk1m_Rtd_xy_o0_t5_l0, rk1m_Rtd_xy_o0_t4_l0, rk1m_Rtd_xy_o0_t3_l0, rk1m_Rtd_xy_o0_t2_l0, rk1m_Rtd_xy_o0_t1_l0, rk1m_Rtd_xy_o0_t0_l0, rk1Rtd_xy, rk1d_Rtd_yy_o0_t9_m0_l0, rk1m_Rtd_yy_o0_t8_l0, rk1m_Rtd_yy_o0_t7_l0, rk1m_Rtd_yy_o0_t6_l0, rk1m_Rtd_yy_o0_t5_l0, rk1m_Rtd_yy_o0_t4_l0, rk1m_Rtd_yy_o0_t3_l0, rk1m_Rtd_yy_o0_t2_l0, rk1m_Rtd_yy_o0_t1_l0, rk1m_Rtd_yy_o0_t0_l0, rk1Rtd_yy, rk1d_Rtd_yz_o0_t12_m0_l0, rk1m_Rtd_yz_o0_t11_l0, rk1m_Rtd_yz_o0_t10_l0, rk1m_Rtd_yz_o0_t9_l0, rk1m_Rtd_yz_o0_t8_l0, rk1m_Rtd_yz_o0_t7_l0, rk1m_Rtd_yz_o0_t6_l0, rk1m_Rtd_yz_o0_t5_l0, rk1m_Rtd_yz_o0_t4_l0, rk1m_Rtd_yz_o0_t3_l0, rk1m_Rtd_yz_o0_t2_l0, rk1m_Rtd_yz_o0_t1_l0, rk1m_Rtd_yz_o0_t0_l0, rk1Rtd_yz, rk1m_Rpd_xx_o0_t22_l0, rk1m_Rpd_xx_o0_t21_l0, rk1m_Rpd_xx_o0_t20_l0, rk1m_Rpd_xx_o0_t19_l0, rk1m_Rpd_xx_o0_t18_l0, rk1m_Rpd_xx_o0_t17_l0, rk1m_Rpd_xx_o0_t16_l0, rk1m_Rpd_xx_o0_t15_l0, rk1m_Rpd_xx_o0_t14_l0, rk1m_Rpd_xx_o0_t13_l0, rk1m_Rpd_xx_o0_t12_l0, rk1m_Rpd_xx_o0_t11_l0, rk1m_Rpd_xx_o0_t10_l0, rk1m_Rpd_xx_o0_t9_l0, rk1m_Rpd_xx_o0_t8_l0, rk1m_Rpd_xx_o0_t7_l0, rk1m_Rpd_xx_o0_t6_l0, rk1m_Rpd_xx_o0_t5_l0, rk1m_Rpd_xx_o0_t4_l0, rk1m_Rpd_xx_o0_t3_l0, rk1m_Rpd_xx_o0_t2_l0, rk1m_Rpd_xx_o0_t1_l0, rk1m_Rpd_xx_o0_t0_l0, rk1Rpd_xx, rk1m_Rpd_yy_o0_t22_l0, rk1m_Rpd_yy_o0_t21_l0, rk1m_Rpd_yy_o0_t20_l0, rk1m_Rpd_yy_o0_t19_l0, rk1m_Rpd_yy_o0_t18_l0, rk1m_Rpd_yy_o0_t17_l0, rk1m_Rpd_yy_o0_t16_l0, rk1m_Rpd_yy_o0_t15_l0, rk1m_Rpd_yy_o0_t14_l0, rk1m_Rpd_yy_o0_t13_l0, rk1m_Rpd_yy_o0_t12_l0, rk1m_Rpd_yy_o0_t11_l0, rk1m_Rpd_yy_o0_t10_l0, rk1m_Rpd_yy_o0_t9_l0, rk1m_Rpd_yy_o0_t8_l0, rk1m_Rpd_yy_o0_t7_l0, rk1m_Rpd_yy_o0_t6_l0, rk1m_Rpd_yy_o0_t5_l0, rk1m_Rpd_yy_o0_t4_l0, rk1m_Rpd_yy_o0_t3_l0, rk1m_Rpd_yy_o0_t2_l0, rk1m_Rpd_yy_o0_t1_l0, rk1m_Rpd_yy_o0_t0_l0, rk1Rpd_yy, rk1m_Rpd_yz_o0_t25_l0, rk1m_Rpd_yz_o0_t24_l0, rk1m_Rpd_yz_o0_t23_l0, rk1m_Rpd_yz_o0_t22_l0, rk1m_Rpd_yz_o0_t21_l0, rk1m_Rpd_yz_o0_t20_l0, rk1m_Rpd_yz_o0_t19_l0, rk1m_Rpd_yz_o0_t18_l0, rk1m_Rpd_yz_o0_t17_l0, rk1m_Rpd_yz_o0_t16_l0, rk1m_Rpd_yz_o0_t15_l0, rk1m_Rpd_yz_o0_t14_l0, rk1m_Rpd_yz_o0_t13_l0, rk1m_Rpd_yz_o0_t12_l0, rk1m_Rpd_yz_o0_t11_l0, rk1m_Rpd_yz_o0_t10_l0, rk1m_Rpd_yz_o0_t9_l0, rk1m_Rpd_yz_o0_t8_l0, rk1m_Rpd_yz_o0_t7_l0, rk1m_Rpd_yz_o0_t6_l0, rk1m_Rpd_yz_o0_t5_l0, rk1m_Rpd_yz_o0_t4_l0, rk1m_Rpd_yz_o0_t3_l0, rk1m_Rpd_yz_o0_t2_l0, rk1m_Rpd_yz_o0_t1_l0, rk1m_Rpd_yz_o0_t0_l0, rk1Rpd_yz, rk1m_Rpd_xz_o0_t25_l0, rk1m_Rpd_xz_o0_t24_l0, rk1m_Rpd_xz_o0_t23_l0, rk1m_Rpd_xz_o0_t22_l0, rk1m_Rpd_xz_o0_t21_l0, rk1m_Rpd_xz_o0_t20_l0, rk1m_Rpd_xz_o0_t19_l0, rk1m_Rpd_xz_o0_t18_l0, rk1m_Rpd_xz_o0_t17_l0, rk1m_Rpd_xz_o0_t16_l0, rk1m_Rpd_xz_o0_t15_l0, rk1m_Rpd_xz_o0_t14_l0, rk1m_Rpd_xz_o0_t13_l0, rk1m_Rpd_xz_o0_t12_l0, rk1m_Rpd_xz_o0_t11_l0, rk1m_Rpd_xz_o0_t10_l0, rk1m_Rpd_xz_o0_t9_l0, rk1m_Rpd_xz_o0_t8_l0, rk1m_Rpd_xz_o0_t7_l0, rk1m_Rpd_xz_o0_t6_l0, rk1m_Rpd_xz_o0_t5_l0, rk1m_Rpd_xz_o0_t4_l0, rk1m_Rpd_xz_o0_t3_l0, rk1m_Rpd_xz_o0_t2_l0, rk1m_Rpd_xz_o0_t1_l0, rk1m_Rpd_xz_o0_t0_l0, rk1Rpd_xz, rk1m_Rpd_xy_o0_t25_l0, rk1m_Rpd_xy_o0_t24_l0, rk1m_Rpd_xy_o0_t23_l0, rk1m_Rpd_xy_o0_t22_l0, rk1m_Rpd_xy_o0_t21_l0, rk1m_Rpd_xy_o0_t20_l0, rk1m_Rpd_xy_o0_t19_l0, rk1m_Rpd_xy_o0_t18_l0, rk1m_Rpd_xy_o0_t17_l0, rk1m_Rpd_xy_o0_t16_l0, rk1m_Rpd_xy_o0_t15_l0, rk1m_Rpd_xy_o0_t14_l0, rk1m_Rpd_xy_o0_t13_l0, rk1m_Rpd_xy_o0_t12_l0, rk1m_Rpd_xy_o0_t11_l0, rk1m_Rpd_xy_o0_t10_l0, rk1m_Rpd_xy_o0_t9_l0, rk1m_Rpd_xy_o0_t8_l0, rk1m_Rpd_xy_o0_t7_l0, rk1m_Rpd_xy_o0_t6_l0, rk1m_Rpd_xy_o0_t5_l0, rk1m_Rpd_xy_o0_t4_l0, rk1m_Rpd_xy_o0_t3_l0, rk1m_Rpd_xy_o0_t2_l0, rk1m_Rpd_xy_o0_t1_l0, rk1m_Rpd_xy_o0_t0_l0, rk1Rpd_xy, rk1m_Rpd_zz_o0_t22_l0, rk1m_Rpd_zz_o0_t21_l0, rk1m_Rpd_zz_o0_t20_l0, rk1m_Rpd_zz_o0_t19_l0, rk1m_Rpd_zz_o0_t18_l0, rk1m_Rpd_zz_o0_t17_l0, rk1m_Rpd_zz_o0_t16_l0, rk1m_Rpd_zz_o0_t15_l0, rk1m_Rpd_zz_o0_t14_l0, rk1m_Rpd_zz_o0_t13_l0, rk1m_Rpd_zz_o0_t12_l0, rk1m_Rpd_zz_o0_t11_l0, rk1m_Rpd_zz_o0_t10_l0, rk1m_Rpd_zz_o0_t9_l0, rk1m_Rpd_zz_o0_t8_l0, rk1m_Rpd_zz_o0_t7_l0, rk1m_Rpd_zz_o0_t6_l0, rk1m_Rpd_zz_o0_t5_l0, rk1m_Rpd_zz_o0_t4_l0, rk1m_Rpd_zz_o0_t3_l0, rk1m_Rpd_zz_o0_t2_l0, rk1m_Rpd_zz_o0_t1_l0, rk1m_Rpd_zz_o0_t0_l0, rk1Rpd_zz, rk1Rscalar, rk1Tf4u_zz, rk1Tu_zz, rk1Tf4u_yz, rk1Tu_yz, rk1Tf4u_yy, rk1Tu_yy, rk1Tf4u_xy, rk1Tu_xy, rk1Tf4u_xx, rk1Tu_xx, rk1Tf4u_xz, rk1Tu_xz, rk1trSf, rk1Tu_tt, rk1pTtd_ADM_xz, rk1pTtd_ADM_xy, rk1pTtd_ADM_xx, rk1d_Psi1_xx_o0_t5_m0_l0, rk1m_Psi1_xx_o0_t4_l0, rk1m_Psi1_xx_o0_t3_l0, rk1m_Psi1_xx_o0_t2_l0, rk1m_Psi1_xx_o0_t1_l0, rk1m_Psi1_xx_o0_t0_l0, rk1Psi1_xx, rk1d_Psi1_xy_o0_t6_m0_l0, rk1m_Psi1_xy_o0_t5_l0, rk1m_Psi1_xy_o0_t4_l0, rk1m_Psi1_xy_o0_t3_l0, rk1m_Psi1_xy_o0_t2_l0, rk1m_Psi1_xy_o0_t1_l0, rk1m_Psi1_xy_o0_t0_l0, rk1Psi1_xy, rk1d_Psi1_xz_o0_t6_m0_l0, rk1m_Psi1_xz_o0_t5_l0, rk1m_Psi1_xz_o0_t4_l0, rk1m_Psi1_xz_o0_t3_l0, rk1m_Psi1_xz_o0_t2_l0, rk1m_Psi1_xz_o0_t1_l0, rk1m_Psi1_xz_o0_t0_l0, rk1Psi1_xz, rk1Jtd_ADM_z, rk1pTtd_ADM_zz, rk1d_Psi1_zz_o0_t5_m0_l0, rk1m_Psi1_zz_o0_t4_l0, rk1m_Psi1_zz_o0_t3_l0, rk1m_Psi1_zz_o0_t2_l0, rk1m_Psi1_zz_o0_t1_l0, rk1m_Psi1_zz_o0_t0_l0, rk1Psi1_zz, rk1Jtd_ADM_y, rk1Jtd_ADM_x, rk1pTtd_ADM_yz, rk1d_Psi1_yz_o0_t6_m0_l0, rk1m_Psi1_yz_o0_t5_l0, rk1m_Psi1_yz_o0_t4_l0, rk1m_Psi1_yz_o0_t3_l0, rk1m_Psi1_yz_o0_t2_l0, rk1m_Psi1_yz_o0_t1_l0, rk1m_Psi1_yz_o0_t0_l0, rk1Psi1_yz, rk1pTtd_ADM_yy, rk1tr_pT, rk1d_Psi1_yy_o0_t5_m0_l0, rk1m_Psi1_yy_o0_t4_l0, rk1m_Psi1_yy_o0_t3_l0, rk1m_Psi1_yy_o0_t2_l0, rk1m_Psi1_yy_o0_t1_l0, rk1m_Psi1_yy_o0_t0_l0, rk1Psi1_yy, rk1trPsi1, rk1Psi1TF_yz, rk1Psi1TF_yy, rk1Psi1TF_xy, rk1Psi1TF_xx, rk1Psi1TF_xz, rk1Psi1TF_zz, rk1kappa_cc, rk1rho_ADM, d_Sfd_x_o0_t11_m0_l0, d_Sfd_y_o0_t11_m0_l0, d_Sfd_z_o0_t11_m0_l0, d_DYf_o0_t0_m0_l0, d_tauf_o0_t3_m0_l0, d_gtd_xx_o0_t3_m0_l0, d_gtd_xy_o0_t6_m0_l0, d_gtd_xz_o0_t6_m0_l0, d_gtd_yy_o0_t3_m0_l0, d_gtd_yz_o0_t6_m0_l0, d_gtd_zz_o0_t3_m0_l0, d_Atd_xx_o0_t3_m0_l0, d_Atd_xy_o0_t6_m0_l0, d_Atd_xz_o0_t6_m0_l0, d_Atd_yy_o0_t3_m0_l0, d_Atd_yz_o0_t6_m0_l0, d_Atd_zz_o0_t3_m0_l0, d_Gamh_x_o0_t24_m0_l0, d_Gamh_y_o0_t24_m0_l0, d_Gamh_z_o0_t24_m0_l0, d_Betau_x_o0_t0_m0_l0, d_Betau_y_o0_t0_m0_l0, d_Betau_z_o0_t0_m0_l0, d_Alpha_o0_t0_m0_l0, d_chi_o0_t0_m0_l0, d_trK_o0_t21_m0_l0, d_theta_o0_t3_m0_l0, RHS_Bfu_x, RHS_Bfu_y, RHS_Bfu_z, RHS_tauf, RHS_phif, RHS_Sfd_x, RHS_Sfd_y, RHS_Sfd_z, RHS_Df, RHS_DYf, m_theta_o2_t2_l0, m_theta_o2_t1_l0, m_theta_o2_t0_l0, m_theta_o1_t2_l0, m_theta_o1_t1_l0, m_theta_o1_t0_l0, m_theta_o0_t2_l0, m_theta_o0_t1_l0, m_theta_o0_t0_l0, m_trK_o2_t2_l0, m_trK_o2_t1_l0, m_trK_o2_t0_l0, m_trK_o1_t2_l0, m_trK_o1_t1_l0, m_trK_o1_t0_l0, m_trK_o0_t20_l0, m_trK_o0_t19_l0, m_trK_o0_t18_l0, m_trK_o0_t17_l0, m_trK_o0_t16_l0, m_trK_o0_t15_l0, m_trK_o0_t14_l0, m_trK_o0_t13_l0, m_trK_o0_t12_l0, m_trK_o0_t11_l0, m_trK_o0_t10_l0, m_trK_o0_t9_l0, m_trK_o0_t8_l0, m_trK_o0_t7_l0, m_trK_o0_t6_l0, m_trK_o0_t5_l0, m_trK_o0_t4_l0, m_trK_o0_t3_l0, m_trK_o0_t2_l0, m_trK_o0_t1_l0, m_trK_o0_t0_l0, m_chi_o2_t2_l0, m_chi_o2_t1_l0, m_chi_o2_t0_l0, m_chi_o1_t2_l0, m_chi_o1_t1_l0, m_chi_o1_t0_l0, m_Alpha_o2_t2_l0, m_Alpha_o2_t1_l0, m_Alpha_o2_t0_l0, m_Alpha_o1_t2_l0, m_Alpha_o1_t1_l0, m_Alpha_o1_t0_l0, m_Betau_z_o2_t2_l0, m_Betau_z_o2_t1_l0, m_Betau_z_o2_t0_l0, m_Betau_z_o1_t2_l0, m_Betau_z_o1_t1_l0, m_Betau_z_o1_t0_l0, m_Betau_y_o2_t2_l0, m_Betau_y_o2_t1_l0, m_Betau_y_o2_t0_l0, m_Betau_y_o1_t2_l0, m_Betau_y_o1_t1_l0, m_Betau_y_o1_t0_l0, m_Betau_x_o2_t2_l0, m_Betau_x_o2_t1_l0, m_Betau_x_o2_t0_l0, m_Betau_x_o1_t2_l0, m_Betau_x_o1_t1_l0, m_Betau_x_o1_t0_l0, m_Gamh_z_o2_t2_l0, m_Gamh_z_o2_t1_l0, m_Gamh_z_o2_t0_l0, m_Gamh_z_o1_t2_l0, m_Gamh_z_o1_t1_l0, m_Gamh_z_o1_t0_l0, m_Gamh_z_o0_t23_l0, m_Gamh_z_o0_t22_l0, m_Gamh_z_o0_t21_l0, m_Gamh_z_o0_t20_l0, m_Gamh_z_o0_t19_l0, m_Gamh_z_o0_t18_l0, m_Gamh_z_o0_t17_l0, m_Gamh_z_o0_t16_l0, m_Gamh_z_o0_t15_l0, m_Gamh_z_o0_t14_l0, m_Gamh_z_o0_t13_l0, m_Gamh_z_o0_t12_l0, m_Gamh_z_o0_t11_l0, m_Gamh_z_o0_t10_l0, m_Gamh_z_o0_t9_l0, m_Gamh_z_o0_t8_l0, m_Gamh_z_o0_t7_l0, m_Gamh_z_o0_t6_l0, m_Gamh_z_o0_t5_l0, m_Gamh_z_o0_t4_l0, m_Gamh_z_o0_t3_l0, m_Gamh_z_o0_t2_l0, m_Gamh_z_o0_t1_l0, m_Gamh_z_o0_t0_l0, m_Gamh_y_o2_t2_l0, m_Gamh_y_o2_t1_l0, m_Gamh_y_o2_t0_l0, m_Gamh_y_o1_t2_l0, m_Gamh_y_o1_t1_l0, m_Gamh_y_o1_t0_l0, m_Gamh_y_o0_t23_l0, m_Gamh_y_o0_t22_l0, m_Gamh_y_o0_t21_l0, m_Gamh_y_o0_t20_l0, m_Gamh_y_o0_t19_l0, m_Gamh_y_o0_t18_l0, m_Gamh_y_o0_t17_l0, m_Gamh_y_o0_t16_l0, m_Gamh_y_o0_t15_l0, m_Gamh_y_o0_t14_l0, m_Gamh_y_o0_t13_l0, m_Gamh_y_o0_t12_l0, m_Gamh_y_o0_t11_l0, m_Gamh_y_o0_t10_l0, m_Gamh_y_o0_t9_l0, m_Gamh_y_o0_t8_l0, m_Gamh_y_o0_t7_l0, m_Gamh_y_o0_t6_l0, m_Gamh_y_o0_t5_l0, m_Gamh_y_o0_t4_l0, m_Gamh_y_o0_t3_l0, m_Gamh_y_o0_t2_l0, m_Gamh_y_o0_t1_l0, m_Gamh_y_o0_t0_l0, m_Gamh_x_o2_t2_l0, m_Gamh_x_o2_t1_l0, m_Gamh_x_o2_t0_l0, m_Gamh_x_o1_t2_l0, m_Gamh_x_o1_t1_l0, m_Gamh_x_o1_t0_l0, m_Gamh_x_o0_t23_l0, m_Gamh_x_o0_t22_l0, m_Gamh_x_o0_t21_l0, m_Gamh_x_o0_t20_l0, m_Gamh_x_o0_t19_l0, m_Gamh_x_o0_t18_l0, m_Gamh_x_o0_t17_l0, m_Gamh_x_o0_t16_l0, m_Gamh_x_o0_t15_l0, m_Gamh_x_o0_t14_l0, m_Gamh_x_o0_t13_l0, m_Gamh_x_o0_t12_l0, m_Gamh_x_o0_t11_l0, m_Gamh_x_o0_t10_l0, m_Gamh_x_o0_t9_l0, m_Gamh_x_o0_t8_l0, m_Gamh_x_o0_t7_l0, m_Gamh_x_o0_t6_l0, m_Gamh_x_o0_t5_l0, m_Gamh_x_o0_t4_l0, m_Gamh_x_o0_t3_l0, m_Gamh_x_o0_t2_l0, m_Gamh_x_o0_t1_l0, m_Gamh_x_o0_t0_l0, m_Atd_zz_o2_t2_l0, m_Atd_zz_o2_t1_l0, m_Atd_zz_o2_t0_l0, m_Atd_zz_o1_t2_l0, m_Atd_zz_o1_t1_l0, m_Atd_zz_o1_t0_l0, m_Atd_zz_o0_t2_l0, m_Atd_zz_o0_t1_l0, m_Atd_zz_o0_t0_l0, m_Atd_yz_o2_t2_l0, m_Atd_yz_o2_t1_l0, m_Atd_yz_o2_t0_l0, m_Atd_yz_o1_t2_l0, m_Atd_yz_o1_t1_l0, m_Atd_yz_o1_t0_l0, m_Atd_yz_o0_t5_l0, m_Atd_yz_o0_t4_l0, m_Atd_yz_o0_t3_l0, m_Atd_yz_o0_t2_l0, m_Atd_yz_o0_t1_l0, m_Atd_yz_o0_t0_l0, m_Atd_yy_o2_t2_l0, m_Atd_yy_o2_t1_l0, m_Atd_yy_o2_t0_l0, m_Atd_yy_o1_t2_l0, m_Atd_yy_o1_t1_l0, m_Atd_yy_o1_t0_l0, m_Atd_yy_o0_t2_l0, m_Atd_yy_o0_t1_l0, m_Atd_yy_o0_t0_l0, m_Atd_xz_o2_t2_l0, m_Atd_xz_o2_t1_l0, m_Atd_xz_o2_t0_l0, m_Atd_xz_o1_t2_l0, m_Atd_xz_o1_t1_l0, m_Atd_xz_o1_t0_l0, m_Atd_xz_o0_t5_l0, m_Atd_xz_o0_t4_l0, m_Atd_xz_o0_t3_l0, m_Atd_xz_o0_t2_l0, m_Atd_xz_o0_t1_l0, m_Atd_xz_o0_t0_l0, m_Atd_xy_o2_t2_l0, m_Atd_xy_o2_t1_l0, m_Atd_xy_o2_t0_l0, m_Atd_xy_o1_t2_l0, m_Atd_xy_o1_t1_l0, m_Atd_xy_o1_t0_l0, m_Atd_xy_o0_t5_l0, m_Atd_xy_o0_t4_l0, m_Atd_xy_o0_t3_l0, m_Atd_xy_o0_t2_l0, m_Atd_xy_o0_t1_l0, m_Atd_xy_o0_t0_l0, m_Atd_xx_o2_t2_l0, m_Atd_xx_o2_t1_l0, m_Atd_xx_o2_t0_l0, m_Atd_xx_o1_t2_l0, m_Atd_xx_o1_t1_l0, m_Atd_xx_o1_t0_l0, m_Atd_xx_o0_t2_l0, m_Atd_xx_o0_t1_l0, m_Atd_xx_o0_t0_l0, m_gtd_zz_o2_t2_l0, m_gtd_zz_o2_t1_l0, m_gtd_zz_o2_t0_l0, m_gtd_zz_o1_t2_l0, m_gtd_zz_o1_t1_l0, m_gtd_zz_o1_t0_l0, m_gtd_zz_o0_t2_l0, m_gtd_zz_o0_t1_l0, m_gtd_zz_o0_t0_l0, m_gtd_yz_o2_t2_l0, m_gtd_yz_o2_t1_l0, m_gtd_yz_o2_t0_l0, m_gtd_yz_o1_t2_l0, m_gtd_yz_o1_t1_l0, m_gtd_yz_o1_t0_l0, m_gtd_yz_o0_t5_l0, m_gtd_yz_o0_t4_l0, m_gtd_yz_o0_t3_l0, m_gtd_yz_o0_t2_l0, m_gtd_yz_o0_t1_l0, m_gtd_yz_o0_t0_l0, m_gtd_yy_o2_t2_l0, m_gtd_yy_o2_t1_l0, m_gtd_yy_o2_t0_l0, m_gtd_yy_o1_t2_l0, m_gtd_yy_o1_t1_l0, m_gtd_yy_o1_t0_l0, m_gtd_yy_o0_t2_l0, m_gtd_yy_o0_t1_l0, m_gtd_yy_o0_t0_l0, m_gtd_xz_o2_t2_l0, m_gtd_xz_o2_t1_l0, m_gtd_xz_o2_t0_l0, m_gtd_xz_o1_t2_l0, m_gtd_xz_o1_t1_l0, m_gtd_xz_o1_t0_l0, m_gtd_xz_o0_t5_l0, m_gtd_xz_o0_t4_l0, m_gtd_xz_o0_t3_l0, m_gtd_xz_o0_t2_l0, m_gtd_xz_o0_t1_l0, m_gtd_xz_o0_t0_l0, m_gtd_xy_o2_t2_l0, m_gtd_xy_o2_t1_l0, m_gtd_xy_o2_t0_l0, m_gtd_xy_o1_t2_l0, m_gtd_xy_o1_t1_l0, m_gtd_xy_o1_t0_l0, m_gtd_xy_o0_t5_l0, m_gtd_xy_o0_t4_l0, m_gtd_xy_o0_t3_l0, m_gtd_xy_o0_t2_l0, m_gtd_xy_o0_t1_l0, m_gtd_xy_o0_t0_l0, m_gtd_xx_o2_t2_l0, m_gtd_xx_o2_t1_l0, m_gtd_xx_o2_t0_l0, m_gtd_xx_o1_t2_l0, m_gtd_xx_o1_t1_l0, m_gtd_xx_o1_t0_l0, m_gtd_xx_o0_t2_l0, m_gtd_xx_o0_t1_l0, m_gtd_xx_o0_t0_l0, m_phif_o0_t2_l0, m_phif_o0_t1_l0, m_phif_o0_t0_l0, m_tauf_o0_t2_l0, m_tauf_o0_t1_l0, m_tauf_o0_t0_l0, m_Bfu_z_o0_t5_l0, m_Bfu_z_o0_t4_l0, m_Bfu_z_o0_t3_l0, m_Bfu_z_o0_t2_l0, m_Bfu_z_o0_t1_l0, m_Bfu_z_o0_t0_l0, m_Bfu_y_o0_t5_l0, m_Bfu_y_o0_t4_l0, m_Bfu_y_o0_t3_l0, m_Bfu_y_o0_t2_l0, m_Bfu_y_o0_t1_l0, m_Bfu_y_o0_t0_l0, m_Bfu_x_o0_t5_l0, m_Bfu_x_o0_t4_l0, m_Bfu_x_o0_t3_l0, m_Bfu_x_o0_t2_l0, m_Bfu_x_o0_t1_l0, m_Bfu_x_o0_t0_l0, m_Sfd_z_o0_t10_l0, m_Sfd_z_o0_t9_l0, m_Sfd_z_o0_t8_l0, m_Sfd_z_o0_t7_l0, m_Sfd_z_o0_t6_l0, m_Sfd_z_o0_t5_l0, m_Sfd_z_o0_t4_l0, m_Sfd_z_o0_t3_l0, m_Sfd_z_o0_t2_l0, m_Sfd_z_o0_t1_l0, m_Sfd_z_o0_t0_l0, m_Sfd_y_o0_t10_l0, m_Sfd_y_o0_t9_l0, m_Sfd_y_o0_t8_l0, m_Sfd_y_o0_t7_l0, m_Sfd_y_o0_t6_l0, m_Sfd_y_o0_t5_l0, m_Sfd_y_o0_t4_l0, m_Sfd_y_o0_t3_l0, m_Sfd_y_o0_t2_l0, m_Sfd_y_o0_t1_l0, m_Sfd_y_o0_t0_l0, m_Sfd_x_o0_t10_l0, m_Sfd_x_o0_t9_l0, m_Sfd_x_o0_t8_l0, m_Sfd_x_o0_t7_l0, m_Sfd_x_o0_t6_l0, m_Sfd_x_o0_t5_l0, m_Sfd_x_o0_t4_l0, m_Sfd_x_o0_t3_l0, m_Sfd_x_o0_t2_l0, m_Sfd_x_o0_t1_l0, m_Sfd_x_o0_t0_l0, RHS_gtd_xx, RHS_gtd_xy, RHS_gtd_xz, RHS_gtd_yy, RHS_gtd_yz, RHS_gtd_zz, RHS_Atd_xx, RHS_Atd_xy, RHS_Atd_xz, RHS_Atd_yy, RHS_Atd_yz, RHS_Atd_zz, RHS_Gamh_x, RHS_Gamh_y, RHS_Gamh_z, RHS_Betau_x, RHS_Betau_y, RHS_Betau_z, RHS_Alpha, RHS_chi, RHS_trK, RHS_theta, n_x, n_y, n_z, interaction_index, mod_normal, i_d_Atd_xx_o0_t0_m0_l0, i_d_Atd_xx_o0_t1_m0_l0, i_d_Atd_xx_o0_t2_m0_l0, i_d_Atd_xy_o0_t0_m0_l0, i_d_Atd_xy_o0_t1_m0_l0, i_d_Atd_xy_o0_t2_m0_l0, i_d_Atd_xz_o0_t0_m0_l0, i_d_Atd_xz_o0_t1_m0_l0, i_d_Atd_xz_o0_t2_m0_l0, i_d_Atd_yy_o0_t0_m0_l0, i_d_Atd_yy_o0_t1_m0_l0, i_d_Atd_yy_o0_t2_m0_l0, i_d_Atd_yz_o0_t0_m0_l0, i_d_Atd_yz_o0_t1_m0_l0, i_d_Atd_yz_o0_t2_m0_l0, i_d_Atd_zz_o0_t0_m0_l0, i_d_Atd_zz_o0_t1_m0_l0, i_d_Atd_zz_o0_t2_m0_l0, i_d_gtd_xx_o0_t3_m0_l0, i_d_gtd_xy_o0_t3_m0_l0, i_d_gtd_xz_o0_t3_m0_l0, i_d_gtd_yy_o0_t3_m0_l0, i_d_gtd_yz_o0_t3_m0_l0, i_d_gtd_zz_o0_t3_m0_l0, i_d_Atd_xx_o0_t3_m0_l0, i_d_Atd_xy_o0_t3_m0_l0, i_d_Atd_xz_o0_t3_m0_l0, i_d_Atd_yy_o0_t3_m0_l0, i_d_Atd_yz_o0_t3_m0_l0, i_d_Atd_zz_o0_t3_m0_l0, i_d_Gamh_x_o0_t3_m0_l0, i_d_Gamh_y_o0_t3_m0_l0, i_d_Gamh_z_o0_t3_m0_l0, i_d_Betau_x_o0_t3_m0_l0, i_d_Betau_y_o0_t3_m0_l0, i_d_Betau_z_o0_t3_m0_l0, i_d_Alpha_o0_t3_m0_l0, i_d_chi_o0_t3_m0_l0, i_d_trK_o0_t3_m0_l0, i_d_theta_o0_t3_m0_l0, i_m_gtd_xx_o0_t0_l0, i_m_gtd_xx_o0_t1_l0, i_m_gtd_xx_o0_t2_l0, i_m_gtd_xy_o0_t0_l0, i_m_gtd_xy_o0_t1_l0, i_m_gtd_xy_o0_t2_l0, i_m_gtd_xz_o0_t0_l0, i_m_gtd_xz_o0_t1_l0, i_m_gtd_xz_o0_t2_l0, i_m_gtd_yy_o0_t0_l0, i_m_gtd_yy_o0_t1_l0, i_m_gtd_yy_o0_t2_l0, i_m_gtd_yz_o0_t0_l0, i_m_gtd_yz_o0_t1_l0, i_m_gtd_yz_o0_t2_l0, i_m_gtd_zz_o0_t0_l0, i_m_gtd_zz_o0_t1_l0, i_m_gtd_zz_o0_t2_l0, i_m_Atd_xx_o0_t0_l0, i_m_Atd_xx_o0_t1_l0, i_m_Atd_xx_o0_t2_l0, i_m_Atd_xy_o0_t0_l0, i_m_Atd_xy_o0_t1_l0, i_m_Atd_xy_o0_t2_l0, i_m_Atd_xz_o0_t0_l0, i_m_Atd_xz_o0_t1_l0, i_m_Atd_xz_o0_t2_l0, i_m_Atd_yy_o0_t0_l0, i_m_Atd_yy_o0_t1_l0, i_m_Atd_yy_o0_t2_l0, i_m_Atd_yz_o0_t0_l0, i_m_Atd_yz_o0_t1_l0, i_m_Atd_yz_o0_t2_l0, i_m_Atd_zz_o0_t0_l0, i_m_Atd_zz_o0_t1_l0, i_m_Atd_zz_o0_t2_l0, i_m_Gamh_x_o0_t0_l0, i_m_Gamh_x_o0_t1_l0, i_m_Gamh_x_o0_t2_l0, i_m_Gamh_y_o0_t0_l0, i_m_Gamh_y_o0_t1_l0, i_m_Gamh_y_o0_t2_l0, i_m_Gamh_z_o0_t0_l0, i_m_Gamh_z_o0_t1_l0, i_m_Gamh_z_o0_t2_l0, i_m_Betau_x_o0_t0_l0, i_m_Betau_x_o0_t1_l0, i_m_Betau_x_o0_t2_l0, i_m_Betau_y_o0_t0_l0, i_m_Betau_y_o0_t1_l0, i_m_Betau_y_o0_t2_l0, i_m_Betau_z_o0_t0_l0, i_m_Betau_z_o0_t1_l0, i_m_Betau_z_o0_t2_l0, i_m_Alpha_o0_t0_l0, i_m_Alpha_o0_t1_l0, i_m_Alpha_o0_t2_l0, i_m_chi_o0_t0_l0, i_m_chi_o0_t1_l0, i_m_chi_o0_t2_l0, i_m_trK_o0_t0_l0, i_m_trK_o0_t1_l0, i_m_trK_o0_t2_l0, i_m_theta_o0_t0_l0, i_m_theta_o0_t1_l0, i_m_theta_o0_t2_l0;
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					if ((i + 3 < ilast && i - 3 >= 0 && j + 3 < jlast && j - 3 >= 0 && k + 3 < klast && k - 3 >= 0)) {
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_yy, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_zz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_xy, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_xz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1sqBf = vector(rk1Bfu_x, i, j, k) * rk1Bfd_x + vector(rk1Bfu_y, i, j, k) * rk1Bfd_y + vector(rk1Bfu_z, i, j, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) * vector(rk1gtd_xx, i, j, k) + (-vector(rk1gtd_yy, i, j, k) * (vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k))) + (-vector(rk1gtd_zz, i, j, k) * (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k))) + (-(vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k)) * vector(rk1gtd_xx, i, j, k)) + 2.0 * vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yy, i, j, k) - vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) - vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_yz, i, j, k) - vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_zz, i, j, k)) + vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j, k) + rk1gtu_xy * vector(rk1Sfd_y, i, j, k) + rk1gtu_xz * vector(rk1Sfd_z, i, j, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j, k) + rk1gtu_xy * vector(vfd_y, i, j, k) + rk1gtu_xz * vector(vfd_z, i, j, k));
						rk1Bfvf = vector(rk1Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk1Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk1Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_zz, i, j, k) - vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yz, i, j, k)) + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j, k) + rk1gtu_yz * vector(vfd_y, i, j, k) + rk1gtu_zz * vector(vfd_z, i, j, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j, k) + rk1gtu_yy * vector(rk1Sfd_y, i, j, k) + rk1gtu_yz * vector(rk1Sfd_z, i, j, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j, k) + rk1gtu_yz * vector(rk1Sfd_y, i, j, k) + rk1gtu_zz * vector(rk1Sfd_z, i, j, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j, k) + rk1gtu_yy * vector(vfd_y, i, j, k) + rk1gtu_yz * vector(vfd_z, i, j, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j, k) + rk1vfu_y * vector(vfd_y, i, j, k) + rk1vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j, k) * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j, k) * vector(rk1Bfu_x, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k) * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k) * rk1vfu_z + vector(rk1Bfu_z, i, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k) * vector(rk1Bfu_y, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k) * rk1vfu_y + vector(rk1Bfu_y, i, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k) * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k) * rk1vfu_z + vector(rk1Bfu_z, i, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j, k) * vector(rk1Bfu_y, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j, k));
						FluxiSfd_x = FiSfd_x_mainI(vector(rk1Sfd_x, i, j, k), vector(rk1gtd_xx, i, j, k), vector(rk1gtd_xy, i, j, k), vector(rk1gtd_xz, i, j, k), vector(rk1Betau_x, i, j, k), vector(rk1Alpha, i, j, k), vector(TauT_xx, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_xz, i, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xx, dx, simPlat_dt, ilast, jlast);
						FluxiSfd_y = FiSfd_y_mainI(vector(rk1Sfd_y, i, j, k), vector(rk1gtd_xy, i, j, k), vector(rk1gtd_yy, i, j, k), vector(rk1gtd_yz, i, j, k), vector(rk1Betau_x, i, j, k), vector(rk1Alpha, i, j, k), vector(TauT_xx, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_xz, i, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xy, dx, simPlat_dt, ilast, jlast);
						FluxiSfd_z = FiSfd_z_mainI(vector(rk1Sfd_z, i, j, k), vector(rk1gtd_xz, i, j, k), vector(rk1gtd_yz, i, j, k), vector(rk1gtd_zz, i, j, k), vector(rk1Betau_x, i, j, k), vector(rk1Alpha, i, j, k), vector(TauT_xx, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_xz, i, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xz, dx, simPlat_dt, ilast, jlast);
						FluxiBfu_x = FiBfu_x_mainI(vector(rk1phif, i, j, k), vector(rk1Alpha, i, j, k), rk1chi_max, rk1gtu_xx, dx, simPlat_dt, ilast, jlast);
						FluxiBfu_y = FiBfu_y_mainI(vector(rk1Bfu_x, i, j, k), vector(rk1Bfu_y, i, j, k), vector(rk1phif, i, j, k), vector(rk1Betau_x, i, j, k), vector(rk1Betau_y, i, j, k), vector(rk1Alpha, i, j, k), vector(TauM_xy, i, j, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						FluxiBfu_z = FiBfu_z_mainI(vector(rk1Bfu_x, i, j, k), vector(rk1Bfu_z, i, j, k), vector(rk1phif, i, j, k), vector(rk1Betau_x, i, j, k), vector(rk1Betau_z, i, j, k), vector(rk1Alpha, i, j, k), vector(TauM_xz, i, j, k), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxiDf = FiDf_mainI(vector(rk1Df, i, j, k), vector(rk1Betau_x, i, j, k), vector(rk1Alpha, i, j, k), vector(TauN_x, i, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						FluxiDYf = FiDYf_mainI(vector(rk1DYf, i, j, k), vector(rk1Betau_x, i, j, k), vector(rk1Alpha, i, j, k), vector(TauNe_x, i, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxitauf = Fitauf_mainI(vector(rk1Df, i, j, k), vector(rk1tauf, i, j, k), vector(rk1Betau_x, i, j, k), vector(rk1Alpha, i, j, k), vector(TauN_x, i, j, k), rk1sdetg, rk1Sfu_x, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxiphif = Fiphif_mainI(vector(rk1Bfu_x, i, j, k), vector(rk1phif, i, j, k), vector(rk1Betau_x, i, j, k), vector(rk1Alpha, i, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j, k) + rk1sqca - vector(sqcs, i, j, k) * rk1sqca;
						SpeediSfd_x = fabs(fabs(vector(rk1Betau_x, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x))))));
						SpeediSfd_y = fabs(fabs(vector(rk1Betau_x, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x))))));
						SpeediSfd_z = fabs(fabs(vector(rk1Betau_x, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x))))));
						SpeediBfu_x = fabs(fabs(vector(rk1Betau_x, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x))))));
						SpeediBfu_y = fabs(fabs(vector(rk1Betau_x, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x))))));
						SpeediBfu_z = fabs(fabs(vector(rk1Betau_x, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x))))));
						SpeediDf = fabs(fabs(vector(rk1Betau_x, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x))))));
						SpeediDYf = fabs(fabs(vector(rk1Betau_x, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x))))));
						Speeditauf = fabs(fabs(vector(rk1Betau_x, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x))))));
						Speediphif = fabs(fabs(vector(rk1Betau_x, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i - 3, j, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i - 3, j, k) * vector(rk1Bfu_x, i - 3, j, k) + vector(rk1gtd_yy, i - 3, j, k) * vector(rk1Bfu_y, i - 3, j, k) + vector(rk1gtd_yz, i - 3, j, k) * vector(rk1Bfu_z, i - 3, j, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i - 3, j, k) * vector(rk1Bfu_x, i - 3, j, k) + vector(rk1gtd_yz, i - 3, j, k) * vector(rk1Bfu_y, i - 3, j, k) + vector(rk1gtd_zz, i - 3, j, k) * vector(rk1Bfu_z, i - 3, j, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i - 3, j, k) * vector(rk1Bfu_x, i - 3, j, k) + vector(rk1gtd_xy, i - 3, j, k) * vector(rk1Bfu_y, i - 3, j, k) + vector(rk1gtd_xz, i - 3, j, k) * vector(rk1Bfu_z, i - 3, j, k));
						rk1sqBf = vector(rk1Bfu_x, i - 3, j, k) * rk1Bfd_x + vector(rk1Bfu_y, i - 3, j, k) * rk1Bfd_y + vector(rk1Bfu_z, i - 3, j, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i - 3, j, k) * vector(rk1gtd_zz, i - 3, j, k) * vector(rk1gtd_xx, i - 3, j, k) + (-vector(rk1gtd_yy, i - 3, j, k) * (vector(rk1gtd_xz, i - 3, j, k) * vector(rk1gtd_xz, i - 3, j, k))) + (-vector(rk1gtd_zz, i - 3, j, k) * (vector(rk1gtd_xy, i - 3, j, k) * vector(rk1gtd_xy, i - 3, j, k))) + (-(vector(rk1gtd_yz, i - 3, j, k) * vector(rk1gtd_yz, i - 3, j, k)) * vector(rk1gtd_xx, i - 3, j, k)) + 2.0 * vector(rk1gtd_yz, i - 3, j, k) * vector(rk1gtd_xy, i - 3, j, k) * vector(rk1gtd_xz, i - 3, j, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i - 3, j, k) * vector(rk1gtd_yy, i - 3, j, k) - vector(rk1gtd_xy, i - 3, j, k) * vector(rk1gtd_xy, i - 3, j, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i - 3, j, k) * vector(rk1gtd_zz, i - 3, j, k) - vector(rk1gtd_yz, i - 3, j, k) * vector(rk1gtd_yz, i - 3, j, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i - 3, j, k) * vector(rk1gtd_yz, i - 3, j, k) - vector(rk1gtd_yy, i - 3, j, k) * vector(rk1gtd_xz, i - 3, j, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i - 3, j, k) * vector(rk1gtd_zz, i - 3, j, k)) + vector(rk1gtd_yz, i - 3, j, k) * vector(rk1gtd_xz, i - 3, j, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i - 3, j, k) + rk1gtu_xy * vector(rk1Sfd_y, i - 3, j, k) + rk1gtu_xz * vector(rk1Sfd_z, i - 3, j, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i - 3, j, k) + rk1gtu_xy * vector(vfd_y, i - 3, j, k) + rk1gtu_xz * vector(vfd_z, i - 3, j, k));
						rk1Bfvf = vector(rk1Bfu_x, i - 3, j, k) * vector(vfd_x, i - 3, j, k) + vector(rk1Bfu_y, i - 3, j, k) * vector(vfd_y, i - 3, j, k) + vector(rk1Bfu_z, i - 3, j, k) * vector(vfd_z, i - 3, j, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i - 3, j, k) * vector(rk1gtd_zz, i - 3, j, k) - vector(rk1gtd_xz, i - 3, j, k) * vector(rk1gtd_xz, i - 3, j, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i - 3, j, k) * vector(rk1gtd_yz, i - 3, j, k)) + vector(rk1gtd_xy, i - 3, j, k) * vector(rk1gtd_xz, i - 3, j, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i - 3, j, k) + rk1gtu_yz * vector(vfd_y, i - 3, j, k) + rk1gtu_zz * vector(vfd_z, i - 3, j, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i - 3, j, k) + rk1gtu_yy * vector(rk1Sfd_y, i - 3, j, k) + rk1gtu_yz * vector(rk1Sfd_z, i - 3, j, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i - 3, j, k) + rk1gtu_yz * vector(rk1Sfd_y, i - 3, j, k) + rk1gtu_zz * vector(rk1Sfd_z, i - 3, j, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i - 3, j, k) + rk1gtu_yy * vector(vfd_y, i - 3, j, k) + rk1gtu_yz * vector(vfd_z, i - 3, j, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i - 3, j, k) + rk1vfu_y * vector(vfd_y, i - 3, j, k) + rk1vfu_z * vector(vfd_z, i - 3, j, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i - 3, j, k) * vector(rk1Bfu_z, i - 3, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i - 3, j, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i - 3, j, k) * vector(rk1Bfu_x, i - 3, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i - 3, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i - 3, j, k) * vector(rk1Bfu_z, i - 3, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i - 3, j, k) * rk1vfu_z + vector(rk1Bfu_z, i - 3, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i - 3, j, k) * vector(rk1Bfu_y, i - 3, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i - 3, j, k) * rk1vfu_y + vector(rk1Bfu_y, i - 3, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i - 3, j, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i - 3, j, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i - 3, j, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i - 3, j, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i - 3, j, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i - 3, j, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i - 3, j, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i - 3, j, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i - 3, j, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i - 3, j, k) * vector(rk1Bfu_z, i - 3, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i - 3, j, k) * rk1vfu_z + vector(rk1Bfu_z, i - 3, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i - 3, j, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i - 3, j, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i - 3, j, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i - 3, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i - 3, j, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i - 3, j, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i - 3, j, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i - 3, j, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i - 3, j, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i - 3, j, k) * vector(rk1Bfu_y, i - 3, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i - 3, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i - 3, j, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i - 3, j, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i - 3, j, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i - 3, j, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i - 3, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i - 3, j, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i - 3, j, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i - 3, j, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i - 3, j, k));
						Fluxm3iSfd_x = FiSfd_x_mainI(vector(rk1Sfd_x, i - 3, j, k), vector(rk1gtd_xx, i - 3, j, k), vector(rk1gtd_xy, i - 3, j, k), vector(rk1gtd_xz, i - 3, j, k), vector(rk1Betau_x, i - 3, j, k), vector(rk1Alpha, i - 3, j, k), vector(TauT_xx, i - 3, j, k), vector(TauT_xy, i - 3, j, k), vector(TauT_xz, i - 3, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm3iSfd_y = FiSfd_y_mainI(vector(rk1Sfd_y, i - 3, j, k), vector(rk1gtd_xy, i - 3, j, k), vector(rk1gtd_yy, i - 3, j, k), vector(rk1gtd_yz, i - 3, j, k), vector(rk1Betau_x, i - 3, j, k), vector(rk1Alpha, i - 3, j, k), vector(TauT_xx, i - 3, j, k), vector(TauT_xy, i - 3, j, k), vector(TauT_xz, i - 3, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxm3iSfd_z = FiSfd_z_mainI(vector(rk1Sfd_z, i - 3, j, k), vector(rk1gtd_xz, i - 3, j, k), vector(rk1gtd_yz, i - 3, j, k), vector(rk1gtd_zz, i - 3, j, k), vector(rk1Betau_x, i - 3, j, k), vector(rk1Alpha, i - 3, j, k), vector(TauT_xx, i - 3, j, k), vector(TauT_xy, i - 3, j, k), vector(TauT_xz, i - 3, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxm3iBfu_x = FiBfu_x_mainI(vector(rk1phif, i - 3, j, k), vector(rk1Alpha, i - 3, j, k), rk1chi_max, rk1gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm3iBfu_y = FiBfu_y_mainI(vector(rk1Bfu_x, i - 3, j, k), vector(rk1Bfu_y, i - 3, j, k), vector(rk1phif, i - 3, j, k), vector(rk1Betau_x, i - 3, j, k), vector(rk1Betau_y, i - 3, j, k), vector(rk1Alpha, i - 3, j, k), vector(TauM_xy, i - 3, j, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3iBfu_z = FiBfu_z_mainI(vector(rk1Bfu_x, i - 3, j, k), vector(rk1Bfu_z, i - 3, j, k), vector(rk1phif, i - 3, j, k), vector(rk1Betau_x, i - 3, j, k), vector(rk1Betau_z, i - 3, j, k), vector(rk1Alpha, i - 3, j, k), vector(TauM_xz, i - 3, j, k), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3iDf = FiDf_mainI(vector(rk1Df, i - 3, j, k), vector(rk1Betau_x, i - 3, j, k), vector(rk1Alpha, i - 3, j, k), vector(TauN_x, i - 3, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm3iDYf = FiDYf_mainI(vector(rk1DYf, i - 3, j, k), vector(rk1Betau_x, i - 3, j, k), vector(rk1Alpha, i - 3, j, k), vector(TauNe_x, i - 3, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm3itauf = Fitauf_mainI(vector(rk1Df, i - 3, j, k), vector(rk1tauf, i - 3, j, k), vector(rk1Betau_x, i - 3, j, k), vector(rk1Alpha, i - 3, j, k), vector(TauN_x, i - 3, j, k), rk1sdetg, rk1Sfu_x, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm3iphif = Fiphif_mainI(vector(rk1Bfu_x, i - 3, j, k), vector(rk1phif, i - 3, j, k), vector(rk1Betau_x, i - 3, j, k), vector(rk1Alpha, i - 3, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i - 3, j, k) * (1.0 + vector(epsf, i - 3, j, k)) + vector(pf, i - 3, j, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i - 3, j, k) + rk1sqca - vector(sqcs, i - 3, j, k) * rk1sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk1Betau_x, i - 3, j, k)) + vector(rk1Alpha, i - 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk1Betau_x, i - 3, j, k)) + vector(rk1Alpha, i - 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk1Betau_x, i - 3, j, k)) + vector(rk1Alpha, i - 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk1Betau_x, i - 3, j, k)) + vector(rk1Alpha, i - 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk1Betau_x, i - 3, j, k)) + vector(rk1Alpha, i - 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk1Betau_x, i - 3, j, k)) + vector(rk1Alpha, i - 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk1Betau_x, i - 3, j, k)) + vector(rk1Alpha, i - 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk1Betau_x, i - 3, j, k)) + vector(rk1Alpha, i - 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk1Betau_x, i - 3, j, k)) + vector(rk1Alpha, i - 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk1Betau_x, i - 3, j, k)) + vector(rk1Alpha, i - 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i - 2, j, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i - 2, j, k) * vector(rk1Bfu_x, i - 2, j, k) + vector(rk1gtd_yy, i - 2, j, k) * vector(rk1Bfu_y, i - 2, j, k) + vector(rk1gtd_yz, i - 2, j, k) * vector(rk1Bfu_z, i - 2, j, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i - 2, j, k) * vector(rk1Bfu_x, i - 2, j, k) + vector(rk1gtd_yz, i - 2, j, k) * vector(rk1Bfu_y, i - 2, j, k) + vector(rk1gtd_zz, i - 2, j, k) * vector(rk1Bfu_z, i - 2, j, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i - 2, j, k) * vector(rk1Bfu_x, i - 2, j, k) + vector(rk1gtd_xy, i - 2, j, k) * vector(rk1Bfu_y, i - 2, j, k) + vector(rk1gtd_xz, i - 2, j, k) * vector(rk1Bfu_z, i - 2, j, k));
						rk1sqBf = vector(rk1Bfu_x, i - 2, j, k) * rk1Bfd_x + vector(rk1Bfu_y, i - 2, j, k) * rk1Bfd_y + vector(rk1Bfu_z, i - 2, j, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i - 2, j, k) * vector(rk1gtd_zz, i - 2, j, k) * vector(rk1gtd_xx, i - 2, j, k) + (-vector(rk1gtd_yy, i - 2, j, k) * (vector(rk1gtd_xz, i - 2, j, k) * vector(rk1gtd_xz, i - 2, j, k))) + (-vector(rk1gtd_zz, i - 2, j, k) * (vector(rk1gtd_xy, i - 2, j, k) * vector(rk1gtd_xy, i - 2, j, k))) + (-(vector(rk1gtd_yz, i - 2, j, k) * vector(rk1gtd_yz, i - 2, j, k)) * vector(rk1gtd_xx, i - 2, j, k)) + 2.0 * vector(rk1gtd_yz, i - 2, j, k) * vector(rk1gtd_xy, i - 2, j, k) * vector(rk1gtd_xz, i - 2, j, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i - 2, j, k) * vector(rk1gtd_yy, i - 2, j, k) - vector(rk1gtd_xy, i - 2, j, k) * vector(rk1gtd_xy, i - 2, j, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i - 2, j, k) * vector(rk1gtd_zz, i - 2, j, k) - vector(rk1gtd_yz, i - 2, j, k) * vector(rk1gtd_yz, i - 2, j, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i - 2, j, k) * vector(rk1gtd_yz, i - 2, j, k) - vector(rk1gtd_yy, i - 2, j, k) * vector(rk1gtd_xz, i - 2, j, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i - 2, j, k) * vector(rk1gtd_zz, i - 2, j, k)) + vector(rk1gtd_yz, i - 2, j, k) * vector(rk1gtd_xz, i - 2, j, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i - 2, j, k) + rk1gtu_xy * vector(rk1Sfd_y, i - 2, j, k) + rk1gtu_xz * vector(rk1Sfd_z, i - 2, j, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i - 2, j, k) + rk1gtu_xy * vector(vfd_y, i - 2, j, k) + rk1gtu_xz * vector(vfd_z, i - 2, j, k));
						rk1Bfvf = vector(rk1Bfu_x, i - 2, j, k) * vector(vfd_x, i - 2, j, k) + vector(rk1Bfu_y, i - 2, j, k) * vector(vfd_y, i - 2, j, k) + vector(rk1Bfu_z, i - 2, j, k) * vector(vfd_z, i - 2, j, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i - 2, j, k) * vector(rk1gtd_zz, i - 2, j, k) - vector(rk1gtd_xz, i - 2, j, k) * vector(rk1gtd_xz, i - 2, j, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i - 2, j, k) * vector(rk1gtd_yz, i - 2, j, k)) + vector(rk1gtd_xy, i - 2, j, k) * vector(rk1gtd_xz, i - 2, j, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i - 2, j, k) + rk1gtu_yz * vector(vfd_y, i - 2, j, k) + rk1gtu_zz * vector(vfd_z, i - 2, j, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i - 2, j, k) + rk1gtu_yy * vector(rk1Sfd_y, i - 2, j, k) + rk1gtu_yz * vector(rk1Sfd_z, i - 2, j, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i - 2, j, k) + rk1gtu_yz * vector(rk1Sfd_y, i - 2, j, k) + rk1gtu_zz * vector(rk1Sfd_z, i - 2, j, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i - 2, j, k) + rk1gtu_yy * vector(vfd_y, i - 2, j, k) + rk1gtu_yz * vector(vfd_z, i - 2, j, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i - 2, j, k) + rk1vfu_y * vector(vfd_y, i - 2, j, k) + rk1vfu_z * vector(vfd_z, i - 2, j, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i - 2, j, k) * vector(rk1Bfu_z, i - 2, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i - 2, j, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i - 2, j, k) * vector(rk1Bfu_x, i - 2, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i - 2, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i - 2, j, k) * vector(rk1Bfu_z, i - 2, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i - 2, j, k) * rk1vfu_z + vector(rk1Bfu_z, i - 2, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i - 2, j, k) * vector(rk1Bfu_y, i - 2, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i - 2, j, k) * rk1vfu_y + vector(rk1Bfu_y, i - 2, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i - 2, j, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i - 2, j, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i - 2, j, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i - 2, j, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i - 2, j, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i - 2, j, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i - 2, j, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i - 2, j, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i - 2, j, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i - 2, j, k) * vector(rk1Bfu_z, i - 2, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i - 2, j, k) * rk1vfu_z + vector(rk1Bfu_z, i - 2, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i - 2, j, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i - 2, j, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i - 2, j, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i - 2, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i - 2, j, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i - 2, j, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i - 2, j, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i - 2, j, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i - 2, j, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i - 2, j, k) * vector(rk1Bfu_y, i - 2, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i - 2, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i - 2, j, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i - 2, j, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i - 2, j, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i - 2, j, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i - 2, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i - 2, j, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i - 2, j, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i - 2, j, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i - 2, j, k));
						Fluxm2iSfd_x = FiSfd_x_mainI(vector(rk1Sfd_x, i - 2, j, k), vector(rk1gtd_xx, i - 2, j, k), vector(rk1gtd_xy, i - 2, j, k), vector(rk1gtd_xz, i - 2, j, k), vector(rk1Betau_x, i - 2, j, k), vector(rk1Alpha, i - 2, j, k), vector(TauT_xx, i - 2, j, k), vector(TauT_xy, i - 2, j, k), vector(TauT_xz, i - 2, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm2iSfd_y = FiSfd_y_mainI(vector(rk1Sfd_y, i - 2, j, k), vector(rk1gtd_xy, i - 2, j, k), vector(rk1gtd_yy, i - 2, j, k), vector(rk1gtd_yz, i - 2, j, k), vector(rk1Betau_x, i - 2, j, k), vector(rk1Alpha, i - 2, j, k), vector(TauT_xx, i - 2, j, k), vector(TauT_xy, i - 2, j, k), vector(TauT_xz, i - 2, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxm2iSfd_z = FiSfd_z_mainI(vector(rk1Sfd_z, i - 2, j, k), vector(rk1gtd_xz, i - 2, j, k), vector(rk1gtd_yz, i - 2, j, k), vector(rk1gtd_zz, i - 2, j, k), vector(rk1Betau_x, i - 2, j, k), vector(rk1Alpha, i - 2, j, k), vector(TauT_xx, i - 2, j, k), vector(TauT_xy, i - 2, j, k), vector(TauT_xz, i - 2, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxm2iBfu_x = FiBfu_x_mainI(vector(rk1phif, i - 2, j, k), vector(rk1Alpha, i - 2, j, k), rk1chi_max, rk1gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm2iBfu_y = FiBfu_y_mainI(vector(rk1Bfu_x, i - 2, j, k), vector(rk1Bfu_y, i - 2, j, k), vector(rk1phif, i - 2, j, k), vector(rk1Betau_x, i - 2, j, k), vector(rk1Betau_y, i - 2, j, k), vector(rk1Alpha, i - 2, j, k), vector(TauM_xy, i - 2, j, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2iBfu_z = FiBfu_z_mainI(vector(rk1Bfu_x, i - 2, j, k), vector(rk1Bfu_z, i - 2, j, k), vector(rk1phif, i - 2, j, k), vector(rk1Betau_x, i - 2, j, k), vector(rk1Betau_z, i - 2, j, k), vector(rk1Alpha, i - 2, j, k), vector(TauM_xz, i - 2, j, k), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2iDf = FiDf_mainI(vector(rk1Df, i - 2, j, k), vector(rk1Betau_x, i - 2, j, k), vector(rk1Alpha, i - 2, j, k), vector(TauN_x, i - 2, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm2iDYf = FiDYf_mainI(vector(rk1DYf, i - 2, j, k), vector(rk1Betau_x, i - 2, j, k), vector(rk1Alpha, i - 2, j, k), vector(TauNe_x, i - 2, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm2itauf = Fitauf_mainI(vector(rk1Df, i - 2, j, k), vector(rk1tauf, i - 2, j, k), vector(rk1Betau_x, i - 2, j, k), vector(rk1Alpha, i - 2, j, k), vector(TauN_x, i - 2, j, k), rk1sdetg, rk1Sfu_x, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm2iphif = Fiphif_mainI(vector(rk1Bfu_x, i - 2, j, k), vector(rk1phif, i - 2, j, k), vector(rk1Betau_x, i - 2, j, k), vector(rk1Alpha, i - 2, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i - 2, j, k) * (1.0 + vector(epsf, i - 2, j, k)) + vector(pf, i - 2, j, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i - 2, j, k) + rk1sqca - vector(sqcs, i - 2, j, k) * rk1sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk1Betau_x, i - 2, j, k)) + vector(rk1Alpha, i - 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk1Betau_x, i - 2, j, k)) + vector(rk1Alpha, i - 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk1Betau_x, i - 2, j, k)) + vector(rk1Alpha, i - 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk1Betau_x, i - 2, j, k)) + vector(rk1Alpha, i - 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk1Betau_x, i - 2, j, k)) + vector(rk1Alpha, i - 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk1Betau_x, i - 2, j, k)) + vector(rk1Alpha, i - 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk1Betau_x, i - 2, j, k)) + vector(rk1Alpha, i - 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk1Betau_x, i - 2, j, k)) + vector(rk1Alpha, i - 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk1Betau_x, i - 2, j, k)) + vector(rk1Alpha, i - 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk1Betau_x, i - 2, j, k)) + vector(rk1Alpha, i - 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i - 1, j, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i - 1, j, k) * vector(rk1Bfu_x, i - 1, j, k) + vector(rk1gtd_yy, i - 1, j, k) * vector(rk1Bfu_y, i - 1, j, k) + vector(rk1gtd_yz, i - 1, j, k) * vector(rk1Bfu_z, i - 1, j, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i - 1, j, k) * vector(rk1Bfu_x, i - 1, j, k) + vector(rk1gtd_yz, i - 1, j, k) * vector(rk1Bfu_y, i - 1, j, k) + vector(rk1gtd_zz, i - 1, j, k) * vector(rk1Bfu_z, i - 1, j, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i - 1, j, k) * vector(rk1Bfu_x, i - 1, j, k) + vector(rk1gtd_xy, i - 1, j, k) * vector(rk1Bfu_y, i - 1, j, k) + vector(rk1gtd_xz, i - 1, j, k) * vector(rk1Bfu_z, i - 1, j, k));
						rk1sqBf = vector(rk1Bfu_x, i - 1, j, k) * rk1Bfd_x + vector(rk1Bfu_y, i - 1, j, k) * rk1Bfd_y + vector(rk1Bfu_z, i - 1, j, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i - 1, j, k) * vector(rk1gtd_zz, i - 1, j, k) * vector(rk1gtd_xx, i - 1, j, k) + (-vector(rk1gtd_yy, i - 1, j, k) * (vector(rk1gtd_xz, i - 1, j, k) * vector(rk1gtd_xz, i - 1, j, k))) + (-vector(rk1gtd_zz, i - 1, j, k) * (vector(rk1gtd_xy, i - 1, j, k) * vector(rk1gtd_xy, i - 1, j, k))) + (-(vector(rk1gtd_yz, i - 1, j, k) * vector(rk1gtd_yz, i - 1, j, k)) * vector(rk1gtd_xx, i - 1, j, k)) + 2.0 * vector(rk1gtd_yz, i - 1, j, k) * vector(rk1gtd_xy, i - 1, j, k) * vector(rk1gtd_xz, i - 1, j, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i - 1, j, k) * vector(rk1gtd_yy, i - 1, j, k) - vector(rk1gtd_xy, i - 1, j, k) * vector(rk1gtd_xy, i - 1, j, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i - 1, j, k) * vector(rk1gtd_zz, i - 1, j, k) - vector(rk1gtd_yz, i - 1, j, k) * vector(rk1gtd_yz, i - 1, j, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i - 1, j, k) * vector(rk1gtd_yz, i - 1, j, k) - vector(rk1gtd_yy, i - 1, j, k) * vector(rk1gtd_xz, i - 1, j, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i - 1, j, k) * vector(rk1gtd_zz, i - 1, j, k)) + vector(rk1gtd_yz, i - 1, j, k) * vector(rk1gtd_xz, i - 1, j, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i - 1, j, k) + rk1gtu_xy * vector(rk1Sfd_y, i - 1, j, k) + rk1gtu_xz * vector(rk1Sfd_z, i - 1, j, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i - 1, j, k) + rk1gtu_xy * vector(vfd_y, i - 1, j, k) + rk1gtu_xz * vector(vfd_z, i - 1, j, k));
						rk1Bfvf = vector(rk1Bfu_x, i - 1, j, k) * vector(vfd_x, i - 1, j, k) + vector(rk1Bfu_y, i - 1, j, k) * vector(vfd_y, i - 1, j, k) + vector(rk1Bfu_z, i - 1, j, k) * vector(vfd_z, i - 1, j, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i - 1, j, k) * vector(rk1gtd_zz, i - 1, j, k) - vector(rk1gtd_xz, i - 1, j, k) * vector(rk1gtd_xz, i - 1, j, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i - 1, j, k) * vector(rk1gtd_yz, i - 1, j, k)) + vector(rk1gtd_xy, i - 1, j, k) * vector(rk1gtd_xz, i - 1, j, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i - 1, j, k) + rk1gtu_yz * vector(vfd_y, i - 1, j, k) + rk1gtu_zz * vector(vfd_z, i - 1, j, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i - 1, j, k) + rk1gtu_yy * vector(rk1Sfd_y, i - 1, j, k) + rk1gtu_yz * vector(rk1Sfd_z, i - 1, j, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i - 1, j, k) + rk1gtu_yz * vector(rk1Sfd_y, i - 1, j, k) + rk1gtu_zz * vector(rk1Sfd_z, i - 1, j, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i - 1, j, k) + rk1gtu_yy * vector(vfd_y, i - 1, j, k) + rk1gtu_yz * vector(vfd_z, i - 1, j, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i - 1, j, k) + rk1vfu_y * vector(vfd_y, i - 1, j, k) + rk1vfu_z * vector(vfd_z, i - 1, j, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i - 1, j, k) * vector(rk1Bfu_z, i - 1, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i - 1, j, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i - 1, j, k) * vector(rk1Bfu_x, i - 1, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i - 1, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i - 1, j, k) * vector(rk1Bfu_z, i - 1, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i - 1, j, k) * rk1vfu_z + vector(rk1Bfu_z, i - 1, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i - 1, j, k) * vector(rk1Bfu_y, i - 1, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i - 1, j, k) * rk1vfu_y + vector(rk1Bfu_y, i - 1, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i - 1, j, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i - 1, j, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i - 1, j, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i - 1, j, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i - 1, j, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i - 1, j, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i - 1, j, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i - 1, j, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i - 1, j, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i - 1, j, k) * vector(rk1Bfu_z, i - 1, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i - 1, j, k) * rk1vfu_z + vector(rk1Bfu_z, i - 1, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i - 1, j, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i - 1, j, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i - 1, j, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i - 1, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i - 1, j, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i - 1, j, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i - 1, j, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i - 1, j, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i - 1, j, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i - 1, j, k) * vector(rk1Bfu_y, i - 1, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i - 1, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i - 1, j, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i - 1, j, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i - 1, j, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i - 1, j, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i - 1, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i - 1, j, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i - 1, j, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i - 1, j, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i - 1, j, k));
						Fluxm1iSfd_x = FiSfd_x_mainI(vector(rk1Sfd_x, i - 1, j, k), vector(rk1gtd_xx, i - 1, j, k), vector(rk1gtd_xy, i - 1, j, k), vector(rk1gtd_xz, i - 1, j, k), vector(rk1Betau_x, i - 1, j, k), vector(rk1Alpha, i - 1, j, k), vector(TauT_xx, i - 1, j, k), vector(TauT_xy, i - 1, j, k), vector(TauT_xz, i - 1, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm1iSfd_y = FiSfd_y_mainI(vector(rk1Sfd_y, i - 1, j, k), vector(rk1gtd_xy, i - 1, j, k), vector(rk1gtd_yy, i - 1, j, k), vector(rk1gtd_yz, i - 1, j, k), vector(rk1Betau_x, i - 1, j, k), vector(rk1Alpha, i - 1, j, k), vector(TauT_xx, i - 1, j, k), vector(TauT_xy, i - 1, j, k), vector(TauT_xz, i - 1, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxm1iSfd_z = FiSfd_z_mainI(vector(rk1Sfd_z, i - 1, j, k), vector(rk1gtd_xz, i - 1, j, k), vector(rk1gtd_yz, i - 1, j, k), vector(rk1gtd_zz, i - 1, j, k), vector(rk1Betau_x, i - 1, j, k), vector(rk1Alpha, i - 1, j, k), vector(TauT_xx, i - 1, j, k), vector(TauT_xy, i - 1, j, k), vector(TauT_xz, i - 1, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxm1iBfu_x = FiBfu_x_mainI(vector(rk1phif, i - 1, j, k), vector(rk1Alpha, i - 1, j, k), rk1chi_max, rk1gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm1iBfu_y = FiBfu_y_mainI(vector(rk1Bfu_x, i - 1, j, k), vector(rk1Bfu_y, i - 1, j, k), vector(rk1phif, i - 1, j, k), vector(rk1Betau_x, i - 1, j, k), vector(rk1Betau_y, i - 1, j, k), vector(rk1Alpha, i - 1, j, k), vector(TauM_xy, i - 1, j, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1iBfu_z = FiBfu_z_mainI(vector(rk1Bfu_x, i - 1, j, k), vector(rk1Bfu_z, i - 1, j, k), vector(rk1phif, i - 1, j, k), vector(rk1Betau_x, i - 1, j, k), vector(rk1Betau_z, i - 1, j, k), vector(rk1Alpha, i - 1, j, k), vector(TauM_xz, i - 1, j, k), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1iDf = FiDf_mainI(vector(rk1Df, i - 1, j, k), vector(rk1Betau_x, i - 1, j, k), vector(rk1Alpha, i - 1, j, k), vector(TauN_x, i - 1, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm1iDYf = FiDYf_mainI(vector(rk1DYf, i - 1, j, k), vector(rk1Betau_x, i - 1, j, k), vector(rk1Alpha, i - 1, j, k), vector(TauNe_x, i - 1, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm1itauf = Fitauf_mainI(vector(rk1Df, i - 1, j, k), vector(rk1tauf, i - 1, j, k), vector(rk1Betau_x, i - 1, j, k), vector(rk1Alpha, i - 1, j, k), vector(TauN_x, i - 1, j, k), rk1sdetg, rk1Sfu_x, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm1iphif = Fiphif_mainI(vector(rk1Bfu_x, i - 1, j, k), vector(rk1phif, i - 1, j, k), vector(rk1Betau_x, i - 1, j, k), vector(rk1Alpha, i - 1, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i - 1, j, k) * (1.0 + vector(epsf, i - 1, j, k)) + vector(pf, i - 1, j, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i - 1, j, k) + rk1sqca - vector(sqcs, i - 1, j, k) * rk1sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk1Betau_x, i - 1, j, k)) + vector(rk1Alpha, i - 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk1Betau_x, i - 1, j, k)) + vector(rk1Alpha, i - 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk1Betau_x, i - 1, j, k)) + vector(rk1Alpha, i - 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk1Betau_x, i - 1, j, k)) + vector(rk1Alpha, i - 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk1Betau_x, i - 1, j, k)) + vector(rk1Alpha, i - 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk1Betau_x, i - 1, j, k)) + vector(rk1Alpha, i - 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk1Betau_x, i - 1, j, k)) + vector(rk1Alpha, i - 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk1Betau_x, i - 1, j, k)) + vector(rk1Alpha, i - 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk1Betau_x, i - 1, j, k)) + vector(rk1Alpha, i - 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk1Betau_x, i - 1, j, k)) + vector(rk1Alpha, i - 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i + 1, j, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i + 1, j, k) * vector(rk1Bfu_x, i + 1, j, k) + vector(rk1gtd_yy, i + 1, j, k) * vector(rk1Bfu_y, i + 1, j, k) + vector(rk1gtd_yz, i + 1, j, k) * vector(rk1Bfu_z, i + 1, j, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i + 1, j, k) * vector(rk1Bfu_x, i + 1, j, k) + vector(rk1gtd_yz, i + 1, j, k) * vector(rk1Bfu_y, i + 1, j, k) + vector(rk1gtd_zz, i + 1, j, k) * vector(rk1Bfu_z, i + 1, j, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i + 1, j, k) * vector(rk1Bfu_x, i + 1, j, k) + vector(rk1gtd_xy, i + 1, j, k) * vector(rk1Bfu_y, i + 1, j, k) + vector(rk1gtd_xz, i + 1, j, k) * vector(rk1Bfu_z, i + 1, j, k));
						rk1sqBf = vector(rk1Bfu_x, i + 1, j, k) * rk1Bfd_x + vector(rk1Bfu_y, i + 1, j, k) * rk1Bfd_y + vector(rk1Bfu_z, i + 1, j, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i + 1, j, k) * vector(rk1gtd_zz, i + 1, j, k) * vector(rk1gtd_xx, i + 1, j, k) + (-vector(rk1gtd_yy, i + 1, j, k) * (vector(rk1gtd_xz, i + 1, j, k) * vector(rk1gtd_xz, i + 1, j, k))) + (-vector(rk1gtd_zz, i + 1, j, k) * (vector(rk1gtd_xy, i + 1, j, k) * vector(rk1gtd_xy, i + 1, j, k))) + (-(vector(rk1gtd_yz, i + 1, j, k) * vector(rk1gtd_yz, i + 1, j, k)) * vector(rk1gtd_xx, i + 1, j, k)) + 2.0 * vector(rk1gtd_yz, i + 1, j, k) * vector(rk1gtd_xy, i + 1, j, k) * vector(rk1gtd_xz, i + 1, j, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i + 1, j, k) * vector(rk1gtd_yy, i + 1, j, k) - vector(rk1gtd_xy, i + 1, j, k) * vector(rk1gtd_xy, i + 1, j, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i + 1, j, k) * vector(rk1gtd_zz, i + 1, j, k) - vector(rk1gtd_yz, i + 1, j, k) * vector(rk1gtd_yz, i + 1, j, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i + 1, j, k) * vector(rk1gtd_yz, i + 1, j, k) - vector(rk1gtd_yy, i + 1, j, k) * vector(rk1gtd_xz, i + 1, j, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i + 1, j, k) * vector(rk1gtd_zz, i + 1, j, k)) + vector(rk1gtd_yz, i + 1, j, k) * vector(rk1gtd_xz, i + 1, j, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i + 1, j, k) + rk1gtu_xy * vector(rk1Sfd_y, i + 1, j, k) + rk1gtu_xz * vector(rk1Sfd_z, i + 1, j, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i + 1, j, k) + rk1gtu_xy * vector(vfd_y, i + 1, j, k) + rk1gtu_xz * vector(vfd_z, i + 1, j, k));
						rk1Bfvf = vector(rk1Bfu_x, i + 1, j, k) * vector(vfd_x, i + 1, j, k) + vector(rk1Bfu_y, i + 1, j, k) * vector(vfd_y, i + 1, j, k) + vector(rk1Bfu_z, i + 1, j, k) * vector(vfd_z, i + 1, j, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i + 1, j, k) * vector(rk1gtd_zz, i + 1, j, k) - vector(rk1gtd_xz, i + 1, j, k) * vector(rk1gtd_xz, i + 1, j, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i + 1, j, k) * vector(rk1gtd_yz, i + 1, j, k)) + vector(rk1gtd_xy, i + 1, j, k) * vector(rk1gtd_xz, i + 1, j, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i + 1, j, k) + rk1gtu_yz * vector(vfd_y, i + 1, j, k) + rk1gtu_zz * vector(vfd_z, i + 1, j, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i + 1, j, k) + rk1gtu_yy * vector(rk1Sfd_y, i + 1, j, k) + rk1gtu_yz * vector(rk1Sfd_z, i + 1, j, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i + 1, j, k) + rk1gtu_yz * vector(rk1Sfd_y, i + 1, j, k) + rk1gtu_zz * vector(rk1Sfd_z, i + 1, j, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i + 1, j, k) + rk1gtu_yy * vector(vfd_y, i + 1, j, k) + rk1gtu_yz * vector(vfd_z, i + 1, j, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i + 1, j, k) + rk1vfu_y * vector(vfd_y, i + 1, j, k) + rk1vfu_z * vector(vfd_z, i + 1, j, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i + 1, j, k) * vector(rk1Bfu_z, i + 1, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i + 1, j, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i + 1, j, k) * vector(rk1Bfu_x, i + 1, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i + 1, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i + 1, j, k) * vector(rk1Bfu_z, i + 1, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i + 1, j, k) * rk1vfu_z + vector(rk1Bfu_z, i + 1, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i + 1, j, k) * vector(rk1Bfu_y, i + 1, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i + 1, j, k) * rk1vfu_y + vector(rk1Bfu_y, i + 1, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i + 1, j, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i + 1, j, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i + 1, j, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i + 1, j, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i + 1, j, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i + 1, j, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i + 1, j, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i + 1, j, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i + 1, j, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i + 1, j, k) * vector(rk1Bfu_z, i + 1, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i + 1, j, k) * rk1vfu_z + vector(rk1Bfu_z, i + 1, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i + 1, j, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i + 1, j, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i + 1, j, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i + 1, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i + 1, j, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i + 1, j, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i + 1, j, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i + 1, j, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i + 1, j, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i + 1, j, k) * vector(rk1Bfu_y, i + 1, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i + 1, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i + 1, j, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i + 1, j, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i + 1, j, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i + 1, j, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i + 1, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i + 1, j, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i + 1, j, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i + 1, j, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i + 1, j, k));
						Fluxp1iSfd_x = FiSfd_x_mainI(vector(rk1Sfd_x, i + 1, j, k), vector(rk1gtd_xx, i + 1, j, k), vector(rk1gtd_xy, i + 1, j, k), vector(rk1gtd_xz, i + 1, j, k), vector(rk1Betau_x, i + 1, j, k), vector(rk1Alpha, i + 1, j, k), vector(TauT_xx, i + 1, j, k), vector(TauT_xy, i + 1, j, k), vector(TauT_xz, i + 1, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp1iSfd_y = FiSfd_y_mainI(vector(rk1Sfd_y, i + 1, j, k), vector(rk1gtd_xy, i + 1, j, k), vector(rk1gtd_yy, i + 1, j, k), vector(rk1gtd_yz, i + 1, j, k), vector(rk1Betau_x, i + 1, j, k), vector(rk1Alpha, i + 1, j, k), vector(TauT_xx, i + 1, j, k), vector(TauT_xy, i + 1, j, k), vector(TauT_xz, i + 1, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxp1iSfd_z = FiSfd_z_mainI(vector(rk1Sfd_z, i + 1, j, k), vector(rk1gtd_xz, i + 1, j, k), vector(rk1gtd_yz, i + 1, j, k), vector(rk1gtd_zz, i + 1, j, k), vector(rk1Betau_x, i + 1, j, k), vector(rk1Alpha, i + 1, j, k), vector(TauT_xx, i + 1, j, k), vector(TauT_xy, i + 1, j, k), vector(TauT_xz, i + 1, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxp1iBfu_x = FiBfu_x_mainI(vector(rk1phif, i + 1, j, k), vector(rk1Alpha, i + 1, j, k), rk1chi_max, rk1gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp1iBfu_y = FiBfu_y_mainI(vector(rk1Bfu_x, i + 1, j, k), vector(rk1Bfu_y, i + 1, j, k), vector(rk1phif, i + 1, j, k), vector(rk1Betau_x, i + 1, j, k), vector(rk1Betau_y, i + 1, j, k), vector(rk1Alpha, i + 1, j, k), vector(TauM_xy, i + 1, j, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1iBfu_z = FiBfu_z_mainI(vector(rk1Bfu_x, i + 1, j, k), vector(rk1Bfu_z, i + 1, j, k), vector(rk1phif, i + 1, j, k), vector(rk1Betau_x, i + 1, j, k), vector(rk1Betau_z, i + 1, j, k), vector(rk1Alpha, i + 1, j, k), vector(TauM_xz, i + 1, j, k), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1iDf = FiDf_mainI(vector(rk1Df, i + 1, j, k), vector(rk1Betau_x, i + 1, j, k), vector(rk1Alpha, i + 1, j, k), vector(TauN_x, i + 1, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp1iDYf = FiDYf_mainI(vector(rk1DYf, i + 1, j, k), vector(rk1Betau_x, i + 1, j, k), vector(rk1Alpha, i + 1, j, k), vector(TauNe_x, i + 1, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp1itauf = Fitauf_mainI(vector(rk1Df, i + 1, j, k), vector(rk1tauf, i + 1, j, k), vector(rk1Betau_x, i + 1, j, k), vector(rk1Alpha, i + 1, j, k), vector(TauN_x, i + 1, j, k), rk1sdetg, rk1Sfu_x, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp1iphif = Fiphif_mainI(vector(rk1Bfu_x, i + 1, j, k), vector(rk1phif, i + 1, j, k), vector(rk1Betau_x, i + 1, j, k), vector(rk1Alpha, i + 1, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i + 1, j, k) * (1.0 + vector(epsf, i + 1, j, k)) + vector(pf, i + 1, j, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i + 1, j, k) + rk1sqca - vector(sqcs, i + 1, j, k) * rk1sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk1Betau_x, i + 1, j, k)) + vector(rk1Alpha, i + 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk1Betau_x, i + 1, j, k)) + vector(rk1Alpha, i + 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk1Betau_x, i + 1, j, k)) + vector(rk1Alpha, i + 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk1Betau_x, i + 1, j, k)) + vector(rk1Alpha, i + 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk1Betau_x, i + 1, j, k)) + vector(rk1Alpha, i + 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk1Betau_x, i + 1, j, k)) + vector(rk1Alpha, i + 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk1Betau_x, i + 1, j, k)) + vector(rk1Alpha, i + 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk1Betau_x, i + 1, j, k)) + vector(rk1Alpha, i + 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk1Betau_x, i + 1, j, k)) + vector(rk1Alpha, i + 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk1Betau_x, i + 1, j, k)) + vector(rk1Alpha, i + 1, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i + 2, j, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i + 2, j, k) * vector(rk1Bfu_x, i + 2, j, k) + vector(rk1gtd_yy, i + 2, j, k) * vector(rk1Bfu_y, i + 2, j, k) + vector(rk1gtd_yz, i + 2, j, k) * vector(rk1Bfu_z, i + 2, j, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i + 2, j, k) * vector(rk1Bfu_x, i + 2, j, k) + vector(rk1gtd_yz, i + 2, j, k) * vector(rk1Bfu_y, i + 2, j, k) + vector(rk1gtd_zz, i + 2, j, k) * vector(rk1Bfu_z, i + 2, j, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i + 2, j, k) * vector(rk1Bfu_x, i + 2, j, k) + vector(rk1gtd_xy, i + 2, j, k) * vector(rk1Bfu_y, i + 2, j, k) + vector(rk1gtd_xz, i + 2, j, k) * vector(rk1Bfu_z, i + 2, j, k));
						rk1sqBf = vector(rk1Bfu_x, i + 2, j, k) * rk1Bfd_x + vector(rk1Bfu_y, i + 2, j, k) * rk1Bfd_y + vector(rk1Bfu_z, i + 2, j, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i + 2, j, k) * vector(rk1gtd_zz, i + 2, j, k) * vector(rk1gtd_xx, i + 2, j, k) + (-vector(rk1gtd_yy, i + 2, j, k) * (vector(rk1gtd_xz, i + 2, j, k) * vector(rk1gtd_xz, i + 2, j, k))) + (-vector(rk1gtd_zz, i + 2, j, k) * (vector(rk1gtd_xy, i + 2, j, k) * vector(rk1gtd_xy, i + 2, j, k))) + (-(vector(rk1gtd_yz, i + 2, j, k) * vector(rk1gtd_yz, i + 2, j, k)) * vector(rk1gtd_xx, i + 2, j, k)) + 2.0 * vector(rk1gtd_yz, i + 2, j, k) * vector(rk1gtd_xy, i + 2, j, k) * vector(rk1gtd_xz, i + 2, j, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i + 2, j, k) * vector(rk1gtd_yy, i + 2, j, k) - vector(rk1gtd_xy, i + 2, j, k) * vector(rk1gtd_xy, i + 2, j, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i + 2, j, k) * vector(rk1gtd_zz, i + 2, j, k) - vector(rk1gtd_yz, i + 2, j, k) * vector(rk1gtd_yz, i + 2, j, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i + 2, j, k) * vector(rk1gtd_yz, i + 2, j, k) - vector(rk1gtd_yy, i + 2, j, k) * vector(rk1gtd_xz, i + 2, j, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i + 2, j, k) * vector(rk1gtd_zz, i + 2, j, k)) + vector(rk1gtd_yz, i + 2, j, k) * vector(rk1gtd_xz, i + 2, j, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i + 2, j, k) + rk1gtu_xy * vector(rk1Sfd_y, i + 2, j, k) + rk1gtu_xz * vector(rk1Sfd_z, i + 2, j, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i + 2, j, k) + rk1gtu_xy * vector(vfd_y, i + 2, j, k) + rk1gtu_xz * vector(vfd_z, i + 2, j, k));
						rk1Bfvf = vector(rk1Bfu_x, i + 2, j, k) * vector(vfd_x, i + 2, j, k) + vector(rk1Bfu_y, i + 2, j, k) * vector(vfd_y, i + 2, j, k) + vector(rk1Bfu_z, i + 2, j, k) * vector(vfd_z, i + 2, j, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i + 2, j, k) * vector(rk1gtd_zz, i + 2, j, k) - vector(rk1gtd_xz, i + 2, j, k) * vector(rk1gtd_xz, i + 2, j, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i + 2, j, k) * vector(rk1gtd_yz, i + 2, j, k)) + vector(rk1gtd_xy, i + 2, j, k) * vector(rk1gtd_xz, i + 2, j, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i + 2, j, k) + rk1gtu_yz * vector(vfd_y, i + 2, j, k) + rk1gtu_zz * vector(vfd_z, i + 2, j, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i + 2, j, k) + rk1gtu_yy * vector(rk1Sfd_y, i + 2, j, k) + rk1gtu_yz * vector(rk1Sfd_z, i + 2, j, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i + 2, j, k) + rk1gtu_yz * vector(rk1Sfd_y, i + 2, j, k) + rk1gtu_zz * vector(rk1Sfd_z, i + 2, j, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i + 2, j, k) + rk1gtu_yy * vector(vfd_y, i + 2, j, k) + rk1gtu_yz * vector(vfd_z, i + 2, j, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i + 2, j, k) + rk1vfu_y * vector(vfd_y, i + 2, j, k) + rk1vfu_z * vector(vfd_z, i + 2, j, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i + 2, j, k) * vector(rk1Bfu_z, i + 2, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i + 2, j, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i + 2, j, k) * vector(rk1Bfu_x, i + 2, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i + 2, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i + 2, j, k) * vector(rk1Bfu_z, i + 2, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i + 2, j, k) * rk1vfu_z + vector(rk1Bfu_z, i + 2, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i + 2, j, k) * vector(rk1Bfu_y, i + 2, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i + 2, j, k) * rk1vfu_y + vector(rk1Bfu_y, i + 2, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i + 2, j, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i + 2, j, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i + 2, j, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i + 2, j, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i + 2, j, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i + 2, j, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i + 2, j, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i + 2, j, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i + 2, j, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i + 2, j, k) * vector(rk1Bfu_z, i + 2, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i + 2, j, k) * rk1vfu_z + vector(rk1Bfu_z, i + 2, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i + 2, j, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i + 2, j, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i + 2, j, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i + 2, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i + 2, j, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i + 2, j, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i + 2, j, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i + 2, j, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i + 2, j, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i + 2, j, k) * vector(rk1Bfu_y, i + 2, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i + 2, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i + 2, j, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i + 2, j, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i + 2, j, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i + 2, j, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i + 2, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i + 2, j, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i + 2, j, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i + 2, j, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i + 2, j, k));
						Fluxp2iSfd_x = FiSfd_x_mainI(vector(rk1Sfd_x, i + 2, j, k), vector(rk1gtd_xx, i + 2, j, k), vector(rk1gtd_xy, i + 2, j, k), vector(rk1gtd_xz, i + 2, j, k), vector(rk1Betau_x, i + 2, j, k), vector(rk1Alpha, i + 2, j, k), vector(TauT_xx, i + 2, j, k), vector(TauT_xy, i + 2, j, k), vector(TauT_xz, i + 2, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp2iSfd_y = FiSfd_y_mainI(vector(rk1Sfd_y, i + 2, j, k), vector(rk1gtd_xy, i + 2, j, k), vector(rk1gtd_yy, i + 2, j, k), vector(rk1gtd_yz, i + 2, j, k), vector(rk1Betau_x, i + 2, j, k), vector(rk1Alpha, i + 2, j, k), vector(TauT_xx, i + 2, j, k), vector(TauT_xy, i + 2, j, k), vector(TauT_xz, i + 2, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxp2iSfd_z = FiSfd_z_mainI(vector(rk1Sfd_z, i + 2, j, k), vector(rk1gtd_xz, i + 2, j, k), vector(rk1gtd_yz, i + 2, j, k), vector(rk1gtd_zz, i + 2, j, k), vector(rk1Betau_x, i + 2, j, k), vector(rk1Alpha, i + 2, j, k), vector(TauT_xx, i + 2, j, k), vector(TauT_xy, i + 2, j, k), vector(TauT_xz, i + 2, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxp2iBfu_x = FiBfu_x_mainI(vector(rk1phif, i + 2, j, k), vector(rk1Alpha, i + 2, j, k), rk1chi_max, rk1gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp2iBfu_y = FiBfu_y_mainI(vector(rk1Bfu_x, i + 2, j, k), vector(rk1Bfu_y, i + 2, j, k), vector(rk1phif, i + 2, j, k), vector(rk1Betau_x, i + 2, j, k), vector(rk1Betau_y, i + 2, j, k), vector(rk1Alpha, i + 2, j, k), vector(TauM_xy, i + 2, j, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2iBfu_z = FiBfu_z_mainI(vector(rk1Bfu_x, i + 2, j, k), vector(rk1Bfu_z, i + 2, j, k), vector(rk1phif, i + 2, j, k), vector(rk1Betau_x, i + 2, j, k), vector(rk1Betau_z, i + 2, j, k), vector(rk1Alpha, i + 2, j, k), vector(TauM_xz, i + 2, j, k), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2iDf = FiDf_mainI(vector(rk1Df, i + 2, j, k), vector(rk1Betau_x, i + 2, j, k), vector(rk1Alpha, i + 2, j, k), vector(TauN_x, i + 2, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp2iDYf = FiDYf_mainI(vector(rk1DYf, i + 2, j, k), vector(rk1Betau_x, i + 2, j, k), vector(rk1Alpha, i + 2, j, k), vector(TauNe_x, i + 2, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp2itauf = Fitauf_mainI(vector(rk1Df, i + 2, j, k), vector(rk1tauf, i + 2, j, k), vector(rk1Betau_x, i + 2, j, k), vector(rk1Alpha, i + 2, j, k), vector(TauN_x, i + 2, j, k), rk1sdetg, rk1Sfu_x, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp2iphif = Fiphif_mainI(vector(rk1Bfu_x, i + 2, j, k), vector(rk1phif, i + 2, j, k), vector(rk1Betau_x, i + 2, j, k), vector(rk1Alpha, i + 2, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i + 2, j, k) * (1.0 + vector(epsf, i + 2, j, k)) + vector(pf, i + 2, j, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i + 2, j, k) + rk1sqca - vector(sqcs, i + 2, j, k) * rk1sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk1Betau_x, i + 2, j, k)) + vector(rk1Alpha, i + 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk1Betau_x, i + 2, j, k)) + vector(rk1Alpha, i + 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk1Betau_x, i + 2, j, k)) + vector(rk1Alpha, i + 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk1Betau_x, i + 2, j, k)) + vector(rk1Alpha, i + 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk1Betau_x, i + 2, j, k)) + vector(rk1Alpha, i + 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk1Betau_x, i + 2, j, k)) + vector(rk1Alpha, i + 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk1Betau_x, i + 2, j, k)) + vector(rk1Alpha, i + 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk1Betau_x, i + 2, j, k)) + vector(rk1Alpha, i + 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk1Betau_x, i + 2, j, k)) + vector(rk1Alpha, i + 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk1Betau_x, i + 2, j, k)) + vector(rk1Alpha, i + 2, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i + 3, j, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i + 3, j, k) * vector(rk1Bfu_x, i + 3, j, k) + vector(rk1gtd_yy, i + 3, j, k) * vector(rk1Bfu_y, i + 3, j, k) + vector(rk1gtd_yz, i + 3, j, k) * vector(rk1Bfu_z, i + 3, j, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i + 3, j, k) * vector(rk1Bfu_x, i + 3, j, k) + vector(rk1gtd_yz, i + 3, j, k) * vector(rk1Bfu_y, i + 3, j, k) + vector(rk1gtd_zz, i + 3, j, k) * vector(rk1Bfu_z, i + 3, j, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i + 3, j, k) * vector(rk1Bfu_x, i + 3, j, k) + vector(rk1gtd_xy, i + 3, j, k) * vector(rk1Bfu_y, i + 3, j, k) + vector(rk1gtd_xz, i + 3, j, k) * vector(rk1Bfu_z, i + 3, j, k));
						rk1sqBf = vector(rk1Bfu_x, i + 3, j, k) * rk1Bfd_x + vector(rk1Bfu_y, i + 3, j, k) * rk1Bfd_y + vector(rk1Bfu_z, i + 3, j, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i + 3, j, k) * vector(rk1gtd_zz, i + 3, j, k) * vector(rk1gtd_xx, i + 3, j, k) + (-vector(rk1gtd_yy, i + 3, j, k) * (vector(rk1gtd_xz, i + 3, j, k) * vector(rk1gtd_xz, i + 3, j, k))) + (-vector(rk1gtd_zz, i + 3, j, k) * (vector(rk1gtd_xy, i + 3, j, k) * vector(rk1gtd_xy, i + 3, j, k))) + (-(vector(rk1gtd_yz, i + 3, j, k) * vector(rk1gtd_yz, i + 3, j, k)) * vector(rk1gtd_xx, i + 3, j, k)) + 2.0 * vector(rk1gtd_yz, i + 3, j, k) * vector(rk1gtd_xy, i + 3, j, k) * vector(rk1gtd_xz, i + 3, j, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i + 3, j, k) * vector(rk1gtd_yy, i + 3, j, k) - vector(rk1gtd_xy, i + 3, j, k) * vector(rk1gtd_xy, i + 3, j, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i + 3, j, k) * vector(rk1gtd_zz, i + 3, j, k) - vector(rk1gtd_yz, i + 3, j, k) * vector(rk1gtd_yz, i + 3, j, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i + 3, j, k) * vector(rk1gtd_yz, i + 3, j, k) - vector(rk1gtd_yy, i + 3, j, k) * vector(rk1gtd_xz, i + 3, j, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i + 3, j, k) * vector(rk1gtd_zz, i + 3, j, k)) + vector(rk1gtd_yz, i + 3, j, k) * vector(rk1gtd_xz, i + 3, j, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i + 3, j, k) + rk1gtu_xy * vector(rk1Sfd_y, i + 3, j, k) + rk1gtu_xz * vector(rk1Sfd_z, i + 3, j, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i + 3, j, k) + rk1gtu_xy * vector(vfd_y, i + 3, j, k) + rk1gtu_xz * vector(vfd_z, i + 3, j, k));
						rk1Bfvf = vector(rk1Bfu_x, i + 3, j, k) * vector(vfd_x, i + 3, j, k) + vector(rk1Bfu_y, i + 3, j, k) * vector(vfd_y, i + 3, j, k) + vector(rk1Bfu_z, i + 3, j, k) * vector(vfd_z, i + 3, j, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i + 3, j, k) * vector(rk1gtd_zz, i + 3, j, k) - vector(rk1gtd_xz, i + 3, j, k) * vector(rk1gtd_xz, i + 3, j, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i + 3, j, k) * vector(rk1gtd_yz, i + 3, j, k)) + vector(rk1gtd_xy, i + 3, j, k) * vector(rk1gtd_xz, i + 3, j, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i + 3, j, k) + rk1gtu_yz * vector(vfd_y, i + 3, j, k) + rk1gtu_zz * vector(vfd_z, i + 3, j, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i + 3, j, k) + rk1gtu_yy * vector(rk1Sfd_y, i + 3, j, k) + rk1gtu_yz * vector(rk1Sfd_z, i + 3, j, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i + 3, j, k) + rk1gtu_yz * vector(rk1Sfd_y, i + 3, j, k) + rk1gtu_zz * vector(rk1Sfd_z, i + 3, j, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i + 3, j, k) + rk1gtu_yy * vector(vfd_y, i + 3, j, k) + rk1gtu_yz * vector(vfd_z, i + 3, j, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i + 3, j, k) + rk1vfu_y * vector(vfd_y, i + 3, j, k) + rk1vfu_z * vector(vfd_z, i + 3, j, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i + 3, j, k) * vector(rk1Bfu_z, i + 3, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i + 3, j, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i + 3, j, k) * vector(rk1Bfu_x, i + 3, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i + 3, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i + 3, j, k) * vector(rk1Bfu_z, i + 3, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i + 3, j, k) * rk1vfu_z + vector(rk1Bfu_z, i + 3, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i + 3, j, k) * vector(rk1Bfu_y, i + 3, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i + 3, j, k) * rk1vfu_y + vector(rk1Bfu_y, i + 3, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i + 3, j, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i + 3, j, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i + 3, j, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i + 3, j, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i + 3, j, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i + 3, j, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i + 3, j, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i + 3, j, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i + 3, j, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i + 3, j, k) * vector(rk1Bfu_z, i + 3, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i + 3, j, k) * rk1vfu_z + vector(rk1Bfu_z, i + 3, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i + 3, j, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i + 3, j, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i + 3, j, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i + 3, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i + 3, j, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i + 3, j, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i + 3, j, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i + 3, j, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i + 3, j, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i + 3, j, k) * vector(rk1Bfu_y, i + 3, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i + 3, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i + 3, j, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i + 3, j, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i + 3, j, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i + 3, j, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i + 3, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i + 3, j, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i + 3, j, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i + 3, j, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i + 3, j, k));
						Fluxp3iSfd_x = FiSfd_x_mainI(vector(rk1Sfd_x, i + 3, j, k), vector(rk1gtd_xx, i + 3, j, k), vector(rk1gtd_xy, i + 3, j, k), vector(rk1gtd_xz, i + 3, j, k), vector(rk1Betau_x, i + 3, j, k), vector(rk1Alpha, i + 3, j, k), vector(TauT_xx, i + 3, j, k), vector(TauT_xy, i + 3, j, k), vector(TauT_xz, i + 3, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp3iSfd_y = FiSfd_y_mainI(vector(rk1Sfd_y, i + 3, j, k), vector(rk1gtd_xy, i + 3, j, k), vector(rk1gtd_yy, i + 3, j, k), vector(rk1gtd_yz, i + 3, j, k), vector(rk1Betau_x, i + 3, j, k), vector(rk1Alpha, i + 3, j, k), vector(TauT_xx, i + 3, j, k), vector(TauT_xy, i + 3, j, k), vector(TauT_xz, i + 3, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxp3iSfd_z = FiSfd_z_mainI(vector(rk1Sfd_z, i + 3, j, k), vector(rk1gtd_xz, i + 3, j, k), vector(rk1gtd_yz, i + 3, j, k), vector(rk1gtd_zz, i + 3, j, k), vector(rk1Betau_x, i + 3, j, k), vector(rk1Alpha, i + 3, j, k), vector(TauT_xx, i + 3, j, k), vector(TauT_xy, i + 3, j, k), vector(TauT_xz, i + 3, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxp3iBfu_x = FiBfu_x_mainI(vector(rk1phif, i + 3, j, k), vector(rk1Alpha, i + 3, j, k), rk1chi_max, rk1gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp3iBfu_y = FiBfu_y_mainI(vector(rk1Bfu_x, i + 3, j, k), vector(rk1Bfu_y, i + 3, j, k), vector(rk1phif, i + 3, j, k), vector(rk1Betau_x, i + 3, j, k), vector(rk1Betau_y, i + 3, j, k), vector(rk1Alpha, i + 3, j, k), vector(TauM_xy, i + 3, j, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3iBfu_z = FiBfu_z_mainI(vector(rk1Bfu_x, i + 3, j, k), vector(rk1Bfu_z, i + 3, j, k), vector(rk1phif, i + 3, j, k), vector(rk1Betau_x, i + 3, j, k), vector(rk1Betau_z, i + 3, j, k), vector(rk1Alpha, i + 3, j, k), vector(TauM_xz, i + 3, j, k), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3iDf = FiDf_mainI(vector(rk1Df, i + 3, j, k), vector(rk1Betau_x, i + 3, j, k), vector(rk1Alpha, i + 3, j, k), vector(TauN_x, i + 3, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp3iDYf = FiDYf_mainI(vector(rk1DYf, i + 3, j, k), vector(rk1Betau_x, i + 3, j, k), vector(rk1Alpha, i + 3, j, k), vector(TauNe_x, i + 3, j, k), rk1sdetg, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp3itauf = Fitauf_mainI(vector(rk1Df, i + 3, j, k), vector(rk1tauf, i + 3, j, k), vector(rk1Betau_x, i + 3, j, k), vector(rk1Alpha, i + 3, j, k), vector(TauN_x, i + 3, j, k), rk1sdetg, rk1Sfu_x, rk1vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp3iphif = Fiphif_mainI(vector(rk1Bfu_x, i + 3, j, k), vector(rk1phif, i + 3, j, k), vector(rk1Betau_x, i + 3, j, k), vector(rk1Alpha, i + 3, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i + 3, j, k) * (1.0 + vector(epsf, i + 3, j, k)) + vector(pf, i + 3, j, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i + 3, j, k) + rk1sqca - vector(sqcs, i + 3, j, k) * rk1sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk1Betau_x, i + 3, j, k)) + vector(rk1Alpha, i + 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk1Betau_x, i + 3, j, k)) + vector(rk1Alpha, i + 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk1Betau_x, i + 3, j, k)) + vector(rk1Alpha, i + 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk1Betau_x, i + 3, j, k)) + vector(rk1Alpha, i + 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk1Betau_x, i + 3, j, k)) + vector(rk1Alpha, i + 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk1Betau_x, i + 3, j, k)) + vector(rk1Alpha, i + 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk1Betau_x, i + 3, j, k)) + vector(rk1Alpha, i + 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk1Betau_x, i + 3, j, k)) + vector(rk1Alpha, i + 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk1Betau_x, i + 3, j, k)) + vector(rk1Alpha, i + 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk1Betau_x, i + 3, j, k)) + vector(rk1Alpha, i + 3, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_x) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_xx - (1.0 - rk1sqa) * (rk1vfu_x * rk1vfu_x)))))));
						LLFpiSfd_x = 0.5 * (FluxiSfd_x + fabs(SpeediSfd_x) * vector(rk1Sfd_x, i, j, k));
						LLFpiSfd_y = 0.5 * (FluxiSfd_y + fabs(SpeediSfd_y) * vector(rk1Sfd_y, i, j, k));
						LLFpiSfd_z = 0.5 * (FluxiSfd_z + fabs(SpeediSfd_z) * vector(rk1Sfd_z, i, j, k));
						LLFpiBfu_x = 0.5 * (FluxiBfu_x + fabs(SpeediBfu_x) * vector(rk1Bfu_x, i, j, k));
						LLFpiBfu_y = 0.5 * (FluxiBfu_y + fabs(SpeediBfu_y) * vector(rk1Bfu_y, i, j, k));
						LLFpiBfu_z = 0.5 * (FluxiBfu_z + fabs(SpeediBfu_z) * vector(rk1Bfu_z, i, j, k));
						LLFpiDf = 0.5 * (FluxiDf + fabs(SpeediDf) * vector(rk1Df, i, j, k));
						LLFpiDYf = 0.5 * (FluxiDYf + fabs(SpeediDYf) * vector(rk1DYf, i, j, k));
						LLFpitauf = 0.5 * (Fluxitauf + fabs(Speeditauf) * vector(rk1tauf, i, j, k));
						LLFpiphif = 0.5 * (Fluxiphif + fabs(Speediphif) * vector(rk1phif, i, j, k));
						LLFniSfd_x = 0.5 * (FluxiSfd_x - fabs(SpeediSfd_x) * vector(rk1Sfd_x, i, j, k));
						LLFniSfd_y = 0.5 * (FluxiSfd_y - fabs(SpeediSfd_y) * vector(rk1Sfd_y, i, j, k));
						LLFniSfd_z = 0.5 * (FluxiSfd_z - fabs(SpeediSfd_z) * vector(rk1Sfd_z, i, j, k));
						LLFniBfu_x = 0.5 * (FluxiBfu_x - fabs(SpeediBfu_x) * vector(rk1Bfu_x, i, j, k));
						LLFniBfu_y = 0.5 * (FluxiBfu_y - fabs(SpeediBfu_y) * vector(rk1Bfu_y, i, j, k));
						LLFniBfu_z = 0.5 * (FluxiBfu_z - fabs(SpeediBfu_z) * vector(rk1Bfu_z, i, j, k));
						LLFniDf = 0.5 * (FluxiDf - fabs(SpeediDf) * vector(rk1Df, i, j, k));
						LLFniDYf = 0.5 * (FluxiDYf - fabs(SpeediDYf) * vector(rk1DYf, i, j, k));
						LLFnitauf = 0.5 * (Fluxitauf - fabs(Speeditauf) * vector(rk1tauf, i, j, k));
						LLFniphif = 0.5 * (Fluxiphif - fabs(Speediphif) * vector(rk1phif, i, j, k));
						LLFp_m3iSfd_x = 0.5 * (Fluxm3iSfd_x + fabs(SpeediSfd_x) * vector(rk1Sfd_x, i - 3, j, k));
						LLFp_m3iSfd_y = 0.5 * (Fluxm3iSfd_y + fabs(SpeediSfd_y) * vector(rk1Sfd_y, i - 3, j, k));
						LLFp_m3iSfd_z = 0.5 * (Fluxm3iSfd_z + fabs(SpeediSfd_z) * vector(rk1Sfd_z, i - 3, j, k));
						LLFp_m3iBfu_x = 0.5 * (Fluxm3iBfu_x + fabs(SpeediBfu_x) * vector(rk1Bfu_x, i - 3, j, k));
						LLFp_m3iBfu_y = 0.5 * (Fluxm3iBfu_y + fabs(SpeediBfu_y) * vector(rk1Bfu_y, i - 3, j, k));
						LLFp_m3iBfu_z = 0.5 * (Fluxm3iBfu_z + fabs(SpeediBfu_z) * vector(rk1Bfu_z, i - 3, j, k));
						LLFp_m3iDf = 0.5 * (Fluxm3iDf + fabs(SpeediDf) * vector(rk1Df, i - 3, j, k));
						LLFp_m3iDYf = 0.5 * (Fluxm3iDYf + fabs(SpeediDYf) * vector(rk1DYf, i - 3, j, k));
						LLFp_m3itauf = 0.5 * (Fluxm3itauf + fabs(Speeditauf) * vector(rk1tauf, i - 3, j, k));
						LLFp_m3iphif = 0.5 * (Fluxm3iphif + fabs(Speediphif) * vector(rk1phif, i - 3, j, k));
						LLFn_m3iSfd_x = 0.5 * (Fluxm3iSfd_x - fabs(SpeediSfd_x) * vector(rk1Sfd_x, i - 3, j, k));
						LLFn_m3iSfd_y = 0.5 * (Fluxm3iSfd_y - fabs(SpeediSfd_y) * vector(rk1Sfd_y, i - 3, j, k));
						LLFn_m3iSfd_z = 0.5 * (Fluxm3iSfd_z - fabs(SpeediSfd_z) * vector(rk1Sfd_z, i - 3, j, k));
						LLFn_m3iBfu_x = 0.5 * (Fluxm3iBfu_x - fabs(SpeediBfu_x) * vector(rk1Bfu_x, i - 3, j, k));
						LLFn_m3iBfu_y = 0.5 * (Fluxm3iBfu_y - fabs(SpeediBfu_y) * vector(rk1Bfu_y, i - 3, j, k));
						LLFn_m3iBfu_z = 0.5 * (Fluxm3iBfu_z - fabs(SpeediBfu_z) * vector(rk1Bfu_z, i - 3, j, k));
						LLFn_m3iDf = 0.5 * (Fluxm3iDf - fabs(SpeediDf) * vector(rk1Df, i - 3, j, k));
						LLFn_m3iDYf = 0.5 * (Fluxm3iDYf - fabs(SpeediDYf) * vector(rk1DYf, i - 3, j, k));
						LLFn_m3itauf = 0.5 * (Fluxm3itauf - fabs(Speeditauf) * vector(rk1tauf, i - 3, j, k));
						LLFn_m3iphif = 0.5 * (Fluxm3iphif - fabs(Speediphif) * vector(rk1phif, i - 3, j, k));
						LLFp_m2iSfd_x = 0.5 * (Fluxm2iSfd_x + fabs(SpeediSfd_x) * vector(rk1Sfd_x, i - 2, j, k));
						LLFp_m2iSfd_y = 0.5 * (Fluxm2iSfd_y + fabs(SpeediSfd_y) * vector(rk1Sfd_y, i - 2, j, k));
						LLFp_m2iSfd_z = 0.5 * (Fluxm2iSfd_z + fabs(SpeediSfd_z) * vector(rk1Sfd_z, i - 2, j, k));
						LLFp_m2iBfu_x = 0.5 * (Fluxm2iBfu_x + fabs(SpeediBfu_x) * vector(rk1Bfu_x, i - 2, j, k));
						LLFp_m2iBfu_y = 0.5 * (Fluxm2iBfu_y + fabs(SpeediBfu_y) * vector(rk1Bfu_y, i - 2, j, k));
						LLFp_m2iBfu_z = 0.5 * (Fluxm2iBfu_z + fabs(SpeediBfu_z) * vector(rk1Bfu_z, i - 2, j, k));
						LLFp_m2iDf = 0.5 * (Fluxm2iDf + fabs(SpeediDf) * vector(rk1Df, i - 2, j, k));
						LLFp_m2iDYf = 0.5 * (Fluxm2iDYf + fabs(SpeediDYf) * vector(rk1DYf, i - 2, j, k));
						LLFp_m2itauf = 0.5 * (Fluxm2itauf + fabs(Speeditauf) * vector(rk1tauf, i - 2, j, k));
						LLFp_m2iphif = 0.5 * (Fluxm2iphif + fabs(Speediphif) * vector(rk1phif, i - 2, j, k));
						LLFn_m2iSfd_x = 0.5 * (Fluxm2iSfd_x - fabs(SpeediSfd_x) * vector(rk1Sfd_x, i - 2, j, k));
						LLFn_m2iSfd_y = 0.5 * (Fluxm2iSfd_y - fabs(SpeediSfd_y) * vector(rk1Sfd_y, i - 2, j, k));
						LLFn_m2iSfd_z = 0.5 * (Fluxm2iSfd_z - fabs(SpeediSfd_z) * vector(rk1Sfd_z, i - 2, j, k));
						LLFn_m2iBfu_x = 0.5 * (Fluxm2iBfu_x - fabs(SpeediBfu_x) * vector(rk1Bfu_x, i - 2, j, k));
						LLFn_m2iBfu_y = 0.5 * (Fluxm2iBfu_y - fabs(SpeediBfu_y) * vector(rk1Bfu_y, i - 2, j, k));
						LLFn_m2iBfu_z = 0.5 * (Fluxm2iBfu_z - fabs(SpeediBfu_z) * vector(rk1Bfu_z, i - 2, j, k));
						LLFn_m2iDf = 0.5 * (Fluxm2iDf - fabs(SpeediDf) * vector(rk1Df, i - 2, j, k));
						LLFn_m2iDYf = 0.5 * (Fluxm2iDYf - fabs(SpeediDYf) * vector(rk1DYf, i - 2, j, k));
						LLFn_m2itauf = 0.5 * (Fluxm2itauf - fabs(Speeditauf) * vector(rk1tauf, i - 2, j, k));
						LLFn_m2iphif = 0.5 * (Fluxm2iphif - fabs(Speediphif) * vector(rk1phif, i - 2, j, k));
						LLFp_m1iSfd_x = 0.5 * (Fluxm1iSfd_x + fabs(SpeediSfd_x) * vector(rk1Sfd_x, i - 1, j, k));
						LLFp_m1iSfd_y = 0.5 * (Fluxm1iSfd_y + fabs(SpeediSfd_y) * vector(rk1Sfd_y, i - 1, j, k));
						LLFp_m1iSfd_z = 0.5 * (Fluxm1iSfd_z + fabs(SpeediSfd_z) * vector(rk1Sfd_z, i - 1, j, k));
						LLFp_m1iBfu_x = 0.5 * (Fluxm1iBfu_x + fabs(SpeediBfu_x) * vector(rk1Bfu_x, i - 1, j, k));
						LLFp_m1iBfu_y = 0.5 * (Fluxm1iBfu_y + fabs(SpeediBfu_y) * vector(rk1Bfu_y, i - 1, j, k));
						LLFp_m1iBfu_z = 0.5 * (Fluxm1iBfu_z + fabs(SpeediBfu_z) * vector(rk1Bfu_z, i - 1, j, k));
						LLFp_m1iDf = 0.5 * (Fluxm1iDf + fabs(SpeediDf) * vector(rk1Df, i - 1, j, k));
						LLFp_m1iDYf = 0.5 * (Fluxm1iDYf + fabs(SpeediDYf) * vector(rk1DYf, i - 1, j, k));
						LLFp_m1itauf = 0.5 * (Fluxm1itauf + fabs(Speeditauf) * vector(rk1tauf, i - 1, j, k));
						LLFp_m1iphif = 0.5 * (Fluxm1iphif + fabs(Speediphif) * vector(rk1phif, i - 1, j, k));
						LLFn_m1iSfd_x = 0.5 * (Fluxm1iSfd_x - fabs(SpeediSfd_x) * vector(rk1Sfd_x, i - 1, j, k));
						LLFn_m1iSfd_y = 0.5 * (Fluxm1iSfd_y - fabs(SpeediSfd_y) * vector(rk1Sfd_y, i - 1, j, k));
						LLFn_m1iSfd_z = 0.5 * (Fluxm1iSfd_z - fabs(SpeediSfd_z) * vector(rk1Sfd_z, i - 1, j, k));
						LLFn_m1iBfu_x = 0.5 * (Fluxm1iBfu_x - fabs(SpeediBfu_x) * vector(rk1Bfu_x, i - 1, j, k));
						LLFn_m1iBfu_y = 0.5 * (Fluxm1iBfu_y - fabs(SpeediBfu_y) * vector(rk1Bfu_y, i - 1, j, k));
						LLFn_m1iBfu_z = 0.5 * (Fluxm1iBfu_z - fabs(SpeediBfu_z) * vector(rk1Bfu_z, i - 1, j, k));
						LLFn_m1iDf = 0.5 * (Fluxm1iDf - fabs(SpeediDf) * vector(rk1Df, i - 1, j, k));
						LLFn_m1iDYf = 0.5 * (Fluxm1iDYf - fabs(SpeediDYf) * vector(rk1DYf, i - 1, j, k));
						LLFn_m1itauf = 0.5 * (Fluxm1itauf - fabs(Speeditauf) * vector(rk1tauf, i - 1, j, k));
						LLFn_m1iphif = 0.5 * (Fluxm1iphif - fabs(Speediphif) * vector(rk1phif, i - 1, j, k));
						LLFp_p1iSfd_x = 0.5 * (Fluxp1iSfd_x + fabs(SpeediSfd_x) * vector(rk1Sfd_x, i + 1, j, k));
						LLFp_p1iSfd_y = 0.5 * (Fluxp1iSfd_y + fabs(SpeediSfd_y) * vector(rk1Sfd_y, i + 1, j, k));
						LLFp_p1iSfd_z = 0.5 * (Fluxp1iSfd_z + fabs(SpeediSfd_z) * vector(rk1Sfd_z, i + 1, j, k));
						LLFp_p1iBfu_x = 0.5 * (Fluxp1iBfu_x + fabs(SpeediBfu_x) * vector(rk1Bfu_x, i + 1, j, k));
						LLFp_p1iBfu_y = 0.5 * (Fluxp1iBfu_y + fabs(SpeediBfu_y) * vector(rk1Bfu_y, i + 1, j, k));
						LLFp_p1iBfu_z = 0.5 * (Fluxp1iBfu_z + fabs(SpeediBfu_z) * vector(rk1Bfu_z, i + 1, j, k));
						LLFp_p1iDf = 0.5 * (Fluxp1iDf + fabs(SpeediDf) * vector(rk1Df, i + 1, j, k));
						LLFp_p1iDYf = 0.5 * (Fluxp1iDYf + fabs(SpeediDYf) * vector(rk1DYf, i + 1, j, k));
						LLFp_p1itauf = 0.5 * (Fluxp1itauf + fabs(Speeditauf) * vector(rk1tauf, i + 1, j, k));
						LLFp_p1iphif = 0.5 * (Fluxp1iphif + fabs(Speediphif) * vector(rk1phif, i + 1, j, k));
						LLFn_p1iSfd_x = 0.5 * (Fluxp1iSfd_x - fabs(SpeediSfd_x) * vector(rk1Sfd_x, i + 1, j, k));
						LLFn_p1iSfd_y = 0.5 * (Fluxp1iSfd_y - fabs(SpeediSfd_y) * vector(rk1Sfd_y, i + 1, j, k));
						LLFn_p1iSfd_z = 0.5 * (Fluxp1iSfd_z - fabs(SpeediSfd_z) * vector(rk1Sfd_z, i + 1, j, k));
						LLFn_p1iBfu_x = 0.5 * (Fluxp1iBfu_x - fabs(SpeediBfu_x) * vector(rk1Bfu_x, i + 1, j, k));
						LLFn_p1iBfu_y = 0.5 * (Fluxp1iBfu_y - fabs(SpeediBfu_y) * vector(rk1Bfu_y, i + 1, j, k));
						LLFn_p1iBfu_z = 0.5 * (Fluxp1iBfu_z - fabs(SpeediBfu_z) * vector(rk1Bfu_z, i + 1, j, k));
						LLFn_p1iDf = 0.5 * (Fluxp1iDf - fabs(SpeediDf) * vector(rk1Df, i + 1, j, k));
						LLFn_p1iDYf = 0.5 * (Fluxp1iDYf - fabs(SpeediDYf) * vector(rk1DYf, i + 1, j, k));
						LLFn_p1itauf = 0.5 * (Fluxp1itauf - fabs(Speeditauf) * vector(rk1tauf, i + 1, j, k));
						LLFn_p1iphif = 0.5 * (Fluxp1iphif - fabs(Speediphif) * vector(rk1phif, i + 1, j, k));
						LLFp_p2iSfd_x = 0.5 * (Fluxp2iSfd_x + fabs(SpeediSfd_x) * vector(rk1Sfd_x, i + 2, j, k));
						LLFp_p2iSfd_y = 0.5 * (Fluxp2iSfd_y + fabs(SpeediSfd_y) * vector(rk1Sfd_y, i + 2, j, k));
						LLFp_p2iSfd_z = 0.5 * (Fluxp2iSfd_z + fabs(SpeediSfd_z) * vector(rk1Sfd_z, i + 2, j, k));
						LLFp_p2iBfu_x = 0.5 * (Fluxp2iBfu_x + fabs(SpeediBfu_x) * vector(rk1Bfu_x, i + 2, j, k));
						LLFp_p2iBfu_y = 0.5 * (Fluxp2iBfu_y + fabs(SpeediBfu_y) * vector(rk1Bfu_y, i + 2, j, k));
						LLFp_p2iBfu_z = 0.5 * (Fluxp2iBfu_z + fabs(SpeediBfu_z) * vector(rk1Bfu_z, i + 2, j, k));
						LLFp_p2iDf = 0.5 * (Fluxp2iDf + fabs(SpeediDf) * vector(rk1Df, i + 2, j, k));
						LLFp_p2iDYf = 0.5 * (Fluxp2iDYf + fabs(SpeediDYf) * vector(rk1DYf, i + 2, j, k));
						LLFp_p2itauf = 0.5 * (Fluxp2itauf + fabs(Speeditauf) * vector(rk1tauf, i + 2, j, k));
						LLFp_p2iphif = 0.5 * (Fluxp2iphif + fabs(Speediphif) * vector(rk1phif, i + 2, j, k));
						LLFn_p2iSfd_x = 0.5 * (Fluxp2iSfd_x - fabs(SpeediSfd_x) * vector(rk1Sfd_x, i + 2, j, k));
						LLFn_p2iSfd_y = 0.5 * (Fluxp2iSfd_y - fabs(SpeediSfd_y) * vector(rk1Sfd_y, i + 2, j, k));
						LLFn_p2iSfd_z = 0.5 * (Fluxp2iSfd_z - fabs(SpeediSfd_z) * vector(rk1Sfd_z, i + 2, j, k));
						LLFn_p2iBfu_x = 0.5 * (Fluxp2iBfu_x - fabs(SpeediBfu_x) * vector(rk1Bfu_x, i + 2, j, k));
						LLFn_p2iBfu_y = 0.5 * (Fluxp2iBfu_y - fabs(SpeediBfu_y) * vector(rk1Bfu_y, i + 2, j, k));
						LLFn_p2iBfu_z = 0.5 * (Fluxp2iBfu_z - fabs(SpeediBfu_z) * vector(rk1Bfu_z, i + 2, j, k));
						LLFn_p2iDf = 0.5 * (Fluxp2iDf - fabs(SpeediDf) * vector(rk1Df, i + 2, j, k));
						LLFn_p2iDYf = 0.5 * (Fluxp2iDYf - fabs(SpeediDYf) * vector(rk1DYf, i + 2, j, k));
						LLFn_p2itauf = 0.5 * (Fluxp2itauf - fabs(Speeditauf) * vector(rk1tauf, i + 2, j, k));
						LLFn_p2iphif = 0.5 * (Fluxp2iphif - fabs(Speediphif) * vector(rk1phif, i + 2, j, k));
						LLFp_p3iSfd_x = 0.5 * (Fluxp3iSfd_x + fabs(SpeediSfd_x) * vector(rk1Sfd_x, i + 3, j, k));
						LLFp_p3iSfd_y = 0.5 * (Fluxp3iSfd_y + fabs(SpeediSfd_y) * vector(rk1Sfd_y, i + 3, j, k));
						LLFp_p3iSfd_z = 0.5 * (Fluxp3iSfd_z + fabs(SpeediSfd_z) * vector(rk1Sfd_z, i + 3, j, k));
						LLFp_p3iBfu_x = 0.5 * (Fluxp3iBfu_x + fabs(SpeediBfu_x) * vector(rk1Bfu_x, i + 3, j, k));
						LLFp_p3iBfu_y = 0.5 * (Fluxp3iBfu_y + fabs(SpeediBfu_y) * vector(rk1Bfu_y, i + 3, j, k));
						LLFp_p3iBfu_z = 0.5 * (Fluxp3iBfu_z + fabs(SpeediBfu_z) * vector(rk1Bfu_z, i + 3, j, k));
						LLFp_p3iDf = 0.5 * (Fluxp3iDf + fabs(SpeediDf) * vector(rk1Df, i + 3, j, k));
						LLFp_p3iDYf = 0.5 * (Fluxp3iDYf + fabs(SpeediDYf) * vector(rk1DYf, i + 3, j, k));
						LLFp_p3itauf = 0.5 * (Fluxp3itauf + fabs(Speeditauf) * vector(rk1tauf, i + 3, j, k));
						LLFp_p3iphif = 0.5 * (Fluxp3iphif + fabs(Speediphif) * vector(rk1phif, i + 3, j, k));
						LLFn_p3iSfd_x = 0.5 * (Fluxp3iSfd_x - fabs(SpeediSfd_x) * vector(rk1Sfd_x, i + 3, j, k));
						LLFn_p3iSfd_y = 0.5 * (Fluxp3iSfd_y - fabs(SpeediSfd_y) * vector(rk1Sfd_y, i + 3, j, k));
						LLFn_p3iSfd_z = 0.5 * (Fluxp3iSfd_z - fabs(SpeediSfd_z) * vector(rk1Sfd_z, i + 3, j, k));
						LLFn_p3iBfu_x = 0.5 * (Fluxp3iBfu_x - fabs(SpeediBfu_x) * vector(rk1Bfu_x, i + 3, j, k));
						LLFn_p3iBfu_y = 0.5 * (Fluxp3iBfu_y - fabs(SpeediBfu_y) * vector(rk1Bfu_y, i + 3, j, k));
						LLFn_p3iBfu_z = 0.5 * (Fluxp3iBfu_z - fabs(SpeediBfu_z) * vector(rk1Bfu_z, i + 3, j, k));
						LLFn_p3iDf = 0.5 * (Fluxp3iDf - fabs(SpeediDf) * vector(rk1Df, i + 3, j, k));
						LLFn_p3iDYf = 0.5 * (Fluxp3iDYf - fabs(SpeediDYf) * vector(rk1DYf, i + 3, j, k));
						LLFn_p3itauf = 0.5 * (Fluxp3itauf - fabs(Speeditauf) * vector(rk1tauf, i + 3, j, k));
						LLFn_p3iphif = 0.5 * (Fluxp3iphif - fabs(Speediphif) * vector(rk1phif, i + 3, j, k));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_yy, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_zz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_xy, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_xz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1sqBf = vector(rk1Bfu_x, i, j, k) * rk1Bfd_x + vector(rk1Bfu_y, i, j, k) * rk1Bfd_y + vector(rk1Bfu_z, i, j, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) * vector(rk1gtd_xx, i, j, k) + (-vector(rk1gtd_yy, i, j, k) * (vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k))) + (-vector(rk1gtd_zz, i, j, k) * (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k))) + (-(vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k)) * vector(rk1gtd_xx, i, j, k)) + 2.0 * vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yy, i, j, k) - vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) - vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_yz, i, j, k) - vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_zz, i, j, k)) + vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j, k) + rk1gtu_xy * vector(rk1Sfd_y, i, j, k) + rk1gtu_xz * vector(rk1Sfd_z, i, j, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j, k) + rk1gtu_xy * vector(vfd_y, i, j, k) + rk1gtu_xz * vector(vfd_z, i, j, k));
						rk1Bfvf = vector(rk1Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk1Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk1Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_zz, i, j, k) - vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yz, i, j, k)) + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j, k) + rk1gtu_yz * vector(vfd_y, i, j, k) + rk1gtu_zz * vector(vfd_z, i, j, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j, k) + rk1gtu_yy * vector(rk1Sfd_y, i, j, k) + rk1gtu_yz * vector(rk1Sfd_z, i, j, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j, k) + rk1gtu_yz * vector(rk1Sfd_y, i, j, k) + rk1gtu_zz * vector(rk1Sfd_z, i, j, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j, k) + rk1gtu_yy * vector(vfd_y, i, j, k) + rk1gtu_yz * vector(vfd_z, i, j, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j, k) + rk1vfu_y * vector(vfd_y, i, j, k) + rk1vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j, k) * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j, k) * vector(rk1Bfu_x, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k) * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k) * rk1vfu_z + vector(rk1Bfu_z, i, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k) * vector(rk1Bfu_y, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k) * rk1vfu_y + vector(rk1Bfu_y, i, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k) * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k) * rk1vfu_z + vector(rk1Bfu_z, i, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j, k) * vector(rk1Bfu_y, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j, k));
						FluxjSfd_x = FjSfd_x_mainI(vector(rk1Sfd_x, i, j, k), vector(rk1gtd_xx, i, j, k), vector(rk1gtd_xy, i, j, k), vector(rk1gtd_xz, i, j, k), vector(rk1Betau_y, i, j, k), vector(rk1Alpha, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_yy, i, j, k), vector(TauT_yz, i, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_yx, dx, simPlat_dt, ilast, jlast);
						FluxjSfd_y = FjSfd_y_mainI(vector(rk1Sfd_y, i, j, k), vector(rk1gtd_xy, i, j, k), vector(rk1gtd_yy, i, j, k), vector(rk1gtd_yz, i, j, k), vector(rk1Betau_y, i, j, k), vector(rk1Alpha, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_yy, i, j, k), vector(TauT_yz, i, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_yy, dx, simPlat_dt, ilast, jlast);
						FluxjSfd_z = FjSfd_z_mainI(vector(rk1Sfd_z, i, j, k), vector(rk1gtd_xz, i, j, k), vector(rk1gtd_yz, i, j, k), vector(rk1gtd_zz, i, j, k), vector(rk1Betau_y, i, j, k), vector(rk1Alpha, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_yy, i, j, k), vector(TauT_yz, i, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_yz, dx, simPlat_dt, ilast, jlast);
						FluxjBfu_x = FjBfu_x_mainI(vector(rk1Bfu_x, i, j, k), vector(rk1Bfu_y, i, j, k), vector(rk1phif, i, j, k), vector(rk1Betau_x, i, j, k), vector(rk1Betau_y, i, j, k), vector(rk1Alpha, i, j, k), vector(TauM_xy, i, j, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						FluxjBfu_y = FjBfu_y_mainI(vector(rk1phif, i, j, k), vector(rk1Alpha, i, j, k), rk1chi_max, rk1gtu_yy, dx, simPlat_dt, ilast, jlast);
						FluxjBfu_z = FjBfu_z_mainI(vector(rk1Bfu_y, i, j, k), vector(rk1Bfu_z, i, j, k), vector(rk1phif, i, j, k), vector(rk1Betau_y, i, j, k), vector(rk1Betau_z, i, j, k), vector(rk1Alpha, i, j, k), vector(TauM_yz, i, j, k), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxjDf = FjDf_mainI(vector(rk1Df, i, j, k), vector(rk1Betau_y, i, j, k), vector(rk1Alpha, i, j, k), vector(TauN_y, i, j, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						FluxjDYf = FjDYf_mainI(vector(rk1DYf, i, j, k), vector(rk1Betau_y, i, j, k), vector(rk1Alpha, i, j, k), vector(TauNe_y, i, j, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxjtauf = Fjtauf_mainI(vector(rk1Df, i, j, k), vector(rk1tauf, i, j, k), vector(rk1Betau_y, i, j, k), vector(rk1Alpha, i, j, k), vector(TauN_y, i, j, k), rk1sdetg, rk1Sfu_y, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxjphif = Fjphif_mainI(vector(rk1Bfu_y, i, j, k), vector(rk1phif, i, j, k), vector(rk1Betau_y, i, j, k), vector(rk1Alpha, i, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j, k) + rk1sqca - vector(sqcs, i, j, k) * rk1sqca;
						SpeedjSfd_x = fabs(fabs(vector(rk1Betau_y, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y))))));
						SpeedjSfd_y = fabs(fabs(vector(rk1Betau_y, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y))))));
						SpeedjSfd_z = fabs(fabs(vector(rk1Betau_y, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y))))));
						SpeedjBfu_x = fabs(fabs(vector(rk1Betau_y, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y))))));
						SpeedjBfu_y = fabs(fabs(vector(rk1Betau_y, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y))))));
						SpeedjBfu_z = fabs(fabs(vector(rk1Betau_y, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y))))));
						SpeedjDf = fabs(fabs(vector(rk1Betau_y, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y))))));
						SpeedjDYf = fabs(fabs(vector(rk1Betau_y, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y))))));
						Speedjtauf = fabs(fabs(vector(rk1Betau_y, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y))))));
						Speedjphif = fabs(fabs(vector(rk1Betau_y, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j - 3, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j - 3, k) * vector(rk1Bfu_x, i, j - 3, k) + vector(rk1gtd_yy, i, j - 3, k) * vector(rk1Bfu_y, i, j - 3, k) + vector(rk1gtd_yz, i, j - 3, k) * vector(rk1Bfu_z, i, j - 3, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j - 3, k) * vector(rk1Bfu_x, i, j - 3, k) + vector(rk1gtd_yz, i, j - 3, k) * vector(rk1Bfu_y, i, j - 3, k) + vector(rk1gtd_zz, i, j - 3, k) * vector(rk1Bfu_z, i, j - 3, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j - 3, k) * vector(rk1Bfu_x, i, j - 3, k) + vector(rk1gtd_xy, i, j - 3, k) * vector(rk1Bfu_y, i, j - 3, k) + vector(rk1gtd_xz, i, j - 3, k) * vector(rk1Bfu_z, i, j - 3, k));
						rk1sqBf = vector(rk1Bfu_x, i, j - 3, k) * rk1Bfd_x + vector(rk1Bfu_y, i, j - 3, k) * rk1Bfd_y + vector(rk1Bfu_z, i, j - 3, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j - 3, k) * vector(rk1gtd_zz, i, j - 3, k) * vector(rk1gtd_xx, i, j - 3, k) + (-vector(rk1gtd_yy, i, j - 3, k) * (vector(rk1gtd_xz, i, j - 3, k) * vector(rk1gtd_xz, i, j - 3, k))) + (-vector(rk1gtd_zz, i, j - 3, k) * (vector(rk1gtd_xy, i, j - 3, k) * vector(rk1gtd_xy, i, j - 3, k))) + (-(vector(rk1gtd_yz, i, j - 3, k) * vector(rk1gtd_yz, i, j - 3, k)) * vector(rk1gtd_xx, i, j - 3, k)) + 2.0 * vector(rk1gtd_yz, i, j - 3, k) * vector(rk1gtd_xy, i, j - 3, k) * vector(rk1gtd_xz, i, j - 3, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j - 3, k) * vector(rk1gtd_yy, i, j - 3, k) - vector(rk1gtd_xy, i, j - 3, k) * vector(rk1gtd_xy, i, j - 3, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j - 3, k) * vector(rk1gtd_zz, i, j - 3, k) - vector(rk1gtd_yz, i, j - 3, k) * vector(rk1gtd_yz, i, j - 3, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j - 3, k) * vector(rk1gtd_yz, i, j - 3, k) - vector(rk1gtd_yy, i, j - 3, k) * vector(rk1gtd_xz, i, j - 3, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j - 3, k) * vector(rk1gtd_zz, i, j - 3, k)) + vector(rk1gtd_yz, i, j - 3, k) * vector(rk1gtd_xz, i, j - 3, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j - 3, k) + rk1gtu_xy * vector(rk1Sfd_y, i, j - 3, k) + rk1gtu_xz * vector(rk1Sfd_z, i, j - 3, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j - 3, k) + rk1gtu_xy * vector(vfd_y, i, j - 3, k) + rk1gtu_xz * vector(vfd_z, i, j - 3, k));
						rk1Bfvf = vector(rk1Bfu_x, i, j - 3, k) * vector(vfd_x, i, j - 3, k) + vector(rk1Bfu_y, i, j - 3, k) * vector(vfd_y, i, j - 3, k) + vector(rk1Bfu_z, i, j - 3, k) * vector(vfd_z, i, j - 3, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j - 3, k) * vector(rk1gtd_zz, i, j - 3, k) - vector(rk1gtd_xz, i, j - 3, k) * vector(rk1gtd_xz, i, j - 3, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j - 3, k) * vector(rk1gtd_yz, i, j - 3, k)) + vector(rk1gtd_xy, i, j - 3, k) * vector(rk1gtd_xz, i, j - 3, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j - 3, k) + rk1gtu_yz * vector(vfd_y, i, j - 3, k) + rk1gtu_zz * vector(vfd_z, i, j - 3, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j - 3, k) + rk1gtu_yy * vector(rk1Sfd_y, i, j - 3, k) + rk1gtu_yz * vector(rk1Sfd_z, i, j - 3, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j - 3, k) + rk1gtu_yz * vector(rk1Sfd_y, i, j - 3, k) + rk1gtu_zz * vector(rk1Sfd_z, i, j - 3, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j - 3, k) + rk1gtu_yy * vector(vfd_y, i, j - 3, k) + rk1gtu_yz * vector(vfd_z, i, j - 3, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j - 3, k) + rk1vfu_y * vector(vfd_y, i, j - 3, k) + rk1vfu_z * vector(vfd_z, i, j - 3, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j - 3, k) * vector(rk1Bfu_z, i, j - 3, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j - 3, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j - 3, k) * vector(rk1Bfu_x, i, j - 3, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j - 3, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j - 3, k) * vector(rk1Bfu_z, i, j - 3, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j - 3, k) * rk1vfu_z + vector(rk1Bfu_z, i, j - 3, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j - 3, k) * vector(rk1Bfu_y, i, j - 3, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j - 3, k) * rk1vfu_y + vector(rk1Bfu_y, i, j - 3, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j - 3, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j - 3, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j - 3, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j - 3, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j - 3, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j - 3, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j - 3, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j - 3, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j - 3, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j - 3, k) * vector(rk1Bfu_z, i, j - 3, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j - 3, k) * rk1vfu_z + vector(rk1Bfu_z, i, j - 3, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j - 3, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j - 3, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j - 3, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j - 3, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j - 3, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j - 3, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j - 3, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j - 3, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j - 3, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j - 3, k) * vector(rk1Bfu_y, i, j - 3, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j - 3, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j - 3, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j - 3, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j - 3, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j - 3, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j - 3, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j - 3, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j - 3, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j - 3, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j - 3, k));
						Fluxm3jSfd_x = FjSfd_x_mainI(vector(rk1Sfd_x, i, j - 3, k), vector(rk1gtd_xx, i, j - 3, k), vector(rk1gtd_xy, i, j - 3, k), vector(rk1gtd_xz, i, j - 3, k), vector(rk1Betau_y, i, j - 3, k), vector(rk1Alpha, i, j - 3, k), vector(TauT_xy, i, j - 3, k), vector(TauT_yy, i, j - 3, k), vector(TauT_yz, i, j - 3, k), rk1inv_chi, rk1sdetg, rk1Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxm3jSfd_y = FjSfd_y_mainI(vector(rk1Sfd_y, i, j - 3, k), vector(rk1gtd_xy, i, j - 3, k), vector(rk1gtd_yy, i, j - 3, k), vector(rk1gtd_yz, i, j - 3, k), vector(rk1Betau_y, i, j - 3, k), vector(rk1Alpha, i, j - 3, k), vector(TauT_xy, i, j - 3, k), vector(TauT_yy, i, j - 3, k), vector(TauT_yz, i, j - 3, k), rk1inv_chi, rk1sdetg, rk1Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm3jSfd_z = FjSfd_z_mainI(vector(rk1Sfd_z, i, j - 3, k), vector(rk1gtd_xz, i, j - 3, k), vector(rk1gtd_yz, i, j - 3, k), vector(rk1gtd_zz, i, j - 3, k), vector(rk1Betau_y, i, j - 3, k), vector(rk1Alpha, i, j - 3, k), vector(TauT_xy, i, j - 3, k), vector(TauT_yy, i, j - 3, k), vector(TauT_yz, i, j - 3, k), rk1inv_chi, rk1sdetg, rk1Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxm3jBfu_x = FjBfu_x_mainI(vector(rk1Bfu_x, i, j - 3, k), vector(rk1Bfu_y, i, j - 3, k), vector(rk1phif, i, j - 3, k), vector(rk1Betau_x, i, j - 3, k), vector(rk1Betau_y, i, j - 3, k), vector(rk1Alpha, i, j - 3, k), vector(TauM_xy, i, j - 3, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3jBfu_y = FjBfu_y_mainI(vector(rk1phif, i, j - 3, k), vector(rk1Alpha, i, j - 3, k), rk1chi_max, rk1gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm3jBfu_z = FjBfu_z_mainI(vector(rk1Bfu_y, i, j - 3, k), vector(rk1Bfu_z, i, j - 3, k), vector(rk1phif, i, j - 3, k), vector(rk1Betau_y, i, j - 3, k), vector(rk1Betau_z, i, j - 3, k), vector(rk1Alpha, i, j - 3, k), vector(TauM_yz, i, j - 3, k), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3jDf = FjDf_mainI(vector(rk1Df, i, j - 3, k), vector(rk1Betau_y, i, j - 3, k), vector(rk1Alpha, i, j - 3, k), vector(TauN_y, i, j - 3, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3jDYf = FjDYf_mainI(vector(rk1DYf, i, j - 3, k), vector(rk1Betau_y, i, j - 3, k), vector(rk1Alpha, i, j - 3, k), vector(TauNe_y, i, j - 3, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3jtauf = Fjtauf_mainI(vector(rk1Df, i, j - 3, k), vector(rk1tauf, i, j - 3, k), vector(rk1Betau_y, i, j - 3, k), vector(rk1Alpha, i, j - 3, k), vector(TauN_y, i, j - 3, k), rk1sdetg, rk1Sfu_y, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3jphif = Fjphif_mainI(vector(rk1Bfu_y, i, j - 3, k), vector(rk1phif, i, j - 3, k), vector(rk1Betau_y, i, j - 3, k), vector(rk1Alpha, i, j - 3, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j - 3, k) * (1.0 + vector(epsf, i, j - 3, k)) + vector(pf, i, j - 3, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j - 3, k) + rk1sqca - vector(sqcs, i, j - 3, k) * rk1sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk1Betau_y, i, j - 3, k)) + vector(rk1Alpha, i, j - 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk1Betau_y, i, j - 3, k)) + vector(rk1Alpha, i, j - 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk1Betau_y, i, j - 3, k)) + vector(rk1Alpha, i, j - 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk1Betau_y, i, j - 3, k)) + vector(rk1Alpha, i, j - 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk1Betau_y, i, j - 3, k)) + vector(rk1Alpha, i, j - 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk1Betau_y, i, j - 3, k)) + vector(rk1Alpha, i, j - 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk1Betau_y, i, j - 3, k)) + vector(rk1Alpha, i, j - 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk1Betau_y, i, j - 3, k)) + vector(rk1Alpha, i, j - 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk1Betau_y, i, j - 3, k)) + vector(rk1Alpha, i, j - 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk1Betau_y, i, j - 3, k)) + vector(rk1Alpha, i, j - 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j - 2, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j - 2, k) * vector(rk1Bfu_x, i, j - 2, k) + vector(rk1gtd_yy, i, j - 2, k) * vector(rk1Bfu_y, i, j - 2, k) + vector(rk1gtd_yz, i, j - 2, k) * vector(rk1Bfu_z, i, j - 2, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j - 2, k) * vector(rk1Bfu_x, i, j - 2, k) + vector(rk1gtd_yz, i, j - 2, k) * vector(rk1Bfu_y, i, j - 2, k) + vector(rk1gtd_zz, i, j - 2, k) * vector(rk1Bfu_z, i, j - 2, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j - 2, k) * vector(rk1Bfu_x, i, j - 2, k) + vector(rk1gtd_xy, i, j - 2, k) * vector(rk1Bfu_y, i, j - 2, k) + vector(rk1gtd_xz, i, j - 2, k) * vector(rk1Bfu_z, i, j - 2, k));
						rk1sqBf = vector(rk1Bfu_x, i, j - 2, k) * rk1Bfd_x + vector(rk1Bfu_y, i, j - 2, k) * rk1Bfd_y + vector(rk1Bfu_z, i, j - 2, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j - 2, k) * vector(rk1gtd_zz, i, j - 2, k) * vector(rk1gtd_xx, i, j - 2, k) + (-vector(rk1gtd_yy, i, j - 2, k) * (vector(rk1gtd_xz, i, j - 2, k) * vector(rk1gtd_xz, i, j - 2, k))) + (-vector(rk1gtd_zz, i, j - 2, k) * (vector(rk1gtd_xy, i, j - 2, k) * vector(rk1gtd_xy, i, j - 2, k))) + (-(vector(rk1gtd_yz, i, j - 2, k) * vector(rk1gtd_yz, i, j - 2, k)) * vector(rk1gtd_xx, i, j - 2, k)) + 2.0 * vector(rk1gtd_yz, i, j - 2, k) * vector(rk1gtd_xy, i, j - 2, k) * vector(rk1gtd_xz, i, j - 2, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j - 2, k) * vector(rk1gtd_yy, i, j - 2, k) - vector(rk1gtd_xy, i, j - 2, k) * vector(rk1gtd_xy, i, j - 2, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j - 2, k) * vector(rk1gtd_zz, i, j - 2, k) - vector(rk1gtd_yz, i, j - 2, k) * vector(rk1gtd_yz, i, j - 2, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j - 2, k) * vector(rk1gtd_yz, i, j - 2, k) - vector(rk1gtd_yy, i, j - 2, k) * vector(rk1gtd_xz, i, j - 2, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j - 2, k) * vector(rk1gtd_zz, i, j - 2, k)) + vector(rk1gtd_yz, i, j - 2, k) * vector(rk1gtd_xz, i, j - 2, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j - 2, k) + rk1gtu_xy * vector(rk1Sfd_y, i, j - 2, k) + rk1gtu_xz * vector(rk1Sfd_z, i, j - 2, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j - 2, k) + rk1gtu_xy * vector(vfd_y, i, j - 2, k) + rk1gtu_xz * vector(vfd_z, i, j - 2, k));
						rk1Bfvf = vector(rk1Bfu_x, i, j - 2, k) * vector(vfd_x, i, j - 2, k) + vector(rk1Bfu_y, i, j - 2, k) * vector(vfd_y, i, j - 2, k) + vector(rk1Bfu_z, i, j - 2, k) * vector(vfd_z, i, j - 2, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j - 2, k) * vector(rk1gtd_zz, i, j - 2, k) - vector(rk1gtd_xz, i, j - 2, k) * vector(rk1gtd_xz, i, j - 2, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j - 2, k) * vector(rk1gtd_yz, i, j - 2, k)) + vector(rk1gtd_xy, i, j - 2, k) * vector(rk1gtd_xz, i, j - 2, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j - 2, k) + rk1gtu_yz * vector(vfd_y, i, j - 2, k) + rk1gtu_zz * vector(vfd_z, i, j - 2, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j - 2, k) + rk1gtu_yy * vector(rk1Sfd_y, i, j - 2, k) + rk1gtu_yz * vector(rk1Sfd_z, i, j - 2, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j - 2, k) + rk1gtu_yz * vector(rk1Sfd_y, i, j - 2, k) + rk1gtu_zz * vector(rk1Sfd_z, i, j - 2, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j - 2, k) + rk1gtu_yy * vector(vfd_y, i, j - 2, k) + rk1gtu_yz * vector(vfd_z, i, j - 2, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j - 2, k) + rk1vfu_y * vector(vfd_y, i, j - 2, k) + rk1vfu_z * vector(vfd_z, i, j - 2, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j - 2, k) * vector(rk1Bfu_z, i, j - 2, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j - 2, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j - 2, k) * vector(rk1Bfu_x, i, j - 2, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j - 2, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j - 2, k) * vector(rk1Bfu_z, i, j - 2, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j - 2, k) * rk1vfu_z + vector(rk1Bfu_z, i, j - 2, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j - 2, k) * vector(rk1Bfu_y, i, j - 2, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j - 2, k) * rk1vfu_y + vector(rk1Bfu_y, i, j - 2, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j - 2, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j - 2, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j - 2, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j - 2, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j - 2, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j - 2, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j - 2, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j - 2, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j - 2, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j - 2, k) * vector(rk1Bfu_z, i, j - 2, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j - 2, k) * rk1vfu_z + vector(rk1Bfu_z, i, j - 2, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j - 2, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j - 2, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j - 2, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j - 2, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j - 2, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j - 2, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j - 2, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j - 2, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j - 2, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j - 2, k) * vector(rk1Bfu_y, i, j - 2, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j - 2, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j - 2, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j - 2, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j - 2, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j - 2, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j - 2, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j - 2, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j - 2, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j - 2, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j - 2, k));
						Fluxm2jSfd_x = FjSfd_x_mainI(vector(rk1Sfd_x, i, j - 2, k), vector(rk1gtd_xx, i, j - 2, k), vector(rk1gtd_xy, i, j - 2, k), vector(rk1gtd_xz, i, j - 2, k), vector(rk1Betau_y, i, j - 2, k), vector(rk1Alpha, i, j - 2, k), vector(TauT_xy, i, j - 2, k), vector(TauT_yy, i, j - 2, k), vector(TauT_yz, i, j - 2, k), rk1inv_chi, rk1sdetg, rk1Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxm2jSfd_y = FjSfd_y_mainI(vector(rk1Sfd_y, i, j - 2, k), vector(rk1gtd_xy, i, j - 2, k), vector(rk1gtd_yy, i, j - 2, k), vector(rk1gtd_yz, i, j - 2, k), vector(rk1Betau_y, i, j - 2, k), vector(rk1Alpha, i, j - 2, k), vector(TauT_xy, i, j - 2, k), vector(TauT_yy, i, j - 2, k), vector(TauT_yz, i, j - 2, k), rk1inv_chi, rk1sdetg, rk1Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm2jSfd_z = FjSfd_z_mainI(vector(rk1Sfd_z, i, j - 2, k), vector(rk1gtd_xz, i, j - 2, k), vector(rk1gtd_yz, i, j - 2, k), vector(rk1gtd_zz, i, j - 2, k), vector(rk1Betau_y, i, j - 2, k), vector(rk1Alpha, i, j - 2, k), vector(TauT_xy, i, j - 2, k), vector(TauT_yy, i, j - 2, k), vector(TauT_yz, i, j - 2, k), rk1inv_chi, rk1sdetg, rk1Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxm2jBfu_x = FjBfu_x_mainI(vector(rk1Bfu_x, i, j - 2, k), vector(rk1Bfu_y, i, j - 2, k), vector(rk1phif, i, j - 2, k), vector(rk1Betau_x, i, j - 2, k), vector(rk1Betau_y, i, j - 2, k), vector(rk1Alpha, i, j - 2, k), vector(TauM_xy, i, j - 2, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2jBfu_y = FjBfu_y_mainI(vector(rk1phif, i, j - 2, k), vector(rk1Alpha, i, j - 2, k), rk1chi_max, rk1gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm2jBfu_z = FjBfu_z_mainI(vector(rk1Bfu_y, i, j - 2, k), vector(rk1Bfu_z, i, j - 2, k), vector(rk1phif, i, j - 2, k), vector(rk1Betau_y, i, j - 2, k), vector(rk1Betau_z, i, j - 2, k), vector(rk1Alpha, i, j - 2, k), vector(TauM_yz, i, j - 2, k), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2jDf = FjDf_mainI(vector(rk1Df, i, j - 2, k), vector(rk1Betau_y, i, j - 2, k), vector(rk1Alpha, i, j - 2, k), vector(TauN_y, i, j - 2, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2jDYf = FjDYf_mainI(vector(rk1DYf, i, j - 2, k), vector(rk1Betau_y, i, j - 2, k), vector(rk1Alpha, i, j - 2, k), vector(TauNe_y, i, j - 2, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2jtauf = Fjtauf_mainI(vector(rk1Df, i, j - 2, k), vector(rk1tauf, i, j - 2, k), vector(rk1Betau_y, i, j - 2, k), vector(rk1Alpha, i, j - 2, k), vector(TauN_y, i, j - 2, k), rk1sdetg, rk1Sfu_y, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2jphif = Fjphif_mainI(vector(rk1Bfu_y, i, j - 2, k), vector(rk1phif, i, j - 2, k), vector(rk1Betau_y, i, j - 2, k), vector(rk1Alpha, i, j - 2, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j - 2, k) * (1.0 + vector(epsf, i, j - 2, k)) + vector(pf, i, j - 2, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j - 2, k) + rk1sqca - vector(sqcs, i, j - 2, k) * rk1sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk1Betau_y, i, j - 2, k)) + vector(rk1Alpha, i, j - 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk1Betau_y, i, j - 2, k)) + vector(rk1Alpha, i, j - 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk1Betau_y, i, j - 2, k)) + vector(rk1Alpha, i, j - 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk1Betau_y, i, j - 2, k)) + vector(rk1Alpha, i, j - 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk1Betau_y, i, j - 2, k)) + vector(rk1Alpha, i, j - 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk1Betau_y, i, j - 2, k)) + vector(rk1Alpha, i, j - 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk1Betau_y, i, j - 2, k)) + vector(rk1Alpha, i, j - 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk1Betau_y, i, j - 2, k)) + vector(rk1Alpha, i, j - 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk1Betau_y, i, j - 2, k)) + vector(rk1Alpha, i, j - 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk1Betau_y, i, j - 2, k)) + vector(rk1Alpha, i, j - 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j - 1, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j - 1, k) * vector(rk1Bfu_x, i, j - 1, k) + vector(rk1gtd_yy, i, j - 1, k) * vector(rk1Bfu_y, i, j - 1, k) + vector(rk1gtd_yz, i, j - 1, k) * vector(rk1Bfu_z, i, j - 1, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j - 1, k) * vector(rk1Bfu_x, i, j - 1, k) + vector(rk1gtd_yz, i, j - 1, k) * vector(rk1Bfu_y, i, j - 1, k) + vector(rk1gtd_zz, i, j - 1, k) * vector(rk1Bfu_z, i, j - 1, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j - 1, k) * vector(rk1Bfu_x, i, j - 1, k) + vector(rk1gtd_xy, i, j - 1, k) * vector(rk1Bfu_y, i, j - 1, k) + vector(rk1gtd_xz, i, j - 1, k) * vector(rk1Bfu_z, i, j - 1, k));
						rk1sqBf = vector(rk1Bfu_x, i, j - 1, k) * rk1Bfd_x + vector(rk1Bfu_y, i, j - 1, k) * rk1Bfd_y + vector(rk1Bfu_z, i, j - 1, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j - 1, k) * vector(rk1gtd_zz, i, j - 1, k) * vector(rk1gtd_xx, i, j - 1, k) + (-vector(rk1gtd_yy, i, j - 1, k) * (vector(rk1gtd_xz, i, j - 1, k) * vector(rk1gtd_xz, i, j - 1, k))) + (-vector(rk1gtd_zz, i, j - 1, k) * (vector(rk1gtd_xy, i, j - 1, k) * vector(rk1gtd_xy, i, j - 1, k))) + (-(vector(rk1gtd_yz, i, j - 1, k) * vector(rk1gtd_yz, i, j - 1, k)) * vector(rk1gtd_xx, i, j - 1, k)) + 2.0 * vector(rk1gtd_yz, i, j - 1, k) * vector(rk1gtd_xy, i, j - 1, k) * vector(rk1gtd_xz, i, j - 1, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j - 1, k) * vector(rk1gtd_yy, i, j - 1, k) - vector(rk1gtd_xy, i, j - 1, k) * vector(rk1gtd_xy, i, j - 1, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j - 1, k) * vector(rk1gtd_zz, i, j - 1, k) - vector(rk1gtd_yz, i, j - 1, k) * vector(rk1gtd_yz, i, j - 1, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j - 1, k) * vector(rk1gtd_yz, i, j - 1, k) - vector(rk1gtd_yy, i, j - 1, k) * vector(rk1gtd_xz, i, j - 1, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j - 1, k) * vector(rk1gtd_zz, i, j - 1, k)) + vector(rk1gtd_yz, i, j - 1, k) * vector(rk1gtd_xz, i, j - 1, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j - 1, k) + rk1gtu_xy * vector(rk1Sfd_y, i, j - 1, k) + rk1gtu_xz * vector(rk1Sfd_z, i, j - 1, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j - 1, k) + rk1gtu_xy * vector(vfd_y, i, j - 1, k) + rk1gtu_xz * vector(vfd_z, i, j - 1, k));
						rk1Bfvf = vector(rk1Bfu_x, i, j - 1, k) * vector(vfd_x, i, j - 1, k) + vector(rk1Bfu_y, i, j - 1, k) * vector(vfd_y, i, j - 1, k) + vector(rk1Bfu_z, i, j - 1, k) * vector(vfd_z, i, j - 1, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j - 1, k) * vector(rk1gtd_zz, i, j - 1, k) - vector(rk1gtd_xz, i, j - 1, k) * vector(rk1gtd_xz, i, j - 1, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j - 1, k) * vector(rk1gtd_yz, i, j - 1, k)) + vector(rk1gtd_xy, i, j - 1, k) * vector(rk1gtd_xz, i, j - 1, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j - 1, k) + rk1gtu_yz * vector(vfd_y, i, j - 1, k) + rk1gtu_zz * vector(vfd_z, i, j - 1, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j - 1, k) + rk1gtu_yy * vector(rk1Sfd_y, i, j - 1, k) + rk1gtu_yz * vector(rk1Sfd_z, i, j - 1, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j - 1, k) + rk1gtu_yz * vector(rk1Sfd_y, i, j - 1, k) + rk1gtu_zz * vector(rk1Sfd_z, i, j - 1, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j - 1, k) + rk1gtu_yy * vector(vfd_y, i, j - 1, k) + rk1gtu_yz * vector(vfd_z, i, j - 1, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j - 1, k) + rk1vfu_y * vector(vfd_y, i, j - 1, k) + rk1vfu_z * vector(vfd_z, i, j - 1, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j - 1, k) * vector(rk1Bfu_z, i, j - 1, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j - 1, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j - 1, k) * vector(rk1Bfu_x, i, j - 1, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j - 1, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j - 1, k) * vector(rk1Bfu_z, i, j - 1, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j - 1, k) * rk1vfu_z + vector(rk1Bfu_z, i, j - 1, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j - 1, k) * vector(rk1Bfu_y, i, j - 1, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j - 1, k) * rk1vfu_y + vector(rk1Bfu_y, i, j - 1, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j - 1, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j - 1, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j - 1, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j - 1, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j - 1, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j - 1, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j - 1, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j - 1, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j - 1, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j - 1, k) * vector(rk1Bfu_z, i, j - 1, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j - 1, k) * rk1vfu_z + vector(rk1Bfu_z, i, j - 1, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j - 1, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j - 1, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j - 1, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j - 1, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j - 1, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j - 1, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j - 1, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j - 1, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j - 1, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j - 1, k) * vector(rk1Bfu_y, i, j - 1, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j - 1, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j - 1, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j - 1, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j - 1, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j - 1, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j - 1, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j - 1, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j - 1, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j - 1, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j - 1, k));
						Fluxm1jSfd_x = FjSfd_x_mainI(vector(rk1Sfd_x, i, j - 1, k), vector(rk1gtd_xx, i, j - 1, k), vector(rk1gtd_xy, i, j - 1, k), vector(rk1gtd_xz, i, j - 1, k), vector(rk1Betau_y, i, j - 1, k), vector(rk1Alpha, i, j - 1, k), vector(TauT_xy, i, j - 1, k), vector(TauT_yy, i, j - 1, k), vector(TauT_yz, i, j - 1, k), rk1inv_chi, rk1sdetg, rk1Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxm1jSfd_y = FjSfd_y_mainI(vector(rk1Sfd_y, i, j - 1, k), vector(rk1gtd_xy, i, j - 1, k), vector(rk1gtd_yy, i, j - 1, k), vector(rk1gtd_yz, i, j - 1, k), vector(rk1Betau_y, i, j - 1, k), vector(rk1Alpha, i, j - 1, k), vector(TauT_xy, i, j - 1, k), vector(TauT_yy, i, j - 1, k), vector(TauT_yz, i, j - 1, k), rk1inv_chi, rk1sdetg, rk1Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm1jSfd_z = FjSfd_z_mainI(vector(rk1Sfd_z, i, j - 1, k), vector(rk1gtd_xz, i, j - 1, k), vector(rk1gtd_yz, i, j - 1, k), vector(rk1gtd_zz, i, j - 1, k), vector(rk1Betau_y, i, j - 1, k), vector(rk1Alpha, i, j - 1, k), vector(TauT_xy, i, j - 1, k), vector(TauT_yy, i, j - 1, k), vector(TauT_yz, i, j - 1, k), rk1inv_chi, rk1sdetg, rk1Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxm1jBfu_x = FjBfu_x_mainI(vector(rk1Bfu_x, i, j - 1, k), vector(rk1Bfu_y, i, j - 1, k), vector(rk1phif, i, j - 1, k), vector(rk1Betau_x, i, j - 1, k), vector(rk1Betau_y, i, j - 1, k), vector(rk1Alpha, i, j - 1, k), vector(TauM_xy, i, j - 1, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1jBfu_y = FjBfu_y_mainI(vector(rk1phif, i, j - 1, k), vector(rk1Alpha, i, j - 1, k), rk1chi_max, rk1gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm1jBfu_z = FjBfu_z_mainI(vector(rk1Bfu_y, i, j - 1, k), vector(rk1Bfu_z, i, j - 1, k), vector(rk1phif, i, j - 1, k), vector(rk1Betau_y, i, j - 1, k), vector(rk1Betau_z, i, j - 1, k), vector(rk1Alpha, i, j - 1, k), vector(TauM_yz, i, j - 1, k), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1jDf = FjDf_mainI(vector(rk1Df, i, j - 1, k), vector(rk1Betau_y, i, j - 1, k), vector(rk1Alpha, i, j - 1, k), vector(TauN_y, i, j - 1, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1jDYf = FjDYf_mainI(vector(rk1DYf, i, j - 1, k), vector(rk1Betau_y, i, j - 1, k), vector(rk1Alpha, i, j - 1, k), vector(TauNe_y, i, j - 1, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1jtauf = Fjtauf_mainI(vector(rk1Df, i, j - 1, k), vector(rk1tauf, i, j - 1, k), vector(rk1Betau_y, i, j - 1, k), vector(rk1Alpha, i, j - 1, k), vector(TauN_y, i, j - 1, k), rk1sdetg, rk1Sfu_y, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1jphif = Fjphif_mainI(vector(rk1Bfu_y, i, j - 1, k), vector(rk1phif, i, j - 1, k), vector(rk1Betau_y, i, j - 1, k), vector(rk1Alpha, i, j - 1, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j - 1, k) * (1.0 + vector(epsf, i, j - 1, k)) + vector(pf, i, j - 1, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j - 1, k) + rk1sqca - vector(sqcs, i, j - 1, k) * rk1sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk1Betau_y, i, j - 1, k)) + vector(rk1Alpha, i, j - 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk1Betau_y, i, j - 1, k)) + vector(rk1Alpha, i, j - 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk1Betau_y, i, j - 1, k)) + vector(rk1Alpha, i, j - 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk1Betau_y, i, j - 1, k)) + vector(rk1Alpha, i, j - 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk1Betau_y, i, j - 1, k)) + vector(rk1Alpha, i, j - 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk1Betau_y, i, j - 1, k)) + vector(rk1Alpha, i, j - 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk1Betau_y, i, j - 1, k)) + vector(rk1Alpha, i, j - 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk1Betau_y, i, j - 1, k)) + vector(rk1Alpha, i, j - 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk1Betau_y, i, j - 1, k)) + vector(rk1Alpha, i, j - 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk1Betau_y, i, j - 1, k)) + vector(rk1Alpha, i, j - 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j + 1, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j + 1, k) * vector(rk1Bfu_x, i, j + 1, k) + vector(rk1gtd_yy, i, j + 1, k) * vector(rk1Bfu_y, i, j + 1, k) + vector(rk1gtd_yz, i, j + 1, k) * vector(rk1Bfu_z, i, j + 1, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j + 1, k) * vector(rk1Bfu_x, i, j + 1, k) + vector(rk1gtd_yz, i, j + 1, k) * vector(rk1Bfu_y, i, j + 1, k) + vector(rk1gtd_zz, i, j + 1, k) * vector(rk1Bfu_z, i, j + 1, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j + 1, k) * vector(rk1Bfu_x, i, j + 1, k) + vector(rk1gtd_xy, i, j + 1, k) * vector(rk1Bfu_y, i, j + 1, k) + vector(rk1gtd_xz, i, j + 1, k) * vector(rk1Bfu_z, i, j + 1, k));
						rk1sqBf = vector(rk1Bfu_x, i, j + 1, k) * rk1Bfd_x + vector(rk1Bfu_y, i, j + 1, k) * rk1Bfd_y + vector(rk1Bfu_z, i, j + 1, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j + 1, k) * vector(rk1gtd_zz, i, j + 1, k) * vector(rk1gtd_xx, i, j + 1, k) + (-vector(rk1gtd_yy, i, j + 1, k) * (vector(rk1gtd_xz, i, j + 1, k) * vector(rk1gtd_xz, i, j + 1, k))) + (-vector(rk1gtd_zz, i, j + 1, k) * (vector(rk1gtd_xy, i, j + 1, k) * vector(rk1gtd_xy, i, j + 1, k))) + (-(vector(rk1gtd_yz, i, j + 1, k) * vector(rk1gtd_yz, i, j + 1, k)) * vector(rk1gtd_xx, i, j + 1, k)) + 2.0 * vector(rk1gtd_yz, i, j + 1, k) * vector(rk1gtd_xy, i, j + 1, k) * vector(rk1gtd_xz, i, j + 1, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j + 1, k) * vector(rk1gtd_yy, i, j + 1, k) - vector(rk1gtd_xy, i, j + 1, k) * vector(rk1gtd_xy, i, j + 1, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j + 1, k) * vector(rk1gtd_zz, i, j + 1, k) - vector(rk1gtd_yz, i, j + 1, k) * vector(rk1gtd_yz, i, j + 1, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j + 1, k) * vector(rk1gtd_yz, i, j + 1, k) - vector(rk1gtd_yy, i, j + 1, k) * vector(rk1gtd_xz, i, j + 1, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j + 1, k) * vector(rk1gtd_zz, i, j + 1, k)) + vector(rk1gtd_yz, i, j + 1, k) * vector(rk1gtd_xz, i, j + 1, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j + 1, k) + rk1gtu_xy * vector(rk1Sfd_y, i, j + 1, k) + rk1gtu_xz * vector(rk1Sfd_z, i, j + 1, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j + 1, k) + rk1gtu_xy * vector(vfd_y, i, j + 1, k) + rk1gtu_xz * vector(vfd_z, i, j + 1, k));
						rk1Bfvf = vector(rk1Bfu_x, i, j + 1, k) * vector(vfd_x, i, j + 1, k) + vector(rk1Bfu_y, i, j + 1, k) * vector(vfd_y, i, j + 1, k) + vector(rk1Bfu_z, i, j + 1, k) * vector(vfd_z, i, j + 1, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j + 1, k) * vector(rk1gtd_zz, i, j + 1, k) - vector(rk1gtd_xz, i, j + 1, k) * vector(rk1gtd_xz, i, j + 1, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j + 1, k) * vector(rk1gtd_yz, i, j + 1, k)) + vector(rk1gtd_xy, i, j + 1, k) * vector(rk1gtd_xz, i, j + 1, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j + 1, k) + rk1gtu_yz * vector(vfd_y, i, j + 1, k) + rk1gtu_zz * vector(vfd_z, i, j + 1, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j + 1, k) + rk1gtu_yy * vector(rk1Sfd_y, i, j + 1, k) + rk1gtu_yz * vector(rk1Sfd_z, i, j + 1, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j + 1, k) + rk1gtu_yz * vector(rk1Sfd_y, i, j + 1, k) + rk1gtu_zz * vector(rk1Sfd_z, i, j + 1, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j + 1, k) + rk1gtu_yy * vector(vfd_y, i, j + 1, k) + rk1gtu_yz * vector(vfd_z, i, j + 1, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j + 1, k) + rk1vfu_y * vector(vfd_y, i, j + 1, k) + rk1vfu_z * vector(vfd_z, i, j + 1, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j + 1, k) * vector(rk1Bfu_z, i, j + 1, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j + 1, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j + 1, k) * vector(rk1Bfu_x, i, j + 1, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j + 1, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j + 1, k) * vector(rk1Bfu_z, i, j + 1, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j + 1, k) * rk1vfu_z + vector(rk1Bfu_z, i, j + 1, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j + 1, k) * vector(rk1Bfu_y, i, j + 1, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j + 1, k) * rk1vfu_y + vector(rk1Bfu_y, i, j + 1, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j + 1, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j + 1, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j + 1, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j + 1, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j + 1, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j + 1, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j + 1, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j + 1, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j + 1, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j + 1, k) * vector(rk1Bfu_z, i, j + 1, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j + 1, k) * rk1vfu_z + vector(rk1Bfu_z, i, j + 1, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j + 1, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j + 1, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j + 1, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j + 1, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j + 1, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j + 1, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j + 1, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j + 1, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j + 1, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j + 1, k) * vector(rk1Bfu_y, i, j + 1, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j + 1, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j + 1, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j + 1, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j + 1, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j + 1, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j + 1, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j + 1, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j + 1, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j + 1, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j + 1, k));
						Fluxp1jSfd_x = FjSfd_x_mainI(vector(rk1Sfd_x, i, j + 1, k), vector(rk1gtd_xx, i, j + 1, k), vector(rk1gtd_xy, i, j + 1, k), vector(rk1gtd_xz, i, j + 1, k), vector(rk1Betau_y, i, j + 1, k), vector(rk1Alpha, i, j + 1, k), vector(TauT_xy, i, j + 1, k), vector(TauT_yy, i, j + 1, k), vector(TauT_yz, i, j + 1, k), rk1inv_chi, rk1sdetg, rk1Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxp1jSfd_y = FjSfd_y_mainI(vector(rk1Sfd_y, i, j + 1, k), vector(rk1gtd_xy, i, j + 1, k), vector(rk1gtd_yy, i, j + 1, k), vector(rk1gtd_yz, i, j + 1, k), vector(rk1Betau_y, i, j + 1, k), vector(rk1Alpha, i, j + 1, k), vector(TauT_xy, i, j + 1, k), vector(TauT_yy, i, j + 1, k), vector(TauT_yz, i, j + 1, k), rk1inv_chi, rk1sdetg, rk1Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp1jSfd_z = FjSfd_z_mainI(vector(rk1Sfd_z, i, j + 1, k), vector(rk1gtd_xz, i, j + 1, k), vector(rk1gtd_yz, i, j + 1, k), vector(rk1gtd_zz, i, j + 1, k), vector(rk1Betau_y, i, j + 1, k), vector(rk1Alpha, i, j + 1, k), vector(TauT_xy, i, j + 1, k), vector(TauT_yy, i, j + 1, k), vector(TauT_yz, i, j + 1, k), rk1inv_chi, rk1sdetg, rk1Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxp1jBfu_x = FjBfu_x_mainI(vector(rk1Bfu_x, i, j + 1, k), vector(rk1Bfu_y, i, j + 1, k), vector(rk1phif, i, j + 1, k), vector(rk1Betau_x, i, j + 1, k), vector(rk1Betau_y, i, j + 1, k), vector(rk1Alpha, i, j + 1, k), vector(TauM_xy, i, j + 1, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1jBfu_y = FjBfu_y_mainI(vector(rk1phif, i, j + 1, k), vector(rk1Alpha, i, j + 1, k), rk1chi_max, rk1gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp1jBfu_z = FjBfu_z_mainI(vector(rk1Bfu_y, i, j + 1, k), vector(rk1Bfu_z, i, j + 1, k), vector(rk1phif, i, j + 1, k), vector(rk1Betau_y, i, j + 1, k), vector(rk1Betau_z, i, j + 1, k), vector(rk1Alpha, i, j + 1, k), vector(TauM_yz, i, j + 1, k), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1jDf = FjDf_mainI(vector(rk1Df, i, j + 1, k), vector(rk1Betau_y, i, j + 1, k), vector(rk1Alpha, i, j + 1, k), vector(TauN_y, i, j + 1, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1jDYf = FjDYf_mainI(vector(rk1DYf, i, j + 1, k), vector(rk1Betau_y, i, j + 1, k), vector(rk1Alpha, i, j + 1, k), vector(TauNe_y, i, j + 1, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1jtauf = Fjtauf_mainI(vector(rk1Df, i, j + 1, k), vector(rk1tauf, i, j + 1, k), vector(rk1Betau_y, i, j + 1, k), vector(rk1Alpha, i, j + 1, k), vector(TauN_y, i, j + 1, k), rk1sdetg, rk1Sfu_y, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1jphif = Fjphif_mainI(vector(rk1Bfu_y, i, j + 1, k), vector(rk1phif, i, j + 1, k), vector(rk1Betau_y, i, j + 1, k), vector(rk1Alpha, i, j + 1, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j + 1, k) * (1.0 + vector(epsf, i, j + 1, k)) + vector(pf, i, j + 1, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j + 1, k) + rk1sqca - vector(sqcs, i, j + 1, k) * rk1sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk1Betau_y, i, j + 1, k)) + vector(rk1Alpha, i, j + 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk1Betau_y, i, j + 1, k)) + vector(rk1Alpha, i, j + 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk1Betau_y, i, j + 1, k)) + vector(rk1Alpha, i, j + 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk1Betau_y, i, j + 1, k)) + vector(rk1Alpha, i, j + 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk1Betau_y, i, j + 1, k)) + vector(rk1Alpha, i, j + 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk1Betau_y, i, j + 1, k)) + vector(rk1Alpha, i, j + 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk1Betau_y, i, j + 1, k)) + vector(rk1Alpha, i, j + 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk1Betau_y, i, j + 1, k)) + vector(rk1Alpha, i, j + 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk1Betau_y, i, j + 1, k)) + vector(rk1Alpha, i, j + 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk1Betau_y, i, j + 1, k)) + vector(rk1Alpha, i, j + 1, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j + 2, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j + 2, k) * vector(rk1Bfu_x, i, j + 2, k) + vector(rk1gtd_yy, i, j + 2, k) * vector(rk1Bfu_y, i, j + 2, k) + vector(rk1gtd_yz, i, j + 2, k) * vector(rk1Bfu_z, i, j + 2, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j + 2, k) * vector(rk1Bfu_x, i, j + 2, k) + vector(rk1gtd_yz, i, j + 2, k) * vector(rk1Bfu_y, i, j + 2, k) + vector(rk1gtd_zz, i, j + 2, k) * vector(rk1Bfu_z, i, j + 2, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j + 2, k) * vector(rk1Bfu_x, i, j + 2, k) + vector(rk1gtd_xy, i, j + 2, k) * vector(rk1Bfu_y, i, j + 2, k) + vector(rk1gtd_xz, i, j + 2, k) * vector(rk1Bfu_z, i, j + 2, k));
						rk1sqBf = vector(rk1Bfu_x, i, j + 2, k) * rk1Bfd_x + vector(rk1Bfu_y, i, j + 2, k) * rk1Bfd_y + vector(rk1Bfu_z, i, j + 2, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j + 2, k) * vector(rk1gtd_zz, i, j + 2, k) * vector(rk1gtd_xx, i, j + 2, k) + (-vector(rk1gtd_yy, i, j + 2, k) * (vector(rk1gtd_xz, i, j + 2, k) * vector(rk1gtd_xz, i, j + 2, k))) + (-vector(rk1gtd_zz, i, j + 2, k) * (vector(rk1gtd_xy, i, j + 2, k) * vector(rk1gtd_xy, i, j + 2, k))) + (-(vector(rk1gtd_yz, i, j + 2, k) * vector(rk1gtd_yz, i, j + 2, k)) * vector(rk1gtd_xx, i, j + 2, k)) + 2.0 * vector(rk1gtd_yz, i, j + 2, k) * vector(rk1gtd_xy, i, j + 2, k) * vector(rk1gtd_xz, i, j + 2, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j + 2, k) * vector(rk1gtd_yy, i, j + 2, k) - vector(rk1gtd_xy, i, j + 2, k) * vector(rk1gtd_xy, i, j + 2, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j + 2, k) * vector(rk1gtd_zz, i, j + 2, k) - vector(rk1gtd_yz, i, j + 2, k) * vector(rk1gtd_yz, i, j + 2, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j + 2, k) * vector(rk1gtd_yz, i, j + 2, k) - vector(rk1gtd_yy, i, j + 2, k) * vector(rk1gtd_xz, i, j + 2, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j + 2, k) * vector(rk1gtd_zz, i, j + 2, k)) + vector(rk1gtd_yz, i, j + 2, k) * vector(rk1gtd_xz, i, j + 2, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j + 2, k) + rk1gtu_xy * vector(rk1Sfd_y, i, j + 2, k) + rk1gtu_xz * vector(rk1Sfd_z, i, j + 2, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j + 2, k) + rk1gtu_xy * vector(vfd_y, i, j + 2, k) + rk1gtu_xz * vector(vfd_z, i, j + 2, k));
						rk1Bfvf = vector(rk1Bfu_x, i, j + 2, k) * vector(vfd_x, i, j + 2, k) + vector(rk1Bfu_y, i, j + 2, k) * vector(vfd_y, i, j + 2, k) + vector(rk1Bfu_z, i, j + 2, k) * vector(vfd_z, i, j + 2, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j + 2, k) * vector(rk1gtd_zz, i, j + 2, k) - vector(rk1gtd_xz, i, j + 2, k) * vector(rk1gtd_xz, i, j + 2, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j + 2, k) * vector(rk1gtd_yz, i, j + 2, k)) + vector(rk1gtd_xy, i, j + 2, k) * vector(rk1gtd_xz, i, j + 2, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j + 2, k) + rk1gtu_yz * vector(vfd_y, i, j + 2, k) + rk1gtu_zz * vector(vfd_z, i, j + 2, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j + 2, k) + rk1gtu_yy * vector(rk1Sfd_y, i, j + 2, k) + rk1gtu_yz * vector(rk1Sfd_z, i, j + 2, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j + 2, k) + rk1gtu_yz * vector(rk1Sfd_y, i, j + 2, k) + rk1gtu_zz * vector(rk1Sfd_z, i, j + 2, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j + 2, k) + rk1gtu_yy * vector(vfd_y, i, j + 2, k) + rk1gtu_yz * vector(vfd_z, i, j + 2, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j + 2, k) + rk1vfu_y * vector(vfd_y, i, j + 2, k) + rk1vfu_z * vector(vfd_z, i, j + 2, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j + 2, k) * vector(rk1Bfu_z, i, j + 2, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j + 2, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j + 2, k) * vector(rk1Bfu_x, i, j + 2, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j + 2, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j + 2, k) * vector(rk1Bfu_z, i, j + 2, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j + 2, k) * rk1vfu_z + vector(rk1Bfu_z, i, j + 2, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j + 2, k) * vector(rk1Bfu_y, i, j + 2, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j + 2, k) * rk1vfu_y + vector(rk1Bfu_y, i, j + 2, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j + 2, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j + 2, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j + 2, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j + 2, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j + 2, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j + 2, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j + 2, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j + 2, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j + 2, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j + 2, k) * vector(rk1Bfu_z, i, j + 2, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j + 2, k) * rk1vfu_z + vector(rk1Bfu_z, i, j + 2, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j + 2, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j + 2, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j + 2, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j + 2, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j + 2, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j + 2, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j + 2, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j + 2, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j + 2, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j + 2, k) * vector(rk1Bfu_y, i, j + 2, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j + 2, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j + 2, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j + 2, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j + 2, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j + 2, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j + 2, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j + 2, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j + 2, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j + 2, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j + 2, k));
						Fluxp2jSfd_x = FjSfd_x_mainI(vector(rk1Sfd_x, i, j + 2, k), vector(rk1gtd_xx, i, j + 2, k), vector(rk1gtd_xy, i, j + 2, k), vector(rk1gtd_xz, i, j + 2, k), vector(rk1Betau_y, i, j + 2, k), vector(rk1Alpha, i, j + 2, k), vector(TauT_xy, i, j + 2, k), vector(TauT_yy, i, j + 2, k), vector(TauT_yz, i, j + 2, k), rk1inv_chi, rk1sdetg, rk1Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxp2jSfd_y = FjSfd_y_mainI(vector(rk1Sfd_y, i, j + 2, k), vector(rk1gtd_xy, i, j + 2, k), vector(rk1gtd_yy, i, j + 2, k), vector(rk1gtd_yz, i, j + 2, k), vector(rk1Betau_y, i, j + 2, k), vector(rk1Alpha, i, j + 2, k), vector(TauT_xy, i, j + 2, k), vector(TauT_yy, i, j + 2, k), vector(TauT_yz, i, j + 2, k), rk1inv_chi, rk1sdetg, rk1Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp2jSfd_z = FjSfd_z_mainI(vector(rk1Sfd_z, i, j + 2, k), vector(rk1gtd_xz, i, j + 2, k), vector(rk1gtd_yz, i, j + 2, k), vector(rk1gtd_zz, i, j + 2, k), vector(rk1Betau_y, i, j + 2, k), vector(rk1Alpha, i, j + 2, k), vector(TauT_xy, i, j + 2, k), vector(TauT_yy, i, j + 2, k), vector(TauT_yz, i, j + 2, k), rk1inv_chi, rk1sdetg, rk1Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxp2jBfu_x = FjBfu_x_mainI(vector(rk1Bfu_x, i, j + 2, k), vector(rk1Bfu_y, i, j + 2, k), vector(rk1phif, i, j + 2, k), vector(rk1Betau_x, i, j + 2, k), vector(rk1Betau_y, i, j + 2, k), vector(rk1Alpha, i, j + 2, k), vector(TauM_xy, i, j + 2, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2jBfu_y = FjBfu_y_mainI(vector(rk1phif, i, j + 2, k), vector(rk1Alpha, i, j + 2, k), rk1chi_max, rk1gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp2jBfu_z = FjBfu_z_mainI(vector(rk1Bfu_y, i, j + 2, k), vector(rk1Bfu_z, i, j + 2, k), vector(rk1phif, i, j + 2, k), vector(rk1Betau_y, i, j + 2, k), vector(rk1Betau_z, i, j + 2, k), vector(rk1Alpha, i, j + 2, k), vector(TauM_yz, i, j + 2, k), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2jDf = FjDf_mainI(vector(rk1Df, i, j + 2, k), vector(rk1Betau_y, i, j + 2, k), vector(rk1Alpha, i, j + 2, k), vector(TauN_y, i, j + 2, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2jDYf = FjDYf_mainI(vector(rk1DYf, i, j + 2, k), vector(rk1Betau_y, i, j + 2, k), vector(rk1Alpha, i, j + 2, k), vector(TauNe_y, i, j + 2, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2jtauf = Fjtauf_mainI(vector(rk1Df, i, j + 2, k), vector(rk1tauf, i, j + 2, k), vector(rk1Betau_y, i, j + 2, k), vector(rk1Alpha, i, j + 2, k), vector(TauN_y, i, j + 2, k), rk1sdetg, rk1Sfu_y, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2jphif = Fjphif_mainI(vector(rk1Bfu_y, i, j + 2, k), vector(rk1phif, i, j + 2, k), vector(rk1Betau_y, i, j + 2, k), vector(rk1Alpha, i, j + 2, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j + 2, k) * (1.0 + vector(epsf, i, j + 2, k)) + vector(pf, i, j + 2, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j + 2, k) + rk1sqca - vector(sqcs, i, j + 2, k) * rk1sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk1Betau_y, i, j + 2, k)) + vector(rk1Alpha, i, j + 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk1Betau_y, i, j + 2, k)) + vector(rk1Alpha, i, j + 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk1Betau_y, i, j + 2, k)) + vector(rk1Alpha, i, j + 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk1Betau_y, i, j + 2, k)) + vector(rk1Alpha, i, j + 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk1Betau_y, i, j + 2, k)) + vector(rk1Alpha, i, j + 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk1Betau_y, i, j + 2, k)) + vector(rk1Alpha, i, j + 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk1Betau_y, i, j + 2, k)) + vector(rk1Alpha, i, j + 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk1Betau_y, i, j + 2, k)) + vector(rk1Alpha, i, j + 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk1Betau_y, i, j + 2, k)) + vector(rk1Alpha, i, j + 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk1Betau_y, i, j + 2, k)) + vector(rk1Alpha, i, j + 2, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j + 3, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j + 3, k) * vector(rk1Bfu_x, i, j + 3, k) + vector(rk1gtd_yy, i, j + 3, k) * vector(rk1Bfu_y, i, j + 3, k) + vector(rk1gtd_yz, i, j + 3, k) * vector(rk1Bfu_z, i, j + 3, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j + 3, k) * vector(rk1Bfu_x, i, j + 3, k) + vector(rk1gtd_yz, i, j + 3, k) * vector(rk1Bfu_y, i, j + 3, k) + vector(rk1gtd_zz, i, j + 3, k) * vector(rk1Bfu_z, i, j + 3, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j + 3, k) * vector(rk1Bfu_x, i, j + 3, k) + vector(rk1gtd_xy, i, j + 3, k) * vector(rk1Bfu_y, i, j + 3, k) + vector(rk1gtd_xz, i, j + 3, k) * vector(rk1Bfu_z, i, j + 3, k));
						rk1sqBf = vector(rk1Bfu_x, i, j + 3, k) * rk1Bfd_x + vector(rk1Bfu_y, i, j + 3, k) * rk1Bfd_y + vector(rk1Bfu_z, i, j + 3, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j + 3, k) * vector(rk1gtd_zz, i, j + 3, k) * vector(rk1gtd_xx, i, j + 3, k) + (-vector(rk1gtd_yy, i, j + 3, k) * (vector(rk1gtd_xz, i, j + 3, k) * vector(rk1gtd_xz, i, j + 3, k))) + (-vector(rk1gtd_zz, i, j + 3, k) * (vector(rk1gtd_xy, i, j + 3, k) * vector(rk1gtd_xy, i, j + 3, k))) + (-(vector(rk1gtd_yz, i, j + 3, k) * vector(rk1gtd_yz, i, j + 3, k)) * vector(rk1gtd_xx, i, j + 3, k)) + 2.0 * vector(rk1gtd_yz, i, j + 3, k) * vector(rk1gtd_xy, i, j + 3, k) * vector(rk1gtd_xz, i, j + 3, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j + 3, k) * vector(rk1gtd_yy, i, j + 3, k) - vector(rk1gtd_xy, i, j + 3, k) * vector(rk1gtd_xy, i, j + 3, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j + 3, k) * vector(rk1gtd_zz, i, j + 3, k) - vector(rk1gtd_yz, i, j + 3, k) * vector(rk1gtd_yz, i, j + 3, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j + 3, k) * vector(rk1gtd_yz, i, j + 3, k) - vector(rk1gtd_yy, i, j + 3, k) * vector(rk1gtd_xz, i, j + 3, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j + 3, k) * vector(rk1gtd_zz, i, j + 3, k)) + vector(rk1gtd_yz, i, j + 3, k) * vector(rk1gtd_xz, i, j + 3, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j + 3, k) + rk1gtu_xy * vector(rk1Sfd_y, i, j + 3, k) + rk1gtu_xz * vector(rk1Sfd_z, i, j + 3, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j + 3, k) + rk1gtu_xy * vector(vfd_y, i, j + 3, k) + rk1gtu_xz * vector(vfd_z, i, j + 3, k));
						rk1Bfvf = vector(rk1Bfu_x, i, j + 3, k) * vector(vfd_x, i, j + 3, k) + vector(rk1Bfu_y, i, j + 3, k) * vector(vfd_y, i, j + 3, k) + vector(rk1Bfu_z, i, j + 3, k) * vector(vfd_z, i, j + 3, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j + 3, k) * vector(rk1gtd_zz, i, j + 3, k) - vector(rk1gtd_xz, i, j + 3, k) * vector(rk1gtd_xz, i, j + 3, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j + 3, k) * vector(rk1gtd_yz, i, j + 3, k)) + vector(rk1gtd_xy, i, j + 3, k) * vector(rk1gtd_xz, i, j + 3, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j + 3, k) + rk1gtu_yz * vector(vfd_y, i, j + 3, k) + rk1gtu_zz * vector(vfd_z, i, j + 3, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j + 3, k) + rk1gtu_yy * vector(rk1Sfd_y, i, j + 3, k) + rk1gtu_yz * vector(rk1Sfd_z, i, j + 3, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j + 3, k) + rk1gtu_yz * vector(rk1Sfd_y, i, j + 3, k) + rk1gtu_zz * vector(rk1Sfd_z, i, j + 3, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j + 3, k) + rk1gtu_yy * vector(vfd_y, i, j + 3, k) + rk1gtu_yz * vector(vfd_z, i, j + 3, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j + 3, k) + rk1vfu_y * vector(vfd_y, i, j + 3, k) + rk1vfu_z * vector(vfd_z, i, j + 3, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j + 3, k) * vector(rk1Bfu_z, i, j + 3, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j + 3, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j + 3, k) * vector(rk1Bfu_x, i, j + 3, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j + 3, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j + 3, k) * vector(rk1Bfu_z, i, j + 3, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j + 3, k) * rk1vfu_z + vector(rk1Bfu_z, i, j + 3, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j + 3, k) * vector(rk1Bfu_y, i, j + 3, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j + 3, k) * rk1vfu_y + vector(rk1Bfu_y, i, j + 3, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j + 3, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j + 3, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j + 3, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j + 3, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j + 3, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j + 3, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j + 3, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j + 3, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j + 3, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j + 3, k) * vector(rk1Bfu_z, i, j + 3, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j + 3, k) * rk1vfu_z + vector(rk1Bfu_z, i, j + 3, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j + 3, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j + 3, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j + 3, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j + 3, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j + 3, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j + 3, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j + 3, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j + 3, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j + 3, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j + 3, k) * vector(rk1Bfu_y, i, j + 3, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j + 3, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j + 3, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j + 3, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j + 3, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j + 3, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j + 3, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j + 3, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j + 3, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j + 3, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j + 3, k));
						Fluxp3jSfd_x = FjSfd_x_mainI(vector(rk1Sfd_x, i, j + 3, k), vector(rk1gtd_xx, i, j + 3, k), vector(rk1gtd_xy, i, j + 3, k), vector(rk1gtd_xz, i, j + 3, k), vector(rk1Betau_y, i, j + 3, k), vector(rk1Alpha, i, j + 3, k), vector(TauT_xy, i, j + 3, k), vector(TauT_yy, i, j + 3, k), vector(TauT_yz, i, j + 3, k), rk1inv_chi, rk1sdetg, rk1Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxp3jSfd_y = FjSfd_y_mainI(vector(rk1Sfd_y, i, j + 3, k), vector(rk1gtd_xy, i, j + 3, k), vector(rk1gtd_yy, i, j + 3, k), vector(rk1gtd_yz, i, j + 3, k), vector(rk1Betau_y, i, j + 3, k), vector(rk1Alpha, i, j + 3, k), vector(TauT_xy, i, j + 3, k), vector(TauT_yy, i, j + 3, k), vector(TauT_yz, i, j + 3, k), rk1inv_chi, rk1sdetg, rk1Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp3jSfd_z = FjSfd_z_mainI(vector(rk1Sfd_z, i, j + 3, k), vector(rk1gtd_xz, i, j + 3, k), vector(rk1gtd_yz, i, j + 3, k), vector(rk1gtd_zz, i, j + 3, k), vector(rk1Betau_y, i, j + 3, k), vector(rk1Alpha, i, j + 3, k), vector(TauT_xy, i, j + 3, k), vector(TauT_yy, i, j + 3, k), vector(TauT_yz, i, j + 3, k), rk1inv_chi, rk1sdetg, rk1Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxp3jBfu_x = FjBfu_x_mainI(vector(rk1Bfu_x, i, j + 3, k), vector(rk1Bfu_y, i, j + 3, k), vector(rk1phif, i, j + 3, k), vector(rk1Betau_x, i, j + 3, k), vector(rk1Betau_y, i, j + 3, k), vector(rk1Alpha, i, j + 3, k), vector(TauM_xy, i, j + 3, k), rk1chi_max, rk1gtu_xy, rk1sdetg, rk1vfu_x, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3jBfu_y = FjBfu_y_mainI(vector(rk1phif, i, j + 3, k), vector(rk1Alpha, i, j + 3, k), rk1chi_max, rk1gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp3jBfu_z = FjBfu_z_mainI(vector(rk1Bfu_y, i, j + 3, k), vector(rk1Bfu_z, i, j + 3, k), vector(rk1phif, i, j + 3, k), vector(rk1Betau_y, i, j + 3, k), vector(rk1Betau_z, i, j + 3, k), vector(rk1Alpha, i, j + 3, k), vector(TauM_yz, i, j + 3, k), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3jDf = FjDf_mainI(vector(rk1Df, i, j + 3, k), vector(rk1Betau_y, i, j + 3, k), vector(rk1Alpha, i, j + 3, k), vector(TauN_y, i, j + 3, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3jDYf = FjDYf_mainI(vector(rk1DYf, i, j + 3, k), vector(rk1Betau_y, i, j + 3, k), vector(rk1Alpha, i, j + 3, k), vector(TauNe_y, i, j + 3, k), rk1sdetg, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3jtauf = Fjtauf_mainI(vector(rk1Df, i, j + 3, k), vector(rk1tauf, i, j + 3, k), vector(rk1Betau_y, i, j + 3, k), vector(rk1Alpha, i, j + 3, k), vector(TauN_y, i, j + 3, k), rk1sdetg, rk1Sfu_y, rk1vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3jphif = Fjphif_mainI(vector(rk1Bfu_y, i, j + 3, k), vector(rk1phif, i, j + 3, k), vector(rk1Betau_y, i, j + 3, k), vector(rk1Alpha, i, j + 3, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j + 3, k) * (1.0 + vector(epsf, i, j + 3, k)) + vector(pf, i, j + 3, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j + 3, k) + rk1sqca - vector(sqcs, i, j + 3, k) * rk1sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk1Betau_y, i, j + 3, k)) + vector(rk1Alpha, i, j + 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk1Betau_y, i, j + 3, k)) + vector(rk1Alpha, i, j + 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk1Betau_y, i, j + 3, k)) + vector(rk1Alpha, i, j + 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk1Betau_y, i, j + 3, k)) + vector(rk1Alpha, i, j + 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk1Betau_y, i, j + 3, k)) + vector(rk1Alpha, i, j + 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk1Betau_y, i, j + 3, k)) + vector(rk1Alpha, i, j + 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk1Betau_y, i, j + 3, k)) + vector(rk1Alpha, i, j + 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk1Betau_y, i, j + 3, k)) + vector(rk1Alpha, i, j + 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk1Betau_y, i, j + 3, k)) + vector(rk1Alpha, i, j + 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk1Betau_y, i, j + 3, k)) + vector(rk1Alpha, i, j + 3, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_y) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_yy - (1.0 - rk1sqa) * (rk1vfu_y * rk1vfu_y)))))));
						LLFpjSfd_x = 0.5 * (FluxjSfd_x + fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j, k));
						LLFpjSfd_y = 0.5 * (FluxjSfd_y + fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j, k));
						LLFpjSfd_z = 0.5 * (FluxjSfd_z + fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j, k));
						LLFpjBfu_x = 0.5 * (FluxjBfu_x + fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j, k));
						LLFpjBfu_y = 0.5 * (FluxjBfu_y + fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j, k));
						LLFpjBfu_z = 0.5 * (FluxjBfu_z + fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j, k));
						LLFpjDf = 0.5 * (FluxjDf + fabs(SpeedjDf) * vector(rk1Df, i, j, k));
						LLFpjDYf = 0.5 * (FluxjDYf + fabs(SpeedjDYf) * vector(rk1DYf, i, j, k));
						LLFpjtauf = 0.5 * (Fluxjtauf + fabs(Speedjtauf) * vector(rk1tauf, i, j, k));
						LLFpjphif = 0.5 * (Fluxjphif + fabs(Speedjphif) * vector(rk1phif, i, j, k));
						LLFnjSfd_x = 0.5 * (FluxjSfd_x - fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j, k));
						LLFnjSfd_y = 0.5 * (FluxjSfd_y - fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j, k));
						LLFnjSfd_z = 0.5 * (FluxjSfd_z - fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j, k));
						LLFnjBfu_x = 0.5 * (FluxjBfu_x - fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j, k));
						LLFnjBfu_y = 0.5 * (FluxjBfu_y - fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j, k));
						LLFnjBfu_z = 0.5 * (FluxjBfu_z - fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j, k));
						LLFnjDf = 0.5 * (FluxjDf - fabs(SpeedjDf) * vector(rk1Df, i, j, k));
						LLFnjDYf = 0.5 * (FluxjDYf - fabs(SpeedjDYf) * vector(rk1DYf, i, j, k));
						LLFnjtauf = 0.5 * (Fluxjtauf - fabs(Speedjtauf) * vector(rk1tauf, i, j, k));
						LLFnjphif = 0.5 * (Fluxjphif - fabs(Speedjphif) * vector(rk1phif, i, j, k));
						LLFp_m3jSfd_x = 0.5 * (Fluxm3jSfd_x + fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j - 3, k));
						LLFp_m3jSfd_y = 0.5 * (Fluxm3jSfd_y + fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j - 3, k));
						LLFp_m3jSfd_z = 0.5 * (Fluxm3jSfd_z + fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j - 3, k));
						LLFp_m3jBfu_x = 0.5 * (Fluxm3jBfu_x + fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j - 3, k));
						LLFp_m3jBfu_y = 0.5 * (Fluxm3jBfu_y + fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j - 3, k));
						LLFp_m3jBfu_z = 0.5 * (Fluxm3jBfu_z + fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j - 3, k));
						LLFp_m3jDf = 0.5 * (Fluxm3jDf + fabs(SpeedjDf) * vector(rk1Df, i, j - 3, k));
						LLFp_m3jDYf = 0.5 * (Fluxm3jDYf + fabs(SpeedjDYf) * vector(rk1DYf, i, j - 3, k));
						LLFp_m3jtauf = 0.5 * (Fluxm3jtauf + fabs(Speedjtauf) * vector(rk1tauf, i, j - 3, k));
						LLFp_m3jphif = 0.5 * (Fluxm3jphif + fabs(Speedjphif) * vector(rk1phif, i, j - 3, k));
						LLFn_m3jSfd_x = 0.5 * (Fluxm3jSfd_x - fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j - 3, k));
						LLFn_m3jSfd_y = 0.5 * (Fluxm3jSfd_y - fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j - 3, k));
						LLFn_m3jSfd_z = 0.5 * (Fluxm3jSfd_z - fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j - 3, k));
						LLFn_m3jBfu_x = 0.5 * (Fluxm3jBfu_x - fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j - 3, k));
						LLFn_m3jBfu_y = 0.5 * (Fluxm3jBfu_y - fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j - 3, k));
						LLFn_m3jBfu_z = 0.5 * (Fluxm3jBfu_z - fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j - 3, k));
						LLFn_m3jDf = 0.5 * (Fluxm3jDf - fabs(SpeedjDf) * vector(rk1Df, i, j - 3, k));
						LLFn_m3jDYf = 0.5 * (Fluxm3jDYf - fabs(SpeedjDYf) * vector(rk1DYf, i, j - 3, k));
						LLFn_m3jtauf = 0.5 * (Fluxm3jtauf - fabs(Speedjtauf) * vector(rk1tauf, i, j - 3, k));
						LLFn_m3jphif = 0.5 * (Fluxm3jphif - fabs(Speedjphif) * vector(rk1phif, i, j - 3, k));
						LLFp_m2jSfd_x = 0.5 * (Fluxm2jSfd_x + fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j - 2, k));
						LLFp_m2jSfd_y = 0.5 * (Fluxm2jSfd_y + fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j - 2, k));
						LLFp_m2jSfd_z = 0.5 * (Fluxm2jSfd_z + fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j - 2, k));
						LLFp_m2jBfu_x = 0.5 * (Fluxm2jBfu_x + fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j - 2, k));
						LLFp_m2jBfu_y = 0.5 * (Fluxm2jBfu_y + fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j - 2, k));
						LLFp_m2jBfu_z = 0.5 * (Fluxm2jBfu_z + fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j - 2, k));
						LLFp_m2jDf = 0.5 * (Fluxm2jDf + fabs(SpeedjDf) * vector(rk1Df, i, j - 2, k));
						LLFp_m2jDYf = 0.5 * (Fluxm2jDYf + fabs(SpeedjDYf) * vector(rk1DYf, i, j - 2, k));
						LLFp_m2jtauf = 0.5 * (Fluxm2jtauf + fabs(Speedjtauf) * vector(rk1tauf, i, j - 2, k));
						LLFp_m2jphif = 0.5 * (Fluxm2jphif + fabs(Speedjphif) * vector(rk1phif, i, j - 2, k));
						LLFn_m2jSfd_x = 0.5 * (Fluxm2jSfd_x - fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j - 2, k));
						LLFn_m2jSfd_y = 0.5 * (Fluxm2jSfd_y - fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j - 2, k));
						LLFn_m2jSfd_z = 0.5 * (Fluxm2jSfd_z - fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j - 2, k));
						LLFn_m2jBfu_x = 0.5 * (Fluxm2jBfu_x - fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j - 2, k));
						LLFn_m2jBfu_y = 0.5 * (Fluxm2jBfu_y - fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j - 2, k));
						LLFn_m2jBfu_z = 0.5 * (Fluxm2jBfu_z - fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j - 2, k));
						LLFn_m2jDf = 0.5 * (Fluxm2jDf - fabs(SpeedjDf) * vector(rk1Df, i, j - 2, k));
						LLFn_m2jDYf = 0.5 * (Fluxm2jDYf - fabs(SpeedjDYf) * vector(rk1DYf, i, j - 2, k));
						LLFn_m2jtauf = 0.5 * (Fluxm2jtauf - fabs(Speedjtauf) * vector(rk1tauf, i, j - 2, k));
						LLFn_m2jphif = 0.5 * (Fluxm2jphif - fabs(Speedjphif) * vector(rk1phif, i, j - 2, k));
						LLFp_m1jSfd_x = 0.5 * (Fluxm1jSfd_x + fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j - 1, k));
						LLFp_m1jSfd_y = 0.5 * (Fluxm1jSfd_y + fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j - 1, k));
						LLFp_m1jSfd_z = 0.5 * (Fluxm1jSfd_z + fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j - 1, k));
						LLFp_m1jBfu_x = 0.5 * (Fluxm1jBfu_x + fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j - 1, k));
						LLFp_m1jBfu_y = 0.5 * (Fluxm1jBfu_y + fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j - 1, k));
						LLFp_m1jBfu_z = 0.5 * (Fluxm1jBfu_z + fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j - 1, k));
						LLFp_m1jDf = 0.5 * (Fluxm1jDf + fabs(SpeedjDf) * vector(rk1Df, i, j - 1, k));
						LLFp_m1jDYf = 0.5 * (Fluxm1jDYf + fabs(SpeedjDYf) * vector(rk1DYf, i, j - 1, k));
						LLFp_m1jtauf = 0.5 * (Fluxm1jtauf + fabs(Speedjtauf) * vector(rk1tauf, i, j - 1, k));
						LLFp_m1jphif = 0.5 * (Fluxm1jphif + fabs(Speedjphif) * vector(rk1phif, i, j - 1, k));
						LLFn_m1jSfd_x = 0.5 * (Fluxm1jSfd_x - fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j - 1, k));
						LLFn_m1jSfd_y = 0.5 * (Fluxm1jSfd_y - fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j - 1, k));
						LLFn_m1jSfd_z = 0.5 * (Fluxm1jSfd_z - fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j - 1, k));
						LLFn_m1jBfu_x = 0.5 * (Fluxm1jBfu_x - fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j - 1, k));
						LLFn_m1jBfu_y = 0.5 * (Fluxm1jBfu_y - fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j - 1, k));
						LLFn_m1jBfu_z = 0.5 * (Fluxm1jBfu_z - fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j - 1, k));
						LLFn_m1jDf = 0.5 * (Fluxm1jDf - fabs(SpeedjDf) * vector(rk1Df, i, j - 1, k));
						LLFn_m1jDYf = 0.5 * (Fluxm1jDYf - fabs(SpeedjDYf) * vector(rk1DYf, i, j - 1, k));
						LLFn_m1jtauf = 0.5 * (Fluxm1jtauf - fabs(Speedjtauf) * vector(rk1tauf, i, j - 1, k));
						LLFn_m1jphif = 0.5 * (Fluxm1jphif - fabs(Speedjphif) * vector(rk1phif, i, j - 1, k));
						LLFp_p1jSfd_x = 0.5 * (Fluxp1jSfd_x + fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j + 1, k));
						LLFp_p1jSfd_y = 0.5 * (Fluxp1jSfd_y + fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j + 1, k));
						LLFp_p1jSfd_z = 0.5 * (Fluxp1jSfd_z + fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j + 1, k));
						LLFp_p1jBfu_x = 0.5 * (Fluxp1jBfu_x + fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j + 1, k));
						LLFp_p1jBfu_y = 0.5 * (Fluxp1jBfu_y + fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j + 1, k));
						LLFp_p1jBfu_z = 0.5 * (Fluxp1jBfu_z + fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j + 1, k));
						LLFp_p1jDf = 0.5 * (Fluxp1jDf + fabs(SpeedjDf) * vector(rk1Df, i, j + 1, k));
						LLFp_p1jDYf = 0.5 * (Fluxp1jDYf + fabs(SpeedjDYf) * vector(rk1DYf, i, j + 1, k));
						LLFp_p1jtauf = 0.5 * (Fluxp1jtauf + fabs(Speedjtauf) * vector(rk1tauf, i, j + 1, k));
						LLFp_p1jphif = 0.5 * (Fluxp1jphif + fabs(Speedjphif) * vector(rk1phif, i, j + 1, k));
						LLFn_p1jSfd_x = 0.5 * (Fluxp1jSfd_x - fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j + 1, k));
						LLFn_p1jSfd_y = 0.5 * (Fluxp1jSfd_y - fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j + 1, k));
						LLFn_p1jSfd_z = 0.5 * (Fluxp1jSfd_z - fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j + 1, k));
						LLFn_p1jBfu_x = 0.5 * (Fluxp1jBfu_x - fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j + 1, k));
						LLFn_p1jBfu_y = 0.5 * (Fluxp1jBfu_y - fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j + 1, k));
						LLFn_p1jBfu_z = 0.5 * (Fluxp1jBfu_z - fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j + 1, k));
						LLFn_p1jDf = 0.5 * (Fluxp1jDf - fabs(SpeedjDf) * vector(rk1Df, i, j + 1, k));
						LLFn_p1jDYf = 0.5 * (Fluxp1jDYf - fabs(SpeedjDYf) * vector(rk1DYf, i, j + 1, k));
						LLFn_p1jtauf = 0.5 * (Fluxp1jtauf - fabs(Speedjtauf) * vector(rk1tauf, i, j + 1, k));
						LLFn_p1jphif = 0.5 * (Fluxp1jphif - fabs(Speedjphif) * vector(rk1phif, i, j + 1, k));
						LLFp_p2jSfd_x = 0.5 * (Fluxp2jSfd_x + fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j + 2, k));
						LLFp_p2jSfd_y = 0.5 * (Fluxp2jSfd_y + fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j + 2, k));
						LLFp_p2jSfd_z = 0.5 * (Fluxp2jSfd_z + fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j + 2, k));
						LLFp_p2jBfu_x = 0.5 * (Fluxp2jBfu_x + fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j + 2, k));
						LLFp_p2jBfu_y = 0.5 * (Fluxp2jBfu_y + fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j + 2, k));
						LLFp_p2jBfu_z = 0.5 * (Fluxp2jBfu_z + fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j + 2, k));
						LLFp_p2jDf = 0.5 * (Fluxp2jDf + fabs(SpeedjDf) * vector(rk1Df, i, j + 2, k));
						LLFp_p2jDYf = 0.5 * (Fluxp2jDYf + fabs(SpeedjDYf) * vector(rk1DYf, i, j + 2, k));
						LLFp_p2jtauf = 0.5 * (Fluxp2jtauf + fabs(Speedjtauf) * vector(rk1tauf, i, j + 2, k));
						LLFp_p2jphif = 0.5 * (Fluxp2jphif + fabs(Speedjphif) * vector(rk1phif, i, j + 2, k));
						LLFn_p2jSfd_x = 0.5 * (Fluxp2jSfd_x - fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j + 2, k));
						LLFn_p2jSfd_y = 0.5 * (Fluxp2jSfd_y - fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j + 2, k));
						LLFn_p2jSfd_z = 0.5 * (Fluxp2jSfd_z - fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j + 2, k));
						LLFn_p2jBfu_x = 0.5 * (Fluxp2jBfu_x - fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j + 2, k));
						LLFn_p2jBfu_y = 0.5 * (Fluxp2jBfu_y - fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j + 2, k));
						LLFn_p2jBfu_z = 0.5 * (Fluxp2jBfu_z - fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j + 2, k));
						LLFn_p2jDf = 0.5 * (Fluxp2jDf - fabs(SpeedjDf) * vector(rk1Df, i, j + 2, k));
						LLFn_p2jDYf = 0.5 * (Fluxp2jDYf - fabs(SpeedjDYf) * vector(rk1DYf, i, j + 2, k));
						LLFn_p2jtauf = 0.5 * (Fluxp2jtauf - fabs(Speedjtauf) * vector(rk1tauf, i, j + 2, k));
						LLFn_p2jphif = 0.5 * (Fluxp2jphif - fabs(Speedjphif) * vector(rk1phif, i, j + 2, k));
						LLFp_p3jSfd_x = 0.5 * (Fluxp3jSfd_x + fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j + 3, k));
						LLFp_p3jSfd_y = 0.5 * (Fluxp3jSfd_y + fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j + 3, k));
						LLFp_p3jSfd_z = 0.5 * (Fluxp3jSfd_z + fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j + 3, k));
						LLFp_p3jBfu_x = 0.5 * (Fluxp3jBfu_x + fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j + 3, k));
						LLFp_p3jBfu_y = 0.5 * (Fluxp3jBfu_y + fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j + 3, k));
						LLFp_p3jBfu_z = 0.5 * (Fluxp3jBfu_z + fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j + 3, k));
						LLFp_p3jDf = 0.5 * (Fluxp3jDf + fabs(SpeedjDf) * vector(rk1Df, i, j + 3, k));
						LLFp_p3jDYf = 0.5 * (Fluxp3jDYf + fabs(SpeedjDYf) * vector(rk1DYf, i, j + 3, k));
						LLFp_p3jtauf = 0.5 * (Fluxp3jtauf + fabs(Speedjtauf) * vector(rk1tauf, i, j + 3, k));
						LLFp_p3jphif = 0.5 * (Fluxp3jphif + fabs(Speedjphif) * vector(rk1phif, i, j + 3, k));
						LLFn_p3jSfd_x = 0.5 * (Fluxp3jSfd_x - fabs(SpeedjSfd_x) * vector(rk1Sfd_x, i, j + 3, k));
						LLFn_p3jSfd_y = 0.5 * (Fluxp3jSfd_y - fabs(SpeedjSfd_y) * vector(rk1Sfd_y, i, j + 3, k));
						LLFn_p3jSfd_z = 0.5 * (Fluxp3jSfd_z - fabs(SpeedjSfd_z) * vector(rk1Sfd_z, i, j + 3, k));
						LLFn_p3jBfu_x = 0.5 * (Fluxp3jBfu_x - fabs(SpeedjBfu_x) * vector(rk1Bfu_x, i, j + 3, k));
						LLFn_p3jBfu_y = 0.5 * (Fluxp3jBfu_y - fabs(SpeedjBfu_y) * vector(rk1Bfu_y, i, j + 3, k));
						LLFn_p3jBfu_z = 0.5 * (Fluxp3jBfu_z - fabs(SpeedjBfu_z) * vector(rk1Bfu_z, i, j + 3, k));
						LLFn_p3jDf = 0.5 * (Fluxp3jDf - fabs(SpeedjDf) * vector(rk1Df, i, j + 3, k));
						LLFn_p3jDYf = 0.5 * (Fluxp3jDYf - fabs(SpeedjDYf) * vector(rk1DYf, i, j + 3, k));
						LLFn_p3jtauf = 0.5 * (Fluxp3jtauf - fabs(Speedjtauf) * vector(rk1tauf, i, j + 3, k));
						LLFn_p3jphif = 0.5 * (Fluxp3jphif - fabs(Speedjphif) * vector(rk1phif, i, j + 3, k));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j, k));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_yy, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_zz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_xy, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_xz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1sqBf = vector(rk1Bfu_x, i, j, k) * rk1Bfd_x + vector(rk1Bfu_y, i, j, k) * rk1Bfd_y + vector(rk1Bfu_z, i, j, k) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) * vector(rk1gtd_xx, i, j, k) + (-vector(rk1gtd_yy, i, j, k) * (vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k))) + (-vector(rk1gtd_zz, i, j, k) * (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k))) + (-(vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k)) * vector(rk1gtd_xx, i, j, k)) + 2.0 * vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yy, i, j, k) - vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) - vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_yz, i, j, k) - vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_zz, i, j, k)) + vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j, k) + rk1gtu_xy * vector(rk1Sfd_y, i, j, k) + rk1gtu_xz * vector(rk1Sfd_z, i, j, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j, k) + rk1gtu_xy * vector(vfd_y, i, j, k) + rk1gtu_xz * vector(vfd_z, i, j, k));
						rk1Bfvf = vector(rk1Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk1Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk1Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_zz, i, j, k) - vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yz, i, j, k)) + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j, k) + rk1gtu_yz * vector(vfd_y, i, j, k) + rk1gtu_zz * vector(vfd_z, i, j, k));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j, k) + rk1gtu_yy * vector(rk1Sfd_y, i, j, k) + rk1gtu_yz * vector(rk1Sfd_z, i, j, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j, k) + rk1gtu_yz * vector(rk1Sfd_y, i, j, k) + rk1gtu_zz * vector(rk1Sfd_z, i, j, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j, k) + rk1gtu_yy * vector(vfd_y, i, j, k) + rk1gtu_yz * vector(vfd_z, i, j, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j, k) + rk1vfu_y * vector(vfd_y, i, j, k) + rk1vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j, k) * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j, k) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j, k) * vector(rk1Bfu_x, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k) * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k) * rk1vfu_z + vector(rk1Bfu_z, i, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k) * vector(rk1Bfu_y, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k) * rk1vfu_y + vector(rk1Bfu_y, i, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j, k) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j, k));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j, k) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j, k) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j, k));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k) * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k) * rk1vfu_z + vector(rk1Bfu_z, i, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j, k) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j, k));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j, k));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j, k));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j, k) * vector(rk1Bfu_y, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j, k) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j, k));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j, k) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j, k));
						FluxkSfd_x = FkSfd_x_mainI(vector(rk1Sfd_x, i, j, k), vector(rk1gtd_xx, i, j, k), vector(rk1gtd_xy, i, j, k), vector(rk1gtd_xz, i, j, k), vector(rk1Betau_z, i, j, k), vector(rk1Alpha, i, j, k), vector(TauT_xz, i, j, k), vector(TauT_yz, i, j, k), vector(TauT_zz, i, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_zx, dx, simPlat_dt, ilast, jlast);
						FluxkSfd_y = FkSfd_y_mainI(vector(rk1Sfd_y, i, j, k), vector(rk1gtd_xy, i, j, k), vector(rk1gtd_yy, i, j, k), vector(rk1gtd_yz, i, j, k), vector(rk1Betau_z, i, j, k), vector(rk1Alpha, i, j, k), vector(TauT_xz, i, j, k), vector(TauT_yz, i, j, k), vector(TauT_zz, i, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_zy, dx, simPlat_dt, ilast, jlast);
						FluxkSfd_z = FkSfd_z_mainI(vector(rk1Sfd_z, i, j, k), vector(rk1gtd_xz, i, j, k), vector(rk1gtd_yz, i, j, k), vector(rk1gtd_zz, i, j, k), vector(rk1Betau_z, i, j, k), vector(rk1Alpha, i, j, k), vector(TauT_xz, i, j, k), vector(TauT_yz, i, j, k), vector(TauT_zz, i, j, k), rk1inv_chi, rk1sdetg, rk1Sfud_zz, dx, simPlat_dt, ilast, jlast);
						FluxkBfu_x = FkBfu_x_mainI(vector(rk1Bfu_x, i, j, k), vector(rk1Bfu_z, i, j, k), vector(rk1phif, i, j, k), vector(rk1Betau_x, i, j, k), vector(rk1Betau_z, i, j, k), vector(rk1Alpha, i, j, k), vector(TauM_xz, i, j, k), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxkBfu_y = FkBfu_y_mainI(vector(rk1Bfu_y, i, j, k), vector(rk1Bfu_z, i, j, k), vector(rk1phif, i, j, k), vector(rk1Betau_y, i, j, k), vector(rk1Betau_z, i, j, k), vector(rk1Alpha, i, j, k), vector(TauM_yz, i, j, k), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxkBfu_z = FkBfu_z_mainI(vector(rk1phif, i, j, k), vector(rk1Alpha, i, j, k), rk1chi_max, rk1gtu_zz, dx, simPlat_dt, ilast, jlast);
						FluxkDf = FkDf_mainI(vector(rk1Df, i, j, k), vector(rk1Betau_z, i, j, k), vector(rk1Alpha, i, j, k), vector(TauN_z, i, j, k), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxkDYf = FkDYf_mainI(vector(rk1DYf, i, j, k), vector(rk1Betau_z, i, j, k), vector(rk1Alpha, i, j, k), vector(TauNe_z, i, j, k), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxktauf = Fktauf_mainI(vector(rk1Df, i, j, k), vector(rk1tauf, i, j, k), vector(rk1Betau_z, i, j, k), vector(rk1Alpha, i, j, k), vector(TauN_z, i, j, k), rk1sdetg, rk1Sfu_z, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxkphif = Fkphif_mainI(vector(rk1Bfu_z, i, j, k), vector(rk1phif, i, j, k), vector(rk1Betau_z, i, j, k), vector(rk1Alpha, i, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j, k) + rk1sqca - vector(sqcs, i, j, k) * rk1sqca;
						SpeedkSfd_x = fabs(fabs(vector(rk1Betau_z, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z))))));
						SpeedkSfd_y = fabs(fabs(vector(rk1Betau_z, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z))))));
						SpeedkSfd_z = fabs(fabs(vector(rk1Betau_z, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z))))));
						SpeedkBfu_x = fabs(fabs(vector(rk1Betau_z, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z))))));
						SpeedkBfu_y = fabs(fabs(vector(rk1Betau_z, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z))))));
						SpeedkBfu_z = fabs(fabs(vector(rk1Betau_z, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z))))));
						SpeedkDf = fabs(fabs(vector(rk1Betau_z, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z))))));
						SpeedkDYf = fabs(fabs(vector(rk1Betau_z, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z))))));
						Speedktauf = fabs(fabs(vector(rk1Betau_z, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z))))));
						Speedkphif = fabs(fabs(vector(rk1Betau_z, i, j, k)) + vector(rk1Alpha, i, j, k) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j, k - 3));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k - 3) * vector(rk1Bfu_x, i, j, k - 3) + vector(rk1gtd_yy, i, j, k - 3) * vector(rk1Bfu_y, i, j, k - 3) + vector(rk1gtd_yz, i, j, k - 3) * vector(rk1Bfu_z, i, j, k - 3));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k - 3) * vector(rk1Bfu_x, i, j, k - 3) + vector(rk1gtd_yz, i, j, k - 3) * vector(rk1Bfu_y, i, j, k - 3) + vector(rk1gtd_zz, i, j, k - 3) * vector(rk1Bfu_z, i, j, k - 3));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k - 3) * vector(rk1Bfu_x, i, j, k - 3) + vector(rk1gtd_xy, i, j, k - 3) * vector(rk1Bfu_y, i, j, k - 3) + vector(rk1gtd_xz, i, j, k - 3) * vector(rk1Bfu_z, i, j, k - 3));
						rk1sqBf = vector(rk1Bfu_x, i, j, k - 3) * rk1Bfd_x + vector(rk1Bfu_y, i, j, k - 3) * rk1Bfd_y + vector(rk1Bfu_z, i, j, k - 3) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j, k - 3) * vector(rk1gtd_zz, i, j, k - 3) * vector(rk1gtd_xx, i, j, k - 3) + (-vector(rk1gtd_yy, i, j, k - 3) * (vector(rk1gtd_xz, i, j, k - 3) * vector(rk1gtd_xz, i, j, k - 3))) + (-vector(rk1gtd_zz, i, j, k - 3) * (vector(rk1gtd_xy, i, j, k - 3) * vector(rk1gtd_xy, i, j, k - 3))) + (-(vector(rk1gtd_yz, i, j, k - 3) * vector(rk1gtd_yz, i, j, k - 3)) * vector(rk1gtd_xx, i, j, k - 3)) + 2.0 * vector(rk1gtd_yz, i, j, k - 3) * vector(rk1gtd_xy, i, j, k - 3) * vector(rk1gtd_xz, i, j, k - 3);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j, k - 3) * vector(rk1gtd_yy, i, j, k - 3) - vector(rk1gtd_xy, i, j, k - 3) * vector(rk1gtd_xy, i, j, k - 3));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j, k - 3) * vector(rk1gtd_zz, i, j, k - 3) - vector(rk1gtd_yz, i, j, k - 3) * vector(rk1gtd_yz, i, j, k - 3));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j, k - 3) * vector(rk1gtd_yz, i, j, k - 3) - vector(rk1gtd_yy, i, j, k - 3) * vector(rk1gtd_xz, i, j, k - 3));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j, k - 3) * vector(rk1gtd_zz, i, j, k - 3)) + vector(rk1gtd_yz, i, j, k - 3) * vector(rk1gtd_xz, i, j, k - 3));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j, k - 3) + rk1gtu_xy * vector(rk1Sfd_y, i, j, k - 3) + rk1gtu_xz * vector(rk1Sfd_z, i, j, k - 3));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j, k - 3) + rk1gtu_xy * vector(vfd_y, i, j, k - 3) + rk1gtu_xz * vector(vfd_z, i, j, k - 3));
						rk1Bfvf = vector(rk1Bfu_x, i, j, k - 3) * vector(vfd_x, i, j, k - 3) + vector(rk1Bfu_y, i, j, k - 3) * vector(vfd_y, i, j, k - 3) + vector(rk1Bfu_z, i, j, k - 3) * vector(vfd_z, i, j, k - 3);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j, k - 3) * vector(rk1gtd_zz, i, j, k - 3) - vector(rk1gtd_xz, i, j, k - 3) * vector(rk1gtd_xz, i, j, k - 3));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j, k - 3) * vector(rk1gtd_yz, i, j, k - 3)) + vector(rk1gtd_xy, i, j, k - 3) * vector(rk1gtd_xz, i, j, k - 3));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j, k - 3) + rk1gtu_yz * vector(vfd_y, i, j, k - 3) + rk1gtu_zz * vector(vfd_z, i, j, k - 3));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j, k - 3) + rk1gtu_yy * vector(rk1Sfd_y, i, j, k - 3) + rk1gtu_yz * vector(rk1Sfd_z, i, j, k - 3));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j, k - 3) + rk1gtu_yz * vector(rk1Sfd_y, i, j, k - 3) + rk1gtu_zz * vector(rk1Sfd_z, i, j, k - 3));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j, k - 3) + rk1gtu_yy * vector(vfd_y, i, j, k - 3) + rk1gtu_yz * vector(vfd_z, i, j, k - 3));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j, k - 3) + rk1vfu_y * vector(vfd_y, i, j, k - 3) + rk1vfu_z * vector(vfd_z, i, j, k - 3), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j, k - 3) * vector(rk1Bfu_z, i, j, k - 3)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j, k - 3) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j, k - 3) * vector(rk1Bfu_x, i, j, k - 3)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j, k - 3) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k - 3) * vector(rk1Bfu_z, i, j, k - 3)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k - 3) * rk1vfu_z + vector(rk1Bfu_z, i, j, k - 3) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k - 3) * vector(rk1Bfu_y, i, j, k - 3)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k - 3) * rk1vfu_y + vector(rk1Bfu_y, i, j, k - 3) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j, k - 3) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k - 3) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k - 3));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j, k - 3) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j, k - 3) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j, k - 3));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j, k - 3) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j, k - 3) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j, k - 3));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k - 3) * vector(rk1Bfu_z, i, j, k - 3)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k - 3) * rk1vfu_z + vector(rk1Bfu_z, i, j, k - 3) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j, k - 3) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j, k - 3) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j, k - 3));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k - 3) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k - 3) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j, k - 3));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j, k - 3) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j, k - 3) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j, k - 3));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j, k - 3) * vector(rk1Bfu_y, i, j, k - 3)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j, k - 3) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j, k - 3) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j, k - 3) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j, k - 3));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k - 3) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j, k - 3) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k - 3));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j, k - 3) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j, k - 3) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j, k - 3));
						Fluxm3kSfd_x = FkSfd_x_mainI(vector(rk1Sfd_x, i, j, k - 3), vector(rk1gtd_xx, i, j, k - 3), vector(rk1gtd_xy, i, j, k - 3), vector(rk1gtd_xz, i, j, k - 3), vector(rk1Betau_z, i, j, k - 3), vector(rk1Alpha, i, j, k - 3), vector(TauT_xz, i, j, k - 3), vector(TauT_yz, i, j, k - 3), vector(TauT_zz, i, j, k - 3), rk1inv_chi, rk1sdetg, rk1Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxm3kSfd_y = FkSfd_y_mainI(vector(rk1Sfd_y, i, j, k - 3), vector(rk1gtd_xy, i, j, k - 3), vector(rk1gtd_yy, i, j, k - 3), vector(rk1gtd_yz, i, j, k - 3), vector(rk1Betau_z, i, j, k - 3), vector(rk1Alpha, i, j, k - 3), vector(TauT_xz, i, j, k - 3), vector(TauT_yz, i, j, k - 3), vector(TauT_zz, i, j, k - 3), rk1inv_chi, rk1sdetg, rk1Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxm3kSfd_z = FkSfd_z_mainI(vector(rk1Sfd_z, i, j, k - 3), vector(rk1gtd_xz, i, j, k - 3), vector(rk1gtd_yz, i, j, k - 3), vector(rk1gtd_zz, i, j, k - 3), vector(rk1Betau_z, i, j, k - 3), vector(rk1Alpha, i, j, k - 3), vector(TauT_xz, i, j, k - 3), vector(TauT_yz, i, j, k - 3), vector(TauT_zz, i, j, k - 3), rk1inv_chi, rk1sdetg, rk1Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm3kBfu_x = FkBfu_x_mainI(vector(rk1Bfu_x, i, j, k - 3), vector(rk1Bfu_z, i, j, k - 3), vector(rk1phif, i, j, k - 3), vector(rk1Betau_x, i, j, k - 3), vector(rk1Betau_z, i, j, k - 3), vector(rk1Alpha, i, j, k - 3), vector(TauM_xz, i, j, k - 3), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3kBfu_y = FkBfu_y_mainI(vector(rk1Bfu_y, i, j, k - 3), vector(rk1Bfu_z, i, j, k - 3), vector(rk1phif, i, j, k - 3), vector(rk1Betau_y, i, j, k - 3), vector(rk1Betau_z, i, j, k - 3), vector(rk1Alpha, i, j, k - 3), vector(TauM_yz, i, j, k - 3), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3kBfu_z = FkBfu_z_mainI(vector(rk1phif, i, j, k - 3), vector(rk1Alpha, i, j, k - 3), rk1chi_max, rk1gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm3kDf = FkDf_mainI(vector(rk1Df, i, j, k - 3), vector(rk1Betau_z, i, j, k - 3), vector(rk1Alpha, i, j, k - 3), vector(TauN_z, i, j, k - 3), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3kDYf = FkDYf_mainI(vector(rk1DYf, i, j, k - 3), vector(rk1Betau_z, i, j, k - 3), vector(rk1Alpha, i, j, k - 3), vector(TauNe_z, i, j, k - 3), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3ktauf = Fktauf_mainI(vector(rk1Df, i, j, k - 3), vector(rk1tauf, i, j, k - 3), vector(rk1Betau_z, i, j, k - 3), vector(rk1Alpha, i, j, k - 3), vector(TauN_z, i, j, k - 3), rk1sdetg, rk1Sfu_z, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3kphif = Fkphif_mainI(vector(rk1Bfu_z, i, j, k - 3), vector(rk1phif, i, j, k - 3), vector(rk1Betau_z, i, j, k - 3), vector(rk1Alpha, i, j, k - 3), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j, k - 3) * (1.0 + vector(epsf, i, j, k - 3)) + vector(pf, i, j, k - 3);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j, k - 3) + rk1sqca - vector(sqcs, i, j, k - 3) * rk1sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk1Betau_z, i, j, k - 3)) + vector(rk1Alpha, i, j, k - 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk1Betau_z, i, j, k - 3)) + vector(rk1Alpha, i, j, k - 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk1Betau_z, i, j, k - 3)) + vector(rk1Alpha, i, j, k - 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk1Betau_z, i, j, k - 3)) + vector(rk1Alpha, i, j, k - 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk1Betau_z, i, j, k - 3)) + vector(rk1Alpha, i, j, k - 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk1Betau_z, i, j, k - 3)) + vector(rk1Alpha, i, j, k - 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk1Betau_z, i, j, k - 3)) + vector(rk1Alpha, i, j, k - 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk1Betau_z, i, j, k - 3)) + vector(rk1Alpha, i, j, k - 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk1Betau_z, i, j, k - 3)) + vector(rk1Alpha, i, j, k - 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk1Betau_z, i, j, k - 3)) + vector(rk1Alpha, i, j, k - 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j, k - 2));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k - 2) * vector(rk1Bfu_x, i, j, k - 2) + vector(rk1gtd_yy, i, j, k - 2) * vector(rk1Bfu_y, i, j, k - 2) + vector(rk1gtd_yz, i, j, k - 2) * vector(rk1Bfu_z, i, j, k - 2));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k - 2) * vector(rk1Bfu_x, i, j, k - 2) + vector(rk1gtd_yz, i, j, k - 2) * vector(rk1Bfu_y, i, j, k - 2) + vector(rk1gtd_zz, i, j, k - 2) * vector(rk1Bfu_z, i, j, k - 2));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k - 2) * vector(rk1Bfu_x, i, j, k - 2) + vector(rk1gtd_xy, i, j, k - 2) * vector(rk1Bfu_y, i, j, k - 2) + vector(rk1gtd_xz, i, j, k - 2) * vector(rk1Bfu_z, i, j, k - 2));
						rk1sqBf = vector(rk1Bfu_x, i, j, k - 2) * rk1Bfd_x + vector(rk1Bfu_y, i, j, k - 2) * rk1Bfd_y + vector(rk1Bfu_z, i, j, k - 2) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j, k - 2) * vector(rk1gtd_zz, i, j, k - 2) * vector(rk1gtd_xx, i, j, k - 2) + (-vector(rk1gtd_yy, i, j, k - 2) * (vector(rk1gtd_xz, i, j, k - 2) * vector(rk1gtd_xz, i, j, k - 2))) + (-vector(rk1gtd_zz, i, j, k - 2) * (vector(rk1gtd_xy, i, j, k - 2) * vector(rk1gtd_xy, i, j, k - 2))) + (-(vector(rk1gtd_yz, i, j, k - 2) * vector(rk1gtd_yz, i, j, k - 2)) * vector(rk1gtd_xx, i, j, k - 2)) + 2.0 * vector(rk1gtd_yz, i, j, k - 2) * vector(rk1gtd_xy, i, j, k - 2) * vector(rk1gtd_xz, i, j, k - 2);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j, k - 2) * vector(rk1gtd_yy, i, j, k - 2) - vector(rk1gtd_xy, i, j, k - 2) * vector(rk1gtd_xy, i, j, k - 2));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j, k - 2) * vector(rk1gtd_zz, i, j, k - 2) - vector(rk1gtd_yz, i, j, k - 2) * vector(rk1gtd_yz, i, j, k - 2));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j, k - 2) * vector(rk1gtd_yz, i, j, k - 2) - vector(rk1gtd_yy, i, j, k - 2) * vector(rk1gtd_xz, i, j, k - 2));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j, k - 2) * vector(rk1gtd_zz, i, j, k - 2)) + vector(rk1gtd_yz, i, j, k - 2) * vector(rk1gtd_xz, i, j, k - 2));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j, k - 2) + rk1gtu_xy * vector(rk1Sfd_y, i, j, k - 2) + rk1gtu_xz * vector(rk1Sfd_z, i, j, k - 2));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j, k - 2) + rk1gtu_xy * vector(vfd_y, i, j, k - 2) + rk1gtu_xz * vector(vfd_z, i, j, k - 2));
						rk1Bfvf = vector(rk1Bfu_x, i, j, k - 2) * vector(vfd_x, i, j, k - 2) + vector(rk1Bfu_y, i, j, k - 2) * vector(vfd_y, i, j, k - 2) + vector(rk1Bfu_z, i, j, k - 2) * vector(vfd_z, i, j, k - 2);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j, k - 2) * vector(rk1gtd_zz, i, j, k - 2) - vector(rk1gtd_xz, i, j, k - 2) * vector(rk1gtd_xz, i, j, k - 2));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j, k - 2) * vector(rk1gtd_yz, i, j, k - 2)) + vector(rk1gtd_xy, i, j, k - 2) * vector(rk1gtd_xz, i, j, k - 2));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j, k - 2) + rk1gtu_yz * vector(vfd_y, i, j, k - 2) + rk1gtu_zz * vector(vfd_z, i, j, k - 2));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j, k - 2) + rk1gtu_yy * vector(rk1Sfd_y, i, j, k - 2) + rk1gtu_yz * vector(rk1Sfd_z, i, j, k - 2));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j, k - 2) + rk1gtu_yz * vector(rk1Sfd_y, i, j, k - 2) + rk1gtu_zz * vector(rk1Sfd_z, i, j, k - 2));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j, k - 2) + rk1gtu_yy * vector(vfd_y, i, j, k - 2) + rk1gtu_yz * vector(vfd_z, i, j, k - 2));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j, k - 2) + rk1vfu_y * vector(vfd_y, i, j, k - 2) + rk1vfu_z * vector(vfd_z, i, j, k - 2), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j, k - 2) * vector(rk1Bfu_z, i, j, k - 2)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j, k - 2) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j, k - 2) * vector(rk1Bfu_x, i, j, k - 2)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j, k - 2) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k - 2) * vector(rk1Bfu_z, i, j, k - 2)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k - 2) * rk1vfu_z + vector(rk1Bfu_z, i, j, k - 2) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k - 2) * vector(rk1Bfu_y, i, j, k - 2)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k - 2) * rk1vfu_y + vector(rk1Bfu_y, i, j, k - 2) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j, k - 2) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k - 2) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k - 2));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j, k - 2) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j, k - 2) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j, k - 2));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j, k - 2) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j, k - 2) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j, k - 2));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k - 2) * vector(rk1Bfu_z, i, j, k - 2)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k - 2) * rk1vfu_z + vector(rk1Bfu_z, i, j, k - 2) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j, k - 2) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j, k - 2) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j, k - 2));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k - 2) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k - 2) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j, k - 2));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j, k - 2) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j, k - 2) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j, k - 2));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j, k - 2) * vector(rk1Bfu_y, i, j, k - 2)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j, k - 2) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j, k - 2) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j, k - 2) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j, k - 2));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k - 2) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j, k - 2) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k - 2));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j, k - 2) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j, k - 2) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j, k - 2));
						Fluxm2kSfd_x = FkSfd_x_mainI(vector(rk1Sfd_x, i, j, k - 2), vector(rk1gtd_xx, i, j, k - 2), vector(rk1gtd_xy, i, j, k - 2), vector(rk1gtd_xz, i, j, k - 2), vector(rk1Betau_z, i, j, k - 2), vector(rk1Alpha, i, j, k - 2), vector(TauT_xz, i, j, k - 2), vector(TauT_yz, i, j, k - 2), vector(TauT_zz, i, j, k - 2), rk1inv_chi, rk1sdetg, rk1Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxm2kSfd_y = FkSfd_y_mainI(vector(rk1Sfd_y, i, j, k - 2), vector(rk1gtd_xy, i, j, k - 2), vector(rk1gtd_yy, i, j, k - 2), vector(rk1gtd_yz, i, j, k - 2), vector(rk1Betau_z, i, j, k - 2), vector(rk1Alpha, i, j, k - 2), vector(TauT_xz, i, j, k - 2), vector(TauT_yz, i, j, k - 2), vector(TauT_zz, i, j, k - 2), rk1inv_chi, rk1sdetg, rk1Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxm2kSfd_z = FkSfd_z_mainI(vector(rk1Sfd_z, i, j, k - 2), vector(rk1gtd_xz, i, j, k - 2), vector(rk1gtd_yz, i, j, k - 2), vector(rk1gtd_zz, i, j, k - 2), vector(rk1Betau_z, i, j, k - 2), vector(rk1Alpha, i, j, k - 2), vector(TauT_xz, i, j, k - 2), vector(TauT_yz, i, j, k - 2), vector(TauT_zz, i, j, k - 2), rk1inv_chi, rk1sdetg, rk1Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm2kBfu_x = FkBfu_x_mainI(vector(rk1Bfu_x, i, j, k - 2), vector(rk1Bfu_z, i, j, k - 2), vector(rk1phif, i, j, k - 2), vector(rk1Betau_x, i, j, k - 2), vector(rk1Betau_z, i, j, k - 2), vector(rk1Alpha, i, j, k - 2), vector(TauM_xz, i, j, k - 2), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2kBfu_y = FkBfu_y_mainI(vector(rk1Bfu_y, i, j, k - 2), vector(rk1Bfu_z, i, j, k - 2), vector(rk1phif, i, j, k - 2), vector(rk1Betau_y, i, j, k - 2), vector(rk1Betau_z, i, j, k - 2), vector(rk1Alpha, i, j, k - 2), vector(TauM_yz, i, j, k - 2), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2kBfu_z = FkBfu_z_mainI(vector(rk1phif, i, j, k - 2), vector(rk1Alpha, i, j, k - 2), rk1chi_max, rk1gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm2kDf = FkDf_mainI(vector(rk1Df, i, j, k - 2), vector(rk1Betau_z, i, j, k - 2), vector(rk1Alpha, i, j, k - 2), vector(TauN_z, i, j, k - 2), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2kDYf = FkDYf_mainI(vector(rk1DYf, i, j, k - 2), vector(rk1Betau_z, i, j, k - 2), vector(rk1Alpha, i, j, k - 2), vector(TauNe_z, i, j, k - 2), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2ktauf = Fktauf_mainI(vector(rk1Df, i, j, k - 2), vector(rk1tauf, i, j, k - 2), vector(rk1Betau_z, i, j, k - 2), vector(rk1Alpha, i, j, k - 2), vector(TauN_z, i, j, k - 2), rk1sdetg, rk1Sfu_z, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2kphif = Fkphif_mainI(vector(rk1Bfu_z, i, j, k - 2), vector(rk1phif, i, j, k - 2), vector(rk1Betau_z, i, j, k - 2), vector(rk1Alpha, i, j, k - 2), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j, k - 2) * (1.0 + vector(epsf, i, j, k - 2)) + vector(pf, i, j, k - 2);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j, k - 2) + rk1sqca - vector(sqcs, i, j, k - 2) * rk1sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk1Betau_z, i, j, k - 2)) + vector(rk1Alpha, i, j, k - 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk1Betau_z, i, j, k - 2)) + vector(rk1Alpha, i, j, k - 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk1Betau_z, i, j, k - 2)) + vector(rk1Alpha, i, j, k - 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk1Betau_z, i, j, k - 2)) + vector(rk1Alpha, i, j, k - 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk1Betau_z, i, j, k - 2)) + vector(rk1Alpha, i, j, k - 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk1Betau_z, i, j, k - 2)) + vector(rk1Alpha, i, j, k - 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk1Betau_z, i, j, k - 2)) + vector(rk1Alpha, i, j, k - 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk1Betau_z, i, j, k - 2)) + vector(rk1Alpha, i, j, k - 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk1Betau_z, i, j, k - 2)) + vector(rk1Alpha, i, j, k - 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk1Betau_z, i, j, k - 2)) + vector(rk1Alpha, i, j, k - 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j, k - 1));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k - 1) * vector(rk1Bfu_x, i, j, k - 1) + vector(rk1gtd_yy, i, j, k - 1) * vector(rk1Bfu_y, i, j, k - 1) + vector(rk1gtd_yz, i, j, k - 1) * vector(rk1Bfu_z, i, j, k - 1));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k - 1) * vector(rk1Bfu_x, i, j, k - 1) + vector(rk1gtd_yz, i, j, k - 1) * vector(rk1Bfu_y, i, j, k - 1) + vector(rk1gtd_zz, i, j, k - 1) * vector(rk1Bfu_z, i, j, k - 1));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k - 1) * vector(rk1Bfu_x, i, j, k - 1) + vector(rk1gtd_xy, i, j, k - 1) * vector(rk1Bfu_y, i, j, k - 1) + vector(rk1gtd_xz, i, j, k - 1) * vector(rk1Bfu_z, i, j, k - 1));
						rk1sqBf = vector(rk1Bfu_x, i, j, k - 1) * rk1Bfd_x + vector(rk1Bfu_y, i, j, k - 1) * rk1Bfd_y + vector(rk1Bfu_z, i, j, k - 1) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j, k - 1) * vector(rk1gtd_zz, i, j, k - 1) * vector(rk1gtd_xx, i, j, k - 1) + (-vector(rk1gtd_yy, i, j, k - 1) * (vector(rk1gtd_xz, i, j, k - 1) * vector(rk1gtd_xz, i, j, k - 1))) + (-vector(rk1gtd_zz, i, j, k - 1) * (vector(rk1gtd_xy, i, j, k - 1) * vector(rk1gtd_xy, i, j, k - 1))) + (-(vector(rk1gtd_yz, i, j, k - 1) * vector(rk1gtd_yz, i, j, k - 1)) * vector(rk1gtd_xx, i, j, k - 1)) + 2.0 * vector(rk1gtd_yz, i, j, k - 1) * vector(rk1gtd_xy, i, j, k - 1) * vector(rk1gtd_xz, i, j, k - 1);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j, k - 1) * vector(rk1gtd_yy, i, j, k - 1) - vector(rk1gtd_xy, i, j, k - 1) * vector(rk1gtd_xy, i, j, k - 1));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j, k - 1) * vector(rk1gtd_zz, i, j, k - 1) - vector(rk1gtd_yz, i, j, k - 1) * vector(rk1gtd_yz, i, j, k - 1));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j, k - 1) * vector(rk1gtd_yz, i, j, k - 1) - vector(rk1gtd_yy, i, j, k - 1) * vector(rk1gtd_xz, i, j, k - 1));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j, k - 1) * vector(rk1gtd_zz, i, j, k - 1)) + vector(rk1gtd_yz, i, j, k - 1) * vector(rk1gtd_xz, i, j, k - 1));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j, k - 1) + rk1gtu_xy * vector(rk1Sfd_y, i, j, k - 1) + rk1gtu_xz * vector(rk1Sfd_z, i, j, k - 1));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j, k - 1) + rk1gtu_xy * vector(vfd_y, i, j, k - 1) + rk1gtu_xz * vector(vfd_z, i, j, k - 1));
						rk1Bfvf = vector(rk1Bfu_x, i, j, k - 1) * vector(vfd_x, i, j, k - 1) + vector(rk1Bfu_y, i, j, k - 1) * vector(vfd_y, i, j, k - 1) + vector(rk1Bfu_z, i, j, k - 1) * vector(vfd_z, i, j, k - 1);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j, k - 1) * vector(rk1gtd_zz, i, j, k - 1) - vector(rk1gtd_xz, i, j, k - 1) * vector(rk1gtd_xz, i, j, k - 1));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j, k - 1) * vector(rk1gtd_yz, i, j, k - 1)) + vector(rk1gtd_xy, i, j, k - 1) * vector(rk1gtd_xz, i, j, k - 1));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j, k - 1) + rk1gtu_yz * vector(vfd_y, i, j, k - 1) + rk1gtu_zz * vector(vfd_z, i, j, k - 1));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j, k - 1) + rk1gtu_yy * vector(rk1Sfd_y, i, j, k - 1) + rk1gtu_yz * vector(rk1Sfd_z, i, j, k - 1));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j, k - 1) + rk1gtu_yz * vector(rk1Sfd_y, i, j, k - 1) + rk1gtu_zz * vector(rk1Sfd_z, i, j, k - 1));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j, k - 1) + rk1gtu_yy * vector(vfd_y, i, j, k - 1) + rk1gtu_yz * vector(vfd_z, i, j, k - 1));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j, k - 1) + rk1vfu_y * vector(vfd_y, i, j, k - 1) + rk1vfu_z * vector(vfd_z, i, j, k - 1), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j, k - 1) * vector(rk1Bfu_z, i, j, k - 1)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j, k - 1) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j, k - 1) * vector(rk1Bfu_x, i, j, k - 1)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j, k - 1) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k - 1) * vector(rk1Bfu_z, i, j, k - 1)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k - 1) * rk1vfu_z + vector(rk1Bfu_z, i, j, k - 1) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k - 1) * vector(rk1Bfu_y, i, j, k - 1)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k - 1) * rk1vfu_y + vector(rk1Bfu_y, i, j, k - 1) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j, k - 1) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k - 1) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k - 1));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j, k - 1) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j, k - 1) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j, k - 1));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j, k - 1) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j, k - 1) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j, k - 1));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k - 1) * vector(rk1Bfu_z, i, j, k - 1)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k - 1) * rk1vfu_z + vector(rk1Bfu_z, i, j, k - 1) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j, k - 1) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j, k - 1) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j, k - 1));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k - 1) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k - 1) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j, k - 1));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j, k - 1) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j, k - 1) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j, k - 1));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j, k - 1) * vector(rk1Bfu_y, i, j, k - 1)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j, k - 1) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j, k - 1) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j, k - 1) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j, k - 1));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k - 1) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j, k - 1) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k - 1));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j, k - 1) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j, k - 1) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j, k - 1));
						Fluxm1kSfd_x = FkSfd_x_mainI(vector(rk1Sfd_x, i, j, k - 1), vector(rk1gtd_xx, i, j, k - 1), vector(rk1gtd_xy, i, j, k - 1), vector(rk1gtd_xz, i, j, k - 1), vector(rk1Betau_z, i, j, k - 1), vector(rk1Alpha, i, j, k - 1), vector(TauT_xz, i, j, k - 1), vector(TauT_yz, i, j, k - 1), vector(TauT_zz, i, j, k - 1), rk1inv_chi, rk1sdetg, rk1Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxm1kSfd_y = FkSfd_y_mainI(vector(rk1Sfd_y, i, j, k - 1), vector(rk1gtd_xy, i, j, k - 1), vector(rk1gtd_yy, i, j, k - 1), vector(rk1gtd_yz, i, j, k - 1), vector(rk1Betau_z, i, j, k - 1), vector(rk1Alpha, i, j, k - 1), vector(TauT_xz, i, j, k - 1), vector(TauT_yz, i, j, k - 1), vector(TauT_zz, i, j, k - 1), rk1inv_chi, rk1sdetg, rk1Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxm1kSfd_z = FkSfd_z_mainI(vector(rk1Sfd_z, i, j, k - 1), vector(rk1gtd_xz, i, j, k - 1), vector(rk1gtd_yz, i, j, k - 1), vector(rk1gtd_zz, i, j, k - 1), vector(rk1Betau_z, i, j, k - 1), vector(rk1Alpha, i, j, k - 1), vector(TauT_xz, i, j, k - 1), vector(TauT_yz, i, j, k - 1), vector(TauT_zz, i, j, k - 1), rk1inv_chi, rk1sdetg, rk1Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm1kBfu_x = FkBfu_x_mainI(vector(rk1Bfu_x, i, j, k - 1), vector(rk1Bfu_z, i, j, k - 1), vector(rk1phif, i, j, k - 1), vector(rk1Betau_x, i, j, k - 1), vector(rk1Betau_z, i, j, k - 1), vector(rk1Alpha, i, j, k - 1), vector(TauM_xz, i, j, k - 1), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1kBfu_y = FkBfu_y_mainI(vector(rk1Bfu_y, i, j, k - 1), vector(rk1Bfu_z, i, j, k - 1), vector(rk1phif, i, j, k - 1), vector(rk1Betau_y, i, j, k - 1), vector(rk1Betau_z, i, j, k - 1), vector(rk1Alpha, i, j, k - 1), vector(TauM_yz, i, j, k - 1), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1kBfu_z = FkBfu_z_mainI(vector(rk1phif, i, j, k - 1), vector(rk1Alpha, i, j, k - 1), rk1chi_max, rk1gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm1kDf = FkDf_mainI(vector(rk1Df, i, j, k - 1), vector(rk1Betau_z, i, j, k - 1), vector(rk1Alpha, i, j, k - 1), vector(TauN_z, i, j, k - 1), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1kDYf = FkDYf_mainI(vector(rk1DYf, i, j, k - 1), vector(rk1Betau_z, i, j, k - 1), vector(rk1Alpha, i, j, k - 1), vector(TauNe_z, i, j, k - 1), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1ktauf = Fktauf_mainI(vector(rk1Df, i, j, k - 1), vector(rk1tauf, i, j, k - 1), vector(rk1Betau_z, i, j, k - 1), vector(rk1Alpha, i, j, k - 1), vector(TauN_z, i, j, k - 1), rk1sdetg, rk1Sfu_z, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1kphif = Fkphif_mainI(vector(rk1Bfu_z, i, j, k - 1), vector(rk1phif, i, j, k - 1), vector(rk1Betau_z, i, j, k - 1), vector(rk1Alpha, i, j, k - 1), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j, k - 1) * (1.0 + vector(epsf, i, j, k - 1)) + vector(pf, i, j, k - 1);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j, k - 1) + rk1sqca - vector(sqcs, i, j, k - 1) * rk1sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk1Betau_z, i, j, k - 1)) + vector(rk1Alpha, i, j, k - 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk1Betau_z, i, j, k - 1)) + vector(rk1Alpha, i, j, k - 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk1Betau_z, i, j, k - 1)) + vector(rk1Alpha, i, j, k - 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk1Betau_z, i, j, k - 1)) + vector(rk1Alpha, i, j, k - 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk1Betau_z, i, j, k - 1)) + vector(rk1Alpha, i, j, k - 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk1Betau_z, i, j, k - 1)) + vector(rk1Alpha, i, j, k - 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk1Betau_z, i, j, k - 1)) + vector(rk1Alpha, i, j, k - 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk1Betau_z, i, j, k - 1)) + vector(rk1Alpha, i, j, k - 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk1Betau_z, i, j, k - 1)) + vector(rk1Alpha, i, j, k - 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk1Betau_z, i, j, k - 1)) + vector(rk1Alpha, i, j, k - 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j, k + 1));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k + 1) * vector(rk1Bfu_x, i, j, k + 1) + vector(rk1gtd_yy, i, j, k + 1) * vector(rk1Bfu_y, i, j, k + 1) + vector(rk1gtd_yz, i, j, k + 1) * vector(rk1Bfu_z, i, j, k + 1));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k + 1) * vector(rk1Bfu_x, i, j, k + 1) + vector(rk1gtd_yz, i, j, k + 1) * vector(rk1Bfu_y, i, j, k + 1) + vector(rk1gtd_zz, i, j, k + 1) * vector(rk1Bfu_z, i, j, k + 1));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k + 1) * vector(rk1Bfu_x, i, j, k + 1) + vector(rk1gtd_xy, i, j, k + 1) * vector(rk1Bfu_y, i, j, k + 1) + vector(rk1gtd_xz, i, j, k + 1) * vector(rk1Bfu_z, i, j, k + 1));
						rk1sqBf = vector(rk1Bfu_x, i, j, k + 1) * rk1Bfd_x + vector(rk1Bfu_y, i, j, k + 1) * rk1Bfd_y + vector(rk1Bfu_z, i, j, k + 1) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j, k + 1) * vector(rk1gtd_zz, i, j, k + 1) * vector(rk1gtd_xx, i, j, k + 1) + (-vector(rk1gtd_yy, i, j, k + 1) * (vector(rk1gtd_xz, i, j, k + 1) * vector(rk1gtd_xz, i, j, k + 1))) + (-vector(rk1gtd_zz, i, j, k + 1) * (vector(rk1gtd_xy, i, j, k + 1) * vector(rk1gtd_xy, i, j, k + 1))) + (-(vector(rk1gtd_yz, i, j, k + 1) * vector(rk1gtd_yz, i, j, k + 1)) * vector(rk1gtd_xx, i, j, k + 1)) + 2.0 * vector(rk1gtd_yz, i, j, k + 1) * vector(rk1gtd_xy, i, j, k + 1) * vector(rk1gtd_xz, i, j, k + 1);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j, k + 1) * vector(rk1gtd_yy, i, j, k + 1) - vector(rk1gtd_xy, i, j, k + 1) * vector(rk1gtd_xy, i, j, k + 1));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j, k + 1) * vector(rk1gtd_zz, i, j, k + 1) - vector(rk1gtd_yz, i, j, k + 1) * vector(rk1gtd_yz, i, j, k + 1));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j, k + 1) * vector(rk1gtd_yz, i, j, k + 1) - vector(rk1gtd_yy, i, j, k + 1) * vector(rk1gtd_xz, i, j, k + 1));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j, k + 1) * vector(rk1gtd_zz, i, j, k + 1)) + vector(rk1gtd_yz, i, j, k + 1) * vector(rk1gtd_xz, i, j, k + 1));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j, k + 1) + rk1gtu_xy * vector(rk1Sfd_y, i, j, k + 1) + rk1gtu_xz * vector(rk1Sfd_z, i, j, k + 1));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j, k + 1) + rk1gtu_xy * vector(vfd_y, i, j, k + 1) + rk1gtu_xz * vector(vfd_z, i, j, k + 1));
						rk1Bfvf = vector(rk1Bfu_x, i, j, k + 1) * vector(vfd_x, i, j, k + 1) + vector(rk1Bfu_y, i, j, k + 1) * vector(vfd_y, i, j, k + 1) + vector(rk1Bfu_z, i, j, k + 1) * vector(vfd_z, i, j, k + 1);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j, k + 1) * vector(rk1gtd_zz, i, j, k + 1) - vector(rk1gtd_xz, i, j, k + 1) * vector(rk1gtd_xz, i, j, k + 1));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j, k + 1) * vector(rk1gtd_yz, i, j, k + 1)) + vector(rk1gtd_xy, i, j, k + 1) * vector(rk1gtd_xz, i, j, k + 1));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j, k + 1) + rk1gtu_yz * vector(vfd_y, i, j, k + 1) + rk1gtu_zz * vector(vfd_z, i, j, k + 1));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j, k + 1) + rk1gtu_yy * vector(rk1Sfd_y, i, j, k + 1) + rk1gtu_yz * vector(rk1Sfd_z, i, j, k + 1));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j, k + 1) + rk1gtu_yz * vector(rk1Sfd_y, i, j, k + 1) + rk1gtu_zz * vector(rk1Sfd_z, i, j, k + 1));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j, k + 1) + rk1gtu_yy * vector(vfd_y, i, j, k + 1) + rk1gtu_yz * vector(vfd_z, i, j, k + 1));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j, k + 1) + rk1vfu_y * vector(vfd_y, i, j, k + 1) + rk1vfu_z * vector(vfd_z, i, j, k + 1), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j, k + 1) * vector(rk1Bfu_z, i, j, k + 1)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j, k + 1) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j, k + 1) * vector(rk1Bfu_x, i, j, k + 1)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j, k + 1) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k + 1) * vector(rk1Bfu_z, i, j, k + 1)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k + 1) * rk1vfu_z + vector(rk1Bfu_z, i, j, k + 1) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k + 1) * vector(rk1Bfu_y, i, j, k + 1)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k + 1) * rk1vfu_y + vector(rk1Bfu_y, i, j, k + 1) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j, k + 1) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k + 1) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k + 1));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j, k + 1) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j, k + 1) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j, k + 1));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j, k + 1) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j, k + 1) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j, k + 1));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k + 1) * vector(rk1Bfu_z, i, j, k + 1)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k + 1) * rk1vfu_z + vector(rk1Bfu_z, i, j, k + 1) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j, k + 1) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j, k + 1) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j, k + 1));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k + 1) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k + 1) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j, k + 1));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j, k + 1) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j, k + 1) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j, k + 1));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j, k + 1) * vector(rk1Bfu_y, i, j, k + 1)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j, k + 1) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j, k + 1) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j, k + 1) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j, k + 1));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k + 1) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j, k + 1) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k + 1));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j, k + 1) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j, k + 1) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j, k + 1));
						Fluxp1kSfd_x = FkSfd_x_mainI(vector(rk1Sfd_x, i, j, k + 1), vector(rk1gtd_xx, i, j, k + 1), vector(rk1gtd_xy, i, j, k + 1), vector(rk1gtd_xz, i, j, k + 1), vector(rk1Betau_z, i, j, k + 1), vector(rk1Alpha, i, j, k + 1), vector(TauT_xz, i, j, k + 1), vector(TauT_yz, i, j, k + 1), vector(TauT_zz, i, j, k + 1), rk1inv_chi, rk1sdetg, rk1Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxp1kSfd_y = FkSfd_y_mainI(vector(rk1Sfd_y, i, j, k + 1), vector(rk1gtd_xy, i, j, k + 1), vector(rk1gtd_yy, i, j, k + 1), vector(rk1gtd_yz, i, j, k + 1), vector(rk1Betau_z, i, j, k + 1), vector(rk1Alpha, i, j, k + 1), vector(TauT_xz, i, j, k + 1), vector(TauT_yz, i, j, k + 1), vector(TauT_zz, i, j, k + 1), rk1inv_chi, rk1sdetg, rk1Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxp1kSfd_z = FkSfd_z_mainI(vector(rk1Sfd_z, i, j, k + 1), vector(rk1gtd_xz, i, j, k + 1), vector(rk1gtd_yz, i, j, k + 1), vector(rk1gtd_zz, i, j, k + 1), vector(rk1Betau_z, i, j, k + 1), vector(rk1Alpha, i, j, k + 1), vector(TauT_xz, i, j, k + 1), vector(TauT_yz, i, j, k + 1), vector(TauT_zz, i, j, k + 1), rk1inv_chi, rk1sdetg, rk1Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp1kBfu_x = FkBfu_x_mainI(vector(rk1Bfu_x, i, j, k + 1), vector(rk1Bfu_z, i, j, k + 1), vector(rk1phif, i, j, k + 1), vector(rk1Betau_x, i, j, k + 1), vector(rk1Betau_z, i, j, k + 1), vector(rk1Alpha, i, j, k + 1), vector(TauM_xz, i, j, k + 1), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1kBfu_y = FkBfu_y_mainI(vector(rk1Bfu_y, i, j, k + 1), vector(rk1Bfu_z, i, j, k + 1), vector(rk1phif, i, j, k + 1), vector(rk1Betau_y, i, j, k + 1), vector(rk1Betau_z, i, j, k + 1), vector(rk1Alpha, i, j, k + 1), vector(TauM_yz, i, j, k + 1), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1kBfu_z = FkBfu_z_mainI(vector(rk1phif, i, j, k + 1), vector(rk1Alpha, i, j, k + 1), rk1chi_max, rk1gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp1kDf = FkDf_mainI(vector(rk1Df, i, j, k + 1), vector(rk1Betau_z, i, j, k + 1), vector(rk1Alpha, i, j, k + 1), vector(TauN_z, i, j, k + 1), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1kDYf = FkDYf_mainI(vector(rk1DYf, i, j, k + 1), vector(rk1Betau_z, i, j, k + 1), vector(rk1Alpha, i, j, k + 1), vector(TauNe_z, i, j, k + 1), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1ktauf = Fktauf_mainI(vector(rk1Df, i, j, k + 1), vector(rk1tauf, i, j, k + 1), vector(rk1Betau_z, i, j, k + 1), vector(rk1Alpha, i, j, k + 1), vector(TauN_z, i, j, k + 1), rk1sdetg, rk1Sfu_z, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1kphif = Fkphif_mainI(vector(rk1Bfu_z, i, j, k + 1), vector(rk1phif, i, j, k + 1), vector(rk1Betau_z, i, j, k + 1), vector(rk1Alpha, i, j, k + 1), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j, k + 1) * (1.0 + vector(epsf, i, j, k + 1)) + vector(pf, i, j, k + 1);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j, k + 1) + rk1sqca - vector(sqcs, i, j, k + 1) * rk1sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk1Betau_z, i, j, k + 1)) + vector(rk1Alpha, i, j, k + 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk1Betau_z, i, j, k + 1)) + vector(rk1Alpha, i, j, k + 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk1Betau_z, i, j, k + 1)) + vector(rk1Alpha, i, j, k + 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk1Betau_z, i, j, k + 1)) + vector(rk1Alpha, i, j, k + 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk1Betau_z, i, j, k + 1)) + vector(rk1Alpha, i, j, k + 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk1Betau_z, i, j, k + 1)) + vector(rk1Alpha, i, j, k + 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk1Betau_z, i, j, k + 1)) + vector(rk1Alpha, i, j, k + 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk1Betau_z, i, j, k + 1)) + vector(rk1Alpha, i, j, k + 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk1Betau_z, i, j, k + 1)) + vector(rk1Alpha, i, j, k + 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk1Betau_z, i, j, k + 1)) + vector(rk1Alpha, i, j, k + 1) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j, k + 2));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k + 2) * vector(rk1Bfu_x, i, j, k + 2) + vector(rk1gtd_yy, i, j, k + 2) * vector(rk1Bfu_y, i, j, k + 2) + vector(rk1gtd_yz, i, j, k + 2) * vector(rk1Bfu_z, i, j, k + 2));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k + 2) * vector(rk1Bfu_x, i, j, k + 2) + vector(rk1gtd_yz, i, j, k + 2) * vector(rk1Bfu_y, i, j, k + 2) + vector(rk1gtd_zz, i, j, k + 2) * vector(rk1Bfu_z, i, j, k + 2));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k + 2) * vector(rk1Bfu_x, i, j, k + 2) + vector(rk1gtd_xy, i, j, k + 2) * vector(rk1Bfu_y, i, j, k + 2) + vector(rk1gtd_xz, i, j, k + 2) * vector(rk1Bfu_z, i, j, k + 2));
						rk1sqBf = vector(rk1Bfu_x, i, j, k + 2) * rk1Bfd_x + vector(rk1Bfu_y, i, j, k + 2) * rk1Bfd_y + vector(rk1Bfu_z, i, j, k + 2) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j, k + 2) * vector(rk1gtd_zz, i, j, k + 2) * vector(rk1gtd_xx, i, j, k + 2) + (-vector(rk1gtd_yy, i, j, k + 2) * (vector(rk1gtd_xz, i, j, k + 2) * vector(rk1gtd_xz, i, j, k + 2))) + (-vector(rk1gtd_zz, i, j, k + 2) * (vector(rk1gtd_xy, i, j, k + 2) * vector(rk1gtd_xy, i, j, k + 2))) + (-(vector(rk1gtd_yz, i, j, k + 2) * vector(rk1gtd_yz, i, j, k + 2)) * vector(rk1gtd_xx, i, j, k + 2)) + 2.0 * vector(rk1gtd_yz, i, j, k + 2) * vector(rk1gtd_xy, i, j, k + 2) * vector(rk1gtd_xz, i, j, k + 2);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j, k + 2) * vector(rk1gtd_yy, i, j, k + 2) - vector(rk1gtd_xy, i, j, k + 2) * vector(rk1gtd_xy, i, j, k + 2));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j, k + 2) * vector(rk1gtd_zz, i, j, k + 2) - vector(rk1gtd_yz, i, j, k + 2) * vector(rk1gtd_yz, i, j, k + 2));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j, k + 2) * vector(rk1gtd_yz, i, j, k + 2) - vector(rk1gtd_yy, i, j, k + 2) * vector(rk1gtd_xz, i, j, k + 2));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j, k + 2) * vector(rk1gtd_zz, i, j, k + 2)) + vector(rk1gtd_yz, i, j, k + 2) * vector(rk1gtd_xz, i, j, k + 2));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j, k + 2) + rk1gtu_xy * vector(rk1Sfd_y, i, j, k + 2) + rk1gtu_xz * vector(rk1Sfd_z, i, j, k + 2));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j, k + 2) + rk1gtu_xy * vector(vfd_y, i, j, k + 2) + rk1gtu_xz * vector(vfd_z, i, j, k + 2));
						rk1Bfvf = vector(rk1Bfu_x, i, j, k + 2) * vector(vfd_x, i, j, k + 2) + vector(rk1Bfu_y, i, j, k + 2) * vector(vfd_y, i, j, k + 2) + vector(rk1Bfu_z, i, j, k + 2) * vector(vfd_z, i, j, k + 2);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j, k + 2) * vector(rk1gtd_zz, i, j, k + 2) - vector(rk1gtd_xz, i, j, k + 2) * vector(rk1gtd_xz, i, j, k + 2));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j, k + 2) * vector(rk1gtd_yz, i, j, k + 2)) + vector(rk1gtd_xy, i, j, k + 2) * vector(rk1gtd_xz, i, j, k + 2));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j, k + 2) + rk1gtu_yz * vector(vfd_y, i, j, k + 2) + rk1gtu_zz * vector(vfd_z, i, j, k + 2));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j, k + 2) + rk1gtu_yy * vector(rk1Sfd_y, i, j, k + 2) + rk1gtu_yz * vector(rk1Sfd_z, i, j, k + 2));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j, k + 2) + rk1gtu_yz * vector(rk1Sfd_y, i, j, k + 2) + rk1gtu_zz * vector(rk1Sfd_z, i, j, k + 2));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j, k + 2) + rk1gtu_yy * vector(vfd_y, i, j, k + 2) + rk1gtu_yz * vector(vfd_z, i, j, k + 2));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j, k + 2) + rk1vfu_y * vector(vfd_y, i, j, k + 2) + rk1vfu_z * vector(vfd_z, i, j, k + 2), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j, k + 2) * vector(rk1Bfu_z, i, j, k + 2)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j, k + 2) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j, k + 2) * vector(rk1Bfu_x, i, j, k + 2)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j, k + 2) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k + 2) * vector(rk1Bfu_z, i, j, k + 2)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k + 2) * rk1vfu_z + vector(rk1Bfu_z, i, j, k + 2) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k + 2) * vector(rk1Bfu_y, i, j, k + 2)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k + 2) * rk1vfu_y + vector(rk1Bfu_y, i, j, k + 2) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j, k + 2) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k + 2) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k + 2));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j, k + 2) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j, k + 2) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j, k + 2));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j, k + 2) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j, k + 2) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j, k + 2));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k + 2) * vector(rk1Bfu_z, i, j, k + 2)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k + 2) * rk1vfu_z + vector(rk1Bfu_z, i, j, k + 2) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j, k + 2) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j, k + 2) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j, k + 2));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k + 2) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k + 2) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j, k + 2));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j, k + 2) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j, k + 2) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j, k + 2));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j, k + 2) * vector(rk1Bfu_y, i, j, k + 2)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j, k + 2) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j, k + 2) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j, k + 2) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j, k + 2));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k + 2) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j, k + 2) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k + 2));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j, k + 2) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j, k + 2) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j, k + 2));
						Fluxp2kSfd_x = FkSfd_x_mainI(vector(rk1Sfd_x, i, j, k + 2), vector(rk1gtd_xx, i, j, k + 2), vector(rk1gtd_xy, i, j, k + 2), vector(rk1gtd_xz, i, j, k + 2), vector(rk1Betau_z, i, j, k + 2), vector(rk1Alpha, i, j, k + 2), vector(TauT_xz, i, j, k + 2), vector(TauT_yz, i, j, k + 2), vector(TauT_zz, i, j, k + 2), rk1inv_chi, rk1sdetg, rk1Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxp2kSfd_y = FkSfd_y_mainI(vector(rk1Sfd_y, i, j, k + 2), vector(rk1gtd_xy, i, j, k + 2), vector(rk1gtd_yy, i, j, k + 2), vector(rk1gtd_yz, i, j, k + 2), vector(rk1Betau_z, i, j, k + 2), vector(rk1Alpha, i, j, k + 2), vector(TauT_xz, i, j, k + 2), vector(TauT_yz, i, j, k + 2), vector(TauT_zz, i, j, k + 2), rk1inv_chi, rk1sdetg, rk1Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxp2kSfd_z = FkSfd_z_mainI(vector(rk1Sfd_z, i, j, k + 2), vector(rk1gtd_xz, i, j, k + 2), vector(rk1gtd_yz, i, j, k + 2), vector(rk1gtd_zz, i, j, k + 2), vector(rk1Betau_z, i, j, k + 2), vector(rk1Alpha, i, j, k + 2), vector(TauT_xz, i, j, k + 2), vector(TauT_yz, i, j, k + 2), vector(TauT_zz, i, j, k + 2), rk1inv_chi, rk1sdetg, rk1Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp2kBfu_x = FkBfu_x_mainI(vector(rk1Bfu_x, i, j, k + 2), vector(rk1Bfu_z, i, j, k + 2), vector(rk1phif, i, j, k + 2), vector(rk1Betau_x, i, j, k + 2), vector(rk1Betau_z, i, j, k + 2), vector(rk1Alpha, i, j, k + 2), vector(TauM_xz, i, j, k + 2), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2kBfu_y = FkBfu_y_mainI(vector(rk1Bfu_y, i, j, k + 2), vector(rk1Bfu_z, i, j, k + 2), vector(rk1phif, i, j, k + 2), vector(rk1Betau_y, i, j, k + 2), vector(rk1Betau_z, i, j, k + 2), vector(rk1Alpha, i, j, k + 2), vector(TauM_yz, i, j, k + 2), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2kBfu_z = FkBfu_z_mainI(vector(rk1phif, i, j, k + 2), vector(rk1Alpha, i, j, k + 2), rk1chi_max, rk1gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp2kDf = FkDf_mainI(vector(rk1Df, i, j, k + 2), vector(rk1Betau_z, i, j, k + 2), vector(rk1Alpha, i, j, k + 2), vector(TauN_z, i, j, k + 2), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2kDYf = FkDYf_mainI(vector(rk1DYf, i, j, k + 2), vector(rk1Betau_z, i, j, k + 2), vector(rk1Alpha, i, j, k + 2), vector(TauNe_z, i, j, k + 2), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2ktauf = Fktauf_mainI(vector(rk1Df, i, j, k + 2), vector(rk1tauf, i, j, k + 2), vector(rk1Betau_z, i, j, k + 2), vector(rk1Alpha, i, j, k + 2), vector(TauN_z, i, j, k + 2), rk1sdetg, rk1Sfu_z, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2kphif = Fkphif_mainI(vector(rk1Bfu_z, i, j, k + 2), vector(rk1phif, i, j, k + 2), vector(rk1Betau_z, i, j, k + 2), vector(rk1Alpha, i, j, k + 2), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j, k + 2) * (1.0 + vector(epsf, i, j, k + 2)) + vector(pf, i, j, k + 2);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j, k + 2) + rk1sqca - vector(sqcs, i, j, k + 2) * rk1sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk1Betau_z, i, j, k + 2)) + vector(rk1Alpha, i, j, k + 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk1Betau_z, i, j, k + 2)) + vector(rk1Alpha, i, j, k + 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk1Betau_z, i, j, k + 2)) + vector(rk1Alpha, i, j, k + 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk1Betau_z, i, j, k + 2)) + vector(rk1Alpha, i, j, k + 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk1Betau_z, i, j, k + 2)) + vector(rk1Alpha, i, j, k + 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk1Betau_z, i, j, k + 2)) + vector(rk1Alpha, i, j, k + 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk1Betau_z, i, j, k + 2)) + vector(rk1Alpha, i, j, k + 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk1Betau_z, i, j, k + 2)) + vector(rk1Alpha, i, j, k + 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk1Betau_z, i, j, k + 2)) + vector(rk1Alpha, i, j, k + 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk1Betau_z, i, j, k + 2)) + vector(rk1Alpha, i, j, k + 2) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j, k + 3));
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k + 3) * vector(rk1Bfu_x, i, j, k + 3) + vector(rk1gtd_yy, i, j, k + 3) * vector(rk1Bfu_y, i, j, k + 3) + vector(rk1gtd_yz, i, j, k + 3) * vector(rk1Bfu_z, i, j, k + 3));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k + 3) * vector(rk1Bfu_x, i, j, k + 3) + vector(rk1gtd_yz, i, j, k + 3) * vector(rk1Bfu_y, i, j, k + 3) + vector(rk1gtd_zz, i, j, k + 3) * vector(rk1Bfu_z, i, j, k + 3));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k + 3) * vector(rk1Bfu_x, i, j, k + 3) + vector(rk1gtd_xy, i, j, k + 3) * vector(rk1Bfu_y, i, j, k + 3) + vector(rk1gtd_xz, i, j, k + 3) * vector(rk1Bfu_z, i, j, k + 3));
						rk1sqBf = vector(rk1Bfu_x, i, j, k + 3) * rk1Bfd_x + vector(rk1Bfu_y, i, j, k + 3) * rk1Bfd_y + vector(rk1Bfu_z, i, j, k + 3) * rk1Bfd_z;
						rk1detgtd = vector(rk1gtd_yy, i, j, k + 3) * vector(rk1gtd_zz, i, j, k + 3) * vector(rk1gtd_xx, i, j, k + 3) + (-vector(rk1gtd_yy, i, j, k + 3) * (vector(rk1gtd_xz, i, j, k + 3) * vector(rk1gtd_xz, i, j, k + 3))) + (-vector(rk1gtd_zz, i, j, k + 3) * (vector(rk1gtd_xy, i, j, k + 3) * vector(rk1gtd_xy, i, j, k + 3))) + (-(vector(rk1gtd_yz, i, j, k + 3) * vector(rk1gtd_yz, i, j, k + 3)) * vector(rk1gtd_xx, i, j, k + 3)) + 2.0 * vector(rk1gtd_yz, i, j, k + 3) * vector(rk1gtd_xy, i, j, k + 3) * vector(rk1gtd_xz, i, j, k + 3);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j, k + 3) * vector(rk1gtd_yy, i, j, k + 3) - vector(rk1gtd_xy, i, j, k + 3) * vector(rk1gtd_xy, i, j, k + 3));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j, k + 3) * vector(rk1gtd_zz, i, j, k + 3) - vector(rk1gtd_yz, i, j, k + 3) * vector(rk1gtd_yz, i, j, k + 3));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j, k + 3) * vector(rk1gtd_yz, i, j, k + 3) - vector(rk1gtd_yy, i, j, k + 3) * vector(rk1gtd_xz, i, j, k + 3));
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j, k + 3) * vector(rk1gtd_zz, i, j, k + 3)) + vector(rk1gtd_yz, i, j, k + 3) * vector(rk1gtd_xz, i, j, k + 3));
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j, k + 3) + rk1gtu_xy * vector(rk1Sfd_y, i, j, k + 3) + rk1gtu_xz * vector(rk1Sfd_z, i, j, k + 3));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j, k + 3) + rk1gtu_xy * vector(vfd_y, i, j, k + 3) + rk1gtu_xz * vector(vfd_z, i, j, k + 3));
						rk1Bfvf = vector(rk1Bfu_x, i, j, k + 3) * vector(vfd_x, i, j, k + 3) + vector(rk1Bfu_y, i, j, k + 3) * vector(vfd_y, i, j, k + 3) + vector(rk1Bfu_z, i, j, k + 3) * vector(vfd_z, i, j, k + 3);
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j, k + 3) * vector(rk1gtd_zz, i, j, k + 3) - vector(rk1gtd_xz, i, j, k + 3) * vector(rk1gtd_xz, i, j, k + 3));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j, k + 3) * vector(rk1gtd_yz, i, j, k + 3)) + vector(rk1gtd_xy, i, j, k + 3) * vector(rk1gtd_xz, i, j, k + 3));
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j, k + 3) + rk1gtu_yz * vector(vfd_y, i, j, k + 3) + rk1gtu_zz * vector(vfd_z, i, j, k + 3));
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j, k + 3) + rk1gtu_yy * vector(rk1Sfd_y, i, j, k + 3) + rk1gtu_yz * vector(rk1Sfd_z, i, j, k + 3));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j, k + 3) + rk1gtu_yz * vector(rk1Sfd_y, i, j, k + 3) + rk1gtu_zz * vector(rk1Sfd_z, i, j, k + 3));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j, k + 3) + rk1gtu_yy * vector(vfd_y, i, j, k + 3) + rk1gtu_yz * vector(vfd_z, i, j, k + 3));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j, k + 3) + rk1vfu_y * vector(vfd_y, i, j, k + 3) + rk1vfu_z * vector(vfd_z, i, j, k + 3), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j, k + 3) * vector(rk1Bfu_z, i, j, k + 3)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j, k + 3) * rk1vfu_z)) / rk1sdetg);
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j, k + 3) * vector(rk1Bfu_x, i, j, k + 3)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j, k + 3) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k + 3) * vector(rk1Bfu_z, i, j, k + 3)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k + 3) * rk1vfu_z + vector(rk1Bfu_z, i, j, k + 3) * rk1vfu_x)) / rk1sdetg);
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k + 3) * vector(rk1Bfu_y, i, j, k + 3)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k + 3) * rk1vfu_y + vector(rk1Bfu_y, i, j, k + 3) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j, k + 3) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k + 3) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k + 3));
						rk1Sfud_xy = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xy, i, j, k + 3) + rk1Sfuu_xy * vector(rk1gtd_yy, i, j, k + 3) + rk1Sfuu_xz * vector(rk1gtd_yz, i, j, k + 3));
						rk1Sfud_xz = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xz, i, j, k + 3) + rk1Sfuu_xy * vector(rk1gtd_yz, i, j, k + 3) + rk1Sfuu_xz * vector(rk1gtd_zz, i, j, k + 3));
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k + 3) * vector(rk1Bfu_z, i, j, k + 3)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k + 3) * rk1vfu_z + vector(rk1Bfu_z, i, j, k + 3) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_zy = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xy, i, j, k + 3) + rk1Sfuu_yz * vector(rk1gtd_yy, i, j, k + 3) + rk1Sfuu_zz * vector(rk1gtd_yz, i, j, k + 3));
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k + 3) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k + 3) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j, k + 3));
						rk1Sfud_zx = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xx, i, j, k + 3) + rk1Sfuu_yz * vector(rk1gtd_xy, i, j, k + 3) + rk1Sfuu_zz * vector(rk1gtd_xz, i, j, k + 3));
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j, k + 3) * vector(rk1Bfu_y, i, j, k + 3)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j, k + 3) * rk1vfu_y)) / rk1sdetg);
						rk1Sfud_yx = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xx, i, j, k + 3) + rk1Sfuu_yy * vector(rk1gtd_xy, i, j, k + 3) + rk1Sfuu_yz * vector(rk1gtd_xz, i, j, k + 3));
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k + 3) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j, k + 3) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k + 3));
						rk1Sfud_yz = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xz, i, j, k + 3) + rk1Sfuu_yy * vector(rk1gtd_yz, i, j, k + 3) + rk1Sfuu_yz * vector(rk1gtd_zz, i, j, k + 3));
						Fluxp3kSfd_x = FkSfd_x_mainI(vector(rk1Sfd_x, i, j, k + 3), vector(rk1gtd_xx, i, j, k + 3), vector(rk1gtd_xy, i, j, k + 3), vector(rk1gtd_xz, i, j, k + 3), vector(rk1Betau_z, i, j, k + 3), vector(rk1Alpha, i, j, k + 3), vector(TauT_xz, i, j, k + 3), vector(TauT_yz, i, j, k + 3), vector(TauT_zz, i, j, k + 3), rk1inv_chi, rk1sdetg, rk1Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxp3kSfd_y = FkSfd_y_mainI(vector(rk1Sfd_y, i, j, k + 3), vector(rk1gtd_xy, i, j, k + 3), vector(rk1gtd_yy, i, j, k + 3), vector(rk1gtd_yz, i, j, k + 3), vector(rk1Betau_z, i, j, k + 3), vector(rk1Alpha, i, j, k + 3), vector(TauT_xz, i, j, k + 3), vector(TauT_yz, i, j, k + 3), vector(TauT_zz, i, j, k + 3), rk1inv_chi, rk1sdetg, rk1Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxp3kSfd_z = FkSfd_z_mainI(vector(rk1Sfd_z, i, j, k + 3), vector(rk1gtd_xz, i, j, k + 3), vector(rk1gtd_yz, i, j, k + 3), vector(rk1gtd_zz, i, j, k + 3), vector(rk1Betau_z, i, j, k + 3), vector(rk1Alpha, i, j, k + 3), vector(TauT_xz, i, j, k + 3), vector(TauT_yz, i, j, k + 3), vector(TauT_zz, i, j, k + 3), rk1inv_chi, rk1sdetg, rk1Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp3kBfu_x = FkBfu_x_mainI(vector(rk1Bfu_x, i, j, k + 3), vector(rk1Bfu_z, i, j, k + 3), vector(rk1phif, i, j, k + 3), vector(rk1Betau_x, i, j, k + 3), vector(rk1Betau_z, i, j, k + 3), vector(rk1Alpha, i, j, k + 3), vector(TauM_xz, i, j, k + 3), rk1chi_max, rk1gtu_xz, rk1sdetg, rk1vfu_x, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3kBfu_y = FkBfu_y_mainI(vector(rk1Bfu_y, i, j, k + 3), vector(rk1Bfu_z, i, j, k + 3), vector(rk1phif, i, j, k + 3), vector(rk1Betau_y, i, j, k + 3), vector(rk1Betau_z, i, j, k + 3), vector(rk1Alpha, i, j, k + 3), vector(TauM_yz, i, j, k + 3), rk1chi_max, rk1gtu_yz, rk1sdetg, rk1vfu_y, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3kBfu_z = FkBfu_z_mainI(vector(rk1phif, i, j, k + 3), vector(rk1Alpha, i, j, k + 3), rk1chi_max, rk1gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp3kDf = FkDf_mainI(vector(rk1Df, i, j, k + 3), vector(rk1Betau_z, i, j, k + 3), vector(rk1Alpha, i, j, k + 3), vector(TauN_z, i, j, k + 3), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3kDYf = FkDYf_mainI(vector(rk1DYf, i, j, k + 3), vector(rk1Betau_z, i, j, k + 3), vector(rk1Alpha, i, j, k + 3), vector(TauNe_z, i, j, k + 3), rk1sdetg, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3ktauf = Fktauf_mainI(vector(rk1Df, i, j, k + 3), vector(rk1tauf, i, j, k + 3), vector(rk1Betau_z, i, j, k + 3), vector(rk1Alpha, i, j, k + 3), vector(TauN_z, i, j, k + 3), rk1sdetg, rk1Sfu_z, rk1vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3kphif = Fkphif_mainI(vector(rk1Bfu_z, i, j, k + 3), vector(rk1phif, i, j, k + 3), vector(rk1Betau_z, i, j, k + 3), vector(rk1Alpha, i, j, k + 3), ch, dx, simPlat_dt, ilast, jlast);
						rk1h = vector(rhof, i, j, k + 3) * (1.0 + vector(epsf, i, j, k + 3)) + vector(pf, i, j, k + 3);
						rk1sqbc = rk1sqBf / rk1sqW + rk1Bfvf * rk1Bfvf;
						rk1sqca = fabs(rk1sqbc / (rk1h + rk1sqbc));
						rk1sqa = vector(sqcs, i, j, k + 3) + rk1sqca - vector(sqcs, i, j, k + 3) * rk1sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk1Betau_z, i, j, k + 3)) + vector(rk1Alpha, i, j, k + 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk1Betau_z, i, j, k + 3)) + vector(rk1Alpha, i, j, k + 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk1Betau_z, i, j, k + 3)) + vector(rk1Alpha, i, j, k + 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk1Betau_z, i, j, k + 3)) + vector(rk1Alpha, i, j, k + 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk1Betau_z, i, j, k + 3)) + vector(rk1Alpha, i, j, k + 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk1Betau_z, i, j, k + 3)) + vector(rk1Alpha, i, j, k + 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk1Betau_z, i, j, k + 3)) + vector(rk1Alpha, i, j, k + 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk1Betau_z, i, j, k + 3)) + vector(rk1Alpha, i, j, k + 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk1Betau_z, i, j, k + 3)) + vector(rk1Alpha, i, j, k + 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk1Betau_z, i, j, k + 3)) + vector(rk1Alpha, i, j, k + 3) / MAX(1.0 - rk1sqvf * rk1sqa, 0.0000000001) * (fabs((1.0 - rk1sqa) * rk1vfu_z) + sqrt(fabs(rk1sqa * (1.0 - rk1sqvf) * ((1.0 - rk1sqvf * rk1sqa) * rk1chi_max * rk1gtu_zz - (1.0 - rk1sqa) * (rk1vfu_z * rk1vfu_z)))))));
						LLFpkSfd_x = 0.5 * (FluxkSfd_x + fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k));
						LLFpkSfd_y = 0.5 * (FluxkSfd_y + fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k));
						LLFpkSfd_z = 0.5 * (FluxkSfd_z + fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k));
						LLFpkBfu_x = 0.5 * (FluxkBfu_x + fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k));
						LLFpkBfu_y = 0.5 * (FluxkBfu_y + fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k));
						LLFpkBfu_z = 0.5 * (FluxkBfu_z + fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k));
						LLFpkDf = 0.5 * (FluxkDf + fabs(SpeedkDf) * vector(rk1Df, i, j, k));
						LLFpkDYf = 0.5 * (FluxkDYf + fabs(SpeedkDYf) * vector(rk1DYf, i, j, k));
						LLFpktauf = 0.5 * (Fluxktauf + fabs(Speedktauf) * vector(rk1tauf, i, j, k));
						LLFpkphif = 0.5 * (Fluxkphif + fabs(Speedkphif) * vector(rk1phif, i, j, k));
						LLFnkSfd_x = 0.5 * (FluxkSfd_x - fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k));
						LLFnkSfd_y = 0.5 * (FluxkSfd_y - fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k));
						LLFnkSfd_z = 0.5 * (FluxkSfd_z - fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k));
						LLFnkBfu_x = 0.5 * (FluxkBfu_x - fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k));
						LLFnkBfu_y = 0.5 * (FluxkBfu_y - fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k));
						LLFnkBfu_z = 0.5 * (FluxkBfu_z - fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k));
						LLFnkDf = 0.5 * (FluxkDf - fabs(SpeedkDf) * vector(rk1Df, i, j, k));
						LLFnkDYf = 0.5 * (FluxkDYf - fabs(SpeedkDYf) * vector(rk1DYf, i, j, k));
						LLFnktauf = 0.5 * (Fluxktauf - fabs(Speedktauf) * vector(rk1tauf, i, j, k));
						LLFnkphif = 0.5 * (Fluxkphif - fabs(Speedkphif) * vector(rk1phif, i, j, k));
						LLFp_m3kSfd_x = 0.5 * (Fluxm3kSfd_x + fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k - 3));
						LLFp_m3kSfd_y = 0.5 * (Fluxm3kSfd_y + fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k - 3));
						LLFp_m3kSfd_z = 0.5 * (Fluxm3kSfd_z + fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k - 3));
						LLFp_m3kBfu_x = 0.5 * (Fluxm3kBfu_x + fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k - 3));
						LLFp_m3kBfu_y = 0.5 * (Fluxm3kBfu_y + fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k - 3));
						LLFp_m3kBfu_z = 0.5 * (Fluxm3kBfu_z + fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k - 3));
						LLFp_m3kDf = 0.5 * (Fluxm3kDf + fabs(SpeedkDf) * vector(rk1Df, i, j, k - 3));
						LLFp_m3kDYf = 0.5 * (Fluxm3kDYf + fabs(SpeedkDYf) * vector(rk1DYf, i, j, k - 3));
						LLFp_m3ktauf = 0.5 * (Fluxm3ktauf + fabs(Speedktauf) * vector(rk1tauf, i, j, k - 3));
						LLFp_m3kphif = 0.5 * (Fluxm3kphif + fabs(Speedkphif) * vector(rk1phif, i, j, k - 3));
						LLFn_m3kSfd_x = 0.5 * (Fluxm3kSfd_x - fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k - 3));
						LLFn_m3kSfd_y = 0.5 * (Fluxm3kSfd_y - fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k - 3));
						LLFn_m3kSfd_z = 0.5 * (Fluxm3kSfd_z - fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k - 3));
						LLFn_m3kBfu_x = 0.5 * (Fluxm3kBfu_x - fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k - 3));
						LLFn_m3kBfu_y = 0.5 * (Fluxm3kBfu_y - fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k - 3));
						LLFn_m3kBfu_z = 0.5 * (Fluxm3kBfu_z - fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k - 3));
						LLFn_m3kDf = 0.5 * (Fluxm3kDf - fabs(SpeedkDf) * vector(rk1Df, i, j, k - 3));
						LLFn_m3kDYf = 0.5 * (Fluxm3kDYf - fabs(SpeedkDYf) * vector(rk1DYf, i, j, k - 3));
						LLFn_m3ktauf = 0.5 * (Fluxm3ktauf - fabs(Speedktauf) * vector(rk1tauf, i, j, k - 3));
						LLFn_m3kphif = 0.5 * (Fluxm3kphif - fabs(Speedkphif) * vector(rk1phif, i, j, k - 3));
						LLFp_m2kSfd_x = 0.5 * (Fluxm2kSfd_x + fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k - 2));
						LLFp_m2kSfd_y = 0.5 * (Fluxm2kSfd_y + fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k - 2));
						LLFp_m2kSfd_z = 0.5 * (Fluxm2kSfd_z + fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k - 2));
						LLFp_m2kBfu_x = 0.5 * (Fluxm2kBfu_x + fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k - 2));
						LLFp_m2kBfu_y = 0.5 * (Fluxm2kBfu_y + fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k - 2));
						LLFp_m2kBfu_z = 0.5 * (Fluxm2kBfu_z + fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k - 2));
						LLFp_m2kDf = 0.5 * (Fluxm2kDf + fabs(SpeedkDf) * vector(rk1Df, i, j, k - 2));
						LLFp_m2kDYf = 0.5 * (Fluxm2kDYf + fabs(SpeedkDYf) * vector(rk1DYf, i, j, k - 2));
						LLFp_m2ktauf = 0.5 * (Fluxm2ktauf + fabs(Speedktauf) * vector(rk1tauf, i, j, k - 2));
						LLFp_m2kphif = 0.5 * (Fluxm2kphif + fabs(Speedkphif) * vector(rk1phif, i, j, k - 2));
						LLFn_m2kSfd_x = 0.5 * (Fluxm2kSfd_x - fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k - 2));
						LLFn_m2kSfd_y = 0.5 * (Fluxm2kSfd_y - fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k - 2));
						LLFn_m2kSfd_z = 0.5 * (Fluxm2kSfd_z - fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k - 2));
						LLFn_m2kBfu_x = 0.5 * (Fluxm2kBfu_x - fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k - 2));
						LLFn_m2kBfu_y = 0.5 * (Fluxm2kBfu_y - fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k - 2));
						LLFn_m2kBfu_z = 0.5 * (Fluxm2kBfu_z - fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k - 2));
						LLFn_m2kDf = 0.5 * (Fluxm2kDf - fabs(SpeedkDf) * vector(rk1Df, i, j, k - 2));
						LLFn_m2kDYf = 0.5 * (Fluxm2kDYf - fabs(SpeedkDYf) * vector(rk1DYf, i, j, k - 2));
						LLFn_m2ktauf = 0.5 * (Fluxm2ktauf - fabs(Speedktauf) * vector(rk1tauf, i, j, k - 2));
						LLFn_m2kphif = 0.5 * (Fluxm2kphif - fabs(Speedkphif) * vector(rk1phif, i, j, k - 2));
						LLFp_m1kSfd_x = 0.5 * (Fluxm1kSfd_x + fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k - 1));
						LLFp_m1kSfd_y = 0.5 * (Fluxm1kSfd_y + fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k - 1));
						LLFp_m1kSfd_z = 0.5 * (Fluxm1kSfd_z + fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k - 1));
						LLFp_m1kBfu_x = 0.5 * (Fluxm1kBfu_x + fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k - 1));
						LLFp_m1kBfu_y = 0.5 * (Fluxm1kBfu_y + fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k - 1));
						LLFp_m1kBfu_z = 0.5 * (Fluxm1kBfu_z + fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k - 1));
						LLFp_m1kDf = 0.5 * (Fluxm1kDf + fabs(SpeedkDf) * vector(rk1Df, i, j, k - 1));
						LLFp_m1kDYf = 0.5 * (Fluxm1kDYf + fabs(SpeedkDYf) * vector(rk1DYf, i, j, k - 1));
						LLFp_m1ktauf = 0.5 * (Fluxm1ktauf + fabs(Speedktauf) * vector(rk1tauf, i, j, k - 1));
						LLFp_m1kphif = 0.5 * (Fluxm1kphif + fabs(Speedkphif) * vector(rk1phif, i, j, k - 1));
						LLFn_m1kSfd_x = 0.5 * (Fluxm1kSfd_x - fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k - 1));
						LLFn_m1kSfd_y = 0.5 * (Fluxm1kSfd_y - fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k - 1));
						LLFn_m1kSfd_z = 0.5 * (Fluxm1kSfd_z - fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k - 1));
						LLFn_m1kBfu_x = 0.5 * (Fluxm1kBfu_x - fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k - 1));
						LLFn_m1kBfu_y = 0.5 * (Fluxm1kBfu_y - fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k - 1));
						LLFn_m1kBfu_z = 0.5 * (Fluxm1kBfu_z - fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k - 1));
						LLFn_m1kDf = 0.5 * (Fluxm1kDf - fabs(SpeedkDf) * vector(rk1Df, i, j, k - 1));
						LLFn_m1kDYf = 0.5 * (Fluxm1kDYf - fabs(SpeedkDYf) * vector(rk1DYf, i, j, k - 1));
						LLFn_m1ktauf = 0.5 * (Fluxm1ktauf - fabs(Speedktauf) * vector(rk1tauf, i, j, k - 1));
						LLFn_m1kphif = 0.5 * (Fluxm1kphif - fabs(Speedkphif) * vector(rk1phif, i, j, k - 1));
						LLFp_p1kSfd_x = 0.5 * (Fluxp1kSfd_x + fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k + 1));
						LLFp_p1kSfd_y = 0.5 * (Fluxp1kSfd_y + fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k + 1));
						LLFp_p1kSfd_z = 0.5 * (Fluxp1kSfd_z + fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k + 1));
						LLFp_p1kBfu_x = 0.5 * (Fluxp1kBfu_x + fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k + 1));
						LLFp_p1kBfu_y = 0.5 * (Fluxp1kBfu_y + fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k + 1));
						LLFp_p1kBfu_z = 0.5 * (Fluxp1kBfu_z + fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k + 1));
						LLFp_p1kDf = 0.5 * (Fluxp1kDf + fabs(SpeedkDf) * vector(rk1Df, i, j, k + 1));
						LLFp_p1kDYf = 0.5 * (Fluxp1kDYf + fabs(SpeedkDYf) * vector(rk1DYf, i, j, k + 1));
						LLFp_p1ktauf = 0.5 * (Fluxp1ktauf + fabs(Speedktauf) * vector(rk1tauf, i, j, k + 1));
						LLFp_p1kphif = 0.5 * (Fluxp1kphif + fabs(Speedkphif) * vector(rk1phif, i, j, k + 1));
						LLFn_p1kSfd_x = 0.5 * (Fluxp1kSfd_x - fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k + 1));
						LLFn_p1kSfd_y = 0.5 * (Fluxp1kSfd_y - fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k + 1));
						LLFn_p1kSfd_z = 0.5 * (Fluxp1kSfd_z - fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k + 1));
						LLFn_p1kBfu_x = 0.5 * (Fluxp1kBfu_x - fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k + 1));
						LLFn_p1kBfu_y = 0.5 * (Fluxp1kBfu_y - fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k + 1));
						LLFn_p1kBfu_z = 0.5 * (Fluxp1kBfu_z - fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k + 1));
						LLFn_p1kDf = 0.5 * (Fluxp1kDf - fabs(SpeedkDf) * vector(rk1Df, i, j, k + 1));
						LLFn_p1kDYf = 0.5 * (Fluxp1kDYf - fabs(SpeedkDYf) * vector(rk1DYf, i, j, k + 1));
						LLFn_p1ktauf = 0.5 * (Fluxp1ktauf - fabs(Speedktauf) * vector(rk1tauf, i, j, k + 1));
						LLFn_p1kphif = 0.5 * (Fluxp1kphif - fabs(Speedkphif) * vector(rk1phif, i, j, k + 1));
						LLFp_p2kSfd_x = 0.5 * (Fluxp2kSfd_x + fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k + 2));
						LLFp_p2kSfd_y = 0.5 * (Fluxp2kSfd_y + fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k + 2));
						LLFp_p2kSfd_z = 0.5 * (Fluxp2kSfd_z + fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k + 2));
						LLFp_p2kBfu_x = 0.5 * (Fluxp2kBfu_x + fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k + 2));
						LLFp_p2kBfu_y = 0.5 * (Fluxp2kBfu_y + fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k + 2));
						LLFp_p2kBfu_z = 0.5 * (Fluxp2kBfu_z + fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k + 2));
						LLFp_p2kDf = 0.5 * (Fluxp2kDf + fabs(SpeedkDf) * vector(rk1Df, i, j, k + 2));
						LLFp_p2kDYf = 0.5 * (Fluxp2kDYf + fabs(SpeedkDYf) * vector(rk1DYf, i, j, k + 2));
						LLFp_p2ktauf = 0.5 * (Fluxp2ktauf + fabs(Speedktauf) * vector(rk1tauf, i, j, k + 2));
						LLFp_p2kphif = 0.5 * (Fluxp2kphif + fabs(Speedkphif) * vector(rk1phif, i, j, k + 2));
						LLFn_p2kSfd_x = 0.5 * (Fluxp2kSfd_x - fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k + 2));
						LLFn_p2kSfd_y = 0.5 * (Fluxp2kSfd_y - fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k + 2));
						LLFn_p2kSfd_z = 0.5 * (Fluxp2kSfd_z - fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k + 2));
						LLFn_p2kBfu_x = 0.5 * (Fluxp2kBfu_x - fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k + 2));
						LLFn_p2kBfu_y = 0.5 * (Fluxp2kBfu_y - fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k + 2));
						LLFn_p2kBfu_z = 0.5 * (Fluxp2kBfu_z - fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k + 2));
						LLFn_p2kDf = 0.5 * (Fluxp2kDf - fabs(SpeedkDf) * vector(rk1Df, i, j, k + 2));
						LLFn_p2kDYf = 0.5 * (Fluxp2kDYf - fabs(SpeedkDYf) * vector(rk1DYf, i, j, k + 2));
						LLFn_p2ktauf = 0.5 * (Fluxp2ktauf - fabs(Speedktauf) * vector(rk1tauf, i, j, k + 2));
						LLFn_p2kphif = 0.5 * (Fluxp2kphif - fabs(Speedkphif) * vector(rk1phif, i, j, k + 2));
						LLFp_p3kSfd_x = 0.5 * (Fluxp3kSfd_x + fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k + 3));
						LLFp_p3kSfd_y = 0.5 * (Fluxp3kSfd_y + fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k + 3));
						LLFp_p3kSfd_z = 0.5 * (Fluxp3kSfd_z + fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k + 3));
						LLFp_p3kBfu_x = 0.5 * (Fluxp3kBfu_x + fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k + 3));
						LLFp_p3kBfu_y = 0.5 * (Fluxp3kBfu_y + fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k + 3));
						LLFp_p3kBfu_z = 0.5 * (Fluxp3kBfu_z + fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k + 3));
						LLFp_p3kDf = 0.5 * (Fluxp3kDf + fabs(SpeedkDf) * vector(rk1Df, i, j, k + 3));
						LLFp_p3kDYf = 0.5 * (Fluxp3kDYf + fabs(SpeedkDYf) * vector(rk1DYf, i, j, k + 3));
						LLFp_p3ktauf = 0.5 * (Fluxp3ktauf + fabs(Speedktauf) * vector(rk1tauf, i, j, k + 3));
						LLFp_p3kphif = 0.5 * (Fluxp3kphif + fabs(Speedkphif) * vector(rk1phif, i, j, k + 3));
						LLFn_p3kSfd_x = 0.5 * (Fluxp3kSfd_x - fabs(SpeedkSfd_x) * vector(rk1Sfd_x, i, j, k + 3));
						LLFn_p3kSfd_y = 0.5 * (Fluxp3kSfd_y - fabs(SpeedkSfd_y) * vector(rk1Sfd_y, i, j, k + 3));
						LLFn_p3kSfd_z = 0.5 * (Fluxp3kSfd_z - fabs(SpeedkSfd_z) * vector(rk1Sfd_z, i, j, k + 3));
						LLFn_p3kBfu_x = 0.5 * (Fluxp3kBfu_x - fabs(SpeedkBfu_x) * vector(rk1Bfu_x, i, j, k + 3));
						LLFn_p3kBfu_y = 0.5 * (Fluxp3kBfu_y - fabs(SpeedkBfu_y) * vector(rk1Bfu_y, i, j, k + 3));
						LLFn_p3kBfu_z = 0.5 * (Fluxp3kBfu_z - fabs(SpeedkBfu_z) * vector(rk1Bfu_z, i, j, k + 3));
						LLFn_p3kDf = 0.5 * (Fluxp3kDf - fabs(SpeedkDf) * vector(rk1Df, i, j, k + 3));
						LLFn_p3kDYf = 0.5 * (Fluxp3kDYf - fabs(SpeedkDYf) * vector(rk1DYf, i, j, k + 3));
						LLFn_p3ktauf = 0.5 * (Fluxp3ktauf - fabs(Speedktauf) * vector(rk1tauf, i, j, k + 3));
						LLFn_p3kphif = 0.5 * (Fluxp3kphif - fabs(Speedkphif) * vector(rk1phif, i, j, k + 3));
						Sfd_x_lni = MP5(LLFp_m3iSfd_x, LLFp_m2iSfd_x, LLFp_m1iSfd_x, LLFpiSfd_x, LLFp_p1iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lni = MP5(LLFp_m3iSfd_y, LLFp_m2iSfd_y, LLFp_m1iSfd_y, LLFpiSfd_y, LLFp_p1iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lni = MP5(LLFp_m3iSfd_z, LLFp_m2iSfd_z, LLFp_m1iSfd_z, LLFpiSfd_z, LLFp_p1iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lni = MP5(LLFp_m3iBfu_x, LLFp_m2iBfu_x, LLFp_m1iBfu_x, LLFpiBfu_x, LLFp_p1iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lni = MP5(LLFp_m3iBfu_y, LLFp_m2iBfu_y, LLFp_m1iBfu_y, LLFpiBfu_y, LLFp_p1iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lni = MP5(LLFp_m3iBfu_z, LLFp_m2iBfu_z, LLFp_m1iBfu_z, LLFpiBfu_z, LLFp_p1iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lni = MP5(LLFp_m3iDf, LLFp_m2iDf, LLFp_m1iDf, LLFpiDf, LLFp_p1iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lni = MP5(LLFp_m3iDYf, LLFp_m2iDYf, LLFp_m1iDYf, LLFpiDYf, LLFp_p1iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lni = MP5(LLFp_m3itauf, LLFp_m2itauf, LLFp_m1itauf, LLFpitauf, LLFp_p1itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lni = MP5(LLFp_m3iphif, LLFp_m2iphif, LLFp_m1iphif, LLFpiphif, LLFp_p1iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lnj = MP5(LLFp_m3jSfd_x, LLFp_m2jSfd_x, LLFp_m1jSfd_x, LLFpjSfd_x, LLFp_p1jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lnj = MP5(LLFp_m3jSfd_y, LLFp_m2jSfd_y, LLFp_m1jSfd_y, LLFpjSfd_y, LLFp_p1jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lnj = MP5(LLFp_m3jSfd_z, LLFp_m2jSfd_z, LLFp_m1jSfd_z, LLFpjSfd_z, LLFp_p1jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lnj = MP5(LLFp_m3jBfu_x, LLFp_m2jBfu_x, LLFp_m1jBfu_x, LLFpjBfu_x, LLFp_p1jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lnj = MP5(LLFp_m3jBfu_y, LLFp_m2jBfu_y, LLFp_m1jBfu_y, LLFpjBfu_y, LLFp_p1jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lnj = MP5(LLFp_m3jBfu_z, LLFp_m2jBfu_z, LLFp_m1jBfu_z, LLFpjBfu_z, LLFp_p1jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lnj = MP5(LLFp_m3jDf, LLFp_m2jDf, LLFp_m1jDf, LLFpjDf, LLFp_p1jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lnj = MP5(LLFp_m3jDYf, LLFp_m2jDYf, LLFp_m1jDYf, LLFpjDYf, LLFp_p1jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lnj = MP5(LLFp_m3jtauf, LLFp_m2jtauf, LLFp_m1jtauf, LLFpjtauf, LLFp_p1jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lnj = MP5(LLFp_m3jphif, LLFp_m2jphif, LLFp_m1jphif, LLFpjphif, LLFp_p1jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lnk = MP5(LLFp_m3kSfd_x, LLFp_m2kSfd_x, LLFp_m1kSfd_x, LLFpkSfd_x, LLFp_p1kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lnk = MP5(LLFp_m3kSfd_y, LLFp_m2kSfd_y, LLFp_m1kSfd_y, LLFpkSfd_y, LLFp_p1kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lnk = MP5(LLFp_m3kSfd_z, LLFp_m2kSfd_z, LLFp_m1kSfd_z, LLFpkSfd_z, LLFp_p1kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lnk = MP5(LLFp_m3kBfu_x, LLFp_m2kBfu_x, LLFp_m1kBfu_x, LLFpkBfu_x, LLFp_p1kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lnk = MP5(LLFp_m3kBfu_y, LLFp_m2kBfu_y, LLFp_m1kBfu_y, LLFpkBfu_y, LLFp_p1kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lnk = MP5(LLFp_m3kBfu_z, LLFp_m2kBfu_z, LLFp_m1kBfu_z, LLFpkBfu_z, LLFp_p1kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lnk = MP5(LLFp_m3kDf, LLFp_m2kDf, LLFp_m1kDf, LLFpkDf, LLFp_p1kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lnk = MP5(LLFp_m3kDYf, LLFp_m2kDYf, LLFp_m1kDYf, LLFpkDYf, LLFp_p1kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lnk = MP5(LLFp_m3ktauf, LLFp_m2ktauf, LLFp_m1ktauf, LLFpktauf, LLFp_p1ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lnk = MP5(LLFp_m3kphif, LLFp_m2kphif, LLFp_m1kphif, LLFpkphif, LLFp_p1kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rni = MP5(LLFn_p2iSfd_x, LLFn_p1iSfd_x, LLFniSfd_x, LLFn_m1iSfd_x, LLFn_m2iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rni = MP5(LLFn_p2iSfd_y, LLFn_p1iSfd_y, LLFniSfd_y, LLFn_m1iSfd_y, LLFn_m2iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rni = MP5(LLFn_p2iSfd_z, LLFn_p1iSfd_z, LLFniSfd_z, LLFn_m1iSfd_z, LLFn_m2iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rni = MP5(LLFn_p2iBfu_x, LLFn_p1iBfu_x, LLFniBfu_x, LLFn_m1iBfu_x, LLFn_m2iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rni = MP5(LLFn_p2iBfu_y, LLFn_p1iBfu_y, LLFniBfu_y, LLFn_m1iBfu_y, LLFn_m2iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rni = MP5(LLFn_p2iBfu_z, LLFn_p1iBfu_z, LLFniBfu_z, LLFn_m1iBfu_z, LLFn_m2iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rni = MP5(LLFn_p2iDf, LLFn_p1iDf, LLFniDf, LLFn_m1iDf, LLFn_m2iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rni = MP5(LLFn_p2iDYf, LLFn_p1iDYf, LLFniDYf, LLFn_m1iDYf, LLFn_m2iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rni = MP5(LLFn_p2itauf, LLFn_p1itauf, LLFnitauf, LLFn_m1itauf, LLFn_m2itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rni = MP5(LLFn_p2iphif, LLFn_p1iphif, LLFniphif, LLFn_m1iphif, LLFn_m2iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rnj = MP5(LLFn_p2jSfd_x, LLFn_p1jSfd_x, LLFnjSfd_x, LLFn_m1jSfd_x, LLFn_m2jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rnj = MP5(LLFn_p2jSfd_y, LLFn_p1jSfd_y, LLFnjSfd_y, LLFn_m1jSfd_y, LLFn_m2jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rnj = MP5(LLFn_p2jSfd_z, LLFn_p1jSfd_z, LLFnjSfd_z, LLFn_m1jSfd_z, LLFn_m2jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rnj = MP5(LLFn_p2jBfu_x, LLFn_p1jBfu_x, LLFnjBfu_x, LLFn_m1jBfu_x, LLFn_m2jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rnj = MP5(LLFn_p2jBfu_y, LLFn_p1jBfu_y, LLFnjBfu_y, LLFn_m1jBfu_y, LLFn_m2jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rnj = MP5(LLFn_p2jBfu_z, LLFn_p1jBfu_z, LLFnjBfu_z, LLFn_m1jBfu_z, LLFn_m2jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rnj = MP5(LLFn_p2jDf, LLFn_p1jDf, LLFnjDf, LLFn_m1jDf, LLFn_m2jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rnj = MP5(LLFn_p2jDYf, LLFn_p1jDYf, LLFnjDYf, LLFn_m1jDYf, LLFn_m2jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rnj = MP5(LLFn_p2jtauf, LLFn_p1jtauf, LLFnjtauf, LLFn_m1jtauf, LLFn_m2jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rnj = MP5(LLFn_p2jphif, LLFn_p1jphif, LLFnjphif, LLFn_m1jphif, LLFn_m2jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rnk = MP5(LLFn_p2kSfd_x, LLFn_p1kSfd_x, LLFnkSfd_x, LLFn_m1kSfd_x, LLFn_m2kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rnk = MP5(LLFn_p2kSfd_y, LLFn_p1kSfd_y, LLFnkSfd_y, LLFn_m1kSfd_y, LLFn_m2kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rnk = MP5(LLFn_p2kSfd_z, LLFn_p1kSfd_z, LLFnkSfd_z, LLFn_m1kSfd_z, LLFn_m2kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rnk = MP5(LLFn_p2kBfu_x, LLFn_p1kBfu_x, LLFnkBfu_x, LLFn_m1kBfu_x, LLFn_m2kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rnk = MP5(LLFn_p2kBfu_y, LLFn_p1kBfu_y, LLFnkBfu_y, LLFn_m1kBfu_y, LLFn_m2kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rnk = MP5(LLFn_p2kBfu_z, LLFn_p1kBfu_z, LLFnkBfu_z, LLFn_m1kBfu_z, LLFn_m2kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rnk = MP5(LLFn_p2kDf, LLFn_p1kDf, LLFnkDf, LLFn_m1kDf, LLFn_m2kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rnk = MP5(LLFn_p2kDYf, LLFn_p1kDYf, LLFnkDYf, LLFn_m1kDYf, LLFn_m2kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rnk = MP5(LLFn_p2ktauf, LLFn_p1ktauf, LLFnktauf, LLFn_m1ktauf, LLFn_m2ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rnk = MP5(LLFn_p2kphif, LLFn_p1kphif, LLFnkphif, LLFn_m1kphif, LLFn_m2kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lpi = MP5(LLFp_m2iSfd_x, LLFp_m1iSfd_x, LLFpiSfd_x, LLFp_p1iSfd_x, LLFp_p2iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lpi = MP5(LLFp_m2iSfd_y, LLFp_m1iSfd_y, LLFpiSfd_y, LLFp_p1iSfd_y, LLFp_p2iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lpi = MP5(LLFp_m2iSfd_z, LLFp_m1iSfd_z, LLFpiSfd_z, LLFp_p1iSfd_z, LLFp_p2iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lpi = MP5(LLFp_m2iBfu_x, LLFp_m1iBfu_x, LLFpiBfu_x, LLFp_p1iBfu_x, LLFp_p2iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lpi = MP5(LLFp_m2iBfu_y, LLFp_m1iBfu_y, LLFpiBfu_y, LLFp_p1iBfu_y, LLFp_p2iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lpi = MP5(LLFp_m2iBfu_z, LLFp_m1iBfu_z, LLFpiBfu_z, LLFp_p1iBfu_z, LLFp_p2iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lpi = MP5(LLFp_m2iDf, LLFp_m1iDf, LLFpiDf, LLFp_p1iDf, LLFp_p2iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lpi = MP5(LLFp_m2iDYf, LLFp_m1iDYf, LLFpiDYf, LLFp_p1iDYf, LLFp_p2iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lpi = MP5(LLFp_m2itauf, LLFp_m1itauf, LLFpitauf, LLFp_p1itauf, LLFp_p2itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lpi = MP5(LLFp_m2iphif, LLFp_m1iphif, LLFpiphif, LLFp_p1iphif, LLFp_p2iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lpj = MP5(LLFp_m2jSfd_x, LLFp_m1jSfd_x, LLFpjSfd_x, LLFp_p1jSfd_x, LLFp_p2jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lpj = MP5(LLFp_m2jSfd_y, LLFp_m1jSfd_y, LLFpjSfd_y, LLFp_p1jSfd_y, LLFp_p2jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lpj = MP5(LLFp_m2jSfd_z, LLFp_m1jSfd_z, LLFpjSfd_z, LLFp_p1jSfd_z, LLFp_p2jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lpj = MP5(LLFp_m2jBfu_x, LLFp_m1jBfu_x, LLFpjBfu_x, LLFp_p1jBfu_x, LLFp_p2jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lpj = MP5(LLFp_m2jBfu_y, LLFp_m1jBfu_y, LLFpjBfu_y, LLFp_p1jBfu_y, LLFp_p2jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lpj = MP5(LLFp_m2jBfu_z, LLFp_m1jBfu_z, LLFpjBfu_z, LLFp_p1jBfu_z, LLFp_p2jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lpj = MP5(LLFp_m2jDf, LLFp_m1jDf, LLFpjDf, LLFp_p1jDf, LLFp_p2jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lpj = MP5(LLFp_m2jDYf, LLFp_m1jDYf, LLFpjDYf, LLFp_p1jDYf, LLFp_p2jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lpj = MP5(LLFp_m2jtauf, LLFp_m1jtauf, LLFpjtauf, LLFp_p1jtauf, LLFp_p2jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lpj = MP5(LLFp_m2jphif, LLFp_m1jphif, LLFpjphif, LLFp_p1jphif, LLFp_p2jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lpk = MP5(LLFp_m2kSfd_x, LLFp_m1kSfd_x, LLFpkSfd_x, LLFp_p1kSfd_x, LLFp_p2kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lpk = MP5(LLFp_m2kSfd_y, LLFp_m1kSfd_y, LLFpkSfd_y, LLFp_p1kSfd_y, LLFp_p2kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lpk = MP5(LLFp_m2kSfd_z, LLFp_m1kSfd_z, LLFpkSfd_z, LLFp_p1kSfd_z, LLFp_p2kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lpk = MP5(LLFp_m2kBfu_x, LLFp_m1kBfu_x, LLFpkBfu_x, LLFp_p1kBfu_x, LLFp_p2kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lpk = MP5(LLFp_m2kBfu_y, LLFp_m1kBfu_y, LLFpkBfu_y, LLFp_p1kBfu_y, LLFp_p2kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lpk = MP5(LLFp_m2kBfu_z, LLFp_m1kBfu_z, LLFpkBfu_z, LLFp_p1kBfu_z, LLFp_p2kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lpk = MP5(LLFp_m2kDf, LLFp_m1kDf, LLFpkDf, LLFp_p1kDf, LLFp_p2kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lpk = MP5(LLFp_m2kDYf, LLFp_m1kDYf, LLFpkDYf, LLFp_p1kDYf, LLFp_p2kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lpk = MP5(LLFp_m2ktauf, LLFp_m1ktauf, LLFpktauf, LLFp_p1ktauf, LLFp_p2ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lpk = MP5(LLFp_m2kphif, LLFp_m1kphif, LLFpkphif, LLFp_p1kphif, LLFp_p2kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rpi = MP5(LLFn_p3iSfd_x, LLFn_p2iSfd_x, LLFn_p1iSfd_x, LLFniSfd_x, LLFn_m1iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rpi = MP5(LLFn_p3iSfd_y, LLFn_p2iSfd_y, LLFn_p1iSfd_y, LLFniSfd_y, LLFn_m1iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rpi = MP5(LLFn_p3iSfd_z, LLFn_p2iSfd_z, LLFn_p1iSfd_z, LLFniSfd_z, LLFn_m1iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rpi = MP5(LLFn_p3iBfu_x, LLFn_p2iBfu_x, LLFn_p1iBfu_x, LLFniBfu_x, LLFn_m1iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rpi = MP5(LLFn_p3iBfu_y, LLFn_p2iBfu_y, LLFn_p1iBfu_y, LLFniBfu_y, LLFn_m1iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rpi = MP5(LLFn_p3iBfu_z, LLFn_p2iBfu_z, LLFn_p1iBfu_z, LLFniBfu_z, LLFn_m1iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rpi = MP5(LLFn_p3iDf, LLFn_p2iDf, LLFn_p1iDf, LLFniDf, LLFn_m1iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rpi = MP5(LLFn_p3iDYf, LLFn_p2iDYf, LLFn_p1iDYf, LLFniDYf, LLFn_m1iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rpi = MP5(LLFn_p3itauf, LLFn_p2itauf, LLFn_p1itauf, LLFnitauf, LLFn_m1itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rpi = MP5(LLFn_p3iphif, LLFn_p2iphif, LLFn_p1iphif, LLFniphif, LLFn_m1iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rpj = MP5(LLFn_p3jSfd_x, LLFn_p2jSfd_x, LLFn_p1jSfd_x, LLFnjSfd_x, LLFn_m1jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rpj = MP5(LLFn_p3jSfd_y, LLFn_p2jSfd_y, LLFn_p1jSfd_y, LLFnjSfd_y, LLFn_m1jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rpj = MP5(LLFn_p3jSfd_z, LLFn_p2jSfd_z, LLFn_p1jSfd_z, LLFnjSfd_z, LLFn_m1jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rpj = MP5(LLFn_p3jBfu_x, LLFn_p2jBfu_x, LLFn_p1jBfu_x, LLFnjBfu_x, LLFn_m1jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rpj = MP5(LLFn_p3jBfu_y, LLFn_p2jBfu_y, LLFn_p1jBfu_y, LLFnjBfu_y, LLFn_m1jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rpj = MP5(LLFn_p3jBfu_z, LLFn_p2jBfu_z, LLFn_p1jBfu_z, LLFnjBfu_z, LLFn_m1jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rpj = MP5(LLFn_p3jDf, LLFn_p2jDf, LLFn_p1jDf, LLFnjDf, LLFn_m1jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rpj = MP5(LLFn_p3jDYf, LLFn_p2jDYf, LLFn_p1jDYf, LLFnjDYf, LLFn_m1jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rpj = MP5(LLFn_p3jtauf, LLFn_p2jtauf, LLFn_p1jtauf, LLFnjtauf, LLFn_m1jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rpj = MP5(LLFn_p3jphif, LLFn_p2jphif, LLFn_p1jphif, LLFnjphif, LLFn_m1jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rpk = MP5(LLFn_p3kSfd_x, LLFn_p2kSfd_x, LLFn_p1kSfd_x, LLFnkSfd_x, LLFn_m1kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rpk = MP5(LLFn_p3kSfd_y, LLFn_p2kSfd_y, LLFn_p1kSfd_y, LLFnkSfd_y, LLFn_m1kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rpk = MP5(LLFn_p3kSfd_z, LLFn_p2kSfd_z, LLFn_p1kSfd_z, LLFnkSfd_z, LLFn_m1kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rpk = MP5(LLFn_p3kBfu_x, LLFn_p2kBfu_x, LLFn_p1kBfu_x, LLFnkBfu_x, LLFn_m1kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rpk = MP5(LLFn_p3kBfu_y, LLFn_p2kBfu_y, LLFn_p1kBfu_y, LLFnkBfu_y, LLFn_m1kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rpk = MP5(LLFn_p3kBfu_z, LLFn_p2kBfu_z, LLFn_p1kBfu_z, LLFnkBfu_z, LLFn_m1kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rpk = MP5(LLFn_p3kDf, LLFn_p2kDf, LLFn_p1kDf, LLFnkDf, LLFn_m1kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rpk = MP5(LLFn_p3kDYf, LLFn_p2kDYf, LLFn_p1kDYf, LLFnkDYf, LLFn_m1kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rpk = MP5(LLFn_p3ktauf, LLFn_p2ktauf, LLFn_p1ktauf, LLFnktauf, LLFn_m1ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rpk = MP5(LLFn_p3kphif, LLFn_p2kphif, LLFn_p1kphif, LLFnkphif, LLFn_m1kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						FluxniSfd_x = Sfd_x_lni + Sfd_x_rni;
						FluxniSfd_y = Sfd_y_lni + Sfd_y_rni;
						FluxniSfd_z = Sfd_z_lni + Sfd_z_rni;
						FluxniBfu_x = Bfu_x_lni + Bfu_x_rni;
						FluxniBfu_y = Bfu_y_lni + Bfu_y_rni;
						FluxniBfu_z = Bfu_z_lni + Bfu_z_rni;
						FluxniDf = Df_lni + Df_rni;
						FluxniDYf = DYf_lni + DYf_rni;
						Fluxnitauf = tauf_lni + tauf_rni;
						Fluxniphif = phif_lni + phif_rni;
						FluxnjSfd_x = Sfd_x_lnj + Sfd_x_rnj;
						FluxnjSfd_y = Sfd_y_lnj + Sfd_y_rnj;
						FluxnjSfd_z = Sfd_z_lnj + Sfd_z_rnj;
						FluxnjBfu_x = Bfu_x_lnj + Bfu_x_rnj;
						FluxnjBfu_y = Bfu_y_lnj + Bfu_y_rnj;
						FluxnjBfu_z = Bfu_z_lnj + Bfu_z_rnj;
						FluxnjDf = Df_lnj + Df_rnj;
						FluxnjDYf = DYf_lnj + DYf_rnj;
						Fluxnjtauf = tauf_lnj + tauf_rnj;
						Fluxnjphif = phif_lnj + phif_rnj;
						FluxnkSfd_x = Sfd_x_lnk + Sfd_x_rnk;
						FluxnkSfd_y = Sfd_y_lnk + Sfd_y_rnk;
						FluxnkSfd_z = Sfd_z_lnk + Sfd_z_rnk;
						FluxnkBfu_x = Bfu_x_lnk + Bfu_x_rnk;
						FluxnkBfu_y = Bfu_y_lnk + Bfu_y_rnk;
						FluxnkBfu_z = Bfu_z_lnk + Bfu_z_rnk;
						FluxnkDf = Df_lnk + Df_rnk;
						FluxnkDYf = DYf_lnk + DYf_rnk;
						Fluxnktauf = tauf_lnk + tauf_rnk;
						Fluxnkphif = phif_lnk + phif_rnk;
						FluxpiSfd_x = Sfd_x_lpi + Sfd_x_rpi;
						FluxpiSfd_y = Sfd_y_lpi + Sfd_y_rpi;
						FluxpiSfd_z = Sfd_z_lpi + Sfd_z_rpi;
						FluxpiBfu_x = Bfu_x_lpi + Bfu_x_rpi;
						FluxpiBfu_y = Bfu_y_lpi + Bfu_y_rpi;
						FluxpiBfu_z = Bfu_z_lpi + Bfu_z_rpi;
						FluxpiDf = Df_lpi + Df_rpi;
						FluxpiDYf = DYf_lpi + DYf_rpi;
						Fluxpitauf = tauf_lpi + tauf_rpi;
						Fluxpiphif = phif_lpi + phif_rpi;
						FluxpjSfd_x = Sfd_x_lpj + Sfd_x_rpj;
						FluxpjSfd_y = Sfd_y_lpj + Sfd_y_rpj;
						FluxpjSfd_z = Sfd_z_lpj + Sfd_z_rpj;
						FluxpjBfu_x = Bfu_x_lpj + Bfu_x_rpj;
						FluxpjBfu_y = Bfu_y_lpj + Bfu_y_rpj;
						FluxpjBfu_z = Bfu_z_lpj + Bfu_z_rpj;
						FluxpjDf = Df_lpj + Df_rpj;
						FluxpjDYf = DYf_lpj + DYf_rpj;
						Fluxpjtauf = tauf_lpj + tauf_rpj;
						Fluxpjphif = phif_lpj + phif_rpj;
						FluxpkSfd_x = Sfd_x_lpk + Sfd_x_rpk;
						FluxpkSfd_y = Sfd_y_lpk + Sfd_y_rpk;
						FluxpkSfd_z = Sfd_z_lpk + Sfd_z_rpk;
						FluxpkBfu_x = Bfu_x_lpk + Bfu_x_rpk;
						FluxpkBfu_y = Bfu_y_lpk + Bfu_y_rpk;
						FluxpkBfu_z = Bfu_z_lpk + Bfu_z_rpk;
						FluxpkDf = Df_lpk + Df_rpk;
						FluxpkDYf = DYf_lpk + DYf_rpk;
						Fluxpktauf = tauf_lpk + tauf_rpk;
						Fluxpkphif = phif_lpk + phif_rpk;
						d_div_Beta_o0_t0_m0_l0 = D1CDO4_i(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o1_t0_m0_l0 = lieforward_i(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o2_t0_m0_l0 = liebackward_i(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_y_o0_t7_m0_l0 = D1CDO4_j(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o1_t1_m0_l0 = lieforward_j(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o2_t1_m0_l0 = liebackward_j(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_z_o0_t7_m0_l0 = D1CDO4_k(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o1_t2_m0_l0 = lieforward_k(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o2_t2_m0_l0 = liebackward_k(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_x_o0_t0_m0_l0 = D2CDO4_i(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_y_o0_t0_m0_l0 = D1CDO4crossed_ij(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_z_o0_t0_m0_l0 = D1CDO4crossed_ik(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t15_m0_l0 = D2CDO4_j(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t16_m0_l0 = D1CDO4crossed_jk(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t17_m0_l0 = D2CDO4_k(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_x_o0_t8_m0_l0 = D1CDO4_i(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o1_t0_m0_l0 = lieforward_i(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o2_t0_m0_l0 = liebackward_i(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_div_Beta_o0_t1_m0_l0 = D1CDO4_j(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o1_t1_m0_l0 = lieforward_j(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o2_t1_m0_l0 = liebackward_j(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_z_o0_t8_m0_l0 = D1CDO4_k(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o1_t2_m0_l0 = lieforward_k(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o2_t2_m0_l0 = liebackward_k(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o0_t6_m0_l0 = D2CDO4_i(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_x_o0_t1_m0_l0 = D1CDO4crossed_ij(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o0_t8_m0_l0 = D1CDO4crossed_ik(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_y_o0_t1_m0_l0 = D2CDO4_j(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_z_o0_t1_m0_l0 = D1CDO4crossed_jk(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o0_t11_m0_l0 = D2CDO4_k(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_x_o0_t9_m0_l0 = D1CDO4_i(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o1_t0_m0_l0 = lieforward_i(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o2_t0_m0_l0 = liebackward_i(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_y_o0_t9_m0_l0 = D1CDO4_j(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o1_t1_m0_l0 = lieforward_j(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o2_t1_m0_l0 = liebackward_j(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_div_Beta_o0_t2_m0_l0 = D1CDO4_k(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o1_t2_m0_l0 = lieforward_k(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o2_t2_m0_l0 = liebackward_k(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o0_t0_m0_l0 = D2CDO4_i(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o0_t1_m0_l0 = D1CDO4crossed_ij(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_x_o0_t2_m0_l0 = D1CDO4crossed_ik(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o0_t3_m0_l0 = D2CDO4_j(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_y_o0_t2_m0_l0 = D1CDO4crossed_jk(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_z_o0_t2_m0_l0 = D2CDO4_k(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xxx_o0_t0_m0_l0 = D1CDO4_i(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o1_t0_m0_l0 = lieforward_i(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o2_t0_m0_l0 = liebackward_i(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xxy_o0_t0_m0_l0 = D1CDO4_j(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o1_t1_m0_l0 = lieforward_j(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o2_t1_m0_l0 = liebackward_j(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xxz_o0_t0_m0_l0 = D1CDO4_k(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o1_t2_m0_l0 = lieforward_k(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o2_t2_m0_l0 = liebackward_k(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yxx_o0_t0_m0_l0 = D1CDO4_i(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o1_t0_m0_l0 = lieforward_i(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o2_t0_m0_l0 = liebackward_i(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyy_o0_t0_m0_l0 = D1CDO4_j(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o1_t1_m0_l0 = lieforward_j(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o2_t1_m0_l0 = liebackward_j(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyz_o0_t1_m0_l0 = D1CDO4_k(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o1_t2_m0_l0 = lieforward_k(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o2_t2_m0_l0 = liebackward_k(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyy_o0_t1_m0_l0 = D1CDO4_i(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o1_t0_m0_l0 = lieforward_i(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o2_t0_m0_l0 = liebackward_i(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yyy_o0_t0_m0_l0 = D1CDO4_j(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o1_t1_m0_l0 = lieforward_j(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o2_t1_m0_l0 = liebackward_j(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yyz_o0_t0_m0_l0 = D1CDO4_k(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o1_t2_m0_l0 = lieforward_k(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o2_t2_m0_l0 = liebackward_k(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_zxx_o0_t0_m0_l0 = D1CDO4_i(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o1_t0_m0_l0 = lieforward_i(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o2_t0_m0_l0 = liebackward_i(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyz_o0_t0_m0_l0 = D1CDO4_j(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o1_t1_m0_l0 = lieforward_j(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o2_t1_m0_l0 = liebackward_j(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xzz_o0_t0_m0_l0 = D1CDO4_k(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o1_t2_m0_l0 = lieforward_k(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o2_t2_m0_l0 = liebackward_k(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyz_o0_t2_m0_l0 = D1CDO4_i(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o1_t0_m0_l0 = lieforward_i(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o2_t0_m0_l0 = liebackward_i(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_zyy_o0_t0_m0_l0 = D1CDO4_j(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o1_t1_m0_l0 = lieforward_j(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o2_t1_m0_l0 = liebackward_j(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yzz_o0_t0_m0_l0 = D1CDO4_k(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o1_t2_m0_l0 = lieforward_k(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o2_t2_m0_l0 = liebackward_k(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xzz_o0_t1_m0_l0 = D1CDO4_i(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o1_t0_m0_l0 = lieforward_i(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o2_t0_m0_l0 = liebackward_i(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yzz_o0_t1_m0_l0 = D1CDO4_j(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o1_t1_m0_l0 = lieforward_j(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o2_t1_m0_l0 = liebackward_j(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_zzz_o0_t0_m0_l0 = D1CDO4_k(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o1_t2_m0_l0 = lieforward_k(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o2_t2_m0_l0 = liebackward_k(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t0_m0_l0 = D1CDO4_i(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o1_t0_m0_l0 = lieforward_i(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o2_t0_m0_l0 = liebackward_i(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t1_m0_l0 = D1CDO4_j(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o1_t1_m0_l0 = lieforward_j(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o2_t1_m0_l0 = liebackward_j(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t2_m0_l0 = D1CDO4_k(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o1_t2_m0_l0 = lieforward_k(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o2_t2_m0_l0 = liebackward_k(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t6_m0_l0 = D1CDO4_i(rk1Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o1_t0_m0_l0 = liebackward_i(rk1Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o2_t0_m0_l0 = lieforward_i(rk1Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t6_m0_l0 = D1CDO4_j(rk1Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o1_t1_m0_l0 = liebackward_j(rk1Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o2_t1_m0_l0 = lieforward_j(rk1Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t6_m0_l0 = D1CDO4_k(rk1Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o1_t2_m0_l0 = liebackward_k(rk1Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o2_t2_m0_l0 = lieforward_k(rk1Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t7_m0_l0 = D1CDO4_i(rk1Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o1_t0_m0_l0 = liebackward_i(rk1Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o2_t0_m0_l0 = lieforward_i(rk1Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t7_m0_l0 = D1CDO4_j(rk1Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o1_t1_m0_l0 = liebackward_j(rk1Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o2_t1_m0_l0 = lieforward_j(rk1Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t7_m0_l0 = D1CDO4_k(rk1Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o1_t2_m0_l0 = liebackward_k(rk1Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o2_t2_m0_l0 = lieforward_k(rk1Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t8_m0_l0 = D1CDO4_i(rk1Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o1_t0_m0_l0 = liebackward_i(rk1Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o2_t0_m0_l0 = lieforward_i(rk1Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t8_m0_l0 = D1CDO4_j(rk1Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o1_t1_m0_l0 = liebackward_j(rk1Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o2_t1_m0_l0 = lieforward_j(rk1Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t8_m0_l0 = D1CDO4_k(rk1Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o1_t2_m0_l0 = liebackward_k(rk1Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o2_t2_m0_l0 = lieforward_k(rk1Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t1_m0_l0 = D1CDO4_i(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o1_t0_m0_l0 = lieforward_i(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o2_t0_m0_l0 = liebackward_i(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t2_m0_l0 = D1CDO4_j(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o1_t1_m0_l0 = lieforward_j(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o2_t1_m0_l0 = liebackward_j(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t3_m0_l0 = D1CDO4_k(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o1_t2_m0_l0 = lieforward_k(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o2_t2_m0_l0 = liebackward_k(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t0_m0_l0 = D2CDO4_i(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xy_o0_t0_m0_l0 = D1CDO4crossed_ij(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xz_o0_t0_m0_l0 = D1CDO4crossed_ik(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_yy_o0_t0_m0_l0 = D2CDO4_j(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_yz_o0_t0_m0_l0 = D1CDO4crossed_jk(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_zz_o0_t0_m0_l0 = D2CDO4_k(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o1_t0_m0_l0 = lieforward_i(rk1Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o2_t0_m0_l0 = liebackward_i(rk1Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o1_t1_m0_l0 = lieforward_j(rk1Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o2_t1_m0_l0 = liebackward_j(rk1Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o1_t2_m0_l0 = lieforward_k(rk1Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o2_t2_m0_l0 = liebackward_k(rk1Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o1_t0_m0_l0 = lieforward_i(rk1Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o2_t0_m0_l0 = liebackward_i(rk1Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o1_t1_m0_l0 = lieforward_j(rk1Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o2_t1_m0_l0 = liebackward_j(rk1Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o1_t2_m0_l0 = lieforward_k(rk1Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o2_t2_m0_l0 = liebackward_k(rk1Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o1_t0_m0_l0 = lieforward_i(rk1Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o2_t0_m0_l0 = liebackward_i(rk1Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o1_t1_m0_l0 = lieforward_j(rk1Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o2_t1_m0_l0 = liebackward_j(rk1Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o1_t2_m0_l0 = lieforward_k(rk1Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o2_t2_m0_l0 = liebackward_k(rk1Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o1_t0_m0_l0 = lieforward_i(rk1Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o2_t0_m0_l0 = liebackward_i(rk1Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o1_t1_m0_l0 = lieforward_j(rk1Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o2_t1_m0_l0 = liebackward_j(rk1Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o1_t2_m0_l0 = lieforward_k(rk1Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o2_t2_m0_l0 = liebackward_k(rk1Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o1_t0_m0_l0 = lieforward_i(rk1Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o2_t0_m0_l0 = liebackward_i(rk1Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o1_t1_m0_l0 = lieforward_j(rk1Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o2_t1_m0_l0 = liebackward_j(rk1Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o1_t2_m0_l0 = lieforward_k(rk1Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o2_t2_m0_l0 = liebackward_k(rk1Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o1_t0_m0_l0 = lieforward_i(rk1Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o2_t0_m0_l0 = liebackward_i(rk1Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o1_t1_m0_l0 = lieforward_j(rk1Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o2_t1_m0_l0 = liebackward_j(rk1Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o1_t2_m0_l0 = lieforward_k(rk1Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o2_t2_m0_l0 = liebackward_k(rk1Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t0_m0_l0 = D1CDO4_i(rk1trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o1_t0_m0_l0 = liebackward_i(rk1trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o2_t0_m0_l0 = lieforward_i(rk1trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t1_m0_l0 = D1CDO4_j(rk1trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o1_t1_m0_l0 = liebackward_j(rk1trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o2_t1_m0_l0 = lieforward_j(rk1trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t2_m0_l0 = D1CDO4_k(rk1trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o1_t2_m0_l0 = liebackward_k(rk1trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o2_t2_m0_l0 = lieforward_k(rk1trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t3_m0_l0 = D1CDO4_i(rk1theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o1_t0_m0_l0 = lieforward_i(rk1theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o2_t0_m0_l0 = liebackward_i(rk1theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t4_m0_l0 = D1CDO4_j(rk1theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o1_t1_m0_l0 = lieforward_j(rk1theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o2_t1_m0_l0 = liebackward_j(rk1theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t5_m0_l0 = D1CDO4_k(rk1theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o1_t2_m0_l0 = lieforward_k(rk1theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o2_t2_m0_l0 = liebackward_k(rk1theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t0_m0_l0 = D2CDO4_i(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t1_m0_l0 = D1CDO4crossed_ij(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t2_m0_l0 = D1CDO4crossed_ik(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t3_m0_l0 = D2CDO4_j(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t4_m0_l0 = D1CDO4crossed_jk(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t5_m0_l0 = D2CDO4_k(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t0_m0_l0 = D2CDO4_i(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t1_m0_l0 = D1CDO4crossed_ij(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t2_m0_l0 = D1CDO4crossed_ik(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t3_m0_l0 = D2CDO4_j(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t4_m0_l0 = D1CDO4crossed_jk(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t5_m0_l0 = D2CDO4_k(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t0_m0_l0 = D2CDO4_i(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t1_m0_l0 = D1CDO4crossed_ij(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t2_m0_l0 = D1CDO4crossed_ik(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t3_m0_l0 = D2CDO4_j(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t4_m0_l0 = D1CDO4crossed_jk(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t5_m0_l0 = D2CDO4_k(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t0_m0_l0 = D2CDO4_i(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t1_m0_l0 = D1CDO4crossed_ij(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t2_m0_l0 = D1CDO4crossed_ik(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t3_m0_l0 = D2CDO4_j(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t4_m0_l0 = D1CDO4crossed_jk(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t5_m0_l0 = D2CDO4_k(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t0_m0_l0 = D2CDO4_i(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t1_m0_l0 = D1CDO4crossed_ij(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t2_m0_l0 = D1CDO4crossed_ik(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t3_m0_l0 = D2CDO4_j(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t4_m0_l0 = D1CDO4crossed_jk(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t5_m0_l0 = D2CDO4_k(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t0_m0_l0 = D2CDO4_i(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t1_m0_l0 = D1CDO4crossed_ij(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t2_m0_l0 = D1CDO4crossed_ik(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t3_m0_l0 = D2CDO4_j(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t4_m0_l0 = D1CDO4crossed_jk(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t5_m0_l0 = D2CDO4_k(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t3_m0_l0 = D2CDO4_i(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t4_m0_l0 = D1CDO4crossed_ij(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t5_m0_l0 = D1CDO4crossed_ik(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t6_m0_l0 = D2CDO4_j(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t7_m0_l0 = D1CDO4crossed_jk(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t8_m0_l0 = D2CDO4_k(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						rk1m_Ctd_yyz_o0_t0_l0 = 0.5 * d_Ctd_yyz_o0_t0_m0_l0;
						rk1Ctd_yyz = rk1m_Ctd_yyz_o0_t0_l0;
						rk1m_Ctd_yyy_o0_t0_l0 = 0.5 * d_Ctd_yyy_o0_t0_m0_l0;
						rk1Ctd_yyy = rk1m_Ctd_yyy_o0_t0_l0;
						rk1div_Beta = (d_div_Beta_o0_t0_m0_l0 + d_div_Beta_o0_t1_m0_l0) + d_div_Beta_o0_t2_m0_l0;
						rk1m_Ctd_xxz_o0_t0_l0 = 0.5 * d_Ctd_xxz_o0_t0_m0_l0;
						rk1Ctd_xxz = rk1m_Ctd_xxz_o0_t0_l0;
						rk1m_Ctd_xxx_o0_t0_l0 = 0.5 * d_Ctd_xxx_o0_t0_m0_l0;
						rk1Ctd_xxx = rk1m_Ctd_xxx_o0_t0_l0;
						rk1m_Ctd_yzz_o0_t1_l0 = (-0.5) * d_Ctd_yzz_o0_t1_m0_l0;
						rk1Ctd_yzz = d_Ctd_yzz_o0_t0_m0_l0 + rk1m_Ctd_yzz_o0_t1_l0;
						rk1m_Ctd_xxy_o0_t0_l0 = 0.5 * d_Ctd_xxy_o0_t0_m0_l0;
						rk1Ctd_xxy = rk1m_Ctd_xxy_o0_t0_l0;
						rk1kappa_z2 = p_kappa_z2;
						rk1m_Ctd_xyy_o0_t1_l0 = (-0.5) * d_Ctd_xyy_o0_t1_m0_l0;
						rk1Ctd_xyy = d_Ctd_xyy_o0_t0_m0_l0 + rk1m_Ctd_xyy_o0_t1_l0;
						rk1m_Ctd_xyz_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t2_m0_l0;
						rk1m_Ctd_xyz_o0_t1_l0 = 0.5 * d_Ctd_xyz_o0_t1_m0_l0;
						rk1m_Ctd_xyz_o0_t0_l0 = 0.5 * d_Ctd_xyz_o0_t0_m0_l0;
						rk1Ctd_xyz = (rk1m_Ctd_xyz_o0_t0_l0 + rk1m_Ctd_xyz_o0_t1_l0) + rk1m_Ctd_xyz_o0_t2_l0;
						rk1d_div_Beta_y = (d_d_div_Beta_y_o0_t0_m0_l0 + d_d_div_Beta_y_o0_t1_m0_l0) + d_d_div_Beta_y_o0_t2_m0_l0;
						rk1d_div_Beta_x = (d_d_div_Beta_x_o0_t0_m0_l0 + d_d_div_Beta_x_o0_t1_m0_l0) + d_d_div_Beta_x_o0_t2_m0_l0;
						rk1d_div_Beta_z = (d_d_div_Beta_z_o0_t0_m0_l0 + d_d_div_Beta_z_o0_t1_m0_l0) + d_d_div_Beta_z_o0_t2_m0_l0;
						rk1m_Ctd_xzz_o0_t1_l0 = (-0.5) * d_Ctd_xzz_o0_t1_m0_l0;
						rk1Ctd_xzz = d_Ctd_xzz_o0_t0_m0_l0 + rk1m_Ctd_xzz_o0_t1_l0;
						rk1m_Ctd_zxz_o0_t0_l0 = 0.5 * d_Ctd_xzz_o0_t1_m0_l0;
						rk1Ctd_zxz = rk1m_Ctd_zxz_o0_t0_l0;
						rk1m_Ctd_zxx_o0_t1_l0 = (-0.5) * d_Ctd_xxz_o0_t0_m0_l0;
						rk1Ctd_zxx = d_Ctd_zxx_o0_t0_m0_l0 + rk1m_Ctd_zxx_o0_t1_l0;
						rk1detgtd = vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) * vector(rk1gtd_xx, i, j, k) + (-vector(rk1gtd_yy, i, j, k) * (vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k))) + (-vector(rk1gtd_zz, i, j, k) * (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k))) + (-(vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k)) * vector(rk1gtd_xx, i, j, k)) + 2.0 * vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k);
						rk1chi_max = MAX(chi_floor, vector(rk1chi, i, j, k));
						rk1inv_chi = 1.0 / fabs(rk1chi_max);
						rk1Bfd_y = rk1inv_chi * (vector(rk1gtd_xy, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_yy, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1Bfd_z = rk1inv_chi * (vector(rk1gtd_xz, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_yz, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_zz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1Bfd_x = rk1inv_chi * (vector(rk1gtd_xx, i, j, k) * vector(rk1Bfu_x, i, j, k) + vector(rk1gtd_xy, i, j, k) * vector(rk1Bfu_y, i, j, k) + vector(rk1gtd_xz, i, j, k) * vector(rk1Bfu_z, i, j, k));
						rk1sqBf = vector(rk1Bfu_x, i, j, k) * rk1Bfd_x + vector(rk1Bfu_y, i, j, k) * rk1Bfd_y + vector(rk1Bfu_z, i, j, k) * rk1Bfd_z;
						rk1m_Ctd_zyy_o0_t1_l0 = (-0.5) * d_Ctd_yyz_o0_t0_m0_l0;
						rk1Ctd_zyy = d_Ctd_zyy_o0_t0_m0_l0 + rk1m_Ctd_zyy_o0_t1_l0;
						rk1m_Ctd_zyz_o0_t0_l0 = 0.5 * d_Ctd_yzz_o0_t1_m0_l0;
						rk1Ctd_zyz = rk1m_Ctd_zyz_o0_t0_l0;
						rk1sdetg = pow(rk1chi_max, (-1.500000000000000));
						rk1Tf4u_tt = (vector(rk1Df, i, j, k) + vector(rk1tauf, i, j, k)) / (rk1sdetg * (vector(rk1Alpha, i, j, k) * vector(rk1Alpha, i, j, k)));
						rk1Betatd_x = vector(rk1Betau_x, i, j, k) * vector(rk1gtd_xx, i, j, k) + vector(rk1Betau_y, i, j, k) * vector(rk1gtd_xy, i, j, k) + vector(rk1Betau_z, i, j, k) * vector(rk1gtd_xz, i, j, k);
						rk1Betatd_z = vector(rk1Betau_x, i, j, k) * vector(rk1gtd_xz, i, j, k) + vector(rk1Betau_y, i, j, k) * vector(rk1gtd_yz, i, j, k) + vector(rk1Betau_z, i, j, k) * vector(rk1gtd_zz, i, j, k);
						rk1Betatd_y = vector(rk1Betau_x, i, j, k) * vector(rk1gtd_xy, i, j, k) + vector(rk1Betau_y, i, j, k) * vector(rk1gtd_yy, i, j, k) + vector(rk1Betau_z, i, j, k) * vector(rk1gtd_yz, i, j, k);
						rk1idetgtd = 1.0 / rk1detgtd;
						rk1gtu_zz = rk1idetgtd * (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yy, i, j, k) - vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k));
						rk1gtu_yy = rk1idetgtd * (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_zz, i, j, k) - vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1gtu_yz = rk1idetgtd * ((-vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yz, i, j, k)) + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1gtu_xx = rk1idetgtd * (vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) - vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k));
						rk1gtu_xz = rk1idetgtd * (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_yz, i, j, k) - vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1Ct_zyz = rk1gtu_xz * rk1Ctd_xyz + rk1gtu_yz * rk1Ctd_yyz + rk1gtu_zz * rk1Ctd_zyz;
						rk1Ct_zyy = rk1gtu_xz * rk1Ctd_xyy + rk1gtu_yz * rk1Ctd_yyy + rk1gtu_zz * rk1Ctd_zyy;
						rk1Atud_zx = rk1gtu_xz * vector(rk1Atd_xx, i, j, k) + rk1gtu_yz * vector(rk1Atd_xy, i, j, k) + rk1gtu_zz * vector(rk1Atd_xz, i, j, k);
						rk1Atud_zz = rk1gtu_xz * vector(rk1Atd_xz, i, j, k) + rk1gtu_yz * vector(rk1Atd_yz, i, j, k) + rk1gtu_zz * vector(rk1Atd_zz, i, j, k);
						rk1Atud_zy = rk1gtu_xz * vector(rk1Atd_xy, i, j, k) + rk1gtu_yz * vector(rk1Atd_yy, i, j, k) + rk1gtu_zz * vector(rk1Atd_yz, i, j, k);
						rk1Atu_zz = rk1Atud_zx * rk1gtu_xz + rk1Atud_zy * rk1gtu_yz + rk1Atud_zz * rk1gtu_zz;
						rk1vfu_z = rk1chi_max * (rk1gtu_xz * vector(vfd_x, i, j, k) + rk1gtu_yz * vector(vfd_y, i, j, k) + rk1gtu_zz * vector(vfd_z, i, j, k));
						rk1Sfu_z = rk1chi_max * (rk1gtu_xz * vector(rk1Sfd_x, i, j, k) + rk1gtu_yz * vector(rk1Sfd_y, i, j, k) + rk1gtu_zz * vector(rk1Sfd_z, i, j, k));
						rk1Tf4u_tz = rk1Sfu_z / (rk1sdetg * vector(rk1Alpha, i, j, k)) + (-vector(rk1Betau_z, i, j, k) * rk1Tf4u_tt);
						rk1Tu_tz = rk1Tf4u_tz;
						rk1gtu_xy = rk1idetgtd * ((-vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_zz, i, j, k)) + vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xz, i, j, k));
						rk1vfu_y = rk1chi_max * (rk1gtu_xy * vector(vfd_x, i, j, k) + rk1gtu_yy * vector(vfd_y, i, j, k) + rk1gtu_yz * vector(vfd_z, i, j, k));
						rk1vfu_x = rk1chi_max * (rk1gtu_xx * vector(vfd_x, i, j, k) + rk1gtu_xy * vector(vfd_y, i, j, k) + rk1gtu_xz * vector(vfd_z, i, j, k));
						rk1sqvf = MIN(rk1vfu_x * vector(vfd_x, i, j, k) + rk1vfu_y * vector(vfd_y, i, j, k) + rk1vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk1sqW = 1.0 / (1.0 - rk1sqvf);
						rk1Atud_yy = rk1gtu_xy * vector(rk1Atd_xy, i, j, k) + rk1gtu_yy * vector(rk1Atd_yy, i, j, k) + rk1gtu_yz * vector(rk1Atd_yz, i, j, k);
						rk1Atud_yx = rk1gtu_xy * vector(rk1Atd_xx, i, j, k) + rk1gtu_yy * vector(rk1Atd_xy, i, j, k) + rk1gtu_yz * vector(rk1Atd_xz, i, j, k);
						rk1Atud_yz = rk1gtu_xy * vector(rk1Atd_xz, i, j, k) + rk1gtu_yy * vector(rk1Atd_yz, i, j, k) + rk1gtu_yz * vector(rk1Atd_zz, i, j, k);
						rk1Atu_yz = rk1Atud_yx * rk1gtu_xz + rk1Atud_yy * rk1gtu_yz + rk1Atud_yz * rk1gtu_zz;
						rk1Atu_yy = rk1Atud_yx * rk1gtu_xy + rk1Atud_yy * rk1gtu_yy + rk1Atud_yz * rk1gtu_yz;
						rk1Ct_yyy = rk1gtu_xy * rk1Ctd_xyy + rk1gtu_yy * rk1Ctd_yyy + rk1gtu_yz * rk1Ctd_zyy;
						rk1Ct_yyz = rk1gtu_xy * rk1Ctd_xyz + rk1gtu_yy * rk1Ctd_yyz + rk1gtu_yz * rk1Ctd_zyz;
						rk1Atud_xx = rk1gtu_xx * vector(rk1Atd_xx, i, j, k) + rk1gtu_xy * vector(rk1Atd_xy, i, j, k) + rk1gtu_xz * vector(rk1Atd_xz, i, j, k);
						rk1trAt = rk1Atud_xx + rk1Atud_yy + rk1Atud_zz;
						rk1Atud_xz = rk1gtu_xx * vector(rk1Atd_xz, i, j, k) + rk1gtu_xy * vector(rk1Atd_yz, i, j, k) + rk1gtu_xz * vector(rk1Atd_zz, i, j, k);
						rk1Atud_xy = rk1gtu_xx * vector(rk1Atd_xy, i, j, k) + rk1gtu_xy * vector(rk1Atd_yy, i, j, k) + rk1gtu_xz * vector(rk1Atd_yz, i, j, k);
						rk1Atu_xz = rk1Atud_xx * rk1gtu_xz + rk1Atud_xy * rk1gtu_yz + rk1Atud_xz * rk1gtu_zz;
						rk1Atu_xy = rk1Atud_xx * rk1gtu_xy + rk1Atud_xy * rk1gtu_yy + rk1Atud_xz * rk1gtu_yz;
						rk1Atu_xx = rk1Atud_xx * rk1gtu_xx + rk1Atud_xy * rk1gtu_xy + rk1Atud_xz * rk1gtu_xz;
						rk1Sfu_y = rk1chi_max * (rk1gtu_xy * vector(rk1Sfd_x, i, j, k) + rk1gtu_yy * vector(rk1Sfd_y, i, j, k) + rk1gtu_yz * vector(rk1Sfd_z, i, j, k));
						rk1Tf4u_ty = rk1Sfu_y / (rk1sdetg * vector(rk1Alpha, i, j, k)) + (-vector(rk1Betau_y, i, j, k) * rk1Tf4u_tt);
						rk1Tu_ty = rk1Tf4u_ty;
						rk1Sfu_x = rk1chi_max * (rk1gtu_xx * vector(rk1Sfd_x, i, j, k) + rk1gtu_xy * vector(rk1Sfd_y, i, j, k) + rk1gtu_xz * vector(rk1Sfd_z, i, j, k));
						rk1Tf4u_tx = rk1Sfu_x / (rk1sdetg * vector(rk1Alpha, i, j, k)) + (-vector(rk1Betau_x, i, j, k) * rk1Tf4u_tt);
						rk1Tu_tx = rk1Tf4u_tx;
						rk1decay_factor = MIN(1.0, pow((R_0 / MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))), eta_damping_exp));
						rk1kappa_z1 = MIN(p_kappa_z1, p_kappa_z1 * rk1decay_factor) / vector(rk1Alpha, i, j, k);
						rk1kappa_f = MIN(p_kappa_f, p_kappa_f * rk1decay_factor);
						rk1feta = MIN(p_feta, p_feta * rk1decay_factor);
						rk1m_Ctd_zzz_o0_t0_l0 = 0.5 * d_Ctd_zzz_o0_t0_m0_l0;
						rk1Ctd_zzz = rk1m_Ctd_zzz_o0_t0_l0;
						rk1Ct_xzz = rk1gtu_xx * rk1Ctd_xzz + rk1gtu_xy * rk1Ctd_yzz + rk1gtu_xz * rk1Ctd_zzz;
						rk1Ct_zzz = rk1gtu_xz * rk1Ctd_xzz + rk1gtu_yz * rk1Ctd_yzz + rk1gtu_zz * rk1Ctd_zzz;
						rk1Ct_yzz = rk1gtu_xy * rk1Ctd_xzz + rk1gtu_yy * rk1Ctd_yzz + rk1gtu_yz * rk1Ctd_zzz;
						rk1m_Ctd_yxy_o0_t0_l0 = 0.5 * d_Ctd_xyy_o0_t1_m0_l0;
						rk1Ctd_yxy = rk1m_Ctd_yxy_o0_t0_l0;
						rk1Ct_xyz = rk1gtu_xx * rk1Ctd_xyz + rk1gtu_xy * rk1Ctd_yyz + rk1gtu_xz * rk1Ctd_zyz;
						rk1m_Ctd_yxz_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t0_m0_l0;
						rk1m_Ctd_yxz_o0_t0_l0 = 0.5 * d_Ctd_xyz_o0_t2_m0_l0;
						rk1Ctd_yxz = (rk1m_Ctd_yxz_o0_t0_l0 + rk1m_Ctd_xyz_o0_t1_l0) + rk1m_Ctd_yxz_o0_t2_l0;
						rk1Ct_zxz = rk1gtu_xz * rk1Ctd_xxz + rk1gtu_yz * rk1Ctd_yxz + rk1gtu_zz * rk1Ctd_zxz;
						rk1Ct_yxz = rk1gtu_xy * rk1Ctd_xxz + rk1gtu_yy * rk1Ctd_yxz + rk1gtu_yz * rk1Ctd_zxz;
						rk1m_Ctd_zxy_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t1_m0_l0;
						rk1Ctd_zxy = (rk1m_Ctd_yxz_o0_t0_l0 + rk1m_Ctd_xyz_o0_t0_l0) + rk1m_Ctd_zxy_o0_t2_l0;
						rk1Ct_zxy = rk1gtu_xz * rk1Ctd_xxy + rk1gtu_yz * rk1Ctd_yxy + rk1gtu_zz * rk1Ctd_zxy;
						rk1Ct_yxy = rk1gtu_xy * rk1Ctd_xxy + rk1gtu_yy * rk1Ctd_yxy + rk1gtu_yz * rk1Ctd_zxy;
						rk1Ct_xxy = rk1gtu_xx * rk1Ctd_xxy + rk1gtu_xy * rk1Ctd_yxy + rk1gtu_xz * rk1Ctd_zxy;
						rk1Ct_xxz = rk1gtu_xx * rk1Ctd_xxz + rk1gtu_xy * rk1Ctd_yxz + rk1gtu_xz * rk1Ctd_zxz;
						rk1Ct_xyy = rk1gtu_xx * rk1Ctd_xyy + rk1gtu_xy * rk1Ctd_yyy + rk1gtu_xz * rk1Ctd_zyy;
						rk1m_Ctd_yxx_o0_t1_l0 = (-0.5) * d_Ctd_xxy_o0_t0_m0_l0;
						rk1Ctd_yxx = d_Ctd_yxx_o0_t0_m0_l0 + rk1m_Ctd_yxx_o0_t1_l0;
						rk1Ct_zxx = rk1gtu_xz * rk1Ctd_xxx + rk1gtu_yz * rk1Ctd_yxx + rk1gtu_zz * rk1Ctd_zxx;
						rk1Gamt_z = rk1gtu_xx * rk1Ct_zxx + 2.0 * rk1gtu_xy * rk1Ct_zxy + 2.0 * rk1gtu_xz * rk1Ct_zxz + rk1gtu_yy * rk1Ct_zyy + 2.0 * rk1gtu_yz * rk1Ct_zyz + rk1gtu_zz * rk1Ct_zzz;
						rk1Zu_z = 0.5 * rk1chi_max * (vector(rk1Gamh_z, i, j, k) - rk1Gamt_z);
						rk1Ct_yxx = rk1gtu_xy * rk1Ctd_xxx + rk1gtu_yy * rk1Ctd_yxx + rk1gtu_yz * rk1Ctd_zxx;
						rk1Gamt_y = rk1gtu_xx * rk1Ct_yxx + 2.0 * rk1gtu_xy * rk1Ct_yxy + 2.0 * rk1gtu_xz * rk1Ct_yxz + rk1gtu_yy * rk1Ct_yyy + 2.0 * rk1gtu_yz * rk1Ct_yyz + rk1gtu_zz * rk1Ct_yzz;
						rk1Zu_y = 0.5 * rk1chi_max * (vector(rk1Gamh_y, i, j, k) - rk1Gamt_y);
						rk1Ct_xxx = rk1gtu_xx * rk1Ctd_xxx + rk1gtu_xy * rk1Ctd_yxx + rk1gtu_xz * rk1Ctd_zxx;
						rk1Gamt_x = rk1gtu_xx * rk1Ct_xxx + 2.0 * rk1gtu_xy * rk1Ct_xxy + 2.0 * rk1gtu_xz * rk1Ct_xxz + rk1gtu_yy * rk1Ct_xyy + 2.0 * rk1gtu_yz * rk1Ct_xyz + rk1gtu_zz * rk1Ct_xzz;
						rk1Zu_x = 0.5 * rk1chi_max * (vector(rk1Gamh_x, i, j, k) - rk1Gamt_x);
						rk1d_Rtd_xx_o0_t9_m0_l0 = rk1gtu_zz * (3.0 * rk1Ct_xxz * rk1Ctd_xxz + 2.0 * rk1Ct_yxz * rk1Ctd_xyz + rk1Ct_yxz * rk1Ctd_yxz + 2.0 * rk1Ct_zxz * rk1Ctd_xzz + rk1Ct_zxz * rk1Ctd_zxz) + rk1gtu_yz * (3.0 * rk1Ct_xxy * rk1Ctd_xxz + 2.0 * rk1Ct_yxy * rk1Ctd_xyz + rk1Ct_yxy * rk1Ctd_yxz + 2.0 * rk1Ct_zxy * rk1Ctd_xzz + rk1Ct_zxy * rk1Ctd_zxz) + rk1gtu_xz * (3.0 * rk1Ct_xxx * rk1Ctd_xxz + 2.0 * rk1Ct_yxx * rk1Ctd_xyz + rk1Ct_yxx * rk1Ctd_yxz + 2.0 * rk1Ct_zxx * rk1Ctd_xzz + rk1Ct_zxx * rk1Ctd_zxz) + rk1gtu_yz * (3.0 * rk1Ct_xxz * rk1Ctd_xxy + 2.0 * rk1Ct_yxz * rk1Ctd_xyy + rk1Ct_yxz * rk1Ctd_yxy + 2.0 * rk1Ct_zxz * rk1Ctd_xyz + rk1Ct_zxz * rk1Ctd_zxy) + rk1gtu_yy * (3.0 * rk1Ct_xxy * rk1Ctd_xxy + 2.0 * rk1Ct_yxy * rk1Ctd_xyy + rk1Ct_yxy * rk1Ctd_yxy + 2.0 * rk1Ct_zxy * rk1Ctd_xyz + rk1Ct_zxy * rk1Ctd_zxy) + rk1gtu_xy * (3.0 * rk1Ct_xxx * rk1Ctd_xxy + 2.0 * rk1Ct_yxx * rk1Ctd_xyy + rk1Ct_yxx * rk1Ctd_yxy + 2.0 * rk1Ct_zxx * rk1Ctd_xyz + rk1Ct_zxx * rk1Ctd_zxy) + rk1gtu_xz * (3.0 * rk1Ct_xxz * rk1Ctd_xxx + 2.0 * rk1Ct_yxz * rk1Ctd_xxy + rk1Ct_yxz * rk1Ctd_yxx + 2.0 * rk1Ct_zxz * rk1Ctd_xxz + rk1Ct_zxz * rk1Ctd_zxx) + rk1gtu_xy * (3.0 * rk1Ct_xxy * rk1Ctd_xxx + 2.0 * rk1Ct_yxy * rk1Ctd_xxy + rk1Ct_yxy * rk1Ctd_yxx + 2.0 * rk1Ct_zxy * rk1Ctd_xxz + rk1Ct_zxy * rk1Ctd_zxx) + rk1gtu_xx * (3.0 * rk1Ct_xxx * rk1Ctd_xxx + 2.0 * rk1Ct_yxx * rk1Ctd_xxy + rk1Ct_yxx * rk1Ctd_yxx + 2.0 * rk1Ct_zxx * rk1Ctd_xxz + rk1Ct_zxx * rk1Ctd_zxx) + vector(rk1Gamh_z, i, j, k) * rk1Ctd_xxz + vector(rk1Gamh_y, i, j, k) * rk1Ctd_xxy + vector(rk1Gamh_x, i, j, k) * rk1Ctd_xxx;
						rk1m_Rtd_xx_o0_t8_l0 = vector(rk1gtd_xz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
						rk1m_Rtd_xx_o0_t7_l0 = vector(rk1gtd_xy, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
						rk1m_Rtd_xx_o0_t6_l0 = vector(rk1gtd_xx, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
						rk1m_Rtd_xx_o0_t5_l0 = (-0.5 * rk1gtu_zz) * d_Rtd_xx_o0_t5_m0_l0;
						rk1m_Rtd_xx_o0_t4_l0 = (-rk1gtu_yz) * d_Rtd_xx_o0_t4_m0_l0;
						rk1m_Rtd_xx_o0_t3_l0 = (-0.5 * rk1gtu_yy) * d_Rtd_xx_o0_t3_m0_l0;
						rk1m_Rtd_xx_o0_t2_l0 = (-rk1gtu_xz) * d_Rtd_xx_o0_t2_m0_l0;
						rk1m_Rtd_xx_o0_t1_l0 = (-rk1gtu_xy) * d_Rtd_xx_o0_t1_m0_l0;
						rk1m_Rtd_xx_o0_t0_l0 = (-0.5 * rk1gtu_xx) * d_Rtd_xx_o0_t0_m0_l0;
						rk1Rtd_xx = ((((((((rk1m_Rtd_xx_o0_t0_l0 + rk1m_Rtd_xx_o0_t1_l0) + rk1m_Rtd_xx_o0_t2_l0) + rk1m_Rtd_xx_o0_t3_l0) + rk1m_Rtd_xx_o0_t4_l0) + rk1m_Rtd_xx_o0_t5_l0) + rk1m_Rtd_xx_o0_t6_l0) + rk1m_Rtd_xx_o0_t7_l0) + rk1m_Rtd_xx_o0_t8_l0) + rk1d_Rtd_xx_o0_t9_m0_l0;
						rk1d_Rtd_xz_o0_t12_m0_l0 = rk1gtu_zz * (rk1Ct_xxz * rk1Ctd_zxz + rk1Ct_xxz * rk1Ctd_xzz + rk1Ct_xzz * rk1Ctd_xxz + rk1Ct_yxz * rk1Ctd_zyz + rk1Ct_yxz * rk1Ctd_yzz + rk1Ct_yzz * rk1Ctd_xyz + 2.0 * rk1Ct_zxz * rk1Ctd_zzz + rk1Ct_zzz * rk1Ctd_xzz) + rk1gtu_yz * (rk1Ct_xxy * rk1Ctd_zxz + rk1Ct_xxy * rk1Ctd_xzz + rk1Ct_xyz * rk1Ctd_xxz + rk1Ct_yxy * rk1Ctd_zyz + rk1Ct_yxy * rk1Ctd_yzz + rk1Ct_yyz * rk1Ctd_xyz + 2.0 * rk1Ct_zxy * rk1Ctd_zzz + rk1Ct_zyz * rk1Ctd_xzz) + rk1gtu_xz * (rk1Ct_xxx * rk1Ctd_zxz + rk1Ct_xxx * rk1Ctd_xzz + rk1Ct_xxz * rk1Ctd_xxz + rk1Ct_yxx * rk1Ctd_zyz + rk1Ct_yxx * rk1Ctd_yzz + rk1Ct_yxz * rk1Ctd_xyz + 2.0 * rk1Ct_zxx * rk1Ctd_zzz + rk1Ct_zxz * rk1Ctd_xzz) + rk1gtu_yz * (rk1Ct_xxz * rk1Ctd_zxy + rk1Ct_xxz * rk1Ctd_xyz + rk1Ct_xzz * rk1Ctd_xxy + rk1Ct_yxz * rk1Ctd_zyy + rk1Ct_yxz * rk1Ctd_yyz + rk1Ct_yzz * rk1Ctd_xyy + 2.0 * rk1Ct_zxz * rk1Ctd_zyz + rk1Ct_zzz * rk1Ctd_xyz) + rk1gtu_yy * (rk1Ct_xxy * rk1Ctd_zxy + rk1Ct_xxy * rk1Ctd_xyz + rk1Ct_xyz * rk1Ctd_xxy + rk1Ct_yxy * rk1Ctd_zyy + rk1Ct_yxy * rk1Ctd_yyz + rk1Ct_yyz * rk1Ctd_xyy + 2.0 * rk1Ct_zxy * rk1Ctd_zyz + rk1Ct_zyz * rk1Ctd_xyz) + rk1gtu_xy * (rk1Ct_xxx * rk1Ctd_zxy + rk1Ct_xxx * rk1Ctd_xyz + rk1Ct_xxz * rk1Ctd_xxy + rk1Ct_yxx * rk1Ctd_zyy + rk1Ct_yxx * rk1Ctd_yyz + rk1Ct_yxz * rk1Ctd_xyy + 2.0 * rk1Ct_zxx * rk1Ctd_zyz + rk1Ct_zxz * rk1Ctd_xyz) + rk1gtu_xz * (rk1Ct_xxz * rk1Ctd_zxx + rk1Ct_xxz * rk1Ctd_xxz + rk1Ct_xzz * rk1Ctd_xxx + rk1Ct_yxz * rk1Ctd_zxy + rk1Ct_yxz * rk1Ctd_yxz + rk1Ct_yzz * rk1Ctd_xxy + 2.0 * rk1Ct_zxz * rk1Ctd_zxz + rk1Ct_zzz * rk1Ctd_xxz) + rk1gtu_xy * (rk1Ct_xxy * rk1Ctd_zxx + rk1Ct_xxy * rk1Ctd_xxz + rk1Ct_xyz * rk1Ctd_xxx + rk1Ct_yxy * rk1Ctd_zxy + rk1Ct_yxy * rk1Ctd_yxz + rk1Ct_yyz * rk1Ctd_xxy + 2.0 * rk1Ct_zxy * rk1Ctd_zxz + rk1Ct_zyz * rk1Ctd_xxz) + rk1gtu_xx * (rk1Ct_xxx * rk1Ctd_zxx + rk1Ct_xxx * rk1Ctd_xxz + rk1Ct_xxz * rk1Ctd_xxx + rk1Ct_yxx * rk1Ctd_zxy + rk1Ct_yxx * rk1Ctd_yxz + rk1Ct_yxz * rk1Ctd_xxy + 2.0 * rk1Ct_zxx * rk1Ctd_zxz + rk1Ct_zxz * rk1Ctd_xxz) + 0.5 * vector(rk1Gamh_z, i, j, k) * (rk1Ctd_xzz + rk1Ctd_zxz) + 0.5 * vector(rk1Gamh_y, i, j, k) * (rk1Ctd_xyz + rk1Ctd_zxy) + 0.5 * vector(rk1Gamh_x, i, j, k) * (rk1Ctd_xxz + rk1Ctd_zxx);
						rk1m_Rtd_xz_o0_t11_l0 = 0.5 * vector(rk1gtd_zz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
						rk1m_Rtd_xz_o0_t10_l0 = 0.5 * vector(rk1gtd_yz, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
						rk1m_Rtd_xz_o0_t9_l0 = 0.5 * vector(rk1gtd_xz, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
						rk1m_Rtd_xz_o0_t8_l0 = 0.5 * vector(rk1gtd_xz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
						rk1m_Rtd_xz_o0_t7_l0 = 0.5 * vector(rk1gtd_xy, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
						rk1m_Rtd_xz_o0_t6_l0 = 0.5 * vector(rk1gtd_xx, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
						rk1m_Rtd_xz_o0_t5_l0 = (-0.5 * rk1gtu_zz) * d_Rtd_xz_o0_t5_m0_l0;
						rk1m_Rtd_xz_o0_t4_l0 = (-rk1gtu_yz) * d_Rtd_xz_o0_t4_m0_l0;
						rk1m_Rtd_xz_o0_t3_l0 = (-0.5 * rk1gtu_yy) * d_Rtd_xz_o0_t3_m0_l0;
						rk1m_Rtd_xz_o0_t2_l0 = (-rk1gtu_xz) * d_Rtd_xz_o0_t2_m0_l0;
						rk1m_Rtd_xz_o0_t1_l0 = (-rk1gtu_xy) * d_Rtd_xz_o0_t1_m0_l0;
						rk1m_Rtd_xz_o0_t0_l0 = (-0.5 * rk1gtu_xx) * d_Rtd_xz_o0_t0_m0_l0;
						rk1Rtd_xz = (((((((((((rk1m_Rtd_xz_o0_t0_l0 + rk1m_Rtd_xz_o0_t1_l0) + rk1m_Rtd_xz_o0_t2_l0) + rk1m_Rtd_xz_o0_t3_l0) + rk1m_Rtd_xz_o0_t4_l0) + rk1m_Rtd_xz_o0_t5_l0) + rk1m_Rtd_xz_o0_t6_l0) + rk1m_Rtd_xz_o0_t7_l0) + rk1m_Rtd_xz_o0_t8_l0) + rk1m_Rtd_xz_o0_t9_l0) + rk1m_Rtd_xz_o0_t10_l0) + rk1m_Rtd_xz_o0_t11_l0) + rk1d_Rtd_xz_o0_t12_m0_l0;
						rk1d_Rtd_zz_o0_t9_m0_l0 = rk1gtu_zz * (2.0 * rk1Ct_xzz * rk1Ctd_zxz + rk1Ct_xzz * rk1Ctd_xzz + 2.0 * rk1Ct_yzz * rk1Ctd_zyz + rk1Ct_yzz * rk1Ctd_yzz + 3.0 * rk1Ct_zzz * rk1Ctd_zzz) + rk1gtu_yz * (2.0 * rk1Ct_xyz * rk1Ctd_zxz + rk1Ct_xyz * rk1Ctd_xzz + 2.0 * rk1Ct_yyz * rk1Ctd_zyz + rk1Ct_yyz * rk1Ctd_yzz + 3.0 * rk1Ct_zyz * rk1Ctd_zzz) + rk1gtu_xz * (2.0 * rk1Ct_xxz * rk1Ctd_zxz + rk1Ct_xxz * rk1Ctd_xzz + 2.0 * rk1Ct_yxz * rk1Ctd_zyz + rk1Ct_yxz * rk1Ctd_yzz + 3.0 * rk1Ct_zxz * rk1Ctd_zzz) + rk1gtu_yz * (2.0 * rk1Ct_xzz * rk1Ctd_zxy + rk1Ct_xzz * rk1Ctd_xyz + 2.0 * rk1Ct_yzz * rk1Ctd_zyy + rk1Ct_yzz * rk1Ctd_yyz + 3.0 * rk1Ct_zzz * rk1Ctd_zyz) + rk1gtu_yy * (2.0 * rk1Ct_xyz * rk1Ctd_zxy + rk1Ct_xyz * rk1Ctd_xyz + 2.0 * rk1Ct_yyz * rk1Ctd_zyy + rk1Ct_yyz * rk1Ctd_yyz + 3.0 * rk1Ct_zyz * rk1Ctd_zyz) + rk1gtu_xy * (2.0 * rk1Ct_xxz * rk1Ctd_zxy + rk1Ct_xxz * rk1Ctd_xyz + 2.0 * rk1Ct_yxz * rk1Ctd_zyy + rk1Ct_yxz * rk1Ctd_yyz + 3.0 * rk1Ct_zxz * rk1Ctd_zyz) + rk1gtu_xz * (2.0 * rk1Ct_xzz * rk1Ctd_zxx + rk1Ct_xzz * rk1Ctd_xxz + 2.0 * rk1Ct_yzz * rk1Ctd_zxy + rk1Ct_yzz * rk1Ctd_yxz + 3.0 * rk1Ct_zzz * rk1Ctd_zxz) + rk1gtu_xy * (2.0 * rk1Ct_xyz * rk1Ctd_zxx + rk1Ct_xyz * rk1Ctd_xxz + 2.0 * rk1Ct_yyz * rk1Ctd_zxy + rk1Ct_yyz * rk1Ctd_yxz + 3.0 * rk1Ct_zyz * rk1Ctd_zxz) + rk1gtu_xx * (2.0 * rk1Ct_xxz * rk1Ctd_zxx + rk1Ct_xxz * rk1Ctd_xxz + 2.0 * rk1Ct_yxz * rk1Ctd_zxy + rk1Ct_yxz * rk1Ctd_yxz + 3.0 * rk1Ct_zxz * rk1Ctd_zxz) + vector(rk1Gamh_z, i, j, k) * rk1Ctd_zzz + vector(rk1Gamh_y, i, j, k) * rk1Ctd_zyz + vector(rk1Gamh_x, i, j, k) * rk1Ctd_zxz;
						rk1m_Rtd_zz_o0_t8_l0 = vector(rk1gtd_zz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
						rk1m_Rtd_zz_o0_t7_l0 = vector(rk1gtd_yz, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
						rk1m_Rtd_zz_o0_t6_l0 = vector(rk1gtd_xz, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
						rk1m_Rtd_zz_o0_t5_l0 = (-0.5 * rk1gtu_zz) * d_Rtd_zz_o0_t5_m0_l0;
						rk1m_Rtd_zz_o0_t4_l0 = (-rk1gtu_yz) * d_Rtd_zz_o0_t4_m0_l0;
						rk1m_Rtd_zz_o0_t3_l0 = (-0.5 * rk1gtu_yy) * d_Rtd_zz_o0_t3_m0_l0;
						rk1m_Rtd_zz_o0_t2_l0 = (-rk1gtu_xz) * d_Rtd_zz_o0_t2_m0_l0;
						rk1m_Rtd_zz_o0_t1_l0 = (-rk1gtu_xy) * d_Rtd_zz_o0_t1_m0_l0;
						rk1m_Rtd_zz_o0_t0_l0 = (-0.5 * rk1gtu_xx) * d_Rtd_zz_o0_t0_m0_l0;
						rk1Rtd_zz = ((((((((rk1m_Rtd_zz_o0_t0_l0 + rk1m_Rtd_zz_o0_t1_l0) + rk1m_Rtd_zz_o0_t2_l0) + rk1m_Rtd_zz_o0_t3_l0) + rk1m_Rtd_zz_o0_t4_l0) + rk1m_Rtd_zz_o0_t5_l0) + rk1m_Rtd_zz_o0_t6_l0) + rk1m_Rtd_zz_o0_t7_l0) + rk1m_Rtd_zz_o0_t8_l0) + rk1d_Rtd_zz_o0_t9_m0_l0;
						rk1d_Rtd_xy_o0_t12_m0_l0 = rk1gtu_zz * (rk1Ct_xxz * rk1Ctd_yxz + rk1Ct_xxz * rk1Ctd_xyz + rk1Ct_xyz * rk1Ctd_xxz + 2.0 * rk1Ct_yxz * rk1Ctd_yyz + rk1Ct_yyz * rk1Ctd_xyz + rk1Ct_zxz * rk1Ctd_yzz + rk1Ct_zxz * rk1Ctd_zyz + rk1Ct_zyz * rk1Ctd_xzz) + rk1gtu_yz * (rk1Ct_xxy * rk1Ctd_yxz + rk1Ct_xxy * rk1Ctd_xyz + rk1Ct_xyy * rk1Ctd_xxz + 2.0 * rk1Ct_yxy * rk1Ctd_yyz + rk1Ct_yyy * rk1Ctd_xyz + rk1Ct_zxy * rk1Ctd_yzz + rk1Ct_zxy * rk1Ctd_zyz + rk1Ct_zyy * rk1Ctd_xzz) + rk1gtu_xz * (rk1Ct_xxx * rk1Ctd_yxz + rk1Ct_xxx * rk1Ctd_xyz + rk1Ct_xxy * rk1Ctd_xxz + 2.0 * rk1Ct_yxx * rk1Ctd_yyz + rk1Ct_yxy * rk1Ctd_xyz + rk1Ct_zxx * rk1Ctd_yzz + rk1Ct_zxx * rk1Ctd_zyz + rk1Ct_zxy * rk1Ctd_xzz) + rk1gtu_yz * (rk1Ct_xxz * rk1Ctd_yxy + rk1Ct_xxz * rk1Ctd_xyy + rk1Ct_xyz * rk1Ctd_xxy + 2.0 * rk1Ct_yxz * rk1Ctd_yyy + rk1Ct_yyz * rk1Ctd_xyy + rk1Ct_zxz * rk1Ctd_yyz + rk1Ct_zxz * rk1Ctd_zyy + rk1Ct_zyz * rk1Ctd_xyz) + rk1gtu_yy * (rk1Ct_xxy * rk1Ctd_yxy + rk1Ct_xxy * rk1Ctd_xyy + rk1Ct_xyy * rk1Ctd_xxy + 2.0 * rk1Ct_yxy * rk1Ctd_yyy + rk1Ct_yyy * rk1Ctd_xyy + rk1Ct_zxy * rk1Ctd_yyz + rk1Ct_zxy * rk1Ctd_zyy + rk1Ct_zyy * rk1Ctd_xyz) + rk1gtu_xy * (rk1Ct_xxx * rk1Ctd_yxy + rk1Ct_xxx * rk1Ctd_xyy + rk1Ct_xxy * rk1Ctd_xxy + 2.0 * rk1Ct_yxx * rk1Ctd_yyy + rk1Ct_yxy * rk1Ctd_xyy + rk1Ct_zxx * rk1Ctd_yyz + rk1Ct_zxx * rk1Ctd_zyy + rk1Ct_zxy * rk1Ctd_xyz) + rk1gtu_xz * (rk1Ct_xxz * rk1Ctd_yxx + rk1Ct_xxz * rk1Ctd_xxy + rk1Ct_xyz * rk1Ctd_xxx + 2.0 * rk1Ct_yxz * rk1Ctd_yxy + rk1Ct_yyz * rk1Ctd_xxy + rk1Ct_zxz * rk1Ctd_yxz + rk1Ct_zxz * rk1Ctd_zxy + rk1Ct_zyz * rk1Ctd_xxz) + rk1gtu_xy * (rk1Ct_xxy * rk1Ctd_yxx + rk1Ct_xxy * rk1Ctd_xxy + rk1Ct_xyy * rk1Ctd_xxx + 2.0 * rk1Ct_yxy * rk1Ctd_yxy + rk1Ct_yyy * rk1Ctd_xxy + rk1Ct_zxy * rk1Ctd_yxz + rk1Ct_zxy * rk1Ctd_zxy + rk1Ct_zyy * rk1Ctd_xxz) + rk1gtu_xx * (rk1Ct_xxx * rk1Ctd_yxx + rk1Ct_xxx * rk1Ctd_xxy + rk1Ct_xxy * rk1Ctd_xxx + 2.0 * rk1Ct_yxx * rk1Ctd_yxy + rk1Ct_yxy * rk1Ctd_xxy + rk1Ct_zxx * rk1Ctd_yxz + rk1Ct_zxx * rk1Ctd_zxy + rk1Ct_zxy * rk1Ctd_xxz) + 0.5 * vector(rk1Gamh_z, i, j, k) * (rk1Ctd_xyz + rk1Ctd_yxz) + 0.5 * vector(rk1Gamh_y, i, j, k) * (rk1Ctd_xyy + rk1Ctd_yxy) + 0.5 * vector(rk1Gamh_x, i, j, k) * (rk1Ctd_xxy + rk1Ctd_yxx);
						rk1m_Rtd_xy_o0_t11_l0 = 0.5 * vector(rk1gtd_yz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
						rk1m_Rtd_xy_o0_t10_l0 = 0.5 * vector(rk1gtd_yy, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
						rk1m_Rtd_xy_o0_t9_l0 = 0.5 * vector(rk1gtd_xy, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
						rk1m_Rtd_xy_o0_t8_l0 = 0.5 * vector(rk1gtd_xz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
						rk1m_Rtd_xy_o0_t7_l0 = 0.5 * vector(rk1gtd_xy, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
						rk1m_Rtd_xy_o0_t6_l0 = 0.5 * vector(rk1gtd_xx, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
						rk1m_Rtd_xy_o0_t5_l0 = (-0.5 * rk1gtu_zz) * d_Rtd_xy_o0_t5_m0_l0;
						rk1m_Rtd_xy_o0_t4_l0 = (-rk1gtu_yz) * d_Rtd_xy_o0_t4_m0_l0;
						rk1m_Rtd_xy_o0_t3_l0 = (-0.5 * rk1gtu_yy) * d_Rtd_xy_o0_t3_m0_l0;
						rk1m_Rtd_xy_o0_t2_l0 = (-rk1gtu_xz) * d_Rtd_xy_o0_t2_m0_l0;
						rk1m_Rtd_xy_o0_t1_l0 = (-rk1gtu_xy) * d_Rtd_xy_o0_t1_m0_l0;
						rk1m_Rtd_xy_o0_t0_l0 = (-0.5 * rk1gtu_xx) * d_Rtd_xy_o0_t0_m0_l0;
						rk1Rtd_xy = (((((((((((rk1m_Rtd_xy_o0_t0_l0 + rk1m_Rtd_xy_o0_t1_l0) + rk1m_Rtd_xy_o0_t2_l0) + rk1m_Rtd_xy_o0_t3_l0) + rk1m_Rtd_xy_o0_t4_l0) + rk1m_Rtd_xy_o0_t5_l0) + rk1m_Rtd_xy_o0_t6_l0) + rk1m_Rtd_xy_o0_t7_l0) + rk1m_Rtd_xy_o0_t8_l0) + rk1m_Rtd_xy_o0_t9_l0) + rk1m_Rtd_xy_o0_t10_l0) + rk1m_Rtd_xy_o0_t11_l0) + rk1d_Rtd_xy_o0_t12_m0_l0;
						rk1d_Rtd_yy_o0_t9_m0_l0 = rk1gtu_zz * (2.0 * rk1Ct_xyz * rk1Ctd_yxz + rk1Ct_xyz * rk1Ctd_xyz + 3.0 * rk1Ct_yyz * rk1Ctd_yyz + 2.0 * rk1Ct_zyz * rk1Ctd_yzz + rk1Ct_zyz * rk1Ctd_zyz) + rk1gtu_yz * (2.0 * rk1Ct_xyy * rk1Ctd_yxz + rk1Ct_xyy * rk1Ctd_xyz + 3.0 * rk1Ct_yyy * rk1Ctd_yyz + 2.0 * rk1Ct_zyy * rk1Ctd_yzz + rk1Ct_zyy * rk1Ctd_zyz) + rk1gtu_xz * (2.0 * rk1Ct_xxy * rk1Ctd_yxz + rk1Ct_xxy * rk1Ctd_xyz + 3.0 * rk1Ct_yxy * rk1Ctd_yyz + 2.0 * rk1Ct_zxy * rk1Ctd_yzz + rk1Ct_zxy * rk1Ctd_zyz) + rk1gtu_yz * (2.0 * rk1Ct_xyz * rk1Ctd_yxy + rk1Ct_xyz * rk1Ctd_xyy + 3.0 * rk1Ct_yyz * rk1Ctd_yyy + 2.0 * rk1Ct_zyz * rk1Ctd_yyz + rk1Ct_zyz * rk1Ctd_zyy) + rk1gtu_yy * (2.0 * rk1Ct_xyy * rk1Ctd_yxy + rk1Ct_xyy * rk1Ctd_xyy + 3.0 * rk1Ct_yyy * rk1Ctd_yyy + 2.0 * rk1Ct_zyy * rk1Ctd_yyz + rk1Ct_zyy * rk1Ctd_zyy) + rk1gtu_xy * (2.0 * rk1Ct_xxy * rk1Ctd_yxy + rk1Ct_xxy * rk1Ctd_xyy + 3.0 * rk1Ct_yxy * rk1Ctd_yyy + 2.0 * rk1Ct_zxy * rk1Ctd_yyz + rk1Ct_zxy * rk1Ctd_zyy) + rk1gtu_xz * (2.0 * rk1Ct_xyz * rk1Ctd_yxx + rk1Ct_xyz * rk1Ctd_xxy + 3.0 * rk1Ct_yyz * rk1Ctd_yxy + 2.0 * rk1Ct_zyz * rk1Ctd_yxz + rk1Ct_zyz * rk1Ctd_zxy) + rk1gtu_xy * (2.0 * rk1Ct_xyy * rk1Ctd_yxx + rk1Ct_xyy * rk1Ctd_xxy + 3.0 * rk1Ct_yyy * rk1Ctd_yxy + 2.0 * rk1Ct_zyy * rk1Ctd_yxz + rk1Ct_zyy * rk1Ctd_zxy) + rk1gtu_xx * (2.0 * rk1Ct_xxy * rk1Ctd_yxx + rk1Ct_xxy * rk1Ctd_xxy + 3.0 * rk1Ct_yxy * rk1Ctd_yxy + 2.0 * rk1Ct_zxy * rk1Ctd_yxz + rk1Ct_zxy * rk1Ctd_zxy) + vector(rk1Gamh_z, i, j, k) * rk1Ctd_yyz + vector(rk1Gamh_y, i, j, k) * rk1Ctd_yyy + vector(rk1Gamh_x, i, j, k) * rk1Ctd_yxy;
						rk1m_Rtd_yy_o0_t8_l0 = vector(rk1gtd_yz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
						rk1m_Rtd_yy_o0_t7_l0 = vector(rk1gtd_yy, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
						rk1m_Rtd_yy_o0_t6_l0 = vector(rk1gtd_xy, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
						rk1m_Rtd_yy_o0_t5_l0 = (-0.5 * rk1gtu_zz) * d_Rtd_yy_o0_t5_m0_l0;
						rk1m_Rtd_yy_o0_t4_l0 = (-rk1gtu_yz) * d_Rtd_yy_o0_t4_m0_l0;
						rk1m_Rtd_yy_o0_t3_l0 = (-0.5 * rk1gtu_yy) * d_Rtd_yy_o0_t3_m0_l0;
						rk1m_Rtd_yy_o0_t2_l0 = (-rk1gtu_xz) * d_Rtd_yy_o0_t2_m0_l0;
						rk1m_Rtd_yy_o0_t1_l0 = (-rk1gtu_xy) * d_Rtd_yy_o0_t1_m0_l0;
						rk1m_Rtd_yy_o0_t0_l0 = (-0.5 * rk1gtu_xx) * d_Rtd_yy_o0_t0_m0_l0;
						rk1Rtd_yy = ((((((((rk1m_Rtd_yy_o0_t0_l0 + rk1m_Rtd_yy_o0_t1_l0) + rk1m_Rtd_yy_o0_t2_l0) + rk1m_Rtd_yy_o0_t3_l0) + rk1m_Rtd_yy_o0_t4_l0) + rk1m_Rtd_yy_o0_t5_l0) + rk1m_Rtd_yy_o0_t6_l0) + rk1m_Rtd_yy_o0_t7_l0) + rk1m_Rtd_yy_o0_t8_l0) + rk1d_Rtd_yy_o0_t9_m0_l0;
						rk1d_Rtd_yz_o0_t12_m0_l0 = rk1gtu_zz * (rk1Ct_xyz * rk1Ctd_zxz + rk1Ct_xyz * rk1Ctd_xzz + rk1Ct_xzz * rk1Ctd_yxz + rk1Ct_yyz * rk1Ctd_zyz + rk1Ct_yyz * rk1Ctd_yzz + rk1Ct_yzz * rk1Ctd_yyz + 2.0 * rk1Ct_zyz * rk1Ctd_zzz + rk1Ct_zzz * rk1Ctd_yzz) + rk1gtu_yz * (rk1Ct_xyy * rk1Ctd_zxz + rk1Ct_xyy * rk1Ctd_xzz + rk1Ct_xyz * rk1Ctd_yxz + rk1Ct_yyy * rk1Ctd_zyz + rk1Ct_yyy * rk1Ctd_yzz + rk1Ct_yyz * rk1Ctd_yyz + 2.0 * rk1Ct_zyy * rk1Ctd_zzz + rk1Ct_zyz * rk1Ctd_yzz) + rk1gtu_xz * (rk1Ct_xxy * rk1Ctd_zxz + rk1Ct_xxy * rk1Ctd_xzz + rk1Ct_xxz * rk1Ctd_yxz + rk1Ct_yxy * rk1Ctd_zyz + rk1Ct_yxy * rk1Ctd_yzz + rk1Ct_yxz * rk1Ctd_yyz + 2.0 * rk1Ct_zxy * rk1Ctd_zzz + rk1Ct_zxz * rk1Ctd_yzz) + rk1gtu_yz * (rk1Ct_xyz * rk1Ctd_zxy + rk1Ct_xyz * rk1Ctd_xyz + rk1Ct_xzz * rk1Ctd_yxy + rk1Ct_yyz * rk1Ctd_zyy + rk1Ct_yyz * rk1Ctd_yyz + rk1Ct_yzz * rk1Ctd_yyy + 2.0 * rk1Ct_zyz * rk1Ctd_zyz + rk1Ct_zzz * rk1Ctd_yyz) + rk1gtu_yy * (rk1Ct_xyy * rk1Ctd_zxy + rk1Ct_xyy * rk1Ctd_xyz + rk1Ct_xyz * rk1Ctd_yxy + rk1Ct_yyy * rk1Ctd_zyy + rk1Ct_yyy * rk1Ctd_yyz + rk1Ct_yyz * rk1Ctd_yyy + 2.0 * rk1Ct_zyy * rk1Ctd_zyz + rk1Ct_zyz * rk1Ctd_yyz) + rk1gtu_xy * (rk1Ct_xxy * rk1Ctd_zxy + rk1Ct_xxy * rk1Ctd_xyz + rk1Ct_xxz * rk1Ctd_yxy + rk1Ct_yxy * rk1Ctd_zyy + rk1Ct_yxy * rk1Ctd_yyz + rk1Ct_yxz * rk1Ctd_yyy + 2.0 * rk1Ct_zxy * rk1Ctd_zyz + rk1Ct_zxz * rk1Ctd_yyz) + rk1gtu_xz * (rk1Ct_xyz * rk1Ctd_zxx + rk1Ct_xyz * rk1Ctd_xxz + rk1Ct_xzz * rk1Ctd_yxx + rk1Ct_yyz * rk1Ctd_zxy + rk1Ct_yyz * rk1Ctd_yxz + rk1Ct_yzz * rk1Ctd_yxy + 2.0 * rk1Ct_zyz * rk1Ctd_zxz + rk1Ct_zzz * rk1Ctd_yxz) + rk1gtu_xy * (rk1Ct_xyy * rk1Ctd_zxx + rk1Ct_xyy * rk1Ctd_xxz + rk1Ct_xyz * rk1Ctd_yxx + rk1Ct_yyy * rk1Ctd_zxy + rk1Ct_yyy * rk1Ctd_yxz + rk1Ct_yyz * rk1Ctd_yxy + 2.0 * rk1Ct_zyy * rk1Ctd_zxz + rk1Ct_zyz * rk1Ctd_yxz) + rk1gtu_xx * (rk1Ct_xxy * rk1Ctd_zxx + rk1Ct_xxy * rk1Ctd_xxz + rk1Ct_xxz * rk1Ctd_yxx + rk1Ct_yxy * rk1Ctd_zxy + rk1Ct_yxy * rk1Ctd_yxz + rk1Ct_yxz * rk1Ctd_yxy + 2.0 * rk1Ct_zxy * rk1Ctd_zxz + rk1Ct_zxz * rk1Ctd_yxz) + 0.5 * vector(rk1Gamh_z, i, j, k) * (rk1Ctd_yzz + rk1Ctd_zyz) + 0.5 * vector(rk1Gamh_y, i, j, k) * (rk1Ctd_yyz + rk1Ctd_zyy) + 0.5 * vector(rk1Gamh_x, i, j, k) * (rk1Ctd_yxz + rk1Ctd_zxy);
						rk1m_Rtd_yz_o0_t11_l0 = 0.5 * vector(rk1gtd_zz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
						rk1m_Rtd_yz_o0_t10_l0 = 0.5 * vector(rk1gtd_yz, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
						rk1m_Rtd_yz_o0_t9_l0 = 0.5 * vector(rk1gtd_xz, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
						rk1m_Rtd_yz_o0_t8_l0 = 0.5 * vector(rk1gtd_yz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
						rk1m_Rtd_yz_o0_t7_l0 = 0.5 * vector(rk1gtd_yy, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
						rk1m_Rtd_yz_o0_t6_l0 = 0.5 * vector(rk1gtd_xy, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
						rk1m_Rtd_yz_o0_t5_l0 = (-0.5 * rk1gtu_zz) * d_Rtd_yz_o0_t5_m0_l0;
						rk1m_Rtd_yz_o0_t4_l0 = (-rk1gtu_yz) * d_Rtd_yz_o0_t4_m0_l0;
						rk1m_Rtd_yz_o0_t3_l0 = (-0.5 * rk1gtu_yy) * d_Rtd_yz_o0_t3_m0_l0;
						rk1m_Rtd_yz_o0_t2_l0 = (-rk1gtu_xz) * d_Rtd_yz_o0_t2_m0_l0;
						rk1m_Rtd_yz_o0_t1_l0 = (-rk1gtu_xy) * d_Rtd_yz_o0_t1_m0_l0;
						rk1m_Rtd_yz_o0_t0_l0 = (-0.5 * rk1gtu_xx) * d_Rtd_yz_o0_t0_m0_l0;
						rk1Rtd_yz = (((((((((((rk1m_Rtd_yz_o0_t0_l0 + rk1m_Rtd_yz_o0_t1_l0) + rk1m_Rtd_yz_o0_t2_l0) + rk1m_Rtd_yz_o0_t3_l0) + rk1m_Rtd_yz_o0_t4_l0) + rk1m_Rtd_yz_o0_t5_l0) + rk1m_Rtd_yz_o0_t6_l0) + rk1m_Rtd_yz_o0_t7_l0) + rk1m_Rtd_yz_o0_t8_l0) + rk1m_Rtd_yz_o0_t9_l0) + rk1m_Rtd_yz_o0_t10_l0) + rk1m_Rtd_yz_o0_t11_l0) + rk1d_Rtd_yz_o0_t12_m0_l0;
						rk1m_Rpd_xx_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t3_m0_l0;
						rk1m_Rpd_xx_o0_t21_l0 = (-0.75 * vector(rk1gtd_xx, i, j, k) * rk1gtu_zz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xx_o0_t20_l0 = (-0.75 * vector(rk1gtd_xx, i, j, k) * rk1gtu_yy * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xx_o0_t19_l0 = (-0.75 * vector(rk1gtd_xx, i, j, k) * rk1gtu_xx * rk1inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xx_o0_t18_l0 = 2.0 * rk1inv_chi * rk1Zu_z * vector(rk1gtd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xx_o0_t17_l0 = 2.0 * rk1inv_chi * rk1Zu_y * vector(rk1gtd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xx_o0_t16_l0 = 2.0 * rk1inv_chi * rk1Zu_x * vector(rk1gtd_xx, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xx_o0_t15_l0 = (-1.500000000000000 * vector(rk1gtd_xx, i, j, k) * rk1gtu_yz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xx_o0_t14_l0 = (-1.500000000000000 * vector(rk1gtd_xx, i, j, k) * rk1gtu_xz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xx_o0_t13_l0 = (-1.500000000000000 * vector(rk1gtd_xx, i, j, k) * rk1gtu_xy * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xx_o0_t12_l0 = (-0.25 * rk1inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xx_o0_t11_l0 = (-0.5 * rk1Ct_zxx) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xx_o0_t10_l0 = (-0.5 * rk1Ct_yxx) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xx_o0_t9_l0 = (-0.5 * rk1Ct_xxx) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xx_o0_t8_l0 = 0.5 * vector(rk1gtd_xx, i, j, k) * rk1gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk1m_Rpd_xx_o0_t7_l0 = vector(rk1gtd_xx, i, j, k) * rk1gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk1m_Rpd_xx_o0_t6_l0 = 0.5 * vector(rk1gtd_xx, i, j, k) * rk1gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk1m_Rpd_xx_o0_t5_l0 = vector(rk1gtd_xx, i, j, k) * rk1gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk1m_Rpd_xx_o0_t4_l0 = vector(rk1gtd_xx, i, j, k) * rk1gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk1m_Rpd_xx_o0_t3_l0 = 0.5 * vector(rk1gtd_xx, i, j, k) * rk1gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk1m_Rpd_xx_o0_t2_l0 = (-0.5 * vector(rk1gtd_xx, i, j, k) * vector(rk1Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xx_o0_t1_l0 = (-0.5 * vector(rk1gtd_xx, i, j, k) * vector(rk1Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xx_o0_t0_l0 = (-0.5 * vector(rk1gtd_xx, i, j, k) * vector(rk1Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk1Rpd_xx = (((((((((((((((((((((rk1m_Rpd_xx_o0_t0_l0 + rk1m_Rpd_xx_o0_t1_l0) + rk1m_Rpd_xx_o0_t2_l0) + rk1m_Rpd_xx_o0_t3_l0) + rk1m_Rpd_xx_o0_t4_l0) + rk1m_Rpd_xx_o0_t5_l0) + rk1m_Rpd_xx_o0_t6_l0) + rk1m_Rpd_xx_o0_t7_l0) + rk1m_Rpd_xx_o0_t8_l0) + rk1m_Rpd_xx_o0_t9_l0) + rk1m_Rpd_xx_o0_t10_l0) + rk1m_Rpd_xx_o0_t11_l0) + rk1m_Rpd_xx_o0_t12_l0) + rk1m_Rpd_xx_o0_t13_l0) + rk1m_Rpd_xx_o0_t14_l0) + rk1m_Rpd_xx_o0_t15_l0) + rk1m_Rpd_xx_o0_t16_l0) + rk1m_Rpd_xx_o0_t17_l0) + rk1m_Rpd_xx_o0_t18_l0) + rk1m_Rpd_xx_o0_t19_l0) + rk1m_Rpd_xx_o0_t20_l0) + rk1m_Rpd_xx_o0_t21_l0) + rk1m_Rpd_xx_o0_t22_l0;
						rk1m_Rpd_yy_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t6_m0_l0;
						rk1m_Rpd_yy_o0_t21_l0 = (-0.75 * vector(rk1gtd_yy, i, j, k) * rk1gtu_zz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_yy_o0_t20_l0 = (-0.75 * vector(rk1gtd_yy, i, j, k) * rk1gtu_yy * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yy_o0_t19_l0 = (-0.75 * vector(rk1gtd_yy, i, j, k) * rk1gtu_xx * rk1inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_yy_o0_t18_l0 = 2.0 * rk1inv_chi * rk1Zu_z * vector(rk1gtd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yy_o0_t17_l0 = 2.0 * rk1inv_chi * rk1Zu_y * vector(rk1gtd_yy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yy_o0_t16_l0 = 2.0 * rk1inv_chi * rk1Zu_x * vector(rk1gtd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yy_o0_t15_l0 = (-1.500000000000000 * vector(rk1gtd_yy, i, j, k) * rk1gtu_xz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_yy_o0_t14_l0 = (-1.500000000000000 * vector(rk1gtd_yy, i, j, k) * rk1gtu_xy * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_yy_o0_t13_l0 = (-0.25 * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yy_o0_t12_l0 = (-0.5 * rk1Ct_zyy) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_yy_o0_t11_l0 = (-0.5 * rk1Ct_yyy) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yy_o0_t10_l0 = (-0.5 * rk1Ct_xyy) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_yy_o0_t9_l0 = 0.5 * vector(rk1gtd_yy, i, j, k) * rk1gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk1m_Rpd_yy_o0_t8_l0 = vector(rk1gtd_yy, i, j, k) * rk1gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk1m_Rpd_yy_o0_t7_l0 = 0.5 * vector(rk1gtd_yy, i, j, k) * rk1gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk1m_Rpd_yy_o0_t6_l0 = vector(rk1gtd_yy, i, j, k) * rk1gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk1m_Rpd_yy_o0_t5_l0 = vector(rk1gtd_yy, i, j, k) * rk1gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk1m_Rpd_yy_o0_t4_l0 = 0.5 * vector(rk1gtd_yy, i, j, k) * rk1gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk1m_Rpd_yy_o0_t3_l0 = (-0.5 * vector(rk1gtd_yy, i, j, k) * vector(rk1Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_yy_o0_t2_l0 = (-0.5 * vector(rk1gtd_yy, i, j, k) * vector(rk1Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yy_o0_t1_l0 = (-0.5 * vector(rk1gtd_yy, i, j, k) * vector(rk1Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_yy_o0_t0_l0 = (-1.500000000000000 * vector(rk1gtd_yy, i, j, k) * rk1gtu_yz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1Rpd_yy = (((((((((((((((((((((rk1m_Rpd_yy_o0_t0_l0 + rk1m_Rpd_yy_o0_t1_l0) + rk1m_Rpd_yy_o0_t2_l0) + rk1m_Rpd_yy_o0_t3_l0) + rk1m_Rpd_yy_o0_t4_l0) + rk1m_Rpd_yy_o0_t5_l0) + rk1m_Rpd_yy_o0_t6_l0) + rk1m_Rpd_yy_o0_t7_l0) + rk1m_Rpd_yy_o0_t8_l0) + rk1m_Rpd_yy_o0_t9_l0) + rk1m_Rpd_yy_o0_t10_l0) + rk1m_Rpd_yy_o0_t11_l0) + rk1m_Rpd_yy_o0_t12_l0) + rk1m_Rpd_yy_o0_t13_l0) + rk1m_Rpd_yy_o0_t14_l0) + rk1m_Rpd_yy_o0_t15_l0) + rk1m_Rpd_yy_o0_t16_l0) + rk1m_Rpd_yy_o0_t17_l0) + rk1m_Rpd_yy_o0_t18_l0) + rk1m_Rpd_yy_o0_t19_l0) + rk1m_Rpd_yy_o0_t20_l0) + rk1m_Rpd_yy_o0_t21_l0) + rk1m_Rpd_yy_o0_t22_l0;
						rk1m_Rpd_yz_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t7_m0_l0;
						rk1m_Rpd_yz_o0_t24_l0 = (-0.75 * vector(rk1gtd_yz, i, j, k) * rk1gtu_zz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_yz_o0_t23_l0 = (-0.75 * vector(rk1gtd_yz, i, j, k) * rk1gtu_yy * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yz_o0_t22_l0 = (-0.75 * vector(rk1gtd_yz, i, j, k) * rk1gtu_xx * rk1inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_yz_o0_t21_l0 = rk1inv_chi * rk1Zu_z * vector(rk1gtd_zz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yz_o0_t20_l0 = rk1inv_chi * rk1Zu_y * vector(rk1gtd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yz_o0_t19_l0 = rk1inv_chi * rk1Zu_x * vector(rk1gtd_xz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yz_o0_t18_l0 = rk1inv_chi * rk1Zu_z * vector(rk1gtd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_yz_o0_t17_l0 = rk1inv_chi * rk1Zu_y * vector(rk1gtd_yy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_yz_o0_t16_l0 = rk1inv_chi * rk1Zu_x * vector(rk1gtd_xy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_yz_o0_t15_l0 = (-0.5 * rk1Ct_zyz) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_yz_o0_t14_l0 = (-0.5 * rk1Ct_yyz) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yz_o0_t13_l0 = (-0.5 * rk1Ct_xyz) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_yz_o0_t12_l0 = (-0.25 * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_yz_o0_t11_l0 = 0.5 * vector(rk1gtd_yz, i, j, k) * rk1gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk1m_Rpd_yz_o0_t10_l0 = vector(rk1gtd_yz, i, j, k) * rk1gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk1m_Rpd_yz_o0_t9_l0 = 0.5 * vector(rk1gtd_yz, i, j, k) * rk1gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk1m_Rpd_yz_o0_t8_l0 = vector(rk1gtd_yz, i, j, k) * rk1gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk1m_Rpd_yz_o0_t7_l0 = vector(rk1gtd_yz, i, j, k) * rk1gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk1m_Rpd_yz_o0_t6_l0 = 0.5 * vector(rk1gtd_yz, i, j, k) * rk1gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk1m_Rpd_yz_o0_t5_l0 = (-0.5 * vector(rk1gtd_yz, i, j, k) * vector(rk1Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_yz_o0_t4_l0 = (-0.5 * vector(rk1gtd_yz, i, j, k) * vector(rk1Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yz_o0_t3_l0 = (-0.5 * vector(rk1gtd_yz, i, j, k) * vector(rk1Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_yz_o0_t2_l0 = (-1.500000000000000 * vector(rk1gtd_yz, i, j, k) * rk1gtu_yz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_yz_o0_t1_l0 = (-1.500000000000000 * vector(rk1gtd_yz, i, j, k) * rk1gtu_xz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_yz_o0_t0_l0 = (-1.500000000000000 * vector(rk1gtd_yz, i, j, k) * rk1gtu_xy * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1Rpd_yz = ((((((((((((((((((((((((rk1m_Rpd_yz_o0_t0_l0 + rk1m_Rpd_yz_o0_t1_l0) + rk1m_Rpd_yz_o0_t2_l0) + rk1m_Rpd_yz_o0_t3_l0) + rk1m_Rpd_yz_o0_t4_l0) + rk1m_Rpd_yz_o0_t5_l0) + rk1m_Rpd_yz_o0_t6_l0) + rk1m_Rpd_yz_o0_t7_l0) + rk1m_Rpd_yz_o0_t8_l0) + rk1m_Rpd_yz_o0_t9_l0) + rk1m_Rpd_yz_o0_t10_l0) + rk1m_Rpd_yz_o0_t11_l0) + rk1m_Rpd_yz_o0_t12_l0) + rk1m_Rpd_yz_o0_t13_l0) + rk1m_Rpd_yz_o0_t14_l0) + rk1m_Rpd_yz_o0_t15_l0) + rk1m_Rpd_yz_o0_t16_l0) + rk1m_Rpd_yz_o0_t17_l0) + rk1m_Rpd_yz_o0_t18_l0) + rk1m_Rpd_yz_o0_t19_l0) + rk1m_Rpd_yz_o0_t20_l0) + rk1m_Rpd_yz_o0_t21_l0) + rk1m_Rpd_yz_o0_t22_l0) + rk1m_Rpd_yz_o0_t23_l0) + rk1m_Rpd_yz_o0_t24_l0) + rk1m_Rpd_yz_o0_t25_l0;
						rk1m_Rpd_xz_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t5_m0_l0;
						rk1m_Rpd_xz_o0_t24_l0 = (-0.75 * vector(rk1gtd_xz, i, j, k) * rk1gtu_zz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xz_o0_t23_l0 = (-0.75 * vector(rk1gtd_xz, i, j, k) * rk1gtu_yy * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xz_o0_t22_l0 = (-0.75 * vector(rk1gtd_xz, i, j, k) * rk1gtu_xx * rk1inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xz_o0_t21_l0 = rk1inv_chi * rk1Zu_z * vector(rk1gtd_zz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xz_o0_t20_l0 = rk1inv_chi * rk1Zu_y * vector(rk1gtd_yz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xz_o0_t19_l0 = rk1inv_chi * rk1Zu_x * vector(rk1gtd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xz_o0_t18_l0 = rk1inv_chi * rk1Zu_z * vector(rk1gtd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xz_o0_t17_l0 = rk1inv_chi * rk1Zu_y * vector(rk1gtd_xy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xz_o0_t16_l0 = rk1inv_chi * rk1Zu_x * vector(rk1gtd_xx, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xz_o0_t15_l0 = (-1.500000000000000 * vector(rk1gtd_xz, i, j, k) * rk1gtu_yz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xz_o0_t14_l0 = (-1.500000000000000 * vector(rk1gtd_xz, i, j, k) * rk1gtu_xz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xz_o0_t13_l0 = (-1.500000000000000 * vector(rk1gtd_xz, i, j, k) * rk1gtu_xy * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xz_o0_t12_l0 = (-0.5 * rk1Ct_zxz) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xz_o0_t11_l0 = (-0.5 * rk1Ct_yxz) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xz_o0_t10_l0 = (-0.5 * rk1Ct_xxz) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xz_o0_t9_l0 = (-0.25 * rk1inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xz_o0_t8_l0 = 0.5 * vector(rk1gtd_xz, i, j, k) * rk1gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk1m_Rpd_xz_o0_t7_l0 = vector(rk1gtd_xz, i, j, k) * rk1gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk1m_Rpd_xz_o0_t6_l0 = 0.5 * vector(rk1gtd_xz, i, j, k) * rk1gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk1m_Rpd_xz_o0_t5_l0 = vector(rk1gtd_xz, i, j, k) * rk1gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk1m_Rpd_xz_o0_t4_l0 = vector(rk1gtd_xz, i, j, k) * rk1gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk1m_Rpd_xz_o0_t3_l0 = 0.5 * vector(rk1gtd_xz, i, j, k) * rk1gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk1m_Rpd_xz_o0_t2_l0 = (-0.5 * vector(rk1gtd_xz, i, j, k) * vector(rk1Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xz_o0_t1_l0 = (-0.5 * vector(rk1gtd_xz, i, j, k) * vector(rk1Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xz_o0_t0_l0 = (-0.5 * vector(rk1gtd_xz, i, j, k) * vector(rk1Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk1Rpd_xz = ((((((((((((((((((((((((rk1m_Rpd_xz_o0_t0_l0 + rk1m_Rpd_xz_o0_t1_l0) + rk1m_Rpd_xz_o0_t2_l0) + rk1m_Rpd_xz_o0_t3_l0) + rk1m_Rpd_xz_o0_t4_l0) + rk1m_Rpd_xz_o0_t5_l0) + rk1m_Rpd_xz_o0_t6_l0) + rk1m_Rpd_xz_o0_t7_l0) + rk1m_Rpd_xz_o0_t8_l0) + rk1m_Rpd_xz_o0_t9_l0) + rk1m_Rpd_xz_o0_t10_l0) + rk1m_Rpd_xz_o0_t11_l0) + rk1m_Rpd_xz_o0_t12_l0) + rk1m_Rpd_xz_o0_t13_l0) + rk1m_Rpd_xz_o0_t14_l0) + rk1m_Rpd_xz_o0_t15_l0) + rk1m_Rpd_xz_o0_t16_l0) + rk1m_Rpd_xz_o0_t17_l0) + rk1m_Rpd_xz_o0_t18_l0) + rk1m_Rpd_xz_o0_t19_l0) + rk1m_Rpd_xz_o0_t20_l0) + rk1m_Rpd_xz_o0_t21_l0) + rk1m_Rpd_xz_o0_t22_l0) + rk1m_Rpd_xz_o0_t23_l0) + rk1m_Rpd_xz_o0_t24_l0) + rk1m_Rpd_xz_o0_t25_l0;
						rk1m_Rpd_xy_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t4_m0_l0;
						rk1m_Rpd_xy_o0_t24_l0 = (-0.75 * vector(rk1gtd_xy, i, j, k) * rk1gtu_zz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xy_o0_t23_l0 = (-0.75 * vector(rk1gtd_xy, i, j, k) * rk1gtu_yy * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xy_o0_t22_l0 = (-0.75 * vector(rk1gtd_xy, i, j, k) * rk1gtu_xx * rk1inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xy_o0_t21_l0 = rk1inv_chi * rk1Zu_z * vector(rk1gtd_yz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xy_o0_t20_l0 = rk1inv_chi * rk1Zu_y * vector(rk1gtd_yy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xy_o0_t19_l0 = rk1inv_chi * rk1Zu_x * vector(rk1gtd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xy_o0_t18_l0 = rk1inv_chi * rk1Zu_z * vector(rk1gtd_xz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xy_o0_t17_l0 = rk1inv_chi * rk1Zu_y * vector(rk1gtd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xy_o0_t16_l0 = rk1inv_chi * rk1Zu_x * vector(rk1gtd_xx, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xy_o0_t15_l0 = (-1.500000000000000 * vector(rk1gtd_xy, i, j, k) * rk1gtu_yz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xy_o0_t14_l0 = (-1.500000000000000 * vector(rk1gtd_xy, i, j, k) * rk1gtu_xz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xy_o0_t13_l0 = (-1.500000000000000 * vector(rk1gtd_xy, i, j, k) * rk1gtu_xy * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xy_o0_t12_l0 = (-0.5 * rk1Ct_zxy) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xy_o0_t11_l0 = (-0.5 * rk1Ct_yxy) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xy_o0_t10_l0 = (-0.5 * rk1Ct_xxy) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xy_o0_t9_l0 = 0.5 * vector(rk1gtd_xy, i, j, k) * rk1gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk1m_Rpd_xy_o0_t8_l0 = vector(rk1gtd_xy, i, j, k) * rk1gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk1m_Rpd_xy_o0_t7_l0 = 0.5 * vector(rk1gtd_xy, i, j, k) * rk1gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk1m_Rpd_xy_o0_t6_l0 = vector(rk1gtd_xy, i, j, k) * rk1gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk1m_Rpd_xy_o0_t5_l0 = vector(rk1gtd_xy, i, j, k) * rk1gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk1m_Rpd_xy_o0_t4_l0 = 0.5 * vector(rk1gtd_xy, i, j, k) * rk1gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk1m_Rpd_xy_o0_t3_l0 = (-0.5 * vector(rk1gtd_xy, i, j, k) * vector(rk1Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_xy_o0_t2_l0 = (-0.5 * vector(rk1gtd_xy, i, j, k) * vector(rk1Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_xy_o0_t1_l0 = (-0.5 * vector(rk1gtd_xy, i, j, k) * vector(rk1Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_xy_o0_t0_l0 = (-0.25 * rk1inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1Rpd_xy = ((((((((((((((((((((((((rk1m_Rpd_xy_o0_t0_l0 + rk1m_Rpd_xy_o0_t1_l0) + rk1m_Rpd_xy_o0_t2_l0) + rk1m_Rpd_xy_o0_t3_l0) + rk1m_Rpd_xy_o0_t4_l0) + rk1m_Rpd_xy_o0_t5_l0) + rk1m_Rpd_xy_o0_t6_l0) + rk1m_Rpd_xy_o0_t7_l0) + rk1m_Rpd_xy_o0_t8_l0) + rk1m_Rpd_xy_o0_t9_l0) + rk1m_Rpd_xy_o0_t10_l0) + rk1m_Rpd_xy_o0_t11_l0) + rk1m_Rpd_xy_o0_t12_l0) + rk1m_Rpd_xy_o0_t13_l0) + rk1m_Rpd_xy_o0_t14_l0) + rk1m_Rpd_xy_o0_t15_l0) + rk1m_Rpd_xy_o0_t16_l0) + rk1m_Rpd_xy_o0_t17_l0) + rk1m_Rpd_xy_o0_t18_l0) + rk1m_Rpd_xy_o0_t19_l0) + rk1m_Rpd_xy_o0_t20_l0) + rk1m_Rpd_xy_o0_t21_l0) + rk1m_Rpd_xy_o0_t22_l0) + rk1m_Rpd_xy_o0_t23_l0) + rk1m_Rpd_xy_o0_t24_l0) + rk1m_Rpd_xy_o0_t25_l0;
						rk1m_Rpd_zz_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t8_m0_l0;
						rk1m_Rpd_zz_o0_t21_l0 = (-0.75 * vector(rk1gtd_zz, i, j, k) * rk1gtu_zz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_zz_o0_t20_l0 = (-0.75 * vector(rk1gtd_zz, i, j, k) * rk1gtu_yy * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_zz_o0_t19_l0 = (-0.75 * vector(rk1gtd_zz, i, j, k) * rk1gtu_xx * rk1inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_zz_o0_t18_l0 = 2.0 * rk1inv_chi * rk1Zu_z * vector(rk1gtd_zz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_zz_o0_t17_l0 = 2.0 * rk1inv_chi * rk1Zu_y * vector(rk1gtd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_zz_o0_t16_l0 = 2.0 * rk1inv_chi * rk1Zu_x * vector(rk1gtd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_zz_o0_t15_l0 = (-0.25 * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_zz_o0_t14_l0 = (-0.5 * rk1Ct_zzz) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_zz_o0_t13_l0 = (-0.5 * rk1Ct_yzz) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_zz_o0_t12_l0 = (-0.5 * rk1Ct_xzz) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_zz_o0_t11_l0 = vector(rk1gtd_zz, i, j, k) * rk1gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk1m_Rpd_zz_o0_t10_l0 = vector(rk1gtd_zz, i, j, k) * rk1gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk1m_Rpd_zz_o0_t9_l0 = 0.5 * vector(rk1gtd_zz, i, j, k) * rk1gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk1m_Rpd_zz_o0_t8_l0 = (-0.5 * vector(rk1gtd_zz, i, j, k) * vector(rk1Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Rpd_zz_o0_t7_l0 = (-0.5 * vector(rk1gtd_zz, i, j, k) * vector(rk1Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_zz_o0_t6_l0 = (-0.5 * vector(rk1gtd_zz, i, j, k) * vector(rk1Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_zz_o0_t5_l0 = 0.5 * vector(rk1gtd_zz, i, j, k) * rk1gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk1m_Rpd_zz_o0_t4_l0 = vector(rk1gtd_zz, i, j, k) * rk1gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk1m_Rpd_zz_o0_t3_l0 = 0.5 * vector(rk1gtd_zz, i, j, k) * rk1gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk1m_Rpd_zz_o0_t2_l0 = (-1.500000000000000 * vector(rk1gtd_zz, i, j, k) * rk1gtu_yz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Rpd_zz_o0_t1_l0 = (-1.500000000000000 * vector(rk1gtd_zz, i, j, k) * rk1gtu_xz * rk1inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Rpd_zz_o0_t0_l0 = (-1.500000000000000 * vector(rk1gtd_zz, i, j, k) * rk1gtu_xy * rk1inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1Rpd_zz = (((((((((((((((((((((rk1m_Rpd_zz_o0_t0_l0 + rk1m_Rpd_zz_o0_t1_l0) + rk1m_Rpd_zz_o0_t2_l0) + rk1m_Rpd_zz_o0_t3_l0) + rk1m_Rpd_zz_o0_t4_l0) + rk1m_Rpd_zz_o0_t5_l0) + rk1m_Rpd_zz_o0_t6_l0) + rk1m_Rpd_zz_o0_t7_l0) + rk1m_Rpd_zz_o0_t8_l0) + rk1m_Rpd_zz_o0_t9_l0) + rk1m_Rpd_zz_o0_t10_l0) + rk1m_Rpd_zz_o0_t11_l0) + rk1m_Rpd_zz_o0_t12_l0) + rk1m_Rpd_zz_o0_t13_l0) + rk1m_Rpd_zz_o0_t14_l0) + rk1m_Rpd_zz_o0_t15_l0) + rk1m_Rpd_zz_o0_t16_l0) + rk1m_Rpd_zz_o0_t17_l0) + rk1m_Rpd_zz_o0_t18_l0) + rk1m_Rpd_zz_o0_t19_l0) + rk1m_Rpd_zz_o0_t20_l0) + rk1m_Rpd_zz_o0_t21_l0) + rk1m_Rpd_zz_o0_t22_l0;
						rk1Rscalar = (rk1chi_max * rk1Rtd_xx + rk1Rpd_xx) * rk1gtu_xx + 2.0 * (rk1chi_max * rk1Rtd_xy + rk1Rpd_xy) * rk1gtu_xy + 2.0 * (rk1chi_max * rk1Rtd_xz + rk1Rpd_xz) * rk1gtu_xz + (rk1chi_max * rk1Rtd_yy + rk1Rpd_yy) * rk1gtu_yy + 2.0 * (rk1chi_max * rk1Rtd_yz + rk1Rpd_yz) * rk1gtu_yz + (rk1chi_max * rk1Rtd_zz + rk1Rpd_zz) * rk1gtu_zz;
						rk1Bfvf = vector(rk1Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk1Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk1Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
						rk1Sfuu_zz = rk1Sfu_z * rk1vfu_z + rk1chi_max * rk1gtu_zz * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_zz) + vector(rk1Bfu_z, i, j, k) * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_zz) + 2.0 * vector(rk1Bfu_z, i, j, k) * rk1vfu_z)) / rk1sdetg);
						rk1Tf4u_zz = rk1Sfuu_zz / rk1sdetg + (-(rk1Sfu_z * vector(rk1Betau_z, i, j, k)) / (rk1sdetg * vector(rk1Alpha, i, j, k))) + (-vector(rk1Betau_z, i, j, k) * rk1Tf4u_tz);
						rk1Tu_zz = rk1Tf4u_zz;
						rk1Sfuu_yz = 0.5 * rk1Sfu_y * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_y + rk1chi_max * rk1gtu_yz * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k) * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yz) + vector(rk1Bfu_y, i, j, k) * rk1vfu_z + vector(rk1Bfu_z, i, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Tf4u_yz = rk1Sfuu_yz / rk1sdetg + (-(rk1Sfu_y * vector(rk1Betau_z, i, j, k)) / (rk1sdetg * vector(rk1Alpha, i, j, k))) + (-vector(rk1Betau_y, i, j, k) * rk1Tf4u_tz);
						rk1Tu_yz = rk1Tf4u_yz;
						rk1Sfuu_yy = rk1Sfu_y * rk1vfu_y + rk1chi_max * rk1gtu_yy * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_yy) + vector(rk1Bfu_y, i, j, k) * vector(rk1Bfu_y, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_yy) + 2.0 * vector(rk1Bfu_y, i, j, k) * rk1vfu_y)) / rk1sdetg);
						rk1Tf4u_yy = rk1Sfuu_yy / rk1sdetg + (-(rk1Sfu_y * vector(rk1Betau_y, i, j, k)) / (rk1sdetg * vector(rk1Alpha, i, j, k))) + (-vector(rk1Betau_y, i, j, k) * rk1Tf4u_ty);
						rk1Tu_yy = rk1Tf4u_yy;
						rk1Sfuu_xy = 0.5 * rk1Sfu_x * rk1vfu_y + 0.5 * rk1Sfu_y * rk1vfu_x + rk1chi_max * rk1gtu_xy * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k) * vector(rk1Bfu_y, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xy) + vector(rk1Bfu_x, i, j, k) * rk1vfu_y + vector(rk1Bfu_y, i, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_yy = rk1inv_chi * (rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_yy * vector(rk1gtd_yy, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k));
						rk1Tf4u_xy = rk1Sfuu_xy / rk1sdetg + (-(rk1Sfu_x * vector(rk1Betau_y, i, j, k)) / (rk1sdetg * vector(rk1Alpha, i, j, k))) + (-vector(rk1Betau_x, i, j, k) * rk1Tf4u_ty);
						rk1Tu_xy = rk1Tf4u_xy;
						rk1Sfuu_xx = rk1Sfu_x * rk1vfu_x + rk1chi_max * rk1gtu_xx * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xx) + vector(rk1Bfu_x, i, j, k) * vector(rk1Bfu_x, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xx) + 2.0 * vector(rk1Bfu_x, i, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Tf4u_xx = rk1Sfuu_xx / rk1sdetg + (-(rk1Sfu_x * vector(rk1Betau_x, i, j, k)) / (rk1sdetg * vector(rk1Alpha, i, j, k))) + (-vector(rk1Betau_x, i, j, k) * rk1Tf4u_tx);
						rk1Tu_xx = rk1Tf4u_xx;
						rk1Sfuu_xz = 0.5 * rk1Sfu_x * rk1vfu_z + 0.5 * rk1Sfu_z * rk1vfu_x + rk1chi_max * rk1gtu_xz * rk1sdetg * vector(pf, i, j, k) + (-((-0.5 * rk1sqBf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k) * vector(rk1Bfu_z, i, j, k)) / (rk1sdetg * rk1sqW)) + (-0.5 * (rk1Bfvf * ((-rk1Bfvf * rk1chi_max * rk1gtu_xz) + vector(rk1Bfu_x, i, j, k) * rk1vfu_z + vector(rk1Bfu_z, i, j, k) * rk1vfu_x)) / rk1sdetg);
						rk1Sfud_xx = rk1inv_chi * (rk1Sfuu_xx * vector(rk1gtd_xx, i, j, k) + rk1Sfuu_xy * vector(rk1gtd_xy, i, j, k) + rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k));
						rk1Tf4u_xz = rk1Sfuu_xz / rk1sdetg + (-(rk1Sfu_x * vector(rk1Betau_z, i, j, k)) / (rk1sdetg * vector(rk1Alpha, i, j, k))) + (-vector(rk1Betau_x, i, j, k) * rk1Tf4u_tz);
						rk1Tu_xz = rk1Tf4u_xz;
						rk1Sfud_zz = rk1inv_chi * (rk1Sfuu_xz * vector(rk1gtd_xz, i, j, k) + rk1Sfuu_yz * vector(rk1gtd_yz, i, j, k) + rk1Sfuu_zz * vector(rk1gtd_zz, i, j, k));
						rk1trSf = rk1Sfud_xx + rk1Sfud_yy + rk1Sfud_zz;
						rk1Tu_tt = rk1Tf4u_tt;
						rk1pTtd_ADM_xz = rk1Betatd_x * rk1Betatd_z * rk1Tu_tt + rk1Betatd_x * (vector(rk1gtd_xz, i, j, k) * rk1Tu_tx + vector(rk1gtd_yz, i, j, k) * rk1Tu_ty + vector(rk1gtd_zz, i, j, k) * rk1Tu_tz) + rk1Betatd_z * (vector(rk1gtd_xx, i, j, k) * rk1Tu_tx + vector(rk1gtd_xy, i, j, k) * rk1Tu_ty + vector(rk1gtd_xz, i, j, k) * rk1Tu_tz) + vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_xz, i, j, k) * rk1Tu_xx + vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yz, i, j, k) * rk1Tu_xy + vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_zz, i, j, k) * rk1Tu_xz + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k) * rk1Tu_xy + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_yz, i, j, k) * rk1Tu_yy + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_zz, i, j, k) * rk1Tu_yz + (vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k)) * rk1Tu_xz + vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_yz, i, j, k) * rk1Tu_yz + vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_zz, i, j, k) * rk1Tu_zz;
						rk1pTtd_ADM_xy = rk1Betatd_x * rk1Betatd_y * rk1Tu_tt + rk1Betatd_x * (vector(rk1gtd_xy, i, j, k) * rk1Tu_tx + vector(rk1gtd_yy, i, j, k) * rk1Tu_ty + vector(rk1gtd_yz, i, j, k) * rk1Tu_tz) + rk1Betatd_y * (vector(rk1gtd_xx, i, j, k) * rk1Tu_tx + vector(rk1gtd_xy, i, j, k) * rk1Tu_ty + vector(rk1gtd_xz, i, j, k) * rk1Tu_tz) + vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_xy, i, j, k) * rk1Tu_xx + vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yy, i, j, k) * rk1Tu_xy + vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_yz, i, j, k) * rk1Tu_xz + (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k)) * rk1Tu_xy + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_yy, i, j, k) * rk1Tu_yy + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_yz, i, j, k) * rk1Tu_yz + vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xy, i, j, k) * rk1Tu_xz + vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_yy, i, j, k) * rk1Tu_yz + vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_yz, i, j, k) * rk1Tu_zz;
						rk1pTtd_ADM_xx = (rk1Betatd_x * rk1Betatd_x) * rk1Tu_tt + 2.0 * rk1Betatd_x * (vector(rk1gtd_xx, i, j, k) * rk1Tu_tx + vector(rk1gtd_xy, i, j, k) * rk1Tu_ty + vector(rk1gtd_xz, i, j, k) * rk1Tu_tz) + (vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_xx, i, j, k)) * rk1Tu_xx + 2.0 * vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_xy, i, j, k) * rk1Tu_xy + 2.0 * vector(rk1gtd_xx, i, j, k) * vector(rk1gtd_xz, i, j, k) * rk1Tu_xz + (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k)) * rk1Tu_yy + 2.0 * vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k) * rk1Tu_yz + (vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k)) * rk1Tu_zz;
						rk1d_Psi1_xx_o0_t5_m0_l0 = (-8.0) * M_PI * rk1inv_chi * vector(rk1Alpha, i, j, k) * rk1pTtd_ADM_xx + vector(rk1Alpha, i, j, k) * rk1Rpd_xx + rk1chi_max * vector(rk1Alpha, i, j, k) * rk1Rtd_xx;
						rk1m_Psi1_xx_o0_t4_l0 = (-rk1chi_max * rk1inv_chi) * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Psi1_xx_o0_t3_l0 = rk1chi_max * rk1Ct_zxx * d_Psi1_xx_o0_t3_m0_l0;
						rk1m_Psi1_xx_o0_t2_l0 = rk1chi_max * rk1Ct_yxx * d_Psi1_xx_o0_t2_m0_l0;
						rk1m_Psi1_xx_o0_t1_l0 = rk1chi_max * rk1Ct_xxx * d_Psi1_xx_o0_t1_m0_l0;
						rk1m_Psi1_xx_o0_t0_l0 = (-rk1chi_max) * d_Psi1_xx_o0_t0_m0_l0;
						rk1Psi1_xx = ((((rk1m_Psi1_xx_o0_t0_l0 + rk1m_Psi1_xx_o0_t1_l0) + rk1m_Psi1_xx_o0_t2_l0) + rk1m_Psi1_xx_o0_t3_l0) + rk1m_Psi1_xx_o0_t4_l0) + rk1d_Psi1_xx_o0_t5_m0_l0;
						rk1d_Psi1_xy_o0_t6_m0_l0 = (-8.0) * M_PI * rk1inv_chi * vector(rk1Alpha, i, j, k) * rk1pTtd_ADM_xy + vector(rk1Alpha, i, j, k) * rk1Rpd_xy + rk1chi_max * vector(rk1Alpha, i, j, k) * rk1Rtd_xy;
						rk1m_Psi1_xy_o0_t5_l0 = (-0.5 * rk1chi_max * rk1inv_chi) * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Psi1_xy_o0_t4_l0 = (-0.5 * rk1chi_max * rk1inv_chi) * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Psi1_xy_o0_t3_l0 = rk1chi_max * rk1Ct_zxy * d_Psi1_xx_o0_t3_m0_l0;
						rk1m_Psi1_xy_o0_t2_l0 = rk1chi_max * rk1Ct_yxy * d_Psi1_xx_o0_t2_m0_l0;
						rk1m_Psi1_xy_o0_t1_l0 = rk1chi_max * rk1Ct_xxy * d_Psi1_xx_o0_t1_m0_l0;
						rk1m_Psi1_xy_o0_t0_l0 = (-rk1chi_max) * d_Psi1_xy_o0_t0_m0_l0;
						rk1Psi1_xy = (((((rk1m_Psi1_xy_o0_t0_l0 + rk1m_Psi1_xy_o0_t1_l0) + rk1m_Psi1_xy_o0_t2_l0) + rk1m_Psi1_xy_o0_t3_l0) + rk1m_Psi1_xy_o0_t4_l0) + rk1m_Psi1_xy_o0_t5_l0) + rk1d_Psi1_xy_o0_t6_m0_l0;
						rk1d_Psi1_xz_o0_t6_m0_l0 = (-8.0) * M_PI * rk1inv_chi * vector(rk1Alpha, i, j, k) * rk1pTtd_ADM_xz + vector(rk1Alpha, i, j, k) * rk1Rpd_xz + rk1chi_max * vector(rk1Alpha, i, j, k) * rk1Rtd_xz;
						rk1m_Psi1_xz_o0_t5_l0 = (-0.5 * rk1chi_max * rk1inv_chi) * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk1m_Psi1_xz_o0_t4_l0 = (-0.5 * rk1chi_max * rk1inv_chi) * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Psi1_xz_o0_t3_l0 = rk1chi_max * rk1Ct_zxz * d_Psi1_xx_o0_t3_m0_l0;
						rk1m_Psi1_xz_o0_t2_l0 = rk1chi_max * rk1Ct_yxz * d_Psi1_xx_o0_t2_m0_l0;
						rk1m_Psi1_xz_o0_t1_l0 = rk1chi_max * rk1Ct_xxz * d_Psi1_xx_o0_t1_m0_l0;
						rk1m_Psi1_xz_o0_t0_l0 = (-rk1chi_max) * d_Psi1_xz_o0_t0_m0_l0;
						rk1Psi1_xz = (((((rk1m_Psi1_xz_o0_t0_l0 + rk1m_Psi1_xz_o0_t1_l0) + rk1m_Psi1_xz_o0_t2_l0) + rk1m_Psi1_xz_o0_t3_l0) + rk1m_Psi1_xz_o0_t4_l0) + rk1m_Psi1_xz_o0_t5_l0) + rk1d_Psi1_xz_o0_t6_m0_l0;
						rk1Jtd_ADM_z = vector(rk1Alpha, i, j, k) * ((vector(rk1Betau_x, i, j, k) * rk1Tu_tt + rk1Tu_tx) * vector(rk1gtd_xz, i, j, k) + (vector(rk1Betau_y, i, j, k) * rk1Tu_tt + rk1Tu_ty) * vector(rk1gtd_yz, i, j, k) + (vector(rk1Betau_z, i, j, k) * rk1Tu_tt + rk1Tu_tz) * vector(rk1gtd_zz, i, j, k));
						rk1pTtd_ADM_zz = (rk1Betatd_z * rk1Betatd_z) * rk1Tu_tt + 2.0 * rk1Betatd_z * (vector(rk1gtd_xz, i, j, k) * rk1Tu_tx + vector(rk1gtd_yz, i, j, k) * rk1Tu_ty + vector(rk1gtd_zz, i, j, k) * rk1Tu_tz) + (vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_xz, i, j, k)) * rk1Tu_xx + 2.0 * vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_yz, i, j, k) * rk1Tu_xy + 2.0 * vector(rk1gtd_xz, i, j, k) * vector(rk1gtd_zz, i, j, k) * rk1Tu_xz + (vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k)) * rk1Tu_yy + 2.0 * vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_zz, i, j, k) * rk1Tu_yz + (vector(rk1gtd_zz, i, j, k) * vector(rk1gtd_zz, i, j, k)) * rk1Tu_zz;
						rk1d_Psi1_zz_o0_t5_m0_l0 = (-8.0) * M_PI * rk1inv_chi * vector(rk1Alpha, i, j, k) * rk1pTtd_ADM_zz + vector(rk1Alpha, i, j, k) * rk1Rpd_zz + rk1chi_max * vector(rk1Alpha, i, j, k) * rk1Rtd_zz;
						rk1m_Psi1_zz_o0_t4_l0 = (-rk1chi_max * rk1inv_chi) * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Psi1_zz_o0_t3_l0 = rk1chi_max * rk1Ct_zzz * d_Psi1_xx_o0_t3_m0_l0;
						rk1m_Psi1_zz_o0_t2_l0 = rk1chi_max * rk1Ct_yzz * d_Psi1_xx_o0_t2_m0_l0;
						rk1m_Psi1_zz_o0_t1_l0 = rk1chi_max * rk1Ct_xzz * d_Psi1_xx_o0_t1_m0_l0;
						rk1m_Psi1_zz_o0_t0_l0 = (-rk1chi_max) * d_Psi1_zz_o0_t0_m0_l0;
						rk1Psi1_zz = ((((rk1m_Psi1_zz_o0_t0_l0 + rk1m_Psi1_zz_o0_t1_l0) + rk1m_Psi1_zz_o0_t2_l0) + rk1m_Psi1_zz_o0_t3_l0) + rk1m_Psi1_zz_o0_t4_l0) + rk1d_Psi1_zz_o0_t5_m0_l0;
						rk1Jtd_ADM_y = vector(rk1Alpha, i, j, k) * ((vector(rk1Betau_x, i, j, k) * rk1Tu_tt + rk1Tu_tx) * vector(rk1gtd_xy, i, j, k) + (vector(rk1Betau_y, i, j, k) * rk1Tu_tt + rk1Tu_ty) * vector(rk1gtd_yy, i, j, k) + (vector(rk1Betau_z, i, j, k) * rk1Tu_tt + rk1Tu_tz) * vector(rk1gtd_yz, i, j, k));
						rk1Jtd_ADM_x = vector(rk1Alpha, i, j, k) * ((vector(rk1Betau_x, i, j, k) * rk1Tu_tt + rk1Tu_tx) * vector(rk1gtd_xx, i, j, k) + (vector(rk1Betau_y, i, j, k) * rk1Tu_tt + rk1Tu_ty) * vector(rk1gtd_xy, i, j, k) + (vector(rk1Betau_z, i, j, k) * rk1Tu_tt + rk1Tu_tz) * vector(rk1gtd_xz, i, j, k));
						rk1pTtd_ADM_yz = rk1Betatd_y * rk1Betatd_z * rk1Tu_tt + rk1Betatd_y * (vector(rk1gtd_xz, i, j, k) * rk1Tu_tx + vector(rk1gtd_yz, i, j, k) * rk1Tu_ty + vector(rk1gtd_zz, i, j, k) * rk1Tu_tz) + rk1Betatd_z * (vector(rk1gtd_xy, i, j, k) * rk1Tu_tx + vector(rk1gtd_yy, i, j, k) * rk1Tu_ty + vector(rk1gtd_yz, i, j, k) * rk1Tu_tz) + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xz, i, j, k) * rk1Tu_xx + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_yz, i, j, k) * rk1Tu_xy + vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_zz, i, j, k) * rk1Tu_xz + vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_xz, i, j, k) * rk1Tu_xy + vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_yz, i, j, k) * rk1Tu_yy + vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_zz, i, j, k) * rk1Tu_yz + vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_xz, i, j, k) * rk1Tu_xz + (vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k)) * rk1Tu_yz + vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_zz, i, j, k) * rk1Tu_zz;
						rk1d_Psi1_yz_o0_t6_m0_l0 = (-8.0) * M_PI * rk1inv_chi * vector(rk1Alpha, i, j, k) * rk1pTtd_ADM_yz + vector(rk1Alpha, i, j, k) * rk1Rpd_yz + rk1chi_max * vector(rk1Alpha, i, j, k) * rk1Rtd_yz;
						rk1m_Psi1_yz_o0_t5_l0 = (-0.5 * rk1chi_max * rk1inv_chi) * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Psi1_yz_o0_t4_l0 = (-0.5 * rk1chi_max * rk1inv_chi) * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk1m_Psi1_yz_o0_t3_l0 = rk1chi_max * rk1Ct_zyz * d_Psi1_xx_o0_t3_m0_l0;
						rk1m_Psi1_yz_o0_t2_l0 = rk1chi_max * rk1Ct_yyz * d_Psi1_xx_o0_t2_m0_l0;
						rk1m_Psi1_yz_o0_t1_l0 = rk1chi_max * rk1Ct_xyz * d_Psi1_xx_o0_t1_m0_l0;
						rk1m_Psi1_yz_o0_t0_l0 = (-rk1chi_max) * d_Psi1_yz_o0_t0_m0_l0;
						rk1Psi1_yz = (((((rk1m_Psi1_yz_o0_t0_l0 + rk1m_Psi1_yz_o0_t1_l0) + rk1m_Psi1_yz_o0_t2_l0) + rk1m_Psi1_yz_o0_t3_l0) + rk1m_Psi1_yz_o0_t4_l0) + rk1m_Psi1_yz_o0_t5_l0) + rk1d_Psi1_yz_o0_t6_m0_l0;
						rk1pTtd_ADM_yy = (rk1Betatd_y * rk1Betatd_y) * rk1Tu_tt + 2.0 * rk1Betatd_y * (vector(rk1gtd_xy, i, j, k) * rk1Tu_tx + vector(rk1gtd_yy, i, j, k) * rk1Tu_ty + vector(rk1gtd_yz, i, j, k) * rk1Tu_tz) + (vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_xy, i, j, k)) * rk1Tu_xx + 2.0 * vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_yy, i, j, k) * rk1Tu_xy + 2.0 * vector(rk1gtd_xy, i, j, k) * vector(rk1gtd_yz, i, j, k) * rk1Tu_xz + (vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_yy, i, j, k)) * rk1Tu_yy + 2.0 * vector(rk1gtd_yy, i, j, k) * vector(rk1gtd_yz, i, j, k) * rk1Tu_yz + (vector(rk1gtd_yz, i, j, k) * vector(rk1gtd_yz, i, j, k)) * rk1Tu_zz;
						rk1tr_pT = rk1inv_chi * (rk1gtu_xx * rk1pTtd_ADM_xx + 2.0 * rk1gtu_xy * rk1pTtd_ADM_xy + 2.0 * rk1gtu_xz * rk1pTtd_ADM_xz + rk1gtu_yy * rk1pTtd_ADM_yy + 2.0 * rk1gtu_yz * rk1pTtd_ADM_yz + rk1gtu_zz * rk1pTtd_ADM_zz);
						rk1d_Psi1_yy_o0_t5_m0_l0 = (-8.0) * M_PI * rk1inv_chi * vector(rk1Alpha, i, j, k) * rk1pTtd_ADM_yy + vector(rk1Alpha, i, j, k) * rk1Rpd_yy + rk1chi_max * vector(rk1Alpha, i, j, k) * rk1Rtd_yy;
						rk1m_Psi1_yy_o0_t4_l0 = (-rk1chi_max * rk1inv_chi) * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk1m_Psi1_yy_o0_t3_l0 = rk1chi_max * rk1Ct_zyy * d_Psi1_xx_o0_t3_m0_l0;
						rk1m_Psi1_yy_o0_t2_l0 = rk1chi_max * rk1Ct_yyy * d_Psi1_xx_o0_t2_m0_l0;
						rk1m_Psi1_yy_o0_t1_l0 = rk1chi_max * rk1Ct_xyy * d_Psi1_xx_o0_t1_m0_l0;
						rk1m_Psi1_yy_o0_t0_l0 = (-rk1chi_max) * d_Psi1_yy_o0_t0_m0_l0;
						rk1Psi1_yy = ((((rk1m_Psi1_yy_o0_t0_l0 + rk1m_Psi1_yy_o0_t1_l0) + rk1m_Psi1_yy_o0_t2_l0) + rk1m_Psi1_yy_o0_t3_l0) + rk1m_Psi1_yy_o0_t4_l0) + rk1d_Psi1_yy_o0_t5_m0_l0;
						rk1trPsi1 = rk1Psi1_xx * rk1gtu_xx + 2.0 * rk1Psi1_xy * rk1gtu_xy + 2.0 * rk1Psi1_xz * rk1gtu_xz + rk1Psi1_yy * rk1gtu_yy + 2.0 * rk1Psi1_yz * rk1gtu_yz + rk1Psi1_zz * rk1gtu_zz;
						rk1Psi1TF_yz = (-0.3333333333333333 * rk1trPsi1 * vector(rk1gtd_yz, i, j, k)) + rk1Psi1_yz;
						rk1Psi1TF_yy = (-0.3333333333333333 * rk1trPsi1 * vector(rk1gtd_yy, i, j, k)) + rk1Psi1_yy;
						rk1Psi1TF_xy = (-0.3333333333333333 * rk1trPsi1 * vector(rk1gtd_xy, i, j, k)) + rk1Psi1_xy;
						rk1Psi1TF_xx = (-0.3333333333333333 * rk1trPsi1 * vector(rk1gtd_xx, i, j, k)) + rk1Psi1_xx;
						rk1Psi1TF_xz = (-0.3333333333333333 * rk1trPsi1 * vector(rk1gtd_xz, i, j, k)) + rk1Psi1_xz;
						rk1Psi1TF_zz = (-0.3333333333333333 * rk1trPsi1 * vector(rk1gtd_zz, i, j, k)) + rk1Psi1_zz;
						rk1kappa_cc = MIN(p_kappa_cc, p_kappa_cc * rk1decay_factor) / vector(rk1Alpha, i, j, k);
						rk1rho_ADM = (vector(rk1Alpha, i, j, k) * vector(rk1Alpha, i, j, k)) * rk1Tu_tt;
						d_Sfd_x_o0_t11_m0_l0 = vector(rk1Alpha, i, j, k) * rk1sdetg * rk1sqW * vector(qnu, i, j, k) * rk1vfu_x;
						d_Sfd_y_o0_t11_m0_l0 = vector(rk1Alpha, i, j, k) * rk1sdetg * rk1sqW * vector(qnu, i, j, k) * rk1vfu_y;
						d_Sfd_z_o0_t11_m0_l0 = vector(rk1Alpha, i, j, k) * rk1sdetg * rk1sqW * vector(qnu, i, j, k) * rk1vfu_z;
						d_DYf_o0_t0_m0_l0 = vector(rk1Alpha, i, j, k) * rk1sdetg * vector(rnu, i, j, k);
						d_tauf_o0_t3_m0_l0 = vector(rk1Alpha, i, j, k) * rk1sdetg * rk1sqW * vector(qnu, i, j, k);
						d_gtd_xx_o0_t3_m0_l0 = (-0.3333333333333333) * rk1kappa_cc * vector(rk1Alpha, i, j, k) * vector(rk1gtd_xx, i, j, k) * log(rk1detgtd) + (-2.0) * vector(rk1Alpha, i, j, k) * ((-0.3333333333333333 * rk1trAt * lambda_0 * vector(rk1gtd_xx, i, j, k)) + vector(rk1Atd_xx, i, j, k)) + (-0.6666666666666667) * vector(rk1gtd_xx, i, j, k) * rk1div_Beta;
						d_gtd_xy_o0_t6_m0_l0 = (-0.3333333333333333) * rk1kappa_cc * vector(rk1Alpha, i, j, k) * vector(rk1gtd_xy, i, j, k) * log(rk1detgtd) + (-2.0) * vector(rk1Alpha, i, j, k) * ((-0.3333333333333333 * rk1trAt * lambda_0 * vector(rk1gtd_xy, i, j, k)) + vector(rk1Atd_xy, i, j, k)) + (-0.6666666666666667) * vector(rk1gtd_xy, i, j, k) * rk1div_Beta;
						d_gtd_xz_o0_t6_m0_l0 = (-0.3333333333333333) * rk1kappa_cc * vector(rk1Alpha, i, j, k) * vector(rk1gtd_xz, i, j, k) * log(rk1detgtd) + (-2.0) * vector(rk1Alpha, i, j, k) * ((-0.3333333333333333 * rk1trAt * lambda_0 * vector(rk1gtd_xz, i, j, k)) + vector(rk1Atd_xz, i, j, k)) + (-0.6666666666666667) * vector(rk1gtd_xz, i, j, k) * rk1div_Beta;
						d_gtd_yy_o0_t3_m0_l0 = (-0.3333333333333333) * rk1kappa_cc * vector(rk1Alpha, i, j, k) * vector(rk1gtd_yy, i, j, k) * log(rk1detgtd) + (-2.0) * vector(rk1Alpha, i, j, k) * ((-0.3333333333333333 * rk1trAt * lambda_0 * vector(rk1gtd_yy, i, j, k)) + vector(rk1Atd_yy, i, j, k)) + (-0.6666666666666667) * vector(rk1gtd_yy, i, j, k) * rk1div_Beta;
						d_gtd_yz_o0_t6_m0_l0 = (-0.3333333333333333) * rk1kappa_cc * vector(rk1Alpha, i, j, k) * vector(rk1gtd_yz, i, j, k) * log(rk1detgtd) + (-2.0) * vector(rk1Alpha, i, j, k) * ((-0.3333333333333333 * rk1trAt * lambda_0 * vector(rk1gtd_yz, i, j, k)) + vector(rk1Atd_yz, i, j, k)) + (-0.6666666666666667) * vector(rk1gtd_yz, i, j, k) * rk1div_Beta;
						d_gtd_zz_o0_t3_m0_l0 = (-0.3333333333333333) * rk1kappa_cc * vector(rk1Alpha, i, j, k) * vector(rk1gtd_zz, i, j, k) * log(rk1detgtd) + (-2.0) * vector(rk1Alpha, i, j, k) * ((-0.3333333333333333 * rk1trAt * lambda_0 * vector(rk1gtd_zz, i, j, k)) + vector(rk1Atd_zz, i, j, k)) + (-0.6666666666666667) * vector(rk1gtd_zz, i, j, k) * rk1div_Beta;
						d_Atd_xx_o0_t3_m0_l0 = (-0.3333333333333333) * rk1kappa_cc * vector(rk1Alpha, i, j, k) * vector(rk1gtd_xx, i, j, k) * rk1trAt + vector(rk1Alpha, i, j, k) * (vector(rk1trK, i, j, k) * vector(rk1Atd_xx, i, j, k) + (-2.0 * vector(rk1Atd_xx, i, j, k) * rk1Atud_xx) + (-2.0 * vector(rk1Atd_xy, i, j, k) * rk1Atud_yx) + (-2.0 * vector(rk1Atd_xz, i, j, k) * rk1Atud_zx)) + rk1Psi1TF_xx + (-0.6666666666666667) * vector(rk1Atd_xx, i, j, k) * rk1div_Beta;
						d_Atd_xy_o0_t6_m0_l0 = (-0.3333333333333333) * rk1kappa_cc * vector(rk1Alpha, i, j, k) * vector(rk1gtd_xy, i, j, k) * rk1trAt + vector(rk1Alpha, i, j, k) * (vector(rk1trK, i, j, k) * vector(rk1Atd_xy, i, j, k) + (-2.0 * vector(rk1Atd_xx, i, j, k) * rk1Atud_xy) + (-2.0 * vector(rk1Atd_xy, i, j, k) * rk1Atud_yy) + (-2.0 * vector(rk1Atd_xz, i, j, k) * rk1Atud_zy)) + rk1Psi1TF_xy + (-0.6666666666666667) * vector(rk1Atd_xy, i, j, k) * rk1div_Beta;
						d_Atd_xz_o0_t6_m0_l0 = (-0.3333333333333333) * rk1kappa_cc * vector(rk1Alpha, i, j, k) * vector(rk1gtd_xz, i, j, k) * rk1trAt + vector(rk1Alpha, i, j, k) * (vector(rk1trK, i, j, k) * vector(rk1Atd_xz, i, j, k) + (-2.0 * vector(rk1Atd_xx, i, j, k) * rk1Atud_xz) + (-2.0 * vector(rk1Atd_xy, i, j, k) * rk1Atud_yz) + (-2.0 * vector(rk1Atd_xz, i, j, k) * rk1Atud_zz)) + rk1Psi1TF_xz + (-0.6666666666666667) * vector(rk1Atd_xz, i, j, k) * rk1div_Beta;
						d_Atd_yy_o0_t3_m0_l0 = (-0.3333333333333333) * rk1kappa_cc * vector(rk1Alpha, i, j, k) * vector(rk1gtd_yy, i, j, k) * rk1trAt + vector(rk1Alpha, i, j, k) * (vector(rk1trK, i, j, k) * vector(rk1Atd_yy, i, j, k) + (-2.0 * vector(rk1Atd_xy, i, j, k) * rk1Atud_xy) + (-2.0 * vector(rk1Atd_yy, i, j, k) * rk1Atud_yy) + (-2.0 * vector(rk1Atd_yz, i, j, k) * rk1Atud_zy)) + rk1Psi1TF_yy + (-0.6666666666666667) * vector(rk1Atd_yy, i, j, k) * rk1div_Beta;
						d_Atd_yz_o0_t6_m0_l0 = (-0.3333333333333333) * rk1kappa_cc * vector(rk1Alpha, i, j, k) * vector(rk1gtd_yz, i, j, k) * rk1trAt + vector(rk1Alpha, i, j, k) * (vector(rk1trK, i, j, k) * vector(rk1Atd_yz, i, j, k) + (-2.0 * vector(rk1Atd_xy, i, j, k) * rk1Atud_xz) + (-2.0 * vector(rk1Atd_yy, i, j, k) * rk1Atud_yz) + (-2.0 * vector(rk1Atd_yz, i, j, k) * rk1Atud_zz)) + rk1Psi1TF_yz + (-0.6666666666666667) * vector(rk1Atd_yz, i, j, k) * rk1div_Beta;
						d_Atd_zz_o0_t3_m0_l0 = (-0.3333333333333333) * rk1kappa_cc * vector(rk1Alpha, i, j, k) * vector(rk1gtd_zz, i, j, k) * rk1trAt + vector(rk1Alpha, i, j, k) * (vector(rk1trK, i, j, k) * vector(rk1Atd_zz, i, j, k) + (-2.0 * vector(rk1Atd_xz, i, j, k) * rk1Atud_xz) + (-2.0 * vector(rk1Atd_yz, i, j, k) * rk1Atud_yz) + (-2.0 * vector(rk1Atd_zz, i, j, k) * rk1Atud_zz)) + rk1Psi1TF_zz + (-0.6666666666666667) * vector(rk1Atd_zz, i, j, k) * rk1div_Beta;
						d_Gamh_x_o0_t24_m0_l0 = (-2.0) * vector(rk1Alpha, i, j, k) * rk1inv_chi * rk1Zu_x * (rk1kappa_z1 + 1.333333333333333 * vector(rk1theta, i, j, k) + 0.6666666666666667 * vector(rk1trK, i, j, k)) + 0.3333333333333333 * rk1gtu_xz * rk1d_div_Beta_z + 0.3333333333333333 * rk1gtu_xy * rk1d_div_Beta_y + 0.3333333333333333 * rk1gtu_xx * rk1d_div_Beta_x + (-50.26548245743669) * vector(rk1Alpha, i, j, k) * rk1inv_chi * (rk1gtu_xx * rk1Jtd_ADM_x + rk1gtu_xy * rk1Jtd_ADM_y + rk1gtu_xz * rk1Jtd_ADM_z) + 2.0 * vector(rk1Alpha, i, j, k) * (rk1Ct_xxx * rk1Atu_xx + 2.0 * rk1Ct_xxy * rk1Atu_xy + 2.0 * rk1Ct_xxz * rk1Atu_xz + rk1Ct_xyy * rk1Atu_yy + 2.0 * rk1Ct_xyz * rk1Atu_yz + rk1Ct_xzz * rk1Atu_zz) + 0.6666666666666667 * vector(rk1Gamh_x, i, j, k) * rk1div_Beta;
						d_Gamh_y_o0_t24_m0_l0 = (-2.0) * vector(rk1Alpha, i, j, k) * rk1inv_chi * rk1Zu_y * (rk1kappa_z1 + 1.333333333333333 * vector(rk1theta, i, j, k) + 0.6666666666666667 * vector(rk1trK, i, j, k)) + (-50.26548245743669) * vector(rk1Alpha, i, j, k) * rk1inv_chi * (rk1gtu_xy * rk1Jtd_ADM_x + rk1gtu_yy * rk1Jtd_ADM_y + rk1gtu_yz * rk1Jtd_ADM_z) + 0.3333333333333333 * rk1gtu_yy * rk1d_div_Beta_y + 0.3333333333333333 * rk1gtu_xy * rk1d_div_Beta_x + 2.0 * vector(rk1Alpha, i, j, k) * (rk1Ct_yxx * rk1Atu_xx + 2.0 * rk1Ct_yxy * rk1Atu_xy + 2.0 * rk1Ct_yxz * rk1Atu_xz + rk1Ct_yyy * rk1Atu_yy + 2.0 * rk1Ct_yyz * rk1Atu_yz + rk1Ct_yzz * rk1Atu_zz) + 0.6666666666666667 * vector(rk1Gamh_y, i, j, k) * rk1div_Beta + 0.3333333333333333 * rk1gtu_yz * rk1d_div_Beta_z;
						d_Gamh_z_o0_t24_m0_l0 = (-2.0) * vector(rk1Alpha, i, j, k) * rk1inv_chi * rk1Zu_z * (rk1kappa_z1 + 1.333333333333333 * vector(rk1theta, i, j, k) + 0.6666666666666667 * vector(rk1trK, i, j, k)) + (-50.26548245743669) * vector(rk1Alpha, i, j, k) * rk1inv_chi * (rk1gtu_xz * rk1Jtd_ADM_x + rk1gtu_yz * rk1Jtd_ADM_y + rk1gtu_zz * rk1Jtd_ADM_z) + 2.0 * vector(rk1Alpha, i, j, k) * (rk1Ct_zxx * rk1Atu_xx + 2.0 * rk1Ct_zxy * rk1Atu_xy + 2.0 * rk1Ct_zxz * rk1Atu_xz + rk1Ct_zyy * rk1Atu_yy + 2.0 * rk1Ct_zyz * rk1Atu_yz + rk1Ct_zzz * rk1Atu_zz) + 0.6666666666666667 * vector(rk1Gamh_z, i, j, k) * rk1div_Beta + 0.3333333333333333 * rk1gtu_zz * rk1d_div_Beta_z + 0.3333333333333333 * rk1gtu_yz * rk1d_div_Beta_y + 0.3333333333333333 * rk1gtu_xz * rk1d_div_Beta_x;
						d_Betau_x_o0_t0_m0_l0 = (Betau_x_0 - vector(rk1Betau_x, i, j, k)) * rk1feta + 0.75 * (vector(rk1Alpha, i, j, k) * lambda_f1 + lambda_f0) * vector(rk1Gamh_x, i, j, k);
						d_Betau_y_o0_t0_m0_l0 = (Betau_y_0 - vector(rk1Betau_y, i, j, k)) * rk1feta + 0.75 * (vector(rk1Alpha, i, j, k) * lambda_f1 + lambda_f0) * vector(rk1Gamh_y, i, j, k);
						d_Betau_z_o0_t0_m0_l0 = (Betau_z_0 - vector(rk1Betau_z, i, j, k)) * rk1feta + 0.75 * (vector(rk1Alpha, i, j, k) * lambda_f1 + lambda_f0) * vector(rk1Gamh_z, i, j, k);
						d_Alpha_o0_t0_m0_l0 = (-2.0) * (vector(rk1Alpha, i, j, k) * lambda_f3 + lambda_f2) * vector(rk1Alpha, i, j, k) * (vector(rk1trK, i, j, k) - trK0);
						d_chi_o0_t0_m0_l0 = 0.6666666666666667 * rk1chi_max * (vector(rk1Alpha, i, j, k) * (2.0 * vector(rk1theta, i, j, k) + vector(rk1trK, i, j, k)) - rk1div_Beta);
						d_trK_o0_t21_m0_l0 = vector(rk1Alpha, i, j, k) * rk1kappa_z1 * (1.0 - rk1kappa_z2) * vector(rk1theta, i, j, k) + 0.3333333333333333 * vector(rk1Alpha, i, j, k) * ((2.0 * vector(rk1theta, i, j, k) + vector(rk1trK, i, j, k)) * (2.0 * vector(rk1theta, i, j, k) + vector(rk1trK, i, j, k))) + vector(rk1Alpha, i, j, k) * (vector(rk1Atd_xx, i, j, k) * rk1Atu_xx + 2.0 * vector(rk1Atd_xy, i, j, k) * rk1Atu_xy + 2.0 * vector(rk1Atd_xz, i, j, k) * rk1Atu_xz + vector(rk1Atd_yy, i, j, k) * rk1Atu_yy + 2.0 * vector(rk1Atd_yz, i, j, k) * rk1Atu_yz + vector(rk1Atd_zz, i, j, k) * rk1Atu_zz) + 4.0 * vector(rk1Alpha, i, j, k) * M_PI * (rk1rho_ADM + rk1tr_pT);
						d_theta_o0_t3_m0_l0 = (-vector(rk1Alpha, i, j, k) * rk1kappa_z1 * (2.0 + rk1kappa_z2) * vector(rk1theta, i, j, k)) + (-8.0) * vector(rk1Alpha, i, j, k) * M_PI * rk1rho_ADM + 0.5 * vector(rk1Alpha, i, j, k) * (rk1Rscalar + 0.6666666666666667 * (vector(rk1trK, i, j, k) * vector(rk1trK, i, j, k)) + 0.6666666666666667 * vector(rk1theta, i, j, k) * ((-2.0 * vector(rk1theta, i, j, k)) + vector(rk1trK, i, j, k)) + (-vector(rk1Atd_xx, i, j, k) * rk1Atu_xx) + (-2.0 * vector(rk1Atd_xy, i, j, k) * rk1Atu_xy) + (-2.0 * vector(rk1Atd_xz, i, j, k) * rk1Atu_xz) + (-vector(rk1Atd_yy, i, j, k) * rk1Atu_yy) + (-2.0 * vector(rk1Atd_yz, i, j, k) * rk1Atu_yz) + (-vector(rk1Atd_zz, i, j, k) * rk1Atu_zz));
						RHS_Bfu_x = -vector(rk1phif, i, j, k) * vector(rk1Alpha, i, j, k) * rk1chi_max * rk1Gamt_x;
						RHS_Bfu_y = -vector(rk1phif, i, j, k) * vector(rk1Alpha, i, j, k) * rk1chi_max * rk1Gamt_y;
						RHS_Bfu_z = -vector(rk1phif, i, j, k) * vector(rk1Alpha, i, j, k) * rk1chi_max * rk1Gamt_z;
						RHS_tauf = (vector(rk1Alpha, i, j, k) * (rk1Sfuu_xx * vector(rk1Atd_xx, i, j, k) + 2.0 * rk1Sfuu_xy * vector(rk1Atd_xy, i, j, k) + 2.0 * rk1Sfuu_xz * vector(rk1Atd_xz, i, j, k) + rk1Sfuu_yy * vector(rk1Atd_yy, i, j, k) + 2.0 * rk1Sfuu_yz * vector(rk1Atd_yz, i, j, k) + rk1Sfuu_zz * vector(rk1Atd_zz, i, j, k))) / rk1chi_max + 0.3333333333333333 * vector(rk1Alpha, i, j, k) * rk1trSf * vector(rk1trK, i, j, k);
						RHS_phif = (-vector(rk1Alpha, i, j, k) * rk1kappa_f * vector(rk1phif, i, j, k)) + (-vector(rk1Alpha, i, j, k) * vector(rk1phif, i, j, k) * vector(rk1trK, i, j, k));
						RHS_Sfd_x = -FVM_i(FluxpiSfd_x, FluxniSfd_x, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_x = RHS_Sfd_x - FVM_j(FluxpjSfd_x, FluxnjSfd_x, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_x = RHS_Sfd_x - FVM_k(FluxpkSfd_x, FluxnkSfd_x, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_y = -FVM_i(FluxpiSfd_y, FluxniSfd_y, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_y = RHS_Sfd_y - FVM_j(FluxpjSfd_y, FluxnjSfd_y, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_y = RHS_Sfd_y - FVM_k(FluxpkSfd_y, FluxnkSfd_y, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_z = -FVM_i(FluxpiSfd_z, FluxniSfd_z, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_z = RHS_Sfd_z - FVM_j(FluxpjSfd_z, FluxnjSfd_z, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_z = RHS_Sfd_z - FVM_k(FluxpkSfd_z, FluxnkSfd_z, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_x = RHS_Bfu_x - FVM_i(FluxpiBfu_x, FluxniBfu_x, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_x = RHS_Bfu_x - FVM_j(FluxpjBfu_x, FluxnjBfu_x, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_x = RHS_Bfu_x - FVM_k(FluxpkBfu_x, FluxnkBfu_x, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_y = RHS_Bfu_y - FVM_i(FluxpiBfu_y, FluxniBfu_y, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_y = RHS_Bfu_y - FVM_j(FluxpjBfu_y, FluxnjBfu_y, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_y = RHS_Bfu_y - FVM_k(FluxpkBfu_y, FluxnkBfu_y, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_z = RHS_Bfu_z - FVM_i(FluxpiBfu_z, FluxniBfu_z, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_z = RHS_Bfu_z - FVM_j(FluxpjBfu_z, FluxnjBfu_z, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_z = RHS_Bfu_z - FVM_k(FluxpkBfu_z, FluxnkBfu_z, dx, simPlat_dt, ilast, jlast);
						RHS_Df = -FVM_i(FluxpiDf, FluxniDf, dx, simPlat_dt, ilast, jlast);
						RHS_Df = RHS_Df - FVM_j(FluxpjDf, FluxnjDf, dx, simPlat_dt, ilast, jlast);
						RHS_Df = RHS_Df - FVM_k(FluxpkDf, FluxnkDf, dx, simPlat_dt, ilast, jlast);
						RHS_DYf = -FVM_i(FluxpiDYf, FluxniDYf, dx, simPlat_dt, ilast, jlast);
						RHS_DYf = RHS_DYf - FVM_j(FluxpjDYf, FluxnjDYf, dx, simPlat_dt, ilast, jlast);
						RHS_DYf = RHS_DYf - FVM_k(FluxpkDYf, FluxnkDYf, dx, simPlat_dt, ilast, jlast);
						RHS_tauf = RHS_tauf - FVM_i(Fluxpitauf, Fluxnitauf, dx, simPlat_dt, ilast, jlast);
						RHS_tauf = RHS_tauf - FVM_j(Fluxpjtauf, Fluxnjtauf, dx, simPlat_dt, ilast, jlast);
						RHS_tauf = RHS_tauf - FVM_k(Fluxpktauf, Fluxnktauf, dx, simPlat_dt, ilast, jlast);
						RHS_phif = RHS_phif - FVM_i(Fluxpiphif, Fluxniphif, dx, simPlat_dt, ilast, jlast);
						RHS_phif = RHS_phif - FVM_j(Fluxpjphif, Fluxnjphif, dx, simPlat_dt, ilast, jlast);
						RHS_phif = RHS_phif - FVM_k(Fluxpkphif, Fluxnkphif, dx, simPlat_dt, ilast, jlast);
						m_theta_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_theta_o2_t2_m0_l0;
						m_theta_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_theta_o2_t1_m0_l0;
						m_theta_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_theta_o2_t0_m0_l0;
						m_theta_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_theta_o1_t2_m0_l0;
						m_theta_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_theta_o1_t1_m0_l0;
						m_theta_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_theta_o1_t0_m0_l0;
						m_theta_o0_t2_l0 = (-rk1Zu_z) * d_Psi1_xx_o0_t3_m0_l0;
						m_theta_o0_t1_l0 = (-rk1Zu_y) * d_Psi1_xx_o0_t2_m0_l0;
						m_theta_o0_t0_l0 = (-rk1Zu_x) * d_Psi1_xx_o0_t1_m0_l0;
						m_trK_o2_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_trK_o2_t2_m0_l0;
						m_trK_o2_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_trK_o2_t1_m0_l0;
						m_trK_o2_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_trK_o2_t0_m0_l0;
						m_trK_o1_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_trK_o1_t2_m0_l0;
						m_trK_o1_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_trK_o1_t1_m0_l0;
						m_trK_o1_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_trK_o1_t0_m0_l0;
						m_trK_o0_t20_l0 = 2.0 * rk1Zu_z * d_Psi1_xx_o0_t3_m0_l0;
						m_trK_o0_t19_l0 = 2.0 * rk1Zu_y * d_Psi1_xx_o0_t2_m0_l0;
						m_trK_o0_t18_l0 = 2.0 * rk1Zu_x * d_Psi1_xx_o0_t1_m0_l0;
						m_trK_o0_t17_l0 = (-2.0 * rk1chi_max * rk1gtu_xz) * d_Psi1_xz_o0_t0_m0_l0;
						m_trK_o0_t16_l0 = (-2.0 * rk1chi_max * rk1gtu_xy) * d_Psi1_xy_o0_t0_m0_l0;
						m_trK_o0_t15_l0 = 0.5 * rk1gtu_zz * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_trK_o0_t14_l0 = 0.5 * rk1gtu_yz * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_trK_o0_t13_l0 = 0.5 * rk1gtu_xz * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_trK_o0_t12_l0 = 0.5 * rk1gtu_yz * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_trK_o0_t11_l0 = 0.5 * rk1gtu_yy * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_trK_o0_t10_l0 = 0.5 * rk1gtu_xy * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_trK_o0_t9_l0 = 0.5 * rk1gtu_xz * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_trK_o0_t8_l0 = 0.5 * rk1gtu_xy * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_trK_o0_t7_l0 = 0.5 * rk1gtu_xx * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_trK_o0_t6_l0 = rk1chi_max * rk1Gamt_z * d_Psi1_xx_o0_t3_m0_l0;
						m_trK_o0_t5_l0 = rk1chi_max * rk1Gamt_y * d_Psi1_xx_o0_t2_m0_l0;
						m_trK_o0_t4_l0 = rk1chi_max * rk1Gamt_x * d_Psi1_xx_o0_t1_m0_l0;
						m_trK_o0_t3_l0 = (-rk1chi_max * rk1gtu_zz) * d_Psi1_zz_o0_t0_m0_l0;
						m_trK_o0_t2_l0 = (-2.0 * rk1chi_max * rk1gtu_yz) * d_Psi1_yz_o0_t0_m0_l0;
						m_trK_o0_t1_l0 = (-rk1chi_max * rk1gtu_yy) * d_Psi1_yy_o0_t0_m0_l0;
						m_trK_o0_t0_l0 = (-rk1chi_max * rk1gtu_xx) * d_Psi1_xx_o0_t0_m0_l0;
						m_chi_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_chi_o2_t2_m0_l0;
						m_chi_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_chi_o2_t1_m0_l0;
						m_chi_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_chi_o2_t0_m0_l0;
						m_chi_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_chi_o1_t2_m0_l0;
						m_chi_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_chi_o1_t1_m0_l0;
						m_chi_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_chi_o1_t0_m0_l0;
						m_Alpha_o2_t2_l0 = lambda_1 * MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Alpha_o2_t2_m0_l0;
						m_Alpha_o2_t1_l0 = lambda_1 * MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Alpha_o2_t1_m0_l0;
						m_Alpha_o2_t0_l0 = lambda_1 * MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Alpha_o2_t0_m0_l0;
						m_Alpha_o1_t2_l0 = lambda_1 * MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Alpha_o1_t2_m0_l0;
						m_Alpha_o1_t1_l0 = lambda_1 * MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Alpha_o1_t1_m0_l0;
						m_Alpha_o1_t0_l0 = lambda_1 * MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Alpha_o1_t0_m0_l0;
						m_Betau_z_o2_t2_l0 = lambda_2 * MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Betau_z_o2_t2_m0_l0;
						m_Betau_z_o2_t1_l0 = lambda_2 * MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Betau_z_o2_t1_m0_l0;
						m_Betau_z_o2_t0_l0 = lambda_2 * MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Betau_z_o2_t0_m0_l0;
						m_Betau_z_o1_t2_l0 = lambda_2 * MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Betau_z_o1_t2_m0_l0;
						m_Betau_z_o1_t1_l0 = lambda_2 * MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Betau_z_o1_t1_m0_l0;
						m_Betau_z_o1_t0_l0 = lambda_2 * MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Betau_z_o1_t0_m0_l0;
						m_Betau_y_o2_t2_l0 = lambda_2 * MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Betau_y_o2_t2_m0_l0;
						m_Betau_y_o2_t1_l0 = lambda_2 * MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Betau_y_o2_t1_m0_l0;
						m_Betau_y_o2_t0_l0 = lambda_2 * MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Betau_y_o2_t0_m0_l0;
						m_Betau_y_o1_t2_l0 = lambda_2 * MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Betau_y_o1_t2_m0_l0;
						m_Betau_y_o1_t1_l0 = lambda_2 * MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Betau_y_o1_t1_m0_l0;
						m_Betau_y_o1_t0_l0 = lambda_2 * MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Betau_y_o1_t0_m0_l0;
						m_Betau_x_o2_t2_l0 = lambda_2 * MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Betau_x_o2_t2_m0_l0;
						m_Betau_x_o2_t1_l0 = lambda_2 * MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Betau_x_o2_t1_m0_l0;
						m_Betau_x_o2_t0_l0 = lambda_2 * MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Betau_x_o2_t0_m0_l0;
						m_Betau_x_o1_t2_l0 = lambda_2 * MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Betau_x_o1_t2_m0_l0;
						m_Betau_x_o1_t1_l0 = lambda_2 * MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Betau_x_o1_t1_m0_l0;
						m_Betau_x_o1_t0_l0 = lambda_2 * MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Betau_x_o1_t0_m0_l0;
						m_Gamh_z_o2_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Gamh_z_o2_t2_m0_l0;
						m_Gamh_z_o2_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Gamh_z_o2_t1_m0_l0;
						m_Gamh_z_o2_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Gamh_z_o2_t0_m0_l0;
						m_Gamh_z_o1_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Gamh_z_o1_t2_m0_l0;
						m_Gamh_z_o1_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Gamh_z_o1_t1_m0_l0;
						m_Gamh_z_o1_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Gamh_z_o1_t0_m0_l0;
						m_Gamh_z_o0_t23_l0 = (-3.0 * vector(rk1Alpha, i, j, k) * rk1inv_chi * rk1Atu_zz) * d_Rpd_xx_o0_t2_m0_l0;
						m_Gamh_z_o0_t22_l0 = (-3.0 * vector(rk1Alpha, i, j, k) * rk1inv_chi * rk1Atu_yz) * d_Rpd_xx_o0_t1_m0_l0;
						m_Gamh_z_o0_t21_l0 = (-3.0 * vector(rk1Alpha, i, j, k) * rk1inv_chi * rk1Atu_xz) * d_Rpd_xx_o0_t0_m0_l0;
						m_Gamh_z_o0_t20_l0 = (-2.0 * vector(rk1theta, i, j, k) * rk1gtu_zz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_z_o0_t19_l0 = (-2.0 * vector(rk1theta, i, j, k) * rk1gtu_yz) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_z_o0_t18_l0 = (-2.0 * vector(rk1theta, i, j, k) * rk1gtu_xz) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_z_o0_t17_l0 = (-0.6666666666666667 * vector(rk1Alpha, i, j, k) * rk1gtu_zz) * d_Gamh_x_o0_t5_m0_l0;
						m_Gamh_z_o0_t16_l0 = (-0.6666666666666667 * vector(rk1Alpha, i, j, k) * rk1gtu_yz) * d_Gamh_x_o0_t4_m0_l0;
						m_Gamh_z_o0_t15_l0 = (-0.6666666666666667 * vector(rk1Alpha, i, j, k) * rk1gtu_xz) * d_Gamh_x_o0_t3_m0_l0;
						m_Gamh_z_o0_t14_l0 = (-1.333333333333333 * vector(rk1Alpha, i, j, k) * rk1gtu_zz) * d_Gamh_x_o0_t2_m0_l0;
						m_Gamh_z_o0_t13_l0 = (-1.333333333333333 * vector(rk1Alpha, i, j, k) * rk1gtu_yz) * d_Gamh_x_o0_t1_m0_l0;
						m_Gamh_z_o0_t12_l0 = (-1.333333333333333 * vector(rk1Alpha, i, j, k) * rk1gtu_xz) * d_Gamh_x_o0_t0_m0_l0;
						m_Gamh_z_o0_t11_l0 = (-vector(rk1Gamh_z, i, j, k)) * d_div_Beta_o0_t2_m0_l0;
						m_Gamh_z_o0_t10_l0 = (-vector(rk1Gamh_y, i, j, k)) * d_Sfd_y_o0_t9_m0_l0;
						m_Gamh_z_o0_t9_l0 = (-vector(rk1Gamh_x, i, j, k)) * d_Sfd_x_o0_t9_m0_l0;
						m_Gamh_z_o0_t8_l0 = (-2.0 * rk1Atu_zz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_z_o0_t7_l0 = (-2.0 * rk1Atu_yz) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_z_o0_t6_l0 = (-2.0 * rk1Atu_xz) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_z_o0_t5_l0 = rk1gtu_zz * d_d_div_Beta_z_o0_t2_m0_l0;
						m_Gamh_z_o0_t4_l0 = 2.0 * rk1gtu_yz * d_d_div_Beta_y_o0_t2_m0_l0;
						m_Gamh_z_o0_t3_l0 = rk1gtu_yy * d_Gamh_z_o0_t3_m0_l0;
						m_Gamh_z_o0_t2_l0 = 2.0 * rk1gtu_xz * d_d_div_Beta_x_o0_t2_m0_l0;
						m_Gamh_z_o0_t1_l0 = 2.0 * rk1gtu_xy * d_Gamh_z_o0_t1_m0_l0;
						m_Gamh_z_o0_t0_l0 = rk1gtu_xx * d_Gamh_z_o0_t0_m0_l0;
						m_Gamh_y_o2_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Gamh_y_o2_t2_m0_l0;
						m_Gamh_y_o2_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Gamh_y_o2_t1_m0_l0;
						m_Gamh_y_o2_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Gamh_y_o2_t0_m0_l0;
						m_Gamh_y_o1_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Gamh_y_o1_t2_m0_l0;
						m_Gamh_y_o1_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Gamh_y_o1_t1_m0_l0;
						m_Gamh_y_o1_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Gamh_y_o1_t0_m0_l0;
						m_Gamh_y_o0_t23_l0 = (-3.0 * vector(rk1Alpha, i, j, k) * rk1inv_chi * rk1Atu_yz) * d_Rpd_xx_o0_t2_m0_l0;
						m_Gamh_y_o0_t22_l0 = (-3.0 * vector(rk1Alpha, i, j, k) * rk1inv_chi * rk1Atu_yy) * d_Rpd_xx_o0_t1_m0_l0;
						m_Gamh_y_o0_t21_l0 = (-3.0 * vector(rk1Alpha, i, j, k) * rk1inv_chi * rk1Atu_xy) * d_Rpd_xx_o0_t0_m0_l0;
						m_Gamh_y_o0_t20_l0 = (-2.0 * vector(rk1theta, i, j, k) * rk1gtu_yz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_y_o0_t19_l0 = (-2.0 * vector(rk1theta, i, j, k) * rk1gtu_yy) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_y_o0_t18_l0 = (-2.0 * vector(rk1theta, i, j, k) * rk1gtu_xy) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_y_o0_t17_l0 = (-0.6666666666666667 * vector(rk1Alpha, i, j, k) * rk1gtu_yz) * d_Gamh_x_o0_t5_m0_l0;
						m_Gamh_y_o0_t16_l0 = (-0.6666666666666667 * vector(rk1Alpha, i, j, k) * rk1gtu_yy) * d_Gamh_x_o0_t4_m0_l0;
						m_Gamh_y_o0_t15_l0 = (-0.6666666666666667 * vector(rk1Alpha, i, j, k) * rk1gtu_xy) * d_Gamh_x_o0_t3_m0_l0;
						m_Gamh_y_o0_t14_l0 = (-1.333333333333333 * vector(rk1Alpha, i, j, k) * rk1gtu_yz) * d_Gamh_x_o0_t2_m0_l0;
						m_Gamh_y_o0_t13_l0 = (-1.333333333333333 * vector(rk1Alpha, i, j, k) * rk1gtu_yy) * d_Gamh_x_o0_t1_m0_l0;
						m_Gamh_y_o0_t12_l0 = (-1.333333333333333 * vector(rk1Alpha, i, j, k) * rk1gtu_xy) * d_Gamh_x_o0_t0_m0_l0;
						m_Gamh_y_o0_t11_l0 = rk1gtu_zz * d_Gamh_y_o0_t11_m0_l0;
						m_Gamh_y_o0_t10_l0 = 2.0 * rk1gtu_yz * d_d_div_Beta_z_o0_t1_m0_l0;
						m_Gamh_y_o0_t9_l0 = rk1gtu_yy * d_d_div_Beta_y_o0_t1_m0_l0;
						m_Gamh_y_o0_t8_l0 = 2.0 * rk1gtu_xz * d_Gamh_y_o0_t8_m0_l0;
						m_Gamh_y_o0_t7_l0 = 2.0 * rk1gtu_xy * d_d_div_Beta_x_o0_t1_m0_l0;
						m_Gamh_y_o0_t6_l0 = rk1gtu_xx * d_Gamh_y_o0_t6_m0_l0;
						m_Gamh_y_o0_t5_l0 = (-vector(rk1Gamh_z, i, j, k)) * d_Sfd_z_o0_t8_m0_l0;
						m_Gamh_y_o0_t4_l0 = (-vector(rk1Gamh_y, i, j, k)) * d_div_Beta_o0_t1_m0_l0;
						m_Gamh_y_o0_t3_l0 = (-vector(rk1Gamh_x, i, j, k)) * d_Sfd_x_o0_t8_m0_l0;
						m_Gamh_y_o0_t2_l0 = (-2.0 * rk1Atu_yz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_y_o0_t1_l0 = (-2.0 * rk1Atu_yy) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_y_o0_t0_l0 = (-2.0 * rk1Atu_xy) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_x_o2_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Gamh_x_o2_t2_m0_l0;
						m_Gamh_x_o2_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Gamh_x_o2_t1_m0_l0;
						m_Gamh_x_o2_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Gamh_x_o2_t0_m0_l0;
						m_Gamh_x_o1_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Gamh_x_o1_t2_m0_l0;
						m_Gamh_x_o1_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Gamh_x_o1_t1_m0_l0;
						m_Gamh_x_o1_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Gamh_x_o1_t0_m0_l0;
						m_Gamh_x_o0_t23_l0 = (-3.0 * vector(rk1Alpha, i, j, k) * rk1inv_chi * rk1Atu_xz) * d_Rpd_xx_o0_t2_m0_l0;
						m_Gamh_x_o0_t22_l0 = (-3.0 * vector(rk1Alpha, i, j, k) * rk1inv_chi * rk1Atu_xy) * d_Rpd_xx_o0_t1_m0_l0;
						m_Gamh_x_o0_t21_l0 = (-3.0 * vector(rk1Alpha, i, j, k) * rk1inv_chi * rk1Atu_xx) * d_Rpd_xx_o0_t0_m0_l0;
						m_Gamh_x_o0_t20_l0 = (-2.0 * rk1Atu_xz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_x_o0_t19_l0 = (-2.0 * rk1Atu_xy) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_x_o0_t18_l0 = (-2.0 * rk1Atu_xx) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_x_o0_t17_l0 = rk1gtu_zz * d_Gamh_x_o0_t17_m0_l0;
						m_Gamh_x_o0_t16_l0 = 2.0 * rk1gtu_yz * d_Gamh_x_o0_t16_m0_l0;
						m_Gamh_x_o0_t15_l0 = rk1gtu_yy * d_Gamh_x_o0_t15_m0_l0;
						m_Gamh_x_o0_t14_l0 = 2.0 * rk1gtu_xz * d_d_div_Beta_z_o0_t0_m0_l0;
						m_Gamh_x_o0_t13_l0 = 2.0 * rk1gtu_xy * d_d_div_Beta_y_o0_t0_m0_l0;
						m_Gamh_x_o0_t12_l0 = rk1gtu_xx * d_d_div_Beta_x_o0_t0_m0_l0;
						m_Gamh_x_o0_t11_l0 = (-vector(rk1Gamh_z, i, j, k)) * d_Sfd_z_o0_t7_m0_l0;
						m_Gamh_x_o0_t10_l0 = (-vector(rk1Gamh_y, i, j, k)) * d_Sfd_y_o0_t7_m0_l0;
						m_Gamh_x_o0_t9_l0 = (-vector(rk1Gamh_x, i, j, k)) * d_div_Beta_o0_t0_m0_l0;
						m_Gamh_x_o0_t8_l0 = (-2.0 * vector(rk1theta, i, j, k) * rk1gtu_xz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_x_o0_t7_l0 = (-2.0 * vector(rk1theta, i, j, k) * rk1gtu_xy) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_x_o0_t6_l0 = (-2.0 * vector(rk1theta, i, j, k) * rk1gtu_xx) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_x_o0_t5_l0 = (-0.6666666666666667 * vector(rk1Alpha, i, j, k) * rk1gtu_xz) * d_Gamh_x_o0_t5_m0_l0;
						m_Gamh_x_o0_t4_l0 = (-0.6666666666666667 * vector(rk1Alpha, i, j, k) * rk1gtu_xy) * d_Gamh_x_o0_t4_m0_l0;
						m_Gamh_x_o0_t3_l0 = (-0.6666666666666667 * vector(rk1Alpha, i, j, k) * rk1gtu_xx) * d_Gamh_x_o0_t3_m0_l0;
						m_Gamh_x_o0_t2_l0 = (-1.333333333333333 * vector(rk1Alpha, i, j, k) * rk1gtu_xz) * d_Gamh_x_o0_t2_m0_l0;
						m_Gamh_x_o0_t1_l0 = (-1.333333333333333 * vector(rk1Alpha, i, j, k) * rk1gtu_xy) * d_Gamh_x_o0_t1_m0_l0;
						m_Gamh_x_o0_t0_l0 = (-1.333333333333333 * vector(rk1Alpha, i, j, k) * rk1gtu_xx) * d_Gamh_x_o0_t0_m0_l0;
						m_Atd_zz_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Atd_zz_o2_t2_m0_l0;
						m_Atd_zz_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Atd_zz_o2_t1_m0_l0;
						m_Atd_zz_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Atd_zz_o2_t0_m0_l0;
						m_Atd_zz_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Atd_zz_o1_t2_m0_l0;
						m_Atd_zz_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Atd_zz_o1_t1_m0_l0;
						m_Atd_zz_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Atd_zz_o1_t0_m0_l0;
						m_Atd_zz_o0_t2_l0 = 2.0 * vector(rk1Atd_zz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Atd_zz_o0_t1_l0 = 2.0 * vector(rk1Atd_yz, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Atd_zz_o0_t0_l0 = 2.0 * vector(rk1Atd_xz, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Atd_yz_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Atd_yz_o2_t2_m0_l0;
						m_Atd_yz_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Atd_yz_o2_t1_m0_l0;
						m_Atd_yz_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Atd_yz_o2_t0_m0_l0;
						m_Atd_yz_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Atd_yz_o1_t2_m0_l0;
						m_Atd_yz_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Atd_yz_o1_t1_m0_l0;
						m_Atd_yz_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Atd_yz_o1_t0_m0_l0;
						m_Atd_yz_o0_t5_l0 = vector(rk1Atd_zz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Atd_yz_o0_t4_l0 = vector(rk1Atd_yz, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Atd_yz_o0_t3_l0 = vector(rk1Atd_xz, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Atd_yz_o0_t2_l0 = vector(rk1Atd_yz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Atd_yz_o0_t1_l0 = vector(rk1Atd_yy, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Atd_yz_o0_t0_l0 = vector(rk1Atd_xy, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Atd_yy_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Atd_yy_o2_t2_m0_l0;
						m_Atd_yy_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Atd_yy_o2_t1_m0_l0;
						m_Atd_yy_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Atd_yy_o2_t0_m0_l0;
						m_Atd_yy_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Atd_yy_o1_t2_m0_l0;
						m_Atd_yy_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Atd_yy_o1_t1_m0_l0;
						m_Atd_yy_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Atd_yy_o1_t0_m0_l0;
						m_Atd_yy_o0_t2_l0 = 2.0 * vector(rk1Atd_yz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Atd_yy_o0_t1_l0 = 2.0 * vector(rk1Atd_yy, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Atd_yy_o0_t0_l0 = 2.0 * vector(rk1Atd_xy, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Atd_xz_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Atd_xz_o2_t2_m0_l0;
						m_Atd_xz_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Atd_xz_o2_t1_m0_l0;
						m_Atd_xz_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Atd_xz_o2_t0_m0_l0;
						m_Atd_xz_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Atd_xz_o1_t2_m0_l0;
						m_Atd_xz_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Atd_xz_o1_t1_m0_l0;
						m_Atd_xz_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Atd_xz_o1_t0_m0_l0;
						m_Atd_xz_o0_t5_l0 = vector(rk1Atd_zz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Atd_xz_o0_t4_l0 = vector(rk1Atd_yz, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Atd_xz_o0_t3_l0 = vector(rk1Atd_xz, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_Atd_xz_o0_t2_l0 = vector(rk1Atd_xz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Atd_xz_o0_t1_l0 = vector(rk1Atd_xy, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Atd_xz_o0_t0_l0 = vector(rk1Atd_xx, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Atd_xy_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Atd_xy_o2_t2_m0_l0;
						m_Atd_xy_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Atd_xy_o2_t1_m0_l0;
						m_Atd_xy_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Atd_xy_o2_t0_m0_l0;
						m_Atd_xy_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Atd_xy_o1_t2_m0_l0;
						m_Atd_xy_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Atd_xy_o1_t1_m0_l0;
						m_Atd_xy_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Atd_xy_o1_t0_m0_l0;
						m_Atd_xy_o0_t5_l0 = vector(rk1Atd_yz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Atd_xy_o0_t4_l0 = vector(rk1Atd_yy, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Atd_xy_o0_t3_l0 = vector(rk1Atd_xy, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_Atd_xy_o0_t2_l0 = vector(rk1Atd_xz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Atd_xy_o0_t1_l0 = vector(rk1Atd_xy, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Atd_xy_o0_t0_l0 = vector(rk1Atd_xx, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Atd_xx_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_Atd_xx_o2_t2_m0_l0;
						m_Atd_xx_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_Atd_xx_o2_t1_m0_l0;
						m_Atd_xx_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_Atd_xx_o2_t0_m0_l0;
						m_Atd_xx_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_Atd_xx_o1_t2_m0_l0;
						m_Atd_xx_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_Atd_xx_o1_t1_m0_l0;
						m_Atd_xx_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_Atd_xx_o1_t0_m0_l0;
						m_Atd_xx_o0_t2_l0 = 2.0 * vector(rk1Atd_xz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Atd_xx_o0_t1_l0 = 2.0 * vector(rk1Atd_xy, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Atd_xx_o0_t0_l0 = 2.0 * vector(rk1Atd_xx, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_gtd_zz_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_gtd_zz_o2_t2_m0_l0;
						m_gtd_zz_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_gtd_zz_o2_t1_m0_l0;
						m_gtd_zz_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_gtd_zz_o2_t0_m0_l0;
						m_gtd_zz_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_gtd_zz_o1_t2_m0_l0;
						m_gtd_zz_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_gtd_zz_o1_t1_m0_l0;
						m_gtd_zz_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_gtd_zz_o1_t0_m0_l0;
						m_gtd_zz_o0_t2_l0 = 2.0 * vector(rk1gtd_zz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_gtd_zz_o0_t1_l0 = 2.0 * vector(rk1gtd_yz, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_gtd_zz_o0_t0_l0 = 2.0 * vector(rk1gtd_xz, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_gtd_yz_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_gtd_yz_o2_t2_m0_l0;
						m_gtd_yz_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_gtd_yz_o2_t1_m0_l0;
						m_gtd_yz_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_gtd_yz_o2_t0_m0_l0;
						m_gtd_yz_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_gtd_yz_o1_t2_m0_l0;
						m_gtd_yz_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_gtd_yz_o1_t1_m0_l0;
						m_gtd_yz_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_gtd_yz_o1_t0_m0_l0;
						m_gtd_yz_o0_t5_l0 = vector(rk1gtd_zz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_gtd_yz_o0_t4_l0 = vector(rk1gtd_yz, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_gtd_yz_o0_t3_l0 = vector(rk1gtd_xz, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_gtd_yz_o0_t2_l0 = vector(rk1gtd_yz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_gtd_yz_o0_t1_l0 = vector(rk1gtd_yy, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_gtd_yz_o0_t0_l0 = vector(rk1gtd_xy, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_gtd_yy_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_gtd_yy_o2_t2_m0_l0;
						m_gtd_yy_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_gtd_yy_o2_t1_m0_l0;
						m_gtd_yy_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_gtd_yy_o2_t0_m0_l0;
						m_gtd_yy_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_gtd_yy_o1_t2_m0_l0;
						m_gtd_yy_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_gtd_yy_o1_t1_m0_l0;
						m_gtd_yy_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_gtd_yy_o1_t0_m0_l0;
						m_gtd_yy_o0_t2_l0 = 2.0 * vector(rk1gtd_yz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_gtd_yy_o0_t1_l0 = 2.0 * vector(rk1gtd_yy, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_gtd_yy_o0_t0_l0 = 2.0 * vector(rk1gtd_xy, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_gtd_xz_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_gtd_xz_o2_t2_m0_l0;
						m_gtd_xz_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_gtd_xz_o2_t1_m0_l0;
						m_gtd_xz_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_gtd_xz_o2_t0_m0_l0;
						m_gtd_xz_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_gtd_xz_o1_t2_m0_l0;
						m_gtd_xz_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_gtd_xz_o1_t1_m0_l0;
						m_gtd_xz_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_gtd_xz_o1_t0_m0_l0;
						m_gtd_xz_o0_t5_l0 = vector(rk1gtd_zz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_gtd_xz_o0_t4_l0 = vector(rk1gtd_yz, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_gtd_xz_o0_t3_l0 = vector(rk1gtd_xz, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_gtd_xz_o0_t2_l0 = vector(rk1gtd_xz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_gtd_xz_o0_t1_l0 = vector(rk1gtd_xy, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_gtd_xz_o0_t0_l0 = vector(rk1gtd_xx, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_gtd_xy_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_gtd_xy_o2_t2_m0_l0;
						m_gtd_xy_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_gtd_xy_o2_t1_m0_l0;
						m_gtd_xy_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_gtd_xy_o2_t0_m0_l0;
						m_gtd_xy_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_gtd_xy_o1_t2_m0_l0;
						m_gtd_xy_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_gtd_xy_o1_t1_m0_l0;
						m_gtd_xy_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_gtd_xy_o1_t0_m0_l0;
						m_gtd_xy_o0_t5_l0 = vector(rk1gtd_yz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_gtd_xy_o0_t4_l0 = vector(rk1gtd_yy, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_gtd_xy_o0_t3_l0 = vector(rk1gtd_xy, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_gtd_xy_o0_t2_l0 = vector(rk1gtd_xz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_gtd_xy_o0_t1_l0 = vector(rk1gtd_xy, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_gtd_xy_o0_t0_l0 = vector(rk1gtd_xx, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_gtd_xx_o2_t2_l0 = MIN(0.0, vector(rk1Betau_z, i, j, k)) * d_gtd_xx_o2_t2_m0_l0;
						m_gtd_xx_o2_t1_l0 = MIN(0.0, vector(rk1Betau_y, i, j, k)) * d_gtd_xx_o2_t1_m0_l0;
						m_gtd_xx_o2_t0_l0 = MIN(0.0, vector(rk1Betau_x, i, j, k)) * d_gtd_xx_o2_t0_m0_l0;
						m_gtd_xx_o1_t2_l0 = MAX(0.0, vector(rk1Betau_z, i, j, k)) * d_gtd_xx_o1_t2_m0_l0;
						m_gtd_xx_o1_t1_l0 = MAX(0.0, vector(rk1Betau_y, i, j, k)) * d_gtd_xx_o1_t1_m0_l0;
						m_gtd_xx_o1_t0_l0 = MAX(0.0, vector(rk1Betau_x, i, j, k)) * d_gtd_xx_o1_t0_m0_l0;
						m_gtd_xx_o0_t2_l0 = 2.0 * vector(rk1gtd_xz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_gtd_xx_o0_t1_l0 = 2.0 * vector(rk1gtd_xy, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_gtd_xx_o0_t0_l0 = 2.0 * vector(rk1gtd_xx, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_phif_o0_t2_l0 = (ch * ch) * vector(rk1Bfu_z, i, j, k) * d_Psi1_xx_o0_t3_m0_l0;
						m_phif_o0_t1_l0 = (ch * ch) * vector(rk1Bfu_y, i, j, k) * d_Psi1_xx_o0_t2_m0_l0;
						m_phif_o0_t0_l0 = (ch * ch) * vector(rk1Bfu_x, i, j, k) * d_Psi1_xx_o0_t1_m0_l0;
						m_tauf_o0_t2_l0 = (-rk1Sfu_z) * d_Psi1_xx_o0_t3_m0_l0;
						m_tauf_o0_t1_l0 = (-rk1Sfu_y) * d_Psi1_xx_o0_t2_m0_l0;
						m_tauf_o0_t0_l0 = (-rk1Sfu_x) * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_z_o0_t5_l0 = vector(rk1phif, i, j, k) * rk1gtu_zz * rk1chi_max * d_Psi1_xx_o0_t3_m0_l0;
						m_Bfu_z_o0_t4_l0 = vector(rk1phif, i, j, k) * rk1gtu_yz * rk1chi_max * d_Psi1_xx_o0_t2_m0_l0;
						m_Bfu_z_o0_t3_l0 = vector(rk1phif, i, j, k) * rk1gtu_xz * rk1chi_max * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_z_o0_t2_l0 = (-0.5 * vector(rk1phif, i, j, k) * rk1gtu_zz * vector(rk1Alpha, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Bfu_z_o0_t1_l0 = (-0.5 * vector(rk1phif, i, j, k) * rk1gtu_yz * vector(rk1Alpha, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Bfu_z_o0_t0_l0 = (-0.5 * vector(rk1phif, i, j, k) * rk1gtu_xz * vector(rk1Alpha, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Bfu_y_o0_t5_l0 = vector(rk1phif, i, j, k) * rk1gtu_yz * rk1chi_max * d_Psi1_xx_o0_t3_m0_l0;
						m_Bfu_y_o0_t4_l0 = vector(rk1phif, i, j, k) * rk1gtu_yy * rk1chi_max * d_Psi1_xx_o0_t2_m0_l0;
						m_Bfu_y_o0_t3_l0 = vector(rk1phif, i, j, k) * rk1gtu_xy * rk1chi_max * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_y_o0_t2_l0 = (-0.5 * vector(rk1phif, i, j, k) * rk1gtu_yz * vector(rk1Alpha, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Bfu_y_o0_t1_l0 = (-0.5 * vector(rk1phif, i, j, k) * rk1gtu_yy * vector(rk1Alpha, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Bfu_y_o0_t0_l0 = (-0.5 * vector(rk1phif, i, j, k) * rk1gtu_xy * vector(rk1Alpha, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Bfu_x_o0_t5_l0 = vector(rk1phif, i, j, k) * rk1gtu_xz * rk1chi_max * d_Psi1_xx_o0_t3_m0_l0;
						m_Bfu_x_o0_t4_l0 = vector(rk1phif, i, j, k) * rk1gtu_xy * rk1chi_max * d_Psi1_xx_o0_t2_m0_l0;
						m_Bfu_x_o0_t3_l0 = vector(rk1phif, i, j, k) * rk1gtu_xx * rk1chi_max * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_x_o0_t2_l0 = (-0.5 * vector(rk1phif, i, j, k) * rk1gtu_xz * vector(rk1Alpha, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Bfu_x_o0_t1_l0 = (-0.5 * vector(rk1phif, i, j, k) * rk1gtu_xy * vector(rk1Alpha, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Bfu_x_o0_t0_l0 = (-0.5 * vector(rk1phif, i, j, k) * rk1gtu_xx * vector(rk1Alpha, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Sfd_z_o0_t10_l0 = (-(vector(rk1Df, i, j, k) + vector(rk1tauf, i, j, k))) * d_Psi1_xx_o0_t3_m0_l0;
						m_Sfd_z_o0_t9_l0 = vector(rk1Sfd_z, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Sfd_z_o0_t8_l0 = vector(rk1Sfd_y, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Sfd_z_o0_t7_l0 = vector(rk1Sfd_x, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Sfd_z_o0_t6_l0 = (-0.5 * (vector(rk1Alpha, i, j, k) * rk1trSf) / rk1chi_max) * d_Rpd_xx_o0_t2_m0_l0;
						m_Sfd_z_o0_t5_l0 = 0.5 * (vector(rk1Alpha, i, j, k) * rk1Sfuu_zz) / rk1chi_max * d_Ctd_zzz_o0_t0_m0_l0;
						m_Sfd_z_o0_t4_l0 = (vector(rk1Alpha, i, j, k) * rk1Sfuu_yz) / rk1chi_max * d_Ctd_yzz_o0_t0_m0_l0;
						m_Sfd_z_o0_t3_l0 = 0.5 * (vector(rk1Alpha, i, j, k) * rk1Sfuu_yy) / rk1chi_max * d_Ctd_yyz_o0_t0_m0_l0;
						m_Sfd_z_o0_t2_l0 = (vector(rk1Alpha, i, j, k) * rk1Sfuu_xz) / rk1chi_max * d_Ctd_xzz_o0_t0_m0_l0;
						m_Sfd_z_o0_t1_l0 = (vector(rk1Alpha, i, j, k) * rk1Sfuu_xy) / rk1chi_max * d_Ctd_xyz_o0_t1_m0_l0;
						m_Sfd_z_o0_t0_l0 = 0.5 * (vector(rk1Alpha, i, j, k) * rk1Sfuu_xx) / rk1chi_max * d_Ctd_xxz_o0_t0_m0_l0;
						m_Sfd_y_o0_t10_l0 = (-(vector(rk1Df, i, j, k) + vector(rk1tauf, i, j, k))) * d_Psi1_xx_o0_t2_m0_l0;
						m_Sfd_y_o0_t9_l0 = vector(rk1Sfd_z, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Sfd_y_o0_t8_l0 = vector(rk1Sfd_y, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Sfd_y_o0_t7_l0 = vector(rk1Sfd_x, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Sfd_y_o0_t6_l0 = (-0.5 * (vector(rk1Alpha, i, j, k) * rk1trSf) / rk1chi_max) * d_Rpd_xx_o0_t1_m0_l0;
						m_Sfd_y_o0_t5_l0 = 0.5 * (vector(rk1Alpha, i, j, k) * rk1Sfuu_zz) / rk1chi_max * d_Ctd_yzz_o0_t1_m0_l0;
						m_Sfd_y_o0_t4_l0 = (vector(rk1Alpha, i, j, k) * rk1Sfuu_yz) / rk1chi_max * d_Ctd_zyy_o0_t0_m0_l0;
						m_Sfd_y_o0_t3_l0 = 0.5 * (vector(rk1Alpha, i, j, k) * rk1Sfuu_yy) / rk1chi_max * d_Ctd_yyy_o0_t0_m0_l0;
						m_Sfd_y_o0_t2_l0 = (vector(rk1Alpha, i, j, k) * rk1Sfuu_xz) / rk1chi_max * d_Ctd_xyz_o0_t0_m0_l0;
						m_Sfd_y_o0_t1_l0 = (vector(rk1Alpha, i, j, k) * rk1Sfuu_xy) / rk1chi_max * d_Ctd_xyy_o0_t0_m0_l0;
						m_Sfd_y_o0_t0_l0 = 0.5 * (vector(rk1Alpha, i, j, k) * rk1Sfuu_xx) / rk1chi_max * d_Ctd_xxy_o0_t0_m0_l0;
						m_Sfd_x_o0_t10_l0 = (-(vector(rk1Df, i, j, k) + vector(rk1tauf, i, j, k))) * d_Psi1_xx_o0_t1_m0_l0;
						m_Sfd_x_o0_t9_l0 = vector(rk1Sfd_z, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Sfd_x_o0_t8_l0 = vector(rk1Sfd_y, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Sfd_x_o0_t7_l0 = vector(rk1Sfd_x, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_Sfd_x_o0_t6_l0 = (-0.5 * (vector(rk1Alpha, i, j, k) * rk1trSf) / rk1chi_max) * d_Rpd_xx_o0_t0_m0_l0;
						m_Sfd_x_o0_t5_l0 = 0.5 * (vector(rk1Alpha, i, j, k) * rk1Sfuu_zz) / rk1chi_max * d_Ctd_xzz_o0_t1_m0_l0;
						m_Sfd_x_o0_t4_l0 = (vector(rk1Alpha, i, j, k) * rk1Sfuu_yz) / rk1chi_max * d_Ctd_xyz_o0_t2_m0_l0;
						m_Sfd_x_o0_t3_l0 = 0.5 * (vector(rk1Alpha, i, j, k) * rk1Sfuu_yy) / rk1chi_max * d_Ctd_xyy_o0_t1_m0_l0;
						m_Sfd_x_o0_t2_l0 = (vector(rk1Alpha, i, j, k) * rk1Sfuu_xz) / rk1chi_max * d_Ctd_zxx_o0_t0_m0_l0;
						m_Sfd_x_o0_t1_l0 = (vector(rk1Alpha, i, j, k) * rk1Sfuu_xy) / rk1chi_max * d_Ctd_yxx_o0_t0_m0_l0;
						m_Sfd_x_o0_t0_l0 = 0.5 * (vector(rk1Alpha, i, j, k) * rk1Sfuu_xx) / rk1chi_max * d_Ctd_xxx_o0_t0_m0_l0;
						RHS_Sfd_x = RHS_Sfd_x + (((((((((((m_Sfd_x_o0_t0_l0 + m_Sfd_x_o0_t1_l0) + m_Sfd_x_o0_t2_l0) + m_Sfd_x_o0_t3_l0) + m_Sfd_x_o0_t4_l0) + m_Sfd_x_o0_t5_l0) + m_Sfd_x_o0_t6_l0) + m_Sfd_x_o0_t7_l0) + m_Sfd_x_o0_t8_l0) + m_Sfd_x_o0_t9_l0) + m_Sfd_x_o0_t10_l0) + d_Sfd_x_o0_t11_m0_l0);
						RHS_Sfd_y = RHS_Sfd_y + (((((((((((m_Sfd_y_o0_t0_l0 + m_Sfd_y_o0_t1_l0) + m_Sfd_y_o0_t2_l0) + m_Sfd_y_o0_t3_l0) + m_Sfd_y_o0_t4_l0) + m_Sfd_y_o0_t5_l0) + m_Sfd_y_o0_t6_l0) + m_Sfd_y_o0_t7_l0) + m_Sfd_y_o0_t8_l0) + m_Sfd_y_o0_t9_l0) + m_Sfd_y_o0_t10_l0) + d_Sfd_y_o0_t11_m0_l0);
						RHS_Sfd_z = RHS_Sfd_z + (((((((((((m_Sfd_z_o0_t0_l0 + m_Sfd_z_o0_t1_l0) + m_Sfd_z_o0_t2_l0) + m_Sfd_z_o0_t3_l0) + m_Sfd_z_o0_t4_l0) + m_Sfd_z_o0_t5_l0) + m_Sfd_z_o0_t6_l0) + m_Sfd_z_o0_t7_l0) + m_Sfd_z_o0_t8_l0) + m_Sfd_z_o0_t9_l0) + m_Sfd_z_o0_t10_l0) + d_Sfd_z_o0_t11_m0_l0);
						RHS_Bfu_x = RHS_Bfu_x + (((((m_Bfu_x_o0_t0_l0 + m_Bfu_x_o0_t1_l0) + m_Bfu_x_o0_t2_l0) + m_Bfu_x_o0_t3_l0) + m_Bfu_x_o0_t4_l0) + m_Bfu_x_o0_t5_l0);
						RHS_Bfu_y = RHS_Bfu_y + (((((m_Bfu_y_o0_t0_l0 + m_Bfu_y_o0_t1_l0) + m_Bfu_y_o0_t2_l0) + m_Bfu_y_o0_t3_l0) + m_Bfu_y_o0_t4_l0) + m_Bfu_y_o0_t5_l0);
						RHS_Bfu_z = RHS_Bfu_z + (((((m_Bfu_z_o0_t0_l0 + m_Bfu_z_o0_t1_l0) + m_Bfu_z_o0_t2_l0) + m_Bfu_z_o0_t3_l0) + m_Bfu_z_o0_t4_l0) + m_Bfu_z_o0_t5_l0);
						RHS_DYf = RHS_DYf + d_DYf_o0_t0_m0_l0;
						RHS_tauf = RHS_tauf + (((m_tauf_o0_t0_l0 + m_tauf_o0_t1_l0) + m_tauf_o0_t2_l0) + d_tauf_o0_t3_m0_l0);
						RHS_phif = RHS_phif + ((m_phif_o0_t0_l0 + m_phif_o0_t1_l0) + m_phif_o0_t2_l0);
						RHS_gtd_xx = ((((((((m_gtd_xx_o0_t0_l0 + m_gtd_xx_o0_t1_l0) + m_gtd_xx_o0_t2_l0) + d_gtd_xx_o0_t3_m0_l0) + m_gtd_xx_o1_t0_l0) + m_gtd_xx_o1_t1_l0) + m_gtd_xx_o1_t2_l0) + m_gtd_xx_o2_t0_l0) + m_gtd_xx_o2_t1_l0) + m_gtd_xx_o2_t2_l0;
						RHS_gtd_xy = (((((((((((m_gtd_xy_o0_t0_l0 + m_gtd_xy_o0_t1_l0) + m_gtd_xy_o0_t2_l0) + m_gtd_xy_o0_t3_l0) + m_gtd_xy_o0_t4_l0) + m_gtd_xy_o0_t5_l0) + d_gtd_xy_o0_t6_m0_l0) + m_gtd_xy_o1_t0_l0) + m_gtd_xy_o1_t1_l0) + m_gtd_xy_o1_t2_l0) + m_gtd_xy_o2_t0_l0) + m_gtd_xy_o2_t1_l0) + m_gtd_xy_o2_t2_l0;
						RHS_gtd_xz = (((((((((((m_gtd_xz_o0_t0_l0 + m_gtd_xz_o0_t1_l0) + m_gtd_xz_o0_t2_l0) + m_gtd_xz_o0_t3_l0) + m_gtd_xz_o0_t4_l0) + m_gtd_xz_o0_t5_l0) + d_gtd_xz_o0_t6_m0_l0) + m_gtd_xz_o1_t0_l0) + m_gtd_xz_o1_t1_l0) + m_gtd_xz_o1_t2_l0) + m_gtd_xz_o2_t0_l0) + m_gtd_xz_o2_t1_l0) + m_gtd_xz_o2_t2_l0;
						RHS_gtd_yy = ((((((((m_gtd_yy_o0_t0_l0 + m_gtd_yy_o0_t1_l0) + m_gtd_yy_o0_t2_l0) + d_gtd_yy_o0_t3_m0_l0) + m_gtd_yy_o1_t0_l0) + m_gtd_yy_o1_t1_l0) + m_gtd_yy_o1_t2_l0) + m_gtd_yy_o2_t0_l0) + m_gtd_yy_o2_t1_l0) + m_gtd_yy_o2_t2_l0;
						RHS_gtd_yz = (((((((((((m_gtd_yz_o0_t0_l0 + m_gtd_yz_o0_t1_l0) + m_gtd_yz_o0_t2_l0) + m_gtd_yz_o0_t3_l0) + m_gtd_yz_o0_t4_l0) + m_gtd_yz_o0_t5_l0) + d_gtd_yz_o0_t6_m0_l0) + m_gtd_yz_o1_t0_l0) + m_gtd_yz_o1_t1_l0) + m_gtd_yz_o1_t2_l0) + m_gtd_yz_o2_t0_l0) + m_gtd_yz_o2_t1_l0) + m_gtd_yz_o2_t2_l0;
						RHS_gtd_zz = ((((((((m_gtd_zz_o0_t0_l0 + m_gtd_zz_o0_t1_l0) + m_gtd_zz_o0_t2_l0) + d_gtd_zz_o0_t3_m0_l0) + m_gtd_zz_o1_t0_l0) + m_gtd_zz_o1_t1_l0) + m_gtd_zz_o1_t2_l0) + m_gtd_zz_o2_t0_l0) + m_gtd_zz_o2_t1_l0) + m_gtd_zz_o2_t2_l0;
						RHS_Atd_xx = ((((((((m_Atd_xx_o0_t0_l0 + m_Atd_xx_o0_t1_l0) + m_Atd_xx_o0_t2_l0) + d_Atd_xx_o0_t3_m0_l0) + m_Atd_xx_o1_t0_l0) + m_Atd_xx_o1_t1_l0) + m_Atd_xx_o1_t2_l0) + m_Atd_xx_o2_t0_l0) + m_Atd_xx_o2_t1_l0) + m_Atd_xx_o2_t2_l0;
						RHS_Atd_xy = (((((((((((m_Atd_xy_o0_t0_l0 + m_Atd_xy_o0_t1_l0) + m_Atd_xy_o0_t2_l0) + m_Atd_xy_o0_t3_l0) + m_Atd_xy_o0_t4_l0) + m_Atd_xy_o0_t5_l0) + d_Atd_xy_o0_t6_m0_l0) + m_Atd_xy_o1_t0_l0) + m_Atd_xy_o1_t1_l0) + m_Atd_xy_o1_t2_l0) + m_Atd_xy_o2_t0_l0) + m_Atd_xy_o2_t1_l0) + m_Atd_xy_o2_t2_l0;
						RHS_Atd_xz = (((((((((((m_Atd_xz_o0_t0_l0 + m_Atd_xz_o0_t1_l0) + m_Atd_xz_o0_t2_l0) + m_Atd_xz_o0_t3_l0) + m_Atd_xz_o0_t4_l0) + m_Atd_xz_o0_t5_l0) + d_Atd_xz_o0_t6_m0_l0) + m_Atd_xz_o1_t0_l0) + m_Atd_xz_o1_t1_l0) + m_Atd_xz_o1_t2_l0) + m_Atd_xz_o2_t0_l0) + m_Atd_xz_o2_t1_l0) + m_Atd_xz_o2_t2_l0;
						RHS_Atd_yy = ((((((((m_Atd_yy_o0_t0_l0 + m_Atd_yy_o0_t1_l0) + m_Atd_yy_o0_t2_l0) + d_Atd_yy_o0_t3_m0_l0) + m_Atd_yy_o1_t0_l0) + m_Atd_yy_o1_t1_l0) + m_Atd_yy_o1_t2_l0) + m_Atd_yy_o2_t0_l0) + m_Atd_yy_o2_t1_l0) + m_Atd_yy_o2_t2_l0;
						RHS_Atd_yz = (((((((((((m_Atd_yz_o0_t0_l0 + m_Atd_yz_o0_t1_l0) + m_Atd_yz_o0_t2_l0) + m_Atd_yz_o0_t3_l0) + m_Atd_yz_o0_t4_l0) + m_Atd_yz_o0_t5_l0) + d_Atd_yz_o0_t6_m0_l0) + m_Atd_yz_o1_t0_l0) + m_Atd_yz_o1_t1_l0) + m_Atd_yz_o1_t2_l0) + m_Atd_yz_o2_t0_l0) + m_Atd_yz_o2_t1_l0) + m_Atd_yz_o2_t2_l0;
						RHS_Atd_zz = ((((((((m_Atd_zz_o0_t0_l0 + m_Atd_zz_o0_t1_l0) + m_Atd_zz_o0_t2_l0) + d_Atd_zz_o0_t3_m0_l0) + m_Atd_zz_o1_t0_l0) + m_Atd_zz_o1_t1_l0) + m_Atd_zz_o1_t2_l0) + m_Atd_zz_o2_t0_l0) + m_Atd_zz_o2_t1_l0) + m_Atd_zz_o2_t2_l0;
						RHS_Gamh_x = (((((((((((((((((((((((((((((m_Gamh_x_o0_t0_l0 + m_Gamh_x_o0_t1_l0) + m_Gamh_x_o0_t2_l0) + m_Gamh_x_o0_t3_l0) + m_Gamh_x_o0_t4_l0) + m_Gamh_x_o0_t5_l0) + m_Gamh_x_o0_t6_l0) + m_Gamh_x_o0_t7_l0) + m_Gamh_x_o0_t8_l0) + m_Gamh_x_o0_t9_l0) + m_Gamh_x_o0_t10_l0) + m_Gamh_x_o0_t11_l0) + m_Gamh_x_o0_t12_l0) + m_Gamh_x_o0_t13_l0) + m_Gamh_x_o0_t14_l0) + m_Gamh_x_o0_t15_l0) + m_Gamh_x_o0_t16_l0) + m_Gamh_x_o0_t17_l0) + m_Gamh_x_o0_t18_l0) + m_Gamh_x_o0_t19_l0) + m_Gamh_x_o0_t20_l0) + m_Gamh_x_o0_t21_l0) + m_Gamh_x_o0_t22_l0) + m_Gamh_x_o0_t23_l0) + d_Gamh_x_o0_t24_m0_l0) + m_Gamh_x_o1_t0_l0) + m_Gamh_x_o1_t1_l0) + m_Gamh_x_o1_t2_l0) + m_Gamh_x_o2_t0_l0) + m_Gamh_x_o2_t1_l0) + m_Gamh_x_o2_t2_l0;
						RHS_Gamh_y = (((((((((((((((((((((((((((((m_Gamh_y_o0_t0_l0 + m_Gamh_y_o0_t1_l0) + m_Gamh_y_o0_t2_l0) + m_Gamh_y_o0_t3_l0) + m_Gamh_y_o0_t4_l0) + m_Gamh_y_o0_t5_l0) + m_Gamh_y_o0_t6_l0) + m_Gamh_y_o0_t7_l0) + m_Gamh_y_o0_t8_l0) + m_Gamh_y_o0_t9_l0) + m_Gamh_y_o0_t10_l0) + m_Gamh_y_o0_t11_l0) + m_Gamh_y_o0_t12_l0) + m_Gamh_y_o0_t13_l0) + m_Gamh_y_o0_t14_l0) + m_Gamh_y_o0_t15_l0) + m_Gamh_y_o0_t16_l0) + m_Gamh_y_o0_t17_l0) + m_Gamh_y_o0_t18_l0) + m_Gamh_y_o0_t19_l0) + m_Gamh_y_o0_t20_l0) + m_Gamh_y_o0_t21_l0) + m_Gamh_y_o0_t22_l0) + m_Gamh_y_o0_t23_l0) + d_Gamh_y_o0_t24_m0_l0) + m_Gamh_y_o1_t0_l0) + m_Gamh_y_o1_t1_l0) + m_Gamh_y_o1_t2_l0) + m_Gamh_y_o2_t0_l0) + m_Gamh_y_o2_t1_l0) + m_Gamh_y_o2_t2_l0;
						RHS_Gamh_z = (((((((((((((((((((((((((((((m_Gamh_z_o0_t0_l0 + m_Gamh_z_o0_t1_l0) + m_Gamh_z_o0_t2_l0) + m_Gamh_z_o0_t3_l0) + m_Gamh_z_o0_t4_l0) + m_Gamh_z_o0_t5_l0) + m_Gamh_z_o0_t6_l0) + m_Gamh_z_o0_t7_l0) + m_Gamh_z_o0_t8_l0) + m_Gamh_z_o0_t9_l0) + m_Gamh_z_o0_t10_l0) + m_Gamh_z_o0_t11_l0) + m_Gamh_z_o0_t12_l0) + m_Gamh_z_o0_t13_l0) + m_Gamh_z_o0_t14_l0) + m_Gamh_z_o0_t15_l0) + m_Gamh_z_o0_t16_l0) + m_Gamh_z_o0_t17_l0) + m_Gamh_z_o0_t18_l0) + m_Gamh_z_o0_t19_l0) + m_Gamh_z_o0_t20_l0) + m_Gamh_z_o0_t21_l0) + m_Gamh_z_o0_t22_l0) + m_Gamh_z_o0_t23_l0) + d_Gamh_z_o0_t24_m0_l0) + m_Gamh_z_o1_t0_l0) + m_Gamh_z_o1_t1_l0) + m_Gamh_z_o1_t2_l0) + m_Gamh_z_o2_t0_l0) + m_Gamh_z_o2_t1_l0) + m_Gamh_z_o2_t2_l0;
						RHS_Betau_x = (((((d_Betau_x_o0_t0_m0_l0 + m_Betau_x_o1_t0_l0) + m_Betau_x_o1_t1_l0) + m_Betau_x_o1_t2_l0) + m_Betau_x_o2_t0_l0) + m_Betau_x_o2_t1_l0) + m_Betau_x_o2_t2_l0;
						RHS_Betau_y = (((((d_Betau_y_o0_t0_m0_l0 + m_Betau_y_o1_t0_l0) + m_Betau_y_o1_t1_l0) + m_Betau_y_o1_t2_l0) + m_Betau_y_o2_t0_l0) + m_Betau_y_o2_t1_l0) + m_Betau_y_o2_t2_l0;
						RHS_Betau_z = (((((d_Betau_z_o0_t0_m0_l0 + m_Betau_z_o1_t0_l0) + m_Betau_z_o1_t1_l0) + m_Betau_z_o1_t2_l0) + m_Betau_z_o2_t0_l0) + m_Betau_z_o2_t1_l0) + m_Betau_z_o2_t2_l0;
						RHS_Alpha = (((((d_Alpha_o0_t0_m0_l0 + m_Alpha_o1_t0_l0) + m_Alpha_o1_t1_l0) + m_Alpha_o1_t2_l0) + m_Alpha_o2_t0_l0) + m_Alpha_o2_t1_l0) + m_Alpha_o2_t2_l0;
						RHS_chi = (((((d_chi_o0_t0_m0_l0 + m_chi_o1_t0_l0) + m_chi_o1_t1_l0) + m_chi_o1_t2_l0) + m_chi_o2_t0_l0) + m_chi_o2_t1_l0) + m_chi_o2_t2_l0;
						RHS_trK = ((((((((((((((((((((((((((m_trK_o0_t0_l0 + m_trK_o0_t1_l0) + m_trK_o0_t2_l0) + m_trK_o0_t3_l0) + m_trK_o0_t4_l0) + m_trK_o0_t5_l0) + m_trK_o0_t6_l0) + m_trK_o0_t7_l0) + m_trK_o0_t8_l0) + m_trK_o0_t9_l0) + m_trK_o0_t10_l0) + m_trK_o0_t11_l0) + m_trK_o0_t12_l0) + m_trK_o0_t13_l0) + m_trK_o0_t14_l0) + m_trK_o0_t15_l0) + m_trK_o0_t16_l0) + m_trK_o0_t17_l0) + m_trK_o0_t18_l0) + m_trK_o0_t19_l0) + m_trK_o0_t20_l0) + d_trK_o0_t21_m0_l0) + m_trK_o1_t0_l0) + m_trK_o1_t1_l0) + m_trK_o1_t2_l0) + m_trK_o2_t0_l0) + m_trK_o2_t1_l0) + m_trK_o2_t2_l0;
						RHS_theta = ((((((((m_theta_o0_t0_l0 + m_theta_o0_t1_l0) + m_theta_o0_t2_l0) + d_theta_o0_t3_m0_l0) + m_theta_o1_t0_l0) + m_theta_o1_t1_l0) + m_theta_o1_t2_l0) + m_theta_o2_t0_l0) + m_theta_o2_t1_l0) + m_theta_o2_t2_l0;
						n_x = 0.0;
						n_y = 0.0;
						n_z = 0.0;
						interaction_index = 0.0;
						if ((((((vector(FOV_xLower, i + 1, j, k) > 0.0) || (vector(FOV_xUpper, i + 1, j, k) > 0.0)) || (vector(FOV_yLower, i + 1, j, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k) > 0.0)) || (vector(FOV_zLower, i + 1, j, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x + 1.0 / dx[0];
						}
						if ((((((vector(FOV_xLower, i - 1, j, k) > 0.0) || (vector(FOV_xUpper, i - 1, j, k) > 0.0)) || (vector(FOV_yLower, i - 1, j, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k) > 0.0)) || (vector(FOV_zLower, i - 1, j, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x - 1.0 / dx[0];
						}
						if ((((((vector(FOV_xLower, i, j + 1, k) > 0.0) || (vector(FOV_xUpper, i, j + 1, k) > 0.0)) || (vector(FOV_yLower, i, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k) > 0.0)) || (vector(FOV_zLower, i, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y + 1.0 / dx[1];
						}
						if ((((((vector(FOV_xLower, i, j - 1, k) > 0.0) || (vector(FOV_xUpper, i, j - 1, k) > 0.0)) || (vector(FOV_yLower, i, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k) > 0.0)) || (vector(FOV_zLower, i, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y - 1.0 / dx[1];
						}
						if ((((((vector(FOV_xLower, i, j, k + 1) > 0.0) || (vector(FOV_xUpper, i, j, k + 1) > 0.0)) || (vector(FOV_yLower, i, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j, k + 1) > 0.0)) || (vector(FOV_zLower, i, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j, k + 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z + 1.0 / dx[2];
						}
						if ((((((vector(FOV_xLower, i, j, k - 1) > 0.0) || (vector(FOV_xUpper, i, j, k - 1) > 0.0)) || (vector(FOV_yLower, i, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j, k - 1) > 0.0)) || (vector(FOV_zLower, i, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j, k - 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z - 1.0 / dx[2];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k) > 0.0)) || (vector(FOV_yLower, i + 1, j, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k) > 0.0)) || (vector(FOV_zLower, i + 1, j, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x + 1.0 / dx[0];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i - 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k) > 0.0)) || (vector(FOV_yLower, i - 1, j, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k) > 0.0)) || (vector(FOV_zLower, i - 1, j, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x - 1.0 / dx[0];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k) > 0.0)) || (vector(FOV_yLower, i, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k) > 0.0)) || (vector(FOV_zLower, i, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y + 1.0 / dx[1];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j - 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k) > 0.0)) || (vector(FOV_yLower, i, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k) > 0.0)) || (vector(FOV_zLower, i, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y - 1.0 / dx[1];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j, k + 1) > 0.0)) || (vector(FOV_yLower, i, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j, k + 1) > 0.0)) || (vector(FOV_zLower, i, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j, k + 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z + 1.0 / dx[2];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k - 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j, k - 1) > 0.0)) || (vector(FOV_yLower, i, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j, k - 1) > 0.0)) || (vector(FOV_zLower, i, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j, k - 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z - 1.0 / dx[2];
						}
						if ((!equalsEq(n_x, 0.0) || !equalsEq(n_y, 0.0)) || !equalsEq(n_z, 0.0)) {
							mod_normal = sqrt((n_x * n_x + n_y * n_y) + n_z * n_z);
							n_x = n_x / mod_normal;
							n_y = n_y / mod_normal;
							n_z = n_z / mod_normal;
						}
						if (equalsEq(interaction_index, 1.0)) {
							i_d_Atd_xx_o0_t0_m0_l0 = D1CDO4_i(rk1Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xx_o0_t1_m0_l0 = D1CDO4_j(rk1Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xx_o0_t2_m0_l0 = D1CDO4_k(rk1Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xy_o0_t0_m0_l0 = D1CDO4_i(rk1Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xy_o0_t1_m0_l0 = D1CDO4_j(rk1Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xy_o0_t2_m0_l0 = D1CDO4_k(rk1Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xz_o0_t0_m0_l0 = D1CDO4_i(rk1Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xz_o0_t1_m0_l0 = D1CDO4_j(rk1Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xz_o0_t2_m0_l0 = D1CDO4_k(rk1Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yy_o0_t0_m0_l0 = D1CDO4_i(rk1Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yy_o0_t1_m0_l0 = D1CDO4_j(rk1Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yy_o0_t2_m0_l0 = D1CDO4_k(rk1Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yz_o0_t0_m0_l0 = D1CDO4_i(rk1Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yz_o0_t1_m0_l0 = D1CDO4_j(rk1Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yz_o0_t2_m0_l0 = D1CDO4_k(rk1Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_zz_o0_t0_m0_l0 = D1CDO4_i(rk1Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_zz_o0_t1_m0_l0 = D1CDO4_j(rk1Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_zz_o0_t2_m0_l0 = D1CDO4_k(rk1Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_gtd_xx_o0_t3_m0_l0 = -gtd_xx_falloff * (vector(rk1gtd_xx, i, j, k) - gtd_xx_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_xy_o0_t3_m0_l0 = -gtd_xy_falloff * (vector(rk1gtd_xy, i, j, k) - gtd_xy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_xz_o0_t3_m0_l0 = -gtd_xz_falloff * (vector(rk1gtd_xz, i, j, k) - gtd_xz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_yy_o0_t3_m0_l0 = -gtd_yy_falloff * (vector(rk1gtd_yy, i, j, k) - gtd_yy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_yz_o0_t3_m0_l0 = -gtd_yz_falloff * (vector(rk1gtd_yz, i, j, k) - gtd_yz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_zz_o0_t3_m0_l0 = -gtd_zz_falloff * (vector(rk1gtd_zz, i, j, k) - gtd_zz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_xx_o0_t3_m0_l0 = -Atd_xx_falloff * (vector(rk1Atd_xx, i, j, k) - Atd_xx_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_xy_o0_t3_m0_l0 = -Atd_xy_falloff * (vector(rk1Atd_xy, i, j, k) - Atd_xy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_xz_o0_t3_m0_l0 = -Atd_xz_falloff * (vector(rk1Atd_xz, i, j, k) - Atd_xz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_yy_o0_t3_m0_l0 = -Atd_yy_falloff * (vector(rk1Atd_yy, i, j, k) - Atd_yy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_yz_o0_t3_m0_l0 = -Atd_yz_falloff * (vector(rk1Atd_yz, i, j, k) - Atd_yz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_zz_o0_t3_m0_l0 = -Atd_zz_falloff * (vector(rk1Atd_zz, i, j, k) - Atd_zz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Gamh_x_o0_t3_m0_l0 = -Gamh_x_falloff * (vector(rk1Gamh_x, i, j, k) - Gamh_x_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Gamh_y_o0_t3_m0_l0 = -Gamh_y_falloff * (vector(rk1Gamh_y, i, j, k) - Gamh_y_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Gamh_z_o0_t3_m0_l0 = -Gamh_z_falloff * (vector(rk1Gamh_z, i, j, k) - Gamh_z_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Betau_x_o0_t3_m0_l0 = -Betau_x_falloff * (vector(rk1Betau_x, i, j, k) - Betau_x_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Betau_y_o0_t3_m0_l0 = -Betau_y_falloff * (vector(rk1Betau_y, i, j, k) - Betau_y_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Betau_z_o0_t3_m0_l0 = -Betau_z_falloff * (vector(rk1Betau_z, i, j, k) - Betau_z_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Alpha_o0_t3_m0_l0 = -Alpha_falloff * (vector(rk1Alpha, i, j, k) - Alpha_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_chi_o0_t3_m0_l0 = -chi_falloff * (vector(rk1chi, i, j, k) - chi_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_trK_o0_t3_m0_l0 = -trK_falloff * (vector(rk1trK, i, j, k) - trK_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_theta_o0_t3_m0_l0 = -theta_falloff * (vector(rk1theta, i, j, k) - theta_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_m_gtd_xx_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xxx_o0_t0_m0_l0;
							i_m_gtd_xx_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xxy_o0_t0_m0_l0;
							i_m_gtd_xx_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xxz_o0_t0_m0_l0;
							i_m_gtd_xy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yxx_o0_t0_m0_l0;
							i_m_gtd_xy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyy_o0_t0_m0_l0;
							i_m_gtd_xy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyz_o0_t1_m0_l0;
							i_m_gtd_xz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_zxx_o0_t0_m0_l0;
							i_m_gtd_xz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyz_o0_t0_m0_l0;
							i_m_gtd_xz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xzz_o0_t0_m0_l0;
							i_m_gtd_yy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyy_o0_t1_m0_l0;
							i_m_gtd_yy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yyy_o0_t0_m0_l0;
							i_m_gtd_yy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yyz_o0_t0_m0_l0;
							i_m_gtd_yz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyz_o0_t2_m0_l0;
							i_m_gtd_yz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_zyy_o0_t0_m0_l0;
							i_m_gtd_yz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yzz_o0_t0_m0_l0;
							i_m_gtd_zz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xzz_o0_t1_m0_l0;
							i_m_gtd_zz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yzz_o0_t1_m0_l0;
							i_m_gtd_zz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_zzz_o0_t0_m0_l0;
							i_m_Atd_xx_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xx_o0_t0_m0_l0;
							i_m_Atd_xx_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xx_o0_t1_m0_l0;
							i_m_Atd_xx_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xx_o0_t2_m0_l0;
							i_m_Atd_xy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xy_o0_t0_m0_l0;
							i_m_Atd_xy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xy_o0_t1_m0_l0;
							i_m_Atd_xy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xy_o0_t2_m0_l0;
							i_m_Atd_xz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xz_o0_t0_m0_l0;
							i_m_Atd_xz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xz_o0_t1_m0_l0;
							i_m_Atd_xz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xz_o0_t2_m0_l0;
							i_m_Atd_yy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yy_o0_t0_m0_l0;
							i_m_Atd_yy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yy_o0_t1_m0_l0;
							i_m_Atd_yy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yy_o0_t2_m0_l0;
							i_m_Atd_yz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yz_o0_t0_m0_l0;
							i_m_Atd_yz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yz_o0_t1_m0_l0;
							i_m_Atd_yz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yz_o0_t2_m0_l0;
							i_m_Atd_zz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_zz_o0_t0_m0_l0;
							i_m_Atd_zz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_zz_o0_t1_m0_l0;
							i_m_Atd_zz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_zz_o0_t2_m0_l0;
							i_m_Gamh_x_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xx_o0_t6_m0_l0;
							i_m_Gamh_x_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xy_o0_t6_m0_l0;
							i_m_Gamh_x_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xz_o0_t6_m0_l0;
							i_m_Gamh_y_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xx_o0_t7_m0_l0;
							i_m_Gamh_y_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xy_o0_t7_m0_l0;
							i_m_Gamh_y_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xz_o0_t7_m0_l0;
							i_m_Gamh_z_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xx_o0_t8_m0_l0;
							i_m_Gamh_z_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xy_o0_t8_m0_l0;
							i_m_Gamh_z_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xz_o0_t8_m0_l0;
							i_m_Betau_x_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_div_Beta_o0_t0_m0_l0;
							i_m_Betau_x_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_y_o0_t7_m0_l0;
							i_m_Betau_x_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_z_o0_t7_m0_l0;
							i_m_Betau_y_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_x_o0_t8_m0_l0;
							i_m_Betau_y_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_div_Beta_o0_t1_m0_l0;
							i_m_Betau_y_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_z_o0_t8_m0_l0;
							i_m_Betau_z_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_x_o0_t9_m0_l0;
							i_m_Betau_z_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_y_o0_t9_m0_l0;
							i_m_Betau_z_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_div_Beta_o0_t2_m0_l0;
							i_m_Alpha_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Psi1_xx_o0_t1_m0_l0;
							i_m_Alpha_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Psi1_xx_o0_t2_m0_l0;
							i_m_Alpha_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Psi1_xx_o0_t3_m0_l0;
							i_m_chi_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rpd_xx_o0_t0_m0_l0;
							i_m_chi_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rpd_xx_o0_t1_m0_l0;
							i_m_chi_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rpd_xx_o0_t2_m0_l0;
							i_m_trK_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t0_m0_l0;
							i_m_trK_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t1_m0_l0;
							i_m_trK_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t2_m0_l0;
							i_m_theta_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t3_m0_l0;
							i_m_theta_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t4_m0_l0;
							i_m_theta_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t5_m0_l0;
							RHS_gtd_xx = ((i_m_gtd_xx_o0_t0_l0 + i_m_gtd_xx_o0_t1_l0) + i_m_gtd_xx_o0_t2_l0) + i_d_gtd_xx_o0_t3_m0_l0;
							RHS_gtd_xy = ((i_m_gtd_xy_o0_t0_l0 + i_m_gtd_xy_o0_t1_l0) + i_m_gtd_xy_o0_t2_l0) + i_d_gtd_xy_o0_t3_m0_l0;
							RHS_gtd_xz = ((i_m_gtd_xz_o0_t0_l0 + i_m_gtd_xz_o0_t1_l0) + i_m_gtd_xz_o0_t2_l0) + i_d_gtd_xz_o0_t3_m0_l0;
							RHS_gtd_yy = ((i_m_gtd_yy_o0_t0_l0 + i_m_gtd_yy_o0_t1_l0) + i_m_gtd_yy_o0_t2_l0) + i_d_gtd_yy_o0_t3_m0_l0;
							RHS_gtd_yz = ((i_m_gtd_yz_o0_t0_l0 + i_m_gtd_yz_o0_t1_l0) + i_m_gtd_yz_o0_t2_l0) + i_d_gtd_yz_o0_t3_m0_l0;
							RHS_gtd_zz = ((i_m_gtd_zz_o0_t0_l0 + i_m_gtd_zz_o0_t1_l0) + i_m_gtd_zz_o0_t2_l0) + i_d_gtd_zz_o0_t3_m0_l0;
							RHS_Atd_xx = ((i_m_Atd_xx_o0_t0_l0 + i_m_Atd_xx_o0_t1_l0) + i_m_Atd_xx_o0_t2_l0) + i_d_Atd_xx_o0_t3_m0_l0;
							RHS_Atd_xy = ((i_m_Atd_xy_o0_t0_l0 + i_m_Atd_xy_o0_t1_l0) + i_m_Atd_xy_o0_t2_l0) + i_d_Atd_xy_o0_t3_m0_l0;
							RHS_Atd_xz = ((i_m_Atd_xz_o0_t0_l0 + i_m_Atd_xz_o0_t1_l0) + i_m_Atd_xz_o0_t2_l0) + i_d_Atd_xz_o0_t3_m0_l0;
							RHS_Atd_yy = ((i_m_Atd_yy_o0_t0_l0 + i_m_Atd_yy_o0_t1_l0) + i_m_Atd_yy_o0_t2_l0) + i_d_Atd_yy_o0_t3_m0_l0;
							RHS_Atd_yz = ((i_m_Atd_yz_o0_t0_l0 + i_m_Atd_yz_o0_t1_l0) + i_m_Atd_yz_o0_t2_l0) + i_d_Atd_yz_o0_t3_m0_l0;
							RHS_Atd_zz = ((i_m_Atd_zz_o0_t0_l0 + i_m_Atd_zz_o0_t1_l0) + i_m_Atd_zz_o0_t2_l0) + i_d_Atd_zz_o0_t3_m0_l0;
							RHS_Gamh_x = ((i_m_Gamh_x_o0_t0_l0 + i_m_Gamh_x_o0_t1_l0) + i_m_Gamh_x_o0_t2_l0) + i_d_Gamh_x_o0_t3_m0_l0;
							RHS_Gamh_y = ((i_m_Gamh_y_o0_t0_l0 + i_m_Gamh_y_o0_t1_l0) + i_m_Gamh_y_o0_t2_l0) + i_d_Gamh_y_o0_t3_m0_l0;
							RHS_Gamh_z = ((i_m_Gamh_z_o0_t0_l0 + i_m_Gamh_z_o0_t1_l0) + i_m_Gamh_z_o0_t2_l0) + i_d_Gamh_z_o0_t3_m0_l0;
							RHS_Betau_x = ((i_m_Betau_x_o0_t0_l0 + i_m_Betau_x_o0_t1_l0) + i_m_Betau_x_o0_t2_l0) + i_d_Betau_x_o0_t3_m0_l0;
							RHS_Betau_y = ((i_m_Betau_y_o0_t0_l0 + i_m_Betau_y_o0_t1_l0) + i_m_Betau_y_o0_t2_l0) + i_d_Betau_y_o0_t3_m0_l0;
							RHS_Betau_z = ((i_m_Betau_z_o0_t0_l0 + i_m_Betau_z_o0_t1_l0) + i_m_Betau_z_o0_t2_l0) + i_d_Betau_z_o0_t3_m0_l0;
							RHS_Alpha = ((i_m_Alpha_o0_t0_l0 + i_m_Alpha_o0_t1_l0) + i_m_Alpha_o0_t2_l0) + i_d_Alpha_o0_t3_m0_l0;
							RHS_chi = ((i_m_chi_o0_t0_l0 + i_m_chi_o0_t1_l0) + i_m_chi_o0_t2_l0) + i_d_chi_o0_t3_m0_l0;
							RHS_trK = ((i_m_trK_o0_t0_l0 + i_m_trK_o0_t1_l0) + i_m_trK_o0_t2_l0) + i_d_trK_o0_t3_m0_l0;
							RHS_theta = ((i_m_theta_o0_t0_l0 + i_m_theta_o0_t1_l0) + i_m_theta_o0_t2_l0) + i_d_theta_o0_t3_m0_l0;
						}
						if (dissipation_factor_Sfd_x > 0.0) {
							RHS_Sfd_x = RHS_Sfd_x + dissipation_factor_Sfd_x * (meshDissipation_i(rk1Sfd_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Sfd_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Sfd_x, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Sfd_y > 0.0) {
							RHS_Sfd_y = RHS_Sfd_y + dissipation_factor_Sfd_y * (meshDissipation_i(rk1Sfd_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Sfd_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Sfd_y, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Sfd_z > 0.0) {
							RHS_Sfd_z = RHS_Sfd_z + dissipation_factor_Sfd_z * (meshDissipation_i(rk1Sfd_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Sfd_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Sfd_z, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Bfu_x > 0.0) {
							RHS_Bfu_x = RHS_Bfu_x + dissipation_factor_Bfu_x * (meshDissipation_i(rk1Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Bfu_y > 0.0) {
							RHS_Bfu_y = RHS_Bfu_y + dissipation_factor_Bfu_y * (meshDissipation_i(rk1Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Bfu_z > 0.0) {
							RHS_Bfu_z = RHS_Bfu_z + dissipation_factor_Bfu_z * (meshDissipation_i(rk1Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Df > 0.0) {
							RHS_Df = RHS_Df + dissipation_factor_Df * (meshDissipation_i(rk1Df, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Df, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Df, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_DYf > 0.0) {
							RHS_DYf = RHS_DYf + dissipation_factor_DYf * (meshDissipation_i(rk1DYf, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1DYf, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1DYf, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_tauf > 0.0) {
							RHS_tauf = RHS_tauf + dissipation_factor_tauf * (meshDissipation_i(rk1tauf, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1tauf, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1tauf, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_phif > 0.0) {
							RHS_phif = RHS_phif + dissipation_factor_phif * (meshDissipation_i(rk1phif, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1phif, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1phif, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_xx > 0.0) {
							RHS_gtd_xx = RHS_gtd_xx + dissipation_factor_gtd_xx * (meshDissipation_i(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_xy > 0.0) {
							RHS_gtd_xy = RHS_gtd_xy + dissipation_factor_gtd_xy * (meshDissipation_i(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_xz > 0.0) {
							RHS_gtd_xz = RHS_gtd_xz + dissipation_factor_gtd_xz * (meshDissipation_i(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_yy > 0.0) {
							RHS_gtd_yy = RHS_gtd_yy + dissipation_factor_gtd_yy * (meshDissipation_i(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_yz > 0.0) {
							RHS_gtd_yz = RHS_gtd_yz + dissipation_factor_gtd_yz * (meshDissipation_i(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_zz > 0.0) {
							RHS_gtd_zz = RHS_gtd_zz + dissipation_factor_gtd_zz * (meshDissipation_i(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_xx > 0.0) {
							RHS_Atd_xx = RHS_Atd_xx + dissipation_factor_Atd_xx * (meshDissipation_i(rk1Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_xy > 0.0) {
							RHS_Atd_xy = RHS_Atd_xy + dissipation_factor_Atd_xy * (meshDissipation_i(rk1Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_xz > 0.0) {
							RHS_Atd_xz = RHS_Atd_xz + dissipation_factor_Atd_xz * (meshDissipation_i(rk1Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_yy > 0.0) {
							RHS_Atd_yy = RHS_Atd_yy + dissipation_factor_Atd_yy * (meshDissipation_i(rk1Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_yz > 0.0) {
							RHS_Atd_yz = RHS_Atd_yz + dissipation_factor_Atd_yz * (meshDissipation_i(rk1Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_zz > 0.0) {
							RHS_Atd_zz = RHS_Atd_zz + dissipation_factor_Atd_zz * (meshDissipation_i(rk1Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Gamh_x > 0.0) {
							RHS_Gamh_x = RHS_Gamh_x + dissipation_factor_Gamh_x * (meshDissipation_i(rk1Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Gamh_y > 0.0) {
							RHS_Gamh_y = RHS_Gamh_y + dissipation_factor_Gamh_y * (meshDissipation_i(rk1Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Gamh_z > 0.0) {
							RHS_Gamh_z = RHS_Gamh_z + dissipation_factor_Gamh_z * (meshDissipation_i(rk1Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Betau_x > 0.0) {
							RHS_Betau_x = RHS_Betau_x + dissipation_factor_Betau_x * (meshDissipation_i(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Betau_y > 0.0) {
							RHS_Betau_y = RHS_Betau_y + dissipation_factor_Betau_y * (meshDissipation_i(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Betau_z > 0.0) {
							RHS_Betau_z = RHS_Betau_z + dissipation_factor_Betau_z * (meshDissipation_i(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Alpha > 0.0) {
							RHS_Alpha = RHS_Alpha + dissipation_factor_Alpha * (meshDissipation_i(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1Alpha, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_chi > 0.0) {
							RHS_chi = RHS_chi + dissipation_factor_chi * (meshDissipation_i(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1chi, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_trK > 0.0) {
							RHS_trK = RHS_trK + dissipation_factor_trK * (meshDissipation_i(rk1trK, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1trK, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1trK, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_theta > 0.0) {
							RHS_theta = RHS_theta + dissipation_factor_theta * (meshDissipation_i(rk1theta, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk1theta, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk1theta, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						vector(rk2Sfd_x, i, j, k) = RK4P2_(RHS_Sfd_x, vector(Sfd_x_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Sfd_y, i, j, k) = RK4P2_(RHS_Sfd_y, vector(Sfd_y_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Sfd_z, i, j, k) = RK4P2_(RHS_Sfd_z, vector(Sfd_z_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Bfu_x, i, j, k) = RK4P2_(RHS_Bfu_x, vector(Bfu_x_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Bfu_y, i, j, k) = RK4P2_(RHS_Bfu_y, vector(Bfu_y_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Bfu_z, i, j, k) = RK4P2_(RHS_Bfu_z, vector(Bfu_z_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Df, i, j, k) = RK4P2_(RHS_Df, vector(Df_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2DYf, i, j, k) = RK4P2_(RHS_DYf, vector(DYf_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2tauf, i, j, k) = RK4P2_(RHS_tauf, vector(tauf_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2phif, i, j, k) = RK4P2_(RHS_phif, vector(phif_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2gtd_xx, i, j, k) = RK4P2_(RHS_gtd_xx, vector(gtd_xx_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2gtd_xy, i, j, k) = RK4P2_(RHS_gtd_xy, vector(gtd_xy_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2gtd_xz, i, j, k) = RK4P2_(RHS_gtd_xz, vector(gtd_xz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2gtd_yy, i, j, k) = RK4P2_(RHS_gtd_yy, vector(gtd_yy_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2gtd_yz, i, j, k) = RK4P2_(RHS_gtd_yz, vector(gtd_yz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2gtd_zz, i, j, k) = RK4P2_(RHS_gtd_zz, vector(gtd_zz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Atd_xx, i, j, k) = RK4P2_(RHS_Atd_xx, vector(Atd_xx_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Atd_xy, i, j, k) = RK4P2_(RHS_Atd_xy, vector(Atd_xy_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Atd_xz, i, j, k) = RK4P2_(RHS_Atd_xz, vector(Atd_xz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Atd_yy, i, j, k) = RK4P2_(RHS_Atd_yy, vector(Atd_yy_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Atd_yz, i, j, k) = RK4P2_(RHS_Atd_yz, vector(Atd_yz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Atd_zz, i, j, k) = RK4P2_(RHS_Atd_zz, vector(Atd_zz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Gamh_x, i, j, k) = RK4P2_(RHS_Gamh_x, vector(Gamh_x_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Gamh_y, i, j, k) = RK4P2_(RHS_Gamh_y, vector(Gamh_y_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Gamh_z, i, j, k) = RK4P2_(RHS_Gamh_z, vector(Gamh_z_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Betau_x, i, j, k) = RK4P2_(RHS_Betau_x, vector(Betau_x_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Betau_y, i, j, k) = RK4P2_(RHS_Betau_y, vector(Betau_y_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Betau_z, i, j, k) = RK4P2_(RHS_Betau_z, vector(Betau_z_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2Alpha, i, j, k) = RK4P2_(RHS_Alpha, vector(Alpha_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2chi, i, j, k) = RK4P2_(RHS_chi, vector(chi_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2trK, i, j, k) = RK4P2_(RHS_trK, vector(trK_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk2theta, i, j, k) = RK4P2_(RHS_theta, vector(theta_p, i, j, k), dx, simPlat_dt, ilast, jlast);
					}
				}
			}
		}
	}
	//Fill ghosts and periodical boundaries
	time_interpolate_operator_mesh1->setStep(2);
	d_bdry_sched_advance9[ln]->fillData(current_time + simPlat_dt * 0.5, false);
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		//Hard region field distance variables
		double* d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
	
		double* rk2Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_x_id).get())->getPointer();
		double* rk2Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_y_id).get())->getPointer();
		double* rk2Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_z_id).get())->getPointer();
		double* rk2Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_x_id).get())->getPointer();
		double* rk2Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_y_id).get())->getPointer();
		double* rk2Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_z_id).get())->getPointer();
		double* rk2Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Df_id).get())->getPointer();
		double* rk2DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2DYf_id).get())->getPointer();
		double* rk2tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2tauf_id).get())->getPointer();
		double* rk2phif = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2phif_id).get())->getPointer();
		double* rk2gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xx_id).get())->getPointer();
		double* rk2gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xy_id).get())->getPointer();
		double* rk2gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xz_id).get())->getPointer();
		double* rk2gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yy_id).get())->getPointer();
		double* rk2gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yz_id).get())->getPointer();
		double* rk2gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_zz_id).get())->getPointer();
		double* rk2Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xx_id).get())->getPointer();
		double* rk2Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xy_id).get())->getPointer();
		double* rk2Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xz_id).get())->getPointer();
		double* rk2Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_yy_id).get())->getPointer();
		double* rk2Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_yz_id).get())->getPointer();
		double* rk2Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_zz_id).get())->getPointer();
		double* rk2Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_x_id).get())->getPointer();
		double* rk2Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_y_id).get())->getPointer();
		double* rk2Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_z_id).get())->getPointer();
		double* rk2Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_x_id).get())->getPointer();
		double* rk2Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_y_id).get())->getPointer();
		double* rk2Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_z_id).get())->getPointer();
		double* rk2Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Alpha_id).get())->getPointer();
		double* rk2chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2chi_id).get())->getPointer();
		double* rk2trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2trK_id).get())->getPointer();
		double* rk2theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2theta_id).get())->getPointer();
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, d_ghost_width, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, d_ghost_width, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, d_ghost_width, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, d_ghost_width, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, d_ghost_width, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, d_ghost_width, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, d_ghost_width, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, d_ghost_width, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, d_ghost_width, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i - 1, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i - 1, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i - 1, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i - 1, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i - 1, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i - 1, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i - 1, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i - 1, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i - 1, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, d_ghost_width, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, d_ghost_width, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, d_ghost_width, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, d_ghost_width, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, d_ghost_width, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, d_ghost_width, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, d_ghost_width, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, d_ghost_width, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, d_ghost_width, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j - 1, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j - 1, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j - 1, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j - 1, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j - 1, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j - 1, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j - 1, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j - 1, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j - 1, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, d_ghost_width);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, d_ghost_width);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, d_ghost_width);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, d_ghost_width);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, d_ghost_width);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, d_ghost_width);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, d_ghost_width);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, d_ghost_width);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, d_ghost_width);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, k - 1);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, k - 1);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, k - 1);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, k - 1);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, k - 1);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, k - 1);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, k - 1);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, k - 1);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, k - 1);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, d_ghost_width, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, d_ghost_width, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, d_ghost_width, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, d_ghost_width, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, d_ghost_width, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, d_ghost_width, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, d_ghost_width, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, d_ghost_width, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, d_ghost_width, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i - 1, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i - 1, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i - 1, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i - 1, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i - 1, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i - 1, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i - 1, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i - 1, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i - 1, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, d_ghost_width, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, d_ghost_width, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, d_ghost_width, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, d_ghost_width, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, d_ghost_width, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, d_ghost_width, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, d_ghost_width, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, d_ghost_width, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, d_ghost_width, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j - 1, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j - 1, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j - 1, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j - 1, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j - 1, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j - 1, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j - 1, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j - 1, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j - 1, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, d_ghost_width);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, d_ghost_width);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, d_ghost_width);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, d_ghost_width);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, d_ghost_width);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, d_ghost_width);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, d_ghost_width);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, d_ghost_width);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, d_ghost_width);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, k - 1);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, k - 1);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, k - 1);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, k - 1);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, k - 1);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, k - 1);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, k - 1);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, k - 1);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, k - 1);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, d_ghost_width, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, d_ghost_width, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, d_ghost_width, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, d_ghost_width, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, d_ghost_width, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, d_ghost_width, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, d_ghost_width, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, d_ghost_width, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, d_ghost_width, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i - 1, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i - 1, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i - 1, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i - 1, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i - 1, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i - 1, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i - 1, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i - 1, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i - 1, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, d_ghost_width, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, d_ghost_width, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, d_ghost_width, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, d_ghost_width, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, d_ghost_width, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, d_ghost_width, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, d_ghost_width, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, d_ghost_width, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, d_ghost_width, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j - 1, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j - 1, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j - 1, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j - 1, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j - 1, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j - 1, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j - 1, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j - 1, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j - 1, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, d_ghost_width);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, d_ghost_width);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, d_ghost_width);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, d_ghost_width);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, d_ghost_width);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, d_ghost_width);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, d_ghost_width);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, d_ghost_width);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, d_ghost_width);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, k - 1);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, k - 1);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, k - 1);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, k - 1);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, k - 1);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, k - 1);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, k - 1);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, k - 1);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, k - 1);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, d_ghost_width, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, d_ghost_width, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, d_ghost_width, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, d_ghost_width, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, d_ghost_width, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, d_ghost_width, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, d_ghost_width, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, d_ghost_width, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, d_ghost_width, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i - 1, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i - 1, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i - 1, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i - 1, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i - 1, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i - 1, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i - 1, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i - 1, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i - 1, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, d_ghost_width, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, d_ghost_width, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, d_ghost_width, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, d_ghost_width, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, d_ghost_width, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, d_ghost_width, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, d_ghost_width, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, d_ghost_width, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, d_ghost_width, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j - 1, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j - 1, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j - 1, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j - 1, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j - 1, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j - 1, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j - 1, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j - 1, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j - 1, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, d_ghost_width);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, d_ghost_width);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, d_ghost_width);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, d_ghost_width);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, d_ghost_width);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, d_ghost_width);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, d_ghost_width);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, d_ghost_width);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, d_ghost_width);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, k - 1);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, k - 1);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, k - 1);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, k - 1);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, k - 1);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, k - 1);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, k - 1);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, k - 1);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, k - 1);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, d_ghost_width, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, d_ghost_width, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, d_ghost_width, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, d_ghost_width, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, d_ghost_width, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, d_ghost_width, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, d_ghost_width, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, d_ghost_width, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, d_ghost_width, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i - 1, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i - 1, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i - 1, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i - 1, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i - 1, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i - 1, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i - 1, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i - 1, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i - 1, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, d_ghost_width, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, d_ghost_width, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, d_ghost_width, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, d_ghost_width, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, d_ghost_width, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, d_ghost_width, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, d_ghost_width, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, d_ghost_width, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, d_ghost_width, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j - 1, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j - 1, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j - 1, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j - 1, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j - 1, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j - 1, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j - 1, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j - 1, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j - 1, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, d_ghost_width);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, d_ghost_width);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, d_ghost_width);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, d_ghost_width);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, d_ghost_width);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, d_ghost_width);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, d_ghost_width);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, d_ghost_width);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, d_ghost_width);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, k - 1);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, k - 1);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, k - 1);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, k - 1);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, k - 1);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, k - 1);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, k - 1);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, k - 1);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, k - 1);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, d_ghost_width, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, d_ghost_width, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, d_ghost_width, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, d_ghost_width, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, d_ghost_width, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, d_ghost_width, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, d_ghost_width, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, d_ghost_width, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, d_ghost_width, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i - 1, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i - 1, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i - 1, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i - 1, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i - 1, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i - 1, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i - 1, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i - 1, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i - 1, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, d_ghost_width, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, d_ghost_width, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, d_ghost_width, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, d_ghost_width, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, d_ghost_width, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, d_ghost_width, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, d_ghost_width, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, d_ghost_width, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, d_ghost_width, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j - 1, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j - 1, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j - 1, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j - 1, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j - 1, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j - 1, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j - 1, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j - 1, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j - 1, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, d_ghost_width);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, d_ghost_width);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, d_ghost_width);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, d_ghost_width);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, d_ghost_width);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, d_ghost_width);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, d_ghost_width);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, d_ghost_width);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, d_ghost_width);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, k - 1);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, k - 1);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, k - 1);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, k - 1);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, k - 1);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, k - 1);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, k - 1);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, k - 1);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, k - 1);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, d_ghost_width, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, d_ghost_width, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, d_ghost_width, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, d_ghost_width, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, d_ghost_width, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, d_ghost_width, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, d_ghost_width, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, d_ghost_width, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, d_ghost_width, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i - 1, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i - 1, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i - 1, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i - 1, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i - 1, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i - 1, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i - 1, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i - 1, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i - 1, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, d_ghost_width, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, d_ghost_width, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, d_ghost_width, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, d_ghost_width, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, d_ghost_width, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, d_ghost_width, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, d_ghost_width, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, d_ghost_width, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, d_ghost_width, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j - 1, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j - 1, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j - 1, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j - 1, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j - 1, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j - 1, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j - 1, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j - 1, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j - 1, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, d_ghost_width);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, d_ghost_width);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, d_ghost_width);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, d_ghost_width);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, d_ghost_width);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, d_ghost_width);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, d_ghost_width);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, d_ghost_width);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, d_ghost_width);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, k - 1);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, k - 1);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, k - 1);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, k - 1);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, k - 1);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, k - 1);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, k - 1);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, k - 1);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, k - 1);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, d_ghost_width, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, d_ghost_width, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, d_ghost_width, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, d_ghost_width, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, d_ghost_width, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, d_ghost_width, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, d_ghost_width, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, d_ghost_width, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, d_ghost_width, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i - 1, j, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i - 1, j, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i - 1, j, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i - 1, j, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i - 1, j, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i - 1, j, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i - 1, j, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i - 1, j, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i - 1, j, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, d_ghost_width, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, d_ghost_width, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, d_ghost_width, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, d_ghost_width, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, d_ghost_width, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, d_ghost_width, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, d_ghost_width, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, d_ghost_width, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, d_ghost_width, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j - 1, k);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j - 1, k);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j - 1, k);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j - 1, k);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j - 1, k);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j - 1, k);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j - 1, k);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j - 1, k);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j - 1, k);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, d_ghost_width);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, d_ghost_width);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, d_ghost_width);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, d_ghost_width);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, d_ghost_width);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, d_ghost_width);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, d_ghost_width);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, d_ghost_width);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, d_ghost_width);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk2Sfd_x, i, j, k) = vector(rk2Sfd_x, i, j, k - 1);
						vector(rk2Sfd_y, i, j, k) = vector(rk2Sfd_y, i, j, k - 1);
						vector(rk2Sfd_z, i, j, k) = vector(rk2Sfd_z, i, j, k - 1);
						vector(rk2Bfu_x, i, j, k) = vector(rk2Bfu_x, i, j, k - 1);
						vector(rk2Bfu_y, i, j, k) = vector(rk2Bfu_y, i, j, k - 1);
						vector(rk2Bfu_z, i, j, k) = vector(rk2Bfu_z, i, j, k - 1);
						vector(rk2Df, i, j, k) = vector(rk2Df, i, j, k - 1);
						vector(rk2DYf, i, j, k) = vector(rk2DYf, i, j, k - 1);
						vector(rk2tauf, i, j, k) = vector(rk2tauf, i, j, k - 1);
						vector(rk2phif, i, j, k) = vector(rk2phif, i, j, k - 1);
					}
				}
			}
		}
	}
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		double* rk2chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2chi_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* rk2Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_x_id).get())->getPointer();
		double* rk2Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_y_id).get())->getPointer();
		double* rk2Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_z_id).get())->getPointer();
		double* rk2tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2tauf_id).get())->getPointer();
		double* rk2gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xy_id).get())->getPointer();
		double* rk2gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yy_id).get())->getPointer();
		double* rk2gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yz_id).get())->getPointer();
		double* rk2gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xz_id).get())->getPointer();
		double* rk2gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_zz_id).get())->getPointer();
		double* rk2gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xx_id).get())->getPointer();
		double* rk2Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_x_id).get())->getPointer();
		double* rk2Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_y_id).get())->getPointer();
		double* rk2Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_z_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* rk2Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Df_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* rk2Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Alpha_id).get())->getPointer();
		double* rk2DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2DYf_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double rk2chi_max, rk2h, rk2sdetg, rk2Bu_x, rk2Bu_y, rk2Bu_z, rk2tau, rk2inv_chi, rk2Bfd_y, rk2Bfd_z, rk2Bfd_x, rk2decay_factor, rk2detgtd, rk2idetgtd, rk2gtu_zz, rk2gtu_xx, rk2gtu_xz, rk2gtu_xy, rk2Sfu_x, rk2Su_x, rk2vfu_x, rk2Sd_x, rk2Sd_y, rk2gtu_yy, rk2Sd_z, rk2gtu_yz, rk2vfu_z, rk2Sfu_y, rk2Su_y, rk2Sfu_z, rk2Su_z, rk2vfu_y, rk2sqvf, rk2sqW, rk2W, rk2Bd_x, rk2D, rk2Bd_y, rk2Bd_z, rhof_tmp, Yef_tmp, Tf_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, dpfdye_tmp, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tauf_tmp, Df_tmp, DYf_tmp, Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, Efu_x_tmp, Efu_y_tmp, Efu_z_tmp, vacuum_D_tmp, vacuum_tau_tmp, vacuum_rho_reset_tmp;
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j, k));
					rk2h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
					rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
					rk2Bu_x = vector(rk2Bfu_x, i, j, k) / rk2sdetg;
					rk2Bu_y = vector(rk2Bfu_y, i, j, k) / rk2sdetg;
					rk2Bu_z = vector(rk2Bfu_z, i, j, k) / rk2sdetg;
					rk2tau = vector(rk2tauf, i, j, k) / rk2sdetg;
					rk2inv_chi = 1.0 / fabs(rk2chi_max);
					rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_yy, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Bfu_z, i, j, k));
					rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_zz, i, j, k) * vector(rk2Bfu_z, i, j, k));
					rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_xy, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_xz, i, j, k) * vector(rk2Bfu_z, i, j, k));
					rk2decay_factor = MIN(1.0, pow((R_0 / MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))), eta_damping_exp));
					rk2detgtd = vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) * vector(rk2gtd_xx, i, j, k) + (-vector(rk2gtd_yy, i, j, k) * (vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k))) + (-vector(rk2gtd_zz, i, j, k) * (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k))) + (-(vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k)) * vector(rk2gtd_xx, i, j, k)) + 2.0 * vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k);
					rk2idetgtd = 1.0 / rk2detgtd;
					rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yy, i, j, k) - vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k));
					rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) - vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k));
					rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_yz, i, j, k) - vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_xz, i, j, k));
					rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_zz, i, j, k)) + vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xz, i, j, k));
					rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j, k) + rk2gtu_xy * vector(rk2Sfd_y, i, j, k) + rk2gtu_xz * vector(rk2Sfd_z, i, j, k));
					rk2Su_x = rk2Sfu_x / rk2sdetg;
					rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j, k) + rk2gtu_xy * vector(vfd_y, i, j, k) + rk2gtu_xz * vector(vfd_z, i, j, k));
					rk2Sd_x = vector(rk2Sfd_x, i, j, k) / rk2sdetg;
					rk2Sd_y = vector(rk2Sfd_y, i, j, k) / rk2sdetg;
					rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_zz, i, j, k) - vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k));
					rk2Sd_z = vector(rk2Sfd_z, i, j, k) / rk2sdetg;
					rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yz, i, j, k)) + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k));
					rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j, k) + rk2gtu_yz * vector(vfd_y, i, j, k) + rk2gtu_zz * vector(vfd_z, i, j, k));
					rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j, k) + rk2gtu_yy * vector(rk2Sfd_y, i, j, k) + rk2gtu_yz * vector(rk2Sfd_z, i, j, k));
					rk2Su_y = rk2Sfu_y / rk2sdetg;
					rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j, k) + rk2gtu_yz * vector(rk2Sfd_y, i, j, k) + rk2gtu_zz * vector(rk2Sfd_z, i, j, k));
					rk2Su_z = rk2Sfu_z / rk2sdetg;
					rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j, k) + rk2gtu_yy * vector(vfd_y, i, j, k) + rk2gtu_yz * vector(vfd_z, i, j, k));
					rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j, k) + rk2vfu_y * vector(vfd_y, i, j, k) + rk2vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
					rk2sqW = 1.0 / (1.0 - rk2sqvf);
					rk2W = sqrt(rk2sqW);
					rk2Bd_x = rk2Bfd_x / rk2sdetg;
					rk2D = vector(rk2Df, i, j, k) / rk2sdetg;
					rk2Bd_y = rk2Bfd_y / rk2sdetg;
					rk2Bd_z = rk2Bfd_z / rk2sdetg;
					vector(t_chie, i, j, k) = vector(chie, i, j, k);
					vector(t_chia, i, j, k) = vector(chia, i, j, k);
					vector(t_optdepthx, i, j, k) = vector(optdepthx, i, j, k);
					vector(t_optdeptha, i, j, k) = vector(optdeptha, i, j, k);
					vector(t_optdepthe, i, j, k) = vector(optdepthe, i, j, k);
					vector(rk2Alpha, i, j, k) = fabs(vector(rk2Alpha, i, j, k));
					rhof_tmp = vector(rhof, i, j, k);
					Yef_tmp = vector(rk2DYf, i, j, k) / vector(rk2Df, i, j, k);
					if (Yef_tmp < vacuum_ye) {
						vector(rk2DYf, i, j, k) = vacuum_ye_reset * vector(rk2Df, i, j, k);
						Yef_tmp = vacuum_ye_reset;
					}
					if (Yef_tmp > ye_maximum) {
						vector(rk2DYf, i, j, k) = 0.9 * ye_maximum * vector(rk2Df, i, j, k);
						Yef_tmp = 0.9 * ye_maximum;
					}
					Tf_tmp = vector(Tf, i, j, k);
					pf_tmp = vector(pf, i, j, k);
					epsf_tmp = vector(epsf, i, j, k);
					vfd_x_tmp = vector(vfd_x, i, j, k);
					vfd_y_tmp = vector(vfd_y, i, j, k);
					vfd_z_tmp = vector(vfd_z, i, j, k);
					sqcs_tmp = vector(sqcs, i, j, k);
					dpfdrho_tmp = vector(dpfdrho, i, j, k);
					dpfdeps_tmp = vector(dpfdeps, i, j, k);
					dpfdye_tmp = vector(dpfdye, i, j, k);
					Sfd_x_tmp = vector(rk2Sfd_x, i, j, k);
					Sfd_y_tmp = vector(rk2Sfd_y, i, j, k);
					Sfd_z_tmp = vector(rk2Sfd_z, i, j, k);
					tauf_tmp = vector(rk2tauf, i, j, k);
					Df_tmp = vector(rk2Df, i, j, k);
					DYf_tmp = vector(rk2DYf, i, j, k);
					Bfu_x_tmp = vector(rk2Bfu_x, i, j, k);
					Bfu_y_tmp = vector(rk2Bfu_y, i, j, k);
					Bfu_z_tmp = vector(rk2Bfu_z, i, j, k);
					if (equalsEq(externalCon2Prim, 1.0)) {
						Efu_x_tmp = vector(Efu_x, i, j, k);
						Efu_y_tmp = vector(Efu_y, i, j, k);
						Efu_z_tmp = vector(Efu_z, i, j, k);
						externalCon2prim(Efu_x_tmp, Efu_y_tmp, Efu_z_tmp, rhof_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, Df_tmp, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tauf_tmp, rk2W, vector(rk2chi, i, j, k), rk2gtu_xx, rk2gtu_xy, rk2gtu_xz, rk2gtu_yy, rk2gtu_yz, rk2gtu_zz, vector(rk2gtd_xx, i, j, k), vector(rk2gtd_xy, i, j, k), vector(rk2gtd_xz, i, j, k), vector(rk2gtd_yy, i, j, k), vector(rk2gtd_yz, i, j, k), vector(rk2gtd_zz, i, j, k), Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, dx, simPlat_dt, ilast, jlast);
						vector(Efu_x, i, j, k) = Efu_x_tmp;
						vector(Efu_y, i, j, k) = Efu_y_tmp;
						vector(Efu_z, i, j, k) = Efu_z_tmp;
					} else {
						vacuum_D_tmp = vacuum_D * rk2decay_factor;
						vacuum_tau_tmp = vacuum_tau * rk2decay_factor;
						vacuum_rho_reset_tmp = vacuum_rho_reset * rk2decay_factor;
						con2prim(eos_type, rhof_tmp, Yef_tmp, Tf_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, dpfdye_tmp, rho_0, rho_1, rho_2, a_0, a_1, a_2, a_3, K_0, K_1, K_2, K_3, gamma_0, gamma_1, gamma_2, gamma_3, rk2D, vacuum_D_tmp, Df_tmp, DYf_tmp, rk2sdetg, rk2Sd_x, rk2Sd_y, rk2Sd_z, rk2Su_x, rk2Su_y, rk2Su_z, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, rk2tau, vacuum_tau_tmp, tauf_tmp, rk2W, rk2h, vector(rk2chi, i, j, k), rk2gtu_xx, rk2gtu_xy, rk2gtu_xz, rk2gtu_yy, rk2gtu_yz, rk2gtu_zz, vector(rk2gtd_xx, i, j, k), vector(rk2gtd_xy, i, j, k), vector(rk2gtd_xz, i, j, k), vector(rk2gtd_yy, i, j, k), vector(rk2gtd_yz, i, j, k), vector(rk2gtd_zz, i, j, k), rk2Bd_x, rk2Bd_y, rk2Bd_z, rk2Bu_x, rk2Bu_y, rk2Bu_z, Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, max_error, gamma, rk2inv_chi, threshold_vacuum, threshold_sqSmax, threshold_sqBmax, energyShift, vacuum_ye_beta, vacuum_rho, vacuum_rho_reset_tmp, vacuum_P_reset, vacuum_ye_reset, vacuum_temp_reset, vacuum_tau_reset, minTableTemperature, minTableEnergy, coord0_data_Data, coord0_dx_Data, coord0_size_Data, coord1_data_Data, coord1_dx_Data, coord1_size_Data, coord2_data_Data, coord2_dx_Data, coord2_size_Data, vars_data_Data, vars_size_Data, dx, simPlat_dt, ilast, jlast);
						vector(Efu_x, i, j, k) = -((-vector(vfd_z, i, j, k) * rk2Bfd_y) + vector(vfd_y, i, j, k) * rk2Bfd_z) / (rk2sdetg * rk2sdetg);
						vector(Efu_y, i, j, k) = -(vector(vfd_z, i, j, k) * rk2Bfd_x - vector(vfd_x, i, j, k) * rk2Bfd_z) / (rk2sdetg * rk2sdetg);
						vector(Efu_z, i, j, k) = -((-vector(vfd_y, i, j, k) * rk2Bfd_x) + vector(vfd_x, i, j, k) * rk2Bfd_y) / (rk2sdetg * rk2sdetg);
					}
					vector(rhof, i, j, k) = rhof_tmp;
					vector(Yef, i, j, k) = Yef_tmp;
					vector(Tf, i, j, k) = Tf_tmp;
					vector(pf, i, j, k) = pf_tmp;
					vector(epsf, i, j, k) = epsf_tmp;
					vector(vfd_x, i, j, k) = vfd_x_tmp;
					vector(vfd_y, i, j, k) = vfd_y_tmp;
					vector(vfd_z, i, j, k) = vfd_z_tmp;
					vector(sqcs, i, j, k) = sqcs_tmp;
					vector(dpfdye, i, j, k) = dpfdye_tmp;
					vector(dpfdeps, i, j, k) = dpfdeps_tmp;
					vector(dpfdrho, i, j, k) = dpfdrho_tmp;
					vector(rk2Sfd_x, i, j, k) = Sfd_x_tmp;
					vector(rk2Sfd_y, i, j, k) = Sfd_y_tmp;
					vector(rk2Sfd_z, i, j, k) = Sfd_z_tmp;
					vector(rk2tauf, i, j, k) = tauf_tmp;
					vector(rk2Df, i, j, k) = Df_tmp;
					vector(rk2DYf, i, j, k) = DYf_tmp;
					vector(rk2Bfu_x, i, j, k) = Bfu_x_tmp;
					vector(rk2Bfu_y, i, j, k) = Bfu_y_tmp;
					vector(rk2Bfu_z, i, j, k) = Bfu_z_tmp;
					vector(t_chix, i, j, k) = vector(chix, i, j, k);
				}
			}
		}
	}
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* rk2DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2DYf_id).get())->getPointer();
		double* rk2Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Df_id).get())->getPointer();
		double* rk2Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_x_id).get())->getPointer();
		double* rk2Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_y_id).get())->getPointer();
		double* rk2Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_z_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* rk2chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2chi_id).get())->getPointer();
		double* rk2gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xy_id).get())->getPointer();
		double* rk2gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yy_id).get())->getPointer();
		double* rk2gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yz_id).get())->getPointer();
		double* rk2gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xz_id).get())->getPointer();
		double* rk2gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_zz_id).get())->getPointer();
		double* rk2gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xx_id).get())->getPointer();
		double* rk2Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_x_id).get())->getPointer();
		double* rk2Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_y_id).get())->getPointer();
		double* rk2Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_z_id).get())->getPointer();
		double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
		double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
		double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
		double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
		double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
		double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
		double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
		double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
		double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
		double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
		double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
		double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
		double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
		double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
		double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
		double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
		double* qnu_e = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_e_id).get())->getPointer();
		double* qnu_a = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_a_id).get())->getPointer();
		double* qnu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_x_id).get())->getPointer();
		double* rk2Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_x_id).get())->getPointer();
		double* rk2Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_y_id).get())->getPointer();
		double* rk2Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_z_id).get())->getPointer();
		double* rk2Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Alpha_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double d_optdeptheh_o0_t0_m0_l0, d_optdepthev_o0_t0_m0_l0, d_optdepthed_o0_t0_m0_l0, d_optdepthah_o0_t0_m0_l0, d_optdepthav_o0_t0_m0_l0, d_optdepthad_o0_t0_m0_l0, d_optdepthxh_o0_t0_m0_l0, d_optdepthxv_o0_t0_m0_l0, d_optdepthxd_o0_t0_m0_l0, d_chieh_o0_t0_m0_l0, d_chiev_o0_t0_m0_l0, d_chied_o0_t0_m0_l0, d_chiah_o0_t0_m0_l0, d_chiav_o0_t0_m0_l0, d_chiad_o0_t0_m0_l0, d_chixh_o0_t0_m0_l0, d_chixv_o0_t0_m0_l0, d_chixd_o0_t0_m0_l0, d_CovdDYf_z_o0_t0_m0_l0, d_CovdDYf_y_o0_t0_m0_l0, d_CovdDYf_x_o0_t0_m0_l0, d_CovdDf_z_o0_t0_m0_l0, d_CovdDf_y_o0_t0_m0_l0, d_CovdDf_x_o0_t0_m0_l0, d_CovdBfu_xx_o0_t0_m0_l0, d_CovdBfu_xy_o0_t0_m0_l0, d_CovdBfu_xz_o0_t0_m0_l0, d_CovdBfu_yx_o0_t0_m0_l0, d_CovdBfu_yy_o0_t0_m0_l0, d_CovdBfu_yz_o0_t0_m0_l0, d_CovdBfu_zx_o0_t0_m0_l0, d_CovdBfu_zy_o0_t0_m0_l0, d_CovdBfu_zz_o0_t0_m0_l0, d_CovdEfu_xx_o0_t0_m0_l0, d_CovdEfu_xy_o0_t0_m0_l0, d_CovdEfu_xz_o0_t0_m0_l0, d_CovdEfu_yx_o0_t0_m0_l0, d_CovdEfu_yy_o0_t0_m0_l0, d_CovdEfu_yz_o0_t0_m0_l0, d_CovdEfu_zx_o0_t0_m0_l0, d_CovdEfu_zy_o0_t0_m0_l0, d_CovdEfu_zz_o0_t0_m0_l0, d_Covdepsf_x_o0_t0_m0_l0, d_Covdepsf_y_o0_t0_m0_l0, d_Covdepsf_z_o0_t0_m0_l0, d_Covddpfdeps_x_o0_t0_m0_l0, d_Covddpfdeps_y_o0_t0_m0_l0, d_Covddpfdeps_z_o0_t0_m0_l0, d_Covddpfdye_x_o0_t0_m0_l0, d_Covddpfdye_y_o0_t0_m0_l0, d_Covddpfdye_z_o0_t0_m0_l0, d_Covddpfdrho_x_o0_t0_m0_l0, d_Covddpfdrho_y_o0_t0_m0_l0, d_Covddpfdrho_z_o0_t0_m0_l0, d_Covdvfd_xx_o0_t0_m0_l0, d_Covdvfd_xy_o0_t0_m0_l0, d_Covdvfd_xz_o0_t0_m0_l0, d_Covdvfd_yx_o0_t0_m0_l0, d_Covdvfd_yy_o0_t0_m0_l0, d_Covdvfd_yz_o0_t0_m0_l0, d_Covdvfd_zx_o0_t0_m0_l0, d_Covdvfd_zy_o0_t0_m0_l0, d_Covdvfd_zz_o0_t0_m0_l0, d_Covdrhof_x_o0_t0_m0_l0, d_Covdrhof_y_o0_t0_m0_l0, d_Covdrhof_z_o0_t0_m0_l0, d_Covdpf_x_o0_t0_m0_l0, d_Covdpf_y_o0_t0_m0_l0, d_Covdpf_z_o0_t0_m0_l0, d_CovdYef_x_o0_t0_m0_l0, d_CovdYef_y_o0_t0_m0_l0, d_CovdYef_z_o0_t0_m0_l0, rk2chiev, rk2chieh, rk2Covdvfd_xx, rk2Covdvfd_xz, rk2Covdvfd_xy, rk2chied, rk2Covddpfdeps_x, rk2Covddpfdeps_y, rk2Covddpfdeps_z, rk2Covdvfd_zx, rk2Covdvfd_zz, rk2Covdvfd_zy, rk2Covdvfd_yy, rk2Covdvfd_yx, rk2Covdvfd_yz, rk2Covdpf_x, rk2Covdpf_y, rk2Covdpf_z, rk2chi_max, rk2inv_chi, rk2Bfd_y, rk2Bfd_z, rk2Bfd_x, rk2sqBf, rk2chixh, rk2chixd, rk2Covddpfdrho_z, rk2Covddpfdrho_y, rk2Covddpfdrho_x, rk2sdetg, rk2m_CovdDf_x_o0_t0_l0, rk2CovdDf_x, rk2m_CovdDf_y_o0_t0_l0, rk2CovdDf_y, rk2m_CovdDf_z_o0_t0_l0, rk2CovdDf_z, rk2m_CovdBfu_zz_o0_t0_l0, rk2CovdBfu_zz, rk2m_CovdBfu_zy_o0_t0_l0, rk2CovdBfu_zy, rk2m_CovdBfu_zx_o0_t0_l0, rk2CovdBfu_zx, rk2m_CovdBfu_yz_o0_t0_l0, rk2CovdBfu_yz, rk2m_CovdBfu_yy_o0_t0_l0, rk2CovdBfu_yy, rk2m_CovdBfu_yx_o0_t0_l0, rk2CovdBfu_yx, rk2m_CovdBfu_xy_o0_t0_l0, rk2CovdBfu_xy, rk2m_CovdBfu_xx_o0_t0_l0, rk2CovdBfu_xx, rk2m_CovdBfu_xz_o0_t0_l0, rk2CovdBfu_xz, rk2CovdBfd_zx, rk2Covdrhof_z, rk2Covdrhof_y, rk2Covdrhof_x, rk2CovdBfd_zz, rk2CovdBfd_zy, rk2m_CovdDYf_x_o0_t0_l0, rk2CovdDYf_x, rk2m_CovdDYf_y_o0_t0_l0, rk2CovdDYf_y, rk2m_CovdDYf_z_o0_t0_l0, rk2CovdDYf_z, rk2Efd_z, rk2Efd_y, rk2chixv, rk2Efd_x, rk2sqEf, rk2optdepthed, rk2CovdBfd_yy, rk2CovdBfd_yx, rk2CovdBfd_yz, rk2optdeptheh, rk2CovdBvf_z, rk2optdepthev, rk2CovdBvf_y, rk2CovdBvf_x, rk2CovdBfd_xx, rk2CovdBfd_xz, rk2CovdBfd_xy, rk2optdepthxd, rk2optdepthxh, rk2optdepthxv, rk2chiav, rk2CovdEfu_zz, rk2CovdEfu_zy, rk2CovdEfu_zx, rk2CovdEfd_zz, rk2CovdEfd_zx, rk2CovdEfd_zy, rk2Bvf, rk2optdepthah, rk2optdepthad, rk2h, rk2chiah, rk2chiad, rk2CovdYef_y, rk2optdepthav, rk2CovdYef_z, rk2CovdYef_x, rk2detgtd, rk2idetgtd, rk2gtu_zz, rk2gtu_yy, rk2gtu_yz, rk2gtu_xx, rk2gtu_xz, rk2CovuBfu_zx, rk2CovuBfu_zz, rk2CovuBfu_zy, rk2Covdvfu_xz, rk2CovuYef_z, rk2Covdvfu_zz, rk2Covdvfu_yz, rk2Covuvfu_zz, rk2Sfu_z, rk2Covurhof_z, rk2CovuBvf_z, rk2gtu_xy, rk2CovuBfu_yz, rk2CovuBfu_yy, rk2CovuBfu_yx, rk2Covdvfu_xy, rk2Covdvfu_xx, rk2CovdinvsqW_x, rk2CovuYef_x, rk2CovuYef_y, rk2Covdvfu_zy, rk2Covdvfu_zx, rk2CovdinvsqW_z, rk2Covdvfu_yx, rk2Covuvfu_zx, rk2Covdvfu_yy, rk2CovdinvsqW_y, rk2CovuinvsqW_z, rk2Covuvfu_zy, rk2CovuBfu_xz, rk2CovuBfu_xy, rk2CovuBfu_xx, rk2Sfu_y, rk2Sfu_x, rk2Covurhof_x, rk2Covurhof_y, rk2Covuvfu_yz, rk2Covuvfu_yy, rk2Covuvfu_yx, rk2Covuvfu_xz, rk2Covuvfu_xy, rk2Covuvfu_xx, rk2vfu_y, rk2vfu_x, rk2CovuBvf_x, rk2CovuBvf_y, rk2CovuinvsqW_y, rk2CovuinvsqW_x, rk2Covddpfdye_z, rk2Covddpfdye_y, rk2Covddpfdye_x, rk2CovdEfu_xz, rk2CovdEfu_xy, rk2CovdEfu_xx, rk2CovdEfd_xx, rk2CovdEfd_xy, rk2CovdEfd_xz, rk2vfu_z, rk2sqvf, rk2sqW, rk2PhiAh, rk2Epf, rk2Thetaf, rk2Covdepsf_z, rk2Covdhf_z, rk2CovdEpf_z, rk2CovdThetaf_z, rk2Covdepsf_y, rk2Covdhf_y, rk2CovdEpf_y, rk2CovdThetaf_y, rk2Covdepsf_x, rk2Covuepsf_z, rk2Covuepsf_y, rk2Covdhf_x, rk2CovdEpf_x, rk2CovdThetaf_x, rk2CovuThetaf_z, rk2CovuThetaf_x, rk2CovuThetaf_y, rk2Phivh_z, rk2Phivh_x, rk2Phivh_y, rk2PhiMh_yz, rk2PhiMh_xz, rk2PhiMh_xy, rk2Covuepsf_x, rk2faceta, rk2CovdEfu_yz, rk2CovuEfu_zz, rk2CovuEfu_xz, rk2CovdEfu_yy, rk2CovuEfu_zy, rk2CovuEfu_xy, rk2CovdEfu_yx, rk2CovdEfd_yy, rk2CovdEfd_yz, rk2CovdEfd_yx, rk2CovuEfu_zx, rk2CovuEfu_xx, rk2CovuEfu_yx, rk2CovuEfu_yz, rk2CovuEfu_yy, rk2PhiThetah, rk2HPres, rk2HTheta, rk2Hv_z, rk2HN_z, rk2HNe_z, rk2Hv_y, rk2HN_y, rk2HM_yz, rk2HNe_y, rk2Hv_x, rk2HN_x, rk2HM_xz, rk2HM_xy, rk2HE_z, rk2HE_x, rk2HE_y, rk2HT_zz, rk2HT_yy, rk2HT_yz, rk2HT_xx, rk2HT_xz, rk2HT_xy, rk2HNe_x, d_TauN_x_o0_t0_m0_l0, d_TauN_y_o0_t0_m0_l0, d_TauN_z_o0_t0_m0_l0, d_TauNe_x_o0_t0_m0_l0, d_TauNe_y_o0_t0_m0_l0, d_TauNe_z_o0_t0_m0_l0, d_TauM_xy_o0_t0_m0_l0, d_TauM_xz_o0_t0_m0_l0, d_TauM_yz_o0_t0_m0_l0, d_TauT_xx_o0_t0_m0_l0, d_TauT_xy_o0_t0_m0_l0, d_TauT_xz_o0_t0_m0_l0, d_TauT_yy_o0_t0_m0_l0, d_TauT_yz_o0_t0_m0_l0, d_TauT_zz_o0_t0_m0_l0, optdepthe_tmp, optdeptha_tmp, optdepthx_tmp, chie_tmp, chia_tmp, chix_tmp, qnu_tmp, rnu_tmp, qnu_e_tmp, qnu_a_tmp, qnu_x_tmp, betabeta, betav, velr, f_redshift;
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					if ((i + 2 < ilast && i - 2 >= 0 && j + 2 < jlast && j - 2 >= 0 && k + 2 < klast && k - 2 >= 0) && ((i + 3 < ilast || !patch->getPatchGeometry()->getTouchesRegularBoundary(0, 1)) && (i - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(0, 0)) && (j + 3 < jlast || !patch->getPatchGeometry()->getTouchesRegularBoundary(1, 1)) && (j - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(1, 0)) && (k + 3 < klast || !patch->getPatchGeometry()->getTouchesRegularBoundary(2, 1)) && (k - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(2, 0)))) {
						d_optdeptheh_o0_t0_m0_l0 = eikonal_i(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthev_o0_t0_m0_l0 = eikonal_j(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthed_o0_t0_m0_l0 = eikonal_k(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthah_o0_t0_m0_l0 = eikonal_i(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthav_o0_t0_m0_l0 = eikonal_j(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthad_o0_t0_m0_l0 = eikonal_k(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthxh_o0_t0_m0_l0 = eikonal_i(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthxv_o0_t0_m0_l0 = eikonal_j(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthxd_o0_t0_m0_l0 = eikonal_k(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chieh_o0_t0_m0_l0 = eikonal_i(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiev_o0_t0_m0_l0 = eikonal_j(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chied_o0_t0_m0_l0 = eikonal_k(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiah_o0_t0_m0_l0 = eikonal_i(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiav_o0_t0_m0_l0 = eikonal_j(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiad_o0_t0_m0_l0 = eikonal_k(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chixh_o0_t0_m0_l0 = eikonal_i(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chixv_o0_t0_m0_l0 = eikonal_j(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chixd_o0_t0_m0_l0 = eikonal_k(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
						if (calculate_les_terms) {
							d_CovdDYf_z_o0_t0_m0_l0 = D1CDO4_k(rk2DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDYf_y_o0_t0_m0_l0 = D1CDO4_j(rk2DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDYf_x_o0_t0_m0_l0 = D1CDO4_i(rk2DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDf_z_o0_t0_m0_l0 = D1CDO4_k(rk2Df, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDf_y_o0_t0_m0_l0 = D1CDO4_j(rk2Df, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDf_x_o0_t0_m0_l0 = D1CDO4_i(rk2Df, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_xx_o0_t0_m0_l0 = D1CDO4_i(rk2Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_xy_o0_t0_m0_l0 = D1CDO4_i(rk2Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_xz_o0_t0_m0_l0 = D1CDO4_i(rk2Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_yx_o0_t0_m0_l0 = D1CDO4_j(rk2Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_yy_o0_t0_m0_l0 = D1CDO4_j(rk2Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_yz_o0_t0_m0_l0 = D1CDO4_j(rk2Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_zx_o0_t0_m0_l0 = D1CDO4_k(rk2Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_zy_o0_t0_m0_l0 = D1CDO4_k(rk2Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_zz_o0_t0_m0_l0 = D1CDO4_k(rk2Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_xx_o0_t0_m0_l0 = D1CDO4_i(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_xy_o0_t0_m0_l0 = D1CDO4_i(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_xz_o0_t0_m0_l0 = D1CDO4_i(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_yx_o0_t0_m0_l0 = D1CDO4_j(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_yy_o0_t0_m0_l0 = D1CDO4_j(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_yz_o0_t0_m0_l0 = D1CDO4_j(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_zx_o0_t0_m0_l0 = D1CDO4_k(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_zy_o0_t0_m0_l0 = D1CDO4_k(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_zz_o0_t0_m0_l0 = D1CDO4_k(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdepsf_x_o0_t0_m0_l0 = D1CDO4_i(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdepsf_y_o0_t0_m0_l0 = D1CDO4_j(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdepsf_z_o0_t0_m0_l0 = D1CDO4_k(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdeps_x_o0_t0_m0_l0 = D1CDO4_i(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdeps_y_o0_t0_m0_l0 = D1CDO4_j(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdeps_z_o0_t0_m0_l0 = D1CDO4_k(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdye_x_o0_t0_m0_l0 = D1CDO4_i(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdye_y_o0_t0_m0_l0 = D1CDO4_j(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdye_z_o0_t0_m0_l0 = D1CDO4_k(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdrho_x_o0_t0_m0_l0 = D1CDO4_i(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdrho_y_o0_t0_m0_l0 = D1CDO4_j(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdrho_z_o0_t0_m0_l0 = D1CDO4_k(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_xx_o0_t0_m0_l0 = D1CDO4_i(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_xy_o0_t0_m0_l0 = D1CDO4_i(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_xz_o0_t0_m0_l0 = D1CDO4_i(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_yx_o0_t0_m0_l0 = D1CDO4_j(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_yy_o0_t0_m0_l0 = D1CDO4_j(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_yz_o0_t0_m0_l0 = D1CDO4_j(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_zx_o0_t0_m0_l0 = D1CDO4_k(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_zy_o0_t0_m0_l0 = D1CDO4_k(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_zz_o0_t0_m0_l0 = D1CDO4_k(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdrhof_x_o0_t0_m0_l0 = D1CDO4_i(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdrhof_y_o0_t0_m0_l0 = D1CDO4_j(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdrhof_z_o0_t0_m0_l0 = D1CDO4_k(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdpf_x_o0_t0_m0_l0 = D1CDO4_i(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdpf_y_o0_t0_m0_l0 = D1CDO4_j(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdpf_z_o0_t0_m0_l0 = D1CDO4_k(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdYef_x_o0_t0_m0_l0 = D1CDO4_i(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdYef_y_o0_t0_m0_l0 = D1CDO4_j(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdYef_z_o0_t0_m0_l0 = D1CDO4_k(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
						}
						rk2chiev = d_chiev_o0_t0_m0_l0;
						rk2chieh = d_chieh_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk2Covdvfd_xx = d_Covdvfd_xx_o0_t0_m0_l0;
							rk2Covdvfd_xz = d_Covdvfd_xz_o0_t0_m0_l0;
							rk2Covdvfd_xy = d_Covdvfd_xy_o0_t0_m0_l0;
						} else {
							rk2Covdvfd_xx = 0.0;
							rk2Covdvfd_xz = 0.0;
							rk2Covdvfd_xy = 0.0;
						}
						rk2chied = d_chied_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk2Covddpfdeps_x = d_Covddpfdeps_x_o0_t0_m0_l0;
							rk2Covddpfdeps_y = d_Covddpfdeps_y_o0_t0_m0_l0;
							rk2Covddpfdeps_z = d_Covddpfdeps_z_o0_t0_m0_l0;
							rk2Covdvfd_zx = d_Covdvfd_zx_o0_t0_m0_l0;
							rk2Covdvfd_zz = d_Covdvfd_zz_o0_t0_m0_l0;
							rk2Covdvfd_zy = d_Covdvfd_zy_o0_t0_m0_l0;
							rk2Covdvfd_yy = d_Covdvfd_yy_o0_t0_m0_l0;
							rk2Covdvfd_yx = d_Covdvfd_yx_o0_t0_m0_l0;
							rk2Covdvfd_yz = d_Covdvfd_yz_o0_t0_m0_l0;
							rk2Covdpf_x = d_Covdpf_x_o0_t0_m0_l0;
							rk2Covdpf_y = d_Covdpf_y_o0_t0_m0_l0;
							rk2Covdpf_z = d_Covdpf_z_o0_t0_m0_l0;
						} else {
							rk2Covddpfdeps_x = 0.0;
							rk2Covddpfdeps_y = 0.0;
							rk2Covddpfdeps_z = 0.0;
							rk2Covdvfd_zx = 0.0;
							rk2Covdvfd_zz = 0.0;
							rk2Covdvfd_zy = 0.0;
							rk2Covdvfd_yy = 0.0;
							rk2Covdvfd_yx = 0.0;
							rk2Covdvfd_yz = 0.0;
							rk2Covdpf_x = 0.0;
							rk2Covdpf_y = 0.0;
							rk2Covdpf_z = 0.0;
						}
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j, k));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_yy, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_zz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_xy, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_xz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2sqBf = vector(rk2Bfu_x, i, j, k) * rk2Bfd_x + vector(rk2Bfu_y, i, j, k) * rk2Bfd_y + vector(rk2Bfu_z, i, j, k) * rk2Bfd_z;
						rk2chixh = d_chixh_o0_t0_m0_l0;
						rk2chixd = d_chixd_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk2Covddpfdrho_z = d_Covddpfdrho_z_o0_t0_m0_l0;
							rk2Covddpfdrho_y = d_Covddpfdrho_y_o0_t0_m0_l0;
							rk2Covddpfdrho_x = d_Covddpfdrho_x_o0_t0_m0_l0;
						} else {
							rk2Covddpfdrho_z = 0.0;
							rk2Covddpfdrho_y = 0.0;
							rk2Covddpfdrho_x = 0.0;
						}
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						if (calculate_les_terms) {
							rk2m_CovdDf_x_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdDf_x_o0_t0_m0_l0;
							rk2CovdDf_x = rk2m_CovdDf_x_o0_t0_l0;
							rk2m_CovdDf_y_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdDf_y_o0_t0_m0_l0;
							rk2CovdDf_y = rk2m_CovdDf_y_o0_t0_l0;
							rk2m_CovdDf_z_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdDf_z_o0_t0_m0_l0;
							rk2CovdDf_z = rk2m_CovdDf_z_o0_t0_l0;
							rk2m_CovdBfu_zz_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdBfu_zz_o0_t0_m0_l0;
							rk2CovdBfu_zz = rk2m_CovdBfu_zz_o0_t0_l0;
							rk2m_CovdBfu_zy_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdBfu_zy_o0_t0_m0_l0;
							rk2CovdBfu_zy = rk2m_CovdBfu_zy_o0_t0_l0;
							rk2m_CovdBfu_zx_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdBfu_zx_o0_t0_m0_l0;
							rk2CovdBfu_zx = rk2m_CovdBfu_zx_o0_t0_l0;
							rk2m_CovdBfu_yz_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdBfu_yz_o0_t0_m0_l0;
							rk2CovdBfu_yz = rk2m_CovdBfu_yz_o0_t0_l0;
							rk2m_CovdBfu_yy_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdBfu_yy_o0_t0_m0_l0;
							rk2CovdBfu_yy = rk2m_CovdBfu_yy_o0_t0_l0;
							rk2m_CovdBfu_yx_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdBfu_yx_o0_t0_m0_l0;
							rk2CovdBfu_yx = rk2m_CovdBfu_yx_o0_t0_l0;
							rk2m_CovdBfu_xy_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdBfu_xy_o0_t0_m0_l0;
							rk2CovdBfu_xy = rk2m_CovdBfu_xy_o0_t0_l0;
							rk2m_CovdBfu_xx_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdBfu_xx_o0_t0_m0_l0;
							rk2CovdBfu_xx = rk2m_CovdBfu_xx_o0_t0_l0;
							rk2m_CovdBfu_xz_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdBfu_xz_o0_t0_m0_l0;
							rk2CovdBfu_xz = rk2m_CovdBfu_xz_o0_t0_l0;
							rk2CovdBfd_zx = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * rk2CovdBfu_zx + vector(rk2gtd_xy, i, j, k) * rk2CovdBfu_zy + vector(rk2gtd_xz, i, j, k) * rk2CovdBfu_zz);
							rk2Covdrhof_z = d_Covdrhof_z_o0_t0_m0_l0;
							rk2Covdrhof_y = d_Covdrhof_y_o0_t0_m0_l0;
							rk2Covdrhof_x = d_Covdrhof_x_o0_t0_m0_l0;
							rk2CovdBfd_zz = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * rk2CovdBfu_zx + vector(rk2gtd_yz, i, j, k) * rk2CovdBfu_zy + vector(rk2gtd_zz, i, j, k) * rk2CovdBfu_zz);
							rk2CovdBfd_zy = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * rk2CovdBfu_zx + vector(rk2gtd_yy, i, j, k) * rk2CovdBfu_zy + vector(rk2gtd_yz, i, j, k) * rk2CovdBfu_zz);
							rk2m_CovdDYf_x_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdDYf_x_o0_t0_m0_l0;
							rk2CovdDYf_x = rk2m_CovdDYf_x_o0_t0_l0;
							rk2m_CovdDYf_y_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdDYf_y_o0_t0_m0_l0;
							rk2CovdDYf_y = rk2m_CovdDYf_y_o0_t0_l0;
							rk2m_CovdDYf_z_o0_t0_l0 = 1.0 / rk2sdetg * d_CovdDYf_z_o0_t0_m0_l0;
							rk2CovdDYf_z = rk2m_CovdDYf_z_o0_t0_l0;
						} else {
							rk2CovdDf_x = 0.0;
							rk2CovdDf_y = 0.0;
							rk2CovdDf_z = 0.0;
							rk2CovdBfu_zz = 0.0;
							rk2CovdBfu_zy = 0.0;
							rk2CovdBfu_zx = 0.0;
							rk2CovdBfu_yz = 0.0;
							rk2CovdBfu_yy = 0.0;
							rk2CovdBfu_yx = 0.0;
							rk2CovdBfu_xy = 0.0;
							rk2CovdBfu_xx = 0.0;
							rk2CovdBfu_xz = 0.0;
							rk2CovdBfd_zx = 0.0;
							rk2Covdrhof_z = 0.0;
							rk2Covdrhof_y = 0.0;
							rk2Covdrhof_x = 0.0;
							rk2CovdBfd_zz = 0.0;
							rk2CovdBfd_zy = 0.0;
							rk2CovdDYf_x = 0.0;
							rk2CovdDYf_y = 0.0;
							rk2CovdDYf_z = 0.0;
						}
						rk2Efd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * vector(Efu_x, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(Efu_y, i, j, k) + vector(rk2gtd_zz, i, j, k) * vector(Efu_z, i, j, k));
						rk2Efd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * vector(Efu_x, i, j, k) + vector(rk2gtd_yy, i, j, k) * vector(Efu_y, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(Efu_z, i, j, k));
						rk2chixv = d_chixv_o0_t0_m0_l0;
						rk2Efd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * vector(Efu_x, i, j, k) + vector(rk2gtd_xy, i, j, k) * vector(Efu_y, i, j, k) + vector(rk2gtd_xz, i, j, k) * vector(Efu_z, i, j, k));
						rk2sqEf = vector(Efu_x, i, j, k) * rk2Efd_x + vector(Efu_y, i, j, k) * rk2Efd_y + vector(Efu_z, i, j, k) * rk2Efd_z;
						rk2optdepthed = d_optdepthed_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk2CovdBfd_yy = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * rk2CovdBfu_yx + vector(rk2gtd_yy, i, j, k) * rk2CovdBfu_yy + vector(rk2gtd_yz, i, j, k) * rk2CovdBfu_yz);
							rk2CovdBfd_yx = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * rk2CovdBfu_yx + vector(rk2gtd_xy, i, j, k) * rk2CovdBfu_yy + vector(rk2gtd_xz, i, j, k) * rk2CovdBfu_yz);
							rk2CovdBfd_yz = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * rk2CovdBfu_yx + vector(rk2gtd_yz, i, j, k) * rk2CovdBfu_yy + vector(rk2gtd_zz, i, j, k) * rk2CovdBfu_yz);
						} else {
							rk2CovdBfd_yy = 0.0;
							rk2CovdBfd_yx = 0.0;
							rk2CovdBfd_yz = 0.0;
						}
						rk2optdeptheh = d_optdeptheh_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk2CovdBvf_z = vector(rk2Bfu_x, i, j, k) * rk2Covdvfd_zx + vector(vfd_x, i, j, k) * rk2CovdBfu_zx + vector(rk2Bfu_y, i, j, k) * rk2Covdvfd_zy + vector(vfd_y, i, j, k) * rk2CovdBfu_zy + vector(rk2Bfu_z, i, j, k) * rk2Covdvfd_zz + vector(vfd_z, i, j, k) * rk2CovdBfu_zz;
						} else {
							rk2CovdBvf_z = 0.0;
						}
						rk2optdepthev = d_optdepthev_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk2CovdBvf_y = vector(rk2Bfu_x, i, j, k) * rk2Covdvfd_yx + vector(vfd_x, i, j, k) * rk2CovdBfu_yx + vector(rk2Bfu_y, i, j, k) * rk2Covdvfd_yy + vector(vfd_y, i, j, k) * rk2CovdBfu_yy + vector(rk2Bfu_z, i, j, k) * rk2Covdvfd_yz + vector(vfd_z, i, j, k) * rk2CovdBfu_yz;
							rk2CovdBvf_x = vector(rk2Bfu_x, i, j, k) * rk2Covdvfd_xx + vector(vfd_x, i, j, k) * rk2CovdBfu_xx + vector(rk2Bfu_y, i, j, k) * rk2Covdvfd_xy + vector(vfd_y, i, j, k) * rk2CovdBfu_xy + vector(rk2Bfu_z, i, j, k) * rk2Covdvfd_xz + vector(vfd_z, i, j, k) * rk2CovdBfu_xz;
							rk2CovdBfd_xx = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * rk2CovdBfu_xx + vector(rk2gtd_xy, i, j, k) * rk2CovdBfu_xy + vector(rk2gtd_xz, i, j, k) * rk2CovdBfu_xz);
							rk2CovdBfd_xz = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * rk2CovdBfu_xx + vector(rk2gtd_yz, i, j, k) * rk2CovdBfu_xy + vector(rk2gtd_zz, i, j, k) * rk2CovdBfu_xz);
							rk2CovdBfd_xy = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * rk2CovdBfu_xx + vector(rk2gtd_yy, i, j, k) * rk2CovdBfu_xy + vector(rk2gtd_yz, i, j, k) * rk2CovdBfu_xz);
						} else {
							rk2CovdBvf_y = 0.0;
							rk2CovdBvf_x = 0.0;
							rk2CovdBfd_xx = 0.0;
							rk2CovdBfd_xz = 0.0;
							rk2CovdBfd_xy = 0.0;
						}
						rk2optdepthxd = d_optdepthxd_o0_t0_m0_l0;
						rk2optdepthxh = d_optdepthxh_o0_t0_m0_l0;
						rk2optdepthxv = d_optdepthxv_o0_t0_m0_l0;
						rk2chiav = d_chiav_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk2CovdEfu_zz = d_CovdEfu_zz_o0_t0_m0_l0;
							rk2CovdEfu_zy = d_CovdEfu_zy_o0_t0_m0_l0;
							rk2CovdEfu_zx = d_CovdEfu_zx_o0_t0_m0_l0;
							rk2CovdEfd_zz = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * rk2CovdEfu_zx + vector(rk2gtd_yz, i, j, k) * rk2CovdEfu_zy + vector(rk2gtd_zz, i, j, k) * rk2CovdEfu_zz);
							rk2CovdEfd_zx = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * rk2CovdEfu_zx + vector(rk2gtd_xy, i, j, k) * rk2CovdEfu_zy + vector(rk2gtd_xz, i, j, k) * rk2CovdEfu_zz);
							rk2CovdEfd_zy = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * rk2CovdEfu_zx + vector(rk2gtd_yy, i, j, k) * rk2CovdEfu_zy + vector(rk2gtd_yz, i, j, k) * rk2CovdEfu_zz);
						} else {
							rk2CovdEfu_zz = 0.0;
							rk2CovdEfu_zy = 0.0;
							rk2CovdEfu_zx = 0.0;
							rk2CovdEfd_zz = 0.0;
							rk2CovdEfd_zx = 0.0;
							rk2CovdEfd_zy = 0.0;
						}
						rk2Bvf = (vector(rk2Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk2Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk2Bfu_z, i, j, k) * vector(vfd_z, i, j, k)) / rk2sdetg;
						rk2optdepthah = d_optdepthah_o0_t0_m0_l0;
						rk2optdepthad = d_optdepthad_o0_t0_m0_l0;
						rk2h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						rk2chiah = d_chiah_o0_t0_m0_l0;
						rk2chiad = d_chiad_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk2CovdYef_y = d_CovdYef_y_o0_t0_m0_l0;
						} else {
							rk2CovdYef_y = 0.0;
						}
						rk2optdepthav = d_optdepthav_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk2CovdYef_z = d_CovdYef_z_o0_t0_m0_l0;
							rk2CovdYef_x = d_CovdYef_x_o0_t0_m0_l0;
						} else {
							rk2CovdYef_z = 0.0;
							rk2CovdYef_x = 0.0;
						}
						rk2detgtd = vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) * vector(rk2gtd_xx, i, j, k) + (-vector(rk2gtd_yy, i, j, k) * (vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k))) + (-vector(rk2gtd_zz, i, j, k) * (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k))) + (-(vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k)) * vector(rk2gtd_xx, i, j, k)) + 2.0 * vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yy, i, j, k) - vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k));
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_zz, i, j, k) - vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yz, i, j, k)) + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) - vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_yz, i, j, k) - vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_xz, i, j, k));
						if (calculate_les_terms) {
							rk2CovuBfu_zx = rk2chi_max * (rk2gtu_xz * rk2CovdBfu_xx + rk2gtu_yz * rk2CovdBfu_yx + rk2gtu_zz * rk2CovdBfu_zx);
							rk2CovuBfu_zz = rk2chi_max * (rk2gtu_xz * rk2CovdBfu_xz + rk2gtu_yz * rk2CovdBfu_yz + rk2gtu_zz * rk2CovdBfu_zz);
							rk2CovuBfu_zy = rk2chi_max * (rk2gtu_xz * rk2CovdBfu_xy + rk2gtu_yz * rk2CovdBfu_yy + rk2gtu_zz * rk2CovdBfu_zy);
							rk2Covdvfu_xz = rk2chi_max * (rk2gtu_xz * rk2Covdvfd_xx + rk2gtu_yz * rk2Covdvfd_xy + rk2gtu_zz * rk2Covdvfd_xz);
							rk2CovuYef_z = rk2chi_max * (rk2gtu_xz * rk2CovdYef_x + rk2gtu_yz * rk2CovdYef_y + rk2gtu_zz * rk2CovdYef_z);
							rk2Covdvfu_zz = rk2chi_max * (rk2gtu_xz * rk2Covdvfd_zx + rk2gtu_yz * rk2Covdvfd_zy + rk2gtu_zz * rk2Covdvfd_zz);
							rk2Covdvfu_yz = rk2chi_max * (rk2gtu_xz * rk2Covdvfd_yx + rk2gtu_yz * rk2Covdvfd_yy + rk2gtu_zz * rk2Covdvfd_yz);
							rk2Covuvfu_zz = rk2chi_max * (rk2gtu_xz * rk2Covdvfu_xz + rk2gtu_yz * rk2Covdvfu_yz + rk2gtu_zz * rk2Covdvfu_zz);
						} else {
							rk2CovuBfu_zx = 0.0;
							rk2CovuBfu_zz = 0.0;
							rk2CovuBfu_zy = 0.0;
							rk2Covdvfu_xz = 0.0;
							rk2CovuYef_z = 0.0;
							rk2Covdvfu_zz = 0.0;
							rk2Covdvfu_yz = 0.0;
							rk2Covuvfu_zz = 0.0;
						}
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j, k) + rk2gtu_yz * vector(rk2Sfd_y, i, j, k) + rk2gtu_zz * vector(rk2Sfd_z, i, j, k));
						if (calculate_les_terms) {
							rk2Covurhof_z = rk2chi_max * (rk2gtu_xz * rk2Covdrhof_x + rk2gtu_yz * rk2Covdrhof_y + rk2gtu_zz * rk2Covdrhof_z);
							rk2CovuBvf_z = rk2chi_max * (rk2gtu_xz * rk2CovdBvf_x + rk2gtu_yz * rk2CovdBvf_y + rk2gtu_zz * rk2CovdBvf_z);
						} else {
							rk2Covurhof_z = 0.0;
							rk2CovuBvf_z = 0.0;
						}
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_zz, i, j, k)) + vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xz, i, j, k));
						if (calculate_les_terms) {
							rk2CovuBfu_yz = rk2chi_max * (rk2gtu_xy * rk2CovdBfu_xz + rk2gtu_yy * rk2CovdBfu_yz + rk2gtu_yz * rk2CovdBfu_zz);
							rk2CovuBfu_yy = rk2chi_max * (rk2gtu_xy * rk2CovdBfu_xy + rk2gtu_yy * rk2CovdBfu_yy + rk2gtu_yz * rk2CovdBfu_zy);
							rk2CovuBfu_yx = rk2chi_max * (rk2gtu_xy * rk2CovdBfu_xx + rk2gtu_yy * rk2CovdBfu_yx + rk2gtu_yz * rk2CovdBfu_zx);
							rk2Covdvfu_xy = rk2chi_max * (rk2gtu_xy * rk2Covdvfd_xx + rk2gtu_yy * rk2Covdvfd_xy + rk2gtu_yz * rk2Covdvfd_xz);
							rk2Covdvfu_xx = rk2chi_max * (rk2gtu_xx * rk2Covdvfd_xx + rk2gtu_xy * rk2Covdvfd_xy + rk2gtu_xz * rk2Covdvfd_xz);
							rk2CovdinvsqW_x = (-2.0 * vector(vfd_x, i, j, k) * rk2Covdvfu_xx) + (-2.0 * vector(vfd_y, i, j, k) * rk2Covdvfu_xy) + (-2.0 * vector(vfd_z, i, j, k) * rk2Covdvfu_xz);
							rk2CovuYef_x = rk2chi_max * (rk2gtu_xx * rk2CovdYef_x + rk2gtu_xy * rk2CovdYef_y + rk2gtu_xz * rk2CovdYef_z);
							rk2CovuYef_y = rk2chi_max * (rk2gtu_xy * rk2CovdYef_x + rk2gtu_yy * rk2CovdYef_y + rk2gtu_yz * rk2CovdYef_z);
							rk2Covdvfu_zy = rk2chi_max * (rk2gtu_xy * rk2Covdvfd_zx + rk2gtu_yy * rk2Covdvfd_zy + rk2gtu_yz * rk2Covdvfd_zz);
							rk2Covdvfu_zx = rk2chi_max * (rk2gtu_xx * rk2Covdvfd_zx + rk2gtu_xy * rk2Covdvfd_zy + rk2gtu_xz * rk2Covdvfd_zz);
							rk2CovdinvsqW_z = (-2.0 * vector(vfd_x, i, j, k) * rk2Covdvfu_zx) + (-2.0 * vector(vfd_y, i, j, k) * rk2Covdvfu_zy) + (-2.0 * vector(vfd_z, i, j, k) * rk2Covdvfu_zz);
							rk2Covdvfu_yx = rk2chi_max * (rk2gtu_xx * rk2Covdvfd_yx + rk2gtu_xy * rk2Covdvfd_yy + rk2gtu_xz * rk2Covdvfd_yz);
							rk2Covuvfu_zx = rk2chi_max * (rk2gtu_xz * rk2Covdvfu_xx + rk2gtu_yz * rk2Covdvfu_yx + rk2gtu_zz * rk2Covdvfu_zx);
							rk2Covdvfu_yy = rk2chi_max * (rk2gtu_xy * rk2Covdvfd_yx + rk2gtu_yy * rk2Covdvfd_yy + rk2gtu_yz * rk2Covdvfd_yz);
							rk2CovdinvsqW_y = (-2.0 * vector(vfd_x, i, j, k) * rk2Covdvfu_yx) + (-2.0 * vector(vfd_y, i, j, k) * rk2Covdvfu_yy) + (-2.0 * vector(vfd_z, i, j, k) * rk2Covdvfu_yz);
							rk2CovuinvsqW_z = rk2chi_max * (rk2gtu_xz * rk2CovdinvsqW_x + rk2gtu_yz * rk2CovdinvsqW_y + rk2gtu_zz * rk2CovdinvsqW_z);
							rk2Covuvfu_zy = rk2chi_max * (rk2gtu_xz * rk2Covdvfu_xy + rk2gtu_yz * rk2Covdvfu_yy + rk2gtu_zz * rk2Covdvfu_zy);
							rk2CovuBfu_xz = rk2chi_max * (rk2gtu_xx * rk2CovdBfu_xz + rk2gtu_xy * rk2CovdBfu_yz + rk2gtu_xz * rk2CovdBfu_zz);
							rk2CovuBfu_xy = rk2chi_max * (rk2gtu_xx * rk2CovdBfu_xy + rk2gtu_xy * rk2CovdBfu_yy + rk2gtu_xz * rk2CovdBfu_zy);
							rk2CovuBfu_xx = rk2chi_max * (rk2gtu_xx * rk2CovdBfu_xx + rk2gtu_xy * rk2CovdBfu_yx + rk2gtu_xz * rk2CovdBfu_zx);
						} else {
							rk2CovuBfu_yz = 0.0;
							rk2CovuBfu_yy = 0.0;
							rk2CovuBfu_yx = 0.0;
							rk2Covdvfu_xy = 0.0;
							rk2Covdvfu_xx = 0.0;
							rk2CovdinvsqW_x = 0.0;
							rk2CovuYef_x = 0.0;
							rk2CovuYef_y = 0.0;
							rk2Covdvfu_zy = 0.0;
							rk2Covdvfu_zx = 0.0;
							rk2CovdinvsqW_z = 0.0;
							rk2Covdvfu_yx = 0.0;
							rk2Covuvfu_zx = 0.0;
							rk2Covdvfu_yy = 0.0;
							rk2CovdinvsqW_y = 0.0;
							rk2CovuinvsqW_z = 0.0;
							rk2Covuvfu_zy = 0.0;
							rk2CovuBfu_xz = 0.0;
							rk2CovuBfu_xy = 0.0;
							rk2CovuBfu_xx = 0.0;
						}
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j, k) + rk2gtu_yy * vector(rk2Sfd_y, i, j, k) + rk2gtu_yz * vector(rk2Sfd_z, i, j, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j, k) + rk2gtu_xy * vector(rk2Sfd_y, i, j, k) + rk2gtu_xz * vector(rk2Sfd_z, i, j, k));
						if (calculate_les_terms) {
							rk2Covurhof_x = rk2chi_max * (rk2gtu_xx * rk2Covdrhof_x + rk2gtu_xy * rk2Covdrhof_y + rk2gtu_xz * rk2Covdrhof_z);
							rk2Covurhof_y = rk2chi_max * (rk2gtu_xy * rk2Covdrhof_x + rk2gtu_yy * rk2Covdrhof_y + rk2gtu_yz * rk2Covdrhof_z);
							rk2Covuvfu_yz = rk2chi_max * (rk2gtu_xy * rk2Covdvfu_xz + rk2gtu_yy * rk2Covdvfu_yz + rk2gtu_yz * rk2Covdvfu_zz);
							rk2Covuvfu_yy = rk2chi_max * (rk2gtu_xy * rk2Covdvfu_xy + rk2gtu_yy * rk2Covdvfu_yy + rk2gtu_yz * rk2Covdvfu_zy);
							rk2Covuvfu_yx = rk2chi_max * (rk2gtu_xy * rk2Covdvfu_xx + rk2gtu_yy * rk2Covdvfu_yx + rk2gtu_yz * rk2Covdvfu_zx);
							rk2Covuvfu_xz = rk2chi_max * (rk2gtu_xx * rk2Covdvfu_xz + rk2gtu_xy * rk2Covdvfu_yz + rk2gtu_xz * rk2Covdvfu_zz);
							rk2Covuvfu_xy = rk2chi_max * (rk2gtu_xx * rk2Covdvfu_xy + rk2gtu_xy * rk2Covdvfu_yy + rk2gtu_xz * rk2Covdvfu_zy);
							rk2Covuvfu_xx = rk2chi_max * (rk2gtu_xx * rk2Covdvfu_xx + rk2gtu_xy * rk2Covdvfu_yx + rk2gtu_xz * rk2Covdvfu_zx);
						} else {
							rk2Covurhof_x = 0.0;
							rk2Covurhof_y = 0.0;
							rk2Covuvfu_yz = 0.0;
							rk2Covuvfu_yy = 0.0;
							rk2Covuvfu_yx = 0.0;
							rk2Covuvfu_xz = 0.0;
							rk2Covuvfu_xy = 0.0;
							rk2Covuvfu_xx = 0.0;
						}
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j, k) + rk2gtu_yy * vector(vfd_y, i, j, k) + rk2gtu_yz * vector(vfd_z, i, j, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j, k) + rk2gtu_xy * vector(vfd_y, i, j, k) + rk2gtu_xz * vector(vfd_z, i, j, k));
						if (calculate_les_terms) {
							rk2CovuBvf_x = rk2chi_max * (rk2gtu_xx * rk2CovdBvf_x + rk2gtu_xy * rk2CovdBvf_y + rk2gtu_xz * rk2CovdBvf_z);
							rk2CovuBvf_y = rk2chi_max * (rk2gtu_xy * rk2CovdBvf_x + rk2gtu_yy * rk2CovdBvf_y + rk2gtu_yz * rk2CovdBvf_z);
							rk2CovuinvsqW_y = rk2chi_max * (rk2gtu_xy * rk2CovdinvsqW_x + rk2gtu_yy * rk2CovdinvsqW_y + rk2gtu_yz * rk2CovdinvsqW_z);
							rk2CovuinvsqW_x = rk2chi_max * (rk2gtu_xx * rk2CovdinvsqW_x + rk2gtu_xy * rk2CovdinvsqW_y + rk2gtu_xz * rk2CovdinvsqW_z);
							rk2Covddpfdye_z = d_Covddpfdye_z_o0_t0_m0_l0;
							rk2Covddpfdye_y = d_Covddpfdye_y_o0_t0_m0_l0;
							rk2Covddpfdye_x = d_Covddpfdye_x_o0_t0_m0_l0;
							rk2CovdEfu_xz = d_CovdEfu_xz_o0_t0_m0_l0;
							rk2CovdEfu_xy = d_CovdEfu_xy_o0_t0_m0_l0;
							rk2CovdEfu_xx = d_CovdEfu_xx_o0_t0_m0_l0;
							rk2CovdEfd_xx = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * rk2CovdEfu_xx + vector(rk2gtd_xy, i, j, k) * rk2CovdEfu_xy + vector(rk2gtd_xz, i, j, k) * rk2CovdEfu_xz);
							rk2CovdEfd_xy = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * rk2CovdEfu_xx + vector(rk2gtd_yy, i, j, k) * rk2CovdEfu_xy + vector(rk2gtd_yz, i, j, k) * rk2CovdEfu_xz);
							rk2CovdEfd_xz = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * rk2CovdEfu_xx + vector(rk2gtd_yz, i, j, k) * rk2CovdEfu_xy + vector(rk2gtd_zz, i, j, k) * rk2CovdEfu_xz);
						} else {
							rk2CovuBvf_x = 0.0;
							rk2CovuBvf_y = 0.0;
							rk2CovuinvsqW_y = 0.0;
							rk2CovuinvsqW_x = 0.0;
							rk2Covddpfdye_z = 0.0;
							rk2Covddpfdye_y = 0.0;
							rk2Covddpfdye_x = 0.0;
							rk2CovdEfu_xz = 0.0;
							rk2CovdEfu_xy = 0.0;
							rk2CovdEfu_xx = 0.0;
							rk2CovdEfd_xx = 0.0;
							rk2CovdEfd_xy = 0.0;
							rk2CovdEfd_xz = 0.0;
						}
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j, k) + rk2gtu_yz * vector(vfd_y, i, j, k) + rk2gtu_zz * vector(vfd_z, i, j, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j, k) + rk2vfu_y * vector(vfd_y, i, j, k) + rk2vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						if (calculate_les_terms) {
							rk2PhiAh = rk2sqW * (vector(pf, i, j, k) * vector(dpfdeps, i, j, k) + (vector(rhof, i, j, k) * vector(rhof, i, j, k)) * vector(dpfdrho, i, j, k));
						} else {
							rk2PhiAh = 0.0;
						}
						rk2Epf = rk2h * rk2sqW;
						rk2Thetaf = rk2Epf + rk2sqBf / (rk2sdetg * rk2sdetg);
						if (calculate_les_terms) {
							rk2Covdepsf_z = d_Covdepsf_z_o0_t0_m0_l0;
							rk2Covdhf_z = (1.0 + vector(epsf, i, j, k)) * rk2Covdrhof_z + vector(rhof, i, j, k) * rk2Covdepsf_z + rk2Covdpf_z;
							rk2CovdEpf_z = rk2sqW * rk2Covdhf_z + (-rk2h * (rk2sqW * rk2sqW) * rk2CovdinvsqW_z);
							rk2CovdThetaf_z = rk2CovdEpf_z + 2.0 * (rk2Bfd_x * rk2CovdBfu_zx) / rk2sdetg + 2.0 * (rk2Bfd_y * rk2CovdBfu_zy) / rk2sdetg + 2.0 * (rk2Bfd_z * rk2CovdBfu_zz) / rk2sdetg;
							rk2Covdepsf_y = d_Covdepsf_y_o0_t0_m0_l0;
							rk2Covdhf_y = (1.0 + vector(epsf, i, j, k)) * rk2Covdrhof_y + vector(rhof, i, j, k) * rk2Covdepsf_y + rk2Covdpf_y;
							rk2CovdEpf_y = rk2sqW * rk2Covdhf_y + (-rk2h * (rk2sqW * rk2sqW) * rk2CovdinvsqW_y);
							rk2CovdThetaf_y = rk2CovdEpf_y + 2.0 * (rk2Bfd_x * rk2CovdBfu_yx) / rk2sdetg + 2.0 * (rk2Bfd_y * rk2CovdBfu_yy) / rk2sdetg + 2.0 * (rk2Bfd_z * rk2CovdBfu_yz) / rk2sdetg;
							rk2Covdepsf_x = d_Covdepsf_x_o0_t0_m0_l0;
							rk2Covuepsf_z = rk2chi_max * (rk2gtu_xz * rk2Covdepsf_x + rk2gtu_yz * rk2Covdepsf_y + rk2gtu_zz * rk2Covdepsf_z);
							rk2Covuepsf_y = rk2chi_max * (rk2gtu_xy * rk2Covdepsf_x + rk2gtu_yy * rk2Covdepsf_y + rk2gtu_yz * rk2Covdepsf_z);
							rk2Covdhf_x = (1.0 + vector(epsf, i, j, k)) * rk2Covdrhof_x + vector(rhof, i, j, k) * rk2Covdepsf_x + rk2Covdpf_x;
							rk2CovdEpf_x = rk2sqW * rk2Covdhf_x + (-rk2h * (rk2sqW * rk2sqW) * rk2CovdinvsqW_x);
							rk2CovdThetaf_x = rk2CovdEpf_x + 2.0 * (rk2Bfd_x * rk2CovdBfu_xx) / rk2sdetg + 2.0 * (rk2Bfd_y * rk2CovdBfu_xy) / rk2sdetg + 2.0 * (rk2Bfd_z * rk2CovdBfu_xz) / rk2sdetg;
							rk2CovuThetaf_z = rk2chi_max * (rk2gtu_xz * rk2CovdThetaf_x + rk2gtu_yz * rk2CovdThetaf_y + rk2gtu_zz * rk2CovdThetaf_z);
							rk2CovuThetaf_x = rk2chi_max * (rk2gtu_xx * rk2CovdThetaf_x + rk2gtu_xy * rk2CovdThetaf_y + rk2gtu_xz * rk2CovdThetaf_z);
							rk2CovuThetaf_y = rk2chi_max * (rk2gtu_xy * rk2CovdThetaf_x + rk2gtu_yy * rk2CovdThetaf_y + rk2gtu_yz * rk2CovdThetaf_z);
							rk2Phivh_z = 2.0 * (rk2CovuBvf_x * rk2CovdBfu_xz + rk2CovuBvf_y * rk2CovdBfu_yz + rk2CovuBvf_z * rk2CovdBfu_zz + (-rk2CovuThetaf_x * rk2Covdvfu_xz) + (-rk2CovuThetaf_y * rk2Covdvfu_yz) + (-rk2CovuThetaf_z * rk2Covdvfu_zz) + (vector(rk2Bfu_z, i, j, k) * (rk2Thetaf * (rk2CovuBfu_xx * rk2Covdvfd_xx + rk2CovuBfu_yx * rk2Covdvfd_yx + rk2CovuBfu_zx * rk2Covdvfd_zx + rk2CovuBfu_xy * rk2Covdvfd_xy + rk2CovuBfu_yy * rk2Covdvfd_yy + rk2CovuBfu_zy * rk2Covdvfd_zy + rk2CovuBfu_xz * rk2Covdvfd_xz + rk2CovuBfu_yz * rk2Covdvfd_yz + rk2CovuBfu_zz * rk2Covdvfd_zz) + (rk2Bfd_x * rk2CovdBfu_xx * rk2CovuBvf_x) / rk2sdetg + (rk2Bfd_y * rk2CovdBfu_xy * rk2CovuBvf_x) / rk2sdetg + (rk2Bfd_z * rk2CovdBfu_xz * rk2CovuBvf_x) / rk2sdetg + (rk2Bfd_x * rk2CovdBfu_yx * rk2CovuBvf_y) / rk2sdetg + (rk2Bfd_y * rk2CovdBfu_yy * rk2CovuBvf_y) / rk2sdetg + (rk2Bfd_z * rk2CovdBfu_yz * rk2CovuBvf_y) / rk2sdetg + (rk2Bfd_x * rk2CovdBfu_zx * rk2CovuBvf_z) / rk2sdetg + (rk2Bfd_y * rk2CovdBfu_zy * rk2CovuBvf_z) / rk2sdetg + (rk2Bfd_z * rk2CovdBfu_zz * rk2CovuBvf_z) / rk2sdetg + (-(vector(rk2Bfu_x, i, j, k) * rk2Covdvfd_xx * rk2CovuThetaf_x) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2Covdvfd_xy * rk2CovuThetaf_x) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfd_xz * rk2CovuThetaf_x) / rk2sdetg) + (-(vector(rk2Bfu_x, i, j, k) * rk2Covdvfd_yx * rk2CovuThetaf_y) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2Covdvfd_yy * rk2CovuThetaf_y) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfd_yz * rk2CovuThetaf_y) / rk2sdetg) + (-(vector(rk2Bfu_x, i, j, k) * rk2Covdvfd_zx * rk2CovuThetaf_z) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2Covdvfd_zy * rk2CovuThetaf_z) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfd_zz * rk2CovuThetaf_z) / rk2sdetg))) / (rk2sdetg * rk2Epf)) / rk2Thetaf;
							rk2Phivh_x = 2.0 * (rk2CovuBvf_x * rk2CovdBfu_xx + rk2CovuBvf_y * rk2CovdBfu_yx + rk2CovuBvf_z * rk2CovdBfu_zx + (-rk2CovuThetaf_x * rk2Covdvfu_xx) + (-rk2CovuThetaf_y * rk2Covdvfu_yx) + (-rk2CovuThetaf_z * rk2Covdvfu_zx) + (vector(rk2Bfu_x, i, j, k) * (rk2Thetaf * (rk2CovuBfu_xx * rk2Covdvfd_xx + rk2CovuBfu_yx * rk2Covdvfd_yx + rk2CovuBfu_zx * rk2Covdvfd_zx + rk2CovuBfu_xy * rk2Covdvfd_xy + rk2CovuBfu_yy * rk2Covdvfd_yy + rk2CovuBfu_zy * rk2Covdvfd_zy + rk2CovuBfu_xz * rk2Covdvfd_xz + rk2CovuBfu_yz * rk2Covdvfd_yz + rk2CovuBfu_zz * rk2Covdvfd_zz) + (rk2Bfd_x * rk2CovdBfu_xx * rk2CovuBvf_x) / rk2sdetg + (rk2Bfd_y * rk2CovdBfu_xy * rk2CovuBvf_x) / rk2sdetg + (rk2Bfd_z * rk2CovdBfu_xz * rk2CovuBvf_x) / rk2sdetg + (rk2Bfd_x * rk2CovdBfu_yx * rk2CovuBvf_y) / rk2sdetg + (rk2Bfd_y * rk2CovdBfu_yy * rk2CovuBvf_y) / rk2sdetg + (rk2Bfd_z * rk2CovdBfu_yz * rk2CovuBvf_y) / rk2sdetg + (rk2Bfd_x * rk2CovdBfu_zx * rk2CovuBvf_z) / rk2sdetg + (rk2Bfd_y * rk2CovdBfu_zy * rk2CovuBvf_z) / rk2sdetg + (rk2Bfd_z * rk2CovdBfu_zz * rk2CovuBvf_z) / rk2sdetg + (-(vector(rk2Bfu_x, i, j, k) * rk2Covdvfd_xx * rk2CovuThetaf_x) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2Covdvfd_xy * rk2CovuThetaf_x) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfd_xz * rk2CovuThetaf_x) / rk2sdetg) + (-(vector(rk2Bfu_x, i, j, k) * rk2Covdvfd_yx * rk2CovuThetaf_y) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2Covdvfd_yy * rk2CovuThetaf_y) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfd_yz * rk2CovuThetaf_y) / rk2sdetg) + (-(vector(rk2Bfu_x, i, j, k) * rk2Covdvfd_zx * rk2CovuThetaf_z) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2Covdvfd_zy * rk2CovuThetaf_z) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfd_zz * rk2CovuThetaf_z) / rk2sdetg))) / (rk2sdetg * rk2Epf)) / rk2Thetaf;
							rk2Phivh_y = 2.0 * (rk2CovuBvf_x * rk2CovdBfu_xy + rk2CovuBvf_y * rk2CovdBfu_yy + rk2CovuBvf_z * rk2CovdBfu_zy + (-rk2CovuThetaf_x * rk2Covdvfu_xy) + (-rk2CovuThetaf_y * rk2Covdvfu_yy) + (-rk2CovuThetaf_z * rk2Covdvfu_zy) + (vector(rk2Bfu_y, i, j, k) * (rk2Thetaf * (rk2CovuBfu_xx * rk2Covdvfd_xx + rk2CovuBfu_yx * rk2Covdvfd_yx + rk2CovuBfu_zx * rk2Covdvfd_zx + rk2CovuBfu_xy * rk2Covdvfd_xy + rk2CovuBfu_yy * rk2Covdvfd_yy + rk2CovuBfu_zy * rk2Covdvfd_zy + rk2CovuBfu_xz * rk2Covdvfd_xz + rk2CovuBfu_yz * rk2Covdvfd_yz + rk2CovuBfu_zz * rk2Covdvfd_zz) + (rk2Bfd_x * rk2CovdBfu_xx * rk2CovuBvf_x) / rk2sdetg + (rk2Bfd_y * rk2CovdBfu_xy * rk2CovuBvf_x) / rk2sdetg + (rk2Bfd_z * rk2CovdBfu_xz * rk2CovuBvf_x) / rk2sdetg + (rk2Bfd_x * rk2CovdBfu_yx * rk2CovuBvf_y) / rk2sdetg + (rk2Bfd_y * rk2CovdBfu_yy * rk2CovuBvf_y) / rk2sdetg + (rk2Bfd_z * rk2CovdBfu_yz * rk2CovuBvf_y) / rk2sdetg + (rk2Bfd_x * rk2CovdBfu_zx * rk2CovuBvf_z) / rk2sdetg + (rk2Bfd_y * rk2CovdBfu_zy * rk2CovuBvf_z) / rk2sdetg + (rk2Bfd_z * rk2CovdBfu_zz * rk2CovuBvf_z) / rk2sdetg + (-(vector(rk2Bfu_x, i, j, k) * rk2Covdvfd_xx * rk2CovuThetaf_x) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2Covdvfd_xy * rk2CovuThetaf_x) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfd_xz * rk2CovuThetaf_x) / rk2sdetg) + (-(vector(rk2Bfu_x, i, j, k) * rk2Covdvfd_yx * rk2CovuThetaf_y) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2Covdvfd_yy * rk2CovuThetaf_y) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfd_yz * rk2CovuThetaf_y) / rk2sdetg) + (-(vector(rk2Bfu_x, i, j, k) * rk2Covdvfd_zx * rk2CovuThetaf_z) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2Covdvfd_zy * rk2CovuThetaf_z) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfd_zz * rk2CovuThetaf_z) / rk2sdetg))) / (rk2sdetg * rk2Epf)) / rk2Thetaf;
							rk2PhiMh_yz = 2.0 * (rk2Thetaf * ((-rk2Covdvfu_xz * rk2CovuBfu_xy) + rk2Covdvfu_xy * rk2CovuBfu_xz + (-rk2Covdvfu_yz * rk2CovuBfu_yy) + rk2Covdvfu_yy * rk2CovuBfu_yz + (-rk2Covdvfu_zz * rk2CovuBfu_zy) + rk2Covdvfu_zy * rk2CovuBfu_zz) + (vector(rk2Bfu_z, i, j, k) * rk2CovuBfu_xy * rk2CovdBvf_x) / rk2sdetg + (vector(rk2Bfu_z, i, j, k) * rk2CovuBfu_yy * rk2CovdBvf_y) / rk2sdetg + (vector(rk2Bfu_z, i, j, k) * rk2CovuBfu_zy * rk2CovdBvf_z) / rk2sdetg + (-(vector(rk2Bfu_y, i, j, k) * rk2CovuBfu_xz * rk2CovdBvf_x) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2CovuBfu_yz * rk2CovdBvf_y) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2CovuBfu_zz * rk2CovdBvf_z) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfu_xy * rk2CovuThetaf_x) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfu_yy * rk2CovuThetaf_y) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfu_zy * rk2CovuThetaf_z) / rk2sdetg) + (vector(rk2Bfu_y, i, j, k) * rk2Covdvfu_xz * rk2CovuThetaf_x) / rk2sdetg + (vector(rk2Bfu_y, i, j, k) * rk2Covdvfu_yz * rk2CovuThetaf_y) / rk2sdetg + (vector(rk2Bfu_y, i, j, k) * rk2Covdvfu_zz * rk2CovuThetaf_z) / rk2sdetg) / rk2Thetaf;
							rk2PhiMh_xz = 2.0 * (rk2Thetaf * ((-rk2Covdvfu_xz * rk2CovuBfu_xx) + rk2Covdvfu_xx * rk2CovuBfu_xz + (-rk2Covdvfu_yz * rk2CovuBfu_yx) + rk2Covdvfu_yx * rk2CovuBfu_yz + (-rk2Covdvfu_zz * rk2CovuBfu_zx) + rk2Covdvfu_zx * rk2CovuBfu_zz) + (vector(rk2Bfu_z, i, j, k) * rk2CovuBfu_xx * rk2CovdBvf_x) / rk2sdetg + (vector(rk2Bfu_z, i, j, k) * rk2CovuBfu_yx * rk2CovdBvf_y) / rk2sdetg + (vector(rk2Bfu_z, i, j, k) * rk2CovuBfu_zx * rk2CovdBvf_z) / rk2sdetg + (-(vector(rk2Bfu_x, i, j, k) * rk2CovuBfu_xz * rk2CovdBvf_x) / rk2sdetg) + (-(vector(rk2Bfu_x, i, j, k) * rk2CovuBfu_yz * rk2CovdBvf_y) / rk2sdetg) + (-(vector(rk2Bfu_x, i, j, k) * rk2CovuBfu_zz * rk2CovdBvf_z) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfu_xx * rk2CovuThetaf_x) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfu_yx * rk2CovuThetaf_y) / rk2sdetg) + (-(vector(rk2Bfu_z, i, j, k) * rk2Covdvfu_zx * rk2CovuThetaf_z) / rk2sdetg) + (vector(rk2Bfu_x, i, j, k) * rk2Covdvfu_xz * rk2CovuThetaf_x) / rk2sdetg + (vector(rk2Bfu_x, i, j, k) * rk2Covdvfu_yz * rk2CovuThetaf_y) / rk2sdetg + (vector(rk2Bfu_x, i, j, k) * rk2Covdvfu_zz * rk2CovuThetaf_z) / rk2sdetg) / rk2Thetaf;
							rk2PhiMh_xy = 2.0 * (rk2Thetaf * ((-rk2Covdvfu_xy * rk2CovuBfu_xx) + rk2Covdvfu_xx * rk2CovuBfu_xy + (-rk2Covdvfu_yy * rk2CovuBfu_yx) + rk2Covdvfu_yx * rk2CovuBfu_yy + (-rk2Covdvfu_zy * rk2CovuBfu_zx) + rk2Covdvfu_zx * rk2CovuBfu_zy) + (vector(rk2Bfu_y, i, j, k) * rk2CovuBfu_xx * rk2CovdBvf_x) / rk2sdetg + (vector(rk2Bfu_y, i, j, k) * rk2CovuBfu_yx * rk2CovdBvf_y) / rk2sdetg + (vector(rk2Bfu_y, i, j, k) * rk2CovuBfu_zx * rk2CovdBvf_z) / rk2sdetg + (-(vector(rk2Bfu_x, i, j, k) * rk2CovuBfu_xy * rk2CovdBvf_x) / rk2sdetg) + (-(vector(rk2Bfu_x, i, j, k) * rk2CovuBfu_yy * rk2CovdBvf_y) / rk2sdetg) + (-(vector(rk2Bfu_x, i, j, k) * rk2CovuBfu_zy * rk2CovdBvf_z) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2Covdvfu_xx * rk2CovuThetaf_x) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2Covdvfu_yx * rk2CovuThetaf_y) / rk2sdetg) + (-(vector(rk2Bfu_y, i, j, k) * rk2Covdvfu_zx * rk2CovuThetaf_z) / rk2sdetg) + (vector(rk2Bfu_x, i, j, k) * rk2Covdvfu_xy * rk2CovuThetaf_x) / rk2sdetg + (vector(rk2Bfu_x, i, j, k) * rk2Covdvfu_yy * rk2CovuThetaf_y) / rk2sdetg + (vector(rk2Bfu_x, i, j, k) * rk2Covdvfu_zy * rk2CovuThetaf_z) / rk2sdetg) / rk2Thetaf;
							rk2Covuepsf_x = rk2chi_max * (rk2gtu_xx * rk2Covdepsf_x + rk2gtu_xy * rk2Covdepsf_y + rk2gtu_xz * rk2Covdepsf_z);
						} else {
							rk2Covdepsf_z = 0.0;
							rk2Covdhf_z = 0.0;
							rk2CovdEpf_z = 0.0;
							rk2CovdThetaf_z = 0.0;
							rk2Covdepsf_y = 0.0;
							rk2Covdhf_y = 0.0;
							rk2CovdEpf_y = 0.0;
							rk2CovdThetaf_y = 0.0;
							rk2Covdepsf_x = 0.0;
							rk2Covuepsf_z = 0.0;
							rk2Covuepsf_y = 0.0;
							rk2Covdhf_x = 0.0;
							rk2CovdEpf_x = 0.0;
							rk2CovdThetaf_x = 0.0;
							rk2CovuThetaf_z = 0.0;
							rk2CovuThetaf_x = 0.0;
							rk2CovuThetaf_y = 0.0;
							rk2Phivh_z = 0.0;
							rk2Phivh_x = 0.0;
							rk2Phivh_y = 0.0;
							rk2PhiMh_yz = 0.0;
							rk2PhiMh_xz = 0.0;
							rk2PhiMh_xy = 0.0;
							rk2Covuepsf_x = 0.0;
						}
						rk2faceta = 0.04166666666666667 * rk2inv_chi * (dx[0] * dx[0]) * Smooth(threshold_Dles_min, threshold_Dles_max, vector(rk2Df, i, j, k), dx, simPlat_dt, ilast, jlast);
						if (calculate_les_terms) {
							rk2CovdEfu_yz = d_CovdEfu_yz_o0_t0_m0_l0;
							rk2CovuEfu_zz = rk2chi_max * (rk2gtu_xz * rk2CovdEfu_xz + rk2gtu_yz * rk2CovdEfu_yz + rk2gtu_zz * rk2CovdEfu_zz);
							rk2CovuEfu_xz = rk2chi_max * (rk2gtu_xx * rk2CovdEfu_xz + rk2gtu_xy * rk2CovdEfu_yz + rk2gtu_xz * rk2CovdEfu_zz);
							rk2CovdEfu_yy = d_CovdEfu_yy_o0_t0_m0_l0;
							rk2CovuEfu_zy = rk2chi_max * (rk2gtu_xz * rk2CovdEfu_xy + rk2gtu_yz * rk2CovdEfu_yy + rk2gtu_zz * rk2CovdEfu_zy);
							rk2CovuEfu_xy = rk2chi_max * (rk2gtu_xx * rk2CovdEfu_xy + rk2gtu_xy * rk2CovdEfu_yy + rk2gtu_xz * rk2CovdEfu_zy);
							rk2CovdEfu_yx = d_CovdEfu_yx_o0_t0_m0_l0;
							rk2CovdEfd_yy = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * rk2CovdEfu_yx + vector(rk2gtd_yy, i, j, k) * rk2CovdEfu_yy + vector(rk2gtd_yz, i, j, k) * rk2CovdEfu_yz);
							rk2CovdEfd_yz = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * rk2CovdEfu_yx + vector(rk2gtd_yz, i, j, k) * rk2CovdEfu_yy + vector(rk2gtd_zz, i, j, k) * rk2CovdEfu_yz);
							rk2CovdEfd_yx = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * rk2CovdEfu_yx + vector(rk2gtd_xy, i, j, k) * rk2CovdEfu_yy + vector(rk2gtd_xz, i, j, k) * rk2CovdEfu_yz);
							rk2CovuEfu_zx = rk2chi_max * (rk2gtu_xz * rk2CovdEfu_xx + rk2gtu_yz * rk2CovdEfu_yx + rk2gtu_zz * rk2CovdEfu_zx);
							rk2CovuEfu_xx = rk2chi_max * (rk2gtu_xx * rk2CovdEfu_xx + rk2gtu_xy * rk2CovdEfu_yx + rk2gtu_xz * rk2CovdEfu_zx);
							rk2CovuEfu_yx = rk2chi_max * (rk2gtu_xy * rk2CovdEfu_xx + rk2gtu_yy * rk2CovdEfu_yx + rk2gtu_yz * rk2CovdEfu_zx);
							rk2CovuEfu_yz = rk2chi_max * (rk2gtu_xy * rk2CovdEfu_xz + rk2gtu_yy * rk2CovdEfu_yz + rk2gtu_yz * rk2CovdEfu_zz);
							rk2CovuEfu_yy = rk2chi_max * (rk2gtu_xy * rk2CovdEfu_xy + rk2gtu_yy * rk2CovdEfu_yy + rk2gtu_yz * rk2CovdEfu_zy);
							rk2PhiThetah = (rk2Thetaf * (rk2CovdBfd_xx * rk2CovuBfu_xx + rk2CovdBfd_yx * rk2CovuBfu_yx + rk2CovdBfd_zx * rk2CovuBfu_zx + rk2CovdBfd_xy * rk2CovuBfu_xy + rk2CovdBfd_yy * rk2CovuBfu_yy + rk2CovdBfd_zy * rk2CovuBfu_zy + rk2CovdBfd_xz * rk2CovuBfu_xz + rk2CovdBfd_yz * rk2CovuBfu_yz + rk2CovdBfd_zz * rk2CovuBfu_zz + (-rk2CovdEfd_xx * rk2CovuEfu_xx) + (-rk2CovdEfd_yx * rk2CovuEfu_yx) + (-rk2CovdEfd_zx * rk2CovuEfu_zx) + (-rk2CovdEfd_xy * rk2CovuEfu_xy) + (-rk2CovdEfd_yy * rk2CovuEfu_yy) + (-rk2CovdEfd_zy * rk2CovuEfu_zy) + (-rk2CovdEfd_xz * rk2CovuEfu_xz) + (-rk2CovdEfd_yz * rk2CovuEfu_yz) + (-rk2CovdEfd_zz * rk2CovuEfu_zz) + (-0.5 * (((-vector(vfd_y, i, j, k) * rk2Bfd_x) + vector(vfd_x, i, j, k) * rk2Bfd_y) * rk2PhiMh_xy + (vector(vfd_x, i, j, k) * rk2Bfd_z - vector(vfd_z, i, j, k) * rk2Bfd_x) * rk2PhiMh_xz + (-(vector(vfd_y, i, j, k) * rk2Bfd_x - vector(vfd_x, i, j, k) * rk2Bfd_y) * rk2PhiMh_xy) + ((-vector(vfd_z, i, j, k) * rk2Bfd_y) + vector(vfd_y, i, j, k) * rk2Bfd_z) * rk2PhiMh_yz + (-(vector(vfd_z, i, j, k) * rk2Bfd_x - vector(vfd_x, i, j, k) * rk2Bfd_z) * rk2PhiMh_xz) + (-(vector(vfd_z, i, j, k) * rk2Bfd_y - vector(vfd_y, i, j, k) * rk2Bfd_z) * rk2PhiMh_yz)) / rk2sdetg))) / (rk2Thetaf - rk2sqEf);
							rk2HPres = (rk2Epf * rk2sqW * (rk2Thetaf - rk2sqEf) * (vector(rhof, i, j, k) * (rk2Covddpfdrho_x * rk2Covurhof_x + rk2Covddpfdrho_y * rk2Covurhof_y + rk2Covddpfdrho_z * rk2Covurhof_z + rk2Covddpfdeps_x * rk2Covuepsf_x + rk2Covddpfdeps_y * rk2Covuepsf_y + rk2Covddpfdeps_z * rk2Covuepsf_z) + (-2.0 * vector(dpfdeps, i, j, k) * (rk2Covurhof_x * rk2Covdepsf_x + rk2Covurhof_y * rk2Covdepsf_y + rk2Covurhof_z * rk2Covdepsf_z)) + (-(rk2Epf * vector(dpfdeps, i, j, k) - rk2PhiAh) * (0.25 * rk2sqW * (rk2CovdinvsqW_x * rk2CovuinvsqW_x + rk2CovdinvsqW_y * rk2CovuinvsqW_y + rk2CovdinvsqW_z * rk2CovuinvsqW_z) + (rk2CovdinvsqW_x * rk2Covurhof_x + rk2CovdinvsqW_y * rk2Covurhof_y + rk2CovdinvsqW_z * rk2Covurhof_z) / vector(rhof, i, j, k))) + (-2.0 * (vector(dpfdeps, i, j, k) * (rk2CovdBfd_xx * rk2CovuBfu_xx + rk2CovdBfd_yx * rk2CovuBfu_yx + rk2CovdBfd_zx * rk2CovuBfu_zx + rk2CovdBfd_xy * rk2CovuBfu_xy + rk2CovdBfd_yy * rk2CovuBfu_yy + rk2CovdBfd_zy * rk2CovuBfu_zy + rk2CovdBfd_xz * rk2CovuBfu_xz + rk2CovdBfd_yz * rk2CovuBfu_yz + rk2CovdBfd_zz * rk2CovuBfu_zz + (-(rk2sqW * rk2sqW) * (rk2CovuinvsqW_x * rk2Covdhf_x + rk2CovuinvsqW_y * rk2Covdhf_y + rk2CovuinvsqW_z * rk2Covdhf_z)))) / rk2sqW) + (-(rk2Epf * vector(dpfdeps, i, j, k) + rk2PhiAh) * (vector(vfd_x, i, j, k) * rk2Phivh_x + vector(vfd_y, i, j, k) * rk2Phivh_y + vector(vfd_z, i, j, k) * rk2Phivh_z + rk2Covdvfd_xx * rk2Covuvfu_xx + rk2Covdvfd_xy * rk2Covuvfu_xy + rk2Covdvfd_xz * rk2Covuvfu_xz + rk2Covdvfd_yx * rk2Covuvfu_yx + rk2Covdvfd_yy * rk2Covuvfu_yy + rk2Covdvfd_yz * rk2Covuvfu_yz + rk2Covdvfd_zx * rk2Covuvfu_zx + rk2Covdvfd_zy * rk2Covuvfu_zy + rk2Covdvfd_zz * rk2Covuvfu_zz + rk2sqW * (rk2CovdinvsqW_x * rk2CovuinvsqW_x + rk2CovdinvsqW_y * rk2CovuinvsqW_y + rk2CovdinvsqW_z * rk2CovuinvsqW_z))) + (rk2PhiThetah * ((rk2Epf * vector(dpfdeps, i, j, k) + rk2PhiAh) * (rk2Thetaf - rk2sqEf) - (rk2PhiAh * rk2Thetaf) / rk2sqW)) / (rk2Epf * rk2Thetaf))) / ((rk2Epf * vector(rhof, i, j, k) - rk2PhiAh) * (rk2Thetaf - rk2sqEf) * rk2sqW + rk2PhiAh * rk2Thetaf) + rk2Covddpfdye_x * rk2CovuYef_x + rk2Covddpfdye_y * rk2CovuYef_y + rk2Covddpfdye_z * rk2CovuYef_z - 2.0 * vector(dpfdye, i, j, k) / vector(rk2Df, i, j, k) * (rk2CovdDf_x * rk2CovuYef_x + rk2CovdDf_y * rk2CovuYef_y + rk2CovdDf_z * rk2CovuYef_z);
							rk2HTheta = rk2PhiThetah + (rk2Thetaf * rk2HPres) / (rk2Thetaf - rk2sqEf);
							rk2Hv_z = rk2Phivh_z + (-((rk2vfu_z + (rk2Bvf * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2Epf)) * rk2HTheta) / rk2Thetaf);
							rk2HN_z = 2.0 * rk2CovdDf_x * rk2Covuvfu_xz + 2.0 * rk2CovdDf_y * rk2Covuvfu_yz + 2.0 * rk2CovdDf_z * rk2Covuvfu_zz + (vector(rk2Df, i, j, k) * rk2Hv_z) / rk2sdetg;
							rk2HNe_z = 2.0 * rk2CovdDYf_x * rk2Covuvfu_xz + 2.0 * rk2CovdDYf_y * rk2Covuvfu_yz + 2.0 * rk2CovdDYf_z * rk2Covuvfu_zz + (vector(rk2DYf, i, j, k) * rk2Hv_z) / rk2sdetg;
							rk2Hv_y = rk2Phivh_y + (-((rk2vfu_y + (rk2Bvf * vector(rk2Bfu_y, i, j, k)) / (rk2sdetg * rk2Epf)) * rk2HTheta) / rk2Thetaf);
							rk2HN_y = 2.0 * rk2CovdDf_x * rk2Covuvfu_xy + 2.0 * rk2CovdDf_y * rk2Covuvfu_yy + 2.0 * rk2CovdDf_z * rk2Covuvfu_zy + (vector(rk2Df, i, j, k) * rk2Hv_y) / rk2sdetg;
							rk2HM_yz = (-2.0 * rk2Covdvfu_xz * rk2CovuBfu_xy) + 2.0 * rk2Covdvfu_xy * rk2CovuBfu_xz + (-2.0 * rk2Covdvfu_yz * rk2CovuBfu_yy) + 2.0 * rk2Covdvfu_yy * rk2CovuBfu_yz + (-2.0 * rk2Covdvfu_zz * rk2CovuBfu_zy) + 2.0 * rk2Covdvfu_zy * rk2CovuBfu_zz + (vector(rk2Bfu_z, i, j, k) * rk2Hv_y - vector(rk2Bfu_y, i, j, k) * rk2Hv_z) / rk2sdetg;
							rk2HNe_y = 2.0 * rk2CovdDYf_x * rk2Covuvfu_xy + 2.0 * rk2CovdDYf_y * rk2Covuvfu_yy + 2.0 * rk2CovdDYf_z * rk2Covuvfu_zy + (vector(rk2DYf, i, j, k) * rk2Hv_y) / rk2sdetg;
							rk2Hv_x = rk2Phivh_x + (-((rk2vfu_x + (rk2Bvf * vector(rk2Bfu_x, i, j, k)) / (rk2sdetg * rk2Epf)) * rk2HTheta) / rk2Thetaf);
							rk2HN_x = 2.0 * rk2CovdDf_x * rk2Covuvfu_xx + 2.0 * rk2CovdDf_y * rk2Covuvfu_yx + 2.0 * rk2CovdDf_z * rk2Covuvfu_zx + (vector(rk2Df, i, j, k) * rk2Hv_x) / rk2sdetg;
							rk2HM_xz = (-2.0 * rk2Covdvfu_xz * rk2CovuBfu_xx) + 2.0 * rk2Covdvfu_xx * rk2CovuBfu_xz + (-2.0 * rk2Covdvfu_yz * rk2CovuBfu_yx) + 2.0 * rk2Covdvfu_yx * rk2CovuBfu_yz + (-2.0 * rk2Covdvfu_zz * rk2CovuBfu_zx) + 2.0 * rk2Covdvfu_zx * rk2CovuBfu_zz + (vector(rk2Bfu_z, i, j, k) * rk2Hv_x - vector(rk2Bfu_x, i, j, k) * rk2Hv_z) / rk2sdetg;
							rk2HM_xy = (-2.0 * rk2Covdvfu_xy * rk2CovuBfu_xx) + 2.0 * rk2Covdvfu_xx * rk2CovuBfu_xy + (-2.0 * rk2Covdvfu_yy * rk2CovuBfu_yx) + 2.0 * rk2Covdvfu_yx * rk2CovuBfu_yy + (-2.0 * rk2Covdvfu_zy * rk2CovuBfu_zx) + 2.0 * rk2Covdvfu_zx * rk2CovuBfu_zy + (vector(rk2Bfu_y, i, j, k) * rk2Hv_x - vector(rk2Bfu_x, i, j, k) * rk2Hv_y) / rk2sdetg;
							rk2HE_z = 0.5 * ((-2.0 * rk2gtu_yz * rk2HM_xz) + 2.0 * rk2gtu_zz * rk2HM_xy + 2.0 * rk2gtu_xz * rk2HM_yz) / sqrt(rk2chi_max);
							rk2HE_x = 0.5 * ((-2.0 * rk2gtu_xy * rk2HM_xz) + 2.0 * rk2gtu_xz * rk2HM_xy + 2.0 * rk2gtu_xx * rk2HM_yz) / sqrt(rk2chi_max);
							rk2HE_y = 0.5 * ((-2.0 * rk2gtu_yy * rk2HM_xz) + 2.0 * rk2gtu_yz * rk2HM_xy + 2.0 * rk2gtu_xy * rk2HM_yz) / sqrt(rk2chi_max);
							rk2HT_zz = 4.0 * rk2CovdEpf_x * rk2Covuvfu_xz * rk2vfu_z + 4.0 * rk2CovdEpf_y * rk2Covuvfu_yz * rk2vfu_z + 4.0 * rk2CovdEpf_z * rk2Covuvfu_zz * rk2vfu_z + rk2Epf * (2.0 * rk2vfu_z * rk2Hv_z + 2.0 * rk2Covdvfu_xz * rk2Covuvfu_xz + 2.0 * rk2Covdvfu_yz * rk2Covuvfu_yz + 2.0 * rk2Covdvfu_zz * rk2Covuvfu_zz) + 2.0 * (rk2vfu_z * rk2vfu_z) * rk2HPres + (-2.0 * rk2CovdBfu_xz * rk2CovuBfu_xz) + (-2.0 * rk2CovdBfu_yz * rk2CovuBfu_yz) + (-2.0 * rk2CovdBfu_zz * rk2CovuBfu_zz) + (-2.0 * rk2CovdEfu_xz * rk2CovuEfu_xz) + (-2.0 * rk2CovdEfu_yz * rk2CovuEfu_yz) + (-2.0 * rk2CovdEfu_zz * rk2CovuEfu_zz) + (-2.00 * vector(Efu_z, i, j, k) * rk2HE_z) + (rk2chi_max * rk2gtu_zz - rk2vfu_z * rk2vfu_z) * (rk2HPres + rk2CovdBfd_xx * rk2CovuBfu_xx + rk2CovdBfd_yx * rk2CovuBfu_yx + rk2CovdBfd_zx * rk2CovuBfu_zx + rk2CovdBfd_xy * rk2CovuBfu_xy + rk2CovdBfd_yy * rk2CovuBfu_yy + rk2CovdBfd_zy * rk2CovuBfu_zy + rk2CovdBfd_xz * rk2CovuBfu_xz + rk2CovdBfd_yz * rk2CovuBfu_yz + rk2CovdBfd_zz * rk2CovuBfu_zz + rk2CovdEfd_xx * rk2CovuEfu_xx + rk2CovdEfd_yx * rk2CovuEfu_yx + rk2CovdEfd_zx * rk2CovuEfu_zx + rk2CovdEfd_xy * rk2CovuEfu_xy + rk2CovdEfd_yy * rk2CovuEfu_yy + rk2CovdEfd_zy * rk2CovuEfu_zy + rk2CovdEfd_xz * rk2CovuEfu_xz + rk2CovdEfd_yz * rk2CovuEfu_yz + rk2CovdEfd_zz * rk2CovuEfu_zz + rk2Efd_x * rk2HE_x + rk2Efd_y * rk2HE_y + rk2Efd_z * rk2HE_z);
							rk2HT_yy = 4.0 * rk2CovdEpf_x * rk2Covuvfu_xy * rk2vfu_y + 4.0 * rk2CovdEpf_y * rk2Covuvfu_yy * rk2vfu_y + 4.0 * rk2CovdEpf_z * rk2Covuvfu_zy * rk2vfu_y + rk2Epf * (2.0 * rk2vfu_y * rk2Hv_y + 2.0 * rk2Covdvfu_xy * rk2Covuvfu_xy + 2.0 * rk2Covdvfu_yy * rk2Covuvfu_yy + 2.0 * rk2Covdvfu_zy * rk2Covuvfu_zy) + 2.0 * (rk2vfu_y * rk2vfu_y) * rk2HPres + (-2.0 * rk2CovdBfu_xy * rk2CovuBfu_xy) + (-2.0 * rk2CovdBfu_yy * rk2CovuBfu_yy) + (-2.0 * rk2CovdBfu_zy * rk2CovuBfu_zy) + (-2.0 * rk2CovdEfu_xy * rk2CovuEfu_xy) + (-2.0 * rk2CovdEfu_yy * rk2CovuEfu_yy) + (-2.0 * rk2CovdEfu_zy * rk2CovuEfu_zy) + (-2.00 * vector(Efu_y, i, j, k) * rk2HE_y) + (rk2chi_max * rk2gtu_yy - rk2vfu_y * rk2vfu_y) * (rk2HPres + rk2CovdBfd_xx * rk2CovuBfu_xx + rk2CovdBfd_yx * rk2CovuBfu_yx + rk2CovdBfd_zx * rk2CovuBfu_zx + rk2CovdBfd_xy * rk2CovuBfu_xy + rk2CovdBfd_yy * rk2CovuBfu_yy + rk2CovdBfd_zy * rk2CovuBfu_zy + rk2CovdBfd_xz * rk2CovuBfu_xz + rk2CovdBfd_yz * rk2CovuBfu_yz + rk2CovdBfd_zz * rk2CovuBfu_zz + rk2CovdEfd_xx * rk2CovuEfu_xx + rk2CovdEfd_yx * rk2CovuEfu_yx + rk2CovdEfd_zx * rk2CovuEfu_zx + rk2CovdEfd_xy * rk2CovuEfu_xy + rk2CovdEfd_yy * rk2CovuEfu_yy + rk2CovdEfd_zy * rk2CovuEfu_zy + rk2CovdEfd_xz * rk2CovuEfu_xz + rk2CovdEfd_yz * rk2CovuEfu_yz + rk2CovdEfd_zz * rk2CovuEfu_zz + rk2Efd_x * rk2HE_x + rk2Efd_y * rk2HE_y + rk2Efd_z * rk2HE_z);
							rk2HT_yz = 2.0 * rk2CovdEpf_x * (rk2Covuvfu_xz * rk2vfu_y + rk2Covuvfu_xy * rk2vfu_z) + 2.0 * rk2CovdEpf_y * (rk2Covuvfu_yz * rk2vfu_y + rk2Covuvfu_yy * rk2vfu_z) + 2.0 * rk2CovdEpf_z * (rk2Covuvfu_zz * rk2vfu_y + rk2Covuvfu_zy * rk2vfu_z) + rk2Epf * (rk2vfu_y * rk2Hv_z + rk2vfu_z * rk2Hv_y + 2.0 * rk2Covdvfu_xy * rk2Covuvfu_xz + 2.0 * rk2Covdvfu_yy * rk2Covuvfu_yz + 2.0 * rk2Covdvfu_zy * rk2Covuvfu_zz) + 2.0 * rk2vfu_y * rk2vfu_z * rk2HPres + (-2.0 * rk2CovdBfu_xy * rk2CovuBfu_xz) + (-2.0 * rk2CovdBfu_yy * rk2CovuBfu_yz) + (-2.0 * rk2CovdBfu_zy * rk2CovuBfu_zz) + (-2.0 * rk2CovdEfu_xy * rk2CovuEfu_xz) + (-2.0 * rk2CovdEfu_yy * rk2CovuEfu_yz) + (-2.0 * rk2CovdEfu_zy * rk2CovuEfu_zz) + (-1.00 * vector(Efu_z, i, j, k) * rk2HE_y) + (-1.00 * vector(Efu_y, i, j, k) * rk2HE_z) + (rk2chi_max * rk2gtu_yz - rk2vfu_z * rk2vfu_y) * (rk2HPres + rk2CovdBfd_xx * rk2CovuBfu_xx + rk2CovdBfd_yx * rk2CovuBfu_yx + rk2CovdBfd_zx * rk2CovuBfu_zx + rk2CovdBfd_xy * rk2CovuBfu_xy + rk2CovdBfd_yy * rk2CovuBfu_yy + rk2CovdBfd_zy * rk2CovuBfu_zy + rk2CovdBfd_xz * rk2CovuBfu_xz + rk2CovdBfd_yz * rk2CovuBfu_yz + rk2CovdBfd_zz * rk2CovuBfu_zz + rk2CovdEfd_xx * rk2CovuEfu_xx + rk2CovdEfd_yx * rk2CovuEfu_yx + rk2CovdEfd_zx * rk2CovuEfu_zx + rk2CovdEfd_xy * rk2CovuEfu_xy + rk2CovdEfd_yy * rk2CovuEfu_yy + rk2CovdEfd_zy * rk2CovuEfu_zy + rk2CovdEfd_xz * rk2CovuEfu_xz + rk2CovdEfd_yz * rk2CovuEfu_yz + rk2CovdEfd_zz * rk2CovuEfu_zz + rk2Efd_x * rk2HE_x + rk2Efd_y * rk2HE_y + rk2Efd_z * rk2HE_z);
							rk2HT_xx = 4.0 * rk2CovdEpf_x * rk2Covuvfu_xx * rk2vfu_x + 4.0 * rk2CovdEpf_y * rk2Covuvfu_yx * rk2vfu_x + 4.0 * rk2CovdEpf_z * rk2Covuvfu_zx * rk2vfu_x + rk2Epf * (2.0 * rk2vfu_x * rk2Hv_x + 2.0 * rk2Covdvfu_xx * rk2Covuvfu_xx + 2.0 * rk2Covdvfu_yx * rk2Covuvfu_yx + 2.0 * rk2Covdvfu_zx * rk2Covuvfu_zx) + 2.0 * (rk2vfu_x * rk2vfu_x) * rk2HPres + (-2.0 * rk2CovdBfu_xx * rk2CovuBfu_xx) + (-2.0 * rk2CovdBfu_yx * rk2CovuBfu_yx) + (-2.0 * rk2CovdBfu_zx * rk2CovuBfu_zx) + (-2.0 * rk2CovdEfu_xx * rk2CovuEfu_xx) + (-2.0 * rk2CovdEfu_yx * rk2CovuEfu_yx) + (-2.0 * rk2CovdEfu_zx * rk2CovuEfu_zx) + (-2.00 * vector(Efu_x, i, j, k) * rk2HE_x) + (rk2chi_max * rk2gtu_xx - rk2vfu_x * rk2vfu_x) * (rk2HPres + rk2CovdBfd_xx * rk2CovuBfu_xx + rk2CovdBfd_yx * rk2CovuBfu_yx + rk2CovdBfd_zx * rk2CovuBfu_zx + rk2CovdBfd_xy * rk2CovuBfu_xy + rk2CovdBfd_yy * rk2CovuBfu_yy + rk2CovdBfd_zy * rk2CovuBfu_zy + rk2CovdBfd_xz * rk2CovuBfu_xz + rk2CovdBfd_yz * rk2CovuBfu_yz + rk2CovdBfd_zz * rk2CovuBfu_zz + rk2CovdEfd_xx * rk2CovuEfu_xx + rk2CovdEfd_yx * rk2CovuEfu_yx + rk2CovdEfd_zx * rk2CovuEfu_zx + rk2CovdEfd_xy * rk2CovuEfu_xy + rk2CovdEfd_yy * rk2CovuEfu_yy + rk2CovdEfd_zy * rk2CovuEfu_zy + rk2CovdEfd_xz * rk2CovuEfu_xz + rk2CovdEfd_yz * rk2CovuEfu_yz + rk2CovdEfd_zz * rk2CovuEfu_zz + rk2Efd_x * rk2HE_x + rk2Efd_y * rk2HE_y + rk2Efd_z * rk2HE_z);
							rk2HT_xz = 2.0 * rk2CovdEpf_x * (rk2Covuvfu_xz * rk2vfu_x + rk2Covuvfu_xx * rk2vfu_z) + 2.0 * rk2CovdEpf_y * (rk2Covuvfu_yz * rk2vfu_x + rk2Covuvfu_yx * rk2vfu_z) + 2.0 * rk2CovdEpf_z * (rk2Covuvfu_zz * rk2vfu_x + rk2Covuvfu_zx * rk2vfu_z) + rk2Epf * (rk2vfu_x * rk2Hv_z + rk2vfu_z * rk2Hv_x + 2.0 * rk2Covdvfu_xx * rk2Covuvfu_xz + 2.0 * rk2Covdvfu_yx * rk2Covuvfu_yz + 2.0 * rk2Covdvfu_zx * rk2Covuvfu_zz) + 2.0 * rk2vfu_x * rk2vfu_z * rk2HPres + (-2.0 * rk2CovdBfu_xx * rk2CovuBfu_xz) + (-2.0 * rk2CovdBfu_yx * rk2CovuBfu_yz) + (-2.0 * rk2CovdBfu_zx * rk2CovuBfu_zz) + (-2.0 * rk2CovdEfu_xx * rk2CovuEfu_xz) + (-2.0 * rk2CovdEfu_yx * rk2CovuEfu_yz) + (-2.0 * rk2CovdEfu_zx * rk2CovuEfu_zz) + (-1.00 * vector(Efu_z, i, j, k) * rk2HE_x) + (-1.00 * vector(Efu_x, i, j, k) * rk2HE_z) + (rk2chi_max * rk2gtu_xz - rk2vfu_z * rk2vfu_x) * (rk2HPres + rk2CovdBfd_xx * rk2CovuBfu_xx + rk2CovdBfd_yx * rk2CovuBfu_yx + rk2CovdBfd_zx * rk2CovuBfu_zx + rk2CovdBfd_xy * rk2CovuBfu_xy + rk2CovdBfd_yy * rk2CovuBfu_yy + rk2CovdBfd_zy * rk2CovuBfu_zy + rk2CovdBfd_xz * rk2CovuBfu_xz + rk2CovdBfd_yz * rk2CovuBfu_yz + rk2CovdBfd_zz * rk2CovuBfu_zz + rk2CovdEfd_xx * rk2CovuEfu_xx + rk2CovdEfd_yx * rk2CovuEfu_yx + rk2CovdEfd_zx * rk2CovuEfu_zx + rk2CovdEfd_xy * rk2CovuEfu_xy + rk2CovdEfd_yy * rk2CovuEfu_yy + rk2CovdEfd_zy * rk2CovuEfu_zy + rk2CovdEfd_xz * rk2CovuEfu_xz + rk2CovdEfd_yz * rk2CovuEfu_yz + rk2CovdEfd_zz * rk2CovuEfu_zz + rk2Efd_x * rk2HE_x + rk2Efd_y * rk2HE_y + rk2Efd_z * rk2HE_z);
							rk2HT_xy = 2.0 * rk2CovdEpf_x * (rk2Covuvfu_xy * rk2vfu_x + rk2Covuvfu_xx * rk2vfu_y) + 2.0 * rk2CovdEpf_y * (rk2Covuvfu_yy * rk2vfu_x + rk2Covuvfu_yx * rk2vfu_y) + 2.0 * rk2CovdEpf_z * (rk2Covuvfu_zy * rk2vfu_x + rk2Covuvfu_zx * rk2vfu_y) + rk2Epf * (rk2vfu_x * rk2Hv_y + rk2vfu_y * rk2Hv_x + 2.0 * rk2Covdvfu_xx * rk2Covuvfu_xy + 2.0 * rk2Covdvfu_yx * rk2Covuvfu_yy + 2.0 * rk2Covdvfu_zx * rk2Covuvfu_zy) + 2.0 * rk2vfu_x * rk2vfu_y * rk2HPres + (-2.0 * rk2CovdBfu_xx * rk2CovuBfu_xy) + (-2.0 * rk2CovdBfu_yx * rk2CovuBfu_yy) + (-2.0 * rk2CovdBfu_zx * rk2CovuBfu_zy) + (-2.0 * rk2CovdEfu_xx * rk2CovuEfu_xy) + (-2.0 * rk2CovdEfu_yx * rk2CovuEfu_yy) + (-2.0 * rk2CovdEfu_zx * rk2CovuEfu_zy) + (-1.00 * vector(Efu_y, i, j, k) * rk2HE_x) + (-1.00 * vector(Efu_x, i, j, k) * rk2HE_y) + (rk2chi_max * rk2gtu_xy - rk2vfu_y * rk2vfu_x) * (rk2HPres + rk2CovdBfd_xx * rk2CovuBfu_xx + rk2CovdBfd_yx * rk2CovuBfu_yx + rk2CovdBfd_zx * rk2CovuBfu_zx + rk2CovdBfd_xy * rk2CovuBfu_xy + rk2CovdBfd_yy * rk2CovuBfu_yy + rk2CovdBfd_zy * rk2CovuBfu_zy + rk2CovdBfd_xz * rk2CovuBfu_xz + rk2CovdBfd_yz * rk2CovuBfu_yz + rk2CovdBfd_zz * rk2CovuBfu_zz + rk2CovdEfd_xx * rk2CovuEfu_xx + rk2CovdEfd_yx * rk2CovuEfu_yx + rk2CovdEfd_zx * rk2CovuEfu_zx + rk2CovdEfd_xy * rk2CovuEfu_xy + rk2CovdEfd_yy * rk2CovuEfu_yy + rk2CovdEfd_zy * rk2CovuEfu_zy + rk2CovdEfd_xz * rk2CovuEfu_xz + rk2CovdEfd_yz * rk2CovuEfu_yz + rk2CovdEfd_zz * rk2CovuEfu_zz + rk2Efd_x * rk2HE_x + rk2Efd_y * rk2HE_y + rk2Efd_z * rk2HE_z);
							rk2HNe_x = 2.0 * rk2CovdDYf_x * rk2Covuvfu_xx + 2.0 * rk2CovdDYf_y * rk2Covuvfu_yx + 2.0 * rk2CovdDYf_z * rk2Covuvfu_zx + (vector(rk2DYf, i, j, k) * rk2Hv_x) / rk2sdetg;
							d_TauN_x_o0_t0_m0_l0 = -rk2faceta * C_N * rk2HN_x;
							d_TauN_y_o0_t0_m0_l0 = -rk2faceta * C_N * rk2HN_y;
							d_TauN_z_o0_t0_m0_l0 = -rk2faceta * C_N * rk2HN_z;
							d_TauNe_x_o0_t0_m0_l0 = -rk2faceta * C_N * rk2HNe_x;
							d_TauNe_y_o0_t0_m0_l0 = -rk2faceta * C_N * rk2HNe_y;
							d_TauNe_z_o0_t0_m0_l0 = -rk2faceta * C_N * rk2HNe_z;
							d_TauM_xy_o0_t0_m0_l0 = -rk2faceta * C_M * rk2HM_xy;
							d_TauM_xz_o0_t0_m0_l0 = -rk2faceta * C_M * rk2HM_xz;
							d_TauM_yz_o0_t0_m0_l0 = -rk2faceta * C_M * rk2HM_yz;
							d_TauT_xx_o0_t0_m0_l0 = -rk2faceta * C_T * rk2HT_xx;
							d_TauT_xy_o0_t0_m0_l0 = -rk2faceta * C_T * rk2HT_xy;
							d_TauT_xz_o0_t0_m0_l0 = -rk2faceta * C_T * rk2HT_xz;
							d_TauT_yy_o0_t0_m0_l0 = -rk2faceta * C_T * rk2HT_yy;
							d_TauT_yz_o0_t0_m0_l0 = -rk2faceta * C_T * rk2HT_yz;
							d_TauT_zz_o0_t0_m0_l0 = -rk2faceta * C_T * rk2HT_zz;
							vector(TauNe_x, i, j, k) = d_TauNe_x_o0_t0_m0_l0;
							vector(TauNe_y, i, j, k) = d_TauNe_y_o0_t0_m0_l0;
							vector(TauT_yz, i, j, k) = d_TauT_yz_o0_t0_m0_l0;
							vector(TauT_xy, i, j, k) = d_TauT_xy_o0_t0_m0_l0;
							vector(TauM_yz, i, j, k) = d_TauM_yz_o0_t0_m0_l0;
							vector(TauM_xy, i, j, k) = d_TauM_xy_o0_t0_m0_l0;
							vector(TauNe_z, i, j, k) = d_TauNe_z_o0_t0_m0_l0;
							vector(TauT_xz, i, j, k) = d_TauT_xz_o0_t0_m0_l0;
							vector(TauN_x, i, j, k) = d_TauN_x_o0_t0_m0_l0;
							vector(TauT_zz, i, j, k) = d_TauT_zz_o0_t0_m0_l0;
							vector(TauT_yy, i, j, k) = d_TauT_yy_o0_t0_m0_l0;
							vector(TauT_xx, i, j, k) = d_TauT_xx_o0_t0_m0_l0;
							vector(TauM_xz, i, j, k) = d_TauM_xz_o0_t0_m0_l0;
							vector(TauN_z, i, j, k) = d_TauN_z_o0_t0_m0_l0;
							vector(TauN_y, i, j, k) = d_TauN_y_o0_t0_m0_l0;
						} else {
							rk2CovdEfu_yz = 0.0;
							rk2CovuEfu_zz = 0.0;
							rk2CovuEfu_xz = 0.0;
							rk2CovdEfu_yy = 0.0;
							rk2CovuEfu_zy = 0.0;
							rk2CovuEfu_xy = 0.0;
							rk2CovdEfu_yx = 0.0;
							rk2CovdEfd_yy = 0.0;
							rk2CovdEfd_yz = 0.0;
							rk2CovdEfd_yx = 0.0;
							rk2CovuEfu_zx = 0.0;
							rk2CovuEfu_xx = 0.0;
							rk2CovuEfu_yx = 0.0;
							rk2CovuEfu_yz = 0.0;
							rk2CovuEfu_yy = 0.0;
							rk2PhiThetah = 0.0;
							rk2HPres = 0.0;
							rk2HTheta = 0.0;
							rk2Hv_z = 0.0;
							rk2HN_z = 0.0;
							rk2HNe_z = 0.0;
							rk2Hv_y = 0.0;
							rk2HN_y = 0.0;
							rk2HM_yz = 0.0;
							rk2HNe_y = 0.0;
							rk2Hv_x = 0.0;
							rk2HN_x = 0.0;
							rk2HM_xz = 0.0;
							rk2HM_xy = 0.0;
							rk2HE_z = 0.0;
							rk2HE_x = 0.0;
							rk2HE_y = 0.0;
							rk2HT_zz = 0.0;
							rk2HT_yy = 0.0;
							rk2HT_yz = 0.0;
							rk2HT_xx = 0.0;
							rk2HT_xz = 0.0;
							rk2HT_xy = 0.0;
							rk2HNe_x = 0.0;
						}
						optdepthe_tmp = vector(optdepthe, i, j, k);
						optdeptha_tmp = vector(optdeptha, i, j, k);
						optdepthx_tmp = vector(optdepthx, i, j, k);
						chie_tmp = vector(chie, i, j, k);
						chia_tmp = vector(chia, i, j, k);
						chix_tmp = vector(chix, i, j, k);
						qnu_tmp = vector(qnu, i, j, k);
						rnu_tmp = vector(rnu, i, j, k);
						qnu_e_tmp = vector(qnu_e, i, j, k);
						qnu_a_tmp = vector(qnu_a, i, j, k);
						qnu_x_tmp = vector(qnu_x, i, j, k);
						betabeta = vector(rk2gtd_xx, i, j, k) * vector(rk2Betau_x, i, j, k) * vector(rk2Betau_x, i, j, k) + vector(rk2gtd_yy, i, j, k) * vector(rk2Betau_y, i, j, k) * vector(rk2Betau_y, i, j, k) + vector(rk2gtd_zz, i, j, k) * vector(rk2Betau_z, i, j, k) * vector(rk2Betau_z, i, j, k) + 2.0 * vector(rk2gtd_xy, i, j, k) * vector(rk2Betau_x, i, j, k) * vector(rk2Betau_y, i, j, k) + 2.0 * vector(rk2gtd_xz, i, j, k) * vector(rk2Betau_x, i, j, k) * vector(rk2Betau_z, i, j, k) + 2.0 * vector(rk2gtd_yz, i, j, k) * vector(rk2Betau_y, i, j, k) * vector(rk2Betau_z, i, j, k);
						betav = vector(rk2gtd_xx, i, j, k) * vector(rk2Betau_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk2gtd_yy, i, j, k) * vector(rk2Betau_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk2gtd_zz, i, j, k) * vector(rk2Betau_z, i, j, k) * vector(vfd_z, i, j, k) + vector(rk2gtd_xy, i, j, k) * vector(rk2Betau_x, i, j, k) * vector(vfd_y, i, j, k) + vector(rk2gtd_xy, i, j, k) * vector(rk2Betau_y, i, j, k) * vector(vfd_x, i, j, k) + vector(rk2gtd_xz, i, j, k) * vector(rk2Betau_x, i, j, k) * vector(vfd_z, i, j, k) + vector(rk2gtd_xz, i, j, k) * vector(rk2Betau_z, i, j, k) * vector(vfd_x, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Betau_y, i, j, k) * vector(vfd_z, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Betau_z, i, j, k) * vector(vfd_y, i, j, k);
						velr = vector(vfd_x, i, j, k) * xcoord(i) + vector(vfd_y, i, j, k) * ycoord(j) + vector(vfd_z, i, j, k) * zcoord(k);
						velr = velr / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k) + 0.01);
						f_redshift = rk2sqW * (vector(rk2Alpha, i, j, k) - betav) * (1.0 + velr);
						f_redshift = f_redshift * sqrt(vector(rk2Alpha, i, j, k) * vector(rk2Alpha, i, j, k) - betabeta);
						leakage(do_leakage, vacuum_rho, threshold_leakage_vacuum, energyShift, optdepthe_tmp, optdeptha_tmp, optdepthx_tmp, qnu_e_tmp, qnu_a_tmp, qnu_x_tmp, chie_tmp, chia_tmp, chix_tmp, qnu_tmp, rnu_tmp, vector(rhof, i, j, k), vector(epsf, i, j, k), vector(Yef, i, j, k), vector(Tf, i, j, k), rk2sdetg, rk2optdeptheh, rk2optdepthah, rk2optdepthxh, rk2optdepthev, rk2optdepthav, rk2optdepthxv, rk2optdepthed, rk2optdepthad, rk2optdepthxd, rk2chieh, rk2chiah, rk2chixh, rk2chiev, rk2chiav, rk2chixv, rk2chied, rk2chiad, rk2chixd, f_redshift, coord0_data_DataLeakage, coord0_dx_DataLeakage, coord0_size_DataLeakage, coord1_data_DataLeakage, coord1_dx_DataLeakage, coord1_size_DataLeakage, coord2_data_DataLeakage, coord2_dx_DataLeakage, coord2_size_DataLeakage, vars_data_DataLeakage, vars_size_DataLeakage, dx, simPlat_dt, ilast, jlast);
						vector(optdepthe, i, j, k) = optdepthe_tmp;
						vector(optdeptha, i, j, k) = optdeptha_tmp;
						vector(optdepthx, i, j, k) = optdepthx_tmp;
						vector(chie, i, j, k) = chie_tmp;
						vector(chia, i, j, k) = chia_tmp;
						vector(chix, i, j, k) = chix_tmp;
						vector(qnu_e, i, j, k) = qnu_e_tmp;
						vector(qnu_a, i, j, k) = qnu_a_tmp;
						vector(qnu_x, i, j, k) = qnu_x_tmp;
						vector(qnu, i, j, k) = qnu_tmp;
						vector(rnu, i, j, k) = rnu_tmp;
					}
				}
			}
		}
	}
	//Fill ghosts and periodical boundaries
	d_bdry_sched_advance12[ln]->fillData(current_time, false);
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		//Hard region field distance variables
		double* d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
	
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
		double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
		double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
		double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
		double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
		double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
		double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
		double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
		double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
		double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
		double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
		double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
		double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
		double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
		double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
		double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
		double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
		double* rk2gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xx_id).get())->getPointer();
		double* rk2gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xy_id).get())->getPointer();
		double* rk2gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xz_id).get())->getPointer();
		double* rk2gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yy_id).get())->getPointer();
		double* rk2gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yz_id).get())->getPointer();
		double* rk2gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_zz_id).get())->getPointer();
		double* rk2Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xx_id).get())->getPointer();
		double* rk2Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xy_id).get())->getPointer();
		double* rk2Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xz_id).get())->getPointer();
		double* rk2Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_yy_id).get())->getPointer();
		double* rk2Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_yz_id).get())->getPointer();
		double* rk2Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_zz_id).get())->getPointer();
		double* rk2Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_x_id).get())->getPointer();
		double* rk2Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_y_id).get())->getPointer();
		double* rk2Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_z_id).get())->getPointer();
		double* rk2Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_x_id).get())->getPointer();
		double* rk2Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_y_id).get())->getPointer();
		double* rk2Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_z_id).get())->getPointer();
		double* rk2Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Alpha_id).get())->getPointer();
		double* rk2chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2chi_id).get())->getPointer();
		double* rk2trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2trK_id).get())->getPointer();
		double* rk2theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2theta_id).get())->getPointer();
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk2theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
	}
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		double* rk2chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2chi_id).get())->getPointer();
		double* rk2gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xy_id).get())->getPointer();
		double* rk2Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_x_id).get())->getPointer();
		double* rk2gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yy_id).get())->getPointer();
		double* rk2Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_y_id).get())->getPointer();
		double* rk2gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yz_id).get())->getPointer();
		double* rk2Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_z_id).get())->getPointer();
		double* rk2gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xz_id).get())->getPointer();
		double* rk2gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_zz_id).get())->getPointer();
		double* rk2gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xx_id).get())->getPointer();
		double* rk2Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_x_id).get())->getPointer();
		double* rk2Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_y_id).get())->getPointer();
		double* rk2Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_z_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* rk2Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_x_id).get())->getPointer();
		double* rk2Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Alpha_id).get())->getPointer();
		double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
		double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
		double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
		double* rk2phif = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2phif_id).get())->getPointer();
		double* rk2Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_y_id).get())->getPointer();
		double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
		double* rk2Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_z_id).get())->getPointer();
		double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
		double* rk2Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Df_id).get())->getPointer();
		double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
		double* rk2DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2DYf_id).get())->getPointer();
		double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
		double* rk2tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2tauf_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
		double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
		double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
		double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
		double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
		double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
		double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
		double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
		double* rk2Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_x_id).get())->getPointer();
		double* rk2Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_y_id).get())->getPointer();
		double* rk2Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_z_id).get())->getPointer();
		double* rk2Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xx_id).get())->getPointer();
		double* rk2Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xy_id).get())->getPointer();
		double* rk2Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xz_id).get())->getPointer();
		double* rk2Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_yy_id).get())->getPointer();
		double* rk2Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_yz_id).get())->getPointer();
		double* rk2Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_zz_id).get())->getPointer();
		double* rk2trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2trK_id).get())->getPointer();
		double* rk2theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2theta_id).get())->getPointer();
		double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
		double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
		double* rk3Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_x_id).get())->getPointer();
		double* Sfd_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_p_id).get())->getPointer();
		double* rk3Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_y_id).get())->getPointer();
		double* Sfd_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_p_id).get())->getPointer();
		double* rk3Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_z_id).get())->getPointer();
		double* Sfd_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_p_id).get())->getPointer();
		double* rk3Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_x_id).get())->getPointer();
		double* Bfu_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_p_id).get())->getPointer();
		double* rk3Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_y_id).get())->getPointer();
		double* Bfu_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_p_id).get())->getPointer();
		double* rk3Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_z_id).get())->getPointer();
		double* Bfu_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_p_id).get())->getPointer();
		double* rk3Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Df_id).get())->getPointer();
		double* Df_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_p_id).get())->getPointer();
		double* rk3DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3DYf_id).get())->getPointer();
		double* DYf_p = ((pdat::NodeData<double> *) patch->getPatchData(d_DYf_p_id).get())->getPointer();
		double* rk3tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3tauf_id).get())->getPointer();
		double* tauf_p = ((pdat::NodeData<double> *) patch->getPatchData(d_tauf_p_id).get())->getPointer();
		double* rk3phif = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3phif_id).get())->getPointer();
		double* phif_p = ((pdat::NodeData<double> *) patch->getPatchData(d_phif_p_id).get())->getPointer();
		double* rk3gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xx_id).get())->getPointer();
		double* gtd_xx_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_p_id).get())->getPointer();
		double* rk3gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xy_id).get())->getPointer();
		double* gtd_xy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_p_id).get())->getPointer();
		double* rk3gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xz_id).get())->getPointer();
		double* gtd_xz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_p_id).get())->getPointer();
		double* rk3gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_yy_id).get())->getPointer();
		double* gtd_yy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_p_id).get())->getPointer();
		double* rk3gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_yz_id).get())->getPointer();
		double* gtd_yz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_p_id).get())->getPointer();
		double* rk3gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_zz_id).get())->getPointer();
		double* gtd_zz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_p_id).get())->getPointer();
		double* rk3Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_xx_id).get())->getPointer();
		double* Atd_xx_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xx_p_id).get())->getPointer();
		double* rk3Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_xy_id).get())->getPointer();
		double* Atd_xy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xy_p_id).get())->getPointer();
		double* rk3Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_xz_id).get())->getPointer();
		double* Atd_xz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xz_p_id).get())->getPointer();
		double* rk3Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_yy_id).get())->getPointer();
		double* Atd_yy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yy_p_id).get())->getPointer();
		double* rk3Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_yz_id).get())->getPointer();
		double* Atd_yz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yz_p_id).get())->getPointer();
		double* rk3Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_zz_id).get())->getPointer();
		double* Atd_zz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_zz_p_id).get())->getPointer();
		double* rk3Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Gamh_x_id).get())->getPointer();
		double* Gamh_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_x_p_id).get())->getPointer();
		double* rk3Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Gamh_y_id).get())->getPointer();
		double* Gamh_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_y_p_id).get())->getPointer();
		double* rk3Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Gamh_z_id).get())->getPointer();
		double* Gamh_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_z_p_id).get())->getPointer();
		double* rk3Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_x_id).get())->getPointer();
		double* Betau_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_p_id).get())->getPointer();
		double* rk3Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_y_id).get())->getPointer();
		double* Betau_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_p_id).get())->getPointer();
		double* rk3Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_z_id).get())->getPointer();
		double* Betau_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_p_id).get())->getPointer();
		double* rk3Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Alpha_id).get())->getPointer();
		double* Alpha_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_p_id).get())->getPointer();
		double* rk3chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3chi_id).get())->getPointer();
		double* chi_p = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_p_id).get())->getPointer();
		double* rk3trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3trK_id).get())->getPointer();
		double* trK_p = ((pdat::NodeData<double> *) patch->getPatchData(d_trK_p_id).get())->getPointer();
		double* rk3theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3theta_id).get())->getPointer();
		double* theta_p = ((pdat::NodeData<double> *) patch->getPatchData(d_theta_p_id).get())->getPointer();
		double rk2chi_max, rk2sdetg, rk2inv_chi, rk2Bfd_y, rk2Bfd_z, rk2Bfd_x, rk2sqBf, rk2detgtd, rk2idetgtd, rk2gtu_zz, rk2gtu_xx, rk2gtu_xz, rk2gtu_xy, rk2Sfu_x, rk2vfu_x, rk2Bfvf, rk2gtu_yy, rk2gtu_yz, rk2vfu_z, rk2Sfu_y, rk2Sfu_z, rk2vfu_y, rk2sqvf, rk2sqW, rk2Sfuu_zz, rk2Sfuu_xx, rk2Sfuu_xz, rk2Sfuu_xy, rk2Sfud_xx, rk2Sfud_xy, rk2Sfud_xz, rk2Sfuu_yz, rk2Sfud_zy, rk2Sfud_zz, rk2Sfud_zx, rk2Sfuu_yy, rk2Sfud_yx, rk2Sfud_yy, rk2Sfud_yz, FluxiSfd_x, FluxiSfd_y, FluxiSfd_z, FluxiBfu_x, FluxiBfu_y, FluxiBfu_z, FluxiDf, FluxiDYf, Fluxitauf, Fluxiphif, rk2h, rk2sqbc, rk2sqca, rk2sqa, SpeediSfd_x, SpeediSfd_y, SpeediSfd_z, SpeediBfu_x, SpeediBfu_y, SpeediBfu_z, SpeediDf, SpeediDYf, Speeditauf, Speediphif, Fluxm3iSfd_x, Fluxm3iSfd_y, Fluxm3iSfd_z, Fluxm3iBfu_x, Fluxm3iBfu_y, Fluxm3iBfu_z, Fluxm3iDf, Fluxm3iDYf, Fluxm3itauf, Fluxm3iphif, Fluxm2iSfd_x, Fluxm2iSfd_y, Fluxm2iSfd_z, Fluxm2iBfu_x, Fluxm2iBfu_y, Fluxm2iBfu_z, Fluxm2iDf, Fluxm2iDYf, Fluxm2itauf, Fluxm2iphif, Fluxm1iSfd_x, Fluxm1iSfd_y, Fluxm1iSfd_z, Fluxm1iBfu_x, Fluxm1iBfu_y, Fluxm1iBfu_z, Fluxm1iDf, Fluxm1iDYf, Fluxm1itauf, Fluxm1iphif, Fluxp1iSfd_x, Fluxp1iSfd_y, Fluxp1iSfd_z, Fluxp1iBfu_x, Fluxp1iBfu_y, Fluxp1iBfu_z, Fluxp1iDf, Fluxp1iDYf, Fluxp1itauf, Fluxp1iphif, Fluxp2iSfd_x, Fluxp2iSfd_y, Fluxp2iSfd_z, Fluxp2iBfu_x, Fluxp2iBfu_y, Fluxp2iBfu_z, Fluxp2iDf, Fluxp2iDYf, Fluxp2itauf, Fluxp2iphif, Fluxp3iSfd_x, Fluxp3iSfd_y, Fluxp3iSfd_z, Fluxp3iBfu_x, Fluxp3iBfu_y, Fluxp3iBfu_z, Fluxp3iDf, Fluxp3iDYf, Fluxp3itauf, Fluxp3iphif, LLFpiSfd_x, LLFpiSfd_y, LLFpiSfd_z, LLFpiBfu_x, LLFpiBfu_y, LLFpiBfu_z, LLFpiDf, LLFpiDYf, LLFpitauf, LLFpiphif, LLFniSfd_x, LLFniSfd_y, LLFniSfd_z, LLFniBfu_x, LLFniBfu_y, LLFniBfu_z, LLFniDf, LLFniDYf, LLFnitauf, LLFniphif, LLFp_m3iSfd_x, LLFp_m3iSfd_y, LLFp_m3iSfd_z, LLFp_m3iBfu_x, LLFp_m3iBfu_y, LLFp_m3iBfu_z, LLFp_m3iDf, LLFp_m3iDYf, LLFp_m3itauf, LLFp_m3iphif, LLFn_m3iSfd_x, LLFn_m3iSfd_y, LLFn_m3iSfd_z, LLFn_m3iBfu_x, LLFn_m3iBfu_y, LLFn_m3iBfu_z, LLFn_m3iDf, LLFn_m3iDYf, LLFn_m3itauf, LLFn_m3iphif, LLFp_m2iSfd_x, LLFp_m2iSfd_y, LLFp_m2iSfd_z, LLFp_m2iBfu_x, LLFp_m2iBfu_y, LLFp_m2iBfu_z, LLFp_m2iDf, LLFp_m2iDYf, LLFp_m2itauf, LLFp_m2iphif, LLFn_m2iSfd_x, LLFn_m2iSfd_y, LLFn_m2iSfd_z, LLFn_m2iBfu_x, LLFn_m2iBfu_y, LLFn_m2iBfu_z, LLFn_m2iDf, LLFn_m2iDYf, LLFn_m2itauf, LLFn_m2iphif, LLFp_m1iSfd_x, LLFp_m1iSfd_y, LLFp_m1iSfd_z, LLFp_m1iBfu_x, LLFp_m1iBfu_y, LLFp_m1iBfu_z, LLFp_m1iDf, LLFp_m1iDYf, LLFp_m1itauf, LLFp_m1iphif, LLFn_m1iSfd_x, LLFn_m1iSfd_y, LLFn_m1iSfd_z, LLFn_m1iBfu_x, LLFn_m1iBfu_y, LLFn_m1iBfu_z, LLFn_m1iDf, LLFn_m1iDYf, LLFn_m1itauf, LLFn_m1iphif, LLFp_p1iSfd_x, LLFp_p1iSfd_y, LLFp_p1iSfd_z, LLFp_p1iBfu_x, LLFp_p1iBfu_y, LLFp_p1iBfu_z, LLFp_p1iDf, LLFp_p1iDYf, LLFp_p1itauf, LLFp_p1iphif, LLFn_p1iSfd_x, LLFn_p1iSfd_y, LLFn_p1iSfd_z, LLFn_p1iBfu_x, LLFn_p1iBfu_y, LLFn_p1iBfu_z, LLFn_p1iDf, LLFn_p1iDYf, LLFn_p1itauf, LLFn_p1iphif, LLFp_p2iSfd_x, LLFp_p2iSfd_y, LLFp_p2iSfd_z, LLFp_p2iBfu_x, LLFp_p2iBfu_y, LLFp_p2iBfu_z, LLFp_p2iDf, LLFp_p2iDYf, LLFp_p2itauf, LLFp_p2iphif, LLFn_p2iSfd_x, LLFn_p2iSfd_y, LLFn_p2iSfd_z, LLFn_p2iBfu_x, LLFn_p2iBfu_y, LLFn_p2iBfu_z, LLFn_p2iDf, LLFn_p2iDYf, LLFn_p2itauf, LLFn_p2iphif, LLFp_p3iSfd_x, LLFp_p3iSfd_y, LLFp_p3iSfd_z, LLFp_p3iBfu_x, LLFp_p3iBfu_y, LLFp_p3iBfu_z, LLFp_p3iDf, LLFp_p3iDYf, LLFp_p3itauf, LLFp_p3iphif, LLFn_p3iSfd_x, LLFn_p3iSfd_y, LLFn_p3iSfd_z, LLFn_p3iBfu_x, LLFn_p3iBfu_y, LLFn_p3iBfu_z, LLFn_p3iDf, LLFn_p3iDYf, LLFn_p3itauf, LLFn_p3iphif, FluxjSfd_x, FluxjSfd_y, FluxjSfd_z, FluxjBfu_x, FluxjBfu_y, FluxjBfu_z, FluxjDf, FluxjDYf, Fluxjtauf, Fluxjphif, SpeedjSfd_x, SpeedjSfd_y, SpeedjSfd_z, SpeedjBfu_x, SpeedjBfu_y, SpeedjBfu_z, SpeedjDf, SpeedjDYf, Speedjtauf, Speedjphif, Fluxm3jSfd_x, Fluxm3jSfd_y, Fluxm3jSfd_z, Fluxm3jBfu_x, Fluxm3jBfu_y, Fluxm3jBfu_z, Fluxm3jDf, Fluxm3jDYf, Fluxm3jtauf, Fluxm3jphif, Fluxm2jSfd_x, Fluxm2jSfd_y, Fluxm2jSfd_z, Fluxm2jBfu_x, Fluxm2jBfu_y, Fluxm2jBfu_z, Fluxm2jDf, Fluxm2jDYf, Fluxm2jtauf, Fluxm2jphif, Fluxm1jSfd_x, Fluxm1jSfd_y, Fluxm1jSfd_z, Fluxm1jBfu_x, Fluxm1jBfu_y, Fluxm1jBfu_z, Fluxm1jDf, Fluxm1jDYf, Fluxm1jtauf, Fluxm1jphif, Fluxp1jSfd_x, Fluxp1jSfd_y, Fluxp1jSfd_z, Fluxp1jBfu_x, Fluxp1jBfu_y, Fluxp1jBfu_z, Fluxp1jDf, Fluxp1jDYf, Fluxp1jtauf, Fluxp1jphif, Fluxp2jSfd_x, Fluxp2jSfd_y, Fluxp2jSfd_z, Fluxp2jBfu_x, Fluxp2jBfu_y, Fluxp2jBfu_z, Fluxp2jDf, Fluxp2jDYf, Fluxp2jtauf, Fluxp2jphif, Fluxp3jSfd_x, Fluxp3jSfd_y, Fluxp3jSfd_z, Fluxp3jBfu_x, Fluxp3jBfu_y, Fluxp3jBfu_z, Fluxp3jDf, Fluxp3jDYf, Fluxp3jtauf, Fluxp3jphif, LLFpjSfd_x, LLFpjSfd_y, LLFpjSfd_z, LLFpjBfu_x, LLFpjBfu_y, LLFpjBfu_z, LLFpjDf, LLFpjDYf, LLFpjtauf, LLFpjphif, LLFnjSfd_x, LLFnjSfd_y, LLFnjSfd_z, LLFnjBfu_x, LLFnjBfu_y, LLFnjBfu_z, LLFnjDf, LLFnjDYf, LLFnjtauf, LLFnjphif, LLFp_m3jSfd_x, LLFp_m3jSfd_y, LLFp_m3jSfd_z, LLFp_m3jBfu_x, LLFp_m3jBfu_y, LLFp_m3jBfu_z, LLFp_m3jDf, LLFp_m3jDYf, LLFp_m3jtauf, LLFp_m3jphif, LLFn_m3jSfd_x, LLFn_m3jSfd_y, LLFn_m3jSfd_z, LLFn_m3jBfu_x, LLFn_m3jBfu_y, LLFn_m3jBfu_z, LLFn_m3jDf, LLFn_m3jDYf, LLFn_m3jtauf, LLFn_m3jphif, LLFp_m2jSfd_x, LLFp_m2jSfd_y, LLFp_m2jSfd_z, LLFp_m2jBfu_x, LLFp_m2jBfu_y, LLFp_m2jBfu_z, LLFp_m2jDf, LLFp_m2jDYf, LLFp_m2jtauf, LLFp_m2jphif, LLFn_m2jSfd_x, LLFn_m2jSfd_y, LLFn_m2jSfd_z, LLFn_m2jBfu_x, LLFn_m2jBfu_y, LLFn_m2jBfu_z, LLFn_m2jDf, LLFn_m2jDYf, LLFn_m2jtauf, LLFn_m2jphif, LLFp_m1jSfd_x, LLFp_m1jSfd_y, LLFp_m1jSfd_z, LLFp_m1jBfu_x, LLFp_m1jBfu_y, LLFp_m1jBfu_z, LLFp_m1jDf, LLFp_m1jDYf, LLFp_m1jtauf, LLFp_m1jphif, LLFn_m1jSfd_x, LLFn_m1jSfd_y, LLFn_m1jSfd_z, LLFn_m1jBfu_x, LLFn_m1jBfu_y, LLFn_m1jBfu_z, LLFn_m1jDf, LLFn_m1jDYf, LLFn_m1jtauf, LLFn_m1jphif, LLFp_p1jSfd_x, LLFp_p1jSfd_y, LLFp_p1jSfd_z, LLFp_p1jBfu_x, LLFp_p1jBfu_y, LLFp_p1jBfu_z, LLFp_p1jDf, LLFp_p1jDYf, LLFp_p1jtauf, LLFp_p1jphif, LLFn_p1jSfd_x, LLFn_p1jSfd_y, LLFn_p1jSfd_z, LLFn_p1jBfu_x, LLFn_p1jBfu_y, LLFn_p1jBfu_z, LLFn_p1jDf, LLFn_p1jDYf, LLFn_p1jtauf, LLFn_p1jphif, LLFp_p2jSfd_x, LLFp_p2jSfd_y, LLFp_p2jSfd_z, LLFp_p2jBfu_x, LLFp_p2jBfu_y, LLFp_p2jBfu_z, LLFp_p2jDf, LLFp_p2jDYf, LLFp_p2jtauf, LLFp_p2jphif, LLFn_p2jSfd_x, LLFn_p2jSfd_y, LLFn_p2jSfd_z, LLFn_p2jBfu_x, LLFn_p2jBfu_y, LLFn_p2jBfu_z, LLFn_p2jDf, LLFn_p2jDYf, LLFn_p2jtauf, LLFn_p2jphif, LLFp_p3jSfd_x, LLFp_p3jSfd_y, LLFp_p3jSfd_z, LLFp_p3jBfu_x, LLFp_p3jBfu_y, LLFp_p3jBfu_z, LLFp_p3jDf, LLFp_p3jDYf, LLFp_p3jtauf, LLFp_p3jphif, LLFn_p3jSfd_x, LLFn_p3jSfd_y, LLFn_p3jSfd_z, LLFn_p3jBfu_x, LLFn_p3jBfu_y, LLFn_p3jBfu_z, LLFn_p3jDf, LLFn_p3jDYf, LLFn_p3jtauf, LLFn_p3jphif, FluxkSfd_x, FluxkSfd_y, FluxkSfd_z, FluxkBfu_x, FluxkBfu_y, FluxkBfu_z, FluxkDf, FluxkDYf, Fluxktauf, Fluxkphif, SpeedkSfd_x, SpeedkSfd_y, SpeedkSfd_z, SpeedkBfu_x, SpeedkBfu_y, SpeedkBfu_z, SpeedkDf, SpeedkDYf, Speedktauf, Speedkphif, Fluxm3kSfd_x, Fluxm3kSfd_y, Fluxm3kSfd_z, Fluxm3kBfu_x, Fluxm3kBfu_y, Fluxm3kBfu_z, Fluxm3kDf, Fluxm3kDYf, Fluxm3ktauf, Fluxm3kphif, Fluxm2kSfd_x, Fluxm2kSfd_y, Fluxm2kSfd_z, Fluxm2kBfu_x, Fluxm2kBfu_y, Fluxm2kBfu_z, Fluxm2kDf, Fluxm2kDYf, Fluxm2ktauf, Fluxm2kphif, Fluxm1kSfd_x, Fluxm1kSfd_y, Fluxm1kSfd_z, Fluxm1kBfu_x, Fluxm1kBfu_y, Fluxm1kBfu_z, Fluxm1kDf, Fluxm1kDYf, Fluxm1ktauf, Fluxm1kphif, Fluxp1kSfd_x, Fluxp1kSfd_y, Fluxp1kSfd_z, Fluxp1kBfu_x, Fluxp1kBfu_y, Fluxp1kBfu_z, Fluxp1kDf, Fluxp1kDYf, Fluxp1ktauf, Fluxp1kphif, Fluxp2kSfd_x, Fluxp2kSfd_y, Fluxp2kSfd_z, Fluxp2kBfu_x, Fluxp2kBfu_y, Fluxp2kBfu_z, Fluxp2kDf, Fluxp2kDYf, Fluxp2ktauf, Fluxp2kphif, Fluxp3kSfd_x, Fluxp3kSfd_y, Fluxp3kSfd_z, Fluxp3kBfu_x, Fluxp3kBfu_y, Fluxp3kBfu_z, Fluxp3kDf, Fluxp3kDYf, Fluxp3ktauf, Fluxp3kphif, LLFpkSfd_x, LLFpkSfd_y, LLFpkSfd_z, LLFpkBfu_x, LLFpkBfu_y, LLFpkBfu_z, LLFpkDf, LLFpkDYf, LLFpktauf, LLFpkphif, LLFnkSfd_x, LLFnkSfd_y, LLFnkSfd_z, LLFnkBfu_x, LLFnkBfu_y, LLFnkBfu_z, LLFnkDf, LLFnkDYf, LLFnktauf, LLFnkphif, LLFp_m3kSfd_x, LLFp_m3kSfd_y, LLFp_m3kSfd_z, LLFp_m3kBfu_x, LLFp_m3kBfu_y, LLFp_m3kBfu_z, LLFp_m3kDf, LLFp_m3kDYf, LLFp_m3ktauf, LLFp_m3kphif, LLFn_m3kSfd_x, LLFn_m3kSfd_y, LLFn_m3kSfd_z, LLFn_m3kBfu_x, LLFn_m3kBfu_y, LLFn_m3kBfu_z, LLFn_m3kDf, LLFn_m3kDYf, LLFn_m3ktauf, LLFn_m3kphif, LLFp_m2kSfd_x, LLFp_m2kSfd_y, LLFp_m2kSfd_z, LLFp_m2kBfu_x, LLFp_m2kBfu_y, LLFp_m2kBfu_z, LLFp_m2kDf, LLFp_m2kDYf, LLFp_m2ktauf, LLFp_m2kphif, LLFn_m2kSfd_x, LLFn_m2kSfd_y, LLFn_m2kSfd_z, LLFn_m2kBfu_x, LLFn_m2kBfu_y, LLFn_m2kBfu_z, LLFn_m2kDf, LLFn_m2kDYf, LLFn_m2ktauf, LLFn_m2kphif, LLFp_m1kSfd_x, LLFp_m1kSfd_y, LLFp_m1kSfd_z, LLFp_m1kBfu_x, LLFp_m1kBfu_y, LLFp_m1kBfu_z, LLFp_m1kDf, LLFp_m1kDYf, LLFp_m1ktauf, LLFp_m1kphif, LLFn_m1kSfd_x, LLFn_m1kSfd_y, LLFn_m1kSfd_z, LLFn_m1kBfu_x, LLFn_m1kBfu_y, LLFn_m1kBfu_z, LLFn_m1kDf, LLFn_m1kDYf, LLFn_m1ktauf, LLFn_m1kphif, LLFp_p1kSfd_x, LLFp_p1kSfd_y, LLFp_p1kSfd_z, LLFp_p1kBfu_x, LLFp_p1kBfu_y, LLFp_p1kBfu_z, LLFp_p1kDf, LLFp_p1kDYf, LLFp_p1ktauf, LLFp_p1kphif, LLFn_p1kSfd_x, LLFn_p1kSfd_y, LLFn_p1kSfd_z, LLFn_p1kBfu_x, LLFn_p1kBfu_y, LLFn_p1kBfu_z, LLFn_p1kDf, LLFn_p1kDYf, LLFn_p1ktauf, LLFn_p1kphif, LLFp_p2kSfd_x, LLFp_p2kSfd_y, LLFp_p2kSfd_z, LLFp_p2kBfu_x, LLFp_p2kBfu_y, LLFp_p2kBfu_z, LLFp_p2kDf, LLFp_p2kDYf, LLFp_p2ktauf, LLFp_p2kphif, LLFn_p2kSfd_x, LLFn_p2kSfd_y, LLFn_p2kSfd_z, LLFn_p2kBfu_x, LLFn_p2kBfu_y, LLFn_p2kBfu_z, LLFn_p2kDf, LLFn_p2kDYf, LLFn_p2ktauf, LLFn_p2kphif, LLFp_p3kSfd_x, LLFp_p3kSfd_y, LLFp_p3kSfd_z, LLFp_p3kBfu_x, LLFp_p3kBfu_y, LLFp_p3kBfu_z, LLFp_p3kDf, LLFp_p3kDYf, LLFp_p3ktauf, LLFp_p3kphif, LLFn_p3kSfd_x, LLFn_p3kSfd_y, LLFn_p3kSfd_z, LLFn_p3kBfu_x, LLFn_p3kBfu_y, LLFn_p3kBfu_z, LLFn_p3kDf, LLFn_p3kDYf, LLFn_p3ktauf, LLFn_p3kphif, Sfd_x_lni, Sfd_y_lni, Sfd_z_lni, Bfu_x_lni, Bfu_y_lni, Bfu_z_lni, Df_lni, DYf_lni, tauf_lni, phif_lni, Sfd_x_lnj, Sfd_y_lnj, Sfd_z_lnj, Bfu_x_lnj, Bfu_y_lnj, Bfu_z_lnj, Df_lnj, DYf_lnj, tauf_lnj, phif_lnj, Sfd_x_lnk, Sfd_y_lnk, Sfd_z_lnk, Bfu_x_lnk, Bfu_y_lnk, Bfu_z_lnk, Df_lnk, DYf_lnk, tauf_lnk, phif_lnk, Sfd_x_rni, Sfd_y_rni, Sfd_z_rni, Bfu_x_rni, Bfu_y_rni, Bfu_z_rni, Df_rni, DYf_rni, tauf_rni, phif_rni, Sfd_x_rnj, Sfd_y_rnj, Sfd_z_rnj, Bfu_x_rnj, Bfu_y_rnj, Bfu_z_rnj, Df_rnj, DYf_rnj, tauf_rnj, phif_rnj, Sfd_x_rnk, Sfd_y_rnk, Sfd_z_rnk, Bfu_x_rnk, Bfu_y_rnk, Bfu_z_rnk, Df_rnk, DYf_rnk, tauf_rnk, phif_rnk, Sfd_x_lpi, Sfd_y_lpi, Sfd_z_lpi, Bfu_x_lpi, Bfu_y_lpi, Bfu_z_lpi, Df_lpi, DYf_lpi, tauf_lpi, phif_lpi, Sfd_x_lpj, Sfd_y_lpj, Sfd_z_lpj, Bfu_x_lpj, Bfu_y_lpj, Bfu_z_lpj, Df_lpj, DYf_lpj, tauf_lpj, phif_lpj, Sfd_x_lpk, Sfd_y_lpk, Sfd_z_lpk, Bfu_x_lpk, Bfu_y_lpk, Bfu_z_lpk, Df_lpk, DYf_lpk, tauf_lpk, phif_lpk, Sfd_x_rpi, Sfd_y_rpi, Sfd_z_rpi, Bfu_x_rpi, Bfu_y_rpi, Bfu_z_rpi, Df_rpi, DYf_rpi, tauf_rpi, phif_rpi, Sfd_x_rpj, Sfd_y_rpj, Sfd_z_rpj, Bfu_x_rpj, Bfu_y_rpj, Bfu_z_rpj, Df_rpj, DYf_rpj, tauf_rpj, phif_rpj, Sfd_x_rpk, Sfd_y_rpk, Sfd_z_rpk, Bfu_x_rpk, Bfu_y_rpk, Bfu_z_rpk, Df_rpk, DYf_rpk, tauf_rpk, phif_rpk, FluxniSfd_x, FluxniSfd_y, FluxniSfd_z, FluxniBfu_x, FluxniBfu_y, FluxniBfu_z, FluxniDf, FluxniDYf, Fluxnitauf, Fluxniphif, FluxnjSfd_x, FluxnjSfd_y, FluxnjSfd_z, FluxnjBfu_x, FluxnjBfu_y, FluxnjBfu_z, FluxnjDf, FluxnjDYf, Fluxnjtauf, Fluxnjphif, FluxnkSfd_x, FluxnkSfd_y, FluxnkSfd_z, FluxnkBfu_x, FluxnkBfu_y, FluxnkBfu_z, FluxnkDf, FluxnkDYf, Fluxnktauf, Fluxnkphif, FluxpiSfd_x, FluxpiSfd_y, FluxpiSfd_z, FluxpiBfu_x, FluxpiBfu_y, FluxpiBfu_z, FluxpiDf, FluxpiDYf, Fluxpitauf, Fluxpiphif, FluxpjSfd_x, FluxpjSfd_y, FluxpjSfd_z, FluxpjBfu_x, FluxpjBfu_y, FluxpjBfu_z, FluxpjDf, FluxpjDYf, Fluxpjtauf, Fluxpjphif, FluxpkSfd_x, FluxpkSfd_y, FluxpkSfd_z, FluxpkBfu_x, FluxpkBfu_y, FluxpkBfu_z, FluxpkDf, FluxpkDYf, Fluxpktauf, Fluxpkphif, d_div_Beta_o0_t0_m0_l0, d_Betau_x_o1_t0_m0_l0, d_Betau_x_o2_t0_m0_l0, d_Sfd_y_o0_t7_m0_l0, d_Betau_x_o1_t1_m0_l0, d_Betau_x_o2_t1_m0_l0, d_Sfd_z_o0_t7_m0_l0, d_Betau_x_o1_t2_m0_l0, d_Betau_x_o2_t2_m0_l0, d_d_div_Beta_x_o0_t0_m0_l0, d_d_div_Beta_y_o0_t0_m0_l0, d_d_div_Beta_z_o0_t0_m0_l0, d_Gamh_x_o0_t15_m0_l0, d_Gamh_x_o0_t16_m0_l0, d_Gamh_x_o0_t17_m0_l0, d_Sfd_x_o0_t8_m0_l0, d_Betau_y_o1_t0_m0_l0, d_Betau_y_o2_t0_m0_l0, d_div_Beta_o0_t1_m0_l0, d_Betau_y_o1_t1_m0_l0, d_Betau_y_o2_t1_m0_l0, d_Sfd_z_o0_t8_m0_l0, d_Betau_y_o1_t2_m0_l0, d_Betau_y_o2_t2_m0_l0, d_Gamh_y_o0_t6_m0_l0, d_d_div_Beta_x_o0_t1_m0_l0, d_Gamh_y_o0_t8_m0_l0, d_d_div_Beta_y_o0_t1_m0_l0, d_d_div_Beta_z_o0_t1_m0_l0, d_Gamh_y_o0_t11_m0_l0, d_Sfd_x_o0_t9_m0_l0, d_Betau_z_o1_t0_m0_l0, d_Betau_z_o2_t0_m0_l0, d_Sfd_y_o0_t9_m0_l0, d_Betau_z_o1_t1_m0_l0, d_Betau_z_o2_t1_m0_l0, d_div_Beta_o0_t2_m0_l0, d_Betau_z_o1_t2_m0_l0, d_Betau_z_o2_t2_m0_l0, d_Gamh_z_o0_t0_m0_l0, d_Gamh_z_o0_t1_m0_l0, d_d_div_Beta_x_o0_t2_m0_l0, d_Gamh_z_o0_t3_m0_l0, d_d_div_Beta_y_o0_t2_m0_l0, d_d_div_Beta_z_o0_t2_m0_l0, d_Ctd_xxx_o0_t0_m0_l0, d_gtd_xx_o1_t0_m0_l0, d_gtd_xx_o2_t0_m0_l0, d_Ctd_xxy_o0_t0_m0_l0, d_gtd_xx_o1_t1_m0_l0, d_gtd_xx_o2_t1_m0_l0, d_Ctd_xxz_o0_t0_m0_l0, d_gtd_xx_o1_t2_m0_l0, d_gtd_xx_o2_t2_m0_l0, d_Ctd_yxx_o0_t0_m0_l0, d_gtd_xy_o1_t0_m0_l0, d_gtd_xy_o2_t0_m0_l0, d_Ctd_xyy_o0_t0_m0_l0, d_gtd_xy_o1_t1_m0_l0, d_gtd_xy_o2_t1_m0_l0, d_Ctd_xyz_o0_t1_m0_l0, d_gtd_xy_o1_t2_m0_l0, d_gtd_xy_o2_t2_m0_l0, d_Ctd_xyy_o0_t1_m0_l0, d_gtd_yy_o1_t0_m0_l0, d_gtd_yy_o2_t0_m0_l0, d_Ctd_yyy_o0_t0_m0_l0, d_gtd_yy_o1_t1_m0_l0, d_gtd_yy_o2_t1_m0_l0, d_Ctd_yyz_o0_t0_m0_l0, d_gtd_yy_o1_t2_m0_l0, d_gtd_yy_o2_t2_m0_l0, d_Ctd_zxx_o0_t0_m0_l0, d_gtd_xz_o1_t0_m0_l0, d_gtd_xz_o2_t0_m0_l0, d_Ctd_xyz_o0_t0_m0_l0, d_gtd_xz_o1_t1_m0_l0, d_gtd_xz_o2_t1_m0_l0, d_Ctd_xzz_o0_t0_m0_l0, d_gtd_xz_o1_t2_m0_l0, d_gtd_xz_o2_t2_m0_l0, d_Ctd_xyz_o0_t2_m0_l0, d_gtd_yz_o1_t0_m0_l0, d_gtd_yz_o2_t0_m0_l0, d_Ctd_zyy_o0_t0_m0_l0, d_gtd_yz_o1_t1_m0_l0, d_gtd_yz_o2_t1_m0_l0, d_Ctd_yzz_o0_t0_m0_l0, d_gtd_yz_o1_t2_m0_l0, d_gtd_yz_o2_t2_m0_l0, d_Ctd_xzz_o0_t1_m0_l0, d_gtd_zz_o1_t0_m0_l0, d_gtd_zz_o2_t0_m0_l0, d_Ctd_yzz_o0_t1_m0_l0, d_gtd_zz_o1_t1_m0_l0, d_gtd_zz_o2_t1_m0_l0, d_Ctd_zzz_o0_t0_m0_l0, d_gtd_zz_o1_t2_m0_l0, d_gtd_zz_o2_t2_m0_l0, d_Rpd_xx_o0_t0_m0_l0, d_chi_o1_t0_m0_l0, d_chi_o2_t0_m0_l0, d_Rpd_xx_o0_t1_m0_l0, d_chi_o1_t1_m0_l0, d_chi_o2_t1_m0_l0, d_Rpd_xx_o0_t2_m0_l0, d_chi_o1_t2_m0_l0, d_chi_o2_t2_m0_l0, d_Rtd_xx_o0_t6_m0_l0, d_Gamh_x_o1_t0_m0_l0, d_Gamh_x_o2_t0_m0_l0, d_Rtd_xy_o0_t6_m0_l0, d_Gamh_x_o1_t1_m0_l0, d_Gamh_x_o2_t1_m0_l0, d_Rtd_xz_o0_t6_m0_l0, d_Gamh_x_o1_t2_m0_l0, d_Gamh_x_o2_t2_m0_l0, d_Rtd_xx_o0_t7_m0_l0, d_Gamh_y_o1_t0_m0_l0, d_Gamh_y_o2_t0_m0_l0, d_Rtd_xy_o0_t7_m0_l0, d_Gamh_y_o1_t1_m0_l0, d_Gamh_y_o2_t1_m0_l0, d_Rtd_xz_o0_t7_m0_l0, d_Gamh_y_o1_t2_m0_l0, d_Gamh_y_o2_t2_m0_l0, d_Rtd_xx_o0_t8_m0_l0, d_Gamh_z_o1_t0_m0_l0, d_Gamh_z_o2_t0_m0_l0, d_Rtd_xy_o0_t8_m0_l0, d_Gamh_z_o1_t1_m0_l0, d_Gamh_z_o2_t1_m0_l0, d_Rtd_xz_o0_t8_m0_l0, d_Gamh_z_o1_t2_m0_l0, d_Gamh_z_o2_t2_m0_l0, d_Psi1_xx_o0_t1_m0_l0, d_Alpha_o1_t0_m0_l0, d_Alpha_o2_t0_m0_l0, d_Psi1_xx_o0_t2_m0_l0, d_Alpha_o1_t1_m0_l0, d_Alpha_o2_t1_m0_l0, d_Psi1_xx_o0_t3_m0_l0, d_Alpha_o1_t2_m0_l0, d_Alpha_o2_t2_m0_l0, d_Psi1_xx_o0_t0_m0_l0, d_Psi1_xy_o0_t0_m0_l0, d_Psi1_xz_o0_t0_m0_l0, d_Psi1_yy_o0_t0_m0_l0, d_Psi1_yz_o0_t0_m0_l0, d_Psi1_zz_o0_t0_m0_l0, d_Atd_xx_o1_t0_m0_l0, d_Atd_xx_o2_t0_m0_l0, d_Atd_xx_o1_t1_m0_l0, d_Atd_xx_o2_t1_m0_l0, d_Atd_xx_o1_t2_m0_l0, d_Atd_xx_o2_t2_m0_l0, d_Atd_xy_o1_t0_m0_l0, d_Atd_xy_o2_t0_m0_l0, d_Atd_xy_o1_t1_m0_l0, d_Atd_xy_o2_t1_m0_l0, d_Atd_xy_o1_t2_m0_l0, d_Atd_xy_o2_t2_m0_l0, d_Atd_xz_o1_t0_m0_l0, d_Atd_xz_o2_t0_m0_l0, d_Atd_xz_o1_t1_m0_l0, d_Atd_xz_o2_t1_m0_l0, d_Atd_xz_o1_t2_m0_l0, d_Atd_xz_o2_t2_m0_l0, d_Atd_yy_o1_t0_m0_l0, d_Atd_yy_o2_t0_m0_l0, d_Atd_yy_o1_t1_m0_l0, d_Atd_yy_o2_t1_m0_l0, d_Atd_yy_o1_t2_m0_l0, d_Atd_yy_o2_t2_m0_l0, d_Atd_yz_o1_t0_m0_l0, d_Atd_yz_o2_t0_m0_l0, d_Atd_yz_o1_t1_m0_l0, d_Atd_yz_o2_t1_m0_l0, d_Atd_yz_o1_t2_m0_l0, d_Atd_yz_o2_t2_m0_l0, d_Atd_zz_o1_t0_m0_l0, d_Atd_zz_o2_t0_m0_l0, d_Atd_zz_o1_t1_m0_l0, d_Atd_zz_o2_t1_m0_l0, d_Atd_zz_o1_t2_m0_l0, d_Atd_zz_o2_t2_m0_l0, d_Gamh_x_o0_t0_m0_l0, d_trK_o1_t0_m0_l0, d_trK_o2_t0_m0_l0, d_Gamh_x_o0_t1_m0_l0, d_trK_o1_t1_m0_l0, d_trK_o2_t1_m0_l0, d_Gamh_x_o0_t2_m0_l0, d_trK_o1_t2_m0_l0, d_trK_o2_t2_m0_l0, d_Gamh_x_o0_t3_m0_l0, d_theta_o1_t0_m0_l0, d_theta_o2_t0_m0_l0, d_Gamh_x_o0_t4_m0_l0, d_theta_o1_t1_m0_l0, d_theta_o2_t1_m0_l0, d_Gamh_x_o0_t5_m0_l0, d_theta_o1_t2_m0_l0, d_theta_o2_t2_m0_l0, d_Rtd_xx_o0_t0_m0_l0, d_Rtd_xx_o0_t1_m0_l0, d_Rtd_xx_o0_t2_m0_l0, d_Rtd_xx_o0_t3_m0_l0, d_Rtd_xx_o0_t4_m0_l0, d_Rtd_xx_o0_t5_m0_l0, d_Rtd_xy_o0_t0_m0_l0, d_Rtd_xy_o0_t1_m0_l0, d_Rtd_xy_o0_t2_m0_l0, d_Rtd_xy_o0_t3_m0_l0, d_Rtd_xy_o0_t4_m0_l0, d_Rtd_xy_o0_t5_m0_l0, d_Rtd_yy_o0_t0_m0_l0, d_Rtd_yy_o0_t1_m0_l0, d_Rtd_yy_o0_t2_m0_l0, d_Rtd_yy_o0_t3_m0_l0, d_Rtd_yy_o0_t4_m0_l0, d_Rtd_yy_o0_t5_m0_l0, d_Rtd_xz_o0_t0_m0_l0, d_Rtd_xz_o0_t1_m0_l0, d_Rtd_xz_o0_t2_m0_l0, d_Rtd_xz_o0_t3_m0_l0, d_Rtd_xz_o0_t4_m0_l0, d_Rtd_xz_o0_t5_m0_l0, d_Rtd_yz_o0_t0_m0_l0, d_Rtd_yz_o0_t1_m0_l0, d_Rtd_yz_o0_t2_m0_l0, d_Rtd_yz_o0_t3_m0_l0, d_Rtd_yz_o0_t4_m0_l0, d_Rtd_yz_o0_t5_m0_l0, d_Rtd_zz_o0_t0_m0_l0, d_Rtd_zz_o0_t1_m0_l0, d_Rtd_zz_o0_t2_m0_l0, d_Rtd_zz_o0_t3_m0_l0, d_Rtd_zz_o0_t4_m0_l0, d_Rtd_zz_o0_t5_m0_l0, d_Rpd_xx_o0_t3_m0_l0, d_Rpd_xx_o0_t4_m0_l0, d_Rpd_xx_o0_t5_m0_l0, d_Rpd_xx_o0_t6_m0_l0, d_Rpd_xx_o0_t7_m0_l0, d_Rpd_xx_o0_t8_m0_l0, rk2m_Ctd_yyz_o0_t0_l0, rk2Ctd_yyz, rk2m_Ctd_yyy_o0_t0_l0, rk2Ctd_yyy, rk2div_Beta, rk2m_Ctd_xxz_o0_t0_l0, rk2Ctd_xxz, rk2m_Ctd_xxx_o0_t0_l0, rk2Ctd_xxx, rk2m_Ctd_yzz_o0_t1_l0, rk2Ctd_yzz, rk2m_Ctd_xxy_o0_t0_l0, rk2Ctd_xxy, rk2kappa_z2, rk2m_Ctd_xyy_o0_t1_l0, rk2Ctd_xyy, rk2m_Ctd_xyz_o0_t2_l0, rk2m_Ctd_xyz_o0_t1_l0, rk2m_Ctd_xyz_o0_t0_l0, rk2Ctd_xyz, rk2d_div_Beta_y, rk2d_div_Beta_x, rk2d_div_Beta_z, rk2m_Ctd_xzz_o0_t1_l0, rk2Ctd_xzz, rk2m_Ctd_zxz_o0_t0_l0, rk2Ctd_zxz, rk2m_Ctd_zxx_o0_t1_l0, rk2Ctd_zxx, rk2m_Ctd_zyy_o0_t1_l0, rk2Ctd_zyy, rk2m_Ctd_zyz_o0_t0_l0, rk2Ctd_zyz, rk2Tf4u_tt, rk2Betatd_x, rk2Betatd_z, rk2Betatd_y, rk2Ct_zyz, rk2Ct_zyy, rk2Atud_zx, rk2Atud_zz, rk2Atud_zy, rk2Atu_zz, rk2Tf4u_tz, rk2Tu_tz, rk2Atud_yy, rk2Atud_yx, rk2Atud_yz, rk2Atu_yz, rk2Atu_yy, rk2Ct_yyy, rk2Ct_yyz, rk2Atud_xx, rk2trAt, rk2Atud_xz, rk2Atud_xy, rk2Atu_xz, rk2Atu_xy, rk2Atu_xx, rk2Tf4u_ty, rk2Tu_ty, rk2Tf4u_tx, rk2Tu_tx, rk2decay_factor, rk2kappa_z1, rk2kappa_f, rk2feta, rk2m_Ctd_zzz_o0_t0_l0, rk2Ctd_zzz, rk2Ct_xzz, rk2Ct_zzz, rk2Ct_yzz, rk2m_Ctd_yxy_o0_t0_l0, rk2Ctd_yxy, rk2Ct_xyz, rk2m_Ctd_yxz_o0_t2_l0, rk2m_Ctd_yxz_o0_t0_l0, rk2Ctd_yxz, rk2Ct_zxz, rk2Ct_yxz, rk2m_Ctd_zxy_o0_t2_l0, rk2Ctd_zxy, rk2Ct_zxy, rk2Ct_yxy, rk2Ct_xxy, rk2Ct_xxz, rk2Ct_xyy, rk2m_Ctd_yxx_o0_t1_l0, rk2Ctd_yxx, rk2Ct_zxx, rk2Gamt_z, rk2Zu_z, rk2Ct_yxx, rk2Gamt_y, rk2Zu_y, rk2Ct_xxx, rk2Gamt_x, rk2Zu_x, rk2d_Rtd_xx_o0_t9_m0_l0, rk2m_Rtd_xx_o0_t8_l0, rk2m_Rtd_xx_o0_t7_l0, rk2m_Rtd_xx_o0_t6_l0, rk2m_Rtd_xx_o0_t5_l0, rk2m_Rtd_xx_o0_t4_l0, rk2m_Rtd_xx_o0_t3_l0, rk2m_Rtd_xx_o0_t2_l0, rk2m_Rtd_xx_o0_t1_l0, rk2m_Rtd_xx_o0_t0_l0, rk2Rtd_xx, rk2d_Rtd_xz_o0_t12_m0_l0, rk2m_Rtd_xz_o0_t11_l0, rk2m_Rtd_xz_o0_t10_l0, rk2m_Rtd_xz_o0_t9_l0, rk2m_Rtd_xz_o0_t8_l0, rk2m_Rtd_xz_o0_t7_l0, rk2m_Rtd_xz_o0_t6_l0, rk2m_Rtd_xz_o0_t5_l0, rk2m_Rtd_xz_o0_t4_l0, rk2m_Rtd_xz_o0_t3_l0, rk2m_Rtd_xz_o0_t2_l0, rk2m_Rtd_xz_o0_t1_l0, rk2m_Rtd_xz_o0_t0_l0, rk2Rtd_xz, rk2d_Rtd_zz_o0_t9_m0_l0, rk2m_Rtd_zz_o0_t8_l0, rk2m_Rtd_zz_o0_t7_l0, rk2m_Rtd_zz_o0_t6_l0, rk2m_Rtd_zz_o0_t5_l0, rk2m_Rtd_zz_o0_t4_l0, rk2m_Rtd_zz_o0_t3_l0, rk2m_Rtd_zz_o0_t2_l0, rk2m_Rtd_zz_o0_t1_l0, rk2m_Rtd_zz_o0_t0_l0, rk2Rtd_zz, rk2d_Rtd_xy_o0_t12_m0_l0, rk2m_Rtd_xy_o0_t11_l0, rk2m_Rtd_xy_o0_t10_l0, rk2m_Rtd_xy_o0_t9_l0, rk2m_Rtd_xy_o0_t8_l0, rk2m_Rtd_xy_o0_t7_l0, rk2m_Rtd_xy_o0_t6_l0, rk2m_Rtd_xy_o0_t5_l0, rk2m_Rtd_xy_o0_t4_l0, rk2m_Rtd_xy_o0_t3_l0, rk2m_Rtd_xy_o0_t2_l0, rk2m_Rtd_xy_o0_t1_l0, rk2m_Rtd_xy_o0_t0_l0, rk2Rtd_xy, rk2d_Rtd_yy_o0_t9_m0_l0, rk2m_Rtd_yy_o0_t8_l0, rk2m_Rtd_yy_o0_t7_l0, rk2m_Rtd_yy_o0_t6_l0, rk2m_Rtd_yy_o0_t5_l0, rk2m_Rtd_yy_o0_t4_l0, rk2m_Rtd_yy_o0_t3_l0, rk2m_Rtd_yy_o0_t2_l0, rk2m_Rtd_yy_o0_t1_l0, rk2m_Rtd_yy_o0_t0_l0, rk2Rtd_yy, rk2d_Rtd_yz_o0_t12_m0_l0, rk2m_Rtd_yz_o0_t11_l0, rk2m_Rtd_yz_o0_t10_l0, rk2m_Rtd_yz_o0_t9_l0, rk2m_Rtd_yz_o0_t8_l0, rk2m_Rtd_yz_o0_t7_l0, rk2m_Rtd_yz_o0_t6_l0, rk2m_Rtd_yz_o0_t5_l0, rk2m_Rtd_yz_o0_t4_l0, rk2m_Rtd_yz_o0_t3_l0, rk2m_Rtd_yz_o0_t2_l0, rk2m_Rtd_yz_o0_t1_l0, rk2m_Rtd_yz_o0_t0_l0, rk2Rtd_yz, rk2m_Rpd_xx_o0_t22_l0, rk2m_Rpd_xx_o0_t21_l0, rk2m_Rpd_xx_o0_t20_l0, rk2m_Rpd_xx_o0_t19_l0, rk2m_Rpd_xx_o0_t18_l0, rk2m_Rpd_xx_o0_t17_l0, rk2m_Rpd_xx_o0_t16_l0, rk2m_Rpd_xx_o0_t15_l0, rk2m_Rpd_xx_o0_t14_l0, rk2m_Rpd_xx_o0_t13_l0, rk2m_Rpd_xx_o0_t12_l0, rk2m_Rpd_xx_o0_t11_l0, rk2m_Rpd_xx_o0_t10_l0, rk2m_Rpd_xx_o0_t9_l0, rk2m_Rpd_xx_o0_t8_l0, rk2m_Rpd_xx_o0_t7_l0, rk2m_Rpd_xx_o0_t6_l0, rk2m_Rpd_xx_o0_t5_l0, rk2m_Rpd_xx_o0_t4_l0, rk2m_Rpd_xx_o0_t3_l0, rk2m_Rpd_xx_o0_t2_l0, rk2m_Rpd_xx_o0_t1_l0, rk2m_Rpd_xx_o0_t0_l0, rk2Rpd_xx, rk2m_Rpd_yy_o0_t22_l0, rk2m_Rpd_yy_o0_t21_l0, rk2m_Rpd_yy_o0_t20_l0, rk2m_Rpd_yy_o0_t19_l0, rk2m_Rpd_yy_o0_t18_l0, rk2m_Rpd_yy_o0_t17_l0, rk2m_Rpd_yy_o0_t16_l0, rk2m_Rpd_yy_o0_t15_l0, rk2m_Rpd_yy_o0_t14_l0, rk2m_Rpd_yy_o0_t13_l0, rk2m_Rpd_yy_o0_t12_l0, rk2m_Rpd_yy_o0_t11_l0, rk2m_Rpd_yy_o0_t10_l0, rk2m_Rpd_yy_o0_t9_l0, rk2m_Rpd_yy_o0_t8_l0, rk2m_Rpd_yy_o0_t7_l0, rk2m_Rpd_yy_o0_t6_l0, rk2m_Rpd_yy_o0_t5_l0, rk2m_Rpd_yy_o0_t4_l0, rk2m_Rpd_yy_o0_t3_l0, rk2m_Rpd_yy_o0_t2_l0, rk2m_Rpd_yy_o0_t1_l0, rk2m_Rpd_yy_o0_t0_l0, rk2Rpd_yy, rk2m_Rpd_yz_o0_t25_l0, rk2m_Rpd_yz_o0_t24_l0, rk2m_Rpd_yz_o0_t23_l0, rk2m_Rpd_yz_o0_t22_l0, rk2m_Rpd_yz_o0_t21_l0, rk2m_Rpd_yz_o0_t20_l0, rk2m_Rpd_yz_o0_t19_l0, rk2m_Rpd_yz_o0_t18_l0, rk2m_Rpd_yz_o0_t17_l0, rk2m_Rpd_yz_o0_t16_l0, rk2m_Rpd_yz_o0_t15_l0, rk2m_Rpd_yz_o0_t14_l0, rk2m_Rpd_yz_o0_t13_l0, rk2m_Rpd_yz_o0_t12_l0, rk2m_Rpd_yz_o0_t11_l0, rk2m_Rpd_yz_o0_t10_l0, rk2m_Rpd_yz_o0_t9_l0, rk2m_Rpd_yz_o0_t8_l0, rk2m_Rpd_yz_o0_t7_l0, rk2m_Rpd_yz_o0_t6_l0, rk2m_Rpd_yz_o0_t5_l0, rk2m_Rpd_yz_o0_t4_l0, rk2m_Rpd_yz_o0_t3_l0, rk2m_Rpd_yz_o0_t2_l0, rk2m_Rpd_yz_o0_t1_l0, rk2m_Rpd_yz_o0_t0_l0, rk2Rpd_yz, rk2m_Rpd_xz_o0_t25_l0, rk2m_Rpd_xz_o0_t24_l0, rk2m_Rpd_xz_o0_t23_l0, rk2m_Rpd_xz_o0_t22_l0, rk2m_Rpd_xz_o0_t21_l0, rk2m_Rpd_xz_o0_t20_l0, rk2m_Rpd_xz_o0_t19_l0, rk2m_Rpd_xz_o0_t18_l0, rk2m_Rpd_xz_o0_t17_l0, rk2m_Rpd_xz_o0_t16_l0, rk2m_Rpd_xz_o0_t15_l0, rk2m_Rpd_xz_o0_t14_l0, rk2m_Rpd_xz_o0_t13_l0, rk2m_Rpd_xz_o0_t12_l0, rk2m_Rpd_xz_o0_t11_l0, rk2m_Rpd_xz_o0_t10_l0, rk2m_Rpd_xz_o0_t9_l0, rk2m_Rpd_xz_o0_t8_l0, rk2m_Rpd_xz_o0_t7_l0, rk2m_Rpd_xz_o0_t6_l0, rk2m_Rpd_xz_o0_t5_l0, rk2m_Rpd_xz_o0_t4_l0, rk2m_Rpd_xz_o0_t3_l0, rk2m_Rpd_xz_o0_t2_l0, rk2m_Rpd_xz_o0_t1_l0, rk2m_Rpd_xz_o0_t0_l0, rk2Rpd_xz, rk2m_Rpd_xy_o0_t25_l0, rk2m_Rpd_xy_o0_t24_l0, rk2m_Rpd_xy_o0_t23_l0, rk2m_Rpd_xy_o0_t22_l0, rk2m_Rpd_xy_o0_t21_l0, rk2m_Rpd_xy_o0_t20_l0, rk2m_Rpd_xy_o0_t19_l0, rk2m_Rpd_xy_o0_t18_l0, rk2m_Rpd_xy_o0_t17_l0, rk2m_Rpd_xy_o0_t16_l0, rk2m_Rpd_xy_o0_t15_l0, rk2m_Rpd_xy_o0_t14_l0, rk2m_Rpd_xy_o0_t13_l0, rk2m_Rpd_xy_o0_t12_l0, rk2m_Rpd_xy_o0_t11_l0, rk2m_Rpd_xy_o0_t10_l0, rk2m_Rpd_xy_o0_t9_l0, rk2m_Rpd_xy_o0_t8_l0, rk2m_Rpd_xy_o0_t7_l0, rk2m_Rpd_xy_o0_t6_l0, rk2m_Rpd_xy_o0_t5_l0, rk2m_Rpd_xy_o0_t4_l0, rk2m_Rpd_xy_o0_t3_l0, rk2m_Rpd_xy_o0_t2_l0, rk2m_Rpd_xy_o0_t1_l0, rk2m_Rpd_xy_o0_t0_l0, rk2Rpd_xy, rk2m_Rpd_zz_o0_t22_l0, rk2m_Rpd_zz_o0_t21_l0, rk2m_Rpd_zz_o0_t20_l0, rk2m_Rpd_zz_o0_t19_l0, rk2m_Rpd_zz_o0_t18_l0, rk2m_Rpd_zz_o0_t17_l0, rk2m_Rpd_zz_o0_t16_l0, rk2m_Rpd_zz_o0_t15_l0, rk2m_Rpd_zz_o0_t14_l0, rk2m_Rpd_zz_o0_t13_l0, rk2m_Rpd_zz_o0_t12_l0, rk2m_Rpd_zz_o0_t11_l0, rk2m_Rpd_zz_o0_t10_l0, rk2m_Rpd_zz_o0_t9_l0, rk2m_Rpd_zz_o0_t8_l0, rk2m_Rpd_zz_o0_t7_l0, rk2m_Rpd_zz_o0_t6_l0, rk2m_Rpd_zz_o0_t5_l0, rk2m_Rpd_zz_o0_t4_l0, rk2m_Rpd_zz_o0_t3_l0, rk2m_Rpd_zz_o0_t2_l0, rk2m_Rpd_zz_o0_t1_l0, rk2m_Rpd_zz_o0_t0_l0, rk2Rpd_zz, rk2Rscalar, rk2Tf4u_zz, rk2Tu_zz, rk2Tf4u_yz, rk2Tu_yz, rk2Tf4u_yy, rk2Tu_yy, rk2Tf4u_xy, rk2Tu_xy, rk2Tf4u_xx, rk2Tu_xx, rk2Tf4u_xz, rk2Tu_xz, rk2trSf, rk2Tu_tt, rk2pTtd_ADM_xz, rk2pTtd_ADM_xy, rk2pTtd_ADM_xx, rk2d_Psi1_xx_o0_t5_m0_l0, rk2m_Psi1_xx_o0_t4_l0, rk2m_Psi1_xx_o0_t3_l0, rk2m_Psi1_xx_o0_t2_l0, rk2m_Psi1_xx_o0_t1_l0, rk2m_Psi1_xx_o0_t0_l0, rk2Psi1_xx, rk2d_Psi1_xy_o0_t6_m0_l0, rk2m_Psi1_xy_o0_t5_l0, rk2m_Psi1_xy_o0_t4_l0, rk2m_Psi1_xy_o0_t3_l0, rk2m_Psi1_xy_o0_t2_l0, rk2m_Psi1_xy_o0_t1_l0, rk2m_Psi1_xy_o0_t0_l0, rk2Psi1_xy, rk2d_Psi1_xz_o0_t6_m0_l0, rk2m_Psi1_xz_o0_t5_l0, rk2m_Psi1_xz_o0_t4_l0, rk2m_Psi1_xz_o0_t3_l0, rk2m_Psi1_xz_o0_t2_l0, rk2m_Psi1_xz_o0_t1_l0, rk2m_Psi1_xz_o0_t0_l0, rk2Psi1_xz, rk2Jtd_ADM_z, rk2pTtd_ADM_zz, rk2d_Psi1_zz_o0_t5_m0_l0, rk2m_Psi1_zz_o0_t4_l0, rk2m_Psi1_zz_o0_t3_l0, rk2m_Psi1_zz_o0_t2_l0, rk2m_Psi1_zz_o0_t1_l0, rk2m_Psi1_zz_o0_t0_l0, rk2Psi1_zz, rk2Jtd_ADM_y, rk2Jtd_ADM_x, rk2pTtd_ADM_yz, rk2d_Psi1_yz_o0_t6_m0_l0, rk2m_Psi1_yz_o0_t5_l0, rk2m_Psi1_yz_o0_t4_l0, rk2m_Psi1_yz_o0_t3_l0, rk2m_Psi1_yz_o0_t2_l0, rk2m_Psi1_yz_o0_t1_l0, rk2m_Psi1_yz_o0_t0_l0, rk2Psi1_yz, rk2pTtd_ADM_yy, rk2tr_pT, rk2d_Psi1_yy_o0_t5_m0_l0, rk2m_Psi1_yy_o0_t4_l0, rk2m_Psi1_yy_o0_t3_l0, rk2m_Psi1_yy_o0_t2_l0, rk2m_Psi1_yy_o0_t1_l0, rk2m_Psi1_yy_o0_t0_l0, rk2Psi1_yy, rk2trPsi1, rk2Psi1TF_yz, rk2Psi1TF_yy, rk2Psi1TF_xy, rk2Psi1TF_xx, rk2Psi1TF_xz, rk2Psi1TF_zz, rk2kappa_cc, rk2rho_ADM, d_Sfd_x_o0_t11_m0_l0, d_Sfd_y_o0_t11_m0_l0, d_Sfd_z_o0_t11_m0_l0, d_DYf_o0_t0_m0_l0, d_tauf_o0_t3_m0_l0, d_gtd_xx_o0_t3_m0_l0, d_gtd_xy_o0_t6_m0_l0, d_gtd_xz_o0_t6_m0_l0, d_gtd_yy_o0_t3_m0_l0, d_gtd_yz_o0_t6_m0_l0, d_gtd_zz_o0_t3_m0_l0, d_Atd_xx_o0_t3_m0_l0, d_Atd_xy_o0_t6_m0_l0, d_Atd_xz_o0_t6_m0_l0, d_Atd_yy_o0_t3_m0_l0, d_Atd_yz_o0_t6_m0_l0, d_Atd_zz_o0_t3_m0_l0, d_Gamh_x_o0_t24_m0_l0, d_Gamh_y_o0_t24_m0_l0, d_Gamh_z_o0_t24_m0_l0, d_Betau_x_o0_t0_m0_l0, d_Betau_y_o0_t0_m0_l0, d_Betau_z_o0_t0_m0_l0, d_Alpha_o0_t0_m0_l0, d_chi_o0_t0_m0_l0, d_trK_o0_t21_m0_l0, d_theta_o0_t3_m0_l0, RHS_Bfu_x, RHS_Bfu_y, RHS_Bfu_z, RHS_tauf, RHS_phif, RHS_Sfd_x, RHS_Sfd_y, RHS_Sfd_z, RHS_Df, RHS_DYf, m_theta_o2_t2_l0, m_theta_o2_t1_l0, m_theta_o2_t0_l0, m_theta_o1_t2_l0, m_theta_o1_t1_l0, m_theta_o1_t0_l0, m_theta_o0_t2_l0, m_theta_o0_t1_l0, m_theta_o0_t0_l0, m_trK_o2_t2_l0, m_trK_o2_t1_l0, m_trK_o2_t0_l0, m_trK_o1_t2_l0, m_trK_o1_t1_l0, m_trK_o1_t0_l0, m_trK_o0_t20_l0, m_trK_o0_t19_l0, m_trK_o0_t18_l0, m_trK_o0_t17_l0, m_trK_o0_t16_l0, m_trK_o0_t15_l0, m_trK_o0_t14_l0, m_trK_o0_t13_l0, m_trK_o0_t12_l0, m_trK_o0_t11_l0, m_trK_o0_t10_l0, m_trK_o0_t9_l0, m_trK_o0_t8_l0, m_trK_o0_t7_l0, m_trK_o0_t6_l0, m_trK_o0_t5_l0, m_trK_o0_t4_l0, m_trK_o0_t3_l0, m_trK_o0_t2_l0, m_trK_o0_t1_l0, m_trK_o0_t0_l0, m_chi_o2_t2_l0, m_chi_o2_t1_l0, m_chi_o2_t0_l0, m_chi_o1_t2_l0, m_chi_o1_t1_l0, m_chi_o1_t0_l0, m_Alpha_o2_t2_l0, m_Alpha_o2_t1_l0, m_Alpha_o2_t0_l0, m_Alpha_o1_t2_l0, m_Alpha_o1_t1_l0, m_Alpha_o1_t0_l0, m_Betau_z_o2_t2_l0, m_Betau_z_o2_t1_l0, m_Betau_z_o2_t0_l0, m_Betau_z_o1_t2_l0, m_Betau_z_o1_t1_l0, m_Betau_z_o1_t0_l0, m_Betau_y_o2_t2_l0, m_Betau_y_o2_t1_l0, m_Betau_y_o2_t0_l0, m_Betau_y_o1_t2_l0, m_Betau_y_o1_t1_l0, m_Betau_y_o1_t0_l0, m_Betau_x_o2_t2_l0, m_Betau_x_o2_t1_l0, m_Betau_x_o2_t0_l0, m_Betau_x_o1_t2_l0, m_Betau_x_o1_t1_l0, m_Betau_x_o1_t0_l0, m_Gamh_z_o2_t2_l0, m_Gamh_z_o2_t1_l0, m_Gamh_z_o2_t0_l0, m_Gamh_z_o1_t2_l0, m_Gamh_z_o1_t1_l0, m_Gamh_z_o1_t0_l0, m_Gamh_z_o0_t23_l0, m_Gamh_z_o0_t22_l0, m_Gamh_z_o0_t21_l0, m_Gamh_z_o0_t20_l0, m_Gamh_z_o0_t19_l0, m_Gamh_z_o0_t18_l0, m_Gamh_z_o0_t17_l0, m_Gamh_z_o0_t16_l0, m_Gamh_z_o0_t15_l0, m_Gamh_z_o0_t14_l0, m_Gamh_z_o0_t13_l0, m_Gamh_z_o0_t12_l0, m_Gamh_z_o0_t11_l0, m_Gamh_z_o0_t10_l0, m_Gamh_z_o0_t9_l0, m_Gamh_z_o0_t8_l0, m_Gamh_z_o0_t7_l0, m_Gamh_z_o0_t6_l0, m_Gamh_z_o0_t5_l0, m_Gamh_z_o0_t4_l0, m_Gamh_z_o0_t3_l0, m_Gamh_z_o0_t2_l0, m_Gamh_z_o0_t1_l0, m_Gamh_z_o0_t0_l0, m_Gamh_y_o2_t2_l0, m_Gamh_y_o2_t1_l0, m_Gamh_y_o2_t0_l0, m_Gamh_y_o1_t2_l0, m_Gamh_y_o1_t1_l0, m_Gamh_y_o1_t0_l0, m_Gamh_y_o0_t23_l0, m_Gamh_y_o0_t22_l0, m_Gamh_y_o0_t21_l0, m_Gamh_y_o0_t20_l0, m_Gamh_y_o0_t19_l0, m_Gamh_y_o0_t18_l0, m_Gamh_y_o0_t17_l0, m_Gamh_y_o0_t16_l0, m_Gamh_y_o0_t15_l0, m_Gamh_y_o0_t14_l0, m_Gamh_y_o0_t13_l0, m_Gamh_y_o0_t12_l0, m_Gamh_y_o0_t11_l0, m_Gamh_y_o0_t10_l0, m_Gamh_y_o0_t9_l0, m_Gamh_y_o0_t8_l0, m_Gamh_y_o0_t7_l0, m_Gamh_y_o0_t6_l0, m_Gamh_y_o0_t5_l0, m_Gamh_y_o0_t4_l0, m_Gamh_y_o0_t3_l0, m_Gamh_y_o0_t2_l0, m_Gamh_y_o0_t1_l0, m_Gamh_y_o0_t0_l0, m_Gamh_x_o2_t2_l0, m_Gamh_x_o2_t1_l0, m_Gamh_x_o2_t0_l0, m_Gamh_x_o1_t2_l0, m_Gamh_x_o1_t1_l0, m_Gamh_x_o1_t0_l0, m_Gamh_x_o0_t23_l0, m_Gamh_x_o0_t22_l0, m_Gamh_x_o0_t21_l0, m_Gamh_x_o0_t20_l0, m_Gamh_x_o0_t19_l0, m_Gamh_x_o0_t18_l0, m_Gamh_x_o0_t17_l0, m_Gamh_x_o0_t16_l0, m_Gamh_x_o0_t15_l0, m_Gamh_x_o0_t14_l0, m_Gamh_x_o0_t13_l0, m_Gamh_x_o0_t12_l0, m_Gamh_x_o0_t11_l0, m_Gamh_x_o0_t10_l0, m_Gamh_x_o0_t9_l0, m_Gamh_x_o0_t8_l0, m_Gamh_x_o0_t7_l0, m_Gamh_x_o0_t6_l0, m_Gamh_x_o0_t5_l0, m_Gamh_x_o0_t4_l0, m_Gamh_x_o0_t3_l0, m_Gamh_x_o0_t2_l0, m_Gamh_x_o0_t1_l0, m_Gamh_x_o0_t0_l0, m_Atd_zz_o2_t2_l0, m_Atd_zz_o2_t1_l0, m_Atd_zz_o2_t0_l0, m_Atd_zz_o1_t2_l0, m_Atd_zz_o1_t1_l0, m_Atd_zz_o1_t0_l0, m_Atd_zz_o0_t2_l0, m_Atd_zz_o0_t1_l0, m_Atd_zz_o0_t0_l0, m_Atd_yz_o2_t2_l0, m_Atd_yz_o2_t1_l0, m_Atd_yz_o2_t0_l0, m_Atd_yz_o1_t2_l0, m_Atd_yz_o1_t1_l0, m_Atd_yz_o1_t0_l0, m_Atd_yz_o0_t5_l0, m_Atd_yz_o0_t4_l0, m_Atd_yz_o0_t3_l0, m_Atd_yz_o0_t2_l0, m_Atd_yz_o0_t1_l0, m_Atd_yz_o0_t0_l0, m_Atd_yy_o2_t2_l0, m_Atd_yy_o2_t1_l0, m_Atd_yy_o2_t0_l0, m_Atd_yy_o1_t2_l0, m_Atd_yy_o1_t1_l0, m_Atd_yy_o1_t0_l0, m_Atd_yy_o0_t2_l0, m_Atd_yy_o0_t1_l0, m_Atd_yy_o0_t0_l0, m_Atd_xz_o2_t2_l0, m_Atd_xz_o2_t1_l0, m_Atd_xz_o2_t0_l0, m_Atd_xz_o1_t2_l0, m_Atd_xz_o1_t1_l0, m_Atd_xz_o1_t0_l0, m_Atd_xz_o0_t5_l0, m_Atd_xz_o0_t4_l0, m_Atd_xz_o0_t3_l0, m_Atd_xz_o0_t2_l0, m_Atd_xz_o0_t1_l0, m_Atd_xz_o0_t0_l0, m_Atd_xy_o2_t2_l0, m_Atd_xy_o2_t1_l0, m_Atd_xy_o2_t0_l0, m_Atd_xy_o1_t2_l0, m_Atd_xy_o1_t1_l0, m_Atd_xy_o1_t0_l0, m_Atd_xy_o0_t5_l0, m_Atd_xy_o0_t4_l0, m_Atd_xy_o0_t3_l0, m_Atd_xy_o0_t2_l0, m_Atd_xy_o0_t1_l0, m_Atd_xy_o0_t0_l0, m_Atd_xx_o2_t2_l0, m_Atd_xx_o2_t1_l0, m_Atd_xx_o2_t0_l0, m_Atd_xx_o1_t2_l0, m_Atd_xx_o1_t1_l0, m_Atd_xx_o1_t0_l0, m_Atd_xx_o0_t2_l0, m_Atd_xx_o0_t1_l0, m_Atd_xx_o0_t0_l0, m_gtd_zz_o2_t2_l0, m_gtd_zz_o2_t1_l0, m_gtd_zz_o2_t0_l0, m_gtd_zz_o1_t2_l0, m_gtd_zz_o1_t1_l0, m_gtd_zz_o1_t0_l0, m_gtd_zz_o0_t2_l0, m_gtd_zz_o0_t1_l0, m_gtd_zz_o0_t0_l0, m_gtd_yz_o2_t2_l0, m_gtd_yz_o2_t1_l0, m_gtd_yz_o2_t0_l0, m_gtd_yz_o1_t2_l0, m_gtd_yz_o1_t1_l0, m_gtd_yz_o1_t0_l0, m_gtd_yz_o0_t5_l0, m_gtd_yz_o0_t4_l0, m_gtd_yz_o0_t3_l0, m_gtd_yz_o0_t2_l0, m_gtd_yz_o0_t1_l0, m_gtd_yz_o0_t0_l0, m_gtd_yy_o2_t2_l0, m_gtd_yy_o2_t1_l0, m_gtd_yy_o2_t0_l0, m_gtd_yy_o1_t2_l0, m_gtd_yy_o1_t1_l0, m_gtd_yy_o1_t0_l0, m_gtd_yy_o0_t2_l0, m_gtd_yy_o0_t1_l0, m_gtd_yy_o0_t0_l0, m_gtd_xz_o2_t2_l0, m_gtd_xz_o2_t1_l0, m_gtd_xz_o2_t0_l0, m_gtd_xz_o1_t2_l0, m_gtd_xz_o1_t1_l0, m_gtd_xz_o1_t0_l0, m_gtd_xz_o0_t5_l0, m_gtd_xz_o0_t4_l0, m_gtd_xz_o0_t3_l0, m_gtd_xz_o0_t2_l0, m_gtd_xz_o0_t1_l0, m_gtd_xz_o0_t0_l0, m_gtd_xy_o2_t2_l0, m_gtd_xy_o2_t1_l0, m_gtd_xy_o2_t0_l0, m_gtd_xy_o1_t2_l0, m_gtd_xy_o1_t1_l0, m_gtd_xy_o1_t0_l0, m_gtd_xy_o0_t5_l0, m_gtd_xy_o0_t4_l0, m_gtd_xy_o0_t3_l0, m_gtd_xy_o0_t2_l0, m_gtd_xy_o0_t1_l0, m_gtd_xy_o0_t0_l0, m_gtd_xx_o2_t2_l0, m_gtd_xx_o2_t1_l0, m_gtd_xx_o2_t0_l0, m_gtd_xx_o1_t2_l0, m_gtd_xx_o1_t1_l0, m_gtd_xx_o1_t0_l0, m_gtd_xx_o0_t2_l0, m_gtd_xx_o0_t1_l0, m_gtd_xx_o0_t0_l0, m_phif_o0_t2_l0, m_phif_o0_t1_l0, m_phif_o0_t0_l0, m_tauf_o0_t2_l0, m_tauf_o0_t1_l0, m_tauf_o0_t0_l0, m_Bfu_z_o0_t5_l0, m_Bfu_z_o0_t4_l0, m_Bfu_z_o0_t3_l0, m_Bfu_z_o0_t2_l0, m_Bfu_z_o0_t1_l0, m_Bfu_z_o0_t0_l0, m_Bfu_y_o0_t5_l0, m_Bfu_y_o0_t4_l0, m_Bfu_y_o0_t3_l0, m_Bfu_y_o0_t2_l0, m_Bfu_y_o0_t1_l0, m_Bfu_y_o0_t0_l0, m_Bfu_x_o0_t5_l0, m_Bfu_x_o0_t4_l0, m_Bfu_x_o0_t3_l0, m_Bfu_x_o0_t2_l0, m_Bfu_x_o0_t1_l0, m_Bfu_x_o0_t0_l0, m_Sfd_z_o0_t10_l0, m_Sfd_z_o0_t9_l0, m_Sfd_z_o0_t8_l0, m_Sfd_z_o0_t7_l0, m_Sfd_z_o0_t6_l0, m_Sfd_z_o0_t5_l0, m_Sfd_z_o0_t4_l0, m_Sfd_z_o0_t3_l0, m_Sfd_z_o0_t2_l0, m_Sfd_z_o0_t1_l0, m_Sfd_z_o0_t0_l0, m_Sfd_y_o0_t10_l0, m_Sfd_y_o0_t9_l0, m_Sfd_y_o0_t8_l0, m_Sfd_y_o0_t7_l0, m_Sfd_y_o0_t6_l0, m_Sfd_y_o0_t5_l0, m_Sfd_y_o0_t4_l0, m_Sfd_y_o0_t3_l0, m_Sfd_y_o0_t2_l0, m_Sfd_y_o0_t1_l0, m_Sfd_y_o0_t0_l0, m_Sfd_x_o0_t10_l0, m_Sfd_x_o0_t9_l0, m_Sfd_x_o0_t8_l0, m_Sfd_x_o0_t7_l0, m_Sfd_x_o0_t6_l0, m_Sfd_x_o0_t5_l0, m_Sfd_x_o0_t4_l0, m_Sfd_x_o0_t3_l0, m_Sfd_x_o0_t2_l0, m_Sfd_x_o0_t1_l0, m_Sfd_x_o0_t0_l0, RHS_gtd_xx, RHS_gtd_xy, RHS_gtd_xz, RHS_gtd_yy, RHS_gtd_yz, RHS_gtd_zz, RHS_Atd_xx, RHS_Atd_xy, RHS_Atd_xz, RHS_Atd_yy, RHS_Atd_yz, RHS_Atd_zz, RHS_Gamh_x, RHS_Gamh_y, RHS_Gamh_z, RHS_Betau_x, RHS_Betau_y, RHS_Betau_z, RHS_Alpha, RHS_chi, RHS_trK, RHS_theta, n_x, n_y, n_z, interaction_index, mod_normal, i_d_Atd_xx_o0_t0_m0_l0, i_d_Atd_xx_o0_t1_m0_l0, i_d_Atd_xx_o0_t2_m0_l0, i_d_Atd_xy_o0_t0_m0_l0, i_d_Atd_xy_o0_t1_m0_l0, i_d_Atd_xy_o0_t2_m0_l0, i_d_Atd_xz_o0_t0_m0_l0, i_d_Atd_xz_o0_t1_m0_l0, i_d_Atd_xz_o0_t2_m0_l0, i_d_Atd_yy_o0_t0_m0_l0, i_d_Atd_yy_o0_t1_m0_l0, i_d_Atd_yy_o0_t2_m0_l0, i_d_Atd_yz_o0_t0_m0_l0, i_d_Atd_yz_o0_t1_m0_l0, i_d_Atd_yz_o0_t2_m0_l0, i_d_Atd_zz_o0_t0_m0_l0, i_d_Atd_zz_o0_t1_m0_l0, i_d_Atd_zz_o0_t2_m0_l0, i_d_gtd_xx_o0_t3_m0_l0, i_d_gtd_xy_o0_t3_m0_l0, i_d_gtd_xz_o0_t3_m0_l0, i_d_gtd_yy_o0_t3_m0_l0, i_d_gtd_yz_o0_t3_m0_l0, i_d_gtd_zz_o0_t3_m0_l0, i_d_Atd_xx_o0_t3_m0_l0, i_d_Atd_xy_o0_t3_m0_l0, i_d_Atd_xz_o0_t3_m0_l0, i_d_Atd_yy_o0_t3_m0_l0, i_d_Atd_yz_o0_t3_m0_l0, i_d_Atd_zz_o0_t3_m0_l0, i_d_Gamh_x_o0_t3_m0_l0, i_d_Gamh_y_o0_t3_m0_l0, i_d_Gamh_z_o0_t3_m0_l0, i_d_Betau_x_o0_t3_m0_l0, i_d_Betau_y_o0_t3_m0_l0, i_d_Betau_z_o0_t3_m0_l0, i_d_Alpha_o0_t3_m0_l0, i_d_chi_o0_t3_m0_l0, i_d_trK_o0_t3_m0_l0, i_d_theta_o0_t3_m0_l0, i_m_gtd_xx_o0_t0_l0, i_m_gtd_xx_o0_t1_l0, i_m_gtd_xx_o0_t2_l0, i_m_gtd_xy_o0_t0_l0, i_m_gtd_xy_o0_t1_l0, i_m_gtd_xy_o0_t2_l0, i_m_gtd_xz_o0_t0_l0, i_m_gtd_xz_o0_t1_l0, i_m_gtd_xz_o0_t2_l0, i_m_gtd_yy_o0_t0_l0, i_m_gtd_yy_o0_t1_l0, i_m_gtd_yy_o0_t2_l0, i_m_gtd_yz_o0_t0_l0, i_m_gtd_yz_o0_t1_l0, i_m_gtd_yz_o0_t2_l0, i_m_gtd_zz_o0_t0_l0, i_m_gtd_zz_o0_t1_l0, i_m_gtd_zz_o0_t2_l0, i_m_Atd_xx_o0_t0_l0, i_m_Atd_xx_o0_t1_l0, i_m_Atd_xx_o0_t2_l0, i_m_Atd_xy_o0_t0_l0, i_m_Atd_xy_o0_t1_l0, i_m_Atd_xy_o0_t2_l0, i_m_Atd_xz_o0_t0_l0, i_m_Atd_xz_o0_t1_l0, i_m_Atd_xz_o0_t2_l0, i_m_Atd_yy_o0_t0_l0, i_m_Atd_yy_o0_t1_l0, i_m_Atd_yy_o0_t2_l0, i_m_Atd_yz_o0_t0_l0, i_m_Atd_yz_o0_t1_l0, i_m_Atd_yz_o0_t2_l0, i_m_Atd_zz_o0_t0_l0, i_m_Atd_zz_o0_t1_l0, i_m_Atd_zz_o0_t2_l0, i_m_Gamh_x_o0_t0_l0, i_m_Gamh_x_o0_t1_l0, i_m_Gamh_x_o0_t2_l0, i_m_Gamh_y_o0_t0_l0, i_m_Gamh_y_o0_t1_l0, i_m_Gamh_y_o0_t2_l0, i_m_Gamh_z_o0_t0_l0, i_m_Gamh_z_o0_t1_l0, i_m_Gamh_z_o0_t2_l0, i_m_Betau_x_o0_t0_l0, i_m_Betau_x_o0_t1_l0, i_m_Betau_x_o0_t2_l0, i_m_Betau_y_o0_t0_l0, i_m_Betau_y_o0_t1_l0, i_m_Betau_y_o0_t2_l0, i_m_Betau_z_o0_t0_l0, i_m_Betau_z_o0_t1_l0, i_m_Betau_z_o0_t2_l0, i_m_Alpha_o0_t0_l0, i_m_Alpha_o0_t1_l0, i_m_Alpha_o0_t2_l0, i_m_chi_o0_t0_l0, i_m_chi_o0_t1_l0, i_m_chi_o0_t2_l0, i_m_trK_o0_t0_l0, i_m_trK_o0_t1_l0, i_m_trK_o0_t2_l0, i_m_theta_o0_t0_l0, i_m_theta_o0_t1_l0, i_m_theta_o0_t2_l0;
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					if ((i + 3 < ilast && i - 3 >= 0 && j + 3 < jlast && j - 3 >= 0 && k + 3 < klast && k - 3 >= 0)) {
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_yy, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_zz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_xy, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_xz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2sqBf = vector(rk2Bfu_x, i, j, k) * rk2Bfd_x + vector(rk2Bfu_y, i, j, k) * rk2Bfd_y + vector(rk2Bfu_z, i, j, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) * vector(rk2gtd_xx, i, j, k) + (-vector(rk2gtd_yy, i, j, k) * (vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k))) + (-vector(rk2gtd_zz, i, j, k) * (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k))) + (-(vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k)) * vector(rk2gtd_xx, i, j, k)) + 2.0 * vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yy, i, j, k) - vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) - vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_yz, i, j, k) - vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_zz, i, j, k)) + vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j, k) + rk2gtu_xy * vector(rk2Sfd_y, i, j, k) + rk2gtu_xz * vector(rk2Sfd_z, i, j, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j, k) + rk2gtu_xy * vector(vfd_y, i, j, k) + rk2gtu_xz * vector(vfd_z, i, j, k));
						rk2Bfvf = vector(rk2Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk2Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk2Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_zz, i, j, k) - vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yz, i, j, k)) + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j, k) + rk2gtu_yz * vector(vfd_y, i, j, k) + rk2gtu_zz * vector(vfd_z, i, j, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j, k) + rk2gtu_yy * vector(rk2Sfd_y, i, j, k) + rk2gtu_yz * vector(rk2Sfd_z, i, j, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j, k) + rk2gtu_yz * vector(rk2Sfd_y, i, j, k) + rk2gtu_zz * vector(rk2Sfd_z, i, j, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j, k) + rk2gtu_yy * vector(vfd_y, i, j, k) + rk2gtu_yz * vector(vfd_z, i, j, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j, k) + rk2vfu_y * vector(vfd_y, i, j, k) + rk2vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j, k) * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j, k) * vector(rk2Bfu_x, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k) * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k) * rk2vfu_z + vector(rk2Bfu_z, i, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k) * vector(rk2Bfu_y, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k) * rk2vfu_y + vector(rk2Bfu_y, i, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k) * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k) * rk2vfu_z + vector(rk2Bfu_z, i, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j, k) * vector(rk2Bfu_y, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j, k));
						FluxiSfd_x = FiSfd_x_mainI(vector(rk2Sfd_x, i, j, k), vector(rk2gtd_xx, i, j, k), vector(rk2gtd_xy, i, j, k), vector(rk2gtd_xz, i, j, k), vector(rk2Betau_x, i, j, k), vector(rk2Alpha, i, j, k), vector(TauT_xx, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_xz, i, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xx, dx, simPlat_dt, ilast, jlast);
						FluxiSfd_y = FiSfd_y_mainI(vector(rk2Sfd_y, i, j, k), vector(rk2gtd_xy, i, j, k), vector(rk2gtd_yy, i, j, k), vector(rk2gtd_yz, i, j, k), vector(rk2Betau_x, i, j, k), vector(rk2Alpha, i, j, k), vector(TauT_xx, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_xz, i, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xy, dx, simPlat_dt, ilast, jlast);
						FluxiSfd_z = FiSfd_z_mainI(vector(rk2Sfd_z, i, j, k), vector(rk2gtd_xz, i, j, k), vector(rk2gtd_yz, i, j, k), vector(rk2gtd_zz, i, j, k), vector(rk2Betau_x, i, j, k), vector(rk2Alpha, i, j, k), vector(TauT_xx, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_xz, i, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xz, dx, simPlat_dt, ilast, jlast);
						FluxiBfu_x = FiBfu_x_mainI(vector(rk2phif, i, j, k), vector(rk2Alpha, i, j, k), rk2chi_max, rk2gtu_xx, dx, simPlat_dt, ilast, jlast);
						FluxiBfu_y = FiBfu_y_mainI(vector(rk2Bfu_x, i, j, k), vector(rk2Bfu_y, i, j, k), vector(rk2phif, i, j, k), vector(rk2Betau_x, i, j, k), vector(rk2Betau_y, i, j, k), vector(rk2Alpha, i, j, k), vector(TauM_xy, i, j, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						FluxiBfu_z = FiBfu_z_mainI(vector(rk2Bfu_x, i, j, k), vector(rk2Bfu_z, i, j, k), vector(rk2phif, i, j, k), vector(rk2Betau_x, i, j, k), vector(rk2Betau_z, i, j, k), vector(rk2Alpha, i, j, k), vector(TauM_xz, i, j, k), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxiDf = FiDf_mainI(vector(rk2Df, i, j, k), vector(rk2Betau_x, i, j, k), vector(rk2Alpha, i, j, k), vector(TauN_x, i, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						FluxiDYf = FiDYf_mainI(vector(rk2DYf, i, j, k), vector(rk2Betau_x, i, j, k), vector(rk2Alpha, i, j, k), vector(TauNe_x, i, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxitauf = Fitauf_mainI(vector(rk2Df, i, j, k), vector(rk2tauf, i, j, k), vector(rk2Betau_x, i, j, k), vector(rk2Alpha, i, j, k), vector(TauN_x, i, j, k), rk2sdetg, rk2Sfu_x, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxiphif = Fiphif_mainI(vector(rk2Bfu_x, i, j, k), vector(rk2phif, i, j, k), vector(rk2Betau_x, i, j, k), vector(rk2Alpha, i, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j, k) + rk2sqca - vector(sqcs, i, j, k) * rk2sqca;
						SpeediSfd_x = fabs(fabs(vector(rk2Betau_x, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x))))));
						SpeediSfd_y = fabs(fabs(vector(rk2Betau_x, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x))))));
						SpeediSfd_z = fabs(fabs(vector(rk2Betau_x, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x))))));
						SpeediBfu_x = fabs(fabs(vector(rk2Betau_x, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x))))));
						SpeediBfu_y = fabs(fabs(vector(rk2Betau_x, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x))))));
						SpeediBfu_z = fabs(fabs(vector(rk2Betau_x, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x))))));
						SpeediDf = fabs(fabs(vector(rk2Betau_x, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x))))));
						SpeediDYf = fabs(fabs(vector(rk2Betau_x, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x))))));
						Speeditauf = fabs(fabs(vector(rk2Betau_x, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x))))));
						Speediphif = fabs(fabs(vector(rk2Betau_x, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i - 3, j, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i - 3, j, k) * vector(rk2Bfu_x, i - 3, j, k) + vector(rk2gtd_yy, i - 3, j, k) * vector(rk2Bfu_y, i - 3, j, k) + vector(rk2gtd_yz, i - 3, j, k) * vector(rk2Bfu_z, i - 3, j, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i - 3, j, k) * vector(rk2Bfu_x, i - 3, j, k) + vector(rk2gtd_yz, i - 3, j, k) * vector(rk2Bfu_y, i - 3, j, k) + vector(rk2gtd_zz, i - 3, j, k) * vector(rk2Bfu_z, i - 3, j, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i - 3, j, k) * vector(rk2Bfu_x, i - 3, j, k) + vector(rk2gtd_xy, i - 3, j, k) * vector(rk2Bfu_y, i - 3, j, k) + vector(rk2gtd_xz, i - 3, j, k) * vector(rk2Bfu_z, i - 3, j, k));
						rk2sqBf = vector(rk2Bfu_x, i - 3, j, k) * rk2Bfd_x + vector(rk2Bfu_y, i - 3, j, k) * rk2Bfd_y + vector(rk2Bfu_z, i - 3, j, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i - 3, j, k) * vector(rk2gtd_zz, i - 3, j, k) * vector(rk2gtd_xx, i - 3, j, k) + (-vector(rk2gtd_yy, i - 3, j, k) * (vector(rk2gtd_xz, i - 3, j, k) * vector(rk2gtd_xz, i - 3, j, k))) + (-vector(rk2gtd_zz, i - 3, j, k) * (vector(rk2gtd_xy, i - 3, j, k) * vector(rk2gtd_xy, i - 3, j, k))) + (-(vector(rk2gtd_yz, i - 3, j, k) * vector(rk2gtd_yz, i - 3, j, k)) * vector(rk2gtd_xx, i - 3, j, k)) + 2.0 * vector(rk2gtd_yz, i - 3, j, k) * vector(rk2gtd_xy, i - 3, j, k) * vector(rk2gtd_xz, i - 3, j, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i - 3, j, k) * vector(rk2gtd_yy, i - 3, j, k) - vector(rk2gtd_xy, i - 3, j, k) * vector(rk2gtd_xy, i - 3, j, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i - 3, j, k) * vector(rk2gtd_zz, i - 3, j, k) - vector(rk2gtd_yz, i - 3, j, k) * vector(rk2gtd_yz, i - 3, j, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i - 3, j, k) * vector(rk2gtd_yz, i - 3, j, k) - vector(rk2gtd_yy, i - 3, j, k) * vector(rk2gtd_xz, i - 3, j, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i - 3, j, k) * vector(rk2gtd_zz, i - 3, j, k)) + vector(rk2gtd_yz, i - 3, j, k) * vector(rk2gtd_xz, i - 3, j, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i - 3, j, k) + rk2gtu_xy * vector(rk2Sfd_y, i - 3, j, k) + rk2gtu_xz * vector(rk2Sfd_z, i - 3, j, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i - 3, j, k) + rk2gtu_xy * vector(vfd_y, i - 3, j, k) + rk2gtu_xz * vector(vfd_z, i - 3, j, k));
						rk2Bfvf = vector(rk2Bfu_x, i - 3, j, k) * vector(vfd_x, i - 3, j, k) + vector(rk2Bfu_y, i - 3, j, k) * vector(vfd_y, i - 3, j, k) + vector(rk2Bfu_z, i - 3, j, k) * vector(vfd_z, i - 3, j, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i - 3, j, k) * vector(rk2gtd_zz, i - 3, j, k) - vector(rk2gtd_xz, i - 3, j, k) * vector(rk2gtd_xz, i - 3, j, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i - 3, j, k) * vector(rk2gtd_yz, i - 3, j, k)) + vector(rk2gtd_xy, i - 3, j, k) * vector(rk2gtd_xz, i - 3, j, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i - 3, j, k) + rk2gtu_yz * vector(vfd_y, i - 3, j, k) + rk2gtu_zz * vector(vfd_z, i - 3, j, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i - 3, j, k) + rk2gtu_yy * vector(rk2Sfd_y, i - 3, j, k) + rk2gtu_yz * vector(rk2Sfd_z, i - 3, j, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i - 3, j, k) + rk2gtu_yz * vector(rk2Sfd_y, i - 3, j, k) + rk2gtu_zz * vector(rk2Sfd_z, i - 3, j, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i - 3, j, k) + rk2gtu_yy * vector(vfd_y, i - 3, j, k) + rk2gtu_yz * vector(vfd_z, i - 3, j, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i - 3, j, k) + rk2vfu_y * vector(vfd_y, i - 3, j, k) + rk2vfu_z * vector(vfd_z, i - 3, j, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i - 3, j, k) * vector(rk2Bfu_z, i - 3, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i - 3, j, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i - 3, j, k) * vector(rk2Bfu_x, i - 3, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i - 3, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i - 3, j, k) * vector(rk2Bfu_z, i - 3, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i - 3, j, k) * rk2vfu_z + vector(rk2Bfu_z, i - 3, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i - 3, j, k) * vector(rk2Bfu_y, i - 3, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i - 3, j, k) * rk2vfu_y + vector(rk2Bfu_y, i - 3, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i - 3, j, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i - 3, j, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i - 3, j, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i - 3, j, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i - 3, j, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i - 3, j, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i - 3, j, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i - 3, j, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i - 3, j, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i - 3, j, k) * vector(rk2Bfu_z, i - 3, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i - 3, j, k) * rk2vfu_z + vector(rk2Bfu_z, i - 3, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i - 3, j, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i - 3, j, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i - 3, j, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i - 3, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i - 3, j, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i - 3, j, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i - 3, j, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i - 3, j, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i - 3, j, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i - 3, j, k) * vector(rk2Bfu_y, i - 3, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i - 3, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i - 3, j, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i - 3, j, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i - 3, j, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i - 3, j, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i - 3, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i - 3, j, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i - 3, j, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i - 3, j, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i - 3, j, k));
						Fluxm3iSfd_x = FiSfd_x_mainI(vector(rk2Sfd_x, i - 3, j, k), vector(rk2gtd_xx, i - 3, j, k), vector(rk2gtd_xy, i - 3, j, k), vector(rk2gtd_xz, i - 3, j, k), vector(rk2Betau_x, i - 3, j, k), vector(rk2Alpha, i - 3, j, k), vector(TauT_xx, i - 3, j, k), vector(TauT_xy, i - 3, j, k), vector(TauT_xz, i - 3, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm3iSfd_y = FiSfd_y_mainI(vector(rk2Sfd_y, i - 3, j, k), vector(rk2gtd_xy, i - 3, j, k), vector(rk2gtd_yy, i - 3, j, k), vector(rk2gtd_yz, i - 3, j, k), vector(rk2Betau_x, i - 3, j, k), vector(rk2Alpha, i - 3, j, k), vector(TauT_xx, i - 3, j, k), vector(TauT_xy, i - 3, j, k), vector(TauT_xz, i - 3, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxm3iSfd_z = FiSfd_z_mainI(vector(rk2Sfd_z, i - 3, j, k), vector(rk2gtd_xz, i - 3, j, k), vector(rk2gtd_yz, i - 3, j, k), vector(rk2gtd_zz, i - 3, j, k), vector(rk2Betau_x, i - 3, j, k), vector(rk2Alpha, i - 3, j, k), vector(TauT_xx, i - 3, j, k), vector(TauT_xy, i - 3, j, k), vector(TauT_xz, i - 3, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxm3iBfu_x = FiBfu_x_mainI(vector(rk2phif, i - 3, j, k), vector(rk2Alpha, i - 3, j, k), rk2chi_max, rk2gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm3iBfu_y = FiBfu_y_mainI(vector(rk2Bfu_x, i - 3, j, k), vector(rk2Bfu_y, i - 3, j, k), vector(rk2phif, i - 3, j, k), vector(rk2Betau_x, i - 3, j, k), vector(rk2Betau_y, i - 3, j, k), vector(rk2Alpha, i - 3, j, k), vector(TauM_xy, i - 3, j, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3iBfu_z = FiBfu_z_mainI(vector(rk2Bfu_x, i - 3, j, k), vector(rk2Bfu_z, i - 3, j, k), vector(rk2phif, i - 3, j, k), vector(rk2Betau_x, i - 3, j, k), vector(rk2Betau_z, i - 3, j, k), vector(rk2Alpha, i - 3, j, k), vector(TauM_xz, i - 3, j, k), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3iDf = FiDf_mainI(vector(rk2Df, i - 3, j, k), vector(rk2Betau_x, i - 3, j, k), vector(rk2Alpha, i - 3, j, k), vector(TauN_x, i - 3, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm3iDYf = FiDYf_mainI(vector(rk2DYf, i - 3, j, k), vector(rk2Betau_x, i - 3, j, k), vector(rk2Alpha, i - 3, j, k), vector(TauNe_x, i - 3, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm3itauf = Fitauf_mainI(vector(rk2Df, i - 3, j, k), vector(rk2tauf, i - 3, j, k), vector(rk2Betau_x, i - 3, j, k), vector(rk2Alpha, i - 3, j, k), vector(TauN_x, i - 3, j, k), rk2sdetg, rk2Sfu_x, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm3iphif = Fiphif_mainI(vector(rk2Bfu_x, i - 3, j, k), vector(rk2phif, i - 3, j, k), vector(rk2Betau_x, i - 3, j, k), vector(rk2Alpha, i - 3, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i - 3, j, k) * (1.0 + vector(epsf, i - 3, j, k)) + vector(pf, i - 3, j, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i - 3, j, k) + rk2sqca - vector(sqcs, i - 3, j, k) * rk2sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk2Betau_x, i - 3, j, k)) + vector(rk2Alpha, i - 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk2Betau_x, i - 3, j, k)) + vector(rk2Alpha, i - 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk2Betau_x, i - 3, j, k)) + vector(rk2Alpha, i - 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk2Betau_x, i - 3, j, k)) + vector(rk2Alpha, i - 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk2Betau_x, i - 3, j, k)) + vector(rk2Alpha, i - 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk2Betau_x, i - 3, j, k)) + vector(rk2Alpha, i - 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk2Betau_x, i - 3, j, k)) + vector(rk2Alpha, i - 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk2Betau_x, i - 3, j, k)) + vector(rk2Alpha, i - 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk2Betau_x, i - 3, j, k)) + vector(rk2Alpha, i - 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk2Betau_x, i - 3, j, k)) + vector(rk2Alpha, i - 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i - 2, j, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i - 2, j, k) * vector(rk2Bfu_x, i - 2, j, k) + vector(rk2gtd_yy, i - 2, j, k) * vector(rk2Bfu_y, i - 2, j, k) + vector(rk2gtd_yz, i - 2, j, k) * vector(rk2Bfu_z, i - 2, j, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i - 2, j, k) * vector(rk2Bfu_x, i - 2, j, k) + vector(rk2gtd_yz, i - 2, j, k) * vector(rk2Bfu_y, i - 2, j, k) + vector(rk2gtd_zz, i - 2, j, k) * vector(rk2Bfu_z, i - 2, j, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i - 2, j, k) * vector(rk2Bfu_x, i - 2, j, k) + vector(rk2gtd_xy, i - 2, j, k) * vector(rk2Bfu_y, i - 2, j, k) + vector(rk2gtd_xz, i - 2, j, k) * vector(rk2Bfu_z, i - 2, j, k));
						rk2sqBf = vector(rk2Bfu_x, i - 2, j, k) * rk2Bfd_x + vector(rk2Bfu_y, i - 2, j, k) * rk2Bfd_y + vector(rk2Bfu_z, i - 2, j, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i - 2, j, k) * vector(rk2gtd_zz, i - 2, j, k) * vector(rk2gtd_xx, i - 2, j, k) + (-vector(rk2gtd_yy, i - 2, j, k) * (vector(rk2gtd_xz, i - 2, j, k) * vector(rk2gtd_xz, i - 2, j, k))) + (-vector(rk2gtd_zz, i - 2, j, k) * (vector(rk2gtd_xy, i - 2, j, k) * vector(rk2gtd_xy, i - 2, j, k))) + (-(vector(rk2gtd_yz, i - 2, j, k) * vector(rk2gtd_yz, i - 2, j, k)) * vector(rk2gtd_xx, i - 2, j, k)) + 2.0 * vector(rk2gtd_yz, i - 2, j, k) * vector(rk2gtd_xy, i - 2, j, k) * vector(rk2gtd_xz, i - 2, j, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i - 2, j, k) * vector(rk2gtd_yy, i - 2, j, k) - vector(rk2gtd_xy, i - 2, j, k) * vector(rk2gtd_xy, i - 2, j, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i - 2, j, k) * vector(rk2gtd_zz, i - 2, j, k) - vector(rk2gtd_yz, i - 2, j, k) * vector(rk2gtd_yz, i - 2, j, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i - 2, j, k) * vector(rk2gtd_yz, i - 2, j, k) - vector(rk2gtd_yy, i - 2, j, k) * vector(rk2gtd_xz, i - 2, j, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i - 2, j, k) * vector(rk2gtd_zz, i - 2, j, k)) + vector(rk2gtd_yz, i - 2, j, k) * vector(rk2gtd_xz, i - 2, j, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i - 2, j, k) + rk2gtu_xy * vector(rk2Sfd_y, i - 2, j, k) + rk2gtu_xz * vector(rk2Sfd_z, i - 2, j, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i - 2, j, k) + rk2gtu_xy * vector(vfd_y, i - 2, j, k) + rk2gtu_xz * vector(vfd_z, i - 2, j, k));
						rk2Bfvf = vector(rk2Bfu_x, i - 2, j, k) * vector(vfd_x, i - 2, j, k) + vector(rk2Bfu_y, i - 2, j, k) * vector(vfd_y, i - 2, j, k) + vector(rk2Bfu_z, i - 2, j, k) * vector(vfd_z, i - 2, j, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i - 2, j, k) * vector(rk2gtd_zz, i - 2, j, k) - vector(rk2gtd_xz, i - 2, j, k) * vector(rk2gtd_xz, i - 2, j, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i - 2, j, k) * vector(rk2gtd_yz, i - 2, j, k)) + vector(rk2gtd_xy, i - 2, j, k) * vector(rk2gtd_xz, i - 2, j, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i - 2, j, k) + rk2gtu_yz * vector(vfd_y, i - 2, j, k) + rk2gtu_zz * vector(vfd_z, i - 2, j, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i - 2, j, k) + rk2gtu_yy * vector(rk2Sfd_y, i - 2, j, k) + rk2gtu_yz * vector(rk2Sfd_z, i - 2, j, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i - 2, j, k) + rk2gtu_yz * vector(rk2Sfd_y, i - 2, j, k) + rk2gtu_zz * vector(rk2Sfd_z, i - 2, j, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i - 2, j, k) + rk2gtu_yy * vector(vfd_y, i - 2, j, k) + rk2gtu_yz * vector(vfd_z, i - 2, j, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i - 2, j, k) + rk2vfu_y * vector(vfd_y, i - 2, j, k) + rk2vfu_z * vector(vfd_z, i - 2, j, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i - 2, j, k) * vector(rk2Bfu_z, i - 2, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i - 2, j, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i - 2, j, k) * vector(rk2Bfu_x, i - 2, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i - 2, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i - 2, j, k) * vector(rk2Bfu_z, i - 2, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i - 2, j, k) * rk2vfu_z + vector(rk2Bfu_z, i - 2, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i - 2, j, k) * vector(rk2Bfu_y, i - 2, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i - 2, j, k) * rk2vfu_y + vector(rk2Bfu_y, i - 2, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i - 2, j, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i - 2, j, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i - 2, j, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i - 2, j, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i - 2, j, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i - 2, j, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i - 2, j, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i - 2, j, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i - 2, j, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i - 2, j, k) * vector(rk2Bfu_z, i - 2, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i - 2, j, k) * rk2vfu_z + vector(rk2Bfu_z, i - 2, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i - 2, j, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i - 2, j, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i - 2, j, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i - 2, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i - 2, j, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i - 2, j, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i - 2, j, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i - 2, j, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i - 2, j, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i - 2, j, k) * vector(rk2Bfu_y, i - 2, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i - 2, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i - 2, j, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i - 2, j, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i - 2, j, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i - 2, j, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i - 2, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i - 2, j, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i - 2, j, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i - 2, j, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i - 2, j, k));
						Fluxm2iSfd_x = FiSfd_x_mainI(vector(rk2Sfd_x, i - 2, j, k), vector(rk2gtd_xx, i - 2, j, k), vector(rk2gtd_xy, i - 2, j, k), vector(rk2gtd_xz, i - 2, j, k), vector(rk2Betau_x, i - 2, j, k), vector(rk2Alpha, i - 2, j, k), vector(TauT_xx, i - 2, j, k), vector(TauT_xy, i - 2, j, k), vector(TauT_xz, i - 2, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm2iSfd_y = FiSfd_y_mainI(vector(rk2Sfd_y, i - 2, j, k), vector(rk2gtd_xy, i - 2, j, k), vector(rk2gtd_yy, i - 2, j, k), vector(rk2gtd_yz, i - 2, j, k), vector(rk2Betau_x, i - 2, j, k), vector(rk2Alpha, i - 2, j, k), vector(TauT_xx, i - 2, j, k), vector(TauT_xy, i - 2, j, k), vector(TauT_xz, i - 2, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxm2iSfd_z = FiSfd_z_mainI(vector(rk2Sfd_z, i - 2, j, k), vector(rk2gtd_xz, i - 2, j, k), vector(rk2gtd_yz, i - 2, j, k), vector(rk2gtd_zz, i - 2, j, k), vector(rk2Betau_x, i - 2, j, k), vector(rk2Alpha, i - 2, j, k), vector(TauT_xx, i - 2, j, k), vector(TauT_xy, i - 2, j, k), vector(TauT_xz, i - 2, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxm2iBfu_x = FiBfu_x_mainI(vector(rk2phif, i - 2, j, k), vector(rk2Alpha, i - 2, j, k), rk2chi_max, rk2gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm2iBfu_y = FiBfu_y_mainI(vector(rk2Bfu_x, i - 2, j, k), vector(rk2Bfu_y, i - 2, j, k), vector(rk2phif, i - 2, j, k), vector(rk2Betau_x, i - 2, j, k), vector(rk2Betau_y, i - 2, j, k), vector(rk2Alpha, i - 2, j, k), vector(TauM_xy, i - 2, j, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2iBfu_z = FiBfu_z_mainI(vector(rk2Bfu_x, i - 2, j, k), vector(rk2Bfu_z, i - 2, j, k), vector(rk2phif, i - 2, j, k), vector(rk2Betau_x, i - 2, j, k), vector(rk2Betau_z, i - 2, j, k), vector(rk2Alpha, i - 2, j, k), vector(TauM_xz, i - 2, j, k), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2iDf = FiDf_mainI(vector(rk2Df, i - 2, j, k), vector(rk2Betau_x, i - 2, j, k), vector(rk2Alpha, i - 2, j, k), vector(TauN_x, i - 2, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm2iDYf = FiDYf_mainI(vector(rk2DYf, i - 2, j, k), vector(rk2Betau_x, i - 2, j, k), vector(rk2Alpha, i - 2, j, k), vector(TauNe_x, i - 2, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm2itauf = Fitauf_mainI(vector(rk2Df, i - 2, j, k), vector(rk2tauf, i - 2, j, k), vector(rk2Betau_x, i - 2, j, k), vector(rk2Alpha, i - 2, j, k), vector(TauN_x, i - 2, j, k), rk2sdetg, rk2Sfu_x, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm2iphif = Fiphif_mainI(vector(rk2Bfu_x, i - 2, j, k), vector(rk2phif, i - 2, j, k), vector(rk2Betau_x, i - 2, j, k), vector(rk2Alpha, i - 2, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i - 2, j, k) * (1.0 + vector(epsf, i - 2, j, k)) + vector(pf, i - 2, j, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i - 2, j, k) + rk2sqca - vector(sqcs, i - 2, j, k) * rk2sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk2Betau_x, i - 2, j, k)) + vector(rk2Alpha, i - 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk2Betau_x, i - 2, j, k)) + vector(rk2Alpha, i - 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk2Betau_x, i - 2, j, k)) + vector(rk2Alpha, i - 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk2Betau_x, i - 2, j, k)) + vector(rk2Alpha, i - 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk2Betau_x, i - 2, j, k)) + vector(rk2Alpha, i - 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk2Betau_x, i - 2, j, k)) + vector(rk2Alpha, i - 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk2Betau_x, i - 2, j, k)) + vector(rk2Alpha, i - 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk2Betau_x, i - 2, j, k)) + vector(rk2Alpha, i - 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk2Betau_x, i - 2, j, k)) + vector(rk2Alpha, i - 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk2Betau_x, i - 2, j, k)) + vector(rk2Alpha, i - 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i - 1, j, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i - 1, j, k) * vector(rk2Bfu_x, i - 1, j, k) + vector(rk2gtd_yy, i - 1, j, k) * vector(rk2Bfu_y, i - 1, j, k) + vector(rk2gtd_yz, i - 1, j, k) * vector(rk2Bfu_z, i - 1, j, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i - 1, j, k) * vector(rk2Bfu_x, i - 1, j, k) + vector(rk2gtd_yz, i - 1, j, k) * vector(rk2Bfu_y, i - 1, j, k) + vector(rk2gtd_zz, i - 1, j, k) * vector(rk2Bfu_z, i - 1, j, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i - 1, j, k) * vector(rk2Bfu_x, i - 1, j, k) + vector(rk2gtd_xy, i - 1, j, k) * vector(rk2Bfu_y, i - 1, j, k) + vector(rk2gtd_xz, i - 1, j, k) * vector(rk2Bfu_z, i - 1, j, k));
						rk2sqBf = vector(rk2Bfu_x, i - 1, j, k) * rk2Bfd_x + vector(rk2Bfu_y, i - 1, j, k) * rk2Bfd_y + vector(rk2Bfu_z, i - 1, j, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i - 1, j, k) * vector(rk2gtd_zz, i - 1, j, k) * vector(rk2gtd_xx, i - 1, j, k) + (-vector(rk2gtd_yy, i - 1, j, k) * (vector(rk2gtd_xz, i - 1, j, k) * vector(rk2gtd_xz, i - 1, j, k))) + (-vector(rk2gtd_zz, i - 1, j, k) * (vector(rk2gtd_xy, i - 1, j, k) * vector(rk2gtd_xy, i - 1, j, k))) + (-(vector(rk2gtd_yz, i - 1, j, k) * vector(rk2gtd_yz, i - 1, j, k)) * vector(rk2gtd_xx, i - 1, j, k)) + 2.0 * vector(rk2gtd_yz, i - 1, j, k) * vector(rk2gtd_xy, i - 1, j, k) * vector(rk2gtd_xz, i - 1, j, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i - 1, j, k) * vector(rk2gtd_yy, i - 1, j, k) - vector(rk2gtd_xy, i - 1, j, k) * vector(rk2gtd_xy, i - 1, j, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i - 1, j, k) * vector(rk2gtd_zz, i - 1, j, k) - vector(rk2gtd_yz, i - 1, j, k) * vector(rk2gtd_yz, i - 1, j, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i - 1, j, k) * vector(rk2gtd_yz, i - 1, j, k) - vector(rk2gtd_yy, i - 1, j, k) * vector(rk2gtd_xz, i - 1, j, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i - 1, j, k) * vector(rk2gtd_zz, i - 1, j, k)) + vector(rk2gtd_yz, i - 1, j, k) * vector(rk2gtd_xz, i - 1, j, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i - 1, j, k) + rk2gtu_xy * vector(rk2Sfd_y, i - 1, j, k) + rk2gtu_xz * vector(rk2Sfd_z, i - 1, j, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i - 1, j, k) + rk2gtu_xy * vector(vfd_y, i - 1, j, k) + rk2gtu_xz * vector(vfd_z, i - 1, j, k));
						rk2Bfvf = vector(rk2Bfu_x, i - 1, j, k) * vector(vfd_x, i - 1, j, k) + vector(rk2Bfu_y, i - 1, j, k) * vector(vfd_y, i - 1, j, k) + vector(rk2Bfu_z, i - 1, j, k) * vector(vfd_z, i - 1, j, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i - 1, j, k) * vector(rk2gtd_zz, i - 1, j, k) - vector(rk2gtd_xz, i - 1, j, k) * vector(rk2gtd_xz, i - 1, j, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i - 1, j, k) * vector(rk2gtd_yz, i - 1, j, k)) + vector(rk2gtd_xy, i - 1, j, k) * vector(rk2gtd_xz, i - 1, j, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i - 1, j, k) + rk2gtu_yz * vector(vfd_y, i - 1, j, k) + rk2gtu_zz * vector(vfd_z, i - 1, j, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i - 1, j, k) + rk2gtu_yy * vector(rk2Sfd_y, i - 1, j, k) + rk2gtu_yz * vector(rk2Sfd_z, i - 1, j, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i - 1, j, k) + rk2gtu_yz * vector(rk2Sfd_y, i - 1, j, k) + rk2gtu_zz * vector(rk2Sfd_z, i - 1, j, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i - 1, j, k) + rk2gtu_yy * vector(vfd_y, i - 1, j, k) + rk2gtu_yz * vector(vfd_z, i - 1, j, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i - 1, j, k) + rk2vfu_y * vector(vfd_y, i - 1, j, k) + rk2vfu_z * vector(vfd_z, i - 1, j, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i - 1, j, k) * vector(rk2Bfu_z, i - 1, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i - 1, j, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i - 1, j, k) * vector(rk2Bfu_x, i - 1, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i - 1, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i - 1, j, k) * vector(rk2Bfu_z, i - 1, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i - 1, j, k) * rk2vfu_z + vector(rk2Bfu_z, i - 1, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i - 1, j, k) * vector(rk2Bfu_y, i - 1, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i - 1, j, k) * rk2vfu_y + vector(rk2Bfu_y, i - 1, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i - 1, j, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i - 1, j, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i - 1, j, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i - 1, j, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i - 1, j, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i - 1, j, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i - 1, j, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i - 1, j, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i - 1, j, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i - 1, j, k) * vector(rk2Bfu_z, i - 1, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i - 1, j, k) * rk2vfu_z + vector(rk2Bfu_z, i - 1, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i - 1, j, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i - 1, j, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i - 1, j, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i - 1, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i - 1, j, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i - 1, j, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i - 1, j, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i - 1, j, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i - 1, j, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i - 1, j, k) * vector(rk2Bfu_y, i - 1, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i - 1, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i - 1, j, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i - 1, j, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i - 1, j, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i - 1, j, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i - 1, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i - 1, j, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i - 1, j, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i - 1, j, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i - 1, j, k));
						Fluxm1iSfd_x = FiSfd_x_mainI(vector(rk2Sfd_x, i - 1, j, k), vector(rk2gtd_xx, i - 1, j, k), vector(rk2gtd_xy, i - 1, j, k), vector(rk2gtd_xz, i - 1, j, k), vector(rk2Betau_x, i - 1, j, k), vector(rk2Alpha, i - 1, j, k), vector(TauT_xx, i - 1, j, k), vector(TauT_xy, i - 1, j, k), vector(TauT_xz, i - 1, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm1iSfd_y = FiSfd_y_mainI(vector(rk2Sfd_y, i - 1, j, k), vector(rk2gtd_xy, i - 1, j, k), vector(rk2gtd_yy, i - 1, j, k), vector(rk2gtd_yz, i - 1, j, k), vector(rk2Betau_x, i - 1, j, k), vector(rk2Alpha, i - 1, j, k), vector(TauT_xx, i - 1, j, k), vector(TauT_xy, i - 1, j, k), vector(TauT_xz, i - 1, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxm1iSfd_z = FiSfd_z_mainI(vector(rk2Sfd_z, i - 1, j, k), vector(rk2gtd_xz, i - 1, j, k), vector(rk2gtd_yz, i - 1, j, k), vector(rk2gtd_zz, i - 1, j, k), vector(rk2Betau_x, i - 1, j, k), vector(rk2Alpha, i - 1, j, k), vector(TauT_xx, i - 1, j, k), vector(TauT_xy, i - 1, j, k), vector(TauT_xz, i - 1, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxm1iBfu_x = FiBfu_x_mainI(vector(rk2phif, i - 1, j, k), vector(rk2Alpha, i - 1, j, k), rk2chi_max, rk2gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm1iBfu_y = FiBfu_y_mainI(vector(rk2Bfu_x, i - 1, j, k), vector(rk2Bfu_y, i - 1, j, k), vector(rk2phif, i - 1, j, k), vector(rk2Betau_x, i - 1, j, k), vector(rk2Betau_y, i - 1, j, k), vector(rk2Alpha, i - 1, j, k), vector(TauM_xy, i - 1, j, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1iBfu_z = FiBfu_z_mainI(vector(rk2Bfu_x, i - 1, j, k), vector(rk2Bfu_z, i - 1, j, k), vector(rk2phif, i - 1, j, k), vector(rk2Betau_x, i - 1, j, k), vector(rk2Betau_z, i - 1, j, k), vector(rk2Alpha, i - 1, j, k), vector(TauM_xz, i - 1, j, k), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1iDf = FiDf_mainI(vector(rk2Df, i - 1, j, k), vector(rk2Betau_x, i - 1, j, k), vector(rk2Alpha, i - 1, j, k), vector(TauN_x, i - 1, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm1iDYf = FiDYf_mainI(vector(rk2DYf, i - 1, j, k), vector(rk2Betau_x, i - 1, j, k), vector(rk2Alpha, i - 1, j, k), vector(TauNe_x, i - 1, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm1itauf = Fitauf_mainI(vector(rk2Df, i - 1, j, k), vector(rk2tauf, i - 1, j, k), vector(rk2Betau_x, i - 1, j, k), vector(rk2Alpha, i - 1, j, k), vector(TauN_x, i - 1, j, k), rk2sdetg, rk2Sfu_x, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm1iphif = Fiphif_mainI(vector(rk2Bfu_x, i - 1, j, k), vector(rk2phif, i - 1, j, k), vector(rk2Betau_x, i - 1, j, k), vector(rk2Alpha, i - 1, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i - 1, j, k) * (1.0 + vector(epsf, i - 1, j, k)) + vector(pf, i - 1, j, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i - 1, j, k) + rk2sqca - vector(sqcs, i - 1, j, k) * rk2sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk2Betau_x, i - 1, j, k)) + vector(rk2Alpha, i - 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk2Betau_x, i - 1, j, k)) + vector(rk2Alpha, i - 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk2Betau_x, i - 1, j, k)) + vector(rk2Alpha, i - 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk2Betau_x, i - 1, j, k)) + vector(rk2Alpha, i - 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk2Betau_x, i - 1, j, k)) + vector(rk2Alpha, i - 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk2Betau_x, i - 1, j, k)) + vector(rk2Alpha, i - 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk2Betau_x, i - 1, j, k)) + vector(rk2Alpha, i - 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk2Betau_x, i - 1, j, k)) + vector(rk2Alpha, i - 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk2Betau_x, i - 1, j, k)) + vector(rk2Alpha, i - 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk2Betau_x, i - 1, j, k)) + vector(rk2Alpha, i - 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i + 1, j, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i + 1, j, k) * vector(rk2Bfu_x, i + 1, j, k) + vector(rk2gtd_yy, i + 1, j, k) * vector(rk2Bfu_y, i + 1, j, k) + vector(rk2gtd_yz, i + 1, j, k) * vector(rk2Bfu_z, i + 1, j, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i + 1, j, k) * vector(rk2Bfu_x, i + 1, j, k) + vector(rk2gtd_yz, i + 1, j, k) * vector(rk2Bfu_y, i + 1, j, k) + vector(rk2gtd_zz, i + 1, j, k) * vector(rk2Bfu_z, i + 1, j, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i + 1, j, k) * vector(rk2Bfu_x, i + 1, j, k) + vector(rk2gtd_xy, i + 1, j, k) * vector(rk2Bfu_y, i + 1, j, k) + vector(rk2gtd_xz, i + 1, j, k) * vector(rk2Bfu_z, i + 1, j, k));
						rk2sqBf = vector(rk2Bfu_x, i + 1, j, k) * rk2Bfd_x + vector(rk2Bfu_y, i + 1, j, k) * rk2Bfd_y + vector(rk2Bfu_z, i + 1, j, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i + 1, j, k) * vector(rk2gtd_zz, i + 1, j, k) * vector(rk2gtd_xx, i + 1, j, k) + (-vector(rk2gtd_yy, i + 1, j, k) * (vector(rk2gtd_xz, i + 1, j, k) * vector(rk2gtd_xz, i + 1, j, k))) + (-vector(rk2gtd_zz, i + 1, j, k) * (vector(rk2gtd_xy, i + 1, j, k) * vector(rk2gtd_xy, i + 1, j, k))) + (-(vector(rk2gtd_yz, i + 1, j, k) * vector(rk2gtd_yz, i + 1, j, k)) * vector(rk2gtd_xx, i + 1, j, k)) + 2.0 * vector(rk2gtd_yz, i + 1, j, k) * vector(rk2gtd_xy, i + 1, j, k) * vector(rk2gtd_xz, i + 1, j, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i + 1, j, k) * vector(rk2gtd_yy, i + 1, j, k) - vector(rk2gtd_xy, i + 1, j, k) * vector(rk2gtd_xy, i + 1, j, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i + 1, j, k) * vector(rk2gtd_zz, i + 1, j, k) - vector(rk2gtd_yz, i + 1, j, k) * vector(rk2gtd_yz, i + 1, j, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i + 1, j, k) * vector(rk2gtd_yz, i + 1, j, k) - vector(rk2gtd_yy, i + 1, j, k) * vector(rk2gtd_xz, i + 1, j, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i + 1, j, k) * vector(rk2gtd_zz, i + 1, j, k)) + vector(rk2gtd_yz, i + 1, j, k) * vector(rk2gtd_xz, i + 1, j, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i + 1, j, k) + rk2gtu_xy * vector(rk2Sfd_y, i + 1, j, k) + rk2gtu_xz * vector(rk2Sfd_z, i + 1, j, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i + 1, j, k) + rk2gtu_xy * vector(vfd_y, i + 1, j, k) + rk2gtu_xz * vector(vfd_z, i + 1, j, k));
						rk2Bfvf = vector(rk2Bfu_x, i + 1, j, k) * vector(vfd_x, i + 1, j, k) + vector(rk2Bfu_y, i + 1, j, k) * vector(vfd_y, i + 1, j, k) + vector(rk2Bfu_z, i + 1, j, k) * vector(vfd_z, i + 1, j, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i + 1, j, k) * vector(rk2gtd_zz, i + 1, j, k) - vector(rk2gtd_xz, i + 1, j, k) * vector(rk2gtd_xz, i + 1, j, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i + 1, j, k) * vector(rk2gtd_yz, i + 1, j, k)) + vector(rk2gtd_xy, i + 1, j, k) * vector(rk2gtd_xz, i + 1, j, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i + 1, j, k) + rk2gtu_yz * vector(vfd_y, i + 1, j, k) + rk2gtu_zz * vector(vfd_z, i + 1, j, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i + 1, j, k) + rk2gtu_yy * vector(rk2Sfd_y, i + 1, j, k) + rk2gtu_yz * vector(rk2Sfd_z, i + 1, j, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i + 1, j, k) + rk2gtu_yz * vector(rk2Sfd_y, i + 1, j, k) + rk2gtu_zz * vector(rk2Sfd_z, i + 1, j, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i + 1, j, k) + rk2gtu_yy * vector(vfd_y, i + 1, j, k) + rk2gtu_yz * vector(vfd_z, i + 1, j, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i + 1, j, k) + rk2vfu_y * vector(vfd_y, i + 1, j, k) + rk2vfu_z * vector(vfd_z, i + 1, j, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i + 1, j, k) * vector(rk2Bfu_z, i + 1, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i + 1, j, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i + 1, j, k) * vector(rk2Bfu_x, i + 1, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i + 1, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i + 1, j, k) * vector(rk2Bfu_z, i + 1, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i + 1, j, k) * rk2vfu_z + vector(rk2Bfu_z, i + 1, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i + 1, j, k) * vector(rk2Bfu_y, i + 1, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i + 1, j, k) * rk2vfu_y + vector(rk2Bfu_y, i + 1, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i + 1, j, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i + 1, j, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i + 1, j, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i + 1, j, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i + 1, j, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i + 1, j, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i + 1, j, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i + 1, j, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i + 1, j, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i + 1, j, k) * vector(rk2Bfu_z, i + 1, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i + 1, j, k) * rk2vfu_z + vector(rk2Bfu_z, i + 1, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i + 1, j, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i + 1, j, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i + 1, j, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i + 1, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i + 1, j, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i + 1, j, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i + 1, j, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i + 1, j, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i + 1, j, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i + 1, j, k) * vector(rk2Bfu_y, i + 1, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i + 1, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i + 1, j, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i + 1, j, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i + 1, j, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i + 1, j, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i + 1, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i + 1, j, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i + 1, j, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i + 1, j, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i + 1, j, k));
						Fluxp1iSfd_x = FiSfd_x_mainI(vector(rk2Sfd_x, i + 1, j, k), vector(rk2gtd_xx, i + 1, j, k), vector(rk2gtd_xy, i + 1, j, k), vector(rk2gtd_xz, i + 1, j, k), vector(rk2Betau_x, i + 1, j, k), vector(rk2Alpha, i + 1, j, k), vector(TauT_xx, i + 1, j, k), vector(TauT_xy, i + 1, j, k), vector(TauT_xz, i + 1, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp1iSfd_y = FiSfd_y_mainI(vector(rk2Sfd_y, i + 1, j, k), vector(rk2gtd_xy, i + 1, j, k), vector(rk2gtd_yy, i + 1, j, k), vector(rk2gtd_yz, i + 1, j, k), vector(rk2Betau_x, i + 1, j, k), vector(rk2Alpha, i + 1, j, k), vector(TauT_xx, i + 1, j, k), vector(TauT_xy, i + 1, j, k), vector(TauT_xz, i + 1, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxp1iSfd_z = FiSfd_z_mainI(vector(rk2Sfd_z, i + 1, j, k), vector(rk2gtd_xz, i + 1, j, k), vector(rk2gtd_yz, i + 1, j, k), vector(rk2gtd_zz, i + 1, j, k), vector(rk2Betau_x, i + 1, j, k), vector(rk2Alpha, i + 1, j, k), vector(TauT_xx, i + 1, j, k), vector(TauT_xy, i + 1, j, k), vector(TauT_xz, i + 1, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxp1iBfu_x = FiBfu_x_mainI(vector(rk2phif, i + 1, j, k), vector(rk2Alpha, i + 1, j, k), rk2chi_max, rk2gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp1iBfu_y = FiBfu_y_mainI(vector(rk2Bfu_x, i + 1, j, k), vector(rk2Bfu_y, i + 1, j, k), vector(rk2phif, i + 1, j, k), vector(rk2Betau_x, i + 1, j, k), vector(rk2Betau_y, i + 1, j, k), vector(rk2Alpha, i + 1, j, k), vector(TauM_xy, i + 1, j, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1iBfu_z = FiBfu_z_mainI(vector(rk2Bfu_x, i + 1, j, k), vector(rk2Bfu_z, i + 1, j, k), vector(rk2phif, i + 1, j, k), vector(rk2Betau_x, i + 1, j, k), vector(rk2Betau_z, i + 1, j, k), vector(rk2Alpha, i + 1, j, k), vector(TauM_xz, i + 1, j, k), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1iDf = FiDf_mainI(vector(rk2Df, i + 1, j, k), vector(rk2Betau_x, i + 1, j, k), vector(rk2Alpha, i + 1, j, k), vector(TauN_x, i + 1, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp1iDYf = FiDYf_mainI(vector(rk2DYf, i + 1, j, k), vector(rk2Betau_x, i + 1, j, k), vector(rk2Alpha, i + 1, j, k), vector(TauNe_x, i + 1, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp1itauf = Fitauf_mainI(vector(rk2Df, i + 1, j, k), vector(rk2tauf, i + 1, j, k), vector(rk2Betau_x, i + 1, j, k), vector(rk2Alpha, i + 1, j, k), vector(TauN_x, i + 1, j, k), rk2sdetg, rk2Sfu_x, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp1iphif = Fiphif_mainI(vector(rk2Bfu_x, i + 1, j, k), vector(rk2phif, i + 1, j, k), vector(rk2Betau_x, i + 1, j, k), vector(rk2Alpha, i + 1, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i + 1, j, k) * (1.0 + vector(epsf, i + 1, j, k)) + vector(pf, i + 1, j, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i + 1, j, k) + rk2sqca - vector(sqcs, i + 1, j, k) * rk2sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk2Betau_x, i + 1, j, k)) + vector(rk2Alpha, i + 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk2Betau_x, i + 1, j, k)) + vector(rk2Alpha, i + 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk2Betau_x, i + 1, j, k)) + vector(rk2Alpha, i + 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk2Betau_x, i + 1, j, k)) + vector(rk2Alpha, i + 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk2Betau_x, i + 1, j, k)) + vector(rk2Alpha, i + 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk2Betau_x, i + 1, j, k)) + vector(rk2Alpha, i + 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk2Betau_x, i + 1, j, k)) + vector(rk2Alpha, i + 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk2Betau_x, i + 1, j, k)) + vector(rk2Alpha, i + 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk2Betau_x, i + 1, j, k)) + vector(rk2Alpha, i + 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk2Betau_x, i + 1, j, k)) + vector(rk2Alpha, i + 1, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i + 2, j, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i + 2, j, k) * vector(rk2Bfu_x, i + 2, j, k) + vector(rk2gtd_yy, i + 2, j, k) * vector(rk2Bfu_y, i + 2, j, k) + vector(rk2gtd_yz, i + 2, j, k) * vector(rk2Bfu_z, i + 2, j, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i + 2, j, k) * vector(rk2Bfu_x, i + 2, j, k) + vector(rk2gtd_yz, i + 2, j, k) * vector(rk2Bfu_y, i + 2, j, k) + vector(rk2gtd_zz, i + 2, j, k) * vector(rk2Bfu_z, i + 2, j, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i + 2, j, k) * vector(rk2Bfu_x, i + 2, j, k) + vector(rk2gtd_xy, i + 2, j, k) * vector(rk2Bfu_y, i + 2, j, k) + vector(rk2gtd_xz, i + 2, j, k) * vector(rk2Bfu_z, i + 2, j, k));
						rk2sqBf = vector(rk2Bfu_x, i + 2, j, k) * rk2Bfd_x + vector(rk2Bfu_y, i + 2, j, k) * rk2Bfd_y + vector(rk2Bfu_z, i + 2, j, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i + 2, j, k) * vector(rk2gtd_zz, i + 2, j, k) * vector(rk2gtd_xx, i + 2, j, k) + (-vector(rk2gtd_yy, i + 2, j, k) * (vector(rk2gtd_xz, i + 2, j, k) * vector(rk2gtd_xz, i + 2, j, k))) + (-vector(rk2gtd_zz, i + 2, j, k) * (vector(rk2gtd_xy, i + 2, j, k) * vector(rk2gtd_xy, i + 2, j, k))) + (-(vector(rk2gtd_yz, i + 2, j, k) * vector(rk2gtd_yz, i + 2, j, k)) * vector(rk2gtd_xx, i + 2, j, k)) + 2.0 * vector(rk2gtd_yz, i + 2, j, k) * vector(rk2gtd_xy, i + 2, j, k) * vector(rk2gtd_xz, i + 2, j, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i + 2, j, k) * vector(rk2gtd_yy, i + 2, j, k) - vector(rk2gtd_xy, i + 2, j, k) * vector(rk2gtd_xy, i + 2, j, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i + 2, j, k) * vector(rk2gtd_zz, i + 2, j, k) - vector(rk2gtd_yz, i + 2, j, k) * vector(rk2gtd_yz, i + 2, j, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i + 2, j, k) * vector(rk2gtd_yz, i + 2, j, k) - vector(rk2gtd_yy, i + 2, j, k) * vector(rk2gtd_xz, i + 2, j, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i + 2, j, k) * vector(rk2gtd_zz, i + 2, j, k)) + vector(rk2gtd_yz, i + 2, j, k) * vector(rk2gtd_xz, i + 2, j, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i + 2, j, k) + rk2gtu_xy * vector(rk2Sfd_y, i + 2, j, k) + rk2gtu_xz * vector(rk2Sfd_z, i + 2, j, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i + 2, j, k) + rk2gtu_xy * vector(vfd_y, i + 2, j, k) + rk2gtu_xz * vector(vfd_z, i + 2, j, k));
						rk2Bfvf = vector(rk2Bfu_x, i + 2, j, k) * vector(vfd_x, i + 2, j, k) + vector(rk2Bfu_y, i + 2, j, k) * vector(vfd_y, i + 2, j, k) + vector(rk2Bfu_z, i + 2, j, k) * vector(vfd_z, i + 2, j, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i + 2, j, k) * vector(rk2gtd_zz, i + 2, j, k) - vector(rk2gtd_xz, i + 2, j, k) * vector(rk2gtd_xz, i + 2, j, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i + 2, j, k) * vector(rk2gtd_yz, i + 2, j, k)) + vector(rk2gtd_xy, i + 2, j, k) * vector(rk2gtd_xz, i + 2, j, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i + 2, j, k) + rk2gtu_yz * vector(vfd_y, i + 2, j, k) + rk2gtu_zz * vector(vfd_z, i + 2, j, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i + 2, j, k) + rk2gtu_yy * vector(rk2Sfd_y, i + 2, j, k) + rk2gtu_yz * vector(rk2Sfd_z, i + 2, j, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i + 2, j, k) + rk2gtu_yz * vector(rk2Sfd_y, i + 2, j, k) + rk2gtu_zz * vector(rk2Sfd_z, i + 2, j, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i + 2, j, k) + rk2gtu_yy * vector(vfd_y, i + 2, j, k) + rk2gtu_yz * vector(vfd_z, i + 2, j, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i + 2, j, k) + rk2vfu_y * vector(vfd_y, i + 2, j, k) + rk2vfu_z * vector(vfd_z, i + 2, j, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i + 2, j, k) * vector(rk2Bfu_z, i + 2, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i + 2, j, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i + 2, j, k) * vector(rk2Bfu_x, i + 2, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i + 2, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i + 2, j, k) * vector(rk2Bfu_z, i + 2, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i + 2, j, k) * rk2vfu_z + vector(rk2Bfu_z, i + 2, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i + 2, j, k) * vector(rk2Bfu_y, i + 2, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i + 2, j, k) * rk2vfu_y + vector(rk2Bfu_y, i + 2, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i + 2, j, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i + 2, j, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i + 2, j, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i + 2, j, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i + 2, j, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i + 2, j, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i + 2, j, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i + 2, j, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i + 2, j, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i + 2, j, k) * vector(rk2Bfu_z, i + 2, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i + 2, j, k) * rk2vfu_z + vector(rk2Bfu_z, i + 2, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i + 2, j, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i + 2, j, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i + 2, j, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i + 2, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i + 2, j, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i + 2, j, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i + 2, j, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i + 2, j, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i + 2, j, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i + 2, j, k) * vector(rk2Bfu_y, i + 2, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i + 2, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i + 2, j, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i + 2, j, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i + 2, j, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i + 2, j, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i + 2, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i + 2, j, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i + 2, j, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i + 2, j, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i + 2, j, k));
						Fluxp2iSfd_x = FiSfd_x_mainI(vector(rk2Sfd_x, i + 2, j, k), vector(rk2gtd_xx, i + 2, j, k), vector(rk2gtd_xy, i + 2, j, k), vector(rk2gtd_xz, i + 2, j, k), vector(rk2Betau_x, i + 2, j, k), vector(rk2Alpha, i + 2, j, k), vector(TauT_xx, i + 2, j, k), vector(TauT_xy, i + 2, j, k), vector(TauT_xz, i + 2, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp2iSfd_y = FiSfd_y_mainI(vector(rk2Sfd_y, i + 2, j, k), vector(rk2gtd_xy, i + 2, j, k), vector(rk2gtd_yy, i + 2, j, k), vector(rk2gtd_yz, i + 2, j, k), vector(rk2Betau_x, i + 2, j, k), vector(rk2Alpha, i + 2, j, k), vector(TauT_xx, i + 2, j, k), vector(TauT_xy, i + 2, j, k), vector(TauT_xz, i + 2, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxp2iSfd_z = FiSfd_z_mainI(vector(rk2Sfd_z, i + 2, j, k), vector(rk2gtd_xz, i + 2, j, k), vector(rk2gtd_yz, i + 2, j, k), vector(rk2gtd_zz, i + 2, j, k), vector(rk2Betau_x, i + 2, j, k), vector(rk2Alpha, i + 2, j, k), vector(TauT_xx, i + 2, j, k), vector(TauT_xy, i + 2, j, k), vector(TauT_xz, i + 2, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxp2iBfu_x = FiBfu_x_mainI(vector(rk2phif, i + 2, j, k), vector(rk2Alpha, i + 2, j, k), rk2chi_max, rk2gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp2iBfu_y = FiBfu_y_mainI(vector(rk2Bfu_x, i + 2, j, k), vector(rk2Bfu_y, i + 2, j, k), vector(rk2phif, i + 2, j, k), vector(rk2Betau_x, i + 2, j, k), vector(rk2Betau_y, i + 2, j, k), vector(rk2Alpha, i + 2, j, k), vector(TauM_xy, i + 2, j, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2iBfu_z = FiBfu_z_mainI(vector(rk2Bfu_x, i + 2, j, k), vector(rk2Bfu_z, i + 2, j, k), vector(rk2phif, i + 2, j, k), vector(rk2Betau_x, i + 2, j, k), vector(rk2Betau_z, i + 2, j, k), vector(rk2Alpha, i + 2, j, k), vector(TauM_xz, i + 2, j, k), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2iDf = FiDf_mainI(vector(rk2Df, i + 2, j, k), vector(rk2Betau_x, i + 2, j, k), vector(rk2Alpha, i + 2, j, k), vector(TauN_x, i + 2, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp2iDYf = FiDYf_mainI(vector(rk2DYf, i + 2, j, k), vector(rk2Betau_x, i + 2, j, k), vector(rk2Alpha, i + 2, j, k), vector(TauNe_x, i + 2, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp2itauf = Fitauf_mainI(vector(rk2Df, i + 2, j, k), vector(rk2tauf, i + 2, j, k), vector(rk2Betau_x, i + 2, j, k), vector(rk2Alpha, i + 2, j, k), vector(TauN_x, i + 2, j, k), rk2sdetg, rk2Sfu_x, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp2iphif = Fiphif_mainI(vector(rk2Bfu_x, i + 2, j, k), vector(rk2phif, i + 2, j, k), vector(rk2Betau_x, i + 2, j, k), vector(rk2Alpha, i + 2, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i + 2, j, k) * (1.0 + vector(epsf, i + 2, j, k)) + vector(pf, i + 2, j, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i + 2, j, k) + rk2sqca - vector(sqcs, i + 2, j, k) * rk2sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk2Betau_x, i + 2, j, k)) + vector(rk2Alpha, i + 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk2Betau_x, i + 2, j, k)) + vector(rk2Alpha, i + 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk2Betau_x, i + 2, j, k)) + vector(rk2Alpha, i + 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk2Betau_x, i + 2, j, k)) + vector(rk2Alpha, i + 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk2Betau_x, i + 2, j, k)) + vector(rk2Alpha, i + 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk2Betau_x, i + 2, j, k)) + vector(rk2Alpha, i + 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk2Betau_x, i + 2, j, k)) + vector(rk2Alpha, i + 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk2Betau_x, i + 2, j, k)) + vector(rk2Alpha, i + 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk2Betau_x, i + 2, j, k)) + vector(rk2Alpha, i + 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk2Betau_x, i + 2, j, k)) + vector(rk2Alpha, i + 2, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i + 3, j, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i + 3, j, k) * vector(rk2Bfu_x, i + 3, j, k) + vector(rk2gtd_yy, i + 3, j, k) * vector(rk2Bfu_y, i + 3, j, k) + vector(rk2gtd_yz, i + 3, j, k) * vector(rk2Bfu_z, i + 3, j, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i + 3, j, k) * vector(rk2Bfu_x, i + 3, j, k) + vector(rk2gtd_yz, i + 3, j, k) * vector(rk2Bfu_y, i + 3, j, k) + vector(rk2gtd_zz, i + 3, j, k) * vector(rk2Bfu_z, i + 3, j, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i + 3, j, k) * vector(rk2Bfu_x, i + 3, j, k) + vector(rk2gtd_xy, i + 3, j, k) * vector(rk2Bfu_y, i + 3, j, k) + vector(rk2gtd_xz, i + 3, j, k) * vector(rk2Bfu_z, i + 3, j, k));
						rk2sqBf = vector(rk2Bfu_x, i + 3, j, k) * rk2Bfd_x + vector(rk2Bfu_y, i + 3, j, k) * rk2Bfd_y + vector(rk2Bfu_z, i + 3, j, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i + 3, j, k) * vector(rk2gtd_zz, i + 3, j, k) * vector(rk2gtd_xx, i + 3, j, k) + (-vector(rk2gtd_yy, i + 3, j, k) * (vector(rk2gtd_xz, i + 3, j, k) * vector(rk2gtd_xz, i + 3, j, k))) + (-vector(rk2gtd_zz, i + 3, j, k) * (vector(rk2gtd_xy, i + 3, j, k) * vector(rk2gtd_xy, i + 3, j, k))) + (-(vector(rk2gtd_yz, i + 3, j, k) * vector(rk2gtd_yz, i + 3, j, k)) * vector(rk2gtd_xx, i + 3, j, k)) + 2.0 * vector(rk2gtd_yz, i + 3, j, k) * vector(rk2gtd_xy, i + 3, j, k) * vector(rk2gtd_xz, i + 3, j, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i + 3, j, k) * vector(rk2gtd_yy, i + 3, j, k) - vector(rk2gtd_xy, i + 3, j, k) * vector(rk2gtd_xy, i + 3, j, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i + 3, j, k) * vector(rk2gtd_zz, i + 3, j, k) - vector(rk2gtd_yz, i + 3, j, k) * vector(rk2gtd_yz, i + 3, j, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i + 3, j, k) * vector(rk2gtd_yz, i + 3, j, k) - vector(rk2gtd_yy, i + 3, j, k) * vector(rk2gtd_xz, i + 3, j, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i + 3, j, k) * vector(rk2gtd_zz, i + 3, j, k)) + vector(rk2gtd_yz, i + 3, j, k) * vector(rk2gtd_xz, i + 3, j, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i + 3, j, k) + rk2gtu_xy * vector(rk2Sfd_y, i + 3, j, k) + rk2gtu_xz * vector(rk2Sfd_z, i + 3, j, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i + 3, j, k) + rk2gtu_xy * vector(vfd_y, i + 3, j, k) + rk2gtu_xz * vector(vfd_z, i + 3, j, k));
						rk2Bfvf = vector(rk2Bfu_x, i + 3, j, k) * vector(vfd_x, i + 3, j, k) + vector(rk2Bfu_y, i + 3, j, k) * vector(vfd_y, i + 3, j, k) + vector(rk2Bfu_z, i + 3, j, k) * vector(vfd_z, i + 3, j, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i + 3, j, k) * vector(rk2gtd_zz, i + 3, j, k) - vector(rk2gtd_xz, i + 3, j, k) * vector(rk2gtd_xz, i + 3, j, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i + 3, j, k) * vector(rk2gtd_yz, i + 3, j, k)) + vector(rk2gtd_xy, i + 3, j, k) * vector(rk2gtd_xz, i + 3, j, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i + 3, j, k) + rk2gtu_yz * vector(vfd_y, i + 3, j, k) + rk2gtu_zz * vector(vfd_z, i + 3, j, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i + 3, j, k) + rk2gtu_yy * vector(rk2Sfd_y, i + 3, j, k) + rk2gtu_yz * vector(rk2Sfd_z, i + 3, j, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i + 3, j, k) + rk2gtu_yz * vector(rk2Sfd_y, i + 3, j, k) + rk2gtu_zz * vector(rk2Sfd_z, i + 3, j, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i + 3, j, k) + rk2gtu_yy * vector(vfd_y, i + 3, j, k) + rk2gtu_yz * vector(vfd_z, i + 3, j, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i + 3, j, k) + rk2vfu_y * vector(vfd_y, i + 3, j, k) + rk2vfu_z * vector(vfd_z, i + 3, j, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i + 3, j, k) * vector(rk2Bfu_z, i + 3, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i + 3, j, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i + 3, j, k) * vector(rk2Bfu_x, i + 3, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i + 3, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i + 3, j, k) * vector(rk2Bfu_z, i + 3, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i + 3, j, k) * rk2vfu_z + vector(rk2Bfu_z, i + 3, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i + 3, j, k) * vector(rk2Bfu_y, i + 3, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i + 3, j, k) * rk2vfu_y + vector(rk2Bfu_y, i + 3, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i + 3, j, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i + 3, j, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i + 3, j, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i + 3, j, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i + 3, j, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i + 3, j, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i + 3, j, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i + 3, j, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i + 3, j, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i + 3, j, k) * vector(rk2Bfu_z, i + 3, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i + 3, j, k) * rk2vfu_z + vector(rk2Bfu_z, i + 3, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i + 3, j, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i + 3, j, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i + 3, j, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i + 3, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i + 3, j, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i + 3, j, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i + 3, j, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i + 3, j, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i + 3, j, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i + 3, j, k) * vector(rk2Bfu_y, i + 3, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i + 3, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i + 3, j, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i + 3, j, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i + 3, j, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i + 3, j, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i + 3, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i + 3, j, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i + 3, j, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i + 3, j, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i + 3, j, k));
						Fluxp3iSfd_x = FiSfd_x_mainI(vector(rk2Sfd_x, i + 3, j, k), vector(rk2gtd_xx, i + 3, j, k), vector(rk2gtd_xy, i + 3, j, k), vector(rk2gtd_xz, i + 3, j, k), vector(rk2Betau_x, i + 3, j, k), vector(rk2Alpha, i + 3, j, k), vector(TauT_xx, i + 3, j, k), vector(TauT_xy, i + 3, j, k), vector(TauT_xz, i + 3, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp3iSfd_y = FiSfd_y_mainI(vector(rk2Sfd_y, i + 3, j, k), vector(rk2gtd_xy, i + 3, j, k), vector(rk2gtd_yy, i + 3, j, k), vector(rk2gtd_yz, i + 3, j, k), vector(rk2Betau_x, i + 3, j, k), vector(rk2Alpha, i + 3, j, k), vector(TauT_xx, i + 3, j, k), vector(TauT_xy, i + 3, j, k), vector(TauT_xz, i + 3, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxp3iSfd_z = FiSfd_z_mainI(vector(rk2Sfd_z, i + 3, j, k), vector(rk2gtd_xz, i + 3, j, k), vector(rk2gtd_yz, i + 3, j, k), vector(rk2gtd_zz, i + 3, j, k), vector(rk2Betau_x, i + 3, j, k), vector(rk2Alpha, i + 3, j, k), vector(TauT_xx, i + 3, j, k), vector(TauT_xy, i + 3, j, k), vector(TauT_xz, i + 3, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxp3iBfu_x = FiBfu_x_mainI(vector(rk2phif, i + 3, j, k), vector(rk2Alpha, i + 3, j, k), rk2chi_max, rk2gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp3iBfu_y = FiBfu_y_mainI(vector(rk2Bfu_x, i + 3, j, k), vector(rk2Bfu_y, i + 3, j, k), vector(rk2phif, i + 3, j, k), vector(rk2Betau_x, i + 3, j, k), vector(rk2Betau_y, i + 3, j, k), vector(rk2Alpha, i + 3, j, k), vector(TauM_xy, i + 3, j, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3iBfu_z = FiBfu_z_mainI(vector(rk2Bfu_x, i + 3, j, k), vector(rk2Bfu_z, i + 3, j, k), vector(rk2phif, i + 3, j, k), vector(rk2Betau_x, i + 3, j, k), vector(rk2Betau_z, i + 3, j, k), vector(rk2Alpha, i + 3, j, k), vector(TauM_xz, i + 3, j, k), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3iDf = FiDf_mainI(vector(rk2Df, i + 3, j, k), vector(rk2Betau_x, i + 3, j, k), vector(rk2Alpha, i + 3, j, k), vector(TauN_x, i + 3, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp3iDYf = FiDYf_mainI(vector(rk2DYf, i + 3, j, k), vector(rk2Betau_x, i + 3, j, k), vector(rk2Alpha, i + 3, j, k), vector(TauNe_x, i + 3, j, k), rk2sdetg, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp3itauf = Fitauf_mainI(vector(rk2Df, i + 3, j, k), vector(rk2tauf, i + 3, j, k), vector(rk2Betau_x, i + 3, j, k), vector(rk2Alpha, i + 3, j, k), vector(TauN_x, i + 3, j, k), rk2sdetg, rk2Sfu_x, rk2vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp3iphif = Fiphif_mainI(vector(rk2Bfu_x, i + 3, j, k), vector(rk2phif, i + 3, j, k), vector(rk2Betau_x, i + 3, j, k), vector(rk2Alpha, i + 3, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i + 3, j, k) * (1.0 + vector(epsf, i + 3, j, k)) + vector(pf, i + 3, j, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i + 3, j, k) + rk2sqca - vector(sqcs, i + 3, j, k) * rk2sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk2Betau_x, i + 3, j, k)) + vector(rk2Alpha, i + 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk2Betau_x, i + 3, j, k)) + vector(rk2Alpha, i + 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk2Betau_x, i + 3, j, k)) + vector(rk2Alpha, i + 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk2Betau_x, i + 3, j, k)) + vector(rk2Alpha, i + 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk2Betau_x, i + 3, j, k)) + vector(rk2Alpha, i + 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk2Betau_x, i + 3, j, k)) + vector(rk2Alpha, i + 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk2Betau_x, i + 3, j, k)) + vector(rk2Alpha, i + 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk2Betau_x, i + 3, j, k)) + vector(rk2Alpha, i + 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk2Betau_x, i + 3, j, k)) + vector(rk2Alpha, i + 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk2Betau_x, i + 3, j, k)) + vector(rk2Alpha, i + 3, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_x) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_xx - (1.0 - rk2sqa) * (rk2vfu_x * rk2vfu_x)))))));
						LLFpiSfd_x = 0.5 * (FluxiSfd_x + fabs(SpeediSfd_x) * vector(rk2Sfd_x, i, j, k));
						LLFpiSfd_y = 0.5 * (FluxiSfd_y + fabs(SpeediSfd_y) * vector(rk2Sfd_y, i, j, k));
						LLFpiSfd_z = 0.5 * (FluxiSfd_z + fabs(SpeediSfd_z) * vector(rk2Sfd_z, i, j, k));
						LLFpiBfu_x = 0.5 * (FluxiBfu_x + fabs(SpeediBfu_x) * vector(rk2Bfu_x, i, j, k));
						LLFpiBfu_y = 0.5 * (FluxiBfu_y + fabs(SpeediBfu_y) * vector(rk2Bfu_y, i, j, k));
						LLFpiBfu_z = 0.5 * (FluxiBfu_z + fabs(SpeediBfu_z) * vector(rk2Bfu_z, i, j, k));
						LLFpiDf = 0.5 * (FluxiDf + fabs(SpeediDf) * vector(rk2Df, i, j, k));
						LLFpiDYf = 0.5 * (FluxiDYf + fabs(SpeediDYf) * vector(rk2DYf, i, j, k));
						LLFpitauf = 0.5 * (Fluxitauf + fabs(Speeditauf) * vector(rk2tauf, i, j, k));
						LLFpiphif = 0.5 * (Fluxiphif + fabs(Speediphif) * vector(rk2phif, i, j, k));
						LLFniSfd_x = 0.5 * (FluxiSfd_x - fabs(SpeediSfd_x) * vector(rk2Sfd_x, i, j, k));
						LLFniSfd_y = 0.5 * (FluxiSfd_y - fabs(SpeediSfd_y) * vector(rk2Sfd_y, i, j, k));
						LLFniSfd_z = 0.5 * (FluxiSfd_z - fabs(SpeediSfd_z) * vector(rk2Sfd_z, i, j, k));
						LLFniBfu_x = 0.5 * (FluxiBfu_x - fabs(SpeediBfu_x) * vector(rk2Bfu_x, i, j, k));
						LLFniBfu_y = 0.5 * (FluxiBfu_y - fabs(SpeediBfu_y) * vector(rk2Bfu_y, i, j, k));
						LLFniBfu_z = 0.5 * (FluxiBfu_z - fabs(SpeediBfu_z) * vector(rk2Bfu_z, i, j, k));
						LLFniDf = 0.5 * (FluxiDf - fabs(SpeediDf) * vector(rk2Df, i, j, k));
						LLFniDYf = 0.5 * (FluxiDYf - fabs(SpeediDYf) * vector(rk2DYf, i, j, k));
						LLFnitauf = 0.5 * (Fluxitauf - fabs(Speeditauf) * vector(rk2tauf, i, j, k));
						LLFniphif = 0.5 * (Fluxiphif - fabs(Speediphif) * vector(rk2phif, i, j, k));
						LLFp_m3iSfd_x = 0.5 * (Fluxm3iSfd_x + fabs(SpeediSfd_x) * vector(rk2Sfd_x, i - 3, j, k));
						LLFp_m3iSfd_y = 0.5 * (Fluxm3iSfd_y + fabs(SpeediSfd_y) * vector(rk2Sfd_y, i - 3, j, k));
						LLFp_m3iSfd_z = 0.5 * (Fluxm3iSfd_z + fabs(SpeediSfd_z) * vector(rk2Sfd_z, i - 3, j, k));
						LLFp_m3iBfu_x = 0.5 * (Fluxm3iBfu_x + fabs(SpeediBfu_x) * vector(rk2Bfu_x, i - 3, j, k));
						LLFp_m3iBfu_y = 0.5 * (Fluxm3iBfu_y + fabs(SpeediBfu_y) * vector(rk2Bfu_y, i - 3, j, k));
						LLFp_m3iBfu_z = 0.5 * (Fluxm3iBfu_z + fabs(SpeediBfu_z) * vector(rk2Bfu_z, i - 3, j, k));
						LLFp_m3iDf = 0.5 * (Fluxm3iDf + fabs(SpeediDf) * vector(rk2Df, i - 3, j, k));
						LLFp_m3iDYf = 0.5 * (Fluxm3iDYf + fabs(SpeediDYf) * vector(rk2DYf, i - 3, j, k));
						LLFp_m3itauf = 0.5 * (Fluxm3itauf + fabs(Speeditauf) * vector(rk2tauf, i - 3, j, k));
						LLFp_m3iphif = 0.5 * (Fluxm3iphif + fabs(Speediphif) * vector(rk2phif, i - 3, j, k));
						LLFn_m3iSfd_x = 0.5 * (Fluxm3iSfd_x - fabs(SpeediSfd_x) * vector(rk2Sfd_x, i - 3, j, k));
						LLFn_m3iSfd_y = 0.5 * (Fluxm3iSfd_y - fabs(SpeediSfd_y) * vector(rk2Sfd_y, i - 3, j, k));
						LLFn_m3iSfd_z = 0.5 * (Fluxm3iSfd_z - fabs(SpeediSfd_z) * vector(rk2Sfd_z, i - 3, j, k));
						LLFn_m3iBfu_x = 0.5 * (Fluxm3iBfu_x - fabs(SpeediBfu_x) * vector(rk2Bfu_x, i - 3, j, k));
						LLFn_m3iBfu_y = 0.5 * (Fluxm3iBfu_y - fabs(SpeediBfu_y) * vector(rk2Bfu_y, i - 3, j, k));
						LLFn_m3iBfu_z = 0.5 * (Fluxm3iBfu_z - fabs(SpeediBfu_z) * vector(rk2Bfu_z, i - 3, j, k));
						LLFn_m3iDf = 0.5 * (Fluxm3iDf - fabs(SpeediDf) * vector(rk2Df, i - 3, j, k));
						LLFn_m3iDYf = 0.5 * (Fluxm3iDYf - fabs(SpeediDYf) * vector(rk2DYf, i - 3, j, k));
						LLFn_m3itauf = 0.5 * (Fluxm3itauf - fabs(Speeditauf) * vector(rk2tauf, i - 3, j, k));
						LLFn_m3iphif = 0.5 * (Fluxm3iphif - fabs(Speediphif) * vector(rk2phif, i - 3, j, k));
						LLFp_m2iSfd_x = 0.5 * (Fluxm2iSfd_x + fabs(SpeediSfd_x) * vector(rk2Sfd_x, i - 2, j, k));
						LLFp_m2iSfd_y = 0.5 * (Fluxm2iSfd_y + fabs(SpeediSfd_y) * vector(rk2Sfd_y, i - 2, j, k));
						LLFp_m2iSfd_z = 0.5 * (Fluxm2iSfd_z + fabs(SpeediSfd_z) * vector(rk2Sfd_z, i - 2, j, k));
						LLFp_m2iBfu_x = 0.5 * (Fluxm2iBfu_x + fabs(SpeediBfu_x) * vector(rk2Bfu_x, i - 2, j, k));
						LLFp_m2iBfu_y = 0.5 * (Fluxm2iBfu_y + fabs(SpeediBfu_y) * vector(rk2Bfu_y, i - 2, j, k));
						LLFp_m2iBfu_z = 0.5 * (Fluxm2iBfu_z + fabs(SpeediBfu_z) * vector(rk2Bfu_z, i - 2, j, k));
						LLFp_m2iDf = 0.5 * (Fluxm2iDf + fabs(SpeediDf) * vector(rk2Df, i - 2, j, k));
						LLFp_m2iDYf = 0.5 * (Fluxm2iDYf + fabs(SpeediDYf) * vector(rk2DYf, i - 2, j, k));
						LLFp_m2itauf = 0.5 * (Fluxm2itauf + fabs(Speeditauf) * vector(rk2tauf, i - 2, j, k));
						LLFp_m2iphif = 0.5 * (Fluxm2iphif + fabs(Speediphif) * vector(rk2phif, i - 2, j, k));
						LLFn_m2iSfd_x = 0.5 * (Fluxm2iSfd_x - fabs(SpeediSfd_x) * vector(rk2Sfd_x, i - 2, j, k));
						LLFn_m2iSfd_y = 0.5 * (Fluxm2iSfd_y - fabs(SpeediSfd_y) * vector(rk2Sfd_y, i - 2, j, k));
						LLFn_m2iSfd_z = 0.5 * (Fluxm2iSfd_z - fabs(SpeediSfd_z) * vector(rk2Sfd_z, i - 2, j, k));
						LLFn_m2iBfu_x = 0.5 * (Fluxm2iBfu_x - fabs(SpeediBfu_x) * vector(rk2Bfu_x, i - 2, j, k));
						LLFn_m2iBfu_y = 0.5 * (Fluxm2iBfu_y - fabs(SpeediBfu_y) * vector(rk2Bfu_y, i - 2, j, k));
						LLFn_m2iBfu_z = 0.5 * (Fluxm2iBfu_z - fabs(SpeediBfu_z) * vector(rk2Bfu_z, i - 2, j, k));
						LLFn_m2iDf = 0.5 * (Fluxm2iDf - fabs(SpeediDf) * vector(rk2Df, i - 2, j, k));
						LLFn_m2iDYf = 0.5 * (Fluxm2iDYf - fabs(SpeediDYf) * vector(rk2DYf, i - 2, j, k));
						LLFn_m2itauf = 0.5 * (Fluxm2itauf - fabs(Speeditauf) * vector(rk2tauf, i - 2, j, k));
						LLFn_m2iphif = 0.5 * (Fluxm2iphif - fabs(Speediphif) * vector(rk2phif, i - 2, j, k));
						LLFp_m1iSfd_x = 0.5 * (Fluxm1iSfd_x + fabs(SpeediSfd_x) * vector(rk2Sfd_x, i - 1, j, k));
						LLFp_m1iSfd_y = 0.5 * (Fluxm1iSfd_y + fabs(SpeediSfd_y) * vector(rk2Sfd_y, i - 1, j, k));
						LLFp_m1iSfd_z = 0.5 * (Fluxm1iSfd_z + fabs(SpeediSfd_z) * vector(rk2Sfd_z, i - 1, j, k));
						LLFp_m1iBfu_x = 0.5 * (Fluxm1iBfu_x + fabs(SpeediBfu_x) * vector(rk2Bfu_x, i - 1, j, k));
						LLFp_m1iBfu_y = 0.5 * (Fluxm1iBfu_y + fabs(SpeediBfu_y) * vector(rk2Bfu_y, i - 1, j, k));
						LLFp_m1iBfu_z = 0.5 * (Fluxm1iBfu_z + fabs(SpeediBfu_z) * vector(rk2Bfu_z, i - 1, j, k));
						LLFp_m1iDf = 0.5 * (Fluxm1iDf + fabs(SpeediDf) * vector(rk2Df, i - 1, j, k));
						LLFp_m1iDYf = 0.5 * (Fluxm1iDYf + fabs(SpeediDYf) * vector(rk2DYf, i - 1, j, k));
						LLFp_m1itauf = 0.5 * (Fluxm1itauf + fabs(Speeditauf) * vector(rk2tauf, i - 1, j, k));
						LLFp_m1iphif = 0.5 * (Fluxm1iphif + fabs(Speediphif) * vector(rk2phif, i - 1, j, k));
						LLFn_m1iSfd_x = 0.5 * (Fluxm1iSfd_x - fabs(SpeediSfd_x) * vector(rk2Sfd_x, i - 1, j, k));
						LLFn_m1iSfd_y = 0.5 * (Fluxm1iSfd_y - fabs(SpeediSfd_y) * vector(rk2Sfd_y, i - 1, j, k));
						LLFn_m1iSfd_z = 0.5 * (Fluxm1iSfd_z - fabs(SpeediSfd_z) * vector(rk2Sfd_z, i - 1, j, k));
						LLFn_m1iBfu_x = 0.5 * (Fluxm1iBfu_x - fabs(SpeediBfu_x) * vector(rk2Bfu_x, i - 1, j, k));
						LLFn_m1iBfu_y = 0.5 * (Fluxm1iBfu_y - fabs(SpeediBfu_y) * vector(rk2Bfu_y, i - 1, j, k));
						LLFn_m1iBfu_z = 0.5 * (Fluxm1iBfu_z - fabs(SpeediBfu_z) * vector(rk2Bfu_z, i - 1, j, k));
						LLFn_m1iDf = 0.5 * (Fluxm1iDf - fabs(SpeediDf) * vector(rk2Df, i - 1, j, k));
						LLFn_m1iDYf = 0.5 * (Fluxm1iDYf - fabs(SpeediDYf) * vector(rk2DYf, i - 1, j, k));
						LLFn_m1itauf = 0.5 * (Fluxm1itauf - fabs(Speeditauf) * vector(rk2tauf, i - 1, j, k));
						LLFn_m1iphif = 0.5 * (Fluxm1iphif - fabs(Speediphif) * vector(rk2phif, i - 1, j, k));
						LLFp_p1iSfd_x = 0.5 * (Fluxp1iSfd_x + fabs(SpeediSfd_x) * vector(rk2Sfd_x, i + 1, j, k));
						LLFp_p1iSfd_y = 0.5 * (Fluxp1iSfd_y + fabs(SpeediSfd_y) * vector(rk2Sfd_y, i + 1, j, k));
						LLFp_p1iSfd_z = 0.5 * (Fluxp1iSfd_z + fabs(SpeediSfd_z) * vector(rk2Sfd_z, i + 1, j, k));
						LLFp_p1iBfu_x = 0.5 * (Fluxp1iBfu_x + fabs(SpeediBfu_x) * vector(rk2Bfu_x, i + 1, j, k));
						LLFp_p1iBfu_y = 0.5 * (Fluxp1iBfu_y + fabs(SpeediBfu_y) * vector(rk2Bfu_y, i + 1, j, k));
						LLFp_p1iBfu_z = 0.5 * (Fluxp1iBfu_z + fabs(SpeediBfu_z) * vector(rk2Bfu_z, i + 1, j, k));
						LLFp_p1iDf = 0.5 * (Fluxp1iDf + fabs(SpeediDf) * vector(rk2Df, i + 1, j, k));
						LLFp_p1iDYf = 0.5 * (Fluxp1iDYf + fabs(SpeediDYf) * vector(rk2DYf, i + 1, j, k));
						LLFp_p1itauf = 0.5 * (Fluxp1itauf + fabs(Speeditauf) * vector(rk2tauf, i + 1, j, k));
						LLFp_p1iphif = 0.5 * (Fluxp1iphif + fabs(Speediphif) * vector(rk2phif, i + 1, j, k));
						LLFn_p1iSfd_x = 0.5 * (Fluxp1iSfd_x - fabs(SpeediSfd_x) * vector(rk2Sfd_x, i + 1, j, k));
						LLFn_p1iSfd_y = 0.5 * (Fluxp1iSfd_y - fabs(SpeediSfd_y) * vector(rk2Sfd_y, i + 1, j, k));
						LLFn_p1iSfd_z = 0.5 * (Fluxp1iSfd_z - fabs(SpeediSfd_z) * vector(rk2Sfd_z, i + 1, j, k));
						LLFn_p1iBfu_x = 0.5 * (Fluxp1iBfu_x - fabs(SpeediBfu_x) * vector(rk2Bfu_x, i + 1, j, k));
						LLFn_p1iBfu_y = 0.5 * (Fluxp1iBfu_y - fabs(SpeediBfu_y) * vector(rk2Bfu_y, i + 1, j, k));
						LLFn_p1iBfu_z = 0.5 * (Fluxp1iBfu_z - fabs(SpeediBfu_z) * vector(rk2Bfu_z, i + 1, j, k));
						LLFn_p1iDf = 0.5 * (Fluxp1iDf - fabs(SpeediDf) * vector(rk2Df, i + 1, j, k));
						LLFn_p1iDYf = 0.5 * (Fluxp1iDYf - fabs(SpeediDYf) * vector(rk2DYf, i + 1, j, k));
						LLFn_p1itauf = 0.5 * (Fluxp1itauf - fabs(Speeditauf) * vector(rk2tauf, i + 1, j, k));
						LLFn_p1iphif = 0.5 * (Fluxp1iphif - fabs(Speediphif) * vector(rk2phif, i + 1, j, k));
						LLFp_p2iSfd_x = 0.5 * (Fluxp2iSfd_x + fabs(SpeediSfd_x) * vector(rk2Sfd_x, i + 2, j, k));
						LLFp_p2iSfd_y = 0.5 * (Fluxp2iSfd_y + fabs(SpeediSfd_y) * vector(rk2Sfd_y, i + 2, j, k));
						LLFp_p2iSfd_z = 0.5 * (Fluxp2iSfd_z + fabs(SpeediSfd_z) * vector(rk2Sfd_z, i + 2, j, k));
						LLFp_p2iBfu_x = 0.5 * (Fluxp2iBfu_x + fabs(SpeediBfu_x) * vector(rk2Bfu_x, i + 2, j, k));
						LLFp_p2iBfu_y = 0.5 * (Fluxp2iBfu_y + fabs(SpeediBfu_y) * vector(rk2Bfu_y, i + 2, j, k));
						LLFp_p2iBfu_z = 0.5 * (Fluxp2iBfu_z + fabs(SpeediBfu_z) * vector(rk2Bfu_z, i + 2, j, k));
						LLFp_p2iDf = 0.5 * (Fluxp2iDf + fabs(SpeediDf) * vector(rk2Df, i + 2, j, k));
						LLFp_p2iDYf = 0.5 * (Fluxp2iDYf + fabs(SpeediDYf) * vector(rk2DYf, i + 2, j, k));
						LLFp_p2itauf = 0.5 * (Fluxp2itauf + fabs(Speeditauf) * vector(rk2tauf, i + 2, j, k));
						LLFp_p2iphif = 0.5 * (Fluxp2iphif + fabs(Speediphif) * vector(rk2phif, i + 2, j, k));
						LLFn_p2iSfd_x = 0.5 * (Fluxp2iSfd_x - fabs(SpeediSfd_x) * vector(rk2Sfd_x, i + 2, j, k));
						LLFn_p2iSfd_y = 0.5 * (Fluxp2iSfd_y - fabs(SpeediSfd_y) * vector(rk2Sfd_y, i + 2, j, k));
						LLFn_p2iSfd_z = 0.5 * (Fluxp2iSfd_z - fabs(SpeediSfd_z) * vector(rk2Sfd_z, i + 2, j, k));
						LLFn_p2iBfu_x = 0.5 * (Fluxp2iBfu_x - fabs(SpeediBfu_x) * vector(rk2Bfu_x, i + 2, j, k));
						LLFn_p2iBfu_y = 0.5 * (Fluxp2iBfu_y - fabs(SpeediBfu_y) * vector(rk2Bfu_y, i + 2, j, k));
						LLFn_p2iBfu_z = 0.5 * (Fluxp2iBfu_z - fabs(SpeediBfu_z) * vector(rk2Bfu_z, i + 2, j, k));
						LLFn_p2iDf = 0.5 * (Fluxp2iDf - fabs(SpeediDf) * vector(rk2Df, i + 2, j, k));
						LLFn_p2iDYf = 0.5 * (Fluxp2iDYf - fabs(SpeediDYf) * vector(rk2DYf, i + 2, j, k));
						LLFn_p2itauf = 0.5 * (Fluxp2itauf - fabs(Speeditauf) * vector(rk2tauf, i + 2, j, k));
						LLFn_p2iphif = 0.5 * (Fluxp2iphif - fabs(Speediphif) * vector(rk2phif, i + 2, j, k));
						LLFp_p3iSfd_x = 0.5 * (Fluxp3iSfd_x + fabs(SpeediSfd_x) * vector(rk2Sfd_x, i + 3, j, k));
						LLFp_p3iSfd_y = 0.5 * (Fluxp3iSfd_y + fabs(SpeediSfd_y) * vector(rk2Sfd_y, i + 3, j, k));
						LLFp_p3iSfd_z = 0.5 * (Fluxp3iSfd_z + fabs(SpeediSfd_z) * vector(rk2Sfd_z, i + 3, j, k));
						LLFp_p3iBfu_x = 0.5 * (Fluxp3iBfu_x + fabs(SpeediBfu_x) * vector(rk2Bfu_x, i + 3, j, k));
						LLFp_p3iBfu_y = 0.5 * (Fluxp3iBfu_y + fabs(SpeediBfu_y) * vector(rk2Bfu_y, i + 3, j, k));
						LLFp_p3iBfu_z = 0.5 * (Fluxp3iBfu_z + fabs(SpeediBfu_z) * vector(rk2Bfu_z, i + 3, j, k));
						LLFp_p3iDf = 0.5 * (Fluxp3iDf + fabs(SpeediDf) * vector(rk2Df, i + 3, j, k));
						LLFp_p3iDYf = 0.5 * (Fluxp3iDYf + fabs(SpeediDYf) * vector(rk2DYf, i + 3, j, k));
						LLFp_p3itauf = 0.5 * (Fluxp3itauf + fabs(Speeditauf) * vector(rk2tauf, i + 3, j, k));
						LLFp_p3iphif = 0.5 * (Fluxp3iphif + fabs(Speediphif) * vector(rk2phif, i + 3, j, k));
						LLFn_p3iSfd_x = 0.5 * (Fluxp3iSfd_x - fabs(SpeediSfd_x) * vector(rk2Sfd_x, i + 3, j, k));
						LLFn_p3iSfd_y = 0.5 * (Fluxp3iSfd_y - fabs(SpeediSfd_y) * vector(rk2Sfd_y, i + 3, j, k));
						LLFn_p3iSfd_z = 0.5 * (Fluxp3iSfd_z - fabs(SpeediSfd_z) * vector(rk2Sfd_z, i + 3, j, k));
						LLFn_p3iBfu_x = 0.5 * (Fluxp3iBfu_x - fabs(SpeediBfu_x) * vector(rk2Bfu_x, i + 3, j, k));
						LLFn_p3iBfu_y = 0.5 * (Fluxp3iBfu_y - fabs(SpeediBfu_y) * vector(rk2Bfu_y, i + 3, j, k));
						LLFn_p3iBfu_z = 0.5 * (Fluxp3iBfu_z - fabs(SpeediBfu_z) * vector(rk2Bfu_z, i + 3, j, k));
						LLFn_p3iDf = 0.5 * (Fluxp3iDf - fabs(SpeediDf) * vector(rk2Df, i + 3, j, k));
						LLFn_p3iDYf = 0.5 * (Fluxp3iDYf - fabs(SpeediDYf) * vector(rk2DYf, i + 3, j, k));
						LLFn_p3itauf = 0.5 * (Fluxp3itauf - fabs(Speeditauf) * vector(rk2tauf, i + 3, j, k));
						LLFn_p3iphif = 0.5 * (Fluxp3iphif - fabs(Speediphif) * vector(rk2phif, i + 3, j, k));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_yy, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_zz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_xy, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_xz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2sqBf = vector(rk2Bfu_x, i, j, k) * rk2Bfd_x + vector(rk2Bfu_y, i, j, k) * rk2Bfd_y + vector(rk2Bfu_z, i, j, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) * vector(rk2gtd_xx, i, j, k) + (-vector(rk2gtd_yy, i, j, k) * (vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k))) + (-vector(rk2gtd_zz, i, j, k) * (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k))) + (-(vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k)) * vector(rk2gtd_xx, i, j, k)) + 2.0 * vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yy, i, j, k) - vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) - vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_yz, i, j, k) - vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_zz, i, j, k)) + vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j, k) + rk2gtu_xy * vector(rk2Sfd_y, i, j, k) + rk2gtu_xz * vector(rk2Sfd_z, i, j, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j, k) + rk2gtu_xy * vector(vfd_y, i, j, k) + rk2gtu_xz * vector(vfd_z, i, j, k));
						rk2Bfvf = vector(rk2Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk2Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk2Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_zz, i, j, k) - vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yz, i, j, k)) + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j, k) + rk2gtu_yz * vector(vfd_y, i, j, k) + rk2gtu_zz * vector(vfd_z, i, j, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j, k) + rk2gtu_yy * vector(rk2Sfd_y, i, j, k) + rk2gtu_yz * vector(rk2Sfd_z, i, j, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j, k) + rk2gtu_yz * vector(rk2Sfd_y, i, j, k) + rk2gtu_zz * vector(rk2Sfd_z, i, j, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j, k) + rk2gtu_yy * vector(vfd_y, i, j, k) + rk2gtu_yz * vector(vfd_z, i, j, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j, k) + rk2vfu_y * vector(vfd_y, i, j, k) + rk2vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j, k) * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j, k) * vector(rk2Bfu_x, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k) * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k) * rk2vfu_z + vector(rk2Bfu_z, i, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k) * vector(rk2Bfu_y, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k) * rk2vfu_y + vector(rk2Bfu_y, i, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k) * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k) * rk2vfu_z + vector(rk2Bfu_z, i, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j, k) * vector(rk2Bfu_y, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j, k));
						FluxjSfd_x = FjSfd_x_mainI(vector(rk2Sfd_x, i, j, k), vector(rk2gtd_xx, i, j, k), vector(rk2gtd_xy, i, j, k), vector(rk2gtd_xz, i, j, k), vector(rk2Betau_y, i, j, k), vector(rk2Alpha, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_yy, i, j, k), vector(TauT_yz, i, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_yx, dx, simPlat_dt, ilast, jlast);
						FluxjSfd_y = FjSfd_y_mainI(vector(rk2Sfd_y, i, j, k), vector(rk2gtd_xy, i, j, k), vector(rk2gtd_yy, i, j, k), vector(rk2gtd_yz, i, j, k), vector(rk2Betau_y, i, j, k), vector(rk2Alpha, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_yy, i, j, k), vector(TauT_yz, i, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_yy, dx, simPlat_dt, ilast, jlast);
						FluxjSfd_z = FjSfd_z_mainI(vector(rk2Sfd_z, i, j, k), vector(rk2gtd_xz, i, j, k), vector(rk2gtd_yz, i, j, k), vector(rk2gtd_zz, i, j, k), vector(rk2Betau_y, i, j, k), vector(rk2Alpha, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_yy, i, j, k), vector(TauT_yz, i, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_yz, dx, simPlat_dt, ilast, jlast);
						FluxjBfu_x = FjBfu_x_mainI(vector(rk2Bfu_x, i, j, k), vector(rk2Bfu_y, i, j, k), vector(rk2phif, i, j, k), vector(rk2Betau_x, i, j, k), vector(rk2Betau_y, i, j, k), vector(rk2Alpha, i, j, k), vector(TauM_xy, i, j, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						FluxjBfu_y = FjBfu_y_mainI(vector(rk2phif, i, j, k), vector(rk2Alpha, i, j, k), rk2chi_max, rk2gtu_yy, dx, simPlat_dt, ilast, jlast);
						FluxjBfu_z = FjBfu_z_mainI(vector(rk2Bfu_y, i, j, k), vector(rk2Bfu_z, i, j, k), vector(rk2phif, i, j, k), vector(rk2Betau_y, i, j, k), vector(rk2Betau_z, i, j, k), vector(rk2Alpha, i, j, k), vector(TauM_yz, i, j, k), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxjDf = FjDf_mainI(vector(rk2Df, i, j, k), vector(rk2Betau_y, i, j, k), vector(rk2Alpha, i, j, k), vector(TauN_y, i, j, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						FluxjDYf = FjDYf_mainI(vector(rk2DYf, i, j, k), vector(rk2Betau_y, i, j, k), vector(rk2Alpha, i, j, k), vector(TauNe_y, i, j, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxjtauf = Fjtauf_mainI(vector(rk2Df, i, j, k), vector(rk2tauf, i, j, k), vector(rk2Betau_y, i, j, k), vector(rk2Alpha, i, j, k), vector(TauN_y, i, j, k), rk2sdetg, rk2Sfu_y, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxjphif = Fjphif_mainI(vector(rk2Bfu_y, i, j, k), vector(rk2phif, i, j, k), vector(rk2Betau_y, i, j, k), vector(rk2Alpha, i, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j, k) + rk2sqca - vector(sqcs, i, j, k) * rk2sqca;
						SpeedjSfd_x = fabs(fabs(vector(rk2Betau_y, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y))))));
						SpeedjSfd_y = fabs(fabs(vector(rk2Betau_y, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y))))));
						SpeedjSfd_z = fabs(fabs(vector(rk2Betau_y, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y))))));
						SpeedjBfu_x = fabs(fabs(vector(rk2Betau_y, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y))))));
						SpeedjBfu_y = fabs(fabs(vector(rk2Betau_y, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y))))));
						SpeedjBfu_z = fabs(fabs(vector(rk2Betau_y, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y))))));
						SpeedjDf = fabs(fabs(vector(rk2Betau_y, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y))))));
						SpeedjDYf = fabs(fabs(vector(rk2Betau_y, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y))))));
						Speedjtauf = fabs(fabs(vector(rk2Betau_y, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y))))));
						Speedjphif = fabs(fabs(vector(rk2Betau_y, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j - 3, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j - 3, k) * vector(rk2Bfu_x, i, j - 3, k) + vector(rk2gtd_yy, i, j - 3, k) * vector(rk2Bfu_y, i, j - 3, k) + vector(rk2gtd_yz, i, j - 3, k) * vector(rk2Bfu_z, i, j - 3, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j - 3, k) * vector(rk2Bfu_x, i, j - 3, k) + vector(rk2gtd_yz, i, j - 3, k) * vector(rk2Bfu_y, i, j - 3, k) + vector(rk2gtd_zz, i, j - 3, k) * vector(rk2Bfu_z, i, j - 3, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j - 3, k) * vector(rk2Bfu_x, i, j - 3, k) + vector(rk2gtd_xy, i, j - 3, k) * vector(rk2Bfu_y, i, j - 3, k) + vector(rk2gtd_xz, i, j - 3, k) * vector(rk2Bfu_z, i, j - 3, k));
						rk2sqBf = vector(rk2Bfu_x, i, j - 3, k) * rk2Bfd_x + vector(rk2Bfu_y, i, j - 3, k) * rk2Bfd_y + vector(rk2Bfu_z, i, j - 3, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j - 3, k) * vector(rk2gtd_zz, i, j - 3, k) * vector(rk2gtd_xx, i, j - 3, k) + (-vector(rk2gtd_yy, i, j - 3, k) * (vector(rk2gtd_xz, i, j - 3, k) * vector(rk2gtd_xz, i, j - 3, k))) + (-vector(rk2gtd_zz, i, j - 3, k) * (vector(rk2gtd_xy, i, j - 3, k) * vector(rk2gtd_xy, i, j - 3, k))) + (-(vector(rk2gtd_yz, i, j - 3, k) * vector(rk2gtd_yz, i, j - 3, k)) * vector(rk2gtd_xx, i, j - 3, k)) + 2.0 * vector(rk2gtd_yz, i, j - 3, k) * vector(rk2gtd_xy, i, j - 3, k) * vector(rk2gtd_xz, i, j - 3, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j - 3, k) * vector(rk2gtd_yy, i, j - 3, k) - vector(rk2gtd_xy, i, j - 3, k) * vector(rk2gtd_xy, i, j - 3, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j - 3, k) * vector(rk2gtd_zz, i, j - 3, k) - vector(rk2gtd_yz, i, j - 3, k) * vector(rk2gtd_yz, i, j - 3, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j - 3, k) * vector(rk2gtd_yz, i, j - 3, k) - vector(rk2gtd_yy, i, j - 3, k) * vector(rk2gtd_xz, i, j - 3, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j - 3, k) * vector(rk2gtd_zz, i, j - 3, k)) + vector(rk2gtd_yz, i, j - 3, k) * vector(rk2gtd_xz, i, j - 3, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j - 3, k) + rk2gtu_xy * vector(rk2Sfd_y, i, j - 3, k) + rk2gtu_xz * vector(rk2Sfd_z, i, j - 3, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j - 3, k) + rk2gtu_xy * vector(vfd_y, i, j - 3, k) + rk2gtu_xz * vector(vfd_z, i, j - 3, k));
						rk2Bfvf = vector(rk2Bfu_x, i, j - 3, k) * vector(vfd_x, i, j - 3, k) + vector(rk2Bfu_y, i, j - 3, k) * vector(vfd_y, i, j - 3, k) + vector(rk2Bfu_z, i, j - 3, k) * vector(vfd_z, i, j - 3, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j - 3, k) * vector(rk2gtd_zz, i, j - 3, k) - vector(rk2gtd_xz, i, j - 3, k) * vector(rk2gtd_xz, i, j - 3, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j - 3, k) * vector(rk2gtd_yz, i, j - 3, k)) + vector(rk2gtd_xy, i, j - 3, k) * vector(rk2gtd_xz, i, j - 3, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j - 3, k) + rk2gtu_yz * vector(vfd_y, i, j - 3, k) + rk2gtu_zz * vector(vfd_z, i, j - 3, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j - 3, k) + rk2gtu_yy * vector(rk2Sfd_y, i, j - 3, k) + rk2gtu_yz * vector(rk2Sfd_z, i, j - 3, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j - 3, k) + rk2gtu_yz * vector(rk2Sfd_y, i, j - 3, k) + rk2gtu_zz * vector(rk2Sfd_z, i, j - 3, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j - 3, k) + rk2gtu_yy * vector(vfd_y, i, j - 3, k) + rk2gtu_yz * vector(vfd_z, i, j - 3, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j - 3, k) + rk2vfu_y * vector(vfd_y, i, j - 3, k) + rk2vfu_z * vector(vfd_z, i, j - 3, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j - 3, k) * vector(rk2Bfu_z, i, j - 3, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j - 3, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j - 3, k) * vector(rk2Bfu_x, i, j - 3, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j - 3, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j - 3, k) * vector(rk2Bfu_z, i, j - 3, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j - 3, k) * rk2vfu_z + vector(rk2Bfu_z, i, j - 3, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j - 3, k) * vector(rk2Bfu_y, i, j - 3, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j - 3, k) * rk2vfu_y + vector(rk2Bfu_y, i, j - 3, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j - 3, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j - 3, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j - 3, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j - 3, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j - 3, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j - 3, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j - 3, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j - 3, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j - 3, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j - 3, k) * vector(rk2Bfu_z, i, j - 3, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j - 3, k) * rk2vfu_z + vector(rk2Bfu_z, i, j - 3, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j - 3, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j - 3, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j - 3, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j - 3, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j - 3, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j - 3, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j - 3, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j - 3, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j - 3, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j - 3, k) * vector(rk2Bfu_y, i, j - 3, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j - 3, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j - 3, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j - 3, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j - 3, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j - 3, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j - 3, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j - 3, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j - 3, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j - 3, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j - 3, k));
						Fluxm3jSfd_x = FjSfd_x_mainI(vector(rk2Sfd_x, i, j - 3, k), vector(rk2gtd_xx, i, j - 3, k), vector(rk2gtd_xy, i, j - 3, k), vector(rk2gtd_xz, i, j - 3, k), vector(rk2Betau_y, i, j - 3, k), vector(rk2Alpha, i, j - 3, k), vector(TauT_xy, i, j - 3, k), vector(TauT_yy, i, j - 3, k), vector(TauT_yz, i, j - 3, k), rk2inv_chi, rk2sdetg, rk2Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxm3jSfd_y = FjSfd_y_mainI(vector(rk2Sfd_y, i, j - 3, k), vector(rk2gtd_xy, i, j - 3, k), vector(rk2gtd_yy, i, j - 3, k), vector(rk2gtd_yz, i, j - 3, k), vector(rk2Betau_y, i, j - 3, k), vector(rk2Alpha, i, j - 3, k), vector(TauT_xy, i, j - 3, k), vector(TauT_yy, i, j - 3, k), vector(TauT_yz, i, j - 3, k), rk2inv_chi, rk2sdetg, rk2Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm3jSfd_z = FjSfd_z_mainI(vector(rk2Sfd_z, i, j - 3, k), vector(rk2gtd_xz, i, j - 3, k), vector(rk2gtd_yz, i, j - 3, k), vector(rk2gtd_zz, i, j - 3, k), vector(rk2Betau_y, i, j - 3, k), vector(rk2Alpha, i, j - 3, k), vector(TauT_xy, i, j - 3, k), vector(TauT_yy, i, j - 3, k), vector(TauT_yz, i, j - 3, k), rk2inv_chi, rk2sdetg, rk2Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxm3jBfu_x = FjBfu_x_mainI(vector(rk2Bfu_x, i, j - 3, k), vector(rk2Bfu_y, i, j - 3, k), vector(rk2phif, i, j - 3, k), vector(rk2Betau_x, i, j - 3, k), vector(rk2Betau_y, i, j - 3, k), vector(rk2Alpha, i, j - 3, k), vector(TauM_xy, i, j - 3, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3jBfu_y = FjBfu_y_mainI(vector(rk2phif, i, j - 3, k), vector(rk2Alpha, i, j - 3, k), rk2chi_max, rk2gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm3jBfu_z = FjBfu_z_mainI(vector(rk2Bfu_y, i, j - 3, k), vector(rk2Bfu_z, i, j - 3, k), vector(rk2phif, i, j - 3, k), vector(rk2Betau_y, i, j - 3, k), vector(rk2Betau_z, i, j - 3, k), vector(rk2Alpha, i, j - 3, k), vector(TauM_yz, i, j - 3, k), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3jDf = FjDf_mainI(vector(rk2Df, i, j - 3, k), vector(rk2Betau_y, i, j - 3, k), vector(rk2Alpha, i, j - 3, k), vector(TauN_y, i, j - 3, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3jDYf = FjDYf_mainI(vector(rk2DYf, i, j - 3, k), vector(rk2Betau_y, i, j - 3, k), vector(rk2Alpha, i, j - 3, k), vector(TauNe_y, i, j - 3, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3jtauf = Fjtauf_mainI(vector(rk2Df, i, j - 3, k), vector(rk2tauf, i, j - 3, k), vector(rk2Betau_y, i, j - 3, k), vector(rk2Alpha, i, j - 3, k), vector(TauN_y, i, j - 3, k), rk2sdetg, rk2Sfu_y, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3jphif = Fjphif_mainI(vector(rk2Bfu_y, i, j - 3, k), vector(rk2phif, i, j - 3, k), vector(rk2Betau_y, i, j - 3, k), vector(rk2Alpha, i, j - 3, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j - 3, k) * (1.0 + vector(epsf, i, j - 3, k)) + vector(pf, i, j - 3, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j - 3, k) + rk2sqca - vector(sqcs, i, j - 3, k) * rk2sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk2Betau_y, i, j - 3, k)) + vector(rk2Alpha, i, j - 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk2Betau_y, i, j - 3, k)) + vector(rk2Alpha, i, j - 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk2Betau_y, i, j - 3, k)) + vector(rk2Alpha, i, j - 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk2Betau_y, i, j - 3, k)) + vector(rk2Alpha, i, j - 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk2Betau_y, i, j - 3, k)) + vector(rk2Alpha, i, j - 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk2Betau_y, i, j - 3, k)) + vector(rk2Alpha, i, j - 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk2Betau_y, i, j - 3, k)) + vector(rk2Alpha, i, j - 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk2Betau_y, i, j - 3, k)) + vector(rk2Alpha, i, j - 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk2Betau_y, i, j - 3, k)) + vector(rk2Alpha, i, j - 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk2Betau_y, i, j - 3, k)) + vector(rk2Alpha, i, j - 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j - 2, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j - 2, k) * vector(rk2Bfu_x, i, j - 2, k) + vector(rk2gtd_yy, i, j - 2, k) * vector(rk2Bfu_y, i, j - 2, k) + vector(rk2gtd_yz, i, j - 2, k) * vector(rk2Bfu_z, i, j - 2, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j - 2, k) * vector(rk2Bfu_x, i, j - 2, k) + vector(rk2gtd_yz, i, j - 2, k) * vector(rk2Bfu_y, i, j - 2, k) + vector(rk2gtd_zz, i, j - 2, k) * vector(rk2Bfu_z, i, j - 2, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j - 2, k) * vector(rk2Bfu_x, i, j - 2, k) + vector(rk2gtd_xy, i, j - 2, k) * vector(rk2Bfu_y, i, j - 2, k) + vector(rk2gtd_xz, i, j - 2, k) * vector(rk2Bfu_z, i, j - 2, k));
						rk2sqBf = vector(rk2Bfu_x, i, j - 2, k) * rk2Bfd_x + vector(rk2Bfu_y, i, j - 2, k) * rk2Bfd_y + vector(rk2Bfu_z, i, j - 2, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j - 2, k) * vector(rk2gtd_zz, i, j - 2, k) * vector(rk2gtd_xx, i, j - 2, k) + (-vector(rk2gtd_yy, i, j - 2, k) * (vector(rk2gtd_xz, i, j - 2, k) * vector(rk2gtd_xz, i, j - 2, k))) + (-vector(rk2gtd_zz, i, j - 2, k) * (vector(rk2gtd_xy, i, j - 2, k) * vector(rk2gtd_xy, i, j - 2, k))) + (-(vector(rk2gtd_yz, i, j - 2, k) * vector(rk2gtd_yz, i, j - 2, k)) * vector(rk2gtd_xx, i, j - 2, k)) + 2.0 * vector(rk2gtd_yz, i, j - 2, k) * vector(rk2gtd_xy, i, j - 2, k) * vector(rk2gtd_xz, i, j - 2, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j - 2, k) * vector(rk2gtd_yy, i, j - 2, k) - vector(rk2gtd_xy, i, j - 2, k) * vector(rk2gtd_xy, i, j - 2, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j - 2, k) * vector(rk2gtd_zz, i, j - 2, k) - vector(rk2gtd_yz, i, j - 2, k) * vector(rk2gtd_yz, i, j - 2, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j - 2, k) * vector(rk2gtd_yz, i, j - 2, k) - vector(rk2gtd_yy, i, j - 2, k) * vector(rk2gtd_xz, i, j - 2, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j - 2, k) * vector(rk2gtd_zz, i, j - 2, k)) + vector(rk2gtd_yz, i, j - 2, k) * vector(rk2gtd_xz, i, j - 2, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j - 2, k) + rk2gtu_xy * vector(rk2Sfd_y, i, j - 2, k) + rk2gtu_xz * vector(rk2Sfd_z, i, j - 2, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j - 2, k) + rk2gtu_xy * vector(vfd_y, i, j - 2, k) + rk2gtu_xz * vector(vfd_z, i, j - 2, k));
						rk2Bfvf = vector(rk2Bfu_x, i, j - 2, k) * vector(vfd_x, i, j - 2, k) + vector(rk2Bfu_y, i, j - 2, k) * vector(vfd_y, i, j - 2, k) + vector(rk2Bfu_z, i, j - 2, k) * vector(vfd_z, i, j - 2, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j - 2, k) * vector(rk2gtd_zz, i, j - 2, k) - vector(rk2gtd_xz, i, j - 2, k) * vector(rk2gtd_xz, i, j - 2, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j - 2, k) * vector(rk2gtd_yz, i, j - 2, k)) + vector(rk2gtd_xy, i, j - 2, k) * vector(rk2gtd_xz, i, j - 2, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j - 2, k) + rk2gtu_yz * vector(vfd_y, i, j - 2, k) + rk2gtu_zz * vector(vfd_z, i, j - 2, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j - 2, k) + rk2gtu_yy * vector(rk2Sfd_y, i, j - 2, k) + rk2gtu_yz * vector(rk2Sfd_z, i, j - 2, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j - 2, k) + rk2gtu_yz * vector(rk2Sfd_y, i, j - 2, k) + rk2gtu_zz * vector(rk2Sfd_z, i, j - 2, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j - 2, k) + rk2gtu_yy * vector(vfd_y, i, j - 2, k) + rk2gtu_yz * vector(vfd_z, i, j - 2, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j - 2, k) + rk2vfu_y * vector(vfd_y, i, j - 2, k) + rk2vfu_z * vector(vfd_z, i, j - 2, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j - 2, k) * vector(rk2Bfu_z, i, j - 2, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j - 2, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j - 2, k) * vector(rk2Bfu_x, i, j - 2, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j - 2, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j - 2, k) * vector(rk2Bfu_z, i, j - 2, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j - 2, k) * rk2vfu_z + vector(rk2Bfu_z, i, j - 2, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j - 2, k) * vector(rk2Bfu_y, i, j - 2, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j - 2, k) * rk2vfu_y + vector(rk2Bfu_y, i, j - 2, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j - 2, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j - 2, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j - 2, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j - 2, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j - 2, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j - 2, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j - 2, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j - 2, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j - 2, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j - 2, k) * vector(rk2Bfu_z, i, j - 2, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j - 2, k) * rk2vfu_z + vector(rk2Bfu_z, i, j - 2, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j - 2, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j - 2, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j - 2, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j - 2, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j - 2, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j - 2, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j - 2, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j - 2, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j - 2, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j - 2, k) * vector(rk2Bfu_y, i, j - 2, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j - 2, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j - 2, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j - 2, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j - 2, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j - 2, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j - 2, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j - 2, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j - 2, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j - 2, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j - 2, k));
						Fluxm2jSfd_x = FjSfd_x_mainI(vector(rk2Sfd_x, i, j - 2, k), vector(rk2gtd_xx, i, j - 2, k), vector(rk2gtd_xy, i, j - 2, k), vector(rk2gtd_xz, i, j - 2, k), vector(rk2Betau_y, i, j - 2, k), vector(rk2Alpha, i, j - 2, k), vector(TauT_xy, i, j - 2, k), vector(TauT_yy, i, j - 2, k), vector(TauT_yz, i, j - 2, k), rk2inv_chi, rk2sdetg, rk2Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxm2jSfd_y = FjSfd_y_mainI(vector(rk2Sfd_y, i, j - 2, k), vector(rk2gtd_xy, i, j - 2, k), vector(rk2gtd_yy, i, j - 2, k), vector(rk2gtd_yz, i, j - 2, k), vector(rk2Betau_y, i, j - 2, k), vector(rk2Alpha, i, j - 2, k), vector(TauT_xy, i, j - 2, k), vector(TauT_yy, i, j - 2, k), vector(TauT_yz, i, j - 2, k), rk2inv_chi, rk2sdetg, rk2Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm2jSfd_z = FjSfd_z_mainI(vector(rk2Sfd_z, i, j - 2, k), vector(rk2gtd_xz, i, j - 2, k), vector(rk2gtd_yz, i, j - 2, k), vector(rk2gtd_zz, i, j - 2, k), vector(rk2Betau_y, i, j - 2, k), vector(rk2Alpha, i, j - 2, k), vector(TauT_xy, i, j - 2, k), vector(TauT_yy, i, j - 2, k), vector(TauT_yz, i, j - 2, k), rk2inv_chi, rk2sdetg, rk2Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxm2jBfu_x = FjBfu_x_mainI(vector(rk2Bfu_x, i, j - 2, k), vector(rk2Bfu_y, i, j - 2, k), vector(rk2phif, i, j - 2, k), vector(rk2Betau_x, i, j - 2, k), vector(rk2Betau_y, i, j - 2, k), vector(rk2Alpha, i, j - 2, k), vector(TauM_xy, i, j - 2, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2jBfu_y = FjBfu_y_mainI(vector(rk2phif, i, j - 2, k), vector(rk2Alpha, i, j - 2, k), rk2chi_max, rk2gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm2jBfu_z = FjBfu_z_mainI(vector(rk2Bfu_y, i, j - 2, k), vector(rk2Bfu_z, i, j - 2, k), vector(rk2phif, i, j - 2, k), vector(rk2Betau_y, i, j - 2, k), vector(rk2Betau_z, i, j - 2, k), vector(rk2Alpha, i, j - 2, k), vector(TauM_yz, i, j - 2, k), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2jDf = FjDf_mainI(vector(rk2Df, i, j - 2, k), vector(rk2Betau_y, i, j - 2, k), vector(rk2Alpha, i, j - 2, k), vector(TauN_y, i, j - 2, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2jDYf = FjDYf_mainI(vector(rk2DYf, i, j - 2, k), vector(rk2Betau_y, i, j - 2, k), vector(rk2Alpha, i, j - 2, k), vector(TauNe_y, i, j - 2, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2jtauf = Fjtauf_mainI(vector(rk2Df, i, j - 2, k), vector(rk2tauf, i, j - 2, k), vector(rk2Betau_y, i, j - 2, k), vector(rk2Alpha, i, j - 2, k), vector(TauN_y, i, j - 2, k), rk2sdetg, rk2Sfu_y, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2jphif = Fjphif_mainI(vector(rk2Bfu_y, i, j - 2, k), vector(rk2phif, i, j - 2, k), vector(rk2Betau_y, i, j - 2, k), vector(rk2Alpha, i, j - 2, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j - 2, k) * (1.0 + vector(epsf, i, j - 2, k)) + vector(pf, i, j - 2, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j - 2, k) + rk2sqca - vector(sqcs, i, j - 2, k) * rk2sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk2Betau_y, i, j - 2, k)) + vector(rk2Alpha, i, j - 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk2Betau_y, i, j - 2, k)) + vector(rk2Alpha, i, j - 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk2Betau_y, i, j - 2, k)) + vector(rk2Alpha, i, j - 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk2Betau_y, i, j - 2, k)) + vector(rk2Alpha, i, j - 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk2Betau_y, i, j - 2, k)) + vector(rk2Alpha, i, j - 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk2Betau_y, i, j - 2, k)) + vector(rk2Alpha, i, j - 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk2Betau_y, i, j - 2, k)) + vector(rk2Alpha, i, j - 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk2Betau_y, i, j - 2, k)) + vector(rk2Alpha, i, j - 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk2Betau_y, i, j - 2, k)) + vector(rk2Alpha, i, j - 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk2Betau_y, i, j - 2, k)) + vector(rk2Alpha, i, j - 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j - 1, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j - 1, k) * vector(rk2Bfu_x, i, j - 1, k) + vector(rk2gtd_yy, i, j - 1, k) * vector(rk2Bfu_y, i, j - 1, k) + vector(rk2gtd_yz, i, j - 1, k) * vector(rk2Bfu_z, i, j - 1, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j - 1, k) * vector(rk2Bfu_x, i, j - 1, k) + vector(rk2gtd_yz, i, j - 1, k) * vector(rk2Bfu_y, i, j - 1, k) + vector(rk2gtd_zz, i, j - 1, k) * vector(rk2Bfu_z, i, j - 1, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j - 1, k) * vector(rk2Bfu_x, i, j - 1, k) + vector(rk2gtd_xy, i, j - 1, k) * vector(rk2Bfu_y, i, j - 1, k) + vector(rk2gtd_xz, i, j - 1, k) * vector(rk2Bfu_z, i, j - 1, k));
						rk2sqBf = vector(rk2Bfu_x, i, j - 1, k) * rk2Bfd_x + vector(rk2Bfu_y, i, j - 1, k) * rk2Bfd_y + vector(rk2Bfu_z, i, j - 1, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j - 1, k) * vector(rk2gtd_zz, i, j - 1, k) * vector(rk2gtd_xx, i, j - 1, k) + (-vector(rk2gtd_yy, i, j - 1, k) * (vector(rk2gtd_xz, i, j - 1, k) * vector(rk2gtd_xz, i, j - 1, k))) + (-vector(rk2gtd_zz, i, j - 1, k) * (vector(rk2gtd_xy, i, j - 1, k) * vector(rk2gtd_xy, i, j - 1, k))) + (-(vector(rk2gtd_yz, i, j - 1, k) * vector(rk2gtd_yz, i, j - 1, k)) * vector(rk2gtd_xx, i, j - 1, k)) + 2.0 * vector(rk2gtd_yz, i, j - 1, k) * vector(rk2gtd_xy, i, j - 1, k) * vector(rk2gtd_xz, i, j - 1, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j - 1, k) * vector(rk2gtd_yy, i, j - 1, k) - vector(rk2gtd_xy, i, j - 1, k) * vector(rk2gtd_xy, i, j - 1, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j - 1, k) * vector(rk2gtd_zz, i, j - 1, k) - vector(rk2gtd_yz, i, j - 1, k) * vector(rk2gtd_yz, i, j - 1, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j - 1, k) * vector(rk2gtd_yz, i, j - 1, k) - vector(rk2gtd_yy, i, j - 1, k) * vector(rk2gtd_xz, i, j - 1, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j - 1, k) * vector(rk2gtd_zz, i, j - 1, k)) + vector(rk2gtd_yz, i, j - 1, k) * vector(rk2gtd_xz, i, j - 1, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j - 1, k) + rk2gtu_xy * vector(rk2Sfd_y, i, j - 1, k) + rk2gtu_xz * vector(rk2Sfd_z, i, j - 1, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j - 1, k) + rk2gtu_xy * vector(vfd_y, i, j - 1, k) + rk2gtu_xz * vector(vfd_z, i, j - 1, k));
						rk2Bfvf = vector(rk2Bfu_x, i, j - 1, k) * vector(vfd_x, i, j - 1, k) + vector(rk2Bfu_y, i, j - 1, k) * vector(vfd_y, i, j - 1, k) + vector(rk2Bfu_z, i, j - 1, k) * vector(vfd_z, i, j - 1, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j - 1, k) * vector(rk2gtd_zz, i, j - 1, k) - vector(rk2gtd_xz, i, j - 1, k) * vector(rk2gtd_xz, i, j - 1, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j - 1, k) * vector(rk2gtd_yz, i, j - 1, k)) + vector(rk2gtd_xy, i, j - 1, k) * vector(rk2gtd_xz, i, j - 1, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j - 1, k) + rk2gtu_yz * vector(vfd_y, i, j - 1, k) + rk2gtu_zz * vector(vfd_z, i, j - 1, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j - 1, k) + rk2gtu_yy * vector(rk2Sfd_y, i, j - 1, k) + rk2gtu_yz * vector(rk2Sfd_z, i, j - 1, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j - 1, k) + rk2gtu_yz * vector(rk2Sfd_y, i, j - 1, k) + rk2gtu_zz * vector(rk2Sfd_z, i, j - 1, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j - 1, k) + rk2gtu_yy * vector(vfd_y, i, j - 1, k) + rk2gtu_yz * vector(vfd_z, i, j - 1, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j - 1, k) + rk2vfu_y * vector(vfd_y, i, j - 1, k) + rk2vfu_z * vector(vfd_z, i, j - 1, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j - 1, k) * vector(rk2Bfu_z, i, j - 1, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j - 1, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j - 1, k) * vector(rk2Bfu_x, i, j - 1, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j - 1, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j - 1, k) * vector(rk2Bfu_z, i, j - 1, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j - 1, k) * rk2vfu_z + vector(rk2Bfu_z, i, j - 1, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j - 1, k) * vector(rk2Bfu_y, i, j - 1, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j - 1, k) * rk2vfu_y + vector(rk2Bfu_y, i, j - 1, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j - 1, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j - 1, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j - 1, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j - 1, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j - 1, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j - 1, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j - 1, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j - 1, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j - 1, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j - 1, k) * vector(rk2Bfu_z, i, j - 1, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j - 1, k) * rk2vfu_z + vector(rk2Bfu_z, i, j - 1, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j - 1, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j - 1, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j - 1, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j - 1, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j - 1, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j - 1, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j - 1, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j - 1, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j - 1, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j - 1, k) * vector(rk2Bfu_y, i, j - 1, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j - 1, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j - 1, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j - 1, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j - 1, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j - 1, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j - 1, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j - 1, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j - 1, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j - 1, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j - 1, k));
						Fluxm1jSfd_x = FjSfd_x_mainI(vector(rk2Sfd_x, i, j - 1, k), vector(rk2gtd_xx, i, j - 1, k), vector(rk2gtd_xy, i, j - 1, k), vector(rk2gtd_xz, i, j - 1, k), vector(rk2Betau_y, i, j - 1, k), vector(rk2Alpha, i, j - 1, k), vector(TauT_xy, i, j - 1, k), vector(TauT_yy, i, j - 1, k), vector(TauT_yz, i, j - 1, k), rk2inv_chi, rk2sdetg, rk2Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxm1jSfd_y = FjSfd_y_mainI(vector(rk2Sfd_y, i, j - 1, k), vector(rk2gtd_xy, i, j - 1, k), vector(rk2gtd_yy, i, j - 1, k), vector(rk2gtd_yz, i, j - 1, k), vector(rk2Betau_y, i, j - 1, k), vector(rk2Alpha, i, j - 1, k), vector(TauT_xy, i, j - 1, k), vector(TauT_yy, i, j - 1, k), vector(TauT_yz, i, j - 1, k), rk2inv_chi, rk2sdetg, rk2Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm1jSfd_z = FjSfd_z_mainI(vector(rk2Sfd_z, i, j - 1, k), vector(rk2gtd_xz, i, j - 1, k), vector(rk2gtd_yz, i, j - 1, k), vector(rk2gtd_zz, i, j - 1, k), vector(rk2Betau_y, i, j - 1, k), vector(rk2Alpha, i, j - 1, k), vector(TauT_xy, i, j - 1, k), vector(TauT_yy, i, j - 1, k), vector(TauT_yz, i, j - 1, k), rk2inv_chi, rk2sdetg, rk2Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxm1jBfu_x = FjBfu_x_mainI(vector(rk2Bfu_x, i, j - 1, k), vector(rk2Bfu_y, i, j - 1, k), vector(rk2phif, i, j - 1, k), vector(rk2Betau_x, i, j - 1, k), vector(rk2Betau_y, i, j - 1, k), vector(rk2Alpha, i, j - 1, k), vector(TauM_xy, i, j - 1, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1jBfu_y = FjBfu_y_mainI(vector(rk2phif, i, j - 1, k), vector(rk2Alpha, i, j - 1, k), rk2chi_max, rk2gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm1jBfu_z = FjBfu_z_mainI(vector(rk2Bfu_y, i, j - 1, k), vector(rk2Bfu_z, i, j - 1, k), vector(rk2phif, i, j - 1, k), vector(rk2Betau_y, i, j - 1, k), vector(rk2Betau_z, i, j - 1, k), vector(rk2Alpha, i, j - 1, k), vector(TauM_yz, i, j - 1, k), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1jDf = FjDf_mainI(vector(rk2Df, i, j - 1, k), vector(rk2Betau_y, i, j - 1, k), vector(rk2Alpha, i, j - 1, k), vector(TauN_y, i, j - 1, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1jDYf = FjDYf_mainI(vector(rk2DYf, i, j - 1, k), vector(rk2Betau_y, i, j - 1, k), vector(rk2Alpha, i, j - 1, k), vector(TauNe_y, i, j - 1, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1jtauf = Fjtauf_mainI(vector(rk2Df, i, j - 1, k), vector(rk2tauf, i, j - 1, k), vector(rk2Betau_y, i, j - 1, k), vector(rk2Alpha, i, j - 1, k), vector(TauN_y, i, j - 1, k), rk2sdetg, rk2Sfu_y, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1jphif = Fjphif_mainI(vector(rk2Bfu_y, i, j - 1, k), vector(rk2phif, i, j - 1, k), vector(rk2Betau_y, i, j - 1, k), vector(rk2Alpha, i, j - 1, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j - 1, k) * (1.0 + vector(epsf, i, j - 1, k)) + vector(pf, i, j - 1, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j - 1, k) + rk2sqca - vector(sqcs, i, j - 1, k) * rk2sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk2Betau_y, i, j - 1, k)) + vector(rk2Alpha, i, j - 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk2Betau_y, i, j - 1, k)) + vector(rk2Alpha, i, j - 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk2Betau_y, i, j - 1, k)) + vector(rk2Alpha, i, j - 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk2Betau_y, i, j - 1, k)) + vector(rk2Alpha, i, j - 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk2Betau_y, i, j - 1, k)) + vector(rk2Alpha, i, j - 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk2Betau_y, i, j - 1, k)) + vector(rk2Alpha, i, j - 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk2Betau_y, i, j - 1, k)) + vector(rk2Alpha, i, j - 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk2Betau_y, i, j - 1, k)) + vector(rk2Alpha, i, j - 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk2Betau_y, i, j - 1, k)) + vector(rk2Alpha, i, j - 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk2Betau_y, i, j - 1, k)) + vector(rk2Alpha, i, j - 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j + 1, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j + 1, k) * vector(rk2Bfu_x, i, j + 1, k) + vector(rk2gtd_yy, i, j + 1, k) * vector(rk2Bfu_y, i, j + 1, k) + vector(rk2gtd_yz, i, j + 1, k) * vector(rk2Bfu_z, i, j + 1, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j + 1, k) * vector(rk2Bfu_x, i, j + 1, k) + vector(rk2gtd_yz, i, j + 1, k) * vector(rk2Bfu_y, i, j + 1, k) + vector(rk2gtd_zz, i, j + 1, k) * vector(rk2Bfu_z, i, j + 1, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j + 1, k) * vector(rk2Bfu_x, i, j + 1, k) + vector(rk2gtd_xy, i, j + 1, k) * vector(rk2Bfu_y, i, j + 1, k) + vector(rk2gtd_xz, i, j + 1, k) * vector(rk2Bfu_z, i, j + 1, k));
						rk2sqBf = vector(rk2Bfu_x, i, j + 1, k) * rk2Bfd_x + vector(rk2Bfu_y, i, j + 1, k) * rk2Bfd_y + vector(rk2Bfu_z, i, j + 1, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j + 1, k) * vector(rk2gtd_zz, i, j + 1, k) * vector(rk2gtd_xx, i, j + 1, k) + (-vector(rk2gtd_yy, i, j + 1, k) * (vector(rk2gtd_xz, i, j + 1, k) * vector(rk2gtd_xz, i, j + 1, k))) + (-vector(rk2gtd_zz, i, j + 1, k) * (vector(rk2gtd_xy, i, j + 1, k) * vector(rk2gtd_xy, i, j + 1, k))) + (-(vector(rk2gtd_yz, i, j + 1, k) * vector(rk2gtd_yz, i, j + 1, k)) * vector(rk2gtd_xx, i, j + 1, k)) + 2.0 * vector(rk2gtd_yz, i, j + 1, k) * vector(rk2gtd_xy, i, j + 1, k) * vector(rk2gtd_xz, i, j + 1, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j + 1, k) * vector(rk2gtd_yy, i, j + 1, k) - vector(rk2gtd_xy, i, j + 1, k) * vector(rk2gtd_xy, i, j + 1, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j + 1, k) * vector(rk2gtd_zz, i, j + 1, k) - vector(rk2gtd_yz, i, j + 1, k) * vector(rk2gtd_yz, i, j + 1, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j + 1, k) * vector(rk2gtd_yz, i, j + 1, k) - vector(rk2gtd_yy, i, j + 1, k) * vector(rk2gtd_xz, i, j + 1, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j + 1, k) * vector(rk2gtd_zz, i, j + 1, k)) + vector(rk2gtd_yz, i, j + 1, k) * vector(rk2gtd_xz, i, j + 1, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j + 1, k) + rk2gtu_xy * vector(rk2Sfd_y, i, j + 1, k) + rk2gtu_xz * vector(rk2Sfd_z, i, j + 1, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j + 1, k) + rk2gtu_xy * vector(vfd_y, i, j + 1, k) + rk2gtu_xz * vector(vfd_z, i, j + 1, k));
						rk2Bfvf = vector(rk2Bfu_x, i, j + 1, k) * vector(vfd_x, i, j + 1, k) + vector(rk2Bfu_y, i, j + 1, k) * vector(vfd_y, i, j + 1, k) + vector(rk2Bfu_z, i, j + 1, k) * vector(vfd_z, i, j + 1, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j + 1, k) * vector(rk2gtd_zz, i, j + 1, k) - vector(rk2gtd_xz, i, j + 1, k) * vector(rk2gtd_xz, i, j + 1, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j + 1, k) * vector(rk2gtd_yz, i, j + 1, k)) + vector(rk2gtd_xy, i, j + 1, k) * vector(rk2gtd_xz, i, j + 1, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j + 1, k) + rk2gtu_yz * vector(vfd_y, i, j + 1, k) + rk2gtu_zz * vector(vfd_z, i, j + 1, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j + 1, k) + rk2gtu_yy * vector(rk2Sfd_y, i, j + 1, k) + rk2gtu_yz * vector(rk2Sfd_z, i, j + 1, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j + 1, k) + rk2gtu_yz * vector(rk2Sfd_y, i, j + 1, k) + rk2gtu_zz * vector(rk2Sfd_z, i, j + 1, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j + 1, k) + rk2gtu_yy * vector(vfd_y, i, j + 1, k) + rk2gtu_yz * vector(vfd_z, i, j + 1, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j + 1, k) + rk2vfu_y * vector(vfd_y, i, j + 1, k) + rk2vfu_z * vector(vfd_z, i, j + 1, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j + 1, k) * vector(rk2Bfu_z, i, j + 1, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j + 1, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j + 1, k) * vector(rk2Bfu_x, i, j + 1, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j + 1, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j + 1, k) * vector(rk2Bfu_z, i, j + 1, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j + 1, k) * rk2vfu_z + vector(rk2Bfu_z, i, j + 1, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j + 1, k) * vector(rk2Bfu_y, i, j + 1, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j + 1, k) * rk2vfu_y + vector(rk2Bfu_y, i, j + 1, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j + 1, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j + 1, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j + 1, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j + 1, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j + 1, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j + 1, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j + 1, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j + 1, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j + 1, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j + 1, k) * vector(rk2Bfu_z, i, j + 1, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j + 1, k) * rk2vfu_z + vector(rk2Bfu_z, i, j + 1, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j + 1, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j + 1, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j + 1, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j + 1, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j + 1, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j + 1, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j + 1, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j + 1, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j + 1, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j + 1, k) * vector(rk2Bfu_y, i, j + 1, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j + 1, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j + 1, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j + 1, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j + 1, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j + 1, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j + 1, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j + 1, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j + 1, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j + 1, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j + 1, k));
						Fluxp1jSfd_x = FjSfd_x_mainI(vector(rk2Sfd_x, i, j + 1, k), vector(rk2gtd_xx, i, j + 1, k), vector(rk2gtd_xy, i, j + 1, k), vector(rk2gtd_xz, i, j + 1, k), vector(rk2Betau_y, i, j + 1, k), vector(rk2Alpha, i, j + 1, k), vector(TauT_xy, i, j + 1, k), vector(TauT_yy, i, j + 1, k), vector(TauT_yz, i, j + 1, k), rk2inv_chi, rk2sdetg, rk2Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxp1jSfd_y = FjSfd_y_mainI(vector(rk2Sfd_y, i, j + 1, k), vector(rk2gtd_xy, i, j + 1, k), vector(rk2gtd_yy, i, j + 1, k), vector(rk2gtd_yz, i, j + 1, k), vector(rk2Betau_y, i, j + 1, k), vector(rk2Alpha, i, j + 1, k), vector(TauT_xy, i, j + 1, k), vector(TauT_yy, i, j + 1, k), vector(TauT_yz, i, j + 1, k), rk2inv_chi, rk2sdetg, rk2Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp1jSfd_z = FjSfd_z_mainI(vector(rk2Sfd_z, i, j + 1, k), vector(rk2gtd_xz, i, j + 1, k), vector(rk2gtd_yz, i, j + 1, k), vector(rk2gtd_zz, i, j + 1, k), vector(rk2Betau_y, i, j + 1, k), vector(rk2Alpha, i, j + 1, k), vector(TauT_xy, i, j + 1, k), vector(TauT_yy, i, j + 1, k), vector(TauT_yz, i, j + 1, k), rk2inv_chi, rk2sdetg, rk2Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxp1jBfu_x = FjBfu_x_mainI(vector(rk2Bfu_x, i, j + 1, k), vector(rk2Bfu_y, i, j + 1, k), vector(rk2phif, i, j + 1, k), vector(rk2Betau_x, i, j + 1, k), vector(rk2Betau_y, i, j + 1, k), vector(rk2Alpha, i, j + 1, k), vector(TauM_xy, i, j + 1, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1jBfu_y = FjBfu_y_mainI(vector(rk2phif, i, j + 1, k), vector(rk2Alpha, i, j + 1, k), rk2chi_max, rk2gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp1jBfu_z = FjBfu_z_mainI(vector(rk2Bfu_y, i, j + 1, k), vector(rk2Bfu_z, i, j + 1, k), vector(rk2phif, i, j + 1, k), vector(rk2Betau_y, i, j + 1, k), vector(rk2Betau_z, i, j + 1, k), vector(rk2Alpha, i, j + 1, k), vector(TauM_yz, i, j + 1, k), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1jDf = FjDf_mainI(vector(rk2Df, i, j + 1, k), vector(rk2Betau_y, i, j + 1, k), vector(rk2Alpha, i, j + 1, k), vector(TauN_y, i, j + 1, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1jDYf = FjDYf_mainI(vector(rk2DYf, i, j + 1, k), vector(rk2Betau_y, i, j + 1, k), vector(rk2Alpha, i, j + 1, k), vector(TauNe_y, i, j + 1, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1jtauf = Fjtauf_mainI(vector(rk2Df, i, j + 1, k), vector(rk2tauf, i, j + 1, k), vector(rk2Betau_y, i, j + 1, k), vector(rk2Alpha, i, j + 1, k), vector(TauN_y, i, j + 1, k), rk2sdetg, rk2Sfu_y, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1jphif = Fjphif_mainI(vector(rk2Bfu_y, i, j + 1, k), vector(rk2phif, i, j + 1, k), vector(rk2Betau_y, i, j + 1, k), vector(rk2Alpha, i, j + 1, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j + 1, k) * (1.0 + vector(epsf, i, j + 1, k)) + vector(pf, i, j + 1, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j + 1, k) + rk2sqca - vector(sqcs, i, j + 1, k) * rk2sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk2Betau_y, i, j + 1, k)) + vector(rk2Alpha, i, j + 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk2Betau_y, i, j + 1, k)) + vector(rk2Alpha, i, j + 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk2Betau_y, i, j + 1, k)) + vector(rk2Alpha, i, j + 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk2Betau_y, i, j + 1, k)) + vector(rk2Alpha, i, j + 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk2Betau_y, i, j + 1, k)) + vector(rk2Alpha, i, j + 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk2Betau_y, i, j + 1, k)) + vector(rk2Alpha, i, j + 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk2Betau_y, i, j + 1, k)) + vector(rk2Alpha, i, j + 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk2Betau_y, i, j + 1, k)) + vector(rk2Alpha, i, j + 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk2Betau_y, i, j + 1, k)) + vector(rk2Alpha, i, j + 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk2Betau_y, i, j + 1, k)) + vector(rk2Alpha, i, j + 1, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j + 2, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j + 2, k) * vector(rk2Bfu_x, i, j + 2, k) + vector(rk2gtd_yy, i, j + 2, k) * vector(rk2Bfu_y, i, j + 2, k) + vector(rk2gtd_yz, i, j + 2, k) * vector(rk2Bfu_z, i, j + 2, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j + 2, k) * vector(rk2Bfu_x, i, j + 2, k) + vector(rk2gtd_yz, i, j + 2, k) * vector(rk2Bfu_y, i, j + 2, k) + vector(rk2gtd_zz, i, j + 2, k) * vector(rk2Bfu_z, i, j + 2, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j + 2, k) * vector(rk2Bfu_x, i, j + 2, k) + vector(rk2gtd_xy, i, j + 2, k) * vector(rk2Bfu_y, i, j + 2, k) + vector(rk2gtd_xz, i, j + 2, k) * vector(rk2Bfu_z, i, j + 2, k));
						rk2sqBf = vector(rk2Bfu_x, i, j + 2, k) * rk2Bfd_x + vector(rk2Bfu_y, i, j + 2, k) * rk2Bfd_y + vector(rk2Bfu_z, i, j + 2, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j + 2, k) * vector(rk2gtd_zz, i, j + 2, k) * vector(rk2gtd_xx, i, j + 2, k) + (-vector(rk2gtd_yy, i, j + 2, k) * (vector(rk2gtd_xz, i, j + 2, k) * vector(rk2gtd_xz, i, j + 2, k))) + (-vector(rk2gtd_zz, i, j + 2, k) * (vector(rk2gtd_xy, i, j + 2, k) * vector(rk2gtd_xy, i, j + 2, k))) + (-(vector(rk2gtd_yz, i, j + 2, k) * vector(rk2gtd_yz, i, j + 2, k)) * vector(rk2gtd_xx, i, j + 2, k)) + 2.0 * vector(rk2gtd_yz, i, j + 2, k) * vector(rk2gtd_xy, i, j + 2, k) * vector(rk2gtd_xz, i, j + 2, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j + 2, k) * vector(rk2gtd_yy, i, j + 2, k) - vector(rk2gtd_xy, i, j + 2, k) * vector(rk2gtd_xy, i, j + 2, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j + 2, k) * vector(rk2gtd_zz, i, j + 2, k) - vector(rk2gtd_yz, i, j + 2, k) * vector(rk2gtd_yz, i, j + 2, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j + 2, k) * vector(rk2gtd_yz, i, j + 2, k) - vector(rk2gtd_yy, i, j + 2, k) * vector(rk2gtd_xz, i, j + 2, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j + 2, k) * vector(rk2gtd_zz, i, j + 2, k)) + vector(rk2gtd_yz, i, j + 2, k) * vector(rk2gtd_xz, i, j + 2, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j + 2, k) + rk2gtu_xy * vector(rk2Sfd_y, i, j + 2, k) + rk2gtu_xz * vector(rk2Sfd_z, i, j + 2, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j + 2, k) + rk2gtu_xy * vector(vfd_y, i, j + 2, k) + rk2gtu_xz * vector(vfd_z, i, j + 2, k));
						rk2Bfvf = vector(rk2Bfu_x, i, j + 2, k) * vector(vfd_x, i, j + 2, k) + vector(rk2Bfu_y, i, j + 2, k) * vector(vfd_y, i, j + 2, k) + vector(rk2Bfu_z, i, j + 2, k) * vector(vfd_z, i, j + 2, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j + 2, k) * vector(rk2gtd_zz, i, j + 2, k) - vector(rk2gtd_xz, i, j + 2, k) * vector(rk2gtd_xz, i, j + 2, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j + 2, k) * vector(rk2gtd_yz, i, j + 2, k)) + vector(rk2gtd_xy, i, j + 2, k) * vector(rk2gtd_xz, i, j + 2, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j + 2, k) + rk2gtu_yz * vector(vfd_y, i, j + 2, k) + rk2gtu_zz * vector(vfd_z, i, j + 2, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j + 2, k) + rk2gtu_yy * vector(rk2Sfd_y, i, j + 2, k) + rk2gtu_yz * vector(rk2Sfd_z, i, j + 2, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j + 2, k) + rk2gtu_yz * vector(rk2Sfd_y, i, j + 2, k) + rk2gtu_zz * vector(rk2Sfd_z, i, j + 2, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j + 2, k) + rk2gtu_yy * vector(vfd_y, i, j + 2, k) + rk2gtu_yz * vector(vfd_z, i, j + 2, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j + 2, k) + rk2vfu_y * vector(vfd_y, i, j + 2, k) + rk2vfu_z * vector(vfd_z, i, j + 2, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j + 2, k) * vector(rk2Bfu_z, i, j + 2, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j + 2, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j + 2, k) * vector(rk2Bfu_x, i, j + 2, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j + 2, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j + 2, k) * vector(rk2Bfu_z, i, j + 2, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j + 2, k) * rk2vfu_z + vector(rk2Bfu_z, i, j + 2, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j + 2, k) * vector(rk2Bfu_y, i, j + 2, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j + 2, k) * rk2vfu_y + vector(rk2Bfu_y, i, j + 2, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j + 2, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j + 2, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j + 2, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j + 2, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j + 2, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j + 2, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j + 2, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j + 2, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j + 2, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j + 2, k) * vector(rk2Bfu_z, i, j + 2, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j + 2, k) * rk2vfu_z + vector(rk2Bfu_z, i, j + 2, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j + 2, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j + 2, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j + 2, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j + 2, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j + 2, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j + 2, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j + 2, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j + 2, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j + 2, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j + 2, k) * vector(rk2Bfu_y, i, j + 2, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j + 2, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j + 2, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j + 2, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j + 2, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j + 2, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j + 2, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j + 2, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j + 2, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j + 2, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j + 2, k));
						Fluxp2jSfd_x = FjSfd_x_mainI(vector(rk2Sfd_x, i, j + 2, k), vector(rk2gtd_xx, i, j + 2, k), vector(rk2gtd_xy, i, j + 2, k), vector(rk2gtd_xz, i, j + 2, k), vector(rk2Betau_y, i, j + 2, k), vector(rk2Alpha, i, j + 2, k), vector(TauT_xy, i, j + 2, k), vector(TauT_yy, i, j + 2, k), vector(TauT_yz, i, j + 2, k), rk2inv_chi, rk2sdetg, rk2Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxp2jSfd_y = FjSfd_y_mainI(vector(rk2Sfd_y, i, j + 2, k), vector(rk2gtd_xy, i, j + 2, k), vector(rk2gtd_yy, i, j + 2, k), vector(rk2gtd_yz, i, j + 2, k), vector(rk2Betau_y, i, j + 2, k), vector(rk2Alpha, i, j + 2, k), vector(TauT_xy, i, j + 2, k), vector(TauT_yy, i, j + 2, k), vector(TauT_yz, i, j + 2, k), rk2inv_chi, rk2sdetg, rk2Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp2jSfd_z = FjSfd_z_mainI(vector(rk2Sfd_z, i, j + 2, k), vector(rk2gtd_xz, i, j + 2, k), vector(rk2gtd_yz, i, j + 2, k), vector(rk2gtd_zz, i, j + 2, k), vector(rk2Betau_y, i, j + 2, k), vector(rk2Alpha, i, j + 2, k), vector(TauT_xy, i, j + 2, k), vector(TauT_yy, i, j + 2, k), vector(TauT_yz, i, j + 2, k), rk2inv_chi, rk2sdetg, rk2Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxp2jBfu_x = FjBfu_x_mainI(vector(rk2Bfu_x, i, j + 2, k), vector(rk2Bfu_y, i, j + 2, k), vector(rk2phif, i, j + 2, k), vector(rk2Betau_x, i, j + 2, k), vector(rk2Betau_y, i, j + 2, k), vector(rk2Alpha, i, j + 2, k), vector(TauM_xy, i, j + 2, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2jBfu_y = FjBfu_y_mainI(vector(rk2phif, i, j + 2, k), vector(rk2Alpha, i, j + 2, k), rk2chi_max, rk2gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp2jBfu_z = FjBfu_z_mainI(vector(rk2Bfu_y, i, j + 2, k), vector(rk2Bfu_z, i, j + 2, k), vector(rk2phif, i, j + 2, k), vector(rk2Betau_y, i, j + 2, k), vector(rk2Betau_z, i, j + 2, k), vector(rk2Alpha, i, j + 2, k), vector(TauM_yz, i, j + 2, k), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2jDf = FjDf_mainI(vector(rk2Df, i, j + 2, k), vector(rk2Betau_y, i, j + 2, k), vector(rk2Alpha, i, j + 2, k), vector(TauN_y, i, j + 2, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2jDYf = FjDYf_mainI(vector(rk2DYf, i, j + 2, k), vector(rk2Betau_y, i, j + 2, k), vector(rk2Alpha, i, j + 2, k), vector(TauNe_y, i, j + 2, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2jtauf = Fjtauf_mainI(vector(rk2Df, i, j + 2, k), vector(rk2tauf, i, j + 2, k), vector(rk2Betau_y, i, j + 2, k), vector(rk2Alpha, i, j + 2, k), vector(TauN_y, i, j + 2, k), rk2sdetg, rk2Sfu_y, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2jphif = Fjphif_mainI(vector(rk2Bfu_y, i, j + 2, k), vector(rk2phif, i, j + 2, k), vector(rk2Betau_y, i, j + 2, k), vector(rk2Alpha, i, j + 2, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j + 2, k) * (1.0 + vector(epsf, i, j + 2, k)) + vector(pf, i, j + 2, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j + 2, k) + rk2sqca - vector(sqcs, i, j + 2, k) * rk2sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk2Betau_y, i, j + 2, k)) + vector(rk2Alpha, i, j + 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk2Betau_y, i, j + 2, k)) + vector(rk2Alpha, i, j + 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk2Betau_y, i, j + 2, k)) + vector(rk2Alpha, i, j + 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk2Betau_y, i, j + 2, k)) + vector(rk2Alpha, i, j + 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk2Betau_y, i, j + 2, k)) + vector(rk2Alpha, i, j + 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk2Betau_y, i, j + 2, k)) + vector(rk2Alpha, i, j + 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk2Betau_y, i, j + 2, k)) + vector(rk2Alpha, i, j + 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk2Betau_y, i, j + 2, k)) + vector(rk2Alpha, i, j + 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk2Betau_y, i, j + 2, k)) + vector(rk2Alpha, i, j + 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk2Betau_y, i, j + 2, k)) + vector(rk2Alpha, i, j + 2, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j + 3, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j + 3, k) * vector(rk2Bfu_x, i, j + 3, k) + vector(rk2gtd_yy, i, j + 3, k) * vector(rk2Bfu_y, i, j + 3, k) + vector(rk2gtd_yz, i, j + 3, k) * vector(rk2Bfu_z, i, j + 3, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j + 3, k) * vector(rk2Bfu_x, i, j + 3, k) + vector(rk2gtd_yz, i, j + 3, k) * vector(rk2Bfu_y, i, j + 3, k) + vector(rk2gtd_zz, i, j + 3, k) * vector(rk2Bfu_z, i, j + 3, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j + 3, k) * vector(rk2Bfu_x, i, j + 3, k) + vector(rk2gtd_xy, i, j + 3, k) * vector(rk2Bfu_y, i, j + 3, k) + vector(rk2gtd_xz, i, j + 3, k) * vector(rk2Bfu_z, i, j + 3, k));
						rk2sqBf = vector(rk2Bfu_x, i, j + 3, k) * rk2Bfd_x + vector(rk2Bfu_y, i, j + 3, k) * rk2Bfd_y + vector(rk2Bfu_z, i, j + 3, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j + 3, k) * vector(rk2gtd_zz, i, j + 3, k) * vector(rk2gtd_xx, i, j + 3, k) + (-vector(rk2gtd_yy, i, j + 3, k) * (vector(rk2gtd_xz, i, j + 3, k) * vector(rk2gtd_xz, i, j + 3, k))) + (-vector(rk2gtd_zz, i, j + 3, k) * (vector(rk2gtd_xy, i, j + 3, k) * vector(rk2gtd_xy, i, j + 3, k))) + (-(vector(rk2gtd_yz, i, j + 3, k) * vector(rk2gtd_yz, i, j + 3, k)) * vector(rk2gtd_xx, i, j + 3, k)) + 2.0 * vector(rk2gtd_yz, i, j + 3, k) * vector(rk2gtd_xy, i, j + 3, k) * vector(rk2gtd_xz, i, j + 3, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j + 3, k) * vector(rk2gtd_yy, i, j + 3, k) - vector(rk2gtd_xy, i, j + 3, k) * vector(rk2gtd_xy, i, j + 3, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j + 3, k) * vector(rk2gtd_zz, i, j + 3, k) - vector(rk2gtd_yz, i, j + 3, k) * vector(rk2gtd_yz, i, j + 3, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j + 3, k) * vector(rk2gtd_yz, i, j + 3, k) - vector(rk2gtd_yy, i, j + 3, k) * vector(rk2gtd_xz, i, j + 3, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j + 3, k) * vector(rk2gtd_zz, i, j + 3, k)) + vector(rk2gtd_yz, i, j + 3, k) * vector(rk2gtd_xz, i, j + 3, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j + 3, k) + rk2gtu_xy * vector(rk2Sfd_y, i, j + 3, k) + rk2gtu_xz * vector(rk2Sfd_z, i, j + 3, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j + 3, k) + rk2gtu_xy * vector(vfd_y, i, j + 3, k) + rk2gtu_xz * vector(vfd_z, i, j + 3, k));
						rk2Bfvf = vector(rk2Bfu_x, i, j + 3, k) * vector(vfd_x, i, j + 3, k) + vector(rk2Bfu_y, i, j + 3, k) * vector(vfd_y, i, j + 3, k) + vector(rk2Bfu_z, i, j + 3, k) * vector(vfd_z, i, j + 3, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j + 3, k) * vector(rk2gtd_zz, i, j + 3, k) - vector(rk2gtd_xz, i, j + 3, k) * vector(rk2gtd_xz, i, j + 3, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j + 3, k) * vector(rk2gtd_yz, i, j + 3, k)) + vector(rk2gtd_xy, i, j + 3, k) * vector(rk2gtd_xz, i, j + 3, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j + 3, k) + rk2gtu_yz * vector(vfd_y, i, j + 3, k) + rk2gtu_zz * vector(vfd_z, i, j + 3, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j + 3, k) + rk2gtu_yy * vector(rk2Sfd_y, i, j + 3, k) + rk2gtu_yz * vector(rk2Sfd_z, i, j + 3, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j + 3, k) + rk2gtu_yz * vector(rk2Sfd_y, i, j + 3, k) + rk2gtu_zz * vector(rk2Sfd_z, i, j + 3, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j + 3, k) + rk2gtu_yy * vector(vfd_y, i, j + 3, k) + rk2gtu_yz * vector(vfd_z, i, j + 3, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j + 3, k) + rk2vfu_y * vector(vfd_y, i, j + 3, k) + rk2vfu_z * vector(vfd_z, i, j + 3, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j + 3, k) * vector(rk2Bfu_z, i, j + 3, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j + 3, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j + 3, k) * vector(rk2Bfu_x, i, j + 3, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j + 3, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j + 3, k) * vector(rk2Bfu_z, i, j + 3, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j + 3, k) * rk2vfu_z + vector(rk2Bfu_z, i, j + 3, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j + 3, k) * vector(rk2Bfu_y, i, j + 3, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j + 3, k) * rk2vfu_y + vector(rk2Bfu_y, i, j + 3, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j + 3, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j + 3, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j + 3, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j + 3, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j + 3, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j + 3, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j + 3, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j + 3, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j + 3, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j + 3, k) * vector(rk2Bfu_z, i, j + 3, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j + 3, k) * rk2vfu_z + vector(rk2Bfu_z, i, j + 3, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j + 3, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j + 3, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j + 3, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j + 3, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j + 3, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j + 3, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j + 3, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j + 3, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j + 3, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j + 3, k) * vector(rk2Bfu_y, i, j + 3, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j + 3, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j + 3, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j + 3, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j + 3, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j + 3, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j + 3, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j + 3, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j + 3, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j + 3, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j + 3, k));
						Fluxp3jSfd_x = FjSfd_x_mainI(vector(rk2Sfd_x, i, j + 3, k), vector(rk2gtd_xx, i, j + 3, k), vector(rk2gtd_xy, i, j + 3, k), vector(rk2gtd_xz, i, j + 3, k), vector(rk2Betau_y, i, j + 3, k), vector(rk2Alpha, i, j + 3, k), vector(TauT_xy, i, j + 3, k), vector(TauT_yy, i, j + 3, k), vector(TauT_yz, i, j + 3, k), rk2inv_chi, rk2sdetg, rk2Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxp3jSfd_y = FjSfd_y_mainI(vector(rk2Sfd_y, i, j + 3, k), vector(rk2gtd_xy, i, j + 3, k), vector(rk2gtd_yy, i, j + 3, k), vector(rk2gtd_yz, i, j + 3, k), vector(rk2Betau_y, i, j + 3, k), vector(rk2Alpha, i, j + 3, k), vector(TauT_xy, i, j + 3, k), vector(TauT_yy, i, j + 3, k), vector(TauT_yz, i, j + 3, k), rk2inv_chi, rk2sdetg, rk2Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp3jSfd_z = FjSfd_z_mainI(vector(rk2Sfd_z, i, j + 3, k), vector(rk2gtd_xz, i, j + 3, k), vector(rk2gtd_yz, i, j + 3, k), vector(rk2gtd_zz, i, j + 3, k), vector(rk2Betau_y, i, j + 3, k), vector(rk2Alpha, i, j + 3, k), vector(TauT_xy, i, j + 3, k), vector(TauT_yy, i, j + 3, k), vector(TauT_yz, i, j + 3, k), rk2inv_chi, rk2sdetg, rk2Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxp3jBfu_x = FjBfu_x_mainI(vector(rk2Bfu_x, i, j + 3, k), vector(rk2Bfu_y, i, j + 3, k), vector(rk2phif, i, j + 3, k), vector(rk2Betau_x, i, j + 3, k), vector(rk2Betau_y, i, j + 3, k), vector(rk2Alpha, i, j + 3, k), vector(TauM_xy, i, j + 3, k), rk2chi_max, rk2gtu_xy, rk2sdetg, rk2vfu_x, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3jBfu_y = FjBfu_y_mainI(vector(rk2phif, i, j + 3, k), vector(rk2Alpha, i, j + 3, k), rk2chi_max, rk2gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp3jBfu_z = FjBfu_z_mainI(vector(rk2Bfu_y, i, j + 3, k), vector(rk2Bfu_z, i, j + 3, k), vector(rk2phif, i, j + 3, k), vector(rk2Betau_y, i, j + 3, k), vector(rk2Betau_z, i, j + 3, k), vector(rk2Alpha, i, j + 3, k), vector(TauM_yz, i, j + 3, k), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3jDf = FjDf_mainI(vector(rk2Df, i, j + 3, k), vector(rk2Betau_y, i, j + 3, k), vector(rk2Alpha, i, j + 3, k), vector(TauN_y, i, j + 3, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3jDYf = FjDYf_mainI(vector(rk2DYf, i, j + 3, k), vector(rk2Betau_y, i, j + 3, k), vector(rk2Alpha, i, j + 3, k), vector(TauNe_y, i, j + 3, k), rk2sdetg, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3jtauf = Fjtauf_mainI(vector(rk2Df, i, j + 3, k), vector(rk2tauf, i, j + 3, k), vector(rk2Betau_y, i, j + 3, k), vector(rk2Alpha, i, j + 3, k), vector(TauN_y, i, j + 3, k), rk2sdetg, rk2Sfu_y, rk2vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3jphif = Fjphif_mainI(vector(rk2Bfu_y, i, j + 3, k), vector(rk2phif, i, j + 3, k), vector(rk2Betau_y, i, j + 3, k), vector(rk2Alpha, i, j + 3, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j + 3, k) * (1.0 + vector(epsf, i, j + 3, k)) + vector(pf, i, j + 3, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j + 3, k) + rk2sqca - vector(sqcs, i, j + 3, k) * rk2sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk2Betau_y, i, j + 3, k)) + vector(rk2Alpha, i, j + 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk2Betau_y, i, j + 3, k)) + vector(rk2Alpha, i, j + 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk2Betau_y, i, j + 3, k)) + vector(rk2Alpha, i, j + 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk2Betau_y, i, j + 3, k)) + vector(rk2Alpha, i, j + 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk2Betau_y, i, j + 3, k)) + vector(rk2Alpha, i, j + 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk2Betau_y, i, j + 3, k)) + vector(rk2Alpha, i, j + 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk2Betau_y, i, j + 3, k)) + vector(rk2Alpha, i, j + 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk2Betau_y, i, j + 3, k)) + vector(rk2Alpha, i, j + 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk2Betau_y, i, j + 3, k)) + vector(rk2Alpha, i, j + 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk2Betau_y, i, j + 3, k)) + vector(rk2Alpha, i, j + 3, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_y) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_yy - (1.0 - rk2sqa) * (rk2vfu_y * rk2vfu_y)))))));
						LLFpjSfd_x = 0.5 * (FluxjSfd_x + fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j, k));
						LLFpjSfd_y = 0.5 * (FluxjSfd_y + fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j, k));
						LLFpjSfd_z = 0.5 * (FluxjSfd_z + fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j, k));
						LLFpjBfu_x = 0.5 * (FluxjBfu_x + fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j, k));
						LLFpjBfu_y = 0.5 * (FluxjBfu_y + fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j, k));
						LLFpjBfu_z = 0.5 * (FluxjBfu_z + fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j, k));
						LLFpjDf = 0.5 * (FluxjDf + fabs(SpeedjDf) * vector(rk2Df, i, j, k));
						LLFpjDYf = 0.5 * (FluxjDYf + fabs(SpeedjDYf) * vector(rk2DYf, i, j, k));
						LLFpjtauf = 0.5 * (Fluxjtauf + fabs(Speedjtauf) * vector(rk2tauf, i, j, k));
						LLFpjphif = 0.5 * (Fluxjphif + fabs(Speedjphif) * vector(rk2phif, i, j, k));
						LLFnjSfd_x = 0.5 * (FluxjSfd_x - fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j, k));
						LLFnjSfd_y = 0.5 * (FluxjSfd_y - fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j, k));
						LLFnjSfd_z = 0.5 * (FluxjSfd_z - fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j, k));
						LLFnjBfu_x = 0.5 * (FluxjBfu_x - fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j, k));
						LLFnjBfu_y = 0.5 * (FluxjBfu_y - fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j, k));
						LLFnjBfu_z = 0.5 * (FluxjBfu_z - fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j, k));
						LLFnjDf = 0.5 * (FluxjDf - fabs(SpeedjDf) * vector(rk2Df, i, j, k));
						LLFnjDYf = 0.5 * (FluxjDYf - fabs(SpeedjDYf) * vector(rk2DYf, i, j, k));
						LLFnjtauf = 0.5 * (Fluxjtauf - fabs(Speedjtauf) * vector(rk2tauf, i, j, k));
						LLFnjphif = 0.5 * (Fluxjphif - fabs(Speedjphif) * vector(rk2phif, i, j, k));
						LLFp_m3jSfd_x = 0.5 * (Fluxm3jSfd_x + fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j - 3, k));
						LLFp_m3jSfd_y = 0.5 * (Fluxm3jSfd_y + fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j - 3, k));
						LLFp_m3jSfd_z = 0.5 * (Fluxm3jSfd_z + fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j - 3, k));
						LLFp_m3jBfu_x = 0.5 * (Fluxm3jBfu_x + fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j - 3, k));
						LLFp_m3jBfu_y = 0.5 * (Fluxm3jBfu_y + fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j - 3, k));
						LLFp_m3jBfu_z = 0.5 * (Fluxm3jBfu_z + fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j - 3, k));
						LLFp_m3jDf = 0.5 * (Fluxm3jDf + fabs(SpeedjDf) * vector(rk2Df, i, j - 3, k));
						LLFp_m3jDYf = 0.5 * (Fluxm3jDYf + fabs(SpeedjDYf) * vector(rk2DYf, i, j - 3, k));
						LLFp_m3jtauf = 0.5 * (Fluxm3jtauf + fabs(Speedjtauf) * vector(rk2tauf, i, j - 3, k));
						LLFp_m3jphif = 0.5 * (Fluxm3jphif + fabs(Speedjphif) * vector(rk2phif, i, j - 3, k));
						LLFn_m3jSfd_x = 0.5 * (Fluxm3jSfd_x - fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j - 3, k));
						LLFn_m3jSfd_y = 0.5 * (Fluxm3jSfd_y - fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j - 3, k));
						LLFn_m3jSfd_z = 0.5 * (Fluxm3jSfd_z - fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j - 3, k));
						LLFn_m3jBfu_x = 0.5 * (Fluxm3jBfu_x - fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j - 3, k));
						LLFn_m3jBfu_y = 0.5 * (Fluxm3jBfu_y - fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j - 3, k));
						LLFn_m3jBfu_z = 0.5 * (Fluxm3jBfu_z - fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j - 3, k));
						LLFn_m3jDf = 0.5 * (Fluxm3jDf - fabs(SpeedjDf) * vector(rk2Df, i, j - 3, k));
						LLFn_m3jDYf = 0.5 * (Fluxm3jDYf - fabs(SpeedjDYf) * vector(rk2DYf, i, j - 3, k));
						LLFn_m3jtauf = 0.5 * (Fluxm3jtauf - fabs(Speedjtauf) * vector(rk2tauf, i, j - 3, k));
						LLFn_m3jphif = 0.5 * (Fluxm3jphif - fabs(Speedjphif) * vector(rk2phif, i, j - 3, k));
						LLFp_m2jSfd_x = 0.5 * (Fluxm2jSfd_x + fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j - 2, k));
						LLFp_m2jSfd_y = 0.5 * (Fluxm2jSfd_y + fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j - 2, k));
						LLFp_m2jSfd_z = 0.5 * (Fluxm2jSfd_z + fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j - 2, k));
						LLFp_m2jBfu_x = 0.5 * (Fluxm2jBfu_x + fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j - 2, k));
						LLFp_m2jBfu_y = 0.5 * (Fluxm2jBfu_y + fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j - 2, k));
						LLFp_m2jBfu_z = 0.5 * (Fluxm2jBfu_z + fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j - 2, k));
						LLFp_m2jDf = 0.5 * (Fluxm2jDf + fabs(SpeedjDf) * vector(rk2Df, i, j - 2, k));
						LLFp_m2jDYf = 0.5 * (Fluxm2jDYf + fabs(SpeedjDYf) * vector(rk2DYf, i, j - 2, k));
						LLFp_m2jtauf = 0.5 * (Fluxm2jtauf + fabs(Speedjtauf) * vector(rk2tauf, i, j - 2, k));
						LLFp_m2jphif = 0.5 * (Fluxm2jphif + fabs(Speedjphif) * vector(rk2phif, i, j - 2, k));
						LLFn_m2jSfd_x = 0.5 * (Fluxm2jSfd_x - fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j - 2, k));
						LLFn_m2jSfd_y = 0.5 * (Fluxm2jSfd_y - fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j - 2, k));
						LLFn_m2jSfd_z = 0.5 * (Fluxm2jSfd_z - fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j - 2, k));
						LLFn_m2jBfu_x = 0.5 * (Fluxm2jBfu_x - fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j - 2, k));
						LLFn_m2jBfu_y = 0.5 * (Fluxm2jBfu_y - fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j - 2, k));
						LLFn_m2jBfu_z = 0.5 * (Fluxm2jBfu_z - fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j - 2, k));
						LLFn_m2jDf = 0.5 * (Fluxm2jDf - fabs(SpeedjDf) * vector(rk2Df, i, j - 2, k));
						LLFn_m2jDYf = 0.5 * (Fluxm2jDYf - fabs(SpeedjDYf) * vector(rk2DYf, i, j - 2, k));
						LLFn_m2jtauf = 0.5 * (Fluxm2jtauf - fabs(Speedjtauf) * vector(rk2tauf, i, j - 2, k));
						LLFn_m2jphif = 0.5 * (Fluxm2jphif - fabs(Speedjphif) * vector(rk2phif, i, j - 2, k));
						LLFp_m1jSfd_x = 0.5 * (Fluxm1jSfd_x + fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j - 1, k));
						LLFp_m1jSfd_y = 0.5 * (Fluxm1jSfd_y + fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j - 1, k));
						LLFp_m1jSfd_z = 0.5 * (Fluxm1jSfd_z + fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j - 1, k));
						LLFp_m1jBfu_x = 0.5 * (Fluxm1jBfu_x + fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j - 1, k));
						LLFp_m1jBfu_y = 0.5 * (Fluxm1jBfu_y + fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j - 1, k));
						LLFp_m1jBfu_z = 0.5 * (Fluxm1jBfu_z + fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j - 1, k));
						LLFp_m1jDf = 0.5 * (Fluxm1jDf + fabs(SpeedjDf) * vector(rk2Df, i, j - 1, k));
						LLFp_m1jDYf = 0.5 * (Fluxm1jDYf + fabs(SpeedjDYf) * vector(rk2DYf, i, j - 1, k));
						LLFp_m1jtauf = 0.5 * (Fluxm1jtauf + fabs(Speedjtauf) * vector(rk2tauf, i, j - 1, k));
						LLFp_m1jphif = 0.5 * (Fluxm1jphif + fabs(Speedjphif) * vector(rk2phif, i, j - 1, k));
						LLFn_m1jSfd_x = 0.5 * (Fluxm1jSfd_x - fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j - 1, k));
						LLFn_m1jSfd_y = 0.5 * (Fluxm1jSfd_y - fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j - 1, k));
						LLFn_m1jSfd_z = 0.5 * (Fluxm1jSfd_z - fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j - 1, k));
						LLFn_m1jBfu_x = 0.5 * (Fluxm1jBfu_x - fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j - 1, k));
						LLFn_m1jBfu_y = 0.5 * (Fluxm1jBfu_y - fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j - 1, k));
						LLFn_m1jBfu_z = 0.5 * (Fluxm1jBfu_z - fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j - 1, k));
						LLFn_m1jDf = 0.5 * (Fluxm1jDf - fabs(SpeedjDf) * vector(rk2Df, i, j - 1, k));
						LLFn_m1jDYf = 0.5 * (Fluxm1jDYf - fabs(SpeedjDYf) * vector(rk2DYf, i, j - 1, k));
						LLFn_m1jtauf = 0.5 * (Fluxm1jtauf - fabs(Speedjtauf) * vector(rk2tauf, i, j - 1, k));
						LLFn_m1jphif = 0.5 * (Fluxm1jphif - fabs(Speedjphif) * vector(rk2phif, i, j - 1, k));
						LLFp_p1jSfd_x = 0.5 * (Fluxp1jSfd_x + fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j + 1, k));
						LLFp_p1jSfd_y = 0.5 * (Fluxp1jSfd_y + fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j + 1, k));
						LLFp_p1jSfd_z = 0.5 * (Fluxp1jSfd_z + fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j + 1, k));
						LLFp_p1jBfu_x = 0.5 * (Fluxp1jBfu_x + fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j + 1, k));
						LLFp_p1jBfu_y = 0.5 * (Fluxp1jBfu_y + fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j + 1, k));
						LLFp_p1jBfu_z = 0.5 * (Fluxp1jBfu_z + fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j + 1, k));
						LLFp_p1jDf = 0.5 * (Fluxp1jDf + fabs(SpeedjDf) * vector(rk2Df, i, j + 1, k));
						LLFp_p1jDYf = 0.5 * (Fluxp1jDYf + fabs(SpeedjDYf) * vector(rk2DYf, i, j + 1, k));
						LLFp_p1jtauf = 0.5 * (Fluxp1jtauf + fabs(Speedjtauf) * vector(rk2tauf, i, j + 1, k));
						LLFp_p1jphif = 0.5 * (Fluxp1jphif + fabs(Speedjphif) * vector(rk2phif, i, j + 1, k));
						LLFn_p1jSfd_x = 0.5 * (Fluxp1jSfd_x - fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j + 1, k));
						LLFn_p1jSfd_y = 0.5 * (Fluxp1jSfd_y - fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j + 1, k));
						LLFn_p1jSfd_z = 0.5 * (Fluxp1jSfd_z - fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j + 1, k));
						LLFn_p1jBfu_x = 0.5 * (Fluxp1jBfu_x - fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j + 1, k));
						LLFn_p1jBfu_y = 0.5 * (Fluxp1jBfu_y - fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j + 1, k));
						LLFn_p1jBfu_z = 0.5 * (Fluxp1jBfu_z - fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j + 1, k));
						LLFn_p1jDf = 0.5 * (Fluxp1jDf - fabs(SpeedjDf) * vector(rk2Df, i, j + 1, k));
						LLFn_p1jDYf = 0.5 * (Fluxp1jDYf - fabs(SpeedjDYf) * vector(rk2DYf, i, j + 1, k));
						LLFn_p1jtauf = 0.5 * (Fluxp1jtauf - fabs(Speedjtauf) * vector(rk2tauf, i, j + 1, k));
						LLFn_p1jphif = 0.5 * (Fluxp1jphif - fabs(Speedjphif) * vector(rk2phif, i, j + 1, k));
						LLFp_p2jSfd_x = 0.5 * (Fluxp2jSfd_x + fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j + 2, k));
						LLFp_p2jSfd_y = 0.5 * (Fluxp2jSfd_y + fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j + 2, k));
						LLFp_p2jSfd_z = 0.5 * (Fluxp2jSfd_z + fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j + 2, k));
						LLFp_p2jBfu_x = 0.5 * (Fluxp2jBfu_x + fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j + 2, k));
						LLFp_p2jBfu_y = 0.5 * (Fluxp2jBfu_y + fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j + 2, k));
						LLFp_p2jBfu_z = 0.5 * (Fluxp2jBfu_z + fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j + 2, k));
						LLFp_p2jDf = 0.5 * (Fluxp2jDf + fabs(SpeedjDf) * vector(rk2Df, i, j + 2, k));
						LLFp_p2jDYf = 0.5 * (Fluxp2jDYf + fabs(SpeedjDYf) * vector(rk2DYf, i, j + 2, k));
						LLFp_p2jtauf = 0.5 * (Fluxp2jtauf + fabs(Speedjtauf) * vector(rk2tauf, i, j + 2, k));
						LLFp_p2jphif = 0.5 * (Fluxp2jphif + fabs(Speedjphif) * vector(rk2phif, i, j + 2, k));
						LLFn_p2jSfd_x = 0.5 * (Fluxp2jSfd_x - fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j + 2, k));
						LLFn_p2jSfd_y = 0.5 * (Fluxp2jSfd_y - fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j + 2, k));
						LLFn_p2jSfd_z = 0.5 * (Fluxp2jSfd_z - fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j + 2, k));
						LLFn_p2jBfu_x = 0.5 * (Fluxp2jBfu_x - fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j + 2, k));
						LLFn_p2jBfu_y = 0.5 * (Fluxp2jBfu_y - fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j + 2, k));
						LLFn_p2jBfu_z = 0.5 * (Fluxp2jBfu_z - fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j + 2, k));
						LLFn_p2jDf = 0.5 * (Fluxp2jDf - fabs(SpeedjDf) * vector(rk2Df, i, j + 2, k));
						LLFn_p2jDYf = 0.5 * (Fluxp2jDYf - fabs(SpeedjDYf) * vector(rk2DYf, i, j + 2, k));
						LLFn_p2jtauf = 0.5 * (Fluxp2jtauf - fabs(Speedjtauf) * vector(rk2tauf, i, j + 2, k));
						LLFn_p2jphif = 0.5 * (Fluxp2jphif - fabs(Speedjphif) * vector(rk2phif, i, j + 2, k));
						LLFp_p3jSfd_x = 0.5 * (Fluxp3jSfd_x + fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j + 3, k));
						LLFp_p3jSfd_y = 0.5 * (Fluxp3jSfd_y + fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j + 3, k));
						LLFp_p3jSfd_z = 0.5 * (Fluxp3jSfd_z + fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j + 3, k));
						LLFp_p3jBfu_x = 0.5 * (Fluxp3jBfu_x + fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j + 3, k));
						LLFp_p3jBfu_y = 0.5 * (Fluxp3jBfu_y + fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j + 3, k));
						LLFp_p3jBfu_z = 0.5 * (Fluxp3jBfu_z + fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j + 3, k));
						LLFp_p3jDf = 0.5 * (Fluxp3jDf + fabs(SpeedjDf) * vector(rk2Df, i, j + 3, k));
						LLFp_p3jDYf = 0.5 * (Fluxp3jDYf + fabs(SpeedjDYf) * vector(rk2DYf, i, j + 3, k));
						LLFp_p3jtauf = 0.5 * (Fluxp3jtauf + fabs(Speedjtauf) * vector(rk2tauf, i, j + 3, k));
						LLFp_p3jphif = 0.5 * (Fluxp3jphif + fabs(Speedjphif) * vector(rk2phif, i, j + 3, k));
						LLFn_p3jSfd_x = 0.5 * (Fluxp3jSfd_x - fabs(SpeedjSfd_x) * vector(rk2Sfd_x, i, j + 3, k));
						LLFn_p3jSfd_y = 0.5 * (Fluxp3jSfd_y - fabs(SpeedjSfd_y) * vector(rk2Sfd_y, i, j + 3, k));
						LLFn_p3jSfd_z = 0.5 * (Fluxp3jSfd_z - fabs(SpeedjSfd_z) * vector(rk2Sfd_z, i, j + 3, k));
						LLFn_p3jBfu_x = 0.5 * (Fluxp3jBfu_x - fabs(SpeedjBfu_x) * vector(rk2Bfu_x, i, j + 3, k));
						LLFn_p3jBfu_y = 0.5 * (Fluxp3jBfu_y - fabs(SpeedjBfu_y) * vector(rk2Bfu_y, i, j + 3, k));
						LLFn_p3jBfu_z = 0.5 * (Fluxp3jBfu_z - fabs(SpeedjBfu_z) * vector(rk2Bfu_z, i, j + 3, k));
						LLFn_p3jDf = 0.5 * (Fluxp3jDf - fabs(SpeedjDf) * vector(rk2Df, i, j + 3, k));
						LLFn_p3jDYf = 0.5 * (Fluxp3jDYf - fabs(SpeedjDYf) * vector(rk2DYf, i, j + 3, k));
						LLFn_p3jtauf = 0.5 * (Fluxp3jtauf - fabs(Speedjtauf) * vector(rk2tauf, i, j + 3, k));
						LLFn_p3jphif = 0.5 * (Fluxp3jphif - fabs(Speedjphif) * vector(rk2phif, i, j + 3, k));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j, k));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_yy, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_zz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_xy, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_xz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2sqBf = vector(rk2Bfu_x, i, j, k) * rk2Bfd_x + vector(rk2Bfu_y, i, j, k) * rk2Bfd_y + vector(rk2Bfu_z, i, j, k) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) * vector(rk2gtd_xx, i, j, k) + (-vector(rk2gtd_yy, i, j, k) * (vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k))) + (-vector(rk2gtd_zz, i, j, k) * (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k))) + (-(vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k)) * vector(rk2gtd_xx, i, j, k)) + 2.0 * vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yy, i, j, k) - vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) - vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_yz, i, j, k) - vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_zz, i, j, k)) + vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j, k) + rk2gtu_xy * vector(rk2Sfd_y, i, j, k) + rk2gtu_xz * vector(rk2Sfd_z, i, j, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j, k) + rk2gtu_xy * vector(vfd_y, i, j, k) + rk2gtu_xz * vector(vfd_z, i, j, k));
						rk2Bfvf = vector(rk2Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk2Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk2Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_zz, i, j, k) - vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yz, i, j, k)) + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j, k) + rk2gtu_yz * vector(vfd_y, i, j, k) + rk2gtu_zz * vector(vfd_z, i, j, k));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j, k) + rk2gtu_yy * vector(rk2Sfd_y, i, j, k) + rk2gtu_yz * vector(rk2Sfd_z, i, j, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j, k) + rk2gtu_yz * vector(rk2Sfd_y, i, j, k) + rk2gtu_zz * vector(rk2Sfd_z, i, j, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j, k) + rk2gtu_yy * vector(vfd_y, i, j, k) + rk2gtu_yz * vector(vfd_z, i, j, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j, k) + rk2vfu_y * vector(vfd_y, i, j, k) + rk2vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j, k) * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j, k) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j, k) * vector(rk2Bfu_x, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k) * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k) * rk2vfu_z + vector(rk2Bfu_z, i, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k) * vector(rk2Bfu_y, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k) * rk2vfu_y + vector(rk2Bfu_y, i, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j, k) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j, k));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j, k) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j, k) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j, k));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k) * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k) * rk2vfu_z + vector(rk2Bfu_z, i, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j, k) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j, k));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j, k));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j, k));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j, k) * vector(rk2Bfu_y, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j, k) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j, k));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j, k) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j, k));
						FluxkSfd_x = FkSfd_x_mainI(vector(rk2Sfd_x, i, j, k), vector(rk2gtd_xx, i, j, k), vector(rk2gtd_xy, i, j, k), vector(rk2gtd_xz, i, j, k), vector(rk2Betau_z, i, j, k), vector(rk2Alpha, i, j, k), vector(TauT_xz, i, j, k), vector(TauT_yz, i, j, k), vector(TauT_zz, i, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_zx, dx, simPlat_dt, ilast, jlast);
						FluxkSfd_y = FkSfd_y_mainI(vector(rk2Sfd_y, i, j, k), vector(rk2gtd_xy, i, j, k), vector(rk2gtd_yy, i, j, k), vector(rk2gtd_yz, i, j, k), vector(rk2Betau_z, i, j, k), vector(rk2Alpha, i, j, k), vector(TauT_xz, i, j, k), vector(TauT_yz, i, j, k), vector(TauT_zz, i, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_zy, dx, simPlat_dt, ilast, jlast);
						FluxkSfd_z = FkSfd_z_mainI(vector(rk2Sfd_z, i, j, k), vector(rk2gtd_xz, i, j, k), vector(rk2gtd_yz, i, j, k), vector(rk2gtd_zz, i, j, k), vector(rk2Betau_z, i, j, k), vector(rk2Alpha, i, j, k), vector(TauT_xz, i, j, k), vector(TauT_yz, i, j, k), vector(TauT_zz, i, j, k), rk2inv_chi, rk2sdetg, rk2Sfud_zz, dx, simPlat_dt, ilast, jlast);
						FluxkBfu_x = FkBfu_x_mainI(vector(rk2Bfu_x, i, j, k), vector(rk2Bfu_z, i, j, k), vector(rk2phif, i, j, k), vector(rk2Betau_x, i, j, k), vector(rk2Betau_z, i, j, k), vector(rk2Alpha, i, j, k), vector(TauM_xz, i, j, k), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxkBfu_y = FkBfu_y_mainI(vector(rk2Bfu_y, i, j, k), vector(rk2Bfu_z, i, j, k), vector(rk2phif, i, j, k), vector(rk2Betau_y, i, j, k), vector(rk2Betau_z, i, j, k), vector(rk2Alpha, i, j, k), vector(TauM_yz, i, j, k), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxkBfu_z = FkBfu_z_mainI(vector(rk2phif, i, j, k), vector(rk2Alpha, i, j, k), rk2chi_max, rk2gtu_zz, dx, simPlat_dt, ilast, jlast);
						FluxkDf = FkDf_mainI(vector(rk2Df, i, j, k), vector(rk2Betau_z, i, j, k), vector(rk2Alpha, i, j, k), vector(TauN_z, i, j, k), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxkDYf = FkDYf_mainI(vector(rk2DYf, i, j, k), vector(rk2Betau_z, i, j, k), vector(rk2Alpha, i, j, k), vector(TauNe_z, i, j, k), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxktauf = Fktauf_mainI(vector(rk2Df, i, j, k), vector(rk2tauf, i, j, k), vector(rk2Betau_z, i, j, k), vector(rk2Alpha, i, j, k), vector(TauN_z, i, j, k), rk2sdetg, rk2Sfu_z, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxkphif = Fkphif_mainI(vector(rk2Bfu_z, i, j, k), vector(rk2phif, i, j, k), vector(rk2Betau_z, i, j, k), vector(rk2Alpha, i, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j, k) + rk2sqca - vector(sqcs, i, j, k) * rk2sqca;
						SpeedkSfd_x = fabs(fabs(vector(rk2Betau_z, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z))))));
						SpeedkSfd_y = fabs(fabs(vector(rk2Betau_z, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z))))));
						SpeedkSfd_z = fabs(fabs(vector(rk2Betau_z, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z))))));
						SpeedkBfu_x = fabs(fabs(vector(rk2Betau_z, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z))))));
						SpeedkBfu_y = fabs(fabs(vector(rk2Betau_z, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z))))));
						SpeedkBfu_z = fabs(fabs(vector(rk2Betau_z, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z))))));
						SpeedkDf = fabs(fabs(vector(rk2Betau_z, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z))))));
						SpeedkDYf = fabs(fabs(vector(rk2Betau_z, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z))))));
						Speedktauf = fabs(fabs(vector(rk2Betau_z, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z))))));
						Speedkphif = fabs(fabs(vector(rk2Betau_z, i, j, k)) + vector(rk2Alpha, i, j, k) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j, k - 3));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k - 3) * vector(rk2Bfu_x, i, j, k - 3) + vector(rk2gtd_yy, i, j, k - 3) * vector(rk2Bfu_y, i, j, k - 3) + vector(rk2gtd_yz, i, j, k - 3) * vector(rk2Bfu_z, i, j, k - 3));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k - 3) * vector(rk2Bfu_x, i, j, k - 3) + vector(rk2gtd_yz, i, j, k - 3) * vector(rk2Bfu_y, i, j, k - 3) + vector(rk2gtd_zz, i, j, k - 3) * vector(rk2Bfu_z, i, j, k - 3));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k - 3) * vector(rk2Bfu_x, i, j, k - 3) + vector(rk2gtd_xy, i, j, k - 3) * vector(rk2Bfu_y, i, j, k - 3) + vector(rk2gtd_xz, i, j, k - 3) * vector(rk2Bfu_z, i, j, k - 3));
						rk2sqBf = vector(rk2Bfu_x, i, j, k - 3) * rk2Bfd_x + vector(rk2Bfu_y, i, j, k - 3) * rk2Bfd_y + vector(rk2Bfu_z, i, j, k - 3) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j, k - 3) * vector(rk2gtd_zz, i, j, k - 3) * vector(rk2gtd_xx, i, j, k - 3) + (-vector(rk2gtd_yy, i, j, k - 3) * (vector(rk2gtd_xz, i, j, k - 3) * vector(rk2gtd_xz, i, j, k - 3))) + (-vector(rk2gtd_zz, i, j, k - 3) * (vector(rk2gtd_xy, i, j, k - 3) * vector(rk2gtd_xy, i, j, k - 3))) + (-(vector(rk2gtd_yz, i, j, k - 3) * vector(rk2gtd_yz, i, j, k - 3)) * vector(rk2gtd_xx, i, j, k - 3)) + 2.0 * vector(rk2gtd_yz, i, j, k - 3) * vector(rk2gtd_xy, i, j, k - 3) * vector(rk2gtd_xz, i, j, k - 3);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j, k - 3) * vector(rk2gtd_yy, i, j, k - 3) - vector(rk2gtd_xy, i, j, k - 3) * vector(rk2gtd_xy, i, j, k - 3));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j, k - 3) * vector(rk2gtd_zz, i, j, k - 3) - vector(rk2gtd_yz, i, j, k - 3) * vector(rk2gtd_yz, i, j, k - 3));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j, k - 3) * vector(rk2gtd_yz, i, j, k - 3) - vector(rk2gtd_yy, i, j, k - 3) * vector(rk2gtd_xz, i, j, k - 3));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j, k - 3) * vector(rk2gtd_zz, i, j, k - 3)) + vector(rk2gtd_yz, i, j, k - 3) * vector(rk2gtd_xz, i, j, k - 3));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j, k - 3) + rk2gtu_xy * vector(rk2Sfd_y, i, j, k - 3) + rk2gtu_xz * vector(rk2Sfd_z, i, j, k - 3));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j, k - 3) + rk2gtu_xy * vector(vfd_y, i, j, k - 3) + rk2gtu_xz * vector(vfd_z, i, j, k - 3));
						rk2Bfvf = vector(rk2Bfu_x, i, j, k - 3) * vector(vfd_x, i, j, k - 3) + vector(rk2Bfu_y, i, j, k - 3) * vector(vfd_y, i, j, k - 3) + vector(rk2Bfu_z, i, j, k - 3) * vector(vfd_z, i, j, k - 3);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j, k - 3) * vector(rk2gtd_zz, i, j, k - 3) - vector(rk2gtd_xz, i, j, k - 3) * vector(rk2gtd_xz, i, j, k - 3));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j, k - 3) * vector(rk2gtd_yz, i, j, k - 3)) + vector(rk2gtd_xy, i, j, k - 3) * vector(rk2gtd_xz, i, j, k - 3));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j, k - 3) + rk2gtu_yz * vector(vfd_y, i, j, k - 3) + rk2gtu_zz * vector(vfd_z, i, j, k - 3));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j, k - 3) + rk2gtu_yy * vector(rk2Sfd_y, i, j, k - 3) + rk2gtu_yz * vector(rk2Sfd_z, i, j, k - 3));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j, k - 3) + rk2gtu_yz * vector(rk2Sfd_y, i, j, k - 3) + rk2gtu_zz * vector(rk2Sfd_z, i, j, k - 3));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j, k - 3) + rk2gtu_yy * vector(vfd_y, i, j, k - 3) + rk2gtu_yz * vector(vfd_z, i, j, k - 3));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j, k - 3) + rk2vfu_y * vector(vfd_y, i, j, k - 3) + rk2vfu_z * vector(vfd_z, i, j, k - 3), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j, k - 3) * vector(rk2Bfu_z, i, j, k - 3)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j, k - 3) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j, k - 3) * vector(rk2Bfu_x, i, j, k - 3)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j, k - 3) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k - 3) * vector(rk2Bfu_z, i, j, k - 3)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k - 3) * rk2vfu_z + vector(rk2Bfu_z, i, j, k - 3) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k - 3) * vector(rk2Bfu_y, i, j, k - 3)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k - 3) * rk2vfu_y + vector(rk2Bfu_y, i, j, k - 3) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j, k - 3) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k - 3) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k - 3));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j, k - 3) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j, k - 3) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j, k - 3));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j, k - 3) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j, k - 3) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j, k - 3));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k - 3) * vector(rk2Bfu_z, i, j, k - 3)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k - 3) * rk2vfu_z + vector(rk2Bfu_z, i, j, k - 3) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j, k - 3) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j, k - 3) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j, k - 3));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k - 3) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k - 3) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j, k - 3));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j, k - 3) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j, k - 3) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j, k - 3));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j, k - 3) * vector(rk2Bfu_y, i, j, k - 3)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j, k - 3) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j, k - 3) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j, k - 3) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j, k - 3));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k - 3) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j, k - 3) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k - 3));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j, k - 3) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j, k - 3) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j, k - 3));
						Fluxm3kSfd_x = FkSfd_x_mainI(vector(rk2Sfd_x, i, j, k - 3), vector(rk2gtd_xx, i, j, k - 3), vector(rk2gtd_xy, i, j, k - 3), vector(rk2gtd_xz, i, j, k - 3), vector(rk2Betau_z, i, j, k - 3), vector(rk2Alpha, i, j, k - 3), vector(TauT_xz, i, j, k - 3), vector(TauT_yz, i, j, k - 3), vector(TauT_zz, i, j, k - 3), rk2inv_chi, rk2sdetg, rk2Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxm3kSfd_y = FkSfd_y_mainI(vector(rk2Sfd_y, i, j, k - 3), vector(rk2gtd_xy, i, j, k - 3), vector(rk2gtd_yy, i, j, k - 3), vector(rk2gtd_yz, i, j, k - 3), vector(rk2Betau_z, i, j, k - 3), vector(rk2Alpha, i, j, k - 3), vector(TauT_xz, i, j, k - 3), vector(TauT_yz, i, j, k - 3), vector(TauT_zz, i, j, k - 3), rk2inv_chi, rk2sdetg, rk2Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxm3kSfd_z = FkSfd_z_mainI(vector(rk2Sfd_z, i, j, k - 3), vector(rk2gtd_xz, i, j, k - 3), vector(rk2gtd_yz, i, j, k - 3), vector(rk2gtd_zz, i, j, k - 3), vector(rk2Betau_z, i, j, k - 3), vector(rk2Alpha, i, j, k - 3), vector(TauT_xz, i, j, k - 3), vector(TauT_yz, i, j, k - 3), vector(TauT_zz, i, j, k - 3), rk2inv_chi, rk2sdetg, rk2Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm3kBfu_x = FkBfu_x_mainI(vector(rk2Bfu_x, i, j, k - 3), vector(rk2Bfu_z, i, j, k - 3), vector(rk2phif, i, j, k - 3), vector(rk2Betau_x, i, j, k - 3), vector(rk2Betau_z, i, j, k - 3), vector(rk2Alpha, i, j, k - 3), vector(TauM_xz, i, j, k - 3), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3kBfu_y = FkBfu_y_mainI(vector(rk2Bfu_y, i, j, k - 3), vector(rk2Bfu_z, i, j, k - 3), vector(rk2phif, i, j, k - 3), vector(rk2Betau_y, i, j, k - 3), vector(rk2Betau_z, i, j, k - 3), vector(rk2Alpha, i, j, k - 3), vector(TauM_yz, i, j, k - 3), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3kBfu_z = FkBfu_z_mainI(vector(rk2phif, i, j, k - 3), vector(rk2Alpha, i, j, k - 3), rk2chi_max, rk2gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm3kDf = FkDf_mainI(vector(rk2Df, i, j, k - 3), vector(rk2Betau_z, i, j, k - 3), vector(rk2Alpha, i, j, k - 3), vector(TauN_z, i, j, k - 3), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3kDYf = FkDYf_mainI(vector(rk2DYf, i, j, k - 3), vector(rk2Betau_z, i, j, k - 3), vector(rk2Alpha, i, j, k - 3), vector(TauNe_z, i, j, k - 3), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3ktauf = Fktauf_mainI(vector(rk2Df, i, j, k - 3), vector(rk2tauf, i, j, k - 3), vector(rk2Betau_z, i, j, k - 3), vector(rk2Alpha, i, j, k - 3), vector(TauN_z, i, j, k - 3), rk2sdetg, rk2Sfu_z, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3kphif = Fkphif_mainI(vector(rk2Bfu_z, i, j, k - 3), vector(rk2phif, i, j, k - 3), vector(rk2Betau_z, i, j, k - 3), vector(rk2Alpha, i, j, k - 3), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j, k - 3) * (1.0 + vector(epsf, i, j, k - 3)) + vector(pf, i, j, k - 3);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j, k - 3) + rk2sqca - vector(sqcs, i, j, k - 3) * rk2sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk2Betau_z, i, j, k - 3)) + vector(rk2Alpha, i, j, k - 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk2Betau_z, i, j, k - 3)) + vector(rk2Alpha, i, j, k - 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk2Betau_z, i, j, k - 3)) + vector(rk2Alpha, i, j, k - 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk2Betau_z, i, j, k - 3)) + vector(rk2Alpha, i, j, k - 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk2Betau_z, i, j, k - 3)) + vector(rk2Alpha, i, j, k - 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk2Betau_z, i, j, k - 3)) + vector(rk2Alpha, i, j, k - 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk2Betau_z, i, j, k - 3)) + vector(rk2Alpha, i, j, k - 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk2Betau_z, i, j, k - 3)) + vector(rk2Alpha, i, j, k - 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk2Betau_z, i, j, k - 3)) + vector(rk2Alpha, i, j, k - 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk2Betau_z, i, j, k - 3)) + vector(rk2Alpha, i, j, k - 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j, k - 2));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k - 2) * vector(rk2Bfu_x, i, j, k - 2) + vector(rk2gtd_yy, i, j, k - 2) * vector(rk2Bfu_y, i, j, k - 2) + vector(rk2gtd_yz, i, j, k - 2) * vector(rk2Bfu_z, i, j, k - 2));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k - 2) * vector(rk2Bfu_x, i, j, k - 2) + vector(rk2gtd_yz, i, j, k - 2) * vector(rk2Bfu_y, i, j, k - 2) + vector(rk2gtd_zz, i, j, k - 2) * vector(rk2Bfu_z, i, j, k - 2));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k - 2) * vector(rk2Bfu_x, i, j, k - 2) + vector(rk2gtd_xy, i, j, k - 2) * vector(rk2Bfu_y, i, j, k - 2) + vector(rk2gtd_xz, i, j, k - 2) * vector(rk2Bfu_z, i, j, k - 2));
						rk2sqBf = vector(rk2Bfu_x, i, j, k - 2) * rk2Bfd_x + vector(rk2Bfu_y, i, j, k - 2) * rk2Bfd_y + vector(rk2Bfu_z, i, j, k - 2) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j, k - 2) * vector(rk2gtd_zz, i, j, k - 2) * vector(rk2gtd_xx, i, j, k - 2) + (-vector(rk2gtd_yy, i, j, k - 2) * (vector(rk2gtd_xz, i, j, k - 2) * vector(rk2gtd_xz, i, j, k - 2))) + (-vector(rk2gtd_zz, i, j, k - 2) * (vector(rk2gtd_xy, i, j, k - 2) * vector(rk2gtd_xy, i, j, k - 2))) + (-(vector(rk2gtd_yz, i, j, k - 2) * vector(rk2gtd_yz, i, j, k - 2)) * vector(rk2gtd_xx, i, j, k - 2)) + 2.0 * vector(rk2gtd_yz, i, j, k - 2) * vector(rk2gtd_xy, i, j, k - 2) * vector(rk2gtd_xz, i, j, k - 2);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j, k - 2) * vector(rk2gtd_yy, i, j, k - 2) - vector(rk2gtd_xy, i, j, k - 2) * vector(rk2gtd_xy, i, j, k - 2));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j, k - 2) * vector(rk2gtd_zz, i, j, k - 2) - vector(rk2gtd_yz, i, j, k - 2) * vector(rk2gtd_yz, i, j, k - 2));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j, k - 2) * vector(rk2gtd_yz, i, j, k - 2) - vector(rk2gtd_yy, i, j, k - 2) * vector(rk2gtd_xz, i, j, k - 2));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j, k - 2) * vector(rk2gtd_zz, i, j, k - 2)) + vector(rk2gtd_yz, i, j, k - 2) * vector(rk2gtd_xz, i, j, k - 2));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j, k - 2) + rk2gtu_xy * vector(rk2Sfd_y, i, j, k - 2) + rk2gtu_xz * vector(rk2Sfd_z, i, j, k - 2));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j, k - 2) + rk2gtu_xy * vector(vfd_y, i, j, k - 2) + rk2gtu_xz * vector(vfd_z, i, j, k - 2));
						rk2Bfvf = vector(rk2Bfu_x, i, j, k - 2) * vector(vfd_x, i, j, k - 2) + vector(rk2Bfu_y, i, j, k - 2) * vector(vfd_y, i, j, k - 2) + vector(rk2Bfu_z, i, j, k - 2) * vector(vfd_z, i, j, k - 2);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j, k - 2) * vector(rk2gtd_zz, i, j, k - 2) - vector(rk2gtd_xz, i, j, k - 2) * vector(rk2gtd_xz, i, j, k - 2));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j, k - 2) * vector(rk2gtd_yz, i, j, k - 2)) + vector(rk2gtd_xy, i, j, k - 2) * vector(rk2gtd_xz, i, j, k - 2));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j, k - 2) + rk2gtu_yz * vector(vfd_y, i, j, k - 2) + rk2gtu_zz * vector(vfd_z, i, j, k - 2));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j, k - 2) + rk2gtu_yy * vector(rk2Sfd_y, i, j, k - 2) + rk2gtu_yz * vector(rk2Sfd_z, i, j, k - 2));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j, k - 2) + rk2gtu_yz * vector(rk2Sfd_y, i, j, k - 2) + rk2gtu_zz * vector(rk2Sfd_z, i, j, k - 2));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j, k - 2) + rk2gtu_yy * vector(vfd_y, i, j, k - 2) + rk2gtu_yz * vector(vfd_z, i, j, k - 2));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j, k - 2) + rk2vfu_y * vector(vfd_y, i, j, k - 2) + rk2vfu_z * vector(vfd_z, i, j, k - 2), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j, k - 2) * vector(rk2Bfu_z, i, j, k - 2)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j, k - 2) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j, k - 2) * vector(rk2Bfu_x, i, j, k - 2)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j, k - 2) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k - 2) * vector(rk2Bfu_z, i, j, k - 2)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k - 2) * rk2vfu_z + vector(rk2Bfu_z, i, j, k - 2) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k - 2) * vector(rk2Bfu_y, i, j, k - 2)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k - 2) * rk2vfu_y + vector(rk2Bfu_y, i, j, k - 2) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j, k - 2) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k - 2) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k - 2));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j, k - 2) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j, k - 2) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j, k - 2));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j, k - 2) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j, k - 2) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j, k - 2));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k - 2) * vector(rk2Bfu_z, i, j, k - 2)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k - 2) * rk2vfu_z + vector(rk2Bfu_z, i, j, k - 2) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j, k - 2) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j, k - 2) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j, k - 2));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k - 2) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k - 2) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j, k - 2));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j, k - 2) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j, k - 2) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j, k - 2));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j, k - 2) * vector(rk2Bfu_y, i, j, k - 2)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j, k - 2) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j, k - 2) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j, k - 2) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j, k - 2));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k - 2) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j, k - 2) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k - 2));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j, k - 2) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j, k - 2) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j, k - 2));
						Fluxm2kSfd_x = FkSfd_x_mainI(vector(rk2Sfd_x, i, j, k - 2), vector(rk2gtd_xx, i, j, k - 2), vector(rk2gtd_xy, i, j, k - 2), vector(rk2gtd_xz, i, j, k - 2), vector(rk2Betau_z, i, j, k - 2), vector(rk2Alpha, i, j, k - 2), vector(TauT_xz, i, j, k - 2), vector(TauT_yz, i, j, k - 2), vector(TauT_zz, i, j, k - 2), rk2inv_chi, rk2sdetg, rk2Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxm2kSfd_y = FkSfd_y_mainI(vector(rk2Sfd_y, i, j, k - 2), vector(rk2gtd_xy, i, j, k - 2), vector(rk2gtd_yy, i, j, k - 2), vector(rk2gtd_yz, i, j, k - 2), vector(rk2Betau_z, i, j, k - 2), vector(rk2Alpha, i, j, k - 2), vector(TauT_xz, i, j, k - 2), vector(TauT_yz, i, j, k - 2), vector(TauT_zz, i, j, k - 2), rk2inv_chi, rk2sdetg, rk2Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxm2kSfd_z = FkSfd_z_mainI(vector(rk2Sfd_z, i, j, k - 2), vector(rk2gtd_xz, i, j, k - 2), vector(rk2gtd_yz, i, j, k - 2), vector(rk2gtd_zz, i, j, k - 2), vector(rk2Betau_z, i, j, k - 2), vector(rk2Alpha, i, j, k - 2), vector(TauT_xz, i, j, k - 2), vector(TauT_yz, i, j, k - 2), vector(TauT_zz, i, j, k - 2), rk2inv_chi, rk2sdetg, rk2Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm2kBfu_x = FkBfu_x_mainI(vector(rk2Bfu_x, i, j, k - 2), vector(rk2Bfu_z, i, j, k - 2), vector(rk2phif, i, j, k - 2), vector(rk2Betau_x, i, j, k - 2), vector(rk2Betau_z, i, j, k - 2), vector(rk2Alpha, i, j, k - 2), vector(TauM_xz, i, j, k - 2), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2kBfu_y = FkBfu_y_mainI(vector(rk2Bfu_y, i, j, k - 2), vector(rk2Bfu_z, i, j, k - 2), vector(rk2phif, i, j, k - 2), vector(rk2Betau_y, i, j, k - 2), vector(rk2Betau_z, i, j, k - 2), vector(rk2Alpha, i, j, k - 2), vector(TauM_yz, i, j, k - 2), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2kBfu_z = FkBfu_z_mainI(vector(rk2phif, i, j, k - 2), vector(rk2Alpha, i, j, k - 2), rk2chi_max, rk2gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm2kDf = FkDf_mainI(vector(rk2Df, i, j, k - 2), vector(rk2Betau_z, i, j, k - 2), vector(rk2Alpha, i, j, k - 2), vector(TauN_z, i, j, k - 2), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2kDYf = FkDYf_mainI(vector(rk2DYf, i, j, k - 2), vector(rk2Betau_z, i, j, k - 2), vector(rk2Alpha, i, j, k - 2), vector(TauNe_z, i, j, k - 2), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2ktauf = Fktauf_mainI(vector(rk2Df, i, j, k - 2), vector(rk2tauf, i, j, k - 2), vector(rk2Betau_z, i, j, k - 2), vector(rk2Alpha, i, j, k - 2), vector(TauN_z, i, j, k - 2), rk2sdetg, rk2Sfu_z, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2kphif = Fkphif_mainI(vector(rk2Bfu_z, i, j, k - 2), vector(rk2phif, i, j, k - 2), vector(rk2Betau_z, i, j, k - 2), vector(rk2Alpha, i, j, k - 2), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j, k - 2) * (1.0 + vector(epsf, i, j, k - 2)) + vector(pf, i, j, k - 2);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j, k - 2) + rk2sqca - vector(sqcs, i, j, k - 2) * rk2sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk2Betau_z, i, j, k - 2)) + vector(rk2Alpha, i, j, k - 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk2Betau_z, i, j, k - 2)) + vector(rk2Alpha, i, j, k - 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk2Betau_z, i, j, k - 2)) + vector(rk2Alpha, i, j, k - 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk2Betau_z, i, j, k - 2)) + vector(rk2Alpha, i, j, k - 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk2Betau_z, i, j, k - 2)) + vector(rk2Alpha, i, j, k - 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk2Betau_z, i, j, k - 2)) + vector(rk2Alpha, i, j, k - 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk2Betau_z, i, j, k - 2)) + vector(rk2Alpha, i, j, k - 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk2Betau_z, i, j, k - 2)) + vector(rk2Alpha, i, j, k - 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk2Betau_z, i, j, k - 2)) + vector(rk2Alpha, i, j, k - 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk2Betau_z, i, j, k - 2)) + vector(rk2Alpha, i, j, k - 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j, k - 1));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k - 1) * vector(rk2Bfu_x, i, j, k - 1) + vector(rk2gtd_yy, i, j, k - 1) * vector(rk2Bfu_y, i, j, k - 1) + vector(rk2gtd_yz, i, j, k - 1) * vector(rk2Bfu_z, i, j, k - 1));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k - 1) * vector(rk2Bfu_x, i, j, k - 1) + vector(rk2gtd_yz, i, j, k - 1) * vector(rk2Bfu_y, i, j, k - 1) + vector(rk2gtd_zz, i, j, k - 1) * vector(rk2Bfu_z, i, j, k - 1));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k - 1) * vector(rk2Bfu_x, i, j, k - 1) + vector(rk2gtd_xy, i, j, k - 1) * vector(rk2Bfu_y, i, j, k - 1) + vector(rk2gtd_xz, i, j, k - 1) * vector(rk2Bfu_z, i, j, k - 1));
						rk2sqBf = vector(rk2Bfu_x, i, j, k - 1) * rk2Bfd_x + vector(rk2Bfu_y, i, j, k - 1) * rk2Bfd_y + vector(rk2Bfu_z, i, j, k - 1) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j, k - 1) * vector(rk2gtd_zz, i, j, k - 1) * vector(rk2gtd_xx, i, j, k - 1) + (-vector(rk2gtd_yy, i, j, k - 1) * (vector(rk2gtd_xz, i, j, k - 1) * vector(rk2gtd_xz, i, j, k - 1))) + (-vector(rk2gtd_zz, i, j, k - 1) * (vector(rk2gtd_xy, i, j, k - 1) * vector(rk2gtd_xy, i, j, k - 1))) + (-(vector(rk2gtd_yz, i, j, k - 1) * vector(rk2gtd_yz, i, j, k - 1)) * vector(rk2gtd_xx, i, j, k - 1)) + 2.0 * vector(rk2gtd_yz, i, j, k - 1) * vector(rk2gtd_xy, i, j, k - 1) * vector(rk2gtd_xz, i, j, k - 1);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j, k - 1) * vector(rk2gtd_yy, i, j, k - 1) - vector(rk2gtd_xy, i, j, k - 1) * vector(rk2gtd_xy, i, j, k - 1));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j, k - 1) * vector(rk2gtd_zz, i, j, k - 1) - vector(rk2gtd_yz, i, j, k - 1) * vector(rk2gtd_yz, i, j, k - 1));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j, k - 1) * vector(rk2gtd_yz, i, j, k - 1) - vector(rk2gtd_yy, i, j, k - 1) * vector(rk2gtd_xz, i, j, k - 1));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j, k - 1) * vector(rk2gtd_zz, i, j, k - 1)) + vector(rk2gtd_yz, i, j, k - 1) * vector(rk2gtd_xz, i, j, k - 1));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j, k - 1) + rk2gtu_xy * vector(rk2Sfd_y, i, j, k - 1) + rk2gtu_xz * vector(rk2Sfd_z, i, j, k - 1));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j, k - 1) + rk2gtu_xy * vector(vfd_y, i, j, k - 1) + rk2gtu_xz * vector(vfd_z, i, j, k - 1));
						rk2Bfvf = vector(rk2Bfu_x, i, j, k - 1) * vector(vfd_x, i, j, k - 1) + vector(rk2Bfu_y, i, j, k - 1) * vector(vfd_y, i, j, k - 1) + vector(rk2Bfu_z, i, j, k - 1) * vector(vfd_z, i, j, k - 1);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j, k - 1) * vector(rk2gtd_zz, i, j, k - 1) - vector(rk2gtd_xz, i, j, k - 1) * vector(rk2gtd_xz, i, j, k - 1));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j, k - 1) * vector(rk2gtd_yz, i, j, k - 1)) + vector(rk2gtd_xy, i, j, k - 1) * vector(rk2gtd_xz, i, j, k - 1));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j, k - 1) + rk2gtu_yz * vector(vfd_y, i, j, k - 1) + rk2gtu_zz * vector(vfd_z, i, j, k - 1));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j, k - 1) + rk2gtu_yy * vector(rk2Sfd_y, i, j, k - 1) + rk2gtu_yz * vector(rk2Sfd_z, i, j, k - 1));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j, k - 1) + rk2gtu_yz * vector(rk2Sfd_y, i, j, k - 1) + rk2gtu_zz * vector(rk2Sfd_z, i, j, k - 1));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j, k - 1) + rk2gtu_yy * vector(vfd_y, i, j, k - 1) + rk2gtu_yz * vector(vfd_z, i, j, k - 1));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j, k - 1) + rk2vfu_y * vector(vfd_y, i, j, k - 1) + rk2vfu_z * vector(vfd_z, i, j, k - 1), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j, k - 1) * vector(rk2Bfu_z, i, j, k - 1)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j, k - 1) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j, k - 1) * vector(rk2Bfu_x, i, j, k - 1)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j, k - 1) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k - 1) * vector(rk2Bfu_z, i, j, k - 1)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k - 1) * rk2vfu_z + vector(rk2Bfu_z, i, j, k - 1) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k - 1) * vector(rk2Bfu_y, i, j, k - 1)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k - 1) * rk2vfu_y + vector(rk2Bfu_y, i, j, k - 1) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j, k - 1) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k - 1) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k - 1));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j, k - 1) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j, k - 1) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j, k - 1));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j, k - 1) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j, k - 1) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j, k - 1));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k - 1) * vector(rk2Bfu_z, i, j, k - 1)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k - 1) * rk2vfu_z + vector(rk2Bfu_z, i, j, k - 1) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j, k - 1) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j, k - 1) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j, k - 1));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k - 1) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k - 1) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j, k - 1));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j, k - 1) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j, k - 1) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j, k - 1));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j, k - 1) * vector(rk2Bfu_y, i, j, k - 1)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j, k - 1) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j, k - 1) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j, k - 1) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j, k - 1));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k - 1) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j, k - 1) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k - 1));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j, k - 1) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j, k - 1) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j, k - 1));
						Fluxm1kSfd_x = FkSfd_x_mainI(vector(rk2Sfd_x, i, j, k - 1), vector(rk2gtd_xx, i, j, k - 1), vector(rk2gtd_xy, i, j, k - 1), vector(rk2gtd_xz, i, j, k - 1), vector(rk2Betau_z, i, j, k - 1), vector(rk2Alpha, i, j, k - 1), vector(TauT_xz, i, j, k - 1), vector(TauT_yz, i, j, k - 1), vector(TauT_zz, i, j, k - 1), rk2inv_chi, rk2sdetg, rk2Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxm1kSfd_y = FkSfd_y_mainI(vector(rk2Sfd_y, i, j, k - 1), vector(rk2gtd_xy, i, j, k - 1), vector(rk2gtd_yy, i, j, k - 1), vector(rk2gtd_yz, i, j, k - 1), vector(rk2Betau_z, i, j, k - 1), vector(rk2Alpha, i, j, k - 1), vector(TauT_xz, i, j, k - 1), vector(TauT_yz, i, j, k - 1), vector(TauT_zz, i, j, k - 1), rk2inv_chi, rk2sdetg, rk2Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxm1kSfd_z = FkSfd_z_mainI(vector(rk2Sfd_z, i, j, k - 1), vector(rk2gtd_xz, i, j, k - 1), vector(rk2gtd_yz, i, j, k - 1), vector(rk2gtd_zz, i, j, k - 1), vector(rk2Betau_z, i, j, k - 1), vector(rk2Alpha, i, j, k - 1), vector(TauT_xz, i, j, k - 1), vector(TauT_yz, i, j, k - 1), vector(TauT_zz, i, j, k - 1), rk2inv_chi, rk2sdetg, rk2Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm1kBfu_x = FkBfu_x_mainI(vector(rk2Bfu_x, i, j, k - 1), vector(rk2Bfu_z, i, j, k - 1), vector(rk2phif, i, j, k - 1), vector(rk2Betau_x, i, j, k - 1), vector(rk2Betau_z, i, j, k - 1), vector(rk2Alpha, i, j, k - 1), vector(TauM_xz, i, j, k - 1), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1kBfu_y = FkBfu_y_mainI(vector(rk2Bfu_y, i, j, k - 1), vector(rk2Bfu_z, i, j, k - 1), vector(rk2phif, i, j, k - 1), vector(rk2Betau_y, i, j, k - 1), vector(rk2Betau_z, i, j, k - 1), vector(rk2Alpha, i, j, k - 1), vector(TauM_yz, i, j, k - 1), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1kBfu_z = FkBfu_z_mainI(vector(rk2phif, i, j, k - 1), vector(rk2Alpha, i, j, k - 1), rk2chi_max, rk2gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm1kDf = FkDf_mainI(vector(rk2Df, i, j, k - 1), vector(rk2Betau_z, i, j, k - 1), vector(rk2Alpha, i, j, k - 1), vector(TauN_z, i, j, k - 1), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1kDYf = FkDYf_mainI(vector(rk2DYf, i, j, k - 1), vector(rk2Betau_z, i, j, k - 1), vector(rk2Alpha, i, j, k - 1), vector(TauNe_z, i, j, k - 1), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1ktauf = Fktauf_mainI(vector(rk2Df, i, j, k - 1), vector(rk2tauf, i, j, k - 1), vector(rk2Betau_z, i, j, k - 1), vector(rk2Alpha, i, j, k - 1), vector(TauN_z, i, j, k - 1), rk2sdetg, rk2Sfu_z, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1kphif = Fkphif_mainI(vector(rk2Bfu_z, i, j, k - 1), vector(rk2phif, i, j, k - 1), vector(rk2Betau_z, i, j, k - 1), vector(rk2Alpha, i, j, k - 1), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j, k - 1) * (1.0 + vector(epsf, i, j, k - 1)) + vector(pf, i, j, k - 1);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j, k - 1) + rk2sqca - vector(sqcs, i, j, k - 1) * rk2sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk2Betau_z, i, j, k - 1)) + vector(rk2Alpha, i, j, k - 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk2Betau_z, i, j, k - 1)) + vector(rk2Alpha, i, j, k - 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk2Betau_z, i, j, k - 1)) + vector(rk2Alpha, i, j, k - 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk2Betau_z, i, j, k - 1)) + vector(rk2Alpha, i, j, k - 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk2Betau_z, i, j, k - 1)) + vector(rk2Alpha, i, j, k - 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk2Betau_z, i, j, k - 1)) + vector(rk2Alpha, i, j, k - 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk2Betau_z, i, j, k - 1)) + vector(rk2Alpha, i, j, k - 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk2Betau_z, i, j, k - 1)) + vector(rk2Alpha, i, j, k - 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk2Betau_z, i, j, k - 1)) + vector(rk2Alpha, i, j, k - 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk2Betau_z, i, j, k - 1)) + vector(rk2Alpha, i, j, k - 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j, k + 1));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k + 1) * vector(rk2Bfu_x, i, j, k + 1) + vector(rk2gtd_yy, i, j, k + 1) * vector(rk2Bfu_y, i, j, k + 1) + vector(rk2gtd_yz, i, j, k + 1) * vector(rk2Bfu_z, i, j, k + 1));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k + 1) * vector(rk2Bfu_x, i, j, k + 1) + vector(rk2gtd_yz, i, j, k + 1) * vector(rk2Bfu_y, i, j, k + 1) + vector(rk2gtd_zz, i, j, k + 1) * vector(rk2Bfu_z, i, j, k + 1));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k + 1) * vector(rk2Bfu_x, i, j, k + 1) + vector(rk2gtd_xy, i, j, k + 1) * vector(rk2Bfu_y, i, j, k + 1) + vector(rk2gtd_xz, i, j, k + 1) * vector(rk2Bfu_z, i, j, k + 1));
						rk2sqBf = vector(rk2Bfu_x, i, j, k + 1) * rk2Bfd_x + vector(rk2Bfu_y, i, j, k + 1) * rk2Bfd_y + vector(rk2Bfu_z, i, j, k + 1) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j, k + 1) * vector(rk2gtd_zz, i, j, k + 1) * vector(rk2gtd_xx, i, j, k + 1) + (-vector(rk2gtd_yy, i, j, k + 1) * (vector(rk2gtd_xz, i, j, k + 1) * vector(rk2gtd_xz, i, j, k + 1))) + (-vector(rk2gtd_zz, i, j, k + 1) * (vector(rk2gtd_xy, i, j, k + 1) * vector(rk2gtd_xy, i, j, k + 1))) + (-(vector(rk2gtd_yz, i, j, k + 1) * vector(rk2gtd_yz, i, j, k + 1)) * vector(rk2gtd_xx, i, j, k + 1)) + 2.0 * vector(rk2gtd_yz, i, j, k + 1) * vector(rk2gtd_xy, i, j, k + 1) * vector(rk2gtd_xz, i, j, k + 1);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j, k + 1) * vector(rk2gtd_yy, i, j, k + 1) - vector(rk2gtd_xy, i, j, k + 1) * vector(rk2gtd_xy, i, j, k + 1));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j, k + 1) * vector(rk2gtd_zz, i, j, k + 1) - vector(rk2gtd_yz, i, j, k + 1) * vector(rk2gtd_yz, i, j, k + 1));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j, k + 1) * vector(rk2gtd_yz, i, j, k + 1) - vector(rk2gtd_yy, i, j, k + 1) * vector(rk2gtd_xz, i, j, k + 1));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j, k + 1) * vector(rk2gtd_zz, i, j, k + 1)) + vector(rk2gtd_yz, i, j, k + 1) * vector(rk2gtd_xz, i, j, k + 1));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j, k + 1) + rk2gtu_xy * vector(rk2Sfd_y, i, j, k + 1) + rk2gtu_xz * vector(rk2Sfd_z, i, j, k + 1));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j, k + 1) + rk2gtu_xy * vector(vfd_y, i, j, k + 1) + rk2gtu_xz * vector(vfd_z, i, j, k + 1));
						rk2Bfvf = vector(rk2Bfu_x, i, j, k + 1) * vector(vfd_x, i, j, k + 1) + vector(rk2Bfu_y, i, j, k + 1) * vector(vfd_y, i, j, k + 1) + vector(rk2Bfu_z, i, j, k + 1) * vector(vfd_z, i, j, k + 1);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j, k + 1) * vector(rk2gtd_zz, i, j, k + 1) - vector(rk2gtd_xz, i, j, k + 1) * vector(rk2gtd_xz, i, j, k + 1));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j, k + 1) * vector(rk2gtd_yz, i, j, k + 1)) + vector(rk2gtd_xy, i, j, k + 1) * vector(rk2gtd_xz, i, j, k + 1));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j, k + 1) + rk2gtu_yz * vector(vfd_y, i, j, k + 1) + rk2gtu_zz * vector(vfd_z, i, j, k + 1));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j, k + 1) + rk2gtu_yy * vector(rk2Sfd_y, i, j, k + 1) + rk2gtu_yz * vector(rk2Sfd_z, i, j, k + 1));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j, k + 1) + rk2gtu_yz * vector(rk2Sfd_y, i, j, k + 1) + rk2gtu_zz * vector(rk2Sfd_z, i, j, k + 1));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j, k + 1) + rk2gtu_yy * vector(vfd_y, i, j, k + 1) + rk2gtu_yz * vector(vfd_z, i, j, k + 1));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j, k + 1) + rk2vfu_y * vector(vfd_y, i, j, k + 1) + rk2vfu_z * vector(vfd_z, i, j, k + 1), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j, k + 1) * vector(rk2Bfu_z, i, j, k + 1)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j, k + 1) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j, k + 1) * vector(rk2Bfu_x, i, j, k + 1)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j, k + 1) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k + 1) * vector(rk2Bfu_z, i, j, k + 1)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k + 1) * rk2vfu_z + vector(rk2Bfu_z, i, j, k + 1) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k + 1) * vector(rk2Bfu_y, i, j, k + 1)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k + 1) * rk2vfu_y + vector(rk2Bfu_y, i, j, k + 1) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j, k + 1) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k + 1) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k + 1));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j, k + 1) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j, k + 1) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j, k + 1));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j, k + 1) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j, k + 1) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j, k + 1));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k + 1) * vector(rk2Bfu_z, i, j, k + 1)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k + 1) * rk2vfu_z + vector(rk2Bfu_z, i, j, k + 1) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j, k + 1) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j, k + 1) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j, k + 1));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k + 1) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k + 1) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j, k + 1));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j, k + 1) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j, k + 1) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j, k + 1));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j, k + 1) * vector(rk2Bfu_y, i, j, k + 1)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j, k + 1) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j, k + 1) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j, k + 1) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j, k + 1));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k + 1) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j, k + 1) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k + 1));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j, k + 1) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j, k + 1) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j, k + 1));
						Fluxp1kSfd_x = FkSfd_x_mainI(vector(rk2Sfd_x, i, j, k + 1), vector(rk2gtd_xx, i, j, k + 1), vector(rk2gtd_xy, i, j, k + 1), vector(rk2gtd_xz, i, j, k + 1), vector(rk2Betau_z, i, j, k + 1), vector(rk2Alpha, i, j, k + 1), vector(TauT_xz, i, j, k + 1), vector(TauT_yz, i, j, k + 1), vector(TauT_zz, i, j, k + 1), rk2inv_chi, rk2sdetg, rk2Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxp1kSfd_y = FkSfd_y_mainI(vector(rk2Sfd_y, i, j, k + 1), vector(rk2gtd_xy, i, j, k + 1), vector(rk2gtd_yy, i, j, k + 1), vector(rk2gtd_yz, i, j, k + 1), vector(rk2Betau_z, i, j, k + 1), vector(rk2Alpha, i, j, k + 1), vector(TauT_xz, i, j, k + 1), vector(TauT_yz, i, j, k + 1), vector(TauT_zz, i, j, k + 1), rk2inv_chi, rk2sdetg, rk2Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxp1kSfd_z = FkSfd_z_mainI(vector(rk2Sfd_z, i, j, k + 1), vector(rk2gtd_xz, i, j, k + 1), vector(rk2gtd_yz, i, j, k + 1), vector(rk2gtd_zz, i, j, k + 1), vector(rk2Betau_z, i, j, k + 1), vector(rk2Alpha, i, j, k + 1), vector(TauT_xz, i, j, k + 1), vector(TauT_yz, i, j, k + 1), vector(TauT_zz, i, j, k + 1), rk2inv_chi, rk2sdetg, rk2Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp1kBfu_x = FkBfu_x_mainI(vector(rk2Bfu_x, i, j, k + 1), vector(rk2Bfu_z, i, j, k + 1), vector(rk2phif, i, j, k + 1), vector(rk2Betau_x, i, j, k + 1), vector(rk2Betau_z, i, j, k + 1), vector(rk2Alpha, i, j, k + 1), vector(TauM_xz, i, j, k + 1), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1kBfu_y = FkBfu_y_mainI(vector(rk2Bfu_y, i, j, k + 1), vector(rk2Bfu_z, i, j, k + 1), vector(rk2phif, i, j, k + 1), vector(rk2Betau_y, i, j, k + 1), vector(rk2Betau_z, i, j, k + 1), vector(rk2Alpha, i, j, k + 1), vector(TauM_yz, i, j, k + 1), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1kBfu_z = FkBfu_z_mainI(vector(rk2phif, i, j, k + 1), vector(rk2Alpha, i, j, k + 1), rk2chi_max, rk2gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp1kDf = FkDf_mainI(vector(rk2Df, i, j, k + 1), vector(rk2Betau_z, i, j, k + 1), vector(rk2Alpha, i, j, k + 1), vector(TauN_z, i, j, k + 1), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1kDYf = FkDYf_mainI(vector(rk2DYf, i, j, k + 1), vector(rk2Betau_z, i, j, k + 1), vector(rk2Alpha, i, j, k + 1), vector(TauNe_z, i, j, k + 1), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1ktauf = Fktauf_mainI(vector(rk2Df, i, j, k + 1), vector(rk2tauf, i, j, k + 1), vector(rk2Betau_z, i, j, k + 1), vector(rk2Alpha, i, j, k + 1), vector(TauN_z, i, j, k + 1), rk2sdetg, rk2Sfu_z, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1kphif = Fkphif_mainI(vector(rk2Bfu_z, i, j, k + 1), vector(rk2phif, i, j, k + 1), vector(rk2Betau_z, i, j, k + 1), vector(rk2Alpha, i, j, k + 1), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j, k + 1) * (1.0 + vector(epsf, i, j, k + 1)) + vector(pf, i, j, k + 1);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j, k + 1) + rk2sqca - vector(sqcs, i, j, k + 1) * rk2sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk2Betau_z, i, j, k + 1)) + vector(rk2Alpha, i, j, k + 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk2Betau_z, i, j, k + 1)) + vector(rk2Alpha, i, j, k + 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk2Betau_z, i, j, k + 1)) + vector(rk2Alpha, i, j, k + 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk2Betau_z, i, j, k + 1)) + vector(rk2Alpha, i, j, k + 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk2Betau_z, i, j, k + 1)) + vector(rk2Alpha, i, j, k + 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk2Betau_z, i, j, k + 1)) + vector(rk2Alpha, i, j, k + 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk2Betau_z, i, j, k + 1)) + vector(rk2Alpha, i, j, k + 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk2Betau_z, i, j, k + 1)) + vector(rk2Alpha, i, j, k + 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk2Betau_z, i, j, k + 1)) + vector(rk2Alpha, i, j, k + 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk2Betau_z, i, j, k + 1)) + vector(rk2Alpha, i, j, k + 1) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j, k + 2));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k + 2) * vector(rk2Bfu_x, i, j, k + 2) + vector(rk2gtd_yy, i, j, k + 2) * vector(rk2Bfu_y, i, j, k + 2) + vector(rk2gtd_yz, i, j, k + 2) * vector(rk2Bfu_z, i, j, k + 2));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k + 2) * vector(rk2Bfu_x, i, j, k + 2) + vector(rk2gtd_yz, i, j, k + 2) * vector(rk2Bfu_y, i, j, k + 2) + vector(rk2gtd_zz, i, j, k + 2) * vector(rk2Bfu_z, i, j, k + 2));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k + 2) * vector(rk2Bfu_x, i, j, k + 2) + vector(rk2gtd_xy, i, j, k + 2) * vector(rk2Bfu_y, i, j, k + 2) + vector(rk2gtd_xz, i, j, k + 2) * vector(rk2Bfu_z, i, j, k + 2));
						rk2sqBf = vector(rk2Bfu_x, i, j, k + 2) * rk2Bfd_x + vector(rk2Bfu_y, i, j, k + 2) * rk2Bfd_y + vector(rk2Bfu_z, i, j, k + 2) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j, k + 2) * vector(rk2gtd_zz, i, j, k + 2) * vector(rk2gtd_xx, i, j, k + 2) + (-vector(rk2gtd_yy, i, j, k + 2) * (vector(rk2gtd_xz, i, j, k + 2) * vector(rk2gtd_xz, i, j, k + 2))) + (-vector(rk2gtd_zz, i, j, k + 2) * (vector(rk2gtd_xy, i, j, k + 2) * vector(rk2gtd_xy, i, j, k + 2))) + (-(vector(rk2gtd_yz, i, j, k + 2) * vector(rk2gtd_yz, i, j, k + 2)) * vector(rk2gtd_xx, i, j, k + 2)) + 2.0 * vector(rk2gtd_yz, i, j, k + 2) * vector(rk2gtd_xy, i, j, k + 2) * vector(rk2gtd_xz, i, j, k + 2);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j, k + 2) * vector(rk2gtd_yy, i, j, k + 2) - vector(rk2gtd_xy, i, j, k + 2) * vector(rk2gtd_xy, i, j, k + 2));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j, k + 2) * vector(rk2gtd_zz, i, j, k + 2) - vector(rk2gtd_yz, i, j, k + 2) * vector(rk2gtd_yz, i, j, k + 2));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j, k + 2) * vector(rk2gtd_yz, i, j, k + 2) - vector(rk2gtd_yy, i, j, k + 2) * vector(rk2gtd_xz, i, j, k + 2));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j, k + 2) * vector(rk2gtd_zz, i, j, k + 2)) + vector(rk2gtd_yz, i, j, k + 2) * vector(rk2gtd_xz, i, j, k + 2));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j, k + 2) + rk2gtu_xy * vector(rk2Sfd_y, i, j, k + 2) + rk2gtu_xz * vector(rk2Sfd_z, i, j, k + 2));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j, k + 2) + rk2gtu_xy * vector(vfd_y, i, j, k + 2) + rk2gtu_xz * vector(vfd_z, i, j, k + 2));
						rk2Bfvf = vector(rk2Bfu_x, i, j, k + 2) * vector(vfd_x, i, j, k + 2) + vector(rk2Bfu_y, i, j, k + 2) * vector(vfd_y, i, j, k + 2) + vector(rk2Bfu_z, i, j, k + 2) * vector(vfd_z, i, j, k + 2);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j, k + 2) * vector(rk2gtd_zz, i, j, k + 2) - vector(rk2gtd_xz, i, j, k + 2) * vector(rk2gtd_xz, i, j, k + 2));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j, k + 2) * vector(rk2gtd_yz, i, j, k + 2)) + vector(rk2gtd_xy, i, j, k + 2) * vector(rk2gtd_xz, i, j, k + 2));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j, k + 2) + rk2gtu_yz * vector(vfd_y, i, j, k + 2) + rk2gtu_zz * vector(vfd_z, i, j, k + 2));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j, k + 2) + rk2gtu_yy * vector(rk2Sfd_y, i, j, k + 2) + rk2gtu_yz * vector(rk2Sfd_z, i, j, k + 2));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j, k + 2) + rk2gtu_yz * vector(rk2Sfd_y, i, j, k + 2) + rk2gtu_zz * vector(rk2Sfd_z, i, j, k + 2));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j, k + 2) + rk2gtu_yy * vector(vfd_y, i, j, k + 2) + rk2gtu_yz * vector(vfd_z, i, j, k + 2));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j, k + 2) + rk2vfu_y * vector(vfd_y, i, j, k + 2) + rk2vfu_z * vector(vfd_z, i, j, k + 2), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j, k + 2) * vector(rk2Bfu_z, i, j, k + 2)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j, k + 2) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j, k + 2) * vector(rk2Bfu_x, i, j, k + 2)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j, k + 2) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k + 2) * vector(rk2Bfu_z, i, j, k + 2)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k + 2) * rk2vfu_z + vector(rk2Bfu_z, i, j, k + 2) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k + 2) * vector(rk2Bfu_y, i, j, k + 2)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k + 2) * rk2vfu_y + vector(rk2Bfu_y, i, j, k + 2) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j, k + 2) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k + 2) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k + 2));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j, k + 2) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j, k + 2) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j, k + 2));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j, k + 2) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j, k + 2) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j, k + 2));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k + 2) * vector(rk2Bfu_z, i, j, k + 2)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k + 2) * rk2vfu_z + vector(rk2Bfu_z, i, j, k + 2) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j, k + 2) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j, k + 2) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j, k + 2));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k + 2) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k + 2) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j, k + 2));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j, k + 2) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j, k + 2) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j, k + 2));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j, k + 2) * vector(rk2Bfu_y, i, j, k + 2)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j, k + 2) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j, k + 2) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j, k + 2) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j, k + 2));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k + 2) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j, k + 2) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k + 2));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j, k + 2) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j, k + 2) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j, k + 2));
						Fluxp2kSfd_x = FkSfd_x_mainI(vector(rk2Sfd_x, i, j, k + 2), vector(rk2gtd_xx, i, j, k + 2), vector(rk2gtd_xy, i, j, k + 2), vector(rk2gtd_xz, i, j, k + 2), vector(rk2Betau_z, i, j, k + 2), vector(rk2Alpha, i, j, k + 2), vector(TauT_xz, i, j, k + 2), vector(TauT_yz, i, j, k + 2), vector(TauT_zz, i, j, k + 2), rk2inv_chi, rk2sdetg, rk2Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxp2kSfd_y = FkSfd_y_mainI(vector(rk2Sfd_y, i, j, k + 2), vector(rk2gtd_xy, i, j, k + 2), vector(rk2gtd_yy, i, j, k + 2), vector(rk2gtd_yz, i, j, k + 2), vector(rk2Betau_z, i, j, k + 2), vector(rk2Alpha, i, j, k + 2), vector(TauT_xz, i, j, k + 2), vector(TauT_yz, i, j, k + 2), vector(TauT_zz, i, j, k + 2), rk2inv_chi, rk2sdetg, rk2Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxp2kSfd_z = FkSfd_z_mainI(vector(rk2Sfd_z, i, j, k + 2), vector(rk2gtd_xz, i, j, k + 2), vector(rk2gtd_yz, i, j, k + 2), vector(rk2gtd_zz, i, j, k + 2), vector(rk2Betau_z, i, j, k + 2), vector(rk2Alpha, i, j, k + 2), vector(TauT_xz, i, j, k + 2), vector(TauT_yz, i, j, k + 2), vector(TauT_zz, i, j, k + 2), rk2inv_chi, rk2sdetg, rk2Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp2kBfu_x = FkBfu_x_mainI(vector(rk2Bfu_x, i, j, k + 2), vector(rk2Bfu_z, i, j, k + 2), vector(rk2phif, i, j, k + 2), vector(rk2Betau_x, i, j, k + 2), vector(rk2Betau_z, i, j, k + 2), vector(rk2Alpha, i, j, k + 2), vector(TauM_xz, i, j, k + 2), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2kBfu_y = FkBfu_y_mainI(vector(rk2Bfu_y, i, j, k + 2), vector(rk2Bfu_z, i, j, k + 2), vector(rk2phif, i, j, k + 2), vector(rk2Betau_y, i, j, k + 2), vector(rk2Betau_z, i, j, k + 2), vector(rk2Alpha, i, j, k + 2), vector(TauM_yz, i, j, k + 2), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2kBfu_z = FkBfu_z_mainI(vector(rk2phif, i, j, k + 2), vector(rk2Alpha, i, j, k + 2), rk2chi_max, rk2gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp2kDf = FkDf_mainI(vector(rk2Df, i, j, k + 2), vector(rk2Betau_z, i, j, k + 2), vector(rk2Alpha, i, j, k + 2), vector(TauN_z, i, j, k + 2), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2kDYf = FkDYf_mainI(vector(rk2DYf, i, j, k + 2), vector(rk2Betau_z, i, j, k + 2), vector(rk2Alpha, i, j, k + 2), vector(TauNe_z, i, j, k + 2), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2ktauf = Fktauf_mainI(vector(rk2Df, i, j, k + 2), vector(rk2tauf, i, j, k + 2), vector(rk2Betau_z, i, j, k + 2), vector(rk2Alpha, i, j, k + 2), vector(TauN_z, i, j, k + 2), rk2sdetg, rk2Sfu_z, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2kphif = Fkphif_mainI(vector(rk2Bfu_z, i, j, k + 2), vector(rk2phif, i, j, k + 2), vector(rk2Betau_z, i, j, k + 2), vector(rk2Alpha, i, j, k + 2), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j, k + 2) * (1.0 + vector(epsf, i, j, k + 2)) + vector(pf, i, j, k + 2);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j, k + 2) + rk2sqca - vector(sqcs, i, j, k + 2) * rk2sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk2Betau_z, i, j, k + 2)) + vector(rk2Alpha, i, j, k + 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk2Betau_z, i, j, k + 2)) + vector(rk2Alpha, i, j, k + 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk2Betau_z, i, j, k + 2)) + vector(rk2Alpha, i, j, k + 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk2Betau_z, i, j, k + 2)) + vector(rk2Alpha, i, j, k + 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk2Betau_z, i, j, k + 2)) + vector(rk2Alpha, i, j, k + 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk2Betau_z, i, j, k + 2)) + vector(rk2Alpha, i, j, k + 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk2Betau_z, i, j, k + 2)) + vector(rk2Alpha, i, j, k + 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk2Betau_z, i, j, k + 2)) + vector(rk2Alpha, i, j, k + 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk2Betau_z, i, j, k + 2)) + vector(rk2Alpha, i, j, k + 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk2Betau_z, i, j, k + 2)) + vector(rk2Alpha, i, j, k + 2) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j, k + 3));
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k + 3) * vector(rk2Bfu_x, i, j, k + 3) + vector(rk2gtd_yy, i, j, k + 3) * vector(rk2Bfu_y, i, j, k + 3) + vector(rk2gtd_yz, i, j, k + 3) * vector(rk2Bfu_z, i, j, k + 3));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k + 3) * vector(rk2Bfu_x, i, j, k + 3) + vector(rk2gtd_yz, i, j, k + 3) * vector(rk2Bfu_y, i, j, k + 3) + vector(rk2gtd_zz, i, j, k + 3) * vector(rk2Bfu_z, i, j, k + 3));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k + 3) * vector(rk2Bfu_x, i, j, k + 3) + vector(rk2gtd_xy, i, j, k + 3) * vector(rk2Bfu_y, i, j, k + 3) + vector(rk2gtd_xz, i, j, k + 3) * vector(rk2Bfu_z, i, j, k + 3));
						rk2sqBf = vector(rk2Bfu_x, i, j, k + 3) * rk2Bfd_x + vector(rk2Bfu_y, i, j, k + 3) * rk2Bfd_y + vector(rk2Bfu_z, i, j, k + 3) * rk2Bfd_z;
						rk2detgtd = vector(rk2gtd_yy, i, j, k + 3) * vector(rk2gtd_zz, i, j, k + 3) * vector(rk2gtd_xx, i, j, k + 3) + (-vector(rk2gtd_yy, i, j, k + 3) * (vector(rk2gtd_xz, i, j, k + 3) * vector(rk2gtd_xz, i, j, k + 3))) + (-vector(rk2gtd_zz, i, j, k + 3) * (vector(rk2gtd_xy, i, j, k + 3) * vector(rk2gtd_xy, i, j, k + 3))) + (-(vector(rk2gtd_yz, i, j, k + 3) * vector(rk2gtd_yz, i, j, k + 3)) * vector(rk2gtd_xx, i, j, k + 3)) + 2.0 * vector(rk2gtd_yz, i, j, k + 3) * vector(rk2gtd_xy, i, j, k + 3) * vector(rk2gtd_xz, i, j, k + 3);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j, k + 3) * vector(rk2gtd_yy, i, j, k + 3) - vector(rk2gtd_xy, i, j, k + 3) * vector(rk2gtd_xy, i, j, k + 3));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j, k + 3) * vector(rk2gtd_zz, i, j, k + 3) - vector(rk2gtd_yz, i, j, k + 3) * vector(rk2gtd_yz, i, j, k + 3));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j, k + 3) * vector(rk2gtd_yz, i, j, k + 3) - vector(rk2gtd_yy, i, j, k + 3) * vector(rk2gtd_xz, i, j, k + 3));
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j, k + 3) * vector(rk2gtd_zz, i, j, k + 3)) + vector(rk2gtd_yz, i, j, k + 3) * vector(rk2gtd_xz, i, j, k + 3));
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j, k + 3) + rk2gtu_xy * vector(rk2Sfd_y, i, j, k + 3) + rk2gtu_xz * vector(rk2Sfd_z, i, j, k + 3));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j, k + 3) + rk2gtu_xy * vector(vfd_y, i, j, k + 3) + rk2gtu_xz * vector(vfd_z, i, j, k + 3));
						rk2Bfvf = vector(rk2Bfu_x, i, j, k + 3) * vector(vfd_x, i, j, k + 3) + vector(rk2Bfu_y, i, j, k + 3) * vector(vfd_y, i, j, k + 3) + vector(rk2Bfu_z, i, j, k + 3) * vector(vfd_z, i, j, k + 3);
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j, k + 3) * vector(rk2gtd_zz, i, j, k + 3) - vector(rk2gtd_xz, i, j, k + 3) * vector(rk2gtd_xz, i, j, k + 3));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j, k + 3) * vector(rk2gtd_yz, i, j, k + 3)) + vector(rk2gtd_xy, i, j, k + 3) * vector(rk2gtd_xz, i, j, k + 3));
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j, k + 3) + rk2gtu_yz * vector(vfd_y, i, j, k + 3) + rk2gtu_zz * vector(vfd_z, i, j, k + 3));
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j, k + 3) + rk2gtu_yy * vector(rk2Sfd_y, i, j, k + 3) + rk2gtu_yz * vector(rk2Sfd_z, i, j, k + 3));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j, k + 3) + rk2gtu_yz * vector(rk2Sfd_y, i, j, k + 3) + rk2gtu_zz * vector(rk2Sfd_z, i, j, k + 3));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j, k + 3) + rk2gtu_yy * vector(vfd_y, i, j, k + 3) + rk2gtu_yz * vector(vfd_z, i, j, k + 3));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j, k + 3) + rk2vfu_y * vector(vfd_y, i, j, k + 3) + rk2vfu_z * vector(vfd_z, i, j, k + 3), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j, k + 3) * vector(rk2Bfu_z, i, j, k + 3)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j, k + 3) * rk2vfu_z)) / rk2sdetg);
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j, k + 3) * vector(rk2Bfu_x, i, j, k + 3)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j, k + 3) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k + 3) * vector(rk2Bfu_z, i, j, k + 3)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k + 3) * rk2vfu_z + vector(rk2Bfu_z, i, j, k + 3) * rk2vfu_x)) / rk2sdetg);
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k + 3) * vector(rk2Bfu_y, i, j, k + 3)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k + 3) * rk2vfu_y + vector(rk2Bfu_y, i, j, k + 3) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j, k + 3) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k + 3) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k + 3));
						rk2Sfud_xy = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xy, i, j, k + 3) + rk2Sfuu_xy * vector(rk2gtd_yy, i, j, k + 3) + rk2Sfuu_xz * vector(rk2gtd_yz, i, j, k + 3));
						rk2Sfud_xz = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xz, i, j, k + 3) + rk2Sfuu_xy * vector(rk2gtd_yz, i, j, k + 3) + rk2Sfuu_xz * vector(rk2gtd_zz, i, j, k + 3));
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k + 3) * vector(rk2Bfu_z, i, j, k + 3)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k + 3) * rk2vfu_z + vector(rk2Bfu_z, i, j, k + 3) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_zy = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xy, i, j, k + 3) + rk2Sfuu_yz * vector(rk2gtd_yy, i, j, k + 3) + rk2Sfuu_zz * vector(rk2gtd_yz, i, j, k + 3));
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k + 3) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k + 3) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j, k + 3));
						rk2Sfud_zx = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xx, i, j, k + 3) + rk2Sfuu_yz * vector(rk2gtd_xy, i, j, k + 3) + rk2Sfuu_zz * vector(rk2gtd_xz, i, j, k + 3));
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j, k + 3) * vector(rk2Bfu_y, i, j, k + 3)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j, k + 3) * rk2vfu_y)) / rk2sdetg);
						rk2Sfud_yx = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xx, i, j, k + 3) + rk2Sfuu_yy * vector(rk2gtd_xy, i, j, k + 3) + rk2Sfuu_yz * vector(rk2gtd_xz, i, j, k + 3));
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k + 3) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j, k + 3) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k + 3));
						rk2Sfud_yz = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xz, i, j, k + 3) + rk2Sfuu_yy * vector(rk2gtd_yz, i, j, k + 3) + rk2Sfuu_yz * vector(rk2gtd_zz, i, j, k + 3));
						Fluxp3kSfd_x = FkSfd_x_mainI(vector(rk2Sfd_x, i, j, k + 3), vector(rk2gtd_xx, i, j, k + 3), vector(rk2gtd_xy, i, j, k + 3), vector(rk2gtd_xz, i, j, k + 3), vector(rk2Betau_z, i, j, k + 3), vector(rk2Alpha, i, j, k + 3), vector(TauT_xz, i, j, k + 3), vector(TauT_yz, i, j, k + 3), vector(TauT_zz, i, j, k + 3), rk2inv_chi, rk2sdetg, rk2Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxp3kSfd_y = FkSfd_y_mainI(vector(rk2Sfd_y, i, j, k + 3), vector(rk2gtd_xy, i, j, k + 3), vector(rk2gtd_yy, i, j, k + 3), vector(rk2gtd_yz, i, j, k + 3), vector(rk2Betau_z, i, j, k + 3), vector(rk2Alpha, i, j, k + 3), vector(TauT_xz, i, j, k + 3), vector(TauT_yz, i, j, k + 3), vector(TauT_zz, i, j, k + 3), rk2inv_chi, rk2sdetg, rk2Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxp3kSfd_z = FkSfd_z_mainI(vector(rk2Sfd_z, i, j, k + 3), vector(rk2gtd_xz, i, j, k + 3), vector(rk2gtd_yz, i, j, k + 3), vector(rk2gtd_zz, i, j, k + 3), vector(rk2Betau_z, i, j, k + 3), vector(rk2Alpha, i, j, k + 3), vector(TauT_xz, i, j, k + 3), vector(TauT_yz, i, j, k + 3), vector(TauT_zz, i, j, k + 3), rk2inv_chi, rk2sdetg, rk2Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp3kBfu_x = FkBfu_x_mainI(vector(rk2Bfu_x, i, j, k + 3), vector(rk2Bfu_z, i, j, k + 3), vector(rk2phif, i, j, k + 3), vector(rk2Betau_x, i, j, k + 3), vector(rk2Betau_z, i, j, k + 3), vector(rk2Alpha, i, j, k + 3), vector(TauM_xz, i, j, k + 3), rk2chi_max, rk2gtu_xz, rk2sdetg, rk2vfu_x, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3kBfu_y = FkBfu_y_mainI(vector(rk2Bfu_y, i, j, k + 3), vector(rk2Bfu_z, i, j, k + 3), vector(rk2phif, i, j, k + 3), vector(rk2Betau_y, i, j, k + 3), vector(rk2Betau_z, i, j, k + 3), vector(rk2Alpha, i, j, k + 3), vector(TauM_yz, i, j, k + 3), rk2chi_max, rk2gtu_yz, rk2sdetg, rk2vfu_y, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3kBfu_z = FkBfu_z_mainI(vector(rk2phif, i, j, k + 3), vector(rk2Alpha, i, j, k + 3), rk2chi_max, rk2gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp3kDf = FkDf_mainI(vector(rk2Df, i, j, k + 3), vector(rk2Betau_z, i, j, k + 3), vector(rk2Alpha, i, j, k + 3), vector(TauN_z, i, j, k + 3), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3kDYf = FkDYf_mainI(vector(rk2DYf, i, j, k + 3), vector(rk2Betau_z, i, j, k + 3), vector(rk2Alpha, i, j, k + 3), vector(TauNe_z, i, j, k + 3), rk2sdetg, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3ktauf = Fktauf_mainI(vector(rk2Df, i, j, k + 3), vector(rk2tauf, i, j, k + 3), vector(rk2Betau_z, i, j, k + 3), vector(rk2Alpha, i, j, k + 3), vector(TauN_z, i, j, k + 3), rk2sdetg, rk2Sfu_z, rk2vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3kphif = Fkphif_mainI(vector(rk2Bfu_z, i, j, k + 3), vector(rk2phif, i, j, k + 3), vector(rk2Betau_z, i, j, k + 3), vector(rk2Alpha, i, j, k + 3), ch, dx, simPlat_dt, ilast, jlast);
						rk2h = vector(rhof, i, j, k + 3) * (1.0 + vector(epsf, i, j, k + 3)) + vector(pf, i, j, k + 3);
						rk2sqbc = rk2sqBf / rk2sqW + rk2Bfvf * rk2Bfvf;
						rk2sqca = fabs(rk2sqbc / (rk2h + rk2sqbc));
						rk2sqa = vector(sqcs, i, j, k + 3) + rk2sqca - vector(sqcs, i, j, k + 3) * rk2sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk2Betau_z, i, j, k + 3)) + vector(rk2Alpha, i, j, k + 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk2Betau_z, i, j, k + 3)) + vector(rk2Alpha, i, j, k + 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk2Betau_z, i, j, k + 3)) + vector(rk2Alpha, i, j, k + 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk2Betau_z, i, j, k + 3)) + vector(rk2Alpha, i, j, k + 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk2Betau_z, i, j, k + 3)) + vector(rk2Alpha, i, j, k + 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk2Betau_z, i, j, k + 3)) + vector(rk2Alpha, i, j, k + 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk2Betau_z, i, j, k + 3)) + vector(rk2Alpha, i, j, k + 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk2Betau_z, i, j, k + 3)) + vector(rk2Alpha, i, j, k + 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk2Betau_z, i, j, k + 3)) + vector(rk2Alpha, i, j, k + 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk2Betau_z, i, j, k + 3)) + vector(rk2Alpha, i, j, k + 3) / MAX(1.0 - rk2sqvf * rk2sqa, 0.0000000001) * (fabs((1.0 - rk2sqa) * rk2vfu_z) + sqrt(fabs(rk2sqa * (1.0 - rk2sqvf) * ((1.0 - rk2sqvf * rk2sqa) * rk2chi_max * rk2gtu_zz - (1.0 - rk2sqa) * (rk2vfu_z * rk2vfu_z)))))));
						LLFpkSfd_x = 0.5 * (FluxkSfd_x + fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k));
						LLFpkSfd_y = 0.5 * (FluxkSfd_y + fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k));
						LLFpkSfd_z = 0.5 * (FluxkSfd_z + fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k));
						LLFpkBfu_x = 0.5 * (FluxkBfu_x + fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k));
						LLFpkBfu_y = 0.5 * (FluxkBfu_y + fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k));
						LLFpkBfu_z = 0.5 * (FluxkBfu_z + fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k));
						LLFpkDf = 0.5 * (FluxkDf + fabs(SpeedkDf) * vector(rk2Df, i, j, k));
						LLFpkDYf = 0.5 * (FluxkDYf + fabs(SpeedkDYf) * vector(rk2DYf, i, j, k));
						LLFpktauf = 0.5 * (Fluxktauf + fabs(Speedktauf) * vector(rk2tauf, i, j, k));
						LLFpkphif = 0.5 * (Fluxkphif + fabs(Speedkphif) * vector(rk2phif, i, j, k));
						LLFnkSfd_x = 0.5 * (FluxkSfd_x - fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k));
						LLFnkSfd_y = 0.5 * (FluxkSfd_y - fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k));
						LLFnkSfd_z = 0.5 * (FluxkSfd_z - fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k));
						LLFnkBfu_x = 0.5 * (FluxkBfu_x - fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k));
						LLFnkBfu_y = 0.5 * (FluxkBfu_y - fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k));
						LLFnkBfu_z = 0.5 * (FluxkBfu_z - fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k));
						LLFnkDf = 0.5 * (FluxkDf - fabs(SpeedkDf) * vector(rk2Df, i, j, k));
						LLFnkDYf = 0.5 * (FluxkDYf - fabs(SpeedkDYf) * vector(rk2DYf, i, j, k));
						LLFnktauf = 0.5 * (Fluxktauf - fabs(Speedktauf) * vector(rk2tauf, i, j, k));
						LLFnkphif = 0.5 * (Fluxkphif - fabs(Speedkphif) * vector(rk2phif, i, j, k));
						LLFp_m3kSfd_x = 0.5 * (Fluxm3kSfd_x + fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k - 3));
						LLFp_m3kSfd_y = 0.5 * (Fluxm3kSfd_y + fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k - 3));
						LLFp_m3kSfd_z = 0.5 * (Fluxm3kSfd_z + fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k - 3));
						LLFp_m3kBfu_x = 0.5 * (Fluxm3kBfu_x + fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k - 3));
						LLFp_m3kBfu_y = 0.5 * (Fluxm3kBfu_y + fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k - 3));
						LLFp_m3kBfu_z = 0.5 * (Fluxm3kBfu_z + fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k - 3));
						LLFp_m3kDf = 0.5 * (Fluxm3kDf + fabs(SpeedkDf) * vector(rk2Df, i, j, k - 3));
						LLFp_m3kDYf = 0.5 * (Fluxm3kDYf + fabs(SpeedkDYf) * vector(rk2DYf, i, j, k - 3));
						LLFp_m3ktauf = 0.5 * (Fluxm3ktauf + fabs(Speedktauf) * vector(rk2tauf, i, j, k - 3));
						LLFp_m3kphif = 0.5 * (Fluxm3kphif + fabs(Speedkphif) * vector(rk2phif, i, j, k - 3));
						LLFn_m3kSfd_x = 0.5 * (Fluxm3kSfd_x - fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k - 3));
						LLFn_m3kSfd_y = 0.5 * (Fluxm3kSfd_y - fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k - 3));
						LLFn_m3kSfd_z = 0.5 * (Fluxm3kSfd_z - fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k - 3));
						LLFn_m3kBfu_x = 0.5 * (Fluxm3kBfu_x - fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k - 3));
						LLFn_m3kBfu_y = 0.5 * (Fluxm3kBfu_y - fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k - 3));
						LLFn_m3kBfu_z = 0.5 * (Fluxm3kBfu_z - fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k - 3));
						LLFn_m3kDf = 0.5 * (Fluxm3kDf - fabs(SpeedkDf) * vector(rk2Df, i, j, k - 3));
						LLFn_m3kDYf = 0.5 * (Fluxm3kDYf - fabs(SpeedkDYf) * vector(rk2DYf, i, j, k - 3));
						LLFn_m3ktauf = 0.5 * (Fluxm3ktauf - fabs(Speedktauf) * vector(rk2tauf, i, j, k - 3));
						LLFn_m3kphif = 0.5 * (Fluxm3kphif - fabs(Speedkphif) * vector(rk2phif, i, j, k - 3));
						LLFp_m2kSfd_x = 0.5 * (Fluxm2kSfd_x + fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k - 2));
						LLFp_m2kSfd_y = 0.5 * (Fluxm2kSfd_y + fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k - 2));
						LLFp_m2kSfd_z = 0.5 * (Fluxm2kSfd_z + fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k - 2));
						LLFp_m2kBfu_x = 0.5 * (Fluxm2kBfu_x + fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k - 2));
						LLFp_m2kBfu_y = 0.5 * (Fluxm2kBfu_y + fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k - 2));
						LLFp_m2kBfu_z = 0.5 * (Fluxm2kBfu_z + fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k - 2));
						LLFp_m2kDf = 0.5 * (Fluxm2kDf + fabs(SpeedkDf) * vector(rk2Df, i, j, k - 2));
						LLFp_m2kDYf = 0.5 * (Fluxm2kDYf + fabs(SpeedkDYf) * vector(rk2DYf, i, j, k - 2));
						LLFp_m2ktauf = 0.5 * (Fluxm2ktauf + fabs(Speedktauf) * vector(rk2tauf, i, j, k - 2));
						LLFp_m2kphif = 0.5 * (Fluxm2kphif + fabs(Speedkphif) * vector(rk2phif, i, j, k - 2));
						LLFn_m2kSfd_x = 0.5 * (Fluxm2kSfd_x - fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k - 2));
						LLFn_m2kSfd_y = 0.5 * (Fluxm2kSfd_y - fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k - 2));
						LLFn_m2kSfd_z = 0.5 * (Fluxm2kSfd_z - fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k - 2));
						LLFn_m2kBfu_x = 0.5 * (Fluxm2kBfu_x - fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k - 2));
						LLFn_m2kBfu_y = 0.5 * (Fluxm2kBfu_y - fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k - 2));
						LLFn_m2kBfu_z = 0.5 * (Fluxm2kBfu_z - fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k - 2));
						LLFn_m2kDf = 0.5 * (Fluxm2kDf - fabs(SpeedkDf) * vector(rk2Df, i, j, k - 2));
						LLFn_m2kDYf = 0.5 * (Fluxm2kDYf - fabs(SpeedkDYf) * vector(rk2DYf, i, j, k - 2));
						LLFn_m2ktauf = 0.5 * (Fluxm2ktauf - fabs(Speedktauf) * vector(rk2tauf, i, j, k - 2));
						LLFn_m2kphif = 0.5 * (Fluxm2kphif - fabs(Speedkphif) * vector(rk2phif, i, j, k - 2));
						LLFp_m1kSfd_x = 0.5 * (Fluxm1kSfd_x + fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k - 1));
						LLFp_m1kSfd_y = 0.5 * (Fluxm1kSfd_y + fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k - 1));
						LLFp_m1kSfd_z = 0.5 * (Fluxm1kSfd_z + fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k - 1));
						LLFp_m1kBfu_x = 0.5 * (Fluxm1kBfu_x + fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k - 1));
						LLFp_m1kBfu_y = 0.5 * (Fluxm1kBfu_y + fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k - 1));
						LLFp_m1kBfu_z = 0.5 * (Fluxm1kBfu_z + fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k - 1));
						LLFp_m1kDf = 0.5 * (Fluxm1kDf + fabs(SpeedkDf) * vector(rk2Df, i, j, k - 1));
						LLFp_m1kDYf = 0.5 * (Fluxm1kDYf + fabs(SpeedkDYf) * vector(rk2DYf, i, j, k - 1));
						LLFp_m1ktauf = 0.5 * (Fluxm1ktauf + fabs(Speedktauf) * vector(rk2tauf, i, j, k - 1));
						LLFp_m1kphif = 0.5 * (Fluxm1kphif + fabs(Speedkphif) * vector(rk2phif, i, j, k - 1));
						LLFn_m1kSfd_x = 0.5 * (Fluxm1kSfd_x - fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k - 1));
						LLFn_m1kSfd_y = 0.5 * (Fluxm1kSfd_y - fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k - 1));
						LLFn_m1kSfd_z = 0.5 * (Fluxm1kSfd_z - fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k - 1));
						LLFn_m1kBfu_x = 0.5 * (Fluxm1kBfu_x - fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k - 1));
						LLFn_m1kBfu_y = 0.5 * (Fluxm1kBfu_y - fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k - 1));
						LLFn_m1kBfu_z = 0.5 * (Fluxm1kBfu_z - fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k - 1));
						LLFn_m1kDf = 0.5 * (Fluxm1kDf - fabs(SpeedkDf) * vector(rk2Df, i, j, k - 1));
						LLFn_m1kDYf = 0.5 * (Fluxm1kDYf - fabs(SpeedkDYf) * vector(rk2DYf, i, j, k - 1));
						LLFn_m1ktauf = 0.5 * (Fluxm1ktauf - fabs(Speedktauf) * vector(rk2tauf, i, j, k - 1));
						LLFn_m1kphif = 0.5 * (Fluxm1kphif - fabs(Speedkphif) * vector(rk2phif, i, j, k - 1));
						LLFp_p1kSfd_x = 0.5 * (Fluxp1kSfd_x + fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k + 1));
						LLFp_p1kSfd_y = 0.5 * (Fluxp1kSfd_y + fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k + 1));
						LLFp_p1kSfd_z = 0.5 * (Fluxp1kSfd_z + fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k + 1));
						LLFp_p1kBfu_x = 0.5 * (Fluxp1kBfu_x + fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k + 1));
						LLFp_p1kBfu_y = 0.5 * (Fluxp1kBfu_y + fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k + 1));
						LLFp_p1kBfu_z = 0.5 * (Fluxp1kBfu_z + fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k + 1));
						LLFp_p1kDf = 0.5 * (Fluxp1kDf + fabs(SpeedkDf) * vector(rk2Df, i, j, k + 1));
						LLFp_p1kDYf = 0.5 * (Fluxp1kDYf + fabs(SpeedkDYf) * vector(rk2DYf, i, j, k + 1));
						LLFp_p1ktauf = 0.5 * (Fluxp1ktauf + fabs(Speedktauf) * vector(rk2tauf, i, j, k + 1));
						LLFp_p1kphif = 0.5 * (Fluxp1kphif + fabs(Speedkphif) * vector(rk2phif, i, j, k + 1));
						LLFn_p1kSfd_x = 0.5 * (Fluxp1kSfd_x - fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k + 1));
						LLFn_p1kSfd_y = 0.5 * (Fluxp1kSfd_y - fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k + 1));
						LLFn_p1kSfd_z = 0.5 * (Fluxp1kSfd_z - fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k + 1));
						LLFn_p1kBfu_x = 0.5 * (Fluxp1kBfu_x - fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k + 1));
						LLFn_p1kBfu_y = 0.5 * (Fluxp1kBfu_y - fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k + 1));
						LLFn_p1kBfu_z = 0.5 * (Fluxp1kBfu_z - fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k + 1));
						LLFn_p1kDf = 0.5 * (Fluxp1kDf - fabs(SpeedkDf) * vector(rk2Df, i, j, k + 1));
						LLFn_p1kDYf = 0.5 * (Fluxp1kDYf - fabs(SpeedkDYf) * vector(rk2DYf, i, j, k + 1));
						LLFn_p1ktauf = 0.5 * (Fluxp1ktauf - fabs(Speedktauf) * vector(rk2tauf, i, j, k + 1));
						LLFn_p1kphif = 0.5 * (Fluxp1kphif - fabs(Speedkphif) * vector(rk2phif, i, j, k + 1));
						LLFp_p2kSfd_x = 0.5 * (Fluxp2kSfd_x + fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k + 2));
						LLFp_p2kSfd_y = 0.5 * (Fluxp2kSfd_y + fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k + 2));
						LLFp_p2kSfd_z = 0.5 * (Fluxp2kSfd_z + fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k + 2));
						LLFp_p2kBfu_x = 0.5 * (Fluxp2kBfu_x + fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k + 2));
						LLFp_p2kBfu_y = 0.5 * (Fluxp2kBfu_y + fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k + 2));
						LLFp_p2kBfu_z = 0.5 * (Fluxp2kBfu_z + fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k + 2));
						LLFp_p2kDf = 0.5 * (Fluxp2kDf + fabs(SpeedkDf) * vector(rk2Df, i, j, k + 2));
						LLFp_p2kDYf = 0.5 * (Fluxp2kDYf + fabs(SpeedkDYf) * vector(rk2DYf, i, j, k + 2));
						LLFp_p2ktauf = 0.5 * (Fluxp2ktauf + fabs(Speedktauf) * vector(rk2tauf, i, j, k + 2));
						LLFp_p2kphif = 0.5 * (Fluxp2kphif + fabs(Speedkphif) * vector(rk2phif, i, j, k + 2));
						LLFn_p2kSfd_x = 0.5 * (Fluxp2kSfd_x - fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k + 2));
						LLFn_p2kSfd_y = 0.5 * (Fluxp2kSfd_y - fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k + 2));
						LLFn_p2kSfd_z = 0.5 * (Fluxp2kSfd_z - fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k + 2));
						LLFn_p2kBfu_x = 0.5 * (Fluxp2kBfu_x - fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k + 2));
						LLFn_p2kBfu_y = 0.5 * (Fluxp2kBfu_y - fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k + 2));
						LLFn_p2kBfu_z = 0.5 * (Fluxp2kBfu_z - fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k + 2));
						LLFn_p2kDf = 0.5 * (Fluxp2kDf - fabs(SpeedkDf) * vector(rk2Df, i, j, k + 2));
						LLFn_p2kDYf = 0.5 * (Fluxp2kDYf - fabs(SpeedkDYf) * vector(rk2DYf, i, j, k + 2));
						LLFn_p2ktauf = 0.5 * (Fluxp2ktauf - fabs(Speedktauf) * vector(rk2tauf, i, j, k + 2));
						LLFn_p2kphif = 0.5 * (Fluxp2kphif - fabs(Speedkphif) * vector(rk2phif, i, j, k + 2));
						LLFp_p3kSfd_x = 0.5 * (Fluxp3kSfd_x + fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k + 3));
						LLFp_p3kSfd_y = 0.5 * (Fluxp3kSfd_y + fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k + 3));
						LLFp_p3kSfd_z = 0.5 * (Fluxp3kSfd_z + fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k + 3));
						LLFp_p3kBfu_x = 0.5 * (Fluxp3kBfu_x + fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k + 3));
						LLFp_p3kBfu_y = 0.5 * (Fluxp3kBfu_y + fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k + 3));
						LLFp_p3kBfu_z = 0.5 * (Fluxp3kBfu_z + fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k + 3));
						LLFp_p3kDf = 0.5 * (Fluxp3kDf + fabs(SpeedkDf) * vector(rk2Df, i, j, k + 3));
						LLFp_p3kDYf = 0.5 * (Fluxp3kDYf + fabs(SpeedkDYf) * vector(rk2DYf, i, j, k + 3));
						LLFp_p3ktauf = 0.5 * (Fluxp3ktauf + fabs(Speedktauf) * vector(rk2tauf, i, j, k + 3));
						LLFp_p3kphif = 0.5 * (Fluxp3kphif + fabs(Speedkphif) * vector(rk2phif, i, j, k + 3));
						LLFn_p3kSfd_x = 0.5 * (Fluxp3kSfd_x - fabs(SpeedkSfd_x) * vector(rk2Sfd_x, i, j, k + 3));
						LLFn_p3kSfd_y = 0.5 * (Fluxp3kSfd_y - fabs(SpeedkSfd_y) * vector(rk2Sfd_y, i, j, k + 3));
						LLFn_p3kSfd_z = 0.5 * (Fluxp3kSfd_z - fabs(SpeedkSfd_z) * vector(rk2Sfd_z, i, j, k + 3));
						LLFn_p3kBfu_x = 0.5 * (Fluxp3kBfu_x - fabs(SpeedkBfu_x) * vector(rk2Bfu_x, i, j, k + 3));
						LLFn_p3kBfu_y = 0.5 * (Fluxp3kBfu_y - fabs(SpeedkBfu_y) * vector(rk2Bfu_y, i, j, k + 3));
						LLFn_p3kBfu_z = 0.5 * (Fluxp3kBfu_z - fabs(SpeedkBfu_z) * vector(rk2Bfu_z, i, j, k + 3));
						LLFn_p3kDf = 0.5 * (Fluxp3kDf - fabs(SpeedkDf) * vector(rk2Df, i, j, k + 3));
						LLFn_p3kDYf = 0.5 * (Fluxp3kDYf - fabs(SpeedkDYf) * vector(rk2DYf, i, j, k + 3));
						LLFn_p3ktauf = 0.5 * (Fluxp3ktauf - fabs(Speedktauf) * vector(rk2tauf, i, j, k + 3));
						LLFn_p3kphif = 0.5 * (Fluxp3kphif - fabs(Speedkphif) * vector(rk2phif, i, j, k + 3));
						Sfd_x_lni = MP5(LLFp_m3iSfd_x, LLFp_m2iSfd_x, LLFp_m1iSfd_x, LLFpiSfd_x, LLFp_p1iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lni = MP5(LLFp_m3iSfd_y, LLFp_m2iSfd_y, LLFp_m1iSfd_y, LLFpiSfd_y, LLFp_p1iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lni = MP5(LLFp_m3iSfd_z, LLFp_m2iSfd_z, LLFp_m1iSfd_z, LLFpiSfd_z, LLFp_p1iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lni = MP5(LLFp_m3iBfu_x, LLFp_m2iBfu_x, LLFp_m1iBfu_x, LLFpiBfu_x, LLFp_p1iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lni = MP5(LLFp_m3iBfu_y, LLFp_m2iBfu_y, LLFp_m1iBfu_y, LLFpiBfu_y, LLFp_p1iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lni = MP5(LLFp_m3iBfu_z, LLFp_m2iBfu_z, LLFp_m1iBfu_z, LLFpiBfu_z, LLFp_p1iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lni = MP5(LLFp_m3iDf, LLFp_m2iDf, LLFp_m1iDf, LLFpiDf, LLFp_p1iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lni = MP5(LLFp_m3iDYf, LLFp_m2iDYf, LLFp_m1iDYf, LLFpiDYf, LLFp_p1iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lni = MP5(LLFp_m3itauf, LLFp_m2itauf, LLFp_m1itauf, LLFpitauf, LLFp_p1itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lni = MP5(LLFp_m3iphif, LLFp_m2iphif, LLFp_m1iphif, LLFpiphif, LLFp_p1iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lnj = MP5(LLFp_m3jSfd_x, LLFp_m2jSfd_x, LLFp_m1jSfd_x, LLFpjSfd_x, LLFp_p1jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lnj = MP5(LLFp_m3jSfd_y, LLFp_m2jSfd_y, LLFp_m1jSfd_y, LLFpjSfd_y, LLFp_p1jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lnj = MP5(LLFp_m3jSfd_z, LLFp_m2jSfd_z, LLFp_m1jSfd_z, LLFpjSfd_z, LLFp_p1jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lnj = MP5(LLFp_m3jBfu_x, LLFp_m2jBfu_x, LLFp_m1jBfu_x, LLFpjBfu_x, LLFp_p1jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lnj = MP5(LLFp_m3jBfu_y, LLFp_m2jBfu_y, LLFp_m1jBfu_y, LLFpjBfu_y, LLFp_p1jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lnj = MP5(LLFp_m3jBfu_z, LLFp_m2jBfu_z, LLFp_m1jBfu_z, LLFpjBfu_z, LLFp_p1jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lnj = MP5(LLFp_m3jDf, LLFp_m2jDf, LLFp_m1jDf, LLFpjDf, LLFp_p1jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lnj = MP5(LLFp_m3jDYf, LLFp_m2jDYf, LLFp_m1jDYf, LLFpjDYf, LLFp_p1jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lnj = MP5(LLFp_m3jtauf, LLFp_m2jtauf, LLFp_m1jtauf, LLFpjtauf, LLFp_p1jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lnj = MP5(LLFp_m3jphif, LLFp_m2jphif, LLFp_m1jphif, LLFpjphif, LLFp_p1jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lnk = MP5(LLFp_m3kSfd_x, LLFp_m2kSfd_x, LLFp_m1kSfd_x, LLFpkSfd_x, LLFp_p1kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lnk = MP5(LLFp_m3kSfd_y, LLFp_m2kSfd_y, LLFp_m1kSfd_y, LLFpkSfd_y, LLFp_p1kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lnk = MP5(LLFp_m3kSfd_z, LLFp_m2kSfd_z, LLFp_m1kSfd_z, LLFpkSfd_z, LLFp_p1kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lnk = MP5(LLFp_m3kBfu_x, LLFp_m2kBfu_x, LLFp_m1kBfu_x, LLFpkBfu_x, LLFp_p1kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lnk = MP5(LLFp_m3kBfu_y, LLFp_m2kBfu_y, LLFp_m1kBfu_y, LLFpkBfu_y, LLFp_p1kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lnk = MP5(LLFp_m3kBfu_z, LLFp_m2kBfu_z, LLFp_m1kBfu_z, LLFpkBfu_z, LLFp_p1kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lnk = MP5(LLFp_m3kDf, LLFp_m2kDf, LLFp_m1kDf, LLFpkDf, LLFp_p1kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lnk = MP5(LLFp_m3kDYf, LLFp_m2kDYf, LLFp_m1kDYf, LLFpkDYf, LLFp_p1kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lnk = MP5(LLFp_m3ktauf, LLFp_m2ktauf, LLFp_m1ktauf, LLFpktauf, LLFp_p1ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lnk = MP5(LLFp_m3kphif, LLFp_m2kphif, LLFp_m1kphif, LLFpkphif, LLFp_p1kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rni = MP5(LLFn_p2iSfd_x, LLFn_p1iSfd_x, LLFniSfd_x, LLFn_m1iSfd_x, LLFn_m2iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rni = MP5(LLFn_p2iSfd_y, LLFn_p1iSfd_y, LLFniSfd_y, LLFn_m1iSfd_y, LLFn_m2iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rni = MP5(LLFn_p2iSfd_z, LLFn_p1iSfd_z, LLFniSfd_z, LLFn_m1iSfd_z, LLFn_m2iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rni = MP5(LLFn_p2iBfu_x, LLFn_p1iBfu_x, LLFniBfu_x, LLFn_m1iBfu_x, LLFn_m2iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rni = MP5(LLFn_p2iBfu_y, LLFn_p1iBfu_y, LLFniBfu_y, LLFn_m1iBfu_y, LLFn_m2iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rni = MP5(LLFn_p2iBfu_z, LLFn_p1iBfu_z, LLFniBfu_z, LLFn_m1iBfu_z, LLFn_m2iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rni = MP5(LLFn_p2iDf, LLFn_p1iDf, LLFniDf, LLFn_m1iDf, LLFn_m2iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rni = MP5(LLFn_p2iDYf, LLFn_p1iDYf, LLFniDYf, LLFn_m1iDYf, LLFn_m2iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rni = MP5(LLFn_p2itauf, LLFn_p1itauf, LLFnitauf, LLFn_m1itauf, LLFn_m2itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rni = MP5(LLFn_p2iphif, LLFn_p1iphif, LLFniphif, LLFn_m1iphif, LLFn_m2iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rnj = MP5(LLFn_p2jSfd_x, LLFn_p1jSfd_x, LLFnjSfd_x, LLFn_m1jSfd_x, LLFn_m2jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rnj = MP5(LLFn_p2jSfd_y, LLFn_p1jSfd_y, LLFnjSfd_y, LLFn_m1jSfd_y, LLFn_m2jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rnj = MP5(LLFn_p2jSfd_z, LLFn_p1jSfd_z, LLFnjSfd_z, LLFn_m1jSfd_z, LLFn_m2jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rnj = MP5(LLFn_p2jBfu_x, LLFn_p1jBfu_x, LLFnjBfu_x, LLFn_m1jBfu_x, LLFn_m2jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rnj = MP5(LLFn_p2jBfu_y, LLFn_p1jBfu_y, LLFnjBfu_y, LLFn_m1jBfu_y, LLFn_m2jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rnj = MP5(LLFn_p2jBfu_z, LLFn_p1jBfu_z, LLFnjBfu_z, LLFn_m1jBfu_z, LLFn_m2jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rnj = MP5(LLFn_p2jDf, LLFn_p1jDf, LLFnjDf, LLFn_m1jDf, LLFn_m2jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rnj = MP5(LLFn_p2jDYf, LLFn_p1jDYf, LLFnjDYf, LLFn_m1jDYf, LLFn_m2jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rnj = MP5(LLFn_p2jtauf, LLFn_p1jtauf, LLFnjtauf, LLFn_m1jtauf, LLFn_m2jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rnj = MP5(LLFn_p2jphif, LLFn_p1jphif, LLFnjphif, LLFn_m1jphif, LLFn_m2jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rnk = MP5(LLFn_p2kSfd_x, LLFn_p1kSfd_x, LLFnkSfd_x, LLFn_m1kSfd_x, LLFn_m2kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rnk = MP5(LLFn_p2kSfd_y, LLFn_p1kSfd_y, LLFnkSfd_y, LLFn_m1kSfd_y, LLFn_m2kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rnk = MP5(LLFn_p2kSfd_z, LLFn_p1kSfd_z, LLFnkSfd_z, LLFn_m1kSfd_z, LLFn_m2kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rnk = MP5(LLFn_p2kBfu_x, LLFn_p1kBfu_x, LLFnkBfu_x, LLFn_m1kBfu_x, LLFn_m2kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rnk = MP5(LLFn_p2kBfu_y, LLFn_p1kBfu_y, LLFnkBfu_y, LLFn_m1kBfu_y, LLFn_m2kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rnk = MP5(LLFn_p2kBfu_z, LLFn_p1kBfu_z, LLFnkBfu_z, LLFn_m1kBfu_z, LLFn_m2kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rnk = MP5(LLFn_p2kDf, LLFn_p1kDf, LLFnkDf, LLFn_m1kDf, LLFn_m2kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rnk = MP5(LLFn_p2kDYf, LLFn_p1kDYf, LLFnkDYf, LLFn_m1kDYf, LLFn_m2kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rnk = MP5(LLFn_p2ktauf, LLFn_p1ktauf, LLFnktauf, LLFn_m1ktauf, LLFn_m2ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rnk = MP5(LLFn_p2kphif, LLFn_p1kphif, LLFnkphif, LLFn_m1kphif, LLFn_m2kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lpi = MP5(LLFp_m2iSfd_x, LLFp_m1iSfd_x, LLFpiSfd_x, LLFp_p1iSfd_x, LLFp_p2iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lpi = MP5(LLFp_m2iSfd_y, LLFp_m1iSfd_y, LLFpiSfd_y, LLFp_p1iSfd_y, LLFp_p2iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lpi = MP5(LLFp_m2iSfd_z, LLFp_m1iSfd_z, LLFpiSfd_z, LLFp_p1iSfd_z, LLFp_p2iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lpi = MP5(LLFp_m2iBfu_x, LLFp_m1iBfu_x, LLFpiBfu_x, LLFp_p1iBfu_x, LLFp_p2iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lpi = MP5(LLFp_m2iBfu_y, LLFp_m1iBfu_y, LLFpiBfu_y, LLFp_p1iBfu_y, LLFp_p2iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lpi = MP5(LLFp_m2iBfu_z, LLFp_m1iBfu_z, LLFpiBfu_z, LLFp_p1iBfu_z, LLFp_p2iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lpi = MP5(LLFp_m2iDf, LLFp_m1iDf, LLFpiDf, LLFp_p1iDf, LLFp_p2iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lpi = MP5(LLFp_m2iDYf, LLFp_m1iDYf, LLFpiDYf, LLFp_p1iDYf, LLFp_p2iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lpi = MP5(LLFp_m2itauf, LLFp_m1itauf, LLFpitauf, LLFp_p1itauf, LLFp_p2itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lpi = MP5(LLFp_m2iphif, LLFp_m1iphif, LLFpiphif, LLFp_p1iphif, LLFp_p2iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lpj = MP5(LLFp_m2jSfd_x, LLFp_m1jSfd_x, LLFpjSfd_x, LLFp_p1jSfd_x, LLFp_p2jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lpj = MP5(LLFp_m2jSfd_y, LLFp_m1jSfd_y, LLFpjSfd_y, LLFp_p1jSfd_y, LLFp_p2jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lpj = MP5(LLFp_m2jSfd_z, LLFp_m1jSfd_z, LLFpjSfd_z, LLFp_p1jSfd_z, LLFp_p2jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lpj = MP5(LLFp_m2jBfu_x, LLFp_m1jBfu_x, LLFpjBfu_x, LLFp_p1jBfu_x, LLFp_p2jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lpj = MP5(LLFp_m2jBfu_y, LLFp_m1jBfu_y, LLFpjBfu_y, LLFp_p1jBfu_y, LLFp_p2jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lpj = MP5(LLFp_m2jBfu_z, LLFp_m1jBfu_z, LLFpjBfu_z, LLFp_p1jBfu_z, LLFp_p2jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lpj = MP5(LLFp_m2jDf, LLFp_m1jDf, LLFpjDf, LLFp_p1jDf, LLFp_p2jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lpj = MP5(LLFp_m2jDYf, LLFp_m1jDYf, LLFpjDYf, LLFp_p1jDYf, LLFp_p2jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lpj = MP5(LLFp_m2jtauf, LLFp_m1jtauf, LLFpjtauf, LLFp_p1jtauf, LLFp_p2jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lpj = MP5(LLFp_m2jphif, LLFp_m1jphif, LLFpjphif, LLFp_p1jphif, LLFp_p2jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lpk = MP5(LLFp_m2kSfd_x, LLFp_m1kSfd_x, LLFpkSfd_x, LLFp_p1kSfd_x, LLFp_p2kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lpk = MP5(LLFp_m2kSfd_y, LLFp_m1kSfd_y, LLFpkSfd_y, LLFp_p1kSfd_y, LLFp_p2kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lpk = MP5(LLFp_m2kSfd_z, LLFp_m1kSfd_z, LLFpkSfd_z, LLFp_p1kSfd_z, LLFp_p2kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lpk = MP5(LLFp_m2kBfu_x, LLFp_m1kBfu_x, LLFpkBfu_x, LLFp_p1kBfu_x, LLFp_p2kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lpk = MP5(LLFp_m2kBfu_y, LLFp_m1kBfu_y, LLFpkBfu_y, LLFp_p1kBfu_y, LLFp_p2kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lpk = MP5(LLFp_m2kBfu_z, LLFp_m1kBfu_z, LLFpkBfu_z, LLFp_p1kBfu_z, LLFp_p2kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lpk = MP5(LLFp_m2kDf, LLFp_m1kDf, LLFpkDf, LLFp_p1kDf, LLFp_p2kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lpk = MP5(LLFp_m2kDYf, LLFp_m1kDYf, LLFpkDYf, LLFp_p1kDYf, LLFp_p2kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lpk = MP5(LLFp_m2ktauf, LLFp_m1ktauf, LLFpktauf, LLFp_p1ktauf, LLFp_p2ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lpk = MP5(LLFp_m2kphif, LLFp_m1kphif, LLFpkphif, LLFp_p1kphif, LLFp_p2kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rpi = MP5(LLFn_p3iSfd_x, LLFn_p2iSfd_x, LLFn_p1iSfd_x, LLFniSfd_x, LLFn_m1iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rpi = MP5(LLFn_p3iSfd_y, LLFn_p2iSfd_y, LLFn_p1iSfd_y, LLFniSfd_y, LLFn_m1iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rpi = MP5(LLFn_p3iSfd_z, LLFn_p2iSfd_z, LLFn_p1iSfd_z, LLFniSfd_z, LLFn_m1iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rpi = MP5(LLFn_p3iBfu_x, LLFn_p2iBfu_x, LLFn_p1iBfu_x, LLFniBfu_x, LLFn_m1iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rpi = MP5(LLFn_p3iBfu_y, LLFn_p2iBfu_y, LLFn_p1iBfu_y, LLFniBfu_y, LLFn_m1iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rpi = MP5(LLFn_p3iBfu_z, LLFn_p2iBfu_z, LLFn_p1iBfu_z, LLFniBfu_z, LLFn_m1iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rpi = MP5(LLFn_p3iDf, LLFn_p2iDf, LLFn_p1iDf, LLFniDf, LLFn_m1iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rpi = MP5(LLFn_p3iDYf, LLFn_p2iDYf, LLFn_p1iDYf, LLFniDYf, LLFn_m1iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rpi = MP5(LLFn_p3itauf, LLFn_p2itauf, LLFn_p1itauf, LLFnitauf, LLFn_m1itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rpi = MP5(LLFn_p3iphif, LLFn_p2iphif, LLFn_p1iphif, LLFniphif, LLFn_m1iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rpj = MP5(LLFn_p3jSfd_x, LLFn_p2jSfd_x, LLFn_p1jSfd_x, LLFnjSfd_x, LLFn_m1jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rpj = MP5(LLFn_p3jSfd_y, LLFn_p2jSfd_y, LLFn_p1jSfd_y, LLFnjSfd_y, LLFn_m1jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rpj = MP5(LLFn_p3jSfd_z, LLFn_p2jSfd_z, LLFn_p1jSfd_z, LLFnjSfd_z, LLFn_m1jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rpj = MP5(LLFn_p3jBfu_x, LLFn_p2jBfu_x, LLFn_p1jBfu_x, LLFnjBfu_x, LLFn_m1jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rpj = MP5(LLFn_p3jBfu_y, LLFn_p2jBfu_y, LLFn_p1jBfu_y, LLFnjBfu_y, LLFn_m1jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rpj = MP5(LLFn_p3jBfu_z, LLFn_p2jBfu_z, LLFn_p1jBfu_z, LLFnjBfu_z, LLFn_m1jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rpj = MP5(LLFn_p3jDf, LLFn_p2jDf, LLFn_p1jDf, LLFnjDf, LLFn_m1jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rpj = MP5(LLFn_p3jDYf, LLFn_p2jDYf, LLFn_p1jDYf, LLFnjDYf, LLFn_m1jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rpj = MP5(LLFn_p3jtauf, LLFn_p2jtauf, LLFn_p1jtauf, LLFnjtauf, LLFn_m1jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rpj = MP5(LLFn_p3jphif, LLFn_p2jphif, LLFn_p1jphif, LLFnjphif, LLFn_m1jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rpk = MP5(LLFn_p3kSfd_x, LLFn_p2kSfd_x, LLFn_p1kSfd_x, LLFnkSfd_x, LLFn_m1kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rpk = MP5(LLFn_p3kSfd_y, LLFn_p2kSfd_y, LLFn_p1kSfd_y, LLFnkSfd_y, LLFn_m1kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rpk = MP5(LLFn_p3kSfd_z, LLFn_p2kSfd_z, LLFn_p1kSfd_z, LLFnkSfd_z, LLFn_m1kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rpk = MP5(LLFn_p3kBfu_x, LLFn_p2kBfu_x, LLFn_p1kBfu_x, LLFnkBfu_x, LLFn_m1kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rpk = MP5(LLFn_p3kBfu_y, LLFn_p2kBfu_y, LLFn_p1kBfu_y, LLFnkBfu_y, LLFn_m1kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rpk = MP5(LLFn_p3kBfu_z, LLFn_p2kBfu_z, LLFn_p1kBfu_z, LLFnkBfu_z, LLFn_m1kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rpk = MP5(LLFn_p3kDf, LLFn_p2kDf, LLFn_p1kDf, LLFnkDf, LLFn_m1kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rpk = MP5(LLFn_p3kDYf, LLFn_p2kDYf, LLFn_p1kDYf, LLFnkDYf, LLFn_m1kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rpk = MP5(LLFn_p3ktauf, LLFn_p2ktauf, LLFn_p1ktauf, LLFnktauf, LLFn_m1ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rpk = MP5(LLFn_p3kphif, LLFn_p2kphif, LLFn_p1kphif, LLFnkphif, LLFn_m1kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						FluxniSfd_x = Sfd_x_lni + Sfd_x_rni;
						FluxniSfd_y = Sfd_y_lni + Sfd_y_rni;
						FluxniSfd_z = Sfd_z_lni + Sfd_z_rni;
						FluxniBfu_x = Bfu_x_lni + Bfu_x_rni;
						FluxniBfu_y = Bfu_y_lni + Bfu_y_rni;
						FluxniBfu_z = Bfu_z_lni + Bfu_z_rni;
						FluxniDf = Df_lni + Df_rni;
						FluxniDYf = DYf_lni + DYf_rni;
						Fluxnitauf = tauf_lni + tauf_rni;
						Fluxniphif = phif_lni + phif_rni;
						FluxnjSfd_x = Sfd_x_lnj + Sfd_x_rnj;
						FluxnjSfd_y = Sfd_y_lnj + Sfd_y_rnj;
						FluxnjSfd_z = Sfd_z_lnj + Sfd_z_rnj;
						FluxnjBfu_x = Bfu_x_lnj + Bfu_x_rnj;
						FluxnjBfu_y = Bfu_y_lnj + Bfu_y_rnj;
						FluxnjBfu_z = Bfu_z_lnj + Bfu_z_rnj;
						FluxnjDf = Df_lnj + Df_rnj;
						FluxnjDYf = DYf_lnj + DYf_rnj;
						Fluxnjtauf = tauf_lnj + tauf_rnj;
						Fluxnjphif = phif_lnj + phif_rnj;
						FluxnkSfd_x = Sfd_x_lnk + Sfd_x_rnk;
						FluxnkSfd_y = Sfd_y_lnk + Sfd_y_rnk;
						FluxnkSfd_z = Sfd_z_lnk + Sfd_z_rnk;
						FluxnkBfu_x = Bfu_x_lnk + Bfu_x_rnk;
						FluxnkBfu_y = Bfu_y_lnk + Bfu_y_rnk;
						FluxnkBfu_z = Bfu_z_lnk + Bfu_z_rnk;
						FluxnkDf = Df_lnk + Df_rnk;
						FluxnkDYf = DYf_lnk + DYf_rnk;
						Fluxnktauf = tauf_lnk + tauf_rnk;
						Fluxnkphif = phif_lnk + phif_rnk;
						FluxpiSfd_x = Sfd_x_lpi + Sfd_x_rpi;
						FluxpiSfd_y = Sfd_y_lpi + Sfd_y_rpi;
						FluxpiSfd_z = Sfd_z_lpi + Sfd_z_rpi;
						FluxpiBfu_x = Bfu_x_lpi + Bfu_x_rpi;
						FluxpiBfu_y = Bfu_y_lpi + Bfu_y_rpi;
						FluxpiBfu_z = Bfu_z_lpi + Bfu_z_rpi;
						FluxpiDf = Df_lpi + Df_rpi;
						FluxpiDYf = DYf_lpi + DYf_rpi;
						Fluxpitauf = tauf_lpi + tauf_rpi;
						Fluxpiphif = phif_lpi + phif_rpi;
						FluxpjSfd_x = Sfd_x_lpj + Sfd_x_rpj;
						FluxpjSfd_y = Sfd_y_lpj + Sfd_y_rpj;
						FluxpjSfd_z = Sfd_z_lpj + Sfd_z_rpj;
						FluxpjBfu_x = Bfu_x_lpj + Bfu_x_rpj;
						FluxpjBfu_y = Bfu_y_lpj + Bfu_y_rpj;
						FluxpjBfu_z = Bfu_z_lpj + Bfu_z_rpj;
						FluxpjDf = Df_lpj + Df_rpj;
						FluxpjDYf = DYf_lpj + DYf_rpj;
						Fluxpjtauf = tauf_lpj + tauf_rpj;
						Fluxpjphif = phif_lpj + phif_rpj;
						FluxpkSfd_x = Sfd_x_lpk + Sfd_x_rpk;
						FluxpkSfd_y = Sfd_y_lpk + Sfd_y_rpk;
						FluxpkSfd_z = Sfd_z_lpk + Sfd_z_rpk;
						FluxpkBfu_x = Bfu_x_lpk + Bfu_x_rpk;
						FluxpkBfu_y = Bfu_y_lpk + Bfu_y_rpk;
						FluxpkBfu_z = Bfu_z_lpk + Bfu_z_rpk;
						FluxpkDf = Df_lpk + Df_rpk;
						FluxpkDYf = DYf_lpk + DYf_rpk;
						Fluxpktauf = tauf_lpk + tauf_rpk;
						Fluxpkphif = phif_lpk + phif_rpk;
						d_div_Beta_o0_t0_m0_l0 = D1CDO4_i(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o1_t0_m0_l0 = lieforward_i(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o2_t0_m0_l0 = liebackward_i(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_y_o0_t7_m0_l0 = D1CDO4_j(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o1_t1_m0_l0 = lieforward_j(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o2_t1_m0_l0 = liebackward_j(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_z_o0_t7_m0_l0 = D1CDO4_k(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o1_t2_m0_l0 = lieforward_k(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o2_t2_m0_l0 = liebackward_k(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_x_o0_t0_m0_l0 = D2CDO4_i(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_y_o0_t0_m0_l0 = D1CDO4crossed_ij(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_z_o0_t0_m0_l0 = D1CDO4crossed_ik(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t15_m0_l0 = D2CDO4_j(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t16_m0_l0 = D1CDO4crossed_jk(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t17_m0_l0 = D2CDO4_k(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_x_o0_t8_m0_l0 = D1CDO4_i(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o1_t0_m0_l0 = lieforward_i(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o2_t0_m0_l0 = liebackward_i(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_div_Beta_o0_t1_m0_l0 = D1CDO4_j(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o1_t1_m0_l0 = lieforward_j(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o2_t1_m0_l0 = liebackward_j(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_z_o0_t8_m0_l0 = D1CDO4_k(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o1_t2_m0_l0 = lieforward_k(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o2_t2_m0_l0 = liebackward_k(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o0_t6_m0_l0 = D2CDO4_i(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_x_o0_t1_m0_l0 = D1CDO4crossed_ij(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o0_t8_m0_l0 = D1CDO4crossed_ik(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_y_o0_t1_m0_l0 = D2CDO4_j(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_z_o0_t1_m0_l0 = D1CDO4crossed_jk(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o0_t11_m0_l0 = D2CDO4_k(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_x_o0_t9_m0_l0 = D1CDO4_i(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o1_t0_m0_l0 = lieforward_i(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o2_t0_m0_l0 = liebackward_i(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_y_o0_t9_m0_l0 = D1CDO4_j(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o1_t1_m0_l0 = lieforward_j(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o2_t1_m0_l0 = liebackward_j(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_div_Beta_o0_t2_m0_l0 = D1CDO4_k(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o1_t2_m0_l0 = lieforward_k(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o2_t2_m0_l0 = liebackward_k(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o0_t0_m0_l0 = D2CDO4_i(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o0_t1_m0_l0 = D1CDO4crossed_ij(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_x_o0_t2_m0_l0 = D1CDO4crossed_ik(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o0_t3_m0_l0 = D2CDO4_j(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_y_o0_t2_m0_l0 = D1CDO4crossed_jk(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_z_o0_t2_m0_l0 = D2CDO4_k(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xxx_o0_t0_m0_l0 = D1CDO4_i(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o1_t0_m0_l0 = lieforward_i(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o2_t0_m0_l0 = liebackward_i(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xxy_o0_t0_m0_l0 = D1CDO4_j(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o1_t1_m0_l0 = lieforward_j(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o2_t1_m0_l0 = liebackward_j(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xxz_o0_t0_m0_l0 = D1CDO4_k(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o1_t2_m0_l0 = lieforward_k(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o2_t2_m0_l0 = liebackward_k(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yxx_o0_t0_m0_l0 = D1CDO4_i(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o1_t0_m0_l0 = lieforward_i(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o2_t0_m0_l0 = liebackward_i(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyy_o0_t0_m0_l0 = D1CDO4_j(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o1_t1_m0_l0 = lieforward_j(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o2_t1_m0_l0 = liebackward_j(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyz_o0_t1_m0_l0 = D1CDO4_k(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o1_t2_m0_l0 = lieforward_k(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o2_t2_m0_l0 = liebackward_k(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyy_o0_t1_m0_l0 = D1CDO4_i(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o1_t0_m0_l0 = lieforward_i(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o2_t0_m0_l0 = liebackward_i(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yyy_o0_t0_m0_l0 = D1CDO4_j(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o1_t1_m0_l0 = lieforward_j(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o2_t1_m0_l0 = liebackward_j(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yyz_o0_t0_m0_l0 = D1CDO4_k(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o1_t2_m0_l0 = lieforward_k(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o2_t2_m0_l0 = liebackward_k(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_zxx_o0_t0_m0_l0 = D1CDO4_i(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o1_t0_m0_l0 = lieforward_i(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o2_t0_m0_l0 = liebackward_i(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyz_o0_t0_m0_l0 = D1CDO4_j(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o1_t1_m0_l0 = lieforward_j(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o2_t1_m0_l0 = liebackward_j(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xzz_o0_t0_m0_l0 = D1CDO4_k(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o1_t2_m0_l0 = lieforward_k(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o2_t2_m0_l0 = liebackward_k(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyz_o0_t2_m0_l0 = D1CDO4_i(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o1_t0_m0_l0 = lieforward_i(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o2_t0_m0_l0 = liebackward_i(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_zyy_o0_t0_m0_l0 = D1CDO4_j(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o1_t1_m0_l0 = lieforward_j(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o2_t1_m0_l0 = liebackward_j(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yzz_o0_t0_m0_l0 = D1CDO4_k(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o1_t2_m0_l0 = lieforward_k(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o2_t2_m0_l0 = liebackward_k(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xzz_o0_t1_m0_l0 = D1CDO4_i(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o1_t0_m0_l0 = lieforward_i(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o2_t0_m0_l0 = liebackward_i(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yzz_o0_t1_m0_l0 = D1CDO4_j(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o1_t1_m0_l0 = lieforward_j(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o2_t1_m0_l0 = liebackward_j(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_zzz_o0_t0_m0_l0 = D1CDO4_k(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o1_t2_m0_l0 = lieforward_k(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o2_t2_m0_l0 = liebackward_k(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t0_m0_l0 = D1CDO4_i(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o1_t0_m0_l0 = lieforward_i(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o2_t0_m0_l0 = liebackward_i(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t1_m0_l0 = D1CDO4_j(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o1_t1_m0_l0 = lieforward_j(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o2_t1_m0_l0 = liebackward_j(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t2_m0_l0 = D1CDO4_k(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o1_t2_m0_l0 = lieforward_k(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o2_t2_m0_l0 = liebackward_k(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t6_m0_l0 = D1CDO4_i(rk2Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o1_t0_m0_l0 = liebackward_i(rk2Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o2_t0_m0_l0 = lieforward_i(rk2Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t6_m0_l0 = D1CDO4_j(rk2Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o1_t1_m0_l0 = liebackward_j(rk2Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o2_t1_m0_l0 = lieforward_j(rk2Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t6_m0_l0 = D1CDO4_k(rk2Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o1_t2_m0_l0 = liebackward_k(rk2Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o2_t2_m0_l0 = lieforward_k(rk2Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t7_m0_l0 = D1CDO4_i(rk2Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o1_t0_m0_l0 = liebackward_i(rk2Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o2_t0_m0_l0 = lieforward_i(rk2Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t7_m0_l0 = D1CDO4_j(rk2Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o1_t1_m0_l0 = liebackward_j(rk2Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o2_t1_m0_l0 = lieforward_j(rk2Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t7_m0_l0 = D1CDO4_k(rk2Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o1_t2_m0_l0 = liebackward_k(rk2Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o2_t2_m0_l0 = lieforward_k(rk2Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t8_m0_l0 = D1CDO4_i(rk2Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o1_t0_m0_l0 = liebackward_i(rk2Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o2_t0_m0_l0 = lieforward_i(rk2Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t8_m0_l0 = D1CDO4_j(rk2Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o1_t1_m0_l0 = liebackward_j(rk2Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o2_t1_m0_l0 = lieforward_j(rk2Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t8_m0_l0 = D1CDO4_k(rk2Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o1_t2_m0_l0 = liebackward_k(rk2Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o2_t2_m0_l0 = lieforward_k(rk2Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t1_m0_l0 = D1CDO4_i(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o1_t0_m0_l0 = lieforward_i(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o2_t0_m0_l0 = liebackward_i(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t2_m0_l0 = D1CDO4_j(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o1_t1_m0_l0 = lieforward_j(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o2_t1_m0_l0 = liebackward_j(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t3_m0_l0 = D1CDO4_k(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o1_t2_m0_l0 = lieforward_k(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o2_t2_m0_l0 = liebackward_k(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t0_m0_l0 = D2CDO4_i(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xy_o0_t0_m0_l0 = D1CDO4crossed_ij(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xz_o0_t0_m0_l0 = D1CDO4crossed_ik(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_yy_o0_t0_m0_l0 = D2CDO4_j(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_yz_o0_t0_m0_l0 = D1CDO4crossed_jk(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_zz_o0_t0_m0_l0 = D2CDO4_k(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o1_t0_m0_l0 = lieforward_i(rk2Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o2_t0_m0_l0 = liebackward_i(rk2Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o1_t1_m0_l0 = lieforward_j(rk2Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o2_t1_m0_l0 = liebackward_j(rk2Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o1_t2_m0_l0 = lieforward_k(rk2Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o2_t2_m0_l0 = liebackward_k(rk2Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o1_t0_m0_l0 = lieforward_i(rk2Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o2_t0_m0_l0 = liebackward_i(rk2Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o1_t1_m0_l0 = lieforward_j(rk2Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o2_t1_m0_l0 = liebackward_j(rk2Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o1_t2_m0_l0 = lieforward_k(rk2Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o2_t2_m0_l0 = liebackward_k(rk2Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o1_t0_m0_l0 = lieforward_i(rk2Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o2_t0_m0_l0 = liebackward_i(rk2Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o1_t1_m0_l0 = lieforward_j(rk2Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o2_t1_m0_l0 = liebackward_j(rk2Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o1_t2_m0_l0 = lieforward_k(rk2Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o2_t2_m0_l0 = liebackward_k(rk2Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o1_t0_m0_l0 = lieforward_i(rk2Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o2_t0_m0_l0 = liebackward_i(rk2Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o1_t1_m0_l0 = lieforward_j(rk2Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o2_t1_m0_l0 = liebackward_j(rk2Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o1_t2_m0_l0 = lieforward_k(rk2Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o2_t2_m0_l0 = liebackward_k(rk2Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o1_t0_m0_l0 = lieforward_i(rk2Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o2_t0_m0_l0 = liebackward_i(rk2Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o1_t1_m0_l0 = lieforward_j(rk2Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o2_t1_m0_l0 = liebackward_j(rk2Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o1_t2_m0_l0 = lieforward_k(rk2Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o2_t2_m0_l0 = liebackward_k(rk2Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o1_t0_m0_l0 = lieforward_i(rk2Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o2_t0_m0_l0 = liebackward_i(rk2Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o1_t1_m0_l0 = lieforward_j(rk2Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o2_t1_m0_l0 = liebackward_j(rk2Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o1_t2_m0_l0 = lieforward_k(rk2Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o2_t2_m0_l0 = liebackward_k(rk2Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t0_m0_l0 = D1CDO4_i(rk2trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o1_t0_m0_l0 = liebackward_i(rk2trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o2_t0_m0_l0 = lieforward_i(rk2trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t1_m0_l0 = D1CDO4_j(rk2trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o1_t1_m0_l0 = liebackward_j(rk2trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o2_t1_m0_l0 = lieforward_j(rk2trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t2_m0_l0 = D1CDO4_k(rk2trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o1_t2_m0_l0 = liebackward_k(rk2trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o2_t2_m0_l0 = lieforward_k(rk2trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t3_m0_l0 = D1CDO4_i(rk2theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o1_t0_m0_l0 = lieforward_i(rk2theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o2_t0_m0_l0 = liebackward_i(rk2theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t4_m0_l0 = D1CDO4_j(rk2theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o1_t1_m0_l0 = lieforward_j(rk2theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o2_t1_m0_l0 = liebackward_j(rk2theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t5_m0_l0 = D1CDO4_k(rk2theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o1_t2_m0_l0 = lieforward_k(rk2theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o2_t2_m0_l0 = liebackward_k(rk2theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t0_m0_l0 = D2CDO4_i(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t1_m0_l0 = D1CDO4crossed_ij(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t2_m0_l0 = D1CDO4crossed_ik(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t3_m0_l0 = D2CDO4_j(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t4_m0_l0 = D1CDO4crossed_jk(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t5_m0_l0 = D2CDO4_k(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t0_m0_l0 = D2CDO4_i(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t1_m0_l0 = D1CDO4crossed_ij(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t2_m0_l0 = D1CDO4crossed_ik(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t3_m0_l0 = D2CDO4_j(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t4_m0_l0 = D1CDO4crossed_jk(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t5_m0_l0 = D2CDO4_k(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t0_m0_l0 = D2CDO4_i(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t1_m0_l0 = D1CDO4crossed_ij(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t2_m0_l0 = D1CDO4crossed_ik(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t3_m0_l0 = D2CDO4_j(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t4_m0_l0 = D1CDO4crossed_jk(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t5_m0_l0 = D2CDO4_k(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t0_m0_l0 = D2CDO4_i(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t1_m0_l0 = D1CDO4crossed_ij(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t2_m0_l0 = D1CDO4crossed_ik(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t3_m0_l0 = D2CDO4_j(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t4_m0_l0 = D1CDO4crossed_jk(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t5_m0_l0 = D2CDO4_k(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t0_m0_l0 = D2CDO4_i(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t1_m0_l0 = D1CDO4crossed_ij(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t2_m0_l0 = D1CDO4crossed_ik(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t3_m0_l0 = D2CDO4_j(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t4_m0_l0 = D1CDO4crossed_jk(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t5_m0_l0 = D2CDO4_k(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t0_m0_l0 = D2CDO4_i(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t1_m0_l0 = D1CDO4crossed_ij(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t2_m0_l0 = D1CDO4crossed_ik(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t3_m0_l0 = D2CDO4_j(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t4_m0_l0 = D1CDO4crossed_jk(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t5_m0_l0 = D2CDO4_k(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t3_m0_l0 = D2CDO4_i(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t4_m0_l0 = D1CDO4crossed_ij(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t5_m0_l0 = D1CDO4crossed_ik(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t6_m0_l0 = D2CDO4_j(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t7_m0_l0 = D1CDO4crossed_jk(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t8_m0_l0 = D2CDO4_k(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						rk2m_Ctd_yyz_o0_t0_l0 = 0.5 * d_Ctd_yyz_o0_t0_m0_l0;
						rk2Ctd_yyz = rk2m_Ctd_yyz_o0_t0_l0;
						rk2m_Ctd_yyy_o0_t0_l0 = 0.5 * d_Ctd_yyy_o0_t0_m0_l0;
						rk2Ctd_yyy = rk2m_Ctd_yyy_o0_t0_l0;
						rk2div_Beta = (d_div_Beta_o0_t0_m0_l0 + d_div_Beta_o0_t1_m0_l0) + d_div_Beta_o0_t2_m0_l0;
						rk2m_Ctd_xxz_o0_t0_l0 = 0.5 * d_Ctd_xxz_o0_t0_m0_l0;
						rk2Ctd_xxz = rk2m_Ctd_xxz_o0_t0_l0;
						rk2m_Ctd_xxx_o0_t0_l0 = 0.5 * d_Ctd_xxx_o0_t0_m0_l0;
						rk2Ctd_xxx = rk2m_Ctd_xxx_o0_t0_l0;
						rk2m_Ctd_yzz_o0_t1_l0 = (-0.5) * d_Ctd_yzz_o0_t1_m0_l0;
						rk2Ctd_yzz = d_Ctd_yzz_o0_t0_m0_l0 + rk2m_Ctd_yzz_o0_t1_l0;
						rk2m_Ctd_xxy_o0_t0_l0 = 0.5 * d_Ctd_xxy_o0_t0_m0_l0;
						rk2Ctd_xxy = rk2m_Ctd_xxy_o0_t0_l0;
						rk2kappa_z2 = p_kappa_z2;
						rk2m_Ctd_xyy_o0_t1_l0 = (-0.5) * d_Ctd_xyy_o0_t1_m0_l0;
						rk2Ctd_xyy = d_Ctd_xyy_o0_t0_m0_l0 + rk2m_Ctd_xyy_o0_t1_l0;
						rk2m_Ctd_xyz_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t2_m0_l0;
						rk2m_Ctd_xyz_o0_t1_l0 = 0.5 * d_Ctd_xyz_o0_t1_m0_l0;
						rk2m_Ctd_xyz_o0_t0_l0 = 0.5 * d_Ctd_xyz_o0_t0_m0_l0;
						rk2Ctd_xyz = (rk2m_Ctd_xyz_o0_t0_l0 + rk2m_Ctd_xyz_o0_t1_l0) + rk2m_Ctd_xyz_o0_t2_l0;
						rk2d_div_Beta_y = (d_d_div_Beta_y_o0_t0_m0_l0 + d_d_div_Beta_y_o0_t1_m0_l0) + d_d_div_Beta_y_o0_t2_m0_l0;
						rk2d_div_Beta_x = (d_d_div_Beta_x_o0_t0_m0_l0 + d_d_div_Beta_x_o0_t1_m0_l0) + d_d_div_Beta_x_o0_t2_m0_l0;
						rk2d_div_Beta_z = (d_d_div_Beta_z_o0_t0_m0_l0 + d_d_div_Beta_z_o0_t1_m0_l0) + d_d_div_Beta_z_o0_t2_m0_l0;
						rk2m_Ctd_xzz_o0_t1_l0 = (-0.5) * d_Ctd_xzz_o0_t1_m0_l0;
						rk2Ctd_xzz = d_Ctd_xzz_o0_t0_m0_l0 + rk2m_Ctd_xzz_o0_t1_l0;
						rk2m_Ctd_zxz_o0_t0_l0 = 0.5 * d_Ctd_xzz_o0_t1_m0_l0;
						rk2Ctd_zxz = rk2m_Ctd_zxz_o0_t0_l0;
						rk2m_Ctd_zxx_o0_t1_l0 = (-0.5) * d_Ctd_xxz_o0_t0_m0_l0;
						rk2Ctd_zxx = d_Ctd_zxx_o0_t0_m0_l0 + rk2m_Ctd_zxx_o0_t1_l0;
						rk2detgtd = vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) * vector(rk2gtd_xx, i, j, k) + (-vector(rk2gtd_yy, i, j, k) * (vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k))) + (-vector(rk2gtd_zz, i, j, k) * (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k))) + (-(vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k)) * vector(rk2gtd_xx, i, j, k)) + 2.0 * vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k);
						rk2chi_max = MAX(chi_floor, vector(rk2chi, i, j, k));
						rk2inv_chi = 1.0 / fabs(rk2chi_max);
						rk2Bfd_y = rk2inv_chi * (vector(rk2gtd_xy, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_yy, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2Bfd_z = rk2inv_chi * (vector(rk2gtd_xz, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_yz, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_zz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2Bfd_x = rk2inv_chi * (vector(rk2gtd_xx, i, j, k) * vector(rk2Bfu_x, i, j, k) + vector(rk2gtd_xy, i, j, k) * vector(rk2Bfu_y, i, j, k) + vector(rk2gtd_xz, i, j, k) * vector(rk2Bfu_z, i, j, k));
						rk2sqBf = vector(rk2Bfu_x, i, j, k) * rk2Bfd_x + vector(rk2Bfu_y, i, j, k) * rk2Bfd_y + vector(rk2Bfu_z, i, j, k) * rk2Bfd_z;
						rk2m_Ctd_zyy_o0_t1_l0 = (-0.5) * d_Ctd_yyz_o0_t0_m0_l0;
						rk2Ctd_zyy = d_Ctd_zyy_o0_t0_m0_l0 + rk2m_Ctd_zyy_o0_t1_l0;
						rk2m_Ctd_zyz_o0_t0_l0 = 0.5 * d_Ctd_yzz_o0_t1_m0_l0;
						rk2Ctd_zyz = rk2m_Ctd_zyz_o0_t0_l0;
						rk2sdetg = pow(rk2chi_max, (-1.500000000000000));
						rk2Tf4u_tt = (vector(rk2Df, i, j, k) + vector(rk2tauf, i, j, k)) / (rk2sdetg * (vector(rk2Alpha, i, j, k) * vector(rk2Alpha, i, j, k)));
						rk2Betatd_x = vector(rk2Betau_x, i, j, k) * vector(rk2gtd_xx, i, j, k) + vector(rk2Betau_y, i, j, k) * vector(rk2gtd_xy, i, j, k) + vector(rk2Betau_z, i, j, k) * vector(rk2gtd_xz, i, j, k);
						rk2Betatd_z = vector(rk2Betau_x, i, j, k) * vector(rk2gtd_xz, i, j, k) + vector(rk2Betau_y, i, j, k) * vector(rk2gtd_yz, i, j, k) + vector(rk2Betau_z, i, j, k) * vector(rk2gtd_zz, i, j, k);
						rk2Betatd_y = vector(rk2Betau_x, i, j, k) * vector(rk2gtd_xy, i, j, k) + vector(rk2Betau_y, i, j, k) * vector(rk2gtd_yy, i, j, k) + vector(rk2Betau_z, i, j, k) * vector(rk2gtd_yz, i, j, k);
						rk2idetgtd = 1.0 / rk2detgtd;
						rk2gtu_zz = rk2idetgtd * (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yy, i, j, k) - vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k));
						rk2gtu_yy = rk2idetgtd * (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_zz, i, j, k) - vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2gtu_yz = rk2idetgtd * ((-vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yz, i, j, k)) + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2gtu_xx = rk2idetgtd * (vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) - vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k));
						rk2gtu_xz = rk2idetgtd * (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_yz, i, j, k) - vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2Ct_zyz = rk2gtu_xz * rk2Ctd_xyz + rk2gtu_yz * rk2Ctd_yyz + rk2gtu_zz * rk2Ctd_zyz;
						rk2Ct_zyy = rk2gtu_xz * rk2Ctd_xyy + rk2gtu_yz * rk2Ctd_yyy + rk2gtu_zz * rk2Ctd_zyy;
						rk2Atud_zx = rk2gtu_xz * vector(rk2Atd_xx, i, j, k) + rk2gtu_yz * vector(rk2Atd_xy, i, j, k) + rk2gtu_zz * vector(rk2Atd_xz, i, j, k);
						rk2Atud_zz = rk2gtu_xz * vector(rk2Atd_xz, i, j, k) + rk2gtu_yz * vector(rk2Atd_yz, i, j, k) + rk2gtu_zz * vector(rk2Atd_zz, i, j, k);
						rk2Atud_zy = rk2gtu_xz * vector(rk2Atd_xy, i, j, k) + rk2gtu_yz * vector(rk2Atd_yy, i, j, k) + rk2gtu_zz * vector(rk2Atd_yz, i, j, k);
						rk2Atu_zz = rk2Atud_zx * rk2gtu_xz + rk2Atud_zy * rk2gtu_yz + rk2Atud_zz * rk2gtu_zz;
						rk2vfu_z = rk2chi_max * (rk2gtu_xz * vector(vfd_x, i, j, k) + rk2gtu_yz * vector(vfd_y, i, j, k) + rk2gtu_zz * vector(vfd_z, i, j, k));
						rk2Sfu_z = rk2chi_max * (rk2gtu_xz * vector(rk2Sfd_x, i, j, k) + rk2gtu_yz * vector(rk2Sfd_y, i, j, k) + rk2gtu_zz * vector(rk2Sfd_z, i, j, k));
						rk2Tf4u_tz = rk2Sfu_z / (rk2sdetg * vector(rk2Alpha, i, j, k)) + (-vector(rk2Betau_z, i, j, k) * rk2Tf4u_tt);
						rk2Tu_tz = rk2Tf4u_tz;
						rk2gtu_xy = rk2idetgtd * ((-vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_zz, i, j, k)) + vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xz, i, j, k));
						rk2vfu_y = rk2chi_max * (rk2gtu_xy * vector(vfd_x, i, j, k) + rk2gtu_yy * vector(vfd_y, i, j, k) + rk2gtu_yz * vector(vfd_z, i, j, k));
						rk2vfu_x = rk2chi_max * (rk2gtu_xx * vector(vfd_x, i, j, k) + rk2gtu_xy * vector(vfd_y, i, j, k) + rk2gtu_xz * vector(vfd_z, i, j, k));
						rk2sqvf = MIN(rk2vfu_x * vector(vfd_x, i, j, k) + rk2vfu_y * vector(vfd_y, i, j, k) + rk2vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk2sqW = 1.0 / (1.0 - rk2sqvf);
						rk2Atud_yy = rk2gtu_xy * vector(rk2Atd_xy, i, j, k) + rk2gtu_yy * vector(rk2Atd_yy, i, j, k) + rk2gtu_yz * vector(rk2Atd_yz, i, j, k);
						rk2Atud_yx = rk2gtu_xy * vector(rk2Atd_xx, i, j, k) + rk2gtu_yy * vector(rk2Atd_xy, i, j, k) + rk2gtu_yz * vector(rk2Atd_xz, i, j, k);
						rk2Atud_yz = rk2gtu_xy * vector(rk2Atd_xz, i, j, k) + rk2gtu_yy * vector(rk2Atd_yz, i, j, k) + rk2gtu_yz * vector(rk2Atd_zz, i, j, k);
						rk2Atu_yz = rk2Atud_yx * rk2gtu_xz + rk2Atud_yy * rk2gtu_yz + rk2Atud_yz * rk2gtu_zz;
						rk2Atu_yy = rk2Atud_yx * rk2gtu_xy + rk2Atud_yy * rk2gtu_yy + rk2Atud_yz * rk2gtu_yz;
						rk2Ct_yyy = rk2gtu_xy * rk2Ctd_xyy + rk2gtu_yy * rk2Ctd_yyy + rk2gtu_yz * rk2Ctd_zyy;
						rk2Ct_yyz = rk2gtu_xy * rk2Ctd_xyz + rk2gtu_yy * rk2Ctd_yyz + rk2gtu_yz * rk2Ctd_zyz;
						rk2Atud_xx = rk2gtu_xx * vector(rk2Atd_xx, i, j, k) + rk2gtu_xy * vector(rk2Atd_xy, i, j, k) + rk2gtu_xz * vector(rk2Atd_xz, i, j, k);
						rk2trAt = rk2Atud_xx + rk2Atud_yy + rk2Atud_zz;
						rk2Atud_xz = rk2gtu_xx * vector(rk2Atd_xz, i, j, k) + rk2gtu_xy * vector(rk2Atd_yz, i, j, k) + rk2gtu_xz * vector(rk2Atd_zz, i, j, k);
						rk2Atud_xy = rk2gtu_xx * vector(rk2Atd_xy, i, j, k) + rk2gtu_xy * vector(rk2Atd_yy, i, j, k) + rk2gtu_xz * vector(rk2Atd_yz, i, j, k);
						rk2Atu_xz = rk2Atud_xx * rk2gtu_xz + rk2Atud_xy * rk2gtu_yz + rk2Atud_xz * rk2gtu_zz;
						rk2Atu_xy = rk2Atud_xx * rk2gtu_xy + rk2Atud_xy * rk2gtu_yy + rk2Atud_xz * rk2gtu_yz;
						rk2Atu_xx = rk2Atud_xx * rk2gtu_xx + rk2Atud_xy * rk2gtu_xy + rk2Atud_xz * rk2gtu_xz;
						rk2Sfu_y = rk2chi_max * (rk2gtu_xy * vector(rk2Sfd_x, i, j, k) + rk2gtu_yy * vector(rk2Sfd_y, i, j, k) + rk2gtu_yz * vector(rk2Sfd_z, i, j, k));
						rk2Tf4u_ty = rk2Sfu_y / (rk2sdetg * vector(rk2Alpha, i, j, k)) + (-vector(rk2Betau_y, i, j, k) * rk2Tf4u_tt);
						rk2Tu_ty = rk2Tf4u_ty;
						rk2Sfu_x = rk2chi_max * (rk2gtu_xx * vector(rk2Sfd_x, i, j, k) + rk2gtu_xy * vector(rk2Sfd_y, i, j, k) + rk2gtu_xz * vector(rk2Sfd_z, i, j, k));
						rk2Tf4u_tx = rk2Sfu_x / (rk2sdetg * vector(rk2Alpha, i, j, k)) + (-vector(rk2Betau_x, i, j, k) * rk2Tf4u_tt);
						rk2Tu_tx = rk2Tf4u_tx;
						rk2decay_factor = MIN(1.0, pow((R_0 / MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))), eta_damping_exp));
						rk2kappa_z1 = MIN(p_kappa_z1, p_kappa_z1 * rk2decay_factor) / vector(rk2Alpha, i, j, k);
						rk2kappa_f = MIN(p_kappa_f, p_kappa_f * rk2decay_factor);
						rk2feta = MIN(p_feta, p_feta * rk2decay_factor);
						rk2m_Ctd_zzz_o0_t0_l0 = 0.5 * d_Ctd_zzz_o0_t0_m0_l0;
						rk2Ctd_zzz = rk2m_Ctd_zzz_o0_t0_l0;
						rk2Ct_xzz = rk2gtu_xx * rk2Ctd_xzz + rk2gtu_xy * rk2Ctd_yzz + rk2gtu_xz * rk2Ctd_zzz;
						rk2Ct_zzz = rk2gtu_xz * rk2Ctd_xzz + rk2gtu_yz * rk2Ctd_yzz + rk2gtu_zz * rk2Ctd_zzz;
						rk2Ct_yzz = rk2gtu_xy * rk2Ctd_xzz + rk2gtu_yy * rk2Ctd_yzz + rk2gtu_yz * rk2Ctd_zzz;
						rk2m_Ctd_yxy_o0_t0_l0 = 0.5 * d_Ctd_xyy_o0_t1_m0_l0;
						rk2Ctd_yxy = rk2m_Ctd_yxy_o0_t0_l0;
						rk2Ct_xyz = rk2gtu_xx * rk2Ctd_xyz + rk2gtu_xy * rk2Ctd_yyz + rk2gtu_xz * rk2Ctd_zyz;
						rk2m_Ctd_yxz_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t0_m0_l0;
						rk2m_Ctd_yxz_o0_t0_l0 = 0.5 * d_Ctd_xyz_o0_t2_m0_l0;
						rk2Ctd_yxz = (rk2m_Ctd_yxz_o0_t0_l0 + rk2m_Ctd_xyz_o0_t1_l0) + rk2m_Ctd_yxz_o0_t2_l0;
						rk2Ct_zxz = rk2gtu_xz * rk2Ctd_xxz + rk2gtu_yz * rk2Ctd_yxz + rk2gtu_zz * rk2Ctd_zxz;
						rk2Ct_yxz = rk2gtu_xy * rk2Ctd_xxz + rk2gtu_yy * rk2Ctd_yxz + rk2gtu_yz * rk2Ctd_zxz;
						rk2m_Ctd_zxy_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t1_m0_l0;
						rk2Ctd_zxy = (rk2m_Ctd_yxz_o0_t0_l0 + rk2m_Ctd_xyz_o0_t0_l0) + rk2m_Ctd_zxy_o0_t2_l0;
						rk2Ct_zxy = rk2gtu_xz * rk2Ctd_xxy + rk2gtu_yz * rk2Ctd_yxy + rk2gtu_zz * rk2Ctd_zxy;
						rk2Ct_yxy = rk2gtu_xy * rk2Ctd_xxy + rk2gtu_yy * rk2Ctd_yxy + rk2gtu_yz * rk2Ctd_zxy;
						rk2Ct_xxy = rk2gtu_xx * rk2Ctd_xxy + rk2gtu_xy * rk2Ctd_yxy + rk2gtu_xz * rk2Ctd_zxy;
						rk2Ct_xxz = rk2gtu_xx * rk2Ctd_xxz + rk2gtu_xy * rk2Ctd_yxz + rk2gtu_xz * rk2Ctd_zxz;
						rk2Ct_xyy = rk2gtu_xx * rk2Ctd_xyy + rk2gtu_xy * rk2Ctd_yyy + rk2gtu_xz * rk2Ctd_zyy;
						rk2m_Ctd_yxx_o0_t1_l0 = (-0.5) * d_Ctd_xxy_o0_t0_m0_l0;
						rk2Ctd_yxx = d_Ctd_yxx_o0_t0_m0_l0 + rk2m_Ctd_yxx_o0_t1_l0;
						rk2Ct_zxx = rk2gtu_xz * rk2Ctd_xxx + rk2gtu_yz * rk2Ctd_yxx + rk2gtu_zz * rk2Ctd_zxx;
						rk2Gamt_z = rk2gtu_xx * rk2Ct_zxx + 2.0 * rk2gtu_xy * rk2Ct_zxy + 2.0 * rk2gtu_xz * rk2Ct_zxz + rk2gtu_yy * rk2Ct_zyy + 2.0 * rk2gtu_yz * rk2Ct_zyz + rk2gtu_zz * rk2Ct_zzz;
						rk2Zu_z = 0.5 * rk2chi_max * (vector(rk2Gamh_z, i, j, k) - rk2Gamt_z);
						rk2Ct_yxx = rk2gtu_xy * rk2Ctd_xxx + rk2gtu_yy * rk2Ctd_yxx + rk2gtu_yz * rk2Ctd_zxx;
						rk2Gamt_y = rk2gtu_xx * rk2Ct_yxx + 2.0 * rk2gtu_xy * rk2Ct_yxy + 2.0 * rk2gtu_xz * rk2Ct_yxz + rk2gtu_yy * rk2Ct_yyy + 2.0 * rk2gtu_yz * rk2Ct_yyz + rk2gtu_zz * rk2Ct_yzz;
						rk2Zu_y = 0.5 * rk2chi_max * (vector(rk2Gamh_y, i, j, k) - rk2Gamt_y);
						rk2Ct_xxx = rk2gtu_xx * rk2Ctd_xxx + rk2gtu_xy * rk2Ctd_yxx + rk2gtu_xz * rk2Ctd_zxx;
						rk2Gamt_x = rk2gtu_xx * rk2Ct_xxx + 2.0 * rk2gtu_xy * rk2Ct_xxy + 2.0 * rk2gtu_xz * rk2Ct_xxz + rk2gtu_yy * rk2Ct_xyy + 2.0 * rk2gtu_yz * rk2Ct_xyz + rk2gtu_zz * rk2Ct_xzz;
						rk2Zu_x = 0.5 * rk2chi_max * (vector(rk2Gamh_x, i, j, k) - rk2Gamt_x);
						rk2d_Rtd_xx_o0_t9_m0_l0 = rk2gtu_zz * (3.0 * rk2Ct_xxz * rk2Ctd_xxz + 2.0 * rk2Ct_yxz * rk2Ctd_xyz + rk2Ct_yxz * rk2Ctd_yxz + 2.0 * rk2Ct_zxz * rk2Ctd_xzz + rk2Ct_zxz * rk2Ctd_zxz) + rk2gtu_yz * (3.0 * rk2Ct_xxy * rk2Ctd_xxz + 2.0 * rk2Ct_yxy * rk2Ctd_xyz + rk2Ct_yxy * rk2Ctd_yxz + 2.0 * rk2Ct_zxy * rk2Ctd_xzz + rk2Ct_zxy * rk2Ctd_zxz) + rk2gtu_xz * (3.0 * rk2Ct_xxx * rk2Ctd_xxz + 2.0 * rk2Ct_yxx * rk2Ctd_xyz + rk2Ct_yxx * rk2Ctd_yxz + 2.0 * rk2Ct_zxx * rk2Ctd_xzz + rk2Ct_zxx * rk2Ctd_zxz) + rk2gtu_yz * (3.0 * rk2Ct_xxz * rk2Ctd_xxy + 2.0 * rk2Ct_yxz * rk2Ctd_xyy + rk2Ct_yxz * rk2Ctd_yxy + 2.0 * rk2Ct_zxz * rk2Ctd_xyz + rk2Ct_zxz * rk2Ctd_zxy) + rk2gtu_yy * (3.0 * rk2Ct_xxy * rk2Ctd_xxy + 2.0 * rk2Ct_yxy * rk2Ctd_xyy + rk2Ct_yxy * rk2Ctd_yxy + 2.0 * rk2Ct_zxy * rk2Ctd_xyz + rk2Ct_zxy * rk2Ctd_zxy) + rk2gtu_xy * (3.0 * rk2Ct_xxx * rk2Ctd_xxy + 2.0 * rk2Ct_yxx * rk2Ctd_xyy + rk2Ct_yxx * rk2Ctd_yxy + 2.0 * rk2Ct_zxx * rk2Ctd_xyz + rk2Ct_zxx * rk2Ctd_zxy) + rk2gtu_xz * (3.0 * rk2Ct_xxz * rk2Ctd_xxx + 2.0 * rk2Ct_yxz * rk2Ctd_xxy + rk2Ct_yxz * rk2Ctd_yxx + 2.0 * rk2Ct_zxz * rk2Ctd_xxz + rk2Ct_zxz * rk2Ctd_zxx) + rk2gtu_xy * (3.0 * rk2Ct_xxy * rk2Ctd_xxx + 2.0 * rk2Ct_yxy * rk2Ctd_xxy + rk2Ct_yxy * rk2Ctd_yxx + 2.0 * rk2Ct_zxy * rk2Ctd_xxz + rk2Ct_zxy * rk2Ctd_zxx) + rk2gtu_xx * (3.0 * rk2Ct_xxx * rk2Ctd_xxx + 2.0 * rk2Ct_yxx * rk2Ctd_xxy + rk2Ct_yxx * rk2Ctd_yxx + 2.0 * rk2Ct_zxx * rk2Ctd_xxz + rk2Ct_zxx * rk2Ctd_zxx) + vector(rk2Gamh_z, i, j, k) * rk2Ctd_xxz + vector(rk2Gamh_y, i, j, k) * rk2Ctd_xxy + vector(rk2Gamh_x, i, j, k) * rk2Ctd_xxx;
						rk2m_Rtd_xx_o0_t8_l0 = vector(rk2gtd_xz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
						rk2m_Rtd_xx_o0_t7_l0 = vector(rk2gtd_xy, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
						rk2m_Rtd_xx_o0_t6_l0 = vector(rk2gtd_xx, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
						rk2m_Rtd_xx_o0_t5_l0 = (-0.5 * rk2gtu_zz) * d_Rtd_xx_o0_t5_m0_l0;
						rk2m_Rtd_xx_o0_t4_l0 = (-rk2gtu_yz) * d_Rtd_xx_o0_t4_m0_l0;
						rk2m_Rtd_xx_o0_t3_l0 = (-0.5 * rk2gtu_yy) * d_Rtd_xx_o0_t3_m0_l0;
						rk2m_Rtd_xx_o0_t2_l0 = (-rk2gtu_xz) * d_Rtd_xx_o0_t2_m0_l0;
						rk2m_Rtd_xx_o0_t1_l0 = (-rk2gtu_xy) * d_Rtd_xx_o0_t1_m0_l0;
						rk2m_Rtd_xx_o0_t0_l0 = (-0.5 * rk2gtu_xx) * d_Rtd_xx_o0_t0_m0_l0;
						rk2Rtd_xx = ((((((((rk2m_Rtd_xx_o0_t0_l0 + rk2m_Rtd_xx_o0_t1_l0) + rk2m_Rtd_xx_o0_t2_l0) + rk2m_Rtd_xx_o0_t3_l0) + rk2m_Rtd_xx_o0_t4_l0) + rk2m_Rtd_xx_o0_t5_l0) + rk2m_Rtd_xx_o0_t6_l0) + rk2m_Rtd_xx_o0_t7_l0) + rk2m_Rtd_xx_o0_t8_l0) + rk2d_Rtd_xx_o0_t9_m0_l0;
						rk2d_Rtd_xz_o0_t12_m0_l0 = rk2gtu_zz * (rk2Ct_xxz * rk2Ctd_zxz + rk2Ct_xxz * rk2Ctd_xzz + rk2Ct_xzz * rk2Ctd_xxz + rk2Ct_yxz * rk2Ctd_zyz + rk2Ct_yxz * rk2Ctd_yzz + rk2Ct_yzz * rk2Ctd_xyz + 2.0 * rk2Ct_zxz * rk2Ctd_zzz + rk2Ct_zzz * rk2Ctd_xzz) + rk2gtu_yz * (rk2Ct_xxy * rk2Ctd_zxz + rk2Ct_xxy * rk2Ctd_xzz + rk2Ct_xyz * rk2Ctd_xxz + rk2Ct_yxy * rk2Ctd_zyz + rk2Ct_yxy * rk2Ctd_yzz + rk2Ct_yyz * rk2Ctd_xyz + 2.0 * rk2Ct_zxy * rk2Ctd_zzz + rk2Ct_zyz * rk2Ctd_xzz) + rk2gtu_xz * (rk2Ct_xxx * rk2Ctd_zxz + rk2Ct_xxx * rk2Ctd_xzz + rk2Ct_xxz * rk2Ctd_xxz + rk2Ct_yxx * rk2Ctd_zyz + rk2Ct_yxx * rk2Ctd_yzz + rk2Ct_yxz * rk2Ctd_xyz + 2.0 * rk2Ct_zxx * rk2Ctd_zzz + rk2Ct_zxz * rk2Ctd_xzz) + rk2gtu_yz * (rk2Ct_xxz * rk2Ctd_zxy + rk2Ct_xxz * rk2Ctd_xyz + rk2Ct_xzz * rk2Ctd_xxy + rk2Ct_yxz * rk2Ctd_zyy + rk2Ct_yxz * rk2Ctd_yyz + rk2Ct_yzz * rk2Ctd_xyy + 2.0 * rk2Ct_zxz * rk2Ctd_zyz + rk2Ct_zzz * rk2Ctd_xyz) + rk2gtu_yy * (rk2Ct_xxy * rk2Ctd_zxy + rk2Ct_xxy * rk2Ctd_xyz + rk2Ct_xyz * rk2Ctd_xxy + rk2Ct_yxy * rk2Ctd_zyy + rk2Ct_yxy * rk2Ctd_yyz + rk2Ct_yyz * rk2Ctd_xyy + 2.0 * rk2Ct_zxy * rk2Ctd_zyz + rk2Ct_zyz * rk2Ctd_xyz) + rk2gtu_xy * (rk2Ct_xxx * rk2Ctd_zxy + rk2Ct_xxx * rk2Ctd_xyz + rk2Ct_xxz * rk2Ctd_xxy + rk2Ct_yxx * rk2Ctd_zyy + rk2Ct_yxx * rk2Ctd_yyz + rk2Ct_yxz * rk2Ctd_xyy + 2.0 * rk2Ct_zxx * rk2Ctd_zyz + rk2Ct_zxz * rk2Ctd_xyz) + rk2gtu_xz * (rk2Ct_xxz * rk2Ctd_zxx + rk2Ct_xxz * rk2Ctd_xxz + rk2Ct_xzz * rk2Ctd_xxx + rk2Ct_yxz * rk2Ctd_zxy + rk2Ct_yxz * rk2Ctd_yxz + rk2Ct_yzz * rk2Ctd_xxy + 2.0 * rk2Ct_zxz * rk2Ctd_zxz + rk2Ct_zzz * rk2Ctd_xxz) + rk2gtu_xy * (rk2Ct_xxy * rk2Ctd_zxx + rk2Ct_xxy * rk2Ctd_xxz + rk2Ct_xyz * rk2Ctd_xxx + rk2Ct_yxy * rk2Ctd_zxy + rk2Ct_yxy * rk2Ctd_yxz + rk2Ct_yyz * rk2Ctd_xxy + 2.0 * rk2Ct_zxy * rk2Ctd_zxz + rk2Ct_zyz * rk2Ctd_xxz) + rk2gtu_xx * (rk2Ct_xxx * rk2Ctd_zxx + rk2Ct_xxx * rk2Ctd_xxz + rk2Ct_xxz * rk2Ctd_xxx + rk2Ct_yxx * rk2Ctd_zxy + rk2Ct_yxx * rk2Ctd_yxz + rk2Ct_yxz * rk2Ctd_xxy + 2.0 * rk2Ct_zxx * rk2Ctd_zxz + rk2Ct_zxz * rk2Ctd_xxz) + 0.5 * vector(rk2Gamh_z, i, j, k) * (rk2Ctd_xzz + rk2Ctd_zxz) + 0.5 * vector(rk2Gamh_y, i, j, k) * (rk2Ctd_xyz + rk2Ctd_zxy) + 0.5 * vector(rk2Gamh_x, i, j, k) * (rk2Ctd_xxz + rk2Ctd_zxx);
						rk2m_Rtd_xz_o0_t11_l0 = 0.5 * vector(rk2gtd_zz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
						rk2m_Rtd_xz_o0_t10_l0 = 0.5 * vector(rk2gtd_yz, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
						rk2m_Rtd_xz_o0_t9_l0 = 0.5 * vector(rk2gtd_xz, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
						rk2m_Rtd_xz_o0_t8_l0 = 0.5 * vector(rk2gtd_xz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
						rk2m_Rtd_xz_o0_t7_l0 = 0.5 * vector(rk2gtd_xy, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
						rk2m_Rtd_xz_o0_t6_l0 = 0.5 * vector(rk2gtd_xx, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
						rk2m_Rtd_xz_o0_t5_l0 = (-0.5 * rk2gtu_zz) * d_Rtd_xz_o0_t5_m0_l0;
						rk2m_Rtd_xz_o0_t4_l0 = (-rk2gtu_yz) * d_Rtd_xz_o0_t4_m0_l0;
						rk2m_Rtd_xz_o0_t3_l0 = (-0.5 * rk2gtu_yy) * d_Rtd_xz_o0_t3_m0_l0;
						rk2m_Rtd_xz_o0_t2_l0 = (-rk2gtu_xz) * d_Rtd_xz_o0_t2_m0_l0;
						rk2m_Rtd_xz_o0_t1_l0 = (-rk2gtu_xy) * d_Rtd_xz_o0_t1_m0_l0;
						rk2m_Rtd_xz_o0_t0_l0 = (-0.5 * rk2gtu_xx) * d_Rtd_xz_o0_t0_m0_l0;
						rk2Rtd_xz = (((((((((((rk2m_Rtd_xz_o0_t0_l0 + rk2m_Rtd_xz_o0_t1_l0) + rk2m_Rtd_xz_o0_t2_l0) + rk2m_Rtd_xz_o0_t3_l0) + rk2m_Rtd_xz_o0_t4_l0) + rk2m_Rtd_xz_o0_t5_l0) + rk2m_Rtd_xz_o0_t6_l0) + rk2m_Rtd_xz_o0_t7_l0) + rk2m_Rtd_xz_o0_t8_l0) + rk2m_Rtd_xz_o0_t9_l0) + rk2m_Rtd_xz_o0_t10_l0) + rk2m_Rtd_xz_o0_t11_l0) + rk2d_Rtd_xz_o0_t12_m0_l0;
						rk2d_Rtd_zz_o0_t9_m0_l0 = rk2gtu_zz * (2.0 * rk2Ct_xzz * rk2Ctd_zxz + rk2Ct_xzz * rk2Ctd_xzz + 2.0 * rk2Ct_yzz * rk2Ctd_zyz + rk2Ct_yzz * rk2Ctd_yzz + 3.0 * rk2Ct_zzz * rk2Ctd_zzz) + rk2gtu_yz * (2.0 * rk2Ct_xyz * rk2Ctd_zxz + rk2Ct_xyz * rk2Ctd_xzz + 2.0 * rk2Ct_yyz * rk2Ctd_zyz + rk2Ct_yyz * rk2Ctd_yzz + 3.0 * rk2Ct_zyz * rk2Ctd_zzz) + rk2gtu_xz * (2.0 * rk2Ct_xxz * rk2Ctd_zxz + rk2Ct_xxz * rk2Ctd_xzz + 2.0 * rk2Ct_yxz * rk2Ctd_zyz + rk2Ct_yxz * rk2Ctd_yzz + 3.0 * rk2Ct_zxz * rk2Ctd_zzz) + rk2gtu_yz * (2.0 * rk2Ct_xzz * rk2Ctd_zxy + rk2Ct_xzz * rk2Ctd_xyz + 2.0 * rk2Ct_yzz * rk2Ctd_zyy + rk2Ct_yzz * rk2Ctd_yyz + 3.0 * rk2Ct_zzz * rk2Ctd_zyz) + rk2gtu_yy * (2.0 * rk2Ct_xyz * rk2Ctd_zxy + rk2Ct_xyz * rk2Ctd_xyz + 2.0 * rk2Ct_yyz * rk2Ctd_zyy + rk2Ct_yyz * rk2Ctd_yyz + 3.0 * rk2Ct_zyz * rk2Ctd_zyz) + rk2gtu_xy * (2.0 * rk2Ct_xxz * rk2Ctd_zxy + rk2Ct_xxz * rk2Ctd_xyz + 2.0 * rk2Ct_yxz * rk2Ctd_zyy + rk2Ct_yxz * rk2Ctd_yyz + 3.0 * rk2Ct_zxz * rk2Ctd_zyz) + rk2gtu_xz * (2.0 * rk2Ct_xzz * rk2Ctd_zxx + rk2Ct_xzz * rk2Ctd_xxz + 2.0 * rk2Ct_yzz * rk2Ctd_zxy + rk2Ct_yzz * rk2Ctd_yxz + 3.0 * rk2Ct_zzz * rk2Ctd_zxz) + rk2gtu_xy * (2.0 * rk2Ct_xyz * rk2Ctd_zxx + rk2Ct_xyz * rk2Ctd_xxz + 2.0 * rk2Ct_yyz * rk2Ctd_zxy + rk2Ct_yyz * rk2Ctd_yxz + 3.0 * rk2Ct_zyz * rk2Ctd_zxz) + rk2gtu_xx * (2.0 * rk2Ct_xxz * rk2Ctd_zxx + rk2Ct_xxz * rk2Ctd_xxz + 2.0 * rk2Ct_yxz * rk2Ctd_zxy + rk2Ct_yxz * rk2Ctd_yxz + 3.0 * rk2Ct_zxz * rk2Ctd_zxz) + vector(rk2Gamh_z, i, j, k) * rk2Ctd_zzz + vector(rk2Gamh_y, i, j, k) * rk2Ctd_zyz + vector(rk2Gamh_x, i, j, k) * rk2Ctd_zxz;
						rk2m_Rtd_zz_o0_t8_l0 = vector(rk2gtd_zz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
						rk2m_Rtd_zz_o0_t7_l0 = vector(rk2gtd_yz, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
						rk2m_Rtd_zz_o0_t6_l0 = vector(rk2gtd_xz, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
						rk2m_Rtd_zz_o0_t5_l0 = (-0.5 * rk2gtu_zz) * d_Rtd_zz_o0_t5_m0_l0;
						rk2m_Rtd_zz_o0_t4_l0 = (-rk2gtu_yz) * d_Rtd_zz_o0_t4_m0_l0;
						rk2m_Rtd_zz_o0_t3_l0 = (-0.5 * rk2gtu_yy) * d_Rtd_zz_o0_t3_m0_l0;
						rk2m_Rtd_zz_o0_t2_l0 = (-rk2gtu_xz) * d_Rtd_zz_o0_t2_m0_l0;
						rk2m_Rtd_zz_o0_t1_l0 = (-rk2gtu_xy) * d_Rtd_zz_o0_t1_m0_l0;
						rk2m_Rtd_zz_o0_t0_l0 = (-0.5 * rk2gtu_xx) * d_Rtd_zz_o0_t0_m0_l0;
						rk2Rtd_zz = ((((((((rk2m_Rtd_zz_o0_t0_l0 + rk2m_Rtd_zz_o0_t1_l0) + rk2m_Rtd_zz_o0_t2_l0) + rk2m_Rtd_zz_o0_t3_l0) + rk2m_Rtd_zz_o0_t4_l0) + rk2m_Rtd_zz_o0_t5_l0) + rk2m_Rtd_zz_o0_t6_l0) + rk2m_Rtd_zz_o0_t7_l0) + rk2m_Rtd_zz_o0_t8_l0) + rk2d_Rtd_zz_o0_t9_m0_l0;
						rk2d_Rtd_xy_o0_t12_m0_l0 = rk2gtu_zz * (rk2Ct_xxz * rk2Ctd_yxz + rk2Ct_xxz * rk2Ctd_xyz + rk2Ct_xyz * rk2Ctd_xxz + 2.0 * rk2Ct_yxz * rk2Ctd_yyz + rk2Ct_yyz * rk2Ctd_xyz + rk2Ct_zxz * rk2Ctd_yzz + rk2Ct_zxz * rk2Ctd_zyz + rk2Ct_zyz * rk2Ctd_xzz) + rk2gtu_yz * (rk2Ct_xxy * rk2Ctd_yxz + rk2Ct_xxy * rk2Ctd_xyz + rk2Ct_xyy * rk2Ctd_xxz + 2.0 * rk2Ct_yxy * rk2Ctd_yyz + rk2Ct_yyy * rk2Ctd_xyz + rk2Ct_zxy * rk2Ctd_yzz + rk2Ct_zxy * rk2Ctd_zyz + rk2Ct_zyy * rk2Ctd_xzz) + rk2gtu_xz * (rk2Ct_xxx * rk2Ctd_yxz + rk2Ct_xxx * rk2Ctd_xyz + rk2Ct_xxy * rk2Ctd_xxz + 2.0 * rk2Ct_yxx * rk2Ctd_yyz + rk2Ct_yxy * rk2Ctd_xyz + rk2Ct_zxx * rk2Ctd_yzz + rk2Ct_zxx * rk2Ctd_zyz + rk2Ct_zxy * rk2Ctd_xzz) + rk2gtu_yz * (rk2Ct_xxz * rk2Ctd_yxy + rk2Ct_xxz * rk2Ctd_xyy + rk2Ct_xyz * rk2Ctd_xxy + 2.0 * rk2Ct_yxz * rk2Ctd_yyy + rk2Ct_yyz * rk2Ctd_xyy + rk2Ct_zxz * rk2Ctd_yyz + rk2Ct_zxz * rk2Ctd_zyy + rk2Ct_zyz * rk2Ctd_xyz) + rk2gtu_yy * (rk2Ct_xxy * rk2Ctd_yxy + rk2Ct_xxy * rk2Ctd_xyy + rk2Ct_xyy * rk2Ctd_xxy + 2.0 * rk2Ct_yxy * rk2Ctd_yyy + rk2Ct_yyy * rk2Ctd_xyy + rk2Ct_zxy * rk2Ctd_yyz + rk2Ct_zxy * rk2Ctd_zyy + rk2Ct_zyy * rk2Ctd_xyz) + rk2gtu_xy * (rk2Ct_xxx * rk2Ctd_yxy + rk2Ct_xxx * rk2Ctd_xyy + rk2Ct_xxy * rk2Ctd_xxy + 2.0 * rk2Ct_yxx * rk2Ctd_yyy + rk2Ct_yxy * rk2Ctd_xyy + rk2Ct_zxx * rk2Ctd_yyz + rk2Ct_zxx * rk2Ctd_zyy + rk2Ct_zxy * rk2Ctd_xyz) + rk2gtu_xz * (rk2Ct_xxz * rk2Ctd_yxx + rk2Ct_xxz * rk2Ctd_xxy + rk2Ct_xyz * rk2Ctd_xxx + 2.0 * rk2Ct_yxz * rk2Ctd_yxy + rk2Ct_yyz * rk2Ctd_xxy + rk2Ct_zxz * rk2Ctd_yxz + rk2Ct_zxz * rk2Ctd_zxy + rk2Ct_zyz * rk2Ctd_xxz) + rk2gtu_xy * (rk2Ct_xxy * rk2Ctd_yxx + rk2Ct_xxy * rk2Ctd_xxy + rk2Ct_xyy * rk2Ctd_xxx + 2.0 * rk2Ct_yxy * rk2Ctd_yxy + rk2Ct_yyy * rk2Ctd_xxy + rk2Ct_zxy * rk2Ctd_yxz + rk2Ct_zxy * rk2Ctd_zxy + rk2Ct_zyy * rk2Ctd_xxz) + rk2gtu_xx * (rk2Ct_xxx * rk2Ctd_yxx + rk2Ct_xxx * rk2Ctd_xxy + rk2Ct_xxy * rk2Ctd_xxx + 2.0 * rk2Ct_yxx * rk2Ctd_yxy + rk2Ct_yxy * rk2Ctd_xxy + rk2Ct_zxx * rk2Ctd_yxz + rk2Ct_zxx * rk2Ctd_zxy + rk2Ct_zxy * rk2Ctd_xxz) + 0.5 * vector(rk2Gamh_z, i, j, k) * (rk2Ctd_xyz + rk2Ctd_yxz) + 0.5 * vector(rk2Gamh_y, i, j, k) * (rk2Ctd_xyy + rk2Ctd_yxy) + 0.5 * vector(rk2Gamh_x, i, j, k) * (rk2Ctd_xxy + rk2Ctd_yxx);
						rk2m_Rtd_xy_o0_t11_l0 = 0.5 * vector(rk2gtd_yz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
						rk2m_Rtd_xy_o0_t10_l0 = 0.5 * vector(rk2gtd_yy, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
						rk2m_Rtd_xy_o0_t9_l0 = 0.5 * vector(rk2gtd_xy, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
						rk2m_Rtd_xy_o0_t8_l0 = 0.5 * vector(rk2gtd_xz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
						rk2m_Rtd_xy_o0_t7_l0 = 0.5 * vector(rk2gtd_xy, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
						rk2m_Rtd_xy_o0_t6_l0 = 0.5 * vector(rk2gtd_xx, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
						rk2m_Rtd_xy_o0_t5_l0 = (-0.5 * rk2gtu_zz) * d_Rtd_xy_o0_t5_m0_l0;
						rk2m_Rtd_xy_o0_t4_l0 = (-rk2gtu_yz) * d_Rtd_xy_o0_t4_m0_l0;
						rk2m_Rtd_xy_o0_t3_l0 = (-0.5 * rk2gtu_yy) * d_Rtd_xy_o0_t3_m0_l0;
						rk2m_Rtd_xy_o0_t2_l0 = (-rk2gtu_xz) * d_Rtd_xy_o0_t2_m0_l0;
						rk2m_Rtd_xy_o0_t1_l0 = (-rk2gtu_xy) * d_Rtd_xy_o0_t1_m0_l0;
						rk2m_Rtd_xy_o0_t0_l0 = (-0.5 * rk2gtu_xx) * d_Rtd_xy_o0_t0_m0_l0;
						rk2Rtd_xy = (((((((((((rk2m_Rtd_xy_o0_t0_l0 + rk2m_Rtd_xy_o0_t1_l0) + rk2m_Rtd_xy_o0_t2_l0) + rk2m_Rtd_xy_o0_t3_l0) + rk2m_Rtd_xy_o0_t4_l0) + rk2m_Rtd_xy_o0_t5_l0) + rk2m_Rtd_xy_o0_t6_l0) + rk2m_Rtd_xy_o0_t7_l0) + rk2m_Rtd_xy_o0_t8_l0) + rk2m_Rtd_xy_o0_t9_l0) + rk2m_Rtd_xy_o0_t10_l0) + rk2m_Rtd_xy_o0_t11_l0) + rk2d_Rtd_xy_o0_t12_m0_l0;
						rk2d_Rtd_yy_o0_t9_m0_l0 = rk2gtu_zz * (2.0 * rk2Ct_xyz * rk2Ctd_yxz + rk2Ct_xyz * rk2Ctd_xyz + 3.0 * rk2Ct_yyz * rk2Ctd_yyz + 2.0 * rk2Ct_zyz * rk2Ctd_yzz + rk2Ct_zyz * rk2Ctd_zyz) + rk2gtu_yz * (2.0 * rk2Ct_xyy * rk2Ctd_yxz + rk2Ct_xyy * rk2Ctd_xyz + 3.0 * rk2Ct_yyy * rk2Ctd_yyz + 2.0 * rk2Ct_zyy * rk2Ctd_yzz + rk2Ct_zyy * rk2Ctd_zyz) + rk2gtu_xz * (2.0 * rk2Ct_xxy * rk2Ctd_yxz + rk2Ct_xxy * rk2Ctd_xyz + 3.0 * rk2Ct_yxy * rk2Ctd_yyz + 2.0 * rk2Ct_zxy * rk2Ctd_yzz + rk2Ct_zxy * rk2Ctd_zyz) + rk2gtu_yz * (2.0 * rk2Ct_xyz * rk2Ctd_yxy + rk2Ct_xyz * rk2Ctd_xyy + 3.0 * rk2Ct_yyz * rk2Ctd_yyy + 2.0 * rk2Ct_zyz * rk2Ctd_yyz + rk2Ct_zyz * rk2Ctd_zyy) + rk2gtu_yy * (2.0 * rk2Ct_xyy * rk2Ctd_yxy + rk2Ct_xyy * rk2Ctd_xyy + 3.0 * rk2Ct_yyy * rk2Ctd_yyy + 2.0 * rk2Ct_zyy * rk2Ctd_yyz + rk2Ct_zyy * rk2Ctd_zyy) + rk2gtu_xy * (2.0 * rk2Ct_xxy * rk2Ctd_yxy + rk2Ct_xxy * rk2Ctd_xyy + 3.0 * rk2Ct_yxy * rk2Ctd_yyy + 2.0 * rk2Ct_zxy * rk2Ctd_yyz + rk2Ct_zxy * rk2Ctd_zyy) + rk2gtu_xz * (2.0 * rk2Ct_xyz * rk2Ctd_yxx + rk2Ct_xyz * rk2Ctd_xxy + 3.0 * rk2Ct_yyz * rk2Ctd_yxy + 2.0 * rk2Ct_zyz * rk2Ctd_yxz + rk2Ct_zyz * rk2Ctd_zxy) + rk2gtu_xy * (2.0 * rk2Ct_xyy * rk2Ctd_yxx + rk2Ct_xyy * rk2Ctd_xxy + 3.0 * rk2Ct_yyy * rk2Ctd_yxy + 2.0 * rk2Ct_zyy * rk2Ctd_yxz + rk2Ct_zyy * rk2Ctd_zxy) + rk2gtu_xx * (2.0 * rk2Ct_xxy * rk2Ctd_yxx + rk2Ct_xxy * rk2Ctd_xxy + 3.0 * rk2Ct_yxy * rk2Ctd_yxy + 2.0 * rk2Ct_zxy * rk2Ctd_yxz + rk2Ct_zxy * rk2Ctd_zxy) + vector(rk2Gamh_z, i, j, k) * rk2Ctd_yyz + vector(rk2Gamh_y, i, j, k) * rk2Ctd_yyy + vector(rk2Gamh_x, i, j, k) * rk2Ctd_yxy;
						rk2m_Rtd_yy_o0_t8_l0 = vector(rk2gtd_yz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
						rk2m_Rtd_yy_o0_t7_l0 = vector(rk2gtd_yy, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
						rk2m_Rtd_yy_o0_t6_l0 = vector(rk2gtd_xy, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
						rk2m_Rtd_yy_o0_t5_l0 = (-0.5 * rk2gtu_zz) * d_Rtd_yy_o0_t5_m0_l0;
						rk2m_Rtd_yy_o0_t4_l0 = (-rk2gtu_yz) * d_Rtd_yy_o0_t4_m0_l0;
						rk2m_Rtd_yy_o0_t3_l0 = (-0.5 * rk2gtu_yy) * d_Rtd_yy_o0_t3_m0_l0;
						rk2m_Rtd_yy_o0_t2_l0 = (-rk2gtu_xz) * d_Rtd_yy_o0_t2_m0_l0;
						rk2m_Rtd_yy_o0_t1_l0 = (-rk2gtu_xy) * d_Rtd_yy_o0_t1_m0_l0;
						rk2m_Rtd_yy_o0_t0_l0 = (-0.5 * rk2gtu_xx) * d_Rtd_yy_o0_t0_m0_l0;
						rk2Rtd_yy = ((((((((rk2m_Rtd_yy_o0_t0_l0 + rk2m_Rtd_yy_o0_t1_l0) + rk2m_Rtd_yy_o0_t2_l0) + rk2m_Rtd_yy_o0_t3_l0) + rk2m_Rtd_yy_o0_t4_l0) + rk2m_Rtd_yy_o0_t5_l0) + rk2m_Rtd_yy_o0_t6_l0) + rk2m_Rtd_yy_o0_t7_l0) + rk2m_Rtd_yy_o0_t8_l0) + rk2d_Rtd_yy_o0_t9_m0_l0;
						rk2d_Rtd_yz_o0_t12_m0_l0 = rk2gtu_zz * (rk2Ct_xyz * rk2Ctd_zxz + rk2Ct_xyz * rk2Ctd_xzz + rk2Ct_xzz * rk2Ctd_yxz + rk2Ct_yyz * rk2Ctd_zyz + rk2Ct_yyz * rk2Ctd_yzz + rk2Ct_yzz * rk2Ctd_yyz + 2.0 * rk2Ct_zyz * rk2Ctd_zzz + rk2Ct_zzz * rk2Ctd_yzz) + rk2gtu_yz * (rk2Ct_xyy * rk2Ctd_zxz + rk2Ct_xyy * rk2Ctd_xzz + rk2Ct_xyz * rk2Ctd_yxz + rk2Ct_yyy * rk2Ctd_zyz + rk2Ct_yyy * rk2Ctd_yzz + rk2Ct_yyz * rk2Ctd_yyz + 2.0 * rk2Ct_zyy * rk2Ctd_zzz + rk2Ct_zyz * rk2Ctd_yzz) + rk2gtu_xz * (rk2Ct_xxy * rk2Ctd_zxz + rk2Ct_xxy * rk2Ctd_xzz + rk2Ct_xxz * rk2Ctd_yxz + rk2Ct_yxy * rk2Ctd_zyz + rk2Ct_yxy * rk2Ctd_yzz + rk2Ct_yxz * rk2Ctd_yyz + 2.0 * rk2Ct_zxy * rk2Ctd_zzz + rk2Ct_zxz * rk2Ctd_yzz) + rk2gtu_yz * (rk2Ct_xyz * rk2Ctd_zxy + rk2Ct_xyz * rk2Ctd_xyz + rk2Ct_xzz * rk2Ctd_yxy + rk2Ct_yyz * rk2Ctd_zyy + rk2Ct_yyz * rk2Ctd_yyz + rk2Ct_yzz * rk2Ctd_yyy + 2.0 * rk2Ct_zyz * rk2Ctd_zyz + rk2Ct_zzz * rk2Ctd_yyz) + rk2gtu_yy * (rk2Ct_xyy * rk2Ctd_zxy + rk2Ct_xyy * rk2Ctd_xyz + rk2Ct_xyz * rk2Ctd_yxy + rk2Ct_yyy * rk2Ctd_zyy + rk2Ct_yyy * rk2Ctd_yyz + rk2Ct_yyz * rk2Ctd_yyy + 2.0 * rk2Ct_zyy * rk2Ctd_zyz + rk2Ct_zyz * rk2Ctd_yyz) + rk2gtu_xy * (rk2Ct_xxy * rk2Ctd_zxy + rk2Ct_xxy * rk2Ctd_xyz + rk2Ct_xxz * rk2Ctd_yxy + rk2Ct_yxy * rk2Ctd_zyy + rk2Ct_yxy * rk2Ctd_yyz + rk2Ct_yxz * rk2Ctd_yyy + 2.0 * rk2Ct_zxy * rk2Ctd_zyz + rk2Ct_zxz * rk2Ctd_yyz) + rk2gtu_xz * (rk2Ct_xyz * rk2Ctd_zxx + rk2Ct_xyz * rk2Ctd_xxz + rk2Ct_xzz * rk2Ctd_yxx + rk2Ct_yyz * rk2Ctd_zxy + rk2Ct_yyz * rk2Ctd_yxz + rk2Ct_yzz * rk2Ctd_yxy + 2.0 * rk2Ct_zyz * rk2Ctd_zxz + rk2Ct_zzz * rk2Ctd_yxz) + rk2gtu_xy * (rk2Ct_xyy * rk2Ctd_zxx + rk2Ct_xyy * rk2Ctd_xxz + rk2Ct_xyz * rk2Ctd_yxx + rk2Ct_yyy * rk2Ctd_zxy + rk2Ct_yyy * rk2Ctd_yxz + rk2Ct_yyz * rk2Ctd_yxy + 2.0 * rk2Ct_zyy * rk2Ctd_zxz + rk2Ct_zyz * rk2Ctd_yxz) + rk2gtu_xx * (rk2Ct_xxy * rk2Ctd_zxx + rk2Ct_xxy * rk2Ctd_xxz + rk2Ct_xxz * rk2Ctd_yxx + rk2Ct_yxy * rk2Ctd_zxy + rk2Ct_yxy * rk2Ctd_yxz + rk2Ct_yxz * rk2Ctd_yxy + 2.0 * rk2Ct_zxy * rk2Ctd_zxz + rk2Ct_zxz * rk2Ctd_yxz) + 0.5 * vector(rk2Gamh_z, i, j, k) * (rk2Ctd_yzz + rk2Ctd_zyz) + 0.5 * vector(rk2Gamh_y, i, j, k) * (rk2Ctd_yyz + rk2Ctd_zyy) + 0.5 * vector(rk2Gamh_x, i, j, k) * (rk2Ctd_yxz + rk2Ctd_zxy);
						rk2m_Rtd_yz_o0_t11_l0 = 0.5 * vector(rk2gtd_zz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
						rk2m_Rtd_yz_o0_t10_l0 = 0.5 * vector(rk2gtd_yz, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
						rk2m_Rtd_yz_o0_t9_l0 = 0.5 * vector(rk2gtd_xz, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
						rk2m_Rtd_yz_o0_t8_l0 = 0.5 * vector(rk2gtd_yz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
						rk2m_Rtd_yz_o0_t7_l0 = 0.5 * vector(rk2gtd_yy, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
						rk2m_Rtd_yz_o0_t6_l0 = 0.5 * vector(rk2gtd_xy, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
						rk2m_Rtd_yz_o0_t5_l0 = (-0.5 * rk2gtu_zz) * d_Rtd_yz_o0_t5_m0_l0;
						rk2m_Rtd_yz_o0_t4_l0 = (-rk2gtu_yz) * d_Rtd_yz_o0_t4_m0_l0;
						rk2m_Rtd_yz_o0_t3_l0 = (-0.5 * rk2gtu_yy) * d_Rtd_yz_o0_t3_m0_l0;
						rk2m_Rtd_yz_o0_t2_l0 = (-rk2gtu_xz) * d_Rtd_yz_o0_t2_m0_l0;
						rk2m_Rtd_yz_o0_t1_l0 = (-rk2gtu_xy) * d_Rtd_yz_o0_t1_m0_l0;
						rk2m_Rtd_yz_o0_t0_l0 = (-0.5 * rk2gtu_xx) * d_Rtd_yz_o0_t0_m0_l0;
						rk2Rtd_yz = (((((((((((rk2m_Rtd_yz_o0_t0_l0 + rk2m_Rtd_yz_o0_t1_l0) + rk2m_Rtd_yz_o0_t2_l0) + rk2m_Rtd_yz_o0_t3_l0) + rk2m_Rtd_yz_o0_t4_l0) + rk2m_Rtd_yz_o0_t5_l0) + rk2m_Rtd_yz_o0_t6_l0) + rk2m_Rtd_yz_o0_t7_l0) + rk2m_Rtd_yz_o0_t8_l0) + rk2m_Rtd_yz_o0_t9_l0) + rk2m_Rtd_yz_o0_t10_l0) + rk2m_Rtd_yz_o0_t11_l0) + rk2d_Rtd_yz_o0_t12_m0_l0;
						rk2m_Rpd_xx_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t3_m0_l0;
						rk2m_Rpd_xx_o0_t21_l0 = (-0.75 * vector(rk2gtd_xx, i, j, k) * rk2gtu_zz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xx_o0_t20_l0 = (-0.75 * vector(rk2gtd_xx, i, j, k) * rk2gtu_yy * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xx_o0_t19_l0 = (-0.75 * vector(rk2gtd_xx, i, j, k) * rk2gtu_xx * rk2inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xx_o0_t18_l0 = 2.0 * rk2inv_chi * rk2Zu_z * vector(rk2gtd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xx_o0_t17_l0 = 2.0 * rk2inv_chi * rk2Zu_y * vector(rk2gtd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xx_o0_t16_l0 = 2.0 * rk2inv_chi * rk2Zu_x * vector(rk2gtd_xx, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xx_o0_t15_l0 = (-1.500000000000000 * vector(rk2gtd_xx, i, j, k) * rk2gtu_yz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xx_o0_t14_l0 = (-1.500000000000000 * vector(rk2gtd_xx, i, j, k) * rk2gtu_xz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xx_o0_t13_l0 = (-1.500000000000000 * vector(rk2gtd_xx, i, j, k) * rk2gtu_xy * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xx_o0_t12_l0 = (-0.25 * rk2inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xx_o0_t11_l0 = (-0.5 * rk2Ct_zxx) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xx_o0_t10_l0 = (-0.5 * rk2Ct_yxx) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xx_o0_t9_l0 = (-0.5 * rk2Ct_xxx) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xx_o0_t8_l0 = 0.5 * vector(rk2gtd_xx, i, j, k) * rk2gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk2m_Rpd_xx_o0_t7_l0 = vector(rk2gtd_xx, i, j, k) * rk2gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk2m_Rpd_xx_o0_t6_l0 = 0.5 * vector(rk2gtd_xx, i, j, k) * rk2gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk2m_Rpd_xx_o0_t5_l0 = vector(rk2gtd_xx, i, j, k) * rk2gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk2m_Rpd_xx_o0_t4_l0 = vector(rk2gtd_xx, i, j, k) * rk2gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk2m_Rpd_xx_o0_t3_l0 = 0.5 * vector(rk2gtd_xx, i, j, k) * rk2gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk2m_Rpd_xx_o0_t2_l0 = (-0.5 * vector(rk2gtd_xx, i, j, k) * vector(rk2Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xx_o0_t1_l0 = (-0.5 * vector(rk2gtd_xx, i, j, k) * vector(rk2Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xx_o0_t0_l0 = (-0.5 * vector(rk2gtd_xx, i, j, k) * vector(rk2Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk2Rpd_xx = (((((((((((((((((((((rk2m_Rpd_xx_o0_t0_l0 + rk2m_Rpd_xx_o0_t1_l0) + rk2m_Rpd_xx_o0_t2_l0) + rk2m_Rpd_xx_o0_t3_l0) + rk2m_Rpd_xx_o0_t4_l0) + rk2m_Rpd_xx_o0_t5_l0) + rk2m_Rpd_xx_o0_t6_l0) + rk2m_Rpd_xx_o0_t7_l0) + rk2m_Rpd_xx_o0_t8_l0) + rk2m_Rpd_xx_o0_t9_l0) + rk2m_Rpd_xx_o0_t10_l0) + rk2m_Rpd_xx_o0_t11_l0) + rk2m_Rpd_xx_o0_t12_l0) + rk2m_Rpd_xx_o0_t13_l0) + rk2m_Rpd_xx_o0_t14_l0) + rk2m_Rpd_xx_o0_t15_l0) + rk2m_Rpd_xx_o0_t16_l0) + rk2m_Rpd_xx_o0_t17_l0) + rk2m_Rpd_xx_o0_t18_l0) + rk2m_Rpd_xx_o0_t19_l0) + rk2m_Rpd_xx_o0_t20_l0) + rk2m_Rpd_xx_o0_t21_l0) + rk2m_Rpd_xx_o0_t22_l0;
						rk2m_Rpd_yy_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t6_m0_l0;
						rk2m_Rpd_yy_o0_t21_l0 = (-0.75 * vector(rk2gtd_yy, i, j, k) * rk2gtu_zz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_yy_o0_t20_l0 = (-0.75 * vector(rk2gtd_yy, i, j, k) * rk2gtu_yy * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yy_o0_t19_l0 = (-0.75 * vector(rk2gtd_yy, i, j, k) * rk2gtu_xx * rk2inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_yy_o0_t18_l0 = 2.0 * rk2inv_chi * rk2Zu_z * vector(rk2gtd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yy_o0_t17_l0 = 2.0 * rk2inv_chi * rk2Zu_y * vector(rk2gtd_yy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yy_o0_t16_l0 = 2.0 * rk2inv_chi * rk2Zu_x * vector(rk2gtd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yy_o0_t15_l0 = (-1.500000000000000 * vector(rk2gtd_yy, i, j, k) * rk2gtu_xz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_yy_o0_t14_l0 = (-1.500000000000000 * vector(rk2gtd_yy, i, j, k) * rk2gtu_xy * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_yy_o0_t13_l0 = (-0.25 * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yy_o0_t12_l0 = (-0.5 * rk2Ct_zyy) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_yy_o0_t11_l0 = (-0.5 * rk2Ct_yyy) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yy_o0_t10_l0 = (-0.5 * rk2Ct_xyy) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_yy_o0_t9_l0 = 0.5 * vector(rk2gtd_yy, i, j, k) * rk2gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk2m_Rpd_yy_o0_t8_l0 = vector(rk2gtd_yy, i, j, k) * rk2gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk2m_Rpd_yy_o0_t7_l0 = 0.5 * vector(rk2gtd_yy, i, j, k) * rk2gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk2m_Rpd_yy_o0_t6_l0 = vector(rk2gtd_yy, i, j, k) * rk2gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk2m_Rpd_yy_o0_t5_l0 = vector(rk2gtd_yy, i, j, k) * rk2gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk2m_Rpd_yy_o0_t4_l0 = 0.5 * vector(rk2gtd_yy, i, j, k) * rk2gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk2m_Rpd_yy_o0_t3_l0 = (-0.5 * vector(rk2gtd_yy, i, j, k) * vector(rk2Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_yy_o0_t2_l0 = (-0.5 * vector(rk2gtd_yy, i, j, k) * vector(rk2Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yy_o0_t1_l0 = (-0.5 * vector(rk2gtd_yy, i, j, k) * vector(rk2Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_yy_o0_t0_l0 = (-1.500000000000000 * vector(rk2gtd_yy, i, j, k) * rk2gtu_yz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2Rpd_yy = (((((((((((((((((((((rk2m_Rpd_yy_o0_t0_l0 + rk2m_Rpd_yy_o0_t1_l0) + rk2m_Rpd_yy_o0_t2_l0) + rk2m_Rpd_yy_o0_t3_l0) + rk2m_Rpd_yy_o0_t4_l0) + rk2m_Rpd_yy_o0_t5_l0) + rk2m_Rpd_yy_o0_t6_l0) + rk2m_Rpd_yy_o0_t7_l0) + rk2m_Rpd_yy_o0_t8_l0) + rk2m_Rpd_yy_o0_t9_l0) + rk2m_Rpd_yy_o0_t10_l0) + rk2m_Rpd_yy_o0_t11_l0) + rk2m_Rpd_yy_o0_t12_l0) + rk2m_Rpd_yy_o0_t13_l0) + rk2m_Rpd_yy_o0_t14_l0) + rk2m_Rpd_yy_o0_t15_l0) + rk2m_Rpd_yy_o0_t16_l0) + rk2m_Rpd_yy_o0_t17_l0) + rk2m_Rpd_yy_o0_t18_l0) + rk2m_Rpd_yy_o0_t19_l0) + rk2m_Rpd_yy_o0_t20_l0) + rk2m_Rpd_yy_o0_t21_l0) + rk2m_Rpd_yy_o0_t22_l0;
						rk2m_Rpd_yz_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t7_m0_l0;
						rk2m_Rpd_yz_o0_t24_l0 = (-0.75 * vector(rk2gtd_yz, i, j, k) * rk2gtu_zz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_yz_o0_t23_l0 = (-0.75 * vector(rk2gtd_yz, i, j, k) * rk2gtu_yy * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yz_o0_t22_l0 = (-0.75 * vector(rk2gtd_yz, i, j, k) * rk2gtu_xx * rk2inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_yz_o0_t21_l0 = rk2inv_chi * rk2Zu_z * vector(rk2gtd_zz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yz_o0_t20_l0 = rk2inv_chi * rk2Zu_y * vector(rk2gtd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yz_o0_t19_l0 = rk2inv_chi * rk2Zu_x * vector(rk2gtd_xz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yz_o0_t18_l0 = rk2inv_chi * rk2Zu_z * vector(rk2gtd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_yz_o0_t17_l0 = rk2inv_chi * rk2Zu_y * vector(rk2gtd_yy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_yz_o0_t16_l0 = rk2inv_chi * rk2Zu_x * vector(rk2gtd_xy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_yz_o0_t15_l0 = (-0.5 * rk2Ct_zyz) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_yz_o0_t14_l0 = (-0.5 * rk2Ct_yyz) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yz_o0_t13_l0 = (-0.5 * rk2Ct_xyz) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_yz_o0_t12_l0 = (-0.25 * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_yz_o0_t11_l0 = 0.5 * vector(rk2gtd_yz, i, j, k) * rk2gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk2m_Rpd_yz_o0_t10_l0 = vector(rk2gtd_yz, i, j, k) * rk2gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk2m_Rpd_yz_o0_t9_l0 = 0.5 * vector(rk2gtd_yz, i, j, k) * rk2gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk2m_Rpd_yz_o0_t8_l0 = vector(rk2gtd_yz, i, j, k) * rk2gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk2m_Rpd_yz_o0_t7_l0 = vector(rk2gtd_yz, i, j, k) * rk2gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk2m_Rpd_yz_o0_t6_l0 = 0.5 * vector(rk2gtd_yz, i, j, k) * rk2gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk2m_Rpd_yz_o0_t5_l0 = (-0.5 * vector(rk2gtd_yz, i, j, k) * vector(rk2Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_yz_o0_t4_l0 = (-0.5 * vector(rk2gtd_yz, i, j, k) * vector(rk2Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yz_o0_t3_l0 = (-0.5 * vector(rk2gtd_yz, i, j, k) * vector(rk2Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_yz_o0_t2_l0 = (-1.500000000000000 * vector(rk2gtd_yz, i, j, k) * rk2gtu_yz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_yz_o0_t1_l0 = (-1.500000000000000 * vector(rk2gtd_yz, i, j, k) * rk2gtu_xz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_yz_o0_t0_l0 = (-1.500000000000000 * vector(rk2gtd_yz, i, j, k) * rk2gtu_xy * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2Rpd_yz = ((((((((((((((((((((((((rk2m_Rpd_yz_o0_t0_l0 + rk2m_Rpd_yz_o0_t1_l0) + rk2m_Rpd_yz_o0_t2_l0) + rk2m_Rpd_yz_o0_t3_l0) + rk2m_Rpd_yz_o0_t4_l0) + rk2m_Rpd_yz_o0_t5_l0) + rk2m_Rpd_yz_o0_t6_l0) + rk2m_Rpd_yz_o0_t7_l0) + rk2m_Rpd_yz_o0_t8_l0) + rk2m_Rpd_yz_o0_t9_l0) + rk2m_Rpd_yz_o0_t10_l0) + rk2m_Rpd_yz_o0_t11_l0) + rk2m_Rpd_yz_o0_t12_l0) + rk2m_Rpd_yz_o0_t13_l0) + rk2m_Rpd_yz_o0_t14_l0) + rk2m_Rpd_yz_o0_t15_l0) + rk2m_Rpd_yz_o0_t16_l0) + rk2m_Rpd_yz_o0_t17_l0) + rk2m_Rpd_yz_o0_t18_l0) + rk2m_Rpd_yz_o0_t19_l0) + rk2m_Rpd_yz_o0_t20_l0) + rk2m_Rpd_yz_o0_t21_l0) + rk2m_Rpd_yz_o0_t22_l0) + rk2m_Rpd_yz_o0_t23_l0) + rk2m_Rpd_yz_o0_t24_l0) + rk2m_Rpd_yz_o0_t25_l0;
						rk2m_Rpd_xz_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t5_m0_l0;
						rk2m_Rpd_xz_o0_t24_l0 = (-0.75 * vector(rk2gtd_xz, i, j, k) * rk2gtu_zz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xz_o0_t23_l0 = (-0.75 * vector(rk2gtd_xz, i, j, k) * rk2gtu_yy * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xz_o0_t22_l0 = (-0.75 * vector(rk2gtd_xz, i, j, k) * rk2gtu_xx * rk2inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xz_o0_t21_l0 = rk2inv_chi * rk2Zu_z * vector(rk2gtd_zz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xz_o0_t20_l0 = rk2inv_chi * rk2Zu_y * vector(rk2gtd_yz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xz_o0_t19_l0 = rk2inv_chi * rk2Zu_x * vector(rk2gtd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xz_o0_t18_l0 = rk2inv_chi * rk2Zu_z * vector(rk2gtd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xz_o0_t17_l0 = rk2inv_chi * rk2Zu_y * vector(rk2gtd_xy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xz_o0_t16_l0 = rk2inv_chi * rk2Zu_x * vector(rk2gtd_xx, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xz_o0_t15_l0 = (-1.500000000000000 * vector(rk2gtd_xz, i, j, k) * rk2gtu_yz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xz_o0_t14_l0 = (-1.500000000000000 * vector(rk2gtd_xz, i, j, k) * rk2gtu_xz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xz_o0_t13_l0 = (-1.500000000000000 * vector(rk2gtd_xz, i, j, k) * rk2gtu_xy * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xz_o0_t12_l0 = (-0.5 * rk2Ct_zxz) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xz_o0_t11_l0 = (-0.5 * rk2Ct_yxz) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xz_o0_t10_l0 = (-0.5 * rk2Ct_xxz) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xz_o0_t9_l0 = (-0.25 * rk2inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xz_o0_t8_l0 = 0.5 * vector(rk2gtd_xz, i, j, k) * rk2gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk2m_Rpd_xz_o0_t7_l0 = vector(rk2gtd_xz, i, j, k) * rk2gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk2m_Rpd_xz_o0_t6_l0 = 0.5 * vector(rk2gtd_xz, i, j, k) * rk2gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk2m_Rpd_xz_o0_t5_l0 = vector(rk2gtd_xz, i, j, k) * rk2gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk2m_Rpd_xz_o0_t4_l0 = vector(rk2gtd_xz, i, j, k) * rk2gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk2m_Rpd_xz_o0_t3_l0 = 0.5 * vector(rk2gtd_xz, i, j, k) * rk2gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk2m_Rpd_xz_o0_t2_l0 = (-0.5 * vector(rk2gtd_xz, i, j, k) * vector(rk2Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xz_o0_t1_l0 = (-0.5 * vector(rk2gtd_xz, i, j, k) * vector(rk2Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xz_o0_t0_l0 = (-0.5 * vector(rk2gtd_xz, i, j, k) * vector(rk2Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk2Rpd_xz = ((((((((((((((((((((((((rk2m_Rpd_xz_o0_t0_l0 + rk2m_Rpd_xz_o0_t1_l0) + rk2m_Rpd_xz_o0_t2_l0) + rk2m_Rpd_xz_o0_t3_l0) + rk2m_Rpd_xz_o0_t4_l0) + rk2m_Rpd_xz_o0_t5_l0) + rk2m_Rpd_xz_o0_t6_l0) + rk2m_Rpd_xz_o0_t7_l0) + rk2m_Rpd_xz_o0_t8_l0) + rk2m_Rpd_xz_o0_t9_l0) + rk2m_Rpd_xz_o0_t10_l0) + rk2m_Rpd_xz_o0_t11_l0) + rk2m_Rpd_xz_o0_t12_l0) + rk2m_Rpd_xz_o0_t13_l0) + rk2m_Rpd_xz_o0_t14_l0) + rk2m_Rpd_xz_o0_t15_l0) + rk2m_Rpd_xz_o0_t16_l0) + rk2m_Rpd_xz_o0_t17_l0) + rk2m_Rpd_xz_o0_t18_l0) + rk2m_Rpd_xz_o0_t19_l0) + rk2m_Rpd_xz_o0_t20_l0) + rk2m_Rpd_xz_o0_t21_l0) + rk2m_Rpd_xz_o0_t22_l0) + rk2m_Rpd_xz_o0_t23_l0) + rk2m_Rpd_xz_o0_t24_l0) + rk2m_Rpd_xz_o0_t25_l0;
						rk2m_Rpd_xy_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t4_m0_l0;
						rk2m_Rpd_xy_o0_t24_l0 = (-0.75 * vector(rk2gtd_xy, i, j, k) * rk2gtu_zz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xy_o0_t23_l0 = (-0.75 * vector(rk2gtd_xy, i, j, k) * rk2gtu_yy * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xy_o0_t22_l0 = (-0.75 * vector(rk2gtd_xy, i, j, k) * rk2gtu_xx * rk2inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xy_o0_t21_l0 = rk2inv_chi * rk2Zu_z * vector(rk2gtd_yz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xy_o0_t20_l0 = rk2inv_chi * rk2Zu_y * vector(rk2gtd_yy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xy_o0_t19_l0 = rk2inv_chi * rk2Zu_x * vector(rk2gtd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xy_o0_t18_l0 = rk2inv_chi * rk2Zu_z * vector(rk2gtd_xz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xy_o0_t17_l0 = rk2inv_chi * rk2Zu_y * vector(rk2gtd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xy_o0_t16_l0 = rk2inv_chi * rk2Zu_x * vector(rk2gtd_xx, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xy_o0_t15_l0 = (-1.500000000000000 * vector(rk2gtd_xy, i, j, k) * rk2gtu_yz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xy_o0_t14_l0 = (-1.500000000000000 * vector(rk2gtd_xy, i, j, k) * rk2gtu_xz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xy_o0_t13_l0 = (-1.500000000000000 * vector(rk2gtd_xy, i, j, k) * rk2gtu_xy * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xy_o0_t12_l0 = (-0.5 * rk2Ct_zxy) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xy_o0_t11_l0 = (-0.5 * rk2Ct_yxy) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xy_o0_t10_l0 = (-0.5 * rk2Ct_xxy) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xy_o0_t9_l0 = 0.5 * vector(rk2gtd_xy, i, j, k) * rk2gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk2m_Rpd_xy_o0_t8_l0 = vector(rk2gtd_xy, i, j, k) * rk2gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk2m_Rpd_xy_o0_t7_l0 = 0.5 * vector(rk2gtd_xy, i, j, k) * rk2gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk2m_Rpd_xy_o0_t6_l0 = vector(rk2gtd_xy, i, j, k) * rk2gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk2m_Rpd_xy_o0_t5_l0 = vector(rk2gtd_xy, i, j, k) * rk2gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk2m_Rpd_xy_o0_t4_l0 = 0.5 * vector(rk2gtd_xy, i, j, k) * rk2gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk2m_Rpd_xy_o0_t3_l0 = (-0.5 * vector(rk2gtd_xy, i, j, k) * vector(rk2Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_xy_o0_t2_l0 = (-0.5 * vector(rk2gtd_xy, i, j, k) * vector(rk2Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_xy_o0_t1_l0 = (-0.5 * vector(rk2gtd_xy, i, j, k) * vector(rk2Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_xy_o0_t0_l0 = (-0.25 * rk2inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2Rpd_xy = ((((((((((((((((((((((((rk2m_Rpd_xy_o0_t0_l0 + rk2m_Rpd_xy_o0_t1_l0) + rk2m_Rpd_xy_o0_t2_l0) + rk2m_Rpd_xy_o0_t3_l0) + rk2m_Rpd_xy_o0_t4_l0) + rk2m_Rpd_xy_o0_t5_l0) + rk2m_Rpd_xy_o0_t6_l0) + rk2m_Rpd_xy_o0_t7_l0) + rk2m_Rpd_xy_o0_t8_l0) + rk2m_Rpd_xy_o0_t9_l0) + rk2m_Rpd_xy_o0_t10_l0) + rk2m_Rpd_xy_o0_t11_l0) + rk2m_Rpd_xy_o0_t12_l0) + rk2m_Rpd_xy_o0_t13_l0) + rk2m_Rpd_xy_o0_t14_l0) + rk2m_Rpd_xy_o0_t15_l0) + rk2m_Rpd_xy_o0_t16_l0) + rk2m_Rpd_xy_o0_t17_l0) + rk2m_Rpd_xy_o0_t18_l0) + rk2m_Rpd_xy_o0_t19_l0) + rk2m_Rpd_xy_o0_t20_l0) + rk2m_Rpd_xy_o0_t21_l0) + rk2m_Rpd_xy_o0_t22_l0) + rk2m_Rpd_xy_o0_t23_l0) + rk2m_Rpd_xy_o0_t24_l0) + rk2m_Rpd_xy_o0_t25_l0;
						rk2m_Rpd_zz_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t8_m0_l0;
						rk2m_Rpd_zz_o0_t21_l0 = (-0.75 * vector(rk2gtd_zz, i, j, k) * rk2gtu_zz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_zz_o0_t20_l0 = (-0.75 * vector(rk2gtd_zz, i, j, k) * rk2gtu_yy * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_zz_o0_t19_l0 = (-0.75 * vector(rk2gtd_zz, i, j, k) * rk2gtu_xx * rk2inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_zz_o0_t18_l0 = 2.0 * rk2inv_chi * rk2Zu_z * vector(rk2gtd_zz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_zz_o0_t17_l0 = 2.0 * rk2inv_chi * rk2Zu_y * vector(rk2gtd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_zz_o0_t16_l0 = 2.0 * rk2inv_chi * rk2Zu_x * vector(rk2gtd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_zz_o0_t15_l0 = (-0.25 * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_zz_o0_t14_l0 = (-0.5 * rk2Ct_zzz) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_zz_o0_t13_l0 = (-0.5 * rk2Ct_yzz) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_zz_o0_t12_l0 = (-0.5 * rk2Ct_xzz) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_zz_o0_t11_l0 = vector(rk2gtd_zz, i, j, k) * rk2gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk2m_Rpd_zz_o0_t10_l0 = vector(rk2gtd_zz, i, j, k) * rk2gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk2m_Rpd_zz_o0_t9_l0 = 0.5 * vector(rk2gtd_zz, i, j, k) * rk2gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk2m_Rpd_zz_o0_t8_l0 = (-0.5 * vector(rk2gtd_zz, i, j, k) * vector(rk2Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Rpd_zz_o0_t7_l0 = (-0.5 * vector(rk2gtd_zz, i, j, k) * vector(rk2Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_zz_o0_t6_l0 = (-0.5 * vector(rk2gtd_zz, i, j, k) * vector(rk2Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_zz_o0_t5_l0 = 0.5 * vector(rk2gtd_zz, i, j, k) * rk2gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk2m_Rpd_zz_o0_t4_l0 = vector(rk2gtd_zz, i, j, k) * rk2gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk2m_Rpd_zz_o0_t3_l0 = 0.5 * vector(rk2gtd_zz, i, j, k) * rk2gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk2m_Rpd_zz_o0_t2_l0 = (-1.500000000000000 * vector(rk2gtd_zz, i, j, k) * rk2gtu_yz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Rpd_zz_o0_t1_l0 = (-1.500000000000000 * vector(rk2gtd_zz, i, j, k) * rk2gtu_xz * rk2inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Rpd_zz_o0_t0_l0 = (-1.500000000000000 * vector(rk2gtd_zz, i, j, k) * rk2gtu_xy * rk2inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2Rpd_zz = (((((((((((((((((((((rk2m_Rpd_zz_o0_t0_l0 + rk2m_Rpd_zz_o0_t1_l0) + rk2m_Rpd_zz_o0_t2_l0) + rk2m_Rpd_zz_o0_t3_l0) + rk2m_Rpd_zz_o0_t4_l0) + rk2m_Rpd_zz_o0_t5_l0) + rk2m_Rpd_zz_o0_t6_l0) + rk2m_Rpd_zz_o0_t7_l0) + rk2m_Rpd_zz_o0_t8_l0) + rk2m_Rpd_zz_o0_t9_l0) + rk2m_Rpd_zz_o0_t10_l0) + rk2m_Rpd_zz_o0_t11_l0) + rk2m_Rpd_zz_o0_t12_l0) + rk2m_Rpd_zz_o0_t13_l0) + rk2m_Rpd_zz_o0_t14_l0) + rk2m_Rpd_zz_o0_t15_l0) + rk2m_Rpd_zz_o0_t16_l0) + rk2m_Rpd_zz_o0_t17_l0) + rk2m_Rpd_zz_o0_t18_l0) + rk2m_Rpd_zz_o0_t19_l0) + rk2m_Rpd_zz_o0_t20_l0) + rk2m_Rpd_zz_o0_t21_l0) + rk2m_Rpd_zz_o0_t22_l0;
						rk2Rscalar = (rk2chi_max * rk2Rtd_xx + rk2Rpd_xx) * rk2gtu_xx + 2.0 * (rk2chi_max * rk2Rtd_xy + rk2Rpd_xy) * rk2gtu_xy + 2.0 * (rk2chi_max * rk2Rtd_xz + rk2Rpd_xz) * rk2gtu_xz + (rk2chi_max * rk2Rtd_yy + rk2Rpd_yy) * rk2gtu_yy + 2.0 * (rk2chi_max * rk2Rtd_yz + rk2Rpd_yz) * rk2gtu_yz + (rk2chi_max * rk2Rtd_zz + rk2Rpd_zz) * rk2gtu_zz;
						rk2Bfvf = vector(rk2Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk2Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk2Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
						rk2Sfuu_zz = rk2Sfu_z * rk2vfu_z + rk2chi_max * rk2gtu_zz * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_zz) + vector(rk2Bfu_z, i, j, k) * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_zz) + 2.0 * vector(rk2Bfu_z, i, j, k) * rk2vfu_z)) / rk2sdetg);
						rk2Tf4u_zz = rk2Sfuu_zz / rk2sdetg + (-(rk2Sfu_z * vector(rk2Betau_z, i, j, k)) / (rk2sdetg * vector(rk2Alpha, i, j, k))) + (-vector(rk2Betau_z, i, j, k) * rk2Tf4u_tz);
						rk2Tu_zz = rk2Tf4u_zz;
						rk2Sfuu_yz = 0.5 * rk2Sfu_y * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_y + rk2chi_max * rk2gtu_yz * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k) * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yz) + vector(rk2Bfu_y, i, j, k) * rk2vfu_z + vector(rk2Bfu_z, i, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Tf4u_yz = rk2Sfuu_yz / rk2sdetg + (-(rk2Sfu_y * vector(rk2Betau_z, i, j, k)) / (rk2sdetg * vector(rk2Alpha, i, j, k))) + (-vector(rk2Betau_y, i, j, k) * rk2Tf4u_tz);
						rk2Tu_yz = rk2Tf4u_yz;
						rk2Sfuu_yy = rk2Sfu_y * rk2vfu_y + rk2chi_max * rk2gtu_yy * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_yy) + vector(rk2Bfu_y, i, j, k) * vector(rk2Bfu_y, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_yy) + 2.0 * vector(rk2Bfu_y, i, j, k) * rk2vfu_y)) / rk2sdetg);
						rk2Tf4u_yy = rk2Sfuu_yy / rk2sdetg + (-(rk2Sfu_y * vector(rk2Betau_y, i, j, k)) / (rk2sdetg * vector(rk2Alpha, i, j, k))) + (-vector(rk2Betau_y, i, j, k) * rk2Tf4u_ty);
						rk2Tu_yy = rk2Tf4u_yy;
						rk2Sfuu_xy = 0.5 * rk2Sfu_x * rk2vfu_y + 0.5 * rk2Sfu_y * rk2vfu_x + rk2chi_max * rk2gtu_xy * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k) * vector(rk2Bfu_y, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xy) + vector(rk2Bfu_x, i, j, k) * rk2vfu_y + vector(rk2Bfu_y, i, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_yy = rk2inv_chi * (rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_yy * vector(rk2gtd_yy, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k));
						rk2Tf4u_xy = rk2Sfuu_xy / rk2sdetg + (-(rk2Sfu_x * vector(rk2Betau_y, i, j, k)) / (rk2sdetg * vector(rk2Alpha, i, j, k))) + (-vector(rk2Betau_x, i, j, k) * rk2Tf4u_ty);
						rk2Tu_xy = rk2Tf4u_xy;
						rk2Sfuu_xx = rk2Sfu_x * rk2vfu_x + rk2chi_max * rk2gtu_xx * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xx) + vector(rk2Bfu_x, i, j, k) * vector(rk2Bfu_x, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xx) + 2.0 * vector(rk2Bfu_x, i, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Tf4u_xx = rk2Sfuu_xx / rk2sdetg + (-(rk2Sfu_x * vector(rk2Betau_x, i, j, k)) / (rk2sdetg * vector(rk2Alpha, i, j, k))) + (-vector(rk2Betau_x, i, j, k) * rk2Tf4u_tx);
						rk2Tu_xx = rk2Tf4u_xx;
						rk2Sfuu_xz = 0.5 * rk2Sfu_x * rk2vfu_z + 0.5 * rk2Sfu_z * rk2vfu_x + rk2chi_max * rk2gtu_xz * rk2sdetg * vector(pf, i, j, k) + (-((-0.5 * rk2sqBf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k) * vector(rk2Bfu_z, i, j, k)) / (rk2sdetg * rk2sqW)) + (-0.5 * (rk2Bfvf * ((-rk2Bfvf * rk2chi_max * rk2gtu_xz) + vector(rk2Bfu_x, i, j, k) * rk2vfu_z + vector(rk2Bfu_z, i, j, k) * rk2vfu_x)) / rk2sdetg);
						rk2Sfud_xx = rk2inv_chi * (rk2Sfuu_xx * vector(rk2gtd_xx, i, j, k) + rk2Sfuu_xy * vector(rk2gtd_xy, i, j, k) + rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k));
						rk2Tf4u_xz = rk2Sfuu_xz / rk2sdetg + (-(rk2Sfu_x * vector(rk2Betau_z, i, j, k)) / (rk2sdetg * vector(rk2Alpha, i, j, k))) + (-vector(rk2Betau_x, i, j, k) * rk2Tf4u_tz);
						rk2Tu_xz = rk2Tf4u_xz;
						rk2Sfud_zz = rk2inv_chi * (rk2Sfuu_xz * vector(rk2gtd_xz, i, j, k) + rk2Sfuu_yz * vector(rk2gtd_yz, i, j, k) + rk2Sfuu_zz * vector(rk2gtd_zz, i, j, k));
						rk2trSf = rk2Sfud_xx + rk2Sfud_yy + rk2Sfud_zz;
						rk2Tu_tt = rk2Tf4u_tt;
						rk2pTtd_ADM_xz = rk2Betatd_x * rk2Betatd_z * rk2Tu_tt + rk2Betatd_x * (vector(rk2gtd_xz, i, j, k) * rk2Tu_tx + vector(rk2gtd_yz, i, j, k) * rk2Tu_ty + vector(rk2gtd_zz, i, j, k) * rk2Tu_tz) + rk2Betatd_z * (vector(rk2gtd_xx, i, j, k) * rk2Tu_tx + vector(rk2gtd_xy, i, j, k) * rk2Tu_ty + vector(rk2gtd_xz, i, j, k) * rk2Tu_tz) + vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_xz, i, j, k) * rk2Tu_xx + vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yz, i, j, k) * rk2Tu_xy + vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_zz, i, j, k) * rk2Tu_xz + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k) * rk2Tu_xy + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_yz, i, j, k) * rk2Tu_yy + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_zz, i, j, k) * rk2Tu_yz + (vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k)) * rk2Tu_xz + vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_yz, i, j, k) * rk2Tu_yz + vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_zz, i, j, k) * rk2Tu_zz;
						rk2pTtd_ADM_xy = rk2Betatd_x * rk2Betatd_y * rk2Tu_tt + rk2Betatd_x * (vector(rk2gtd_xy, i, j, k) * rk2Tu_tx + vector(rk2gtd_yy, i, j, k) * rk2Tu_ty + vector(rk2gtd_yz, i, j, k) * rk2Tu_tz) + rk2Betatd_y * (vector(rk2gtd_xx, i, j, k) * rk2Tu_tx + vector(rk2gtd_xy, i, j, k) * rk2Tu_ty + vector(rk2gtd_xz, i, j, k) * rk2Tu_tz) + vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_xy, i, j, k) * rk2Tu_xx + vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yy, i, j, k) * rk2Tu_xy + vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_yz, i, j, k) * rk2Tu_xz + (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k)) * rk2Tu_xy + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_yy, i, j, k) * rk2Tu_yy + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_yz, i, j, k) * rk2Tu_yz + vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xy, i, j, k) * rk2Tu_xz + vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_yy, i, j, k) * rk2Tu_yz + vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_yz, i, j, k) * rk2Tu_zz;
						rk2pTtd_ADM_xx = (rk2Betatd_x * rk2Betatd_x) * rk2Tu_tt + 2.0 * rk2Betatd_x * (vector(rk2gtd_xx, i, j, k) * rk2Tu_tx + vector(rk2gtd_xy, i, j, k) * rk2Tu_ty + vector(rk2gtd_xz, i, j, k) * rk2Tu_tz) + (vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_xx, i, j, k)) * rk2Tu_xx + 2.0 * vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_xy, i, j, k) * rk2Tu_xy + 2.0 * vector(rk2gtd_xx, i, j, k) * vector(rk2gtd_xz, i, j, k) * rk2Tu_xz + (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k)) * rk2Tu_yy + 2.0 * vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k) * rk2Tu_yz + (vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k)) * rk2Tu_zz;
						rk2d_Psi1_xx_o0_t5_m0_l0 = (-8.0) * M_PI * rk2inv_chi * vector(rk2Alpha, i, j, k) * rk2pTtd_ADM_xx + vector(rk2Alpha, i, j, k) * rk2Rpd_xx + rk2chi_max * vector(rk2Alpha, i, j, k) * rk2Rtd_xx;
						rk2m_Psi1_xx_o0_t4_l0 = (-rk2chi_max * rk2inv_chi) * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Psi1_xx_o0_t3_l0 = rk2chi_max * rk2Ct_zxx * d_Psi1_xx_o0_t3_m0_l0;
						rk2m_Psi1_xx_o0_t2_l0 = rk2chi_max * rk2Ct_yxx * d_Psi1_xx_o0_t2_m0_l0;
						rk2m_Psi1_xx_o0_t1_l0 = rk2chi_max * rk2Ct_xxx * d_Psi1_xx_o0_t1_m0_l0;
						rk2m_Psi1_xx_o0_t0_l0 = (-rk2chi_max) * d_Psi1_xx_o0_t0_m0_l0;
						rk2Psi1_xx = ((((rk2m_Psi1_xx_o0_t0_l0 + rk2m_Psi1_xx_o0_t1_l0) + rk2m_Psi1_xx_o0_t2_l0) + rk2m_Psi1_xx_o0_t3_l0) + rk2m_Psi1_xx_o0_t4_l0) + rk2d_Psi1_xx_o0_t5_m0_l0;
						rk2d_Psi1_xy_o0_t6_m0_l0 = (-8.0) * M_PI * rk2inv_chi * vector(rk2Alpha, i, j, k) * rk2pTtd_ADM_xy + vector(rk2Alpha, i, j, k) * rk2Rpd_xy + rk2chi_max * vector(rk2Alpha, i, j, k) * rk2Rtd_xy;
						rk2m_Psi1_xy_o0_t5_l0 = (-0.5 * rk2chi_max * rk2inv_chi) * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Psi1_xy_o0_t4_l0 = (-0.5 * rk2chi_max * rk2inv_chi) * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Psi1_xy_o0_t3_l0 = rk2chi_max * rk2Ct_zxy * d_Psi1_xx_o0_t3_m0_l0;
						rk2m_Psi1_xy_o0_t2_l0 = rk2chi_max * rk2Ct_yxy * d_Psi1_xx_o0_t2_m0_l0;
						rk2m_Psi1_xy_o0_t1_l0 = rk2chi_max * rk2Ct_xxy * d_Psi1_xx_o0_t1_m0_l0;
						rk2m_Psi1_xy_o0_t0_l0 = (-rk2chi_max) * d_Psi1_xy_o0_t0_m0_l0;
						rk2Psi1_xy = (((((rk2m_Psi1_xy_o0_t0_l0 + rk2m_Psi1_xy_o0_t1_l0) + rk2m_Psi1_xy_o0_t2_l0) + rk2m_Psi1_xy_o0_t3_l0) + rk2m_Psi1_xy_o0_t4_l0) + rk2m_Psi1_xy_o0_t5_l0) + rk2d_Psi1_xy_o0_t6_m0_l0;
						rk2d_Psi1_xz_o0_t6_m0_l0 = (-8.0) * M_PI * rk2inv_chi * vector(rk2Alpha, i, j, k) * rk2pTtd_ADM_xz + vector(rk2Alpha, i, j, k) * rk2Rpd_xz + rk2chi_max * vector(rk2Alpha, i, j, k) * rk2Rtd_xz;
						rk2m_Psi1_xz_o0_t5_l0 = (-0.5 * rk2chi_max * rk2inv_chi) * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk2m_Psi1_xz_o0_t4_l0 = (-0.5 * rk2chi_max * rk2inv_chi) * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Psi1_xz_o0_t3_l0 = rk2chi_max * rk2Ct_zxz * d_Psi1_xx_o0_t3_m0_l0;
						rk2m_Psi1_xz_o0_t2_l0 = rk2chi_max * rk2Ct_yxz * d_Psi1_xx_o0_t2_m0_l0;
						rk2m_Psi1_xz_o0_t1_l0 = rk2chi_max * rk2Ct_xxz * d_Psi1_xx_o0_t1_m0_l0;
						rk2m_Psi1_xz_o0_t0_l0 = (-rk2chi_max) * d_Psi1_xz_o0_t0_m0_l0;
						rk2Psi1_xz = (((((rk2m_Psi1_xz_o0_t0_l0 + rk2m_Psi1_xz_o0_t1_l0) + rk2m_Psi1_xz_o0_t2_l0) + rk2m_Psi1_xz_o0_t3_l0) + rk2m_Psi1_xz_o0_t4_l0) + rk2m_Psi1_xz_o0_t5_l0) + rk2d_Psi1_xz_o0_t6_m0_l0;
						rk2Jtd_ADM_z = vector(rk2Alpha, i, j, k) * ((vector(rk2Betau_x, i, j, k) * rk2Tu_tt + rk2Tu_tx) * vector(rk2gtd_xz, i, j, k) + (vector(rk2Betau_y, i, j, k) * rk2Tu_tt + rk2Tu_ty) * vector(rk2gtd_yz, i, j, k) + (vector(rk2Betau_z, i, j, k) * rk2Tu_tt + rk2Tu_tz) * vector(rk2gtd_zz, i, j, k));
						rk2pTtd_ADM_zz = (rk2Betatd_z * rk2Betatd_z) * rk2Tu_tt + 2.0 * rk2Betatd_z * (vector(rk2gtd_xz, i, j, k) * rk2Tu_tx + vector(rk2gtd_yz, i, j, k) * rk2Tu_ty + vector(rk2gtd_zz, i, j, k) * rk2Tu_tz) + (vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_xz, i, j, k)) * rk2Tu_xx + 2.0 * vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_yz, i, j, k) * rk2Tu_xy + 2.0 * vector(rk2gtd_xz, i, j, k) * vector(rk2gtd_zz, i, j, k) * rk2Tu_xz + (vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k)) * rk2Tu_yy + 2.0 * vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_zz, i, j, k) * rk2Tu_yz + (vector(rk2gtd_zz, i, j, k) * vector(rk2gtd_zz, i, j, k)) * rk2Tu_zz;
						rk2d_Psi1_zz_o0_t5_m0_l0 = (-8.0) * M_PI * rk2inv_chi * vector(rk2Alpha, i, j, k) * rk2pTtd_ADM_zz + vector(rk2Alpha, i, j, k) * rk2Rpd_zz + rk2chi_max * vector(rk2Alpha, i, j, k) * rk2Rtd_zz;
						rk2m_Psi1_zz_o0_t4_l0 = (-rk2chi_max * rk2inv_chi) * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Psi1_zz_o0_t3_l0 = rk2chi_max * rk2Ct_zzz * d_Psi1_xx_o0_t3_m0_l0;
						rk2m_Psi1_zz_o0_t2_l0 = rk2chi_max * rk2Ct_yzz * d_Psi1_xx_o0_t2_m0_l0;
						rk2m_Psi1_zz_o0_t1_l0 = rk2chi_max * rk2Ct_xzz * d_Psi1_xx_o0_t1_m0_l0;
						rk2m_Psi1_zz_o0_t0_l0 = (-rk2chi_max) * d_Psi1_zz_o0_t0_m0_l0;
						rk2Psi1_zz = ((((rk2m_Psi1_zz_o0_t0_l0 + rk2m_Psi1_zz_o0_t1_l0) + rk2m_Psi1_zz_o0_t2_l0) + rk2m_Psi1_zz_o0_t3_l0) + rk2m_Psi1_zz_o0_t4_l0) + rk2d_Psi1_zz_o0_t5_m0_l0;
						rk2Jtd_ADM_y = vector(rk2Alpha, i, j, k) * ((vector(rk2Betau_x, i, j, k) * rk2Tu_tt + rk2Tu_tx) * vector(rk2gtd_xy, i, j, k) + (vector(rk2Betau_y, i, j, k) * rk2Tu_tt + rk2Tu_ty) * vector(rk2gtd_yy, i, j, k) + (vector(rk2Betau_z, i, j, k) * rk2Tu_tt + rk2Tu_tz) * vector(rk2gtd_yz, i, j, k));
						rk2Jtd_ADM_x = vector(rk2Alpha, i, j, k) * ((vector(rk2Betau_x, i, j, k) * rk2Tu_tt + rk2Tu_tx) * vector(rk2gtd_xx, i, j, k) + (vector(rk2Betau_y, i, j, k) * rk2Tu_tt + rk2Tu_ty) * vector(rk2gtd_xy, i, j, k) + (vector(rk2Betau_z, i, j, k) * rk2Tu_tt + rk2Tu_tz) * vector(rk2gtd_xz, i, j, k));
						rk2pTtd_ADM_yz = rk2Betatd_y * rk2Betatd_z * rk2Tu_tt + rk2Betatd_y * (vector(rk2gtd_xz, i, j, k) * rk2Tu_tx + vector(rk2gtd_yz, i, j, k) * rk2Tu_ty + vector(rk2gtd_zz, i, j, k) * rk2Tu_tz) + rk2Betatd_z * (vector(rk2gtd_xy, i, j, k) * rk2Tu_tx + vector(rk2gtd_yy, i, j, k) * rk2Tu_ty + vector(rk2gtd_yz, i, j, k) * rk2Tu_tz) + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xz, i, j, k) * rk2Tu_xx + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_yz, i, j, k) * rk2Tu_xy + vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_zz, i, j, k) * rk2Tu_xz + vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_xz, i, j, k) * rk2Tu_xy + vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_yz, i, j, k) * rk2Tu_yy + vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_zz, i, j, k) * rk2Tu_yz + vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_xz, i, j, k) * rk2Tu_xz + (vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k)) * rk2Tu_yz + vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_zz, i, j, k) * rk2Tu_zz;
						rk2d_Psi1_yz_o0_t6_m0_l0 = (-8.0) * M_PI * rk2inv_chi * vector(rk2Alpha, i, j, k) * rk2pTtd_ADM_yz + vector(rk2Alpha, i, j, k) * rk2Rpd_yz + rk2chi_max * vector(rk2Alpha, i, j, k) * rk2Rtd_yz;
						rk2m_Psi1_yz_o0_t5_l0 = (-0.5 * rk2chi_max * rk2inv_chi) * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Psi1_yz_o0_t4_l0 = (-0.5 * rk2chi_max * rk2inv_chi) * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk2m_Psi1_yz_o0_t3_l0 = rk2chi_max * rk2Ct_zyz * d_Psi1_xx_o0_t3_m0_l0;
						rk2m_Psi1_yz_o0_t2_l0 = rk2chi_max * rk2Ct_yyz * d_Psi1_xx_o0_t2_m0_l0;
						rk2m_Psi1_yz_o0_t1_l0 = rk2chi_max * rk2Ct_xyz * d_Psi1_xx_o0_t1_m0_l0;
						rk2m_Psi1_yz_o0_t0_l0 = (-rk2chi_max) * d_Psi1_yz_o0_t0_m0_l0;
						rk2Psi1_yz = (((((rk2m_Psi1_yz_o0_t0_l0 + rk2m_Psi1_yz_o0_t1_l0) + rk2m_Psi1_yz_o0_t2_l0) + rk2m_Psi1_yz_o0_t3_l0) + rk2m_Psi1_yz_o0_t4_l0) + rk2m_Psi1_yz_o0_t5_l0) + rk2d_Psi1_yz_o0_t6_m0_l0;
						rk2pTtd_ADM_yy = (rk2Betatd_y * rk2Betatd_y) * rk2Tu_tt + 2.0 * rk2Betatd_y * (vector(rk2gtd_xy, i, j, k) * rk2Tu_tx + vector(rk2gtd_yy, i, j, k) * rk2Tu_ty + vector(rk2gtd_yz, i, j, k) * rk2Tu_tz) + (vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_xy, i, j, k)) * rk2Tu_xx + 2.0 * vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_yy, i, j, k) * rk2Tu_xy + 2.0 * vector(rk2gtd_xy, i, j, k) * vector(rk2gtd_yz, i, j, k) * rk2Tu_xz + (vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_yy, i, j, k)) * rk2Tu_yy + 2.0 * vector(rk2gtd_yy, i, j, k) * vector(rk2gtd_yz, i, j, k) * rk2Tu_yz + (vector(rk2gtd_yz, i, j, k) * vector(rk2gtd_yz, i, j, k)) * rk2Tu_zz;
						rk2tr_pT = rk2inv_chi * (rk2gtu_xx * rk2pTtd_ADM_xx + 2.0 * rk2gtu_xy * rk2pTtd_ADM_xy + 2.0 * rk2gtu_xz * rk2pTtd_ADM_xz + rk2gtu_yy * rk2pTtd_ADM_yy + 2.0 * rk2gtu_yz * rk2pTtd_ADM_yz + rk2gtu_zz * rk2pTtd_ADM_zz);
						rk2d_Psi1_yy_o0_t5_m0_l0 = (-8.0) * M_PI * rk2inv_chi * vector(rk2Alpha, i, j, k) * rk2pTtd_ADM_yy + vector(rk2Alpha, i, j, k) * rk2Rpd_yy + rk2chi_max * vector(rk2Alpha, i, j, k) * rk2Rtd_yy;
						rk2m_Psi1_yy_o0_t4_l0 = (-rk2chi_max * rk2inv_chi) * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk2m_Psi1_yy_o0_t3_l0 = rk2chi_max * rk2Ct_zyy * d_Psi1_xx_o0_t3_m0_l0;
						rk2m_Psi1_yy_o0_t2_l0 = rk2chi_max * rk2Ct_yyy * d_Psi1_xx_o0_t2_m0_l0;
						rk2m_Psi1_yy_o0_t1_l0 = rk2chi_max * rk2Ct_xyy * d_Psi1_xx_o0_t1_m0_l0;
						rk2m_Psi1_yy_o0_t0_l0 = (-rk2chi_max) * d_Psi1_yy_o0_t0_m0_l0;
						rk2Psi1_yy = ((((rk2m_Psi1_yy_o0_t0_l0 + rk2m_Psi1_yy_o0_t1_l0) + rk2m_Psi1_yy_o0_t2_l0) + rk2m_Psi1_yy_o0_t3_l0) + rk2m_Psi1_yy_o0_t4_l0) + rk2d_Psi1_yy_o0_t5_m0_l0;
						rk2trPsi1 = rk2Psi1_xx * rk2gtu_xx + 2.0 * rk2Psi1_xy * rk2gtu_xy + 2.0 * rk2Psi1_xz * rk2gtu_xz + rk2Psi1_yy * rk2gtu_yy + 2.0 * rk2Psi1_yz * rk2gtu_yz + rk2Psi1_zz * rk2gtu_zz;
						rk2Psi1TF_yz = (-0.3333333333333333 * rk2trPsi1 * vector(rk2gtd_yz, i, j, k)) + rk2Psi1_yz;
						rk2Psi1TF_yy = (-0.3333333333333333 * rk2trPsi1 * vector(rk2gtd_yy, i, j, k)) + rk2Psi1_yy;
						rk2Psi1TF_xy = (-0.3333333333333333 * rk2trPsi1 * vector(rk2gtd_xy, i, j, k)) + rk2Psi1_xy;
						rk2Psi1TF_xx = (-0.3333333333333333 * rk2trPsi1 * vector(rk2gtd_xx, i, j, k)) + rk2Psi1_xx;
						rk2Psi1TF_xz = (-0.3333333333333333 * rk2trPsi1 * vector(rk2gtd_xz, i, j, k)) + rk2Psi1_xz;
						rk2Psi1TF_zz = (-0.3333333333333333 * rk2trPsi1 * vector(rk2gtd_zz, i, j, k)) + rk2Psi1_zz;
						rk2kappa_cc = MIN(p_kappa_cc, p_kappa_cc * rk2decay_factor) / vector(rk2Alpha, i, j, k);
						rk2rho_ADM = (vector(rk2Alpha, i, j, k) * vector(rk2Alpha, i, j, k)) * rk2Tu_tt;
						d_Sfd_x_o0_t11_m0_l0 = vector(rk2Alpha, i, j, k) * rk2sdetg * rk2sqW * vector(qnu, i, j, k) * rk2vfu_x;
						d_Sfd_y_o0_t11_m0_l0 = vector(rk2Alpha, i, j, k) * rk2sdetg * rk2sqW * vector(qnu, i, j, k) * rk2vfu_y;
						d_Sfd_z_o0_t11_m0_l0 = vector(rk2Alpha, i, j, k) * rk2sdetg * rk2sqW * vector(qnu, i, j, k) * rk2vfu_z;
						d_DYf_o0_t0_m0_l0 = vector(rk2Alpha, i, j, k) * rk2sdetg * vector(rnu, i, j, k);
						d_tauf_o0_t3_m0_l0 = vector(rk2Alpha, i, j, k) * rk2sdetg * rk2sqW * vector(qnu, i, j, k);
						d_gtd_xx_o0_t3_m0_l0 = (-0.3333333333333333) * rk2kappa_cc * vector(rk2Alpha, i, j, k) * vector(rk2gtd_xx, i, j, k) * log(rk2detgtd) + (-2.0) * vector(rk2Alpha, i, j, k) * ((-0.3333333333333333 * rk2trAt * lambda_0 * vector(rk2gtd_xx, i, j, k)) + vector(rk2Atd_xx, i, j, k)) + (-0.6666666666666667) * vector(rk2gtd_xx, i, j, k) * rk2div_Beta;
						d_gtd_xy_o0_t6_m0_l0 = (-0.3333333333333333) * rk2kappa_cc * vector(rk2Alpha, i, j, k) * vector(rk2gtd_xy, i, j, k) * log(rk2detgtd) + (-2.0) * vector(rk2Alpha, i, j, k) * ((-0.3333333333333333 * rk2trAt * lambda_0 * vector(rk2gtd_xy, i, j, k)) + vector(rk2Atd_xy, i, j, k)) + (-0.6666666666666667) * vector(rk2gtd_xy, i, j, k) * rk2div_Beta;
						d_gtd_xz_o0_t6_m0_l0 = (-0.3333333333333333) * rk2kappa_cc * vector(rk2Alpha, i, j, k) * vector(rk2gtd_xz, i, j, k) * log(rk2detgtd) + (-2.0) * vector(rk2Alpha, i, j, k) * ((-0.3333333333333333 * rk2trAt * lambda_0 * vector(rk2gtd_xz, i, j, k)) + vector(rk2Atd_xz, i, j, k)) + (-0.6666666666666667) * vector(rk2gtd_xz, i, j, k) * rk2div_Beta;
						d_gtd_yy_o0_t3_m0_l0 = (-0.3333333333333333) * rk2kappa_cc * vector(rk2Alpha, i, j, k) * vector(rk2gtd_yy, i, j, k) * log(rk2detgtd) + (-2.0) * vector(rk2Alpha, i, j, k) * ((-0.3333333333333333 * rk2trAt * lambda_0 * vector(rk2gtd_yy, i, j, k)) + vector(rk2Atd_yy, i, j, k)) + (-0.6666666666666667) * vector(rk2gtd_yy, i, j, k) * rk2div_Beta;
						d_gtd_yz_o0_t6_m0_l0 = (-0.3333333333333333) * rk2kappa_cc * vector(rk2Alpha, i, j, k) * vector(rk2gtd_yz, i, j, k) * log(rk2detgtd) + (-2.0) * vector(rk2Alpha, i, j, k) * ((-0.3333333333333333 * rk2trAt * lambda_0 * vector(rk2gtd_yz, i, j, k)) + vector(rk2Atd_yz, i, j, k)) + (-0.6666666666666667) * vector(rk2gtd_yz, i, j, k) * rk2div_Beta;
						d_gtd_zz_o0_t3_m0_l0 = (-0.3333333333333333) * rk2kappa_cc * vector(rk2Alpha, i, j, k) * vector(rk2gtd_zz, i, j, k) * log(rk2detgtd) + (-2.0) * vector(rk2Alpha, i, j, k) * ((-0.3333333333333333 * rk2trAt * lambda_0 * vector(rk2gtd_zz, i, j, k)) + vector(rk2Atd_zz, i, j, k)) + (-0.6666666666666667) * vector(rk2gtd_zz, i, j, k) * rk2div_Beta;
						d_Atd_xx_o0_t3_m0_l0 = (-0.3333333333333333) * rk2kappa_cc * vector(rk2Alpha, i, j, k) * vector(rk2gtd_xx, i, j, k) * rk2trAt + vector(rk2Alpha, i, j, k) * (vector(rk2trK, i, j, k) * vector(rk2Atd_xx, i, j, k) + (-2.0 * vector(rk2Atd_xx, i, j, k) * rk2Atud_xx) + (-2.0 * vector(rk2Atd_xy, i, j, k) * rk2Atud_yx) + (-2.0 * vector(rk2Atd_xz, i, j, k) * rk2Atud_zx)) + rk2Psi1TF_xx + (-0.6666666666666667) * vector(rk2Atd_xx, i, j, k) * rk2div_Beta;
						d_Atd_xy_o0_t6_m0_l0 = (-0.3333333333333333) * rk2kappa_cc * vector(rk2Alpha, i, j, k) * vector(rk2gtd_xy, i, j, k) * rk2trAt + vector(rk2Alpha, i, j, k) * (vector(rk2trK, i, j, k) * vector(rk2Atd_xy, i, j, k) + (-2.0 * vector(rk2Atd_xx, i, j, k) * rk2Atud_xy) + (-2.0 * vector(rk2Atd_xy, i, j, k) * rk2Atud_yy) + (-2.0 * vector(rk2Atd_xz, i, j, k) * rk2Atud_zy)) + rk2Psi1TF_xy + (-0.6666666666666667) * vector(rk2Atd_xy, i, j, k) * rk2div_Beta;
						d_Atd_xz_o0_t6_m0_l0 = (-0.3333333333333333) * rk2kappa_cc * vector(rk2Alpha, i, j, k) * vector(rk2gtd_xz, i, j, k) * rk2trAt + vector(rk2Alpha, i, j, k) * (vector(rk2trK, i, j, k) * vector(rk2Atd_xz, i, j, k) + (-2.0 * vector(rk2Atd_xx, i, j, k) * rk2Atud_xz) + (-2.0 * vector(rk2Atd_xy, i, j, k) * rk2Atud_yz) + (-2.0 * vector(rk2Atd_xz, i, j, k) * rk2Atud_zz)) + rk2Psi1TF_xz + (-0.6666666666666667) * vector(rk2Atd_xz, i, j, k) * rk2div_Beta;
						d_Atd_yy_o0_t3_m0_l0 = (-0.3333333333333333) * rk2kappa_cc * vector(rk2Alpha, i, j, k) * vector(rk2gtd_yy, i, j, k) * rk2trAt + vector(rk2Alpha, i, j, k) * (vector(rk2trK, i, j, k) * vector(rk2Atd_yy, i, j, k) + (-2.0 * vector(rk2Atd_xy, i, j, k) * rk2Atud_xy) + (-2.0 * vector(rk2Atd_yy, i, j, k) * rk2Atud_yy) + (-2.0 * vector(rk2Atd_yz, i, j, k) * rk2Atud_zy)) + rk2Psi1TF_yy + (-0.6666666666666667) * vector(rk2Atd_yy, i, j, k) * rk2div_Beta;
						d_Atd_yz_o0_t6_m0_l0 = (-0.3333333333333333) * rk2kappa_cc * vector(rk2Alpha, i, j, k) * vector(rk2gtd_yz, i, j, k) * rk2trAt + vector(rk2Alpha, i, j, k) * (vector(rk2trK, i, j, k) * vector(rk2Atd_yz, i, j, k) + (-2.0 * vector(rk2Atd_xy, i, j, k) * rk2Atud_xz) + (-2.0 * vector(rk2Atd_yy, i, j, k) * rk2Atud_yz) + (-2.0 * vector(rk2Atd_yz, i, j, k) * rk2Atud_zz)) + rk2Psi1TF_yz + (-0.6666666666666667) * vector(rk2Atd_yz, i, j, k) * rk2div_Beta;
						d_Atd_zz_o0_t3_m0_l0 = (-0.3333333333333333) * rk2kappa_cc * vector(rk2Alpha, i, j, k) * vector(rk2gtd_zz, i, j, k) * rk2trAt + vector(rk2Alpha, i, j, k) * (vector(rk2trK, i, j, k) * vector(rk2Atd_zz, i, j, k) + (-2.0 * vector(rk2Atd_xz, i, j, k) * rk2Atud_xz) + (-2.0 * vector(rk2Atd_yz, i, j, k) * rk2Atud_yz) + (-2.0 * vector(rk2Atd_zz, i, j, k) * rk2Atud_zz)) + rk2Psi1TF_zz + (-0.6666666666666667) * vector(rk2Atd_zz, i, j, k) * rk2div_Beta;
						d_Gamh_x_o0_t24_m0_l0 = (-2.0) * vector(rk2Alpha, i, j, k) * rk2inv_chi * rk2Zu_x * (rk2kappa_z1 + 1.333333333333333 * vector(rk2theta, i, j, k) + 0.6666666666666667 * vector(rk2trK, i, j, k)) + 0.3333333333333333 * rk2gtu_xz * rk2d_div_Beta_z + 0.3333333333333333 * rk2gtu_xy * rk2d_div_Beta_y + 0.3333333333333333 * rk2gtu_xx * rk2d_div_Beta_x + (-50.26548245743669) * vector(rk2Alpha, i, j, k) * rk2inv_chi * (rk2gtu_xx * rk2Jtd_ADM_x + rk2gtu_xy * rk2Jtd_ADM_y + rk2gtu_xz * rk2Jtd_ADM_z) + 2.0 * vector(rk2Alpha, i, j, k) * (rk2Ct_xxx * rk2Atu_xx + 2.0 * rk2Ct_xxy * rk2Atu_xy + 2.0 * rk2Ct_xxz * rk2Atu_xz + rk2Ct_xyy * rk2Atu_yy + 2.0 * rk2Ct_xyz * rk2Atu_yz + rk2Ct_xzz * rk2Atu_zz) + 0.6666666666666667 * vector(rk2Gamh_x, i, j, k) * rk2div_Beta;
						d_Gamh_y_o0_t24_m0_l0 = (-2.0) * vector(rk2Alpha, i, j, k) * rk2inv_chi * rk2Zu_y * (rk2kappa_z1 + 1.333333333333333 * vector(rk2theta, i, j, k) + 0.6666666666666667 * vector(rk2trK, i, j, k)) + (-50.26548245743669) * vector(rk2Alpha, i, j, k) * rk2inv_chi * (rk2gtu_xy * rk2Jtd_ADM_x + rk2gtu_yy * rk2Jtd_ADM_y + rk2gtu_yz * rk2Jtd_ADM_z) + 0.3333333333333333 * rk2gtu_yy * rk2d_div_Beta_y + 0.3333333333333333 * rk2gtu_xy * rk2d_div_Beta_x + 2.0 * vector(rk2Alpha, i, j, k) * (rk2Ct_yxx * rk2Atu_xx + 2.0 * rk2Ct_yxy * rk2Atu_xy + 2.0 * rk2Ct_yxz * rk2Atu_xz + rk2Ct_yyy * rk2Atu_yy + 2.0 * rk2Ct_yyz * rk2Atu_yz + rk2Ct_yzz * rk2Atu_zz) + 0.6666666666666667 * vector(rk2Gamh_y, i, j, k) * rk2div_Beta + 0.3333333333333333 * rk2gtu_yz * rk2d_div_Beta_z;
						d_Gamh_z_o0_t24_m0_l0 = (-2.0) * vector(rk2Alpha, i, j, k) * rk2inv_chi * rk2Zu_z * (rk2kappa_z1 + 1.333333333333333 * vector(rk2theta, i, j, k) + 0.6666666666666667 * vector(rk2trK, i, j, k)) + (-50.26548245743669) * vector(rk2Alpha, i, j, k) * rk2inv_chi * (rk2gtu_xz * rk2Jtd_ADM_x + rk2gtu_yz * rk2Jtd_ADM_y + rk2gtu_zz * rk2Jtd_ADM_z) + 2.0 * vector(rk2Alpha, i, j, k) * (rk2Ct_zxx * rk2Atu_xx + 2.0 * rk2Ct_zxy * rk2Atu_xy + 2.0 * rk2Ct_zxz * rk2Atu_xz + rk2Ct_zyy * rk2Atu_yy + 2.0 * rk2Ct_zyz * rk2Atu_yz + rk2Ct_zzz * rk2Atu_zz) + 0.6666666666666667 * vector(rk2Gamh_z, i, j, k) * rk2div_Beta + 0.3333333333333333 * rk2gtu_zz * rk2d_div_Beta_z + 0.3333333333333333 * rk2gtu_yz * rk2d_div_Beta_y + 0.3333333333333333 * rk2gtu_xz * rk2d_div_Beta_x;
						d_Betau_x_o0_t0_m0_l0 = (Betau_x_0 - vector(rk2Betau_x, i, j, k)) * rk2feta + 0.75 * (vector(rk2Alpha, i, j, k) * lambda_f1 + lambda_f0) * vector(rk2Gamh_x, i, j, k);
						d_Betau_y_o0_t0_m0_l0 = (Betau_y_0 - vector(rk2Betau_y, i, j, k)) * rk2feta + 0.75 * (vector(rk2Alpha, i, j, k) * lambda_f1 + lambda_f0) * vector(rk2Gamh_y, i, j, k);
						d_Betau_z_o0_t0_m0_l0 = (Betau_z_0 - vector(rk2Betau_z, i, j, k)) * rk2feta + 0.75 * (vector(rk2Alpha, i, j, k) * lambda_f1 + lambda_f0) * vector(rk2Gamh_z, i, j, k);
						d_Alpha_o0_t0_m0_l0 = (-2.0) * (vector(rk2Alpha, i, j, k) * lambda_f3 + lambda_f2) * vector(rk2Alpha, i, j, k) * (vector(rk2trK, i, j, k) - trK0);
						d_chi_o0_t0_m0_l0 = 0.6666666666666667 * rk2chi_max * (vector(rk2Alpha, i, j, k) * (2.0 * vector(rk2theta, i, j, k) + vector(rk2trK, i, j, k)) - rk2div_Beta);
						d_trK_o0_t21_m0_l0 = vector(rk2Alpha, i, j, k) * rk2kappa_z1 * (1.0 - rk2kappa_z2) * vector(rk2theta, i, j, k) + 0.3333333333333333 * vector(rk2Alpha, i, j, k) * ((2.0 * vector(rk2theta, i, j, k) + vector(rk2trK, i, j, k)) * (2.0 * vector(rk2theta, i, j, k) + vector(rk2trK, i, j, k))) + vector(rk2Alpha, i, j, k) * (vector(rk2Atd_xx, i, j, k) * rk2Atu_xx + 2.0 * vector(rk2Atd_xy, i, j, k) * rk2Atu_xy + 2.0 * vector(rk2Atd_xz, i, j, k) * rk2Atu_xz + vector(rk2Atd_yy, i, j, k) * rk2Atu_yy + 2.0 * vector(rk2Atd_yz, i, j, k) * rk2Atu_yz + vector(rk2Atd_zz, i, j, k) * rk2Atu_zz) + 4.0 * vector(rk2Alpha, i, j, k) * M_PI * (rk2rho_ADM + rk2tr_pT);
						d_theta_o0_t3_m0_l0 = (-vector(rk2Alpha, i, j, k) * rk2kappa_z1 * (2.0 + rk2kappa_z2) * vector(rk2theta, i, j, k)) + (-8.0) * vector(rk2Alpha, i, j, k) * M_PI * rk2rho_ADM + 0.5 * vector(rk2Alpha, i, j, k) * (rk2Rscalar + 0.6666666666666667 * (vector(rk2trK, i, j, k) * vector(rk2trK, i, j, k)) + 0.6666666666666667 * vector(rk2theta, i, j, k) * ((-2.0 * vector(rk2theta, i, j, k)) + vector(rk2trK, i, j, k)) + (-vector(rk2Atd_xx, i, j, k) * rk2Atu_xx) + (-2.0 * vector(rk2Atd_xy, i, j, k) * rk2Atu_xy) + (-2.0 * vector(rk2Atd_xz, i, j, k) * rk2Atu_xz) + (-vector(rk2Atd_yy, i, j, k) * rk2Atu_yy) + (-2.0 * vector(rk2Atd_yz, i, j, k) * rk2Atu_yz) + (-vector(rk2Atd_zz, i, j, k) * rk2Atu_zz));
						RHS_Bfu_x = -vector(rk2phif, i, j, k) * vector(rk2Alpha, i, j, k) * rk2chi_max * rk2Gamt_x;
						RHS_Bfu_y = -vector(rk2phif, i, j, k) * vector(rk2Alpha, i, j, k) * rk2chi_max * rk2Gamt_y;
						RHS_Bfu_z = -vector(rk2phif, i, j, k) * vector(rk2Alpha, i, j, k) * rk2chi_max * rk2Gamt_z;
						RHS_tauf = (vector(rk2Alpha, i, j, k) * (rk2Sfuu_xx * vector(rk2Atd_xx, i, j, k) + 2.0 * rk2Sfuu_xy * vector(rk2Atd_xy, i, j, k) + 2.0 * rk2Sfuu_xz * vector(rk2Atd_xz, i, j, k) + rk2Sfuu_yy * vector(rk2Atd_yy, i, j, k) + 2.0 * rk2Sfuu_yz * vector(rk2Atd_yz, i, j, k) + rk2Sfuu_zz * vector(rk2Atd_zz, i, j, k))) / rk2chi_max + 0.3333333333333333 * vector(rk2Alpha, i, j, k) * rk2trSf * vector(rk2trK, i, j, k);
						RHS_phif = (-vector(rk2Alpha, i, j, k) * rk2kappa_f * vector(rk2phif, i, j, k)) + (-vector(rk2Alpha, i, j, k) * vector(rk2phif, i, j, k) * vector(rk2trK, i, j, k));
						RHS_Sfd_x = -FVM_i(FluxpiSfd_x, FluxniSfd_x, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_x = RHS_Sfd_x - FVM_j(FluxpjSfd_x, FluxnjSfd_x, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_x = RHS_Sfd_x - FVM_k(FluxpkSfd_x, FluxnkSfd_x, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_y = -FVM_i(FluxpiSfd_y, FluxniSfd_y, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_y = RHS_Sfd_y - FVM_j(FluxpjSfd_y, FluxnjSfd_y, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_y = RHS_Sfd_y - FVM_k(FluxpkSfd_y, FluxnkSfd_y, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_z = -FVM_i(FluxpiSfd_z, FluxniSfd_z, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_z = RHS_Sfd_z - FVM_j(FluxpjSfd_z, FluxnjSfd_z, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_z = RHS_Sfd_z - FVM_k(FluxpkSfd_z, FluxnkSfd_z, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_x = RHS_Bfu_x - FVM_i(FluxpiBfu_x, FluxniBfu_x, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_x = RHS_Bfu_x - FVM_j(FluxpjBfu_x, FluxnjBfu_x, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_x = RHS_Bfu_x - FVM_k(FluxpkBfu_x, FluxnkBfu_x, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_y = RHS_Bfu_y - FVM_i(FluxpiBfu_y, FluxniBfu_y, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_y = RHS_Bfu_y - FVM_j(FluxpjBfu_y, FluxnjBfu_y, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_y = RHS_Bfu_y - FVM_k(FluxpkBfu_y, FluxnkBfu_y, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_z = RHS_Bfu_z - FVM_i(FluxpiBfu_z, FluxniBfu_z, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_z = RHS_Bfu_z - FVM_j(FluxpjBfu_z, FluxnjBfu_z, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_z = RHS_Bfu_z - FVM_k(FluxpkBfu_z, FluxnkBfu_z, dx, simPlat_dt, ilast, jlast);
						RHS_Df = -FVM_i(FluxpiDf, FluxniDf, dx, simPlat_dt, ilast, jlast);
						RHS_Df = RHS_Df - FVM_j(FluxpjDf, FluxnjDf, dx, simPlat_dt, ilast, jlast);
						RHS_Df = RHS_Df - FVM_k(FluxpkDf, FluxnkDf, dx, simPlat_dt, ilast, jlast);
						RHS_DYf = -FVM_i(FluxpiDYf, FluxniDYf, dx, simPlat_dt, ilast, jlast);
						RHS_DYf = RHS_DYf - FVM_j(FluxpjDYf, FluxnjDYf, dx, simPlat_dt, ilast, jlast);
						RHS_DYf = RHS_DYf - FVM_k(FluxpkDYf, FluxnkDYf, dx, simPlat_dt, ilast, jlast);
						RHS_tauf = RHS_tauf - FVM_i(Fluxpitauf, Fluxnitauf, dx, simPlat_dt, ilast, jlast);
						RHS_tauf = RHS_tauf - FVM_j(Fluxpjtauf, Fluxnjtauf, dx, simPlat_dt, ilast, jlast);
						RHS_tauf = RHS_tauf - FVM_k(Fluxpktauf, Fluxnktauf, dx, simPlat_dt, ilast, jlast);
						RHS_phif = RHS_phif - FVM_i(Fluxpiphif, Fluxniphif, dx, simPlat_dt, ilast, jlast);
						RHS_phif = RHS_phif - FVM_j(Fluxpjphif, Fluxnjphif, dx, simPlat_dt, ilast, jlast);
						RHS_phif = RHS_phif - FVM_k(Fluxpkphif, Fluxnkphif, dx, simPlat_dt, ilast, jlast);
						m_theta_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_theta_o2_t2_m0_l0;
						m_theta_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_theta_o2_t1_m0_l0;
						m_theta_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_theta_o2_t0_m0_l0;
						m_theta_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_theta_o1_t2_m0_l0;
						m_theta_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_theta_o1_t1_m0_l0;
						m_theta_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_theta_o1_t0_m0_l0;
						m_theta_o0_t2_l0 = (-rk2Zu_z) * d_Psi1_xx_o0_t3_m0_l0;
						m_theta_o0_t1_l0 = (-rk2Zu_y) * d_Psi1_xx_o0_t2_m0_l0;
						m_theta_o0_t0_l0 = (-rk2Zu_x) * d_Psi1_xx_o0_t1_m0_l0;
						m_trK_o2_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_trK_o2_t2_m0_l0;
						m_trK_o2_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_trK_o2_t1_m0_l0;
						m_trK_o2_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_trK_o2_t0_m0_l0;
						m_trK_o1_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_trK_o1_t2_m0_l0;
						m_trK_o1_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_trK_o1_t1_m0_l0;
						m_trK_o1_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_trK_o1_t0_m0_l0;
						m_trK_o0_t20_l0 = 2.0 * rk2Zu_z * d_Psi1_xx_o0_t3_m0_l0;
						m_trK_o0_t19_l0 = 2.0 * rk2Zu_y * d_Psi1_xx_o0_t2_m0_l0;
						m_trK_o0_t18_l0 = 2.0 * rk2Zu_x * d_Psi1_xx_o0_t1_m0_l0;
						m_trK_o0_t17_l0 = (-2.0 * rk2chi_max * rk2gtu_xz) * d_Psi1_xz_o0_t0_m0_l0;
						m_trK_o0_t16_l0 = (-2.0 * rk2chi_max * rk2gtu_xy) * d_Psi1_xy_o0_t0_m0_l0;
						m_trK_o0_t15_l0 = 0.5 * rk2gtu_zz * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_trK_o0_t14_l0 = 0.5 * rk2gtu_yz * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_trK_o0_t13_l0 = 0.5 * rk2gtu_xz * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_trK_o0_t12_l0 = 0.5 * rk2gtu_yz * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_trK_o0_t11_l0 = 0.5 * rk2gtu_yy * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_trK_o0_t10_l0 = 0.5 * rk2gtu_xy * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_trK_o0_t9_l0 = 0.5 * rk2gtu_xz * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_trK_o0_t8_l0 = 0.5 * rk2gtu_xy * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_trK_o0_t7_l0 = 0.5 * rk2gtu_xx * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_trK_o0_t6_l0 = rk2chi_max * rk2Gamt_z * d_Psi1_xx_o0_t3_m0_l0;
						m_trK_o0_t5_l0 = rk2chi_max * rk2Gamt_y * d_Psi1_xx_o0_t2_m0_l0;
						m_trK_o0_t4_l0 = rk2chi_max * rk2Gamt_x * d_Psi1_xx_o0_t1_m0_l0;
						m_trK_o0_t3_l0 = (-rk2chi_max * rk2gtu_zz) * d_Psi1_zz_o0_t0_m0_l0;
						m_trK_o0_t2_l0 = (-2.0 * rk2chi_max * rk2gtu_yz) * d_Psi1_yz_o0_t0_m0_l0;
						m_trK_o0_t1_l0 = (-rk2chi_max * rk2gtu_yy) * d_Psi1_yy_o0_t0_m0_l0;
						m_trK_o0_t0_l0 = (-rk2chi_max * rk2gtu_xx) * d_Psi1_xx_o0_t0_m0_l0;
						m_chi_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_chi_o2_t2_m0_l0;
						m_chi_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_chi_o2_t1_m0_l0;
						m_chi_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_chi_o2_t0_m0_l0;
						m_chi_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_chi_o1_t2_m0_l0;
						m_chi_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_chi_o1_t1_m0_l0;
						m_chi_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_chi_o1_t0_m0_l0;
						m_Alpha_o2_t2_l0 = lambda_1 * MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Alpha_o2_t2_m0_l0;
						m_Alpha_o2_t1_l0 = lambda_1 * MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Alpha_o2_t1_m0_l0;
						m_Alpha_o2_t0_l0 = lambda_1 * MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Alpha_o2_t0_m0_l0;
						m_Alpha_o1_t2_l0 = lambda_1 * MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Alpha_o1_t2_m0_l0;
						m_Alpha_o1_t1_l0 = lambda_1 * MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Alpha_o1_t1_m0_l0;
						m_Alpha_o1_t0_l0 = lambda_1 * MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Alpha_o1_t0_m0_l0;
						m_Betau_z_o2_t2_l0 = lambda_2 * MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Betau_z_o2_t2_m0_l0;
						m_Betau_z_o2_t1_l0 = lambda_2 * MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Betau_z_o2_t1_m0_l0;
						m_Betau_z_o2_t0_l0 = lambda_2 * MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Betau_z_o2_t0_m0_l0;
						m_Betau_z_o1_t2_l0 = lambda_2 * MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Betau_z_o1_t2_m0_l0;
						m_Betau_z_o1_t1_l0 = lambda_2 * MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Betau_z_o1_t1_m0_l0;
						m_Betau_z_o1_t0_l0 = lambda_2 * MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Betau_z_o1_t0_m0_l0;
						m_Betau_y_o2_t2_l0 = lambda_2 * MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Betau_y_o2_t2_m0_l0;
						m_Betau_y_o2_t1_l0 = lambda_2 * MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Betau_y_o2_t1_m0_l0;
						m_Betau_y_o2_t0_l0 = lambda_2 * MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Betau_y_o2_t0_m0_l0;
						m_Betau_y_o1_t2_l0 = lambda_2 * MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Betau_y_o1_t2_m0_l0;
						m_Betau_y_o1_t1_l0 = lambda_2 * MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Betau_y_o1_t1_m0_l0;
						m_Betau_y_o1_t0_l0 = lambda_2 * MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Betau_y_o1_t0_m0_l0;
						m_Betau_x_o2_t2_l0 = lambda_2 * MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Betau_x_o2_t2_m0_l0;
						m_Betau_x_o2_t1_l0 = lambda_2 * MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Betau_x_o2_t1_m0_l0;
						m_Betau_x_o2_t0_l0 = lambda_2 * MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Betau_x_o2_t0_m0_l0;
						m_Betau_x_o1_t2_l0 = lambda_2 * MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Betau_x_o1_t2_m0_l0;
						m_Betau_x_o1_t1_l0 = lambda_2 * MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Betau_x_o1_t1_m0_l0;
						m_Betau_x_o1_t0_l0 = lambda_2 * MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Betau_x_o1_t0_m0_l0;
						m_Gamh_z_o2_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Gamh_z_o2_t2_m0_l0;
						m_Gamh_z_o2_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Gamh_z_o2_t1_m0_l0;
						m_Gamh_z_o2_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Gamh_z_o2_t0_m0_l0;
						m_Gamh_z_o1_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Gamh_z_o1_t2_m0_l0;
						m_Gamh_z_o1_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Gamh_z_o1_t1_m0_l0;
						m_Gamh_z_o1_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Gamh_z_o1_t0_m0_l0;
						m_Gamh_z_o0_t23_l0 = (-3.0 * vector(rk2Alpha, i, j, k) * rk2inv_chi * rk2Atu_zz) * d_Rpd_xx_o0_t2_m0_l0;
						m_Gamh_z_o0_t22_l0 = (-3.0 * vector(rk2Alpha, i, j, k) * rk2inv_chi * rk2Atu_yz) * d_Rpd_xx_o0_t1_m0_l0;
						m_Gamh_z_o0_t21_l0 = (-3.0 * vector(rk2Alpha, i, j, k) * rk2inv_chi * rk2Atu_xz) * d_Rpd_xx_o0_t0_m0_l0;
						m_Gamh_z_o0_t20_l0 = (-2.0 * vector(rk2theta, i, j, k) * rk2gtu_zz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_z_o0_t19_l0 = (-2.0 * vector(rk2theta, i, j, k) * rk2gtu_yz) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_z_o0_t18_l0 = (-2.0 * vector(rk2theta, i, j, k) * rk2gtu_xz) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_z_o0_t17_l0 = (-0.6666666666666667 * vector(rk2Alpha, i, j, k) * rk2gtu_zz) * d_Gamh_x_o0_t5_m0_l0;
						m_Gamh_z_o0_t16_l0 = (-0.6666666666666667 * vector(rk2Alpha, i, j, k) * rk2gtu_yz) * d_Gamh_x_o0_t4_m0_l0;
						m_Gamh_z_o0_t15_l0 = (-0.6666666666666667 * vector(rk2Alpha, i, j, k) * rk2gtu_xz) * d_Gamh_x_o0_t3_m0_l0;
						m_Gamh_z_o0_t14_l0 = (-1.333333333333333 * vector(rk2Alpha, i, j, k) * rk2gtu_zz) * d_Gamh_x_o0_t2_m0_l0;
						m_Gamh_z_o0_t13_l0 = (-1.333333333333333 * vector(rk2Alpha, i, j, k) * rk2gtu_yz) * d_Gamh_x_o0_t1_m0_l0;
						m_Gamh_z_o0_t12_l0 = (-1.333333333333333 * vector(rk2Alpha, i, j, k) * rk2gtu_xz) * d_Gamh_x_o0_t0_m0_l0;
						m_Gamh_z_o0_t11_l0 = (-vector(rk2Gamh_z, i, j, k)) * d_div_Beta_o0_t2_m0_l0;
						m_Gamh_z_o0_t10_l0 = (-vector(rk2Gamh_y, i, j, k)) * d_Sfd_y_o0_t9_m0_l0;
						m_Gamh_z_o0_t9_l0 = (-vector(rk2Gamh_x, i, j, k)) * d_Sfd_x_o0_t9_m0_l0;
						m_Gamh_z_o0_t8_l0 = (-2.0 * rk2Atu_zz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_z_o0_t7_l0 = (-2.0 * rk2Atu_yz) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_z_o0_t6_l0 = (-2.0 * rk2Atu_xz) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_z_o0_t5_l0 = rk2gtu_zz * d_d_div_Beta_z_o0_t2_m0_l0;
						m_Gamh_z_o0_t4_l0 = 2.0 * rk2gtu_yz * d_d_div_Beta_y_o0_t2_m0_l0;
						m_Gamh_z_o0_t3_l0 = rk2gtu_yy * d_Gamh_z_o0_t3_m0_l0;
						m_Gamh_z_o0_t2_l0 = 2.0 * rk2gtu_xz * d_d_div_Beta_x_o0_t2_m0_l0;
						m_Gamh_z_o0_t1_l0 = 2.0 * rk2gtu_xy * d_Gamh_z_o0_t1_m0_l0;
						m_Gamh_z_o0_t0_l0 = rk2gtu_xx * d_Gamh_z_o0_t0_m0_l0;
						m_Gamh_y_o2_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Gamh_y_o2_t2_m0_l0;
						m_Gamh_y_o2_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Gamh_y_o2_t1_m0_l0;
						m_Gamh_y_o2_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Gamh_y_o2_t0_m0_l0;
						m_Gamh_y_o1_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Gamh_y_o1_t2_m0_l0;
						m_Gamh_y_o1_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Gamh_y_o1_t1_m0_l0;
						m_Gamh_y_o1_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Gamh_y_o1_t0_m0_l0;
						m_Gamh_y_o0_t23_l0 = (-3.0 * vector(rk2Alpha, i, j, k) * rk2inv_chi * rk2Atu_yz) * d_Rpd_xx_o0_t2_m0_l0;
						m_Gamh_y_o0_t22_l0 = (-3.0 * vector(rk2Alpha, i, j, k) * rk2inv_chi * rk2Atu_yy) * d_Rpd_xx_o0_t1_m0_l0;
						m_Gamh_y_o0_t21_l0 = (-3.0 * vector(rk2Alpha, i, j, k) * rk2inv_chi * rk2Atu_xy) * d_Rpd_xx_o0_t0_m0_l0;
						m_Gamh_y_o0_t20_l0 = (-2.0 * vector(rk2theta, i, j, k) * rk2gtu_yz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_y_o0_t19_l0 = (-2.0 * vector(rk2theta, i, j, k) * rk2gtu_yy) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_y_o0_t18_l0 = (-2.0 * vector(rk2theta, i, j, k) * rk2gtu_xy) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_y_o0_t17_l0 = (-0.6666666666666667 * vector(rk2Alpha, i, j, k) * rk2gtu_yz) * d_Gamh_x_o0_t5_m0_l0;
						m_Gamh_y_o0_t16_l0 = (-0.6666666666666667 * vector(rk2Alpha, i, j, k) * rk2gtu_yy) * d_Gamh_x_o0_t4_m0_l0;
						m_Gamh_y_o0_t15_l0 = (-0.6666666666666667 * vector(rk2Alpha, i, j, k) * rk2gtu_xy) * d_Gamh_x_o0_t3_m0_l0;
						m_Gamh_y_o0_t14_l0 = (-1.333333333333333 * vector(rk2Alpha, i, j, k) * rk2gtu_yz) * d_Gamh_x_o0_t2_m0_l0;
						m_Gamh_y_o0_t13_l0 = (-1.333333333333333 * vector(rk2Alpha, i, j, k) * rk2gtu_yy) * d_Gamh_x_o0_t1_m0_l0;
						m_Gamh_y_o0_t12_l0 = (-1.333333333333333 * vector(rk2Alpha, i, j, k) * rk2gtu_xy) * d_Gamh_x_o0_t0_m0_l0;
						m_Gamh_y_o0_t11_l0 = rk2gtu_zz * d_Gamh_y_o0_t11_m0_l0;
						m_Gamh_y_o0_t10_l0 = 2.0 * rk2gtu_yz * d_d_div_Beta_z_o0_t1_m0_l0;
						m_Gamh_y_o0_t9_l0 = rk2gtu_yy * d_d_div_Beta_y_o0_t1_m0_l0;
						m_Gamh_y_o0_t8_l0 = 2.0 * rk2gtu_xz * d_Gamh_y_o0_t8_m0_l0;
						m_Gamh_y_o0_t7_l0 = 2.0 * rk2gtu_xy * d_d_div_Beta_x_o0_t1_m0_l0;
						m_Gamh_y_o0_t6_l0 = rk2gtu_xx * d_Gamh_y_o0_t6_m0_l0;
						m_Gamh_y_o0_t5_l0 = (-vector(rk2Gamh_z, i, j, k)) * d_Sfd_z_o0_t8_m0_l0;
						m_Gamh_y_o0_t4_l0 = (-vector(rk2Gamh_y, i, j, k)) * d_div_Beta_o0_t1_m0_l0;
						m_Gamh_y_o0_t3_l0 = (-vector(rk2Gamh_x, i, j, k)) * d_Sfd_x_o0_t8_m0_l0;
						m_Gamh_y_o0_t2_l0 = (-2.0 * rk2Atu_yz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_y_o0_t1_l0 = (-2.0 * rk2Atu_yy) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_y_o0_t0_l0 = (-2.0 * rk2Atu_xy) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_x_o2_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Gamh_x_o2_t2_m0_l0;
						m_Gamh_x_o2_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Gamh_x_o2_t1_m0_l0;
						m_Gamh_x_o2_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Gamh_x_o2_t0_m0_l0;
						m_Gamh_x_o1_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Gamh_x_o1_t2_m0_l0;
						m_Gamh_x_o1_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Gamh_x_o1_t1_m0_l0;
						m_Gamh_x_o1_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Gamh_x_o1_t0_m0_l0;
						m_Gamh_x_o0_t23_l0 = (-3.0 * vector(rk2Alpha, i, j, k) * rk2inv_chi * rk2Atu_xz) * d_Rpd_xx_o0_t2_m0_l0;
						m_Gamh_x_o0_t22_l0 = (-3.0 * vector(rk2Alpha, i, j, k) * rk2inv_chi * rk2Atu_xy) * d_Rpd_xx_o0_t1_m0_l0;
						m_Gamh_x_o0_t21_l0 = (-3.0 * vector(rk2Alpha, i, j, k) * rk2inv_chi * rk2Atu_xx) * d_Rpd_xx_o0_t0_m0_l0;
						m_Gamh_x_o0_t20_l0 = (-2.0 * rk2Atu_xz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_x_o0_t19_l0 = (-2.0 * rk2Atu_xy) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_x_o0_t18_l0 = (-2.0 * rk2Atu_xx) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_x_o0_t17_l0 = rk2gtu_zz * d_Gamh_x_o0_t17_m0_l0;
						m_Gamh_x_o0_t16_l0 = 2.0 * rk2gtu_yz * d_Gamh_x_o0_t16_m0_l0;
						m_Gamh_x_o0_t15_l0 = rk2gtu_yy * d_Gamh_x_o0_t15_m0_l0;
						m_Gamh_x_o0_t14_l0 = 2.0 * rk2gtu_xz * d_d_div_Beta_z_o0_t0_m0_l0;
						m_Gamh_x_o0_t13_l0 = 2.0 * rk2gtu_xy * d_d_div_Beta_y_o0_t0_m0_l0;
						m_Gamh_x_o0_t12_l0 = rk2gtu_xx * d_d_div_Beta_x_o0_t0_m0_l0;
						m_Gamh_x_o0_t11_l0 = (-vector(rk2Gamh_z, i, j, k)) * d_Sfd_z_o0_t7_m0_l0;
						m_Gamh_x_o0_t10_l0 = (-vector(rk2Gamh_y, i, j, k)) * d_Sfd_y_o0_t7_m0_l0;
						m_Gamh_x_o0_t9_l0 = (-vector(rk2Gamh_x, i, j, k)) * d_div_Beta_o0_t0_m0_l0;
						m_Gamh_x_o0_t8_l0 = (-2.0 * vector(rk2theta, i, j, k) * rk2gtu_xz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_x_o0_t7_l0 = (-2.0 * vector(rk2theta, i, j, k) * rk2gtu_xy) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_x_o0_t6_l0 = (-2.0 * vector(rk2theta, i, j, k) * rk2gtu_xx) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_x_o0_t5_l0 = (-0.6666666666666667 * vector(rk2Alpha, i, j, k) * rk2gtu_xz) * d_Gamh_x_o0_t5_m0_l0;
						m_Gamh_x_o0_t4_l0 = (-0.6666666666666667 * vector(rk2Alpha, i, j, k) * rk2gtu_xy) * d_Gamh_x_o0_t4_m0_l0;
						m_Gamh_x_o0_t3_l0 = (-0.6666666666666667 * vector(rk2Alpha, i, j, k) * rk2gtu_xx) * d_Gamh_x_o0_t3_m0_l0;
						m_Gamh_x_o0_t2_l0 = (-1.333333333333333 * vector(rk2Alpha, i, j, k) * rk2gtu_xz) * d_Gamh_x_o0_t2_m0_l0;
						m_Gamh_x_o0_t1_l0 = (-1.333333333333333 * vector(rk2Alpha, i, j, k) * rk2gtu_xy) * d_Gamh_x_o0_t1_m0_l0;
						m_Gamh_x_o0_t0_l0 = (-1.333333333333333 * vector(rk2Alpha, i, j, k) * rk2gtu_xx) * d_Gamh_x_o0_t0_m0_l0;
						m_Atd_zz_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Atd_zz_o2_t2_m0_l0;
						m_Atd_zz_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Atd_zz_o2_t1_m0_l0;
						m_Atd_zz_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Atd_zz_o2_t0_m0_l0;
						m_Atd_zz_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Atd_zz_o1_t2_m0_l0;
						m_Atd_zz_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Atd_zz_o1_t1_m0_l0;
						m_Atd_zz_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Atd_zz_o1_t0_m0_l0;
						m_Atd_zz_o0_t2_l0 = 2.0 * vector(rk2Atd_zz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Atd_zz_o0_t1_l0 = 2.0 * vector(rk2Atd_yz, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Atd_zz_o0_t0_l0 = 2.0 * vector(rk2Atd_xz, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Atd_yz_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Atd_yz_o2_t2_m0_l0;
						m_Atd_yz_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Atd_yz_o2_t1_m0_l0;
						m_Atd_yz_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Atd_yz_o2_t0_m0_l0;
						m_Atd_yz_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Atd_yz_o1_t2_m0_l0;
						m_Atd_yz_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Atd_yz_o1_t1_m0_l0;
						m_Atd_yz_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Atd_yz_o1_t0_m0_l0;
						m_Atd_yz_o0_t5_l0 = vector(rk2Atd_zz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Atd_yz_o0_t4_l0 = vector(rk2Atd_yz, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Atd_yz_o0_t3_l0 = vector(rk2Atd_xz, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Atd_yz_o0_t2_l0 = vector(rk2Atd_yz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Atd_yz_o0_t1_l0 = vector(rk2Atd_yy, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Atd_yz_o0_t0_l0 = vector(rk2Atd_xy, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Atd_yy_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Atd_yy_o2_t2_m0_l0;
						m_Atd_yy_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Atd_yy_o2_t1_m0_l0;
						m_Atd_yy_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Atd_yy_o2_t0_m0_l0;
						m_Atd_yy_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Atd_yy_o1_t2_m0_l0;
						m_Atd_yy_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Atd_yy_o1_t1_m0_l0;
						m_Atd_yy_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Atd_yy_o1_t0_m0_l0;
						m_Atd_yy_o0_t2_l0 = 2.0 * vector(rk2Atd_yz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Atd_yy_o0_t1_l0 = 2.0 * vector(rk2Atd_yy, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Atd_yy_o0_t0_l0 = 2.0 * vector(rk2Atd_xy, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Atd_xz_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Atd_xz_o2_t2_m0_l0;
						m_Atd_xz_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Atd_xz_o2_t1_m0_l0;
						m_Atd_xz_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Atd_xz_o2_t0_m0_l0;
						m_Atd_xz_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Atd_xz_o1_t2_m0_l0;
						m_Atd_xz_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Atd_xz_o1_t1_m0_l0;
						m_Atd_xz_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Atd_xz_o1_t0_m0_l0;
						m_Atd_xz_o0_t5_l0 = vector(rk2Atd_zz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Atd_xz_o0_t4_l0 = vector(rk2Atd_yz, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Atd_xz_o0_t3_l0 = vector(rk2Atd_xz, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_Atd_xz_o0_t2_l0 = vector(rk2Atd_xz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Atd_xz_o0_t1_l0 = vector(rk2Atd_xy, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Atd_xz_o0_t0_l0 = vector(rk2Atd_xx, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Atd_xy_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Atd_xy_o2_t2_m0_l0;
						m_Atd_xy_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Atd_xy_o2_t1_m0_l0;
						m_Atd_xy_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Atd_xy_o2_t0_m0_l0;
						m_Atd_xy_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Atd_xy_o1_t2_m0_l0;
						m_Atd_xy_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Atd_xy_o1_t1_m0_l0;
						m_Atd_xy_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Atd_xy_o1_t0_m0_l0;
						m_Atd_xy_o0_t5_l0 = vector(rk2Atd_yz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Atd_xy_o0_t4_l0 = vector(rk2Atd_yy, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Atd_xy_o0_t3_l0 = vector(rk2Atd_xy, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_Atd_xy_o0_t2_l0 = vector(rk2Atd_xz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Atd_xy_o0_t1_l0 = vector(rk2Atd_xy, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Atd_xy_o0_t0_l0 = vector(rk2Atd_xx, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Atd_xx_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_Atd_xx_o2_t2_m0_l0;
						m_Atd_xx_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_Atd_xx_o2_t1_m0_l0;
						m_Atd_xx_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_Atd_xx_o2_t0_m0_l0;
						m_Atd_xx_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_Atd_xx_o1_t2_m0_l0;
						m_Atd_xx_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_Atd_xx_o1_t1_m0_l0;
						m_Atd_xx_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_Atd_xx_o1_t0_m0_l0;
						m_Atd_xx_o0_t2_l0 = 2.0 * vector(rk2Atd_xz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Atd_xx_o0_t1_l0 = 2.0 * vector(rk2Atd_xy, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Atd_xx_o0_t0_l0 = 2.0 * vector(rk2Atd_xx, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_gtd_zz_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_gtd_zz_o2_t2_m0_l0;
						m_gtd_zz_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_gtd_zz_o2_t1_m0_l0;
						m_gtd_zz_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_gtd_zz_o2_t0_m0_l0;
						m_gtd_zz_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_gtd_zz_o1_t2_m0_l0;
						m_gtd_zz_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_gtd_zz_o1_t1_m0_l0;
						m_gtd_zz_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_gtd_zz_o1_t0_m0_l0;
						m_gtd_zz_o0_t2_l0 = 2.0 * vector(rk2gtd_zz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_gtd_zz_o0_t1_l0 = 2.0 * vector(rk2gtd_yz, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_gtd_zz_o0_t0_l0 = 2.0 * vector(rk2gtd_xz, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_gtd_yz_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_gtd_yz_o2_t2_m0_l0;
						m_gtd_yz_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_gtd_yz_o2_t1_m0_l0;
						m_gtd_yz_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_gtd_yz_o2_t0_m0_l0;
						m_gtd_yz_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_gtd_yz_o1_t2_m0_l0;
						m_gtd_yz_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_gtd_yz_o1_t1_m0_l0;
						m_gtd_yz_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_gtd_yz_o1_t0_m0_l0;
						m_gtd_yz_o0_t5_l0 = vector(rk2gtd_zz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_gtd_yz_o0_t4_l0 = vector(rk2gtd_yz, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_gtd_yz_o0_t3_l0 = vector(rk2gtd_xz, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_gtd_yz_o0_t2_l0 = vector(rk2gtd_yz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_gtd_yz_o0_t1_l0 = vector(rk2gtd_yy, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_gtd_yz_o0_t0_l0 = vector(rk2gtd_xy, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_gtd_yy_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_gtd_yy_o2_t2_m0_l0;
						m_gtd_yy_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_gtd_yy_o2_t1_m0_l0;
						m_gtd_yy_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_gtd_yy_o2_t0_m0_l0;
						m_gtd_yy_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_gtd_yy_o1_t2_m0_l0;
						m_gtd_yy_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_gtd_yy_o1_t1_m0_l0;
						m_gtd_yy_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_gtd_yy_o1_t0_m0_l0;
						m_gtd_yy_o0_t2_l0 = 2.0 * vector(rk2gtd_yz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_gtd_yy_o0_t1_l0 = 2.0 * vector(rk2gtd_yy, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_gtd_yy_o0_t0_l0 = 2.0 * vector(rk2gtd_xy, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_gtd_xz_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_gtd_xz_o2_t2_m0_l0;
						m_gtd_xz_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_gtd_xz_o2_t1_m0_l0;
						m_gtd_xz_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_gtd_xz_o2_t0_m0_l0;
						m_gtd_xz_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_gtd_xz_o1_t2_m0_l0;
						m_gtd_xz_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_gtd_xz_o1_t1_m0_l0;
						m_gtd_xz_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_gtd_xz_o1_t0_m0_l0;
						m_gtd_xz_o0_t5_l0 = vector(rk2gtd_zz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_gtd_xz_o0_t4_l0 = vector(rk2gtd_yz, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_gtd_xz_o0_t3_l0 = vector(rk2gtd_xz, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_gtd_xz_o0_t2_l0 = vector(rk2gtd_xz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_gtd_xz_o0_t1_l0 = vector(rk2gtd_xy, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_gtd_xz_o0_t0_l0 = vector(rk2gtd_xx, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_gtd_xy_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_gtd_xy_o2_t2_m0_l0;
						m_gtd_xy_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_gtd_xy_o2_t1_m0_l0;
						m_gtd_xy_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_gtd_xy_o2_t0_m0_l0;
						m_gtd_xy_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_gtd_xy_o1_t2_m0_l0;
						m_gtd_xy_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_gtd_xy_o1_t1_m0_l0;
						m_gtd_xy_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_gtd_xy_o1_t0_m0_l0;
						m_gtd_xy_o0_t5_l0 = vector(rk2gtd_yz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_gtd_xy_o0_t4_l0 = vector(rk2gtd_yy, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_gtd_xy_o0_t3_l0 = vector(rk2gtd_xy, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_gtd_xy_o0_t2_l0 = vector(rk2gtd_xz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_gtd_xy_o0_t1_l0 = vector(rk2gtd_xy, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_gtd_xy_o0_t0_l0 = vector(rk2gtd_xx, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_gtd_xx_o2_t2_l0 = MIN(0.0, vector(rk2Betau_z, i, j, k)) * d_gtd_xx_o2_t2_m0_l0;
						m_gtd_xx_o2_t1_l0 = MIN(0.0, vector(rk2Betau_y, i, j, k)) * d_gtd_xx_o2_t1_m0_l0;
						m_gtd_xx_o2_t0_l0 = MIN(0.0, vector(rk2Betau_x, i, j, k)) * d_gtd_xx_o2_t0_m0_l0;
						m_gtd_xx_o1_t2_l0 = MAX(0.0, vector(rk2Betau_z, i, j, k)) * d_gtd_xx_o1_t2_m0_l0;
						m_gtd_xx_o1_t1_l0 = MAX(0.0, vector(rk2Betau_y, i, j, k)) * d_gtd_xx_o1_t1_m0_l0;
						m_gtd_xx_o1_t0_l0 = MAX(0.0, vector(rk2Betau_x, i, j, k)) * d_gtd_xx_o1_t0_m0_l0;
						m_gtd_xx_o0_t2_l0 = 2.0 * vector(rk2gtd_xz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_gtd_xx_o0_t1_l0 = 2.0 * vector(rk2gtd_xy, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_gtd_xx_o0_t0_l0 = 2.0 * vector(rk2gtd_xx, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_phif_o0_t2_l0 = (ch * ch) * vector(rk2Bfu_z, i, j, k) * d_Psi1_xx_o0_t3_m0_l0;
						m_phif_o0_t1_l0 = (ch * ch) * vector(rk2Bfu_y, i, j, k) * d_Psi1_xx_o0_t2_m0_l0;
						m_phif_o0_t0_l0 = (ch * ch) * vector(rk2Bfu_x, i, j, k) * d_Psi1_xx_o0_t1_m0_l0;
						m_tauf_o0_t2_l0 = (-rk2Sfu_z) * d_Psi1_xx_o0_t3_m0_l0;
						m_tauf_o0_t1_l0 = (-rk2Sfu_y) * d_Psi1_xx_o0_t2_m0_l0;
						m_tauf_o0_t0_l0 = (-rk2Sfu_x) * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_z_o0_t5_l0 = vector(rk2phif, i, j, k) * rk2gtu_zz * rk2chi_max * d_Psi1_xx_o0_t3_m0_l0;
						m_Bfu_z_o0_t4_l0 = vector(rk2phif, i, j, k) * rk2gtu_yz * rk2chi_max * d_Psi1_xx_o0_t2_m0_l0;
						m_Bfu_z_o0_t3_l0 = vector(rk2phif, i, j, k) * rk2gtu_xz * rk2chi_max * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_z_o0_t2_l0 = (-0.5 * vector(rk2phif, i, j, k) * rk2gtu_zz * vector(rk2Alpha, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Bfu_z_o0_t1_l0 = (-0.5 * vector(rk2phif, i, j, k) * rk2gtu_yz * vector(rk2Alpha, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Bfu_z_o0_t0_l0 = (-0.5 * vector(rk2phif, i, j, k) * rk2gtu_xz * vector(rk2Alpha, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Bfu_y_o0_t5_l0 = vector(rk2phif, i, j, k) * rk2gtu_yz * rk2chi_max * d_Psi1_xx_o0_t3_m0_l0;
						m_Bfu_y_o0_t4_l0 = vector(rk2phif, i, j, k) * rk2gtu_yy * rk2chi_max * d_Psi1_xx_o0_t2_m0_l0;
						m_Bfu_y_o0_t3_l0 = vector(rk2phif, i, j, k) * rk2gtu_xy * rk2chi_max * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_y_o0_t2_l0 = (-0.5 * vector(rk2phif, i, j, k) * rk2gtu_yz * vector(rk2Alpha, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Bfu_y_o0_t1_l0 = (-0.5 * vector(rk2phif, i, j, k) * rk2gtu_yy * vector(rk2Alpha, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Bfu_y_o0_t0_l0 = (-0.5 * vector(rk2phif, i, j, k) * rk2gtu_xy * vector(rk2Alpha, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Bfu_x_o0_t5_l0 = vector(rk2phif, i, j, k) * rk2gtu_xz * rk2chi_max * d_Psi1_xx_o0_t3_m0_l0;
						m_Bfu_x_o0_t4_l0 = vector(rk2phif, i, j, k) * rk2gtu_xy * rk2chi_max * d_Psi1_xx_o0_t2_m0_l0;
						m_Bfu_x_o0_t3_l0 = vector(rk2phif, i, j, k) * rk2gtu_xx * rk2chi_max * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_x_o0_t2_l0 = (-0.5 * vector(rk2phif, i, j, k) * rk2gtu_xz * vector(rk2Alpha, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Bfu_x_o0_t1_l0 = (-0.5 * vector(rk2phif, i, j, k) * rk2gtu_xy * vector(rk2Alpha, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Bfu_x_o0_t0_l0 = (-0.5 * vector(rk2phif, i, j, k) * rk2gtu_xx * vector(rk2Alpha, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Sfd_z_o0_t10_l0 = (-(vector(rk2Df, i, j, k) + vector(rk2tauf, i, j, k))) * d_Psi1_xx_o0_t3_m0_l0;
						m_Sfd_z_o0_t9_l0 = vector(rk2Sfd_z, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Sfd_z_o0_t8_l0 = vector(rk2Sfd_y, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Sfd_z_o0_t7_l0 = vector(rk2Sfd_x, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Sfd_z_o0_t6_l0 = (-0.5 * (vector(rk2Alpha, i, j, k) * rk2trSf) / rk2chi_max) * d_Rpd_xx_o0_t2_m0_l0;
						m_Sfd_z_o0_t5_l0 = 0.5 * (vector(rk2Alpha, i, j, k) * rk2Sfuu_zz) / rk2chi_max * d_Ctd_zzz_o0_t0_m0_l0;
						m_Sfd_z_o0_t4_l0 = (vector(rk2Alpha, i, j, k) * rk2Sfuu_yz) / rk2chi_max * d_Ctd_yzz_o0_t0_m0_l0;
						m_Sfd_z_o0_t3_l0 = 0.5 * (vector(rk2Alpha, i, j, k) * rk2Sfuu_yy) / rk2chi_max * d_Ctd_yyz_o0_t0_m0_l0;
						m_Sfd_z_o0_t2_l0 = (vector(rk2Alpha, i, j, k) * rk2Sfuu_xz) / rk2chi_max * d_Ctd_xzz_o0_t0_m0_l0;
						m_Sfd_z_o0_t1_l0 = (vector(rk2Alpha, i, j, k) * rk2Sfuu_xy) / rk2chi_max * d_Ctd_xyz_o0_t1_m0_l0;
						m_Sfd_z_o0_t0_l0 = 0.5 * (vector(rk2Alpha, i, j, k) * rk2Sfuu_xx) / rk2chi_max * d_Ctd_xxz_o0_t0_m0_l0;
						m_Sfd_y_o0_t10_l0 = (-(vector(rk2Df, i, j, k) + vector(rk2tauf, i, j, k))) * d_Psi1_xx_o0_t2_m0_l0;
						m_Sfd_y_o0_t9_l0 = vector(rk2Sfd_z, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Sfd_y_o0_t8_l0 = vector(rk2Sfd_y, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Sfd_y_o0_t7_l0 = vector(rk2Sfd_x, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Sfd_y_o0_t6_l0 = (-0.5 * (vector(rk2Alpha, i, j, k) * rk2trSf) / rk2chi_max) * d_Rpd_xx_o0_t1_m0_l0;
						m_Sfd_y_o0_t5_l0 = 0.5 * (vector(rk2Alpha, i, j, k) * rk2Sfuu_zz) / rk2chi_max * d_Ctd_yzz_o0_t1_m0_l0;
						m_Sfd_y_o0_t4_l0 = (vector(rk2Alpha, i, j, k) * rk2Sfuu_yz) / rk2chi_max * d_Ctd_zyy_o0_t0_m0_l0;
						m_Sfd_y_o0_t3_l0 = 0.5 * (vector(rk2Alpha, i, j, k) * rk2Sfuu_yy) / rk2chi_max * d_Ctd_yyy_o0_t0_m0_l0;
						m_Sfd_y_o0_t2_l0 = (vector(rk2Alpha, i, j, k) * rk2Sfuu_xz) / rk2chi_max * d_Ctd_xyz_o0_t0_m0_l0;
						m_Sfd_y_o0_t1_l0 = (vector(rk2Alpha, i, j, k) * rk2Sfuu_xy) / rk2chi_max * d_Ctd_xyy_o0_t0_m0_l0;
						m_Sfd_y_o0_t0_l0 = 0.5 * (vector(rk2Alpha, i, j, k) * rk2Sfuu_xx) / rk2chi_max * d_Ctd_xxy_o0_t0_m0_l0;
						m_Sfd_x_o0_t10_l0 = (-(vector(rk2Df, i, j, k) + vector(rk2tauf, i, j, k))) * d_Psi1_xx_o0_t1_m0_l0;
						m_Sfd_x_o0_t9_l0 = vector(rk2Sfd_z, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Sfd_x_o0_t8_l0 = vector(rk2Sfd_y, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Sfd_x_o0_t7_l0 = vector(rk2Sfd_x, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_Sfd_x_o0_t6_l0 = (-0.5 * (vector(rk2Alpha, i, j, k) * rk2trSf) / rk2chi_max) * d_Rpd_xx_o0_t0_m0_l0;
						m_Sfd_x_o0_t5_l0 = 0.5 * (vector(rk2Alpha, i, j, k) * rk2Sfuu_zz) / rk2chi_max * d_Ctd_xzz_o0_t1_m0_l0;
						m_Sfd_x_o0_t4_l0 = (vector(rk2Alpha, i, j, k) * rk2Sfuu_yz) / rk2chi_max * d_Ctd_xyz_o0_t2_m0_l0;
						m_Sfd_x_o0_t3_l0 = 0.5 * (vector(rk2Alpha, i, j, k) * rk2Sfuu_yy) / rk2chi_max * d_Ctd_xyy_o0_t1_m0_l0;
						m_Sfd_x_o0_t2_l0 = (vector(rk2Alpha, i, j, k) * rk2Sfuu_xz) / rk2chi_max * d_Ctd_zxx_o0_t0_m0_l0;
						m_Sfd_x_o0_t1_l0 = (vector(rk2Alpha, i, j, k) * rk2Sfuu_xy) / rk2chi_max * d_Ctd_yxx_o0_t0_m0_l0;
						m_Sfd_x_o0_t0_l0 = 0.5 * (vector(rk2Alpha, i, j, k) * rk2Sfuu_xx) / rk2chi_max * d_Ctd_xxx_o0_t0_m0_l0;
						RHS_Sfd_x = RHS_Sfd_x + (((((((((((m_Sfd_x_o0_t0_l0 + m_Sfd_x_o0_t1_l0) + m_Sfd_x_o0_t2_l0) + m_Sfd_x_o0_t3_l0) + m_Sfd_x_o0_t4_l0) + m_Sfd_x_o0_t5_l0) + m_Sfd_x_o0_t6_l0) + m_Sfd_x_o0_t7_l0) + m_Sfd_x_o0_t8_l0) + m_Sfd_x_o0_t9_l0) + m_Sfd_x_o0_t10_l0) + d_Sfd_x_o0_t11_m0_l0);
						RHS_Sfd_y = RHS_Sfd_y + (((((((((((m_Sfd_y_o0_t0_l0 + m_Sfd_y_o0_t1_l0) + m_Sfd_y_o0_t2_l0) + m_Sfd_y_o0_t3_l0) + m_Sfd_y_o0_t4_l0) + m_Sfd_y_o0_t5_l0) + m_Sfd_y_o0_t6_l0) + m_Sfd_y_o0_t7_l0) + m_Sfd_y_o0_t8_l0) + m_Sfd_y_o0_t9_l0) + m_Sfd_y_o0_t10_l0) + d_Sfd_y_o0_t11_m0_l0);
						RHS_Sfd_z = RHS_Sfd_z + (((((((((((m_Sfd_z_o0_t0_l0 + m_Sfd_z_o0_t1_l0) + m_Sfd_z_o0_t2_l0) + m_Sfd_z_o0_t3_l0) + m_Sfd_z_o0_t4_l0) + m_Sfd_z_o0_t5_l0) + m_Sfd_z_o0_t6_l0) + m_Sfd_z_o0_t7_l0) + m_Sfd_z_o0_t8_l0) + m_Sfd_z_o0_t9_l0) + m_Sfd_z_o0_t10_l0) + d_Sfd_z_o0_t11_m0_l0);
						RHS_Bfu_x = RHS_Bfu_x + (((((m_Bfu_x_o0_t0_l0 + m_Bfu_x_o0_t1_l0) + m_Bfu_x_o0_t2_l0) + m_Bfu_x_o0_t3_l0) + m_Bfu_x_o0_t4_l0) + m_Bfu_x_o0_t5_l0);
						RHS_Bfu_y = RHS_Bfu_y + (((((m_Bfu_y_o0_t0_l0 + m_Bfu_y_o0_t1_l0) + m_Bfu_y_o0_t2_l0) + m_Bfu_y_o0_t3_l0) + m_Bfu_y_o0_t4_l0) + m_Bfu_y_o0_t5_l0);
						RHS_Bfu_z = RHS_Bfu_z + (((((m_Bfu_z_o0_t0_l0 + m_Bfu_z_o0_t1_l0) + m_Bfu_z_o0_t2_l0) + m_Bfu_z_o0_t3_l0) + m_Bfu_z_o0_t4_l0) + m_Bfu_z_o0_t5_l0);
						RHS_DYf = RHS_DYf + d_DYf_o0_t0_m0_l0;
						RHS_tauf = RHS_tauf + (((m_tauf_o0_t0_l0 + m_tauf_o0_t1_l0) + m_tauf_o0_t2_l0) + d_tauf_o0_t3_m0_l0);
						RHS_phif = RHS_phif + ((m_phif_o0_t0_l0 + m_phif_o0_t1_l0) + m_phif_o0_t2_l0);
						RHS_gtd_xx = ((((((((m_gtd_xx_o0_t0_l0 + m_gtd_xx_o0_t1_l0) + m_gtd_xx_o0_t2_l0) + d_gtd_xx_o0_t3_m0_l0) + m_gtd_xx_o1_t0_l0) + m_gtd_xx_o1_t1_l0) + m_gtd_xx_o1_t2_l0) + m_gtd_xx_o2_t0_l0) + m_gtd_xx_o2_t1_l0) + m_gtd_xx_o2_t2_l0;
						RHS_gtd_xy = (((((((((((m_gtd_xy_o0_t0_l0 + m_gtd_xy_o0_t1_l0) + m_gtd_xy_o0_t2_l0) + m_gtd_xy_o0_t3_l0) + m_gtd_xy_o0_t4_l0) + m_gtd_xy_o0_t5_l0) + d_gtd_xy_o0_t6_m0_l0) + m_gtd_xy_o1_t0_l0) + m_gtd_xy_o1_t1_l0) + m_gtd_xy_o1_t2_l0) + m_gtd_xy_o2_t0_l0) + m_gtd_xy_o2_t1_l0) + m_gtd_xy_o2_t2_l0;
						RHS_gtd_xz = (((((((((((m_gtd_xz_o0_t0_l0 + m_gtd_xz_o0_t1_l0) + m_gtd_xz_o0_t2_l0) + m_gtd_xz_o0_t3_l0) + m_gtd_xz_o0_t4_l0) + m_gtd_xz_o0_t5_l0) + d_gtd_xz_o0_t6_m0_l0) + m_gtd_xz_o1_t0_l0) + m_gtd_xz_o1_t1_l0) + m_gtd_xz_o1_t2_l0) + m_gtd_xz_o2_t0_l0) + m_gtd_xz_o2_t1_l0) + m_gtd_xz_o2_t2_l0;
						RHS_gtd_yy = ((((((((m_gtd_yy_o0_t0_l0 + m_gtd_yy_o0_t1_l0) + m_gtd_yy_o0_t2_l0) + d_gtd_yy_o0_t3_m0_l0) + m_gtd_yy_o1_t0_l0) + m_gtd_yy_o1_t1_l0) + m_gtd_yy_o1_t2_l0) + m_gtd_yy_o2_t0_l0) + m_gtd_yy_o2_t1_l0) + m_gtd_yy_o2_t2_l0;
						RHS_gtd_yz = (((((((((((m_gtd_yz_o0_t0_l0 + m_gtd_yz_o0_t1_l0) + m_gtd_yz_o0_t2_l0) + m_gtd_yz_o0_t3_l0) + m_gtd_yz_o0_t4_l0) + m_gtd_yz_o0_t5_l0) + d_gtd_yz_o0_t6_m0_l0) + m_gtd_yz_o1_t0_l0) + m_gtd_yz_o1_t1_l0) + m_gtd_yz_o1_t2_l0) + m_gtd_yz_o2_t0_l0) + m_gtd_yz_o2_t1_l0) + m_gtd_yz_o2_t2_l0;
						RHS_gtd_zz = ((((((((m_gtd_zz_o0_t0_l0 + m_gtd_zz_o0_t1_l0) + m_gtd_zz_o0_t2_l0) + d_gtd_zz_o0_t3_m0_l0) + m_gtd_zz_o1_t0_l0) + m_gtd_zz_o1_t1_l0) + m_gtd_zz_o1_t2_l0) + m_gtd_zz_o2_t0_l0) + m_gtd_zz_o2_t1_l0) + m_gtd_zz_o2_t2_l0;
						RHS_Atd_xx = ((((((((m_Atd_xx_o0_t0_l0 + m_Atd_xx_o0_t1_l0) + m_Atd_xx_o0_t2_l0) + d_Atd_xx_o0_t3_m0_l0) + m_Atd_xx_o1_t0_l0) + m_Atd_xx_o1_t1_l0) + m_Atd_xx_o1_t2_l0) + m_Atd_xx_o2_t0_l0) + m_Atd_xx_o2_t1_l0) + m_Atd_xx_o2_t2_l0;
						RHS_Atd_xy = (((((((((((m_Atd_xy_o0_t0_l0 + m_Atd_xy_o0_t1_l0) + m_Atd_xy_o0_t2_l0) + m_Atd_xy_o0_t3_l0) + m_Atd_xy_o0_t4_l0) + m_Atd_xy_o0_t5_l0) + d_Atd_xy_o0_t6_m0_l0) + m_Atd_xy_o1_t0_l0) + m_Atd_xy_o1_t1_l0) + m_Atd_xy_o1_t2_l0) + m_Atd_xy_o2_t0_l0) + m_Atd_xy_o2_t1_l0) + m_Atd_xy_o2_t2_l0;
						RHS_Atd_xz = (((((((((((m_Atd_xz_o0_t0_l0 + m_Atd_xz_o0_t1_l0) + m_Atd_xz_o0_t2_l0) + m_Atd_xz_o0_t3_l0) + m_Atd_xz_o0_t4_l0) + m_Atd_xz_o0_t5_l0) + d_Atd_xz_o0_t6_m0_l0) + m_Atd_xz_o1_t0_l0) + m_Atd_xz_o1_t1_l0) + m_Atd_xz_o1_t2_l0) + m_Atd_xz_o2_t0_l0) + m_Atd_xz_o2_t1_l0) + m_Atd_xz_o2_t2_l0;
						RHS_Atd_yy = ((((((((m_Atd_yy_o0_t0_l0 + m_Atd_yy_o0_t1_l0) + m_Atd_yy_o0_t2_l0) + d_Atd_yy_o0_t3_m0_l0) + m_Atd_yy_o1_t0_l0) + m_Atd_yy_o1_t1_l0) + m_Atd_yy_o1_t2_l0) + m_Atd_yy_o2_t0_l0) + m_Atd_yy_o2_t1_l0) + m_Atd_yy_o2_t2_l0;
						RHS_Atd_yz = (((((((((((m_Atd_yz_o0_t0_l0 + m_Atd_yz_o0_t1_l0) + m_Atd_yz_o0_t2_l0) + m_Atd_yz_o0_t3_l0) + m_Atd_yz_o0_t4_l0) + m_Atd_yz_o0_t5_l0) + d_Atd_yz_o0_t6_m0_l0) + m_Atd_yz_o1_t0_l0) + m_Atd_yz_o1_t1_l0) + m_Atd_yz_o1_t2_l0) + m_Atd_yz_o2_t0_l0) + m_Atd_yz_o2_t1_l0) + m_Atd_yz_o2_t2_l0;
						RHS_Atd_zz = ((((((((m_Atd_zz_o0_t0_l0 + m_Atd_zz_o0_t1_l0) + m_Atd_zz_o0_t2_l0) + d_Atd_zz_o0_t3_m0_l0) + m_Atd_zz_o1_t0_l0) + m_Atd_zz_o1_t1_l0) + m_Atd_zz_o1_t2_l0) + m_Atd_zz_o2_t0_l0) + m_Atd_zz_o2_t1_l0) + m_Atd_zz_o2_t2_l0;
						RHS_Gamh_x = (((((((((((((((((((((((((((((m_Gamh_x_o0_t0_l0 + m_Gamh_x_o0_t1_l0) + m_Gamh_x_o0_t2_l0) + m_Gamh_x_o0_t3_l0) + m_Gamh_x_o0_t4_l0) + m_Gamh_x_o0_t5_l0) + m_Gamh_x_o0_t6_l0) + m_Gamh_x_o0_t7_l0) + m_Gamh_x_o0_t8_l0) + m_Gamh_x_o0_t9_l0) + m_Gamh_x_o0_t10_l0) + m_Gamh_x_o0_t11_l0) + m_Gamh_x_o0_t12_l0) + m_Gamh_x_o0_t13_l0) + m_Gamh_x_o0_t14_l0) + m_Gamh_x_o0_t15_l0) + m_Gamh_x_o0_t16_l0) + m_Gamh_x_o0_t17_l0) + m_Gamh_x_o0_t18_l0) + m_Gamh_x_o0_t19_l0) + m_Gamh_x_o0_t20_l0) + m_Gamh_x_o0_t21_l0) + m_Gamh_x_o0_t22_l0) + m_Gamh_x_o0_t23_l0) + d_Gamh_x_o0_t24_m0_l0) + m_Gamh_x_o1_t0_l0) + m_Gamh_x_o1_t1_l0) + m_Gamh_x_o1_t2_l0) + m_Gamh_x_o2_t0_l0) + m_Gamh_x_o2_t1_l0) + m_Gamh_x_o2_t2_l0;
						RHS_Gamh_y = (((((((((((((((((((((((((((((m_Gamh_y_o0_t0_l0 + m_Gamh_y_o0_t1_l0) + m_Gamh_y_o0_t2_l0) + m_Gamh_y_o0_t3_l0) + m_Gamh_y_o0_t4_l0) + m_Gamh_y_o0_t5_l0) + m_Gamh_y_o0_t6_l0) + m_Gamh_y_o0_t7_l0) + m_Gamh_y_o0_t8_l0) + m_Gamh_y_o0_t9_l0) + m_Gamh_y_o0_t10_l0) + m_Gamh_y_o0_t11_l0) + m_Gamh_y_o0_t12_l0) + m_Gamh_y_o0_t13_l0) + m_Gamh_y_o0_t14_l0) + m_Gamh_y_o0_t15_l0) + m_Gamh_y_o0_t16_l0) + m_Gamh_y_o0_t17_l0) + m_Gamh_y_o0_t18_l0) + m_Gamh_y_o0_t19_l0) + m_Gamh_y_o0_t20_l0) + m_Gamh_y_o0_t21_l0) + m_Gamh_y_o0_t22_l0) + m_Gamh_y_o0_t23_l0) + d_Gamh_y_o0_t24_m0_l0) + m_Gamh_y_o1_t0_l0) + m_Gamh_y_o1_t1_l0) + m_Gamh_y_o1_t2_l0) + m_Gamh_y_o2_t0_l0) + m_Gamh_y_o2_t1_l0) + m_Gamh_y_o2_t2_l0;
						RHS_Gamh_z = (((((((((((((((((((((((((((((m_Gamh_z_o0_t0_l0 + m_Gamh_z_o0_t1_l0) + m_Gamh_z_o0_t2_l0) + m_Gamh_z_o0_t3_l0) + m_Gamh_z_o0_t4_l0) + m_Gamh_z_o0_t5_l0) + m_Gamh_z_o0_t6_l0) + m_Gamh_z_o0_t7_l0) + m_Gamh_z_o0_t8_l0) + m_Gamh_z_o0_t9_l0) + m_Gamh_z_o0_t10_l0) + m_Gamh_z_o0_t11_l0) + m_Gamh_z_o0_t12_l0) + m_Gamh_z_o0_t13_l0) + m_Gamh_z_o0_t14_l0) + m_Gamh_z_o0_t15_l0) + m_Gamh_z_o0_t16_l0) + m_Gamh_z_o0_t17_l0) + m_Gamh_z_o0_t18_l0) + m_Gamh_z_o0_t19_l0) + m_Gamh_z_o0_t20_l0) + m_Gamh_z_o0_t21_l0) + m_Gamh_z_o0_t22_l0) + m_Gamh_z_o0_t23_l0) + d_Gamh_z_o0_t24_m0_l0) + m_Gamh_z_o1_t0_l0) + m_Gamh_z_o1_t1_l0) + m_Gamh_z_o1_t2_l0) + m_Gamh_z_o2_t0_l0) + m_Gamh_z_o2_t1_l0) + m_Gamh_z_o2_t2_l0;
						RHS_Betau_x = (((((d_Betau_x_o0_t0_m0_l0 + m_Betau_x_o1_t0_l0) + m_Betau_x_o1_t1_l0) + m_Betau_x_o1_t2_l0) + m_Betau_x_o2_t0_l0) + m_Betau_x_o2_t1_l0) + m_Betau_x_o2_t2_l0;
						RHS_Betau_y = (((((d_Betau_y_o0_t0_m0_l0 + m_Betau_y_o1_t0_l0) + m_Betau_y_o1_t1_l0) + m_Betau_y_o1_t2_l0) + m_Betau_y_o2_t0_l0) + m_Betau_y_o2_t1_l0) + m_Betau_y_o2_t2_l0;
						RHS_Betau_z = (((((d_Betau_z_o0_t0_m0_l0 + m_Betau_z_o1_t0_l0) + m_Betau_z_o1_t1_l0) + m_Betau_z_o1_t2_l0) + m_Betau_z_o2_t0_l0) + m_Betau_z_o2_t1_l0) + m_Betau_z_o2_t2_l0;
						RHS_Alpha = (((((d_Alpha_o0_t0_m0_l0 + m_Alpha_o1_t0_l0) + m_Alpha_o1_t1_l0) + m_Alpha_o1_t2_l0) + m_Alpha_o2_t0_l0) + m_Alpha_o2_t1_l0) + m_Alpha_o2_t2_l0;
						RHS_chi = (((((d_chi_o0_t0_m0_l0 + m_chi_o1_t0_l0) + m_chi_o1_t1_l0) + m_chi_o1_t2_l0) + m_chi_o2_t0_l0) + m_chi_o2_t1_l0) + m_chi_o2_t2_l0;
						RHS_trK = ((((((((((((((((((((((((((m_trK_o0_t0_l0 + m_trK_o0_t1_l0) + m_trK_o0_t2_l0) + m_trK_o0_t3_l0) + m_trK_o0_t4_l0) + m_trK_o0_t5_l0) + m_trK_o0_t6_l0) + m_trK_o0_t7_l0) + m_trK_o0_t8_l0) + m_trK_o0_t9_l0) + m_trK_o0_t10_l0) + m_trK_o0_t11_l0) + m_trK_o0_t12_l0) + m_trK_o0_t13_l0) + m_trK_o0_t14_l0) + m_trK_o0_t15_l0) + m_trK_o0_t16_l0) + m_trK_o0_t17_l0) + m_trK_o0_t18_l0) + m_trK_o0_t19_l0) + m_trK_o0_t20_l0) + d_trK_o0_t21_m0_l0) + m_trK_o1_t0_l0) + m_trK_o1_t1_l0) + m_trK_o1_t2_l0) + m_trK_o2_t0_l0) + m_trK_o2_t1_l0) + m_trK_o2_t2_l0;
						RHS_theta = ((((((((m_theta_o0_t0_l0 + m_theta_o0_t1_l0) + m_theta_o0_t2_l0) + d_theta_o0_t3_m0_l0) + m_theta_o1_t0_l0) + m_theta_o1_t1_l0) + m_theta_o1_t2_l0) + m_theta_o2_t0_l0) + m_theta_o2_t1_l0) + m_theta_o2_t2_l0;
						n_x = 0.0;
						n_y = 0.0;
						n_z = 0.0;
						interaction_index = 0.0;
						if ((((((vector(FOV_xLower, i + 1, j, k) > 0.0) || (vector(FOV_xUpper, i + 1, j, k) > 0.0)) || (vector(FOV_yLower, i + 1, j, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k) > 0.0)) || (vector(FOV_zLower, i + 1, j, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x + 1.0 / dx[0];
						}
						if ((((((vector(FOV_xLower, i - 1, j, k) > 0.0) || (vector(FOV_xUpper, i - 1, j, k) > 0.0)) || (vector(FOV_yLower, i - 1, j, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k) > 0.0)) || (vector(FOV_zLower, i - 1, j, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x - 1.0 / dx[0];
						}
						if ((((((vector(FOV_xLower, i, j + 1, k) > 0.0) || (vector(FOV_xUpper, i, j + 1, k) > 0.0)) || (vector(FOV_yLower, i, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k) > 0.0)) || (vector(FOV_zLower, i, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y + 1.0 / dx[1];
						}
						if ((((((vector(FOV_xLower, i, j - 1, k) > 0.0) || (vector(FOV_xUpper, i, j - 1, k) > 0.0)) || (vector(FOV_yLower, i, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k) > 0.0)) || (vector(FOV_zLower, i, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y - 1.0 / dx[1];
						}
						if ((((((vector(FOV_xLower, i, j, k + 1) > 0.0) || (vector(FOV_xUpper, i, j, k + 1) > 0.0)) || (vector(FOV_yLower, i, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j, k + 1) > 0.0)) || (vector(FOV_zLower, i, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j, k + 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z + 1.0 / dx[2];
						}
						if ((((((vector(FOV_xLower, i, j, k - 1) > 0.0) || (vector(FOV_xUpper, i, j, k - 1) > 0.0)) || (vector(FOV_yLower, i, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j, k - 1) > 0.0)) || (vector(FOV_zLower, i, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j, k - 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z - 1.0 / dx[2];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k) > 0.0)) || (vector(FOV_yLower, i + 1, j, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k) > 0.0)) || (vector(FOV_zLower, i + 1, j, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x + 1.0 / dx[0];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i - 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k) > 0.0)) || (vector(FOV_yLower, i - 1, j, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k) > 0.0)) || (vector(FOV_zLower, i - 1, j, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x - 1.0 / dx[0];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k) > 0.0)) || (vector(FOV_yLower, i, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k) > 0.0)) || (vector(FOV_zLower, i, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y + 1.0 / dx[1];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j - 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k) > 0.0)) || (vector(FOV_yLower, i, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k) > 0.0)) || (vector(FOV_zLower, i, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y - 1.0 / dx[1];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j, k + 1) > 0.0)) || (vector(FOV_yLower, i, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j, k + 1) > 0.0)) || (vector(FOV_zLower, i, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j, k + 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z + 1.0 / dx[2];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k - 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j, k - 1) > 0.0)) || (vector(FOV_yLower, i, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j, k - 1) > 0.0)) || (vector(FOV_zLower, i, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j, k - 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z - 1.0 / dx[2];
						}
						if ((!equalsEq(n_x, 0.0) || !equalsEq(n_y, 0.0)) || !equalsEq(n_z, 0.0)) {
							mod_normal = sqrt((n_x * n_x + n_y * n_y) + n_z * n_z);
							n_x = n_x / mod_normal;
							n_y = n_y / mod_normal;
							n_z = n_z / mod_normal;
						}
						if (equalsEq(interaction_index, 1.0)) {
							i_d_Atd_xx_o0_t0_m0_l0 = D1CDO4_i(rk2Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xx_o0_t1_m0_l0 = D1CDO4_j(rk2Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xx_o0_t2_m0_l0 = D1CDO4_k(rk2Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xy_o0_t0_m0_l0 = D1CDO4_i(rk2Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xy_o0_t1_m0_l0 = D1CDO4_j(rk2Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xy_o0_t2_m0_l0 = D1CDO4_k(rk2Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xz_o0_t0_m0_l0 = D1CDO4_i(rk2Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xz_o0_t1_m0_l0 = D1CDO4_j(rk2Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xz_o0_t2_m0_l0 = D1CDO4_k(rk2Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yy_o0_t0_m0_l0 = D1CDO4_i(rk2Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yy_o0_t1_m0_l0 = D1CDO4_j(rk2Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yy_o0_t2_m0_l0 = D1CDO4_k(rk2Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yz_o0_t0_m0_l0 = D1CDO4_i(rk2Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yz_o0_t1_m0_l0 = D1CDO4_j(rk2Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yz_o0_t2_m0_l0 = D1CDO4_k(rk2Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_zz_o0_t0_m0_l0 = D1CDO4_i(rk2Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_zz_o0_t1_m0_l0 = D1CDO4_j(rk2Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_zz_o0_t2_m0_l0 = D1CDO4_k(rk2Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_gtd_xx_o0_t3_m0_l0 = -gtd_xx_falloff * (vector(rk2gtd_xx, i, j, k) - gtd_xx_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_xy_o0_t3_m0_l0 = -gtd_xy_falloff * (vector(rk2gtd_xy, i, j, k) - gtd_xy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_xz_o0_t3_m0_l0 = -gtd_xz_falloff * (vector(rk2gtd_xz, i, j, k) - gtd_xz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_yy_o0_t3_m0_l0 = -gtd_yy_falloff * (vector(rk2gtd_yy, i, j, k) - gtd_yy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_yz_o0_t3_m0_l0 = -gtd_yz_falloff * (vector(rk2gtd_yz, i, j, k) - gtd_yz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_zz_o0_t3_m0_l0 = -gtd_zz_falloff * (vector(rk2gtd_zz, i, j, k) - gtd_zz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_xx_o0_t3_m0_l0 = -Atd_xx_falloff * (vector(rk2Atd_xx, i, j, k) - Atd_xx_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_xy_o0_t3_m0_l0 = -Atd_xy_falloff * (vector(rk2Atd_xy, i, j, k) - Atd_xy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_xz_o0_t3_m0_l0 = -Atd_xz_falloff * (vector(rk2Atd_xz, i, j, k) - Atd_xz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_yy_o0_t3_m0_l0 = -Atd_yy_falloff * (vector(rk2Atd_yy, i, j, k) - Atd_yy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_yz_o0_t3_m0_l0 = -Atd_yz_falloff * (vector(rk2Atd_yz, i, j, k) - Atd_yz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_zz_o0_t3_m0_l0 = -Atd_zz_falloff * (vector(rk2Atd_zz, i, j, k) - Atd_zz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Gamh_x_o0_t3_m0_l0 = -Gamh_x_falloff * (vector(rk2Gamh_x, i, j, k) - Gamh_x_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Gamh_y_o0_t3_m0_l0 = -Gamh_y_falloff * (vector(rk2Gamh_y, i, j, k) - Gamh_y_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Gamh_z_o0_t3_m0_l0 = -Gamh_z_falloff * (vector(rk2Gamh_z, i, j, k) - Gamh_z_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Betau_x_o0_t3_m0_l0 = -Betau_x_falloff * (vector(rk2Betau_x, i, j, k) - Betau_x_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Betau_y_o0_t3_m0_l0 = -Betau_y_falloff * (vector(rk2Betau_y, i, j, k) - Betau_y_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Betau_z_o0_t3_m0_l0 = -Betau_z_falloff * (vector(rk2Betau_z, i, j, k) - Betau_z_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Alpha_o0_t3_m0_l0 = -Alpha_falloff * (vector(rk2Alpha, i, j, k) - Alpha_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_chi_o0_t3_m0_l0 = -chi_falloff * (vector(rk2chi, i, j, k) - chi_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_trK_o0_t3_m0_l0 = -trK_falloff * (vector(rk2trK, i, j, k) - trK_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_theta_o0_t3_m0_l0 = -theta_falloff * (vector(rk2theta, i, j, k) - theta_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_m_gtd_xx_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xxx_o0_t0_m0_l0;
							i_m_gtd_xx_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xxy_o0_t0_m0_l0;
							i_m_gtd_xx_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xxz_o0_t0_m0_l0;
							i_m_gtd_xy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yxx_o0_t0_m0_l0;
							i_m_gtd_xy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyy_o0_t0_m0_l0;
							i_m_gtd_xy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyz_o0_t1_m0_l0;
							i_m_gtd_xz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_zxx_o0_t0_m0_l0;
							i_m_gtd_xz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyz_o0_t0_m0_l0;
							i_m_gtd_xz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xzz_o0_t0_m0_l0;
							i_m_gtd_yy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyy_o0_t1_m0_l0;
							i_m_gtd_yy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yyy_o0_t0_m0_l0;
							i_m_gtd_yy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yyz_o0_t0_m0_l0;
							i_m_gtd_yz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyz_o0_t2_m0_l0;
							i_m_gtd_yz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_zyy_o0_t0_m0_l0;
							i_m_gtd_yz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yzz_o0_t0_m0_l0;
							i_m_gtd_zz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xzz_o0_t1_m0_l0;
							i_m_gtd_zz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yzz_o0_t1_m0_l0;
							i_m_gtd_zz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_zzz_o0_t0_m0_l0;
							i_m_Atd_xx_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xx_o0_t0_m0_l0;
							i_m_Atd_xx_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xx_o0_t1_m0_l0;
							i_m_Atd_xx_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xx_o0_t2_m0_l0;
							i_m_Atd_xy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xy_o0_t0_m0_l0;
							i_m_Atd_xy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xy_o0_t1_m0_l0;
							i_m_Atd_xy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xy_o0_t2_m0_l0;
							i_m_Atd_xz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xz_o0_t0_m0_l0;
							i_m_Atd_xz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xz_o0_t1_m0_l0;
							i_m_Atd_xz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xz_o0_t2_m0_l0;
							i_m_Atd_yy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yy_o0_t0_m0_l0;
							i_m_Atd_yy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yy_o0_t1_m0_l0;
							i_m_Atd_yy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yy_o0_t2_m0_l0;
							i_m_Atd_yz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yz_o0_t0_m0_l0;
							i_m_Atd_yz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yz_o0_t1_m0_l0;
							i_m_Atd_yz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yz_o0_t2_m0_l0;
							i_m_Atd_zz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_zz_o0_t0_m0_l0;
							i_m_Atd_zz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_zz_o0_t1_m0_l0;
							i_m_Atd_zz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_zz_o0_t2_m0_l0;
							i_m_Gamh_x_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xx_o0_t6_m0_l0;
							i_m_Gamh_x_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xy_o0_t6_m0_l0;
							i_m_Gamh_x_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xz_o0_t6_m0_l0;
							i_m_Gamh_y_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xx_o0_t7_m0_l0;
							i_m_Gamh_y_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xy_o0_t7_m0_l0;
							i_m_Gamh_y_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xz_o0_t7_m0_l0;
							i_m_Gamh_z_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xx_o0_t8_m0_l0;
							i_m_Gamh_z_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xy_o0_t8_m0_l0;
							i_m_Gamh_z_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xz_o0_t8_m0_l0;
							i_m_Betau_x_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_div_Beta_o0_t0_m0_l0;
							i_m_Betau_x_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_y_o0_t7_m0_l0;
							i_m_Betau_x_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_z_o0_t7_m0_l0;
							i_m_Betau_y_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_x_o0_t8_m0_l0;
							i_m_Betau_y_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_div_Beta_o0_t1_m0_l0;
							i_m_Betau_y_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_z_o0_t8_m0_l0;
							i_m_Betau_z_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_x_o0_t9_m0_l0;
							i_m_Betau_z_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_y_o0_t9_m0_l0;
							i_m_Betau_z_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_div_Beta_o0_t2_m0_l0;
							i_m_Alpha_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Psi1_xx_o0_t1_m0_l0;
							i_m_Alpha_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Psi1_xx_o0_t2_m0_l0;
							i_m_Alpha_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Psi1_xx_o0_t3_m0_l0;
							i_m_chi_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rpd_xx_o0_t0_m0_l0;
							i_m_chi_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rpd_xx_o0_t1_m0_l0;
							i_m_chi_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rpd_xx_o0_t2_m0_l0;
							i_m_trK_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t0_m0_l0;
							i_m_trK_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t1_m0_l0;
							i_m_trK_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t2_m0_l0;
							i_m_theta_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t3_m0_l0;
							i_m_theta_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t4_m0_l0;
							i_m_theta_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t5_m0_l0;
							RHS_gtd_xx = ((i_m_gtd_xx_o0_t0_l0 + i_m_gtd_xx_o0_t1_l0) + i_m_gtd_xx_o0_t2_l0) + i_d_gtd_xx_o0_t3_m0_l0;
							RHS_gtd_xy = ((i_m_gtd_xy_o0_t0_l0 + i_m_gtd_xy_o0_t1_l0) + i_m_gtd_xy_o0_t2_l0) + i_d_gtd_xy_o0_t3_m0_l0;
							RHS_gtd_xz = ((i_m_gtd_xz_o0_t0_l0 + i_m_gtd_xz_o0_t1_l0) + i_m_gtd_xz_o0_t2_l0) + i_d_gtd_xz_o0_t3_m0_l0;
							RHS_gtd_yy = ((i_m_gtd_yy_o0_t0_l0 + i_m_gtd_yy_o0_t1_l0) + i_m_gtd_yy_o0_t2_l0) + i_d_gtd_yy_o0_t3_m0_l0;
							RHS_gtd_yz = ((i_m_gtd_yz_o0_t0_l0 + i_m_gtd_yz_o0_t1_l0) + i_m_gtd_yz_o0_t2_l0) + i_d_gtd_yz_o0_t3_m0_l0;
							RHS_gtd_zz = ((i_m_gtd_zz_o0_t0_l0 + i_m_gtd_zz_o0_t1_l0) + i_m_gtd_zz_o0_t2_l0) + i_d_gtd_zz_o0_t3_m0_l0;
							RHS_Atd_xx = ((i_m_Atd_xx_o0_t0_l0 + i_m_Atd_xx_o0_t1_l0) + i_m_Atd_xx_o0_t2_l0) + i_d_Atd_xx_o0_t3_m0_l0;
							RHS_Atd_xy = ((i_m_Atd_xy_o0_t0_l0 + i_m_Atd_xy_o0_t1_l0) + i_m_Atd_xy_o0_t2_l0) + i_d_Atd_xy_o0_t3_m0_l0;
							RHS_Atd_xz = ((i_m_Atd_xz_o0_t0_l0 + i_m_Atd_xz_o0_t1_l0) + i_m_Atd_xz_o0_t2_l0) + i_d_Atd_xz_o0_t3_m0_l0;
							RHS_Atd_yy = ((i_m_Atd_yy_o0_t0_l0 + i_m_Atd_yy_o0_t1_l0) + i_m_Atd_yy_o0_t2_l0) + i_d_Atd_yy_o0_t3_m0_l0;
							RHS_Atd_yz = ((i_m_Atd_yz_o0_t0_l0 + i_m_Atd_yz_o0_t1_l0) + i_m_Atd_yz_o0_t2_l0) + i_d_Atd_yz_o0_t3_m0_l0;
							RHS_Atd_zz = ((i_m_Atd_zz_o0_t0_l0 + i_m_Atd_zz_o0_t1_l0) + i_m_Atd_zz_o0_t2_l0) + i_d_Atd_zz_o0_t3_m0_l0;
							RHS_Gamh_x = ((i_m_Gamh_x_o0_t0_l0 + i_m_Gamh_x_o0_t1_l0) + i_m_Gamh_x_o0_t2_l0) + i_d_Gamh_x_o0_t3_m0_l0;
							RHS_Gamh_y = ((i_m_Gamh_y_o0_t0_l0 + i_m_Gamh_y_o0_t1_l0) + i_m_Gamh_y_o0_t2_l0) + i_d_Gamh_y_o0_t3_m0_l0;
							RHS_Gamh_z = ((i_m_Gamh_z_o0_t0_l0 + i_m_Gamh_z_o0_t1_l0) + i_m_Gamh_z_o0_t2_l0) + i_d_Gamh_z_o0_t3_m0_l0;
							RHS_Betau_x = ((i_m_Betau_x_o0_t0_l0 + i_m_Betau_x_o0_t1_l0) + i_m_Betau_x_o0_t2_l0) + i_d_Betau_x_o0_t3_m0_l0;
							RHS_Betau_y = ((i_m_Betau_y_o0_t0_l0 + i_m_Betau_y_o0_t1_l0) + i_m_Betau_y_o0_t2_l0) + i_d_Betau_y_o0_t3_m0_l0;
							RHS_Betau_z = ((i_m_Betau_z_o0_t0_l0 + i_m_Betau_z_o0_t1_l0) + i_m_Betau_z_o0_t2_l0) + i_d_Betau_z_o0_t3_m0_l0;
							RHS_Alpha = ((i_m_Alpha_o0_t0_l0 + i_m_Alpha_o0_t1_l0) + i_m_Alpha_o0_t2_l0) + i_d_Alpha_o0_t3_m0_l0;
							RHS_chi = ((i_m_chi_o0_t0_l0 + i_m_chi_o0_t1_l0) + i_m_chi_o0_t2_l0) + i_d_chi_o0_t3_m0_l0;
							RHS_trK = ((i_m_trK_o0_t0_l0 + i_m_trK_o0_t1_l0) + i_m_trK_o0_t2_l0) + i_d_trK_o0_t3_m0_l0;
							RHS_theta = ((i_m_theta_o0_t0_l0 + i_m_theta_o0_t1_l0) + i_m_theta_o0_t2_l0) + i_d_theta_o0_t3_m0_l0;
						}
						if (dissipation_factor_Sfd_x > 0.0) {
							RHS_Sfd_x = RHS_Sfd_x + dissipation_factor_Sfd_x * (meshDissipation_i(rk2Sfd_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Sfd_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Sfd_x, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Sfd_y > 0.0) {
							RHS_Sfd_y = RHS_Sfd_y + dissipation_factor_Sfd_y * (meshDissipation_i(rk2Sfd_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Sfd_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Sfd_y, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Sfd_z > 0.0) {
							RHS_Sfd_z = RHS_Sfd_z + dissipation_factor_Sfd_z * (meshDissipation_i(rk2Sfd_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Sfd_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Sfd_z, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Bfu_x > 0.0) {
							RHS_Bfu_x = RHS_Bfu_x + dissipation_factor_Bfu_x * (meshDissipation_i(rk2Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Bfu_y > 0.0) {
							RHS_Bfu_y = RHS_Bfu_y + dissipation_factor_Bfu_y * (meshDissipation_i(rk2Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Bfu_z > 0.0) {
							RHS_Bfu_z = RHS_Bfu_z + dissipation_factor_Bfu_z * (meshDissipation_i(rk2Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Df > 0.0) {
							RHS_Df = RHS_Df + dissipation_factor_Df * (meshDissipation_i(rk2Df, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Df, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Df, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_DYf > 0.0) {
							RHS_DYf = RHS_DYf + dissipation_factor_DYf * (meshDissipation_i(rk2DYf, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2DYf, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2DYf, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_tauf > 0.0) {
							RHS_tauf = RHS_tauf + dissipation_factor_tauf * (meshDissipation_i(rk2tauf, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2tauf, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2tauf, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_phif > 0.0) {
							RHS_phif = RHS_phif + dissipation_factor_phif * (meshDissipation_i(rk2phif, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2phif, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2phif, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_xx > 0.0) {
							RHS_gtd_xx = RHS_gtd_xx + dissipation_factor_gtd_xx * (meshDissipation_i(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_xy > 0.0) {
							RHS_gtd_xy = RHS_gtd_xy + dissipation_factor_gtd_xy * (meshDissipation_i(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_xz > 0.0) {
							RHS_gtd_xz = RHS_gtd_xz + dissipation_factor_gtd_xz * (meshDissipation_i(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_yy > 0.0) {
							RHS_gtd_yy = RHS_gtd_yy + dissipation_factor_gtd_yy * (meshDissipation_i(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_yz > 0.0) {
							RHS_gtd_yz = RHS_gtd_yz + dissipation_factor_gtd_yz * (meshDissipation_i(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_zz > 0.0) {
							RHS_gtd_zz = RHS_gtd_zz + dissipation_factor_gtd_zz * (meshDissipation_i(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_xx > 0.0) {
							RHS_Atd_xx = RHS_Atd_xx + dissipation_factor_Atd_xx * (meshDissipation_i(rk2Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_xy > 0.0) {
							RHS_Atd_xy = RHS_Atd_xy + dissipation_factor_Atd_xy * (meshDissipation_i(rk2Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_xz > 0.0) {
							RHS_Atd_xz = RHS_Atd_xz + dissipation_factor_Atd_xz * (meshDissipation_i(rk2Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_yy > 0.0) {
							RHS_Atd_yy = RHS_Atd_yy + dissipation_factor_Atd_yy * (meshDissipation_i(rk2Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_yz > 0.0) {
							RHS_Atd_yz = RHS_Atd_yz + dissipation_factor_Atd_yz * (meshDissipation_i(rk2Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_zz > 0.0) {
							RHS_Atd_zz = RHS_Atd_zz + dissipation_factor_Atd_zz * (meshDissipation_i(rk2Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Gamh_x > 0.0) {
							RHS_Gamh_x = RHS_Gamh_x + dissipation_factor_Gamh_x * (meshDissipation_i(rk2Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Gamh_y > 0.0) {
							RHS_Gamh_y = RHS_Gamh_y + dissipation_factor_Gamh_y * (meshDissipation_i(rk2Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Gamh_z > 0.0) {
							RHS_Gamh_z = RHS_Gamh_z + dissipation_factor_Gamh_z * (meshDissipation_i(rk2Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Betau_x > 0.0) {
							RHS_Betau_x = RHS_Betau_x + dissipation_factor_Betau_x * (meshDissipation_i(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Betau_y > 0.0) {
							RHS_Betau_y = RHS_Betau_y + dissipation_factor_Betau_y * (meshDissipation_i(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Betau_z > 0.0) {
							RHS_Betau_z = RHS_Betau_z + dissipation_factor_Betau_z * (meshDissipation_i(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Alpha > 0.0) {
							RHS_Alpha = RHS_Alpha + dissipation_factor_Alpha * (meshDissipation_i(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2Alpha, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_chi > 0.0) {
							RHS_chi = RHS_chi + dissipation_factor_chi * (meshDissipation_i(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2chi, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_trK > 0.0) {
							RHS_trK = RHS_trK + dissipation_factor_trK * (meshDissipation_i(rk2trK, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2trK, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2trK, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_theta > 0.0) {
							RHS_theta = RHS_theta + dissipation_factor_theta * (meshDissipation_i(rk2theta, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk2theta, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk2theta, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						vector(rk3Sfd_x, i, j, k) = RK4P3_(RHS_Sfd_x, vector(Sfd_x_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Sfd_y, i, j, k) = RK4P3_(RHS_Sfd_y, vector(Sfd_y_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Sfd_z, i, j, k) = RK4P3_(RHS_Sfd_z, vector(Sfd_z_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Bfu_x, i, j, k) = RK4P3_(RHS_Bfu_x, vector(Bfu_x_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Bfu_y, i, j, k) = RK4P3_(RHS_Bfu_y, vector(Bfu_y_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Bfu_z, i, j, k) = RK4P3_(RHS_Bfu_z, vector(Bfu_z_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Df, i, j, k) = RK4P3_(RHS_Df, vector(Df_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3DYf, i, j, k) = RK4P3_(RHS_DYf, vector(DYf_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3tauf, i, j, k) = RK4P3_(RHS_tauf, vector(tauf_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3phif, i, j, k) = RK4P3_(RHS_phif, vector(phif_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3gtd_xx, i, j, k) = RK4P3_(RHS_gtd_xx, vector(gtd_xx_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3gtd_xy, i, j, k) = RK4P3_(RHS_gtd_xy, vector(gtd_xy_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3gtd_xz, i, j, k) = RK4P3_(RHS_gtd_xz, vector(gtd_xz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3gtd_yy, i, j, k) = RK4P3_(RHS_gtd_yy, vector(gtd_yy_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3gtd_yz, i, j, k) = RK4P3_(RHS_gtd_yz, vector(gtd_yz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3gtd_zz, i, j, k) = RK4P3_(RHS_gtd_zz, vector(gtd_zz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Atd_xx, i, j, k) = RK4P3_(RHS_Atd_xx, vector(Atd_xx_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Atd_xy, i, j, k) = RK4P3_(RHS_Atd_xy, vector(Atd_xy_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Atd_xz, i, j, k) = RK4P3_(RHS_Atd_xz, vector(Atd_xz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Atd_yy, i, j, k) = RK4P3_(RHS_Atd_yy, vector(Atd_yy_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Atd_yz, i, j, k) = RK4P3_(RHS_Atd_yz, vector(Atd_yz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Atd_zz, i, j, k) = RK4P3_(RHS_Atd_zz, vector(Atd_zz_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Gamh_x, i, j, k) = RK4P3_(RHS_Gamh_x, vector(Gamh_x_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Gamh_y, i, j, k) = RK4P3_(RHS_Gamh_y, vector(Gamh_y_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Gamh_z, i, j, k) = RK4P3_(RHS_Gamh_z, vector(Gamh_z_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Betau_x, i, j, k) = RK4P3_(RHS_Betau_x, vector(Betau_x_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Betau_y, i, j, k) = RK4P3_(RHS_Betau_y, vector(Betau_y_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Betau_z, i, j, k) = RK4P3_(RHS_Betau_z, vector(Betau_z_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3Alpha, i, j, k) = RK4P3_(RHS_Alpha, vector(Alpha_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3chi, i, j, k) = RK4P3_(RHS_chi, vector(chi_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3trK, i, j, k) = RK4P3_(RHS_trK, vector(trK_p, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(rk3theta, i, j, k) = RK4P3_(RHS_theta, vector(theta_p, i, j, k), dx, simPlat_dt, ilast, jlast);
					}
				}
			}
		}
	}
	//Fill ghosts and periodical boundaries
	time_interpolate_operator_mesh1->setStep(3);
	d_bdry_sched_advance16[ln]->fillData(current_time + simPlat_dt, false);
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		//Hard region field distance variables
		double* d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
	
		double* rk3Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_x_id).get())->getPointer();
		double* rk3Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_y_id).get())->getPointer();
		double* rk3Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_z_id).get())->getPointer();
		double* rk3Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_x_id).get())->getPointer();
		double* rk3Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_y_id).get())->getPointer();
		double* rk3Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_z_id).get())->getPointer();
		double* rk3Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Df_id).get())->getPointer();
		double* rk3DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3DYf_id).get())->getPointer();
		double* rk3tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3tauf_id).get())->getPointer();
		double* rk3phif = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3phif_id).get())->getPointer();
		double* rk3gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xx_id).get())->getPointer();
		double* rk3gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xy_id).get())->getPointer();
		double* rk3gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xz_id).get())->getPointer();
		double* rk3gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_yy_id).get())->getPointer();
		double* rk3gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_yz_id).get())->getPointer();
		double* rk3gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_zz_id).get())->getPointer();
		double* rk3Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_xx_id).get())->getPointer();
		double* rk3Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_xy_id).get())->getPointer();
		double* rk3Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_xz_id).get())->getPointer();
		double* rk3Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_yy_id).get())->getPointer();
		double* rk3Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_yz_id).get())->getPointer();
		double* rk3Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_zz_id).get())->getPointer();
		double* rk3Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Gamh_x_id).get())->getPointer();
		double* rk3Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Gamh_y_id).get())->getPointer();
		double* rk3Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Gamh_z_id).get())->getPointer();
		double* rk3Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_x_id).get())->getPointer();
		double* rk3Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_y_id).get())->getPointer();
		double* rk3Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_z_id).get())->getPointer();
		double* rk3Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Alpha_id).get())->getPointer();
		double* rk3chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3chi_id).get())->getPointer();
		double* rk3trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3trK_id).get())->getPointer();
		double* rk3theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3theta_id).get())->getPointer();
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, d_ghost_width, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, d_ghost_width, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, d_ghost_width, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, d_ghost_width, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, d_ghost_width, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, d_ghost_width, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, d_ghost_width, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, d_ghost_width, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, d_ghost_width, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i - 1, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i - 1, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i - 1, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i - 1, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i - 1, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i - 1, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i - 1, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i - 1, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i - 1, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, d_ghost_width, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, d_ghost_width, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, d_ghost_width, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, d_ghost_width, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, d_ghost_width, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, d_ghost_width, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, d_ghost_width, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, d_ghost_width, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, d_ghost_width, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j - 1, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j - 1, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j - 1, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j - 1, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j - 1, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j - 1, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j - 1, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j - 1, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j - 1, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, d_ghost_width);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, d_ghost_width);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, d_ghost_width);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, d_ghost_width);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, d_ghost_width);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, d_ghost_width);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, d_ghost_width);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, d_ghost_width);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, d_ghost_width);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, k - 1);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, k - 1);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, k - 1);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, k - 1);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, k - 1);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, k - 1);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, k - 1);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, k - 1);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, k - 1);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, d_ghost_width, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, d_ghost_width, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, d_ghost_width, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, d_ghost_width, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, d_ghost_width, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, d_ghost_width, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, d_ghost_width, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, d_ghost_width, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, d_ghost_width, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i - 1, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i - 1, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i - 1, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i - 1, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i - 1, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i - 1, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i - 1, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i - 1, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i - 1, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, d_ghost_width, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, d_ghost_width, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, d_ghost_width, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, d_ghost_width, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, d_ghost_width, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, d_ghost_width, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, d_ghost_width, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, d_ghost_width, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, d_ghost_width, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j - 1, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j - 1, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j - 1, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j - 1, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j - 1, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j - 1, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j - 1, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j - 1, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j - 1, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, d_ghost_width);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, d_ghost_width);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, d_ghost_width);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, d_ghost_width);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, d_ghost_width);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, d_ghost_width);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, d_ghost_width);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, d_ghost_width);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, d_ghost_width);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, k - 1);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, k - 1);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, k - 1);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, k - 1);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, k - 1);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, k - 1);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, k - 1);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, k - 1);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, k - 1);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, d_ghost_width, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, d_ghost_width, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, d_ghost_width, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, d_ghost_width, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, d_ghost_width, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, d_ghost_width, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, d_ghost_width, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, d_ghost_width, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, d_ghost_width, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i - 1, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i - 1, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i - 1, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i - 1, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i - 1, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i - 1, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i - 1, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i - 1, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i - 1, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, d_ghost_width, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, d_ghost_width, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, d_ghost_width, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, d_ghost_width, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, d_ghost_width, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, d_ghost_width, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, d_ghost_width, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, d_ghost_width, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, d_ghost_width, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j - 1, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j - 1, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j - 1, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j - 1, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j - 1, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j - 1, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j - 1, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j - 1, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j - 1, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, d_ghost_width);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, d_ghost_width);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, d_ghost_width);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, d_ghost_width);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, d_ghost_width);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, d_ghost_width);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, d_ghost_width);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, d_ghost_width);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, d_ghost_width);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, k - 1);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, k - 1);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, k - 1);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, k - 1);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, k - 1);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, k - 1);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, k - 1);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, k - 1);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, k - 1);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, d_ghost_width, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, d_ghost_width, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, d_ghost_width, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, d_ghost_width, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, d_ghost_width, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, d_ghost_width, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, d_ghost_width, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, d_ghost_width, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, d_ghost_width, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i - 1, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i - 1, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i - 1, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i - 1, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i - 1, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i - 1, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i - 1, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i - 1, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i - 1, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, d_ghost_width, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, d_ghost_width, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, d_ghost_width, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, d_ghost_width, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, d_ghost_width, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, d_ghost_width, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, d_ghost_width, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, d_ghost_width, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, d_ghost_width, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j - 1, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j - 1, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j - 1, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j - 1, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j - 1, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j - 1, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j - 1, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j - 1, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j - 1, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, d_ghost_width);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, d_ghost_width);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, d_ghost_width);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, d_ghost_width);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, d_ghost_width);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, d_ghost_width);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, d_ghost_width);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, d_ghost_width);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, d_ghost_width);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, k - 1);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, k - 1);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, k - 1);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, k - 1);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, k - 1);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, k - 1);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, k - 1);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, k - 1);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, k - 1);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, d_ghost_width, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, d_ghost_width, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, d_ghost_width, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, d_ghost_width, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, d_ghost_width, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, d_ghost_width, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, d_ghost_width, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, d_ghost_width, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, d_ghost_width, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i - 1, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i - 1, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i - 1, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i - 1, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i - 1, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i - 1, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i - 1, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i - 1, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i - 1, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, d_ghost_width, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, d_ghost_width, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, d_ghost_width, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, d_ghost_width, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, d_ghost_width, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, d_ghost_width, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, d_ghost_width, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, d_ghost_width, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, d_ghost_width, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j - 1, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j - 1, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j - 1, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j - 1, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j - 1, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j - 1, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j - 1, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j - 1, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j - 1, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, d_ghost_width);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, d_ghost_width);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, d_ghost_width);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, d_ghost_width);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, d_ghost_width);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, d_ghost_width);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, d_ghost_width);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, d_ghost_width);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, d_ghost_width);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, k - 1);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, k - 1);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, k - 1);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, k - 1);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, k - 1);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, k - 1);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, k - 1);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, k - 1);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, k - 1);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, d_ghost_width, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, d_ghost_width, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, d_ghost_width, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, d_ghost_width, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, d_ghost_width, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, d_ghost_width, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, d_ghost_width, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, d_ghost_width, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, d_ghost_width, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i - 1, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i - 1, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i - 1, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i - 1, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i - 1, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i - 1, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i - 1, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i - 1, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i - 1, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, d_ghost_width, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, d_ghost_width, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, d_ghost_width, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, d_ghost_width, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, d_ghost_width, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, d_ghost_width, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, d_ghost_width, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, d_ghost_width, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, d_ghost_width, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j - 1, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j - 1, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j - 1, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j - 1, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j - 1, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j - 1, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j - 1, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j - 1, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j - 1, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, d_ghost_width);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, d_ghost_width);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, d_ghost_width);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, d_ghost_width);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, d_ghost_width);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, d_ghost_width);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, d_ghost_width);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, d_ghost_width);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, d_ghost_width);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, k - 1);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, k - 1);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, k - 1);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, k - 1);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, k - 1);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, k - 1);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, k - 1);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, k - 1);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, k - 1);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, d_ghost_width, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, d_ghost_width, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, d_ghost_width, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, d_ghost_width, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, d_ghost_width, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, d_ghost_width, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, d_ghost_width, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, d_ghost_width, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, d_ghost_width, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i - 1, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i - 1, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i - 1, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i - 1, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i - 1, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i - 1, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i - 1, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i - 1, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i - 1, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, d_ghost_width, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, d_ghost_width, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, d_ghost_width, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, d_ghost_width, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, d_ghost_width, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, d_ghost_width, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, d_ghost_width, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, d_ghost_width, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, d_ghost_width, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j - 1, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j - 1, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j - 1, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j - 1, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j - 1, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j - 1, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j - 1, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j - 1, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j - 1, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, d_ghost_width);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, d_ghost_width);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, d_ghost_width);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, d_ghost_width);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, d_ghost_width);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, d_ghost_width);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, d_ghost_width);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, d_ghost_width);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, d_ghost_width);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, k - 1);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, k - 1);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, k - 1);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, k - 1);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, k - 1);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, k - 1);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, k - 1);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, k - 1);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, k - 1);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, d_ghost_width, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, d_ghost_width, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, d_ghost_width, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, d_ghost_width, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, d_ghost_width, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, d_ghost_width, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, d_ghost_width, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, d_ghost_width, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, d_ghost_width, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i - 1, j, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i - 1, j, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i - 1, j, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i - 1, j, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i - 1, j, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i - 1, j, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i - 1, j, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i - 1, j, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i - 1, j, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, d_ghost_width, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, d_ghost_width, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, d_ghost_width, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, d_ghost_width, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, d_ghost_width, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, d_ghost_width, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, d_ghost_width, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, d_ghost_width, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, d_ghost_width, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j - 1, k);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j - 1, k);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j - 1, k);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j - 1, k);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j - 1, k);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j - 1, k);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j - 1, k);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j - 1, k);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j - 1, k);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, d_ghost_width);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, d_ghost_width);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, d_ghost_width);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, d_ghost_width);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, d_ghost_width);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, d_ghost_width);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, d_ghost_width);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, d_ghost_width);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, d_ghost_width);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(rk3Sfd_x, i, j, k) = vector(rk3Sfd_x, i, j, k - 1);
						vector(rk3Sfd_y, i, j, k) = vector(rk3Sfd_y, i, j, k - 1);
						vector(rk3Sfd_z, i, j, k) = vector(rk3Sfd_z, i, j, k - 1);
						vector(rk3Bfu_x, i, j, k) = vector(rk3Bfu_x, i, j, k - 1);
						vector(rk3Bfu_y, i, j, k) = vector(rk3Bfu_y, i, j, k - 1);
						vector(rk3Bfu_z, i, j, k) = vector(rk3Bfu_z, i, j, k - 1);
						vector(rk3Df, i, j, k) = vector(rk3Df, i, j, k - 1);
						vector(rk3DYf, i, j, k) = vector(rk3DYf, i, j, k - 1);
						vector(rk3tauf, i, j, k) = vector(rk3tauf, i, j, k - 1);
						vector(rk3phif, i, j, k) = vector(rk3phif, i, j, k - 1);
					}
				}
			}
		}
	}
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		double* rk3chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3chi_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* rk3Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_x_id).get())->getPointer();
		double* rk3Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_y_id).get())->getPointer();
		double* rk3Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_z_id).get())->getPointer();
		double* rk3tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3tauf_id).get())->getPointer();
		double* rk3gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xy_id).get())->getPointer();
		double* rk3gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_yy_id).get())->getPointer();
		double* rk3gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_yz_id).get())->getPointer();
		double* rk3gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xz_id).get())->getPointer();
		double* rk3gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_zz_id).get())->getPointer();
		double* rk3gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xx_id).get())->getPointer();
		double* rk3Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_x_id).get())->getPointer();
		double* rk3Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_y_id).get())->getPointer();
		double* rk3Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_z_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* rk3Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Df_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* rk3Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Alpha_id).get())->getPointer();
		double* rk3DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3DYf_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double rk3chi_max, rk3h, rk3sdetg, rk3Bu_x, rk3Bu_y, rk3Bu_z, rk3tau, rk3inv_chi, rk3Bfd_y, rk3Bfd_z, rk3Bfd_x, rk3decay_factor, rk3detgtd, rk3idetgtd, rk3gtu_zz, rk3gtu_xx, rk3gtu_xz, rk3gtu_xy, rk3Sfu_x, rk3Su_x, rk3vfu_x, rk3Sd_x, rk3Sd_y, rk3gtu_yy, rk3Sd_z, rk3gtu_yz, rk3vfu_z, rk3Sfu_y, rk3Su_y, rk3Sfu_z, rk3Su_z, rk3vfu_y, rk3sqvf, rk3sqW, rk3W, rk3Bd_x, rk3D, rk3Bd_y, rk3Bd_z, rhof_tmp, Yef_tmp, Tf_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, dpfdye_tmp, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tauf_tmp, Df_tmp, DYf_tmp, Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, Efu_x_tmp, Efu_y_tmp, Efu_z_tmp, vacuum_D_tmp, vacuum_tau_tmp, vacuum_rho_reset_tmp;
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j, k));
					rk3h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
					rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
					rk3Bu_x = vector(rk3Bfu_x, i, j, k) / rk3sdetg;
					rk3Bu_y = vector(rk3Bfu_y, i, j, k) / rk3sdetg;
					rk3Bu_z = vector(rk3Bfu_z, i, j, k) / rk3sdetg;
					rk3tau = vector(rk3tauf, i, j, k) / rk3sdetg;
					rk3inv_chi = 1.0 / fabs(rk3chi_max);
					rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_yy, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Bfu_z, i, j, k));
					rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_zz, i, j, k) * vector(rk3Bfu_z, i, j, k));
					rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_xy, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_xz, i, j, k) * vector(rk3Bfu_z, i, j, k));
					rk3decay_factor = MIN(1.0, pow((R_0 / MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))), eta_damping_exp));
					rk3detgtd = vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) * vector(rk3gtd_xx, i, j, k) + (-vector(rk3gtd_yy, i, j, k) * (vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k))) + (-vector(rk3gtd_zz, i, j, k) * (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k))) + (-(vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k)) * vector(rk3gtd_xx, i, j, k)) + 2.0 * vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k);
					rk3idetgtd = 1.0 / rk3detgtd;
					rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yy, i, j, k) - vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k));
					rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) - vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k));
					rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_yz, i, j, k) - vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_xz, i, j, k));
					rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_zz, i, j, k)) + vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xz, i, j, k));
					rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j, k) + rk3gtu_xy * vector(rk3Sfd_y, i, j, k) + rk3gtu_xz * vector(rk3Sfd_z, i, j, k));
					rk3Su_x = rk3Sfu_x / rk3sdetg;
					rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j, k) + rk3gtu_xy * vector(vfd_y, i, j, k) + rk3gtu_xz * vector(vfd_z, i, j, k));
					rk3Sd_x = vector(rk3Sfd_x, i, j, k) / rk3sdetg;
					rk3Sd_y = vector(rk3Sfd_y, i, j, k) / rk3sdetg;
					rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_zz, i, j, k) - vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k));
					rk3Sd_z = vector(rk3Sfd_z, i, j, k) / rk3sdetg;
					rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yz, i, j, k)) + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k));
					rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j, k) + rk3gtu_yz * vector(vfd_y, i, j, k) + rk3gtu_zz * vector(vfd_z, i, j, k));
					rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j, k) + rk3gtu_yy * vector(rk3Sfd_y, i, j, k) + rk3gtu_yz * vector(rk3Sfd_z, i, j, k));
					rk3Su_y = rk3Sfu_y / rk3sdetg;
					rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j, k) + rk3gtu_yz * vector(rk3Sfd_y, i, j, k) + rk3gtu_zz * vector(rk3Sfd_z, i, j, k));
					rk3Su_z = rk3Sfu_z / rk3sdetg;
					rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j, k) + rk3gtu_yy * vector(vfd_y, i, j, k) + rk3gtu_yz * vector(vfd_z, i, j, k));
					rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j, k) + rk3vfu_y * vector(vfd_y, i, j, k) + rk3vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
					rk3sqW = 1.0 / (1.0 - rk3sqvf);
					rk3W = sqrt(rk3sqW);
					rk3Bd_x = rk3Bfd_x / rk3sdetg;
					rk3D = vector(rk3Df, i, j, k) / rk3sdetg;
					rk3Bd_y = rk3Bfd_y / rk3sdetg;
					rk3Bd_z = rk3Bfd_z / rk3sdetg;
					vector(t_chie, i, j, k) = vector(chie, i, j, k);
					vector(t_chia, i, j, k) = vector(chia, i, j, k);
					vector(t_optdepthx, i, j, k) = vector(optdepthx, i, j, k);
					vector(t_optdeptha, i, j, k) = vector(optdeptha, i, j, k);
					vector(t_optdepthe, i, j, k) = vector(optdepthe, i, j, k);
					vector(rk3Alpha, i, j, k) = fabs(vector(rk3Alpha, i, j, k));
					rhof_tmp = vector(rhof, i, j, k);
					Yef_tmp = vector(rk3DYf, i, j, k) / vector(rk3Df, i, j, k);
					if (Yef_tmp < vacuum_ye) {
						vector(rk3DYf, i, j, k) = vacuum_ye_reset * vector(rk3Df, i, j, k);
						Yef_tmp = vacuum_ye_reset;
					}
					if (Yef_tmp > ye_maximum) {
						vector(rk3DYf, i, j, k) = 0.9 * ye_maximum * vector(rk3Df, i, j, k);
						Yef_tmp = 0.9 * ye_maximum;
					}
					Tf_tmp = vector(Tf, i, j, k);
					pf_tmp = vector(pf, i, j, k);
					epsf_tmp = vector(epsf, i, j, k);
					vfd_x_tmp = vector(vfd_x, i, j, k);
					vfd_y_tmp = vector(vfd_y, i, j, k);
					vfd_z_tmp = vector(vfd_z, i, j, k);
					sqcs_tmp = vector(sqcs, i, j, k);
					dpfdrho_tmp = vector(dpfdrho, i, j, k);
					dpfdeps_tmp = vector(dpfdeps, i, j, k);
					dpfdye_tmp = vector(dpfdye, i, j, k);
					Sfd_x_tmp = vector(rk3Sfd_x, i, j, k);
					Sfd_y_tmp = vector(rk3Sfd_y, i, j, k);
					Sfd_z_tmp = vector(rk3Sfd_z, i, j, k);
					tauf_tmp = vector(rk3tauf, i, j, k);
					Df_tmp = vector(rk3Df, i, j, k);
					DYf_tmp = vector(rk3DYf, i, j, k);
					Bfu_x_tmp = vector(rk3Bfu_x, i, j, k);
					Bfu_y_tmp = vector(rk3Bfu_y, i, j, k);
					Bfu_z_tmp = vector(rk3Bfu_z, i, j, k);
					if (equalsEq(externalCon2Prim, 1.0)) {
						Efu_x_tmp = vector(Efu_x, i, j, k);
						Efu_y_tmp = vector(Efu_y, i, j, k);
						Efu_z_tmp = vector(Efu_z, i, j, k);
						externalCon2prim(Efu_x_tmp, Efu_y_tmp, Efu_z_tmp, rhof_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, Df_tmp, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tauf_tmp, rk3W, vector(rk3chi, i, j, k), rk3gtu_xx, rk3gtu_xy, rk3gtu_xz, rk3gtu_yy, rk3gtu_yz, rk3gtu_zz, vector(rk3gtd_xx, i, j, k), vector(rk3gtd_xy, i, j, k), vector(rk3gtd_xz, i, j, k), vector(rk3gtd_yy, i, j, k), vector(rk3gtd_yz, i, j, k), vector(rk3gtd_zz, i, j, k), Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, dx, simPlat_dt, ilast, jlast);
						vector(Efu_x, i, j, k) = Efu_x_tmp;
						vector(Efu_y, i, j, k) = Efu_y_tmp;
						vector(Efu_z, i, j, k) = Efu_z_tmp;
					} else {
						vacuum_D_tmp = vacuum_D * rk3decay_factor;
						vacuum_tau_tmp = vacuum_tau * rk3decay_factor;
						vacuum_rho_reset_tmp = vacuum_rho_reset * rk3decay_factor;
						con2prim(eos_type, rhof_tmp, Yef_tmp, Tf_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, dpfdye_tmp, rho_0, rho_1, rho_2, a_0, a_1, a_2, a_3, K_0, K_1, K_2, K_3, gamma_0, gamma_1, gamma_2, gamma_3, rk3D, vacuum_D_tmp, Df_tmp, DYf_tmp, rk3sdetg, rk3Sd_x, rk3Sd_y, rk3Sd_z, rk3Su_x, rk3Su_y, rk3Su_z, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, rk3tau, vacuum_tau_tmp, tauf_tmp, rk3W, rk3h, vector(rk3chi, i, j, k), rk3gtu_xx, rk3gtu_xy, rk3gtu_xz, rk3gtu_yy, rk3gtu_yz, rk3gtu_zz, vector(rk3gtd_xx, i, j, k), vector(rk3gtd_xy, i, j, k), vector(rk3gtd_xz, i, j, k), vector(rk3gtd_yy, i, j, k), vector(rk3gtd_yz, i, j, k), vector(rk3gtd_zz, i, j, k), rk3Bd_x, rk3Bd_y, rk3Bd_z, rk3Bu_x, rk3Bu_y, rk3Bu_z, Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, max_error, gamma, rk3inv_chi, threshold_vacuum, threshold_sqSmax, threshold_sqBmax, energyShift, vacuum_ye_beta, vacuum_rho, vacuum_rho_reset_tmp, vacuum_P_reset, vacuum_ye_reset, vacuum_temp_reset, vacuum_tau_reset, minTableTemperature, minTableEnergy, coord0_data_Data, coord0_dx_Data, coord0_size_Data, coord1_data_Data, coord1_dx_Data, coord1_size_Data, coord2_data_Data, coord2_dx_Data, coord2_size_Data, vars_data_Data, vars_size_Data, dx, simPlat_dt, ilast, jlast);
						vector(Efu_x, i, j, k) = -((-vector(vfd_z, i, j, k) * rk3Bfd_y) + vector(vfd_y, i, j, k) * rk3Bfd_z) / (rk3sdetg * rk3sdetg);
						vector(Efu_y, i, j, k) = -(vector(vfd_z, i, j, k) * rk3Bfd_x - vector(vfd_x, i, j, k) * rk3Bfd_z) / (rk3sdetg * rk3sdetg);
						vector(Efu_z, i, j, k) = -((-vector(vfd_y, i, j, k) * rk3Bfd_x) + vector(vfd_x, i, j, k) * rk3Bfd_y) / (rk3sdetg * rk3sdetg);
					}
					vector(rhof, i, j, k) = rhof_tmp;
					vector(Yef, i, j, k) = Yef_tmp;
					vector(Tf, i, j, k) = Tf_tmp;
					vector(pf, i, j, k) = pf_tmp;
					vector(epsf, i, j, k) = epsf_tmp;
					vector(vfd_x, i, j, k) = vfd_x_tmp;
					vector(vfd_y, i, j, k) = vfd_y_tmp;
					vector(vfd_z, i, j, k) = vfd_z_tmp;
					vector(sqcs, i, j, k) = sqcs_tmp;
					vector(dpfdye, i, j, k) = dpfdye_tmp;
					vector(dpfdeps, i, j, k) = dpfdeps_tmp;
					vector(dpfdrho, i, j, k) = dpfdrho_tmp;
					vector(rk3Sfd_x, i, j, k) = Sfd_x_tmp;
					vector(rk3Sfd_y, i, j, k) = Sfd_y_tmp;
					vector(rk3Sfd_z, i, j, k) = Sfd_z_tmp;
					vector(rk3tauf, i, j, k) = tauf_tmp;
					vector(rk3Df, i, j, k) = Df_tmp;
					vector(rk3DYf, i, j, k) = DYf_tmp;
					vector(rk3Bfu_x, i, j, k) = Bfu_x_tmp;
					vector(rk3Bfu_y, i, j, k) = Bfu_y_tmp;
					vector(rk3Bfu_z, i, j, k) = Bfu_z_tmp;
					vector(t_chix, i, j, k) = vector(chix, i, j, k);
				}
			}
		}
	}
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* rk3DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3DYf_id).get())->getPointer();
		double* rk3Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Df_id).get())->getPointer();
		double* rk3Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_x_id).get())->getPointer();
		double* rk3Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_y_id).get())->getPointer();
		double* rk3Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_z_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* rk3chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3chi_id).get())->getPointer();
		double* rk3gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xy_id).get())->getPointer();
		double* rk3gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_yy_id).get())->getPointer();
		double* rk3gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_yz_id).get())->getPointer();
		double* rk3gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xz_id).get())->getPointer();
		double* rk3gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_zz_id).get())->getPointer();
		double* rk3gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xx_id).get())->getPointer();
		double* rk3Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_x_id).get())->getPointer();
		double* rk3Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_y_id).get())->getPointer();
		double* rk3Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_z_id).get())->getPointer();
		double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
		double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
		double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
		double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
		double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
		double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
		double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
		double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
		double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
		double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
		double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
		double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
		double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
		double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
		double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
		double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
		double* qnu_e = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_e_id).get())->getPointer();
		double* qnu_a = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_a_id).get())->getPointer();
		double* qnu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_x_id).get())->getPointer();
		double* rk3Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_x_id).get())->getPointer();
		double* rk3Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_y_id).get())->getPointer();
		double* rk3Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_z_id).get())->getPointer();
		double* rk3Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Alpha_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double d_optdeptheh_o0_t0_m0_l0, d_optdepthev_o0_t0_m0_l0, d_optdepthed_o0_t0_m0_l0, d_optdepthah_o0_t0_m0_l0, d_optdepthav_o0_t0_m0_l0, d_optdepthad_o0_t0_m0_l0, d_optdepthxh_o0_t0_m0_l0, d_optdepthxv_o0_t0_m0_l0, d_optdepthxd_o0_t0_m0_l0, d_chieh_o0_t0_m0_l0, d_chiev_o0_t0_m0_l0, d_chied_o0_t0_m0_l0, d_chiah_o0_t0_m0_l0, d_chiav_o0_t0_m0_l0, d_chiad_o0_t0_m0_l0, d_chixh_o0_t0_m0_l0, d_chixv_o0_t0_m0_l0, d_chixd_o0_t0_m0_l0, d_CovdDYf_z_o0_t0_m0_l0, d_CovdDYf_y_o0_t0_m0_l0, d_CovdDYf_x_o0_t0_m0_l0, d_CovdDf_z_o0_t0_m0_l0, d_CovdDf_y_o0_t0_m0_l0, d_CovdDf_x_o0_t0_m0_l0, d_CovdBfu_xx_o0_t0_m0_l0, d_CovdBfu_xy_o0_t0_m0_l0, d_CovdBfu_xz_o0_t0_m0_l0, d_CovdBfu_yx_o0_t0_m0_l0, d_CovdBfu_yy_o0_t0_m0_l0, d_CovdBfu_yz_o0_t0_m0_l0, d_CovdBfu_zx_o0_t0_m0_l0, d_CovdBfu_zy_o0_t0_m0_l0, d_CovdBfu_zz_o0_t0_m0_l0, d_CovdEfu_xx_o0_t0_m0_l0, d_CovdEfu_xy_o0_t0_m0_l0, d_CovdEfu_xz_o0_t0_m0_l0, d_CovdEfu_yx_o0_t0_m0_l0, d_CovdEfu_yy_o0_t0_m0_l0, d_CovdEfu_yz_o0_t0_m0_l0, d_CovdEfu_zx_o0_t0_m0_l0, d_CovdEfu_zy_o0_t0_m0_l0, d_CovdEfu_zz_o0_t0_m0_l0, d_Covdepsf_x_o0_t0_m0_l0, d_Covdepsf_y_o0_t0_m0_l0, d_Covdepsf_z_o0_t0_m0_l0, d_Covddpfdeps_x_o0_t0_m0_l0, d_Covddpfdeps_y_o0_t0_m0_l0, d_Covddpfdeps_z_o0_t0_m0_l0, d_Covddpfdye_x_o0_t0_m0_l0, d_Covddpfdye_y_o0_t0_m0_l0, d_Covddpfdye_z_o0_t0_m0_l0, d_Covddpfdrho_x_o0_t0_m0_l0, d_Covddpfdrho_y_o0_t0_m0_l0, d_Covddpfdrho_z_o0_t0_m0_l0, d_Covdvfd_xx_o0_t0_m0_l0, d_Covdvfd_xy_o0_t0_m0_l0, d_Covdvfd_xz_o0_t0_m0_l0, d_Covdvfd_yx_o0_t0_m0_l0, d_Covdvfd_yy_o0_t0_m0_l0, d_Covdvfd_yz_o0_t0_m0_l0, d_Covdvfd_zx_o0_t0_m0_l0, d_Covdvfd_zy_o0_t0_m0_l0, d_Covdvfd_zz_o0_t0_m0_l0, d_Covdrhof_x_o0_t0_m0_l0, d_Covdrhof_y_o0_t0_m0_l0, d_Covdrhof_z_o0_t0_m0_l0, d_Covdpf_x_o0_t0_m0_l0, d_Covdpf_y_o0_t0_m0_l0, d_Covdpf_z_o0_t0_m0_l0, d_CovdYef_x_o0_t0_m0_l0, d_CovdYef_y_o0_t0_m0_l0, d_CovdYef_z_o0_t0_m0_l0, rk3chiev, rk3chieh, rk3Covdvfd_xx, rk3Covdvfd_xz, rk3Covdvfd_xy, rk3chied, rk3Covddpfdeps_x, rk3Covddpfdeps_y, rk3Covddpfdeps_z, rk3Covdvfd_zx, rk3Covdvfd_zz, rk3Covdvfd_zy, rk3Covdvfd_yy, rk3Covdvfd_yx, rk3Covdvfd_yz, rk3Covdpf_x, rk3Covdpf_y, rk3Covdpf_z, rk3chi_max, rk3inv_chi, rk3Bfd_y, rk3Bfd_z, rk3Bfd_x, rk3sqBf, rk3chixh, rk3chixd, rk3Covddpfdrho_z, rk3Covddpfdrho_y, rk3Covddpfdrho_x, rk3sdetg, rk3m_CovdDf_x_o0_t0_l0, rk3CovdDf_x, rk3m_CovdDf_y_o0_t0_l0, rk3CovdDf_y, rk3m_CovdDf_z_o0_t0_l0, rk3CovdDf_z, rk3m_CovdBfu_zz_o0_t0_l0, rk3CovdBfu_zz, rk3m_CovdBfu_zy_o0_t0_l0, rk3CovdBfu_zy, rk3m_CovdBfu_zx_o0_t0_l0, rk3CovdBfu_zx, rk3m_CovdBfu_yz_o0_t0_l0, rk3CovdBfu_yz, rk3m_CovdBfu_yy_o0_t0_l0, rk3CovdBfu_yy, rk3m_CovdBfu_yx_o0_t0_l0, rk3CovdBfu_yx, rk3m_CovdBfu_xy_o0_t0_l0, rk3CovdBfu_xy, rk3m_CovdBfu_xx_o0_t0_l0, rk3CovdBfu_xx, rk3m_CovdBfu_xz_o0_t0_l0, rk3CovdBfu_xz, rk3CovdBfd_zx, rk3Covdrhof_z, rk3Covdrhof_y, rk3Covdrhof_x, rk3CovdBfd_zz, rk3CovdBfd_zy, rk3m_CovdDYf_x_o0_t0_l0, rk3CovdDYf_x, rk3m_CovdDYf_y_o0_t0_l0, rk3CovdDYf_y, rk3m_CovdDYf_z_o0_t0_l0, rk3CovdDYf_z, rk3Efd_z, rk3Efd_y, rk3chixv, rk3Efd_x, rk3sqEf, rk3optdepthed, rk3CovdBfd_yy, rk3CovdBfd_yx, rk3CovdBfd_yz, rk3optdeptheh, rk3CovdBvf_z, rk3optdepthev, rk3CovdBvf_y, rk3CovdBvf_x, rk3CovdBfd_xx, rk3CovdBfd_xz, rk3CovdBfd_xy, rk3optdepthxd, rk3optdepthxh, rk3optdepthxv, rk3chiav, rk3CovdEfu_zz, rk3CovdEfu_zy, rk3CovdEfu_zx, rk3CovdEfd_zz, rk3CovdEfd_zx, rk3CovdEfd_zy, rk3Bvf, rk3optdepthah, rk3optdepthad, rk3h, rk3chiah, rk3chiad, rk3CovdYef_y, rk3optdepthav, rk3CovdYef_z, rk3CovdYef_x, rk3detgtd, rk3idetgtd, rk3gtu_zz, rk3gtu_yy, rk3gtu_yz, rk3gtu_xx, rk3gtu_xz, rk3CovuBfu_zx, rk3CovuBfu_zz, rk3CovuBfu_zy, rk3Covdvfu_xz, rk3CovuYef_z, rk3Covdvfu_zz, rk3Covdvfu_yz, rk3Covuvfu_zz, rk3Sfu_z, rk3Covurhof_z, rk3CovuBvf_z, rk3gtu_xy, rk3CovuBfu_yz, rk3CovuBfu_yy, rk3CovuBfu_yx, rk3Covdvfu_xy, rk3Covdvfu_xx, rk3CovdinvsqW_x, rk3CovuYef_x, rk3CovuYef_y, rk3Covdvfu_zy, rk3Covdvfu_zx, rk3CovdinvsqW_z, rk3Covdvfu_yx, rk3Covuvfu_zx, rk3Covdvfu_yy, rk3CovdinvsqW_y, rk3CovuinvsqW_z, rk3Covuvfu_zy, rk3CovuBfu_xz, rk3CovuBfu_xy, rk3CovuBfu_xx, rk3Sfu_y, rk3Sfu_x, rk3Covurhof_x, rk3Covurhof_y, rk3Covuvfu_yz, rk3Covuvfu_yy, rk3Covuvfu_yx, rk3Covuvfu_xz, rk3Covuvfu_xy, rk3Covuvfu_xx, rk3vfu_y, rk3vfu_x, rk3CovuBvf_x, rk3CovuBvf_y, rk3CovuinvsqW_y, rk3CovuinvsqW_x, rk3Covddpfdye_z, rk3Covddpfdye_y, rk3Covddpfdye_x, rk3CovdEfu_xz, rk3CovdEfu_xy, rk3CovdEfu_xx, rk3CovdEfd_xx, rk3CovdEfd_xy, rk3CovdEfd_xz, rk3vfu_z, rk3sqvf, rk3sqW, rk3PhiAh, rk3Epf, rk3Thetaf, rk3Covdepsf_z, rk3Covdhf_z, rk3CovdEpf_z, rk3CovdThetaf_z, rk3Covdepsf_y, rk3Covdhf_y, rk3CovdEpf_y, rk3CovdThetaf_y, rk3Covdepsf_x, rk3Covuepsf_z, rk3Covuepsf_y, rk3Covdhf_x, rk3CovdEpf_x, rk3CovdThetaf_x, rk3CovuThetaf_z, rk3CovuThetaf_x, rk3CovuThetaf_y, rk3Phivh_z, rk3Phivh_x, rk3Phivh_y, rk3PhiMh_yz, rk3PhiMh_xz, rk3PhiMh_xy, rk3Covuepsf_x, rk3faceta, rk3CovdEfu_yz, rk3CovuEfu_zz, rk3CovuEfu_xz, rk3CovdEfu_yy, rk3CovuEfu_zy, rk3CovuEfu_xy, rk3CovdEfu_yx, rk3CovdEfd_yy, rk3CovdEfd_yz, rk3CovdEfd_yx, rk3CovuEfu_zx, rk3CovuEfu_xx, rk3CovuEfu_yx, rk3CovuEfu_yz, rk3CovuEfu_yy, rk3PhiThetah, rk3HPres, rk3HTheta, rk3Hv_z, rk3HN_z, rk3HNe_z, rk3Hv_y, rk3HN_y, rk3HM_yz, rk3HNe_y, rk3Hv_x, rk3HN_x, rk3HM_xz, rk3HM_xy, rk3HE_z, rk3HE_x, rk3HE_y, rk3HT_zz, rk3HT_yy, rk3HT_yz, rk3HT_xx, rk3HT_xz, rk3HT_xy, rk3HNe_x, d_TauN_x_o0_t0_m0_l0, d_TauN_y_o0_t0_m0_l0, d_TauN_z_o0_t0_m0_l0, d_TauNe_x_o0_t0_m0_l0, d_TauNe_y_o0_t0_m0_l0, d_TauNe_z_o0_t0_m0_l0, d_TauM_xy_o0_t0_m0_l0, d_TauM_xz_o0_t0_m0_l0, d_TauM_yz_o0_t0_m0_l0, d_TauT_xx_o0_t0_m0_l0, d_TauT_xy_o0_t0_m0_l0, d_TauT_xz_o0_t0_m0_l0, d_TauT_yy_o0_t0_m0_l0, d_TauT_yz_o0_t0_m0_l0, d_TauT_zz_o0_t0_m0_l0, optdepthe_tmp, optdeptha_tmp, optdepthx_tmp, chie_tmp, chia_tmp, chix_tmp, qnu_tmp, rnu_tmp, qnu_e_tmp, qnu_a_tmp, qnu_x_tmp, betabeta, betav, velr, f_redshift;
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					if ((i + 2 < ilast && i - 2 >= 0 && j + 2 < jlast && j - 2 >= 0 && k + 2 < klast && k - 2 >= 0) && ((i + 3 < ilast || !patch->getPatchGeometry()->getTouchesRegularBoundary(0, 1)) && (i - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(0, 0)) && (j + 3 < jlast || !patch->getPatchGeometry()->getTouchesRegularBoundary(1, 1)) && (j - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(1, 0)) && (k + 3 < klast || !patch->getPatchGeometry()->getTouchesRegularBoundary(2, 1)) && (k - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(2, 0)))) {
						d_optdeptheh_o0_t0_m0_l0 = eikonal_i(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthev_o0_t0_m0_l0 = eikonal_j(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthed_o0_t0_m0_l0 = eikonal_k(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthah_o0_t0_m0_l0 = eikonal_i(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthav_o0_t0_m0_l0 = eikonal_j(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthad_o0_t0_m0_l0 = eikonal_k(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthxh_o0_t0_m0_l0 = eikonal_i(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthxv_o0_t0_m0_l0 = eikonal_j(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthxd_o0_t0_m0_l0 = eikonal_k(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chieh_o0_t0_m0_l0 = eikonal_i(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiev_o0_t0_m0_l0 = eikonal_j(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chied_o0_t0_m0_l0 = eikonal_k(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiah_o0_t0_m0_l0 = eikonal_i(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiav_o0_t0_m0_l0 = eikonal_j(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiad_o0_t0_m0_l0 = eikonal_k(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chixh_o0_t0_m0_l0 = eikonal_i(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chixv_o0_t0_m0_l0 = eikonal_j(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chixd_o0_t0_m0_l0 = eikonal_k(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
						if (calculate_les_terms) {
							d_CovdDYf_z_o0_t0_m0_l0 = D1CDO4_k(rk3DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDYf_y_o0_t0_m0_l0 = D1CDO4_j(rk3DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDYf_x_o0_t0_m0_l0 = D1CDO4_i(rk3DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDf_z_o0_t0_m0_l0 = D1CDO4_k(rk3Df, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDf_y_o0_t0_m0_l0 = D1CDO4_j(rk3Df, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDf_x_o0_t0_m0_l0 = D1CDO4_i(rk3Df, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_xx_o0_t0_m0_l0 = D1CDO4_i(rk3Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_xy_o0_t0_m0_l0 = D1CDO4_i(rk3Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_xz_o0_t0_m0_l0 = D1CDO4_i(rk3Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_yx_o0_t0_m0_l0 = D1CDO4_j(rk3Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_yy_o0_t0_m0_l0 = D1CDO4_j(rk3Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_yz_o0_t0_m0_l0 = D1CDO4_j(rk3Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_zx_o0_t0_m0_l0 = D1CDO4_k(rk3Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_zy_o0_t0_m0_l0 = D1CDO4_k(rk3Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_zz_o0_t0_m0_l0 = D1CDO4_k(rk3Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_xx_o0_t0_m0_l0 = D1CDO4_i(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_xy_o0_t0_m0_l0 = D1CDO4_i(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_xz_o0_t0_m0_l0 = D1CDO4_i(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_yx_o0_t0_m0_l0 = D1CDO4_j(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_yy_o0_t0_m0_l0 = D1CDO4_j(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_yz_o0_t0_m0_l0 = D1CDO4_j(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_zx_o0_t0_m0_l0 = D1CDO4_k(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_zy_o0_t0_m0_l0 = D1CDO4_k(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_zz_o0_t0_m0_l0 = D1CDO4_k(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdepsf_x_o0_t0_m0_l0 = D1CDO4_i(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdepsf_y_o0_t0_m0_l0 = D1CDO4_j(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdepsf_z_o0_t0_m0_l0 = D1CDO4_k(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdeps_x_o0_t0_m0_l0 = D1CDO4_i(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdeps_y_o0_t0_m0_l0 = D1CDO4_j(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdeps_z_o0_t0_m0_l0 = D1CDO4_k(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdye_x_o0_t0_m0_l0 = D1CDO4_i(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdye_y_o0_t0_m0_l0 = D1CDO4_j(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdye_z_o0_t0_m0_l0 = D1CDO4_k(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdrho_x_o0_t0_m0_l0 = D1CDO4_i(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdrho_y_o0_t0_m0_l0 = D1CDO4_j(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdrho_z_o0_t0_m0_l0 = D1CDO4_k(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_xx_o0_t0_m0_l0 = D1CDO4_i(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_xy_o0_t0_m0_l0 = D1CDO4_i(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_xz_o0_t0_m0_l0 = D1CDO4_i(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_yx_o0_t0_m0_l0 = D1CDO4_j(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_yy_o0_t0_m0_l0 = D1CDO4_j(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_yz_o0_t0_m0_l0 = D1CDO4_j(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_zx_o0_t0_m0_l0 = D1CDO4_k(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_zy_o0_t0_m0_l0 = D1CDO4_k(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_zz_o0_t0_m0_l0 = D1CDO4_k(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdrhof_x_o0_t0_m0_l0 = D1CDO4_i(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdrhof_y_o0_t0_m0_l0 = D1CDO4_j(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdrhof_z_o0_t0_m0_l0 = D1CDO4_k(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdpf_x_o0_t0_m0_l0 = D1CDO4_i(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdpf_y_o0_t0_m0_l0 = D1CDO4_j(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdpf_z_o0_t0_m0_l0 = D1CDO4_k(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdYef_x_o0_t0_m0_l0 = D1CDO4_i(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdYef_y_o0_t0_m0_l0 = D1CDO4_j(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdYef_z_o0_t0_m0_l0 = D1CDO4_k(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
						}
						rk3chiev = d_chiev_o0_t0_m0_l0;
						rk3chieh = d_chieh_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk3Covdvfd_xx = d_Covdvfd_xx_o0_t0_m0_l0;
							rk3Covdvfd_xz = d_Covdvfd_xz_o0_t0_m0_l0;
							rk3Covdvfd_xy = d_Covdvfd_xy_o0_t0_m0_l0;
						} else {
							rk3Covdvfd_xx = 0.0;
							rk3Covdvfd_xz = 0.0;
							rk3Covdvfd_xy = 0.0;
						}
						rk3chied = d_chied_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk3Covddpfdeps_x = d_Covddpfdeps_x_o0_t0_m0_l0;
							rk3Covddpfdeps_y = d_Covddpfdeps_y_o0_t0_m0_l0;
							rk3Covddpfdeps_z = d_Covddpfdeps_z_o0_t0_m0_l0;
							rk3Covdvfd_zx = d_Covdvfd_zx_o0_t0_m0_l0;
							rk3Covdvfd_zz = d_Covdvfd_zz_o0_t0_m0_l0;
							rk3Covdvfd_zy = d_Covdvfd_zy_o0_t0_m0_l0;
							rk3Covdvfd_yy = d_Covdvfd_yy_o0_t0_m0_l0;
							rk3Covdvfd_yx = d_Covdvfd_yx_o0_t0_m0_l0;
							rk3Covdvfd_yz = d_Covdvfd_yz_o0_t0_m0_l0;
							rk3Covdpf_x = d_Covdpf_x_o0_t0_m0_l0;
							rk3Covdpf_y = d_Covdpf_y_o0_t0_m0_l0;
							rk3Covdpf_z = d_Covdpf_z_o0_t0_m0_l0;
						} else {
							rk3Covddpfdeps_x = 0.0;
							rk3Covddpfdeps_y = 0.0;
							rk3Covddpfdeps_z = 0.0;
							rk3Covdvfd_zx = 0.0;
							rk3Covdvfd_zz = 0.0;
							rk3Covdvfd_zy = 0.0;
							rk3Covdvfd_yy = 0.0;
							rk3Covdvfd_yx = 0.0;
							rk3Covdvfd_yz = 0.0;
							rk3Covdpf_x = 0.0;
							rk3Covdpf_y = 0.0;
							rk3Covdpf_z = 0.0;
						}
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j, k));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_yy, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_zz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_xy, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_xz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3sqBf = vector(rk3Bfu_x, i, j, k) * rk3Bfd_x + vector(rk3Bfu_y, i, j, k) * rk3Bfd_y + vector(rk3Bfu_z, i, j, k) * rk3Bfd_z;
						rk3chixh = d_chixh_o0_t0_m0_l0;
						rk3chixd = d_chixd_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk3Covddpfdrho_z = d_Covddpfdrho_z_o0_t0_m0_l0;
							rk3Covddpfdrho_y = d_Covddpfdrho_y_o0_t0_m0_l0;
							rk3Covddpfdrho_x = d_Covddpfdrho_x_o0_t0_m0_l0;
						} else {
							rk3Covddpfdrho_z = 0.0;
							rk3Covddpfdrho_y = 0.0;
							rk3Covddpfdrho_x = 0.0;
						}
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						if (calculate_les_terms) {
							rk3m_CovdDf_x_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdDf_x_o0_t0_m0_l0;
							rk3CovdDf_x = rk3m_CovdDf_x_o0_t0_l0;
							rk3m_CovdDf_y_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdDf_y_o0_t0_m0_l0;
							rk3CovdDf_y = rk3m_CovdDf_y_o0_t0_l0;
							rk3m_CovdDf_z_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdDf_z_o0_t0_m0_l0;
							rk3CovdDf_z = rk3m_CovdDf_z_o0_t0_l0;
							rk3m_CovdBfu_zz_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdBfu_zz_o0_t0_m0_l0;
							rk3CovdBfu_zz = rk3m_CovdBfu_zz_o0_t0_l0;
							rk3m_CovdBfu_zy_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdBfu_zy_o0_t0_m0_l0;
							rk3CovdBfu_zy = rk3m_CovdBfu_zy_o0_t0_l0;
							rk3m_CovdBfu_zx_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdBfu_zx_o0_t0_m0_l0;
							rk3CovdBfu_zx = rk3m_CovdBfu_zx_o0_t0_l0;
							rk3m_CovdBfu_yz_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdBfu_yz_o0_t0_m0_l0;
							rk3CovdBfu_yz = rk3m_CovdBfu_yz_o0_t0_l0;
							rk3m_CovdBfu_yy_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdBfu_yy_o0_t0_m0_l0;
							rk3CovdBfu_yy = rk3m_CovdBfu_yy_o0_t0_l0;
							rk3m_CovdBfu_yx_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdBfu_yx_o0_t0_m0_l0;
							rk3CovdBfu_yx = rk3m_CovdBfu_yx_o0_t0_l0;
							rk3m_CovdBfu_xy_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdBfu_xy_o0_t0_m0_l0;
							rk3CovdBfu_xy = rk3m_CovdBfu_xy_o0_t0_l0;
							rk3m_CovdBfu_xx_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdBfu_xx_o0_t0_m0_l0;
							rk3CovdBfu_xx = rk3m_CovdBfu_xx_o0_t0_l0;
							rk3m_CovdBfu_xz_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdBfu_xz_o0_t0_m0_l0;
							rk3CovdBfu_xz = rk3m_CovdBfu_xz_o0_t0_l0;
							rk3CovdBfd_zx = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * rk3CovdBfu_zx + vector(rk3gtd_xy, i, j, k) * rk3CovdBfu_zy + vector(rk3gtd_xz, i, j, k) * rk3CovdBfu_zz);
							rk3Covdrhof_z = d_Covdrhof_z_o0_t0_m0_l0;
							rk3Covdrhof_y = d_Covdrhof_y_o0_t0_m0_l0;
							rk3Covdrhof_x = d_Covdrhof_x_o0_t0_m0_l0;
							rk3CovdBfd_zz = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * rk3CovdBfu_zx + vector(rk3gtd_yz, i, j, k) * rk3CovdBfu_zy + vector(rk3gtd_zz, i, j, k) * rk3CovdBfu_zz);
							rk3CovdBfd_zy = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * rk3CovdBfu_zx + vector(rk3gtd_yy, i, j, k) * rk3CovdBfu_zy + vector(rk3gtd_yz, i, j, k) * rk3CovdBfu_zz);
							rk3m_CovdDYf_x_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdDYf_x_o0_t0_m0_l0;
							rk3CovdDYf_x = rk3m_CovdDYf_x_o0_t0_l0;
							rk3m_CovdDYf_y_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdDYf_y_o0_t0_m0_l0;
							rk3CovdDYf_y = rk3m_CovdDYf_y_o0_t0_l0;
							rk3m_CovdDYf_z_o0_t0_l0 = 1.0 / rk3sdetg * d_CovdDYf_z_o0_t0_m0_l0;
							rk3CovdDYf_z = rk3m_CovdDYf_z_o0_t0_l0;
						} else {
							rk3CovdDf_x = 0.0;
							rk3CovdDf_y = 0.0;
							rk3CovdDf_z = 0.0;
							rk3CovdBfu_zz = 0.0;
							rk3CovdBfu_zy = 0.0;
							rk3CovdBfu_zx = 0.0;
							rk3CovdBfu_yz = 0.0;
							rk3CovdBfu_yy = 0.0;
							rk3CovdBfu_yx = 0.0;
							rk3CovdBfu_xy = 0.0;
							rk3CovdBfu_xx = 0.0;
							rk3CovdBfu_xz = 0.0;
							rk3CovdBfd_zx = 0.0;
							rk3Covdrhof_z = 0.0;
							rk3Covdrhof_y = 0.0;
							rk3Covdrhof_x = 0.0;
							rk3CovdBfd_zz = 0.0;
							rk3CovdBfd_zy = 0.0;
							rk3CovdDYf_x = 0.0;
							rk3CovdDYf_y = 0.0;
							rk3CovdDYf_z = 0.0;
						}
						rk3Efd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * vector(Efu_x, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(Efu_y, i, j, k) + vector(rk3gtd_zz, i, j, k) * vector(Efu_z, i, j, k));
						rk3Efd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * vector(Efu_x, i, j, k) + vector(rk3gtd_yy, i, j, k) * vector(Efu_y, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(Efu_z, i, j, k));
						rk3chixv = d_chixv_o0_t0_m0_l0;
						rk3Efd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * vector(Efu_x, i, j, k) + vector(rk3gtd_xy, i, j, k) * vector(Efu_y, i, j, k) + vector(rk3gtd_xz, i, j, k) * vector(Efu_z, i, j, k));
						rk3sqEf = vector(Efu_x, i, j, k) * rk3Efd_x + vector(Efu_y, i, j, k) * rk3Efd_y + vector(Efu_z, i, j, k) * rk3Efd_z;
						rk3optdepthed = d_optdepthed_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk3CovdBfd_yy = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * rk3CovdBfu_yx + vector(rk3gtd_yy, i, j, k) * rk3CovdBfu_yy + vector(rk3gtd_yz, i, j, k) * rk3CovdBfu_yz);
							rk3CovdBfd_yx = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * rk3CovdBfu_yx + vector(rk3gtd_xy, i, j, k) * rk3CovdBfu_yy + vector(rk3gtd_xz, i, j, k) * rk3CovdBfu_yz);
							rk3CovdBfd_yz = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * rk3CovdBfu_yx + vector(rk3gtd_yz, i, j, k) * rk3CovdBfu_yy + vector(rk3gtd_zz, i, j, k) * rk3CovdBfu_yz);
						} else {
							rk3CovdBfd_yy = 0.0;
							rk3CovdBfd_yx = 0.0;
							rk3CovdBfd_yz = 0.0;
						}
						rk3optdeptheh = d_optdeptheh_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk3CovdBvf_z = vector(rk3Bfu_x, i, j, k) * rk3Covdvfd_zx + vector(vfd_x, i, j, k) * rk3CovdBfu_zx + vector(rk3Bfu_y, i, j, k) * rk3Covdvfd_zy + vector(vfd_y, i, j, k) * rk3CovdBfu_zy + vector(rk3Bfu_z, i, j, k) * rk3Covdvfd_zz + vector(vfd_z, i, j, k) * rk3CovdBfu_zz;
						} else {
							rk3CovdBvf_z = 0.0;
						}
						rk3optdepthev = d_optdepthev_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk3CovdBvf_y = vector(rk3Bfu_x, i, j, k) * rk3Covdvfd_yx + vector(vfd_x, i, j, k) * rk3CovdBfu_yx + vector(rk3Bfu_y, i, j, k) * rk3Covdvfd_yy + vector(vfd_y, i, j, k) * rk3CovdBfu_yy + vector(rk3Bfu_z, i, j, k) * rk3Covdvfd_yz + vector(vfd_z, i, j, k) * rk3CovdBfu_yz;
							rk3CovdBvf_x = vector(rk3Bfu_x, i, j, k) * rk3Covdvfd_xx + vector(vfd_x, i, j, k) * rk3CovdBfu_xx + vector(rk3Bfu_y, i, j, k) * rk3Covdvfd_xy + vector(vfd_y, i, j, k) * rk3CovdBfu_xy + vector(rk3Bfu_z, i, j, k) * rk3Covdvfd_xz + vector(vfd_z, i, j, k) * rk3CovdBfu_xz;
							rk3CovdBfd_xx = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * rk3CovdBfu_xx + vector(rk3gtd_xy, i, j, k) * rk3CovdBfu_xy + vector(rk3gtd_xz, i, j, k) * rk3CovdBfu_xz);
							rk3CovdBfd_xz = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * rk3CovdBfu_xx + vector(rk3gtd_yz, i, j, k) * rk3CovdBfu_xy + vector(rk3gtd_zz, i, j, k) * rk3CovdBfu_xz);
							rk3CovdBfd_xy = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * rk3CovdBfu_xx + vector(rk3gtd_yy, i, j, k) * rk3CovdBfu_xy + vector(rk3gtd_yz, i, j, k) * rk3CovdBfu_xz);
						} else {
							rk3CovdBvf_y = 0.0;
							rk3CovdBvf_x = 0.0;
							rk3CovdBfd_xx = 0.0;
							rk3CovdBfd_xz = 0.0;
							rk3CovdBfd_xy = 0.0;
						}
						rk3optdepthxd = d_optdepthxd_o0_t0_m0_l0;
						rk3optdepthxh = d_optdepthxh_o0_t0_m0_l0;
						rk3optdepthxv = d_optdepthxv_o0_t0_m0_l0;
						rk3chiav = d_chiav_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk3CovdEfu_zz = d_CovdEfu_zz_o0_t0_m0_l0;
							rk3CovdEfu_zy = d_CovdEfu_zy_o0_t0_m0_l0;
							rk3CovdEfu_zx = d_CovdEfu_zx_o0_t0_m0_l0;
							rk3CovdEfd_zz = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * rk3CovdEfu_zx + vector(rk3gtd_yz, i, j, k) * rk3CovdEfu_zy + vector(rk3gtd_zz, i, j, k) * rk3CovdEfu_zz);
							rk3CovdEfd_zx = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * rk3CovdEfu_zx + vector(rk3gtd_xy, i, j, k) * rk3CovdEfu_zy + vector(rk3gtd_xz, i, j, k) * rk3CovdEfu_zz);
							rk3CovdEfd_zy = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * rk3CovdEfu_zx + vector(rk3gtd_yy, i, j, k) * rk3CovdEfu_zy + vector(rk3gtd_yz, i, j, k) * rk3CovdEfu_zz);
						} else {
							rk3CovdEfu_zz = 0.0;
							rk3CovdEfu_zy = 0.0;
							rk3CovdEfu_zx = 0.0;
							rk3CovdEfd_zz = 0.0;
							rk3CovdEfd_zx = 0.0;
							rk3CovdEfd_zy = 0.0;
						}
						rk3Bvf = (vector(rk3Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk3Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk3Bfu_z, i, j, k) * vector(vfd_z, i, j, k)) / rk3sdetg;
						rk3optdepthah = d_optdepthah_o0_t0_m0_l0;
						rk3optdepthad = d_optdepthad_o0_t0_m0_l0;
						rk3h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						rk3chiah = d_chiah_o0_t0_m0_l0;
						rk3chiad = d_chiad_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk3CovdYef_y = d_CovdYef_y_o0_t0_m0_l0;
						} else {
							rk3CovdYef_y = 0.0;
						}
						rk3optdepthav = d_optdepthav_o0_t0_m0_l0;
						if (calculate_les_terms) {
							rk3CovdYef_z = d_CovdYef_z_o0_t0_m0_l0;
							rk3CovdYef_x = d_CovdYef_x_o0_t0_m0_l0;
						} else {
							rk3CovdYef_z = 0.0;
							rk3CovdYef_x = 0.0;
						}
						rk3detgtd = vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) * vector(rk3gtd_xx, i, j, k) + (-vector(rk3gtd_yy, i, j, k) * (vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k))) + (-vector(rk3gtd_zz, i, j, k) * (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k))) + (-(vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k)) * vector(rk3gtd_xx, i, j, k)) + 2.0 * vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yy, i, j, k) - vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k));
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_zz, i, j, k) - vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yz, i, j, k)) + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) - vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_yz, i, j, k) - vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_xz, i, j, k));
						if (calculate_les_terms) {
							rk3CovuBfu_zx = rk3chi_max * (rk3gtu_xz * rk3CovdBfu_xx + rk3gtu_yz * rk3CovdBfu_yx + rk3gtu_zz * rk3CovdBfu_zx);
							rk3CovuBfu_zz = rk3chi_max * (rk3gtu_xz * rk3CovdBfu_xz + rk3gtu_yz * rk3CovdBfu_yz + rk3gtu_zz * rk3CovdBfu_zz);
							rk3CovuBfu_zy = rk3chi_max * (rk3gtu_xz * rk3CovdBfu_xy + rk3gtu_yz * rk3CovdBfu_yy + rk3gtu_zz * rk3CovdBfu_zy);
							rk3Covdvfu_xz = rk3chi_max * (rk3gtu_xz * rk3Covdvfd_xx + rk3gtu_yz * rk3Covdvfd_xy + rk3gtu_zz * rk3Covdvfd_xz);
							rk3CovuYef_z = rk3chi_max * (rk3gtu_xz * rk3CovdYef_x + rk3gtu_yz * rk3CovdYef_y + rk3gtu_zz * rk3CovdYef_z);
							rk3Covdvfu_zz = rk3chi_max * (rk3gtu_xz * rk3Covdvfd_zx + rk3gtu_yz * rk3Covdvfd_zy + rk3gtu_zz * rk3Covdvfd_zz);
							rk3Covdvfu_yz = rk3chi_max * (rk3gtu_xz * rk3Covdvfd_yx + rk3gtu_yz * rk3Covdvfd_yy + rk3gtu_zz * rk3Covdvfd_yz);
							rk3Covuvfu_zz = rk3chi_max * (rk3gtu_xz * rk3Covdvfu_xz + rk3gtu_yz * rk3Covdvfu_yz + rk3gtu_zz * rk3Covdvfu_zz);
						} else {
							rk3CovuBfu_zx = 0.0;
							rk3CovuBfu_zz = 0.0;
							rk3CovuBfu_zy = 0.0;
							rk3Covdvfu_xz = 0.0;
							rk3CovuYef_z = 0.0;
							rk3Covdvfu_zz = 0.0;
							rk3Covdvfu_yz = 0.0;
							rk3Covuvfu_zz = 0.0;
						}
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j, k) + rk3gtu_yz * vector(rk3Sfd_y, i, j, k) + rk3gtu_zz * vector(rk3Sfd_z, i, j, k));
						if (calculate_les_terms) {
							rk3Covurhof_z = rk3chi_max * (rk3gtu_xz * rk3Covdrhof_x + rk3gtu_yz * rk3Covdrhof_y + rk3gtu_zz * rk3Covdrhof_z);
							rk3CovuBvf_z = rk3chi_max * (rk3gtu_xz * rk3CovdBvf_x + rk3gtu_yz * rk3CovdBvf_y + rk3gtu_zz * rk3CovdBvf_z);
						} else {
							rk3Covurhof_z = 0.0;
							rk3CovuBvf_z = 0.0;
						}
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_zz, i, j, k)) + vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xz, i, j, k));
						if (calculate_les_terms) {
							rk3CovuBfu_yz = rk3chi_max * (rk3gtu_xy * rk3CovdBfu_xz + rk3gtu_yy * rk3CovdBfu_yz + rk3gtu_yz * rk3CovdBfu_zz);
							rk3CovuBfu_yy = rk3chi_max * (rk3gtu_xy * rk3CovdBfu_xy + rk3gtu_yy * rk3CovdBfu_yy + rk3gtu_yz * rk3CovdBfu_zy);
							rk3CovuBfu_yx = rk3chi_max * (rk3gtu_xy * rk3CovdBfu_xx + rk3gtu_yy * rk3CovdBfu_yx + rk3gtu_yz * rk3CovdBfu_zx);
							rk3Covdvfu_xy = rk3chi_max * (rk3gtu_xy * rk3Covdvfd_xx + rk3gtu_yy * rk3Covdvfd_xy + rk3gtu_yz * rk3Covdvfd_xz);
							rk3Covdvfu_xx = rk3chi_max * (rk3gtu_xx * rk3Covdvfd_xx + rk3gtu_xy * rk3Covdvfd_xy + rk3gtu_xz * rk3Covdvfd_xz);
							rk3CovdinvsqW_x = (-2.0 * vector(vfd_x, i, j, k) * rk3Covdvfu_xx) + (-2.0 * vector(vfd_y, i, j, k) * rk3Covdvfu_xy) + (-2.0 * vector(vfd_z, i, j, k) * rk3Covdvfu_xz);
							rk3CovuYef_x = rk3chi_max * (rk3gtu_xx * rk3CovdYef_x + rk3gtu_xy * rk3CovdYef_y + rk3gtu_xz * rk3CovdYef_z);
							rk3CovuYef_y = rk3chi_max * (rk3gtu_xy * rk3CovdYef_x + rk3gtu_yy * rk3CovdYef_y + rk3gtu_yz * rk3CovdYef_z);
							rk3Covdvfu_zy = rk3chi_max * (rk3gtu_xy * rk3Covdvfd_zx + rk3gtu_yy * rk3Covdvfd_zy + rk3gtu_yz * rk3Covdvfd_zz);
							rk3Covdvfu_zx = rk3chi_max * (rk3gtu_xx * rk3Covdvfd_zx + rk3gtu_xy * rk3Covdvfd_zy + rk3gtu_xz * rk3Covdvfd_zz);
							rk3CovdinvsqW_z = (-2.0 * vector(vfd_x, i, j, k) * rk3Covdvfu_zx) + (-2.0 * vector(vfd_y, i, j, k) * rk3Covdvfu_zy) + (-2.0 * vector(vfd_z, i, j, k) * rk3Covdvfu_zz);
							rk3Covdvfu_yx = rk3chi_max * (rk3gtu_xx * rk3Covdvfd_yx + rk3gtu_xy * rk3Covdvfd_yy + rk3gtu_xz * rk3Covdvfd_yz);
							rk3Covuvfu_zx = rk3chi_max * (rk3gtu_xz * rk3Covdvfu_xx + rk3gtu_yz * rk3Covdvfu_yx + rk3gtu_zz * rk3Covdvfu_zx);
							rk3Covdvfu_yy = rk3chi_max * (rk3gtu_xy * rk3Covdvfd_yx + rk3gtu_yy * rk3Covdvfd_yy + rk3gtu_yz * rk3Covdvfd_yz);
							rk3CovdinvsqW_y = (-2.0 * vector(vfd_x, i, j, k) * rk3Covdvfu_yx) + (-2.0 * vector(vfd_y, i, j, k) * rk3Covdvfu_yy) + (-2.0 * vector(vfd_z, i, j, k) * rk3Covdvfu_yz);
							rk3CovuinvsqW_z = rk3chi_max * (rk3gtu_xz * rk3CovdinvsqW_x + rk3gtu_yz * rk3CovdinvsqW_y + rk3gtu_zz * rk3CovdinvsqW_z);
							rk3Covuvfu_zy = rk3chi_max * (rk3gtu_xz * rk3Covdvfu_xy + rk3gtu_yz * rk3Covdvfu_yy + rk3gtu_zz * rk3Covdvfu_zy);
							rk3CovuBfu_xz = rk3chi_max * (rk3gtu_xx * rk3CovdBfu_xz + rk3gtu_xy * rk3CovdBfu_yz + rk3gtu_xz * rk3CovdBfu_zz);
							rk3CovuBfu_xy = rk3chi_max * (rk3gtu_xx * rk3CovdBfu_xy + rk3gtu_xy * rk3CovdBfu_yy + rk3gtu_xz * rk3CovdBfu_zy);
							rk3CovuBfu_xx = rk3chi_max * (rk3gtu_xx * rk3CovdBfu_xx + rk3gtu_xy * rk3CovdBfu_yx + rk3gtu_xz * rk3CovdBfu_zx);
						} else {
							rk3CovuBfu_yz = 0.0;
							rk3CovuBfu_yy = 0.0;
							rk3CovuBfu_yx = 0.0;
							rk3Covdvfu_xy = 0.0;
							rk3Covdvfu_xx = 0.0;
							rk3CovdinvsqW_x = 0.0;
							rk3CovuYef_x = 0.0;
							rk3CovuYef_y = 0.0;
							rk3Covdvfu_zy = 0.0;
							rk3Covdvfu_zx = 0.0;
							rk3CovdinvsqW_z = 0.0;
							rk3Covdvfu_yx = 0.0;
							rk3Covuvfu_zx = 0.0;
							rk3Covdvfu_yy = 0.0;
							rk3CovdinvsqW_y = 0.0;
							rk3CovuinvsqW_z = 0.0;
							rk3Covuvfu_zy = 0.0;
							rk3CovuBfu_xz = 0.0;
							rk3CovuBfu_xy = 0.0;
							rk3CovuBfu_xx = 0.0;
						}
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j, k) + rk3gtu_yy * vector(rk3Sfd_y, i, j, k) + rk3gtu_yz * vector(rk3Sfd_z, i, j, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j, k) + rk3gtu_xy * vector(rk3Sfd_y, i, j, k) + rk3gtu_xz * vector(rk3Sfd_z, i, j, k));
						if (calculate_les_terms) {
							rk3Covurhof_x = rk3chi_max * (rk3gtu_xx * rk3Covdrhof_x + rk3gtu_xy * rk3Covdrhof_y + rk3gtu_xz * rk3Covdrhof_z);
							rk3Covurhof_y = rk3chi_max * (rk3gtu_xy * rk3Covdrhof_x + rk3gtu_yy * rk3Covdrhof_y + rk3gtu_yz * rk3Covdrhof_z);
							rk3Covuvfu_yz = rk3chi_max * (rk3gtu_xy * rk3Covdvfu_xz + rk3gtu_yy * rk3Covdvfu_yz + rk3gtu_yz * rk3Covdvfu_zz);
							rk3Covuvfu_yy = rk3chi_max * (rk3gtu_xy * rk3Covdvfu_xy + rk3gtu_yy * rk3Covdvfu_yy + rk3gtu_yz * rk3Covdvfu_zy);
							rk3Covuvfu_yx = rk3chi_max * (rk3gtu_xy * rk3Covdvfu_xx + rk3gtu_yy * rk3Covdvfu_yx + rk3gtu_yz * rk3Covdvfu_zx);
							rk3Covuvfu_xz = rk3chi_max * (rk3gtu_xx * rk3Covdvfu_xz + rk3gtu_xy * rk3Covdvfu_yz + rk3gtu_xz * rk3Covdvfu_zz);
							rk3Covuvfu_xy = rk3chi_max * (rk3gtu_xx * rk3Covdvfu_xy + rk3gtu_xy * rk3Covdvfu_yy + rk3gtu_xz * rk3Covdvfu_zy);
							rk3Covuvfu_xx = rk3chi_max * (rk3gtu_xx * rk3Covdvfu_xx + rk3gtu_xy * rk3Covdvfu_yx + rk3gtu_xz * rk3Covdvfu_zx);
						} else {
							rk3Covurhof_x = 0.0;
							rk3Covurhof_y = 0.0;
							rk3Covuvfu_yz = 0.0;
							rk3Covuvfu_yy = 0.0;
							rk3Covuvfu_yx = 0.0;
							rk3Covuvfu_xz = 0.0;
							rk3Covuvfu_xy = 0.0;
							rk3Covuvfu_xx = 0.0;
						}
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j, k) + rk3gtu_yy * vector(vfd_y, i, j, k) + rk3gtu_yz * vector(vfd_z, i, j, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j, k) + rk3gtu_xy * vector(vfd_y, i, j, k) + rk3gtu_xz * vector(vfd_z, i, j, k));
						if (calculate_les_terms) {
							rk3CovuBvf_x = rk3chi_max * (rk3gtu_xx * rk3CovdBvf_x + rk3gtu_xy * rk3CovdBvf_y + rk3gtu_xz * rk3CovdBvf_z);
							rk3CovuBvf_y = rk3chi_max * (rk3gtu_xy * rk3CovdBvf_x + rk3gtu_yy * rk3CovdBvf_y + rk3gtu_yz * rk3CovdBvf_z);
							rk3CovuinvsqW_y = rk3chi_max * (rk3gtu_xy * rk3CovdinvsqW_x + rk3gtu_yy * rk3CovdinvsqW_y + rk3gtu_yz * rk3CovdinvsqW_z);
							rk3CovuinvsqW_x = rk3chi_max * (rk3gtu_xx * rk3CovdinvsqW_x + rk3gtu_xy * rk3CovdinvsqW_y + rk3gtu_xz * rk3CovdinvsqW_z);
							rk3Covddpfdye_z = d_Covddpfdye_z_o0_t0_m0_l0;
							rk3Covddpfdye_y = d_Covddpfdye_y_o0_t0_m0_l0;
							rk3Covddpfdye_x = d_Covddpfdye_x_o0_t0_m0_l0;
							rk3CovdEfu_xz = d_CovdEfu_xz_o0_t0_m0_l0;
							rk3CovdEfu_xy = d_CovdEfu_xy_o0_t0_m0_l0;
							rk3CovdEfu_xx = d_CovdEfu_xx_o0_t0_m0_l0;
							rk3CovdEfd_xx = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * rk3CovdEfu_xx + vector(rk3gtd_xy, i, j, k) * rk3CovdEfu_xy + vector(rk3gtd_xz, i, j, k) * rk3CovdEfu_xz);
							rk3CovdEfd_xy = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * rk3CovdEfu_xx + vector(rk3gtd_yy, i, j, k) * rk3CovdEfu_xy + vector(rk3gtd_yz, i, j, k) * rk3CovdEfu_xz);
							rk3CovdEfd_xz = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * rk3CovdEfu_xx + vector(rk3gtd_yz, i, j, k) * rk3CovdEfu_xy + vector(rk3gtd_zz, i, j, k) * rk3CovdEfu_xz);
						} else {
							rk3CovuBvf_x = 0.0;
							rk3CovuBvf_y = 0.0;
							rk3CovuinvsqW_y = 0.0;
							rk3CovuinvsqW_x = 0.0;
							rk3Covddpfdye_z = 0.0;
							rk3Covddpfdye_y = 0.0;
							rk3Covddpfdye_x = 0.0;
							rk3CovdEfu_xz = 0.0;
							rk3CovdEfu_xy = 0.0;
							rk3CovdEfu_xx = 0.0;
							rk3CovdEfd_xx = 0.0;
							rk3CovdEfd_xy = 0.0;
							rk3CovdEfd_xz = 0.0;
						}
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j, k) + rk3gtu_yz * vector(vfd_y, i, j, k) + rk3gtu_zz * vector(vfd_z, i, j, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j, k) + rk3vfu_y * vector(vfd_y, i, j, k) + rk3vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						if (calculate_les_terms) {
							rk3PhiAh = rk3sqW * (vector(pf, i, j, k) * vector(dpfdeps, i, j, k) + (vector(rhof, i, j, k) * vector(rhof, i, j, k)) * vector(dpfdrho, i, j, k));
						} else {
							rk3PhiAh = 0.0;
						}
						rk3Epf = rk3h * rk3sqW;
						rk3Thetaf = rk3Epf + rk3sqBf / (rk3sdetg * rk3sdetg);
						if (calculate_les_terms) {
							rk3Covdepsf_z = d_Covdepsf_z_o0_t0_m0_l0;
							rk3Covdhf_z = (1.0 + vector(epsf, i, j, k)) * rk3Covdrhof_z + vector(rhof, i, j, k) * rk3Covdepsf_z + rk3Covdpf_z;
							rk3CovdEpf_z = rk3sqW * rk3Covdhf_z + (-rk3h * (rk3sqW * rk3sqW) * rk3CovdinvsqW_z);
							rk3CovdThetaf_z = rk3CovdEpf_z + 2.0 * (rk3Bfd_x * rk3CovdBfu_zx) / rk3sdetg + 2.0 * (rk3Bfd_y * rk3CovdBfu_zy) / rk3sdetg + 2.0 * (rk3Bfd_z * rk3CovdBfu_zz) / rk3sdetg;
							rk3Covdepsf_y = d_Covdepsf_y_o0_t0_m0_l0;
							rk3Covdhf_y = (1.0 + vector(epsf, i, j, k)) * rk3Covdrhof_y + vector(rhof, i, j, k) * rk3Covdepsf_y + rk3Covdpf_y;
							rk3CovdEpf_y = rk3sqW * rk3Covdhf_y + (-rk3h * (rk3sqW * rk3sqW) * rk3CovdinvsqW_y);
							rk3CovdThetaf_y = rk3CovdEpf_y + 2.0 * (rk3Bfd_x * rk3CovdBfu_yx) / rk3sdetg + 2.0 * (rk3Bfd_y * rk3CovdBfu_yy) / rk3sdetg + 2.0 * (rk3Bfd_z * rk3CovdBfu_yz) / rk3sdetg;
							rk3Covdepsf_x = d_Covdepsf_x_o0_t0_m0_l0;
							rk3Covuepsf_z = rk3chi_max * (rk3gtu_xz * rk3Covdepsf_x + rk3gtu_yz * rk3Covdepsf_y + rk3gtu_zz * rk3Covdepsf_z);
							rk3Covuepsf_y = rk3chi_max * (rk3gtu_xy * rk3Covdepsf_x + rk3gtu_yy * rk3Covdepsf_y + rk3gtu_yz * rk3Covdepsf_z);
							rk3Covdhf_x = (1.0 + vector(epsf, i, j, k)) * rk3Covdrhof_x + vector(rhof, i, j, k) * rk3Covdepsf_x + rk3Covdpf_x;
							rk3CovdEpf_x = rk3sqW * rk3Covdhf_x + (-rk3h * (rk3sqW * rk3sqW) * rk3CovdinvsqW_x);
							rk3CovdThetaf_x = rk3CovdEpf_x + 2.0 * (rk3Bfd_x * rk3CovdBfu_xx) / rk3sdetg + 2.0 * (rk3Bfd_y * rk3CovdBfu_xy) / rk3sdetg + 2.0 * (rk3Bfd_z * rk3CovdBfu_xz) / rk3sdetg;
							rk3CovuThetaf_z = rk3chi_max * (rk3gtu_xz * rk3CovdThetaf_x + rk3gtu_yz * rk3CovdThetaf_y + rk3gtu_zz * rk3CovdThetaf_z);
							rk3CovuThetaf_x = rk3chi_max * (rk3gtu_xx * rk3CovdThetaf_x + rk3gtu_xy * rk3CovdThetaf_y + rk3gtu_xz * rk3CovdThetaf_z);
							rk3CovuThetaf_y = rk3chi_max * (rk3gtu_xy * rk3CovdThetaf_x + rk3gtu_yy * rk3CovdThetaf_y + rk3gtu_yz * rk3CovdThetaf_z);
							rk3Phivh_z = 2.0 * (rk3CovuBvf_x * rk3CovdBfu_xz + rk3CovuBvf_y * rk3CovdBfu_yz + rk3CovuBvf_z * rk3CovdBfu_zz + (-rk3CovuThetaf_x * rk3Covdvfu_xz) + (-rk3CovuThetaf_y * rk3Covdvfu_yz) + (-rk3CovuThetaf_z * rk3Covdvfu_zz) + (vector(rk3Bfu_z, i, j, k) * (rk3Thetaf * (rk3CovuBfu_xx * rk3Covdvfd_xx + rk3CovuBfu_yx * rk3Covdvfd_yx + rk3CovuBfu_zx * rk3Covdvfd_zx + rk3CovuBfu_xy * rk3Covdvfd_xy + rk3CovuBfu_yy * rk3Covdvfd_yy + rk3CovuBfu_zy * rk3Covdvfd_zy + rk3CovuBfu_xz * rk3Covdvfd_xz + rk3CovuBfu_yz * rk3Covdvfd_yz + rk3CovuBfu_zz * rk3Covdvfd_zz) + (rk3Bfd_x * rk3CovdBfu_xx * rk3CovuBvf_x) / rk3sdetg + (rk3Bfd_y * rk3CovdBfu_xy * rk3CovuBvf_x) / rk3sdetg + (rk3Bfd_z * rk3CovdBfu_xz * rk3CovuBvf_x) / rk3sdetg + (rk3Bfd_x * rk3CovdBfu_yx * rk3CovuBvf_y) / rk3sdetg + (rk3Bfd_y * rk3CovdBfu_yy * rk3CovuBvf_y) / rk3sdetg + (rk3Bfd_z * rk3CovdBfu_yz * rk3CovuBvf_y) / rk3sdetg + (rk3Bfd_x * rk3CovdBfu_zx * rk3CovuBvf_z) / rk3sdetg + (rk3Bfd_y * rk3CovdBfu_zy * rk3CovuBvf_z) / rk3sdetg + (rk3Bfd_z * rk3CovdBfu_zz * rk3CovuBvf_z) / rk3sdetg + (-(vector(rk3Bfu_x, i, j, k) * rk3Covdvfd_xx * rk3CovuThetaf_x) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3Covdvfd_xy * rk3CovuThetaf_x) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfd_xz * rk3CovuThetaf_x) / rk3sdetg) + (-(vector(rk3Bfu_x, i, j, k) * rk3Covdvfd_yx * rk3CovuThetaf_y) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3Covdvfd_yy * rk3CovuThetaf_y) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfd_yz * rk3CovuThetaf_y) / rk3sdetg) + (-(vector(rk3Bfu_x, i, j, k) * rk3Covdvfd_zx * rk3CovuThetaf_z) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3Covdvfd_zy * rk3CovuThetaf_z) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfd_zz * rk3CovuThetaf_z) / rk3sdetg))) / (rk3sdetg * rk3Epf)) / rk3Thetaf;
							rk3Phivh_x = 2.0 * (rk3CovuBvf_x * rk3CovdBfu_xx + rk3CovuBvf_y * rk3CovdBfu_yx + rk3CovuBvf_z * rk3CovdBfu_zx + (-rk3CovuThetaf_x * rk3Covdvfu_xx) + (-rk3CovuThetaf_y * rk3Covdvfu_yx) + (-rk3CovuThetaf_z * rk3Covdvfu_zx) + (vector(rk3Bfu_x, i, j, k) * (rk3Thetaf * (rk3CovuBfu_xx * rk3Covdvfd_xx + rk3CovuBfu_yx * rk3Covdvfd_yx + rk3CovuBfu_zx * rk3Covdvfd_zx + rk3CovuBfu_xy * rk3Covdvfd_xy + rk3CovuBfu_yy * rk3Covdvfd_yy + rk3CovuBfu_zy * rk3Covdvfd_zy + rk3CovuBfu_xz * rk3Covdvfd_xz + rk3CovuBfu_yz * rk3Covdvfd_yz + rk3CovuBfu_zz * rk3Covdvfd_zz) + (rk3Bfd_x * rk3CovdBfu_xx * rk3CovuBvf_x) / rk3sdetg + (rk3Bfd_y * rk3CovdBfu_xy * rk3CovuBvf_x) / rk3sdetg + (rk3Bfd_z * rk3CovdBfu_xz * rk3CovuBvf_x) / rk3sdetg + (rk3Bfd_x * rk3CovdBfu_yx * rk3CovuBvf_y) / rk3sdetg + (rk3Bfd_y * rk3CovdBfu_yy * rk3CovuBvf_y) / rk3sdetg + (rk3Bfd_z * rk3CovdBfu_yz * rk3CovuBvf_y) / rk3sdetg + (rk3Bfd_x * rk3CovdBfu_zx * rk3CovuBvf_z) / rk3sdetg + (rk3Bfd_y * rk3CovdBfu_zy * rk3CovuBvf_z) / rk3sdetg + (rk3Bfd_z * rk3CovdBfu_zz * rk3CovuBvf_z) / rk3sdetg + (-(vector(rk3Bfu_x, i, j, k) * rk3Covdvfd_xx * rk3CovuThetaf_x) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3Covdvfd_xy * rk3CovuThetaf_x) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfd_xz * rk3CovuThetaf_x) / rk3sdetg) + (-(vector(rk3Bfu_x, i, j, k) * rk3Covdvfd_yx * rk3CovuThetaf_y) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3Covdvfd_yy * rk3CovuThetaf_y) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfd_yz * rk3CovuThetaf_y) / rk3sdetg) + (-(vector(rk3Bfu_x, i, j, k) * rk3Covdvfd_zx * rk3CovuThetaf_z) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3Covdvfd_zy * rk3CovuThetaf_z) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfd_zz * rk3CovuThetaf_z) / rk3sdetg))) / (rk3sdetg * rk3Epf)) / rk3Thetaf;
							rk3Phivh_y = 2.0 * (rk3CovuBvf_x * rk3CovdBfu_xy + rk3CovuBvf_y * rk3CovdBfu_yy + rk3CovuBvf_z * rk3CovdBfu_zy + (-rk3CovuThetaf_x * rk3Covdvfu_xy) + (-rk3CovuThetaf_y * rk3Covdvfu_yy) + (-rk3CovuThetaf_z * rk3Covdvfu_zy) + (vector(rk3Bfu_y, i, j, k) * (rk3Thetaf * (rk3CovuBfu_xx * rk3Covdvfd_xx + rk3CovuBfu_yx * rk3Covdvfd_yx + rk3CovuBfu_zx * rk3Covdvfd_zx + rk3CovuBfu_xy * rk3Covdvfd_xy + rk3CovuBfu_yy * rk3Covdvfd_yy + rk3CovuBfu_zy * rk3Covdvfd_zy + rk3CovuBfu_xz * rk3Covdvfd_xz + rk3CovuBfu_yz * rk3Covdvfd_yz + rk3CovuBfu_zz * rk3Covdvfd_zz) + (rk3Bfd_x * rk3CovdBfu_xx * rk3CovuBvf_x) / rk3sdetg + (rk3Bfd_y * rk3CovdBfu_xy * rk3CovuBvf_x) / rk3sdetg + (rk3Bfd_z * rk3CovdBfu_xz * rk3CovuBvf_x) / rk3sdetg + (rk3Bfd_x * rk3CovdBfu_yx * rk3CovuBvf_y) / rk3sdetg + (rk3Bfd_y * rk3CovdBfu_yy * rk3CovuBvf_y) / rk3sdetg + (rk3Bfd_z * rk3CovdBfu_yz * rk3CovuBvf_y) / rk3sdetg + (rk3Bfd_x * rk3CovdBfu_zx * rk3CovuBvf_z) / rk3sdetg + (rk3Bfd_y * rk3CovdBfu_zy * rk3CovuBvf_z) / rk3sdetg + (rk3Bfd_z * rk3CovdBfu_zz * rk3CovuBvf_z) / rk3sdetg + (-(vector(rk3Bfu_x, i, j, k) * rk3Covdvfd_xx * rk3CovuThetaf_x) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3Covdvfd_xy * rk3CovuThetaf_x) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfd_xz * rk3CovuThetaf_x) / rk3sdetg) + (-(vector(rk3Bfu_x, i, j, k) * rk3Covdvfd_yx * rk3CovuThetaf_y) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3Covdvfd_yy * rk3CovuThetaf_y) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfd_yz * rk3CovuThetaf_y) / rk3sdetg) + (-(vector(rk3Bfu_x, i, j, k) * rk3Covdvfd_zx * rk3CovuThetaf_z) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3Covdvfd_zy * rk3CovuThetaf_z) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfd_zz * rk3CovuThetaf_z) / rk3sdetg))) / (rk3sdetg * rk3Epf)) / rk3Thetaf;
							rk3PhiMh_yz = 2.0 * (rk3Thetaf * ((-rk3Covdvfu_xz * rk3CovuBfu_xy) + rk3Covdvfu_xy * rk3CovuBfu_xz + (-rk3Covdvfu_yz * rk3CovuBfu_yy) + rk3Covdvfu_yy * rk3CovuBfu_yz + (-rk3Covdvfu_zz * rk3CovuBfu_zy) + rk3Covdvfu_zy * rk3CovuBfu_zz) + (vector(rk3Bfu_z, i, j, k) * rk3CovuBfu_xy * rk3CovdBvf_x) / rk3sdetg + (vector(rk3Bfu_z, i, j, k) * rk3CovuBfu_yy * rk3CovdBvf_y) / rk3sdetg + (vector(rk3Bfu_z, i, j, k) * rk3CovuBfu_zy * rk3CovdBvf_z) / rk3sdetg + (-(vector(rk3Bfu_y, i, j, k) * rk3CovuBfu_xz * rk3CovdBvf_x) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3CovuBfu_yz * rk3CovdBvf_y) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3CovuBfu_zz * rk3CovdBvf_z) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfu_xy * rk3CovuThetaf_x) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfu_yy * rk3CovuThetaf_y) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfu_zy * rk3CovuThetaf_z) / rk3sdetg) + (vector(rk3Bfu_y, i, j, k) * rk3Covdvfu_xz * rk3CovuThetaf_x) / rk3sdetg + (vector(rk3Bfu_y, i, j, k) * rk3Covdvfu_yz * rk3CovuThetaf_y) / rk3sdetg + (vector(rk3Bfu_y, i, j, k) * rk3Covdvfu_zz * rk3CovuThetaf_z) / rk3sdetg) / rk3Thetaf;
							rk3PhiMh_xz = 2.0 * (rk3Thetaf * ((-rk3Covdvfu_xz * rk3CovuBfu_xx) + rk3Covdvfu_xx * rk3CovuBfu_xz + (-rk3Covdvfu_yz * rk3CovuBfu_yx) + rk3Covdvfu_yx * rk3CovuBfu_yz + (-rk3Covdvfu_zz * rk3CovuBfu_zx) + rk3Covdvfu_zx * rk3CovuBfu_zz) + (vector(rk3Bfu_z, i, j, k) * rk3CovuBfu_xx * rk3CovdBvf_x) / rk3sdetg + (vector(rk3Bfu_z, i, j, k) * rk3CovuBfu_yx * rk3CovdBvf_y) / rk3sdetg + (vector(rk3Bfu_z, i, j, k) * rk3CovuBfu_zx * rk3CovdBvf_z) / rk3sdetg + (-(vector(rk3Bfu_x, i, j, k) * rk3CovuBfu_xz * rk3CovdBvf_x) / rk3sdetg) + (-(vector(rk3Bfu_x, i, j, k) * rk3CovuBfu_yz * rk3CovdBvf_y) / rk3sdetg) + (-(vector(rk3Bfu_x, i, j, k) * rk3CovuBfu_zz * rk3CovdBvf_z) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfu_xx * rk3CovuThetaf_x) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfu_yx * rk3CovuThetaf_y) / rk3sdetg) + (-(vector(rk3Bfu_z, i, j, k) * rk3Covdvfu_zx * rk3CovuThetaf_z) / rk3sdetg) + (vector(rk3Bfu_x, i, j, k) * rk3Covdvfu_xz * rk3CovuThetaf_x) / rk3sdetg + (vector(rk3Bfu_x, i, j, k) * rk3Covdvfu_yz * rk3CovuThetaf_y) / rk3sdetg + (vector(rk3Bfu_x, i, j, k) * rk3Covdvfu_zz * rk3CovuThetaf_z) / rk3sdetg) / rk3Thetaf;
							rk3PhiMh_xy = 2.0 * (rk3Thetaf * ((-rk3Covdvfu_xy * rk3CovuBfu_xx) + rk3Covdvfu_xx * rk3CovuBfu_xy + (-rk3Covdvfu_yy * rk3CovuBfu_yx) + rk3Covdvfu_yx * rk3CovuBfu_yy + (-rk3Covdvfu_zy * rk3CovuBfu_zx) + rk3Covdvfu_zx * rk3CovuBfu_zy) + (vector(rk3Bfu_y, i, j, k) * rk3CovuBfu_xx * rk3CovdBvf_x) / rk3sdetg + (vector(rk3Bfu_y, i, j, k) * rk3CovuBfu_yx * rk3CovdBvf_y) / rk3sdetg + (vector(rk3Bfu_y, i, j, k) * rk3CovuBfu_zx * rk3CovdBvf_z) / rk3sdetg + (-(vector(rk3Bfu_x, i, j, k) * rk3CovuBfu_xy * rk3CovdBvf_x) / rk3sdetg) + (-(vector(rk3Bfu_x, i, j, k) * rk3CovuBfu_yy * rk3CovdBvf_y) / rk3sdetg) + (-(vector(rk3Bfu_x, i, j, k) * rk3CovuBfu_zy * rk3CovdBvf_z) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3Covdvfu_xx * rk3CovuThetaf_x) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3Covdvfu_yx * rk3CovuThetaf_y) / rk3sdetg) + (-(vector(rk3Bfu_y, i, j, k) * rk3Covdvfu_zx * rk3CovuThetaf_z) / rk3sdetg) + (vector(rk3Bfu_x, i, j, k) * rk3Covdvfu_xy * rk3CovuThetaf_x) / rk3sdetg + (vector(rk3Bfu_x, i, j, k) * rk3Covdvfu_yy * rk3CovuThetaf_y) / rk3sdetg + (vector(rk3Bfu_x, i, j, k) * rk3Covdvfu_zy * rk3CovuThetaf_z) / rk3sdetg) / rk3Thetaf;
							rk3Covuepsf_x = rk3chi_max * (rk3gtu_xx * rk3Covdepsf_x + rk3gtu_xy * rk3Covdepsf_y + rk3gtu_xz * rk3Covdepsf_z);
						} else {
							rk3Covdepsf_z = 0.0;
							rk3Covdhf_z = 0.0;
							rk3CovdEpf_z = 0.0;
							rk3CovdThetaf_z = 0.0;
							rk3Covdepsf_y = 0.0;
							rk3Covdhf_y = 0.0;
							rk3CovdEpf_y = 0.0;
							rk3CovdThetaf_y = 0.0;
							rk3Covdepsf_x = 0.0;
							rk3Covuepsf_z = 0.0;
							rk3Covuepsf_y = 0.0;
							rk3Covdhf_x = 0.0;
							rk3CovdEpf_x = 0.0;
							rk3CovdThetaf_x = 0.0;
							rk3CovuThetaf_z = 0.0;
							rk3CovuThetaf_x = 0.0;
							rk3CovuThetaf_y = 0.0;
							rk3Phivh_z = 0.0;
							rk3Phivh_x = 0.0;
							rk3Phivh_y = 0.0;
							rk3PhiMh_yz = 0.0;
							rk3PhiMh_xz = 0.0;
							rk3PhiMh_xy = 0.0;
							rk3Covuepsf_x = 0.0;
						}
						rk3faceta = 0.04166666666666667 * rk3inv_chi * (dx[0] * dx[0]) * Smooth(threshold_Dles_min, threshold_Dles_max, vector(rk3Df, i, j, k), dx, simPlat_dt, ilast, jlast);
						if (calculate_les_terms) {
							rk3CovdEfu_yz = d_CovdEfu_yz_o0_t0_m0_l0;
							rk3CovuEfu_zz = rk3chi_max * (rk3gtu_xz * rk3CovdEfu_xz + rk3gtu_yz * rk3CovdEfu_yz + rk3gtu_zz * rk3CovdEfu_zz);
							rk3CovuEfu_xz = rk3chi_max * (rk3gtu_xx * rk3CovdEfu_xz + rk3gtu_xy * rk3CovdEfu_yz + rk3gtu_xz * rk3CovdEfu_zz);
							rk3CovdEfu_yy = d_CovdEfu_yy_o0_t0_m0_l0;
							rk3CovuEfu_zy = rk3chi_max * (rk3gtu_xz * rk3CovdEfu_xy + rk3gtu_yz * rk3CovdEfu_yy + rk3gtu_zz * rk3CovdEfu_zy);
							rk3CovuEfu_xy = rk3chi_max * (rk3gtu_xx * rk3CovdEfu_xy + rk3gtu_xy * rk3CovdEfu_yy + rk3gtu_xz * rk3CovdEfu_zy);
							rk3CovdEfu_yx = d_CovdEfu_yx_o0_t0_m0_l0;
							rk3CovdEfd_yy = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * rk3CovdEfu_yx + vector(rk3gtd_yy, i, j, k) * rk3CovdEfu_yy + vector(rk3gtd_yz, i, j, k) * rk3CovdEfu_yz);
							rk3CovdEfd_yz = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * rk3CovdEfu_yx + vector(rk3gtd_yz, i, j, k) * rk3CovdEfu_yy + vector(rk3gtd_zz, i, j, k) * rk3CovdEfu_yz);
							rk3CovdEfd_yx = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * rk3CovdEfu_yx + vector(rk3gtd_xy, i, j, k) * rk3CovdEfu_yy + vector(rk3gtd_xz, i, j, k) * rk3CovdEfu_yz);
							rk3CovuEfu_zx = rk3chi_max * (rk3gtu_xz * rk3CovdEfu_xx + rk3gtu_yz * rk3CovdEfu_yx + rk3gtu_zz * rk3CovdEfu_zx);
							rk3CovuEfu_xx = rk3chi_max * (rk3gtu_xx * rk3CovdEfu_xx + rk3gtu_xy * rk3CovdEfu_yx + rk3gtu_xz * rk3CovdEfu_zx);
							rk3CovuEfu_yx = rk3chi_max * (rk3gtu_xy * rk3CovdEfu_xx + rk3gtu_yy * rk3CovdEfu_yx + rk3gtu_yz * rk3CovdEfu_zx);
							rk3CovuEfu_yz = rk3chi_max * (rk3gtu_xy * rk3CovdEfu_xz + rk3gtu_yy * rk3CovdEfu_yz + rk3gtu_yz * rk3CovdEfu_zz);
							rk3CovuEfu_yy = rk3chi_max * (rk3gtu_xy * rk3CovdEfu_xy + rk3gtu_yy * rk3CovdEfu_yy + rk3gtu_yz * rk3CovdEfu_zy);
							rk3PhiThetah = (rk3Thetaf * (rk3CovdBfd_xx * rk3CovuBfu_xx + rk3CovdBfd_yx * rk3CovuBfu_yx + rk3CovdBfd_zx * rk3CovuBfu_zx + rk3CovdBfd_xy * rk3CovuBfu_xy + rk3CovdBfd_yy * rk3CovuBfu_yy + rk3CovdBfd_zy * rk3CovuBfu_zy + rk3CovdBfd_xz * rk3CovuBfu_xz + rk3CovdBfd_yz * rk3CovuBfu_yz + rk3CovdBfd_zz * rk3CovuBfu_zz + (-rk3CovdEfd_xx * rk3CovuEfu_xx) + (-rk3CovdEfd_yx * rk3CovuEfu_yx) + (-rk3CovdEfd_zx * rk3CovuEfu_zx) + (-rk3CovdEfd_xy * rk3CovuEfu_xy) + (-rk3CovdEfd_yy * rk3CovuEfu_yy) + (-rk3CovdEfd_zy * rk3CovuEfu_zy) + (-rk3CovdEfd_xz * rk3CovuEfu_xz) + (-rk3CovdEfd_yz * rk3CovuEfu_yz) + (-rk3CovdEfd_zz * rk3CovuEfu_zz) + (-0.5 * (((-vector(vfd_y, i, j, k) * rk3Bfd_x) + vector(vfd_x, i, j, k) * rk3Bfd_y) * rk3PhiMh_xy + (vector(vfd_x, i, j, k) * rk3Bfd_z - vector(vfd_z, i, j, k) * rk3Bfd_x) * rk3PhiMh_xz + (-(vector(vfd_y, i, j, k) * rk3Bfd_x - vector(vfd_x, i, j, k) * rk3Bfd_y) * rk3PhiMh_xy) + ((-vector(vfd_z, i, j, k) * rk3Bfd_y) + vector(vfd_y, i, j, k) * rk3Bfd_z) * rk3PhiMh_yz + (-(vector(vfd_z, i, j, k) * rk3Bfd_x - vector(vfd_x, i, j, k) * rk3Bfd_z) * rk3PhiMh_xz) + (-(vector(vfd_z, i, j, k) * rk3Bfd_y - vector(vfd_y, i, j, k) * rk3Bfd_z) * rk3PhiMh_yz)) / rk3sdetg))) / (rk3Thetaf - rk3sqEf);
							rk3HPres = (rk3Epf * rk3sqW * (rk3Thetaf - rk3sqEf) * (vector(rhof, i, j, k) * (rk3Covddpfdrho_x * rk3Covurhof_x + rk3Covddpfdrho_y * rk3Covurhof_y + rk3Covddpfdrho_z * rk3Covurhof_z + rk3Covddpfdeps_x * rk3Covuepsf_x + rk3Covddpfdeps_y * rk3Covuepsf_y + rk3Covddpfdeps_z * rk3Covuepsf_z) + (-2.0 * vector(dpfdeps, i, j, k) * (rk3Covurhof_x * rk3Covdepsf_x + rk3Covurhof_y * rk3Covdepsf_y + rk3Covurhof_z * rk3Covdepsf_z)) + (-(rk3Epf * vector(dpfdeps, i, j, k) - rk3PhiAh) * (0.25 * rk3sqW * (rk3CovdinvsqW_x * rk3CovuinvsqW_x + rk3CovdinvsqW_y * rk3CovuinvsqW_y + rk3CovdinvsqW_z * rk3CovuinvsqW_z) + (rk3CovdinvsqW_x * rk3Covurhof_x + rk3CovdinvsqW_y * rk3Covurhof_y + rk3CovdinvsqW_z * rk3Covurhof_z) / vector(rhof, i, j, k))) + (-2.0 * (vector(dpfdeps, i, j, k) * (rk3CovdBfd_xx * rk3CovuBfu_xx + rk3CovdBfd_yx * rk3CovuBfu_yx + rk3CovdBfd_zx * rk3CovuBfu_zx + rk3CovdBfd_xy * rk3CovuBfu_xy + rk3CovdBfd_yy * rk3CovuBfu_yy + rk3CovdBfd_zy * rk3CovuBfu_zy + rk3CovdBfd_xz * rk3CovuBfu_xz + rk3CovdBfd_yz * rk3CovuBfu_yz + rk3CovdBfd_zz * rk3CovuBfu_zz + (-(rk3sqW * rk3sqW) * (rk3CovuinvsqW_x * rk3Covdhf_x + rk3CovuinvsqW_y * rk3Covdhf_y + rk3CovuinvsqW_z * rk3Covdhf_z)))) / rk3sqW) + (-(rk3Epf * vector(dpfdeps, i, j, k) + rk3PhiAh) * (vector(vfd_x, i, j, k) * rk3Phivh_x + vector(vfd_y, i, j, k) * rk3Phivh_y + vector(vfd_z, i, j, k) * rk3Phivh_z + rk3Covdvfd_xx * rk3Covuvfu_xx + rk3Covdvfd_xy * rk3Covuvfu_xy + rk3Covdvfd_xz * rk3Covuvfu_xz + rk3Covdvfd_yx * rk3Covuvfu_yx + rk3Covdvfd_yy * rk3Covuvfu_yy + rk3Covdvfd_yz * rk3Covuvfu_yz + rk3Covdvfd_zx * rk3Covuvfu_zx + rk3Covdvfd_zy * rk3Covuvfu_zy + rk3Covdvfd_zz * rk3Covuvfu_zz + rk3sqW * (rk3CovdinvsqW_x * rk3CovuinvsqW_x + rk3CovdinvsqW_y * rk3CovuinvsqW_y + rk3CovdinvsqW_z * rk3CovuinvsqW_z))) + (rk3PhiThetah * ((rk3Epf * vector(dpfdeps, i, j, k) + rk3PhiAh) * (rk3Thetaf - rk3sqEf) - (rk3PhiAh * rk3Thetaf) / rk3sqW)) / (rk3Epf * rk3Thetaf))) / ((rk3Epf * vector(rhof, i, j, k) - rk3PhiAh) * (rk3Thetaf - rk3sqEf) * rk3sqW + rk3PhiAh * rk3Thetaf) + rk3Covddpfdye_x * rk3CovuYef_x + rk3Covddpfdye_y * rk3CovuYef_y + rk3Covddpfdye_z * rk3CovuYef_z - 2.0 * vector(dpfdye, i, j, k) / vector(rk3Df, i, j, k) * (rk3CovdDf_x * rk3CovuYef_x + rk3CovdDf_y * rk3CovuYef_y + rk3CovdDf_z * rk3CovuYef_z);
							rk3HTheta = rk3PhiThetah + (rk3Thetaf * rk3HPres) / (rk3Thetaf - rk3sqEf);
							rk3Hv_z = rk3Phivh_z + (-((rk3vfu_z + (rk3Bvf * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3Epf)) * rk3HTheta) / rk3Thetaf);
							rk3HN_z = 2.0 * rk3CovdDf_x * rk3Covuvfu_xz + 2.0 * rk3CovdDf_y * rk3Covuvfu_yz + 2.0 * rk3CovdDf_z * rk3Covuvfu_zz + (vector(rk3Df, i, j, k) * rk3Hv_z) / rk3sdetg;
							rk3HNe_z = 2.0 * rk3CovdDYf_x * rk3Covuvfu_xz + 2.0 * rk3CovdDYf_y * rk3Covuvfu_yz + 2.0 * rk3CovdDYf_z * rk3Covuvfu_zz + (vector(rk3DYf, i, j, k) * rk3Hv_z) / rk3sdetg;
							rk3Hv_y = rk3Phivh_y + (-((rk3vfu_y + (rk3Bvf * vector(rk3Bfu_y, i, j, k)) / (rk3sdetg * rk3Epf)) * rk3HTheta) / rk3Thetaf);
							rk3HN_y = 2.0 * rk3CovdDf_x * rk3Covuvfu_xy + 2.0 * rk3CovdDf_y * rk3Covuvfu_yy + 2.0 * rk3CovdDf_z * rk3Covuvfu_zy + (vector(rk3Df, i, j, k) * rk3Hv_y) / rk3sdetg;
							rk3HM_yz = (-2.0 * rk3Covdvfu_xz * rk3CovuBfu_xy) + 2.0 * rk3Covdvfu_xy * rk3CovuBfu_xz + (-2.0 * rk3Covdvfu_yz * rk3CovuBfu_yy) + 2.0 * rk3Covdvfu_yy * rk3CovuBfu_yz + (-2.0 * rk3Covdvfu_zz * rk3CovuBfu_zy) + 2.0 * rk3Covdvfu_zy * rk3CovuBfu_zz + (vector(rk3Bfu_z, i, j, k) * rk3Hv_y - vector(rk3Bfu_y, i, j, k) * rk3Hv_z) / rk3sdetg;
							rk3HNe_y = 2.0 * rk3CovdDYf_x * rk3Covuvfu_xy + 2.0 * rk3CovdDYf_y * rk3Covuvfu_yy + 2.0 * rk3CovdDYf_z * rk3Covuvfu_zy + (vector(rk3DYf, i, j, k) * rk3Hv_y) / rk3sdetg;
							rk3Hv_x = rk3Phivh_x + (-((rk3vfu_x + (rk3Bvf * vector(rk3Bfu_x, i, j, k)) / (rk3sdetg * rk3Epf)) * rk3HTheta) / rk3Thetaf);
							rk3HN_x = 2.0 * rk3CovdDf_x * rk3Covuvfu_xx + 2.0 * rk3CovdDf_y * rk3Covuvfu_yx + 2.0 * rk3CovdDf_z * rk3Covuvfu_zx + (vector(rk3Df, i, j, k) * rk3Hv_x) / rk3sdetg;
							rk3HM_xz = (-2.0 * rk3Covdvfu_xz * rk3CovuBfu_xx) + 2.0 * rk3Covdvfu_xx * rk3CovuBfu_xz + (-2.0 * rk3Covdvfu_yz * rk3CovuBfu_yx) + 2.0 * rk3Covdvfu_yx * rk3CovuBfu_yz + (-2.0 * rk3Covdvfu_zz * rk3CovuBfu_zx) + 2.0 * rk3Covdvfu_zx * rk3CovuBfu_zz + (vector(rk3Bfu_z, i, j, k) * rk3Hv_x - vector(rk3Bfu_x, i, j, k) * rk3Hv_z) / rk3sdetg;
							rk3HM_xy = (-2.0 * rk3Covdvfu_xy * rk3CovuBfu_xx) + 2.0 * rk3Covdvfu_xx * rk3CovuBfu_xy + (-2.0 * rk3Covdvfu_yy * rk3CovuBfu_yx) + 2.0 * rk3Covdvfu_yx * rk3CovuBfu_yy + (-2.0 * rk3Covdvfu_zy * rk3CovuBfu_zx) + 2.0 * rk3Covdvfu_zx * rk3CovuBfu_zy + (vector(rk3Bfu_y, i, j, k) * rk3Hv_x - vector(rk3Bfu_x, i, j, k) * rk3Hv_y) / rk3sdetg;
							rk3HE_z = 0.5 * ((-2.0 * rk3gtu_yz * rk3HM_xz) + 2.0 * rk3gtu_zz * rk3HM_xy + 2.0 * rk3gtu_xz * rk3HM_yz) / sqrt(rk3chi_max);
							rk3HE_x = 0.5 * ((-2.0 * rk3gtu_xy * rk3HM_xz) + 2.0 * rk3gtu_xz * rk3HM_xy + 2.0 * rk3gtu_xx * rk3HM_yz) / sqrt(rk3chi_max);
							rk3HE_y = 0.5 * ((-2.0 * rk3gtu_yy * rk3HM_xz) + 2.0 * rk3gtu_yz * rk3HM_xy + 2.0 * rk3gtu_xy * rk3HM_yz) / sqrt(rk3chi_max);
							rk3HT_zz = 4.0 * rk3CovdEpf_x * rk3Covuvfu_xz * rk3vfu_z + 4.0 * rk3CovdEpf_y * rk3Covuvfu_yz * rk3vfu_z + 4.0 * rk3CovdEpf_z * rk3Covuvfu_zz * rk3vfu_z + rk3Epf * (2.0 * rk3vfu_z * rk3Hv_z + 2.0 * rk3Covdvfu_xz * rk3Covuvfu_xz + 2.0 * rk3Covdvfu_yz * rk3Covuvfu_yz + 2.0 * rk3Covdvfu_zz * rk3Covuvfu_zz) + 2.0 * (rk3vfu_z * rk3vfu_z) * rk3HPres + (-2.0 * rk3CovdBfu_xz * rk3CovuBfu_xz) + (-2.0 * rk3CovdBfu_yz * rk3CovuBfu_yz) + (-2.0 * rk3CovdBfu_zz * rk3CovuBfu_zz) + (-2.0 * rk3CovdEfu_xz * rk3CovuEfu_xz) + (-2.0 * rk3CovdEfu_yz * rk3CovuEfu_yz) + (-2.0 * rk3CovdEfu_zz * rk3CovuEfu_zz) + (-2.00 * vector(Efu_z, i, j, k) * rk3HE_z) + (rk3chi_max * rk3gtu_zz - rk3vfu_z * rk3vfu_z) * (rk3HPres + rk3CovdBfd_xx * rk3CovuBfu_xx + rk3CovdBfd_yx * rk3CovuBfu_yx + rk3CovdBfd_zx * rk3CovuBfu_zx + rk3CovdBfd_xy * rk3CovuBfu_xy + rk3CovdBfd_yy * rk3CovuBfu_yy + rk3CovdBfd_zy * rk3CovuBfu_zy + rk3CovdBfd_xz * rk3CovuBfu_xz + rk3CovdBfd_yz * rk3CovuBfu_yz + rk3CovdBfd_zz * rk3CovuBfu_zz + rk3CovdEfd_xx * rk3CovuEfu_xx + rk3CovdEfd_yx * rk3CovuEfu_yx + rk3CovdEfd_zx * rk3CovuEfu_zx + rk3CovdEfd_xy * rk3CovuEfu_xy + rk3CovdEfd_yy * rk3CovuEfu_yy + rk3CovdEfd_zy * rk3CovuEfu_zy + rk3CovdEfd_xz * rk3CovuEfu_xz + rk3CovdEfd_yz * rk3CovuEfu_yz + rk3CovdEfd_zz * rk3CovuEfu_zz + rk3Efd_x * rk3HE_x + rk3Efd_y * rk3HE_y + rk3Efd_z * rk3HE_z);
							rk3HT_yy = 4.0 * rk3CovdEpf_x * rk3Covuvfu_xy * rk3vfu_y + 4.0 * rk3CovdEpf_y * rk3Covuvfu_yy * rk3vfu_y + 4.0 * rk3CovdEpf_z * rk3Covuvfu_zy * rk3vfu_y + rk3Epf * (2.0 * rk3vfu_y * rk3Hv_y + 2.0 * rk3Covdvfu_xy * rk3Covuvfu_xy + 2.0 * rk3Covdvfu_yy * rk3Covuvfu_yy + 2.0 * rk3Covdvfu_zy * rk3Covuvfu_zy) + 2.0 * (rk3vfu_y * rk3vfu_y) * rk3HPres + (-2.0 * rk3CovdBfu_xy * rk3CovuBfu_xy) + (-2.0 * rk3CovdBfu_yy * rk3CovuBfu_yy) + (-2.0 * rk3CovdBfu_zy * rk3CovuBfu_zy) + (-2.0 * rk3CovdEfu_xy * rk3CovuEfu_xy) + (-2.0 * rk3CovdEfu_yy * rk3CovuEfu_yy) + (-2.0 * rk3CovdEfu_zy * rk3CovuEfu_zy) + (-2.00 * vector(Efu_y, i, j, k) * rk3HE_y) + (rk3chi_max * rk3gtu_yy - rk3vfu_y * rk3vfu_y) * (rk3HPres + rk3CovdBfd_xx * rk3CovuBfu_xx + rk3CovdBfd_yx * rk3CovuBfu_yx + rk3CovdBfd_zx * rk3CovuBfu_zx + rk3CovdBfd_xy * rk3CovuBfu_xy + rk3CovdBfd_yy * rk3CovuBfu_yy + rk3CovdBfd_zy * rk3CovuBfu_zy + rk3CovdBfd_xz * rk3CovuBfu_xz + rk3CovdBfd_yz * rk3CovuBfu_yz + rk3CovdBfd_zz * rk3CovuBfu_zz + rk3CovdEfd_xx * rk3CovuEfu_xx + rk3CovdEfd_yx * rk3CovuEfu_yx + rk3CovdEfd_zx * rk3CovuEfu_zx + rk3CovdEfd_xy * rk3CovuEfu_xy + rk3CovdEfd_yy * rk3CovuEfu_yy + rk3CovdEfd_zy * rk3CovuEfu_zy + rk3CovdEfd_xz * rk3CovuEfu_xz + rk3CovdEfd_yz * rk3CovuEfu_yz + rk3CovdEfd_zz * rk3CovuEfu_zz + rk3Efd_x * rk3HE_x + rk3Efd_y * rk3HE_y + rk3Efd_z * rk3HE_z);
							rk3HT_yz = 2.0 * rk3CovdEpf_x * (rk3Covuvfu_xz * rk3vfu_y + rk3Covuvfu_xy * rk3vfu_z) + 2.0 * rk3CovdEpf_y * (rk3Covuvfu_yz * rk3vfu_y + rk3Covuvfu_yy * rk3vfu_z) + 2.0 * rk3CovdEpf_z * (rk3Covuvfu_zz * rk3vfu_y + rk3Covuvfu_zy * rk3vfu_z) + rk3Epf * (rk3vfu_y * rk3Hv_z + rk3vfu_z * rk3Hv_y + 2.0 * rk3Covdvfu_xy * rk3Covuvfu_xz + 2.0 * rk3Covdvfu_yy * rk3Covuvfu_yz + 2.0 * rk3Covdvfu_zy * rk3Covuvfu_zz) + 2.0 * rk3vfu_y * rk3vfu_z * rk3HPres + (-2.0 * rk3CovdBfu_xy * rk3CovuBfu_xz) + (-2.0 * rk3CovdBfu_yy * rk3CovuBfu_yz) + (-2.0 * rk3CovdBfu_zy * rk3CovuBfu_zz) + (-2.0 * rk3CovdEfu_xy * rk3CovuEfu_xz) + (-2.0 * rk3CovdEfu_yy * rk3CovuEfu_yz) + (-2.0 * rk3CovdEfu_zy * rk3CovuEfu_zz) + (-1.00 * vector(Efu_z, i, j, k) * rk3HE_y) + (-1.00 * vector(Efu_y, i, j, k) * rk3HE_z) + (rk3chi_max * rk3gtu_yz - rk3vfu_z * rk3vfu_y) * (rk3HPres + rk3CovdBfd_xx * rk3CovuBfu_xx + rk3CovdBfd_yx * rk3CovuBfu_yx + rk3CovdBfd_zx * rk3CovuBfu_zx + rk3CovdBfd_xy * rk3CovuBfu_xy + rk3CovdBfd_yy * rk3CovuBfu_yy + rk3CovdBfd_zy * rk3CovuBfu_zy + rk3CovdBfd_xz * rk3CovuBfu_xz + rk3CovdBfd_yz * rk3CovuBfu_yz + rk3CovdBfd_zz * rk3CovuBfu_zz + rk3CovdEfd_xx * rk3CovuEfu_xx + rk3CovdEfd_yx * rk3CovuEfu_yx + rk3CovdEfd_zx * rk3CovuEfu_zx + rk3CovdEfd_xy * rk3CovuEfu_xy + rk3CovdEfd_yy * rk3CovuEfu_yy + rk3CovdEfd_zy * rk3CovuEfu_zy + rk3CovdEfd_xz * rk3CovuEfu_xz + rk3CovdEfd_yz * rk3CovuEfu_yz + rk3CovdEfd_zz * rk3CovuEfu_zz + rk3Efd_x * rk3HE_x + rk3Efd_y * rk3HE_y + rk3Efd_z * rk3HE_z);
							rk3HT_xx = 4.0 * rk3CovdEpf_x * rk3Covuvfu_xx * rk3vfu_x + 4.0 * rk3CovdEpf_y * rk3Covuvfu_yx * rk3vfu_x + 4.0 * rk3CovdEpf_z * rk3Covuvfu_zx * rk3vfu_x + rk3Epf * (2.0 * rk3vfu_x * rk3Hv_x + 2.0 * rk3Covdvfu_xx * rk3Covuvfu_xx + 2.0 * rk3Covdvfu_yx * rk3Covuvfu_yx + 2.0 * rk3Covdvfu_zx * rk3Covuvfu_zx) + 2.0 * (rk3vfu_x * rk3vfu_x) * rk3HPres + (-2.0 * rk3CovdBfu_xx * rk3CovuBfu_xx) + (-2.0 * rk3CovdBfu_yx * rk3CovuBfu_yx) + (-2.0 * rk3CovdBfu_zx * rk3CovuBfu_zx) + (-2.0 * rk3CovdEfu_xx * rk3CovuEfu_xx) + (-2.0 * rk3CovdEfu_yx * rk3CovuEfu_yx) + (-2.0 * rk3CovdEfu_zx * rk3CovuEfu_zx) + (-2.00 * vector(Efu_x, i, j, k) * rk3HE_x) + (rk3chi_max * rk3gtu_xx - rk3vfu_x * rk3vfu_x) * (rk3HPres + rk3CovdBfd_xx * rk3CovuBfu_xx + rk3CovdBfd_yx * rk3CovuBfu_yx + rk3CovdBfd_zx * rk3CovuBfu_zx + rk3CovdBfd_xy * rk3CovuBfu_xy + rk3CovdBfd_yy * rk3CovuBfu_yy + rk3CovdBfd_zy * rk3CovuBfu_zy + rk3CovdBfd_xz * rk3CovuBfu_xz + rk3CovdBfd_yz * rk3CovuBfu_yz + rk3CovdBfd_zz * rk3CovuBfu_zz + rk3CovdEfd_xx * rk3CovuEfu_xx + rk3CovdEfd_yx * rk3CovuEfu_yx + rk3CovdEfd_zx * rk3CovuEfu_zx + rk3CovdEfd_xy * rk3CovuEfu_xy + rk3CovdEfd_yy * rk3CovuEfu_yy + rk3CovdEfd_zy * rk3CovuEfu_zy + rk3CovdEfd_xz * rk3CovuEfu_xz + rk3CovdEfd_yz * rk3CovuEfu_yz + rk3CovdEfd_zz * rk3CovuEfu_zz + rk3Efd_x * rk3HE_x + rk3Efd_y * rk3HE_y + rk3Efd_z * rk3HE_z);
							rk3HT_xz = 2.0 * rk3CovdEpf_x * (rk3Covuvfu_xz * rk3vfu_x + rk3Covuvfu_xx * rk3vfu_z) + 2.0 * rk3CovdEpf_y * (rk3Covuvfu_yz * rk3vfu_x + rk3Covuvfu_yx * rk3vfu_z) + 2.0 * rk3CovdEpf_z * (rk3Covuvfu_zz * rk3vfu_x + rk3Covuvfu_zx * rk3vfu_z) + rk3Epf * (rk3vfu_x * rk3Hv_z + rk3vfu_z * rk3Hv_x + 2.0 * rk3Covdvfu_xx * rk3Covuvfu_xz + 2.0 * rk3Covdvfu_yx * rk3Covuvfu_yz + 2.0 * rk3Covdvfu_zx * rk3Covuvfu_zz) + 2.0 * rk3vfu_x * rk3vfu_z * rk3HPres + (-2.0 * rk3CovdBfu_xx * rk3CovuBfu_xz) + (-2.0 * rk3CovdBfu_yx * rk3CovuBfu_yz) + (-2.0 * rk3CovdBfu_zx * rk3CovuBfu_zz) + (-2.0 * rk3CovdEfu_xx * rk3CovuEfu_xz) + (-2.0 * rk3CovdEfu_yx * rk3CovuEfu_yz) + (-2.0 * rk3CovdEfu_zx * rk3CovuEfu_zz) + (-1.00 * vector(Efu_z, i, j, k) * rk3HE_x) + (-1.00 * vector(Efu_x, i, j, k) * rk3HE_z) + (rk3chi_max * rk3gtu_xz - rk3vfu_z * rk3vfu_x) * (rk3HPres + rk3CovdBfd_xx * rk3CovuBfu_xx + rk3CovdBfd_yx * rk3CovuBfu_yx + rk3CovdBfd_zx * rk3CovuBfu_zx + rk3CovdBfd_xy * rk3CovuBfu_xy + rk3CovdBfd_yy * rk3CovuBfu_yy + rk3CovdBfd_zy * rk3CovuBfu_zy + rk3CovdBfd_xz * rk3CovuBfu_xz + rk3CovdBfd_yz * rk3CovuBfu_yz + rk3CovdBfd_zz * rk3CovuBfu_zz + rk3CovdEfd_xx * rk3CovuEfu_xx + rk3CovdEfd_yx * rk3CovuEfu_yx + rk3CovdEfd_zx * rk3CovuEfu_zx + rk3CovdEfd_xy * rk3CovuEfu_xy + rk3CovdEfd_yy * rk3CovuEfu_yy + rk3CovdEfd_zy * rk3CovuEfu_zy + rk3CovdEfd_xz * rk3CovuEfu_xz + rk3CovdEfd_yz * rk3CovuEfu_yz + rk3CovdEfd_zz * rk3CovuEfu_zz + rk3Efd_x * rk3HE_x + rk3Efd_y * rk3HE_y + rk3Efd_z * rk3HE_z);
							rk3HT_xy = 2.0 * rk3CovdEpf_x * (rk3Covuvfu_xy * rk3vfu_x + rk3Covuvfu_xx * rk3vfu_y) + 2.0 * rk3CovdEpf_y * (rk3Covuvfu_yy * rk3vfu_x + rk3Covuvfu_yx * rk3vfu_y) + 2.0 * rk3CovdEpf_z * (rk3Covuvfu_zy * rk3vfu_x + rk3Covuvfu_zx * rk3vfu_y) + rk3Epf * (rk3vfu_x * rk3Hv_y + rk3vfu_y * rk3Hv_x + 2.0 * rk3Covdvfu_xx * rk3Covuvfu_xy + 2.0 * rk3Covdvfu_yx * rk3Covuvfu_yy + 2.0 * rk3Covdvfu_zx * rk3Covuvfu_zy) + 2.0 * rk3vfu_x * rk3vfu_y * rk3HPres + (-2.0 * rk3CovdBfu_xx * rk3CovuBfu_xy) + (-2.0 * rk3CovdBfu_yx * rk3CovuBfu_yy) + (-2.0 * rk3CovdBfu_zx * rk3CovuBfu_zy) + (-2.0 * rk3CovdEfu_xx * rk3CovuEfu_xy) + (-2.0 * rk3CovdEfu_yx * rk3CovuEfu_yy) + (-2.0 * rk3CovdEfu_zx * rk3CovuEfu_zy) + (-1.00 * vector(Efu_y, i, j, k) * rk3HE_x) + (-1.00 * vector(Efu_x, i, j, k) * rk3HE_y) + (rk3chi_max * rk3gtu_xy - rk3vfu_y * rk3vfu_x) * (rk3HPres + rk3CovdBfd_xx * rk3CovuBfu_xx + rk3CovdBfd_yx * rk3CovuBfu_yx + rk3CovdBfd_zx * rk3CovuBfu_zx + rk3CovdBfd_xy * rk3CovuBfu_xy + rk3CovdBfd_yy * rk3CovuBfu_yy + rk3CovdBfd_zy * rk3CovuBfu_zy + rk3CovdBfd_xz * rk3CovuBfu_xz + rk3CovdBfd_yz * rk3CovuBfu_yz + rk3CovdBfd_zz * rk3CovuBfu_zz + rk3CovdEfd_xx * rk3CovuEfu_xx + rk3CovdEfd_yx * rk3CovuEfu_yx + rk3CovdEfd_zx * rk3CovuEfu_zx + rk3CovdEfd_xy * rk3CovuEfu_xy + rk3CovdEfd_yy * rk3CovuEfu_yy + rk3CovdEfd_zy * rk3CovuEfu_zy + rk3CovdEfd_xz * rk3CovuEfu_xz + rk3CovdEfd_yz * rk3CovuEfu_yz + rk3CovdEfd_zz * rk3CovuEfu_zz + rk3Efd_x * rk3HE_x + rk3Efd_y * rk3HE_y + rk3Efd_z * rk3HE_z);
							rk3HNe_x = 2.0 * rk3CovdDYf_x * rk3Covuvfu_xx + 2.0 * rk3CovdDYf_y * rk3Covuvfu_yx + 2.0 * rk3CovdDYf_z * rk3Covuvfu_zx + (vector(rk3DYf, i, j, k) * rk3Hv_x) / rk3sdetg;
							d_TauN_x_o0_t0_m0_l0 = -rk3faceta * C_N * rk3HN_x;
							d_TauN_y_o0_t0_m0_l0 = -rk3faceta * C_N * rk3HN_y;
							d_TauN_z_o0_t0_m0_l0 = -rk3faceta * C_N * rk3HN_z;
							d_TauNe_x_o0_t0_m0_l0 = -rk3faceta * C_N * rk3HNe_x;
							d_TauNe_y_o0_t0_m0_l0 = -rk3faceta * C_N * rk3HNe_y;
							d_TauNe_z_o0_t0_m0_l0 = -rk3faceta * C_N * rk3HNe_z;
							d_TauM_xy_o0_t0_m0_l0 = -rk3faceta * C_M * rk3HM_xy;
							d_TauM_xz_o0_t0_m0_l0 = -rk3faceta * C_M * rk3HM_xz;
							d_TauM_yz_o0_t0_m0_l0 = -rk3faceta * C_M * rk3HM_yz;
							d_TauT_xx_o0_t0_m0_l0 = -rk3faceta * C_T * rk3HT_xx;
							d_TauT_xy_o0_t0_m0_l0 = -rk3faceta * C_T * rk3HT_xy;
							d_TauT_xz_o0_t0_m0_l0 = -rk3faceta * C_T * rk3HT_xz;
							d_TauT_yy_o0_t0_m0_l0 = -rk3faceta * C_T * rk3HT_yy;
							d_TauT_yz_o0_t0_m0_l0 = -rk3faceta * C_T * rk3HT_yz;
							d_TauT_zz_o0_t0_m0_l0 = -rk3faceta * C_T * rk3HT_zz;
							vector(TauNe_x, i, j, k) = d_TauNe_x_o0_t0_m0_l0;
							vector(TauNe_y, i, j, k) = d_TauNe_y_o0_t0_m0_l0;
							vector(TauT_yz, i, j, k) = d_TauT_yz_o0_t0_m0_l0;
							vector(TauT_xy, i, j, k) = d_TauT_xy_o0_t0_m0_l0;
							vector(TauM_yz, i, j, k) = d_TauM_yz_o0_t0_m0_l0;
							vector(TauM_xy, i, j, k) = d_TauM_xy_o0_t0_m0_l0;
							vector(TauNe_z, i, j, k) = d_TauNe_z_o0_t0_m0_l0;
							vector(TauT_xz, i, j, k) = d_TauT_xz_o0_t0_m0_l0;
							vector(TauN_x, i, j, k) = d_TauN_x_o0_t0_m0_l0;
							vector(TauT_zz, i, j, k) = d_TauT_zz_o0_t0_m0_l0;
							vector(TauT_yy, i, j, k) = d_TauT_yy_o0_t0_m0_l0;
							vector(TauT_xx, i, j, k) = d_TauT_xx_o0_t0_m0_l0;
							vector(TauM_xz, i, j, k) = d_TauM_xz_o0_t0_m0_l0;
							vector(TauN_z, i, j, k) = d_TauN_z_o0_t0_m0_l0;
							vector(TauN_y, i, j, k) = d_TauN_y_o0_t0_m0_l0;
						} else {
							rk3CovdEfu_yz = 0.0;
							rk3CovuEfu_zz = 0.0;
							rk3CovuEfu_xz = 0.0;
							rk3CovdEfu_yy = 0.0;
							rk3CovuEfu_zy = 0.0;
							rk3CovuEfu_xy = 0.0;
							rk3CovdEfu_yx = 0.0;
							rk3CovdEfd_yy = 0.0;
							rk3CovdEfd_yz = 0.0;
							rk3CovdEfd_yx = 0.0;
							rk3CovuEfu_zx = 0.0;
							rk3CovuEfu_xx = 0.0;
							rk3CovuEfu_yx = 0.0;
							rk3CovuEfu_yz = 0.0;
							rk3CovuEfu_yy = 0.0;
							rk3PhiThetah = 0.0;
							rk3HPres = 0.0;
							rk3HTheta = 0.0;
							rk3Hv_z = 0.0;
							rk3HN_z = 0.0;
							rk3HNe_z = 0.0;
							rk3Hv_y = 0.0;
							rk3HN_y = 0.0;
							rk3HM_yz = 0.0;
							rk3HNe_y = 0.0;
							rk3Hv_x = 0.0;
							rk3HN_x = 0.0;
							rk3HM_xz = 0.0;
							rk3HM_xy = 0.0;
							rk3HE_z = 0.0;
							rk3HE_x = 0.0;
							rk3HE_y = 0.0;
							rk3HT_zz = 0.0;
							rk3HT_yy = 0.0;
							rk3HT_yz = 0.0;
							rk3HT_xx = 0.0;
							rk3HT_xz = 0.0;
							rk3HT_xy = 0.0;
							rk3HNe_x = 0.0;
						}
						optdepthe_tmp = vector(optdepthe, i, j, k);
						optdeptha_tmp = vector(optdeptha, i, j, k);
						optdepthx_tmp = vector(optdepthx, i, j, k);
						chie_tmp = vector(chie, i, j, k);
						chia_tmp = vector(chia, i, j, k);
						chix_tmp = vector(chix, i, j, k);
						qnu_tmp = vector(qnu, i, j, k);
						rnu_tmp = vector(rnu, i, j, k);
						qnu_e_tmp = vector(qnu_e, i, j, k);
						qnu_a_tmp = vector(qnu_a, i, j, k);
						qnu_x_tmp = vector(qnu_x, i, j, k);
						betabeta = vector(rk3gtd_xx, i, j, k) * vector(rk3Betau_x, i, j, k) * vector(rk3Betau_x, i, j, k) + vector(rk3gtd_yy, i, j, k) * vector(rk3Betau_y, i, j, k) * vector(rk3Betau_y, i, j, k) + vector(rk3gtd_zz, i, j, k) * vector(rk3Betau_z, i, j, k) * vector(rk3Betau_z, i, j, k) + 2.0 * vector(rk3gtd_xy, i, j, k) * vector(rk3Betau_x, i, j, k) * vector(rk3Betau_y, i, j, k) + 2.0 * vector(rk3gtd_xz, i, j, k) * vector(rk3Betau_x, i, j, k) * vector(rk3Betau_z, i, j, k) + 2.0 * vector(rk3gtd_yz, i, j, k) * vector(rk3Betau_y, i, j, k) * vector(rk3Betau_z, i, j, k);
						betav = vector(rk3gtd_xx, i, j, k) * vector(rk3Betau_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk3gtd_yy, i, j, k) * vector(rk3Betau_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk3gtd_zz, i, j, k) * vector(rk3Betau_z, i, j, k) * vector(vfd_z, i, j, k) + vector(rk3gtd_xy, i, j, k) * vector(rk3Betau_x, i, j, k) * vector(vfd_y, i, j, k) + vector(rk3gtd_xy, i, j, k) * vector(rk3Betau_y, i, j, k) * vector(vfd_x, i, j, k) + vector(rk3gtd_xz, i, j, k) * vector(rk3Betau_x, i, j, k) * vector(vfd_z, i, j, k) + vector(rk3gtd_xz, i, j, k) * vector(rk3Betau_z, i, j, k) * vector(vfd_x, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Betau_y, i, j, k) * vector(vfd_z, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Betau_z, i, j, k) * vector(vfd_y, i, j, k);
						velr = vector(vfd_x, i, j, k) * xcoord(i) + vector(vfd_y, i, j, k) * ycoord(j) + vector(vfd_z, i, j, k) * zcoord(k);
						velr = velr / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k) + 0.01);
						f_redshift = rk3sqW * (vector(rk3Alpha, i, j, k) - betav) * (1.0 + velr);
						f_redshift = f_redshift * sqrt(vector(rk3Alpha, i, j, k) * vector(rk3Alpha, i, j, k) - betabeta);
						leakage(do_leakage, vacuum_rho, threshold_leakage_vacuum, energyShift, optdepthe_tmp, optdeptha_tmp, optdepthx_tmp, qnu_e_tmp, qnu_a_tmp, qnu_x_tmp, chie_tmp, chia_tmp, chix_tmp, qnu_tmp, rnu_tmp, vector(rhof, i, j, k), vector(epsf, i, j, k), vector(Yef, i, j, k), vector(Tf, i, j, k), rk3sdetg, rk3optdeptheh, rk3optdepthah, rk3optdepthxh, rk3optdepthev, rk3optdepthav, rk3optdepthxv, rk3optdepthed, rk3optdepthad, rk3optdepthxd, rk3chieh, rk3chiah, rk3chixh, rk3chiev, rk3chiav, rk3chixv, rk3chied, rk3chiad, rk3chixd, f_redshift, coord0_data_DataLeakage, coord0_dx_DataLeakage, coord0_size_DataLeakage, coord1_data_DataLeakage, coord1_dx_DataLeakage, coord1_size_DataLeakage, coord2_data_DataLeakage, coord2_dx_DataLeakage, coord2_size_DataLeakage, vars_data_DataLeakage, vars_size_DataLeakage, dx, simPlat_dt, ilast, jlast);
						vector(optdepthe, i, j, k) = optdepthe_tmp;
						vector(optdeptha, i, j, k) = optdeptha_tmp;
						vector(optdepthx, i, j, k) = optdepthx_tmp;
						vector(chie, i, j, k) = chie_tmp;
						vector(chia, i, j, k) = chia_tmp;
						vector(chix, i, j, k) = chix_tmp;
						vector(qnu_e, i, j, k) = qnu_e_tmp;
						vector(qnu_a, i, j, k) = qnu_a_tmp;
						vector(qnu_x, i, j, k) = qnu_x_tmp;
						vector(qnu, i, j, k) = qnu_tmp;
						vector(rnu, i, j, k) = rnu_tmp;
					}
				}
			}
		}
	}
	//Fill ghosts and periodical boundaries
	d_bdry_sched_advance19[ln]->fillData(current_time, false);
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		//Hard region field distance variables
		double* d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
	
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
		double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
		double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
		double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
		double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
		double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
		double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
		double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
		double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
		double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
		double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
		double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
		double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
		double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
		double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
		double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
		double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
		double* rk3gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xx_id).get())->getPointer();
		double* rk3gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xy_id).get())->getPointer();
		double* rk3gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xz_id).get())->getPointer();
		double* rk3gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_yy_id).get())->getPointer();
		double* rk3gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_yz_id).get())->getPointer();
		double* rk3gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_zz_id).get())->getPointer();
		double* rk3Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_xx_id).get())->getPointer();
		double* rk3Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_xy_id).get())->getPointer();
		double* rk3Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_xz_id).get())->getPointer();
		double* rk3Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_yy_id).get())->getPointer();
		double* rk3Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_yz_id).get())->getPointer();
		double* rk3Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_zz_id).get())->getPointer();
		double* rk3Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Gamh_x_id).get())->getPointer();
		double* rk3Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Gamh_y_id).get())->getPointer();
		double* rk3Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Gamh_z_id).get())->getPointer();
		double* rk3Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_x_id).get())->getPointer();
		double* rk3Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_y_id).get())->getPointer();
		double* rk3Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_z_id).get())->getPointer();
		double* rk3Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Alpha_id).get())->getPointer();
		double* rk3chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3chi_id).get())->getPointer();
		double* rk3trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3trK_id).get())->getPointer();
		double* rk3theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3theta_id).get())->getPointer();
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, rk3theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
	}
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		double* rk3chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3chi_id).get())->getPointer();
		double* rk3gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xy_id).get())->getPointer();
		double* rk3Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_x_id).get())->getPointer();
		double* rk3gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_yy_id).get())->getPointer();
		double* rk3Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_y_id).get())->getPointer();
		double* rk3gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_yz_id).get())->getPointer();
		double* rk3Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Bfu_z_id).get())->getPointer();
		double* rk3gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xz_id).get())->getPointer();
		double* rk3gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_zz_id).get())->getPointer();
		double* rk3gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3gtd_xx_id).get())->getPointer();
		double* rk3Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_x_id).get())->getPointer();
		double* rk3Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_y_id).get())->getPointer();
		double* rk3Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Sfd_z_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* rk3Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_x_id).get())->getPointer();
		double* rk3Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Alpha_id).get())->getPointer();
		double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
		double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
		double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
		double* rk3phif = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3phif_id).get())->getPointer();
		double* rk3Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_y_id).get())->getPointer();
		double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
		double* rk3Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Betau_z_id).get())->getPointer();
		double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
		double* rk3Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Df_id).get())->getPointer();
		double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
		double* rk3DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3DYf_id).get())->getPointer();
		double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
		double* rk3tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3tauf_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
		double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
		double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
		double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
		double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
		double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
		double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
		double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
		double* rk3Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Gamh_x_id).get())->getPointer();
		double* rk3Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Gamh_y_id).get())->getPointer();
		double* rk3Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Gamh_z_id).get())->getPointer();
		double* rk3Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_xx_id).get())->getPointer();
		double* rk3Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_xy_id).get())->getPointer();
		double* rk3Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_xz_id).get())->getPointer();
		double* rk3Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_yy_id).get())->getPointer();
		double* rk3Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_yz_id).get())->getPointer();
		double* rk3Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3Atd_zz_id).get())->getPointer();
		double* rk3trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3trK_id).get())->getPointer();
		double* rk3theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk3theta_id).get())->getPointer();
		double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
		double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
		double* Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_id).get())->getPointer();
		double* Sfd_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_p_id).get())->getPointer();
		double* rk1Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_x_id).get())->getPointer();
		double* rk2Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_x_id).get())->getPointer();
		double* Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_id).get())->getPointer();
		double* Sfd_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_p_id).get())->getPointer();
		double* rk1Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_y_id).get())->getPointer();
		double* rk2Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_y_id).get())->getPointer();
		double* Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_id).get())->getPointer();
		double* Sfd_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_p_id).get())->getPointer();
		double* rk1Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Sfd_z_id).get())->getPointer();
		double* rk2Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Sfd_z_id).get())->getPointer();
		double* Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_id).get())->getPointer();
		double* Bfu_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_p_id).get())->getPointer();
		double* rk1Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_x_id).get())->getPointer();
		double* rk2Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_x_id).get())->getPointer();
		double* Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_id).get())->getPointer();
		double* Bfu_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_p_id).get())->getPointer();
		double* rk1Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_y_id).get())->getPointer();
		double* rk2Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_y_id).get())->getPointer();
		double* Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_id).get())->getPointer();
		double* Bfu_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_p_id).get())->getPointer();
		double* rk1Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Bfu_z_id).get())->getPointer();
		double* rk2Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Bfu_z_id).get())->getPointer();
		double* Df = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_id).get())->getPointer();
		double* Df_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_p_id).get())->getPointer();
		double* rk1Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Df_id).get())->getPointer();
		double* rk2Df = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Df_id).get())->getPointer();
		double* DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_DYf_id).get())->getPointer();
		double* DYf_p = ((pdat::NodeData<double> *) patch->getPatchData(d_DYf_p_id).get())->getPointer();
		double* rk1DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1DYf_id).get())->getPointer();
		double* rk2DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2DYf_id).get())->getPointer();
		double* tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_tauf_id).get())->getPointer();
		double* tauf_p = ((pdat::NodeData<double> *) patch->getPatchData(d_tauf_p_id).get())->getPointer();
		double* rk1tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1tauf_id).get())->getPointer();
		double* rk2tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2tauf_id).get())->getPointer();
		double* phif = ((pdat::NodeData<double> *) patch->getPatchData(d_phif_id).get())->getPointer();
		double* phif_p = ((pdat::NodeData<double> *) patch->getPatchData(d_phif_p_id).get())->getPointer();
		double* rk1phif = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1phif_id).get())->getPointer();
		double* rk2phif = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2phif_id).get())->getPointer();
		double* gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_id).get())->getPointer();
		double* gtd_xx_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_p_id).get())->getPointer();
		double* rk1gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xx_id).get())->getPointer();
		double* rk2gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xx_id).get())->getPointer();
		double* gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_id).get())->getPointer();
		double* gtd_xy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_p_id).get())->getPointer();
		double* rk1gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xy_id).get())->getPointer();
		double* rk2gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xy_id).get())->getPointer();
		double* gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_id).get())->getPointer();
		double* gtd_xz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_p_id).get())->getPointer();
		double* rk1gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_xz_id).get())->getPointer();
		double* rk2gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_xz_id).get())->getPointer();
		double* gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_id).get())->getPointer();
		double* gtd_yy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_p_id).get())->getPointer();
		double* rk1gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yy_id).get())->getPointer();
		double* rk2gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yy_id).get())->getPointer();
		double* gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_id).get())->getPointer();
		double* gtd_yz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_p_id).get())->getPointer();
		double* rk1gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_yz_id).get())->getPointer();
		double* rk2gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_yz_id).get())->getPointer();
		double* gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_id).get())->getPointer();
		double* gtd_zz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_p_id).get())->getPointer();
		double* rk1gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1gtd_zz_id).get())->getPointer();
		double* rk2gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2gtd_zz_id).get())->getPointer();
		double* Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xx_id).get())->getPointer();
		double* Atd_xx_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xx_p_id).get())->getPointer();
		double* rk1Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xx_id).get())->getPointer();
		double* rk2Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xx_id).get())->getPointer();
		double* Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xy_id).get())->getPointer();
		double* Atd_xy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xy_p_id).get())->getPointer();
		double* rk1Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xy_id).get())->getPointer();
		double* rk2Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xy_id).get())->getPointer();
		double* Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xz_id).get())->getPointer();
		double* Atd_xz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xz_p_id).get())->getPointer();
		double* rk1Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_xz_id).get())->getPointer();
		double* rk2Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_xz_id).get())->getPointer();
		double* Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yy_id).get())->getPointer();
		double* Atd_yy_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yy_p_id).get())->getPointer();
		double* rk1Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_yy_id).get())->getPointer();
		double* rk2Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_yy_id).get())->getPointer();
		double* Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yz_id).get())->getPointer();
		double* Atd_yz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yz_p_id).get())->getPointer();
		double* rk1Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_yz_id).get())->getPointer();
		double* rk2Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_yz_id).get())->getPointer();
		double* Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_zz_id).get())->getPointer();
		double* Atd_zz_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_zz_p_id).get())->getPointer();
		double* rk1Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Atd_zz_id).get())->getPointer();
		double* rk2Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Atd_zz_id).get())->getPointer();
		double* Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_x_id).get())->getPointer();
		double* Gamh_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_x_p_id).get())->getPointer();
		double* rk1Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_x_id).get())->getPointer();
		double* rk2Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_x_id).get())->getPointer();
		double* Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_y_id).get())->getPointer();
		double* Gamh_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_y_p_id).get())->getPointer();
		double* rk1Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_y_id).get())->getPointer();
		double* rk2Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_y_id).get())->getPointer();
		double* Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_z_id).get())->getPointer();
		double* Gamh_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_z_p_id).get())->getPointer();
		double* rk1Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Gamh_z_id).get())->getPointer();
		double* rk2Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Gamh_z_id).get())->getPointer();
		double* Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_id).get())->getPointer();
		double* Betau_x_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_p_id).get())->getPointer();
		double* rk1Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_x_id).get())->getPointer();
		double* rk2Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_x_id).get())->getPointer();
		double* Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_id).get())->getPointer();
		double* Betau_y_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_p_id).get())->getPointer();
		double* rk1Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_y_id).get())->getPointer();
		double* rk2Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_y_id).get())->getPointer();
		double* Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_id).get())->getPointer();
		double* Betau_z_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_p_id).get())->getPointer();
		double* rk1Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Betau_z_id).get())->getPointer();
		double* rk2Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Betau_z_id).get())->getPointer();
		double* Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_id).get())->getPointer();
		double* Alpha_p = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_p_id).get())->getPointer();
		double* rk1Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1Alpha_id).get())->getPointer();
		double* rk2Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2Alpha_id).get())->getPointer();
		double* chi = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_id).get())->getPointer();
		double* chi_p = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_p_id).get())->getPointer();
		double* rk1chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1chi_id).get())->getPointer();
		double* rk2chi = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2chi_id).get())->getPointer();
		double* trK = ((pdat::NodeData<double> *) patch->getPatchData(d_trK_id).get())->getPointer();
		double* trK_p = ((pdat::NodeData<double> *) patch->getPatchData(d_trK_p_id).get())->getPointer();
		double* rk1trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1trK_id).get())->getPointer();
		double* rk2trK = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2trK_id).get())->getPointer();
		double* theta = ((pdat::NodeData<double> *) patch->getPatchData(d_theta_id).get())->getPointer();
		double* theta_p = ((pdat::NodeData<double> *) patch->getPatchData(d_theta_p_id).get())->getPointer();
		double* rk1theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk1theta_id).get())->getPointer();
		double* rk2theta = ((pdat::NodeData<double> *) patch->getPatchData(d_rk2theta_id).get())->getPointer();
		double rk3chi_max, rk3sdetg, rk3inv_chi, rk3Bfd_y, rk3Bfd_z, rk3Bfd_x, rk3sqBf, rk3detgtd, rk3idetgtd, rk3gtu_zz, rk3gtu_xx, rk3gtu_xz, rk3gtu_xy, rk3Sfu_x, rk3vfu_x, rk3Bfvf, rk3gtu_yy, rk3gtu_yz, rk3vfu_z, rk3Sfu_y, rk3Sfu_z, rk3vfu_y, rk3sqvf, rk3sqW, rk3Sfuu_zz, rk3Sfuu_xx, rk3Sfuu_xz, rk3Sfuu_xy, rk3Sfud_xx, rk3Sfud_xy, rk3Sfud_xz, rk3Sfuu_yz, rk3Sfud_zy, rk3Sfud_zz, rk3Sfud_zx, rk3Sfuu_yy, rk3Sfud_yx, rk3Sfud_yy, rk3Sfud_yz, FluxiSfd_x, FluxiSfd_y, FluxiSfd_z, FluxiBfu_x, FluxiBfu_y, FluxiBfu_z, FluxiDf, FluxiDYf, Fluxitauf, Fluxiphif, rk3h, rk3sqbc, rk3sqca, rk3sqa, SpeediSfd_x, SpeediSfd_y, SpeediSfd_z, SpeediBfu_x, SpeediBfu_y, SpeediBfu_z, SpeediDf, SpeediDYf, Speeditauf, Speediphif, Fluxm3iSfd_x, Fluxm3iSfd_y, Fluxm3iSfd_z, Fluxm3iBfu_x, Fluxm3iBfu_y, Fluxm3iBfu_z, Fluxm3iDf, Fluxm3iDYf, Fluxm3itauf, Fluxm3iphif, Fluxm2iSfd_x, Fluxm2iSfd_y, Fluxm2iSfd_z, Fluxm2iBfu_x, Fluxm2iBfu_y, Fluxm2iBfu_z, Fluxm2iDf, Fluxm2iDYf, Fluxm2itauf, Fluxm2iphif, Fluxm1iSfd_x, Fluxm1iSfd_y, Fluxm1iSfd_z, Fluxm1iBfu_x, Fluxm1iBfu_y, Fluxm1iBfu_z, Fluxm1iDf, Fluxm1iDYf, Fluxm1itauf, Fluxm1iphif, Fluxp1iSfd_x, Fluxp1iSfd_y, Fluxp1iSfd_z, Fluxp1iBfu_x, Fluxp1iBfu_y, Fluxp1iBfu_z, Fluxp1iDf, Fluxp1iDYf, Fluxp1itauf, Fluxp1iphif, Fluxp2iSfd_x, Fluxp2iSfd_y, Fluxp2iSfd_z, Fluxp2iBfu_x, Fluxp2iBfu_y, Fluxp2iBfu_z, Fluxp2iDf, Fluxp2iDYf, Fluxp2itauf, Fluxp2iphif, Fluxp3iSfd_x, Fluxp3iSfd_y, Fluxp3iSfd_z, Fluxp3iBfu_x, Fluxp3iBfu_y, Fluxp3iBfu_z, Fluxp3iDf, Fluxp3iDYf, Fluxp3itauf, Fluxp3iphif, LLFpiSfd_x, LLFpiSfd_y, LLFpiSfd_z, LLFpiBfu_x, LLFpiBfu_y, LLFpiBfu_z, LLFpiDf, LLFpiDYf, LLFpitauf, LLFpiphif, LLFniSfd_x, LLFniSfd_y, LLFniSfd_z, LLFniBfu_x, LLFniBfu_y, LLFniBfu_z, LLFniDf, LLFniDYf, LLFnitauf, LLFniphif, LLFp_m3iSfd_x, LLFp_m3iSfd_y, LLFp_m3iSfd_z, LLFp_m3iBfu_x, LLFp_m3iBfu_y, LLFp_m3iBfu_z, LLFp_m3iDf, LLFp_m3iDYf, LLFp_m3itauf, LLFp_m3iphif, LLFn_m3iSfd_x, LLFn_m3iSfd_y, LLFn_m3iSfd_z, LLFn_m3iBfu_x, LLFn_m3iBfu_y, LLFn_m3iBfu_z, LLFn_m3iDf, LLFn_m3iDYf, LLFn_m3itauf, LLFn_m3iphif, LLFp_m2iSfd_x, LLFp_m2iSfd_y, LLFp_m2iSfd_z, LLFp_m2iBfu_x, LLFp_m2iBfu_y, LLFp_m2iBfu_z, LLFp_m2iDf, LLFp_m2iDYf, LLFp_m2itauf, LLFp_m2iphif, LLFn_m2iSfd_x, LLFn_m2iSfd_y, LLFn_m2iSfd_z, LLFn_m2iBfu_x, LLFn_m2iBfu_y, LLFn_m2iBfu_z, LLFn_m2iDf, LLFn_m2iDYf, LLFn_m2itauf, LLFn_m2iphif, LLFp_m1iSfd_x, LLFp_m1iSfd_y, LLFp_m1iSfd_z, LLFp_m1iBfu_x, LLFp_m1iBfu_y, LLFp_m1iBfu_z, LLFp_m1iDf, LLFp_m1iDYf, LLFp_m1itauf, LLFp_m1iphif, LLFn_m1iSfd_x, LLFn_m1iSfd_y, LLFn_m1iSfd_z, LLFn_m1iBfu_x, LLFn_m1iBfu_y, LLFn_m1iBfu_z, LLFn_m1iDf, LLFn_m1iDYf, LLFn_m1itauf, LLFn_m1iphif, LLFp_p1iSfd_x, LLFp_p1iSfd_y, LLFp_p1iSfd_z, LLFp_p1iBfu_x, LLFp_p1iBfu_y, LLFp_p1iBfu_z, LLFp_p1iDf, LLFp_p1iDYf, LLFp_p1itauf, LLFp_p1iphif, LLFn_p1iSfd_x, LLFn_p1iSfd_y, LLFn_p1iSfd_z, LLFn_p1iBfu_x, LLFn_p1iBfu_y, LLFn_p1iBfu_z, LLFn_p1iDf, LLFn_p1iDYf, LLFn_p1itauf, LLFn_p1iphif, LLFp_p2iSfd_x, LLFp_p2iSfd_y, LLFp_p2iSfd_z, LLFp_p2iBfu_x, LLFp_p2iBfu_y, LLFp_p2iBfu_z, LLFp_p2iDf, LLFp_p2iDYf, LLFp_p2itauf, LLFp_p2iphif, LLFn_p2iSfd_x, LLFn_p2iSfd_y, LLFn_p2iSfd_z, LLFn_p2iBfu_x, LLFn_p2iBfu_y, LLFn_p2iBfu_z, LLFn_p2iDf, LLFn_p2iDYf, LLFn_p2itauf, LLFn_p2iphif, LLFp_p3iSfd_x, LLFp_p3iSfd_y, LLFp_p3iSfd_z, LLFp_p3iBfu_x, LLFp_p3iBfu_y, LLFp_p3iBfu_z, LLFp_p3iDf, LLFp_p3iDYf, LLFp_p3itauf, LLFp_p3iphif, LLFn_p3iSfd_x, LLFn_p3iSfd_y, LLFn_p3iSfd_z, LLFn_p3iBfu_x, LLFn_p3iBfu_y, LLFn_p3iBfu_z, LLFn_p3iDf, LLFn_p3iDYf, LLFn_p3itauf, LLFn_p3iphif, FluxjSfd_x, FluxjSfd_y, FluxjSfd_z, FluxjBfu_x, FluxjBfu_y, FluxjBfu_z, FluxjDf, FluxjDYf, Fluxjtauf, Fluxjphif, SpeedjSfd_x, SpeedjSfd_y, SpeedjSfd_z, SpeedjBfu_x, SpeedjBfu_y, SpeedjBfu_z, SpeedjDf, SpeedjDYf, Speedjtauf, Speedjphif, Fluxm3jSfd_x, Fluxm3jSfd_y, Fluxm3jSfd_z, Fluxm3jBfu_x, Fluxm3jBfu_y, Fluxm3jBfu_z, Fluxm3jDf, Fluxm3jDYf, Fluxm3jtauf, Fluxm3jphif, Fluxm2jSfd_x, Fluxm2jSfd_y, Fluxm2jSfd_z, Fluxm2jBfu_x, Fluxm2jBfu_y, Fluxm2jBfu_z, Fluxm2jDf, Fluxm2jDYf, Fluxm2jtauf, Fluxm2jphif, Fluxm1jSfd_x, Fluxm1jSfd_y, Fluxm1jSfd_z, Fluxm1jBfu_x, Fluxm1jBfu_y, Fluxm1jBfu_z, Fluxm1jDf, Fluxm1jDYf, Fluxm1jtauf, Fluxm1jphif, Fluxp1jSfd_x, Fluxp1jSfd_y, Fluxp1jSfd_z, Fluxp1jBfu_x, Fluxp1jBfu_y, Fluxp1jBfu_z, Fluxp1jDf, Fluxp1jDYf, Fluxp1jtauf, Fluxp1jphif, Fluxp2jSfd_x, Fluxp2jSfd_y, Fluxp2jSfd_z, Fluxp2jBfu_x, Fluxp2jBfu_y, Fluxp2jBfu_z, Fluxp2jDf, Fluxp2jDYf, Fluxp2jtauf, Fluxp2jphif, Fluxp3jSfd_x, Fluxp3jSfd_y, Fluxp3jSfd_z, Fluxp3jBfu_x, Fluxp3jBfu_y, Fluxp3jBfu_z, Fluxp3jDf, Fluxp3jDYf, Fluxp3jtauf, Fluxp3jphif, LLFpjSfd_x, LLFpjSfd_y, LLFpjSfd_z, LLFpjBfu_x, LLFpjBfu_y, LLFpjBfu_z, LLFpjDf, LLFpjDYf, LLFpjtauf, LLFpjphif, LLFnjSfd_x, LLFnjSfd_y, LLFnjSfd_z, LLFnjBfu_x, LLFnjBfu_y, LLFnjBfu_z, LLFnjDf, LLFnjDYf, LLFnjtauf, LLFnjphif, LLFp_m3jSfd_x, LLFp_m3jSfd_y, LLFp_m3jSfd_z, LLFp_m3jBfu_x, LLFp_m3jBfu_y, LLFp_m3jBfu_z, LLFp_m3jDf, LLFp_m3jDYf, LLFp_m3jtauf, LLFp_m3jphif, LLFn_m3jSfd_x, LLFn_m3jSfd_y, LLFn_m3jSfd_z, LLFn_m3jBfu_x, LLFn_m3jBfu_y, LLFn_m3jBfu_z, LLFn_m3jDf, LLFn_m3jDYf, LLFn_m3jtauf, LLFn_m3jphif, LLFp_m2jSfd_x, LLFp_m2jSfd_y, LLFp_m2jSfd_z, LLFp_m2jBfu_x, LLFp_m2jBfu_y, LLFp_m2jBfu_z, LLFp_m2jDf, LLFp_m2jDYf, LLFp_m2jtauf, LLFp_m2jphif, LLFn_m2jSfd_x, LLFn_m2jSfd_y, LLFn_m2jSfd_z, LLFn_m2jBfu_x, LLFn_m2jBfu_y, LLFn_m2jBfu_z, LLFn_m2jDf, LLFn_m2jDYf, LLFn_m2jtauf, LLFn_m2jphif, LLFp_m1jSfd_x, LLFp_m1jSfd_y, LLFp_m1jSfd_z, LLFp_m1jBfu_x, LLFp_m1jBfu_y, LLFp_m1jBfu_z, LLFp_m1jDf, LLFp_m1jDYf, LLFp_m1jtauf, LLFp_m1jphif, LLFn_m1jSfd_x, LLFn_m1jSfd_y, LLFn_m1jSfd_z, LLFn_m1jBfu_x, LLFn_m1jBfu_y, LLFn_m1jBfu_z, LLFn_m1jDf, LLFn_m1jDYf, LLFn_m1jtauf, LLFn_m1jphif, LLFp_p1jSfd_x, LLFp_p1jSfd_y, LLFp_p1jSfd_z, LLFp_p1jBfu_x, LLFp_p1jBfu_y, LLFp_p1jBfu_z, LLFp_p1jDf, LLFp_p1jDYf, LLFp_p1jtauf, LLFp_p1jphif, LLFn_p1jSfd_x, LLFn_p1jSfd_y, LLFn_p1jSfd_z, LLFn_p1jBfu_x, LLFn_p1jBfu_y, LLFn_p1jBfu_z, LLFn_p1jDf, LLFn_p1jDYf, LLFn_p1jtauf, LLFn_p1jphif, LLFp_p2jSfd_x, LLFp_p2jSfd_y, LLFp_p2jSfd_z, LLFp_p2jBfu_x, LLFp_p2jBfu_y, LLFp_p2jBfu_z, LLFp_p2jDf, LLFp_p2jDYf, LLFp_p2jtauf, LLFp_p2jphif, LLFn_p2jSfd_x, LLFn_p2jSfd_y, LLFn_p2jSfd_z, LLFn_p2jBfu_x, LLFn_p2jBfu_y, LLFn_p2jBfu_z, LLFn_p2jDf, LLFn_p2jDYf, LLFn_p2jtauf, LLFn_p2jphif, LLFp_p3jSfd_x, LLFp_p3jSfd_y, LLFp_p3jSfd_z, LLFp_p3jBfu_x, LLFp_p3jBfu_y, LLFp_p3jBfu_z, LLFp_p3jDf, LLFp_p3jDYf, LLFp_p3jtauf, LLFp_p3jphif, LLFn_p3jSfd_x, LLFn_p3jSfd_y, LLFn_p3jSfd_z, LLFn_p3jBfu_x, LLFn_p3jBfu_y, LLFn_p3jBfu_z, LLFn_p3jDf, LLFn_p3jDYf, LLFn_p3jtauf, LLFn_p3jphif, FluxkSfd_x, FluxkSfd_y, FluxkSfd_z, FluxkBfu_x, FluxkBfu_y, FluxkBfu_z, FluxkDf, FluxkDYf, Fluxktauf, Fluxkphif, SpeedkSfd_x, SpeedkSfd_y, SpeedkSfd_z, SpeedkBfu_x, SpeedkBfu_y, SpeedkBfu_z, SpeedkDf, SpeedkDYf, Speedktauf, Speedkphif, Fluxm3kSfd_x, Fluxm3kSfd_y, Fluxm3kSfd_z, Fluxm3kBfu_x, Fluxm3kBfu_y, Fluxm3kBfu_z, Fluxm3kDf, Fluxm3kDYf, Fluxm3ktauf, Fluxm3kphif, Fluxm2kSfd_x, Fluxm2kSfd_y, Fluxm2kSfd_z, Fluxm2kBfu_x, Fluxm2kBfu_y, Fluxm2kBfu_z, Fluxm2kDf, Fluxm2kDYf, Fluxm2ktauf, Fluxm2kphif, Fluxm1kSfd_x, Fluxm1kSfd_y, Fluxm1kSfd_z, Fluxm1kBfu_x, Fluxm1kBfu_y, Fluxm1kBfu_z, Fluxm1kDf, Fluxm1kDYf, Fluxm1ktauf, Fluxm1kphif, Fluxp1kSfd_x, Fluxp1kSfd_y, Fluxp1kSfd_z, Fluxp1kBfu_x, Fluxp1kBfu_y, Fluxp1kBfu_z, Fluxp1kDf, Fluxp1kDYf, Fluxp1ktauf, Fluxp1kphif, Fluxp2kSfd_x, Fluxp2kSfd_y, Fluxp2kSfd_z, Fluxp2kBfu_x, Fluxp2kBfu_y, Fluxp2kBfu_z, Fluxp2kDf, Fluxp2kDYf, Fluxp2ktauf, Fluxp2kphif, Fluxp3kSfd_x, Fluxp3kSfd_y, Fluxp3kSfd_z, Fluxp3kBfu_x, Fluxp3kBfu_y, Fluxp3kBfu_z, Fluxp3kDf, Fluxp3kDYf, Fluxp3ktauf, Fluxp3kphif, LLFpkSfd_x, LLFpkSfd_y, LLFpkSfd_z, LLFpkBfu_x, LLFpkBfu_y, LLFpkBfu_z, LLFpkDf, LLFpkDYf, LLFpktauf, LLFpkphif, LLFnkSfd_x, LLFnkSfd_y, LLFnkSfd_z, LLFnkBfu_x, LLFnkBfu_y, LLFnkBfu_z, LLFnkDf, LLFnkDYf, LLFnktauf, LLFnkphif, LLFp_m3kSfd_x, LLFp_m3kSfd_y, LLFp_m3kSfd_z, LLFp_m3kBfu_x, LLFp_m3kBfu_y, LLFp_m3kBfu_z, LLFp_m3kDf, LLFp_m3kDYf, LLFp_m3ktauf, LLFp_m3kphif, LLFn_m3kSfd_x, LLFn_m3kSfd_y, LLFn_m3kSfd_z, LLFn_m3kBfu_x, LLFn_m3kBfu_y, LLFn_m3kBfu_z, LLFn_m3kDf, LLFn_m3kDYf, LLFn_m3ktauf, LLFn_m3kphif, LLFp_m2kSfd_x, LLFp_m2kSfd_y, LLFp_m2kSfd_z, LLFp_m2kBfu_x, LLFp_m2kBfu_y, LLFp_m2kBfu_z, LLFp_m2kDf, LLFp_m2kDYf, LLFp_m2ktauf, LLFp_m2kphif, LLFn_m2kSfd_x, LLFn_m2kSfd_y, LLFn_m2kSfd_z, LLFn_m2kBfu_x, LLFn_m2kBfu_y, LLFn_m2kBfu_z, LLFn_m2kDf, LLFn_m2kDYf, LLFn_m2ktauf, LLFn_m2kphif, LLFp_m1kSfd_x, LLFp_m1kSfd_y, LLFp_m1kSfd_z, LLFp_m1kBfu_x, LLFp_m1kBfu_y, LLFp_m1kBfu_z, LLFp_m1kDf, LLFp_m1kDYf, LLFp_m1ktauf, LLFp_m1kphif, LLFn_m1kSfd_x, LLFn_m1kSfd_y, LLFn_m1kSfd_z, LLFn_m1kBfu_x, LLFn_m1kBfu_y, LLFn_m1kBfu_z, LLFn_m1kDf, LLFn_m1kDYf, LLFn_m1ktauf, LLFn_m1kphif, LLFp_p1kSfd_x, LLFp_p1kSfd_y, LLFp_p1kSfd_z, LLFp_p1kBfu_x, LLFp_p1kBfu_y, LLFp_p1kBfu_z, LLFp_p1kDf, LLFp_p1kDYf, LLFp_p1ktauf, LLFp_p1kphif, LLFn_p1kSfd_x, LLFn_p1kSfd_y, LLFn_p1kSfd_z, LLFn_p1kBfu_x, LLFn_p1kBfu_y, LLFn_p1kBfu_z, LLFn_p1kDf, LLFn_p1kDYf, LLFn_p1ktauf, LLFn_p1kphif, LLFp_p2kSfd_x, LLFp_p2kSfd_y, LLFp_p2kSfd_z, LLFp_p2kBfu_x, LLFp_p2kBfu_y, LLFp_p2kBfu_z, LLFp_p2kDf, LLFp_p2kDYf, LLFp_p2ktauf, LLFp_p2kphif, LLFn_p2kSfd_x, LLFn_p2kSfd_y, LLFn_p2kSfd_z, LLFn_p2kBfu_x, LLFn_p2kBfu_y, LLFn_p2kBfu_z, LLFn_p2kDf, LLFn_p2kDYf, LLFn_p2ktauf, LLFn_p2kphif, LLFp_p3kSfd_x, LLFp_p3kSfd_y, LLFp_p3kSfd_z, LLFp_p3kBfu_x, LLFp_p3kBfu_y, LLFp_p3kBfu_z, LLFp_p3kDf, LLFp_p3kDYf, LLFp_p3ktauf, LLFp_p3kphif, LLFn_p3kSfd_x, LLFn_p3kSfd_y, LLFn_p3kSfd_z, LLFn_p3kBfu_x, LLFn_p3kBfu_y, LLFn_p3kBfu_z, LLFn_p3kDf, LLFn_p3kDYf, LLFn_p3ktauf, LLFn_p3kphif, Sfd_x_lni, Sfd_y_lni, Sfd_z_lni, Bfu_x_lni, Bfu_y_lni, Bfu_z_lni, Df_lni, DYf_lni, tauf_lni, phif_lni, Sfd_x_lnj, Sfd_y_lnj, Sfd_z_lnj, Bfu_x_lnj, Bfu_y_lnj, Bfu_z_lnj, Df_lnj, DYf_lnj, tauf_lnj, phif_lnj, Sfd_x_lnk, Sfd_y_lnk, Sfd_z_lnk, Bfu_x_lnk, Bfu_y_lnk, Bfu_z_lnk, Df_lnk, DYf_lnk, tauf_lnk, phif_lnk, Sfd_x_rni, Sfd_y_rni, Sfd_z_rni, Bfu_x_rni, Bfu_y_rni, Bfu_z_rni, Df_rni, DYf_rni, tauf_rni, phif_rni, Sfd_x_rnj, Sfd_y_rnj, Sfd_z_rnj, Bfu_x_rnj, Bfu_y_rnj, Bfu_z_rnj, Df_rnj, DYf_rnj, tauf_rnj, phif_rnj, Sfd_x_rnk, Sfd_y_rnk, Sfd_z_rnk, Bfu_x_rnk, Bfu_y_rnk, Bfu_z_rnk, Df_rnk, DYf_rnk, tauf_rnk, phif_rnk, Sfd_x_lpi, Sfd_y_lpi, Sfd_z_lpi, Bfu_x_lpi, Bfu_y_lpi, Bfu_z_lpi, Df_lpi, DYf_lpi, tauf_lpi, phif_lpi, Sfd_x_lpj, Sfd_y_lpj, Sfd_z_lpj, Bfu_x_lpj, Bfu_y_lpj, Bfu_z_lpj, Df_lpj, DYf_lpj, tauf_lpj, phif_lpj, Sfd_x_lpk, Sfd_y_lpk, Sfd_z_lpk, Bfu_x_lpk, Bfu_y_lpk, Bfu_z_lpk, Df_lpk, DYf_lpk, tauf_lpk, phif_lpk, Sfd_x_rpi, Sfd_y_rpi, Sfd_z_rpi, Bfu_x_rpi, Bfu_y_rpi, Bfu_z_rpi, Df_rpi, DYf_rpi, tauf_rpi, phif_rpi, Sfd_x_rpj, Sfd_y_rpj, Sfd_z_rpj, Bfu_x_rpj, Bfu_y_rpj, Bfu_z_rpj, Df_rpj, DYf_rpj, tauf_rpj, phif_rpj, Sfd_x_rpk, Sfd_y_rpk, Sfd_z_rpk, Bfu_x_rpk, Bfu_y_rpk, Bfu_z_rpk, Df_rpk, DYf_rpk, tauf_rpk, phif_rpk, FluxniSfd_x, FluxniSfd_y, FluxniSfd_z, FluxniBfu_x, FluxniBfu_y, FluxniBfu_z, FluxniDf, FluxniDYf, Fluxnitauf, Fluxniphif, FluxnjSfd_x, FluxnjSfd_y, FluxnjSfd_z, FluxnjBfu_x, FluxnjBfu_y, FluxnjBfu_z, FluxnjDf, FluxnjDYf, Fluxnjtauf, Fluxnjphif, FluxnkSfd_x, FluxnkSfd_y, FluxnkSfd_z, FluxnkBfu_x, FluxnkBfu_y, FluxnkBfu_z, FluxnkDf, FluxnkDYf, Fluxnktauf, Fluxnkphif, FluxpiSfd_x, FluxpiSfd_y, FluxpiSfd_z, FluxpiBfu_x, FluxpiBfu_y, FluxpiBfu_z, FluxpiDf, FluxpiDYf, Fluxpitauf, Fluxpiphif, FluxpjSfd_x, FluxpjSfd_y, FluxpjSfd_z, FluxpjBfu_x, FluxpjBfu_y, FluxpjBfu_z, FluxpjDf, FluxpjDYf, Fluxpjtauf, Fluxpjphif, FluxpkSfd_x, FluxpkSfd_y, FluxpkSfd_z, FluxpkBfu_x, FluxpkBfu_y, FluxpkBfu_z, FluxpkDf, FluxpkDYf, Fluxpktauf, Fluxpkphif, d_div_Beta_o0_t0_m0_l0, d_Betau_x_o1_t0_m0_l0, d_Betau_x_o2_t0_m0_l0, d_Sfd_y_o0_t7_m0_l0, d_Betau_x_o1_t1_m0_l0, d_Betau_x_o2_t1_m0_l0, d_Sfd_z_o0_t7_m0_l0, d_Betau_x_o1_t2_m0_l0, d_Betau_x_o2_t2_m0_l0, d_d_div_Beta_x_o0_t0_m0_l0, d_d_div_Beta_y_o0_t0_m0_l0, d_d_div_Beta_z_o0_t0_m0_l0, d_Gamh_x_o0_t15_m0_l0, d_Gamh_x_o0_t16_m0_l0, d_Gamh_x_o0_t17_m0_l0, d_Sfd_x_o0_t8_m0_l0, d_Betau_y_o1_t0_m0_l0, d_Betau_y_o2_t0_m0_l0, d_div_Beta_o0_t1_m0_l0, d_Betau_y_o1_t1_m0_l0, d_Betau_y_o2_t1_m0_l0, d_Sfd_z_o0_t8_m0_l0, d_Betau_y_o1_t2_m0_l0, d_Betau_y_o2_t2_m0_l0, d_Gamh_y_o0_t6_m0_l0, d_d_div_Beta_x_o0_t1_m0_l0, d_Gamh_y_o0_t8_m0_l0, d_d_div_Beta_y_o0_t1_m0_l0, d_d_div_Beta_z_o0_t1_m0_l0, d_Gamh_y_o0_t11_m0_l0, d_Sfd_x_o0_t9_m0_l0, d_Betau_z_o1_t0_m0_l0, d_Betau_z_o2_t0_m0_l0, d_Sfd_y_o0_t9_m0_l0, d_Betau_z_o1_t1_m0_l0, d_Betau_z_o2_t1_m0_l0, d_div_Beta_o0_t2_m0_l0, d_Betau_z_o1_t2_m0_l0, d_Betau_z_o2_t2_m0_l0, d_Gamh_z_o0_t0_m0_l0, d_Gamh_z_o0_t1_m0_l0, d_d_div_Beta_x_o0_t2_m0_l0, d_Gamh_z_o0_t3_m0_l0, d_d_div_Beta_y_o0_t2_m0_l0, d_d_div_Beta_z_o0_t2_m0_l0, d_Ctd_xxx_o0_t0_m0_l0, d_gtd_xx_o1_t0_m0_l0, d_gtd_xx_o2_t0_m0_l0, d_Ctd_xxy_o0_t0_m0_l0, d_gtd_xx_o1_t1_m0_l0, d_gtd_xx_o2_t1_m0_l0, d_Ctd_xxz_o0_t0_m0_l0, d_gtd_xx_o1_t2_m0_l0, d_gtd_xx_o2_t2_m0_l0, d_Ctd_yxx_o0_t0_m0_l0, d_gtd_xy_o1_t0_m0_l0, d_gtd_xy_o2_t0_m0_l0, d_Ctd_xyy_o0_t0_m0_l0, d_gtd_xy_o1_t1_m0_l0, d_gtd_xy_o2_t1_m0_l0, d_Ctd_xyz_o0_t1_m0_l0, d_gtd_xy_o1_t2_m0_l0, d_gtd_xy_o2_t2_m0_l0, d_Ctd_xyy_o0_t1_m0_l0, d_gtd_yy_o1_t0_m0_l0, d_gtd_yy_o2_t0_m0_l0, d_Ctd_yyy_o0_t0_m0_l0, d_gtd_yy_o1_t1_m0_l0, d_gtd_yy_o2_t1_m0_l0, d_Ctd_yyz_o0_t0_m0_l0, d_gtd_yy_o1_t2_m0_l0, d_gtd_yy_o2_t2_m0_l0, d_Ctd_zxx_o0_t0_m0_l0, d_gtd_xz_o1_t0_m0_l0, d_gtd_xz_o2_t0_m0_l0, d_Ctd_xyz_o0_t0_m0_l0, d_gtd_xz_o1_t1_m0_l0, d_gtd_xz_o2_t1_m0_l0, d_Ctd_xzz_o0_t0_m0_l0, d_gtd_xz_o1_t2_m0_l0, d_gtd_xz_o2_t2_m0_l0, d_Ctd_xyz_o0_t2_m0_l0, d_gtd_yz_o1_t0_m0_l0, d_gtd_yz_o2_t0_m0_l0, d_Ctd_zyy_o0_t0_m0_l0, d_gtd_yz_o1_t1_m0_l0, d_gtd_yz_o2_t1_m0_l0, d_Ctd_yzz_o0_t0_m0_l0, d_gtd_yz_o1_t2_m0_l0, d_gtd_yz_o2_t2_m0_l0, d_Ctd_xzz_o0_t1_m0_l0, d_gtd_zz_o1_t0_m0_l0, d_gtd_zz_o2_t0_m0_l0, d_Ctd_yzz_o0_t1_m0_l0, d_gtd_zz_o1_t1_m0_l0, d_gtd_zz_o2_t1_m0_l0, d_Ctd_zzz_o0_t0_m0_l0, d_gtd_zz_o1_t2_m0_l0, d_gtd_zz_o2_t2_m0_l0, d_Rpd_xx_o0_t0_m0_l0, d_chi_o1_t0_m0_l0, d_chi_o2_t0_m0_l0, d_Rpd_xx_o0_t1_m0_l0, d_chi_o1_t1_m0_l0, d_chi_o2_t1_m0_l0, d_Rpd_xx_o0_t2_m0_l0, d_chi_o1_t2_m0_l0, d_chi_o2_t2_m0_l0, d_Rtd_xx_o0_t6_m0_l0, d_Gamh_x_o1_t0_m0_l0, d_Gamh_x_o2_t0_m0_l0, d_Rtd_xy_o0_t6_m0_l0, d_Gamh_x_o1_t1_m0_l0, d_Gamh_x_o2_t1_m0_l0, d_Rtd_xz_o0_t6_m0_l0, d_Gamh_x_o1_t2_m0_l0, d_Gamh_x_o2_t2_m0_l0, d_Rtd_xx_o0_t7_m0_l0, d_Gamh_y_o1_t0_m0_l0, d_Gamh_y_o2_t0_m0_l0, d_Rtd_xy_o0_t7_m0_l0, d_Gamh_y_o1_t1_m0_l0, d_Gamh_y_o2_t1_m0_l0, d_Rtd_xz_o0_t7_m0_l0, d_Gamh_y_o1_t2_m0_l0, d_Gamh_y_o2_t2_m0_l0, d_Rtd_xx_o0_t8_m0_l0, d_Gamh_z_o1_t0_m0_l0, d_Gamh_z_o2_t0_m0_l0, d_Rtd_xy_o0_t8_m0_l0, d_Gamh_z_o1_t1_m0_l0, d_Gamh_z_o2_t1_m0_l0, d_Rtd_xz_o0_t8_m0_l0, d_Gamh_z_o1_t2_m0_l0, d_Gamh_z_o2_t2_m0_l0, d_Psi1_xx_o0_t1_m0_l0, d_Alpha_o1_t0_m0_l0, d_Alpha_o2_t0_m0_l0, d_Psi1_xx_o0_t2_m0_l0, d_Alpha_o1_t1_m0_l0, d_Alpha_o2_t1_m0_l0, d_Psi1_xx_o0_t3_m0_l0, d_Alpha_o1_t2_m0_l0, d_Alpha_o2_t2_m0_l0, d_Psi1_xx_o0_t0_m0_l0, d_Psi1_xy_o0_t0_m0_l0, d_Psi1_xz_o0_t0_m0_l0, d_Psi1_yy_o0_t0_m0_l0, d_Psi1_yz_o0_t0_m0_l0, d_Psi1_zz_o0_t0_m0_l0, d_Atd_xx_o1_t0_m0_l0, d_Atd_xx_o2_t0_m0_l0, d_Atd_xx_o1_t1_m0_l0, d_Atd_xx_o2_t1_m0_l0, d_Atd_xx_o1_t2_m0_l0, d_Atd_xx_o2_t2_m0_l0, d_Atd_xy_o1_t0_m0_l0, d_Atd_xy_o2_t0_m0_l0, d_Atd_xy_o1_t1_m0_l0, d_Atd_xy_o2_t1_m0_l0, d_Atd_xy_o1_t2_m0_l0, d_Atd_xy_o2_t2_m0_l0, d_Atd_xz_o1_t0_m0_l0, d_Atd_xz_o2_t0_m0_l0, d_Atd_xz_o1_t1_m0_l0, d_Atd_xz_o2_t1_m0_l0, d_Atd_xz_o1_t2_m0_l0, d_Atd_xz_o2_t2_m0_l0, d_Atd_yy_o1_t0_m0_l0, d_Atd_yy_o2_t0_m0_l0, d_Atd_yy_o1_t1_m0_l0, d_Atd_yy_o2_t1_m0_l0, d_Atd_yy_o1_t2_m0_l0, d_Atd_yy_o2_t2_m0_l0, d_Atd_yz_o1_t0_m0_l0, d_Atd_yz_o2_t0_m0_l0, d_Atd_yz_o1_t1_m0_l0, d_Atd_yz_o2_t1_m0_l0, d_Atd_yz_o1_t2_m0_l0, d_Atd_yz_o2_t2_m0_l0, d_Atd_zz_o1_t0_m0_l0, d_Atd_zz_o2_t0_m0_l0, d_Atd_zz_o1_t1_m0_l0, d_Atd_zz_o2_t1_m0_l0, d_Atd_zz_o1_t2_m0_l0, d_Atd_zz_o2_t2_m0_l0, d_Gamh_x_o0_t0_m0_l0, d_trK_o1_t0_m0_l0, d_trK_o2_t0_m0_l0, d_Gamh_x_o0_t1_m0_l0, d_trK_o1_t1_m0_l0, d_trK_o2_t1_m0_l0, d_Gamh_x_o0_t2_m0_l0, d_trK_o1_t2_m0_l0, d_trK_o2_t2_m0_l0, d_Gamh_x_o0_t3_m0_l0, d_theta_o1_t0_m0_l0, d_theta_o2_t0_m0_l0, d_Gamh_x_o0_t4_m0_l0, d_theta_o1_t1_m0_l0, d_theta_o2_t1_m0_l0, d_Gamh_x_o0_t5_m0_l0, d_theta_o1_t2_m0_l0, d_theta_o2_t2_m0_l0, d_Rtd_xx_o0_t0_m0_l0, d_Rtd_xx_o0_t1_m0_l0, d_Rtd_xx_o0_t2_m0_l0, d_Rtd_xx_o0_t3_m0_l0, d_Rtd_xx_o0_t4_m0_l0, d_Rtd_xx_o0_t5_m0_l0, d_Rtd_xy_o0_t0_m0_l0, d_Rtd_xy_o0_t1_m0_l0, d_Rtd_xy_o0_t2_m0_l0, d_Rtd_xy_o0_t3_m0_l0, d_Rtd_xy_o0_t4_m0_l0, d_Rtd_xy_o0_t5_m0_l0, d_Rtd_yy_o0_t0_m0_l0, d_Rtd_yy_o0_t1_m0_l0, d_Rtd_yy_o0_t2_m0_l0, d_Rtd_yy_o0_t3_m0_l0, d_Rtd_yy_o0_t4_m0_l0, d_Rtd_yy_o0_t5_m0_l0, d_Rtd_xz_o0_t0_m0_l0, d_Rtd_xz_o0_t1_m0_l0, d_Rtd_xz_o0_t2_m0_l0, d_Rtd_xz_o0_t3_m0_l0, d_Rtd_xz_o0_t4_m0_l0, d_Rtd_xz_o0_t5_m0_l0, d_Rtd_yz_o0_t0_m0_l0, d_Rtd_yz_o0_t1_m0_l0, d_Rtd_yz_o0_t2_m0_l0, d_Rtd_yz_o0_t3_m0_l0, d_Rtd_yz_o0_t4_m0_l0, d_Rtd_yz_o0_t5_m0_l0, d_Rtd_zz_o0_t0_m0_l0, d_Rtd_zz_o0_t1_m0_l0, d_Rtd_zz_o0_t2_m0_l0, d_Rtd_zz_o0_t3_m0_l0, d_Rtd_zz_o0_t4_m0_l0, d_Rtd_zz_o0_t5_m0_l0, d_Rpd_xx_o0_t3_m0_l0, d_Rpd_xx_o0_t4_m0_l0, d_Rpd_xx_o0_t5_m0_l0, d_Rpd_xx_o0_t6_m0_l0, d_Rpd_xx_o0_t7_m0_l0, d_Rpd_xx_o0_t8_m0_l0, rk3m_Ctd_yyz_o0_t0_l0, rk3Ctd_yyz, rk3m_Ctd_yyy_o0_t0_l0, rk3Ctd_yyy, rk3div_Beta, rk3m_Ctd_xxz_o0_t0_l0, rk3Ctd_xxz, rk3m_Ctd_xxx_o0_t0_l0, rk3Ctd_xxx, rk3m_Ctd_yzz_o0_t1_l0, rk3Ctd_yzz, rk3m_Ctd_xxy_o0_t0_l0, rk3Ctd_xxy, rk3kappa_z2, rk3m_Ctd_xyy_o0_t1_l0, rk3Ctd_xyy, rk3m_Ctd_xyz_o0_t2_l0, rk3m_Ctd_xyz_o0_t1_l0, rk3m_Ctd_xyz_o0_t0_l0, rk3Ctd_xyz, rk3d_div_Beta_y, rk3d_div_Beta_x, rk3d_div_Beta_z, rk3m_Ctd_xzz_o0_t1_l0, rk3Ctd_xzz, rk3m_Ctd_zxz_o0_t0_l0, rk3Ctd_zxz, rk3m_Ctd_zxx_o0_t1_l0, rk3Ctd_zxx, rk3m_Ctd_zyy_o0_t1_l0, rk3Ctd_zyy, rk3m_Ctd_zyz_o0_t0_l0, rk3Ctd_zyz, rk3Tf4u_tt, rk3Betatd_x, rk3Betatd_z, rk3Betatd_y, rk3Ct_zyz, rk3Ct_zyy, rk3Atud_zx, rk3Atud_zz, rk3Atud_zy, rk3Atu_zz, rk3Tf4u_tz, rk3Tu_tz, rk3Atud_yy, rk3Atud_yx, rk3Atud_yz, rk3Atu_yz, rk3Atu_yy, rk3Ct_yyy, rk3Ct_yyz, rk3Atud_xx, rk3trAt, rk3Atud_xz, rk3Atud_xy, rk3Atu_xz, rk3Atu_xy, rk3Atu_xx, rk3Tf4u_ty, rk3Tu_ty, rk3Tf4u_tx, rk3Tu_tx, rk3decay_factor, rk3kappa_z1, rk3kappa_f, rk3feta, rk3m_Ctd_zzz_o0_t0_l0, rk3Ctd_zzz, rk3Ct_xzz, rk3Ct_zzz, rk3Ct_yzz, rk3m_Ctd_yxy_o0_t0_l0, rk3Ctd_yxy, rk3Ct_xyz, rk3m_Ctd_yxz_o0_t2_l0, rk3m_Ctd_yxz_o0_t0_l0, rk3Ctd_yxz, rk3Ct_zxz, rk3Ct_yxz, rk3m_Ctd_zxy_o0_t2_l0, rk3Ctd_zxy, rk3Ct_zxy, rk3Ct_yxy, rk3Ct_xxy, rk3Ct_xxz, rk3Ct_xyy, rk3m_Ctd_yxx_o0_t1_l0, rk3Ctd_yxx, rk3Ct_zxx, rk3Gamt_z, rk3Zu_z, rk3Ct_yxx, rk3Gamt_y, rk3Zu_y, rk3Ct_xxx, rk3Gamt_x, rk3Zu_x, rk3d_Rtd_xx_o0_t9_m0_l0, rk3m_Rtd_xx_o0_t8_l0, rk3m_Rtd_xx_o0_t7_l0, rk3m_Rtd_xx_o0_t6_l0, rk3m_Rtd_xx_o0_t5_l0, rk3m_Rtd_xx_o0_t4_l0, rk3m_Rtd_xx_o0_t3_l0, rk3m_Rtd_xx_o0_t2_l0, rk3m_Rtd_xx_o0_t1_l0, rk3m_Rtd_xx_o0_t0_l0, rk3Rtd_xx, rk3d_Rtd_xz_o0_t12_m0_l0, rk3m_Rtd_xz_o0_t11_l0, rk3m_Rtd_xz_o0_t10_l0, rk3m_Rtd_xz_o0_t9_l0, rk3m_Rtd_xz_o0_t8_l0, rk3m_Rtd_xz_o0_t7_l0, rk3m_Rtd_xz_o0_t6_l0, rk3m_Rtd_xz_o0_t5_l0, rk3m_Rtd_xz_o0_t4_l0, rk3m_Rtd_xz_o0_t3_l0, rk3m_Rtd_xz_o0_t2_l0, rk3m_Rtd_xz_o0_t1_l0, rk3m_Rtd_xz_o0_t0_l0, rk3Rtd_xz, rk3d_Rtd_zz_o0_t9_m0_l0, rk3m_Rtd_zz_o0_t8_l0, rk3m_Rtd_zz_o0_t7_l0, rk3m_Rtd_zz_o0_t6_l0, rk3m_Rtd_zz_o0_t5_l0, rk3m_Rtd_zz_o0_t4_l0, rk3m_Rtd_zz_o0_t3_l0, rk3m_Rtd_zz_o0_t2_l0, rk3m_Rtd_zz_o0_t1_l0, rk3m_Rtd_zz_o0_t0_l0, rk3Rtd_zz, rk3d_Rtd_xy_o0_t12_m0_l0, rk3m_Rtd_xy_o0_t11_l0, rk3m_Rtd_xy_o0_t10_l0, rk3m_Rtd_xy_o0_t9_l0, rk3m_Rtd_xy_o0_t8_l0, rk3m_Rtd_xy_o0_t7_l0, rk3m_Rtd_xy_o0_t6_l0, rk3m_Rtd_xy_o0_t5_l0, rk3m_Rtd_xy_o0_t4_l0, rk3m_Rtd_xy_o0_t3_l0, rk3m_Rtd_xy_o0_t2_l0, rk3m_Rtd_xy_o0_t1_l0, rk3m_Rtd_xy_o0_t0_l0, rk3Rtd_xy, rk3d_Rtd_yy_o0_t9_m0_l0, rk3m_Rtd_yy_o0_t8_l0, rk3m_Rtd_yy_o0_t7_l0, rk3m_Rtd_yy_o0_t6_l0, rk3m_Rtd_yy_o0_t5_l0, rk3m_Rtd_yy_o0_t4_l0, rk3m_Rtd_yy_o0_t3_l0, rk3m_Rtd_yy_o0_t2_l0, rk3m_Rtd_yy_o0_t1_l0, rk3m_Rtd_yy_o0_t0_l0, rk3Rtd_yy, rk3d_Rtd_yz_o0_t12_m0_l0, rk3m_Rtd_yz_o0_t11_l0, rk3m_Rtd_yz_o0_t10_l0, rk3m_Rtd_yz_o0_t9_l0, rk3m_Rtd_yz_o0_t8_l0, rk3m_Rtd_yz_o0_t7_l0, rk3m_Rtd_yz_o0_t6_l0, rk3m_Rtd_yz_o0_t5_l0, rk3m_Rtd_yz_o0_t4_l0, rk3m_Rtd_yz_o0_t3_l0, rk3m_Rtd_yz_o0_t2_l0, rk3m_Rtd_yz_o0_t1_l0, rk3m_Rtd_yz_o0_t0_l0, rk3Rtd_yz, rk3m_Rpd_xx_o0_t22_l0, rk3m_Rpd_xx_o0_t21_l0, rk3m_Rpd_xx_o0_t20_l0, rk3m_Rpd_xx_o0_t19_l0, rk3m_Rpd_xx_o0_t18_l0, rk3m_Rpd_xx_o0_t17_l0, rk3m_Rpd_xx_o0_t16_l0, rk3m_Rpd_xx_o0_t15_l0, rk3m_Rpd_xx_o0_t14_l0, rk3m_Rpd_xx_o0_t13_l0, rk3m_Rpd_xx_o0_t12_l0, rk3m_Rpd_xx_o0_t11_l0, rk3m_Rpd_xx_o0_t10_l0, rk3m_Rpd_xx_o0_t9_l0, rk3m_Rpd_xx_o0_t8_l0, rk3m_Rpd_xx_o0_t7_l0, rk3m_Rpd_xx_o0_t6_l0, rk3m_Rpd_xx_o0_t5_l0, rk3m_Rpd_xx_o0_t4_l0, rk3m_Rpd_xx_o0_t3_l0, rk3m_Rpd_xx_o0_t2_l0, rk3m_Rpd_xx_o0_t1_l0, rk3m_Rpd_xx_o0_t0_l0, rk3Rpd_xx, rk3m_Rpd_yy_o0_t22_l0, rk3m_Rpd_yy_o0_t21_l0, rk3m_Rpd_yy_o0_t20_l0, rk3m_Rpd_yy_o0_t19_l0, rk3m_Rpd_yy_o0_t18_l0, rk3m_Rpd_yy_o0_t17_l0, rk3m_Rpd_yy_o0_t16_l0, rk3m_Rpd_yy_o0_t15_l0, rk3m_Rpd_yy_o0_t14_l0, rk3m_Rpd_yy_o0_t13_l0, rk3m_Rpd_yy_o0_t12_l0, rk3m_Rpd_yy_o0_t11_l0, rk3m_Rpd_yy_o0_t10_l0, rk3m_Rpd_yy_o0_t9_l0, rk3m_Rpd_yy_o0_t8_l0, rk3m_Rpd_yy_o0_t7_l0, rk3m_Rpd_yy_o0_t6_l0, rk3m_Rpd_yy_o0_t5_l0, rk3m_Rpd_yy_o0_t4_l0, rk3m_Rpd_yy_o0_t3_l0, rk3m_Rpd_yy_o0_t2_l0, rk3m_Rpd_yy_o0_t1_l0, rk3m_Rpd_yy_o0_t0_l0, rk3Rpd_yy, rk3m_Rpd_yz_o0_t25_l0, rk3m_Rpd_yz_o0_t24_l0, rk3m_Rpd_yz_o0_t23_l0, rk3m_Rpd_yz_o0_t22_l0, rk3m_Rpd_yz_o0_t21_l0, rk3m_Rpd_yz_o0_t20_l0, rk3m_Rpd_yz_o0_t19_l0, rk3m_Rpd_yz_o0_t18_l0, rk3m_Rpd_yz_o0_t17_l0, rk3m_Rpd_yz_o0_t16_l0, rk3m_Rpd_yz_o0_t15_l0, rk3m_Rpd_yz_o0_t14_l0, rk3m_Rpd_yz_o0_t13_l0, rk3m_Rpd_yz_o0_t12_l0, rk3m_Rpd_yz_o0_t11_l0, rk3m_Rpd_yz_o0_t10_l0, rk3m_Rpd_yz_o0_t9_l0, rk3m_Rpd_yz_o0_t8_l0, rk3m_Rpd_yz_o0_t7_l0, rk3m_Rpd_yz_o0_t6_l0, rk3m_Rpd_yz_o0_t5_l0, rk3m_Rpd_yz_o0_t4_l0, rk3m_Rpd_yz_o0_t3_l0, rk3m_Rpd_yz_o0_t2_l0, rk3m_Rpd_yz_o0_t1_l0, rk3m_Rpd_yz_o0_t0_l0, rk3Rpd_yz, rk3m_Rpd_xz_o0_t25_l0, rk3m_Rpd_xz_o0_t24_l0, rk3m_Rpd_xz_o0_t23_l0, rk3m_Rpd_xz_o0_t22_l0, rk3m_Rpd_xz_o0_t21_l0, rk3m_Rpd_xz_o0_t20_l0, rk3m_Rpd_xz_o0_t19_l0, rk3m_Rpd_xz_o0_t18_l0, rk3m_Rpd_xz_o0_t17_l0, rk3m_Rpd_xz_o0_t16_l0, rk3m_Rpd_xz_o0_t15_l0, rk3m_Rpd_xz_o0_t14_l0, rk3m_Rpd_xz_o0_t13_l0, rk3m_Rpd_xz_o0_t12_l0, rk3m_Rpd_xz_o0_t11_l0, rk3m_Rpd_xz_o0_t10_l0, rk3m_Rpd_xz_o0_t9_l0, rk3m_Rpd_xz_o0_t8_l0, rk3m_Rpd_xz_o0_t7_l0, rk3m_Rpd_xz_o0_t6_l0, rk3m_Rpd_xz_o0_t5_l0, rk3m_Rpd_xz_o0_t4_l0, rk3m_Rpd_xz_o0_t3_l0, rk3m_Rpd_xz_o0_t2_l0, rk3m_Rpd_xz_o0_t1_l0, rk3m_Rpd_xz_o0_t0_l0, rk3Rpd_xz, rk3m_Rpd_xy_o0_t25_l0, rk3m_Rpd_xy_o0_t24_l0, rk3m_Rpd_xy_o0_t23_l0, rk3m_Rpd_xy_o0_t22_l0, rk3m_Rpd_xy_o0_t21_l0, rk3m_Rpd_xy_o0_t20_l0, rk3m_Rpd_xy_o0_t19_l0, rk3m_Rpd_xy_o0_t18_l0, rk3m_Rpd_xy_o0_t17_l0, rk3m_Rpd_xy_o0_t16_l0, rk3m_Rpd_xy_o0_t15_l0, rk3m_Rpd_xy_o0_t14_l0, rk3m_Rpd_xy_o0_t13_l0, rk3m_Rpd_xy_o0_t12_l0, rk3m_Rpd_xy_o0_t11_l0, rk3m_Rpd_xy_o0_t10_l0, rk3m_Rpd_xy_o0_t9_l0, rk3m_Rpd_xy_o0_t8_l0, rk3m_Rpd_xy_o0_t7_l0, rk3m_Rpd_xy_o0_t6_l0, rk3m_Rpd_xy_o0_t5_l0, rk3m_Rpd_xy_o0_t4_l0, rk3m_Rpd_xy_o0_t3_l0, rk3m_Rpd_xy_o0_t2_l0, rk3m_Rpd_xy_o0_t1_l0, rk3m_Rpd_xy_o0_t0_l0, rk3Rpd_xy, rk3m_Rpd_zz_o0_t22_l0, rk3m_Rpd_zz_o0_t21_l0, rk3m_Rpd_zz_o0_t20_l0, rk3m_Rpd_zz_o0_t19_l0, rk3m_Rpd_zz_o0_t18_l0, rk3m_Rpd_zz_o0_t17_l0, rk3m_Rpd_zz_o0_t16_l0, rk3m_Rpd_zz_o0_t15_l0, rk3m_Rpd_zz_o0_t14_l0, rk3m_Rpd_zz_o0_t13_l0, rk3m_Rpd_zz_o0_t12_l0, rk3m_Rpd_zz_o0_t11_l0, rk3m_Rpd_zz_o0_t10_l0, rk3m_Rpd_zz_o0_t9_l0, rk3m_Rpd_zz_o0_t8_l0, rk3m_Rpd_zz_o0_t7_l0, rk3m_Rpd_zz_o0_t6_l0, rk3m_Rpd_zz_o0_t5_l0, rk3m_Rpd_zz_o0_t4_l0, rk3m_Rpd_zz_o0_t3_l0, rk3m_Rpd_zz_o0_t2_l0, rk3m_Rpd_zz_o0_t1_l0, rk3m_Rpd_zz_o0_t0_l0, rk3Rpd_zz, rk3Rscalar, rk3Tf4u_zz, rk3Tu_zz, rk3Tf4u_yz, rk3Tu_yz, rk3Tf4u_yy, rk3Tu_yy, rk3Tf4u_xy, rk3Tu_xy, rk3Tf4u_xx, rk3Tu_xx, rk3Tf4u_xz, rk3Tu_xz, rk3trSf, rk3Tu_tt, rk3pTtd_ADM_xz, rk3pTtd_ADM_xy, rk3pTtd_ADM_xx, rk3d_Psi1_xx_o0_t5_m0_l0, rk3m_Psi1_xx_o0_t4_l0, rk3m_Psi1_xx_o0_t3_l0, rk3m_Psi1_xx_o0_t2_l0, rk3m_Psi1_xx_o0_t1_l0, rk3m_Psi1_xx_o0_t0_l0, rk3Psi1_xx, rk3d_Psi1_xy_o0_t6_m0_l0, rk3m_Psi1_xy_o0_t5_l0, rk3m_Psi1_xy_o0_t4_l0, rk3m_Psi1_xy_o0_t3_l0, rk3m_Psi1_xy_o0_t2_l0, rk3m_Psi1_xy_o0_t1_l0, rk3m_Psi1_xy_o0_t0_l0, rk3Psi1_xy, rk3d_Psi1_xz_o0_t6_m0_l0, rk3m_Psi1_xz_o0_t5_l0, rk3m_Psi1_xz_o0_t4_l0, rk3m_Psi1_xz_o0_t3_l0, rk3m_Psi1_xz_o0_t2_l0, rk3m_Psi1_xz_o0_t1_l0, rk3m_Psi1_xz_o0_t0_l0, rk3Psi1_xz, rk3Jtd_ADM_z, rk3pTtd_ADM_zz, rk3d_Psi1_zz_o0_t5_m0_l0, rk3m_Psi1_zz_o0_t4_l0, rk3m_Psi1_zz_o0_t3_l0, rk3m_Psi1_zz_o0_t2_l0, rk3m_Psi1_zz_o0_t1_l0, rk3m_Psi1_zz_o0_t0_l0, rk3Psi1_zz, rk3Jtd_ADM_y, rk3Jtd_ADM_x, rk3pTtd_ADM_yz, rk3d_Psi1_yz_o0_t6_m0_l0, rk3m_Psi1_yz_o0_t5_l0, rk3m_Psi1_yz_o0_t4_l0, rk3m_Psi1_yz_o0_t3_l0, rk3m_Psi1_yz_o0_t2_l0, rk3m_Psi1_yz_o0_t1_l0, rk3m_Psi1_yz_o0_t0_l0, rk3Psi1_yz, rk3pTtd_ADM_yy, rk3tr_pT, rk3d_Psi1_yy_o0_t5_m0_l0, rk3m_Psi1_yy_o0_t4_l0, rk3m_Psi1_yy_o0_t3_l0, rk3m_Psi1_yy_o0_t2_l0, rk3m_Psi1_yy_o0_t1_l0, rk3m_Psi1_yy_o0_t0_l0, rk3Psi1_yy, rk3trPsi1, rk3Psi1TF_yz, rk3Psi1TF_yy, rk3Psi1TF_xy, rk3Psi1TF_xx, rk3Psi1TF_xz, rk3Psi1TF_zz, rk3kappa_cc, rk3rho_ADM, d_Sfd_x_o0_t11_m0_l0, d_Sfd_y_o0_t11_m0_l0, d_Sfd_z_o0_t11_m0_l0, d_DYf_o0_t0_m0_l0, d_tauf_o0_t3_m0_l0, d_gtd_xx_o0_t3_m0_l0, d_gtd_xy_o0_t6_m0_l0, d_gtd_xz_o0_t6_m0_l0, d_gtd_yy_o0_t3_m0_l0, d_gtd_yz_o0_t6_m0_l0, d_gtd_zz_o0_t3_m0_l0, d_Atd_xx_o0_t3_m0_l0, d_Atd_xy_o0_t6_m0_l0, d_Atd_xz_o0_t6_m0_l0, d_Atd_yy_o0_t3_m0_l0, d_Atd_yz_o0_t6_m0_l0, d_Atd_zz_o0_t3_m0_l0, d_Gamh_x_o0_t24_m0_l0, d_Gamh_y_o0_t24_m0_l0, d_Gamh_z_o0_t24_m0_l0, d_Betau_x_o0_t0_m0_l0, d_Betau_y_o0_t0_m0_l0, d_Betau_z_o0_t0_m0_l0, d_Alpha_o0_t0_m0_l0, d_chi_o0_t0_m0_l0, d_trK_o0_t21_m0_l0, d_theta_o0_t3_m0_l0, RHS_Bfu_x, RHS_Bfu_y, RHS_Bfu_z, RHS_tauf, RHS_phif, RHS_Sfd_x, RHS_Sfd_y, RHS_Sfd_z, RHS_Df, RHS_DYf, m_theta_o2_t2_l0, m_theta_o2_t1_l0, m_theta_o2_t0_l0, m_theta_o1_t2_l0, m_theta_o1_t1_l0, m_theta_o1_t0_l0, m_theta_o0_t2_l0, m_theta_o0_t1_l0, m_theta_o0_t0_l0, m_trK_o2_t2_l0, m_trK_o2_t1_l0, m_trK_o2_t0_l0, m_trK_o1_t2_l0, m_trK_o1_t1_l0, m_trK_o1_t0_l0, m_trK_o0_t20_l0, m_trK_o0_t19_l0, m_trK_o0_t18_l0, m_trK_o0_t17_l0, m_trK_o0_t16_l0, m_trK_o0_t15_l0, m_trK_o0_t14_l0, m_trK_o0_t13_l0, m_trK_o0_t12_l0, m_trK_o0_t11_l0, m_trK_o0_t10_l0, m_trK_o0_t9_l0, m_trK_o0_t8_l0, m_trK_o0_t7_l0, m_trK_o0_t6_l0, m_trK_o0_t5_l0, m_trK_o0_t4_l0, m_trK_o0_t3_l0, m_trK_o0_t2_l0, m_trK_o0_t1_l0, m_trK_o0_t0_l0, m_chi_o2_t2_l0, m_chi_o2_t1_l0, m_chi_o2_t0_l0, m_chi_o1_t2_l0, m_chi_o1_t1_l0, m_chi_o1_t0_l0, m_Alpha_o2_t2_l0, m_Alpha_o2_t1_l0, m_Alpha_o2_t0_l0, m_Alpha_o1_t2_l0, m_Alpha_o1_t1_l0, m_Alpha_o1_t0_l0, m_Betau_z_o2_t2_l0, m_Betau_z_o2_t1_l0, m_Betau_z_o2_t0_l0, m_Betau_z_o1_t2_l0, m_Betau_z_o1_t1_l0, m_Betau_z_o1_t0_l0, m_Betau_y_o2_t2_l0, m_Betau_y_o2_t1_l0, m_Betau_y_o2_t0_l0, m_Betau_y_o1_t2_l0, m_Betau_y_o1_t1_l0, m_Betau_y_o1_t0_l0, m_Betau_x_o2_t2_l0, m_Betau_x_o2_t1_l0, m_Betau_x_o2_t0_l0, m_Betau_x_o1_t2_l0, m_Betau_x_o1_t1_l0, m_Betau_x_o1_t0_l0, m_Gamh_z_o2_t2_l0, m_Gamh_z_o2_t1_l0, m_Gamh_z_o2_t0_l0, m_Gamh_z_o1_t2_l0, m_Gamh_z_o1_t1_l0, m_Gamh_z_o1_t0_l0, m_Gamh_z_o0_t23_l0, m_Gamh_z_o0_t22_l0, m_Gamh_z_o0_t21_l0, m_Gamh_z_o0_t20_l0, m_Gamh_z_o0_t19_l0, m_Gamh_z_o0_t18_l0, m_Gamh_z_o0_t17_l0, m_Gamh_z_o0_t16_l0, m_Gamh_z_o0_t15_l0, m_Gamh_z_o0_t14_l0, m_Gamh_z_o0_t13_l0, m_Gamh_z_o0_t12_l0, m_Gamh_z_o0_t11_l0, m_Gamh_z_o0_t10_l0, m_Gamh_z_o0_t9_l0, m_Gamh_z_o0_t8_l0, m_Gamh_z_o0_t7_l0, m_Gamh_z_o0_t6_l0, m_Gamh_z_o0_t5_l0, m_Gamh_z_o0_t4_l0, m_Gamh_z_o0_t3_l0, m_Gamh_z_o0_t2_l0, m_Gamh_z_o0_t1_l0, m_Gamh_z_o0_t0_l0, m_Gamh_y_o2_t2_l0, m_Gamh_y_o2_t1_l0, m_Gamh_y_o2_t0_l0, m_Gamh_y_o1_t2_l0, m_Gamh_y_o1_t1_l0, m_Gamh_y_o1_t0_l0, m_Gamh_y_o0_t23_l0, m_Gamh_y_o0_t22_l0, m_Gamh_y_o0_t21_l0, m_Gamh_y_o0_t20_l0, m_Gamh_y_o0_t19_l0, m_Gamh_y_o0_t18_l0, m_Gamh_y_o0_t17_l0, m_Gamh_y_o0_t16_l0, m_Gamh_y_o0_t15_l0, m_Gamh_y_o0_t14_l0, m_Gamh_y_o0_t13_l0, m_Gamh_y_o0_t12_l0, m_Gamh_y_o0_t11_l0, m_Gamh_y_o0_t10_l0, m_Gamh_y_o0_t9_l0, m_Gamh_y_o0_t8_l0, m_Gamh_y_o0_t7_l0, m_Gamh_y_o0_t6_l0, m_Gamh_y_o0_t5_l0, m_Gamh_y_o0_t4_l0, m_Gamh_y_o0_t3_l0, m_Gamh_y_o0_t2_l0, m_Gamh_y_o0_t1_l0, m_Gamh_y_o0_t0_l0, m_Gamh_x_o2_t2_l0, m_Gamh_x_o2_t1_l0, m_Gamh_x_o2_t0_l0, m_Gamh_x_o1_t2_l0, m_Gamh_x_o1_t1_l0, m_Gamh_x_o1_t0_l0, m_Gamh_x_o0_t23_l0, m_Gamh_x_o0_t22_l0, m_Gamh_x_o0_t21_l0, m_Gamh_x_o0_t20_l0, m_Gamh_x_o0_t19_l0, m_Gamh_x_o0_t18_l0, m_Gamh_x_o0_t17_l0, m_Gamh_x_o0_t16_l0, m_Gamh_x_o0_t15_l0, m_Gamh_x_o0_t14_l0, m_Gamh_x_o0_t13_l0, m_Gamh_x_o0_t12_l0, m_Gamh_x_o0_t11_l0, m_Gamh_x_o0_t10_l0, m_Gamh_x_o0_t9_l0, m_Gamh_x_o0_t8_l0, m_Gamh_x_o0_t7_l0, m_Gamh_x_o0_t6_l0, m_Gamh_x_o0_t5_l0, m_Gamh_x_o0_t4_l0, m_Gamh_x_o0_t3_l0, m_Gamh_x_o0_t2_l0, m_Gamh_x_o0_t1_l0, m_Gamh_x_o0_t0_l0, m_Atd_zz_o2_t2_l0, m_Atd_zz_o2_t1_l0, m_Atd_zz_o2_t0_l0, m_Atd_zz_o1_t2_l0, m_Atd_zz_o1_t1_l0, m_Atd_zz_o1_t0_l0, m_Atd_zz_o0_t2_l0, m_Atd_zz_o0_t1_l0, m_Atd_zz_o0_t0_l0, m_Atd_yz_o2_t2_l0, m_Atd_yz_o2_t1_l0, m_Atd_yz_o2_t0_l0, m_Atd_yz_o1_t2_l0, m_Atd_yz_o1_t1_l0, m_Atd_yz_o1_t0_l0, m_Atd_yz_o0_t5_l0, m_Atd_yz_o0_t4_l0, m_Atd_yz_o0_t3_l0, m_Atd_yz_o0_t2_l0, m_Atd_yz_o0_t1_l0, m_Atd_yz_o0_t0_l0, m_Atd_yy_o2_t2_l0, m_Atd_yy_o2_t1_l0, m_Atd_yy_o2_t0_l0, m_Atd_yy_o1_t2_l0, m_Atd_yy_o1_t1_l0, m_Atd_yy_o1_t0_l0, m_Atd_yy_o0_t2_l0, m_Atd_yy_o0_t1_l0, m_Atd_yy_o0_t0_l0, m_Atd_xz_o2_t2_l0, m_Atd_xz_o2_t1_l0, m_Atd_xz_o2_t0_l0, m_Atd_xz_o1_t2_l0, m_Atd_xz_o1_t1_l0, m_Atd_xz_o1_t0_l0, m_Atd_xz_o0_t5_l0, m_Atd_xz_o0_t4_l0, m_Atd_xz_o0_t3_l0, m_Atd_xz_o0_t2_l0, m_Atd_xz_o0_t1_l0, m_Atd_xz_o0_t0_l0, m_Atd_xy_o2_t2_l0, m_Atd_xy_o2_t1_l0, m_Atd_xy_o2_t0_l0, m_Atd_xy_o1_t2_l0, m_Atd_xy_o1_t1_l0, m_Atd_xy_o1_t0_l0, m_Atd_xy_o0_t5_l0, m_Atd_xy_o0_t4_l0, m_Atd_xy_o0_t3_l0, m_Atd_xy_o0_t2_l0, m_Atd_xy_o0_t1_l0, m_Atd_xy_o0_t0_l0, m_Atd_xx_o2_t2_l0, m_Atd_xx_o2_t1_l0, m_Atd_xx_o2_t0_l0, m_Atd_xx_o1_t2_l0, m_Atd_xx_o1_t1_l0, m_Atd_xx_o1_t0_l0, m_Atd_xx_o0_t2_l0, m_Atd_xx_o0_t1_l0, m_Atd_xx_o0_t0_l0, m_gtd_zz_o2_t2_l0, m_gtd_zz_o2_t1_l0, m_gtd_zz_o2_t0_l0, m_gtd_zz_o1_t2_l0, m_gtd_zz_o1_t1_l0, m_gtd_zz_o1_t0_l0, m_gtd_zz_o0_t2_l0, m_gtd_zz_o0_t1_l0, m_gtd_zz_o0_t0_l0, m_gtd_yz_o2_t2_l0, m_gtd_yz_o2_t1_l0, m_gtd_yz_o2_t0_l0, m_gtd_yz_o1_t2_l0, m_gtd_yz_o1_t1_l0, m_gtd_yz_o1_t0_l0, m_gtd_yz_o0_t5_l0, m_gtd_yz_o0_t4_l0, m_gtd_yz_o0_t3_l0, m_gtd_yz_o0_t2_l0, m_gtd_yz_o0_t1_l0, m_gtd_yz_o0_t0_l0, m_gtd_yy_o2_t2_l0, m_gtd_yy_o2_t1_l0, m_gtd_yy_o2_t0_l0, m_gtd_yy_o1_t2_l0, m_gtd_yy_o1_t1_l0, m_gtd_yy_o1_t0_l0, m_gtd_yy_o0_t2_l0, m_gtd_yy_o0_t1_l0, m_gtd_yy_o0_t0_l0, m_gtd_xz_o2_t2_l0, m_gtd_xz_o2_t1_l0, m_gtd_xz_o2_t0_l0, m_gtd_xz_o1_t2_l0, m_gtd_xz_o1_t1_l0, m_gtd_xz_o1_t0_l0, m_gtd_xz_o0_t5_l0, m_gtd_xz_o0_t4_l0, m_gtd_xz_o0_t3_l0, m_gtd_xz_o0_t2_l0, m_gtd_xz_o0_t1_l0, m_gtd_xz_o0_t0_l0, m_gtd_xy_o2_t2_l0, m_gtd_xy_o2_t1_l0, m_gtd_xy_o2_t0_l0, m_gtd_xy_o1_t2_l0, m_gtd_xy_o1_t1_l0, m_gtd_xy_o1_t0_l0, m_gtd_xy_o0_t5_l0, m_gtd_xy_o0_t4_l0, m_gtd_xy_o0_t3_l0, m_gtd_xy_o0_t2_l0, m_gtd_xy_o0_t1_l0, m_gtd_xy_o0_t0_l0, m_gtd_xx_o2_t2_l0, m_gtd_xx_o2_t1_l0, m_gtd_xx_o2_t0_l0, m_gtd_xx_o1_t2_l0, m_gtd_xx_o1_t1_l0, m_gtd_xx_o1_t0_l0, m_gtd_xx_o0_t2_l0, m_gtd_xx_o0_t1_l0, m_gtd_xx_o0_t0_l0, m_phif_o0_t2_l0, m_phif_o0_t1_l0, m_phif_o0_t0_l0, m_tauf_o0_t2_l0, m_tauf_o0_t1_l0, m_tauf_o0_t0_l0, m_Bfu_z_o0_t5_l0, m_Bfu_z_o0_t4_l0, m_Bfu_z_o0_t3_l0, m_Bfu_z_o0_t2_l0, m_Bfu_z_o0_t1_l0, m_Bfu_z_o0_t0_l0, m_Bfu_y_o0_t5_l0, m_Bfu_y_o0_t4_l0, m_Bfu_y_o0_t3_l0, m_Bfu_y_o0_t2_l0, m_Bfu_y_o0_t1_l0, m_Bfu_y_o0_t0_l0, m_Bfu_x_o0_t5_l0, m_Bfu_x_o0_t4_l0, m_Bfu_x_o0_t3_l0, m_Bfu_x_o0_t2_l0, m_Bfu_x_o0_t1_l0, m_Bfu_x_o0_t0_l0, m_Sfd_z_o0_t10_l0, m_Sfd_z_o0_t9_l0, m_Sfd_z_o0_t8_l0, m_Sfd_z_o0_t7_l0, m_Sfd_z_o0_t6_l0, m_Sfd_z_o0_t5_l0, m_Sfd_z_o0_t4_l0, m_Sfd_z_o0_t3_l0, m_Sfd_z_o0_t2_l0, m_Sfd_z_o0_t1_l0, m_Sfd_z_o0_t0_l0, m_Sfd_y_o0_t10_l0, m_Sfd_y_o0_t9_l0, m_Sfd_y_o0_t8_l0, m_Sfd_y_o0_t7_l0, m_Sfd_y_o0_t6_l0, m_Sfd_y_o0_t5_l0, m_Sfd_y_o0_t4_l0, m_Sfd_y_o0_t3_l0, m_Sfd_y_o0_t2_l0, m_Sfd_y_o0_t1_l0, m_Sfd_y_o0_t0_l0, m_Sfd_x_o0_t10_l0, m_Sfd_x_o0_t9_l0, m_Sfd_x_o0_t8_l0, m_Sfd_x_o0_t7_l0, m_Sfd_x_o0_t6_l0, m_Sfd_x_o0_t5_l0, m_Sfd_x_o0_t4_l0, m_Sfd_x_o0_t3_l0, m_Sfd_x_o0_t2_l0, m_Sfd_x_o0_t1_l0, m_Sfd_x_o0_t0_l0, RHS_gtd_xx, RHS_gtd_xy, RHS_gtd_xz, RHS_gtd_yy, RHS_gtd_yz, RHS_gtd_zz, RHS_Atd_xx, RHS_Atd_xy, RHS_Atd_xz, RHS_Atd_yy, RHS_Atd_yz, RHS_Atd_zz, RHS_Gamh_x, RHS_Gamh_y, RHS_Gamh_z, RHS_Betau_x, RHS_Betau_y, RHS_Betau_z, RHS_Alpha, RHS_chi, RHS_trK, RHS_theta, n_x, n_y, n_z, interaction_index, mod_normal, i_d_Atd_xx_o0_t0_m0_l0, i_d_Atd_xx_o0_t1_m0_l0, i_d_Atd_xx_o0_t2_m0_l0, i_d_Atd_xy_o0_t0_m0_l0, i_d_Atd_xy_o0_t1_m0_l0, i_d_Atd_xy_o0_t2_m0_l0, i_d_Atd_xz_o0_t0_m0_l0, i_d_Atd_xz_o0_t1_m0_l0, i_d_Atd_xz_o0_t2_m0_l0, i_d_Atd_yy_o0_t0_m0_l0, i_d_Atd_yy_o0_t1_m0_l0, i_d_Atd_yy_o0_t2_m0_l0, i_d_Atd_yz_o0_t0_m0_l0, i_d_Atd_yz_o0_t1_m0_l0, i_d_Atd_yz_o0_t2_m0_l0, i_d_Atd_zz_o0_t0_m0_l0, i_d_Atd_zz_o0_t1_m0_l0, i_d_Atd_zz_o0_t2_m0_l0, i_d_gtd_xx_o0_t3_m0_l0, i_d_gtd_xy_o0_t3_m0_l0, i_d_gtd_xz_o0_t3_m0_l0, i_d_gtd_yy_o0_t3_m0_l0, i_d_gtd_yz_o0_t3_m0_l0, i_d_gtd_zz_o0_t3_m0_l0, i_d_Atd_xx_o0_t3_m0_l0, i_d_Atd_xy_o0_t3_m0_l0, i_d_Atd_xz_o0_t3_m0_l0, i_d_Atd_yy_o0_t3_m0_l0, i_d_Atd_yz_o0_t3_m0_l0, i_d_Atd_zz_o0_t3_m0_l0, i_d_Gamh_x_o0_t3_m0_l0, i_d_Gamh_y_o0_t3_m0_l0, i_d_Gamh_z_o0_t3_m0_l0, i_d_Betau_x_o0_t3_m0_l0, i_d_Betau_y_o0_t3_m0_l0, i_d_Betau_z_o0_t3_m0_l0, i_d_Alpha_o0_t3_m0_l0, i_d_chi_o0_t3_m0_l0, i_d_trK_o0_t3_m0_l0, i_d_theta_o0_t3_m0_l0, i_m_gtd_xx_o0_t0_l0, i_m_gtd_xx_o0_t1_l0, i_m_gtd_xx_o0_t2_l0, i_m_gtd_xy_o0_t0_l0, i_m_gtd_xy_o0_t1_l0, i_m_gtd_xy_o0_t2_l0, i_m_gtd_xz_o0_t0_l0, i_m_gtd_xz_o0_t1_l0, i_m_gtd_xz_o0_t2_l0, i_m_gtd_yy_o0_t0_l0, i_m_gtd_yy_o0_t1_l0, i_m_gtd_yy_o0_t2_l0, i_m_gtd_yz_o0_t0_l0, i_m_gtd_yz_o0_t1_l0, i_m_gtd_yz_o0_t2_l0, i_m_gtd_zz_o0_t0_l0, i_m_gtd_zz_o0_t1_l0, i_m_gtd_zz_o0_t2_l0, i_m_Atd_xx_o0_t0_l0, i_m_Atd_xx_o0_t1_l0, i_m_Atd_xx_o0_t2_l0, i_m_Atd_xy_o0_t0_l0, i_m_Atd_xy_o0_t1_l0, i_m_Atd_xy_o0_t2_l0, i_m_Atd_xz_o0_t0_l0, i_m_Atd_xz_o0_t1_l0, i_m_Atd_xz_o0_t2_l0, i_m_Atd_yy_o0_t0_l0, i_m_Atd_yy_o0_t1_l0, i_m_Atd_yy_o0_t2_l0, i_m_Atd_yz_o0_t0_l0, i_m_Atd_yz_o0_t1_l0, i_m_Atd_yz_o0_t2_l0, i_m_Atd_zz_o0_t0_l0, i_m_Atd_zz_o0_t1_l0, i_m_Atd_zz_o0_t2_l0, i_m_Gamh_x_o0_t0_l0, i_m_Gamh_x_o0_t1_l0, i_m_Gamh_x_o0_t2_l0, i_m_Gamh_y_o0_t0_l0, i_m_Gamh_y_o0_t1_l0, i_m_Gamh_y_o0_t2_l0, i_m_Gamh_z_o0_t0_l0, i_m_Gamh_z_o0_t1_l0, i_m_Gamh_z_o0_t2_l0, i_m_Betau_x_o0_t0_l0, i_m_Betau_x_o0_t1_l0, i_m_Betau_x_o0_t2_l0, i_m_Betau_y_o0_t0_l0, i_m_Betau_y_o0_t1_l0, i_m_Betau_y_o0_t2_l0, i_m_Betau_z_o0_t0_l0, i_m_Betau_z_o0_t1_l0, i_m_Betau_z_o0_t2_l0, i_m_Alpha_o0_t0_l0, i_m_Alpha_o0_t1_l0, i_m_Alpha_o0_t2_l0, i_m_chi_o0_t0_l0, i_m_chi_o0_t1_l0, i_m_chi_o0_t2_l0, i_m_trK_o0_t0_l0, i_m_trK_o0_t1_l0, i_m_trK_o0_t2_l0, i_m_theta_o0_t0_l0, i_m_theta_o0_t1_l0, i_m_theta_o0_t2_l0;
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					if ((i + 3 < ilast && i - 3 >= 0 && j + 3 < jlast && j - 3 >= 0 && k + 3 < klast && k - 3 >= 0)) {
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_yy, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_zz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_xy, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_xz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3sqBf = vector(rk3Bfu_x, i, j, k) * rk3Bfd_x + vector(rk3Bfu_y, i, j, k) * rk3Bfd_y + vector(rk3Bfu_z, i, j, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) * vector(rk3gtd_xx, i, j, k) + (-vector(rk3gtd_yy, i, j, k) * (vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k))) + (-vector(rk3gtd_zz, i, j, k) * (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k))) + (-(vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k)) * vector(rk3gtd_xx, i, j, k)) + 2.0 * vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yy, i, j, k) - vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) - vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_yz, i, j, k) - vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_zz, i, j, k)) + vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j, k) + rk3gtu_xy * vector(rk3Sfd_y, i, j, k) + rk3gtu_xz * vector(rk3Sfd_z, i, j, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j, k) + rk3gtu_xy * vector(vfd_y, i, j, k) + rk3gtu_xz * vector(vfd_z, i, j, k));
						rk3Bfvf = vector(rk3Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk3Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk3Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_zz, i, j, k) - vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yz, i, j, k)) + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j, k) + rk3gtu_yz * vector(vfd_y, i, j, k) + rk3gtu_zz * vector(vfd_z, i, j, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j, k) + rk3gtu_yy * vector(rk3Sfd_y, i, j, k) + rk3gtu_yz * vector(rk3Sfd_z, i, j, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j, k) + rk3gtu_yz * vector(rk3Sfd_y, i, j, k) + rk3gtu_zz * vector(rk3Sfd_z, i, j, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j, k) + rk3gtu_yy * vector(vfd_y, i, j, k) + rk3gtu_yz * vector(vfd_z, i, j, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j, k) + rk3vfu_y * vector(vfd_y, i, j, k) + rk3vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j, k) * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j, k) * vector(rk3Bfu_x, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k) * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k) * rk3vfu_z + vector(rk3Bfu_z, i, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k) * vector(rk3Bfu_y, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k) * rk3vfu_y + vector(rk3Bfu_y, i, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k) * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k) * rk3vfu_z + vector(rk3Bfu_z, i, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j, k) * vector(rk3Bfu_y, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j, k));
						FluxiSfd_x = FiSfd_x_mainI(vector(rk3Sfd_x, i, j, k), vector(rk3gtd_xx, i, j, k), vector(rk3gtd_xy, i, j, k), vector(rk3gtd_xz, i, j, k), vector(rk3Betau_x, i, j, k), vector(rk3Alpha, i, j, k), vector(TauT_xx, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_xz, i, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xx, dx, simPlat_dt, ilast, jlast);
						FluxiSfd_y = FiSfd_y_mainI(vector(rk3Sfd_y, i, j, k), vector(rk3gtd_xy, i, j, k), vector(rk3gtd_yy, i, j, k), vector(rk3gtd_yz, i, j, k), vector(rk3Betau_x, i, j, k), vector(rk3Alpha, i, j, k), vector(TauT_xx, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_xz, i, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xy, dx, simPlat_dt, ilast, jlast);
						FluxiSfd_z = FiSfd_z_mainI(vector(rk3Sfd_z, i, j, k), vector(rk3gtd_xz, i, j, k), vector(rk3gtd_yz, i, j, k), vector(rk3gtd_zz, i, j, k), vector(rk3Betau_x, i, j, k), vector(rk3Alpha, i, j, k), vector(TauT_xx, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_xz, i, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xz, dx, simPlat_dt, ilast, jlast);
						FluxiBfu_x = FiBfu_x_mainI(vector(rk3phif, i, j, k), vector(rk3Alpha, i, j, k), rk3chi_max, rk3gtu_xx, dx, simPlat_dt, ilast, jlast);
						FluxiBfu_y = FiBfu_y_mainI(vector(rk3Bfu_x, i, j, k), vector(rk3Bfu_y, i, j, k), vector(rk3phif, i, j, k), vector(rk3Betau_x, i, j, k), vector(rk3Betau_y, i, j, k), vector(rk3Alpha, i, j, k), vector(TauM_xy, i, j, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						FluxiBfu_z = FiBfu_z_mainI(vector(rk3Bfu_x, i, j, k), vector(rk3Bfu_z, i, j, k), vector(rk3phif, i, j, k), vector(rk3Betau_x, i, j, k), vector(rk3Betau_z, i, j, k), vector(rk3Alpha, i, j, k), vector(TauM_xz, i, j, k), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxiDf = FiDf_mainI(vector(rk3Df, i, j, k), vector(rk3Betau_x, i, j, k), vector(rk3Alpha, i, j, k), vector(TauN_x, i, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						FluxiDYf = FiDYf_mainI(vector(rk3DYf, i, j, k), vector(rk3Betau_x, i, j, k), vector(rk3Alpha, i, j, k), vector(TauNe_x, i, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxitauf = Fitauf_mainI(vector(rk3Df, i, j, k), vector(rk3tauf, i, j, k), vector(rk3Betau_x, i, j, k), vector(rk3Alpha, i, j, k), vector(TauN_x, i, j, k), rk3sdetg, rk3Sfu_x, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxiphif = Fiphif_mainI(vector(rk3Bfu_x, i, j, k), vector(rk3phif, i, j, k), vector(rk3Betau_x, i, j, k), vector(rk3Alpha, i, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j, k) + rk3sqca - vector(sqcs, i, j, k) * rk3sqca;
						SpeediSfd_x = fabs(fabs(vector(rk3Betau_x, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x))))));
						SpeediSfd_y = fabs(fabs(vector(rk3Betau_x, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x))))));
						SpeediSfd_z = fabs(fabs(vector(rk3Betau_x, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x))))));
						SpeediBfu_x = fabs(fabs(vector(rk3Betau_x, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x))))));
						SpeediBfu_y = fabs(fabs(vector(rk3Betau_x, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x))))));
						SpeediBfu_z = fabs(fabs(vector(rk3Betau_x, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x))))));
						SpeediDf = fabs(fabs(vector(rk3Betau_x, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x))))));
						SpeediDYf = fabs(fabs(vector(rk3Betau_x, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x))))));
						Speeditauf = fabs(fabs(vector(rk3Betau_x, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x))))));
						Speediphif = fabs(fabs(vector(rk3Betau_x, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i - 3, j, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i - 3, j, k) * vector(rk3Bfu_x, i - 3, j, k) + vector(rk3gtd_yy, i - 3, j, k) * vector(rk3Bfu_y, i - 3, j, k) + vector(rk3gtd_yz, i - 3, j, k) * vector(rk3Bfu_z, i - 3, j, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i - 3, j, k) * vector(rk3Bfu_x, i - 3, j, k) + vector(rk3gtd_yz, i - 3, j, k) * vector(rk3Bfu_y, i - 3, j, k) + vector(rk3gtd_zz, i - 3, j, k) * vector(rk3Bfu_z, i - 3, j, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i - 3, j, k) * vector(rk3Bfu_x, i - 3, j, k) + vector(rk3gtd_xy, i - 3, j, k) * vector(rk3Bfu_y, i - 3, j, k) + vector(rk3gtd_xz, i - 3, j, k) * vector(rk3Bfu_z, i - 3, j, k));
						rk3sqBf = vector(rk3Bfu_x, i - 3, j, k) * rk3Bfd_x + vector(rk3Bfu_y, i - 3, j, k) * rk3Bfd_y + vector(rk3Bfu_z, i - 3, j, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i - 3, j, k) * vector(rk3gtd_zz, i - 3, j, k) * vector(rk3gtd_xx, i - 3, j, k) + (-vector(rk3gtd_yy, i - 3, j, k) * (vector(rk3gtd_xz, i - 3, j, k) * vector(rk3gtd_xz, i - 3, j, k))) + (-vector(rk3gtd_zz, i - 3, j, k) * (vector(rk3gtd_xy, i - 3, j, k) * vector(rk3gtd_xy, i - 3, j, k))) + (-(vector(rk3gtd_yz, i - 3, j, k) * vector(rk3gtd_yz, i - 3, j, k)) * vector(rk3gtd_xx, i - 3, j, k)) + 2.0 * vector(rk3gtd_yz, i - 3, j, k) * vector(rk3gtd_xy, i - 3, j, k) * vector(rk3gtd_xz, i - 3, j, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i - 3, j, k) * vector(rk3gtd_yy, i - 3, j, k) - vector(rk3gtd_xy, i - 3, j, k) * vector(rk3gtd_xy, i - 3, j, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i - 3, j, k) * vector(rk3gtd_zz, i - 3, j, k) - vector(rk3gtd_yz, i - 3, j, k) * vector(rk3gtd_yz, i - 3, j, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i - 3, j, k) * vector(rk3gtd_yz, i - 3, j, k) - vector(rk3gtd_yy, i - 3, j, k) * vector(rk3gtd_xz, i - 3, j, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i - 3, j, k) * vector(rk3gtd_zz, i - 3, j, k)) + vector(rk3gtd_yz, i - 3, j, k) * vector(rk3gtd_xz, i - 3, j, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i - 3, j, k) + rk3gtu_xy * vector(rk3Sfd_y, i - 3, j, k) + rk3gtu_xz * vector(rk3Sfd_z, i - 3, j, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i - 3, j, k) + rk3gtu_xy * vector(vfd_y, i - 3, j, k) + rk3gtu_xz * vector(vfd_z, i - 3, j, k));
						rk3Bfvf = vector(rk3Bfu_x, i - 3, j, k) * vector(vfd_x, i - 3, j, k) + vector(rk3Bfu_y, i - 3, j, k) * vector(vfd_y, i - 3, j, k) + vector(rk3Bfu_z, i - 3, j, k) * vector(vfd_z, i - 3, j, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i - 3, j, k) * vector(rk3gtd_zz, i - 3, j, k) - vector(rk3gtd_xz, i - 3, j, k) * vector(rk3gtd_xz, i - 3, j, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i - 3, j, k) * vector(rk3gtd_yz, i - 3, j, k)) + vector(rk3gtd_xy, i - 3, j, k) * vector(rk3gtd_xz, i - 3, j, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i - 3, j, k) + rk3gtu_yz * vector(vfd_y, i - 3, j, k) + rk3gtu_zz * vector(vfd_z, i - 3, j, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i - 3, j, k) + rk3gtu_yy * vector(rk3Sfd_y, i - 3, j, k) + rk3gtu_yz * vector(rk3Sfd_z, i - 3, j, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i - 3, j, k) + rk3gtu_yz * vector(rk3Sfd_y, i - 3, j, k) + rk3gtu_zz * vector(rk3Sfd_z, i - 3, j, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i - 3, j, k) + rk3gtu_yy * vector(vfd_y, i - 3, j, k) + rk3gtu_yz * vector(vfd_z, i - 3, j, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i - 3, j, k) + rk3vfu_y * vector(vfd_y, i - 3, j, k) + rk3vfu_z * vector(vfd_z, i - 3, j, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i - 3, j, k) * vector(rk3Bfu_z, i - 3, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i - 3, j, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i - 3, j, k) * vector(rk3Bfu_x, i - 3, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i - 3, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i - 3, j, k) * vector(rk3Bfu_z, i - 3, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i - 3, j, k) * rk3vfu_z + vector(rk3Bfu_z, i - 3, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i - 3, j, k) * vector(rk3Bfu_y, i - 3, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i - 3, j, k) * rk3vfu_y + vector(rk3Bfu_y, i - 3, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i - 3, j, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i - 3, j, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i - 3, j, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i - 3, j, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i - 3, j, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i - 3, j, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i - 3, j, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i - 3, j, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i - 3, j, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i - 3, j, k) * vector(rk3Bfu_z, i - 3, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i - 3, j, k) * rk3vfu_z + vector(rk3Bfu_z, i - 3, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i - 3, j, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i - 3, j, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i - 3, j, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i - 3, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i - 3, j, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i - 3, j, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i - 3, j, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i - 3, j, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i - 3, j, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i - 3, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i - 3, j, k) * vector(rk3Bfu_y, i - 3, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i - 3, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i - 3, j, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i - 3, j, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i - 3, j, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i - 3, j, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i - 3, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i - 3, j, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i - 3, j, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i - 3, j, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i - 3, j, k));
						Fluxm3iSfd_x = FiSfd_x_mainI(vector(rk3Sfd_x, i - 3, j, k), vector(rk3gtd_xx, i - 3, j, k), vector(rk3gtd_xy, i - 3, j, k), vector(rk3gtd_xz, i - 3, j, k), vector(rk3Betau_x, i - 3, j, k), vector(rk3Alpha, i - 3, j, k), vector(TauT_xx, i - 3, j, k), vector(TauT_xy, i - 3, j, k), vector(TauT_xz, i - 3, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm3iSfd_y = FiSfd_y_mainI(vector(rk3Sfd_y, i - 3, j, k), vector(rk3gtd_xy, i - 3, j, k), vector(rk3gtd_yy, i - 3, j, k), vector(rk3gtd_yz, i - 3, j, k), vector(rk3Betau_x, i - 3, j, k), vector(rk3Alpha, i - 3, j, k), vector(TauT_xx, i - 3, j, k), vector(TauT_xy, i - 3, j, k), vector(TauT_xz, i - 3, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxm3iSfd_z = FiSfd_z_mainI(vector(rk3Sfd_z, i - 3, j, k), vector(rk3gtd_xz, i - 3, j, k), vector(rk3gtd_yz, i - 3, j, k), vector(rk3gtd_zz, i - 3, j, k), vector(rk3Betau_x, i - 3, j, k), vector(rk3Alpha, i - 3, j, k), vector(TauT_xx, i - 3, j, k), vector(TauT_xy, i - 3, j, k), vector(TauT_xz, i - 3, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxm3iBfu_x = FiBfu_x_mainI(vector(rk3phif, i - 3, j, k), vector(rk3Alpha, i - 3, j, k), rk3chi_max, rk3gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm3iBfu_y = FiBfu_y_mainI(vector(rk3Bfu_x, i - 3, j, k), vector(rk3Bfu_y, i - 3, j, k), vector(rk3phif, i - 3, j, k), vector(rk3Betau_x, i - 3, j, k), vector(rk3Betau_y, i - 3, j, k), vector(rk3Alpha, i - 3, j, k), vector(TauM_xy, i - 3, j, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3iBfu_z = FiBfu_z_mainI(vector(rk3Bfu_x, i - 3, j, k), vector(rk3Bfu_z, i - 3, j, k), vector(rk3phif, i - 3, j, k), vector(rk3Betau_x, i - 3, j, k), vector(rk3Betau_z, i - 3, j, k), vector(rk3Alpha, i - 3, j, k), vector(TauM_xz, i - 3, j, k), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3iDf = FiDf_mainI(vector(rk3Df, i - 3, j, k), vector(rk3Betau_x, i - 3, j, k), vector(rk3Alpha, i - 3, j, k), vector(TauN_x, i - 3, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm3iDYf = FiDYf_mainI(vector(rk3DYf, i - 3, j, k), vector(rk3Betau_x, i - 3, j, k), vector(rk3Alpha, i - 3, j, k), vector(TauNe_x, i - 3, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm3itauf = Fitauf_mainI(vector(rk3Df, i - 3, j, k), vector(rk3tauf, i - 3, j, k), vector(rk3Betau_x, i - 3, j, k), vector(rk3Alpha, i - 3, j, k), vector(TauN_x, i - 3, j, k), rk3sdetg, rk3Sfu_x, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm3iphif = Fiphif_mainI(vector(rk3Bfu_x, i - 3, j, k), vector(rk3phif, i - 3, j, k), vector(rk3Betau_x, i - 3, j, k), vector(rk3Alpha, i - 3, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i - 3, j, k) * (1.0 + vector(epsf, i - 3, j, k)) + vector(pf, i - 3, j, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i - 3, j, k) + rk3sqca - vector(sqcs, i - 3, j, k) * rk3sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk3Betau_x, i - 3, j, k)) + vector(rk3Alpha, i - 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk3Betau_x, i - 3, j, k)) + vector(rk3Alpha, i - 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk3Betau_x, i - 3, j, k)) + vector(rk3Alpha, i - 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk3Betau_x, i - 3, j, k)) + vector(rk3Alpha, i - 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk3Betau_x, i - 3, j, k)) + vector(rk3Alpha, i - 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk3Betau_x, i - 3, j, k)) + vector(rk3Alpha, i - 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk3Betau_x, i - 3, j, k)) + vector(rk3Alpha, i - 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk3Betau_x, i - 3, j, k)) + vector(rk3Alpha, i - 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk3Betau_x, i - 3, j, k)) + vector(rk3Alpha, i - 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk3Betau_x, i - 3, j, k)) + vector(rk3Alpha, i - 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i - 2, j, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i - 2, j, k) * vector(rk3Bfu_x, i - 2, j, k) + vector(rk3gtd_yy, i - 2, j, k) * vector(rk3Bfu_y, i - 2, j, k) + vector(rk3gtd_yz, i - 2, j, k) * vector(rk3Bfu_z, i - 2, j, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i - 2, j, k) * vector(rk3Bfu_x, i - 2, j, k) + vector(rk3gtd_yz, i - 2, j, k) * vector(rk3Bfu_y, i - 2, j, k) + vector(rk3gtd_zz, i - 2, j, k) * vector(rk3Bfu_z, i - 2, j, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i - 2, j, k) * vector(rk3Bfu_x, i - 2, j, k) + vector(rk3gtd_xy, i - 2, j, k) * vector(rk3Bfu_y, i - 2, j, k) + vector(rk3gtd_xz, i - 2, j, k) * vector(rk3Bfu_z, i - 2, j, k));
						rk3sqBf = vector(rk3Bfu_x, i - 2, j, k) * rk3Bfd_x + vector(rk3Bfu_y, i - 2, j, k) * rk3Bfd_y + vector(rk3Bfu_z, i - 2, j, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i - 2, j, k) * vector(rk3gtd_zz, i - 2, j, k) * vector(rk3gtd_xx, i - 2, j, k) + (-vector(rk3gtd_yy, i - 2, j, k) * (vector(rk3gtd_xz, i - 2, j, k) * vector(rk3gtd_xz, i - 2, j, k))) + (-vector(rk3gtd_zz, i - 2, j, k) * (vector(rk3gtd_xy, i - 2, j, k) * vector(rk3gtd_xy, i - 2, j, k))) + (-(vector(rk3gtd_yz, i - 2, j, k) * vector(rk3gtd_yz, i - 2, j, k)) * vector(rk3gtd_xx, i - 2, j, k)) + 2.0 * vector(rk3gtd_yz, i - 2, j, k) * vector(rk3gtd_xy, i - 2, j, k) * vector(rk3gtd_xz, i - 2, j, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i - 2, j, k) * vector(rk3gtd_yy, i - 2, j, k) - vector(rk3gtd_xy, i - 2, j, k) * vector(rk3gtd_xy, i - 2, j, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i - 2, j, k) * vector(rk3gtd_zz, i - 2, j, k) - vector(rk3gtd_yz, i - 2, j, k) * vector(rk3gtd_yz, i - 2, j, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i - 2, j, k) * vector(rk3gtd_yz, i - 2, j, k) - vector(rk3gtd_yy, i - 2, j, k) * vector(rk3gtd_xz, i - 2, j, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i - 2, j, k) * vector(rk3gtd_zz, i - 2, j, k)) + vector(rk3gtd_yz, i - 2, j, k) * vector(rk3gtd_xz, i - 2, j, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i - 2, j, k) + rk3gtu_xy * vector(rk3Sfd_y, i - 2, j, k) + rk3gtu_xz * vector(rk3Sfd_z, i - 2, j, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i - 2, j, k) + rk3gtu_xy * vector(vfd_y, i - 2, j, k) + rk3gtu_xz * vector(vfd_z, i - 2, j, k));
						rk3Bfvf = vector(rk3Bfu_x, i - 2, j, k) * vector(vfd_x, i - 2, j, k) + vector(rk3Bfu_y, i - 2, j, k) * vector(vfd_y, i - 2, j, k) + vector(rk3Bfu_z, i - 2, j, k) * vector(vfd_z, i - 2, j, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i - 2, j, k) * vector(rk3gtd_zz, i - 2, j, k) - vector(rk3gtd_xz, i - 2, j, k) * vector(rk3gtd_xz, i - 2, j, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i - 2, j, k) * vector(rk3gtd_yz, i - 2, j, k)) + vector(rk3gtd_xy, i - 2, j, k) * vector(rk3gtd_xz, i - 2, j, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i - 2, j, k) + rk3gtu_yz * vector(vfd_y, i - 2, j, k) + rk3gtu_zz * vector(vfd_z, i - 2, j, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i - 2, j, k) + rk3gtu_yy * vector(rk3Sfd_y, i - 2, j, k) + rk3gtu_yz * vector(rk3Sfd_z, i - 2, j, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i - 2, j, k) + rk3gtu_yz * vector(rk3Sfd_y, i - 2, j, k) + rk3gtu_zz * vector(rk3Sfd_z, i - 2, j, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i - 2, j, k) + rk3gtu_yy * vector(vfd_y, i - 2, j, k) + rk3gtu_yz * vector(vfd_z, i - 2, j, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i - 2, j, k) + rk3vfu_y * vector(vfd_y, i - 2, j, k) + rk3vfu_z * vector(vfd_z, i - 2, j, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i - 2, j, k) * vector(rk3Bfu_z, i - 2, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i - 2, j, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i - 2, j, k) * vector(rk3Bfu_x, i - 2, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i - 2, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i - 2, j, k) * vector(rk3Bfu_z, i - 2, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i - 2, j, k) * rk3vfu_z + vector(rk3Bfu_z, i - 2, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i - 2, j, k) * vector(rk3Bfu_y, i - 2, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i - 2, j, k) * rk3vfu_y + vector(rk3Bfu_y, i - 2, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i - 2, j, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i - 2, j, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i - 2, j, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i - 2, j, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i - 2, j, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i - 2, j, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i - 2, j, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i - 2, j, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i - 2, j, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i - 2, j, k) * vector(rk3Bfu_z, i - 2, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i - 2, j, k) * rk3vfu_z + vector(rk3Bfu_z, i - 2, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i - 2, j, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i - 2, j, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i - 2, j, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i - 2, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i - 2, j, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i - 2, j, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i - 2, j, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i - 2, j, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i - 2, j, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i - 2, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i - 2, j, k) * vector(rk3Bfu_y, i - 2, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i - 2, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i - 2, j, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i - 2, j, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i - 2, j, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i - 2, j, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i - 2, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i - 2, j, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i - 2, j, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i - 2, j, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i - 2, j, k));
						Fluxm2iSfd_x = FiSfd_x_mainI(vector(rk3Sfd_x, i - 2, j, k), vector(rk3gtd_xx, i - 2, j, k), vector(rk3gtd_xy, i - 2, j, k), vector(rk3gtd_xz, i - 2, j, k), vector(rk3Betau_x, i - 2, j, k), vector(rk3Alpha, i - 2, j, k), vector(TauT_xx, i - 2, j, k), vector(TauT_xy, i - 2, j, k), vector(TauT_xz, i - 2, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm2iSfd_y = FiSfd_y_mainI(vector(rk3Sfd_y, i - 2, j, k), vector(rk3gtd_xy, i - 2, j, k), vector(rk3gtd_yy, i - 2, j, k), vector(rk3gtd_yz, i - 2, j, k), vector(rk3Betau_x, i - 2, j, k), vector(rk3Alpha, i - 2, j, k), vector(TauT_xx, i - 2, j, k), vector(TauT_xy, i - 2, j, k), vector(TauT_xz, i - 2, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxm2iSfd_z = FiSfd_z_mainI(vector(rk3Sfd_z, i - 2, j, k), vector(rk3gtd_xz, i - 2, j, k), vector(rk3gtd_yz, i - 2, j, k), vector(rk3gtd_zz, i - 2, j, k), vector(rk3Betau_x, i - 2, j, k), vector(rk3Alpha, i - 2, j, k), vector(TauT_xx, i - 2, j, k), vector(TauT_xy, i - 2, j, k), vector(TauT_xz, i - 2, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxm2iBfu_x = FiBfu_x_mainI(vector(rk3phif, i - 2, j, k), vector(rk3Alpha, i - 2, j, k), rk3chi_max, rk3gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm2iBfu_y = FiBfu_y_mainI(vector(rk3Bfu_x, i - 2, j, k), vector(rk3Bfu_y, i - 2, j, k), vector(rk3phif, i - 2, j, k), vector(rk3Betau_x, i - 2, j, k), vector(rk3Betau_y, i - 2, j, k), vector(rk3Alpha, i - 2, j, k), vector(TauM_xy, i - 2, j, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2iBfu_z = FiBfu_z_mainI(vector(rk3Bfu_x, i - 2, j, k), vector(rk3Bfu_z, i - 2, j, k), vector(rk3phif, i - 2, j, k), vector(rk3Betau_x, i - 2, j, k), vector(rk3Betau_z, i - 2, j, k), vector(rk3Alpha, i - 2, j, k), vector(TauM_xz, i - 2, j, k), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2iDf = FiDf_mainI(vector(rk3Df, i - 2, j, k), vector(rk3Betau_x, i - 2, j, k), vector(rk3Alpha, i - 2, j, k), vector(TauN_x, i - 2, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm2iDYf = FiDYf_mainI(vector(rk3DYf, i - 2, j, k), vector(rk3Betau_x, i - 2, j, k), vector(rk3Alpha, i - 2, j, k), vector(TauNe_x, i - 2, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm2itauf = Fitauf_mainI(vector(rk3Df, i - 2, j, k), vector(rk3tauf, i - 2, j, k), vector(rk3Betau_x, i - 2, j, k), vector(rk3Alpha, i - 2, j, k), vector(TauN_x, i - 2, j, k), rk3sdetg, rk3Sfu_x, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm2iphif = Fiphif_mainI(vector(rk3Bfu_x, i - 2, j, k), vector(rk3phif, i - 2, j, k), vector(rk3Betau_x, i - 2, j, k), vector(rk3Alpha, i - 2, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i - 2, j, k) * (1.0 + vector(epsf, i - 2, j, k)) + vector(pf, i - 2, j, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i - 2, j, k) + rk3sqca - vector(sqcs, i - 2, j, k) * rk3sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk3Betau_x, i - 2, j, k)) + vector(rk3Alpha, i - 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk3Betau_x, i - 2, j, k)) + vector(rk3Alpha, i - 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk3Betau_x, i - 2, j, k)) + vector(rk3Alpha, i - 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk3Betau_x, i - 2, j, k)) + vector(rk3Alpha, i - 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk3Betau_x, i - 2, j, k)) + vector(rk3Alpha, i - 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk3Betau_x, i - 2, j, k)) + vector(rk3Alpha, i - 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk3Betau_x, i - 2, j, k)) + vector(rk3Alpha, i - 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk3Betau_x, i - 2, j, k)) + vector(rk3Alpha, i - 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk3Betau_x, i - 2, j, k)) + vector(rk3Alpha, i - 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk3Betau_x, i - 2, j, k)) + vector(rk3Alpha, i - 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i - 1, j, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i - 1, j, k) * vector(rk3Bfu_x, i - 1, j, k) + vector(rk3gtd_yy, i - 1, j, k) * vector(rk3Bfu_y, i - 1, j, k) + vector(rk3gtd_yz, i - 1, j, k) * vector(rk3Bfu_z, i - 1, j, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i - 1, j, k) * vector(rk3Bfu_x, i - 1, j, k) + vector(rk3gtd_yz, i - 1, j, k) * vector(rk3Bfu_y, i - 1, j, k) + vector(rk3gtd_zz, i - 1, j, k) * vector(rk3Bfu_z, i - 1, j, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i - 1, j, k) * vector(rk3Bfu_x, i - 1, j, k) + vector(rk3gtd_xy, i - 1, j, k) * vector(rk3Bfu_y, i - 1, j, k) + vector(rk3gtd_xz, i - 1, j, k) * vector(rk3Bfu_z, i - 1, j, k));
						rk3sqBf = vector(rk3Bfu_x, i - 1, j, k) * rk3Bfd_x + vector(rk3Bfu_y, i - 1, j, k) * rk3Bfd_y + vector(rk3Bfu_z, i - 1, j, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i - 1, j, k) * vector(rk3gtd_zz, i - 1, j, k) * vector(rk3gtd_xx, i - 1, j, k) + (-vector(rk3gtd_yy, i - 1, j, k) * (vector(rk3gtd_xz, i - 1, j, k) * vector(rk3gtd_xz, i - 1, j, k))) + (-vector(rk3gtd_zz, i - 1, j, k) * (vector(rk3gtd_xy, i - 1, j, k) * vector(rk3gtd_xy, i - 1, j, k))) + (-(vector(rk3gtd_yz, i - 1, j, k) * vector(rk3gtd_yz, i - 1, j, k)) * vector(rk3gtd_xx, i - 1, j, k)) + 2.0 * vector(rk3gtd_yz, i - 1, j, k) * vector(rk3gtd_xy, i - 1, j, k) * vector(rk3gtd_xz, i - 1, j, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i - 1, j, k) * vector(rk3gtd_yy, i - 1, j, k) - vector(rk3gtd_xy, i - 1, j, k) * vector(rk3gtd_xy, i - 1, j, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i - 1, j, k) * vector(rk3gtd_zz, i - 1, j, k) - vector(rk3gtd_yz, i - 1, j, k) * vector(rk3gtd_yz, i - 1, j, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i - 1, j, k) * vector(rk3gtd_yz, i - 1, j, k) - vector(rk3gtd_yy, i - 1, j, k) * vector(rk3gtd_xz, i - 1, j, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i - 1, j, k) * vector(rk3gtd_zz, i - 1, j, k)) + vector(rk3gtd_yz, i - 1, j, k) * vector(rk3gtd_xz, i - 1, j, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i - 1, j, k) + rk3gtu_xy * vector(rk3Sfd_y, i - 1, j, k) + rk3gtu_xz * vector(rk3Sfd_z, i - 1, j, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i - 1, j, k) + rk3gtu_xy * vector(vfd_y, i - 1, j, k) + rk3gtu_xz * vector(vfd_z, i - 1, j, k));
						rk3Bfvf = vector(rk3Bfu_x, i - 1, j, k) * vector(vfd_x, i - 1, j, k) + vector(rk3Bfu_y, i - 1, j, k) * vector(vfd_y, i - 1, j, k) + vector(rk3Bfu_z, i - 1, j, k) * vector(vfd_z, i - 1, j, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i - 1, j, k) * vector(rk3gtd_zz, i - 1, j, k) - vector(rk3gtd_xz, i - 1, j, k) * vector(rk3gtd_xz, i - 1, j, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i - 1, j, k) * vector(rk3gtd_yz, i - 1, j, k)) + vector(rk3gtd_xy, i - 1, j, k) * vector(rk3gtd_xz, i - 1, j, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i - 1, j, k) + rk3gtu_yz * vector(vfd_y, i - 1, j, k) + rk3gtu_zz * vector(vfd_z, i - 1, j, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i - 1, j, k) + rk3gtu_yy * vector(rk3Sfd_y, i - 1, j, k) + rk3gtu_yz * vector(rk3Sfd_z, i - 1, j, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i - 1, j, k) + rk3gtu_yz * vector(rk3Sfd_y, i - 1, j, k) + rk3gtu_zz * vector(rk3Sfd_z, i - 1, j, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i - 1, j, k) + rk3gtu_yy * vector(vfd_y, i - 1, j, k) + rk3gtu_yz * vector(vfd_z, i - 1, j, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i - 1, j, k) + rk3vfu_y * vector(vfd_y, i - 1, j, k) + rk3vfu_z * vector(vfd_z, i - 1, j, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i - 1, j, k) * vector(rk3Bfu_z, i - 1, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i - 1, j, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i - 1, j, k) * vector(rk3Bfu_x, i - 1, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i - 1, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i - 1, j, k) * vector(rk3Bfu_z, i - 1, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i - 1, j, k) * rk3vfu_z + vector(rk3Bfu_z, i - 1, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i - 1, j, k) * vector(rk3Bfu_y, i - 1, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i - 1, j, k) * rk3vfu_y + vector(rk3Bfu_y, i - 1, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i - 1, j, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i - 1, j, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i - 1, j, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i - 1, j, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i - 1, j, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i - 1, j, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i - 1, j, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i - 1, j, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i - 1, j, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i - 1, j, k) * vector(rk3Bfu_z, i - 1, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i - 1, j, k) * rk3vfu_z + vector(rk3Bfu_z, i - 1, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i - 1, j, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i - 1, j, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i - 1, j, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i - 1, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i - 1, j, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i - 1, j, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i - 1, j, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i - 1, j, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i - 1, j, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i - 1, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i - 1, j, k) * vector(rk3Bfu_y, i - 1, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i - 1, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i - 1, j, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i - 1, j, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i - 1, j, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i - 1, j, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i - 1, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i - 1, j, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i - 1, j, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i - 1, j, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i - 1, j, k));
						Fluxm1iSfd_x = FiSfd_x_mainI(vector(rk3Sfd_x, i - 1, j, k), vector(rk3gtd_xx, i - 1, j, k), vector(rk3gtd_xy, i - 1, j, k), vector(rk3gtd_xz, i - 1, j, k), vector(rk3Betau_x, i - 1, j, k), vector(rk3Alpha, i - 1, j, k), vector(TauT_xx, i - 1, j, k), vector(TauT_xy, i - 1, j, k), vector(TauT_xz, i - 1, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm1iSfd_y = FiSfd_y_mainI(vector(rk3Sfd_y, i - 1, j, k), vector(rk3gtd_xy, i - 1, j, k), vector(rk3gtd_yy, i - 1, j, k), vector(rk3gtd_yz, i - 1, j, k), vector(rk3Betau_x, i - 1, j, k), vector(rk3Alpha, i - 1, j, k), vector(TauT_xx, i - 1, j, k), vector(TauT_xy, i - 1, j, k), vector(TauT_xz, i - 1, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxm1iSfd_z = FiSfd_z_mainI(vector(rk3Sfd_z, i - 1, j, k), vector(rk3gtd_xz, i - 1, j, k), vector(rk3gtd_yz, i - 1, j, k), vector(rk3gtd_zz, i - 1, j, k), vector(rk3Betau_x, i - 1, j, k), vector(rk3Alpha, i - 1, j, k), vector(TauT_xx, i - 1, j, k), vector(TauT_xy, i - 1, j, k), vector(TauT_xz, i - 1, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxm1iBfu_x = FiBfu_x_mainI(vector(rk3phif, i - 1, j, k), vector(rk3Alpha, i - 1, j, k), rk3chi_max, rk3gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxm1iBfu_y = FiBfu_y_mainI(vector(rk3Bfu_x, i - 1, j, k), vector(rk3Bfu_y, i - 1, j, k), vector(rk3phif, i - 1, j, k), vector(rk3Betau_x, i - 1, j, k), vector(rk3Betau_y, i - 1, j, k), vector(rk3Alpha, i - 1, j, k), vector(TauM_xy, i - 1, j, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1iBfu_z = FiBfu_z_mainI(vector(rk3Bfu_x, i - 1, j, k), vector(rk3Bfu_z, i - 1, j, k), vector(rk3phif, i - 1, j, k), vector(rk3Betau_x, i - 1, j, k), vector(rk3Betau_z, i - 1, j, k), vector(rk3Alpha, i - 1, j, k), vector(TauM_xz, i - 1, j, k), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1iDf = FiDf_mainI(vector(rk3Df, i - 1, j, k), vector(rk3Betau_x, i - 1, j, k), vector(rk3Alpha, i - 1, j, k), vector(TauN_x, i - 1, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm1iDYf = FiDYf_mainI(vector(rk3DYf, i - 1, j, k), vector(rk3Betau_x, i - 1, j, k), vector(rk3Alpha, i - 1, j, k), vector(TauNe_x, i - 1, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm1itauf = Fitauf_mainI(vector(rk3Df, i - 1, j, k), vector(rk3tauf, i - 1, j, k), vector(rk3Betau_x, i - 1, j, k), vector(rk3Alpha, i - 1, j, k), vector(TauN_x, i - 1, j, k), rk3sdetg, rk3Sfu_x, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxm1iphif = Fiphif_mainI(vector(rk3Bfu_x, i - 1, j, k), vector(rk3phif, i - 1, j, k), vector(rk3Betau_x, i - 1, j, k), vector(rk3Alpha, i - 1, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i - 1, j, k) * (1.0 + vector(epsf, i - 1, j, k)) + vector(pf, i - 1, j, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i - 1, j, k) + rk3sqca - vector(sqcs, i - 1, j, k) * rk3sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk3Betau_x, i - 1, j, k)) + vector(rk3Alpha, i - 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk3Betau_x, i - 1, j, k)) + vector(rk3Alpha, i - 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk3Betau_x, i - 1, j, k)) + vector(rk3Alpha, i - 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk3Betau_x, i - 1, j, k)) + vector(rk3Alpha, i - 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk3Betau_x, i - 1, j, k)) + vector(rk3Alpha, i - 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk3Betau_x, i - 1, j, k)) + vector(rk3Alpha, i - 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk3Betau_x, i - 1, j, k)) + vector(rk3Alpha, i - 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk3Betau_x, i - 1, j, k)) + vector(rk3Alpha, i - 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk3Betau_x, i - 1, j, k)) + vector(rk3Alpha, i - 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk3Betau_x, i - 1, j, k)) + vector(rk3Alpha, i - 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i + 1, j, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i + 1, j, k) * vector(rk3Bfu_x, i + 1, j, k) + vector(rk3gtd_yy, i + 1, j, k) * vector(rk3Bfu_y, i + 1, j, k) + vector(rk3gtd_yz, i + 1, j, k) * vector(rk3Bfu_z, i + 1, j, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i + 1, j, k) * vector(rk3Bfu_x, i + 1, j, k) + vector(rk3gtd_yz, i + 1, j, k) * vector(rk3Bfu_y, i + 1, j, k) + vector(rk3gtd_zz, i + 1, j, k) * vector(rk3Bfu_z, i + 1, j, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i + 1, j, k) * vector(rk3Bfu_x, i + 1, j, k) + vector(rk3gtd_xy, i + 1, j, k) * vector(rk3Bfu_y, i + 1, j, k) + vector(rk3gtd_xz, i + 1, j, k) * vector(rk3Bfu_z, i + 1, j, k));
						rk3sqBf = vector(rk3Bfu_x, i + 1, j, k) * rk3Bfd_x + vector(rk3Bfu_y, i + 1, j, k) * rk3Bfd_y + vector(rk3Bfu_z, i + 1, j, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i + 1, j, k) * vector(rk3gtd_zz, i + 1, j, k) * vector(rk3gtd_xx, i + 1, j, k) + (-vector(rk3gtd_yy, i + 1, j, k) * (vector(rk3gtd_xz, i + 1, j, k) * vector(rk3gtd_xz, i + 1, j, k))) + (-vector(rk3gtd_zz, i + 1, j, k) * (vector(rk3gtd_xy, i + 1, j, k) * vector(rk3gtd_xy, i + 1, j, k))) + (-(vector(rk3gtd_yz, i + 1, j, k) * vector(rk3gtd_yz, i + 1, j, k)) * vector(rk3gtd_xx, i + 1, j, k)) + 2.0 * vector(rk3gtd_yz, i + 1, j, k) * vector(rk3gtd_xy, i + 1, j, k) * vector(rk3gtd_xz, i + 1, j, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i + 1, j, k) * vector(rk3gtd_yy, i + 1, j, k) - vector(rk3gtd_xy, i + 1, j, k) * vector(rk3gtd_xy, i + 1, j, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i + 1, j, k) * vector(rk3gtd_zz, i + 1, j, k) - vector(rk3gtd_yz, i + 1, j, k) * vector(rk3gtd_yz, i + 1, j, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i + 1, j, k) * vector(rk3gtd_yz, i + 1, j, k) - vector(rk3gtd_yy, i + 1, j, k) * vector(rk3gtd_xz, i + 1, j, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i + 1, j, k) * vector(rk3gtd_zz, i + 1, j, k)) + vector(rk3gtd_yz, i + 1, j, k) * vector(rk3gtd_xz, i + 1, j, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i + 1, j, k) + rk3gtu_xy * vector(rk3Sfd_y, i + 1, j, k) + rk3gtu_xz * vector(rk3Sfd_z, i + 1, j, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i + 1, j, k) + rk3gtu_xy * vector(vfd_y, i + 1, j, k) + rk3gtu_xz * vector(vfd_z, i + 1, j, k));
						rk3Bfvf = vector(rk3Bfu_x, i + 1, j, k) * vector(vfd_x, i + 1, j, k) + vector(rk3Bfu_y, i + 1, j, k) * vector(vfd_y, i + 1, j, k) + vector(rk3Bfu_z, i + 1, j, k) * vector(vfd_z, i + 1, j, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i + 1, j, k) * vector(rk3gtd_zz, i + 1, j, k) - vector(rk3gtd_xz, i + 1, j, k) * vector(rk3gtd_xz, i + 1, j, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i + 1, j, k) * vector(rk3gtd_yz, i + 1, j, k)) + vector(rk3gtd_xy, i + 1, j, k) * vector(rk3gtd_xz, i + 1, j, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i + 1, j, k) + rk3gtu_yz * vector(vfd_y, i + 1, j, k) + rk3gtu_zz * vector(vfd_z, i + 1, j, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i + 1, j, k) + rk3gtu_yy * vector(rk3Sfd_y, i + 1, j, k) + rk3gtu_yz * vector(rk3Sfd_z, i + 1, j, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i + 1, j, k) + rk3gtu_yz * vector(rk3Sfd_y, i + 1, j, k) + rk3gtu_zz * vector(rk3Sfd_z, i + 1, j, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i + 1, j, k) + rk3gtu_yy * vector(vfd_y, i + 1, j, k) + rk3gtu_yz * vector(vfd_z, i + 1, j, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i + 1, j, k) + rk3vfu_y * vector(vfd_y, i + 1, j, k) + rk3vfu_z * vector(vfd_z, i + 1, j, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i + 1, j, k) * vector(rk3Bfu_z, i + 1, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i + 1, j, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i + 1, j, k) * vector(rk3Bfu_x, i + 1, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i + 1, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i + 1, j, k) * vector(rk3Bfu_z, i + 1, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i + 1, j, k) * rk3vfu_z + vector(rk3Bfu_z, i + 1, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i + 1, j, k) * vector(rk3Bfu_y, i + 1, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i + 1, j, k) * rk3vfu_y + vector(rk3Bfu_y, i + 1, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i + 1, j, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i + 1, j, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i + 1, j, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i + 1, j, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i + 1, j, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i + 1, j, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i + 1, j, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i + 1, j, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i + 1, j, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i + 1, j, k) * vector(rk3Bfu_z, i + 1, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i + 1, j, k) * rk3vfu_z + vector(rk3Bfu_z, i + 1, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i + 1, j, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i + 1, j, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i + 1, j, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i + 1, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i + 1, j, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i + 1, j, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i + 1, j, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i + 1, j, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i + 1, j, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i + 1, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i + 1, j, k) * vector(rk3Bfu_y, i + 1, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i + 1, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i + 1, j, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i + 1, j, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i + 1, j, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i + 1, j, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i + 1, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i + 1, j, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i + 1, j, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i + 1, j, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i + 1, j, k));
						Fluxp1iSfd_x = FiSfd_x_mainI(vector(rk3Sfd_x, i + 1, j, k), vector(rk3gtd_xx, i + 1, j, k), vector(rk3gtd_xy, i + 1, j, k), vector(rk3gtd_xz, i + 1, j, k), vector(rk3Betau_x, i + 1, j, k), vector(rk3Alpha, i + 1, j, k), vector(TauT_xx, i + 1, j, k), vector(TauT_xy, i + 1, j, k), vector(TauT_xz, i + 1, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp1iSfd_y = FiSfd_y_mainI(vector(rk3Sfd_y, i + 1, j, k), vector(rk3gtd_xy, i + 1, j, k), vector(rk3gtd_yy, i + 1, j, k), vector(rk3gtd_yz, i + 1, j, k), vector(rk3Betau_x, i + 1, j, k), vector(rk3Alpha, i + 1, j, k), vector(TauT_xx, i + 1, j, k), vector(TauT_xy, i + 1, j, k), vector(TauT_xz, i + 1, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxp1iSfd_z = FiSfd_z_mainI(vector(rk3Sfd_z, i + 1, j, k), vector(rk3gtd_xz, i + 1, j, k), vector(rk3gtd_yz, i + 1, j, k), vector(rk3gtd_zz, i + 1, j, k), vector(rk3Betau_x, i + 1, j, k), vector(rk3Alpha, i + 1, j, k), vector(TauT_xx, i + 1, j, k), vector(TauT_xy, i + 1, j, k), vector(TauT_xz, i + 1, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxp1iBfu_x = FiBfu_x_mainI(vector(rk3phif, i + 1, j, k), vector(rk3Alpha, i + 1, j, k), rk3chi_max, rk3gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp1iBfu_y = FiBfu_y_mainI(vector(rk3Bfu_x, i + 1, j, k), vector(rk3Bfu_y, i + 1, j, k), vector(rk3phif, i + 1, j, k), vector(rk3Betau_x, i + 1, j, k), vector(rk3Betau_y, i + 1, j, k), vector(rk3Alpha, i + 1, j, k), vector(TauM_xy, i + 1, j, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1iBfu_z = FiBfu_z_mainI(vector(rk3Bfu_x, i + 1, j, k), vector(rk3Bfu_z, i + 1, j, k), vector(rk3phif, i + 1, j, k), vector(rk3Betau_x, i + 1, j, k), vector(rk3Betau_z, i + 1, j, k), vector(rk3Alpha, i + 1, j, k), vector(TauM_xz, i + 1, j, k), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1iDf = FiDf_mainI(vector(rk3Df, i + 1, j, k), vector(rk3Betau_x, i + 1, j, k), vector(rk3Alpha, i + 1, j, k), vector(TauN_x, i + 1, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp1iDYf = FiDYf_mainI(vector(rk3DYf, i + 1, j, k), vector(rk3Betau_x, i + 1, j, k), vector(rk3Alpha, i + 1, j, k), vector(TauNe_x, i + 1, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp1itauf = Fitauf_mainI(vector(rk3Df, i + 1, j, k), vector(rk3tauf, i + 1, j, k), vector(rk3Betau_x, i + 1, j, k), vector(rk3Alpha, i + 1, j, k), vector(TauN_x, i + 1, j, k), rk3sdetg, rk3Sfu_x, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp1iphif = Fiphif_mainI(vector(rk3Bfu_x, i + 1, j, k), vector(rk3phif, i + 1, j, k), vector(rk3Betau_x, i + 1, j, k), vector(rk3Alpha, i + 1, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i + 1, j, k) * (1.0 + vector(epsf, i + 1, j, k)) + vector(pf, i + 1, j, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i + 1, j, k) + rk3sqca - vector(sqcs, i + 1, j, k) * rk3sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk3Betau_x, i + 1, j, k)) + vector(rk3Alpha, i + 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk3Betau_x, i + 1, j, k)) + vector(rk3Alpha, i + 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk3Betau_x, i + 1, j, k)) + vector(rk3Alpha, i + 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk3Betau_x, i + 1, j, k)) + vector(rk3Alpha, i + 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk3Betau_x, i + 1, j, k)) + vector(rk3Alpha, i + 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk3Betau_x, i + 1, j, k)) + vector(rk3Alpha, i + 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk3Betau_x, i + 1, j, k)) + vector(rk3Alpha, i + 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk3Betau_x, i + 1, j, k)) + vector(rk3Alpha, i + 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk3Betau_x, i + 1, j, k)) + vector(rk3Alpha, i + 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk3Betau_x, i + 1, j, k)) + vector(rk3Alpha, i + 1, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i + 2, j, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i + 2, j, k) * vector(rk3Bfu_x, i + 2, j, k) + vector(rk3gtd_yy, i + 2, j, k) * vector(rk3Bfu_y, i + 2, j, k) + vector(rk3gtd_yz, i + 2, j, k) * vector(rk3Bfu_z, i + 2, j, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i + 2, j, k) * vector(rk3Bfu_x, i + 2, j, k) + vector(rk3gtd_yz, i + 2, j, k) * vector(rk3Bfu_y, i + 2, j, k) + vector(rk3gtd_zz, i + 2, j, k) * vector(rk3Bfu_z, i + 2, j, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i + 2, j, k) * vector(rk3Bfu_x, i + 2, j, k) + vector(rk3gtd_xy, i + 2, j, k) * vector(rk3Bfu_y, i + 2, j, k) + vector(rk3gtd_xz, i + 2, j, k) * vector(rk3Bfu_z, i + 2, j, k));
						rk3sqBf = vector(rk3Bfu_x, i + 2, j, k) * rk3Bfd_x + vector(rk3Bfu_y, i + 2, j, k) * rk3Bfd_y + vector(rk3Bfu_z, i + 2, j, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i + 2, j, k) * vector(rk3gtd_zz, i + 2, j, k) * vector(rk3gtd_xx, i + 2, j, k) + (-vector(rk3gtd_yy, i + 2, j, k) * (vector(rk3gtd_xz, i + 2, j, k) * vector(rk3gtd_xz, i + 2, j, k))) + (-vector(rk3gtd_zz, i + 2, j, k) * (vector(rk3gtd_xy, i + 2, j, k) * vector(rk3gtd_xy, i + 2, j, k))) + (-(vector(rk3gtd_yz, i + 2, j, k) * vector(rk3gtd_yz, i + 2, j, k)) * vector(rk3gtd_xx, i + 2, j, k)) + 2.0 * vector(rk3gtd_yz, i + 2, j, k) * vector(rk3gtd_xy, i + 2, j, k) * vector(rk3gtd_xz, i + 2, j, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i + 2, j, k) * vector(rk3gtd_yy, i + 2, j, k) - vector(rk3gtd_xy, i + 2, j, k) * vector(rk3gtd_xy, i + 2, j, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i + 2, j, k) * vector(rk3gtd_zz, i + 2, j, k) - vector(rk3gtd_yz, i + 2, j, k) * vector(rk3gtd_yz, i + 2, j, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i + 2, j, k) * vector(rk3gtd_yz, i + 2, j, k) - vector(rk3gtd_yy, i + 2, j, k) * vector(rk3gtd_xz, i + 2, j, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i + 2, j, k) * vector(rk3gtd_zz, i + 2, j, k)) + vector(rk3gtd_yz, i + 2, j, k) * vector(rk3gtd_xz, i + 2, j, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i + 2, j, k) + rk3gtu_xy * vector(rk3Sfd_y, i + 2, j, k) + rk3gtu_xz * vector(rk3Sfd_z, i + 2, j, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i + 2, j, k) + rk3gtu_xy * vector(vfd_y, i + 2, j, k) + rk3gtu_xz * vector(vfd_z, i + 2, j, k));
						rk3Bfvf = vector(rk3Bfu_x, i + 2, j, k) * vector(vfd_x, i + 2, j, k) + vector(rk3Bfu_y, i + 2, j, k) * vector(vfd_y, i + 2, j, k) + vector(rk3Bfu_z, i + 2, j, k) * vector(vfd_z, i + 2, j, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i + 2, j, k) * vector(rk3gtd_zz, i + 2, j, k) - vector(rk3gtd_xz, i + 2, j, k) * vector(rk3gtd_xz, i + 2, j, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i + 2, j, k) * vector(rk3gtd_yz, i + 2, j, k)) + vector(rk3gtd_xy, i + 2, j, k) * vector(rk3gtd_xz, i + 2, j, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i + 2, j, k) + rk3gtu_yz * vector(vfd_y, i + 2, j, k) + rk3gtu_zz * vector(vfd_z, i + 2, j, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i + 2, j, k) + rk3gtu_yy * vector(rk3Sfd_y, i + 2, j, k) + rk3gtu_yz * vector(rk3Sfd_z, i + 2, j, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i + 2, j, k) + rk3gtu_yz * vector(rk3Sfd_y, i + 2, j, k) + rk3gtu_zz * vector(rk3Sfd_z, i + 2, j, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i + 2, j, k) + rk3gtu_yy * vector(vfd_y, i + 2, j, k) + rk3gtu_yz * vector(vfd_z, i + 2, j, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i + 2, j, k) + rk3vfu_y * vector(vfd_y, i + 2, j, k) + rk3vfu_z * vector(vfd_z, i + 2, j, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i + 2, j, k) * vector(rk3Bfu_z, i + 2, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i + 2, j, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i + 2, j, k) * vector(rk3Bfu_x, i + 2, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i + 2, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i + 2, j, k) * vector(rk3Bfu_z, i + 2, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i + 2, j, k) * rk3vfu_z + vector(rk3Bfu_z, i + 2, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i + 2, j, k) * vector(rk3Bfu_y, i + 2, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i + 2, j, k) * rk3vfu_y + vector(rk3Bfu_y, i + 2, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i + 2, j, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i + 2, j, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i + 2, j, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i + 2, j, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i + 2, j, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i + 2, j, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i + 2, j, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i + 2, j, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i + 2, j, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i + 2, j, k) * vector(rk3Bfu_z, i + 2, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i + 2, j, k) * rk3vfu_z + vector(rk3Bfu_z, i + 2, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i + 2, j, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i + 2, j, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i + 2, j, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i + 2, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i + 2, j, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i + 2, j, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i + 2, j, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i + 2, j, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i + 2, j, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i + 2, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i + 2, j, k) * vector(rk3Bfu_y, i + 2, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i + 2, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i + 2, j, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i + 2, j, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i + 2, j, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i + 2, j, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i + 2, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i + 2, j, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i + 2, j, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i + 2, j, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i + 2, j, k));
						Fluxp2iSfd_x = FiSfd_x_mainI(vector(rk3Sfd_x, i + 2, j, k), vector(rk3gtd_xx, i + 2, j, k), vector(rk3gtd_xy, i + 2, j, k), vector(rk3gtd_xz, i + 2, j, k), vector(rk3Betau_x, i + 2, j, k), vector(rk3Alpha, i + 2, j, k), vector(TauT_xx, i + 2, j, k), vector(TauT_xy, i + 2, j, k), vector(TauT_xz, i + 2, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp2iSfd_y = FiSfd_y_mainI(vector(rk3Sfd_y, i + 2, j, k), vector(rk3gtd_xy, i + 2, j, k), vector(rk3gtd_yy, i + 2, j, k), vector(rk3gtd_yz, i + 2, j, k), vector(rk3Betau_x, i + 2, j, k), vector(rk3Alpha, i + 2, j, k), vector(TauT_xx, i + 2, j, k), vector(TauT_xy, i + 2, j, k), vector(TauT_xz, i + 2, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxp2iSfd_z = FiSfd_z_mainI(vector(rk3Sfd_z, i + 2, j, k), vector(rk3gtd_xz, i + 2, j, k), vector(rk3gtd_yz, i + 2, j, k), vector(rk3gtd_zz, i + 2, j, k), vector(rk3Betau_x, i + 2, j, k), vector(rk3Alpha, i + 2, j, k), vector(TauT_xx, i + 2, j, k), vector(TauT_xy, i + 2, j, k), vector(TauT_xz, i + 2, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxp2iBfu_x = FiBfu_x_mainI(vector(rk3phif, i + 2, j, k), vector(rk3Alpha, i + 2, j, k), rk3chi_max, rk3gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp2iBfu_y = FiBfu_y_mainI(vector(rk3Bfu_x, i + 2, j, k), vector(rk3Bfu_y, i + 2, j, k), vector(rk3phif, i + 2, j, k), vector(rk3Betau_x, i + 2, j, k), vector(rk3Betau_y, i + 2, j, k), vector(rk3Alpha, i + 2, j, k), vector(TauM_xy, i + 2, j, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2iBfu_z = FiBfu_z_mainI(vector(rk3Bfu_x, i + 2, j, k), vector(rk3Bfu_z, i + 2, j, k), vector(rk3phif, i + 2, j, k), vector(rk3Betau_x, i + 2, j, k), vector(rk3Betau_z, i + 2, j, k), vector(rk3Alpha, i + 2, j, k), vector(TauM_xz, i + 2, j, k), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2iDf = FiDf_mainI(vector(rk3Df, i + 2, j, k), vector(rk3Betau_x, i + 2, j, k), vector(rk3Alpha, i + 2, j, k), vector(TauN_x, i + 2, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp2iDYf = FiDYf_mainI(vector(rk3DYf, i + 2, j, k), vector(rk3Betau_x, i + 2, j, k), vector(rk3Alpha, i + 2, j, k), vector(TauNe_x, i + 2, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp2itauf = Fitauf_mainI(vector(rk3Df, i + 2, j, k), vector(rk3tauf, i + 2, j, k), vector(rk3Betau_x, i + 2, j, k), vector(rk3Alpha, i + 2, j, k), vector(TauN_x, i + 2, j, k), rk3sdetg, rk3Sfu_x, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp2iphif = Fiphif_mainI(vector(rk3Bfu_x, i + 2, j, k), vector(rk3phif, i + 2, j, k), vector(rk3Betau_x, i + 2, j, k), vector(rk3Alpha, i + 2, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i + 2, j, k) * (1.0 + vector(epsf, i + 2, j, k)) + vector(pf, i + 2, j, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i + 2, j, k) + rk3sqca - vector(sqcs, i + 2, j, k) * rk3sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk3Betau_x, i + 2, j, k)) + vector(rk3Alpha, i + 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk3Betau_x, i + 2, j, k)) + vector(rk3Alpha, i + 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk3Betau_x, i + 2, j, k)) + vector(rk3Alpha, i + 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk3Betau_x, i + 2, j, k)) + vector(rk3Alpha, i + 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk3Betau_x, i + 2, j, k)) + vector(rk3Alpha, i + 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk3Betau_x, i + 2, j, k)) + vector(rk3Alpha, i + 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk3Betau_x, i + 2, j, k)) + vector(rk3Alpha, i + 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk3Betau_x, i + 2, j, k)) + vector(rk3Alpha, i + 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk3Betau_x, i + 2, j, k)) + vector(rk3Alpha, i + 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk3Betau_x, i + 2, j, k)) + vector(rk3Alpha, i + 2, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i + 3, j, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i + 3, j, k) * vector(rk3Bfu_x, i + 3, j, k) + vector(rk3gtd_yy, i + 3, j, k) * vector(rk3Bfu_y, i + 3, j, k) + vector(rk3gtd_yz, i + 3, j, k) * vector(rk3Bfu_z, i + 3, j, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i + 3, j, k) * vector(rk3Bfu_x, i + 3, j, k) + vector(rk3gtd_yz, i + 3, j, k) * vector(rk3Bfu_y, i + 3, j, k) + vector(rk3gtd_zz, i + 3, j, k) * vector(rk3Bfu_z, i + 3, j, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i + 3, j, k) * vector(rk3Bfu_x, i + 3, j, k) + vector(rk3gtd_xy, i + 3, j, k) * vector(rk3Bfu_y, i + 3, j, k) + vector(rk3gtd_xz, i + 3, j, k) * vector(rk3Bfu_z, i + 3, j, k));
						rk3sqBf = vector(rk3Bfu_x, i + 3, j, k) * rk3Bfd_x + vector(rk3Bfu_y, i + 3, j, k) * rk3Bfd_y + vector(rk3Bfu_z, i + 3, j, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i + 3, j, k) * vector(rk3gtd_zz, i + 3, j, k) * vector(rk3gtd_xx, i + 3, j, k) + (-vector(rk3gtd_yy, i + 3, j, k) * (vector(rk3gtd_xz, i + 3, j, k) * vector(rk3gtd_xz, i + 3, j, k))) + (-vector(rk3gtd_zz, i + 3, j, k) * (vector(rk3gtd_xy, i + 3, j, k) * vector(rk3gtd_xy, i + 3, j, k))) + (-(vector(rk3gtd_yz, i + 3, j, k) * vector(rk3gtd_yz, i + 3, j, k)) * vector(rk3gtd_xx, i + 3, j, k)) + 2.0 * vector(rk3gtd_yz, i + 3, j, k) * vector(rk3gtd_xy, i + 3, j, k) * vector(rk3gtd_xz, i + 3, j, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i + 3, j, k) * vector(rk3gtd_yy, i + 3, j, k) - vector(rk3gtd_xy, i + 3, j, k) * vector(rk3gtd_xy, i + 3, j, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i + 3, j, k) * vector(rk3gtd_zz, i + 3, j, k) - vector(rk3gtd_yz, i + 3, j, k) * vector(rk3gtd_yz, i + 3, j, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i + 3, j, k) * vector(rk3gtd_yz, i + 3, j, k) - vector(rk3gtd_yy, i + 3, j, k) * vector(rk3gtd_xz, i + 3, j, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i + 3, j, k) * vector(rk3gtd_zz, i + 3, j, k)) + vector(rk3gtd_yz, i + 3, j, k) * vector(rk3gtd_xz, i + 3, j, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i + 3, j, k) + rk3gtu_xy * vector(rk3Sfd_y, i + 3, j, k) + rk3gtu_xz * vector(rk3Sfd_z, i + 3, j, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i + 3, j, k) + rk3gtu_xy * vector(vfd_y, i + 3, j, k) + rk3gtu_xz * vector(vfd_z, i + 3, j, k));
						rk3Bfvf = vector(rk3Bfu_x, i + 3, j, k) * vector(vfd_x, i + 3, j, k) + vector(rk3Bfu_y, i + 3, j, k) * vector(vfd_y, i + 3, j, k) + vector(rk3Bfu_z, i + 3, j, k) * vector(vfd_z, i + 3, j, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i + 3, j, k) * vector(rk3gtd_zz, i + 3, j, k) - vector(rk3gtd_xz, i + 3, j, k) * vector(rk3gtd_xz, i + 3, j, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i + 3, j, k) * vector(rk3gtd_yz, i + 3, j, k)) + vector(rk3gtd_xy, i + 3, j, k) * vector(rk3gtd_xz, i + 3, j, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i + 3, j, k) + rk3gtu_yz * vector(vfd_y, i + 3, j, k) + rk3gtu_zz * vector(vfd_z, i + 3, j, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i + 3, j, k) + rk3gtu_yy * vector(rk3Sfd_y, i + 3, j, k) + rk3gtu_yz * vector(rk3Sfd_z, i + 3, j, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i + 3, j, k) + rk3gtu_yz * vector(rk3Sfd_y, i + 3, j, k) + rk3gtu_zz * vector(rk3Sfd_z, i + 3, j, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i + 3, j, k) + rk3gtu_yy * vector(vfd_y, i + 3, j, k) + rk3gtu_yz * vector(vfd_z, i + 3, j, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i + 3, j, k) + rk3vfu_y * vector(vfd_y, i + 3, j, k) + rk3vfu_z * vector(vfd_z, i + 3, j, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i + 3, j, k) * vector(rk3Bfu_z, i + 3, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i + 3, j, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i + 3, j, k) * vector(rk3Bfu_x, i + 3, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i + 3, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i + 3, j, k) * vector(rk3Bfu_z, i + 3, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i + 3, j, k) * rk3vfu_z + vector(rk3Bfu_z, i + 3, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i + 3, j, k) * vector(rk3Bfu_y, i + 3, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i + 3, j, k) * rk3vfu_y + vector(rk3Bfu_y, i + 3, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i + 3, j, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i + 3, j, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i + 3, j, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i + 3, j, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i + 3, j, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i + 3, j, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i + 3, j, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i + 3, j, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i + 3, j, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i + 3, j, k) * vector(rk3Bfu_z, i + 3, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i + 3, j, k) * rk3vfu_z + vector(rk3Bfu_z, i + 3, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i + 3, j, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i + 3, j, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i + 3, j, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i + 3, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i + 3, j, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i + 3, j, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i + 3, j, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i + 3, j, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i + 3, j, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i + 3, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i + 3, j, k) * vector(rk3Bfu_y, i + 3, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i + 3, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i + 3, j, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i + 3, j, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i + 3, j, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i + 3, j, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i + 3, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i + 3, j, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i + 3, j, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i + 3, j, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i + 3, j, k));
						Fluxp3iSfd_x = FiSfd_x_mainI(vector(rk3Sfd_x, i + 3, j, k), vector(rk3gtd_xx, i + 3, j, k), vector(rk3gtd_xy, i + 3, j, k), vector(rk3gtd_xz, i + 3, j, k), vector(rk3Betau_x, i + 3, j, k), vector(rk3Alpha, i + 3, j, k), vector(TauT_xx, i + 3, j, k), vector(TauT_xy, i + 3, j, k), vector(TauT_xz, i + 3, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp3iSfd_y = FiSfd_y_mainI(vector(rk3Sfd_y, i + 3, j, k), vector(rk3gtd_xy, i + 3, j, k), vector(rk3gtd_yy, i + 3, j, k), vector(rk3gtd_yz, i + 3, j, k), vector(rk3Betau_x, i + 3, j, k), vector(rk3Alpha, i + 3, j, k), vector(TauT_xx, i + 3, j, k), vector(TauT_xy, i + 3, j, k), vector(TauT_xz, i + 3, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xy, dx, simPlat_dt, ilast, jlast);
						Fluxp3iSfd_z = FiSfd_z_mainI(vector(rk3Sfd_z, i + 3, j, k), vector(rk3gtd_xz, i + 3, j, k), vector(rk3gtd_yz, i + 3, j, k), vector(rk3gtd_zz, i + 3, j, k), vector(rk3Betau_x, i + 3, j, k), vector(rk3Alpha, i + 3, j, k), vector(TauT_xx, i + 3, j, k), vector(TauT_xy, i + 3, j, k), vector(TauT_xz, i + 3, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_xz, dx, simPlat_dt, ilast, jlast);
						Fluxp3iBfu_x = FiBfu_x_mainI(vector(rk3phif, i + 3, j, k), vector(rk3Alpha, i + 3, j, k), rk3chi_max, rk3gtu_xx, dx, simPlat_dt, ilast, jlast);
						Fluxp3iBfu_y = FiBfu_y_mainI(vector(rk3Bfu_x, i + 3, j, k), vector(rk3Bfu_y, i + 3, j, k), vector(rk3phif, i + 3, j, k), vector(rk3Betau_x, i + 3, j, k), vector(rk3Betau_y, i + 3, j, k), vector(rk3Alpha, i + 3, j, k), vector(TauM_xy, i + 3, j, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3iBfu_z = FiBfu_z_mainI(vector(rk3Bfu_x, i + 3, j, k), vector(rk3Bfu_z, i + 3, j, k), vector(rk3phif, i + 3, j, k), vector(rk3Betau_x, i + 3, j, k), vector(rk3Betau_z, i + 3, j, k), vector(rk3Alpha, i + 3, j, k), vector(TauM_xz, i + 3, j, k), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3iDf = FiDf_mainI(vector(rk3Df, i + 3, j, k), vector(rk3Betau_x, i + 3, j, k), vector(rk3Alpha, i + 3, j, k), vector(TauN_x, i + 3, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp3iDYf = FiDYf_mainI(vector(rk3DYf, i + 3, j, k), vector(rk3Betau_x, i + 3, j, k), vector(rk3Alpha, i + 3, j, k), vector(TauNe_x, i + 3, j, k), rk3sdetg, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp3itauf = Fitauf_mainI(vector(rk3Df, i + 3, j, k), vector(rk3tauf, i + 3, j, k), vector(rk3Betau_x, i + 3, j, k), vector(rk3Alpha, i + 3, j, k), vector(TauN_x, i + 3, j, k), rk3sdetg, rk3Sfu_x, rk3vfu_x, dx, simPlat_dt, ilast, jlast);
						Fluxp3iphif = Fiphif_mainI(vector(rk3Bfu_x, i + 3, j, k), vector(rk3phif, i + 3, j, k), vector(rk3Betau_x, i + 3, j, k), vector(rk3Alpha, i + 3, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i + 3, j, k) * (1.0 + vector(epsf, i + 3, j, k)) + vector(pf, i + 3, j, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i + 3, j, k) + rk3sqca - vector(sqcs, i + 3, j, k) * rk3sqca;
						SpeediSfd_x = MAX(SpeediSfd_x, fabs(fabs(vector(rk3Betau_x, i + 3, j, k)) + vector(rk3Alpha, i + 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediSfd_y = MAX(SpeediSfd_y, fabs(fabs(vector(rk3Betau_x, i + 3, j, k)) + vector(rk3Alpha, i + 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediSfd_z = MAX(SpeediSfd_z, fabs(fabs(vector(rk3Betau_x, i + 3, j, k)) + vector(rk3Alpha, i + 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_x = MAX(SpeediBfu_x, fabs(fabs(vector(rk3Betau_x, i + 3, j, k)) + vector(rk3Alpha, i + 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_y = MAX(SpeediBfu_y, fabs(fabs(vector(rk3Betau_x, i + 3, j, k)) + vector(rk3Alpha, i + 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediBfu_z = MAX(SpeediBfu_z, fabs(fabs(vector(rk3Betau_x, i + 3, j, k)) + vector(rk3Alpha, i + 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediDf = MAX(SpeediDf, fabs(fabs(vector(rk3Betau_x, i + 3, j, k)) + vector(rk3Alpha, i + 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						SpeediDYf = MAX(SpeediDYf, fabs(fabs(vector(rk3Betau_x, i + 3, j, k)) + vector(rk3Alpha, i + 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						Speeditauf = MAX(Speeditauf, fabs(fabs(vector(rk3Betau_x, i + 3, j, k)) + vector(rk3Alpha, i + 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						Speediphif = MAX(Speediphif, fabs(fabs(vector(rk3Betau_x, i + 3, j, k)) + vector(rk3Alpha, i + 3, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_x) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_xx - (1.0 - rk3sqa) * (rk3vfu_x * rk3vfu_x)))))));
						LLFpiSfd_x = 0.5 * (FluxiSfd_x + fabs(SpeediSfd_x) * vector(rk3Sfd_x, i, j, k));
						LLFpiSfd_y = 0.5 * (FluxiSfd_y + fabs(SpeediSfd_y) * vector(rk3Sfd_y, i, j, k));
						LLFpiSfd_z = 0.5 * (FluxiSfd_z + fabs(SpeediSfd_z) * vector(rk3Sfd_z, i, j, k));
						LLFpiBfu_x = 0.5 * (FluxiBfu_x + fabs(SpeediBfu_x) * vector(rk3Bfu_x, i, j, k));
						LLFpiBfu_y = 0.5 * (FluxiBfu_y + fabs(SpeediBfu_y) * vector(rk3Bfu_y, i, j, k));
						LLFpiBfu_z = 0.5 * (FluxiBfu_z + fabs(SpeediBfu_z) * vector(rk3Bfu_z, i, j, k));
						LLFpiDf = 0.5 * (FluxiDf + fabs(SpeediDf) * vector(rk3Df, i, j, k));
						LLFpiDYf = 0.5 * (FluxiDYf + fabs(SpeediDYf) * vector(rk3DYf, i, j, k));
						LLFpitauf = 0.5 * (Fluxitauf + fabs(Speeditauf) * vector(rk3tauf, i, j, k));
						LLFpiphif = 0.5 * (Fluxiphif + fabs(Speediphif) * vector(rk3phif, i, j, k));
						LLFniSfd_x = 0.5 * (FluxiSfd_x - fabs(SpeediSfd_x) * vector(rk3Sfd_x, i, j, k));
						LLFniSfd_y = 0.5 * (FluxiSfd_y - fabs(SpeediSfd_y) * vector(rk3Sfd_y, i, j, k));
						LLFniSfd_z = 0.5 * (FluxiSfd_z - fabs(SpeediSfd_z) * vector(rk3Sfd_z, i, j, k));
						LLFniBfu_x = 0.5 * (FluxiBfu_x - fabs(SpeediBfu_x) * vector(rk3Bfu_x, i, j, k));
						LLFniBfu_y = 0.5 * (FluxiBfu_y - fabs(SpeediBfu_y) * vector(rk3Bfu_y, i, j, k));
						LLFniBfu_z = 0.5 * (FluxiBfu_z - fabs(SpeediBfu_z) * vector(rk3Bfu_z, i, j, k));
						LLFniDf = 0.5 * (FluxiDf - fabs(SpeediDf) * vector(rk3Df, i, j, k));
						LLFniDYf = 0.5 * (FluxiDYf - fabs(SpeediDYf) * vector(rk3DYf, i, j, k));
						LLFnitauf = 0.5 * (Fluxitauf - fabs(Speeditauf) * vector(rk3tauf, i, j, k));
						LLFniphif = 0.5 * (Fluxiphif - fabs(Speediphif) * vector(rk3phif, i, j, k));
						LLFp_m3iSfd_x = 0.5 * (Fluxm3iSfd_x + fabs(SpeediSfd_x) * vector(rk3Sfd_x, i - 3, j, k));
						LLFp_m3iSfd_y = 0.5 * (Fluxm3iSfd_y + fabs(SpeediSfd_y) * vector(rk3Sfd_y, i - 3, j, k));
						LLFp_m3iSfd_z = 0.5 * (Fluxm3iSfd_z + fabs(SpeediSfd_z) * vector(rk3Sfd_z, i - 3, j, k));
						LLFp_m3iBfu_x = 0.5 * (Fluxm3iBfu_x + fabs(SpeediBfu_x) * vector(rk3Bfu_x, i - 3, j, k));
						LLFp_m3iBfu_y = 0.5 * (Fluxm3iBfu_y + fabs(SpeediBfu_y) * vector(rk3Bfu_y, i - 3, j, k));
						LLFp_m3iBfu_z = 0.5 * (Fluxm3iBfu_z + fabs(SpeediBfu_z) * vector(rk3Bfu_z, i - 3, j, k));
						LLFp_m3iDf = 0.5 * (Fluxm3iDf + fabs(SpeediDf) * vector(rk3Df, i - 3, j, k));
						LLFp_m3iDYf = 0.5 * (Fluxm3iDYf + fabs(SpeediDYf) * vector(rk3DYf, i - 3, j, k));
						LLFp_m3itauf = 0.5 * (Fluxm3itauf + fabs(Speeditauf) * vector(rk3tauf, i - 3, j, k));
						LLFp_m3iphif = 0.5 * (Fluxm3iphif + fabs(Speediphif) * vector(rk3phif, i - 3, j, k));
						LLFn_m3iSfd_x = 0.5 * (Fluxm3iSfd_x - fabs(SpeediSfd_x) * vector(rk3Sfd_x, i - 3, j, k));
						LLFn_m3iSfd_y = 0.5 * (Fluxm3iSfd_y - fabs(SpeediSfd_y) * vector(rk3Sfd_y, i - 3, j, k));
						LLFn_m3iSfd_z = 0.5 * (Fluxm3iSfd_z - fabs(SpeediSfd_z) * vector(rk3Sfd_z, i - 3, j, k));
						LLFn_m3iBfu_x = 0.5 * (Fluxm3iBfu_x - fabs(SpeediBfu_x) * vector(rk3Bfu_x, i - 3, j, k));
						LLFn_m3iBfu_y = 0.5 * (Fluxm3iBfu_y - fabs(SpeediBfu_y) * vector(rk3Bfu_y, i - 3, j, k));
						LLFn_m3iBfu_z = 0.5 * (Fluxm3iBfu_z - fabs(SpeediBfu_z) * vector(rk3Bfu_z, i - 3, j, k));
						LLFn_m3iDf = 0.5 * (Fluxm3iDf - fabs(SpeediDf) * vector(rk3Df, i - 3, j, k));
						LLFn_m3iDYf = 0.5 * (Fluxm3iDYf - fabs(SpeediDYf) * vector(rk3DYf, i - 3, j, k));
						LLFn_m3itauf = 0.5 * (Fluxm3itauf - fabs(Speeditauf) * vector(rk3tauf, i - 3, j, k));
						LLFn_m3iphif = 0.5 * (Fluxm3iphif - fabs(Speediphif) * vector(rk3phif, i - 3, j, k));
						LLFp_m2iSfd_x = 0.5 * (Fluxm2iSfd_x + fabs(SpeediSfd_x) * vector(rk3Sfd_x, i - 2, j, k));
						LLFp_m2iSfd_y = 0.5 * (Fluxm2iSfd_y + fabs(SpeediSfd_y) * vector(rk3Sfd_y, i - 2, j, k));
						LLFp_m2iSfd_z = 0.5 * (Fluxm2iSfd_z + fabs(SpeediSfd_z) * vector(rk3Sfd_z, i - 2, j, k));
						LLFp_m2iBfu_x = 0.5 * (Fluxm2iBfu_x + fabs(SpeediBfu_x) * vector(rk3Bfu_x, i - 2, j, k));
						LLFp_m2iBfu_y = 0.5 * (Fluxm2iBfu_y + fabs(SpeediBfu_y) * vector(rk3Bfu_y, i - 2, j, k));
						LLFp_m2iBfu_z = 0.5 * (Fluxm2iBfu_z + fabs(SpeediBfu_z) * vector(rk3Bfu_z, i - 2, j, k));
						LLFp_m2iDf = 0.5 * (Fluxm2iDf + fabs(SpeediDf) * vector(rk3Df, i - 2, j, k));
						LLFp_m2iDYf = 0.5 * (Fluxm2iDYf + fabs(SpeediDYf) * vector(rk3DYf, i - 2, j, k));
						LLFp_m2itauf = 0.5 * (Fluxm2itauf + fabs(Speeditauf) * vector(rk3tauf, i - 2, j, k));
						LLFp_m2iphif = 0.5 * (Fluxm2iphif + fabs(Speediphif) * vector(rk3phif, i - 2, j, k));
						LLFn_m2iSfd_x = 0.5 * (Fluxm2iSfd_x - fabs(SpeediSfd_x) * vector(rk3Sfd_x, i - 2, j, k));
						LLFn_m2iSfd_y = 0.5 * (Fluxm2iSfd_y - fabs(SpeediSfd_y) * vector(rk3Sfd_y, i - 2, j, k));
						LLFn_m2iSfd_z = 0.5 * (Fluxm2iSfd_z - fabs(SpeediSfd_z) * vector(rk3Sfd_z, i - 2, j, k));
						LLFn_m2iBfu_x = 0.5 * (Fluxm2iBfu_x - fabs(SpeediBfu_x) * vector(rk3Bfu_x, i - 2, j, k));
						LLFn_m2iBfu_y = 0.5 * (Fluxm2iBfu_y - fabs(SpeediBfu_y) * vector(rk3Bfu_y, i - 2, j, k));
						LLFn_m2iBfu_z = 0.5 * (Fluxm2iBfu_z - fabs(SpeediBfu_z) * vector(rk3Bfu_z, i - 2, j, k));
						LLFn_m2iDf = 0.5 * (Fluxm2iDf - fabs(SpeediDf) * vector(rk3Df, i - 2, j, k));
						LLFn_m2iDYf = 0.5 * (Fluxm2iDYf - fabs(SpeediDYf) * vector(rk3DYf, i - 2, j, k));
						LLFn_m2itauf = 0.5 * (Fluxm2itauf - fabs(Speeditauf) * vector(rk3tauf, i - 2, j, k));
						LLFn_m2iphif = 0.5 * (Fluxm2iphif - fabs(Speediphif) * vector(rk3phif, i - 2, j, k));
						LLFp_m1iSfd_x = 0.5 * (Fluxm1iSfd_x + fabs(SpeediSfd_x) * vector(rk3Sfd_x, i - 1, j, k));
						LLFp_m1iSfd_y = 0.5 * (Fluxm1iSfd_y + fabs(SpeediSfd_y) * vector(rk3Sfd_y, i - 1, j, k));
						LLFp_m1iSfd_z = 0.5 * (Fluxm1iSfd_z + fabs(SpeediSfd_z) * vector(rk3Sfd_z, i - 1, j, k));
						LLFp_m1iBfu_x = 0.5 * (Fluxm1iBfu_x + fabs(SpeediBfu_x) * vector(rk3Bfu_x, i - 1, j, k));
						LLFp_m1iBfu_y = 0.5 * (Fluxm1iBfu_y + fabs(SpeediBfu_y) * vector(rk3Bfu_y, i - 1, j, k));
						LLFp_m1iBfu_z = 0.5 * (Fluxm1iBfu_z + fabs(SpeediBfu_z) * vector(rk3Bfu_z, i - 1, j, k));
						LLFp_m1iDf = 0.5 * (Fluxm1iDf + fabs(SpeediDf) * vector(rk3Df, i - 1, j, k));
						LLFp_m1iDYf = 0.5 * (Fluxm1iDYf + fabs(SpeediDYf) * vector(rk3DYf, i - 1, j, k));
						LLFp_m1itauf = 0.5 * (Fluxm1itauf + fabs(Speeditauf) * vector(rk3tauf, i - 1, j, k));
						LLFp_m1iphif = 0.5 * (Fluxm1iphif + fabs(Speediphif) * vector(rk3phif, i - 1, j, k));
						LLFn_m1iSfd_x = 0.5 * (Fluxm1iSfd_x - fabs(SpeediSfd_x) * vector(rk3Sfd_x, i - 1, j, k));
						LLFn_m1iSfd_y = 0.5 * (Fluxm1iSfd_y - fabs(SpeediSfd_y) * vector(rk3Sfd_y, i - 1, j, k));
						LLFn_m1iSfd_z = 0.5 * (Fluxm1iSfd_z - fabs(SpeediSfd_z) * vector(rk3Sfd_z, i - 1, j, k));
						LLFn_m1iBfu_x = 0.5 * (Fluxm1iBfu_x - fabs(SpeediBfu_x) * vector(rk3Bfu_x, i - 1, j, k));
						LLFn_m1iBfu_y = 0.5 * (Fluxm1iBfu_y - fabs(SpeediBfu_y) * vector(rk3Bfu_y, i - 1, j, k));
						LLFn_m1iBfu_z = 0.5 * (Fluxm1iBfu_z - fabs(SpeediBfu_z) * vector(rk3Bfu_z, i - 1, j, k));
						LLFn_m1iDf = 0.5 * (Fluxm1iDf - fabs(SpeediDf) * vector(rk3Df, i - 1, j, k));
						LLFn_m1iDYf = 0.5 * (Fluxm1iDYf - fabs(SpeediDYf) * vector(rk3DYf, i - 1, j, k));
						LLFn_m1itauf = 0.5 * (Fluxm1itauf - fabs(Speeditauf) * vector(rk3tauf, i - 1, j, k));
						LLFn_m1iphif = 0.5 * (Fluxm1iphif - fabs(Speediphif) * vector(rk3phif, i - 1, j, k));
						LLFp_p1iSfd_x = 0.5 * (Fluxp1iSfd_x + fabs(SpeediSfd_x) * vector(rk3Sfd_x, i + 1, j, k));
						LLFp_p1iSfd_y = 0.5 * (Fluxp1iSfd_y + fabs(SpeediSfd_y) * vector(rk3Sfd_y, i + 1, j, k));
						LLFp_p1iSfd_z = 0.5 * (Fluxp1iSfd_z + fabs(SpeediSfd_z) * vector(rk3Sfd_z, i + 1, j, k));
						LLFp_p1iBfu_x = 0.5 * (Fluxp1iBfu_x + fabs(SpeediBfu_x) * vector(rk3Bfu_x, i + 1, j, k));
						LLFp_p1iBfu_y = 0.5 * (Fluxp1iBfu_y + fabs(SpeediBfu_y) * vector(rk3Bfu_y, i + 1, j, k));
						LLFp_p1iBfu_z = 0.5 * (Fluxp1iBfu_z + fabs(SpeediBfu_z) * vector(rk3Bfu_z, i + 1, j, k));
						LLFp_p1iDf = 0.5 * (Fluxp1iDf + fabs(SpeediDf) * vector(rk3Df, i + 1, j, k));
						LLFp_p1iDYf = 0.5 * (Fluxp1iDYf + fabs(SpeediDYf) * vector(rk3DYf, i + 1, j, k));
						LLFp_p1itauf = 0.5 * (Fluxp1itauf + fabs(Speeditauf) * vector(rk3tauf, i + 1, j, k));
						LLFp_p1iphif = 0.5 * (Fluxp1iphif + fabs(Speediphif) * vector(rk3phif, i + 1, j, k));
						LLFn_p1iSfd_x = 0.5 * (Fluxp1iSfd_x - fabs(SpeediSfd_x) * vector(rk3Sfd_x, i + 1, j, k));
						LLFn_p1iSfd_y = 0.5 * (Fluxp1iSfd_y - fabs(SpeediSfd_y) * vector(rk3Sfd_y, i + 1, j, k));
						LLFn_p1iSfd_z = 0.5 * (Fluxp1iSfd_z - fabs(SpeediSfd_z) * vector(rk3Sfd_z, i + 1, j, k));
						LLFn_p1iBfu_x = 0.5 * (Fluxp1iBfu_x - fabs(SpeediBfu_x) * vector(rk3Bfu_x, i + 1, j, k));
						LLFn_p1iBfu_y = 0.5 * (Fluxp1iBfu_y - fabs(SpeediBfu_y) * vector(rk3Bfu_y, i + 1, j, k));
						LLFn_p1iBfu_z = 0.5 * (Fluxp1iBfu_z - fabs(SpeediBfu_z) * vector(rk3Bfu_z, i + 1, j, k));
						LLFn_p1iDf = 0.5 * (Fluxp1iDf - fabs(SpeediDf) * vector(rk3Df, i + 1, j, k));
						LLFn_p1iDYf = 0.5 * (Fluxp1iDYf - fabs(SpeediDYf) * vector(rk3DYf, i + 1, j, k));
						LLFn_p1itauf = 0.5 * (Fluxp1itauf - fabs(Speeditauf) * vector(rk3tauf, i + 1, j, k));
						LLFn_p1iphif = 0.5 * (Fluxp1iphif - fabs(Speediphif) * vector(rk3phif, i + 1, j, k));
						LLFp_p2iSfd_x = 0.5 * (Fluxp2iSfd_x + fabs(SpeediSfd_x) * vector(rk3Sfd_x, i + 2, j, k));
						LLFp_p2iSfd_y = 0.5 * (Fluxp2iSfd_y + fabs(SpeediSfd_y) * vector(rk3Sfd_y, i + 2, j, k));
						LLFp_p2iSfd_z = 0.5 * (Fluxp2iSfd_z + fabs(SpeediSfd_z) * vector(rk3Sfd_z, i + 2, j, k));
						LLFp_p2iBfu_x = 0.5 * (Fluxp2iBfu_x + fabs(SpeediBfu_x) * vector(rk3Bfu_x, i + 2, j, k));
						LLFp_p2iBfu_y = 0.5 * (Fluxp2iBfu_y + fabs(SpeediBfu_y) * vector(rk3Bfu_y, i + 2, j, k));
						LLFp_p2iBfu_z = 0.5 * (Fluxp2iBfu_z + fabs(SpeediBfu_z) * vector(rk3Bfu_z, i + 2, j, k));
						LLFp_p2iDf = 0.5 * (Fluxp2iDf + fabs(SpeediDf) * vector(rk3Df, i + 2, j, k));
						LLFp_p2iDYf = 0.5 * (Fluxp2iDYf + fabs(SpeediDYf) * vector(rk3DYf, i + 2, j, k));
						LLFp_p2itauf = 0.5 * (Fluxp2itauf + fabs(Speeditauf) * vector(rk3tauf, i + 2, j, k));
						LLFp_p2iphif = 0.5 * (Fluxp2iphif + fabs(Speediphif) * vector(rk3phif, i + 2, j, k));
						LLFn_p2iSfd_x = 0.5 * (Fluxp2iSfd_x - fabs(SpeediSfd_x) * vector(rk3Sfd_x, i + 2, j, k));
						LLFn_p2iSfd_y = 0.5 * (Fluxp2iSfd_y - fabs(SpeediSfd_y) * vector(rk3Sfd_y, i + 2, j, k));
						LLFn_p2iSfd_z = 0.5 * (Fluxp2iSfd_z - fabs(SpeediSfd_z) * vector(rk3Sfd_z, i + 2, j, k));
						LLFn_p2iBfu_x = 0.5 * (Fluxp2iBfu_x - fabs(SpeediBfu_x) * vector(rk3Bfu_x, i + 2, j, k));
						LLFn_p2iBfu_y = 0.5 * (Fluxp2iBfu_y - fabs(SpeediBfu_y) * vector(rk3Bfu_y, i + 2, j, k));
						LLFn_p2iBfu_z = 0.5 * (Fluxp2iBfu_z - fabs(SpeediBfu_z) * vector(rk3Bfu_z, i + 2, j, k));
						LLFn_p2iDf = 0.5 * (Fluxp2iDf - fabs(SpeediDf) * vector(rk3Df, i + 2, j, k));
						LLFn_p2iDYf = 0.5 * (Fluxp2iDYf - fabs(SpeediDYf) * vector(rk3DYf, i + 2, j, k));
						LLFn_p2itauf = 0.5 * (Fluxp2itauf - fabs(Speeditauf) * vector(rk3tauf, i + 2, j, k));
						LLFn_p2iphif = 0.5 * (Fluxp2iphif - fabs(Speediphif) * vector(rk3phif, i + 2, j, k));
						LLFp_p3iSfd_x = 0.5 * (Fluxp3iSfd_x + fabs(SpeediSfd_x) * vector(rk3Sfd_x, i + 3, j, k));
						LLFp_p3iSfd_y = 0.5 * (Fluxp3iSfd_y + fabs(SpeediSfd_y) * vector(rk3Sfd_y, i + 3, j, k));
						LLFp_p3iSfd_z = 0.5 * (Fluxp3iSfd_z + fabs(SpeediSfd_z) * vector(rk3Sfd_z, i + 3, j, k));
						LLFp_p3iBfu_x = 0.5 * (Fluxp3iBfu_x + fabs(SpeediBfu_x) * vector(rk3Bfu_x, i + 3, j, k));
						LLFp_p3iBfu_y = 0.5 * (Fluxp3iBfu_y + fabs(SpeediBfu_y) * vector(rk3Bfu_y, i + 3, j, k));
						LLFp_p3iBfu_z = 0.5 * (Fluxp3iBfu_z + fabs(SpeediBfu_z) * vector(rk3Bfu_z, i + 3, j, k));
						LLFp_p3iDf = 0.5 * (Fluxp3iDf + fabs(SpeediDf) * vector(rk3Df, i + 3, j, k));
						LLFp_p3iDYf = 0.5 * (Fluxp3iDYf + fabs(SpeediDYf) * vector(rk3DYf, i + 3, j, k));
						LLFp_p3itauf = 0.5 * (Fluxp3itauf + fabs(Speeditauf) * vector(rk3tauf, i + 3, j, k));
						LLFp_p3iphif = 0.5 * (Fluxp3iphif + fabs(Speediphif) * vector(rk3phif, i + 3, j, k));
						LLFn_p3iSfd_x = 0.5 * (Fluxp3iSfd_x - fabs(SpeediSfd_x) * vector(rk3Sfd_x, i + 3, j, k));
						LLFn_p3iSfd_y = 0.5 * (Fluxp3iSfd_y - fabs(SpeediSfd_y) * vector(rk3Sfd_y, i + 3, j, k));
						LLFn_p3iSfd_z = 0.5 * (Fluxp3iSfd_z - fabs(SpeediSfd_z) * vector(rk3Sfd_z, i + 3, j, k));
						LLFn_p3iBfu_x = 0.5 * (Fluxp3iBfu_x - fabs(SpeediBfu_x) * vector(rk3Bfu_x, i + 3, j, k));
						LLFn_p3iBfu_y = 0.5 * (Fluxp3iBfu_y - fabs(SpeediBfu_y) * vector(rk3Bfu_y, i + 3, j, k));
						LLFn_p3iBfu_z = 0.5 * (Fluxp3iBfu_z - fabs(SpeediBfu_z) * vector(rk3Bfu_z, i + 3, j, k));
						LLFn_p3iDf = 0.5 * (Fluxp3iDf - fabs(SpeediDf) * vector(rk3Df, i + 3, j, k));
						LLFn_p3iDYf = 0.5 * (Fluxp3iDYf - fabs(SpeediDYf) * vector(rk3DYf, i + 3, j, k));
						LLFn_p3itauf = 0.5 * (Fluxp3itauf - fabs(Speeditauf) * vector(rk3tauf, i + 3, j, k));
						LLFn_p3iphif = 0.5 * (Fluxp3iphif - fabs(Speediphif) * vector(rk3phif, i + 3, j, k));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_yy, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_zz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_xy, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_xz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3sqBf = vector(rk3Bfu_x, i, j, k) * rk3Bfd_x + vector(rk3Bfu_y, i, j, k) * rk3Bfd_y + vector(rk3Bfu_z, i, j, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) * vector(rk3gtd_xx, i, j, k) + (-vector(rk3gtd_yy, i, j, k) * (vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k))) + (-vector(rk3gtd_zz, i, j, k) * (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k))) + (-(vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k)) * vector(rk3gtd_xx, i, j, k)) + 2.0 * vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yy, i, j, k) - vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) - vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_yz, i, j, k) - vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_zz, i, j, k)) + vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j, k) + rk3gtu_xy * vector(rk3Sfd_y, i, j, k) + rk3gtu_xz * vector(rk3Sfd_z, i, j, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j, k) + rk3gtu_xy * vector(vfd_y, i, j, k) + rk3gtu_xz * vector(vfd_z, i, j, k));
						rk3Bfvf = vector(rk3Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk3Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk3Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_zz, i, j, k) - vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yz, i, j, k)) + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j, k) + rk3gtu_yz * vector(vfd_y, i, j, k) + rk3gtu_zz * vector(vfd_z, i, j, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j, k) + rk3gtu_yy * vector(rk3Sfd_y, i, j, k) + rk3gtu_yz * vector(rk3Sfd_z, i, j, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j, k) + rk3gtu_yz * vector(rk3Sfd_y, i, j, k) + rk3gtu_zz * vector(rk3Sfd_z, i, j, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j, k) + rk3gtu_yy * vector(vfd_y, i, j, k) + rk3gtu_yz * vector(vfd_z, i, j, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j, k) + rk3vfu_y * vector(vfd_y, i, j, k) + rk3vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j, k) * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j, k) * vector(rk3Bfu_x, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k) * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k) * rk3vfu_z + vector(rk3Bfu_z, i, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k) * vector(rk3Bfu_y, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k) * rk3vfu_y + vector(rk3Bfu_y, i, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k) * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k) * rk3vfu_z + vector(rk3Bfu_z, i, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j, k) * vector(rk3Bfu_y, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j, k));
						FluxjSfd_x = FjSfd_x_mainI(vector(rk3Sfd_x, i, j, k), vector(rk3gtd_xx, i, j, k), vector(rk3gtd_xy, i, j, k), vector(rk3gtd_xz, i, j, k), vector(rk3Betau_y, i, j, k), vector(rk3Alpha, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_yy, i, j, k), vector(TauT_yz, i, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_yx, dx, simPlat_dt, ilast, jlast);
						FluxjSfd_y = FjSfd_y_mainI(vector(rk3Sfd_y, i, j, k), vector(rk3gtd_xy, i, j, k), vector(rk3gtd_yy, i, j, k), vector(rk3gtd_yz, i, j, k), vector(rk3Betau_y, i, j, k), vector(rk3Alpha, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_yy, i, j, k), vector(TauT_yz, i, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_yy, dx, simPlat_dt, ilast, jlast);
						FluxjSfd_z = FjSfd_z_mainI(vector(rk3Sfd_z, i, j, k), vector(rk3gtd_xz, i, j, k), vector(rk3gtd_yz, i, j, k), vector(rk3gtd_zz, i, j, k), vector(rk3Betau_y, i, j, k), vector(rk3Alpha, i, j, k), vector(TauT_xy, i, j, k), vector(TauT_yy, i, j, k), vector(TauT_yz, i, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_yz, dx, simPlat_dt, ilast, jlast);
						FluxjBfu_x = FjBfu_x_mainI(vector(rk3Bfu_x, i, j, k), vector(rk3Bfu_y, i, j, k), vector(rk3phif, i, j, k), vector(rk3Betau_x, i, j, k), vector(rk3Betau_y, i, j, k), vector(rk3Alpha, i, j, k), vector(TauM_xy, i, j, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						FluxjBfu_y = FjBfu_y_mainI(vector(rk3phif, i, j, k), vector(rk3Alpha, i, j, k), rk3chi_max, rk3gtu_yy, dx, simPlat_dt, ilast, jlast);
						FluxjBfu_z = FjBfu_z_mainI(vector(rk3Bfu_y, i, j, k), vector(rk3Bfu_z, i, j, k), vector(rk3phif, i, j, k), vector(rk3Betau_y, i, j, k), vector(rk3Betau_z, i, j, k), vector(rk3Alpha, i, j, k), vector(TauM_yz, i, j, k), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxjDf = FjDf_mainI(vector(rk3Df, i, j, k), vector(rk3Betau_y, i, j, k), vector(rk3Alpha, i, j, k), vector(TauN_y, i, j, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						FluxjDYf = FjDYf_mainI(vector(rk3DYf, i, j, k), vector(rk3Betau_y, i, j, k), vector(rk3Alpha, i, j, k), vector(TauNe_y, i, j, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxjtauf = Fjtauf_mainI(vector(rk3Df, i, j, k), vector(rk3tauf, i, j, k), vector(rk3Betau_y, i, j, k), vector(rk3Alpha, i, j, k), vector(TauN_y, i, j, k), rk3sdetg, rk3Sfu_y, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxjphif = Fjphif_mainI(vector(rk3Bfu_y, i, j, k), vector(rk3phif, i, j, k), vector(rk3Betau_y, i, j, k), vector(rk3Alpha, i, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j, k) + rk3sqca - vector(sqcs, i, j, k) * rk3sqca;
						SpeedjSfd_x = fabs(fabs(vector(rk3Betau_y, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y))))));
						SpeedjSfd_y = fabs(fabs(vector(rk3Betau_y, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y))))));
						SpeedjSfd_z = fabs(fabs(vector(rk3Betau_y, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y))))));
						SpeedjBfu_x = fabs(fabs(vector(rk3Betau_y, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y))))));
						SpeedjBfu_y = fabs(fabs(vector(rk3Betau_y, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y))))));
						SpeedjBfu_z = fabs(fabs(vector(rk3Betau_y, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y))))));
						SpeedjDf = fabs(fabs(vector(rk3Betau_y, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y))))));
						SpeedjDYf = fabs(fabs(vector(rk3Betau_y, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y))))));
						Speedjtauf = fabs(fabs(vector(rk3Betau_y, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y))))));
						Speedjphif = fabs(fabs(vector(rk3Betau_y, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j - 3, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j - 3, k) * vector(rk3Bfu_x, i, j - 3, k) + vector(rk3gtd_yy, i, j - 3, k) * vector(rk3Bfu_y, i, j - 3, k) + vector(rk3gtd_yz, i, j - 3, k) * vector(rk3Bfu_z, i, j - 3, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j - 3, k) * vector(rk3Bfu_x, i, j - 3, k) + vector(rk3gtd_yz, i, j - 3, k) * vector(rk3Bfu_y, i, j - 3, k) + vector(rk3gtd_zz, i, j - 3, k) * vector(rk3Bfu_z, i, j - 3, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j - 3, k) * vector(rk3Bfu_x, i, j - 3, k) + vector(rk3gtd_xy, i, j - 3, k) * vector(rk3Bfu_y, i, j - 3, k) + vector(rk3gtd_xz, i, j - 3, k) * vector(rk3Bfu_z, i, j - 3, k));
						rk3sqBf = vector(rk3Bfu_x, i, j - 3, k) * rk3Bfd_x + vector(rk3Bfu_y, i, j - 3, k) * rk3Bfd_y + vector(rk3Bfu_z, i, j - 3, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j - 3, k) * vector(rk3gtd_zz, i, j - 3, k) * vector(rk3gtd_xx, i, j - 3, k) + (-vector(rk3gtd_yy, i, j - 3, k) * (vector(rk3gtd_xz, i, j - 3, k) * vector(rk3gtd_xz, i, j - 3, k))) + (-vector(rk3gtd_zz, i, j - 3, k) * (vector(rk3gtd_xy, i, j - 3, k) * vector(rk3gtd_xy, i, j - 3, k))) + (-(vector(rk3gtd_yz, i, j - 3, k) * vector(rk3gtd_yz, i, j - 3, k)) * vector(rk3gtd_xx, i, j - 3, k)) + 2.0 * vector(rk3gtd_yz, i, j - 3, k) * vector(rk3gtd_xy, i, j - 3, k) * vector(rk3gtd_xz, i, j - 3, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j - 3, k) * vector(rk3gtd_yy, i, j - 3, k) - vector(rk3gtd_xy, i, j - 3, k) * vector(rk3gtd_xy, i, j - 3, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j - 3, k) * vector(rk3gtd_zz, i, j - 3, k) - vector(rk3gtd_yz, i, j - 3, k) * vector(rk3gtd_yz, i, j - 3, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j - 3, k) * vector(rk3gtd_yz, i, j - 3, k) - vector(rk3gtd_yy, i, j - 3, k) * vector(rk3gtd_xz, i, j - 3, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j - 3, k) * vector(rk3gtd_zz, i, j - 3, k)) + vector(rk3gtd_yz, i, j - 3, k) * vector(rk3gtd_xz, i, j - 3, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j - 3, k) + rk3gtu_xy * vector(rk3Sfd_y, i, j - 3, k) + rk3gtu_xz * vector(rk3Sfd_z, i, j - 3, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j - 3, k) + rk3gtu_xy * vector(vfd_y, i, j - 3, k) + rk3gtu_xz * vector(vfd_z, i, j - 3, k));
						rk3Bfvf = vector(rk3Bfu_x, i, j - 3, k) * vector(vfd_x, i, j - 3, k) + vector(rk3Bfu_y, i, j - 3, k) * vector(vfd_y, i, j - 3, k) + vector(rk3Bfu_z, i, j - 3, k) * vector(vfd_z, i, j - 3, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j - 3, k) * vector(rk3gtd_zz, i, j - 3, k) - vector(rk3gtd_xz, i, j - 3, k) * vector(rk3gtd_xz, i, j - 3, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j - 3, k) * vector(rk3gtd_yz, i, j - 3, k)) + vector(rk3gtd_xy, i, j - 3, k) * vector(rk3gtd_xz, i, j - 3, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j - 3, k) + rk3gtu_yz * vector(vfd_y, i, j - 3, k) + rk3gtu_zz * vector(vfd_z, i, j - 3, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j - 3, k) + rk3gtu_yy * vector(rk3Sfd_y, i, j - 3, k) + rk3gtu_yz * vector(rk3Sfd_z, i, j - 3, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j - 3, k) + rk3gtu_yz * vector(rk3Sfd_y, i, j - 3, k) + rk3gtu_zz * vector(rk3Sfd_z, i, j - 3, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j - 3, k) + rk3gtu_yy * vector(vfd_y, i, j - 3, k) + rk3gtu_yz * vector(vfd_z, i, j - 3, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j - 3, k) + rk3vfu_y * vector(vfd_y, i, j - 3, k) + rk3vfu_z * vector(vfd_z, i, j - 3, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j - 3, k) * vector(rk3Bfu_z, i, j - 3, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j - 3, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j - 3, k) * vector(rk3Bfu_x, i, j - 3, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j - 3, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j - 3, k) * vector(rk3Bfu_z, i, j - 3, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j - 3, k) * rk3vfu_z + vector(rk3Bfu_z, i, j - 3, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j - 3, k) * vector(rk3Bfu_y, i, j - 3, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j - 3, k) * rk3vfu_y + vector(rk3Bfu_y, i, j - 3, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j - 3, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j - 3, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j - 3, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j - 3, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j - 3, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j - 3, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j - 3, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j - 3, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j - 3, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j - 3, k) * vector(rk3Bfu_z, i, j - 3, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j - 3, k) * rk3vfu_z + vector(rk3Bfu_z, i, j - 3, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j - 3, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j - 3, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j - 3, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j - 3, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j - 3, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j - 3, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j - 3, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j - 3, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j - 3, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j - 3, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j - 3, k) * vector(rk3Bfu_y, i, j - 3, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j - 3, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j - 3, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j - 3, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j - 3, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j - 3, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j - 3, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j - 3, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j - 3, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j - 3, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j - 3, k));
						Fluxm3jSfd_x = FjSfd_x_mainI(vector(rk3Sfd_x, i, j - 3, k), vector(rk3gtd_xx, i, j - 3, k), vector(rk3gtd_xy, i, j - 3, k), vector(rk3gtd_xz, i, j - 3, k), vector(rk3Betau_y, i, j - 3, k), vector(rk3Alpha, i, j - 3, k), vector(TauT_xy, i, j - 3, k), vector(TauT_yy, i, j - 3, k), vector(TauT_yz, i, j - 3, k), rk3inv_chi, rk3sdetg, rk3Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxm3jSfd_y = FjSfd_y_mainI(vector(rk3Sfd_y, i, j - 3, k), vector(rk3gtd_xy, i, j - 3, k), vector(rk3gtd_yy, i, j - 3, k), vector(rk3gtd_yz, i, j - 3, k), vector(rk3Betau_y, i, j - 3, k), vector(rk3Alpha, i, j - 3, k), vector(TauT_xy, i, j - 3, k), vector(TauT_yy, i, j - 3, k), vector(TauT_yz, i, j - 3, k), rk3inv_chi, rk3sdetg, rk3Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm3jSfd_z = FjSfd_z_mainI(vector(rk3Sfd_z, i, j - 3, k), vector(rk3gtd_xz, i, j - 3, k), vector(rk3gtd_yz, i, j - 3, k), vector(rk3gtd_zz, i, j - 3, k), vector(rk3Betau_y, i, j - 3, k), vector(rk3Alpha, i, j - 3, k), vector(TauT_xy, i, j - 3, k), vector(TauT_yy, i, j - 3, k), vector(TauT_yz, i, j - 3, k), rk3inv_chi, rk3sdetg, rk3Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxm3jBfu_x = FjBfu_x_mainI(vector(rk3Bfu_x, i, j - 3, k), vector(rk3Bfu_y, i, j - 3, k), vector(rk3phif, i, j - 3, k), vector(rk3Betau_x, i, j - 3, k), vector(rk3Betau_y, i, j - 3, k), vector(rk3Alpha, i, j - 3, k), vector(TauM_xy, i, j - 3, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3jBfu_y = FjBfu_y_mainI(vector(rk3phif, i, j - 3, k), vector(rk3Alpha, i, j - 3, k), rk3chi_max, rk3gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm3jBfu_z = FjBfu_z_mainI(vector(rk3Bfu_y, i, j - 3, k), vector(rk3Bfu_z, i, j - 3, k), vector(rk3phif, i, j - 3, k), vector(rk3Betau_y, i, j - 3, k), vector(rk3Betau_z, i, j - 3, k), vector(rk3Alpha, i, j - 3, k), vector(TauM_yz, i, j - 3, k), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3jDf = FjDf_mainI(vector(rk3Df, i, j - 3, k), vector(rk3Betau_y, i, j - 3, k), vector(rk3Alpha, i, j - 3, k), vector(TauN_y, i, j - 3, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3jDYf = FjDYf_mainI(vector(rk3DYf, i, j - 3, k), vector(rk3Betau_y, i, j - 3, k), vector(rk3Alpha, i, j - 3, k), vector(TauNe_y, i, j - 3, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3jtauf = Fjtauf_mainI(vector(rk3Df, i, j - 3, k), vector(rk3tauf, i, j - 3, k), vector(rk3Betau_y, i, j - 3, k), vector(rk3Alpha, i, j - 3, k), vector(TauN_y, i, j - 3, k), rk3sdetg, rk3Sfu_y, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm3jphif = Fjphif_mainI(vector(rk3Bfu_y, i, j - 3, k), vector(rk3phif, i, j - 3, k), vector(rk3Betau_y, i, j - 3, k), vector(rk3Alpha, i, j - 3, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j - 3, k) * (1.0 + vector(epsf, i, j - 3, k)) + vector(pf, i, j - 3, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j - 3, k) + rk3sqca - vector(sqcs, i, j - 3, k) * rk3sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk3Betau_y, i, j - 3, k)) + vector(rk3Alpha, i, j - 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk3Betau_y, i, j - 3, k)) + vector(rk3Alpha, i, j - 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk3Betau_y, i, j - 3, k)) + vector(rk3Alpha, i, j - 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk3Betau_y, i, j - 3, k)) + vector(rk3Alpha, i, j - 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk3Betau_y, i, j - 3, k)) + vector(rk3Alpha, i, j - 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk3Betau_y, i, j - 3, k)) + vector(rk3Alpha, i, j - 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk3Betau_y, i, j - 3, k)) + vector(rk3Alpha, i, j - 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk3Betau_y, i, j - 3, k)) + vector(rk3Alpha, i, j - 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk3Betau_y, i, j - 3, k)) + vector(rk3Alpha, i, j - 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk3Betau_y, i, j - 3, k)) + vector(rk3Alpha, i, j - 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j - 2, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j - 2, k) * vector(rk3Bfu_x, i, j - 2, k) + vector(rk3gtd_yy, i, j - 2, k) * vector(rk3Bfu_y, i, j - 2, k) + vector(rk3gtd_yz, i, j - 2, k) * vector(rk3Bfu_z, i, j - 2, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j - 2, k) * vector(rk3Bfu_x, i, j - 2, k) + vector(rk3gtd_yz, i, j - 2, k) * vector(rk3Bfu_y, i, j - 2, k) + vector(rk3gtd_zz, i, j - 2, k) * vector(rk3Bfu_z, i, j - 2, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j - 2, k) * vector(rk3Bfu_x, i, j - 2, k) + vector(rk3gtd_xy, i, j - 2, k) * vector(rk3Bfu_y, i, j - 2, k) + vector(rk3gtd_xz, i, j - 2, k) * vector(rk3Bfu_z, i, j - 2, k));
						rk3sqBf = vector(rk3Bfu_x, i, j - 2, k) * rk3Bfd_x + vector(rk3Bfu_y, i, j - 2, k) * rk3Bfd_y + vector(rk3Bfu_z, i, j - 2, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j - 2, k) * vector(rk3gtd_zz, i, j - 2, k) * vector(rk3gtd_xx, i, j - 2, k) + (-vector(rk3gtd_yy, i, j - 2, k) * (vector(rk3gtd_xz, i, j - 2, k) * vector(rk3gtd_xz, i, j - 2, k))) + (-vector(rk3gtd_zz, i, j - 2, k) * (vector(rk3gtd_xy, i, j - 2, k) * vector(rk3gtd_xy, i, j - 2, k))) + (-(vector(rk3gtd_yz, i, j - 2, k) * vector(rk3gtd_yz, i, j - 2, k)) * vector(rk3gtd_xx, i, j - 2, k)) + 2.0 * vector(rk3gtd_yz, i, j - 2, k) * vector(rk3gtd_xy, i, j - 2, k) * vector(rk3gtd_xz, i, j - 2, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j - 2, k) * vector(rk3gtd_yy, i, j - 2, k) - vector(rk3gtd_xy, i, j - 2, k) * vector(rk3gtd_xy, i, j - 2, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j - 2, k) * vector(rk3gtd_zz, i, j - 2, k) - vector(rk3gtd_yz, i, j - 2, k) * vector(rk3gtd_yz, i, j - 2, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j - 2, k) * vector(rk3gtd_yz, i, j - 2, k) - vector(rk3gtd_yy, i, j - 2, k) * vector(rk3gtd_xz, i, j - 2, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j - 2, k) * vector(rk3gtd_zz, i, j - 2, k)) + vector(rk3gtd_yz, i, j - 2, k) * vector(rk3gtd_xz, i, j - 2, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j - 2, k) + rk3gtu_xy * vector(rk3Sfd_y, i, j - 2, k) + rk3gtu_xz * vector(rk3Sfd_z, i, j - 2, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j - 2, k) + rk3gtu_xy * vector(vfd_y, i, j - 2, k) + rk3gtu_xz * vector(vfd_z, i, j - 2, k));
						rk3Bfvf = vector(rk3Bfu_x, i, j - 2, k) * vector(vfd_x, i, j - 2, k) + vector(rk3Bfu_y, i, j - 2, k) * vector(vfd_y, i, j - 2, k) + vector(rk3Bfu_z, i, j - 2, k) * vector(vfd_z, i, j - 2, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j - 2, k) * vector(rk3gtd_zz, i, j - 2, k) - vector(rk3gtd_xz, i, j - 2, k) * vector(rk3gtd_xz, i, j - 2, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j - 2, k) * vector(rk3gtd_yz, i, j - 2, k)) + vector(rk3gtd_xy, i, j - 2, k) * vector(rk3gtd_xz, i, j - 2, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j - 2, k) + rk3gtu_yz * vector(vfd_y, i, j - 2, k) + rk3gtu_zz * vector(vfd_z, i, j - 2, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j - 2, k) + rk3gtu_yy * vector(rk3Sfd_y, i, j - 2, k) + rk3gtu_yz * vector(rk3Sfd_z, i, j - 2, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j - 2, k) + rk3gtu_yz * vector(rk3Sfd_y, i, j - 2, k) + rk3gtu_zz * vector(rk3Sfd_z, i, j - 2, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j - 2, k) + rk3gtu_yy * vector(vfd_y, i, j - 2, k) + rk3gtu_yz * vector(vfd_z, i, j - 2, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j - 2, k) + rk3vfu_y * vector(vfd_y, i, j - 2, k) + rk3vfu_z * vector(vfd_z, i, j - 2, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j - 2, k) * vector(rk3Bfu_z, i, j - 2, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j - 2, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j - 2, k) * vector(rk3Bfu_x, i, j - 2, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j - 2, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j - 2, k) * vector(rk3Bfu_z, i, j - 2, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j - 2, k) * rk3vfu_z + vector(rk3Bfu_z, i, j - 2, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j - 2, k) * vector(rk3Bfu_y, i, j - 2, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j - 2, k) * rk3vfu_y + vector(rk3Bfu_y, i, j - 2, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j - 2, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j - 2, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j - 2, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j - 2, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j - 2, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j - 2, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j - 2, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j - 2, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j - 2, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j - 2, k) * vector(rk3Bfu_z, i, j - 2, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j - 2, k) * rk3vfu_z + vector(rk3Bfu_z, i, j - 2, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j - 2, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j - 2, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j - 2, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j - 2, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j - 2, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j - 2, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j - 2, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j - 2, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j - 2, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j - 2, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j - 2, k) * vector(rk3Bfu_y, i, j - 2, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j - 2, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j - 2, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j - 2, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j - 2, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j - 2, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j - 2, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j - 2, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j - 2, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j - 2, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j - 2, k));
						Fluxm2jSfd_x = FjSfd_x_mainI(vector(rk3Sfd_x, i, j - 2, k), vector(rk3gtd_xx, i, j - 2, k), vector(rk3gtd_xy, i, j - 2, k), vector(rk3gtd_xz, i, j - 2, k), vector(rk3Betau_y, i, j - 2, k), vector(rk3Alpha, i, j - 2, k), vector(TauT_xy, i, j - 2, k), vector(TauT_yy, i, j - 2, k), vector(TauT_yz, i, j - 2, k), rk3inv_chi, rk3sdetg, rk3Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxm2jSfd_y = FjSfd_y_mainI(vector(rk3Sfd_y, i, j - 2, k), vector(rk3gtd_xy, i, j - 2, k), vector(rk3gtd_yy, i, j - 2, k), vector(rk3gtd_yz, i, j - 2, k), vector(rk3Betau_y, i, j - 2, k), vector(rk3Alpha, i, j - 2, k), vector(TauT_xy, i, j - 2, k), vector(TauT_yy, i, j - 2, k), vector(TauT_yz, i, j - 2, k), rk3inv_chi, rk3sdetg, rk3Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm2jSfd_z = FjSfd_z_mainI(vector(rk3Sfd_z, i, j - 2, k), vector(rk3gtd_xz, i, j - 2, k), vector(rk3gtd_yz, i, j - 2, k), vector(rk3gtd_zz, i, j - 2, k), vector(rk3Betau_y, i, j - 2, k), vector(rk3Alpha, i, j - 2, k), vector(TauT_xy, i, j - 2, k), vector(TauT_yy, i, j - 2, k), vector(TauT_yz, i, j - 2, k), rk3inv_chi, rk3sdetg, rk3Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxm2jBfu_x = FjBfu_x_mainI(vector(rk3Bfu_x, i, j - 2, k), vector(rk3Bfu_y, i, j - 2, k), vector(rk3phif, i, j - 2, k), vector(rk3Betau_x, i, j - 2, k), vector(rk3Betau_y, i, j - 2, k), vector(rk3Alpha, i, j - 2, k), vector(TauM_xy, i, j - 2, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2jBfu_y = FjBfu_y_mainI(vector(rk3phif, i, j - 2, k), vector(rk3Alpha, i, j - 2, k), rk3chi_max, rk3gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm2jBfu_z = FjBfu_z_mainI(vector(rk3Bfu_y, i, j - 2, k), vector(rk3Bfu_z, i, j - 2, k), vector(rk3phif, i, j - 2, k), vector(rk3Betau_y, i, j - 2, k), vector(rk3Betau_z, i, j - 2, k), vector(rk3Alpha, i, j - 2, k), vector(TauM_yz, i, j - 2, k), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2jDf = FjDf_mainI(vector(rk3Df, i, j - 2, k), vector(rk3Betau_y, i, j - 2, k), vector(rk3Alpha, i, j - 2, k), vector(TauN_y, i, j - 2, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2jDYf = FjDYf_mainI(vector(rk3DYf, i, j - 2, k), vector(rk3Betau_y, i, j - 2, k), vector(rk3Alpha, i, j - 2, k), vector(TauNe_y, i, j - 2, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2jtauf = Fjtauf_mainI(vector(rk3Df, i, j - 2, k), vector(rk3tauf, i, j - 2, k), vector(rk3Betau_y, i, j - 2, k), vector(rk3Alpha, i, j - 2, k), vector(TauN_y, i, j - 2, k), rk3sdetg, rk3Sfu_y, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm2jphif = Fjphif_mainI(vector(rk3Bfu_y, i, j - 2, k), vector(rk3phif, i, j - 2, k), vector(rk3Betau_y, i, j - 2, k), vector(rk3Alpha, i, j - 2, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j - 2, k) * (1.0 + vector(epsf, i, j - 2, k)) + vector(pf, i, j - 2, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j - 2, k) + rk3sqca - vector(sqcs, i, j - 2, k) * rk3sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk3Betau_y, i, j - 2, k)) + vector(rk3Alpha, i, j - 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk3Betau_y, i, j - 2, k)) + vector(rk3Alpha, i, j - 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk3Betau_y, i, j - 2, k)) + vector(rk3Alpha, i, j - 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk3Betau_y, i, j - 2, k)) + vector(rk3Alpha, i, j - 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk3Betau_y, i, j - 2, k)) + vector(rk3Alpha, i, j - 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk3Betau_y, i, j - 2, k)) + vector(rk3Alpha, i, j - 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk3Betau_y, i, j - 2, k)) + vector(rk3Alpha, i, j - 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk3Betau_y, i, j - 2, k)) + vector(rk3Alpha, i, j - 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk3Betau_y, i, j - 2, k)) + vector(rk3Alpha, i, j - 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk3Betau_y, i, j - 2, k)) + vector(rk3Alpha, i, j - 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j - 1, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j - 1, k) * vector(rk3Bfu_x, i, j - 1, k) + vector(rk3gtd_yy, i, j - 1, k) * vector(rk3Bfu_y, i, j - 1, k) + vector(rk3gtd_yz, i, j - 1, k) * vector(rk3Bfu_z, i, j - 1, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j - 1, k) * vector(rk3Bfu_x, i, j - 1, k) + vector(rk3gtd_yz, i, j - 1, k) * vector(rk3Bfu_y, i, j - 1, k) + vector(rk3gtd_zz, i, j - 1, k) * vector(rk3Bfu_z, i, j - 1, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j - 1, k) * vector(rk3Bfu_x, i, j - 1, k) + vector(rk3gtd_xy, i, j - 1, k) * vector(rk3Bfu_y, i, j - 1, k) + vector(rk3gtd_xz, i, j - 1, k) * vector(rk3Bfu_z, i, j - 1, k));
						rk3sqBf = vector(rk3Bfu_x, i, j - 1, k) * rk3Bfd_x + vector(rk3Bfu_y, i, j - 1, k) * rk3Bfd_y + vector(rk3Bfu_z, i, j - 1, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j - 1, k) * vector(rk3gtd_zz, i, j - 1, k) * vector(rk3gtd_xx, i, j - 1, k) + (-vector(rk3gtd_yy, i, j - 1, k) * (vector(rk3gtd_xz, i, j - 1, k) * vector(rk3gtd_xz, i, j - 1, k))) + (-vector(rk3gtd_zz, i, j - 1, k) * (vector(rk3gtd_xy, i, j - 1, k) * vector(rk3gtd_xy, i, j - 1, k))) + (-(vector(rk3gtd_yz, i, j - 1, k) * vector(rk3gtd_yz, i, j - 1, k)) * vector(rk3gtd_xx, i, j - 1, k)) + 2.0 * vector(rk3gtd_yz, i, j - 1, k) * vector(rk3gtd_xy, i, j - 1, k) * vector(rk3gtd_xz, i, j - 1, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j - 1, k) * vector(rk3gtd_yy, i, j - 1, k) - vector(rk3gtd_xy, i, j - 1, k) * vector(rk3gtd_xy, i, j - 1, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j - 1, k) * vector(rk3gtd_zz, i, j - 1, k) - vector(rk3gtd_yz, i, j - 1, k) * vector(rk3gtd_yz, i, j - 1, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j - 1, k) * vector(rk3gtd_yz, i, j - 1, k) - vector(rk3gtd_yy, i, j - 1, k) * vector(rk3gtd_xz, i, j - 1, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j - 1, k) * vector(rk3gtd_zz, i, j - 1, k)) + vector(rk3gtd_yz, i, j - 1, k) * vector(rk3gtd_xz, i, j - 1, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j - 1, k) + rk3gtu_xy * vector(rk3Sfd_y, i, j - 1, k) + rk3gtu_xz * vector(rk3Sfd_z, i, j - 1, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j - 1, k) + rk3gtu_xy * vector(vfd_y, i, j - 1, k) + rk3gtu_xz * vector(vfd_z, i, j - 1, k));
						rk3Bfvf = vector(rk3Bfu_x, i, j - 1, k) * vector(vfd_x, i, j - 1, k) + vector(rk3Bfu_y, i, j - 1, k) * vector(vfd_y, i, j - 1, k) + vector(rk3Bfu_z, i, j - 1, k) * vector(vfd_z, i, j - 1, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j - 1, k) * vector(rk3gtd_zz, i, j - 1, k) - vector(rk3gtd_xz, i, j - 1, k) * vector(rk3gtd_xz, i, j - 1, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j - 1, k) * vector(rk3gtd_yz, i, j - 1, k)) + vector(rk3gtd_xy, i, j - 1, k) * vector(rk3gtd_xz, i, j - 1, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j - 1, k) + rk3gtu_yz * vector(vfd_y, i, j - 1, k) + rk3gtu_zz * vector(vfd_z, i, j - 1, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j - 1, k) + rk3gtu_yy * vector(rk3Sfd_y, i, j - 1, k) + rk3gtu_yz * vector(rk3Sfd_z, i, j - 1, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j - 1, k) + rk3gtu_yz * vector(rk3Sfd_y, i, j - 1, k) + rk3gtu_zz * vector(rk3Sfd_z, i, j - 1, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j - 1, k) + rk3gtu_yy * vector(vfd_y, i, j - 1, k) + rk3gtu_yz * vector(vfd_z, i, j - 1, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j - 1, k) + rk3vfu_y * vector(vfd_y, i, j - 1, k) + rk3vfu_z * vector(vfd_z, i, j - 1, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j - 1, k) * vector(rk3Bfu_z, i, j - 1, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j - 1, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j - 1, k) * vector(rk3Bfu_x, i, j - 1, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j - 1, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j - 1, k) * vector(rk3Bfu_z, i, j - 1, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j - 1, k) * rk3vfu_z + vector(rk3Bfu_z, i, j - 1, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j - 1, k) * vector(rk3Bfu_y, i, j - 1, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j - 1, k) * rk3vfu_y + vector(rk3Bfu_y, i, j - 1, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j - 1, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j - 1, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j - 1, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j - 1, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j - 1, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j - 1, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j - 1, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j - 1, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j - 1, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j - 1, k) * vector(rk3Bfu_z, i, j - 1, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j - 1, k) * rk3vfu_z + vector(rk3Bfu_z, i, j - 1, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j - 1, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j - 1, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j - 1, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j - 1, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j - 1, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j - 1, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j - 1, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j - 1, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j - 1, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j - 1, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j - 1, k) * vector(rk3Bfu_y, i, j - 1, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j - 1, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j - 1, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j - 1, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j - 1, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j - 1, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j - 1, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j - 1, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j - 1, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j - 1, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j - 1, k));
						Fluxm1jSfd_x = FjSfd_x_mainI(vector(rk3Sfd_x, i, j - 1, k), vector(rk3gtd_xx, i, j - 1, k), vector(rk3gtd_xy, i, j - 1, k), vector(rk3gtd_xz, i, j - 1, k), vector(rk3Betau_y, i, j - 1, k), vector(rk3Alpha, i, j - 1, k), vector(TauT_xy, i, j - 1, k), vector(TauT_yy, i, j - 1, k), vector(TauT_yz, i, j - 1, k), rk3inv_chi, rk3sdetg, rk3Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxm1jSfd_y = FjSfd_y_mainI(vector(rk3Sfd_y, i, j - 1, k), vector(rk3gtd_xy, i, j - 1, k), vector(rk3gtd_yy, i, j - 1, k), vector(rk3gtd_yz, i, j - 1, k), vector(rk3Betau_y, i, j - 1, k), vector(rk3Alpha, i, j - 1, k), vector(TauT_xy, i, j - 1, k), vector(TauT_yy, i, j - 1, k), vector(TauT_yz, i, j - 1, k), rk3inv_chi, rk3sdetg, rk3Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm1jSfd_z = FjSfd_z_mainI(vector(rk3Sfd_z, i, j - 1, k), vector(rk3gtd_xz, i, j - 1, k), vector(rk3gtd_yz, i, j - 1, k), vector(rk3gtd_zz, i, j - 1, k), vector(rk3Betau_y, i, j - 1, k), vector(rk3Alpha, i, j - 1, k), vector(TauT_xy, i, j - 1, k), vector(TauT_yy, i, j - 1, k), vector(TauT_yz, i, j - 1, k), rk3inv_chi, rk3sdetg, rk3Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxm1jBfu_x = FjBfu_x_mainI(vector(rk3Bfu_x, i, j - 1, k), vector(rk3Bfu_y, i, j - 1, k), vector(rk3phif, i, j - 1, k), vector(rk3Betau_x, i, j - 1, k), vector(rk3Betau_y, i, j - 1, k), vector(rk3Alpha, i, j - 1, k), vector(TauM_xy, i, j - 1, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1jBfu_y = FjBfu_y_mainI(vector(rk3phif, i, j - 1, k), vector(rk3Alpha, i, j - 1, k), rk3chi_max, rk3gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxm1jBfu_z = FjBfu_z_mainI(vector(rk3Bfu_y, i, j - 1, k), vector(rk3Bfu_z, i, j - 1, k), vector(rk3phif, i, j - 1, k), vector(rk3Betau_y, i, j - 1, k), vector(rk3Betau_z, i, j - 1, k), vector(rk3Alpha, i, j - 1, k), vector(TauM_yz, i, j - 1, k), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1jDf = FjDf_mainI(vector(rk3Df, i, j - 1, k), vector(rk3Betau_y, i, j - 1, k), vector(rk3Alpha, i, j - 1, k), vector(TauN_y, i, j - 1, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1jDYf = FjDYf_mainI(vector(rk3DYf, i, j - 1, k), vector(rk3Betau_y, i, j - 1, k), vector(rk3Alpha, i, j - 1, k), vector(TauNe_y, i, j - 1, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1jtauf = Fjtauf_mainI(vector(rk3Df, i, j - 1, k), vector(rk3tauf, i, j - 1, k), vector(rk3Betau_y, i, j - 1, k), vector(rk3Alpha, i, j - 1, k), vector(TauN_y, i, j - 1, k), rk3sdetg, rk3Sfu_y, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxm1jphif = Fjphif_mainI(vector(rk3Bfu_y, i, j - 1, k), vector(rk3phif, i, j - 1, k), vector(rk3Betau_y, i, j - 1, k), vector(rk3Alpha, i, j - 1, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j - 1, k) * (1.0 + vector(epsf, i, j - 1, k)) + vector(pf, i, j - 1, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j - 1, k) + rk3sqca - vector(sqcs, i, j - 1, k) * rk3sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk3Betau_y, i, j - 1, k)) + vector(rk3Alpha, i, j - 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk3Betau_y, i, j - 1, k)) + vector(rk3Alpha, i, j - 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk3Betau_y, i, j - 1, k)) + vector(rk3Alpha, i, j - 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk3Betau_y, i, j - 1, k)) + vector(rk3Alpha, i, j - 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk3Betau_y, i, j - 1, k)) + vector(rk3Alpha, i, j - 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk3Betau_y, i, j - 1, k)) + vector(rk3Alpha, i, j - 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk3Betau_y, i, j - 1, k)) + vector(rk3Alpha, i, j - 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk3Betau_y, i, j - 1, k)) + vector(rk3Alpha, i, j - 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk3Betau_y, i, j - 1, k)) + vector(rk3Alpha, i, j - 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk3Betau_y, i, j - 1, k)) + vector(rk3Alpha, i, j - 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j + 1, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j + 1, k) * vector(rk3Bfu_x, i, j + 1, k) + vector(rk3gtd_yy, i, j + 1, k) * vector(rk3Bfu_y, i, j + 1, k) + vector(rk3gtd_yz, i, j + 1, k) * vector(rk3Bfu_z, i, j + 1, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j + 1, k) * vector(rk3Bfu_x, i, j + 1, k) + vector(rk3gtd_yz, i, j + 1, k) * vector(rk3Bfu_y, i, j + 1, k) + vector(rk3gtd_zz, i, j + 1, k) * vector(rk3Bfu_z, i, j + 1, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j + 1, k) * vector(rk3Bfu_x, i, j + 1, k) + vector(rk3gtd_xy, i, j + 1, k) * vector(rk3Bfu_y, i, j + 1, k) + vector(rk3gtd_xz, i, j + 1, k) * vector(rk3Bfu_z, i, j + 1, k));
						rk3sqBf = vector(rk3Bfu_x, i, j + 1, k) * rk3Bfd_x + vector(rk3Bfu_y, i, j + 1, k) * rk3Bfd_y + vector(rk3Bfu_z, i, j + 1, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j + 1, k) * vector(rk3gtd_zz, i, j + 1, k) * vector(rk3gtd_xx, i, j + 1, k) + (-vector(rk3gtd_yy, i, j + 1, k) * (vector(rk3gtd_xz, i, j + 1, k) * vector(rk3gtd_xz, i, j + 1, k))) + (-vector(rk3gtd_zz, i, j + 1, k) * (vector(rk3gtd_xy, i, j + 1, k) * vector(rk3gtd_xy, i, j + 1, k))) + (-(vector(rk3gtd_yz, i, j + 1, k) * vector(rk3gtd_yz, i, j + 1, k)) * vector(rk3gtd_xx, i, j + 1, k)) + 2.0 * vector(rk3gtd_yz, i, j + 1, k) * vector(rk3gtd_xy, i, j + 1, k) * vector(rk3gtd_xz, i, j + 1, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j + 1, k) * vector(rk3gtd_yy, i, j + 1, k) - vector(rk3gtd_xy, i, j + 1, k) * vector(rk3gtd_xy, i, j + 1, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j + 1, k) * vector(rk3gtd_zz, i, j + 1, k) - vector(rk3gtd_yz, i, j + 1, k) * vector(rk3gtd_yz, i, j + 1, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j + 1, k) * vector(rk3gtd_yz, i, j + 1, k) - vector(rk3gtd_yy, i, j + 1, k) * vector(rk3gtd_xz, i, j + 1, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j + 1, k) * vector(rk3gtd_zz, i, j + 1, k)) + vector(rk3gtd_yz, i, j + 1, k) * vector(rk3gtd_xz, i, j + 1, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j + 1, k) + rk3gtu_xy * vector(rk3Sfd_y, i, j + 1, k) + rk3gtu_xz * vector(rk3Sfd_z, i, j + 1, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j + 1, k) + rk3gtu_xy * vector(vfd_y, i, j + 1, k) + rk3gtu_xz * vector(vfd_z, i, j + 1, k));
						rk3Bfvf = vector(rk3Bfu_x, i, j + 1, k) * vector(vfd_x, i, j + 1, k) + vector(rk3Bfu_y, i, j + 1, k) * vector(vfd_y, i, j + 1, k) + vector(rk3Bfu_z, i, j + 1, k) * vector(vfd_z, i, j + 1, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j + 1, k) * vector(rk3gtd_zz, i, j + 1, k) - vector(rk3gtd_xz, i, j + 1, k) * vector(rk3gtd_xz, i, j + 1, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j + 1, k) * vector(rk3gtd_yz, i, j + 1, k)) + vector(rk3gtd_xy, i, j + 1, k) * vector(rk3gtd_xz, i, j + 1, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j + 1, k) + rk3gtu_yz * vector(vfd_y, i, j + 1, k) + rk3gtu_zz * vector(vfd_z, i, j + 1, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j + 1, k) + rk3gtu_yy * vector(rk3Sfd_y, i, j + 1, k) + rk3gtu_yz * vector(rk3Sfd_z, i, j + 1, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j + 1, k) + rk3gtu_yz * vector(rk3Sfd_y, i, j + 1, k) + rk3gtu_zz * vector(rk3Sfd_z, i, j + 1, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j + 1, k) + rk3gtu_yy * vector(vfd_y, i, j + 1, k) + rk3gtu_yz * vector(vfd_z, i, j + 1, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j + 1, k) + rk3vfu_y * vector(vfd_y, i, j + 1, k) + rk3vfu_z * vector(vfd_z, i, j + 1, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j + 1, k) * vector(rk3Bfu_z, i, j + 1, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j + 1, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j + 1, k) * vector(rk3Bfu_x, i, j + 1, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j + 1, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j + 1, k) * vector(rk3Bfu_z, i, j + 1, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j + 1, k) * rk3vfu_z + vector(rk3Bfu_z, i, j + 1, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j + 1, k) * vector(rk3Bfu_y, i, j + 1, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j + 1, k) * rk3vfu_y + vector(rk3Bfu_y, i, j + 1, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j + 1, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j + 1, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j + 1, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j + 1, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j + 1, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j + 1, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j + 1, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j + 1, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j + 1, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j + 1, k) * vector(rk3Bfu_z, i, j + 1, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j + 1, k) * rk3vfu_z + vector(rk3Bfu_z, i, j + 1, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j + 1, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j + 1, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j + 1, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j + 1, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j + 1, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j + 1, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j + 1, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j + 1, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j + 1, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j + 1, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j + 1, k) * vector(rk3Bfu_y, i, j + 1, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j + 1, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j + 1, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j + 1, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j + 1, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j + 1, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j + 1, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j + 1, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j + 1, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j + 1, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j + 1, k));
						Fluxp1jSfd_x = FjSfd_x_mainI(vector(rk3Sfd_x, i, j + 1, k), vector(rk3gtd_xx, i, j + 1, k), vector(rk3gtd_xy, i, j + 1, k), vector(rk3gtd_xz, i, j + 1, k), vector(rk3Betau_y, i, j + 1, k), vector(rk3Alpha, i, j + 1, k), vector(TauT_xy, i, j + 1, k), vector(TauT_yy, i, j + 1, k), vector(TauT_yz, i, j + 1, k), rk3inv_chi, rk3sdetg, rk3Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxp1jSfd_y = FjSfd_y_mainI(vector(rk3Sfd_y, i, j + 1, k), vector(rk3gtd_xy, i, j + 1, k), vector(rk3gtd_yy, i, j + 1, k), vector(rk3gtd_yz, i, j + 1, k), vector(rk3Betau_y, i, j + 1, k), vector(rk3Alpha, i, j + 1, k), vector(TauT_xy, i, j + 1, k), vector(TauT_yy, i, j + 1, k), vector(TauT_yz, i, j + 1, k), rk3inv_chi, rk3sdetg, rk3Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp1jSfd_z = FjSfd_z_mainI(vector(rk3Sfd_z, i, j + 1, k), vector(rk3gtd_xz, i, j + 1, k), vector(rk3gtd_yz, i, j + 1, k), vector(rk3gtd_zz, i, j + 1, k), vector(rk3Betau_y, i, j + 1, k), vector(rk3Alpha, i, j + 1, k), vector(TauT_xy, i, j + 1, k), vector(TauT_yy, i, j + 1, k), vector(TauT_yz, i, j + 1, k), rk3inv_chi, rk3sdetg, rk3Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxp1jBfu_x = FjBfu_x_mainI(vector(rk3Bfu_x, i, j + 1, k), vector(rk3Bfu_y, i, j + 1, k), vector(rk3phif, i, j + 1, k), vector(rk3Betau_x, i, j + 1, k), vector(rk3Betau_y, i, j + 1, k), vector(rk3Alpha, i, j + 1, k), vector(TauM_xy, i, j + 1, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1jBfu_y = FjBfu_y_mainI(vector(rk3phif, i, j + 1, k), vector(rk3Alpha, i, j + 1, k), rk3chi_max, rk3gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp1jBfu_z = FjBfu_z_mainI(vector(rk3Bfu_y, i, j + 1, k), vector(rk3Bfu_z, i, j + 1, k), vector(rk3phif, i, j + 1, k), vector(rk3Betau_y, i, j + 1, k), vector(rk3Betau_z, i, j + 1, k), vector(rk3Alpha, i, j + 1, k), vector(TauM_yz, i, j + 1, k), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1jDf = FjDf_mainI(vector(rk3Df, i, j + 1, k), vector(rk3Betau_y, i, j + 1, k), vector(rk3Alpha, i, j + 1, k), vector(TauN_y, i, j + 1, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1jDYf = FjDYf_mainI(vector(rk3DYf, i, j + 1, k), vector(rk3Betau_y, i, j + 1, k), vector(rk3Alpha, i, j + 1, k), vector(TauNe_y, i, j + 1, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1jtauf = Fjtauf_mainI(vector(rk3Df, i, j + 1, k), vector(rk3tauf, i, j + 1, k), vector(rk3Betau_y, i, j + 1, k), vector(rk3Alpha, i, j + 1, k), vector(TauN_y, i, j + 1, k), rk3sdetg, rk3Sfu_y, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp1jphif = Fjphif_mainI(vector(rk3Bfu_y, i, j + 1, k), vector(rk3phif, i, j + 1, k), vector(rk3Betau_y, i, j + 1, k), vector(rk3Alpha, i, j + 1, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j + 1, k) * (1.0 + vector(epsf, i, j + 1, k)) + vector(pf, i, j + 1, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j + 1, k) + rk3sqca - vector(sqcs, i, j + 1, k) * rk3sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk3Betau_y, i, j + 1, k)) + vector(rk3Alpha, i, j + 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk3Betau_y, i, j + 1, k)) + vector(rk3Alpha, i, j + 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk3Betau_y, i, j + 1, k)) + vector(rk3Alpha, i, j + 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk3Betau_y, i, j + 1, k)) + vector(rk3Alpha, i, j + 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk3Betau_y, i, j + 1, k)) + vector(rk3Alpha, i, j + 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk3Betau_y, i, j + 1, k)) + vector(rk3Alpha, i, j + 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk3Betau_y, i, j + 1, k)) + vector(rk3Alpha, i, j + 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk3Betau_y, i, j + 1, k)) + vector(rk3Alpha, i, j + 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk3Betau_y, i, j + 1, k)) + vector(rk3Alpha, i, j + 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk3Betau_y, i, j + 1, k)) + vector(rk3Alpha, i, j + 1, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j + 2, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j + 2, k) * vector(rk3Bfu_x, i, j + 2, k) + vector(rk3gtd_yy, i, j + 2, k) * vector(rk3Bfu_y, i, j + 2, k) + vector(rk3gtd_yz, i, j + 2, k) * vector(rk3Bfu_z, i, j + 2, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j + 2, k) * vector(rk3Bfu_x, i, j + 2, k) + vector(rk3gtd_yz, i, j + 2, k) * vector(rk3Bfu_y, i, j + 2, k) + vector(rk3gtd_zz, i, j + 2, k) * vector(rk3Bfu_z, i, j + 2, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j + 2, k) * vector(rk3Bfu_x, i, j + 2, k) + vector(rk3gtd_xy, i, j + 2, k) * vector(rk3Bfu_y, i, j + 2, k) + vector(rk3gtd_xz, i, j + 2, k) * vector(rk3Bfu_z, i, j + 2, k));
						rk3sqBf = vector(rk3Bfu_x, i, j + 2, k) * rk3Bfd_x + vector(rk3Bfu_y, i, j + 2, k) * rk3Bfd_y + vector(rk3Bfu_z, i, j + 2, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j + 2, k) * vector(rk3gtd_zz, i, j + 2, k) * vector(rk3gtd_xx, i, j + 2, k) + (-vector(rk3gtd_yy, i, j + 2, k) * (vector(rk3gtd_xz, i, j + 2, k) * vector(rk3gtd_xz, i, j + 2, k))) + (-vector(rk3gtd_zz, i, j + 2, k) * (vector(rk3gtd_xy, i, j + 2, k) * vector(rk3gtd_xy, i, j + 2, k))) + (-(vector(rk3gtd_yz, i, j + 2, k) * vector(rk3gtd_yz, i, j + 2, k)) * vector(rk3gtd_xx, i, j + 2, k)) + 2.0 * vector(rk3gtd_yz, i, j + 2, k) * vector(rk3gtd_xy, i, j + 2, k) * vector(rk3gtd_xz, i, j + 2, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j + 2, k) * vector(rk3gtd_yy, i, j + 2, k) - vector(rk3gtd_xy, i, j + 2, k) * vector(rk3gtd_xy, i, j + 2, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j + 2, k) * vector(rk3gtd_zz, i, j + 2, k) - vector(rk3gtd_yz, i, j + 2, k) * vector(rk3gtd_yz, i, j + 2, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j + 2, k) * vector(rk3gtd_yz, i, j + 2, k) - vector(rk3gtd_yy, i, j + 2, k) * vector(rk3gtd_xz, i, j + 2, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j + 2, k) * vector(rk3gtd_zz, i, j + 2, k)) + vector(rk3gtd_yz, i, j + 2, k) * vector(rk3gtd_xz, i, j + 2, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j + 2, k) + rk3gtu_xy * vector(rk3Sfd_y, i, j + 2, k) + rk3gtu_xz * vector(rk3Sfd_z, i, j + 2, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j + 2, k) + rk3gtu_xy * vector(vfd_y, i, j + 2, k) + rk3gtu_xz * vector(vfd_z, i, j + 2, k));
						rk3Bfvf = vector(rk3Bfu_x, i, j + 2, k) * vector(vfd_x, i, j + 2, k) + vector(rk3Bfu_y, i, j + 2, k) * vector(vfd_y, i, j + 2, k) + vector(rk3Bfu_z, i, j + 2, k) * vector(vfd_z, i, j + 2, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j + 2, k) * vector(rk3gtd_zz, i, j + 2, k) - vector(rk3gtd_xz, i, j + 2, k) * vector(rk3gtd_xz, i, j + 2, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j + 2, k) * vector(rk3gtd_yz, i, j + 2, k)) + vector(rk3gtd_xy, i, j + 2, k) * vector(rk3gtd_xz, i, j + 2, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j + 2, k) + rk3gtu_yz * vector(vfd_y, i, j + 2, k) + rk3gtu_zz * vector(vfd_z, i, j + 2, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j + 2, k) + rk3gtu_yy * vector(rk3Sfd_y, i, j + 2, k) + rk3gtu_yz * vector(rk3Sfd_z, i, j + 2, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j + 2, k) + rk3gtu_yz * vector(rk3Sfd_y, i, j + 2, k) + rk3gtu_zz * vector(rk3Sfd_z, i, j + 2, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j + 2, k) + rk3gtu_yy * vector(vfd_y, i, j + 2, k) + rk3gtu_yz * vector(vfd_z, i, j + 2, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j + 2, k) + rk3vfu_y * vector(vfd_y, i, j + 2, k) + rk3vfu_z * vector(vfd_z, i, j + 2, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j + 2, k) * vector(rk3Bfu_z, i, j + 2, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j + 2, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j + 2, k) * vector(rk3Bfu_x, i, j + 2, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j + 2, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j + 2, k) * vector(rk3Bfu_z, i, j + 2, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j + 2, k) * rk3vfu_z + vector(rk3Bfu_z, i, j + 2, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j + 2, k) * vector(rk3Bfu_y, i, j + 2, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j + 2, k) * rk3vfu_y + vector(rk3Bfu_y, i, j + 2, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j + 2, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j + 2, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j + 2, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j + 2, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j + 2, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j + 2, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j + 2, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j + 2, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j + 2, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j + 2, k) * vector(rk3Bfu_z, i, j + 2, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j + 2, k) * rk3vfu_z + vector(rk3Bfu_z, i, j + 2, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j + 2, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j + 2, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j + 2, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j + 2, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j + 2, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j + 2, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j + 2, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j + 2, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j + 2, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j + 2, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j + 2, k) * vector(rk3Bfu_y, i, j + 2, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j + 2, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j + 2, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j + 2, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j + 2, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j + 2, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j + 2, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j + 2, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j + 2, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j + 2, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j + 2, k));
						Fluxp2jSfd_x = FjSfd_x_mainI(vector(rk3Sfd_x, i, j + 2, k), vector(rk3gtd_xx, i, j + 2, k), vector(rk3gtd_xy, i, j + 2, k), vector(rk3gtd_xz, i, j + 2, k), vector(rk3Betau_y, i, j + 2, k), vector(rk3Alpha, i, j + 2, k), vector(TauT_xy, i, j + 2, k), vector(TauT_yy, i, j + 2, k), vector(TauT_yz, i, j + 2, k), rk3inv_chi, rk3sdetg, rk3Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxp2jSfd_y = FjSfd_y_mainI(vector(rk3Sfd_y, i, j + 2, k), vector(rk3gtd_xy, i, j + 2, k), vector(rk3gtd_yy, i, j + 2, k), vector(rk3gtd_yz, i, j + 2, k), vector(rk3Betau_y, i, j + 2, k), vector(rk3Alpha, i, j + 2, k), vector(TauT_xy, i, j + 2, k), vector(TauT_yy, i, j + 2, k), vector(TauT_yz, i, j + 2, k), rk3inv_chi, rk3sdetg, rk3Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp2jSfd_z = FjSfd_z_mainI(vector(rk3Sfd_z, i, j + 2, k), vector(rk3gtd_xz, i, j + 2, k), vector(rk3gtd_yz, i, j + 2, k), vector(rk3gtd_zz, i, j + 2, k), vector(rk3Betau_y, i, j + 2, k), vector(rk3Alpha, i, j + 2, k), vector(TauT_xy, i, j + 2, k), vector(TauT_yy, i, j + 2, k), vector(TauT_yz, i, j + 2, k), rk3inv_chi, rk3sdetg, rk3Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxp2jBfu_x = FjBfu_x_mainI(vector(rk3Bfu_x, i, j + 2, k), vector(rk3Bfu_y, i, j + 2, k), vector(rk3phif, i, j + 2, k), vector(rk3Betau_x, i, j + 2, k), vector(rk3Betau_y, i, j + 2, k), vector(rk3Alpha, i, j + 2, k), vector(TauM_xy, i, j + 2, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2jBfu_y = FjBfu_y_mainI(vector(rk3phif, i, j + 2, k), vector(rk3Alpha, i, j + 2, k), rk3chi_max, rk3gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp2jBfu_z = FjBfu_z_mainI(vector(rk3Bfu_y, i, j + 2, k), vector(rk3Bfu_z, i, j + 2, k), vector(rk3phif, i, j + 2, k), vector(rk3Betau_y, i, j + 2, k), vector(rk3Betau_z, i, j + 2, k), vector(rk3Alpha, i, j + 2, k), vector(TauM_yz, i, j + 2, k), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2jDf = FjDf_mainI(vector(rk3Df, i, j + 2, k), vector(rk3Betau_y, i, j + 2, k), vector(rk3Alpha, i, j + 2, k), vector(TauN_y, i, j + 2, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2jDYf = FjDYf_mainI(vector(rk3DYf, i, j + 2, k), vector(rk3Betau_y, i, j + 2, k), vector(rk3Alpha, i, j + 2, k), vector(TauNe_y, i, j + 2, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2jtauf = Fjtauf_mainI(vector(rk3Df, i, j + 2, k), vector(rk3tauf, i, j + 2, k), vector(rk3Betau_y, i, j + 2, k), vector(rk3Alpha, i, j + 2, k), vector(TauN_y, i, j + 2, k), rk3sdetg, rk3Sfu_y, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp2jphif = Fjphif_mainI(vector(rk3Bfu_y, i, j + 2, k), vector(rk3phif, i, j + 2, k), vector(rk3Betau_y, i, j + 2, k), vector(rk3Alpha, i, j + 2, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j + 2, k) * (1.0 + vector(epsf, i, j + 2, k)) + vector(pf, i, j + 2, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j + 2, k) + rk3sqca - vector(sqcs, i, j + 2, k) * rk3sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk3Betau_y, i, j + 2, k)) + vector(rk3Alpha, i, j + 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk3Betau_y, i, j + 2, k)) + vector(rk3Alpha, i, j + 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk3Betau_y, i, j + 2, k)) + vector(rk3Alpha, i, j + 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk3Betau_y, i, j + 2, k)) + vector(rk3Alpha, i, j + 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk3Betau_y, i, j + 2, k)) + vector(rk3Alpha, i, j + 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk3Betau_y, i, j + 2, k)) + vector(rk3Alpha, i, j + 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk3Betau_y, i, j + 2, k)) + vector(rk3Alpha, i, j + 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk3Betau_y, i, j + 2, k)) + vector(rk3Alpha, i, j + 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk3Betau_y, i, j + 2, k)) + vector(rk3Alpha, i, j + 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk3Betau_y, i, j + 2, k)) + vector(rk3Alpha, i, j + 2, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j + 3, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j + 3, k) * vector(rk3Bfu_x, i, j + 3, k) + vector(rk3gtd_yy, i, j + 3, k) * vector(rk3Bfu_y, i, j + 3, k) + vector(rk3gtd_yz, i, j + 3, k) * vector(rk3Bfu_z, i, j + 3, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j + 3, k) * vector(rk3Bfu_x, i, j + 3, k) + vector(rk3gtd_yz, i, j + 3, k) * vector(rk3Bfu_y, i, j + 3, k) + vector(rk3gtd_zz, i, j + 3, k) * vector(rk3Bfu_z, i, j + 3, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j + 3, k) * vector(rk3Bfu_x, i, j + 3, k) + vector(rk3gtd_xy, i, j + 3, k) * vector(rk3Bfu_y, i, j + 3, k) + vector(rk3gtd_xz, i, j + 3, k) * vector(rk3Bfu_z, i, j + 3, k));
						rk3sqBf = vector(rk3Bfu_x, i, j + 3, k) * rk3Bfd_x + vector(rk3Bfu_y, i, j + 3, k) * rk3Bfd_y + vector(rk3Bfu_z, i, j + 3, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j + 3, k) * vector(rk3gtd_zz, i, j + 3, k) * vector(rk3gtd_xx, i, j + 3, k) + (-vector(rk3gtd_yy, i, j + 3, k) * (vector(rk3gtd_xz, i, j + 3, k) * vector(rk3gtd_xz, i, j + 3, k))) + (-vector(rk3gtd_zz, i, j + 3, k) * (vector(rk3gtd_xy, i, j + 3, k) * vector(rk3gtd_xy, i, j + 3, k))) + (-(vector(rk3gtd_yz, i, j + 3, k) * vector(rk3gtd_yz, i, j + 3, k)) * vector(rk3gtd_xx, i, j + 3, k)) + 2.0 * vector(rk3gtd_yz, i, j + 3, k) * vector(rk3gtd_xy, i, j + 3, k) * vector(rk3gtd_xz, i, j + 3, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j + 3, k) * vector(rk3gtd_yy, i, j + 3, k) - vector(rk3gtd_xy, i, j + 3, k) * vector(rk3gtd_xy, i, j + 3, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j + 3, k) * vector(rk3gtd_zz, i, j + 3, k) - vector(rk3gtd_yz, i, j + 3, k) * vector(rk3gtd_yz, i, j + 3, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j + 3, k) * vector(rk3gtd_yz, i, j + 3, k) - vector(rk3gtd_yy, i, j + 3, k) * vector(rk3gtd_xz, i, j + 3, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j + 3, k) * vector(rk3gtd_zz, i, j + 3, k)) + vector(rk3gtd_yz, i, j + 3, k) * vector(rk3gtd_xz, i, j + 3, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j + 3, k) + rk3gtu_xy * vector(rk3Sfd_y, i, j + 3, k) + rk3gtu_xz * vector(rk3Sfd_z, i, j + 3, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j + 3, k) + rk3gtu_xy * vector(vfd_y, i, j + 3, k) + rk3gtu_xz * vector(vfd_z, i, j + 3, k));
						rk3Bfvf = vector(rk3Bfu_x, i, j + 3, k) * vector(vfd_x, i, j + 3, k) + vector(rk3Bfu_y, i, j + 3, k) * vector(vfd_y, i, j + 3, k) + vector(rk3Bfu_z, i, j + 3, k) * vector(vfd_z, i, j + 3, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j + 3, k) * vector(rk3gtd_zz, i, j + 3, k) - vector(rk3gtd_xz, i, j + 3, k) * vector(rk3gtd_xz, i, j + 3, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j + 3, k) * vector(rk3gtd_yz, i, j + 3, k)) + vector(rk3gtd_xy, i, j + 3, k) * vector(rk3gtd_xz, i, j + 3, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j + 3, k) + rk3gtu_yz * vector(vfd_y, i, j + 3, k) + rk3gtu_zz * vector(vfd_z, i, j + 3, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j + 3, k) + rk3gtu_yy * vector(rk3Sfd_y, i, j + 3, k) + rk3gtu_yz * vector(rk3Sfd_z, i, j + 3, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j + 3, k) + rk3gtu_yz * vector(rk3Sfd_y, i, j + 3, k) + rk3gtu_zz * vector(rk3Sfd_z, i, j + 3, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j + 3, k) + rk3gtu_yy * vector(vfd_y, i, j + 3, k) + rk3gtu_yz * vector(vfd_z, i, j + 3, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j + 3, k) + rk3vfu_y * vector(vfd_y, i, j + 3, k) + rk3vfu_z * vector(vfd_z, i, j + 3, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j + 3, k) * vector(rk3Bfu_z, i, j + 3, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j + 3, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j + 3, k) * vector(rk3Bfu_x, i, j + 3, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j + 3, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j + 3, k) * vector(rk3Bfu_z, i, j + 3, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j + 3, k) * rk3vfu_z + vector(rk3Bfu_z, i, j + 3, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j + 3, k) * vector(rk3Bfu_y, i, j + 3, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j + 3, k) * rk3vfu_y + vector(rk3Bfu_y, i, j + 3, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j + 3, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j + 3, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j + 3, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j + 3, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j + 3, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j + 3, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j + 3, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j + 3, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j + 3, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j + 3, k) * vector(rk3Bfu_z, i, j + 3, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j + 3, k) * rk3vfu_z + vector(rk3Bfu_z, i, j + 3, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j + 3, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j + 3, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j + 3, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j + 3, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j + 3, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j + 3, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j + 3, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j + 3, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j + 3, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j + 3, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j + 3, k) * vector(rk3Bfu_y, i, j + 3, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j + 3, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j + 3, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j + 3, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j + 3, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j + 3, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j + 3, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j + 3, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j + 3, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j + 3, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j + 3, k));
						Fluxp3jSfd_x = FjSfd_x_mainI(vector(rk3Sfd_x, i, j + 3, k), vector(rk3gtd_xx, i, j + 3, k), vector(rk3gtd_xy, i, j + 3, k), vector(rk3gtd_xz, i, j + 3, k), vector(rk3Betau_y, i, j + 3, k), vector(rk3Alpha, i, j + 3, k), vector(TauT_xy, i, j + 3, k), vector(TauT_yy, i, j + 3, k), vector(TauT_yz, i, j + 3, k), rk3inv_chi, rk3sdetg, rk3Sfud_yx, dx, simPlat_dt, ilast, jlast);
						Fluxp3jSfd_y = FjSfd_y_mainI(vector(rk3Sfd_y, i, j + 3, k), vector(rk3gtd_xy, i, j + 3, k), vector(rk3gtd_yy, i, j + 3, k), vector(rk3gtd_yz, i, j + 3, k), vector(rk3Betau_y, i, j + 3, k), vector(rk3Alpha, i, j + 3, k), vector(TauT_xy, i, j + 3, k), vector(TauT_yy, i, j + 3, k), vector(TauT_yz, i, j + 3, k), rk3inv_chi, rk3sdetg, rk3Sfud_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp3jSfd_z = FjSfd_z_mainI(vector(rk3Sfd_z, i, j + 3, k), vector(rk3gtd_xz, i, j + 3, k), vector(rk3gtd_yz, i, j + 3, k), vector(rk3gtd_zz, i, j + 3, k), vector(rk3Betau_y, i, j + 3, k), vector(rk3Alpha, i, j + 3, k), vector(TauT_xy, i, j + 3, k), vector(TauT_yy, i, j + 3, k), vector(TauT_yz, i, j + 3, k), rk3inv_chi, rk3sdetg, rk3Sfud_yz, dx, simPlat_dt, ilast, jlast);
						Fluxp3jBfu_x = FjBfu_x_mainI(vector(rk3Bfu_x, i, j + 3, k), vector(rk3Bfu_y, i, j + 3, k), vector(rk3phif, i, j + 3, k), vector(rk3Betau_x, i, j + 3, k), vector(rk3Betau_y, i, j + 3, k), vector(rk3Alpha, i, j + 3, k), vector(TauM_xy, i, j + 3, k), rk3chi_max, rk3gtu_xy, rk3sdetg, rk3vfu_x, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3jBfu_y = FjBfu_y_mainI(vector(rk3phif, i, j + 3, k), vector(rk3Alpha, i, j + 3, k), rk3chi_max, rk3gtu_yy, dx, simPlat_dt, ilast, jlast);
						Fluxp3jBfu_z = FjBfu_z_mainI(vector(rk3Bfu_y, i, j + 3, k), vector(rk3Bfu_z, i, j + 3, k), vector(rk3phif, i, j + 3, k), vector(rk3Betau_y, i, j + 3, k), vector(rk3Betau_z, i, j + 3, k), vector(rk3Alpha, i, j + 3, k), vector(TauM_yz, i, j + 3, k), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3jDf = FjDf_mainI(vector(rk3Df, i, j + 3, k), vector(rk3Betau_y, i, j + 3, k), vector(rk3Alpha, i, j + 3, k), vector(TauN_y, i, j + 3, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3jDYf = FjDYf_mainI(vector(rk3DYf, i, j + 3, k), vector(rk3Betau_y, i, j + 3, k), vector(rk3Alpha, i, j + 3, k), vector(TauNe_y, i, j + 3, k), rk3sdetg, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3jtauf = Fjtauf_mainI(vector(rk3Df, i, j + 3, k), vector(rk3tauf, i, j + 3, k), vector(rk3Betau_y, i, j + 3, k), vector(rk3Alpha, i, j + 3, k), vector(TauN_y, i, j + 3, k), rk3sdetg, rk3Sfu_y, rk3vfu_y, dx, simPlat_dt, ilast, jlast);
						Fluxp3jphif = Fjphif_mainI(vector(rk3Bfu_y, i, j + 3, k), vector(rk3phif, i, j + 3, k), vector(rk3Betau_y, i, j + 3, k), vector(rk3Alpha, i, j + 3, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j + 3, k) * (1.0 + vector(epsf, i, j + 3, k)) + vector(pf, i, j + 3, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j + 3, k) + rk3sqca - vector(sqcs, i, j + 3, k) * rk3sqca;
						SpeedjSfd_x = MAX(SpeedjSfd_x, fabs(fabs(vector(rk3Betau_y, i, j + 3, k)) + vector(rk3Alpha, i, j + 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjSfd_y = MAX(SpeedjSfd_y, fabs(fabs(vector(rk3Betau_y, i, j + 3, k)) + vector(rk3Alpha, i, j + 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjSfd_z = MAX(SpeedjSfd_z, fabs(fabs(vector(rk3Betau_y, i, j + 3, k)) + vector(rk3Alpha, i, j + 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_x = MAX(SpeedjBfu_x, fabs(fabs(vector(rk3Betau_y, i, j + 3, k)) + vector(rk3Alpha, i, j + 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_y = MAX(SpeedjBfu_y, fabs(fabs(vector(rk3Betau_y, i, j + 3, k)) + vector(rk3Alpha, i, j + 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjBfu_z = MAX(SpeedjBfu_z, fabs(fabs(vector(rk3Betau_y, i, j + 3, k)) + vector(rk3Alpha, i, j + 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjDf = MAX(SpeedjDf, fabs(fabs(vector(rk3Betau_y, i, j + 3, k)) + vector(rk3Alpha, i, j + 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						SpeedjDYf = MAX(SpeedjDYf, fabs(fabs(vector(rk3Betau_y, i, j + 3, k)) + vector(rk3Alpha, i, j + 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						Speedjtauf = MAX(Speedjtauf, fabs(fabs(vector(rk3Betau_y, i, j + 3, k)) + vector(rk3Alpha, i, j + 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						Speedjphif = MAX(Speedjphif, fabs(fabs(vector(rk3Betau_y, i, j + 3, k)) + vector(rk3Alpha, i, j + 3, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_y) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_yy - (1.0 - rk3sqa) * (rk3vfu_y * rk3vfu_y)))))));
						LLFpjSfd_x = 0.5 * (FluxjSfd_x + fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j, k));
						LLFpjSfd_y = 0.5 * (FluxjSfd_y + fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j, k));
						LLFpjSfd_z = 0.5 * (FluxjSfd_z + fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j, k));
						LLFpjBfu_x = 0.5 * (FluxjBfu_x + fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j, k));
						LLFpjBfu_y = 0.5 * (FluxjBfu_y + fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j, k));
						LLFpjBfu_z = 0.5 * (FluxjBfu_z + fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j, k));
						LLFpjDf = 0.5 * (FluxjDf + fabs(SpeedjDf) * vector(rk3Df, i, j, k));
						LLFpjDYf = 0.5 * (FluxjDYf + fabs(SpeedjDYf) * vector(rk3DYf, i, j, k));
						LLFpjtauf = 0.5 * (Fluxjtauf + fabs(Speedjtauf) * vector(rk3tauf, i, j, k));
						LLFpjphif = 0.5 * (Fluxjphif + fabs(Speedjphif) * vector(rk3phif, i, j, k));
						LLFnjSfd_x = 0.5 * (FluxjSfd_x - fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j, k));
						LLFnjSfd_y = 0.5 * (FluxjSfd_y - fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j, k));
						LLFnjSfd_z = 0.5 * (FluxjSfd_z - fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j, k));
						LLFnjBfu_x = 0.5 * (FluxjBfu_x - fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j, k));
						LLFnjBfu_y = 0.5 * (FluxjBfu_y - fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j, k));
						LLFnjBfu_z = 0.5 * (FluxjBfu_z - fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j, k));
						LLFnjDf = 0.5 * (FluxjDf - fabs(SpeedjDf) * vector(rk3Df, i, j, k));
						LLFnjDYf = 0.5 * (FluxjDYf - fabs(SpeedjDYf) * vector(rk3DYf, i, j, k));
						LLFnjtauf = 0.5 * (Fluxjtauf - fabs(Speedjtauf) * vector(rk3tauf, i, j, k));
						LLFnjphif = 0.5 * (Fluxjphif - fabs(Speedjphif) * vector(rk3phif, i, j, k));
						LLFp_m3jSfd_x = 0.5 * (Fluxm3jSfd_x + fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j - 3, k));
						LLFp_m3jSfd_y = 0.5 * (Fluxm3jSfd_y + fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j - 3, k));
						LLFp_m3jSfd_z = 0.5 * (Fluxm3jSfd_z + fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j - 3, k));
						LLFp_m3jBfu_x = 0.5 * (Fluxm3jBfu_x + fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j - 3, k));
						LLFp_m3jBfu_y = 0.5 * (Fluxm3jBfu_y + fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j - 3, k));
						LLFp_m3jBfu_z = 0.5 * (Fluxm3jBfu_z + fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j - 3, k));
						LLFp_m3jDf = 0.5 * (Fluxm3jDf + fabs(SpeedjDf) * vector(rk3Df, i, j - 3, k));
						LLFp_m3jDYf = 0.5 * (Fluxm3jDYf + fabs(SpeedjDYf) * vector(rk3DYf, i, j - 3, k));
						LLFp_m3jtauf = 0.5 * (Fluxm3jtauf + fabs(Speedjtauf) * vector(rk3tauf, i, j - 3, k));
						LLFp_m3jphif = 0.5 * (Fluxm3jphif + fabs(Speedjphif) * vector(rk3phif, i, j - 3, k));
						LLFn_m3jSfd_x = 0.5 * (Fluxm3jSfd_x - fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j - 3, k));
						LLFn_m3jSfd_y = 0.5 * (Fluxm3jSfd_y - fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j - 3, k));
						LLFn_m3jSfd_z = 0.5 * (Fluxm3jSfd_z - fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j - 3, k));
						LLFn_m3jBfu_x = 0.5 * (Fluxm3jBfu_x - fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j - 3, k));
						LLFn_m3jBfu_y = 0.5 * (Fluxm3jBfu_y - fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j - 3, k));
						LLFn_m3jBfu_z = 0.5 * (Fluxm3jBfu_z - fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j - 3, k));
						LLFn_m3jDf = 0.5 * (Fluxm3jDf - fabs(SpeedjDf) * vector(rk3Df, i, j - 3, k));
						LLFn_m3jDYf = 0.5 * (Fluxm3jDYf - fabs(SpeedjDYf) * vector(rk3DYf, i, j - 3, k));
						LLFn_m3jtauf = 0.5 * (Fluxm3jtauf - fabs(Speedjtauf) * vector(rk3tauf, i, j - 3, k));
						LLFn_m3jphif = 0.5 * (Fluxm3jphif - fabs(Speedjphif) * vector(rk3phif, i, j - 3, k));
						LLFp_m2jSfd_x = 0.5 * (Fluxm2jSfd_x + fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j - 2, k));
						LLFp_m2jSfd_y = 0.5 * (Fluxm2jSfd_y + fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j - 2, k));
						LLFp_m2jSfd_z = 0.5 * (Fluxm2jSfd_z + fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j - 2, k));
						LLFp_m2jBfu_x = 0.5 * (Fluxm2jBfu_x + fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j - 2, k));
						LLFp_m2jBfu_y = 0.5 * (Fluxm2jBfu_y + fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j - 2, k));
						LLFp_m2jBfu_z = 0.5 * (Fluxm2jBfu_z + fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j - 2, k));
						LLFp_m2jDf = 0.5 * (Fluxm2jDf + fabs(SpeedjDf) * vector(rk3Df, i, j - 2, k));
						LLFp_m2jDYf = 0.5 * (Fluxm2jDYf + fabs(SpeedjDYf) * vector(rk3DYf, i, j - 2, k));
						LLFp_m2jtauf = 0.5 * (Fluxm2jtauf + fabs(Speedjtauf) * vector(rk3tauf, i, j - 2, k));
						LLFp_m2jphif = 0.5 * (Fluxm2jphif + fabs(Speedjphif) * vector(rk3phif, i, j - 2, k));
						LLFn_m2jSfd_x = 0.5 * (Fluxm2jSfd_x - fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j - 2, k));
						LLFn_m2jSfd_y = 0.5 * (Fluxm2jSfd_y - fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j - 2, k));
						LLFn_m2jSfd_z = 0.5 * (Fluxm2jSfd_z - fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j - 2, k));
						LLFn_m2jBfu_x = 0.5 * (Fluxm2jBfu_x - fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j - 2, k));
						LLFn_m2jBfu_y = 0.5 * (Fluxm2jBfu_y - fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j - 2, k));
						LLFn_m2jBfu_z = 0.5 * (Fluxm2jBfu_z - fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j - 2, k));
						LLFn_m2jDf = 0.5 * (Fluxm2jDf - fabs(SpeedjDf) * vector(rk3Df, i, j - 2, k));
						LLFn_m2jDYf = 0.5 * (Fluxm2jDYf - fabs(SpeedjDYf) * vector(rk3DYf, i, j - 2, k));
						LLFn_m2jtauf = 0.5 * (Fluxm2jtauf - fabs(Speedjtauf) * vector(rk3tauf, i, j - 2, k));
						LLFn_m2jphif = 0.5 * (Fluxm2jphif - fabs(Speedjphif) * vector(rk3phif, i, j - 2, k));
						LLFp_m1jSfd_x = 0.5 * (Fluxm1jSfd_x + fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j - 1, k));
						LLFp_m1jSfd_y = 0.5 * (Fluxm1jSfd_y + fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j - 1, k));
						LLFp_m1jSfd_z = 0.5 * (Fluxm1jSfd_z + fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j - 1, k));
						LLFp_m1jBfu_x = 0.5 * (Fluxm1jBfu_x + fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j - 1, k));
						LLFp_m1jBfu_y = 0.5 * (Fluxm1jBfu_y + fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j - 1, k));
						LLFp_m1jBfu_z = 0.5 * (Fluxm1jBfu_z + fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j - 1, k));
						LLFp_m1jDf = 0.5 * (Fluxm1jDf + fabs(SpeedjDf) * vector(rk3Df, i, j - 1, k));
						LLFp_m1jDYf = 0.5 * (Fluxm1jDYf + fabs(SpeedjDYf) * vector(rk3DYf, i, j - 1, k));
						LLFp_m1jtauf = 0.5 * (Fluxm1jtauf + fabs(Speedjtauf) * vector(rk3tauf, i, j - 1, k));
						LLFp_m1jphif = 0.5 * (Fluxm1jphif + fabs(Speedjphif) * vector(rk3phif, i, j - 1, k));
						LLFn_m1jSfd_x = 0.5 * (Fluxm1jSfd_x - fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j - 1, k));
						LLFn_m1jSfd_y = 0.5 * (Fluxm1jSfd_y - fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j - 1, k));
						LLFn_m1jSfd_z = 0.5 * (Fluxm1jSfd_z - fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j - 1, k));
						LLFn_m1jBfu_x = 0.5 * (Fluxm1jBfu_x - fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j - 1, k));
						LLFn_m1jBfu_y = 0.5 * (Fluxm1jBfu_y - fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j - 1, k));
						LLFn_m1jBfu_z = 0.5 * (Fluxm1jBfu_z - fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j - 1, k));
						LLFn_m1jDf = 0.5 * (Fluxm1jDf - fabs(SpeedjDf) * vector(rk3Df, i, j - 1, k));
						LLFn_m1jDYf = 0.5 * (Fluxm1jDYf - fabs(SpeedjDYf) * vector(rk3DYf, i, j - 1, k));
						LLFn_m1jtauf = 0.5 * (Fluxm1jtauf - fabs(Speedjtauf) * vector(rk3tauf, i, j - 1, k));
						LLFn_m1jphif = 0.5 * (Fluxm1jphif - fabs(Speedjphif) * vector(rk3phif, i, j - 1, k));
						LLFp_p1jSfd_x = 0.5 * (Fluxp1jSfd_x + fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j + 1, k));
						LLFp_p1jSfd_y = 0.5 * (Fluxp1jSfd_y + fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j + 1, k));
						LLFp_p1jSfd_z = 0.5 * (Fluxp1jSfd_z + fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j + 1, k));
						LLFp_p1jBfu_x = 0.5 * (Fluxp1jBfu_x + fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j + 1, k));
						LLFp_p1jBfu_y = 0.5 * (Fluxp1jBfu_y + fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j + 1, k));
						LLFp_p1jBfu_z = 0.5 * (Fluxp1jBfu_z + fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j + 1, k));
						LLFp_p1jDf = 0.5 * (Fluxp1jDf + fabs(SpeedjDf) * vector(rk3Df, i, j + 1, k));
						LLFp_p1jDYf = 0.5 * (Fluxp1jDYf + fabs(SpeedjDYf) * vector(rk3DYf, i, j + 1, k));
						LLFp_p1jtauf = 0.5 * (Fluxp1jtauf + fabs(Speedjtauf) * vector(rk3tauf, i, j + 1, k));
						LLFp_p1jphif = 0.5 * (Fluxp1jphif + fabs(Speedjphif) * vector(rk3phif, i, j + 1, k));
						LLFn_p1jSfd_x = 0.5 * (Fluxp1jSfd_x - fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j + 1, k));
						LLFn_p1jSfd_y = 0.5 * (Fluxp1jSfd_y - fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j + 1, k));
						LLFn_p1jSfd_z = 0.5 * (Fluxp1jSfd_z - fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j + 1, k));
						LLFn_p1jBfu_x = 0.5 * (Fluxp1jBfu_x - fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j + 1, k));
						LLFn_p1jBfu_y = 0.5 * (Fluxp1jBfu_y - fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j + 1, k));
						LLFn_p1jBfu_z = 0.5 * (Fluxp1jBfu_z - fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j + 1, k));
						LLFn_p1jDf = 0.5 * (Fluxp1jDf - fabs(SpeedjDf) * vector(rk3Df, i, j + 1, k));
						LLFn_p1jDYf = 0.5 * (Fluxp1jDYf - fabs(SpeedjDYf) * vector(rk3DYf, i, j + 1, k));
						LLFn_p1jtauf = 0.5 * (Fluxp1jtauf - fabs(Speedjtauf) * vector(rk3tauf, i, j + 1, k));
						LLFn_p1jphif = 0.5 * (Fluxp1jphif - fabs(Speedjphif) * vector(rk3phif, i, j + 1, k));
						LLFp_p2jSfd_x = 0.5 * (Fluxp2jSfd_x + fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j + 2, k));
						LLFp_p2jSfd_y = 0.5 * (Fluxp2jSfd_y + fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j + 2, k));
						LLFp_p2jSfd_z = 0.5 * (Fluxp2jSfd_z + fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j + 2, k));
						LLFp_p2jBfu_x = 0.5 * (Fluxp2jBfu_x + fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j + 2, k));
						LLFp_p2jBfu_y = 0.5 * (Fluxp2jBfu_y + fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j + 2, k));
						LLFp_p2jBfu_z = 0.5 * (Fluxp2jBfu_z + fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j + 2, k));
						LLFp_p2jDf = 0.5 * (Fluxp2jDf + fabs(SpeedjDf) * vector(rk3Df, i, j + 2, k));
						LLFp_p2jDYf = 0.5 * (Fluxp2jDYf + fabs(SpeedjDYf) * vector(rk3DYf, i, j + 2, k));
						LLFp_p2jtauf = 0.5 * (Fluxp2jtauf + fabs(Speedjtauf) * vector(rk3tauf, i, j + 2, k));
						LLFp_p2jphif = 0.5 * (Fluxp2jphif + fabs(Speedjphif) * vector(rk3phif, i, j + 2, k));
						LLFn_p2jSfd_x = 0.5 * (Fluxp2jSfd_x - fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j + 2, k));
						LLFn_p2jSfd_y = 0.5 * (Fluxp2jSfd_y - fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j + 2, k));
						LLFn_p2jSfd_z = 0.5 * (Fluxp2jSfd_z - fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j + 2, k));
						LLFn_p2jBfu_x = 0.5 * (Fluxp2jBfu_x - fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j + 2, k));
						LLFn_p2jBfu_y = 0.5 * (Fluxp2jBfu_y - fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j + 2, k));
						LLFn_p2jBfu_z = 0.5 * (Fluxp2jBfu_z - fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j + 2, k));
						LLFn_p2jDf = 0.5 * (Fluxp2jDf - fabs(SpeedjDf) * vector(rk3Df, i, j + 2, k));
						LLFn_p2jDYf = 0.5 * (Fluxp2jDYf - fabs(SpeedjDYf) * vector(rk3DYf, i, j + 2, k));
						LLFn_p2jtauf = 0.5 * (Fluxp2jtauf - fabs(Speedjtauf) * vector(rk3tauf, i, j + 2, k));
						LLFn_p2jphif = 0.5 * (Fluxp2jphif - fabs(Speedjphif) * vector(rk3phif, i, j + 2, k));
						LLFp_p3jSfd_x = 0.5 * (Fluxp3jSfd_x + fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j + 3, k));
						LLFp_p3jSfd_y = 0.5 * (Fluxp3jSfd_y + fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j + 3, k));
						LLFp_p3jSfd_z = 0.5 * (Fluxp3jSfd_z + fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j + 3, k));
						LLFp_p3jBfu_x = 0.5 * (Fluxp3jBfu_x + fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j + 3, k));
						LLFp_p3jBfu_y = 0.5 * (Fluxp3jBfu_y + fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j + 3, k));
						LLFp_p3jBfu_z = 0.5 * (Fluxp3jBfu_z + fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j + 3, k));
						LLFp_p3jDf = 0.5 * (Fluxp3jDf + fabs(SpeedjDf) * vector(rk3Df, i, j + 3, k));
						LLFp_p3jDYf = 0.5 * (Fluxp3jDYf + fabs(SpeedjDYf) * vector(rk3DYf, i, j + 3, k));
						LLFp_p3jtauf = 0.5 * (Fluxp3jtauf + fabs(Speedjtauf) * vector(rk3tauf, i, j + 3, k));
						LLFp_p3jphif = 0.5 * (Fluxp3jphif + fabs(Speedjphif) * vector(rk3phif, i, j + 3, k));
						LLFn_p3jSfd_x = 0.5 * (Fluxp3jSfd_x - fabs(SpeedjSfd_x) * vector(rk3Sfd_x, i, j + 3, k));
						LLFn_p3jSfd_y = 0.5 * (Fluxp3jSfd_y - fabs(SpeedjSfd_y) * vector(rk3Sfd_y, i, j + 3, k));
						LLFn_p3jSfd_z = 0.5 * (Fluxp3jSfd_z - fabs(SpeedjSfd_z) * vector(rk3Sfd_z, i, j + 3, k));
						LLFn_p3jBfu_x = 0.5 * (Fluxp3jBfu_x - fabs(SpeedjBfu_x) * vector(rk3Bfu_x, i, j + 3, k));
						LLFn_p3jBfu_y = 0.5 * (Fluxp3jBfu_y - fabs(SpeedjBfu_y) * vector(rk3Bfu_y, i, j + 3, k));
						LLFn_p3jBfu_z = 0.5 * (Fluxp3jBfu_z - fabs(SpeedjBfu_z) * vector(rk3Bfu_z, i, j + 3, k));
						LLFn_p3jDf = 0.5 * (Fluxp3jDf - fabs(SpeedjDf) * vector(rk3Df, i, j + 3, k));
						LLFn_p3jDYf = 0.5 * (Fluxp3jDYf - fabs(SpeedjDYf) * vector(rk3DYf, i, j + 3, k));
						LLFn_p3jtauf = 0.5 * (Fluxp3jtauf - fabs(Speedjtauf) * vector(rk3tauf, i, j + 3, k));
						LLFn_p3jphif = 0.5 * (Fluxp3jphif - fabs(Speedjphif) * vector(rk3phif, i, j + 3, k));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j, k));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_yy, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_zz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_xy, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_xz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3sqBf = vector(rk3Bfu_x, i, j, k) * rk3Bfd_x + vector(rk3Bfu_y, i, j, k) * rk3Bfd_y + vector(rk3Bfu_z, i, j, k) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) * vector(rk3gtd_xx, i, j, k) + (-vector(rk3gtd_yy, i, j, k) * (vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k))) + (-vector(rk3gtd_zz, i, j, k) * (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k))) + (-(vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k)) * vector(rk3gtd_xx, i, j, k)) + 2.0 * vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yy, i, j, k) - vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) - vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_yz, i, j, k) - vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_zz, i, j, k)) + vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j, k) + rk3gtu_xy * vector(rk3Sfd_y, i, j, k) + rk3gtu_xz * vector(rk3Sfd_z, i, j, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j, k) + rk3gtu_xy * vector(vfd_y, i, j, k) + rk3gtu_xz * vector(vfd_z, i, j, k));
						rk3Bfvf = vector(rk3Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk3Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk3Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_zz, i, j, k) - vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yz, i, j, k)) + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j, k) + rk3gtu_yz * vector(vfd_y, i, j, k) + rk3gtu_zz * vector(vfd_z, i, j, k));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j, k) + rk3gtu_yy * vector(rk3Sfd_y, i, j, k) + rk3gtu_yz * vector(rk3Sfd_z, i, j, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j, k) + rk3gtu_yz * vector(rk3Sfd_y, i, j, k) + rk3gtu_zz * vector(rk3Sfd_z, i, j, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j, k) + rk3gtu_yy * vector(vfd_y, i, j, k) + rk3gtu_yz * vector(vfd_z, i, j, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j, k) + rk3vfu_y * vector(vfd_y, i, j, k) + rk3vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j, k) * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j, k) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j, k) * vector(rk3Bfu_x, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k) * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k) * rk3vfu_z + vector(rk3Bfu_z, i, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k) * vector(rk3Bfu_y, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k) * rk3vfu_y + vector(rk3Bfu_y, i, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j, k) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j, k));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j, k) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j, k) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j, k));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k) * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k) * rk3vfu_z + vector(rk3Bfu_z, i, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j, k) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j, k));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j, k));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j, k));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j, k) * vector(rk3Bfu_y, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j, k) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j, k));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j, k) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j, k));
						FluxkSfd_x = FkSfd_x_mainI(vector(rk3Sfd_x, i, j, k), vector(rk3gtd_xx, i, j, k), vector(rk3gtd_xy, i, j, k), vector(rk3gtd_xz, i, j, k), vector(rk3Betau_z, i, j, k), vector(rk3Alpha, i, j, k), vector(TauT_xz, i, j, k), vector(TauT_yz, i, j, k), vector(TauT_zz, i, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_zx, dx, simPlat_dt, ilast, jlast);
						FluxkSfd_y = FkSfd_y_mainI(vector(rk3Sfd_y, i, j, k), vector(rk3gtd_xy, i, j, k), vector(rk3gtd_yy, i, j, k), vector(rk3gtd_yz, i, j, k), vector(rk3Betau_z, i, j, k), vector(rk3Alpha, i, j, k), vector(TauT_xz, i, j, k), vector(TauT_yz, i, j, k), vector(TauT_zz, i, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_zy, dx, simPlat_dt, ilast, jlast);
						FluxkSfd_z = FkSfd_z_mainI(vector(rk3Sfd_z, i, j, k), vector(rk3gtd_xz, i, j, k), vector(rk3gtd_yz, i, j, k), vector(rk3gtd_zz, i, j, k), vector(rk3Betau_z, i, j, k), vector(rk3Alpha, i, j, k), vector(TauT_xz, i, j, k), vector(TauT_yz, i, j, k), vector(TauT_zz, i, j, k), rk3inv_chi, rk3sdetg, rk3Sfud_zz, dx, simPlat_dt, ilast, jlast);
						FluxkBfu_x = FkBfu_x_mainI(vector(rk3Bfu_x, i, j, k), vector(rk3Bfu_z, i, j, k), vector(rk3phif, i, j, k), vector(rk3Betau_x, i, j, k), vector(rk3Betau_z, i, j, k), vector(rk3Alpha, i, j, k), vector(TauM_xz, i, j, k), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxkBfu_y = FkBfu_y_mainI(vector(rk3Bfu_y, i, j, k), vector(rk3Bfu_z, i, j, k), vector(rk3phif, i, j, k), vector(rk3Betau_y, i, j, k), vector(rk3Betau_z, i, j, k), vector(rk3Alpha, i, j, k), vector(TauM_yz, i, j, k), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxkBfu_z = FkBfu_z_mainI(vector(rk3phif, i, j, k), vector(rk3Alpha, i, j, k), rk3chi_max, rk3gtu_zz, dx, simPlat_dt, ilast, jlast);
						FluxkDf = FkDf_mainI(vector(rk3Df, i, j, k), vector(rk3Betau_z, i, j, k), vector(rk3Alpha, i, j, k), vector(TauN_z, i, j, k), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						FluxkDYf = FkDYf_mainI(vector(rk3DYf, i, j, k), vector(rk3Betau_z, i, j, k), vector(rk3Alpha, i, j, k), vector(TauNe_z, i, j, k), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxktauf = Fktauf_mainI(vector(rk3Df, i, j, k), vector(rk3tauf, i, j, k), vector(rk3Betau_z, i, j, k), vector(rk3Alpha, i, j, k), vector(TauN_z, i, j, k), rk3sdetg, rk3Sfu_z, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxkphif = Fkphif_mainI(vector(rk3Bfu_z, i, j, k), vector(rk3phif, i, j, k), vector(rk3Betau_z, i, j, k), vector(rk3Alpha, i, j, k), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j, k) + rk3sqca - vector(sqcs, i, j, k) * rk3sqca;
						SpeedkSfd_x = fabs(fabs(vector(rk3Betau_z, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z))))));
						SpeedkSfd_y = fabs(fabs(vector(rk3Betau_z, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z))))));
						SpeedkSfd_z = fabs(fabs(vector(rk3Betau_z, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z))))));
						SpeedkBfu_x = fabs(fabs(vector(rk3Betau_z, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z))))));
						SpeedkBfu_y = fabs(fabs(vector(rk3Betau_z, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z))))));
						SpeedkBfu_z = fabs(fabs(vector(rk3Betau_z, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z))))));
						SpeedkDf = fabs(fabs(vector(rk3Betau_z, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z))))));
						SpeedkDYf = fabs(fabs(vector(rk3Betau_z, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z))))));
						Speedktauf = fabs(fabs(vector(rk3Betau_z, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z))))));
						Speedkphif = fabs(fabs(vector(rk3Betau_z, i, j, k)) + vector(rk3Alpha, i, j, k) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j, k - 3));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k - 3) * vector(rk3Bfu_x, i, j, k - 3) + vector(rk3gtd_yy, i, j, k - 3) * vector(rk3Bfu_y, i, j, k - 3) + vector(rk3gtd_yz, i, j, k - 3) * vector(rk3Bfu_z, i, j, k - 3));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k - 3) * vector(rk3Bfu_x, i, j, k - 3) + vector(rk3gtd_yz, i, j, k - 3) * vector(rk3Bfu_y, i, j, k - 3) + vector(rk3gtd_zz, i, j, k - 3) * vector(rk3Bfu_z, i, j, k - 3));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k - 3) * vector(rk3Bfu_x, i, j, k - 3) + vector(rk3gtd_xy, i, j, k - 3) * vector(rk3Bfu_y, i, j, k - 3) + vector(rk3gtd_xz, i, j, k - 3) * vector(rk3Bfu_z, i, j, k - 3));
						rk3sqBf = vector(rk3Bfu_x, i, j, k - 3) * rk3Bfd_x + vector(rk3Bfu_y, i, j, k - 3) * rk3Bfd_y + vector(rk3Bfu_z, i, j, k - 3) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j, k - 3) * vector(rk3gtd_zz, i, j, k - 3) * vector(rk3gtd_xx, i, j, k - 3) + (-vector(rk3gtd_yy, i, j, k - 3) * (vector(rk3gtd_xz, i, j, k - 3) * vector(rk3gtd_xz, i, j, k - 3))) + (-vector(rk3gtd_zz, i, j, k - 3) * (vector(rk3gtd_xy, i, j, k - 3) * vector(rk3gtd_xy, i, j, k - 3))) + (-(vector(rk3gtd_yz, i, j, k - 3) * vector(rk3gtd_yz, i, j, k - 3)) * vector(rk3gtd_xx, i, j, k - 3)) + 2.0 * vector(rk3gtd_yz, i, j, k - 3) * vector(rk3gtd_xy, i, j, k - 3) * vector(rk3gtd_xz, i, j, k - 3);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j, k - 3) * vector(rk3gtd_yy, i, j, k - 3) - vector(rk3gtd_xy, i, j, k - 3) * vector(rk3gtd_xy, i, j, k - 3));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j, k - 3) * vector(rk3gtd_zz, i, j, k - 3) - vector(rk3gtd_yz, i, j, k - 3) * vector(rk3gtd_yz, i, j, k - 3));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j, k - 3) * vector(rk3gtd_yz, i, j, k - 3) - vector(rk3gtd_yy, i, j, k - 3) * vector(rk3gtd_xz, i, j, k - 3));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j, k - 3) * vector(rk3gtd_zz, i, j, k - 3)) + vector(rk3gtd_yz, i, j, k - 3) * vector(rk3gtd_xz, i, j, k - 3));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j, k - 3) + rk3gtu_xy * vector(rk3Sfd_y, i, j, k - 3) + rk3gtu_xz * vector(rk3Sfd_z, i, j, k - 3));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j, k - 3) + rk3gtu_xy * vector(vfd_y, i, j, k - 3) + rk3gtu_xz * vector(vfd_z, i, j, k - 3));
						rk3Bfvf = vector(rk3Bfu_x, i, j, k - 3) * vector(vfd_x, i, j, k - 3) + vector(rk3Bfu_y, i, j, k - 3) * vector(vfd_y, i, j, k - 3) + vector(rk3Bfu_z, i, j, k - 3) * vector(vfd_z, i, j, k - 3);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j, k - 3) * vector(rk3gtd_zz, i, j, k - 3) - vector(rk3gtd_xz, i, j, k - 3) * vector(rk3gtd_xz, i, j, k - 3));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j, k - 3) * vector(rk3gtd_yz, i, j, k - 3)) + vector(rk3gtd_xy, i, j, k - 3) * vector(rk3gtd_xz, i, j, k - 3));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j, k - 3) + rk3gtu_yz * vector(vfd_y, i, j, k - 3) + rk3gtu_zz * vector(vfd_z, i, j, k - 3));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j, k - 3) + rk3gtu_yy * vector(rk3Sfd_y, i, j, k - 3) + rk3gtu_yz * vector(rk3Sfd_z, i, j, k - 3));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j, k - 3) + rk3gtu_yz * vector(rk3Sfd_y, i, j, k - 3) + rk3gtu_zz * vector(rk3Sfd_z, i, j, k - 3));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j, k - 3) + rk3gtu_yy * vector(vfd_y, i, j, k - 3) + rk3gtu_yz * vector(vfd_z, i, j, k - 3));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j, k - 3) + rk3vfu_y * vector(vfd_y, i, j, k - 3) + rk3vfu_z * vector(vfd_z, i, j, k - 3), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j, k - 3) * vector(rk3Bfu_z, i, j, k - 3)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j, k - 3) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j, k - 3) * vector(rk3Bfu_x, i, j, k - 3)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j, k - 3) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k - 3) * vector(rk3Bfu_z, i, j, k - 3)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k - 3) * rk3vfu_z + vector(rk3Bfu_z, i, j, k - 3) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k - 3) * vector(rk3Bfu_y, i, j, k - 3)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k - 3) * rk3vfu_y + vector(rk3Bfu_y, i, j, k - 3) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j, k - 3) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k - 3) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k - 3));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j, k - 3) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j, k - 3) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j, k - 3));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j, k - 3) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j, k - 3) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j, k - 3));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k - 3) * vector(rk3Bfu_z, i, j, k - 3)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k - 3) * rk3vfu_z + vector(rk3Bfu_z, i, j, k - 3) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j, k - 3) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j, k - 3) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j, k - 3));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k - 3) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k - 3) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j, k - 3));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j, k - 3) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j, k - 3) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j, k - 3));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j, k - 3) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j, k - 3) * vector(rk3Bfu_y, i, j, k - 3)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j, k - 3) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j, k - 3) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j, k - 3) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j, k - 3));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k - 3) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j, k - 3) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k - 3));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j, k - 3) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j, k - 3) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j, k - 3));
						Fluxm3kSfd_x = FkSfd_x_mainI(vector(rk3Sfd_x, i, j, k - 3), vector(rk3gtd_xx, i, j, k - 3), vector(rk3gtd_xy, i, j, k - 3), vector(rk3gtd_xz, i, j, k - 3), vector(rk3Betau_z, i, j, k - 3), vector(rk3Alpha, i, j, k - 3), vector(TauT_xz, i, j, k - 3), vector(TauT_yz, i, j, k - 3), vector(TauT_zz, i, j, k - 3), rk3inv_chi, rk3sdetg, rk3Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxm3kSfd_y = FkSfd_y_mainI(vector(rk3Sfd_y, i, j, k - 3), vector(rk3gtd_xy, i, j, k - 3), vector(rk3gtd_yy, i, j, k - 3), vector(rk3gtd_yz, i, j, k - 3), vector(rk3Betau_z, i, j, k - 3), vector(rk3Alpha, i, j, k - 3), vector(TauT_xz, i, j, k - 3), vector(TauT_yz, i, j, k - 3), vector(TauT_zz, i, j, k - 3), rk3inv_chi, rk3sdetg, rk3Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxm3kSfd_z = FkSfd_z_mainI(vector(rk3Sfd_z, i, j, k - 3), vector(rk3gtd_xz, i, j, k - 3), vector(rk3gtd_yz, i, j, k - 3), vector(rk3gtd_zz, i, j, k - 3), vector(rk3Betau_z, i, j, k - 3), vector(rk3Alpha, i, j, k - 3), vector(TauT_xz, i, j, k - 3), vector(TauT_yz, i, j, k - 3), vector(TauT_zz, i, j, k - 3), rk3inv_chi, rk3sdetg, rk3Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm3kBfu_x = FkBfu_x_mainI(vector(rk3Bfu_x, i, j, k - 3), vector(rk3Bfu_z, i, j, k - 3), vector(rk3phif, i, j, k - 3), vector(rk3Betau_x, i, j, k - 3), vector(rk3Betau_z, i, j, k - 3), vector(rk3Alpha, i, j, k - 3), vector(TauM_xz, i, j, k - 3), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3kBfu_y = FkBfu_y_mainI(vector(rk3Bfu_y, i, j, k - 3), vector(rk3Bfu_z, i, j, k - 3), vector(rk3phif, i, j, k - 3), vector(rk3Betau_y, i, j, k - 3), vector(rk3Betau_z, i, j, k - 3), vector(rk3Alpha, i, j, k - 3), vector(TauM_yz, i, j, k - 3), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3kBfu_z = FkBfu_z_mainI(vector(rk3phif, i, j, k - 3), vector(rk3Alpha, i, j, k - 3), rk3chi_max, rk3gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm3kDf = FkDf_mainI(vector(rk3Df, i, j, k - 3), vector(rk3Betau_z, i, j, k - 3), vector(rk3Alpha, i, j, k - 3), vector(TauN_z, i, j, k - 3), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3kDYf = FkDYf_mainI(vector(rk3DYf, i, j, k - 3), vector(rk3Betau_z, i, j, k - 3), vector(rk3Alpha, i, j, k - 3), vector(TauNe_z, i, j, k - 3), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3ktauf = Fktauf_mainI(vector(rk3Df, i, j, k - 3), vector(rk3tauf, i, j, k - 3), vector(rk3Betau_z, i, j, k - 3), vector(rk3Alpha, i, j, k - 3), vector(TauN_z, i, j, k - 3), rk3sdetg, rk3Sfu_z, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm3kphif = Fkphif_mainI(vector(rk3Bfu_z, i, j, k - 3), vector(rk3phif, i, j, k - 3), vector(rk3Betau_z, i, j, k - 3), vector(rk3Alpha, i, j, k - 3), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j, k - 3) * (1.0 + vector(epsf, i, j, k - 3)) + vector(pf, i, j, k - 3);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j, k - 3) + rk3sqca - vector(sqcs, i, j, k - 3) * rk3sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk3Betau_z, i, j, k - 3)) + vector(rk3Alpha, i, j, k - 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk3Betau_z, i, j, k - 3)) + vector(rk3Alpha, i, j, k - 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk3Betau_z, i, j, k - 3)) + vector(rk3Alpha, i, j, k - 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk3Betau_z, i, j, k - 3)) + vector(rk3Alpha, i, j, k - 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk3Betau_z, i, j, k - 3)) + vector(rk3Alpha, i, j, k - 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk3Betau_z, i, j, k - 3)) + vector(rk3Alpha, i, j, k - 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk3Betau_z, i, j, k - 3)) + vector(rk3Alpha, i, j, k - 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk3Betau_z, i, j, k - 3)) + vector(rk3Alpha, i, j, k - 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk3Betau_z, i, j, k - 3)) + vector(rk3Alpha, i, j, k - 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk3Betau_z, i, j, k - 3)) + vector(rk3Alpha, i, j, k - 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j, k - 2));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k - 2) * vector(rk3Bfu_x, i, j, k - 2) + vector(rk3gtd_yy, i, j, k - 2) * vector(rk3Bfu_y, i, j, k - 2) + vector(rk3gtd_yz, i, j, k - 2) * vector(rk3Bfu_z, i, j, k - 2));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k - 2) * vector(rk3Bfu_x, i, j, k - 2) + vector(rk3gtd_yz, i, j, k - 2) * vector(rk3Bfu_y, i, j, k - 2) + vector(rk3gtd_zz, i, j, k - 2) * vector(rk3Bfu_z, i, j, k - 2));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k - 2) * vector(rk3Bfu_x, i, j, k - 2) + vector(rk3gtd_xy, i, j, k - 2) * vector(rk3Bfu_y, i, j, k - 2) + vector(rk3gtd_xz, i, j, k - 2) * vector(rk3Bfu_z, i, j, k - 2));
						rk3sqBf = vector(rk3Bfu_x, i, j, k - 2) * rk3Bfd_x + vector(rk3Bfu_y, i, j, k - 2) * rk3Bfd_y + vector(rk3Bfu_z, i, j, k - 2) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j, k - 2) * vector(rk3gtd_zz, i, j, k - 2) * vector(rk3gtd_xx, i, j, k - 2) + (-vector(rk3gtd_yy, i, j, k - 2) * (vector(rk3gtd_xz, i, j, k - 2) * vector(rk3gtd_xz, i, j, k - 2))) + (-vector(rk3gtd_zz, i, j, k - 2) * (vector(rk3gtd_xy, i, j, k - 2) * vector(rk3gtd_xy, i, j, k - 2))) + (-(vector(rk3gtd_yz, i, j, k - 2) * vector(rk3gtd_yz, i, j, k - 2)) * vector(rk3gtd_xx, i, j, k - 2)) + 2.0 * vector(rk3gtd_yz, i, j, k - 2) * vector(rk3gtd_xy, i, j, k - 2) * vector(rk3gtd_xz, i, j, k - 2);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j, k - 2) * vector(rk3gtd_yy, i, j, k - 2) - vector(rk3gtd_xy, i, j, k - 2) * vector(rk3gtd_xy, i, j, k - 2));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j, k - 2) * vector(rk3gtd_zz, i, j, k - 2) - vector(rk3gtd_yz, i, j, k - 2) * vector(rk3gtd_yz, i, j, k - 2));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j, k - 2) * vector(rk3gtd_yz, i, j, k - 2) - vector(rk3gtd_yy, i, j, k - 2) * vector(rk3gtd_xz, i, j, k - 2));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j, k - 2) * vector(rk3gtd_zz, i, j, k - 2)) + vector(rk3gtd_yz, i, j, k - 2) * vector(rk3gtd_xz, i, j, k - 2));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j, k - 2) + rk3gtu_xy * vector(rk3Sfd_y, i, j, k - 2) + rk3gtu_xz * vector(rk3Sfd_z, i, j, k - 2));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j, k - 2) + rk3gtu_xy * vector(vfd_y, i, j, k - 2) + rk3gtu_xz * vector(vfd_z, i, j, k - 2));
						rk3Bfvf = vector(rk3Bfu_x, i, j, k - 2) * vector(vfd_x, i, j, k - 2) + vector(rk3Bfu_y, i, j, k - 2) * vector(vfd_y, i, j, k - 2) + vector(rk3Bfu_z, i, j, k - 2) * vector(vfd_z, i, j, k - 2);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j, k - 2) * vector(rk3gtd_zz, i, j, k - 2) - vector(rk3gtd_xz, i, j, k - 2) * vector(rk3gtd_xz, i, j, k - 2));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j, k - 2) * vector(rk3gtd_yz, i, j, k - 2)) + vector(rk3gtd_xy, i, j, k - 2) * vector(rk3gtd_xz, i, j, k - 2));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j, k - 2) + rk3gtu_yz * vector(vfd_y, i, j, k - 2) + rk3gtu_zz * vector(vfd_z, i, j, k - 2));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j, k - 2) + rk3gtu_yy * vector(rk3Sfd_y, i, j, k - 2) + rk3gtu_yz * vector(rk3Sfd_z, i, j, k - 2));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j, k - 2) + rk3gtu_yz * vector(rk3Sfd_y, i, j, k - 2) + rk3gtu_zz * vector(rk3Sfd_z, i, j, k - 2));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j, k - 2) + rk3gtu_yy * vector(vfd_y, i, j, k - 2) + rk3gtu_yz * vector(vfd_z, i, j, k - 2));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j, k - 2) + rk3vfu_y * vector(vfd_y, i, j, k - 2) + rk3vfu_z * vector(vfd_z, i, j, k - 2), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j, k - 2) * vector(rk3Bfu_z, i, j, k - 2)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j, k - 2) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j, k - 2) * vector(rk3Bfu_x, i, j, k - 2)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j, k - 2) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k - 2) * vector(rk3Bfu_z, i, j, k - 2)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k - 2) * rk3vfu_z + vector(rk3Bfu_z, i, j, k - 2) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k - 2) * vector(rk3Bfu_y, i, j, k - 2)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k - 2) * rk3vfu_y + vector(rk3Bfu_y, i, j, k - 2) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j, k - 2) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k - 2) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k - 2));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j, k - 2) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j, k - 2) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j, k - 2));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j, k - 2) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j, k - 2) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j, k - 2));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k - 2) * vector(rk3Bfu_z, i, j, k - 2)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k - 2) * rk3vfu_z + vector(rk3Bfu_z, i, j, k - 2) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j, k - 2) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j, k - 2) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j, k - 2));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k - 2) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k - 2) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j, k - 2));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j, k - 2) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j, k - 2) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j, k - 2));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j, k - 2) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j, k - 2) * vector(rk3Bfu_y, i, j, k - 2)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j, k - 2) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j, k - 2) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j, k - 2) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j, k - 2));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k - 2) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j, k - 2) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k - 2));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j, k - 2) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j, k - 2) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j, k - 2));
						Fluxm2kSfd_x = FkSfd_x_mainI(vector(rk3Sfd_x, i, j, k - 2), vector(rk3gtd_xx, i, j, k - 2), vector(rk3gtd_xy, i, j, k - 2), vector(rk3gtd_xz, i, j, k - 2), vector(rk3Betau_z, i, j, k - 2), vector(rk3Alpha, i, j, k - 2), vector(TauT_xz, i, j, k - 2), vector(TauT_yz, i, j, k - 2), vector(TauT_zz, i, j, k - 2), rk3inv_chi, rk3sdetg, rk3Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxm2kSfd_y = FkSfd_y_mainI(vector(rk3Sfd_y, i, j, k - 2), vector(rk3gtd_xy, i, j, k - 2), vector(rk3gtd_yy, i, j, k - 2), vector(rk3gtd_yz, i, j, k - 2), vector(rk3Betau_z, i, j, k - 2), vector(rk3Alpha, i, j, k - 2), vector(TauT_xz, i, j, k - 2), vector(TauT_yz, i, j, k - 2), vector(TauT_zz, i, j, k - 2), rk3inv_chi, rk3sdetg, rk3Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxm2kSfd_z = FkSfd_z_mainI(vector(rk3Sfd_z, i, j, k - 2), vector(rk3gtd_xz, i, j, k - 2), vector(rk3gtd_yz, i, j, k - 2), vector(rk3gtd_zz, i, j, k - 2), vector(rk3Betau_z, i, j, k - 2), vector(rk3Alpha, i, j, k - 2), vector(TauT_xz, i, j, k - 2), vector(TauT_yz, i, j, k - 2), vector(TauT_zz, i, j, k - 2), rk3inv_chi, rk3sdetg, rk3Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm2kBfu_x = FkBfu_x_mainI(vector(rk3Bfu_x, i, j, k - 2), vector(rk3Bfu_z, i, j, k - 2), vector(rk3phif, i, j, k - 2), vector(rk3Betau_x, i, j, k - 2), vector(rk3Betau_z, i, j, k - 2), vector(rk3Alpha, i, j, k - 2), vector(TauM_xz, i, j, k - 2), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2kBfu_y = FkBfu_y_mainI(vector(rk3Bfu_y, i, j, k - 2), vector(rk3Bfu_z, i, j, k - 2), vector(rk3phif, i, j, k - 2), vector(rk3Betau_y, i, j, k - 2), vector(rk3Betau_z, i, j, k - 2), vector(rk3Alpha, i, j, k - 2), vector(TauM_yz, i, j, k - 2), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2kBfu_z = FkBfu_z_mainI(vector(rk3phif, i, j, k - 2), vector(rk3Alpha, i, j, k - 2), rk3chi_max, rk3gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm2kDf = FkDf_mainI(vector(rk3Df, i, j, k - 2), vector(rk3Betau_z, i, j, k - 2), vector(rk3Alpha, i, j, k - 2), vector(TauN_z, i, j, k - 2), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2kDYf = FkDYf_mainI(vector(rk3DYf, i, j, k - 2), vector(rk3Betau_z, i, j, k - 2), vector(rk3Alpha, i, j, k - 2), vector(TauNe_z, i, j, k - 2), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2ktauf = Fktauf_mainI(vector(rk3Df, i, j, k - 2), vector(rk3tauf, i, j, k - 2), vector(rk3Betau_z, i, j, k - 2), vector(rk3Alpha, i, j, k - 2), vector(TauN_z, i, j, k - 2), rk3sdetg, rk3Sfu_z, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm2kphif = Fkphif_mainI(vector(rk3Bfu_z, i, j, k - 2), vector(rk3phif, i, j, k - 2), vector(rk3Betau_z, i, j, k - 2), vector(rk3Alpha, i, j, k - 2), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j, k - 2) * (1.0 + vector(epsf, i, j, k - 2)) + vector(pf, i, j, k - 2);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j, k - 2) + rk3sqca - vector(sqcs, i, j, k - 2) * rk3sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk3Betau_z, i, j, k - 2)) + vector(rk3Alpha, i, j, k - 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk3Betau_z, i, j, k - 2)) + vector(rk3Alpha, i, j, k - 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk3Betau_z, i, j, k - 2)) + vector(rk3Alpha, i, j, k - 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk3Betau_z, i, j, k - 2)) + vector(rk3Alpha, i, j, k - 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk3Betau_z, i, j, k - 2)) + vector(rk3Alpha, i, j, k - 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk3Betau_z, i, j, k - 2)) + vector(rk3Alpha, i, j, k - 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk3Betau_z, i, j, k - 2)) + vector(rk3Alpha, i, j, k - 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk3Betau_z, i, j, k - 2)) + vector(rk3Alpha, i, j, k - 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk3Betau_z, i, j, k - 2)) + vector(rk3Alpha, i, j, k - 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk3Betau_z, i, j, k - 2)) + vector(rk3Alpha, i, j, k - 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j, k - 1));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k - 1) * vector(rk3Bfu_x, i, j, k - 1) + vector(rk3gtd_yy, i, j, k - 1) * vector(rk3Bfu_y, i, j, k - 1) + vector(rk3gtd_yz, i, j, k - 1) * vector(rk3Bfu_z, i, j, k - 1));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k - 1) * vector(rk3Bfu_x, i, j, k - 1) + vector(rk3gtd_yz, i, j, k - 1) * vector(rk3Bfu_y, i, j, k - 1) + vector(rk3gtd_zz, i, j, k - 1) * vector(rk3Bfu_z, i, j, k - 1));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k - 1) * vector(rk3Bfu_x, i, j, k - 1) + vector(rk3gtd_xy, i, j, k - 1) * vector(rk3Bfu_y, i, j, k - 1) + vector(rk3gtd_xz, i, j, k - 1) * vector(rk3Bfu_z, i, j, k - 1));
						rk3sqBf = vector(rk3Bfu_x, i, j, k - 1) * rk3Bfd_x + vector(rk3Bfu_y, i, j, k - 1) * rk3Bfd_y + vector(rk3Bfu_z, i, j, k - 1) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j, k - 1) * vector(rk3gtd_zz, i, j, k - 1) * vector(rk3gtd_xx, i, j, k - 1) + (-vector(rk3gtd_yy, i, j, k - 1) * (vector(rk3gtd_xz, i, j, k - 1) * vector(rk3gtd_xz, i, j, k - 1))) + (-vector(rk3gtd_zz, i, j, k - 1) * (vector(rk3gtd_xy, i, j, k - 1) * vector(rk3gtd_xy, i, j, k - 1))) + (-(vector(rk3gtd_yz, i, j, k - 1) * vector(rk3gtd_yz, i, j, k - 1)) * vector(rk3gtd_xx, i, j, k - 1)) + 2.0 * vector(rk3gtd_yz, i, j, k - 1) * vector(rk3gtd_xy, i, j, k - 1) * vector(rk3gtd_xz, i, j, k - 1);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j, k - 1) * vector(rk3gtd_yy, i, j, k - 1) - vector(rk3gtd_xy, i, j, k - 1) * vector(rk3gtd_xy, i, j, k - 1));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j, k - 1) * vector(rk3gtd_zz, i, j, k - 1) - vector(rk3gtd_yz, i, j, k - 1) * vector(rk3gtd_yz, i, j, k - 1));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j, k - 1) * vector(rk3gtd_yz, i, j, k - 1) - vector(rk3gtd_yy, i, j, k - 1) * vector(rk3gtd_xz, i, j, k - 1));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j, k - 1) * vector(rk3gtd_zz, i, j, k - 1)) + vector(rk3gtd_yz, i, j, k - 1) * vector(rk3gtd_xz, i, j, k - 1));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j, k - 1) + rk3gtu_xy * vector(rk3Sfd_y, i, j, k - 1) + rk3gtu_xz * vector(rk3Sfd_z, i, j, k - 1));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j, k - 1) + rk3gtu_xy * vector(vfd_y, i, j, k - 1) + rk3gtu_xz * vector(vfd_z, i, j, k - 1));
						rk3Bfvf = vector(rk3Bfu_x, i, j, k - 1) * vector(vfd_x, i, j, k - 1) + vector(rk3Bfu_y, i, j, k - 1) * vector(vfd_y, i, j, k - 1) + vector(rk3Bfu_z, i, j, k - 1) * vector(vfd_z, i, j, k - 1);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j, k - 1) * vector(rk3gtd_zz, i, j, k - 1) - vector(rk3gtd_xz, i, j, k - 1) * vector(rk3gtd_xz, i, j, k - 1));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j, k - 1) * vector(rk3gtd_yz, i, j, k - 1)) + vector(rk3gtd_xy, i, j, k - 1) * vector(rk3gtd_xz, i, j, k - 1));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j, k - 1) + rk3gtu_yz * vector(vfd_y, i, j, k - 1) + rk3gtu_zz * vector(vfd_z, i, j, k - 1));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j, k - 1) + rk3gtu_yy * vector(rk3Sfd_y, i, j, k - 1) + rk3gtu_yz * vector(rk3Sfd_z, i, j, k - 1));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j, k - 1) + rk3gtu_yz * vector(rk3Sfd_y, i, j, k - 1) + rk3gtu_zz * vector(rk3Sfd_z, i, j, k - 1));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j, k - 1) + rk3gtu_yy * vector(vfd_y, i, j, k - 1) + rk3gtu_yz * vector(vfd_z, i, j, k - 1));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j, k - 1) + rk3vfu_y * vector(vfd_y, i, j, k - 1) + rk3vfu_z * vector(vfd_z, i, j, k - 1), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j, k - 1) * vector(rk3Bfu_z, i, j, k - 1)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j, k - 1) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j, k - 1) * vector(rk3Bfu_x, i, j, k - 1)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j, k - 1) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k - 1) * vector(rk3Bfu_z, i, j, k - 1)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k - 1) * rk3vfu_z + vector(rk3Bfu_z, i, j, k - 1) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k - 1) * vector(rk3Bfu_y, i, j, k - 1)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k - 1) * rk3vfu_y + vector(rk3Bfu_y, i, j, k - 1) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j, k - 1) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k - 1) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k - 1));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j, k - 1) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j, k - 1) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j, k - 1));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j, k - 1) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j, k - 1) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j, k - 1));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k - 1) * vector(rk3Bfu_z, i, j, k - 1)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k - 1) * rk3vfu_z + vector(rk3Bfu_z, i, j, k - 1) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j, k - 1) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j, k - 1) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j, k - 1));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k - 1) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k - 1) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j, k - 1));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j, k - 1) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j, k - 1) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j, k - 1));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j, k - 1) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j, k - 1) * vector(rk3Bfu_y, i, j, k - 1)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j, k - 1) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j, k - 1) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j, k - 1) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j, k - 1));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k - 1) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j, k - 1) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k - 1));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j, k - 1) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j, k - 1) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j, k - 1));
						Fluxm1kSfd_x = FkSfd_x_mainI(vector(rk3Sfd_x, i, j, k - 1), vector(rk3gtd_xx, i, j, k - 1), vector(rk3gtd_xy, i, j, k - 1), vector(rk3gtd_xz, i, j, k - 1), vector(rk3Betau_z, i, j, k - 1), vector(rk3Alpha, i, j, k - 1), vector(TauT_xz, i, j, k - 1), vector(TauT_yz, i, j, k - 1), vector(TauT_zz, i, j, k - 1), rk3inv_chi, rk3sdetg, rk3Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxm1kSfd_y = FkSfd_y_mainI(vector(rk3Sfd_y, i, j, k - 1), vector(rk3gtd_xy, i, j, k - 1), vector(rk3gtd_yy, i, j, k - 1), vector(rk3gtd_yz, i, j, k - 1), vector(rk3Betau_z, i, j, k - 1), vector(rk3Alpha, i, j, k - 1), vector(TauT_xz, i, j, k - 1), vector(TauT_yz, i, j, k - 1), vector(TauT_zz, i, j, k - 1), rk3inv_chi, rk3sdetg, rk3Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxm1kSfd_z = FkSfd_z_mainI(vector(rk3Sfd_z, i, j, k - 1), vector(rk3gtd_xz, i, j, k - 1), vector(rk3gtd_yz, i, j, k - 1), vector(rk3gtd_zz, i, j, k - 1), vector(rk3Betau_z, i, j, k - 1), vector(rk3Alpha, i, j, k - 1), vector(TauT_xz, i, j, k - 1), vector(TauT_yz, i, j, k - 1), vector(TauT_zz, i, j, k - 1), rk3inv_chi, rk3sdetg, rk3Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm1kBfu_x = FkBfu_x_mainI(vector(rk3Bfu_x, i, j, k - 1), vector(rk3Bfu_z, i, j, k - 1), vector(rk3phif, i, j, k - 1), vector(rk3Betau_x, i, j, k - 1), vector(rk3Betau_z, i, j, k - 1), vector(rk3Alpha, i, j, k - 1), vector(TauM_xz, i, j, k - 1), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1kBfu_y = FkBfu_y_mainI(vector(rk3Bfu_y, i, j, k - 1), vector(rk3Bfu_z, i, j, k - 1), vector(rk3phif, i, j, k - 1), vector(rk3Betau_y, i, j, k - 1), vector(rk3Betau_z, i, j, k - 1), vector(rk3Alpha, i, j, k - 1), vector(TauM_yz, i, j, k - 1), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1kBfu_z = FkBfu_z_mainI(vector(rk3phif, i, j, k - 1), vector(rk3Alpha, i, j, k - 1), rk3chi_max, rk3gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxm1kDf = FkDf_mainI(vector(rk3Df, i, j, k - 1), vector(rk3Betau_z, i, j, k - 1), vector(rk3Alpha, i, j, k - 1), vector(TauN_z, i, j, k - 1), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1kDYf = FkDYf_mainI(vector(rk3DYf, i, j, k - 1), vector(rk3Betau_z, i, j, k - 1), vector(rk3Alpha, i, j, k - 1), vector(TauNe_z, i, j, k - 1), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1ktauf = Fktauf_mainI(vector(rk3Df, i, j, k - 1), vector(rk3tauf, i, j, k - 1), vector(rk3Betau_z, i, j, k - 1), vector(rk3Alpha, i, j, k - 1), vector(TauN_z, i, j, k - 1), rk3sdetg, rk3Sfu_z, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxm1kphif = Fkphif_mainI(vector(rk3Bfu_z, i, j, k - 1), vector(rk3phif, i, j, k - 1), vector(rk3Betau_z, i, j, k - 1), vector(rk3Alpha, i, j, k - 1), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j, k - 1) * (1.0 + vector(epsf, i, j, k - 1)) + vector(pf, i, j, k - 1);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j, k - 1) + rk3sqca - vector(sqcs, i, j, k - 1) * rk3sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk3Betau_z, i, j, k - 1)) + vector(rk3Alpha, i, j, k - 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk3Betau_z, i, j, k - 1)) + vector(rk3Alpha, i, j, k - 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk3Betau_z, i, j, k - 1)) + vector(rk3Alpha, i, j, k - 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk3Betau_z, i, j, k - 1)) + vector(rk3Alpha, i, j, k - 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk3Betau_z, i, j, k - 1)) + vector(rk3Alpha, i, j, k - 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk3Betau_z, i, j, k - 1)) + vector(rk3Alpha, i, j, k - 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk3Betau_z, i, j, k - 1)) + vector(rk3Alpha, i, j, k - 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk3Betau_z, i, j, k - 1)) + vector(rk3Alpha, i, j, k - 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk3Betau_z, i, j, k - 1)) + vector(rk3Alpha, i, j, k - 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk3Betau_z, i, j, k - 1)) + vector(rk3Alpha, i, j, k - 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j, k + 1));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k + 1) * vector(rk3Bfu_x, i, j, k + 1) + vector(rk3gtd_yy, i, j, k + 1) * vector(rk3Bfu_y, i, j, k + 1) + vector(rk3gtd_yz, i, j, k + 1) * vector(rk3Bfu_z, i, j, k + 1));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k + 1) * vector(rk3Bfu_x, i, j, k + 1) + vector(rk3gtd_yz, i, j, k + 1) * vector(rk3Bfu_y, i, j, k + 1) + vector(rk3gtd_zz, i, j, k + 1) * vector(rk3Bfu_z, i, j, k + 1));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k + 1) * vector(rk3Bfu_x, i, j, k + 1) + vector(rk3gtd_xy, i, j, k + 1) * vector(rk3Bfu_y, i, j, k + 1) + vector(rk3gtd_xz, i, j, k + 1) * vector(rk3Bfu_z, i, j, k + 1));
						rk3sqBf = vector(rk3Bfu_x, i, j, k + 1) * rk3Bfd_x + vector(rk3Bfu_y, i, j, k + 1) * rk3Bfd_y + vector(rk3Bfu_z, i, j, k + 1) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j, k + 1) * vector(rk3gtd_zz, i, j, k + 1) * vector(rk3gtd_xx, i, j, k + 1) + (-vector(rk3gtd_yy, i, j, k + 1) * (vector(rk3gtd_xz, i, j, k + 1) * vector(rk3gtd_xz, i, j, k + 1))) + (-vector(rk3gtd_zz, i, j, k + 1) * (vector(rk3gtd_xy, i, j, k + 1) * vector(rk3gtd_xy, i, j, k + 1))) + (-(vector(rk3gtd_yz, i, j, k + 1) * vector(rk3gtd_yz, i, j, k + 1)) * vector(rk3gtd_xx, i, j, k + 1)) + 2.0 * vector(rk3gtd_yz, i, j, k + 1) * vector(rk3gtd_xy, i, j, k + 1) * vector(rk3gtd_xz, i, j, k + 1);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j, k + 1) * vector(rk3gtd_yy, i, j, k + 1) - vector(rk3gtd_xy, i, j, k + 1) * vector(rk3gtd_xy, i, j, k + 1));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j, k + 1) * vector(rk3gtd_zz, i, j, k + 1) - vector(rk3gtd_yz, i, j, k + 1) * vector(rk3gtd_yz, i, j, k + 1));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j, k + 1) * vector(rk3gtd_yz, i, j, k + 1) - vector(rk3gtd_yy, i, j, k + 1) * vector(rk3gtd_xz, i, j, k + 1));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j, k + 1) * vector(rk3gtd_zz, i, j, k + 1)) + vector(rk3gtd_yz, i, j, k + 1) * vector(rk3gtd_xz, i, j, k + 1));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j, k + 1) + rk3gtu_xy * vector(rk3Sfd_y, i, j, k + 1) + rk3gtu_xz * vector(rk3Sfd_z, i, j, k + 1));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j, k + 1) + rk3gtu_xy * vector(vfd_y, i, j, k + 1) + rk3gtu_xz * vector(vfd_z, i, j, k + 1));
						rk3Bfvf = vector(rk3Bfu_x, i, j, k + 1) * vector(vfd_x, i, j, k + 1) + vector(rk3Bfu_y, i, j, k + 1) * vector(vfd_y, i, j, k + 1) + vector(rk3Bfu_z, i, j, k + 1) * vector(vfd_z, i, j, k + 1);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j, k + 1) * vector(rk3gtd_zz, i, j, k + 1) - vector(rk3gtd_xz, i, j, k + 1) * vector(rk3gtd_xz, i, j, k + 1));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j, k + 1) * vector(rk3gtd_yz, i, j, k + 1)) + vector(rk3gtd_xy, i, j, k + 1) * vector(rk3gtd_xz, i, j, k + 1));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j, k + 1) + rk3gtu_yz * vector(vfd_y, i, j, k + 1) + rk3gtu_zz * vector(vfd_z, i, j, k + 1));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j, k + 1) + rk3gtu_yy * vector(rk3Sfd_y, i, j, k + 1) + rk3gtu_yz * vector(rk3Sfd_z, i, j, k + 1));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j, k + 1) + rk3gtu_yz * vector(rk3Sfd_y, i, j, k + 1) + rk3gtu_zz * vector(rk3Sfd_z, i, j, k + 1));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j, k + 1) + rk3gtu_yy * vector(vfd_y, i, j, k + 1) + rk3gtu_yz * vector(vfd_z, i, j, k + 1));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j, k + 1) + rk3vfu_y * vector(vfd_y, i, j, k + 1) + rk3vfu_z * vector(vfd_z, i, j, k + 1), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j, k + 1) * vector(rk3Bfu_z, i, j, k + 1)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j, k + 1) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j, k + 1) * vector(rk3Bfu_x, i, j, k + 1)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j, k + 1) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k + 1) * vector(rk3Bfu_z, i, j, k + 1)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k + 1) * rk3vfu_z + vector(rk3Bfu_z, i, j, k + 1) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k + 1) * vector(rk3Bfu_y, i, j, k + 1)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k + 1) * rk3vfu_y + vector(rk3Bfu_y, i, j, k + 1) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j, k + 1) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k + 1) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k + 1));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j, k + 1) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j, k + 1) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j, k + 1));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j, k + 1) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j, k + 1) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j, k + 1));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k + 1) * vector(rk3Bfu_z, i, j, k + 1)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k + 1) * rk3vfu_z + vector(rk3Bfu_z, i, j, k + 1) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j, k + 1) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j, k + 1) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j, k + 1));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k + 1) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k + 1) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j, k + 1));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j, k + 1) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j, k + 1) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j, k + 1));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j, k + 1) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j, k + 1) * vector(rk3Bfu_y, i, j, k + 1)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j, k + 1) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j, k + 1) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j, k + 1) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j, k + 1));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k + 1) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j, k + 1) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k + 1));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j, k + 1) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j, k + 1) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j, k + 1));
						Fluxp1kSfd_x = FkSfd_x_mainI(vector(rk3Sfd_x, i, j, k + 1), vector(rk3gtd_xx, i, j, k + 1), vector(rk3gtd_xy, i, j, k + 1), vector(rk3gtd_xz, i, j, k + 1), vector(rk3Betau_z, i, j, k + 1), vector(rk3Alpha, i, j, k + 1), vector(TauT_xz, i, j, k + 1), vector(TauT_yz, i, j, k + 1), vector(TauT_zz, i, j, k + 1), rk3inv_chi, rk3sdetg, rk3Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxp1kSfd_y = FkSfd_y_mainI(vector(rk3Sfd_y, i, j, k + 1), vector(rk3gtd_xy, i, j, k + 1), vector(rk3gtd_yy, i, j, k + 1), vector(rk3gtd_yz, i, j, k + 1), vector(rk3Betau_z, i, j, k + 1), vector(rk3Alpha, i, j, k + 1), vector(TauT_xz, i, j, k + 1), vector(TauT_yz, i, j, k + 1), vector(TauT_zz, i, j, k + 1), rk3inv_chi, rk3sdetg, rk3Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxp1kSfd_z = FkSfd_z_mainI(vector(rk3Sfd_z, i, j, k + 1), vector(rk3gtd_xz, i, j, k + 1), vector(rk3gtd_yz, i, j, k + 1), vector(rk3gtd_zz, i, j, k + 1), vector(rk3Betau_z, i, j, k + 1), vector(rk3Alpha, i, j, k + 1), vector(TauT_xz, i, j, k + 1), vector(TauT_yz, i, j, k + 1), vector(TauT_zz, i, j, k + 1), rk3inv_chi, rk3sdetg, rk3Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp1kBfu_x = FkBfu_x_mainI(vector(rk3Bfu_x, i, j, k + 1), vector(rk3Bfu_z, i, j, k + 1), vector(rk3phif, i, j, k + 1), vector(rk3Betau_x, i, j, k + 1), vector(rk3Betau_z, i, j, k + 1), vector(rk3Alpha, i, j, k + 1), vector(TauM_xz, i, j, k + 1), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1kBfu_y = FkBfu_y_mainI(vector(rk3Bfu_y, i, j, k + 1), vector(rk3Bfu_z, i, j, k + 1), vector(rk3phif, i, j, k + 1), vector(rk3Betau_y, i, j, k + 1), vector(rk3Betau_z, i, j, k + 1), vector(rk3Alpha, i, j, k + 1), vector(TauM_yz, i, j, k + 1), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1kBfu_z = FkBfu_z_mainI(vector(rk3phif, i, j, k + 1), vector(rk3Alpha, i, j, k + 1), rk3chi_max, rk3gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp1kDf = FkDf_mainI(vector(rk3Df, i, j, k + 1), vector(rk3Betau_z, i, j, k + 1), vector(rk3Alpha, i, j, k + 1), vector(TauN_z, i, j, k + 1), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1kDYf = FkDYf_mainI(vector(rk3DYf, i, j, k + 1), vector(rk3Betau_z, i, j, k + 1), vector(rk3Alpha, i, j, k + 1), vector(TauNe_z, i, j, k + 1), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1ktauf = Fktauf_mainI(vector(rk3Df, i, j, k + 1), vector(rk3tauf, i, j, k + 1), vector(rk3Betau_z, i, j, k + 1), vector(rk3Alpha, i, j, k + 1), vector(TauN_z, i, j, k + 1), rk3sdetg, rk3Sfu_z, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp1kphif = Fkphif_mainI(vector(rk3Bfu_z, i, j, k + 1), vector(rk3phif, i, j, k + 1), vector(rk3Betau_z, i, j, k + 1), vector(rk3Alpha, i, j, k + 1), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j, k + 1) * (1.0 + vector(epsf, i, j, k + 1)) + vector(pf, i, j, k + 1);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j, k + 1) + rk3sqca - vector(sqcs, i, j, k + 1) * rk3sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk3Betau_z, i, j, k + 1)) + vector(rk3Alpha, i, j, k + 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk3Betau_z, i, j, k + 1)) + vector(rk3Alpha, i, j, k + 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk3Betau_z, i, j, k + 1)) + vector(rk3Alpha, i, j, k + 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk3Betau_z, i, j, k + 1)) + vector(rk3Alpha, i, j, k + 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk3Betau_z, i, j, k + 1)) + vector(rk3Alpha, i, j, k + 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk3Betau_z, i, j, k + 1)) + vector(rk3Alpha, i, j, k + 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk3Betau_z, i, j, k + 1)) + vector(rk3Alpha, i, j, k + 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk3Betau_z, i, j, k + 1)) + vector(rk3Alpha, i, j, k + 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk3Betau_z, i, j, k + 1)) + vector(rk3Alpha, i, j, k + 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk3Betau_z, i, j, k + 1)) + vector(rk3Alpha, i, j, k + 1) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j, k + 2));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k + 2) * vector(rk3Bfu_x, i, j, k + 2) + vector(rk3gtd_yy, i, j, k + 2) * vector(rk3Bfu_y, i, j, k + 2) + vector(rk3gtd_yz, i, j, k + 2) * vector(rk3Bfu_z, i, j, k + 2));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k + 2) * vector(rk3Bfu_x, i, j, k + 2) + vector(rk3gtd_yz, i, j, k + 2) * vector(rk3Bfu_y, i, j, k + 2) + vector(rk3gtd_zz, i, j, k + 2) * vector(rk3Bfu_z, i, j, k + 2));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k + 2) * vector(rk3Bfu_x, i, j, k + 2) + vector(rk3gtd_xy, i, j, k + 2) * vector(rk3Bfu_y, i, j, k + 2) + vector(rk3gtd_xz, i, j, k + 2) * vector(rk3Bfu_z, i, j, k + 2));
						rk3sqBf = vector(rk3Bfu_x, i, j, k + 2) * rk3Bfd_x + vector(rk3Bfu_y, i, j, k + 2) * rk3Bfd_y + vector(rk3Bfu_z, i, j, k + 2) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j, k + 2) * vector(rk3gtd_zz, i, j, k + 2) * vector(rk3gtd_xx, i, j, k + 2) + (-vector(rk3gtd_yy, i, j, k + 2) * (vector(rk3gtd_xz, i, j, k + 2) * vector(rk3gtd_xz, i, j, k + 2))) + (-vector(rk3gtd_zz, i, j, k + 2) * (vector(rk3gtd_xy, i, j, k + 2) * vector(rk3gtd_xy, i, j, k + 2))) + (-(vector(rk3gtd_yz, i, j, k + 2) * vector(rk3gtd_yz, i, j, k + 2)) * vector(rk3gtd_xx, i, j, k + 2)) + 2.0 * vector(rk3gtd_yz, i, j, k + 2) * vector(rk3gtd_xy, i, j, k + 2) * vector(rk3gtd_xz, i, j, k + 2);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j, k + 2) * vector(rk3gtd_yy, i, j, k + 2) - vector(rk3gtd_xy, i, j, k + 2) * vector(rk3gtd_xy, i, j, k + 2));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j, k + 2) * vector(rk3gtd_zz, i, j, k + 2) - vector(rk3gtd_yz, i, j, k + 2) * vector(rk3gtd_yz, i, j, k + 2));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j, k + 2) * vector(rk3gtd_yz, i, j, k + 2) - vector(rk3gtd_yy, i, j, k + 2) * vector(rk3gtd_xz, i, j, k + 2));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j, k + 2) * vector(rk3gtd_zz, i, j, k + 2)) + vector(rk3gtd_yz, i, j, k + 2) * vector(rk3gtd_xz, i, j, k + 2));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j, k + 2) + rk3gtu_xy * vector(rk3Sfd_y, i, j, k + 2) + rk3gtu_xz * vector(rk3Sfd_z, i, j, k + 2));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j, k + 2) + rk3gtu_xy * vector(vfd_y, i, j, k + 2) + rk3gtu_xz * vector(vfd_z, i, j, k + 2));
						rk3Bfvf = vector(rk3Bfu_x, i, j, k + 2) * vector(vfd_x, i, j, k + 2) + vector(rk3Bfu_y, i, j, k + 2) * vector(vfd_y, i, j, k + 2) + vector(rk3Bfu_z, i, j, k + 2) * vector(vfd_z, i, j, k + 2);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j, k + 2) * vector(rk3gtd_zz, i, j, k + 2) - vector(rk3gtd_xz, i, j, k + 2) * vector(rk3gtd_xz, i, j, k + 2));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j, k + 2) * vector(rk3gtd_yz, i, j, k + 2)) + vector(rk3gtd_xy, i, j, k + 2) * vector(rk3gtd_xz, i, j, k + 2));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j, k + 2) + rk3gtu_yz * vector(vfd_y, i, j, k + 2) + rk3gtu_zz * vector(vfd_z, i, j, k + 2));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j, k + 2) + rk3gtu_yy * vector(rk3Sfd_y, i, j, k + 2) + rk3gtu_yz * vector(rk3Sfd_z, i, j, k + 2));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j, k + 2) + rk3gtu_yz * vector(rk3Sfd_y, i, j, k + 2) + rk3gtu_zz * vector(rk3Sfd_z, i, j, k + 2));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j, k + 2) + rk3gtu_yy * vector(vfd_y, i, j, k + 2) + rk3gtu_yz * vector(vfd_z, i, j, k + 2));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j, k + 2) + rk3vfu_y * vector(vfd_y, i, j, k + 2) + rk3vfu_z * vector(vfd_z, i, j, k + 2), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j, k + 2) * vector(rk3Bfu_z, i, j, k + 2)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j, k + 2) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j, k + 2) * vector(rk3Bfu_x, i, j, k + 2)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j, k + 2) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k + 2) * vector(rk3Bfu_z, i, j, k + 2)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k + 2) * rk3vfu_z + vector(rk3Bfu_z, i, j, k + 2) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k + 2) * vector(rk3Bfu_y, i, j, k + 2)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k + 2) * rk3vfu_y + vector(rk3Bfu_y, i, j, k + 2) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j, k + 2) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k + 2) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k + 2));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j, k + 2) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j, k + 2) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j, k + 2));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j, k + 2) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j, k + 2) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j, k + 2));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k + 2) * vector(rk3Bfu_z, i, j, k + 2)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k + 2) * rk3vfu_z + vector(rk3Bfu_z, i, j, k + 2) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j, k + 2) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j, k + 2) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j, k + 2));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k + 2) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k + 2) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j, k + 2));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j, k + 2) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j, k + 2) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j, k + 2));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j, k + 2) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j, k + 2) * vector(rk3Bfu_y, i, j, k + 2)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j, k + 2) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j, k + 2) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j, k + 2) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j, k + 2));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k + 2) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j, k + 2) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k + 2));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j, k + 2) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j, k + 2) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j, k + 2));
						Fluxp2kSfd_x = FkSfd_x_mainI(vector(rk3Sfd_x, i, j, k + 2), vector(rk3gtd_xx, i, j, k + 2), vector(rk3gtd_xy, i, j, k + 2), vector(rk3gtd_xz, i, j, k + 2), vector(rk3Betau_z, i, j, k + 2), vector(rk3Alpha, i, j, k + 2), vector(TauT_xz, i, j, k + 2), vector(TauT_yz, i, j, k + 2), vector(TauT_zz, i, j, k + 2), rk3inv_chi, rk3sdetg, rk3Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxp2kSfd_y = FkSfd_y_mainI(vector(rk3Sfd_y, i, j, k + 2), vector(rk3gtd_xy, i, j, k + 2), vector(rk3gtd_yy, i, j, k + 2), vector(rk3gtd_yz, i, j, k + 2), vector(rk3Betau_z, i, j, k + 2), vector(rk3Alpha, i, j, k + 2), vector(TauT_xz, i, j, k + 2), vector(TauT_yz, i, j, k + 2), vector(TauT_zz, i, j, k + 2), rk3inv_chi, rk3sdetg, rk3Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxp2kSfd_z = FkSfd_z_mainI(vector(rk3Sfd_z, i, j, k + 2), vector(rk3gtd_xz, i, j, k + 2), vector(rk3gtd_yz, i, j, k + 2), vector(rk3gtd_zz, i, j, k + 2), vector(rk3Betau_z, i, j, k + 2), vector(rk3Alpha, i, j, k + 2), vector(TauT_xz, i, j, k + 2), vector(TauT_yz, i, j, k + 2), vector(TauT_zz, i, j, k + 2), rk3inv_chi, rk3sdetg, rk3Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp2kBfu_x = FkBfu_x_mainI(vector(rk3Bfu_x, i, j, k + 2), vector(rk3Bfu_z, i, j, k + 2), vector(rk3phif, i, j, k + 2), vector(rk3Betau_x, i, j, k + 2), vector(rk3Betau_z, i, j, k + 2), vector(rk3Alpha, i, j, k + 2), vector(TauM_xz, i, j, k + 2), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2kBfu_y = FkBfu_y_mainI(vector(rk3Bfu_y, i, j, k + 2), vector(rk3Bfu_z, i, j, k + 2), vector(rk3phif, i, j, k + 2), vector(rk3Betau_y, i, j, k + 2), vector(rk3Betau_z, i, j, k + 2), vector(rk3Alpha, i, j, k + 2), vector(TauM_yz, i, j, k + 2), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2kBfu_z = FkBfu_z_mainI(vector(rk3phif, i, j, k + 2), vector(rk3Alpha, i, j, k + 2), rk3chi_max, rk3gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp2kDf = FkDf_mainI(vector(rk3Df, i, j, k + 2), vector(rk3Betau_z, i, j, k + 2), vector(rk3Alpha, i, j, k + 2), vector(TauN_z, i, j, k + 2), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2kDYf = FkDYf_mainI(vector(rk3DYf, i, j, k + 2), vector(rk3Betau_z, i, j, k + 2), vector(rk3Alpha, i, j, k + 2), vector(TauNe_z, i, j, k + 2), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2ktauf = Fktauf_mainI(vector(rk3Df, i, j, k + 2), vector(rk3tauf, i, j, k + 2), vector(rk3Betau_z, i, j, k + 2), vector(rk3Alpha, i, j, k + 2), vector(TauN_z, i, j, k + 2), rk3sdetg, rk3Sfu_z, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp2kphif = Fkphif_mainI(vector(rk3Bfu_z, i, j, k + 2), vector(rk3phif, i, j, k + 2), vector(rk3Betau_z, i, j, k + 2), vector(rk3Alpha, i, j, k + 2), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j, k + 2) * (1.0 + vector(epsf, i, j, k + 2)) + vector(pf, i, j, k + 2);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j, k + 2) + rk3sqca - vector(sqcs, i, j, k + 2) * rk3sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk3Betau_z, i, j, k + 2)) + vector(rk3Alpha, i, j, k + 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk3Betau_z, i, j, k + 2)) + vector(rk3Alpha, i, j, k + 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk3Betau_z, i, j, k + 2)) + vector(rk3Alpha, i, j, k + 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk3Betau_z, i, j, k + 2)) + vector(rk3Alpha, i, j, k + 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk3Betau_z, i, j, k + 2)) + vector(rk3Alpha, i, j, k + 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk3Betau_z, i, j, k + 2)) + vector(rk3Alpha, i, j, k + 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk3Betau_z, i, j, k + 2)) + vector(rk3Alpha, i, j, k + 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk3Betau_z, i, j, k + 2)) + vector(rk3Alpha, i, j, k + 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk3Betau_z, i, j, k + 2)) + vector(rk3Alpha, i, j, k + 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk3Betau_z, i, j, k + 2)) + vector(rk3Alpha, i, j, k + 2) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j, k + 3));
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k + 3) * vector(rk3Bfu_x, i, j, k + 3) + vector(rk3gtd_yy, i, j, k + 3) * vector(rk3Bfu_y, i, j, k + 3) + vector(rk3gtd_yz, i, j, k + 3) * vector(rk3Bfu_z, i, j, k + 3));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k + 3) * vector(rk3Bfu_x, i, j, k + 3) + vector(rk3gtd_yz, i, j, k + 3) * vector(rk3Bfu_y, i, j, k + 3) + vector(rk3gtd_zz, i, j, k + 3) * vector(rk3Bfu_z, i, j, k + 3));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k + 3) * vector(rk3Bfu_x, i, j, k + 3) + vector(rk3gtd_xy, i, j, k + 3) * vector(rk3Bfu_y, i, j, k + 3) + vector(rk3gtd_xz, i, j, k + 3) * vector(rk3Bfu_z, i, j, k + 3));
						rk3sqBf = vector(rk3Bfu_x, i, j, k + 3) * rk3Bfd_x + vector(rk3Bfu_y, i, j, k + 3) * rk3Bfd_y + vector(rk3Bfu_z, i, j, k + 3) * rk3Bfd_z;
						rk3detgtd = vector(rk3gtd_yy, i, j, k + 3) * vector(rk3gtd_zz, i, j, k + 3) * vector(rk3gtd_xx, i, j, k + 3) + (-vector(rk3gtd_yy, i, j, k + 3) * (vector(rk3gtd_xz, i, j, k + 3) * vector(rk3gtd_xz, i, j, k + 3))) + (-vector(rk3gtd_zz, i, j, k + 3) * (vector(rk3gtd_xy, i, j, k + 3) * vector(rk3gtd_xy, i, j, k + 3))) + (-(vector(rk3gtd_yz, i, j, k + 3) * vector(rk3gtd_yz, i, j, k + 3)) * vector(rk3gtd_xx, i, j, k + 3)) + 2.0 * vector(rk3gtd_yz, i, j, k + 3) * vector(rk3gtd_xy, i, j, k + 3) * vector(rk3gtd_xz, i, j, k + 3);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j, k + 3) * vector(rk3gtd_yy, i, j, k + 3) - vector(rk3gtd_xy, i, j, k + 3) * vector(rk3gtd_xy, i, j, k + 3));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j, k + 3) * vector(rk3gtd_zz, i, j, k + 3) - vector(rk3gtd_yz, i, j, k + 3) * vector(rk3gtd_yz, i, j, k + 3));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j, k + 3) * vector(rk3gtd_yz, i, j, k + 3) - vector(rk3gtd_yy, i, j, k + 3) * vector(rk3gtd_xz, i, j, k + 3));
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j, k + 3) * vector(rk3gtd_zz, i, j, k + 3)) + vector(rk3gtd_yz, i, j, k + 3) * vector(rk3gtd_xz, i, j, k + 3));
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j, k + 3) + rk3gtu_xy * vector(rk3Sfd_y, i, j, k + 3) + rk3gtu_xz * vector(rk3Sfd_z, i, j, k + 3));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j, k + 3) + rk3gtu_xy * vector(vfd_y, i, j, k + 3) + rk3gtu_xz * vector(vfd_z, i, j, k + 3));
						rk3Bfvf = vector(rk3Bfu_x, i, j, k + 3) * vector(vfd_x, i, j, k + 3) + vector(rk3Bfu_y, i, j, k + 3) * vector(vfd_y, i, j, k + 3) + vector(rk3Bfu_z, i, j, k + 3) * vector(vfd_z, i, j, k + 3);
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j, k + 3) * vector(rk3gtd_zz, i, j, k + 3) - vector(rk3gtd_xz, i, j, k + 3) * vector(rk3gtd_xz, i, j, k + 3));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j, k + 3) * vector(rk3gtd_yz, i, j, k + 3)) + vector(rk3gtd_xy, i, j, k + 3) * vector(rk3gtd_xz, i, j, k + 3));
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j, k + 3) + rk3gtu_yz * vector(vfd_y, i, j, k + 3) + rk3gtu_zz * vector(vfd_z, i, j, k + 3));
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j, k + 3) + rk3gtu_yy * vector(rk3Sfd_y, i, j, k + 3) + rk3gtu_yz * vector(rk3Sfd_z, i, j, k + 3));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j, k + 3) + rk3gtu_yz * vector(rk3Sfd_y, i, j, k + 3) + rk3gtu_zz * vector(rk3Sfd_z, i, j, k + 3));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j, k + 3) + rk3gtu_yy * vector(vfd_y, i, j, k + 3) + rk3gtu_yz * vector(vfd_z, i, j, k + 3));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j, k + 3) + rk3vfu_y * vector(vfd_y, i, j, k + 3) + rk3vfu_z * vector(vfd_z, i, j, k + 3), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j, k + 3) * vector(rk3Bfu_z, i, j, k + 3)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j, k + 3) * rk3vfu_z)) / rk3sdetg);
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j, k + 3) * vector(rk3Bfu_x, i, j, k + 3)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j, k + 3) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k + 3) * vector(rk3Bfu_z, i, j, k + 3)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k + 3) * rk3vfu_z + vector(rk3Bfu_z, i, j, k + 3) * rk3vfu_x)) / rk3sdetg);
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k + 3) * vector(rk3Bfu_y, i, j, k + 3)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k + 3) * rk3vfu_y + vector(rk3Bfu_y, i, j, k + 3) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j, k + 3) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k + 3) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k + 3));
						rk3Sfud_xy = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xy, i, j, k + 3) + rk3Sfuu_xy * vector(rk3gtd_yy, i, j, k + 3) + rk3Sfuu_xz * vector(rk3gtd_yz, i, j, k + 3));
						rk3Sfud_xz = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xz, i, j, k + 3) + rk3Sfuu_xy * vector(rk3gtd_yz, i, j, k + 3) + rk3Sfuu_xz * vector(rk3gtd_zz, i, j, k + 3));
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k + 3) * vector(rk3Bfu_z, i, j, k + 3)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k + 3) * rk3vfu_z + vector(rk3Bfu_z, i, j, k + 3) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_zy = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xy, i, j, k + 3) + rk3Sfuu_yz * vector(rk3gtd_yy, i, j, k + 3) + rk3Sfuu_zz * vector(rk3gtd_yz, i, j, k + 3));
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k + 3) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k + 3) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j, k + 3));
						rk3Sfud_zx = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xx, i, j, k + 3) + rk3Sfuu_yz * vector(rk3gtd_xy, i, j, k + 3) + rk3Sfuu_zz * vector(rk3gtd_xz, i, j, k + 3));
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j, k + 3) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j, k + 3) * vector(rk3Bfu_y, i, j, k + 3)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j, k + 3) * rk3vfu_y)) / rk3sdetg);
						rk3Sfud_yx = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xx, i, j, k + 3) + rk3Sfuu_yy * vector(rk3gtd_xy, i, j, k + 3) + rk3Sfuu_yz * vector(rk3gtd_xz, i, j, k + 3));
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k + 3) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j, k + 3) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k + 3));
						rk3Sfud_yz = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xz, i, j, k + 3) + rk3Sfuu_yy * vector(rk3gtd_yz, i, j, k + 3) + rk3Sfuu_yz * vector(rk3gtd_zz, i, j, k + 3));
						Fluxp3kSfd_x = FkSfd_x_mainI(vector(rk3Sfd_x, i, j, k + 3), vector(rk3gtd_xx, i, j, k + 3), vector(rk3gtd_xy, i, j, k + 3), vector(rk3gtd_xz, i, j, k + 3), vector(rk3Betau_z, i, j, k + 3), vector(rk3Alpha, i, j, k + 3), vector(TauT_xz, i, j, k + 3), vector(TauT_yz, i, j, k + 3), vector(TauT_zz, i, j, k + 3), rk3inv_chi, rk3sdetg, rk3Sfud_zx, dx, simPlat_dt, ilast, jlast);
						Fluxp3kSfd_y = FkSfd_y_mainI(vector(rk3Sfd_y, i, j, k + 3), vector(rk3gtd_xy, i, j, k + 3), vector(rk3gtd_yy, i, j, k + 3), vector(rk3gtd_yz, i, j, k + 3), vector(rk3Betau_z, i, j, k + 3), vector(rk3Alpha, i, j, k + 3), vector(TauT_xz, i, j, k + 3), vector(TauT_yz, i, j, k + 3), vector(TauT_zz, i, j, k + 3), rk3inv_chi, rk3sdetg, rk3Sfud_zy, dx, simPlat_dt, ilast, jlast);
						Fluxp3kSfd_z = FkSfd_z_mainI(vector(rk3Sfd_z, i, j, k + 3), vector(rk3gtd_xz, i, j, k + 3), vector(rk3gtd_yz, i, j, k + 3), vector(rk3gtd_zz, i, j, k + 3), vector(rk3Betau_z, i, j, k + 3), vector(rk3Alpha, i, j, k + 3), vector(TauT_xz, i, j, k + 3), vector(TauT_yz, i, j, k + 3), vector(TauT_zz, i, j, k + 3), rk3inv_chi, rk3sdetg, rk3Sfud_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp3kBfu_x = FkBfu_x_mainI(vector(rk3Bfu_x, i, j, k + 3), vector(rk3Bfu_z, i, j, k + 3), vector(rk3phif, i, j, k + 3), vector(rk3Betau_x, i, j, k + 3), vector(rk3Betau_z, i, j, k + 3), vector(rk3Alpha, i, j, k + 3), vector(TauM_xz, i, j, k + 3), rk3chi_max, rk3gtu_xz, rk3sdetg, rk3vfu_x, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3kBfu_y = FkBfu_y_mainI(vector(rk3Bfu_y, i, j, k + 3), vector(rk3Bfu_z, i, j, k + 3), vector(rk3phif, i, j, k + 3), vector(rk3Betau_y, i, j, k + 3), vector(rk3Betau_z, i, j, k + 3), vector(rk3Alpha, i, j, k + 3), vector(TauM_yz, i, j, k + 3), rk3chi_max, rk3gtu_yz, rk3sdetg, rk3vfu_y, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3kBfu_z = FkBfu_z_mainI(vector(rk3phif, i, j, k + 3), vector(rk3Alpha, i, j, k + 3), rk3chi_max, rk3gtu_zz, dx, simPlat_dt, ilast, jlast);
						Fluxp3kDf = FkDf_mainI(vector(rk3Df, i, j, k + 3), vector(rk3Betau_z, i, j, k + 3), vector(rk3Alpha, i, j, k + 3), vector(TauN_z, i, j, k + 3), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3kDYf = FkDYf_mainI(vector(rk3DYf, i, j, k + 3), vector(rk3Betau_z, i, j, k + 3), vector(rk3Alpha, i, j, k + 3), vector(TauNe_z, i, j, k + 3), rk3sdetg, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3ktauf = Fktauf_mainI(vector(rk3Df, i, j, k + 3), vector(rk3tauf, i, j, k + 3), vector(rk3Betau_z, i, j, k + 3), vector(rk3Alpha, i, j, k + 3), vector(TauN_z, i, j, k + 3), rk3sdetg, rk3Sfu_z, rk3vfu_z, dx, simPlat_dt, ilast, jlast);
						Fluxp3kphif = Fkphif_mainI(vector(rk3Bfu_z, i, j, k + 3), vector(rk3phif, i, j, k + 3), vector(rk3Betau_z, i, j, k + 3), vector(rk3Alpha, i, j, k + 3), ch, dx, simPlat_dt, ilast, jlast);
						rk3h = vector(rhof, i, j, k + 3) * (1.0 + vector(epsf, i, j, k + 3)) + vector(pf, i, j, k + 3);
						rk3sqbc = rk3sqBf / rk3sqW + rk3Bfvf * rk3Bfvf;
						rk3sqca = fabs(rk3sqbc / (rk3h + rk3sqbc));
						rk3sqa = vector(sqcs, i, j, k + 3) + rk3sqca - vector(sqcs, i, j, k + 3) * rk3sqca;
						SpeedkSfd_x = MAX(SpeedkSfd_x, fabs(fabs(vector(rk3Betau_z, i, j, k + 3)) + vector(rk3Alpha, i, j, k + 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkSfd_y = MAX(SpeedkSfd_y, fabs(fabs(vector(rk3Betau_z, i, j, k + 3)) + vector(rk3Alpha, i, j, k + 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkSfd_z = MAX(SpeedkSfd_z, fabs(fabs(vector(rk3Betau_z, i, j, k + 3)) + vector(rk3Alpha, i, j, k + 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_x = MAX(SpeedkBfu_x, fabs(fabs(vector(rk3Betau_z, i, j, k + 3)) + vector(rk3Alpha, i, j, k + 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_y = MAX(SpeedkBfu_y, fabs(fabs(vector(rk3Betau_z, i, j, k + 3)) + vector(rk3Alpha, i, j, k + 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkBfu_z = MAX(SpeedkBfu_z, fabs(fabs(vector(rk3Betau_z, i, j, k + 3)) + vector(rk3Alpha, i, j, k + 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkDf = MAX(SpeedkDf, fabs(fabs(vector(rk3Betau_z, i, j, k + 3)) + vector(rk3Alpha, i, j, k + 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						SpeedkDYf = MAX(SpeedkDYf, fabs(fabs(vector(rk3Betau_z, i, j, k + 3)) + vector(rk3Alpha, i, j, k + 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						Speedktauf = MAX(Speedktauf, fabs(fabs(vector(rk3Betau_z, i, j, k + 3)) + vector(rk3Alpha, i, j, k + 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						Speedkphif = MAX(Speedkphif, fabs(fabs(vector(rk3Betau_z, i, j, k + 3)) + vector(rk3Alpha, i, j, k + 3) / MAX(1.0 - rk3sqvf * rk3sqa, 0.0000000001) * (fabs((1.0 - rk3sqa) * rk3vfu_z) + sqrt(fabs(rk3sqa * (1.0 - rk3sqvf) * ((1.0 - rk3sqvf * rk3sqa) * rk3chi_max * rk3gtu_zz - (1.0 - rk3sqa) * (rk3vfu_z * rk3vfu_z)))))));
						LLFpkSfd_x = 0.5 * (FluxkSfd_x + fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k));
						LLFpkSfd_y = 0.5 * (FluxkSfd_y + fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k));
						LLFpkSfd_z = 0.5 * (FluxkSfd_z + fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k));
						LLFpkBfu_x = 0.5 * (FluxkBfu_x + fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k));
						LLFpkBfu_y = 0.5 * (FluxkBfu_y + fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k));
						LLFpkBfu_z = 0.5 * (FluxkBfu_z + fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k));
						LLFpkDf = 0.5 * (FluxkDf + fabs(SpeedkDf) * vector(rk3Df, i, j, k));
						LLFpkDYf = 0.5 * (FluxkDYf + fabs(SpeedkDYf) * vector(rk3DYf, i, j, k));
						LLFpktauf = 0.5 * (Fluxktauf + fabs(Speedktauf) * vector(rk3tauf, i, j, k));
						LLFpkphif = 0.5 * (Fluxkphif + fabs(Speedkphif) * vector(rk3phif, i, j, k));
						LLFnkSfd_x = 0.5 * (FluxkSfd_x - fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k));
						LLFnkSfd_y = 0.5 * (FluxkSfd_y - fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k));
						LLFnkSfd_z = 0.5 * (FluxkSfd_z - fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k));
						LLFnkBfu_x = 0.5 * (FluxkBfu_x - fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k));
						LLFnkBfu_y = 0.5 * (FluxkBfu_y - fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k));
						LLFnkBfu_z = 0.5 * (FluxkBfu_z - fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k));
						LLFnkDf = 0.5 * (FluxkDf - fabs(SpeedkDf) * vector(rk3Df, i, j, k));
						LLFnkDYf = 0.5 * (FluxkDYf - fabs(SpeedkDYf) * vector(rk3DYf, i, j, k));
						LLFnktauf = 0.5 * (Fluxktauf - fabs(Speedktauf) * vector(rk3tauf, i, j, k));
						LLFnkphif = 0.5 * (Fluxkphif - fabs(Speedkphif) * vector(rk3phif, i, j, k));
						LLFp_m3kSfd_x = 0.5 * (Fluxm3kSfd_x + fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k - 3));
						LLFp_m3kSfd_y = 0.5 * (Fluxm3kSfd_y + fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k - 3));
						LLFp_m3kSfd_z = 0.5 * (Fluxm3kSfd_z + fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k - 3));
						LLFp_m3kBfu_x = 0.5 * (Fluxm3kBfu_x + fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k - 3));
						LLFp_m3kBfu_y = 0.5 * (Fluxm3kBfu_y + fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k - 3));
						LLFp_m3kBfu_z = 0.5 * (Fluxm3kBfu_z + fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k - 3));
						LLFp_m3kDf = 0.5 * (Fluxm3kDf + fabs(SpeedkDf) * vector(rk3Df, i, j, k - 3));
						LLFp_m3kDYf = 0.5 * (Fluxm3kDYf + fabs(SpeedkDYf) * vector(rk3DYf, i, j, k - 3));
						LLFp_m3ktauf = 0.5 * (Fluxm3ktauf + fabs(Speedktauf) * vector(rk3tauf, i, j, k - 3));
						LLFp_m3kphif = 0.5 * (Fluxm3kphif + fabs(Speedkphif) * vector(rk3phif, i, j, k - 3));
						LLFn_m3kSfd_x = 0.5 * (Fluxm3kSfd_x - fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k - 3));
						LLFn_m3kSfd_y = 0.5 * (Fluxm3kSfd_y - fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k - 3));
						LLFn_m3kSfd_z = 0.5 * (Fluxm3kSfd_z - fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k - 3));
						LLFn_m3kBfu_x = 0.5 * (Fluxm3kBfu_x - fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k - 3));
						LLFn_m3kBfu_y = 0.5 * (Fluxm3kBfu_y - fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k - 3));
						LLFn_m3kBfu_z = 0.5 * (Fluxm3kBfu_z - fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k - 3));
						LLFn_m3kDf = 0.5 * (Fluxm3kDf - fabs(SpeedkDf) * vector(rk3Df, i, j, k - 3));
						LLFn_m3kDYf = 0.5 * (Fluxm3kDYf - fabs(SpeedkDYf) * vector(rk3DYf, i, j, k - 3));
						LLFn_m3ktauf = 0.5 * (Fluxm3ktauf - fabs(Speedktauf) * vector(rk3tauf, i, j, k - 3));
						LLFn_m3kphif = 0.5 * (Fluxm3kphif - fabs(Speedkphif) * vector(rk3phif, i, j, k - 3));
						LLFp_m2kSfd_x = 0.5 * (Fluxm2kSfd_x + fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k - 2));
						LLFp_m2kSfd_y = 0.5 * (Fluxm2kSfd_y + fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k - 2));
						LLFp_m2kSfd_z = 0.5 * (Fluxm2kSfd_z + fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k - 2));
						LLFp_m2kBfu_x = 0.5 * (Fluxm2kBfu_x + fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k - 2));
						LLFp_m2kBfu_y = 0.5 * (Fluxm2kBfu_y + fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k - 2));
						LLFp_m2kBfu_z = 0.5 * (Fluxm2kBfu_z + fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k - 2));
						LLFp_m2kDf = 0.5 * (Fluxm2kDf + fabs(SpeedkDf) * vector(rk3Df, i, j, k - 2));
						LLFp_m2kDYf = 0.5 * (Fluxm2kDYf + fabs(SpeedkDYf) * vector(rk3DYf, i, j, k - 2));
						LLFp_m2ktauf = 0.5 * (Fluxm2ktauf + fabs(Speedktauf) * vector(rk3tauf, i, j, k - 2));
						LLFp_m2kphif = 0.5 * (Fluxm2kphif + fabs(Speedkphif) * vector(rk3phif, i, j, k - 2));
						LLFn_m2kSfd_x = 0.5 * (Fluxm2kSfd_x - fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k - 2));
						LLFn_m2kSfd_y = 0.5 * (Fluxm2kSfd_y - fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k - 2));
						LLFn_m2kSfd_z = 0.5 * (Fluxm2kSfd_z - fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k - 2));
						LLFn_m2kBfu_x = 0.5 * (Fluxm2kBfu_x - fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k - 2));
						LLFn_m2kBfu_y = 0.5 * (Fluxm2kBfu_y - fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k - 2));
						LLFn_m2kBfu_z = 0.5 * (Fluxm2kBfu_z - fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k - 2));
						LLFn_m2kDf = 0.5 * (Fluxm2kDf - fabs(SpeedkDf) * vector(rk3Df, i, j, k - 2));
						LLFn_m2kDYf = 0.5 * (Fluxm2kDYf - fabs(SpeedkDYf) * vector(rk3DYf, i, j, k - 2));
						LLFn_m2ktauf = 0.5 * (Fluxm2ktauf - fabs(Speedktauf) * vector(rk3tauf, i, j, k - 2));
						LLFn_m2kphif = 0.5 * (Fluxm2kphif - fabs(Speedkphif) * vector(rk3phif, i, j, k - 2));
						LLFp_m1kSfd_x = 0.5 * (Fluxm1kSfd_x + fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k - 1));
						LLFp_m1kSfd_y = 0.5 * (Fluxm1kSfd_y + fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k - 1));
						LLFp_m1kSfd_z = 0.5 * (Fluxm1kSfd_z + fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k - 1));
						LLFp_m1kBfu_x = 0.5 * (Fluxm1kBfu_x + fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k - 1));
						LLFp_m1kBfu_y = 0.5 * (Fluxm1kBfu_y + fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k - 1));
						LLFp_m1kBfu_z = 0.5 * (Fluxm1kBfu_z + fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k - 1));
						LLFp_m1kDf = 0.5 * (Fluxm1kDf + fabs(SpeedkDf) * vector(rk3Df, i, j, k - 1));
						LLFp_m1kDYf = 0.5 * (Fluxm1kDYf + fabs(SpeedkDYf) * vector(rk3DYf, i, j, k - 1));
						LLFp_m1ktauf = 0.5 * (Fluxm1ktauf + fabs(Speedktauf) * vector(rk3tauf, i, j, k - 1));
						LLFp_m1kphif = 0.5 * (Fluxm1kphif + fabs(Speedkphif) * vector(rk3phif, i, j, k - 1));
						LLFn_m1kSfd_x = 0.5 * (Fluxm1kSfd_x - fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k - 1));
						LLFn_m1kSfd_y = 0.5 * (Fluxm1kSfd_y - fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k - 1));
						LLFn_m1kSfd_z = 0.5 * (Fluxm1kSfd_z - fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k - 1));
						LLFn_m1kBfu_x = 0.5 * (Fluxm1kBfu_x - fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k - 1));
						LLFn_m1kBfu_y = 0.5 * (Fluxm1kBfu_y - fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k - 1));
						LLFn_m1kBfu_z = 0.5 * (Fluxm1kBfu_z - fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k - 1));
						LLFn_m1kDf = 0.5 * (Fluxm1kDf - fabs(SpeedkDf) * vector(rk3Df, i, j, k - 1));
						LLFn_m1kDYf = 0.5 * (Fluxm1kDYf - fabs(SpeedkDYf) * vector(rk3DYf, i, j, k - 1));
						LLFn_m1ktauf = 0.5 * (Fluxm1ktauf - fabs(Speedktauf) * vector(rk3tauf, i, j, k - 1));
						LLFn_m1kphif = 0.5 * (Fluxm1kphif - fabs(Speedkphif) * vector(rk3phif, i, j, k - 1));
						LLFp_p1kSfd_x = 0.5 * (Fluxp1kSfd_x + fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k + 1));
						LLFp_p1kSfd_y = 0.5 * (Fluxp1kSfd_y + fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k + 1));
						LLFp_p1kSfd_z = 0.5 * (Fluxp1kSfd_z + fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k + 1));
						LLFp_p1kBfu_x = 0.5 * (Fluxp1kBfu_x + fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k + 1));
						LLFp_p1kBfu_y = 0.5 * (Fluxp1kBfu_y + fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k + 1));
						LLFp_p1kBfu_z = 0.5 * (Fluxp1kBfu_z + fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k + 1));
						LLFp_p1kDf = 0.5 * (Fluxp1kDf + fabs(SpeedkDf) * vector(rk3Df, i, j, k + 1));
						LLFp_p1kDYf = 0.5 * (Fluxp1kDYf + fabs(SpeedkDYf) * vector(rk3DYf, i, j, k + 1));
						LLFp_p1ktauf = 0.5 * (Fluxp1ktauf + fabs(Speedktauf) * vector(rk3tauf, i, j, k + 1));
						LLFp_p1kphif = 0.5 * (Fluxp1kphif + fabs(Speedkphif) * vector(rk3phif, i, j, k + 1));
						LLFn_p1kSfd_x = 0.5 * (Fluxp1kSfd_x - fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k + 1));
						LLFn_p1kSfd_y = 0.5 * (Fluxp1kSfd_y - fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k + 1));
						LLFn_p1kSfd_z = 0.5 * (Fluxp1kSfd_z - fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k + 1));
						LLFn_p1kBfu_x = 0.5 * (Fluxp1kBfu_x - fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k + 1));
						LLFn_p1kBfu_y = 0.5 * (Fluxp1kBfu_y - fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k + 1));
						LLFn_p1kBfu_z = 0.5 * (Fluxp1kBfu_z - fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k + 1));
						LLFn_p1kDf = 0.5 * (Fluxp1kDf - fabs(SpeedkDf) * vector(rk3Df, i, j, k + 1));
						LLFn_p1kDYf = 0.5 * (Fluxp1kDYf - fabs(SpeedkDYf) * vector(rk3DYf, i, j, k + 1));
						LLFn_p1ktauf = 0.5 * (Fluxp1ktauf - fabs(Speedktauf) * vector(rk3tauf, i, j, k + 1));
						LLFn_p1kphif = 0.5 * (Fluxp1kphif - fabs(Speedkphif) * vector(rk3phif, i, j, k + 1));
						LLFp_p2kSfd_x = 0.5 * (Fluxp2kSfd_x + fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k + 2));
						LLFp_p2kSfd_y = 0.5 * (Fluxp2kSfd_y + fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k + 2));
						LLFp_p2kSfd_z = 0.5 * (Fluxp2kSfd_z + fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k + 2));
						LLFp_p2kBfu_x = 0.5 * (Fluxp2kBfu_x + fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k + 2));
						LLFp_p2kBfu_y = 0.5 * (Fluxp2kBfu_y + fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k + 2));
						LLFp_p2kBfu_z = 0.5 * (Fluxp2kBfu_z + fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k + 2));
						LLFp_p2kDf = 0.5 * (Fluxp2kDf + fabs(SpeedkDf) * vector(rk3Df, i, j, k + 2));
						LLFp_p2kDYf = 0.5 * (Fluxp2kDYf + fabs(SpeedkDYf) * vector(rk3DYf, i, j, k + 2));
						LLFp_p2ktauf = 0.5 * (Fluxp2ktauf + fabs(Speedktauf) * vector(rk3tauf, i, j, k + 2));
						LLFp_p2kphif = 0.5 * (Fluxp2kphif + fabs(Speedkphif) * vector(rk3phif, i, j, k + 2));
						LLFn_p2kSfd_x = 0.5 * (Fluxp2kSfd_x - fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k + 2));
						LLFn_p2kSfd_y = 0.5 * (Fluxp2kSfd_y - fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k + 2));
						LLFn_p2kSfd_z = 0.5 * (Fluxp2kSfd_z - fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k + 2));
						LLFn_p2kBfu_x = 0.5 * (Fluxp2kBfu_x - fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k + 2));
						LLFn_p2kBfu_y = 0.5 * (Fluxp2kBfu_y - fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k + 2));
						LLFn_p2kBfu_z = 0.5 * (Fluxp2kBfu_z - fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k + 2));
						LLFn_p2kDf = 0.5 * (Fluxp2kDf - fabs(SpeedkDf) * vector(rk3Df, i, j, k + 2));
						LLFn_p2kDYf = 0.5 * (Fluxp2kDYf - fabs(SpeedkDYf) * vector(rk3DYf, i, j, k + 2));
						LLFn_p2ktauf = 0.5 * (Fluxp2ktauf - fabs(Speedktauf) * vector(rk3tauf, i, j, k + 2));
						LLFn_p2kphif = 0.5 * (Fluxp2kphif - fabs(Speedkphif) * vector(rk3phif, i, j, k + 2));
						LLFp_p3kSfd_x = 0.5 * (Fluxp3kSfd_x + fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k + 3));
						LLFp_p3kSfd_y = 0.5 * (Fluxp3kSfd_y + fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k + 3));
						LLFp_p3kSfd_z = 0.5 * (Fluxp3kSfd_z + fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k + 3));
						LLFp_p3kBfu_x = 0.5 * (Fluxp3kBfu_x + fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k + 3));
						LLFp_p3kBfu_y = 0.5 * (Fluxp3kBfu_y + fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k + 3));
						LLFp_p3kBfu_z = 0.5 * (Fluxp3kBfu_z + fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k + 3));
						LLFp_p3kDf = 0.5 * (Fluxp3kDf + fabs(SpeedkDf) * vector(rk3Df, i, j, k + 3));
						LLFp_p3kDYf = 0.5 * (Fluxp3kDYf + fabs(SpeedkDYf) * vector(rk3DYf, i, j, k + 3));
						LLFp_p3ktauf = 0.5 * (Fluxp3ktauf + fabs(Speedktauf) * vector(rk3tauf, i, j, k + 3));
						LLFp_p3kphif = 0.5 * (Fluxp3kphif + fabs(Speedkphif) * vector(rk3phif, i, j, k + 3));
						LLFn_p3kSfd_x = 0.5 * (Fluxp3kSfd_x - fabs(SpeedkSfd_x) * vector(rk3Sfd_x, i, j, k + 3));
						LLFn_p3kSfd_y = 0.5 * (Fluxp3kSfd_y - fabs(SpeedkSfd_y) * vector(rk3Sfd_y, i, j, k + 3));
						LLFn_p3kSfd_z = 0.5 * (Fluxp3kSfd_z - fabs(SpeedkSfd_z) * vector(rk3Sfd_z, i, j, k + 3));
						LLFn_p3kBfu_x = 0.5 * (Fluxp3kBfu_x - fabs(SpeedkBfu_x) * vector(rk3Bfu_x, i, j, k + 3));
						LLFn_p3kBfu_y = 0.5 * (Fluxp3kBfu_y - fabs(SpeedkBfu_y) * vector(rk3Bfu_y, i, j, k + 3));
						LLFn_p3kBfu_z = 0.5 * (Fluxp3kBfu_z - fabs(SpeedkBfu_z) * vector(rk3Bfu_z, i, j, k + 3));
						LLFn_p3kDf = 0.5 * (Fluxp3kDf - fabs(SpeedkDf) * vector(rk3Df, i, j, k + 3));
						LLFn_p3kDYf = 0.5 * (Fluxp3kDYf - fabs(SpeedkDYf) * vector(rk3DYf, i, j, k + 3));
						LLFn_p3ktauf = 0.5 * (Fluxp3ktauf - fabs(Speedktauf) * vector(rk3tauf, i, j, k + 3));
						LLFn_p3kphif = 0.5 * (Fluxp3kphif - fabs(Speedkphif) * vector(rk3phif, i, j, k + 3));
						Sfd_x_lni = MP5(LLFp_m3iSfd_x, LLFp_m2iSfd_x, LLFp_m1iSfd_x, LLFpiSfd_x, LLFp_p1iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lni = MP5(LLFp_m3iSfd_y, LLFp_m2iSfd_y, LLFp_m1iSfd_y, LLFpiSfd_y, LLFp_p1iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lni = MP5(LLFp_m3iSfd_z, LLFp_m2iSfd_z, LLFp_m1iSfd_z, LLFpiSfd_z, LLFp_p1iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lni = MP5(LLFp_m3iBfu_x, LLFp_m2iBfu_x, LLFp_m1iBfu_x, LLFpiBfu_x, LLFp_p1iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lni = MP5(LLFp_m3iBfu_y, LLFp_m2iBfu_y, LLFp_m1iBfu_y, LLFpiBfu_y, LLFp_p1iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lni = MP5(LLFp_m3iBfu_z, LLFp_m2iBfu_z, LLFp_m1iBfu_z, LLFpiBfu_z, LLFp_p1iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lni = MP5(LLFp_m3iDf, LLFp_m2iDf, LLFp_m1iDf, LLFpiDf, LLFp_p1iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lni = MP5(LLFp_m3iDYf, LLFp_m2iDYf, LLFp_m1iDYf, LLFpiDYf, LLFp_p1iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lni = MP5(LLFp_m3itauf, LLFp_m2itauf, LLFp_m1itauf, LLFpitauf, LLFp_p1itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lni = MP5(LLFp_m3iphif, LLFp_m2iphif, LLFp_m1iphif, LLFpiphif, LLFp_p1iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lnj = MP5(LLFp_m3jSfd_x, LLFp_m2jSfd_x, LLFp_m1jSfd_x, LLFpjSfd_x, LLFp_p1jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lnj = MP5(LLFp_m3jSfd_y, LLFp_m2jSfd_y, LLFp_m1jSfd_y, LLFpjSfd_y, LLFp_p1jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lnj = MP5(LLFp_m3jSfd_z, LLFp_m2jSfd_z, LLFp_m1jSfd_z, LLFpjSfd_z, LLFp_p1jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lnj = MP5(LLFp_m3jBfu_x, LLFp_m2jBfu_x, LLFp_m1jBfu_x, LLFpjBfu_x, LLFp_p1jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lnj = MP5(LLFp_m3jBfu_y, LLFp_m2jBfu_y, LLFp_m1jBfu_y, LLFpjBfu_y, LLFp_p1jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lnj = MP5(LLFp_m3jBfu_z, LLFp_m2jBfu_z, LLFp_m1jBfu_z, LLFpjBfu_z, LLFp_p1jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lnj = MP5(LLFp_m3jDf, LLFp_m2jDf, LLFp_m1jDf, LLFpjDf, LLFp_p1jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lnj = MP5(LLFp_m3jDYf, LLFp_m2jDYf, LLFp_m1jDYf, LLFpjDYf, LLFp_p1jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lnj = MP5(LLFp_m3jtauf, LLFp_m2jtauf, LLFp_m1jtauf, LLFpjtauf, LLFp_p1jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lnj = MP5(LLFp_m3jphif, LLFp_m2jphif, LLFp_m1jphif, LLFpjphif, LLFp_p1jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lnk = MP5(LLFp_m3kSfd_x, LLFp_m2kSfd_x, LLFp_m1kSfd_x, LLFpkSfd_x, LLFp_p1kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lnk = MP5(LLFp_m3kSfd_y, LLFp_m2kSfd_y, LLFp_m1kSfd_y, LLFpkSfd_y, LLFp_p1kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lnk = MP5(LLFp_m3kSfd_z, LLFp_m2kSfd_z, LLFp_m1kSfd_z, LLFpkSfd_z, LLFp_p1kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lnk = MP5(LLFp_m3kBfu_x, LLFp_m2kBfu_x, LLFp_m1kBfu_x, LLFpkBfu_x, LLFp_p1kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lnk = MP5(LLFp_m3kBfu_y, LLFp_m2kBfu_y, LLFp_m1kBfu_y, LLFpkBfu_y, LLFp_p1kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lnk = MP5(LLFp_m3kBfu_z, LLFp_m2kBfu_z, LLFp_m1kBfu_z, LLFpkBfu_z, LLFp_p1kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lnk = MP5(LLFp_m3kDf, LLFp_m2kDf, LLFp_m1kDf, LLFpkDf, LLFp_p1kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lnk = MP5(LLFp_m3kDYf, LLFp_m2kDYf, LLFp_m1kDYf, LLFpkDYf, LLFp_p1kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lnk = MP5(LLFp_m3ktauf, LLFp_m2ktauf, LLFp_m1ktauf, LLFpktauf, LLFp_p1ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lnk = MP5(LLFp_m3kphif, LLFp_m2kphif, LLFp_m1kphif, LLFpkphif, LLFp_p1kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rni = MP5(LLFn_p2iSfd_x, LLFn_p1iSfd_x, LLFniSfd_x, LLFn_m1iSfd_x, LLFn_m2iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rni = MP5(LLFn_p2iSfd_y, LLFn_p1iSfd_y, LLFniSfd_y, LLFn_m1iSfd_y, LLFn_m2iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rni = MP5(LLFn_p2iSfd_z, LLFn_p1iSfd_z, LLFniSfd_z, LLFn_m1iSfd_z, LLFn_m2iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rni = MP5(LLFn_p2iBfu_x, LLFn_p1iBfu_x, LLFniBfu_x, LLFn_m1iBfu_x, LLFn_m2iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rni = MP5(LLFn_p2iBfu_y, LLFn_p1iBfu_y, LLFniBfu_y, LLFn_m1iBfu_y, LLFn_m2iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rni = MP5(LLFn_p2iBfu_z, LLFn_p1iBfu_z, LLFniBfu_z, LLFn_m1iBfu_z, LLFn_m2iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rni = MP5(LLFn_p2iDf, LLFn_p1iDf, LLFniDf, LLFn_m1iDf, LLFn_m2iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rni = MP5(LLFn_p2iDYf, LLFn_p1iDYf, LLFniDYf, LLFn_m1iDYf, LLFn_m2iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rni = MP5(LLFn_p2itauf, LLFn_p1itauf, LLFnitauf, LLFn_m1itauf, LLFn_m2itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rni = MP5(LLFn_p2iphif, LLFn_p1iphif, LLFniphif, LLFn_m1iphif, LLFn_m2iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rnj = MP5(LLFn_p2jSfd_x, LLFn_p1jSfd_x, LLFnjSfd_x, LLFn_m1jSfd_x, LLFn_m2jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rnj = MP5(LLFn_p2jSfd_y, LLFn_p1jSfd_y, LLFnjSfd_y, LLFn_m1jSfd_y, LLFn_m2jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rnj = MP5(LLFn_p2jSfd_z, LLFn_p1jSfd_z, LLFnjSfd_z, LLFn_m1jSfd_z, LLFn_m2jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rnj = MP5(LLFn_p2jBfu_x, LLFn_p1jBfu_x, LLFnjBfu_x, LLFn_m1jBfu_x, LLFn_m2jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rnj = MP5(LLFn_p2jBfu_y, LLFn_p1jBfu_y, LLFnjBfu_y, LLFn_m1jBfu_y, LLFn_m2jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rnj = MP5(LLFn_p2jBfu_z, LLFn_p1jBfu_z, LLFnjBfu_z, LLFn_m1jBfu_z, LLFn_m2jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rnj = MP5(LLFn_p2jDf, LLFn_p1jDf, LLFnjDf, LLFn_m1jDf, LLFn_m2jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rnj = MP5(LLFn_p2jDYf, LLFn_p1jDYf, LLFnjDYf, LLFn_m1jDYf, LLFn_m2jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rnj = MP5(LLFn_p2jtauf, LLFn_p1jtauf, LLFnjtauf, LLFn_m1jtauf, LLFn_m2jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rnj = MP5(LLFn_p2jphif, LLFn_p1jphif, LLFnjphif, LLFn_m1jphif, LLFn_m2jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rnk = MP5(LLFn_p2kSfd_x, LLFn_p1kSfd_x, LLFnkSfd_x, LLFn_m1kSfd_x, LLFn_m2kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rnk = MP5(LLFn_p2kSfd_y, LLFn_p1kSfd_y, LLFnkSfd_y, LLFn_m1kSfd_y, LLFn_m2kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rnk = MP5(LLFn_p2kSfd_z, LLFn_p1kSfd_z, LLFnkSfd_z, LLFn_m1kSfd_z, LLFn_m2kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rnk = MP5(LLFn_p2kBfu_x, LLFn_p1kBfu_x, LLFnkBfu_x, LLFn_m1kBfu_x, LLFn_m2kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rnk = MP5(LLFn_p2kBfu_y, LLFn_p1kBfu_y, LLFnkBfu_y, LLFn_m1kBfu_y, LLFn_m2kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rnk = MP5(LLFn_p2kBfu_z, LLFn_p1kBfu_z, LLFnkBfu_z, LLFn_m1kBfu_z, LLFn_m2kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rnk = MP5(LLFn_p2kDf, LLFn_p1kDf, LLFnkDf, LLFn_m1kDf, LLFn_m2kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rnk = MP5(LLFn_p2kDYf, LLFn_p1kDYf, LLFnkDYf, LLFn_m1kDYf, LLFn_m2kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rnk = MP5(LLFn_p2ktauf, LLFn_p1ktauf, LLFnktauf, LLFn_m1ktauf, LLFn_m2ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rnk = MP5(LLFn_p2kphif, LLFn_p1kphif, LLFnkphif, LLFn_m1kphif, LLFn_m2kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lpi = MP5(LLFp_m2iSfd_x, LLFp_m1iSfd_x, LLFpiSfd_x, LLFp_p1iSfd_x, LLFp_p2iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lpi = MP5(LLFp_m2iSfd_y, LLFp_m1iSfd_y, LLFpiSfd_y, LLFp_p1iSfd_y, LLFp_p2iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lpi = MP5(LLFp_m2iSfd_z, LLFp_m1iSfd_z, LLFpiSfd_z, LLFp_p1iSfd_z, LLFp_p2iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lpi = MP5(LLFp_m2iBfu_x, LLFp_m1iBfu_x, LLFpiBfu_x, LLFp_p1iBfu_x, LLFp_p2iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lpi = MP5(LLFp_m2iBfu_y, LLFp_m1iBfu_y, LLFpiBfu_y, LLFp_p1iBfu_y, LLFp_p2iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lpi = MP5(LLFp_m2iBfu_z, LLFp_m1iBfu_z, LLFpiBfu_z, LLFp_p1iBfu_z, LLFp_p2iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lpi = MP5(LLFp_m2iDf, LLFp_m1iDf, LLFpiDf, LLFp_p1iDf, LLFp_p2iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lpi = MP5(LLFp_m2iDYf, LLFp_m1iDYf, LLFpiDYf, LLFp_p1iDYf, LLFp_p2iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lpi = MP5(LLFp_m2itauf, LLFp_m1itauf, LLFpitauf, LLFp_p1itauf, LLFp_p2itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lpi = MP5(LLFp_m2iphif, LLFp_m1iphif, LLFpiphif, LLFp_p1iphif, LLFp_p2iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lpj = MP5(LLFp_m2jSfd_x, LLFp_m1jSfd_x, LLFpjSfd_x, LLFp_p1jSfd_x, LLFp_p2jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lpj = MP5(LLFp_m2jSfd_y, LLFp_m1jSfd_y, LLFpjSfd_y, LLFp_p1jSfd_y, LLFp_p2jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lpj = MP5(LLFp_m2jSfd_z, LLFp_m1jSfd_z, LLFpjSfd_z, LLFp_p1jSfd_z, LLFp_p2jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lpj = MP5(LLFp_m2jBfu_x, LLFp_m1jBfu_x, LLFpjBfu_x, LLFp_p1jBfu_x, LLFp_p2jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lpj = MP5(LLFp_m2jBfu_y, LLFp_m1jBfu_y, LLFpjBfu_y, LLFp_p1jBfu_y, LLFp_p2jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lpj = MP5(LLFp_m2jBfu_z, LLFp_m1jBfu_z, LLFpjBfu_z, LLFp_p1jBfu_z, LLFp_p2jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lpj = MP5(LLFp_m2jDf, LLFp_m1jDf, LLFpjDf, LLFp_p1jDf, LLFp_p2jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lpj = MP5(LLFp_m2jDYf, LLFp_m1jDYf, LLFpjDYf, LLFp_p1jDYf, LLFp_p2jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lpj = MP5(LLFp_m2jtauf, LLFp_m1jtauf, LLFpjtauf, LLFp_p1jtauf, LLFp_p2jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lpj = MP5(LLFp_m2jphif, LLFp_m1jphif, LLFpjphif, LLFp_p1jphif, LLFp_p2jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_lpk = MP5(LLFp_m2kSfd_x, LLFp_m1kSfd_x, LLFpkSfd_x, LLFp_p1kSfd_x, LLFp_p2kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_lpk = MP5(LLFp_m2kSfd_y, LLFp_m1kSfd_y, LLFpkSfd_y, LLFp_p1kSfd_y, LLFp_p2kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_lpk = MP5(LLFp_m2kSfd_z, LLFp_m1kSfd_z, LLFpkSfd_z, LLFp_p1kSfd_z, LLFp_p2kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_lpk = MP5(LLFp_m2kBfu_x, LLFp_m1kBfu_x, LLFpkBfu_x, LLFp_p1kBfu_x, LLFp_p2kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_lpk = MP5(LLFp_m2kBfu_y, LLFp_m1kBfu_y, LLFpkBfu_y, LLFp_p1kBfu_y, LLFp_p2kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_lpk = MP5(LLFp_m2kBfu_z, LLFp_m1kBfu_z, LLFpkBfu_z, LLFp_p1kBfu_z, LLFp_p2kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_lpk = MP5(LLFp_m2kDf, LLFp_m1kDf, LLFpkDf, LLFp_p1kDf, LLFp_p2kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_lpk = MP5(LLFp_m2kDYf, LLFp_m1kDYf, LLFpkDYf, LLFp_p1kDYf, LLFp_p2kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_lpk = MP5(LLFp_m2ktauf, LLFp_m1ktauf, LLFpktauf, LLFp_p1ktauf, LLFp_p2ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_lpk = MP5(LLFp_m2kphif, LLFp_m1kphif, LLFpkphif, LLFp_p1kphif, LLFp_p2kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rpi = MP5(LLFn_p3iSfd_x, LLFn_p2iSfd_x, LLFn_p1iSfd_x, LLFniSfd_x, LLFn_m1iSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rpi = MP5(LLFn_p3iSfd_y, LLFn_p2iSfd_y, LLFn_p1iSfd_y, LLFniSfd_y, LLFn_m1iSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rpi = MP5(LLFn_p3iSfd_z, LLFn_p2iSfd_z, LLFn_p1iSfd_z, LLFniSfd_z, LLFn_m1iSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rpi = MP5(LLFn_p3iBfu_x, LLFn_p2iBfu_x, LLFn_p1iBfu_x, LLFniBfu_x, LLFn_m1iBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rpi = MP5(LLFn_p3iBfu_y, LLFn_p2iBfu_y, LLFn_p1iBfu_y, LLFniBfu_y, LLFn_m1iBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rpi = MP5(LLFn_p3iBfu_z, LLFn_p2iBfu_z, LLFn_p1iBfu_z, LLFniBfu_z, LLFn_m1iBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rpi = MP5(LLFn_p3iDf, LLFn_p2iDf, LLFn_p1iDf, LLFniDf, LLFn_m1iDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rpi = MP5(LLFn_p3iDYf, LLFn_p2iDYf, LLFn_p1iDYf, LLFniDYf, LLFn_m1iDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rpi = MP5(LLFn_p3itauf, LLFn_p2itauf, LLFn_p1itauf, LLFnitauf, LLFn_m1itauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rpi = MP5(LLFn_p3iphif, LLFn_p2iphif, LLFn_p1iphif, LLFniphif, LLFn_m1iphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rpj = MP5(LLFn_p3jSfd_x, LLFn_p2jSfd_x, LLFn_p1jSfd_x, LLFnjSfd_x, LLFn_m1jSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rpj = MP5(LLFn_p3jSfd_y, LLFn_p2jSfd_y, LLFn_p1jSfd_y, LLFnjSfd_y, LLFn_m1jSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rpj = MP5(LLFn_p3jSfd_z, LLFn_p2jSfd_z, LLFn_p1jSfd_z, LLFnjSfd_z, LLFn_m1jSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rpj = MP5(LLFn_p3jBfu_x, LLFn_p2jBfu_x, LLFn_p1jBfu_x, LLFnjBfu_x, LLFn_m1jBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rpj = MP5(LLFn_p3jBfu_y, LLFn_p2jBfu_y, LLFn_p1jBfu_y, LLFnjBfu_y, LLFn_m1jBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rpj = MP5(LLFn_p3jBfu_z, LLFn_p2jBfu_z, LLFn_p1jBfu_z, LLFnjBfu_z, LLFn_m1jBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rpj = MP5(LLFn_p3jDf, LLFn_p2jDf, LLFn_p1jDf, LLFnjDf, LLFn_m1jDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rpj = MP5(LLFn_p3jDYf, LLFn_p2jDYf, LLFn_p1jDYf, LLFnjDYf, LLFn_m1jDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rpj = MP5(LLFn_p3jtauf, LLFn_p2jtauf, LLFn_p1jtauf, LLFnjtauf, LLFn_m1jtauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rpj = MP5(LLFn_p3jphif, LLFn_p2jphif, LLFn_p1jphif, LLFnjphif, LLFn_m1jphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_x_rpk = MP5(LLFn_p3kSfd_x, LLFn_p2kSfd_x, LLFn_p1kSfd_x, LLFnkSfd_x, LLFn_m1kSfd_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_y_rpk = MP5(LLFn_p3kSfd_y, LLFn_p2kSfd_y, LLFn_p1kSfd_y, LLFnkSfd_y, LLFn_m1kSfd_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Sfd_z_rpk = MP5(LLFn_p3kSfd_z, LLFn_p2kSfd_z, LLFn_p1kSfd_z, LLFnkSfd_z, LLFn_m1kSfd_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_x_rpk = MP5(LLFn_p3kBfu_x, LLFn_p2kBfu_x, LLFn_p1kBfu_x, LLFnkBfu_x, LLFn_m1kBfu_x, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_y_rpk = MP5(LLFn_p3kBfu_y, LLFn_p2kBfu_y, LLFn_p1kBfu_y, LLFnkBfu_y, LLFn_m1kBfu_y, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Bfu_z_rpk = MP5(LLFn_p3kBfu_z, LLFn_p2kBfu_z, LLFn_p1kBfu_z, LLFnkBfu_z, LLFn_m1kBfu_z, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						Df_rpk = MP5(LLFn_p3kDf, LLFn_p2kDf, LLFn_p1kDf, LLFnkDf, LLFn_m1kDf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						DYf_rpk = MP5(LLFn_p3kDYf, LLFn_p2kDYf, LLFn_p1kDYf, LLFnkDYf, LLFn_m1kDYf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						tauf_rpk = MP5(LLFn_p3ktauf, LLFn_p2ktauf, LLFn_p1ktauf, LLFnktauf, LLFn_m1ktauf, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						phif_rpk = MP5(LLFn_p3kphif, LLFn_p2kphif, LLFn_p1kphif, LLFnkphif, LLFn_m1kphif, alpha_main, epsilon_main, dx, simPlat_dt, ilast, jlast);
						FluxniSfd_x = Sfd_x_lni + Sfd_x_rni;
						FluxniSfd_y = Sfd_y_lni + Sfd_y_rni;
						FluxniSfd_z = Sfd_z_lni + Sfd_z_rni;
						FluxniBfu_x = Bfu_x_lni + Bfu_x_rni;
						FluxniBfu_y = Bfu_y_lni + Bfu_y_rni;
						FluxniBfu_z = Bfu_z_lni + Bfu_z_rni;
						FluxniDf = Df_lni + Df_rni;
						FluxniDYf = DYf_lni + DYf_rni;
						Fluxnitauf = tauf_lni + tauf_rni;
						Fluxniphif = phif_lni + phif_rni;
						FluxnjSfd_x = Sfd_x_lnj + Sfd_x_rnj;
						FluxnjSfd_y = Sfd_y_lnj + Sfd_y_rnj;
						FluxnjSfd_z = Sfd_z_lnj + Sfd_z_rnj;
						FluxnjBfu_x = Bfu_x_lnj + Bfu_x_rnj;
						FluxnjBfu_y = Bfu_y_lnj + Bfu_y_rnj;
						FluxnjBfu_z = Bfu_z_lnj + Bfu_z_rnj;
						FluxnjDf = Df_lnj + Df_rnj;
						FluxnjDYf = DYf_lnj + DYf_rnj;
						Fluxnjtauf = tauf_lnj + tauf_rnj;
						Fluxnjphif = phif_lnj + phif_rnj;
						FluxnkSfd_x = Sfd_x_lnk + Sfd_x_rnk;
						FluxnkSfd_y = Sfd_y_lnk + Sfd_y_rnk;
						FluxnkSfd_z = Sfd_z_lnk + Sfd_z_rnk;
						FluxnkBfu_x = Bfu_x_lnk + Bfu_x_rnk;
						FluxnkBfu_y = Bfu_y_lnk + Bfu_y_rnk;
						FluxnkBfu_z = Bfu_z_lnk + Bfu_z_rnk;
						FluxnkDf = Df_lnk + Df_rnk;
						FluxnkDYf = DYf_lnk + DYf_rnk;
						Fluxnktauf = tauf_lnk + tauf_rnk;
						Fluxnkphif = phif_lnk + phif_rnk;
						FluxpiSfd_x = Sfd_x_lpi + Sfd_x_rpi;
						FluxpiSfd_y = Sfd_y_lpi + Sfd_y_rpi;
						FluxpiSfd_z = Sfd_z_lpi + Sfd_z_rpi;
						FluxpiBfu_x = Bfu_x_lpi + Bfu_x_rpi;
						FluxpiBfu_y = Bfu_y_lpi + Bfu_y_rpi;
						FluxpiBfu_z = Bfu_z_lpi + Bfu_z_rpi;
						FluxpiDf = Df_lpi + Df_rpi;
						FluxpiDYf = DYf_lpi + DYf_rpi;
						Fluxpitauf = tauf_lpi + tauf_rpi;
						Fluxpiphif = phif_lpi + phif_rpi;
						FluxpjSfd_x = Sfd_x_lpj + Sfd_x_rpj;
						FluxpjSfd_y = Sfd_y_lpj + Sfd_y_rpj;
						FluxpjSfd_z = Sfd_z_lpj + Sfd_z_rpj;
						FluxpjBfu_x = Bfu_x_lpj + Bfu_x_rpj;
						FluxpjBfu_y = Bfu_y_lpj + Bfu_y_rpj;
						FluxpjBfu_z = Bfu_z_lpj + Bfu_z_rpj;
						FluxpjDf = Df_lpj + Df_rpj;
						FluxpjDYf = DYf_lpj + DYf_rpj;
						Fluxpjtauf = tauf_lpj + tauf_rpj;
						Fluxpjphif = phif_lpj + phif_rpj;
						FluxpkSfd_x = Sfd_x_lpk + Sfd_x_rpk;
						FluxpkSfd_y = Sfd_y_lpk + Sfd_y_rpk;
						FluxpkSfd_z = Sfd_z_lpk + Sfd_z_rpk;
						FluxpkBfu_x = Bfu_x_lpk + Bfu_x_rpk;
						FluxpkBfu_y = Bfu_y_lpk + Bfu_y_rpk;
						FluxpkBfu_z = Bfu_z_lpk + Bfu_z_rpk;
						FluxpkDf = Df_lpk + Df_rpk;
						FluxpkDYf = DYf_lpk + DYf_rpk;
						Fluxpktauf = tauf_lpk + tauf_rpk;
						Fluxpkphif = phif_lpk + phif_rpk;
						d_div_Beta_o0_t0_m0_l0 = D1CDO4_i(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o1_t0_m0_l0 = lieforward_i(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o2_t0_m0_l0 = liebackward_i(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_y_o0_t7_m0_l0 = D1CDO4_j(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o1_t1_m0_l0 = lieforward_j(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o2_t1_m0_l0 = liebackward_j(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_z_o0_t7_m0_l0 = D1CDO4_k(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o1_t2_m0_l0 = lieforward_k(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_x_o2_t2_m0_l0 = liebackward_k(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_x_o0_t0_m0_l0 = D2CDO4_i(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_y_o0_t0_m0_l0 = D1CDO4crossed_ij(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_z_o0_t0_m0_l0 = D1CDO4crossed_ik(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t15_m0_l0 = D2CDO4_j(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t16_m0_l0 = D1CDO4crossed_jk(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t17_m0_l0 = D2CDO4_k(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_x_o0_t8_m0_l0 = D1CDO4_i(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o1_t0_m0_l0 = lieforward_i(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o2_t0_m0_l0 = liebackward_i(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_div_Beta_o0_t1_m0_l0 = D1CDO4_j(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o1_t1_m0_l0 = lieforward_j(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o2_t1_m0_l0 = liebackward_j(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_z_o0_t8_m0_l0 = D1CDO4_k(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o1_t2_m0_l0 = lieforward_k(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_y_o2_t2_m0_l0 = liebackward_k(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o0_t6_m0_l0 = D2CDO4_i(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_x_o0_t1_m0_l0 = D1CDO4crossed_ij(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o0_t8_m0_l0 = D1CDO4crossed_ik(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_y_o0_t1_m0_l0 = D2CDO4_j(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_z_o0_t1_m0_l0 = D1CDO4crossed_jk(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o0_t11_m0_l0 = D2CDO4_k(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_x_o0_t9_m0_l0 = D1CDO4_i(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o1_t0_m0_l0 = lieforward_i(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o2_t0_m0_l0 = liebackward_i(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Sfd_y_o0_t9_m0_l0 = D1CDO4_j(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o1_t1_m0_l0 = lieforward_j(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o2_t1_m0_l0 = liebackward_j(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_div_Beta_o0_t2_m0_l0 = D1CDO4_k(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o1_t2_m0_l0 = lieforward_k(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Betau_z_o2_t2_m0_l0 = liebackward_k(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o0_t0_m0_l0 = D2CDO4_i(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o0_t1_m0_l0 = D1CDO4crossed_ij(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_x_o0_t2_m0_l0 = D1CDO4crossed_ik(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o0_t3_m0_l0 = D2CDO4_j(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_y_o0_t2_m0_l0 = D1CDO4crossed_jk(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_d_div_Beta_z_o0_t2_m0_l0 = D2CDO4_k(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xxx_o0_t0_m0_l0 = D1CDO4_i(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o1_t0_m0_l0 = lieforward_i(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o2_t0_m0_l0 = liebackward_i(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xxy_o0_t0_m0_l0 = D1CDO4_j(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o1_t1_m0_l0 = lieforward_j(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o2_t1_m0_l0 = liebackward_j(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xxz_o0_t0_m0_l0 = D1CDO4_k(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o1_t2_m0_l0 = lieforward_k(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xx_o2_t2_m0_l0 = liebackward_k(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yxx_o0_t0_m0_l0 = D1CDO4_i(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o1_t0_m0_l0 = lieforward_i(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o2_t0_m0_l0 = liebackward_i(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyy_o0_t0_m0_l0 = D1CDO4_j(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o1_t1_m0_l0 = lieforward_j(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o2_t1_m0_l0 = liebackward_j(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyz_o0_t1_m0_l0 = D1CDO4_k(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o1_t2_m0_l0 = lieforward_k(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xy_o2_t2_m0_l0 = liebackward_k(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyy_o0_t1_m0_l0 = D1CDO4_i(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o1_t0_m0_l0 = lieforward_i(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o2_t0_m0_l0 = liebackward_i(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yyy_o0_t0_m0_l0 = D1CDO4_j(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o1_t1_m0_l0 = lieforward_j(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o2_t1_m0_l0 = liebackward_j(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yyz_o0_t0_m0_l0 = D1CDO4_k(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o1_t2_m0_l0 = lieforward_k(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yy_o2_t2_m0_l0 = liebackward_k(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_zxx_o0_t0_m0_l0 = D1CDO4_i(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o1_t0_m0_l0 = lieforward_i(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o2_t0_m0_l0 = liebackward_i(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyz_o0_t0_m0_l0 = D1CDO4_j(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o1_t1_m0_l0 = lieforward_j(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o2_t1_m0_l0 = liebackward_j(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xzz_o0_t0_m0_l0 = D1CDO4_k(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o1_t2_m0_l0 = lieforward_k(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_xz_o2_t2_m0_l0 = liebackward_k(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xyz_o0_t2_m0_l0 = D1CDO4_i(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o1_t0_m0_l0 = lieforward_i(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o2_t0_m0_l0 = liebackward_i(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_zyy_o0_t0_m0_l0 = D1CDO4_j(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o1_t1_m0_l0 = lieforward_j(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o2_t1_m0_l0 = liebackward_j(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yzz_o0_t0_m0_l0 = D1CDO4_k(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o1_t2_m0_l0 = lieforward_k(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_yz_o2_t2_m0_l0 = liebackward_k(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_xzz_o0_t1_m0_l0 = D1CDO4_i(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o1_t0_m0_l0 = lieforward_i(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o2_t0_m0_l0 = liebackward_i(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_yzz_o0_t1_m0_l0 = D1CDO4_j(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o1_t1_m0_l0 = lieforward_j(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o2_t1_m0_l0 = liebackward_j(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Ctd_zzz_o0_t0_m0_l0 = D1CDO4_k(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o1_t2_m0_l0 = lieforward_k(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_gtd_zz_o2_t2_m0_l0 = liebackward_k(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t0_m0_l0 = D1CDO4_i(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o1_t0_m0_l0 = lieforward_i(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o2_t0_m0_l0 = liebackward_i(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t1_m0_l0 = D1CDO4_j(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o1_t1_m0_l0 = lieforward_j(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o2_t1_m0_l0 = liebackward_j(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t2_m0_l0 = D1CDO4_k(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o1_t2_m0_l0 = lieforward_k(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chi_o2_t2_m0_l0 = liebackward_k(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t6_m0_l0 = D1CDO4_i(rk3Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o1_t0_m0_l0 = liebackward_i(rk3Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o2_t0_m0_l0 = lieforward_i(rk3Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t6_m0_l0 = D1CDO4_j(rk3Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o1_t1_m0_l0 = liebackward_j(rk3Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o2_t1_m0_l0 = lieforward_j(rk3Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t6_m0_l0 = D1CDO4_k(rk3Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o1_t2_m0_l0 = liebackward_k(rk3Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o2_t2_m0_l0 = lieforward_k(rk3Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t7_m0_l0 = D1CDO4_i(rk3Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o1_t0_m0_l0 = liebackward_i(rk3Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o2_t0_m0_l0 = lieforward_i(rk3Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t7_m0_l0 = D1CDO4_j(rk3Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o1_t1_m0_l0 = liebackward_j(rk3Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o2_t1_m0_l0 = lieforward_j(rk3Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t7_m0_l0 = D1CDO4_k(rk3Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o1_t2_m0_l0 = liebackward_k(rk3Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_y_o2_t2_m0_l0 = lieforward_k(rk3Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t8_m0_l0 = D1CDO4_i(rk3Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o1_t0_m0_l0 = liebackward_i(rk3Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o2_t0_m0_l0 = lieforward_i(rk3Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t8_m0_l0 = D1CDO4_j(rk3Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o1_t1_m0_l0 = liebackward_j(rk3Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o2_t1_m0_l0 = lieforward_j(rk3Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t8_m0_l0 = D1CDO4_k(rk3Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o1_t2_m0_l0 = liebackward_k(rk3Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_z_o2_t2_m0_l0 = lieforward_k(rk3Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t1_m0_l0 = D1CDO4_i(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o1_t0_m0_l0 = lieforward_i(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o2_t0_m0_l0 = liebackward_i(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t2_m0_l0 = D1CDO4_j(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o1_t1_m0_l0 = lieforward_j(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o2_t1_m0_l0 = liebackward_j(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t3_m0_l0 = D1CDO4_k(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o1_t2_m0_l0 = lieforward_k(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Alpha_o2_t2_m0_l0 = liebackward_k(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xx_o0_t0_m0_l0 = D2CDO4_i(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xy_o0_t0_m0_l0 = D1CDO4crossed_ij(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_xz_o0_t0_m0_l0 = D1CDO4crossed_ik(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_yy_o0_t0_m0_l0 = D2CDO4_j(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_yz_o0_t0_m0_l0 = D1CDO4crossed_jk(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Psi1_zz_o0_t0_m0_l0 = D2CDO4_k(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o1_t0_m0_l0 = lieforward_i(rk3Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o2_t0_m0_l0 = liebackward_i(rk3Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o1_t1_m0_l0 = lieforward_j(rk3Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o2_t1_m0_l0 = liebackward_j(rk3Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o1_t2_m0_l0 = lieforward_k(rk3Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xx_o2_t2_m0_l0 = liebackward_k(rk3Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o1_t0_m0_l0 = lieforward_i(rk3Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o2_t0_m0_l0 = liebackward_i(rk3Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o1_t1_m0_l0 = lieforward_j(rk3Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o2_t1_m0_l0 = liebackward_j(rk3Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o1_t2_m0_l0 = lieforward_k(rk3Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xy_o2_t2_m0_l0 = liebackward_k(rk3Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o1_t0_m0_l0 = lieforward_i(rk3Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o2_t0_m0_l0 = liebackward_i(rk3Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o1_t1_m0_l0 = lieforward_j(rk3Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o2_t1_m0_l0 = liebackward_j(rk3Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o1_t2_m0_l0 = lieforward_k(rk3Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_xz_o2_t2_m0_l0 = liebackward_k(rk3Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o1_t0_m0_l0 = lieforward_i(rk3Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o2_t0_m0_l0 = liebackward_i(rk3Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o1_t1_m0_l0 = lieforward_j(rk3Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o2_t1_m0_l0 = liebackward_j(rk3Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o1_t2_m0_l0 = lieforward_k(rk3Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yy_o2_t2_m0_l0 = liebackward_k(rk3Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o1_t0_m0_l0 = lieforward_i(rk3Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o2_t0_m0_l0 = liebackward_i(rk3Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o1_t1_m0_l0 = lieforward_j(rk3Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o2_t1_m0_l0 = liebackward_j(rk3Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o1_t2_m0_l0 = lieforward_k(rk3Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_yz_o2_t2_m0_l0 = liebackward_k(rk3Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o1_t0_m0_l0 = lieforward_i(rk3Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o2_t0_m0_l0 = liebackward_i(rk3Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o1_t1_m0_l0 = lieforward_j(rk3Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o2_t1_m0_l0 = liebackward_j(rk3Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o1_t2_m0_l0 = lieforward_k(rk3Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Atd_zz_o2_t2_m0_l0 = liebackward_k(rk3Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t0_m0_l0 = D1CDO4_i(rk3trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o1_t0_m0_l0 = liebackward_i(rk3trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o2_t0_m0_l0 = lieforward_i(rk3trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t1_m0_l0 = D1CDO4_j(rk3trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o1_t1_m0_l0 = liebackward_j(rk3trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o2_t1_m0_l0 = lieforward_j(rk3trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t2_m0_l0 = D1CDO4_k(rk3trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o1_t2_m0_l0 = liebackward_k(rk3trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_trK_o2_t2_m0_l0 = lieforward_k(rk3trK, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t3_m0_l0 = D1CDO4_i(rk3theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o1_t0_m0_l0 = lieforward_i(rk3theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o2_t0_m0_l0 = liebackward_i(rk3theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t4_m0_l0 = D1CDO4_j(rk3theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o1_t1_m0_l0 = lieforward_j(rk3theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o2_t1_m0_l0 = liebackward_j(rk3theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Gamh_x_o0_t5_m0_l0 = D1CDO4_k(rk3theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o1_t2_m0_l0 = lieforward_k(rk3theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_theta_o2_t2_m0_l0 = liebackward_k(rk3theta, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t0_m0_l0 = D2CDO4_i(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t1_m0_l0 = D1CDO4crossed_ij(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t2_m0_l0 = D1CDO4crossed_ik(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t3_m0_l0 = D2CDO4_j(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t4_m0_l0 = D1CDO4crossed_jk(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xx_o0_t5_m0_l0 = D2CDO4_k(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t0_m0_l0 = D2CDO4_i(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t1_m0_l0 = D1CDO4crossed_ij(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t2_m0_l0 = D1CDO4crossed_ik(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t3_m0_l0 = D2CDO4_j(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t4_m0_l0 = D1CDO4crossed_jk(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xy_o0_t5_m0_l0 = D2CDO4_k(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t0_m0_l0 = D2CDO4_i(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t1_m0_l0 = D1CDO4crossed_ij(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t2_m0_l0 = D1CDO4crossed_ik(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t3_m0_l0 = D2CDO4_j(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t4_m0_l0 = D1CDO4crossed_jk(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yy_o0_t5_m0_l0 = D2CDO4_k(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t0_m0_l0 = D2CDO4_i(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t1_m0_l0 = D1CDO4crossed_ij(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t2_m0_l0 = D1CDO4crossed_ik(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t3_m0_l0 = D2CDO4_j(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t4_m0_l0 = D1CDO4crossed_jk(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_xz_o0_t5_m0_l0 = D2CDO4_k(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t0_m0_l0 = D2CDO4_i(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t1_m0_l0 = D1CDO4crossed_ij(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t2_m0_l0 = D1CDO4crossed_ik(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t3_m0_l0 = D2CDO4_j(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t4_m0_l0 = D1CDO4crossed_jk(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_yz_o0_t5_m0_l0 = D2CDO4_k(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t0_m0_l0 = D2CDO4_i(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t1_m0_l0 = D1CDO4crossed_ij(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t2_m0_l0 = D1CDO4crossed_ik(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t3_m0_l0 = D2CDO4_j(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t4_m0_l0 = D1CDO4crossed_jk(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rtd_zz_o0_t5_m0_l0 = D2CDO4_k(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t3_m0_l0 = D2CDO4_i(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t4_m0_l0 = D1CDO4crossed_ij(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t5_m0_l0 = D1CDO4crossed_ik(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t6_m0_l0 = D2CDO4_j(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t7_m0_l0 = D1CDO4crossed_jk(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_Rpd_xx_o0_t8_m0_l0 = D2CDO4_k(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast);
						rk3m_Ctd_yyz_o0_t0_l0 = 0.5 * d_Ctd_yyz_o0_t0_m0_l0;
						rk3Ctd_yyz = rk3m_Ctd_yyz_o0_t0_l0;
						rk3m_Ctd_yyy_o0_t0_l0 = 0.5 * d_Ctd_yyy_o0_t0_m0_l0;
						rk3Ctd_yyy = rk3m_Ctd_yyy_o0_t0_l0;
						rk3div_Beta = (d_div_Beta_o0_t0_m0_l0 + d_div_Beta_o0_t1_m0_l0) + d_div_Beta_o0_t2_m0_l0;
						rk3m_Ctd_xxz_o0_t0_l0 = 0.5 * d_Ctd_xxz_o0_t0_m0_l0;
						rk3Ctd_xxz = rk3m_Ctd_xxz_o0_t0_l0;
						rk3m_Ctd_xxx_o0_t0_l0 = 0.5 * d_Ctd_xxx_o0_t0_m0_l0;
						rk3Ctd_xxx = rk3m_Ctd_xxx_o0_t0_l0;
						rk3m_Ctd_yzz_o0_t1_l0 = (-0.5) * d_Ctd_yzz_o0_t1_m0_l0;
						rk3Ctd_yzz = d_Ctd_yzz_o0_t0_m0_l0 + rk3m_Ctd_yzz_o0_t1_l0;
						rk3m_Ctd_xxy_o0_t0_l0 = 0.5 * d_Ctd_xxy_o0_t0_m0_l0;
						rk3Ctd_xxy = rk3m_Ctd_xxy_o0_t0_l0;
						rk3kappa_z2 = p_kappa_z2;
						rk3m_Ctd_xyy_o0_t1_l0 = (-0.5) * d_Ctd_xyy_o0_t1_m0_l0;
						rk3Ctd_xyy = d_Ctd_xyy_o0_t0_m0_l0 + rk3m_Ctd_xyy_o0_t1_l0;
						rk3m_Ctd_xyz_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t2_m0_l0;
						rk3m_Ctd_xyz_o0_t1_l0 = 0.5 * d_Ctd_xyz_o0_t1_m0_l0;
						rk3m_Ctd_xyz_o0_t0_l0 = 0.5 * d_Ctd_xyz_o0_t0_m0_l0;
						rk3Ctd_xyz = (rk3m_Ctd_xyz_o0_t0_l0 + rk3m_Ctd_xyz_o0_t1_l0) + rk3m_Ctd_xyz_o0_t2_l0;
						rk3d_div_Beta_y = (d_d_div_Beta_y_o0_t0_m0_l0 + d_d_div_Beta_y_o0_t1_m0_l0) + d_d_div_Beta_y_o0_t2_m0_l0;
						rk3d_div_Beta_x = (d_d_div_Beta_x_o0_t0_m0_l0 + d_d_div_Beta_x_o0_t1_m0_l0) + d_d_div_Beta_x_o0_t2_m0_l0;
						rk3d_div_Beta_z = (d_d_div_Beta_z_o0_t0_m0_l0 + d_d_div_Beta_z_o0_t1_m0_l0) + d_d_div_Beta_z_o0_t2_m0_l0;
						rk3m_Ctd_xzz_o0_t1_l0 = (-0.5) * d_Ctd_xzz_o0_t1_m0_l0;
						rk3Ctd_xzz = d_Ctd_xzz_o0_t0_m0_l0 + rk3m_Ctd_xzz_o0_t1_l0;
						rk3m_Ctd_zxz_o0_t0_l0 = 0.5 * d_Ctd_xzz_o0_t1_m0_l0;
						rk3Ctd_zxz = rk3m_Ctd_zxz_o0_t0_l0;
						rk3m_Ctd_zxx_o0_t1_l0 = (-0.5) * d_Ctd_xxz_o0_t0_m0_l0;
						rk3Ctd_zxx = d_Ctd_zxx_o0_t0_m0_l0 + rk3m_Ctd_zxx_o0_t1_l0;
						rk3detgtd = vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) * vector(rk3gtd_xx, i, j, k) + (-vector(rk3gtd_yy, i, j, k) * (vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k))) + (-vector(rk3gtd_zz, i, j, k) * (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k))) + (-(vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k)) * vector(rk3gtd_xx, i, j, k)) + 2.0 * vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k);
						rk3chi_max = MAX(chi_floor, vector(rk3chi, i, j, k));
						rk3inv_chi = 1.0 / fabs(rk3chi_max);
						rk3Bfd_y = rk3inv_chi * (vector(rk3gtd_xy, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_yy, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3Bfd_z = rk3inv_chi * (vector(rk3gtd_xz, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_yz, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_zz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3Bfd_x = rk3inv_chi * (vector(rk3gtd_xx, i, j, k) * vector(rk3Bfu_x, i, j, k) + vector(rk3gtd_xy, i, j, k) * vector(rk3Bfu_y, i, j, k) + vector(rk3gtd_xz, i, j, k) * vector(rk3Bfu_z, i, j, k));
						rk3sqBf = vector(rk3Bfu_x, i, j, k) * rk3Bfd_x + vector(rk3Bfu_y, i, j, k) * rk3Bfd_y + vector(rk3Bfu_z, i, j, k) * rk3Bfd_z;
						rk3m_Ctd_zyy_o0_t1_l0 = (-0.5) * d_Ctd_yyz_o0_t0_m0_l0;
						rk3Ctd_zyy = d_Ctd_zyy_o0_t0_m0_l0 + rk3m_Ctd_zyy_o0_t1_l0;
						rk3m_Ctd_zyz_o0_t0_l0 = 0.5 * d_Ctd_yzz_o0_t1_m0_l0;
						rk3Ctd_zyz = rk3m_Ctd_zyz_o0_t0_l0;
						rk3sdetg = pow(rk3chi_max, (-1.500000000000000));
						rk3Tf4u_tt = (vector(rk3Df, i, j, k) + vector(rk3tauf, i, j, k)) / (rk3sdetg * (vector(rk3Alpha, i, j, k) * vector(rk3Alpha, i, j, k)));
						rk3Betatd_x = vector(rk3Betau_x, i, j, k) * vector(rk3gtd_xx, i, j, k) + vector(rk3Betau_y, i, j, k) * vector(rk3gtd_xy, i, j, k) + vector(rk3Betau_z, i, j, k) * vector(rk3gtd_xz, i, j, k);
						rk3Betatd_z = vector(rk3Betau_x, i, j, k) * vector(rk3gtd_xz, i, j, k) + vector(rk3Betau_y, i, j, k) * vector(rk3gtd_yz, i, j, k) + vector(rk3Betau_z, i, j, k) * vector(rk3gtd_zz, i, j, k);
						rk3Betatd_y = vector(rk3Betau_x, i, j, k) * vector(rk3gtd_xy, i, j, k) + vector(rk3Betau_y, i, j, k) * vector(rk3gtd_yy, i, j, k) + vector(rk3Betau_z, i, j, k) * vector(rk3gtd_yz, i, j, k);
						rk3idetgtd = 1.0 / rk3detgtd;
						rk3gtu_zz = rk3idetgtd * (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yy, i, j, k) - vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k));
						rk3gtu_yy = rk3idetgtd * (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_zz, i, j, k) - vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3gtu_yz = rk3idetgtd * ((-vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yz, i, j, k)) + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3gtu_xx = rk3idetgtd * (vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) - vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k));
						rk3gtu_xz = rk3idetgtd * (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_yz, i, j, k) - vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3Ct_zyz = rk3gtu_xz * rk3Ctd_xyz + rk3gtu_yz * rk3Ctd_yyz + rk3gtu_zz * rk3Ctd_zyz;
						rk3Ct_zyy = rk3gtu_xz * rk3Ctd_xyy + rk3gtu_yz * rk3Ctd_yyy + rk3gtu_zz * rk3Ctd_zyy;
						rk3Atud_zx = rk3gtu_xz * vector(rk3Atd_xx, i, j, k) + rk3gtu_yz * vector(rk3Atd_xy, i, j, k) + rk3gtu_zz * vector(rk3Atd_xz, i, j, k);
						rk3Atud_zz = rk3gtu_xz * vector(rk3Atd_xz, i, j, k) + rk3gtu_yz * vector(rk3Atd_yz, i, j, k) + rk3gtu_zz * vector(rk3Atd_zz, i, j, k);
						rk3Atud_zy = rk3gtu_xz * vector(rk3Atd_xy, i, j, k) + rk3gtu_yz * vector(rk3Atd_yy, i, j, k) + rk3gtu_zz * vector(rk3Atd_yz, i, j, k);
						rk3Atu_zz = rk3Atud_zx * rk3gtu_xz + rk3Atud_zy * rk3gtu_yz + rk3Atud_zz * rk3gtu_zz;
						rk3vfu_z = rk3chi_max * (rk3gtu_xz * vector(vfd_x, i, j, k) + rk3gtu_yz * vector(vfd_y, i, j, k) + rk3gtu_zz * vector(vfd_z, i, j, k));
						rk3Sfu_z = rk3chi_max * (rk3gtu_xz * vector(rk3Sfd_x, i, j, k) + rk3gtu_yz * vector(rk3Sfd_y, i, j, k) + rk3gtu_zz * vector(rk3Sfd_z, i, j, k));
						rk3Tf4u_tz = rk3Sfu_z / (rk3sdetg * vector(rk3Alpha, i, j, k)) + (-vector(rk3Betau_z, i, j, k) * rk3Tf4u_tt);
						rk3Tu_tz = rk3Tf4u_tz;
						rk3gtu_xy = rk3idetgtd * ((-vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_zz, i, j, k)) + vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xz, i, j, k));
						rk3vfu_y = rk3chi_max * (rk3gtu_xy * vector(vfd_x, i, j, k) + rk3gtu_yy * vector(vfd_y, i, j, k) + rk3gtu_yz * vector(vfd_z, i, j, k));
						rk3vfu_x = rk3chi_max * (rk3gtu_xx * vector(vfd_x, i, j, k) + rk3gtu_xy * vector(vfd_y, i, j, k) + rk3gtu_xz * vector(vfd_z, i, j, k));
						rk3sqvf = MIN(rk3vfu_x * vector(vfd_x, i, j, k) + rk3vfu_y * vector(vfd_y, i, j, k) + rk3vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						rk3sqW = 1.0 / (1.0 - rk3sqvf);
						rk3Atud_yy = rk3gtu_xy * vector(rk3Atd_xy, i, j, k) + rk3gtu_yy * vector(rk3Atd_yy, i, j, k) + rk3gtu_yz * vector(rk3Atd_yz, i, j, k);
						rk3Atud_yx = rk3gtu_xy * vector(rk3Atd_xx, i, j, k) + rk3gtu_yy * vector(rk3Atd_xy, i, j, k) + rk3gtu_yz * vector(rk3Atd_xz, i, j, k);
						rk3Atud_yz = rk3gtu_xy * vector(rk3Atd_xz, i, j, k) + rk3gtu_yy * vector(rk3Atd_yz, i, j, k) + rk3gtu_yz * vector(rk3Atd_zz, i, j, k);
						rk3Atu_yz = rk3Atud_yx * rk3gtu_xz + rk3Atud_yy * rk3gtu_yz + rk3Atud_yz * rk3gtu_zz;
						rk3Atu_yy = rk3Atud_yx * rk3gtu_xy + rk3Atud_yy * rk3gtu_yy + rk3Atud_yz * rk3gtu_yz;
						rk3Ct_yyy = rk3gtu_xy * rk3Ctd_xyy + rk3gtu_yy * rk3Ctd_yyy + rk3gtu_yz * rk3Ctd_zyy;
						rk3Ct_yyz = rk3gtu_xy * rk3Ctd_xyz + rk3gtu_yy * rk3Ctd_yyz + rk3gtu_yz * rk3Ctd_zyz;
						rk3Atud_xx = rk3gtu_xx * vector(rk3Atd_xx, i, j, k) + rk3gtu_xy * vector(rk3Atd_xy, i, j, k) + rk3gtu_xz * vector(rk3Atd_xz, i, j, k);
						rk3trAt = rk3Atud_xx + rk3Atud_yy + rk3Atud_zz;
						rk3Atud_xz = rk3gtu_xx * vector(rk3Atd_xz, i, j, k) + rk3gtu_xy * vector(rk3Atd_yz, i, j, k) + rk3gtu_xz * vector(rk3Atd_zz, i, j, k);
						rk3Atud_xy = rk3gtu_xx * vector(rk3Atd_xy, i, j, k) + rk3gtu_xy * vector(rk3Atd_yy, i, j, k) + rk3gtu_xz * vector(rk3Atd_yz, i, j, k);
						rk3Atu_xz = rk3Atud_xx * rk3gtu_xz + rk3Atud_xy * rk3gtu_yz + rk3Atud_xz * rk3gtu_zz;
						rk3Atu_xy = rk3Atud_xx * rk3gtu_xy + rk3Atud_xy * rk3gtu_yy + rk3Atud_xz * rk3gtu_yz;
						rk3Atu_xx = rk3Atud_xx * rk3gtu_xx + rk3Atud_xy * rk3gtu_xy + rk3Atud_xz * rk3gtu_xz;
						rk3Sfu_y = rk3chi_max * (rk3gtu_xy * vector(rk3Sfd_x, i, j, k) + rk3gtu_yy * vector(rk3Sfd_y, i, j, k) + rk3gtu_yz * vector(rk3Sfd_z, i, j, k));
						rk3Tf4u_ty = rk3Sfu_y / (rk3sdetg * vector(rk3Alpha, i, j, k)) + (-vector(rk3Betau_y, i, j, k) * rk3Tf4u_tt);
						rk3Tu_ty = rk3Tf4u_ty;
						rk3Sfu_x = rk3chi_max * (rk3gtu_xx * vector(rk3Sfd_x, i, j, k) + rk3gtu_xy * vector(rk3Sfd_y, i, j, k) + rk3gtu_xz * vector(rk3Sfd_z, i, j, k));
						rk3Tf4u_tx = rk3Sfu_x / (rk3sdetg * vector(rk3Alpha, i, j, k)) + (-vector(rk3Betau_x, i, j, k) * rk3Tf4u_tt);
						rk3Tu_tx = rk3Tf4u_tx;
						rk3decay_factor = MIN(1.0, pow((R_0 / MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))), eta_damping_exp));
						rk3kappa_z1 = MIN(p_kappa_z1, p_kappa_z1 * rk3decay_factor) / vector(rk3Alpha, i, j, k);
						rk3kappa_f = MIN(p_kappa_f, p_kappa_f * rk3decay_factor);
						rk3feta = MIN(p_feta, p_feta * rk3decay_factor);
						rk3m_Ctd_zzz_o0_t0_l0 = 0.5 * d_Ctd_zzz_o0_t0_m0_l0;
						rk3Ctd_zzz = rk3m_Ctd_zzz_o0_t0_l0;
						rk3Ct_xzz = rk3gtu_xx * rk3Ctd_xzz + rk3gtu_xy * rk3Ctd_yzz + rk3gtu_xz * rk3Ctd_zzz;
						rk3Ct_zzz = rk3gtu_xz * rk3Ctd_xzz + rk3gtu_yz * rk3Ctd_yzz + rk3gtu_zz * rk3Ctd_zzz;
						rk3Ct_yzz = rk3gtu_xy * rk3Ctd_xzz + rk3gtu_yy * rk3Ctd_yzz + rk3gtu_yz * rk3Ctd_zzz;
						rk3m_Ctd_yxy_o0_t0_l0 = 0.5 * d_Ctd_xyy_o0_t1_m0_l0;
						rk3Ctd_yxy = rk3m_Ctd_yxy_o0_t0_l0;
						rk3Ct_xyz = rk3gtu_xx * rk3Ctd_xyz + rk3gtu_xy * rk3Ctd_yyz + rk3gtu_xz * rk3Ctd_zyz;
						rk3m_Ctd_yxz_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t0_m0_l0;
						rk3m_Ctd_yxz_o0_t0_l0 = 0.5 * d_Ctd_xyz_o0_t2_m0_l0;
						rk3Ctd_yxz = (rk3m_Ctd_yxz_o0_t0_l0 + rk3m_Ctd_xyz_o0_t1_l0) + rk3m_Ctd_yxz_o0_t2_l0;
						rk3Ct_zxz = rk3gtu_xz * rk3Ctd_xxz + rk3gtu_yz * rk3Ctd_yxz + rk3gtu_zz * rk3Ctd_zxz;
						rk3Ct_yxz = rk3gtu_xy * rk3Ctd_xxz + rk3gtu_yy * rk3Ctd_yxz + rk3gtu_yz * rk3Ctd_zxz;
						rk3m_Ctd_zxy_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t1_m0_l0;
						rk3Ctd_zxy = (rk3m_Ctd_yxz_o0_t0_l0 + rk3m_Ctd_xyz_o0_t0_l0) + rk3m_Ctd_zxy_o0_t2_l0;
						rk3Ct_zxy = rk3gtu_xz * rk3Ctd_xxy + rk3gtu_yz * rk3Ctd_yxy + rk3gtu_zz * rk3Ctd_zxy;
						rk3Ct_yxy = rk3gtu_xy * rk3Ctd_xxy + rk3gtu_yy * rk3Ctd_yxy + rk3gtu_yz * rk3Ctd_zxy;
						rk3Ct_xxy = rk3gtu_xx * rk3Ctd_xxy + rk3gtu_xy * rk3Ctd_yxy + rk3gtu_xz * rk3Ctd_zxy;
						rk3Ct_xxz = rk3gtu_xx * rk3Ctd_xxz + rk3gtu_xy * rk3Ctd_yxz + rk3gtu_xz * rk3Ctd_zxz;
						rk3Ct_xyy = rk3gtu_xx * rk3Ctd_xyy + rk3gtu_xy * rk3Ctd_yyy + rk3gtu_xz * rk3Ctd_zyy;
						rk3m_Ctd_yxx_o0_t1_l0 = (-0.5) * d_Ctd_xxy_o0_t0_m0_l0;
						rk3Ctd_yxx = d_Ctd_yxx_o0_t0_m0_l0 + rk3m_Ctd_yxx_o0_t1_l0;
						rk3Ct_zxx = rk3gtu_xz * rk3Ctd_xxx + rk3gtu_yz * rk3Ctd_yxx + rk3gtu_zz * rk3Ctd_zxx;
						rk3Gamt_z = rk3gtu_xx * rk3Ct_zxx + 2.0 * rk3gtu_xy * rk3Ct_zxy + 2.0 * rk3gtu_xz * rk3Ct_zxz + rk3gtu_yy * rk3Ct_zyy + 2.0 * rk3gtu_yz * rk3Ct_zyz + rk3gtu_zz * rk3Ct_zzz;
						rk3Zu_z = 0.5 * rk3chi_max * (vector(rk3Gamh_z, i, j, k) - rk3Gamt_z);
						rk3Ct_yxx = rk3gtu_xy * rk3Ctd_xxx + rk3gtu_yy * rk3Ctd_yxx + rk3gtu_yz * rk3Ctd_zxx;
						rk3Gamt_y = rk3gtu_xx * rk3Ct_yxx + 2.0 * rk3gtu_xy * rk3Ct_yxy + 2.0 * rk3gtu_xz * rk3Ct_yxz + rk3gtu_yy * rk3Ct_yyy + 2.0 * rk3gtu_yz * rk3Ct_yyz + rk3gtu_zz * rk3Ct_yzz;
						rk3Zu_y = 0.5 * rk3chi_max * (vector(rk3Gamh_y, i, j, k) - rk3Gamt_y);
						rk3Ct_xxx = rk3gtu_xx * rk3Ctd_xxx + rk3gtu_xy * rk3Ctd_yxx + rk3gtu_xz * rk3Ctd_zxx;
						rk3Gamt_x = rk3gtu_xx * rk3Ct_xxx + 2.0 * rk3gtu_xy * rk3Ct_xxy + 2.0 * rk3gtu_xz * rk3Ct_xxz + rk3gtu_yy * rk3Ct_xyy + 2.0 * rk3gtu_yz * rk3Ct_xyz + rk3gtu_zz * rk3Ct_xzz;
						rk3Zu_x = 0.5 * rk3chi_max * (vector(rk3Gamh_x, i, j, k) - rk3Gamt_x);
						rk3d_Rtd_xx_o0_t9_m0_l0 = rk3gtu_zz * (3.0 * rk3Ct_xxz * rk3Ctd_xxz + 2.0 * rk3Ct_yxz * rk3Ctd_xyz + rk3Ct_yxz * rk3Ctd_yxz + 2.0 * rk3Ct_zxz * rk3Ctd_xzz + rk3Ct_zxz * rk3Ctd_zxz) + rk3gtu_yz * (3.0 * rk3Ct_xxy * rk3Ctd_xxz + 2.0 * rk3Ct_yxy * rk3Ctd_xyz + rk3Ct_yxy * rk3Ctd_yxz + 2.0 * rk3Ct_zxy * rk3Ctd_xzz + rk3Ct_zxy * rk3Ctd_zxz) + rk3gtu_xz * (3.0 * rk3Ct_xxx * rk3Ctd_xxz + 2.0 * rk3Ct_yxx * rk3Ctd_xyz + rk3Ct_yxx * rk3Ctd_yxz + 2.0 * rk3Ct_zxx * rk3Ctd_xzz + rk3Ct_zxx * rk3Ctd_zxz) + rk3gtu_yz * (3.0 * rk3Ct_xxz * rk3Ctd_xxy + 2.0 * rk3Ct_yxz * rk3Ctd_xyy + rk3Ct_yxz * rk3Ctd_yxy + 2.0 * rk3Ct_zxz * rk3Ctd_xyz + rk3Ct_zxz * rk3Ctd_zxy) + rk3gtu_yy * (3.0 * rk3Ct_xxy * rk3Ctd_xxy + 2.0 * rk3Ct_yxy * rk3Ctd_xyy + rk3Ct_yxy * rk3Ctd_yxy + 2.0 * rk3Ct_zxy * rk3Ctd_xyz + rk3Ct_zxy * rk3Ctd_zxy) + rk3gtu_xy * (3.0 * rk3Ct_xxx * rk3Ctd_xxy + 2.0 * rk3Ct_yxx * rk3Ctd_xyy + rk3Ct_yxx * rk3Ctd_yxy + 2.0 * rk3Ct_zxx * rk3Ctd_xyz + rk3Ct_zxx * rk3Ctd_zxy) + rk3gtu_xz * (3.0 * rk3Ct_xxz * rk3Ctd_xxx + 2.0 * rk3Ct_yxz * rk3Ctd_xxy + rk3Ct_yxz * rk3Ctd_yxx + 2.0 * rk3Ct_zxz * rk3Ctd_xxz + rk3Ct_zxz * rk3Ctd_zxx) + rk3gtu_xy * (3.0 * rk3Ct_xxy * rk3Ctd_xxx + 2.0 * rk3Ct_yxy * rk3Ctd_xxy + rk3Ct_yxy * rk3Ctd_yxx + 2.0 * rk3Ct_zxy * rk3Ctd_xxz + rk3Ct_zxy * rk3Ctd_zxx) + rk3gtu_xx * (3.0 * rk3Ct_xxx * rk3Ctd_xxx + 2.0 * rk3Ct_yxx * rk3Ctd_xxy + rk3Ct_yxx * rk3Ctd_yxx + 2.0 * rk3Ct_zxx * rk3Ctd_xxz + rk3Ct_zxx * rk3Ctd_zxx) + vector(rk3Gamh_z, i, j, k) * rk3Ctd_xxz + vector(rk3Gamh_y, i, j, k) * rk3Ctd_xxy + vector(rk3Gamh_x, i, j, k) * rk3Ctd_xxx;
						rk3m_Rtd_xx_o0_t8_l0 = vector(rk3gtd_xz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
						rk3m_Rtd_xx_o0_t7_l0 = vector(rk3gtd_xy, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
						rk3m_Rtd_xx_o0_t6_l0 = vector(rk3gtd_xx, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
						rk3m_Rtd_xx_o0_t5_l0 = (-0.5 * rk3gtu_zz) * d_Rtd_xx_o0_t5_m0_l0;
						rk3m_Rtd_xx_o0_t4_l0 = (-rk3gtu_yz) * d_Rtd_xx_o0_t4_m0_l0;
						rk3m_Rtd_xx_o0_t3_l0 = (-0.5 * rk3gtu_yy) * d_Rtd_xx_o0_t3_m0_l0;
						rk3m_Rtd_xx_o0_t2_l0 = (-rk3gtu_xz) * d_Rtd_xx_o0_t2_m0_l0;
						rk3m_Rtd_xx_o0_t1_l0 = (-rk3gtu_xy) * d_Rtd_xx_o0_t1_m0_l0;
						rk3m_Rtd_xx_o0_t0_l0 = (-0.5 * rk3gtu_xx) * d_Rtd_xx_o0_t0_m0_l0;
						rk3Rtd_xx = ((((((((rk3m_Rtd_xx_o0_t0_l0 + rk3m_Rtd_xx_o0_t1_l0) + rk3m_Rtd_xx_o0_t2_l0) + rk3m_Rtd_xx_o0_t3_l0) + rk3m_Rtd_xx_o0_t4_l0) + rk3m_Rtd_xx_o0_t5_l0) + rk3m_Rtd_xx_o0_t6_l0) + rk3m_Rtd_xx_o0_t7_l0) + rk3m_Rtd_xx_o0_t8_l0) + rk3d_Rtd_xx_o0_t9_m0_l0;
						rk3d_Rtd_xz_o0_t12_m0_l0 = rk3gtu_zz * (rk3Ct_xxz * rk3Ctd_zxz + rk3Ct_xxz * rk3Ctd_xzz + rk3Ct_xzz * rk3Ctd_xxz + rk3Ct_yxz * rk3Ctd_zyz + rk3Ct_yxz * rk3Ctd_yzz + rk3Ct_yzz * rk3Ctd_xyz + 2.0 * rk3Ct_zxz * rk3Ctd_zzz + rk3Ct_zzz * rk3Ctd_xzz) + rk3gtu_yz * (rk3Ct_xxy * rk3Ctd_zxz + rk3Ct_xxy * rk3Ctd_xzz + rk3Ct_xyz * rk3Ctd_xxz + rk3Ct_yxy * rk3Ctd_zyz + rk3Ct_yxy * rk3Ctd_yzz + rk3Ct_yyz * rk3Ctd_xyz + 2.0 * rk3Ct_zxy * rk3Ctd_zzz + rk3Ct_zyz * rk3Ctd_xzz) + rk3gtu_xz * (rk3Ct_xxx * rk3Ctd_zxz + rk3Ct_xxx * rk3Ctd_xzz + rk3Ct_xxz * rk3Ctd_xxz + rk3Ct_yxx * rk3Ctd_zyz + rk3Ct_yxx * rk3Ctd_yzz + rk3Ct_yxz * rk3Ctd_xyz + 2.0 * rk3Ct_zxx * rk3Ctd_zzz + rk3Ct_zxz * rk3Ctd_xzz) + rk3gtu_yz * (rk3Ct_xxz * rk3Ctd_zxy + rk3Ct_xxz * rk3Ctd_xyz + rk3Ct_xzz * rk3Ctd_xxy + rk3Ct_yxz * rk3Ctd_zyy + rk3Ct_yxz * rk3Ctd_yyz + rk3Ct_yzz * rk3Ctd_xyy + 2.0 * rk3Ct_zxz * rk3Ctd_zyz + rk3Ct_zzz * rk3Ctd_xyz) + rk3gtu_yy * (rk3Ct_xxy * rk3Ctd_zxy + rk3Ct_xxy * rk3Ctd_xyz + rk3Ct_xyz * rk3Ctd_xxy + rk3Ct_yxy * rk3Ctd_zyy + rk3Ct_yxy * rk3Ctd_yyz + rk3Ct_yyz * rk3Ctd_xyy + 2.0 * rk3Ct_zxy * rk3Ctd_zyz + rk3Ct_zyz * rk3Ctd_xyz) + rk3gtu_xy * (rk3Ct_xxx * rk3Ctd_zxy + rk3Ct_xxx * rk3Ctd_xyz + rk3Ct_xxz * rk3Ctd_xxy + rk3Ct_yxx * rk3Ctd_zyy + rk3Ct_yxx * rk3Ctd_yyz + rk3Ct_yxz * rk3Ctd_xyy + 2.0 * rk3Ct_zxx * rk3Ctd_zyz + rk3Ct_zxz * rk3Ctd_xyz) + rk3gtu_xz * (rk3Ct_xxz * rk3Ctd_zxx + rk3Ct_xxz * rk3Ctd_xxz + rk3Ct_xzz * rk3Ctd_xxx + rk3Ct_yxz * rk3Ctd_zxy + rk3Ct_yxz * rk3Ctd_yxz + rk3Ct_yzz * rk3Ctd_xxy + 2.0 * rk3Ct_zxz * rk3Ctd_zxz + rk3Ct_zzz * rk3Ctd_xxz) + rk3gtu_xy * (rk3Ct_xxy * rk3Ctd_zxx + rk3Ct_xxy * rk3Ctd_xxz + rk3Ct_xyz * rk3Ctd_xxx + rk3Ct_yxy * rk3Ctd_zxy + rk3Ct_yxy * rk3Ctd_yxz + rk3Ct_yyz * rk3Ctd_xxy + 2.0 * rk3Ct_zxy * rk3Ctd_zxz + rk3Ct_zyz * rk3Ctd_xxz) + rk3gtu_xx * (rk3Ct_xxx * rk3Ctd_zxx + rk3Ct_xxx * rk3Ctd_xxz + rk3Ct_xxz * rk3Ctd_xxx + rk3Ct_yxx * rk3Ctd_zxy + rk3Ct_yxx * rk3Ctd_yxz + rk3Ct_yxz * rk3Ctd_xxy + 2.0 * rk3Ct_zxx * rk3Ctd_zxz + rk3Ct_zxz * rk3Ctd_xxz) + 0.5 * vector(rk3Gamh_z, i, j, k) * (rk3Ctd_xzz + rk3Ctd_zxz) + 0.5 * vector(rk3Gamh_y, i, j, k) * (rk3Ctd_xyz + rk3Ctd_zxy) + 0.5 * vector(rk3Gamh_x, i, j, k) * (rk3Ctd_xxz + rk3Ctd_zxx);
						rk3m_Rtd_xz_o0_t11_l0 = 0.5 * vector(rk3gtd_zz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
						rk3m_Rtd_xz_o0_t10_l0 = 0.5 * vector(rk3gtd_yz, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
						rk3m_Rtd_xz_o0_t9_l0 = 0.5 * vector(rk3gtd_xz, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
						rk3m_Rtd_xz_o0_t8_l0 = 0.5 * vector(rk3gtd_xz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
						rk3m_Rtd_xz_o0_t7_l0 = 0.5 * vector(rk3gtd_xy, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
						rk3m_Rtd_xz_o0_t6_l0 = 0.5 * vector(rk3gtd_xx, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
						rk3m_Rtd_xz_o0_t5_l0 = (-0.5 * rk3gtu_zz) * d_Rtd_xz_o0_t5_m0_l0;
						rk3m_Rtd_xz_o0_t4_l0 = (-rk3gtu_yz) * d_Rtd_xz_o0_t4_m0_l0;
						rk3m_Rtd_xz_o0_t3_l0 = (-0.5 * rk3gtu_yy) * d_Rtd_xz_o0_t3_m0_l0;
						rk3m_Rtd_xz_o0_t2_l0 = (-rk3gtu_xz) * d_Rtd_xz_o0_t2_m0_l0;
						rk3m_Rtd_xz_o0_t1_l0 = (-rk3gtu_xy) * d_Rtd_xz_o0_t1_m0_l0;
						rk3m_Rtd_xz_o0_t0_l0 = (-0.5 * rk3gtu_xx) * d_Rtd_xz_o0_t0_m0_l0;
						rk3Rtd_xz = (((((((((((rk3m_Rtd_xz_o0_t0_l0 + rk3m_Rtd_xz_o0_t1_l0) + rk3m_Rtd_xz_o0_t2_l0) + rk3m_Rtd_xz_o0_t3_l0) + rk3m_Rtd_xz_o0_t4_l0) + rk3m_Rtd_xz_o0_t5_l0) + rk3m_Rtd_xz_o0_t6_l0) + rk3m_Rtd_xz_o0_t7_l0) + rk3m_Rtd_xz_o0_t8_l0) + rk3m_Rtd_xz_o0_t9_l0) + rk3m_Rtd_xz_o0_t10_l0) + rk3m_Rtd_xz_o0_t11_l0) + rk3d_Rtd_xz_o0_t12_m0_l0;
						rk3d_Rtd_zz_o0_t9_m0_l0 = rk3gtu_zz * (2.0 * rk3Ct_xzz * rk3Ctd_zxz + rk3Ct_xzz * rk3Ctd_xzz + 2.0 * rk3Ct_yzz * rk3Ctd_zyz + rk3Ct_yzz * rk3Ctd_yzz + 3.0 * rk3Ct_zzz * rk3Ctd_zzz) + rk3gtu_yz * (2.0 * rk3Ct_xyz * rk3Ctd_zxz + rk3Ct_xyz * rk3Ctd_xzz + 2.0 * rk3Ct_yyz * rk3Ctd_zyz + rk3Ct_yyz * rk3Ctd_yzz + 3.0 * rk3Ct_zyz * rk3Ctd_zzz) + rk3gtu_xz * (2.0 * rk3Ct_xxz * rk3Ctd_zxz + rk3Ct_xxz * rk3Ctd_xzz + 2.0 * rk3Ct_yxz * rk3Ctd_zyz + rk3Ct_yxz * rk3Ctd_yzz + 3.0 * rk3Ct_zxz * rk3Ctd_zzz) + rk3gtu_yz * (2.0 * rk3Ct_xzz * rk3Ctd_zxy + rk3Ct_xzz * rk3Ctd_xyz + 2.0 * rk3Ct_yzz * rk3Ctd_zyy + rk3Ct_yzz * rk3Ctd_yyz + 3.0 * rk3Ct_zzz * rk3Ctd_zyz) + rk3gtu_yy * (2.0 * rk3Ct_xyz * rk3Ctd_zxy + rk3Ct_xyz * rk3Ctd_xyz + 2.0 * rk3Ct_yyz * rk3Ctd_zyy + rk3Ct_yyz * rk3Ctd_yyz + 3.0 * rk3Ct_zyz * rk3Ctd_zyz) + rk3gtu_xy * (2.0 * rk3Ct_xxz * rk3Ctd_zxy + rk3Ct_xxz * rk3Ctd_xyz + 2.0 * rk3Ct_yxz * rk3Ctd_zyy + rk3Ct_yxz * rk3Ctd_yyz + 3.0 * rk3Ct_zxz * rk3Ctd_zyz) + rk3gtu_xz * (2.0 * rk3Ct_xzz * rk3Ctd_zxx + rk3Ct_xzz * rk3Ctd_xxz + 2.0 * rk3Ct_yzz * rk3Ctd_zxy + rk3Ct_yzz * rk3Ctd_yxz + 3.0 * rk3Ct_zzz * rk3Ctd_zxz) + rk3gtu_xy * (2.0 * rk3Ct_xyz * rk3Ctd_zxx + rk3Ct_xyz * rk3Ctd_xxz + 2.0 * rk3Ct_yyz * rk3Ctd_zxy + rk3Ct_yyz * rk3Ctd_yxz + 3.0 * rk3Ct_zyz * rk3Ctd_zxz) + rk3gtu_xx * (2.0 * rk3Ct_xxz * rk3Ctd_zxx + rk3Ct_xxz * rk3Ctd_xxz + 2.0 * rk3Ct_yxz * rk3Ctd_zxy + rk3Ct_yxz * rk3Ctd_yxz + 3.0 * rk3Ct_zxz * rk3Ctd_zxz) + vector(rk3Gamh_z, i, j, k) * rk3Ctd_zzz + vector(rk3Gamh_y, i, j, k) * rk3Ctd_zyz + vector(rk3Gamh_x, i, j, k) * rk3Ctd_zxz;
						rk3m_Rtd_zz_o0_t8_l0 = vector(rk3gtd_zz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
						rk3m_Rtd_zz_o0_t7_l0 = vector(rk3gtd_yz, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
						rk3m_Rtd_zz_o0_t6_l0 = vector(rk3gtd_xz, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
						rk3m_Rtd_zz_o0_t5_l0 = (-0.5 * rk3gtu_zz) * d_Rtd_zz_o0_t5_m0_l0;
						rk3m_Rtd_zz_o0_t4_l0 = (-rk3gtu_yz) * d_Rtd_zz_o0_t4_m0_l0;
						rk3m_Rtd_zz_o0_t3_l0 = (-0.5 * rk3gtu_yy) * d_Rtd_zz_o0_t3_m0_l0;
						rk3m_Rtd_zz_o0_t2_l0 = (-rk3gtu_xz) * d_Rtd_zz_o0_t2_m0_l0;
						rk3m_Rtd_zz_o0_t1_l0 = (-rk3gtu_xy) * d_Rtd_zz_o0_t1_m0_l0;
						rk3m_Rtd_zz_o0_t0_l0 = (-0.5 * rk3gtu_xx) * d_Rtd_zz_o0_t0_m0_l0;
						rk3Rtd_zz = ((((((((rk3m_Rtd_zz_o0_t0_l0 + rk3m_Rtd_zz_o0_t1_l0) + rk3m_Rtd_zz_o0_t2_l0) + rk3m_Rtd_zz_o0_t3_l0) + rk3m_Rtd_zz_o0_t4_l0) + rk3m_Rtd_zz_o0_t5_l0) + rk3m_Rtd_zz_o0_t6_l0) + rk3m_Rtd_zz_o0_t7_l0) + rk3m_Rtd_zz_o0_t8_l0) + rk3d_Rtd_zz_o0_t9_m0_l0;
						rk3d_Rtd_xy_o0_t12_m0_l0 = rk3gtu_zz * (rk3Ct_xxz * rk3Ctd_yxz + rk3Ct_xxz * rk3Ctd_xyz + rk3Ct_xyz * rk3Ctd_xxz + 2.0 * rk3Ct_yxz * rk3Ctd_yyz + rk3Ct_yyz * rk3Ctd_xyz + rk3Ct_zxz * rk3Ctd_yzz + rk3Ct_zxz * rk3Ctd_zyz + rk3Ct_zyz * rk3Ctd_xzz) + rk3gtu_yz * (rk3Ct_xxy * rk3Ctd_yxz + rk3Ct_xxy * rk3Ctd_xyz + rk3Ct_xyy * rk3Ctd_xxz + 2.0 * rk3Ct_yxy * rk3Ctd_yyz + rk3Ct_yyy * rk3Ctd_xyz + rk3Ct_zxy * rk3Ctd_yzz + rk3Ct_zxy * rk3Ctd_zyz + rk3Ct_zyy * rk3Ctd_xzz) + rk3gtu_xz * (rk3Ct_xxx * rk3Ctd_yxz + rk3Ct_xxx * rk3Ctd_xyz + rk3Ct_xxy * rk3Ctd_xxz + 2.0 * rk3Ct_yxx * rk3Ctd_yyz + rk3Ct_yxy * rk3Ctd_xyz + rk3Ct_zxx * rk3Ctd_yzz + rk3Ct_zxx * rk3Ctd_zyz + rk3Ct_zxy * rk3Ctd_xzz) + rk3gtu_yz * (rk3Ct_xxz * rk3Ctd_yxy + rk3Ct_xxz * rk3Ctd_xyy + rk3Ct_xyz * rk3Ctd_xxy + 2.0 * rk3Ct_yxz * rk3Ctd_yyy + rk3Ct_yyz * rk3Ctd_xyy + rk3Ct_zxz * rk3Ctd_yyz + rk3Ct_zxz * rk3Ctd_zyy + rk3Ct_zyz * rk3Ctd_xyz) + rk3gtu_yy * (rk3Ct_xxy * rk3Ctd_yxy + rk3Ct_xxy * rk3Ctd_xyy + rk3Ct_xyy * rk3Ctd_xxy + 2.0 * rk3Ct_yxy * rk3Ctd_yyy + rk3Ct_yyy * rk3Ctd_xyy + rk3Ct_zxy * rk3Ctd_yyz + rk3Ct_zxy * rk3Ctd_zyy + rk3Ct_zyy * rk3Ctd_xyz) + rk3gtu_xy * (rk3Ct_xxx * rk3Ctd_yxy + rk3Ct_xxx * rk3Ctd_xyy + rk3Ct_xxy * rk3Ctd_xxy + 2.0 * rk3Ct_yxx * rk3Ctd_yyy + rk3Ct_yxy * rk3Ctd_xyy + rk3Ct_zxx * rk3Ctd_yyz + rk3Ct_zxx * rk3Ctd_zyy + rk3Ct_zxy * rk3Ctd_xyz) + rk3gtu_xz * (rk3Ct_xxz * rk3Ctd_yxx + rk3Ct_xxz * rk3Ctd_xxy + rk3Ct_xyz * rk3Ctd_xxx + 2.0 * rk3Ct_yxz * rk3Ctd_yxy + rk3Ct_yyz * rk3Ctd_xxy + rk3Ct_zxz * rk3Ctd_yxz + rk3Ct_zxz * rk3Ctd_zxy + rk3Ct_zyz * rk3Ctd_xxz) + rk3gtu_xy * (rk3Ct_xxy * rk3Ctd_yxx + rk3Ct_xxy * rk3Ctd_xxy + rk3Ct_xyy * rk3Ctd_xxx + 2.0 * rk3Ct_yxy * rk3Ctd_yxy + rk3Ct_yyy * rk3Ctd_xxy + rk3Ct_zxy * rk3Ctd_yxz + rk3Ct_zxy * rk3Ctd_zxy + rk3Ct_zyy * rk3Ctd_xxz) + rk3gtu_xx * (rk3Ct_xxx * rk3Ctd_yxx + rk3Ct_xxx * rk3Ctd_xxy + rk3Ct_xxy * rk3Ctd_xxx + 2.0 * rk3Ct_yxx * rk3Ctd_yxy + rk3Ct_yxy * rk3Ctd_xxy + rk3Ct_zxx * rk3Ctd_yxz + rk3Ct_zxx * rk3Ctd_zxy + rk3Ct_zxy * rk3Ctd_xxz) + 0.5 * vector(rk3Gamh_z, i, j, k) * (rk3Ctd_xyz + rk3Ctd_yxz) + 0.5 * vector(rk3Gamh_y, i, j, k) * (rk3Ctd_xyy + rk3Ctd_yxy) + 0.5 * vector(rk3Gamh_x, i, j, k) * (rk3Ctd_xxy + rk3Ctd_yxx);
						rk3m_Rtd_xy_o0_t11_l0 = 0.5 * vector(rk3gtd_yz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
						rk3m_Rtd_xy_o0_t10_l0 = 0.5 * vector(rk3gtd_yy, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
						rk3m_Rtd_xy_o0_t9_l0 = 0.5 * vector(rk3gtd_xy, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
						rk3m_Rtd_xy_o0_t8_l0 = 0.5 * vector(rk3gtd_xz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
						rk3m_Rtd_xy_o0_t7_l0 = 0.5 * vector(rk3gtd_xy, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
						rk3m_Rtd_xy_o0_t6_l0 = 0.5 * vector(rk3gtd_xx, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
						rk3m_Rtd_xy_o0_t5_l0 = (-0.5 * rk3gtu_zz) * d_Rtd_xy_o0_t5_m0_l0;
						rk3m_Rtd_xy_o0_t4_l0 = (-rk3gtu_yz) * d_Rtd_xy_o0_t4_m0_l0;
						rk3m_Rtd_xy_o0_t3_l0 = (-0.5 * rk3gtu_yy) * d_Rtd_xy_o0_t3_m0_l0;
						rk3m_Rtd_xy_o0_t2_l0 = (-rk3gtu_xz) * d_Rtd_xy_o0_t2_m0_l0;
						rk3m_Rtd_xy_o0_t1_l0 = (-rk3gtu_xy) * d_Rtd_xy_o0_t1_m0_l0;
						rk3m_Rtd_xy_o0_t0_l0 = (-0.5 * rk3gtu_xx) * d_Rtd_xy_o0_t0_m0_l0;
						rk3Rtd_xy = (((((((((((rk3m_Rtd_xy_o0_t0_l0 + rk3m_Rtd_xy_o0_t1_l0) + rk3m_Rtd_xy_o0_t2_l0) + rk3m_Rtd_xy_o0_t3_l0) + rk3m_Rtd_xy_o0_t4_l0) + rk3m_Rtd_xy_o0_t5_l0) + rk3m_Rtd_xy_o0_t6_l0) + rk3m_Rtd_xy_o0_t7_l0) + rk3m_Rtd_xy_o0_t8_l0) + rk3m_Rtd_xy_o0_t9_l0) + rk3m_Rtd_xy_o0_t10_l0) + rk3m_Rtd_xy_o0_t11_l0) + rk3d_Rtd_xy_o0_t12_m0_l0;
						rk3d_Rtd_yy_o0_t9_m0_l0 = rk3gtu_zz * (2.0 * rk3Ct_xyz * rk3Ctd_yxz + rk3Ct_xyz * rk3Ctd_xyz + 3.0 * rk3Ct_yyz * rk3Ctd_yyz + 2.0 * rk3Ct_zyz * rk3Ctd_yzz + rk3Ct_zyz * rk3Ctd_zyz) + rk3gtu_yz * (2.0 * rk3Ct_xyy * rk3Ctd_yxz + rk3Ct_xyy * rk3Ctd_xyz + 3.0 * rk3Ct_yyy * rk3Ctd_yyz + 2.0 * rk3Ct_zyy * rk3Ctd_yzz + rk3Ct_zyy * rk3Ctd_zyz) + rk3gtu_xz * (2.0 * rk3Ct_xxy * rk3Ctd_yxz + rk3Ct_xxy * rk3Ctd_xyz + 3.0 * rk3Ct_yxy * rk3Ctd_yyz + 2.0 * rk3Ct_zxy * rk3Ctd_yzz + rk3Ct_zxy * rk3Ctd_zyz) + rk3gtu_yz * (2.0 * rk3Ct_xyz * rk3Ctd_yxy + rk3Ct_xyz * rk3Ctd_xyy + 3.0 * rk3Ct_yyz * rk3Ctd_yyy + 2.0 * rk3Ct_zyz * rk3Ctd_yyz + rk3Ct_zyz * rk3Ctd_zyy) + rk3gtu_yy * (2.0 * rk3Ct_xyy * rk3Ctd_yxy + rk3Ct_xyy * rk3Ctd_xyy + 3.0 * rk3Ct_yyy * rk3Ctd_yyy + 2.0 * rk3Ct_zyy * rk3Ctd_yyz + rk3Ct_zyy * rk3Ctd_zyy) + rk3gtu_xy * (2.0 * rk3Ct_xxy * rk3Ctd_yxy + rk3Ct_xxy * rk3Ctd_xyy + 3.0 * rk3Ct_yxy * rk3Ctd_yyy + 2.0 * rk3Ct_zxy * rk3Ctd_yyz + rk3Ct_zxy * rk3Ctd_zyy) + rk3gtu_xz * (2.0 * rk3Ct_xyz * rk3Ctd_yxx + rk3Ct_xyz * rk3Ctd_xxy + 3.0 * rk3Ct_yyz * rk3Ctd_yxy + 2.0 * rk3Ct_zyz * rk3Ctd_yxz + rk3Ct_zyz * rk3Ctd_zxy) + rk3gtu_xy * (2.0 * rk3Ct_xyy * rk3Ctd_yxx + rk3Ct_xyy * rk3Ctd_xxy + 3.0 * rk3Ct_yyy * rk3Ctd_yxy + 2.0 * rk3Ct_zyy * rk3Ctd_yxz + rk3Ct_zyy * rk3Ctd_zxy) + rk3gtu_xx * (2.0 * rk3Ct_xxy * rk3Ctd_yxx + rk3Ct_xxy * rk3Ctd_xxy + 3.0 * rk3Ct_yxy * rk3Ctd_yxy + 2.0 * rk3Ct_zxy * rk3Ctd_yxz + rk3Ct_zxy * rk3Ctd_zxy) + vector(rk3Gamh_z, i, j, k) * rk3Ctd_yyz + vector(rk3Gamh_y, i, j, k) * rk3Ctd_yyy + vector(rk3Gamh_x, i, j, k) * rk3Ctd_yxy;
						rk3m_Rtd_yy_o0_t8_l0 = vector(rk3gtd_yz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
						rk3m_Rtd_yy_o0_t7_l0 = vector(rk3gtd_yy, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
						rk3m_Rtd_yy_o0_t6_l0 = vector(rk3gtd_xy, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
						rk3m_Rtd_yy_o0_t5_l0 = (-0.5 * rk3gtu_zz) * d_Rtd_yy_o0_t5_m0_l0;
						rk3m_Rtd_yy_o0_t4_l0 = (-rk3gtu_yz) * d_Rtd_yy_o0_t4_m0_l0;
						rk3m_Rtd_yy_o0_t3_l0 = (-0.5 * rk3gtu_yy) * d_Rtd_yy_o0_t3_m0_l0;
						rk3m_Rtd_yy_o0_t2_l0 = (-rk3gtu_xz) * d_Rtd_yy_o0_t2_m0_l0;
						rk3m_Rtd_yy_o0_t1_l0 = (-rk3gtu_xy) * d_Rtd_yy_o0_t1_m0_l0;
						rk3m_Rtd_yy_o0_t0_l0 = (-0.5 * rk3gtu_xx) * d_Rtd_yy_o0_t0_m0_l0;
						rk3Rtd_yy = ((((((((rk3m_Rtd_yy_o0_t0_l0 + rk3m_Rtd_yy_o0_t1_l0) + rk3m_Rtd_yy_o0_t2_l0) + rk3m_Rtd_yy_o0_t3_l0) + rk3m_Rtd_yy_o0_t4_l0) + rk3m_Rtd_yy_o0_t5_l0) + rk3m_Rtd_yy_o0_t6_l0) + rk3m_Rtd_yy_o0_t7_l0) + rk3m_Rtd_yy_o0_t8_l0) + rk3d_Rtd_yy_o0_t9_m0_l0;
						rk3d_Rtd_yz_o0_t12_m0_l0 = rk3gtu_zz * (rk3Ct_xyz * rk3Ctd_zxz + rk3Ct_xyz * rk3Ctd_xzz + rk3Ct_xzz * rk3Ctd_yxz + rk3Ct_yyz * rk3Ctd_zyz + rk3Ct_yyz * rk3Ctd_yzz + rk3Ct_yzz * rk3Ctd_yyz + 2.0 * rk3Ct_zyz * rk3Ctd_zzz + rk3Ct_zzz * rk3Ctd_yzz) + rk3gtu_yz * (rk3Ct_xyy * rk3Ctd_zxz + rk3Ct_xyy * rk3Ctd_xzz + rk3Ct_xyz * rk3Ctd_yxz + rk3Ct_yyy * rk3Ctd_zyz + rk3Ct_yyy * rk3Ctd_yzz + rk3Ct_yyz * rk3Ctd_yyz + 2.0 * rk3Ct_zyy * rk3Ctd_zzz + rk3Ct_zyz * rk3Ctd_yzz) + rk3gtu_xz * (rk3Ct_xxy * rk3Ctd_zxz + rk3Ct_xxy * rk3Ctd_xzz + rk3Ct_xxz * rk3Ctd_yxz + rk3Ct_yxy * rk3Ctd_zyz + rk3Ct_yxy * rk3Ctd_yzz + rk3Ct_yxz * rk3Ctd_yyz + 2.0 * rk3Ct_zxy * rk3Ctd_zzz + rk3Ct_zxz * rk3Ctd_yzz) + rk3gtu_yz * (rk3Ct_xyz * rk3Ctd_zxy + rk3Ct_xyz * rk3Ctd_xyz + rk3Ct_xzz * rk3Ctd_yxy + rk3Ct_yyz * rk3Ctd_zyy + rk3Ct_yyz * rk3Ctd_yyz + rk3Ct_yzz * rk3Ctd_yyy + 2.0 * rk3Ct_zyz * rk3Ctd_zyz + rk3Ct_zzz * rk3Ctd_yyz) + rk3gtu_yy * (rk3Ct_xyy * rk3Ctd_zxy + rk3Ct_xyy * rk3Ctd_xyz + rk3Ct_xyz * rk3Ctd_yxy + rk3Ct_yyy * rk3Ctd_zyy + rk3Ct_yyy * rk3Ctd_yyz + rk3Ct_yyz * rk3Ctd_yyy + 2.0 * rk3Ct_zyy * rk3Ctd_zyz + rk3Ct_zyz * rk3Ctd_yyz) + rk3gtu_xy * (rk3Ct_xxy * rk3Ctd_zxy + rk3Ct_xxy * rk3Ctd_xyz + rk3Ct_xxz * rk3Ctd_yxy + rk3Ct_yxy * rk3Ctd_zyy + rk3Ct_yxy * rk3Ctd_yyz + rk3Ct_yxz * rk3Ctd_yyy + 2.0 * rk3Ct_zxy * rk3Ctd_zyz + rk3Ct_zxz * rk3Ctd_yyz) + rk3gtu_xz * (rk3Ct_xyz * rk3Ctd_zxx + rk3Ct_xyz * rk3Ctd_xxz + rk3Ct_xzz * rk3Ctd_yxx + rk3Ct_yyz * rk3Ctd_zxy + rk3Ct_yyz * rk3Ctd_yxz + rk3Ct_yzz * rk3Ctd_yxy + 2.0 * rk3Ct_zyz * rk3Ctd_zxz + rk3Ct_zzz * rk3Ctd_yxz) + rk3gtu_xy * (rk3Ct_xyy * rk3Ctd_zxx + rk3Ct_xyy * rk3Ctd_xxz + rk3Ct_xyz * rk3Ctd_yxx + rk3Ct_yyy * rk3Ctd_zxy + rk3Ct_yyy * rk3Ctd_yxz + rk3Ct_yyz * rk3Ctd_yxy + 2.0 * rk3Ct_zyy * rk3Ctd_zxz + rk3Ct_zyz * rk3Ctd_yxz) + rk3gtu_xx * (rk3Ct_xxy * rk3Ctd_zxx + rk3Ct_xxy * rk3Ctd_xxz + rk3Ct_xxz * rk3Ctd_yxx + rk3Ct_yxy * rk3Ctd_zxy + rk3Ct_yxy * rk3Ctd_yxz + rk3Ct_yxz * rk3Ctd_yxy + 2.0 * rk3Ct_zxy * rk3Ctd_zxz + rk3Ct_zxz * rk3Ctd_yxz) + 0.5 * vector(rk3Gamh_z, i, j, k) * (rk3Ctd_yzz + rk3Ctd_zyz) + 0.5 * vector(rk3Gamh_y, i, j, k) * (rk3Ctd_yyz + rk3Ctd_zyy) + 0.5 * vector(rk3Gamh_x, i, j, k) * (rk3Ctd_yxz + rk3Ctd_zxy);
						rk3m_Rtd_yz_o0_t11_l0 = 0.5 * vector(rk3gtd_zz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
						rk3m_Rtd_yz_o0_t10_l0 = 0.5 * vector(rk3gtd_yz, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
						rk3m_Rtd_yz_o0_t9_l0 = 0.5 * vector(rk3gtd_xz, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
						rk3m_Rtd_yz_o0_t8_l0 = 0.5 * vector(rk3gtd_yz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
						rk3m_Rtd_yz_o0_t7_l0 = 0.5 * vector(rk3gtd_yy, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
						rk3m_Rtd_yz_o0_t6_l0 = 0.5 * vector(rk3gtd_xy, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
						rk3m_Rtd_yz_o0_t5_l0 = (-0.5 * rk3gtu_zz) * d_Rtd_yz_o0_t5_m0_l0;
						rk3m_Rtd_yz_o0_t4_l0 = (-rk3gtu_yz) * d_Rtd_yz_o0_t4_m0_l0;
						rk3m_Rtd_yz_o0_t3_l0 = (-0.5 * rk3gtu_yy) * d_Rtd_yz_o0_t3_m0_l0;
						rk3m_Rtd_yz_o0_t2_l0 = (-rk3gtu_xz) * d_Rtd_yz_o0_t2_m0_l0;
						rk3m_Rtd_yz_o0_t1_l0 = (-rk3gtu_xy) * d_Rtd_yz_o0_t1_m0_l0;
						rk3m_Rtd_yz_o0_t0_l0 = (-0.5 * rk3gtu_xx) * d_Rtd_yz_o0_t0_m0_l0;
						rk3Rtd_yz = (((((((((((rk3m_Rtd_yz_o0_t0_l0 + rk3m_Rtd_yz_o0_t1_l0) + rk3m_Rtd_yz_o0_t2_l0) + rk3m_Rtd_yz_o0_t3_l0) + rk3m_Rtd_yz_o0_t4_l0) + rk3m_Rtd_yz_o0_t5_l0) + rk3m_Rtd_yz_o0_t6_l0) + rk3m_Rtd_yz_o0_t7_l0) + rk3m_Rtd_yz_o0_t8_l0) + rk3m_Rtd_yz_o0_t9_l0) + rk3m_Rtd_yz_o0_t10_l0) + rk3m_Rtd_yz_o0_t11_l0) + rk3d_Rtd_yz_o0_t12_m0_l0;
						rk3m_Rpd_xx_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t3_m0_l0;
						rk3m_Rpd_xx_o0_t21_l0 = (-0.75 * vector(rk3gtd_xx, i, j, k) * rk3gtu_zz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xx_o0_t20_l0 = (-0.75 * vector(rk3gtd_xx, i, j, k) * rk3gtu_yy * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xx_o0_t19_l0 = (-0.75 * vector(rk3gtd_xx, i, j, k) * rk3gtu_xx * rk3inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xx_o0_t18_l0 = 2.0 * rk3inv_chi * rk3Zu_z * vector(rk3gtd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xx_o0_t17_l0 = 2.0 * rk3inv_chi * rk3Zu_y * vector(rk3gtd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xx_o0_t16_l0 = 2.0 * rk3inv_chi * rk3Zu_x * vector(rk3gtd_xx, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xx_o0_t15_l0 = (-1.500000000000000 * vector(rk3gtd_xx, i, j, k) * rk3gtu_yz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xx_o0_t14_l0 = (-1.500000000000000 * vector(rk3gtd_xx, i, j, k) * rk3gtu_xz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xx_o0_t13_l0 = (-1.500000000000000 * vector(rk3gtd_xx, i, j, k) * rk3gtu_xy * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xx_o0_t12_l0 = (-0.25 * rk3inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xx_o0_t11_l0 = (-0.5 * rk3Ct_zxx) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xx_o0_t10_l0 = (-0.5 * rk3Ct_yxx) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xx_o0_t9_l0 = (-0.5 * rk3Ct_xxx) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xx_o0_t8_l0 = 0.5 * vector(rk3gtd_xx, i, j, k) * rk3gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk3m_Rpd_xx_o0_t7_l0 = vector(rk3gtd_xx, i, j, k) * rk3gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk3m_Rpd_xx_o0_t6_l0 = 0.5 * vector(rk3gtd_xx, i, j, k) * rk3gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk3m_Rpd_xx_o0_t5_l0 = vector(rk3gtd_xx, i, j, k) * rk3gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk3m_Rpd_xx_o0_t4_l0 = vector(rk3gtd_xx, i, j, k) * rk3gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk3m_Rpd_xx_o0_t3_l0 = 0.5 * vector(rk3gtd_xx, i, j, k) * rk3gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk3m_Rpd_xx_o0_t2_l0 = (-0.5 * vector(rk3gtd_xx, i, j, k) * vector(rk3Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xx_o0_t1_l0 = (-0.5 * vector(rk3gtd_xx, i, j, k) * vector(rk3Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xx_o0_t0_l0 = (-0.5 * vector(rk3gtd_xx, i, j, k) * vector(rk3Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk3Rpd_xx = (((((((((((((((((((((rk3m_Rpd_xx_o0_t0_l0 + rk3m_Rpd_xx_o0_t1_l0) + rk3m_Rpd_xx_o0_t2_l0) + rk3m_Rpd_xx_o0_t3_l0) + rk3m_Rpd_xx_o0_t4_l0) + rk3m_Rpd_xx_o0_t5_l0) + rk3m_Rpd_xx_o0_t6_l0) + rk3m_Rpd_xx_o0_t7_l0) + rk3m_Rpd_xx_o0_t8_l0) + rk3m_Rpd_xx_o0_t9_l0) + rk3m_Rpd_xx_o0_t10_l0) + rk3m_Rpd_xx_o0_t11_l0) + rk3m_Rpd_xx_o0_t12_l0) + rk3m_Rpd_xx_o0_t13_l0) + rk3m_Rpd_xx_o0_t14_l0) + rk3m_Rpd_xx_o0_t15_l0) + rk3m_Rpd_xx_o0_t16_l0) + rk3m_Rpd_xx_o0_t17_l0) + rk3m_Rpd_xx_o0_t18_l0) + rk3m_Rpd_xx_o0_t19_l0) + rk3m_Rpd_xx_o0_t20_l0) + rk3m_Rpd_xx_o0_t21_l0) + rk3m_Rpd_xx_o0_t22_l0;
						rk3m_Rpd_yy_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t6_m0_l0;
						rk3m_Rpd_yy_o0_t21_l0 = (-0.75 * vector(rk3gtd_yy, i, j, k) * rk3gtu_zz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_yy_o0_t20_l0 = (-0.75 * vector(rk3gtd_yy, i, j, k) * rk3gtu_yy * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yy_o0_t19_l0 = (-0.75 * vector(rk3gtd_yy, i, j, k) * rk3gtu_xx * rk3inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_yy_o0_t18_l0 = 2.0 * rk3inv_chi * rk3Zu_z * vector(rk3gtd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yy_o0_t17_l0 = 2.0 * rk3inv_chi * rk3Zu_y * vector(rk3gtd_yy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yy_o0_t16_l0 = 2.0 * rk3inv_chi * rk3Zu_x * vector(rk3gtd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yy_o0_t15_l0 = (-1.500000000000000 * vector(rk3gtd_yy, i, j, k) * rk3gtu_xz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_yy_o0_t14_l0 = (-1.500000000000000 * vector(rk3gtd_yy, i, j, k) * rk3gtu_xy * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_yy_o0_t13_l0 = (-0.25 * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yy_o0_t12_l0 = (-0.5 * rk3Ct_zyy) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_yy_o0_t11_l0 = (-0.5 * rk3Ct_yyy) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yy_o0_t10_l0 = (-0.5 * rk3Ct_xyy) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_yy_o0_t9_l0 = 0.5 * vector(rk3gtd_yy, i, j, k) * rk3gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk3m_Rpd_yy_o0_t8_l0 = vector(rk3gtd_yy, i, j, k) * rk3gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk3m_Rpd_yy_o0_t7_l0 = 0.5 * vector(rk3gtd_yy, i, j, k) * rk3gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk3m_Rpd_yy_o0_t6_l0 = vector(rk3gtd_yy, i, j, k) * rk3gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk3m_Rpd_yy_o0_t5_l0 = vector(rk3gtd_yy, i, j, k) * rk3gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk3m_Rpd_yy_o0_t4_l0 = 0.5 * vector(rk3gtd_yy, i, j, k) * rk3gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk3m_Rpd_yy_o0_t3_l0 = (-0.5 * vector(rk3gtd_yy, i, j, k) * vector(rk3Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_yy_o0_t2_l0 = (-0.5 * vector(rk3gtd_yy, i, j, k) * vector(rk3Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yy_o0_t1_l0 = (-0.5 * vector(rk3gtd_yy, i, j, k) * vector(rk3Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_yy_o0_t0_l0 = (-1.500000000000000 * vector(rk3gtd_yy, i, j, k) * rk3gtu_yz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3Rpd_yy = (((((((((((((((((((((rk3m_Rpd_yy_o0_t0_l0 + rk3m_Rpd_yy_o0_t1_l0) + rk3m_Rpd_yy_o0_t2_l0) + rk3m_Rpd_yy_o0_t3_l0) + rk3m_Rpd_yy_o0_t4_l0) + rk3m_Rpd_yy_o0_t5_l0) + rk3m_Rpd_yy_o0_t6_l0) + rk3m_Rpd_yy_o0_t7_l0) + rk3m_Rpd_yy_o0_t8_l0) + rk3m_Rpd_yy_o0_t9_l0) + rk3m_Rpd_yy_o0_t10_l0) + rk3m_Rpd_yy_o0_t11_l0) + rk3m_Rpd_yy_o0_t12_l0) + rk3m_Rpd_yy_o0_t13_l0) + rk3m_Rpd_yy_o0_t14_l0) + rk3m_Rpd_yy_o0_t15_l0) + rk3m_Rpd_yy_o0_t16_l0) + rk3m_Rpd_yy_o0_t17_l0) + rk3m_Rpd_yy_o0_t18_l0) + rk3m_Rpd_yy_o0_t19_l0) + rk3m_Rpd_yy_o0_t20_l0) + rk3m_Rpd_yy_o0_t21_l0) + rk3m_Rpd_yy_o0_t22_l0;
						rk3m_Rpd_yz_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t7_m0_l0;
						rk3m_Rpd_yz_o0_t24_l0 = (-0.75 * vector(rk3gtd_yz, i, j, k) * rk3gtu_zz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_yz_o0_t23_l0 = (-0.75 * vector(rk3gtd_yz, i, j, k) * rk3gtu_yy * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yz_o0_t22_l0 = (-0.75 * vector(rk3gtd_yz, i, j, k) * rk3gtu_xx * rk3inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_yz_o0_t21_l0 = rk3inv_chi * rk3Zu_z * vector(rk3gtd_zz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yz_o0_t20_l0 = rk3inv_chi * rk3Zu_y * vector(rk3gtd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yz_o0_t19_l0 = rk3inv_chi * rk3Zu_x * vector(rk3gtd_xz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yz_o0_t18_l0 = rk3inv_chi * rk3Zu_z * vector(rk3gtd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_yz_o0_t17_l0 = rk3inv_chi * rk3Zu_y * vector(rk3gtd_yy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_yz_o0_t16_l0 = rk3inv_chi * rk3Zu_x * vector(rk3gtd_xy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_yz_o0_t15_l0 = (-0.5 * rk3Ct_zyz) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_yz_o0_t14_l0 = (-0.5 * rk3Ct_yyz) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yz_o0_t13_l0 = (-0.5 * rk3Ct_xyz) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_yz_o0_t12_l0 = (-0.25 * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_yz_o0_t11_l0 = 0.5 * vector(rk3gtd_yz, i, j, k) * rk3gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk3m_Rpd_yz_o0_t10_l0 = vector(rk3gtd_yz, i, j, k) * rk3gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk3m_Rpd_yz_o0_t9_l0 = 0.5 * vector(rk3gtd_yz, i, j, k) * rk3gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk3m_Rpd_yz_o0_t8_l0 = vector(rk3gtd_yz, i, j, k) * rk3gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk3m_Rpd_yz_o0_t7_l0 = vector(rk3gtd_yz, i, j, k) * rk3gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk3m_Rpd_yz_o0_t6_l0 = 0.5 * vector(rk3gtd_yz, i, j, k) * rk3gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk3m_Rpd_yz_o0_t5_l0 = (-0.5 * vector(rk3gtd_yz, i, j, k) * vector(rk3Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_yz_o0_t4_l0 = (-0.5 * vector(rk3gtd_yz, i, j, k) * vector(rk3Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yz_o0_t3_l0 = (-0.5 * vector(rk3gtd_yz, i, j, k) * vector(rk3Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_yz_o0_t2_l0 = (-1.500000000000000 * vector(rk3gtd_yz, i, j, k) * rk3gtu_yz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_yz_o0_t1_l0 = (-1.500000000000000 * vector(rk3gtd_yz, i, j, k) * rk3gtu_xz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_yz_o0_t0_l0 = (-1.500000000000000 * vector(rk3gtd_yz, i, j, k) * rk3gtu_xy * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3Rpd_yz = ((((((((((((((((((((((((rk3m_Rpd_yz_o0_t0_l0 + rk3m_Rpd_yz_o0_t1_l0) + rk3m_Rpd_yz_o0_t2_l0) + rk3m_Rpd_yz_o0_t3_l0) + rk3m_Rpd_yz_o0_t4_l0) + rk3m_Rpd_yz_o0_t5_l0) + rk3m_Rpd_yz_o0_t6_l0) + rk3m_Rpd_yz_o0_t7_l0) + rk3m_Rpd_yz_o0_t8_l0) + rk3m_Rpd_yz_o0_t9_l0) + rk3m_Rpd_yz_o0_t10_l0) + rk3m_Rpd_yz_o0_t11_l0) + rk3m_Rpd_yz_o0_t12_l0) + rk3m_Rpd_yz_o0_t13_l0) + rk3m_Rpd_yz_o0_t14_l0) + rk3m_Rpd_yz_o0_t15_l0) + rk3m_Rpd_yz_o0_t16_l0) + rk3m_Rpd_yz_o0_t17_l0) + rk3m_Rpd_yz_o0_t18_l0) + rk3m_Rpd_yz_o0_t19_l0) + rk3m_Rpd_yz_o0_t20_l0) + rk3m_Rpd_yz_o0_t21_l0) + rk3m_Rpd_yz_o0_t22_l0) + rk3m_Rpd_yz_o0_t23_l0) + rk3m_Rpd_yz_o0_t24_l0) + rk3m_Rpd_yz_o0_t25_l0;
						rk3m_Rpd_xz_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t5_m0_l0;
						rk3m_Rpd_xz_o0_t24_l0 = (-0.75 * vector(rk3gtd_xz, i, j, k) * rk3gtu_zz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xz_o0_t23_l0 = (-0.75 * vector(rk3gtd_xz, i, j, k) * rk3gtu_yy * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xz_o0_t22_l0 = (-0.75 * vector(rk3gtd_xz, i, j, k) * rk3gtu_xx * rk3inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xz_o0_t21_l0 = rk3inv_chi * rk3Zu_z * vector(rk3gtd_zz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xz_o0_t20_l0 = rk3inv_chi * rk3Zu_y * vector(rk3gtd_yz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xz_o0_t19_l0 = rk3inv_chi * rk3Zu_x * vector(rk3gtd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xz_o0_t18_l0 = rk3inv_chi * rk3Zu_z * vector(rk3gtd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xz_o0_t17_l0 = rk3inv_chi * rk3Zu_y * vector(rk3gtd_xy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xz_o0_t16_l0 = rk3inv_chi * rk3Zu_x * vector(rk3gtd_xx, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xz_o0_t15_l0 = (-1.500000000000000 * vector(rk3gtd_xz, i, j, k) * rk3gtu_yz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xz_o0_t14_l0 = (-1.500000000000000 * vector(rk3gtd_xz, i, j, k) * rk3gtu_xz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xz_o0_t13_l0 = (-1.500000000000000 * vector(rk3gtd_xz, i, j, k) * rk3gtu_xy * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xz_o0_t12_l0 = (-0.5 * rk3Ct_zxz) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xz_o0_t11_l0 = (-0.5 * rk3Ct_yxz) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xz_o0_t10_l0 = (-0.5 * rk3Ct_xxz) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xz_o0_t9_l0 = (-0.25 * rk3inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xz_o0_t8_l0 = 0.5 * vector(rk3gtd_xz, i, j, k) * rk3gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk3m_Rpd_xz_o0_t7_l0 = vector(rk3gtd_xz, i, j, k) * rk3gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk3m_Rpd_xz_o0_t6_l0 = 0.5 * vector(rk3gtd_xz, i, j, k) * rk3gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk3m_Rpd_xz_o0_t5_l0 = vector(rk3gtd_xz, i, j, k) * rk3gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk3m_Rpd_xz_o0_t4_l0 = vector(rk3gtd_xz, i, j, k) * rk3gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk3m_Rpd_xz_o0_t3_l0 = 0.5 * vector(rk3gtd_xz, i, j, k) * rk3gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk3m_Rpd_xz_o0_t2_l0 = (-0.5 * vector(rk3gtd_xz, i, j, k) * vector(rk3Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xz_o0_t1_l0 = (-0.5 * vector(rk3gtd_xz, i, j, k) * vector(rk3Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xz_o0_t0_l0 = (-0.5 * vector(rk3gtd_xz, i, j, k) * vector(rk3Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk3Rpd_xz = ((((((((((((((((((((((((rk3m_Rpd_xz_o0_t0_l0 + rk3m_Rpd_xz_o0_t1_l0) + rk3m_Rpd_xz_o0_t2_l0) + rk3m_Rpd_xz_o0_t3_l0) + rk3m_Rpd_xz_o0_t4_l0) + rk3m_Rpd_xz_o0_t5_l0) + rk3m_Rpd_xz_o0_t6_l0) + rk3m_Rpd_xz_o0_t7_l0) + rk3m_Rpd_xz_o0_t8_l0) + rk3m_Rpd_xz_o0_t9_l0) + rk3m_Rpd_xz_o0_t10_l0) + rk3m_Rpd_xz_o0_t11_l0) + rk3m_Rpd_xz_o0_t12_l0) + rk3m_Rpd_xz_o0_t13_l0) + rk3m_Rpd_xz_o0_t14_l0) + rk3m_Rpd_xz_o0_t15_l0) + rk3m_Rpd_xz_o0_t16_l0) + rk3m_Rpd_xz_o0_t17_l0) + rk3m_Rpd_xz_o0_t18_l0) + rk3m_Rpd_xz_o0_t19_l0) + rk3m_Rpd_xz_o0_t20_l0) + rk3m_Rpd_xz_o0_t21_l0) + rk3m_Rpd_xz_o0_t22_l0) + rk3m_Rpd_xz_o0_t23_l0) + rk3m_Rpd_xz_o0_t24_l0) + rk3m_Rpd_xz_o0_t25_l0;
						rk3m_Rpd_xy_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t4_m0_l0;
						rk3m_Rpd_xy_o0_t24_l0 = (-0.75 * vector(rk3gtd_xy, i, j, k) * rk3gtu_zz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xy_o0_t23_l0 = (-0.75 * vector(rk3gtd_xy, i, j, k) * rk3gtu_yy * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xy_o0_t22_l0 = (-0.75 * vector(rk3gtd_xy, i, j, k) * rk3gtu_xx * rk3inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xy_o0_t21_l0 = rk3inv_chi * rk3Zu_z * vector(rk3gtd_yz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xy_o0_t20_l0 = rk3inv_chi * rk3Zu_y * vector(rk3gtd_yy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xy_o0_t19_l0 = rk3inv_chi * rk3Zu_x * vector(rk3gtd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xy_o0_t18_l0 = rk3inv_chi * rk3Zu_z * vector(rk3gtd_xz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xy_o0_t17_l0 = rk3inv_chi * rk3Zu_y * vector(rk3gtd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xy_o0_t16_l0 = rk3inv_chi * rk3Zu_x * vector(rk3gtd_xx, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xy_o0_t15_l0 = (-1.500000000000000 * vector(rk3gtd_xy, i, j, k) * rk3gtu_yz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xy_o0_t14_l0 = (-1.500000000000000 * vector(rk3gtd_xy, i, j, k) * rk3gtu_xz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xy_o0_t13_l0 = (-1.500000000000000 * vector(rk3gtd_xy, i, j, k) * rk3gtu_xy * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xy_o0_t12_l0 = (-0.5 * rk3Ct_zxy) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xy_o0_t11_l0 = (-0.5 * rk3Ct_yxy) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xy_o0_t10_l0 = (-0.5 * rk3Ct_xxy) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xy_o0_t9_l0 = 0.5 * vector(rk3gtd_xy, i, j, k) * rk3gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk3m_Rpd_xy_o0_t8_l0 = vector(rk3gtd_xy, i, j, k) * rk3gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk3m_Rpd_xy_o0_t7_l0 = 0.5 * vector(rk3gtd_xy, i, j, k) * rk3gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk3m_Rpd_xy_o0_t6_l0 = vector(rk3gtd_xy, i, j, k) * rk3gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk3m_Rpd_xy_o0_t5_l0 = vector(rk3gtd_xy, i, j, k) * rk3gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk3m_Rpd_xy_o0_t4_l0 = 0.5 * vector(rk3gtd_xy, i, j, k) * rk3gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk3m_Rpd_xy_o0_t3_l0 = (-0.5 * vector(rk3gtd_xy, i, j, k) * vector(rk3Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_xy_o0_t2_l0 = (-0.5 * vector(rk3gtd_xy, i, j, k) * vector(rk3Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_xy_o0_t1_l0 = (-0.5 * vector(rk3gtd_xy, i, j, k) * vector(rk3Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_xy_o0_t0_l0 = (-0.25 * rk3inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3Rpd_xy = ((((((((((((((((((((((((rk3m_Rpd_xy_o0_t0_l0 + rk3m_Rpd_xy_o0_t1_l0) + rk3m_Rpd_xy_o0_t2_l0) + rk3m_Rpd_xy_o0_t3_l0) + rk3m_Rpd_xy_o0_t4_l0) + rk3m_Rpd_xy_o0_t5_l0) + rk3m_Rpd_xy_o0_t6_l0) + rk3m_Rpd_xy_o0_t7_l0) + rk3m_Rpd_xy_o0_t8_l0) + rk3m_Rpd_xy_o0_t9_l0) + rk3m_Rpd_xy_o0_t10_l0) + rk3m_Rpd_xy_o0_t11_l0) + rk3m_Rpd_xy_o0_t12_l0) + rk3m_Rpd_xy_o0_t13_l0) + rk3m_Rpd_xy_o0_t14_l0) + rk3m_Rpd_xy_o0_t15_l0) + rk3m_Rpd_xy_o0_t16_l0) + rk3m_Rpd_xy_o0_t17_l0) + rk3m_Rpd_xy_o0_t18_l0) + rk3m_Rpd_xy_o0_t19_l0) + rk3m_Rpd_xy_o0_t20_l0) + rk3m_Rpd_xy_o0_t21_l0) + rk3m_Rpd_xy_o0_t22_l0) + rk3m_Rpd_xy_o0_t23_l0) + rk3m_Rpd_xy_o0_t24_l0) + rk3m_Rpd_xy_o0_t25_l0;
						rk3m_Rpd_zz_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t8_m0_l0;
						rk3m_Rpd_zz_o0_t21_l0 = (-0.75 * vector(rk3gtd_zz, i, j, k) * rk3gtu_zz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_zz_o0_t20_l0 = (-0.75 * vector(rk3gtd_zz, i, j, k) * rk3gtu_yy * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_zz_o0_t19_l0 = (-0.75 * vector(rk3gtd_zz, i, j, k) * rk3gtu_xx * rk3inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_zz_o0_t18_l0 = 2.0 * rk3inv_chi * rk3Zu_z * vector(rk3gtd_zz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_zz_o0_t17_l0 = 2.0 * rk3inv_chi * rk3Zu_y * vector(rk3gtd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_zz_o0_t16_l0 = 2.0 * rk3inv_chi * rk3Zu_x * vector(rk3gtd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_zz_o0_t15_l0 = (-0.25 * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_zz_o0_t14_l0 = (-0.5 * rk3Ct_zzz) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_zz_o0_t13_l0 = (-0.5 * rk3Ct_yzz) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_zz_o0_t12_l0 = (-0.5 * rk3Ct_xzz) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_zz_o0_t11_l0 = vector(rk3gtd_zz, i, j, k) * rk3gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
						rk3m_Rpd_zz_o0_t10_l0 = vector(rk3gtd_zz, i, j, k) * rk3gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
						rk3m_Rpd_zz_o0_t9_l0 = 0.5 * vector(rk3gtd_zz, i, j, k) * rk3gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
						rk3m_Rpd_zz_o0_t8_l0 = (-0.5 * vector(rk3gtd_zz, i, j, k) * vector(rk3Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Rpd_zz_o0_t7_l0 = (-0.5 * vector(rk3gtd_zz, i, j, k) * vector(rk3Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_zz_o0_t6_l0 = (-0.5 * vector(rk3gtd_zz, i, j, k) * vector(rk3Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_zz_o0_t5_l0 = 0.5 * vector(rk3gtd_zz, i, j, k) * rk3gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
						rk3m_Rpd_zz_o0_t4_l0 = vector(rk3gtd_zz, i, j, k) * rk3gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
						rk3m_Rpd_zz_o0_t3_l0 = 0.5 * vector(rk3gtd_zz, i, j, k) * rk3gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
						rk3m_Rpd_zz_o0_t2_l0 = (-1.500000000000000 * vector(rk3gtd_zz, i, j, k) * rk3gtu_yz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Rpd_zz_o0_t1_l0 = (-1.500000000000000 * vector(rk3gtd_zz, i, j, k) * rk3gtu_xz * rk3inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Rpd_zz_o0_t0_l0 = (-1.500000000000000 * vector(rk3gtd_zz, i, j, k) * rk3gtu_xy * rk3inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3Rpd_zz = (((((((((((((((((((((rk3m_Rpd_zz_o0_t0_l0 + rk3m_Rpd_zz_o0_t1_l0) + rk3m_Rpd_zz_o0_t2_l0) + rk3m_Rpd_zz_o0_t3_l0) + rk3m_Rpd_zz_o0_t4_l0) + rk3m_Rpd_zz_o0_t5_l0) + rk3m_Rpd_zz_o0_t6_l0) + rk3m_Rpd_zz_o0_t7_l0) + rk3m_Rpd_zz_o0_t8_l0) + rk3m_Rpd_zz_o0_t9_l0) + rk3m_Rpd_zz_o0_t10_l0) + rk3m_Rpd_zz_o0_t11_l0) + rk3m_Rpd_zz_o0_t12_l0) + rk3m_Rpd_zz_o0_t13_l0) + rk3m_Rpd_zz_o0_t14_l0) + rk3m_Rpd_zz_o0_t15_l0) + rk3m_Rpd_zz_o0_t16_l0) + rk3m_Rpd_zz_o0_t17_l0) + rk3m_Rpd_zz_o0_t18_l0) + rk3m_Rpd_zz_o0_t19_l0) + rk3m_Rpd_zz_o0_t20_l0) + rk3m_Rpd_zz_o0_t21_l0) + rk3m_Rpd_zz_o0_t22_l0;
						rk3Rscalar = (rk3chi_max * rk3Rtd_xx + rk3Rpd_xx) * rk3gtu_xx + 2.0 * (rk3chi_max * rk3Rtd_xy + rk3Rpd_xy) * rk3gtu_xy + 2.0 * (rk3chi_max * rk3Rtd_xz + rk3Rpd_xz) * rk3gtu_xz + (rk3chi_max * rk3Rtd_yy + rk3Rpd_yy) * rk3gtu_yy + 2.0 * (rk3chi_max * rk3Rtd_yz + rk3Rpd_yz) * rk3gtu_yz + (rk3chi_max * rk3Rtd_zz + rk3Rpd_zz) * rk3gtu_zz;
						rk3Bfvf = vector(rk3Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(rk3Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(rk3Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
						rk3Sfuu_zz = rk3Sfu_z * rk3vfu_z + rk3chi_max * rk3gtu_zz * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_zz) + vector(rk3Bfu_z, i, j, k) * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_zz) + 2.0 * vector(rk3Bfu_z, i, j, k) * rk3vfu_z)) / rk3sdetg);
						rk3Tf4u_zz = rk3Sfuu_zz / rk3sdetg + (-(rk3Sfu_z * vector(rk3Betau_z, i, j, k)) / (rk3sdetg * vector(rk3Alpha, i, j, k))) + (-vector(rk3Betau_z, i, j, k) * rk3Tf4u_tz);
						rk3Tu_zz = rk3Tf4u_zz;
						rk3Sfuu_yz = 0.5 * rk3Sfu_y * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_y + rk3chi_max * rk3gtu_yz * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k) * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yz) + vector(rk3Bfu_y, i, j, k) * rk3vfu_z + vector(rk3Bfu_z, i, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Tf4u_yz = rk3Sfuu_yz / rk3sdetg + (-(rk3Sfu_y * vector(rk3Betau_z, i, j, k)) / (rk3sdetg * vector(rk3Alpha, i, j, k))) + (-vector(rk3Betau_y, i, j, k) * rk3Tf4u_tz);
						rk3Tu_yz = rk3Tf4u_yz;
						rk3Sfuu_yy = rk3Sfu_y * rk3vfu_y + rk3chi_max * rk3gtu_yy * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_yy) + vector(rk3Bfu_y, i, j, k) * vector(rk3Bfu_y, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_yy) + 2.0 * vector(rk3Bfu_y, i, j, k) * rk3vfu_y)) / rk3sdetg);
						rk3Tf4u_yy = rk3Sfuu_yy / rk3sdetg + (-(rk3Sfu_y * vector(rk3Betau_y, i, j, k)) / (rk3sdetg * vector(rk3Alpha, i, j, k))) + (-vector(rk3Betau_y, i, j, k) * rk3Tf4u_ty);
						rk3Tu_yy = rk3Tf4u_yy;
						rk3Sfuu_xy = 0.5 * rk3Sfu_x * rk3vfu_y + 0.5 * rk3Sfu_y * rk3vfu_x + rk3chi_max * rk3gtu_xy * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k) * vector(rk3Bfu_y, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xy) + vector(rk3Bfu_x, i, j, k) * rk3vfu_y + vector(rk3Bfu_y, i, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_yy = rk3inv_chi * (rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_yy * vector(rk3gtd_yy, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k));
						rk3Tf4u_xy = rk3Sfuu_xy / rk3sdetg + (-(rk3Sfu_x * vector(rk3Betau_y, i, j, k)) / (rk3sdetg * vector(rk3Alpha, i, j, k))) + (-vector(rk3Betau_x, i, j, k) * rk3Tf4u_ty);
						rk3Tu_xy = rk3Tf4u_xy;
						rk3Sfuu_xx = rk3Sfu_x * rk3vfu_x + rk3chi_max * rk3gtu_xx * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xx) + vector(rk3Bfu_x, i, j, k) * vector(rk3Bfu_x, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xx) + 2.0 * vector(rk3Bfu_x, i, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Tf4u_xx = rk3Sfuu_xx / rk3sdetg + (-(rk3Sfu_x * vector(rk3Betau_x, i, j, k)) / (rk3sdetg * vector(rk3Alpha, i, j, k))) + (-vector(rk3Betau_x, i, j, k) * rk3Tf4u_tx);
						rk3Tu_xx = rk3Tf4u_xx;
						rk3Sfuu_xz = 0.5 * rk3Sfu_x * rk3vfu_z + 0.5 * rk3Sfu_z * rk3vfu_x + rk3chi_max * rk3gtu_xz * rk3sdetg * vector(pf, i, j, k) + (-((-0.5 * rk3sqBf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k) * vector(rk3Bfu_z, i, j, k)) / (rk3sdetg * rk3sqW)) + (-0.5 * (rk3Bfvf * ((-rk3Bfvf * rk3chi_max * rk3gtu_xz) + vector(rk3Bfu_x, i, j, k) * rk3vfu_z + vector(rk3Bfu_z, i, j, k) * rk3vfu_x)) / rk3sdetg);
						rk3Sfud_xx = rk3inv_chi * (rk3Sfuu_xx * vector(rk3gtd_xx, i, j, k) + rk3Sfuu_xy * vector(rk3gtd_xy, i, j, k) + rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k));
						rk3Tf4u_xz = rk3Sfuu_xz / rk3sdetg + (-(rk3Sfu_x * vector(rk3Betau_z, i, j, k)) / (rk3sdetg * vector(rk3Alpha, i, j, k))) + (-vector(rk3Betau_x, i, j, k) * rk3Tf4u_tz);
						rk3Tu_xz = rk3Tf4u_xz;
						rk3Sfud_zz = rk3inv_chi * (rk3Sfuu_xz * vector(rk3gtd_xz, i, j, k) + rk3Sfuu_yz * vector(rk3gtd_yz, i, j, k) + rk3Sfuu_zz * vector(rk3gtd_zz, i, j, k));
						rk3trSf = rk3Sfud_xx + rk3Sfud_yy + rk3Sfud_zz;
						rk3Tu_tt = rk3Tf4u_tt;
						rk3pTtd_ADM_xz = rk3Betatd_x * rk3Betatd_z * rk3Tu_tt + rk3Betatd_x * (vector(rk3gtd_xz, i, j, k) * rk3Tu_tx + vector(rk3gtd_yz, i, j, k) * rk3Tu_ty + vector(rk3gtd_zz, i, j, k) * rk3Tu_tz) + rk3Betatd_z * (vector(rk3gtd_xx, i, j, k) * rk3Tu_tx + vector(rk3gtd_xy, i, j, k) * rk3Tu_ty + vector(rk3gtd_xz, i, j, k) * rk3Tu_tz) + vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_xz, i, j, k) * rk3Tu_xx + vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yz, i, j, k) * rk3Tu_xy + vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_zz, i, j, k) * rk3Tu_xz + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k) * rk3Tu_xy + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_yz, i, j, k) * rk3Tu_yy + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_zz, i, j, k) * rk3Tu_yz + (vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k)) * rk3Tu_xz + vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_yz, i, j, k) * rk3Tu_yz + vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_zz, i, j, k) * rk3Tu_zz;
						rk3pTtd_ADM_xy = rk3Betatd_x * rk3Betatd_y * rk3Tu_tt + rk3Betatd_x * (vector(rk3gtd_xy, i, j, k) * rk3Tu_tx + vector(rk3gtd_yy, i, j, k) * rk3Tu_ty + vector(rk3gtd_yz, i, j, k) * rk3Tu_tz) + rk3Betatd_y * (vector(rk3gtd_xx, i, j, k) * rk3Tu_tx + vector(rk3gtd_xy, i, j, k) * rk3Tu_ty + vector(rk3gtd_xz, i, j, k) * rk3Tu_tz) + vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_xy, i, j, k) * rk3Tu_xx + vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yy, i, j, k) * rk3Tu_xy + vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_yz, i, j, k) * rk3Tu_xz + (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k)) * rk3Tu_xy + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_yy, i, j, k) * rk3Tu_yy + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_yz, i, j, k) * rk3Tu_yz + vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xy, i, j, k) * rk3Tu_xz + vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_yy, i, j, k) * rk3Tu_yz + vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_yz, i, j, k) * rk3Tu_zz;
						rk3pTtd_ADM_xx = (rk3Betatd_x * rk3Betatd_x) * rk3Tu_tt + 2.0 * rk3Betatd_x * (vector(rk3gtd_xx, i, j, k) * rk3Tu_tx + vector(rk3gtd_xy, i, j, k) * rk3Tu_ty + vector(rk3gtd_xz, i, j, k) * rk3Tu_tz) + (vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_xx, i, j, k)) * rk3Tu_xx + 2.0 * vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_xy, i, j, k) * rk3Tu_xy + 2.0 * vector(rk3gtd_xx, i, j, k) * vector(rk3gtd_xz, i, j, k) * rk3Tu_xz + (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k)) * rk3Tu_yy + 2.0 * vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k) * rk3Tu_yz + (vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k)) * rk3Tu_zz;
						rk3d_Psi1_xx_o0_t5_m0_l0 = (-8.0) * M_PI * rk3inv_chi * vector(rk3Alpha, i, j, k) * rk3pTtd_ADM_xx + vector(rk3Alpha, i, j, k) * rk3Rpd_xx + rk3chi_max * vector(rk3Alpha, i, j, k) * rk3Rtd_xx;
						rk3m_Psi1_xx_o0_t4_l0 = (-rk3chi_max * rk3inv_chi) * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Psi1_xx_o0_t3_l0 = rk3chi_max * rk3Ct_zxx * d_Psi1_xx_o0_t3_m0_l0;
						rk3m_Psi1_xx_o0_t2_l0 = rk3chi_max * rk3Ct_yxx * d_Psi1_xx_o0_t2_m0_l0;
						rk3m_Psi1_xx_o0_t1_l0 = rk3chi_max * rk3Ct_xxx * d_Psi1_xx_o0_t1_m0_l0;
						rk3m_Psi1_xx_o0_t0_l0 = (-rk3chi_max) * d_Psi1_xx_o0_t0_m0_l0;
						rk3Psi1_xx = ((((rk3m_Psi1_xx_o0_t0_l0 + rk3m_Psi1_xx_o0_t1_l0) + rk3m_Psi1_xx_o0_t2_l0) + rk3m_Psi1_xx_o0_t3_l0) + rk3m_Psi1_xx_o0_t4_l0) + rk3d_Psi1_xx_o0_t5_m0_l0;
						rk3d_Psi1_xy_o0_t6_m0_l0 = (-8.0) * M_PI * rk3inv_chi * vector(rk3Alpha, i, j, k) * rk3pTtd_ADM_xy + vector(rk3Alpha, i, j, k) * rk3Rpd_xy + rk3chi_max * vector(rk3Alpha, i, j, k) * rk3Rtd_xy;
						rk3m_Psi1_xy_o0_t5_l0 = (-0.5 * rk3chi_max * rk3inv_chi) * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Psi1_xy_o0_t4_l0 = (-0.5 * rk3chi_max * rk3inv_chi) * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Psi1_xy_o0_t3_l0 = rk3chi_max * rk3Ct_zxy * d_Psi1_xx_o0_t3_m0_l0;
						rk3m_Psi1_xy_o0_t2_l0 = rk3chi_max * rk3Ct_yxy * d_Psi1_xx_o0_t2_m0_l0;
						rk3m_Psi1_xy_o0_t1_l0 = rk3chi_max * rk3Ct_xxy * d_Psi1_xx_o0_t1_m0_l0;
						rk3m_Psi1_xy_o0_t0_l0 = (-rk3chi_max) * d_Psi1_xy_o0_t0_m0_l0;
						rk3Psi1_xy = (((((rk3m_Psi1_xy_o0_t0_l0 + rk3m_Psi1_xy_o0_t1_l0) + rk3m_Psi1_xy_o0_t2_l0) + rk3m_Psi1_xy_o0_t3_l0) + rk3m_Psi1_xy_o0_t4_l0) + rk3m_Psi1_xy_o0_t5_l0) + rk3d_Psi1_xy_o0_t6_m0_l0;
						rk3d_Psi1_xz_o0_t6_m0_l0 = (-8.0) * M_PI * rk3inv_chi * vector(rk3Alpha, i, j, k) * rk3pTtd_ADM_xz + vector(rk3Alpha, i, j, k) * rk3Rpd_xz + rk3chi_max * vector(rk3Alpha, i, j, k) * rk3Rtd_xz;
						rk3m_Psi1_xz_o0_t5_l0 = (-0.5 * rk3chi_max * rk3inv_chi) * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						rk3m_Psi1_xz_o0_t4_l0 = (-0.5 * rk3chi_max * rk3inv_chi) * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Psi1_xz_o0_t3_l0 = rk3chi_max * rk3Ct_zxz * d_Psi1_xx_o0_t3_m0_l0;
						rk3m_Psi1_xz_o0_t2_l0 = rk3chi_max * rk3Ct_yxz * d_Psi1_xx_o0_t2_m0_l0;
						rk3m_Psi1_xz_o0_t1_l0 = rk3chi_max * rk3Ct_xxz * d_Psi1_xx_o0_t1_m0_l0;
						rk3m_Psi1_xz_o0_t0_l0 = (-rk3chi_max) * d_Psi1_xz_o0_t0_m0_l0;
						rk3Psi1_xz = (((((rk3m_Psi1_xz_o0_t0_l0 + rk3m_Psi1_xz_o0_t1_l0) + rk3m_Psi1_xz_o0_t2_l0) + rk3m_Psi1_xz_o0_t3_l0) + rk3m_Psi1_xz_o0_t4_l0) + rk3m_Psi1_xz_o0_t5_l0) + rk3d_Psi1_xz_o0_t6_m0_l0;
						rk3Jtd_ADM_z = vector(rk3Alpha, i, j, k) * ((vector(rk3Betau_x, i, j, k) * rk3Tu_tt + rk3Tu_tx) * vector(rk3gtd_xz, i, j, k) + (vector(rk3Betau_y, i, j, k) * rk3Tu_tt + rk3Tu_ty) * vector(rk3gtd_yz, i, j, k) + (vector(rk3Betau_z, i, j, k) * rk3Tu_tt + rk3Tu_tz) * vector(rk3gtd_zz, i, j, k));
						rk3pTtd_ADM_zz = (rk3Betatd_z * rk3Betatd_z) * rk3Tu_tt + 2.0 * rk3Betatd_z * (vector(rk3gtd_xz, i, j, k) * rk3Tu_tx + vector(rk3gtd_yz, i, j, k) * rk3Tu_ty + vector(rk3gtd_zz, i, j, k) * rk3Tu_tz) + (vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_xz, i, j, k)) * rk3Tu_xx + 2.0 * vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_yz, i, j, k) * rk3Tu_xy + 2.0 * vector(rk3gtd_xz, i, j, k) * vector(rk3gtd_zz, i, j, k) * rk3Tu_xz + (vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k)) * rk3Tu_yy + 2.0 * vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_zz, i, j, k) * rk3Tu_yz + (vector(rk3gtd_zz, i, j, k) * vector(rk3gtd_zz, i, j, k)) * rk3Tu_zz;
						rk3d_Psi1_zz_o0_t5_m0_l0 = (-8.0) * M_PI * rk3inv_chi * vector(rk3Alpha, i, j, k) * rk3pTtd_ADM_zz + vector(rk3Alpha, i, j, k) * rk3Rpd_zz + rk3chi_max * vector(rk3Alpha, i, j, k) * rk3Rtd_zz;
						rk3m_Psi1_zz_o0_t4_l0 = (-rk3chi_max * rk3inv_chi) * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Psi1_zz_o0_t3_l0 = rk3chi_max * rk3Ct_zzz * d_Psi1_xx_o0_t3_m0_l0;
						rk3m_Psi1_zz_o0_t2_l0 = rk3chi_max * rk3Ct_yzz * d_Psi1_xx_o0_t2_m0_l0;
						rk3m_Psi1_zz_o0_t1_l0 = rk3chi_max * rk3Ct_xzz * d_Psi1_xx_o0_t1_m0_l0;
						rk3m_Psi1_zz_o0_t0_l0 = (-rk3chi_max) * d_Psi1_zz_o0_t0_m0_l0;
						rk3Psi1_zz = ((((rk3m_Psi1_zz_o0_t0_l0 + rk3m_Psi1_zz_o0_t1_l0) + rk3m_Psi1_zz_o0_t2_l0) + rk3m_Psi1_zz_o0_t3_l0) + rk3m_Psi1_zz_o0_t4_l0) + rk3d_Psi1_zz_o0_t5_m0_l0;
						rk3Jtd_ADM_y = vector(rk3Alpha, i, j, k) * ((vector(rk3Betau_x, i, j, k) * rk3Tu_tt + rk3Tu_tx) * vector(rk3gtd_xy, i, j, k) + (vector(rk3Betau_y, i, j, k) * rk3Tu_tt + rk3Tu_ty) * vector(rk3gtd_yy, i, j, k) + (vector(rk3Betau_z, i, j, k) * rk3Tu_tt + rk3Tu_tz) * vector(rk3gtd_yz, i, j, k));
						rk3Jtd_ADM_x = vector(rk3Alpha, i, j, k) * ((vector(rk3Betau_x, i, j, k) * rk3Tu_tt + rk3Tu_tx) * vector(rk3gtd_xx, i, j, k) + (vector(rk3Betau_y, i, j, k) * rk3Tu_tt + rk3Tu_ty) * vector(rk3gtd_xy, i, j, k) + (vector(rk3Betau_z, i, j, k) * rk3Tu_tt + rk3Tu_tz) * vector(rk3gtd_xz, i, j, k));
						rk3pTtd_ADM_yz = rk3Betatd_y * rk3Betatd_z * rk3Tu_tt + rk3Betatd_y * (vector(rk3gtd_xz, i, j, k) * rk3Tu_tx + vector(rk3gtd_yz, i, j, k) * rk3Tu_ty + vector(rk3gtd_zz, i, j, k) * rk3Tu_tz) + rk3Betatd_z * (vector(rk3gtd_xy, i, j, k) * rk3Tu_tx + vector(rk3gtd_yy, i, j, k) * rk3Tu_ty + vector(rk3gtd_yz, i, j, k) * rk3Tu_tz) + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xz, i, j, k) * rk3Tu_xx + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_yz, i, j, k) * rk3Tu_xy + vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_zz, i, j, k) * rk3Tu_xz + vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_xz, i, j, k) * rk3Tu_xy + vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_yz, i, j, k) * rk3Tu_yy + vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_zz, i, j, k) * rk3Tu_yz + vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_xz, i, j, k) * rk3Tu_xz + (vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k)) * rk3Tu_yz + vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_zz, i, j, k) * rk3Tu_zz;
						rk3d_Psi1_yz_o0_t6_m0_l0 = (-8.0) * M_PI * rk3inv_chi * vector(rk3Alpha, i, j, k) * rk3pTtd_ADM_yz + vector(rk3Alpha, i, j, k) * rk3Rpd_yz + rk3chi_max * vector(rk3Alpha, i, j, k) * rk3Rtd_yz;
						rk3m_Psi1_yz_o0_t5_l0 = (-0.5 * rk3chi_max * rk3inv_chi) * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Psi1_yz_o0_t4_l0 = (-0.5 * rk3chi_max * rk3inv_chi) * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						rk3m_Psi1_yz_o0_t3_l0 = rk3chi_max * rk3Ct_zyz * d_Psi1_xx_o0_t3_m0_l0;
						rk3m_Psi1_yz_o0_t2_l0 = rk3chi_max * rk3Ct_yyz * d_Psi1_xx_o0_t2_m0_l0;
						rk3m_Psi1_yz_o0_t1_l0 = rk3chi_max * rk3Ct_xyz * d_Psi1_xx_o0_t1_m0_l0;
						rk3m_Psi1_yz_o0_t0_l0 = (-rk3chi_max) * d_Psi1_yz_o0_t0_m0_l0;
						rk3Psi1_yz = (((((rk3m_Psi1_yz_o0_t0_l0 + rk3m_Psi1_yz_o0_t1_l0) + rk3m_Psi1_yz_o0_t2_l0) + rk3m_Psi1_yz_o0_t3_l0) + rk3m_Psi1_yz_o0_t4_l0) + rk3m_Psi1_yz_o0_t5_l0) + rk3d_Psi1_yz_o0_t6_m0_l0;
						rk3pTtd_ADM_yy = (rk3Betatd_y * rk3Betatd_y) * rk3Tu_tt + 2.0 * rk3Betatd_y * (vector(rk3gtd_xy, i, j, k) * rk3Tu_tx + vector(rk3gtd_yy, i, j, k) * rk3Tu_ty + vector(rk3gtd_yz, i, j, k) * rk3Tu_tz) + (vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_xy, i, j, k)) * rk3Tu_xx + 2.0 * vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_yy, i, j, k) * rk3Tu_xy + 2.0 * vector(rk3gtd_xy, i, j, k) * vector(rk3gtd_yz, i, j, k) * rk3Tu_xz + (vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_yy, i, j, k)) * rk3Tu_yy + 2.0 * vector(rk3gtd_yy, i, j, k) * vector(rk3gtd_yz, i, j, k) * rk3Tu_yz + (vector(rk3gtd_yz, i, j, k) * vector(rk3gtd_yz, i, j, k)) * rk3Tu_zz;
						rk3tr_pT = rk3inv_chi * (rk3gtu_xx * rk3pTtd_ADM_xx + 2.0 * rk3gtu_xy * rk3pTtd_ADM_xy + 2.0 * rk3gtu_xz * rk3pTtd_ADM_xz + rk3gtu_yy * rk3pTtd_ADM_yy + 2.0 * rk3gtu_yz * rk3pTtd_ADM_yz + rk3gtu_zz * rk3pTtd_ADM_zz);
						rk3d_Psi1_yy_o0_t5_m0_l0 = (-8.0) * M_PI * rk3inv_chi * vector(rk3Alpha, i, j, k) * rk3pTtd_ADM_yy + vector(rk3Alpha, i, j, k) * rk3Rpd_yy + rk3chi_max * vector(rk3Alpha, i, j, k) * rk3Rtd_yy;
						rk3m_Psi1_yy_o0_t4_l0 = (-rk3chi_max * rk3inv_chi) * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						rk3m_Psi1_yy_o0_t3_l0 = rk3chi_max * rk3Ct_zyy * d_Psi1_xx_o0_t3_m0_l0;
						rk3m_Psi1_yy_o0_t2_l0 = rk3chi_max * rk3Ct_yyy * d_Psi1_xx_o0_t2_m0_l0;
						rk3m_Psi1_yy_o0_t1_l0 = rk3chi_max * rk3Ct_xyy * d_Psi1_xx_o0_t1_m0_l0;
						rk3m_Psi1_yy_o0_t0_l0 = (-rk3chi_max) * d_Psi1_yy_o0_t0_m0_l0;
						rk3Psi1_yy = ((((rk3m_Psi1_yy_o0_t0_l0 + rk3m_Psi1_yy_o0_t1_l0) + rk3m_Psi1_yy_o0_t2_l0) + rk3m_Psi1_yy_o0_t3_l0) + rk3m_Psi1_yy_o0_t4_l0) + rk3d_Psi1_yy_o0_t5_m0_l0;
						rk3trPsi1 = rk3Psi1_xx * rk3gtu_xx + 2.0 * rk3Psi1_xy * rk3gtu_xy + 2.0 * rk3Psi1_xz * rk3gtu_xz + rk3Psi1_yy * rk3gtu_yy + 2.0 * rk3Psi1_yz * rk3gtu_yz + rk3Psi1_zz * rk3gtu_zz;
						rk3Psi1TF_yz = (-0.3333333333333333 * rk3trPsi1 * vector(rk3gtd_yz, i, j, k)) + rk3Psi1_yz;
						rk3Psi1TF_yy = (-0.3333333333333333 * rk3trPsi1 * vector(rk3gtd_yy, i, j, k)) + rk3Psi1_yy;
						rk3Psi1TF_xy = (-0.3333333333333333 * rk3trPsi1 * vector(rk3gtd_xy, i, j, k)) + rk3Psi1_xy;
						rk3Psi1TF_xx = (-0.3333333333333333 * rk3trPsi1 * vector(rk3gtd_xx, i, j, k)) + rk3Psi1_xx;
						rk3Psi1TF_xz = (-0.3333333333333333 * rk3trPsi1 * vector(rk3gtd_xz, i, j, k)) + rk3Psi1_xz;
						rk3Psi1TF_zz = (-0.3333333333333333 * rk3trPsi1 * vector(rk3gtd_zz, i, j, k)) + rk3Psi1_zz;
						rk3kappa_cc = MIN(p_kappa_cc, p_kappa_cc * rk3decay_factor) / vector(rk3Alpha, i, j, k);
						rk3rho_ADM = (vector(rk3Alpha, i, j, k) * vector(rk3Alpha, i, j, k)) * rk3Tu_tt;
						d_Sfd_x_o0_t11_m0_l0 = vector(rk3Alpha, i, j, k) * rk3sdetg * rk3sqW * vector(qnu, i, j, k) * rk3vfu_x;
						d_Sfd_y_o0_t11_m0_l0 = vector(rk3Alpha, i, j, k) * rk3sdetg * rk3sqW * vector(qnu, i, j, k) * rk3vfu_y;
						d_Sfd_z_o0_t11_m0_l0 = vector(rk3Alpha, i, j, k) * rk3sdetg * rk3sqW * vector(qnu, i, j, k) * rk3vfu_z;
						d_DYf_o0_t0_m0_l0 = vector(rk3Alpha, i, j, k) * rk3sdetg * vector(rnu, i, j, k);
						d_tauf_o0_t3_m0_l0 = vector(rk3Alpha, i, j, k) * rk3sdetg * rk3sqW * vector(qnu, i, j, k);
						d_gtd_xx_o0_t3_m0_l0 = (-0.3333333333333333) * rk3kappa_cc * vector(rk3Alpha, i, j, k) * vector(rk3gtd_xx, i, j, k) * log(rk3detgtd) + (-2.0) * vector(rk3Alpha, i, j, k) * ((-0.3333333333333333 * rk3trAt * lambda_0 * vector(rk3gtd_xx, i, j, k)) + vector(rk3Atd_xx, i, j, k)) + (-0.6666666666666667) * vector(rk3gtd_xx, i, j, k) * rk3div_Beta;
						d_gtd_xy_o0_t6_m0_l0 = (-0.3333333333333333) * rk3kappa_cc * vector(rk3Alpha, i, j, k) * vector(rk3gtd_xy, i, j, k) * log(rk3detgtd) + (-2.0) * vector(rk3Alpha, i, j, k) * ((-0.3333333333333333 * rk3trAt * lambda_0 * vector(rk3gtd_xy, i, j, k)) + vector(rk3Atd_xy, i, j, k)) + (-0.6666666666666667) * vector(rk3gtd_xy, i, j, k) * rk3div_Beta;
						d_gtd_xz_o0_t6_m0_l0 = (-0.3333333333333333) * rk3kappa_cc * vector(rk3Alpha, i, j, k) * vector(rk3gtd_xz, i, j, k) * log(rk3detgtd) + (-2.0) * vector(rk3Alpha, i, j, k) * ((-0.3333333333333333 * rk3trAt * lambda_0 * vector(rk3gtd_xz, i, j, k)) + vector(rk3Atd_xz, i, j, k)) + (-0.6666666666666667) * vector(rk3gtd_xz, i, j, k) * rk3div_Beta;
						d_gtd_yy_o0_t3_m0_l0 = (-0.3333333333333333) * rk3kappa_cc * vector(rk3Alpha, i, j, k) * vector(rk3gtd_yy, i, j, k) * log(rk3detgtd) + (-2.0) * vector(rk3Alpha, i, j, k) * ((-0.3333333333333333 * rk3trAt * lambda_0 * vector(rk3gtd_yy, i, j, k)) + vector(rk3Atd_yy, i, j, k)) + (-0.6666666666666667) * vector(rk3gtd_yy, i, j, k) * rk3div_Beta;
						d_gtd_yz_o0_t6_m0_l0 = (-0.3333333333333333) * rk3kappa_cc * vector(rk3Alpha, i, j, k) * vector(rk3gtd_yz, i, j, k) * log(rk3detgtd) + (-2.0) * vector(rk3Alpha, i, j, k) * ((-0.3333333333333333 * rk3trAt * lambda_0 * vector(rk3gtd_yz, i, j, k)) + vector(rk3Atd_yz, i, j, k)) + (-0.6666666666666667) * vector(rk3gtd_yz, i, j, k) * rk3div_Beta;
						d_gtd_zz_o0_t3_m0_l0 = (-0.3333333333333333) * rk3kappa_cc * vector(rk3Alpha, i, j, k) * vector(rk3gtd_zz, i, j, k) * log(rk3detgtd) + (-2.0) * vector(rk3Alpha, i, j, k) * ((-0.3333333333333333 * rk3trAt * lambda_0 * vector(rk3gtd_zz, i, j, k)) + vector(rk3Atd_zz, i, j, k)) + (-0.6666666666666667) * vector(rk3gtd_zz, i, j, k) * rk3div_Beta;
						d_Atd_xx_o0_t3_m0_l0 = (-0.3333333333333333) * rk3kappa_cc * vector(rk3Alpha, i, j, k) * vector(rk3gtd_xx, i, j, k) * rk3trAt + vector(rk3Alpha, i, j, k) * (vector(rk3trK, i, j, k) * vector(rk3Atd_xx, i, j, k) + (-2.0 * vector(rk3Atd_xx, i, j, k) * rk3Atud_xx) + (-2.0 * vector(rk3Atd_xy, i, j, k) * rk3Atud_yx) + (-2.0 * vector(rk3Atd_xz, i, j, k) * rk3Atud_zx)) + rk3Psi1TF_xx + (-0.6666666666666667) * vector(rk3Atd_xx, i, j, k) * rk3div_Beta;
						d_Atd_xy_o0_t6_m0_l0 = (-0.3333333333333333) * rk3kappa_cc * vector(rk3Alpha, i, j, k) * vector(rk3gtd_xy, i, j, k) * rk3trAt + vector(rk3Alpha, i, j, k) * (vector(rk3trK, i, j, k) * vector(rk3Atd_xy, i, j, k) + (-2.0 * vector(rk3Atd_xx, i, j, k) * rk3Atud_xy) + (-2.0 * vector(rk3Atd_xy, i, j, k) * rk3Atud_yy) + (-2.0 * vector(rk3Atd_xz, i, j, k) * rk3Atud_zy)) + rk3Psi1TF_xy + (-0.6666666666666667) * vector(rk3Atd_xy, i, j, k) * rk3div_Beta;
						d_Atd_xz_o0_t6_m0_l0 = (-0.3333333333333333) * rk3kappa_cc * vector(rk3Alpha, i, j, k) * vector(rk3gtd_xz, i, j, k) * rk3trAt + vector(rk3Alpha, i, j, k) * (vector(rk3trK, i, j, k) * vector(rk3Atd_xz, i, j, k) + (-2.0 * vector(rk3Atd_xx, i, j, k) * rk3Atud_xz) + (-2.0 * vector(rk3Atd_xy, i, j, k) * rk3Atud_yz) + (-2.0 * vector(rk3Atd_xz, i, j, k) * rk3Atud_zz)) + rk3Psi1TF_xz + (-0.6666666666666667) * vector(rk3Atd_xz, i, j, k) * rk3div_Beta;
						d_Atd_yy_o0_t3_m0_l0 = (-0.3333333333333333) * rk3kappa_cc * vector(rk3Alpha, i, j, k) * vector(rk3gtd_yy, i, j, k) * rk3trAt + vector(rk3Alpha, i, j, k) * (vector(rk3trK, i, j, k) * vector(rk3Atd_yy, i, j, k) + (-2.0 * vector(rk3Atd_xy, i, j, k) * rk3Atud_xy) + (-2.0 * vector(rk3Atd_yy, i, j, k) * rk3Atud_yy) + (-2.0 * vector(rk3Atd_yz, i, j, k) * rk3Atud_zy)) + rk3Psi1TF_yy + (-0.6666666666666667) * vector(rk3Atd_yy, i, j, k) * rk3div_Beta;
						d_Atd_yz_o0_t6_m0_l0 = (-0.3333333333333333) * rk3kappa_cc * vector(rk3Alpha, i, j, k) * vector(rk3gtd_yz, i, j, k) * rk3trAt + vector(rk3Alpha, i, j, k) * (vector(rk3trK, i, j, k) * vector(rk3Atd_yz, i, j, k) + (-2.0 * vector(rk3Atd_xy, i, j, k) * rk3Atud_xz) + (-2.0 * vector(rk3Atd_yy, i, j, k) * rk3Atud_yz) + (-2.0 * vector(rk3Atd_yz, i, j, k) * rk3Atud_zz)) + rk3Psi1TF_yz + (-0.6666666666666667) * vector(rk3Atd_yz, i, j, k) * rk3div_Beta;
						d_Atd_zz_o0_t3_m0_l0 = (-0.3333333333333333) * rk3kappa_cc * vector(rk3Alpha, i, j, k) * vector(rk3gtd_zz, i, j, k) * rk3trAt + vector(rk3Alpha, i, j, k) * (vector(rk3trK, i, j, k) * vector(rk3Atd_zz, i, j, k) + (-2.0 * vector(rk3Atd_xz, i, j, k) * rk3Atud_xz) + (-2.0 * vector(rk3Atd_yz, i, j, k) * rk3Atud_yz) + (-2.0 * vector(rk3Atd_zz, i, j, k) * rk3Atud_zz)) + rk3Psi1TF_zz + (-0.6666666666666667) * vector(rk3Atd_zz, i, j, k) * rk3div_Beta;
						d_Gamh_x_o0_t24_m0_l0 = (-2.0) * vector(rk3Alpha, i, j, k) * rk3inv_chi * rk3Zu_x * (rk3kappa_z1 + 1.333333333333333 * vector(rk3theta, i, j, k) + 0.6666666666666667 * vector(rk3trK, i, j, k)) + 0.3333333333333333 * rk3gtu_xz * rk3d_div_Beta_z + 0.3333333333333333 * rk3gtu_xy * rk3d_div_Beta_y + 0.3333333333333333 * rk3gtu_xx * rk3d_div_Beta_x + (-50.26548245743669) * vector(rk3Alpha, i, j, k) * rk3inv_chi * (rk3gtu_xx * rk3Jtd_ADM_x + rk3gtu_xy * rk3Jtd_ADM_y + rk3gtu_xz * rk3Jtd_ADM_z) + 2.0 * vector(rk3Alpha, i, j, k) * (rk3Ct_xxx * rk3Atu_xx + 2.0 * rk3Ct_xxy * rk3Atu_xy + 2.0 * rk3Ct_xxz * rk3Atu_xz + rk3Ct_xyy * rk3Atu_yy + 2.0 * rk3Ct_xyz * rk3Atu_yz + rk3Ct_xzz * rk3Atu_zz) + 0.6666666666666667 * vector(rk3Gamh_x, i, j, k) * rk3div_Beta;
						d_Gamh_y_o0_t24_m0_l0 = (-2.0) * vector(rk3Alpha, i, j, k) * rk3inv_chi * rk3Zu_y * (rk3kappa_z1 + 1.333333333333333 * vector(rk3theta, i, j, k) + 0.6666666666666667 * vector(rk3trK, i, j, k)) + (-50.26548245743669) * vector(rk3Alpha, i, j, k) * rk3inv_chi * (rk3gtu_xy * rk3Jtd_ADM_x + rk3gtu_yy * rk3Jtd_ADM_y + rk3gtu_yz * rk3Jtd_ADM_z) + 0.3333333333333333 * rk3gtu_yy * rk3d_div_Beta_y + 0.3333333333333333 * rk3gtu_xy * rk3d_div_Beta_x + 2.0 * vector(rk3Alpha, i, j, k) * (rk3Ct_yxx * rk3Atu_xx + 2.0 * rk3Ct_yxy * rk3Atu_xy + 2.0 * rk3Ct_yxz * rk3Atu_xz + rk3Ct_yyy * rk3Atu_yy + 2.0 * rk3Ct_yyz * rk3Atu_yz + rk3Ct_yzz * rk3Atu_zz) + 0.6666666666666667 * vector(rk3Gamh_y, i, j, k) * rk3div_Beta + 0.3333333333333333 * rk3gtu_yz * rk3d_div_Beta_z;
						d_Gamh_z_o0_t24_m0_l0 = (-2.0) * vector(rk3Alpha, i, j, k) * rk3inv_chi * rk3Zu_z * (rk3kappa_z1 + 1.333333333333333 * vector(rk3theta, i, j, k) + 0.6666666666666667 * vector(rk3trK, i, j, k)) + (-50.26548245743669) * vector(rk3Alpha, i, j, k) * rk3inv_chi * (rk3gtu_xz * rk3Jtd_ADM_x + rk3gtu_yz * rk3Jtd_ADM_y + rk3gtu_zz * rk3Jtd_ADM_z) + 2.0 * vector(rk3Alpha, i, j, k) * (rk3Ct_zxx * rk3Atu_xx + 2.0 * rk3Ct_zxy * rk3Atu_xy + 2.0 * rk3Ct_zxz * rk3Atu_xz + rk3Ct_zyy * rk3Atu_yy + 2.0 * rk3Ct_zyz * rk3Atu_yz + rk3Ct_zzz * rk3Atu_zz) + 0.6666666666666667 * vector(rk3Gamh_z, i, j, k) * rk3div_Beta + 0.3333333333333333 * rk3gtu_zz * rk3d_div_Beta_z + 0.3333333333333333 * rk3gtu_yz * rk3d_div_Beta_y + 0.3333333333333333 * rk3gtu_xz * rk3d_div_Beta_x;
						d_Betau_x_o0_t0_m0_l0 = (Betau_x_0 - vector(rk3Betau_x, i, j, k)) * rk3feta + 0.75 * (vector(rk3Alpha, i, j, k) * lambda_f1 + lambda_f0) * vector(rk3Gamh_x, i, j, k);
						d_Betau_y_o0_t0_m0_l0 = (Betau_y_0 - vector(rk3Betau_y, i, j, k)) * rk3feta + 0.75 * (vector(rk3Alpha, i, j, k) * lambda_f1 + lambda_f0) * vector(rk3Gamh_y, i, j, k);
						d_Betau_z_o0_t0_m0_l0 = (Betau_z_0 - vector(rk3Betau_z, i, j, k)) * rk3feta + 0.75 * (vector(rk3Alpha, i, j, k) * lambda_f1 + lambda_f0) * vector(rk3Gamh_z, i, j, k);
						d_Alpha_o0_t0_m0_l0 = (-2.0) * (vector(rk3Alpha, i, j, k) * lambda_f3 + lambda_f2) * vector(rk3Alpha, i, j, k) * (vector(rk3trK, i, j, k) - trK0);
						d_chi_o0_t0_m0_l0 = 0.6666666666666667 * rk3chi_max * (vector(rk3Alpha, i, j, k) * (2.0 * vector(rk3theta, i, j, k) + vector(rk3trK, i, j, k)) - rk3div_Beta);
						d_trK_o0_t21_m0_l0 = vector(rk3Alpha, i, j, k) * rk3kappa_z1 * (1.0 - rk3kappa_z2) * vector(rk3theta, i, j, k) + 0.3333333333333333 * vector(rk3Alpha, i, j, k) * ((2.0 * vector(rk3theta, i, j, k) + vector(rk3trK, i, j, k)) * (2.0 * vector(rk3theta, i, j, k) + vector(rk3trK, i, j, k))) + vector(rk3Alpha, i, j, k) * (vector(rk3Atd_xx, i, j, k) * rk3Atu_xx + 2.0 * vector(rk3Atd_xy, i, j, k) * rk3Atu_xy + 2.0 * vector(rk3Atd_xz, i, j, k) * rk3Atu_xz + vector(rk3Atd_yy, i, j, k) * rk3Atu_yy + 2.0 * vector(rk3Atd_yz, i, j, k) * rk3Atu_yz + vector(rk3Atd_zz, i, j, k) * rk3Atu_zz) + 4.0 * vector(rk3Alpha, i, j, k) * M_PI * (rk3rho_ADM + rk3tr_pT);
						d_theta_o0_t3_m0_l0 = (-vector(rk3Alpha, i, j, k) * rk3kappa_z1 * (2.0 + rk3kappa_z2) * vector(rk3theta, i, j, k)) + (-8.0) * vector(rk3Alpha, i, j, k) * M_PI * rk3rho_ADM + 0.5 * vector(rk3Alpha, i, j, k) * (rk3Rscalar + 0.6666666666666667 * (vector(rk3trK, i, j, k) * vector(rk3trK, i, j, k)) + 0.6666666666666667 * vector(rk3theta, i, j, k) * ((-2.0 * vector(rk3theta, i, j, k)) + vector(rk3trK, i, j, k)) + (-vector(rk3Atd_xx, i, j, k) * rk3Atu_xx) + (-2.0 * vector(rk3Atd_xy, i, j, k) * rk3Atu_xy) + (-2.0 * vector(rk3Atd_xz, i, j, k) * rk3Atu_xz) + (-vector(rk3Atd_yy, i, j, k) * rk3Atu_yy) + (-2.0 * vector(rk3Atd_yz, i, j, k) * rk3Atu_yz) + (-vector(rk3Atd_zz, i, j, k) * rk3Atu_zz));
						RHS_Bfu_x = -vector(rk3phif, i, j, k) * vector(rk3Alpha, i, j, k) * rk3chi_max * rk3Gamt_x;
						RHS_Bfu_y = -vector(rk3phif, i, j, k) * vector(rk3Alpha, i, j, k) * rk3chi_max * rk3Gamt_y;
						RHS_Bfu_z = -vector(rk3phif, i, j, k) * vector(rk3Alpha, i, j, k) * rk3chi_max * rk3Gamt_z;
						RHS_tauf = (vector(rk3Alpha, i, j, k) * (rk3Sfuu_xx * vector(rk3Atd_xx, i, j, k) + 2.0 * rk3Sfuu_xy * vector(rk3Atd_xy, i, j, k) + 2.0 * rk3Sfuu_xz * vector(rk3Atd_xz, i, j, k) + rk3Sfuu_yy * vector(rk3Atd_yy, i, j, k) + 2.0 * rk3Sfuu_yz * vector(rk3Atd_yz, i, j, k) + rk3Sfuu_zz * vector(rk3Atd_zz, i, j, k))) / rk3chi_max + 0.3333333333333333 * vector(rk3Alpha, i, j, k) * rk3trSf * vector(rk3trK, i, j, k);
						RHS_phif = (-vector(rk3Alpha, i, j, k) * rk3kappa_f * vector(rk3phif, i, j, k)) + (-vector(rk3Alpha, i, j, k) * vector(rk3phif, i, j, k) * vector(rk3trK, i, j, k));
						RHS_Sfd_x = -FVM_i(FluxpiSfd_x, FluxniSfd_x, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_x = RHS_Sfd_x - FVM_j(FluxpjSfd_x, FluxnjSfd_x, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_x = RHS_Sfd_x - FVM_k(FluxpkSfd_x, FluxnkSfd_x, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_y = -FVM_i(FluxpiSfd_y, FluxniSfd_y, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_y = RHS_Sfd_y - FVM_j(FluxpjSfd_y, FluxnjSfd_y, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_y = RHS_Sfd_y - FVM_k(FluxpkSfd_y, FluxnkSfd_y, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_z = -FVM_i(FluxpiSfd_z, FluxniSfd_z, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_z = RHS_Sfd_z - FVM_j(FluxpjSfd_z, FluxnjSfd_z, dx, simPlat_dt, ilast, jlast);
						RHS_Sfd_z = RHS_Sfd_z - FVM_k(FluxpkSfd_z, FluxnkSfd_z, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_x = RHS_Bfu_x - FVM_i(FluxpiBfu_x, FluxniBfu_x, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_x = RHS_Bfu_x - FVM_j(FluxpjBfu_x, FluxnjBfu_x, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_x = RHS_Bfu_x - FVM_k(FluxpkBfu_x, FluxnkBfu_x, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_y = RHS_Bfu_y - FVM_i(FluxpiBfu_y, FluxniBfu_y, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_y = RHS_Bfu_y - FVM_j(FluxpjBfu_y, FluxnjBfu_y, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_y = RHS_Bfu_y - FVM_k(FluxpkBfu_y, FluxnkBfu_y, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_z = RHS_Bfu_z - FVM_i(FluxpiBfu_z, FluxniBfu_z, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_z = RHS_Bfu_z - FVM_j(FluxpjBfu_z, FluxnjBfu_z, dx, simPlat_dt, ilast, jlast);
						RHS_Bfu_z = RHS_Bfu_z - FVM_k(FluxpkBfu_z, FluxnkBfu_z, dx, simPlat_dt, ilast, jlast);
						RHS_Df = -FVM_i(FluxpiDf, FluxniDf, dx, simPlat_dt, ilast, jlast);
						RHS_Df = RHS_Df - FVM_j(FluxpjDf, FluxnjDf, dx, simPlat_dt, ilast, jlast);
						RHS_Df = RHS_Df - FVM_k(FluxpkDf, FluxnkDf, dx, simPlat_dt, ilast, jlast);
						RHS_DYf = -FVM_i(FluxpiDYf, FluxniDYf, dx, simPlat_dt, ilast, jlast);
						RHS_DYf = RHS_DYf - FVM_j(FluxpjDYf, FluxnjDYf, dx, simPlat_dt, ilast, jlast);
						RHS_DYf = RHS_DYf - FVM_k(FluxpkDYf, FluxnkDYf, dx, simPlat_dt, ilast, jlast);
						RHS_tauf = RHS_tauf - FVM_i(Fluxpitauf, Fluxnitauf, dx, simPlat_dt, ilast, jlast);
						RHS_tauf = RHS_tauf - FVM_j(Fluxpjtauf, Fluxnjtauf, dx, simPlat_dt, ilast, jlast);
						RHS_tauf = RHS_tauf - FVM_k(Fluxpktauf, Fluxnktauf, dx, simPlat_dt, ilast, jlast);
						RHS_phif = RHS_phif - FVM_i(Fluxpiphif, Fluxniphif, dx, simPlat_dt, ilast, jlast);
						RHS_phif = RHS_phif - FVM_j(Fluxpjphif, Fluxnjphif, dx, simPlat_dt, ilast, jlast);
						RHS_phif = RHS_phif - FVM_k(Fluxpkphif, Fluxnkphif, dx, simPlat_dt, ilast, jlast);
						m_theta_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_theta_o2_t2_m0_l0;
						m_theta_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_theta_o2_t1_m0_l0;
						m_theta_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_theta_o2_t0_m0_l0;
						m_theta_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_theta_o1_t2_m0_l0;
						m_theta_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_theta_o1_t1_m0_l0;
						m_theta_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_theta_o1_t0_m0_l0;
						m_theta_o0_t2_l0 = (-rk3Zu_z) * d_Psi1_xx_o0_t3_m0_l0;
						m_theta_o0_t1_l0 = (-rk3Zu_y) * d_Psi1_xx_o0_t2_m0_l0;
						m_theta_o0_t0_l0 = (-rk3Zu_x) * d_Psi1_xx_o0_t1_m0_l0;
						m_trK_o2_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_trK_o2_t2_m0_l0;
						m_trK_o2_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_trK_o2_t1_m0_l0;
						m_trK_o2_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_trK_o2_t0_m0_l0;
						m_trK_o1_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_trK_o1_t2_m0_l0;
						m_trK_o1_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_trK_o1_t1_m0_l0;
						m_trK_o1_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_trK_o1_t0_m0_l0;
						m_trK_o0_t20_l0 = 2.0 * rk3Zu_z * d_Psi1_xx_o0_t3_m0_l0;
						m_trK_o0_t19_l0 = 2.0 * rk3Zu_y * d_Psi1_xx_o0_t2_m0_l0;
						m_trK_o0_t18_l0 = 2.0 * rk3Zu_x * d_Psi1_xx_o0_t1_m0_l0;
						m_trK_o0_t17_l0 = (-2.0 * rk3chi_max * rk3gtu_xz) * d_Psi1_xz_o0_t0_m0_l0;
						m_trK_o0_t16_l0 = (-2.0 * rk3chi_max * rk3gtu_xy) * d_Psi1_xy_o0_t0_m0_l0;
						m_trK_o0_t15_l0 = 0.5 * rk3gtu_zz * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_trK_o0_t14_l0 = 0.5 * rk3gtu_yz * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_trK_o0_t13_l0 = 0.5 * rk3gtu_xz * d_Psi1_xx_o0_t3_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_trK_o0_t12_l0 = 0.5 * rk3gtu_yz * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_trK_o0_t11_l0 = 0.5 * rk3gtu_yy * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_trK_o0_t10_l0 = 0.5 * rk3gtu_xy * d_Psi1_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_trK_o0_t9_l0 = 0.5 * rk3gtu_xz * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
						m_trK_o0_t8_l0 = 0.5 * rk3gtu_xy * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
						m_trK_o0_t7_l0 = 0.5 * rk3gtu_xx * d_Psi1_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
						m_trK_o0_t6_l0 = rk3chi_max * rk3Gamt_z * d_Psi1_xx_o0_t3_m0_l0;
						m_trK_o0_t5_l0 = rk3chi_max * rk3Gamt_y * d_Psi1_xx_o0_t2_m0_l0;
						m_trK_o0_t4_l0 = rk3chi_max * rk3Gamt_x * d_Psi1_xx_o0_t1_m0_l0;
						m_trK_o0_t3_l0 = (-rk3chi_max * rk3gtu_zz) * d_Psi1_zz_o0_t0_m0_l0;
						m_trK_o0_t2_l0 = (-2.0 * rk3chi_max * rk3gtu_yz) * d_Psi1_yz_o0_t0_m0_l0;
						m_trK_o0_t1_l0 = (-rk3chi_max * rk3gtu_yy) * d_Psi1_yy_o0_t0_m0_l0;
						m_trK_o0_t0_l0 = (-rk3chi_max * rk3gtu_xx) * d_Psi1_xx_o0_t0_m0_l0;
						m_chi_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_chi_o2_t2_m0_l0;
						m_chi_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_chi_o2_t1_m0_l0;
						m_chi_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_chi_o2_t0_m0_l0;
						m_chi_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_chi_o1_t2_m0_l0;
						m_chi_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_chi_o1_t1_m0_l0;
						m_chi_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_chi_o1_t0_m0_l0;
						m_Alpha_o2_t2_l0 = lambda_1 * MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Alpha_o2_t2_m0_l0;
						m_Alpha_o2_t1_l0 = lambda_1 * MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Alpha_o2_t1_m0_l0;
						m_Alpha_o2_t0_l0 = lambda_1 * MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Alpha_o2_t0_m0_l0;
						m_Alpha_o1_t2_l0 = lambda_1 * MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Alpha_o1_t2_m0_l0;
						m_Alpha_o1_t1_l0 = lambda_1 * MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Alpha_o1_t1_m0_l0;
						m_Alpha_o1_t0_l0 = lambda_1 * MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Alpha_o1_t0_m0_l0;
						m_Betau_z_o2_t2_l0 = lambda_2 * MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Betau_z_o2_t2_m0_l0;
						m_Betau_z_o2_t1_l0 = lambda_2 * MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Betau_z_o2_t1_m0_l0;
						m_Betau_z_o2_t0_l0 = lambda_2 * MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Betau_z_o2_t0_m0_l0;
						m_Betau_z_o1_t2_l0 = lambda_2 * MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Betau_z_o1_t2_m0_l0;
						m_Betau_z_o1_t1_l0 = lambda_2 * MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Betau_z_o1_t1_m0_l0;
						m_Betau_z_o1_t0_l0 = lambda_2 * MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Betau_z_o1_t0_m0_l0;
						m_Betau_y_o2_t2_l0 = lambda_2 * MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Betau_y_o2_t2_m0_l0;
						m_Betau_y_o2_t1_l0 = lambda_2 * MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Betau_y_o2_t1_m0_l0;
						m_Betau_y_o2_t0_l0 = lambda_2 * MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Betau_y_o2_t0_m0_l0;
						m_Betau_y_o1_t2_l0 = lambda_2 * MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Betau_y_o1_t2_m0_l0;
						m_Betau_y_o1_t1_l0 = lambda_2 * MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Betau_y_o1_t1_m0_l0;
						m_Betau_y_o1_t0_l0 = lambda_2 * MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Betau_y_o1_t0_m0_l0;
						m_Betau_x_o2_t2_l0 = lambda_2 * MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Betau_x_o2_t2_m0_l0;
						m_Betau_x_o2_t1_l0 = lambda_2 * MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Betau_x_o2_t1_m0_l0;
						m_Betau_x_o2_t0_l0 = lambda_2 * MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Betau_x_o2_t0_m0_l0;
						m_Betau_x_o1_t2_l0 = lambda_2 * MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Betau_x_o1_t2_m0_l0;
						m_Betau_x_o1_t1_l0 = lambda_2 * MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Betau_x_o1_t1_m0_l0;
						m_Betau_x_o1_t0_l0 = lambda_2 * MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Betau_x_o1_t0_m0_l0;
						m_Gamh_z_o2_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Gamh_z_o2_t2_m0_l0;
						m_Gamh_z_o2_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Gamh_z_o2_t1_m0_l0;
						m_Gamh_z_o2_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Gamh_z_o2_t0_m0_l0;
						m_Gamh_z_o1_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Gamh_z_o1_t2_m0_l0;
						m_Gamh_z_o1_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Gamh_z_o1_t1_m0_l0;
						m_Gamh_z_o1_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Gamh_z_o1_t0_m0_l0;
						m_Gamh_z_o0_t23_l0 = (-3.0 * vector(rk3Alpha, i, j, k) * rk3inv_chi * rk3Atu_zz) * d_Rpd_xx_o0_t2_m0_l0;
						m_Gamh_z_o0_t22_l0 = (-3.0 * vector(rk3Alpha, i, j, k) * rk3inv_chi * rk3Atu_yz) * d_Rpd_xx_o0_t1_m0_l0;
						m_Gamh_z_o0_t21_l0 = (-3.0 * vector(rk3Alpha, i, j, k) * rk3inv_chi * rk3Atu_xz) * d_Rpd_xx_o0_t0_m0_l0;
						m_Gamh_z_o0_t20_l0 = (-2.0 * vector(rk3theta, i, j, k) * rk3gtu_zz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_z_o0_t19_l0 = (-2.0 * vector(rk3theta, i, j, k) * rk3gtu_yz) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_z_o0_t18_l0 = (-2.0 * vector(rk3theta, i, j, k) * rk3gtu_xz) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_z_o0_t17_l0 = (-0.6666666666666667 * vector(rk3Alpha, i, j, k) * rk3gtu_zz) * d_Gamh_x_o0_t5_m0_l0;
						m_Gamh_z_o0_t16_l0 = (-0.6666666666666667 * vector(rk3Alpha, i, j, k) * rk3gtu_yz) * d_Gamh_x_o0_t4_m0_l0;
						m_Gamh_z_o0_t15_l0 = (-0.6666666666666667 * vector(rk3Alpha, i, j, k) * rk3gtu_xz) * d_Gamh_x_o0_t3_m0_l0;
						m_Gamh_z_o0_t14_l0 = (-1.333333333333333 * vector(rk3Alpha, i, j, k) * rk3gtu_zz) * d_Gamh_x_o0_t2_m0_l0;
						m_Gamh_z_o0_t13_l0 = (-1.333333333333333 * vector(rk3Alpha, i, j, k) * rk3gtu_yz) * d_Gamh_x_o0_t1_m0_l0;
						m_Gamh_z_o0_t12_l0 = (-1.333333333333333 * vector(rk3Alpha, i, j, k) * rk3gtu_xz) * d_Gamh_x_o0_t0_m0_l0;
						m_Gamh_z_o0_t11_l0 = (-vector(rk3Gamh_z, i, j, k)) * d_div_Beta_o0_t2_m0_l0;
						m_Gamh_z_o0_t10_l0 = (-vector(rk3Gamh_y, i, j, k)) * d_Sfd_y_o0_t9_m0_l0;
						m_Gamh_z_o0_t9_l0 = (-vector(rk3Gamh_x, i, j, k)) * d_Sfd_x_o0_t9_m0_l0;
						m_Gamh_z_o0_t8_l0 = (-2.0 * rk3Atu_zz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_z_o0_t7_l0 = (-2.0 * rk3Atu_yz) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_z_o0_t6_l0 = (-2.0 * rk3Atu_xz) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_z_o0_t5_l0 = rk3gtu_zz * d_d_div_Beta_z_o0_t2_m0_l0;
						m_Gamh_z_o0_t4_l0 = 2.0 * rk3gtu_yz * d_d_div_Beta_y_o0_t2_m0_l0;
						m_Gamh_z_o0_t3_l0 = rk3gtu_yy * d_Gamh_z_o0_t3_m0_l0;
						m_Gamh_z_o0_t2_l0 = 2.0 * rk3gtu_xz * d_d_div_Beta_x_o0_t2_m0_l0;
						m_Gamh_z_o0_t1_l0 = 2.0 * rk3gtu_xy * d_Gamh_z_o0_t1_m0_l0;
						m_Gamh_z_o0_t0_l0 = rk3gtu_xx * d_Gamh_z_o0_t0_m0_l0;
						m_Gamh_y_o2_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Gamh_y_o2_t2_m0_l0;
						m_Gamh_y_o2_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Gamh_y_o2_t1_m0_l0;
						m_Gamh_y_o2_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Gamh_y_o2_t0_m0_l0;
						m_Gamh_y_o1_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Gamh_y_o1_t2_m0_l0;
						m_Gamh_y_o1_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Gamh_y_o1_t1_m0_l0;
						m_Gamh_y_o1_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Gamh_y_o1_t0_m0_l0;
						m_Gamh_y_o0_t23_l0 = (-3.0 * vector(rk3Alpha, i, j, k) * rk3inv_chi * rk3Atu_yz) * d_Rpd_xx_o0_t2_m0_l0;
						m_Gamh_y_o0_t22_l0 = (-3.0 * vector(rk3Alpha, i, j, k) * rk3inv_chi * rk3Atu_yy) * d_Rpd_xx_o0_t1_m0_l0;
						m_Gamh_y_o0_t21_l0 = (-3.0 * vector(rk3Alpha, i, j, k) * rk3inv_chi * rk3Atu_xy) * d_Rpd_xx_o0_t0_m0_l0;
						m_Gamh_y_o0_t20_l0 = (-2.0 * vector(rk3theta, i, j, k) * rk3gtu_yz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_y_o0_t19_l0 = (-2.0 * vector(rk3theta, i, j, k) * rk3gtu_yy) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_y_o0_t18_l0 = (-2.0 * vector(rk3theta, i, j, k) * rk3gtu_xy) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_y_o0_t17_l0 = (-0.6666666666666667 * vector(rk3Alpha, i, j, k) * rk3gtu_yz) * d_Gamh_x_o0_t5_m0_l0;
						m_Gamh_y_o0_t16_l0 = (-0.6666666666666667 * vector(rk3Alpha, i, j, k) * rk3gtu_yy) * d_Gamh_x_o0_t4_m0_l0;
						m_Gamh_y_o0_t15_l0 = (-0.6666666666666667 * vector(rk3Alpha, i, j, k) * rk3gtu_xy) * d_Gamh_x_o0_t3_m0_l0;
						m_Gamh_y_o0_t14_l0 = (-1.333333333333333 * vector(rk3Alpha, i, j, k) * rk3gtu_yz) * d_Gamh_x_o0_t2_m0_l0;
						m_Gamh_y_o0_t13_l0 = (-1.333333333333333 * vector(rk3Alpha, i, j, k) * rk3gtu_yy) * d_Gamh_x_o0_t1_m0_l0;
						m_Gamh_y_o0_t12_l0 = (-1.333333333333333 * vector(rk3Alpha, i, j, k) * rk3gtu_xy) * d_Gamh_x_o0_t0_m0_l0;
						m_Gamh_y_o0_t11_l0 = rk3gtu_zz * d_Gamh_y_o0_t11_m0_l0;
						m_Gamh_y_o0_t10_l0 = 2.0 * rk3gtu_yz * d_d_div_Beta_z_o0_t1_m0_l0;
						m_Gamh_y_o0_t9_l0 = rk3gtu_yy * d_d_div_Beta_y_o0_t1_m0_l0;
						m_Gamh_y_o0_t8_l0 = 2.0 * rk3gtu_xz * d_Gamh_y_o0_t8_m0_l0;
						m_Gamh_y_o0_t7_l0 = 2.0 * rk3gtu_xy * d_d_div_Beta_x_o0_t1_m0_l0;
						m_Gamh_y_o0_t6_l0 = rk3gtu_xx * d_Gamh_y_o0_t6_m0_l0;
						m_Gamh_y_o0_t5_l0 = (-vector(rk3Gamh_z, i, j, k)) * d_Sfd_z_o0_t8_m0_l0;
						m_Gamh_y_o0_t4_l0 = (-vector(rk3Gamh_y, i, j, k)) * d_div_Beta_o0_t1_m0_l0;
						m_Gamh_y_o0_t3_l0 = (-vector(rk3Gamh_x, i, j, k)) * d_Sfd_x_o0_t8_m0_l0;
						m_Gamh_y_o0_t2_l0 = (-2.0 * rk3Atu_yz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_y_o0_t1_l0 = (-2.0 * rk3Atu_yy) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_y_o0_t0_l0 = (-2.0 * rk3Atu_xy) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_x_o2_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Gamh_x_o2_t2_m0_l0;
						m_Gamh_x_o2_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Gamh_x_o2_t1_m0_l0;
						m_Gamh_x_o2_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Gamh_x_o2_t0_m0_l0;
						m_Gamh_x_o1_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Gamh_x_o1_t2_m0_l0;
						m_Gamh_x_o1_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Gamh_x_o1_t1_m0_l0;
						m_Gamh_x_o1_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Gamh_x_o1_t0_m0_l0;
						m_Gamh_x_o0_t23_l0 = (-3.0 * vector(rk3Alpha, i, j, k) * rk3inv_chi * rk3Atu_xz) * d_Rpd_xx_o0_t2_m0_l0;
						m_Gamh_x_o0_t22_l0 = (-3.0 * vector(rk3Alpha, i, j, k) * rk3inv_chi * rk3Atu_xy) * d_Rpd_xx_o0_t1_m0_l0;
						m_Gamh_x_o0_t21_l0 = (-3.0 * vector(rk3Alpha, i, j, k) * rk3inv_chi * rk3Atu_xx) * d_Rpd_xx_o0_t0_m0_l0;
						m_Gamh_x_o0_t20_l0 = (-2.0 * rk3Atu_xz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_x_o0_t19_l0 = (-2.0 * rk3Atu_xy) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_x_o0_t18_l0 = (-2.0 * rk3Atu_xx) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_x_o0_t17_l0 = rk3gtu_zz * d_Gamh_x_o0_t17_m0_l0;
						m_Gamh_x_o0_t16_l0 = 2.0 * rk3gtu_yz * d_Gamh_x_o0_t16_m0_l0;
						m_Gamh_x_o0_t15_l0 = rk3gtu_yy * d_Gamh_x_o0_t15_m0_l0;
						m_Gamh_x_o0_t14_l0 = 2.0 * rk3gtu_xz * d_d_div_Beta_z_o0_t0_m0_l0;
						m_Gamh_x_o0_t13_l0 = 2.0 * rk3gtu_xy * d_d_div_Beta_y_o0_t0_m0_l0;
						m_Gamh_x_o0_t12_l0 = rk3gtu_xx * d_d_div_Beta_x_o0_t0_m0_l0;
						m_Gamh_x_o0_t11_l0 = (-vector(rk3Gamh_z, i, j, k)) * d_Sfd_z_o0_t7_m0_l0;
						m_Gamh_x_o0_t10_l0 = (-vector(rk3Gamh_y, i, j, k)) * d_Sfd_y_o0_t7_m0_l0;
						m_Gamh_x_o0_t9_l0 = (-vector(rk3Gamh_x, i, j, k)) * d_div_Beta_o0_t0_m0_l0;
						m_Gamh_x_o0_t8_l0 = (-2.0 * vector(rk3theta, i, j, k) * rk3gtu_xz) * d_Psi1_xx_o0_t3_m0_l0;
						m_Gamh_x_o0_t7_l0 = (-2.0 * vector(rk3theta, i, j, k) * rk3gtu_xy) * d_Psi1_xx_o0_t2_m0_l0;
						m_Gamh_x_o0_t6_l0 = (-2.0 * vector(rk3theta, i, j, k) * rk3gtu_xx) * d_Psi1_xx_o0_t1_m0_l0;
						m_Gamh_x_o0_t5_l0 = (-0.6666666666666667 * vector(rk3Alpha, i, j, k) * rk3gtu_xz) * d_Gamh_x_o0_t5_m0_l0;
						m_Gamh_x_o0_t4_l0 = (-0.6666666666666667 * vector(rk3Alpha, i, j, k) * rk3gtu_xy) * d_Gamh_x_o0_t4_m0_l0;
						m_Gamh_x_o0_t3_l0 = (-0.6666666666666667 * vector(rk3Alpha, i, j, k) * rk3gtu_xx) * d_Gamh_x_o0_t3_m0_l0;
						m_Gamh_x_o0_t2_l0 = (-1.333333333333333 * vector(rk3Alpha, i, j, k) * rk3gtu_xz) * d_Gamh_x_o0_t2_m0_l0;
						m_Gamh_x_o0_t1_l0 = (-1.333333333333333 * vector(rk3Alpha, i, j, k) * rk3gtu_xy) * d_Gamh_x_o0_t1_m0_l0;
						m_Gamh_x_o0_t0_l0 = (-1.333333333333333 * vector(rk3Alpha, i, j, k) * rk3gtu_xx) * d_Gamh_x_o0_t0_m0_l0;
						m_Atd_zz_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Atd_zz_o2_t2_m0_l0;
						m_Atd_zz_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Atd_zz_o2_t1_m0_l0;
						m_Atd_zz_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Atd_zz_o2_t0_m0_l0;
						m_Atd_zz_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Atd_zz_o1_t2_m0_l0;
						m_Atd_zz_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Atd_zz_o1_t1_m0_l0;
						m_Atd_zz_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Atd_zz_o1_t0_m0_l0;
						m_Atd_zz_o0_t2_l0 = 2.0 * vector(rk3Atd_zz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Atd_zz_o0_t1_l0 = 2.0 * vector(rk3Atd_yz, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Atd_zz_o0_t0_l0 = 2.0 * vector(rk3Atd_xz, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Atd_yz_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Atd_yz_o2_t2_m0_l0;
						m_Atd_yz_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Atd_yz_o2_t1_m0_l0;
						m_Atd_yz_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Atd_yz_o2_t0_m0_l0;
						m_Atd_yz_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Atd_yz_o1_t2_m0_l0;
						m_Atd_yz_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Atd_yz_o1_t1_m0_l0;
						m_Atd_yz_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Atd_yz_o1_t0_m0_l0;
						m_Atd_yz_o0_t5_l0 = vector(rk3Atd_zz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Atd_yz_o0_t4_l0 = vector(rk3Atd_yz, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Atd_yz_o0_t3_l0 = vector(rk3Atd_xz, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Atd_yz_o0_t2_l0 = vector(rk3Atd_yz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Atd_yz_o0_t1_l0 = vector(rk3Atd_yy, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Atd_yz_o0_t0_l0 = vector(rk3Atd_xy, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Atd_yy_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Atd_yy_o2_t2_m0_l0;
						m_Atd_yy_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Atd_yy_o2_t1_m0_l0;
						m_Atd_yy_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Atd_yy_o2_t0_m0_l0;
						m_Atd_yy_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Atd_yy_o1_t2_m0_l0;
						m_Atd_yy_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Atd_yy_o1_t1_m0_l0;
						m_Atd_yy_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Atd_yy_o1_t0_m0_l0;
						m_Atd_yy_o0_t2_l0 = 2.0 * vector(rk3Atd_yz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Atd_yy_o0_t1_l0 = 2.0 * vector(rk3Atd_yy, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Atd_yy_o0_t0_l0 = 2.0 * vector(rk3Atd_xy, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Atd_xz_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Atd_xz_o2_t2_m0_l0;
						m_Atd_xz_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Atd_xz_o2_t1_m0_l0;
						m_Atd_xz_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Atd_xz_o2_t0_m0_l0;
						m_Atd_xz_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Atd_xz_o1_t2_m0_l0;
						m_Atd_xz_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Atd_xz_o1_t1_m0_l0;
						m_Atd_xz_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Atd_xz_o1_t0_m0_l0;
						m_Atd_xz_o0_t5_l0 = vector(rk3Atd_zz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Atd_xz_o0_t4_l0 = vector(rk3Atd_yz, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Atd_xz_o0_t3_l0 = vector(rk3Atd_xz, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_Atd_xz_o0_t2_l0 = vector(rk3Atd_xz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Atd_xz_o0_t1_l0 = vector(rk3Atd_xy, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Atd_xz_o0_t0_l0 = vector(rk3Atd_xx, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Atd_xy_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Atd_xy_o2_t2_m0_l0;
						m_Atd_xy_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Atd_xy_o2_t1_m0_l0;
						m_Atd_xy_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Atd_xy_o2_t0_m0_l0;
						m_Atd_xy_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Atd_xy_o1_t2_m0_l0;
						m_Atd_xy_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Atd_xy_o1_t1_m0_l0;
						m_Atd_xy_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Atd_xy_o1_t0_m0_l0;
						m_Atd_xy_o0_t5_l0 = vector(rk3Atd_yz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Atd_xy_o0_t4_l0 = vector(rk3Atd_yy, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Atd_xy_o0_t3_l0 = vector(rk3Atd_xy, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_Atd_xy_o0_t2_l0 = vector(rk3Atd_xz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Atd_xy_o0_t1_l0 = vector(rk3Atd_xy, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Atd_xy_o0_t0_l0 = vector(rk3Atd_xx, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Atd_xx_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_Atd_xx_o2_t2_m0_l0;
						m_Atd_xx_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_Atd_xx_o2_t1_m0_l0;
						m_Atd_xx_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_Atd_xx_o2_t0_m0_l0;
						m_Atd_xx_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_Atd_xx_o1_t2_m0_l0;
						m_Atd_xx_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_Atd_xx_o1_t1_m0_l0;
						m_Atd_xx_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_Atd_xx_o1_t0_m0_l0;
						m_Atd_xx_o0_t2_l0 = 2.0 * vector(rk3Atd_xz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Atd_xx_o0_t1_l0 = 2.0 * vector(rk3Atd_xy, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Atd_xx_o0_t0_l0 = 2.0 * vector(rk3Atd_xx, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_gtd_zz_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_gtd_zz_o2_t2_m0_l0;
						m_gtd_zz_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_gtd_zz_o2_t1_m0_l0;
						m_gtd_zz_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_gtd_zz_o2_t0_m0_l0;
						m_gtd_zz_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_gtd_zz_o1_t2_m0_l0;
						m_gtd_zz_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_gtd_zz_o1_t1_m0_l0;
						m_gtd_zz_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_gtd_zz_o1_t0_m0_l0;
						m_gtd_zz_o0_t2_l0 = 2.0 * vector(rk3gtd_zz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_gtd_zz_o0_t1_l0 = 2.0 * vector(rk3gtd_yz, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_gtd_zz_o0_t0_l0 = 2.0 * vector(rk3gtd_xz, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_gtd_yz_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_gtd_yz_o2_t2_m0_l0;
						m_gtd_yz_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_gtd_yz_o2_t1_m0_l0;
						m_gtd_yz_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_gtd_yz_o2_t0_m0_l0;
						m_gtd_yz_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_gtd_yz_o1_t2_m0_l0;
						m_gtd_yz_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_gtd_yz_o1_t1_m0_l0;
						m_gtd_yz_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_gtd_yz_o1_t0_m0_l0;
						m_gtd_yz_o0_t5_l0 = vector(rk3gtd_zz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_gtd_yz_o0_t4_l0 = vector(rk3gtd_yz, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_gtd_yz_o0_t3_l0 = vector(rk3gtd_xz, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_gtd_yz_o0_t2_l0 = vector(rk3gtd_yz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_gtd_yz_o0_t1_l0 = vector(rk3gtd_yy, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_gtd_yz_o0_t0_l0 = vector(rk3gtd_xy, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_gtd_yy_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_gtd_yy_o2_t2_m0_l0;
						m_gtd_yy_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_gtd_yy_o2_t1_m0_l0;
						m_gtd_yy_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_gtd_yy_o2_t0_m0_l0;
						m_gtd_yy_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_gtd_yy_o1_t2_m0_l0;
						m_gtd_yy_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_gtd_yy_o1_t1_m0_l0;
						m_gtd_yy_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_gtd_yy_o1_t0_m0_l0;
						m_gtd_yy_o0_t2_l0 = 2.0 * vector(rk3gtd_yz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_gtd_yy_o0_t1_l0 = 2.0 * vector(rk3gtd_yy, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_gtd_yy_o0_t0_l0 = 2.0 * vector(rk3gtd_xy, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_gtd_xz_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_gtd_xz_o2_t2_m0_l0;
						m_gtd_xz_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_gtd_xz_o2_t1_m0_l0;
						m_gtd_xz_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_gtd_xz_o2_t0_m0_l0;
						m_gtd_xz_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_gtd_xz_o1_t2_m0_l0;
						m_gtd_xz_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_gtd_xz_o1_t1_m0_l0;
						m_gtd_xz_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_gtd_xz_o1_t0_m0_l0;
						m_gtd_xz_o0_t5_l0 = vector(rk3gtd_zz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_gtd_xz_o0_t4_l0 = vector(rk3gtd_yz, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_gtd_xz_o0_t3_l0 = vector(rk3gtd_xz, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_gtd_xz_o0_t2_l0 = vector(rk3gtd_xz, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_gtd_xz_o0_t1_l0 = vector(rk3gtd_xy, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_gtd_xz_o0_t0_l0 = vector(rk3gtd_xx, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_gtd_xy_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_gtd_xy_o2_t2_m0_l0;
						m_gtd_xy_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_gtd_xy_o2_t1_m0_l0;
						m_gtd_xy_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_gtd_xy_o2_t0_m0_l0;
						m_gtd_xy_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_gtd_xy_o1_t2_m0_l0;
						m_gtd_xy_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_gtd_xy_o1_t1_m0_l0;
						m_gtd_xy_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_gtd_xy_o1_t0_m0_l0;
						m_gtd_xy_o0_t5_l0 = vector(rk3gtd_yz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_gtd_xy_o0_t4_l0 = vector(rk3gtd_yy, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_gtd_xy_o0_t3_l0 = vector(rk3gtd_xy, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_gtd_xy_o0_t2_l0 = vector(rk3gtd_xz, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_gtd_xy_o0_t1_l0 = vector(rk3gtd_xy, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_gtd_xy_o0_t0_l0 = vector(rk3gtd_xx, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_gtd_xx_o2_t2_l0 = MIN(0.0, vector(rk3Betau_z, i, j, k)) * d_gtd_xx_o2_t2_m0_l0;
						m_gtd_xx_o2_t1_l0 = MIN(0.0, vector(rk3Betau_y, i, j, k)) * d_gtd_xx_o2_t1_m0_l0;
						m_gtd_xx_o2_t0_l0 = MIN(0.0, vector(rk3Betau_x, i, j, k)) * d_gtd_xx_o2_t0_m0_l0;
						m_gtd_xx_o1_t2_l0 = MAX(0.0, vector(rk3Betau_z, i, j, k)) * d_gtd_xx_o1_t2_m0_l0;
						m_gtd_xx_o1_t1_l0 = MAX(0.0, vector(rk3Betau_y, i, j, k)) * d_gtd_xx_o1_t1_m0_l0;
						m_gtd_xx_o1_t0_l0 = MAX(0.0, vector(rk3Betau_x, i, j, k)) * d_gtd_xx_o1_t0_m0_l0;
						m_gtd_xx_o0_t2_l0 = 2.0 * vector(rk3gtd_xz, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_gtd_xx_o0_t1_l0 = 2.0 * vector(rk3gtd_xy, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_gtd_xx_o0_t0_l0 = 2.0 * vector(rk3gtd_xx, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_phif_o0_t2_l0 = (ch * ch) * vector(rk3Bfu_z, i, j, k) * d_Psi1_xx_o0_t3_m0_l0;
						m_phif_o0_t1_l0 = (ch * ch) * vector(rk3Bfu_y, i, j, k) * d_Psi1_xx_o0_t2_m0_l0;
						m_phif_o0_t0_l0 = (ch * ch) * vector(rk3Bfu_x, i, j, k) * d_Psi1_xx_o0_t1_m0_l0;
						m_tauf_o0_t2_l0 = (-rk3Sfu_z) * d_Psi1_xx_o0_t3_m0_l0;
						m_tauf_o0_t1_l0 = (-rk3Sfu_y) * d_Psi1_xx_o0_t2_m0_l0;
						m_tauf_o0_t0_l0 = (-rk3Sfu_x) * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_z_o0_t5_l0 = vector(rk3phif, i, j, k) * rk3gtu_zz * rk3chi_max * d_Psi1_xx_o0_t3_m0_l0;
						m_Bfu_z_o0_t4_l0 = vector(rk3phif, i, j, k) * rk3gtu_yz * rk3chi_max * d_Psi1_xx_o0_t2_m0_l0;
						m_Bfu_z_o0_t3_l0 = vector(rk3phif, i, j, k) * rk3gtu_xz * rk3chi_max * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_z_o0_t2_l0 = (-0.5 * vector(rk3phif, i, j, k) * rk3gtu_zz * vector(rk3Alpha, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Bfu_z_o0_t1_l0 = (-0.5 * vector(rk3phif, i, j, k) * rk3gtu_yz * vector(rk3Alpha, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Bfu_z_o0_t0_l0 = (-0.5 * vector(rk3phif, i, j, k) * rk3gtu_xz * vector(rk3Alpha, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Bfu_y_o0_t5_l0 = vector(rk3phif, i, j, k) * rk3gtu_yz * rk3chi_max * d_Psi1_xx_o0_t3_m0_l0;
						m_Bfu_y_o0_t4_l0 = vector(rk3phif, i, j, k) * rk3gtu_yy * rk3chi_max * d_Psi1_xx_o0_t2_m0_l0;
						m_Bfu_y_o0_t3_l0 = vector(rk3phif, i, j, k) * rk3gtu_xy * rk3chi_max * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_y_o0_t2_l0 = (-0.5 * vector(rk3phif, i, j, k) * rk3gtu_yz * vector(rk3Alpha, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Bfu_y_o0_t1_l0 = (-0.5 * vector(rk3phif, i, j, k) * rk3gtu_yy * vector(rk3Alpha, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Bfu_y_o0_t0_l0 = (-0.5 * vector(rk3phif, i, j, k) * rk3gtu_xy * vector(rk3Alpha, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Bfu_x_o0_t5_l0 = vector(rk3phif, i, j, k) * rk3gtu_xz * rk3chi_max * d_Psi1_xx_o0_t3_m0_l0;
						m_Bfu_x_o0_t4_l0 = vector(rk3phif, i, j, k) * rk3gtu_xy * rk3chi_max * d_Psi1_xx_o0_t2_m0_l0;
						m_Bfu_x_o0_t3_l0 = vector(rk3phif, i, j, k) * rk3gtu_xx * rk3chi_max * d_Psi1_xx_o0_t1_m0_l0;
						m_Bfu_x_o0_t2_l0 = (-0.5 * vector(rk3phif, i, j, k) * rk3gtu_xz * vector(rk3Alpha, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
						m_Bfu_x_o0_t1_l0 = (-0.5 * vector(rk3phif, i, j, k) * rk3gtu_xy * vector(rk3Alpha, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
						m_Bfu_x_o0_t0_l0 = (-0.5 * vector(rk3phif, i, j, k) * rk3gtu_xx * vector(rk3Alpha, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
						m_Sfd_z_o0_t10_l0 = (-(vector(rk3Df, i, j, k) + vector(rk3tauf, i, j, k))) * d_Psi1_xx_o0_t3_m0_l0;
						m_Sfd_z_o0_t9_l0 = vector(rk3Sfd_z, i, j, k) * d_div_Beta_o0_t2_m0_l0;
						m_Sfd_z_o0_t8_l0 = vector(rk3Sfd_y, i, j, k) * d_Sfd_z_o0_t8_m0_l0;
						m_Sfd_z_o0_t7_l0 = vector(rk3Sfd_x, i, j, k) * d_Sfd_z_o0_t7_m0_l0;
						m_Sfd_z_o0_t6_l0 = (-0.5 * (vector(rk3Alpha, i, j, k) * rk3trSf) / rk3chi_max) * d_Rpd_xx_o0_t2_m0_l0;
						m_Sfd_z_o0_t5_l0 = 0.5 * (vector(rk3Alpha, i, j, k) * rk3Sfuu_zz) / rk3chi_max * d_Ctd_zzz_o0_t0_m0_l0;
						m_Sfd_z_o0_t4_l0 = (vector(rk3Alpha, i, j, k) * rk3Sfuu_yz) / rk3chi_max * d_Ctd_yzz_o0_t0_m0_l0;
						m_Sfd_z_o0_t3_l0 = 0.5 * (vector(rk3Alpha, i, j, k) * rk3Sfuu_yy) / rk3chi_max * d_Ctd_yyz_o0_t0_m0_l0;
						m_Sfd_z_o0_t2_l0 = (vector(rk3Alpha, i, j, k) * rk3Sfuu_xz) / rk3chi_max * d_Ctd_xzz_o0_t0_m0_l0;
						m_Sfd_z_o0_t1_l0 = (vector(rk3Alpha, i, j, k) * rk3Sfuu_xy) / rk3chi_max * d_Ctd_xyz_o0_t1_m0_l0;
						m_Sfd_z_o0_t0_l0 = 0.5 * (vector(rk3Alpha, i, j, k) * rk3Sfuu_xx) / rk3chi_max * d_Ctd_xxz_o0_t0_m0_l0;
						m_Sfd_y_o0_t10_l0 = (-(vector(rk3Df, i, j, k) + vector(rk3tauf, i, j, k))) * d_Psi1_xx_o0_t2_m0_l0;
						m_Sfd_y_o0_t9_l0 = vector(rk3Sfd_z, i, j, k) * d_Sfd_y_o0_t9_m0_l0;
						m_Sfd_y_o0_t8_l0 = vector(rk3Sfd_y, i, j, k) * d_div_Beta_o0_t1_m0_l0;
						m_Sfd_y_o0_t7_l0 = vector(rk3Sfd_x, i, j, k) * d_Sfd_y_o0_t7_m0_l0;
						m_Sfd_y_o0_t6_l0 = (-0.5 * (vector(rk3Alpha, i, j, k) * rk3trSf) / rk3chi_max) * d_Rpd_xx_o0_t1_m0_l0;
						m_Sfd_y_o0_t5_l0 = 0.5 * (vector(rk3Alpha, i, j, k) * rk3Sfuu_zz) / rk3chi_max * d_Ctd_yzz_o0_t1_m0_l0;
						m_Sfd_y_o0_t4_l0 = (vector(rk3Alpha, i, j, k) * rk3Sfuu_yz) / rk3chi_max * d_Ctd_zyy_o0_t0_m0_l0;
						m_Sfd_y_o0_t3_l0 = 0.5 * (vector(rk3Alpha, i, j, k) * rk3Sfuu_yy) / rk3chi_max * d_Ctd_yyy_o0_t0_m0_l0;
						m_Sfd_y_o0_t2_l0 = (vector(rk3Alpha, i, j, k) * rk3Sfuu_xz) / rk3chi_max * d_Ctd_xyz_o0_t0_m0_l0;
						m_Sfd_y_o0_t1_l0 = (vector(rk3Alpha, i, j, k) * rk3Sfuu_xy) / rk3chi_max * d_Ctd_xyy_o0_t0_m0_l0;
						m_Sfd_y_o0_t0_l0 = 0.5 * (vector(rk3Alpha, i, j, k) * rk3Sfuu_xx) / rk3chi_max * d_Ctd_xxy_o0_t0_m0_l0;
						m_Sfd_x_o0_t10_l0 = (-(vector(rk3Df, i, j, k) + vector(rk3tauf, i, j, k))) * d_Psi1_xx_o0_t1_m0_l0;
						m_Sfd_x_o0_t9_l0 = vector(rk3Sfd_z, i, j, k) * d_Sfd_x_o0_t9_m0_l0;
						m_Sfd_x_o0_t8_l0 = vector(rk3Sfd_y, i, j, k) * d_Sfd_x_o0_t8_m0_l0;
						m_Sfd_x_o0_t7_l0 = vector(rk3Sfd_x, i, j, k) * d_div_Beta_o0_t0_m0_l0;
						m_Sfd_x_o0_t6_l0 = (-0.5 * (vector(rk3Alpha, i, j, k) * rk3trSf) / rk3chi_max) * d_Rpd_xx_o0_t0_m0_l0;
						m_Sfd_x_o0_t5_l0 = 0.5 * (vector(rk3Alpha, i, j, k) * rk3Sfuu_zz) / rk3chi_max * d_Ctd_xzz_o0_t1_m0_l0;
						m_Sfd_x_o0_t4_l0 = (vector(rk3Alpha, i, j, k) * rk3Sfuu_yz) / rk3chi_max * d_Ctd_xyz_o0_t2_m0_l0;
						m_Sfd_x_o0_t3_l0 = 0.5 * (vector(rk3Alpha, i, j, k) * rk3Sfuu_yy) / rk3chi_max * d_Ctd_xyy_o0_t1_m0_l0;
						m_Sfd_x_o0_t2_l0 = (vector(rk3Alpha, i, j, k) * rk3Sfuu_xz) / rk3chi_max * d_Ctd_zxx_o0_t0_m0_l0;
						m_Sfd_x_o0_t1_l0 = (vector(rk3Alpha, i, j, k) * rk3Sfuu_xy) / rk3chi_max * d_Ctd_yxx_o0_t0_m0_l0;
						m_Sfd_x_o0_t0_l0 = 0.5 * (vector(rk3Alpha, i, j, k) * rk3Sfuu_xx) / rk3chi_max * d_Ctd_xxx_o0_t0_m0_l0;
						RHS_Sfd_x = RHS_Sfd_x + (((((((((((m_Sfd_x_o0_t0_l0 + m_Sfd_x_o0_t1_l0) + m_Sfd_x_o0_t2_l0) + m_Sfd_x_o0_t3_l0) + m_Sfd_x_o0_t4_l0) + m_Sfd_x_o0_t5_l0) + m_Sfd_x_o0_t6_l0) + m_Sfd_x_o0_t7_l0) + m_Sfd_x_o0_t8_l0) + m_Sfd_x_o0_t9_l0) + m_Sfd_x_o0_t10_l0) + d_Sfd_x_o0_t11_m0_l0);
						RHS_Sfd_y = RHS_Sfd_y + (((((((((((m_Sfd_y_o0_t0_l0 + m_Sfd_y_o0_t1_l0) + m_Sfd_y_o0_t2_l0) + m_Sfd_y_o0_t3_l0) + m_Sfd_y_o0_t4_l0) + m_Sfd_y_o0_t5_l0) + m_Sfd_y_o0_t6_l0) + m_Sfd_y_o0_t7_l0) + m_Sfd_y_o0_t8_l0) + m_Sfd_y_o0_t9_l0) + m_Sfd_y_o0_t10_l0) + d_Sfd_y_o0_t11_m0_l0);
						RHS_Sfd_z = RHS_Sfd_z + (((((((((((m_Sfd_z_o0_t0_l0 + m_Sfd_z_o0_t1_l0) + m_Sfd_z_o0_t2_l0) + m_Sfd_z_o0_t3_l0) + m_Sfd_z_o0_t4_l0) + m_Sfd_z_o0_t5_l0) + m_Sfd_z_o0_t6_l0) + m_Sfd_z_o0_t7_l0) + m_Sfd_z_o0_t8_l0) + m_Sfd_z_o0_t9_l0) + m_Sfd_z_o0_t10_l0) + d_Sfd_z_o0_t11_m0_l0);
						RHS_Bfu_x = RHS_Bfu_x + (((((m_Bfu_x_o0_t0_l0 + m_Bfu_x_o0_t1_l0) + m_Bfu_x_o0_t2_l0) + m_Bfu_x_o0_t3_l0) + m_Bfu_x_o0_t4_l0) + m_Bfu_x_o0_t5_l0);
						RHS_Bfu_y = RHS_Bfu_y + (((((m_Bfu_y_o0_t0_l0 + m_Bfu_y_o0_t1_l0) + m_Bfu_y_o0_t2_l0) + m_Bfu_y_o0_t3_l0) + m_Bfu_y_o0_t4_l0) + m_Bfu_y_o0_t5_l0);
						RHS_Bfu_z = RHS_Bfu_z + (((((m_Bfu_z_o0_t0_l0 + m_Bfu_z_o0_t1_l0) + m_Bfu_z_o0_t2_l0) + m_Bfu_z_o0_t3_l0) + m_Bfu_z_o0_t4_l0) + m_Bfu_z_o0_t5_l0);
						RHS_DYf = RHS_DYf + d_DYf_o0_t0_m0_l0;
						RHS_tauf = RHS_tauf + (((m_tauf_o0_t0_l0 + m_tauf_o0_t1_l0) + m_tauf_o0_t2_l0) + d_tauf_o0_t3_m0_l0);
						RHS_phif = RHS_phif + ((m_phif_o0_t0_l0 + m_phif_o0_t1_l0) + m_phif_o0_t2_l0);
						RHS_gtd_xx = ((((((((m_gtd_xx_o0_t0_l0 + m_gtd_xx_o0_t1_l0) + m_gtd_xx_o0_t2_l0) + d_gtd_xx_o0_t3_m0_l0) + m_gtd_xx_o1_t0_l0) + m_gtd_xx_o1_t1_l0) + m_gtd_xx_o1_t2_l0) + m_gtd_xx_o2_t0_l0) + m_gtd_xx_o2_t1_l0) + m_gtd_xx_o2_t2_l0;
						RHS_gtd_xy = (((((((((((m_gtd_xy_o0_t0_l0 + m_gtd_xy_o0_t1_l0) + m_gtd_xy_o0_t2_l0) + m_gtd_xy_o0_t3_l0) + m_gtd_xy_o0_t4_l0) + m_gtd_xy_o0_t5_l0) + d_gtd_xy_o0_t6_m0_l0) + m_gtd_xy_o1_t0_l0) + m_gtd_xy_o1_t1_l0) + m_gtd_xy_o1_t2_l0) + m_gtd_xy_o2_t0_l0) + m_gtd_xy_o2_t1_l0) + m_gtd_xy_o2_t2_l0;
						RHS_gtd_xz = (((((((((((m_gtd_xz_o0_t0_l0 + m_gtd_xz_o0_t1_l0) + m_gtd_xz_o0_t2_l0) + m_gtd_xz_o0_t3_l0) + m_gtd_xz_o0_t4_l0) + m_gtd_xz_o0_t5_l0) + d_gtd_xz_o0_t6_m0_l0) + m_gtd_xz_o1_t0_l0) + m_gtd_xz_o1_t1_l0) + m_gtd_xz_o1_t2_l0) + m_gtd_xz_o2_t0_l0) + m_gtd_xz_o2_t1_l0) + m_gtd_xz_o2_t2_l0;
						RHS_gtd_yy = ((((((((m_gtd_yy_o0_t0_l0 + m_gtd_yy_o0_t1_l0) + m_gtd_yy_o0_t2_l0) + d_gtd_yy_o0_t3_m0_l0) + m_gtd_yy_o1_t0_l0) + m_gtd_yy_o1_t1_l0) + m_gtd_yy_o1_t2_l0) + m_gtd_yy_o2_t0_l0) + m_gtd_yy_o2_t1_l0) + m_gtd_yy_o2_t2_l0;
						RHS_gtd_yz = (((((((((((m_gtd_yz_o0_t0_l0 + m_gtd_yz_o0_t1_l0) + m_gtd_yz_o0_t2_l0) + m_gtd_yz_o0_t3_l0) + m_gtd_yz_o0_t4_l0) + m_gtd_yz_o0_t5_l0) + d_gtd_yz_o0_t6_m0_l0) + m_gtd_yz_o1_t0_l0) + m_gtd_yz_o1_t1_l0) + m_gtd_yz_o1_t2_l0) + m_gtd_yz_o2_t0_l0) + m_gtd_yz_o2_t1_l0) + m_gtd_yz_o2_t2_l0;
						RHS_gtd_zz = ((((((((m_gtd_zz_o0_t0_l0 + m_gtd_zz_o0_t1_l0) + m_gtd_zz_o0_t2_l0) + d_gtd_zz_o0_t3_m0_l0) + m_gtd_zz_o1_t0_l0) + m_gtd_zz_o1_t1_l0) + m_gtd_zz_o1_t2_l0) + m_gtd_zz_o2_t0_l0) + m_gtd_zz_o2_t1_l0) + m_gtd_zz_o2_t2_l0;
						RHS_Atd_xx = ((((((((m_Atd_xx_o0_t0_l0 + m_Atd_xx_o0_t1_l0) + m_Atd_xx_o0_t2_l0) + d_Atd_xx_o0_t3_m0_l0) + m_Atd_xx_o1_t0_l0) + m_Atd_xx_o1_t1_l0) + m_Atd_xx_o1_t2_l0) + m_Atd_xx_o2_t0_l0) + m_Atd_xx_o2_t1_l0) + m_Atd_xx_o2_t2_l0;
						RHS_Atd_xy = (((((((((((m_Atd_xy_o0_t0_l0 + m_Atd_xy_o0_t1_l0) + m_Atd_xy_o0_t2_l0) + m_Atd_xy_o0_t3_l0) + m_Atd_xy_o0_t4_l0) + m_Atd_xy_o0_t5_l0) + d_Atd_xy_o0_t6_m0_l0) + m_Atd_xy_o1_t0_l0) + m_Atd_xy_o1_t1_l0) + m_Atd_xy_o1_t2_l0) + m_Atd_xy_o2_t0_l0) + m_Atd_xy_o2_t1_l0) + m_Atd_xy_o2_t2_l0;
						RHS_Atd_xz = (((((((((((m_Atd_xz_o0_t0_l0 + m_Atd_xz_o0_t1_l0) + m_Atd_xz_o0_t2_l0) + m_Atd_xz_o0_t3_l0) + m_Atd_xz_o0_t4_l0) + m_Atd_xz_o0_t5_l0) + d_Atd_xz_o0_t6_m0_l0) + m_Atd_xz_o1_t0_l0) + m_Atd_xz_o1_t1_l0) + m_Atd_xz_o1_t2_l0) + m_Atd_xz_o2_t0_l0) + m_Atd_xz_o2_t1_l0) + m_Atd_xz_o2_t2_l0;
						RHS_Atd_yy = ((((((((m_Atd_yy_o0_t0_l0 + m_Atd_yy_o0_t1_l0) + m_Atd_yy_o0_t2_l0) + d_Atd_yy_o0_t3_m0_l0) + m_Atd_yy_o1_t0_l0) + m_Atd_yy_o1_t1_l0) + m_Atd_yy_o1_t2_l0) + m_Atd_yy_o2_t0_l0) + m_Atd_yy_o2_t1_l0) + m_Atd_yy_o2_t2_l0;
						RHS_Atd_yz = (((((((((((m_Atd_yz_o0_t0_l0 + m_Atd_yz_o0_t1_l0) + m_Atd_yz_o0_t2_l0) + m_Atd_yz_o0_t3_l0) + m_Atd_yz_o0_t4_l0) + m_Atd_yz_o0_t5_l0) + d_Atd_yz_o0_t6_m0_l0) + m_Atd_yz_o1_t0_l0) + m_Atd_yz_o1_t1_l0) + m_Atd_yz_o1_t2_l0) + m_Atd_yz_o2_t0_l0) + m_Atd_yz_o2_t1_l0) + m_Atd_yz_o2_t2_l0;
						RHS_Atd_zz = ((((((((m_Atd_zz_o0_t0_l0 + m_Atd_zz_o0_t1_l0) + m_Atd_zz_o0_t2_l0) + d_Atd_zz_o0_t3_m0_l0) + m_Atd_zz_o1_t0_l0) + m_Atd_zz_o1_t1_l0) + m_Atd_zz_o1_t2_l0) + m_Atd_zz_o2_t0_l0) + m_Atd_zz_o2_t1_l0) + m_Atd_zz_o2_t2_l0;
						RHS_Gamh_x = (((((((((((((((((((((((((((((m_Gamh_x_o0_t0_l0 + m_Gamh_x_o0_t1_l0) + m_Gamh_x_o0_t2_l0) + m_Gamh_x_o0_t3_l0) + m_Gamh_x_o0_t4_l0) + m_Gamh_x_o0_t5_l0) + m_Gamh_x_o0_t6_l0) + m_Gamh_x_o0_t7_l0) + m_Gamh_x_o0_t8_l0) + m_Gamh_x_o0_t9_l0) + m_Gamh_x_o0_t10_l0) + m_Gamh_x_o0_t11_l0) + m_Gamh_x_o0_t12_l0) + m_Gamh_x_o0_t13_l0) + m_Gamh_x_o0_t14_l0) + m_Gamh_x_o0_t15_l0) + m_Gamh_x_o0_t16_l0) + m_Gamh_x_o0_t17_l0) + m_Gamh_x_o0_t18_l0) + m_Gamh_x_o0_t19_l0) + m_Gamh_x_o0_t20_l0) + m_Gamh_x_o0_t21_l0) + m_Gamh_x_o0_t22_l0) + m_Gamh_x_o0_t23_l0) + d_Gamh_x_o0_t24_m0_l0) + m_Gamh_x_o1_t0_l0) + m_Gamh_x_o1_t1_l0) + m_Gamh_x_o1_t2_l0) + m_Gamh_x_o2_t0_l0) + m_Gamh_x_o2_t1_l0) + m_Gamh_x_o2_t2_l0;
						RHS_Gamh_y = (((((((((((((((((((((((((((((m_Gamh_y_o0_t0_l0 + m_Gamh_y_o0_t1_l0) + m_Gamh_y_o0_t2_l0) + m_Gamh_y_o0_t3_l0) + m_Gamh_y_o0_t4_l0) + m_Gamh_y_o0_t5_l0) + m_Gamh_y_o0_t6_l0) + m_Gamh_y_o0_t7_l0) + m_Gamh_y_o0_t8_l0) + m_Gamh_y_o0_t9_l0) + m_Gamh_y_o0_t10_l0) + m_Gamh_y_o0_t11_l0) + m_Gamh_y_o0_t12_l0) + m_Gamh_y_o0_t13_l0) + m_Gamh_y_o0_t14_l0) + m_Gamh_y_o0_t15_l0) + m_Gamh_y_o0_t16_l0) + m_Gamh_y_o0_t17_l0) + m_Gamh_y_o0_t18_l0) + m_Gamh_y_o0_t19_l0) + m_Gamh_y_o0_t20_l0) + m_Gamh_y_o0_t21_l0) + m_Gamh_y_o0_t22_l0) + m_Gamh_y_o0_t23_l0) + d_Gamh_y_o0_t24_m0_l0) + m_Gamh_y_o1_t0_l0) + m_Gamh_y_o1_t1_l0) + m_Gamh_y_o1_t2_l0) + m_Gamh_y_o2_t0_l0) + m_Gamh_y_o2_t1_l0) + m_Gamh_y_o2_t2_l0;
						RHS_Gamh_z = (((((((((((((((((((((((((((((m_Gamh_z_o0_t0_l0 + m_Gamh_z_o0_t1_l0) + m_Gamh_z_o0_t2_l0) + m_Gamh_z_o0_t3_l0) + m_Gamh_z_o0_t4_l0) + m_Gamh_z_o0_t5_l0) + m_Gamh_z_o0_t6_l0) + m_Gamh_z_o0_t7_l0) + m_Gamh_z_o0_t8_l0) + m_Gamh_z_o0_t9_l0) + m_Gamh_z_o0_t10_l0) + m_Gamh_z_o0_t11_l0) + m_Gamh_z_o0_t12_l0) + m_Gamh_z_o0_t13_l0) + m_Gamh_z_o0_t14_l0) + m_Gamh_z_o0_t15_l0) + m_Gamh_z_o0_t16_l0) + m_Gamh_z_o0_t17_l0) + m_Gamh_z_o0_t18_l0) + m_Gamh_z_o0_t19_l0) + m_Gamh_z_o0_t20_l0) + m_Gamh_z_o0_t21_l0) + m_Gamh_z_o0_t22_l0) + m_Gamh_z_o0_t23_l0) + d_Gamh_z_o0_t24_m0_l0) + m_Gamh_z_o1_t0_l0) + m_Gamh_z_o1_t1_l0) + m_Gamh_z_o1_t2_l0) + m_Gamh_z_o2_t0_l0) + m_Gamh_z_o2_t1_l0) + m_Gamh_z_o2_t2_l0;
						RHS_Betau_x = (((((d_Betau_x_o0_t0_m0_l0 + m_Betau_x_o1_t0_l0) + m_Betau_x_o1_t1_l0) + m_Betau_x_o1_t2_l0) + m_Betau_x_o2_t0_l0) + m_Betau_x_o2_t1_l0) + m_Betau_x_o2_t2_l0;
						RHS_Betau_y = (((((d_Betau_y_o0_t0_m0_l0 + m_Betau_y_o1_t0_l0) + m_Betau_y_o1_t1_l0) + m_Betau_y_o1_t2_l0) + m_Betau_y_o2_t0_l0) + m_Betau_y_o2_t1_l0) + m_Betau_y_o2_t2_l0;
						RHS_Betau_z = (((((d_Betau_z_o0_t0_m0_l0 + m_Betau_z_o1_t0_l0) + m_Betau_z_o1_t1_l0) + m_Betau_z_o1_t2_l0) + m_Betau_z_o2_t0_l0) + m_Betau_z_o2_t1_l0) + m_Betau_z_o2_t2_l0;
						RHS_Alpha = (((((d_Alpha_o0_t0_m0_l0 + m_Alpha_o1_t0_l0) + m_Alpha_o1_t1_l0) + m_Alpha_o1_t2_l0) + m_Alpha_o2_t0_l0) + m_Alpha_o2_t1_l0) + m_Alpha_o2_t2_l0;
						RHS_chi = (((((d_chi_o0_t0_m0_l0 + m_chi_o1_t0_l0) + m_chi_o1_t1_l0) + m_chi_o1_t2_l0) + m_chi_o2_t0_l0) + m_chi_o2_t1_l0) + m_chi_o2_t2_l0;
						RHS_trK = ((((((((((((((((((((((((((m_trK_o0_t0_l0 + m_trK_o0_t1_l0) + m_trK_o0_t2_l0) + m_trK_o0_t3_l0) + m_trK_o0_t4_l0) + m_trK_o0_t5_l0) + m_trK_o0_t6_l0) + m_trK_o0_t7_l0) + m_trK_o0_t8_l0) + m_trK_o0_t9_l0) + m_trK_o0_t10_l0) + m_trK_o0_t11_l0) + m_trK_o0_t12_l0) + m_trK_o0_t13_l0) + m_trK_o0_t14_l0) + m_trK_o0_t15_l0) + m_trK_o0_t16_l0) + m_trK_o0_t17_l0) + m_trK_o0_t18_l0) + m_trK_o0_t19_l0) + m_trK_o0_t20_l0) + d_trK_o0_t21_m0_l0) + m_trK_o1_t0_l0) + m_trK_o1_t1_l0) + m_trK_o1_t2_l0) + m_trK_o2_t0_l0) + m_trK_o2_t1_l0) + m_trK_o2_t2_l0;
						RHS_theta = ((((((((m_theta_o0_t0_l0 + m_theta_o0_t1_l0) + m_theta_o0_t2_l0) + d_theta_o0_t3_m0_l0) + m_theta_o1_t0_l0) + m_theta_o1_t1_l0) + m_theta_o1_t2_l0) + m_theta_o2_t0_l0) + m_theta_o2_t1_l0) + m_theta_o2_t2_l0;
						n_x = 0.0;
						n_y = 0.0;
						n_z = 0.0;
						interaction_index = 0.0;
						if ((((((vector(FOV_xLower, i + 1, j, k) > 0.0) || (vector(FOV_xUpper, i + 1, j, k) > 0.0)) || (vector(FOV_yLower, i + 1, j, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k) > 0.0)) || (vector(FOV_zLower, i + 1, j, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x + 1.0 / dx[0];
						}
						if ((((((vector(FOV_xLower, i - 1, j, k) > 0.0) || (vector(FOV_xUpper, i - 1, j, k) > 0.0)) || (vector(FOV_yLower, i - 1, j, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k) > 0.0)) || (vector(FOV_zLower, i - 1, j, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x - 1.0 / dx[0];
						}
						if ((((((vector(FOV_xLower, i, j + 1, k) > 0.0) || (vector(FOV_xUpper, i, j + 1, k) > 0.0)) || (vector(FOV_yLower, i, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k) > 0.0)) || (vector(FOV_zLower, i, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y + 1.0 / dx[1];
						}
						if ((((((vector(FOV_xLower, i, j - 1, k) > 0.0) || (vector(FOV_xUpper, i, j - 1, k) > 0.0)) || (vector(FOV_yLower, i, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k) > 0.0)) || (vector(FOV_zLower, i, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y - 1.0 / dx[1];
						}
						if ((((((vector(FOV_xLower, i, j, k + 1) > 0.0) || (vector(FOV_xUpper, i, j, k + 1) > 0.0)) || (vector(FOV_yLower, i, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j, k + 1) > 0.0)) || (vector(FOV_zLower, i, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j, k + 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z + 1.0 / dx[2];
						}
						if ((((((vector(FOV_xLower, i, j, k - 1) > 0.0) || (vector(FOV_xUpper, i, j, k - 1) > 0.0)) || (vector(FOV_yLower, i, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j, k - 1) > 0.0)) || (vector(FOV_zLower, i, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j, k - 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z - 1.0 / dx[2];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k) > 0.0)) || (vector(FOV_yLower, i + 1, j, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k) > 0.0)) || (vector(FOV_zLower, i + 1, j, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x + 1.0 / dx[0];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i - 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k) > 0.0)) || (vector(FOV_yLower, i - 1, j, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k) > 0.0)) || (vector(FOV_zLower, i - 1, j, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k) > 0.0)) {
							interaction_index = 1.0;
							n_x = n_x - 1.0 / dx[0];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k) > 0.0)) || (vector(FOV_xLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k) > 0.0)) || (vector(FOV_yLower, i, j + 1, k) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k) > 0.0)) || (vector(FOV_zLower, i, j + 1, k) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y + 1.0 / dx[1];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j - 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k) > 0.0)) || (vector(FOV_xLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k) > 0.0)) || (vector(FOV_yLower, i, j - 1, k) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k) > 0.0)) || (vector(FOV_zLower, i, j - 1, k) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k) > 0.0)) {
							interaction_index = 1.0;
							n_y = n_y - 1.0 / dx[1];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k + 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k + 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k + 1) > 0.0)) || (vector(FOV_xLower, i, j, k + 1) > 0.0)) || (vector(FOV_xUpper, i, j, k + 1) > 0.0)) || (vector(FOV_yLower, i, j, k + 1) > 0.0)) || (vector(FOV_yUpper, i, j, k + 1) > 0.0)) || (vector(FOV_zLower, i, j, k + 1) > 0.0)) || (vector(FOV_zUpper, i, j, k + 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z + 1.0 / dx[2];
						}
						if ((((((((((((((((((((((((((((((((((((((((((((((((((((((vector(FOV_xLower, i + 1, j + 1, k - 1) > 0.0) || (vector(FOV_xUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i + 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i + 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zLower, i - 1, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i - 1, j, k - 1) > 0.0)) || (vector(FOV_xLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j + 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j + 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zLower, i, j - 1, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j - 1, k - 1) > 0.0)) || (vector(FOV_xLower, i, j, k - 1) > 0.0)) || (vector(FOV_xUpper, i, j, k - 1) > 0.0)) || (vector(FOV_yLower, i, j, k - 1) > 0.0)) || (vector(FOV_yUpper, i, j, k - 1) > 0.0)) || (vector(FOV_zLower, i, j, k - 1) > 0.0)) || (vector(FOV_zUpper, i, j, k - 1) > 0.0)) {
							interaction_index = 1.0;
							n_z = n_z - 1.0 / dx[2];
						}
						if ((!equalsEq(n_x, 0.0) || !equalsEq(n_y, 0.0)) || !equalsEq(n_z, 0.0)) {
							mod_normal = sqrt((n_x * n_x + n_y * n_y) + n_z * n_z);
							n_x = n_x / mod_normal;
							n_y = n_y / mod_normal;
							n_z = n_z / mod_normal;
						}
						if (equalsEq(interaction_index, 1.0)) {
							i_d_Atd_xx_o0_t0_m0_l0 = D1CDO4_i(rk3Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xx_o0_t1_m0_l0 = D1CDO4_j(rk3Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xx_o0_t2_m0_l0 = D1CDO4_k(rk3Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xy_o0_t0_m0_l0 = D1CDO4_i(rk3Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xy_o0_t1_m0_l0 = D1CDO4_j(rk3Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xy_o0_t2_m0_l0 = D1CDO4_k(rk3Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xz_o0_t0_m0_l0 = D1CDO4_i(rk3Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xz_o0_t1_m0_l0 = D1CDO4_j(rk3Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_xz_o0_t2_m0_l0 = D1CDO4_k(rk3Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yy_o0_t0_m0_l0 = D1CDO4_i(rk3Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yy_o0_t1_m0_l0 = D1CDO4_j(rk3Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yy_o0_t2_m0_l0 = D1CDO4_k(rk3Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yz_o0_t0_m0_l0 = D1CDO4_i(rk3Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yz_o0_t1_m0_l0 = D1CDO4_j(rk3Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_yz_o0_t2_m0_l0 = D1CDO4_k(rk3Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_zz_o0_t0_m0_l0 = D1CDO4_i(rk3Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_zz_o0_t1_m0_l0 = D1CDO4_j(rk3Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_Atd_zz_o0_t2_m0_l0 = D1CDO4_k(rk3Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
							i_d_gtd_xx_o0_t3_m0_l0 = -gtd_xx_falloff * (vector(rk3gtd_xx, i, j, k) - gtd_xx_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_xy_o0_t3_m0_l0 = -gtd_xy_falloff * (vector(rk3gtd_xy, i, j, k) - gtd_xy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_xz_o0_t3_m0_l0 = -gtd_xz_falloff * (vector(rk3gtd_xz, i, j, k) - gtd_xz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_yy_o0_t3_m0_l0 = -gtd_yy_falloff * (vector(rk3gtd_yy, i, j, k) - gtd_yy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_yz_o0_t3_m0_l0 = -gtd_yz_falloff * (vector(rk3gtd_yz, i, j, k) - gtd_yz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_gtd_zz_o0_t3_m0_l0 = -gtd_zz_falloff * (vector(rk3gtd_zz, i, j, k) - gtd_zz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_xx_o0_t3_m0_l0 = -Atd_xx_falloff * (vector(rk3Atd_xx, i, j, k) - Atd_xx_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_xy_o0_t3_m0_l0 = -Atd_xy_falloff * (vector(rk3Atd_xy, i, j, k) - Atd_xy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_xz_o0_t3_m0_l0 = -Atd_xz_falloff * (vector(rk3Atd_xz, i, j, k) - Atd_xz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_yy_o0_t3_m0_l0 = -Atd_yy_falloff * (vector(rk3Atd_yy, i, j, k) - Atd_yy_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_yz_o0_t3_m0_l0 = -Atd_yz_falloff * (vector(rk3Atd_yz, i, j, k) - Atd_yz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Atd_zz_o0_t3_m0_l0 = -Atd_zz_falloff * (vector(rk3Atd_zz, i, j, k) - Atd_zz_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Gamh_x_o0_t3_m0_l0 = -Gamh_x_falloff * (vector(rk3Gamh_x, i, j, k) - Gamh_x_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Gamh_y_o0_t3_m0_l0 = -Gamh_y_falloff * (vector(rk3Gamh_y, i, j, k) - Gamh_y_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Gamh_z_o0_t3_m0_l0 = -Gamh_z_falloff * (vector(rk3Gamh_z, i, j, k) - Gamh_z_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Betau_x_o0_t3_m0_l0 = -Betau_x_falloff * (vector(rk3Betau_x, i, j, k) - Betau_x_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Betau_y_o0_t3_m0_l0 = -Betau_y_falloff * (vector(rk3Betau_y, i, j, k) - Betau_y_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Betau_z_o0_t3_m0_l0 = -Betau_z_falloff * (vector(rk3Betau_z, i, j, k) - Betau_z_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_Alpha_o0_t3_m0_l0 = -Alpha_falloff * (vector(rk3Alpha, i, j, k) - Alpha_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_chi_o0_t3_m0_l0 = -chi_falloff * (vector(rk3chi, i, j, k) - chi_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_trK_o0_t3_m0_l0 = -trK_falloff * (vector(rk3trK, i, j, k) - trK_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_d_theta_o0_t3_m0_l0 = -theta_falloff * (vector(rk3theta, i, j, k) - theta_asymptotic) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k));
							i_m_gtd_xx_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xxx_o0_t0_m0_l0;
							i_m_gtd_xx_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xxy_o0_t0_m0_l0;
							i_m_gtd_xx_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xxz_o0_t0_m0_l0;
							i_m_gtd_xy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yxx_o0_t0_m0_l0;
							i_m_gtd_xy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyy_o0_t0_m0_l0;
							i_m_gtd_xy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyz_o0_t1_m0_l0;
							i_m_gtd_xz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_zxx_o0_t0_m0_l0;
							i_m_gtd_xz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyz_o0_t0_m0_l0;
							i_m_gtd_xz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xzz_o0_t0_m0_l0;
							i_m_gtd_yy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyy_o0_t1_m0_l0;
							i_m_gtd_yy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yyy_o0_t0_m0_l0;
							i_m_gtd_yy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yyz_o0_t0_m0_l0;
							i_m_gtd_yz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xyz_o0_t2_m0_l0;
							i_m_gtd_yz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_zyy_o0_t0_m0_l0;
							i_m_gtd_yz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yzz_o0_t0_m0_l0;
							i_m_gtd_zz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_xzz_o0_t1_m0_l0;
							i_m_gtd_zz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_yzz_o0_t1_m0_l0;
							i_m_gtd_zz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Ctd_zzz_o0_t0_m0_l0;
							i_m_Atd_xx_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xx_o0_t0_m0_l0;
							i_m_Atd_xx_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xx_o0_t1_m0_l0;
							i_m_Atd_xx_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xx_o0_t2_m0_l0;
							i_m_Atd_xy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xy_o0_t0_m0_l0;
							i_m_Atd_xy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xy_o0_t1_m0_l0;
							i_m_Atd_xy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xy_o0_t2_m0_l0;
							i_m_Atd_xz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xz_o0_t0_m0_l0;
							i_m_Atd_xz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xz_o0_t1_m0_l0;
							i_m_Atd_xz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_xz_o0_t2_m0_l0;
							i_m_Atd_yy_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yy_o0_t0_m0_l0;
							i_m_Atd_yy_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yy_o0_t1_m0_l0;
							i_m_Atd_yy_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yy_o0_t2_m0_l0;
							i_m_Atd_yz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yz_o0_t0_m0_l0;
							i_m_Atd_yz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yz_o0_t1_m0_l0;
							i_m_Atd_yz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_yz_o0_t2_m0_l0;
							i_m_Atd_zz_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_zz_o0_t0_m0_l0;
							i_m_Atd_zz_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_zz_o0_t1_m0_l0;
							i_m_Atd_zz_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * i_d_Atd_zz_o0_t2_m0_l0;
							i_m_Gamh_x_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xx_o0_t6_m0_l0;
							i_m_Gamh_x_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xy_o0_t6_m0_l0;
							i_m_Gamh_x_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xz_o0_t6_m0_l0;
							i_m_Gamh_y_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xx_o0_t7_m0_l0;
							i_m_Gamh_y_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xy_o0_t7_m0_l0;
							i_m_Gamh_y_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xz_o0_t7_m0_l0;
							i_m_Gamh_z_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xx_o0_t8_m0_l0;
							i_m_Gamh_z_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xy_o0_t8_m0_l0;
							i_m_Gamh_z_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rtd_xz_o0_t8_m0_l0;
							i_m_Betau_x_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_div_Beta_o0_t0_m0_l0;
							i_m_Betau_x_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_y_o0_t7_m0_l0;
							i_m_Betau_x_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_z_o0_t7_m0_l0;
							i_m_Betau_y_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_x_o0_t8_m0_l0;
							i_m_Betau_y_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_div_Beta_o0_t1_m0_l0;
							i_m_Betau_y_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_z_o0_t8_m0_l0;
							i_m_Betau_z_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_x_o0_t9_m0_l0;
							i_m_Betau_z_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Sfd_y_o0_t9_m0_l0;
							i_m_Betau_z_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_div_Beta_o0_t2_m0_l0;
							i_m_Alpha_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Psi1_xx_o0_t1_m0_l0;
							i_m_Alpha_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Psi1_xx_o0_t2_m0_l0;
							i_m_Alpha_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Psi1_xx_o0_t3_m0_l0;
							i_m_chi_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rpd_xx_o0_t0_m0_l0;
							i_m_chi_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rpd_xx_o0_t1_m0_l0;
							i_m_chi_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Rpd_xx_o0_t2_m0_l0;
							i_m_trK_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t0_m0_l0;
							i_m_trK_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t1_m0_l0;
							i_m_trK_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t2_m0_l0;
							i_m_theta_o0_t0_l0 = (-xcoord(i) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t3_m0_l0;
							i_m_theta_o0_t1_l0 = (-ycoord(j) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t4_m0_l0;
							i_m_theta_o0_t2_l0 = (-zcoord(k) / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * d_Gamh_x_o0_t5_m0_l0;
							RHS_gtd_xx = ((i_m_gtd_xx_o0_t0_l0 + i_m_gtd_xx_o0_t1_l0) + i_m_gtd_xx_o0_t2_l0) + i_d_gtd_xx_o0_t3_m0_l0;
							RHS_gtd_xy = ((i_m_gtd_xy_o0_t0_l0 + i_m_gtd_xy_o0_t1_l0) + i_m_gtd_xy_o0_t2_l0) + i_d_gtd_xy_o0_t3_m0_l0;
							RHS_gtd_xz = ((i_m_gtd_xz_o0_t0_l0 + i_m_gtd_xz_o0_t1_l0) + i_m_gtd_xz_o0_t2_l0) + i_d_gtd_xz_o0_t3_m0_l0;
							RHS_gtd_yy = ((i_m_gtd_yy_o0_t0_l0 + i_m_gtd_yy_o0_t1_l0) + i_m_gtd_yy_o0_t2_l0) + i_d_gtd_yy_o0_t3_m0_l0;
							RHS_gtd_yz = ((i_m_gtd_yz_o0_t0_l0 + i_m_gtd_yz_o0_t1_l0) + i_m_gtd_yz_o0_t2_l0) + i_d_gtd_yz_o0_t3_m0_l0;
							RHS_gtd_zz = ((i_m_gtd_zz_o0_t0_l0 + i_m_gtd_zz_o0_t1_l0) + i_m_gtd_zz_o0_t2_l0) + i_d_gtd_zz_o0_t3_m0_l0;
							RHS_Atd_xx = ((i_m_Atd_xx_o0_t0_l0 + i_m_Atd_xx_o0_t1_l0) + i_m_Atd_xx_o0_t2_l0) + i_d_Atd_xx_o0_t3_m0_l0;
							RHS_Atd_xy = ((i_m_Atd_xy_o0_t0_l0 + i_m_Atd_xy_o0_t1_l0) + i_m_Atd_xy_o0_t2_l0) + i_d_Atd_xy_o0_t3_m0_l0;
							RHS_Atd_xz = ((i_m_Atd_xz_o0_t0_l0 + i_m_Atd_xz_o0_t1_l0) + i_m_Atd_xz_o0_t2_l0) + i_d_Atd_xz_o0_t3_m0_l0;
							RHS_Atd_yy = ((i_m_Atd_yy_o0_t0_l0 + i_m_Atd_yy_o0_t1_l0) + i_m_Atd_yy_o0_t2_l0) + i_d_Atd_yy_o0_t3_m0_l0;
							RHS_Atd_yz = ((i_m_Atd_yz_o0_t0_l0 + i_m_Atd_yz_o0_t1_l0) + i_m_Atd_yz_o0_t2_l0) + i_d_Atd_yz_o0_t3_m0_l0;
							RHS_Atd_zz = ((i_m_Atd_zz_o0_t0_l0 + i_m_Atd_zz_o0_t1_l0) + i_m_Atd_zz_o0_t2_l0) + i_d_Atd_zz_o0_t3_m0_l0;
							RHS_Gamh_x = ((i_m_Gamh_x_o0_t0_l0 + i_m_Gamh_x_o0_t1_l0) + i_m_Gamh_x_o0_t2_l0) + i_d_Gamh_x_o0_t3_m0_l0;
							RHS_Gamh_y = ((i_m_Gamh_y_o0_t0_l0 + i_m_Gamh_y_o0_t1_l0) + i_m_Gamh_y_o0_t2_l0) + i_d_Gamh_y_o0_t3_m0_l0;
							RHS_Gamh_z = ((i_m_Gamh_z_o0_t0_l0 + i_m_Gamh_z_o0_t1_l0) + i_m_Gamh_z_o0_t2_l0) + i_d_Gamh_z_o0_t3_m0_l0;
							RHS_Betau_x = ((i_m_Betau_x_o0_t0_l0 + i_m_Betau_x_o0_t1_l0) + i_m_Betau_x_o0_t2_l0) + i_d_Betau_x_o0_t3_m0_l0;
							RHS_Betau_y = ((i_m_Betau_y_o0_t0_l0 + i_m_Betau_y_o0_t1_l0) + i_m_Betau_y_o0_t2_l0) + i_d_Betau_y_o0_t3_m0_l0;
							RHS_Betau_z = ((i_m_Betau_z_o0_t0_l0 + i_m_Betau_z_o0_t1_l0) + i_m_Betau_z_o0_t2_l0) + i_d_Betau_z_o0_t3_m0_l0;
							RHS_Alpha = ((i_m_Alpha_o0_t0_l0 + i_m_Alpha_o0_t1_l0) + i_m_Alpha_o0_t2_l0) + i_d_Alpha_o0_t3_m0_l0;
							RHS_chi = ((i_m_chi_o0_t0_l0 + i_m_chi_o0_t1_l0) + i_m_chi_o0_t2_l0) + i_d_chi_o0_t3_m0_l0;
							RHS_trK = ((i_m_trK_o0_t0_l0 + i_m_trK_o0_t1_l0) + i_m_trK_o0_t2_l0) + i_d_trK_o0_t3_m0_l0;
							RHS_theta = ((i_m_theta_o0_t0_l0 + i_m_theta_o0_t1_l0) + i_m_theta_o0_t2_l0) + i_d_theta_o0_t3_m0_l0;
						}
						if (dissipation_factor_Sfd_x > 0.0) {
							RHS_Sfd_x = RHS_Sfd_x + dissipation_factor_Sfd_x * (meshDissipation_i(rk3Sfd_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Sfd_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Sfd_x, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Sfd_y > 0.0) {
							RHS_Sfd_y = RHS_Sfd_y + dissipation_factor_Sfd_y * (meshDissipation_i(rk3Sfd_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Sfd_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Sfd_y, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Sfd_z > 0.0) {
							RHS_Sfd_z = RHS_Sfd_z + dissipation_factor_Sfd_z * (meshDissipation_i(rk3Sfd_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Sfd_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Sfd_z, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Bfu_x > 0.0) {
							RHS_Bfu_x = RHS_Bfu_x + dissipation_factor_Bfu_x * (meshDissipation_i(rk3Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Bfu_y > 0.0) {
							RHS_Bfu_y = RHS_Bfu_y + dissipation_factor_Bfu_y * (meshDissipation_i(rk3Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Bfu_z > 0.0) {
							RHS_Bfu_z = RHS_Bfu_z + dissipation_factor_Bfu_z * (meshDissipation_i(rk3Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Df > 0.0) {
							RHS_Df = RHS_Df + dissipation_factor_Df * (meshDissipation_i(rk3Df, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Df, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Df, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_DYf > 0.0) {
							RHS_DYf = RHS_DYf + dissipation_factor_DYf * (meshDissipation_i(rk3DYf, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3DYf, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3DYf, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_tauf > 0.0) {
							RHS_tauf = RHS_tauf + dissipation_factor_tauf * (meshDissipation_i(rk3tauf, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3tauf, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3tauf, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_phif > 0.0) {
							RHS_phif = RHS_phif + dissipation_factor_phif * (meshDissipation_i(rk3phif, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3phif, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3phif, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_xx > 0.0) {
							RHS_gtd_xx = RHS_gtd_xx + dissipation_factor_gtd_xx * (meshDissipation_i(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_xy > 0.0) {
							RHS_gtd_xy = RHS_gtd_xy + dissipation_factor_gtd_xy * (meshDissipation_i(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_xz > 0.0) {
							RHS_gtd_xz = RHS_gtd_xz + dissipation_factor_gtd_xz * (meshDissipation_i(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_yy > 0.0) {
							RHS_gtd_yy = RHS_gtd_yy + dissipation_factor_gtd_yy * (meshDissipation_i(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_yz > 0.0) {
							RHS_gtd_yz = RHS_gtd_yz + dissipation_factor_gtd_yz * (meshDissipation_i(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_gtd_zz > 0.0) {
							RHS_gtd_zz = RHS_gtd_zz + dissipation_factor_gtd_zz * (meshDissipation_i(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_xx > 0.0) {
							RHS_Atd_xx = RHS_Atd_xx + dissipation_factor_Atd_xx * (meshDissipation_i(rk3Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_xy > 0.0) {
							RHS_Atd_xy = RHS_Atd_xy + dissipation_factor_Atd_xy * (meshDissipation_i(rk3Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_xz > 0.0) {
							RHS_Atd_xz = RHS_Atd_xz + dissipation_factor_Atd_xz * (meshDissipation_i(rk3Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_yy > 0.0) {
							RHS_Atd_yy = RHS_Atd_yy + dissipation_factor_Atd_yy * (meshDissipation_i(rk3Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_yz > 0.0) {
							RHS_Atd_yz = RHS_Atd_yz + dissipation_factor_Atd_yz * (meshDissipation_i(rk3Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Atd_zz > 0.0) {
							RHS_Atd_zz = RHS_Atd_zz + dissipation_factor_Atd_zz * (meshDissipation_i(rk3Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Gamh_x > 0.0) {
							RHS_Gamh_x = RHS_Gamh_x + dissipation_factor_Gamh_x * (meshDissipation_i(rk3Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Gamh_y > 0.0) {
							RHS_Gamh_y = RHS_Gamh_y + dissipation_factor_Gamh_y * (meshDissipation_i(rk3Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Gamh_z > 0.0) {
							RHS_Gamh_z = RHS_Gamh_z + dissipation_factor_Gamh_z * (meshDissipation_i(rk3Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Betau_x > 0.0) {
							RHS_Betau_x = RHS_Betau_x + dissipation_factor_Betau_x * (meshDissipation_i(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Betau_x, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Betau_y > 0.0) {
							RHS_Betau_y = RHS_Betau_y + dissipation_factor_Betau_y * (meshDissipation_i(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Betau_y, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Betau_z > 0.0) {
							RHS_Betau_z = RHS_Betau_z + dissipation_factor_Betau_z * (meshDissipation_i(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Betau_z, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_Alpha > 0.0) {
							RHS_Alpha = RHS_Alpha + dissipation_factor_Alpha * (meshDissipation_i(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3Alpha, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_chi > 0.0) {
							RHS_chi = RHS_chi + dissipation_factor_chi * (meshDissipation_i(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3chi, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_trK > 0.0) {
							RHS_trK = RHS_trK + dissipation_factor_trK * (meshDissipation_i(rk3trK, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3trK, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3trK, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						if (dissipation_factor_theta > 0.0) {
							RHS_theta = RHS_theta + dissipation_factor_theta * (meshDissipation_i(rk3theta, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_j(rk3theta, i, j, k, dx, simPlat_dt, ilast, jlast) + meshDissipation_k(rk3theta, i, j, k, dx, simPlat_dt, ilast, jlast));
						}
						vector(Sfd_x, i, j, k) = RK4P4_(RHS_Sfd_x, vector(Sfd_x_p, i, j, k), vector(rk1Sfd_x, i, j, k), vector(rk2Sfd_x, i, j, k), vector(rk3Sfd_x, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Sfd_y, i, j, k) = RK4P4_(RHS_Sfd_y, vector(Sfd_y_p, i, j, k), vector(rk1Sfd_y, i, j, k), vector(rk2Sfd_y, i, j, k), vector(rk3Sfd_y, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Sfd_z, i, j, k) = RK4P4_(RHS_Sfd_z, vector(Sfd_z_p, i, j, k), vector(rk1Sfd_z, i, j, k), vector(rk2Sfd_z, i, j, k), vector(rk3Sfd_z, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Bfu_x, i, j, k) = RK4P4_(RHS_Bfu_x, vector(Bfu_x_p, i, j, k), vector(rk1Bfu_x, i, j, k), vector(rk2Bfu_x, i, j, k), vector(rk3Bfu_x, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Bfu_y, i, j, k) = RK4P4_(RHS_Bfu_y, vector(Bfu_y_p, i, j, k), vector(rk1Bfu_y, i, j, k), vector(rk2Bfu_y, i, j, k), vector(rk3Bfu_y, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Bfu_z, i, j, k) = RK4P4_(RHS_Bfu_z, vector(Bfu_z_p, i, j, k), vector(rk1Bfu_z, i, j, k), vector(rk2Bfu_z, i, j, k), vector(rk3Bfu_z, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Df, i, j, k) = RK4P4_(RHS_Df, vector(Df_p, i, j, k), vector(rk1Df, i, j, k), vector(rk2Df, i, j, k), vector(rk3Df, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(DYf, i, j, k) = RK4P4_(RHS_DYf, vector(DYf_p, i, j, k), vector(rk1DYf, i, j, k), vector(rk2DYf, i, j, k), vector(rk3DYf, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(tauf, i, j, k) = RK4P4_(RHS_tauf, vector(tauf_p, i, j, k), vector(rk1tauf, i, j, k), vector(rk2tauf, i, j, k), vector(rk3tauf, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(phif, i, j, k) = RK4P4_(RHS_phif, vector(phif_p, i, j, k), vector(rk1phif, i, j, k), vector(rk2phif, i, j, k), vector(rk3phif, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(gtd_xx, i, j, k) = RK4P4_(RHS_gtd_xx, vector(gtd_xx_p, i, j, k), vector(rk1gtd_xx, i, j, k), vector(rk2gtd_xx, i, j, k), vector(rk3gtd_xx, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(gtd_xy, i, j, k) = RK4P4_(RHS_gtd_xy, vector(gtd_xy_p, i, j, k), vector(rk1gtd_xy, i, j, k), vector(rk2gtd_xy, i, j, k), vector(rk3gtd_xy, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(gtd_xz, i, j, k) = RK4P4_(RHS_gtd_xz, vector(gtd_xz_p, i, j, k), vector(rk1gtd_xz, i, j, k), vector(rk2gtd_xz, i, j, k), vector(rk3gtd_xz, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(gtd_yy, i, j, k) = RK4P4_(RHS_gtd_yy, vector(gtd_yy_p, i, j, k), vector(rk1gtd_yy, i, j, k), vector(rk2gtd_yy, i, j, k), vector(rk3gtd_yy, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(gtd_yz, i, j, k) = RK4P4_(RHS_gtd_yz, vector(gtd_yz_p, i, j, k), vector(rk1gtd_yz, i, j, k), vector(rk2gtd_yz, i, j, k), vector(rk3gtd_yz, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(gtd_zz, i, j, k) = RK4P4_(RHS_gtd_zz, vector(gtd_zz_p, i, j, k), vector(rk1gtd_zz, i, j, k), vector(rk2gtd_zz, i, j, k), vector(rk3gtd_zz, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Atd_xx, i, j, k) = RK4P4_(RHS_Atd_xx, vector(Atd_xx_p, i, j, k), vector(rk1Atd_xx, i, j, k), vector(rk2Atd_xx, i, j, k), vector(rk3Atd_xx, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Atd_xy, i, j, k) = RK4P4_(RHS_Atd_xy, vector(Atd_xy_p, i, j, k), vector(rk1Atd_xy, i, j, k), vector(rk2Atd_xy, i, j, k), vector(rk3Atd_xy, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Atd_xz, i, j, k) = RK4P4_(RHS_Atd_xz, vector(Atd_xz_p, i, j, k), vector(rk1Atd_xz, i, j, k), vector(rk2Atd_xz, i, j, k), vector(rk3Atd_xz, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Atd_yy, i, j, k) = RK4P4_(RHS_Atd_yy, vector(Atd_yy_p, i, j, k), vector(rk1Atd_yy, i, j, k), vector(rk2Atd_yy, i, j, k), vector(rk3Atd_yy, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Atd_yz, i, j, k) = RK4P4_(RHS_Atd_yz, vector(Atd_yz_p, i, j, k), vector(rk1Atd_yz, i, j, k), vector(rk2Atd_yz, i, j, k), vector(rk3Atd_yz, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Atd_zz, i, j, k) = RK4P4_(RHS_Atd_zz, vector(Atd_zz_p, i, j, k), vector(rk1Atd_zz, i, j, k), vector(rk2Atd_zz, i, j, k), vector(rk3Atd_zz, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Gamh_x, i, j, k) = RK4P4_(RHS_Gamh_x, vector(Gamh_x_p, i, j, k), vector(rk1Gamh_x, i, j, k), vector(rk2Gamh_x, i, j, k), vector(rk3Gamh_x, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Gamh_y, i, j, k) = RK4P4_(RHS_Gamh_y, vector(Gamh_y_p, i, j, k), vector(rk1Gamh_y, i, j, k), vector(rk2Gamh_y, i, j, k), vector(rk3Gamh_y, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Gamh_z, i, j, k) = RK4P4_(RHS_Gamh_z, vector(Gamh_z_p, i, j, k), vector(rk1Gamh_z, i, j, k), vector(rk2Gamh_z, i, j, k), vector(rk3Gamh_z, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Betau_x, i, j, k) = RK4P4_(RHS_Betau_x, vector(Betau_x_p, i, j, k), vector(rk1Betau_x, i, j, k), vector(rk2Betau_x, i, j, k), vector(rk3Betau_x, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Betau_y, i, j, k) = RK4P4_(RHS_Betau_y, vector(Betau_y_p, i, j, k), vector(rk1Betau_y, i, j, k), vector(rk2Betau_y, i, j, k), vector(rk3Betau_y, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Betau_z, i, j, k) = RK4P4_(RHS_Betau_z, vector(Betau_z_p, i, j, k), vector(rk1Betau_z, i, j, k), vector(rk2Betau_z, i, j, k), vector(rk3Betau_z, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(Alpha, i, j, k) = RK4P4_(RHS_Alpha, vector(Alpha_p, i, j, k), vector(rk1Alpha, i, j, k), vector(rk2Alpha, i, j, k), vector(rk3Alpha, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(chi, i, j, k) = RK4P4_(RHS_chi, vector(chi_p, i, j, k), vector(rk1chi, i, j, k), vector(rk2chi, i, j, k), vector(rk3chi, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(trK, i, j, k) = RK4P4_(RHS_trK, vector(trK_p, i, j, k), vector(rk1trK, i, j, k), vector(rk2trK, i, j, k), vector(rk3trK, i, j, k), dx, simPlat_dt, ilast, jlast);
						vector(theta, i, j, k) = RK4P4_(RHS_theta, vector(theta_p, i, j, k), vector(rk1theta, i, j, k), vector(rk2theta, i, j, k), vector(rk3theta, i, j, k), dx, simPlat_dt, ilast, jlast);
					}
				}
			}
		}
	}
	//Fill ghosts and periodical boundaries
	time_interpolate_operator_mesh1->setStep(4);
	d_bdry_sched_advance23[ln]->fillData(current_time + simPlat_dt, false);
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		//Hard region field distance variables
		double* d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
	
		double* Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_id).get())->getPointer();
		double* Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_id).get())->getPointer();
		double* Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_id).get())->getPointer();
		double* Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_id).get())->getPointer();
		double* Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_id).get())->getPointer();
		double* Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_id).get())->getPointer();
		double* Df = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_id).get())->getPointer();
		double* DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_DYf_id).get())->getPointer();
		double* tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_tauf_id).get())->getPointer();
		double* phif = ((pdat::NodeData<double> *) patch->getPatchData(d_phif_id).get())->getPointer();
		double* gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_id).get())->getPointer();
		double* gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_id).get())->getPointer();
		double* gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_id).get())->getPointer();
		double* gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_id).get())->getPointer();
		double* gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_id).get())->getPointer();
		double* gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_id).get())->getPointer();
		double* Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xx_id).get())->getPointer();
		double* Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xy_id).get())->getPointer();
		double* Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xz_id).get())->getPointer();
		double* Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yy_id).get())->getPointer();
		double* Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yz_id).get())->getPointer();
		double* Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_zz_id).get())->getPointer();
		double* Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_x_id).get())->getPointer();
		double* Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_y_id).get())->getPointer();
		double* Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_z_id).get())->getPointer();
		double* Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_id).get())->getPointer();
		double* Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_id).get())->getPointer();
		double* Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_id).get())->getPointer();
		double* Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_id).get())->getPointer();
		double* chi = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_id).get())->getPointer();
		double* trK = ((pdat::NodeData<double> *) patch->getPatchData(d_trK_id).get())->getPointer();
		double* theta = ((pdat::NodeData<double> *) patch->getPatchData(d_theta_id).get())->getPointer();
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
						vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
						vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
						vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
						vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
						vector(Df, i, j, k) = vector(Df, i - 1, j, k);
						vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
						vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
						vector(phif, i, j, k) = vector(phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
						vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
						vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
						vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
						vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
						vector(Df, i, j, k) = vector(Df, i, j - 1, k);
						vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
						vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
						vector(phif, i, j, k) = vector(phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
						vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
						vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
						vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
						vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
						vector(Df, i, j, k) = vector(Df, i, j, k - 1);
						vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
						vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
						vector(phif, i, j, k) = vector(phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
						vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
						vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
						vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
						vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
						vector(Df, i, j, k) = vector(Df, i - 1, j, k);
						vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
						vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
						vector(phif, i, j, k) = vector(phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
						vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
						vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
						vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
						vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
						vector(Df, i, j, k) = vector(Df, i, j - 1, k);
						vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
						vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
						vector(phif, i, j, k) = vector(phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
						vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
						vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
						vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
						vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
						vector(Df, i, j, k) = vector(Df, i, j, k - 1);
						vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
						vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
						vector(phif, i, j, k) = vector(phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
						vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
						vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
						vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
						vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
						vector(Df, i, j, k) = vector(Df, i - 1, j, k);
						vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
						vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
						vector(phif, i, j, k) = vector(phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
						vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
						vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
						vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
						vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
						vector(Df, i, j, k) = vector(Df, i, j - 1, k);
						vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
						vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
						vector(phif, i, j, k) = vector(phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
						vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
						vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
						vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
						vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
						vector(Df, i, j, k) = vector(Df, i, j, k - 1);
						vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
						vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
						vector(phif, i, j, k) = vector(phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
						vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
						vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
						vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
						vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
						vector(Df, i, j, k) = vector(Df, i - 1, j, k);
						vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
						vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
						vector(phif, i, j, k) = vector(phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
						vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
						vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
						vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
						vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
						vector(Df, i, j, k) = vector(Df, i, j - 1, k);
						vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
						vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
						vector(phif, i, j, k) = vector(phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
						vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
						vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
						vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
						vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
						vector(Df, i, j, k) = vector(Df, i, j, k - 1);
						vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
						vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
						vector(phif, i, j, k) = vector(phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
						vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
						vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
						vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
						vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
						vector(Df, i, j, k) = vector(Df, i - 1, j, k);
						vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
						vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
						vector(phif, i, j, k) = vector(phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
						vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
						vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
						vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
						vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
						vector(Df, i, j, k) = vector(Df, i, j - 1, k);
						vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
						vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
						vector(phif, i, j, k) = vector(phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
						vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
						vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
						vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
						vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
						vector(Df, i, j, k) = vector(Df, i, j, k - 1);
						vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
						vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
						vector(phif, i, j, k) = vector(phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
						vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
						vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
						vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
						vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
						vector(Df, i, j, k) = vector(Df, i - 1, j, k);
						vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
						vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
						vector(phif, i, j, k) = vector(phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
						vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
						vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
						vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
						vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
						vector(Df, i, j, k) = vector(Df, i, j - 1, k);
						vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
						vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
						vector(phif, i, j, k) = vector(phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
						vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
						vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
						vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
						vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
						vector(Df, i, j, k) = vector(Df, i, j, k - 1);
						vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
						vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
						vector(phif, i, j, k) = vector(phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
						vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
						vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
						vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
						vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
						vector(Df, i, j, k) = vector(Df, i - 1, j, k);
						vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
						vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
						vector(phif, i, j, k) = vector(phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
						vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
						vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
						vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
						vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
						vector(Df, i, j, k) = vector(Df, i, j - 1, k);
						vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
						vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
						vector(phif, i, j, k) = vector(phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
						vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
						vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
						vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
						vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
						vector(Df, i, j, k) = vector(Df, i, j, k - 1);
						vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
						vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
						vector(phif, i, j, k) = vector(phif, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, d_ghost_width, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, d_ghost_width, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, d_ghost_width, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, d_ghost_width, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, d_ghost_width, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, d_ghost_width, j, k);
						vector(Df, i, j, k) = vector(Df, d_ghost_width, j, k);
						vector(DYf, i, j, k) = vector(DYf, d_ghost_width, j, k);
						vector(tauf, i, j, k) = vector(tauf, d_ghost_width, j, k);
						vector(phif, i, j, k) = vector(phif, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i - 1, j, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i - 1, j, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i - 1, j, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i - 1, j, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i - 1, j, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i - 1, j, k);
						vector(Df, i, j, k) = vector(Df, i - 1, j, k);
						vector(DYf, i, j, k) = vector(DYf, i - 1, j, k);
						vector(tauf, i, j, k) = vector(tauf, i - 1, j, k);
						vector(phif, i, j, k) = vector(phif, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, d_ghost_width, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, d_ghost_width, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, d_ghost_width, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, d_ghost_width, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, d_ghost_width, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, d_ghost_width, k);
						vector(Df, i, j, k) = vector(Df, i, d_ghost_width, k);
						vector(DYf, i, j, k) = vector(DYf, i, d_ghost_width, k);
						vector(tauf, i, j, k) = vector(tauf, i, d_ghost_width, k);
						vector(phif, i, j, k) = vector(phif, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j - 1, k);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j - 1, k);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j - 1, k);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j - 1, k);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j - 1, k);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j - 1, k);
						vector(Df, i, j, k) = vector(Df, i, j - 1, k);
						vector(DYf, i, j, k) = vector(DYf, i, j - 1, k);
						vector(tauf, i, j, k) = vector(tauf, i, j - 1, k);
						vector(phif, i, j, k) = vector(phif, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, d_ghost_width);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, d_ghost_width);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, d_ghost_width);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, d_ghost_width);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, d_ghost_width);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, d_ghost_width);
						vector(Df, i, j, k) = vector(Df, i, j, d_ghost_width);
						vector(DYf, i, j, k) = vector(DYf, i, j, d_ghost_width);
						vector(tauf, i, j, k) = vector(tauf, i, j, d_ghost_width);
						vector(phif, i, j, k) = vector(phif, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(Sfd_x, i, j, k) = vector(Sfd_x, i, j, k - 1);
						vector(Sfd_y, i, j, k) = vector(Sfd_y, i, j, k - 1);
						vector(Sfd_z, i, j, k) = vector(Sfd_z, i, j, k - 1);
						vector(Bfu_x, i, j, k) = vector(Bfu_x, i, j, k - 1);
						vector(Bfu_y, i, j, k) = vector(Bfu_y, i, j, k - 1);
						vector(Bfu_z, i, j, k) = vector(Bfu_z, i, j, k - 1);
						vector(Df, i, j, k) = vector(Df, i, j, k - 1);
						vector(DYf, i, j, k) = vector(DYf, i, j, k - 1);
						vector(tauf, i, j, k) = vector(tauf, i, j, k - 1);
						vector(phif, i, j, k) = vector(phif, i, j, k - 1);
					}
				}
			}
		}
	}
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		double* chi = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_id).get())->getPointer();
		double* Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_id).get())->getPointer();
		double* Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_id).get())->getPointer();
		double* tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_tauf_id).get())->getPointer();
		double* gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_id).get())->getPointer();
		double* gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_id).get())->getPointer();
		double* gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_id).get())->getPointer();
		double* gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_id).get())->getPointer();
		double* gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_id).get())->getPointer();
		double* gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_id).get())->getPointer();
		double* Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_id).get())->getPointer();
		double* Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_id).get())->getPointer();
		double* Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* Df = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_id).get())->getPointer();
		double* DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_DYf_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double chi_max, h, sdetg, Bu_x, Bu_y, Bu_z, tau, inv_chi, Bfd_y, Bfd_z, Bfd_x, decay_factor, detgtd, idetgtd, gtu_zz, gtu_xx, gtu_xz, gtu_xy, Sfu_x, Su_x, vfu_x, Sd_x, Sd_y, gtu_yy, Sd_z, gtu_yz, vfu_z, Sfu_y, Su_y, Sfu_z, Su_z, vfu_y, sqvf, sqW, W, Bd_x, D, Bd_y, Bd_z, rhof_tmp, Yef_tmp, Tf_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, dpfdye_tmp, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tauf_tmp, Df_tmp, DYf_tmp, Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, Efu_x_tmp, Efu_y_tmp, Efu_z_tmp, vacuum_D_tmp, vacuum_tau_tmp, vacuum_rho_reset_tmp;
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					chi_max = MAX(chi_floor, vector(chi, i, j, k));
					h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
					sdetg = pow(chi_max, (-1.500000000000000));
					Bu_x = vector(Bfu_x, i, j, k) / sdetg;
					Bu_y = vector(Bfu_y, i, j, k) / sdetg;
					Bu_z = vector(Bfu_z, i, j, k) / sdetg;
					tau = vector(tauf, i, j, k) / sdetg;
					inv_chi = 1.0 / fabs(chi_max);
					Bfd_y = inv_chi * (vector(gtd_xy, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_yz, i, j, k) * vector(Bfu_z, i, j, k));
					Bfd_z = inv_chi * (vector(gtd_xz, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_yz, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Bfu_z, i, j, k));
					Bfd_x = inv_chi * (vector(gtd_xx, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_xy, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_xz, i, j, k) * vector(Bfu_z, i, j, k));
					decay_factor = MIN(1.0, pow((R_0 / MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))), eta_damping_exp));
					detgtd = vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) * vector(gtd_xx, i, j, k) + (-vector(gtd_yy, i, j, k) * (vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k))) + (-vector(gtd_zz, i, j, k) * (vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k))) + (-(vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k)) * vector(gtd_xx, i, j, k)) + 2.0 * vector(gtd_yz, i, j, k) * vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k);
					idetgtd = 1.0 / detgtd;
					gtu_zz = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_yy, i, j, k) - vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k));
					gtu_xx = idetgtd * (vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k));
					gtu_xz = idetgtd * (vector(gtd_xy, i, j, k) * vector(gtd_yz, i, j, k) - vector(gtd_yy, i, j, k) * vector(gtd_xz, i, j, k));
					gtu_xy = idetgtd * ((-vector(gtd_xy, i, j, k) * vector(gtd_zz, i, j, k)) + vector(gtd_yz, i, j, k) * vector(gtd_xz, i, j, k));
					Sfu_x = chi_max * (gtu_xx * vector(Sfd_x, i, j, k) + gtu_xy * vector(Sfd_y, i, j, k) + gtu_xz * vector(Sfd_z, i, j, k));
					Su_x = Sfu_x / sdetg;
					vfu_x = chi_max * (gtu_xx * vector(vfd_x, i, j, k) + gtu_xy * vector(vfd_y, i, j, k) + gtu_xz * vector(vfd_z, i, j, k));
					Sd_x = vector(Sfd_x, i, j, k) / sdetg;
					Sd_y = vector(Sfd_y, i, j, k) / sdetg;
					gtu_yy = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k));
					Sd_z = vector(Sfd_z, i, j, k) / sdetg;
					gtu_yz = idetgtd * ((-vector(gtd_xx, i, j, k) * vector(gtd_yz, i, j, k)) + vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k));
					vfu_z = chi_max * (gtu_xz * vector(vfd_x, i, j, k) + gtu_yz * vector(vfd_y, i, j, k) + gtu_zz * vector(vfd_z, i, j, k));
					Sfu_y = chi_max * (gtu_xy * vector(Sfd_x, i, j, k) + gtu_yy * vector(Sfd_y, i, j, k) + gtu_yz * vector(Sfd_z, i, j, k));
					Su_y = Sfu_y / sdetg;
					Sfu_z = chi_max * (gtu_xz * vector(Sfd_x, i, j, k) + gtu_yz * vector(Sfd_y, i, j, k) + gtu_zz * vector(Sfd_z, i, j, k));
					Su_z = Sfu_z / sdetg;
					vfu_y = chi_max * (gtu_xy * vector(vfd_x, i, j, k) + gtu_yy * vector(vfd_y, i, j, k) + gtu_yz * vector(vfd_z, i, j, k));
					sqvf = MIN(vfu_x * vector(vfd_x, i, j, k) + vfu_y * vector(vfd_y, i, j, k) + vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
					sqW = 1.0 / (1.0 - sqvf);
					W = sqrt(sqW);
					Bd_x = Bfd_x / sdetg;
					D = vector(Df, i, j, k) / sdetg;
					Bd_y = Bfd_y / sdetg;
					Bd_z = Bfd_z / sdetg;
					vector(t_chie, i, j, k) = vector(chie, i, j, k);
					vector(t_chia, i, j, k) = vector(chia, i, j, k);
					vector(t_optdepthx, i, j, k) = vector(optdepthx, i, j, k);
					vector(t_optdeptha, i, j, k) = vector(optdeptha, i, j, k);
					vector(t_optdepthe, i, j, k) = vector(optdepthe, i, j, k);
					vector(Alpha, i, j, k) = fabs(vector(Alpha, i, j, k));
					rhof_tmp = vector(rhof, i, j, k);
					Yef_tmp = vector(DYf, i, j, k) / vector(Df, i, j, k);
					if (Yef_tmp < vacuum_ye) {
						vector(DYf, i, j, k) = vacuum_ye_reset * vector(Df, i, j, k);
						Yef_tmp = vacuum_ye_reset;
					}
					if (Yef_tmp > ye_maximum) {
						vector(DYf, i, j, k) = 0.9 * ye_maximum * vector(Df, i, j, k);
						Yef_tmp = 0.9 * ye_maximum;
					}
					Tf_tmp = vector(Tf, i, j, k);
					pf_tmp = vector(pf, i, j, k);
					epsf_tmp = vector(epsf, i, j, k);
					vfd_x_tmp = vector(vfd_x, i, j, k);
					vfd_y_tmp = vector(vfd_y, i, j, k);
					vfd_z_tmp = vector(vfd_z, i, j, k);
					sqcs_tmp = vector(sqcs, i, j, k);
					dpfdrho_tmp = vector(dpfdrho, i, j, k);
					dpfdeps_tmp = vector(dpfdeps, i, j, k);
					dpfdye_tmp = vector(dpfdye, i, j, k);
					Sfd_x_tmp = vector(Sfd_x, i, j, k);
					Sfd_y_tmp = vector(Sfd_y, i, j, k);
					Sfd_z_tmp = vector(Sfd_z, i, j, k);
					tauf_tmp = vector(tauf, i, j, k);
					Df_tmp = vector(Df, i, j, k);
					DYf_tmp = vector(DYf, i, j, k);
					Bfu_x_tmp = vector(Bfu_x, i, j, k);
					Bfu_y_tmp = vector(Bfu_y, i, j, k);
					Bfu_z_tmp = vector(Bfu_z, i, j, k);
					if (equalsEq(externalCon2Prim, 1.0)) {
						Efu_x_tmp = vector(Efu_x, i, j, k);
						Efu_y_tmp = vector(Efu_y, i, j, k);
						Efu_z_tmp = vector(Efu_z, i, j, k);
						externalCon2prim(Efu_x_tmp, Efu_y_tmp, Efu_z_tmp, rhof_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, Df_tmp, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tauf_tmp, W, vector(chi, i, j, k), gtu_xx, gtu_xy, gtu_xz, gtu_yy, gtu_yz, gtu_zz, vector(gtd_xx, i, j, k), vector(gtd_xy, i, j, k), vector(gtd_xz, i, j, k), vector(gtd_yy, i, j, k), vector(gtd_yz, i, j, k), vector(gtd_zz, i, j, k), Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, dx, simPlat_dt, ilast, jlast);
						vector(Efu_x, i, j, k) = Efu_x_tmp;
						vector(Efu_y, i, j, k) = Efu_y_tmp;
						vector(Efu_z, i, j, k) = Efu_z_tmp;
					} else {
						vacuum_D_tmp = vacuum_D * decay_factor;
						vacuum_tau_tmp = vacuum_tau * decay_factor;
						vacuum_rho_reset_tmp = vacuum_rho_reset * decay_factor;
						con2prim(eos_type, rhof_tmp, Yef_tmp, Tf_tmp, pf_tmp, epsf_tmp, vfd_x_tmp, vfd_y_tmp, vfd_z_tmp, sqcs_tmp, dpfdrho_tmp, dpfdeps_tmp, dpfdye_tmp, rho_0, rho_1, rho_2, a_0, a_1, a_2, a_3, K_0, K_1, K_2, K_3, gamma_0, gamma_1, gamma_2, gamma_3, D, vacuum_D_tmp, Df_tmp, DYf_tmp, sdetg, Sd_x, Sd_y, Sd_z, Su_x, Su_y, Su_z, Sfd_x_tmp, Sfd_y_tmp, Sfd_z_tmp, tau, vacuum_tau_tmp, tauf_tmp, W, h, vector(chi, i, j, k), gtu_xx, gtu_xy, gtu_xz, gtu_yy, gtu_yz, gtu_zz, vector(gtd_xx, i, j, k), vector(gtd_xy, i, j, k), vector(gtd_xz, i, j, k), vector(gtd_yy, i, j, k), vector(gtd_yz, i, j, k), vector(gtd_zz, i, j, k), Bd_x, Bd_y, Bd_z, Bu_x, Bu_y, Bu_z, Bfu_x_tmp, Bfu_y_tmp, Bfu_z_tmp, max_error, gamma, inv_chi, threshold_vacuum, threshold_sqSmax, threshold_sqBmax, energyShift, vacuum_ye_beta, vacuum_rho, vacuum_rho_reset_tmp, vacuum_P_reset, vacuum_ye_reset, vacuum_temp_reset, vacuum_tau_reset, minTableTemperature, minTableEnergy, coord0_data_Data, coord0_dx_Data, coord0_size_Data, coord1_data_Data, coord1_dx_Data, coord1_size_Data, coord2_data_Data, coord2_dx_Data, coord2_size_Data, vars_data_Data, vars_size_Data, dx, simPlat_dt, ilast, jlast);
						vector(Efu_x, i, j, k) = -((-vector(vfd_z, i, j, k) * Bfd_y) + vector(vfd_y, i, j, k) * Bfd_z) / (sdetg * sdetg);
						vector(Efu_y, i, j, k) = -(vector(vfd_z, i, j, k) * Bfd_x - vector(vfd_x, i, j, k) * Bfd_z) / (sdetg * sdetg);
						vector(Efu_z, i, j, k) = -((-vector(vfd_y, i, j, k) * Bfd_x) + vector(vfd_x, i, j, k) * Bfd_y) / (sdetg * sdetg);
					}
					vector(rhof, i, j, k) = rhof_tmp;
					vector(Yef, i, j, k) = Yef_tmp;
					vector(Tf, i, j, k) = Tf_tmp;
					vector(pf, i, j, k) = pf_tmp;
					vector(epsf, i, j, k) = epsf_tmp;
					vector(vfd_x, i, j, k) = vfd_x_tmp;
					vector(vfd_y, i, j, k) = vfd_y_tmp;
					vector(vfd_z, i, j, k) = vfd_z_tmp;
					vector(sqcs, i, j, k) = sqcs_tmp;
					vector(dpfdye, i, j, k) = dpfdye_tmp;
					vector(dpfdeps, i, j, k) = dpfdeps_tmp;
					vector(dpfdrho, i, j, k) = dpfdrho_tmp;
					vector(Sfd_x, i, j, k) = Sfd_x_tmp;
					vector(Sfd_y, i, j, k) = Sfd_y_tmp;
					vector(Sfd_z, i, j, k) = Sfd_z_tmp;
					vector(tauf, i, j, k) = tauf_tmp;
					vector(Df, i, j, k) = Df_tmp;
					vector(DYf, i, j, k) = DYf_tmp;
					vector(Bfu_x, i, j, k) = Bfu_x_tmp;
					vector(Bfu_y, i, j, k) = Bfu_y_tmp;
					vector(Bfu_z, i, j, k) = Bfu_z_tmp;
					vector(t_chix, i, j, k) = vector(chix, i, j, k);
				}
			}
		}
	}
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* DYf = ((pdat::NodeData<double> *) patch->getPatchData(d_DYf_id).get())->getPointer();
		double* Df = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_id).get())->getPointer();
		double* Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_id).get())->getPointer();
		double* Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_id).get())->getPointer();
		double* Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* chi = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_id).get())->getPointer();
		double* gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_id).get())->getPointer();
		double* gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_id).get())->getPointer();
		double* gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_id).get())->getPointer();
		double* gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_id).get())->getPointer();
		double* gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_id).get())->getPointer();
		double* gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_id).get())->getPointer();
		double* Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_id).get())->getPointer();
		double* Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_id).get())->getPointer();
		double* Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_id).get())->getPointer();
		double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
		double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
		double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
		double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
		double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
		double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
		double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
		double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
		double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
		double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
		double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
		double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
		double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
		double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
		double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
		double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
		double* qnu_e = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_e_id).get())->getPointer();
		double* qnu_a = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_a_id).get())->getPointer();
		double* qnu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_x_id).get())->getPointer();
		double* Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_id).get())->getPointer();
		double* Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_id).get())->getPointer();
		double* Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_id).get())->getPointer();
		double* Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double d_optdeptheh_o0_t0_m0_l0, d_optdepthev_o0_t0_m0_l0, d_optdepthed_o0_t0_m0_l0, d_optdepthah_o0_t0_m0_l0, d_optdepthav_o0_t0_m0_l0, d_optdepthad_o0_t0_m0_l0, d_optdepthxh_o0_t0_m0_l0, d_optdepthxv_o0_t0_m0_l0, d_optdepthxd_o0_t0_m0_l0, d_chieh_o0_t0_m0_l0, d_chiev_o0_t0_m0_l0, d_chied_o0_t0_m0_l0, d_chiah_o0_t0_m0_l0, d_chiav_o0_t0_m0_l0, d_chiad_o0_t0_m0_l0, d_chixh_o0_t0_m0_l0, d_chixv_o0_t0_m0_l0, d_chixd_o0_t0_m0_l0, d_CovdDYf_z_o0_t0_m0_l0, d_CovdDYf_y_o0_t0_m0_l0, d_CovdDYf_x_o0_t0_m0_l0, d_CovdDf_z_o0_t0_m0_l0, d_CovdDf_y_o0_t0_m0_l0, d_CovdDf_x_o0_t0_m0_l0, d_CovdBfu_xx_o0_t0_m0_l0, d_CovdBfu_xy_o0_t0_m0_l0, d_CovdBfu_xz_o0_t0_m0_l0, d_CovdBfu_yx_o0_t0_m0_l0, d_CovdBfu_yy_o0_t0_m0_l0, d_CovdBfu_yz_o0_t0_m0_l0, d_CovdBfu_zx_o0_t0_m0_l0, d_CovdBfu_zy_o0_t0_m0_l0, d_CovdBfu_zz_o0_t0_m0_l0, d_CovdEfu_xx_o0_t0_m0_l0, d_CovdEfu_xy_o0_t0_m0_l0, d_CovdEfu_xz_o0_t0_m0_l0, d_CovdEfu_yx_o0_t0_m0_l0, d_CovdEfu_yy_o0_t0_m0_l0, d_CovdEfu_yz_o0_t0_m0_l0, d_CovdEfu_zx_o0_t0_m0_l0, d_CovdEfu_zy_o0_t0_m0_l0, d_CovdEfu_zz_o0_t0_m0_l0, d_Covdepsf_x_o0_t0_m0_l0, d_Covdepsf_y_o0_t0_m0_l0, d_Covdepsf_z_o0_t0_m0_l0, d_Covddpfdeps_x_o0_t0_m0_l0, d_Covddpfdeps_y_o0_t0_m0_l0, d_Covddpfdeps_z_o0_t0_m0_l0, d_Covddpfdye_x_o0_t0_m0_l0, d_Covddpfdye_y_o0_t0_m0_l0, d_Covddpfdye_z_o0_t0_m0_l0, d_Covddpfdrho_x_o0_t0_m0_l0, d_Covddpfdrho_y_o0_t0_m0_l0, d_Covddpfdrho_z_o0_t0_m0_l0, d_Covdvfd_xx_o0_t0_m0_l0, d_Covdvfd_xy_o0_t0_m0_l0, d_Covdvfd_xz_o0_t0_m0_l0, d_Covdvfd_yx_o0_t0_m0_l0, d_Covdvfd_yy_o0_t0_m0_l0, d_Covdvfd_yz_o0_t0_m0_l0, d_Covdvfd_zx_o0_t0_m0_l0, d_Covdvfd_zy_o0_t0_m0_l0, d_Covdvfd_zz_o0_t0_m0_l0, d_Covdrhof_x_o0_t0_m0_l0, d_Covdrhof_y_o0_t0_m0_l0, d_Covdrhof_z_o0_t0_m0_l0, d_Covdpf_x_o0_t0_m0_l0, d_Covdpf_y_o0_t0_m0_l0, d_Covdpf_z_o0_t0_m0_l0, d_CovdYef_x_o0_t0_m0_l0, d_CovdYef_y_o0_t0_m0_l0, d_CovdYef_z_o0_t0_m0_l0, chiev, chieh, Covdvfd_xx, Covdvfd_xz, Covdvfd_xy, chied, Covddpfdeps_x, Covddpfdeps_y, Covddpfdeps_z, Covdvfd_zx, Covdvfd_zz, Covdvfd_zy, Covdvfd_yy, Covdvfd_yx, Covdvfd_yz, Covdpf_x, Covdpf_y, Covdpf_z, chi_max, inv_chi, Bfd_y, Bfd_z, Bfd_x, sqBf, chixh, chixd, Covddpfdrho_z, Covddpfdrho_y, Covddpfdrho_x, sdetg, m_CovdDf_x_o0_t0_l0, CovdDf_x, m_CovdDf_y_o0_t0_l0, CovdDf_y, m_CovdDf_z_o0_t0_l0, CovdDf_z, m_CovdBfu_zz_o0_t0_l0, CovdBfu_zz, m_CovdBfu_zy_o0_t0_l0, CovdBfu_zy, m_CovdBfu_zx_o0_t0_l0, CovdBfu_zx, m_CovdBfu_yz_o0_t0_l0, CovdBfu_yz, m_CovdBfu_yy_o0_t0_l0, CovdBfu_yy, m_CovdBfu_yx_o0_t0_l0, CovdBfu_yx, m_CovdBfu_xy_o0_t0_l0, CovdBfu_xy, m_CovdBfu_xx_o0_t0_l0, CovdBfu_xx, m_CovdBfu_xz_o0_t0_l0, CovdBfu_xz, CovdBfd_zx, Covdrhof_z, Covdrhof_y, Covdrhof_x, CovdBfd_zz, CovdBfd_zy, m_CovdDYf_x_o0_t0_l0, CovdDYf_x, m_CovdDYf_y_o0_t0_l0, CovdDYf_y, m_CovdDYf_z_o0_t0_l0, CovdDYf_z, Efd_z, Efd_y, chixv, Efd_x, sqEf, optdepthed, CovdBfd_yy, CovdBfd_yx, CovdBfd_yz, optdeptheh, CovdBvf_z, optdepthev, CovdBvf_y, CovdBvf_x, CovdBfd_xx, CovdBfd_xz, CovdBfd_xy, optdepthxd, optdepthxh, optdepthxv, chiav, CovdEfu_zz, CovdEfu_zy, CovdEfu_zx, CovdEfd_zz, CovdEfd_zx, CovdEfd_zy, Bvf, optdepthah, optdepthad, h, chiah, chiad, CovdYef_y, optdepthav, CovdYef_z, CovdYef_x, detgtd, idetgtd, gtu_zz, gtu_yy, gtu_yz, gtu_xx, gtu_xz, CovuBfu_zx, CovuBfu_zz, CovuBfu_zy, Covdvfu_xz, CovuYef_z, Covdvfu_zz, Covdvfu_yz, Covuvfu_zz, Sfu_z, Covurhof_z, CovuBvf_z, gtu_xy, CovuBfu_yz, CovuBfu_yy, CovuBfu_yx, Covdvfu_xy, Covdvfu_xx, CovdinvsqW_x, CovuYef_x, CovuYef_y, Covdvfu_zy, Covdvfu_zx, CovdinvsqW_z, Covdvfu_yx, Covuvfu_zx, Covdvfu_yy, CovdinvsqW_y, CovuinvsqW_z, Covuvfu_zy, CovuBfu_xz, CovuBfu_xy, CovuBfu_xx, Sfu_y, Sfu_x, Covurhof_x, Covurhof_y, Covuvfu_yz, Covuvfu_yy, Covuvfu_yx, Covuvfu_xz, Covuvfu_xy, Covuvfu_xx, vfu_y, vfu_x, CovuBvf_x, CovuBvf_y, CovuinvsqW_y, CovuinvsqW_x, Covddpfdye_z, Covddpfdye_y, Covddpfdye_x, CovdEfu_xz, CovdEfu_xy, CovdEfu_xx, CovdEfd_xx, CovdEfd_xy, CovdEfd_xz, vfu_z, sqvf, sqW, PhiAh, Epf, Thetaf, Covdepsf_z, Covdhf_z, CovdEpf_z, CovdThetaf_z, Covdepsf_y, Covdhf_y, CovdEpf_y, CovdThetaf_y, Covdepsf_x, Covuepsf_z, Covuepsf_y, Covdhf_x, CovdEpf_x, CovdThetaf_x, CovuThetaf_z, CovuThetaf_x, CovuThetaf_y, Phivh_z, Phivh_x, Phivh_y, PhiMh_yz, PhiMh_xz, PhiMh_xy, Covuepsf_x, faceta, CovdEfu_yz, CovuEfu_zz, CovuEfu_xz, CovdEfu_yy, CovuEfu_zy, CovuEfu_xy, CovdEfu_yx, CovdEfd_yy, CovdEfd_yz, CovdEfd_yx, CovuEfu_zx, CovuEfu_xx, CovuEfu_yx, CovuEfu_yz, CovuEfu_yy, PhiThetah, HPres, HTheta, Hv_z, HN_z, HNe_z, Hv_y, HN_y, HM_yz, HNe_y, Hv_x, HN_x, HM_xz, HM_xy, HE_z, HE_x, HE_y, HT_zz, HT_yy, HT_yz, HT_xx, HT_xz, HT_xy, HNe_x, d_TauN_x_o0_t0_m0_l0, d_TauN_y_o0_t0_m0_l0, d_TauN_z_o0_t0_m0_l0, d_TauNe_x_o0_t0_m0_l0, d_TauNe_y_o0_t0_m0_l0, d_TauNe_z_o0_t0_m0_l0, d_TauM_xy_o0_t0_m0_l0, d_TauM_xz_o0_t0_m0_l0, d_TauM_yz_o0_t0_m0_l0, d_TauT_xx_o0_t0_m0_l0, d_TauT_xy_o0_t0_m0_l0, d_TauT_xz_o0_t0_m0_l0, d_TauT_yy_o0_t0_m0_l0, d_TauT_yz_o0_t0_m0_l0, d_TauT_zz_o0_t0_m0_l0, optdepthe_tmp, optdeptha_tmp, optdepthx_tmp, chie_tmp, chia_tmp, chix_tmp, qnu_tmp, rnu_tmp, qnu_e_tmp, qnu_a_tmp, qnu_x_tmp, betabeta, betav, velr, f_redshift;
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
		for(int k = 0; k < klast; k++) {
			for(int j = 0; j < jlast; j++) {
				for(int i = 0; i < ilast; i++) {
					if ((i + 2 < ilast && i - 2 >= 0 && j + 2 < jlast && j - 2 >= 0 && k + 2 < klast && k - 2 >= 0) && ((i + 3 < ilast || !patch->getPatchGeometry()->getTouchesRegularBoundary(0, 1)) && (i - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(0, 0)) && (j + 3 < jlast || !patch->getPatchGeometry()->getTouchesRegularBoundary(1, 1)) && (j - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(1, 0)) && (k + 3 < klast || !patch->getPatchGeometry()->getTouchesRegularBoundary(2, 1)) && (k - 3 >= 0 || !patch->getPatchGeometry()->getTouchesRegularBoundary(2, 0)))) {
						d_optdeptheh_o0_t0_m0_l0 = eikonal_i(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthev_o0_t0_m0_l0 = eikonal_j(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthed_o0_t0_m0_l0 = eikonal_k(t_optdepthe, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthah_o0_t0_m0_l0 = eikonal_i(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthav_o0_t0_m0_l0 = eikonal_j(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthad_o0_t0_m0_l0 = eikonal_k(t_optdeptha, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthxh_o0_t0_m0_l0 = eikonal_i(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthxv_o0_t0_m0_l0 = eikonal_j(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_optdepthxd_o0_t0_m0_l0 = eikonal_k(t_optdepthx, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chieh_o0_t0_m0_l0 = eikonal_i(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiev_o0_t0_m0_l0 = eikonal_j(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chied_o0_t0_m0_l0 = eikonal_k(t_chie, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiah_o0_t0_m0_l0 = eikonal_i(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiav_o0_t0_m0_l0 = eikonal_j(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chiad_o0_t0_m0_l0 = eikonal_k(t_chia, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chixh_o0_t0_m0_l0 = eikonal_i(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chixv_o0_t0_m0_l0 = eikonal_j(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
						d_chixd_o0_t0_m0_l0 = eikonal_k(t_chix, i, j, k, dx, simPlat_dt, ilast, jlast);
						if (calculate_les_terms) {
							d_CovdDYf_z_o0_t0_m0_l0 = D1CDO4_k(DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDYf_y_o0_t0_m0_l0 = D1CDO4_j(DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDYf_x_o0_t0_m0_l0 = D1CDO4_i(DYf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDf_z_o0_t0_m0_l0 = D1CDO4_k(Df, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDf_y_o0_t0_m0_l0 = D1CDO4_j(Df, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdDf_x_o0_t0_m0_l0 = D1CDO4_i(Df, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_xx_o0_t0_m0_l0 = D1CDO4_i(Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_xy_o0_t0_m0_l0 = D1CDO4_i(Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_xz_o0_t0_m0_l0 = D1CDO4_i(Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_yx_o0_t0_m0_l0 = D1CDO4_j(Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_yy_o0_t0_m0_l0 = D1CDO4_j(Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_yz_o0_t0_m0_l0 = D1CDO4_j(Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_zx_o0_t0_m0_l0 = D1CDO4_k(Bfu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_zy_o0_t0_m0_l0 = D1CDO4_k(Bfu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdBfu_zz_o0_t0_m0_l0 = D1CDO4_k(Bfu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_xx_o0_t0_m0_l0 = D1CDO4_i(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_xy_o0_t0_m0_l0 = D1CDO4_i(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_xz_o0_t0_m0_l0 = D1CDO4_i(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_yx_o0_t0_m0_l0 = D1CDO4_j(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_yy_o0_t0_m0_l0 = D1CDO4_j(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_yz_o0_t0_m0_l0 = D1CDO4_j(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_zx_o0_t0_m0_l0 = D1CDO4_k(Efu_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_zy_o0_t0_m0_l0 = D1CDO4_k(Efu_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdEfu_zz_o0_t0_m0_l0 = D1CDO4_k(Efu_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdepsf_x_o0_t0_m0_l0 = D1CDO4_i(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdepsf_y_o0_t0_m0_l0 = D1CDO4_j(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdepsf_z_o0_t0_m0_l0 = D1CDO4_k(epsf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdeps_x_o0_t0_m0_l0 = D1CDO4_i(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdeps_y_o0_t0_m0_l0 = D1CDO4_j(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdeps_z_o0_t0_m0_l0 = D1CDO4_k(dpfdeps, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdye_x_o0_t0_m0_l0 = D1CDO4_i(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdye_y_o0_t0_m0_l0 = D1CDO4_j(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdye_z_o0_t0_m0_l0 = D1CDO4_k(dpfdye, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdrho_x_o0_t0_m0_l0 = D1CDO4_i(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdrho_y_o0_t0_m0_l0 = D1CDO4_j(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covddpfdrho_z_o0_t0_m0_l0 = D1CDO4_k(dpfdrho, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_xx_o0_t0_m0_l0 = D1CDO4_i(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_xy_o0_t0_m0_l0 = D1CDO4_i(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_xz_o0_t0_m0_l0 = D1CDO4_i(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_yx_o0_t0_m0_l0 = D1CDO4_j(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_yy_o0_t0_m0_l0 = D1CDO4_j(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_yz_o0_t0_m0_l0 = D1CDO4_j(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_zx_o0_t0_m0_l0 = D1CDO4_k(vfd_x, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_zy_o0_t0_m0_l0 = D1CDO4_k(vfd_y, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdvfd_zz_o0_t0_m0_l0 = D1CDO4_k(vfd_z, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdrhof_x_o0_t0_m0_l0 = D1CDO4_i(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdrhof_y_o0_t0_m0_l0 = D1CDO4_j(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdrhof_z_o0_t0_m0_l0 = D1CDO4_k(rhof, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdpf_x_o0_t0_m0_l0 = D1CDO4_i(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdpf_y_o0_t0_m0_l0 = D1CDO4_j(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_Covdpf_z_o0_t0_m0_l0 = D1CDO4_k(pf, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdYef_x_o0_t0_m0_l0 = D1CDO4_i(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdYef_y_o0_t0_m0_l0 = D1CDO4_j(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
							d_CovdYef_z_o0_t0_m0_l0 = D1CDO4_k(Yef, i, j, k, dx, simPlat_dt, ilast, jlast);
						}
						chiev = d_chiev_o0_t0_m0_l0;
						chieh = d_chieh_o0_t0_m0_l0;
						if (calculate_les_terms) {
							Covdvfd_xx = d_Covdvfd_xx_o0_t0_m0_l0;
							Covdvfd_xz = d_Covdvfd_xz_o0_t0_m0_l0;
							Covdvfd_xy = d_Covdvfd_xy_o0_t0_m0_l0;
						} else {
							Covdvfd_xx = 0.0;
							Covdvfd_xz = 0.0;
							Covdvfd_xy = 0.0;
						}
						chied = d_chied_o0_t0_m0_l0;
						if (calculate_les_terms) {
							Covddpfdeps_x = d_Covddpfdeps_x_o0_t0_m0_l0;
							Covddpfdeps_y = d_Covddpfdeps_y_o0_t0_m0_l0;
							Covddpfdeps_z = d_Covddpfdeps_z_o0_t0_m0_l0;
							Covdvfd_zx = d_Covdvfd_zx_o0_t0_m0_l0;
							Covdvfd_zz = d_Covdvfd_zz_o0_t0_m0_l0;
							Covdvfd_zy = d_Covdvfd_zy_o0_t0_m0_l0;
							Covdvfd_yy = d_Covdvfd_yy_o0_t0_m0_l0;
							Covdvfd_yx = d_Covdvfd_yx_o0_t0_m0_l0;
							Covdvfd_yz = d_Covdvfd_yz_o0_t0_m0_l0;
							Covdpf_x = d_Covdpf_x_o0_t0_m0_l0;
							Covdpf_y = d_Covdpf_y_o0_t0_m0_l0;
							Covdpf_z = d_Covdpf_z_o0_t0_m0_l0;
						} else {
							Covddpfdeps_x = 0.0;
							Covddpfdeps_y = 0.0;
							Covddpfdeps_z = 0.0;
							Covdvfd_zx = 0.0;
							Covdvfd_zz = 0.0;
							Covdvfd_zy = 0.0;
							Covdvfd_yy = 0.0;
							Covdvfd_yx = 0.0;
							Covdvfd_yz = 0.0;
							Covdpf_x = 0.0;
							Covdpf_y = 0.0;
							Covdpf_z = 0.0;
						}
						chi_max = MAX(chi_floor, vector(chi, i, j, k));
						inv_chi = 1.0 / fabs(chi_max);
						Bfd_y = inv_chi * (vector(gtd_xy, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_yz, i, j, k) * vector(Bfu_z, i, j, k));
						Bfd_z = inv_chi * (vector(gtd_xz, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_yz, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Bfu_z, i, j, k));
						Bfd_x = inv_chi * (vector(gtd_xx, i, j, k) * vector(Bfu_x, i, j, k) + vector(gtd_xy, i, j, k) * vector(Bfu_y, i, j, k) + vector(gtd_xz, i, j, k) * vector(Bfu_z, i, j, k));
						sqBf = vector(Bfu_x, i, j, k) * Bfd_x + vector(Bfu_y, i, j, k) * Bfd_y + vector(Bfu_z, i, j, k) * Bfd_z;
						chixh = d_chixh_o0_t0_m0_l0;
						chixd = d_chixd_o0_t0_m0_l0;
						if (calculate_les_terms) {
							Covddpfdrho_z = d_Covddpfdrho_z_o0_t0_m0_l0;
							Covddpfdrho_y = d_Covddpfdrho_y_o0_t0_m0_l0;
							Covddpfdrho_x = d_Covddpfdrho_x_o0_t0_m0_l0;
						} else {
							Covddpfdrho_z = 0.0;
							Covddpfdrho_y = 0.0;
							Covddpfdrho_x = 0.0;
						}
						sdetg = pow(chi_max, (-1.500000000000000));
						if (calculate_les_terms) {
							m_CovdDf_x_o0_t0_l0 = 1.0 / sdetg * d_CovdDf_x_o0_t0_m0_l0;
							CovdDf_x = m_CovdDf_x_o0_t0_l0;
							m_CovdDf_y_o0_t0_l0 = 1.0 / sdetg * d_CovdDf_y_o0_t0_m0_l0;
							CovdDf_y = m_CovdDf_y_o0_t0_l0;
							m_CovdDf_z_o0_t0_l0 = 1.0 / sdetg * d_CovdDf_z_o0_t0_m0_l0;
							CovdDf_z = m_CovdDf_z_o0_t0_l0;
							m_CovdBfu_zz_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_zz_o0_t0_m0_l0;
							CovdBfu_zz = m_CovdBfu_zz_o0_t0_l0;
							m_CovdBfu_zy_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_zy_o0_t0_m0_l0;
							CovdBfu_zy = m_CovdBfu_zy_o0_t0_l0;
							m_CovdBfu_zx_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_zx_o0_t0_m0_l0;
							CovdBfu_zx = m_CovdBfu_zx_o0_t0_l0;
							m_CovdBfu_yz_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_yz_o0_t0_m0_l0;
							CovdBfu_yz = m_CovdBfu_yz_o0_t0_l0;
							m_CovdBfu_yy_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_yy_o0_t0_m0_l0;
							CovdBfu_yy = m_CovdBfu_yy_o0_t0_l0;
							m_CovdBfu_yx_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_yx_o0_t0_m0_l0;
							CovdBfu_yx = m_CovdBfu_yx_o0_t0_l0;
							m_CovdBfu_xy_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_xy_o0_t0_m0_l0;
							CovdBfu_xy = m_CovdBfu_xy_o0_t0_l0;
							m_CovdBfu_xx_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_xx_o0_t0_m0_l0;
							CovdBfu_xx = m_CovdBfu_xx_o0_t0_l0;
							m_CovdBfu_xz_o0_t0_l0 = 1.0 / sdetg * d_CovdBfu_xz_o0_t0_m0_l0;
							CovdBfu_xz = m_CovdBfu_xz_o0_t0_l0;
							CovdBfd_zx = inv_chi * (vector(gtd_xx, i, j, k) * CovdBfu_zx + vector(gtd_xy, i, j, k) * CovdBfu_zy + vector(gtd_xz, i, j, k) * CovdBfu_zz);
							Covdrhof_z = d_Covdrhof_z_o0_t0_m0_l0;
							Covdrhof_y = d_Covdrhof_y_o0_t0_m0_l0;
							Covdrhof_x = d_Covdrhof_x_o0_t0_m0_l0;
							CovdBfd_zz = inv_chi * (vector(gtd_xz, i, j, k) * CovdBfu_zx + vector(gtd_yz, i, j, k) * CovdBfu_zy + vector(gtd_zz, i, j, k) * CovdBfu_zz);
							CovdBfd_zy = inv_chi * (vector(gtd_xy, i, j, k) * CovdBfu_zx + vector(gtd_yy, i, j, k) * CovdBfu_zy + vector(gtd_yz, i, j, k) * CovdBfu_zz);
							m_CovdDYf_x_o0_t0_l0 = 1.0 / sdetg * d_CovdDYf_x_o0_t0_m0_l0;
							CovdDYf_x = m_CovdDYf_x_o0_t0_l0;
							m_CovdDYf_y_o0_t0_l0 = 1.0 / sdetg * d_CovdDYf_y_o0_t0_m0_l0;
							CovdDYf_y = m_CovdDYf_y_o0_t0_l0;
							m_CovdDYf_z_o0_t0_l0 = 1.0 / sdetg * d_CovdDYf_z_o0_t0_m0_l0;
							CovdDYf_z = m_CovdDYf_z_o0_t0_l0;
						} else {
							CovdDf_x = 0.0;
							CovdDf_y = 0.0;
							CovdDf_z = 0.0;
							CovdBfu_zz = 0.0;
							CovdBfu_zy = 0.0;
							CovdBfu_zx = 0.0;
							CovdBfu_yz = 0.0;
							CovdBfu_yy = 0.0;
							CovdBfu_yx = 0.0;
							CovdBfu_xy = 0.0;
							CovdBfu_xx = 0.0;
							CovdBfu_xz = 0.0;
							CovdBfd_zx = 0.0;
							Covdrhof_z = 0.0;
							Covdrhof_y = 0.0;
							Covdrhof_x = 0.0;
							CovdBfd_zz = 0.0;
							CovdBfd_zy = 0.0;
							CovdDYf_x = 0.0;
							CovdDYf_y = 0.0;
							CovdDYf_z = 0.0;
						}
						Efd_z = inv_chi * (vector(gtd_xz, i, j, k) * vector(Efu_x, i, j, k) + vector(gtd_yz, i, j, k) * vector(Efu_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Efu_z, i, j, k));
						Efd_y = inv_chi * (vector(gtd_xy, i, j, k) * vector(Efu_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Efu_y, i, j, k) + vector(gtd_yz, i, j, k) * vector(Efu_z, i, j, k));
						chixv = d_chixv_o0_t0_m0_l0;
						Efd_x = inv_chi * (vector(gtd_xx, i, j, k) * vector(Efu_x, i, j, k) + vector(gtd_xy, i, j, k) * vector(Efu_y, i, j, k) + vector(gtd_xz, i, j, k) * vector(Efu_z, i, j, k));
						sqEf = vector(Efu_x, i, j, k) * Efd_x + vector(Efu_y, i, j, k) * Efd_y + vector(Efu_z, i, j, k) * Efd_z;
						optdepthed = d_optdepthed_o0_t0_m0_l0;
						if (calculate_les_terms) {
							CovdBfd_yy = inv_chi * (vector(gtd_xy, i, j, k) * CovdBfu_yx + vector(gtd_yy, i, j, k) * CovdBfu_yy + vector(gtd_yz, i, j, k) * CovdBfu_yz);
							CovdBfd_yx = inv_chi * (vector(gtd_xx, i, j, k) * CovdBfu_yx + vector(gtd_xy, i, j, k) * CovdBfu_yy + vector(gtd_xz, i, j, k) * CovdBfu_yz);
							CovdBfd_yz = inv_chi * (vector(gtd_xz, i, j, k) * CovdBfu_yx + vector(gtd_yz, i, j, k) * CovdBfu_yy + vector(gtd_zz, i, j, k) * CovdBfu_yz);
						} else {
							CovdBfd_yy = 0.0;
							CovdBfd_yx = 0.0;
							CovdBfd_yz = 0.0;
						}
						optdeptheh = d_optdeptheh_o0_t0_m0_l0;
						if (calculate_les_terms) {
							CovdBvf_z = vector(Bfu_x, i, j, k) * Covdvfd_zx + vector(vfd_x, i, j, k) * CovdBfu_zx + vector(Bfu_y, i, j, k) * Covdvfd_zy + vector(vfd_y, i, j, k) * CovdBfu_zy + vector(Bfu_z, i, j, k) * Covdvfd_zz + vector(vfd_z, i, j, k) * CovdBfu_zz;
						} else {
							CovdBvf_z = 0.0;
						}
						optdepthev = d_optdepthev_o0_t0_m0_l0;
						if (calculate_les_terms) {
							CovdBvf_y = vector(Bfu_x, i, j, k) * Covdvfd_yx + vector(vfd_x, i, j, k) * CovdBfu_yx + vector(Bfu_y, i, j, k) * Covdvfd_yy + vector(vfd_y, i, j, k) * CovdBfu_yy + vector(Bfu_z, i, j, k) * Covdvfd_yz + vector(vfd_z, i, j, k) * CovdBfu_yz;
							CovdBvf_x = vector(Bfu_x, i, j, k) * Covdvfd_xx + vector(vfd_x, i, j, k) * CovdBfu_xx + vector(Bfu_y, i, j, k) * Covdvfd_xy + vector(vfd_y, i, j, k) * CovdBfu_xy + vector(Bfu_z, i, j, k) * Covdvfd_xz + vector(vfd_z, i, j, k) * CovdBfu_xz;
							CovdBfd_xx = inv_chi * (vector(gtd_xx, i, j, k) * CovdBfu_xx + vector(gtd_xy, i, j, k) * CovdBfu_xy + vector(gtd_xz, i, j, k) * CovdBfu_xz);
							CovdBfd_xz = inv_chi * (vector(gtd_xz, i, j, k) * CovdBfu_xx + vector(gtd_yz, i, j, k) * CovdBfu_xy + vector(gtd_zz, i, j, k) * CovdBfu_xz);
							CovdBfd_xy = inv_chi * (vector(gtd_xy, i, j, k) * CovdBfu_xx + vector(gtd_yy, i, j, k) * CovdBfu_xy + vector(gtd_yz, i, j, k) * CovdBfu_xz);
						} else {
							CovdBvf_y = 0.0;
							CovdBvf_x = 0.0;
							CovdBfd_xx = 0.0;
							CovdBfd_xz = 0.0;
							CovdBfd_xy = 0.0;
						}
						optdepthxd = d_optdepthxd_o0_t0_m0_l0;
						optdepthxh = d_optdepthxh_o0_t0_m0_l0;
						optdepthxv = d_optdepthxv_o0_t0_m0_l0;
						chiav = d_chiav_o0_t0_m0_l0;
						if (calculate_les_terms) {
							CovdEfu_zz = d_CovdEfu_zz_o0_t0_m0_l0;
							CovdEfu_zy = d_CovdEfu_zy_o0_t0_m0_l0;
							CovdEfu_zx = d_CovdEfu_zx_o0_t0_m0_l0;
							CovdEfd_zz = inv_chi * (vector(gtd_xz, i, j, k) * CovdEfu_zx + vector(gtd_yz, i, j, k) * CovdEfu_zy + vector(gtd_zz, i, j, k) * CovdEfu_zz);
							CovdEfd_zx = inv_chi * (vector(gtd_xx, i, j, k) * CovdEfu_zx + vector(gtd_xy, i, j, k) * CovdEfu_zy + vector(gtd_xz, i, j, k) * CovdEfu_zz);
							CovdEfd_zy = inv_chi * (vector(gtd_xy, i, j, k) * CovdEfu_zx + vector(gtd_yy, i, j, k) * CovdEfu_zy + vector(gtd_yz, i, j, k) * CovdEfu_zz);
						} else {
							CovdEfu_zz = 0.0;
							CovdEfu_zy = 0.0;
							CovdEfu_zx = 0.0;
							CovdEfd_zz = 0.0;
							CovdEfd_zx = 0.0;
							CovdEfd_zy = 0.0;
						}
						Bvf = (vector(Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(Bfu_z, i, j, k) * vector(vfd_z, i, j, k)) / sdetg;
						optdepthah = d_optdepthah_o0_t0_m0_l0;
						optdepthad = d_optdepthad_o0_t0_m0_l0;
						h = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
						chiah = d_chiah_o0_t0_m0_l0;
						chiad = d_chiad_o0_t0_m0_l0;
						if (calculate_les_terms) {
							CovdYef_y = d_CovdYef_y_o0_t0_m0_l0;
						} else {
							CovdYef_y = 0.0;
						}
						optdepthav = d_optdepthav_o0_t0_m0_l0;
						if (calculate_les_terms) {
							CovdYef_z = d_CovdYef_z_o0_t0_m0_l0;
							CovdYef_x = d_CovdYef_x_o0_t0_m0_l0;
						} else {
							CovdYef_z = 0.0;
							CovdYef_x = 0.0;
						}
						detgtd = vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) * vector(gtd_xx, i, j, k) + (-vector(gtd_yy, i, j, k) * (vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k))) + (-vector(gtd_zz, i, j, k) * (vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k))) + (-(vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k)) * vector(gtd_xx, i, j, k)) + 2.0 * vector(gtd_yz, i, j, k) * vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k);
						idetgtd = 1.0 / detgtd;
						gtu_zz = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_yy, i, j, k) - vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k));
						gtu_yy = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k));
						gtu_yz = idetgtd * ((-vector(gtd_xx, i, j, k) * vector(gtd_yz, i, j, k)) + vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k));
						gtu_xx = idetgtd * (vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k));
						gtu_xz = idetgtd * (vector(gtd_xy, i, j, k) * vector(gtd_yz, i, j, k) - vector(gtd_yy, i, j, k) * vector(gtd_xz, i, j, k));
						if (calculate_les_terms) {
							CovuBfu_zx = chi_max * (gtu_xz * CovdBfu_xx + gtu_yz * CovdBfu_yx + gtu_zz * CovdBfu_zx);
							CovuBfu_zz = chi_max * (gtu_xz * CovdBfu_xz + gtu_yz * CovdBfu_yz + gtu_zz * CovdBfu_zz);
							CovuBfu_zy = chi_max * (gtu_xz * CovdBfu_xy + gtu_yz * CovdBfu_yy + gtu_zz * CovdBfu_zy);
							Covdvfu_xz = chi_max * (gtu_xz * Covdvfd_xx + gtu_yz * Covdvfd_xy + gtu_zz * Covdvfd_xz);
							CovuYef_z = chi_max * (gtu_xz * CovdYef_x + gtu_yz * CovdYef_y + gtu_zz * CovdYef_z);
							Covdvfu_zz = chi_max * (gtu_xz * Covdvfd_zx + gtu_yz * Covdvfd_zy + gtu_zz * Covdvfd_zz);
							Covdvfu_yz = chi_max * (gtu_xz * Covdvfd_yx + gtu_yz * Covdvfd_yy + gtu_zz * Covdvfd_yz);
							Covuvfu_zz = chi_max * (gtu_xz * Covdvfu_xz + gtu_yz * Covdvfu_yz + gtu_zz * Covdvfu_zz);
						} else {
							CovuBfu_zx = 0.0;
							CovuBfu_zz = 0.0;
							CovuBfu_zy = 0.0;
							Covdvfu_xz = 0.0;
							CovuYef_z = 0.0;
							Covdvfu_zz = 0.0;
							Covdvfu_yz = 0.0;
							Covuvfu_zz = 0.0;
						}
						Sfu_z = chi_max * (gtu_xz * vector(Sfd_x, i, j, k) + gtu_yz * vector(Sfd_y, i, j, k) + gtu_zz * vector(Sfd_z, i, j, k));
						if (calculate_les_terms) {
							Covurhof_z = chi_max * (gtu_xz * Covdrhof_x + gtu_yz * Covdrhof_y + gtu_zz * Covdrhof_z);
							CovuBvf_z = chi_max * (gtu_xz * CovdBvf_x + gtu_yz * CovdBvf_y + gtu_zz * CovdBvf_z);
						} else {
							Covurhof_z = 0.0;
							CovuBvf_z = 0.0;
						}
						gtu_xy = idetgtd * ((-vector(gtd_xy, i, j, k) * vector(gtd_zz, i, j, k)) + vector(gtd_yz, i, j, k) * vector(gtd_xz, i, j, k));
						if (calculate_les_terms) {
							CovuBfu_yz = chi_max * (gtu_xy * CovdBfu_xz + gtu_yy * CovdBfu_yz + gtu_yz * CovdBfu_zz);
							CovuBfu_yy = chi_max * (gtu_xy * CovdBfu_xy + gtu_yy * CovdBfu_yy + gtu_yz * CovdBfu_zy);
							CovuBfu_yx = chi_max * (gtu_xy * CovdBfu_xx + gtu_yy * CovdBfu_yx + gtu_yz * CovdBfu_zx);
							Covdvfu_xy = chi_max * (gtu_xy * Covdvfd_xx + gtu_yy * Covdvfd_xy + gtu_yz * Covdvfd_xz);
							Covdvfu_xx = chi_max * (gtu_xx * Covdvfd_xx + gtu_xy * Covdvfd_xy + gtu_xz * Covdvfd_xz);
							CovdinvsqW_x = (-2.0 * vector(vfd_x, i, j, k) * Covdvfu_xx) + (-2.0 * vector(vfd_y, i, j, k) * Covdvfu_xy) + (-2.0 * vector(vfd_z, i, j, k) * Covdvfu_xz);
							CovuYef_x = chi_max * (gtu_xx * CovdYef_x + gtu_xy * CovdYef_y + gtu_xz * CovdYef_z);
							CovuYef_y = chi_max * (gtu_xy * CovdYef_x + gtu_yy * CovdYef_y + gtu_yz * CovdYef_z);
							Covdvfu_zy = chi_max * (gtu_xy * Covdvfd_zx + gtu_yy * Covdvfd_zy + gtu_yz * Covdvfd_zz);
							Covdvfu_zx = chi_max * (gtu_xx * Covdvfd_zx + gtu_xy * Covdvfd_zy + gtu_xz * Covdvfd_zz);
							CovdinvsqW_z = (-2.0 * vector(vfd_x, i, j, k) * Covdvfu_zx) + (-2.0 * vector(vfd_y, i, j, k) * Covdvfu_zy) + (-2.0 * vector(vfd_z, i, j, k) * Covdvfu_zz);
							Covdvfu_yx = chi_max * (gtu_xx * Covdvfd_yx + gtu_xy * Covdvfd_yy + gtu_xz * Covdvfd_yz);
							Covuvfu_zx = chi_max * (gtu_xz * Covdvfu_xx + gtu_yz * Covdvfu_yx + gtu_zz * Covdvfu_zx);
							Covdvfu_yy = chi_max * (gtu_xy * Covdvfd_yx + gtu_yy * Covdvfd_yy + gtu_yz * Covdvfd_yz);
							CovdinvsqW_y = (-2.0 * vector(vfd_x, i, j, k) * Covdvfu_yx) + (-2.0 * vector(vfd_y, i, j, k) * Covdvfu_yy) + (-2.0 * vector(vfd_z, i, j, k) * Covdvfu_yz);
							CovuinvsqW_z = chi_max * (gtu_xz * CovdinvsqW_x + gtu_yz * CovdinvsqW_y + gtu_zz * CovdinvsqW_z);
							Covuvfu_zy = chi_max * (gtu_xz * Covdvfu_xy + gtu_yz * Covdvfu_yy + gtu_zz * Covdvfu_zy);
							CovuBfu_xz = chi_max * (gtu_xx * CovdBfu_xz + gtu_xy * CovdBfu_yz + gtu_xz * CovdBfu_zz);
							CovuBfu_xy = chi_max * (gtu_xx * CovdBfu_xy + gtu_xy * CovdBfu_yy + gtu_xz * CovdBfu_zy);
							CovuBfu_xx = chi_max * (gtu_xx * CovdBfu_xx + gtu_xy * CovdBfu_yx + gtu_xz * CovdBfu_zx);
						} else {
							CovuBfu_yz = 0.0;
							CovuBfu_yy = 0.0;
							CovuBfu_yx = 0.0;
							Covdvfu_xy = 0.0;
							Covdvfu_xx = 0.0;
							CovdinvsqW_x = 0.0;
							CovuYef_x = 0.0;
							CovuYef_y = 0.0;
							Covdvfu_zy = 0.0;
							Covdvfu_zx = 0.0;
							CovdinvsqW_z = 0.0;
							Covdvfu_yx = 0.0;
							Covuvfu_zx = 0.0;
							Covdvfu_yy = 0.0;
							CovdinvsqW_y = 0.0;
							CovuinvsqW_z = 0.0;
							Covuvfu_zy = 0.0;
							CovuBfu_xz = 0.0;
							CovuBfu_xy = 0.0;
							CovuBfu_xx = 0.0;
						}
						Sfu_y = chi_max * (gtu_xy * vector(Sfd_x, i, j, k) + gtu_yy * vector(Sfd_y, i, j, k) + gtu_yz * vector(Sfd_z, i, j, k));
						Sfu_x = chi_max * (gtu_xx * vector(Sfd_x, i, j, k) + gtu_xy * vector(Sfd_y, i, j, k) + gtu_xz * vector(Sfd_z, i, j, k));
						if (calculate_les_terms) {
							Covurhof_x = chi_max * (gtu_xx * Covdrhof_x + gtu_xy * Covdrhof_y + gtu_xz * Covdrhof_z);
							Covurhof_y = chi_max * (gtu_xy * Covdrhof_x + gtu_yy * Covdrhof_y + gtu_yz * Covdrhof_z);
							Covuvfu_yz = chi_max * (gtu_xy * Covdvfu_xz + gtu_yy * Covdvfu_yz + gtu_yz * Covdvfu_zz);
							Covuvfu_yy = chi_max * (gtu_xy * Covdvfu_xy + gtu_yy * Covdvfu_yy + gtu_yz * Covdvfu_zy);
							Covuvfu_yx = chi_max * (gtu_xy * Covdvfu_xx + gtu_yy * Covdvfu_yx + gtu_yz * Covdvfu_zx);
							Covuvfu_xz = chi_max * (gtu_xx * Covdvfu_xz + gtu_xy * Covdvfu_yz + gtu_xz * Covdvfu_zz);
							Covuvfu_xy = chi_max * (gtu_xx * Covdvfu_xy + gtu_xy * Covdvfu_yy + gtu_xz * Covdvfu_zy);
							Covuvfu_xx = chi_max * (gtu_xx * Covdvfu_xx + gtu_xy * Covdvfu_yx + gtu_xz * Covdvfu_zx);
						} else {
							Covurhof_x = 0.0;
							Covurhof_y = 0.0;
							Covuvfu_yz = 0.0;
							Covuvfu_yy = 0.0;
							Covuvfu_yx = 0.0;
							Covuvfu_xz = 0.0;
							Covuvfu_xy = 0.0;
							Covuvfu_xx = 0.0;
						}
						vfu_y = chi_max * (gtu_xy * vector(vfd_x, i, j, k) + gtu_yy * vector(vfd_y, i, j, k) + gtu_yz * vector(vfd_z, i, j, k));
						vfu_x = chi_max * (gtu_xx * vector(vfd_x, i, j, k) + gtu_xy * vector(vfd_y, i, j, k) + gtu_xz * vector(vfd_z, i, j, k));
						if (calculate_les_terms) {
							CovuBvf_x = chi_max * (gtu_xx * CovdBvf_x + gtu_xy * CovdBvf_y + gtu_xz * CovdBvf_z);
							CovuBvf_y = chi_max * (gtu_xy * CovdBvf_x + gtu_yy * CovdBvf_y + gtu_yz * CovdBvf_z);
							CovuinvsqW_y = chi_max * (gtu_xy * CovdinvsqW_x + gtu_yy * CovdinvsqW_y + gtu_yz * CovdinvsqW_z);
							CovuinvsqW_x = chi_max * (gtu_xx * CovdinvsqW_x + gtu_xy * CovdinvsqW_y + gtu_xz * CovdinvsqW_z);
							Covddpfdye_z = d_Covddpfdye_z_o0_t0_m0_l0;
							Covddpfdye_y = d_Covddpfdye_y_o0_t0_m0_l0;
							Covddpfdye_x = d_Covddpfdye_x_o0_t0_m0_l0;
							CovdEfu_xz = d_CovdEfu_xz_o0_t0_m0_l0;
							CovdEfu_xy = d_CovdEfu_xy_o0_t0_m0_l0;
							CovdEfu_xx = d_CovdEfu_xx_o0_t0_m0_l0;
							CovdEfd_xx = inv_chi * (vector(gtd_xx, i, j, k) * CovdEfu_xx + vector(gtd_xy, i, j, k) * CovdEfu_xy + vector(gtd_xz, i, j, k) * CovdEfu_xz);
							CovdEfd_xy = inv_chi * (vector(gtd_xy, i, j, k) * CovdEfu_xx + vector(gtd_yy, i, j, k) * CovdEfu_xy + vector(gtd_yz, i, j, k) * CovdEfu_xz);
							CovdEfd_xz = inv_chi * (vector(gtd_xz, i, j, k) * CovdEfu_xx + vector(gtd_yz, i, j, k) * CovdEfu_xy + vector(gtd_zz, i, j, k) * CovdEfu_xz);
						} else {
							CovuBvf_x = 0.0;
							CovuBvf_y = 0.0;
							CovuinvsqW_y = 0.0;
							CovuinvsqW_x = 0.0;
							Covddpfdye_z = 0.0;
							Covddpfdye_y = 0.0;
							Covddpfdye_x = 0.0;
							CovdEfu_xz = 0.0;
							CovdEfu_xy = 0.0;
							CovdEfu_xx = 0.0;
							CovdEfd_xx = 0.0;
							CovdEfd_xy = 0.0;
							CovdEfd_xz = 0.0;
						}
						vfu_z = chi_max * (gtu_xz * vector(vfd_x, i, j, k) + gtu_yz * vector(vfd_y, i, j, k) + gtu_zz * vector(vfd_z, i, j, k));
						sqvf = MIN(vfu_x * vector(vfd_x, i, j, k) + vfu_y * vector(vfd_y, i, j, k) + vfu_z * vector(vfd_z, i, j, k), 0.9999999999);
						sqW = 1.0 / (1.0 - sqvf);
						if (calculate_les_terms) {
							PhiAh = sqW * (vector(pf, i, j, k) * vector(dpfdeps, i, j, k) + (vector(rhof, i, j, k) * vector(rhof, i, j, k)) * vector(dpfdrho, i, j, k));
						} else {
							PhiAh = 0.0;
						}
						Epf = h * sqW;
						Thetaf = Epf + sqBf / (sdetg * sdetg);
						if (calculate_les_terms) {
							Covdepsf_z = d_Covdepsf_z_o0_t0_m0_l0;
							Covdhf_z = (1.0 + vector(epsf, i, j, k)) * Covdrhof_z + vector(rhof, i, j, k) * Covdepsf_z + Covdpf_z;
							CovdEpf_z = sqW * Covdhf_z + (-h * (sqW * sqW) * CovdinvsqW_z);
							CovdThetaf_z = CovdEpf_z + 2.0 * (Bfd_x * CovdBfu_zx) / sdetg + 2.0 * (Bfd_y * CovdBfu_zy) / sdetg + 2.0 * (Bfd_z * CovdBfu_zz) / sdetg;
							Covdepsf_y = d_Covdepsf_y_o0_t0_m0_l0;
							Covdhf_y = (1.0 + vector(epsf, i, j, k)) * Covdrhof_y + vector(rhof, i, j, k) * Covdepsf_y + Covdpf_y;
							CovdEpf_y = sqW * Covdhf_y + (-h * (sqW * sqW) * CovdinvsqW_y);
							CovdThetaf_y = CovdEpf_y + 2.0 * (Bfd_x * CovdBfu_yx) / sdetg + 2.0 * (Bfd_y * CovdBfu_yy) / sdetg + 2.0 * (Bfd_z * CovdBfu_yz) / sdetg;
							Covdepsf_x = d_Covdepsf_x_o0_t0_m0_l0;
							Covuepsf_z = chi_max * (gtu_xz * Covdepsf_x + gtu_yz * Covdepsf_y + gtu_zz * Covdepsf_z);
							Covuepsf_y = chi_max * (gtu_xy * Covdepsf_x + gtu_yy * Covdepsf_y + gtu_yz * Covdepsf_z);
							Covdhf_x = (1.0 + vector(epsf, i, j, k)) * Covdrhof_x + vector(rhof, i, j, k) * Covdepsf_x + Covdpf_x;
							CovdEpf_x = sqW * Covdhf_x + (-h * (sqW * sqW) * CovdinvsqW_x);
							CovdThetaf_x = CovdEpf_x + 2.0 * (Bfd_x * CovdBfu_xx) / sdetg + 2.0 * (Bfd_y * CovdBfu_xy) / sdetg + 2.0 * (Bfd_z * CovdBfu_xz) / sdetg;
							CovuThetaf_z = chi_max * (gtu_xz * CovdThetaf_x + gtu_yz * CovdThetaf_y + gtu_zz * CovdThetaf_z);
							CovuThetaf_x = chi_max * (gtu_xx * CovdThetaf_x + gtu_xy * CovdThetaf_y + gtu_xz * CovdThetaf_z);
							CovuThetaf_y = chi_max * (gtu_xy * CovdThetaf_x + gtu_yy * CovdThetaf_y + gtu_yz * CovdThetaf_z);
							Phivh_z = 2.0 * (CovuBvf_x * CovdBfu_xz + CovuBvf_y * CovdBfu_yz + CovuBvf_z * CovdBfu_zz + (-CovuThetaf_x * Covdvfu_xz) + (-CovuThetaf_y * Covdvfu_yz) + (-CovuThetaf_z * Covdvfu_zz) + (vector(Bfu_z, i, j, k) * (Thetaf * (CovuBfu_xx * Covdvfd_xx + CovuBfu_yx * Covdvfd_yx + CovuBfu_zx * Covdvfd_zx + CovuBfu_xy * Covdvfd_xy + CovuBfu_yy * Covdvfd_yy + CovuBfu_zy * Covdvfd_zy + CovuBfu_xz * Covdvfd_xz + CovuBfu_yz * Covdvfd_yz + CovuBfu_zz * Covdvfd_zz) + (Bfd_x * CovdBfu_xx * CovuBvf_x) / sdetg + (Bfd_y * CovdBfu_xy * CovuBvf_x) / sdetg + (Bfd_z * CovdBfu_xz * CovuBvf_x) / sdetg + (Bfd_x * CovdBfu_yx * CovuBvf_y) / sdetg + (Bfd_y * CovdBfu_yy * CovuBvf_y) / sdetg + (Bfd_z * CovdBfu_yz * CovuBvf_y) / sdetg + (Bfd_x * CovdBfu_zx * CovuBvf_z) / sdetg + (Bfd_y * CovdBfu_zy * CovuBvf_z) / sdetg + (Bfd_z * CovdBfu_zz * CovuBvf_z) / sdetg + (-(vector(Bfu_x, i, j, k) * Covdvfd_xx * CovuThetaf_x) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_xy * CovuThetaf_x) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_xz * CovuThetaf_x) / sdetg) + (-(vector(Bfu_x, i, j, k) * Covdvfd_yx * CovuThetaf_y) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_yy * CovuThetaf_y) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_yz * CovuThetaf_y) / sdetg) + (-(vector(Bfu_x, i, j, k) * Covdvfd_zx * CovuThetaf_z) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_zy * CovuThetaf_z) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_zz * CovuThetaf_z) / sdetg))) / (sdetg * Epf)) / Thetaf;
							Phivh_x = 2.0 * (CovuBvf_x * CovdBfu_xx + CovuBvf_y * CovdBfu_yx + CovuBvf_z * CovdBfu_zx + (-CovuThetaf_x * Covdvfu_xx) + (-CovuThetaf_y * Covdvfu_yx) + (-CovuThetaf_z * Covdvfu_zx) + (vector(Bfu_x, i, j, k) * (Thetaf * (CovuBfu_xx * Covdvfd_xx + CovuBfu_yx * Covdvfd_yx + CovuBfu_zx * Covdvfd_zx + CovuBfu_xy * Covdvfd_xy + CovuBfu_yy * Covdvfd_yy + CovuBfu_zy * Covdvfd_zy + CovuBfu_xz * Covdvfd_xz + CovuBfu_yz * Covdvfd_yz + CovuBfu_zz * Covdvfd_zz) + (Bfd_x * CovdBfu_xx * CovuBvf_x) / sdetg + (Bfd_y * CovdBfu_xy * CovuBvf_x) / sdetg + (Bfd_z * CovdBfu_xz * CovuBvf_x) / sdetg + (Bfd_x * CovdBfu_yx * CovuBvf_y) / sdetg + (Bfd_y * CovdBfu_yy * CovuBvf_y) / sdetg + (Bfd_z * CovdBfu_yz * CovuBvf_y) / sdetg + (Bfd_x * CovdBfu_zx * CovuBvf_z) / sdetg + (Bfd_y * CovdBfu_zy * CovuBvf_z) / sdetg + (Bfd_z * CovdBfu_zz * CovuBvf_z) / sdetg + (-(vector(Bfu_x, i, j, k) * Covdvfd_xx * CovuThetaf_x) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_xy * CovuThetaf_x) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_xz * CovuThetaf_x) / sdetg) + (-(vector(Bfu_x, i, j, k) * Covdvfd_yx * CovuThetaf_y) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_yy * CovuThetaf_y) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_yz * CovuThetaf_y) / sdetg) + (-(vector(Bfu_x, i, j, k) * Covdvfd_zx * CovuThetaf_z) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_zy * CovuThetaf_z) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_zz * CovuThetaf_z) / sdetg))) / (sdetg * Epf)) / Thetaf;
							Phivh_y = 2.0 * (CovuBvf_x * CovdBfu_xy + CovuBvf_y * CovdBfu_yy + CovuBvf_z * CovdBfu_zy + (-CovuThetaf_x * Covdvfu_xy) + (-CovuThetaf_y * Covdvfu_yy) + (-CovuThetaf_z * Covdvfu_zy) + (vector(Bfu_y, i, j, k) * (Thetaf * (CovuBfu_xx * Covdvfd_xx + CovuBfu_yx * Covdvfd_yx + CovuBfu_zx * Covdvfd_zx + CovuBfu_xy * Covdvfd_xy + CovuBfu_yy * Covdvfd_yy + CovuBfu_zy * Covdvfd_zy + CovuBfu_xz * Covdvfd_xz + CovuBfu_yz * Covdvfd_yz + CovuBfu_zz * Covdvfd_zz) + (Bfd_x * CovdBfu_xx * CovuBvf_x) / sdetg + (Bfd_y * CovdBfu_xy * CovuBvf_x) / sdetg + (Bfd_z * CovdBfu_xz * CovuBvf_x) / sdetg + (Bfd_x * CovdBfu_yx * CovuBvf_y) / sdetg + (Bfd_y * CovdBfu_yy * CovuBvf_y) / sdetg + (Bfd_z * CovdBfu_yz * CovuBvf_y) / sdetg + (Bfd_x * CovdBfu_zx * CovuBvf_z) / sdetg + (Bfd_y * CovdBfu_zy * CovuBvf_z) / sdetg + (Bfd_z * CovdBfu_zz * CovuBvf_z) / sdetg + (-(vector(Bfu_x, i, j, k) * Covdvfd_xx * CovuThetaf_x) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_xy * CovuThetaf_x) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_xz * CovuThetaf_x) / sdetg) + (-(vector(Bfu_x, i, j, k) * Covdvfd_yx * CovuThetaf_y) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_yy * CovuThetaf_y) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_yz * CovuThetaf_y) / sdetg) + (-(vector(Bfu_x, i, j, k) * Covdvfd_zx * CovuThetaf_z) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfd_zy * CovuThetaf_z) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfd_zz * CovuThetaf_z) / sdetg))) / (sdetg * Epf)) / Thetaf;
							PhiMh_yz = 2.0 * (Thetaf * ((-Covdvfu_xz * CovuBfu_xy) + Covdvfu_xy * CovuBfu_xz + (-Covdvfu_yz * CovuBfu_yy) + Covdvfu_yy * CovuBfu_yz + (-Covdvfu_zz * CovuBfu_zy) + Covdvfu_zy * CovuBfu_zz) + (vector(Bfu_z, i, j, k) * CovuBfu_xy * CovdBvf_x) / sdetg + (vector(Bfu_z, i, j, k) * CovuBfu_yy * CovdBvf_y) / sdetg + (vector(Bfu_z, i, j, k) * CovuBfu_zy * CovdBvf_z) / sdetg + (-(vector(Bfu_y, i, j, k) * CovuBfu_xz * CovdBvf_x) / sdetg) + (-(vector(Bfu_y, i, j, k) * CovuBfu_yz * CovdBvf_y) / sdetg) + (-(vector(Bfu_y, i, j, k) * CovuBfu_zz * CovdBvf_z) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfu_xy * CovuThetaf_x) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfu_yy * CovuThetaf_y) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfu_zy * CovuThetaf_z) / sdetg) + (vector(Bfu_y, i, j, k) * Covdvfu_xz * CovuThetaf_x) / sdetg + (vector(Bfu_y, i, j, k) * Covdvfu_yz * CovuThetaf_y) / sdetg + (vector(Bfu_y, i, j, k) * Covdvfu_zz * CovuThetaf_z) / sdetg) / Thetaf;
							PhiMh_xz = 2.0 * (Thetaf * ((-Covdvfu_xz * CovuBfu_xx) + Covdvfu_xx * CovuBfu_xz + (-Covdvfu_yz * CovuBfu_yx) + Covdvfu_yx * CovuBfu_yz + (-Covdvfu_zz * CovuBfu_zx) + Covdvfu_zx * CovuBfu_zz) + (vector(Bfu_z, i, j, k) * CovuBfu_xx * CovdBvf_x) / sdetg + (vector(Bfu_z, i, j, k) * CovuBfu_yx * CovdBvf_y) / sdetg + (vector(Bfu_z, i, j, k) * CovuBfu_zx * CovdBvf_z) / sdetg + (-(vector(Bfu_x, i, j, k) * CovuBfu_xz * CovdBvf_x) / sdetg) + (-(vector(Bfu_x, i, j, k) * CovuBfu_yz * CovdBvf_y) / sdetg) + (-(vector(Bfu_x, i, j, k) * CovuBfu_zz * CovdBvf_z) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfu_xx * CovuThetaf_x) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfu_yx * CovuThetaf_y) / sdetg) + (-(vector(Bfu_z, i, j, k) * Covdvfu_zx * CovuThetaf_z) / sdetg) + (vector(Bfu_x, i, j, k) * Covdvfu_xz * CovuThetaf_x) / sdetg + (vector(Bfu_x, i, j, k) * Covdvfu_yz * CovuThetaf_y) / sdetg + (vector(Bfu_x, i, j, k) * Covdvfu_zz * CovuThetaf_z) / sdetg) / Thetaf;
							PhiMh_xy = 2.0 * (Thetaf * ((-Covdvfu_xy * CovuBfu_xx) + Covdvfu_xx * CovuBfu_xy + (-Covdvfu_yy * CovuBfu_yx) + Covdvfu_yx * CovuBfu_yy + (-Covdvfu_zy * CovuBfu_zx) + Covdvfu_zx * CovuBfu_zy) + (vector(Bfu_y, i, j, k) * CovuBfu_xx * CovdBvf_x) / sdetg + (vector(Bfu_y, i, j, k) * CovuBfu_yx * CovdBvf_y) / sdetg + (vector(Bfu_y, i, j, k) * CovuBfu_zx * CovdBvf_z) / sdetg + (-(vector(Bfu_x, i, j, k) * CovuBfu_xy * CovdBvf_x) / sdetg) + (-(vector(Bfu_x, i, j, k) * CovuBfu_yy * CovdBvf_y) / sdetg) + (-(vector(Bfu_x, i, j, k) * CovuBfu_zy * CovdBvf_z) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfu_xx * CovuThetaf_x) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfu_yx * CovuThetaf_y) / sdetg) + (-(vector(Bfu_y, i, j, k) * Covdvfu_zx * CovuThetaf_z) / sdetg) + (vector(Bfu_x, i, j, k) * Covdvfu_xy * CovuThetaf_x) / sdetg + (vector(Bfu_x, i, j, k) * Covdvfu_yy * CovuThetaf_y) / sdetg + (vector(Bfu_x, i, j, k) * Covdvfu_zy * CovuThetaf_z) / sdetg) / Thetaf;
							Covuepsf_x = chi_max * (gtu_xx * Covdepsf_x + gtu_xy * Covdepsf_y + gtu_xz * Covdepsf_z);
						} else {
							Covdepsf_z = 0.0;
							Covdhf_z = 0.0;
							CovdEpf_z = 0.0;
							CovdThetaf_z = 0.0;
							Covdepsf_y = 0.0;
							Covdhf_y = 0.0;
							CovdEpf_y = 0.0;
							CovdThetaf_y = 0.0;
							Covdepsf_x = 0.0;
							Covuepsf_z = 0.0;
							Covuepsf_y = 0.0;
							Covdhf_x = 0.0;
							CovdEpf_x = 0.0;
							CovdThetaf_x = 0.0;
							CovuThetaf_z = 0.0;
							CovuThetaf_x = 0.0;
							CovuThetaf_y = 0.0;
							Phivh_z = 0.0;
							Phivh_x = 0.0;
							Phivh_y = 0.0;
							PhiMh_yz = 0.0;
							PhiMh_xz = 0.0;
							PhiMh_xy = 0.0;
							Covuepsf_x = 0.0;
						}
						faceta = 0.04166666666666667 * inv_chi * (dx[0] * dx[0]) * Smooth(threshold_Dles_min, threshold_Dles_max, vector(Df, i, j, k), dx, simPlat_dt, ilast, jlast);
						if (calculate_les_terms) {
							CovdEfu_yz = d_CovdEfu_yz_o0_t0_m0_l0;
							CovuEfu_zz = chi_max * (gtu_xz * CovdEfu_xz + gtu_yz * CovdEfu_yz + gtu_zz * CovdEfu_zz);
							CovuEfu_xz = chi_max * (gtu_xx * CovdEfu_xz + gtu_xy * CovdEfu_yz + gtu_xz * CovdEfu_zz);
							CovdEfu_yy = d_CovdEfu_yy_o0_t0_m0_l0;
							CovuEfu_zy = chi_max * (gtu_xz * CovdEfu_xy + gtu_yz * CovdEfu_yy + gtu_zz * CovdEfu_zy);
							CovuEfu_xy = chi_max * (gtu_xx * CovdEfu_xy + gtu_xy * CovdEfu_yy + gtu_xz * CovdEfu_zy);
							CovdEfu_yx = d_CovdEfu_yx_o0_t0_m0_l0;
							CovdEfd_yy = inv_chi * (vector(gtd_xy, i, j, k) * CovdEfu_yx + vector(gtd_yy, i, j, k) * CovdEfu_yy + vector(gtd_yz, i, j, k) * CovdEfu_yz);
							CovdEfd_yz = inv_chi * (vector(gtd_xz, i, j, k) * CovdEfu_yx + vector(gtd_yz, i, j, k) * CovdEfu_yy + vector(gtd_zz, i, j, k) * CovdEfu_yz);
							CovdEfd_yx = inv_chi * (vector(gtd_xx, i, j, k) * CovdEfu_yx + vector(gtd_xy, i, j, k) * CovdEfu_yy + vector(gtd_xz, i, j, k) * CovdEfu_yz);
							CovuEfu_zx = chi_max * (gtu_xz * CovdEfu_xx + gtu_yz * CovdEfu_yx + gtu_zz * CovdEfu_zx);
							CovuEfu_xx = chi_max * (gtu_xx * CovdEfu_xx + gtu_xy * CovdEfu_yx + gtu_xz * CovdEfu_zx);
							CovuEfu_yx = chi_max * (gtu_xy * CovdEfu_xx + gtu_yy * CovdEfu_yx + gtu_yz * CovdEfu_zx);
							CovuEfu_yz = chi_max * (gtu_xy * CovdEfu_xz + gtu_yy * CovdEfu_yz + gtu_yz * CovdEfu_zz);
							CovuEfu_yy = chi_max * (gtu_xy * CovdEfu_xy + gtu_yy * CovdEfu_yy + gtu_yz * CovdEfu_zy);
							PhiThetah = (Thetaf * (CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + (-CovdEfd_xx * CovuEfu_xx) + (-CovdEfd_yx * CovuEfu_yx) + (-CovdEfd_zx * CovuEfu_zx) + (-CovdEfd_xy * CovuEfu_xy) + (-CovdEfd_yy * CovuEfu_yy) + (-CovdEfd_zy * CovuEfu_zy) + (-CovdEfd_xz * CovuEfu_xz) + (-CovdEfd_yz * CovuEfu_yz) + (-CovdEfd_zz * CovuEfu_zz) + (-0.5 * (((-vector(vfd_y, i, j, k) * Bfd_x) + vector(vfd_x, i, j, k) * Bfd_y) * PhiMh_xy + (vector(vfd_x, i, j, k) * Bfd_z - vector(vfd_z, i, j, k) * Bfd_x) * PhiMh_xz + (-(vector(vfd_y, i, j, k) * Bfd_x - vector(vfd_x, i, j, k) * Bfd_y) * PhiMh_xy) + ((-vector(vfd_z, i, j, k) * Bfd_y) + vector(vfd_y, i, j, k) * Bfd_z) * PhiMh_yz + (-(vector(vfd_z, i, j, k) * Bfd_x - vector(vfd_x, i, j, k) * Bfd_z) * PhiMh_xz) + (-(vector(vfd_z, i, j, k) * Bfd_y - vector(vfd_y, i, j, k) * Bfd_z) * PhiMh_yz)) / sdetg))) / (Thetaf - sqEf);
							HPres = (Epf * sqW * (Thetaf - sqEf) * (vector(rhof, i, j, k) * (Covddpfdrho_x * Covurhof_x + Covddpfdrho_y * Covurhof_y + Covddpfdrho_z * Covurhof_z + Covddpfdeps_x * Covuepsf_x + Covddpfdeps_y * Covuepsf_y + Covddpfdeps_z * Covuepsf_z) + (-2.0 * vector(dpfdeps, i, j, k) * (Covurhof_x * Covdepsf_x + Covurhof_y * Covdepsf_y + Covurhof_z * Covdepsf_z)) + (-(Epf * vector(dpfdeps, i, j, k) - PhiAh) * (0.25 * sqW * (CovdinvsqW_x * CovuinvsqW_x + CovdinvsqW_y * CovuinvsqW_y + CovdinvsqW_z * CovuinvsqW_z) + (CovdinvsqW_x * Covurhof_x + CovdinvsqW_y * Covurhof_y + CovdinvsqW_z * Covurhof_z) / vector(rhof, i, j, k))) + (-2.0 * (vector(dpfdeps, i, j, k) * (CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + (-(sqW * sqW) * (CovuinvsqW_x * Covdhf_x + CovuinvsqW_y * Covdhf_y + CovuinvsqW_z * Covdhf_z)))) / sqW) + (-(Epf * vector(dpfdeps, i, j, k) + PhiAh) * (vector(vfd_x, i, j, k) * Phivh_x + vector(vfd_y, i, j, k) * Phivh_y + vector(vfd_z, i, j, k) * Phivh_z + Covdvfd_xx * Covuvfu_xx + Covdvfd_xy * Covuvfu_xy + Covdvfd_xz * Covuvfu_xz + Covdvfd_yx * Covuvfu_yx + Covdvfd_yy * Covuvfu_yy + Covdvfd_yz * Covuvfu_yz + Covdvfd_zx * Covuvfu_zx + Covdvfd_zy * Covuvfu_zy + Covdvfd_zz * Covuvfu_zz + sqW * (CovdinvsqW_x * CovuinvsqW_x + CovdinvsqW_y * CovuinvsqW_y + CovdinvsqW_z * CovuinvsqW_z))) + (PhiThetah * ((Epf * vector(dpfdeps, i, j, k) + PhiAh) * (Thetaf - sqEf) - (PhiAh * Thetaf) / sqW)) / (Epf * Thetaf))) / ((Epf * vector(rhof, i, j, k) - PhiAh) * (Thetaf - sqEf) * sqW + PhiAh * Thetaf) + Covddpfdye_x * CovuYef_x + Covddpfdye_y * CovuYef_y + Covddpfdye_z * CovuYef_z - 2.0 * vector(dpfdye, i, j, k) / vector(Df, i, j, k) * (CovdDf_x * CovuYef_x + CovdDf_y * CovuYef_y + CovdDf_z * CovuYef_z);
							HTheta = PhiThetah + (Thetaf * HPres) / (Thetaf - sqEf);
							Hv_z = Phivh_z + (-((vfu_z + (Bvf * vector(Bfu_z, i, j, k)) / (sdetg * Epf)) * HTheta) / Thetaf);
							HN_z = 2.0 * CovdDf_x * Covuvfu_xz + 2.0 * CovdDf_y * Covuvfu_yz + 2.0 * CovdDf_z * Covuvfu_zz + (vector(Df, i, j, k) * Hv_z) / sdetg;
							HNe_z = 2.0 * CovdDYf_x * Covuvfu_xz + 2.0 * CovdDYf_y * Covuvfu_yz + 2.0 * CovdDYf_z * Covuvfu_zz + (vector(DYf, i, j, k) * Hv_z) / sdetg;
							Hv_y = Phivh_y + (-((vfu_y + (Bvf * vector(Bfu_y, i, j, k)) / (sdetg * Epf)) * HTheta) / Thetaf);
							HN_y = 2.0 * CovdDf_x * Covuvfu_xy + 2.0 * CovdDf_y * Covuvfu_yy + 2.0 * CovdDf_z * Covuvfu_zy + (vector(Df, i, j, k) * Hv_y) / sdetg;
							HM_yz = (-2.0 * Covdvfu_xz * CovuBfu_xy) + 2.0 * Covdvfu_xy * CovuBfu_xz + (-2.0 * Covdvfu_yz * CovuBfu_yy) + 2.0 * Covdvfu_yy * CovuBfu_yz + (-2.0 * Covdvfu_zz * CovuBfu_zy) + 2.0 * Covdvfu_zy * CovuBfu_zz + (vector(Bfu_z, i, j, k) * Hv_y - vector(Bfu_y, i, j, k) * Hv_z) / sdetg;
							HNe_y = 2.0 * CovdDYf_x * Covuvfu_xy + 2.0 * CovdDYf_y * Covuvfu_yy + 2.0 * CovdDYf_z * Covuvfu_zy + (vector(DYf, i, j, k) * Hv_y) / sdetg;
							Hv_x = Phivh_x + (-((vfu_x + (Bvf * vector(Bfu_x, i, j, k)) / (sdetg * Epf)) * HTheta) / Thetaf);
							HN_x = 2.0 * CovdDf_x * Covuvfu_xx + 2.0 * CovdDf_y * Covuvfu_yx + 2.0 * CovdDf_z * Covuvfu_zx + (vector(Df, i, j, k) * Hv_x) / sdetg;
							HM_xz = (-2.0 * Covdvfu_xz * CovuBfu_xx) + 2.0 * Covdvfu_xx * CovuBfu_xz + (-2.0 * Covdvfu_yz * CovuBfu_yx) + 2.0 * Covdvfu_yx * CovuBfu_yz + (-2.0 * Covdvfu_zz * CovuBfu_zx) + 2.0 * Covdvfu_zx * CovuBfu_zz + (vector(Bfu_z, i, j, k) * Hv_x - vector(Bfu_x, i, j, k) * Hv_z) / sdetg;
							HM_xy = (-2.0 * Covdvfu_xy * CovuBfu_xx) + 2.0 * Covdvfu_xx * CovuBfu_xy + (-2.0 * Covdvfu_yy * CovuBfu_yx) + 2.0 * Covdvfu_yx * CovuBfu_yy + (-2.0 * Covdvfu_zy * CovuBfu_zx) + 2.0 * Covdvfu_zx * CovuBfu_zy + (vector(Bfu_y, i, j, k) * Hv_x - vector(Bfu_x, i, j, k) * Hv_y) / sdetg;
							HE_z = 0.5 * ((-2.0 * gtu_yz * HM_xz) + 2.0 * gtu_zz * HM_xy + 2.0 * gtu_xz * HM_yz) / sqrt(chi_max);
							HE_x = 0.5 * ((-2.0 * gtu_xy * HM_xz) + 2.0 * gtu_xz * HM_xy + 2.0 * gtu_xx * HM_yz) / sqrt(chi_max);
							HE_y = 0.5 * ((-2.0 * gtu_yy * HM_xz) + 2.0 * gtu_yz * HM_xy + 2.0 * gtu_xy * HM_yz) / sqrt(chi_max);
							HT_zz = 4.0 * CovdEpf_x * Covuvfu_xz * vfu_z + 4.0 * CovdEpf_y * Covuvfu_yz * vfu_z + 4.0 * CovdEpf_z * Covuvfu_zz * vfu_z + Epf * (2.0 * vfu_z * Hv_z + 2.0 * Covdvfu_xz * Covuvfu_xz + 2.0 * Covdvfu_yz * Covuvfu_yz + 2.0 * Covdvfu_zz * Covuvfu_zz) + 2.0 * (vfu_z * vfu_z) * HPres + (-2.0 * CovdBfu_xz * CovuBfu_xz) + (-2.0 * CovdBfu_yz * CovuBfu_yz) + (-2.0 * CovdBfu_zz * CovuBfu_zz) + (-2.0 * CovdEfu_xz * CovuEfu_xz) + (-2.0 * CovdEfu_yz * CovuEfu_yz) + (-2.0 * CovdEfu_zz * CovuEfu_zz) + (-2.00 * vector(Efu_z, i, j, k) * HE_z) + (chi_max * gtu_zz - vfu_z * vfu_z) * (HPres + CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + CovdEfd_xx * CovuEfu_xx + CovdEfd_yx * CovuEfu_yx + CovdEfd_zx * CovuEfu_zx + CovdEfd_xy * CovuEfu_xy + CovdEfd_yy * CovuEfu_yy + CovdEfd_zy * CovuEfu_zy + CovdEfd_xz * CovuEfu_xz + CovdEfd_yz * CovuEfu_yz + CovdEfd_zz * CovuEfu_zz + Efd_x * HE_x + Efd_y * HE_y + Efd_z * HE_z);
							HT_yy = 4.0 * CovdEpf_x * Covuvfu_xy * vfu_y + 4.0 * CovdEpf_y * Covuvfu_yy * vfu_y + 4.0 * CovdEpf_z * Covuvfu_zy * vfu_y + Epf * (2.0 * vfu_y * Hv_y + 2.0 * Covdvfu_xy * Covuvfu_xy + 2.0 * Covdvfu_yy * Covuvfu_yy + 2.0 * Covdvfu_zy * Covuvfu_zy) + 2.0 * (vfu_y * vfu_y) * HPres + (-2.0 * CovdBfu_xy * CovuBfu_xy) + (-2.0 * CovdBfu_yy * CovuBfu_yy) + (-2.0 * CovdBfu_zy * CovuBfu_zy) + (-2.0 * CovdEfu_xy * CovuEfu_xy) + (-2.0 * CovdEfu_yy * CovuEfu_yy) + (-2.0 * CovdEfu_zy * CovuEfu_zy) + (-2.00 * vector(Efu_y, i, j, k) * HE_y) + (chi_max * gtu_yy - vfu_y * vfu_y) * (HPres + CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + CovdEfd_xx * CovuEfu_xx + CovdEfd_yx * CovuEfu_yx + CovdEfd_zx * CovuEfu_zx + CovdEfd_xy * CovuEfu_xy + CovdEfd_yy * CovuEfu_yy + CovdEfd_zy * CovuEfu_zy + CovdEfd_xz * CovuEfu_xz + CovdEfd_yz * CovuEfu_yz + CovdEfd_zz * CovuEfu_zz + Efd_x * HE_x + Efd_y * HE_y + Efd_z * HE_z);
							HT_yz = 2.0 * CovdEpf_x * (Covuvfu_xz * vfu_y + Covuvfu_xy * vfu_z) + 2.0 * CovdEpf_y * (Covuvfu_yz * vfu_y + Covuvfu_yy * vfu_z) + 2.0 * CovdEpf_z * (Covuvfu_zz * vfu_y + Covuvfu_zy * vfu_z) + Epf * (vfu_y * Hv_z + vfu_z * Hv_y + 2.0 * Covdvfu_xy * Covuvfu_xz + 2.0 * Covdvfu_yy * Covuvfu_yz + 2.0 * Covdvfu_zy * Covuvfu_zz) + 2.0 * vfu_y * vfu_z * HPres + (-2.0 * CovdBfu_xy * CovuBfu_xz) + (-2.0 * CovdBfu_yy * CovuBfu_yz) + (-2.0 * CovdBfu_zy * CovuBfu_zz) + (-2.0 * CovdEfu_xy * CovuEfu_xz) + (-2.0 * CovdEfu_yy * CovuEfu_yz) + (-2.0 * CovdEfu_zy * CovuEfu_zz) + (-1.00 * vector(Efu_z, i, j, k) * HE_y) + (-1.00 * vector(Efu_y, i, j, k) * HE_z) + (chi_max * gtu_yz - vfu_z * vfu_y) * (HPres + CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + CovdEfd_xx * CovuEfu_xx + CovdEfd_yx * CovuEfu_yx + CovdEfd_zx * CovuEfu_zx + CovdEfd_xy * CovuEfu_xy + CovdEfd_yy * CovuEfu_yy + CovdEfd_zy * CovuEfu_zy + CovdEfd_xz * CovuEfu_xz + CovdEfd_yz * CovuEfu_yz + CovdEfd_zz * CovuEfu_zz + Efd_x * HE_x + Efd_y * HE_y + Efd_z * HE_z);
							HT_xx = 4.0 * CovdEpf_x * Covuvfu_xx * vfu_x + 4.0 * CovdEpf_y * Covuvfu_yx * vfu_x + 4.0 * CovdEpf_z * Covuvfu_zx * vfu_x + Epf * (2.0 * vfu_x * Hv_x + 2.0 * Covdvfu_xx * Covuvfu_xx + 2.0 * Covdvfu_yx * Covuvfu_yx + 2.0 * Covdvfu_zx * Covuvfu_zx) + 2.0 * (vfu_x * vfu_x) * HPres + (-2.0 * CovdBfu_xx * CovuBfu_xx) + (-2.0 * CovdBfu_yx * CovuBfu_yx) + (-2.0 * CovdBfu_zx * CovuBfu_zx) + (-2.0 * CovdEfu_xx * CovuEfu_xx) + (-2.0 * CovdEfu_yx * CovuEfu_yx) + (-2.0 * CovdEfu_zx * CovuEfu_zx) + (-2.00 * vector(Efu_x, i, j, k) * HE_x) + (chi_max * gtu_xx - vfu_x * vfu_x) * (HPres + CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + CovdEfd_xx * CovuEfu_xx + CovdEfd_yx * CovuEfu_yx + CovdEfd_zx * CovuEfu_zx + CovdEfd_xy * CovuEfu_xy + CovdEfd_yy * CovuEfu_yy + CovdEfd_zy * CovuEfu_zy + CovdEfd_xz * CovuEfu_xz + CovdEfd_yz * CovuEfu_yz + CovdEfd_zz * CovuEfu_zz + Efd_x * HE_x + Efd_y * HE_y + Efd_z * HE_z);
							HT_xz = 2.0 * CovdEpf_x * (Covuvfu_xz * vfu_x + Covuvfu_xx * vfu_z) + 2.0 * CovdEpf_y * (Covuvfu_yz * vfu_x + Covuvfu_yx * vfu_z) + 2.0 * CovdEpf_z * (Covuvfu_zz * vfu_x + Covuvfu_zx * vfu_z) + Epf * (vfu_x * Hv_z + vfu_z * Hv_x + 2.0 * Covdvfu_xx * Covuvfu_xz + 2.0 * Covdvfu_yx * Covuvfu_yz + 2.0 * Covdvfu_zx * Covuvfu_zz) + 2.0 * vfu_x * vfu_z * HPres + (-2.0 * CovdBfu_xx * CovuBfu_xz) + (-2.0 * CovdBfu_yx * CovuBfu_yz) + (-2.0 * CovdBfu_zx * CovuBfu_zz) + (-2.0 * CovdEfu_xx * CovuEfu_xz) + (-2.0 * CovdEfu_yx * CovuEfu_yz) + (-2.0 * CovdEfu_zx * CovuEfu_zz) + (-1.00 * vector(Efu_z, i, j, k) * HE_x) + (-1.00 * vector(Efu_x, i, j, k) * HE_z) + (chi_max * gtu_xz - vfu_z * vfu_x) * (HPres + CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + CovdEfd_xx * CovuEfu_xx + CovdEfd_yx * CovuEfu_yx + CovdEfd_zx * CovuEfu_zx + CovdEfd_xy * CovuEfu_xy + CovdEfd_yy * CovuEfu_yy + CovdEfd_zy * CovuEfu_zy + CovdEfd_xz * CovuEfu_xz + CovdEfd_yz * CovuEfu_yz + CovdEfd_zz * CovuEfu_zz + Efd_x * HE_x + Efd_y * HE_y + Efd_z * HE_z);
							HT_xy = 2.0 * CovdEpf_x * (Covuvfu_xy * vfu_x + Covuvfu_xx * vfu_y) + 2.0 * CovdEpf_y * (Covuvfu_yy * vfu_x + Covuvfu_yx * vfu_y) + 2.0 * CovdEpf_z * (Covuvfu_zy * vfu_x + Covuvfu_zx * vfu_y) + Epf * (vfu_x * Hv_y + vfu_y * Hv_x + 2.0 * Covdvfu_xx * Covuvfu_xy + 2.0 * Covdvfu_yx * Covuvfu_yy + 2.0 * Covdvfu_zx * Covuvfu_zy) + 2.0 * vfu_x * vfu_y * HPres + (-2.0 * CovdBfu_xx * CovuBfu_xy) + (-2.0 * CovdBfu_yx * CovuBfu_yy) + (-2.0 * CovdBfu_zx * CovuBfu_zy) + (-2.0 * CovdEfu_xx * CovuEfu_xy) + (-2.0 * CovdEfu_yx * CovuEfu_yy) + (-2.0 * CovdEfu_zx * CovuEfu_zy) + (-1.00 * vector(Efu_y, i, j, k) * HE_x) + (-1.00 * vector(Efu_x, i, j, k) * HE_y) + (chi_max * gtu_xy - vfu_y * vfu_x) * (HPres + CovdBfd_xx * CovuBfu_xx + CovdBfd_yx * CovuBfu_yx + CovdBfd_zx * CovuBfu_zx + CovdBfd_xy * CovuBfu_xy + CovdBfd_yy * CovuBfu_yy + CovdBfd_zy * CovuBfu_zy + CovdBfd_xz * CovuBfu_xz + CovdBfd_yz * CovuBfu_yz + CovdBfd_zz * CovuBfu_zz + CovdEfd_xx * CovuEfu_xx + CovdEfd_yx * CovuEfu_yx + CovdEfd_zx * CovuEfu_zx + CovdEfd_xy * CovuEfu_xy + CovdEfd_yy * CovuEfu_yy + CovdEfd_zy * CovuEfu_zy + CovdEfd_xz * CovuEfu_xz + CovdEfd_yz * CovuEfu_yz + CovdEfd_zz * CovuEfu_zz + Efd_x * HE_x + Efd_y * HE_y + Efd_z * HE_z);
							HNe_x = 2.0 * CovdDYf_x * Covuvfu_xx + 2.0 * CovdDYf_y * Covuvfu_yx + 2.0 * CovdDYf_z * Covuvfu_zx + (vector(DYf, i, j, k) * Hv_x) / sdetg;
							d_TauN_x_o0_t0_m0_l0 = -faceta * C_N * HN_x;
							d_TauN_y_o0_t0_m0_l0 = -faceta * C_N * HN_y;
							d_TauN_z_o0_t0_m0_l0 = -faceta * C_N * HN_z;
							d_TauNe_x_o0_t0_m0_l0 = -faceta * C_N * HNe_x;
							d_TauNe_y_o0_t0_m0_l0 = -faceta * C_N * HNe_y;
							d_TauNe_z_o0_t0_m0_l0 = -faceta * C_N * HNe_z;
							d_TauM_xy_o0_t0_m0_l0 = -faceta * C_M * HM_xy;
							d_TauM_xz_o0_t0_m0_l0 = -faceta * C_M * HM_xz;
							d_TauM_yz_o0_t0_m0_l0 = -faceta * C_M * HM_yz;
							d_TauT_xx_o0_t0_m0_l0 = -faceta * C_T * HT_xx;
							d_TauT_xy_o0_t0_m0_l0 = -faceta * C_T * HT_xy;
							d_TauT_xz_o0_t0_m0_l0 = -faceta * C_T * HT_xz;
							d_TauT_yy_o0_t0_m0_l0 = -faceta * C_T * HT_yy;
							d_TauT_yz_o0_t0_m0_l0 = -faceta * C_T * HT_yz;
							d_TauT_zz_o0_t0_m0_l0 = -faceta * C_T * HT_zz;
							vector(TauNe_x, i, j, k) = d_TauNe_x_o0_t0_m0_l0;
							vector(TauNe_y, i, j, k) = d_TauNe_y_o0_t0_m0_l0;
							vector(TauT_yz, i, j, k) = d_TauT_yz_o0_t0_m0_l0;
							vector(TauT_xy, i, j, k) = d_TauT_xy_o0_t0_m0_l0;
							vector(TauM_yz, i, j, k) = d_TauM_yz_o0_t0_m0_l0;
							vector(TauM_xy, i, j, k) = d_TauM_xy_o0_t0_m0_l0;
							vector(TauNe_z, i, j, k) = d_TauNe_z_o0_t0_m0_l0;
							vector(TauT_xz, i, j, k) = d_TauT_xz_o0_t0_m0_l0;
							vector(TauN_x, i, j, k) = d_TauN_x_o0_t0_m0_l0;
							vector(TauT_zz, i, j, k) = d_TauT_zz_o0_t0_m0_l0;
							vector(TauT_yy, i, j, k) = d_TauT_yy_o0_t0_m0_l0;
							vector(TauT_xx, i, j, k) = d_TauT_xx_o0_t0_m0_l0;
							vector(TauM_xz, i, j, k) = d_TauM_xz_o0_t0_m0_l0;
							vector(TauN_z, i, j, k) = d_TauN_z_o0_t0_m0_l0;
							vector(TauN_y, i, j, k) = d_TauN_y_o0_t0_m0_l0;
						} else {
							CovdEfu_yz = 0.0;
							CovuEfu_zz = 0.0;
							CovuEfu_xz = 0.0;
							CovdEfu_yy = 0.0;
							CovuEfu_zy = 0.0;
							CovuEfu_xy = 0.0;
							CovdEfu_yx = 0.0;
							CovdEfd_yy = 0.0;
							CovdEfd_yz = 0.0;
							CovdEfd_yx = 0.0;
							CovuEfu_zx = 0.0;
							CovuEfu_xx = 0.0;
							CovuEfu_yx = 0.0;
							CovuEfu_yz = 0.0;
							CovuEfu_yy = 0.0;
							PhiThetah = 0.0;
							HPres = 0.0;
							HTheta = 0.0;
							Hv_z = 0.0;
							HN_z = 0.0;
							HNe_z = 0.0;
							Hv_y = 0.0;
							HN_y = 0.0;
							HM_yz = 0.0;
							HNe_y = 0.0;
							Hv_x = 0.0;
							HN_x = 0.0;
							HM_xz = 0.0;
							HM_xy = 0.0;
							HE_z = 0.0;
							HE_x = 0.0;
							HE_y = 0.0;
							HT_zz = 0.0;
							HT_yy = 0.0;
							HT_yz = 0.0;
							HT_xx = 0.0;
							HT_xz = 0.0;
							HT_xy = 0.0;
							HNe_x = 0.0;
						}
						optdepthe_tmp = vector(optdepthe, i, j, k);
						optdeptha_tmp = vector(optdeptha, i, j, k);
						optdepthx_tmp = vector(optdepthx, i, j, k);
						chie_tmp = vector(chie, i, j, k);
						chia_tmp = vector(chia, i, j, k);
						chix_tmp = vector(chix, i, j, k);
						qnu_tmp = vector(qnu, i, j, k);
						rnu_tmp = vector(rnu, i, j, k);
						qnu_e_tmp = vector(qnu_e, i, j, k);
						qnu_a_tmp = vector(qnu_a, i, j, k);
						qnu_x_tmp = vector(qnu_x, i, j, k);
						betabeta = vector(gtd_xx, i, j, k) * vector(Betau_x, i, j, k) * vector(Betau_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Betau_y, i, j, k) * vector(Betau_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Betau_z, i, j, k) * vector(Betau_z, i, j, k) + 2.0 * vector(gtd_xy, i, j, k) * vector(Betau_x, i, j, k) * vector(Betau_y, i, j, k) + 2.0 * vector(gtd_xz, i, j, k) * vector(Betau_x, i, j, k) * vector(Betau_z, i, j, k) + 2.0 * vector(gtd_yz, i, j, k) * vector(Betau_y, i, j, k) * vector(Betau_z, i, j, k);
						betav = vector(gtd_xx, i, j, k) * vector(Betau_x, i, j, k) * vector(vfd_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Betau_y, i, j, k) * vector(vfd_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Betau_z, i, j, k) * vector(vfd_z, i, j, k) + vector(gtd_xy, i, j, k) * vector(Betau_x, i, j, k) * vector(vfd_y, i, j, k) + vector(gtd_xy, i, j, k) * vector(Betau_y, i, j, k) * vector(vfd_x, i, j, k) + vector(gtd_xz, i, j, k) * vector(Betau_x, i, j, k) * vector(vfd_z, i, j, k) + vector(gtd_xz, i, j, k) * vector(Betau_z, i, j, k) * vector(vfd_x, i, j, k) + vector(gtd_yz, i, j, k) * vector(Betau_y, i, j, k) * vector(vfd_z, i, j, k) + vector(gtd_yz, i, j, k) * vector(Betau_z, i, j, k) * vector(vfd_y, i, j, k);
						velr = vector(vfd_x, i, j, k) * xcoord(i) + vector(vfd_y, i, j, k) * ycoord(j) + vector(vfd_z, i, j, k) * zcoord(k);
						velr = velr / sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k) + 0.01);
						f_redshift = sqW * (vector(Alpha, i, j, k) - betav) * (1.0 + velr);
						f_redshift = f_redshift * sqrt(vector(Alpha, i, j, k) * vector(Alpha, i, j, k) - betabeta);
						leakage(do_leakage, vacuum_rho, threshold_leakage_vacuum, energyShift, optdepthe_tmp, optdeptha_tmp, optdepthx_tmp, qnu_e_tmp, qnu_a_tmp, qnu_x_tmp, chie_tmp, chia_tmp, chix_tmp, qnu_tmp, rnu_tmp, vector(rhof, i, j, k), vector(epsf, i, j, k), vector(Yef, i, j, k), vector(Tf, i, j, k), sdetg, optdeptheh, optdepthah, optdepthxh, optdepthev, optdepthav, optdepthxv, optdepthed, optdepthad, optdepthxd, chieh, chiah, chixh, chiev, chiav, chixv, chied, chiad, chixd, f_redshift, coord0_data_DataLeakage, coord0_dx_DataLeakage, coord0_size_DataLeakage, coord1_data_DataLeakage, coord1_dx_DataLeakage, coord1_size_DataLeakage, coord2_data_DataLeakage, coord2_dx_DataLeakage, coord2_size_DataLeakage, vars_data_DataLeakage, vars_size_DataLeakage, dx, simPlat_dt, ilast, jlast);
						vector(optdepthe, i, j, k) = optdepthe_tmp;
						vector(optdeptha, i, j, k) = optdeptha_tmp;
						vector(optdepthx, i, j, k) = optdepthx_tmp;
						vector(chie, i, j, k) = chie_tmp;
						vector(chia, i, j, k) = chia_tmp;
						vector(chix, i, j, k) = chix_tmp;
						vector(qnu_e, i, j, k) = qnu_e_tmp;
						vector(qnu_a, i, j, k) = qnu_a_tmp;
						vector(qnu_x, i, j, k) = qnu_x_tmp;
						vector(qnu, i, j, k) = qnu_tmp;
						vector(rnu, i, j, k) = rnu_tmp;
					}
				}
			}
		}
	}
	//Fill ghosts and periodical boundaries
	d_bdry_sched_advance26[ln]->fillData(current_time, false);
	for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
		const std::shared_ptr<hier::Patch >& patch = *p_it;
		double* FOV_1 = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_1_id).get())->getPointer();
		double* FOV_xLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xLower_id).get())->getPointer();
		double* FOV_xUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_xUpper_id).get())->getPointer();
		double* FOV_yLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yLower_id).get())->getPointer();
		double* FOV_yUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_yUpper_id).get())->getPointer();
		double* FOV_zLower = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zLower_id).get())->getPointer();
		double* FOV_zUpper = ((pdat::NodeData<double> *) patch->getPatchData(d_FOV_zUpper_id).get())->getPointer();
	
		//Hard region field distance variables
		double* d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
		double* d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz_id).get())->getPointer();
	
		double* t_optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthe_id).get())->getPointer();
		double* t_optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdeptha_id).get())->getPointer();
		double* t_optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_t_optdepthx_id).get())->getPointer();
		double* t_chie = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chie_id).get())->getPointer();
		double* t_chia = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chia_id).get())->getPointer();
		double* t_chix = ((pdat::NodeData<double> *) patch->getPatchData(d_t_chix_id).get())->getPointer();
		double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
		double* Yef = ((pdat::NodeData<double> *) patch->getPatchData(d_Yef_id).get())->getPointer();
		double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
		double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
		double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
		double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
		double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
		double* sqcs = ((pdat::NodeData<double> *) patch->getPatchData(d_sqcs_id).get())->getPointer();
		double* Tf = ((pdat::NodeData<double> *) patch->getPatchData(d_Tf_id).get())->getPointer();
		double* dpfdrho = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdrho_id).get())->getPointer();
		double* dpfdeps = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdeps_id).get())->getPointer();
		double* dpfdye = ((pdat::NodeData<double> *) patch->getPatchData(d_dpfdye_id).get())->getPointer();
		double* Efu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_x_id).get())->getPointer();
		double* Efu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_y_id).get())->getPointer();
		double* Efu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Efu_z_id).get())->getPointer();
		double* optdepthe = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthe_id).get())->getPointer();
		double* optdeptha = ((pdat::NodeData<double> *) patch->getPatchData(d_optdeptha_id).get())->getPointer();
		double* optdepthx = ((pdat::NodeData<double> *) patch->getPatchData(d_optdepthx_id).get())->getPointer();
		double* chie = ((pdat::NodeData<double> *) patch->getPatchData(d_chie_id).get())->getPointer();
		double* chia = ((pdat::NodeData<double> *) patch->getPatchData(d_chia_id).get())->getPointer();
		double* chix = ((pdat::NodeData<double> *) patch->getPatchData(d_chix_id).get())->getPointer();
		double* qnu = ((pdat::NodeData<double> *) patch->getPatchData(d_qnu_id).get())->getPointer();
		double* rnu = ((pdat::NodeData<double> *) patch->getPatchData(d_rnu_id).get())->getPointer();
		double* TauN_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_x_id).get())->getPointer();
		double* TauN_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_y_id).get())->getPointer();
		double* TauN_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauN_z_id).get())->getPointer();
		double* TauNe_x = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_x_id).get())->getPointer();
		double* TauNe_y = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_y_id).get())->getPointer();
		double* TauNe_z = ((pdat::NodeData<double> *) patch->getPatchData(d_TauNe_z_id).get())->getPointer();
		double* TauM_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xy_id).get())->getPointer();
		double* TauM_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_xz_id).get())->getPointer();
		double* TauM_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauM_yz_id).get())->getPointer();
		double* TauT_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xx_id).get())->getPointer();
		double* TauT_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xy_id).get())->getPointer();
		double* TauT_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_xz_id).get())->getPointer();
		double* TauT_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yy_id).get())->getPointer();
		double* TauT_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_yz_id).get())->getPointer();
		double* TauT_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_TauT_zz_id).get())->getPointer();
		double* gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_id).get())->getPointer();
		double* gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_id).get())->getPointer();
		double* gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_id).get())->getPointer();
		double* gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_id).get())->getPointer();
		double* gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_id).get())->getPointer();
		double* gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_id).get())->getPointer();
		double* Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xx_id).get())->getPointer();
		double* Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xy_id).get())->getPointer();
		double* Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xz_id).get())->getPointer();
		double* Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yy_id).get())->getPointer();
		double* Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yz_id).get())->getPointer();
		double* Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_zz_id).get())->getPointer();
		double* Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_x_id).get())->getPointer();
		double* Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_y_id).get())->getPointer();
		double* Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_z_id).get())->getPointer();
		double* Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_id).get())->getPointer();
		double* Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_id).get())->getPointer();
		double* Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_id).get())->getPointer();
		double* Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_id).get())->getPointer();
		double* chi = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_id).get())->getPointer();
		double* trK = ((pdat::NodeData<double> *) patch->getPatchData(d_trK_id).get())->getPointer();
		double* theta = ((pdat::NodeData<double> *) patch->getPatchData(d_theta_id).get())->getPointer();
	
		//Get the dimensions of the patch
		const hier::Index boxfirst = patch->getBox().lower();
		const hier::Index boxlast = patch->getBox().upper();
	
		//Get delta spaces into an array. dx, dy, dz.
		std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
		const double* dx  = patch_geom->getDx();
	
		//Auxiliary definitions
		int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
		int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
		int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2; i < ilast; i++) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2; j < jlast; j++) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2 - 1; k >= 0; k--) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
			for(int k = klast/2; k < klast; k++) {
				for(int j = jlast/2 - 1; j >= 0; j--) {
					for(int i = ilast/2 - 1; i >= 0; i--) {
					if ((vector(FOV_xLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, d_ghost_width, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, d_ghost_width, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, d_ghost_width, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, d_ghost_width, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, d_ghost_width, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, d_ghost_width, j, k);
						vector(rhof, i, j, k) = vector(rhof, d_ghost_width, j, k);
						vector(Yef, i, j, k) = vector(Yef, d_ghost_width, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, d_ghost_width, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, d_ghost_width, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, d_ghost_width, j, k);
						vector(epsf, i, j, k) = vector(epsf, d_ghost_width, j, k);
						vector(pf, i, j, k) = vector(pf, d_ghost_width, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, d_ghost_width, j, k);
						vector(Tf, i, j, k) = vector(Tf, d_ghost_width, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, d_ghost_width, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, d_ghost_width, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, d_ghost_width, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, d_ghost_width, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, d_ghost_width, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, d_ghost_width, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, d_ghost_width, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, d_ghost_width, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, d_ghost_width, j, k);
						vector(chie, i, j, k) = vector(chie, d_ghost_width, j, k);
						vector(chia, i, j, k) = vector(chia, d_ghost_width, j, k);
						vector(chix, i, j, k) = vector(chix, d_ghost_width, j, k);
						vector(qnu, i, j, k) = vector(qnu, d_ghost_width, j, k);
						vector(rnu, i, j, k) = vector(rnu, d_ghost_width, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, d_ghost_width, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, d_ghost_width, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, d_ghost_width, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, d_ghost_width, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, d_ghost_width, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, d_ghost_width, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, d_ghost_width, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, d_ghost_width, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, d_ghost_width, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, d_ghost_width, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, d_ghost_width, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, d_ghost_width, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, d_ghost_width, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, d_ghost_width, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, d_ghost_width, j, k);
					}
					if ((vector(FOV_xUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i - 1, j, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i - 1, j, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i - 1, j, k);
						vector(t_chie, i, j, k) = vector(t_chie, i - 1, j, k);
						vector(t_chia, i, j, k) = vector(t_chia, i - 1, j, k);
						vector(t_chix, i, j, k) = vector(t_chix, i - 1, j, k);
						vector(rhof, i, j, k) = vector(rhof, i - 1, j, k);
						vector(Yef, i, j, k) = vector(Yef, i - 1, j, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i - 1, j, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i - 1, j, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i - 1, j, k);
						vector(epsf, i, j, k) = vector(epsf, i - 1, j, k);
						vector(pf, i, j, k) = vector(pf, i - 1, j, k);
						vector(sqcs, i, j, k) = vector(sqcs, i - 1, j, k);
						vector(Tf, i, j, k) = vector(Tf, i - 1, j, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i - 1, j, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i - 1, j, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i - 1, j, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i - 1, j, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i - 1, j, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i - 1, j, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i - 1, j, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i - 1, j, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i - 1, j, k);
						vector(chie, i, j, k) = vector(chie, i - 1, j, k);
						vector(chia, i, j, k) = vector(chia, i - 1, j, k);
						vector(chix, i, j, k) = vector(chix, i - 1, j, k);
						vector(qnu, i, j, k) = vector(qnu, i - 1, j, k);
						vector(rnu, i, j, k) = vector(rnu, i - 1, j, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i - 1, j, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i - 1, j, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i - 1, j, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i - 1, j, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i - 1, j, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i - 1, j, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i - 1, j, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i - 1, j, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i - 1, j, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i - 1, j, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i - 1, j, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i - 1, j, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i - 1, j, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i - 1, j, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i - 1, j, k);
					}
					if ((vector(FOV_yLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, d_ghost_width, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, d_ghost_width, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, d_ghost_width, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, d_ghost_width, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, d_ghost_width, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, d_ghost_width, k);
						vector(rhof, i, j, k) = vector(rhof, i, d_ghost_width, k);
						vector(Yef, i, j, k) = vector(Yef, i, d_ghost_width, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, d_ghost_width, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, d_ghost_width, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, d_ghost_width, k);
						vector(epsf, i, j, k) = vector(epsf, i, d_ghost_width, k);
						vector(pf, i, j, k) = vector(pf, i, d_ghost_width, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, d_ghost_width, k);
						vector(Tf, i, j, k) = vector(Tf, i, d_ghost_width, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, d_ghost_width, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, d_ghost_width, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, d_ghost_width, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, d_ghost_width, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, d_ghost_width, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, d_ghost_width, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, d_ghost_width, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, d_ghost_width, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, d_ghost_width, k);
						vector(chie, i, j, k) = vector(chie, i, d_ghost_width, k);
						vector(chia, i, j, k) = vector(chia, i, d_ghost_width, k);
						vector(chix, i, j, k) = vector(chix, i, d_ghost_width, k);
						vector(qnu, i, j, k) = vector(qnu, i, d_ghost_width, k);
						vector(rnu, i, j, k) = vector(rnu, i, d_ghost_width, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, d_ghost_width, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, d_ghost_width, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, d_ghost_width, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, d_ghost_width, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, d_ghost_width, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, d_ghost_width, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, d_ghost_width, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, d_ghost_width, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, d_ghost_width, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, d_ghost_width, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, d_ghost_width, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, d_ghost_width, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, d_ghost_width, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, d_ghost_width, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, d_ghost_width, k);
					}
					if ((vector(FOV_yUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j - 1, k);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j - 1, k);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j - 1, k);
						vector(t_chie, i, j, k) = vector(t_chie, i, j - 1, k);
						vector(t_chia, i, j, k) = vector(t_chia, i, j - 1, k);
						vector(t_chix, i, j, k) = vector(t_chix, i, j - 1, k);
						vector(rhof, i, j, k) = vector(rhof, i, j - 1, k);
						vector(Yef, i, j, k) = vector(Yef, i, j - 1, k);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j - 1, k);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j - 1, k);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j - 1, k);
						vector(epsf, i, j, k) = vector(epsf, i, j - 1, k);
						vector(pf, i, j, k) = vector(pf, i, j - 1, k);
						vector(sqcs, i, j, k) = vector(sqcs, i, j - 1, k);
						vector(Tf, i, j, k) = vector(Tf, i, j - 1, k);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j - 1, k);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j - 1, k);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j - 1, k);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j - 1, k);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j - 1, k);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j - 1, k);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j - 1, k);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j - 1, k);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j - 1, k);
						vector(chie, i, j, k) = vector(chie, i, j - 1, k);
						vector(chia, i, j, k) = vector(chia, i, j - 1, k);
						vector(chix, i, j, k) = vector(chix, i, j - 1, k);
						vector(qnu, i, j, k) = vector(qnu, i, j - 1, k);
						vector(rnu, i, j, k) = vector(rnu, i, j - 1, k);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j - 1, k);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j - 1, k);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j - 1, k);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j - 1, k);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j - 1, k);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j - 1, k);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j - 1, k);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j - 1, k);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j - 1, k);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j - 1, k);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j - 1, k);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j - 1, k);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j - 1, k);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j - 1, k);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j - 1, k);
					}
					if ((vector(FOV_zLower, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, d_ghost_width);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, d_ghost_width);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, d_ghost_width);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, d_ghost_width);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, d_ghost_width);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, d_ghost_width);
						vector(rhof, i, j, k) = vector(rhof, i, j, d_ghost_width);
						vector(Yef, i, j, k) = vector(Yef, i, j, d_ghost_width);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, d_ghost_width);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, d_ghost_width);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, d_ghost_width);
						vector(epsf, i, j, k) = vector(epsf, i, j, d_ghost_width);
						vector(pf, i, j, k) = vector(pf, i, j, d_ghost_width);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, d_ghost_width);
						vector(Tf, i, j, k) = vector(Tf, i, j, d_ghost_width);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, d_ghost_width);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, d_ghost_width);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, d_ghost_width);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, d_ghost_width);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, d_ghost_width);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, d_ghost_width);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, d_ghost_width);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, d_ghost_width);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, d_ghost_width);
						vector(chie, i, j, k) = vector(chie, i, j, d_ghost_width);
						vector(chia, i, j, k) = vector(chia, i, j, d_ghost_width);
						vector(chix, i, j, k) = vector(chix, i, j, d_ghost_width);
						vector(qnu, i, j, k) = vector(qnu, i, j, d_ghost_width);
						vector(rnu, i, j, k) = vector(rnu, i, j, d_ghost_width);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, d_ghost_width);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, d_ghost_width);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, d_ghost_width);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, d_ghost_width);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, d_ghost_width);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, d_ghost_width);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, d_ghost_width);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, d_ghost_width);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, d_ghost_width);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, d_ghost_width);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, d_ghost_width);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, d_ghost_width);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, d_ghost_width);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, d_ghost_width);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, d_ghost_width);
					}
					if ((vector(FOV_zUpper, i, j, k) > 0)) {
						//Region field extrapolations
						if ((vector(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0 || vector(d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, j, k) != 0)) {
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, gtd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xx, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_xz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yy, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_yz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Atd_zz, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Gamh_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_x, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_y, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Betau_z, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, Alpha, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, chi, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, trK, FOV_1, dx, ilast, jlast);
							extrapolate_field(d_i_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, i, d_j_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, j, d_k_Gamh_x_Gamh_y_Gamh_z_Betau_x_Betau_y_Betau_z_Alpha_chi_trK_theta_gtd_xx_gtd_xy_gtd_xz_gtd_yy_gtd_yz_gtd_zz_Atd_xx_Atd_xy_Atd_xz_Atd_yy_Atd_yz_Atd_zz, k, theta, FOV_1, dx, ilast, jlast);
						}
						vector(t_optdepthe, i, j, k) = vector(t_optdepthe, i, j, k - 1);
						vector(t_optdeptha, i, j, k) = vector(t_optdeptha, i, j, k - 1);
						vector(t_optdepthx, i, j, k) = vector(t_optdepthx, i, j, k - 1);
						vector(t_chie, i, j, k) = vector(t_chie, i, j, k - 1);
						vector(t_chia, i, j, k) = vector(t_chia, i, j, k - 1);
						vector(t_chix, i, j, k) = vector(t_chix, i, j, k - 1);
						vector(rhof, i, j, k) = vector(rhof, i, j, k - 1);
						vector(Yef, i, j, k) = vector(Yef, i, j, k - 1);
						vector(vfd_x, i, j, k) = vector(vfd_x, i, j, k - 1);
						vector(vfd_y, i, j, k) = vector(vfd_y, i, j, k - 1);
						vector(vfd_z, i, j, k) = vector(vfd_z, i, j, k - 1);
						vector(epsf, i, j, k) = vector(epsf, i, j, k - 1);
						vector(pf, i, j, k) = vector(pf, i, j, k - 1);
						vector(sqcs, i, j, k) = vector(sqcs, i, j, k - 1);
						vector(Tf, i, j, k) = vector(Tf, i, j, k - 1);
						vector(dpfdrho, i, j, k) = vector(dpfdrho, i, j, k - 1);
						vector(dpfdeps, i, j, k) = vector(dpfdeps, i, j, k - 1);
						vector(dpfdye, i, j, k) = vector(dpfdye, i, j, k - 1);
						vector(Efu_x, i, j, k) = vector(Efu_x, i, j, k - 1);
						vector(Efu_y, i, j, k) = vector(Efu_y, i, j, k - 1);
						vector(Efu_z, i, j, k) = vector(Efu_z, i, j, k - 1);
						vector(optdepthe, i, j, k) = vector(optdepthe, i, j, k - 1);
						vector(optdeptha, i, j, k) = vector(optdeptha, i, j, k - 1);
						vector(optdepthx, i, j, k) = vector(optdepthx, i, j, k - 1);
						vector(chie, i, j, k) = vector(chie, i, j, k - 1);
						vector(chia, i, j, k) = vector(chia, i, j, k - 1);
						vector(chix, i, j, k) = vector(chix, i, j, k - 1);
						vector(qnu, i, j, k) = vector(qnu, i, j, k - 1);
						vector(rnu, i, j, k) = vector(rnu, i, j, k - 1);
						vector(TauN_x, i, j, k) = vector(TauN_x, i, j, k - 1);
						vector(TauN_y, i, j, k) = vector(TauN_y, i, j, k - 1);
						vector(TauN_z, i, j, k) = vector(TauN_z, i, j, k - 1);
						vector(TauNe_x, i, j, k) = vector(TauNe_x, i, j, k - 1);
						vector(TauNe_y, i, j, k) = vector(TauNe_y, i, j, k - 1);
						vector(TauNe_z, i, j, k) = vector(TauNe_z, i, j, k - 1);
						vector(TauM_xy, i, j, k) = vector(TauM_xy, i, j, k - 1);
						vector(TauM_xz, i, j, k) = vector(TauM_xz, i, j, k - 1);
						vector(TauM_yz, i, j, k) = vector(TauM_yz, i, j, k - 1);
						vector(TauT_xx, i, j, k) = vector(TauT_xx, i, j, k - 1);
						vector(TauT_xy, i, j, k) = vector(TauT_xy, i, j, k - 1);
						vector(TauT_xz, i, j, k) = vector(TauT_xz, i, j, k - 1);
						vector(TauT_yy, i, j, k) = vector(TauT_yy, i, j, k - 1);
						vector(TauT_yz, i, j, k) = vector(TauT_yz, i, j, k - 1);
						vector(TauT_zz, i, j, k) = vector(TauT_zz, i, j, k - 1);
					}
				}
			}
		}
	}
	if (d_refinedTimeStepping) {
		if (!hierarchy->finerLevelExists(ln) && last_step) {
			int currentLevelNumber = ln;
			while (currentLevelNumber > 0 && current_iteration[currentLevelNumber] % hierarchy->getRatioToCoarserLevel(currentLevelNumber).max() == 0) {
				d_coarsen_schedule[currentLevelNumber]->coarsenData();
				d_bdry_sched_postCoarsen[currentLevelNumber - 1]->fillData(current_time, false);
				currentLevelNumber--;
			}
		}
	} else {
		if (ln > 0) {
			d_coarsen_schedule[ln]->coarsenData();
			d_bdry_sched_postCoarsen[ln - 1]->fillData(current_time, false);
		}
	}
	

	t_step->stop();
	//Analysis
	bool calculate = false;
	if (previous_iteration < next_mesh_dump_iteration && outputCycle >= next_mesh_dump_iteration) {
		calculate = true;
	}
	for (int i = 0; i < next_slice_dump_iteration.size(); i++) {
		if (analysis_slice_dump[i] && previous_iteration < next_slice_dump_iteration[i] && outputCycle >= next_slice_dump_iteration[i]) {
			calculate = true;
		}
	}
	for (int i = 0; i < next_sphere_dump_iteration.size(); i++) {
		if (analysis_sphere_dump[i] && previous_iteration < next_sphere_dump_iteration[i] && outputCycle >= next_sphere_dump_iteration[i]) {
			calculate = true;
		}
	}
	for (int i = 0; i < next_integration_dump_iteration.size(); i++) {
		if (analysis_integration_dump[i] && previous_iteration < next_integration_dump_iteration[i] && outputCycle >= next_integration_dump_iteration[i]) {
			calculate = true;
		}
	}
	for (int i = 0; i < next_point_dump_iteration.size(); i++) {
		if (analysis_point_dump[i] && previous_iteration < next_point_dump_iteration[i] && outputCycle >= next_point_dump_iteration[i]) {
			calculate = true;
		}
	}
	if (calculate) {
		for (int ln=0; ln<=d_patch_hierarchy->getFinestLevelNumber(); ++ln ) {
			std::shared_ptr<hier::PatchLevel > level(d_patch_hierarchy->getPatchLevel(ln));
			for (hier::PatchLevel::iterator p_it(level->begin()); p_it != level->end(); ++p_it) {
				const std::shared_ptr<hier::Patch >& patch = *p_it;
		
				double* chi = ((pdat::NodeData<double> *) patch->getPatchData(d_chi_id).get())->getPointer();
				double* gtd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xx_id).get())->getPointer();
				double* Gamh_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_x_id).get())->getPointer();
				double* Gamh_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_y_id).get())->getPointer();
				double* Gamh_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Gamh_z_id).get())->getPointer();
				double* gtd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xy_id).get())->getPointer();
				double* gtd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_xz_id).get())->getPointer();
				double* gtd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yy_id).get())->getPointer();
				double* gtd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_yz_id).get())->getPointer();
				double* gtd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_gtd_zz_id).get())->getPointer();
				double* Atd_xy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xy_id).get())->getPointer();
				double* trK = ((pdat::NodeData<double> *) patch->getPatchData(d_trK_id).get())->getPointer();
				double* Atd_xz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xz_id).get())->getPointer();
				double* Atd_xx = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_xx_id).get())->getPointer();
				double* Atd_yy = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yy_id).get())->getPointer();
				double* Atd_yz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_yz_id).get())->getPointer();
				double* Atd_zz = ((pdat::NodeData<double> *) patch->getPatchData(d_Atd_zz_id).get())->getPointer();
				double* Bfu_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_x_id).get())->getPointer();
				double* Bfu_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_y_id).get())->getPointer();
				double* Bfu_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Bfu_z_id).get())->getPointer();
				double* Df = ((pdat::NodeData<double> *) patch->getPatchData(d_Df_id).get())->getPointer();
				double* tauf = ((pdat::NodeData<double> *) patch->getPatchData(d_tauf_id).get())->getPointer();
				double* Alpha = ((pdat::NodeData<double> *) patch->getPatchData(d_Alpha_id).get())->getPointer();
				double* rhof = ((pdat::NodeData<double> *) patch->getPatchData(d_rhof_id).get())->getPointer();
				double* epsf = ((pdat::NodeData<double> *) patch->getPatchData(d_epsf_id).get())->getPointer();
				double* pf = ((pdat::NodeData<double> *) patch->getPatchData(d_pf_id).get())->getPointer();
				double* Betau_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_x_id).get())->getPointer();
				double* Betau_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_y_id).get())->getPointer();
				double* Betau_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Betau_z_id).get())->getPointer();
				double* Sfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_x_id).get())->getPointer();
				double* Sfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_y_id).get())->getPointer();
				double* Sfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Sfd_z_id).get())->getPointer();
				double* vfd_x = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_x_id).get())->getPointer();
				double* vfd_y = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_y_id).get())->getPointer();
				double* vfd_z = ((pdat::NodeData<double> *) patch->getPatchData(d_vfd_z_id).get())->getPointer();
				double* Rscalar = ((pdat::NodeData<double> *) patch->getPatchData(d_Rscalar_id).get())->getPointer();
				double* HamCon = ((pdat::NodeData<double> *) patch->getPatchData(d_HamCon_id).get())->getPointer();
				double* MomCon_x = ((pdat::NodeData<double> *) patch->getPatchData(d_MomCon_x_id).get())->getPointer();
				double* MomCon_y = ((pdat::NodeData<double> *) patch->getPatchData(d_MomCon_y_id).get())->getPointer();
				double* MomCon_z = ((pdat::NodeData<double> *) patch->getPatchData(d_MomCon_z_id).get())->getPointer();
				double* trA = ((pdat::NodeData<double> *) patch->getPatchData(d_trA_id).get())->getPointer();
				double* detgtm1 = ((pdat::NodeData<double> *) patch->getPatchData(d_detgtm1_id).get())->getPointer();
				double* M_ADM_surf = ((pdat::NodeData<double> *) patch->getPatchData(d_M_ADM_surf_id).get())->getPointer();
				double* Jz_ADM_surf = ((pdat::NodeData<double> *) patch->getPatchData(d_Jz_ADM_surf_id).get())->getPointer();
				double* M_Komar = ((pdat::NodeData<double> *) patch->getPatchData(d_M_Komar_id).get())->getPointer();
				double* Jz_Komar = ((pdat::NodeData<double> *) patch->getPatchData(d_Jz_Komar_id).get())->getPointer();
				double* psi4R = ((pdat::NodeData<double> *) patch->getPatchData(d_psi4R_id).get())->getPointer();
				double* psi4I = ((pdat::NodeData<double> *) patch->getPatchData(d_psi4I_id).get())->getPointer();
				double* Z_x = ((pdat::NodeData<double> *) patch->getPatchData(d_Z_x_id).get())->getPointer();
				double* Z_y = ((pdat::NodeData<double> *) patch->getPatchData(d_Z_y_id).get())->getPointer();
				double* Z_z = ((pdat::NodeData<double> *) patch->getPatchData(d_Z_z_id).get())->getPointer();
				double* flux_mass = ((pdat::NodeData<double> *) patch->getPatchData(d_flux_mass_id).get())->getPointer();
				double d_Rpd_xx_o0_t0_m0_l0, d_Rpd_xx_o0_t1_m0_l0, d_Rpd_xx_o0_t2_m0_l0, d_Rpd_xx_o0_t3_m0_l0, d_Rpd_xx_o0_t4_m0_l0, d_Rpd_xx_o0_t5_m0_l0, d_Rpd_xx_o0_t6_m0_l0, d_Rpd_xx_o0_t7_m0_l0, d_Rpd_xx_o0_t8_m0_l0, d_Ctd_xxx_o0_t0_m0_l0, d_Ctd_xxy_o0_t0_m0_l0, d_Ctd_xxz_o0_t0_m0_l0, d_Rtd_xx_o0_t0_m0_l0, d_Rtd_xx_o0_t1_m0_l0, d_Rtd_xx_o0_t2_m0_l0, d_Rtd_xx_o0_t3_m0_l0, d_Rtd_xx_o0_t4_m0_l0, d_Rtd_xx_o0_t5_m0_l0, d_Rtd_xx_o0_t6_m0_l0, d_Rtd_xy_o0_t6_m0_l0, d_Rtd_xz_o0_t6_m0_l0, d_Rtd_xx_o0_t7_m0_l0, d_Rtd_xy_o0_t7_m0_l0, d_Rtd_xz_o0_t7_m0_l0, d_Rtd_xx_o0_t8_m0_l0, d_Rtd_xy_o0_t8_m0_l0, d_Rtd_xz_o0_t8_m0_l0, d_Ctd_yxx_o0_t0_m0_l0, d_Ctd_xyy_o0_t0_m0_l0, d_Ctd_xyz_o0_t1_m0_l0, d_Rtd_xy_o0_t0_m0_l0, d_Rtd_xy_o0_t1_m0_l0, d_Rtd_xy_o0_t2_m0_l0, d_Rtd_xy_o0_t3_m0_l0, d_Rtd_xy_o0_t4_m0_l0, d_Rtd_xy_o0_t5_m0_l0, d_Ctd_zxx_o0_t0_m0_l0, d_Ctd_xyz_o0_t0_m0_l0, d_Ctd_xzz_o0_t0_m0_l0, d_Rtd_xz_o0_t0_m0_l0, d_Rtd_xz_o0_t1_m0_l0, d_Rtd_xz_o0_t2_m0_l0, d_Rtd_xz_o0_t3_m0_l0, d_Rtd_xz_o0_t4_m0_l0, d_Rtd_xz_o0_t5_m0_l0, d_Ctd_xyy_o0_t1_m0_l0, d_Ctd_yyy_o0_t0_m0_l0, d_Ctd_yyz_o0_t0_m0_l0, d_Rtd_yy_o0_t0_m0_l0, d_Rtd_yy_o0_t1_m0_l0, d_Rtd_yy_o0_t2_m0_l0, d_Rtd_yy_o0_t3_m0_l0, d_Rtd_yy_o0_t4_m0_l0, d_Rtd_yy_o0_t5_m0_l0, d_Ctd_xyz_o0_t2_m0_l0, d_Ctd_zyy_o0_t0_m0_l0, d_Ctd_yzz_o0_t0_m0_l0, d_Rtd_yz_o0_t0_m0_l0, d_Rtd_yz_o0_t1_m0_l0, d_Rtd_yz_o0_t2_m0_l0, d_Rtd_yz_o0_t3_m0_l0, d_Rtd_yz_o0_t4_m0_l0, d_Rtd_yz_o0_t5_m0_l0, d_Ctd_xzz_o0_t1_m0_l0, d_Ctd_yzz_o0_t1_m0_l0, d_Ctd_zzz_o0_t0_m0_l0, d_Rtd_zz_o0_t0_m0_l0, d_Rtd_zz_o0_t1_m0_l0, d_Rtd_zz_o0_t2_m0_l0, d_Rtd_zz_o0_t3_m0_l0, d_Rtd_zz_o0_t4_m0_l0, d_Rtd_zz_o0_t5_m0_l0, d_Del_Kd_xxy_o0_t0_m0_l0, d_Del_Kd_yxy_o0_t0_m0_l0, d_Del_Kd_zxy_o0_t0_m0_l0, d_Del_Kd_xxy_o0_t2_m0_l0, d_Del_Kd_yxx_o0_t2_m0_l0, d_Del_Kd_zxx_o0_t2_m0_l0, d_Del_Kd_xxz_o0_t0_m0_l0, d_Del_Kd_yxz_o0_t0_m0_l0, d_Del_Kd_zxz_o0_t0_m0_l0, d_MomCon_x_o0_t0_m0_l0, d_Del_Kd_yxx_o0_t0_m0_l0, d_Del_Kd_zxx_o0_t0_m0_l0, d_Del_Kd_xyy_o0_t0_m0_l0, d_MomCon_y_o0_t4_m0_l0, d_Del_Kd_zyy_o0_t0_m0_l0, d_Del_Kd_xyz_o0_t0_m0_l0, d_Del_Kd_yyz_o0_t0_m0_l0, d_Del_Kd_zyz_o0_t0_m0_l0, d_Del_Kd_xzz_o0_t0_m0_l0, d_Del_Kd_yzz_o0_t0_m0_l0, d_MomCon_z_o0_t11_m0_l0, inv_chi, g4d_zz, g4d_yz, g4d_yy, Bfd_y, Bfd_z, Bfd_x, sqBf, r_pos, ur_z, ur_y, uph_z, uph_x, uph_y, ur_x, uthd_y, uthd_x, uthd_z, sdetg, Tf4u_tt, m_Ctd_zzz_o0_t0_l0, Ctd_zzz, hf, wrr, m_Ctd_yyz_o0_t0_l0, Ctd_yyz, m_Ctd_zyy_o0_t1_l0, m_Ctd_zyy_o0_t0_l0, Ctd_zyy, m_Ctd_yyy_o0_t0_l0, Ctd_yyy, m_Ctd_xxz_o0_t0_l0, Ctd_xxz, m_Ctd_zxx_o0_t1_l0, m_Ctd_zxx_o0_t0_l0, Ctd_zxx, m_Ctd_xxx_o0_t0_l0, Ctd_xxx, m_Ctd_yzz_o0_t1_l0, m_Ctd_yzz_o0_t0_l0, Ctd_yzz, m_Ctd_zyz_o0_t0_l0, Ctd_zyz, m_Ctd_xxy_o0_t0_l0, Ctd_xxy, m_Ctd_yxx_o0_t1_l0, m_Ctd_yxx_o0_t0_l0, Ctd_yxx, m_Ctd_xyy_o0_t1_l0, m_Ctd_xyy_o0_t0_l0, Ctd_xyy, m_Ctd_yxy_o0_t0_l0, Ctd_yxy, m_Ctd_xyz_o0_t2_l0, m_Ctd_xyz_o0_t1_l0, m_Ctd_xyz_o0_t0_l0, Ctd_xyz, m_Ctd_yxz_o0_t2_l0, m_Ctd_yxz_o0_t0_l0, Ctd_yxz, m_Ctd_zxy_o0_t2_l0, Ctd_zxy, g4d_tt, g4d_ty, g4d_tx, g4d_tz, m_Ctd_xzz_o0_t1_l0, m_Ctd_xzz_o0_t0_l0, Ctd_xzz, m_Ctd_zxz_o0_t0_l0, Ctd_zxz, dSigma_y, dSigma_x, dSigma_z, detgtd, idetgtd, gtu_zz, gtu_yy, gtu_yz, gtu_xx, gtu_xz, uth_z, Atud_zx, Atud_zz, Atud_zy, Atu_zz, Sfu_z, Tf4u_tz, Ct_zxz, Ct_zxy, Ct_zxx, Ct_zyz, Ct_zyy, Ct_zzz, gtu_xy, Ct_xzz, uth_x, uth_y, Gamt_z, Zu_z, Atud_yy, Atud_yx, Atud_yz, Atu_yz, Atud_xx, Atud_xz, Atud_xy, Atu_xz, Sfu_y, Tf4u_ty, Sfu_x, Tf4u_tx, vfu_y, Ju_y, vfu_x, Ju_x, Atu_yy, Ct_yxx, Ct_yxz, Ct_yxy, Atu_xy, Atu_xx, Ct_yyy, Ct_yyz, Tu_ty, Tu_tx, Tu_tz, Ct_yzz, Gamt_y, Zu_y, Ct_xxy, vfu_z, sqvf, sqW, Ju_z, Ct_xxx, Ct_xxz, d_Rtd_xx_o0_t9_m0_l0, m_Rtd_xx_o0_t8_l0, m_Rtd_xx_o0_t7_l0, m_Rtd_xx_o0_t6_l0, m_Rtd_xx_o0_t5_l0, m_Rtd_xx_o0_t4_l0, m_Rtd_xx_o0_t3_l0, m_Rtd_xx_o0_t2_l0, m_Rtd_xx_o0_t1_l0, m_Rtd_xx_o0_t0_l0, Rtd_xx, g4d_xy, wphph, vph_y, vph_z, vph_x, wthph, g4d_xx, wrph, vr_x, vr_y, vr_z, wthr, g4d_xz, wthth, vth_x, mmR_xx, mmI_xx, vth_y, mmR_yy, mmR_xy, mmI_xy, mmI_yy, vth_z, mmR_yz, mmR_zz, mmR_xz, mmI_zz, mmI_xz, mmI_yz, Ct_xyz, d_Rtd_xz_o0_t12_m0_l0, m_Rtd_xz_o0_t11_l0, m_Rtd_xz_o0_t10_l0, m_Rtd_xz_o0_t9_l0, m_Rtd_xz_o0_t8_l0, m_Rtd_xz_o0_t7_l0, m_Rtd_xz_o0_t6_l0, m_Rtd_xz_o0_t5_l0, m_Rtd_xz_o0_t4_l0, m_Rtd_xz_o0_t3_l0, m_Rtd_xz_o0_t2_l0, m_Rtd_xz_o0_t1_l0, m_Rtd_xz_o0_t0_l0, Rtd_xz, d_Rtd_zz_o0_t9_m0_l0, m_Rtd_zz_o0_t8_l0, m_Rtd_zz_o0_t7_l0, m_Rtd_zz_o0_t6_l0, m_Rtd_zz_o0_t5_l0, m_Rtd_zz_o0_t4_l0, m_Rtd_zz_o0_t3_l0, m_Rtd_zz_o0_t2_l0, m_Rtd_zz_o0_t1_l0, m_Rtd_zz_o0_t0_l0, Rtd_zz, Ct_xyy, Gamt_x, Zu_x, m_Rpd_xx_o0_t22_l0, m_Rpd_xx_o0_t21_l0, m_Rpd_xx_o0_t20_l0, m_Rpd_xx_o0_t19_l0, m_Rpd_xx_o0_t18_l0, m_Rpd_xx_o0_t17_l0, m_Rpd_xx_o0_t16_l0, m_Rpd_xx_o0_t15_l0, m_Rpd_xx_o0_t14_l0, m_Rpd_xx_o0_t13_l0, m_Rpd_xx_o0_t12_l0, m_Rpd_xx_o0_t11_l0, m_Rpd_xx_o0_t10_l0, m_Rpd_xx_o0_t9_l0, m_Rpd_xx_o0_t8_l0, m_Rpd_xx_o0_t7_l0, m_Rpd_xx_o0_t6_l0, m_Rpd_xx_o0_t5_l0, m_Rpd_xx_o0_t4_l0, m_Rpd_xx_o0_t3_l0, m_Rpd_xx_o0_t2_l0, m_Rpd_xx_o0_t1_l0, m_Rpd_xx_o0_t0_l0, Rpd_xx, d_Del_Kd_yxx_o0_t1_m0_l0, m_Del_Kd_yxx_o0_t6_l0, m_Del_Kd_yxx_o0_t5_l0, m_Del_Kd_yxx_o0_t4_l0, m_Del_Kd_yxx_o0_t3_l0, m_Del_Kd_yxx_o0_t2_l0, m_Del_Kd_yxx_o0_t0_l0, Del_Kd_yxx, d_Del_Kd_yxz_o0_t1_m0_l0, m_Del_Kd_yxz_o0_t10_l0, m_Del_Kd_yxz_o0_t9_l0, m_Del_Kd_yxz_o0_t8_l0, m_Del_Kd_yxz_o0_t7_l0, m_Del_Kd_yxz_o0_t6_l0, m_Del_Kd_yxz_o0_t5_l0, m_Del_Kd_yxz_o0_t4_l0, m_Del_Kd_yxz_o0_t3_l0, m_Del_Kd_yxz_o0_t2_l0, m_Del_Kd_yxz_o0_t0_l0, Del_Kd_yxz, EWeyl_xx, d_Del_Kd_zxx_o0_t1_m0_l0, m_Del_Kd_zxx_o0_t6_l0, m_Del_Kd_zxx_o0_t5_l0, m_Del_Kd_zxx_o0_t4_l0, m_Del_Kd_zxx_o0_t3_l0, m_Del_Kd_zxx_o0_t2_l0, m_Del_Kd_zxx_o0_t0_l0, Del_Kd_zxx, d_Del_Kd_zyz_o0_t1_m0_l0, m_Del_Kd_zyz_o0_t10_l0, m_Del_Kd_zyz_o0_t9_l0, m_Del_Kd_zyz_o0_t8_l0, m_Del_Kd_zyz_o0_t7_l0, m_Del_Kd_zyz_o0_t6_l0, m_Del_Kd_zyz_o0_t5_l0, m_Del_Kd_zyz_o0_t4_l0, m_Del_Kd_zyz_o0_t3_l0, m_Del_Kd_zyz_o0_t2_l0, m_Del_Kd_zyz_o0_t0_l0, Del_Kd_zyz, d_Del_Kd_zyy_o0_t1_m0_l0, m_Del_Kd_zyy_o0_t6_l0, m_Del_Kd_zyy_o0_t5_l0, m_Del_Kd_zyy_o0_t4_l0, m_Del_Kd_zyy_o0_t3_l0, m_Del_Kd_zyy_o0_t2_l0, m_Del_Kd_zyy_o0_t0_l0, Del_Kd_zyy, d_Del_Kd_zxy_o0_t1_m0_l0, m_Del_Kd_zxy_o0_t7_l0, m_Del_Kd_zxy_o0_t6_l0, m_Del_Kd_zxy_o0_t5_l0, m_Del_Kd_zxy_o0_t4_l0, m_Del_Kd_zxy_o0_t2_l0, m_Del_Kd_zxy_o0_t0_l0, Del_Kd_zxy, d_Del_Kd_zxz_o0_t1_m0_l0, m_Del_Kd_zxz_o0_t10_l0, m_Del_Kd_zxz_o0_t9_l0, m_Del_Kd_zxz_o0_t8_l0, m_Del_Kd_zxz_o0_t7_l0, m_Del_Kd_zxz_o0_t6_l0, m_Del_Kd_zxz_o0_t5_l0, m_Del_Kd_zxz_o0_t4_l0, m_Del_Kd_zxz_o0_t3_l0, m_Del_Kd_zxz_o0_t2_l0, m_Del_Kd_zxz_o0_t0_l0, Del_Kd_zxz, m_Rpd_yz_o0_t25_l0, m_Rpd_yz_o0_t24_l0, m_Rpd_yz_o0_t23_l0, m_Rpd_yz_o0_t22_l0, m_Rpd_yz_o0_t21_l0, m_Rpd_yz_o0_t20_l0, m_Rpd_yz_o0_t19_l0, m_Rpd_yz_o0_t18_l0, m_Rpd_yz_o0_t17_l0, m_Rpd_yz_o0_t16_l0, m_Rpd_yz_o0_t15_l0, m_Rpd_yz_o0_t14_l0, m_Rpd_yz_o0_t13_l0, m_Rpd_yz_o0_t12_l0, m_Rpd_yz_o0_t11_l0, m_Rpd_yz_o0_t10_l0, m_Rpd_yz_o0_t9_l0, m_Rpd_yz_o0_t8_l0, m_Rpd_yz_o0_t7_l0, m_Rpd_yz_o0_t6_l0, m_Rpd_yz_o0_t5_l0, m_Rpd_yz_o0_t4_l0, m_Rpd_yz_o0_t3_l0, m_Rpd_yz_o0_t2_l0, m_Rpd_yz_o0_t1_l0, m_Rpd_yz_o0_t0_l0, Rpd_yz, d_Del_Kd_yzz_o0_t1_m0_l0, m_Del_Kd_yzz_o0_t6_l0, m_Del_Kd_yzz_o0_t5_l0, m_Del_Kd_yzz_o0_t4_l0, m_Del_Kd_yzz_o0_t3_l0, m_Del_Kd_yzz_o0_t2_l0, m_Del_Kd_yzz_o0_t0_l0, Del_Kd_yzz, d_Del_Kd_xxy_o0_t1_m0_l0, m_Del_Kd_xxy_o0_t10_l0, m_Del_Kd_xxy_o0_t9_l0, m_Del_Kd_xxy_o0_t8_l0, m_Del_Kd_xxy_o0_t7_l0, m_Del_Kd_xxy_o0_t6_l0, m_Del_Kd_xxy_o0_t5_l0, m_Del_Kd_xxy_o0_t4_l0, m_Del_Kd_xxy_o0_t3_l0, m_Del_Kd_xxy_o0_t2_l0, m_Del_Kd_xxy_o0_t0_l0, Del_Kd_xxy, d_Del_Kd_xzz_o0_t1_m0_l0, m_Del_Kd_xzz_o0_t6_l0, m_Del_Kd_xzz_o0_t5_l0, m_Del_Kd_xzz_o0_t4_l0, m_Del_Kd_xzz_o0_t3_l0, m_Del_Kd_xzz_o0_t2_l0, m_Del_Kd_xzz_o0_t0_l0, Del_Kd_xzz, d_Del_Kd_xyz_o0_t1_m0_l0, m_Del_Kd_xyz_o0_t2_l0, m_Del_Kd_xyz_o0_t0_l0, Del_Kd_xyz, BWeyl_xz, BWeyl_zz, BWeyl_yz, d_Del_Kd_xyy_o0_t1_m0_l0, m_Del_Kd_xyy_o0_t6_l0, m_Del_Kd_xyy_o0_t5_l0, m_Del_Kd_xyy_o0_t4_l0, m_Del_Kd_xyy_o0_t3_l0, m_Del_Kd_xyy_o0_t2_l0, m_Del_Kd_xyy_o0_t0_l0, Del_Kd_xyy, d_Del_Kd_xxz_o0_t1_m0_l0, m_Del_Kd_xxz_o0_t10_l0, m_Del_Kd_xxz_o0_t9_l0, m_Del_Kd_xxz_o0_t8_l0, m_Del_Kd_xxz_o0_t7_l0, m_Del_Kd_xxz_o0_t6_l0, m_Del_Kd_xxz_o0_t5_l0, m_Del_Kd_xxz_o0_t4_l0, m_Del_Kd_xxz_o0_t3_l0, m_Del_Kd_xxz_o0_t2_l0, m_Del_Kd_xxz_o0_t0_l0, Del_Kd_xxz, BWeyl_xx, BWeyl_zx, BWeyl_yx, m_Rpd_xz_o0_t25_l0, m_Rpd_xz_o0_t24_l0, m_Rpd_xz_o0_t23_l0, m_Rpd_xz_o0_t22_l0, m_Rpd_xz_o0_t21_l0, m_Rpd_xz_o0_t20_l0, m_Rpd_xz_o0_t19_l0, m_Rpd_xz_o0_t18_l0, m_Rpd_xz_o0_t17_l0, m_Rpd_xz_o0_t16_l0, m_Rpd_xz_o0_t15_l0, m_Rpd_xz_o0_t14_l0, m_Rpd_xz_o0_t13_l0, m_Rpd_xz_o0_t12_l0, m_Rpd_xz_o0_t11_l0, m_Rpd_xz_o0_t10_l0, m_Rpd_xz_o0_t9_l0, m_Rpd_xz_o0_t8_l0, m_Rpd_xz_o0_t7_l0, m_Rpd_xz_o0_t6_l0, m_Rpd_xz_o0_t5_l0, m_Rpd_xz_o0_t4_l0, m_Rpd_xz_o0_t3_l0, m_Rpd_xz_o0_t2_l0, m_Rpd_xz_o0_t1_l0, m_Rpd_xz_o0_t0_l0, Rpd_xz, EWeyl_xz, m_Rpd_xy_o0_t25_l0, m_Rpd_xy_o0_t24_l0, m_Rpd_xy_o0_t23_l0, m_Rpd_xy_o0_t22_l0, m_Rpd_xy_o0_t21_l0, m_Rpd_xy_o0_t20_l0, m_Rpd_xy_o0_t19_l0, m_Rpd_xy_o0_t18_l0, m_Rpd_xy_o0_t17_l0, m_Rpd_xy_o0_t16_l0, m_Rpd_xy_o0_t15_l0, m_Rpd_xy_o0_t14_l0, m_Rpd_xy_o0_t13_l0, m_Rpd_xy_o0_t12_l0, m_Rpd_xy_o0_t11_l0, m_Rpd_xy_o0_t10_l0, m_Rpd_xy_o0_t9_l0, m_Rpd_xy_o0_t8_l0, m_Rpd_xy_o0_t7_l0, m_Rpd_xy_o0_t6_l0, m_Rpd_xy_o0_t5_l0, m_Rpd_xy_o0_t4_l0, m_Rpd_xy_o0_t3_l0, m_Rpd_xy_o0_t2_l0, m_Rpd_xy_o0_t1_l0, m_Rpd_xy_o0_t0_l0, Rpd_xy, m_Rpd_zz_o0_t22_l0, m_Rpd_zz_o0_t21_l0, m_Rpd_zz_o0_t20_l0, m_Rpd_zz_o0_t19_l0, m_Rpd_zz_o0_t18_l0, m_Rpd_zz_o0_t17_l0, m_Rpd_zz_o0_t16_l0, m_Rpd_zz_o0_t15_l0, m_Rpd_zz_o0_t14_l0, m_Rpd_zz_o0_t13_l0, m_Rpd_zz_o0_t12_l0, m_Rpd_zz_o0_t11_l0, m_Rpd_zz_o0_t10_l0, m_Rpd_zz_o0_t9_l0, m_Rpd_zz_o0_t8_l0, m_Rpd_zz_o0_t7_l0, m_Rpd_zz_o0_t6_l0, m_Rpd_zz_o0_t5_l0, m_Rpd_zz_o0_t4_l0, m_Rpd_zz_o0_t3_l0, m_Rpd_zz_o0_t2_l0, m_Rpd_zz_o0_t1_l0, m_Rpd_zz_o0_t0_l0, Rpd_zz, EWeyl_zz, d_Del_Kd_yyz_o0_t1_m0_l0, m_Del_Kd_yyz_o0_t10_l0, m_Del_Kd_yyz_o0_t9_l0, m_Del_Kd_yyz_o0_t8_l0, m_Del_Kd_yyz_o0_t7_l0, m_Del_Kd_yyz_o0_t6_l0, m_Del_Kd_yyz_o0_t5_l0, m_Del_Kd_yyz_o0_t4_l0, m_Del_Kd_yyz_o0_t3_l0, m_Del_Kd_yyz_o0_t2_l0, m_Del_Kd_yyz_o0_t0_l0, Del_Kd_yyz, d_Del_Kd_yxy_o0_t1_m0_l0, m_Del_Kd_yxy_o0_t10_l0, m_Del_Kd_yxy_o0_t9_l0, m_Del_Kd_yxy_o0_t8_l0, m_Del_Kd_yxy_o0_t7_l0, m_Del_Kd_yxy_o0_t6_l0, m_Del_Kd_yxy_o0_t5_l0, m_Del_Kd_yxy_o0_t4_l0, m_Del_Kd_yxy_o0_t3_l0, m_Del_Kd_yxy_o0_t2_l0, m_Del_Kd_yxy_o0_t0_l0, Del_Kd_yxy, BWeyl_xy, BWeyl_zy, BWeyl_yy, m_Rpd_yy_o0_t22_l0, m_Rpd_yy_o0_t21_l0, m_Rpd_yy_o0_t20_l0, m_Rpd_yy_o0_t19_l0, m_Rpd_yy_o0_t18_l0, m_Rpd_yy_o0_t17_l0, m_Rpd_yy_o0_t16_l0, m_Rpd_yy_o0_t15_l0, m_Rpd_yy_o0_t14_l0, m_Rpd_yy_o0_t13_l0, m_Rpd_yy_o0_t12_l0, m_Rpd_yy_o0_t11_l0, m_Rpd_yy_o0_t10_l0, m_Rpd_yy_o0_t9_l0, m_Rpd_yy_o0_t8_l0, m_Rpd_yy_o0_t7_l0, m_Rpd_yy_o0_t6_l0, m_Rpd_yy_o0_t5_l0, m_Rpd_yy_o0_t4_l0, m_Rpd_yy_o0_t3_l0, m_Rpd_yy_o0_t2_l0, m_Rpd_yy_o0_t1_l0, m_Rpd_yy_o0_t0_l0, Rpd_yy, d_Rtd_xy_o0_t12_m0_l0, m_Rtd_xy_o0_t11_l0, m_Rtd_xy_o0_t10_l0, m_Rtd_xy_o0_t9_l0, m_Rtd_xy_o0_t8_l0, m_Rtd_xy_o0_t7_l0, m_Rtd_xy_o0_t6_l0, m_Rtd_xy_o0_t5_l0, m_Rtd_xy_o0_t4_l0, m_Rtd_xy_o0_t3_l0, m_Rtd_xy_o0_t2_l0, m_Rtd_xy_o0_t1_l0, m_Rtd_xy_o0_t0_l0, Rtd_xy, EWeyl_xy, d_Rtd_yy_o0_t9_m0_l0, m_Rtd_yy_o0_t8_l0, m_Rtd_yy_o0_t7_l0, m_Rtd_yy_o0_t6_l0, m_Rtd_yy_o0_t5_l0, m_Rtd_yy_o0_t4_l0, m_Rtd_yy_o0_t3_l0, m_Rtd_yy_o0_t2_l0, m_Rtd_yy_o0_t1_l0, m_Rtd_yy_o0_t0_l0, Rtd_yy, EWeyl_yy, d_Rtd_yz_o0_t12_m0_l0, m_Rtd_yz_o0_t11_l0, m_Rtd_yz_o0_t10_l0, m_Rtd_yz_o0_t9_l0, m_Rtd_yz_o0_t8_l0, m_Rtd_yz_o0_t7_l0, m_Rtd_yz_o0_t6_l0, m_Rtd_yz_o0_t5_l0, m_Rtd_yz_o0_t4_l0, m_Rtd_yz_o0_t3_l0, m_Rtd_yz_o0_t2_l0, m_Rtd_yz_o0_t1_l0, m_Rtd_yz_o0_t0_l0, Rtd_yz, EWeyl_yz, Bfvf, Sfuu_zz, Tf4u_zz, Tu_zz, Sfuu_yz, Tf4u_yz, Tu_yz, Sfuu_yy, Tf4u_yy, Tu_yy, Sfuu_xy, Tf4u_xy, Tu_xy, Sfuu_xx, Tf4u_xx, Tu_xx, Sfuu_xz, Tf4u_xz, Tu_xz, Tu_tt, Td_yz, Td_yy, Td_xz, Td_xy, Td_xx, Td_zz, TT, Td_tx, Td_tz, Td_ty, Jtd_ADM_z, Td_tt, Jtd_ADM_y, Jtd_ADM_x, rho_ADM, d_Rscalar_o0_t0_m0_l0, d_HamCon_o0_t0_m0_l0, d_MomCon_x_o0_t13_m0_l0, d_MomCon_y_o0_t13_m0_l0, d_MomCon_z_o0_t13_m0_l0, d_trA_o0_t0_m0_l0, d_detgtm1_o0_t0_m0_l0, d_M_ADM_surf_o0_t9_m0_l0, d_Jz_ADM_surf_o0_t0_m0_l0, d_M_Komar_o0_t0_m0_l0, d_Jz_Komar_o0_t0_m0_l0, d_psi4R_o0_t0_m0_l0, d_psi4I_o0_t0_m0_l0, d_Z_x_o0_t0_m0_l0, d_Z_y_o0_t0_m0_l0, d_Z_z_o0_t0_m0_l0, d_flux_mass_o0_t0_m0_l0, m_M_ADM_surf_o0_t8_l0, m_M_ADM_surf_o0_t7_l0, m_M_ADM_surf_o0_t6_l0, m_M_ADM_surf_o0_t5_l0, m_M_ADM_surf_o0_t4_l0, m_M_ADM_surf_o0_t3_l0, m_M_ADM_surf_o0_t2_l0, m_M_ADM_surf_o0_t1_l0, m_M_ADM_surf_o0_t0_l0, m_MomCon_z_o0_t12_l0, m_MomCon_z_o0_t11_l0, m_MomCon_z_o0_t10_l0, m_MomCon_z_o0_t9_l0, m_MomCon_z_o0_t8_l0, m_MomCon_z_o0_t7_l0, m_MomCon_z_o0_t6_l0, m_MomCon_z_o0_t5_l0, m_MomCon_z_o0_t4_l0, m_MomCon_z_o0_t3_l0, m_MomCon_z_o0_t2_l0, m_MomCon_z_o0_t1_l0, m_MomCon_z_o0_t0_l0, m_MomCon_y_o0_t12_l0, m_MomCon_y_o0_t11_l0, m_MomCon_y_o0_t10_l0, m_MomCon_y_o0_t9_l0, m_MomCon_y_o0_t8_l0, m_MomCon_y_o0_t7_l0, m_MomCon_y_o0_t6_l0, m_MomCon_y_o0_t5_l0, m_MomCon_y_o0_t4_l0, m_MomCon_y_o0_t3_l0, m_MomCon_y_o0_t2_l0, m_MomCon_y_o0_t1_l0, m_MomCon_y_o0_t0_l0, m_MomCon_x_o0_t12_l0, m_MomCon_x_o0_t11_l0, m_MomCon_x_o0_t10_l0, m_MomCon_x_o0_t9_l0, m_MomCon_x_o0_t8_l0, m_MomCon_x_o0_t7_l0, m_MomCon_x_o0_t6_l0, m_MomCon_x_o0_t5_l0, m_MomCon_x_o0_t4_l0, m_MomCon_x_o0_t3_l0, m_MomCon_x_o0_t2_l0, m_MomCon_x_o0_t1_l0, m_MomCon_x_o0_t0_l0;
		
				//Get the dimensions of the patch
				hier::Box pbox = patch->getBox();
				const hier::Index boxfirst = patch->getBox().lower();
				const hier::Index boxlast = patch->getBox().upper();
		
				//Get delta spaces into an array. dx, dy, dz.
				std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
				const double* dx  = patch_geom->getDx();
		
				//Auxiliary definitions
				int ilast = boxlast(0)-boxfirst(0) + 2 + 2 * d_ghost_width;
				int jlast = boxlast(1)-boxfirst(1) + 2 + 2 * d_ghost_width;
				int klast = boxlast(2)-boxfirst(2) + 2 + 2 * d_ghost_width;
			for(int k = 0; k < klast; k++) {
				for(int j = 0; j < jlast; j++) {
					for(int i = 0; i < ilast; i++) {
							if ((i + 2 < ilast && i - 2 >= 0 && j + 2 < jlast && j - 2 >= 0 && k + 2 < klast && k - 2 >= 0)) {
								d_Rpd_xx_o0_t0_m0_l0 = D1CDO4_i(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rpd_xx_o0_t1_m0_l0 = D1CDO4_j(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rpd_xx_o0_t2_m0_l0 = D1CDO4_k(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rpd_xx_o0_t3_m0_l0 = D2CDO4_i(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rpd_xx_o0_t4_m0_l0 = D1CDO4crossed_ij(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rpd_xx_o0_t5_m0_l0 = D1CDO4crossed_ik(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rpd_xx_o0_t6_m0_l0 = D2CDO4_j(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rpd_xx_o0_t7_m0_l0 = D1CDO4crossed_jk(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rpd_xx_o0_t8_m0_l0 = D2CDO4_k(chi, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_xxx_o0_t0_m0_l0 = D1CDO4_i(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_xxy_o0_t0_m0_l0 = D1CDO4_j(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_xxz_o0_t0_m0_l0 = D1CDO4_k(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xx_o0_t0_m0_l0 = D2CDO4_i(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xx_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xx_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xx_o0_t3_m0_l0 = D2CDO4_j(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xx_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xx_o0_t5_m0_l0 = D2CDO4_k(gtd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xx_o0_t6_m0_l0 = D1CDO4_i(Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xy_o0_t6_m0_l0 = D1CDO4_j(Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xz_o0_t6_m0_l0 = D1CDO4_k(Gamh_x, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xx_o0_t7_m0_l0 = D1CDO4_i(Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xy_o0_t7_m0_l0 = D1CDO4_j(Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xz_o0_t7_m0_l0 = D1CDO4_k(Gamh_y, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xx_o0_t8_m0_l0 = D1CDO4_i(Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xy_o0_t8_m0_l0 = D1CDO4_j(Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xz_o0_t8_m0_l0 = D1CDO4_k(Gamh_z, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_yxx_o0_t0_m0_l0 = D1CDO4_i(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_xyy_o0_t0_m0_l0 = D1CDO4_j(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_xyz_o0_t1_m0_l0 = D1CDO4_k(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xy_o0_t0_m0_l0 = D2CDO4_i(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xy_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xy_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xy_o0_t3_m0_l0 = D2CDO4_j(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xy_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xy_o0_t5_m0_l0 = D2CDO4_k(gtd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_zxx_o0_t0_m0_l0 = D1CDO4_i(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_xyz_o0_t0_m0_l0 = D1CDO4_j(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_xzz_o0_t0_m0_l0 = D1CDO4_k(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xz_o0_t0_m0_l0 = D2CDO4_i(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xz_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xz_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xz_o0_t3_m0_l0 = D2CDO4_j(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xz_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_xz_o0_t5_m0_l0 = D2CDO4_k(gtd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_xyy_o0_t1_m0_l0 = D1CDO4_i(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_yyy_o0_t0_m0_l0 = D1CDO4_j(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_yyz_o0_t0_m0_l0 = D1CDO4_k(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_yy_o0_t0_m0_l0 = D2CDO4_i(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_yy_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_yy_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_yy_o0_t3_m0_l0 = D2CDO4_j(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_yy_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_yy_o0_t5_m0_l0 = D2CDO4_k(gtd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_xyz_o0_t2_m0_l0 = D1CDO4_i(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_zyy_o0_t0_m0_l0 = D1CDO4_j(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_yzz_o0_t0_m0_l0 = D1CDO4_k(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_yz_o0_t0_m0_l0 = D2CDO4_i(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_yz_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_yz_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_yz_o0_t3_m0_l0 = D2CDO4_j(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_yz_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_yz_o0_t5_m0_l0 = D2CDO4_k(gtd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_xzz_o0_t1_m0_l0 = D1CDO4_i(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_yzz_o0_t1_m0_l0 = D1CDO4_j(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Ctd_zzz_o0_t0_m0_l0 = D1CDO4_k(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_zz_o0_t0_m0_l0 = D2CDO4_i(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_zz_o0_t1_m0_l0 = D1CDO4crossed_ij(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_zz_o0_t2_m0_l0 = D1CDO4crossed_ik(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_zz_o0_t3_m0_l0 = D2CDO4_j(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_zz_o0_t4_m0_l0 = D1CDO4crossed_jk(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Rtd_zz_o0_t5_m0_l0 = D2CDO4_k(gtd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_xxy_o0_t0_m0_l0 = D1CDO4_i(Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_yxy_o0_t0_m0_l0 = D1CDO4_j(Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_zxy_o0_t0_m0_l0 = D1CDO4_k(Atd_xy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_xxy_o0_t2_m0_l0 = D1CDO4_i(trK, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_yxx_o0_t2_m0_l0 = D1CDO4_j(trK, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_zxx_o0_t2_m0_l0 = D1CDO4_k(trK, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_xxz_o0_t0_m0_l0 = D1CDO4_i(Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_yxz_o0_t0_m0_l0 = D1CDO4_j(Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_zxz_o0_t0_m0_l0 = D1CDO4_k(Atd_xz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_MomCon_x_o0_t0_m0_l0 = D1CDO4_i(Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_yxx_o0_t0_m0_l0 = D1CDO4_j(Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_zxx_o0_t0_m0_l0 = D1CDO4_k(Atd_xx, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_xyy_o0_t0_m0_l0 = D1CDO4_i(Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_MomCon_y_o0_t4_m0_l0 = D1CDO4_j(Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_zyy_o0_t0_m0_l0 = D1CDO4_k(Atd_yy, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_xyz_o0_t0_m0_l0 = D1CDO4_i(Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_yyz_o0_t0_m0_l0 = D1CDO4_j(Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_zyz_o0_t0_m0_l0 = D1CDO4_k(Atd_yz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_xzz_o0_t0_m0_l0 = D1CDO4_i(Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_Del_Kd_yzz_o0_t0_m0_l0 = D1CDO4_j(Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
								d_MomCon_z_o0_t11_m0_l0 = D1CDO4_k(Atd_zz, i, j, k, dx, simPlat_dt, ilast, jlast);
								inv_chi = 1.0 / fabs(vector(chi, i, j, k));
								g4d_zz = inv_chi * vector(gtd_zz, i, j, k);
								g4d_yz = inv_chi * vector(gtd_yz, i, j, k);
								g4d_yy = inv_chi * vector(gtd_yy, i, j, k);
								Bfd_y = inv_chi * vector(gtd_xy, i, j, k) * vector(Bfu_x, i, j, k) + inv_chi * vector(gtd_yy, i, j, k) * vector(Bfu_y, i, j, k) + inv_chi * vector(gtd_yz, i, j, k) * vector(Bfu_z, i, j, k);
								Bfd_z = inv_chi * vector(gtd_xz, i, j, k) * vector(Bfu_x, i, j, k) + inv_chi * vector(gtd_yz, i, j, k) * vector(Bfu_y, i, j, k) + inv_chi * vector(gtd_zz, i, j, k) * vector(Bfu_z, i, j, k);
								Bfd_x = inv_chi * vector(gtd_xx, i, j, k) * vector(Bfu_x, i, j, k) + inv_chi * vector(gtd_xy, i, j, k) * vector(Bfu_y, i, j, k) + inv_chi * vector(gtd_xz, i, j, k) * vector(Bfu_z, i, j, k);
								sqBf = vector(Bfu_x, i, j, k) * Bfd_x + vector(Bfu_y, i, j, k) * Bfd_y + vector(Bfu_z, i, j, k) * Bfd_z;
								r_pos = MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)));
								ur_z = zcoord(k);
								ur_y = ycoord(j);
								uph_z = 0.0;
								uph_x = -ycoord(j);
								uph_y = xcoord(i);
								ur_x = xcoord(i);
								uthd_y = ycoord(j) * zcoord(k);
								uthd_x = xcoord(i) * zcoord(k);
								uthd_z = (-xcoord(i) * xcoord(i)) + (-ycoord(j) * ycoord(j));
								sdetg = pow(vector(chi, i, j, k), (-1.500000000000000));
								Tf4u_tt = (vector(Df, i, j, k) + vector(tauf, i, j, k)) / (sdetg * (vector(Alpha, i, j, k) * vector(Alpha, i, j, k)));
								m_Ctd_zzz_o0_t0_l0 = 0.5 * d_Ctd_zzz_o0_t0_m0_l0;
								Ctd_zzz = m_Ctd_zzz_o0_t0_l0;
								hf = vector(rhof, i, j, k) * (1.0 + vector(epsf, i, j, k)) + vector(pf, i, j, k);
								wrr = inv_chi * vector(gtd_xx, i, j, k) * (ur_x * ur_x) + 2.0 * inv_chi * vector(gtd_xy, i, j, k) * ur_x * ur_y + 2.0 * inv_chi * vector(gtd_xz, i, j, k) * ur_x * ur_z + inv_chi * vector(gtd_yy, i, j, k) * (ur_y * ur_y) + 2.0 * inv_chi * vector(gtd_yz, i, j, k) * ur_y * ur_z + inv_chi * vector(gtd_zz, i, j, k) * (ur_z * ur_z);
								m_Ctd_yyz_o0_t0_l0 = 0.5 * d_Ctd_yyz_o0_t0_m0_l0;
								Ctd_yyz = m_Ctd_yyz_o0_t0_l0;
								m_Ctd_zyy_o0_t1_l0 = (-0.5) * d_Ctd_yyz_o0_t0_m0_l0;
								m_Ctd_zyy_o0_t0_l0 = 1.0 * d_Ctd_zyy_o0_t0_m0_l0;
								Ctd_zyy = m_Ctd_zyy_o0_t0_l0 + m_Ctd_zyy_o0_t1_l0;
								m_Ctd_yyy_o0_t0_l0 = 0.5 * d_Ctd_yyy_o0_t0_m0_l0;
								Ctd_yyy = m_Ctd_yyy_o0_t0_l0;
								m_Ctd_xxz_o0_t0_l0 = 0.5 * d_Ctd_xxz_o0_t0_m0_l0;
								Ctd_xxz = m_Ctd_xxz_o0_t0_l0;
								m_Ctd_zxx_o0_t1_l0 = (-0.5) * d_Ctd_xxz_o0_t0_m0_l0;
								m_Ctd_zxx_o0_t0_l0 = 1.0 * d_Ctd_zxx_o0_t0_m0_l0;
								Ctd_zxx = m_Ctd_zxx_o0_t0_l0 + m_Ctd_zxx_o0_t1_l0;
								m_Ctd_xxx_o0_t0_l0 = 0.5 * d_Ctd_xxx_o0_t0_m0_l0;
								Ctd_xxx = m_Ctd_xxx_o0_t0_l0;
								m_Ctd_yzz_o0_t1_l0 = (-0.5) * d_Ctd_yzz_o0_t1_m0_l0;
								m_Ctd_yzz_o0_t0_l0 = 1.0 * d_Ctd_yzz_o0_t0_m0_l0;
								Ctd_yzz = m_Ctd_yzz_o0_t0_l0 + m_Ctd_yzz_o0_t1_l0;
								m_Ctd_zyz_o0_t0_l0 = 0.5 * d_Ctd_yzz_o0_t1_m0_l0;
								Ctd_zyz = m_Ctd_zyz_o0_t0_l0;
								m_Ctd_xxy_o0_t0_l0 = 0.5 * d_Ctd_xxy_o0_t0_m0_l0;
								Ctd_xxy = m_Ctd_xxy_o0_t0_l0;
								m_Ctd_yxx_o0_t1_l0 = (-0.5) * d_Ctd_xxy_o0_t0_m0_l0;
								m_Ctd_yxx_o0_t0_l0 = 1.0 * d_Ctd_yxx_o0_t0_m0_l0;
								Ctd_yxx = m_Ctd_yxx_o0_t0_l0 + m_Ctd_yxx_o0_t1_l0;
								m_Ctd_xyy_o0_t1_l0 = (-0.5) * d_Ctd_xyy_o0_t1_m0_l0;
								m_Ctd_xyy_o0_t0_l0 = 1.0 * d_Ctd_xyy_o0_t0_m0_l0;
								Ctd_xyy = m_Ctd_xyy_o0_t0_l0 + m_Ctd_xyy_o0_t1_l0;
								m_Ctd_yxy_o0_t0_l0 = 0.5 * d_Ctd_xyy_o0_t1_m0_l0;
								Ctd_yxy = m_Ctd_yxy_o0_t0_l0;
								m_Ctd_xyz_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t2_m0_l0;
								m_Ctd_xyz_o0_t1_l0 = 0.5 * d_Ctd_xyz_o0_t1_m0_l0;
								m_Ctd_xyz_o0_t0_l0 = 0.5 * d_Ctd_xyz_o0_t0_m0_l0;
								Ctd_xyz = (m_Ctd_xyz_o0_t0_l0 + m_Ctd_xyz_o0_t1_l0) + m_Ctd_xyz_o0_t2_l0;
								m_Ctd_yxz_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t0_m0_l0;
								m_Ctd_yxz_o0_t0_l0 = 0.5 * d_Ctd_xyz_o0_t2_m0_l0;
								Ctd_yxz = (m_Ctd_yxz_o0_t0_l0 + m_Ctd_xyz_o0_t1_l0) + m_Ctd_yxz_o0_t2_l0;
								m_Ctd_zxy_o0_t2_l0 = (-0.5) * d_Ctd_xyz_o0_t1_m0_l0;
								Ctd_zxy = (m_Ctd_yxz_o0_t0_l0 + m_Ctd_xyz_o0_t0_l0) + m_Ctd_zxy_o0_t2_l0;
								g4d_tt = (-vector(Alpha, i, j, k) * vector(Alpha, i, j, k)) + inv_chi * (vector(gtd_xx, i, j, k) * (vector(Betau_x, i, j, k) * vector(Betau_x, i, j, k)) + 2.0 * vector(gtd_xy, i, j, k) * vector(Betau_y, i, j, k) * vector(Betau_x, i, j, k) + 2.0 * vector(gtd_xz, i, j, k) * vector(Betau_z, i, j, k) * vector(Betau_x, i, j, k) + vector(gtd_yy, i, j, k) * (vector(Betau_y, i, j, k) * vector(Betau_y, i, j, k)) + 2.0 * vector(gtd_yz, i, j, k) * vector(Betau_z, i, j, k) * vector(Betau_y, i, j, k) + vector(gtd_zz, i, j, k) * (vector(Betau_z, i, j, k) * vector(Betau_z, i, j, k)));
								g4d_ty = inv_chi * (vector(gtd_xy, i, j, k) * vector(Betau_x, i, j, k) + vector(gtd_yy, i, j, k) * vector(Betau_y, i, j, k) + vector(gtd_yz, i, j, k) * vector(Betau_z, i, j, k));
								g4d_tx = inv_chi * (vector(gtd_xx, i, j, k) * vector(Betau_x, i, j, k) + vector(gtd_xy, i, j, k) * vector(Betau_y, i, j, k) + vector(gtd_xz, i, j, k) * vector(Betau_z, i, j, k));
								g4d_tz = inv_chi * (vector(gtd_xz, i, j, k) * vector(Betau_x, i, j, k) + vector(gtd_yz, i, j, k) * vector(Betau_y, i, j, k) + vector(gtd_zz, i, j, k) * vector(Betau_z, i, j, k));
								m_Ctd_xzz_o0_t1_l0 = (-0.5) * d_Ctd_xzz_o0_t1_m0_l0;
								m_Ctd_xzz_o0_t0_l0 = 1.0 * d_Ctd_xzz_o0_t0_m0_l0;
								Ctd_xzz = m_Ctd_xzz_o0_t0_l0 + m_Ctd_xzz_o0_t1_l0;
								m_Ctd_zxz_o0_t0_l0 = 0.5 * d_Ctd_xzz_o0_t1_m0_l0;
								Ctd_zxz = m_Ctd_zxz_o0_t0_l0;
								dSigma_y = (1.0 * ur_y * MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))) / (pow(vector(chi, i, j, k), 1.500000000000000));
								dSigma_x = (1.0 * ur_x * MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))) / (pow(vector(chi, i, j, k), 1.500000000000000));
								dSigma_z = (1.0 * ur_z * MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k)))) / (pow(vector(chi, i, j, k), 1.500000000000000));
								detgtd = vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) * vector(gtd_xx, i, j, k) + (-vector(gtd_yy, i, j, k) * (vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k))) + (-vector(gtd_zz, i, j, k) * (vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k))) + (-(vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k)) * vector(gtd_xx, i, j, k)) + 2.0 * vector(gtd_yz, i, j, k) * vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k);
								idetgtd = 1.0 / detgtd;
								gtu_zz = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_yy, i, j, k) - vector(gtd_xy, i, j, k) * vector(gtd_xy, i, j, k));
								gtu_yy = idetgtd * (vector(gtd_xx, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_xz, i, j, k) * vector(gtd_xz, i, j, k));
								gtu_yz = idetgtd * ((-vector(gtd_xx, i, j, k) * vector(gtd_yz, i, j, k)) + vector(gtd_xy, i, j, k) * vector(gtd_xz, i, j, k));
								gtu_xx = idetgtd * (vector(gtd_yy, i, j, k) * vector(gtd_zz, i, j, k) - vector(gtd_yz, i, j, k) * vector(gtd_yz, i, j, k));
								gtu_xz = idetgtd * (vector(gtd_xy, i, j, k) * vector(gtd_yz, i, j, k) - vector(gtd_yy, i, j, k) * vector(gtd_xz, i, j, k));
								uth_z = vector(chi, i, j, k) * gtu_xz * uthd_x + vector(chi, i, j, k) * gtu_yz * uthd_y + vector(chi, i, j, k) * gtu_zz * uthd_z;
								Atud_zx = gtu_xz * vector(Atd_xx, i, j, k) + gtu_yz * vector(Atd_xy, i, j, k) + gtu_zz * vector(Atd_xz, i, j, k);
								Atud_zz = gtu_xz * vector(Atd_xz, i, j, k) + gtu_yz * vector(Atd_yz, i, j, k) + gtu_zz * vector(Atd_zz, i, j, k);
								Atud_zy = gtu_xz * vector(Atd_xy, i, j, k) + gtu_yz * vector(Atd_yy, i, j, k) + gtu_zz * vector(Atd_yz, i, j, k);
								Atu_zz = Atud_zx * gtu_xz + Atud_zy * gtu_yz + Atud_zz * gtu_zz;
								Sfu_z = vector(chi, i, j, k) * gtu_xz * vector(Sfd_x, i, j, k) + vector(chi, i, j, k) * gtu_yz * vector(Sfd_y, i, j, k) + vector(chi, i, j, k) * gtu_zz * vector(Sfd_z, i, j, k);
								Tf4u_tz = Sfu_z / (sdetg * vector(Alpha, i, j, k)) + (-vector(Betau_z, i, j, k) * Tf4u_tt);
								Ct_zxz = gtu_xz * Ctd_xxz + gtu_yz * Ctd_yxz + gtu_zz * Ctd_zxz;
								Ct_zxy = gtu_xz * Ctd_xxy + gtu_yz * Ctd_yxy + gtu_zz * Ctd_zxy;
								Ct_zxx = gtu_xz * Ctd_xxx + gtu_yz * Ctd_yxx + gtu_zz * Ctd_zxx;
								Ct_zyz = gtu_xz * Ctd_xyz + gtu_yz * Ctd_yyz + gtu_zz * Ctd_zyz;
								Ct_zyy = gtu_xz * Ctd_xyy + gtu_yz * Ctd_yyy + gtu_zz * Ctd_zyy;
								Ct_zzz = gtu_xz * Ctd_xzz + gtu_yz * Ctd_yzz + gtu_zz * Ctd_zzz;
								gtu_xy = idetgtd * ((-vector(gtd_xy, i, j, k) * vector(gtd_zz, i, j, k)) + vector(gtd_yz, i, j, k) * vector(gtd_xz, i, j, k));
								Ct_xzz = gtu_xx * Ctd_xzz + gtu_xy * Ctd_yzz + gtu_xz * Ctd_zzz;
								uth_x = vector(chi, i, j, k) * gtu_xx * uthd_x + vector(chi, i, j, k) * gtu_xy * uthd_y + vector(chi, i, j, k) * gtu_xz * uthd_z;
								uth_y = vector(chi, i, j, k) * gtu_xy * uthd_x + vector(chi, i, j, k) * gtu_yy * uthd_y + vector(chi, i, j, k) * gtu_yz * uthd_z;
								Gamt_z = gtu_xx * Ct_zxx + 2.0 * gtu_xy * Ct_zxy + 2.0 * gtu_xz * Ct_zxz + gtu_yy * Ct_zyy + 2.0 * gtu_yz * Ct_zyz + gtu_zz * Ct_zzz;
								Zu_z = 0.5 * vector(chi, i, j, k) * (vector(Gamh_z, i, j, k) - Gamt_z);
								Atud_yy = gtu_xy * vector(Atd_xy, i, j, k) + gtu_yy * vector(Atd_yy, i, j, k) + gtu_yz * vector(Atd_yz, i, j, k);
								Atud_yx = gtu_xy * vector(Atd_xx, i, j, k) + gtu_yy * vector(Atd_xy, i, j, k) + gtu_yz * vector(Atd_xz, i, j, k);
								Atud_yz = gtu_xy * vector(Atd_xz, i, j, k) + gtu_yy * vector(Atd_yz, i, j, k) + gtu_yz * vector(Atd_zz, i, j, k);
								Atu_yz = Atud_yx * gtu_xz + Atud_yy * gtu_yz + Atud_yz * gtu_zz;
								Atud_xx = gtu_xx * vector(Atd_xx, i, j, k) + gtu_xy * vector(Atd_xy, i, j, k) + gtu_xz * vector(Atd_xz, i, j, k);
								Atud_xz = gtu_xx * vector(Atd_xz, i, j, k) + gtu_xy * vector(Atd_yz, i, j, k) + gtu_xz * vector(Atd_zz, i, j, k);
								Atud_xy = gtu_xx * vector(Atd_xy, i, j, k) + gtu_xy * vector(Atd_yy, i, j, k) + gtu_xz * vector(Atd_yz, i, j, k);
								Atu_xz = Atud_xx * gtu_xz + Atud_xy * gtu_yz + Atud_xz * gtu_zz;
								Sfu_y = vector(chi, i, j, k) * gtu_xy * vector(Sfd_x, i, j, k) + vector(chi, i, j, k) * gtu_yy * vector(Sfd_y, i, j, k) + vector(chi, i, j, k) * gtu_yz * vector(Sfd_z, i, j, k);
								Tf4u_ty = Sfu_y / (sdetg * vector(Alpha, i, j, k)) + (-vector(Betau_y, i, j, k) * Tf4u_tt);
								Sfu_x = vector(chi, i, j, k) * gtu_xx * vector(Sfd_x, i, j, k) + vector(chi, i, j, k) * gtu_xy * vector(Sfd_y, i, j, k) + vector(chi, i, j, k) * gtu_xz * vector(Sfd_z, i, j, k);
								Tf4u_tx = Sfu_x / (sdetg * vector(Alpha, i, j, k)) + (-vector(Betau_x, i, j, k) * Tf4u_tt);
								vfu_y = vector(chi, i, j, k) * gtu_xy * vector(vfd_x, i, j, k) + vector(chi, i, j, k) * gtu_yy * vector(vfd_y, i, j, k) + vector(chi, i, j, k) * gtu_yz * vector(vfd_z, i, j, k);
								Ju_y = vector(Df, i, j, k) * (vector(Alpha, i, j, k) * vfu_y - vector(Betau_y, i, j, k));
								vfu_x = vector(chi, i, j, k) * gtu_xx * vector(vfd_x, i, j, k) + vector(chi, i, j, k) * gtu_xy * vector(vfd_y, i, j, k) + vector(chi, i, j, k) * gtu_xz * vector(vfd_z, i, j, k);
								Ju_x = vector(Df, i, j, k) * (vector(Alpha, i, j, k) * vfu_x - vector(Betau_x, i, j, k));
								Atu_yy = Atud_yx * gtu_xy + Atud_yy * gtu_yy + Atud_yz * gtu_yz;
								Ct_yxx = gtu_xy * Ctd_xxx + gtu_yy * Ctd_yxx + gtu_yz * Ctd_zxx;
								Ct_yxz = gtu_xy * Ctd_xxz + gtu_yy * Ctd_yxz + gtu_yz * Ctd_zxz;
								Ct_yxy = gtu_xy * Ctd_xxy + gtu_yy * Ctd_yxy + gtu_yz * Ctd_zxy;
								Atu_xy = Atud_xx * gtu_xy + Atud_xy * gtu_yy + Atud_xz * gtu_yz;
								Atu_xx = Atud_xx * gtu_xx + Atud_xy * gtu_xy + Atud_xz * gtu_xz;
								Ct_yyy = gtu_xy * Ctd_xyy + gtu_yy * Ctd_yyy + gtu_yz * Ctd_zyy;
								Ct_yyz = gtu_xy * Ctd_xyz + gtu_yy * Ctd_yyz + gtu_yz * Ctd_zyz;
								Tu_ty = Tf4u_ty;
								Tu_tx = Tf4u_tx;
								Tu_tz = Tf4u_tz;
								Ct_yzz = gtu_xy * Ctd_xzz + gtu_yy * Ctd_yzz + gtu_yz * Ctd_zzz;
								Gamt_y = gtu_xx * Ct_yxx + 2.0 * gtu_xy * Ct_yxy + 2.0 * gtu_xz * Ct_yxz + gtu_yy * Ct_yyy + 2.0 * gtu_yz * Ct_yyz + gtu_zz * Ct_yzz;
								Zu_y = 0.5 * vector(chi, i, j, k) * (vector(Gamh_y, i, j, k) - Gamt_y);
								Ct_xxy = gtu_xx * Ctd_xxy + gtu_xy * Ctd_yxy + gtu_xz * Ctd_zxy;
								vfu_z = vector(chi, i, j, k) * gtu_xz * vector(vfd_x, i, j, k) + vector(chi, i, j, k) * gtu_yz * vector(vfd_y, i, j, k) + vector(chi, i, j, k) * gtu_zz * vector(vfd_z, i, j, k);
								sqvf = vfu_x * vector(vfd_x, i, j, k) + vfu_y * vector(vfd_y, i, j, k) + vfu_z * vector(vfd_z, i, j, k);
								sqW = 1.0 / (1.0 - sqvf);
								Ju_z = vector(Df, i, j, k) * (vector(Alpha, i, j, k) * vfu_z - vector(Betau_z, i, j, k));
								Ct_xxx = gtu_xx * Ctd_xxx + gtu_xy * Ctd_yxx + gtu_xz * Ctd_zxx;
								Ct_xxz = gtu_xx * Ctd_xxz + gtu_xy * Ctd_yxz + gtu_xz * Ctd_zxz;
								d_Rtd_xx_o0_t9_m0_l0 = gtu_zz * (3.0 * Ct_xxz * Ctd_xxz + 2.0 * Ct_yxz * Ctd_xyz + Ct_yxz * Ctd_yxz + 2.0 * Ct_zxz * Ctd_xzz + Ct_zxz * Ctd_zxz) + gtu_yz * (3.0 * Ct_xxy * Ctd_xxz + 2.0 * Ct_yxy * Ctd_xyz + Ct_yxy * Ctd_yxz + 2.0 * Ct_zxy * Ctd_xzz + Ct_zxy * Ctd_zxz) + gtu_xz * (3.0 * Ct_xxx * Ctd_xxz + 2.0 * Ct_yxx * Ctd_xyz + Ct_yxx * Ctd_yxz + 2.0 * Ct_zxx * Ctd_xzz + Ct_zxx * Ctd_zxz) + gtu_yz * (3.0 * Ct_xxz * Ctd_xxy + 2.0 * Ct_yxz * Ctd_xyy + Ct_yxz * Ctd_yxy + 2.0 * Ct_zxz * Ctd_xyz + Ct_zxz * Ctd_zxy) + gtu_yy * (3.0 * Ct_xxy * Ctd_xxy + 2.0 * Ct_yxy * Ctd_xyy + Ct_yxy * Ctd_yxy + 2.0 * Ct_zxy * Ctd_xyz + Ct_zxy * Ctd_zxy) + gtu_xy * (3.0 * Ct_xxx * Ctd_xxy + 2.0 * Ct_yxx * Ctd_xyy + Ct_yxx * Ctd_yxy + 2.0 * Ct_zxx * Ctd_xyz + Ct_zxx * Ctd_zxy) + gtu_xz * (3.0 * Ct_xxz * Ctd_xxx + 2.0 * Ct_yxz * Ctd_xxy + Ct_yxz * Ctd_yxx + 2.0 * Ct_zxz * Ctd_xxz + Ct_zxz * Ctd_zxx) + gtu_xy * (3.0 * Ct_xxy * Ctd_xxx + 2.0 * Ct_yxy * Ctd_xxy + Ct_yxy * Ctd_yxx + 2.0 * Ct_zxy * Ctd_xxz + Ct_zxy * Ctd_zxx) + gtu_xx * (3.0 * Ct_xxx * Ctd_xxx + 2.0 * Ct_yxx * Ctd_xxy + Ct_yxx * Ctd_yxx + 2.0 * Ct_zxx * Ctd_xxz + Ct_zxx * Ctd_zxx) + 1.0 * vector(Gamh_z, i, j, k) * Ctd_xxz + 1.0 * vector(Gamh_y, i, j, k) * Ctd_xxy + 1.0 * vector(Gamh_x, i, j, k) * Ctd_xxx;
								m_Rtd_xx_o0_t8_l0 = 1.0 * vector(gtd_xz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
								m_Rtd_xx_o0_t7_l0 = 1.0 * vector(gtd_xy, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
								m_Rtd_xx_o0_t6_l0 = 1.0 * vector(gtd_xx, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
								m_Rtd_xx_o0_t5_l0 = (-0.5 * gtu_zz) * d_Rtd_xx_o0_t5_m0_l0;
								m_Rtd_xx_o0_t4_l0 = (-1.0 * gtu_yz) * d_Rtd_xx_o0_t4_m0_l0;
								m_Rtd_xx_o0_t3_l0 = (-0.5 * gtu_yy) * d_Rtd_xx_o0_t3_m0_l0;
								m_Rtd_xx_o0_t2_l0 = (-1.0 * gtu_xz) * d_Rtd_xx_o0_t2_m0_l0;
								m_Rtd_xx_o0_t1_l0 = (-1.0 * gtu_xy) * d_Rtd_xx_o0_t1_m0_l0;
								m_Rtd_xx_o0_t0_l0 = (-0.5 * gtu_xx) * d_Rtd_xx_o0_t0_m0_l0;
								Rtd_xx = ((((((((m_Rtd_xx_o0_t0_l0 + m_Rtd_xx_o0_t1_l0) + m_Rtd_xx_o0_t2_l0) + m_Rtd_xx_o0_t3_l0) + m_Rtd_xx_o0_t4_l0) + m_Rtd_xx_o0_t5_l0) + m_Rtd_xx_o0_t6_l0) + m_Rtd_xx_o0_t7_l0) + m_Rtd_xx_o0_t8_l0) + d_Rtd_xx_o0_t9_m0_l0;
								g4d_xy = inv_chi * vector(gtd_xy, i, j, k);
								wphph = inv_chi * vector(gtd_xx, i, j, k) * (uph_x * uph_x) + 2.0 * inv_chi * vector(gtd_xy, i, j, k) * uph_x * uph_y + 2.0 * inv_chi * vector(gtd_xz, i, j, k) * uph_x * uph_z + inv_chi * vector(gtd_yy, i, j, k) * (uph_y * uph_y) + 2.0 * inv_chi * vector(gtd_yz, i, j, k) * uph_y * uph_z + inv_chi * vector(gtd_zz, i, j, k) * (uph_z * uph_z);
								vph_y = uph_y / MAX(0.0001, sqrt(wphph));
								vph_z = uph_z / MAX(0.0001, sqrt(wphph));
								vph_x = uph_x / MAX(0.0001, sqrt(wphph));
								wthph = inv_chi * vector(gtd_xx, i, j, k) * uth_x * vph_x + inv_chi * vector(gtd_xy, i, j, k) * uth_x * vph_y + inv_chi * vector(gtd_xz, i, j, k) * uth_x * vph_z + inv_chi * vector(gtd_xy, i, j, k) * uth_y * vph_x + inv_chi * vector(gtd_yy, i, j, k) * uth_y * vph_y + inv_chi * vector(gtd_yz, i, j, k) * uth_y * vph_z + inv_chi * vector(gtd_xz, i, j, k) * uth_z * vph_x + inv_chi * vector(gtd_yz, i, j, k) * uth_z * vph_y + inv_chi * vector(gtd_zz, i, j, k) * uth_z * vph_z;
								g4d_xx = inv_chi * vector(gtd_xx, i, j, k);
								wrph = inv_chi * vector(gtd_xx, i, j, k) * ur_x * vph_x + inv_chi * vector(gtd_xy, i, j, k) * ur_x * vph_y + inv_chi * vector(gtd_xz, i, j, k) * ur_x * vph_z + inv_chi * vector(gtd_xy, i, j, k) * ur_y * vph_x + inv_chi * vector(gtd_yy, i, j, k) * ur_y * vph_y + inv_chi * vector(gtd_yz, i, j, k) * ur_y * vph_z + inv_chi * vector(gtd_xz, i, j, k) * ur_z * vph_x + inv_chi * vector(gtd_yz, i, j, k) * ur_z * vph_y + inv_chi * vector(gtd_zz, i, j, k) * ur_z * vph_z;
								vr_x = ((-wrph * vph_x) + ur_x) / MAX(0.0001, sqrt(wrr));
								vr_y = ((-wrph * vph_y) + ur_y) / MAX(0.0001, sqrt(wrr));
								vr_z = ((-wrph * vph_z) + ur_z) / MAX(0.0001, sqrt(wrr));
								wthr = inv_chi * vector(gtd_xx, i, j, k) * uth_x * vr_x + inv_chi * vector(gtd_xy, i, j, k) * uth_x * vr_y + inv_chi * vector(gtd_xz, i, j, k) * uth_x * vr_z + inv_chi * vector(gtd_xy, i, j, k) * uth_y * vr_x + inv_chi * vector(gtd_yy, i, j, k) * uth_y * vr_y + inv_chi * vector(gtd_yz, i, j, k) * uth_y * vr_z + inv_chi * vector(gtd_xz, i, j, k) * uth_z * vr_x + inv_chi * vector(gtd_yz, i, j, k) * uth_z * vr_y + inv_chi * vector(gtd_zz, i, j, k) * uth_z * vr_z;
								g4d_xz = inv_chi * vector(gtd_xz, i, j, k);
								wthth = inv_chi * vector(gtd_xx, i, j, k) * (uth_x * uth_x) + 2.0 * inv_chi * vector(gtd_xy, i, j, k) * uth_x * uth_y + 2.0 * inv_chi * vector(gtd_xz, i, j, k) * uth_x * uth_z + inv_chi * vector(gtd_yy, i, j, k) * (uth_y * uth_y) + 2.0 * inv_chi * vector(gtd_yz, i, j, k) * uth_y * uth_z + inv_chi * vector(gtd_zz, i, j, k) * (uth_z * uth_z);
								vth_x = ((-wthph * vph_x) + (-wthr * vr_x) + uth_x) / MAX(0.0001, sqrt(wthth));
								mmR_xx = (-0.5 * (vph_x * vph_x)) + 0.5 * (vth_x * vth_x);
								mmI_xx = (-1.0) * vph_x * vth_x;
								vth_y = ((-wthph * vph_y) + (-wthr * vr_y) + uth_y) / MAX(0.0001, sqrt(wthth));
								mmR_yy = (-0.5 * (vph_y * vph_y)) + 0.5 * (vth_y * vth_y);
								mmR_xy = (-0.5 * vph_x * vph_y) + 0.5 * vth_x * vth_y;
								mmI_xy = (-0.5 * vph_x * vth_y) + (-0.5 * vph_y * vth_x);
								mmI_yy = (-1.0) * vph_y * vth_y;
								vth_z = ((-wthph * vph_z) + (-wthr * vr_z) + uth_z) / MAX(0.0001, sqrt(wthth));
								mmR_yz = (-0.5 * vph_y * vph_z) + 0.5 * vth_y * vth_z;
								mmR_zz = (-0.5 * (vph_z * vph_z)) + 0.5 * (vth_z * vth_z);
								mmR_xz = (-0.5 * vph_x * vph_z) + 0.5 * vth_x * vth_z;
								mmI_zz = (-1.0) * vph_z * vth_z;
								mmI_xz = (-0.5 * vph_x * vth_z) + (-0.5 * vph_z * vth_x);
								mmI_yz = (-0.5 * vph_y * vth_z) + (-0.5 * vph_z * vth_y);
								Ct_xyz = gtu_xx * Ctd_xyz + gtu_xy * Ctd_yyz + gtu_xz * Ctd_zyz;
								d_Rtd_xz_o0_t12_m0_l0 = gtu_zz * (Ct_xxz * Ctd_zxz + Ct_xxz * Ctd_xzz + Ct_xzz * Ctd_xxz + Ct_yxz * Ctd_zyz + Ct_yxz * Ctd_yzz + Ct_yzz * Ctd_xyz + 2.0 * Ct_zxz * Ctd_zzz + Ct_zzz * Ctd_xzz) + gtu_yz * (Ct_xxy * Ctd_zxz + Ct_xxy * Ctd_xzz + Ct_xyz * Ctd_xxz + Ct_yxy * Ctd_zyz + Ct_yxy * Ctd_yzz + Ct_yyz * Ctd_xyz + 2.0 * Ct_zxy * Ctd_zzz + Ct_zyz * Ctd_xzz) + gtu_xz * (Ct_xxx * Ctd_zxz + Ct_xxx * Ctd_xzz + Ct_xxz * Ctd_xxz + Ct_yxx * Ctd_zyz + Ct_yxx * Ctd_yzz + Ct_yxz * Ctd_xyz + 2.0 * Ct_zxx * Ctd_zzz + Ct_zxz * Ctd_xzz) + gtu_yz * (Ct_xxz * Ctd_zxy + Ct_xxz * Ctd_xyz + Ct_xzz * Ctd_xxy + Ct_yxz * Ctd_zyy + Ct_yxz * Ctd_yyz + Ct_yzz * Ctd_xyy + 2.0 * Ct_zxz * Ctd_zyz + Ct_zzz * Ctd_xyz) + gtu_yy * (Ct_xxy * Ctd_zxy + Ct_xxy * Ctd_xyz + Ct_xyz * Ctd_xxy + Ct_yxy * Ctd_zyy + Ct_yxy * Ctd_yyz + Ct_yyz * Ctd_xyy + 2.0 * Ct_zxy * Ctd_zyz + Ct_zyz * Ctd_xyz) + gtu_xy * (Ct_xxx * Ctd_zxy + Ct_xxx * Ctd_xyz + Ct_xxz * Ctd_xxy + Ct_yxx * Ctd_zyy + Ct_yxx * Ctd_yyz + Ct_yxz * Ctd_xyy + 2.0 * Ct_zxx * Ctd_zyz + Ct_zxz * Ctd_xyz) + gtu_xz * (Ct_xxz * Ctd_zxx + Ct_xxz * Ctd_xxz + Ct_xzz * Ctd_xxx + Ct_yxz * Ctd_zxy + Ct_yxz * Ctd_yxz + Ct_yzz * Ctd_xxy + 2.0 * Ct_zxz * Ctd_zxz + Ct_zzz * Ctd_xxz) + gtu_xy * (Ct_xxy * Ctd_zxx + Ct_xxy * Ctd_xxz + Ct_xyz * Ctd_xxx + Ct_yxy * Ctd_zxy + Ct_yxy * Ctd_yxz + Ct_yyz * Ctd_xxy + 2.0 * Ct_zxy * Ctd_zxz + Ct_zyz * Ctd_xxz) + gtu_xx * (Ct_xxx * Ctd_zxx + Ct_xxx * Ctd_xxz + Ct_xxz * Ctd_xxx + Ct_yxx * Ctd_zxy + Ct_yxx * Ctd_yxz + Ct_yxz * Ctd_xxy + 2.0 * Ct_zxx * Ctd_zxz + Ct_zxz * Ctd_xxz) + 0.5 * vector(Gamh_z, i, j, k) * (Ctd_xzz + Ctd_zxz) + 0.5 * vector(Gamh_y, i, j, k) * (Ctd_xyz + Ctd_zxy) + 0.5 * vector(Gamh_x, i, j, k) * (Ctd_xxz + Ctd_zxx);
								m_Rtd_xz_o0_t11_l0 = 0.5 * vector(gtd_zz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
								m_Rtd_xz_o0_t10_l0 = 0.5 * vector(gtd_yz, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
								m_Rtd_xz_o0_t9_l0 = 0.5 * vector(gtd_xz, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
								m_Rtd_xz_o0_t8_l0 = 0.5 * vector(gtd_xz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
								m_Rtd_xz_o0_t7_l0 = 0.5 * vector(gtd_xy, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
								m_Rtd_xz_o0_t6_l0 = 0.5 * vector(gtd_xx, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
								m_Rtd_xz_o0_t5_l0 = (-0.5 * gtu_zz) * d_Rtd_xz_o0_t5_m0_l0;
								m_Rtd_xz_o0_t4_l0 = (-1.0 * gtu_yz) * d_Rtd_xz_o0_t4_m0_l0;
								m_Rtd_xz_o0_t3_l0 = (-0.5 * gtu_yy) * d_Rtd_xz_o0_t3_m0_l0;
								m_Rtd_xz_o0_t2_l0 = (-1.0 * gtu_xz) * d_Rtd_xz_o0_t2_m0_l0;
								m_Rtd_xz_o0_t1_l0 = (-1.0 * gtu_xy) * d_Rtd_xz_o0_t1_m0_l0;
								m_Rtd_xz_o0_t0_l0 = (-0.5 * gtu_xx) * d_Rtd_xz_o0_t0_m0_l0;
								Rtd_xz = (((((((((((m_Rtd_xz_o0_t0_l0 + m_Rtd_xz_o0_t1_l0) + m_Rtd_xz_o0_t2_l0) + m_Rtd_xz_o0_t3_l0) + m_Rtd_xz_o0_t4_l0) + m_Rtd_xz_o0_t5_l0) + m_Rtd_xz_o0_t6_l0) + m_Rtd_xz_o0_t7_l0) + m_Rtd_xz_o0_t8_l0) + m_Rtd_xz_o0_t9_l0) + m_Rtd_xz_o0_t10_l0) + m_Rtd_xz_o0_t11_l0) + d_Rtd_xz_o0_t12_m0_l0;
								d_Rtd_zz_o0_t9_m0_l0 = gtu_zz * (2.0 * Ct_xzz * Ctd_zxz + Ct_xzz * Ctd_xzz + 2.0 * Ct_yzz * Ctd_zyz + Ct_yzz * Ctd_yzz + 3.0 * Ct_zzz * Ctd_zzz) + gtu_yz * (2.0 * Ct_xyz * Ctd_zxz + Ct_xyz * Ctd_xzz + 2.0 * Ct_yyz * Ctd_zyz + Ct_yyz * Ctd_yzz + 3.0 * Ct_zyz * Ctd_zzz) + gtu_xz * (2.0 * Ct_xxz * Ctd_zxz + Ct_xxz * Ctd_xzz + 2.0 * Ct_yxz * Ctd_zyz + Ct_yxz * Ctd_yzz + 3.0 * Ct_zxz * Ctd_zzz) + gtu_yz * (2.0 * Ct_xzz * Ctd_zxy + Ct_xzz * Ctd_xyz + 2.0 * Ct_yzz * Ctd_zyy + Ct_yzz * Ctd_yyz + 3.0 * Ct_zzz * Ctd_zyz) + gtu_yy * (2.0 * Ct_xyz * Ctd_zxy + Ct_xyz * Ctd_xyz + 2.0 * Ct_yyz * Ctd_zyy + Ct_yyz * Ctd_yyz + 3.0 * Ct_zyz * Ctd_zyz) + gtu_xy * (2.0 * Ct_xxz * Ctd_zxy + Ct_xxz * Ctd_xyz + 2.0 * Ct_yxz * Ctd_zyy + Ct_yxz * Ctd_yyz + 3.0 * Ct_zxz * Ctd_zyz) + gtu_xz * (2.0 * Ct_xzz * Ctd_zxx + Ct_xzz * Ctd_xxz + 2.0 * Ct_yzz * Ctd_zxy + Ct_yzz * Ctd_yxz + 3.0 * Ct_zzz * Ctd_zxz) + gtu_xy * (2.0 * Ct_xyz * Ctd_zxx + Ct_xyz * Ctd_xxz + 2.0 * Ct_yyz * Ctd_zxy + Ct_yyz * Ctd_yxz + 3.0 * Ct_zyz * Ctd_zxz) + gtu_xx * (2.0 * Ct_xxz * Ctd_zxx + Ct_xxz * Ctd_xxz + 2.0 * Ct_yxz * Ctd_zxy + Ct_yxz * Ctd_yxz + 3.0 * Ct_zxz * Ctd_zxz) + 1.0 * vector(Gamh_z, i, j, k) * Ctd_zzz + 1.0 * vector(Gamh_y, i, j, k) * Ctd_zyz + 1.0 * vector(Gamh_x, i, j, k) * Ctd_zxz;
								m_Rtd_zz_o0_t8_l0 = 1.0 * vector(gtd_zz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
								m_Rtd_zz_o0_t7_l0 = 1.0 * vector(gtd_yz, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
								m_Rtd_zz_o0_t6_l0 = 1.0 * vector(gtd_xz, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
								m_Rtd_zz_o0_t5_l0 = (-0.5 * gtu_zz) * d_Rtd_zz_o0_t5_m0_l0;
								m_Rtd_zz_o0_t4_l0 = (-1.0 * gtu_yz) * d_Rtd_zz_o0_t4_m0_l0;
								m_Rtd_zz_o0_t3_l0 = (-0.5 * gtu_yy) * d_Rtd_zz_o0_t3_m0_l0;
								m_Rtd_zz_o0_t2_l0 = (-1.0 * gtu_xz) * d_Rtd_zz_o0_t2_m0_l0;
								m_Rtd_zz_o0_t1_l0 = (-1.0 * gtu_xy) * d_Rtd_zz_o0_t1_m0_l0;
								m_Rtd_zz_o0_t0_l0 = (-0.5 * gtu_xx) * d_Rtd_zz_o0_t0_m0_l0;
								Rtd_zz = ((((((((m_Rtd_zz_o0_t0_l0 + m_Rtd_zz_o0_t1_l0) + m_Rtd_zz_o0_t2_l0) + m_Rtd_zz_o0_t3_l0) + m_Rtd_zz_o0_t4_l0) + m_Rtd_zz_o0_t5_l0) + m_Rtd_zz_o0_t6_l0) + m_Rtd_zz_o0_t7_l0) + m_Rtd_zz_o0_t8_l0) + d_Rtd_zz_o0_t9_m0_l0;
								Ct_xyy = gtu_xx * Ctd_xyy + gtu_xy * Ctd_yyy + gtu_xz * Ctd_zyy;
								Gamt_x = gtu_xx * Ct_xxx + 2.0 * gtu_xy * Ct_xxy + 2.0 * gtu_xz * Ct_xxz + gtu_yy * Ct_xyy + 2.0 * gtu_yz * Ct_xyz + gtu_zz * Ct_xzz;
								Zu_x = 0.5 * vector(chi, i, j, k) * (vector(Gamh_x, i, j, k) - Gamt_x);
								m_Rpd_xx_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t3_m0_l0;
								m_Rpd_xx_o0_t21_l0 = (-0.75 * vector(gtd_xx, i, j, k) * gtu_zz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xx_o0_t20_l0 = (-0.75 * vector(gtd_xx, i, j, k) * gtu_yy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xx_o0_t19_l0 = (-0.75 * vector(gtd_xx, i, j, k) * gtu_xx * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xx_o0_t18_l0 = 2.0 * inv_chi * Zu_z * vector(gtd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xx_o0_t17_l0 = 2.0 * inv_chi * Zu_y * vector(gtd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xx_o0_t16_l0 = 2.0 * inv_chi * Zu_x * vector(gtd_xx, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xx_o0_t15_l0 = (-1.500000000000000 * vector(gtd_xx, i, j, k) * gtu_yz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xx_o0_t14_l0 = (-1.500000000000000 * vector(gtd_xx, i, j, k) * gtu_xz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xx_o0_t13_l0 = (-1.500000000000000 * vector(gtd_xx, i, j, k) * gtu_xy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xx_o0_t12_l0 = (-0.5 * Ct_xxx) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xx_o0_t11_l0 = (-0.25 * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xx_o0_t10_l0 = (-0.5 * Ct_zxx) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xx_o0_t9_l0 = (-0.5 * Ct_yxx) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xx_o0_t8_l0 = 0.5 * vector(gtd_xx, i, j, k) * gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
								m_Rpd_xx_o0_t7_l0 = 1.0 * vector(gtd_xx, i, j, k) * gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
								m_Rpd_xx_o0_t6_l0 = 0.5 * vector(gtd_xx, i, j, k) * gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
								m_Rpd_xx_o0_t5_l0 = 1.0 * vector(gtd_xx, i, j, k) * gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
								m_Rpd_xx_o0_t4_l0 = 1.0 * vector(gtd_xx, i, j, k) * gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
								m_Rpd_xx_o0_t3_l0 = 0.5 * vector(gtd_xx, i, j, k) * gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
								m_Rpd_xx_o0_t2_l0 = (-0.5 * vector(gtd_xx, i, j, k) * vector(Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xx_o0_t1_l0 = (-0.5 * vector(gtd_xx, i, j, k) * vector(Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xx_o0_t0_l0 = (-0.5 * vector(gtd_xx, i, j, k) * vector(Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
								Rpd_xx = (((((((((((((((((((((m_Rpd_xx_o0_t0_l0 + m_Rpd_xx_o0_t1_l0) + m_Rpd_xx_o0_t2_l0) + m_Rpd_xx_o0_t3_l0) + m_Rpd_xx_o0_t4_l0) + m_Rpd_xx_o0_t5_l0) + m_Rpd_xx_o0_t6_l0) + m_Rpd_xx_o0_t7_l0) + m_Rpd_xx_o0_t8_l0) + m_Rpd_xx_o0_t9_l0) + m_Rpd_xx_o0_t10_l0) + m_Rpd_xx_o0_t11_l0) + m_Rpd_xx_o0_t12_l0) + m_Rpd_xx_o0_t13_l0) + m_Rpd_xx_o0_t14_l0) + m_Rpd_xx_o0_t15_l0) + m_Rpd_xx_o0_t16_l0) + m_Rpd_xx_o0_t17_l0) + m_Rpd_xx_o0_t18_l0) + m_Rpd_xx_o0_t19_l0) + m_Rpd_xx_o0_t20_l0) + m_Rpd_xx_o0_t21_l0) + m_Rpd_xx_o0_t22_l0;
								d_Del_Kd_yxx_o0_t1_m0_l0 = -inv_chi * (2.0 * vector(Atd_xx, i, j, k) * Ct_xxy + 2.0 * vector(Atd_xy, i, j, k) * Ct_yxy + 2.0 * vector(Atd_xz, i, j, k) * Ct_zxy);
								m_Del_Kd_yxx_o0_t6_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_yxx_o0_t5_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_yxx_o0_t4_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_yxx_o0_t3_l0 = 1.0 * (inv_chi * inv_chi) * vector(Atd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_yxx_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xx, i, j, k) * d_Del_Kd_yxx_o0_t2_m0_l0;
								m_Del_Kd_yxx_o0_t0_l0 = inv_chi * d_Del_Kd_yxx_o0_t0_m0_l0;
								Del_Kd_yxx = (((((m_Del_Kd_yxx_o0_t0_l0 + d_Del_Kd_yxx_o0_t1_m0_l0) + m_Del_Kd_yxx_o0_t2_l0) + m_Del_Kd_yxx_o0_t3_l0) + m_Del_Kd_yxx_o0_t4_l0) + m_Del_Kd_yxx_o0_t5_l0) + m_Del_Kd_yxx_o0_t6_l0;
								d_Del_Kd_yxz_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xz, i, j, k) * Ct_xxy + vector(Atd_xx, i, j, k) * Ct_xyz + vector(Atd_yz, i, j, k) * Ct_yxy + vector(Atd_xy, i, j, k) * Ct_yyz + vector(Atd_zz, i, j, k) * Ct_zxy + vector(Atd_xz, i, j, k) * Ct_zyz);
								m_Del_Kd_yxz_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_yxz_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_yxz_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_yxz_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_yxz_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_yxz_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_yxz_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_yxz_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_yz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_yxz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xz, i, j, k) * d_Del_Kd_yxx_o0_t2_m0_l0;
								m_Del_Kd_yxz_o0_t0_l0 = inv_chi * d_Del_Kd_yxz_o0_t0_m0_l0;
								Del_Kd_yxz = (((((((((m_Del_Kd_yxz_o0_t0_l0 + d_Del_Kd_yxz_o0_t1_m0_l0) + m_Del_Kd_yxz_o0_t2_l0) + m_Del_Kd_yxz_o0_t3_l0) + m_Del_Kd_yxz_o0_t4_l0) + m_Del_Kd_yxz_o0_t5_l0) + m_Del_Kd_yxz_o0_t6_l0) + m_Del_Kd_yxz_o0_t7_l0) + m_Del_Kd_yxz_o0_t8_l0) + m_Del_Kd_yxz_o0_t9_l0) + m_Del_Kd_yxz_o0_t10_l0;
								EWeyl_xx = (-inv_chi * Rpd_xx) + (-Rtd_xx) + inv_chi * (vector(Atd_xx, i, j, k) * Atud_xx + vector(Atd_xy, i, j, k) * Atud_yx + vector(Atd_xz, i, j, k) * Atud_zx) + (-0.3333333333333333 * inv_chi * vector(trK, i, j, k) * vector(Atd_xx, i, j, k)) + (-0.2222222222222222 * inv_chi * vector(gtd_xx, i, j, k) * (vector(trK, i, j, k) * vector(trK, i, j, k)));
								d_Del_Kd_zxx_o0_t1_m0_l0 = -inv_chi * (2.0 * vector(Atd_xx, i, j, k) * Ct_xxz + 2.0 * vector(Atd_xy, i, j, k) * Ct_yxz + 2.0 * vector(Atd_xz, i, j, k) * Ct_zxz);
								m_Del_Kd_zxx_o0_t6_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_zxx_o0_t5_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_zxx_o0_t4_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_zxx_o0_t3_l0 = 1.0 * (inv_chi * inv_chi) * vector(Atd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_zxx_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xx, i, j, k) * d_Del_Kd_zxx_o0_t2_m0_l0;
								m_Del_Kd_zxx_o0_t0_l0 = inv_chi * d_Del_Kd_zxx_o0_t0_m0_l0;
								Del_Kd_zxx = (((((m_Del_Kd_zxx_o0_t0_l0 + d_Del_Kd_zxx_o0_t1_m0_l0) + m_Del_Kd_zxx_o0_t2_l0) + m_Del_Kd_zxx_o0_t3_l0) + m_Del_Kd_zxx_o0_t4_l0) + m_Del_Kd_zxx_o0_t5_l0) + m_Del_Kd_zxx_o0_t6_l0;
								d_Del_Kd_zyz_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xz, i, j, k) * Ct_xyz + vector(Atd_xy, i, j, k) * Ct_xzz + vector(Atd_yz, i, j, k) * Ct_yyz + vector(Atd_yy, i, j, k) * Ct_yzz + vector(Atd_zz, i, j, k) * Ct_zyz + vector(Atd_yz, i, j, k) * Ct_zzz);
								m_Del_Kd_zyz_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_zz, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_zyz_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_zz, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_zyz_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_zz, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_zyz_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_zyz_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_zyz_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_zyz_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_zyz_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_zz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_zyz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_yz, i, j, k) * d_Del_Kd_zxx_o0_t2_m0_l0;
								m_Del_Kd_zyz_o0_t0_l0 = inv_chi * d_Del_Kd_zyz_o0_t0_m0_l0;
								Del_Kd_zyz = (((((((((m_Del_Kd_zyz_o0_t0_l0 + d_Del_Kd_zyz_o0_t1_m0_l0) + m_Del_Kd_zyz_o0_t2_l0) + m_Del_Kd_zyz_o0_t3_l0) + m_Del_Kd_zyz_o0_t4_l0) + m_Del_Kd_zyz_o0_t5_l0) + m_Del_Kd_zyz_o0_t6_l0) + m_Del_Kd_zyz_o0_t7_l0) + m_Del_Kd_zyz_o0_t8_l0) + m_Del_Kd_zyz_o0_t9_l0) + m_Del_Kd_zyz_o0_t10_l0;
								d_Del_Kd_zyy_o0_t1_m0_l0 = -inv_chi * (2.0 * vector(Atd_xy, i, j, k) * Ct_xyz + 2.0 * vector(Atd_yy, i, j, k) * Ct_yyz + 2.0 * vector(Atd_yz, i, j, k) * Ct_zyz);
								m_Del_Kd_zyy_o0_t6_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_zyy_o0_t5_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_zyy_o0_t4_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_zyy_o0_t3_l0 = 1.0 * (inv_chi * inv_chi) * vector(Atd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_zyy_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_yy, i, j, k) * d_Del_Kd_zxx_o0_t2_m0_l0;
								m_Del_Kd_zyy_o0_t0_l0 = inv_chi * d_Del_Kd_zyy_o0_t0_m0_l0;
								Del_Kd_zyy = (((((m_Del_Kd_zyy_o0_t0_l0 + d_Del_Kd_zyy_o0_t1_m0_l0) + m_Del_Kd_zyy_o0_t2_l0) + m_Del_Kd_zyy_o0_t3_l0) + m_Del_Kd_zyy_o0_t4_l0) + m_Del_Kd_zyy_o0_t5_l0) + m_Del_Kd_zyy_o0_t6_l0;
								d_Del_Kd_zxy_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xy, i, j, k) * Ct_xxz + vector(Atd_xx, i, j, k) * Ct_xyz + vector(Atd_yy, i, j, k) * Ct_yxz + vector(Atd_xy, i, j, k) * Ct_yyz + vector(Atd_yz, i, j, k) * Ct_zxz + vector(Atd_xz, i, j, k) * Ct_zyz);
								m_Del_Kd_zxy_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_zxy_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_zxy_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_zxy_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_zxy_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xy, i, j, k) * d_Del_Kd_zxx_o0_t2_m0_l0;
								m_Del_Kd_zxy_o0_t0_l0 = inv_chi * d_Del_Kd_zxy_o0_t0_m0_l0;
								Del_Kd_zxy = (((((((((m_Del_Kd_zxy_o0_t0_l0 + d_Del_Kd_zxy_o0_t1_m0_l0) + m_Del_Kd_zxy_o0_t2_l0) + m_Del_Kd_yxz_o0_t3_l0) + m_Del_Kd_zxy_o0_t4_l0) + m_Del_Kd_zxy_o0_t5_l0) + m_Del_Kd_zxy_o0_t6_l0) + m_Del_Kd_zxy_o0_t7_l0) + m_Del_Kd_yxz_o0_t8_l0) + m_Del_Kd_yxz_o0_t9_l0) + m_Del_Kd_yxz_o0_t10_l0;
								d_Del_Kd_zxz_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xz, i, j, k) * Ct_xxz + vector(Atd_xx, i, j, k) * Ct_xzz + vector(Atd_yz, i, j, k) * Ct_yxz + vector(Atd_xy, i, j, k) * Ct_yzz + vector(Atd_zz, i, j, k) * Ct_zxz + vector(Atd_xz, i, j, k) * Ct_zzz);
								m_Del_Kd_zxz_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_zz, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_zxz_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_zz, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_zxz_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_zz, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_zxz_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_zxz_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_zxz_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_zxz_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_zxz_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_zz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_zxz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xz, i, j, k) * d_Del_Kd_zxx_o0_t2_m0_l0;
								m_Del_Kd_zxz_o0_t0_l0 = inv_chi * d_Del_Kd_zxz_o0_t0_m0_l0;
								Del_Kd_zxz = (((((((((m_Del_Kd_zxz_o0_t0_l0 + d_Del_Kd_zxz_o0_t1_m0_l0) + m_Del_Kd_zxz_o0_t2_l0) + m_Del_Kd_zxz_o0_t3_l0) + m_Del_Kd_zxz_o0_t4_l0) + m_Del_Kd_zxz_o0_t5_l0) + m_Del_Kd_zxz_o0_t6_l0) + m_Del_Kd_zxz_o0_t7_l0) + m_Del_Kd_zxz_o0_t8_l0) + m_Del_Kd_zxz_o0_t9_l0) + m_Del_Kd_zxz_o0_t10_l0;
								m_Rpd_yz_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t7_m0_l0;
								m_Rpd_yz_o0_t24_l0 = (-0.75 * vector(gtd_yz, i, j, k) * gtu_zz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_yz_o0_t23_l0 = (-0.75 * vector(gtd_yz, i, j, k) * gtu_yy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yz_o0_t22_l0 = (-0.75 * vector(gtd_yz, i, j, k) * gtu_xx * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_yz_o0_t21_l0 = inv_chi * Zu_z * vector(gtd_zz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yz_o0_t20_l0 = inv_chi * Zu_y * vector(gtd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yz_o0_t19_l0 = inv_chi * Zu_x * vector(gtd_xz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yz_o0_t18_l0 = inv_chi * Zu_z * vector(gtd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_yz_o0_t17_l0 = inv_chi * Zu_y * vector(gtd_yy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_yz_o0_t16_l0 = inv_chi * Zu_x * vector(gtd_xy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_yz_o0_t15_l0 = (-1.500000000000000 * vector(gtd_yz, i, j, k) * gtu_yz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yz_o0_t14_l0 = (-1.500000000000000 * vector(gtd_yz, i, j, k) * gtu_xz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_yz_o0_t13_l0 = (-1.500000000000000 * vector(gtd_yz, i, j, k) * gtu_xy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_yz_o0_t12_l0 = (-0.5 * Ct_zyz) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_yz_o0_t11_l0 = (-0.5 * Ct_yyz) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yz_o0_t10_l0 = (-0.5 * Ct_xyz) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_yz_o0_t9_l0 = (-0.25 * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_yz_o0_t8_l0 = 0.5 * vector(gtd_yz, i, j, k) * gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
								m_Rpd_yz_o0_t7_l0 = 1.0 * vector(gtd_yz, i, j, k) * gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
								m_Rpd_yz_o0_t6_l0 = 0.5 * vector(gtd_yz, i, j, k) * gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
								m_Rpd_yz_o0_t5_l0 = 1.0 * vector(gtd_yz, i, j, k) * gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
								m_Rpd_yz_o0_t4_l0 = 1.0 * vector(gtd_yz, i, j, k) * gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
								m_Rpd_yz_o0_t3_l0 = 0.5 * vector(gtd_yz, i, j, k) * gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
								m_Rpd_yz_o0_t2_l0 = (-0.5 * vector(gtd_yz, i, j, k) * vector(Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_yz_o0_t1_l0 = (-0.5 * vector(gtd_yz, i, j, k) * vector(Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yz_o0_t0_l0 = (-0.5 * vector(gtd_yz, i, j, k) * vector(Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
								Rpd_yz = ((((((((((((((((((((((((m_Rpd_yz_o0_t0_l0 + m_Rpd_yz_o0_t1_l0) + m_Rpd_yz_o0_t2_l0) + m_Rpd_yz_o0_t3_l0) + m_Rpd_yz_o0_t4_l0) + m_Rpd_yz_o0_t5_l0) + m_Rpd_yz_o0_t6_l0) + m_Rpd_yz_o0_t7_l0) + m_Rpd_yz_o0_t8_l0) + m_Rpd_yz_o0_t9_l0) + m_Rpd_yz_o0_t10_l0) + m_Rpd_yz_o0_t11_l0) + m_Rpd_yz_o0_t12_l0) + m_Rpd_yz_o0_t13_l0) + m_Rpd_yz_o0_t14_l0) + m_Rpd_yz_o0_t15_l0) + m_Rpd_yz_o0_t16_l0) + m_Rpd_yz_o0_t17_l0) + m_Rpd_yz_o0_t18_l0) + m_Rpd_yz_o0_t19_l0) + m_Rpd_yz_o0_t20_l0) + m_Rpd_yz_o0_t21_l0) + m_Rpd_yz_o0_t22_l0) + m_Rpd_yz_o0_t23_l0) + m_Rpd_yz_o0_t24_l0) + m_Rpd_yz_o0_t25_l0;
								d_Del_Kd_yzz_o0_t1_m0_l0 = -inv_chi * (2.0 * vector(Atd_xz, i, j, k) * Ct_xyz + 2.0 * vector(Atd_yz, i, j, k) * Ct_yyz + 2.0 * vector(Atd_zz, i, j, k) * Ct_zyz);
								m_Del_Kd_yzz_o0_t6_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_yzz_o0_t5_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_yzz_o0_t4_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_yzz_o0_t3_l0 = 1.0 * (inv_chi * inv_chi) * vector(Atd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_yzz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_zz, i, j, k) * d_Del_Kd_yxx_o0_t2_m0_l0;
								m_Del_Kd_yzz_o0_t0_l0 = inv_chi * d_Del_Kd_yzz_o0_t0_m0_l0;
								Del_Kd_yzz = (((((m_Del_Kd_yzz_o0_t0_l0 + d_Del_Kd_yzz_o0_t1_m0_l0) + m_Del_Kd_yzz_o0_t2_l0) + m_Del_Kd_yzz_o0_t3_l0) + m_Del_Kd_yzz_o0_t4_l0) + m_Del_Kd_yzz_o0_t5_l0) + m_Del_Kd_yzz_o0_t6_l0;
								d_Del_Kd_xxy_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xy, i, j, k) * Ct_xxx + vector(Atd_xx, i, j, k) * Ct_xxy + vector(Atd_yy, i, j, k) * Ct_yxx + vector(Atd_xy, i, j, k) * Ct_yxy + vector(Atd_yz, i, j, k) * Ct_zxx + vector(Atd_xz, i, j, k) * Ct_zxy);
								m_Del_Kd_xxy_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_xxy_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_xxy_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_xxy_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xx, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_xxy_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xx, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_xxy_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xx, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_xxy_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xx, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_xxy_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_xxy_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xy, i, j, k) * d_Del_Kd_xxy_o0_t2_m0_l0;
								m_Del_Kd_xxy_o0_t0_l0 = inv_chi * d_Del_Kd_xxy_o0_t0_m0_l0;
								Del_Kd_xxy = (((((((((m_Del_Kd_xxy_o0_t0_l0 + d_Del_Kd_xxy_o0_t1_m0_l0) + m_Del_Kd_xxy_o0_t2_l0) + m_Del_Kd_xxy_o0_t3_l0) + m_Del_Kd_xxy_o0_t4_l0) + m_Del_Kd_xxy_o0_t5_l0) + m_Del_Kd_xxy_o0_t6_l0) + m_Del_Kd_xxy_o0_t7_l0) + m_Del_Kd_xxy_o0_t8_l0) + m_Del_Kd_xxy_o0_t9_l0) + m_Del_Kd_xxy_o0_t10_l0;
								d_Del_Kd_xzz_o0_t1_m0_l0 = -inv_chi * (2.0 * vector(Atd_xz, i, j, k) * Ct_xxz + 2.0 * vector(Atd_yz, i, j, k) * Ct_yxz + 2.0 * vector(Atd_zz, i, j, k) * Ct_zxz);
								m_Del_Kd_xzz_o0_t6_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_xzz_o0_t5_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_xzz_o0_t4_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_xzz_o0_t3_l0 = 1.0 * (inv_chi * inv_chi) * vector(Atd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_xzz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_zz, i, j, k) * d_Del_Kd_xxy_o0_t2_m0_l0;
								m_Del_Kd_xzz_o0_t0_l0 = inv_chi * d_Del_Kd_xzz_o0_t0_m0_l0;
								Del_Kd_xzz = (((((m_Del_Kd_xzz_o0_t0_l0 + d_Del_Kd_xzz_o0_t1_m0_l0) + m_Del_Kd_xzz_o0_t2_l0) + m_Del_Kd_xzz_o0_t3_l0) + m_Del_Kd_xzz_o0_t4_l0) + m_Del_Kd_xzz_o0_t5_l0) + m_Del_Kd_xzz_o0_t6_l0;
								d_Del_Kd_xyz_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xz, i, j, k) * Ct_xxy + vector(Atd_xy, i, j, k) * Ct_xxz + vector(Atd_yz, i, j, k) * Ct_yxy + vector(Atd_yy, i, j, k) * Ct_yxz + vector(Atd_zz, i, j, k) * Ct_zxy + vector(Atd_yz, i, j, k) * Ct_zxz);
								m_Del_Kd_xyz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_yz, i, j, k) * d_Del_Kd_xxy_o0_t2_m0_l0;
								m_Del_Kd_xyz_o0_t0_l0 = inv_chi * d_Del_Kd_xyz_o0_t0_m0_l0;
								Del_Kd_xyz = (((((((((m_Del_Kd_xyz_o0_t0_l0 + d_Del_Kd_xyz_o0_t1_m0_l0) + m_Del_Kd_xyz_o0_t2_l0) + m_Del_Kd_zxy_o0_t4_l0) + m_Del_Kd_yxz_o0_t4_l0) + m_Del_Kd_yxz_o0_t5_l0) + m_Del_Kd_yxz_o0_t6_l0) + m_Del_Kd_yxz_o0_t7_l0) + m_Del_Kd_zxy_o0_t5_l0) + m_Del_Kd_zxy_o0_t6_l0) + m_Del_Kd_zxy_o0_t7_l0;
								BWeyl_xz = (-vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxz) + vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxz + vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyz + (-vector(gtd_xx, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zyz) + (-vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xzz) + vector(gtd_xx, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yzz;
								BWeyl_zz = (-vector(gtd_zz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxz) + vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxz + vector(gtd_zz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyz + (-vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zyz) + (-vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xzz) + vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yzz;
								BWeyl_yz = (-vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxz) + vector(gtd_yy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxz + vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyz + (-vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zyz) + (-vector(gtd_yy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xzz) + vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yzz;
								d_Del_Kd_xyy_o0_t1_m0_l0 = -inv_chi * (2.0 * vector(Atd_xy, i, j, k) * Ct_xxy + 2.0 * vector(Atd_yy, i, j, k) * Ct_yxy + 2.0 * vector(Atd_yz, i, j, k) * Ct_zxy);
								m_Del_Kd_xyy_o0_t6_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_xyy_o0_t5_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_xyy_o0_t4_l0 = (-1.0 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_xyy_o0_t3_l0 = 1.0 * (inv_chi * inv_chi) * vector(Atd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_xyy_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_yy, i, j, k) * d_Del_Kd_xxy_o0_t2_m0_l0;
								m_Del_Kd_xyy_o0_t0_l0 = inv_chi * d_Del_Kd_xyy_o0_t0_m0_l0;
								Del_Kd_xyy = (((((m_Del_Kd_xyy_o0_t0_l0 + d_Del_Kd_xyy_o0_t1_m0_l0) + m_Del_Kd_xyy_o0_t2_l0) + m_Del_Kd_xyy_o0_t3_l0) + m_Del_Kd_xyy_o0_t4_l0) + m_Del_Kd_xyy_o0_t5_l0) + m_Del_Kd_xyy_o0_t6_l0;
								d_Del_Kd_xxz_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xz, i, j, k) * Ct_xxx + vector(Atd_xx, i, j, k) * Ct_xxz + vector(Atd_yz, i, j, k) * Ct_yxx + vector(Atd_xy, i, j, k) * Ct_yxz + vector(Atd_zz, i, j, k) * Ct_zxx + vector(Atd_xz, i, j, k) * Ct_zxz);
								m_Del_Kd_xxz_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_xxz_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_xxz_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xz, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_xxz_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xx, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_xxz_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xx, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_xxz_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xx, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_xxz_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xx, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_xxz_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_xxz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xz, i, j, k) * d_Del_Kd_xxy_o0_t2_m0_l0;
								m_Del_Kd_xxz_o0_t0_l0 = inv_chi * d_Del_Kd_xxz_o0_t0_m0_l0;
								Del_Kd_xxz = (((((((((m_Del_Kd_xxz_o0_t0_l0 + d_Del_Kd_xxz_o0_t1_m0_l0) + m_Del_Kd_xxz_o0_t2_l0) + m_Del_Kd_xxz_o0_t3_l0) + m_Del_Kd_xxz_o0_t4_l0) + m_Del_Kd_xxz_o0_t5_l0) + m_Del_Kd_xxz_o0_t6_l0) + m_Del_Kd_xxz_o0_t7_l0) + m_Del_Kd_xxz_o0_t8_l0) + m_Del_Kd_xxz_o0_t9_l0) + m_Del_Kd_xxz_o0_t10_l0;
								BWeyl_xx = (-vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxx) + vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxx + vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xxy + (-vector(gtd_xx, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxy) + (-vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xxz) + vector(gtd_xx, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxz;
								BWeyl_zx = (-vector(gtd_zz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxx) + vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxx + vector(gtd_zz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xxy + (-vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxy) + (-vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xxz) + vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxz;
								BWeyl_yx = (-vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxx) + vector(gtd_yy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxx + vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xxy + (-vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxy) + (-vector(gtd_yy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xxz) + vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxz;
								m_Rpd_xz_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t5_m0_l0;
								m_Rpd_xz_o0_t24_l0 = (-0.75 * vector(gtd_xz, i, j, k) * gtu_zz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xz_o0_t23_l0 = (-0.75 * vector(gtd_xz, i, j, k) * gtu_yy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xz_o0_t22_l0 = (-0.75 * vector(gtd_xz, i, j, k) * gtu_xx * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xz_o0_t21_l0 = inv_chi * Zu_z * vector(gtd_zz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xz_o0_t20_l0 = inv_chi * Zu_y * vector(gtd_yz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xz_o0_t19_l0 = inv_chi * Zu_x * vector(gtd_xz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xz_o0_t18_l0 = inv_chi * Zu_z * vector(gtd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xz_o0_t17_l0 = inv_chi * Zu_y * vector(gtd_xy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xz_o0_t16_l0 = inv_chi * Zu_x * vector(gtd_xx, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xz_o0_t15_l0 = (-1.500000000000000 * vector(gtd_xz, i, j, k) * gtu_yz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xz_o0_t14_l0 = (-1.500000000000000 * vector(gtd_xz, i, j, k) * gtu_xz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xz_o0_t13_l0 = (-1.500000000000000 * vector(gtd_xz, i, j, k) * gtu_xy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xz_o0_t12_l0 = (-0.25 * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xz_o0_t11_l0 = 0.5 * vector(gtd_xz, i, j, k) * gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
								m_Rpd_xz_o0_t10_l0 = 1.0 * vector(gtd_xz, i, j, k) * gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
								m_Rpd_xz_o0_t9_l0 = 0.5 * vector(gtd_xz, i, j, k) * gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
								m_Rpd_xz_o0_t8_l0 = 1.0 * vector(gtd_xz, i, j, k) * gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
								m_Rpd_xz_o0_t7_l0 = 1.0 * vector(gtd_xz, i, j, k) * gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
								m_Rpd_xz_o0_t6_l0 = 0.5 * vector(gtd_xz, i, j, k) * gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
								m_Rpd_xz_o0_t5_l0 = (-0.5 * vector(gtd_xz, i, j, k) * vector(Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xz_o0_t4_l0 = (-0.5 * vector(gtd_xz, i, j, k) * vector(Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xz_o0_t3_l0 = (-0.5 * vector(gtd_xz, i, j, k) * vector(Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xz_o0_t2_l0 = (-0.5 * Ct_zxz) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xz_o0_t1_l0 = (-0.5 * Ct_yxz) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xz_o0_t0_l0 = (-0.5 * Ct_xxz) * d_Rpd_xx_o0_t0_m0_l0;
								Rpd_xz = ((((((((((((((((((((((((m_Rpd_xz_o0_t0_l0 + m_Rpd_xz_o0_t1_l0) + m_Rpd_xz_o0_t2_l0) + m_Rpd_xz_o0_t3_l0) + m_Rpd_xz_o0_t4_l0) + m_Rpd_xz_o0_t5_l0) + m_Rpd_xz_o0_t6_l0) + m_Rpd_xz_o0_t7_l0) + m_Rpd_xz_o0_t8_l0) + m_Rpd_xz_o0_t9_l0) + m_Rpd_xz_o0_t10_l0) + m_Rpd_xz_o0_t11_l0) + m_Rpd_xz_o0_t12_l0) + m_Rpd_xz_o0_t13_l0) + m_Rpd_xz_o0_t14_l0) + m_Rpd_xz_o0_t15_l0) + m_Rpd_xz_o0_t16_l0) + m_Rpd_xz_o0_t17_l0) + m_Rpd_xz_o0_t18_l0) + m_Rpd_xz_o0_t19_l0) + m_Rpd_xz_o0_t20_l0) + m_Rpd_xz_o0_t21_l0) + m_Rpd_xz_o0_t22_l0) + m_Rpd_xz_o0_t23_l0) + m_Rpd_xz_o0_t24_l0) + m_Rpd_xz_o0_t25_l0;
								EWeyl_xz = (-inv_chi * Rpd_xz) + (-Rtd_xz) + inv_chi * (vector(Atd_xx, i, j, k) * Atud_xz + vector(Atd_xy, i, j, k) * Atud_yz + vector(Atd_xz, i, j, k) * Atud_zz) + (-0.3333333333333333 * inv_chi * vector(trK, i, j, k) * vector(Atd_xz, i, j, k)) + (-0.2222222222222222 * inv_chi * vector(gtd_xz, i, j, k) * (vector(trK, i, j, k) * vector(trK, i, j, k)));
								m_Rpd_xy_o0_t25_l0 = 0.5 * d_Rpd_xx_o0_t4_m0_l0;
								m_Rpd_xy_o0_t24_l0 = (-0.75 * vector(gtd_xy, i, j, k) * gtu_zz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xy_o0_t23_l0 = (-0.75 * vector(gtd_xy, i, j, k) * gtu_yy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xy_o0_t22_l0 = (-0.75 * vector(gtd_xy, i, j, k) * gtu_xx * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xy_o0_t21_l0 = inv_chi * Zu_z * vector(gtd_yz, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xy_o0_t20_l0 = inv_chi * Zu_y * vector(gtd_yy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xy_o0_t19_l0 = inv_chi * Zu_x * vector(gtd_xy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xy_o0_t18_l0 = inv_chi * Zu_z * vector(gtd_xz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xy_o0_t17_l0 = inv_chi * Zu_y * vector(gtd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xy_o0_t16_l0 = inv_chi * Zu_x * vector(gtd_xx, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xy_o0_t15_l0 = (-1.500000000000000 * vector(gtd_xy, i, j, k) * gtu_yz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xy_o0_t14_l0 = (-1.500000000000000 * vector(gtd_xy, i, j, k) * gtu_xz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xy_o0_t13_l0 = (-1.500000000000000 * vector(gtd_xy, i, j, k) * gtu_xy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xy_o0_t12_l0 = (-0.25 * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xy_o0_t11_l0 = 0.5 * vector(gtd_xy, i, j, k) * gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
								m_Rpd_xy_o0_t10_l0 = 1.0 * vector(gtd_xy, i, j, k) * gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
								m_Rpd_xy_o0_t9_l0 = 0.5 * vector(gtd_xy, i, j, k) * gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
								m_Rpd_xy_o0_t8_l0 = 1.0 * vector(gtd_xy, i, j, k) * gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
								m_Rpd_xy_o0_t7_l0 = 1.0 * vector(gtd_xy, i, j, k) * gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
								m_Rpd_xy_o0_t6_l0 = 0.5 * vector(gtd_xy, i, j, k) * gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
								m_Rpd_xy_o0_t5_l0 = (-0.5 * vector(gtd_xy, i, j, k) * vector(Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xy_o0_t4_l0 = (-0.5 * vector(gtd_xy, i, j, k) * vector(Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xy_o0_t3_l0 = (-0.5 * vector(gtd_xy, i, j, k) * vector(Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_xy_o0_t2_l0 = (-0.5 * Ct_zxy) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_xy_o0_t1_l0 = (-0.5 * Ct_yxy) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_xy_o0_t0_l0 = (-0.5 * Ct_xxy) * d_Rpd_xx_o0_t0_m0_l0;
								Rpd_xy = ((((((((((((((((((((((((m_Rpd_xy_o0_t0_l0 + m_Rpd_xy_o0_t1_l0) + m_Rpd_xy_o0_t2_l0) + m_Rpd_xy_o0_t3_l0) + m_Rpd_xy_o0_t4_l0) + m_Rpd_xy_o0_t5_l0) + m_Rpd_xy_o0_t6_l0) + m_Rpd_xy_o0_t7_l0) + m_Rpd_xy_o0_t8_l0) + m_Rpd_xy_o0_t9_l0) + m_Rpd_xy_o0_t10_l0) + m_Rpd_xy_o0_t11_l0) + m_Rpd_xy_o0_t12_l0) + m_Rpd_xy_o0_t13_l0) + m_Rpd_xy_o0_t14_l0) + m_Rpd_xy_o0_t15_l0) + m_Rpd_xy_o0_t16_l0) + m_Rpd_xy_o0_t17_l0) + m_Rpd_xy_o0_t18_l0) + m_Rpd_xy_o0_t19_l0) + m_Rpd_xy_o0_t20_l0) + m_Rpd_xy_o0_t21_l0) + m_Rpd_xy_o0_t22_l0) + m_Rpd_xy_o0_t23_l0) + m_Rpd_xy_o0_t24_l0) + m_Rpd_xy_o0_t25_l0;
								m_Rpd_zz_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t8_m0_l0;
								m_Rpd_zz_o0_t21_l0 = (-0.75 * vector(gtd_zz, i, j, k) * gtu_zz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_zz_o0_t20_l0 = (-0.75 * vector(gtd_zz, i, j, k) * gtu_yy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_zz_o0_t19_l0 = (-0.75 * vector(gtd_zz, i, j, k) * gtu_xx * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_zz_o0_t18_l0 = 2.0 * inv_chi * Zu_z * vector(gtd_zz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_zz_o0_t17_l0 = 2.0 * inv_chi * Zu_y * vector(gtd_yz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_zz_o0_t16_l0 = 2.0 * inv_chi * Zu_x * vector(gtd_xz, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_zz_o0_t15_l0 = (-1.500000000000000 * vector(gtd_zz, i, j, k) * gtu_yz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_zz_o0_t14_l0 = (-1.500000000000000 * vector(gtd_zz, i, j, k) * gtu_xz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_zz_o0_t13_l0 = (-1.500000000000000 * vector(gtd_zz, i, j, k) * gtu_xy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_zz_o0_t12_l0 = (-0.25 * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_zz_o0_t11_l0 = (-0.5 * Ct_zzz) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_zz_o0_t10_l0 = (-0.5 * Ct_yzz) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_zz_o0_t9_l0 = (-0.5 * Ct_xzz) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_zz_o0_t8_l0 = 0.5 * vector(gtd_zz, i, j, k) * gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
								m_Rpd_zz_o0_t7_l0 = 1.0 * vector(gtd_zz, i, j, k) * gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
								m_Rpd_zz_o0_t6_l0 = 0.5 * vector(gtd_zz, i, j, k) * gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
								m_Rpd_zz_o0_t5_l0 = 1.0 * vector(gtd_zz, i, j, k) * gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
								m_Rpd_zz_o0_t4_l0 = 1.0 * vector(gtd_zz, i, j, k) * gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
								m_Rpd_zz_o0_t3_l0 = 0.5 * vector(gtd_zz, i, j, k) * gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
								m_Rpd_zz_o0_t2_l0 = (-0.5 * vector(gtd_zz, i, j, k) * vector(Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_zz_o0_t1_l0 = (-0.5 * vector(gtd_zz, i, j, k) * vector(Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_zz_o0_t0_l0 = (-0.5 * vector(gtd_zz, i, j, k) * vector(Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
								Rpd_zz = (((((((((((((((((((((m_Rpd_zz_o0_t0_l0 + m_Rpd_zz_o0_t1_l0) + m_Rpd_zz_o0_t2_l0) + m_Rpd_zz_o0_t3_l0) + m_Rpd_zz_o0_t4_l0) + m_Rpd_zz_o0_t5_l0) + m_Rpd_zz_o0_t6_l0) + m_Rpd_zz_o0_t7_l0) + m_Rpd_zz_o0_t8_l0) + m_Rpd_zz_o0_t9_l0) + m_Rpd_zz_o0_t10_l0) + m_Rpd_zz_o0_t11_l0) + m_Rpd_zz_o0_t12_l0) + m_Rpd_zz_o0_t13_l0) + m_Rpd_zz_o0_t14_l0) + m_Rpd_zz_o0_t15_l0) + m_Rpd_zz_o0_t16_l0) + m_Rpd_zz_o0_t17_l0) + m_Rpd_zz_o0_t18_l0) + m_Rpd_zz_o0_t19_l0) + m_Rpd_zz_o0_t20_l0) + m_Rpd_zz_o0_t21_l0) + m_Rpd_zz_o0_t22_l0;
								EWeyl_zz = (-inv_chi * Rpd_zz) + (-Rtd_zz) + inv_chi * (vector(Atd_xz, i, j, k) * Atud_xz + vector(Atd_yz, i, j, k) * Atud_yz + vector(Atd_zz, i, j, k) * Atud_zz) + (-0.3333333333333333 * inv_chi * vector(trK, i, j, k) * vector(Atd_zz, i, j, k)) + (-0.2222222222222222 * inv_chi * vector(gtd_zz, i, j, k) * (vector(trK, i, j, k) * vector(trK, i, j, k)));
								d_Del_Kd_yyz_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xz, i, j, k) * Ct_xyy + vector(Atd_xy, i, j, k) * Ct_xyz + vector(Atd_yz, i, j, k) * Ct_yyy + vector(Atd_yy, i, j, k) * Ct_yyz + vector(Atd_zz, i, j, k) * Ct_zyy + vector(Atd_yz, i, j, k) * Ct_zyz);
								m_Del_Kd_yyz_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_yyz_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_yyz_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yz, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_yyz_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yy, i, j, k) * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_yyz_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yy, i, j, k) * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_yyz_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yy, i, j, k) * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_yyz_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_yy, i, j, k) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_yyz_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_yyz_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_yz, i, j, k) * d_Del_Kd_yxx_o0_t2_m0_l0;
								m_Del_Kd_yyz_o0_t0_l0 = inv_chi * d_Del_Kd_yyz_o0_t0_m0_l0;
								Del_Kd_yyz = (((((((((m_Del_Kd_yyz_o0_t0_l0 + d_Del_Kd_yyz_o0_t1_m0_l0) + m_Del_Kd_yyz_o0_t2_l0) + m_Del_Kd_yyz_o0_t3_l0) + m_Del_Kd_yyz_o0_t4_l0) + m_Del_Kd_yyz_o0_t5_l0) + m_Del_Kd_yyz_o0_t6_l0) + m_Del_Kd_yyz_o0_t7_l0) + m_Del_Kd_yyz_o0_t8_l0) + m_Del_Kd_yyz_o0_t9_l0) + m_Del_Kd_yyz_o0_t10_l0;
								d_Del_Kd_yxy_o0_t1_m0_l0 = -inv_chi * (vector(Atd_xy, i, j, k) * Ct_xxy + vector(Atd_xx, i, j, k) * Ct_xyy + vector(Atd_yy, i, j, k) * Ct_yxy + vector(Atd_xy, i, j, k) * Ct_yyy + vector(Atd_yz, i, j, k) * Ct_zxy + vector(Atd_xz, i, j, k) * Ct_zyy);
								m_Del_Kd_yxy_o0_t10_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yy, i, j, k) * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_yxy_o0_t9_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yy, i, j, k) * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_yxy_o0_t8_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_yy, i, j, k) * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_yxy_o0_t7_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
								m_Del_Kd_yxy_o0_t6_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_yxy_o0_t5_l0 = (-0.5 * (inv_chi * inv_chi) * vector(gtd_xy, i, j, k) * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_yxy_o0_t4_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Del_Kd_yxy_o0_t3_l0 = 0.5 * (inv_chi * inv_chi) * vector(Atd_yy, i, j, k) * d_Rpd_xx_o0_t0_m0_l0;
								m_Del_Kd_yxy_o0_t2_l0 = 0.3333333333333333 * inv_chi * vector(gtd_xy, i, j, k) * d_Del_Kd_yxx_o0_t2_m0_l0;
								m_Del_Kd_yxy_o0_t0_l0 = inv_chi * d_Del_Kd_yxy_o0_t0_m0_l0;
								Del_Kd_yxy = (((((((((m_Del_Kd_yxy_o0_t0_l0 + d_Del_Kd_yxy_o0_t1_m0_l0) + m_Del_Kd_yxy_o0_t2_l0) + m_Del_Kd_yxy_o0_t3_l0) + m_Del_Kd_yxy_o0_t4_l0) + m_Del_Kd_yxy_o0_t5_l0) + m_Del_Kd_yxy_o0_t6_l0) + m_Del_Kd_yxy_o0_t7_l0) + m_Del_Kd_yxy_o0_t8_l0) + m_Del_Kd_yxy_o0_t9_l0) + m_Del_Kd_yxy_o0_t10_l0;
								BWeyl_xy = (-vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxy) + vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxy + vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyy + (-vector(gtd_xx, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zyy) + (-vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyz) + vector(gtd_xx, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yyz;
								BWeyl_zy = (-vector(gtd_zz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxy) + vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxy + vector(gtd_zz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyy + (-vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zyy) + (-vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyz) + vector(gtd_xz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yyz;
								BWeyl_yy = (-vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yxy) + vector(gtd_yy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zxy + vector(gtd_yz, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyy + (-vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_zyy) + (-vector(gtd_yy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_xyz) + vector(gtd_xy, i, j, k) * (pow(inv_chi, 2.5)) * Del_Kd_yyz;
								m_Rpd_yy_o0_t22_l0 = 0.5 * d_Rpd_xx_o0_t6_m0_l0;
								m_Rpd_yy_o0_t21_l0 = (-0.75 * vector(gtd_yy, i, j, k) * gtu_zz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_yy_o0_t20_l0 = (-0.75 * vector(gtd_yy, i, j, k) * gtu_yy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yy_o0_t19_l0 = (-0.75 * vector(gtd_yy, i, j, k) * gtu_xx * inv_chi) * d_Rpd_xx_o0_t0_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_yy_o0_t18_l0 = 2.0 * inv_chi * Zu_z * vector(gtd_yz, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yy_o0_t17_l0 = 2.0 * inv_chi * Zu_y * vector(gtd_yy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yy_o0_t16_l0 = 2.0 * inv_chi * Zu_x * vector(gtd_xy, i, j, k) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yy_o0_t15_l0 = (-1.500000000000000 * vector(gtd_yy, i, j, k) * gtu_yz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yy_o0_t14_l0 = (-1.500000000000000 * vector(gtd_yy, i, j, k) * gtu_xz * inv_chi) * d_Rpd_xx_o0_t2_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_yy_o0_t13_l0 = (-1.500000000000000 * vector(gtd_yy, i, j, k) * gtu_xy * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_yy_o0_t12_l0 = (-0.5 * vector(gtd_yy, i, j, k) * vector(Gamh_z, i, j, k)) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_yy_o0_t11_l0 = (-0.5 * vector(gtd_yy, i, j, k) * vector(Gamh_y, i, j, k)) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yy_o0_t10_l0 = (-0.5 * vector(gtd_yy, i, j, k) * vector(Gamh_x, i, j, k)) * d_Rpd_xx_o0_t0_m0_l0;
								m_Rpd_yy_o0_t9_l0 = 0.5 * vector(gtd_yy, i, j, k) * gtu_zz * d_Rpd_xx_o0_t8_m0_l0;
								m_Rpd_yy_o0_t8_l0 = 1.0 * vector(gtd_yy, i, j, k) * gtu_yz * d_Rpd_xx_o0_t7_m0_l0;
								m_Rpd_yy_o0_t7_l0 = 0.5 * vector(gtd_yy, i, j, k) * gtu_yy * d_Rpd_xx_o0_t6_m0_l0;
								m_Rpd_yy_o0_t6_l0 = 1.0 * vector(gtd_yy, i, j, k) * gtu_xz * d_Rpd_xx_o0_t5_m0_l0;
								m_Rpd_yy_o0_t5_l0 = 1.0 * vector(gtd_yy, i, j, k) * gtu_xy * d_Rpd_xx_o0_t4_m0_l0;
								m_Rpd_yy_o0_t4_l0 = 0.5 * vector(gtd_yy, i, j, k) * gtu_xx * d_Rpd_xx_o0_t3_m0_l0;
								m_Rpd_yy_o0_t3_l0 = (-0.25 * inv_chi) * d_Rpd_xx_o0_t1_m0_l0 * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yy_o0_t2_l0 = (-0.5 * Ct_zyy) * d_Rpd_xx_o0_t2_m0_l0;
								m_Rpd_yy_o0_t1_l0 = (-0.5 * Ct_yyy) * d_Rpd_xx_o0_t1_m0_l0;
								m_Rpd_yy_o0_t0_l0 = (-0.5 * Ct_xyy) * d_Rpd_xx_o0_t0_m0_l0;
								Rpd_yy = (((((((((((((((((((((m_Rpd_yy_o0_t0_l0 + m_Rpd_yy_o0_t1_l0) + m_Rpd_yy_o0_t2_l0) + m_Rpd_yy_o0_t3_l0) + m_Rpd_yy_o0_t4_l0) + m_Rpd_yy_o0_t5_l0) + m_Rpd_yy_o0_t6_l0) + m_Rpd_yy_o0_t7_l0) + m_Rpd_yy_o0_t8_l0) + m_Rpd_yy_o0_t9_l0) + m_Rpd_yy_o0_t10_l0) + m_Rpd_yy_o0_t11_l0) + m_Rpd_yy_o0_t12_l0) + m_Rpd_yy_o0_t13_l0) + m_Rpd_yy_o0_t14_l0) + m_Rpd_yy_o0_t15_l0) + m_Rpd_yy_o0_t16_l0) + m_Rpd_yy_o0_t17_l0) + m_Rpd_yy_o0_t18_l0) + m_Rpd_yy_o0_t19_l0) + m_Rpd_yy_o0_t20_l0) + m_Rpd_yy_o0_t21_l0) + m_Rpd_yy_o0_t22_l0;
								d_Rtd_xy_o0_t12_m0_l0 = gtu_zz * (Ct_xxz * Ctd_yxz + Ct_xxz * Ctd_xyz + Ct_xyz * Ctd_xxz + 2.0 * Ct_yxz * Ctd_yyz + Ct_yyz * Ctd_xyz + Ct_zxz * Ctd_yzz + Ct_zxz * Ctd_zyz + Ct_zyz * Ctd_xzz) + gtu_yz * (Ct_xxy * Ctd_yxz + Ct_xxy * Ctd_xyz + Ct_xyy * Ctd_xxz + 2.0 * Ct_yxy * Ctd_yyz + Ct_yyy * Ctd_xyz + Ct_zxy * Ctd_yzz + Ct_zxy * Ctd_zyz + Ct_zyy * Ctd_xzz) + gtu_xz * (Ct_xxx * Ctd_yxz + Ct_xxx * Ctd_xyz + Ct_xxy * Ctd_xxz + 2.0 * Ct_yxx * Ctd_yyz + Ct_yxy * Ctd_xyz + Ct_zxx * Ctd_yzz + Ct_zxx * Ctd_zyz + Ct_zxy * Ctd_xzz) + gtu_yz * (Ct_xxz * Ctd_yxy + Ct_xxz * Ctd_xyy + Ct_xyz * Ctd_xxy + 2.0 * Ct_yxz * Ctd_yyy + Ct_yyz * Ctd_xyy + Ct_zxz * Ctd_yyz + Ct_zxz * Ctd_zyy + Ct_zyz * Ctd_xyz) + gtu_yy * (Ct_xxy * Ctd_yxy + Ct_xxy * Ctd_xyy + Ct_xyy * Ctd_xxy + 2.0 * Ct_yxy * Ctd_yyy + Ct_yyy * Ctd_xyy + Ct_zxy * Ctd_yyz + Ct_zxy * Ctd_zyy + Ct_zyy * Ctd_xyz) + gtu_xy * (Ct_xxx * Ctd_yxy + Ct_xxx * Ctd_xyy + Ct_xxy * Ctd_xxy + 2.0 * Ct_yxx * Ctd_yyy + Ct_yxy * Ctd_xyy + Ct_zxx * Ctd_yyz + Ct_zxx * Ctd_zyy + Ct_zxy * Ctd_xyz) + gtu_xz * (Ct_xxz * Ctd_yxx + Ct_xxz * Ctd_xxy + Ct_xyz * Ctd_xxx + 2.0 * Ct_yxz * Ctd_yxy + Ct_yyz * Ctd_xxy + Ct_zxz * Ctd_yxz + Ct_zxz * Ctd_zxy + Ct_zyz * Ctd_xxz) + gtu_xy * (Ct_xxy * Ctd_yxx + Ct_xxy * Ctd_xxy + Ct_xyy * Ctd_xxx + 2.0 * Ct_yxy * Ctd_yxy + Ct_yyy * Ctd_xxy + Ct_zxy * Ctd_yxz + Ct_zxy * Ctd_zxy + Ct_zyy * Ctd_xxz) + gtu_xx * (Ct_xxx * Ctd_yxx + Ct_xxx * Ctd_xxy + Ct_xxy * Ctd_xxx + 2.0 * Ct_yxx * Ctd_yxy + Ct_yxy * Ctd_xxy + Ct_zxx * Ctd_yxz + Ct_zxx * Ctd_zxy + Ct_zxy * Ctd_xxz) + 0.5 * vector(Gamh_z, i, j, k) * (Ctd_xyz + Ctd_yxz) + 0.5 * vector(Gamh_y, i, j, k) * (Ctd_xyy + Ctd_yxy) + 0.5 * vector(Gamh_x, i, j, k) * (Ctd_xxy + Ctd_yxx);
								m_Rtd_xy_o0_t11_l0 = 0.5 * vector(gtd_yz, i, j, k) * d_Rtd_xx_o0_t8_m0_l0;
								m_Rtd_xy_o0_t10_l0 = 0.5 * vector(gtd_yy, i, j, k) * d_Rtd_xx_o0_t7_m0_l0;
								m_Rtd_xy_o0_t9_l0 = 0.5 * vector(gtd_xy, i, j, k) * d_Rtd_xx_o0_t6_m0_l0;
								m_Rtd_xy_o0_t8_l0 = 0.5 * vector(gtd_xz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
								m_Rtd_xy_o0_t7_l0 = 0.5 * vector(gtd_xy, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
								m_Rtd_xy_o0_t6_l0 = 0.5 * vector(gtd_xx, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
								m_Rtd_xy_o0_t5_l0 = (-0.5 * gtu_zz) * d_Rtd_xy_o0_t5_m0_l0;
								m_Rtd_xy_o0_t4_l0 = (-1.0 * gtu_yz) * d_Rtd_xy_o0_t4_m0_l0;
								m_Rtd_xy_o0_t3_l0 = (-0.5 * gtu_yy) * d_Rtd_xy_o0_t3_m0_l0;
								m_Rtd_xy_o0_t2_l0 = (-1.0 * gtu_xz) * d_Rtd_xy_o0_t2_m0_l0;
								m_Rtd_xy_o0_t1_l0 = (-1.0 * gtu_xy) * d_Rtd_xy_o0_t1_m0_l0;
								m_Rtd_xy_o0_t0_l0 = (-0.5 * gtu_xx) * d_Rtd_xy_o0_t0_m0_l0;
								Rtd_xy = (((((((((((m_Rtd_xy_o0_t0_l0 + m_Rtd_xy_o0_t1_l0) + m_Rtd_xy_o0_t2_l0) + m_Rtd_xy_o0_t3_l0) + m_Rtd_xy_o0_t4_l0) + m_Rtd_xy_o0_t5_l0) + m_Rtd_xy_o0_t6_l0) + m_Rtd_xy_o0_t7_l0) + m_Rtd_xy_o0_t8_l0) + m_Rtd_xy_o0_t9_l0) + m_Rtd_xy_o0_t10_l0) + m_Rtd_xy_o0_t11_l0) + d_Rtd_xy_o0_t12_m0_l0;
								EWeyl_xy = (-inv_chi * Rpd_xy) + (-Rtd_xy) + inv_chi * (vector(Atd_xx, i, j, k) * Atud_xy + vector(Atd_xy, i, j, k) * Atud_yy + vector(Atd_xz, i, j, k) * Atud_zy) + (-0.3333333333333333 * inv_chi * vector(trK, i, j, k) * vector(Atd_xy, i, j, k)) + (-0.2222222222222222 * inv_chi * vector(gtd_xy, i, j, k) * (vector(trK, i, j, k) * vector(trK, i, j, k)));
								d_Rtd_yy_o0_t9_m0_l0 = gtu_zz * (2.0 * Ct_xyz * Ctd_yxz + Ct_xyz * Ctd_xyz + 3.0 * Ct_yyz * Ctd_yyz + 2.0 * Ct_zyz * Ctd_yzz + Ct_zyz * Ctd_zyz) + gtu_yz * (2.0 * Ct_xyy * Ctd_yxz + Ct_xyy * Ctd_xyz + 3.0 * Ct_yyy * Ctd_yyz + 2.0 * Ct_zyy * Ctd_yzz + Ct_zyy * Ctd_zyz) + gtu_xz * (2.0 * Ct_xxy * Ctd_yxz + Ct_xxy * Ctd_xyz + 3.0 * Ct_yxy * Ctd_yyz + 2.0 * Ct_zxy * Ctd_yzz + Ct_zxy * Ctd_zyz) + gtu_yz * (2.0 * Ct_xyz * Ctd_yxy + Ct_xyz * Ctd_xyy + 3.0 * Ct_yyz * Ctd_yyy + 2.0 * Ct_zyz * Ctd_yyz + Ct_zyz * Ctd_zyy) + gtu_yy * (2.0 * Ct_xyy * Ctd_yxy + Ct_xyy * Ctd_xyy + 3.0 * Ct_yyy * Ctd_yyy + 2.0 * Ct_zyy * Ctd_yyz + Ct_zyy * Ctd_zyy) + gtu_xy * (2.0 * Ct_xxy * Ctd_yxy + Ct_xxy * Ctd_xyy + 3.0 * Ct_yxy * Ctd_yyy + 2.0 * Ct_zxy * Ctd_yyz + Ct_zxy * Ctd_zyy) + gtu_xz * (2.0 * Ct_xyz * Ctd_yxx + Ct_xyz * Ctd_xxy + 3.0 * Ct_yyz * Ctd_yxy + 2.0 * Ct_zyz * Ctd_yxz + Ct_zyz * Ctd_zxy) + gtu_xy * (2.0 * Ct_xyy * Ctd_yxx + Ct_xyy * Ctd_xxy + 3.0 * Ct_yyy * Ctd_yxy + 2.0 * Ct_zyy * Ctd_yxz + Ct_zyy * Ctd_zxy) + gtu_xx * (2.0 * Ct_xxy * Ctd_yxx + Ct_xxy * Ctd_xxy + 3.0 * Ct_yxy * Ctd_yxy + 2.0 * Ct_zxy * Ctd_yxz + Ct_zxy * Ctd_zxy) + 1.0 * vector(Gamh_z, i, j, k) * Ctd_yyz + 1.0 * vector(Gamh_y, i, j, k) * Ctd_yyy + 1.0 * vector(Gamh_x, i, j, k) * Ctd_yxy;
								m_Rtd_yy_o0_t8_l0 = 1.0 * vector(gtd_yz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
								m_Rtd_yy_o0_t7_l0 = 1.0 * vector(gtd_yy, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
								m_Rtd_yy_o0_t6_l0 = 1.0 * vector(gtd_xy, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
								m_Rtd_yy_o0_t5_l0 = (-0.5 * gtu_zz) * d_Rtd_yy_o0_t5_m0_l0;
								m_Rtd_yy_o0_t4_l0 = (-1.0 * gtu_yz) * d_Rtd_yy_o0_t4_m0_l0;
								m_Rtd_yy_o0_t3_l0 = (-0.5 * gtu_yy) * d_Rtd_yy_o0_t3_m0_l0;
								m_Rtd_yy_o0_t2_l0 = (-1.0 * gtu_xz) * d_Rtd_yy_o0_t2_m0_l0;
								m_Rtd_yy_o0_t1_l0 = (-1.0 * gtu_xy) * d_Rtd_yy_o0_t1_m0_l0;
								m_Rtd_yy_o0_t0_l0 = (-0.5 * gtu_xx) * d_Rtd_yy_o0_t0_m0_l0;
								Rtd_yy = ((((((((m_Rtd_yy_o0_t0_l0 + m_Rtd_yy_o0_t1_l0) + m_Rtd_yy_o0_t2_l0) + m_Rtd_yy_o0_t3_l0) + m_Rtd_yy_o0_t4_l0) + m_Rtd_yy_o0_t5_l0) + m_Rtd_yy_o0_t6_l0) + m_Rtd_yy_o0_t7_l0) + m_Rtd_yy_o0_t8_l0) + d_Rtd_yy_o0_t9_m0_l0;
								EWeyl_yy = (-inv_chi * Rpd_yy) + (-Rtd_yy) + inv_chi * (vector(Atd_xy, i, j, k) * Atud_xy + vector(Atd_yy, i, j, k) * Atud_yy + vector(Atd_yz, i, j, k) * Atud_zy) + (-0.3333333333333333 * inv_chi * vector(trK, i, j, k) * vector(Atd_yy, i, j, k)) + (-0.2222222222222222 * inv_chi * vector(gtd_yy, i, j, k) * (vector(trK, i, j, k) * vector(trK, i, j, k)));
								d_Rtd_yz_o0_t12_m0_l0 = gtu_zz * (Ct_xyz * Ctd_zxz + Ct_xyz * Ctd_xzz + Ct_xzz * Ctd_yxz + Ct_yyz * Ctd_zyz + Ct_yyz * Ctd_yzz + Ct_yzz * Ctd_yyz + 2.0 * Ct_zyz * Ctd_zzz + Ct_zzz * Ctd_yzz) + gtu_yz * (Ct_xyy * Ctd_zxz + Ct_xyy * Ctd_xzz + Ct_xyz * Ctd_yxz + Ct_yyy * Ctd_zyz + Ct_yyy * Ctd_yzz + Ct_yyz * Ctd_yyz + 2.0 * Ct_zyy * Ctd_zzz + Ct_zyz * Ctd_yzz) + gtu_xz * (Ct_xxy * Ctd_zxz + Ct_xxy * Ctd_xzz + Ct_xxz * Ctd_yxz + Ct_yxy * Ctd_zyz + Ct_yxy * Ctd_yzz + Ct_yxz * Ctd_yyz + 2.0 * Ct_zxy * Ctd_zzz + Ct_zxz * Ctd_yzz) + gtu_yz * (Ct_xyz * Ctd_zxy + Ct_xyz * Ctd_xyz + Ct_xzz * Ctd_yxy + Ct_yyz * Ctd_zyy + Ct_yyz * Ctd_yyz + Ct_yzz * Ctd_yyy + 2.0 * Ct_zyz * Ctd_zyz + Ct_zzz * Ctd_yyz) + gtu_yy * (Ct_xyy * Ctd_zxy + Ct_xyy * Ctd_xyz + Ct_xyz * Ctd_yxy + Ct_yyy * Ctd_zyy + Ct_yyy * Ctd_yyz + Ct_yyz * Ctd_yyy + 2.0 * Ct_zyy * Ctd_zyz + Ct_zyz * Ctd_yyz) + gtu_xy * (Ct_xxy * Ctd_zxy + Ct_xxy * Ctd_xyz + Ct_xxz * Ctd_yxy + Ct_yxy * Ctd_zyy + Ct_yxy * Ctd_yyz + Ct_yxz * Ctd_yyy + 2.0 * Ct_zxy * Ctd_zyz + Ct_zxz * Ctd_yyz) + gtu_xz * (Ct_xyz * Ctd_zxx + Ct_xyz * Ctd_xxz + Ct_xzz * Ctd_yxx + Ct_yyz * Ctd_zxy + Ct_yyz * Ctd_yxz + Ct_yzz * Ctd_yxy + 2.0 * Ct_zyz * Ctd_zxz + Ct_zzz * Ctd_yxz) + gtu_xy * (Ct_xyy * Ctd_zxx + Ct_xyy * Ctd_xxz + Ct_xyz * Ctd_yxx + Ct_yyy * Ctd_zxy + Ct_yyy * Ctd_yxz + Ct_yyz * Ctd_yxy + 2.0 * Ct_zyy * Ctd_zxz + Ct_zyz * Ctd_yxz) + gtu_xx * (Ct_xxy * Ctd_zxx + Ct_xxy * Ctd_xxz + Ct_xxz * Ctd_yxx + Ct_yxy * Ctd_zxy + Ct_yxy * Ctd_yxz + Ct_yxz * Ctd_yxy + 2.0 * Ct_zxy * Ctd_zxz + Ct_zxz * Ctd_yxz) + 0.5 * vector(Gamh_z, i, j, k) * (Ctd_yzz + Ctd_zyz) + 0.5 * vector(Gamh_y, i, j, k) * (Ctd_yyz + Ctd_zyy) + 0.5 * vector(Gamh_x, i, j, k) * (Ctd_yxz + Ctd_zxy);
								m_Rtd_yz_o0_t11_l0 = 0.5 * vector(gtd_zz, i, j, k) * d_Rtd_xy_o0_t8_m0_l0;
								m_Rtd_yz_o0_t10_l0 = 0.5 * vector(gtd_yz, i, j, k) * d_Rtd_xy_o0_t7_m0_l0;
								m_Rtd_yz_o0_t9_l0 = 0.5 * vector(gtd_xz, i, j, k) * d_Rtd_xy_o0_t6_m0_l0;
								m_Rtd_yz_o0_t8_l0 = 0.5 * vector(gtd_yz, i, j, k) * d_Rtd_xz_o0_t8_m0_l0;
								m_Rtd_yz_o0_t7_l0 = 0.5 * vector(gtd_yy, i, j, k) * d_Rtd_xz_o0_t7_m0_l0;
								m_Rtd_yz_o0_t6_l0 = 0.5 * vector(gtd_xy, i, j, k) * d_Rtd_xz_o0_t6_m0_l0;
								m_Rtd_yz_o0_t5_l0 = (-0.5 * gtu_zz) * d_Rtd_yz_o0_t5_m0_l0;
								m_Rtd_yz_o0_t4_l0 = (-1.0 * gtu_yz) * d_Rtd_yz_o0_t4_m0_l0;
								m_Rtd_yz_o0_t3_l0 = (-0.5 * gtu_yy) * d_Rtd_yz_o0_t3_m0_l0;
								m_Rtd_yz_o0_t2_l0 = (-1.0 * gtu_xz) * d_Rtd_yz_o0_t2_m0_l0;
								m_Rtd_yz_o0_t1_l0 = (-1.0 * gtu_xy) * d_Rtd_yz_o0_t1_m0_l0;
								m_Rtd_yz_o0_t0_l0 = (-0.5 * gtu_xx) * d_Rtd_yz_o0_t0_m0_l0;
								Rtd_yz = (((((((((((m_Rtd_yz_o0_t0_l0 + m_Rtd_yz_o0_t1_l0) + m_Rtd_yz_o0_t2_l0) + m_Rtd_yz_o0_t3_l0) + m_Rtd_yz_o0_t4_l0) + m_Rtd_yz_o0_t5_l0) + m_Rtd_yz_o0_t6_l0) + m_Rtd_yz_o0_t7_l0) + m_Rtd_yz_o0_t8_l0) + m_Rtd_yz_o0_t9_l0) + m_Rtd_yz_o0_t10_l0) + m_Rtd_yz_o0_t11_l0) + d_Rtd_yz_o0_t12_m0_l0;
								EWeyl_yz = (-inv_chi * Rpd_yz) + (-Rtd_yz) + inv_chi * (vector(Atd_xy, i, j, k) * Atud_xz + vector(Atd_yy, i, j, k) * Atud_yz + vector(Atd_yz, i, j, k) * Atud_zz) + (-0.3333333333333333 * inv_chi * vector(trK, i, j, k) * vector(Atd_yz, i, j, k)) + (-0.2222222222222222 * inv_chi * vector(gtd_yz, i, j, k) * (vector(trK, i, j, k) * vector(trK, i, j, k)));
								Bfvf = vector(Bfu_x, i, j, k) * vector(vfd_x, i, j, k) + vector(Bfu_y, i, j, k) * vector(vfd_y, i, j, k) + vector(Bfu_z, i, j, k) * vector(vfd_z, i, j, k);
								Sfuu_zz = 1.0 * Sfu_z * vfu_z + vector(chi, i, j, k) * gtu_zz * sdetg * vector(pf, i, j, k) + (-((-0.5 * vector(chi, i, j, k) * sqBf * gtu_zz) + vector(Bfu_z, i, j, k) * vector(Bfu_z, i, j, k)) / (sdetg * sqW)) + (-0.5 * (Bfvf * ((-Bfvf * vector(chi, i, j, k) * gtu_zz) + 2.0 * vector(Bfu_z, i, j, k) * vfu_z)) / sdetg);
								Tf4u_zz = Sfuu_zz / sdetg + (-(Sfu_z * vector(Betau_z, i, j, k)) / (sdetg * vector(Alpha, i, j, k))) + (-vector(Betau_z, i, j, k) * Tf4u_tz);
								Tu_zz = Tf4u_zz;
								Sfuu_yz = 0.5 * Sfu_y * vfu_z + 0.5 * Sfu_z * vfu_y + vector(chi, i, j, k) * gtu_yz * sdetg * vector(pf, i, j, k) + (-((-0.5 * vector(chi, i, j, k) * sqBf * gtu_yz) + vector(Bfu_y, i, j, k) * vector(Bfu_z, i, j, k)) / (sdetg * sqW)) + (-0.5 * (Bfvf * ((-Bfvf * vector(chi, i, j, k) * gtu_yz) + vector(Bfu_y, i, j, k) * vfu_z + vector(Bfu_z, i, j, k) * vfu_y)) / sdetg);
								Tf4u_yz = Sfuu_yz / sdetg + (-(Sfu_y * vector(Betau_z, i, j, k)) / (sdetg * vector(Alpha, i, j, k))) + (-vector(Betau_y, i, j, k) * Tf4u_tz);
								Tu_yz = Tf4u_yz;
								Sfuu_yy = 1.0 * Sfu_y * vfu_y + vector(chi, i, j, k) * gtu_yy * sdetg * vector(pf, i, j, k) + (-((-0.5 * vector(chi, i, j, k) * sqBf * gtu_yy) + vector(Bfu_y, i, j, k) * vector(Bfu_y, i, j, k)) / (sdetg * sqW)) + (-0.5 * (Bfvf * ((-Bfvf * vector(chi, i, j, k) * gtu_yy) + 2.0 * vector(Bfu_y, i, j, k) * vfu_y)) / sdetg);
								Tf4u_yy = Sfuu_yy / sdetg + (-(Sfu_y * vector(Betau_y, i, j, k)) / (sdetg * vector(Alpha, i, j, k))) + (-vector(Betau_y, i, j, k) * Tf4u_ty);
								Tu_yy = Tf4u_yy;
								Sfuu_xy = 0.5 * Sfu_x * vfu_y + 0.5 * Sfu_y * vfu_x + vector(chi, i, j, k) * gtu_xy * sdetg * vector(pf, i, j, k) + (-((-0.5 * vector(chi, i, j, k) * sqBf * gtu_xy) + vector(Bfu_x, i, j, k) * vector(Bfu_y, i, j, k)) / (sdetg * sqW)) + (-0.5 * (Bfvf * ((-Bfvf * vector(chi, i, j, k) * gtu_xy) + vector(Bfu_x, i, j, k) * vfu_y + vector(Bfu_y, i, j, k) * vfu_x)) / sdetg);
								Tf4u_xy = Sfuu_xy / sdetg + (-(Sfu_x * vector(Betau_y, i, j, k)) / (sdetg * vector(Alpha, i, j, k))) + (-vector(Betau_x, i, j, k) * Tf4u_ty);
								Tu_xy = Tf4u_xy;
								Sfuu_xx = 1.0 * Sfu_x * vfu_x + vector(chi, i, j, k) * gtu_xx * sdetg * vector(pf, i, j, k) + (-((-0.5 * vector(chi, i, j, k) * sqBf * gtu_xx) + vector(Bfu_x, i, j, k) * vector(Bfu_x, i, j, k)) / (sdetg * sqW)) + (-0.5 * (Bfvf * ((-Bfvf * vector(chi, i, j, k) * gtu_xx) + 2.0 * vector(Bfu_x, i, j, k) * vfu_x)) / sdetg);
								Tf4u_xx = Sfuu_xx / sdetg + (-(Sfu_x * vector(Betau_x, i, j, k)) / (sdetg * vector(Alpha, i, j, k))) + (-vector(Betau_x, i, j, k) * Tf4u_tx);
								Tu_xx = Tf4u_xx;
								Sfuu_xz = 0.5 * Sfu_x * vfu_z + 0.5 * Sfu_z * vfu_x + vector(chi, i, j, k) * gtu_xz * sdetg * vector(pf, i, j, k) + (-((-0.5 * vector(chi, i, j, k) * sqBf * gtu_xz) + vector(Bfu_x, i, j, k) * vector(Bfu_z, i, j, k)) / (sdetg * sqW)) + (-0.5 * (Bfvf * ((-Bfvf * vector(chi, i, j, k) * gtu_xz) + vector(Bfu_x, i, j, k) * vfu_z + vector(Bfu_z, i, j, k) * vfu_x)) / sdetg);
								Tf4u_xz = Sfuu_xz / sdetg + (-(Sfu_x * vector(Betau_z, i, j, k)) / (sdetg * vector(Alpha, i, j, k))) + (-vector(Betau_x, i, j, k) * Tf4u_tz);
								Tu_xz = Tf4u_xz;
								Tu_tt = Tf4u_tt;
								Td_yz = g4d_ty * g4d_tz * Tu_tt + g4d_xy * g4d_tz * Tu_tx + g4d_yy * g4d_tz * Tu_ty + g4d_yz * g4d_tz * Tu_tz + g4d_ty * g4d_xz * Tu_tx + g4d_xy * g4d_xz * Tu_xx + g4d_yy * g4d_xz * Tu_xy + g4d_yz * g4d_xz * Tu_xz + g4d_ty * g4d_yz * Tu_ty + g4d_xy * g4d_yz * Tu_xy + g4d_yy * g4d_yz * Tu_yy + (g4d_yz * g4d_yz) * Tu_yz + g4d_ty * g4d_zz * Tu_tz + g4d_xy * g4d_zz * Tu_xz + g4d_yy * g4d_zz * Tu_yz + g4d_yz * g4d_zz * Tu_zz;
								Td_yy = (g4d_ty * g4d_ty) * Tu_tt + 2.0 * g4d_xy * g4d_ty * Tu_tx + 2.0 * g4d_yy * g4d_ty * Tu_ty + 2.0 * g4d_yz * g4d_ty * Tu_tz + (g4d_xy * g4d_xy) * Tu_xx + 2.0 * g4d_yy * g4d_xy * Tu_xy + 2.0 * g4d_yz * g4d_xy * Tu_xz + (g4d_yy * g4d_yy) * Tu_yy + 2.0 * g4d_yz * g4d_yy * Tu_yz + (g4d_yz * g4d_yz) * Tu_zz;
								Td_xz = g4d_tx * g4d_tz * Tu_tt + g4d_xx * g4d_tz * Tu_tx + g4d_xy * g4d_tz * Tu_ty + g4d_xz * g4d_tz * Tu_tz + g4d_tx * g4d_xz * Tu_tx + g4d_xx * g4d_xz * Tu_xx + g4d_xy * g4d_xz * Tu_xy + (g4d_xz * g4d_xz) * Tu_xz + g4d_tx * g4d_yz * Tu_ty + g4d_xx * g4d_yz * Tu_xy + g4d_xy * g4d_yz * Tu_yy + g4d_xz * g4d_yz * Tu_yz + g4d_tx * g4d_zz * Tu_tz + g4d_xx * g4d_zz * Tu_xz + g4d_xy * g4d_zz * Tu_yz + g4d_xz * g4d_zz * Tu_zz;
								Td_xy = g4d_tx * g4d_ty * Tu_tt + g4d_xx * g4d_ty * Tu_tx + g4d_xy * g4d_ty * Tu_ty + g4d_xz * g4d_ty * Tu_tz + g4d_tx * g4d_xy * Tu_tx + g4d_xx * g4d_xy * Tu_xx + (g4d_xy * g4d_xy) * Tu_xy + g4d_xz * g4d_xy * Tu_xz + g4d_tx * g4d_yy * Tu_ty + g4d_xx * g4d_yy * Tu_xy + g4d_xy * g4d_yy * Tu_yy + g4d_xz * g4d_yy * Tu_yz + g4d_tx * g4d_yz * Tu_tz + g4d_xx * g4d_yz * Tu_xz + g4d_xy * g4d_yz * Tu_yz + g4d_xz * g4d_yz * Tu_zz;
								Td_xx = (g4d_tx * g4d_tx) * Tu_tt + 2.0 * g4d_xx * g4d_tx * Tu_tx + 2.0 * g4d_xy * g4d_tx * Tu_ty + 2.0 * g4d_xz * g4d_tx * Tu_tz + (g4d_xx * g4d_xx) * Tu_xx + 2.0 * g4d_xy * g4d_xx * Tu_xy + 2.0 * g4d_xz * g4d_xx * Tu_xz + (g4d_xy * g4d_xy) * Tu_yy + 2.0 * g4d_xz * g4d_xy * Tu_yz + (g4d_xz * g4d_xz) * Tu_zz;
								Td_zz = (g4d_tz * g4d_tz) * Tu_tt + 2.0 * g4d_xz * g4d_tz * Tu_tx + 2.0 * g4d_yz * g4d_tz * Tu_ty + 2.0 * g4d_zz * g4d_tz * Tu_tz + (g4d_xz * g4d_xz) * Tu_xx + 2.0 * g4d_yz * g4d_xz * Tu_xy + 2.0 * g4d_zz * g4d_xz * Tu_xz + (g4d_yz * g4d_yz) * Tu_yy + 2.0 * g4d_zz * g4d_yz * Tu_yz + (g4d_zz * g4d_zz) * Tu_zz;
								TT = g4d_tt * Tu_tt + 2.0 * g4d_tx * Tu_tx + 2.0 * g4d_ty * Tu_ty + 2.0 * g4d_tz * Tu_tz + g4d_xx * Tu_xx + 2.0 * g4d_xy * Tu_xy + 2.0 * g4d_xz * Tu_xz + g4d_yy * Tu_yy + 2.0 * g4d_yz * Tu_yz + g4d_zz * Tu_zz;
								Td_tx = g4d_tt * g4d_tx * Tu_tt + (g4d_tx * g4d_tx) * Tu_tx + g4d_ty * g4d_tx * Tu_ty + g4d_tz * g4d_tx * Tu_tz + g4d_tt * g4d_xx * Tu_tx + g4d_tx * g4d_xx * Tu_xx + g4d_ty * g4d_xx * Tu_xy + g4d_tz * g4d_xx * Tu_xz + g4d_tt * g4d_xy * Tu_ty + g4d_tx * g4d_xy * Tu_xy + g4d_ty * g4d_xy * Tu_yy + g4d_tz * g4d_xy * Tu_yz + g4d_tt * g4d_xz * Tu_tz + g4d_tx * g4d_xz * Tu_xz + g4d_ty * g4d_xz * Tu_yz + g4d_tz * g4d_xz * Tu_zz;
								Td_tz = g4d_tt * g4d_tz * Tu_tt + g4d_tx * g4d_tz * Tu_tx + g4d_ty * g4d_tz * Tu_ty + (g4d_tz * g4d_tz) * Tu_tz + g4d_tt * g4d_xz * Tu_tx + g4d_tx * g4d_xz * Tu_xx + g4d_ty * g4d_xz * Tu_xy + g4d_tz * g4d_xz * Tu_xz + g4d_tt * g4d_yz * Tu_ty + g4d_tx * g4d_yz * Tu_xy + g4d_ty * g4d_yz * Tu_yy + g4d_tz * g4d_yz * Tu_yz + g4d_tt * g4d_zz * Tu_tz + g4d_tx * g4d_zz * Tu_xz + g4d_ty * g4d_zz * Tu_yz + g4d_tz * g4d_zz * Tu_zz;
								Td_ty = g4d_tt * g4d_ty * Tu_tt + g4d_tx * g4d_ty * Tu_tx + (g4d_ty * g4d_ty) * Tu_ty + g4d_tz * g4d_ty * Tu_tz + g4d_tt * g4d_xy * Tu_tx + g4d_tx * g4d_xy * Tu_xx + g4d_ty * g4d_xy * Tu_xy + g4d_tz * g4d_xy * Tu_xz + g4d_tt * g4d_yy * Tu_ty + g4d_tx * g4d_yy * Tu_xy + g4d_ty * g4d_yy * Tu_yy + g4d_tz * g4d_yy * Tu_yz + g4d_tt * g4d_yz * Tu_tz + g4d_tx * g4d_yz * Tu_xz + g4d_ty * g4d_yz * Tu_yz + g4d_tz * g4d_yz * Tu_zz;
								Jtd_ADM_z = vector(Alpha, i, j, k) * ((vector(Betau_x, i, j, k) * Tu_tt + Tu_tx) * vector(gtd_xz, i, j, k) + (vector(Betau_y, i, j, k) * Tu_tt + Tu_ty) * vector(gtd_yz, i, j, k) + (vector(Betau_z, i, j, k) * Tu_tt + Tu_tz) * vector(gtd_zz, i, j, k));
								Td_tt = (g4d_tt * g4d_tt) * Tu_tt + 2.0 * g4d_tx * g4d_tt * Tu_tx + 2.0 * g4d_ty * g4d_tt * Tu_ty + 2.0 * g4d_tz * g4d_tt * Tu_tz + (g4d_tx * g4d_tx) * Tu_xx + 2.0 * g4d_ty * g4d_tx * Tu_xy + 2.0 * g4d_tz * g4d_tx * Tu_xz + (g4d_ty * g4d_ty) * Tu_yy + 2.0 * g4d_tz * g4d_ty * Tu_yz + (g4d_tz * g4d_tz) * Tu_zz;
								Jtd_ADM_y = vector(Alpha, i, j, k) * ((vector(Betau_x, i, j, k) * Tu_tt + Tu_tx) * vector(gtd_xy, i, j, k) + (vector(Betau_y, i, j, k) * Tu_tt + Tu_ty) * vector(gtd_yy, i, j, k) + (vector(Betau_z, i, j, k) * Tu_tt + Tu_tz) * vector(gtd_yz, i, j, k));
								Jtd_ADM_x = vector(Alpha, i, j, k) * ((vector(Betau_x, i, j, k) * Tu_tt + Tu_tx) * vector(gtd_xx, i, j, k) + (vector(Betau_y, i, j, k) * Tu_tt + Tu_ty) * vector(gtd_xy, i, j, k) + (vector(Betau_z, i, j, k) * Tu_tt + Tu_tz) * vector(gtd_xz, i, j, k));
								rho_ADM = (vector(Alpha, i, j, k) * vector(Alpha, i, j, k)) * Tu_tt;
								d_Rscalar_o0_t0_m0_l0 = (vector(chi, i, j, k) * Rtd_xx + Rpd_xx) * gtu_xx + 2.0 * (vector(chi, i, j, k) * Rtd_xy + Rpd_xy) * gtu_xy + 2.0 * (vector(chi, i, j, k) * Rtd_xz + Rpd_xz) * gtu_xz + (vector(chi, i, j, k) * Rtd_yy + Rpd_yy) * gtu_yy + 2.0 * (vector(chi, i, j, k) * Rtd_yz + Rpd_yz) * gtu_yz + (vector(chi, i, j, k) * Rtd_zz + Rpd_zz) * gtu_zz;
								d_HamCon_o0_t0_m0_l0 = (-16.0 * M_PI * rho_ADM) + (vector(chi, i, j, k) * Rtd_xx + Rpd_xx) * gtu_xx + 2.0 * (vector(chi, i, j, k) * Rtd_xy + Rpd_xy) * gtu_xy + 2.0 * (vector(chi, i, j, k) * Rtd_xz + Rpd_xz) * gtu_xz + (vector(chi, i, j, k) * Rtd_yy + Rpd_yy) * gtu_yy + 2.0 * (vector(chi, i, j, k) * Rtd_yz + Rpd_yz) * gtu_yz + (vector(chi, i, j, k) * Rtd_zz + Rpd_zz) * gtu_zz + (-vector(Atd_xx, i, j, k) * Atu_xx) + (-2.0 * vector(Atd_xy, i, j, k) * Atu_xy) + (-2.0 * vector(Atd_xz, i, j, k) * Atu_xz) + (-vector(Atd_yy, i, j, k) * Atu_yy) + (-2.0 * vector(Atd_yz, i, j, k) * Atu_yz) + (-vector(Atd_zz, i, j, k) * Atu_zz) + 0.6666666666666667 * (vector(trK, i, j, k) * vector(trK, i, j, k));
								d_MomCon_x_o0_t13_m0_l0 = (-8.0) * M_PI * Jtd_ADM_x * inv_chi + (-Ct_xxz * Atud_zx) + (-Ct_xxy * Atud_yx) + (-Ct_xxx * Atud_xx) + (-Gamt_z * vector(Atd_xz, i, j, k)) + (-Gamt_y * vector(Atd_xy, i, j, k)) + (-Gamt_x * vector(Atd_xx, i, j, k)) + (-Ct_zxz * Atud_zz) + (-Ct_zxy * Atud_yz) + (-Ct_zxx * Atud_xz) + (-Ct_yxz * Atud_zy) + (-Ct_yxy * Atud_yy) + (-Ct_yxx * Atud_xy);
								d_MomCon_y_o0_t13_m0_l0 = (-8.0) * M_PI * Jtd_ADM_y * inv_chi + (-Gamt_z * vector(Atd_yz, i, j, k)) + (-Gamt_y * vector(Atd_yy, i, j, k)) + (-Gamt_x * vector(Atd_xy, i, j, k)) + (-Ct_zyz * Atud_zz) + (-Ct_zyy * Atud_yz) + (-Ct_zxy * Atud_xz) + (-Ct_yyz * Atud_zy) + (-Ct_yyy * Atud_yy) + (-Ct_yxy * Atud_xy) + (-Ct_xyz * Atud_zx) + (-Ct_xyy * Atud_yx) + (-Ct_xxy * Atud_xx);
								d_MomCon_z_o0_t13_m0_l0 = (-Gamt_z * vector(Atd_zz, i, j, k)) + (-Gamt_y * vector(Atd_yz, i, j, k)) + (-Gamt_x * vector(Atd_xz, i, j, k)) + (-Ct_zzz * Atud_zz) + (-Ct_zyz * Atud_yz) + (-Ct_zxz * Atud_xz) + (-Ct_yzz * Atud_zy) + (-Ct_yyz * Atud_yy) + (-Ct_yxz * Atud_xy) + (-Ct_xzz * Atud_zx) + (-Ct_xyz * Atud_yx) + (-Ct_xxz * Atud_xx) + (-8.0) * M_PI * Jtd_ADM_z * inv_chi;
								d_trA_o0_t0_m0_l0 = Atud_xx + Atud_yy + Atud_zz;
								d_detgtm1_o0_t0_m0_l0 = detgtd + (-1.0);
								d_M_ADM_surf_o0_t9_m0_l0 = 0.0198943678864869 * vector(chi, i, j, k) * Gamt_z * dSigma_z + 0.0198943678864869 * vector(chi, i, j, k) * Gamt_y * dSigma_y + 0.0198943678864869 * vector(chi, i, j, k) * Gamt_x * dSigma_x;
								d_Jz_ADM_surf_o0_t0_m0_l0 = 0.0397887357729738 * vector(chi, i, j, k) * (uph_x * (0.3333333333333333 * vector(trK, i, j, k) * gtu_xx + Atu_xx) + uph_y * (0.3333333333333333 * vector(trK, i, j, k) * gtu_xy + Atu_xy) + uph_z * (0.3333333333333333 * vector(trK, i, j, k) * gtu_xz + Atu_xz)) * dSigma_x + 0.0397887357729738 * vector(chi, i, j, k) * (uph_x * (0.3333333333333333 * vector(trK, i, j, k) * gtu_xy + Atu_xy) + uph_y * (0.3333333333333333 * vector(trK, i, j, k) * gtu_yy + Atu_yy) + uph_z * (0.3333333333333333 * vector(trK, i, j, k) * gtu_yz + Atu_yz)) * dSigma_y + 0.0397887357729738 * vector(chi, i, j, k) * (uph_x * (0.3333333333333333 * vector(trK, i, j, k) * gtu_xz + Atu_xz) + uph_y * (0.3333333333333333 * vector(trK, i, j, k) * gtu_yz + Atu_yz) + uph_z * (0.3333333333333333 * vector(trK, i, j, k) * gtu_zz + Atu_zz)) * dSigma_z;
								d_M_Komar_o0_t0_m0_l0 = 2.0 * (((-0.5 * TT * g4d_tt) + Td_tt) / vector(Alpha, i, j, k) + (-(((-0.5 * TT * g4d_tx) + Td_tx) * vector(Betau_x, i, j, k)) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_ty) + Td_ty) * vector(Betau_y, i, j, k)) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_tz) + Td_tz) * vector(Betau_z, i, j, k)) / vector(Alpha, i, j, k))) / (pow(vector(chi, i, j, k), 1.500000000000000));
								d_Jz_Komar_o0_t0_m0_l0 = -((((-0.5 * TT * g4d_tx) + Td_tx) * vph_x) / vector(Alpha, i, j, k) + (-(((-0.5 * TT * g4d_xx) + Td_xx) * vector(Betau_x, i, j, k) * vph_x) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_xy) + Td_xy) * vector(Betau_y, i, j, k) * vph_x) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_xz) + Td_xz) * vector(Betau_z, i, j, k) * vph_x) / vector(Alpha, i, j, k)) + (((-0.5 * TT * g4d_ty) + Td_ty) * vph_y) / vector(Alpha, i, j, k) + (-(((-0.5 * TT * g4d_xy) + Td_xy) * vector(Betau_x, i, j, k) * vph_y) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_yy) + Td_yy) * vector(Betau_y, i, j, k) * vph_y) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_yz) + Td_yz) * vector(Betau_z, i, j, k) * vph_y) / vector(Alpha, i, j, k)) + (((-0.5 * TT * g4d_tz) + Td_tz) * vph_z) / vector(Alpha, i, j, k) + (-(((-0.5 * TT * g4d_xz) + Td_xz) * vector(Betau_x, i, j, k) * vph_z) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_yz) + Td_yz) * vector(Betau_y, i, j, k) * vph_z) / vector(Alpha, i, j, k)) + (-(((-0.5 * TT * g4d_zz) + Td_zz) * vector(Betau_z, i, j, k) * vph_z) / vector(Alpha, i, j, k))) / (pow(vector(chi, i, j, k), 1.500000000000000));
								d_psi4R_o0_t0_m0_l0 = MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * ((-BWeyl_xx * mmI_xx) + EWeyl_xx * mmR_xx + (-BWeyl_yx * mmI_xy) + 2.0 * EWeyl_xy * mmR_xy + (-BWeyl_zx * mmI_xz) + 2.0 * EWeyl_xz * mmR_xz + (-BWeyl_xy * mmI_xy) + (-BWeyl_yy * mmI_yy) + EWeyl_yy * mmR_yy + (-BWeyl_zy * mmI_yz) + 2.0 * EWeyl_yz * mmR_yz + (-BWeyl_xz * mmI_xz) + (-BWeyl_yz * mmI_yz) + (-BWeyl_zz * mmI_zz) + EWeyl_zz * mmR_zz);
								d_psi4I_o0_t0_m0_l0 = MAX(0.0000000001, sqrt(xcoord(i) * xcoord(i) + ycoord(j) * ycoord(j) + zcoord(k) * zcoord(k))) * (BWeyl_xx * mmR_xx + EWeyl_xx * mmI_xx + BWeyl_yx * mmR_xy + 2.0 * EWeyl_xy * mmI_xy + BWeyl_zx * mmR_xz + 2.0 * EWeyl_xz * mmI_xz + BWeyl_xy * mmR_xy + BWeyl_yy * mmR_yy + EWeyl_yy * mmI_yy + BWeyl_zy * mmR_yz + 2.0 * EWeyl_yz * mmI_yz + BWeyl_xz * mmR_xz + BWeyl_yz * mmR_yz + BWeyl_zz * mmR_zz + EWeyl_zz * mmI_zz);
								d_Z_x_o0_t0_m0_l0 = Zu_x;
								d_Z_y_o0_t0_m0_l0 = Zu_y;
								d_Z_z_o0_t0_m0_l0 = Zu_z;
								d_flux_mass_o0_t0_m0_l0 = (Ju_x * xcoord(i) + Ju_y * ycoord(j) + Ju_z * zcoord(k)) * r_pos * SIGN(MAX(vector(Alpha, i, j, k) * hf * sqrt(sqW) - 1.0, 0.0));
								m_M_ADM_surf_o0_t8_l0 = 0.125 * (gtu_zz * dSigma_z) / M_PI * d_Rpd_xx_o0_t2_m0_l0;
								m_M_ADM_surf_o0_t7_l0 = 0.125 * (gtu_yz * dSigma_z) / M_PI * d_Rpd_xx_o0_t1_m0_l0;
								m_M_ADM_surf_o0_t6_l0 = 0.125 * (gtu_xz * dSigma_z) / M_PI * d_Rpd_xx_o0_t0_m0_l0;
								m_M_ADM_surf_o0_t5_l0 = 0.125 * (gtu_yz * dSigma_y) / M_PI * d_Rpd_xx_o0_t2_m0_l0;
								m_M_ADM_surf_o0_t4_l0 = 0.125 * (gtu_yy * dSigma_y) / M_PI * d_Rpd_xx_o0_t1_m0_l0;
								m_M_ADM_surf_o0_t3_l0 = 0.125 * (gtu_xy * dSigma_y) / M_PI * d_Rpd_xx_o0_t0_m0_l0;
								m_M_ADM_surf_o0_t2_l0 = 0.125 * (gtu_xz * dSigma_x) / M_PI * d_Rpd_xx_o0_t2_m0_l0;
								m_M_ADM_surf_o0_t1_l0 = 0.125 * (gtu_xy * dSigma_x) / M_PI * d_Rpd_xx_o0_t1_m0_l0;
								m_M_ADM_surf_o0_t0_l0 = 0.125 * (gtu_xx * dSigma_x) / M_PI * d_Rpd_xx_o0_t0_m0_l0;
								m_MomCon_z_o0_t12_l0 = (-0.6666666666666667) * d_Del_Kd_zxx_o0_t2_m0_l0;
								m_MomCon_z_o0_t11_l0 = gtu_zz * d_MomCon_z_o0_t11_m0_l0;
								m_MomCon_z_o0_t10_l0 = gtu_yz * d_Del_Kd_zyz_o0_t0_m0_l0;
								m_MomCon_z_o0_t9_l0 = gtu_xz * d_Del_Kd_zxz_o0_t0_m0_l0;
								m_MomCon_z_o0_t8_l0 = gtu_yz * d_Del_Kd_yzz_o0_t0_m0_l0;
								m_MomCon_z_o0_t7_l0 = gtu_yy * d_Del_Kd_yyz_o0_t0_m0_l0;
								m_MomCon_z_o0_t6_l0 = gtu_xy * d_Del_Kd_yxz_o0_t0_m0_l0;
								m_MomCon_z_o0_t5_l0 = gtu_xz * d_Del_Kd_xzz_o0_t0_m0_l0;
								m_MomCon_z_o0_t4_l0 = gtu_xy * d_Del_Kd_xyz_o0_t0_m0_l0;
								m_MomCon_z_o0_t3_l0 = gtu_xx * d_Del_Kd_xxz_o0_t0_m0_l0;
								m_MomCon_z_o0_t2_l0 = (-1.500000000000000 * inv_chi * Atud_zz) * d_Rpd_xx_o0_t2_m0_l0;
								m_MomCon_z_o0_t1_l0 = (-1.500000000000000 * inv_chi * Atud_yz) * d_Rpd_xx_o0_t1_m0_l0;
								m_MomCon_z_o0_t0_l0 = (-1.500000000000000 * inv_chi * Atud_xz) * d_Rpd_xx_o0_t0_m0_l0;
								m_MomCon_y_o0_t12_l0 = (-0.6666666666666667) * d_Del_Kd_yxx_o0_t2_m0_l0;
								m_MomCon_y_o0_t11_l0 = (-1.500000000000000 * inv_chi * Atud_zy) * d_Rpd_xx_o0_t2_m0_l0;
								m_MomCon_y_o0_t10_l0 = (-1.500000000000000 * inv_chi * Atud_yy) * d_Rpd_xx_o0_t1_m0_l0;
								m_MomCon_y_o0_t9_l0 = (-1.500000000000000 * inv_chi * Atud_xy) * d_Rpd_xx_o0_t0_m0_l0;
								m_MomCon_y_o0_t8_l0 = gtu_zz * d_Del_Kd_zyz_o0_t0_m0_l0;
								m_MomCon_y_o0_t7_l0 = gtu_yz * d_Del_Kd_zyy_o0_t0_m0_l0;
								m_MomCon_y_o0_t6_l0 = gtu_xz * d_Del_Kd_zxy_o0_t0_m0_l0;
								m_MomCon_y_o0_t5_l0 = gtu_yz * d_Del_Kd_yyz_o0_t0_m0_l0;
								m_MomCon_y_o0_t4_l0 = gtu_yy * d_MomCon_y_o0_t4_m0_l0;
								m_MomCon_y_o0_t3_l0 = gtu_xy * d_Del_Kd_yxy_o0_t0_m0_l0;
								m_MomCon_y_o0_t2_l0 = gtu_xz * d_Del_Kd_xyz_o0_t0_m0_l0;
								m_MomCon_y_o0_t1_l0 = gtu_xy * d_Del_Kd_xyy_o0_t0_m0_l0;
								m_MomCon_y_o0_t0_l0 = gtu_xx * d_Del_Kd_xxy_o0_t0_m0_l0;
								m_MomCon_x_o0_t12_l0 = (-0.6666666666666667) * d_Del_Kd_xxy_o0_t2_m0_l0;
								m_MomCon_x_o0_t11_l0 = (-1.500000000000000 * inv_chi * Atud_yx) * d_Rpd_xx_o0_t1_m0_l0;
								m_MomCon_x_o0_t10_l0 = (-1.500000000000000 * inv_chi * Atud_xx) * d_Rpd_xx_o0_t0_m0_l0;
								m_MomCon_x_o0_t9_l0 = (-1.500000000000000 * inv_chi * Atud_zx) * d_Rpd_xx_o0_t2_m0_l0;
								m_MomCon_x_o0_t8_l0 = gtu_zz * d_Del_Kd_zxz_o0_t0_m0_l0;
								m_MomCon_x_o0_t7_l0 = gtu_yz * d_Del_Kd_zxy_o0_t0_m0_l0;
								m_MomCon_x_o0_t6_l0 = gtu_xz * d_Del_Kd_zxx_o0_t0_m0_l0;
								m_MomCon_x_o0_t5_l0 = gtu_yz * d_Del_Kd_yxz_o0_t0_m0_l0;
								m_MomCon_x_o0_t4_l0 = gtu_yy * d_Del_Kd_yxy_o0_t0_m0_l0;
								m_MomCon_x_o0_t3_l0 = gtu_xy * d_Del_Kd_yxx_o0_t0_m0_l0;
								m_MomCon_x_o0_t2_l0 = gtu_xz * d_Del_Kd_xxz_o0_t0_m0_l0;
								m_MomCon_x_o0_t1_l0 = gtu_xy * d_Del_Kd_xxy_o0_t0_m0_l0;
								m_MomCon_x_o0_t0_l0 = gtu_xx * d_MomCon_x_o0_t0_m0_l0;
								vector(Rscalar, i, j, k) = d_Rscalar_o0_t0_m0_l0;
								vector(HamCon, i, j, k) = d_HamCon_o0_t0_m0_l0;
								vector(MomCon_x, i, j, k) = ((((((((((((m_MomCon_x_o0_t0_l0 + m_MomCon_x_o0_t1_l0) + m_MomCon_x_o0_t2_l0) + m_MomCon_x_o0_t3_l0) + m_MomCon_x_o0_t4_l0) + m_MomCon_x_o0_t5_l0) + m_MomCon_x_o0_t6_l0) + m_MomCon_x_o0_t7_l0) + m_MomCon_x_o0_t8_l0) + m_MomCon_x_o0_t9_l0) + m_MomCon_x_o0_t10_l0) + m_MomCon_x_o0_t11_l0) + m_MomCon_x_o0_t12_l0) + d_MomCon_x_o0_t13_m0_l0;
								vector(MomCon_y, i, j, k) = ((((((((((((m_MomCon_y_o0_t0_l0 + m_MomCon_y_o0_t1_l0) + m_MomCon_y_o0_t2_l0) + m_MomCon_y_o0_t3_l0) + m_MomCon_y_o0_t4_l0) + m_MomCon_y_o0_t5_l0) + m_MomCon_y_o0_t6_l0) + m_MomCon_y_o0_t7_l0) + m_MomCon_y_o0_t8_l0) + m_MomCon_y_o0_t9_l0) + m_MomCon_y_o0_t10_l0) + m_MomCon_y_o0_t11_l0) + m_MomCon_y_o0_t12_l0) + d_MomCon_y_o0_t13_m0_l0;
								vector(MomCon_z, i, j, k) = ((((((((((((m_MomCon_z_o0_t0_l0 + m_MomCon_z_o0_t1_l0) + m_MomCon_z_o0_t2_l0) + m_MomCon_z_o0_t3_l0) + m_MomCon_z_o0_t4_l0) + m_MomCon_z_o0_t5_l0) + m_MomCon_z_o0_t6_l0) + m_MomCon_z_o0_t7_l0) + m_MomCon_z_o0_t8_l0) + m_MomCon_z_o0_t9_l0) + m_MomCon_z_o0_t10_l0) + m_MomCon_z_o0_t11_l0) + m_MomCon_z_o0_t12_l0) + d_MomCon_z_o0_t13_m0_l0;
								vector(trA, i, j, k) = d_trA_o0_t0_m0_l0;
								vector(detgtm1, i, j, k) = d_detgtm1_o0_t0_m0_l0;
								vector(M_ADM_surf, i, j, k) = ((((((((m_M_ADM_surf_o0_t0_l0 + m_M_ADM_surf_o0_t1_l0) + m_M_ADM_surf_o0_t2_l0) + m_M_ADM_surf_o0_t3_l0) + m_M_ADM_surf_o0_t4_l0) + m_M_ADM_surf_o0_t5_l0) + m_M_ADM_surf_o0_t6_l0) + m_M_ADM_surf_o0_t7_l0) + m_M_ADM_surf_o0_t8_l0) + d_M_ADM_surf_o0_t9_m0_l0;
								vector(Jz_ADM_surf, i, j, k) = d_Jz_ADM_surf_o0_t0_m0_l0;
								vector(M_Komar, i, j, k) = d_M_Komar_o0_t0_m0_l0;
								vector(Jz_Komar, i, j, k) = d_Jz_Komar_o0_t0_m0_l0;
								vector(psi4R, i, j, k) = d_psi4R_o0_t0_m0_l0;
								vector(psi4I, i, j, k) = d_psi4I_o0_t0_m0_l0;
								vector(Z_x, i, j, k) = d_Z_x_o0_t0_m0_l0;
								vector(Z_y, i, j, k) = d_Z_y_o0_t0_m0_l0;
								vector(Z_z, i, j, k) = d_Z_z_o0_t0_m0_l0;
								vector(flux_mass, i, j, k) = d_flux_mass_o0_t0_m0_l0;
							}
						}
					}
				}
			}
		}
		

	}


	//Output
	t_output->start();
	if (ln == hierarchy->getFinestLevelNumber() && viz_mesh_dump_interval > 0) {
		if (previous_iteration < next_mesh_dump_iteration && outputCycle >= next_mesh_dump_iteration) {
			d_visit_data_writer->writePlotData(hierarchy, outputCycle, new_time);
			next_mesh_dump_iteration = next_mesh_dump_iteration + viz_mesh_dump_interval;
			while (outputCycle >= next_mesh_dump_iteration) {
				next_mesh_dump_iteration = next_mesh_dump_iteration + viz_mesh_dump_interval;
			}
		}
	}
	//Slicer output
	if (ln == hierarchy->getFinestLevelNumber() && d_slicer_output_period.size() > 0) {
		int i = 0;
		for (std::vector<std::shared_ptr<SlicerDataWriter> >::iterator it = d_sliceWriters.begin(); it != d_sliceWriters.end(); ++it) {
			if (d_slicer_output_period[i] > 0 && previous_iteration < next_slice_dump_iteration[i] && outputCycle >= next_slice_dump_iteration[i]) {
				(*it)->writePlotData(hierarchy, outputCycle, new_time);
				next_slice_dump_iteration[i] = next_slice_dump_iteration[i] + d_slicer_output_period[i];
				while (outputCycle >= next_slice_dump_iteration[i]) {
					next_slice_dump_iteration[i] = next_slice_dump_iteration[i] + d_slicer_output_period[i];
				}
			}
			i++;
		}
	}

	//Spherical output
	if (ln == hierarchy->getFinestLevelNumber() && d_sphere_output_period.size() > 0) {
		int i = 0;
		for (std::vector<std::shared_ptr<SphereDataWriter> >::iterator it = d_sphereWriters.begin(); it != d_sphereWriters.end(); ++it) {
			if (d_sphere_output_period[i] > 0 && previous_iteration < next_sphere_dump_iteration[i] && outputCycle >= next_sphere_dump_iteration[i]) {
				(*it)->writePlotData(hierarchy, outputCycle, new_time);
				next_sphere_dump_iteration[i] = next_sphere_dump_iteration[i] + d_sphere_output_period[i];
				while (outputCycle >= next_sphere_dump_iteration[i]) {
					next_sphere_dump_iteration[i] = next_sphere_dump_iteration[i] + d_sphere_output_period[i];
				}
			}
			i++;
		}
	}

	//Integration output
	if (ln == hierarchy->getFinestLevelNumber() && d_integration_output_period.size() > 0) {
		int i = 0;
		for (std::vector<std::shared_ptr<IntegrateDataWriter> >::iterator it = d_integrateDataWriters.begin(); it != d_integrateDataWriters.end(); ++it) {
			if (d_integration_output_period[i] > 0 && previous_iteration < next_integration_dump_iteration[i] && outputCycle >= next_integration_dump_iteration[i]) {
				(*it)->writePlotData(hierarchy, outputCycle, new_time);
				next_integration_dump_iteration[i] = next_integration_dump_iteration[i] + d_integration_output_period[i];
				while (outputCycle >= next_integration_dump_iteration[i]) {
					next_integration_dump_iteration[i] = next_integration_dump_iteration[i] + d_integration_output_period[i];
				}
			}
			i++;
		}
	}

	//Point output
	if (ln == hierarchy->getFinestLevelNumber() && d_point_output_period.size() > 0) {
		int i = 0;
		for (std::vector<std::shared_ptr<PointDataWriter> >::iterator it = d_pointDataWriters.begin(); it != d_pointDataWriters.end(); ++it) {
			if (d_point_output_period[i] > 0 && previous_iteration < next_point_dump_iteration[i] && outputCycle >= next_point_dump_iteration[i]) {
				(*it)->writePlotData(hierarchy, outputCycle, new_time);
				next_point_dump_iteration[i] = next_point_dump_iteration[i] + d_point_output_period[i];
				while (outputCycle >= next_point_dump_iteration[i]) {
					next_point_dump_iteration[i] = next_point_dump_iteration[i] + d_point_output_period[i];
				}
			}
			i++;
		}
	}


	t_output->stop();

	if (mpi.getRank() == 0 && d_output_interval > 0 ) {
		if (previous_iteration < next_console_output && outputCycle >= next_console_output) {
			int currentLevelNumber = ln;
			while (currentLevelNumber > 0) {
				currentLevelNumber--;
				cout <<"  ";
			}

			cout << "Level "<<ln<<". Iteration " << current_iteration[ln]<<". Time "<<current_time<<"."<< endl;
			if (ln == hierarchy->getFinestLevelNumber()) {
				next_console_output = next_console_output + d_output_interval;
				while (outputCycle >= next_console_output) {
					next_console_output = next_console_output + d_output_interval;
				}
			
			}
		}
	}

	if (d_timer_output_interval > 0 ) {
		if (previous_iteration < next_timer_output && outputCycle >= next_timer_output) {
			if (ln == hierarchy->getFinestLevelNumber()) {
				//Print timers
				if (mpi.getRank() == 0) {
					tbox::TimerManager::getManager()->print(cout);
				}
				else {
					if (ln == hierarchy->getFinestLevelNumber()) {
						//Dispose other processor timers
						//SAMRAI needs all processors run tbox::TimerManager::getManager()->print, otherwise it hungs
						std::ofstream ofs;
						ofs.setstate(std::ios_base::badbit);
						tbox::TimerManager::getManager()->print(ofs);
					}
				}
				next_timer_output = next_timer_output + d_timer_output_interval;
				while (outputCycle >= next_timer_output) {
					next_timer_output = next_timer_output + d_timer_output_interval;
				}
			
			}
		}
	}

	return simPlat_dt;
}

/*
 * Checks the finalization conditions              
 */
bool Problem::checkFinalization(const double current_time, const double simPlat_dt)
{
	if (greaterEq(current_time, tend)) { 
		return true;
	}
	return false;
	
	

}

void Problem::putToRestart(MainRestartData& mrd) {
	mrd.setNextSliceDumpIteration(next_slice_dump_iteration);
	mrd.setNextSphereDumpIteration(next_sphere_dump_iteration);
	mrd.setNextMeshDumpIteration(next_mesh_dump_iteration);
	mrd.setNextIntegrationDumpIteration(next_integration_dump_iteration);
	mrd.setNextPointDumpIteration(next_point_dump_iteration);

	mrd.setCurrentIteration(current_iteration);
	mrd.setNextConsoleOutputIteration(next_console_output);
	mrd.setNextTimerOutputIteration(next_timer_output);
}

void Problem::getFromRestart(MainRestartData& mrd) {
	next_slice_dump_iteration = mrd.getNextSliceDumpIteration();
	next_sphere_dump_iteration = mrd.getNextSphereDumpIteration();
	next_mesh_dump_iteration = mrd.getNextMeshDumpIteration();
	next_integration_dump_iteration = mrd.getNextIntegrationDumpIteration();
	next_point_dump_iteration = mrd.getNextPointDumpIteration();

	current_iteration = mrd.getCurrentIteration();
	next_console_output = mrd.getNextConsoleOutputIteration();
	next_timer_output = mrd.getNextTimerOutputIteration();
}

void Problem::allocateAfterRestart() {
	for (int il = 0; il < d_patch_hierarchy->getNumberOfLevels(); il++) {
		std::shared_ptr< hier::PatchLevel > level(d_patch_hierarchy->getPatchLevel(il));
		level->allocatePatchData(d_mask_id);
		level->allocatePatchData(d_interior_regridding_value_id);
		level->allocatePatchData(d_nonSync_regridding_tag_id);
		level->allocatePatchData(d_interior_i_id);
		level->allocatePatchData(d_interior_j_id);
		level->allocatePatchData(d_interior_k_id);
		level->allocatePatchData(d_Rscalar_id);
		level->allocatePatchData(d_HamCon_id);
		level->allocatePatchData(d_MomCon_x_id);
		level->allocatePatchData(d_MomCon_y_id);
		level->allocatePatchData(d_MomCon_z_id);
		level->allocatePatchData(d_trA_id);
		level->allocatePatchData(d_detgtm1_id);
		level->allocatePatchData(d_M_ADM_surf_id);
		level->allocatePatchData(d_Jz_ADM_surf_id);
		level->allocatePatchData(d_M_Komar_id);
		level->allocatePatchData(d_Jz_Komar_id);
		level->allocatePatchData(d_psi4R_id);
		level->allocatePatchData(d_psi4I_id);
		level->allocatePatchData(d_Z_x_id);
		level->allocatePatchData(d_Z_y_id);
		level->allocatePatchData(d_Z_z_id);
		level->allocatePatchData(d_flux_mass_id);
		level->allocatePatchData(d_rk1Sfd_x_id);
		level->allocatePatchData(d_rk1Sfd_y_id);
		level->allocatePatchData(d_rk1Sfd_z_id);
		level->allocatePatchData(d_rk1Bfu_x_id);
		level->allocatePatchData(d_rk1Bfu_y_id);
		level->allocatePatchData(d_rk1Bfu_z_id);
		level->allocatePatchData(d_rk1Df_id);
		level->allocatePatchData(d_rk1DYf_id);
		level->allocatePatchData(d_rk1tauf_id);
		level->allocatePatchData(d_rk1phif_id);
		level->allocatePatchData(d_rk1gtd_xx_id);
		level->allocatePatchData(d_rk1gtd_xy_id);
		level->allocatePatchData(d_rk1gtd_xz_id);
		level->allocatePatchData(d_rk1gtd_yy_id);
		level->allocatePatchData(d_rk1gtd_yz_id);
		level->allocatePatchData(d_rk1gtd_zz_id);
		level->allocatePatchData(d_rk1Atd_xx_id);
		level->allocatePatchData(d_rk1Atd_xy_id);
		level->allocatePatchData(d_rk1Atd_xz_id);
		level->allocatePatchData(d_rk1Atd_yy_id);
		level->allocatePatchData(d_rk1Atd_yz_id);
		level->allocatePatchData(d_rk1Atd_zz_id);
		level->allocatePatchData(d_rk1Gamh_x_id);
		level->allocatePatchData(d_rk1Gamh_y_id);
		level->allocatePatchData(d_rk1Gamh_z_id);
		level->allocatePatchData(d_rk1Betau_x_id);
		level->allocatePatchData(d_rk1Betau_y_id);
		level->allocatePatchData(d_rk1Betau_z_id);
		level->allocatePatchData(d_rk1Alpha_id);
		level->allocatePatchData(d_rk1chi_id);
		level->allocatePatchData(d_rk1trK_id);
		level->allocatePatchData(d_rk1theta_id);
		level->allocatePatchData(d_rk2Sfd_x_id);
		level->allocatePatchData(d_rk2Sfd_y_id);
		level->allocatePatchData(d_rk2Sfd_z_id);
		level->allocatePatchData(d_rk2Bfu_x_id);
		level->allocatePatchData(d_rk2Bfu_y_id);
		level->allocatePatchData(d_rk2Bfu_z_id);
		level->allocatePatchData(d_rk2Df_id);
		level->allocatePatchData(d_rk2DYf_id);
		level->allocatePatchData(d_rk2tauf_id);
		level->allocatePatchData(d_rk2phif_id);
		level->allocatePatchData(d_rk2gtd_xx_id);
		level->allocatePatchData(d_rk2gtd_xy_id);
		level->allocatePatchData(d_rk2gtd_xz_id);
		level->allocatePatchData(d_rk2gtd_yy_id);
		level->allocatePatchData(d_rk2gtd_yz_id);
		level->allocatePatchData(d_rk2gtd_zz_id);
		level->allocatePatchData(d_rk2Atd_xx_id);
		level->allocatePatchData(d_rk2Atd_xy_id);
		level->allocatePatchData(d_rk2Atd_xz_id);
		level->allocatePatchData(d_rk2Atd_yy_id);
		level->allocatePatchData(d_rk2Atd_yz_id);
		level->allocatePatchData(d_rk2Atd_zz_id);
		level->allocatePatchData(d_rk2Gamh_x_id);
		level->allocatePatchData(d_rk2Gamh_y_id);
		level->allocatePatchData(d_rk2Gamh_z_id);
		level->allocatePatchData(d_rk2Betau_x_id);
		level->allocatePatchData(d_rk2Betau_y_id);
		level->allocatePatchData(d_rk2Betau_z_id);
		level->allocatePatchData(d_rk2Alpha_id);
		level->allocatePatchData(d_rk2chi_id);
		level->allocatePatchData(d_rk2trK_id);
		level->allocatePatchData(d_rk2theta_id);
		level->allocatePatchData(d_rk3Sfd_x_id);
		level->allocatePatchData(d_rk3Sfd_y_id);
		level->allocatePatchData(d_rk3Sfd_z_id);
		level->allocatePatchData(d_rk3Bfu_x_id);
		level->allocatePatchData(d_rk3Bfu_y_id);
		level->allocatePatchData(d_rk3Bfu_z_id);
		level->allocatePatchData(d_rk3Df_id);
		level->allocatePatchData(d_rk3DYf_id);
		level->allocatePatchData(d_rk3tauf_id);
		level->allocatePatchData(d_rk3phif_id);
		level->allocatePatchData(d_rk3gtd_xx_id);
		level->allocatePatchData(d_rk3gtd_xy_id);
		level->allocatePatchData(d_rk3gtd_xz_id);
		level->allocatePatchData(d_rk3gtd_yy_id);
		level->allocatePatchData(d_rk3gtd_yz_id);
		level->allocatePatchData(d_rk3gtd_zz_id);
		level->allocatePatchData(d_rk3Atd_xx_id);
		level->allocatePatchData(d_rk3Atd_xy_id);
		level->allocatePatchData(d_rk3Atd_xz_id);
		level->allocatePatchData(d_rk3Atd_yy_id);
		level->allocatePatchData(d_rk3Atd_yz_id);
		level->allocatePatchData(d_rk3Atd_zz_id);
		level->allocatePatchData(d_rk3Gamh_x_id);
		level->allocatePatchData(d_rk3Gamh_y_id);
		level->allocatePatchData(d_rk3Gamh_z_id);
		level->allocatePatchData(d_rk3Betau_x_id);
		level->allocatePatchData(d_rk3Betau_y_id);
		level->allocatePatchData(d_rk3Betau_z_id);
		level->allocatePatchData(d_rk3Alpha_id);
		level->allocatePatchData(d_rk3chi_id);
		level->allocatePatchData(d_rk3trK_id);
		level->allocatePatchData(d_rk3theta_id);
		level->allocatePatchData(d_Sfd_x_p_id);
		level->allocatePatchData(d_Sfd_y_p_id);
		level->allocatePatchData(d_Sfd_z_p_id);
		level->allocatePatchData(d_Bfu_x_p_id);
		level->allocatePatchData(d_Bfu_y_p_id);
		level->allocatePatchData(d_Bfu_z_p_id);
		level->allocatePatchData(d_Df_p_id);
		level->allocatePatchData(d_DYf_p_id);
		level->allocatePatchData(d_tauf_p_id);
		level->allocatePatchData(d_phif_p_id);
		level->allocatePatchData(d_gtd_xx_p_id);
		level->allocatePatchData(d_gtd_xy_p_id);
		level->allocatePatchData(d_gtd_xz_p_id);
		level->allocatePatchData(d_gtd_yy_p_id);
		level->allocatePatchData(d_gtd_yz_p_id);
		level->allocatePatchData(d_gtd_zz_p_id);
		level->allocatePatchData(d_Atd_xx_p_id);
		level->allocatePatchData(d_Atd_xy_p_id);
		level->allocatePatchData(d_Atd_xz_p_id);
		level->allocatePatchData(d_Atd_yy_p_id);
		level->allocatePatchData(d_Atd_yz_p_id);
		level->allocatePatchData(d_Atd_zz_p_id);
		level->allocatePatchData(d_Gamh_x_p_id);
		level->allocatePatchData(d_Gamh_y_p_id);
		level->allocatePatchData(d_Gamh_z_p_id);
		level->allocatePatchData(d_Betau_x_p_id);
		level->allocatePatchData(d_Betau_y_p_id);
		level->allocatePatchData(d_Betau_z_p_id);
		level->allocatePatchData(d_Alpha_p_id);
		level->allocatePatchData(d_chi_p_id);
		level->allocatePatchData(d_trK_p_id);
		level->allocatePatchData(d_theta_p_id);
	}

}

/*
 *  Cell tagging routine - tag cells that require refinement based on a provided condition. 
 */
void Problem::applyGradientDetector(
   const std::shared_ptr< hier::PatchHierarchy >& hierarchy, 
   const int level_number,
   const double time, 
   const int tag_index,
   const bool initial_time,
   const bool uses_richardson_extrapolation_too) 
{
	if (d_regridding && level_number + 1 >= d_regridding_min_level && level_number + 1 <= d_regridding_max_level) {
		hier::VariableDatabase *vdb = hier::VariableDatabase::getDatabase();
		if (!(vdb->checkVariableExists(d_regridding_field))) {
			TBOX_ERROR(d_object_name << ": Regridding field selected not found:n"					<<  d_regridding_field<<  "n");
		}

		std::shared_ptr< hier::PatchLevel > level(hierarchy->getPatchLevel(level_number));

		for (hier::PatchLevel::iterator ip(level->begin()); ip != level->end(); ++ip) {
			const std::shared_ptr< hier::Patch >& patch = *ip;
			int* tags = ((pdat::CellData<int> *) patch->getPatchData(tag_index).get())->getPointer();
			int* regridding_tag = ((pdat::NodeData<int> *) patch->getPatchData(d_nonSync_regridding_tag_id).get())->getPointer();
			const std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
			const double* dx  = patch_geom->getDx();
			const hier::Index tfirst = patch->getPatchData(tag_index)->getGhostBox().lower();
			const hier::Index tlast  = patch->getPatchData(tag_index)->getGhostBox().upper();
			const hier::Index boxfirst = patch->getBox().lower();
			const hier::Index boxlast  = patch->getBox().upper();
			int ilast = boxlast(0)-boxfirst(0)+2+2*d_ghost_width;
			int itlast = tlast(0)-tfirst(0)+1;
			int jlast = boxlast(1)-boxfirst(1)+2+2*d_ghost_width;
			int jtlast = tlast(1)-tfirst(1)+1;
			int klast = boxlast(2)-boxfirst(2)+2+2*d_ghost_width;
			int ktlast = tlast(2)-tfirst(2)+1;
			for(int index2 = 0; index2 < (tlast(2)-tfirst(2))+1; index2++) {
				for(int index1 = 0; index1 < (tlast(1)-tfirst(1))+1; index1++) {
					for(int index0 = 0; index0 < (tlast(0)-tfirst(0))+1; index0++) {
						vectorT(tags,index0, index1, index2) = 0;
					}
				}
			}
			for(int index2 = 0; index2 < klast; index2++) {
				for(int index1 = 0; index1 < jlast; index1++) {
					for(int index0 = 0; index0 < ilast; index0++) {
						vector(regridding_tag,index0, index1, index2) = 0;
					}
				}
			}
			if (vdb->checkVariableExists(d_regridding_field)) {
				//Mesh
				if (d_regridding_type == "GRADIENT") {
					int regrid_field_id = vdb->getVariable(d_regridding_field)->getInstanceIdentifier();
					double* regrid_field = ((pdat::NodeData<double> *) patch->getPatchData(regrid_field_id).get())->getPointer();
					double* regridding_value = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_regridding_value_id).get())->getPointer();
					for(int index2 = 0; index2 < klast - 1; index2++) {
						for(int index1 = 0; index1 < jlast - 1; index1++) {
							for(int index0 = 0; index0 < ilast - 1; index0++) {
								if (vector(regrid_field, index0, index1, index2)!=0) {
									if ((fabs(vector(regrid_field, index0+1+d_ghost_width, index1+d_ghost_width, index2+d_ghost_width)-vector(regrid_field, index0+d_ghost_width, index1+d_ghost_width, index2+d_ghost_width)) > d_regridding_compressionFactor * MIN(fabs(vector(regrid_field, index0+2+d_ghost_width, index1+d_ghost_width, index2+d_ghost_width)-vector(regrid_field, index0+1+d_ghost_width, index1+d_ghost_width, index2+d_ghost_width)) + d_regridding_mOffset * pow(dx[0], 2), fabs(vector(regrid_field, index0+d_ghost_width, index1+d_ghost_width, index2+d_ghost_width)-vector(regrid_field, index0-1+d_ghost_width, index1+d_ghost_width, index2+d_ghost_width)) + d_regridding_mOffset * pow(dx[0], 2)) ) ||(fabs(vector(regrid_field, index0+d_ghost_width, index1+1+d_ghost_width, index2+d_ghost_width)-vector(regrid_field, index0+d_ghost_width, index1+d_ghost_width, index2+d_ghost_width)) > d_regridding_compressionFactor * MIN(fabs(vector(regrid_field, index0+d_ghost_width, index1+2+d_ghost_width, index2+d_ghost_width)-vector(regrid_field, index0+d_ghost_width, index1+1+d_ghost_width, index2+d_ghost_width)) + d_regridding_mOffset * pow(dx[1], 2), fabs(vector(regrid_field, index0+d_ghost_width, index1+d_ghost_width, index2+d_ghost_width)-vector(regrid_field, index0+d_ghost_width, index1-1+d_ghost_width, index2+d_ghost_width)) + d_regridding_mOffset * pow(dx[1], 2)) ) ||(fabs(vector(regrid_field, index0+d_ghost_width, index1+d_ghost_width, index2+1+d_ghost_width)-vector(regrid_field, index0+d_ghost_width, index1+d_ghost_width, index2+d_ghost_width)) > d_regridding_compressionFactor * MIN(fabs(vector(regrid_field, index0+d_ghost_width, index1+d_ghost_width, index2+2+d_ghost_width)-vector(regrid_field, index0+d_ghost_width, index1+d_ghost_width, index2+1+d_ghost_width)) + d_regridding_mOffset * pow(dx[2], 2), fabs(vector(regrid_field, index0+d_ghost_width, index1+d_ghost_width, index2+d_ghost_width)-vector(regrid_field, index0+d_ghost_width, index1+d_ghost_width, index2-1+d_ghost_width)) + d_regridding_mOffset * pow(dx[2], 2)) ) ) {
										if (index0 >= d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 >= d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 >= d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags,index0 - d_ghost_width, index1 - d_ghost_width, index2 - d_ghost_width) = 1;
										}
										vector(regridding_tag,index0, index1, index2) = 1;
										//SAMRAI tagging
										if (index0 > d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 > d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 > d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width - 1, index1-d_ghost_width - 1, index2-d_ghost_width - 1) = 1;
										}
										if (index0 >= d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 > d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 > d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width, index1-d_ghost_width - 1, index2-d_ghost_width - 1) = 1;
										}
										if (index0 > d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 >= d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 > d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width - 1, index1-d_ghost_width, index2-d_ghost_width - 1) = 1;
										}
										if (index0 >= d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 >= d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 > d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width, index1-d_ghost_width, index2-d_ghost_width - 1) = 1;
										}
										if (index0 > d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 > d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 >= d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width - 1, index1-d_ghost_width - 1, index2-d_ghost_width) = 1;
										}
										if (index0 > d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 >= d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 >= d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width - 1, index1-d_ghost_width, index2-d_ghost_width) = 1;
										}
										if (index0 >= d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 > d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 >= d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width, index1-d_ghost_width - 1, index2-d_ghost_width) = 1;
										}
										//Informative tagging
										if (index0 > 0 && index1 > 0 && index2 > 0) {
											if (vector(regridding_tag,index0-1,index1-1,index2-1) != 1)
												vector(regridding_tag,index0-1,index1-1,index2-1) = 1;
										}
										if (index1 > 0 && index2 > 0) {
											if (vector(regridding_tag,index0,index1-1,index2-1) != 1)
												vector(regridding_tag,index0,index1-1,index2-1) = 1;
										}
										if (index0 > 0 && index2 > 0) {
											if (vector(regridding_tag,index0-1,index1,index2-1) != 1)
												vector(regridding_tag,index0-1,index1,index2-1) = 1;
										}
										if (index2 > 0) {
											if (vector(regridding_tag,index0,index1,index2-1) != 1)
												vector(regridding_tag,index0,index1,index2-1) = 1;
										}
										if (index0 > 0 && index1 > 0) {
											if (vector(regridding_tag,index0-1,index1-1,index2) != 1)
												vector(regridding_tag,index0-1,index1-1,index2) = 1;
										}
										if (index0 > 0) {
											if (vector(regridding_tag,index0-1,index1,index2) != 1)
												vector(regridding_tag,index0-1,index1,index2) = 1;
										}
										if (index1 > 0) {
											if (vector(regridding_tag,index0,index1-1,index2) != 1)
												vector(regridding_tag,index0,index1-1,index2) = 1;
										}
										if (d_regridding_buffer > 0) {
											int distance;
											for(int index2b = MAX(0, index2 - d_regridding_buffer - 1); index2b < MIN(index2 + d_regridding_buffer + 1, klast); index2b++) {
												for(int index1b = MAX(0, index1 - d_regridding_buffer - 1); index1b < MIN(index1 + d_regridding_buffer + 1, jlast); index1b++) {
													for(int index0b = MAX(0, index0 - d_regridding_buffer - 1); index0b < MIN(index0 + d_regridding_buffer + 1, ilast); index0b++) {
														int distx = (index0b - index0);
														if (distx < 0) {
															distx++;
														}
														int disty = (index1b - index1);
														if (disty < 0) {
															disty++;
														}
														int distz = (index2b - index2);
														if (distz < 0) {
															distz++;
														}
														distance = 1 + MAX(MAX(abs(distx), abs(disty)), abs(distz));
														if (index0b >= d_ghost_width && index0b < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1b >= d_ghost_width && index1b < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2b >= d_ghost_width && index2b < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
															vectorT(tags,index0b-d_ghost_width,index1b-d_ghost_width,index2b-d_ghost_width) = 1;
														}
														if (vector(regridding_tag,index0b,index1b,index2b) == 0 || vector(regridding_tag,index0b,index1b,index2b) > distance) {
															vector(regridding_tag,index0b,index1b,index2b) = distance;
														}
													}
												}
											}
										}
			
									}
								}
							}
						}
					}
		
				} else {
					if (d_regridding_type == "FUNCTION") {
						int regrid_field_id = vdb->getVariable(d_regridding_field)->getInstanceIdentifier();
						double* regrid_field = ((pdat::NodeData<double> *) patch->getPatchData(regrid_field_id).get())->getPointer();
						double* regridding_value = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_regridding_value_id).get())->getPointer();
						for(int index2 = 0; index2 < klast; index2++) {
							for(int index1 = 0; index1 < jlast; index1++) {
								for(int index0 = 0; index0 < ilast; index0++) {
									vector(regridding_value, index0, index1, index2) = vector(regrid_field, index0, index1, index2);
									if (vector(regrid_field, index0, index1, index2) > d_regridding_threshold) {
										if (index0 >= d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 >= d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 >= d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags,index0 - d_ghost_width, index1 - d_ghost_width, index2 - d_ghost_width) = 1;
										}
										vector(regridding_tag,index0, index1, index2) = 1;
										//SAMRAI tagging
										if (index0 > d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 > d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 > d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width - 1, index1-d_ghost_width - 1, index2-d_ghost_width - 1) = 1;
										}
										if (index0 >= d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 > d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 > d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width, index1-d_ghost_width - 1, index2-d_ghost_width - 1) = 1;
										}
										if (index0 > d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 >= d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 > d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width - 1, index1-d_ghost_width, index2-d_ghost_width - 1) = 1;
										}
										if (index0 >= d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 >= d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 > d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width, index1-d_ghost_width, index2-d_ghost_width - 1) = 1;
										}
										if (index0 > d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 > d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 >= d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width - 1, index1-d_ghost_width - 1, index2-d_ghost_width) = 1;
										}
										if (index0 > d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 >= d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 >= d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width - 1, index1-d_ghost_width, index2-d_ghost_width) = 1;
										}
										if (index0 >= d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 > d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 >= d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
											vectorT(tags, index0-d_ghost_width, index1-d_ghost_width - 1, index2-d_ghost_width) = 1;
										}
										//Informative tagging
										if (index0 > 0 && index1 > 0 && index2 > 0) {
											if (vector(regridding_tag,index0-1,index1-1,index2-1) != 1)
												vector(regridding_tag,index0-1,index1-1,index2-1) = 1;
										}
										if (index1 > 0 && index2 > 0) {
											if (vector(regridding_tag,index0,index1-1,index2-1) != 1)
												vector(regridding_tag,index0,index1-1,index2-1) = 1;
										}
										if (index0 > 0 && index2 > 0) {
											if (vector(regridding_tag,index0-1,index1,index2-1) != 1)
												vector(regridding_tag,index0-1,index1,index2-1) = 1;
										}
										if (index2 > 0) {
											if (vector(regridding_tag,index0,index1,index2-1) != 1)
												vector(regridding_tag,index0,index1,index2-1) = 1;
										}
										if (index0 > 0 && index1 > 0) {
											if (vector(regridding_tag,index0-1,index1-1,index2) != 1)
												vector(regridding_tag,index0-1,index1-1,index2) = 1;
										}
										if (index0 > 0) {
											if (vector(regridding_tag,index0-1,index1,index2) != 1)
												vector(regridding_tag,index0-1,index1,index2) = 1;
										}
										if (index1 > 0) {
											if (vector(regridding_tag,index0,index1-1,index2) != 1)
												vector(regridding_tag,index0,index1-1,index2) = 1;
										}
										if (d_regridding_buffer > 0) {
											int distance;
											for(int index2b = MAX(0, index2 - d_regridding_buffer - 1); index2b < MIN(index2 + d_regridding_buffer + 1, klast); index2b++) {
												for(int index1b = MAX(0, index1 - d_regridding_buffer - 1); index1b < MIN(index1 + d_regridding_buffer + 1, jlast); index1b++) {
													for(int index0b = MAX(0, index0 - d_regridding_buffer - 1); index0b < MIN(index0 + d_regridding_buffer + 1, ilast); index0b++) {
														int distx = (index0b - index0);
														if (distx < 0) {
															distx++;
														}
														int disty = (index1b - index1);
														if (disty < 0) {
															disty++;
														}
														int distz = (index2b - index2);
														if (distz < 0) {
															distz++;
														}
														distance = 1 + MAX(MAX(abs(distx), abs(disty)), abs(distz));
														if (index0b >= d_ghost_width && index0b < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1b >= d_ghost_width && index1b < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2b >= d_ghost_width && index2b < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
															vectorT(tags,index0b-d_ghost_width,index1b-d_ghost_width,index2b-d_ghost_width) = 1;
														}
														if (vector(regridding_tag,index0b,index1b,index2b) == 0 || vector(regridding_tag,index0b,index1b,index2b) > distance) {
															vector(regridding_tag,index0b,index1b,index2b) = distance;
														}
													}
												}
											}
										}
									}
								}
							}
						}
			
					} else {
						if (d_regridding_type == "SHADOW") {
							if (!initial_time) {
								if (!(vdb->checkVariableExists(d_regridding_field_shadow))) {
									TBOX_ERROR(d_object_name << ": Regridding field selected not found:" <<  d_regridding_field_shadow<<  "");
								}
								int regrid_field_id = vdb->getVariable(d_regridding_field)->getInstanceIdentifier();
								double* regrid_field1 = ((pdat::NodeData<double> *) patch->getPatchData(regrid_field_id).get())->getPointer();
								int regrid_field_shadow_id = vdb->getVariable(d_regridding_field_shadow)->getInstanceIdentifier();
								double* regrid_field2 = ((pdat::NodeData<double> *) patch->getPatchData(regrid_field_shadow_id).get())->getPointer();
								double* regridding_value = ((pdat::NodeData<double> *) patch->getPatchData(d_interior_regridding_value_id).get())->getPointer();
								for(int index2 = 0; index2 < klast; index2++) {
									for(int index1 = 0; index1 < jlast; index1++) {
										for(int index0 = 0; index0 < ilast; index0++) {
					
											double error = 2 * fabs(vector(regrid_field1, index0, index1, index2) - vector(regrid_field2, index0, index1, index2))/fabs(vector(regrid_field1, index0, index1, index2) + vector(regrid_field2, index0, index1, index2));
											vector(regridding_value, index0, index1, index2) = error;
											if (error > d_regridding_error) {
												if (index0 >= d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 >= d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 >= d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
													vectorT(tags,index0 - d_ghost_width, index1 - d_ghost_width, index2 - d_ghost_width) = 1;
												}
												vector(regridding_tag,index0, index1, index2) = 1;
												//SAMRAI tagging
												if (index0 > d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 > d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 > d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
													vectorT(tags, index0-d_ghost_width - 1, index1-d_ghost_width - 1, index2-d_ghost_width - 1) = 1;
												}
												if (index0 >= d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 > d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 > d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
													vectorT(tags, index0-d_ghost_width, index1-d_ghost_width - 1, index2-d_ghost_width - 1) = 1;
												}
												if (index0 > d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 >= d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 > d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
													vectorT(tags, index0-d_ghost_width - 1, index1-d_ghost_width, index2-d_ghost_width - 1) = 1;
												}
												if (index0 >= d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 >= d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 > d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
													vectorT(tags, index0-d_ghost_width, index1-d_ghost_width, index2-d_ghost_width - 1) = 1;
												}
												if (index0 > d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 > d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 >= d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
													vectorT(tags, index0-d_ghost_width - 1, index1-d_ghost_width - 1, index2-d_ghost_width) = 1;
												}
												if (index0 > d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 >= d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 >= d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
													vectorT(tags, index0-d_ghost_width - 1, index1-d_ghost_width, index2-d_ghost_width) = 1;
												}
												if (index0 >= d_ghost_width && index0 < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1 > d_ghost_width && index1 < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2 >= d_ghost_width && index2 < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
													vectorT(tags, index0-d_ghost_width, index1-d_ghost_width - 1, index2-d_ghost_width) = 1;
												}
												//Informative tagging
												if (index0 > 0 && index1 > 0 && index2 > 0) {
													if (vector(regridding_tag,index0-1,index1-1,index2-1) != 1)
														vector(regridding_tag,index0-1,index1-1,index2-1) = 1;
												}
												if (index1 > 0 && index2 > 0) {
													if (vector(regridding_tag,index0,index1-1,index2-1) != 1)
														vector(regridding_tag,index0,index1-1,index2-1) = 1;
												}
												if (index0 > 0 && index2 > 0) {
													if (vector(regridding_tag,index0-1,index1,index2-1) != 1)
														vector(regridding_tag,index0-1,index1,index2-1) = 1;
												}
												if (index2 > 0) {
													if (vector(regridding_tag,index0,index1,index2-1) != 1)
														vector(regridding_tag,index0,index1,index2-1) = 1;
												}
												if (index0 > 0 && index1 > 0) {
													if (vector(regridding_tag,index0-1,index1-1,index2) != 1)
														vector(regridding_tag,index0-1,index1-1,index2) = 1;
												}
												if (index0 > 0) {
													if (vector(regridding_tag,index0-1,index1,index2) != 1)
														vector(regridding_tag,index0-1,index1,index2) = 1;
												}
												if (index1 > 0) {
													if (vector(regridding_tag,index0,index1-1,index2) != 1)
														vector(regridding_tag,index0,index1-1,index2) = 1;
												}
												if (d_regridding_buffer > 0) {
													int distance;
													for(int index2b = MAX(0, index2 - d_regridding_buffer - 1); index2b < MIN(index2 + d_regridding_buffer + 1, klast); index2b++) {
														for(int index1b = MAX(0, index1 - d_regridding_buffer - 1); index1b < MIN(index1 + d_regridding_buffer + 1, jlast); index1b++) {
															for(int index0b = MAX(0, index0 - d_regridding_buffer - 1); index0b < MIN(index0 + d_regridding_buffer + 1, ilast); index0b++) {
																int distx = (index0b - index0);
																if (distx < 0) {
																	distx++;
																}
																int disty = (index1b - index1);
																if (disty < 0) {
																	disty++;
																}
																int distz = (index2b - index2);
																if (distz < 0) {
																	distz++;
																}
																distance = 1 + MAX(MAX(abs(distx), abs(disty)), abs(distz));
																if (index0b >= d_ghost_width && index0b < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1b >= d_ghost_width && index1b < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2b >= d_ghost_width && index2b < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
																	vectorT(tags,index0b-d_ghost_width,index1b-d_ghost_width,index2b-d_ghost_width) = 1;
																}
																if (vector(regridding_tag,index0b,index1b,index2b) == 0 || vector(regridding_tag,index0b,index1b,index2b) > distance) {
																	vector(regridding_tag,index0b,index1b,index2b) = distance;
																}
															}
														}
													}
												}
					
											}
										}
									}
								}
					
							}
				
						}
					}
				}
			}
		}
		//Buffer synchronization if needed
		if (d_regridding_buffer > d_ghost_width) {
			d_tagging_fill->createSchedule(level)->fillData(0, false);
		}
		if (d_regridding_buffer > 0) {
			for (hier::PatchLevel::iterator ip(level->begin()); ip != level->end(); ++ip) {
				const std::shared_ptr< hier::Patch >& patch = *ip;
				int* tags = ((pdat::CellData<int> *) patch->getPatchData(tag_index).get())->getPointer();
				int* regridding_tag = ((pdat::NodeData<int> *) patch->getPatchData(d_nonSync_regridding_tag_id).get())->getPointer();
				const std::shared_ptr<geom::CartesianPatchGeometry > patch_geom(SAMRAI_SHARED_PTR_CAST<geom::CartesianPatchGeometry, hier::PatchGeometry>(patch->getPatchGeometry()));
				const double* dx  = patch_geom->getDx();
				const hier::Index tfirst = patch->getPatchData(tag_index)->getGhostBox().lower();
				const hier::Index tlast  = patch->getPatchData(tag_index)->getGhostBox().upper();
				const hier::Index boxfirst = patch->getBox().lower();
				const hier::Index boxlast  = patch->getBox().upper();
				int ilast = boxlast(0)-boxfirst(0)+2+2*d_ghost_width;
				int itlast = tlast(0)-tfirst(0)+1;
				int jlast = boxlast(1)-boxfirst(1)+2+2*d_ghost_width;
				int jtlast = tlast(1)-tfirst(1)+1;
				int klast = boxlast(2)-boxfirst(2)+2+2*d_ghost_width;
				int ktlast = tlast(2)-tfirst(2)+1;
	
				for(int index2 = 0; index2 < klast; index2++) {
					for(int index1 = 0; index1 < jlast; index1++) {
						for(int index0 = 0; index0 < ilast; index0++) {
	
							int value = vector(regridding_tag, index0, index1, index2);
							if (value > 0 && value < 1 + d_regridding_buffer) {
								int buffer_left = 1 + d_regridding_buffer - value;
								int distance;
								for(int index2b = MAX(0, index2 - buffer_left); index2b < MIN(index2 + buffer_left + 1, klast); index2b++) {
									for(int index1b = MAX(0, index1 - buffer_left); index1b < MIN(index1 + buffer_left + 1, jlast); index1b++) {
										for(int index0b = MAX(0, index0 - buffer_left); index0b < MIN(index0 + buffer_left + 1, ilast); index0b++) {
								
											distance = MAX(MAX(abs(index0b - index0), abs(index1b - index1)), abs(index2b - index2));
											if (distance > 0 && index0b >= d_ghost_width && index0b < (tlast(0)-tfirst(0))+1 + d_ghost_width && index1b >= d_ghost_width && index1b < (tlast(1)-tfirst(1))+1 + d_ghost_width && index2b >= d_ghost_width && index2b < (tlast(2)-tfirst(2))+1 + d_ghost_width) {
												vectorT(tags, index0b - d_ghost_width, index1b - d_ghost_width, index2b - d_ghost_width) = 1;
											}
											if (distance > 0 && (vector(regridding_tag,index0b, index1b, index2b) == 0 || vector(regridding_tag, index0b, index1b, index2b) > value + distance)) {
												vector(regridding_tag, index0b, index1b, index2b) = value + distance;
											}
										}
									}
								}
								
							}
						}
					}
				}
	
			}
		}
	}
}



